/****************************************************************/
/*								*/
/*	WARNING! WARNING! DANGER! DANGER! WARNING! WARNING!	*/
/*								*/
/*	These definitions also have to match the header file	*/
/*	src/alarmvu/alarmString.h in addition to matching	*/
/*	whatever is in choiceGbl.ascii.				*/
/*								*/
/****************************************************************/

/* alarm.h Alarm definitions (Must Match choiceGbl.ascii) */
/* share/epicsH @(#)alarm.h	1.6     6/7/91 */

/*
 *      Original Author: Bob Dalesio
 *      Current Author:  Marty Kraimer
 *      Date:            11-7-90
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  mm-dd-yy        iii     Comment
 * .02  03-24-93	ems	added CMD_NOT_ALLOWED
 *                              added VOLUME_ALARM
 * .03  04-16-93        ems     added LEVEL_ALARM, TEMPERATURE_ALARM
 *                                    PRESSURE_ALARM, BSW_ALARM, TANK_VALVE_ALARM
 * .04  05-07-93	ems	added VALVE_OPEN_ALARM and VALVE_CLOSE_ALARM
 * .05  05-10-93	ems	added VALVE_LIMIT_TORQUE_LAG_ALARM
 * .06  06-04-93	ems	added VIBRATION_ALARM,PUMP_START_ALARM,PUMP_STOP_ALARM
 * .07	06-09-93	ems	added POT_DROP_READY_ALARM
 * .08  09-13-93	ems	added MANUAL_OVERRIDE
 * .09  10-04-93	ems	added SETPT ERROR and CASC MODE ERROR
 * .10  11-19-93	ems	added PARAMETER alarm
 * .11  01-25-94	ems	chg PID_ALM_NOCMD to PID_ALM_TRACKING
 * .12  03-07-94	ems	chg TF_UNEXPECTED to PERMISSIVE_ALM, UNEXPECTED no
 *				longer used
 * .13  03-15-94	ems	added new TANk level and temperature alarms
 * .14  04-15-94	ems	added NO_INIT_ALARM
 * .15	03-07-95	ems	chg BAD_SUB_ALARM to BAD_UDB_ALARM
 * .16	08-07-95	wep	chg COMM_ALARM to OPENTC_ALARM
 * .17	04-01-96	ems	added typedef alarm_limits and defines for
 *				dbCommon.h LARF last alarm flag
 * .18	04-28-96	ems	added PID_ CLAMP alarm defs back for use by
 *				other programs
 */

#ifndef INCalarmh
#define INCalarmh 1

/* Maximum number of values to be used in a ROC alarm */

#define	MAX_ROC_VALUES	60

/* scan enable */

#define	ON_SCAN		1
#define	OFF_SCAN	0

/* alarm enabled */

#define	ALARM_ENABLED	1
#define	ALARM_DISABLED	0

/* tagged point */

#define	TAGGED		2

/* define used by recBi.c */

#define	NORMAL		0
#define	USE_RAW_VALUE	2

/* defines for the choice fields */
/* ALARM SEVERITIES - NOTE: must match defs in choiceGbl.ascii GBL_ALARM_SEV */

#define NO_ALARM		0x0
#define INFORM			0x1
#define INFO_ALARM		INFORM
#define	ALM_EVENT		0x2
#define EVENT_ALARM		ALM_EVENT
#define	MINOR_ALARM		0x3
#define	MAJOR_ALARM		0x4
#define VALID_ALARM		0x5
#define ALARM_NSEV		VALID_ALARM+1

/* ALARM STATUS  -NOTE: 
	must match defs in choiceGbl.ascii GBL_ALARM_STATi,
	epicsH/alarmStrings.h and epicsH/decTalkalarmStrings.h
*/
/* NO_ALARM = 0 as above */

#define	READ_ALARM		1
#define	WRITE_ALARM		2

/* ANALOG ALARMS */

#define	HIHI_ALARM		3
#define	HIGH_ALARM		4
#define	LOLO_ALARM		5
#define	LOW_ALARM		6
#define SBAD_ALARM		7
#define SIGNAL_ALARM		SBAD_ALARM
#define ROC_ALARM		8

/* BINARY ALARMS */

#define	STATE_ALARM		9
#define	COS_ALARM		10

/* other alarms */

#define OPENTC_ALARM		11
#define	TIMEOUT_ALARM		12
#define	HW_LIMIT_ALARM		13
#define LIMIT_ALARM		HW_LIMIT_ALARM
#define	CALC_ALARM		14
#define	SCAN_ALARM		15
#define	LINK_ALARM		16
#define	SOFT_ALARM		17
#define	BAD_UDB_ALARM		18
#define	UDF_ALARM		19
#define INHIBIT_ALARM		20
#define MANUAL_ALARM		21
#define TAGGED_ALARM		MANUAL_ALARM
#define CMD_NOT_ALLOWED		22
#define VOLUME	        	23
#define HIHI_LEVEL              24
#define TEMPERATURE	        25
#define PRESSURE		26
#define BSW	        	27
#define TANK_VALVE      	28
#define VALVE_OPEN_ALM  	29
#define VALVE_CLOSE_ALM  	30
#define VALVE_TORQUE_LAG	31
#define VALVE_FAULT_ALM         32
#define VALVE_POSITION	        33
#define AUTOMATIC		34
#define VIBRATION		35
#define PUMP_START	        36
#define PUMP_STOP		37
#define POT_DROP_READY	        38
#define PERMISSIVE_ALM          39
#define TF_UNCOMMANDED          40
#define THRESHOLD_EXCEEDED	41
#define	PID_ALM_NOMV		42
#define	PID_ALM_NOOUT		43
#define DEVIATION_ALARM		44
#define	PID_ALM_DEV		DEVIATION_ALARM		/* ems done for compatability */
#define	PID_ALM_TRACKING	45
#define	HIGH_SETPT_CLAMP	46
#define	PID_ALM_HIGH_STP_CLAMP	HIGH_SETPT_CLAMP	/* ems done for compatability */
#define	LOW_SETPT_CLAMP		47
#define	PID_ALM_LOW_STP_CLAMP	LOW_SETPT_CLAMP		/* ems done for compatability */
#define	HIGH_OUTPUT_CLAMP	48
#define	PID_ALM_HIGH_OUT_CLAMP	HIGH_OUTPUT_CLAMP	/* ems done for compatability */
#define	LOW_OUTPUT_CLAMP	49
#define	PID_ALM_LOW_OUT_CLAMP	LOW_OUTPUT_CLAMP	/* ems done for compatability */
#define MANUAL_OVERRIDE		50
#define SETPT_ERROR             51
#define OUT_MODE_ERROR		52
#define PARAMETER		53
#define HIGH_LEVEL		54
#define LOW_LEVEL		55
#define LOLO_LEVEL		56
#define HIHI_TEMPERATURE	57
#define HIGH_TEMPERATURE	58
#define LOW_TEMPERATURE		59
#define LOLO_TEMPERATURE	60
#define NO_INIT_ALARM		61
#define ALARM_NSTATUS		(NO_INIT_ALARM + 1)

#endif


/* this bit of sillyness is needed because dbCommon.ascii includes
   alarm.h and the build can not handle struct defs, so 
   INCalarm_limitsh is defined in dbCommon.ascii
*/
#ifndef INCalarm_limitsh
#define INCalarm_limitsh 1

/* used by those blocks that have alarm limits */

typedef struct alarm_limits
{
    float		hihi;		/* hihi alarm value */
    float		high;		/* high alarm value */
    float		low;		/* low alarm value */
    float		lolo;		/* lolo alarm value */
    float		devl;		/* deviation limit alarm value */
    float		devs;		/* deviation setpoint */
    double		hyst;		/* hysteresis alarm deadband */
    unsigned short	hhsv;		/* hihi alarm severity */
    unsigned short	hsv;		/* high alarm severity */
    unsigned short	lsv;		/* low alarm severity */
    unsigned short	llsv;		/* lolo alarm severity */
    unsigned short	dvsv;		/* deviation alarm severity */
}ALARM_LIMITS;

/* define bits for different alarm limits for dbCommon/LARF (last alarm flag) */

#define	LARF_NO_ALARM		0x00000000
#define	LARF_HIHI		0x00000001
#define	LARF_HIGH		0x00000002
#define	LARF_LOW		0x00000004
#define	LARF_LOLO		0x00000008
#define	LARF_DEV		0x00000010
#define	LARF_SBAD_LOW		0x00000020
#define	LARF_SBAD_HIGH		0x00000040
#define	LARF_OUTPUT_HIGH	0x00000080
#define	LARF_OUTPUT_LOW		0x00000100
#define	LARF_SETPT_HIGH		0x00000200
#define	LARF_SETPT_LOW		0x00000400
#define	ALDL_MAX		32760		/* max alarm delay count */

#ifdef vxWorks
extern void AlarmLimits (ALARM_LIMITS *, double, void *);
extern void AlarmStates (void *, int);
extern int  AlarmDisableLink (void *);
#endif

#endif
