/*channelControl.h,v 1.2 1994/04/21 12:47:44 uchenick Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		CHANNEL CONTROL STRUCTURES			*/
/*								*/
/*	These structures are used by Channel Acquisition and	*/
/*	Control to keep track of all the channels that are	*/
/*	known to the workstation. "Oh what a tangled web we	*/
/*	weave...."						*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifndef	channelcontrolH
#define	channelcontrolH

#ifdef AXP
#ifndef POINTER64
#define POINTER64
#endif
#ifndef Pointer64fromToken
#define Pointer64fromToken(tok)	TokenPointers[tok-500000]
extern  int   TokenFromPointer64 (void *);
extern	void *TokenPointers[];
#endif
#endif

#ifndef WIN32
#    ifndef SOCKET
#	define SOCKET int
#    endif SOCKET
#    ifndef INVALID_SOCKET
#	define	INVALID_SOCKET	(-1)
#    endif INVALID_SOCKET
#endif WIN32

/****************************************************************/
/*								*/
/*		IOC STATUS BLOCK (ISB)				*/
/*								*/
/*	This structure is used to tell us if the IOC is up	*/
/*	and also to point to a linked list of the channels	*/
/*	that are associated with that IOC.			*/
/*								*/
/****************************************************************/

#include <dbDefs.h>
#include <tsDefs.h>

typedef struct IOCSB
    {
    struct in_addr	Inet;		/* Internet address */
    struct
	{
	int		secPastEpoch;
	int		nsec;
	}		lastAlarm;	/* Last alarm from this IOC */
    struct CHDR		*pFirstCHB;	/* Pointer to first CHB */
    struct IOCSB	*nextISB;	/* Pointer to next ISB */
    int			connected;	/* Connected flag */
    int			splicable;	/* Can channel splicing be used? */
#ifdef POINTER64
    int			isbToken;	/* 32 bit ID value replacing ptr */
#endif
    } ISB;


/****************************************************************/
/*								*/
/*		CHANNEL HEADER BLOCK (CHB)			*/
/*								*/
/*	There is one Channel Header Block for each point,	*/
/*	regardless of the number of applications that have	*/
/*	channels open to that point. It contains the current	*/
/*	status as well as a lot of links for status.		*/
/*								*/
/****************************************************************/

typedef struct CHDR
    {
    char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
					/* Channel Name */
    ISB			*pISB1;		/* Pointer to first ISB */
    ISB			*pISB2;		/* Pointer to second ISB */
    ISB			*pCurrISB;	/* Pointer to current ISB */
    int			handleIOC1;	/* Handle for IOC 1 */
    int			handleIOC2;	/* Handle for IOC 2 */
    short		currType;	/* Current data type */
    short		currMask;	/* Current event mask */
    short		nativeType;	/* Native data type, old way */
    short		nativeCount;	/* Native element count */
    unsigned short	checksum;	/* Record CRC from PVMAP */
    short		pointType;	/* Record type from PVMAP */
    unsigned int	creationTime;	/* Record creation time from PVMAP */
    int			currCount;	/* Current element count */
    struct CHAB		*pFirstCAB;	/* First CAB, 0 -> delete pending */
    struct CHDR		*pNextIOC1;	/* Next CHB, IOC 1 */
    struct CHDR		*pPrevIOC1;	/* Previous CHB, IOC 1 */
    struct CHDR		*pNextIOC2;	/* Next CHB, IOC 2 */
    struct CHDR		*pPrevIOC2;	/* Previous CHB, IOC 2 */
    struct CHDR		*pNextRes;	/* Next CHB, Resource List */
    struct CHDR		*pPrevRes;	/* Previous CHB, Resource List */
    struct RECB		*pRes;		/* RCB for this point */
    struct CHDR		*pNextHash;	/* Next CHB, Hash Table Order */
    struct HSHTB	*pHash;		/* Hash table row addr */
#ifdef POINTER64
    int			chbToken;	/* 32 bit ID value replacing ptr */
#endif
    } CHB;


/****************************************************************/
/*								*/
/*		CHANNEL APPLICATION BLOCK (CAB)			*/
/*								*/
/*	There is one Channel Application Block for each		*/
/*	application that is connected to a channel. It holds	*/
/*	all of the stuff that we need to send data to that	*/
/*	application as well as links for bookkeeping.		*/
/*								*/
/****************************************************************/

typedef struct CHAB
    {
    CHB			*pCHB;		/* Pointer back to CHB */
    struct APCB		*pACB;		/* Pointer back to ACB */
    int			handle;		/* User's handle */
    short		mask;		/* Event mask */
    short		dataType;	/* Data type */
    short		count;		/* Array size */
    short		readType;	/* Data type for pending directed read
					   AGNT_NOCONN if no read pending. */
    short		awaitAck;	/* Application awaiting handle exchange */
    short		awaitMon;	/* Application awaiting initial monitor */
    struct CHAB		*pNextChan;	/* Next CAB, this channel */
    struct CHAB		*pPrevApp;	/* Prev CAB, this application */
    struct CHAB		*pNextApp;	/* Next CAB, this application */
#ifdef POINTER64
    int			cabToken;	/* 32 bit ID value replacing ptr */
#endif
    } CAB;


/****************************************************************/
/*								*/
/*		APPLICATION CONTROL BLOCK (ACB)			*/
/*								*/
/*	There is one Application Control Block for each		*/
/*	application program. It is a header that points to	*/
/*	a double linked list of CABs that are all the points	*/
/*	used by this application.				*/
/*								*/
/****************************************************************/

typedef struct APCB
    {
    SOCKET		sockFd;		/* Fd for socket I/O, 0 -> delete pending */
    int			appRxSize;	/* Receive buffer size for sockFd */
    int			appTxSize;	/* Transmit buffer size for sockFd */
    int			appRxTopped;	/* Receive buffer maxout flag for sockFd */
    CAB			*pFirstCAB;	/* First CAB, this app */
    struct APCB		*pNextACB;	/* Next ACB */
    short		trendHandler;	/* Is app a trend handler? */
    short		alarmHandler;	/* Is app an alarm handler? */
    int			buffSize;	/* Number bytes in txBuffer */
    int			nchib;		/* Number bytes in rxBuffer */
    int			rcvOutPos;	/* Index of next byte to be extracted from rx buffer */
    char		appBuff[5120];	/* The transmit buffer */
    char		appRxBuff[2048];	/* The receive buffer */
#ifdef POINTER64
    int			acbToken;	/* 32 bit ID value replacing ptr */
#endif
    } ACB;


/****************************************************************/
/*								*/
/*		RESOURCE CONTROL BLOCK (RCB)			*/
/*								*/
/*	There is one Resource Control Block for each resource	*/
/*	that is known to the system. We use this to keep track	*/
/*	of a chain of CHBs so that we know all the points that	*/
/*	are open for a given resource. When a resource changes	*/
/*	state, we should generate disconnect events for all	*/
/*	the points that are on that resource.			*/
/*								*/
/****************************************************************/

typedef struct RECB
    {
    char		*resName;	/* Resource Name */
    ISB			*pISB1;		/* Pointer to first ISB */
    ISB			*pISB2;		/* Pointer to second ISB */
    ISB			*currISB;	/* Current master ISB, 0 if none */
    CHB			*pFirstCHB;	/* First CHB, this resource */
    struct RECB		*pNextRCB;	/* Next RCB */
    struct NSQUEUE	*pFirstParam;	/* List of outstanding param requests */
    struct NSQUEUE	*pFirstAck;	/* List of outstanding alarm acks */
    struct NSQUEUE	*pFirstCtl;	/* List of outstanding alarm enables/disables */
    } RCB;


/****************************************************************/
/*								*/
/*			HASH TABLE (HTB)			*/
/*								*/
/*	All point names are in a hash table that has the	*/
/*	following layout.					*/
/*								*/
/****************************************************************/

typedef struct HSHTB
    {
    CHB			*pFirstCHB;	/* First CHB, this hash */
    } HTB;


/****************************************************************/
/*								*/
/*			NAME SERVER QUEUE			*/
/*								*/
/*	Channel Acquisition and Control keeps a queue of	*/
/*	the names that are pending resolution. Each of		*/
/*	these entries has enough data to complete the request	*/
/*	as well as a pointer back to the Application Control	*/
/*	block that originated the request. This is so that	*/
/*	we can purge the queue of pending entries if the	*/
/*	application dies.					*/
/*								*/
/****************************************************************/

typedef struct NSQUEUE
    {
    int			handle;		/* User's Handle */
    int			dueTime;	/* Timeout (vms & vxWorks only) */
    struct
	{
	int		secPastEpoch;
	int		nsec;
	}		ackTime;	/* For alarm acknowledgment */
    short		type;		/* Data Type */
    short		count;		/* Element Count */
    ACB			*pACB;		/* Pointer to ACB */
    struct NSQUEUE	*pNextNSQ;	/* Link to next one */
    struct NSQUEUE	*nameChain;	/* Other req's, same name */
    char		name[PVNAME_SZ + FLDNAME_SZ + 2];
#ifdef POINTER64
    int			nsqToken;	/* 32 bit ID value replacing ptr */
#endif
    } NSQ;


/****************************************************************/
/*								*/
/*		FUNCTION PROTOTYPES FOR BUFFER MGT		*/
/*								*/
/*	The following routines are provided to manage the	*/
/*	control blocks that are acquired and released.		*/
/*	Routines are not provided if control blocks are only	*/
/*	acquired and never released.				*/
/*								*/
/****************************************************************/

CHB	*newCHB(void);		/* Allocate a Channel Header Block */
void	deleteCHB(CHB *oldChb);	/* Free a Channel Header Block */
CAB	*newCAB(void);		/* Allocate a Channel Application Block */
void	deleteCAB(CAB *oldCab);	/* Free a Channel Application Block */
ACB	*newACB(void);		/* Allocate an Application Control Block */
void	deleteACB(ACB *oldAcb);	/* Free an Application Control Block */
NSQ	*newNSQ(void);		/* Allocate a Name Server Queue element */
void	deleteNSQ(NSQ *oldNsq);	/* Free a Name Server Queue element */

#endif	channelcontrolH
