/*dataTypes.h,v 1.1.1.1 1999/08/03 14:41:21 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	DATA TYPE DEFINITIONS FOR CHANNEL ACCESS AGENT		*/
/*								*/
/*	Because the data type for a channel can be promoted	*/
/*	or demoted at any time, a new designation for data	*/
/*	types is required. A data type is now a 16 bit number.	*/
/*	The low order 8 bits is the elemental type (i.e, char,	*/
/*	short, float, etc.). The upper 8 bits consist of the	*/
/*	embellishments to that type (i.e., status, time,	*/
/*	graphics, etc.). We are also providing a vector that	*/
/*	can translate between the old epics DBR types and	*/
/*	our new list.						*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*			ELEMENTAL DATA TYPES			*/
/*								*/
/****************************************************************/

#ifndef INCLdataTypes
#define INCLdataTypes
#define AGNT_NOCONN	-1
#define AGNT_STRING	0
#define AGNT_INT	1
#define	AGNT_SHORT	1
#define AGNT_FLOAT	2
#define AGNT_ENUM	3
#define AGNT_CHAR	4
#define AGNT_LONG	5
#define AGNT_DOUBLE	6
#define AGNT_ADAPTIVE	7

/*
AGNT_ADAPTIVE is new with the Tate implementation. This means
to open a channel to the native type of the field being
referenced. Also, the remote server task in the IOC is also
to generate a read for the control variant of the native data
type  (see the macro dbf_type_to_DBR_CTRL in db_access.h) and
also generate a monitor for the status variant of the native
data type (see the macro dbf_type_to_DBR_STS in db_access.h).
The purpose of this new data type is to reduce the number of
round trips that are required between the workstation and the
IOC to get graphics to appear on the screen.

Note that the remote server task never sees the AGNT_ADAPTIVE
data type. chanAcq will send him a CHAN_BLDRDMON message instead
of the normal CHAN_BUILD message.
*/

#define AGNT_TYPE_MASK	0xFF


/****************************************************************/
/*								*/
/*			DATA TYPE EMBELLISHMENTS		*/
/*								*/
/****************************************************************/

#define	AGNT_STAT	0x100	/* Status and Severity */
#define AGNT_TIME	0x200	/* Time Stamp */
#define AGNT_GRAPH	0x400	/* Graphics Fields */
#define AGNT_CNTRL	0x800	/* Control Information */

/******

The following table gives the meaning of all the fields that 
make up each embellishment. Strings are different than the other 
types because of enhancements that were made to optimize alarm 
management. In general, the embellishment fields are the same
data type as the elemental type except for .EGU, .DESC, .ACK,
and AENB which are always strings. The .STAT and .SEVR fields
are always shorts, and time stamps are in the EPICS format.



ELEMENTAL	-----------------------EMBELLISHMENT-----------------------
  TYPE		AGNT_STAT	AGNT_TIME	AGNT_GRAPH	AGNT_CNTRL
		
AGNT_STRING	.STAT		stamp		not		not
		.SEVR				implemented	implemented
		.AENB
		.ACK
		.DESC
		.EGU

AGNT_INT	.STAT		stamp		.EGU		.DRVH
		.SEVR				.HOPR		.DRVL
						.LOPR
						.HIHI
						.HIGH
						.LOW
						.LOLO

AGNT_SHORT	.STAT		stamp		.EGU		.DRVH
		.SEVR				.HOPR		.DRVL
						.LOPR
						.HIHI
						.HIGH
						.LOW
						.LOLO

AGNT_FLOAT	.STAT		stamp		.EGU		.DRVH
		.SEVR				.HOPR		.DRVL
						.LOPR
						.HIHI
						.HIGH
						.LOW
						.LOLO
						.PREC

AGNT_ENUM	.STAT		stamp		# Strings	# Strings
		.SEVR				strings		strings

AGNT_CHAR	.STAT		stamp		.EGU		.DRVH
		.SEVR				.HOPR		.DRVL
						.LOPR
						.HIHI
						.HIGH
						.LOW
						.LOLO

AGNT_LONG	.STAT		stamp		.EGU		.DRVH
		.SEVR				.HOPR		.DRVL
						.LOPR
						.HIHI
						.LOW
						.LOLO

AGNT_DOUBLE	.STAT		stamp		.EGU		.DRVH
		.SEVR				.HOPR		.DRVL
						.LOPR
						.HIHI
						.LOW
						.LOLO
						.PREC
******/


/******
	The following table translates between the DBR types
	as defined in db_access.h to the newer agent types.
	See dataTypes.c
******/

extern short dbrXlate[35];


/****************************************************************/
/*								*/
/*		EVENT MASKS FOR MONITORS			*/
/*	NOTE: the 3 have the same value as the DBE versions in	*/
/*	      dbAcccess.h, DO NOT CHANGE THESE			*/
/*								*/
/****************************************************************/

#define	AGNT_VALUE	0x01
#define AGNT_LOG	0x02
#define	AGNT_ALARM	0x04


/****************************************************************/
/*								*/
/*		MASKS FOR CHANNEL SPLICING			*/
/*								*/
/*	These masks are applied to the MSB of the event mask	*/
/*								*/
/****************************************************************/

#define AGNT_SPL_STAT	0x100
#define AGNT_SPL_SEVR	0x200
#define AGNT_SPL_ACK	0x400
#define AGNT_SPL_VAL	0x800
#define AGNT_SPLICED	(AGNT_SPL_ACK | AGNT_SPL_SEVR | AGNT_SPL_STAT)
#endif
