/* dbDefs.h			*/
/* share/epicsH @(#)dbDefs.h	1.9     7/31/91 */

/*
 *      Author:          Marty Kraimer
 *      Date:            6-1-90
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  mm-dd-yy        iii     Comment
 */

#ifndef INCdbDefsh
#define INCdbDefsh 1

#ifndef NULL
#define NULL 0
#endif

#ifndef NO
#define NO	0
#define YES	1
#endif

#ifndef FALSE
#define FALSE	0
#define	TRUE	1
#endif

/* define the default path for the default.dctsdr */

#ifdef WIN32
#define	TIS_DATABASE_PATH	"\\tis4000db\\"
#else
#define	TIS_DATABASE_PATH	"/tis4000db/"
#endif WIN32
#define	TIS_SDR_NAME		"default.dctsdr"

/* PVNAME_SZ MUST be a multiple of 4. FLDNAME_SZ MUST be 4 */
#define	PVNAME_SZ	28	/*Process Variable Name Size	*/
#define	FLDNAME_SZ	4	/*Field Name Size		*/

/* defines for various database fields */

#define	ALARMINFO_DISPLAY_SZ  250 /* alarminfo display size, MUST agree with
				     dataVu/alarminfo.h */
#define	ALARMINFO_GUIDANCE_SZ 240 /* alarm info guidance size, MUST agree with
				     dataVu/alarminfo.h */
#define	API_STRAP_NAME_SZ 30	/* api,strapping table name size, MUST agree with
				   dataVu/apitbl.h and straptbl.h */
#define	ARCHIVE_METH_SZ	3	/* archive method size, MUST agree with dataVu/ar_reduc.h */
#define	ARCHIVE_RATE_SZ	5	/* archive rate size, MUST agree with dataVu/ar_reduc.h */
#define	ALARMGRP_SZ	16	/* alarm group size, MUST agree with dbCommon.ascii */
#define	CALCEXPR_SZ	72	/* size of the calc expression */
#define	DESCRIPTOR_SZ	41	/* descriptor field size, MUST agree with 
				   dbCommon.ascii */
#define	DESCRIPTION_SZ 30	/* desc fld in tisusers.dbf,tisgrps.dbf,protgrps.dbf */
#define	DIGITALSTATE_SZ	20	/* size of state fields in bi/bo, MUST agree with
				   <bi,bo>Record.ascii and dataVu/<di,do>.h */
#define	DEVICETYPE_SZ	23	/* device type size, MUST agree with pvmap.h
				   in dataVu */
#define	DPA_DATA_SZ	200	/* MUST agree with dataVu/dpacfg.h */
#define	DPA_STRING_SZ	32	/* MUST agree with dataVu/dpacfg.h */
#define	EGUTAG_SZ	16	/* engineering units tag, used in various point types */
#define	FDPROTFLD_SZ	12	/* fieldname in dataVu/fdlprott.h */
#define	IOCNAME_SZ	16	/* ioc name, MUST agree with value in dataVu/iocList.h */
#define	DBC_IOCNAMESIZE	IOCNAME_SZ /* used is some source, include it so dont have to chg
				      the source */
#define	LOCATION_SZ	16	/* location field size, MUST agree with pvmap.h 
				   in dataVu */
#define MULTIDIGITALSTATE_SZ 20	/* mbbi, mbbo, descriptor (state) size, MUST agree
				   with <mbbi,mbbo>.h in dataVu */
/* operator name & group are kept in the same fields of some databases, */
/* so they must be <= size of those fields */
#define	OPERATOR_SZ	30	/* operator id size, MUST agree with op_trak.h and */
#define	GROUPNAME_SZ	30	/* privtab.dbf, sysusers.dbf,tisgrps.dbf,utilacc.dbf */
#define	ENCRYPTPASSWD_SZ 30	/* tisusers.dbf */
#define	UNENCRYPTPASSWD_SZ 14	/* unecrypted password size */
#define	POINTTYPE_SZ	15	/* point type size, MUST agree with pvamp.h and fldprott.h
				   in dataVu */
#define	PROTECTIONGRP_SZ 16	/* protectiongroup size, MUST agree with 
				   dataVu/pvamp.h */
#define	PRIV_LEVEL_SZ	2	/* privilege level size, MUST agree with privtab.h and fldprott.h
				   in dataVu */
#define	RESOURCE_SZ	16	/* resource name size, MUST agree with 
				   dataVu/pvmap.h */
#define	SCRIPTID_SZ	28	/* script id name, MUST agree with dataVu/sript.h */
#define	SCRIPTPROC_SZ	60	/* script process size, MUST agree with dataVu/script.h */
#define	SCRIPTDESC_SZ	50	/* script description size, MUST agree with dataVu/script.h */
#define	SCRIPTPARAM_SZ	25	/* script parameter size, MUST agree with dataVu/script.h */
#define	STATEVAL_SZ	80	/* Val size in state block, MUST agree with dataVu/state.h */
#define	STRNGVAL_SZ	80	/* MUST agree with dataVu/strng<in,out>.h */
#define	TFCMDSTS_SZ	18	/* tankfarm command and status strings, see various
				   tankfarm .h files in dataVu */
#define	UDBPROGRAMNAME_SZ 16	/* Udb program name, MUST agree with dataVu/udb.h */
#define	UDBSTRINGS_SZ	16	/* string sizes in UDB, MUST agree with dataVu/udb.h */
#define	UDBLABEL_SZ	24	/* MUST agree with dataVu/udbcfg.h */
#define	UTILACCPRIV_SZ	45	/* utility privilege fld size, utilacc.dbf */


/* defines for hardware address sizes */
#define	IOLINE_SZ	4
#define	IOCARD_SZ	4
#define	IOADDR_SZ	4
#define	IOSIGNAL_SZ	5

#define SUPERVISORY 0
#define CLOSED_LOOP 1

#define NTO1FIRST	0
#define NTO1LOW		1
#define NTO1HIGH	2
#define NTO1AVE		3

#define vxTicksPerSecond 60	/*clock ticks per second*/

#define priorityLow	0
#define priorityMedium	1
#define priorityHigh	2
#endif INCdbDefsh
