/* dbRecords.h	Record Location		*/
/* share/epicsH @(#)dbRecords.h	1.4     6/7/91 */

/*
 *      Original Author: Bob Dalesio
 *      Current Author:  Marty Kraimer
 *      Date:            6-1-90
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  mm-dd-yy        iii     Comment
 */

#ifndef INCdbRecordsh
#define INCdbRecordsh 1

struct recLoc{	/* record location*/
	long	no_records;	/*number of records of this type*/
	long	no_used;	/*number actually used		*/
	long	rec_size;	/*record size in bytes		*/
	long	record_type;	/*record type			*/
	char	*pFirst;	/*ptr to struct recChain for first record */
};

struct recChain{ /* individual record */
	struct recChain	*next;	/* Pointer to next record */
	char	*pRecord;	/* addr of the record */
	};

struct recHeader{ /*record header*/
	long		number;		/*number of record types*/
	struct recLoc	**papRecLoc;	/*ptr to arr of ptr to recLoc*/
};
extern struct recHeader *dbRecords;

#define S_record_noRecords   (M_record| 1) /*No records */

/***********************************************************************
 * Except that some ptr could be null the following is a valid reference
 *
 * dbRecords->papRecLoc[i]->pFirst->pRecord
 *
 * The memory layout is:
 *
 * dbRecords->recHeader
 *		 number
 *		 papRecLoc->[1]
 *			    [2]->recLoc
 *				    no_records
 *				    no_used
 *				    rec_size
 *				    pFirst----->next------>next (etc.)
 *						pRecord    pRecord
 *						  |	      |
 *						  V	      V
 *					      db Record    db Record
 ************************************************************************/

/************************************************************************
 * In the self defining record the following appears
 *
 *	sdrHeader
 *	recLoc
 *	record 1
 *	...
 *	record n (n = recLoc.no_records
 **************************************************************************/

/************************************************************************
 * The following macro returns NULL or a ptr to struct recLoc
 * A typical usage is:
 *
 *	struct recLoc *precLoc
 *	if(!(precLoc=GET_PRECLOC(type))) {<not found>}
 ***********************************************************************/
#define GET_PRECLOC(REC_TYPE)\
(\
    dbRecords\
    ?(\
	( ((REC_TYPE)<1) || ((REC_TYPE)>=dbRecords->number) )\
	? NULL\
	: (dbRecords->papRecLoc[(REC_TYPE)])\
    )\
    : NULL\
)

/************************************************************************
 * The following macro returns NULL or a ptr to a record
 *
 * A typical usage is:
 *
 *	char * precord;
 *	if(!(precord=GET_PRECORD(precLoc,recnum))) {<not found>}
 ************************************************************************/

/**************************************************/
/*						  */
/* This macro was replaced with a function. See	  */
/* the file getPrecord.c (GMU)			  */
/*						  */
/**************************************************/

/****************************************************************************************/
/*											*/
/*#define GET_PRECORD(PRECLOC,REC_NUM)\							*/
/*(\											*/
/*    (PRECLOC)\									*/
/*    ?(\										*/
/*	( ((REC_NUM)<0) || ((REC_NUM)>=((struct recLoc *)(PRECLOC))->no_records) )\	*/
/*	? NULL\										*/
/*	:(\										*/
/*	    ((struct recLoc *)(PRECLOC))->pFirst\					*/
/*		+ ((REC_NUM)*((struct recLoc *)(PRECLOC))->rec_size) \			*/
/*	)\										*/
/*    )\										*/
/*    : NULL\										*/
/*)											*/
/*											*/
/****************************************************************************************/
#endif
