/* devSup.h	Device Support		*/
/* share/epicsH @(#)devSup.h	1.4     6/7/91 */

/*
 *      Author:          Marty Kraimer
 *      Date:            6-1-90
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  mm-dd-yy        iii     Comment
 * .02  03-29-94	ems	GET_PDSET no longer explicitly checks for 
 *				< 0, since it is unsigned
 */

#ifndef INCdevSuph
#define INCdevSuph 1

typedef long (*DEVSUPFUN) ();	/* ptr to device support function*/

struct dset {	/* device support entry table */
	long		number;		/*number of support routines*/
	DEVSUPFUN	report;		/*print report*/
	DEVSUPFUN	init;		/*init support*/
	DEVSUPFUN	init_record;	/*init support for particular record*/
	DEVSUPFUN	get_ioint_info;	/* get io interrupt information*/
	/*other functions are record dependent*/
	};

struct devSup {
	long		number;		/*number of dset		*/
	char		**dsetName;	/*ptr of arr of ptr to name	*/
	struct dset	**papDset;	/*ptr to arr of ptr to dset	*/
	};

struct recDevSup {
	long		number;		/*number of record types	*/
	struct devSup	**papDevSup;	/*ptr to arr of ptr to devSup	*/
	};

extern struct recDevSup *devSup;

#define S_dev_noDevSup   (M_devSup| 1) /*SDR_DEVSUP: Device support missing*/
#define S_dev_noDSET     (M_devSup| 3) /*Missing device support entry table*/
#define S_dev_missingSup (M_devSup| 5) /*Missing device support routine*/
#define S_dev_badInpType (M_devSup| 7) /*Bad INP link type*/
#define S_dev_badOutType (M_devSup| 9) /*Bad OUT link type*/

/***************************************************************************
 * except that some fields could be null the following is a valid reference
 *
 * devSup->papDevSup[i]->papDset[j]->report(<args>);
 *
 * The memory layout is
 *
 * devSup->recDevSup
 *            number
 *            papDevSup->[1]
 *                       [2]->devSup
 *                               number
 *                               dsetName->[0]->"name"
 *                               papDset-->[0]
 *                                         [1]->dset
 *                                                 number
 *                                                 report->report()
 *                                                  ...
 ***************************************************************************/

/***************************************************************************
 * The following macro returns NULL or ptr to struct devSup
 * A typical usage is:
 *
 *	struct devSup *pdevSup;
 *	if(!(pdevSup=GET_DEVSUP(rectype))) {<null>}
 **************************************************************************/
#define GET_DEVSUP(REC_TYPE)\
(\
    devSup\
    ?(\
	( ((REC_TYPE)<1) || ((REC_TYPE)>=devSup->number) )\
	? NULL\
	: (devSup->papDevSup[(REC_TYPE)])\
    )\
    : NULL\
)


/***************************************************************************
 * The following macro returns NULL of ptr to struct dset
 * A typical usage is:
 *
 *	struct dset *pdset;
 *	if(!(pdset=GET_PDSET(pdevSup,dtyp))) {<null>}
 **************************************************************************/
/* original, emsteel
#define GET_PDSET(PDEVSUP,DTYP)\
(\
    (PDEVSUP)\
    ?(\
	( (DTYP)<0 || ((DTYP)>=((struct devSup*)(PDEVSUP))->number) )\
	? NULL\
	: ( ((struct devSup*)(PDEVSUP))->papDset[(DTYP)] )\
    )\
    : NULL\
)
*/

#define GET_PDSET(PDEVSUP,DTYP)\
(\
    (PDEVSUP)\
    ?(\
	((unsigned)(DTYP)>=((struct devSup*)(PDEVSUP))->number) \
	? NULL\
	: ( ((struct devSup*)(PDEVSUP))->papDset[(DTYP)] )\
    )\
    : NULL\
)
#endif
