/* drvSup.h	Driver Support	*/
/* share/epicsH @(#)drvSup.h	1.4     6/7/91 */

/*
 *      Author:          Marty Kraimer
 *      Date:            6-1-90
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  mm-dd-yy        iii     Comment
 */

#ifndef INCdrvSuph
#define INCdrvSuph 1

typedef long (*DRVSUPFUN) ();	/* ptr to driver support function*/

struct drvet {	/* driver entry table */
	long		number;		/*number of support routines*/
	DRVSUPFUN	report;		/*print report*/
	DRVSUPFUN	init;		/*init support*/
	/*other functions are device dependent*/
	};

struct drvSup {
	long		number;		/*number of dset		*/
	char		**drvetName;	/*ptr to arr of ptr to drvetName*/
	struct drvet	**papDrvet;	/*ptr to arr ptr to drvet	*/
	};
#define DRVETNUMBER ( (sizeof(struct drvet) -sizeof(long))/sizeof(DRVSUPFUN) )
extern struct drvSup *drvSup;

#define S_drv_noDrvSup   (M_drvSup| 1) /*SDR_DRVSUP: Driver support missing*/
#define S_drv_noDrvet    (M_drvSup| 3) /*Missing driver support entry table*/

/**************************************************************************
 *
 *The following are valid references
 *
 * drvSup->drvetName[i]  This is address of name string
 * drvSup->papDrvet[i]->report(<args>);
 *
 * The memory layout is
 *
 * drvSup->drvSup
 *            number
 *            drvetName->[0]->"name"
 *            papDrvet-->[0]
 *                       [1]->drvet
 *                               number
 *                               report->report()
 *                                ...
 *************************************************************************/
#endif
