#ifndef INC_guiSubr_h
#define INC_guiSubr_h
/*	@(#)guiSubr.h	1.7 3/15/92
 *	Author:	Roger A. Cole
 *	Date:	02-08-91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991-92, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 *  .00 02-08-91 rac	initial version
 *  .01 07-30-91 rac	installed in SCCS
 *  .02 08-14-91 rac	added guiNoticeName
 *  .03 09-06-91 rac	added guiCheckbox..., guiGetNameFromSeln,
 *			guiIconCreate
 *  .04 11-18-91 rac	added guiInit, guiFileSelect, guiShellCmd,
 *			guiTextswXxx
 *  .05 02-26-92 rac	added guiEditor, guiPrinter, guiTextswPuts
 *  .06 03-15-92 rac    add guiLock as a locking mechanism
 *
 * make options
 *	-DXWINDOWS	to use xview/X Window System
 */
/*+/mod***********************************************************************
* TITLE	guiSubr.h - definitions for guiSubr.c
*
* DESCRIPTION
*
*-***************************************************************************/

#define GUI_TDIM 120

typedef struct {
    Frame	baseFrame;	/* base frame for this whole shootin' match */
    Display	*pDisp;		/* X display pointer */
    Xv_server	server;		/* xview server handle */
    Frame	fsFrame;	/* command frame for file selector */
    Textsw	fsTextsw;	/* text subwindow for file selection */
    Panel_item	fsDir_PT;	/* directory panel text for file sel */
    Panel_item	fsFile_PT;	/* file panel text for file sel */
    Panel_item	fsStatus_PM;	/* panel message for file sel status */
    void	(*fsCallFn)();	/* user's callback function for file sel */
    void	*fsCallArg;	/* first arg for user's callback function */
    char	fsLsFile[GUI_TDIM];/* file to contain file sel listing */
    char	host[GUI_TDIM];	/* name of host machine, or X display */
    char	user[GUI_TDIM];	/* name of user */
    char	pwd[GUI_TDIM];	/* present working directory */
} GUI_CTX;

typedef struct {
    GUI_CTX	*pGuiCtx;	/* pointer to gui context */
    Frame	print_CF;	/* command frame for printer setup */
    Panel_item	printer_PT;	/* printer name panel text item */
    Panel_item	landscape_PCS;	/* PANEL_CHOICE_STACK item for printer mode */
    Panel_item	prog_PCS;	/* PANEL_CHOICE_STACK item for print program */
    Panel_item	point_PT;	/* PANEL_TEXT item for point size */
    int		pointSize;
} GUI_PRT;

typedef struct {
    GUI_CTX	*pGuiCtx;	/* pointer to gui context */
    GUI_PRT	*pPrtCtx;	/* pointer to printer context */
    Frame	edit_CF;	/* the actual command frame */
    char	title[GUI_TDIM];/* title for command frames */
    char	path[GUI_TDIM*2];/* full path name */
    Panel_item	dir_PT;		/* directory panel text item */
    char	dir[GUI_TDIM];	/* directory */
    Panel_item	file_PT;	/* file panel text item */
    char	file[GUI_TDIM];	/* file name */
    Textsw	text_TSW;	/* text subwindow for browsing/editing */
    Panel_item	update_PCB;	/* state of read-only/update checkbox */
    int		updFlag;	/* 0,1 for read-only/update */
    char	lockFile[GUI_TDIM*2+5];/* path of lockfile */
    int		loaded;		/* 0,1 for not loaded, loaded */
} GUI_EDIT;

/*-----------------------------------------------------------------------------
* XV_KEY_DATA keys
*----------------------------------------------------------------------------*/
enum guiDataKeys {
    GUI_PCTX=1,		/* pointer to gui context */
    GUI_CF,		/* command frame for guiCFshow_mi */
    GUI_PEDIT,		/* pointer to gui editor context */
    GUI_PPRT,		/* pointer to gui printer context */
};
/*-----------------------------------------------------------------------------
* notifier client numbers; user client numbers are expected to be
* in the range 100 to 200
*----------------------------------------------------------------------------*/
enum guiClientNums {
    GUI_FILE_SEL_CLIENT=1,
    GUI_HOSTNAME_CLIENT,
    GUI_PRINT_CLIENT,
};


Textsw guiBrowser();
Panel_item guiButton();
void guiButtonCenter();
void guiCFdismiss_xvo();
void guiCFshow_mi(), guiCFshowPin_mi(), guiCFshow_xvo(), guiCFshowPin_xvo();
Panel_item guiCheckbox();
void guiCheckboxItem();
GUI_EDIT *guiEditor();
void guiEditorNewEntry_pb();
void guiEditorShowCF(), guiEditorShowCF_xvo(), guiEditorShowCF_mi();
void guiFileSelect();
char *guiGetNameFromSeln();
Frame guiFrame(), guiFrameCmd();
Icon guiIconCreate();
Panel guiInit();
int guiLock();
Menu guiMenu();
Menu_item guiMenuItem();
Panel_item guiMessage();
void guiNotice(), guiNoticeFile(), guiNoticeName();
GUI_PRT *guiPrinter();
void guiPrinterShowCF_mi(), guiPrinterShowCF_xvo(), guiPrintFile();
void guiShellCmd();
void guiShellCmd_execv_argv();
Panel_item guiTextField();
char *guiTextswGets();
int guiTextswPuts();
void guiTextswReset();

#endif /* INC_guiSubr_h */
