/* module_types.h */
/* share/epicsH @(#)module_types.h	1.20     7/23/92 */
/*
 * 	Author:      Bob Dalesio
 * 	Date:        12-07-88
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	12-07-88	lrd	moved IO_TYPES.H into here
 *				added interruptable indicator for each io card
 *				added arrays of pointers for info on the io cards
 * .02	02-01-89	lrd	add Allen-Bradley Modules
 *				rearrange I/O card order
 * .03	02-04-89	lrd	add bus specification
 *
 * .04	02-06-89	joh	add timing int vec #  
 * .05	02-09-89	lrd	move vectors and I/O card base addresses here
 * .06	02-11-89	lrd	changed AB num cards to 12 and OFE num chan to 4
 * .04	02-13-89	joh	fixed length of timing address array 
 * .05	04-22-89	lrd	changed AB num channels to 8 for ABBI cards
 * .06	04-24-89	lrd	moved AB 6008SV info here
 * .07  05-26-89	joh	added Joerger VTR1 transient recorder
 * .08	09-13-89	lrd	add VXI module specification to each type
 *				remove unused modules
 * .09	09-25-89	lrd	AB adapters to 8 cards to 16 and IFE channels to
 *				8
 * .10	10-17-89	lrd	AB changed number of cards from 12 to 16 for
 *				binary inputs and outputs
 * .11	11-21-89	joh	fixed card and chnl count for at5 vxi timing 
 * .12	11-30-89	lrd	add the OMS 6 axis motor driver
 * .13	12-12-89	joh	moved XY220 base to match changes made by the
 *				hardware section. 
 * .14	05-16-90	cr	add Single Ended IFE module
 * .15	05-24-90	mk	add 4-20Ma IFE
 * .16  05-23-90        lrd/ms  add DVX-2502
 * .17	07-19-90	ms	changed the mizar interrupt vector to e8
 *				wouldn't work with the CMC ethernet card
 *				used the same interrupt vector
 * .18	12-19-90	lrd	add 0-5V IFE
 * .19	03-26-91	lrd	modified the AT5-VXIBO from 16 to 32
 * .20	07-23-91	ges	Add AT830X card base addresses and vectors.
 * .21	08-14-91	ges	Add AT830X_1 card base addresses and ivectors.
 *				(These are the new gen 8308B and 8309C cards)
 * .22	06-24-92	joh	added the hp E1368a VXI video switching module
 * .23	06-25-92	 bg	stripped intitial 1 off the standard addresses
 *				in AB_BASE_ADDR,ai_memaddrs,wf_memaddrs,AT830X_
 *                               ADDRS, and AT830X_1_ADDRS so they can be mapped
				in sysBusToLocalAdrs
 * .24  06-25-92	 bg	stripped initial ff off all short addresses so
				they can be mapped in sysBusToLocalAdrs
 * .25	06-25-92	 bg	Moved SM_BASE_ADDR frin oms_driver.h to sm_addrs
 				in this file.
 * .26	06-25-92	 bg	Moved dvx addresses,interrupt vectors,channel numbers
                                and card number definitions from dvx_driver.h to here.
 	
 * .26	06-25-92	 bg	Moved dvx addresses,interrupt vectors,channel numbers
 * .27	06-25-92	 bg	Added definitions for xy240_driver   
 * .28	06-25-92	 bg	Removed all reference to the Zio085
 * .29	06-26-92	 bg	Temporarily returned all reference to the Zio085
 * .30	06-26-92	joh	added binary input capability to the hp E1368a 
 *				VXI video switching module
 * .31	06-30-92	joh	added constant for the number of VXI LA
 * .32	07-06-92	joh	added base and size VXIA24 & A32 devices 
 * .33	07-07-92	joh	added KSC V215 VXI AI
 * .34	07-09-92	joh	added  AT8 FP10S master/slave fast 
 *				protect module
 * .35  07-10-92	joh	added FP int vec base
 * .36  07-17-92	 bg	moved addresses for Comet card from drvComet.c to
 *				here.
 * .37	07-23-92	joh	number of AI channels for AT5 VXI should be 8 not 6
 * 				number of BI channels for AT5 VXI should be 32 not 16
 * 				number of TI channels for AT5 VXI should be 10 not 5
 *				(these changes reflect the AT5 hardware upgrade)
 *
 */

/* Device module types */
/*
 * all devices have corresponding entries in ~operator/db/src/menus.c
 * changes must be made in both areas to keep the database and drivers in sync
 */
/* & in comment indicates tested with card  0 */
/* % in comment indicates tested with card other than card 0 */
/* # in comment indicates that the Nth card has been tested */
/* !! never been tested */

/*@#  If any changes are made to this file, check the procedures ab_card, and vme_card
	in signallist.c, and get_address in sigmenu.c.
 */

#ifdef MODULE_TYPES_INIT
#define MODULE_TYPES_DEF(MT_DEF_PARM)  MT_DEF_PARM
#else
#define MODULE_TYPES_DEF(MT_DEF_PARM) extern MT_DEF_PARM;
#endif

/* Xycom defines including those for the XY240 */
#define	XY_LED		0x3	/* set the Xycom status LEDs to OK */
#define	XY_SOFTRESET	0x10	/* reset the IO card */
#define XY566_MEM_INCR  0x10000	/* size of data memory area */
#define XY240           10
#define XY240_ADDR0     0xd000  /* Address of card 0 of the xy240 binary in/out */
#define XY240_MAX_CARDS 2
#define XY240_MAX_CHANS 32

/* VMIC define */
#define	VMI_MAXCHAN	16

/* I/O types */
#define IO_AI		0
#define IO_AO		1
#define IO_BI		2
#define IO_BO		3
#define IO_SM   	4
#define IO_WF  		5
#define	IO_TIMER	6
#define MAX_IO_TYPE	IO_TIMER

/* bus types */
/* must correspond to the values in link types */
/* these defines are in ~gta/dbcon/h/link.h */

/* system controller card */
#define SRM010_ADDR     0x0000	/* base addr of first XY010 cntroller */

/* Number of columns used in io_report. */
#define IOR_MAX_COLS 4

/* equates for the Allen-Bradley cards. */
#define	AB_BASE_ADDR	0xc00000	/* base addr of first AB6008SV */
#define AB_MAX_LINKS    2       /* number of serial links from VME */
#define AB_MAX_ADAPTERS 8       /* number of physical adapters on a link */
#define AB_MAX_CARDS    16      /* max number of IO cards per adapter */
#define AB_CARD_ADAPTER 16      /* cards per logical adapter */
#define AB_CHAN_CARD    16      /* max channels per card */

/* analog inputs */
#define AB1771IL	0	/* &% Allen-Bradley low level analog input */
#define AB1771IFE	1	/* &% Allen-Bradley low level analog input */
#define	AB1771IXE	2	/* &% Allen-Bradley millivolt input */
#define XY566SE		3	/* & Xycom 12-bit Single Ended Scanned*/
#define XY566DI		4	/* &% Xycom 12-bit Differential Scanned */
#define XY566DIL	5	/* &% Xycom 12-bit Differential Latched */
#define	VXI_AT5_AI	6	/* % AT-5 VXI module's Analog Inputs */
#define AB1771IFE_SE	7	/* % A-B IFE in 16 single-ended input mode */
#define AB1771IFE_4to20MA 8	/* % A-B IFE in 8 double-ended 4to20Ma	*/
#define DVX2502         9       /* &% DVX_2502 128 chan 16 bit differential */
#define AB1771IFE_0to5V 10	/* % A-B IFE in 8 double-ended 4to20Ma	*/
#define KSCV215		11	/* % KSC V215 VXI 16 bit differential */
#define MAX_AI_TYPES	KSCV215	
MODULE_TYPES_DEF(short ai_num_cards[MAX_AI_TYPES+1])
#ifdef MODULE_TYPES_INIT
		={12,12,12, 4, 4, 6,32,12,12,  5, 12, 32};
#endif
MODULE_TYPES_DEF(short ai_num_channels[MAX_AI_TYPES+1])
#ifdef MODULE_TYPES_INIT
		={ 8, 8, 8,32,16,16, 8,16, 8, 127,  8, 32};
#endif
MODULE_TYPES_DEF(short ai_interruptable[MAX_AI_TYPES+1])
#ifdef MODULE_TYPES_INIT
		={0, 0, 0, 0, 0, 1, 1, 0, 0,  1,  0, 0};
#endif
MODULE_TYPES_DEF(short ai_bus[MAX_AI_TYPES+1])
#ifdef MODULE_TYPES_INIT
		={ 4, 4, 4, 2, 2, 2, 2, 4, 4,  2,  4, 2};
#endif
MODULE_TYPES_DEF(unsigned short ai_addrs[MAX_AI_TYPES+1])
#ifdef MODULE_TYPES_INIT
		={ 0,0,0,0x6000,0x7000,0xe000, 0xc014,0,0, 0xff00, 0, 0};
#endif
MODULE_TYPES_DEF(long ai_memaddrs[MAX_AI_TYPES+1])
#ifdef MODULE_TYPES_INIT
		={0,0,0,0x000000,0x040000,0x0c0000, 0,0,0, 0x100000, 0, 0};
#endif

/* analog outputs */
#define AB1771OFE	0	/* &% Allen-Bradley 12 bit Analog Output */
#define VMI4100		1	/* & VMIC VMIVME 4100 */
#define ZIO085		2 	/* & Ziomek 085 */
#define	VXI_AT5_AO	3	/* !! AT-5 VXI modules analog outputs */
#define MAX_AO_TYPES	VXI_AT5_AO
MODULE_TYPES_DEF(short ao_num_cards[MAX_AO_TYPES+1])
#ifdef MODULE_TYPES_INIT
                = {12,  4,  1, 32};
#endif
MODULE_TYPES_DEF(short ao_num_channels[MAX_AO_TYPES+1])
#ifdef MODULE_TYPES_INIT
                =  { 4, 16, 32, 16};
#endif
MODULE_TYPES_DEF(short ao_interruptable[MAX_AO_TYPES+1] )
#ifdef MODULE_TYPES_INIT
                = { 0,  0,  0,  1};
#endif
MODULE_TYPES_DEF(short ao_bus[MAX_AO_TYPES+1])
#ifdef MODULE_TYPES_INIT
                ={ 4,  2,  2,  2};
#endif
MODULE_TYPES_DEF(unsigned short ao_addrs[MAX_AO_TYPES+1])
#ifdef MODULE_TYPES_INIT
                ={ 0,0x4100,0x0800, 0xc000};
#endif

/* binary inputs */
#define	ABBI_08_BIT	0	/* &% Allen-Bradley generic Binary In  8 bit */
#define	ABBI_16_BIT	1	/* &% Allen-Bradley generic Binary In 16 bit */
#define BB910		2	/* & BURR BROWN MPV 910 (relay) */
#define XY210		3	/* &% XYcom 32 bit binary in */
#define	VXI_AT5_BI	4	/* !! AT-5 VXI modules binary inputs */
#define HPE1368A_BI	5	/* !! HP E1368A video switch */
#define AT8_FP10S_BI	6	/* !! AT8 FP10 slave fast protect */
#define XY240_BI	7	/* !! Xycom 32 bit binary in / 32 bit binary out */
#define	MAX_BI_TYPES	XY240_BI	
MODULE_TYPES_DEF(short bi_num_cards[MAX_BI_TYPES+1] )
#ifdef MODULE_TYPES_INIT
		={ 12, 12,  4,  4, 32, 32, 8, 2};
#endif
MODULE_TYPES_DEF(short bi_num_channels[MAX_BI_TYPES+1] )
#ifdef MODULE_TYPES_INIT
		={ 8, 16, 32, 32, 32, 16, 32, 32};
#endif
MODULE_TYPES_DEF(short bi_interruptable[MAX_BI_TYPES+1] )
#ifdef MODULE_TYPES_INIT
		={  1,  1,  0,  0,  1, 1, 1, 1};
#endif
MODULE_TYPES_DEF(short bi_bus[MAX_BI_TYPES+1])
#ifdef MODULE_TYPES_INIT
		={  4,  4,  2,  2,  2, 2, 2, 2};
#endif
MODULE_TYPES_DEF(unsigned short bi_addrs[MAX_BI_TYPES+1])
#ifdef MODULE_TYPES_INIT
		={ 0,0,0xb800,0xa000, 0xc000, 0xc000, 0x0e00, 0xd400};
#endif

/* binary outputs */
#define	ABBO_08_BIT	0	/* &% Allen-Bradley 8 bit binary out */
#define	ABBO_16_BIT	1	/* &% Allen-Bradley 16 bit binary out */
#define BB902		2	/* &% BURR BROWN MPV 902 (relay) */
#define XY220		3	/* &% XYcom 32 bit binary out */
#define	VXI_AT5_BO	4	/* !! AT-5 VXI modules binary outputs */
#define HPE1368A_BO	5	/* !! HP E1368A video switch */
#define AT8_FP10M_BO	6	/* !! AT8 FP10 master fast protect */
#define XY240_BO	7	/* !! Xycom 32 bit binary in / 32 bit binary out */
#define	MAX_BO_TYPES	XY240_BO	
MODULE_TYPES_DEF(short bo_num_cards[MAX_BO_TYPES+1] )
#ifdef MODULE_TYPES_INIT
		={12, 12,  4,  4, 32, 32, 2, 2};
#endif
MODULE_TYPES_DEF(short bo_num_channels[MAX_BO_TYPES+1] )
#ifdef MODULE_TYPES_INIT
		={ 8, 16, 32, 32, 32, 16, 32, 32};
#endif
MODULE_TYPES_DEF(short bo_interruptable[MAX_BO_TYPES+1] )
#ifdef MODULE_TYPES_INIT
		={ 0,  0,  0,  0,  1, 0, 0, 1 };
#endif
MODULE_TYPES_DEF(short bo_bus[MAX_BO_TYPES+1])
#ifdef MODULE_TYPES_INIT
		={ 4,  4,  2,  2,  2, 2, 2, 2 };
#endif
MODULE_TYPES_DEF(unsigned short bo_addrs[MAX_BO_TYPES+1])
#ifdef MODULE_TYPES_INIT
		={ 0,0,0xd800,0xc800, 0xc000, 0xc000, 0x0c00, 0xd400};
#endif

/* stepper motor drivers */
#define	CM57_83E	0	/* & Compumotor 57-83E motor controller */
#define	OMS_6AXIS	1	/* & OMS six axis motor controller */
#define	MAX_SM_TYPES	OMS_6AXIS
MODULE_TYPES_DEF(short sm_num_cards[MAX_SM_TYPES+1] )
#ifdef MODULE_TYPES_INIT
                ={ 8, 8 };
#endif
MODULE_TYPES_DEF(short sm_num_channels[MAX_SM_TYPES+1] )
#ifdef MODULE_TYPES_INIT
                =  { 1, 6};
#endif
MODULE_TYPES_DEF(short sm_interruptable[MAX_SM_TYPES+1] )
#ifdef MODULE_TYPES_INIT
                = { 0, 0 };
#endif
MODULE_TYPES_DEF(short sm_bus[MAX_SM_TYPES+1])
#ifdef MODULE_TYPES_INIT
                ={ 2, 2 };
#endif
MODULE_TYPES_DEF(unsigned short sm_addrs[MAX_SM_TYPES+1])
#ifdef MODULE_TYPES_INIT
                ={ 0x8000, 0xfc00 };
#endif

/* waveforms */
#define XY566WF		0	/* & Xycom 566 as a waveform */
#define CAMAC_THING	1	/* !! CAMAC waveform digitizer */
#define JGVTR1		2	/* & Joerger transient recorder */
#define	COMET		3	/* !! COMET transient recorder */
#define MAX_WF_TYPES	COMET
MODULE_TYPES_DEF(short wf_num_cards[MAX_WF_TYPES+1] )
#ifdef MODULE_TYPES_INIT
                ={4, 4, 8, 4};
#endif
MODULE_TYPES_DEF(short wf_num_channels[MAX_WF_TYPES+1] )
#ifdef MODULE_TYPES_INIT
                ={1, 1, 1, 4};
#endif
MODULE_TYPES_DEF(short wf_interruptable[MAX_WF_TYPES+1] )
#ifdef MODULE_TYPES_INIT
                = {0, 0, 0, 0};
#endif
MODULE_TYPES_DEF(short wf_bus[MAX_WF_TYPES+1])
#ifdef MODULE_TYPES_INIT
                ={2, 3, 2, 2};
#endif
MODULE_TYPES_DEF(unsigned short wf_addrs[MAX_WF_TYPES+1])
#ifdef MODULE_TYPES_INIT
                ={0x9000, 0, 0xB000, 0xa000};
#endif
MODULE_TYPES_DEF(unsigned short wf_armaddrs[MAX_WF_TYPES+1])
#ifdef MODULE_TYPES_INIT
                = {0x5400, 0, 0, 0};
#endif
MODULE_TYPES_DEF(long wf_memaddrs[MAX_WF_TYPES+1])
#ifdef MODULE_TYPES_INIT
                ={0x080000, 0, 0xb80000, 0xe0000000};
#endif


/* timing cards */
#define MZ8310		0	/* &% Mizar Timing Module */
#define DG535		1	/* !! GPIB timing instrument */
#define	VXI_AT5_TIME	2	/* !! AT-5 VXI modules timing channels */
#define MAX_TM_TYPES	VXI_AT5_TIME
MODULE_TYPES_DEF(short tm_num_cards[MAX_TM_TYPES+1] )
#ifdef MODULE_TYPES_INIT
                ={ 4, 1,  32 };
#endif
MODULE_TYPES_DEF(short tm_num_channels[MAX_TM_TYPES+1] )
#ifdef MODULE_TYPES_INIT
                ={10, 1, 10};
#endif
MODULE_TYPES_DEF(short tm_interruptable[MAX_TM_TYPES+1] )
#ifdef MODULE_TYPES_INIT
                = { 1, 0,  1 };
#endif
MODULE_TYPES_DEF(short tm_bus[MAX_TM_TYPES+1])
#ifdef MODULE_TYPES_INIT
                ={ 2, 5,  2 };
#endif
MODULE_TYPES_DEF(unsigned short tm_addrs[MAX_TM_TYPES+1])
#ifdef MODULE_TYPES_INIT
                ={0xf800, 0, 0xc000 };
#endif

/* AT830X_1 cards */
MODULE_TYPES_DEF(long AT830X_1_addrs )
#ifdef MODULE_TYPES_INIT
                = 0x0400;
#endif
MODULE_TYPES_DEF(short AT830X_1_num_cards )
#ifdef MODULE_TYPES_INIT
                = 2;
#endif

/* AT830X cards */
MODULE_TYPES_DEF(long AT830X_addrs )
#ifdef MODULE_TYPES_INIT
                = 0xaa0000;
#endif
MODULE_TYPES_DEF(short AT830X_num_cards )
#ifdef MODULE_TYPES_INIT
                = 2;
#endif

/* pointer to the interruptable card arrays for each I/O type */
ai_interruptable,
ao_interruptable,
bi_interruptable,
bo_interruptable,
sm_interruptable,
wf_interruptable,
tm_interruptable
};

/* pointer to the number of channel arrays for each I/O type */
ai_num_channels,
ao_num_channels,
bi_num_channels,
bo_num_channels,
sm_num_channels,
wf_num_channels,
tm_num_channels
};

/* pointer to the number of card arrays for each I/O type */
ai_num_cards,
ao_num_cards,
bi_num_cards,
bo_num_cards,
sm_num_cards,
wf_num_cards,
tm_num_cards
};

/* pointer to the IO bus arrays for each I/O type */
ai_bus,
ao_bus,
bi_bus,
bo_bus,
sm_bus,
wf_bus,
tm_bus
};

/* pointer to the VME address arrays for each I/O type */
ai_addrs,
ao_addrs,
bi_addrs,
bo_addrs,
sm_addrs,
wf_addrs,
tm_addrs,
};

/* maximum card types for each I/O type */
MAX_AI_TYPES,
MAX_AO_TYPES,
MAX_BI_TYPES,
MAX_BO_TYPES,
MAX_SM_TYPES,
MAX_WF_TYPES,
MAX_TM_TYPES,
};

/* 
 *	limit the size of the VXI logical address space
 *
 *	<VXI LA BASE> = <VME SHORT ADDR BASE> + 0xc000
 *	
 *	LA			VME address
 *	0	 		<VXI LA BASE>
 *	EPICS_VXI_LA_COUNT	<VXI LA BASE> + (EPICS_VXI_LA_COUNT-1)*64
 */
MODULE_TYPES_DEF(unsigned char EPICS_VXI_LA_COUNT)
#ifdef MODULE_TYPES_INIT
		= 32;
#endif

/*
 *
 *	address ranges for VXI A24 and A32 devices
 *
 */
MODULE_TYPES_DEF(char *EPICS_VXI_A24_BASE)
#ifdef MODULE_TYPES_INIT
		= (char *) 0x900000;
#endif
MODULE_TYPES_DEF(unsigned long EPICS_VXI_A24_SIZE)
#ifdef MODULE_TYPES_INIT
		= 0x100000;
#endif
MODULE_TYPES_DEF(char *EPICS_VXI_A32_BASE)
#ifdef MODULE_TYPES_INIT
		= (char *) 0x90000000;
#endif
MODULE_TYPES_DEF(unsigned long EPICS_VXI_A32_SIZE)
#ifdef MODULE_TYPES_INIT
		= 0x10000000;
#endif


/* interrupt vector allocation - one for each XY566 DIL card */
#define AI566_VNUM		0xf8	/* Xycom 566 Differential Latched */

/* stepper motor interrupt vector - one for each motor */
#define MD_INT_BASE		0xf0	/* base of the motor int vector */

/* I reserve from here up to num_cards * 4 interrupting chans/card - joh */
#define MZ8310_INT_VEC_BASE	0xe8

/* Allen-Bradley Serial Driver - MAX_AB_LINKS number of vectors */
#define	AB_VEC_BASE		0x60

/* only one interrupt vector allocated for all Joerger VTR1 boards joh */
#define JGVTR1_INT_VEC		0xe0

/* AT830X_1 cards have 1 intr vector for each AT830X_1_num_cards (presently 2) */
#define	AT830X_1_IVEC0		0xd4

/* AT830X cards have 1 intr vector for each AT830X_num_cards (presently 2) */
#define	AT830X_IVEC0		0xd6
 
/* interrupt vector for  dvx_card 0 and one for each DVX card (presently 5) */
#define DVX_IVEC0       0xD0            /* interrupt vector for card 0 */

/* AT8 fast protect interrupt vector base */
#define AT8FP_IVEC_BASE		0xa2	/* usually <= 8 cards */
#define AT8FPM_IVEC_BASE	0xa0	/* usually <= 2 cards */
