/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	STRUCTURES FOR CONNECTIONLESS PARAMETER GETS		*/
/*								*/
/*	These data types are for the connectionless parameter	*/
/*	gets. DBR_SHORT_PARAM should be equal to the symbol	*/
/*	LAST_BUFFER_TYPE + 1. LAST_BUFFER_TYPE is defined in	*/
/*	db_access.h. It couldn't be included here because of	*/
/*	colisions in rsrvTask and many other places.		*/
/*								*/
/*	Author: Gordon Ucheick					*/
/*								*/
/****************************************************************/

#define DBR_SHORT_PARAM		35	/* WARNING! SHOULD BE LAST_BUFFER_TYPE + 1
					   from db_access.h! */
#define DBR_ENUM_PARAM		(DBR_SHORT_PARAM + 1)
#define DBR_STRING_PARAM 	(DBR_ENUM_PARAM + 1)
#define DBR_FLOAT_PARAM		(DBR_STRING_PARAM + 1)
#define DBR_LONG_PARAM		(DBR_FLOAT_PARAM + 1)
#define DBR_DOUBLE_PARAM	(DBR_LONG_PARAM + 1)

/* Structures for connectionless parameter gets */

struct dbr_short_param {
	short	hopr;			/* High operating range */
	short	lopr;			/* Low operating range */
	short	drvh;			/* Upper control limit */
	short	drvl;			/* Lower control limit */
	short	nativeType;		/* Native field type */
	short	nativeCount;		/* Native field count */
	short	prec;			/* digits after decimal */
	char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];	/* Channel Name */
	char	units[8];		/* Egu tag */
	char	resName[16];		/* Resource name */
	char	desc[28];		/* Point descriptor */
	};

struct dbr_enum_param {
	short	no_str;
	char	strs[64][26];
	char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];	/* Channel Name */
	char	units[8];		/* Egu tag */
	char	resName[16];		/* Resource name */
	char	desc[28];		/* Point descriptor */
	};

struct dbr_string_param {
	char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];	/* Channel Name */
	char	units[8];		/* Egu tag */
	char	resName[16];		/* Resource name */
	char	desc[28];		/* Point descriptor */
	};

struct dbr_float_param {
	float	hopr;			/* High operating range */
	float	lopr;			/* Low operating range */
	float	drvh;			/* Upper control limit */
	float	drvl;			/* Lower control limit */
	short	nativeType;		/* Native field type */
	short	nativeCount;		/* Native field count */
	short	prec;			/* digits after decimal */
	char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];	/* Channel Name */
	char	units[8];		/* Egu tag */
	char	resName[16];		/* Resource name */
	char	desc[28];		/* Point descriptor */
	};

struct dbr_long_param {
	int	hopr;			/* High operating range */
	int	lopr;			/* Low operating range */
	int	drvh;			/* Upper control limit */
	int	drvl;			/* Lower control limit */
	short	nativeType;		/* Native field type */
	short	nativeCount;		/* Native field count */
	short	prec;			/* digits after decimal */
	char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];	/* Channel Name */
	char	units[8];		/* Egu tag */
	char	resName[16];		/* Resource name */
	char	desc[28];		/* Point descriptor */
	};

struct dbr_double_param {
	double	hopr;			/* High operating range */
	double	lopr;			/* Low operating range */
	double	drvh;			/* Upper control limit */
	double	drvl;			/* Lower control limit */
	short	nativeType;		/* Native field type */
	short	nativeCount;		/* Native field count */
	short	prec;			/* digits after decimal */
	char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];	/* Channel Name */
	char	units[8];		/* Egu tag */
	char	resName[16];		/* Resource name */
	char	desc[28];		/* Point descriptor */
	};
