/****************************************************************/
/*								*/
/*		LIST OF FIELDS REQUIRING REDUNDANT UPDATE	*/
/*								*/
/*	The variable papRedFldList is globally available.	*/
/*	This pointer points to a list that is indexed by	*/
/*	the point type. Each entry in this list points to 	*/
/*	a list of pointers to struct fldDes (see dbRecDes.h)	*/
/*	terminated by a null pointer. This list defines		*/
/*	the master set of all fields that might ever have to	*/
/*	participate in a redundant update for that type of	*/
/*	point.							*/
/*								*/
/*	dbServer initializes papRedFldList and creates the	*/
/*	list of pointers indexed by type but leaves all of	*/
/*	those pointers NULL. Each record type initialization	*/
/*	routine allocates its own list and fills it by		*/
/*	calling the various routines listed below.		*/
/*								*/
/****************************************************************/

#include "dbRecDes.h"

extern struct fldDes ***papRedFldList;

struct fldDes **redFldAlloc(int type, int nfields);
/******
	This routine is called to allocate a list of pointers
	for a specific point type. Room for one more field
	than specified is actually allocated so that the list
	can end with a null pointer. This routine will also
	store the pointer to this new array in the proper list
	so that it is accessable by all parties.

	Syntax:
		redFldAlloc(type, nfields); 
		int	type;
		int	nfields;
	Arguments:
		The point type is in type; The number of fields
		that the user intends to define is contained in
		nfields.
	Returns:
		The address of the array that has been allocated.
		This should be retained by the record initialization
		routine for later used. NULL is returned if the
		buffer can't be allocated.
******/


int redFldAdd(int type, int index, char *field);
/******
	This routine is called to add a field to the list in a
	given position.

	Syntax:
		redFldAdd(type, index, field);
		int		type;
		int		index;
		char		*field;
	Arguments:
		The point type is in point.  The desired
		index position is in index. The field tag
		(i.e., the four char max name, not the long one)
		is passed in field.
	Returns:
		The routine returns 0 on success or -1 if the
		list for the point type hasn't been allocated,
		if the point type number passed is illegal,
		or if the field can't be found.
******/
