/*
**  sslShared.h - systemwide ssl routines
**
*/

#ifndef SSLSHARED_H
#define SSLSHARED_H

/* define to use SSLC from RSA (remember to redirect include search paths) */
/*
	#define USE_SSLC
*/

/* include files from the SSL packages */
#include "rsa.h"
#include "x509.h"
#include "pem.h"
#include "ssl.h"
#include "err.h"

#include "e_os.h"
#include "buffer.h"
#include "bio.h"
#include "asn1.h"


/* maximum x509 used in auditing/privileges */
#define X509_MAX_DN 71


/* Set disk certificate usage */
void UseDiskCertificate (int value);

/* Set up the context used by all programs in this computer */
int sslSharedInit (void);

/* Release the common context and associated storage */
int sslSharedExit (void);

/* tracking of ssl connection blocks to fd values */
int sslSharedRegConn (SSL *sslConn, int fd);

/* close tracking of ssl connection blocks to fd values */
int sslSharedCloseConn (SSL *sslConn);

/* tracking of ssl connection blocks to fd values */
SSL *sslSharedGetConn (int fd);

/* Routine to listen on a socket, accept clients, and spawn a task */
/* to handle them. */
/* This provides most of the startup functionality required for a server */
/* The routine incorporates init_server, do_accept, setup_ssl_accept calls */
int do_ssl_server (int port, int *ret, void (*cb)(char *, SSL *),
		   char *name, int pri, int stsize);

/* convenience function to open the server listen socket */
int init_server (int *sock, int port);

/* convenience function to perform accept, and identify other host name or IP */
int do_accept (int acc_sock, int *new_sock, char **host);

/* convenience function to prepare for SSL promotion on accept socket */
SSL *setup_ssl_accept (int sockfd);

/* Attempt to perform SSL promotion on accept socket */
int accept_ssl_connection (SSL *con);

/* Attempt to perform SSL promotion on accept socket (with timeout) */
int accept_ssl_connection_timed (SSL *con, int timeout /* in seconds */);

/* Attempt to perform SSL connect to specified host/port, returns SSL conn */
SSL *connect_ssl_connection (long hostaddr  /* in net order */,
			    int port	    /* in host order */);

/* Attempt to perform SSL connect to specified host/port(with timeout),
** returns SSL conn */
SSL *connect_ssl_connection_timed (long hostaddr /* in net order */,
			     int port,	    /* in host order */
			     int timeout);  /* in seconds */

/* Set password string for shared certificate */
int passwd (char *new_passwd);

/* Routine to retrieve DN field from env certificate */
char *GetMyDN (char *retName, int retSize);

/* Routine to retrieve CN field from env certificate */
char *GetMyCN (char *retName, int retSize);

/* Routine to retrieve DN field from peer certificate */
char *GetPeerDN (SSL *ssl_con, char *retName, int retSize);

/* Routine to retrieve CN field from peer certificate */
char *GetPeerCN (SSL *ssl_con, char *retName, int retSize);

/* Routine to get SN and valid dates from peer certificate */
char *stringCertInfo (SSL *ssl_con);

#endif SSLSHARED_H

