#ifndef INC_sydDefs_h
#define INC_sydDefs_h
/*	@(#)sydDefs.h	1.9 3/9/92
 *	Author:	Roger A. Cole
 *	Date:	12-04-90
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991-92, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .00	12-04-90	rac	initial version
 * .01	06-19-91	rac	installed in SCCS
 * .02	08-16-91	rac	added macros; changed some sydXxx names
 * .03	09-22-91	rac	added more sydXxx names; added structures
 *				to handle triggering
 * .04	11-21-91	rac	added EGU to channel structure; added time
 *				stamp rounding factor; add more sydXxx names
 * .05	01-20-92	rac	fix SydChanDbfType to return TYPENOTCONN
 *				when appropriate
 * .06	02-04-92	rac	add text for notConn status code
 * .07	03-08-92	rac	add dflt deadband and sydSetAttr
 *
 * include options
 *	-DSYD_PRIVATE	makes structures visible--use only by sydXxx routines
 *	-DSYD_PRIVATE_DATA defines global text fields--use only by sydSubr.c
 */
/*+/mod***********************************************************************
* TITLE	sydDefs.h - definitions for synchronous data routines
*
* DESCRIPTION
*
*-***************************************************************************/
#ifndef INC_genDefs_h
#   include <genDefs.h>
#endif
#ifndef INC_tsDefs_h
#   include <tsDefs.h>
#endif
#ifndef INCLcadefh
#   include <cadef.h>
#endif
#include <errMdef.h>	/* get M_syd for this subsystem's `number' */
/*-----------------------------------------------------------------------------
* some MACROS for handling synchronous data structures
*----------------------------------------------------------------------------*/
#define NEXT_INBUF(pSChan, sub) (((sub)+1 < pSChan->nInBufs) ? (sub)+1 : 0)


/*/subhead SYD_ATTR------------------------------------------------------------
*
*----------------------------------------------------------------------------*/
typedef enum sydAttr {
    SYD_ATTR_DEADBAND,
} SYD_ATTR;


/*/subhead SYD_TRIG---------------------------------------------------------
* SYD_TRIG
*
*    .type	type of triggering to do
*    .cond	trigger condition for conditional trigger types
*    .pSChan	sync channel ptr for types of triggering which use channels
*    .trigTs	is the time at which to trigger for DELAY and PERIODIC
*    .preTs	is the amount of time (relative to the trigger time) prior
*		to the trigger time for which samples are to be taken
*    .postTs	is the amount of time (relative to the trigger time) after
*		the trigger time for which samples are to be taken
*    .preCount	is the number of samples to be taken prior to the trigger
*    .postCount	is the number of samples to be taken after the trigger
*    .compVal	is the value for comparison trigger types
*----------------------------------------------------------------------------*/
enum sydTrigType {
    SYD_TR_DEMAND,
    SYD_TR_DELAY,
    SYD_TR_PERIODIC,
    SYD_TR_ALWAYS,
    SYD_TR_NEVER,
    SYD_TR_EVENT,
    SYD_TR_WHILE,
};
enum sydTrigCond {
    SYD_TRC_CHANGE,
    SYD_TRC_EQ,
    SYD_TRC_NE,
    SYD_TRC_GT,
    SYD_TRC_GE,
    SYD_TRC_LT,
    SYD_TRC_LE,
};

typedef struct sydTrigger {
    enum sydTrigType type;
    enum sydTrigCond cond;
    struct sydChan *pSChan;
    TS_STAMP	trigTs;
    TS_STAMP	preTs;
    TS_STAMP	postTs;
    long	preCount;
    long	postCount;
    union {
	double	dblVal;
	long	lngVal;
	char	strVal[db_strval_dim];
    } compVal;
} SYD_TRIG;

/*/subhead SYD_CHAN--------------------------------------------------------
*
* input buffer status should not be confused with the status of input data.
* In general, the status of the data is a part of the structure containing
* the data.  The data structures are of type DBR_TIME_xxx, which has
* .status and .severity items.  The input buffer status codes are:
*
*    SYD_B_EMPTY	the buffer is empty of data (and status)
*    SYD_B_FULL		the buffer contains data with a valid time stamp
*    SYD_B_RESTART	similar to SYD_B_FULL, except the value restarts
*		a "run" following missing data.  This status is used at
*		the beginning of file and when an archive request is
*		activated.  (For "retrievals" for real-time data, this
*		is used following either the initial connection or a
*		resumed connection after a disconnect.)
*    SYD_B_EOF		channel's stream is at end of file
*    SYD_B_MISSING	there is no data for the channel.  This status
*		indicates the beginning of a "hiatus" in data for the
*		channel, as between inactivate and activate for an
*		archive request or between disconnect and reconnect for
*		a real-time channel.  This status is different from
*		SYD_B_EMPTY because this status has a valid time stamp,
*		which is important in deciding how to construct synchronous
*		samples.
*    SYD_B_SAMPLED	the buffer has valid data which has been used
*		in at least one sample.
*----------------------------------------------------------------------------*/
enum sydBStatus {
    SYD_B_EMPTY,
    SYD_B_MISSING,
    SYD_B_RESTART,
    SYD_B_FULL,
    SYD_B_SAMPLED,
    SYD_B_EOF
};
enum sydChanStep {
    SYD_ST_UNDEF,
    SYD_ST_STEP,
    SYD_ST_SMOOTH,
    SYD_ST_PAST
};
enum sydChanSync {
    SYD_SY_UNDEF,
    SYD_SY_NONF,
    SYD_SY_FILLED,
    SYD_SY_PAST
};
/*-----------------------------------------------------------------------------
* NOTE WELL!!! If you add to these flags, then update the ugly initializer
*	in sydInputStoreInSet (static struct sydChanFlags flags0={ ... };)
*----------------------------------------------------------------------------*/
struct sydChanFlags {
    unsigned int ok : 1;	/* value is OK; often with other flags */
    unsigned int restart : 1;	/* value restarts a "run"; often with other
				flags */
    unsigned int filled : 1;	/* value filled from prior value */
    unsigned int missing : 1;	/* no value for time stamp; filled values
				will not have this flag set.  This will
				often be set in conjunction with other
				flags, such as EOF */
    unsigned int eof : 1;	/* end of file */
    unsigned int discon : 1;	/* channel was disconnected */
    unsigned int inact : 1;	/* archive request was inactivated */
};

/*/subhead SYD_CHAN------------------------------------------------------------
* SYD_CHAN
*----------------------------------------------------------------------------*/
typedef struct sydChan {
    struct sydChan *pPrev;      /* link to previous channel */
    struct sydChan *pNext;      /* link to next channel */
    struct sydSpec *pSspec;	/* pointer to synchronous data spec */
    void	*pHandle;
    evid	evid;
    short	conn;		/* 1 if connected at least once */
    short	discon;		/* 1 if currently disconnected */
    short	gotGr;
    short	dataChan;
    short	trigChan;
    short	restart;
#define SYD_CHAN_MAX_IN 10
    short	nInBufs;
    short	firstInBuf;
    short	lastInBuf;
    short	sampInBuf;
    short	reused;
    union db_access_val	*pInBuf[SYD_CHAN_MAX_IN];
    enum sydBStatus inStatus[SYD_CHAN_MAX_IN];
    double	maxDataVal;
    double	minDataVal;
    void	*pData;
    UCHAR	*pDataAlStat;
    UCHAR	*pDataAlSev;
    char	*pDataCodeL;
    char	*pDataCodeR;
    struct sydChanFlags *pFlags;
    char        name[db_name_dim];
    char	EGU[db_units_dim];
    char	label[120];	/* label using chanName and units */
    short	precision;	/* number of decimal places to print;
				0 for integers, -1 for ENUM */
    unsigned short elCount;
    chtype      dbrType;
    chtype	dbfType;
    enum sydChanStep step;
    enum sydChanSync sync;
    union db_access_val	grBuf;
    chtype	dbrGrType;
} SYD_CHAN;
/*-----------------------------------------------------------------------------
*       A synchronous data channel descriptor provides for:
*	o   buffering data during "acquisition"
*	o   preparing "synchronous samples"
*	o   building an accumulation of "synchronous samples" (perhaps in
*	    conjunction with the select specification from the Sspec)
*
* SYD_CHAN_MAX_IN
*	This symbol limits the number of input buffers for saving data
*	while trying to build a coincidence sample.  When retrieving
*	from AR "by channel" streams, 2 input buffers are required.
*	For AR "sample set" streams, 2 input buffers are required.
*	For pfoPlot "streams", 2 input buffers are required.
*	When retrieving from real-time channels, this number, in
*	conjunction with the beam rep rate, indicates the delay which
*	will be tolerated.  A value of 10 with a rep rate of 10 indicates
*	that a 1 second delay can be handled.
*
* interact with the "source"
*
*	pSspec->pFunc	points to a function which is called to open
*			and close the source and to acquire data.  See
*			SYD_SPEC for more information.
*
*	.pHandle	a "handle" to be used by pFunc() for interacting
*			with the "source" for this channel.  The nature
*			of this handle depends on Cset type:
*			o  for SYD_TY_CF pHandle is (AR_CHAN_DESC *)
*			o  for SYD_TY_SSF pHandle is not used
*			o  for SYD_TY_CA pHandle is (chid)
*			o  for SYD_TY_PFO pHandle is (PFO_WIN *) for OPI
*	.evid		for use only with SYD_TY_CA; (evid)
*	.gotGr		for use only with SYD_TY_CA; 1 says have gr info
*	.restart	for use only with SYD_TY_CA; 1 says restart a run
*
* acquire data from the "source"
*
*	Data acquired from the source is stored in input buffers.  There
*	is a fixed size array of pointers to input buffers; the actual
*	number of input buffers which are available to hold data is
*	available in nInBufs.  There are several arrays which parallel
*	the array of input buffer pointers.
*
*	.nInBufs	the number of input buffers available
*	.firstInBuf	subscript for oldest input buffer containing data.
*	.lastInBuf	subscript for last input buffer to receive data.
*	.sampInBuf	subscript of input buffer which is to be part
*			of a coincidence sample; if < 0, this channel doesn't
*			contribute to the sample.
*	.reused		if 1, the value indicated by .sampInBuf has been
*			used in previous samples; else 0  .
*	.pInBuf[]	pointers to input buffers
*	.inStatus[]	statuses for the input buffers; one of SYD_B_xxx
*
* synchronous sample set storage
*	.maxDataVal	maximum value in the coincidence set
*	.minDataVal	minimum value in the coincidence set
*	.pData		pointer to array of data values.  For array channels,
*			a "data value" in *pData is actually an array of
*			dimension [.elCount] of values.
*	.pDataAlStat	pointer to array of data alarm status codes
*	.pDataAlSev	pointer to array of data alarm severity codes
*	.pFlags		pointer to array of flag structures
*	.pDataCodeL	pointer to array of left-hand status code char's
*	.pDataCodeR	pointer to array of right-hand status code char's
*
* channel characteristics
*	.name		name of the channel (as entered by user)
*	.EGU		engineering units text
*	.label		a label consisting of name and units
*	.precision	number of dec. places to print; 0 for int, -1 for ENUM
*	.elCount	array element count for channel
*	.dbrType	form in which data is to be acquired and stored
*	.step		indicates how the channel's value changes:
*			SYD_ST_STEP	changes as a step function
*			SYD_ST_SMOOTH	changes smoothly between values
*	.sync		indicates if channel is:
*			SYD_SY_NONF	not filled, or
*			SYD_SY_FILLED	filled
*	.grBuf		a (union db_access_val) containing graphics info
*	.dbrGrType	type of graphics buffer--one of DBR_GR_xxx
*----------------------------------------------------------------------------*/

/*/subhead SYD_SPEC-------------------------------------------------------
* SYD_SPEC
*----------------------------------------------------------------------------*/
enum sydFuncCode {
    SYD_FC_INIT,
    SYD_FC_OPEN,
    SYD_FC_READ,
    SYD_FC_READ_SYNC,
    SYD_FC_POSITION,
    SYD_FC_CLOSE,
    SYD_FC_FILEINFO,
    SYD_FC_WRAPUP
};
enum sydType {
    SYD_TY_CF,		/* AR channel file */
    SYD_TY_SSF,		/* AR sample set file */
    SYD_TY_CA,		/* Channel Access */
    SYD_TY_PFO		/* plotting for OPI data */
};

typedef struct sydSpec {
    long	(*pFunc)();
    enum sydType type;
    void	*pHandle;
    double      sampleSec;      /* `X' coordinate, as seconds past
                                           the reference time of the
                                           entire sync sample set */
    short	partial;	/* partial sample flag */
    SYD_TRIG	*pTrigger;      /* pointer to `trigger' structure */
    SYD_CHAN	*pChanHead;     /* pointer to head of channel list */
    SYD_CHAN	*pChanTail;     /* pointer to tail of channel list */
    TS_STAMP    sampleTs;       /* time for sync sample */
    TS_STAMP    priorTs;        /* time for prior sync sample */
    TS_STAMP    refTs;          /* reference time for entire sync sample set */
    int         reqCount;       /* number of sets to acquire */
    int         sampleCount;    /* number of sets acquired */
    int         chanCount;      /* count of channels selected */
    short       eofCount;       /* count of channels with EOF */
    short	dataDim;
    short	lastData;
    short	firstData;
    double	*pDeltaSec;
    TS_STAMP	*pTimeStamp;
    char	*pPartial;
    short	nInBufs;
    unsigned long roundNsec;	/* nano-second rounding interval */
    unsigned long deadband;	/* deadband flag for ca_add_event */
} SYD_SPEC;

/*------------------------------------------------------------------------
*
*       A synchronous set spec is the `master structure' which is used for
*       accessing data.  The SYD_SPEC is created empty, and then filled
*       in with additional calls.
*
* interact with the "source"
*	.pFunc		points to a function which is called to open
*			and close the source and to acquire data.
*			(In some cases, data is stored by some outside
*			agent, rather than by calling pFunc().)  When
*			pFunc() is called, it is passed a function
*			code telling it what to do:
*			SYD_FC_INIT	initialize the "source"
*			SYD_FC_OPEN	"open" a channel
*			SYD_FC_READ	"read" a data item for a channel
*			SYD_FC_CLOSE	"close" a channel
*			SYD_FC_WRAPUP	wrapup the "source"
*
*			sydCFFunc()	for "by channel" data
*			sydSSFFunc()	for "sample set" data
*	.type		type of source:
*			SYD_TY_CF   "by channel"
*			SYD_TY_SSF  "sample set"
*
*	.pHandle	an argument to be used by pFunc().  Typically,
*			this will be an AR_CF_DESC, a FILE *, or NULL.
*
* "time base" storage for coincidence set
*	.dataDim	the number of data "slots" available
*	.firstData	subscript of first "slot" item containing data.
*	.lastData	subscript of last "slot" item containing data. All
*			"slots" from [firstData] through [.lastData]
*			have data (wrapping around, if needed).
*	.pDeltaSec[]	pointer to array of delta times past first sample
*	.pTimeStamp[]	pointer to array of time stamps for samples
*	.pPartial[]	pointer to array of partial sample flags: 1 indicates
*			indicates a partial sample--one or more
*			channels in the sample has a status flag which is
*			not one of the SYD_VAL_OKxxx flags.
*----------------------------------------------------------------------------*/


/*/subhead status codes-------------------------------------------------------
*		S T A T U S   C O D E S
*
*	
*    Status codes of the form S_ts_briefMessage are used with the AR access
*    library routines.
*
*    The macro SydStatusToText(stat) can be used to obtain a text string
*    corresponding to a status code.
*
*----------------------------------------------------------------------------*/
#define S_syd_OK		(M_syd|0| 0<<1) /* success */
#define S_syd_noMem		(M_syd|1| 1<<1) /* couldn't malloc memory */
#define S_syd_chanNotFound	(M_syd|1| 2<<1) /* couldn't find channel */
#define S_syd_partial		(M_syd|1| 3<<1) /* partial retrieval set */
#define S_syd_EOF		(M_syd|1| 4<<1) /* end of file */
#define S_syd_ERROR		(M_syd|1| 5<<1) /* other error */
#define S_syd_noDataNow		(M_syd|1| 6<<1) /* no data available now */
#define S_syd_chanNotConn	(M_syd|1| 7<<1) /* chan not connected to CA */

#define SYD_STAT_S_PAST		 8	/* one past last legal code */

#define SydStatusToIndex(status) \
	(  ((status&0xffff0000)!=M_syd)  \
		? SYD_STAT_S_PAST  \
		: (  (((status&0xffff)>>1)>=SYD_STAT_S_PAST)  \
		     ? SYD_STAT_S_PAST  \
		     : ((status&0xffff)>>1)  \
		  )  \
	)

#define SydStatusToText(status) \
	(glSydStatText[SydStatusToIndex(status)])

#ifndef SYD_PRIVATE_DATA
    extern char *glSydStatText[];
#else
    char *glSydStatText[] = {
	/* S_syd_OK            */ "success",
	/* S_syd_noMem         */ "couldn't malloc memory",
	/* S_syd_chanNotFound  */ "couldn't find channel",
	/* S_syd_partial       */ "partial retrieval set",
	/* S_syd_EOF           */ "end of file",
	/* S_syd_ERROR         */ "error",
	/* S_syd_noDataNow     */ "no data available now",
	/* S_syd_chanNotConn   */ "channel not connected",

	/* SYD_STAT_S_PAST          */ "illegal syd status code",
    };
#endif

/*/subhead macros--------------------------------------------------------------
* macros
*----------------------------------------------------------------------------*/
#define SydChanArrayCount(pSChan) ((pSChan)->elCount)
#define SydChanDbfType(pSChan) \
	((pSChan)->discon \
	    ? TYPENOTCONN \
	    : (pSChan)->dbfType)
#define SydChanDbrType(pSChan) ((pSChan)->dbrType)
#define SydChanName(pSChan) ((pSChan)->name)
#define SydChanEGU(pSChan) ((pSChan)->EGU)
#define SydInputTs(pSspec) (pSspec)->sampleTs
#define SydInputStatus(pSChan) \
	((pSChan)->sampInBuf < 0 \
	    ? SYD_B_MISSING \
	    : (pSChan)->inStatus[(pSChan)->sampInBuf])
#define SydInputMarkAsSampled(pSChan) \
	((pSChan)->sampInBuf < 0 \
	    ? SYD_B_MISSING \
	    : ((pSChan)->inStatus[(pSChan)->sampInBuf]=SYD_B_SAMPLED))
#define SydInputValAsVoidPtr(pSChan) (dbr_value_ptr( \
    (void *)(pSChan)->pInBuf[(pSChan)->sampInBuf], SydChanDbrType(pSChan)))
#define SydInputValAsFloat(pSChan) (*(float *)SydInputValAsVoidPtr(pSChan))
#define SydInputValAsDouble(pSChan) (*(double *)SydInputValAsVoidPtr(pSChan))
#define SydInputValAsShort(pSChan) (*(short *)SydInputValAsVoidPtr(pSChan))
#define SydInputValAsLong(pSChan) (*(long *)SydInputValAsVoidPtr(pSChan))
#define SydInputValAsChar(pSChan) (*(char *)SydInputValAsVoidPtr(pSChan))
#define SydInputValAsString(pSChan) ((char *)SydInputValAsVoidPtr(pSChan))


/*/subhead prototypes----------------------------------------------------------
* prototypes
*----------------------------------------------------------------------------*/
int  nextAlphField (char **ppText, char **ppField, char *pDelim);
long sydChanClose(SYD_SPEC *pSspec, SYD_CHAN *pSChan);
SYD_CHAN *sydChanFind(SYD_SPEC *pSspec, char *chanName);
long sydChanOpen(SYD_SPEC *pSspec, SYD_CHAN **ppSChan, char *chanName, enum sydChanSync sync, void *pArg, int trig);
long sydChanOpen1(SYD_SPEC *pSspec, SYD_CHAN *pSChan);
long sydChanOpenGR(SYD_CHAN *pSChan);
long sydClose(SYD_SPEC *pSspec);
long sydFileInfo(SYD_SPEC *pSspec, FILE *outStream);
long sydInputGet(SYD_SPEC *pSspe, int *pMoreFlag);
void sydInputReset(SYD_SPEC *pSspec);
void sydInputResetKeepNewest(SYD_SPEC *pSspec);
void sydInputResetSampled(SYD_SPEC *pSspec);
void sydInputStoreInSet(SYD_SPEC *pSspec, int ignorePartial);
long sydInputSync(SYD_SPEC *pSspec);
long sydOpen(SYD_SPEC **ppSspec);
long sydPosition(SYD_SPEC *pSspec, TS_STAMP *pStamp);
long sydSampleExport(SYD_SPEC *pSspec, FILE *out, int option, int samp);
long sydSamplePrint(SYD_SPEC *pSspec, FILE *out, int formatFlag,
		    int headerFlag, int nCol, int colWidth, int samp);
long sydSampleWriteSSF(SYD_SPEC *pSspec, FILE *pFile, char *reqDesc,
		       char *sampDesc, int samp);
long sydSampleSetAlloc(SYD_SPEC *pSspec, int reqCount);
long sydSampleSetExport(SYD_SPEC *pSspec, FILE *out, int option);
long sydSampleSetFree(SYD_SPEC *pSspec);
long sydSampleSetGet(SYD_SPEC *pSspec);
long sydSampleSetPrint(SYD_SPEC *pSspec, FILE *out, int formatFlag,
		       int nCol, int colWidth);
long sydSampleSetWriteSSF(SYD_SPEC *pSspec, FILE *pFile, char *reqDesc,
			  char *sampDesc);
long sydSetAttr(SYD_SPEC *pSspec, SYD_ATTR attr, int value, void *pArg);
long sydTrigger(SYD_SPEC *pSspec);
long sydTriggerClose (SYD_SPEC *pSspec);
long sydTriggerAddFromText (SYD_SPEC *pSspec, char *text);
int sydTsRound (TS_STAMP *pStamp, int roundNsec);

long sydOpenCA(SYD_SPEC **ppSspec, void *pHandle);
long sydOpenCF(SYD_SPEC **ppSspec, char *filePath);
long sydOpenPFO(SYD_SPEC **ppSspec, void *pHandle);
long sydOpenSSF(SYD_SPEC **ppSspec, char *filePath);
void sydInputMarkAsSampled(SYD_SPEC *pSspec);
long sydInputFetch();
long sydChanConnHandler();
#endif
