#ifndef INC_sydPlotDefs_h
#define INC_sydPlotDefs_h
/*	@(#)sydPlotDefs.h	1.3 9/23/91
 *	Author:	Roger A. Cole
 *	Date:	12-04-90
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .00	12-04-90	rac	initial version
 * .01	06-19-91	rac	installed in SCCS
 * .02	09-19-91	rac	always compile for X11, never SunView
 * .03	09-23-91	rac	undefine XWINDOWS from `make'
 *
 * include options
 *	-DSYD_PLOT_PRIVATE compiles private structures
 *	-DXWINDOWS	compiles X11 structure items, etc.
 */
/*+/mod***********************************************************************
* TITLE	sydPlotDefs.h - definitions for synchronous data plotting routines
*
* DESCRIPTION
*
*-***************************************************************************/

#undef XWINDOWS
#define XWINDOWS

#include <genDefs.h>
#ifndef INC_sydDefs_h
#   include <sydDefs.h>
#endif
#ifndef INC_pprPlotDefs_h
#   include <pprPlotDefs.h>
#endif
#ifdef XWINDOWS
#   include <X11/Xlib.h>
#endif


/*/subhead SYD_PL_xxx----------------------------------------------------------
* plot attributes
*	SYD_PLATTER_FG1 & 2 are used for Smith Chart axes
* plot type
*----------------------------------------------------------------------------*/

typedef enum sydPlAttr {
    SYD_PLATTR_LINE,	/* plot with lines */
    SYD_PLATTR_POINT,	/* plot with points */
    SYD_PLATTR_MARK,	/* plot with marks */
    SYD_PLATTR_SHOW,	/* show point status on plot */
    SYD_PLATTR_UNDER,	/* "fill under" plotting */
    SYD_PLATTR_XLAB,	/* do label for x axis */
    SYD_PLATTR_XANN,	/* do annotation for x axis */
    SYD_PLATTR_YLAB,	/* do label for y axis */
    SYD_PLATTR_YANN,	/* do annotation for y axis */
    SYD_PLATTR_MONO,	/* plotting medium is monochrome */
    SYD_PLATTR_BG,	/* set background pixel value */
    SYD_PLATTR_FG,	/* set foreground pixel value */
    SYD_PLATTR_FG1,	/* set first alternate fg pixel value */
    SYD_PLATTR_FG2,	/* set second alternate fg pixel value */
    SYD_PLATTR_XCHAN,	/* channel is an X-axis channel */
    SYD_PLATTR_WRAP,	/* wrap-around if value exceeds extent */
} SYD_PLATTR;

#ifndef SYD_PLOT_PRIVATE
    extern char *glArPltyText;
#else
    char *glArPltyText[] = {
       /* SYD_PLAX_UNDEF     */	"undefined",
       /* SYD_PLAX_SMITH_IMP */	"Smith chart, impedance",
       /* SYD_PLAX_SMITH_ADM */	"Smith chart, admittance",
       /* SYD_PLAX_SMITH_IMM */	"Smith chart, immitance",
       /* SYD_PLAX_TY        */	"time vs y",
       /* SYD_PLAX_TYY       */	"time vs multiple y",
       /* SYD_PLAX_XY        */	"x vs y",
       /* SYD_PLAX_XYY       */	"x vs multiple y",
       /* SYD_PLAX_Y         */	"y, uniform spacing",
       /* SYD_PLAX_YY        */	"multiple y, uniform spacing",
       /* SYD_PLAX_PAST      */	"illegal"
    };
#endif

typedef enum sydPlType {
    SYD_PLAX_UNDEF,	/* MUST BE ZERO!! */
    SYD_PLAX_SMITH_IMP,
    SYD_PLAX_SMITH_ADM,
    SYD_PLAX_SMITH_IMM,
    SYD_PLAX_TY,
    SYD_PLAX_TYY,
    SYD_PLAX_XY,
    SYD_PLAX_XYY,
    SYD_PLAX_Y,
    SYD_PLAX_YY,
    SYD_PLAX_PAST	/* last legal + 1 */
} SYD_PLAX;

/*/subhead plot_slave_struct---------------------------------------------------
*
*----------------------------------------------------------------------------*/
typedef struct sydPlSlave {
    struct sydPlSlave *pPrev;   /* pointer to previous slave */
    struct sydPlSlave *pNext;   /* pointer to next slave */
    PPR_AREA	*pArea;		/* ptr to area structure for pprXxx routines */
    SYD_CHAN	*pSChan;	/* pointer to synchronous channel struct */
    double	extentVal;	/* value at end of axis */
    double	originVal;	/* value at origin of axis */
    int		nInt;		/* number of intervals for axis */
    int		lineKey;	/* line key number for slave */
    int		markNum;	/* mark number for slave */
    char	timeLabel[120];	/* label using chanName and start stamp */
    short	xChan;		/* 1 indicates an x-axis channel */
    char	*pAnnot[db_state_dim];/* pointers to annotations (for ENUM) */
    char	**ppAnnot;	/* NULL or pointer to annotation array */
    unsigned long bg;		/* X11 background pixel value, or 0 */
    unsigned long fg;		/* X11 foreground pixel value, or 0 */
    double	xFracLeft;	/* window fraction at left x data value */
    double	xFracRight;	/* window fraction at right x data value */
    double	yFracBot;	/* window fraction at bottom y data value */
    double	yFracTop;	/* window fraction at top y data value */
    short	first;
    short	skip;
    double	oldX;
    double	oldY;
} SYD_PL_SLAVE;

/*/subhead plot_master_struct--------------------------------------------------
* SYD_PL_MSTR
*
*	A plot master controls the plotting of data for one or more
*	channels.  It contains:
*
*	o   a list of plot slaves, one slave per channel
*	o   a pointer to the retrieval set master; all slaves share this
*	o   some overall information about the master plot, such as its
*	    text title (not to be confused with the _window_ title)
*	o   some x axis information for the case where time is used as the
*	    x axis
*	o   "hidden" information about the window in which plotting occurs
*----------------------------------------------------------------------------*/

typedef struct sydPlMstr {
    PPR_WIN_TY	winType;	/* type of "window" being used */
    SYD_PLAX	plotAxis;	/* type of plot axes--one of SYD_PLAX_xxx */
    short	linePlot;	/* 1 for line plot (can be used with marks) */
    short	pointPlot;	/* 1 for point plot (can be used with lines) */
    short	markPlot;	/* 1 for mark plot */
    short	showStat;	/* 1 for show "alarm" status */
    short	fillUnder;	/* 1 for "fill under" plot */
    short	noColor;	/* 1 for monochrome plotting */
#ifdef XWINDOWS
    Display	*pDisp;		/* pointer to X11 display, or NULL */
    Window	window;		/* X11 window */
    unsigned long fg;		/* initial foreground pixel value for master */
    unsigned long bg;		/* initial background pixel value for master */
    unsigned long altPixel1;	/* alternate pixel value, or 0 */
    unsigned long altPixel2;	/* alternate pixel value, or 0 */
#endif
    short	useXlabel;	/* 1 to label x axis */
    short	useXannot;	/* 1 to annotate x axis */
    short	useYlabel;	/* 1 to label y axis */
    short	useYannot;	/* 1 to annotate y axis */
    SYD_SPEC	*pSspec;	/* pointer to synchronous set descriptor */
    PPR_WIN	*pWin;		/* ptr to window struct for pprXxx routines */
    char	title[120];	/* top title for plot */
    char	label[120];	/* label if time is used for x axis */
    char	lTitle[120];	/* left title for plot */
    char	bTitle[120];	/* bottom title for plot */
    char	rTitle[120];	/* right title for plot */
    SYD_PL_SLAVE *pHead;	/* pointer to head of list of slave plots */
    SYD_PL_SLAVE *pTail;	/* pointer to tail of list of slave plots */
    long        nSlaves;        /* number of slave plots */
    double	originVal;	/* elapsed time at origin of x axis */
    double	extentVal;	/* elapsed time at end of x axis */
    int		nInt;		/* number of intervals for x axis */
    short	wrapX;		/* wrap around if time exceeds extentVal */
    char	refText[28];	/* text for reference time stamp for x axis */
    int		x;		/* x pos of window on the display */
    int		y;		/* y pos of window on the display */
    int		width;		/* width of the window */
    int		height;		/* height of the window */
} SYD_PL_MSTR;

/*-----------------------------------------------------------------------------
* prototypes
*----------------------------------------------------------------------------*/
SYD_PL_SLAVE *sydPlotChanAdd();
void sydPlotAxisAutoRange();
void sydPlotAxisSetup();
long sydPlotAxisSetAttr();
long sydPlotDone();
long sydPlotInit();
long sydPlotInitUW();
long sydPlotSetAttr();
long sydPlotSetTitles();
long sydPlotWinLoop();
void sydPlot();
void sydPlotSamples();
void sydPlot_SmithPlot();
void sydPlot_SmithGrid();
void sydPlot_SmithSamples();
void sydPlot_TYPlot();
void sydPlot_TYGrid();
void sydPlot_TYSamples();
void sydPlot_TYYPlot();
void sydPlot_TYYGrid();
void sydPlot_XYPlot();
void sydPlot_XYGrid();
void sydPlot_XYSamples();
void sydPlot_XYarray();
void sydPlot_XYYPlot();
void sydPlot_XYYGrid();
void sydPlot_YPlot();
void sydPlot_YGrid();
void sydPlot_YSamples();
void sydPlot_Yarray();
void sydPlot_YYPlot();
void sydPlot_YYGrid();

#endif
