/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
/*	File:	drvArcom.h							       	
/*	Auth:	Walter E. Palmer
/*	Edit:	03/8/94        WEP: based on smod_driver
/*
/*	This file contains the data structures used by the ARCOM I/O driver.
/*                                                                             	      	*/

#define TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d

/*	number of lines of modbus 	*/
#define		MODLINES	32
/*	number of modbus plcs per line */
#define		NUMMODPLCS	70

/*                                                                           		
/*	The IOC will assume that the data to be exchanged with each PLC will           	
/*	come from a fixed number of coils, digital inputs, analog inputs               	
/*	or registers. It will read these in groups to minimize the amount of           	
/*	communications overhead. It will also try to build as large a fetch            	
/*	message as possible & required by remembering which consecutive groups         	
/*	have been requested by the SCAN task.                                          	
*/

#define MCOILS		  0			/* Card number for coils 		*/
#define MDINS		  1			/* Card number for digital inputs 	*/
#define MAINS		  3			/* Card number for analog inputs 	*/
#define MREGS		  4			/* Card number for registers 		*/
#define MPSEUDO		 33			/* Card number for pseudo i/o 		*/

/* Number of Groups of items */
#define NUMMODCOILS	 96			/* 96 groups of 8 coils 		*/
#define NUMMODDIS	 64			/* 64 groups of 8 digitals		*/
#define NUMMODAIS	 64			/* 64 groups of 4 analogs		*/
#define NUMREGISTERS	800			/* 64 groups of 4 registers		*/
#define BITBLOCK	  8			/* size of coil/digital groups	        */
#define REGBLOCK	  4			/* size of analog/register groups	*/

/* Number of items */
#define MINCOILNO       1                       /* min number of individual coil 	*/
#define MINDINO         1                       /* max number of individual digital 	*/
#define MINAINO         1                       /* max number of individual analog      */
#define MINREGNO        1                       /* max number of individual register	*/

#define MAXCOILNO	NUMMODCOILS*BITBLOCK	/* max number of individual coils 	*/
#define MAXDINO		NUMMODDIS*BITBLOCK	/* max number of individual digitals 	*/
#define MAXAINO		NUMMODAIS*REGBLOCK	/* max number of analogs 		*/
#define MAXREGNO	NUMREGISTERS*REGBLOCK	/* max number of registers 		*/

#define MAXDIGBLKSIZE	64			/* Maximum number of digital groups 	*/
						/*  to request in one packet        	*/
#define MAXREGBLKSIZE	16			/* Maximum number of digital groups 	*/
						/*  to request in one packet        	*/
/* Polled Report By Exception Declarations */

/* PRBE Mask Bit Definitions */
#define PRBEDIGITALBIT     1     /* 500 */      /* Digital Input Status Change          */
#define PRBEACCUMBIT       2     /* 501 */      /* Accumulator/Frequency Status Change  */
#define PRBEANALOGBIT      4     /* 502 */      /* Analog Input Deadband Change         */
#define PRBEHARTSTATUSBIT  8     /* 503 */      /* HART Transmitter STATUS Change       */
#define PRBEHARTPV1BIT    16     /* 504 */      /* HART PV1 Deadband Change             */
#define PRBESLLCOILBIT    32     /* 505 */      /* HART SLL Coil Change                 */
#define PRBESLLHOLDBIT    64     /* 506 */      /* HART SLL Holding Register Change     */
#define PRBESLLFLOATBIT  128     /* 507 */      /* HART SLL Floating Point Reg Change   */
#define PRBESLLLONGBIT   256     /* 508 */      /* HART SLL Long Register Change        */

/* PRBE Coil Definitions */
#define PRBELOWCOIL      500
#define PRBEHIGHCOIL     504
#define PRBECOILMASK     0x000003ff             /* only the first 10 bits are defined.  */ 
#define PRBEDIGITAL      500                    /* Digital Input Status Change          */
#define PRBEACCUM        501                    /* Accumulator/Frequency Status Change  */
#define PRBEANALOG       502                    /* Analog Input Deadband Change         */
#define PRBEHARTSTATUS   503                    /* HART Transmitter STATUS Change       */
#define PRBEHARTPV1      504                    /* HART PV1 Deadband Change             */
#define PRBESLLCOIL      505                    /* SLL Coils State Change               */
#define PRBESLLHOLD      506                    /* SLL Holding Register Exception       */
#define PRBESLLFLOAT     507                    /* SLL Floating Point Register Exception*/
#define PRBESLLLONG      508                    /* SLL Long Integer Register Exception  */
/*#define PRBEIOMSTATUS    509                    /* IOM STATUS Change                    */


/* PRBE Address Range Definitions */

/* I/O Address: 0,XXXX */
#define FIRSTCOIL1         1                    /*                                      */
#define LASTCOIL1        400                    /*                                      */
#define FIRSTSLLCOIL     401                    /*  505                                 */
#define LASTSLLCOIL      499                    /*  505                                 */
#define FIRSTCOIL2       500                    /*                                      */
#define LASTCOIL2        768                    /*                                      */

/* I/O Address: 1,XXXX */
#define FIRSTDIGITAL       1                    /*  500                                 */
#define LASTDIGITAL      256                    /*  500                                 */
#define FIRSTDIGLATCH    257                    /*  500                                 */
#define LASTDIGLATCH     512                    /*  500                                 */

/* I/O Address: 3,XXXX */
#define FIRSTANALOGIN      1                    /*  502                                 */
#define LASTANALOGIN      64                    /*  502                                 */
#define FIRSTACCUM        65                    /*  501                                 */
#define LASTACCUM         72                    /*  501                                 */
#define FIRSTFREQUENCY    73                    /*  501                                 */
#define LASTFREQUENCY     76                    /*  501                                 */
#define FIRSTHARTSTATUS   77                    /*  503                                 */
#define LASTHARTSTATUS   121                    /*  503                                 */
#define FIRSTAI1         122                    /*                                      */
#define LASTAI1          256                    /*                                      */

/* I/O Address: 4,XXXX */

#define FIRSTANALOGOUT     1                    /*                                      */
#define LASTANALOGOUT     16                    /*                                      */
#define FIRSTAO1          17                    /*                                      */

#define RTC_START         17

#define RTC_SECONDS       17
#define RTC_MINUTES       18
#define RTC_HOURS         19
#define RTC_DAYS          20
#define RTC_MONTHS        21
#define RTC_YEARS         22
#define RTC_HOLDFLAG      23
#define RTC_SETFLAG       24

#define LASTAO1          100                    /*                                      */
#define FIRSTSLLHOLD     101                    /*  506                                 */
#define LASTSLLHOLD      200                    /*  506                                 */
#define FIRSTAO2         201                    /*                                      */
#define LASTAO2         2048                    /*                                      */

#define FIRSTHARTFPHR   2049                    /*                                      */
#define LASTHARTFPHR    2487                    /*                                      */

#define FIRSTHARTFLOAT  2489                    /*  504                                 */
#define LASTHARTFLOAT   2517                    /*  504                                 */
#define FIRSTSLLFLOAT   2519                    /*  507                                 */
#define LASTSLLFLOAT    2549                    /*  507                                 */
#define FIRSTGENFLOAT   2551                    /*                                      */
#define LASTGENFLOAT    3071                    /*                                      */

#define FIRSTSLLLONG    3073                    /*  508                                 */
#define LASTSLLLONG     3099                    /*  508                                 */
#define FIRSTGENLONG    3101                    /*                                      */
#define LASTGENLONG     3199                    /*                                      */



/*
/*	Conversion types for analog inputs	*/
/*
*/
#define		AICNVINTEGER	10
#define		AICNVDOUBLEINT	11
#define		AICNVFLOAT	12
#define		AICNVDOUBLEFLT	13
#define		AICNVASCIIFLT	14
#define		AICNVSTATUS	15
#define		AICNVTIME	16
#define		AICNVASCII_2	17
#define		AICNVASCII_4	18
#define		AICNVASCII_12	19
#define		AICNVASCII_20	20
#define		MAXCONV		20

/* The following must match the definition in recChoice.ascii */
#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2



/*
/*	Communications I/O Structures
/*
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	256
#define		MAXTXMSG	256


typedef	struct {
	unsigned short		rxstate;
	unsigned short		fixedlen;
	unsigned short		rxmsglenexpected;
        unsigned short          rxmsglen;
       	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
        unsigned char           txaddress;
        unsigned char           txfunction;
       	unsigned char		rxbuf[MAXRXMSG];
		} 	RXLINE;


typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;
/*
/*	Communications I/O Constants
/*
*/

#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		60	/* 16 ms ticks ( one second)            */
#define		ROMTIMEOUT		80*60	/* 16 ms ticks (80 seconds)             */
#define		ACTIVITYTIME		300	/* must see activity in 5 seconds       */

#define		VARRESP			FALSE
#define		FIXEDRESP		TRUE
#define		ONERETRY		0x8000	/* OR'ed into line number for xfermsg 	*/
#define		RXIGNORE		0
#define		RXADRST			1
#define		RXFNCST			2
#define		RXLENST			3
#define		RXDATAST		4

#define		TXGOOD			1
#define		TXBAD			2

#define		TXTIMEOUT		1255	/* 16 ms ticks: just under 21 seconds 	*/ 

#define		IO_TIMEOUT		3	/* Code returned by timer 		*/

#define		MAXRETRYS		3	/* retransmission trys 			*/

#define		MAXCFGCK		8	/* Try to Bring online each cycle 	*/
#define		BRINGBACKCNT		4	/* Poll loops per cycle 		*/

#define		XFERGOOD		1
#define		XFERBAD			2

#define		RESNAMELEN		16	/* length of a resource name 		*/
#define		SLWWINDOW		900	/* # of ticks to allow slave to try 	*/


/* Periodic Scan, Demand Scan, and Time Sync default parameters */

#define     SCANPERIOD            2*60*60 /*    5*60*60   /* Periodic Scan Interval: 5 minutes    */
#define     DEMANDPERIODDEFAULT   1*60*60 /*   10*60*60   /* Demand Scan Duration: 10 minutes     */
#define     DEMANDINTERVALDEFAULT    2*60 /*      10*60   /* Demand Scan Interval: 10 seconds     */
#define     SYNC_DEFAULT          5*60*60 /*24*60*60*60   /* IOC/RTU Time Sync Interval: 24 hours */



/*
/*	PLC PROGRAM LOAD STATES
*/
#define		PLC_NOT_LOADING		0
#define		PLC_LOAD_REQUESTED	1
#define		PLC_LOAD_IN_PROGRESS	2
#define		PLC_LOAD_FAILED		3
#define		PLC_LOAD_SUCCESSFUL	4

/*
/* Output Queue Management Structures for Modbus Protocol
/*
*/
#define         AOUTTYPE        1
#define         BOUTTYPE        2
#define         SOUTTYPE        3

#define         OUTQSIZE        128
#define		STRINGMSGLEN	80


typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	short           aovalue;
	double		dblaovalue;
	unsigned short	conversion;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bomask;
	long            bovalue;
		} BOUT, *BOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  sosignal;
	long            solen;
	unsigned char   sovalue[STRINGMSGLEN];
		} SOUT, *SOUTPTR;

typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			SOUT	sonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque[OUTQSIZE];
		} OUTPUTQUE;
/*
/*	Internal IOC Structure for MODICON PLC
/*
*/
typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned long		last_read;
		}	MODCOILS;

typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned long		last_read;
		}	MODDIS;

typedef struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		ioregs[REGBLOCK];
		}	REGISTERS;

typedef struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	short			aivals[REGBLOCK];
		}	MODAIS;

typedef	struct	{
	unsigned short  was_requested;		/* 0  The scan task has accesses this RTU 	*/
	unsigned short  is_responding;		/* 1  The RTU is responding 			*/
	unsigned short  holdflag;		/*    The RTU data is in hold 			*/
	unsigned short  loading;		/*    Program load Status: TRUE/FALSE           */
	unsigned short  prog_load_stat;		/* 2  Program load Status: multiple states      */
	unsigned short  prog_load_page;		/* 3  Program Load Block Number 		*/
	unsigned short  prog_load_number;	/* 4  Program Number to Load 			*/
	unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs 			*/
	unsigned short  no_datain_msgs;		/* 6  Count Data Requests 			*/
	unsigned short  no_rxnoresp;		/* 7  Count No Responses 			*/
	unsigned short  no_txnoresp;		/* 8  Count No Responses 			*/
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors 		*/
	unsigned short  no_retries;		/* 10 Count Retries 				*/
	unsigned short  no_rxhwerrs;		/* 11 Count Comm Hardware Errors 		*/
	unsigned short  no_txhwerrs;		/* 12 Count Comm Hardware Errors 		*/
	unsigned short  no_crcerrs;		/* 13 Count CRC Errors 				*/
	unsigned short  no_offlines;		/* 14 Count Off Line Transitions		*/
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats			*/
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline 			*/
	unsigned short  badoutcnt;		/* 17 Count Failed output commands 		*/
						/* 18 Back up Mode 1 = Auto 0 = Man		*/
						/* 19 Active Channel 1= Priamry 0 = Backup 	*/
	unsigned short	prog_load_code;		/* 20 Program Load Error Code 			*/
	unsigned short	plc_start;		/* 21 Command the PLC to Start 			*/
	unsigned short	plc_stop;		/* 22 Command the PLC to Stop 			*/

	unsigned short	prog_load_interval;     /* 24 Program Load Interval 			*/
	unsigned short	prog_load_window;       /* 25 Program Load Window			*/

        unsigned short  do_timesync;            /* 0 = no action, 1 = send timesync to RTU.     */

	unsigned short  moretoscan;		/* 0 = scan completed. 0 <> more to scan.	*/
        unsigned short  demandscan;             /* 0 = normal/periodic scan, 1 = demand scan.   */
        unsigned long   lastscantime;           /* holds getTick value from last scan attempt.  */
        unsigned long   demandstarttime;        /* holds getTick value from beginning of demand.*/ 

	MODCOILS	coils[NUMMODCOILS];
	MODDIS		dis[NUMMODDIS];
	MODAIS		ais[NUMMODAIS];
	REGISTERS	regs[NUMREGISTERS];
	}	MODPLC;
/*
/* MODBUS Protocol function codes
*/

#define MODCOILREAD	0x01	/* READ COIL STATUS: (ON/OFF) read a group of logic coils          */
#define MODDIREAD	0x02	/* READ INPUT STATUS: (ON/OFF) of a group of discrete inputs       */
#define MODREGREAD	0x03	/* READ HOLDING REGISTERS: get current binary value in one or more */
                                /* holding registers                                               */
#define MODAIREAD	0x04	/* READ INPUT REGISTERS: get current binary value in one or more   */
                                /* input registers                                                 */

#define MODCOILWRITE	0x05	/* FORCE SINGLE COIL: force logic coil to a state of ON or OFF     */
#define MODREGWRITE	0x06	/* PRESET SINGLE REGISTER: place a specific binary value into a    */
                                /* holding register                                                */

#define MODLOOPBACK	0x08	/* LOOPBACK DIAGNOSTIC TEST */

#define MODMCOILWRITE	0x0F	/* FORCE MULTIPLE COILS: force series of consecutive logic coils   */
                                /* to defined ON/OFF states                                        */
#define MODMREGWRITE	0x10	/* PRESET MULTIPLE REGISTERS: put binary values into consecutive   */
                                /* series of holding regs                                          */

#define FNCERRBIT	0x80	/* Modbus error function code */


/* Various message lengths used in message transmission and response validation */

#define MODWRITEHDRLEN		0x06	/* wep: not currently used */
#define MODREADMSGLEN		0x06
#define MODSETCOILLEN		0x06
#define MODSETREGLEN		0x06

#define MODSETTIMELEN           0x13    /* 6 registers are used to define and control time.*/

#define MODSET2REGLEN		0x0b
#define MODMSETCOILLEN		0x07	/* wep: not currently used */
#define MODMSETCOILRESPLEN	0x06
#define MODCRCLEN		0x02
#define MODLOOPBACKLEN		0x06
#define MODLOOPBACKRESPLEN	0x08	/* Including CRC */
#define MODWRITERESPLEN		0x08	/* Including CRC */

#define MODSETMREGRESPLEN	0x06
#define NAKMSGLEN		0x03	/* bytes left get */

/*
/* MODBUS Exception Codes
*/

#define MODEXCEPTFNC		0x01	/* wep: not used */
#define MODEXCEPTADR		0x02	/* "    "   "    */
#define MODEXCEPTDATA		0x03	/* "    "   "    */
#define MODEXCEPSLAVE		0x04	/* "    "   "    */
#define MODEXCEPTSACK		0x05	/* "    "   "    */
#define MODEXCEPTBUSY		0x06	/* "    "   "    */


/* task definitions */

#define MOD_STACK		8000   /* Stack size              */
#define	MOD_PRIORITY		152    /* default priority        */
#define	MOD_TASK_OPTIONS	8      /* Floating point          */




/*
/* Mutual-Exclusion Control for diagnostic/data radio channel sharing */
/*
*/

#define		MUXTIMEOUT	60*30	/* 16.67 ms ticks (60Hz): 30 seconds 	*/

/* Loader Header Information */


/*	Constants	*/


/*	File Offsets and PLC Memory Offsets	*/
/*#define PAGE0			71 */
/*#define PAGEF			80 */
#define ASCII			144
/*#define FUTURE			160 */
/*#define DXCFG			98 */

/*#define NUMDUMPAREAS		4	/* Number of Dump Blocks 				*/
/*#define DUMPAREASIZE		18	/* Number of Bytes in a dump block 			*/
/*#define AREASPERBLK		4	/* Number of areas to dump per block 			*/

/*#define CONFIGAREA		0	/* Identifier of Configuration Area 			*/

/*#define LOGICAREATYPE		0x08	/* Bit mask in specifier to indicate logic		*/
/*#define ASCIIAREATYPE		0x80	/* Bit mask in specifier to indicate ascii msgs 	*/
/*#define MEMORYAREATYPE		0x00	/* Bit mask in specifier to indicate memory 		*/
/*#define AREATYPEMASK		0x88	/* Or of Above 						*/
/*#define MACHINESTATEADDR	101	/* PLC Memory Address of Running and Memory Protect 	*/
/*#define RUNNINGSTATE		0x40	/* Value of Running controller 				*/
/*#define MEMPROTECT		0x20	/* Value of Memory Protected Controller 		*/
/*#define RUNLOADSWITCHADDR	110	/* PLC Memory Addres of the Run Load Switch Value 	*/
/*#define LOADSWITCH		0x02	/* Switch Value for LOAD MODE 				*/
/*#define RUNSWITCH		0x01	/* Switch Value for RUN MODE 				*/

/*	File Related Constants */
/*#define EOLNODE			5	/* End of logic node type 				*/
/*#define BOCNODE			4	/* Begining of column codes are less this 		*/ 
/*#define NODELEN			3	/* Length of a logic node 				*/
/*#define WORDLEN			2	/* Length of a PLC memory word 				*/

/* Message Related Constants */
/*#define MAXMESSAGELEN		50	/* Greater than any message we will send 		*/
/*#define NOSTARTNODEMSGLEN	15	/* Maximum size of logic load msg without start found 	*/
/*#define MAXMSGNODES		11	/* Maximum size of logic load msg 			*/
/*#define MAXWORDMSGLEN		32	/* Maximum size of memory load message 			*/

/*#define MODCMDSTARTED		0x05	/* Code for Command Started (must poll for done) 	*/
/*#define MODCMDDONE		0x06	/* Code for Command Completed 				*/
/*#define MAXPOLLSFORDONE		0x04	/* Maximum polls to get command done 			*/
/*#define MODLOAD			0x0D	/* LOAD Function Code 					*/
/*#define MODPOLL			0x0E	/* POLL Function Code 					*/
/*#define MODSFSTOP		0x01	/* STOP Subfunction Code 				*/
/*#define MODSFSTART		0x02	/* START Subfunction Code 				*/
/*#define MODSFREADMEM		0x03	/* READ MEMORY Subfunction Code 			*/
/*#define MODSFWRITEMEM		0x04	/* WRITE MEMORY Subfunction Code 			*/
/*#define MODSFWRITENODE		0x07	/* WRITE NODE Subfunction Code 				*/
/*#define MODSFLOGIN		0x19	/* LOGIN Subfunction Code 				*/
/*#define MODSFLOGOUT		0x1A	/* LOGOUT Subfunction Code 				*/
/*#define MODSFINITIALIZE		0x2B	/* INITIALIZE Subfunction Code 				*/
/*#define MODSFMYSTERY1		0x31	/* MYSTERY1 Subfunction Code 				*/
/*#define MODSFMYSTERY2		0x34	/* MYSTERY2 Subfunction Code 				*/
/*#define MODSFENTERDIM		0x35	/* INVOKE DIM Subfunction Code 				*/
/*#define MODSFCHKDIM		0x36	/* CONFIRM DIM Subfunction Code 			*/
/*#define MODSFWRITEASCII		0x39	/* WRITE ASCII MSG Subfunction Code 			*/

#define DIAG_CMD8               0x08    /* diagnostic command 8 */
#define SUBCMD_BRK              0x40    /* Sucommand 64: Break Point */
#define SUBCMD_RUN              0x42    /* Sucommand 66: Continuous Execution */
#define SUBCMD_PGM              0x43    /* Sucommand 67: Halt and Receive New Program */
#define SUBCMD_ROM              0x46    /* Sucommand 70: Burn SLL Programs into FLASH ROM */

#define RTUDIAGCMDRESPLEN       0x08    /* Including CRC */

#define SENDINGPROGRAMLOGIC     0x00    /* sending program logic */
#define SENDINGPROGRAMNAME      0x01    /* sending program name */

#define PROGNAMELENGTH          0x0c    /* filename length 8 body, 1 space, 3 extension. */


/* Termination Error Codes */

#define   LOADOK                0x00
#define   CANNOTOPENFILE        0x01
#define   CANNOTSEEKTO          0x02
#define   CANNOTHALTRTU         0x03
#define   BOOLEANLINETOOLONG    0x04
#define   CANNOTSETUPLOADMODE   0x05
#define   CANNOTSENDRUNGDATA    0x06
#define   CANNOTSENDNAME        0x07
#define   CANNOTSENDSTART       0x08
#define   CANNOTSENDSAVE        0x09


/*#define	CANNOTREAD		0x03 */
/*#define	CANNOTLOGOUT		0x04 */
/*#define	CANNOTLOGIN		0x05 */
/*#define	MEMORYPROTECT		0x06 */
/*#define	PLCWONTSTOP		0x07 */
/*#define	PLCWONTGETDIM		0x08 */
/*#define	PLCWONTSETDIM		0x09 */
/*#define	PLCWONTBEDIM		0x0A */
/*#define	PLCWONTSTART		0x0B */
/*#define	PLCWONTINIT		0x0C */
/*#define	CANTGETMODE		0x0D */
/*#define	CANTSETLOADMODE		0x0E */
/*#define	UNEXPECTDEOF		0x0F */
/*#define	LOGICWRITEERROR		0x10 */
/*#define	MEMORYWRITEERROR	0x11 */
/*#define	CANTSETRUNMODE		0x12 */
/*#define	CANTSETLOADEDMODE	0x13 */



/******************************************/
/* ladder logic tokens for MODBUS download*/
/******************************************/

#define  OR_C          -1
#define  AND_C         -2
#define  OUTPUT_C      -3 
#define  ROOT_C        -4 
#define  ADD_C         - 5
#define  SUBTRACT_C     -6
#define  MULTIPLY_C     -7
#define  DIVIDE_C       -8
#define  SKIP_RUNGS     -9
#define  GREATER_C      -11
#define  LESS_C         -12
#define  EQUAL_C        -13
#define  INVERT_C       -19
#define  CONSTANT_C     -20
#define  BREAK_POINT    -48
#define  END_RUNG       -49      /* Arcom eliminated the END_RUNG token on 2-20-92. */
#define  END_CIRCUIT    -50


#define MAXPATH         80

/*	Structure Definitions */
/*typedef struct {
	unsigned short		length;
	unsigned short		strtaddr;
		}	AREADESC;

typedef struct {
	unsigned char	areaspec;
	unsigned char	page;
	AREADESC	areas[AREASPERBLK];
		}	DUMPAREA;
*/
typedef struct {
	unsigned short	line;
	unsigned short	loading;
	MODPLC		*pmod;
	unsigned short	mod;
	FILE		*loadfile;
	long		fileaddr;
	short		rlmode;
	short		machstate;
	unsigned char	filename[30];

	unsigned short	ldblock;
	unsigned short	ldarea;
        unsigned short  rung;
        unsigned short  length;
        unsigned int    pgmoffset;

        unsigned short  loadinterval;        /* length of time between load actions. */
        unsigned short  loadperiod;          /* length of time to load in one shot.  */
        unsigned long   lastloadaction;      /* time last load action was taken.     */

/*	DUMPAREA	dadesc[NUMDUMPAREAS]; */
		} 	LOADINFO;

/* Trace Bit Definitions */

#define ALLTRAFFICTRACE      0x0001
#define DATALINKTRACE        0x0002
#define NETWORKTRACE         0x0004
#define APPLICATIONTRACE     0x0008
#define CONFIGURATIONTRACE   0x0010        /* trace configuration errors in driver calls */
#define RUNTIMEERRORTRACE    0x0020        /* trace run time errors in driver calls */
#define CALLREJECTTRACE      0x0040        /* trace non-error based driver call rejections */
#define TIMESYNCTRACE        0x0080
/*	End of File:	drvArcom.h */
