/*drvBendix.h,v 1.1.1.1 1999/08/03 14:41:54 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/*
/*	File:	drvbendix.h
/*	Auth:	John L. Dalesio
/*	Edit:	4/24/93	JLD
/*
/*	This file contains the data structures used by the Bendix LMP
/*	I/O driver.
/*
*/

#define	TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d
#define LF	0x0a
#define EXPCHAR	'%'

/*	number of lines of LMP devices */
#define		LMPLINES	18
/*	number of LMP devices per line */	
#define		NUMLMPDEVS	1

/* Define the operating modes for the Bendix driver */
#define		OPMODELMP	0
#define		OPMODEASCII	1
/* Down Converter */
#define		OPMODEMIMIC1	2
/* LNA */
#define		OPMODEMIMIC2	3
#define		OPMODEHPA	4
/* Up Converter */
#define		OPMODEMIMIC3	5





/*	segment number for directive responses	*/
#define		DIRSEG		256
/*	segment number for directive responses	*/
#define		DIRCODESEG	257
/*	segment number for asynchronous responses	*/
#define		ASYNSEG		258
/*	segment number for asynchronous responses	*/
#define		SCROLLSEG	275
/*	segment number for ascii device values	*/
#define		ASCIISEG	300
/*	segment number for ascii time representation */
#define		TIMESEG		301
/*	segment number for ascii time representation */
#define		ASCIICNVSEG		302
/*	segment number for mimic 1 values	*/
#define		MIMICUDCSEG	400
/*	segment number for mimic 2 values	*/
#define		MIMICLNASEG	410
/*	segment number for mimic 3 values	*/
#define		MIMICUPSEG	420
/*   segment number for high power amp values	*/
#define		HPASEG		500
/*   segment number for high power amp values	*/
#define		HPASEG2		510
/*	segment number for pseudo values	*/
#define		PSEUDO		999






/*	Maximum segment size			*/
#define		MAXSEGSIZE	1024

/*	Conversion types for analog inputs	*/
#define		LMPCNVINTEGER	10
#define		LMPCNVDOUBLEINT	11
#define		LMPCNVFLOAT	12
#define		LMPCNVDOUBLEFLT	13
#define		LMPCNVASCIIFLT	14
#define		LMPCNVSTATUS	15
#define		LMPCNVTIME	16
#define		LMPCNVASCII_2	17
#define		LMPCNVASCII_4	18
#define		LMPCNVASCII_12	19
#define		LMPCNVASCII_20	20
#define		MAXCONV		20


/*
/*	Communications I/O Structures
/*
*/

#define		MAXRXMSG	1500
#define		RXSEM		0
#define		TXSEM		1

typedef	struct {
	unsigned short		protocol;
	unsigned short		terminator;
	unsigned short		rxstate;
	unsigned short		rxtocnt;
	unsigned short		rxhwerrcnt;
	unsigned short		fixedlen;
	unsigned short		rxmsglen;
	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		rxbuf[MAXRXMSG];
		} 	RXLINE;


#define		MAXTXMSG	100	

typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;

/*
/*	Communications I/O Constants
/*
*/

#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		50 	/* 16 ms ticks */

#define		VARRESP			FALSE
#define		FIXEDRESP		TRUE
#define		ONERETRY		0x8000	/* ored into line number for xfermsg */
#define		RXIGNORE		0
#define		RXSOHST			1
#define		RXADRST			2
#define		RXLEN1ST		3
#define		RXLEN2ST		4
#define		RXDATAST		5
#define		RXCRC1ST		6
#define		RXCRC2ST		7

#define		TXGOOD			1
#define		TXBAD			2

#define		TXTIMEOUT		125	/* 16 ms ticks */

#define		IO_TIMEOUT		3	/* Code returned by timer */

#define		MAXRETRYS		3	/* retransmission trys */

#define		MAXCFGCK		8	/* Try to Bring online each cycle */
#define		BRINGBACKCNT		4	/* Poll loops per cycle */

#define		XFERGOOD		1
#define		XFERBAD			2

#define		LMPSOH			1	/* Start of messsage character */
#define		RESNAMELEN		16	/* length of a resource name */
#define		SLWWINDOW		900	/* # of ticks to allow slave to try */
/*
/*	PLC PROGRAM LOAD STATES
*/
#define		PLC_NOT_LOADED		0
#define		PLC_LOAD_REQUESTED	1
#define		PLC_LOAD_IN_PROGRESS	2
#define		PLC_LOAD_FAILED		3
#define		PLC_LOAD_SUCCESSFUL	4



/*
* Output Que Management Structures for Modbus Protocol
*/
#define         AOUTTYPE        1
#define         BOUTTYPE        2
#define         SOUTTYPE        3

#define         OUTQSIZE        100
#define		LMPMSGLEN	80


typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	unsigned int    aovalue;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	unsigned long   bomask;
	unsigned long	bobvalue;
	unsigned char   bovalue[LMPMSGLEN];
		} BOUT, *BOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  sosignal;
	long            solen;
	unsigned char   sovalue[LMPMSGLEN];
		} SOUT, *SOUTPTR;

typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			SOUT	sonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque[OUTQSIZE];
		} OUTPUTQUE;



/*
/* Data Structures for  Storage for a Bendix LMP Device
/*
*/

/*	Size definitions			    */

#define RESPSEQQLEN	32
#define RESBUFLEN	128
#define MAXSEGS		128
#define MAXASYNEVENTS	64
#define MAXSCROLL	16
#define MAXASCEVENTS	64
#define MAXASCBITOFF	512
#define MAXHPASTATBITS	64
#define MAXHPAOPPARAMS	32
#define MAXHPAHWLIMITS	16
#define MAXHPASWLLIMITS	32
#define MAXHPASWHLIMITS	32
#define MAXHPACHCHGFREQ	24
#define WFDATASIZE	601

/*	Signal Number Definitions for HPA Interface */

#define	SHPAFLTCNT	0
#define	SHPAALMCNT	1
#define	SHPAFTD		2
#define	SHPAOPPARMS	3
#define SHPAHWLIMS	50
#define	SHPASWLLIMS	70
#define	SHPASWHLIMS	110
#define SHPANUMCHAN	150
#define SHPACURCHN	151
#define	SHPAFREQS	152
#define MAXHPAAI	(SHPAFREQS+MAXHPACHCHGFREQ-1)

#define SHPARESETBO	99

/* Structure Element Definitions */

typedef struct {
		unsigned short	dirseqno;
		unsigned short  responseix;
		} RESPSEQQENT;

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned short	respstrlen;
		unsigned short	response_code;
		unsigned char   response[LMPMSGLEN+1];
		} RESPBUFENT;

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned short	asynstrlen;
		unsigned short	asyncmsg_code;
		unsigned char   asyncmsg[LMPMSGLEN+1];
		} ASYNCEVENT;

#define MAXM1AISIGNAL	1
#define MAXM1BISIGNAL	7

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned short	mstrlen;
		short		atten;
		long		status;
		} MIMIC1STAT;

#define MAXM2BISIGNAL	7
typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		long		status;
		} MIMIC2STAT;

#define MAXM3AISIGNAL	2
#define MAXM3BISIGNAL	7
typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		short		atten[2];
		long		status;
		} MIMIC3STAT;

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned short	astrlen;
		unsigned char   adatabuf[LMPMSGLEN+1];
		} ASCIIDATA;

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned char   statbits[((MAXHPASTATBITS+7)/8)];
		unsigned char	fltcnt;
		unsigned char	almcnt;
		unsigned short	ftd;
		unsigned short	opparams[MAXHPAOPPARAMS];
		} HPASTATUSDATA;

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned short	limits[MAXHPAHWLIMITS];
		} HPAHWLIMITDATA;

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned short	limits[MAXHPASWLLIMITS];
		} HPASWLLIMITDATA;

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned short	limits[MAXHPASWHLIMITS];
		} HPASWHLIMITDATA;

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned char	numchans;
		unsigned char	curchan;
		float		freqs[MAXHPACHCHGFREQ];
		} HPACHNCHGFREQ;

typedef struct {
	unsigned short	is_responding;
	HPASTATUSDATA	hpastat;
	HPAHWLIMITDATA	hpahwlimits;
	HPASWLLIMITDATA	hpaswllimits;
	HPASWHLIMITDATA	hpaswhlimits;
	HPACHNCHGFREQ	hpachnfreq;
		} HPADEF;


typedef struct {
		unsigned short	was_logged;
		unsigned char   databuf[LMPMSGLEN+1];
		} SCROLLENT;

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned short	size;
		unsigned short	has_descriptor;
		unsigned short	force_fetch;
		unsigned short	fetchfreq;
		long		last_time_fetched;
		unsigned char	*segdatabuf;
		} SEGMENT;

typedef struct {
		unsigned short	was_requested;
		unsigned short	data_valid;
		unsigned short	was_read;
		unsigned short	data[WFDATASIZE];
		} WFDATA;


typedef	struct	{
	unsigned short	was_requested;		/* 0  The scan task has accesses this RTU */
	unsigned short	is_responding;		/* 1  The RTU is responding	*/
	unsigned short	holdflag;		/*    The RTU data is in hold */
	unsigned short	prog_load_stat;		/* 2  Program load Status */
	unsigned short  prog_load_page;		/* 3  Program Load Block Number */
	unsigned short  prog_load_number;	/* 4  Program Number to Load */
	unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs */
	unsigned short  no_datain_msgs;		/* 6  Count Data Requests */
	unsigned short  no_rxnoresp;		/* 7  Count No Responses */
	unsigned short  no_txnoresp;		/* 8  Count No Responses */
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors */
	unsigned short  no_retries;		/* 10 Count Retries */
	unsigned short  no_rxhwerrs;		/* 11 Count Comm Hardware Errors */
	unsigned short  no_txhwerrs;		/* 12 Count Comm Hardware Errors */
	unsigned short  no_crcerrs;		/* 13 Count CRC Errors */
	unsigned short  no_offlines;		/* 14 Count Off Line Transitions */
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats*/
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline */
	unsigned short  badoutcnt;		/* 17 Count Failed output commands */
						/* 18 Back up Mode 1 = Auto 0 = Man*/
						/* 19 Active Channel 1= Priamry 0 = Backup */
	unsigned short  dirseqno;		/* 20 Direcive Sequence Number */
	unsigned short  asyneventseqno;		/* 21 Asynchronous Event Sequence Number */
	unsigned short	asyneventcount;		/* 22 Number of Async Events Requested */
	unsigned short	haveasyncevent;		/*    LMP has an Async Event In Que */
	unsigned short	dirseqqueix;		/*    Current index into respseqp */
	RESPSEQQENT	respseqq[RESPSEQQLEN];	
	RESPBUFENT	responses[RESBUFLEN];
	ASYNCEVENT	asynevents[MAXASYNEVENTS];
	SEGMENT		segments[MAXSEGS];
	SCROLLENT	scrolllines[MAXSCROLL];
	MIMIC1STAT	mimic1;
	MIMIC2STAT	mimic2;
	MIMIC3STAT	mimic3;
	ASCIIDATA	ascmsg[MAXASCEVENTS];
	HPADEF		hpa[2];
	WFDATA		wf;

	}	LMPDEV;

/*
/* LMP Protocol Definitions
/*
*/
#define 	LMPSOM		0x01
#define 	DIR		0x12
#define		DIRRESP		0x13
#define 	SEGREQ		0x18
#define		SEGREQLEN	10
#define 	SEG		0x19
#define 	ASYNCREQ	0x17
#define		ASYNCREQLEN	10
#define 	ASYNC		0x16
#define		DIRRESPOVERHEAD	10
#define		SEGRESPOVERHEAD	10
#define		ASYNRESPOVERHEAD	10

/*
/* Message offsets
*/
/*	Header */

#define 	SOMSGIX		0x00
#define 	ADDRIX		0x01
#define 	LENIX		0x02
#define 	DATAIX		0x04

#define		DIRSEQIX	0x04
#define		MSGIDIX		0x05
#define		DIRDATAIX	0x06

#define		MAXDIRDATALEN	80

#define		DIRRESPCODEIX	0x06
#define		DIRRESPDATAIX	0x08

#define		MAXDIRRESPDATA	64

#define		REQSEGNOIX	0x07

#define		SEGNUMIX	0x06
#define		SEGLENIXLOW	0x07
#define		SEGDATAIX	0x08

#define		MAXSEGDATA	80

#define		ASYNCLASSIX	0x04
#define		ASYNSEQIX	0x07
#define		ASYNDATAIX	0x08
#define		EVENTIDIX	0x06
#define		EVENTDATAIX	0x08

#define		MAXEVENTDATA	80

/* Comm definitions for dumb ascii device */
/* String terminating modes */
#define		TNULL	0
#define		TCR	1
#define		TLF	2
#define		TCRLF	3
#define		T5D	4
#define		WFRESP	5
#define		ENDTIME	6
#define		ENDAACRLF 7
#define		TFLEN	10	

/* Comm definitions for mimic 1 device */
/* State Codes */
#define		MIMGETCLOSE	2
#define		MIMGETCSUM	3


/* Message Codes and response lengths */
#define		MIMICOPEN	'{'
#define		MIMICCLOSE	'}'
#define		MIMICADDR	'A'
#define		LNAADDR		'@'

#define		M1DECATTEN	'<'
#define		M1INCATTEN	'>'
#define		M1STATREQ	'A'
#define		M1MUTE		'M'
#define		M1UNMUTE	'U'
#define		M1FREQSET	'F'
#define		M1ATTENSET	'T'

#define		M1DECATTENTLEN	1
#define		M1INCATTENTLEN	1
#define		M1STATREQTLEN	1
#define		M1FREQSETTLEN	8
#define		M1ATTENSETTLEN	4

#define		M1DECATTENRLEN	1
#define		M1INCATTENRLEN	1
#define		M1STATREQRLEN	22
#define		M1FREQSETRLEN	1
#define		M1ATTENSETRLEN	1

/* Comm definitions for mimic 2 device */
/* Message Codes and response lengths */
#define		M2STATREQ	'A'
#define		M2ONOFFSET	'P'
#define		M2AUTOMANSET	'U'

#define		M2STATREQTLEN		1
#define		M2ONOFFSETTLEN		3
#define		M2AUTOMANSETTLEN	2

#define		M2STATREQRLEN		10
#define		M2ONOFFSETRLEN		1
#define		M2AUTOMANSETRLEN	1

/* Comm definitions for mimic 3 device */
/* Message Codes and response lengths */

#define		M3DECATTEN	'<'
#define		M3INCATTEN	'>'
#define		M3STATREQ	'A'
#define		M3MUTE		'M'
#define		M3UNMUTE	'U'
#define		M3FREQSET	'F'
#define		M3ATTENSET	'T'


/* Comm definitions for HPA device */
/* State Codes */
#define		HPAGETLEN	2
#define		HPAGETMSG	3
#define		HPAGETCSUM	4

#define		HPASOM		0XAA
#define		HPAONOFF	'H'
#define		HPAOFF		'0'
#define		HPAON		'1'
#define		HPADUMMY	'F'
#define		HPADUMMYLOAD	'D'
#define		HPAANT		'A'
#define		HPARESET	'R'
#define		HPARFPWR	'P'
#define		HPARFPWRINHIB	'I'
#define		HPARFPWRENA	'E'
#define		HPASELFTEST	'T'
#define		HPALOCAL	'L'
#define		HPAAUTOMAN	'O'
#define		HPAGOAUTO	'A'
#define		HPAGOMAN	'M'
#define		HPASTAT		'S'
#define		HPASTATUPDATE	'U'
#define		HPASTATHW	'H'
#define		HPASTATLOW	'L'
#define		HPASTATHIGH	'G'
#define		HPASTATCHN	'C'
#define		HPAOUTPWR	'A'
#define		HPAOUTPWVAL	'V'
#define		HPAOUTPWRRAISE	'R'
#define		HPAOUTPWRLOWER	'L'
#define		HPACHGSLL	'B'
#define		HPACHGSHL	'C'
#define		HPACHGCHN	'Q'
#define		HPAAUTOSW	'W'
#define		HPAAUTOSWDIS	'D'
#define		HPAAUTOSWENB	'E'


#define		HPAONOFFTLEN	2
#define		HPADUMMYTLEN	2
#define		HPARESETTLEN	1
#define		HPARFPWRTLEN	2
#define		HPASELFTESTTLEN	1
#define		HPALOCALTLEN	1
#define		HPAAUTOMANTLEN	2
#define		HPASTATTLEN	2
#define		HPAOUTPWRTLEN	4
#define		HPACHGSLLTLEN	4
#define		HPACHGSHLTLEN	4
#define		HPACHGCHNTLEN	2
#define		HPAAUTOSWTLEN	2

#define		HPASTATUPDATERLEN	79
#define		HPASTATHWRLEN		35
#define		HPASTATLOWRLEN		67
#define		HPASTATHIGHRLEN		67
#define		HPASTATCHNRLEN		101
#define		GPIBHEADERLEN		12



/* task definitions */

#define 	LMP_STACK		5000   /* Stack size              */
#define		LMP_PRIORITY		200    /* default proirity        */
#define		LMP_TASK_OPTIONS	8      /* Floating point	  */
 
