/*drvBrooks.h,v 1.1.1.1 1999/08/03 14:41:57 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
/*	File:	drvbrooks.h
/*	Auth:	John L. Dalesio
/*	Edit:	6/24/93	JLD
/*
/*	This file contains the data structures used by the Brooks
/*	I/O driver.
/*
*/

#define	TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d
#define LF	0x0a

/*	number of brooks meter lines */
#define		BRKLINES	20
/*	number of Brooks devices per line */	
#define		NUMBRKDEVS	32
/* Address will be used to specify the control unit address
/* Card will be used to specify the parameter number or pseudo parameter no.
/* Signal will be used with strings or binary inputs to specify an offset
*/
#define		MAXPARAMS	1000
/*	Executable tasks will be issued by outputs to cards >TASKS and <MAKTASKS */
#define		TASKS		2000
#define		MAXTASKS	2100

/* Pseudo values will be specified using the following parameter number */
/* Signal will be used to select the appropriate pseudo value */
#define		PSEUDO		3000
/* Force Fetches will be activated by writes to address >=4000 and <5000 */
#define		FORCEFETCH	4000
#define		STATPARAM	806


/*	Conversion types for analog inputs	*/
#define		AICNVINTEGER	10
#define		AICNVDOUBLEINT	11
#define		AICNVFLOAT	12
#define		AICNVDOUBLEFLT	13
#define		AICNVASCIIFLT	14
#define		AICNVSTATUS	15
#define		AICNVTIME	16
#define		AICNVASCII_2	17
#define		AICNVASCII_4	18
#define		AICNVASCII_12	19
#define		AICNVASCII_20	20
#define		MAXCONV		20

/* The following must match the definition in recChoice.ascii */
#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2



/*
/*	Communications I/O Structures
/*
*/

#define		MAXRXMSG	1200
#define		RXSEM		0
#define		TXSEM		1

typedef	struct {
	unsigned short		protocol;
	unsigned short		terminator;
	unsigned short		rxstate;
	unsigned short		rxtocnt;
	unsigned short		rxhwerrcnt;
	unsigned short		fixedlen;
	unsigned short		rxmsglen;
	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		rxbuf[MAXRXMSG];
		} 	RXLINE;


#define		MAXTXMSG	100	

typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;

/*
/*	Communications I/O Constants
/*
*/

#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		50	/* 16 ms ticks */

#define		VARRESP			FALSE
#define		FIXEDRESP		TRUE
#define		ONERETRY		0x8000	/* ored into line number for xfermsg */
#define		RXIGNORE		0

#define		TXGOOD			1
#define		TXBAD			2

#define		TXTIMEOUT		125	/* 16 ms ticks */

#define		IO_TIMEOUT		3	/* Code returned by timer */

#define		MAXRETRYS		3	/* retransmission trys */

#define		MAXCFGCK		8	/* Try to Bring online each cycle */
#define		BRINGBACKCNT		4	/* Poll loops per cycle */

#define		XFERGOOD		1
#define		XFERBAD			2

#define		RESNAMELEN		16	/* length of a resource name */
#define		SLWWINDOW		900	/* # of ticks to allow slave to try */

/*
/*	PLC PROGRAM LOAD STATES
*/
#define		PLC_NOT_LOADED		0
#define		PLC_LOAD_REQUESTED	1
#define		PLC_LOAD_IN_PROGRESS	2
#define		PLC_LOAD_FAILED		3
#define		PLC_LOAD_SUCCESSFUL	4



/*
* Output Que Management Structures for Modbus Protocol
*/
#define         AOUTTYPE        1
#define         BOUTTYPE        2
#define         SOUTTYPE        3

#define         OUTQSIZE        100
#define		STRMSGLEN	80
#define		BRKPARAMASCLEN  16

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	unsigned int    aovalue;
	double		dblaovalue;
	unsigned short	conversion;
	unsigned short  prec;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	unsigned long   bomask;
	unsigned long	bobvalue;
	unsigned char   bovalue[STRMSGLEN];
		} BOUT, *BOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  sosignal;
	long            solen;
	unsigned char   sovalue[STRMSGLEN];
		} SOUT, *SOUTPTR;

typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			SOUT	sonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque[OUTQSIZE];
		} OUTPUTQUE;





typedef struct {
		unsigned char	was_requested;
		unsigned char	data_valid;
		unsigned char	force_fetch;
		unsigned char	sdatabuf[BRKPARAMASCLEN];
		double		fdata;
		long		intdata;
		long		maskdata;
		} BRKSPARAMETER;



typedef	struct	{
	unsigned short	was_requested;		/* 0  The scan task has accesses this RTU */
	unsigned short	is_responding;		/* 1  The RTU is responding	*/
	unsigned short	holdflag;		/*    The RTU data is in hold */
	unsigned short	prog_load_stat;		/* 2  Program load Status */
	unsigned short  prog_load_page;		/* 3  Program Load Block Number */
	unsigned short  prog_load_number;	/* 4  Program Number to Load */
	unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs */
	unsigned short  no_datain_msgs;		/* 6  Count Data Requests */
	unsigned short  no_rxnoresp;		/* 7  Count No Responses */
	unsigned short  no_txnoresp;		/* 8  Count No Responses */
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors */
	unsigned short  no_retries;		/* 10 Count Retries */
	unsigned short  no_rxhwerrs;		/* 11 Count Comm Hardware Errors */
	unsigned short  no_txhwerrs;		/* 12 Count Comm Hardware Errors */
	unsigned short  no_crcerrs;		/* 13 Count CRC Errors */
	unsigned short  no_offlines;		/* 14 Count Off Line Transitions */
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats*/
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline */
	unsigned short  badoutcnt;		/* 17 Count Failed output commands */
	unsigned short	scanmode;		/* 18 Sacn mode determines fetch control*/

	BRKSPARAMETER	*params[MAXPARAMS];	/*  Storage for the pointers to the Parameters */

	}	BRKDEV;

/*
/* Brooks Protocol Definitions
/*
*/
#define 	BRKSOH		0x01
#define		BRKSTX		0x02
#define 	BRKETX		0x03
#define		BRKACK		0x06
#define		BRKNAK		0x15

#define		BRKADDRLEN	0x03
#define		BRKPARNOLEN	0x03

#define		BRKREAD		'R'
#define		BRKWRITERET	'W'
#define		BRKWRITEACK	'A'
#define		BRKWRITENORESP	'B'
#define		BRKWRITECFRM	'C'
#define		BRKCONFIRM	'K'
#define		BRKEXECUTE	'X'
#define		BRKACK		0x06
#define		BRKNAK		0x15
#define		BRKPSEP		';'
#define		BRKVALSEP	'='
#define		BRKDISPTASK	2023

#define		RXSOHST		0x01
#define 	BRKADDRST	0x02
#define 	BRKSTXST	0x03
#define		BRKDATAST	0x04
#define 	BRKBCC1ST	0x05
#define		BRKBCC2ST	0x06
#define		BRSMSTRADDR1	'5'
#define		BRSMSTRADDR2	'5'
#define		BRSMSTRADDR3	'5'


/* task definitions */

#define 	BRK_STACK		5000   /* Stack size              */
#define		BRK_PRIORITY		200    /* default proirity        */
#define		BRK_TASK_OPTIONS	8      /* Floating point	  */
