/*drvCsi.h,v 1.1.1.1 1999/08/03 14:42:11 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
**	File:	drvCsi.h
**	Auth:	John L. Dalesio
**
**	This file contains the data structures used by the CSI I/O driver.
**

#define TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d

**	number of lines of modbus 	*/
#define		CSILINES	20
/*	number of modbus plcs per line */
#define		NUMCSIRTUS	64

/*
**	The IOC will assume that the data to be exchanged with each RTU will
**	come from a fixed number of dis,acc,and ais.
**	It will read these initially in polls for all data, and the in
**	requests for upsets.
*/

#define CSIDIS		1			/* Card number for digital inputs 	*/
#define CSIDAS		2			/* Card number for digital alarmss 	*/
#define CSIDMS		3			/* Card number for digital monitors 	*/
#define CSIDCS		4			/* Card number for digital controls 	*/
#define CSIDOS		5			/* Card number for digital outputs 		*/
#define CSIAIS		6			/* Card number for analog inputs 		*/
#define CSIAOS		7			/* Card number for analog outputs 		*/
#define CSIPIS		10			/* Card number for accumulators 	*/
#define PSEUDO		33			/* Card number for pseudo i/o 		*/

#define CSIVALUE	0			/* Value property of object */
#define CSIAUTO		1			/* Auto/Man property of object */
#define CSITEST		2			/* Test Mode property of object */
#define CSIOLOW		3			/* Low limit property of object */
#define CSIOHIGH	4			/* HIGH limit property of object */

#define MAXPOINT	32			/* Maximum Number of points per type */
#define MAXBOFF		10			/* 10 bit offset possibilities per point */
#define MAXDIGS		MAXPOINT*MAXBOFF

/*
**	Conversion types for analog inputs	*/
/*
*/
#define		AICNVINTEGER	10
#define		AICNVDOUBLEINT	11
#define		AICNVFLOAT	12
#define		AICNVDOUBLEFLT	13
#define		AICNVASCIIFLT	14
#define		AICNVSTATUS	15
#define		AICNVTIME	16
#define		AICNVASCII_2	17
#define		AICNVASCII_4	18
#define		AICNVASCII_12	19
#define		AICNVASCII_20	20
#define		MAXCONV		20

/* The following must match the definition in recChoice.ascii */
#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2

/*
**	Communications I/O Structures
**
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	(256*4)+10
#define		MAXTXMSG	256


typedef	struct {
	unsigned short		protocol;
	unsigned short		rxstate;
	unsigned short		rxtocnt;
	unsigned short		rxhwerrcnt;
	unsigned short		fixedlen;
	unsigned short		rxmsglen;
	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		rxbuf[MAXRXMSG];
		} 	RXLINE;


typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;
/*
**	Communications I/O Constants
**
*/
#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		60	/* 16 ms ticks 				*/
#define		ACTIVITYTIME		300	/* must see activity in 5 seconds 	*/

#define		VARRESP			FALSE
#define		FIXEDRESP		TRUE
#define		ONERETRY		0x8000	/* OR'ed into line number for xfermsg 	*/
#define		RXIGNORE		0
#define		RXBLKCNT		1
#define		RXDATA			2

#define		TXGOOD			1
#define		TXBAD			2

#define		TXTIMEOUT		1255	/* 16 ms ticks: just under 21 seconds 	*/ 

#define		IO_TIMEOUT		3	/* Code returned by timer 		*/

#define		MAXRETRYS		3	/* retransmission trys 			*/

#define		MAXCFGCK		8	/* Try to Bring online each cycle 	*/
#define		BRINGBACKCNT		4	/* Poll loops per cycle 		*/

#define		XFERGOOD		1
#define		XFERBAD			2

#define		RESNAMELEN		16	/* length of a resource name 		*/
#define		SLWWINDOW		900	/* # of ticks to allow slave to try 	*/


/*
** Output Queue Management Structures for Modbus Protocol
**
*/

#define         AOUTTYPE        1
#define         BOUTTYPE        2
#define         OUTQSIZE        128

typedef struct {
	char id[5];
    } ACTID;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	short           aovalue;
	double		dblaovalue;
	unsigned short	conversion;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bomask;
	long            bovalue;
		} BOUT, *BOUTPTR;

typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque[OUTQSIZE];
		} OUTPUTQUE;


typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		automode;
	unsigned short		testmode;
	unsigned short		last_read;
	unsigned long		last_time;
		}	CSIDTYPE;

typedef struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		automode;
	unsigned short		testmode;
	float			last_read;
	unsigned long		last_time;
		}	CSIAITYPE;

typedef struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		automode;
	unsigned short		testmode;
	unsigned short		limit_req;
	unsigned short		limit_valid;
	float			outhigh;
	float			outlow;
	float			last_read;
	unsigned long		last_time;
	unsigned long		lim_last_time;
		}	CSIAOTYPE;

typedef	struct	{
	unsigned short  was_requested;		/* 0  The scan task has accesses this RTU 	*/
	unsigned short  is_responding;		/* 1  The RTU is responding 			*/
	unsigned short  holdflag;		/*    The RTU data is in hold 			*/
	unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs 			*/
	unsigned short  no_datain_msgs;		/* 6  Count Data Requests 			*/
	unsigned short  no_rxnoresp;		/* 7  Count No Responses 			*/
	unsigned short  no_txnoresp;		/* 8  Count No Responses 			*/
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors 		*/
	unsigned short  no_retries;		/* 10 Count Retries 				*/
	unsigned short  no_rxhwerrs;		/* 11 Count Comm Hardware Errors 		*/
	unsigned short  no_txhwerrs;		/* 12 Count Comm Hardware Errors 		*/
	unsigned short  no_crcerrs;		/* 13 Count CRC Errors 				*/
	unsigned short  no_offlines;		/* 14 Count Off Line Transitions		*/
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats			*/
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline 			*/
	unsigned short  badoutcnt;		/* 17 Count Failed output commands 		*/

	CSIDTYPE	dis[MAXPOINT*MAXBOFF];	/* digital inputs */
	CSIDTYPE	das[MAXPOINT*MAXBOFF];	/* digital alarms */
	CSIDTYPE	dms[MAXPOINT*MAXBOFF];	/* digital monitors */
	CSIDTYPE	dcs[MAXPOINT*MAXBOFF];	/* digital controls */
	CSIDTYPE	dos[MAXPOINT*MAXBOFF];	/* digital outputs */

	CSIAITYPE	ais[MAXPOINT];		/* analog inputs */
	CSIAITYPE	pis[MAXPOINT*MAXBOFF];	/* pulse inputs */

	CSIAOTYPE	aos[MAXPOINT];		/* Analog outputs */


	}	CSIRTU;


/* CSI Communications constants */

#define	CSIREADCMD			0x12
#define	CSIREADSBCMD			0x84

#define	CSIWRITECMD			0x13
#define	CSIWRITESBCMD			0x80


#define	CSIREADAORECCMD			0x12
#define	CSIREADAORECSBCMD		0x00

#define	CSIWRITEAORECCMD		0x13
#define	CSIWRITEAORECSBCMD		0x00

#define	CSIACTIVATETESTCMD		0x13
#define	CSIACTIVATETESTSBCMD		0x84

#define	CSIDEACTIVATETESTCMD		0x13
#define	CSIDEACTIVATETESTSBCMD		0x85

#define	CSIACTIVATEMANUALCMD		0x13
#define	CSIACTIVATEMANUALSBCMD		0x86

#define	CSIDEACTIVATEMANUALCMD		0x13
#define	CSIDEACTIVATEMANUALSBCMD	0x87

#define CSIPOLLCMD			0x18
#define CSIREPOLLCMD			0x19
#define CSIUNSOLICITED			0x16
#define CSIBROADCAST			0x17
#define COMMANDERRMASK			0x80

#define EVENTFMT			0
#define ACTIONFMT			1
#define CONTROLFMT			2
#define SYSTEMFMT			3


typedef struct {
	unsigned short	type;
	unsigned short	offset;
	unsigned short	number;
    } FETCHPOINT;

#define FETCHMAX	18


/* task definitions */

#define CSI_STACK		8000   /* Stack size              */
#define	CSI_PRIORITY		152    /* default priority        */
#define	CSI_TASK_OPTIONS	8      /* Floating point          */




/*
** Mutual-Exclusion Control for diagnostic/data radio channel sharing */
/*
*/

#define		MUXTIMEOUT	60*30	/* 16.67 ms ticks (60Hz): 30 seconds 	*/

/*	End of File:	drvCSI.h */

