/*drvEabplc.c,v 1.1.1.1 1999/08/03 14:42:55 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
**  File:	drvEabplc.c
**  Edit:	12/18/92 12:33 JLD
**        	01/18/94:	WEP	name changes only
**		03-30-94	ems	startabplc(), remove test ln >= 0
**					it is unsigned
**		04/13/94	WHS	upgrading driver to EPICS
**              12/02/94        WEP     modified eabplc_eabplc_aidriver to mask out MSBit of statistics counts.
**              05/10/95        WEP     added 'abplc_clear' to clear statistics via 'initstats' call.
**              05/22/95        WEP     modified master loop logic to exit when local state is not
**                                      master. This causes lockstep logic to function correctly.   
**                                      test for and conditionally display an illegal lstate msg.
**		03/26/96	ems	chg'd struct drvSabplc to drvEabplc to conform to
**					drvSup.ascii
**
*/



/*#define TEST*/

#ifdef TEST

/* #include "testmode.h" */

#include	"bioscomm.h"
#include	"dummyos.h"
#include 	"drvsabplc.h"
#include 	"link.h"
#include	"sabproto.h"
#define		logMsg printf

#endif

#ifndef TEST

#include	<ctype.h>
#include	<vxWorks.h>
#include	<types.h>
#include	<in.h>
#include	<netinet/tcp.h>
#include	<sys/times.h>
#include	<bootLib.h>
#include	<inetLib.h>
#include	<sysLib.h>
#include	<semLib.h>
#include	<errno.h>
#include	<socket.h>
#include	<selectLib.h>
#include	<fcntl.h>
#include	<ioLib.h>
#include	<wdLib.h>
#include	<rngLib.h>
#include	<task_params.h>
#include	<stdioLib.h>

#include	<dbDefs.h>
#include	<drvSup.h>
#include	<module_types.h>
#include	<drvEabplc.h>
#include	<drvEabplc_proto.h>
#include	<restable.h>
#endif

struct
{
	long		number;
	DRVSUPFUN	report;
	DRVSUPFUN	init;
} drvEabplc =
	{2,
	NULL,
	init
};

/**********
*
*      External Line Level Data Structures (declared in drvcommon.c)
*
**********/

/*	array of resource name strings.		*/

extern	unsigned	char	resnames [ABLINES] [RESNAMELEN];

/*	array of pointers to resource tables 	*/

extern	RES_TABLE		*prestable [ABLINES];

/*	array of resource response flags. 		*/
/*	False = no active resource (slave),		*/
/*	True = active resource (master) 		*/

extern	unsigned	short	resresponse [ABLINES];

/*	true / false line active flag. calculated	*/
/*	using activity_time below. 			*/

extern	unsigned	short	line_activity [ABLINES];

/*	assigned value of line_activity prior to	*/
/*	it's recalculation.   			*/

extern	unsigned	short	lastactivity [ABLINES];

/*	controls output of communication error log	*/
/*	messages (true / false) 			*/

extern	unsigned	short	modtrace [ABLINES];

/*	controls output of certain resource related	*/
/*	log messages (level) 				*/

extern	unsigned	short	mltrace [ABLINES];

/*	controls output of MASTER / SLAVE arbitration	*/
/*	log messages (binary) 			*/

extern	unsigned	short	modrestrace [ABLINES];

/*	holds value passed in task spawn. used for	*/
/*	each msgxfer attempt 				*/

extern	unsigned	short	intermsgdelay [ABLINES];

/*	holds post RTS delay before TX. 		*/

extern	unsigned	short	pretxdelay [ABLINES];

/*	holds post TX delay before dropping RTS. 	*/

extern	unsigned	short	posttxdelay [ABLINES];

/*	holds timetick of last activity. used to	*/
/*	set line_activity flag  			*/
/*	shared resource of the data link 		*/
/*	controls access to common path shared by	*/
/*	data and diagnostic channels. 			*/
/*	originally designed for use with MDS remote	*/
/*	radio diagnostics. 				*/

extern	long			activity_time [ABLINES];

extern	char			consoleName [];
 
/**********
*
*      Storage for Allen Bradley Comm Line Structures
*
**********/

/* array of pointers to dynamically allocated comm controls / buffers */


static RXLINE		*rxlinetab [ABLINES] =
 {
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL
};

static TXLINE          *txlinetab [ABLINES] = 
{
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL
};

static	unsigned short	enet_tmout [ABLINES];

/**********
*
*      Storage for Allen Bradley PLC Structures
*
**********/

static	ABPLC		*plcs [ABLINES] [NUMABPLCS + 1];

static	OUTPUTQUE	*oq [ABLINES];

/*	holds id and password of the user that       */
/*	spawned the driver task required by host to  */
/*	check file read privileges on PLC load files */

static	char		user [40], password [40];

static	char    	*bools [] =
{

	"TRUE ",
	"FALSE"
};

static	int		taskId [ABLINES];

static	char		*tasknames [] =
{
	"EABPLC Ch0", "EABPLC Ch1", "EABPLC Ch2", "EABPLC Ch3",
	"EABPLC Ch4", "EABPLC Ch5", "EABPLC Ch6", "EABPLC Ch7",
	"EABPLC Ch8", "EABPLC Ch9", "EABPLC Ch10", "EABPLC Ch11",
	"EABPLC Ch12", "EABPLC Ch13", "EABPLC Ch14", "EABPLC Ch15",
	"EABPLC CH16", "EABPLC Ch17", "EABPLC Ch18", "EABPLC Ch19",
	"EABPLC CH20", "EABPLC Ch21", "EABPLC Ch22", "EABPLC Ch23",
	"EABPLC CH24", "EABPLC Ch25", "EABPLC Ch26", "EABPLC Ch27",
	"EABPLC CH28", "EABPLC Ch29", "EABPLC Ch30", "EABPLC Ch31"
};

static	short		linemon [ABLINES];
static	char		partner_alive [ABLINES];

/****************************************************************/
/*								*/
/*	STRUCTURE TO KEEP STATUS FOR ONE PLC-5			*/
/*								*/
/****************************************************************/

struct ab_plc5
    {
    int			inetAddr;	/* Always in network byte order */
    int			sockFd;		/* -1 = none */
    int			rxBuffSize;	/* Receive buffer size */
    int			rxBuffTopped;	/* Buffer Topped? */
    int			txBuffSize;	/* Transmit buffer size */
    unsigned int	sessionID;	/* -1 = none */
    unsigned int	requestID;	/* Just a message counter */
    };

static int enetSend();
static int enetRead();
static void abortConnection();

static struct ab_plc5	*eabplc_ip [NUMABPLCS];

/**********
*
*	This routine is required by iocinit in the EPICS system.
*
**********/

static long init ()
{
	return (0);
}
/********
*
*	This routine enables logging of comm trace messages
*
********/

void eabplctraceon (unsigned short line)
{
	modtrace [line] = TRUE;
}

/********
*
*	This routine disables logging of comm trace messages
*
********/

void eabplctraceoff (unsigned short line)
{
	modtrace [line] = FALSE;
}
/********
*
*	This routine enables logging of comm trace messages
*
********/

void eablnmonon (unsigned short line)
{
	linemon [line] = TRUE;
}

/********
*
*	This routine disables logging of comm trace messages
*
********/

void eablnmonoff (unsigned short line)
{
	linemon [line] = FALSE;
}

/********
*
*	This routine enables logging of resource trace messages
*
********/

void eabplcreseabplctraceon (unsigned short line)
{
	modrestrace [line] = TRUE;
}

/********
*
*	This routine disables logging of resource trace messages
*
********/

void eabplcreseabplctraceoff (unsigned short line)
{
	modrestrace [line] = FALSE;
}

/********
*
*	This routine controls logging of trace messages
*
********/

void eabplctrlv (unsigned short line, unsigned short level)
{
	mltrace [line] = level;
}

int stopeabplc (short ln)
{
	if (comm_disconnect (ln) != -1)
	{
	    taskDelete (taskId [ln]);
	    return (1);
	}
	else
	{
	    return (-1);
	}
}

int starteabplc (unsigned short ln, char *inetaddr, unsigned short delay, unsigned short etmout)
{
	if ( ln < ABLINES )
	{
	    taskId [ln] = taskSpawn (tasknames [ln],
                                     AB_PRIORITY,
                                     AB_TASK_OPTIONS,
                                     AB_STACK, eabplc_driver,
                                     ln,
				     inetaddr,
                                     delay,
				     etmout);

	    logMsg ("Started the EABPLC driver on line %d\n", ln);

	    remCurIdGet (user, password);
	    return (1);
	}
	else
	{
	    logMsg ("Invalid line number %d\n", ln);
	    return (-1);
	}
}

/********
*
* Set communication logic time delays to handle half/full duplex,
* radio delays, etc 
*
********/

void eabplcsetdelays (unsigned short line,
		unsigned short tickcount1,
		unsigned short etmout)
{
	/* at beginning new message / response sequence. */

	intermsgdelay [line] = tickcount1;

	enet_tmout [line] = etmout;
}

/**********
*
*      abplc_driver
*
*      The purpose of this task is to scan the allen bradley plc
*      lines. It will eabplc_drivertain a configuration list supporting online additions
*      and repplacement of modules, and  eabplc_drivertain a table of most recently
*      scanned values.
*
**********/
void eabplc_driver (unsigned short line, unsigned char *inetaddr, unsigned short delay,
		unsigned short etmout)
{
	unsigned	short 	mod, cfgchk, cnt, modspolled, onlinefreq;
	unsigned	short 	activity, pstate, lstate, execute, command,i;
	long	        	this_time, last_time;
	char			tyconame [16];
	int			tycoport;
	ABPLC          		*pmod;
	RXLINE          	*rxp;
	TXLINE          	*txp;

	/* Initialize the data structures */

	intermsgdelay [line] = delay;
	pretxdelay [line] = 0;
	posttxdelay [line] = 0;
	enet_tmout [line] = etmout;

	resresponse [line] = FALSE;
	resnames [line] [0] = 0;
	prestable [line] = NULL;

	modtrace [line] = FALSE;
	mltrace [line] = 0;
	modrestrace [line] = TRUE;

	activity_time [line] = tickGet ();
	lastactivity [line] = line_activity [line] = FALSE;

	oq [line] = (OUTPUTQUE *) calloc (1, (sizeof (OUTPUTQUE)));
	if (!oq [line])
	{
	    logMsg ("Not enough memory for line %d data structure OUTPUTQUE\n",
			line);
	    return;
	}

	rxlinetab [line] = (RXLINE *) calloc (1, (sizeof (RXLINE)));
	if (!rxlinetab [line])
	{
	    logMsg ("Not enough memory for line %d data structure RXLINE\n",
			line);
	    return;
	}
	txlinetab [line] = (TXLINE *) calloc (1, (sizeof (TXLINE)));
	if (!txlinetab [line])
	{
	    logMsg ("Not enough memory for line %d data structure TXLINE\n",
			line);
	    return;
	}

	eabplc_ip [line] = (struct ab_plc5 *) enetInit (inetaddr);

	/* Attach to the serial redundant task */

	attachSerialRedund (line, 500);

	onlinefreq = 1;	/* causes on_line calls to begin on first pass if master */

	rxp = rxlinetab [line];
	rxp->rxstate = RXIGNORE;

	/* Wait for resource to be defined via an initial call	*/
	/* to 'register_resource'.				*/

	while ((resnames [line] [0] == 0) || (prestable [line] == NULL))
	{
	    if (mltrace [line] == 1)
	    {
	    	logMsg ( "No Resource Defined for line - %d \n", line);
	    }
	    taskDelay (120);
	}

	/* Start out as slave */

	resClearCaAnswer (prestable [line]);		/* start quiet */
	resPutLocalState (prestable [line], RES_SLV);
	if (modrestrace [line])
	{
	    logMsg ("1] Became RES_SLV\n");
	}
	taskDelay (300);
	resGetStates (prestable [line], &lstate, &pstate);
	activity = check_line_activity (line, ACTIVITYTIME);

	/* if the partner is not there, be master */

	if ((pstate != RES_MSTR) && (pstate != RES_SLVWANT) && (!activity))
	{
	    resSetCaAnswer (prestable [line]);
	    resPutLocalState (prestable [line], RES_MSTR);
	    lstate = RES_MSTR;
	    if (modrestrace [line])
	    {
	    	logMsg ("2] Became RES_MSTR\n");
	    }
	}

	/* Task will process all modules sending outputs and getting inputs */
	/* It will then try to configure requested non responding modules   */

	cfgchk = 0;     	/* Init the new config pointer */
	last_time = tickGet (); /* Initialize the time  */

	while (TRUE)
	{
	    activity_time [line] = tickGet ();
	    lastactivity [line] = line_activity [line] = FALSE;

	    /* we arrive here because we are a slave.		*/
	    /* we stay here until:				*/
	    /*		asked to be master by partner.		*/
	    /*		if we reach PLCs, become master.	*/
	    /*		the partner is not master and there is	*/
	    /*		no activity on the line. 		*/

	    while (lstate == RES_SLV)
	    {
		if (pstate != RES_SLVWANT)
		{
		    clear_rts (line); 	/* make sure line is left for master */
		}

		resresponse [line] = FALSE; /* make sure driver is not responding 	*/
		resPutCommand (prestable [line], 0); 	/* clear out commands 			*/

		pulseSerialRedund (line, 'S');
		if ((partner_alive [line] = remoteSerialRedund (line)) != 'M')
		{
		    line_activity [line] = FALSE;
		}
		else
		{
		    line_activity [line] = TRUE;
		}

		for (mod = 1; (mod <= NUMABPLCS); mod++)
		{
		    pmod = plcs [line] [mod];
		    if (pmod)    /* verify pointer validity prior to use */
		    {
			if ((partner_alive [line] == 'M') ||
			    (partner_alive [line] == 'W'))
			{
			    pmod->partner_online = TRUE;
			}
			else
			{
			    pmod->partner_online = FALSE;
			}
		    }
		}
		
		/* wait 10 seconds before checking 	*/

		activity = check_line_activity (line, 10);
		resGetStates (prestable [line], &lstate, &pstate);

		/* see if needed by partner to be master. */
		/* if yes, do it and break out of slave state */

		if (pstate == RES_SLVWANTCA)
		{
		    resSetCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_MSTR);
		    lstate = RES_MSTR;
		    if (modrestrace [line])
		    {
			logMsg ("3] Became RES_MSTR  P= %x A= %d \n", pstate, activity);
		    }
		    break;
		}

		/* If partner is there and wants to give it up, try to take it */

		if (pstate == RES_SLVWANT)
		{
		    execute = FALSE;

		    taskDelay (2); /* Time for switch to occur */

		    /* See if anyone is out there and able to respond to a loopback command */
		    for (mod = 1; (mod <= NUMABPLCS); mod++)
		    {
			pmod = plcs [line] [mod];
			if (pmod)    /* verify pointer validity prior to use */
			{
			    if ((pmod->was_requested) && (!pmod->cmd_force_offline))
			    {
			        on_line (line, mod);	
			        if ((pmod->is_responding) && (!pmod->cmd_force_offline))
			        {   
				    execute = TRUE;
				    break;
			        }
			    }
			}
		    }

		    /* If someone answered, become master and make sure partner really gave it up */

		    if (execute)
		    {
			resPutLocalState (prestable [line], RES_MSTR);
			lstate = RES_MSTR;
			if (modrestrace [line])
			{
			    logMsg ("4] Became RES_MSTR  P= %x A= %d \n",
					pstate, activity);
			}
			execute = FALSE;
			activity_time [line] = tickGet ();
			lastactivity [line] = line_activity [line] = FALSE;
			for (i = 0; (i < 60); i++)
			{
			    activity = check_line_activity (line, 2);    /* pause before testing */
			    resGetStates (prestable [line], &lstate, &pstate);
			    if ((pstate != RES_MSTR) &&
				(pstate != RES_SLVWANT) && (!activity))
			    {
				execute = TRUE; /* Partner Gave It Up so take CA */
				break;
			    }
			}
			if (!execute)
			{   /* Partner did not give it up go back to slave */

			    resPutLocalState (prestable [line], RES_SLV);
			    lstate = RES_SLV;
			    if (modrestrace [line])
			    {
			    	logMsg ("5] Became RES_SLV  P= %x A= %d \n",
					pstate, activity);
			    }
			}
			else
			{ /* Assume control by taking the CA responsibilities */ 

			    resSetCaAnswer (prestable [line]);
			    prepresponses (line);
			    break;
			}
		    }
		}

		/* if the partner is not talking, be master */

		if ((pstate != RES_MSTR) &&
			(pstate != RES_SLVWANT) && (!activity))
		{
		    resSetCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_MSTR);
		    lstate = RES_MSTR;
		    if (modrestrace [line])
		    {
		    	logMsg ("6] Became RES_MSTR P= %x A= %d \n", 
				pstate, activity);
		    }
		    prepresponses (line);
		}
	    }

            /* to arrive here, we used to be master		*/
	    /* and had been told to relinquish master status.	*/
	    /* stay in this loop until the partner takes over	*/
	    /* or the command is canceled			*/

	    while (lstate == RES_SLVWANTCA)
	    {
	        taskDelay(2);		/* Give OS a chance...  */
		clear_rts (line); 	/* Make sure line is left for master */
		resresponse [line] = FALSE; 	/* Make Sure driver is not responding */
		pulseSerialRedund (line, 'W');
		if (remoteSerialRedund (line) != 'M')
		{
		    line_activity [line] = FALSE;
		}
		else
		{
		    line_activity [line] = TRUE;
		}

		for (mod = 1; (mod <= NUMABPLCS); mod++)
		{
		    pmod = plcs [line] [mod];
		    if (pmod)    /* verify pointer validity prior to use */
		    {
			pmod->partner_online = line_activity [line];
		    }
		}
		
 		resGetStates (prestable [line], &lstate, &pstate);
		if ((pstate == RES_MSTR) || (pstate == RES_SLVWANT))
		{
		    resClearCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_SLV);
		    lstate = RES_SLV;
		    if (modrestrace [line])
		    {
		    	logMsg ("7] Became RES_SLV  P= %x A= %d \n",
				pstate, activity);
		    }
		    break;
		}
		resGetCommand (prestable [line], &command);
		resPutCommand (prestable [line], 0); /* clear out commands */
		if (command == RES_CANCELREL)
		{
		    resSetCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_MSTR);
		    lstate = RES_MSTR;
		    if (modrestrace [line])
		    {
		    	logMsg ("8] Became RES_MSTR  P= %x A= %d \n",
				pstate, activity);
		    }
		    prepresponses (line);
		}
 	    }


	    /* in normal operation this code should never execute.	*/
	    /* it has, however, been seen to execute when the database	*/
	    /* has been inadvertently configured with the same resource	*/
	    /* assigned to two different drivers in the same IOC.	*/
	    /* The state does funny things.				*/

	    while (lstate == RES_SLVWANT)
	    {
		pulseSerialRedund (line, 'W');
		if (remoteSerialRedund (line) != 'M')
		{
		    line_activity [line] = FALSE;
		}
		else
		{
		    line_activity [line] = TRUE;
		}
		for (mod = 1; (mod <= NUMABPLCS); mod++)
		{
		    pmod = plcs [line] [mod];
		    if (pmod)    /* verify pointer validity prior to use */
		    {
			pmod->partner_online = line_activity [line];
		    }
		}
		
                if (modrestrace [line])
		{
                    /* condition should NEVER occur. print error message every 5 seconds. */
                    logMsg ("Illegal Local Resource State.\n");
                    taskDelay (600);
                }
                taskDelay (2);
            }

	    /* we come here because we have gone to, or are looping in,	*/
	    /* the master state. we stay in this loop until:		*/
	    /*	we have been told to give master state to the partner	*/
	    /*	we break out to RES_SLVWANTCA.				*/
	    /*	we have successfully handed master state to partner.	*/
	    /* While master, this task will process all modules sending	*/
	    /* outputs and getting inputs. It will then try to		*/
	    /* configure requested non-responding modules.		*/
	    /* it doesn't appear that we ever come here in RES_SLVWANT	*/
	    /* lstate mode. it happens within this loop, but is never	*/
	    /* seen by the while loop conditional evaluation.		*/

	    last_time = tickGet ();		/* initialize gap */
	    while ((lstate == RES_MSTR) || (lstate == RES_SLVWANT))
	    {
		/* initial stuff */

		resresponse [line] = TRUE; /* Make sure driver is responding. */

		resGetCommand (prestable [line], &command);
		resPutCommand (prestable [line], 0);    /* clear out commands */

		/* if relinquish command sent, change state and break out to outer loop */
		pulseSerialRedund (line, 'M');
		if (remoteSerialRedund (line) != 'S')
		{
		    line_activity [line] = FALSE;
		}
		else
		{
		    line_activity [line] = TRUE;
		}

		for (mod = 1; (mod <= NUMABPLCS); mod++)
		{
		    pmod = plcs [line] [mod];
		    if (pmod)    /* verify pointer validity prior to use */
		    {
			pmod->partner_online = line_activity [line];
		    }
		}
		
		if (command == RES_RELINQUISH)
		{
		    resClearCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_SLVWANTCA);
		    lstate = RES_SLVWANTCA;
		    if (modrestrace [line])
		    {
                       logMsg ("Became RES_SLVWANTCA  P= %x A= %d \n",
				pstate, activity);
		    }
		    break;
		}

		modspolled = 0;

		/* call 'scan_plc' for each PLC device, unless user has */
                /* taken it off line or it is not responding.           */

		for (mod = 1; (mod <= NUMABPLCS); mod++ )
		{
		    pmod = plcs [line] [mod];		/* shorthand pointer 			*/
		    if (pmod)				/* verify pointer validity prior to use */
		    {
			if (pmod->cmd_force_offline)
			{ 
			    /* if user has turned this plc offline via		*/
			    /* pseudo commands, mark it as not responding.	*/

 			    pmod->is_responding = FALSE;
	    		}
			else if (pmod->is_responding)
			{
			    modspolled++;


			    /* local/partner lockstep recovery */

			    resGetStates (prestable [line], &lstate, &pstate);
			    if (pstate == RES_MSTR)
			    {
				resClearCaAnswer (prestable [line]);
				resPutLocalState (prestable [line], RES_SLV);
				lstate = RES_SLV;
				clear_rts(line); 		/* Make sure line is left for master */
				for (mod = 1; (mod <= NUMABPLCS); mod++ )
				{
		    		    pmod = plcs [line] [mod];		/* shorthand pointer 	*/
		    		    if (pmod)				/* verify pointer validity prior to use */
		    		    {
		     			pmod->is_responding = FALSE;
	    			    }
				}
				if (modrestrace [line])
				{
				    logMsg ("13] Became RES_SLV  P= %x A= %d \n", pstate, activity);
				}
				taskDelay (((tickGet () % 8) + 5) * 60); /* random 5 to 12 second delay */
				break;
			    }

			    scan_plc (pmod, line, mod);

			    if (!pmod->is_responding)
			    {
				mark_not_valid (pmod);
                            }
			}
		    }

		    /* end of single module action */

		    /* if told to relinquish our master state, stop plc	*/
		    /* scanning immediately. don't trash the command	*/
		    /*so that we can still read it on the next loop	*/
                    /* through this while statement.			*/

		    resGetCommand (prestable [line], &command);
		    if (command == RES_RELINQUISH)
		    {
			break;
		    }
		}					/* end of scan loop */

                /* test for lockstep detection. */

                if (lstate != RES_MSTR)
		{
		    break;
		}

		/* try and bring all nonresponding plcs on line, but */
                /* don't bring in more than a specified number.      */

	        cnt = 0;  /* count the number to bring on in one pass */

	        if (--onlinefreq == 0)
		{
		    onlinefreq = BRINGBACKCNT;

		    for ( mod = 1; (mod < NUMABPLCS); mod++)
		    {
		        cfgchk++;
			pulseSerialRedund (line, 'M');
		        if (cfgchk == NUMABPLCS)
			{
			    cfgchk = 1;
			}
		        pmod = plcs [line] [cfgchk];
			if (pmod)
			{
		            if ((!pmod->is_responding) &&
				(pmod->was_requested) &&
				(!pmod->cmd_force_offline))
			    {
				on_line (line, cfgchk);
				modspolled++;
				cnt++;
				if (cnt == MAXCFGCK)
				{
				    break;
				}
			    }
			}
		    }
	        }


		/* see if anyone on the line has been responding. */
                /* if not, try to get partner to become slave.    */
		/* Check for no one responding */

		if (!any_responding (line))
		{
		    execute = FALSE;

		    for (mod = 1; (mod <= NUMABPLCS); mod++)
		    {
			pulseSerialRedund (line, 'M');
			pmod = plcs [line] [mod];
			if (pmod)    /* verify pointer validity prior to use */
			{
			    if ((pmod->was_requested) &&
				(!pmod->cmd_force_offline))
			    {
			        on_line (line, mod);	

			        if ((pmod->is_responding) &&
				    (!pmod->cmd_force_offline))
			        {   
				    execute = TRUE;
				    break;
			    	}
			    }
			}
		    }
		    if (execute)
		    {
			break;
		    }

		    /* See if it is time to let partner try */

		    this_time = tickGet ();
		    if ((this_time - last_time) > (3 * SLWWINDOW))
		    {   /* # ticks for slave to try */

			/* Switch States to let partner know it's its time to try */

			resPutLocalState (prestable [line], RES_SLVWANT);
			lstate = RES_SLVWANT;
			if (modrestrace [line])
			{
			    logMsg ("10] Became RES_SLVWANT  P= %x A= %d \n",
				pstate, activity);
			}

			/* Remember the time of this try */

			last_time = tickGet ();
			lastactivity [line] = line_activity [line] = FALSE;
			for (i = 0; (i < SLWWINDOW); i += 10)
			{ /* SLWINDOW is the # of ticks for slave to try */
			    taskDelay (10);

			    /* Let partner know that we are alive via serial link */

			    pulseSerialRedund (line, 'W');

		    	    /* If partner picked up let it have control */

			    activity = check_line_activity (line, 0);  /* check immediately */
			    resGetStates (prestable [line], &lstate, &pstate);
 			    if ((pstate == RES_MSTR) || (pstate == RES_SLVWANT) ||
				(activity))
			    {
				resClearCaAnswer (prestable [line]);
				resPutLocalState (prestable [line], RES_SLV);
				lstate = RES_SLV;
				if (modrestrace [line])
				{
				    logMsg ("11] Became RES_SLV  P= %x A= %d \n",
					 pstate, activity);
				}
				taskDelay (240);
				break;
			    }
			}

			/* Go Back to being the Master and take control of the hardware */
			if (lstate != RES_SLV)
			{
                            resSetCaAnswer (prestable [line]);
                            resPutLocalState (prestable [line], RES_MSTR);
			    lstate = RES_MSTR;
			    if (modrestrace [line])
			    {
			    	logMsg ("12] Became RES_MSTR  P= %x A= %d \n",
					pstate, activity);
			    }
			}
		    }
		}

		/* take plcs off hold status. */
		/* the plc's data are not to be held off since now master. */

		for (mod = 1; (mod <= NUMABPLCS); mod++ )
		{
		    pmod = plcs [line] [mod];
		    if (pmod)		/* verify pointer validity prior to use */
		    {        
			pmod->holdflag = FALSE;
                    }
		}
	    }

#ifndef TEST
	    if (modspolled == 0)
	    {
		taskDelay (60);
	    }
	    else 
	    {
		taskDelay (10);
	    }
#endif
#ifdef  TEST
	    scansome();
#endif

	} /* end task loop */
}

/*********
*
* Try to bring a PLC on line
*
**********/

static	void    on_line (unsigned short line, unsigned short mod)
{
	ABPLC			*pmod;
	unsigned	short	i, j, offlines;

	pmod = plcs [line] [mod];

/*
**      EVERY PLC MUST HAVE AT LEAST THE DEFAULT STATUS FILE DEFINED
**      & IF I CAN READ THEM THE PLC IS RESPONDING
**
*/

	if (pmod)
	{
	    pmod->nodatastrike = 0;
	    offlines = pmod->no_offlines;	/* avoid extra counts */
	    if (!abdiagstat (pmod, line, mod))
	    {
		pmod->no_offlines = offlines;
		return;
	    }
	    mark_not_valid (pmod);	/* PLC is online - Init data structures */

	    pmod->scanpass = 0;
	    pmod->is_responding = TRUE;
	}
}

/**********
*							
* Mark all plc data elements as not valid     		
*                                             		
**********/                                             	

/* called from on_line and prepresponses      		*/    	

static	void	mark_not_valid (ABPLC *pmod)
{
	unsigned	short	i;
/*
**      INITIALIZE ALL VALUES AS
**              - NOT REQUESTED
**              - DATA NOT VALID
**
*/

	for (i = 0; (i < NUMABOUTS); i++)
	{
	    pmod->abfouts [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINS); i++)
	{
	    pmod->abfins [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABSTATS); i++)
	{
	    pmod->abfstats [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABBITS); i++)
	{
	    pmod->abfbits [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABTMRS); i++)	
	{
	    pmod->abftimers [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABCNTRS); i++)
	{
	    pmod->abfcntrs [i].data_valid = FALSE;
	}
	for (i=0; (i < NUMABCNTRLS); i++)
	{
	    pmod->abfcntrls [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINTS); i++)
	{
	    pmod->abfints [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINTS); i++)
	{
	    pmod->abfexints1 [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINTS); i++)
	{
	    pmod->abfexints2 [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINTS); i++)
	{
	    pmod->abfexints3 [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINTS); i++)
	{
	    pmod->abfexints4 [i].data_valid = FALSE;
	}
}

/**********
*
* Scan an Allen Bradley PLC -  get inputs an set outputs as required
*
**********/

static	void    scan_plc (ABPLC *pmod, unsigned short line, unsigned short mod)
{
	unsigned	short	i, j, k, val, maxints, numgroups;
	unsigned	short	ignore_holes;
	unsigned	long	bitmask;

	pmod->scanpass++;
	pmod->tscanpass++;
	checkoutputs (line);

	/* read the Output Image values to be scanned */

	for (i = 0; (i < NUMABOUTS); i++)
	{
	    if (pmod->abfouts [i].was_requested)
	    {
		for (j = 1; ((j < MAXABOUTS) && ((i + j) < NUMABOUTS)); j++)
		{
		    if (!pmod->abfouts [i + j].was_requested)
		    {
			break;
		    }
		}
		numgroups = j;
		if (!readabouts (line, mod, pmod, i, numgroups))
		{
		    return;
		}
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}

	/* read the Input Image values to be scanned */

	for (i = 0; (i < NUMABINS); i++)
	{
	    if (pmod->abfins [i].was_requested)
	    {
		for (j = 1; ((j < MAXABINS) && ((i + j) < NUMABINS )); j++)
		{
		    if (!pmod->abfins [i + j].was_requested)
		    {
			break;
		    }
		}
		numgroups = j;
		if (!readabins (line, mod, pmod, i, numgroups))
		{
		    return;
		}
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}

	/* read the Status Image values to be scanned */

	for (i = 0; (i < NUMABSTATS); i++)
	{
	    if (pmod->abfstats [i].was_requested)
	    {
		for (j = 1; ((j < MAXABSTATS) && ((i + j) < NUMABSTATS)); j++)
		{
		    if (!pmod->abfstats [i + j].was_requested)
		    {
			break;
		    }
		}
		numgroups = j;
		if (!readabstats (line, mod, pmod, i, numgroups))
		{
		    return;
		}
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}

	/* read the Bit File values to be scanned */

	for (i = 0; (i < NUMABBITS); i++)
	{
	    if (pmod->abfbits [i].was_requested)
	    {
		for (j = 1; ((j < MAXABBITS) && ((i + j) < NUMABBITS )); j++)
		{
		    if (!pmod->abfbits [i + j].was_requested)
		    {
			break;
		    }
		}
		numgroups = j;
		if (!readabbits (line, mod, pmod, i, numgroups))
		{
		    return;
		}
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}

	/* Adjust the maximum number of integers based on scan pass */
/*
	maxints = NUMFASTABINTS;

	if ((pmod->scanpass % SLOWPASS) == 0)
	{
	    pmod->scanpass = 0;
	    maxints =  NUMABINTS;
	}
*/
	maxints =  NUMABINTS;

	/* read the Default Integer File  values to be scanned */

	for (i = 0; (i < maxints); i++)
	{
	    if (pmod->abfints [i].was_requested)
	    {
		for (j = 1; ((j < MAXABINTS) && ( (i + j) < maxints )); j++)
		{
		    if (!pmod->abfints [i + j].was_requested)
		    {
			ignore_holes = FALSE;
			for (k = 1; (((k + j) < MAXABINTS) && ((i + j + k) < maxints) && (k < 4)); k++)
			{
			    if (pmod->abfints [i + j + k].was_requested)
			    {
				ignore_holes = TRUE;
			    }
			}
		    }
		    else
		    {
			continue;
		    }
		    if (ignore_holes)
		    {
			j += (k - 1);
		    }
		    else
		    {
			break;
		    }
		}
		numgroups = j;
		if (!readabints (line, mod, pmod, ABINTS, pmod->abfints,
				i, numgroups))
		{
		    return;
		}
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}

	/* read the First Extended Integer File  values to be scanned */

	if (pmod->extintfile1 < MAXFILE)
	{
	    for (i = 0; (i < maxints); i++)
	    {
		if (pmod->abfexints1 [i].was_requested)
		{
		    for (j = 1; ((j < MAXABINTS) && ((i + j) < maxints )); j++)
		    {
			if (!pmod->abfexints1 [i + j].was_requested)
			{
			    ignore_holes = FALSE;
			    for (k = 1; (((k + j) < MAXABINTS) && ((i + j + k) < maxints) && (k < 4)); k++)
			    {
				if (pmod->abfexints1 [i + j + k].was_requested)
				{
				    ignore_holes = TRUE;
				}
			    }
			}
			else
			{
			    continue;
			}
			if (ignore_holes)
			{
			    j += (k - 1);
			}
			else
			{
			    break;
			}
		    }
		    numgroups = j;
		    if (!readabints (line, mod, pmod, pmod->extintfile1,
				pmod->abfexints1 ,i, numgroups))
		    {
			return;
		    }
		    checkoutputs (line);
		    i = (i + j) - 1;
		}
	    }
	}

	/* read the Second Extended Integer File  values to be scanned */

	if (pmod->extintfile2 < MAXFILE)
	{
	    for (i = 0; (i < maxints); i++)
	    {
		if (pmod->abfexints2 [i].was_requested)
		{
		    for (j = 1; ((j < MAXABINTS) && ((i + j)<maxints)); j++)
		    {
			if (!pmod->abfexints2 [i + j].was_requested)
			{
			    ignore_holes = FALSE;
			    for (k = 1; (((k + j) < MAXABINTS) && ((i + j + k) < maxints) && (k < 4)); k++)
			    {
				if (pmod->abfexints2 [i + j + k].was_requested)
				{
				    ignore_holes = TRUE;
				}
			    }
			}
			else
			{
			    continue;
			}
			if (ignore_holes)
			{
			    j += (k - 1);
			}
			else
			{
			    break;
			}
		    }
		    numgroups = j;
		    if (!readabints (line, mod, pmod, pmod->extintfile2,
				pmod->abfexints2, i, numgroups))
		    {
			return;
		    }
		    checkoutputs (line);
		    i = (i + j) - 1;
		}
	    }
	}

	/* read the Third Extended Integer File  values to be scanned */

	if (pmod->extintfile3 < MAXFILE)
	{
	    for (i = 0; (i < maxints); i++)
	    {
		if (pmod->abfexints3 [i].was_requested)
		{
		    for (j = 1; ((j < MAXABINTS) && ((i + j) < maxints)); j++)
		    {
			if (!pmod->abfexints3 [i + j].was_requested)
			{
			    ignore_holes = FALSE;
			    for (k = 1; (((k + j) < MAXABINTS) && ((i + j + k) < maxints) && (k < 4)); k++)
			    {
				if (pmod->abfexints3 [i + j + k].was_requested)
				{
				    ignore_holes = TRUE;
				}
			    }
			}
			else
			{
			    continue;
			}
			if (ignore_holes)
			{
			    j += (k - 1);
			}
			else
			{
			    break;
			}
		    }
		    numgroups = j;
		    if (!readabints (line, mod, pmod, pmod->extintfile3,
				pmod->abfexints3, i, numgroups))
		    {
			return;
		    }
		    checkoutputs (line);
		    i = (i + j) - 1;
		}
	    }
	}

	/* read the Fourth Extended Integer File  values to be scanned */

	if (pmod->extintfile4 < MAXFILE)
	{
	    for (i = 0; (i < maxints); i++)
	    {
		if (pmod->abfexints4 [i].was_requested)
		{
		    for (j = 1; ((j < MAXABINTS) && ((i + j) < maxints)); j++)
		    {
			if (!pmod->abfexints4 [i + j].was_requested)
			{
			    ignore_holes = FALSE;
			    for (k = 1; (((k + j) < MAXABINTS) && ((i + j + k) < maxints) && (k < 4)); k++)
			    {
				if (pmod->abfexints4 [i + j + k].was_requested)
				{
				    ignore_holes = TRUE;
				}
			    }
			}
			else
			{
			    continue;
			}
			if (ignore_holes)
			{
			    j += (k - 1);
			}
			else
			{
			    break;
			}
		    }
		    numgroups = j;
		    if (!readabints (line, mod, pmod, pmod->extintfile4, pmod->abfexints4, i, numgroups))
		    {
			return;
		    }
		    checkoutputs (line);
		    i = (i + j) - 1;
		}
	    }
	}

	/* read the Timer File values to be scanned */

	for (i = 0; (i < NUMABTMRS); i++)
	{
	    if (pmod->abftimers [i].was_requested)
	    {
		for (j = 1; ((j < MAXABTMRS) && ((i + j) < NUMABTMRS)); j++)
		{
		    if (!pmod->abftimers [i + j].was_requested)
		    {
			break;
		    }
		}
		numgroups = j;
		if (!readabtimers (line, mod, pmod, i, numgroups))
		{
		    return;
		}
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}

	/* read the Counter File values to be scanned */

	for (i = 0; (i < NUMABCNTRS); i++)
	{
	    if (pmod->abfcntrs [i].was_requested)
	    {
		for (j = 1; ((j < MAXABCNTRS) && ((i + j) < NUMABCNTRS)); j++)
		{
		    if (!pmod->abfcntrs [i + j].was_requested)
		    {
			break;
		    }
		}
		numgroups = j;
		if (!readabcntrs (line, mod, pmod, i, numgroups))
		{
		    return;
		}
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}

	/* read the Controllers File values to be scanned */

	for (i = 0; (i < NUMABCNTRLS); i++)
	{
	    if (pmod->abfcntrls [i].was_requested)
	    {
		for (j = 1; ((j < MAXABCNTRLS) && ((i + j) < NUMABCNTRLS)); j++)
		{
		    if (!pmod->abfcntrls [i + j].was_requested)
		    {
			break;
		    }
		}
		numgroups = j;
		if (!readabcntrls (line, mod, pmod, i, numgroups))
		{
		    return;
		}
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}
}

/***********
*
*	checkoutputs
*
*	checks output que and returns if empty.
*	if there is something on the que it attempts to process all
*	pending outputs doing imediate retries. If attempts fail the
*	request is dropped
*
**********/
static	void	checkoutputs (unsigned short line)
{
	OUTQENT			*optr;
	unsigned	short	signal, value, link, adapter, card;
	unsigned	long	bomask, boval;
	ABPLC			*pmod;

	while ((optr = ogetnextp (line)) != NULLP)
	{
	    if ((optr->otype) == AOUTTYPE)
	    {
		link = optr->out.aonode.link;
		adapter = optr->out.aonode.adapter;

		pmod = plcs [link] [adapter];
		if (pmod == NULL)
		{
		    if (!allocate_plc (link, adapter))
		    {
			return;		/* if not allocated, try. if can't allocate, get out. */
		    }
		    pmod = plcs [link] [adapter];
		}
		card = optr->out.aonode.card;
		signal = optr->out.aonode.aosignal;
		value = optr->out.aonode.aovalue;

		if (mltrace [line] == 2)
		{ 
		    logMsg ("Analog Output to line %d, address %d, card %d signal %d,m val %d\n",
				link, adapter, card, signal, value);
		}

		if ((!pmod->cmd_force_offline)  && (resresponse [link]) && (pmod->is_responding))
		{
		    if (!setaouts (link, adapter, pmod, card, signal, value))
		    {
			pmod->badoutcnt++;
			return;
		    }
		}
	    }
	    else if ((optr->otype) == BOUTTYPE)
	    {
		link = optr->out.bonode.link;
		adapter = optr->out.bonode.adapter;

		pmod = plcs [link] [adapter];
		if (pmod == NULL)
		{
		    if (!allocate_plc (link, adapter))
		    {
			return;		/* if not allocated, try. if can't allocate, get out. */
		    }
		    pmod = plcs [link] [adapter];
		}

		card = optr->out.bonode.card;
		signal = optr->out.bonode.bosignal;
		boval = optr->out.bonode.bovalue;
		bomask = optr->out.bonode.bomask;

		if (mltrace [line] == 2)
		{ 
		    logMsg ( "Binary Output to line %d, address %d, card %d signal %d\n",
				link, adapter, card, signal);
		}
		if ((!pmod->cmd_force_offline) && (resresponse [link]) && (pmod->is_responding))
		{
		    if (!setbouts (line, adapter, pmod, card, signal, bomask, boval))
		    {
			pmod->badoutcnt++;
		    }
		}
	    }
	    oreleasefull (line);     /* release the output que entry */
	}
	return;
}

/**********
*
*      eabplc_eabplc_aidriver
*
*      This routines eabplc_driver purpose is to return the analog input value for the
*      specified module. For a module value
*      a (-1) will be returned if there is no valid data ready.
*
**********/
int	eabplc_eabplc_aidriver (char *pointname, unsigned short link, unsigned short adapter,
			unsigned short card, unsigned short signal,
			unsigned short *pvalue, unsigned short conversion,
			unsigned char *resname)
{
	ABPLC 			*pmod;
	unsigned	short 	blkix, elemix, subelement;

	if (!rangeok (link, adapter, card, signal))
	{
	    return (-1);
	}
	if (conversion != conversion)
	{
	    return (-1);        /* No warnings */
	}
	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		return (-1); /* if not allocated, try. if can't allocate, get out. */
	    }
	    pmod = plcs [link] [adapter];
	}

	pmod->was_requested = TRUE;

	if (card == PSEUDO)
	{
	    switch (signal)
	    {
	    case 0: /* Were the PLC values requested */
		*pvalue = pmod->was_requested;
		break;

	    case 1: /* Is the PLC responding */
		*pvalue = pmod->is_responding;
		break;

	    case 2: /* The program Load request */
		*pvalue = pmod->prog_load_stat;
		break;

	    case 3: /* PLC Program Load Block Number */
		*pvalue = pmod->prog_load_page;
		break;

	    case 4: /* Set the program number to Load */
		*pvalue = pmod->prog_load_number;
		break;

	    case 5: /* Output Messages */
		*pvalue = pmod->no_dataout_msgs & 0x7fff;
		break;

	    case 6: /* Input Messages */
		*pvalue = pmod->no_datain_msgs & 0x7fff;
		break;

	    case 7: /* No Receiver Responses */
		*pvalue = pmod->no_rxnoresp & 0x7fff;
		break;

	    case 8: /* No Transmitter Response */
		*pvalue = pmod->no_txnoresp & 0x7fff;
		break;

	    case 9: /* Data Echo Errors */
		*pvalue = pmod->no_dataerr & 0x7fff;
		break;

	    case 10: /* Message Retries */
		*pvalue = pmod->no_retries & 0x7fff;
		break;

	    case 11: /* RX Hardware errors */
		*pvalue = pmod->no_rxhwerrs & 0x7fff;
		break;

	    case 12: /* TX Hardware errors */
		*pvalue = pmod->no_txhwerrs & 0x7fff;
		break;

	    case 13: /* CRC errors */
		*pvalue = pmod->no_crcerrs & 0x7fff;
		break;

	    case 14: /* Offline transitions */
		*pvalue = pmod->no_offlines & 0x7fff;
		break;

	    case 15: /* reset statistics */
		*pvalue = pmod->cmd_reset_stat;
		break;

	    case 16: /* Force the PLC offline */
		*pvalue = pmod->cmd_force_offline;
		break;

	    case 17: /* Failed output commands */
		*pvalue = pmod->badoutcnt & 0x7fff;
		break;

	    default:
		return (-1);
	    }
	    return (0);
	}

	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}

	if ((card == ABTIMERS) || (card == ABCOUNTERS) || (card == ABCNTRLS))
	{
	    blkix = (signal / 3) / ABNOGRPELEMS;
	    elemix = ((signal / 3) % ABNOGRPELEMS);
	    subelement = signal % 3;
	}
	else
	{
	    blkix = (signal / ABNOGRPELEMS);
	    elemix = (signal % ABNOGRPELEMS);
	}

	setrequested (pmod, card, blkix);

	if (!resresponse [link])
	{
	    return (-1);
	}
	if (pmod->holdflag)
	{
	    return (-2);
	}
	if (!pmod->is_responding)
	{
	    return (-1);
	}
	return (fetchvalue (pmod, card, blkix, elemix, subelement, pvalue));
}

/**********
*
*	abplc_eabplc_aodriver
*
*	This routines eabplc_driver purpose is to set requested analog outputs to the
*	specified module.
*
**********/
int	eabplc_aodriver (char *pointname, unsigned short link, unsigned short adapter,
			unsigned short card, unsigned short signal,
			double dval, unsigned short value,
			unsigned short conversion, unsigned char *resname)
{
	ABPLC 			*pmod;
	unsigned	short 	blkix;
	OUTQENT			*oqp;

	if (!rangeok (link, adapter, card, signal))
	{
	    return (-1);
	}
	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		return (-1); /* if not allocated, try. if can't allocate, get out. */
	    }
	    pmod = plcs [link] [adapter];
	}
	if (card == PSEUDO)
	{
	    switch (signal)
	    {
	    case 2: /* Set the program Load request */
		if (value != 0)
		{
		    pmod->prog_load_stat = PLC_LOAD_REQUESTED;
		}
		break;

	    case 4: /* Set the program number to Load */
		pmod->prog_load_number = value;
		break;

	    case 15: /* Reset the accumulated statistics */
		if (value != 0)
		{
		    initstats (pmod);
		}
		break;

	    case 16: /* Force the PLC offline */
		pmod->cmd_force_offline = (value != 0);
		pmod->is_responding = (value != 0);
		break;

	    default:
		return (-1);
	    }
	    return (0);
	}

	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}
	if ((card == ABTIMERS) || (card == ABCOUNTERS) ||(card == ABCNTRLS))
	{
	    blkix = (signal / 3) / ABNOGRPELEMS;
	}
	else
	{
	    blkix = (signal / ABNOGRPELEMS);
	}

	setrequested (pmod, card, blkix);

	if (!resresponse [link])
	{
	    return (-2);
	}
	if ((!pmod->is_responding) || (pmod->cmd_force_offline))
	{
	    return (-1);
	}

	oqp = ogetemptyp (link);
	if (oqp == NULLP)
	{
	    return (-1); /* output que is full */
	}
	oqp->otype = AOUTTYPE;
	oqp->out.aonode.link = link;
	oqp->out.aonode.adapter = adapter;
	oqp->out.aonode.card = card;
	oqp->out.aonode.aosignal = signal;
	oqp->out.aonode.aovalue = value;

	omarkfull (link); /* Mark this entry as ready for output */

	return (0);
}

/**********
*
*      abplc_eabplc_aoread
*
*      This routines eabplc_driver purpose is to return the value last read back
*      from the analog output hardware. This is used to initialize analog
*      outputs.
*
**********/
int	eabplc_aoread (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned short *pvalue, unsigned short conversion, 
		unsigned char *resname)
{
	ABPLC			*pmod;
	unsigned	short 	blkix, elemix, subelement;

	if (!rangeok (link, adapter, card, signal))
	{
	    return (-1);
	}
	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		return (-1);   /* if not allocated, try. if can't allocate, get out. */
	    }
	    pmod = plcs [link] [adapter];
	}
	if (card == PSEUDO)
	{
	    switch (signal)
	    {
	    case 2: /* Set the program Load request */
		*pvalue = pmod->prog_load_stat;
		break;

	    case 4: /* Set the program number to Load */
		*pvalue = pmod->prog_load_number;
		break;

	    case 15: /* reset statistics */
		*pvalue = pmod->cmd_reset_stat;
		break;

	    case 16: /* Force the PLC offline */
		*pvalue = pmod->cmd_force_offline;
		break;

	    default:
		return(-1);
	    }
	    return (0);
	}

	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}

	pmod->was_requested = TRUE;

	if ((card == ABTIMERS) || (card == ABCOUNTERS) || (card == ABCNTRLS))
	{

	    blkix = (signal / 3) / ABNOGRPELEMS;
	    elemix = ((signal / 3) % ABNOGRPELEMS);
	    subelement = signal % 3;
	}
	else
	{
	    blkix = (signal / ABNOGRPELEMS);
	    elemix = (signal % ABNOGRPELEMS);
	}

	setrequested (pmod, card, blkix);

	if (!resresponse [link])
	{
		return (-1);
	}
	if (pmod->holdflag)
	{
		return (-2);
	}
	if ((!pmod->is_responding) || (pmod->cmd_force_offline))
	{
	    return (-1);
	}

	return (fetchvalue (pmod, card, blkix, elemix, subelement, pvalue));
}

/***********
*
*      abplc_eabplc_bidriver
*
*      The eabplc_driver purpose of this routine is to return the request digital input
*      value(s) from the requested card. This routine will return a (-1) if the
*      requested module is the wrong type or the data is not yet valid.
*
***********/
int	eabplc_bidriver (char *pointname, unsigned short link, unsigned short adapter,
			unsigned short card, unsigned short signal,
			unsigned long mask, unsigned long *pvalue,
			unsigned char *resname)
{
	ABPLC			*pmod;
	unsigned	short	blkix, elemix, subelement, tvallow, tvalhigh;
	unsigned	short	tsignal;
	long			retval;

	if (!plcs [link])
	{
	    return (-1);
	}
	blkix = (signal / 16) / ABNOGRPELEMS;
	elemix = ((signal / 16) % ABNOGRPELEMS);
	subelement = 0;

	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		return (-1); /* if not allocated, try. if can't allocate, get out. */
	    }
	    pmod = plcs [link] [adapter];
	}

	pmod->was_requested = TRUE;

	if (card == PSEUDO)
	{
	    switch (signal)
	    {
	    case 0: /* Were the PLC values requested */
		*pvalue = pmod->was_requested;
		break;

	    case 1: /* Is the PLC responding */
		*pvalue = pmod->is_responding;
		break;

	    case 15: /* reset statistics */
		*pvalue = pmod->cmd_reset_stat;
		break;

	    case 16: /* Force the PLC offline */
		*pvalue = pmod->cmd_force_offline;
		break;

	    case 18: /* Force the PLC offline */
		*pvalue = pmod->partner_online;
		break;

	    default:
		return (-1);
	    }

	    return (0);
	}

	if ((card == ABTIMERS) || (card == ABCOUNTERS) || (card == ABCNTRLS))
	{
	    tsignal = (signal / 16) * 3; /* adjust for skip of data */
	}
	else
	{
	    tsignal = (signal / 16);
	}
	if (!rangeok (link, adapter, card, tsignal))
	{
	    return (-1);
	}
	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}
	setrequested (pmod, card, blkix);

	if (!resresponse [link])
	{
	    return (-1);
	}
	if (pmod->holdflag)
	{
	    return (-2);
	}

	if ((!pmod->is_responding) || (pmod->cmd_force_offline))
	{
	    return (-1);
	}

	if ((fetchvalue (pmod, card, blkix, (elemix & 0xFFFE), subelement, &tvallow)) == -1)
	{
		return (-1);
	}
	if ((fetchvalue (pmod, card, blkix, (elemix | 1), subelement, &tvalhigh)) == -1)
	{
		return (-1);
	}
	retval = ((long) tvalhigh << 16);
	retval = retval + tvallow;
	*pvalue = (retval & mask);
	return (0);
}

/**********
*
*      abplc_eabplc_bodriver
*
*      The eabplc_driver purpose of this routine is to set the request digital output
*      value(s) on the requested card. This routine will return a (-1) if the
*      requested module is the wrong type or not configured yet.
*
***********/
int	eabplc_bodriver (char *pointname, unsigned short link, unsigned short adapter,
			unsigned short card, unsigned short signal,
			unsigned long value, unsigned long mask,
			unsigned char *resname)
{
	ABPLC			*pmod;
	OUTQENT			*oqp;
	unsigned	short	tsignal;
	unsigned	short	blkix;

	blkix = (signal / 16) / ABNOGRPELEMS;

	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		return (-1); /* if not allocated, try. if can't allocate, get out. */
	    }
	    pmod = plcs [link] [adapter];
	}

	pmod->was_requested = TRUE;

	if (card == PSEUDO)
	{
	    switch (signal & 0x1f)
	    {
	    case 2: /* Set the program Load request */
		if (value != 0)
		{
		    pmod->prog_load_stat = PLC_LOAD_REQUESTED;
		}
		break;

	    case 15: /* Reset the accumulated statistics */
		if (value != 0)
		{
		    initstats (pmod);
		}
		break;

	    case 16: /* Force the PLC offline */
		pmod->cmd_force_offline = (value != 0);
		pmod->is_responding = (value != 0);
		break;

	    default:
		return (-1);
	    }
	    return (0);
	}

	if ((card == ABTIMERS) || (card == ABCOUNTERS) || (card == ABCNTRLS))
	{
	    tsignal = (signal / 16) * 3; /* adjust for skip of data */
	}
	else
	{
	    tsignal = (signal / 16);
	}
	if (!rangeok (link, adapter, card, tsignal))
	{
	    return (-1);
	}
	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}

	setrequested (pmod, card, blkix);

	if (!resresponse [link])
	{
	    return (-1);
	}
	if ((!pmod->is_responding) || (pmod->cmd_force_offline))
	{
	    return (-1);
	}

	oqp = ogetemptyp (link);

	if (oqp == NULLP)
	{
	    return (-1); /* output que is full */
	}
	oqp->otype = BOUTTYPE;
	oqp->out.bonode.link = link;
	oqp->out.bonode.adapter = adapter;
	oqp->out.bonode.card = card;
	oqp->out.bonode.bosignal = signal;
	oqp->out.bonode.bomask = mask;
	oqp->out.bonode.bovalue = value;

	omarkfull (link); /* Mark this entry as ready for output */

	return (0);
}

/**********
*
*      abplc_eabplc_boread
*
*      This routines eabplc_driver purpose is to return the value last read back
*      from the digital output hardware. This is used to initialize digital
*      outputs. This will return a (-1) if the module is the wrong type or
*      the data is not yet valid
*
**********/
int	eabplc_boread (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned long *pvalue, unsigned long mask,
		unsigned char *resname)
{
	ABPLC			*pmod;
	unsigned	short	blkix, elemix, subelement, tvallow, tvalhigh;
	unsigned	short	tsignal;
	long			retval;

	if (!plcs [link])
	{
	    return (-1);
	}
	blkix = (signal / 16) / ABNOGRPELEMS;
	elemix = ((signal / 16) % ABNOGRPELEMS);
	subelement = 0;

	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		return (-1); /* if not allocated, try. if can't allocate, get out. */
	    }
	    pmod = plcs [link] [adapter];
	}

	pmod->was_requested = TRUE;

	if (card == PSEUDO)
	{
	    switch (signal)
	    {
	    case 2: /* Set the program Load request */
		*pvalue = 0;
		break;

	    case 15: /* reset statistics */
		*pvalue = pmod->cmd_reset_stat;
		break;

	    case 16: /* Force the PLC offline */
		*pvalue = pmod->cmd_force_offline;
		break;

	    default:
		return (-1);
	    }
	    return (0);
	}

	if ((card == ABTIMERS) || (card == ABCOUNTERS) || (card == ABCNTRLS))
	{
	    tsignal = (signal / 16) * 3; /* adjust for skip of data */
	}
	else
	{
	    tsignal = (signal / 16);
	}

	if (!rangeok (link, adapter, card, tsignal))
	{
	    return (-1);
	}
	if (resname != 0)
	{
	    register_resource (link, pmod, resname, pointname); 
	}
	if (!resresponse [link])
	{
	    return (-1);
	}
	if (pmod->holdflag)
	{
	    return (-2);
	}
	if ((!pmod->is_responding) || (pmod->cmd_force_offline))
	{
	    return (-1);
	}
	setrequested (pmod, card, blkix);

	if ((fetchvalue (pmod, card, blkix, (elemix & 0xFFFE), subelement, &tvallow)) == -1)
	{
	    return (-1);
	}
	if ((fetchvalue (pmod, card, blkix, (elemix | 1), subelement, &tvalhigh)) == -1)
	{
	    return (-1);
	}

	retval = ((long) tvalhigh << 16);
	retval = retval + tvallow;
	*pvalue = (retval & mask);
	return (0);
}

/***********
*
*	register_resource
*
*	This routine is used to initialize the resource pointers
*	If this is the first time the resource for this line is being
*	referenced the pointer to the resource table entry will be set
*	and the resource name will be saved. If the resouce name has
*	already been set, this routine will verify that all other points
*	for this driver are using the same resource name.
*
**********/                                                                                    	
/* this routine called from eabplc_eabplc_aidriver, eabplc_aodriver, eabplc_aoread, eabplc_boread,				
			    eabplc_bidriver, and eabplc_bodriver */                               	

static	void 	register_resource (unsigned short link, ABPLC *pmod,
				unsigned char *resname,
				unsigned char *pointname)
{
	if (resnames [link] [0] == 0) /* This is the first reference */
	{
	    if (resGetByName (resname, &prestable [link]) != RES_OK)
	    {
		logMsg ("Resource not found in table - %s, %s\n",
				pointname,
				resname );
	    }
	    else
	    {
		strcpy (resnames [link], resname);
	    }
	}
	else
	{
	    if (strcmp (resnames [link], resname) != 0)
	    {
		logMsg ("Resource name conflict - %s,  %s, %s\n",
				pointname,
				resnames [link],
				resname );
	    }
	}
}

/**********
*
*	rangeok
*
*	routine to check the boundaries for link adapter card and signal
*
**********/
static	int	rangeok (unsigned short link,unsigned short adapter,
		unsigned short card,unsigned short signal)
{
	ABPLC  *pmod;

	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		return (FALSE);			/* if can't allocate, get out. */
	    }
	    pmod = plcs [link] [adapter];
	}

	if (link >= ABLINES)
	{
	    return (FALSE);
	}
	if (adapter >= NUMABPLCS)
	{
	    return (FALSE);
	}
	if ((card > ABINTS) &&
	    (card != pmod->extintfile1) &&
	    (card != pmod->extintfile2) &&
	    (card != pmod->extintfile3) &&
	    (card != pmod->extintfile4) &&
	    (card != PSEUDO))
	    {

	    /* Look for an empty slot to define a new extended integer file */

	    if (pmod->extintfile1 == MAXFILE + 1)
	    {
		pmod->extintfile1 = card;
		return (TRUE);
	    }
	    if (pmod->extintfile2 == MAXFILE + 1)
	    {
		pmod->extintfile2 = card;
		return (TRUE);
	    }
	    if (pmod->extintfile3 == MAXFILE + 1)
	    {
		pmod->extintfile3 = card;
		return (TRUE);
	    }
	    if (pmod->extintfile4 == MAXFILE + 1)
	    {
		pmod->extintfile4 = card;
		return (TRUE);
	    }
	    return (FALSE);
	}
	if ((card == ABOUTIMG)   && (signal > MAXOUTIMG))
	{
	    return (FALSE);
	}
	if ((card == ABINIMG)    && (signal > MAXINIMG))
	{
	    return (FALSE);
	}
	if ((card == ABSTATUS)   && (signal > MAXSTATUS))
	{
	    return (FALSE);
	}
	if ((card == ABBITS)     && (signal > MAXBITS))
	{
	    return (FALSE);
	}
	if ((card == ABTIMERS)   && (signal > MAXTIMERS * 3))
	{
	    return (FALSE);
	}
	if ((card == ABCOUNTERS) && (signal > MAXCOUNTERS * 3))
	{
	    return (FALSE);
	}
	if ((card == ABCNTRLS)   && (signal > MAXCNTRLS * 3))
	{
	    return (FALSE);
	}
	if ((card >= ABINTS)     && (signal > MAXINTS))
	{
	    return (FALSE);
	}
	return (TRUE);
}

/**********
*
*	setrequested
*	routine to set the requested flag for the selected block
*
**********/
static	void	setrequested (ABPLC *pmod, unsigned short card, unsigned short blkix)
{
	if (card == ABINTS)
	{
	    pmod->abfints[blkix].was_requested = TRUE;
	    return;
	}
	if (card == ABOUTIMG)
	{
	    pmod->abfouts[blkix].was_requested = TRUE;
	    return;
	}
	if (card == ABINIMG)
	{
	    pmod->abfins[blkix].was_requested = TRUE;
	    return;
	}
	if (card == ABSTATUS)
	{
	    pmod->abfstats[blkix].was_requested = TRUE;
	    return;
	}
	if (card == ABBITS)
	{
	    pmod->abfbits[blkix].was_requested = TRUE;
	    return;
	}
	if (card == ABTIMERS)
	{
	    pmod->abftimers[blkix].was_requested = TRUE;
	    return;
	}
	if (card == ABCOUNTERS)
	{
	    pmod->abfcntrs[blkix].was_requested = TRUE;
	    return;
	}
	if (card == ABCNTRLS)
	{
	    pmod->abfcntrls[blkix].was_requested = TRUE;
	    return;
	}
	if (card == pmod->extintfile1)
	{
	    pmod->abfexints1[blkix].was_requested = TRUE;
	    return;
	}
	if (card == pmod->extintfile2)
	{
	    pmod->abfexints2[blkix].was_requested = TRUE;
	    return;
	}
	if (card == pmod->extintfile3)
	{
	    pmod->abfexints3[blkix].was_requested = TRUE;
	    return;
	}
	if (card == pmod->extintfile4)
	{
	    pmod->abfexints4[blkix].was_requested = TRUE;
	    return;
	}
}

/**********
*
*	fetchvalue
*	routine to get a value from the Allen Bradley data structures
*
**********/
static	int  fetchvalue (ABPLC *pmod, unsigned short card,
		 unsigned short blkix, unsigned short elemix,
		 unsigned short subelement, unsigned short *pvalue)
{
	if (card == ABOUTIMG)
	{
	    if (!pmod->abfouts [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfouts [blkix].oimgdata [elemix];
	    return (0);
	}
	if (card == ABINIMG)
	{
	    if (!pmod->abfins [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfins [blkix].iimgdata [elemix];
	    return (0);
	}
	if (card == ABSTATUS)
	{
	    if (!pmod->abfstats [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfstats [blkix].statdata[ elemix];
	    return (0);
	}
	if (card == ABBITS)
	{
	    if (!pmod->abfbits [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfbits [blkix].bitdata [elemix];
	    return (0);
	}
	if (card == ABTIMERS)
	{
	    if (!pmod->abftimers [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue =     pmod->abftimers [blkix].timerdata [elemix].cdata [subelement];
	    return (0);
	}
	if (card == ABCOUNTERS)
	{
	    if (!pmod->abfcntrs [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfcntrs [blkix].cntrdata [elemix].cdata [subelement];
	    return (0);
	}
	if (card == ABCNTRLS)
	{
	    if (!pmod->abfcntrls [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfcntrls [blkix].cntrldata [elemix].cdata [subelement];
	    return (0);
	}
	if (card == ABINTS)
	{
	    if (!pmod->abfints [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfints [blkix].intdata [elemix];
	    return (0);
	}
	if (card == pmod->extintfile1)
	{
	    if (!pmod->abfexints1 [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfexints1 [blkix].intdata [elemix];
	    return (0);
	}
	if (card == pmod->extintfile2)
	{
	    if (!pmod->abfexints2 [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfexints2 [blkix].intdata [elemix];
	    return (0);
	}
	if (card == pmod->extintfile3)
	{
	    if (!pmod->abfexints3 [blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfexints3 [blkix].intdata [elemix];
	    return (0);
	}
	if (card == pmod->extintfile4)
	{
	    if (!pmod->abfexints4[blkix].data_valid)
	    {
		return (-1);
	    }
	    *pvalue = pmod->abfexints4 [blkix].intdata [elemix];
	    return (0);
	}
	return (-1);
}

/**********
*
*      Initialize the Queue
*
**********/
static	void	outqueinit (unsigned short line)
{
	oq [line]->ooutix = oq [line]->oinix = 0;
}

/**********
*
*  ogetempty - get the next available slot on the output list
*
**********/
static	OUTQENT	*ogetemptyp (unsigned short line)
{
	if (((oq [line]->oinix + 1) % OUTQSIZE) != oq [line]->ooutix)
	{
	    return (&oq [line]->outque [oq [line]->oinix]);
	}
	else
	{
	    return (NULLP);
	}
}

/**********
*
* omarkfull - mark the next block as full
*
**********/
static	void	omarkfull (unsigned short line)
{
	oq [line]->oinix = ((oq [line]->oinix + 1) % OUTQSIZE);
}


/**********
*
*  ogetfullp - get the next full slot on the output list
*
**********/
static	OUTQENT	*ogetnextp (unsigned short line)
{
	if (oq [line]->ooutix != oq [line]->oinix)
	{
	    return (&oq [line]->outque [oq [line]->ooutix]);
	}
	else
	{
	    return (NULLP);
	}
}

/**********
*
* oreleasefull - release the current full output block
*
**********/
static	void	oreleasefull (unsigned short line)
{
	if (oq [line]->ooutix == oq [line]->oinix)
	{
	    return;
	}
	oq [line]->ooutix = ((oq [line]->ooutix + 1) % OUTQSIZE);
}

/***********
*
*      link_status
*
*      The purpose of this routine is to print out the state of each of the
*      analog devices lines and the statistics gathered about them.
*
**********/
static	void	link_status (unsigned short line)
{
	if (line > ABLINES - 1)
	{
	    return;
	}
}

/**********
*                                                                              
*	allocate_plc                                                            
*                                                                              
*	The purpose of this routine is to allocate the memory for a single      
*	plc data structure. If not successful, a console error message is  	
*	printed and an error flag is returned to the caller.                    
*	If successful, the newly allocated data structures are initialized.     
*                                                                             	
**********/
static	int	allocate_plc (unsigned short line, unsigned short plc )
{
	plcs [line] [plc] = (ABPLC *) calloc (1, (sizeof (ABPLC)));
	if (!plcs [line] [plc])
	{
	    logMsg ("Unable to allocate memory");
            logMsg (" for ABPLC data structure on line %d, plc %d\n",line ,plc);
	    return (FALSE);
	}
	else
	{
	    logMsg ("Allocated memory for ABPLC data structure on line %d, plc %d\n",line ,plc);
	    init_plc_tab (line, plc);	     /*	initialize the data structures. */
	}
	return (TRUE);
}

/**********
*
*      check_line_activity
*                                                                             	
*      The purpose of this routine is to process a receive character interrupt	
*      for a given line.                                                      	
*                                                                             	
**********/

/* only called from eabplc_driver  */

static	int	check_line_activity (unsigned short line, unsigned short predelay)
{
	RXLINE          *rxp;
	long		thistime;
	short		status;

	if (predelay != 0)
	{
	    taskDelay (predelay);
	}
	thistime = tickGet ();
	if ((status = remoteSerialRedund (line)) != 'M' &&
	    (status != 'W'))
	{
	    line_activity [line] = FALSE;
	}
	else
	{
	    line_activity [line] = TRUE;
	}
	if ((thistime - activity_time [line]) < ACTIVITYTIME)
	{
	    return (lastactivity [line] || (line_activity [line]));
	}
	lastactivity [line] = status = line_activity [line];
	line_activity [line] = FALSE;
	activity_time [line] = thistime;
	return (status);
}

/**********
*
*      msgxfer
*
*      The purpose of this routine is to transmit a buffer to the
*      Allen Bradley PLCs and to await a response. The transmit and receive
*      buffers are in their respective line tables. This routine will retry
*      the message up to a maximum number of times or until a data response
*      is received with a proper checksum.
*              returns         XFERGOOD if response received
*                              XFERBAD  if no good repsonse is received
*
**********/
static	unsigned	short	msgxfer (unsigned short fline,
				unsigned short resptype, ABPLC *pmod)
{
	unsigned short  retry, rxlen, no_char_tmout;
	unsigned short	line, wrongreply, wrflag, ix;
	short		rxstatus;
	RXLINE          *rxp;
	TXLINE          *txp;
	unsigned char   csums [2];
	unsigned char	txlnbuf [512];
	unsigned char	rxlnbuf [512];

	line = (fline & 0x1f);          /* strip off the no retry flag*/

	rxp = rxlinetab [line];
	txp = txlinetab [line];

	wrongreply = 0;
	no_char_tmout = 0;

	for (retry = ((fline == line) ? 0 : MAXRETRYS - 1); (retry < MAXRETRYS); retry++ )
	{
	    if (intermsgdelay [line] != 0)
	    {
		taskDelay (intermsgdelay [line]);
	    }

	    wrflag = FALSE;
	    if (linemon [line])
	    {
		txlnbuf [0] = 0;
		for (ix = 0; ix < txp->txmsglen; ix++)
		{
		    sprintf (&txlnbuf [ix * 3], "%02x ", txp->txbuf [ix]);
		}
		txlnbuf [ix * 3] = 0;
		logMsg ("TX: %s\n", txlnbuf);
	    }

	    rxstatus = enetXfer (txp->txbuf, txp->txmsglen, rxp->rxbuf, MAXRXMSG, eabplc_ip [line], 4000, line);

	    if ((rxstatus != -1) && (rxstatus != 0))
	    {
		if (resptype == NOREPLY)
		{
		    return (XFERGOOD);
		}

		if (linemon [line])
		{
		    rxlnbuf [0] = 0;
		    for (ix = 0; ix < rxp->rxix; ix++)
		    {
			sprintf (&rxlnbuf [ix * 3], "%02x ", rxp->rxbuf [ix]);
		    }
		    rxlnbuf [ix * 3] = 0;
		    logMsg ("RX: %s\n", rxlnbuf);
		}

		rxlen = rxp->rxmsglen = rxstatus;
		if (rxp->rxbuf [4] == resptype)
		{
		    return (XFERGOOD);
		}
		else
		{
		    wrflag = TRUE;
		    wrongreply++;
		}
	    }
	    else if (rxstatus == 0)
	    {
		if (modtrace [line])
		{
		    logMsg ("Receive Timeout... Length: %d\n", rxlen);
		}
		pmod->no_rxnoresp++;
	    }
	    else
	    {
		if (modtrace [line])
		{
		    logMsg ("Receive Error - Connection not established\n");
		}
		pmod->no_rxhwerrs++;
	    }
	    if ((retry != MAXRETRYS) && (!wrflag))
	    {
		pmod->no_retries++;
	    }
	}
	if (modtrace [line])
	{
	    logMsg ("Xfer Bad\n");
	}
	return (XFERBAD);
}

/**********
*
*      init_plc_tab
*
*      The purpose of this routine is to initialize the line tables for the
*      Allen Bradley I/O
*
**********/

/* only called from allocate_plc  */

static	void	init_plc_tab (unsigned short line, unsigned short mod)
{
	ABPLC  *pmod;

	pmod = plcs [line] [mod];
	if (!pmod)
	{
		return;
	}
	pmod->is_responding = FALSE;
	pmod->holdflag = FALSE;
	pmod->was_requested = FALSE;
	pmod->cmd_force_offline = FALSE;
	pmod->prog_load_stat = PLC_NOT_LOADED;
	pmod->prog_load_number = 0;
	pmod->prog_load_page = 0;

	initplc (pmod);
	initstats (pmod);
}

/**********
*
* Cold Init a PLC
*
**********/

/* only called from init_line_tabs  */

static	void	initplc (ABPLC *pmod)
{
	unsigned	short	i;

	for (i = 0; (i < NUMABOUTS); i++)
	{
	    pmod->abfouts [i].was_requested = FALSE;
	    pmod->abfouts [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINS); i++)
	{
	    pmod->abfins [i].was_requested = FALSE;
	    pmod->abfins [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABSTATS); i++)
	{
	    pmod->abfstats [i].was_requested = FALSE;
	    pmod->abfstats [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABBITS); i++)
	{
	    pmod->abfbits [i].was_requested = FALSE;
	    pmod->abfbits [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABTMRS); i++)
	{
	    pmod->abftimers [i].was_requested = FALSE;
	    pmod->abftimers [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABCNTRS); i++)
	{
	    pmod->abfcntrs [i].was_requested = FALSE;
	    pmod->abfcntrs [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABCNTRLS); i++)
	{
	    pmod->abfcntrls [i].was_requested = FALSE;
	    pmod->abfcntrls [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINTS); i++)
	{
	    pmod->abfints [i].was_requested = FALSE;
	    pmod->abfints [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINTS); i++)
	{
	    pmod->abfexints1 [i].was_requested = FALSE;
	    pmod->abfexints1 [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINTS); i++)
	{
	    pmod->abfexints2 [i].was_requested = FALSE;
	    pmod->abfexints2 [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINTS); i++)
	{
	    pmod->abfexints3 [i].was_requested = FALSE;
	    pmod->abfexints3 [i].data_valid = FALSE;
	}
	for (i = 0; (i < NUMABINTS); i++)
	{
	    pmod->abfexints4 [i].was_requested = FALSE;
	    pmod->abfexints4 [i].data_valid = FALSE;
	}
	pmod->extintfile1 = MAXFILE + 1;
	pmod->extintfile2 = MAXFILE + 1;
	pmod->extintfile3 = MAXFILE + 1;
	pmod->extintfile4 = MAXFILE + 1;
}

/**********
*
* initialize the statistics for an ALLEN BRADLEY PLC
*
**********/
static	void	initstats (ABPLC *pmod)
{
	pmod->no_dataout_msgs = 0;      /* 5  Count Output Msgs */
	pmod->no_datain_msgs = 0;       /* 6  Count Data Requests */
	pmod->no_rxnoresp = 0;          /* 7  Count No Responses */
	pmod->no_txnoresp = 0;          /* 8  Count No Responses */
	pmod->no_dataerr = 0;           /* 9  Count Message Echo Errors */
	pmod->no_retries = 0;           /* 10 Count Retries */
	pmod->no_rxhwerrs = 0;          /* 11 Count Comm Hardware Errors */
	pmod->no_txhwerrs = 0;          /* 12 Count Comm Hardware Errors */
	pmod->no_crcerrs = 0;           /* 13 Count CRC Errors */
	pmod->no_offlines = 0;          /* 14 Count Off Line Transitions */
	pmod->badoutcnt = 0;		/* 17 Count failed output commands */

        pmod->tscanpass = 0;            /* plc scan pass counter            */
        pmod->msgs = 0;                 /* good response message counts     */
        pmod->responsetime = 0;         /* total wait time for good messages*/

}

/**********
*	prepresponses                                                            
*                                                                             	
*       The purpose of this routine is to mark the PLCs as on hold and the    	
*	data as not valid until read again from the PLC                       	
*                                                                           	
**********/

/* only called from eabplc_driver */

static	void	prepresponses (unsigned short line)
{
	unsigned	short 	mod;
	ABPLC			*pmod;

	for (mod = 1; (mod <= NUMABPLCS); mod++)
	{
            pmod = plcs [line] [mod];
	    if (pmod)			/* verify pointer validity prior to use */
	    {
	    	pmod->holdflag = TRUE;
		mark_not_valid (pmod);
            }
	}
}

/**********
*
*      any_responding                                               	
*                                                                   	
*      The purpose of this routine is to see if any units are       	
*	responding on this line                                      	
*                                                                   	
**********/

/* only called from eabplc_driver */

static	int	any_responding (unsigned short line)
{
	unsigned short  mod;
	ABPLC  *pmod;

	for (mod = 1; (mod <= NUMABPLCS); mod++)
	{
	    pmod = plcs [line] [mod];
	    if (pmod)			/* verify pointer validity prior to use */
	    {
		if (pmod->is_responding)
		{
		    return (TRUE);
		}
	    }
	}
	return (FALSE);
}

/**********
*
*	Readabouts
*
*	Reads the Allen Bradley Output Image File
*
**********/
static	unsigned	short	readabouts (unsigned short line,
				unsigned short mod, ABPLC *pmod,
				unsigned short group, unsigned short numgroups)
{
	unsigned	short 	wval, elementno, numelements;
	unsigned	short	dataix, i, j, lastgroup;
	unsigned	char	data [MAXRXMSG];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Get the data */

	elementno = ABNOGRPELEMS * group;
	numelements = ABNOGRPELEMS * numgroups;

	if (!(abgetdata ( line, mod, pmod, ABOUTIMG,
			 elementno, numelements, data)))
	{
	    return (FALSE);
	}
	lastgroup = group + numgroups - 1;
	dataix = 0;
	for (i = group; i <= lastgroup; i++)
	{
	    for (j = 0; j < ABNOGRPELEMS; j++)
	    {
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abfouts [i].oimgdata [j] = wval;
	    }
	    pmod->abfouts [i].data_valid = TRUE;
	}
	return (TRUE);
}

/**********
*
*	Readabins
*	Reads the Allen Bradley Input Image File
*
**********/
static	unsigned	short	readabins (unsigned short line,
				unsigned short mod, ABPLC *pmod,
				unsigned short group, unsigned short numgroups)
{
	unsigned	short	wval, elementno, numelements;
	unsigned	short	dataix, i, j, lastgroup;
	unsigned	char	data [MAXRXMSG];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Get the data */

	elementno = ABNOGRPELEMS * group;
	numelements =  ABNOGRPELEMS * numgroups;

	if (!(abgetdata (line, mod, pmod, ABINIMG,
			 elementno, numelements, data)))
	{
	    return (FALSE);
	}
	lastgroup = group + numgroups - 1;
	dataix = 0;
	for (i = group; i <= lastgroup; i++)
	{
	    for (j = 0; j < ABNOGRPELEMS; j++)
	    {
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abfins [i].iimgdata [j] = wval;
	    }
	    pmod->abfins [i].data_valid = TRUE;
	}
	return (TRUE);
}

/**********
*
*	Readabstats
*	Reads the Allen Bradley Status File
*
**********/
static	unsigned	short	readabstats (unsigned short line,
				unsigned short mod, ABPLC *pmod,
				unsigned short group, unsigned short numgroups)
{
	unsigned	short	wval, elementno, numelements;
	unsigned	short	dataix, i, j, lastgroup;
	unsigned	char	data [MAXRXMSG];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Get the data */

	elementno = ABNOGRPELEMS * group;
	numelements =  ABNOGRPELEMS * numgroups;

	if (!(abgetdata ( line, mod, pmod, ABSTATUS,
			 elementno, numelements, data)))
	{
	    return (FALSE);
	}
	lastgroup = group + numgroups - 1;
	dataix = 0;
	for (i = group; i <= lastgroup; i++)
	{
	    for (j = 0; j < ABNOGRPELEMS; j++)
	    {
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abfstats [i].statdata [j] = wval;
	    }
	    pmod->abfstats [i].data_valid = TRUE;
	}
	return (TRUE);
}

/**********
*
*	Readabbits
*	Reads the Allen Bradley Bit File
*
**********/
static	unsigned	short	readabbits (unsigned short line,
				unsigned short mod, ABPLC *pmod,
				unsigned short group, unsigned short numgroups)
{
	unsigned	short	wval, elementno, numelements;
	unsigned	short	dataix, i, j, lastgroup;
	unsigned	char	data [MAXRXMSG];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Get the data */

	elementno = ABNOGRPELEMS * group;
	numelements =  ABNOGRPELEMS * numgroups;

	if (!(abgetdata ( line, mod, pmod, ABBITS,
			 elementno, numelements, data)))
	{
	    return (FALSE);
	}
	lastgroup = group + numgroups - 1;
	dataix = 0;
	for (i = group; i <= lastgroup; i++)
	{
	    for (j = 0; j < ABNOGRPELEMS; j++)
	    {
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abfbits [i].bitdata [j] = wval;
	    }
	    pmod->abfbits [i].data_valid = TRUE;
	}
	return (TRUE);
}

/**********
*
*	Readabintss
*	Reads an Allen Bradley Integer File
*
**********/
static	unsigned short readabints (unsigned short line, unsigned short mod,
			  ABPLC *pmod, unsigned short fileno ,ABINTSS abfints[],
			  unsigned short group,unsigned short numgroups)
{

	unsigned	short	wval, elementno, numelements;
	unsigned	short	dataix, i, j, lastgroup;
	unsigned	char	data [MAXRXMSG];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Get the data */

	elementno = ABNOGRPELEMS * group;
	numelements =  ABNOGRPELEMS * numgroups;

	if (!(abgetdata (line, mod, pmod, fileno,
			 elementno, numelements, data)))
	{
	    return (FALSE);
	}
	lastgroup = group + numgroups - 1;
	dataix = 0;
	for (i = group; i <= lastgroup; i++)
	{
	    for (j = 0; j < ABNOGRPELEMS; j++)
	    {
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		abfints [i].intdata [j] = wval;
	    }
	    abfints [i].data_valid = TRUE;
	}
	return (TRUE);
}

/**********
*
*	Readabtimers
*	Reads the Allen Bradley Timer File
*
**********/
static	unsigned short readabtimers (unsigned short line, unsigned short mod,
			  ABPLC *pmod,unsigned short group,
			  unsigned short numgroups)
{
	unsigned	short	wval, elementno, numelements;
	unsigned	short	dataix, i, j, lastgroup;
	unsigned	char	data [MAXRXMSG];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Get the data */

	elementno = ABNOGRPELEMS * group;
	numelements =  ABNOGRPELEMS * numgroups;

	if (!(abgetdata (line, mod, pmod, ABTIMERS,
			 elementno, numelements, data)))
	{
	    return (FALSE);
	}
	lastgroup = group + numgroups - 1;
	dataix = 0;
	for (i = group; i <= lastgroup; i++)
	{
	    for (j = 0; j < ABNOGRPELEMS; j++)
	    {
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abftimers [i].timerdata [j].cdata[0] = wval;
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abftimers [i].timerdata [j].cdata [1] = wval;
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abftimers [i].timerdata [j].cdata [2] = wval;
	    }
	    pmod->abftimers [i].data_valid = TRUE;
	}
	return (TRUE);
}

/**********
*
*	Readabcntrs
*	Reads the Allen Bradley Counter File
*
**********/
static	unsigned short readabcntrs (unsigned short line, unsigned short mod,
			  ABPLC *pmod,unsigned short group,
			  unsigned short numgroups)
{
	unsigned	short	wval, elementno, numelements;
	unsigned	short	dataix, i, j, lastgroup;
	unsigned	char	data [MAXRXMSG];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Get the data */

	elementno = ABNOGRPELEMS * group;
	numelements =  ABNOGRPELEMS * numgroups;

	if (!(abgetdata (line, mod, pmod, ABCOUNTERS,
			 elementno, numelements, data)))
	{
	    return (FALSE);
	}
	lastgroup = group + numgroups - 1;
	dataix = 0;
	for (i = group; i <= lastgroup; i++)
	{
	    for (j = 0; j < ABNOGRPELEMS; j++)
	    {
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abfcntrs [i].cntrdata [j].cdata [0] = wval;
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abfcntrs [i].cntrdata [j].cdata [1] = wval;
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abfcntrs [i].cntrdata [j].cdata [2] = wval;
	    }
	    pmod->abfcntrs [i].data_valid = TRUE;
	}
	return (TRUE);
}

/**********
*
*	Readabcntrls
*	Reads the Allen Bradley Controllers File
*
**********/
static	unsigned short readabcntrls (unsigned short line, unsigned short mod,
			  ABPLC *pmod,unsigned short group,
			  unsigned short numgroups)
{
	unsigned	short 	wval, elementno, numelements;
	unsigned	short	dataix, i, j, lastgroup;
	unsigned	char	data [MAXRXMSG];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Get the data */

	elementno = ABNOGRPELEMS * group;
	numelements =  ABNOGRPELEMS * numgroups;

	if (!(abgetdata (line, mod, pmod, ABCNTRLS,
			 elementno, numelements, data)))
			 {
	    return (FALSE);
	}
	lastgroup = group + numgroups - 1;
	dataix = 0;
	for (i = group; i <= lastgroup; i++)
	{
	    for (j = 0; j < ABNOGRPELEMS; j++)
	    {
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abfcntrls [i].cntrldata [j].cdata [0] = wval;
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abfcntrls [i].cntrldata [j].cdata [1] = wval;
		wval = data [dataix++];
		wval += ((unsigned short) data [dataix++] << 8);
		pmod->abfcntrls [i].cntrldata [j].cdata [2] = wval;
	    }
	    pmod->abfcntrls [i].data_valid = TRUE;
	}
	return (TRUE);
}

/**********
*
*	Setaouts
*	Set analog outputs in Allen Bradley PLC files.
*
**********/
static	unsigned short setaouts (unsigned short line, unsigned short mod,
			 ABPLC *pmod, unsigned short card,
			 unsigned short signal, unsigned short val)
{
	TXLINE  		*txp;
	unsigned	short	ix, element, subelement, checksum;
	unsigned	char	cval, retbuf [50];
	int			retcnt;
	unsigned long	msgstart;

	msgstart = tickGet();

	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];

	/* Count number of data request messages */

	pmod->no_dataout_msgs++;

	/* Compute the element and subelement portions of the Logical Address */

	if ( (card == ABTIMERS) ||
	     (card == ABCNTRLS) ||
	     (card == ABCNTRLS) )
	{
	     element = (signal / 3);
	     subelement = (signal % 3);
	}
	else
	{
	     element = signal;
	     subelement = 0;
	}

	/* Format the output message */

	ix = 0;

	checksum = 0;

	abheader (&txp->txbuf [0], &ix, &checksum, pmod, line, mod, ABWWORDS,
		TRUE, 0 /* pkt */, 1 /* ttrans */);

	ablogaddr (pmod, &txp->txbuf [0], &ix, &checksum, card, element, subelement /*subelement*/);

	cval = ((val & 0xFF));		/* Value */
	txp->txbuf [ix++]  = cval;
	cval =  (val >> 8);		/* Value */
	txp->txbuf [ix++]  = cval;

	/* Bump TNS */

	pmod->txtsn++;

	/* Time to send the message and get an ACK */

	txp->txmsglen = ix;
	if (msgxfer ( line, AB5CMDRESP, pmod) == XFERBAD)
	{
	    pmod->no_offlines++;
	    pmod->badoutcnt++;
	    pmod->holdflag = TRUE;
	    pmod->is_responding = FALSE;
	    return (FALSE);
	}

	pmod->msgs++;
	pmod->responsetime += (tickGet() - msgstart);

	return (TRUE);
}

/**********
*
*	Setbouts
*	Set binary outputs in Allen Bradley PLC files.
*
**********/
static	unsigned short setbouts (unsigned short line, unsigned short mod,
			 ABPLC *pmod, unsigned short card,
			 unsigned short signal,
			 long mask, long val)
{
	TXLINE			*txp;
	unsigned	short	ix, checksum, baseelement, element, subelement;
	unsigned	short	tmask, tvalue, andmask, ormask;
	unsigned	short	blkix, elemix, subelemix, tvallow, tvalhigh;
	unsigned	char 	cval, retbuf [50];
	int			retcnt;
	unsigned	char	addrset = FALSE;
	unsigned long	msgstart;

	msgstart = tickGet();


	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];

	/* Count number of data request messages */

	pmod->no_dataout_msgs++;

	subelement = 0;
	baseelement = ((signal / 16) & 0xFFFE);

	/* Format the output message */

	ix = 0;

	checksum = 0;

	abheader (&txp->txbuf [0], &ix, &checksum, pmod, line, mod, ABWBITSFNC,
	       FALSE, 0 /* pkt */, 1 /* ttrans */);

	if ((mask & 0xFFFF) != 0)
	{
	    element = baseelement;
	    tvalue = val & 0xFFFF;
	    tmask = mask & 0xFFFF;
	    andmask = (~tmask | tvalue);
	    ormask = (tmask & tvalue);
	    ablogaddr (pmod, &txp->txbuf [0], &ix, &checksum, card,
				element, 0 /*subelement*/);

	    cval =  (andmask & 0xFF);			/* And Mask low */
	    txp->txbuf [ix++]  = cval;
	    cval =  (andmask >> 8);			/* And Mask High */
	    txp->txbuf [ix++]  = cval;
	    cval =  (ormask & 0xFF);		/* Or Mask low */
	    txp->txbuf [ix++]  = cval;
	    cval =  (ormask >> 8);			/* Or Mask High */
	    txp->txbuf [ix++]  = cval;
	}
	if ((mask & 0xFFFF0000) != 0)
	{
	    element = baseelement + 1;
	    tvalue = val >> 16;
	    tmask = mask >> 16;
	    andmask = (~tmask | tvalue);
	    ormask = (tmask & tvalue);

	    ablogaddr (pmod, &txp->txbuf [0], &ix, &checksum, card,
				element, subelement /*subelement*/);

	    cval =  (andmask & 0xFF);		/* And Mask low */
	    txp->txbuf [ix++]  = cval;
	    cval =  (andmask >> 8);		/* And Mask High */
	    txp->txbuf [ix++]  = cval;
	    cval =  (ormask & 0xFF);		/* Or Mask low */
	    txp->txbuf [ix++]  = cval;
	    cval =  (ormask >> 8);		/* Or Mask High */
	    txp->txbuf [ix++]  = cval;
	}

	/* Bump TNS */

	pmod->txtsn++;

	/* Time to send the message and get an ACK */

	txp->txmsglen = ix;
	if (msgxfer (line, AB5CMDRESP, pmod) == XFERBAD)
	{
	    pmod->no_offlines++;
	    pmod->badoutcnt++;
	    pmod->holdflag = TRUE;
	    pmod->is_responding = FALSE;
	    return (FALSE);
	}

	pmod->msgs++;
	pmod->responsetime += (tickGet() - msgstart);

	return (TRUE);
}

/**********
*
*	Routine to fetch data from an Allen Bradley PLC
*
**********/
static	int abgetdata ( unsigned short fline, unsigned short mod,
		ABPLC *pmod, unsigned short file,
		unsigned short element, unsigned short cnt,
		unsigned char *retbuf)
{
    int			bytecnt;
    unsigned	short	ix, outix, checksum, polltries ,line;
    TXLINE		*txp;
    RXLINE		*rxp;
    unsigned	char	cval;
    unsigned long	msgstart;

	msgstart = tickGet();


	if ((file ==  ABOUTIMG) || (file ==  ABINIMG) ||
	    (file ==  ABSTATUS) || (file ==  ABBITS) ||
	    (file ==  ABINTS) || (file ==  ABOUTIMG) ||
	    (file >= 9) )
	{
	    bytecnt = (cnt * 2);
	}
	else
	{
	    bytecnt = (cnt * 6);
	}
	line = fline & (~ONERETRY);

	txp = txlinetab [line];
	rxp = rxlinetab [line];

	ix = 0;

	checksum = 0;

	abheader (&txp->txbuf [0], &ix, &checksum, pmod, line, mod, ABRWORDS,
	       TRUE, 0 /* pkt */, (bytecnt / 2) /* ttrans */);

	ablogaddr (pmod, &txp->txbuf [0], &ix, &checksum, file, element, 0 /*subelement*/);

	cval =  (bytecnt);       		/* Total Trans  Byte Count */
	txp->txbuf [ix++]  = cval;

	/* Bump TNS */

	pmod->txtsn++;

	/* Time to send the message and get an ACK */

	txp->txmsglen = ix;
	if (msgxfer (fline, AB5CMDRESP, pmod) == XFERBAD)
	{
	    pmod->no_offlines++;
	    pmod->holdflag = TRUE;
	    pmod->is_responding = FALSE;
	    return (FALSE);
	}

	pmod->msgs++;
	pmod->responsetime += (tickGet() - msgstart);

	for (ix = 8; ix < rxp->rxmsglen; ix++)
	{
	    retbuf [ix - 8] = rxp->rxbuf [ix];
	}
	return (TRUE);
}


/**********
*
* Routine to format an Allen Bradley transmission header
*
**********/
static	void abheader ( unsigned char *buf, unsigned short *ix, unsigned short *checksum,
	       ABPLC *pmod, unsigned short line, unsigned short mod, unsigned short fnc,
	       unsigned short pktnttans, unsigned short pkt, unsigned short ttrans)
{
	unsigned	char	cval;

	buf [(*ix)++] = 0;		/* Magic */
	buf [(*ix)++] = 5;		/* Magic */
	buf [(*ix)++] = 0;		/* Magic */
	buf [(*ix)++] = 0;		/* Magic */
	buf [(*ix)++] = AB5CMD;		/* CMD */
	buf [(*ix)++] = 0;         	/* Status Byte Ok */
	cval = (pmod->txtsn & 0xFF);	/* TNS Low */
	buf [(*ix)++] = cval;
	cval = (pmod->txtsn >> 8);	/* TNS High */
	buf [(*ix)++] = cval;
	cval = fnc;                       	/* FNC */
	buf [(*ix)++] = cval;
	if (!pktnttans)
	{
	    return;			/* Exit early if not required */
	}
	cval = (pkt & 0xFF);		/* Packet Offset Low */
	buf [(*ix)++] = cval;
	cval = (pkt >> 8);		/* Packet Offset High */
	buf [(*ix)++] = cval;
	cval = (ttrans & 0xFF);		/* Total Trans Low */
	buf [(*ix)++] = cval;
	cval = (ttrans >> 8);       	/* Total Trans High */
	buf [(*ix)++] = cval;
   }

/**********
*
* Routine to send diagnostic status messsage to AB device
*
**********/
static	int abdiagstat (ABPLC *pmod, unsigned short line, unsigned short mod)
{
	unsigned	char	cval;
	unsigned	short	ix = 0;
	unsigned	short	checksum = 0;
	unsigned	short	outix;
	int			bytecnt = 1;
	TXLINE			*txp;
	RXLINE			*rxp;

	txp = txlinetab [line];
	txp->txbuf [ix++] = 0;		/* Magic */
	txp->txbuf [ix++] = 5;		/* Magic */
	txp->txbuf [ix++] = 0;		/* Magic */
	txp->txbuf [ix++] = 0;		/* Magic */
	txp->txbuf [ix++] = ABDIAGCMD;
	cval = 0;		
	txp->txbuf [ix++] = 0;		/* Status Byte Ok */
	cval = (pmod->txtsn & 0xFF);	/* TNS Low */
	txp->txbuf [ix++] = cval;
	cval = (pmod->txtsn >> 8);	/* TNS High */
	txp->txbuf [ix++] = cval;
	cval = ABDSTATFNC;	
	txp->txbuf [ix++] = cval;	/* FNC */

	/* Bump TNS */

	pmod->txtsn++;

	/* Time to send the message and get an ACK */

	txp->txmsglen = ix;
	if (msgxfer ((ONERETRY | line), ABDIAGRESP, pmod) == XFERBAD)
	{
	    pmod->no_offlines++;
	    pmod->holdflag = TRUE;
	    pmod->is_responding = FALSE;
	    return (FALSE);
	}
	return (TRUE);
}

/**********
*
* Routine to enter Allen Bradley logical address into buffer
*
**********/
static	void	ablogaddr (ABPLC *pmod, unsigned char *buf, unsigned short *ix,
			unsigned short *checksum, unsigned short file,
			unsigned short element, unsigned short subelement)
{
	unsigned	char	cval;

	buf [(*ix)++] = FULLADDR;		/* Full Address (4 parts) */ 
	buf [(*ix)++] = DATATABLE;		/* Data Table */
	cval = file;       			/* File Number */
	buf [(*ix)++] = cval;
	buf [(*ix)++] = FULLELEMFLAG;		/* Flag Full Element Addr */
	cval = (element & 0xFF);       		/* Element Number Low */
	buf [(*ix)++] = cval;
	cval = (element >> 8);            	/* Element Number High */
	buf [(*ix)++] = cval;
	cval = subelement;	       		/* Sub Element Address */
	buf [(*ix)++] = cval;
}

/**********
*
*      dump_mod_table
*
*      The purpose of this routine is to dump the module description
*      portion of the analog devices driver tables for the selected line.
*      It will print out two lines for each module which is responding.
*      The first line will give the general module header, the second line
*      will give module type specific info.
*
**********/
void	abplc_dump (unsigned short line,unsigned short addr)
{
	ABPLC		*pmod;
	short           ix;
        unsigned long   microsecs;

	pmod = plcs [line] [addr];

	if (!pmod)
	{
	    logMsg ("Line/ModbusAddress Not Defined \n");
	    return;
	}

	printf (" Was requested = %d \n", pmod->was_requested);
	printf (" Is responding = %d \n", pmod->is_responding);
	printf (" Resource Responding = %d \n", resresponse[line]);
	printf (" Hold flag  = %d \n", pmod->holdflag );
	printf (" Force Offline = %d \n", pmod->cmd_force_offline);

	printf (" # Output messages = %d \n", pmod->no_dataout_msgs & 0x7fff);
	printf (" # Data in messages = %d \n", pmod->no_datain_msgs & 0x7fff);
	printf (" #No Rx Response = %d \n", pmod->no_rxnoresp & 0x7fff);
	printf (" #No Tx Responses = %d \n", pmod->no_txnoresp & 0x7fff);
	printf (" #Data Errors = %d \n", pmod->no_dataerr & 0x7fff);
	printf (" #Retries = %d \n", pmod->no_retries & 0x7fff);
	printf (" #Rx Hw Errors = %d \n", pmod->no_rxhwerrs & 0x7fff);
	printf (" #Tx Hw Errors = %d \n", pmod->no_txhwerrs & 0x7fff);
	printf (" #Crc Errors = %d \n", pmod->no_crcerrs & 0x7fff);
	printf (" #Off Lines = %d \n", pmod->no_offlines & 0x7fff);
	printf (" #Failed Outputs = %d \n", pmod->badoutcnt & 0x7fff);

	printf ("\n");
	printf (" # of Messages %d \n", pmod->msgs);
	printf (" # Total response time (ticks) = %lu \n", pmod->responsetime);
	printf ("   Average response time (ticks) = %d \n", (pmod->responsetime/pmod->msgs));
        microsecs = (16667*pmod->responsetime)/pmod->msgs;
	printf ("   Average response time (msec) = %d.%1d \n", microsecs/1000, (microsecs%1000)/100);
	printf (" # of scan passes = %d \n", pmod->tscanpass);
}

/**********
*
*      abplc_clear
*
*      The purpose of this routine is to clear the statistics for the
*      specified line and address.
*
**********/
void	abplc_clear (unsigned short line,unsigned short addr)
{
	ABPLC		*pmod;

	pmod = plcs [line] [addr];

	if (!pmod)
	{
	    logMsg ("Line / EABPLC Address Not Defined\n");
	    return;
	}

	initstats (pmod); /* use existing procedure */
}

/****************************************************************/
/*								*/
/*	CONVERT DATA TO ALLEN-BRADLEY (i.e., little) ENDIAN	*/
/*								*/
/*	This routine is written to work on either big or	*/
/*	little endian processors.				*/
/*								*/
/****************************************************************/

static short abShort(inpShort)
short	inpShort;
{
    short	endian = 0x1234;

    if (*(char *)&endian != 0x12)
	return (inpShort);
    else
	return ( ((inpShort >> 8) & 0xff) | ((inpShort << 8)  & 0xff00) );
}


static int abInt(inpInt)
int	inpInt;
{
    short	endian = 0x1234;

    if (*(char *)&endian != 0x12)
	return (inpInt);
    else
	return (  ((inpInt >> 24) & 0xff)
		| ((inpInt >> 8)  & 0xff00)
		| ((inpInt << 8)  & 0xff0000)
		| ((inpInt << 24) & 0xff000000)
	       );
}

/****************************************************************/
/*								*/
/*		INITIALIZE PLC-5 DATA STRUCTURE			*/
/*								*/
/****************************************************************/

static void *enetInit(inetString)
char	*inetString;			/* In "dot" notation */
{
    struct ab_plc5	*plc5;

    plc5 = (struct ab_plc5 *) malloc(sizeof(struct ab_plc5));
    if (!plc5)
	return (plc5);
    plc5 -> inetAddr = inet_addr(inetString);
    plc5 -> sockFd = -1;
    plc5 -> sessionID = -1;

    return (plc5);
}

/****************************************************************/
/*								*/
/*	SEND DATA BLOCK TO PLC-5 AND READ RESPONSE		*/
/*								*/
/*	Return a -1 if we don't have a connection and can't	*/
/*	create one.						*/
/*	Return a zero on timeout.				*/
/*	Else, return the number of bytes read.			*/
/*								*/
/****************************************************************/

static int enetXfer(outBuff, outBytes, inBuff, inBytes, plc5, timeOut, line)
char		*outBuff;			/* Output Buffer */
int		outBytes;			/* Output Buffer Size */
char		*inBuff;			/* Reply Buffer */
int		inBytes;			/* Input Buffer Size */
struct ab_plc5	*plc5;				/* Where we keep socket data */
int		timeOut;			/* Milliseconds */
unsigned short	line;
{

    struct sockaddr_in	plcAddr;
    int			intSize;
    PCCHEADER		header;
    short		abLength;
    struct timeval	tmOut;
    fd_set		fdMask;
    char		xmitBuff[MAXTXMSG + sizeof(header)];

    /*
    First, do we have a socket?
    */

    if (plc5 -> sockFd == -1)
	{

	/*
	No. Create one and try to connect it to the plc.
	*/

	plc5 -> sockFd = socket(PF_INET, SOCK_STREAM, 0);
	if (plc5 -> sockFd == ERROR)
	    {
	    plc5 -> sockFd = -1;
	    return (-1);
	    }
	/*
	Set up a mask for the fd.
	*/

	FD_ZERO (&fdMask);
	FD_SET (plc5->sockFd, &fdMask);

	/*
	Set up the timeout
	*/

	tmOut.tv_sec = enet_tmout [line];
	tmOut.tv_usec = 0;

	intSize = 1;					/* Non-blocking I/O! */
	ioctl (plc5->sockFd, FIONBIO, (int) &intSize);

	plcAddr.sin_family = AF_INET;
	plcAddr.sin_addr.s_addr = plc5->inetAddr;
	plcAddr.sin_port = htons(2222);
	if (connect (plc5->sockFd, (struct sockaddr *) &plcAddr, sizeof (plcAddr)) == ERROR)
	{
	    if (errno != EINPROGRESS)
	    {
		close(plc5 -> sockFd);
		plc5 -> sockFd = -1;
		return (-1);
	    }
	}
	if (select (plc5->sockFd + 1, &fdMask, NULL, NULL, &tmOut) <= 0)
	{
	    close (plc5 -> sockFd);
	    plc5 -> sockFd = -1;
	    return (-1);
	}

	intSize = 1;				/* Don't coalesce buffers! */
	setsockopt(plc5 -> sockFd, IPPROTO_TCP, TCP_NODELAY, &intSize, sizeof(intSize));
	intSize = 1;
	setsockopt(plc5 -> sockFd, SOL_SOCKET, SO_KEEPALIVE, &intSize, sizeof(intSize));
	plc5 -> rxBuffTopped = 0;
	intSize = sizeof(plc5 -> rxBuffSize);
	getsockopt(plc5 -> sockFd, SOL_SOCKET, SO_RCVBUF, &(plc5 -> rxBuffSize), &intSize);
	intSize = sizeof(plc5 -> txBuffSize);
	getsockopt(plc5 -> sockFd, SOL_SOCKET, SO_SNDBUF, &(plc5 -> rxBuffSize), &intSize);


	/*
	Try to get a session ID.
	*/

	memset((void *) &header, 0, sizeof(header));
	header.flag = FLAG_REQUEST;
	header.command = COMMAND_NEW_SESSION;
	*(short *)(header.custom) = htons(4);
	*(short *)(header.custom + 2) = htons(5);
	if (enetSend(plc5, (char *) &header, sizeof(header)) == ERROR)
	    return (-1);
	if (enetRead(plc5, (char *) &header, sizeof(header), 1000) <= 0)
	    return (-1);
	if (header.flag != FLAG_REPLY)
	    {
	    abortConnection(plc5);
	    return (-1);
	    }
	if (header.command != COMMAND_NEW_SESSION)
	    {
	    abortConnection(plc5);
	    return (-1);
	    }
	plc5 -> sessionID = header.session;
	plc5 -> requestID = 0;
	}

    /*
    Send the message.
    */

    memset((void *) &header, 0, sizeof(header));
    header.flag = FLAG_REQUEST;
    header.command = COMMAND_PCCC_LOCAL_CMD;
    header.length = htons(outBytes);
    header.session = plc5 -> sessionID;
    plc5 -> requestID++;
    *(int *)header.custom = htonl(plc5 -> requestID);
    memcpy(xmitBuff, &header, sizeof(header));
    memcpy(xmitBuff + sizeof(header), outBuff, outBytes);
    if (enetSend(plc5, xmitBuff, sizeof(header) + outBytes) == ERROR)
	return (-1);
/*
    if (enetSend(plc5, (char *) &header, sizeof(header)) == ERROR)
	return (-1);
    if (enetSend(plc5, outBuff, outBytes) == ERROR)
	return (-1);
*/
    /*
    Get the response.
    */

    if ((intSize = enetRead(plc5, (char *) &header, sizeof(header), timeOut)) <= 0)
	return (intSize);
    if (header.flag != FLAG_REPLY)
	{
	abortConnection(plc5);
	return (-1);
	}
    if (header.command != COMMAND_PCCC_LOCAL_CMD)
	{
	abortConnection(plc5);
	return (-1);
	}
    abLength = ntohs(header.length);
    if ((intSize = enetRead(plc5, inBuff, abLength, timeOut)) <= 0)
	return (intSize);
    return (abLength);
}

/****************************************************************/
/*								*/
/*		SEND A BUFFER ON THE ETHERNET			*/
/*								*/
/*	Return ERROR if any problem or if timeout.		*/
/*								*/
/****************************************************************/

static int enetSend(plc, buffer, nbytes)
struct ab_plc5	*plc;
char		*buffer;
int		nbytes;
{
    int		bytesSent;

    bytesSent = send(plc -> sockFd, buffer, nbytes, 0);
    if (bytesSent == ERROR)
	{

	/*
	Nothing went out. See if we can expand the output buffer
	size to contain what we are trying to send.
	*/

	if (errno == EWOULDBLOCK)
	    {
	    plc -> txBuffSize += nbytes;
	    if (setsockopt(plc -> sockFd, SOL_SOCKET, SO_SNDBUF, &(plc -> txBuffSize),
				sizeof(plc -> txBuffSize)) == ERROR)
		{
		abortConnection(plc);
		return (-1);
		}
	    return (enetSend(plc, buffer, nbytes));
	    }
	else
	    {
	    abortConnection(plc);
	    return (-1);
	    }
	}

    else if (bytesSent != nbytes)
	{

	/*
	Some went out, but not all. This is only possible when we are
	using non-blocking I/O. Resend the bytes that didn't go out
	the door.
	*/

	return (enetSend(plc, buffer + bytesSent, nbytes - bytesSent));
	}

    /*
    All is well. Return a value that is clearly not ERROR.
    */

    return (~ERROR);
}

/****************************************************************/
/*								*/
/*		READ A BUFFER FROM THE ETHERNET			*/
/*								*/
/*	Return ERROR if any problem. Return a zero on timeout	*/
/*	Return a 1 if OK.					*/
/*								*/
/****************************************************************/

static int enetRead(plc, buffer, nbytes, maxTime)
struct ab_plc5	*plc;
char		*buffer;
int		maxTime;	/* Milliseconds */
{
    struct timeval	timeOut;
    fd_set		fdMask;
    int			startTime;
    int			bytesGot;
    int			intSize;

    /*
    Set up a mask for the fd.
    */

    FD_ZERO(&fdMask);
    FD_SET(plc -> sockFd, &fdMask);

    /*
    Set up the timeout
    */

    timeOut.tv_sec = maxTime / 1000;
    timeOut.tv_usec = (maxTime % 1000) * 1000;

    /*
    Get all the data that we need.
    */

    while (nbytes > 0)
	{
	startTime = tickGet();
	if (select(plc -> sockFd + 1, &fdMask, NULL, NULL, &timeOut) <= 0)
	    {

	    /*
	    Timeout. Drop this connection so we can start over later.
	    */

	    abortConnection(plc);
	    return (0);
	    }
	
	if (!(plc -> rxBuffTopped))
	    {
	    ioctl(plc -> sockFd, FIONREAD, (int) &intSize);
	    if (intSize == plc -> rxBuffSize)
		{
		plc -> rxBuffSize += 1024;
		if (setsockopt(plc -> sockFd, SOL_SOCKET, SO_RCVBUF, &(plc -> rxBuffSize),
				sizeof(plc -> rxBuffSize)) == ERROR)
		    plc -> rxBuffTopped = 1;
		}
	    }
	bytesGot = recv(plc -> sockFd, buffer, nbytes, 0);
	if (bytesGot <= 0)
	    {

	    /*
	    Error or we were dropped by the other end. It doesn't matter which,
	    we handle them in the same way.
	    */

	    abortConnection(plc);
	    return (ERROR);
	    }

	nbytes -= bytesGot;
	buffer += bytesGot;

	/*
	We spent some time getting this far. Deduct that from the time that we
	were allowed to get the entire message. Assume 16 ms per tick just
	to keep the arithmetic easy. That is close enough.
	*/

	timeOut.tv_usec -= ((tickGet() - startTime) << 4) * 1000;	/* 16 ms per tick */
	while (timeOut.tv_usec < 0)
	    {
	    timeOut.tv_usec += 1000000;
	    timeOut.tv_sec -= 1;
	    }
	if (timeOut.tv_sec < 0)
	    {
	    abortConnection(plc);
	    return (ERROR);
	    }
	}
    return (1);
}

/****************************************************************/
/*								*/
/*		ABORT THE CURRENT CONNECTION			*/
/*								*/
/****************************************************************/

static void abortConnection(plc)
struct ab_plc5	*plc;
{
    if (plc -> sockFd != -1)
	{
	close(plc -> sockFd);
	plc -> sockFd = -1;
	}
    plc -> sessionID = -1;
}
