/*drvEabplc.h,v 1.1.1.1 1999/08/03 14:42:55 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/**********
*
*	File:	drvEabplc.h
*	Auth:	John L. Dalesio
*	Edit:	11/11/92	JLD
*	Edit:	04/14/92	WHS	Update to EPICS
*
*	This file contains the data structures used by the Allen Bradley
*	I/O driver.
*
**********/

#ifndef	INCdrveabplcH
#define	INCdrveabplcH

#define	AB_PORT			5800

#define	TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d

/*	number of lines of abplc */

#define		ABLINES			32

/*	number of allen bradley plcsper line */

#define		NUMABPLCS		20

/*
**	The IOC will assume the data to exchanged with each PLC will
**	come from the a fixed number of file elements It will read
**	these groups to minimize the amount of
**	communications overhead. It will also try to build as large a fetch
**	message as possible & required by remembering which consecutive groups
**	have been requested by the SCAN task.
*/

#define	ABOUTIMG	0		/* File Number for OUTPUT IMAGE	*/
#define	ABINIMG		1		/* File Number for INPUT IMAGE  */
#define	ABSTATUS	2		/* File Number for STATUS       */
#define	ABBITS		3		/* File Number for BITS         */
#define	ABTIMERS	4		/* File Number for TIMERS       */
#define	ABCOUNTERS	5		/* File Number for COUNTERS     */
#define	ABCNTRLS	6		/* File Number for CONTROLS     */
#define	ABINTS		7		/* File Number for INTEGERS     */
#define MAXFILE		1000		/* Out of Range file number     */
#define PSEUDO		1000		/* File Number for Statistics   */

#define ABNOGRPELEMS    4		/* Number of Elements in Each Group */

#define	NUMABOUTS	32		/* 32 groups of 4 OUTPUT REGISTERS */
#define	NUMABINS	32		/* 32 groups of 4 INPUT REGISTERS  */
#define	NUMABSTATS	32		/* 32  groups of 4 STATUS WORDS     */
#define	NUMABBITS	250		/* 250 groups of 4 WORDS(16) OF BITS*/
#define	NUMABTMRS	32		/* 32 groups of 4 TIMERS           */
#define	NUMABCNTRS	32		/* 32 groups of 4 COUNTERS         */
#define	NUMABCNTRLS	32		/* 32 groups of 4 CONTROLLERS      */
#define	NUMABINTS	250		/* 250 groups of 4 INTEGERS         */
#define NUMFASTABINTS	50		/* 50 groups of 4 INTEGERS	   */
#define SLOWPASS	8		/* Number of passes to skip slow points */

/* MAXIMUM ELEMENTS PER MEMORY TYPE */

#define	MAXOUTIMG	NUMABOUTS * ABNOGRPELEMS
#define	MAXINIMG	NUMABINS * ABNOGRPELEMS
#define	MAXSTATUS	NUMABSTATS * ABNOGRPELEMS
#define	MAXBITS		NUMABBITS * ABNOGRPELEMS
#define	MAXTIMERS	NUMABTMRS * ABNOGRPELEMS
#define	MAXCOUNTERS	NUMABCNTRS * ABNOGRPELEMS
#define	MAXCNTRLS	NUMABCNTRLS * ABNOGRPELEMS
#define	MAXINTS		NUMABINTS * ABNOGRPELEMS

/* MAXIMUM ELEMENTS TO ASK FOR IN A COMM REQUEST */

#define	MAXABOUTS	16	/*128 BYTES- 16 groups of 4 OUTPUT REGISTERS */
#define	MAXABINS	16	/*128 BYTES- 16 groups of 4 INPUT REGISTERS  */
#define	MAXABSTATS	8	/* 64 BYTES- 8  groups of 4 STATUS WORDS     */
#define	MAXABBITS	16	/*128 BYTES- 16 groups of 4 WORDS(16) OF BITS*/
#define	MAXABTMRS	8	/*192 BYTES- 8  groups of 4 TIMERS           */
#define	MAXABCNTRS	8	/*192 BYTES- 8  groups of 4 COUNTERS         */
#define	MAXABCNTRLS	8	/*192 BYTES- 8  groups of 4 CONTROLLERS      */
#define	MAXABINTS	25	/*200 BYTES- 50 groups of 4 INTEGERS         */

/**********
*
*	Communications I/O Structures
*
**********/

#define		RXSEM		0
#define		TXSEM		1

/* Consecutive No responses to data to force offline */

#define		MAXDATASTRIKE	3	

#define		MAXRXMSG	300

/* LINEMODE role definitions */

#define		ROLEPRIMARY	0
#define		ROLEBACKUP	1

/* LINEMODE stat definitions */

#define		ACTIVE		TRUE
#define		INACTIVE	FALSE
	
/* Allen Bradley Receive Message tokens */

#define ABNOMSG		0	/* EOT */
#define ABACK		1	/* ACK */
#define ABDATA		2	/* DATA MSG */
#define ABBADDATA	3	/* CRC of Other Data Error */

typedef	struct {
	unsigned short		dualchannel;
	unsigned short		role;
	unsigned short		stat;
		}	LINEMODE;

typedef	struct {
	unsigned short		protocol;
	unsigned short		rxstate;
	unsigned short		rxtocnt;
	unsigned short		rxhwerrcnt;
	unsigned short		token;
	unsigned short		fixedlen;
	unsigned short		rxmsglen;
	unsigned short		rxix;
	unsigned short		line_role;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		rxbuf [MAXRXMSG + 2];
		} 	RXLINE;

#define		MAXTXMSG	300

typedef	struct {
	unsigned short		protocol;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	unsigned short		line_role;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned short		txdisable;
	unsigned char		txbuf [MAXTXMSG + 2];
		} 	TXLINE;

/**********
*
*	Communications I/O Constants
*
**********/

#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		60 	/* 16 ms ticks */
#define		ACTIVITYTIME		300	/* 5 seconds activity time */

#ifdef TEST
#define		RX_ERR_MASK		(TIMEOUT | FRAMERR | PARITYERR | OVERRUN)
#define		TX_ERR_MASK		0	/* No Error Indication */
#endif

#define		VARRESP			FALSE
#define		FIXEDRESP		TRUE

/* ored into line number for xfermsg */

#define		ONERETRY		0x8000

#define		RXIGNORE		0
#define		RXINIT			1
#define		RX2NDCHAR		2
#define		RXGETDATA1		3
#define	     	RXGOTESC		4
#define		RXGETCRC1		5
#define		RXGETCRC2		6

#define		TXGOOD			1
#define		TXBAD			2

#define		TXTIMEOUT		1255	/* 16 ms ticks */

#define		IO_TIMEOUT		3	/* Code returned by timer */

#define		MAXRETRYS		3	/* retransmission trys */

#define		MAXCFGCK		8	/* Bring online each cycle */
#define		BRINGBACKCNT		4	/* Poll loops per cycle */

#define		XFERGOOD		1
#define		XFERBAD			2

#define		RESNAMELEN		16	/* length of a resource name */

/* # of ticks to allow slave to try 	*/

#define		SLWWINDOW		900
/**********
*
*	PLC PROGRAM LOAD STATES
*
**********/

#define		PLC_NOT_LOADED		0
#define		PLC_LOAD_REQUESTED	1
#define		PLC_LOAD_IN_PROGRESS	2
#define		PLC_LOAD_FAILED		3
#define		PLC_LOAD_SUCCESSFUL	4

/**********
*
* Output Que Management Structures for Allen Bradley Protocol
*
**********/

#define         AOUTTYPE        1
#define         BOUTTYPE        2
#define         OUTQSIZE        128

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	short           aovalue;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bomask;
	long            bovalue;
		} BOUT, *BOUTPTR;

typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque [OUTQSIZE];
		} OUTPUTQUE;

/**********
*
*	Internal IOC Structure for ALLEN BRADLEY PLCS
*
**********/

#define TIMERENB	0x8000
#define TIMERTIMING	0x4000
#define TIMERDONE	0x2000

#define COUNTUPENB	0x8000
#define COUNTDOWNENB	0x4000
#define CNTDONE		0x2000
#define CNTOVRFLW	0x1000
#define CNTUNDRFLW	0x0800

typedef struct {
	unsigned short	status;
	short		preset;
	short           accum;
		} ABTIMER;

typedef struct {
	unsigned short	status;
	short		preset;
	short           accum;

		} ABCOUNTER;

typedef struct {
	unsigned short	status;
	unsigned short	length;
	unsigned short  position;
		} ABCONTROLLER;

typedef struct {
	unsigned short	cdata [3];
		} ABCOMPLEX;

typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		oimgdata [ABNOGRPELEMS];
		}	ABOIMGSS;

typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		iimgdata [ABNOGRPELEMS];
		}	ABIIMGSS;

typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		statdata [ABNOGRPELEMS];
		}	ABSTATSS;

typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		bitdata [ABNOGRPELEMS];
		}	ABBITSS;

typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	ABCOMPLEX		timerdata [ABNOGRPELEMS];
		}	ABTIMERSS;

typedef	struct {
	unsigned short	      	was_requested;
	unsigned short		data_valid;
	ABCOMPLEX		cntrdata [ABNOGRPELEMS];
		}	ABCNTRSS;

typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	ABCOMPLEX		cntrldata [ABNOGRPELEMS];
		}	ABCNTRLSS;

typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		intdata [ABNOGRPELEMS];
		}	ABINTSS;

typedef	struct	{
	unsigned short	was_requested;		/* 0  The scan task accessed this RTU */
	unsigned short	is_responding;		/* 1  The RTU is responding */
	unsigned short	holdflag;		/*    The RTU data is in hold */
	unsigned short	prog_load_stat;		/* 2  Program load Status */
	unsigned short	prog_load_page;		/* 3  Program Load Block Number */
	unsigned short  prog_load_number;	/* 4  Program Number to Load */
	unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs */
	unsigned short  no_datain_msgs;		/* 6  Count Data Requests */
	unsigned short  no_rxnoresp;		/* 7  Count No Responses */
	unsigned short  no_txnoresp;		/* 8  Count No Responses */
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors */
	unsigned short  no_retries;		/* 10 Count Retries */
	unsigned short  no_rxhwerrs;		/* 11 Count Comm Hardware Errors */
	unsigned short  no_txhwerrs;		/* 12 Count Comm Hardware Errors */
	unsigned short  no_crcerrs;		/* 13 Count CRC Errors */
	unsigned short  no_offlines;		/* 14 Count Off Line Transitions */
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats*/
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline */
	unsigned short  badoutcnt;		/* 17 Count Failed output commands */
	unsigned short	partner_online;		/* 18 Redundant partner is alive */
	unsigned short	tscanpass;		/* Total number of scanpasses */
	unsigned short	msgs;			/* Count of messags which got go0d responses */
	unsigned long	responsetime;		/* Total amount of time waiting for god responses */
	unsigned short	txtsn;			/* Transmit Sequence Number */
	unsigned short	rxcmd;			/* Last Received Message Type */
	unsigned short	rxtsn;			/* Last Received Message Sequence Number */
	unsigned short	scanpass;               /* Scan Pass Counter */
	unsigned short	extintfile1;		/* Extended Integer file Number 1 */
	unsigned short	extintfile2;		/* Extended Integer file Number 2 */
	unsigned short	extintfile3;		/* Extended Integer file Number 3 */
	unsigned short	extintfile4;		/* Extended Integer file Number 4 */
	unsigned short	nodatastrike;		/* Count consequtive data noresponses */
	unsigned short	devtype;
	unsigned short	devidx;
	ABOIMGSS  	abfouts [NUMABOUTS];
	ABIIMGSS        abfins [NUMABINS];
	ABSTATSS        abfstats [NUMABSTATS];
	ABBITSS         abfbits [NUMABBITS];
	ABTIMERSS       abftimers [NUMABTMRS];
	ABCNTRSS        abfcntrs [NUMABCNTRS];
	ABCNTRLSS       abfcntrls [NUMABCNTRLS];
	ABINTSS		abfints [NUMABINTS];
	ABINTSS		abfexints1 [NUMABINTS];
	ABINTSS		abfexints2 [NUMABINTS];
	ABINTSS		abfexints3 [NUMABINTS];
	ABINTSS		abfexints4 [NUMABINTS];
	} ABPLC;

/**********
*
* AB Protocol function codes
*
**********/

#define	SOH	0X01
#define	STX     0X02
#define	ETX     0X03
#define	EOT	0X04
#define	ENQ     0X05
#define	ACK     0X06
#define	DLE     0X10
#define	NAK	0X15
#define NOREPLY	0xFF
#define MASTERADDR	0

/* Allen Bradley Message function codes */

#define AB5CMD			0x0F
#define AB5CMDRESP		0x4F
#define ABWBITSFNC		0x26
#define ABRWORDS		0x01
#define ABWWORDS		0x00
#define	ABUREAD			0x01
#define	ABUREADRESP		0x41
#define	ABUWRITE		0x08
#define	ABUWRITERESP		0x48
#define	ABDIAGCMD		0x06
#define	ABDSTATFNC		0x03
#define	ABDIAGRESP		0x46

/* Allen Bradley Datamessage offsets */

#define	WORDREADSRCOFFSET	2
#define	WORDREADCMDOFFSET	3
#define	WORDREADSTSOFFSET	4
#define	WORDREADTNSLOFFSET	5
#define	WORDREADTNSHOFFSET	6

#define WORDREADDATAOFFSET	7
#define	DIAGTYPEOFFSET		8
#define	DIAGEXTTYPEOFFSET	9
#define ABSLC501		0xEE
#define DATATABLE		0
#define FULLADDR		0x0F
#define FULLELEMFLAG		0xFF

/* Allen Bradley Receive Datamessage offsets */

#define EXTSTS			6
#define RXDATASTART		6

#define MAXPOLLTRY		6

/**********
*
* Allen Bradley STS Codes
*
**********/

#define ABNOERROR	0x00
#define ABEXTERROR	0xF0
#define	RNODEOFFLINE	0x02


/* task definitions */

#define 	AB_STACK		5000   /* Stack size              */
#define		AB_PRIORITY		200    /* default proirity        */
#define		AB_TASK_OPTIONS		8      /* Floating point	  */

/* Message definitions for redundant updates */

#define	AB_ACK		1
#define AB_REDUPDT	2

typedef struct
    {
    unsigned int	ackSeqNo;
    } AB_ACK_MSG;

typedef struct
    {
    unsigned int	ackSeqNo;
    short		blockNo;
    unsigned short	value;
    } AB_UPDT_MSG;
/*
**	Redundant Communications I/O Structures
**
*/

typedef	struct
{
       	unsigned char		rxbuf;
	unsigned char		rxnew;
} REDRXLINE;


typedef	struct
{
	unsigned short		txmsglen;
	unsigned short		txix;
	unsigned char		txbuf;
} REDTXLINE;


/****************************************************************/
/*								*/
/*	DEFINITIONS FOR ALLEN-BRADLEY INTERCHANGE PROTOCOL	*/
/*								*/
/*	Source: Allen-Bradley confidential information		*/
/*	document titled: INTERCHANGE(tm) Software: Client	*/
/*	Server Protocol, Aug 20, 1993, CID#CS1532.		*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

/****************************************************************/
/*								*/
/*			PCCC PACKET HEADER			*/
/*								*/
/****************************************************************/

typedef struct
    {
    unsigned char		flag;
    unsigned char		command;
    short			length;
    unsigned int		session;
    unsigned int		status;
    unsigned char		custom[16];
    } PCCHEADER;

/****************************************************************/
/*								*/
/*		VALUES FOR THE FLAG FIELD			*/
/*								*/
/****************************************************************/

#define FLAG_REQUEST		1
#define FLAG_REPLY		2


/****************************************************************/
/*								*/
/*		VALUES FOR THE COMMAND FIELD			*/
/*								*/
/****************************************************************/

#define COMMAND_NOP		0
#define COMMAND_NEW_SESSION	1
#define COMMAND_BACKLOG		3
#define COMMAND_UNSOL_DEF	4
#define COMMAND_UNSOL_UNDEF	5
#define COMMAND_DH_PLUS_CMD	6
#define COMMAND_PCCC_LOCAL_CMD	7
#define UNSOL_GETALL		8

/****************************************************************/
/*								*/
/*		Subroutines					*/
/*								*/
/****************************************************************/

int	eabplc_eabplc_aidriver (char *pointname, unsigned short link, unsigned short adapter,
			unsigned short card, unsigned short signal,
			unsigned short *pvalue, unsigned short conversion,
			unsigned char *resname);

int	eabplc_aodriver (char *pointname, unsigned short link, unsigned short adapter,
			unsigned short card, unsigned short signal,
			double dval, unsigned short value,
			unsigned short conversion, unsigned char *resname);

int	eabplc_aoread (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned short *pvalue, unsigned short conversion, 
		unsigned char *resname);

#endif	INCdrveabplcH
/*	End of File:	drvEabplc.h */
