/*drvMultilin.h,v 1.1.1.1 1999/08/03 14:43:25 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
**  File: drvMultilin.h
**  Edit: 03/03/97      WEP: created based on drvModicon.h                                           
**        06/14/97	WEP: removed modbus bi calls       
**        09/04/97	WEP: added support for Resource Group Lock function
**        02/05/98	WEP: added logic to continue to support the 'dumbed down'     
**                         : generic version while adding support targeted to         
**                         : specific models of Multilin devices: 269 Plus, SR469,    
**                         : SR750, SR735, and PQM. each has a new and seperate       
**                         : command for starting the driver, and a unique task name. 
**
*/

#define TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d

/*	number of lines of multilin 	*/
#define		MULLINES	32
/*	number of multilin plcs per line */
#define		NUMMULPLCS	32

/*      number of groups which can be defined in the PCU/LGU */
#define         MAXGROUPNUM     16

#define         RES_LOCKED       0

/*
**	The IOC will assume that the data to be exchanged with each PLC will           	
**	come from a fixed number of coils, digital inputs, analog inputs               	
**	or registers. It will read these in groups to minimize the amount of           	
**	communications overhead. It will also try to build as large a fetch            	
**	message as possible & required by remembering which consecutive groups         	
**	have been requested by the SCAN task.                                          	
*/

#define MCMDS		0			/* Card number for coils 		*/
#define MEINS		2			/* Card number for exception inputs 	*/
#define MAINS		3			/* Card number for analog inputs 	*/
#define MREGS		4			/* Card number for registers 		*/
#define MPSEUDO		33			/* Card number for pseudo i/o 		*/

/* Number of Groups of items */
#define NUMMULCMDS	5        /*     33 */   /*    2 groups of 8 coils		*/
#define NUMMULEIS	1        /*      8 */   /*    1 group  of 8 exceptions	        */
#define NUMMULAIS	4096     /* 16,384 */   /* 4096 groups of 4 analogs		*/
#define NUMMULREGS	4096     /* 16,384 */   /* 4096 groups of 4 registers		*/
#define BITBLOCK        8			/* size of coil/digital groups	        */
#define REGBLOCK	4			/* size of analog/register groups	*/

#define NUM269CMDS	0        /*      0 */   /*    0 groups of 8 coils		*/
#define NUM269EIS	0        /*      0 */   /*    0 groups of 8 exceptions	        */
#define NUM269AIS	64       /*    256 */   /*   64 groups of 4 analogs		*/
#define NUM269REGS	64       /*    256 */   /*   64 groups of 4 registers		*/

#define NUM469CMDS	1        /*      3 */   /*    1 group  of 8 coils		*/
#define NUM469EIS	1        /*      8 */   /*    1 group  of 8 exceptions	        */
#define NUM469AIS	4096     /* 16,384 */   /* 4096 groups of 4 analogs		*/
#define NUM469REGS	4096     /* 16,384 */   /* 4096 groups of 4 registers		*/

#define NUM750CMDS	2        /*     15 */   /*    2 groups of 8 coils		*/
#define NUM750EIS	0        /*      0 */   /*    0 groups of 8 exceptions	        */
#define NUM750AIS	3072     /* 12,288 */   /* 3072 groups of 4 analogs		*/
#define NUM750REGS	3072     /* 12,288 */   /* 3072 groups of 4 registers		*/

#define NUM735CMDS	2        /*     15 */   /*    2 groups of 8 coils		*/
#define NUM735EIS	1        /*      8 */   /*    1 group  of 8 exceptions	        */
#define NUM735AIS	28       /*    110 */   /*   28 groups of 4 analogs		*/
#define NUM735REGS	28       /*    110 */   /*   28 groups of 4 registers		*/

#define NUMPQMCMDS	5        /*     33 */   /*    5 groups of 8 coils		*/
#define NUMPQMEIS	1        /*      8 */   /*    1 group  of 8 exceptions	        */
#define NUMPQMAIS	1224     /*  4,896 */   /* 1224 groups of 4 analogs		*/
#define NUMPQMREGS	1224     /*  4,896 */   /* 1224 groups of 4 registers		*/

/* Number of items */
#define MINMULCMDNO        0                    /* min number of individual coils 	*/
#define MINMULEINO         0                    /* max number of individual exceptions 	*/
#define MINMULAINO         0                    /* max number of individual analogs     */
#define MINMULREGNO        0                    /* max number of individual registers	*/

#define MIN269CMDNO        0                    /* min number of individual coils 	*/
#define MIN269EINO         0                    /* max number of individual exceptions 	*/
#define MIN269AINO         0                    /* max number of individual analogs     */
#define MIN269REGNO        0                    /* max number of individual registers	*/

#define MIN469CMDNO        1                    /* min number of individual coils 	*/
#define MIN469EINO         0                    /* max number of individual exceptions 	*/
#define MIN469AINO         0                    /* max number of individual analogs     */
#define MIN469REGNO        0                    /* max number of individual registers	*/

#define MIN750CMDNO        0                    /* min number of individual coils 	*/
#define MIN750EINO         0                    /* max number of individual exceptions 	*/
#define MIN750AINO         0                    /* max number of individual analogs     */
#define MIN750REGNO        0                    /* max number of individual registers	*/

#define MIN735CMDNO        1                    /* min number of individual coils 	*/
#define MIN735EINO         0                    /* max number of individual exceptions 	*/
#define MIN735AINO         0                    /* max number of individual analogs     */
#define MIN735REGNO        0                    /* max number of individual registers	*/

#define MINPQMCMDNO        1                    /* min number of individual coils 	*/
#define MINPQMEINO         0                    /* max number of individual exceptions 	*/
#define MINPQMAINO         0                    /* max number of individual analogs     */
#define MINPQMREGNO        0                    /* max number of individual registers	*/

/* Number of items */
#define MAXMULCMDNO        14                          /* max number of individual coils 	*/
#define MAXMULEINO         7                           /* max number of individual exceptions 	*/
#define MAXMULAINO         16383                       /* max number of analogs 		*/
#define MAXMULREGNO        16383                       /* max number of registers 		*/

#define MAX269CMDNO        0                           /* max number of individual coils 	*/
#define MAX269EINO         0                           /* max number of individual exceptions 	*/
#define MAX269AINO         255                         /* max number of analogs 		*/
#define MAX269REGNO        255                         /* max number of registers 		*/

#define MAX469CMDNO        3                           /* max number of individual coils 	*/
#define MAX469EINO         7                           /* max number of individual exceptions 	*/
#define MAX469AINO         16363                       /* max number of analogs 		*/
#define MAX469REGNO        16363                       /* max number of registers 		*/

#define MAX750CMDNO        14                          /* max number of individual coils 	*/
#define MAX750EINO         0                           /* max number of individual exceptions 	*/
#define MAX750AINO         12287                       /* max number of analogs 		*/
#define MAX750REGNO        12287                       /* max number of registers 		*/

#define MAX735CMDNO        14                          /* max number of individual coils 	*/
#define MAX735EINO         7                           /* max number of individual exceptions 	*/
#define MAX735AINO         110                         /* max number of analogs 		*/
#define MAX735REGNO        110                         /* max number of registers 		*/

#define MAXPQMCMDNO        33                          /* max number of individual coils 	*/
#define MAXPQMEINO         7                           /* max number of individual exceptions 	*/
#define MAXPQMAINO         4895                        /* max number of analogs 		*/
#define MAXPQMREGNO        4895                        /* max number of registers 		*/

#define MAXREGBLKSIZE	16			/* Maximum number of digital groups 	*/
						/*  to request in one packet        	*/

/*
**	Conversion types for analog inputs	*/
**
*/
#define		AICNVINTEGER	10
#define		AICNVDOUBLEINT	11
#define		AICNVFLOAT	12
#define		AICNVDOUBLEFLT	13
#define		AICNVASCIIFLT	14
#define		AICNVSTATUS	15
#define		AICNVTIME	16
#define		AICNVASCII_2	17
#define		AICNVASCII_4	18
#define		AICNVASCII_12	19
#define		AICNVASCII_20	20
#define		MAXCONV		20

/* The following must match the definition in recChoice.ascii */
#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2


#define    WATCHDOGTIMEOUT       10*60 /* watchdog monitor timeout.  */


/*
**	Communications I/O Structures
**
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	256
#define		MAXTXMSG	256


typedef	struct {
	unsigned short		rxstate;
	unsigned short		fixedlen;
	unsigned short		rxmsglenexpected;
        unsigned short          rxmsglen;
       	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
        unsigned char           txaddress;
        unsigned char           txfunction;
       	unsigned char		rxbuf[MAXRXMSG];
		} 	RXLINE;


typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;
/*
**	Communications I/O Constants
**
*/

#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		60	/* 16 ms ticks 				*/
#define		ACTIVITYTIME		300	/* must see activity in 5 seconds 	*/

#define		VARRESP			FALSE
#define		FIXEDRESP		TRUE
#define		ONERETRY		0x8000	/* OR'ed into line number for xfermsg 	*/
#define		RXIGNORE		0
#define		RXADRST			1
#define		RXFNCST			2
#define		RXLENST			3
#define		RXDATAST		4

#define		TXGOOD			1
#define		TXBAD			2
#define		IO_TIMEOUT		3	/* Code returned by timer 		*/


#define		TXTIMEOUT		60	/* 16 ms ticks: one second              */

#define		MAXRETRYS		3	/* retransmission trys 			*/

#define		MAXCFGCK		8	/* Try to Bring online each cycle 	*/
#define		BRINGBACKCNT		4	/* Poll loops per cycle 		*/

#define		XFERGOOD		1
#define		XFERBAD			2

#define		RESNAMELEN		16	/* length of a resource name 		*/
#define		SLWWINDOW		900	/* # of ticks to allow slave to try 	*/

/*
** Output Queue Management Structures for Multilin/Modbus Protocol
**
*/

#define         AOUTTYPE        1
#define         BOUTTYPE        2

#define         OUTQSIZE        128
#define		STRINGMSGLEN	80

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	short           aovalue;
	double		dblaovalue;
	unsigned short	conversion;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bomask;
	long            bovalue;
		} BOUT, *BOUTPTR;


typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque[OUTQSIZE];
		} OUTPUTQUE;
/*
**	Internal IOC Structure for Multilin PLC
**
*/
typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned long		last_read;
		}	MULCMDS;

typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned long		last_read;
		}	MULEXCEPT;

typedef struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		ioregs[REGBLOCK];
	unsigned short		data_togo[REGBLOCK];
	unsigned short		regs_togo[REGBLOCK];

		}	REGISTERS;

typedef struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	short			aivals[REGBLOCK];
		}	MULAIS;



typedef	struct	{
	unsigned short  was_requested;		/* 0  The scan task has accesses this RTU 	*/
	unsigned short  is_responding;		/* 1  The RTU is responding 			*/
	unsigned short  holdflag;		/*    The RTU data is in hold 			*/
	unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs 			*/
	unsigned short  no_datain_msgs;		/* 6  Count Data Requests 			*/
	unsigned short  no_rxnoresp;		/* 7  Count No Responses 			*/
	unsigned short  no_txnoresp;		/* 8  Count No Responses 			*/
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors 		*/
	unsigned short  no_retries;		/* 10 Count Retries 				*/
	unsigned short  no_rxhwerrs;		/* 11 Count Comm Hardware Errors 		*/
	unsigned short  no_txhwerrs;		/* 12 Count Comm Hardware Errors 		*/
	unsigned short  no_crcerrs;		/* 13 Count CRC Errors 				*/
	unsigned short  no_offlines;		/* 14 Count Off Line Transitions		*/
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats			*/
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline 			*/
	unsigned short  badoutcnt;		/* 17 Count Failed output commands 		*/

        unsigned short  tscanpass;		/* Total number of scanpasses. incremented at   */
                                                /*  the start of each plc scan.                 */
        unsigned short  msgs;			/* Count of messages which got good responses.  */
                                                /*  incremented on each exit from 'good' msgxfer*/
        unsigned long   responsetime;		/* Total time(ticks) waiting for good responses.*/
                                                /* accumu (exit-entry) time for 'good' msgxfers */

	MULCMDS	        coils[NUMMULCMDS];
        MULEXCEPT       exception;
	MULAIS          ais[NUMMULAIS];
	REGISTERS       regs[NUMMULREGS];
	}	MULPLC;
/*
** MODBUS Protocol function codes
*/

/*#define MODCOILREAD		0x01  ** READ COIL STATUS: (ON/OFF) read a group of logic coils.     */
/*#define MODDIREAD		0x02  ** READ INPUT STATUS:(ON/OFF) of a group of discrete inputs.   */
#define MODREGREAD		0x03  /* READ HOLDING REGISTERS:    get current binary value in one  */
                                      /*                            or more holding registers.       */
#define MODAIREAD		0x04  /* READ INPUT REGISTERS:      get current binary value in one  */
                                      /*                            or more input registers.         */

#define MODCOILWRITE		0x05  /* FORCE SINGLE COIL:         force logic coil to ON/OFF state.*/
#define MODREGWRITE		0x06  /* PRESET SINGLE REGISTER:    place a specific binary value    */
                                      /*                            into a holding register.         */

#define MODEXCEPTREAD           0x07  /* READ EXCEPTION STATUS:     get the 8 predefined exception   */
                                      /*                            status bits.             .       */
#define MODLOOPBACK		0x08  /* LOOPBACK DIAGNOSTIC TEST                                    */

/*#define MODMCOILWRITE		0x0F  ** FORCE MULTIPLE COILS:      force series of consecutive      */
                                      /*                            logic coils to ON/OFF states.    */
#define MODMREGWRITE		0x10  /* PRESET MULTIPLE REGISTERS: put specific binary values into  */
                                      /*                            consecutive series of holding    */
                                      /*                            registers.                       */

#define FNCERRBIT		0x80  /* Modbus error function code */


/* Various message lengths used in message transmission and response validation */

#define MODWRITEHDRLEN		0x06	/* wep: not currently used */
#define MODREADMSGLEN		0x06
#define MODSETCOILLEN		0x06
#define MODSETREGLEN		0x06
#define MODSET1REGLEN		0x09
#define MODSET2REGLEN		0x0b
#define MODSET3REGLEN		0x0d
#define MODSET4REGLEN		0x0f

#define MODEXCEPTLEN            0x02	/* wep: specially added for Multilin */
#define MODEXCEPTRESPLEN        0x05    /* Including CRC */

#define MODMSETCOILLEN		0x07	/* wep: not currently used */
#define MODMSETCOILRESPLEN	0x06
#define MODCRCLEN		0x02
#define MODLOOPBACKLEN		0x06
#define MODLOOPBACKRESPLEN	0x08	/* Including CRC */
#define MODWRITERESPLEN		0x08	/* Including CRC */
#define MODSETMREGRESPLEN	0x06
#define NAKMSGLEN		0x03	/* bytes left get */

/*
** MODBUS Exception Codes
*/

#define MODEXCEPTFNC		0x01	/* wep: not used */
#define MODEXCEPTADR		0x02	/* "    "   "    */
#define MODEXCEPTDATA		0x03	/* "    "   "    */
#define MODEXCEPSLAVE		0x04	/* "    "   "    */
#define MODEXCEPTSACK		0x05	/* "    "   "    */
#define MODEXCEPTBUSY		0x06	/* "    "   "    */

/* Command Results */

#define CMDDONE                 0x00     /* the command was sent and the response was error free.   */
#define CMDFAIL                 0x01     /* there were errors in sending or receiving the command.  */
#define CMDNACK                 0x02     /* response was error free, but command was not completed. */
#define CMDREJECT               0x03     /* driver rejected command due to unsupported data type.   */
#define CMDBUSY			0x05	 /* The command was sent, but the device is busy */

/* task definitions */

#define MUL_STACK		8000   /* Stack size              */
#define	MUL_PRIORITY		152    /* default priority        */
#define	MUL_TASK_OPTIONS	8      /* Floating point          */



/*
** Mutual-Exclusion Control for diagnostic/data radio channel sharing */
/*
*/

#define		MUXTIMEOUT	60*30	/* 16.67 ms ticks (60Hz): 30 seconds 	*/


typedef struct {
        unsigned long   TXintcount;
        unsigned long   RXintcount;
        unsigned long   RXbadcount;
        unsigned long   msgxfercount;
        unsigned long   startingtick;
               }        LINESTATS;

/* Trace Bit Definitions */

#define ALLTRAFFICTRACE      0x0001
#define MODBUSNACKTRACE      0x0002        /* MODBUS NACK errors.                               */
#define CONFIGURATIONTRACE   0x0010        /* trace configuration errors in driver calls.       */
#define RUNTIMEERRORTRACE    0x0020        /* trace run time errors in driver calls.            */
#define CALLREJECTTRACE      0x0040        /* trace non-error based driver call rejections.     */


/*	End of File:	drvMultilin.h */
