/****************************************************************/
/*                                                              */
/*    This software is the proprietary property of Tate         */
/*    Integrated Systems, L.P.  Patents and copyrights are      */
/*    pending.                                                  */
/*                                                              */
/*    This software is furnished under a license and may        */
/*    be used and copied only in accordance with the terms      */
/*    of such license and with the inclusion of the above       */
/*    notice. This software or any other copies thereof         */
/*    may not be provided or otherwise made available to        */
/*    any other person. No title to and ownership of the        */
/*    software is hereby transferred.                           */
/*                                                              */
/*    The information in this software is subject to            */
/*    change without notice and should not be construed         */
/*    as a commitment by Tate Integrated Systems L.P. or        */
/*    its third party suppliers.                                */
/*                                                              */
/*    Tate Integrated Systems, L.P. (TIS) and its third         */
/*    party suppliers assume no responsibility for the          */
/*    use or inability to use any of its software. TIS          */
/*    software is provided "as is" without warranty of          */
/*    any kind and TIS expressly disclaims all implied          */
/*    warranties, including but not limited to the implied      */
/*    warranties of merchantability and fitness for a           */
/*    particular purpose.                                       */
/*                                                              */
/*    Notice: Notwithstanding any other lease or license        */
/*    that may pertain to or accompany the delivery of this     */
/*    computer software, the rights of the Government           */
/*    regarding its use, reproduction, and disclosure are       */
/*    as set forth in Section 52.227-19 of the FARS             */
/*    Computer Software-Restricted Rights clause.               */
/*                                                              */
/****************************************************************/
/*                                                                           
**  File: drvSec.c                                                           
**  Edit:  1/07/99  wep : created as sneaky way to test Sec driver in t51    
**         2/17/99  wep : added support for reading Status Group 3 (Output)  
**                      : parameters using the ST3 command.                  
**         2/22/99  wep : merged in new static storage for mode strings      
**         3/12/99  wep : in private mode, register resource via rmPrivate() 
**         4/28/99  wep : fixed faulty private mode logic.                   
**         4/29/99  wep : modified sectableinfo () diagnostic to display the 
**                      : upstable [value].support variable with the data.   
**                      : modfied parsing error logic to ignore single       
**                      : instances of parsing errors, to have repeated      
**                      : parsing errors eventually mark the data as invalid 
**                      : and put the plc offline. fixed an annoying but     
**                      : harmless detailed trace message errors. modified   
**                      : retry logic to properly go offline after three     
**                      : communication errors.                              
*/

#define  VERSIONNUM     13     /* NOTE: change here */

#include <vxWorks.h>
#include <sysLib.h>
#include <semLib.h>
#include <wdLib.h>
#include <rngLib.h>
#include <logLib.h>
#include <tickLib.h>
#include <remLib.h>
#include <stdioLib.h>
#include <string.h>

#include "resmgr.h"
#include "task_params.h"
#include "dbDefs.h"
#include "drvSup.h"
#include "restable.h"

#include "iocsiodvr_bridge.h"
#include <serialFile.h>
#include "resgrouptask.h"

#include "drvSec.h"                          /* NOTE: change here */
#include "drvSec_proto.h"                    /* NOTE: change here */

/*
#include <resGroupTask.h>
*/

/***********************************/
/***********************************/
/******** Begin Prototypes *********/
/***********************************/
/***********************************/

/**********
*
*    This routine is required by iocinit in the EPICS system.
*
**********/
static  long init ();

        void sechelp (void);

        void sectraceon (unsigned short line);

        void sectraceoff (unsigned short line);

        void sectraffic (unsigned short line, unsigned char add, unsigned type);

        void sectrafficon (unsigned short line);

        void sectrafficoff (unsigned short line);

        void secasciion (unsigned short line);

        void secasciioff (unsigned short line);

        void secrestrlv (unsigned short line, unsigned short level);

        void secrestraceon (unsigned short line);

        void secrestraceoff (unsigned short line);

        void sectrlv (unsigned short line, unsigned short level);

        void seccontraceon (unsigned short line);

        void seccontraceoff (unsigned short line);

        void secruntraceon (unsigned short line);

        void secruntraceoff (unsigned short line);

        void secrejecttraceon (unsigned short line);

        void secrejecttraceoff (unsigned short line);

        void secnacktraceon (unsigned short line);

        void secnacktraceoff (unsigned short line);

        void npallstatuson (unsigned short line);

        void npallstatusoff (unsigned short line);

        void npbadstatuson (unsigned short line);

        void npbadstatusoff (unsigned short line);

        void secdefaulttrace (unsigned short line);

        int startsec (unsigned short line,
                         char *mode,
                         unsigned short delay,
                         unsigned short x,
                         unsigned short delay1,
                         unsigned short delay2,
                         unsigned short private);

        void seccomm (unsigned short line,
                      unsigned short delay,
                      unsigned short x,
                      unsigned short delay1,
                      unsigned short delay2);

        void secinfo (unsigned short line);

        void secdriver (unsigned short line,
                   unsigned char *mode,
                   unsigned short delay,
                   unsigned short hdflag,
                   unsigned short delay1,
                   unsigned short delay2,
                   unsigned short private);

static  int checkgroup (unsigned short line, unsigned short *lstate, unsigned short tracenum);

static  void on_line (unsigned short line);

static void mark_not_valid (SEC  *pmod);

static  void scan_module (SEC *pmod, unsigned short line);

        int secaodriver (char *pointname, unsigned short link, unsigned short adapter,
                      unsigned short card, unsigned short signal, double dvalue,
                      unsigned short value, unsigned short conversion, unsigned char *resname);

        int secaoread (char *pointname, unsigned short link, unsigned short adapter,
                    unsigned short card, unsigned short signal, unsigned short *pvalue,
                    unsigned short conversion, unsigned char *resname);

static  void  register_resource (unsigned short link, SEC *pmod,
                                 unsigned char *resname, char *pointname);

static  int allocate_module (unsigned short line);

static  void start_rx (unsigned short line, SEMAPHORE *semid, unsigned short *pstatus);

static  void stop_rx (unsigned short line);

static  void rx_int_proc (unsigned short line);

static  int check_line_activity (unsigned short line, unsigned short predelay);

static  void start_tx (unsigned short line, SEMAPHORE *semid, unsigned short *pstatus);

static  void stop_tx (unsigned short line);

static  void tx_int_proc (unsigned short line);

static  unsigned short  msgxfer (unsigned short fline, SEC *pmod, unsigned short rxlen);

static  void init_module_tab (unsigned short line);

static  void initmodule (SEC  *pmod);

static  void initstats (SEC *pmod);

static  unsigned short readparameters1 (unsigned short line, SEC *pmod);

static  unsigned short readparameters2 (unsigned short line, SEC *pmod);

static  unsigned short readstatusgroup1 (unsigned short line, SEC *pmod);

static  unsigned short readstatusgroup2 (unsigned short line, SEC *pmod);

static  unsigned short readstatusgroup3 (unsigned short line, SEC *pmod);

static  unsigned short readstatusgroup5 (unsigned short line, SEC *pmod);

static  void prepresponses (unsigned short line);

static  int any_responding (unsigned short line);

        void secdumpmsgcnts (unsigned short line);

        void secclearmsgcnts (unsigned short line);

        void secdumpstats (unsigned short line);

        void secclearstats (unsigned short line);

        void sectableinfo (unsigned short line);

        void secmodinfo (unsigned short line);

static  void printmodhdr (unsigned char *buffer, unsigned short cnt);

        void seclineinfo (void);

/*********************************/
/*********************************/
/******** End Prototypes *********/
/*********************************/
/*********************************/

struct {
            long        number;
            DRVSUPFUN   report;
            DRVSUPFUN   init;
       }
        drvSec={ 2, NULL, init};

/*
**      External Line Level Data Structures (declared in drvcommon.c)
**
*/
extern  unsigned char   resnames [SECLINES] [RESNAMELEN]; /* array of resource name strings.       */

extern  RES_TABLE      *prestable [SECLINES];        /* array of pointers to resource tables         */

extern  unsigned short  resresponse [SECLINES];       /* array of resource response flags.            */
                                                      /* False = not active resource(slave),          */
                                                      /* True = active resource(master).              */

extern  unsigned short  line_activity [SECLINES];     /* true/false line active flag. calculated      */
                                                      /* using activity_time below.                   */
extern  unsigned short  lastactivity [SECLINES];      /* assigned value of line_activity prior to     */
                                                      /* it's recalculation.                          */

extern  unsigned short  modtrace [SECLINES];          /* controls output of communication error log   */
                                                      /* messages (true/false)                        */
extern  unsigned short  mltrace [SECLINES];           /* controls output of certain resource related  */
                                                      /* log messages (level)                         */
extern  unsigned short  modrestrace [SECLINES];       /* controls output of MASTER/SLAVE arbitration  */
                                                      /* log messages (binary)                        */

extern  SEM_ID          semids [SECLINES] [2];        /* 'new' semaphore logic. the old SEMAPHORE is  */
                                                      /* still used.                                  */

extern  unsigned short  intermsgdelay [SECLINES];     /* holds value passed in task spawn. used for   */
                                                      /* each msgxfer attempt                         */
extern	unsigned short	pretxdelay [SECLINES];        /* holds post RTS delay before TX.              */
extern	unsigned short	posttxdelay [SECLINES];       /* holds post TX delay before dropping RTS.     */

extern  long            activity_time [SECLINES];     /* holds timetick of last activity. used to     */
                                                      /* set line_activity flag                       */

extern  unsigned short  halfduplex [SECLINES];        /* store half/full duplex line status           */

extern  unsigned short  pstatewait [SECLINES];        /* holds max time in seconds that driver will   */
                                                      /* wait to obtain partner state on first        */
                                                      /* attempting to decide if it should be master. */


extern short resGroupLock[]; /* holds line driver group lock flags. */
extern char *resGroupTripName[]; /* line driver trip name. */

/*
**      Storage for Sec Comm Line Structures
**
*/

static RXLINE          *rxlinetab [SECLINES] =
			{NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL,} ; /* array of pointers to dynamically           */
                                        /* allocated comm controls/buffers            */

static TXLINE          *txlinetab [SECLINES] =
			{NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL,} ; /* array of pointers to dynamically           */
                                        /* allocated comm controls/buffers            */


/*
**      Storage for Sec Structures
**
*/
static unsigned short   parseinvalidcount [SECLINES];

static unsigned short   parseofflinecount [SECLINES];

static unsigned short   parseofflinecounter [SECLINES];

static unsigned short   pryvate [SECLINES];     /* 0 = redundant mode,   !0=private mode        */

static unsigned int     tick [SECLINES][5];

static unsigned int     lastscan [SECLINES];    /* holds timestamp of start of secdriver loop.  */

static unsigned int     scantime [SECLINES];    /* holds actual time taken to scan modules.     */

static unsigned int     loopoutmsgs [SECLINES];

static unsigned int     loopmsgs [SECLINES];

static LINESTATS        linestats [SECLINES];   /* array of line level statistics.              */


static SEC             *modules [SECLINES]; 	/* one dimensional array of pointers to         */
                                                /* fixed length dynamically allocated Sec       */
static unsigned short   link [SECLINES];        /* line Number started with for file I/O        */
                                                /* control and data structures.                 */

static  unsigned short  trace [SECLINES];       /* controls output of system error messages:    */
                                                /* bit mapped controls                          */
static  unsigned char   traceadr [SECLINES];
static  unsigned char   tracetype [SECLINES];

static           char   user [40], password [40];/* holds id and password of the user that      */
                                                 /* spawned the driver task required by host to */
                                                 /* check file read privileges on load files    */

#define ATR  1
#define MAN  2
#define MOD  3
#define NOM  4
#define NUL  5
#define PSD  6
#define RWD  7
#define SDA  8
#define ST1  9
#define ST2 10
#define ST3 11
#define ST4 12
#define ST5 13
#define STD 14
#define STR 15
#define TST 16
#define UBR 17
#define UID 18
#define VER 19

/* this table defines the comands implemented by this driver. */
#define COMMANDCOUNT 20
static short validcommands [COMMANDCOUNT] =
{
/*      0 */  FALSE,
/* ATR  1 */  FALSE,
/* MAN  2 */  FALSE,
/* MOD  3 */  FALSE,
/* NOM  4 */  FALSE,
/* NUL  5 */  FALSE,
/* PSD  6 */  FALSE,
/* RWD  7 */  FALSE,
/* SDA  8 */  FALSE,
/* ST1  9 */  TRUE,
/* ST2 10 */  TRUE,

/* ST3 11 */ /* FALSE, */
/* ST3 11 */  TRUE,

/* ST4 12 */  FALSE,
/* ST5 13 */  TRUE,
/* STD 14 */  FALSE,
/* STR 15 */  FALSE,
/* TST 16 */  FALSE,
/* UBR 17 */  FALSE,
/* UID 18 */  FALSE,
/* VER 19 */  FALSE
};

static struct UPSENTRY
{
    char id;
    char support;
    char *name;
    char command;
    char offset;
    char aiflag;
    char biflag;
    char siflag;
}               upstable [] = {
/* Num, Support,"parameter name",    command type, group,                AI,    BI,    SI */
   { 0, FALSE, "Illegal Command Zero",        NUL, ILLEGALCOMMAND,    FALSE, FALSE, FALSE},
   { 1, FALSE, "Alarm Awaiting Power",        ST5, ALARMAWAITINGPOWER, TRUE,  TRUE, FALSE},
   { 2, FALSE, "Alarm Bypass Bad",            ST5, ALARMBYPASSBAD,     TRUE,  TRUE, FALSE},
   { 3, FALSE, "Alarm Charger Failure",       ST5, ALARMCHARGERFAILURE,TRUE,  TRUE, FALSE},
   { 4, FALSE, "Alarm Fan Failure",           ST5, ALARMFANFAILURE,    TRUE,  TRUE, FALSE},
   { 5, FALSE, "Alarm Fuse Failure",          ST5, ALARMFUSEFAILURE,   TRUE,  TRUE, FALSE},
   { 6, FALSE, "Alarm General Fault",         ST5, ALARMGENERALFAULT,  TRUE,  TRUE, FALSE},
   { 7, FALSE, "Alarm Input Bad",             ST5, ALARMINPUTBAD,      TRUE,  TRUE, FALSE},
   { 8, FALSE, "Alarm Output Bad",            ST5, ALARMOUTPUTBAD,     TRUE,  TRUE, FALSE},
   { 9, FALSE, "Alarm Output Off",            ST5, ALARMOUTPUTOFF,     TRUE,  TRUE, FALSE},
   {10, FALSE, "Alarm Overload",              ST5, ALARMOVERLOAD,      TRUE,  TRUE, FALSE},
   {11, FALSE, "Alarm Shutdown Imminent",     ST5, ALARMSHUTDNIMMINENT,TRUE,  TRUE, FALSE},
   {12, FALSE, "Alarm Shutdown Pending",      ST5, ALARMSHUTDNPENDING, TRUE,  TRUE, FALSE},
   {13, FALSE, "Alarm System Off",            ST5, ALARMSYSTEMOFF,     TRUE,  TRUE, FALSE},
   {14, FALSE, "Alarm Temperature",           ST5, ALARMTEMPERATURE,   TRUE,  TRUE, FALSE},
   {15, FALSE, "Alarm UPS Shutdown",          ST5, ALARMUPSSHUTDOWN,   TRUE,  TRUE, FALSE},
   {16, FALSE, "Audible Alarm",               NOM, AUDIBLEALARM,       TRUE,  TRUE, FALSE},
   {17, FALSE, "Auto Restart",                ATR, AUTORESTART,        TRUE, FALSE, FALSE},
   {18, FALSE, "Battery Charge",              ST1, BATTERYCHARGE,      TRUE,  TRUE, FALSE},
   {19, FALSE, "Battery Condition",           ST1, BATTERYCONDITION,   TRUE,  TRUE, FALSE},
   {20, FALSE, "Battery Current",             ST1, BATTERYCURRENT,     TRUE,  TRUE, FALSE},
   {21, FALSE, "Battery Installed Date",      NOM, BATTERYINSTALLDATE,FALSE, FALSE,  TRUE},
   {22, FALSE, "Battery Status",              ST1, BATTERYSTATUS,      TRUE,  TRUE, FALSE},
   {23, FALSE, "Battery Temperature",         ST1, BATTERYTEMPERATURE, TRUE, FALSE, FALSE},
   {24, FALSE, "Battery Voltage",             ST1, BATTERYVOLTAGE,     TRUE, FALSE, FALSE},
   {25, FALSE, "Bypass Current 1",            ST4, BYPASSCURRENT1,     TRUE, FALSE, FALSE},
   {26, FALSE, "Bypass Current 2",            ST4, BYPASSCURRENT2,     TRUE, FALSE, FALSE},
   {27, FALSE, "Bypass Current 3",            ST4, BYPASSCURRENT3,     TRUE, FALSE, FALSE},
   {28, FALSE, "Bypass Frequency",            ST4, BYPASSFREQUENCY,    TRUE, FALSE, FALSE},
   {29, FALSE, "Bypass Num Lines",            ST4, BYPASSNUMLINES,     TRUE, FALSE, FALSE},
   {30, FALSE, "Bypass Power 1",              ST4, BYPASSPOWER1,       TRUE, FALSE, FALSE},
   {31, FALSE, "Bypass Power 2",              ST4, BYPASSPOWER2,       TRUE, FALSE, FALSE},
   {32, FALSE, "Bypass Power 3",              ST4, BYPASSPOWER3,       TRUE, FALSE, FALSE},
   {33, FALSE, "Bypass Voltage 1",            ST4, BYPASSVOLTAGE1,     TRUE, FALSE, FALSE},
   {34, FALSE, "Bypass Voltage 2",            ST4, BYPASSVOLTAGE2,     TRUE, FALSE, FALSE},
   {35, FALSE, "Bypass Voltage 3",            ST4, BYPASSVOLTAGE3,     TRUE, FALSE, FALSE},
   {36, FALSE, "Estimated Charge",            ST1, ESTIMATEDCHARGE,    TRUE, FALSE, FALSE},
   {37, FALSE, "Estimated Minutes",           ST1, ESTIMATEDMINUTES,   TRUE, FALSE, FALSE},
   {38, FALSE, "High Voltage Transfer Point", NOM, HIGHVOLTAGETRANSFER,TRUE, FALSE, FALSE},
   {39, FALSE, "Identification",              UID, IDENTIFICATION,    FALSE, FALSE,  TRUE},
   {40, FALSE, "Input Current 1",             ST2, INPUTCURRENT1,      TRUE,  TRUE, FALSE},
   {41, FALSE, "Input Current 2",             ST2, INPUTCURRENT2,      TRUE,  TRUE, FALSE},
   {42, FALSE, "Input Current 3",             ST2, INPUTCURRENT3,      TRUE,  TRUE, FALSE},
   {43, FALSE, "Input Frequency 1",           ST2, INPUTFREQUENCY1,    TRUE,  TRUE, FALSE},
   {44, FALSE, "Input Frequency 2",           ST2, INPUTFREQUENCY2,    TRUE,  TRUE, FALSE},
   {45, FALSE, "Input Frequency 3",           ST2, INPUTFREQUENCY3,    TRUE,  TRUE, FALSE},
   {46, FALSE, "Input Line Bads",             ST2, INPUTLINEBADS,      TRUE,  TRUE, FALSE},
   {47, FALSE, "Input Num Lines",             ST2, INPUTNUMLINES,      TRUE,  TRUE, FALSE},
   {48, FALSE, "Input Power 1",               ST2, INPUTPOWER1,        TRUE,  TRUE, FALSE},
   {49, FALSE, "Input Power 2",               ST2, INPUTPOWER2,        TRUE,  TRUE, FALSE},
   {50, FALSE, "Input Power 3",               ST2, INPUTPOWER3,        TRUE,  TRUE, FALSE},
   {51, FALSE, "Input Voltage 1",             ST2, INPUTVOLTAGE1,      TRUE,  TRUE, FALSE},
   {52, FALSE, "Input Voltage 2",             ST2, INPUTVOLTAGE2,      TRUE,  TRUE, FALSE},
   {53, FALSE, "Input Voltage 3",             ST2, INPUTVOLTAGE3,      TRUE,  TRUE, FALSE},
   {54, FALSE, "Low Voltage Transfer Point",  NOM, LOWVOLTAGETRANSFER, TRUE, FALSE, FALSE},
   {55, FALSE, "Manufacturer",                MAN, MANUFACTURER,      FALSE, FALSE,  TRUE},
   {56, FALSE, "Model",                       MOD, MODEL,             FALSE, FALSE,  TRUE},
   {57, FALSE, "Nominal Battery Life",        NOM, NOMBATTERYLIFE,    TRUE, FALSE, FALSE},
   {58, FALSE, "Nominal Input Frequency",     NOM, MONINPUTFREQUENCY,  TRUE, FALSE, FALSE},
   {59, FALSE, "Nominal Input Voltage",       NOM, NOMINPUTVOLTAGE,    TRUE, FALSE, FALSE},
   {60, FALSE, "Nominal Low Battery Time",    NOM, NOMLOWBATTERYTIME,  TRUE, FALSE, FALSE},
   {61, FALSE, "Nominal Output Frequency",    NOM, NOMOUTPUTFREQUENCY, TRUE, FALSE, FALSE},
   {62, FALSE, "Nominal Output Power",        NOM, NOMOUTPUTPOWER,     TRUE, FALSE, FALSE},
   {63, FALSE, "Nominal Output Voltage",      NOM, NOMOUTPUTVOLTAGE,   TRUE, FALSE, FALSE},
   {64, FALSE, "Nominal Volt-Amp Rating",     NOM, NOMVOLTAMPRATING,   TRUE, FALSE, FALSE},
   {65, FALSE, "Output Current 1",            ST3, OUTPUTCURRENT1,     TRUE, FALSE, FALSE},
   {66, FALSE, "Output Current 2",            ST3, OUTPUTCURRENT2,     TRUE, FALSE, FALSE},
   {67, FALSE, "Output Current 3",            ST3, OUTPUTCURRENT3,     TRUE, FALSE, FALSE},
   {68, FALSE, "Output Frequency",            ST3, OUTPUTFREQENCY,     TRUE, FALSE, FALSE},
   {69, FALSE, "Output Load 1",               ST3, OUTPUTLOAD1,        TRUE, FALSE, FALSE},
   {70, FALSE, "Output Load 2",               ST3, OUTPUTLOAD2,        TRUE, FALSE, FALSE},
   {71, FALSE, "Output Load 3",               ST3, OUTPUTLOAD3,        TRUE, FALSE, FALSE},
   {72, FALSE, "Output Num Lines",            ST3, OUTPUTNUMLINES,     TRUE, FALSE, FALSE},
   {73, FALSE, "Output Power 1",              ST3, OUTPUTPOWER1,       TRUE, FALSE, FALSE},
   {74, FALSE, "Output Power 2",              ST3, OUTPUTPOWER2,       TRUE, FALSE, FALSE},
   {75, FALSE, "Output Power 3",              ST3, OUTPUTPOWER3,       TRUE, FALSE, FALSE},
   {76, FALSE, "Output Source",               ST3, OUTPUTSOURCE,       TRUE,  TRUE, FALSE},
   {77, FALSE, "Output Voltage 1",            ST3, OUTPUTVOLTAGE1,     TRUE, FALSE, FALSE},
   {78, FALSE, "Output Voltage 2",            ST3, OUTPUTVOLTAGE2,     TRUE, FALSE, FALSE},
   {79, FALSE, "Output Voltage 3",            ST3, OUTPUTVOLTAGE3,     TRUE, FALSE, FALSE},
   {80, FALSE, "Reboot With Duration",        RWD, REBOOTWITHDURATION,FALSE, FALSE, FALSE},
   {81, FALSE, "Seconds on Battery",          ST1, SECONDSONBATTERY,   TRUE, FALSE, FALSE},
   {82, FALSE, "Shutdown Type",               SDA, SHUTDOWNTYPE,       TRUE, FALSE, FALSE},
   {83, FALSE, "Shutdown After Delay",        PSD, SHUTDOWNAFTERDELAY,FALSE, FALSE, FALSE},
   {84, FALSE, "Software Version",            VER, SOFTWAREVERSION,   FALSE, FALSE,  TRUE},
   {85, FALSE, "Startup After Delay",         STD, STARTUPAFTERDELAY, FALSE, FALSE, FALSE},
   {86, FALSE, "Test Results Detail",         STR, TESTRESULTSDETAIL, FALSE, FALSE,  TRUE},
   {87, FALSE, "Test Results Summary",        STR, TESTRESULTSSUMMARY, TRUE,  TRUE, FALSE},
   {88, FALSE, "Test Type",                   TST, TESTTYPE,          FALSE, FALSE, FALSE},
   {89, FALSE, "UPS Baud Rate",               UBR, UPSBAUDRATE,        TRUE, FALSE, FALSE}
};

#define PARAMETERCOUNT (sizeof upstable / sizeof (struct upsentry))
#define MAXPARAMETERCOUNT 90

static char modes [SECLINES][16];                       /* Room to save the mode strings */

static           int    taskId [SECLINES];

static           char  *tasknames [] =
                           {"SEC Ch0", "SEC Ch1", "SEC Ch2", "SEC Ch3",
                            "SEC Ch4", "SEC Ch5", "SEC Ch6", "SEC Ch7",
                            "SEC Ch8", "SEC Ch9", "SEC Ch10","SEC Ch11",
                            "SEC Ch12","SEC Ch13","SEC Ch14","SEC Ch15",
                            "SEC Ch16","SEC Ch17","SEC Ch18","SEC Ch19",
                            "SEC Ch20","SEC Ch21","SEC Ch22","SEC Ch23",
                            "SEC Ch24","SEC Ch25","SEC Ch26","SEC Ch27",
                            "SEC Ch28","SEC Ch29","SEC Ch30","SEC Ch31",
                            "SEC Ch32"};

static unsigned short secResGroups [SECLINES] =
			{0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0} ; /* line driver group assignments   */


/**********
*
*    This routine is required by iocinit in the EPICS system.
*
**********/
static long init ()
{
    return (0);
}

void secset_rts(unsigned short line)
{
    set_rts(link[line]);
}

void secclear_rts(unsigned short line)
{
    clear_rts(link[line]);
}


/********
**
**    These routine call common help routine
**
********/
void secHelp (void)
{
    sechelp();
}
void Sechelp (void)
{
    sechelp();
}
void Sec_help (void)
{
    sechelp();
}
void sec_help (void)
{
    sechelp();
}
/********
**
**    This routine outputs information to help with debugging.
**
********/
void sechelp (void)
{
    printf ("Starting the Sec Driver:\n");
    printf ("\n");
    printf (" startsec(N,C,T1,0,T2,T3,P)                                                       \n");
    printf ("                      N: line number                                              \n");
    printf ("                      C: communications parameter string (comma delimited ascii)  \n"); 
    printf ("                     T1: intermessage delay (in ticks).                           \n");
    printf ("                      0: always set to zero.                                      \n");
    printf ("                     T2: post RTS assert, pre TX delay (in ticks).                \n");
    printf ("                     T3: post TX, pre RTS drop delay (in ticks).                  \n");
    printf ("                      P: Private/Redundant mode.  0 = redundant, !0 = private.    \n");
    printf ("                                                                                  \n");
    printf (" Example: startsec(2,\"19200,N,8,1\",4,1,0,0,0)                                   \n");
    printf ("                                                                                  \n");
    printf ("  (v%d)                                                               \n", VERSIONNUM);
    printf ("\n");
    printf ("\n");

    taskDelay (30); /* Time for operator to see it and pause the output */

    printf ("Sec Driver Control & Display Functions (* = default)\n");
    printf ("\n");
    printf (" sectraceon(N)      Enable the following debug messages for line N:             \n");
    printf ("                         1) Major Scan loop Info:                                \n");
    printf ("                           scan time, total output messages, and total messages .\n"); 
    printf ("                         3) All communication errors encountered in msgxfer.     \n");
    printf ("*sectraceoff(N)     Disable all of the modtraceon debug messages for line N.    \n");
    printf ("\n");
    printf (" secrestrace(N,X)   Resource Trace Level: 0 = off. 1 = state change; 2 = delay. \n");
    printf ("*secrestraceon(N)   Enable display of master/slave/etc state changes in secdriver. \n");
    printf (" secrestraceoff(N)  Disable display of master/slave/etc state changes in secdriver.\n");
    printf ("\n");

    taskDelay (30); /* Time for operator to see it and pause the output */

    printf (" seccontraceon(N)     Enable display of configuration errors in driver calls.   \n");
    printf ("*seccontraceoff(N)    Disable display of configuration errors in driver calls.  \n");
    printf (" secruntraceon(N)     Enable display of runtime errors (queue overflow).        \n");
    printf ("*secruntraceoff(N)    Disable display of runtime errors (queue overflow).       \n");
    printf (" secrejecttraceon(N)  Enable display of driver call rejections in driver calls. \n");
    printf ("*secrejecttraceoff(N) Disable display of driver call rejections in driver calls.\n");
    printf (" secdetailtraceon(N)  Enable display of command response parsing.               \n");
    printf ("*secdetailtraceoff(N) Disable display of  command response parsing.             \n");
    printf ("\n");

    taskDelay (30); /* Time for operator to see it and pause the output */

    printf (" sectrlv(N,X)       Enable the following message for line N:                   \n");
    printf ("                         X=0,  display none of the messages below.              \n");
    printf ("                         X=1,  beginning of scan loop, bringing modules online. \n");
    printf ("                         X=2,  beginning of scan.                               \n");
    printf ("                         X=3 to 7,   reserved                                   \n");
    printf ("                         X=8,  AIDRIVER: normal call return.                    \n");
    printf ("                         X=9,  BIDRIVER: normal call return.                    \n");
    printf ("                         X=10 to 254,  reserved.                                \n");
    printf ("\n");

    taskDelay (30); /* Time for operator to see it and pause the output */

    printf (" sectrafficon(N)    Enable traffic tracing for all module addresses            \n");
    printf ("                         and all modbus functions on line N.                    \n");
    printf ("*sectrafficoff(N)   Disable all traffic tracing on line N.                     \n");
    printf (" secdefaulttrace(N) Restores default setting for all traces.                   \n");
    printf ("\n");

    taskDelay (30); /* Time for operator to see it and pause the output */

    printf (" secinfo(N)            Display startup and current state information.            \n");
    printf (" seccomm(N,t1,h,t2,t3) Set communication timing parameters for line N as follows:\n");
    printf ("                         t1: intermessage delay (in ticks).                       \n");
    printf ("                          h: 1 = half duplex.                                     \n");
    printf ("                         t2: post RTS assert, pre TX delay (in ticks).            \n");
    printf ("                         t3: post TX, pre RTS drop delay (in ticks).              \n");
    printf ("\n");
    printf (" secdumpmsgcnts(N)      Display line N communication statistics.                 \n");
    printf (" secclearmsgcnts(N)     Display and clear line N communication statistics.       \n");
    printf (" secdumpstats(N)        Display line N, module X, communication statistics.      \n");
    printf (" secclearstats(N)       Display and clear line N, xfers, status & errors.        \n");
    printf (" sectableinfo(N)        Display line N,xfers, status, errors and input data.     \n");
    printf (" seclineinfo            Display simple status info for all sec driver lines.     \n");

    printf (" sechelp                Display this information.                                \n");
    printf ("\n");
    printf (" secgrouphelp       Display information on Resource Group commands.               \n");

    printf (" secparseinvalidcount(N) Set number of sequential parse errors allowed for each   \n");
    printf ("                          'read status group N' command before marking the data   \n");
    printf ("                          returned by that command as invalid. Default = %d.\n",PARSEINVALIDDEFAULT);
    printf (" secparseofflinecount(N) Set number of sequential parse errors allowed for all    \n");
    printf ("                          'read status group N' commands before marking the plc   \n");
    printf ("                          as offline.  Default = %d.\n",PARSEOFFLINEDEFAULT);
}




void secparseinvalidcount (unsigned short line, unsigned short count)
{
    if (line >= SECLINES)
    {
        logMsg ("Invalid line number %d: line number must be less than %d.\n", line, SECLINES, 0, 0, 0, 0);
        return;
    }
    if (count)
    {
        logMsg ("The number of sequential parse errors allowed for each 'read status group N' command\n", 0, 0, 0, 0, 0, 0);
        logMsg (" before marking the data returned by that command as invalid has been set to %d.\n", count, 0, 0, 0, 0, 0);
        parseinvalidcount [line] = count;
        return;
    }
    logMsg ("The number of sequential parse errors allowed for each 'read status group N' command\n", 0, 0, 0, 0, 0, 0);
    logMsg (" before marking the data returned by that command as invalid is currently set to %d.\n", parseinvalidcount [line], 0, 0, 0, 0, 0);
    return;
}

void secparseofflinecount  (unsigned short line, unsigned short count)
{
    if (line >= SECLINES)
    {
        logMsg ("Invalid line number %d: line number must be less than %d.\n", line, SECLINES, 0, 0, 0, 0);
        return;
    }
    if (count)
    {
        logMsg ("The number of sequential parse errors allowed for each 'read status group N' command\n", 0, 0, 0, 0, 0, 0);
        logMsg (" before marking the data returned by that command as invalid has been set to %d.\n", count, 0, 0, 0, 0, 0);
        parseofflinecount [line] = count;
        return;
    }
    logMsg ("The number of sequential parse errors allowed for all read status group N' commands \n", 0, 0, 0, 0, 0, 0);
    logMsg (" before marking the plc as offline is currently set to %d.\n", parseofflinecount [line], 0, 0, 0, 0, 0);
    return;
}


/********
**
**    This routine enables logging of comm trace messages
**
********/
void sectraceon (unsigned short line)
{
    modtrace [line] = TRUE;
}

/********
**
**    This routine disables logging of comm trace messages
**
********/
void sectraceoff (unsigned short line)
{
    modtrace [line] = FALSE;
}

/********
**
**    This routine enables detailed display of normal communications traffic
**
********/
void sectrafficon (unsigned short line)
{
    trace [line] = trace [line] | ASCIITRAFFICTRACE;
    trace [line] = trace [line] | ALLTRAFFICTRACE;
        traceadr [line] = 255;  /* every address */
        tracetype [line] = 255; /* every message type */
}

/********
**
**    This routine disables detailed display of normal communications traffic
**
********/
void sectrafficoff (unsigned short line)
{
    trace [line] = trace [line] & ~ALLTRAFFICTRACE;
    trace [line] = trace [line] & ~ASCIITRAFFICTRACE;
}

/********
**
**    This routine enables detailed display of normal communications traffic in ascii
**
********/
void secasciion (unsigned short line)
{
    trace [line] = trace [line] | ASCIITRAFFICTRACE;
}

/********
**
**    This routine returns detailed display of normal communications traffic in ascii-hex
**
********/
void secasciioff (unsigned short line)
{
    trace [line] = trace [line] & ~ASCIITRAFFICTRACE;
}

/********
**
**    This routine enables logging of resource trace messages
**
********/
void secrestrlv (unsigned short line, unsigned short level)
{
    modrestrace [line] = level;
}

/********
**
**    This routine enables logging of resource trace messages
**
********/
void secrestraceon (unsigned short line)
{
    modrestrace [line] = 1;
}

/********
**
**    This routine disables logging of resource trace messages
**
********/
void secrestraceoff (unsigned short line)
{
    modrestrace [line] = 0;
}

/********
**
**    This routine controls logging of trace messages
**
********/
void sectrlv (unsigned short line, unsigned short level)
{
    mltrace [line] = level;
}

/********
**
**    These routines enable and disable tracing of database configuration errors in driver calls
**
********/
void seccontraceon (unsigned short line)
{
    trace[line] = trace[line] | CONFIGURATIONTRACE;
}

void seccontraceoff (unsigned short line)
{
    trace[line] = trace[line] & ~CONFIGURATIONTRACE;
}

/********
**
**    These routines enable and disable tracing of runtime errors in driver calls
**
********/
void secruntraceon (unsigned short line)
{
    trace[line] = trace[line] | RUNTIMEERRORTRACE;
}

void secruntraceoff (unsigned short line)
{
    trace[line] = trace[line] & ~RUNTIMEERRORTRACE;
}

/********
**
**    These routines enable and disable tracing of call rejection errors in driver calls
**
********/
void secrejecttraceon (unsigned short line)
{
    trace[line] = trace[line] | CALLREJECTTRACE;
}

void secrejecttraceoff (unsigned short line)
{
    trace[line] = trace[line] & ~CALLREJECTTRACE;
}


/********
**
**    These routines enable and disable tracing of details of command parsing.
**
********/
void secdetailtraceon (unsigned short line)
{
    trace[line] = trace[line] | DETAILEDTRACE;
}

void secdetailtraceoff (unsigned short line)
{
    trace[line] = trace[line] & ~DETAILEDTRACE;
}

/********
**
**    These routines enable and disable tracing of call rejection errors in driver calls
**
********/
void npallstatuson (unsigned short line)
{
    trace[line] = trace[line] |
                  OKTRACE |
                  COMMTRACE |
                  PROGTRACE |
                  FATALTRACE |
                  CONFIGTRACE |
                  UNKNOWNTRACE |
                  0;
}

void npallstatusoff (unsigned short line)
{
    trace[line] = trace[line] &
                  ~OKTRACE &
                  ~COMMTRACE &
                  ~PROGTRACE &
                  ~FATALTRACE &
                  ~CONFIGTRACE &
                  ~UNKNOWNTRACE &
                  ~0;
}

/********
**
**    These routines enable and disable tracing of call rejection errors in driver calls
**
********/
void npbadstatuson (unsigned short line)
{
    trace[line] = trace[line] |
                  COMMTRACE |
                  PROGTRACE |
                  FATALTRACE |
                  CONFIGTRACE |
                  UNKNOWNTRACE |
                  0;
}

void npbadstatusoff (unsigned short line)
{
    trace[line] = trace[line] &
                  ~COMMTRACE &
                  ~PROGTRACE &
                  ~FATALTRACE &
                  ~CONFIGTRACE &
                  ~UNKNOWNTRACE &
                  ~0;
}

/********
**
**    This routine restores default setting for all traces.
**
********/
void secdefaulttrace (unsigned short line)
{
    /* debug structures */
    modtrace [line] = FALSE;
    mltrace [line] = 0;
    modrestrace [line] = 1;
    trace [line] = 0;
    traceadr [line] = 255;  /* every address      */
    tracetype [line] = 255; /* every message type */
}



int startsec (unsigned short tlink,       /* the line number of this driver.                      */
                  char *mode,             /* UART initializations string (comma delimited ascii). */
                  unsigned short delay,   /* intermessage delay value, expressed in ticks.        */
                  unsigned short slug,    /* half duplex flag placeholder: ignored.               */
                  unsigned short delay1,  /* half duplex post RTS assert, pre TX delay.           */
                  unsigned short delay2,  /* half duplex post TX, pre RTS drop delay.             */
                  unsigned short private) /* private flag. if 0, normal. if !0, no partner driver.*/
{
    unsigned short	line;

	if (tlink >=FILELINESTART) {
	    line = tlink - FILELINESTART;
	    if (line >= SECLINES)
		return(-1); 
	} else {
	    line = tlink;
	}


    if (line < SECLINES)
    {
	strncpy (modes [line],mode,sizeof(modes[line]));
	link[line] = tlink;
        taskId [line] = taskSpawn (tasknames [line],
                                   SEC_PRIORITY,
                                   SEC_TASK_OPTIONS,
                                   SEC_STACK,
                                   (FUNCPTR) secdriver,
                                   (int) line,
                                   (int) modes[line],
                                   (int) delay,
                                   (int) slug,
                                   (int) delay1,
                                   (int) delay2,
                                   (int) private, 0, 0, 0);

        logMsg ("Started the SEC UPS Standard Protocol driver on line %d\n",line, 0, 0, 0, 0, 0);

        remCurIdGet (user,password);
        return (1);
    }
    else
    {
        logMsg ("Invalid line number %d\n", line, 0, 0, 0, 0, 0);
        return (-1);
    }
}

/********
**
** Set communication logic time delays to handle radio delays and half duplex delays.
**
********/
void seccomm (unsigned short line,
              unsigned short tickcount1,
              unsigned short hdflag,
              unsigned short tickcount2,
              unsigned short tickcount3)
{
	intermsgdelay [line] = tickcount1;  /* at beginning new message/response sequence.               */
	halfduplex [line]    = hdflag;
	pretxdelay [line]    = tickcount2;  /* after asserting RTS but before beginning transmission.    */
	posttxdelay [line]   = tickcount3;  /* after sending last character but before dropping RTS.     */
}


int secResGroup (unsigned short line, unsigned short group)
{
    if (line < SECLINES)
    {
        if (group <= MAXGROUPNUM)
        {
            secResGroups [line] = group;
            logMsg ("Assigned Line %d to Group %d\n", line, group, 0, 0, 0, 0);
            return (1);
        }
        else
        {
            logMsg ("Invalid Group Assignment: line %d: group number must be less than %d.\n", line, MAXGROUPNUM+1, 0, 0, 0, 0);
            return (-1);
        }
    }
    else
    {
        logMsg ("Invalid Group Assignment: line %d: line number must be less than %d\n", line, SECLINES, 0, 0, 0, 0);
        return (-1);
    }
}

void secgrouptable (void)
{
    unsigned short x, y, i, j, index;

    printf (" Resource Group Lock Manager:  Limitorque Driver Line to Group Map Table\n");
    printf (" -----------------------------------------------------------------------\n");
    printf ("  Ln:Grp   Ln:Grp   Ln:Grp   Ln:Grp   Ln:Grp   Ln:Grp   Ln:Grp   Ln:Grp \n");

    x = 8;
    y = SECLINES/x;

    for (i = 0; i < y; i++)
    {
        for (j = 0; j < x; j++)
        {
            index = i + j*y;
            printf ("  %d%d:%d%d  ",
                    index/10,
                    index%10,
                    secResGroups [index]/10,
                    secResGroups [index]%10);
        }
        printf ("\n");
    }
}

void seclocktable (void)
{
    unsigned short x, y, i, j, index;

    printf (" Resource Group Lock Manager Global Array\n");
    printf (" ----------------------------------------\n");
    printf ("  Grp:Flg   Grp:Flg   Grp:Flg   Grp:Flg\n");

    x = 4;
    y = MAXGROUPNUM/x;

    for (i = 0; i < y; i++)
    {
        for (j = 0; j < x; j++)
        {
            index = i + j*y;
            printf ("  %d%d:%d%d   ",
                    (index+1)/10,
                    (index+1)%10,
                    resGroupLock [index]/10,
                    resGroupLock [index]%10);
        }
        printf ("\n");
    }
}

void seclockall (void)
{
    unsigned short i;

    printf (" Resource Group Lock Manager Global Array\n");
    printf (" ----------------------------------------\n");
    printf ("  Group  Flag  Trip Name\n");
    for (i = 0; i < MAXGROUPNUM; i++)
    {
        printf ("   %d%d     %d%d",
                    (i+1)/10,
                    (i+1)%10,
                    resGroupLock [i]/10,
                    resGroupLock [i]%10);
        if (resGroupTripName[i])
        {
            printf ("   %s\n", resGroupTripName[i]);
        }
        else
        {
            printf ("\n");
        }
    }
}

void secsetlock (unsigned short group)
{
    if (group <= MAXGROUPNUM)
    {
        printf ("Setting Resource Lock Flag for Group %d.\n", group);
        taskDelay(60);
        resGroupLock [group -1] = 1;
    }
    else
    {
        logMsg ("Invalid Group %d: group number must be less than %d.\n", group, MAXGROUPNUM+1, 0, 0, 0, 0);
    }
}

void secclrlock (unsigned short group)
{
    if (group <= MAXGROUPNUM)
    {
        printf ("Clearing Resource Lock Flag for Group %d.\n", group);
        taskDelay(60);
        resGroupLock [group -1] = 0;
    }
    else
    {
        logMsg ("Invalid Group %d: group number must be less than %d.\n", group, MAXGROUPNUM+1, 0, 0, 0, 0);
    }
}

void secgrouphelp (void)
{
    printf (" SEC UPS Standard Protocol Driver Group Related Functions\n");
    printf ("\n");
    printf (" secResGroup(N,G)      Assign line N to group G (valid G range is 1 to 16)         \n");
    printf (" secgrouptable         Display the Limitorque's line to group assignment table.    \n");
    printf (" seclocktable          Display the resource group lock Global Array.               \n");
    printf (" seclockall            Display the resource group lock and trip name Global Array. \n");
    printf (" secsetlock (G)        Set the resource group lock for group G. (1 to 16)          \n");
    printf (" secclrlock (G)        Clear the resource group lock for group G. (1 to 16)        \n");
    printf (" secgrouphelp          Display this help message.                                  \n");
    printf ("\n");
}


/********
**
** Show Driver Info and Current State Information  
**
********/
void secinfo (unsigned short line)
{
    printf ("intermsgdelay [%d]     = %d\n", line, intermsgdelay[line]);
    printf ("halfduplex [%d]        = %d\n", line, halfduplex[line]);
    printf ("pretxdelay [%d]        = %d\n", line, pretxdelay[line]);
    printf ("posttxdelay [%d]       = %d\n", line, posttxdelay[line]);
    printf ("pryvate [%d]           = %d\n", line, pryvate[line]);
    printf ("resresponse [%d]       = %d\n", line, resresponse[line]);
    printf ("activity_time [%d]     = %d\n", line, (int) activity_time[line]);
    printf ("line_activity [%d]     = %d\n", line, line_activity[line]);
    printf ("lastactivity [%d]      = %d\n", line, lastactivity[line]);
    printf ("modtrace [%d]          = %d\n", line, modtrace[line]);
    printf ("mltrace [%d]           = %d\n", line, mltrace[line]);
    printf ("modrestrace [%d]       = %d\n", line, modrestrace[line]);
    printf ("parseinvalidcount [%d] = %d\n", line, parseinvalidcount[line]);
    printf ("parseofflinecount [%d] = %d\n", line, parseofflinecount[line]);
}

/*                                                                              
**    secdriver                                                                 
**                                                                              
**      The purpose of this task is to scan the sec module lines.               
**      It will maintain a configuration list supporting online additions       
**      and replacement of modules, and maintain a table of most recently       
**      scanned values.                                                         
**                                                                              
*/                                                                                         
/* not called from within this driver     */

void secdriver (unsigned short line,
           unsigned char *mode,
           unsigned short delay,
           unsigned short hdflag,
           unsigned short delay1,
           unsigned short delay2,
           unsigned short private)
{
    unsigned short  mod, cfgchk, cnt, modspolled, onlinefreq;
    unsigned short  activity, pstate, lstate, execute, command,i;
    long            this_time, last_time;
    SEC             *pmod;
    RXLINE          *rxp;
    unsigned int    temptick, tempdiff;
    int             temp;
    unsigned short  group;

    /* Initialize the data structures */
    intermsgdelay [line] = delay;
    halfduplex [line]    = hdflag;
    pretxdelay [line]    = delay1;
    posttxdelay [line]   = delay2;
    pryvate [line]       = private;

    parseinvalidcount [line] = PARSEINVALIDDEFAULT;

    parseofflinecount [line] = PARSEOFFLINEDEFAULT;

    activity = FALSE;

    resresponse [line] = FALSE;
    resnames [line] [0] = 0;
    prestable [line] = NULL;

    activity_time [line] = tickGet ();
    lastactivity [line] = line_activity [line] = FALSE;

    pstatewait[line] = 60;      /* will wait up to one minute for partner to become valid */

    /* debug structures */
    /* moddebugtrace (line); */
    secdefaulttrace (line);

    tick [line] [0] = tick [line] [1] = tick [line] [2] = tick [line] [3] = 0;

    linestats [line].TXintcount = 0;
    linestats [line].RXintcount = 0;
    linestats [line].RXbadcount = 0;
    linestats [line].msgxfercount = 0;
    linestats [line].startingtick = tickGet ();

    /* communication initialization */
    rxlinetab [line] = (RXLINE *) calloc (1, (sizeof (RXLINE)));
    if (!rxlinetab [line])
    {
        logMsg ("Not enough memory for line %d data structure (comm: rx)\n",line, 0, 0, 0, 0, 0);
        return;
    }
    txlinetab [line] = (TXLINE *) calloc (1, (sizeof (TXLINE)));
    if (!txlinetab [line])
    {
        logMsg ("Not enough memory for line %d data structure (comm: tx)\n",line, 0, 0, 0, 0, 0);
        return;
    }

    taskLock ();  /* make terminal display easier to read - avoids interleaving of other task data */
    comm_init (link[line], mode, (FUNCPTR) tx_int_proc, (FUNCPTR) rx_int_proc);
    clear_rts (link[line]);
    taskUnlock ();

    onlinefreq = 1;   /* causes on_line calls to begin on first pass if master */

    /* Initialize the Rx interrupts to receive once for activity check         */
    /* enable the receiver interrupt after set to ignore                       */

    rxp = rxlinetab [line];
    rxp->rxstate = RXIGNORE;
    enable_rx_ints (link[line]);

    /* Initialize the line communication semaphores */
    semids [line] [0] = semBCreate (SEM_Q_PRIORITY, SEM_EMPTY);
    semids [line] [1] = semBCreate (SEM_Q_PRIORITY, SEM_EMPTY);

    /* Wait for resource to be defined via an initial call to 'register_resource' via driver call.  */
    while ((resnames [line] [0] == 0) || (prestable [line] == NULL))
    {
        if (modrestrace [line] > 1)
        {
            logMsg ( "No Resource Defined for this line.\n", 0, 0, 0, 0, 0, 0);
        }
        taskDelay (600);
    }
    /* diagnostic message */
    if (modrestrace [line] > 1)
    {
        logMsg ( "\nResource has now been defined for this line.\n", 0, 0, 0, 0, 0, 0);
    } 
    taskDelay (60);

    if (pryvate [line])
    {
            /* mark the restable as private. */
            rmPrivate (resnames [line]);

        /*****************************/
        /* Immediately become master */
        /*****************************/

        resSetCaAnswer (prestable [line]);
        resPutLocalState (prestable [line], RES_MSTR);

        lastscan [line] = tickGet ();
        scantime [line] = 0;

        lstate = RES_MSTR;
        if (modrestrace [line])
        {
            logMsg ("0] Private: Initialize as RES_MSTR\n", 0, 0, 0, 0, 0, 0);
        }
    }
    else /* this is the 'normal' path */
    {
        /**********************/
        /* Start out as slave */

        resClearCaAnswer (prestable [line]);        /* start quiet */
        resPutLocalState (prestable [line], RES_SLV);

        if (modrestrace [line])
        {
            logMsg ("1] Initialized as RES_SLV\n", 0, 0, 0, 0, 0, 0);
        }

        /* wait finite time for pstate to be resolved */
        logMsg ("Begin wait for pstate [%d] to become valid.\n", pstatewait [line], 0, 0, 0, 0, 0);
        for (i = 0; i <= pstatewait [line]; i++)
        {
            resGetStates (prestable [line], &lstate, &pstate);
            if (pstate != 0)
            {
                break;
            }
            taskDelay (60);
        }

        if (pstate == 0)
        {
            logMsg ("Timeout of pstate wait.\n", 0, 0, 0, 0, 0, 0);
        }
        else
        {
            logMsg ("Completion of pstate wait.\n", 0, 0, 0, 0, 0, 0);
        }

        for (i = 0; i < ACTIVITYTIME/10; i++) ;  /* ten seconds */
        {
            taskDelay(10);
        }

        activity = check_line_activity(line, 0);
        resGetStates (prestable [line], &lstate, &pstate);

        /******************************************/
        /* if the partner is not there, be master */

        if ((pstate != RES_MSTR) && (pstate != RES_SLVWANT) && (!activity))
        {
            resSetCaAnswer (prestable [line]);
            resPutLocalState (prestable [line], RES_MSTR);
            lastscan [line] = tickGet ();
            scantime [line] = 0;
            lstate = RES_MSTR;
            if (modrestrace [line])
            {
                logMsg ("2] Became RES_MSTR   P= %x A= %d \n", pstate, activity, 0, 0, 0, 0);
            }
            prepresponses (line);
        }
    }

    cfgchk = 0;               /* Init the new config pointer.  */
    last_time = tickGet ();   /* Initialize the time.          */

    while (TRUE)
    {
        activity_time [line] = tickGet ();
        lastactivity [line] = line_activity [line] = FALSE;

        /************************************/
        /* check for line group disablement */
        /************************************/
        group = secResGroups [line];
        if ((group) && (resGroupLock [group-1]))
        {
            if (modrestrace [line])
            {
                logMsg ("20] Forced to RES_SLV.\n", 0, 0, 0, 0, 0, 0);
            }
            resClearCaAnswer (prestable [line]);
            resPutLocalState (prestable [line], RES_SLV);
            lstate = RES_SLV;
            resresponse [line] = FALSE; 	/* Make Sure driver is not responding   */
            if (modrestrace [line])
            {
                logMsg ("21] Resource Group Locked     P= %x A= %d \n", pstate, activity, 0, 0, 0, 0);
            }
            while (resGroupLock [group-1])
            {
                group = secResGroups [line];
                if (group)
                {
                    clear_rts (link[line]); 		/* Make sure line is left for master    */
                    taskDelay (2);
                }
                else /* the line has been taken out of the group */
                {
                    break;
                }
            }
            if (modrestrace [line])
            {
                logMsg ("22] Resource Group Unlocked   P= %x A= %d \n", pstate, activity, 0, 0, 0, 0);
                logMsg ("23] Released to RES_SLV.\n", 0, 0, 0, 0, 0, 0);
            }
        }

        /*********************************************************************************************/
        /* we arrive here because we are a slave.                                                    */
        /* we stay until:   asked to be master by partner. if we can reach modules, we become master.*/
        /*                  the partner is not master and there is no activity on the line.          */
        /*********************************************************************************************/

        while (lstate == RES_SLV)
        {
            if (pstate != RES_SLVWANT)
            {
                clear_rts (link[line]);                      /* make sure line is left for master    */
            }
            resresponse [line] = FALSE;                /* make sure driver is not responding   */
            resPutCommand (prestable [line], 0);       /* clear out commands                   */
            activity = check_line_activity (line, 10); /* wait 10 seconds before checking      */
            resGetStates (prestable [line], &lstate, &pstate);

            /************************************/
            /* check for line group disablement */
            if (checkgroup (line, &lstate, 24)) break;

            /****************************************************************************************/
            /* see if needed by partner to be master. if yes, do it and break out of slave state.   */
            if (pstate == RES_SLVWANTCA)
            {
                resPutLocalState (prestable [line], RES_MSTR);
                lstate = RES_MSTR;

                lastscan [line] = tickGet ();
                scantime [line] = 0;

                if (modrestrace [line])
                {
                    logMsg ("3] Became RES_MSTR  P= %x A= %d \n",pstate,activity, 0, 0, 0, 0);
                }
                taskDelay (60);
                prepresponses (line);
                resSetCaAnswer (prestable [line]);
                onlinefreq = 1;   /* causes on_line calls to begin on first pass as master */
                break;
            }
            /***************************************************************/
            /* If partner is there and wants to give it up, try to take it */

            if (pstate == RES_SLVWANT)
            {
                execute = FALSE;
                taskDelay (2); /* Time for switch to occur */

                /* See if anyone is out there and able to respond to a loopback command */
                for (mod = 0; (mod < NUMMODS); mod++)
                {
                    /************************************/
                    /* check for line group disablement */
                    if (checkgroup (line, &lstate, 25)) break;

                    pmod = modules [line];
                    if ((pmod) &&            /* verify pointer validity prior to use */
                        (pmod->was_requested) &&    /* data points have been requested      */
                        (!pmod->cmd_force_offline)) /* user has NOT forced module offline      */
                    {
                        on_line (line);
                        /* check cmd_force_offline flag before AND after due to asynchronous tasks */
                        if ((pmod->is_responding) && (!pmod->cmd_force_offline))
                        {
                            execute = TRUE;
                            break;
                        }
                    }
                }
                /************************************/
                /* check for line group disablement */
                if (checkgroup (line, &lstate, 26)) break;

                /* If someone answered, become master and make sure partner really gave it up */
                if (execute)
                {
                    resPutLocalState (prestable [line], RES_MSTR);
                    lstate = RES_MSTR;

                    lastscan [line] = tickGet ();
                    scantime [line] = 0;

                    if (modrestrace [line])
                    {
                        logMsg ("4] Became RES_MSTR  P= %x A= %d \n",pstate,activity, 0, 0, 0, 0);
                    }
                    execute = FALSE;
                    activity_time [line] = tickGet ();
                    lastactivity [line] = line_activity [line] = FALSE;
                    for (i = 0; (i < 60); i++)
                    {
                        activity = check_line_activity (line, 2); /* pause before testing */
                        resGetStates (prestable [line], &lstate, &pstate);
                        if ((pstate != RES_MSTR) && (pstate != RES_SLVWANT) && (!activity) )
                        {
                            execute = TRUE; /* Partner Gave It Up so take CA */
                            break;
                        }
                    }
                    if (!execute)
                    {   /* Partner did not give it up go back to slave */
                        resPutLocalState (prestable [line], RES_SLV);
                        lstate = RES_SLV;
                        if (modrestrace [line])
                        {
                            logMsg ("5] Became RES_SLV  P= %x A= %d \n",pstate,activity, 0, 0, 0, 0);
                        }
                    }
                    else
                    { /* Assume control by taking the CA responsibilities */
                        resSetCaAnswer (prestable [line]);
                        prepresponses (line);
                        break;
                    }
                }
            }

            /********************************************/
            /* if the partner is not talking, be master */

            if ((pstate != RES_MSTR) &&  (pstate != RES_SLVWANT) && (!activity))
            {
                resSetCaAnswer (prestable [line]);
                resPutLocalState (prestable [line], RES_MSTR);
                lstate = RES_MSTR;

                lastscan [line] = tickGet ();
                scantime [line] = 0;

                if (modrestrace [line])
                {
                    logMsg ("6] Became RES_MSTR P= %x A= %d \n",pstate,activity, 0, 0, 0, 0);
                }
                prepresponses (line);
                onlinefreq = 1;   /* causes on_line calls to begin on first pass as master */
            }
        }

        /****************************************************************************************/
        /* to arrive here, we used to be master, and had been told to relinquish master status. */
        /* stay in this loop until the partner takes over, or the command is canceled.          */
        /****************************************************************************************/
        while (lstate == RES_SLVWANTCA)
        {
            taskDelay (2);                   /* Give OS a chance...  */

            /************************************/
            /* check for line group disablement */
            if (checkgroup (line, &lstate, 27)) break;

            clear_rts (link[line]);                /* Make sure line is left for master    */
            resresponse [line] = FALSE;      /* Make Sure driver is not responding   */

            resGetStates (prestable [line], &lstate, &pstate);
            if ((pstate == RES_MSTR) || (pstate == RES_SLVWANT))
            {
                resClearCaAnswer (prestable [line]);
                resPutLocalState (prestable [line], RES_SLV);
                lstate = RES_SLV;
                if (modrestrace [line])
                {
                    logMsg ("7] Became RES_SLV  P= %x A= %d \n",pstate,activity, 0, 0, 0, 0);
                }
                break;
            }
            resGetCommand (prestable [line], &command);
            resPutCommand (prestable [line], 0); /* clear out commands */
            if (command == RES_CANCELREL)
            {
                resSetCaAnswer (prestable [line]);
                resPutLocalState (prestable [line], RES_MSTR);
                lstate = RES_MSTR;
                if (modrestrace [line])
                {
                    logMsg ("8] Became RES_MSTR  P= %x A= %d \n",pstate,activity, 0, 0, 0, 0);
                }
                prepresponses (line);
            }
        }

        /****************************************************************************************/
        /* in normal operation this code should never execute. it has, however, been seen to    */
        /* execute when the database has been inadvertently configured with the same resource   */
        /* assigned to two different drivers in the same IOC. The state does funny things.      */
        /****************************************************************************************/
        while (lstate == RES_SLVWANT)
        {
            /************************************/
            /* check for line group disablement */
            if (checkgroup (line, &lstate, 28)) break;

            if ((modrestrace [line]) || (trace[line] & CONFIGURATIONTRACE))
            {
                /* condition should NEVER occur. print error message every 5 seconds. */
                logMsg ("Illegal Local Resource State.\n", 0, 0, 0, 0, 0, 0);
                if (trace[line] & CONFIGURATIONTRACE)
                {
                    logMsg ("  Possible Resource Configuration Error:\n", 0, 0, 0, 0, 0, 0);
                    logMsg ("   Check for multiple drivers sharing single Resource.\n", 0, 0, 0, 0, 0, 0);
                }
                taskDelay (600);
            }
            taskDelay (2);
        }

        /****************************************************************************************/
        /* we come here because we have gone to, or are looping in, the master state.           */
        /* we stay in this loop until:    we have been told to give master state to the partner.*/
        /*                                we break out to RES_SLVWANTCA.                        */
        /*                                we have successfully handed master state to partner.  */
        /* While master, this task will process all modules sending outputs and getting inputs. */
        /* It will then try to configure requested non-responding modules.                      */
        /****************************************************************************************/
        last_time = tickGet ();        /* initialize gap */
        while (lstate == RES_MSTR)
        {
            if (mltrace [line] == 1)
            {
                logMsg ( "Top of Master Loop... \n", 0, 0, 0, 0, 0, 0);
            }
            /************************************/
            /* test for line group disablement. */
            if (lstate == RES_LOCKED) break;

            /* initial stuff */
            resresponse [line] = TRUE; /* Make sure driver is responding. driver won't   */
                                       /* issue commands if not set.                     */

            resGetCommand (prestable [line], &command);
            resPutCommand (prestable [line], 0);        /* clear out commands */

            /************************************************************************/
            /* if relinquish command sent, change state and break out to outer loop */
            if (command == RES_RELINQUISH)
            {
                resClearCaAnswer (prestable [line]);
                resPutLocalState (prestable [line], RES_SLVWANTCA);
                lstate = RES_SLVWANTCA;
                if (modrestrace [line])
                {
                    logMsg ("9] Became RES_SLVWANTCA  P= %x A= %d \n",pstate,activity, 0, 0, 0, 0);
                }
                break;
            }

            /* measure time between scans of the entire set of modules. display if requested. */
            temptick = tickGet ();
            tempdiff = temptick - lastscan [line];

            if (modtrace [line])
            {
                logMsg ("ScanTime: =%4d, # OutputMsgs =%5d, # TotalMsgs =%3d\n",
                         scantime[line], loopoutmsgs[line], loopmsgs[line], 0, 0, 0);
            }

            loopoutmsgs [line] = 0;
            loopmsgs [line] = 0;
            lastscan [line] = temptick;

            /* scan modules for data i/o, if allowed */

            modspolled = 0;

            /* call 'scan_module' for each Sec Module, unless the   */
            /* user has taken it off line or it is not responding.     */
            for (mod = 0; (mod < NUMMODS); mod++)
            {
                pmod = modules [line];  /* shorthand pointer                    */
                if (pmod)               /* verify pointer validity prior to use */
                {
                    /************************************/
                    /* check for line group disablement */
                    if (checkgroup (line, &lstate, 30)) break;

                    if (pmod->cmd_force_offline)
                    {
                        /* if user has turned this module offline via pseudo commands,   */
                        /* mark it as not responding.                                    */
                        pmod->is_responding = FALSE;
                        mark_not_valid (pmod);
                    }
                    else if (pmod->is_responding)
                    {
                        /* local/partner lockstep recovery */
                         resGetStates (prestable [line], &lstate, &pstate);
                        if (pstate == RES_MSTR)
                        {
                            resClearCaAnswer (prestable [line]);
                            resPutLocalState (prestable [line], RES_SLV);
                            lstate = RES_SLV;
                            clear_rts (link[line]);         /* Make sure line is left for master */
                            if (modrestrace [line])
                            {
                                logMsg ("13] Became RES_SLV  P= %x A= %d \n", pstate, activity, 0, 0, 0, 0);
                            }
                            temp = ((tickGet () % 8) + 5); /* random 5 to 12 second delay */
                            for ( i = 0; i < temp; i++)
                            {
                                taskDelay(60);
                            }
                            break;
                        }
                        modspolled++;

                        scan_module (pmod, line);          /* I/O done here!              */
                        if (!pmod->is_responding)
                        {
                            mark_not_valid (pmod);
                        }
                    }
                }
                /* end of single module action */

                /* if told to relinquish our master state, stop module scanning immediately. */
                /* immediately, don't trash the command, so that we can still read it on     */
                /* the next loop through this while statement.                               */
                resGetCommand (prestable [line], &command);
                if (command == RES_RELINQUISH)
                {
                    break;
                }
            } /* end of scan loop */

            /* test for lockstep detection. */
            if (lstate == RES_SLV) break;
            /* test for line group disablement. */
            if (lstate == RES_LOCKED) break;

            /* measure how long the scan took. to be displayed at beginning of next loop. */
            temptick = tickGet ();
            scantime [line] = temptick - lastscan [line];

            /* try and bring all nonresponding modules on line, but */
            /* don't bring in more than a specified number.         */

            cnt = 0;
            if (--onlinefreq == 0)         /* do we try to bring any in on this pass? */
            {
                onlinefreq = BRINGBACKCNT;
                for (mod = 0; (mod < NUMMODS); mod++)
                {
                    /************************************/
                    /* check for line group disablement */
                    if (checkgroup (line, &lstate, 31)) break;

                    cfgchk++;           /* manage module index */
                    if (cfgchk == NUMMODS)
                    {
                        cfgchk=0;    /* wrap module index */
                    }
                    pmod = modules [line];
                    if ((pmod) &&        /* verify pointer validity prior to use*/
                        (!pmod->is_responding) &&
                        (pmod->was_requested) &&
                        (!pmod->cmd_force_offline))
                    {
                        on_line (line);    /* loopback testing done here!     */

                        modspolled++;
                        cnt++;
                        if (cnt == MAXCFGCK)
                        {
                            break;
                        }
                    }
                }
            }

            /************************************/
            /* test for line group disablement. */
            if (lstate == RES_LOCKED) break;

            /* see if anyone on the line has been responding. */
            /* if not, try to get partner to become slave.    */
            /* Check for no one responding.                   */

            if ((!any_responding (line)) && (!pryvate [line]))
            {
                /* See if it is time to let partner try */
                this_time = tickGet ();
                if ((this_time - last_time) > (3 * SLWWINDOW))
                {   /* #ticks for slave to try */

                    /* Switch States to let partner know it's its time to try */

                    resPutLocalState (prestable [line], RES_SLVWANT);
                    lstate = RES_SLVWANT;
                    if (modrestrace [line])
                    {
                        logMsg ("10] Became RES_SLVWANT  P= %x A= %d \n",pstate, activity, 0, 0, 0, 0);
                    }
                    /* Give partner access to the hardware */
                    clear_rts (link[line]);

                    /* Remember the time of this try */
                    last_time = tickGet ();
                    lastactivity [line] = line_activity [line] = FALSE;
                    for (i = 0; (i < SLWWINDOW); i += 10)
                    {               /* SLWINDOW is the # of ticks for slave to try */
                        /************************************/
                        /* check for line group disablement */
                        if (checkgroup (line, &lstate, 32)) break;

                        taskDelay (10);
                        /* If partner picked up let it have control */
                        activity = check_line_activity (line, 0);  /* check immediately */
                        resGetStates (prestable [line], &lstate, &pstate);
                        if ((pstate == RES_MSTR) || (pstate == RES_SLVWANT) || (activity))
                        {
                            resClearCaAnswer (prestable [line]);
                            resPutLocalState (prestable [line], RES_SLV);
                            lstate = RES_SLV;
                            if (modrestrace [line])
                            {
                                logMsg ("11] Became RES_SLV  P= %x A= %d \n",pstate,activity, 0, 0, 0, 0);
                            }
                            taskDelay(60);
                            taskDelay(60);
                            taskDelay(60);
                            break;
                        }
                    }
                    /************************************/
                    /* test for line group disablement. */
                    if (lstate == RES_LOCKED) break;

                    /* Go Back to being the Master and take control of the hardware */
                    if (lstate != RES_SLV)
                    {
                        resSetCaAnswer (prestable [line]);
                        resPutLocalState (prestable [line], RES_MSTR);
                        lstate = RES_MSTR;
                        if (modrestrace [line])
                        {
                            logMsg ("12] Became RES_MSTR  P= %x A= %d \n",pstate,activity, 0, 0, 0, 0);
                        }
                        prepresponses (line);
                    }
                }
            }

            /* take modules off hold status. the module's data are not to be held off since now master. */
            for (mod = 0; (mod < NUMMODS); mod++ )
            {
                pmod = modules [line];
                if (pmod)            /* verify pointer validity prior to use */
                {
                    pmod->holdflag = FALSE;
                }
            }
            if (modspolled == 0)
            {
                taskDelay(6);
            }
            else
            {
                taskDelay(2);
            }
        }
    } /* end infinite task loop */
}

/************************************/
/* check for line group disablement */
/************************************/
static int checkgroup (unsigned short line, unsigned short *lstate, unsigned short tracenum)
{
    unsigned short group;

    group = secResGroups [line];
    if ((group) && (resGroupLock [group-1]))
    {
        *lstate = RES_LOCKED;
        if (modrestrace [line])
        {
            logMsg ("%d] Resource Group Lock Detected.\n", tracenum, 0, 0, 0, 0, 0);
        }
        return (TRUE);
    }
    return (FALSE);
}

/*
** Try to bring a Sec module on line
**
*/

/* this routine only called from secdriver */

static void on_line (unsigned short line)
{
    SEC          *pmod;

    pmod = modules [line];
    if (pmod)            /* verify pointer prior to use */
    {
        if (mltrace [line] == 1)
        {
            logMsg ( "Bring Online SEC UPS\n", 0, 0, 0, 0, 0, 0);
        }


        /* Use the loop back function to see if the unit is there and responding */
        if (!readparameters1 (ONERETRY + line, pmod))
            return; /* if no response, get out */
        if (!readparameters2 (ONERETRY + line, pmod))
            return; /* if no response, get out */

        mark_not_valid (pmod);    /* a module that has just been brought online must  */
                                  /* have it's data marked as being not valid.        */
        pmod->is_responding = TRUE;
    }
}


/*                                                          
** Mark all module data elements as not valid               
**                                                          
*/                                                 
/* called from on_line and prepresponses              */        

static void mark_not_valid (SEC  *pmod)
{
    unsigned    short i;

/*                                                        
**       values are initalized in the driver as not valid 
*/
    pmod->st1.data_valid = FALSE;

    for (i = 0; (i < MAXDG1); i++)
    {
        pmod->st1.status [i] = FALSE;
    }

    pmod->st2.data_valid = FALSE;

    for (i = 0; (i < MAXDG2); i++)
    {
        pmod->st2.status [i] = FALSE;
    }

    pmod->st3.data_valid = FALSE;

    for (i = 0; (i < MAXDG3); i++)
    {
        pmod->st3.status [i] = FALSE;
    }

    pmod->st5.data_valid = FALSE;

    for (i = 0; (i < MAXDG5); i++)
    {
        pmod->st5.status [i] = FALSE;
    }
}



/*                                                                  
** Scan a SEC Module -  get inputs and set outputs as required      
**                                                                  
*/
/* this routine only called from secdriver         */

static void scan_module (SEC *pmod, unsigned short line)
{

    pmod->tscanpass++;

    if (mltrace [line] == 2)
    {
        logMsg ( "Scanning SEC UPS.\n", 0, 0, 0, 0, 0, 0);
    }

    if (pmod->st1.was_requested)
    {
        if (readstatusgroup1 (line, pmod) == FALSE)
        {
            return;
        }
    }

    if (pmod->st2.was_requested)
    {
        if (readstatusgroup2 (line, pmod) == FALSE)
        {
            return;
        }
    }

    if (pmod->st3.was_requested)
    {
        if (readstatusgroup3 (line, pmod) == FALSE)
        {
            return;
        }
    }

    if (pmod->st5.was_requested)
    {
        if (readstatusgroup5 (line, pmod) == FALSE)
        {
            return;
        }
    }
    return;
}



/*                                                                                     
**      secaidriver                                                                   
**                                                                                     
**      This routines secdriver purpose is to return the analog input value for       
**      the specified module. For a module value                                       
**      a (-1) will be returned if there is no valid data ready.                       
**
*/
/* this routine not called from within this driver */

int secaidriver (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
        unsigned short signal, float *pvalue, unsigned short conversion,
        unsigned char *resname)
{
    SEC            *pmod;
    unsigned short data_valid = FALSE;
    unsigned short datastatus = FALSE;
    unsigned long  datavalue  = 0;
 
    unsigned short aidataflag;
    unsigned short commandtype;
    unsigned short commandvalid;
    unsigned short commandoffset;
 
    if (link > SECLINES)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AIDRIVER: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }
    if (adapter != 0)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AIDRIVER: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d\n",
                   (int) pointname, link, adapter, card, signal, 0);
         return (-1);
    }

    if ((card != 0) && (card != MPSEUDO))
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AIDRIVER: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    pmod = modules [link];
    if (pmod == NULL)
    {
        if (!allocate_module (link))
        {
            return (-1); /* if not allocated, try. if can't allocate, get out. */
        }
        pmod = modules [link];
    }

    if (card == MPSEUDO)
    {
        switch (signal)
        {
            case 0: /* Were the Module values requested */
                *pvalue = pmod->was_requested;
                break;

            case 1: /* Is the Module responding */
                *pvalue = pmod->is_responding;
                break;

            case 5: /* Output Messages */
                *pvalue = pmod->no_dataout_msgs & 0x7fff;
                break;

            case 6: /* Input Messages */
                *pvalue = pmod->no_datain_msgs & 0x7fff;
                break;

            case 7: /* No Receiver Responses */
                *pvalue = pmod->no_rxnoresp & 0x7fff;
                break;

            case 8: /* No Transmitter Response */
                *pvalue = pmod->no_txnoresp & 0x7fff;
                break;

            case 9: /* Message Reply Errors */
                *pvalue = pmod->no_dataerr & 0x7fff;
                break;

            case 10: /* Message Retries */
                *pvalue = pmod->no_retries & 0x7fff;
                break;

            case 11: /* RX Hardware errors */
                *pvalue = pmod->no_rxhwerrs & 0x7fff;
                break;

            case 12: /* TX Hardware errors */
                *pvalue = pmod->no_txhwerrs & 0x7fff;
                break;

            case 13: /* Sec reported checksum errors */
                *pvalue = pmod->no_chksumerrs & 0x7fff;
                break;

            case 14: /* Offline transitions */
                *pvalue = pmod->no_offlines & 0x7fff;
                break;

            case 15: /* reset statistics */
                *pvalue = pmod->cmd_reset_stat;
                break;

            case 16: /* Force the Module offline */
                *pvalue = pmod->cmd_force_offline;
                break;

	    case 17: /* Failed output commands */
		*pvalue = pmod->badoutcnt & 0x7fff;
		break;

            case 18: /* Sec reported framing errors */
                *pvalue = pmod->no_frameerrs & 0x7fff;
                break;

            default:
                if (trace[link] & CONFIGURATIONTRACE)
                    logMsg("AIDRIVER: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal, 0);
                return (-1);
        }
        /* check for illegal conversion type */
        if (conversion != NOCONV)
        {
            if (trace[link] & CONFIGURATIONTRACE)
                logMsg("AIDRIVER: Illegal Pseudo Point Conversion Type:[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
            return (-1);
        }
        if (mltrace[link] == 8)
            logMsg("AIDRIVER: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, V=0x%x\n",
                    (int) pointname, link, adapter, card, signal, *pvalue);
        return (0);
    }

    /* check for illegal signal: Below Min Value */
    if (signal < MINSIGNAL)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AIDRIVER: Illegal Signal (1) :[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    /* check for illegal signal: Above Max Value */
    if (signal > MAXSIGNAL)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AIDRIVER: Illegal Signal (2) :[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    /* check for illegal signal: Parameter not AI Data Type */
    aidataflag = upstable [signal].aiflag;
    if (aidataflag != TRUE)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AIDRIVER: Illegal Signal (3) :[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }
    /* check for illegal signal: Acquistion Command Unsupported By Driver */
    commandtype = upstable [signal].command;
    commandvalid = validcommands [commandtype];
    
    if (commandvalid != TRUE)
    {
        if (trace[link] & CONFIGURATIONTRACE)
        {
            logMsg("AIDRIVER: Illegal Signal (4) :[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
            logMsg("AIDRIVER: Illegal Signal (4) :[%s] commandtype = %d, commandvalid = %d.\n",
                    (int) pointname, commandtype, commandvalid, 0, 0, 0);
         }
        return (-1);
    }

    /* now that address is validated, register the resource. */
    if (resname != 0)
    { 
        register_resource (link, pmod, resname, pointname);
    }

    pmod->was_requested = TRUE;

    /* check for illegal conversion type */
    if ((conversion != NOCONV) && (conversion != LINEAR) && (conversion != AICNVSQRROOT))
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AIDRIVER: Illegal Conversion Type:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }
    if (!resresponse[link])
    {
        if (trace[link] & CALLREJECTTRACE)
            logMsg("AIDRIVER: Driver Offline(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    if (pmod->holdflag)
    {
        if (trace[link] & CALLREJECTTRACE)
            logMsg("AIDRIVER: On Hold(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-2);
    }

    if (!pmod->is_responding)
    {
        if (trace[link] & CALLREJECTTRACE)
            logMsg("AIDRIVER: Not Responding(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    commandoffset = upstable [signal].offset;
    
    /* is the data valid? */
    switch (commandtype)
    {
        case  ST1:
            pmod->st1.was_requested = TRUE;
            data_valid    = pmod->st1.data_valid;
            datastatus    = pmod->st1.status [commandoffset];
            datavalue     = pmod->st1.data   [commandoffset];
            break;

        case  ST2:
            pmod->st2.was_requested = TRUE;
            data_valid    = pmod->st2.data_valid;
            datastatus    = pmod->st2.status [commandoffset];
            datavalue     = pmod->st2.data   [commandoffset];
            break;

        case  ST3:
            pmod->st3.was_requested = TRUE;
            data_valid    = pmod->st3.data_valid;
            datastatus    = pmod->st3.status [commandoffset];
            datavalue     = pmod->st3.data   [commandoffset];
            break;

        case  ST5:
            pmod->st5.was_requested = TRUE;
            data_valid    = pmod->st5.data_valid;
            datastatus    = pmod->st5.status [commandoffset];
            datavalue     = pmod->st5.data   [commandoffset];
            break;

        case  ATR:
        case  MAN:
        case  MOD:
        case  NOM:
        case  NUL:
        case  PSD:
        case  RWD:
        case  SDA:
/*        case  ST3: */
        case  ST4:
        case  STD:
        case  STR:
        case  TST:
        case  UBR:
        case  UID:
        case  VER:
        default:
            if (trace[link] & CONFIGURATIONTRACE)
                logMsg("AIDRIVER: Illegal Signal (5):[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
            return (-1);
    }

    if (!data_valid)
    {
        if (trace[link] & CALLREJECTTRACE)
            logMsg("AIDRIVER: Data Not Valid(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    if (!datastatus)
    {
        if (trace[link] & CALLREJECTTRACE)
            logMsg("AIDRIVER: Data Not Valid(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    *pvalue = datavalue;

    if (mltrace[link] == 8)
        logMsg("AIDRIVER: Data Returned(1):[%s] L=%d, A=%d, C=%d, S=%d, V=%12.6f\n",
                (int) pointname, link, adapter, card, signal, *pvalue);

    return(0);
}

/*                                                                                
**      secaodriver                                                              
**                                                                                
**      This routines secdriver purpose is to set requested analog outputs to    
**      the specified module.                                                     
*/
/* this routine not called from within this driver */


int secaodriver (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
              unsigned short signal, double dvalue, unsigned short value,
              unsigned short conversion, unsigned char *resname)
{
    SEC         *pmod;

    if (link > SECLINES)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AODRIVER: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                    (int) pointname, link, adapter, card, signal, value);
        return (-1);
    }

    if (adapter >= NUMMODS)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AODRIVER: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                    (int) pointname, link, adapter, card, signal, value);
        return (-1);
    }

    if (card != MPSEUDO)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AODRIVER: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                    (int) pointname, link, adapter, card, signal, value);
        return (-1);
    }

    pmod = modules [link];
    if (pmod == NULL)
    {
        if (!allocate_module (link))
        {
            return (-1); /* if not allocated, try. if can't allocate, get out. */
        }
        pmod = modules [link];
    }
    if (card == MPSEUDO)
    {
        switch (signal)
        {
            case 15: /* Reset the accumulated statistics */
                if (value != 0)
                {
                    initstats(pmod);
                }
                break;

            case 16: /* Force the Module online/offline */
                pmod->cmd_force_offline = (value != 0);
                if (value != 0)
                {
                    pmod->is_responding = FALSE; /* only required when asserting cmd_force_offline */
                }
                break;

            default:
                if (trace[link] & CONFIGURATIONTRACE)
                    logMsg("AODRIVER: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                            (int) pointname, link, adapter, card, signal, value);
            return (-1);
        }
        return (0);
    }
    return (-1);
}


/*                                                                               
**      secaoread                                                               
**                                                                               
**      This routines secdriver purpose is to return the value last readback    
**      from the analog output hardware. This is used to initialize analog       
**      outputs.                                                                 
*/
/* this routine not called from within this driver */


int secaoread (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
      unsigned short signal, unsigned short *pvalue,
      unsigned short conversion, unsigned char *resname)

{
    SEC         *pmod;
    
    if (link > SECLINES)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AOREAD: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    if (adapter >= NUMMODS)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AOREAD: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    if (card != MPSEUDO)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("AOREAD: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    pmod = modules [link];
    if (pmod == NULL)
    {
        if (!allocate_module (link))
        {
                return (-1); /* if not allocated, try. if can't allocate, get out. */
        }
        pmod = modules [link];
    }

    if (card == MPSEUDO)
    {
        switch (signal)
        {
            case 15: /* read statistics command value */
                *pvalue = pmod->cmd_reset_stat;
                break;

            case 16: /* Read the current Module offline command status */
                *pvalue = pmod->cmd_force_offline;
                break;

            default:
                if (trace[link] & CONFIGURATIONTRACE)
                    logMsg("AOREAD: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal, 0);
                return(-1);
        }
        return (0);
    }
    return(-1);
}


/*                                                                                    
**      secbidriver                                                                  
**                                                                                    
**      The secdriver purpose of this routine is to return the request digital       
**      input value(s) from the requested card. This routine will return a (-1) if    
**      the requested module is the wrong type or the data is not yet valid.          
**                                                                                    
*/
/* this routine not called from within this driver */

int secbidriver (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
        unsigned short signal, unsigned long mask, unsigned long *pvalue,
        unsigned char *resname)
{
    SEC          *pmod;
    unsigned short data_valid = FALSE;
    unsigned short datastatus = FALSE;
    unsigned long  datavalue  = 0;
 
    unsigned short bidataflag;
    unsigned short commandtype;
    unsigned short commandvalid;
    unsigned short commandoffset;

    if (link > SECLINES)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BIDRIVER: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    if (adapter != 0)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BIDRIVER: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    if ((card != 0) && (card != MPSEUDO))
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BIDRIVER: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    pmod = modules [link];
    if (pmod == NULL)
    {
        if (!allocate_module (link))
        {
            return (-1); /* if not allocated, try. if can't allocate, get out. */
        }
        pmod = modules [link];
    }

    if (card == MPSEUDO)
    {
        switch (signal)
        {
            case 0: /* Were the Module values requested */
                *pvalue = pmod->was_requested;
                break;

            case 1: /* Is the Module responding */
                *pvalue = pmod->is_responding;
                break;

            case 15: /* reset statistics */
                *pvalue = pmod->cmd_reset_stat;
                break;

            case 16: /* Force the Module offline */
                *pvalue = pmod->cmd_force_offline;
                break;

            default:
                if (trace[link] & CONFIGURATIONTRACE)
                    logMsg("BIDRIVER: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal, 0);
                return (-1);
        }
        if (mltrace[link] == 9)
            logMsg("BIDRIVER: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, V=0x%x\n",
                    (int) pointname, link, adapter, card, signal, *pvalue);
        return (0);
    }

    /* check for illegal signal: Below Min Value */
    if (signal < MINSIGNAL)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BIDRIVER: Illegal Signal (1) :[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    /* check for illegal signal: Above Max Value */
    if (signal > MAXSIGNAL)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BIDRIVER: Illegal Signal (2) :[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    /* check for illegal signal: Parameter not BI Data Type */
    bidataflag = upstable [signal].biflag;
    if (bidataflag != TRUE)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BIDRIVER: Illegal Signal (3) :[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }
    /* check for illegal signal: Acquistion Command Unsupported By Driver */
    commandtype = upstable [signal].command;
    commandvalid = validcommands [commandtype];
    
    if (commandvalid != TRUE)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BIDRIVER: Illegal Signal (4) :[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    /* now that address is validated, register the resource. */
    if (resname != 0)
    { 
        register_resource (link, pmod, resname, pointname);
    }

    pmod->was_requested = TRUE;

    if (!resresponse[link])
    {
        if (trace[link] & CALLREJECTTRACE)
            logMsg("BIDRIVER: Driver Offline(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    if (pmod->holdflag)
    {
        if (trace[link] & CALLREJECTTRACE)
            logMsg("BIDRIVER: On Hold(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-2);
    }

    if (!pmod->is_responding)
    {
        if (trace[link] & CALLREJECTTRACE)
            logMsg("BIDRIVER: Not Responding(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    commandoffset = upstable [signal].offset;
    
    /* is the data valid? */
    switch (commandtype)
    {
        case  ST1:
            pmod->st1.was_requested = TRUE;
            data_valid    = pmod->st1.data_valid;
            datastatus    = pmod->st1.status [commandoffset];
            datavalue     = pmod->st1.data   [commandoffset];
            break;

        case  ST2:
            pmod->st2.was_requested = TRUE;
            data_valid    = pmod->st2.data_valid;
            datastatus    = pmod->st2.status [commandoffset];
            datavalue     = pmod->st2.data   [commandoffset];
            break;

        case  ST3:
            pmod->st3.was_requested = TRUE;
            data_valid    = pmod->st3.data_valid;
            datastatus    = pmod->st3.status [commandoffset];
            datavalue     = pmod->st3.data   [commandoffset];
            break;

        case  ST5:
            pmod->st5.was_requested = TRUE;
            data_valid    = pmod->st5.data_valid;
            datastatus    = pmod->st5.status [commandoffset];
            datavalue     = pmod->st5.data   [commandoffset];
            break;

        case  ATR:
        case  MAN:
        case  MOD:
        case  NOM:
        case  NUL:
        case  PSD:
        case  RWD:
        case  SDA:
/*        case  ST3: */
        case  ST4:
        case  STD:
        case  STR:
        case  TST:
        case  UBR:
        case  UID:
        case  VER:
        default:
            if (trace[link] & CONFIGURATIONTRACE)
                logMsg("BIDRIVER: Illegal Signal (5):[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
            return (-1);
    }

    if (!data_valid)
    {
        if (trace[link] & CALLREJECTTRACE)
            logMsg("BIDRIVER: Data Not Valid(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    if (!datastatus)
    {
        if (trace[link] & CALLREJECTTRACE)
            logMsg("BIDRIVER: Data Not Valid(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    *pvalue = datavalue;

    if (mltrace[link] == 8)
        logMsg("BIDRIVER: Data Returned(1):[%s] L=%d, A=%d, C=%d, S=%d, V=%12.6f\n",
                (int) pointname, link, adapter, card, signal, *pvalue);

    return(0);
}


/*                                                                                     
**      secbodriver                                                                
**                                                                                     
**      The secdriver purpose of this routine is to set the request digital        
**      output value(s) on the requested card. This routine will return a (-1) if      
**      the requested module is the wrong type or not configured yet.                  
**                                                                                     
*/
/* this routine not called from within this driver */

int secbodriver (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
        unsigned short signal, unsigned long value, unsigned long mask,
        unsigned char *resname)
{
    SEC         *pmod;

    if (link > SECLINES)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BODRIVER: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                    (int) pointname, link, adapter, card, signal, value);
        return (-1);
    }

    if (adapter >= NUMMODS)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BODRIVER: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                    (int) pointname, link, adapter, card, signal, value);
        return (-1);
    }

    if (card != MPSEUDO)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BODRIVER: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                    (int) pointname, link, adapter, card, signal, value);
            return (-1);
    }

    pmod = modules [link];
    if (pmod == NULL)
    {
        if (!allocate_module (link))
        {
            return(-1); /* if not allocated, try. if can't allocate, get out. */
        }
        pmod = modules [link];
    }

    if (card == MPSEUDO)
    {
        switch (signal & 0x1f)
        {
            case 15: /* Reset the accumulated statistics */
                if (value != 0)
                {
                    initstats (pmod);
                }
                break;

            case 16: /* Force the Module offline */
                pmod->cmd_force_offline = (value != 0);
                if (value != 0)
                {
                    pmod->is_responding = FALSE;/* only required when asserting cmd_force_offline */
                }
                break;

            default:
                if (trace[link] & CONFIGURATIONTRACE)
                    logMsg("BODRIVER: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                            (int) pointname, link, adapter, card, signal, value);
                return (-1);
        }
        return (0);
    }
    /* this point should never be reached. */
    return (-1);
}

/*                                                                              
**      secboread                                                           
**                                                                              
**      This routines secdriver purpose is to return the value last read    
**      back from the digital output hardware. This is used to initialize       
**      digital outputs. This will return a (-1) if the module is the wrong     
**      type or the data is not yet valid                                       
*/
/* this routine not called from within this driver */


int secboread (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
        unsigned short signal, unsigned long *pvalue, unsigned long mask,
        unsigned char *resname)

{
    SEC          *pmod;

    if (link > SECLINES)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BOREAD: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }
    if (adapter >= NUMMODS)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BOREAD: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
        return (-1);
    }
    if (card != MPSEUDO)
    {
        if (trace[link] & CONFIGURATIONTRACE)
            logMsg("BOREAD: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    (int) pointname, link, adapter, card, signal, 0);
         return (-1);
    }

    pmod = modules [link];
    if (pmod == NULL)
    {
        if (!allocate_module (link))
        {
            return(-1); /* if not allocated, try. if can't allocate, get out. */
        }
        pmod = modules [link];
    }

    if (card == MPSEUDO)
    {
        switch (signal)
        {
            case 15: /* reset statistics */
                *pvalue = pmod->cmd_reset_stat;
                break;

            case 16: /* Force the Module offline */
                *pvalue = pmod->cmd_force_offline;
                break;

            default:
                if (trace[link] & CONFIGURATIONTRACE)
                    logMsg("BOREAD: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal, 0);
                return (-1);
        }
        return (0);
    }
    /* this point should never be reached. */
    return (-1);
}
/*
** register_resource
**
**    This routine is used to initialize the resource pointers
**  If this is the first time the resource for this line is being
**  referenced the pointer to the resource table entry will be set
**  and the resource name will be saved. If the resource name has
**  already been set, this routine will verify that all other points
**  for this driver are using the same resource name.
**
*/                                                                                        

static void  register_resource (unsigned short link, SEC *pmod,
                                unsigned char *resname, char *pointname)
{
    int group;

    if (resnames [link] [0] == 0) /* This is the first reference */
    {
        if (resGetByName (resname, &prestable [link]) != RES_OK)
        {
        logMsg ("Resource not found in table.(Point Name = %s, Resource Name = %s)\n",
                 (int) pointname, (int) resname, 0, 0, 0, 0);
        }
        else
        {
            strcpy (resnames [link], resname);
            group = secResGroups [link];
            if (group)
            {
                if (assignLockGroup (group, resname) == ERROR)
                {
                    logMsg ("assignLockGroup error: Line = %d, Group = %d, Point Name = %s, Resource Name = %s\n",
                             link, group, (int) pointname, (int) resname, 0, 0);
                }
            }
        }
    }
    else
    {
        if (strcmp (resnames [link], resname) != 0)
        {
        logMsg ("Resource name[%s] conflict.(Point Name = %s, Resource Name = %s)\n",
                (int) resnames[link], (int) pointname, (int) resname, 0, 0, 0);
        }
    }
}


/*                                                                              
**                                                                              
**    allocate_module                                                           
**                                                                              
**    The purpose of this routine is to allocate the memory for a single        
**    module data structure. If not successful, a console error message is      
**    printed and an error flag is returned to the caller.                      
**    If successful, the newly allocated data structures are initialized.       
**                                                                              
*/

static int allocate_module (unsigned short line)
{
    modules [line] = (SEC *) calloc (1, (sizeof (SEC)));
    if (!modules [line])
    {
        logMsg ("Unable to allocate memory for SEC data structure.\n", 0, 0, 0, 0, 0, 0);
        return(FALSE);
    }
    else
    {
        logMsg ("Allocated memory for SEC data structure.\n", 0, 0, 0, 0, 0, 0);
        init_module_tab (line);         /*    initialize the data structures. */
    }
    return (TRUE);
}



/*                                                                          
**      start_rx                                                            
**                                                                          
**      The purpose of this routine is to initialize a line table           
**      for receive and to enable the receiver.                             
**                                                                          
*/
/* only called from msgxfer */

static  void start_rx (unsigned short line, SEMAPHORE *semid, unsigned short *pstatus)
{
    RXLINE        *rxp;

    rxp = rxlinetab [line];
    rxp->rxix = 0;
    rxp->semid = semid;
    rxp->status = pstatus;
    rxp->rxstate = RXTOPHAT; 
}

/*                                                                               
**      stop_rx                                                                  
**                                                                               
**      The purpose of this routine is to initialize a line table                
**      for receive and to disable the receiver.                                 
**                                                                               
*/
/* only called from msgxfer */

static  void stop_rx (unsigned short line )

{
    RXLINE        *rxp;

    rxp = rxlinetab [line];
    rxp->rxstate = RXIGNORE;
}


/*                                                                                 
**      NEW rx_int_proc                                                            
**                                                                                 
**      The purpose of this routine is to process a receive character interrupt    
**      for a given line.                                                          
*/
/* referenced in system call in secdriver. indirectly called through actions in msgxfer. */

static  void rx_int_proc (unsigned short line)
{
    unsigned char   rxchar;
    unsigned short  rxstat;
    unsigned short  rx_err_mask;
    RXLINE         *rxp;

    read_rx_hw (link[line], &rxchar, &rxstat, &rx_err_mask);
    /* only execute if no comm error */
    if ((rxstat & rx_err_mask) == 0)
    {
        /* Something is sending in data. */
        line_activity [line] = TRUE;
        linestats [line].RXintcount++;
    }
    else
    {
        linestats [line].RXbadcount++;
    }

    rxp = rxlinetab [line];

    if (rxp->rxstate == RXIGNORE)
    {
        return;
    }
    /* no matter what the state, errors cause a logical reset of the message processing */
    if ((rxstat & rx_err_mask) != 0)
    {
        rxp->rxstate = RXTOPHAT;   /* if comm goes bad, assume it was garbage,  */
                                   /* and go wait for the good stuff.           */ 
        rxp->rxix = 0;             /* reset the buffer index/counter            */
        return;
    }

    rxp->rxbuf [rxp->rxix++] = rxchar; /* store the character in the buffer and bump the index. */
    switch (rxp->rxstate)
    {
        case RXTOPHAT:           /* Waiting for Top Hat (caret) character to begin response   */
            if (rxchar == '^')
            {
                rxp->rxstate = RXMESSAGETYPE;
                return;
            }
            else
            {
                rxp->rxstate = RXTOPHAT;               /* presume it was a false input.  */
                                                       /* go wait for the real one.      */ 
                rxp->rxix = 0;                         /* reset the buffer index/counter.*/
                return;
            }
            break;

        case RXMESSAGETYPE:      /* Waiting for valid message type */
            switch (rxchar)
            {
                /* these messages have no data length or data bytes */
                case '0':
                case '1':
                    *(rxp->status) = RXGOOD;
                    rxp->rxstate = RXIGNORE;
                    semGive (rxp->semid);
                    return;
                    break;

                /* these messages have a three character data length specified */
                case 'D':
                case '*':
                case 'P':
                case 'S':
                    rxp->rxstate = RXDATALENGTH1;
                    rxp->datalength = 0; 
                    return;
                    break;

                default:
                    rxp->rxstate = RXTOPHAT;               /* presume it was a false input.  */
                                                           /* go wait for the real one.      */ 
                    rxp->rxix = 0;                         /* reset the buffer index/counter.*/
                    return;
                    break;
            }
            break;

        case RXDATALENGTH1:      /* Waiting for first digit of length */
            switch (rxchar)
            {
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    rxp->datalength = rxchar - '0';
                    rxp->rxstate = RXDATALENGTH2;
                      return;
                    break;

                default:
                    rxp->rxstate = RXTOPHAT;               /* presume it was a false input.  */
                                                           /* go wait for the real one.      */ 
                    rxp->rxix = 0;                         /* reset the buffer index/counter.*/
                    return;
                    break;
            }
            break;

        case RXDATALENGTH2:      /* Waiting for second digit of length */
            switch (rxchar)
            {
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    rxp->datalength = (rxp->datalength * 10) + rxchar - '0';
                    rxp->rxstate = RXDATALENGTH3;
                    return;
                    break;

                default:
                    rxp->rxstate = RXTOPHAT;               /* presume it was a false input.  */
                                                           /* go wait for the real one.      */ 
                    rxp->rxix = 0;                         /* reset the buffer index/counter.*/
                    return;
                    break;
            }

        case RXDATALENGTH3:      /* Waiting for third digit of length */
            switch (rxchar)
            {
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    rxp->datalength = (rxp->datalength * 10) + rxchar - '0';
                    rxp->rxstate = RXDATA;
                    return;
                    break;

                default:
                    rxp->rxstate = RXTOPHAT;               /* presume it was a false input.  */
                                                           /* go wait for the real one.      */ 
                    rxp->rxix = 0;                         /* reset the buffer index/counter.*/
                    return;
                    break;
            }
            break;

        case RXDATA:                /* Waiting for message data */

            /* if datalength characters have come in, the message is finished. */
            if (rxp->rxix >= (rxp->datalength + 5))
            {
                *(rxp->status) = RXGOOD;
                rxp->rxstate = RXIGNORE;
                semGive (rxp->semid);
                return;
            }
            if (rxp->rxix >= MAXRXMSG)       /* don't overflow the buffer.     */
            {
                *(rxp->status) = RXBAD;
                rxp->rxstate = RXIGNORE;
                semGive (rxp->semid);
                return;
            }
            break;
    }
    return;
}


/*                                                                             
**      check_line_activity                                                    
**                                                                             
*/
/* only called from secdriver  */

static  int check_line_activity (unsigned short line, unsigned short predelay )
{
    RXLINE      *rxp;
    long         thistime;
    short        status;

    if (predelay != 0)
    {
        taskDelay (predelay);
    }
    thistime = tickGet ();
    if ((thistime - activity_time [line]) < ACTIVITYTIME)
    {
        return (lastactivity [line] || (line_activity [line]));
    }
    lastactivity [line] = status = line_activity [line];
    line_activity [line] = FALSE;
    activity_time [line] = thistime;
    rxp = rxlinetab [line];
    rxp->rxstate = RXIGNORE;

    return (status);
}


/*                                                                                 
**      start_tx                                                                   
**                                                                                 
**      The purpose of this routine is to initialize a line table                  
**      for send and to enable the transmitter.                                    
**                                                                                 
*/
/* only called from msgxfer  */

static  void start_tx (unsigned short line, SEMAPHORE *semid, unsigned short *pstatus )
{
    TXLINE        *txp;

    disable_tx_ints (link[line]);
    txp = txlinetab [line];
    txp->txix = 0;
    txp->semid = semid;
    txp->status = pstatus;
    txp->txdisable = FALSE;

    enable_tx_ints (link[line]);
}


/*                                                                                 
**      stop_tx                                                                    
**                                                                                 
**      The purpose of this routine is to initialize a line table                  
**      for receive and to disable the receiver.                                   
**                                                                                 
*/
/* only called from msgxfer  */

static  void stop_tx (unsigned short line )
{
    TXLINE        *txp;

    disable_tx_ints (link[line]);
    txp = txlinetab [line];
    txp->txdisable = TRUE;

    /* the 485 interface uses RTS to control direction of line transmission. */
    clear_rts (link[line]);
}


/*                                                                                
**      tx_int_proc                                                               
**                                                                                
**      The purpose of this routine is to process a transmit character interrupt  
**      for a given line.                                                         
**                                                                                
*/
/* only called from secdriver  */

static  void tx_int_proc (unsigned short line )
{
    char            txchar;
    unsigned short  txstat;
    unsigned short  tx_err_mask;
    TXLINE         *txp;

    linestats [line].TXintcount++;

    txp = txlinetab [line];

    if (txp->txdisable)
    {
        return;
    }

    if (txp->txix < txp->txmsglen)
    {
        txchar = txp->txbuf [txp->txix++];
        write_tx_hw (link[line], txchar, &txstat, &tx_err_mask);
        if ((txstat & tx_err_mask) != 0 )
        {
            txp->txdisable = TRUE;
            disable_tx_ints (link[line]);
            *(txp->status) = TXBAD;
            semGive (txp->semid);
            return;
        }
        if (txp->txix >= MAXTXMSG)
        {
            txp->txdisable = TRUE;
            disable_tx_ints (link[line]);
            *(txp->status) = TXBAD;
            semGive (txp->semid);
            return;
        }
        return;
    }

    if (txp->txix == txp->txmsglen)
    {
        /* the 485 interface uses RTS to control direction of line transmission.        */
        /* we transmit an extra character (NULL) in order to provide time for the last  */
        /* 'real' character to be transmitted before completing tranmission interrupt   */
        /* servicing and reversing the RTS line in order to receive the reply.          */

        txchar = 0;
        txp->txix++;
        write_tx_hw (link[line], txchar, &txstat, &tx_err_mask);
        if ((txstat & tx_err_mask) != 0 )
        {
            txp->txdisable = TRUE;
            disable_tx_ints (link[line]);
            *(txp->status) = TXBAD;
            semGive (txp->semid);
        }
        return;
    }

    if (txp->txix > txp->txmsglen)
    {
        txp->txdisable = TRUE;
        disable_tx_ints (link[line]);
        *(txp->status) = TXGOOD;
        semGive (txp->semid);
        return;
    }
}


/*                                                                               
**      msgxfer                                                                  
**                                                                               
**      The purpose of this routine is to transmit a buffer to the               
**      Sec UPS and to await a response. The transmit and receive buffers        
**      are in their respective line tables. This routine will retry the         
**      message up to a maximum number of times or until a response is received  
**      with a proper checksum.                                                  
**              returns         XFERGOOD if response received                    
**                              XFERBAD  if no good response is received         
**                                                                               
*/

static  unsigned short  msgxfer (unsigned short fline, SEC *pmod, unsigned short responselen)
{
    unsigned short  retry, txstatus, rxstatus, txlen, rxlen, line, ix;
    RXLINE         *rxp;
    TXLINE         *txp;

    unsigned char   buffer[200];
    unsigned char   buffet[200];

    unsigned short  oldrxix;
    unsigned int    delta40, delta01, delta12, delta23, delta34;
    unsigned long   msgstart;


    msgstart = tickGet();
    line = (fline & (~ONERETRY));          /* strip off the no retry flag*/
    delta40 = delta01 = delta12 = delta23 = delta34 = 0;

    rxp = rxlinetab [line];
    txp = txlinetab [line];

    for (retry = ((fline == line) ? 0 : MAXRETRYS - 1); (retry < MAXRETRYS); retry++)
    {
        linestats [line].msgxfercount++;
        loopmsgs [line]++;

        /* time since end of last msgxfer */
        tick [line] [0] = tickGet ();
        delta40 = (tick [line] [0] - tick [line] [4]);

        if (intermsgdelay [line] != 0)
        {
            taskDelay (intermsgdelay [line]);
        }

        txstatus = IO_TIMEOUT;
        rxstatus = IO_TIMEOUT;

        semTake (semids [line] [TXSEM], NO_WAIT);
        semTake (semids [line] [RXSEM], NO_WAIT);

        txlen = txp->txmsglen;

        if ((trace [line] & ALLTRAFFICTRACE) &&
            ((traceadr [line] == txp->txbuf [0]) || (traceadr [line] == 255)) &&
            ((tracetype [line] == txp->txbuf [1]) || (tracetype [line] == 255)))
        {
            sprintf (buffer, "TX:");
            if (trace [line] & ASCIITRAFFICTRACE)
            {
                for (ix = 0; ix < txlen ; ix++) /* just print what is there to print */
                {
                    if (!((txp->txbuf [ix]) < 0x20))
                    {
                        sprintf(buffet, "%c", txp->txbuf [ix]);
                        strcat(buffer, buffet);
                    }
                    else
                    {
                        sprintf(buffet, "<%x>", txp->txbuf [ix]);
                        strcat(buffer, buffet);
                    }
                }
            }
            else
            {
                for (ix = 0; ix < 12 ; ix++) /* 12 is maximum TX message length */
                {
                    if (ix < txlen)
                    {
                        sprintf(buffet, "_%2x", txp->txbuf [ix]);
                        strcat(buffer, buffet);
                    }
                    else
                    {
                        strcat(buffer, "...");
                    }
                }
            }
            logMsg( "%s\n",  (int) buffer, 0, 0, 0, 0, 0);
        }

       set_rts (link[line]);

        if (pretxdelay [line] != 0)
        {
            taskDelay (pretxdelay [line]);
        }

        /* time from entering msgxfer to starting transmission. (including delays & diags) */
        tick [line] [1] = tickGet ();
        delta01 = (tick [line] [1] - tick [line] [0]);

        start_rx (line, semids [line] [RXSEM], &rxstatus);
        start_tx (line, semids [line] [TXSEM], &txstatus); /* this fires off the transmission. */

        if ((semTake (semids [line] [TXSEM], TXTIMEOUT)) == ERROR)
        {
            txstatus = IO_TIMEOUT;
        }

        stop_tx (line);

        /* time to transmit message. */
        tick [line] [2] = tickGet ();
        delta12 = (tick [line] [2] - tick [line] [1]);

        if (posttxdelay [line] != 0)
        {
            taskDelay (posttxdelay [line]);
        }

        if (halfduplex [line] == TRUE)
        {
            clear_rts (link[line]);
        }

        /* intermessage wait time. */
        tick [line] [3] = tickGet ();
        delta23 = (tick [line] [3] - tick [line] [2]);

        if (txstatus == TXGOOD)
        {
            oldrxix = -1;    /* make it pass through timeout loop at least once */
            while ((semTake (semids [line] [RXSEM], (int) (RXTIMEOUT))) == ERROR)
            {
                if (oldrxix == rxp->rxix)
                {
                    if (modtrace [line])
                    {
                        logMsg("RX timeout\n", 0, 0, 0, 0, 0, 0);
                    }
                    rxstatus = IO_TIMEOUT;
                    break;
                }
                else
                {
                    if ((modtrace [line]) || (mltrace[line] == 12))
                    {
                         logMsg("waiting...", 0, 0, 0, 0, 0, 0);
                    }
                }
                oldrxix = rxp->rxix;
            }
            stop_rx (line);

            /* time to receive message. */
            tick [line] [4] = tickGet();
            delta34 = (tick [line] [4] - tick [line] [3]);

            rxp->rxmsglen = rxp->rxix;
            rxlen = rxp->rxix; /* index is just past last received character. length EQUALS index */

            if (rxstatus == RXGOOD)
            {
                if ((trace [line] & ALLTRAFFICTRACE) &&
                    ((traceadr [line] == txp->txbuf [0]) || (traceadr [line] == 255)) &&
                    ((tracetype [line] == txp->txbuf [1]) || (tracetype [line] == 255)))
                {
                    sprintf(buffer, "RX:");
                    if (trace [line] & ASCIITRAFFICTRACE)
                    {
                        for (ix = 0; ix < rxlen ; ix++) /* just print what is there to print */
                        {
                            if (((ix-2)%70 == 0) && (ix != 2))
                            {
                                logMsg( "%s\n", *buffer, 0, 0, 0, 0, 0); /* avoid line wrap */
                                sprintf(buffer, "         ");
                            }

                            if (!((rxp->rxbuf [ix]) < 0x20))
                            {
                                sprintf(buffet, "%c", rxp->rxbuf [ix]);
                                strcat(buffer, buffet);
                            }
                            else
                            {
                                sprintf(buffet, "<%x>", rxp->rxbuf [ix]);
                                strcat(buffer, buffet);
                            }
                        }

                    }
                    else
                    {
                        for (ix = 0; ix < 20 ; ix++) /* 20 is maximum RX message length */
                        {
                            if (ix < rxlen)
                            {
                                sprintf(buffet, "_%2x", rxp->rxbuf [ix]);
                                strcat(buffer, buffet);
			    }
                            else
			    {
                                strcat(buffer, "...");
                            }
                        }
                    }
                    logMsg( "%s\n",  (int) buffer, 0, 0, 0, 0, 0);
                    logMsg(" ............. DELTA: %3d %3d %3d %3d %3d\n",
                            delta40, delta01, delta12, delta23, delta34, 0);
                }
                /**********************************************/
                /**********************************************/
                pmod->msgs++;
                pmod->responsetime += (tickGet() - msgstart);
                return (XFERGOOD);
            } /* bottom of rxstatus == RXGOOD */
            else
            {
                if (rxstatus == IO_TIMEOUT)
                {
                    if (modtrace [line])
                    {
                        logMsg ("Receive Timeout... Length: %d\n", rxlen, 0, 0, 0, 0, 0);
                    }
                    if ((trace [line] & ALLTRAFFICTRACE) &&
                        ((traceadr [line] == txp->txbuf [0]) || (traceadr [line] == 255)) &&
                        ((tracetype [line] == txp->txbuf [1]) || (tracetype [line] == 255)))
                    {
                        logMsg ("DELTA: %3d %3d %3d %3d %3d\n",
                                delta40, delta01, delta12, delta23, delta34, 0);
                    }
                    pmod->no_rxnoresp++;
                }
                else
                {
                    if (modtrace [line])
                    {
                        logMsg ("Receive Error:   RXHWERR: %2x, RXCHR: %2x, RXCNT: %2x\n",
                         rxp->rxbuf[100],
                         rxp->rxbuf[101],
                         rxp->rxix, 0, 0, 0);
                    }
                    pmod->no_rxhwerrs++;
                }
            }
        }
        else
        {
            if (txstatus == IO_TIMEOUT)
            {
                if (modtrace [line])
                {
                    logMsg ("Transmit Timeout\n", 0, 0, 0, 0, 0, 0);
                }
                if ((trace [line] & ALLTRAFFICTRACE) &&
                    ((traceadr [line] == txp->txbuf [0]) || (traceadr [line] == 255)) &&
                    ((tracetype [line] == txp->txbuf [1]) || (tracetype [line] == 255)))
                {
                    logMsg("DELTA: %3d %3d %3d %3d %3d\n",
                            delta40, delta01, delta12, delta23, delta34, 0);
                }
                pmod->no_txnoresp++;
            }
            else
            {
                if (modtrace [line])
                {
                    logMsg ("Transmit Error\n", 0, 0, 0, 0, 0, 0);
                }
                pmod->no_txhwerrs++;
            }
        }
        /* bottom of retry loop */
        if (retry != MAXRETRYS)
            pmod->no_retries++;
    }
    if (modtrace [line])
    {
        logMsg ("Xfer Bad\n", 0, 0, 0, 0, 0, 0);
    }
    return (XFERBAD);
}



/*
**      init_module_tab
**
**      The purpose of this routine is to initialize the line tables for the
**      analog devices I/O
**
*/
/* only called from allocate_module  */


static  void init_module_tab (unsigned short line)
{
    SEC  *pmod;

    pmod = modules [line];
    if (!pmod)
    {
        return;
    }
    pmod->is_responding = FALSE;
    pmod->holdflag = FALSE;
    pmod->was_requested = FALSE;
    pmod->cmd_force_offline = FALSE;

    initmodule (pmod);
    initstats (pmod);
}

/*
** Cold Init a Module
**
*/
/* only called from init_line_tabs  */


static void initmodule (SEC  *pmod)
{
    unsigned short    i, j, k;
                                                                    
/*                                                                  
**      All data are initialized as not requested & data not valid. 
**                                                                  
*/
    pmod->st1.was_requested = FALSE;
    pmod->st1.data_valid = FALSE;
    for (i = 0; (i < MAXDG1); i++)
    {
        /* find the corresponding signal number from the upstable. */
        pmod->st1.number [i] = 0;
        for (k = 0; k < MAXPARAMETERCOUNT; k++)
        {
            if ((upstable [k].command == ST1) && (upstable [k].offset == i))
            {
                pmod->st1.number [i] = k;
                break;
            }
        }
	/* initialize data parameters */
        pmod->st1.status [i] = FALSE;
        pmod->st1.data [i] = FALSE;
        for (j = 0; j < 7; j++)
        {
            pmod->st1.raw_read [i][j] = 'U';
        }
        pmod->st1.raw_read [i][j] = 0;
    }

    pmod->st2.was_requested = FALSE;
    pmod->st2.data_valid = FALSE;
    for (i = 0; (i < MAXDG2); i++)
    {
        /* find the corresponding signal number from the upstable. */
        pmod->st2.number [i] = 0;
        for (k = 0; k < MAXPARAMETERCOUNT; k++)
        {
            if ((upstable [k].command == ST2) && (upstable [k].offset == i))
            {
                pmod->st2.number [i] = k;
                break;
            }
        }
	/* initialize data parameters */
        pmod->st2.status [i] = FALSE;
        pmod->st2.data [i] = FALSE;
        for (j = 0; j < 7; j++)
        {
            pmod->st2.raw_read [i][j] = 'U';
        }
        pmod->st2.raw_read [i][j] = 0;
    }

    pmod->st3.was_requested = FALSE;
    pmod->st3.data_valid = FALSE;
    for (i = 0; (i < MAXDG3); i++)
    {
        /* find the corresponding signal number from the upstable. */
        pmod->st3.number [i] = 0;
        for (k = 0; k < MAXPARAMETERCOUNT; k++)
        {
            if ((upstable [k].command == ST3) && (upstable [k].offset == i))
            {
                pmod->st3.number [i] = k;
                break;
            }
        }
	/* initialize data parameters */
        pmod->st3.status [i] = FALSE;
        pmod->st3.data [i] = FALSE;
        for (j = 0; j < 7; j++)
        {
            pmod->st3.raw_read [i][j] = 'U';
        }
        pmod->st3.raw_read [i][j] = 0;
    }

    pmod->st5.was_requested = FALSE;
    pmod->st5.data_valid = FALSE;
    for (i = 0; (i < MAXDG5); i++)
    {
        /* find the corresponding signal number from the upstable. */
        pmod->st5.number [i] = 0;
        for (k = 0; k < (MAXPARAMETERCOUNT); k++)
        {
            if ((upstable [k].command == ST5) && (upstable [k].offset == i))
            {
                pmod->st5.number [i] = k;
                break;
            }
        }
	/* initialize data parameters */
        pmod->st5.status [i] = FALSE;
        pmod->st5.data [i] = FALSE;
        for (j = 0; j < 7; j++)
        {
            pmod->st5.raw_read [i][j] = 'U';
        }
        pmod->st5.raw_read [i][j] = 0;
    }
}


/*                                                                    
**                                                                    
** initialize the statistics for a Sec Module                         
**                                                                    
*/                                                                   
/* called from AODRIVER, BODRIVER, and init_line_tabs  */

static  void initstats (SEC *pmod)
{
    pmod->no_dataout_msgs = 0;  /* 5  Count Output Msgs               */
    pmod->no_datain_msgs = 0;   /* 6  Count Data Requests             */
    pmod->no_rxnoresp = 0;      /* 7  Count No Responses              */
    pmod->no_txnoresp = 0;      /* 8  Count No Responses              */
    pmod->no_dataerr = 0;       /* 9  Count Message Reply Errors      */
    pmod->no_retries = 0;       /* 10 Count Retries                   */              
    pmod->no_rxhwerrs = 0;      /* 11 Count Comm Hardware Errors      */
    pmod->no_txhwerrs = 0;      /* 12 Count Comm Hardware Errors      */
    pmod->no_chksumerrs = 0;    /* 13 Count Checksum Errors           */
    pmod->no_offlines = 0;      /* 14 Count Off Line Transitions      */
    pmod->no_frameerrs = 0;     /*    Count Framing Errors            */

    pmod->tscanpass = 0;        /* module scan pass counter           */
    pmod->msgs = 0;             /* good response message counts       */
    pmod->responsetime = 0;     /* total wait time for good messages  */
}


static  short getvalue (unsigned char  *pbuffer,
                        unsigned short *value, 
                        unsigned short *length)
{
    unsigned short  status, i, maxparselength;

    maxparselength = *length;
    *value = 0;
    *length = 0;
    status = FALSE;

    for (i = 0; i < maxparselength; i++)
    {
        *length = *length + 1;
        if (*pbuffer == ',')
        {
            pbuffer++;
            return (status);
        }
        if ((*pbuffer >= '0') && (*pbuffer <='9'))
        {
            /* valid numeric character */
            status = TRUE;
            *value = ((*value) * 10) + (*pbuffer - '0');
/*
printf ("i=%d, value=%d\n", i, *value);
*/
            pbuffer++;
        }
        else
        {
            /* illegal character */
            pbuffer++;
            status = FALSE;
            return (status);
        }
    }
    return (status);
}




/*                                                              
** readparameters1                                              
**                                                              
*/                                                                              
/* only called from on_line */


static  unsigned short readparameters1 (unsigned short line, SEC *pmod)
{
    TXLINE         *txp;
    RXLINE         *rxp;
      signed short  count;
    unsigned short  i, tempi, value, length, status;

    /* Set up pointers for comm buffers */
    txp = txlinetab [line & (~ONERETRY)];
    rxp = rxlinetab [line & (~ONERETRY)];

    /* Format the output message */
    txp->txbuf [0] = '^';
    txp->txbuf [1] = 'P';
    txp->txbuf [2] = '0';
    txp->txbuf [3] = '0';
    txp->txbuf [4] = '3';
    txp->txbuf [5] = 'A';
    txp->txbuf [6] = 'P';
    txp->txbuf [7] = '1';
    txp->txmsglen = 8;
    pmod->no_datain_msgs++;

    /* Send out the message */
    if (msgxfer (line, pmod, MAXDATALENGTH) == XFERBAD)
    {
        pmod->is_responding = FALSE;
        return (FALSE);
    }

    /* now that communications are completed, strip off the retry bit */
    line = line & (~ONERETRY);

    /* expecting a Data Returned response. if anything else, its a mistake */
    switch (rxp->rxbuf [1])
    {
        case 'D':
            /* so far so good. */
            break;

        case '0':
            /* most likely an unsuppported command has been sent,  */
            /*  although this one is supposedly always valid.      */
            pmod->no_dataerr++;
            return (FALSE);
            break;

        case '1':
            /* Yikes! The UPS obviously it thinks was sent a       */
            /* legitimate command, which it accepted and executed. */
            /* As often said, "this condition should never occur." */
            pmod->no_dataerr++;
            return (FALSE);
            break;

        case '*':
            /* Double Yikes! Someone is out there trying to send   */
            /* unsolicited data, and I was told this feature was   */
            /* never implemented in the SEC UPS code. What we need */
            /* do here is simply ignore it, and go wait for the    */
            /* correct response, or a timeout. Should I bother to  */
            /* say it again?  "this condition should never occur." */
            pmod->no_dataerr++;
            return (FALSE);
            break;

        case 'P':
        case 'S':
            /* Triple Yikes! Someone is out there trying to tell   */
            /* ME what to do! Most likely this is not a UPS but    */
            /* another SEC UPS driver, and someone has got their   */
            /* wires crossed! "this condition should never occur." */
            pmod->no_dataerr++;
            return (FALSE);
            break;

        default:
            /* This is most likely a data error. Ignore it and go  */
            /* wait for a valid response.                          */
            pmod->no_dataerr++;
            return (FALSE);
            break;
    }

    /* get length of the data character portion of the message */
    count = (rxp->rxmsglen - 5);
    if (count < 1) return (FALSE);
    /* clear out parametertable flags */
    for (i = 0; i <= 46; i++)
    {
	upstable [i].support = FALSE;
    }

    /* parse response string, set valid flags in parametertable */
    i = 5;
    while (count > 0)
    {
        length = count;
        status = getvalue (&rxp->rxbuf[i], &value, &length);
/*
printf ("count = %d, length = %d, value = %d\n", count, length, value);
*/
        tempi = i;
        i = i + length; 
        count = count - length;

        if (status != TRUE)
        {
            if (rxp->rxbuf[tempi] == ',')
            {
                /* no error. skip to next. */
            }
            else
            {
/*
    printf("*********** Count = %d ****************\n", count);
*/
                pmod->no_dataerr++;
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readparameters1: Illegal character '%s'[0x%x] in %dth position in message.\n", 
                             rxp->rxbuf[i - 1], rxp->rxbuf[i - 1], i, 0, 0, 0);
                return (FALSE);
            }
        }

        if ((value > 0) && (value <= 46) )
        {
            if (trace[line] & DETAILEDTRACE)
                logMsg ("readparameters1: Parameter Index Value (%d) is Valid.\n", value, 0, 0, 0, 0, 0);
            upstable [value].support = TRUE;
        }
        else
        {
            if (value > 46) 
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readparameters1: Parameter Index Value (%d) Greater Than Expected Maximum (%d).\n",
                             value, 46, 0, 0, 0, 0);
            }
            else
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readparameters1: Parameter Index Value (%d) Less Than Expected Minimum (%d).\n",
                             value, 1, 0, 0, 0, 0);
            }
        }
    }
    return (TRUE);
}

/*                                                              
** readparameters2                                              
**                                                              
*/                                                                              
/* only called from on_line */


static  unsigned short readparameters2 (unsigned short line, SEC *pmod)
{
    TXLINE         *txp;
    RXLINE         *rxp;
    signed short   count;
    unsigned short  i, tempi, value, length, status;

    /* Set up pointers for comm buffers */
    txp = txlinetab [line & (~ONERETRY)];
    rxp = rxlinetab [line & (~ONERETRY)];

    /* Format the output message */
    txp->txbuf [0] = '^';
    txp->txbuf [1] = 'P';
    txp->txbuf [2] = '0';
    txp->txbuf [3] = '0';
    txp->txbuf [4] = '3';
    txp->txbuf [5] = 'A';
    txp->txbuf [6] = 'P';
    txp->txbuf [7] = '2';
    txp->txmsglen = 8;
    pmod->no_datain_msgs++;

    /* Send out the message */
    if (msgxfer (line, pmod, MAXDATALENGTH) == XFERBAD)
    {
        pmod->is_responding = FALSE;
        return (FALSE);
    }

    /* now that communications are completed, strip off the retry bit */
    line = line & (~ONERETRY);

    /* expecting a Data Returned response. if anything else, its a mistake */
    switch (rxp->rxbuf [1])
    {
        case 'D':
            /* so far so good. */
            break;

        case '0':
            /* most likely an unsuppported command has been sent,  */
            /*  although this one is supposedly always valid.      */
            pmod->no_dataerr++;
            return (FALSE);
            break;

        case '1':
            /* Yikes! The UPS obviously it thinks was sent a       */
            /* legitimate command, which it accepted and executed. */
            /* As often said, "this condition should never occur." */
            pmod->no_dataerr++;
            return (FALSE);
            break;

        case '*':
            /* Double Yikes! Someone is out there trying to send   */
            /* unsolicited data, and I was told this feature was   */
            /* never implemented in the SEC UPS code. What we need */
            /* do here is simply ignore it, and go wait for the    */
            /* correct response, or a timeout. Should I bother to  */
            /* say it again?  "this condition should never occur." */
            pmod->no_dataerr++;
            return (FALSE);
            break;

        case 'P':
        case 'S':
            /* Triple Yikes! Someone is out there trying to tell   */
            /* ME what to do! Most likely this is not a UPS but    */
            /* another SEC UPS driver, and someone has got their   */
            /* wires crossed! "this condition should never occur." */
            pmod->no_dataerr++;
            return (FALSE);
            break;

        default:
            /* This is most likely a data error. Ignore it and go  */
            /* wait for a valid response.                          */
            pmod->no_dataerr++;
            return (FALSE);
            break;
    }

    /* get character count */
    count = (rxp->rxmsglen - 5);
    if (count < 1) return (FALSE);

    /* clear out parametertable flags */
    for (i = 47; i <= 89; i++)
    {
	upstable [i].support = FALSE;
    }

    /* parse response string, set valid flags in parametertable */
    i = 5;
    while (count > 0)
    {
        length = count;
        status = getvalue (&rxp->rxbuf[i], &value, &length);

        tempi = i;
        i = i + length; 
        count = count - length;

        if (status != TRUE)
        {
            if (rxp->rxbuf[tempi] == ',')
            {
                /* no error. skip to next. */
            }
            else
            {
                pmod->no_dataerr++;
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readparameters2: Illegal character '%s'[0x%x] in %dth position in message.\n", 
                             rxp->rxbuf[i - 1], rxp->rxbuf[i - 1], i, 0, 0, 0);
                return (FALSE);
            }
        }

        if ((value >= 47) && (value <= 89) )
        {
            if (trace[line] & DETAILEDTRACE)
                logMsg ("readparameters2: Parameter Index Value (%d) is Valid.\n", value, 0, 0, 0, 0, 0);
            upstable [value].support = TRUE;
        }
        else
        {
            if (value > 89) 
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readparameter2: Parameter Index Value (%d) Greater Than Expected Maximum (%d).\n",
                             value, 89, 0, 0, 0, 0);
            }
            else
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readparameters2: Parameter Index Value (%d) Less Than Expected Minimum (%d).\n",
                             value, 46, 0, 0, 0, 0);
            }
        }
        count = count - length;
    }
    return (TRUE);
}


/*                                                              
** readstatusgroup1                                              
**                                                              
*/                                                                              
/* only called from on_line */


static  unsigned short readstatusgroup1 (unsigned short line, SEC *pmod)
{
       TXLINE         *txp;
       RXLINE         *rxp;
       signed short    count;
       unsigned short  i, tempi, j, value, length, status;
       unsigned short  index;

static short  parsingerrors;

    /* Set up pointers for comm buffers */
    txp = txlinetab [line & (~ONERETRY)];
    rxp = rxlinetab [line & (~ONERETRY)];

    /* Format the output message */
    txp->txbuf [0] = '^';
    txp->txbuf [1] = 'P';
    txp->txbuf [2] = '0';
    txp->txbuf [3] = '0';
    txp->txbuf [4] = '3';
    txp->txbuf [5] = 'S';
    txp->txbuf [6] = 'T';
    txp->txbuf [7] = '1';
    txp->txmsglen = 8;
    pmod->no_datain_msgs++;

    /* Send out the message */
    if (msgxfer (line, pmod, MAXDATALENGTH) == XFERBAD)
    {
        pmod->is_responding = FALSE;
        return (FALSE);
    }

    /* now that communications are completed, strip off the retry bit */
    line = line & (~ONERETRY);

    /* after N sequential parsing errors, mark the data as invalid. */
    if (parsingerrors >= parseinvalidcount [line])
    {
        pmod->st1.data_valid = FALSE;
    }

    /* after N sequential parsing errors, mark the plc as offline. */
    if (parseofflinecounter [line] >= parseofflinecount [line])
    {
        pmod->st1.data_valid = FALSE;
        pmod->is_responding = FALSE;
        return (FALSE);
    }

    /* expecting a Data Returned response. if anything else, its a mistake */
    switch (rxp->rxbuf [1])
    {
        case 'D':
            /* so far so good. */
            break;

        case '0':
            /* most likely an unsuppported command has been sent,  */
            /*  although this one is supposedly always valid.      */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case '1':
            /* Yikes! The UPS obviously it thinks was sent a       */
            /* legitimate command, which it accepted and executed. */
            /* As often said, "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case '*':
            /* Double Yikes! Someone is out there trying to send   */
            /* unsolicited data, and I was told this feature was   */
            /* never implemented in the SEC UPS code. What we need */
            /* do here is simply ignore it, and go wait for the    */
            /* correct response, or a timeout. Should I bother to  */
            /* say it again?  "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case 'P':
        case 'S':
            /* Triple Yikes! Someone is out there trying to tell   */
            /* ME what to do! Most likely this is not a UPS but    */
            /* another SEC UPS driver, and someone has got their   */
            /* wires crossed! "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        default:
            /* This is most likely a data error. Ignore it and go  */
            /* wait for a valid response.                          */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;
    }

    /* get character count */
    count = (rxp->rxmsglen - 5);
    if (count < 1)
    {
        pmod->no_dataerr++;
        parsingerrors++;
        parseofflinecounter [line]++;
        return (TRUE);
    }

    /* parse response string, storing valid flag and value, if any */
    i = 5;
    j = 0;      /* parameter counter */

    while (count > 0)
    {
        length = count;
        status = getvalue (&rxp->rxbuf[i], &value, &length);
        index = pmod->st1.number [j];

        tempi = i;
        i = i + length; 
        count = count - length;

        if (status == TRUE)
        {
            /* good data. mark as valid. for now, null raw string. */
            pmod->st1.status [j] = TRUE;
            pmod->st1.data [j] = value;
            pmod->st1.raw_read [j][0] = 0;
            if (upstable [pmod->st1.number [j]].support)
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup1: Data Group 1 [%d], which is supported by the UPS, has a Valid Entry.\n",
                             j, 0, 0, 0, 0, 0);
            }
            else
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup1: Data Group 1 [%d], although unsupported by the UPS, has a Valid Entry. (suspect a parsing error)\n",
                             j, 0, 0, 0, 0, 0);
            }
        }
        else 
        {
            if (rxp->rxbuf[tempi] == ',')
            {
                /* no available data. mark as invalid. null raw string. */
                pmod->st1.status [j] = FALSE;
                pmod->st1.data [j] = 0;
                pmod->st1.raw_read [j][0] = 0;
                if (upstable [pmod->st1.number [j]].support)
                {
                    if (trace[line] & DETAILEDTRACE)
                        logMsg ("readstatusgroup1: Data Group 1 [%d], although supported by the UPS, has No Valid Entry. (suspect a parsing error)\n",
                                 j, 0, 0, 0, 0, 0);
                }
                else
                {
                    if (trace[line] & DETAILEDTRACE)
                        logMsg ("readstatusgroup1: Data Group 1 [%d], which is unsupported by the UPS, has No Valid Entry.\n", 
                                 j, 0, 0, 0, 0, 0);
                }
            }
            else
            {
                pmod->no_dataerr++;
                parsingerrors++;
                parseofflinecounter [line]++;
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup1: Illegal character '%s'[0x%x] in %dth position in message.\n", 
                             rxp->rxbuf[tempi], rxp->rxbuf[tempi], tempi, 0, 0, 0);
                return (TRUE);
            }

        }
        if (++j >= MAXDG1)
        {
            if (trace[line] & DETAILEDTRACE)
                logMsg ("readstatusgroup1: Maximum Parameter Index Reached. Exiting.\n", MAXDG1, 0, 0, 0, 0, 0);
            /* reset cumulative parsing error counter. */
            parsingerrors = 0;
            parseofflinecounter [line] = 0;
            return (TRUE);
        }
    }

    /* reset cumulative parsing error counter. */
    parsingerrors = 0;
    parseofflinecounter [line] = 0;
    pmod->st1.data_valid = TRUE;

    return (TRUE);
}



/*                                                              
** readstatusgroup2                                              
**                                                              
*/                                                                              
/* only called from on_line */


static  unsigned short readstatusgroup2 (unsigned short line, SEC *pmod)
{
       TXLINE         *txp;
       RXLINE         *rxp;
       signed short    count;
       unsigned short  i, tempi, j, value, length, status;
       unsigned short  index;

static short  parsingerrors;

    /* Set up pointers for comm buffers */
    txp = txlinetab [line & (~ONERETRY)];
    rxp = rxlinetab [line & (~ONERETRY)];

    /* Format the output message */
    txp->txbuf [0] = '^';
    txp->txbuf [1] = 'P';
    txp->txbuf [2] = '0';
    txp->txbuf [3] = '0';
    txp->txbuf [4] = '3';
    txp->txbuf [5] = 'S';
    txp->txbuf [6] = 'T';
    txp->txbuf [7] = '2';
    txp->txmsglen = 8;
    pmod->no_datain_msgs++;

    /* Send out the message */
    if (msgxfer (line, pmod, MAXDATALENGTH) == XFERBAD)
    {
        pmod->is_responding = FALSE;
        return (FALSE);
    }

    /* now that communications are completed, strip off the retry bit */
    line = line & (~ONERETRY);

    /* after N sequential parsing errors, mark the data as invalid. */
    if (parsingerrors >= parseinvalidcount [line])
    {
        pmod->st2.data_valid = FALSE;
    }

    /* after N sequential parsing errors, mark the plc as offline. */
    if (parseofflinecounter [line] >= parseofflinecount [line])
    {
        pmod->st2.data_valid = FALSE;
        pmod->is_responding = FALSE;
        return (FALSE);
    }

    /* expecting a Data Returned response. if anything else, its a mistake */
    switch (rxp->rxbuf [1])
    {
        case 'D':
            /* so far so good. */
            break;

        case '0':
            /* most likely an unsuppported command has been sent,  */
            /*  although this one is supposedly always valid.      */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case '1':
            /* Yikes! The UPS obviously it thinks was sent a       */
            /* legitimate command, which it accepted and executed. */
            /* As often said, "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case '*':
            /* Double Yikes! Someone is out there trying to send   */
            /* unsolicited data, and I was told this feature was   */
            /* never implemented in the SEC UPS code. What we need */
            /* do here is simply ignore it, and go wait for the    */
            /* correct response, or a timeout. Should I bother to  */
            /* say it again?  "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case 'P':
        case 'S':
            /* Triple Yikes! Someone is out there trying to tell   */
            /* ME what to do! Most likely this is not a UPS but    */
            /* another SEC UPS driver, and someone has got their   */
            /* wires crossed! "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        default:
            /* This is most likely a data error. Ignore it and go  */
            /* wait for a valid response.                          */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;
    }

    /* get character count */
    count = (rxp->rxmsglen - 5);
    if (count < 1)
    {
        pmod->no_dataerr++;
        parsingerrors++;
        parseofflinecounter [line]++;
        return (TRUE);
    }

    /* parse response string, storing valid flag and value, if any */
    i = 5;
    j = 0;      /* parameter counter */

    while (count > 0)
    {
        length = count;
        status = getvalue (&rxp->rxbuf[i], &value, &length);
        index = pmod->st2.number [j];

        tempi = i;
        i = i + length; 
        count = count - length;

        if (status == TRUE)
        {
            /* good data. mark as valid. for now, null raw string. */
            pmod->st2.status [j] = TRUE;
            pmod->st2.data [j] = value;
            pmod->st2.raw_read [j][0] = 0;
            if (upstable [pmod->st2.number [j]].support)
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup2: Data Group 2 [%d], which is supported by the UPS, has a Valid Entry.\n",
                             j, 0, 0, 0, 0, 0);
            }
            else
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup2: Data Group 2 [%d], although unsupported by the UPS, has a Valid Entry. (suspect a parsing error)\n",
                             j, 0, 0, 0, 0, 0);
            }
        }
        else 
        {
            if (rxp->rxbuf[tempi] == ',')
            {
                /* no available data. mark as invalid. null raw string. */
                pmod->st2.status [j] = FALSE;
                pmod->st2.data [j] = 0;
                pmod->st2.raw_read [j][0] = 0;
                if (upstable [pmod->st2.number [j]].support)
                {
                    if (trace[line] & DETAILEDTRACE)
                        logMsg ("readstatusgroup2: Data Group 2 [%d], although supported by the UPS, has No Valid Entry. (suspect a parsing error)\n",
                                 j, 0, 0, 0, 0, 0);
                }
                else
                {
                    if (trace[line] & DETAILEDTRACE)
                        logMsg ("readstatusgroup2: Data Group 2 [%d], which is unsupported by the UPS, has No Valid Entry.\n", 
                                 j, 0, 0, 0, 0, 0);
                }
            }
            else
            {
                pmod->no_dataerr++;
                parsingerrors++;
                parseofflinecounter [line]++;
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup2: Illegal character '%s'[0x%x] in %dth position in message.\n", 
                             rxp->rxbuf[i - 1], rxp->rxbuf[i - 1], i, 0, 0, 0);
                return (TRUE);
            }

        }
        if (++j >= MAXDG2)
        {
            if (trace[line] & DETAILEDTRACE)
                logMsg ("readstatusgroup2: Maximum Parameter Index Reached. Exiting.\n", MAXDG1, 0, 0, 0, 0, 0);
            /* reset cumulative parsing error counter. */
            parsingerrors = 0;
            parseofflinecounter [line] = 0;
            return (TRUE);
        }
    }

    /* reset cumulative parsing error counter. */
    parsingerrors = 0;
    parseofflinecounter [line] = 0;
    pmod->st2.data_valid = TRUE;

    return (TRUE);
}

/*                                                              
** readstatusgroup3                                              
**                                                              
*/                                                                              
/* only called from on_line */


static  unsigned short readstatusgroup3 (unsigned short line, SEC *pmod)
{
       TXLINE         *txp;
       RXLINE         *rxp;
       signed short    count;
       unsigned short  i, tempi, j, value, length, status;
       unsigned short  index;

static short  parsingerrors;

    /* Set up pointers for comm buffers */
    txp = txlinetab [line & (~ONERETRY)];
    rxp = rxlinetab [line & (~ONERETRY)];

    /* Format the output message */
    txp->txbuf [0] = '^';
    txp->txbuf [1] = 'P';
    txp->txbuf [2] = '0';
    txp->txbuf [3] = '0';
    txp->txbuf [4] = '3';
    txp->txbuf [5] = 'S';
    txp->txbuf [6] = 'T';
    txp->txbuf [7] = '3';
    txp->txmsglen = 8;
    pmod->no_datain_msgs++;

    /* Send out the message */
    if (msgxfer (line, pmod, MAXDATALENGTH) == XFERBAD)
    {
        pmod->is_responding = FALSE;
        return (FALSE);
    }

    /* now that communications are completed, strip off the retry bit */
    line = line & (~ONERETRY);

    /* after N sequential parsing errors, mark the data as invalid. */
    if (parsingerrors >= parseinvalidcount [line])
    {
        pmod->st3.data_valid = FALSE;
    }

    /* after N sequential parsing errors, mark the plc as offline. */
    if (parseofflinecounter [line] >= parseofflinecount [line])
    {
        pmod->st3.data_valid = FALSE;
        pmod->is_responding = FALSE;
        return (FALSE);
    }

    /* expecting a Data Returned response. if anything else, its a mistake */
    switch (rxp->rxbuf [1])
    {
        case 'D':
            /* so far so good. */
            break;

        case '0':
            /* most likely an unsuppported command has been sent,  */
            /*  although this one is supposedly always valid.      */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case '1':
            /* Yikes! The UPS obviously it thinks was sent a       */
            /* legitimate command, which it accepted and executed. */
            /* As often said, "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case '*':
            /* Double Yikes! Someone is out there trying to send   */
            /* unsolicited data, and I was told this feature was   */
            /* never implemented in the SEC UPS code. What we need */
            /* do here is simply ignore it, and go wait for the    */
            /* correct response, or a timeout. Should I bother to  */
            /* say it again?  "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case 'P':
        case 'S':
            /* Triple Yikes! Someone is out there trying to tell   */
            /* ME what to do! Most likely this is not a UPS but    */
            /* another SEC UPS driver, and someone has got their   */
            /* wires crossed! "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        default:
            /* This is most likely a data error. Ignore it and go  */
            /* wait for a valid response.                          */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;
    }

    /* get character count */
    count = (rxp->rxmsglen - 5);
    if (count < 1)
    {
        pmod->no_dataerr++;
        parsingerrors++;
        parseofflinecounter [line]++;
        return (TRUE);
    }

    /* parse response string, storing valid flag and value, if any */
    i = 5;
    j = 0;      /* parameter counter */

    while (count > 0)
    {
        length = count;
        status = getvalue (&rxp->rxbuf[i], &value, &length);
        index = pmod->st3.number [j];

        tempi = i;
        i = i + length; 
        count = count - length;

        if (status == TRUE)
        {
            /* good data. mark as valid. for now, null raw string. */
            pmod->st3.status [j] = TRUE;
            pmod->st3.data [j] = value;
            pmod->st3.raw_read [j][0] = 0;
            if (upstable [pmod->st3.number [j]].support)
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup3: Data Group 3 [%d], which is supported by the UPS, has a Valid Entry.\n",
                             j, 0, 0, 0, 0, 0);
            }
            else
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup3: Data Group 3 [%d], although unsupported by the UPS, has a Valid Entry. (suspect a parsing error)\n",
                             j, 0, 0, 0, 0, 0);
            }
        }
        else 
        {
            if (rxp->rxbuf[tempi] == ',')
            {
                /* no available data. mark as invalid. null raw string. */
                pmod->st3.status [j] = FALSE;
                pmod->st3.data [j] = 0;
                pmod->st3.raw_read [j][0] = 0;
                if (upstable [pmod->st3.number [j]].support)
                {
                    if (trace[line] & DETAILEDTRACE)
                        logMsg ("readstatusgroup3: Data Group 3 [%d], although supported by the UPS, has No Valid Entry. (suspect a parsing error)\n",
                                 j, 0, 0, 0, 0, 0);
                }
                else
                {
                    if (trace[line] & DETAILEDTRACE)
                        logMsg ("readstatusgroup3: Data Group 3 [%d], which is unsupported by the UPS, has No Valid Entry.\n", 
                                 j, 0, 0, 0, 0, 0);
                }
            }
            else
            {
                pmod->no_dataerr++;
                parsingerrors++;
                parseofflinecounter [line]++;
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup3: Illegal character '%s'[0x%x] in %dth position in message.\n", 
                             rxp->rxbuf[i - 1], rxp->rxbuf[i - 1], i, 0, 0, 0);
                return (TRUE);
            }

        }
        if (++j >= MAXDG3)
        {
            if (trace[line] & DETAILEDTRACE)
                logMsg ("readstatusgroup3: Maximum Parameter Index Reached. Exiting.\n", MAXDG1, 0, 0, 0, 0, 0);
            /* reset cumulative parsing error counter. */
            parsingerrors = 0;
            parseofflinecounter [line] = 0;
            return (TRUE);
        }
    }

    /* reset cumulative parsing error counter. */
    parsingerrors = 0;
    parseofflinecounter [line] = 0;
    pmod->st3.data_valid = TRUE;

    return (TRUE);
}

/*                                                              
** readstatusgroup5                                             
**                                                              
*/                                                                              
/* only called from on_line */


static unsigned short readstatusgroup5 (unsigned short line, SEC *pmod)
{
       TXLINE         *txp;
       RXLINE         *rxp;
       signed short    count;
       unsigned short  i, tempi, j, value, length, status;
       unsigned short  index;

static short  parsingerrors;

    /* Set up pointers for comm buffers */
    txp = txlinetab [line & (~ONERETRY)];
    rxp = rxlinetab [line & (~ONERETRY)];

    /* Format the output message */
    txp->txbuf [0] = '^';
    txp->txbuf [1] = 'P';
    txp->txbuf [2] = '0';
    txp->txbuf [3] = '0';
    txp->txbuf [4] = '3';
    txp->txbuf [5] = 'S';
    txp->txbuf [6] = 'T';
    txp->txbuf [7] = '5';
    txp->txmsglen = 8;
    pmod->no_datain_msgs++;

    /* Send out the message */
    if (msgxfer (line, pmod, MAXDATALENGTH) == XFERBAD)
    {
        pmod->is_responding = FALSE;
        return (FALSE);
    }

    /* now that communications are completed, strip off the retry bit */
    line = line & (~ONERETRY);

    /* after N sequential parsing errors, mark the data as invalid. */
    if (parsingerrors >= parseinvalidcount [line])
    {
        pmod->st5.data_valid = FALSE;
    }

    /* after N sequential parsing errors, mark the plc as offline. */
    if (parseofflinecounter [line] >= parseofflinecount [line])
    {
        pmod->st5.data_valid = FALSE;
        pmod->is_responding = FALSE;
        return (FALSE);
    }

    /* expecting a Data Returned response. if anything else, its a mistake */
    switch (rxp->rxbuf [1])
    {
        case 'D':
            /* so far so good. */
            break;

        case '0':
            /* most likely an unsuppported command has been sent,  */
            /*  although this one is supposedly always valid.      */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case '1':
            /* Yikes! The UPS obviously it thinks was sent a       */
            /* legitimate command, which it accepted and executed. */
            /* As often said, "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case '*':
            /* Double Yikes! Someone is out there trying to send   */
            /* unsolicited data, and I was told this feature was   */
            /* never implemented in the SEC UPS code. What we need */
            /* do here is simply ignore it, and go wait for the    */
            /* correct response, or a timeout. Should I bother to  */
            /* say it again?  "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        case 'P':
        case 'S':
            /* Triple Yikes! Someone is out there trying to tell   */
            /* ME what to do! Most likely this is not a UPS but    */
            /* another SEC UPS driver, and someone has got their   */
            /* wires crossed! "this condition should never occur." */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;

        default:
            /* This is most likely a data error. Ignore it and go  */
            /* wait for a valid response.                          */
            pmod->no_dataerr++;
            parsingerrors++;
            parseofflinecounter [line]++;
            return (TRUE);
            break;
    }

    /* get character count */
    count = (rxp->rxmsglen - 5);
    if (count < 1)
    {
        pmod->no_dataerr++;
        parsingerrors++;
        parseofflinecounter [line]++;
        return (TRUE);
    }

    /* parse response string, storing valid flag and value, if any */
    i = 5;
    j = 0;      /* parameter counter */

    while (count > 0)
    {
        length = count;
        status = getvalue (&rxp->rxbuf[i], &value, &length);
        index = pmod->st5.number [j];

        tempi = i;
        i = i + length; 
        count = count - length;

        if (status == TRUE)
        {
            /* good data. mark as valid. for now, null raw string. */
            pmod->st5.status [j] = TRUE;
            pmod->st5.data [j] = value;
            pmod->st5.raw_read [j][0] = 0;
            if (upstable [pmod->st5.number [j]].support)
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup5: Data Group 5 [%d], which is supported by the UPS, has a Valid Entry.\n",
                             j, 0, 0, 0, 0, 0);
            }
            else
            {
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup5: Data Group 5 [%d], although unsupported by the UPS, has a Valid Entry. (suspect a parsing error)\n",
                             j, 0, 0, 0, 0, 0);
            }
        }
        else 
        {
            if (rxp->rxbuf[tempi] == ',')
            {
                /* no available data. mark as invalid. null raw string. */
                pmod->st5.status [j] = FALSE;
                pmod->st5.data [j] = 0;
                pmod->st5.raw_read [j][0] = 0;
                if (upstable [pmod->st5.number [j]].support)
                {
                    if (trace[line] & DETAILEDTRACE)
                        logMsg ("readstatusgroup5: Data Group 5 [%d], although supported by the UPS, has No Valid Entry. (suspect a parsing error)\n",
                                 j, 0, 0, 0, 0, 0);
                }
                else
                {
                    if (trace[line] & DETAILEDTRACE)
                        logMsg ("readstatusgroup5: Data Group 5 [%d], which is unsupported by the UPS, has No Valid Entry.\n", 
                                 j, 0, 0, 0, 0, 0);
                }
            }
            else
            {
                pmod->no_dataerr++;
                parsingerrors++;
                parseofflinecounter [line]++;
                if (trace[line] & DETAILEDTRACE)
                    logMsg ("readstatusgroup5: Illegal character '%s'[0x%x] in %dth position in message.\n", 
                             rxp->rxbuf[tempi], rxp->rxbuf[tempi], tempi, 0, 0, 0);
                return (TRUE);
            }

        }
        if (++j >= MAXDG5)
        {
            if (trace[line] & DETAILEDTRACE)
                logMsg ("readstatusgroup5: Maximum Parameter Index Reached. Exiting.\n", MAXDG1, 0, 0, 0, 0, 0);

            /* reset cumulative parsing error counter. */
            parsingerrors = 0;
            parseofflinecounter [line] = 0;
            return (TRUE);
        }
    }

    /* reset cumulative parsing error counter. */
    parsingerrors = 0;
    parseofflinecounter [line] = 0;
    pmod->st5.data_valid = TRUE;

    return (TRUE);
}

/*                                                                              
**     prepresponses                                                            
**                                                                              
**      The purpose of this routine is to mark the Modules as on hold and the   
**    data as not valid until read again from the Module                        
**                                                                              
*/
/* only called from secdriver */

static  void prepresponses (unsigned short line)
{
    unsigned short   mod;
    SEC          *pmod;

    for (mod = 0; (mod < NUMMODS); mod++)
    {
        pmod = modules [line];
        if (pmod)                /* verify pointer validity prior to use */
        {
            pmod->holdflag = TRUE;
            mark_not_valid (pmod);
        }
    }
}


/*                                                                       
**      any_responding                                                   
**                                                                       
**      The purpose of this routine is to see if any units are           
**    responding on this line                                            
**                                                                       
*/
/* only called from secdriver */

static  int any_responding (unsigned short line)
{
    unsigned short  mod;
    SEC        *pmod;

    for (mod = 0; (mod < NUMMODS); mod++)
    {
        pmod = modules [line];
        if (pmod)            /* verify pointer validity prior to use */
        {
            /* check cmd_force_offline flag because it is set in an asynchronous task */
            if ((pmod->is_responding) && (!pmod->cmd_force_offline))
            {
                return (TRUE);
            }
        }
    }
    return (FALSE);
}


/*
**      secdumpmsgcnts
**
**      The purpose of this routine is to dump the line level statistics for this driver. 
**
*/
/* called from sectableinfo */

void secdumpmsgcnts (unsigned short line)

    {
        unsigned long	temp, t, x, y, e, z;

        temp = tickGet ();
        t = (temp - linestats [line].startingtick);

 	x = linestats [line].TXintcount;
	y = linestats [line].RXintcount;
	e = linestats [line].RXbadcount;
	z = linestats [line].msgxfercount;

        if ((int)z == 0)
        {
            printf ("TX Interrupt Count    = %10d\n", (int)x);
            printf ("RX Interrupt Count    = %10d\n", (int)y);
            printf ("BadRX Intrpt Count    = %10d\n", (int)e);
            printf (" ******  No Messages Transferred ****** \n");
            return;
        }

 	printf ("\n");
	printf ("Measurement Interval: clock ticks = %d, seconds = %d.%1d \n", (int)t, (int)(t/60), (int)((t%60)/6));
 	printf ("\n");
 	printf ("TX Interrupt Count    = %10d   Avg TX cnts/sec        = %d \n", (int)x, (int)((60*x)/t));
 	printf ("RX Interrupt Count    = %10d   Avg RX cnts/sec        = %d \n", (int)y, (int)((60*y)/t));
 	printf ("                                     Avg RX cnts/Msgxfer    = %d.\n", (int)(x/z));
 	printf ("                                     Avg TX cnts/Msgxfer    = %d.\n", (int)(y/z));
        printf ("BadRX Intrpt Count    = %10d\n", (int)e);
 	printf ("\n");
 	printf ("RX+TX Interrupt Count = %10d   Avg RX+TX cnts/sec     = %d \n", (int)(x+y), (int)((60*(x+y))/t));
 	printf ("                                     Avg RX+TX cnts/Msgxfer = %d \n" ,(int)((x+y)/z));
 	printf ("\n");
	printf ("Msgxfer Count         = %10d   Avg Msgxfers/sec       = %d \n", (int)z, (int)((60*z)/t));
	printf ("                                     Avg msec/Msgxfer       = %d \n", (int)((t*1000)/(z*60)));
 	printf ("\n");
        return;
}

/*
**      secclearmsgcnts
**
**      The purpose of this routine is to dump and clear the line level statistics
**      for this driver.
**
*/
/* called from sectableinfo */

void secclearmsgcnts (unsigned short line)
{
    secdumpmsgcnts (line);

    linestats [line].TXintcount = 0;
    linestats [line].RXintcount = 0;
    linestats [line].RXbadcount = 0;
    linestats [line].msgxfercount = 0;

    linestats [line].startingtick = tickGet ();
    printf (" ******  All Counts Cleared  ****** \n");
    printf ("\n");
    return;
}



/*                                                                                           
**      secdumpstats                                                                         
**                                                                                           
**      The purpose of this routine is to dump the module level statistics for this driver.  
**                                                                                           
*/                                                                                           
/* called from sectableinfo */

void secdumpstats (unsigned short line)
{
    SEC            *pmod;
    unsigned long   microsecs;

    pmod = modules [line];
    if (!pmod)
    {
        logMsg ("Line/Module Address Not Defined \n", 0, 0, 0, 0, 0, 0);
        return;
    }

    printf (" Was requested        = %5d         ", pmod->was_requested);
    printf (" Is responding        = %5d       \n", pmod->is_responding);
    printf (" # Data out messages  = %5d         ", pmod->no_dataout_msgs & 0x7fff);
    printf (" # Data in messages   = %5d       \n", pmod->no_datain_msgs & 0x7fff);
    printf (" #No Rx Response      = %5d         ", pmod->no_rxnoresp & 0x7fff);
    printf (" #No Tx Responses     = %5d       \n", pmod->no_txnoresp & 0x7fff);
    printf (" #Data Errors         = %5d         ", pmod->no_dataerr & 0x7fff);
    printf (" #Retries             = %5d       \n", pmod->no_retries & 0x7fff);
    printf (" #Rx Hw Errors        = %5d         ", pmod->no_rxhwerrs & 0x7fff);
    printf (" #Tx Hw Errors        = %5d       \n", pmod->no_txhwerrs & 0x7fff);
    printf (" #Checksum Errors     = %5d         ", pmod->no_chksumerrs & 0x7fff);
    printf (" #Framing Errors      = %5d       \n", pmod->no_frameerrs & 0x7fff);
    printf (" #Off Lines           = %5d         ", pmod->no_offlines & 0x7fff);
    printf (" Reset Statistics     = %5d       \n", pmod->cmd_reset_stat);
    printf (" Force Offline        = %5d       \n", pmod->cmd_force_offline);


    printf ("Error Class Counts\n\n");
/*
    printf (" # No Errors          = %5d         ", pmod->noerror & 0x7fff);
    printf (" # Programming        = %5d       \n", pmod->progerror & 0x7fff);
    printf (" # Communication      = %5d         ", pmod->commerror & 0x7fff);
    printf (" # Fatal              = %5d       \n", pmod->fatalerror & 0x7fff);
    printf (" # Configuration      = %5d         ", pmod->configerror & 0x7fff);
    printf (" # Unknown            = %5d       \n", pmod->unknownerror & 0x7fff);

    printf ("Sec Status Counts\n\n");

    printf (" # No Err (00)        = %5d         ", pmod->error00 & 0x7fff);
    printf (" # Command OK (01)    = %5d       \n", pmod->error01 & 0x7fff);
    printf (" # Prog Err (02)      = %5d         ", pmod->error02 & 0x7fff);
    printf (" # Power-up Err (03)  = %5d       \n", pmod->error03 & 0x7fff);
    printf (" # Framing Err (04)   = %5d         ", pmod->error04 & 0x7fff);
    printf (" # PROM Err (10)      = %5d       \n", pmod->error10 & 0x7fff);
    printf (" # RAM Err (11)       = %5d         ", pmod->error11 & 0x7fff);
    printf (" # VCO Err (12)       = %5d       \n", pmod->error12 & 0x7fff);
    printf (" # Chan Num Err (40)  = %5d         ", pmod->error40 & 0x7fff);
    printf (" # Card Num Err (41)  = %5d       \n", pmod->error41 & 0x7fff);
    printf (" # EV Err (42)        = %5d         ", pmod->error42 & 0x7fff);
    printf (" # Val Range Err (43) = %5d       \n", pmod->error43 & 0x7fff);
    printf (" # Overrange Err (44) = %5d         ", pmod->error44 & 0x7fff);
    printf (" # PowerFailure (45)  = %5d       \n", pmod->error45 & 0x7fff);
    printf (" # All Checksum (50)  = %5d       \n", pmod->error50 & 0x7fff);

*/


    printf ("\n");
    printf (" # of Messages %d \n", pmod->msgs);
    printf (" # Total response time (ticks) = %lu \n", pmod->responsetime);
    if (pmod->msgs)
    {
        printf ("   Average response time (ticks) = %d \n", (int) (pmod->responsetime/pmod->msgs));
        microsecs = (16667*pmod->responsetime)/pmod->msgs;
        printf ("   Average response time (msec) = %d.%1d \n", (int) (microsecs/1000), (int) ((microsecs%1000)/100));
    }

    printf (" # of scan passes = %d \n", pmod->tscanpass);
    printf ("\n");

    printf (" Resource Responding = %d \n", resresponse[line]);
    printf (" Hold flag  = %d \n", pmod->holdflag );
        return;
}

/*
**      secclearstats
**
**      The purpose of this routine is to clear the module level statistics for this driver.
**
*/
/* called from sectableinfo */

void secclearstats (unsigned short line)
{
    SEC        *pmod;

    pmod = modules [line];
    if (!pmod)
    {
        logMsg ("Line/Module Address Not Defined \n", 0, 0, 0, 0, 0, 0);
        return;
    }
    initstats (pmod); /* use existing procedure */
    return;
}



/*
**      sectableinfo                                                  
**                                                                      
**      The purpose of this routine is twofold:                         
**         1) dump the module level statistics for the modbus driver.   
**         2) dump the data for the specified Sec module.            
**                                                                      
*/
/* not called from within this driver */

void sectableinfo (unsigned short line)
{
    SEC            *pmod;
    unsigned short  i;
    int             index;
    unsigned short  linecnt;

    pmod = modules [line];
    if (!pmod)
    {
        logMsg("Line/Module Address Not Defined \n", 0, 0, 0, 0, 0, 0);
        return;
    }

    /* previously integral code moved to seperate procedure for debugging convenience */

    secdumpstats (line);

    linecnt = 1;

    printf("ST1 Command Requested = %d\n", pmod->st1.was_requested);
    printf("ST1 Command Successful = %d\n", pmod->st1.data_valid);
    printf("---------------------------------------------------------------------------------------\n");
    printf("Offset Status   Decimal Value    Hex Value    [RawData]    Parameter Num/Name\n");
    for (i = 0; i < MAXDG1; i++)
    {
        index = pmod->st1.number [i];
        printf("   %3d:   %3d     %10ld   0x%8lx   [%8s]    %3d   %s [%s]\n",
               i,
               pmod->st1.status   [i],
               pmod->st1.data     [i],
               pmod->st1.data     [i],
               &pmod->st1.raw_read [i][0],
               pmod->st1.number   [i],
               upstable [index].name,
               (upstable [index].support?" ":"*"));
    }

    printf("\n");
    printf("ST2 Command Requested = %d\n", pmod->st2.was_requested);
    printf("ST2 Command Successful = %d\n", pmod->st2.data_valid);
    printf("---------------------------------------------------------------------------------------\n");
    printf("Offset Status   Decimal Value    Hex Value    [RawData]    Parameter Num/Name\n");
    for (i = 0; i < MAXDG2; i++)
    {
        index = pmod->st2.number [i];
        printf("   %3d:   %3d     %10ld   0x%8lx   [%8s]    %3d   %s [%s]\n",
               i,
               pmod->st2.status   [i],
               pmod->st2.data     [i],
               pmod->st2.data     [i],
               &pmod->st2.raw_read [i][0],
               pmod->st2.number   [i],
               upstable [index].name,
               (upstable [index].support?" ":"*"));
    }

    printf("\n");
    printf("ST3 Command Requested = %d\n", pmod->st3.was_requested);
    printf("ST3 Command Successful = %d\n", pmod->st3.data_valid);
    printf("---------------------------------------------------------------------------------------\n");
    printf("Offset Status   Decimal Value    Hex Value    [RawData]    Parameter Num/Name\n");
    for (i = 0; i < MAXDG3; i++)
    {
        index = pmod->st3.number [i];
        printf("   %3d:   %3d     %10ld   0x%8lx   [%8s]    %3d   %s [%s]\n",
               i,
               pmod->st3.status   [i],
               pmod->st3.data     [i],
               pmod->st3.data     [i],
               &pmod->st3.raw_read [i][0],
               pmod->st3.number   [i],
               upstable [index].name,
               (upstable [index].support?" ":"*"));
    }

    printf("\n");
    printf("ST5 Command Requested = %d\n", pmod->st5.was_requested);
    printf("ST5 Command Successful = %d\n", pmod->st5.data_valid);
    printf("---------------------------------------------------------------------------------------\n");
    printf("Offset Status   Decimal Value    Hex Value    [RawData]    Parameter Num/Name\n");
    for (i = 0; i < MAXDG5; i++)
    {
        index = pmod->st5.number [i];
        printf("   %3d:   %3d     %10ld   0x%8lx   [%8s]    %3d   %s [%s]\n",
               i,
               pmod->st5.status   [i],
               pmod->st5.data     [i],
               pmod->st5.data     [i],
               &pmod->st5.raw_read [i][0],
               pmod->st5.number   [i],
               upstable [index].name,
               (upstable [index].support?" ":"*"));
    }
}

/*
**      secmodinfo                                                        
**                                                                      
**      The purpose of this routine is to list summary information      
**        for all modules allocated and referenced on this line.        
**
*/
/* not called from within this driver */

void secmodinfo (unsigned short line)
{
    SEC          *pmod;
    unsigned short   i;

    unsigned short   linecnt;
    unsigned char    buffer [200];

    linecnt = 0;

    pmod = modules [line];

    printmodhdr (buffer, linecnt);
    for (i = 0; i < NUMMODS; i++)
    {
        pmod = modules [line];
        if (pmod)
        {
            linecnt = 1;
            printf(" %2d:   %1d   %1d   %1d   %5d %5d %5d %5d %5d %5d %5d %5d %5d  %1d",
                    i,
                    pmod->was_requested,
                    pmod->is_responding,
                    pmod->cmd_force_offline,
                    pmod->no_dataout_msgs & 0x7fff,
                    pmod->no_datain_msgs & 0x7fff,
                    pmod->no_rxnoresp & 0x7fff,
                    pmod->no_txnoresp & 0x7fff,
                    pmod->no_dataerr & 0x7fff,
                    pmod->no_retries & 0x7fff,
                    pmod->no_chksumerrs & 0x7fff,
                    pmod->no_frameerrs & 0x7fff,
                    pmod->no_offlines & 0x7fff, 
                    0);
            linecnt += 1;
            if ((linecnt >= 20) && ((i + 1)  < NUMMODS))
            {
                linecnt = 1;
                printmodhdr(buffer, linecnt);
            }
        }
    }
}
/*
** only called from dump_mod_table 
*/
static void printmodhdr (unsigned char *buffer, unsigned short cnt)
{    
    taskDelay (30);
    printf ("\n");
    sprintf (buffer,"                        SEC MODULE CONFIGURATION INFO");
    printf ( "%s\n", buffer);
/*    cnt += 1; */
    sprintf (buffer,
      " MOD   --- Status ---  # Messages   ------- Comm Error Counts by Type ------- ");
    printf ( "%s\n", buffer);
    sprintf (buffer,
      " NUM   Req  Res  FOL   In    Out    NoTx  NoRx  DatEr Retry Chksm Frame Offln ");
    printf ( "%s\n", buffer);
/*    cnt += 1; */
}

/*
**                                                                
**    The purpose of this routine is to display the contents      
**    of the static variables contained within this driver.       
**                                                                
*/

/* this routine not called from within this driver */

void seclineinfo (void)
{
    short    i;
    char    buffer [132];
    char    *resrsp;
    char    *lineact;


    printf ("                                 Sec Line Information Table          \n");
    printf ("\n");
    printf (" Line        Pointer to        Resource      Resource");
    printf ("       Line      Intermessage   Activity\n");
    printf ("Number      SEC Table         Name        Response");
    printf ("     Activity      IM  .TX  TX.         Time\n");
    printf ("\n");

    for (i = 0; i < SECLINES; i++)
    {
        if(modules [i] == NULL)
        {
        continue;
        }
        if (resresponse [i] == 0)
        {
        resrsp = "FALSE";
        }
        else
        {
        resrsp = "TRUE";
        }
        if (line_activity [i] == 0)
        {
        lineact = "FALSE";
        }
        else
        {
        lineact = "TRUE";
        }
        sprintf (buffer,
                       "  %2d         %4x   %16s        %5s        %5s        %2d   %2d   %2d        %4d\n",
            i,
            (int) &modules[i],
            resnames[i],
            &resrsp[0],
            &lineact[0],
            intermsgdelay[i],
            pretxdelay[i],
            posttxdelay[i],
            (int) activity_time[i]);
        printf ( "%s\n", &buffer[0]);
    }
    return;
}

/*    End of File:    drvSec.c */
