/*devTankway.c,v 1.1.1.1 1999/08/03 14:44:58 tis4000 Exp */

/****************************************************************/
/*					                        */
/*   This software is the proprietary property of Tate	        */
/*   Integrated Systems, L.P.  Patents and copyrights are       */
/*   pending.                  				        */
/*                        					*/
/*   This software is furnished under a license and may         */
/*   be used and copied only in accordance with the terms       */
/*   of such license and with the inclusion of the above        */
/*   notice. This software or any other copies thereof   	*/
/*   may not be provided or otherwise made available to   	*/
/*   any other person. No title to and ownership of the   	*/
/*   software is hereby transferred.            		*/
/*                        					*/
/*   The information in this software is subject to      	*/
/*   change without notice and should not be construed    	*/
/*   as a commitment by Tate Integrated Systems L.P. or   	*/
/*   its third party suppliers.            			*/
/*                        					*/
/*   Tate Integrated Systems, L.P. (TIS) and its third   	*/
/*   party suppliers assume no responsibility for the   	*/
/*   use or inability to use any of its software. TIS   	*/
/*   software is provided "as is" without warranty of   	*/
/*   any kind and TIS expressly disclaims all implied   	*/
/*   warranties, including but not limited to the implied   	*/
/*   warranties of merchantability and fitness for a      	*/
/*   particular purpose.               				*/
/*                        					*/
/*   Notice: Notwithstanding any other lease or license   	*/
/*   that may pertain to or accompany the delivery of this   	*/
/*   computer software, the rights of the Government      	*/
/*   regarding its use, reproduction, and disclosure are   	*/
/*   as set forth in Section 52.227-19 of the FARS      	*/
/*   Computer Software-Restricted Rights clause.      		*/
/*                        					*/
/****************************************************************/


/****************************************************************************
 *
 *  MODULE NAME            : devTankway.c
 *  AUTHOR                 : wep
 *  DATE                   : 21 July 97
 *
 *
 * DESCRIPTION: device support for Tankway
 *
 * (based on original devModicon.c code by steel 22 Dec 93)
 *
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : 
 *  REVISION DATE          : 
 *  REVISION               : 
 *
****************************************************************************/

/****************************************************************************
 *
 * Programmer's Notes:
 *    This is a compendium of all dev**Modicon.c files in .../share/src/dev
 *    Where noted, change appropriately to create a new device support
 *    module
 *
 *
****************************************************************************/

#include   <vxWorks.h>
#include   <types.h>
#include   <stdioLib.h>

#include   <alarm.h>
#include   <cvtTable.h>
#include   <dbAccess.h>
#include   <dbDefs.h>
#include   <devSup.h>
#include   <link.h>
/*#include   <module_types.h>*/
#include   <special.h>
#include   <cvnTypes.h>   /* defines for conversion types */
#include   <rec/aiRecord.h>
#include   <rec/aoRecord.h>
#include   <rec/biRecord.h>
#include   <rec/boRecord.h>


#define  IO_TYPE     TANKWAY_IO           /* NOTE: change here */
#define  AIDRIVER    tank_aidriver         /* NOTE: change here */
#define  AODRIVER    tank_aodriver         /* NOTE: change here */
#define  AOREAD      tank_aoread           /* NOTE: change here */
#define  BIDRIVER    tank_bidriver         /* NOTE: change here */
#define  BODRIVER    tank_bodriver         /* NOTE: change here */
#define  BOREAD      tank_boread           /* NOTE: change here */

static char *io_name = "Tankway";         /* NOTE: change here */


/* Create the dset for devAiTankway */

static long ai_init_record();
static long read_ai();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   read_ai;
   DEVSUPFUN   special_linconv;
} devAiTankway = {                        /* NOTE: change here */
   6,
   NULL,
   NULL,
   ai_init_record,
   NULL,
   read_ai,
   NULL};

/* Create the dset for devAoTankway */

static long ao_init_record();
static int read_ao ();
static long write_ao();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   write_ao;
   DEVSUPFUN   special_linconv;
   DEVSUPFUN   readback_ao;
} devAoTankway = {                        /* NOTE: change here */
   7,
   NULL,
   NULL,
   ao_init_record,
   NULL,
   write_ao,
   NULL,
   NULL};

/* Create the dset for devBiTankway */

static long bi_init_record ();
static long read_bi ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   read_bi;
} devBiTankway = {                        /* NOTE: change here */
   5,
   NULL,
   NULL,
   bi_init_record,
   NULL,
   read_bi};

/* Create the dset for devBoTankway */

static long bo_init_record ();
static long write_bo ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   write_bo;
   DEVSUPFUN   readback_bo;
} devBoTankway = {                        /* NOTE: change here */
   6,
   NULL,
   NULL,
   bo_init_record,
   NULL,
   write_bo,
   NULL};


static long masks [] = {
   0x00000001, 0x00000002, 0x00000004, 0x00000008,
   0x00000010, 0x00000020, 0x00000040, 0x00000080,
   0x00000100, 0x00000200, 0x00000400, 0x00000800,
   0x00001000, 0x00002000, 0x00004000, 0x00008000,
   0x00010000, 0x00020000, 0x00040000, 0x00080000,
   0x00100000, 0x00200000, 0x00400000, 0x00800000,
   0x01000000, 0x02000000, 0x04000000, 0x08000000,
   0x10000000, 0x20000000, 0x40000000, 0x80000000,
};




/********************************** AI ************************************/

static long ai_init_record (pai)
   struct   aiRecord *pai;
{
   struct   tateio   *ptateio;
   char              message[128];
   unsigned short    value;
   int               status;
   union u_value {
      short    ivalue;
      long     lvalue;
      float    fvalue;
  }  uval;

   switch (pai->inp.type)
   {

      case (IO_TYPE):
         break;

      default:
         strcpy (message, pai->name);
         strcat (message, ": devAi");
         strcat (message, io_name);
         strcat (message, " (init_record) Illegal INP field");
         errMessage (S_db_badField, message);
         return (S_db_badField);
   }

   /* call driver so that it configures card */
   ptateio = (struct tateio *) &(pai->inp.value);
   status = AIDRIVER (pai->name,ptateio->link, ptateio->adapter, ptateio->card,
                      ptateio->signal, &uval, pai->linr, pai->resn);

    /* setting return (status) to 2 tells rec ai's process() not to convert value */
    if (status == 0)
    {
        pai->rval = uval.fvalue;
        switch(pai->linr)
        {
            case NOCONV:
            case LINEAR:
            case CNVSQRROOT:
                convertfloat (pai, uval.fvalue);
                status = 3;         /* don't convert at rec level. */
                break;
         
            case CNVINTEGER:
            case CNVDOUBLEINT:
            case CNVFLOAT:
            case CNVDOUBLEFLT:
            case CNVASCIIFLT:
                default:
                status = -1;
                break;
        }
    }
    return (status);
}

static long read_ai (pai)
   struct aiRecord   *pai;
{
   struct   tateio   *ptateio;
   int               status;
   unsigned short    value;

   union u_value {
      short    ivalue;
      long     lvalue;
      float    fvalue;
  }  uval;

   ptateio = (struct tateio *) &(pai->inp.value);
   status = AIDRIVER (pai->name,ptateio->link, ptateio->adapter, ptateio->card,
                      ptateio->signal, &uval, pai->linr, 0);

    /* setting return (status) to 2 tells rec ai's process() not to convert value */
    if (status == 0)
    {
        pai->rval = uval.fvalue;
        switch(pai->linr)
        {
            case NOCONV:
            case LINEAR:
            case CNVSQRROOT:
                convertfloat (pai, uval.fvalue);
                status = 2;         /* don't convert at rec level. */
                break;
         
            case CNVINTEGER:
            case CNVDOUBLEINT:
            case CNVFLOAT:
            case CNVDOUBLEFLT:
            case CNVASCIIFLT:
                default:
                status = -1;
                break;
        }
    }

    /* special case: if status == open thermocouple */
    if (status == -3)
    {
        if ((pai -> aenb != ALARM_DISABLED) && (pai->nsev < VALID_ALARM))
        {
            pai->nsta = OPENTC_ALARM;
            pai->nsev = VALID_ALARM;
        }
        status = 2;         /* don't convert */
    }

    if (status == -1)
    {
        if ((pai -> aenb != ALARM_DISABLED) && (pai->nsev < VALID_ALARM))
        {
            pai->nsta = READ_ALARM;
            pai->nsev = VALID_ALARM;
        }
        status = 2;         /* don't convert */
    }

    return (status);
}

static void convertfloat (pai, fvalue)
      struct aiRecord *pai;
      float fvalue;
{
    double	val;
    double      sqrt_val;
    double      temp_dbl;
    float	aslo = pai->aslo;
    float	aoff = pai->aoff;

    /* check value in instrument range */
    val = fvalue;
    if ((pai->irhi != 0) || (pai->irlo != 0))
    {
        if ((val > (pai->irhi + pai->idev)) || (val < (pai->irlo - pai->idev))) 
        {
            if (pai -> nsev < VALID_ALARM)
            {
                pai->nsta = SBAD_ALARM;
                pai->nsev = VALID_ALARM;
	    }
            return;
        }
    }

    /* adjust slope and offset */
    if (aslo != 0.0) 
    {
        val *= aslo;
    }
    if (aoff != 0.0) 
    {
        val += aoff;
    }

    switch (pai->linr)
    {
        case NOCONV:
             break;

        case LINEAR:
            /* y = mx + b, m = slope = eslo, b = offset = roff calc'd in slope_and_offset()  */
            val = (pai -> eslo * val) + pai -> roff;
            break;

        case CNVSQRROOT:
            /* val = eng units full - eng units low * square root of */
            /*       raw aval - instrument range low divided by      */
            /*       instrument range high - instrument range low    */
            /*       plus eng units low                              */

            if (pai -> irhi == pai -> irlo)
            {
                sqrt_val = 0.0;
            }
            else
            {
                temp_dbl = ( (val - pai -> irlo) / (pai -> irhi - pai ->irlo) );
                if (temp_dbl < 0.0)
                {
                    if (pai -> nsev < MAJOR_ALARM)
                    {
                        pai -> nsta = CALC_ALARM;
                        pai -> nsev = VALID_ALARM;
                    }
                    return;
                }
                sqrt_val = sqrt (temp_dbl);
            }
            val = ((pai -> eguf - pai -> egul) * sqrt_val) + pai -> egul;
            break;

        default:
            break;
    }

    /* apply smoothing algorithm only if fully init'd, since   */
    /*  we do not have a valid value in pai->val until then. */
    if ( (pai->smoo != 0.0) && (pai -> ini2) )
    {
        pai->val = val * (1.00 - pai->smoo) + (pai->val * pai->smoo);
    }
    else
    {
        pai->val = val;
    }

    pai->udf = FALSE;
    return;
}

/*********************************** AO **********************************/

static long ao_init_record (pao)
   struct   aoRecord *pao;
{
   struct   tateio   *ptateio;
   char              message[128];
   short             value;

   switch (pao->out.type)
   {

      case (IO_TYPE):
         break;

      default:
         strcpy (message, pao->name);
         strcat (message, ": devAo");
         strcat (message, io_name);
         strcat (message, " (init_record) Illegal OUT field");
         errMessage (S_db_badField,message);
         return (S_db_badField);
   }

      /* call driver so that it configures card */

   return (read_ao (pao));
}

static long write_ao (pao)
   struct aoRecord   *pao;
{
   struct   tateio   *ptateio;
   int               status;
   int               value;
   double            dvalue;


   ptateio = (struct tateio *) &(pao->out.value);
   dvalue = pao->oval;
   value = pao->rval;
   status = AODRIVER (pao->name,ptateio->link, ptateio->adapter,ptateio->card,
                      ptateio->signal,dvalue, value, pao->linr,0);
   if(status == -1)
   {
      if((pao -> aenb != ALARM_DISABLED) && (pao->nsev < VALID_ALARM))
      {
         pao->nsta = WRITE_ALARM;
         pao->nsev = VALID_ALARM;
      }
   }
   else if (status == -2)
   {
      status = 0;
      if (pao->nsev < VALID_ALARM )
      {
         pao->nsta = HW_LIMIT_ALARM;
         pao->nsev = VALID_ALARM;
      }
   }

   return (status);
}


static int read_ao (pao)
   struct   aoRecord *pao;
{
   struct   tateio   *ptateio = &pao->out.value.tateio;
   int               status;

   union u_value {
      short    ivalue;
      long     lvalue;
      float    fvalue;
      double   dvalue;
      char     afvalue[100];
   }  uval;

      /* get the value from the ao driver */

   status = AOREAD (pao->name,ptateio->link, ptateio->adapter, ptateio->card,
                    ptateio->signal, &uval,  pao->linr, &pao->resn [0]);

   if (status == 0)
   {
      switch (pao -> linr)
      {
         case ASCFLT:   /* ascii float conversion */
            pao -> val = uval.fvalue;
            break;

         case NOCONV:   /* no conversion */
         case LINEAR:   /* linear conversion */
         default:
            pao -> rval = uval.ivalue;
            break;
      }
   }

   return (status);
}


/*********************************** BI ***********************************/

static long bi_init_record (pbi)
   struct   biRecord *pbi;
{
   struct   tateio   *ptateio;
   char              message[128];
   int               status = 0;
   unsigned long     value;

   switch (pbi->inp.type)
   {
      case (IO_TYPE) :
         ptateio = (struct tateio *) &(pbi->inp.value);
/*         pbi->mask = masks[((ptateio->signal - 1) % 32)]; */
         pbi->mask = masks[((ptateio->signal) % 32)];
         status = BIDRIVER (pbi->name,ptateio->link, ptateio->adapter,
                            ptateio->card, ptateio->signal, pbi->mask, &value,
                            &pbi->resn [0]);
	 if (!status)
	    pbi -> rval = value;

         break;

      default :
         strcpy (message, pbi->name);
         strcat (message, ": devBi");
         strcat (message, io_name);
         strcat (message, " (init_record) Illegal INP field");
         errMessage (S_db_badField, message);
         status = S_db_badField;
	 break;
   }
   return (status);
}

static long read_bi(pbi)
   struct   biRecord *pbi;
{
   struct   tateio   *ptateio;
   int               status;
   unsigned long     value;

   ptateio = (struct tateio *) &(pbi->inp.value);
   status = BIDRIVER (pbi->name,ptateio->link, ptateio->adapter, ptateio->card,
                      ptateio->signal, pbi->mask, &value, 0);
   if (status == 0)
   {
      pbi->rval = value;
      return (status);
   }
   else
   {
      if ((pbi -> aenb != ALARM_DISABLED) && (pbi->nsev < VALID_ALARM))
      {
         pbi->nsta = READ_ALARM;
         pbi->nsev = VALID_ALARM;
      }
      return (2);
   }
}


/*********************************** BO ***********************************/

static long bo_init_record (pbo)
   struct   boRecord *pbo;
{
   struct   tateio   *ptateio;
   char              message[128];
   unsigned long     value;
   int               status = 0;

   switch (pbo->out.type)
   {
      case (IO_TYPE) :
         ptateio = (struct tateio *) &(pbo->out.value);

           /* read the value via the tankway driver */

         pbo->mask = masks [((ptateio->signal - 1) % 32)];
         status = BOREAD (pbo->name,ptateio->link, ptateio->adapter, 
                          ptateio->card, ptateio->signal, &value, pbo->mask, 
                          &pbo->resn [0]);
         if(status == 0)
         {
            pbo->rval = value;
         }
         break;

      default :
         strcpy (message, pbo->name);
         strcat (message, ": devBo");
         strcat (message, io_name);
         strcat (message, " (init_record) Illegal OUT field");
         errMessage (S_db_badField, message);
         status = S_db_badField;
	 break;
   }
   return (status);
}

static long write_bo (pbo)
   struct   boRecord *pbo;
{
   struct   tateio   *ptateio;
   int               status;

   ptateio = (struct tateio *) &(pbo->out.value);
   status = BODRIVER (pbo->name,ptateio->link, ptateio->adapter, ptateio->card,
                      ptateio->signal, pbo->rval, pbo->mask, 0);
   if (status != 0)
   {
      if ((pbo -> aenb != ALARM_DISABLED) && (pbo -> nsev < VALID_ALARM))
      {
         pbo->nsta = WRITE_ALARM;
         pbo->nsev = VALID_ALARM;
      }
   }
   return (status);
}

