/*drvTosg3.h,v 1.1.1.1 1999/08/03 14:45:08 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
** File:    drvTosg3.h
** Edit:    10/13/95   wep: initial version                                             
**                                                                                      
**	This file contains the data structures used by the Toshiba G3 Inverter Driver   
**	I/O driver.                                                            	       	
**                                                                             	      	
*/

#define TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d

/* number of lines of Toshiba G3 Inverter Units  */
#define	    G3LINES	20
/* number of Toshiba G3 Inverter Units per line  */
#define	    NUMG3UNITS	100
/* number of Toshiba G3 Banks supported by driver  */
#define	    NUMG3BANKS	2

/*                                                                           	  
**	The IOC will assume that the data to be exchanged with each G3 Unit will  
**	come from a fixed number of coils, digital inputs, analog inputs          
**	or registers. It will read these in groups to minimize the amount of      
**	communications overhead. It will also try to build as large a fetch       
**	message as possible & required by remembering which consecutive groups    
**	have been requested by the SCAN task.                                     
*/

#define PSEUDOCARD    3333          /* Card number for pseudo i/o        */

#define SPECIALCASE   99            /* special signal number for oddly masked locations */

/* define base addresses */
#define MINBANKCARD       960       /* min address accessable in bank 0  */
#define MAXBANKCARD      1440       /* max address accessable in bank 0  */
#define MINBANK0CARD  MINBANKCARD   /* min address accessable in bank 0  */
#define MAXBANK0CARD  MAXBANKCARD   /* max address accessable in bank 0  */
#define MINBANK1CARD  MINBANKCARD   /* min address accessable in bank 1  */
#define MAXBANK1CARD  MAXBANKCARD   /* max address accessable in bank 1  */


/* Number of items */
#define BANK0DATASIZE (1 + MAXBANK0CARD - MINBANK0CARD)/2   /* word size */
#define BANK1DATASIZE (1 + MAXBANK1CARD - MINBANK1CARD)/2   /* word size */


/*
**	Conversion types for analog inputs	*/
/*
*/
/* The following must match the definition in recChoice.ascii */
#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2

#define		AICNVINTEGER	10
#define		AICNVDOUBLEINT	11
#define		AICNVFLOAT	12
#define		AICNVDOUBLEFLT	13
#define		AICNVASCIIFLT	14

#define		AICNVSQRROOT    15

#define		AICNVSTATUS	15
#define		AICNVTIME	16
#define		AICNVASCII_2	17
#define		AICNVASCII_4	18
#define		AICNVASCII_12	19
#define		AICNVASCII_20	20
#define		MAXCONV		20


#define WATCHDOGTIMEOUT      10*60 /* watchdog monitor timeout.  */

/*
**	Communications I/O Structures
**
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	256
#define		MAXTXMSG	256


typedef	struct {
	unsigned short		rxstate;
	unsigned short		fixedlen;
	unsigned short		rxmsglenexpected;
        unsigned short          rxmsglen;
       	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
       	unsigned char		rxbuf[MAXRXMSG];
		} 	RXLINE;


typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
        unsigned char           address;
        unsigned char           function;
        unsigned char           number;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;
/*
**	Communications I/O Constants
**
*/

#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		60	/* 16 ms ticks 				*/
#define		ACTIVITYTIME		300	/* must see activity in 5 seconds 	*/

#define         ONERETRY                0x8000	/* OR'ed into line number for xfermsg 	*/

#define         RXIGNORE                0
#define         RXHEADER                1
#define         RXBODY                  2


#define		TXGOOD			1
#define		TXBAD			2
#define		IO_TIMEOUT		3	/* Code returned by timer 		*/


#define		TXTIMEOUT		60	/* 16 ms ticks: one second              */

#define		MAXRETRYS		3	/* retransmission trys 			*/

#define		MAXCFGCK		8	/* Try to Bring online each cycle 	*/
#define		BRINGBACKCNT		4	/* Poll loops per cycle 		*/

#define		XFERGOOD		1
#define		XFERBAD			2
#define		XFERREJECT		3

#define		RESNAMELEN		16	/* length of a resource name 		*/
#define		SLWWINDOW		900	/* # of ticks to allow slave to try 	*/


/*
** Output Queue Management Structures for Modbus Protocol
**
*/

#define         AOUTTYPE        1
#define         BOUTTYPE        2

#define         OUTQSIZE        128

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	unsigned short  aovalue;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bovalue;
	long            bomask;
		} BOUT, *BOUTPTR;

typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque[OUTQSIZE];
		} OUTPUTQUE;
/*
**	Internal IOC Structure for Toshiba G3 Inverter Unit
**
*/
typedef	struct {
    unsigned short     was_requested;
    unsigned short     data_valid;
    unsigned short     last_read;
               } BANK0;

typedef	struct {
    unsigned short     was_requested;
    unsigned short     data_valid;
    unsigned short     last_read;
               } BANK1;



/*******************/
/*******************/

#define  G3HOLDOFFTICKS  (5*60) /* hold off commands for 5 seconds after sending a trip clear command */


typedef	struct {
    unsigned short  was_requested;      /* 0  The scan task has accesses this RTU 	*/
    unsigned short  is_responding;      /* 1  The RTU is responding 			*/
    unsigned short  holdflag;           /*    The RTU data is in hold 			*/
    unsigned short  no_dataout_msgs;    /* 5  Count Output Msgs 			*/
    unsigned short  no_datain_msgs;     /* 6  Count Data Requests 			*/
    unsigned short  no_rxnoresp;        /* 7  Count No Responses 			*/
    unsigned short  no_txnoresp;        /* 8  Count No Responses 			*/
    unsigned short  no_dataerr;         /* 9  Count Message Echo Errors 		*/
    unsigned short  no_retries;         /* 10 Count Retries 				*/
    unsigned short  no_rxhwerrs;        /* 11 Count Comm Hardware Errors 		*/
    unsigned short  no_txhwerrs;        /* 12 Count Comm Hardware Errors 		*/
    unsigned short  no_chksumerrs;      /* 13 Count Checksum Errors 			*/
    unsigned short  no_offlines;        /* 14 Count Off Line Transitions		*/
    unsigned short  cmd_reset_stat;     /* 15 Comand to reset stats			*/
    unsigned short  cmd_force_offline;  /* 16 Command to Force Offline 			*/
    unsigned short  badoutcnt;          /* 17 Count Failed output commands 		*/

    unsigned short  holdoff;            /* flag indicating driver is not talking to unit*/
    unsigned short  holdofftick;        /* tick time stored when holdoff first set TRUE */

    unsigned short  tscanpass;          /* Total number of scanpasses. incremented at   */
                                        /*  the start of each plc scan.                 */
    unsigned short  msgs;               /* Count of messages which got good responses.  */
                                        /*  incremented on each exit from 'good' msgxfer*/
    unsigned long   responsetime;       /* Total time(ticks) waiting for good responses.*/
                                        /* accumu (exit-entry) time for 'good' msgxfers */
    unsigned short  tripped;
    unsigned short  tripstatevalid;

    unsigned short   tempvalid;
    unsigned short   tempvalue;

    BANK0    bank0    [BANK0DATASIZE];
    BANK1    bank1    [BANK1DATASIZE];

} G3UNIT;

/*
** Toshiba G3 Protocol function codes
*/

#define ADDRSET          'A'   /* Address Setting Command                                     */
#define ADDRSETLEN        8

#define READWORD         'R'   /* Read Word Command                                           */
#define READWORDLEN       4

#define WRITEWORD        'W'   /* Write Word Command                                          */
#define WRITEWORDLEN      8

#define MASKSET          'M'   /* Mask Setting Command                                        */
#define MASKSETLEN        8

#define BANKSET          'B'   /* Bank Setting Command                                        */
#define BANKSETLEN        5

#define TESTCOMM         'T'   /* Test Communications Command                                 */
#define TESTCOMMLEN       5    /* Test Communications Command Length                          */

#define BODYLENGTH        9    /* Length of body of the reply message (checksummed part)      */
#define REPLYLENGTH      13    /* Length of total reply message                               */

/* special case */

#define NACKRESPONSE     'N'   /* Unexpected Command Response from one of the above Commands  */






/* Command Results */

#define CMDDONE           0x00     /* the command was sent and the response was error free.   */
#define CMDFAIL           0x01     /* there were errors in sending or receiving the command.  */
#define CMDCORRUPT        0x02     /* the response msg was error free, but data was corrupted.*/
#define CMDREJECT         0x03     /* driver rejected command due to unsupported data type.   */
#define CMDBUSY           0x05     /* The command was sent, but the device is busy.           */

/* task definitions */

#define G3_STACK          8000     /* Stack size              */
#define	G3_PRIORITY        152     /* default priority        */
#define	G3_TASK_OPTIONS      8     /* Floating point          */


/*
** Mutual-Exclusion Control for diagnostic/data radio channel sharing */
/*
*/

#define	MUXTIMEOUT       60*30    /* 16.67 ms ticks (60Hz): 30 seconds */


typedef struct {
        unsigned long   TXintcount;
        unsigned long   RXintcount;
        unsigned long   msgxfercount;
        unsigned long   startingtick;
               }        LINESTATS;

/* Trace Type Definitions */

#define READBANK0            0x0001
#define READBANK1            0x0002
#define WRITEBANK0           0x0004


/* Trace Bit Definitions */

#define ALLTRAFFICTRACE      0x0001
#define G3NACKTRACE          0x0002     /* G3 NACK errors.                                   */
#define CONFIGURATIONTRACE   0x0010     /* trace configuration errors in driver calls.       */
#define RUNTIMEERRORTRACE    0x0020     /* trace run time errors in driver calls.            */
#define CALLREJECTTRACE      0x0040     /* trace non-error based driver call rejections.     */

#define ASCIITRAFFICTRACE    0x4000     /* when combined with ALLTRAFFICTRACE,               */
                                        /*  prints comm traffic in ascii.                    */

/*	End of File:	drvTosg3.h */
