/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/



/************************************************************************
*
*    alarmTalkIO.c
*
*    modifications
*    -------------
*  .01	09-22-96	ems	creation date
*
************************************************************************/


#include <string.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/termios.h>
#include <sys/ttold.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <malloc.h>
#ifndef AXP
#include <net/if.h>
#include <sys/sockio.h>
#include <sys/filio.h>
#endif

#ifdef MIPSEL
#include "error.h"
#endif

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "dbDefs.h"		/* .../share/epicsH */
#include "shareMem.h"		/* .../share/src/alarm */
#include "alarmLib.h"		/* .../share/src/alarm */
#include "cadef.h"		/* .../share/epicsH */
#include "messages.h"		/* .../share/src/alarm */
#include "alarmDefs.h"		/* .../share/src/alarm */
#include "alarmvu.h"		/* .../share/src/alarmVuCC */
#include "slmDefs.h"		/* .../share/epicsH */
#include "OperPriv.h"		/* .../share/src/libtis */
#include "tsDefs.h"		/* .../share/epicsH, TS_STAMP_def */
#include "msgMon.h"		/* .../share/epicsH */
#include "alarmTalk.h"

extern char	decTalkRecvBuffer[DECTALK_MSG_SIZE + 1];
extern char	newDecTalkMsg[DECTALK_MSG_SIZE + 1];
extern char	newAlarmMsg[ALARM_MSG_SIZE + 1];
extern char	decTalkName[64];

extern int	alarmMsgCnt;
extern int	alarmPrtListenFd;
extern int	alarmPrtFd;
extern int	alarmMgrFd;
extern int	decTalkFd;
extern int	decTalkMsgCnt;
extern int	debug;
extern int	dbOpened;
extern int	mgrAvailable;
extern int	firstDbOpen;
extern int	decWriteFailed;
extern int	settty;
extern short	decTalkPort;

extern struct sockaddr_in alPrtPort;	/* for alarmPrinter */
extern struct termios	savetty;
extern time_t		configLastModified;
extern ALARMCFG_PTR	currentAlarmConfig;
extern ALARMCFG		alarmConfigs[MAX_ALARM_CFGS];
extern CODE4		cb;
extern DATA4		*alarmTalkDb;



/***************************************************
*
*    acceptAlarmPrinter ()
*	accept a connection from the alarmPrinter
*
***************************************************/

int	acceptAlarmPrinter (void)
{
    int		addrSize;

    addrSize = sizeof (alPrtPort);
    alarmPrtFd = accept (alarmPrtListenFd, (struct sockaddr *)&alPrtPort, 
			 &addrSize);
    if (alarmPrtFd == -1)
    {
	tisSysLog (0,0, "acceptAlarmPrinter: accept failed (%m)");
	return (1);
    }
    if (setNoBlock (alarmPrtFd))
    {
	tisSysLog (0,0, "acceptAlarmPrinter; non-blocking IO failed");
	return (1);
    }
    if (answerAlarmPrinter())
    {
	return (1);
    }

if(debug)
{
tisSysLog(0,0,
"acceptAlarmPrinter: return (0), fd=%d",
alarmPrtFd);
}
    return (0);
}


/***************************************************************
*
*    answerAlarmPrinter ()
*	when alarmPrinter send me something, I must respond
*	since it hangs on a read select before it does a 
*	send again
*
***************************************************************/

int	answerAlarmPrinter (void)
{
    int		bytesSent;
    char	msg[2];

    msg[0] = 'D';
    msg[1] = NULL;
    bytesSent = send (alarmPrtFd, msg, strlen (msg), 0);
    if (bytesSent != strlen (msg))
    {
	tisSysLog (0,0, "answerAlarmPrinter: send failed (%m)");
	return (1);
    }

    return (0);
}


/*********************************************************************
*
*    connectToAlarmMgr ()
*	connnect to the alarm manager
*	NOTE: MSGPORT is defined in 
*		.../share/src/newalarm/alarmvu.h as 8583
*	      VUPORT is defined in 
*		.../share/src/alarm/alarmDefs.h as 8583
*	      this is the port that a program uses so that can 
*	      look like an alarmVu to the alarmMgr, you would think 
*	      it would be defined in 1 place
*
**********************************************************************/

int	connectToAlarmMgr (void)
{
    struct hostent	*hostval;
    struct sockaddr_in	setPort;
    short		msgPort = MSGPORT;

    /* create a socket */

    if ((alarmMgrFd = socket (AF_INET, SOCK_STREAM, 0)) == -1)
    {
	tisSysLog (0,0, "connectToAlarmMgr: can't create socket (%m)");
	return (1);
    }

    /* make connection to alarmMgr's existing socket */

    memset ((char*)&setPort, 0, sizeof (setPort));
    setPort.sin_family = AF_INET;
    hostval = gethostbyname ("localhost");
    if (hostval == NULL)
    {
	tisSysLog (0,0, "connectToAlarmMgr: can't get hostname (%m)");
	return (1);
    }
    setPort.sin_addr.s_addr = **((long **)(hostval -> h_addr_list));
    setPort.sin_port = htons(msgPort);
    if ((connect (alarmMgrFd, (struct sockaddr *)&setPort, 
						sizeof (setPort))) == -1)
    {
	tisSysLog (0,0, "connectToAlarmMgr: failed to connect to alarmMgr (%m)");
	return (1);
    }

    if (setNoBlock (alarmMgrFd))
    {
	tisSysLog (0,0, "connectToAlarmMgr: failed to set non-blocking IO (%m)");
	return (1);
    }

if (debug)
{
tisSysLog (0,0,
"connected to alarmMgr: return (0)");
}
    return (0);
}


/****************************************
*
*    createAlarmPrtSocket ()
*	connect to the alarmPrinter
*
***************************************/

int	createAlarmPrtSocket (void)
{
    int			optVal = 1;
    short		decTalkPort = DEC_TALK_PORT;

    /* create a socket */

    if ((alarmPrtListenFd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
    {
	tisSysLog (0,0, "createAlarmPrtSocket: can't create socket (%m)");
	return (1);
    }

    /* make it so that if we die, TCP will allow use to run again
       without TCP waiting before it realizes we have died
    */
    if ((setsockopt (alarmPrtListenFd, SOL_SOCKET, SO_REUSEADDR, (caddr_t)&optVal, 
							sizeof (optVal))) == -1)
    {
	tisSysLog (0,0, "createAlarmPrtSocket: can't set REUSEADDR (%m)");
	return (1);
    }

    /* bind to alarmPrinter port */

    memset ((char*)&alPrtPort, 0, sizeof (alPrtPort));
    alPrtPort.sin_family = AF_INET;
    alPrtPort.sin_addr.s_addr = htonl(INADDR_ANY);
    alPrtPort.sin_port = htons(decTalkPort);
    if ((bind (alarmPrtListenFd, (struct sockaddr *)&alPrtPort, sizeof (alPrtPort))) == -1)
    {
	tisSysLog (0,0, "createAlarmPrtSocket: can't bind to port (%m)");
	return (1);
    }

    /* enable it for connection */

    if ((listen (alarmPrtListenFd, 2)) == -1)
    {
	tisSysLog (0,0, "createAlarmPrtSocket: can't listen on socket (%m)");
	return (1);
    }

if (debug)
{
tisSysLog (0,0,
"createAlarmPrtSocket: return (0)");
}

    return (0);
}


/*********************************************
*
*    fflushDecTalkQue ()
*	clean up any messages from decTalk
*
*********************************************/

void	fflushDecTalkQue (void)
{
    int			numread;
    int			maxFd = -1;
    int			status;
    int			numTimes = 3;
    fd_set		chanmask;
    struct timeval	selectWait;	/* timeout setting for select() */

    maxFd = decTalkFd;

    while (numTimes > 0)
    {
	FD_ZERO(&chanmask);
	FD_SET(decTalkFd, &chanmask);

	selectWait.tv_sec = 5;
	selectWait.tv_usec = 0;
	status = select (maxFd + 1, &chanmask, NULL, NULL, &selectWait);
	if (status == -1)
	{
	    tisSysLog (0,0, "fflushDecTalkQue: select error (%m)");
	    return;
	}
	if (status == 0)
	{
	    numTimes--;
	    continue;
	}
	memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
	numread = read (decTalkFd, newDecTalkMsg, sizeof (newDecTalkMsg));
if(debug)
{
tisSysLog(0,0,
"fflush, numread = %d,msg = %s",
numread,convertDecTalkMsg (newDecTalkMsg));
}
    }
    memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
    memset ((char *)&decTalkRecvBuffer[0], 0, sizeof (decTalkRecvBuffer));
    decTalkMsgCnt = 0;

if(debug)
{
tisSysLog(0,0,
"fflush return");
}
    return;
}


/******************************************************
*
*    initDecTalk ()
*	open the decTalk and send some initialization 
*	sequences to it
*
*******************************************************/

int	initDecTalk (void)
{

    if ((strstr (decTalkName, "/dev/")) != NULL)
    {
	if (openSerialDecTalk ())
	{
	    return (1);
	}
    }
    else
    {
	if (openNetworkDecTalk ())
	{
	    return (1);
	}
    }

    if (decTalkReset ())
    {
	return (1);
    }

if (debug)
{
tisSysLog (0,0,
"initDecTalk: return (0)");
}

    return (0);
}


/********************************************************
*
*    openAlarmTalkDbf ()
*	open the alarmTalk config file, also init some
*	codebase stuff
*
********************************************************/

int	openAlarmTalkDbf (void)
{
    long	count = 0;
    int		status = 0;
    int		configIdx;
    int		nextAuto;
    char	sevr[ALARMSEVER_SZ + 1];
    struct stat	statbuff;
    FIELD4	*alarmSevr, *annunciation, *spkrDevice;
    FIELD4	*pageDevice, *autoDevice, *conNumber;
    FIELD4	*conRedial, *pageNumber, *pageRedial;
    FIELD4	*autoNumbr1, *autoNumbr2, *autoNumbr3;
    FIELD4	*autoNumbr4, *autoNumbr5, *autoNumbr6;
    FIELD4	*autoNumbr7, *autoNumbr8, *autoFailAnsTimeout;
    FIELD4	*autoRestart, *idTimeout, *repackTimeout;

    dbOpened = 0;
    configLastModified = 0;
    memset ((char *)&alarmConfigs[0], 0, sizeof (alarmConfigs));

    status = stat (AT_DATABASE, &statbuff);
    if (status == -1)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: stat on %s failed (%m)",AT_DATABASE);
	return (1);
    }

    if (firstDbOpen)
    {
	firstDbOpen = 0;
	d4init (&cb);
#if defined(USECB63) || defined(WIN32)
	cb.errGo = 0;
	cb.errOpen = 0;
	cb.errCreate = 0;
	cb.errFieldName = 0;
#else
	cb.go_error = 0;
	cb.open_error = 0;
	cb.create_error = 0;
	cb.field_name_error = 0;
#endif	USECB63
    }
    if ( (alarmTalkDb = d4open (&cb, AT_DATABASE)) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: failed to open %s", AT_DATABASE);
	return (1);
    }
    dbOpened = 1;
    count = d4reccount (alarmTalkDb);
    if (count == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: no records");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((alarmSevr = d4field (alarmTalkDb, "ALARMSEVER")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find ALARMSEVER field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((annunciation = d4field (alarmTalkDb, "ANNUNCTION")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find ANNUNCTION field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((spkrDevice = d4field (alarmTalkDb, "SPKRDEVICE")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find SPKRDEVICE field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((pageDevice = d4field (alarmTalkDb, "PAGEDEVICE")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find PAGEDEVICE field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoDevice = d4field (alarmTalkDb, "AUTODEVICE")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUTODEVICE field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((conNumber = d4field (alarmTalkDb, "CONNUMBER")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find CONNUMBER field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((conRedial = d4field (alarmTalkDb, "CONREDIAL")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find CONREDIAL field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((pageNumber = d4field (alarmTalkDb, "PAGENUMBER")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find PAGENUMBER field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((pageRedial = d4field (alarmTalkDb, "PAGEREDIAL")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find PAGEREDIAL field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoNumbr1 = d4field (alarmTalkDb, "AUTONUMBR1")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUTONUMBR1 field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoNumbr2 = d4field (alarmTalkDb, "AUTONUMBR2")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUTONUMBR2 field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoNumbr3 = d4field (alarmTalkDb, "AUTONUMBR3")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUTONUMBR3 field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoNumbr4 = d4field (alarmTalkDb, "AUTONUMBR4")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUTONUMBR4 field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoNumbr5 = d4field (alarmTalkDb, "AUTONUMBR5")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUTONUMBR5 field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoNumbr6 = d4field (alarmTalkDb, "AUTONUMBR6")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUTONUMBR6 field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoNumbr7 = d4field (alarmTalkDb, "AUTONUMBR7")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUTONUMBR7 field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoNumbr8 = d4field (alarmTalkDb, "AUTONUMBR8")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUTONUMBR8 field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoFailAnsTimeout = d4field (alarmTalkDb, "AUFAILANST")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUFAILANST field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((autoRestart = d4field (alarmTalkDb, "AUTORESTRT")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find AUTORESTRT field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((idTimeout = d4field (alarmTalkDb, "IDTIMEOUT")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find IDTIMEOUT field");
	d4close (alarmTalkDb);
	dbOpened = 0;
	return (1);
    }
    if ((repackTimeout = d4field (alarmTalkDb, "REPACKTOUT")) == 0)
    {
	tisSysLog (0,0, "openAlarmTalkDbf: did not find REPACKTOUT field");
	d4close (alarmTalkDb); dbOpened = 0;
	return (1);
    }

    /* read in config data */

    for (d4top (alarmTalkDb); !d4eof (alarmTalkDb); d4skip (alarmTalkDb, 1))
    {
	strncpy (sevr, f4str (alarmSevr), ALARMSEVER_SZ);
	sevr[ALARMSEVER_SZ] = NULL;
	removeTrailingSpaces (sevr);
	if (sevr[1] == 'A')
	    configIdx = MAJOR_CFG;
	else if (sevr[1] == 'I')
	    configIdx = MINOR_CFG;
	else if (sevr[1] == 'N')
	    configIdx = INFO_CFG;
	else
	{
	    tisSysLog (0,0, "openAlarmTalkDbf: unknown severity %s", sevr);
	    d4close (alarmTalkDb);
	    dbOpened = 0;
	    return (1);
	}
	strncpy (alarmConfigs[configIdx].alarmSevr, f4str (alarmSevr), ALARMSEVER_SZ);
	alarmConfigs[configIdx].alarmSevr[ALARMSEVER_SZ] = NULL;
	removeTrailingSpaces (alarmConfigs[configIdx].alarmSevr);

	strncpy (alarmConfigs[configIdx].annunciation, f4str (annunciation), ANNUNCTION_SZ);
	alarmConfigs[configIdx].annunciation[ANNUNCTION_SZ] = NULL;
	removeTrailingSpaces (alarmConfigs[configIdx].annunciation);

	if (f4char (spkrDevice) == 'T')
	    alarmConfigs[configIdx].spkrDevice = 1;
	else
	    alarmConfigs[configIdx].spkrDevice = 0;

	if (f4char (pageDevice) == 'T')
	    alarmConfigs[configIdx].pageDevice = 1;
	else
	    alarmConfigs[configIdx].pageDevice = 0;

	if (f4char (autoDevice) == 'T')
	    alarmConfigs[configIdx].autoDevice = 1;
	else
	    alarmConfigs[configIdx].autoDevice = 0;

	strncpy (alarmConfigs[configIdx].conNumber, f4str (conNumber), PHONENUMBR_SZ);
	alarmConfigs[configIdx].conNumber[PHONENUMBR_SZ] = NULL;
	removeTrailingSpaces (alarmConfigs[configIdx].conNumber);

	alarmConfigs[configIdx].conRedial = f4int (conRedial);

	strncpy (alarmConfigs[configIdx].pageNumber, f4str (pageNumber), PHONENUMBR_SZ);
	alarmConfigs[configIdx].pageNumber[PHONENUMBR_SZ] = NULL;
	removeTrailingSpaces (alarmConfigs[configIdx].pageNumber);

	alarmConfigs[configIdx].pageRedial = f4int (pageRedial);

	nextAuto = 0;
	strncpy ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto], f4str (autoNumbr1), 
								PHONENUMBR_SZ);
	alarmConfigs[configIdx].autoNumbrs[nextAuto][PHONENUMBR_SZ] = NULL;
	removeTrailingSpaces ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto]);
	nextAuto++;

	strncpy ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto], f4str (autoNumbr2), 
								PHONENUMBR_SZ);
	alarmConfigs[configIdx].autoNumbrs[nextAuto][PHONENUMBR_SZ] = NULL;
	removeTrailingSpaces ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto]);
	nextAuto++;

	strncpy ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto], f4str (autoNumbr3), 
								PHONENUMBR_SZ);
	alarmConfigs[configIdx].autoNumbrs[nextAuto][PHONENUMBR_SZ] = NULL;
	removeTrailingSpaces ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto]);
	nextAuto++;

	strncpy ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto], f4str (autoNumbr4), 
								PHONENUMBR_SZ);
	alarmConfigs[configIdx].autoNumbrs[nextAuto][PHONENUMBR_SZ] = NULL;
	removeTrailingSpaces ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto]);
	nextAuto++;

	strncpy ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto], f4str (autoNumbr5), 
								PHONENUMBR_SZ);
	alarmConfigs[configIdx].autoNumbrs[nextAuto][PHONENUMBR_SZ] = NULL;
	removeTrailingSpaces ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto]);
	nextAuto++;

	strncpy ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto], f4str (autoNumbr6), 
								PHONENUMBR_SZ);
	alarmConfigs[configIdx].autoNumbrs[nextAuto][PHONENUMBR_SZ] = NULL;
	removeTrailingSpaces ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto]);
	nextAuto++;

	strncpy ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto], f4str (autoNumbr7), 
								PHONENUMBR_SZ);
	alarmConfigs[configIdx].autoNumbrs[nextAuto][PHONENUMBR_SZ] = NULL;
	removeTrailingSpaces ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto]);
	nextAuto++;

	strncpy ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto], f4str (autoNumbr8), 
								PHONENUMBR_SZ);
	alarmConfigs[configIdx].autoNumbrs[nextAuto][PHONENUMBR_SZ] = NULL;
	removeTrailingSpaces ((char *)&alarmConfigs[configIdx].autoNumbrs[nextAuto]);
	nextAuto++;

	alarmConfigs[configIdx].autoFailAnsTimeout = f4int (autoFailAnsTimeout);
	alarmConfigs[configIdx].autoRestart = f4int (autoRestart);
	alarmConfigs[configIdx].idTimeout = f4int (idTimeout);
	alarmConfigs[configIdx].repackTimeout = f4int (repackTimeout);

    }  /* for () */

    configLastModified = statbuff.st_mtime;
    d4close (alarmTalkDb);
    dbOpened = 0;

if(debug)
{
tisSysLog(0,0,
"openAlarmTalkDbf: return (0)");
}
    return (0);
}


/******************************************************
*
*    openNetworkDecTalk ()
*	open the decTalk on the network
*
******************************************************/

int	openNetworkDecTalk (void)
{
    char		*pChar;
    long		hAddr;
    int			numTries;
    int			connected;
    struct sockaddr_in	netPort;
    struct hostent	*netDecTalk;

    if ((pChar = strchr (decTalkName, ':')) == NULL)
    {
	tisSysLog (0,0, "openNetworkdecTalk, usage: alarmTalk netname:[nnnn] (%s)",
		decTalkName);
	return (1);
    } 
    *pChar = NULL;
    pChar++;
    if (*pChar == NULL)
    {
	tisSysLog (0,0, "openNetworkdecTalk, usage: alarmTalk netname:[nnnn] (%s)",
		decTalkName);
	return (1);
    }
    decTalkPort = atoi (pChar);
    connected = 0;
    for (numTries = 0; numTries < 3; numTries++)
    {
	if ( (decTalkFd = socket (AF_INET, SOCK_STREAM, 0)) == -1)
	{
	    tisSysLog (0,0, "openNetworkDecTalk: can't create socket (%m)");
	    return (1);
	}
	if ((netDecTalk = gethostbyname (decTalkName)) == NULL)
	{
	    tisSysLog (0,0, "openNetworkDecTalk: nothing know about network decTalk %s (%m)",
		    decTalkName);
	    return (1);
	}
	memset ((char *)&netPort, 0, sizeof (struct sockaddr_in));
	netPort.sin_family = AF_INET;
	hAddr = *(long *)netDecTalk -> h_addr_list[0];
	netPort.sin_addr.s_addr = htonl (hAddr);
	netPort.sin_port = htons (decTalkPort);

	if ((connect (decTalkFd, (struct sockaddr *)&netPort, sizeof (struct sockaddr))) == -1)
	{
	    close (decTalkFd);
	    decTalkFd = -1;
	    sleep (5);
	}
	else
	{
	    connected = 1;
	    break;
	}
    }
    if (connected == 0)
    {
	tisSysLog (0,0, "openNetworkDecTalk: connect error (%m)");
	return (1);
    }
    if (setNoBlock (decTalkFd))
    {
	tisSysLog (0,0, "openNetworkDecTalk: failed to set non-blocking IO (%m)");
	return (1);
    }

if(debug)
{
tisSysLog(0,0,
"openNetworkDecTalk: return (0), fd = %d",
decTalkFd);
}
    return (0);
}


/******************************************************
*
*    openSerialDecTalk ()
*	open the decTalk on a serial line
*
*******************************************************/

int	openSerialDecTalk (void)
{
    int			i;
    int			numTries;
    int			opened;
    struct termios	ti;

    settty = 0;
    opened = 0;
    for (numTries = 0; numTries < 3; numTries++)
    {
	if ((decTalkFd = open (decTalkName, O_RDWR | O_NDELAY)) < 0)
	{
	    sleep (5);
	}
	else
	{
	    opened = 1;
	    break;
	}
    }
    if (opened == 0)
    {
	tisSysLog (0,0, "openSerialDecTalk: failed to open the decTalk %s (%m)",
		decTalkName);
	return (1);
    }

    memset ((char *)&ti, 0, sizeof (struct termios));

    if (tcgetattr (decTalkFd, &ti) < 0)
    {
	tisSysLog (0,0, "openSerialDecTalk: failed to get port settings (%m)");
	return (1);
    }

    /* save the original settings to restore later */
    savetty = ti;

    /* now that the tty is open set up its characteristics */

    /* set input modes */
    ti.c_iflag |= IGNBRK;	/* ignore breaks */
    ti.c_iflag |= IGNPAR;	/* ignore parity errors */
    ti.c_iflag |= ISTRIP;	/* strip the 8th bit */
    ti.c_iflag |= IXON;		/* set start/stop output control */
    ti.c_iflag &= ~IMAXBEL;	/* no bell */
    ti.c_iflag &= ~IGNCR;	/* don't ignore CR */
    ti.c_iflag &= ~ICRNL;	/* dont CR -> NL */

    /* set output modes */
    ti.c_oflag = 0;

    /* set control modes */
    /* set output baud rate, this also sets the input baud rate
       since we are not specify an input baud rate diff from
       the output
    */
    if (cfsetospeed (&ti, B9600) < 0)
    {
	tisSysLog (0,0, "openSerialDecTalk: failed to set output baud rate (%m)");
	return (1);
    }
    ti.c_cflag |= CREAD;	/* enable receiver */
    ti.c_cflag |= CLOCAL;	/* direct control, no modem control */

    /* set line discipline modes */
    ti.c_lflag = 0;

    /* set controls:
       min chars requested or # chars available
       return w/o waiting for more chars
    */
    ti.c_cc[VMIN] = 0;
    ti.c_cc[VTIME] = 0;

    /* set new line attributes NOW */
    if (tcsetattr (decTalkFd, TCSANOW, &ti) < 0)
    {
	tisSysLog (0,0, "openSerialDecTalk: failed to set tty characteristics (%m)");
	return (1);
    }

    /* set exclusive use of this tty, the arg i is ignored for this call */
    if (ioctl (decTalkFd, TIOCEXCL, &i) < 0)
    {
	tisSysLog (0,0, "openSerialDecTalk: failed to set exclusive tty use (%m)");
	return (1);
    }

    /* turn on soft carrier flag */
    /*
    i = 1;
    if (ioctl (decTalkFd, TIOCSSOFTCAR, &i) < 0)
    {
	tisSysLog (0,0, "openSerialDecTalk: failed to set carrier flag (%m)");
	return (1);
    }
    */

    settty = 1;

if (debug)
{
tisSysLog (0,0,
"openSerialDecTalk: return (0), fd = %d",
decTalkFd);
}

    return (0);
}


/*********************************************
*
*    readAlarmMgr ()
*	process msgs from the alarm manager
*
*********************************************/

int	readAlarmMgr (void)
{
    int			numbytes;
    TypeOnlyMessage	alarmMgrMsg;

    numbytes = recv (alarmMgrFd, (char *)&alarmMgrMsg, sizeof (TypeOnlyMessage), 0);
    if (numbytes == -1)
    {
	tisSysLog (0,0, "readAlarmMgr: read error (%m)");
	return (1);
    }

    if (numbytes == 0)
    {
if (debug)
{
tisSysLog (0,0,
"readAlamrMgr, 0 length read");
}
	return (1);
    }

    switch (alarmMgrMsg.msgType)
    {
	case UNAVAILABLE:
	    send (alarmMgrFd, (char *)&alarmMgrMsg, sizeof (TypeOnlyMessage), 0);
	    mgrAvailable = 0;
	    break;

	case AVAILABLE:
	    if (mgrAvailable == 0)
	    {
		updateActiveList ();
	    }
	    mgrAvailable = 1;
	    break;

	default:
	    break;
    }
    return (0);
}


/****************************
*
*    readAlarmMsg ()
*	read an alarm msg
*
****************************/

int	readAlarmMsg (void)
{
    int		i;
    int		bytesGot = 0;

    memset ((char *)&newAlarmMsg[alarmMsgCnt], 0, ALARM_MSG_SIZE - alarmMsgCnt);
    bytesGot = recv (alarmPrtFd, (char *)&newAlarmMsg[alarmMsgCnt],
		     ALARM_MSG_SIZE - alarmMsgCnt, 0);
    if (bytesGot == -1)
    {
	tisSysLog (0,0, "readAlarmMsg: read error (%m)");
	return (1);
    }
    if (bytesGot == 0)
    {
if(debug)
{
tisSysLog(0,0,
"readAlarmMsg, 0 length read");
}
	return (1);
    }

    i = 0;
    alarmMsgCnt += bytesGot;
    while (1)
    {
	if (i >= ALARM_MSG_SIZE)
	    break;
	if (newAlarmMsg[i] == NULL)
    	    break;
	if (newAlarmMsg[i] == '\n')
	{
	    addNewAlarmPoint ();
    	    memcpy ((char *)&newAlarmMsg[0], (char *)&newAlarmMsg[i + 1], 
		    ALARM_MSG_SIZE - (i + 1));
	    alarmMsgCnt -= (i + 1);
	    memset ((char *)&newAlarmMsg[alarmMsgCnt], 0, 
		    ALARM_MSG_SIZE - alarmMsgCnt);
	    i = 0;
	    continue;
	}
	i++;
    }
    if (answerAlarmPrinter ())
	return (1);

if (debug)
{
tisSysLog(0,0,
"readAlarmMsg, bytes=%d",bytesGot);
}

    return (0);
}


/*********************************************************
*
*    readCurrentAlarmConfig ()
*	called from processAlarms ()
*	set global ptr currentAlarmConfig
*	for this alarm severity
*
*********************************************************/

int	readCurrentAlarmConfig (QUEMSG_PTR newMsg)
{
    int		status = 0;
    int		configIdx;
    struct stat	statbuff;

    status = stat (AT_DATABASE, &statbuff);
    if (status == -1)
    {
	tisSysLog (0,0, "readCurrentAlarmConfig:, stat error for %s (%m)", AT_DATABASE);
	return (1);
    }
    /* if the file has been modified since the last time we
       used its data, re-read it
    */
    if (statbuff.st_mtime > configLastModified)
    {
	if (openAlarmTalkDbf ())
	{
	    return (1);
	}
    }

    if (newMsg -> alarmSeverity == MAJOR_ALARM)
	configIdx = MAJOR_CFG;
    else if (newMsg -> alarmSeverity == MINOR_ALARM)
	configIdx = MINOR_CFG;
    else if (newMsg -> alarmSeverity == INFO_ALARM)
	configIdx = INFO_CFG;
    else
    {
	tisSysLog (0,0, "readCurrentAlarmConfig: illegal alarm severity %d",
		newMsg -> alarmSeverity);
	return (1);
    }

    currentAlarmConfig = &alarmConfigs[configIdx];

    return (0);
}


/*******************************************
*
*    readDecTalk ()
*	read a message from the decTalk 
*
*******************************************/

int	readDecTalk (void)
{
    int		bytesGot;

    memset ((char *)&decTalkRecvBuffer[decTalkMsgCnt], 0, 
	    DECTALK_MSG_SIZE - decTalkMsgCnt);

    if ((strstr (decTalkName, "/dev/")) != NULL)
    {
	bytesGot = read (decTalkFd, (char *)&decTalkRecvBuffer[decTalkMsgCnt], 
			 DECTALK_MSG_SIZE - decTalkMsgCnt);
    }
    else
    {
	bytesGot = recv (decTalkFd, (char *)&decTalkRecvBuffer[decTalkMsgCnt], 
			 DECTALK_MSG_SIZE - decTalkMsgCnt, 0);
    }

    if (bytesGot == -1)
    {
	tisSysLog (0,0, "readDecTalk: read error (%m)");
	return (1);
    }

    if (bytesGot == 0)
    {
if (debug)
{
tisSysLog(0,0,
"readDecTalk, 0 length read");
}
	return (1);
    }

    decTalkMsgCnt += bytesGot;

    return (0);
}


/*****************************************************
*
*    setNoBlock ()
*	set non-blocking IO
*	we do some work-around for the SOLARIS since
*	non-blocking IO does not seem to work, so
*	we increase the send buff to be as large as
*	it can be
*
*****************************************************/

int	setNoBlock (int fd)
{
    int		stat;
    int		true = 1;
    int		txSize = 4096;

#ifdef SOLARIS
    stat = fcntl (fd, F_GETFL, true);
    if (stat == -1)
    {
	tisSysLog (0,0, "setNoBlock: failed fcntl (%m)");
	return (1);
    }
    /* SOL_SOCKET = manipulate options at socket level
       SO_SNDBUF  = set output buffer size
    */
    do
    {
	if (setsockopt (fd, SOL_SOCKET, SO_SNDBUF, (char *)&txSize, 
			sizeof (txSize)) == -1)
	    break;
	txSize += 4096;
    } while (txSize < 65536);

    /* SO_RCVBUF, set input buffer size */
    txSize = 4096;
    setsockopt (fd, SOL_SOCKET, SO_RCVBUF, (char *)&txSize, sizeof (txSize));

    /* set no delay on the fd */
    stat |= O_NDELAY;
    stat = fcntl (fd, F_SETFL, stat);
#else
    /* set non-blocking IO */
    if (stat = ioctl (fd, FIONBIO, &true) == -1)
    {
	tisSysLog (0,0, "setNoBlock: ioctl failed (%m)");
	return (1);
    }
#endif
    return (0);
}
	


/***********************************
*
*   writeDecTalk ()
*	send a msg to the decTalk
*
***********************************/

int	writeDecTalk (char *msg, char *func)
{
    int		numwritten;
    int		totalwritten = 0;
    int		numTries;
    int		numToSend;
    char	outputBuff[DECTALK_MSG_SIZE + 1];
    char	*pChar;

    strcpy (outputBuff, msg);
    numToSend = strlen (msg);
    pChar = &outputBuff[0];

    for (numTries = 0; numTries < 3; numTries++)
    {
	if ((strstr (decTalkName, "/dev/")) != NULL)
	{
	    numwritten = write (decTalkFd, pChar, numToSend);
	}
	else
	{
	    numwritten = send (decTalkFd, pChar, numToSend, 0);
	}
	if (numwritten == -1)
	{
	    tisSysLog (0,0, "writeDecTalk: %s, write error (%m)", func);
	    decWriteFailed = 1;
	    return (1);
	}
	totalwritten += numwritten;
	if (numwritten == numToSend)
	{
	    break;
	}
	else
	{
	    pChar += numwritten;
	    numToSend -= numwritten;
	}
    }
    if (totalwritten != strlen (msg))
    {
	tisSysLog (0,0, "writeDecTalk: %s, failed to send entire msg (%d/%d) (%m)", 
		totalwritten, strlen (msg), func);
	decWriteFailed = 1;
	return (1);
    }
if (debug)
{
tisSysLog (0,0,
"writeDecTalk: wrote %d out of %d, %s", 
totalwritten, strlen (msg),
convertDecTalkMsg (msg));
}
    return (0);
}

