/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/************************************************************************
*
*    alarmTalk.c
*
*    modifications
*    -------------
*  .01	09-22-96	ems	creation date
*
************************************************************************/


#include <string.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/termios.h>
#include <sys/ttold.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <malloc.h>
#ifndef AXP
#include <net/if.h>
#include <sys/sockio.h>
#include <sys/filio.h>
#endif

#ifdef MIPSEL
#include "error.h"
#endif

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "dbDefs.h"		/* .../share/epicsH */
#include "shareMem.h"		/* .../share/src/alarm */
#include "alarmLib.h"		/* .../share/src/alarm */
#include "cadef.h"		/* .../share/epicsH */
#include "messages.h"		/* .../share/src/alarm */
#include "decTalkalarmStrings.h"/* .../share/epicsH */
#include "alarmDefs.h"		/* .../share/src/alarm */
#include "alarmvu.h"		/* .../share/src/alarmVuCC */
#include "slmDefs.h"		/* .../share/epicsH */
#include "OperPriv.h"		/* .../share/src/libtis */
#include "tsDefs.h"		/* .../share/epicsH, TS_STAMP_def */
#include "msgMon.h"		/* .../share/epicsH */
#include "alarmTalk.h"

extern ALARMCFG_PTR	currentAlarmConfig;
extern QUEMSG_PTR	currentAlarm;
extern int		debug;



/*******************************************************************
*
*    composeAlarmMessage ()
*	compose alarm message to be sent to decTalk using values
*	in currentAlarm, chg hyphens and periods to spaces so
*	hyphens not pronounced and . does not act as end of a 
*	phrase/sentence
*	if there is a decimal point in the value, convert it to
*	point, the [_] is a silence phoneme to place a brief pause
*	between that to the left of teh decimal and that to the right
*	to aid in pronunciation
*	NOTE: array decTalkAlarmStrings[] is found in 
*		.../share/epicsH/decTalkalarmStrings.h
*
*******************************************************************/

void	composeAlarmMessage (char *msgBuff)
{
    char	*pChar;
    char	*pName;
    char	*pValue;
    char	*pNewValue;
    char	pointName[PVNAME_SZ + 1];
    char	descriptor[DESCRIPTOR_SZ + 1];
    char	egutag[AL_EGUTAG_SIZE + 1];
    char	value[AL_VALUE_SIZE + 1];
    char	newValue[AL_VALUE_SIZE + 16];

    if ((strncmp (currentAlarmConfig -> annunciation, STD_MALE_VOICE, 
							ANNUNCTION_SZ)) == 0)
	pName = STD_MALE_VOICE_CMD;
    else if ((strncmp (currentAlarmConfig -> annunciation, STD_FEMALE_VOICE, 
							ANNUNCTION_SZ)) == 0)
	pName = STD_FEMALE_VOICE_CMD;
    else if ((strncmp (currentAlarmConfig -> annunciation, DEEP_MALE_VOICE, 
							ANNUNCTION_SZ)) == 0)
	pName = DEEP_MALE_VOICE_CMD;
    else 
	pName = DEEP_FEMALE_VOICE_CMD;

    strncpy (pointName, currentAlarm -> pointName, PVNAME_SZ);
    pointName[PVNAME_SZ] = NULL;
    pChar = pointName;
    while ( (*pChar) != NULL)
    {
	if ( ((*pChar) == '-') ||
	     ((*pChar) == '.') ||
	     ((*pChar) == '_') )
	{
	    *pChar = ' ';
	}
	pChar++;
    }
    strncpy (descriptor, currentAlarm -> descriptor, DESCRIPTOR_SZ);
    descriptor[DESCRIPTOR_SZ] = NULL;
    pChar = descriptor;
    while ( (*pChar) != NULL)
    {
	if ( ((*pChar) == '-') ||
	     ((*pChar) == '.') ||
	     ((*pChar) == '_') )
	{
	    *pChar = ' ';
	}
	pChar++;
    }
    strncpy (egutag, currentAlarm -> egutag, AL_EGUTAG_SIZE);
    egutag[AL_EGUTAG_SIZE] = NULL;
    pChar = egutag;
    while ( (*pChar) != NULL)
    {
	if ( ((*pChar) == '-') ||
	     ((*pChar) == '.') ||
	     ((*pChar) == '_') )
	{
	    *pChar = ' ';
	}
	pChar++;
    }
    strncpy (value, currentAlarm -> value, AL_VALUE_SIZE);
    value[AL_VALUE_SIZE] = NULL;
    /* if there is a decimal point inthe value, chg it to " point " */
    pValue = value;
    if ((pChar = strchr (value, '.')) != NULL)
    {
	memset ((char *)&newValue[0], 0, sizeof (newValue));
	pNewValue = newValue;
	while ((*pValue) != NULL)
	{
	    if ((*pValue) == '.')
	    {
		strcat (pNewValue, "[_] point ");
		/* only 9 bewcause bump 1 below */
		pNewValue += 9;
	    }
	    else
		*pNewValue = *pValue;
	    pValue++;
	    pNewValue++;
	}
	pValue = newValue;
    }
    sprintf (msgBuff, "%s Point %s, %s, is in %s alarm with value %s %s",
		      pName,
		      pointName,
		      descriptor,
		      decTalkAlarmStrings[currentAlarm -> alarmStatus],
		      pValue,
		      egutag);
    msgBuff[DECTALK_MSG_SIZE] = NULL;
    /* there can not be any spaces between the last character and the
       period terminating the line or else the decTalk will pronoune
       the . as period rather than looking at it as an end of a
       sentence
    */
    removeTrailingSpaces (msgBuff);
    strcat (msgBuff, ".");
if(debug)
{
tisSysLog(0,0,
"return composeAlMsg");
}
    return;
}


/*************************************************
*
*    sendAckdPtsMsg ()
*	tell user all points were acknowwledged
*
**************************************************/

void	sendAckdPtsMsg (void)
{
    writeDecTalk (DTNOKEYPAD, "sendAckdPtsMsg");
    writeDecTalk (DTSPEAK, "sendAckdPtsMsg");
    writeDecTalk (ALL_PTS_ACKD_MSG, "sendAckdPtsMsg");
    writeDecTalk (DTWINK, "sendAckdPtsMsg");
    writeDecTalk (HANGUP_THE_PHONE_MSG, "sendAckdPtsMsg");
    writeDecTalk (DTSYNC, "sendAckdPtsMsg");
    return;
}


/*******************************************************
*
*    sendAlarmMsg ()
*	responeses if auto-dial
*
*******************************************************/

void	sendAlarmMsg (int sendGreeting)
{
    char	msgBuff[DECTALK_MSG_SIZE + 1];
    int		winkSent = 0;

    writeDecTalk (DTNOKEYPAD, "sendAlarmMsg");
    writeDecTalk (DTSPEAK, "sendAlarmMsg");
    if (sendGreeting)
    {
	writeDecTalk (ALARMTALK_GREETING,"sendAlarmMsg");
	writeDecTalk (DTWINK, "sendAlarmMsg");
	winkSent = 1;
    }
    if (!winkSent)
    {
	writeDecTalk (DTWINK, "sendAlarmMsg");
    }
    composeAlarmMessage (msgBuff);
    writeDecTalk (msgBuff, "sendAlarmMsg");
    writeDecTalk (DTSYNC, "sendAlarmMsg");
    writeDecTalk (DTINDEXQUERY, "sendAlarmMsg");

    return;
}


/**************************************************
*
*    sendAllPtsNotAckdMsg ()
*	tell user all points were acknowwledged
*
**************************************************/

void	sendAllPtsNotAckdMsg (void)
{
    writeDecTalk (DTNOKEYPAD, "sendAllPtsNotAckdMsg");
    writeDecTalk (DTSPEAK, "sendAllPtsNotAckdMsg");
    writeDecTalk (ALL_PTS_NOT_ACKD_MSG, "sendAllPtsNotAckdMsg");
    writeDecTalk (DTWINK, "sendAllPtsNotAckdMsg");
    writeDecTalk (HANGUP_THE_PHONE_MSG, "sendAllPtsNotAckdMsg");
    writeDecTalk (DTSYNC, "sendAllPtsNotAckdMsg");
    return;
}


/************************************************
*
*    sendAutoIdMsg ()
*	send request for caller ID when in 
*	auto dialer mode, DTSYNC will cause the
*	message to be spoken before the timer
*	starts
*
************************************************/

int	sendAutoIdMsg (void)
{
    char	msgBuff[DECTALK_MSG_SIZE + 1];
    int		timeout;

    writeDecTalk (DTKEYPAD, "sendAutoIdMsg");
    writeDecTalk (DTSPEAK, "sendAutoIdMsg");
    writeDecTalk (AUTO_ID_REQ_MSG, "sendAutoIdMsg");
    writeDecTalk (DTWINK, "sendAutoIdMsg");
    writeDecTalk (DTSYNC, "sendAutoIdMsg");
    timeout = currentAlarmConfig -> idTimeout;
    if (timeout == 0)
    {
	timeout = TIMEOUT_DEADBAND;
    }
    sprintf (msgBuff, "%s%d%s", DTTIMEOUT, timeout, DTEND);
    writeDecTalk (msgBuff, "sendAutoIdMsg");
    return (timeout);
}


/*************************************************
*
*    sendFailRepeatAckMsg ()
*	user failed to respond to REPEAT_ACK_MSG
*
*************************************************/

int	sendFailRepeatAckMsg (void)
{
    char	msgBuff[DECTALK_MSG_SIZE + 1];
    int		timeout;

    writeDecTalk (DTKEYPAD, "sendFailRepeatAckMsg");
    writeDecTalk (DTSPEAK, "sendFailRepeatAckMsg");
    writeDecTalk (FAILED_REPEAT_ACK_MSG, "sendFailRepeatAckMsg");
    writeDecTalk (DTWINK, "sendFailRepeatAckMsg");
    writeDecTalk (REPEAT_ACK_MSG, "sendFailRepeatAckMsg");
    writeDecTalk (DTSYNC, "sendFailRepeatAckMsg");
    timeout = currentAlarmConfig -> repackTimeout;
    if (timeout == 0)
    {
	timeout = TIMEOUT_DEADBAND;
    }
    sprintf (msgBuff, "%s%d%s", DTTIMEOUT, timeout, DTEND);
    writeDecTalk (msgBuff, "sendFailRepeatAckMsg");
    return (timeout);
}


/***************************************************
*
*    sendFailToEnterIdMsg ()
*	user failed to enter id code within timeout
*	period. tell them, let them try again
*
***************************************************/

int	sendFailToEnterIdMsg (void)
{
    char	msgBuff[DECTALK_MSG_SIZE + 1];
    int		timeout;

    writeDecTalk (DTKEYPAD, "sendFailToEnterIdMsg");
    writeDecTalk (DTSPEAK, "sendFailToEnterIdMsg");
    writeDecTalk (NO_ID_ENTRY, "sendFailToEnterIdMsg");
    writeDecTalk (DTWINK, "sendFailToEnterIdMsg");
    writeDecTalk (AUTO_ID_REQ_MSG, "sendFailToEnterIdMsg");
    writeDecTalk (DTSYNC, "sendFailToEnterIdMsg");
    timeout = currentAlarmConfig -> idTimeout;
    if (timeout == 0)
    {
	timeout = TIMEOUT_DEADBAND;
    }
    sprintf (msgBuff, "%s%d%s", DTTIMEOUT, timeout, DTEND);
    writeDecTalk (msgBuff, "sendFailToEnterIdMsg");
    return (timeout);
}


/****************************************************************
*
*    sendFinalBadIdMsg ()
*	notify user that they did not enter a valid password
*	after 2 tries and that they are to hangup the phone
*
****************************************************************/

void	sendFinalBadIdMsg (void)
{
    writeDecTalk (DTNOKEYPAD, "sendFinalBadIdMsg");
    writeDecTalk (DTSPEAK, "sendFinalBadIdMsg");
    writeDecTalk (BAD_ID_ENTRY, "sendFinalBadIdMsg");
    writeDecTalk (DTWINK, "sendFinalBadIdMsg");
    writeDecTalk (HANGUP_THE_PHONE_MSG, "sendFinalBadIdMsg");
    writeDecTalk (DTSYNC, "sendFinalBadIdMsg");
    return;
}


/*****************************************************
*
*    sendFinalFailedIdMsg ()
*	user failed to enter Id for 2nd time, let
*	them know and ask for a hangup
*
*****************************************************/

void	sendFinalFailedIdMsg (void)
{
    writeDecTalk (DTNOKEYPAD, "sendFinalFailedIdMsg");
    writeDecTalk (DTSPEAK, "sendFinalFailedIdMsg");
    writeDecTalk (NO_ID_ENTRY, "sendFinalFailedIdMsg");
    writeDecTalk (DTWINK, "sendFinalFailedIdMsg");
    writeDecTalk (HANGUP_THE_PHONE_MSG, "sendFinalFailedIdMsg");
    writeDecTalk (DTSYNC, "sendFinalFailedIdMsg");
    return;
}


/*****************************************************
*
*    sendFinalMgrNotAvailableMsg ()
*	alarm manager was not available after 2 tries
*
*****************************************************/

void	sendFinalMgrNotAvailableMsg (char *msg)
{
    writeDecTalk (DTNOKEYPAD, "sendFinalMgrNotAvailableMsg");
    writeDecTalk (DTSPEAK, "sendFinalMgrNotAvailableMsg");
    writeDecTalk (FINAL_MGR_NOT_AVAILABLE_MSG, "sendFinalMgrNotAvailableMsg");
    writeDecTalk (DTWINK, "sendFinalMgrNotAvailableMsg");
    writeDecTalk (msg, "sendFinalMgrNotAvailableMsg");
    writeDecTalk (HANGUP_THE_PHONE_MSG, "sendFinalMgrNotAvailableMsg");
    writeDecTalk (DTSYNC, "sendFinalMgrNotAvailableMsg");
    return;
}


/*****************************************************
*
*    sendFinalRepeatAckMsg ()
*	user failed to enter Id for 2nd time, let
*	them know and ask for a hangup
*
*****************************************************/

void	sendFinalRepeatAckMsg (void)
{
    writeDecTalk (DTNOKEYPAD, "sendFinalRepeatAckMsg");
    writeDecTalk (DTSPEAK, "sendFinalRepeatAckMsg");
    writeDecTalk (FAILED_REPEAT_ACK_MSG, "sendFinalRepeatAckMsg");
    writeDecTalk (DTWINK, "sendFinalRepeatAckMsg");
    writeDecTalk (HANGUP_THE_PHONE_MSG, "sendFinalRepeatAckMsg");
    writeDecTalk (DTSYNC, "sendFinalRepeatAckMsg");
    return;
}


/*******************************************
*
*    sendHangUpMsg ()
*	tell someone to hangup the phone
*
*******************************************/

void	sendHangUpMsg (void)
{
    writeDecTalk (DTNOKEYPAD, "sendHangUpMsg");
    writeDecTalk (DTSPEAK, "sendHangUpMsg");
    writeDecTalk (ALARMTALK_HANGUP_MSG, "sendHangUpMsg");
    writeDecTalk (DTWINK, "sendHangUpMsg");
    writeDecTalk (DTSYNC, "sendHangUpMsg");
    return;
}


/*******************************************
*
*    sendMgrAvailableMsg ()
*	tell someone the alarm manager is
*	available again
*
*******************************************/

void	sendMgrAvailableMsg (void)
{
    writeDecTalk (DTNOKEYPAD, "sendMgrAvailableMsg");
    writeDecTalk (DTSPEAK, "sendMgrAvailableMsg");
    writeDecTalk (DTWINK, "sendMgrAvailableMsg");
    writeDecTalk (MGR_AVAILABLE_MSG, "sendMgrAvailableMsg");
    writeDecTalk (DTSYNC, "sendMgrAvailableMsg");
    return;
}


/*******************************************
*
*    sendMgrNotAvailableMsg ()
*	alarm mgr is not available
*
*******************************************/

void	sendMgrNotAvailableMsg (void)
{
    writeDecTalk (DTNOKEYPAD, "sendMgrNotAvailableMsg");
    writeDecTalk (DTSPEAK, "sendMgrNotAvailableMsg");
    writeDecTalk (MGR_NOT_AVAILABLE_MSG, "sendMgrNotAvailableMsg");
    writeDecTalk (DTWINK, "sendMgrNotAvailableMsg");
    writeDecTalk (DTSYNC, "sendMgrNotAvailableMsg");
    return;
}



/************************************************
*
*    sendNoAlarmsMsg ()
*	in auto dialer mode there are no alarms
*	for which the user has ack privileges
*	or there was an error building the 
*	list, see buildUnAckList ()
*	or could not find any in list to send
*
************************************************/

void	sendNoAlarmsMsg (char *msg)
{
    writeDecTalk (DTNOKEYPAD, "sendNoAlarmsMsg");
    writeDecTalk (DTSPEAK, "sendNoAlarmsMsg");
    writeDecTalk (DTWINK, "sendNoAlarmsMsg");
    writeDecTalk (msg, "sendNoAlarmsMsg");
    writeDecTalk (HANGUP_THE_PHONE_MSG, "sendNoAlarmsMsg");
    writeDecTalk (DTSYNC, "sendNoAlarmsMsg");
    return;
}


/***********************************************
*
*    sendRepeatAckMsg ()
*	ask user if they want the unack list
*	to be repeated, or to ack the alarms
*
***********************************************/

int	sendRepeatAckMsg (void)
{
    char	msgBuff[DECTALK_MSG_SIZE + 1];
    int		timeout;

    writeDecTalk (DTKEYPAD, "sendRepeatAckMsg");
    writeDecTalk (DTSPEAK, "sendRepeatAckMsg");
    writeDecTalk (REPEAT_ACK_MSG, "sendRepeatAckMsg");
    writeDecTalk (DTWINK, "sendRepeatAckMsg");
    writeDecTalk (DTSYNC, "sendRepeatAckMsg");
    timeout = currentAlarmConfig -> repackTimeout;
    if (timeout == 0)
    {
	timeout = TIMEOUT_DEADBAND;
    }
    sprintf (msgBuff, "%s%d%s", DTTIMEOUT, timeout, DTEND);
    writeDecTalk (msgBuff, "sendRepeatAckMsg");
    return (timeout);
}


/***********************************************
*
*    sendRepeatIdMsg ()
*	tell user to reenter their password
*
***********************************************/

int	sendRepeatIdMsg (void)
{
    char	msgBuff[DECTALK_MSG_SIZE + 1];
    int		timeout;

    writeDecTalk (DTKEYPAD, "sendRepeatIdMsg");
    writeDecTalk (DTSPEAK, "sendRepeatIdMsg");
    writeDecTalk (BAD_ID_ENTRY, "sendRepeatIdMsg");
    writeDecTalk (DTWINK, "sendRepeatIdMsg");
    writeDecTalk (AUTO_ID_REQ_MSG, "sendRepeatIdMsg");
    writeDecTalk (DTSYNC, "sendRepeatIdMsg");
    timeout = currentAlarmConfig -> idTimeout;
    if (timeout == 0)
    {
	timeout = TIMEOUT_DEADBAND;
    }
    sprintf (msgBuff, "%s%d%s", DTTIMEOUT, timeout, DTEND);
    writeDecTalk (msgBuff, "sendRepeatIdMsg");
    return (timeout);
}


/************************************************************
*
*    sendUnexpectedMsg ()
*	send msg saying we recived an unexpected msg from
*	the decTalk, msg is the consequence
*
************************************************************/

void	sendUnexpectedMsg (char *msg)
{
    writeDecTalk (DTNOKEYPAD, "sendUnexpectedMsg");
    writeDecTalk (DTSPEAK, "sendUnexpectedMsg");
    writeDecTalk (UNEXPECTED_MSG, "sendUnexpectedMsg");
    writeDecTalk (DTWINK, "sendUnexpectedMsg");
    writeDecTalk (msg, "sendUnexpectedMsg");
    writeDecTalk (DTSYNC, "sendUnexpectedMsg");
    return;
}


/*******************************************************
*
*    sendWaitBuildListMsg ()
*	send a message that tells listener that we are
*	building an unacknowledged list
*
*******************************************************/

void	sendWaitBuildListMsg (void)
{
    writeDecTalk (DTNOKEYPAD, "sendWaitBuildListMsg");
    writeDecTalk (DTSPEAK, "sendWaitBuildListMsg");
    writeDecTalk (WAIT_BUILD_LIST_MSG,"sendWaitBuildListMsg");
    writeDecTalk (DTWINK, "sendWaitBuildListMsg");
    writeDecTalk (DTSYNC, "sendWaitBuildListMsg");
    writeDecTalk (DTINDEXQUERY, "sendWaitBuildListMsg");

    return;
}
