/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <string.h>

#include <cadef.h>
#include <tisAlarm.h>
#include <alarmDefs.h>
#include <stringSpace.h>
#include <ChannelList.h>

/* Functions defined in this module */

#if defined(__STDC__) || defined(WIN32)
#define	VOID	void
	void	clInitialize (void);
	void	clTerminate (void);
	int	clAdd (char *);
	void	clConnect (void);
static	void	connectHandler (struct connection_handler_args args);;
static	void	eventHandler (struct event_handler_args args);
	int	clGetValue (int);
#else
#define	VOID
	void	clInitialize (void);
	void	clTerminate (void);
	int	clAdd (char *);
	void	clConnect (void);
static	void	connectHandler (struct connection_handler_args args);;
static	void	eventHandler (struct event_handler_args args);
	int	clGetValue (int);
#endif

static	ChannelList	channelList;
	int		channelChanged;

/*----------*/
void clInitialize (VOID)

    {

	/* Entry 0 always has the value zero so that a group having */
	/* no auxPv specified is always in "main" mode. */

	channelList.chans [0].value = 0;
	channelList.chans [0].chanName = addNewString ("dummy");
	channelList.numChannels = 1;

    }

/*----------*/
void clTerminate (VOID)

    {

	int	i;

	for (i=1 ; i<channelList.numChannels ; i++)
	    /* guard against the channel not being built yet */
	    if (channelList.chans [i].chanId)
		ca_clear_channel (channelList.chans [i].chanId);

    }

/*----------*/
int clAdd (char *channelName)

    {

	int	i;

	/* Check to see if the channel is already in list */

	for (i=0 ; i<channelList.numChannels ; i++)
	    if (strcmp (channelName, channelList.chans [i].chanName) == 0)
		return i;

	/* Place the channel in the list */

	channelList.chans [channelList.numChannels].chanName = addNewString (channelName);
	channelList.chans [channelList.numChannels].value = 0;	/* Normal mode until connection made */
	return channelList.numChannels++;

    }

/*----------*/
void clConnect (VOID)	/* Connect any channels that are not already */

    {

	int	i;

	for (i=0 ; i<channelList.numChannels ; i++) {
	    ca_build_and_connect (channelList.chans [i].chanName, TYPENOTCONN, 0, 
				&channelList.chans [i].chanId, NULL, connectHandler, NULL);
	}

    }

/*----------*/
static void connectHandler (struct connection_handler_args args)

    {

	int	i;

	/* Find the channel that has connected */

	for (i=1 ; i<channelList.numChannels ; i++)
	    if (args.chid == channelList.chans [i].chanId)
		break;

	/* Channel has connected, now set callback for event. */

	ca_add_event (OLD_DBR_INT, args.chid, eventHandler,
			(void *)&channelList.chans [i].value,
			&channelList.chans [i].eventId);
	    
    }

/*----------*/
static void eventHandler (struct event_handler_args args)

    {

	*(short *)args.usr = *(short *)args.dbr;
	channelChanged = 1;

    }

/*----------*/
int clGetValue (int i)
{

	return channelList.chans [i].value;

}
