/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include "alarmDefs.h"
#include "shareMem.h"
#include "groupInfo.h"
#include "tsDefs.h"
#include "dbase.h"
#include "resolveLinks.h"

/* Functions defined in this module */

#if defined(__STDC__) || defined(WIN32)
#define VOID	void
	void	initializeDatabase (void);
	int	createGroupFile (void);
	int	openGroupFile (void);
	void	closeGroupFile (void);
	int	getNextGroup (char *);
	void	writeGroupName (char *);
	int	getNumGroups (void);
	int	createTreeFile (void);
	int	openTreeFile (void);
	void	closeTreeFile (void);
	int	getNextTreeEdge (char *, char *);
	void	writeTreeEdge (char *, char *);
	int	createInfoFile (void);
	int	openInfoFile (void);
	void	closeInfoFile (void);
	int	getInfo (char *, GroupInfo *);
	void	writeInfo (char *, GroupInfo *);
	int	openPvMapFile (void);
	void	closePvMapFile (void);
	int	getNumPVs (void);
	int	getNextPv (PvMapInfo *);
	int	getPv (char *, char *, char *, char *);
	int	openIOCListFile (void);
	void	closeIOCListFile (void);
	int	getNextIOCEntry (char *, int *);
	int	createLogFile (void);
	int	openLogFile (void);
	void	closeLogFile (void);
	int	getLogEntry (char *, LogEntry *, int *);
	int	getNextLogEntry (LogEntry *, int *);
static	void	getLogFields (LogEntry *);
	void	writeLogEntry (LogEntry *);
	int	writeLogMemo (char *, TS_STAMP *, char *msg);
	void	makeLogKeyFromTS (char *, TS_STAMP *, int, char *);
	void	makeLogKeyFromEntry (LogEntry *, int, char *);
	int	setAckInLog (char *, TS_STAMP *, TS_STAMP *);		 
        void    openResListFile (void);
        void    closeResListFile (void);
        int     getResListEntry (char *, ResListInfo *);
#ifndef CONFIG
	int	createPointStateFile (void);
	int	openPointStateFile (void);
	void	closePointStateFile (void);
	int	getPointState (char *, PvData *, int *);
	void	writePointState (PvData *);
	int	deletePointState (char *);
#endif
	int	createPrinterFile (void);
	int	openPrinterFile (void);
	void	closePrinterFile (void);
	int	getNextPrinterGroup (PrinterGroupEntry *);
	void	writePrinterGroup (PrinterGroupEntry *);
	int	getNumPrinters (void);
	void	updateDBaseFile (D4DATA *);
static	void	stringCopy (char *, char *, int);
	void	stripBlanks (char *s);
#else
	void	initializeDatabase ();
	int	createGroupFile ();
	int	openGroupFile ();
	void	closeGroupFile ();
	int	getNextGroup ();
	void	writeGroupName ();
	int	getNumGroups ();
	int	createTreeFile ();
	int	openTreeFile ();
	void	closeTreeFile ();
	int	getNextTreeEdge ();
	void	writeTreeEdge ();
	int	createInfoFile ();
	int	openInfoFile ();
	void	closeInfoFile ();
	int	getInfo ();
	void	writeInfo ();
	int	openPvMapFile ();
	void	closePvMapFile ();
	int	getNumPVs ();
	int	getNextPv ();
	int	getPv ();
	int	openIOCListFile ();
	void	closeIOCListFile ();
	int	getNextIOCEntry ();
	int	createLogFile ();
	int	openLogFile ();
	void	closeLogFile ();
	int	getLogEntry ();
	int	getNextLogEntry ();
static	void	getLogFields ();
	void	writeLogEntry ();
	int	writeLogMemo ();
	void	makeLogKeyFromTS ();
	void	makeLogKeyFromEntry ();
	int	setAckInLog ();		
        void    openResListFile ();
        void    closeResListFile ();
        int     getResListEntry ();
#ifndef CONFIG
	int	createPointStateFile ();
	int	openPointStateFile ();
	void	closePointStateFile ();
	int	getPointState ();
	void	writePointState ();
	int	deletePointState ();
#endif
	int	createPrinterFile ();
	int	openPrinterFile ();
	void	closePrinterFile ();
	int	getNextPrinterGroup ();
	void	writePrinterGroup ();
	int	getNumPrinters ();
	void	updateDBaseFile ();
static	void	stringCopy ();
	void	stripBlanks ();
#endif

/* This file contains information that is global to all alarms. It will */
/* only ever contain a single record. */
 
#include <dbFormats.c>

/* Codebase related things */

C4CODE	codeBase;

D4DATA	*groupData;
F4FIELD	*groupDataNameFld;

D4DATA	*treeData;
F4FIELD	*parentFld, *childFld;

D4DATA	*infoData;
F4FIELD	*groupNameFld;
F4FIELD	*autoScriptFld, *auxAutoScriptFld, *manScriptFld, *auxManScriptFld;
F4FIELD	*displayNameFld, *helpTopicFld, *guidanceFld;
F4FIELD	*majorPrintFld, *minorPrintFld, *infoPrintFld, *eventPrintFld, *normPrintFld;
F4FIELD	*auxMajorPrintFld, *auxMinorPrintFld, *auxInfoPrintFld, *auxEventPrintFld, *auxNormPrintFld;
F4FIELD	*majorSoundFld, *minorSoundFld, *infoSoundFld, *eventSoundFld;
F4FIELD	*auxMajorSoundFld, *auxMinorSoundFld, *auxInfoSoundFld, *auxEventSoundFld;
F4FIELD	*majorRateFld, *minorRateFld, *infoRateFld, *eventRateFld;
F4FIELD	*auxMajorRateFld, *auxMinorRateFld, *auxInfoRateFld, *auxEventRateFld;
F4FIELD	*auxSelectPVFld;

D4DATA	*pvMapData;
F4FIELD	*pvNameFld, *pvGroupFld, *pvDescriptorFld, *pvDeleteFlagFld, *pvResNameFld;

D4DATA	*pdbXrefData;
char	dbfilename [AL_MAXFILENAME_SIZE];
D4DATA	*dbData;


D4DATA	*iocListData;
F4FIELD	*iocNameFld, *iocIsHostFld;

D4DATA  *resListData;
F4FIELD *ioc1Fld, *ioc2Fld;
D4DATA	*logData;
F4FIELD	*logIsAckFld, *logDateFld, *logTimeFld, *logAlarmSecsFld, *logAlarmMsecsFld, *logGroupNameFld;
F4FIELD	*logPointNameFld, *logDescriptorFld, *logStatusFld, *logSeverityFld;
F4FIELD	*logAckedFld, *logOperatorFld, *logResourceFld, *logStationNameFld, *logEguTagFld;
F4FIELD	*logValueFld, *logAckSecsFld, *logAckMsecsFld, *logCommentFld, *logNoPrintFld;
F4FIELD	*logInfo1Fld, *logInfo2Fld, *logInfo3Fld, *logInfo4Fld, *logInfo5Fld;

D4DATA	*psData;
F4FIELD	*psPointNameFld, *psAlarmSecsFld, *psAlarmNsecsFld;
F4FIELD	*psAckSecsFld, *psAckNsecsFld, *psTimeSecsFld, *psTimeNsecsFld;
F4FIELD	*psStatusFld, *psSeverityFld, *psUstatusFld, *psUseverityFld, *psValueFld, *psUvalueFld;
F4FIELD	*psAckFld, *psRtnFld, *psDescriptorFld, *psEgutagFld, *psOperatorFld, *psInSummaryFld;
F4FIELD	*psAenbFld, *psOscnFld;

D4DATA	*printerData;
F4FIELD	*printerGroupNameFld, *printerPrinterFld [MAX_PRINTERS_PER_GROUP];
 
GroupInfo	blankParms;
long		treeRecNum, treeNumRecs;
long		groupRecNum, groupNumRecs;
long		pvMapRecNum, pvMapNumRecs;
long		logRecNum, logNumRecs;
long		printerRecNum, printerNumRecs;

char	severityChars [] = {" IEmM  "};
static	int	initialized = 0;

/*----------*/
void initializeDatabase (VOID)

    {

	if (!initialized) {
	    d4init (&codeBase);
#if defined(USECB63) || defined(WIN32)
	    codeBase.errOpen = 0;
	    codeBase.errCreate = 0;
	    codeBase.errFieldName = 0;
#else
	    codeBase.open_error = 0;
	    codeBase.create_error = 0;
	    codeBase.field_name_error = 0;
#endif WIN32

	    /* This is used when writing parameters to the info database */

	    memset ((char *)&blankParms, 0, sizeof (GroupInfo));
	    initialized = 1;
	}

    }

/*----------*/
int createGroupFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_GROUP_FILE_NAME, pathName);
	unlink (pathName);
	groupData = d4create (&codeBase, pathName, groupNameFields, NULL);
	if (groupData == NULL)
	    return 0;
	d4append_start (groupData, 0);
	f4assign (d4field (groupData, "GROUPNAME"), "MAIN");
	d4append (groupData);
	d4close (groupData);
	return openGroupFile ();

    }

/*----------*/
int openGroupFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_GROUP_FILE_NAME, pathName);
	groupData = d4open (&codeBase, pathName);
	if (groupData == NULL)
	    return 0;
	groupDataNameFld = d4field (groupData, "GROUPNAME");
	groupRecNum = 1;
	groupNumRecs = d4reccount (groupData);
	return 1;

    }

/*----------*/
void closeGroupFile (VOID)

    {

	updateDBaseFile (groupData);
	d4close (groupData);

    }

/*----------*/
int getNextGroup (groupName)

    char	*groupName;

    {

	if (groupRecNum > groupNumRecs)
	    return 0;

	d4go (groupData, groupRecNum);
	stringCopy (groupName, f4str (groupDataNameFld), sizeof (GroupName));
	stripBlanks (groupName);
	groupRecNum++;
	return 1;
		
    }

/*----------*/
void writeGroupName (groupName)

    char	*groupName;

    {

	if (strcmp (groupName, "MAIN") != 0) {
	    d4append_start (groupData, 0);
	    f4assign (groupDataNameFld, groupName);
	    d4append (groupData);
	}

    }

/*----------*/
int getNumGroups (VOID)	/* Get number of groups in group database */

    {

	return d4reccount (groupData);

    }

/*----------*/
int createTreeFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_TREE_FILE_NAME, pathName);
	unlink (pathName);
	treeData = d4create (&codeBase, pathName, treeFields, NULL);
	if (treeData == NULL)
	    return 0;
	d4append_start (treeData, 0);
	f4assign (d4field (treeData, "PARENT"), "");
	f4assign (d4field (treeData, "CHILD"), "MAIN");
	d4append (treeData);
	d4close (treeData);
	return openTreeFile ();

    }

/*----------*/
int openTreeFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_TREE_FILE_NAME, pathName);
	treeData = d4open (&codeBase, pathName);
	if (treeData == NULL)
	    return 0;
	parentFld = d4field (treeData, "PARENT");
	childFld = d4field (treeData, "CHILD");
	treeRecNum = 1;
	treeNumRecs = d4reccount (treeData);
	return 1;

    }

/*----------*/
void closeTreeFile (VOID)

    {

	updateDBaseFile (treeData);
	d4close (treeData);

    }

/*----------*/
int getNextTreeEdge (parent, child)

    char	*parent, *child;

    {

	if (treeRecNum > treeNumRecs)
	    return 0;

	d4go (treeData, treeRecNum);
	stringCopy (parent, f4str (parentFld), sizeof (GroupName));
	stripBlanks (parent);
	stringCopy (child, f4str (childFld), sizeof (GroupName));
	stripBlanks (child);
	treeRecNum++;
	return 1;

    }

/*----------*/
void writeTreeEdge (parent, child)

    char	*parent, *child;

    {

	/* The root is already written when the file is created */

	if (strlen (parent)) {
	    d4append_start (treeData, 0);
	    f4assign (parentFld, parent);
	    f4assign (childFld, child);
	    d4append (treeData);
	}

    }

/*----------*/
int createInfoFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_INFO_IX_FILE_NAME, pathName);
	unlink (pathName);
	resolveLinks (AL_INFO_FILE_NAME, pathName);
	unlink (pathName);
	infoData = d4create (&codeBase, pathName, infoFields, infoTags);
	if (infoData == NULL)
	    return 0;
	d4close (infoData);
	return openInfoFile ();

    }

/*----------*/
int openInfoFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_INFO_FILE_NAME, pathName);
	infoData = d4open (&codeBase, pathName);
	if (infoData == NULL)
	    return 0;

	groupNameFld = d4field (infoData, "GROUPNAME");
	autoScriptFld = d4field (infoData, "ASCRIPT");
	auxAutoScriptFld = d4field (infoData, "AUXASCRIPT");
	manScriptFld = d4field (infoData, "MSCRIPT");
	auxManScriptFld = d4field (infoData, "AUXMSCRIPT");
	displayNameFld = d4field (infoData, "DISPNAME");
	helpTopicFld = d4field (infoData, "HELPTOPIC");
	guidanceFld = d4field (infoData, "GUIDANCE");
	majorPrintFld = d4field (infoData, "MAJPRINT");
	minorPrintFld = d4field (infoData, "MINPRINT");
	infoPrintFld = d4field (infoData, "INFOPRINT");
	eventPrintFld = d4field (infoData, "EVPRINT");
	normPrintFld = d4field (infoData, "NORMPRINT");
	auxMajorPrintFld = d4field (infoData, "AMAJPRINT");
	auxMinorPrintFld = d4field (infoData, "AMINPRINT");
	auxInfoPrintFld = d4field (infoData, "AINFOPRINT");
	auxEventPrintFld = d4field (infoData, "AEVPRINT");
	auxNormPrintFld = d4field (infoData, "ANORMPRINT");
	majorSoundFld = d4field (infoData, "MAJSOUND");
	minorSoundFld = d4field (infoData, "MINSOUND");
	infoSoundFld = d4field (infoData, "INFOSOUND");
	eventSoundFld = d4field (infoData, "EVSOUND");
	auxMajorSoundFld = d4field (infoData, "AMAJSOUND");
	auxMinorSoundFld = d4field (infoData, "AMINSOUND");
	auxInfoSoundFld = d4field (infoData, "AINFOSOUND");
	auxEventSoundFld = d4field (infoData, "AEVSOUND");
	majorRateFld = d4field (infoData, "MAJRATE");
	minorRateFld = d4field (infoData, "MINRATE");
	infoRateFld = d4field (infoData, "INFORATE");
	eventRateFld = d4field (infoData, "EVRATE");
	auxMajorRateFld = d4field (infoData, "AMAJRATE");
	auxMinorRateFld = d4field (infoData, "AMINRATE");
	auxInfoRateFld = d4field (infoData, "AINFORATE");
	auxEventRateFld = d4field (infoData, "AEVRATE");
	auxSelectPVFld = d4field (infoData, "AUXSELPV");
	d4top (infoData);
	return 1;

    }

/*----------*/
void closeInfoFile (VOID)

    {

	updateDBaseFile (infoData);
	d4close (infoData);

    }

/*----------*/
int getInfo (groupName, parms)

    char	*groupName;
    GroupInfo	*parms;

    {

	char	key [50];

	memset ((char *)parms, 0, sizeof (GroupInfo));

	sprintf (key, "%-*s", AL_GROUPNAME_SIZE, groupName);
	if (d4seek (infoData, key) != 0) {
	    d4unlock_all (infoData);
	    return 0;
	}

	/* Copy normal mode parameters */

	stringCopy (parms->printers [MINOR], f4str (minorPrintFld), sizeof (PathName));
	stripBlanks (parms->printers [MINOR]);
	stringCopy (parms->printers [MAJOR], f4str (majorPrintFld), sizeof (PathName));
	stripBlanks (parms->printers [MAJOR]);
	stringCopy (parms->printers [INFO], f4str (infoPrintFld), sizeof (PathName));
	stripBlanks (parms->printers [INFO]);
	stringCopy (parms->printers [EVENT], f4str (eventPrintFld), sizeof (PathName));
	stripBlanks (parms->printers [EVENT]);
	stringCopy (parms->printers [NORM], f4str (normPrintFld), sizeof (PathName));
	stripBlanks (parms->printers [NORM]);

	stringCopy (parms->sounds [MINOR], f4str (minorSoundFld), sizeof (PathName));
	stripBlanks (parms->sounds [MINOR]);
	stringCopy (parms->sounds [MAJOR], f4str (majorSoundFld), sizeof (PathName));
	stripBlanks (parms->sounds [MAJOR]);
	stringCopy (parms->sounds [INFO], f4str (infoSoundFld), sizeof (PathName));
	stripBlanks (parms->sounds [INFO]);
	stringCopy (parms->sounds [EVENT], f4str (eventSoundFld), sizeof (PathName));
	stripBlanks (parms->sounds [EVENT]);

	stringCopy (parms->rates [MINOR], f4str (minorRateFld), sizeof (Rate));
	stripBlanks (parms->rates [MINOR]);
	stringCopy (parms->rates [MAJOR], f4str (majorRateFld), sizeof (Rate));
	stripBlanks (parms->rates [MAJOR]);
	stringCopy (parms->rates [INFO], f4str (infoRateFld), sizeof (Rate));
	stripBlanks (parms->rates [INFO]);
	stringCopy (parms->rates [EVENT], f4str (eventRateFld), sizeof (Rate));
	stripBlanks (parms->rates [EVENT]);

	stringCopy (parms->manualScript, f4str (manScriptFld), sizeof (ScriptName));
	stripBlanks (parms->manualScript);
	stringCopy (parms->autoScript, f4str (autoScriptFld), sizeof (ScriptName));
	stripBlanks (parms->autoScript);

	/* Copy aux mode parameters */

	stringCopy (parms->auxPrinters [MINOR], f4str (auxMinorPrintFld), sizeof (PathName));
	stripBlanks (parms->auxPrinters [MINOR]);
	stringCopy (parms->auxPrinters [MAJOR], f4str (auxMajorPrintFld), sizeof (PathName));
	stripBlanks (parms->auxPrinters [MAJOR]);
	stringCopy (parms->auxPrinters [INFO], f4str (auxInfoPrintFld), sizeof (PathName));
	stripBlanks (parms->auxPrinters [INFO]);
	stringCopy (parms->auxPrinters [EVENT], f4str (auxEventPrintFld), sizeof (PathName));
	stripBlanks (parms->auxPrinters [EVENT]);
	stringCopy (parms->auxPrinters [NORM], f4str (auxNormPrintFld), sizeof (PathName));
	stripBlanks (parms->auxPrinters [NORM]);

	stringCopy (parms->auxSounds [MINOR], f4str (auxMinorSoundFld), sizeof (PathName));
	stripBlanks (parms->auxSounds [MINOR]);
	stringCopy (parms->auxSounds [MAJOR], f4str (auxMajorSoundFld), sizeof (PathName));
	stripBlanks (parms->auxSounds [MAJOR]);
	stringCopy (parms->auxSounds [INFO], f4str (auxInfoSoundFld), sizeof (PathName));
	stripBlanks (parms->auxSounds [INFO]);
	stringCopy (parms->auxSounds [EVENT], f4str (auxEventSoundFld), sizeof (PathName));
	stripBlanks (parms->auxSounds [EVENT]);

	stringCopy (parms->auxRates [MINOR], f4str (auxMinorRateFld), sizeof (Rate));
	stripBlanks (parms->auxRates [MINOR]);
	stringCopy (parms->auxRates [MAJOR], f4str (auxMajorRateFld), sizeof (Rate));
	stripBlanks (parms->auxRates [MAJOR]);
	stringCopy (parms->auxRates [INFO], f4str (auxInfoRateFld), sizeof (Rate));
	stripBlanks (parms->auxRates [INFO]);
	stringCopy (parms->auxRates [EVENT], f4str (auxEventRateFld), sizeof (Rate));
	stripBlanks (parms->auxRates [EVENT]);

	stringCopy (parms->auxManualScript, f4str (auxManScriptFld), sizeof (ScriptName));
	stripBlanks (parms->auxManualScript);
	stringCopy (parms->auxAutoScript, f4str (auxAutoScriptFld), sizeof (ScriptName));
	stripBlanks (parms->auxAutoScript);

	/* Other misc fields */

	stringCopy (parms->auxSelectPV, f4str (auxSelectPVFld), sizeof (PointName));
	stripBlanks (parms->auxSelectPV);
	stringCopy (parms->displayName, f4str (displayNameFld), sizeof (ScriptName));
	stripBlanks (parms->displayName);
	stringCopy (parms->helpTopic, f4str (helpTopicFld), sizeof (HelpTopic));
	stripBlanks (parms->helpTopic);
	stringCopy (parms->guidance, f4str (guidanceFld), sizeof (Guidance));

	return 1;

    }

/*----------*/
void writeInfo (groupName, parms)

    char	*groupName;
    GroupInfo	*parms;

    {

	/* If the record is empty, not write it */

	if (memcmp (parms, &blankParms, sizeof (GroupInfo)) == 0)
	    return;

	d4append_start (infoData, 0);

	/* Copy the group name to the record */

	f4assign (groupNameFld, groupName);

	/* Copy normal mode parameters */

	f4assign (minorPrintFld, parms->printers [MINOR]);
	f4assign (majorPrintFld, parms->printers [MAJOR]);
	f4assign (infoPrintFld, parms->printers [INFO]);
	f4assign (eventPrintFld, parms->printers [EVENT]);
	f4assign (normPrintFld, parms->printers [NORM]);

	f4assign (minorSoundFld, parms->sounds [MINOR]);
	f4assign (majorSoundFld, parms->sounds [MAJOR]);
	f4assign (infoSoundFld, parms->sounds [INFO]);
	f4assign (eventSoundFld, parms->sounds [EVENT]);

	f4assign (minorRateFld, parms->rates [MINOR]);
	f4assign (majorRateFld, parms->rates [MAJOR]);
	f4assign (infoRateFld, parms->rates [INFO]);
	f4assign (eventRateFld, parms->rates [EVENT]);

	f4assign (manScriptFld, parms->manualScript);
	f4assign (autoScriptFld, parms->autoScript);

	/* Copy aux mode parameters */

	f4assign (auxMinorPrintFld, parms->auxPrinters [MINOR]);
	f4assign (auxMajorPrintFld, parms->auxPrinters [MAJOR]);
	f4assign (auxInfoPrintFld, parms->auxPrinters [INFO]);
	f4assign (auxEventPrintFld, parms->auxPrinters [EVENT]);
	f4assign (auxNormPrintFld, parms->auxPrinters [NORM]);

	f4assign (auxMinorSoundFld, parms->auxSounds [MINOR]);
	f4assign (auxMajorSoundFld, parms->auxSounds [MAJOR]);
	f4assign (auxInfoSoundFld, parms->auxSounds [INFO]);
	f4assign (auxEventSoundFld, parms->auxSounds [EVENT]);

	f4assign (auxMinorRateFld, parms->auxRates [MINOR]);
	f4assign (auxMajorRateFld, parms->auxRates [MAJOR]);
	f4assign (auxInfoRateFld, parms->auxRates [INFO]);
	f4assign (auxEventRateFld, parms->auxRates [EVENT]);

	f4assign (auxManScriptFld, parms->auxManualScript);
	f4assign (auxAutoScriptFld, parms->auxAutoScript);

	/* Other misc fields */

	f4assign (auxSelectPVFld, parms->auxSelectPV);
	f4assign (displayNameFld, parms->displayName);
	f4assign (helpTopicFld, parms->helpTopic);
	f4assign (guidanceFld, parms->guidance);

	d4append (infoData);

    }

/*----------*/
int openPvMapFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_PVMAP_FILE_NAME, pathName);
	pvMapData = d4open (&codeBase, pathName);
	if (pvMapData == NULL)
	    return 0;
	pvNameFld = d4field (pvMapData, "NAME");
	pvGroupFld = d4field (pvMapData, "ALARMGROUP");
	pvDescriptorFld = d4field (pvMapData, "DESC");
	pvDeleteFlagFld = d4field (pvMapData, "DELETEFLAG");
	pvResNameFld = d4field (pvMapData, "RESRCNAME");
	d4top (pvMapData);
	d4unlock_all (pvMapData);
	return 1;
	
    }

/*----------*/
void closePvMapFile (VOID)

    {

	d4close (pvMapData);

    }

/*----------*/
int getNumPVs (VOID)

    {

	return d4reccount (pvMapData);

    }

/*----------*/
int getNextPv (pvMapInfo)

    PvMapInfo	*pvMapInfo;

    {

	if (d4eof (pvMapData))
	    return 0;

	/* Skip any deleted records */
	    
	while (f4true (pvDeleteFlagFld)) {
	    d4skip (pvMapData, 1L);
	    d4unlock_all (pvMapData);
	    if (d4eof (pvMapData))
		return 0;
	}

	stringCopy (pvMapInfo->pointName, f4str (pvNameFld), sizeof (PointName));
	stripBlanks (pvMapInfo->pointName);
	stringCopy (pvMapInfo->groupName, f4str (pvGroupFld), sizeof (GroupName));
	stripBlanks (pvMapInfo->groupName);
	stringCopy (pvMapInfo->descriptor, f4str (pvDescriptorFld), sizeof (Descriptor));
	stripBlanks (pvMapInfo->descriptor);
	stringCopy (pvMapInfo->resName, f4str (pvResNameFld), sizeof (Resource));
	stripBlanks (pvMapInfo->resName);

	d4skip (pvMapData, 1L);
	d4unlock_all (pvMapData);
	return 1;
	
    }

/*----------*/
int getPv (pointName, groupName, descriptor, resource)	/* Find a specific point */

    char	*pointName, *groupName, *descriptor, *resource;

    {

	char	key [50];

	sprintf (key, "%-*s", AL_POINTNAME_SIZE, pointName);
	if (d4seek (pvMapData, key) != 0){
	    d4unlock_all (pvMapData);
	    return 0;
	}
	if (f4true (pvDeleteFlagFld)) {
	    d4unlock_all (pvMapData);
	    return 0;
	}
	stringCopy (groupName, f4str (pvGroupFld), sizeof (GroupName));
	stripBlanks (groupName);
	stringCopy (descriptor, f4str (pvDescriptorFld), sizeof (Descriptor));
	stripBlanks (descriptor);
	stringCopy (resource, f4str (pvResNameFld), sizeof (Resource));
	stripBlanks (resource);
	d4unlock_all (pvMapData);
	return 1;

    }

char	currentPoint[50];
char	currentType[16];
char	currentFldname[7];

/*----------*/
int	getPvPosition (pointName)	/* Find a specific point */
char	*pointName;
{
    char	key [50];

    if (strcmp (currentPoint, pointName) == 0)
	return 1;

    sprintf (key, "%-*s", AL_POINTNAME_SIZE, pointName);
    if (d4seek (pvMapData, key) != 0)
    {
	d4unlock_all (pvMapData);
	return 0;
    }
    if (f4true (pvDeleteFlagFld)) {
	d4unlock_all (pvMapData);
	return 0;
    }
    strncpy (currentPoint, pointName, sizeof(currentPoint));
    d4unlock_all (pvMapData);
    return 1;
}

/*----------*/
char	*GetPVPointInfo ( char *pvName, char *dbfld )
{
    if (getPvPosition (pvName))
    {
	return f4str (d4field (pvMapData, dbfld));
    }
    return "***NOT FOUND***";
}

/*----------*/
void ResetPVPoint ( )
{
    initializeDatabase ();
    currentPoint [0] = 0;
    currentType [0] = 0;
}

/*----------*/
int openPdbXrefFile (VOID)
{
    char	pathName [200];

    resolveLinks (AL_PDBXREF_FILE_NAME, pathName);
    pdbXrefData = d4open (&codeBase, pathName);
    if (pdbXrefData == NULL)
	return 0;

    d4unlock_all (pdbXrefData);
    return 1;
}

/*----------*/
void closePdbXrefFile (VOID)
{
    d4close (pdbXrefData);
}

/*----------*/
char	*GetPdbXrefInfo ( char *pointType, char *fldname, char *dbfld)
{
    char	key [50];

    if ((strcmp (currentType, pointType) != 0)
       || (strcmp (currentFldname, fldname) != 0))
    {
	d4tag_select (pdbXrefData, d4tag (pdbXrefData, "XREF"));
	sprintf (key, "%-*s%-*s", AL_POINTTYPE_SIZE, pointType, AL_FIELDNAME_SIZE, fldname);
	if (d4seek (pdbXrefData, key) != r4success)
	{
	    d4unlock_all (pdbXrefData);
	    return NULL;
	}
    	if (d4deleted (pdbXrefData)) {
	    d4unlock_all (pdbXrefData);
	    return NULL;
	}
	strncpy (currentType, pointType, sizeof(currentType));
	strncpy (currentFldname, fldname, sizeof(currentFldname));
	d4unlock_all (pdbXrefData);
    }
    return f4str (d4field (pdbXrefData, dbfld));
}

/*----------*/
char	*GetDbInfo ( char *dbfile, char *pointName, char *dbfld )
{
    char	key [50];
    char	pathName [200];
    static char returnString [200];

    if (strcmp (currentType, dbfile) != 0)
    {
	strncpy (dbfilename, AL_DBASE_DIRECTORY_NAME, sizeof(dbfilename));
	stripBlanks (dbfile);
	strncat (dbfilename, dbfile, sizeof(dbfilename));
	strncat (dbfilename, ".dbf", sizeof(dbfilename));
	resolveLinks (dbfilename, pathName);
	dbData = d4open (&codeBase, pathName);
	if (dbData == NULL)
	    return NULL;
    }
    sprintf (key, "%-*s", AL_POINTNAME_SIZE, pointName);
    if (d4seek (dbData, key) != 0)
    {
	d4unlock_all (dbData);
	return NULL;
    }
    if (d4deleted (dbData)) {
	d4unlock_all (dbData);
	return NULL;
    }
    d4unlock_all (dbData);

    strncpy (returnString, f4str (d4field (dbData, dbfld)), sizeof (returnString));
    d4close (dbData);
    return returnString;
}

#define MAX_LIMITS_FIELDS	4

typedef struct {
	char *names [MAX_LIMITS_FIELDS];
} LIMITFIELDS;

LIMITFIELDS fldNames [] = {
	{ 0, 0, 0, 0 },			/* RETRN NORMAL */
	{ 0, 0, 0, 0 },			/* READ         */
	{ 0, 0, 0, 0 },			/* WRITE        */
	{ "HIHI" },		/* HIHI         */	
	{ "HIGH" },		/* HIGH         */	
	{ "LOLO" },		/* LOLO         */	
	{ "LOW" },		/* LOW          */	
	{ 0, 0, 0, 0 },			/* SBAD         */
	{ "ROC" },		/* ROC          */
	{ 0, 0, 0, 0 },			/* STATE        */
	{ 0, 0, 0, 0 },			/* COS          */
	{ 0, 0, 0, 0 },			/* OPEN T/C     */
	{ 0, 0, 0, 0 },			/* TIMEOUT      */
	{ 0, 0, 0, 0 },			/* HW_LIMIT     */
	{ 0, 0, 0, 0 },			/* CALC         */
	{ 0, 0, 0, 0 },			/* SCAN         */
	{ 0, 0, 0, 0 },			/* LINK         */
	{ 0, 0, 0, 0 },			/* SOFT         */
	{ 0, 0, 0, 0 },			/* BAD UDB      */
	{ 0, 0, 0, 0 },			/* UDF          */
	{ 0, 0, 0, 0 },			/* INHIBIT      */
	{ 0, 0, 0, 0 },			/* TAGGED       */
	{ 0, 0, 0, 0 },			/* CMD NOT ALLW */
	{ "VOLD" },		/* VOLUME       */
	{ "HHLV" },		/* HIHI LEVEL   */
	{ 0, 0, 0, 0 },			/* TEMPERATURE  */
	{ 0, 0, 0, 0 },			/* PRESSURE     */
	{ "HIBV" },		/* BSW          */
	{ 0, 0, 0, 0 },			/* TANK_VALVE   */
	{ "EXOP", "VJOT" },	/* VLV OPEN ALM */
	{ "EXCP", "VJCT" },	/* VLV CLS  ALM */ 
	{ "MLTL" },		/* VLV TORQ LAG */ 
	{ 0, 0, 0, 0 },			/* VLV FAULT    */
	{ "LOPR", "HOPR" },	/* VLV POSITION */ 
	{ 0, 0, 0, 0 },			/* AUTOMATIC    */
	{ 0, 0, 0, 0 },			/* VIBRATION    */
	{ "EXON" },		/* PUMP START   */	
	{ "EXOF" },		/* PUMP STOP    */	
	{ 0, 0, 0, 0 },			/* POT DROP RDY */
	{ 0, 0, 0, 0 },			/* PERMISSIVE   */
	{ 0, 0, 0, 0 },			/* UNCOMMANDED  */
	{ "RLOV", "THLD", "TERR" }, /* THRSHLD EXCD */ 
	{ 0, 0, 0, 0 },			/* MV ERROR     */
	{ 0, 0, 0, 0 },			/* OUTPUT ERROR */
	{ "DEVL" },		/* DEVIATION    */	
	{ 0, 0, 0, 0 },			/* TRACKNG MODE */
	{ "STHC", "CLMH" },	/* SETPT HI CLP */ 
	{ "STLC", "CLML" },	/* SETPT LO CLP */ 
	{ "OUTH" },		/* OUTPT HI CLP */ 
	{ "OUTL" },		/* OUTPT LO CLP */ 
	{ 0, 0, 0, 0 },			/* MAN OVERRIDE */
	{ 0, 0, 0, 0 },			/* SETPT ERROR  */
	{ 0, 0, 0, 0 },			/* OUT MODE ERR */
	{ 0, 0, 0, 0 },			/* PARAMETER    */
	{ "HILV" },		/* HIGH LEVEL   */	
	{ "LOLV" },		/* LOW LEVEL    */	
	{ "LLLV" },		/* LOLO LEVEL   */	
	{ "HHTV" },		/* HIHI TEMP    */	
	{ "HITV" },		/* HIGH TEMP    */	
	{ "LOTV" },		/* LOW TEMP     */	
	{ "LLTV" },		/* LOLO TEMP    */	
	{ 0, 0, 0, 0 },			/* NO INIT      */
};

static char limitsStr [AL_INFO_SIZE];

/*----------*/
char *GetLimitValue (char *pvName, int status)
{
    char	pointType[50];
    char	fieldName[50];
    char	*valString;
    int		i;

    if (fldNames[status].names[0] && getPvPosition (pvName))
    {
	strncpy (pointType, GetPVPointInfo (pvName, "TYPE"), sizeof(pointType));
	for (i = 0, limitsStr[0] = 0; fldNames[status].names[i] && i < MAX_LIMITS_FIELDS; i++)
	{
	    if (GetPdbXrefInfo (pointType, fldNames[status].names[i], "FIELDNAME"))
	    {
		strncpy (fieldName,
			 GetPdbXrefInfo (pointType, fldNames[status].names[i], "FIELD_NAME"),
			 sizeof (fieldName));
		if (i > 0)
		    strcat (limitsStr, ", ");
		valString = GetDbInfo (GetPdbXrefInfo (pointType, fldNames[status].names[i], "TABLE"),
				    pvName,
				    fieldName);
		if (valString)
		    strcat (limitsStr, valString);
	    }
	}
	return limitsStr;
    }
    return 0;
}
/*----------*/
int openIOCListFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_IOCLIST_FILE_NAME, pathName);
	iocListData = d4open (&codeBase, pathName);
	if (iocListData == NULL)
	    return 0;
	iocNameFld = d4field (iocListData, "IOCNAME");
	iocIsHostFld = d4field (iocListData, "IS_HOST");
	d4top (iocListData);
	d4unlock_all (iocListData);
	return 1;
	
    }

/*----------*/
void closeIOCListFile (VOID)

    {

	d4close (iocListData);

    }

/*----------*/
int getNextIOCEntry (iocName, isHost)

    char	*iocName;
    int		*isHost;

    {

	if (d4eof (iocListData))
	    return 0;

	/* Get the name since a zero length name indicates a deleted record */
	
	stringCopy (iocName, f4str (iocNameFld), sizeof (HostName));
	stripBlanks (iocName);

	/* Skip any deleted records */
	    
	while (d4deleted (iocListData) || (strlen (iocName) == 0)) {
	    d4skip (iocListData, 1L);
	    stringCopy (iocName, f4str (iocNameFld), sizeof (HostName));
	    stripBlanks (iocName);
	    if (d4eof (iocListData))
		return 0;
	}

	*isHost = f4true (iocIsHostFld);

	d4skip (iocListData, 1L);
	d4unlock_all (iocListData);
	return 1;
	
    }

#if defined(INFOFIELDS) || defined(WIN32)
/********************************************************************************************************/
#define CONVERTSUB
#ifdef WIN32
#define TIS4000DB   "\\tis4000db\\"
#else
#define TIS4000DB   "/tis4000db/"
#endif WIN32
#define DBCNVRT	    "alarmsave"
#define FIELDINFO   alarmLogFields 
#define TAGINFO	    alarmLogTags
#define MEMOFILE

/* define name of codebase structure in local program */
#define cb	codeBase

#include "convertdb.h"
#endif INFOFIELDS
/*----------*/
int openLogFile (VOID)	/* Open/create the alarm log file */

    {

	T4TAG	*t;
	char	pathName [200];

	resolveLinks (AL_LOG_FILE_NAME, pathName);
	logData = d4open (&codeBase, pathName);
	if (logData == NULL)
	{
	    if (createLogFile () == 0)
		return 0;
	    else
		logData = d4open (&codeBase, pathName);
	}
#if defined(INFOFIELDS) || defined(WIN32)
	else
	{
	    if ((logInfo1Fld = d4field (logData, "INFO1")) == NULL)
	    {
		char filename [32];
		char *cp;

		/* convert to new format */
		d4close( logData );

		/* convertDB needs the filename part of the name! */
		cp = pathName;
		while (strchr (cp, '/'))
		    cp = strchr (cp, '/') + 1;
		/* without the .dbf nonsense */
		strcpy (filename, cp);
		if (strchr (filename, '.'))
		    *strchr (filename, '.') = 0;
		convertDB (filename);

		/* Open the converted file */
		if (!( logData = d4open (&codeBase, pathName)))
		    return 0;		/* Can't open converted file */
	    }
	}
#endif INFOFIELDS
	t = d4tag (logData, "EVENTTIME");
	d4tag_select (logData, t);
	d4top (logData);
	d4unlock_all (logData);

	logIsAckFld = d4field (logData, "ISACK");
	logDateFld = d4field (logData, "DATE");
	logTimeFld = d4field (logData, "TIME");
	logAlarmSecsFld = d4field (logData, "ALARMSECS");
	logAlarmMsecsFld = d4field (logData, "ALARMMSECS");
	logGroupNameFld = d4field (logData, "GROUPNAME");
	logPointNameFld = d4field (logData, "POINTNAME");
	logDescriptorFld = d4field (logData, "DESCRIPTOR");
	logStatusFld = d4field (logData, "STATUS");
	logSeverityFld = d4field (logData, "SEVERITY");
	logAckedFld = d4field (logData, "ACKED");
	logNoPrintFld = d4field (logData, "NOPRINT");
	logOperatorFld = d4field (logData, "OPERATOR");
	logResourceFld = d4field (logData, "RESOURCE");
        logStationNameFld = d4field (logData, "STATION");
	logEguTagFld = d4field (logData, "EGUTAG");
	logValueFld = d4field (logData, "VALUE");
	logAckSecsFld = d4field (logData, "ACKSECS");
	logAckMsecsFld = d4field (logData, "ACKMSECS");
	logCommentFld = d4field (logData, "COMMENT");
#if defined(INFOFIELDS) || defined(WIN32)
	logInfo1Fld = d4field (logData, "INFO1");
	logInfo2Fld = d4field (logData, "INFO2");
	logInfo3Fld = d4field (logData, "INFO3");
	logInfo4Fld = d4field (logData, "INFO4");
	logInfo5Fld = d4field (logData, "INFO5");
#endif INFOFIELDS

	return 1;

    }

/*----------*/
int createLogFile (VOID)

    {
	char	pathName [200];

	resolveLinks (AL_LOG_IX_FILE_NAME, pathName);
	unlink (pathName);
	resolveLinks (AL_LOG_MEMO_FILE_NAME, pathName);
	unlink (pathName);
	resolveLinks (AL_LOG_FILE_NAME, pathName);
	unlink (pathName);
	logData = d4create (&codeBase, pathName, alarmLogFields, alarmLogTags);
	if (logData == NULL)
	    return 0;
	d4close (logData);
	return 1;

    }

/*----------*/
void closeLogFile (VOID)

    {

	updateDBaseFile (logData);
	d4close (logData);

    }

/*----------*/
int getLogEntry (key, entry, isAck)

    LogEntry	*entry;
    int		*isAck;
    char	*key;

    {

	if (d4seek (logData, key) != 0) {
	    d4unlock_all (logData);
	    return 0;
	}

	getLogFields (entry);

	*isAck = f4char (logIsAckFld) == 'Y';

	d4unlock_all (logData);

	return 1;
		
    }

/*----------*/
int getNextLogEntry (entry, isAck)

    LogEntry	*entry;
    int		*isAck;

    {

	if (d4eof (logData)) {
	    d4unlock_all (logData);
	    return 0;
	}

	/* Get the name since a zero length name indicates a deleted record */
	
	stringCopy (entry->pointName, f4str (logPointNameFld), sizeof (PointName));
	stripBlanks (entry->pointName);

	/* Skip any deleted records */
	    
	while (d4deleted (logData) || (strlen (entry->pointName) == 0)) {
	    d4skip (logData, 1L);
	    d4unlock_all (logData);
	    stringCopy (entry->pointName, f4str (logPointNameFld), sizeof (PointName));
	    stripBlanks (entry->pointName);
	    if (d4eof (logData)) {
		d4unlock_all (logData);
		return 0;
	    }
	}

	getLogFields (entry);

	*isAck = f4char (logIsAckFld) == 'Y';

	d4skip (logData, 1L);

	d4unlock_all (logData);

	return 1;
		
    }

/*----------*/
static void getLogFields (entry)

    LogEntry	*entry;

    {

	char	buf [100];

	/* Copy the items to the LogEntry structure */

	stringCopy (entry->groupName, f4str (logGroupNameFld), sizeof (GroupName));
	stripBlanks (entry->groupName);
	stringCopy (entry->pointName, f4str (logPointNameFld), sizeof (PointName));
	stripBlanks (entry->pointName);
	stringCopy (entry->descriptor, f4str (logDescriptorFld), sizeof (Descriptor));
	stripBlanks (entry->descriptor);
	stringCopy (entry->oper, f4str (logOperatorFld), sizeof (Operator));
	stripBlanks (entry->oper);
	stringCopy (entry->resource, f4str (logResourceFld), sizeof (Resource));
	stripBlanks (entry->resource);
	stringCopy (entry->stationName, f4str (logStationNameFld), sizeof (IOCName));
	stripBlanks (entry->stationName);
	stringCopy (entry->eguTag, f4str (logEguTagFld), sizeof (EguTag));
	stripBlanks (entry->eguTag);
	stringCopy (entry->value, f4str (logValueFld), sizeof (PvValue));
	stripBlanks (entry->value);
	stringCopy (entry->status, f4str (logStatusFld), sizeof (alStatus));
	stripBlanks (entry->status);
	stringCopy (entry->comment, (char *)m4str (logCommentFld), sizeof (Comment));
	if (entry->severity != -1)
	    entry->severity = strchr (severityChars, f4char (logSeverityFld)) - severityChars;

	entry->acked = f4true (logAckedFld);
	entry->noPrint = f4true (logNoPrintFld);
#if defined(INFOFIELDS) || defined(WIN32)
	stringCopy (entry->info1, f4str (logInfo1Fld), AL_INFO_SIZE);
	stripBlanks (entry->info1);
	stringCopy (entry->info2, f4str (logInfo2Fld), AL_INFO_SIZE);
	stripBlanks (entry->info2);
	stringCopy (entry->info3, f4str (logInfo3Fld), AL_INFO_SIZE);
	stripBlanks (entry->info3);
	stringCopy (entry->info4, f4str (logInfo4Fld), AL_INFO_SIZE);
	stripBlanks (entry->info4);
	stringCopy (entry->info5, f4str (logInfo5Fld), AL_INFO_SIZE);
	stripBlanks (entry->info5);
#endif INFOFIELDS

	/* The timestamps are stored as seconds and milliseconds */

	entry->logTime.secPastEpoch = atol (f4str (logAlarmSecsFld)) - TS_EPOCH_SEC_PAST_1970;
	entry->logTime.nsec = atol (f4str (logAlarmMsecsFld)) * 1000000;

	stringCopy (buf, f4str (logAckSecsFld), AL_GMTSECS_SIZE);
	stripBlanks (buf);
	if (strlen (buf) > 0) {
	    entry->ackTime.secPastEpoch = atol (buf) - TS_EPOCH_SEC_PAST_1970;
	    entry->ackTime.nsec = atol (f4str (logAckMsecsFld)) * 1000000;
	} else {
	    entry->ackTime.secPastEpoch = 0;
	    entry->ackTime.nsec = 0;
	}

    }

/*----------*/
void writeLogEntry (entry)

    LogEntry	*entry;

    {

	int		doAppend, result;
	char		buf [20];
	time_t		t;
	struct tm	*tim;

	/* This is a trap for acktime == logtime */

	/* Seek to the beginning of the file to determine if we have an empty record */
	/* we can reclaim. */

	doAppend = 1;
	if (d4top (logData) == 0) {
	    if (d4deleted (logData)) {
		d4recall (logData);
		doAppend = 0;
	    }
	}

	if (doAppend) {
	    d4go_eof (logData);
	    result = d4append_start (logData, 0);
	}
	d4blank (logData);

	/* Lets do the simple stuff first */

	f4assign (logGroupNameFld, entry->groupName);
	f4assign (logPointNameFld, entry->pointName);
	f4assign (logDescriptorFld, entry->descriptor);
	f4assign (logOperatorFld, entry->oper);
	f4assign (logResourceFld, entry->resource);
        f4assign (logStationNameFld, entry->stationName);
	f4assign (logEguTagFld, entry->eguTag);
	f4assign (logValueFld, entry->value);
	f4assign (logStatusFld, entry->status);
	m4assign (logCommentFld, entry->comment);
	f4assign_char (logSeverityFld, severityChars [entry->severity]);
	f4assign_char (logAckedFld, entry->acked ? 'Y' : 'N');
	f4assign_char (logNoPrintFld, entry->noPrint ? 'Y' : 'N');
#if defined(INFOFIELDS) || defined(WIN32)
	f4assign ( logInfo1Fld, entry->info1 );
	f4assign ( logInfo2Fld, entry->info2 );
	f4assign ( logInfo3Fld, entry->info3 );
	f4assign ( logInfo4Fld, entry->info4 );
	f4assign ( logInfo5Fld, entry->info5 );
#endif INFOFIELDS

	/* The timestamps are stored as seconds and milliseconds */

	sprintf (buf, "%*ld", AL_GMTSECS_SIZE,
			entry->logTime.secPastEpoch + TS_EPOCH_SEC_PAST_1970);
	f4assign (logAlarmSecsFld, buf);
	sprintf (buf, "%*ld", AL_GMTMSECS_SIZE,
			entry->logTime.nsec / 1000000);
	f4assign (logAlarmMsecsFld, buf);

	if (entry->ackTime.secPastEpoch != 0) {		
	    sprintf (buf, "%*ld", AL_GMTSECS_SIZE,
			entry->ackTime.secPastEpoch + TS_EPOCH_SEC_PAST_1970);
	    f4assign (logAckSecsFld, buf);
	    sprintf (buf, "%*ld", AL_GMTMSECS_SIZE,
			entry->ackTime.nsec / 1000000);
	    f4assign (logAckMsecsFld, buf);
	}

	/* The alarm time is also stored in a nice human readable format */

	t = entry->logTime.secPastEpoch + TS_EPOCH_SEC_PAST_1970;
	tim = localtime (&t);
	sprintf (buf, "%02d/%02d/%02d", tim->tm_mon + 1, tim->tm_mday, tim->tm_year % 100);
	f4assign (logDateFld, buf);
	sprintf (buf, "%02d:%02d:%02d", tim->tm_hour, tim->tm_min, tim->tm_sec);
	f4assign (logTimeFld, buf);

	/* Set record type */

	f4assign_char (logIsAckFld, entry->isAck ? 'Y' : 'N');

	if (doAppend)
	    result = d4append (logData);

	updateDBaseFile (logData);

    }

/*----------*/
int writeLogMemo (pointName, ts, msg)	/* Write a memo to a log field */

    char	*pointName, *msg;
    TS_STAMP	*ts;

    {

	char	key [100];

	/* Lookup the record, insert the memo, and write it. */

	makeLogKeyFromTS (pointName, ts, 0, key);
	if (d4seek (logData, key) != 0) {
	    d4unlock_all (logData);
	    return 0;
	}

	m4assign (logCommentFld, msg);
	
	updateDBaseFile (logData);

	return 1;

    }

/*----------*/
void makeLogKeyFromTS (pointName, ts, isAck, key)	/* Make a key for the log file */

    char	*pointName, *key;
    TS_STAMP	*ts;
    int		isAck;

    {

	long	secs;
	int	msecs;

	secs = ts->secPastEpoch + TS_EPOCH_SEC_PAST_1970;
	msecs = ts->nsec / 1000000;

	sprintf (key, "%*ld%*d%-*s%c",
		AL_GMTSECS_SIZE, secs,
		AL_GMTMSECS_SIZE, msecs,
 		AL_POINTNAME_SIZE, pointName,
		isAck ? 'Y' : 'N');

    }

/*----------*/
void makeLogKeyFromEntry (entry, isAck, key)	/* Make a key for the log file */

    char	*key;
    LogEntry	*entry;
    int		isAck;

    {

	long	secs;
	int	msecs;

	secs = entry->logTime.secPastEpoch + TS_EPOCH_SEC_PAST_1970;
	msecs = entry->logTime.nsec / 1000000;

	sprintf (key, "%*ld%*d%-*s%c",
		AL_GMTSECS_SIZE, secs,
		AL_GMTMSECS_SIZE, msecs,
 		AL_POINTNAME_SIZE, entry->pointName,
		isAck ? 'Y' : 'N');

    }

/*----------*/
int setAckInLog (pointName, ts, ackTime)	/* Set the ACK flag in the specified log record */

    char	*pointName;
    TS_STAMP	*ts, *ackTime;

    {

	char		key [100], buf [20], secBuf [20], msecBuf [20];
	TS_STAMP	save;

	strcpy (secBuf, "");
	strcpy (msecBuf, "");
	if (ackTime->secPastEpoch != 0) {		
	    sprintf (secBuf, "%*ld", AL_GMTSECS_SIZE, ackTime->secPastEpoch + TS_EPOCH_SEC_PAST_1970);
	    sprintf (msecBuf, "%*ld", AL_GMTMSECS_SIZE, ackTime->nsec / 1000000);
	}

	/* Make a key and find the record */

	save = *ts;
	makeLogKeyFromTS (pointName, ts, LOG_ALARM, key);
	if (d4seek (logData, key) == 0) {
	    strcpy (buf, f4str (logAckSecsFld));
	    stripBlanks (buf);
	    if (strlen (buf) > 0) {
		ts->secPastEpoch = atol (buf) - TS_EPOCH_SEC_PAST_1970;
		ts->nsec = atol (f4str (logAckMsecsFld)) * 1000000;
	    } else {
		ts->secPastEpoch = 0;
		ts->nsec = 0;
	    }
	    f4assign_char (logAckedFld, 'Y');
	    f4assign (logAckSecsFld, secBuf);
	    f4assign (logAckMsecsFld, msecBuf);
	    updateDBaseFile (logData);
	    if (TsCmpStampsEQ (&save, ts)) {
		ts->secPastEpoch = 0;
		ts->nsec = 0;
		d4unlock_all (logData);
		return 0;
	    }
	    return 1;
	}
	d4unlock_all (logData);
	ts->secPastEpoch = 0;
	ts->nsec = 0;
	return 0;

    }

/*----------*/
void openResListFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_RESLIST_FILE_NAME, pathName);
	resListData = d4open (&codeBase, pathName);
	if (resListData == NULL)
	    return;
	ioc1Fld = d4field (resListData, "PRIMARYIOC");
	ioc2Fld = d4field (resListData, "SECNDRYIOC");
	d4top (resListData);
	d4unlock_all (resListData);

    }

/*----------*/
void closeResListFile (VOID)

    {

        d4close (resListData);

    }

/*----------*/
int getResListEntry (char *resName, ResListInfo *resListInfo)

    {

        char key [AL_RESOURCE_SIZE+1];

	sprintf (key, "%-*s", AL_RESOURCE_SIZE, resName);
	if (d4seek (resListData, key) == 0) {
	    stringCopy (resListInfo->ioc1, f4str (ioc1Fld), AL_IOCNAME_SIZE);
	    stripBlanks (resListInfo->ioc1);
	    stringCopy (resListInfo->ioc2, f4str (ioc2Fld), AL_IOCNAME_SIZE);
	    stripBlanks (resListInfo->ioc2);
	    return 1;
	} else
	    return 0;

    }

/*----------*/
int createPointStateFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_POINT_STATE_NAME, pathName);
	psData = d4create (&codeBase, pathName, pointStateFields, pointStateTags);
	if (psData == NULL)
	    return 0;
	d4close (psData);
	return openPointStateFile ();

    }

/*----------*/
int openPointStateFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_POINT_STATE_NAME, pathName);
	psData = d4open (&codeBase, pathName);
	if (psData == NULL)
	    return 0;

	psPointNameFld = d4field (psData, "POINTNAME");
	psAlarmSecsFld = d4field (psData, "ALARMSECS");
	psAlarmNsecsFld = d4field (psData, "ALARMNSECS");
	psAckSecsFld = d4field (psData, "ACKSECS");
	psAckNsecsFld = d4field (psData, "ACKNSECS");
	psTimeSecsFld = d4field (psData, "TIMESECS");
	psTimeNsecsFld = d4field (psData, "TIMENSECS");
	psStatusFld = d4field (psData, "STATUS");
	psSeverityFld = d4field (psData, "SEVERITY");
	psValueFld = d4field (psData, "VALUE");
	psUstatusFld = d4field (psData, "USTATUS");
	psUseverityFld = d4field (psData, "USEVERITY");
	psUvalueFld = d4field (psData, "UVALUE");
	psAckFld = d4field (psData, "ACK");
	psRtnFld = d4field (psData, "RTN");
	psAenbFld = d4field (psData, "AENB");
	psOscnFld = d4field (psData, "OSCN");
	psDescriptorFld = d4field (psData, "DESCRIPTOR");
	psEgutagFld = d4field (psData, "EGUTAG");
	psOperatorFld = d4field (psData, "OPERATOR");
	psInSummaryFld = d4field (psData, "INSUM");
	d4top (psData);
	return 1;

    }

/*----------*/
void closePointStateFile (VOID)

    {

	updateDBaseFile (psData);
	d4close (psData);

    }

/*----------*/
int getPointState (pointName, pvData, inSummary)

    char	*pointName;
    PvData	*pvData;
    int		*inSummary;

    {

	char	key [50];

	sprintf (key, "%-*s", AL_POINTNAME_SIZE, pointName);
	if (d4seek (psData, key) != 0) {
	    d4unlock_all (psData);
	    return 0;
	}

	if (d4deleted (psData))
	    return 0;

	pvData->lastAlarm.secPastEpoch = f4long (psAlarmSecsFld);
	pvData->lastAlarm.nsec = f4long (psAlarmNsecsFld);
	pvData->lastAck.secPastEpoch = f4long (psAckSecsFld);
	pvData->lastAck.nsec = f4long (psAckNsecsFld);
	pvData->lastAlarmReceived.secPastEpoch = f4long (psTimeSecsFld);
	pvData->lastAlarmReceived.nsec = f4long (psTimeNsecsFld);
	pvData->status = f4int (psStatusFld);
	pvData->severity = f4int (psSeverityFld);
	stringCopy (pvData->value, f4str (psValueFld), sizeof (PvValue));
	stripBlanks (pvData->value);
	pvData->unAckedStatus = f4int (psUstatusFld);
	pvData->unAckedSeverity = f4int (psUseverityFld);
	stringCopy (pvData->unAckedValue, f4str (psUvalueFld), sizeof (PvValue));
	stripBlanks (pvData->unAckedValue);
	pvData->Ack = f4int (psAckFld);
	pvData->RTN = f4int (psRtnFld);
	pvData->aenb = f4int (psAenbFld);
	pvData->oscn = f4int (psOscnFld);
	stringCopy (pvData->description, f4str (psDescriptorFld), sizeof (Descriptor));
	stripBlanks (pvData->description);
	stringCopy (pvData->eguTag, f4str (psEgutagFld), sizeof (EguTag));
	stripBlanks (pvData->eguTag);
	stringCopy (pvData->op, f4str (psOperatorFld), sizeof (Operator));
	stripBlanks (pvData->op);
	*inSummary = f4int (psInSummaryFld);

	d4unlock_all (psData);

	return 1;

    }

/*----------*/
void writePointState (pvData)

    PvData	*pvData;

    {

	int	doAppend = 0;
	char	key [50];

	sprintf (key, "%-*s", AL_POINTNAME_SIZE, pvData->pvName);
	if (d4seek (psData, key) != 0) {
	    d4append_start (psData, 0);
	    doAppend = 1;
	}

	d4recall (psData);

	f4assign (psPointNameFld, pvData->pvName);
	f4assign_long (psAlarmSecsFld, pvData->lastAlarm.secPastEpoch);
	f4assign_long (psAlarmNsecsFld, pvData->lastAlarm.nsec);
	f4assign_long (psAckSecsFld, pvData->lastAck.secPastEpoch);
	f4assign_long (psAckNsecsFld, pvData->lastAck.nsec);
	f4assign_long (psTimeSecsFld, pvData->lastAlarmReceived.secPastEpoch);
	f4assign_long (psTimeNsecsFld, pvData->lastAlarmReceived.nsec);
	f4assign_int (psStatusFld, (int)(pvData->status));
	f4assign_int (psSeverityFld, (int)(pvData->severity));
	f4assign (psValueFld, pvData->value);
	f4assign_int (psUstatusFld, (int)(pvData->unAckedStatus));
	f4assign_int (psUseverityFld, (int)(pvData->unAckedSeverity));
	f4assign (psUvalueFld, pvData->unAckedValue);
	f4assign_int (psAckFld, (int)(pvData->Ack));
	f4assign_int (psRtnFld, (int)(pvData->RTN));
	f4assign_int (psAenbFld, (int)(pvData->aenb));
	f4assign_int (psOscnFld, (int)(pvData->oscn));
	f4assign (psDescriptorFld, pvData->description);
	f4assign (psEgutagFld, pvData->eguTag);
	f4assign (psOperatorFld, pvData->op);
	f4assign_int (psInSummaryFld, pvData->prevSummaryPv != (PvData *)-1);

	if (doAppend)
	    d4append (psData);

	updateDBaseFile (psData);

    }

/*----------*/
int deletePointState (pointName)	/* Delete a state record */

    char	*pointName;

    {

	char	key [50];

	sprintf (key, "%-*s", AL_POINTNAME_SIZE, pointName);
	if (d4seek (psData, key) != 0) {
	    d4unlock_all (psData);
	    return 0;
	}

	d4delete (psData);
	updateDBaseFile (psData);
	return 1;

    }

/*----------*/
int createPrinterFile (VOID)

    {

	char	pathName [200];

	resolveLinks (AL_PRINTER_FILE_NAME, pathName);
	unlink (pathName);
	printerData = d4create (&codeBase, pathName, alarmPrinterFields, NULL);
	if (printerData == NULL)
	    return 0;
	d4close (printerData);
	return openPrinterFile ();

    }

/*----------*/
int openPrinterFile (VOID)

    {

	char	pathName [200], buf [10];
	int	i;

	resolveLinks (AL_PRINTER_FILE_NAME, pathName);
	printerData = d4open (&codeBase, pathName);
	if (printerData == NULL)
	    return 0;
	printerGroupNameFld = d4field (printerData, "NAME");
	for (i=0 ; i<MAX_PRINTERS_PER_GROUP ; i++) {
	    sprintf (buf, "PTR%d", i + 1);
	    printerPrinterFld [i] = d4field (printerData, buf);
	}
	printerRecNum = 1;
	printerNumRecs = d4reccount (printerData);
	d4unlock_all (printerData);
	return 1;

    }

/*----------*/
void closePrinterFile (VOID)

    {

	updateDBaseFile (printerData);
	d4close (printerData);

    }

/*----------*/
int getNextPrinterGroup (printerEntry)

    PrinterGroupEntry	*printerEntry;

    {

	int	i;

	if (printerRecNum > printerNumRecs)
	    return 0;

	d4go (printerData, printerRecNum);
	d4unlock_all (printerData);
	strcpy (printerEntry->name, f4str (printerGroupNameFld));
	stripBlanks (printerEntry->name);
	for (i=0 ; i<MAX_PRINTERS_PER_GROUP ; i++) {
	    stringCopy (printerEntry->printers [i], f4str (printerPrinterFld [i]),
				sizeof (PrinterGroupEntry));
	    stripBlanks (printerEntry->printers [i]);
	}
	printerRecNum++;
	return 1;
		
    }

/*----------*/
void writePrinterGroup (printerEntry)

    PrinterGroupEntry	*printerEntry;

    {

	int	i;

	d4append_start (printerData, 0);
	f4assign (printerGroupNameFld, printerEntry->name);
	for (i=0 ; i<MAX_PRINTERS_PER_GROUP ; i++)
	    f4assign (printerPrinterFld [i], printerEntry->printers [i]);
	d4append (printerData);
	d4unlock_all (printerData);

    }

/*----------*/
int getNumPrinters (VOID)	/* Get number of printers in printers database */

    {

	return d4reccount (printerData);

    }

/*----------*/
void updateDBaseFile (data)

    D4DATA	*data;

    {

	d4flush_all (data);
#ifndef CB51
	data->file_lock = 1;		/* to force d4update_header() to write the num_recs field */
	d4update_header (data, 1, 1);
	data->file_lock = 0;	/* to force d4unlock_all() to write the num_recs field */
#endif
	d4unlock_all (data);

    }

/*----------*/
void stripBlanks (buf)	/* Strip blanks from a string */

    char	buf [];

    {

	int	last;

	last = strlen (buf) - 1;
	while (last >= 0) {
	    if (buf [last] != ' ')
		break;
	    last--;
	}
	buf [last+1] = '\0';	/* enpty string */
				
    }

/*----------*/
static void stringCopy (dest, src, n)

    char	*dest, *src;
    int		n;

	/* String copy with length protection and NULL termination. */

    {

	strncpy (dest, src, n - 1);
	*(dest + n - 1) = '\0';

    }

