#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vlistitemHEADER
#include vbuttonHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "alarmDefs.h"	// .../share/src/alarm
}

#include "alCfgTreeListClass.hxx"
#include "alarmConfig.hxx"
#include "alCfgMainDialogClass.hxx"
#include "alCfgDatabaseClass.hxx"


/*******************************************************************
*
*    alarmTreeClass
*
*******************************************************************/

/*******************************************************************
*
*    alarmTreeClass constructor
*	build the alarm group tree
*
*******************************************************************/
alarmTreeClass::alarmTreeClass (alCfgApplication *app)
{

    alCfgApp = app;
    treeRoot = NULL;
    clipBoard = NULL;
    db = alCfgApp -> getDatabaseClass ();
    loadTree ();
    treeModified = vFALSE;
    infoModified = vFALSE;
    vmemSet ((void *)&this -> blankInfo, 0, sizeof (struct GroupInfo));
}


/*******************************************************************
*
*    alarmTreeClass destructor
*
*******************************************************************/
alarmTreeClass::~alarmTreeClass (void)
{
    if (treeRoot != NULL)
    {
	if ((treeModified == vTRUE) || (infoModified == vTRUE))
	{
	    if ((db -> writeTreeList ()) == ALCFG_SUCCESS)
	    {
		writeTreeTraverse (treeRoot);
	    }
	    if ((db -> createInfoTable ()) == ALCFG_SUCCESS)
	    {
		writeInfoTraverse (treeRoot);
	    }
	}

	deleteFwdlink (treeRoot);

	if (clipBoard != NULL)
	{
	    deleteFwdlink (clipBoard);
	}
    }
}


/*******************************************************************
*
*    alarmTreeClass add node to tree
*
*******************************************************************/
void alarmTreeClass::addChild (TREENODE_PTR parent, char *name)
{
    TREENODE_PTR child = (TREENODE_PTR)(vmemAllocAndClear (sizeof (TREENODE)));
    child -> parent = parent;
    strncpy (child -> groupName, name, AL_GROUPNAME_SIZE);
    child -> groupName[AL_GROUPNAME_SIZE] = 0;

    if (parent -> sublink == NULL)
    {
	if ((isRoot (parent)) == vFALSE)
	{
	    parent -> boxType = ALCFG_GROUP_BOX; 
	    if (parent -> fwdlink != NULL)
	    {
		parent -> imageType = ALCFG_PARENT;
	    }
	    else
	    {
		parent -> imageType = ALCFG_LAST_PARENT;
	    }
	}
	parent -> sublink = child;
    }
    else
    {
	TREENODE_PTR nextChild = parent -> sublink -> fwdlink;
	TREENODE_PTR lastChild = parent -> sublink;
	while (nextChild)
	{
	    lastChild = nextChild;
	    nextChild = nextChild -> fwdlink;
	}
	lastChild -> imageType = ALCFG_CHILD;
	lastChild -> fwdlink = child;
	if ((isRoot (lastChild)) == vFALSE)
	{
	    if (lastChild -> sublink != NULL)
	    {
		lastChild -> imageType = ALCFG_PARENT;
	    }
	}
    }
    child -> level = parent -> level + 1;
    child -> boxType = ALCFG_NO_GROUP_BOX;
    child -> imageType = ALCFG_LAST_CHILD;
    db -> getInfoData (child);
    treeModified = vTRUE;
}


/*******************************************************************
*
*    alarmTreeClass cut a node from tree to clipBoard
*
*******************************************************************/
void alarmTreeClass::cutSubTree (TREENODE_PTR node)
{
    TREENODE_PTR prevNode = NULL;
    TREENODE_PTR currNode = node -> parent -> sublink;

    while (currNode)
    {
	if (currNode == node)
	{
	    break;
	}
	prevNode = currNode;
	currNode = currNode -> fwdlink;
    }

    clipBoard = currNode;

    // node is parents only child or node is parents 1st child, 
    // change parents box and image type
    if (prevNode == NULL)
    {
	node -> parent -> sublink = node -> fwdlink;
	// parents only child, change parents meaning
	// if it was LAST PARENT then it becomes the LAST CHILD
	// else it was a PARENT and becomes a CHILD
	if (node -> fwdlink == NULL)
	{
	    if (node -> parent -> imageType == ALCFG_LAST_PARENT)
	    {
		node -> parent -> imageType = ALCFG_LAST_CHILD;
	    }
	    else 
	    {
		node -> parent -> imageType = ALCFG_CHILD;
	    }
	    node -> parent -> boxType = ALCFG_NO_GROUP_BOX;
	}
    }
    // parent has other children
    else
    {
	prevNode -> fwdlink = node -> fwdlink;
	if ( (node -> imageType == ALCFG_LAST_PARENT) ||
	     (node -> imageType == ALCFG_LAST_CHILD) )
	{
	    if (prevNode -> imageType == ALCFG_PARENT)
	    {
		prevNode -> imageType = ALCFG_LAST_PARENT;
	    }
	    else
	    {
		prevNode -> imageType = ALCFG_LAST_CHILD;
	    }
	}
    }
    node -> parent = NULL;
    node -> fwdlink = NULL;
    treeModified = vTRUE;
}


/*******************************************************************
*
*    alarmTreeClass delete a node
*
*******************************************************************/
void alarmTreeClass::deleteFwdlink (TREENODE_PTR node)
{
    TREENODE_PTR currNode = node;
    TREENODE_PTR prevNode;

    while (currNode)
    {
	prevNode = currNode;
	if (currNode -> sublink)
	{
	    deleteSublink (currNode -> sublink);
	}
	currNode = currNode -> fwdlink;
	if (prevNode -> parameters != NULL)
	{
	    vmemFree ((void *)prevNode -> parameters);
	    prevNode -> parameters = NULL;
	}
	vmemFree ((void *)prevNode);
	treeModified = vTRUE;
    }
}


/*******************************************************************
*
*    alarmTreeClass delete a node
*
*******************************************************************/
void alarmTreeClass::deleteSublink (TREENODE_PTR node)
{
    TREENODE_PTR currNode = node;
    TREENODE_PTR prevNode;

    while (currNode)
    {
	prevNode = currNode;
	if (currNode -> fwdlink)
	{
	    deleteFwdlink (currNode  -> fwdlink);
	}
	currNode = currNode -> sublink;
	if (prevNode -> parameters != NULL)
	{
	    vmemFree ((void *)prevNode -> parameters);
	    prevNode -> parameters = NULL;
	}
	vmemFree ((void *)prevNode);
    }
}


/*******************************************************************
*
*    alarmTreeClass delete a node subtree
*	traverse the tree of siblings (fwdlinks) and delete the
*	specified subtree
*
*******************************************************************/
void alarmTreeClass::deleteSubTree (TREENODE_PTR node)
{
    TREENODE_PTR prevNode = NULL;
    TREENODE_PTR currNode = node -> parent -> sublink;

    while (currNode)
    {
	if (currNode == node)
	{
	    break;
	}
	prevNode = currNode;
	currNode = currNode -> fwdlink;
    }
    // node is parents only child or node is parents 1st child, 
    // change parents box and image type
    if (prevNode == NULL)
    {
	node -> parent -> sublink = node -> fwdlink;
	// parents only child, change parents meaning
	// if it was LAST PARENT then it becomes the LAST CHILD
	// else it was a PARENT and becomes a CHILD
	if (node -> fwdlink == NULL)
	{
	    if (node -> parent -> imageType == ALCFG_LAST_PARENT)
	    {
		node -> parent -> imageType = ALCFG_LAST_CHILD;
	    }
	    else 
	    {
		node -> parent -> imageType = ALCFG_CHILD;
	    }
	    node -> parent -> boxType = ALCFG_NO_GROUP_BOX;
	}
    }
    // parent has other children
    else
    {
	prevNode -> fwdlink = node -> fwdlink;
	if ( (node -> imageType == ALCFG_LAST_PARENT) ||
	     (node -> imageType == ALCFG_LAST_CHILD) )
	{
	    if (prevNode -> imageType == ALCFG_PARENT)
	    {
		prevNode -> imageType = ALCFG_LAST_PARENT;
	    }
	    else
	    {
		prevNode -> imageType = ALCFG_LAST_CHILD;
	    }
	}
    }
    // delete all nodes in the subtree
    deleteSubTreeNodesTrav (node -> sublink);
    if (node -> parameters != NULL)
    {
	vmemFree ((void *)node -> parameters);
    }
    vmemFree ((void *)node);
    treeModified = vTRUE;
}


/*******************************************************************
*
*    alarmTreeClass delete a node subtree
*	traverse all descendents and delete them
*
*******************************************************************/
void alarmTreeClass::deleteSubTreeNodesTrav (TREENODE_PTR node)
{
    if (node)
    {
	deleteSubTreeNodesTrav (node -> sublink);
	deleteSubTreeNodesTrav (node -> fwdlink);
	if (node -> parameters != NULL)
	{
	    vmemFree ((void *)node -> parameters);
	}
	vmemFree ((void *)node);
    }
}


/*******************************************************************
*
*    alarmTreeClass find a node
*
*******************************************************************/
TREENODE_PTR alarmTreeClass::findNode (TREENODE_PTR node, char *name)
{
    TREENODE_PTR p;
    TREENODE_PTR res = NULL;

    for (p = node; p != NULL; p = p -> fwdlink)
    {
	if ((strncmp (name, p -> groupName, AL_GROUPNAME_SIZE))== 0)
	{
	    res = p;
	    break;
	}
	if (p -> sublink)
	{
	    res = findInSublink (p-> sublink, name);
	    if (res != NULL)
	    {
		break;
	    }
	}
    }

    return (res);
}


/*******************************************************************
*
*    alarmTreeClass find a node in a sublink
*
*******************************************************************/
TREENODE_PTR alarmTreeClass::findInSublink (TREENODE_PTR node, char *name)
{
    TREENODE_PTR p;
    TREENODE_PTR res = NULL;

    for (p = node; p != NULL; p = p -> sublink)
    {
	if ((strncmp (name, p -> groupName, AL_GROUPNAME_SIZE)) == 0)
	{
	    res = p;
	    break;
	}
	if (p -> fwdlink)
	{
	    res = findNode (p -> fwdlink, name);
	    if (res != NULL)
	    {
		break;
	    }
	}
    }

    return (res);
}


/*******************************************************************
*
*    alarmTreeClass 
*	build the alarm group tree
*
*******************************************************************/
void alarmTreeClass::loadTree (void)
{
    char	parent [AL_GROUPNAME_SIZE + 1];
    char	child [AL_GROUPNAME_SIZE + 1];

    int			dataType;
    TREENODE_PTR	node;

    DB_RECORD_PTR rec = DB_Top (db -> alCfgTreeTable);

    while (rec)
    {
	DB_GetFieldValue (rec, db -> parentFld, parent, &dataType);
	DB_GetFieldValue (rec, db -> childFld, child, &dataType);
	// Must be main, since it has no parent, immaculate conception !
	if ( (strlen (parent)) == 0)
	{
	    treeRoot = (TREENODE_PTR)(vmemAllocAndClear (sizeof (TREENODE)));
	    strncpy (treeRoot -> groupName, child, AL_GROUPNAME_SIZE);
	    treeRoot -> groupName[AL_GROUPNAME_SIZE] = 0;
	    treeRoot -> level = 0;
	    treeRoot -> boxType = ALCFG_GROUP_BOX;
	    treeRoot -> imageType = ALCFG_MAIN_PARENT;
	    db -> getInfoData (treeRoot);
	}
	else
	{
	    node = findNode (treeRoot, parent);
	    addChild (node, child);
	}
	rec = (DB_RECORD_PTR)DB_Next (db -> alCfgTreeTable);
    }
    treeModified = vFALSE;
}


/*******************************************************************
*
*    alarmTreeClass paste a node from clipboard into the tree
*	get inserted as nodes 1st child
*
*******************************************************************/
void alarmTreeClass::pasteSubTree (TREENODE_PTR node)
{
    clipBoard -> parent = node;
    clipBoard -> fwdlink = node -> sublink;
    node -> sublink = clipBoard;

    if (node -> imageType == ALCFG_CHILD)
    {
	node -> imageType = ALCFG_PARENT;
    }
    else if (node -> imageType == ALCFG_LAST_CHILD)
    {
	node -> imageType = ALCFG_LAST_PARENT;
    }
    node -> boxType = ALCFG_GROUP_BOX;
    // if has no siblings
    if (clipBoard -> fwdlink == NULL)
    {
	// if it has no children of its own
	if (clipBoard -> sublink == NULL)
	{
	    clipBoard -> imageType = ALCFG_LAST_CHILD;
	}
	// becomes node's last parent
	else
	{
	    clipBoard -> imageType = ALCFG_LAST_PARENT;
	}
    }
    // has siblings
    else
    {
	// if it has no children of its own
	if (clipBoard -> sublink == NULL)
	{
	    clipBoard -> imageType = ALCFG_CHILD;
	}
	// it has its own children, becomes parent
	else
	{
	    clipBoard -> imageType = ALCFG_PARENT;
	}
    }
    updateLevelsTraverse (clipBoard);
    clipBoard = NULL;
    treeModified = vTRUE;
}


/*******************************************************************
*
*    alarmTreeClass update the levels starting from this node
*
*******************************************************************/
void alarmTreeClass::updateLevelsTraverse (TREENODE_PTR node)
{

    if (node != NULL)
    {
	if (node -> parent)
	{
	    node -> level = node -> parent -> level + 1;
	}
	// update children
	updateLevelsTraverse (node -> sublink);
	// update  siblings
	updateLevelsTraverse (node -> fwdlink);
    }
}


/*******************************************************************
*
*    alarmTreeClass write info params
*
*******************************************************************/
void alarmTreeClass::writeInfoTraverse (TREENODE_PTR node)
{

    if (node != NULL)
    {
	if (node -> parameters != NULL)
	{
	    // do not write blank records
	    if ((vmemCompare ((void *)node -> parameters, 
			      (void *)&this -> blankInfo,
			      sizeof (struct GroupInfo))) != 0)
	    {
		db -> writeInfoRecord (node);
	    }
	}
	// write children
	writeInfoTraverse (node -> sublink);
	// write siblings
	writeInfoTraverse (node -> fwdlink);
    }
}


/*******************************************************************
*
*    alarmTreeClass write the tree
*
*******************************************************************/
void alarmTreeClass::writeTreeTraverse (TREENODE_PTR node)
{

    if (node != NULL)
    {
	if (node -> parent)
	{
	    DB_RECORD_PTR rec = DB_NewRecord (db -> alCfgTreeTable, NULL);
	    DB_PutFieldNF (rec, db -> parentFld, node -> parent -> groupName);
	    DB_PutFieldNF (rec, db -> childFld, node -> groupName);
	}
	// write children
	writeTreeTraverse (node -> sublink);
	// write siblings
	writeTreeTraverse (node -> fwdlink);
    }
}

