#ifndef alcfgtreelistclassHXX
#define	alcfgtreelistclassHXX

//** DEFINE ALARM TREE CLASS ITEMS **************************************

typedef enum
{
    ALCFG_NO_GROUP_BOX = 0,
    ALCFG_GROUP_BOX
} ALCFG_BOX_TYPE;

// define imageType
typedef enum
{
    ALCFG_MAIN_PARENT = 0,
    ALCFG_PARENT,
    ALCFG_LAST_PARENT,
    ALCFG_CHILD,
    ALCFG_LAST_CHILD
} ALCFG_IMAGE_TYPE;

typedef struct GroupInfo
{
    char	autoScript [AL_SCRIPTNAME_SIZE + 1];
    char	auxAutoScript [AL_SCRIPTNAME_SIZE + 1];
    char	manualScript [AL_SCRIPTNAME_SIZE + 1];
    char	auxManualScript [AL_SCRIPTNAME_SIZE + 1];
    char	displayName [AL_SCRIPTNAME_SIZE + 1];
    char	helpTopic [AL_HELPTOPIC_SIZE + 1];
    char	printers [AL_NUM_SEVR][AL_PATHNAME_SIZE + 1];
    char	auxPrinters [AL_NUM_SEVR][AL_PATHNAME_SIZE + 1];
    char	sounds [AL_NUM_SEVR][AL_PATHNAME_SIZE + 1];
    char	auxSounds [AL_NUM_SEVR][AL_PATHNAME_SIZE + 1];
    char	rates [AL_NUM_SEVR][AL_PATHNAME_SIZE + 1];
    char	auxRates [AL_NUM_SEVR][AL_PATHNAME_SIZE + 1];
    char	auxSelectPV [AL_POINTNAME_SIZE + 1];
    char	guidance [AL_GUIDANCE_SIZE + 1];
} GROUPINFO, *GROUPINFO_PTR;

typedef struct TreeNode
{
    char		groupName[AL_GROUPNAME_SIZE + 1];
    TreeNode		*fwdlink;	// links to other nodes at the same level, siblings
    TreeNode		*sublink;	// link to next level, children
    TreeNode		*parent;	// MAIN has no parent
    int			numChildren;
    int		  	level;
    ALCFG_BOX_TYPE	boxType;
    ALCFG_IMAGE_TYPE	imageType;
    GROUPINFO_PTR	parameters;
}TREENODE, *TREENODE_PTR;


/*
* NODE LAYOUT
*    Level
*	1	MAIN -> 
*                |
*	2	FRED -> LUCY          -> RICKY -> ETHEL
*                        |
*	3		LUCYA -> LUCYB
*                        |         |
*	4		LUCYA1	 LUCYB1 -> LUCYB2 -> LUCYB3
*/
class alCfgApplication;
class alCfgDatabaseClass;

//** THE ALARM TREE CLASS ************************************************
class alarmTreeClass
{
    alCfgApplication	*alCfgApp;
    alCfgDatabaseClass	*db;
    TREENODE_PTR	treeRoot;
    TREENODE_PTR	clipBoard;
    int			treeModified;
    int			infoModified;
    GROUPINFO		blankInfo;

public:

    // class constructor
    alarmTreeClass (alCfgApplication *app);
    // class destuctor
    ~alarmTreeClass (void);

    // return root of tree
    TREENODE_PTR getTreeRoot (void)
	{ return (treeRoot); }

    // is node passed in the tree root
    int isRoot (TREENODE_PTR p)
	{ return ((p == treeRoot) ? vTRUE : vFALSE); }

    // functs for clipBoard
    TREENODE_PTR getClipBoard (void)
	{ return (clipBoard); }
    int clipBoardActive (void)
	{ return ((clipBoard != NULL) ? vTRUE : vFALSE); }

    // handle infoModfied flag
    void setInfoModified (int flag)
	{ infoModified = flag; }

    void addChild (TREENODE_PTR node, char *child);
    void cutSubTree (TREENODE_PTR node);
    void deleteFwdlink (TREENODE_PTR node);
    void deleteSublink (TREENODE_PTR node);
    void deleteSubTree (TREENODE_PTR node);
    void deleteSubTreeNodesTrav (TREENODE_PTR node);
    TREENODE_PTR findNode (TREENODE_PTR node, char *name);
    TREENODE_PTR findInSublink (TREENODE_PTR node, char *name);
    void loadTree (void);
    void pasteSubTree (TREENODE_PTR node);
    void updateLevelsTraverse (TREENODE_PTR node);
    void writeInfoTraverse (TREENODE_PTR node);
    void writeTreeTraverse (TREENODE_PTR node);
};


#endif	alcfgtreelistclassHXX
