#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vlistitemHEADER
#include vlistHEADER
#include vbuttonHEADER

extern "C"
{
#include "alarmDefs.h"
}

#include "alcfg.h"
#include "alCfgTreeListClass.hxx"
#include "alarmConfig.hxx"
#include "alCfgMainDialogClass.hxx"
#include "alCfgUtils.hxx"


enum
{
    ALCFG_NO_MSG = 0,
    ALCFG_INFO_MSG,
    ALCFG_CAUTION_MSG
};


/*******************************************************************
*
*    Alarm Config sort printerList compare function
*
*******************************************************************/

int alCfgComparePrinters (const void *value1, const void *value2)
{
    PRINTER_ENTRY_PTR entry1 = (PRINTER_ENTRY_PTR)value1;
    PRINTER_ENTRY_PTR entry2 = (PRINTER_ENTRY_PTR)value2;

    return (vcharCompare (entry1 -> printerGroupName,
			  entry2 -> printerGroupName));
}


/*******************************************************************
*
*    alCfgMessageDialogCenter ()
*	passed a message number as defined in alCfgUtils.hxx
*
*******************************************************************/

int alCfgMessageDialogCenter (vdialog *dialog, int msgNumber, ...)
{
    int		result = ALCFG_SUCCESS;
    
    vapplication *alcfgApp = vapplication::GetCurrent ();
    vresource    alcfgRes  = alcfgApp -> GetResources ();
    vresource    msgRes     = vresourceGet (alcfgRes,
					alcfg_alCfgMessagesDict);
    vresource	 titlesRes = vresourceGet (alcfgRes,
					alcfg_alCfgTitlesDict);
    vstr         *infoTitle = vresourceGetString (titlesRes,
					alcfg_alCfgInfoNoticeTitle);
    vstr         *cautionTitle = vresourceGetString (titlesRes,
					alcfg_alCfgCautionNoticeTitle);
    vstr         *msg = NULL;
    vstr	 *formattedMsg = NULL;
    int          msgType = ALCFG_NO_MSG;
    int		 freeMsg = vTRUE;
    vchar	 *vPtr1, *vPtr2;
    va_list	 argptr;

    va_start (argptr, msgNumber);

    switch (msgNumber)
    {
	case ALCFG_QUIT_MSG:
	    msg = vresourceGetString (msgRes, alcfg_alCfgQuitMsg);
	    msgType = ALCFG_CAUTION_MSG;
	    break;

	case ALCFG_TREEFILE_ERR:
	    msg = vresourceGetString (msgRes, alcfg_alCfgTreeFileErr);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_INFOFILE_ERR:
	    msg = vresourceGetString (msgRes, alcfg_alCfgInfoFileErr);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_PRINTERFILE_ERR:
	    msg = vresourceGetString (msgRes, alcfg_alCfgPrinterFileErr);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_GROUPFILE_ERR:
	    msg = vresourceGetString (msgRes, alcfg_alCfgGroupFileErr);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_GROUP_EXISTS_ERR:
	    formattedMsg = vresourceGetString (msgRes, alcfg_alCfgGroupExistsErr);
	    vPtr1 = va_arg (argptr, vchar *);
	    msg = vstrCloneScribed (vcharScribeFormat (formattedMsg, vPtr1));
	    vstrDestroy (formattedMsg);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_DELETE_GRP_REQ:
	    formattedMsg = vresourceGetString (msgRes, alcfg_alCfgDeleteGrpReq);
	    vPtr1 = va_arg (argptr, vchar *);
	    msg = vstrCloneScribed (vcharScribeFormat (formattedMsg, vPtr1));
	    vstrDestroy (formattedMsg);
	    msgType = ALCFG_CAUTION_MSG;
	    break;

	case ALCFG_GRP_INTREE_ERR:
	    formattedMsg = vresourceGetString (msgRes, alcfg_alCfgGrpInTreeErr);
	    vPtr1 = va_arg (argptr, vchar *);
	    msg = vstrCloneScribed (vcharScribeFormat (formattedMsg, vPtr1));
	    vstrDestroy (formattedMsg);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_NO_DELETE_MAIN_ERR:
	    msg = vresourceGetString (msgRes, alcfg_alCfgNoDeleteMainErr);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_NOTHING_SELECTED_ERR:
	    msg = vresourceGetString (msgRes, alcfg_alCfgNothingSelectedErr);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_NO_NODE_SEL_ERR:
	    msg = vresourceGetString (msgRes, alcfg_alCfgNoNodeSelErr);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_INSERT_GRP_TREE_REQ:
	    formattedMsg = vresourceGetString (msgRes, alcfg_alCfgInsertGrpTreeReq);
	    vPtr1 = va_arg (argptr, vchar *);
	    vPtr2 = va_arg (argptr, vchar *);
	    msg = vstrCloneScribed (vcharScribeFormat (formattedMsg, vPtr1, vPtr2));
	    vstrDestroy (formattedMsg);
	    msgType = ALCFG_CAUTION_MSG;
	    break;

	case ALCFG_PRINTER_EXISTS_ERR:
	    formattedMsg = vresourceGetString (msgRes, alcfg_alCfgPrinterExistsErr);
	    vPtr1 = va_arg (argptr, vchar *);
	    msg = vstrCloneScribed (vcharScribeFormat (formattedMsg, vPtr1));
	    vstrDestroy (formattedMsg);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_DATAMOD_MSG:
	    msg = vresourceGetString (msgRes, alcfg_alCfgDataModMsg);
	    msgType = ALCFG_CAUTION_MSG;
	    break;

	case ALCFG_DELETE_PRT_REQ:
	    formattedMsg = vresourceGetString (msgRes, alcfg_alCfgDeletePrtReq);
	    vPtr1 = va_arg (argptr, vchar *);
	    msg = vstrCloneScribed (vcharScribeFormat (formattedMsg, vPtr1));
	    vstrDestroy (formattedMsg);
	    msgType = ALCFG_CAUTION_MSG;
	    break;

	case ALCFG_CLIPBOARD_ACTIVE_MSG:
	    formattedMsg = vresourceGetString (msgRes, alcfg_alCfgClipActMsg);
	    vPtr1 = va_arg (argptr, vchar *);
	    msg = vstrCloneScribed (vcharScribeFormat (formattedMsg, vPtr1));
	    vstrDestroy (formattedMsg);
	    msgType = ALCFG_CAUTION_MSG;
	    break;

	case ALCFG_NOCUT_MAIN_ERR:
	    msg = vresourceGetString (msgRes, alcfg_alCfgNoCutMainErr);
	    msgType = ALCFG_INFO_MSG;
	    break;

	case ALCFG_DELETE_PARAMS_REQ:
	    formattedMsg = vresourceGetString (msgRes, alcfg_alCfgDeleteParamsReq);
	    vPtr1 = va_arg (argptr, vchar *);
	    msg = vstrCloneScribed (vcharScribeFormat (formattedMsg, vPtr1));
	    vstrDestroy (formattedMsg);
	    msgType = ALCFG_CAUTION_MSG;
	    break;

	case ALCFG_CLIP_QUIT_MSG:
	    formattedMsg = vresourceGetString (msgRes, alcfg_alCfgClipActQuitMsg);
	    vPtr1 = va_arg (argptr, vchar *);
	    msg = vstrCloneScribed (vcharScribeFormat (formattedMsg, vPtr1));
	    vstrDestroy (formattedMsg);
	    msgType = ALCFG_CAUTION_MSG;
	    break;

	case ALCFG_INIT_OPERPRIV_ERR:
	    msg = vresourceGetString (msgRes, alcfg_alCfgInitOperPrivErr);
	    msgType = ALCFG_INFO_MSG;
	    break;

	default:
	    break;
    }

    va_end (argptr);

    switch (msgType)
    {
	case ALCFG_NO_MSG:
	    break;

	case ALCFG_INFO_MSG:
	    infoNotice (dialog, infoTitle, msg);
	    break;

	case ALCFG_CAUTION_MSG:
	    result = cautionNotice (dialog, cautionTitle, msg);
	    if (result == vTRUE)
		result = ALCFG_TRUE;
	    else
		result = ALCFG_FALSE;
	    break;

    }

    vstrDestroy (infoTitle);
    vstrDestroy (cautionTitle);
    if ( (msg != NULL) && (freeMsg == vTRUE) )
	vstrDestroy (msg);

    return (result);
}

