/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/* This message is used for simple messages that have no body */

#define	UPDATE			1
#define	UPDATE_NEW_ALARM	2
#define	ALARM_MSG		3
#define	ACK_MSG			4
#define	ACK_GROUP_MSG		5
#define	ENABLE_MSG		6
#define	DISABLE_MSG		7
#define	ALARM_COMMENT_UPDATE	8
#define ALARM_SILENCE		9
#define	AVAILABLE		10
#define	UNAVAILABLE		11
#define	PING			99

typedef struct {
	uint32		msgType;
	char		pad [4];	
} TypeOnlyMessage;

/* This message travels from AlarmVu to the Alarm Manager.  It also travels between Alarm Managers. */

typedef struct {
	uint32			msgType;
	uint32			memberOfGroup;	/* 1 => this ACK resulted from a group ACK */
	uint32			status;
	uint32			severity;
	TS_STAMP		timeStamp;
	TS_STAMP		ackTime;	/* Time ack occurred */
	PointName		pointName;
	union {
	    Operator		operatorName;	/* If memberOfGroup == 0 */
	    GroupName		groupName;	/* If memberOfGroup == 1 */
	} name;
	PvValue			value;		/* Value at time of alarm */
	char			pad [5];
} AlarmMsg;

/* This message is sent from alarmVu to the Alarm Manager to have a comment attached to */
/* an alarm or ACK. */

typedef struct  {
	uint32		msgType;
	TS_STAMP	timeStamp;
	PointName	pointName;	/* Could be a group name as well */
	char		memoText[AL_MAX_MEMO_SIZE+1]; /* if 1st char is zero, delete memo field */
	char		pad [2];
} AlarmMemo;

union allmsgs {
	TypeOnlyMessage	typemsg;
	AlarmMsg	alarmmsg;
	AlarmMemo	memomsg;
};
#define MAX_MSG_SIZE	(sizeof(union allmsgs) + 1)

/****************************************************************************************/
/* 			Messages to/from the alarmMgr to printer servers                   */
/****************************************************************************************/

#define	PRINTER_NAME_MSG	1

typedef struct {
	long		msgType;
	PathName	printerName;
	char		pad [3];
} PrinterNameMsg;
