/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* Alarm Manager */

#include <sys/types.h>
#ifdef WIN32
#include <winsock.h>
#include <process.h>
#include <sys/stat.h>
#include "NTUnixSubs.h"
#else
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#ifndef AXP
#include <net/if.h>
#include <sys/sockio.h>
#include <sys/filio.h>
#endif
#include <netdb.h>
#include <sys/ioctl.h>
#include <signal.h>
#endif WIN32
#include <time.h>
#include <malloc.h>
#include <errno.h>
#include <string.h>
#include "msgMon.h"

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "alarmDefs.h"
#include "shareMem.h"
#include "groupInfo.h"

#include "cadef.h"
#include "dbDefs.h"
#include "db_access.h"
#include "paramStructs.h"
#include "tsDefs.h"
#include "alarm.h"
#include "tisAlarm.h"
#include "alarmString.h"
#include "dbase.h"
#include "alarmq.h"

#include "globals.h"
#include "sharedMem.h"
#include "tree.h"
#include "points.h"
#include "ChannelList.h"
#include "messages.h"

#include "printerGroupList.h"
#include "alarmProcess.h"
#include "archive.h"
#include "opTrkArchive.h"

/* Functions defined in this module */

#define	VOID	void
static	void	reconnectCallback (void *, int, int);
static	void	dataCallback (tisAlarm *);
static	void	connectionCallback (char *, short);
static	int	setupListenForVu (void);
static	void	checkForAlarmVuConnect (void);
static	void	processPendingVuMessages (void);
static	int	getVuMsgType (int);
static	int	getVuBytes (int, char *, int);
static	void	flushVuBuffer (int);
static	void	closeAlarmVuFds (void);
	void	sendSimpleMessageToAlarmvu (int);
	void	oustAlarmVus (int);
	void	allowAlarmVus (int);
static	void	buildPeerList (void);
static	void	connectToPeerAlarmMgrs (void);
static	int	setupListenForMgr (void);
static	void	checkForAlarmMgrConnect (void);
static	void	processPendingMgrMessages (void);
static	int	getMgrMsgType (int);
static	int	getMgrBytes (int, char *, int);
static	void	checkForMgrTimeout (void);
static	void	closeAlarmMgrFds (void);
static	void	sendPingToPeers (void);
	void	sendMessageToPeers (char *, int);
static	int	setupListenForPrinters (void);
static	void	checkForPrinterConnect (void);
	void	addPrinter (char *);
	void	addOnePrinter (char *);
static	void	startPrinterServers (void);
static	void	processPendingPrinterMessages (void);
static	int	getPrinterMsgType (int);
static	int	getPrinterBytes (int, char *, int);
static	void	closePrinterFds (void);
	int	sendMessageToPrinter (char *, char *);
static	void	sendPrinterTestMessages (void);
static	int	sendMsgToPrinterServer (char *, char *);
static	int	sendMsg (int, char *, int);
static	void	readAllFds (void);
static	void	reconfigure (void);
	void	stringCopy (char *, char *, int);
	void	logBytes (char *, int);

/* External functions */

extern	void	setupSignalHandlers ();	
extern	void	serverInit ();
extern	void	startArchiving ();
extern	void	archiveAndClean ();
extern	void	processPendingAlarms ();
extern	int	processAlarm ();
extern	void	processAck ();
extern	void	processAckGroup  ();
extern	void	processEnable ();
extern	void	processDisable ();
extern	void	processMemo ();
extern	void	launchOpTrkArchiver ();

extern	void	dumpTree ();
extern	void	dumpSummaryList ();

/* Data */

static	int			currentHour, currentMin;
static	long			lastClockTime;
static	long			lastPingTime;
static	long			ping_timeout = PING_TIMEOUT;
	int			running = 0;
static	int			doConnect;
static	int			caInitialized;
	AlarmQueue		alarmQueue;
static	int			listeningForAlarmVu = 0, listeningForAlarmMgrs = 0, listeningForPrinters = 0;
static	int			vuListenFd, mgrListenFd, printListenFd;
static	unsigned		printListenPort;
static	short			vuPort = VUPORT;
static	short			mgrPort = MGRPORT;
static	struct sockaddr		saddr;
static	int			saddrLen;
static	int			optionValue = 1;
static	VuConn			vuConn [MAX_ALARMVUS];
static	MgrConn			mgrConn [MAX_PEERS];
static	int			numPeers;
static	PrintConn		printConn [MAX_PRINTERS];
static	PrinterList		printerList [MAX_PRINTERS];
static	int			numPrinters = 0;
static	int			caFd;
static	int			displayPings = 0;
	int			myPid;
	int			alarmsProcessed = 0;
static	unsigned long		myIpAddr;
static	int			printerTest = 0;

extern	char	severityChars [];
extern	long	treeFileTime;
extern	char	*sharedBase;
#ifndef SOLARIS
extern	char	*sys_errlist [];
#endif

/*----------*/
int main (int argc, char **argv)
{

	int	c, i;
	extern	char	*optarg;
	struct tm	*t;
	time_t		clock, lastTime;

#ifdef WIN32
	myPid = _getpid ();
#else
	myPid = getpid ();
#endif WIN32
	tisOpenLog ("alarmMgr", 6, 2, NULL, NULL);

	/* Process command line options */

	while ((c = getopt (argc, argv, "l:dvsptP:")) != -1)
	    switch (c) {
		case 'l':	/* Explicitly set debug level */
		    debugLevel = atoi (optarg);
		    break;
		case 'P':	/* Explicitly set ping_timeout (and ping rate is 1/3 of timeout) */
		    ping_timeout = atoi (optarg);
		    break;
		case 'd':	/* Debugging on */
		    debugLevel = 3;
		    break;
		case 'v':	/* Verbose mode on */
		    debugLevel = 2;
		    break;
		case 's':	/* Status mode */
		    debugLevel = 1;
		    break;
		case 'p':	/* Display ping messages */
		    displayPings = 1;
		    break;
		case 't':	/* Perform printer test */
		    printerTest = 1;
		    break;
	    }

	serverInit ("alarmMgr", "/tis4000/tasks/alarmMgr.lock", "lock");

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Started.\n", myPid);

	/* Set up signal handlers */

	setupSignalHandlers ();

	/* Allocate shared memory segments */

	if (!allocateSharedMemory ())
	    tisSysLog (0, 2, "alarmMgr(%d): Unable to allocate shared memory.\n", myPid);

	/* Initialize channel access */

	if (ca_task_initialize () != ECA_NORMAL) {
	    tisSysLog (0, 2, "alarmMgr(%d): CA won't initialize\n", myPid);
	    exit (-1);
	}
	caInitialized = 1;

	/* Initialize the database system */

	initializeDatabase ();

	/* Initialize all connections to "not connected" */

	for (i=0; i < MAX_ALARMVUS; i++)
	    vuConn[i].fd = -1;

	for (i=0; i < MAX_ALARMMGRS; i++)
	    mgrConn [i].fd = -1;

	for (i=0 ; i<MAX_PRINTERS ; i++) {
	    printConn [i].fd = -1;
	    printConn [i].nameIx = -1;
	}

	/* Build list of peer alarm managers */

	buildPeerList ();

	/* Open the alarm log, pvmap, and reslist files */

	openPvMapFile ();
	openPdbXrefFile ();
	openLogFile ();
	openResListFile ();

	/* Read the printer group list */

	printerGroupListInit ();

	/* Open or create the point state file */

	if (!openPointStateFile ())
	    createPointStateFile ();

	/* Initialize some things. */

	allocateStringSpace (sharedBase + sizeof (ShareHeader), AL_STRINGSIZE);
	initializeTree ();
	initializePoints ();

	/* Initialize channel list */

	clInitialize ();

	/* Load the alarm tree and database points */

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Loading alarm tree\n", myPid);

	if (!loadTree ()) {
	    tisSysLog (0, 2, "alarmMgr(%d): Alarm group tree has not been configured.\n", myPid);
	    clTerminate ();
	    ca_task_exit ();
	    freeSharedMemory ();
	    closePvMapFile ();
	    closePdbXrefFile ();
	    closeLogFile ();
	    exit (-1);
	}

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Loading database points\n", myPid);

	loadPoints ();

	/* Initialize the alarm queue */

	aqInit (&alarmQueue, 1000L);

	/* Force initial hourly and minutely processing */

	currentHour = -1;
	currentMin = -1;
	lastClockTime = 0;			/* Force initial processing */
	lastTime = time (0);		/* Initial time for alarm processing rate determination */

	/* This callback will be called when the connection is established or lost */

	ca_add_fd_registration (reconnectCallback, NULL);
	doConnect = 1;
	caFd = -1;

	if (debugLevel >= 2)
	    tisSysLog (0, 2, "alarmMgr(%d): Registered with CA\n", myPid);

	/* Attempt to setup listening */

	setupListenForVu ();
	setupListenForMgr ();
	setupListenForPrinters ();

	/* Start all print servers */

	startPrinterServers ();

	/* Wait for an incoming message to process or until a timeout occurs */

	running = 1;
	while (running) {

	    /* Wait for any incoming updates */

	    ca_pend_io ((float)0.0);

	    /* Check if a reconnect is necessary */

	    if (doConnect) {

		if (debugLevel >= 2)
		    tisSysLog (0, 2, "alarmMgr(%d): Connecting to CA\n", myPid);

		if (!caInitialized) {
		    ca_task_exit ();
		    if (ca_task_initialize () != ECA_NORMAL) {
			tisSysLog (0, 2, "alarmMgr(%d): CA won't initialize\n", myPid);
#ifdef WIN32
			Sleep (10000);
#else
			sleep (10);
#endif WIN32
			exit (-1);
		    }
		    caInitialized = 1;
		}

		ca_add_fd_registration (reconnectCallback, NULL);

		if (ca_register_as_alarm_handler (dataCallback, connectionCallback) != ECA_NORMAL) {
		    tisSysLog (0, 2, "alarmMgr(%d): Can't register as archiver with CA\n", myPid);
		    exit (-1);
		}

		clConnect ();

		doConnect = 0;

	    }

	    /* Check for any connections from alarmVu or alarmMgr */

	    checkForAlarmVuConnect ();
	    checkForAlarmMgrConnect ();
	    checkForPrinterConnect ();

	    /* Check for any messages from alarmVu, alarmMgr, or print servers */

	    readAllFds ();

	    /* Process any pending messages from alarmVu or alarmMgr */

	    processPendingVuMessages ();
	    processPendingMgrMessages ();
	    processPendingPrinterMessages ();

	    /* Update the values of the aux select PV's */

	    if (channelChanged) {
		updateAuxSelect ();
		channelChanged = 0;
	    }

	    /* Process any incoming requests */

	    processPendingAlarms ();

	    /* Get the current time to determine if it is time to do things. */

	    clock = time (0);
	    t = localtime (&clock);

	    /* These things are done every 5 seconds */

	    if ((clock - lastClockTime) >= 5) {

		/* Set up listen sockets (if not already) */

		setupListenForVu ();
		setupListenForMgr ();
		setupListenForPrinters ();

		/* Connect to peer alarmMgr's if not already */

		connectToPeerAlarmMgrs ();

		/* Check for if any alarm managers have timed out */

		checkForMgrTimeout ();

		lastClockTime = clock;
	    }


	    if ((clock - lastPingTime) >= (ping_timeout / 3)) {

		/* Ping our peers to let then know we are here */

		sendPingToPeers ();

		lastPingTime = clock;
	    }

	    if (t->tm_min != currentMin) {

		/* Perform once per minute processing here */

		if (debugLevel >= 1)
		    tisSysLog (0, 2, "alarmMgr(%d): Processed %.2lf alarms/second",
				myPid, alarmsProcessed / (double)(time (0) - lastTime));
		lastTime = time (0);
		alarmsProcessed = 0;

		/* Send a test message to each printer connected if in printer test mode */

		if (printerTest)
		    sendPrinterTestMessages ();

		/* Reconfigure if the tree has been modified */

		reconfigure ();
		
		currentMin = t->tm_min;
	    }

	    if (t->tm_hour != currentHour) {

		currentHour = t->tm_hour;

		/* Perform once per hour processing here */

		/* Start archiving if the proper time */

		startArchiving (currentHour);

		/* Launch the operator tranking archiver if it is time */

		launchOpTrkArchiver (currentHour);

	    }

	    /* Do some archiving if time */

	    archiveAndClean ();

	}

	/* Close all channels in the channel list */

	clTerminate ();
	
	/* Terminate the tree system */

	terminateTree ();

	/* Free shared memory segments */

	if (!freeSharedMemory ())
	    tisSysLog (0, 2, "alarmMgr(%d): Unable to free shared memory segments.\n", myPid);

	/* Close all the fd's */

	closeAlarmVuFds ();
	closeAlarmMgrFds ();
	closePrinterFds ();
	close (vuListenFd);
	close (mgrListenFd);
	close (printListenFd);

	/* Close the log file, pvmap, and point state file */

	closePvMapFile ();
	closePdbXrefFile ();
	closeLogFile ();
	closeResListFile ();
	closePointStateFile ();

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Terminated.\n", myPid);

	return 0;

}

/*----------*/
static void reconnectCallback (void *arg, int fd, int cond)
    {

	if (cond == 0) {
	    doConnect = 1;
	    caInitialized = 0;
	    if (debugLevel >= 1)
		tisSysLog (0, 2, "alarmMgr(%d): Connection to CA lost.\n", myPid);
	}
	caFd = fd;

    }

/*----------*/
static void dataCallback (tisAlarm *pAlarm)
    {

	tisAlarm	alarm;

	/* Copy structure to local to avoid alignment problems */

	memcpy ((char *)&alarm, (char *)pAlarm, sizeof (tisAlarm));

	if (debugLevel >= 3)
	    tisSysLog (0, 2, "alarmMgr(%d): Callback for %s, stat = %d, sevr = %d, ack = %d, value = %s, ser=%d, ts=%ld,%ld",
				myPid,
				alarm.pvName,
				(int)alarm.alarmStat,
				(int)alarm.alarmSevr,
				(int)alarm.alarmAck,
				alarm.alarmValue,
				htonl(alarm.alarmSerial),
				alarm.alarmTime.secPastEpoch, alarm.alarmTime.nsec);

	/* Check for an invalid timestamp. */

	if (alarm.alarmTime.secPastEpoch < 100000)
	    return;

	/* Insert alarm onto the queue */

	if (!aqInsertAlarm (&alarmQueue, &alarm))
	    if (debugLevel >= 1)
		tisSysLog (0, 2, "alarmMgr(%d): Alarm queue from CA full - message from CA discarded.\n",
			myPid);

    }

/*----------*/
static void connectionCallback (char *resName, short status)

    {


    }

/************************************************************************************/
/*                            Communications with alarmVu			    */
/************************************************************************************/

/*----------*/
static int setupListenForVu (VOID)

    {

	int	optval;
	struct sockaddr_in serverAddr;

	if (listeningForAlarmVu)
	    return 1;

	if ((vuListenFd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't create listen socket for alarmVu", myPid);
	    return 0;
	}

	optval = 1;
	setsockopt (vuListenFd, SOL_SOCKET, SO_REUSEADDR, (char *)&optval, sizeof (optval));

#ifdef WIN32
	if (ioctlsocket (vuListenFd, FIONBIO, &optionValue) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't set alarmVu listen socket to nonblocking", myPid);
	    closesocket (vuListenFd);
#else
	if (ioctl (vuListenFd, FIONBIO, &optionValue) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't set alarmVu listen socket to nonblocking", myPid);
	    close (vuListenFd);
#endif WIN32
	    return 0;
	}

	memset (&serverAddr, 0, sizeof (serverAddr));
	serverAddr.sin_family = AF_INET;
	serverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	serverAddr.sin_port = htons(vuPort);

	if (bind (vuListenFd, (struct sockaddr *) &serverAddr, sizeof(serverAddr)) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't bind to TCP port for alarmVu", myPid);
	    close (vuListenFd);
	    return 0;
	}

	if (listen (vuListenFd, 5) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't listen on TCP port for alarmVu", myPid);
	    close (vuListenFd);
	    return 0;
	}

	listeningForAlarmVu = 1;
	return 1;

    }

/*----------*/
static void checkForAlarmVuConnect (VOID)

    {

	int	vuAcceptFd, i;

	saddrLen = sizeof (saddr);
	if ((vuAcceptFd = accept (vuListenFd, &saddr, &saddrLen)) >= 0) {
	    for (i=0 ; i<MAX_ALARMVUS ; i++) {
		if (vuConn[i].fd == -1) {
		    /* Someone is indeed trying to talk to us. Add them to the list. */
		    if (debugLevel >= 1)
			tisSysLog (0, 2, "alarmMgr(%d): alarmVu has connected on %d.\n",
				myPid, vuAcceptFd);
		    vuConn[i].fd = vuAcceptFd;
		    vuConn[i].numBytes = 0;
		    break;
		}
	    }
	    if (i >= MAX_ALARMVUS) {
		close (vuAcceptFd);
		tisSysLog (0, 2, "alarmMgr(%d): Max number of alarmvus reached: %d\n",
				myPid, MAX_ALARMVUS);
	    }
	}

    }

/*----------*/
static void processPendingVuMessages ()

    {

	int		i, informAlarmVu;
	TS_STAMP	ackTime;

	/* Process messages from alarmVus */

	informAlarmVu = 0;
	for (i=0 ; i<MAX_ALARMVUS ; i++) {
	    int type;
	    tsLocalTime (&ackTime);
	    if (type = getVuMsgType (i)) {	/* Got a message type */
		switch (type) {

		    case ACK_MSG:
			{
			    AlarmMsg msg;
			    if (getVuBytes (i, (char *)&msg, sizeof (msg))) {
				memcpy ((char *)&msg.ackTime, (char *)&ackTime, sizeof (TS_STAMP));
				processAck (&msg, FROM_ALARMVU);
			    }
			}
			break;

		    case ACK_GROUP_MSG:		/* Group ack */
			{
			    AlarmMsg msg;
			    if (getVuBytes (i, (char *)&msg, sizeof (msg))) {
				memcpy ((char *)&msg.ackTime, (char *)&ackTime, sizeof (TS_STAMP));
				processAckGroup (&msg, FROM_ALARMVU);
			    }
			}
			break;

		    case ENABLE_MSG:
			{
			    AlarmMsg msg;
			    if (getVuBytes (i, (char *)&msg, sizeof (msg)))
				processEnable (&msg);
			}
			break;

		    case DISABLE_MSG:
			{
			    AlarmMsg msg;
			    if (getVuBytes (i, (char *)&msg, sizeof (msg)))
				processDisable (&msg);
			}
			break;

		    case UNAVAILABLE:	/* Reply from alarmVu indicating that he */
					/* is clear of the shared memory area. */
			{
			    TypeOnlyMessage msg;
			    if (getVuBytes (i, (char *)&msg, sizeof (msg)))
				vuConn [i].usingShared = 0;
			}
			break;

		    case ALARM_COMMENT_UPDATE:
			{
			    AlarmMemo msg;
			    if (getVuBytes (i, (char *)&msg, sizeof (msg))) {
				processMemo (&msg, FROM_ALARMVU);
				informAlarmVu = 1;
			    }
			}
			break;

		    default:
			if (debugLevel >= 1)
			    tisSysLog (0, 2, "alarmMgr(%d): Synchronization with alarmVu on fd %d lost -- flushing message buffer.\n",
			myPid, vuConn [i].fd);
			flushVuBuffer (i);
			break;
		}
	    }
	}

	if (informAlarmVu)
	    sendSimpleMessageToAlarmvu (UPDATE);

    }

/*----------*/
static int getVuMsgType (int i)		/* Get message type of next message (if present) */

    {

	if (vuConn [i].numBytes >= sizeof (long))
	    return *(long *)vuConn [i].msgBuffer;
	return 0;

    }
    
/*----------*/
static int getVuBytes (int i, char *buf, int numBytes)

    {

	if (vuConn [i].numBytes >= numBytes) {
	    memcpy (buf, vuConn [i].msgBuffer, numBytes);
	    vuConn [i].numBytes -= numBytes;
	    memcpy (vuConn [i].msgBuffer, &vuConn [i].msgBuffer [numBytes],
			vuConn [i].numBytes);
	    return 1;
	}

	return 0;
	
    }

/*----------*/
static void flushVuBuffer (int i)

    {

	vuConn [i].numBytes = 0;

    }

/*----------*/
static void closeAlarmVuFds (VOID)

    {

	int	i;

	for (i=0 ; i<MAX_ALARMVUS ; i++) {
	    if (vuConn [i].fd != -1) {
		close (vuConn [i].fd);
	    }
	}

    }

/*----------*/
/* Send an update message to all connected alarmVu's */
void sendSimpleMessageToAlarmvu (int msgType)

    {

	TypeOnlyMessage	msg;
	int		stat, i;

	msg.msgType = msgType;
	for (i=0 ; i<MAX_ALARMVUS ; i++) {
	    if (vuConn [i].fd != -1) {
		stat = send (vuConn [i].fd, (char *)&msg, sizeof (msg), 0);
		if (stat <= 0) {
		    if (debugLevel >= 1)
			tisSysLog (0, 2, "alarmMgr(%d): alarmVu on %d has disconnected.\n",
					myPid, vuConn [i].fd);
		    close (vuConn [i].fd);
		    vuConn [i].fd = -1;
		}
	    }
	}

    }

/*----------*/
void oustAlarmVus (int oustingType)	/* Tell alarmVus to get out of the shared memory */

    {

	int	i, notClear;

	/* Assume all alarmVu's are using the shared memory. */

	for (i=0 ; i<MAX_ALARMVUS ; i++)
	    vuConn [i].usingShared = 1;

	/* Wait for all alarmVu's to get clear */

	sendSimpleMessageToAlarmvu (oustingType);

	do {

	    /* Check for and process any message that show up */

	    readAllFds ();
	    processPendingVuMessages ();
	    processPendingMgrMessages ();
	    processPendingPrinterMessages ();

	    /* Check if all alarmVu's are now clear */

	    notClear = 0;
	    for (i=0 ; i<MAX_ALARMVUS ; i++)
		if (vuConn [i].fd != -1)
		    notClear += vuConn [i].usingShared;

	    /* Use pend_event to perform delay. This will queue any pending alarms. */

	   if (notClear)
		ca_pend_event ((float)1.0);

	} while (notClear);

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): All alarmVu's have cleared from shared memory\n", myPid);

    }

/*----------*/
void allowAlarmVus (int oustingType)	/* Inform alarmVu that the alarmMgr is available */

    {

	sendSimpleMessageToAlarmvu (oustingType);

    }

/************************************************************************************/
/*                          Communications with peer alarmMgr			    */
/************************************************************************************/

/*----------*/
static void buildPeerList (VOID)

    {

	HostName	hostName;
	int		isHost;
	struct hostent	*hostptr;
	char myName [30];

	openIOCListFile ();

	/* Get my name so I don't make myself a peer */

	gethostname (myName, sizeof (myName));

	/* Read through iocList and find my peers */
 
	numPeers = 0;
	while (getNextIOCEntry (hostName, &isHost)) {
	    if (isHost && (strcmp (hostName, myName) != 0)) {
		stringCopy (mgrConn [numPeers].peerName, hostName, sizeof (HostName));
		mgrConn [numPeers].fd = -1;
		mgrConn [numPeers].connectInProgress = 0;

		if ((hostptr = gethostbyname (hostName)) == NULL) {
		    tisSysLog (0, 2, "alarmMgr(%d): Unable to find address of %s\n",
				myPid, hostName);
		    continue;
		}
  
		mgrConn [numPeers].ipAddr = htonl(((struct in_addr *) *(hostptr->h_addr_list))->s_addr);

		numPeers++;
		if (debugLevel >= 1)
		    tisSysLog (0, 2, "alarmMgr(%d): %s is a peer.\n", myPid, hostName);
	    }
	}

	closeIOCListFile ();

	/* Check if I am lowest IP address */

	hostptr = gethostbyname (myName);
	myIpAddr = htonl(((struct in_addr *) *(hostptr->h_addr_list))->s_addr);

    }

/*----------*/
static void connectToPeerAlarmMgrs (VOID)

    {

	int		alarmMgrFd, i, status;
	struct sockaddr_in peerAddr;

	/* Connect to alarmMgrs with a higher IP address */

	for (i=0 ; i<numPeers ; i++ ) {

	    if ((mgrConn [i].fd == -1) && !mgrConn [i].connectInProgress) {

		if (mgrConn [i].ipAddr <= myIpAddr)
		    continue;

		/* Look up the address of the server host */

		if ((alarmMgrFd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
		    tisSysLog (0, 2, "alarmMgr(%d): Can't create socket to connect to peer alarmMgr",
				myPid);
		    continue;
		}

#ifdef WIN32
		if (ioctlsocket (alarmMgrFd, FIONBIO, &optionValue) == -1) {
#else
		if (ioctl (alarmMgrFd, FIONBIO, &optionValue) == -1) {
#endif WIN32
		    tisSysLog (0, 2, "alarmMgr(%d): Can't set socket nonblocking for peer alarmMgr", myPid);
		    exit (-1);
		}
		memset ((char *) &peerAddr, 0, sizeof (peerAddr));
		peerAddr.sin_family = AF_INET;
		peerAddr.sin_addr.s_addr = htonl (mgrConn [i].ipAddr);
		peerAddr.sin_port = htons(mgrPort);
		if (debugLevel >= 2)
		    tisSysLog (0, 2, "alarmMgr(%d): Performing connect to %s", myPid, mgrConn [i].peerName);
		status = connect (alarmMgrFd, (struct sockaddr *) &peerAddr, sizeof (peerAddr));
		if (debugLevel >= 2)
#ifndef SOLARIS
		    tisSysLog (0, 2, "alarmMgr(%d): Connect returned %d, errno=%d, %s", myPid, status, errno, sys_errlist [errno]);
#else
		    tisSysLog (0, 2, "alarmMgr(%d): Connect returned %d, errno=%d, %s", myPid, status, errno, strerror (errno));
#endif
		if (status == -1) {
#ifdef WIN32
		    if (errno == WSAEINPROGRESS) {
#else
		    if (errno == EINPROGRESS) {
#endif WIN32
			mgrConn [i].fd = alarmMgrFd;
			mgrConn [i].connectInProgress = 1;
			mgrConn [i].lastPing = time (0);
			if (debugLevel >= 1)
			    tisSysLog (0, 2, "alarmMgr(%d): Attempting connection alarmMgr on %s, fd=%d\n",
					myPid, mgrConn [i].peerName, alarmMgrFd);
		    } else {
			if (debugLevel >= 1)
			    tisSysLog (0, 2, "alarmMgr(%d): Connection to alarmMgr on %s, fd=%d refused.\n",
					myPid, mgrConn [i].peerName, alarmMgrFd);
#ifdef WIN32
			closesocket (alarmMgrFd);
#else
			close (alarmMgrFd);
#endif WIN32
			return;
		    }
		} else {
		    mgrConn [i].fd = alarmMgrFd;
		    mgrConn [i].connectInProgress = 0;
		    mgrConn [i].lastPing = time (0);
		    if (debugLevel >= 1)
			tisSysLog (0, 2, "alarmMgr(%d): Connected to alarmMgr on %s, fd=%d\n",
				myPid, mgrConn [i].peerName, alarmMgrFd);
		}

	    }
	}

    }

/*----------*/
static int setupListenForMgr (VOID)

    {

	int	optval;
	struct sockaddr_in serverAddr;

	if (listeningForAlarmMgrs)
	    return 1;

	if ((mgrListenFd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't create listen socket for alarmMgr", myPid);
	    return 0;
	}

	optval = 1;
	setsockopt (mgrListenFd, SOL_SOCKET, SO_REUSEADDR, (char *)&optval, sizeof (optval));

#ifdef WIN32
	if (ioctlsocket (mgrListenFd, FIONBIO, &optionValue) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't set alarmMgr listen socket to nonblocking", myPid);
	    closesocket (mgrListenFd);
#else
	if (ioctl (mgrListenFd, FIONBIO, &optionValue) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't set alarmMgr listen socket to nonblocking", myPid);
	    close (mgrListenFd);
#endif WIN32
	    return 0;
	}

	memset (&serverAddr, 0, sizeof (serverAddr));
	serverAddr.sin_family = AF_INET;
	serverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	serverAddr.sin_port = htons(mgrPort);

	if (bind (mgrListenFd, (struct sockaddr *) &serverAddr, sizeof(serverAddr)) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't bind to TCP port for alarmMgr", myPid);
#ifdef WIN32
	    closesocket (mgrListenFd);
#else
	    close (mgrListenFd);
#endif WIN32
	    return 0;
	}

	if (listen (mgrListenFd, 5) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't listen on TCP port for alarmMgr", myPid);
#ifdef WIN32
	    closesocket (mgrListenFd);
#else
	    close (mgrListenFd);
#endif WIN32
	    return 0;
	}

	listeningForAlarmMgrs = 1;
	return 1;

    }

/*----------*/
static void checkForAlarmMgrConnect (VOID)

    {

	int	mgrAcceptFd, saddrLen, addrLen, i;
	struct sockaddr_in saddr;
	struct hostent *host;

	saddrLen = sizeof (saddr);
	if ((mgrAcceptFd = accept (mgrListenFd, (struct sockaddr *)&saddr, &saddrLen)) >= 0) {
#ifdef WIN32
	    ioctlsocket (mgrAcceptFd, FIONBIO, &optionValue);
#else
	    ioctl (mgrAcceptFd, FIONBIO, &optionValue);
#endif WIN32

	    /* Someone is trying to connect. Get the host name */

	    addrLen = sizeof (saddr.sin_addr);
	    host = gethostbyaddr ((char *)&saddr.sin_addr, addrLen, AF_INET);
	    
	    /* This is really wierd! Somtimes gethostbyaddr returns NULL -- beats me. */

	    if (host == NULL) {
#ifdef WIN32
		closesocket (mgrAcceptFd);
#else
		close (mgrAcceptFd);
#endif WIN32
#ifndef AXP
		tisSysLog (0, 2, "alarmMgr(%d): Peer at %d.%d.%d.%d attempted connection on fd=%d but unable to resolve address into host name.",
				myPid, (int)(saddr.sin_addr.S_un.S_un_b.s_b1), (int)(saddr.sin_addr.S_un.S_un_b.s_b2),
				(int)(saddr.sin_addr.S_un.S_un_b.s_b3), (int)(saddr.sin_addr.S_un.S_un_b.s_b4), mgrAcceptFd);
#else
		tisSysLog (0, 2, "alarmMgr(%d): Peer at %d.%d.%d.%d attempted connection on fd=%d but unable to resolve address into host name.",
				myPid, (int)(saddr.sin_addr.s_addr >> 24) & 0xff, (int)(saddr.sin_addr.s_addr >> 16) & 0xff,
				(int)(saddr.sin_addr.s_addr >> 8) & 0xff, (int)saddr.sin_addr.s_addr & 0xff, mgrAcceptFd);
#endif
		return;
	    }

	    for (i=0 ; i<numPeers ; i++) {
		if (strcmp (mgrConn [i].peerName, host->h_name) == 0) {
		    if (debugLevel >= 1)
		        tisSysLog (0, 2, "alarmMgr(%d): alarmMgr on %s has connected on %d.\n",
					myPid, host->h_name, mgrAcceptFd);
		    mgrConn [i].fd = mgrAcceptFd;
		    mgrConn [i].numBytes = 0;
		    mgrConn [i].lastPing = time (0);
		    mgrConn [i].connectInProgress = 0;
		    break;
		}
	    }
	    if (i == numPeers) {
#ifdef WIN32
		closesocket (mgrAcceptFd);
#else
		close (mgrAcceptFd);
#endif WIN32
		tisSysLog (0, 2, "alarmMgr(%d): Connection attempt from unexpected host %s\n",
				myPid, host->h_name);
	    }
	}

    }

/*----------*/
static void processPendingMgrMessages (VOID)

    {

	int	i;
	static	int	cnt = 0;

	/* Process messages from alarmMgrs */

	for (i=0 ; i<numPeers ; i++) {
	    int type;
	    if (type = getMgrMsgType (i)) {	/* Got a message type */

		switch (type) {

		    case ALARM_COMMENT_UPDATE:
			{
			    AlarmMemo msg;
			    if (getMgrBytes (i, (char *)&msg, sizeof (msg))) {

				/* Convert to host byte ordering */

				msg.msgType = ntohl (msg.msgType);
				msg.timeStamp.secPastEpoch = ntohl (msg.timeStamp.secPastEpoch);
				msg.timeStamp.nsec = ntohl (msg.timeStamp.nsec);

				/* Process the memo message */

				processMemo (&msg, FROM_ALARMMGR);
			    }
			}
			break;

		    case ACK_MSG:
			{
			    AlarmMsg msg;
			    if (getMgrBytes (i, (char *)&msg, sizeof (msg))) {

				/* Convert to host byte ordering */

				msg.msgType = ntohl (msg.msgType);
				msg.timeStamp.secPastEpoch = ntohl (msg.timeStamp.secPastEpoch);
				msg.timeStamp.nsec = ntohl (msg.timeStamp.nsec);
				msg.ackTime.secPastEpoch = ntohl (msg.ackTime.secPastEpoch);
				msg.ackTime.nsec = ntohl (msg.ackTime.nsec);
				msg.memberOfGroup = ntohl (msg.memberOfGroup);
				msg.status = ntohl (msg.status);
				msg.severity = ntohl (msg.severity);

				/* Process the ack */

				processAck (&msg, FROM_ALARMMGR);
			    }
			}
			break;

		    case ACK_GROUP_MSG:		/* Group ack */
			{
			    AlarmMsg msg;


			    if (getMgrBytes (i, (char *)&msg, sizeof (msg))) {

				/* Convert to host byte ordering */

				msg.msgType = ntohl (msg.msgType);
				msg.timeStamp.secPastEpoch = ntohl (msg.timeStamp.secPastEpoch);
				msg.timeStamp.nsec = ntohl (msg.timeStamp.nsec);
				msg.ackTime.secPastEpoch = ntohl (msg.ackTime.secPastEpoch);
				msg.ackTime.nsec = ntohl (msg.ackTime.nsec);
				msg.memberOfGroup = ntohl (msg.memberOfGroup);
				msg.status = ntohl (msg.status);
				msg.severity = ntohl (msg.severity);

				/* Process the group ack */

				processAckGroup (&msg, FROM_ALARMMGR);
			    }
			}
			break;

		    case PING:		/* General ping */
			{

			    TypeOnlyMessage msg;
			    if (getMgrBytes (i, (char *)&msg, sizeof (msg))) {
				if (displayPings)
				    tisSysLog (0, 2, "alarmMgr(%d): Ping %d received from %s\n",
						myPid, cnt++, mgrConn [i].peerName);
				mgrConn [i].lastPing = time (0);
			    }

			}
			break;

		    default:
			tisSysLog (0, 2, "alarmMgr(%d): Unknown message type %d", myPid, type);
			break;
		}
	    }
	}

    }

/*----------*/
static int getMgrMsgType (int i)	/* Get message type of next message (if present) */

    {

	if (mgrConn [i].numBytes >= sizeof (uint32))
	    return ntohl (*(uint32 *)mgrConn [i].msgBuffer);
	    
	return 0;

    }
    
/*----------*/
static int getMgrBytes ( int i, char *buf, int numBytes)

    {

	if (mgrConn [i].numBytes >= numBytes) {
	    memcpy (buf, mgrConn [i].msgBuffer, numBytes);
	    mgrConn [i].numBytes -= numBytes;
	    memcpy (mgrConn [i].msgBuffer, &mgrConn [i].msgBuffer [numBytes],
			mgrConn [i].numBytes);
	    return 1;
	}

	return 0;
	
    }

/*----------*/
static void checkForMgrTimeout (VOID)

    {

	int	i, resetOne = 0;

	for (i=0 ; i<numPeers ; i++) {
	    if (mgrConn [i].fd != -1) {
		if ((time (0) - mgrConn [i].lastPing) > ping_timeout) {
		    if (debugLevel >= 1)
			tisSysLog (0, 2, "alarmMgr(%d): Timeout on fd %d, closed.\n",
					myPid, mgrConn [i].fd);
		    close (mgrConn [i].fd);
		    mgrConn [i].fd = -1;
		    mgrConn [i].connectInProgress = 0;
		    resetOne = 1;
		}
	    }
	}

    }
		    
/*----------*/
static void closeAlarmMgrFds (VOID)

    {

	int	i;

	for (i=0 ; i<numPeers ; i++) {
	    if (mgrConn [i].fd != -1) {
#ifdef WIN32
		closesocket (mgrConn [i].fd);
#else
		close (mgrConn [i].fd);
#endif WIN32
	    }
	}

    }

/*----------*/
static void sendPingToPeers (VOID)

    {

	TypeOnlyMessage	msg;

	msg.msgType = htonl (PING);
	sendMessageToPeers ((char *)&msg, sizeof (msg));

    }

/*----------*/
/* Send a message to all connected alarmMgr's */
void sendMessageToPeers (char *msg, int size)	

    {

	int		stat, i;

	for (i=0 ; i<numPeers ; i++) {
	    if ((mgrConn [i].fd != -1) && !mgrConn [i].connectInProgress) {
		stat = sendMsg (mgrConn [i].fd, msg, size);
		if (stat <= 0) {
		    if (debugLevel >= 1)
			tisSysLog (0, 2, "alarmMgr(%d): alarmMgr on %s has disconnected, fd=%d\n",
					myPid, mgrConn [i].peerName, mgrConn [i].fd);
#ifdef WIN32
		    closesocket (mgrConn [i].fd);
#else
		    close (mgrConn [i].fd);
#endif WIN32
		    mgrConn [i].fd = -1;
		    mgrConn [i].connectInProgress = 0;
		}
	    }
	}

    }

/************************************************************************************/
/*                          Communications with print servers			    */
/************************************************************************************/

/*----------*/
static int setupListenForPrinters (VOID)

    {

	int	optval, addrSize;
	struct sockaddr_in serverAddr;

	if (listeningForPrinters)
	    return 1;

	/* Open a stream socket. */

	if ((printListenFd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
            tisSysLog (0, 2, "alarmMgr(%d): Fd creation error for alarmPrinter listen fd\n",
			myPid);
            return 0;
	}
    
	optval = 1;
	setsockopt (printListenFd, SOL_SOCKET, SO_REUSEADDR, (char *)&optval, sizeof (optval));

#ifdef WIN32
	if (ioctlsocket (printListenFd, FIONBIO, &optionValue) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't set alarmPrinter listen socket to nonblocking",
			myPid);
	    closesocket (printListenFd);
#else
	if (ioctl (printListenFd, FIONBIO, &optionValue) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Can't set alarmPrinter listen socket to nonblocking",
			myPid);
	    close (printListenFd);
#endif WIN32
	    return 0;
	}

	/* Zero the sock_addr structure. */

	memset (&serverAddr, 0, sizeof(serverAddr));
	serverAddr.sin_family = AF_INET;
	serverAddr.sin_port = htons(0);

	/* get server port address. */

	if (bind (printListenFd, (struct sockaddr *)&serverAddr, sizeof(serverAddr)) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Unable to bind to alarmPrinter port.\n", myPid);
#ifdef WIN32
	    closesocket (printListenFd);
#else
	    close (printListenFd);
#endif WIN32
	    return 0;
	}
	addrSize = sizeof (serverAddr);
	if (getsockname (printListenFd, (struct sockaddr *)&serverAddr, &addrSize)) {
	    tisSysLog (0, 2, "alarmMgr(%d): getsockname error for alarmPrinter\n", myPid);
#ifdef WIN32
	    closesocket (printListenFd);
#else
	    close (printListenFd);
#endif WIN32
	    return 0;
	}
	printListenPort = (unsigned)(ntohs (serverAddr.sin_port));

	/* Listen for new connections. */

	if (listen (printListenFd, 5) == -1) {
	    tisSysLog (0, 2, "alarmMgr(%d): Unable to listen on printer server port.\n", myPid);
#ifdef WIN32
	    closesocket (printListenFd);
#else
	    close (printListenFd);
#endif WIN32
	    return 0;
	}

	if (debugLevel >= 2)
	    tisSysLog (0, 2, "alarmMgr(%d): Listening for print servers on port %u",
			myPid, printListenPort);

	listeningForPrinters = 1;
	return 1;

    }

/*----------*/
static void checkForPrinterConnect (VOID)

    {

	int	prAcceptFd, i;

	saddrLen = sizeof (saddr);
	while ((prAcceptFd = accept (printListenFd, &saddr, &saddrLen)) >= 0) {
	    if (debugLevel >= 2)
		tisSysLog (0, 2, "alarmMgr(%d): Accept fired on printer port, fd=%d", myPid, prAcceptFd);

	    for (i=0 ; i<MAX_PRINTERS ; i++) {
		if (printConn [i].fd == -1) {
		    /* Someone is indeed trying to talk to us. Add them to the list. */
		    if (debugLevel >= 1)
			tisSysLog (0, 2, "alarmMgr(%d): Print server has connected on %d.\n",
					myPid, prAcceptFd);
		    printConn [i].fd = prAcceptFd;
		    printConn [i].numBytes = 0;
		    break;
		} /* else if (ioctl (printConn [i].fd, FIONBIO, &optionValue) == -1) {
		    close (printConn[i].fd);
		    printConn [i].fd = -1;
		    tisSysLog (0, 2, "alarmMgr(%d): Can't set socket non-blocking for printer server\n",
				myPid);
		    break;
		} */
	    }
	    if (i >= MAX_PRINTERS) {
#ifdef WIN32
		closesocket (prAcceptFd);
#else
		close (prAcceptFd);
#endif WIN32
		tisSysLog (0, 2, "alarmMgr(%d): Max number of print servers reached: %d\n",
				myPid, MAX_PRINTERS);
	    }
	    /* this is required to prime for next exec in while (above) */
	    saddrLen = sizeof (saddr);
	}

    }

/*----------*/
void addPrinter (char *printerName)

    {

	int	i;

	/* If just a plain printer, add it to the list. If a printer group name, */
	/* add all printers in the group to the list. */

	if (*printerName != '#')
	    addOnePrinter (printerName);
	else {
	    PrinterGroupEntry *p = getPrinterGroup (printerName + 1);
	    for (i=0 ; i<MAX_PRINTERS_PER_GROUP ; i++) {
		if (strlen (p->printers [i])) {
		    addOnePrinter (p->printers [i]);
		}
	    }
	}

    }

/*----------*/
void addOnePrinter (char *printerName)

    {

	int	i;

	/* Find place in printer list. */

	for (i=0 ; i<numPrinters ; i++) {
	    if (strcmp (printerList [i].name, printerName) == 0) {
		return;	/* Found it. */
	    }
	}

	stringCopy (printerList [numPrinters].name, printerName, sizeof (PathName));
	printerList [numPrinters].running = 0;

	numPrinters++;

    }

/*----------*/
static void startPrinterServers (VOID)

    {   

	int	i;

	for (i=0 ; i<numPrinters ; i++) {
	    if (!printerList [i].running) {
#ifdef WIN32
	    char portText [10];

	    sprintf (portText, "%u", printListenPort);
	    _spawnl (_P_DETACH, "\\tis4000\\tasks\\alarmPrinter", "alarmPrinter",
			    printerList [i].name, portText, NULL);
#else
		char cmd [100];
		sprintf (cmd, "sh -c \"/tis4000/tasks/alarmPrinter %s %u &\"",
				printerList [i].name, printListenPort);
		system (cmd);
#endif WIN32
		printerList [i].running = 1;
	    }
	}

    }

/*----------*/
static void processPendingPrinterMessages (VOID)

    {

	int	i, j;
	static	PrinterNameMsg msg;

	/* Process messages from printer servers */

	for (i=0 ; i<numPrinters ; i++) {
	    int type;
	    if (type = getPrinterMsgType (i)) {	/* Got a message type */
		switch (type) {

		    /* A print server is reporting it's name. Associate it with the proper */ 		   
		    /* entry in the printer table. */


		    case PRINTER_NAME_MSG:
			{
			    if (getPrinterBytes (i, (char *)&msg, sizeof (msg))) {
				if (debugLevel >= 3) {
				    tisSysLog (0, 2, "alarmMgr(%d): Received mesasge from printer %s, fd=%d",
						myPid, msg.printerName, printConn [i].fd); 
				    logBytes ((char *)&msg, sizeof (msg));
				}
				for (j=0 ; j<MAX_PRINTERS ; j++) {
				    if (strcmp (printerList [j].name, msg.printerName) == 0) {
					printConn [i].nameIx = j;
					break;
				    }
				}
				if (debugLevel >= 1)
				    tisSysLog (0, 2, "alarmMgr(%d): Printer server for %s is ready\n",
						myPid, msg.printerName);
			    }
			}
			break;
		    default:
			tisSysLog (0, 2, "alarmMgr(%d): Unknown message from printer %s, type = %d",
					myPid, msg.printerName, type);
		}
	    }
	}

    }

/*----------*/
/* Get message type of next printer message (if present) */
static int getPrinterMsgType (int i)	

    {

	if (printConn [i].numBytes >= sizeof (long))
	    return ntohl (*(long *)printConn [i].msgBuffer);
	    
	return 0;

    }
    
/*----------*/
static int getPrinterBytes ( int i, char *buf, int numBytes)

    {

	if (printConn [i].numBytes >= numBytes) {
	    memcpy (buf, printConn [i].msgBuffer, numBytes);
	    printConn [i].numBytes -= numBytes;
	    memcpy (printConn [i].msgBuffer, &printConn [i].msgBuffer [numBytes],
			printConn [i].numBytes);

	    return 1;
	}

	return 0;
	
    }

/*----------*/
static void closePrinterFds (VOID)

    {

	int	i;

	for (i=0 ; i<MAX_PRINTERS ; i++) {
	    if (printConn [i].fd != -1) {
#ifdef WIN32
		closesocket (printConn [i].fd);
#else
		close (printConn [i].fd);
#endif WIN32
	    }
	}

    }

/*----------*/
int sendMsgToPrinter ( char *printer, char *msg)

    {

	int	i;

	/* If the printer is just a single printer, send the message to the appropriate */
	/* print server.  If printer is a group, send to all print servers in the group. */

	if (*printer != '#')
	    sendMsgToPrinterServer (printer, msg);
	else {
	    PrinterGroupEntry *p = getPrinterGroup (printer + 1);
	    for (i=0 ; i<MAX_PRINTERS_PER_GROUP ; i++)
		if (strlen (p->printers [i]))
		    sendMsgToPrinterServer (p->printers [i], msg);
	}

	return 1;	/* At this point we must assume the alarm was printed. */

    }

/*----------*/
static int sendMsgToPrinterServer ( char *printer, char *msg)

    {

	int i, fd = -1;

	/* Find the fd of the desired printer */

	for (i=0 ; i<MAX_PRINTERS ; i++) {
	    if (printConn [i].fd != -1) {
		if (strcmp (printer, printerList [printConn [i].nameIx].name) == 0) {
		    fd = printConn [i].fd;
		    break;
		}
	    }
	}
	if (fd == -1)
	    return 0;

	if (!sendMsg (fd, msg, strlen (msg))) {
	    if (debugLevel >= 1)
		tisSysLog (0, 2, "alarmMgr(%d): Printer on fd %d disconnected.\n", myPid, fd);
	    printConn [i].fd = -1;
	    return 0;
	}

	if (debugLevel >= 3) {
	    tisSysLog (0, 2, "alarmMgr(%d): Sending message to printer on fd=%d", myPid, fd);
	    logBytes (msg, strlen (msg));
	}

	return 1;

    }

/*----------*/
static void sendPrinterTestMessages ()	/* Send a test message to each printer connected */

    {

	int	i;
	char	msg [100];

	for (i=0 ; i<MAX_PRINTERS ; i++) {
	    if (printConn [i].fd != -1) {
		sprintf (msg, "Printer Test: This printer is %s\n", printerList [printConn [i].nameIx].name);
		if (!sendMsg (printConn [i].fd, msg, strlen (msg))) {
		    if (debugLevel >= 1)
			tisSysLog (0, 2, "alarmMgr(%d): Printer on fd %d disconnected.\n", myPid, printConn [i].fd);
		    printConn [i].fd = -1;
		}
		tisSysLog (0, 2, "alarmMgr(%d): Sending test message to printer %s on fd=%d",
			myPid, printerList [printConn [i].nameIx].name, printConn [i].fd);
	    }
	}

    }

/************************************************************************************/
/*	Generalized send							    */
/************************************************************************************/

static int sendMsg ( int fd, char *msg, int size)

    {

	int	bytesLeft, bytesSent;

	bytesLeft = size;
	do {
	    if (debugLevel >= 3) {
		tisSysLog (0, 2, "alarmMgr(%d): Sending %d bytes to fd=%d", myPid, bytesLeft, fd);
		logBytes (msg, size);
	    }
	    bytesSent = send (fd, msg, bytesLeft, 0);
	    if (debugLevel >= 3)
#ifndef SOLARIS
		tisSysLog (0, 2, "alarmMgr(%d): send returned %d, errno=%d, %s", myPid, bytesSent, errno, sys_errlist [errno]);
#else
		tisSysLog (0, 2, "alarmMgr(%d): send returned %d, errno=%d, %s", myPid, bytesSent, errno, strerror (errno));
#endif
	    if (bytesSent >= 0) {
		bytesLeft -= bytesSent;

	    } else {
#ifdef WIN32
		closesocket (fd);
#else
		close (fd);
#endif WIN32
		return 0;
	    }
	} while (bytesLeft > 0);
	    
	return 1;

    }

/************************************************************************************/
/*                     Process all pending incoming messages			    */
/************************************************************************************/

/*----------*/
static void readAllFds (VOID)	/* Read any data present at any of the alarmVu, alarmMgr, or printer sockets */
				/* Also check for any activity from CA */

    {

	int	i, maxFd, result, bytesLeft, bytesRead;
	struct timeval	timeOut;

	/* Build an fd set with all the alarmVu, alarmMgr, printers, and CA fds in it. */

	fd_set fdSet, fdWrSet;
	FD_ZERO (&fdSet);
	FD_ZERO (&fdWrSet);
	maxFd = -1;

	for (i=0 ; i<MAX_ALARMVUS ; i++) {
	    if (vuConn [i].fd != -1) {
		FD_SET(vuConn [i].fd, &fdSet);
		if (vuConn [i].fd > maxFd)
		    maxFd = vuConn [i].fd;
	    }
	}

	for (i=0 ; i<numPeers ; i++) {
	    if ((mgrConn [i].fd != -1) && !mgrConn [i].connectInProgress) {
		FD_SET(mgrConn [i].fd, &fdSet);
		if (mgrConn [i].fd > maxFd)
		    maxFd = mgrConn [i].fd;
	    }
	    if (mgrConn [i].connectInProgress) {
		FD_SET(mgrConn [i].fd, &fdWrSet);
		if (mgrConn [i].fd > maxFd)
		    maxFd = mgrConn [i].fd;
	    }
	}


	for (i=0 ; i<MAX_PRINTERS ; i++) {
	    if (printConn [i].fd != -1) {
		FD_SET(printConn [i].fd, &fdSet);
		if (printConn [i].fd > maxFd)
		    maxFd = printConn [i].fd;
	    }
	}

	if (caFd != -1) {
	    FD_SET(caFd, &fdSet);
	    if (caFd > maxFd)
		maxFd = caFd;
	}

	if (maxFd == -1)
	    return;
	
	/* See if any of the fds are ready for reading */

	timeOut.tv_sec = shareHeader->archivingActive ? 0 : 1;	/* Just poll if archiving */
	timeOut.tv_usec = 0;
	result = select (maxFd+1, &fdSet, &fdWrSet, NULL, &timeOut);
	if (result <= 0)
	    return;

	/* Check if the CA fd fired */

	if (FD_ISSET(caFd, &fdSet))
	    ca_pend_io ((float)0.0001);

	/* Drain any alarmVu data */

	for (i=0 ; i<MAX_ALARMVUS ; i++) {
	    if (vuConn [i].fd != -1) {
		if (FD_ISSET (vuConn [i].fd, &fdSet)) {
		    bytesLeft = AL_RECV_BUFFER_SIZE - vuConn [i].numBytes;
		    bytesRead = recv (vuConn [i].fd,
				&vuConn [i].msgBuffer [vuConn [i].numBytes], bytesLeft, 0);
		    if (bytesRead > 0) {
			vuConn [i].numBytes += bytesRead;
		    } else {
			if (debugLevel >= 1)
			    tisSysLog (0, 2, "alarmMgr(%d): alarmVu on fd %d disconnected.\n",
					myPid, vuConn [i].fd);
			close (vuConn [i].fd);
			vuConn [i].fd = -1;
		    }
		}
	    }
	}

	/* Check for any completed connections by peer alarmMgrs */

	for (i=0 ; i<numPeers ; i++) {
	    if ((mgrConn [i].fd != -1) && mgrConn [i].connectInProgress) {
		if (FD_ISSET (mgrConn [i].fd, &fdWrSet)) {
		    mgrConn [i].connectInProgress = 0;
		    mgrConn [i].lastPing = time (0);
		    if (debugLevel >= 2)
			tisSysLog (0, 2, "alarmMgr(%d): Connection completed for fd=%d", myPid, mgrConn [i].fd);
		}
	    }
	}

	/* Drain any alarmMgr data */

	for (i=0 ; i<numPeers ; i++) {
	    if ((mgrConn [i].fd != -1) && !mgrConn [i].connectInProgress) {
		if (FD_ISSET (mgrConn [i].fd, &fdSet)) {
		    bytesLeft = AL_RECV_BUFFER_SIZE - mgrConn [i].numBytes;
		    if (debugLevel >= 3)
			tisSysLog (0, 2, "alarmMgr(%d): Trying to read %d bytes from fd=%d", myPid, bytesLeft, mgrConn [i].fd);
		    bytesRead = recv (mgrConn [i].fd, &mgrConn [i].msgBuffer [mgrConn [i].numBytes], bytesLeft, 0);
#ifdef WIN32
		    errno = WSAGetLastError ();
#endif WIN32
		    if (debugLevel >= 3) {
#ifndef SOLARIS
			tisSysLog (0, 2, "alarmMgr(%d): Read returns %d, errno=%d, %s", myPid, bytesRead, errno, sys_errlist [errno]);
#else
			tisSysLog (0, 2, "alarmMgr(%d): Read returns %d, errno=%d, %s", myPid, bytesRead, errno, strerror (errno));
#endif
			if (bytesRead > 0)
			    logBytes (&mgrConn [i].msgBuffer [mgrConn [i].numBytes], bytesRead);
		    }
#ifdef WIN32
		    if (errno != WSAEWOULDBLOCK) {
#else
		    if (errno != EWOULDBLOCK) {
#endif WIN32
			if (bytesRead > 0) {
			    mgrConn [i].numBytes += bytesRead;
			} else {
#ifdef WIN32
			    if (errno != WSAECONNREFUSED)
#else
			    if (errno != ECONNREFUSED)
#endif WIN32
				if (debugLevel >= 1)
				    tisSysLog (0, 2, "alarmMgr(%d): alarmMgr on %s disconnected, fd=%d\n",
						myPid, mgrConn [i].peerName, mgrConn [i].fd);
#ifdef WIN32
			    closesocket (mgrConn [i].fd);
#else
			    close (mgrConn [i].fd);
#endif WIN32
			    mgrConn [i].fd = -1;
			    mgrConn [i].connectInProgress = 0;
			}
		    }
		}
	    }
	}

	/* Drain any printer data */

	for (i=0 ; i<MAX_PRINTERS ; i++) {
	    if (printConn [i].fd != -1) {
		if (FD_ISSET (printConn [i].fd, &fdSet)) {
		    bytesLeft = AL_RECV_BUFFER_SIZE - printConn [i].numBytes;
		    bytesRead = recv (printConn [i].fd,
				&printConn [i].msgBuffer [printConn [i].numBytes], bytesLeft, 0);
		    if (bytesRead > 0) {
			printConn [i].numBytes += bytesRead;
		    } else {
			if (debugLevel >= 1)
			    tisSysLog (0, 2, "alarmMgr(%d): alarmPrinter on fd %d disconnected.\n",
						myPid, printConn [i].fd);
#ifdef WIN32
			closesocket (printConn [i].fd);
#else
			close (printConn [i].fd);
#endif WIN32
			printConn [i].fd = -1;
		    }
		}
	    }
	}

    }

/************************************************************************************/
/*                   Reconfigure the alarm tree if it has changed		    */
/************************************************************************************/

/*----------*/
static void reconfigure (VOID)	/* Rebuild tree and point list */

    {

#ifdef WIN32
    struct _stat status;
#else
    struct stat status;
#endif WIN32

	/* Get the last modification date/time of the file.  If the file has not changed since */
	/* the last time we loaded it then don't reload. */

#ifdef WIN32
	_stat (AL_TREE_FILE_NAME, &status);
#else
	stat (AL_TREE_FILE_NAME, &status);
#endif WIN32
	if (treeFileTime == status.st_mtime)
	    return;
	treeFileTime = status.st_mtime;
	
	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Reconfiguring alarm tree -- started.\n", myPid);

	/* Tell alarmVus to vacate the shared memory. */

	oustAlarmVus (UNAVAILABLE);

	/* Initialize some things. */

	allocateStringSpace (sharedBase + sizeof (ShareHeader), AL_STRINGSIZE);
	initializeTree ();
	initializePoints ();

	/* Initialize channel list */

	clTerminate ();
	clInitialize ();

	/* Go to the beginning of the point state file and pvMap */

	closePointStateFile ();
	openPointStateFile ();

	closePvMapFile ();
	openPvMapFile ();

	/* Load the alarm tree and database points */

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Loading alarm tree\n", myPid);

	if (!loadTree ()) {
	    tisSysLog (0, 2, "alarmMgr(%d): Alarm group tree has not been configured.\n", myPid);
	    clTerminate ();
	    ca_task_exit ();
	    freeSharedMemory ();
	    closePvMapFile ();
	    closePdbXrefFile ();
	    closeLogFile ();
	    exit (-1);
	}

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Loading database points\n", myPid);

	shareHeader->summaryHead = NULL;
	shareHeader->summaryTail = NULL;

	loadPoints ();

	/* Tally the group counts */

	tallyCounts ();

	/* Allow alarmVus back into shared memory */

	allowAlarmVus (AVAILABLE);

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Reconfiguration of alarm tree complete.\n", myPid);

dumpSummaryList ();

    }

/*----------*/
void stringCopy ( char *dest, char *src, int n)

    /* String copy with length protection and NULL termination. */

    {

	strncpy (dest, src, n - 1);
	*(dest + n - 1) = '\0';

    }

/*----------*/
/* Dump bytes in hex to syslog */
void logBytes ( char *addr, int size)

    {

	int	i;
	char	buf [100], buf2 [5];

	strcpy (buf, "");
	for (i=0 ; i<size ; i++) {
	    sprintf (buf2, "%2x ", addr [i] & 0xff);
	    strcat (buf, buf2);
	}
	tisSysLog (0, 2, "alarmMgr(%d): %s", myPid, buf);

    }

