/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "msgMon.h"
#include <tsDefs.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#ifdef WIN32
#include "NTUnixSubs.h"
#endif WIN32

#include "alarmDefs.h"
#include "shareMem.h"
#include "groupInfo.h"
#include "dbase.h"
#include "resolveLinks.h"

/* Functions defined in this module */

#if defined(__STDC__) || defined(WIN32)
#define	VOID	void
static	void	readArchiveConfig (void);
	void	startArchiving (int);
	void	archiveAndClean (void);
static	void	makeLogKey (char *, long, long, int, char *);
static	void	copyFields (void);
#else
#define	VOID
static	void	readArchiveConfig (void);
	void	startArchiving (int);
	void	archiveAndClean (void);
static	void	makeLogKey (char *, long, long, int, char *);
static	void	copyFields (void);
#endif

/* Archiving and cleaning information */

char		archiveFileName [100];			/* Name of the file to place archive data into */
int		archiveOlderThan;			/* Archive records older than this many days */
int		cleanOlderThan;				/* Clean records older than this many days */
int		runAtHour;				/* Hour to start archive process */
int		archiveEnabled, cleanEnabled;
time_t		archiveTime, cleanTime, timeNow;
long		numRecords, curRecord;
char		timeStr [50];

/* Various data items */

int		debugLevel;
extern	C4CODE	codeBase;
extern	ShareHeader	*shareHeader;
extern	int	myPid;

/* Field descriptors */

extern	D4DATA	*logData;
extern	F4FIELD	*logIsAckFld, *logDateFld, *logTimeFld, *logAlarmSecsFld, *logAlarmMsecsFld, *logGroupNameFld;
extern	F4FIELD	*logPointNameFld, *logDescriptorFld, *logStatusFld, *logSeverityFld;
extern	F4FIELD	*logAckedFld, *logOperatorFld, *logResourceFld, *logStationNameFld, *logEguTagFld;
extern	F4FIELD	*logValueFld, *logAckSecsFld, *logAckMsecsFld, *logCommentFld, *logNoPrintFld;

D4DATA	*archiveData;
F4FIELD	*archiveIsAckFld, *archiveDateFld, *archiveTimeFld, *archiveAlarmSecsFld, *archiveAlarmMsecsFld, *archiveGroupNameFld;
F4FIELD	*archivePointNameFld, *archiveDescriptorFld, *archiveStatusFld, *archiveSeverityFld;
F4FIELD	*archiveAckedFld, *archiveOperatorFld, *archiveResourceFld, *archiveStationNameFld, *archiveEguTagFld;
F4FIELD	*archiveValueFld, *archiveAckSecsFld, *archiveAckMsecsFld, *archiveCommentFld, *archiveNoPrintFld;

extern	F4FIELD_INFO	alarmLogFields [];
extern	T4TAG_INFO	alarmLogTags [];

#ifdef WIN32
#define	ARCHIVE_INFO_FILE	"\\tis4000db\\archiveInfo"
#else
#define	ARCHIVE_INFO_FILE	"/tis4000db/archiveInfo"
#endif WIN32
#define	SECS_PER_DAY		86400

/*----------*/
static void readArchiveConfig (VOID)

    {

	FILE		*f;
	char		str [100];
	int		i;

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Reading archive configuration file for alarm log.\n", myPid);

	f = fopen (ARCHIVE_INFO_FILE, "r");
	if (f == NULL) {
	    tisSysLog (0, 2, "alarmMgr(%d): Unable to find %s -- No archiving or cleaning will be performed.\n",
				myPid, ARCHIVE_INFO_FILE);
	    archiveEnabled = 0;
	    cleanEnabled = 0;
	    return;
	}

	/* Skip over the first six entries to get to the alarm log archiving parameters */

	for (i=0 ; i<6 ; i++)
	    fgets (archiveFileName, 100, f);

	/* Get the archiving parameters */

	fgets (archiveFileName, 100, f);
	if (strlen (archiveFileName))
	    archiveFileName [strlen (archiveFileName) - 1] = '\0';

	fgets (str, 100, f);
	archiveOlderThan = atoi (str);

	fgets (str, 100, f);
	cleanOlderThan = atoi (str);

	fgets (str, 100, f);
	runAtHour = atoi (str);

	fgets (str, 100, f);
	archiveEnabled = (strncasecmp (str, "enable", 6) == 0);

	fgets (str, 100, f);
	cleanEnabled = (strncasecmp (str, "enable", 6) == 0);

	fclose (f);

    }

/*----------*/
void startArchiving (currentHour)

    int		currentHour;

    {
	char	archPath [200], pathName [200];
	time_t	clock;

	/* Read the archive configuration file */

	readArchiveConfig ();

	/* If both archiving and cleaning are disabled then we can leave now. */

	if (!archiveEnabled && !cleanEnabled)
	    return;

	/* See if this is the hour to start archiving */

	if (currentHour != runAtHour)
	    return;

	if (debugLevel >= 1) {
	    clock = time (0);
	    tisSysLog (0, 2, "alarmMgr(%d): Archiver activated at %s.\n", myPid,
			strcpy (timeStr, ctime (&clock)));
	}

	/* Check if the archive file exists. If it does not then create it. */

	if (archiveEnabled) {
	    sprintf (archPath, "/tisarch/%s", archiveFileName);
	    resolveLinks (archPath, pathName);
	    archiveData = d4open (&codeBase, pathName);
	    if (archiveData == NULL) {
		archiveData = d4create (&codeBase, pathName, alarmLogFields, alarmLogTags);
		if (archiveData == NULL) {
		    tisSysLog (0, 2, "alarmMgr(%d): Unable to create archive file %s\n",
				myPid, archPath);
		    return;
		}
	    }
	    archiveIsAckFld = d4field (archiveData, "ISACK");
	    archiveDateFld = d4field (archiveData, "DATE");
	    archiveTimeFld = d4field (archiveData, "TIME");
	    archiveAlarmSecsFld = d4field (archiveData, "ALARMSECS");
	    archiveAlarmMsecsFld = d4field (archiveData, "ALARMMSECS");
	    archiveGroupNameFld = d4field (archiveData, "GROUPNAME");
	    archivePointNameFld = d4field (archiveData, "POINTNAME");
	    archiveDescriptorFld = d4field (archiveData, "DESCRIPTOR");
	    archiveStatusFld = d4field (archiveData, "STATUS");
	    archiveSeverityFld = d4field (archiveData, "SEVERITY");
	    archiveAckedFld = d4field (archiveData, "ACKED");
	    archiveNoPrintFld = d4field (archiveData, "NOPRINT");
	    archiveOperatorFld = d4field (archiveData, "OPERATOR");
	    archiveResourceFld = d4field (archiveData, "RESOURCE");
	    archiveStationNameFld = d4field (archiveData, "STATION");
	    archiveEguTagFld = d4field (archiveData, "EGUTAG");
	    archiveValueFld = d4field (archiveData, "VALUE");
	    archiveAckSecsFld = d4field (archiveData, "ACKSECS");
	    archiveAckMsecsFld = d4field (archiveData, "ACKMSECS");
	    archiveCommentFld = d4field (archiveData, "COMMENT");
	}

	/* Get the current time so that we can compare to it */

	timeNow = time (0);

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Time now is %s\n",
			myPid, strcpy (timeStr, ctime (&timeNow)));

	/* Get the archive and clean threshold times */

	archiveTime = timeNow - (archiveOlderThan * SECS_PER_DAY);
	if ((debugLevel >= 1) && archiveEnabled) {
	    strcpy (timeStr, ctime (&archiveTime));
	    timeStr [19] = '\0';
	    tisSysLog (0, 2, "alarmMgr(%d): Records older than %s will be archived.\n",
				myPid, timeStr);
	}

	cleanTime = timeNow - (cleanOlderThan * SECS_PER_DAY);
	if ((debugLevel >= 1) && cleanEnabled) {
	    strcpy (timeStr, ctime (&cleanTime));
	    timeStr [19] = '\0';
	    tisSysLog (0, 2, "alarmMgr(%d): Records older than %s will be cleaned.\n",
				myPid, timeStr);
	}

	/* Get number of records to search and set at beginning of file */

	numRecords = d4reccount (logData);
	curRecord = 1;
	shareHeader->archivingActive = 1;

    }

/*----------*/
void archiveAndClean (VOID)

    {
	
	int		stat, isAck;
	char		key [100];
	long		secs, msecs;
	time_t		clock;
	char		pointName [50];

	if (!shareHeader->archivingActive)
	    return;

	/* See if archiving has completed */

	if (curRecord > numRecords) {
	    if (archiveEnabled)
		d4close (archiveData);
	    if (debugLevel >= 1) {
		clock = time (0);
		tisSysLog (0, 2, "alarmMgr(%d): Archiver finished at %s.\n", myPid,
					strcpy (timeStr, ctime (&clock)));
	    }
	    shareHeader->archivingActive = 0;
	    return;
	}

	/* Get to the next record to be considered */

	for (;;) {
	    stat = d4go (logData, curRecord);
	    d4unlock_all (logData);
	    if (stat == 0)
		break;
	    if (stat == r4locked) {
		printf ("alarmLog record %ld locked -- waiting.\n", curRecord);
#ifdef WIN32
		Sleep (1000);
#else
		sleep (1);
#endif WIN32
	    } else if (stat != 0) {
		closeLogFile ();
		if (archiveEnabled)
		    d4close (archiveData);
		printf ("Error during d4go -- archiving terminated\n");
		shareHeader->archivingActive = 0;
		return;
	    }
	}
	curRecord++;

	/* Skip over deleted entries. */

	strcpy (pointName, f4str (logPointNameFld));
	if (d4deleted (logData) || (pointName [0] == ' '))
	    return;

	/* Create a key for this record */

	strcpy (pointName, f4str (logPointNameFld));
	secs = f4long (logAlarmSecsFld);
	msecs = f4long (logAlarmMsecsFld);
	isAck = f4char (logIsAckFld);
	makeLogKey (pointName, secs, msecs, isAck, key);

	/* Check to see if it is time to archive this record */
	/* We archive records that are complete as of the archive cutoff times. */

	if (archiveEnabled) {
	    if (secs < archiveTime) {
		if (d4seek (archiveData, key) == 0) {	/* Overwrite existing record */
		    copyFields ();
		} else {	/* Create new record */
		    d4append_start (archiveData, 0);
		    copyFields ();
		    d4append (archiveData);
		}
		d4unlock_all (archiveData);
		d4flush_all (archiveData);
	    }
	}

	/* Check to see if it is time to clean this record */

	if (cleanEnabled) {
	    if (secs < cleanTime) {
		d4blank (logData);		/* Blank the record so that it will float to the top. */
		d4delete (logData);		/* Mark for deletion */
		updateDBaseFile (logData);	/* Force an update to place blanked record in the index */
	    }
	}

	d4unlock_all (logData);

    }

/*----------*/
static void makeLogKey (pointName, secs, msecs, isAck, key)	/* Make a key for the log file */

    char	*pointName, *key;
    long	secs, msecs;
    int		isAck;

    {

	sprintf (key, "%*ld%*d%-*s%c",
		AL_GMTSECS_SIZE, secs,
		AL_GMTMSECS_SIZE, msecs,
 		AL_POINTNAME_SIZE, pointName,
		(char)isAck);

    }

/*----------*/
static void copyFields (VOID)

    {

	f4assign (archiveIsAckFld, f4str (logIsAckFld));
	f4assign (archiveDateFld, f4str (logDateFld));
	f4assign (archiveTimeFld, f4str (logTimeFld));
	f4assign (archiveAlarmSecsFld, f4str (logAlarmSecsFld));
	f4assign (archiveAlarmMsecsFld, f4str (logAlarmMsecsFld));
	f4assign (archiveGroupNameFld, f4str (logGroupNameFld));
	f4assign (archivePointNameFld, f4str (logPointNameFld));
	f4assign (archiveDescriptorFld, f4str (logDescriptorFld));
	f4assign (archiveStatusFld, f4str (logStatusFld));
	f4assign (archiveSeverityFld, f4str (logSeverityFld));
	f4assign (archiveAckedFld, f4str (logAckedFld));
	f4assign (archiveNoPrintFld, f4str (logNoPrintFld));
	f4assign (archiveOperatorFld, f4str (logOperatorFld));
	f4assign (archiveResourceFld, f4str (logResourceFld));
	f4assign (archiveStationNameFld, f4str (logStationNameFld));
	f4assign (archiveEguTagFld, f4str (logEguTagFld));
	f4assign (archiveValueFld, f4str (logValueFld));
	f4assign (archiveAckSecsFld, f4str (logAckSecsFld));
	f4assign (archiveAckMsecsFld, f4str (logAckMsecsFld));
	m4assign (archiveCommentFld, m4str (logCommentFld));

    }


