/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#ifdef WIN32
#include <process.h>
#include "NTUnixSubs.h"
#else
#include <unistd.h>
#endif WIN32
#if defined(SOLARIS) || defined(WIN32)
#include <string.h>
#else
#include <strings.h>
#endif

#include "msgMon.h"
#include "alarmDefs.h"

#if defined(__STDC__) || defined(WIN32)
void	launchOpTrkArchiver (int);
#else
void	launchOpTrkArchiver ();
#endif

extern	int	myPid;
extern	int	debugLevel;

#ifdef WIN32
#define	ARCHIVE_INFO_FILE	"\\tis4000db\\archiveInfo"
#else
#define	ARCHIVE_INFO_FILE	"/tis4000db/archiveInfo"
#endif WIN32

/*----------*/
void launchOpTrkArchiver (currentHour)
     int currentHour;

    {

	FILE		*f;
	char		str [100], archiveFileName [100];
	int		i, archiveEnabled, cleanEnabled;
	int		archiveOlderThan, cleanOlderThan, runAtHour;
#ifndef WIN32
	int		pid, result;
#endif WIN32

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Reading archive configuration file for operator tracking archiving.\n", myPid);

	f = fopen (ARCHIVE_INFO_FILE, "r");
	if (f == NULL) {
	    tisSysLog (0, 2, "alarmMgr(%d): Unable to find %s -- No operator tracking archiving or cleaning will be performed.\n",
				myPid, ARCHIVE_INFO_FILE);
	    archiveEnabled = 0;
	    cleanEnabled = 0;
	    return;
	}

	/* Skip over the first twelve entries to get to the operating tracking archiving parameters */

	for (i=0 ; i<12 ; i++)
	    fgets (archiveFileName, 100, f);

	/* Get the archiving parameters */

	fgets (archiveFileName, 100, f);
	if (strlen (archiveFileName))
	    archiveFileName [strlen (archiveFileName) - 1] = '\0';

	fgets (str, 100, f);
	archiveOlderThan = atoi (str);

	fgets (str, 100, f);
	cleanOlderThan = atoi (str);

	fgets (str, 100, f);
	runAtHour = atoi (str);

	fgets (str, 100, f);
	archiveEnabled = (strncasecmp (str, "enable", 6) == 0);

	fgets (str, 100, f);
	cleanEnabled = (strncasecmp (str, "enable", 6) == 0);

	fclose (f);

	/* If this is the hour to run archiving and cleaning then launch the program. */

	if (currentHour == runAtHour) {
#ifdef WIN32
	    char arEnab [5], archOlder [5], cleanEnab [5], cleanOlder [5];

	    sprintf (arEnab, "%d", archiveEnabled);
	    sprintf (archOlder, "%d", archiveOlderThan);
	    sprintf (cleanEnab, "%d", cleanEnabled);
	    sprintf (cleanOlder, "%d", cleanOlderThan);
	    _spawnl (_P_DETACH, "\\tis4000\\tasks\\CleanAndArchTrak", "cleanAndArchTrak",
			    arEnab, archOlder, archiveFileName, cleanEnab, cleanOlder, NULL);
#else
	    if ((pid = fork ()) == 0) {
		char arEnab [5], archOlder [5], cleanEnab [5], cleanOlder [5];
		sprintf (arEnab, "%d", archiveEnabled);
		sprintf (archOlder, "%d", archiveOlderThan);
		sprintf (cleanEnab, "%d", cleanEnabled);
		sprintf (cleanOlder, "%d", cleanOlderThan);
		result = execl ("/tis4000/tasks/CleanAndArchTrak", "cleanAndArchTrak",
				arEnab, archOlder, archiveFileName, cleanEnab, cleanOlder, NULL);
		tisSysLog (0, 2, "alarmMgr(%d): Unable to exec CleanAndArchTrak - no operator tracking and cleaning performed.",
			myPid);
		exit (-1);
	    } else if (pid == -1) {
		tisSysLog (0, 2, "alarmMgr(%d): Unable to fork CleanAndArchTrak - no operator tracking and cleaning performed.",
			myPid);
	    }
#endif WIN32
	}
    }

