/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* Alarm Manager - alarm processor */

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <signal.h>
#include <string.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "msgMon.h"
#include "alarmDefs.h"
#include "alarmMgr.h"
#include "shareMem.h"
#include "groupInfo.h"

#include "cadef.h"
#include "dbDefs.h"
#include "db_access.h"
#include "paramStructs.h"
#include "tsDefs.h"
#include "alarm.h"
#include "tisAlarm.h"
#include "../common/dbase.h"
#include "alarmq.h"

#include "globals.h"
#include "shareMem.h"
#include "sharedMem.h"
#include "tree.h"
#include "points.h"
#include "print.h"
#include "messages.h"
#include "dbase.h"

/* Functions defined in this module */

#define	VOID	void
	void	processPendingAlarms (void);
static	int	processAlarm (tisAlarm *);
static	void	buildLogEntry (LogEntry *, PvData *, tisAlarm *);
	void	processAck (AlarmMsg *, int);
static	void	buildLogAckEntry (LogEntry *, PvData *, AlarmMsg *);
	void	processAckGroup (AlarmMsg *, int);
static	void	buildLogGrpAckEntry (LogEntry *, AlarmMsg *);
	void	processEnable (AlarmMsg *);
	void	processDisable (AlarmMsg *);
	void	processMemo (AlarmMemo *, int);
#ifdef	USEAUTOSCRIPT
static	void	launchAutoScript (PvData *);
#endif	USEAUTOSCRIPT

/* Data */

extern	AlarmQueue	alarmQueue;
extern	char		*alarmStatusString [];
extern	int		alarmsProcessed;
extern	int		myPid;
extern  void		dumpSummaryList (VOID);
	int		alarmProcessed;
		
/*----------*/
void processPendingAlarms (VOID)

    {

	tisAlarm	alarm;

	alarmProcessed = 0;
	alarmsProcessed = 0;
	while (aqRemoveHead (&alarmQueue, &alarm)) {
	    processAlarm (&alarm);
	    ca_pend_io ((float)0.0);
	    alarmsProcessed++;
	}

	/* Tally up all the group counts and tell alarmVu. */

	if (alarmsProcessed > 0) {
	    tallyCounts ();
	    sendSimpleMessageToAlarmvu (alarmProcessed ? UPDATE_NEW_ALARM : UPDATE);
	}

    }

/*----------*/
static int processAlarm (alarm)

    tisAlarm *alarm;

    {

	GroupName	groupName;
	Descriptor	descriptor;
	LogEntry	logEntry;
	Resource        resource;
	PvData		*p;
	int		found, firstReport, writeLog;
	char		*c;

	/* Say something if appropriate */

	if (debugLevel >= 2)
	    printf ("alarmMgr(%d): Msg received for %s, stat = %d, sevr = %d, ack = %d, value = %s, ser=%d, ts=%ld,%ld\n",
				myPid,
				alarm->pvName,
				(int)alarm->alarmStat,
				(int)alarm->alarmSevr,
				(int)alarm->alarmAck,
				alarm->alarmValue,
				htonl(alarm->alarmSerial),
				alarm->alarmTime.secPastEpoch, alarm->alarmTime.nsec);

	/* Find the point for this alarm in the tree.  If it does not exist in the tree insert it. */
	/* If it does exist in the tree, check to see if it is in the right group.  If it is not, */
	/* move the point to the right group. */

	p = findPointInList (alarm->pvName);
	found = getPv (alarm->pvName, groupName, descriptor, resource);

	/* If the point was not found in pvmap then it could be a DPA point. */
	/* If the point contains a colon, use the string left of the colon as */
	/* the group. */

	if (!found) {
	    strcpy (groupName, alarm->pvName);
	    if (c = strchr (groupName, ':')) {
		*c = '\0';
		strcpy (descriptor, "");
		found = 1;
	    }
	}

	/* If point not assigned to a group, then considered it not found */

	if (found) {
	    if (strlen (groupName) == 0) {
		ca_recv_alarm (alarm->alarmSerial, alarm->stateLess, alarm->pvName, &alarm->alarmTime);
		if (debugLevel >= 1)
		    tisSysLog (0, 2, "alarmMgr(%d): Point %s not assigned to a group -- recv sent to CA.\n",
				myPid, alarm->pvName);
		return 0;
	    }
	}

	if (p && found) {	/* In a group and in pv_map */
	    if (strcmp (p->parentGroup->groupName, groupName) != 0) {  /* In the wrong group */
		TreeNode *t;
		removePoint (p);
		stringCopy (p->description, descriptor, sizeof (Descriptor));
		t = findNode (groupName);
		if (t == NULL) { 	/* Point moved from an existing group to a non-existant group. */
		    ca_recv_alarm (alarm->alarmSerial, alarm->stateLess, alarm->pvName, &alarm->alarmTime);
		    if (debugLevel >= 1)
			tisSysLog (0, 2, "alarmMgr(%d): Point %s moved from existing group to non-existant group -- recv sent to CA.\n",
					myPid, alarm->pvName);
			return 0;
		}
		p->parentGroup = t;
		addPoint (t, p);
		if (debugLevel >= 1)
		    tisSysLog (0, 2, "alarmMgr(%d): %s moved to group %s\n",
				myPid, alarm->pvName, groupName);
	    }
	} else if (p) {		/* In a group but not in pv_map */
	    deletePointState (p->pvName);
	    if (inSummaryList (p))
		removePointFromSummary (p);
	    removePoint (p);
	    if (debugLevel >= 1)
		tisSysLog (0, 2, "alarmMgr(%d): %s removed\n", myPid, alarm->pvName);
	    return 1;		
	} else if (found) {		/* Not in the tree but in pv_map */
	    p = loadPoint (alarm->pvName, groupName, descriptor, resource);
	    if (p == NULL)
		return 0;		/* Couldn't load for some reason */
	    else
		if (debugLevel >= 1)
		    tisSysLog (0, 2, "alarmMgr(%d): %s added to group %s\n",
				myPid, alarm->pvName, groupName);
	} else {
	    return 0;
	}

	/* Save the time of the last alarm heard from CA */

	firstReport = p->lastAlarmReceived.secPastEpoch == 0;
	if (TsCmpStampsGT (&(alarm->alarmTime), &(p->lastAlarmReceived))) {	/* Changed from GT to GE 8/17/95 dpd */
	    assignTS (p->lastAlarmReceived, alarm->alarmTime);
	} else
	    if (debugLevel >= 2)
		tisSysLog (0, 2, "alarmMgr(%d): Alarm with old timestamp detected -- %s",
				myPid, alarm->pvName);

	/* These are saved for all reports from CA */
 
	p->aenb = alarm->alarmEnabled;
	p->oscn = alarm->alarmOnscan;
	stringCopy (p->description, alarm->pvDesc, sizeof (Descriptor));
	stringCopy (p->eguTag, alarm->pvEgu, sizeof (EguTag));
	
	/* Check to see if the previous severity for this point is VALID. If it is and the */
	/* current severity is equal to the severity just prior to the VALID severity, ack */
	/* the point to the IOC and discard it. */

	if (p->curRawSevr == VALID_ALARM) {
	    if (p->prevRawSevr == alarm->alarmSevr) {
		ca_recv_alarm (alarm->alarmSerial, alarm->stateLess, alarm->pvName, &alarm->alarmTime);
		p->prevRawSevr = p->curRawSevr;
		p->curRawSevr = alarm->alarmSevr;
		p->status = alarm->alarmStat;
		p->severity = alarm->alarmSevr;
		p->Ack = !alarm->alarmAck;
		if (p->Ack && p->RTN && inSummaryList (p))
		    removePointFromSummary (p);
		writePointState (p);
		return 1;
	    }
	}
	p->prevRawSevr = p->curRawSevr;
	p->curRawSevr = alarm->alarmSevr;

	/* Valid alarms are either converted to major or dropped on the floor */

	if (alarm->alarmSevr == VALID_ALARM) {
	    if (alarm->alarmStat == SIGNAL_ALARM)
		alarm->alarmSevr = MAJOR_ALARM;
	    else {
		ca_recv_alarm (alarm->alarmSerial, alarm->stateLess, alarm->pvName, &alarm->alarmTime);
		p->status = alarm->alarmStat;
		p->severity = alarm->alarmSevr;
		p->Ack = !alarm->alarmAck;
		if (p->Ack && p->RTN && inSummaryList (p))
		    removePointFromSummary (p);
		writePointState (p);
		return 1;
	    }
	}


	/* Determine the type of message from CA and process as either */
	/* a new alarm, an RTN or an ACK from CA. */

	writeLog = 0;

	if (((alarm->alarmSevr != NO_ALARM) &&
	     (alarm->alarmSevr != EVENT_ALARM)) &&
	    ((alarm->alarmStat != p->unAckedStatus) ||
	     (alarm->alarmSevr != p->unAckedSeverity) ||
	     p->RTN || ((alarm->alarmStat == STATE_ALARM) &&
			(strncmp(p->value, alarm->alarmValue, sizeof (PvValue)) != 0)))) {  /* Alarm message */

	    TS_STAMP chainTime;

	    if (inSummaryList (p))
		removePointFromSummary (p);

	    if (!p->Ack && (p->unAckedSeverity != NO_ALARM)) 
		p->lastAck = chainTime = p->lastAlarm;
	    else
		chainTime.secPastEpoch = 0;

	    p->RTN = 0;
	    p->status = p->unAckedStatus = alarm->alarmStat;
	    p->severity = p->unAckedSeverity = alarm->alarmSevr;
	    stringCopy (p->value, alarm->alarmValue, sizeof (PvValue));
	    stringCopy (p->unAckedValue, alarm->alarmValue, sizeof (PvValue));
	    strcpy (p->op, "");
	    if (alarm->alarmAck == 0) {	/* Does not need ACK */
		if ((p->lastAck.secPastEpoch != 0)) {
		    TS_STAMP keyTime;
		    keyTime = p->lastAlarm;
		    while (keyTime.secPastEpoch != 0)
			setAckInLog (p->pvName, &keyTime, &logEntry.ackTime);	
		    p->lastAck.secPastEpoch = 0;
		}
	    }
	    p->Ack = !alarm->alarmAck;
	    buildLogEntry (&logEntry, p, alarm);
	    logEntry.ackTime = chainTime;	/* Link onto "+" list */
	    assignTS (p->lastAlarm, alarm->alarmTime);
	    writeLog = 1;
	    addPointToSummary (p);
	    alarmProcessed = 1;
	} else if (((alarm->alarmSevr == NO_ALARM) || (alarm->alarmSevr == EVENT_ALARM)) &&
			(p->unAckedSeverity != NO_ALARM) &&
			(p->severity != NO_ALARM)) {	/* RTN message */
	    p->RTN = 1;
	    p->status = alarm->alarmStat;
	    p->severity = alarm->alarmSevr;
	    stringCopy (p->value, alarm->alarmValue, sizeof (PvValue));
	    buildLogEntry (&logEntry, p, alarm);
	    logEntry.ackTime = p->lastAlarm;	/* Link onto "+" list */
	    assignTS (p->lastAlarm, alarm->alarmTime);
	    if (p->Ack || (alarm->alarmAck == 0)) {
		if (inSummaryList (p))
		    removePointFromSummary (p);
		p->unAckedStatus = alarm->alarmStat;
		p->unAckedSeverity = alarm->alarmSevr;
		stringCopy (p->unAckedValue, alarm->alarmValue, sizeof (PvValue));
		p->RTN = 0;
		if (!p->Ack) {
		    if ((p->lastAck.secPastEpoch != 0)) {
			TS_STAMP keyTime;
			keyTime = p->lastAlarm;
			while (keyTime.secPastEpoch != 0)
			    setAckInLog (p->pvName, &keyTime, &logEntry.ackTime);	
			p->lastAck.secPastEpoch = 0;
		    }
		}
		p->Ack = 0;
	    }
	    writeLog = 1;
	} else if ((alarm->alarmAck == 0) && !p->Ack) {		/* ACK message */
	    p->Ack = 1;
	    /* set lastAck time to the most recent ack */
	    tsLocalTime (&logEntry.ackTime);
	    if ((p->lastAck.secPastEpoch != 0)) {
		TS_STAMP keyTime;
		keyTime = p->lastAlarm;
		while (keyTime.secPastEpoch != 0)
		    setAckInLog (p->pvName, &keyTime, &logEntry.ackTime);
		p->lastAck.secPastEpoch = 0;
	    }
	    p->lastAck = logEntry.ackTime;
	    if (p->RTN) {
		if (inSummaryList (p))
		    removePointFromSummary (p);
		p->unAckedStatus = p->status;
		p->unAckedSeverity = p->severity;
		stringCopy (p->unAckedValue, p->value, sizeof (PvValue));
	    }
	} else if (((alarm->alarmSevr != p->severity) || (alarm->alarmStat != p->status)) &&
			TsCmpStampsNE (&alarm->alarmTime, &p->lastAlarm)) {
	    TS_STAMP chainTime;
	    if (p->lastAck.secPastEpoch != 0)
		p->lastAck = chainTime = p->lastAlarm;
	    else
		chainTime.secPastEpoch = 0;
	    p->status = alarm->alarmStat;
	    p->severity = alarm->alarmSevr;
	    p->Ack = !alarm->alarmAck;
	    stringCopy (p->value, alarm->alarmValue, sizeof (PvValue));
	    buildLogEntry (&logEntry, p, alarm);
	    logEntry.ackTime = chainTime;	/* Link onto "+" list */
	    assignTS (p->lastAlarm, alarm->alarmTime);
	    writeLog = 1;
	} else {
	    if ((debugLevel >= 2) && !firstReport)
		tisSysLog (0, 2, "alarmMgr(%d): Alarm message not processed! %s, stat=%d, sevr=%d, ts=%ld,%ld, ack=%d, ser=%d",
			myPid, alarm->pvName, (int)alarm->alarmStat, (int)alarm->alarmSevr,
	    		alarm->alarmTime.secPastEpoch, alarm->alarmTime.nsec,
			(int)alarm->alarmAck, (int)alarm->alarmSerial);
	    p->Ack = !alarm->alarmAck;
	}

	/* Make sure we remove entry from summary if Ack and RTN */
 
	if (p->Ack && p->RTN && inSummaryList (p))
	    removePointFromSummary (p);

	/* Write the alarm to the log file and tell CA */

	if (writeLog) {
	    logEntry.noPrint = !printAlarm (p, &logEntry);
	    writeLogEntry (&logEntry);
#ifdef	USEAUTOSCRIPT
	    launchAutoScript (p);
#endif	USEAUTOSCRIPT
	}

	/* Acknowledge the receipt of the alarm message to CA */

	ca_recv_alarm (alarm->alarmSerial, alarm->stateLess, alarm->pvName, &alarm->alarmTime);

	/* Update the point state table */

	writePointState (p);

dumpSummaryList ();

	return 1;

    }

/*----------*/
static void buildLogEntry (logEntry, p, alarm)

    LogEntry *logEntry;
    PvData *p;
    tisAlarm *alarm;

    {
	char *violated;

	memset ((char *)logEntry, 0, sizeof (LogEntry));
	logEntry->isAck = 0;	/* An alarm entry */
	assignTS (logEntry->logTime, alarm->alarmTime);
	logEntry->ackTime.secPastEpoch = 0;
	stringCopy (logEntry->status, alarmStatusString [alarm->alarmStat], sizeof (alStatus));
	logEntry->severity = (char)alarm->alarmSevr;
	logEntry->acked = p->Ack;
	stringCopy (logEntry->pointName, alarm->pvName, sizeof (PointName));
	stringCopy (logEntry->descriptor, alarm->pvDesc, sizeof (Descriptor));
	stringCopy (logEntry->eguTag, alarm->pvEgu, sizeof (EguTag));
	stringCopy (logEntry->value, alarm->alarmValue, sizeof (PvValue));
	stringCopy (logEntry->groupName, p->parentGroup->groupName, sizeof (GroupName));
	stringCopy (logEntry->stationName, p->stationName, sizeof (IOCName));
	stringCopy (logEntry->resource, p->resource, sizeof (Resource));
#if defined(INFOFIELDS) || defined(WIN32)
	/* store the violated limit in the info1 field of the database */
	violated = GetLimitValue (alarm->pvName, alarm->alarmStat);
	if (violated)
	    stringCopy (logEntry->info1, violated, AL_INFO_SIZE);
#endif INFOFIELDS
    }

/*----------*/
void processAck (msg, ackSource)
     AlarmMsg *msg;
     int ackSource;

    {

	PvData		*p;
	LogEntry	logEntry;
	GroupName	groupName;
	Descriptor	descriptor;
        Resource        resource;
	TS_STAMP	keyTime;
	int		found;
	PointName	pointName;
	char		*c;

	strcpy (pointName, msg->pointName);

	found = getPv (pointName, groupName, descriptor, resource);

	/* If the point was not found in pvmap then it could be a DPA point. */
	/* If the point contains a colon, use the string left of the colon as */
	/* the group and the string to the right of the colon as the point name. */

	if (!found) {
	    strcpy (groupName, pointName);
	    if (c = strchr (groupName, ':')) {
		*c = '\0';
		/*
		  comment out following line. Point name must be preserved
		  for any search operations.
		  */
		/* strcpy (pointName, c + 1); */	/* All afer the colon is the pointname */
		strcpy (descriptor, "");
		found = 1;
	    }
	}

	/* Look in pv_map to see if this point has been deleted. If it has then remove */
	/* it from the alarm tree and log the ack. */

	if (!found) {

	    p = findPointInList (pointName);
	    if (p == NULL) {	/* This shouldn't happen */
		if (debugLevel >= 2)
		    tisSysLog (0, 2, "alarmMgr(%d): Unexpected ACK for point not in tree -- %s\n",
					myPid, msg->pointName);
	    	return;
	    }
	    buildLogAckEntry (&logEntry, p, msg);
	    deletePointState (p->pvName);
	    if (inSummaryList (p))
		removePointFromSummary (p);
	    removePoint (p);
	    tallyCounts ();
	    sendSimpleMessageToAlarmvu (UPDATE);
	    dumpSummaryList ();
	    return;
	}

	/* See if this point is one we are prepared to deal with. */

	p = findPointInList (pointName);
	if (p == NULL) {	/* This shouldn't happen */
	    if (debugLevel >= 2)
		tisSysLog (0, 2, "alarmMgr(%d): Unexpected ACK for point not in tree -- %s\n",
				myPid, msg->pointName);
	    return;
	}

	/* Say something if appropriate */

	if (debugLevel >= 1)
	    if (ackSource == FROM_ALARMVU)
		tisSysLog (0, 2, "alarmMgr(%d): ACK received from alarmVu for %s, member=%d\n",
			myPid, msg->pointName, (int)(msg->memberOfGroup));
	    else
		tisSysLog (0, 2, "alarmMgr(%d): ACK received from alarmMgr for %s, member=%d\n",
			myPid, msg->pointName, (int)(msg->memberOfGroup));

	/* Send this ACK to my peers */

	if (ackSource == FROM_ALARMVU) {
	    AlarmMsg	peerMsg;
	    peerMsg = *msg;

	    /* Convert to network byte ordering */

	    peerMsg.msgType = htonl (peerMsg.msgType);
	    peerMsg.timeStamp.secPastEpoch = htonl (peerMsg.timeStamp.secPastEpoch);
	    peerMsg.timeStamp.nsec = htonl (peerMsg.timeStamp.nsec);
	    peerMsg.ackTime.secPastEpoch = htonl (peerMsg.ackTime.secPastEpoch);
	    peerMsg.ackTime.nsec = htonl (peerMsg.ackTime.nsec);
	    peerMsg.memberOfGroup = htonl (peerMsg.memberOfGroup);
	    peerMsg.status = htonl (peerMsg.status);
	    peerMsg.severity = htonl (peerMsg.severity);

	    sendMessageToPeers ((char *)&peerMsg, sizeof (AlarmMsg));
	}

	/* Write ack to log */

	buildLogAckEntry (&logEntry, p, msg);

	/* Update the alarm record in the alarm log.  If a chain of alarms exists */
	/* update each of them. */

	logEntry.noPrint = !printAck (p, &logEntry);
	writeLogEntry (&logEntry);
	keyTime = p->lastAlarm;
	while (keyTime.secPastEpoch != 0)
	    setAckInLog (p->pvName, &keyTime, &logEntry.ackTime);	

	/* Fill in operator ack information */	
	/*  We must set op name here because we don't have access to it later! */
	stringCopy (p->op, msg->name.operatorName, sizeof (Operator));

/* should record ack time here, and use the fact that this time points to an ack message */
/* to tell difference between chain and ack time. */
	p->lastAck.secPastEpoch = 0;

	if (ackSource == FROM_ALARMVU)
	    ca_ack_alarm (p->pvName, &(p->lastAlarmReceived));

	/* Update the point state file */

	writePointState (p);

	dumpSummaryList ();

    }

/*----------*/
static void buildLogAckEntry (logEntry, p, msg)
     LogEntry *logEntry;
     PvData *p;
     AlarmMsg *msg;

    {

	memset ((char *)logEntry, 0, sizeof (LogEntry));
	logEntry->isAck = 1;	/* An ack entry */
	logEntry->ackTime = p->lastAlarm;
	logEntry->logTime = msg->ackTime;
	strcpy (logEntry->status, "ACK");
	logEntry->severity = 0;
	stringCopy (logEntry->pointName, p->pvName, sizeof (PointName));
	strcpy (logEntry->value, "");
	stringCopy (logEntry->oper, msg->name.operatorName, sizeof (Operator));
	logEntry->acked = 1;

    }

/*----------*/
void processAckGroup (msg, ackSource)
     AlarmMsg *msg;
     int ackSource;

    {

	LogEntry	logEntry;
	TreeNode	*t;

	/* Say something if appropriate */

	if (debugLevel >= 1)
	    if (ackSource == FROM_ALARMVU)
		tisSysLog (0, 2, "alarmMgr(%d): GRP ACK received from alarmVu for %s, member=%d\n",
			myPid, msg->pointName, (int)(msg->memberOfGroup));
	    else
		tisSysLog (0, 2, "alarmMgr(%d): GRP ACK received from alarmMgr for %s, member=%d\n",
			myPid, msg->pointName, (int)(msg->memberOfGroup));

	/* Send this ACK to my peers */

	if (ackSource == FROM_ALARMVU) {
	    AlarmMsg	peerMsg;
	    peerMsg = *msg;

	    /* Convert to network byte ordering */

	    peerMsg.msgType = htonl (peerMsg.msgType);
	    peerMsg.timeStamp.secPastEpoch = htonl (peerMsg.timeStamp.secPastEpoch);
	    peerMsg.timeStamp.nsec = htonl (peerMsg.timeStamp.nsec);
	    peerMsg.ackTime.secPastEpoch = htonl (peerMsg.ackTime.secPastEpoch);
	    peerMsg.ackTime.nsec = htonl (peerMsg.ackTime.nsec);
	    peerMsg.memberOfGroup = htonl (peerMsg.memberOfGroup);
	    peerMsg.status = htonl (peerMsg.status);
	    peerMsg.severity = htonl (peerMsg.severity);

	    sendMessageToPeers ((char *)&peerMsg, sizeof (AlarmMsg));
	}

	/* Find the group to which this ACK applies */

	t = findNode (msg->pointName);
	if (t == NULL) {
	    if (debugLevel >= 1)
		tisSysLog (0, 2, "alarmMgr(%d): GRP ACK for unknown group %s -- discarded.\n",
			myPid, msg->pointName);
	    return;
	}
	buildLogGrpAckEntry (&logEntry, msg);
	logEntry.noPrint = !printGrpAck (t, &logEntry);
	writeLogEntry (&logEntry);

    }

/*----------*/
static void buildLogGrpAckEntry (logEntry, msg)
     LogEntry *logEntry;
     AlarmMsg *msg;

    {

	memset ((char *)logEntry, 0, sizeof (LogEntry));
	logEntry->isAck = 1;	/* An ack entry */
	logEntry->logTime = msg->ackTime;
	logEntry->ackTime = msg->ackTime;
	strcpy (logEntry->status, "GRP ACK");
	logEntry->severity = 0;
	stringCopy (logEntry->pointName, msg->pointName, sizeof (PointName));
	strcpy (logEntry->value, "");
	stringCopy (logEntry->oper, msg->name.groupName, sizeof (GroupName));
	logEntry->acked = 1;

    }

/*----------*/
void processEnable (msg)
     AlarmMsg *msg;

    {

	ca_enable_alarm (msg->pointName);
	sendSimpleMessageToAlarmvu (UPDATE);
	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Enabling %s\n", myPid, msg->pointName);

    }

/*----------*/
void processDisable (msg)
     AlarmMsg *msg;

    {

	ca_disable_alarm (msg->pointName);
	sendSimpleMessageToAlarmvu (UPDATE);
	if (debugLevel >= 1)
	    tisSysLog (0, 2, "alarmMgr(%d): Disabling %s\n", myPid, msg->pointName);


    }

/*----------*/
void processMemo (msg, memoSource)
     AlarmMemo *msg;
     int memoSource;

    {

	/* Say something if appropriate */

	if (debugLevel >= 1)
	    if (memoSource == FROM_ALARMVU)
		tisSysLog (0, 2, "alarmMgr(%d): Memo received from alarmVu: Point=%s, memo=%s\n", 
				myPid, msg->pointName, msg->memoText);
	    else
		tisSysLog (0, 2, "alarmMgr(%d): Memo received from alarmMgr: Point=%s, memo=%s\n", 		
				myPid, msg->pointName, msg->memoText);

	/* Send this memo to my peers */

	if (memoSource == FROM_ALARMVU) {
	    AlarmMemo	peerMsg;
	    peerMsg = *msg;

	    /* Convert to network byte ordering */

	    peerMsg.msgType = htonl (peerMsg.msgType);
	    peerMsg.timeStamp.secPastEpoch = htonl (peerMsg.timeStamp.secPastEpoch);
	    peerMsg.timeStamp.nsec = htonl (peerMsg.timeStamp.nsec);

	    sendMessageToPeers ((char *)&peerMsg, sizeof (AlarmMemo));
	}

	/* Write the memo to the log file */

	writeLogMemo (msg->pointName, &(msg->timeStamp), msg->memoText);

    }

/*----------*/
#ifdef	USEAUTOSCRIPT
static void launchAutoScript (p)
     PvData *p;

    {

	TreeNode	*group;
	ScriptName	script;
	char		expScript [400];
	int		i, j;
	char		cmd [AL_SCRIPTNAME_SIZE+20];

	/* Get the script from the tree node depending upon the aux PV */

	group = p->parentGroup;
	strcpy (script, "");
	if (group->auxValue) {	/* In aux mode */
	    stringCopy (script, group->groupParms.auxAutoScript, sizeof (ScriptName));
	    if ((*script == '\0') || (strcmp (script, "*") == 0))
		return;
	} else {	/* In normal mode */
	    stringCopy (script, group->groupParms.autoScript, sizeof (ScriptName));
	    if ((*script == '\0') || (strcmp (script, "*") == 0))
		return;
	}

	/* A script specifier can have the form:
	
		myScript %N %T ...
	
	 The flags are replaced by the flag followed by a value
	
		myScript %N groupName %T groupStatus ...
	
	 The following flags are recognized as:
	
		%N	Group Name
		%T	Group Status
		%S	Group Severity
		%n	Point Name
		%t	Point Status
		%s	Point Severity
		%v	Point Value
		%a	Ack FLag (1 => needs acking)
	
	  %% will be translated to a single %
	*/

	i=j=0;
	while (script [i]) {
	    if (script [i] == '%') {	/* Found a flag */
		i++;
		switch (script [i]) {
		    case 'N':		/* Group Name */
			sprintf (&expScript [j], "%s", group->groupName);
			j += strlen (&expScript [j]);
			break;
		    case 'T':		/* Group Status */
			sprintf (&expScript [j], "%d", group->groupData.groupStatus);
			j += strlen (&expScript [j]);
			break;
		    case 'S':		/* Group Severity */
			sprintf (&expScript [j], "%d", group->groupData.groupSeverity);
			j += strlen (&expScript [j]);
			break;
		    case 'n':		/* Point Name */
			sprintf (&expScript [j], "%s", p->pvName);
			j += strlen (&expScript [j]);
			break;
		    case 't':		/* Point Status */
			sprintf (&expScript [j], "%d", p->status);
			j += strlen (&expScript [j]);
			break;
		    case 's':		/* Point Severity */
			sprintf (&expScript [j], "%d", p->severity);
			j += strlen (&expScript [j]);
			break;
		    case 'v':		/* Point Value */
			sprintf (&expScript [j], "\"%s\"", p->value);
			j += strlen (&expScript [j]);
			break;
		    case 'a':		/* Point is ACKed */
			sprintf (&expScript [j], "%d", p->Ack);
			j += strlen (&expScript [j]);
			break;
		    default:
			sprintf (&expScript [j], "%", script [i]);
			j += strlen (&expScript [j]);
			break;
		}
		i++;
	    } else {
		expScript [j++] = script [i++];
	    }
	}
	expScript [j] = '\0';

	sprintf (cmd, "csh -c \"%s &\"", expScript);
	system (cmd);

    }
#endif	USEAUTOSCRIPT

