/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* Shared memory utilities */

#include <stdio.h>
#include <sys/types.h>
#ifdef WIN32
#include "NTUnixSubs.h"
#else
#include <sys/ipc.h>
#include <sys/shm.h>
#endif WIN32
#include <alarmDefs.h>
#include <shareMem.h>
#include "globals.h"

/* For keeping track of the shared segments */

static struct {
    int		memId;
    char	*memBase;
} segList [AL_MAX_SEGMENTS];
int	numSegments;

char	*sharedBase;		/* Base of the shared memory */ 
char	*sharedTop;		/* Top of shared memory */

/*----------*/
int allocateSharedMemory ()

    {

	int	i;

	/* Allocate the initial shared memory segments */

	numSegments = AL_INITIAL_SEGMENTS;
	segList [0].memBase = AL_SHARED_BASE;
	for (i=0 ; i<numSegments ; i++) {
	    segList [i].memId = shmget (AL_SHR_MEM_KEY + i, AL_SEGMENT_SIZE, 01666);
	    if (segList [i].memId == -1)
		return 0;
	    segList [i].memBase = (char *) shmat (segList [i].memId, segList [0].memBase + (i * AL_SEGMENT_SIZE), 0);
	    if (segList [i].memBase == (char *)-1)
		return 0;
	}

	/* Set up pointers to the various segments */

	sharedBase = segList [0].memBase;
	sharedTop = (char *)sharedBase + (AL_SEGMENT_SIZE * numSegments) - 1;

	/* Set up the ShareHeader structure at the beginning of the tree segment. */

	shareHeader = (ShareHeader *)segList [0].memBase;
	shareHeader->shareSignature = AL_SIGNATURE;
	shareHeader->shareSegs = AL_INITIAL_SEGMENTS;
	shareHeader->treeRoot = (TreeNode *)(segList [0].memBase + sizeof (ShareHeader) + AL_STRINGSIZE);
	shareHeader->alarmPoints = (PvData *)(segList [0].memBase + sizeof (ShareHeader) + AL_STRINGSIZE + AL_TREESIZE);
	shareHeader->summaryHead = NULL;
	shareHeader->summaryTail = NULL;

	return 1;

    }

/*----------*/
int freeSharedMemory ()

    {

	int	i;

	for (i=0 ; i<numSegments ; i++) {
	    if (shmdt ((char *)segList [i].memBase) == -1)
		return 0;
	    if (shmctl (segList [i].memId, IPC_RMID, (struct shmid_ds *)0) == -1)
		return 0;
	}

	return 1;

    }

/*----------*/
int extendMemoryIfFull (p)
     char *p;	/* Check if p is within the shared memory. */
				/* If not allocate more shared memory. */

    {

	if (p > sharedTop) {
	    segList [numSegments].memId = shmget (AL_SHR_MEM_KEY + numSegments,
				AL_SEGMENT_SIZE, 01666);
	    if (segList [numSegments].memId == -1)
		return 0;
	    segList [numSegments].memBase = (char *) shmat (segList [numSegments].memId,
				segList [0].memBase + (numSegments * AL_SEGMENT_SIZE), 0);
	    if (segList [numSegments].memBase == (char *)-1)
		return 0;
	    numSegments++;
	    shareHeader->shareSegs++;
	    sharedTop += AL_SEGMENT_SIZE;
	    return 1;
	}
	return 0;
    }

