/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/******************************************************************************
 *                                                                            *
 *      path:  alarmPrinter.c                                                 *
 *      date:  21-Jul-94                                                      *
 *                                                                            *
 ******************************************************************************/
#include <string.h>
#include <stdio.h>
#include <time.h>
#ifdef WIN32
#include <winsock.h>
#include <sys/types.h>
#include <sys/stat.h>
#else
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#ifndef AXP
#include <net/if.h>
#include <sys/sockio.h>
#include <sys/filio.h>
#include <sys/ioctl.h>
#endif AXP
#ifdef MIPSEL
#include "error.h"
#endif MIPSEL
#endif WIN32

#include "tsDefs.h"
#include "alarmDefs.h"
#include "messages.h"
#include "msgMon.h"

#define MAX_PRINT_TRIES	3	/* Max number of attempts to print to network printer */
#define PRINTERPORT 2000	/* default printer port */

/******************************************************************************
 *                                                                            *
 *     external variables                                                     *
 *                                                                            *
 ******************************************************************************/

extern int		errno;

/******************************************************************************
 *                                                                            *
 *     external functions                                                     *
 *                                                                            *
 ******************************************************************************/

extern char		*mktemp ();

/******************************************************************************
 *                                                                            *
 *     static functions                                                       *
 *                                                                            *
 ******************************************************************************/

static int		sendToPrinter ( char * prBuf );
static void		printInit ();
static void		printerError ();

/******************************************************************************
 *                                                                            *
 *     global variables                                                       *
 *                                                                            *
 ******************************************************************************/

short			managerPort;
short			parPrPort;
#ifdef MIPSEL
ERRSYMTAB		errSymTbl;
#endif
#ifdef WIN32
SOCKET			readFd;		/* Fd for incoming print data socket */
SOCKET			printDesc;	/* Fd for outbound print info */
#else
int			readFd;		/* Fd for incoming print data socket */
int			printDesc;	/* Fd for outbound print info */
#endif WIN32
FILE			*printFile;	/* device based access to printer */

/******************************************************************************
 *                                                                            *
 *     static variables                                                       *
 *                                                                            *
 ******************************************************************************/

static struct	hostent	*netPrinter;			/* Buffer for network printer's address */
static int optionValue = 1;
static char printer_name [100];
static struct timeval	zeroDelay;
static struct timeval	printDelay;

/******************************************************************************
 *                                                                            *
 *     TIS 4000 alarm print handler                                           *
 *	    alarmprint <printer_name:[<netprintport>]> <data_port>	      *
 *                                                                            *
 ******************************************************************************/

main (int argc, char **argv)
{
    char		txtBuf[400];
    int			rcvSize;
    struct sockaddr_in	netPort;	/* 6/14/94 DEK - Address of network printer */
    long		hAddr;
    static struct hostent *alarmMgrHost;
    int			numReady;
    fd_set		PrFdSet;
#ifndef	WIN32
    int			i;
#endif	!WIN32

#ifdef WIN32
    WSADATA wsaData; 
    WORD    wVersionRequested = MAKEWORD(1, 1);
    int	    err;

    err = WSAStartup(wVersionRequested, &wsaData); 
 
    if (err != 0) 
	return 0; 
#elif SOLARIS
    {
	struct rlimit rl;
	if (getrlimit (RLIMIT_NOFILE, &rl))
	{
	    /* Just use default */
	    for (i = 19; i >= 3; --i)
		close(i);
	}
	else
	{
	    for (i = rl.rlim_cur - 1; i >= 3; --i)
		close(i);
	}
    }
#else    
    for (i = getdtablesize() - 1; i >= 3; --i)
	close(i);
#endif WIN32

    tisOpenLog ("alarmPrinter", 6, 2, NULL, NULL);

    parPrPort = PRINTERPORT;

    if (argc != 3)
    {
	printf ("usage: alarmprint <printer_name:[nnnn]> <data_port>\n");
	exit (1);
    }
    strcpy (printer_name, argv[1]);
    if (strchr(printer_name, ':'))
    {
	parPrPort = atoi (strchr(printer_name, ':')+1);
    }
    managerPort = atoi (argv[2]);

    memset ((char *)&zeroDelay, 0, sizeof (zeroDelay));
    zeroDelay.tv_sec = 10;	/* Ten-second timeout on select() */
    printDelay.tv_sec = 10;
    printDelay.tv_usec = 0;

#ifndef WIN32
    signal (SIGPIPE, SIG_IGN);
#endif WIN32

    /* Connect with alarm manager */
    if(( readFd = socket( AF_INET, SOCK_STREAM, 0 )) == -1 )
    {
	printerError( "Can't create socket to alarm manager" );
	exit (1);
    }
#ifdef WIN32
    if (ioctlsocket( readFd, FIONBIO, &optionValue) == -1)
#else
    if (ioctl( readFd, FIONBIO, &optionValue) == -1)
#endif WIN32
    {
	printerError( "Can't set manager socket nonblocking" );
	exit (1);
    }
    if(( alarmMgrHost = gethostbyname( "localhost" )) == NULL )
    {
	printerError( "Failure getting local host" );
	exit (1);
    }
    memset( &netPort, 0, sizeof( struct sockaddr_in ));
    netPort.sin_family = AF_INET;
    hAddr = *(long *)alarmMgrHost->h_addr_list[0];
    netPort.sin_addr.s_addr = hAddr;
    netPort.sin_port = htons( managerPort );

    if( connect( readFd, (struct sockaddr *)&netPort, sizeof( struct sockaddr )) == -1 )
    {
#ifdef WIN32
	errno = WSAGetLastError ();
	if( errno != WSAEINPROGRESS )
#else
	if( errno != EINPROGRESS )
#endif WIN32
	{
	    printerError( "Can't connect to alarm manager" );
	    exit (1);
	}
    }
    /* announce ourselves to the manager */
    {
	PrinterNameMsg	msg;
	msg.msgType = htonl (PRINTER_NAME_MSG);
	strcpy (msg.printerName, printer_name);
	if (send (readFd, (char *)&msg, sizeof (msg), 0) != sizeof (msg)) {
#ifdef WIN32
	    errno = WSAGetLastError ();
	    if ((errno != WSAEINPROGRESS) && (errno != WSAEINVAL)) {
#else
	    if ((errno != EINPROGRESS) && (errno != EINVAL)) {
#endif WIN32
		printerError( "Can't send to manager" );
		exit (1);
	    }
	}
    }    

    for (;;)
    {
	memset ( txtBuf, 0, sizeof txtBuf );
	FD_ZERO( &PrFdSet );
	FD_SET( readFd, &PrFdSet );
	numReady = select( readFd+1, &PrFdSet, NULL, NULL, &printDelay );
	if (numReady < 0) {
	    /* select error */
#ifdef WIN32
	    closesocket (readFd);
#else
	    close (readFd);
#endif WIN32
	    if(( readFd = socket( AF_INET, SOCK_STREAM, 0 )) == -1 )
	    {
		printerError( "Can't create socket to alarm manager" );
		exit (1);
	    }
#ifdef WIN32
	    if (ioctlsocket( readFd, FIONBIO, &optionValue) == -1)
#else
	    if (ioctl( readFd, FIONBIO, &optionValue) == -1)
#endif WIN32
	    {
		printerError( "Can't set manager socket nonblocking" );
		exit (1);
	    }
	    if(( netPrinter = gethostbyname( "localhost" )) == NULL )
	    {
		printerError( "Failure getting local host" );
		exit (1);
	    }
	    memset( &netPort, 0, sizeof( struct sockaddr_in ));
	    netPort.sin_family = AF_INET;
	    hAddr = *(long *)netPrinter->h_addr_list[0];
	    netPort.sin_addr.s_addr = hAddr;
	    netPort.sin_port = htons( managerPort );
	    if( connect( readFd, (struct sockaddr *)&netPort, sizeof( struct sockaddr )) == -1 )
	    {
#ifdef WIN32
		errno = WSAGetLastError ();
		if( errno != WSAEINPROGRESS )
#else
		if( errno != EINPROGRESS )
#endif WIN32
		{
		    printerError( "Can't connect to alarm manager" );
		    exit (1);
		}
	    }
	    /* announce ourselves to the manager */
	    {
		PrinterNameMsg	msg;
		msg.msgType = htonl (PRINTER_NAME_MSG);
		strcpy (msg.printerName, printer_name);
		if (send (readFd, (char *)&msg, sizeof (msg), 0) != sizeof (msg)) {
		    printerError( "Can't send to manager" );
		    exit (1);
		}
    	    }
	}
	else if (numReady == 0) /* timeout, let others use printer */
	{
	    /* close the printer */
	    if (printDesc)
	    {
		if (strstr (printer_name, "/dev/") == printer_name)
		{
		    fclose (printFile);
		}
		else
		{
#ifdef WIN32
		    closesocket( printDesc );
#else
		    close( printDesc );
#endif WIN32
		}
	        printDesc = 0;
	    }
	}
	else
	{
	    rcvSize = recv( readFd, txtBuf, sizeof (txtBuf)-1, 0);
	    if( rcvSize )
	    {
		txtBuf [rcvSize] = 0; /* null terminate the incoming data!! */
		sendToPrinter( txtBuf );
	    }
	    else
	    {
#ifdef WIN32
		closesocket (readFd);
#else
		close (readFd);
#endif WIN32
		printerError( "Zero length read from alarm manager, reopening connection" );
		/* Connect with alarm manager */
		if(( readFd = socket( AF_INET, SOCK_STREAM, 0 )) == -1 )
		{
		    printerError( "Can't create socket to alarm manager" );
		    exit (1);
		}
#ifdef WIN32
		if (ioctlsocket( readFd, FIONBIO, &optionValue) == -1)
#else
		if (ioctl( readFd, FIONBIO, &optionValue) == -1)
#endif WIN32
		{
		    printerError( "Can't set manager socket nonblocking" );
		    exit (1);
		}
		if(( alarmMgrHost = gethostbyname( "localhost" )) == NULL )
		{
		    printerError( "Failure getting local host" );
		    exit (1);
		}
		memset( &netPort, 0, sizeof( struct sockaddr_in ));
		netPort.sin_family = AF_INET;
		hAddr = *(long *)alarmMgrHost->h_addr_list[0];
		netPort.sin_addr.s_addr = hAddr;
		netPort.sin_port = htons( managerPort );
		if( connect( readFd, (struct sockaddr *)&netPort, sizeof( struct sockaddr )) == -1 )
		{
#ifdef WIN32
		    errno = WSAGetLastError ();
		    if( errno != WSAEINPROGRESS )
#else
		    if( errno != EINPROGRESS )
#endif WIN32
		    {
			printerError( "Can't connect to alarm manager" );
			exit (1);
		    }
		}
		/* announce ourselves to the manager */
		{
		    PrinterNameMsg	msg;
		    msg.msgType = htonl (PRINTER_NAME_MSG);
		    strcpy (msg.printerName, printer_name);
		    if (send (readFd, (char *)&msg, sizeof (msg), 0) != sizeof (msg)) {
			printerError( "Can't send to manager" );
			exit (1);
		    }
		}
	    }
	}
    }
}

static int sendToPrinter(char *prBuf )
{
    static char	printBuf[800]; /* worst case expansion */      
    int		numToSend;
    int		numSent;
    char	*prptr;
    char	*inptr;
    int		tries;
    int		numReady;
    fd_set	netPrFdSet;
    int		prLen;

    if (printDesc == 0)
    {
	printInit( );
    }

    if ( strstr ( printer_name, "/dev/") == printer_name)
    {
	fprintf (printFile, "%-s\n", prBuf );
	fflush (printFile);
    }
    else
    {
	prLen = strlen(prBuf);
	inptr = prBuf;
	while (prLen)
	{
	     for (prptr = printBuf;
		prLen && ((prptr-printBuf) < sizeof(printBuf));
		inptr++, prLen--, prptr++)
	     {
		*prptr = *inptr;
		if (*prptr == '\n')
		{
		    *prptr = '\r';
		    prptr++;
		    *prptr = '\n';
		}
	    }
	    *prptr = '\0';
	
	    numToSend = strlen( printBuf );
	    prptr = printBuf;
	    for (tries = 0; tries < MAX_PRINT_TRIES; tries++)
	    {
		FD_ZERO( &netPrFdSet );
		FD_SET( printDesc, &netPrFdSet );
		numReady = select( printDesc+1, NULL, &netPrFdSet, NULL, &printDelay );
		if( numReady > 0 )  {
		    numSent = send( printDesc, prptr, numToSend, 0 );
		    if( numSent == -1 )  {
#ifdef WIN32
			errno = WSAGetLastError ();
			if( errno != WSAEWOULDBLOCK )  {
#else
			if( errno != EWOULDBLOCK )  {
#endif WIN32
			    /* socket send error, reset printer stream */
#ifdef WIN32
			    closesocket( printDesc );
#else
			    close( printDesc );
#endif WIN32
			    printDesc = 0;
			    printInit( );
			}
		    }
		    else  {
			if( numSent == numToSend ) {
			    break;
			}
			else {
			    /* partial send, adjust and try to send the rest */
			    prptr += numSent;
			    numToSend -= numSent;
			}
		    }
		}
		else if( numReady  < 0 ) {
		    /* select error, reset print stream */
#ifdef WIN32
		    closesocket( printDesc );
#else
		    close( printDesc );
#endif WIN32
		    printDesc = 0;
		    printInit( );
		}
	    }
	    if (numSent != numToSend) {
		printerError( "Unable to send to network alarm printer" );
#ifdef WIN32
		closesocket( printDesc );
#else
		close( printDesc );
#endif WIN32
		printDesc = 0;
		return strlen( prBuf ) - numToSend;
	    }
	}
	return (strlen(prBuf));
    }
}

static void printerError(char *errMsg )
{       
    tisSysLog (0, 2, "alarmprint:  Error with printer %s - %s.",
					printer_name, errMsg );
}

static void printInit( )
{
    struct sockaddr_in		netPort;	/* 6/14/94 DEK - Address of network printer */
    long			hAddr;
    int				tries;
    static char			net_name [100];

    if ( strstr ( printer_name, "/dev/") == printer_name)
    {
	for (tries = 0; tries < MAX_PRINT_TRIES; tries++)
	{ 
	    if ((printFile = fopen (printer_name, "w")))
		break;
	    else
#ifdef WIN32
		Sleep (printDelay.tv_sec * 1000);
#else
		sleep (printDelay.tv_sec);
#endif WIN32
	}
	if (tries == MAX_PRINT_TRIES)
	{
	    printerError( "Unable to open local alarm printer");
	}
	else
	{
	    printDesc = 1; /* signal that file is open */
/*	    printf ("alarmprint local printer \"%s\" opened\n", printer_name); */
	}
    }
    else
    {
	if(( printDesc = socket( AF_INET, SOCK_STREAM, 0 )) == -1 )
	{
	    printerError( "Can't create socket for network printer" );
	    exit (1);
	}
	strncpy (net_name, printer_name, sizeof(net_name));
	if (strchr(net_name, ':'))
	    *(strchr(net_name, ':')) = 0;

	if(( netPrinter = gethostbyname( net_name )) == NULL )
	{
	    printerError( "Nothing known about network printer" );
	    exit (1);
	}
	memset( &netPort, 0, sizeof( struct sockaddr_in ));
	netPort.sin_family = AF_INET;
	hAddr = *(long *)netPrinter->h_addr_list[0];
	netPort.sin_addr.s_addr = hAddr;
	netPort.sin_port = htons( parPrPort );
	if( connect( printDesc, (struct sockaddr *)&netPort, sizeof( struct sockaddr )) == -1 )
	{
	    printerError( "Can't connect to network printer" );
	}
	else
	{
	   /* printf ("alarmprint network printer '%s' opened\n", printer_name); */
	}
    }
}

