
/****************************************************************/
/*                                                              */
/*    This software is the proprietary property of Tate         */
/*    Integrated Systems, L.P.  Patents and copyrights are      */
/*    pending.                                                  */
/*                                                              */
/*    This software is furnished under a license and may        */
/*    be used and copied only in accordance with the terms      */
/*    of such license and with the inclusion of the above       */
/*    notice. This software or any other copies thereof         */
/*    may not be provided or otherwise made available to        */
/*    any other person. No title to and ownership of the        */
/*    software is hereby transferred.                           */
/*                                                              */
/*    The information in this software is subject to            */
/*    change without notice and should not be construed         */
/*    as a commitment by Tate Integrated Systems L.P. or        */
/*    its third party suppliers.                                */
/*                                                              */
/*    Tate Integrated Systems, L.P. (TIS) and its third         */
/*    party suppliers assume no responsibility for the          */
/*    use or inability to use any of its software. TIS          */
/*    software is provided "as is" without warranty of          */
/*    any kind and TIS expressly disclaims all implied          */
/*    warranties, including but not limited to the implied      */
/*    warranties of merchantability and fitness for a           */
/*    particular purpose.                                       */
/*                                                              */
/*    Notice: Notwithstanding any other lease or license        */
/*    that may pertain to or accompany the delivery of this     */
/*    computer software, the rights of the Government           */
/*    regarding its use, reproduction, and disclosure are       */
/*    as set forth in Section 52.227-19 of the FARS             */
/*    Computer Software-Restricted Rights clause.               */
/*                                                              */
/****************************************************************/

#ifndef alarmfil_h
#define alarmfil_h

#define NO_ERR 0
#define OPEN_ERR 1
#define TEMP_ERR 2
#define CREATE_ERR 3
#define SOCK_ERR 4

#define NORECORD  0
#define RECFOUND  1
#define RECAFTER  2

#define FILTOK			0
#define FILTOUT			1
#define FILTOUTOFRANGELOW	2
#define FILTOUTOFRANGEHIGH	3

typedef struct
{
	char			date[AL_DATE_SIZE+1];	/* local date */
	char			time[AL_TIME_SIZE+1];	/* and local time */
	TS_STAMP		eventtime;		/* time of event (GMT) */
	char			pointName[AL_POINTNAME_SIZE+1];
	char			descriptor[AL_DESCRIPTOR_SIZE+1];
	char			status[AL_STATUS_SIZE+1];
	char			severity[AL_SEVERITY_SIZE+1];
	char			value[AL_VALUE_SIZE+1];
	char			egutag[AL_EGUTAG_SIZE+1];
	int			acked;
	TS_STAMP		acktime;		/* time of ack */
	char			operatorName[AL_OPERATOR_SIZE+1];	/* who acked */
	char			comment[AL_MAX_MEMO_SIZE+1];		/* 3/30/94 DEK */
	char			stationName[AL_IOCNAME_SIZE+1];	/* station reporting alarm */
	char			info1 [AL_INFO_SIZE+1]; /* extra alarm info fields */
	char			info2 [AL_INFO_SIZE+1]; /* extra alarm info fields */
	char			info3 [AL_INFO_SIZE+1]; /* extra alarm info fields */
	char			info4 [AL_INFO_SIZE+1]; /* extra alarm info fields */
	char			info5 [AL_INFO_SIZE+1]; /* extra alarm info fields */
} AlarmLogRecord;

/* NOTE: the defines in the Filter structure reflect the organization of the
**	radio and check box groups in the filter dialogs. if a change is
**	made in the dialogs, these defines must change also!
*/
typedef struct {
    FilterName	name;
    int		modified;	/* not saved in data base! */
    GroupName	group;
    long	ack;
#define ACKFILTER	1
#define UNACKFILTER	2
    long	sev;
#define MAJORFILTER	1
#define MINORFILTER	2
#define INFOFILTER	4
#define EVENTFILTER	8
#define NOALARMFILTER  16
    long	alarmOnly;
#define ALLFILTER	1
#define ALMONLYFILTER	2
    char	start [AL_DATE_SIZE+1+AL_TIME_SIZE+1];
    char	end [AL_DATE_SIZE+1+AL_TIME_SIZE+1];
    Operator	op;
    long	startupDisplay;
#define ALSUMSTART	2
#define GRPSTART	4
#define ALWINSTART	8
#define ALHISTSTART	16
#define EVHISTSTART	32
#define OPTRKSTART	64
#define REPORTSTART	128
#define	SYSTEMSTART	256
} Filter;

#define REPORTNAME_SIZE	35

typedef struct {
    char	name [REPORTNAME_SIZE];	/* report name */
    long	reportitems;	/* selected criteria specification */
/* criteria specification - first column items bits 0-15 */
#define RPTOFFSCAN	0x0001
#define RPTDISABLE	0x0002
#define RPTINITFAIL	0x0004
#define RPTIOFAIL	0x0008
#define RPTBADSENS	0x0010
/* criteria specification - second column items bits 16-31 */
#define RPTSYSTAG	0x0100
#define RPTINFOTAG	0x0200
#define RPTMANUAL	0x0400
#define RPTPUMPFAULT	0x0800
#define RPTVALVFAULT	0x1000
} RepCriteria;

/* packed color manipulation */
#define Color(r,g,b) ((r<<16) + (g<<8) + b)
#define Red(c)       ((c>>16) & 255)
#define Green(c)     ((c>>8) & 255)
#define Blue(c)      (c & 255)

int OpenCodeBase ( void );
int myOperPriv ( char * );
int Log_AlarmLog ( void );
int Log_CloseAlarmLog ( void );
unsigned int Log_NumRecords ( int indexFile );
int Log_RecordByTime ( char *pointname, TS_STAMP *eventtime, AlarmLogRecord *alarmLogRecordPtr);
int Log_Record ( unsigned int recordNum, AlarmLogRecord *alarmLogRecordPtr, int indexFile );
int Log_RecordsByTime ( char *pointname, TS_STAMP *eventtime, 
			AlarmLogRecord alarmLogRecordPtr[], 
			int numRecs, int indexFile, int (*CheckFunction)(), void *arg );
void LoadRecord ( AlarmLogRecord *alarmLogRecordPtr );
int Log_SetFilter ( Filter *filter, int indexFile );
TreeNode *FindGroup (TreeNode *, char *);
PvData	*FindChannel (TreeNode *, char *);
int Log_PrintAlarmLog ( FILE *printerFileDesc );
int Log_PrintEventLog ( FILE *printerFileDesc );
int Log_PrintRecord ( FILE *printerFileDesc );
int PostAlarm ( void );
int LogAlarm ( void );
int Log_LogAlarm ( char *group, char *name, char *descriptor, time_t alarmTime,
		   unsigned int curStat, unsigned int curSevr, unsigned int ack,
		   char *valAckPtr, char *tagOpPtr, char *location );
void alarminfoInit ( void );
void GetAlarmInfo (char *name, char **Guidance, char **Display);
int TrkOpenFile (void);
void TrkInit (void);
int TrkSeek (unsigned int recordNum);
void TrkSetFilter (Filter *filter);
unsigned int TrkNumEntries ();
long TrkEventTime ();
char *TrkDate ();
char *TrkTime ();
char *TrkPointname ();
char *TrkDescriptor ();
char *TrkFieldname ();
char *TrkOldvalue ();
char *TrkNewvalue ();
char *TrkOperator ();
void TrkLogAck (char *group, char *name, char *descriptor, time_t ackTime,
		unsigned int curStat, unsigned int curSevr, char *ackString,
		char *location);
void FiltInit (void);
void GetFilter (char *name, Filter *filt);
void SaveFilter (char *name, Filter *filt);
char **GetFilterNames (void);
void DeleteFilter (char *name);
void ReportInit (void);
void GetReport (char *name, RepCriteria *criteria);
void SaveReport (char *name, RepCriteria *criteria);
char **GetReportNames (void);
void DeleteReport (char *name);

char **GetSubGroups (char *groupName);
void trimblanks (char *string);
void padblanks (char *string, int len); 
#ifdef USEX11
int alloc_color_cells (Widget w, int n);
Pixel get_pixel_from_index (int n);
int get_index_from_pixel (Pixel p);
Pixel set_pixel_from_long (Widget w, int n, long c);
Pixel get_pixel_from_long (Widget, long);
long  get_color_from_pixel (Widget, Pixel);
void ForceUpdate ( Widget  w );
#endif USEX11

long UnixTime (long *);

#endif alarmfil_h
