#ifndef	aldlglistviewHXX
#define	aldlglistviewHXX

class AlarmDialog;
class AckButton;
class ListEntry;

class AldlgItemList : public vdialogItemList 
{

public:

    // the default versions of the necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(AldlgItemList, vdialogItemList);

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(AldlgItemList);

    // function overrides

    void     DetermineBounds (vrect *rect);
    void     DetermineContent (vrect *rect);
    vcursor *DetermineCursor (void);
    void     DetermineColorsFrom (long phlags, vcolor **known, int *ids,
				  int count, vcolor **colors);
    void     DetermineColorSchemeFrom (vcolorScheme *known,
				       unsigned long phlags,
				       vcolorScheme *scheme);
    vdialog *DetermineDialog (void);
    void     DrawDevice (void);
    int      DetermineEnabled (void);
    vfont   *DetermineFont (void);
    int      DetermineLineWidth (void);
    int      DetermineShown (void);
    vdict   *GetKeyMap (void);
    vdict   *GetKeyDispatch (void);
};


// listview subclass, with extra facilities for handling dialog
// item lists

class AldlgListView : public vlistview 
{
    AlarmDialog	    *alarmDialog;
    ListEntry	    *entry;
    vlist	    *lastList;
    int		    firstList;
    AldlgItemList   *itemList;	// the list of dialogitems in the listview

public:

    // the default versions of the necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(AldlgListView, vlistview);

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(AldlgListView);

    // overrides

    void Destroy (void);

    // load dynamically from a resource
    void LoadInit (vresource resource);

    // vobjectAttributed

    // notification of an attribute change in a parent
    void ChangedParentAttribute (const vname *attr, vdict *context);

    // notification of attribute changes
    void ChangedAttribute (const vname *attr, vdict *context);

    // handle focus stuff
    int IssueSelector (int selector, vdict *context);

    // vdialogItem stuff

    // close the item down
    void Close (void);

    void DrawCellBackground (int row, int col, int selected, vrect *rect);

    // draw the cell contents
    void DrawCellContents (int row, int col, int selected, vrect *rect);

    // give the item keyboard focus
    void Focus (void);

    // handle clicks
    int HandleButtonDown (vevent *event);

    // handle button releases
    int HandleButtonUp (vevent *event);

    // handle mouse drags
    int HandleDrag (vevent *event);

    // handle keypresses
    int HandleKeyDown (vevent *event);

    void ObserveScroll(int horzDistance, int vertDistance);
    void ObserveSelect(int state);


    // open the item in a dialog
    void Open (void);

    // scroll to this position
    void ScrollAbsolute (int x, int y);

    // assign a width to a given column
    void SetColumnWidth (int col, int width);

    // assign a height to a given row
    void SetRowHeight (int row, int height);

    // assign a list to be viewed
    void SetList (vlist *list);

    // snatch away keyboard focus
    void Unfocus (void);
    
    // class extensions

    void AdjustItemRect (int row, int col, vrect *rect);
    void CalculateBounds (void);
    vdialogItem *FindItem (const vname *tag);
    int GetCellHeight (void);
    int	GetListNumElements (void);

    // inline member methods

    // return the listview's item list
    AldlgItemList *GetItemList (void) 
	{ return itemList; }
    ListEntry *GetListEntry (void)
	{ return (entry); }
    void SetListEntry (ListEntry *ent)
	{ entry = ent; }

    AckButton		*ackButtons[ALVU_MAXITEMS];

    // class inlines
    void StoreAlarmDialog (AlarmDialog *dlg)
	{ alarmDialog = dlg; }

};


// *** ALARM DIALOG LIST ENTRY CLASS **********************************

class ListEntry
{
    AlarmDialog	*alarmDialog;

    unsigned int	row;	// row number in alarm dialog dataList
    int			AlmColor;
    int			ackMany;
    char		date [24];
    char		name [30];
    char		desc [256];
    TS_STAMP		evttime;
    chid		caChid;
    evid		evId;
    char		value [MAX_STRING_SIZE]; // MAX in db_access.h
    short		curstat;	// alarm status: type
    short		curseverity;	//		 severity
    short		ackstate;	//		 acked
    int			monitored;	// is it monitored

public:

    ListEntry (unsigned int id, int Color, char *Date, char *Name,
	       char *Desc, TS_STAMP *EventTime, AlarmDialog *dlg);
    ListEntry (unsigned int id, int Color, int AckM, char *Date,
	       char *Name, char *Desc, TS_STAMP *EventTime, AlarmDialog *dlg);
    ~ListEntry (void);

    // inline functions
    AckButton	*GetButtonItem (void)
	{ return (buttonItem); }

    unsigned int GetRow (void)
	{ return (row); }
    int AckMany (void)
	{ return (ackMany); }
    char *GetDate (void)
	{ return (date); }
    char *GetName (void)
	{ return (name); }
    char *GetDesc (void)
	{ return (desc); }
    int GetColor (void)
	{ return (AlmColor); }
    void SetColor (int newcolor)
	{ AlmColor = newcolor; }
    void SetAckMany (int newvalue)
	{ ackMany = newvalue; }
    void SetDesc (char *newdesc)
	{
	    strncpy (desc, newdesc, sizeof (desc));
	    desc [sizeof (desc) -1] = 0;
	}
    char *GetValue (void)
	{ return (value); }
    short GetCurrSeverity (void)
	{ return (curseverity); }
    short GetCurrStatus (void)
	{ return (curstat); }
    short GetCurrAck (void)
	{ return (ackstate); }
    TS_STAMP *GetEventTime (void)
	{ return (&evttime); }
    int IsMonitored (void)
    { return (monitored); }

    int ConnCB (connection_handler_args args);
    int MonCB (event_handler_args args);
    int Monitor (void);
    void StopMonitor (void);

    // functions for alarmDialog
    void StoreAlarmDialog (AlarmDialog *dlg)
	{ alarmDialog = dlg; }
    AlarmDialog *GetAlarmDialog (void)
	{ return (alarmDialog); }
    AckButton	*buttonItem;
};


// *** LIST ENTRY ACK BUTTON CLASS **********************************

class AckButton : public vbutton
{
    AlarmDialog	*alarmDialog;
    ListEntry	*entry;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (AckButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (AckButton, vbutton);

    // virtual member function overrides
    int  HandleButtonDown (vevent *event);
    int  HandleButtonUp (vevent *event);
    void DrawContent (void);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }

    // handle the pointer to entry
    void StoreListEntry (ListEntry *ent) 
	{ entry = ent; }
};

#endif	aldlglistviewHXX

