/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "tsDefs.h"
#include "alarmDefs.h"
#include "shareMem.h"
#include "alarmfil.h"
#include "alpvmap.h"

/* Functions defined in this module */


/* Codebase related things */

C4CODE	codeBase;

int	initialized;

D4DATA	*pvMapData;
F4FIELD *pvDeleteFlagFld;
D4DATA	*pdbXrefData;
char	dbfilename [AL_MAXFILENAME_SIZE];
D4DATA	*dbData;

/*----------*/
void initializeDatabase (void)
{
    if (!initialized)
    {
	d4init (&codeBase);
#if defined (WIN32) || defined(USECB63)
	codeBase.errCreate = 0;
	codeBase.errOpen = 0;
#else
	codeBase.open_error = 0;
	codeBase.create_error = 0;
#endif WIN32
	openPvMapFile ();
	openPdbXrefFile ();
	initialized = 1;
    }
}

void resolveLinks (char	*orig, char *resolved);

/*----------*/
int openPvMapFile (void)
{
    char	pathName [200];

    resolveLinks (AL_PVMAP_FILE_NAME, pathName);
    pvMapData = d4open (&codeBase, pathName);
    if (pvMapData == NULL)
	return 0;

    pvDeleteFlagFld = d4field (pvMapData, "DELETEFLAG");
    d4unlock_all (pvMapData);
    return 1;
}

/*----------*/
void closePvMapFile (void)
{
    d4close (pvMapData);
}

/*----------*/
int openPdbXrefFile (void)
{
    char	pathName [200];

    resolveLinks (AL_PDBXREF_FILE_NAME, pathName);
    pdbXrefData = d4open (&codeBase, pathName);
    if (pdbXrefData == NULL)
	return 0;

    d4unlock_all (pdbXrefData);
    return 1;
}

/*----------*/
void closePdbXrefFile (void)
{
    d4close (pdbXrefData);
}

/*----------*/
int getNumPVs (void)
{
    return d4reccount (pvMapData);
}

char	currentPoint[50];
char	currentType[16];
char	currentFldname[7];

/*----------*/
void ResetPVPoint ( )
{
    initializeDatabase ();
    currentPoint [0] = 0;
    currentType [0] = 0;
}

/*----------*/
int	getPv (pointName)	/* Find a specific point */
char	*pointName;
{
    char	key [50];

    if (strcmp (currentPoint, pointName) == 0)
	return 1;

    sprintf (key, "%-*s", AL_POINTNAME_SIZE, pointName);
    if (d4seek (pvMapData, key) != 0)
    {
	d4unlock_all (pvMapData);
	return 0;
    }
    if (f4true (pvDeleteFlagFld)) {
	d4unlock_all (pvMapData);
	return 0;
    }
    strncpy (currentPoint, pointName, sizeof(currentPoint));
    d4unlock_all (pvMapData);
    return 1;
}

/*----------*/
char	*GetPVPointInfo ( char *pvName, char *dbfld )
{
    if (getPv (pvName))
    {
	return f4str (d4field (pvMapData, dbfld));
    }
    return "***NOT FOUND***";
}

/*----------*/
double	GetPVPointValue ( char *pvName, char *dbfld )
{
    if (getPv (pvName))
    {
	if (f4type (d4field (pvMapData, dbfld)) == 'L')
	{
	    return (double)(f4true (d4field (pvMapData, dbfld)));
	}
	else
	{
	    return f4double (d4field (pvMapData, dbfld));
	}
    }
    return 0.0;    
}

/*----------*/
char	*GetPdbXrefInfo ( char *pointType, char *fldname, char *dbfld)
{
    char	key [50];

    if ((strcmp (currentType, pointType) != 0)
       || (strcmp (currentFldname, fldname) != 0))
    {
	d4tag_select (pdbXrefData, d4tag (pdbXrefData, "XREF"));
	sprintf (key, "%-*s%-*s", AL_POINTTYPE_SIZE, pointType, AL_FIELDNAME_SIZE, fldname);
	if (d4seek (pdbXrefData, key) != r4success)
	{
	    d4unlock_all (pdbXrefData);
	    return NULL;
	}
    	if (d4deleted (pdbXrefData)) {
	    d4unlock_all (pdbXrefData);
	    return NULL;
	}
	strncpy (currentType, pointType, sizeof(currentType));
	strncpy (currentFldname, fldname, sizeof(currentFldname));
	d4unlock_all (pdbXrefData);
    }
    return f4str (d4field (pdbXrefData, dbfld));
}

/*----------*/
char	*GetDbInfo ( char *dbfile, char *pointName, char *dbfld )
{
    char	key [50];
    char	pathName [200];
    static char returnString [200];

    if (strcmp (currentType, dbfile) != 0)
    {
	strncpy (dbfilename, AL_DBASE_DIRECTORY_NAME, sizeof(dbfilename));
	trimblanks (dbfile);
	strncat (dbfilename, dbfile, sizeof(dbfilename));
	strncat (dbfilename, ".dbf", sizeof(dbfilename));
	resolveLinks (dbfilename, pathName);
	dbData = d4open (&codeBase, pathName);
	if (dbData == NULL)
	    return NULL;
    }
    sprintf (key, "%-*s", AL_POINTNAME_SIZE, pointName);
    if (d4seek (dbData, key) != 0)
    {
	d4close (dbData);
	return NULL;
    }
    if (d4deleted (dbData)) {
	d4close (dbData);
	return NULL;
    }
    d4unlock_all (dbData);

    strncpy (returnString, f4str (d4field (dbData, dbfld)), sizeof (returnString));
    d4close (dbData);
    return returnString;
}

/*----------*/
int GetManualMode (char *pvName)
{
    char	pointType[50];
    char	fieldName[50];
    char	stateInfo[50];
    char	*valString;

    if (getPv (pvName))
    {
	strncpy (pointType, GetPVPointInfo (pvName, "TYPE"), sizeof(pointType));
	if (GetPdbXrefInfo (pointType, "MODE", "FIELDNAME"))
	{
	    strncpy (fieldName, GetPdbXrefInfo (pointType, "MODE", "FIELD_NAME"), sizeof (fieldName));
	    strncpy (stateInfo, GetPdbXrefInfo (pointType, "MODE", "TRUE_STATE"), sizeof (stateInfo));
	    valString = GetDbInfo (GetPdbXrefInfo (pointType, "MODE", "TABLE"),
					 pvName, fieldName);
	    if (valString)
		return (strcmp (valString, stateInfo) == 0);
	    else
		return 0;
	}
	else
	{
	    if (GetPdbXrefInfo (pointType, "OMSL", "FIELDNAME"))
	    {
		strncpy (fieldName, GetPdbXrefInfo (pointType, "OMSL", "FIELD_NAME"), sizeof (fieldName));
		strncpy (stateInfo, GetPdbXrefInfo (pointType, "OMSL", "TRUE_STATE"), sizeof (stateInfo));
		valString = GetDbInfo (GetPdbXrefInfo (pointType, "OMSL", "TABLE"),
					     pvName, fieldName);
		if (valString)
		    return (strcmp (valString, stateInfo) == 0);
		else
		    return 0;
	    }
	    else
	    {
		return 0;
	    }
	}
    }
    return 0;
}

/*----------*/
int GetInfoTag (char *pvName)
{

    if (getPv (pvName))
    {
	return f4true (d4field (pvMapData, "INFOTAG"));
    }
    return 0;
}

/*----------*/
/* current field values in pdbxref 
Pointtype	Fieldname Field_name	FTable		True_state	OpID HWFlag 	RefField ConstField
"ai",		"MODE",	"AUTOMATIC",	"ai",		F,		"",	F,	"",	""
"bi",		"MODE",	"AUTOMATIC",	"di",		F,		"",	F,	"",	""
"calc",		"MODE",	"AUTOMATIC",	"calc",		F,		"",	F,	"",	""
"mbbi",		"MODE",	"AUTOMATIC",	"mbbi",		F,		"",	F,	"",	""
"stringin",	"MODE",	"AUTOMATIC",	"strngin",	F,		"",	F,	"",	""
"leadlag",	"MODE",	"AUTOMATIC",	"leadlag",	F,		"",	F,	"",	""
"ramp",		"MODE",	"AUTOMATIC",	"ramp",		N,		"",	F,	"",	""
"pipe",		"MODE",	"MODE",		"pipe",		N,		"",	F,	"",	""
"tank",		"MODE",	"MODE",		"tank4",	F,		"",	F,	"",	""
"valve",	"MODE",	"MODE",		"valve",	F,		"",	F,	"",	""
"pump",		"MODE",	"MODE",		"pump",		F,		"",	F,	"",	""
"pot",		"MODE",	"MODE",		"pot",		F,		"",	F,	"",	""
"accum",	"MODE",	"AUTOMATIC",	"accum",	F,		"",	F,	"",	""
"sel",		"MODE",	"AUTOMATIC",	"sel",		F,		"",	F,	"",	""
"deadtime",	"MODE",	"AUTOMATIC",	"deadtime",	F,		"",	F,	"",	""
"datamove",	"MODE",	"AUTOMATIC",	"datamove",	F,		"",	F,	"",	""
"sigchar",	"MODE",	"AUTOMATIC",	"sigchar",	F,		"",	F,	"",	""
"udb",		"MODE",	"MODE",		"udb",		F,		"",	F,	"",	""
"ldfw1",	"MODE",	"AUTOMATIC",	"ldfw1",	F,		"",	F,	"",	""
"ldfw2",	"MODE",	"AUTOMATIC",	"ldfw2",	F,		"",	F,	"",	""
"ddc",		"MODE",	"AUTOMATIC",	"ddc",		F,		"",	F,	"",	""
"ao",		"OMSL",	"CLOSEDLOOP",	"ao",		T,		"",	F,	"",	""
"bo",		"OMSL",	"CLOSEDLOOP",	"do",		T,		"",	F,	"",	""
"mbbo",		"OMSL",	"CLOSEDLOOP",	"mbbo",		T,		"",	F,	"",	""
"stringout",	"OMSL",	"CLOSEDLOOP",	"strngout",	T,		"",	F,	"",	""
"pto",		"OMSL",	"CLOSEDLOOP",	"pto",		T,		"",	F,	"",	""
"td",		"OMSL",	"CLOSEDLOOP",	"td",		T,		"",	F,	"",	""
"pat",		"OMSL",	"CLOSEDLOOP",	"pat",		T,		"",	F,	"",	""
"ddc",		"OMSL",	"CLOSEDLOOP",	"ddc",		T,		"",	F,	"",	""
*/
