#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vbuttonHEADER
#include vcontainerHEADER
#include vdictHEADER
#include vmenuHEADER
#include vmenubarHEADER
#include vcolorHEADER
#include vtimerHEADER
#include vlistviewHEADER
#include vcomboboxHEADER
#include vprintHEADER

extern "C"
{
#ifdef WIN32
#include <stdio.h>
#include <winsock.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#endif WIN32

#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "cadef.h"
#include "shareMem.h"
#include "alarmDefs.h"
#include "alarmfil.h"
#include "alarmvu.h"
#include "bell.h"
#include "messages.h"		// alarm/messages.h
#include "OperPriv.h"
#include "pdfHelpUtils.h"	// func protos for help menu

extern ShareHeader  *shareHeader;
extern TreeNode	    *treeRoot;

}

#include "alvu.h"
#include "alarmvu.hxx"
#include "alvuMainDialog.hxx"
#include "aldlgListView.hxx"
#include "alarmDialogs.hxx"
#include "comboDialog.hxx"
#include "colorDialog.hxx"
#include "filterDialog.hxx"
#include "oneLineDialog.hxx"
#include "popupDialog.hxx"
#include "alvuProgressDialog.hxx"
#include "alvuUtils.hxx"
#include "utils.hxx"

extern int	    alarmSoundEnabled;
extern int	    alarmSummaryDisabled;
extern long	    backgroundColor;
extern vcolor	    *backgroundVcolor;
extern vcolor	    *blackVcolor;
extern char	    **channelNames;
extern PvData	    **channelPtrs;
extern long	    CurrUnackSevColor [];
extern long	    defaultAlarmColors []; 
extern char	    defFilterName [AL_FILTERNAME_SIZE + 1];
extern char	    **groupNames;
extern TreeNode	    **groupPtrs;
extern char	    linePrinter [LINEPRINTER_SIZE + 1];
extern int	    MaxSendsPerSecond;
extern unsigned int popupAlrmEnable;
extern long	    SevColor [];
extern long	    UnackSevColor [];
extern unsigned int winEnable;

// global ptrs to possible Dialogs, see alarmvu.cxx
extern AlarmEventDialog		*alarmEventDialogPtr;
extern AlarmHistoryDialog	*alarmHistoryDialogPtr;
extern AlarmPointsDialog	*reportsDialogPtr;
extern AlarmSummaryDialog	*alarmSummaryDialogPtr;
extern AlarmTrackDialog		*alarmTrackDialogPtr;
extern AlarmWindowDialog	*alarmWindowDialogPtr;
extern DynamicGroupDialog	*dynamicGroupDialogPtr;
extern PopupDialog		*popAlarmPtr;
extern AlarmSystemDialog	*systemDialogPtr;

// gateway function so bell.c can call mainDialog::Beep
extern "C"
{
void MainDialogRingBellGW (void)
{
    vwindow::Beep ();
}
}

static void observeHelpMenu (vmenuItem *item, vevent *event);
vkindTYPED_FULL_DEFINITION(helpMenuItem, vmenuItem, "helpMenuItem");


vkindTYPED_FULL_DEFINITION(alvuMainDialog, vdialog, "alvuMainDialog");

/*******************************************************************
*
*    alvuMainDialog constructor
*
*******************************************************************/

void alvuMainDialog::LoadInit (vresource res)
{
    tisAbout = NULL;

    // init to false
    loadiniterror = ALVU_SUCCESS;

    vmenubar *menubar = vmenubar::RequestSubLoad (this, alvu_MainMenu);

    // load the File menu items
    vmenu *fileMenu = vmenu::RequestSubLoad (menubar, alvu_MainFileMenu);

    vmenuItem *selectFilterItem = vmenuItem::RequestSubLoad (fileMenu,
				    	alvu_MainFileSelect);
    vmenuItem *newFilterItem = vmenuItem::RequestSubLoad (fileMenu,
				    	alvu_MainFileNew);
    vmenuItem *editFilterItem = vmenuItem::RequestSubLoad (fileMenu,
				    	alvu_MainFileEdit);
    vmenuItem *deleteFilterItem = vmenuItem::RequestSubLoad (fileMenu,
				    	alvu_MainFileDelete);
    vmenuItem *printItem = vmenuItem::RequestSubLoad (fileMenu,
				    	alvu_MainFilePrint);
    vmenuItem *exitItem = vmenuItem::RequestSubLoad (fileMenu,
				    	alvu_MainFileExit);

    // load the View menu items
    vmenu *viewMenu = vmenu::RequestSubLoad (menubar, alvu_MainViewMenu);

    vmenuItem *groupItem = vmenuItem::RequestSubLoad (viewMenu,
				    	alvu_MainViewGroup);
    vmenuItem *alWindowItem = vmenuItem::RequestSubLoad (viewMenu,
				    	alvu_MainViewAlWindow);
    vmenuItem *alSumItem = vmenuItem::RequestSubLoad (viewMenu,
				    	alvu_MainViewAlSum);
    vmenuItem *alHistItem = vmenuItem::RequestSubLoad (viewMenu,
				    	alvu_MainViewAlHist);
    vmenuItem *evHistItem = vmenuItem::RequestSubLoad (viewMenu,
				    	alvu_MainViewEvHist);
    vmenuItem *opTrackItem = vmenuItem::RequestSubLoad (viewMenu,
				    	alvu_MainViewOpTrack);
    vmenuItem *reportsItem = vmenuItem::RequestSubLoad (viewMenu,
				    	alvu_MainViewReports);
    vmenuItem *systemItem = vmenuItem::RequestSubLoad (viewMenu,
					alvu_MainViewSysHistory);

    // load the Control menu items
    vmenu *ctrlMenu = vmenu::RequestSubLoad (menubar, alvu_MainControlMenu);

    vmenuItem *disablePtItem = vmenuItem::RequestSubLoad (ctrlMenu,
				    	alvu_MainCtrlDisPt);
    vmenuItem *enablePtItem = vmenuItem::RequestSubLoad (ctrlMenu,
				    	alvu_MainCtrlEnPt);
    vmenuItem *disableGrpItem = vmenuItem::RequestSubLoad (ctrlMenu,
				    	alvu_MainCtrlDisGrp);
    vmenuItem *enableGrpItem = vmenuItem::RequestSubLoad (ctrlMenu,
				    	alvu_MainCtrlEnGrp);

    // load the Setup menu items
    vmenu *setupMenu = vmenu::RequestSubLoad (menubar, alvu_MainSetupMenu);

    vmenuItem *defaultsItem = vmenuItem::RequestSubLoad (setupMenu,
				    	alvu_MainSetupDef);
    vmenuItem *saveItem = vmenuItem::RequestSubLoad (setupMenu,
				    	alvu_MainSetupSave);
    vmenuItem *printerItem = vmenuItem::RequestSubLoad (setupMenu,
				    	alvu_MainSetupPrt);
    vmenuItem *colorsItem = vmenuItem::RequestSubLoad (setupMenu,
				    	alvu_MainSetupColors);
    selFilterMenuItem = vmenuItem::RequestSubLoad (setupMenu,
				    	alvu_MainSetupStFilter);
    setupAlWin = vmenuToggleItem::RequestSubLoad (setupMenu,
				    	alvu_MainSetupAlWin);
    setupPopAl = vmenuToggleItem::RequestSubLoad (setupMenu,
				    	alvu_MainSetupPopAl);

    // load Help menu, no menu items, built at run time
    helpMenu = vmenu::RequestSubLoad (menubar, alvu_MainHelpMenu);

    // load in buttons
    vcontainer *buttonContainer = vcontainer::RequestSubLoad (this,
					alvu_MainContainer);

    silenceButton = mainDialogSilence::RequestSubLoad (buttonContainer,
					alvu_MainSilenceButton);
    summaryButton = mainDialogSummary::RequestSubLoad (buttonContainer,
					alvu_MainSummaryButton);
    exitButton = mainDialogExit::RequestSubLoad (buttonContainer,
					alvu_MainExitButton);
    majorButton = mainDialogMajor::RequestSubLoad (buttonContainer,	
					alvu_MainMajorButton);
    minorButton = mainDialogMinor::RequestSubLoad (buttonContainer,
					alvu_MainMinorButton);
    infoButton = mainDialogInfo::RequestSubLoad (buttonContainer,
					alvu_MainInfoButton);
    vdialog::LoadInit (res);

    // check on privileges
    int	allDisabled;
    /*
    // init privilege checking, see .../share/src/libtis/OperPrvtrk.c
    if ((InitOperPriv (NULL)) != PRIVOK)
    {
	alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_InitOperPrivErr);
	loadiniterror = ALVU_ERROR;
	return;
    }*/
    // do the files menu
    if ((OperPriv (NULL, "*MMI_ALARMVU_FILES_SELECTFILTER")) == 0)
    {
	selectFilterItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_FILES_NEWFILTER")) == 0)
    {
	newFilterItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_FILES_EDITFILTER")) == 0)
    {
	editFilterItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_FILES_DELETEFILTER")) == 0)
    {
	deleteFilterItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_FILES_PRINT")) == 0)
    {
	printItem -> Disable ();
    }
    // disable this menu item until code is writen
    printItem -> Disable ();

    // do the view menu
    allDisabled = 0;
    if ((OperPriv (NULL, "*MMI_ALARMVU_VIEW_GROUP")) == 0)
    {
	groupItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_VIEW_ALARMWINDOW")) == 0)
    {
	alWindowItem -> Disable ();
	// disable setup of alarm window for setup also
	selFilterMenuItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_VIEW_ALARMSUMMARY")) == 0)
    {
	alSumItem -> Disable ();
	// disable the alarm summary button on the main dialog
	summaryButton -> Disable ();
	// set flag so disable the summary button on the alarm popup dialog
	alarmSummaryDisabled = vTRUE;
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_VIEW_ALARMHISTORY")) == 0)
    {
	alHistItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_VIEW_EVENTHISTORY")) == 0)
    {
	evHistItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_VIEW_OPERTRACK")) == 0)
    {
	opTrackItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_VIEW_REPORTS")) == 0)
    {
	reportsItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_VIEW_SYSTEMHISTORY")) == 0)
    {
	systemItem -> Disable ();
	allDisabled++;
    }
    if (allDisabled == 8)
    {
	viewMenu -> Disable ();
    }

    // do the control menu
    allDisabled = 0;
    if ((OperPriv (NULL, "*MMI_ALARMVU_CONTROL_DISABLEPOINT")) == 0)
    {
	disablePtItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_CONTROL_ENABLEPOINT")) == 0)
    {
	enablePtItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_CONTROL_DISABLEGROUP")) == 0)
    {
	disableGrpItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_CONTROL_ENABLEGROUP")) == 0)
    {
	enableGrpItem -> Disable ();
	allDisabled++;
    }
    if (allDisabled == 4)
    {
	ctrlMenu -> Disable ();
    }

    // do the setup menu
    allDisabled = 0;
    if ((OperPriv (NULL, "*MMI_ALARMVU_SETUP_USEDEFAULTS")) == 0)
    {
	defaultsItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_SETUP_SAVESETUP")) == 0)
    {
	saveItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_SETUP_PRINTER")) == 0)
    {
	printerItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_SETUP_COLORS")) == 0)
    {
	colorsItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_SETUP_STARTUPFILTER")) == 0)
    {
	selFilterMenuItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_SETUP_ALARMWINDOW")) == 0)
    {
	selFilterMenuItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMVU_SETUP_POPUPALARM")) == 0)
    {
	setupPopAl -> Disable ();
	allDisabled++;
    }
    if (allDisabled == 7)
    {
	setupMenu -> Disable ();
    }

    // close privilege checking, see .../share/src/libtis/OperPrvtrk.c
/*    CloseOperPriv (); */

    // store main ptr in buttons
    silenceButton -> StoreMainDialog (this);
    summaryButton -> StoreMainDialog (this);
    exitButton    -> StoreMainDialog (this);
    majorButton   -> StoreMainDialog (this);
    minorButton   -> StoreMainDialog (this);
    infoButton    -> StoreMainDialog (this);

   // store internal names for the file menu tags

    commandNames[SELECT_FILTER_MENUITEM] = alvu_MainFileSelect;
    commandNames[NEW_FILTER_MENUITEM]    = alvu_MainFileNew;
    commandNames[EDIT_FILTER_MENUITEM]   = alvu_MainFileEdit;
    commandNames[DELETE_FILTER_MENUITEM] = alvu_MainFileDelete;
    commandNames[PRINT_MENUITEM]         = alvu_MainFilePrint;
    commandNames[EXIT_MENUITEM]          = alvu_MainFileExit;

    // store internal names for the view menu tags

    commandNames[GROUP_MENUITEM]          = alvu_MainViewGroup;
    commandNames[ALARM_WINDOW_MENUITEM]   = alvu_MainViewAlWindow;
    commandNames[ALARM_SUMMARY_MENUITEM]  = alvu_MainViewAlSum;
    commandNames[ALARM_HISTORY_MENUITEM]  = alvu_MainViewAlHist;
    commandNames[EVENT_HISTORY_MENUITEM]  = alvu_MainViewEvHist;
    commandNames[OPER_TRACK_MENUITEM]     = alvu_MainViewOpTrack;
    commandNames[REPORTS_MENUITEM]        = alvu_MainViewReports;
    commandNames[SYSTEM_HISTORY_MENUITEM] = alvu_MainViewSysHistory;

    // store internal names for the control menu tags

    commandNames[DISABLE_PT_MENUITEM]  = alvu_MainCtrlDisPt;
    commandNames[ENABLE_PT_MENUITEM]   = alvu_MainCtrlEnPt;
    commandNames[DISABLE_GRP_MENUITEM] = alvu_MainCtrlDisGrp;
    commandNames[ENABLE_GRP_MENUITEM]  = alvu_MainCtrlEnGrp;

    // store internal names for the setup menu tags

    commandNames [USE_DEF_MENUITEM]           = alvu_MainSetupDef;
    commandNames [SAVE_MENUITEM]              = alvu_MainSetupSave;
    commandNames [PRINTER_MENUITEM]           = alvu_MainSetupPrt;
    commandNames [COLORS_MENUITEM]            = alvu_MainSetupColors;
    commandNames [STARTUP_FILTER_MENUITEM]    = alvu_MainSetupStFilter;
    commandNames [SETUP_ALARMWINDOW_MENUITEM] = alvu_MainSetupAlWin;
    commandNames [POPUP_MENUITEM]             = alvu_MainSetupPopAl;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;
    // populate our dictionary with menu commands
    for (i = SELECT_FILTER_MENUITEM; i < MENUITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	alvuMainCommand *command = new alvuMainCommand;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (commandNames[i], command);
    }

    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}


/*******************************************************************
*
*    alvuMainDialog destructor
*
*******************************************************************/

void alvuMainDialog::Destroy ()
{
    if (helpItems != NULL)
    {
	freeHelpItems (helpItems);
    }
    if (tisAbout != NULL)
    {
	delete tisAbout;
    }
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    alvuMainDialog window observer
*
*******************************************************************/

void alvuMainDialog::ObserveWindow (vevent *evnt)
{
    if ( evnt -> GetType () == veventWINDOW_STATE && evnt -> IsClose ())
    {
	Close ();
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    alvuMainDialog open function
*
*******************************************************************/

void alvuMainDialog::OpenMain (alvuApplication *app)
{
    // so main will iconify properly to task bar in NT
    AddStyle (vwindowSTYLE_GLOBAL);
    alvuApp = app;

    SetFilterMenuTitle (alvu_StartUpFilterTitle, alvuApp -> shellFilter.name);
    helpItems = NULL;
    BuildHelpMenu ();

    // set toggle menu items
    if (winEnable == vTRUE)
    {
	setupAlWin -> SetOn (vTRUE);
    }
    else
    {
	setupAlWin -> SetOn (vFALSE);
    }
    if (popupAlrmEnable == vTRUE)
    {
	setupPopAl -> SetOn (vTRUE);
    }
    else
    {
	setupPopAl -> SetOn (vFALSE);
    }

    // get the timer
    mainTimer = new mainDialogTimer;
    mainTimer -> StoreMainDialog (this);

    // set owner of printer to main
    if (app -> printer != NULL)
    {
	app -> printer -> SetDocumentOwner (this);
    }

    vdialog::Open ();
}


/*******************************************************************
*
*    alvuMainDialog build help menu function
*
*******************************************************************/

void alvuMainDialog::BuildHelpMenu (void)
{
    helpItems = getHelpItems ("alarmvu", "AlarmVu");
    if (helpItems != NULL)
    {
	int i;
	for (i = 0; helpItems [i]; i++)
	{
	    helpMenuItem *menuItem = new helpMenuItem;
	    menuItem -> SetTitleScribed (vcharScribeSystem (helpItems[i]));
	    menuItem -> SetObserveMenuItemProc (observeHelpMenu);
	    menuItem -> StoreMainDialog (this);
	    helpMenu -> AppendItem (menuItem);
	}
    }
    // do the about menu here
    vstr *verMsg = vresourceGetString (alvuApp -> messagesResource, 
				       alvu_VersionMsg);
    tisAbout = new TisAbout (alvuApp -> tisGalaxyResource, verMsg);
    vstrDestroy (verMsg);
    vstr *aboutTitle = vresourceGetString (alvuApp -> titlesResource, 
					   alvu_AboutTitle);
    tisAbout -> SetAboutTitle (aboutTitle);
    vstrDestroy (aboutTitle);
    AboutMenuItem *aboutItem = tisAbout -> GetAboutMenuItem ();
    helpMenu -> AppendItem (aboutItem);
}


/*******************************************************************
*
*    alvuMainDialog refresh display function
*
*******************************************************************/

void alvuMainDialog::RefreshDisplay (void)
{
    vcolorSpec	*spec;
    vcolor	*fgcolor;
    double	red, green, blue;

    spec = new vcolorSpec;

    if (treeRoot->groupData.notRTNnotACK[MAJOR_ALARM])
    {
	GetRGBColor (CurrUnackSevColor [MAJOR_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	majorButton -> SetForeground (fgcolor);
    }
    else if (treeRoot->groupData.notRTNandACK[MAJOR_ALARM])
    {
	GetRGBColor (SevColor [MAJOR_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	majorButton -> SetForeground (fgcolor);
    }
    else if (treeRoot->groupData.RTNnotACK[MAJOR_ALARM])
    {
	GetRGBColor (CurrUnackSevColor [NO_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	majorButton -> SetForeground (fgcolor);
    }
    else
    {
	GetRGBColor (SevColor [NO_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	majorButton -> SetForeground (fgcolor);
    }

    if (treeRoot->groupData.notRTNnotACK[MINOR_ALARM])
    {
	GetRGBColor (CurrUnackSevColor [MINOR_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	minorButton -> SetForeground (fgcolor);
    }
    else if (treeRoot->groupData.notRTNandACK[MINOR_ALARM])
    {
	GetRGBColor (SevColor [MINOR_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	minorButton -> SetForeground (fgcolor);
    }
    else if (treeRoot->groupData.RTNnotACK[MINOR_ALARM])
    {
	GetRGBColor (CurrUnackSevColor [NO_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	minorButton -> SetForeground (fgcolor);
    }
    else
    {
	GetRGBColor (SevColor [NO_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	minorButton -> SetForeground (fgcolor);
    }

    if (treeRoot->groupData.notRTNnotACK[INFO_ALARM])
    {
	GetRGBColor (CurrUnackSevColor [INFO_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	infoButton -> SetForeground (fgcolor);
    }
    else if (treeRoot->groupData.notRTNandACK[INFO_ALARM])
    {
	GetRGBColor (SevColor [INFO_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	infoButton -> SetForeground (fgcolor);
    }
    else if (treeRoot->groupData.RTNnotACK[INFO_ALARM])
    {
	GetRGBColor (CurrUnackSevColor [NO_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	infoButton -> SetForeground (fgcolor);
    }
    else
    {
	GetRGBColor (SevColor [NO_ALARM], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	infoButton -> SetForeground (fgcolor);
    }

    delete spec;
}


/*******************************************************************
*
*    alvuMainDialog set colors function
*	get the background color and set default colors for
*	the MAJOR, MINOR and INFO buttons
*
*******************************************************************/

void alvuMainDialog::SetColors (void)
{
    vcolorSpec	*spec;
    vcolor	*color;
    double	red, green, blue;
    int		bgRed, bgGreen, bgBlue;

    color = majorButton -> DetermineBackground ();
    (spec = color -> GetSpec ()) -> Convert (vcolor::GetRGBModel());
    spec -> FGetRGB (&red, &green, &blue);
    backgroundVcolor = vcolor::Intern (spec);
    bgRed = (int)(red * 255);
    bgGreen = (int)(green * 255);
    bgBlue = (int)(blue * 255);
    backgroundColor = Color(bgRed, bgGreen, bgBlue);

    red = green = blue = 0.0;
    spec = new vcolorSpec;
    spec -> FSetRGB (red, green, blue);
    blackVcolor = vcolor::Intern (spec);
    delete spec;

    GetRGBColor (SevColor [NO_ALARM], &red, &green, &blue);
    spec = new vcolorSpec;
    spec -> FSetRGB (red, green, blue);
    color = vcolor::Intern (spec);

    majorButton -> SetForeground (color);
    minorButton -> SetForeground (color);
    infoButton -> SetForeground (color);

    delete spec;
}


/*******************************************************************
*
*    alvuMainDialog set Setup menu, filter menuitem
*
*******************************************************************/

void alvuMainDialog::SetFilterMenuTitle (const vname *menuTitle, 
					 char *filterName)
{
    vstr *formattedTitle = vresourceGetString (alvuApp -> titlesResource,
					       menuTitle);
    vstr *name  = vstrCloneScribed (vcharScribeSystem (filterName));
    vstr *filterTitle = vstrCloneScribed (vcharScribeFormat (formattedTitle,
						       name));
    selFilterMenuItem -> SetTitle (filterTitle);
    vstrDestroy (formattedTitle);
    vstrDestroy (name);
    vstrDestroy (filterTitle);
}


vkindTYPED_FULL_DEFINITION(alvuMainCommand, vcommandSelector, "alvuMainCommand");
/*******************************************************************
*
*    alvuMainCommand menu selector commands
*
*******************************************************************/

int alvuMainCommand::Issue (vdict *dict)
{
    int			command;
    alvuMainDialog	*mainDialog;
    Filter		tempFilter;
    PvData		*pvdata;
    TreeNode		*nodeptr;
    ComboDialog		*comboDialog;
    vchar		selectedItem[256];
    char		item[256];
    vstr		*strPtr;


    mainDialog = (alvuMainDialog *) dict -> GetElem (vname_Dialog);
    alvuApplication *alvuApp = mainDialog -> alvuApp;
    vresource alvuRes = alvuApp -> alvuResource;
    vresource titlesRes = alvuApp -> titlesResource;
    vresource comboRes = vresourceGet (alvuRes, alvu_ComboDialog);
    vresource alarmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);

    command = GetSelector ();

    switch (command)
    {
	// File menu items

	case SELECT_FILTER_MENUITEM:
	{
	    vresource comboRes = vresourceGet (alvuRes, alvu_ComboDialog);
	    ComboDialog *comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (alvuApp, BASE_FILTER_LIST,
				      GetFilterNames (), NULL);
	    vchar selectedFilter[AL_FILTERNAME_SIZE + 1];
	    vcharCopyBounded (comboDialog -> GetSelected (), 
			      selectedFilter, 
			      AL_FILTERNAME_SIZE);
	    selectedFilter [AL_FILTERNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedFilter[0])
	    {
		char selectedfilter [AL_FILTERNAME_SIZE + 1];
		vcharExportBounded (vcharScribeSystemExport (selectedFilter),
				    selectedfilter,
				    AL_FILTERNAME_SIZE);
		selectedfilter [AL_FILTERNAME_SIZE] = 0;
		GetFilter (selectedfilter, &alvuApp -> shellFilter);
	    }
	    break;
	}
	
	case NEW_FILTER_MENUITEM:
	{
	    vresource filtRes = vresourceGet (alvuRes, alvu_FilterDialog);
	    FilterEditDialog *filterEditDialog = new FilterEditDialog (filtRes);
	    filterEditDialog -> OpenFilter (alvuApp, "", NULL, NEW_FILTER);
	    if (filterEditDialog -> NewFilter (&tempFilter))
	    {
		SaveFilter (tempFilter.name, &tempFilter);
	    }
	    delete filterEditDialog;
	    break;
	}

	case EDIT_FILTER_MENUITEM:
	{
	    vresource comboRes = vresourceGet (alvuRes, alvu_ComboDialog);
	    ComboDialog *comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (alvuApp, EDIT_FILTER_LIST,
				      GetFilterNames (), NULL);
	    vchar selectedFilter[AL_FILTERNAME_SIZE + 1];
	    vcharCopyBounded (comboDialog -> GetSelected (), 
			      selectedFilter, 
			      AL_FILTERNAME_SIZE);
	    selectedFilter [AL_FILTERNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedFilter[0])
	    {
		char selectedfilter [AL_FILTERNAME_SIZE + 1];
		vcharExportBounded (vcharScribeSystemExport (selectedFilter),
				    selectedfilter,
				    AL_FILTERNAME_SIZE);
		selectedfilter [AL_FILTERNAME_SIZE] = 0;
		vresource filtRes = vresourceGet (alvuRes, alvu_FilterDialog);
		FilterEditDialog *filterEditDialog = new FilterEditDialog (filtRes);
		filterEditDialog -> OpenFilter (alvuApp, selectedfilter, NULL, EDIT_FILTER);
		if (filterEditDialog -> NewFilter (&tempFilter))
		{
		    SaveFilter (tempFilter.name, &tempFilter);
		}
		delete filterEditDialog;
	    }
	    break;
	}

	case DELETE_FILTER_MENUITEM:
	{
	    vresource comboRes = vresourceGet (alvuRes, alvu_ComboDialog);
	    ComboDialog *comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (alvuApp, DELETE_FILTER_LIST,
				      GetFilterNames (), NULL);
	    vchar selectedFilter[AL_FILTERNAME_SIZE + 1];
	    vcharCopyBounded (comboDialog -> GetSelected (), 
			      selectedFilter, 
			      AL_FILTERNAME_SIZE);
	    selectedFilter [AL_FILTERNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedFilter[0])
	    {
		char selectedfilter [AL_FILTERNAME_SIZE + 1];
		vcharExportBounded (vcharScribeSystemExport (selectedFilter),
				    selectedfilter,
				    AL_FILTERNAME_SIZE);
		selectedfilter [AL_FILTERNAME_SIZE] = 0;
		DeleteFilter (selectedfilter);
	    }
	    break;
	}

	case PRINT_MENUITEM:
	{
	    break;
	}

	case EXIT_MENUITEM:
	{
	    vevent::StopProcessing ();
	    break;
	}

	case GROUP_MENUITEM:
	{
	    vresource comboRes = vresourceGet (alvuRes, alvu_ComboDialog);
	    ComboDialog *comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (alvuApp, BROWSE_GROUP_LIST, groupNames, NULL);
	    vchar selectedGroup[AL_GROUPNAME_SIZE + 1];
	    vcharCopyBounded (comboDialog -> GetSelected (), 
			      selectedGroup, 
			      AL_GROUPNAME_SIZE);
	    selectedGroup [AL_GROUPNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedGroup[0])
	    {
		char groupname [AL_GROUPNAME_SIZE + 1];
		vcharExportBounded (vcharScribeSystemExport (selectedGroup),
				    groupname,
				    AL_GROUPNAME_SIZE);
		groupname [AL_GROUPNAME_SIZE] = 0;

		if (FindGroup (treeRoot, groupname))
		{
		    if (dynamicGroupDialogPtr != NULL)
		    {
			delete dynamicGroupDialogPtr;
		    }
		    tempFilter = alvuApp -> shellFilter;
		    strncpy (tempFilter.group, groupname, sizeof (tempFilter.group));
		    tempFilter.group [sizeof (tempFilter.group) - 1] = 0;
		    vresource alarmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
		    dynamicGroupDialogPtr = new DynamicGroupDialog (alarmDlgRes);
		    dynamicGroupDialogPtr -> AlarmDialogOpen (alvuApp, &tempFilter);
		}
		else
		{
		    alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE,
					     alvu_GroupNotExistMsg, selectedGroup);
		}
	    }
	    
	    break;
	}

	case ALARM_WINDOW_MENUITEM:
	    if (alarmWindowDialogPtr != NULL)
	    {
		delete alarmWindowDialogPtr;
	    }
	    alarmWindowDialogPtr = new AlarmWindowDialog (alarmDlgRes);
	    alarmWindowDialogPtr -> AlarmDialogOpen (alvuApp, &alvuApp -> shellFilter);
	    break;

	case ALARM_SUMMARY_MENUITEM:
	    if (alarmSummaryDialogPtr != NULL)
	    {
		delete alarmSummaryDialogPtr;
	    }
	    alarmSummaryDialogPtr = new AlarmSummaryDialog (alarmDlgRes);
	    alarmSummaryDialogPtr -> AlarmDialogOpen (alvuApp, &alvuApp -> shellFilter);

	    break;

	case ALARM_HISTORY_MENUITEM:
	    if (alarmHistoryDialogPtr == NULL)
	    {
		tempFilter = alvuApp -> shellFilter;
		tempFilter.sev |= EVENTFILTER;
		tempFilter.modified = True;
		alarmHistoryDialogPtr = new AlarmHistoryDialog (alarmDlgRes);
		alarmHistoryDialogPtr -> AlarmDialogOpen (alvuApp, &tempFilter);
	    }
	    else
	    {
		if (alarmHistoryDialogPtr -> IsIconified ())
		{
		    alarmHistoryDialogPtr -> Uniconify ();
		    alarmHistoryDialogPtr -> Place (vwindow::GetRoot (),
						    vrectPLACE_CENTER,
						    vrectPLACE_CENTER);				
		}
		else
		{
		    delete alarmHistoryDialogPtr;
		    tempFilter = alvuApp -> shellFilter;
		    tempFilter.sev |= EVENTFILTER;
		    tempFilter.modified = True;
		    alarmHistoryDialogPtr = new AlarmHistoryDialog (alarmDlgRes);
		    alarmHistoryDialogPtr -> AlarmDialogOpen (alvuApp, &tempFilter);
	        }
	    }
	    break;

	case EVENT_HISTORY_MENUITEM:
	    if (alarmEventDialogPtr == NULL)
	    {
		tempFilter = alvuApp -> shellFilter;
		tempFilter.sev = NOALARMFILTER | INFOFILTER | MINORFILTER | MAJORFILTER;
		tempFilter.modified = True;
		alarmEventDialogPtr = new AlarmEventDialog (alarmDlgRes);
		alarmEventDialogPtr -> AlarmDialogOpen (alvuApp, &tempFilter);
	    }
	    else
	    {
		if (alarmEventDialogPtr -> IsIconified ())
		{
		    alarmEventDialogPtr -> Uniconify ();
		    alarmEventDialogPtr -> Place (vwindow::GetRoot (),
					          vrectPLACE_CENTER,
					          vrectPLACE_CENTER);				
		}
		else
		{
		    delete alarmEventDialogPtr;
		    tempFilter = alvuApp -> shellFilter;
		    tempFilter.sev |= EVENTFILTER;
		    tempFilter.modified = True;
		    alarmEventDialogPtr = new AlarmEventDialog (alarmDlgRes);
		    alarmEventDialogPtr -> AlarmDialogOpen (alvuApp, &tempFilter);
	        }
	    }
	    break;

	case OPER_TRACK_MENUITEM:
	    if (alarmTrackDialogPtr == NULL)
	    {
		alarmTrackDialogPtr = new AlarmTrackDialog (alarmDlgRes);
		alarmTrackDialogPtr -> AlarmDialogOpen (alvuApp, &alvuApp -> shellFilter);
	    }
	    else
	    {
		if (alarmTrackDialogPtr -> IsIconified ())
		{
		    alarmTrackDialogPtr -> Uniconify ();
		    alarmTrackDialogPtr -> Place (vwindow::GetRoot (),
						  vrectPLACE_CENTER,
						  vrectPLACE_CENTER);				
		}
		else
		{
		    delete alarmTrackDialogPtr;
		    alarmTrackDialogPtr = new AlarmTrackDialog (alarmDlgRes);
		    alarmTrackDialogPtr -> AlarmDialogOpen (alvuApp, &alvuApp -> shellFilter);
	        }
	    }
	    break;

	case REPORTS_MENUITEM:
	    if (reportsDialogPtr == NULL)
	    {
		reportsDialogPtr = new AlarmPointsDialog (alarmDlgRes);
		reportsDialogPtr -> AlarmDialogOpen (alvuApp, &alvuApp -> shellFilter);
	    }
	    else
	    {
		if (reportsDialogPtr -> IsIconified ())
		{
		    reportsDialogPtr -> Uniconify ();
		    reportsDialogPtr -> Place (vwindow::GetRoot (),
					       vrectPLACE_CENTER,
					       vrectPLACE_CENTER);				
		}
		else
		{
		    delete reportsDialogPtr;
		    reportsDialogPtr = new AlarmPointsDialog (alarmDlgRes);
		    reportsDialogPtr -> AlarmDialogOpen (alvuApp, &alvuApp -> shellFilter);
	        }
	    }
	    break;

	case SYSTEM_HISTORY_MENUITEM:
	{
	    if (systemDialogPtr == NULL)
	    {
		systemDialogPtr = new AlarmSystemDialog (alarmDlgRes);
		systemDialogPtr -> AlarmDialogOpen (alvuApp, &alvuApp -> shellFilter);
	    }
	    else
	    {
		if (systemDialogPtr -> IsIconified ())
		{
		    systemDialogPtr -> Uniconify ();
		    systemDialogPtr -> Place (vwindow::GetRoot (),
					      vrectPLACE_CENTER,
					      vrectPLACE_CENTER);				
		}
		else
		{
		    delete systemDialogPtr;
		    systemDialogPtr = new AlarmSystemDialog (alarmDlgRes);
		    systemDialogPtr -> AlarmDialogOpen (alvuApp, &alvuApp -> shellFilter);
	        }
	    }
	    break;
	}

	case DISABLE_PT_MENUITEM:
	    comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (alvuApp, DISABLE_POINT_LIST, channelNames, NULL);
	    vcharCopyBounded (comboDialog -> GetSelected (), selectedItem, 
			      AL_POINTNAME_SIZE);
	    selectedItem [AL_POINTNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedItem[0])
	    {
		vcharExportBounded (vcharScribeSystemExport (selectedItem),
				    item,
				    AL_POINTNAME_SIZE);
		item [AL_POINTNAME_SIZE] = 0;
		if (pvdata = FindChannel (treeRoot, item))
		{
		    if (DisablePoint (pvdata) != True)
		    {
			strPtr = vstrCloneScribed (vcharScribeSystem (item));
			alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE,
						 alvu_NoDisablePriv, strPtr);
			vstrDestroy (strPtr);
		    }
		}
	    }
	    break;

	case ENABLE_PT_MENUITEM:
	    comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (alvuApp, ENABLE_POINT_LIST, channelNames, NULL);
	    vcharCopyBounded (comboDialog -> GetSelected (), selectedItem, 
			      AL_POINTNAME_SIZE);
	    selectedItem [AL_POINTNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedItem[0])
	    {
		vcharExportBounded (vcharScribeSystemExport (selectedItem),
				    item,
				    AL_POINTNAME_SIZE);
		item [AL_POINTNAME_SIZE] = 0;
		if (pvdata = FindChannel (treeRoot, item))
		{
		    if (EnablePoint (pvdata) != True)
		    {
			strPtr = vstrCloneScribed (vcharScribeSystem (item));
			alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE,
						 alvu_NoEnablePriv, strPtr);
			vstrDestroy (strPtr);
		    }
		}
	    }
	    break;

	case DISABLE_GRP_MENUITEM:
	    comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (alvuApp, DISABLE_GROUP_LIST, 
				      groupNames, 
				      NULL);
	    vcharCopyBounded (comboDialog -> GetSelected (), selectedItem, 
			      AL_GROUPNAME_SIZE);
	    selectedItem [AL_GROUPNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedItem[0])
	    {
		vcharExportBounded (vcharScribeSystemExport (selectedItem),
				    item,
				    AL_GROUPNAME_SIZE);
		item [AL_GROUPNAME_SIZE] = 0;
		if (nodeptr = FindGroup (treeRoot, item))
		{
		    DisableGroups (nodeptr);
		}
	    }
	    break;

	case ENABLE_GRP_MENUITEM:
	    comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (alvuApp, ENABLE_GROUP_LIST, 
				      groupNames,
				      NULL);
	    vcharCopyBounded (comboDialog -> GetSelected (), selectedItem, 
			      AL_GROUPNAME_SIZE);
	    selectedItem [AL_GROUPNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedItem[0])
	    {
		vcharExportBounded (vcharScribeSystemExport (selectedItem),
				    item,
				    AL_GROUPNAME_SIZE);
		item [AL_GROUPNAME_SIZE] = 0;
		if (nodeptr = FindGroup (treeRoot, item))
		{
		    EnableGroups (nodeptr);
		}
	    }
	    break;

	case USE_DEF_MENUITEM:
	{
	    strcpy (linePrinter, "");
	    strcpy (defFilterName, ALVU_DEFAULT_FILTER_NAME);
	    strcpy (alvuApp -> shellFilter.name, ALVU_DEFAULT_FILTER_NAME);
	    mainDialog -> SetFilterMenuTitle (alvu_DefaultFilterTitle,
					      defFilterName);
	    strcpy (alvuApp -> shellFilter.group, "");
	    alvuApp -> shellFilter.ack = 0;
	    alvuApp -> shellFilter.sev = 0;
	    alvuApp -> shellFilter.alarmOnly = ALLFILTER;
	    strcpy (alvuApp -> shellFilter.start, "");
	    strcpy (alvuApp -> shellFilter.end, "");
	    strcpy (alvuApp -> shellFilter.op, "");
	    alvuApp -> shellFilter.modified = vFALSE;
	    winEnable = vFALSE;
	    mainDialog -> setupAlWin -> SetOn (vFALSE);
	    popupAlrmEnable = vTRUE;
	    mainDialog -> setupPopAl -> SetOn (vTRUE);
	    // reset alarm colors to defaults
	    int i;
	    for (i = 0; i < NUM_SEV_COLORS; i++)
	    {
		SevColor [i] = defaultAlarmColors [i];
		UnackSevColor [i] = defaultAlarmColors [i];
	    }
	    UpdateBlinkColors ();   // see alvuUtils.cxx
	    break;
	}

	case SAVE_MENUITEM:
	{
	    FILE *fd;
	    vstr  *name;

	    if (fd = fopen (alvuApp -> setupFileName, "w"))
	    {
		unsigned char alarmColors[NUM_SEV_COLORS][3];
		unsigned int  magicNumber = ALARMVU_MAGIC_NUMBER;
		unsigned int  version = ALVU_NEW_VERSION;
		int i;

		for (i = 0; i < NUM_SEV_COLORS; i++)
		{
		    alarmColors [i][0] = (char)(Red (SevColor [i]));
		    alarmColors [i][1] = (char)(Green (SevColor [i]));
		    alarmColors [i][2] = (char)(Blue (SevColor [i]));
		}
		if (fwrite ((char *)&magicNumber, sizeof (magicNumber), 1, fd) != 1 ||
		    fwrite ((char *)&version, sizeof (version), 1, fd) != 1 ||
		    fwrite (linePrinter, sizeof (linePrinter), 1, fd) != 1 ||
		    fwrite ((char *)&alarmColors, sizeof (alarmColors), 1, fd) != 1 ||
		    fwrite (defFilterName, sizeof (defFilterName), 1, fd) != 1 ||
		    fwrite ((char *)&winEnable, sizeof (winEnable), 1, fd) != 1 ||
		    fwrite ((char *)&popupAlrmEnable, sizeof (popupAlrmEnable), 1, fd) != 1)
		{
		    name = vstrCloneScribed (vcharScribeSystem (alvuApp -> setupFileName));
		    alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE,
					     alvu_SetupFileWriteErr, name);
		    vstrDestroy (name);
		}
		fclose (fd);
	    }
	    else
	    {
		name = vstrCloneScribed (vcharScribeSystem (alvuApp -> setupFileName));
		alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE,
					 alvu_SetupFileOpenErr, name);
		vstrDestroy (name);
	    }
	    break;
	}

	case PRINTER_MENUITEM:
	{
	    vresource oneLineRes = vresourceGet (alvuRes, alvu_OneLineDialog);
	    OneLineDialog *oneLineDlg = new OneLineDialog (oneLineRes);
	    oneLineDlg -> OpenOneLine (alvuApp, ONELINE_PRINT_DEVICE);
	    vchar *oneLineText = oneLineDlg -> GetOneLineText ();
	    if (vcharLength (oneLineText) != 0)
	    {
		vcharExportBounded (vcharScribeSystemExport (oneLineText),
				    linePrinter,
				    sizeof (linePrinter));
		linePrinter [sizeof (linePrinter) - 1] = 0;
	    }
	    delete oneLineDlg;
	    break;
	}

	case COLORS_MENUITEM:
	{
	    vresource colorDlgRes = vresourceGet (alvuRes, alvu_ColorChooserDialog);
	    ColorDialog *colorDlg = new ColorDialog (colorDlgRes);
	    colorDlg -> OpenColor (alvuApp, SevColor);
	    colorDlg -> GetNewColors (SevColor);
	    UpdateBlinkColors ();
	    delete colorDlg;
	    break;
	}

	case STARTUP_FILTER_MENUITEM:
	    comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (alvuApp, STARTUP_FILTER_LIST, 
				      GetFilterNames(),
				      NULL);
	    vcharCopyBounded (comboDialog -> GetSelected (), selectedItem, 
			      AL_FILTERNAME_SIZE);
	    selectedItem [AL_FILTERNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedItem[0])
	    {
		vcharExportBounded (vcharScribeSystemExport (selectedItem),
				    item,
				    AL_FILTERNAME_SIZE);
		item [AL_FILTERNAME_SIZE] = 0;
		strncpy (defFilterName, item, AL_FILTERNAME_SIZE);
		defFilterName [AL_FILTERNAME_SIZE] = 0;
		mainDialog -> SetFilterMenuTitle (alvu_StartUpFilterTitle,
						  defFilterName);
	    }
	    break;

	case SETUP_ALARMWINDOW_MENUITEM:
	    if (mainDialog -> setupAlWin -> IsOn () == vTRUE)
	    {
		winEnable = vTRUE;
		if (alarmWindowDialogPtr == NULL)
		{
		    alarmWindowDialogPtr = new AlarmWindowDialog ();
		}
	    }
	    else
	    {
		winEnable = vFALSE;
	    }
	    break;

	case POPUP_MENUITEM:
	    if (mainDialog -> setupPopAl -> IsOn () == vTRUE)
	    {
		popupAlrmEnable = vTRUE;
	    }
	    else
	    {
		popupAlrmEnable = vFALSE;
		if (popAlarmPtr != NULL)
		{
		    delete popAlarmPtr;
		    popAlarmPtr = NULL;
		}
	    }
	    break;

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}
  

vkindTYPED_FULL_DEFINITION(mainDialogSilence, vbutton, "mainDialogSilence");

/*******************************************************************
*
*    mainDialog Silence button Observer
*
*******************************************************************/

void mainDialogSilence::ObserveDialogItem (vevent *event)
{
    // see bell.c
    BellTurnOff ();
}


vkindTYPED_FULL_DEFINITION(mainDialogSummary, vbutton, "mainDialogSummary");

/*******************************************************************
*
*    mainDialogSummary Observer
*
*******************************************************************/

void mainDialogSummary::ObserveDialogItem (vevent *event)
{
    alvuApplication *alvuApp = mainDialog -> alvuApp;
    vresource alvuRes = alvuApp -> alvuResource;
    vresource titlesRes = alvuApp -> titlesResource;

    vstr *summaryStr = vresourceGetString (titlesRes, alvu_SummaryFilterName);
    char summary [AL_FILTERNAME_SIZE + 1];
    vcharExportBounded (vcharScribeSystemExport (summaryStr), summary, AL_FILTERNAME_SIZE);
    summary [AL_FILTERNAME_SIZE] = 0;
    vstrDestroy (summaryStr);

    Filter tempFilter = alvuApp -> shellFilter;
    strncpy (tempFilter.name, summary, sizeof (tempFilter.name));
    tempFilter.name [sizeof (tempFilter.name) - 1] = 0;
    strncpy (tempFilter.group, ALVU_MAINS_GROUPNAME, sizeof (tempFilter.group));
    tempFilter.group [sizeof (tempFilter.group) - 1] = 0;
    tempFilter.alarmOnly = ALLFILTER;
    tempFilter.sev = 0;
    memset ((void *)tempFilter.start, 0, sizeof (tempFilter.start));
    memset ((void *)tempFilter.end, 0, sizeof (tempFilter.end));
    tempFilter.modified = vFALSE;
    memset ((void *)tempFilter.op, 0, sizeof (tempFilter.op));

    if (alarmSummaryDialogPtr != NULL)
    {
	delete alarmSummaryDialogPtr;
    }
    vresource alarmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
    alarmSummaryDialogPtr = new AlarmSummaryDialog (alarmDlgRes);

    alarmSummaryDialogPtr -> AlarmDialogOpen (alvuApp, &tempFilter);
}
  

vkindTYPED_FULL_DEFINITION(mainDialogExit, vbutton, "mainDialogExit");

/*******************************************************************
*
*    mainDialogExit Observer
*
*******************************************************************/

void mainDialogExit::ObserveDialogItem (vevent *event)
{
    vevent::StopProcessing ();
}
  

vkindTYPED_FULL_DEFINITION(mainDialogMajor, vbutton, "mainDialogMajor");

/*******************************************************************
*
*    mainDialogMajor Observer
*
*******************************************************************/

void mainDialogMajor::ObserveDialogItem (vevent *event)
{
    alvuApplication *alvuApp = mainDialog -> alvuApp;
    vresource alvuRes = alvuApp -> alvuResource;
    vresource titlesRes = alvuApp -> titlesResource;

    vstr *majorStr = vresourceGetString (titlesRes, alvu_MajorFilterName);
    char major [AL_FILTERNAME_SIZE + 1];
    vcharExportBounded (vcharScribeSystemExport (majorStr), major, AL_FILTERNAME_SIZE);
    major [AL_FILTERNAME_SIZE] = 0;
    vstrDestroy (majorStr);

    Filter tempFilter = alvuApp -> shellFilter;
    strncpy (tempFilter.name, major, sizeof (tempFilter.name));
    tempFilter.name [sizeof (tempFilter.name) - 1] = 0;
    strncpy (tempFilter.group, ALVU_MAINS_GROUPNAME, sizeof (tempFilter.group));
    tempFilter.group [sizeof (tempFilter.group) - 1] = 0;
    tempFilter.alarmOnly = ALLFILTER;
    tempFilter.sev = MINORFILTER | INFOFILTER | EVENTFILTER;
    memset ((void *)tempFilter.start, 0, sizeof (tempFilter.start));
    memset ((void *)tempFilter.end, 0, sizeof (tempFilter.end));
    tempFilter.modified = vFALSE;
    memset ((void *)tempFilter.op, 0, sizeof (tempFilter.op));

    if (alarmSummaryDialogPtr != NULL)
    {
	delete alarmSummaryDialogPtr;
    }
    vresource alarmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
    alarmSummaryDialogPtr = new AlarmSummaryDialog (alarmDlgRes);

    alarmSummaryDialogPtr -> AlarmDialogOpen (alvuApp, &tempFilter);
}
  

vkindTYPED_FULL_DEFINITION(mainDialogMinor, vbutton, "mainDialogMinor");

/*******************************************************************
*
*    mainDialogMinor Observer
*
*******************************************************************/

void mainDialogMinor::ObserveDialogItem (vevent *event)
{
    alvuApplication *alvuApp = mainDialog -> alvuApp;
    vresource alvuRes = alvuApp -> alvuResource;
    vresource titlesRes = alvuApp -> titlesResource;

    vstr *minorStr = vresourceGetString (titlesRes, alvu_MinorFilterName);
    char minor [AL_FILTERNAME_SIZE + 1];
    vcharExportBounded (vcharScribeSystemExport (minorStr), minor, AL_FILTERNAME_SIZE);
    minor [AL_FILTERNAME_SIZE] = 0;
    vstrDestroy (minorStr);

    Filter tempFilter = alvuApp -> shellFilter;
    strncpy (tempFilter.name, minor, sizeof (tempFilter.name));
    tempFilter.name [sizeof (tempFilter.name) - 1] = 0;
    strncpy (tempFilter.group, ALVU_MAINS_GROUPNAME, sizeof (tempFilter.group));
    tempFilter.group [sizeof (tempFilter.group) - 1] = 0;
    tempFilter.alarmOnly = ALLFILTER;
    tempFilter.sev = MAJORFILTER | INFOFILTER | EVENTFILTER;
    memset ((void *)tempFilter.start, 0, sizeof (tempFilter.start));
    memset ((void *)tempFilter.end, 0, sizeof (tempFilter.end));
    tempFilter.modified = vFALSE;
    memset ((void *)tempFilter.op, 0, sizeof (tempFilter.op));

    if (alarmSummaryDialogPtr != NULL)
    {
	delete alarmSummaryDialogPtr;
    }
    vresource alarmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
    alarmSummaryDialogPtr = new AlarmSummaryDialog (alarmDlgRes);

    alarmSummaryDialogPtr -> AlarmDialogOpen (alvuApp, &tempFilter);
}
  

vkindTYPED_FULL_DEFINITION(mainDialogInfo, vbutton, "mainDialogInfo");

/*******************************************************************
*
*    mainDialogInfo Observer
*
*******************************************************************/

void mainDialogInfo::ObserveDialogItem (vevent *event)
{
    alvuApplication *alvuApp = mainDialog -> alvuApp;
    vresource alvuRes = alvuApp -> alvuResource;
    vresource titlesRes = alvuApp -> titlesResource;

    vstr *infoStr = vresourceGetString (titlesRes, alvu_InfoFilterName);
    char info [AL_FILTERNAME_SIZE + 1];
    vcharExportBounded (vcharScribeSystemExport (infoStr), info, AL_FILTERNAME_SIZE);
    info [AL_FILTERNAME_SIZE] = 0;
    vstrDestroy (infoStr);

    Filter tempFilter = alvuApp -> shellFilter;
    strncpy (tempFilter.name, info, sizeof (tempFilter.name));
    tempFilter.name [sizeof (tempFilter.name) - 1] = 0;
    strncpy (tempFilter.group, ALVU_MAINS_GROUPNAME, sizeof (tempFilter.group));
    tempFilter.group [sizeof (tempFilter.group) - 1] = 0;
    tempFilter.alarmOnly = ALLFILTER;
    tempFilter.sev = MAJORFILTER | MINORFILTER | EVENTFILTER;
    memset ((void *)tempFilter.start, 0, sizeof (tempFilter.start));
    memset ((void *)tempFilter.end, 0, sizeof (tempFilter.end));
    tempFilter.modified = vFALSE;
    memset ((void *)tempFilter.op, 0, sizeof (tempFilter.op));

    if (alarmSummaryDialogPtr != NULL)
    {
	delete alarmSummaryDialogPtr;
    }
    vresource alarmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
    alarmSummaryDialogPtr = new AlarmSummaryDialog (alarmDlgRes);

    alarmSummaryDialogPtr -> AlarmDialogOpen (alvuApp, &tempFilter);
}


static void observeHelpMenu (vmenuItem *item, vevent *event)
{
    helpMenuItem *menuItem = (helpMenuItem *)item;
    vmenuItemList *menuItemList = menuItem -> GetItemListIn ();
    int	index = menuItemList -> GetIndex ((vobjectAttributed *)menuItem);
    alvuMainDialog *mainDialog = menuItem -> GetMainDialog ();
    char **helpItems = mainDialog -> GetHelpItems ();

    printf ("Help menu index = %d\n", index);

    //selHelpItem ("alarmvu", "AlarmVu", helpItem [index]);
}
  

/*******************************************************************
*
*    stuff used by ObsereTimer
*
*******************************************************************/
static char		ProcMsgs[MAX_MSG_SIZE];
static int		rxLen = 0;
static unsigned int	inEvent = vFALSE;
static unsigned int	timeCount = 0;
static unsigned int	updateCount = 1;
static unsigned int	popAlarm = vFALSE;
static unsigned int	popAlarmCount = 0;
static unsigned int	numAlarmLogRecords = 0;
static unsigned int	blinkState = 0;
static unsigned int	deferredUpdate = 0;
static time_t		lastUpdate;
#define UPDATEDELAY	2

AckGroupPvPtrs	*ackLists = NULL;
static char	*sendMsg;
static char	*sendMsgResidue;
static int	sendMsgLength;
static AlarmMsg	controlMsg;
alvuProgressDialog	*inProgressPtr = NULL;
int		caUpdateReq = vFALSE;
long		tickVal;

/*******************************************************************
*
*    mainDialogTimer Observer
*
*******************************************************************/

void mainDialogTimer::ObserveTimer (void)
{
    int		msgStatus;
    int		update = vFALSE;
    TreeNode	*groupPtr;
    TreeNode	*savgroupPtr;
    TreeNode	*highestgroupPtr;
    int		bytesSent;
    PvData	*pvptr;
    TreeNode	*group;
    AckGroupPvPtrs *agptr;
    TypeOnlyMessage *alarmMsg;
    TypeOnlyMessage controlMsg;
    unsigned int rcvSize;
    char	testchannel [AL_POINTNAME_SIZE + 5];
    int		newAlarms = vFALSE;
    int		privStat;
    int		msgsSent;
    alvuApplication *alvuApp = mainDialog -> alvuApp;

    ca_pend_io ((float)0.0); /* call CA, to allow callbacks */

    /* run no more than once in a second */
    if (tickVal == UnixTime(NULL))
	return;
    tickVal = UnixTime(NULL);

    alarmMsg = (TypeOnlyMessage *)ProcMsgs;

    if (inEvent == vFALSE)
    {
	inEvent = vTRUE;

	/* send any message pieces */
	if (sendMsgResidue)
	{
	    bytesSent = send (alvuApp -> alarmhndSocket, sendMsgResidue, sendMsgLength, 0);
	    if (bytesSent < 0)
	    {
		/* socket failure */
	    }
	    else
	    {
		sendMsgResidue += bytesSent;
		sendMsgLength -= bytesSent;
		if (sendMsgLength == 0)
		{
		    free (sendMsg);
		    sendMsgResidue = 0;
		}
	    }
	}
	msgsSent = 0;
	while (ackLists && (sendMsgResidue == 0) && msgsSent < MaxSendsPerSecond)
	{
	    /* send out complete messages */
	    privStat = PRIVOK; /* defaults to priv OK */
	    switch (ackLists->msgtype [ackLists->current])
	    {
		case ACK_MSG:
		case ENABLE_MSG:
		case DISABLE_MSG:
		    pvptr = (PvData *)ackLists->pvptrs [ackLists->current];
		    switch (ackLists->msgtype [ackLists->current])
		    {
			case ACK_MSG:
			    strncpy (testchannel, pvptr->pvName, AL_POINTNAME_SIZE);
			    strcat (testchannel, ".ACK");
			    privStat = CheckPriv (testchannel);
			    // fixup ambiguous deleted point status!
			    if (privStat == PRIVDBERR)
				privStat = PRIVOK;
			    break;
			case ENABLE_MSG:
			    strncpy (testchannel, pvptr->pvName, AL_POINTNAME_SIZE);
			    strcat (testchannel, ".AENB");
			    privStat = CheckTrackPriv (testchannel, "NO ", "YES");
			    break;
			case DISABLE_MSG:
			    strncpy (testchannel, pvptr->pvName, AL_POINTNAME_SIZE);
			    strcat (testchannel, ".AENB");
			    privStat = CheckTrackPriv (testchannel, "YES", "NO ");
			    break;
		    }
		    if (privStat != PRIVOK)
		    {
			break;
		    }
		    sendMsg = (char *)malloc (sizeof (AlarmMsg));
		    if (sendMsg == 0)
		    {
			/* unable to allocate memory */
			break;
		    }
		    memset (sendMsg, 0, sizeof (AlarmMsg));
		    ((AlarmMsg *)sendMsg)->msgType = ackLists->msgtype [ackLists->current];
		    strcpy (((AlarmMsg *)sendMsg)->pointName, pvptr->pvName);
		    GetOperName (((AlarmMsg *)sendMsg)->name.operatorName);
		    ((AlarmMsg *)sendMsg)->timeStamp = pvptr->lastAlarm;
		    sendMsgLength = sizeof (AlarmMsg);
		    break;
		case ACK_GROUP_MSG:
		    sendMsg = (char *)malloc (sizeof (AlarmMsg));
		    if (sendMsg == 0)
		    {
			/* unable to allocate memory */
			break;
		    }
		    memset (sendMsg, 0, sizeof (AlarmMsg));
		    group = (TreeNode *)ackLists->pvptrs [ackLists->current];
		    ((AlarmMsg *)sendMsg)->msgType = ACK_GROUP_MSG;
		    strcpy (((AlarmMsg *)sendMsg)->pointName, group->groupName);
		    GetOperName (((AlarmMsg *)sendMsg)->name.operatorName);
		    sendMsgLength = sizeof (AlarmMsg);
		    break;
		case ALARM_COMMENT_UPDATE:
		    sendMsg = (char *)ackLists->pvptrs [ackLists->current];
		    sendMsgLength = sizeof (AlarmMemo);
		    break;
	    }
	    if (privStat != PRIVOK)
	    {
		ackLists->current++;
		if (ackLists->current == ackLists->total)
		{
		    agptr = ackLists;
		    ackLists = (AckGroupPvPtrs *)ackLists->link;
		    free ((char *)agptr);
		}
	    }
	    else if (sendMsg)
	    {
		msgsSent++;
		ackLists->current++;
		if (ackLists->current == ackLists->total)
		{
		    agptr = ackLists;
		    ackLists = (AckGroupPvPtrs *)ackLists->link;
		    free ((char *)agptr);
		}
		bytesSent = send (alvuApp -> alarmhndSocket, sendMsg, sendMsgLength, 0);
		if (bytesSent < 0)
		{
		    /* socket failure!!! */
		}
		else if (bytesSent < sendMsgLength)
		{
		    sendMsgResidue = sendMsg + bytesSent;
		    sendMsgLength = sendMsgLength - bytesSent;
		    break;
		}
		else
		{
		    free (sendMsg);
		}
	    }
	    else
	    {
		break; /* no memory for now */
	    }
	}
	if (ackLists)
	{  // update progress counter (reuse bytesSent var as msgs left)
	    bytesSent = 0;
	    for (agptr = ackLists; agptr; agptr = ((AckGroupPvPtrs *)agptr->link))
		bytesSent = bytesSent + (agptr->total - agptr->current);
	    if (inProgressPtr)
	    {
		if (inProgressPtr -> IsOpen ())
		{
		    inProgressPtr -> UpdateCount (bytesSent);
		}
		else
		{
		    delete inProgressPtr;
		    inProgressPtr = NULL;
		}
	    }
	    else
	    {
		vresource inprgres = vresourceGet (mainDialog -> alvuApp -> alvuResource, alvu_ProgressDialog);
		inProgressPtr = new alvuProgressDialog (inprgres);
		inProgressPtr -> OpenProgress (bytesSent);
	    }
	}
	else if (inProgressPtr)
	{
	    delete inProgressPtr;
	    inProgressPtr = NULL;
	}

	// get messages from socket.
	rcvSize = sizeof(TypeOnlyMessage);

	for ( ; ; )
	{
	    ca_pend_io ((float)0.0);
	    // read as much as possible this gulp.
 	    msgStatus = TCPRecv (alvuApp -> alarmhndSocket, &ProcMsgs [rxLen], rcvSize - rxLen);
	    if (msgStatus == 0)
	    {
		alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_MgrConnClosed);
		vevent::StopProcessing ();
	    }
	    else if (msgStatus == -1)
	    {
#ifdef WIN32
		if (errno != WSAEWOULDBLOCK)
#else
		if (errno != EWOULDBLOCK)
#endif WIN32
		{
		    alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_MgrRecvErr, errno);
		    vevent::StopProcessing ();
		}
		// adjust length to indicate no data from this read
		msgStatus = 0;
	    }
	    rxLen += msgStatus;

	    // These check for not enough data to process
	    if ((unsigned int)rxLen < rcvSize)
	    {
		break;
	    }

	    if (alarmMsg->msgType == UPDATE)
	    {
		update = vTRUE;
		rxLen -= sizeof(TypeOnlyMessage);
	    }
	    else if (alarmMsg->msgType == UPDATE_NEW_ALARM)
	    {
		update = vTRUE;
		newAlarms = vTRUE;
		rxLen -= sizeof(TypeOnlyMessage);
	    }
	    else if (alarmMsg->msgType == UNAVAILABLE)
	    {
		// Remove all displays
		if (alarmSummaryDialogPtr)
		{
		    delete alarmSummaryDialogPtr;
		    alarmSummaryDialogPtr = NULL;
		}
		if (dynamicGroupDialogPtr)
		{
		    delete dynamicGroupDialogPtr;
		    dynamicGroupDialogPtr = NULL;
		}
		if (alarmHistoryDialogPtr)
		{
		    delete alarmHistoryDialogPtr;
		    alarmHistoryDialogPtr = NULL;
		}
		if (alarmEventDialogPtr)
		{
		    delete alarmEventDialogPtr;
		    alarmEventDialogPtr = NULL;
		}
		if (alarmWindowDialogPtr)
		{
		    delete alarmWindowDialogPtr;
		    alarmWindowDialogPtr = NULL;
		}
		if (popAlarmPtr)
		{
		    delete popAlarmPtr;
		    popAlarmPtr = NULL;
		}
		if (reportsDialogPtr)
		{
		    delete reportsDialogPtr;
		    reportsDialogPtr = NULL;
		}
		alvuApp -> UnMapSharedMemory ();
		rxLen -= sizeof(TypeOnlyMessage);
		controlMsg.msgType = UNAVAILABLE;
		if (send (alvuApp -> alarmhndSocket, (char *)&controlMsg, sizeof (controlMsg), 0) == -1)
		{
		    alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_SendUnavailMsg);
		}
		// process messages until memory is available again.
		while (shareHeader == NULL)
		{
		    alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_AlarmMemReconfig);
		    // read as much as possible this gulp.
	 	    msgStatus = TCPRecv (alvuApp -> alarmhndSocket, &ProcMsgs [rxLen], rcvSize - rxLen);
		    if (msgStatus == 0)
		    {
			alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_MgrConnClosed);
			vevent::StopProcessing ();
		    }
		    else if (msgStatus == -1)
		    {
#ifdef WIN32
			if (errno != WSAEWOULDBLOCK)
#else
			if (errno != EWOULDBLOCK)
#endif WIN32
			{
			    alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_MgrRecvErr, errno);
			    vevent::StopProcessing ();
			}
			// adjust length to indicate no data from this read
			msgStatus = 0;
		    }
		    rxLen += msgStatus;

		    // These check for not enough data to process
		    if ((unsigned int)rxLen < rcvSize)
		    {
			continue;
		    }

		    if (alarmMsg->msgType == AVAILABLE)
		    {
			alvuApp -> MapSharedMemory ();
			rxLen -= sizeof(TypeOnlyMessage);
		    }
		}
		free ((void *)groupPtrs);
		free ((void *)groupNames);
		groupNames = GetGroups (&groupPtrs);
		free ((void *)channelPtrs);
		free ((void *)channelNames);
		channelNames = GetChannels (&channelPtrs);
	    }
	    else if (alarmMsg->msgType == AVAILABLE)
	    {
		// ignore extras!!
	    }
	    else
	    {
		// Illegal op value!!
		alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_UnkMgrMsg, alarmMsg -> msgType);
		vevent::StopProcessing ();
	    }
	}

	ca_pend_io ((float)0.0);

	long *currColors;
	if (blinkState ^= 1)
	{
	    currColors = &UnackSevColor[0];
	}
	else
	{
	    currColors = &SevColor[0];
	}
	int i;
	for (i = 0; i < NUM_SEV_COLORS; i++)
	{
	    CurrUnackSevColor [i] = currColors[i];
	}

	highestgroupPtr = NULL;
	if (HighestSeverity (treeRoot->groupData.numUnackedBySeverity))
	{
	    highestgroupPtr = treeRoot;
	}
	if (highestgroupPtr)
	{
	    savgroupPtr = NULL;
	    groupPtr = treeRoot;
	    while (groupPtr)
	    {
		if (HighestSeverity (groupPtr->groupData.numUnackedBySeverity))
		{
		    if (savgroupPtr)
		    {
			savgroupPtr = NULL;
			break;
		    }
		    else
		    {
			savgroupPtr = groupPtr;
		    }
		}
		if ( !groupPtr->rightSibling)
		{
		    if (savgroupPtr)
		    {
			highestgroupPtr = savgroupPtr;
			groupPtr =  savgroupPtr->leftChild;
			savgroupPtr = NULL;
		    }
		    else
		    {
			break;
		    }
		}
		else
		{
		    groupPtr = groupPtr->rightSibling;
		}
	    }
	}

	++timeCount;

	if ((caUpdateReq == vTRUE) && dynamicGroupDialogPtr)
	{
	    dynamicGroupDialogPtr -> RefreshDisplay ();
	    caUpdateReq = vFALSE;
	}
	if ((update == vTRUE) || deferredUpdate)
	{
	    update = vTRUE;
	    if (time (0) - lastUpdate > UPDATEDELAY)
	    {
		deferredUpdate = vFALSE;
		// Reset pop alarm stuff
		if (newAlarms == vTRUE)
		{
		    popAlarmCount = 0;
		    popAlarm = vTRUE;
		}

		// Refresh the shell display
		mainDialog -> RefreshDisplay ();
		// Refresh any displayed windows
		if (alarmSummaryDialogPtr)
		{
		    alarmSummaryDialogPtr -> RefreshDisplay ();
		}
		if (dynamicGroupDialogPtr)
		{
		    dynamicGroupDialogPtr -> RefreshDisplay ();
		    caUpdateReq = vFALSE;
		}
		if (alarmHistoryDialogPtr)
		{
		    alarmHistoryDialogPtr -> SetIndication ();
		}
		if (alarmEventDialogPtr)
		{
		    alarmEventDialogPtr -> SetIndication ();
		}
		if (alarmWindowDialogPtr)
		{
		    alarmWindowDialogPtr -> Resync ();
		}
		if (reportsDialogPtr)
		{
		    reportsDialogPtr -> Resync ( );
		}
		lastUpdate = time (0);
	    }
	    else
	    {
		deferredUpdate = vTRUE;
	    }

	    // if there is a tree, and the highest severity is 0, turn off the bell.
	    if (treeRoot)
	    {
		if (HighestSeverity (treeRoot->groupData.numUnackedBySeverity) == 0)
		{
		    if (alarmSoundEnabled)
			BellTurnOff ();
		}
	    }
	}
	else
	{
	    // handle blinking here
	    // start with main
	    mainDialog -> RefreshDisplay ();
	    if (alarmSummaryDialogPtr)
	    {
		alarmSummaryDialogPtr -> UpdateBlink ();
	    }
	    if (dynamicGroupDialogPtr)
	    {
		dynamicGroupDialogPtr -> UpdateBlink ();
	    }
	    if (alarmHistoryDialogPtr)
	    {
		alarmHistoryDialogPtr -> UpdateBlink ();
	    }
	    if (alarmEventDialogPtr)
	    {
		alarmEventDialogPtr -> UpdateBlink ();
	    }
	    if (alarmWindowDialogPtr)
	    {
		alarmWindowDialogPtr -> UpdateBlink ();
	    }
	    timeCount = 0;
	    updateCount = 1;
	    if (popAlarmCount < 5)
	    {
		popAlarmCount++;
	    }
	}

	if ((popAlarm == vTRUE) && (popAlarmCount >= 5))
	{
	    popAlarm = vFALSE;
	    if (highestgroupPtr)
	    {
		if (popupAlrmEnable)
		{
		    if ( (popAlarmPtr == NULL) || highestgroupPtr != popAlarmPtr->groupPtr)
		    {
			if (popAlarmPtr)
			{
			    delete popAlarmPtr;
			    popAlarmPtr = NULL;
			}
			vresource popRes = vresourceGet (alvuApp -> alvuResource, 
							 alvu_PopupDialog);
			popAlarmPtr = new PopupDialog (popRes);
			popAlarmPtr -> OpenPopup (alvuApp, highestgroupPtr, &alvuApp -> shellFilter);
		    }
		    else if (popAlarmPtr)
		    {
			popAlarmPtr -> RefreshDisplay ();
		    }
		}
	    }
	}
	if (alarmSoundEnabled)
	{
	    if (newAlarms == vTRUE)
		BellTurnOn ();
	    BellRing (highestgroupPtr ? highestgroupPtr : NULL);
	}

	inEvent = vFALSE;
    }
}


/*******************************************************************
*
*    mainDialogTimer starter
*
*******************************************************************/

void mainDialogTimer::StartMainTimer (void)
{
    // timer restarts as soon as ObserveTimer finishes
    SetRecurrent ();
    // set timer to expire every .1 seconds
    // args are sec, nanosecs
    SetPeriod (0, 100000000L);
    Start ();
}

