#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vbuttonHEADER
#include vsliderHEADER
#include vcolorHEADER
#include vnumHEADER
#include vprintHEADER

extern "C"
{
#ifdef WIN32
#include <stdio.h>
#include <winsock.h>
// this bit of sillyness is needed because windows #defines AppendMenu
// as #define AppendMenu AppendMenuA, which interfers with galaxy class
// vmenubar::AppendMenu
#undef AppendMenu
#endif WIN32
#include "shareMem.h"
#include "tsDefs.h"
#include "alarm.h"
#include "alarmDefs.h"
#include "alarmfil.h"
#include "alarmvu.h"
}

#include "alvu.h"
#include "alarmvu.hxx"
#include "colorDialog.hxx"
#include "alvuUtils.hxx"


/*******************************************************************
*
*    ColorDialog Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(ColorDialog, vdialog, "ColorDialog");

/*******************************************************************
*
*    ColorDialog constructor
*
*******************************************************************/

void ColorDialog::LoadInit (vresource res)
{

    selected = -1;
    colorsOk = vFALSE;

    sevButtons [NO_ALARM] = SevButtons::RequestSubLoad (this,
					alvu_NormalButton);
    sevButtons [INFO_ALARM] = SevButtons::RequestSubLoad (this,
					alvu_InfoButton);
    sevButtons [EVENT_ALARM] = SevButtons::RequestSubLoad (this,
					alvu_EventButton);
    sevButtons [MINOR_ALARM] = SevButtons::RequestSubLoad (this,
					alvu_MinorButton);
    sevButtons [MAJOR_ALARM] = SevButtons::RequestSubLoad (this,
					alvu_MajorButton);
    sevButtons [VALID_ALARM] = SevButtons::RequestSubLoad (this,
					alvu_ValidButton);
    redSlider = RedSlider::RequestSubLoad (this,
					    alvu_RedSlider);
    redValue = vtextitem::RequestSubLoad (this,
					    alvu_RedValue);
    greenSlider = GreenSlider::RequestSubLoad (this,
					    alvu_GreenSlider);
    greenValue = vtextitem::RequestSubLoad (this,
					    alvu_GreenValue);
    blueSlider = BlueSlider::RequestSubLoad (this,
					    alvu_BlueSlider);
    blueValue = vtextitem::RequestSubLoad (this,
					    alvu_BlueValue);
    acceptButton = ColorAcceptButton::RequestSubLoad (this,
						alvu_AcceptButton);
    cancelButton = ColorCancelButton::RequestSubLoad (this,
						alvu_CancelButton);

    vdialog::LoadInit (res);

    // store ptr to dialog in buttons
    int i;
    for (i = 0; i < NUM_SEV_COLORS; i++)
    {
	sevButtons [i] -> StoreColorDialog (this);
	sevButtons [i] -> index = i;
    }
    redSlider -> StoreColorDialog (this);
    greenSlider -> StoreColorDialog (this);
    blueSlider -> StoreColorDialog (this);
    acceptButton -> StoreColorDialog (this);
    cancelButton -> StoreColorDialog (this);
}


/*******************************************************************
*
*    ColorDialog window observer
*
*******************************************************************/

void ColorDialog::ObserveWindow (vevent *event)
{
    if (event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	colorsOk = vFALSE;
	Close ();
	vevent::StopProcessing ();
    }
}

/*******************************************************************
*
*    ColorDialog destructor
*
*******************************************************************/

void ColorDialog::Destroy ()
{
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    ColorDialog open function
*
*******************************************************************/

void ColorDialog::OpenColor (alvuApplication *alvuapp, long *colors)
{
    app = alvuapp;

    LoadColors (colors);
    InitSliders ();

    Place (vwindow::GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    vevent::Process ();
}


/*******************************************************************
*
*    ColorDialog get new colors
*
*******************************************************************/

void ColorDialog::GetNewColors (long *colors)
{
    if (colorsOk == vTRUE)
    {
	int i;

	for (i = 0; i < NUM_SEV_COLORS; i++)
	{
	    colors [i] = newColors[i];
	}
    }
}


/*******************************************************************
*
*    ColorDialog init sliders routine
*
*******************************************************************/

void ColorDialog::InitSliders (void)
{
    redSlider -> SetValue (0);
    redValue -> SetTextScribed (vcharScribeLiteral ("0"));
    greenSlider -> SetValue (0);
    greenValue -> SetTextScribed (vcharScribeLiteral ("0"));
    blueSlider -> SetValue (0);
    blueValue -> SetTextScribed (vcharScribeLiteral ("0"));
}


/*******************************************************************
*
*    ColorDialog load colors
*	grab incoming colors and set severity buttons to
*	its own foreground color
*
*******************************************************************/

void ColorDialog::LoadColors (long *colors)
{
    vcolorSpec	*spec = new vcolorSpec;
    vcolor	*fgcolor;
    double	red, green, blue;
    int		i;

    for (i = 0; i < NUM_SEV_COLORS; i++)
    {
	newColors [i] = colors [i];
	// see alvuUtils.cxx
	GetRGBColor (newColors [i], &red, &green, &blue);
	spec -> FSetRGB (red, green, blue);
	fgcolor = vcolor::Intern (spec);
	sevButtons [i] -> SetForeground (fgcolor);
    }
    delete spec;
}


/*******************************************************************
*
*    ColorDialog Select
*	set color of sliders and their value fields based on the
*	color of the selected button
*
*******************************************************************/

void ColorDialog::Select (int i)
{
    selected = i;

    int red = (int)(Red(newColors[i]));
    redSlider -> SetValue (red);
    redValue -> SetTextScribed (vnum::ScribeInt (red));

    int green = (int)(Green(newColors[i]));
    greenSlider -> SetValue (green);
    greenValue -> SetTextScribed (vnum::ScribeInt (green));

    int blue = (int)(Blue(newColors[i]));
    blueSlider -> SetValue (blue);
    blueValue -> SetTextScribed (vnum::ScribeInt (blue));
}



/*******************************************************************
*
*    SevButtons Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(SevButtons, vbutton, "SevButtons");

void SevButtons::ObserveDialogItem (vevent *event)
{
    colorDialog -> Select (index);
}


/*******************************************************************
*
*    RedSlider Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(RedSlider, vslider, "RedSlider");

void RedSlider::ObserveDialogItem (vevent *event)
{

    int redV = GetValue ();
    colorDialog -> redValue -> SetTextScribed (vnum::ScribeInt (redV));
    if (colorDialog -> selected < 0)
    {
	return;
    }
    long color = colorDialog -> newColors [colorDialog -> selected];
    long newC = Color(redV, Green(color), Blue(color));
    colorDialog -> newColors [colorDialog -> selected] = newC;

    vcolorSpec	*spec = new vcolorSpec;
    vcolor	*fgcolor;
    double	red, green, blue;

    GetRGBColor (newC, &red, &green, &blue);
    spec -> FSetRGB (red, green, blue);
    fgcolor = vcolor::Intern (spec);
    colorDialog -> sevButtons [colorDialog -> selected] -> SetForeground (fgcolor);
    delete spec;
}


/*******************************************************************
*
*    GreenSlider Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(GreenSlider, vslider, "GreenSlider");

void GreenSlider::ObserveDialogItem (vevent *event)
{
    int greenV = GetValue ();
    colorDialog -> greenValue -> SetTextScribed (vnum::ScribeInt (greenV));
    if (colorDialog -> selected < 0)
    {
	return;
    }
    long color = colorDialog -> newColors [colorDialog -> selected];
    long newC = Color(Red(color), greenV, Blue(color));
    colorDialog -> newColors [colorDialog -> selected] = newC;

    vcolorSpec	*spec = new vcolorSpec;
    vcolor	*fgcolor;
    double	red, green, blue;

    GetRGBColor (newC, &red, &green, &blue);
    spec -> FSetRGB (red, green, blue);
    fgcolor = vcolor::Intern (spec);
    colorDialog -> sevButtons [colorDialog -> selected] -> SetForeground (fgcolor);
    delete spec;
}


/*******************************************************************
*
*    BlueSlider Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(BlueSlider, vslider, "BlueSlider");

void BlueSlider::ObserveDialogItem (vevent *event)
{
    int blueV = GetValue ();
    colorDialog -> blueValue -> SetTextScribed (vnum::ScribeInt (blueV));
    if (colorDialog -> selected < 0)
    {
	return;
    }
    long color = colorDialog -> newColors [colorDialog -> selected];
    long newC = Color(Red(color), Green(color), blueV);
    colorDialog -> newColors [colorDialog -> selected] = newC;

    vcolorSpec	*spec = new vcolorSpec;
    vcolor	*fgcolor;
    double	red, green, blue;

    GetRGBColor (newC, &red, &green, &blue);
    spec -> FSetRGB (red, green, blue);
    fgcolor = vcolor::Intern (spec);
    colorDialog -> sevButtons [colorDialog -> selected] -> SetForeground (fgcolor);
    delete spec;
}


/*******************************************************************
*
*    ColorAcceptButton Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(ColorAcceptButton, vbutton, "ColorAcceptButton");

void ColorAcceptButton::ObserveDialogItem (vevent *event)
{
    colorDialog -> colorsOk = vTRUE;
    colorDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    ColorCancelButton Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(ColorCancelButton, vbutton, "ColorCancelButton");

void ColorCancelButton::ObserveDialogItem (vevent *event)
{
    colorDialog -> colorsOk = vFALSE;
    colorDialog -> Close ();
    vevent::StopProcessing ();
}

