#ifndef	filterdialogHXX
#define	filterdialogHXX

enum FilterType {
    NEW_FILTER = 0,
    MODIFY_FILTER,
    EDIT_FILTER
};

class FilterNameContainer;
class StartupContainer;
class FilterAcceptButton;
class FilterBrowseButton;
class FilterCancelButton;

// *** FILTER DIALOG CLASS *********************************************

class FilterDialog : public vdialog
{
    alvuApplication	*app;
    vstr		*title;
    Filter		filterDef;
    char		*filterName;
    FilterType		typeFlag;
    int			newFilter;

    void LoadFilter (void);
    void SetDialogTitle (void);

public:
    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(FilterDialog);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(FilterDialog, vdialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void ObserveWindow (vevent *event);

    // override
    void Destroy ();

    void OpenFilter (alvuApplication *app, char *filtername, 
			Filter *filter, FilterType typeFlag);

    FilterNameContainer		*filterNameContainer;
    filteredTextitem		*filterNameFld;	// AL_FILTERNAME_SIZE
    filteredTextitem		*filterGroup;	// AL_GROUPNAME_SIZE 
    FilterBrowseButton		*browseButton;
    vcontainerExclusiveGroup	*groupContainer;
    vcontrolToggle		*showAll;
    vcontrolToggle		*showAlarm;
    vcontainerGroup		*ackContainer;
    vcontrolToggle		*hideAck;
    vcontrolToggle		*hideUnAck;
    vcontainerGroup		*severityContainer;
    vcontrolToggle		*hideMajor;
    vcontrolToggle		*hideMinor;
    vcontrolToggle		*hideInfo;
    vcontrolToggle		*hideEvent;
    filteredTextitem		*alarmsAfter;	// AL_DATE_SIZE + AL_TIME_SIZE + 2
    filteredTextitem		*alarmsBefore;
    filteredTextitem		*operatorName;	// AL_OPERATOR_SIZE
    StartupContainer		*startupContainer;
    vcontrolOptionMenu		*startupDisplay;
    FilterAcceptButton		*acceptButton;
    FilterCancelButton		*cancelButton;

    alvuApplication *GetApp (void)
	{ return (app); }
    Filter *GetDefFilter (void)
	{ return (&filterDef); }
    FilterType GetFilterType (void)
	{ return (typeFlag); }
    int GetNewFilter (void)
	{ return (newFilter);  }
    void SetNewFilter (int flag)
	{ newFilter = flag; }
    int NewFilter (Filter *filter);
};


// *** FILTER NAME CONTAINER CLASS *************************************

class FilterNameContainer : public vcontainer
{

public:

    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(FilterNameContainer);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(FilterNameContainer, vcontainer);

    // vloadable override
    void LoadInit (vresource resource);

    filteredTextitem	*filterName;
};


// *** STARTUP CONTAINER CLASS ****************************************

class StartupContainer : public vcontainer
{

public:

    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(StartupContainer);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(StartupContainer, vcontainer);

    // vloadable override
    void LoadInit (vresource resource);

    vcontrolOptionMenu	*startupDisplay;
};


// *** ACCEPT BUTTON CLASS **********************************************

class FilterAcceptButton : public vbutton
{
    FilterDialog	*filterDialog;

public:

    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(FilterAcceptButton);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(FilterAcceptButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    // class extensions
    void StoreFilterDialog (FilterDialog *dlg)
	{ filterDialog = dlg; }
};


// *** BROWSE BUTTON CLASS ***********************************************

class FilterBrowseButton : public vbutton
{
    FilterDialog	*filterDialog;

public:
    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(FilterBrowseButton);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(FilterBrowseButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    // class extensions
    void StoreFilterDialog (FilterDialog *dlg)
	{ filterDialog = dlg; }
};


// *** CANCEL BUTTON CLASS **********************************************

class FilterCancelButton : public vbutton
{
    FilterDialog	*filterDialog;

public:
    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(FilterCancelButton);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(FilterCancelButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    // class extensions
    void StoreFilterDialog (FilterDialog *dlg)
	{ filterDialog = dlg; }
};


// *** FILTER EDIT CLASS **********************************************

class FilterEditDialog : public FilterDialog
{

public:
    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(FilterEditDialog);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(FilterEditDialog, FilterDialog);

    // vloadable override
    void LoadInit (vresource resource);

    void OpenFilter (alvuApplication *app, char *filtername,
		     Filter *filter, FilterType typeFlag);
};

#endif	filterdialogHXX

