/*
    This is the Network class for all dataVu network operations. These include:

	1. Online point additions.
	2. Online point deletions.
	3. Remote unit reset.
	4. dbClient (via localhost) notification of point additions/modifications
*/
#define VU_NETWORK_MSG_SIZE 1024
#ifndef USE_UDP
#define DBC_NETMSG_SIZE 4096
#endif USE_UDP

#ifndef	WIN32
#define	SOCKET		int
#define	INVALID_SOCKET	-1
#endif	!WIN32

extern "C"
{
#ifdef	WIN32
#include <winsock.h>
#else
#include <netinet/in.h>
#endif	WIN32
}

class vuNetwork
{
public:
	vuNetwork(vuMainDialogClass *md);		// no dest...use local host and do connect
#ifdef USE_UDP
	vuNetwork(vuMainDialogClass *md,char *dest);	// just save destination and other stuff
#endif USE_UDP
	~vuNetwork();
private:
    SOCKET vuSocket;					// socket for all communications
    int addr;						// inet address for current socket
    struct hostent	*ent;
    struct  sockaddr_in	destAddr;
    int vuNetError;					// set to vTRUE; on any error
    char destName[VU_NETWORK_MSG_SIZE];
#ifndef USE_UDP
    char rxBuffer[DBC_NETMSG_SIZE];
    int rxCount;
    int rxNextChar;
#endif USE_UDP
    char outMsg[VU_NETWORK_MSG_SIZE];
    vuMainDialogClass *mainDialog;
    vresource vuRes;
public:
    //
    // member functions to actually do some work
    //
    int checkError(void);				// return error flag;
#ifdef USE_UDP
    int sendReset(void);				// reset IOC	dest already set
    void sendDelete(char *pvname);			// delete "pvname" from IOC...dest already set
    int sendAdd(void);					// tell IOC that there are new points...dest already set
#else
    int sendReset(char *iocname);			// reset IOC	dest already set
    void sendDelete(char *iocname, char *pvname);	// delete "pvname" from IOC...dest already set
    int sendAdd(char *iocname);				// tell IOC that there are new points...dest already set
#endif USE_UDP
    void sendNotify(char *name);			// notify dbClient of point change
#ifndef USE_UDP
    int sendToLocalClient(short mType, char *msg, int size);
    int recvFromLocalClient (void);
    int readTcp (char *buff, int nbytes);
    int haveCompleteMsg (void);
#endif USE_UDP
};

#ifndef vuInUseDialogclassHXX
#define vuInUseDialogclassHXX

class  vuInUseDialogClass : public vdialog
{
    vresource	IocResource;
    int		singleSelect;
public:
    // macros for standard versions of the
    vkindTYPED_FULL_DECLARATION(vuInUseDialogClass);
    // macros for standard constructors and desstructors
    vloadableINLINE_CONSTRUCTORS(vuInUseDialogClass, vdialog);
    // override to load dialog
    void LoadInit(vresource res);
    void addToList(char *item); 
private:
    vbutton	*dismissButton;
    vlistitem	*iocList;
};
#endif vuInUseDialogclassHXX

