#include <sys/time.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <avl.h>
#include "apxdefs.h"
#include <../newtrend/trendDefs.h>
#include <../newtrend/oracleSnap.h>
#include <../newtrend/oracleSnapLib.h>
#include <../newtrend/archiveLib.h>
#define MAX_ARRAY_SIZE 2000

static double *apxValue = 0;
static double *apxValue2 = 0;
static unsigned char **apxStatus = 0;
static int atdbCount = 0;
static char daytab[2][13] =
{
  {0,31,28,31,30,31,30,31,31,30,31,30,31},
  {0,31,29,31,30,31,30,31,31,30,31,30,31}
};
void apxGetMemory(int size);
int parseTime(char *timeString);

int main(int argc,char *argv[])
{
  double	dispLo,dispHi;
  int	atdbSize;
  int	i,atdbSpan;
  char	charArg[81],eguTag[20],pointName[30];
  char	atdbRed1[RED_STR_SIZE],atdbRed2[RED_STR_SIZE];
  long	atdbRate;
  time_t	apxTime;
  unsigned char *sptr;
  /*
    SnapRec structure for data returns from oracle wrapper functions
  */
  if(argc < 7) exit(-2);
  /*
    Init database(Oracle) stuff next...
  */
  OpenOracle();					/* start oracle initialization */
  /*
    convert command line parameters
  */
  strcpy(pointName,argv[1]);
  apxTime = parseTime(argv[2]);			/* use readable Date/Time string */
  atdbSpan = (int)atoi(argv[3]);
  atdbSize = (int)atoi(argv[4]);
  strncpy(atdbRed1,argv[5],RED_STR_SIZE);
  strncpy(atdbRed2,argv[6],RED_STR_SIZE);

  apxGetMemory(atdbSize);
/*
	Get the sample.
*/
  if(!GetSnapValues(pointName,atdbRed1,atdbRed2,atdbSpan,apxTime,&atdbRate,eguTag,&dispLo,&dispHi,atdbSize,apxValue,apxStatus,apxValue2))
    {
      free(apxValue);
      free(apxValue2);
      free(apxStatus);
      return(-1);
    }
  sptr = (char *)apxStatus;
  for(i = 0;i < atdbSize;i++)
    {
      if((char)sptr[i*3] == (char)0xff)
	apxValue[i] = 0;
      strcpy(charArg,ctime(&apxTime));
      charArg[strlen(charArg)-1] = 0; /* drop \lf */
      printf("%s... Value: %lf... Status \"%3.3s\" [%02x %02x %02x] @%d...%lf\n",
	     charArg,
	     apxValue[i],
	     sptr+(i*3),
	     *(sptr+(i*3)),
	     *(sptr+((i*3)+1)),
	     *(sptr+((i*3)+2)),
	     atdbRate,
	     apxValue2[i]);
      apxTime += atdbSpan;
    }
  free(apxValue);
  free(apxValue2);
  free(apxStatus);
  return(0);
}
void apxGetMemory(int size)
{
if(size > atdbCount)
  {
    atdbCount = size;
    if(apxValue) free(apxValue);
    apxValue = (double *)calloc(size,sizeof(double));
    if(apxValue2) free(apxValue2);
    apxValue2 = (double *)calloc(size,sizeof(double));
    if(apxStatus) free(apxStatus);
    apxStatus = (unsigned char **)calloc(size*3,sizeof(char));
  }
}
/*
  function to parse date/time as entered from ommand line. Return seconds
  since 1-Jan-1970 as an integer;
*/
extern int timezone;
int parseTime(char *timeString)
{
  struct tm t;
  int rTime;
  int leap,days=0;
  short i;

  if((char *)strptime(timeString,"%D %T",&t) != NULL)
    {
      rTime = t.tm_sec+(SECS_PER_MINUTE*t.tm_min)+(SECS_PER_HOUR*(t.tm_hour));
      for(i = 1970;i < t.tm_year+1900;i++)
	{
	  leap = (i%4 == 0) && (i%100 != 0) || (i%400 == 0);
	  days += 365;
	  if(leap) days++;
	}
      leap = (i%4 == 0) && (i%100 != 0) || (i%400 == 0);
      for(i=0;i < t.tm_mon+1;i++) days += daytab[leap][i];
      days += (t.tm_mday-1);
#ifdef SOLARIS
      rTime += (days * SECS_PER_DAY);
#else
      rTime += (days * SECS_PER_DAY);
#endif
      return(rTime);/*+timezone);*/
    }
  else
    return((int)time(0));/*+timezone);*/
}
