
#include <stdio.h>
#include <process.h>
#include "DBDefines.h"

FILE	*dbLog;		/* Log file pointer */
int	dbPid;		/* PID of current program */
int	dbJournal;	/* verbosity level */
char	dbPName[MAXPATHLEN];	/* program name from argv[0] */
int     lockFd;		/* fd of lock file */

char	DB_DatabasePath[MAXPATHLEN];
C4CODE	DB_CodeBase;
static	char first_time = 1;			/* so d4init only called once */

static DB_TABLE_PTR	dbSearchTable(desc)
char	*desc;
{
    DB_TABLE_PTR	temp;

    temp = (DB_TABLE_PTR)lstFirst((LIST *)&DB_Database.tables);
    while(temp)
    {
	if(!strcmp(desc,temp->table_name))
	    break;
	temp = (DB_TABLE_PTR)lstNext((NODE *)temp);
    }
    return(temp);
}

static DB_FIELD_PTR	dbSearchField(tptr,fdesc)
DB_TABLE_PTR	tptr;
char		*fdesc;
{
    DB_FIELD_PTR	temp;

    temp = (DB_FIELD_PTR)lstFirst((LIST *)&tptr -> fields);
    while(temp)
    {
	if(!strcmp(fdesc,temp -> field_name))
		break;
	temp = (DB_FIELD_PTR)lstNext((NODE *)temp);
    }
    return(temp);
}

static int	dbDeleted(table)
D4DATA	*table;
{
    F4FIELD	*del;

    if((del = d4field (table, FLD_DELETEFLAG)) != 0) /* must me pvmap */
	return (f4true (del));
    else
    {
	return (d4deleted (table));	/* otherwise record is blanked and CB delete flag is set */
    }
}

static void	dbClrDeleteFlag(table)
D4DATA	*table;
{
    F4FIELD	*del;

    if((del = d4field (table, FLD_DELETEFLAG)) != 0)
	f4assignChar (del, 'F');	/* set delete flag to false */
    return;
}

static int    dbObtainRecord (D4DATA *table, char *name)
{
    T4TAG   *tag;
    int	    stat;

    if (name)				/* try to find pre-existing record */
    {
	if ((stat = d4seek (table, name)) == 0)	/* if found then see if delete flag set */
	{
	    if (dbDeleted (table))	/* if deleted then return record */
	    {
		if (d4deleted (table))	/* remove dBase delete mark */
		    d4recall (table);
		dbClrDeleteFlag (table);/* set 'delete flag' to FALSE (if there is one) */
		d4unlock_all (table);
		return (stat);
	    }
	}
	else
	{
	    stat = d4top (table);
	    if (stat == 0)
	    {
		if (dbDeleted (table))	/* is record blank ? */
		{
		    if (d4deleted (table))/* remove dBase delete mark */
			d4recall (table);
		    dbClrDeleteFlag (table);/* set 'delete flag' to FALSE (if there is one) */
		}
		else
		{
		    tag = d4tagSelected (table);/* get the selected tag */
		    d4tagSelect (table,0);/* record number processing */
		    d4goEof (table);     /* recalculate EOF(number of records) and position file there */
		    stat = d4appendBlank (table); /* append at new EOF */
		    d4tagSelect (table,tag);      /* restore original tag */
		    if (stat != 0)
			stat = -1;	/* could not append then error */
		}
	    }
	    else if (stat == r4eof)
	    {				/* file empty then create new record */
		stat = d4appendBlank (table);
		if (stat != 0)
		    stat = -1;		/* could not append then error */
	    }
	    else
	    {
		stat = -1;
	    }
	}
    }
    else
    {
	stat = d4top (table);
	if (stat == 0)
	{
	    if (dbDeleted (table))	/* is record blank ? */
	    {
		if (d4deleted (table))	/* remove dBase delete mark */
		    d4recall (table);
		dbClrDeleteFlag (table); /* set 'delete flag' to FALSE (if there is one) */
	    }
	    else
	    {
		tag = d4tagSelected (table); /* get the selected tag */
		d4tagSelect (table, 0); /* record number processing */
		d4goEof (table);     	/* recalculate EOF(number of records) and position file there */
		stat = d4appendBlank (table); /* append at new EOF */
		d4tagSelect (table, tag); /* restore original tag */
		if (stat != 0)
		    stat = -1;     	/* could not append then error */
	    }
	}
	else if (stat == r4eof)
	{				/* file empty then create new record */
	    stat = d4appendBlank (table);
	    if (stat != 0)
		stat = -1;		/* could not append then error */
	}
	else
	{
	    stat = -1;
	}
    }
    d4unlock_all (table);
    return (stat);
}

void	DB_InitDatabase(pgmname)
char	*pgmname;
{
    if(first_time)
    {
	d4init(&DB_CodeBase);
	DB_CodeBase.errFieldName = 0;
	DB_CodeBase.errOpen = 0;
	DB_CodeBase.errTagName = 0;
	DB_CodeBase.readLock = 0;
#if !(defined(CB51) || defined (CB60))
	DB_CodeBase.wait = 0;
#endif
	DB_CodeBase.errCreate = 0;
	DB_CodeBase.errGo = 0;	/* 18-apr-1996 */
	DB_CodeBase.errOff = 1;	/* 19-Apr-1996 turn off error display */
	first_time = 0;

#ifdef WIN32
	dbPid = _getpid();				/* calling process PID */
#else
	dbPid = getpid();				/* calling process PID */
#endif WIN32
	strcpy(dbPName,pgmname);			/* save program name too */
    }
}

void	DB_MapDatabase(dbdesc)
char	*dbdesc;
{
    strcpy(DB_Database.path, dbdesc);
    lstInit((LIST *)&DB_Database.tables);
}

void	DB_UnMapDatabase (dopack)
int		dopack;
{
    DB_TABLE_PTR	tbl;

    tbl = (DB_TABLE_PTR)lstFirst((LIST *)&DB_Database.tables);
    while(tbl)
    {
	lstFree((LIST *)&tbl->fields);		/* free associated fields first */
	d4unlock_all (tbl->table);
	d4flush_all (tbl->table);
	d4unlock_all (tbl->table);
	d4close(tbl->table);			/* close the file */
	tbl = (DB_TABLE_PTR)lstNext((NODE *)tbl); /* next table... */
    }
    lstFree((LIST *)&DB_Database.tables);	/* free all tables last */
}

DB_TABLE_PTR	DB_MapTable (tabdesc)
char	*tabdesc;
{
    DB_TABLE_PTR temp;
    char    name[MAXPATHLEN], buff[MAXPATHLEN];
    int	    linklen;

#ifdef WIN32
    linklen = -1;
#else
    linklen = readlink (DB_Database.path, buff, MAXPATHLEN);
#endif WIN32
    if(linklen < 0)
	sprintf (name,"%s%s.dbf\0", DB_Database.path, tabdesc);	/* copy name */
    else
    {
	buff[linklen] = '\0';
	sprintf (name, "%s%s.dbf\0", buff, tabdesc);		/* copy name */
    }
    if((temp = dbSearchTable (name)) == 0)
    {
	if((temp = (DB_TABLE_PTR)malloc (sizeof(DB_TABLE))) == 0)
	    return ((DB_TABLE_PTR)0);				/* can't get memory then return 0 */
	strcpy(temp -> table_name, name);
	if((temp -> table = d4open(&DB_CodeBase, name)) == 0)	/* open database */
	{
	    free(temp);						/* if error then release memory and exit */
	    return((DB_TABLE_PTR)0);
	}
	temp -> table_records = d4recCount (temp -> table);	/* get number of records in file */
	temp -> table_fields = d4numFields (temp -> table);	/* get number of fields in file */
	temp -> tag = d4tagDefault (temp -> table);		/* get the default tag for this table */
	temp -> current_record = 0;				/* no current record until DB_Seek */
	lstInit ((LIST *)&temp->fields);			/* init field list */
/*
	Add to list of tables
*/
	lstAdd((LIST *)&DB_Database.tables, (NODE *)temp);
    }
    return (temp);
}

DB_TABLE_PTR	DB_CreateTable(char *tabdesc, FIELD4INFO *FieldArray, TAG4INFO *TagArray)
{
    DB_TABLE_PTR temp;
    char    name[MAXPATHLEN],buff[MAXPATHLEN];
    int	    linklen;

#ifdef WIN32
    linklen = -1;
#else
    linklen = readlink(DB_Database.path,buff,MAXPATHLEN);	/* convert link to abs path if possible */
#endif WIN32
    if(linklen < 0)
	sprintf(name,"%s%s.dbf\0",DB_Database.path,tabdesc);	/* path not a link...just copy name */
    else
    {
	buff[linklen] = '\0';					/* null terminate */
	sprintf(name,"%s%s.dbf\0",buff,tabdesc);		/* 'buff' has fully resolved path...copy name */
    }
    if((temp = (DB_TABLE_PTR)malloc(sizeof(DB_TABLE))) == 0)
	return((DB_TABLE_PTR)0);				/* can't get memory then return 0 */
    strcpy(temp -> table_name,name);
    if((temp -> table = d4create(&DB_CodeBase,name,FieldArray,TagArray)) == 0)		/* create database */
    {
	free(temp);						/* if error then release memory and exit */
	return((DB_TABLE_PTR)0);
    }
    d4unlock_all(temp -> table);
    temp -> table_records = d4recCount(temp -> table);		/* get number of records in file */
    temp -> table_fields = d4numFields(temp -> table);		/* get number of fields in file */
    temp -> tag = d4tagNext(temp -> table,0);			/* get the default tag for this table */
    temp -> current_record = 0;					/* no current record until DB_Seek */
    lstInit((LIST *)&temp->fields);					/* init field list */
    /*
	Add to list of tables
    */
    lstAdd((LIST *)&DB_Database.tables, (NODE *)temp);
    return(temp);
}

DB_FIELD_PTR	DB_MapField (TableHandle,FieldName)
DB_TABLE_PTR TableHandle;
char	*FieldName;
{
    DB_FIELD_PTR	temp;
    F4FIELD		*tempfield;

    if(strlen(FieldName) == 0)
	return ((DB_FIELD_PTR)0);
    if((temp = dbSearchField (TableHandle, FieldName)) == 0)
    {
	if((tempfield = d4field (TableHandle -> table, FieldName)) == 0)
		return((DB_FIELD_PTR)0);
	if((temp = (DB_FIELD_PTR)malloc (sizeof(DB_FIELD))) == 0)
		return ((DB_FIELD_PTR)0);
	strcpy(temp -> field_name, FieldName);
	temp -> field = tempfield;
	temp -> field_type	= f4type(temp -> field);
	temp -> field_size	= f4len(temp -> field);
	temp -> field_scale	= f4decimals(temp -> field);
	temp -> field_number	= d4fieldNumber (TableHandle -> table, temp -> field_name);
	temp -> owner		= TableHandle;		/* point to owner table */
/*
	Add to list
*/
	lstAdd ((LIST *)&TableHandle->fields, (NODE *)temp);
    }
    return (temp);
}

int	DB_Top (TableHandle)
DB_TABLE_PTR	TableHandle;
{
    int		stat;

    if (TableHandle == 0)
	return (0);
    d4tagSelect (TableHandle -> table, TableHandle->tag);
    stat = d4top (TableHandle -> table);
    if(stat == r4eof)					/* this means the file was empty */
    {
	TableHandle -> current_record = 0;
	d4unlock_all (TableHandle -> table);
	return (0);
    }
    if(stat == 0)
    {
	while (dbDeleted (TableHandle -> table))		/* skip over deleted records */
	{
	    if((stat = d4skip (TableHandle -> table, 1)) != 0)	/* none 'till EOF or error, return error */
	    {
		TableHandle -> current_record = 0;
		d4unlock_all (TableHandle -> table);
		return (0);
	    }
	}
    }
    else
    {
	TableHandle -> current_record = 0;
	d4unlock_all (TableHandle -> table);
	return (0);
    }
    TableHandle -> current_record = 1;
    d4unlock_all (TableHandle -> table);
    return (1);
}

int	DB_Seek (TableHandle, PrimaryKey)
DB_TABLE_PTR	TableHandle;
char		*PrimaryKey;
{
    int		stat;

    if (TableHandle == 0)
	return (0);
    d4tagSelect (TableHandle -> table, TableHandle -> tag);
    stat = d4seek (TableHandle -> table, PrimaryKey);
    if(stat == 0)
    {
	if(dbDeleted(TableHandle -> table))			/* same as not found */
	{
	    TableHandle -> current_record = 0;	/* else return error */
	    d4unlock_all (TableHandle -> table);
	    return (0);
	}
    }
    else
    {
	d4unlock_all(TableHandle -> table);
	return (0);
    }
    TableHandle -> current_record = 1;
    d4unlock_all (TableHandle -> table);
    return (1);
}

int	DB_SeekNext (TableHandle)
DB_TABLE_PTR	TableHandle;
{
    int	    status;
    int	    rec;

    if((rec = TableHandle -> current_record) != 0)		/* seek next only if DB_Seek done first */
    {
	d4tagSelect(TableHandle -> table,TableHandle -> tag);
	status = d4skip(TableHandle -> table,DB_SKIP_ONE);
	if(status == 0)
	{
	    while (dbDeleted (TableHandle -> table))		/* do NOT return deleted records */
	    {
		if((status = d4skip(TableHandle -> table,1)) != 0)
		{
		    rec = 0;
		    TableHandle -> current_record = 0;
		    d4unlock_all(TableHandle -> table);
		    return (rec);
		}
	    }
	}
	else
	{
	    rec = 0;				/* return zero to indicate error */
	    TableHandle -> current_record = 0;	/* clear for next DB_Seek */	
	}
    }
    d4unlock_all(TableHandle -> table);
    return(rec);
}

int	DB_NewRecord (TableHandle, name)
DB_TABLE_PTR	TableHandle;
char			*name;
{
    if (TableHandle == 0)
	return (0);
    if (dbObtainRecord (TableHandle->table, name) < 0)
	return (0);
    d4unlock_all (TableHandle->table);
    d4flush_all (TableHandle->table);
    TableHandle -> current_record = 1;
    d4unlock_all (TableHandle->table);
    return (1);
}

void	DB_GetFieldValue (RecordHandle, FieldHandle, value, type)
int	RecordHandle;
DB_FIELD_PTR	FieldHandle;
void	*value;
int	*type;
{
    char    *tstr;
    int	    tsize;

    *type = FieldHandle -> field_type;
    switch(FieldHandle -> field_type)
    {
	case 'N':
	case 'F':
	    *((double *)value) = f4double(FieldHandle -> field);
	    break;
	case 'D':
	    *((int *)value) = f4long(FieldHandle -> field);
	    break;
	case 'L':
	    *((char *)value) = f4true(FieldHandle -> field);
	    break;
	case 'C':
	    tstr = f4str(FieldHandle -> field);
	    for(tsize = strlen(tstr)-1;tsize >= 0;tsize--)
		if(isspace(tstr[tsize]))
		    tstr[tsize] = '\0';
		else
		    break;
	    strcpy((char *)value,tstr);
	    break;
    }
    d4unlock_all(FieldHandle->owner->table);
}

void	DB_PutFieldValue(RecordHandle,FieldHandle,value)
int	RecordHandle;
DB_FIELD_PTR	FieldHandle;
void		*value;
{
    DB_TABLE_PTR tbl;

    if(RecordHandle == 0)
	return;
    tbl = FieldHandle -> owner;
    switch(FieldHandle -> field_type)
    {
	case 'N':
	case 'F':
		f4assignDouble (FieldHandle -> field, *((double *)value));
		break;
	case 'D':
		f4assignLong (FieldHandle -> field, *((int *)value));
		break;
	case 'L':
		f4assignChar (FieldHandle -> field, (*((unsigned char *)value) == 1 ? 'T' : 'F'));
		break;
	case 'C':
		f4assignN (FieldHandle -> field, (char *)value, strlen((char *)value));
		break;
    }
    d4unlock_all (tbl->table);		/* make sure everything is unlocked */
    d4flush_all (tbl->table);		/* write everything to disk */
}

int	DB_DeleteRecord(tbl,rec)
int	rec;
DB_TABLE_PTR	tbl;
{
    if(!rec)
	return (0);		    	/* must be non-zero */
    d4blank (tbl->table);		/* blank entire record */
    d4delete (tbl->table);		/* mark for delete also GCT 02-Sep-1994 */
    d4unlock_all (tbl->table);		/* un-lock everything */
    d4flush_all (tbl->table);		/* flush it! */
    d4unlock_all (tbl->table);		/* un-lock everything */
    return (1);				/* record deleted */
}
