/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <DBDefines.h>
#include <DBProto.h>
#ifdef WIN32
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#else
#include <sys/time.h>
#include <sys/errno.h>
#include <stat.h>
#endif WIN32
#include <avmDefs.h>
#include <catlib.h>

FIELD4INFO	avmFieldInfo[] = 
	{
	{AVM_STARTDATE_NAME,	'C',	AVM_STARTDATE_SIZE,	0},
	{AVM_ENDDATE_NAME,	'C',	AVM_ENDDATE_SIZE,	0},
	{AVM_MOUNTED_NAME,	'L',	AVM_MOUNTED_SIZE,      	0},
	{AVM_DISKLABEL_NAME,	'C',	AVM_DISKLABEL_SIZE,	0},
	{AVM_MOUNTPNT_NAME,	'C',	AVM_MOUNTPNT_SIZE,	0},
	{AVM_TABLENAME_NAME,	'C',	AVM_TABLENAME_SIZE,	0},
	{0,0,0,0}
	};
T4TAG_INFO	avmTagInfo[] = 
	{
	{"STARTDATE",	"STARTDATE",	0,	0,	0},
	{0,0,0,0,0}
	};
/*
  Static data
  */
static DB_TABLE_PTR catTable;			/* dbLibrary table */
static DB_FIELD_PTR catStartField;
static DB_FIELD_PTR catEndField;
static DB_FIELD_PTR catMountedField;
static DB_FIELD_PTR catDiskLabelField;
static DB_FIELD_PTR catMountPntField;
static DB_FIELD_PTR catTableNameField;
struct	dbdatabase	DB_Database;		/* for DB functions */
static int firstTime = 1;			/* so dBase stuff is only done once */
static int firstRec = 1;			/* do first record first */
static int currentRec;
static int catFindEntry(time_t);		/* find record in catalog within specified range */
extern int errno;
extern time_t timezone;
/*
  Open or create the catalog file and assign all the field pointers.
*/
int catOpen (CREATE_FLAG create_it,char *id)
{
#ifdef WIN32
    struct _stat buff;
#else
    struct stat buff;
#endif WIN32

    if (firstTime)
    {
	DB_InitDatabase (id);
	DB_MapDatabase (AVM_DEFAULT_DATABASE);	/* map to database path */
	firstTime = 0;
    }
#ifdef WIN32
    if ((_stat(AVM_CATALOG_NAME,&buff)) < 0)
#else
    if ((stat(AVM_CATALOG_NAME,&buff)) < 0)
    {
	if (errno == ENOENT)
#endif WIN32
	{
	    if (create_it == YES_CREATE)
		catTable = (DB_TABLE_PTR)DB_CreateTable(AVM_CATALOG_NAME_ONLY,avmFieldInfo,avmTagInfo);
	}
#ifndef WIN32
	else
	    return (0);
    }
#endif WIN32
    else
	catTable = DB_MapTable(AVM_CATALOG_NAME_ONLY);
  /*
    map fields
  */
    if (catTable)
    {
	catStartField	    = DB_MapField (catTable, AVM_STARTDATE_NAME);
	catEndField	    = DB_MapField (catTable, AVM_ENDDATE_NAME);
	catMountedField	    = DB_MapField (catTable, AVM_MOUNTED_NAME);
	catDiskLabelField   = DB_MapField (catTable, AVM_DISKLABEL_NAME);
	catMountPntField    = DB_MapField (catTable, AVM_MOUNTPNT_NAME);
	catTableNameField   = DB_MapField (catTable, AVM_TABLENAME_NAME);
	return (1);
    }
    return (0);
}

int catAddEntry (time_t startDate,
		 time_t endDate,
		 char *tableName,
		 char *diskName,
		 char *pathName,
		 char onlineFlag)
{
    char startStr[AVM_STARTDATE_SIZE+1], endStr[AVM_ENDDATE_SIZE+1];
    struct tm *tmPtr;

    tmPtr = gmtime(&startDate);
    sprintf (startStr, AVM_DATE_FORMAT,
		tmPtr->tm_mon+1, tmPtr->tm_mday, tmPtr->tm_year % 100,
		tmPtr->tm_hour,tmPtr->tm_min,tmPtr->tm_sec);
    if ((currentRec = DB_NewRecord (catTable,startStr)) != 0)
    {
	DB_PutFieldValue (currentRec, catStartField, startStr);
	if (endDate)
	{
	    tmPtr = gmtime (&endDate);
	    sprintf (endStr, AVM_DATE_FORMAT,
			tmPtr->tm_mon+1, tmPtr->tm_mday, tmPtr->tm_year % 100,
			tmPtr->tm_hour,tmPtr->tm_min,tmPtr->tm_sec);
	    DB_PutFieldValue (currentRec, catEndField, endStr);
	}
	if (diskName)
	    DB_PutFieldValue(currentRec,catDiskLabelField,diskName);
	if (pathName)
	    DB_PutFieldValue(currentRec,catMountPntField,pathName);
	if (tableName)
	    DB_PutFieldValue(currentRec,catTableNameField,tableName);
	if (onlineFlag >= 0)
	    DB_PutFieldValue(currentRec,catMountedField,&onlineFlag);
    }
    return ((int)currentRec);
}

int catUpdateEntry(time_t startDate,time_t endDate,char *tableName,char *diskName,char *pathName,char onlineFlag)
{
    char startStr[AVM_STARTDATE_SIZE+1],endStr[AVM_ENDDATE_SIZE+1];
    struct tm *tmPtr;

    tmPtr = gmtime(&startDate);
    sprintf(startStr,AVM_DATE_FORMAT,tmPtr->tm_mon+1,tmPtr->tm_mday,tmPtr->tm_year % 100,tmPtr->tm_hour,tmPtr->tm_min,tmPtr->tm_sec);
    if ((currentRec = DB_Seek(catTable,startStr)) != 0)
    {
	if (endDate)
	{
	    tmPtr = gmtime(&endDate);
	    sprintf(endStr,AVM_DATE_FORMAT,tmPtr->tm_mon+1,tmPtr->tm_mday,tmPtr->tm_year % 100,tmPtr->tm_hour,tmPtr->tm_min,tmPtr->tm_sec);
	    DB_PutFieldValue(currentRec,catEndField,endStr);
	}
	if (diskName)
	    DB_PutFieldValue (currentRec, catDiskLabelField, diskName);
	if (pathName)
	    DB_PutFieldValue (currentRec, catMountPntField, pathName);
	if (tableName)
	    DB_PutFieldValue (currentRec, catTableNameField, tableName);
	if (onlineFlag >= 0)
	    DB_PutFieldValue (currentRec, catMountedField, &onlineFlag);
    }
    return ((int)currentRec);
}

int catGetEntry (time_t targetDate,
		 time_t *startDate,
		 time_t *endDate,
		 char *tableName,
		 char *diskName,
		 char *pathName,
		 char *onlineFlag)
{
    int type;
    char startStr[AVM_STARTDATE_SIZE+1], endStr[AVM_ENDDATE_SIZE+1];
    struct tm tmBuff;

    if (tableName)
	strcpy(tableName,"");			/* set-up for error return...success will fill it in anyway */
    if ((currentRec = catFindEntry(targetDate)) != 0)
    {
	if (startDate)
	{
	    DB_GetFieldValue (currentRec, catStartField, startStr, &type);
	    memset ((char *)&tmBuff, 0, sizeof(struct tm));
	    sscanf (startStr, AVM_DATE_FORMAT,
			&tmBuff.tm_mon, &tmBuff.tm_mday, &tmBuff.tm_year,
			&tmBuff.tm_hour, &tmBuff.tm_min, &tmBuff.tm_sec);
	    if (tmBuff.tm_year < 70)
		tmBuff.tm_year += 100;
	    tmBuff.tm_mon--;
	    *startDate = mktime (&tmBuff) - timezone;
	}
	if (endDate)
	{
	    DB_GetFieldValue (currentRec, catEndField, endStr, &type);
	    memset ((char *)&tmBuff, 0, sizeof(struct tm));
	    sscanf (endStr, AVM_DATE_FORMAT,
			&tmBuff.tm_mon, &tmBuff.tm_mday, &tmBuff.tm_year,
			&tmBuff.tm_hour, &tmBuff.tm_min, &tmBuff.tm_sec);
	    if (tmBuff.tm_year < 70)
		tmBuff.tm_year += 100;
	    tmBuff.tm_mon--;
	    *endDate = mktime (&tmBuff) - timezone;
	}
	if (diskName)
	    DB_GetFieldValue (currentRec, catDiskLabelField, diskName, &type);
	if (pathName)
	    DB_GetFieldValue (currentRec, catMountPntField, pathName, &type);
	if (tableName)
	    DB_GetFieldValue (currentRec, catTableNameField, tableName, &type);
	if (onlineFlag)
	{
	    DB_GetFieldValue (currentRec, catMountedField, onlineFlag, &type);
/* This prevent offline table recognition!!!!
	    if (*onlineFlag == 0)
	      currentRec = 0;
*/
	}
    }
    return ((int)currentRec);
}

/*
  Get Raw table entry. Any NULL parameter does not attempt to get that parameter
*/
int catGetNextRawEntry(char *startDate,
		       char *endDate,
		       char *tableName,
		       char *diskName,
		       char *pathName,
		       char *onlineFlag)
{
    int type;

    if (firstRec)
    {
	firstRec = 0;
	currentRec = DB_Top (catTable);
    }
    else
	currentRec = DB_SeekNext (catTable);
    if (currentRec)
    {
	if (startDate)
	    DB_GetFieldValue (currentRec, catStartField, startDate, &type);
	if (endDate)
	    DB_GetFieldValue (currentRec, catEndField, endDate, &type);
	if (diskName)
	    DB_GetFieldValue (currentRec, catDiskLabelField, diskName, &type);
	if (pathName)
	    DB_GetFieldValue (currentRec, catMountPntField, pathName, &type);
	if (tableName)
	    DB_GetFieldValue (currentRec, catTableNameField, tableName, &type);
	if (onlineFlag)
	    DB_GetFieldValue (currentRec, catMountedField, onlineFlag, &type);
    }
    return ((int)currentRec);
}

int catDeleteCurrentEntry (void)
{
    if (currentRec)
	DB_DeleteRecord (catTable, currentRec);
    return (currentRec);
}

void catClose (void)
{
    DB_UnMapDatabase(0);
    catStartField	= 0;
    catEndField      	= 0;
    catMountedField	= 0;
    catDiskLabelField	= 0;
    catMountPntField	= 0;
    catTableNameField	= 0;
}

static int catFindEntry (time_t targetDate)
{
    char startStr[AVM_STARTDATE_SIZE+1], endStr[AVM_ENDDATE_SIZE+1];
    time_t startDate, endDate;
    struct tm tmBuff;
    DB_RECORD_PTR rec;

    firstRec = 1;			/* start at top */
    while ((rec = (DB_RECORD_PTR)catGetNextRawEntry(startStr,endStr,NULL,NULL,NULL,NULL)) != 0)
    {
	/* convert time strings to time_t */
	memset ((char *)&tmBuff, 0, sizeof(struct tm));
	sscanf (startStr, AVM_DATE_FORMAT,
		    &tmBuff.tm_mon, &tmBuff.tm_mday, &tmBuff.tm_year,
		    &tmBuff.tm_hour, &tmBuff.tm_min, &tmBuff.tm_sec);
	if (tmBuff.tm_year < 70)
	    tmBuff.tm_year += 100;
	tmBuff.tm_mon--;
	startDate = mktime (&tmBuff) - timezone;
	memset ((char *)&tmBuff, 0, sizeof(struct tm));
	sscanf (endStr, AVM_DATE_FORMAT,
		    &tmBuff.tm_mon, &tmBuff.tm_mday, &tmBuff.tm_year,
		    &tmBuff.tm_hour, &tmBuff.tm_min, &tmBuff.tm_sec);
	if (tmBuff.tm_year < 70)
	    tmBuff.tm_year += 100;
	tmBuff.tm_mon--;
	endDate = mktime (&tmBuff) - timezone;
	/*
	compare dates in record with target date. If target date is between
	the two specified dates then return the current record else
	return zero.
	*/
	if ((targetDate >= startDate) && (targetDate < endDate))
	    break;
    }
    return ((int)rec);
}

/*
  get the minimum (oldest) start date from all records in the file. return the
  date as a time_t value.
*/
time_t  catGetEarliestStartDate (void)
{
    char startStr[AVM_STARTDATE_SIZE+1],endStr[AVM_ENDDATE_SIZE+1];
    time_t startDate,endDate,minDate;
    struct tm tmBuff;
    DB_RECORD_PTR rec;

    firstRec = 1;			/* start at top */
    minDate = (time_t)0;
    while ((rec = (DB_RECORD_PTR)catGetNextRawEntry(startStr,endStr,NULL,NULL,NULL,NULL)) != 0)
    {
	/* convert time strings to time_t */
	memset ((char *)&tmBuff, 0, sizeof(struct tm));
	sscanf (startStr, AVM_DATE_FORMAT,
		    &tmBuff.tm_mon, &tmBuff.tm_mday, &tmBuff.tm_year,
		    &tmBuff.tm_hour, &tmBuff.tm_min, &tmBuff.tm_sec);
	if (tmBuff.tm_year < 70)
	    tmBuff.tm_year += 100;
	tmBuff.tm_mon--;
	startDate = mktime(&tmBuff) - timezone;
	memset ((char *)&tmBuff, 0, sizeof(struct tm));
	sscanf (endStr, AVM_DATE_FORMAT,
		    &tmBuff.tm_mon, &tmBuff.tm_mday, &tmBuff.tm_year,
		    &tmBuff.tm_hour, &tmBuff.tm_min, &tmBuff.tm_sec);
	if (tmBuff.tm_year < 70)
	    tmBuff.tm_year += 100;
	tmBuff.tm_mon--;
	endDate = mktime (&tmBuff) - timezone;
	/*
	Collect the minimum startDate from each record in the file. If there are
	no records in the file then return 0.
	*/
	if ((minDate == 0) || (startDate < minDate))
	    minDate = startDate;
    }
    return (minDate);
}

/*
  get the maximum (newest) end date from all records in the file. return the
  date as a time_t value.
*/
time_t  catGetLatestEndDate (void)
{
    char startStr[AVM_STARTDATE_SIZE+1], endStr[AVM_ENDDATE_SIZE+1];
    time_t startDate, endDate,maxDate;
    struct tm tmBuff;
    DB_RECORD_PTR rec;

    firstRec = 1;			/* start at top */
    maxDate = (time_t)0;
    while ((rec = (DB_RECORD_PTR)catGetNextRawEntry(startStr,endStr,NULL,NULL,NULL,NULL)) != 0)
    {
	/* convert time strings to time_t */
	memset ((char *)&tmBuff, 0, sizeof(struct tm));
	sscanf (startStr, AVM_DATE_FORMAT,
		    &tmBuff.tm_mon, &tmBuff.tm_mday, &tmBuff.tm_year,
		    &tmBuff.tm_hour, &tmBuff.tm_min, &tmBuff.tm_sec);
	if (tmBuff.tm_year < 70)
	    tmBuff.tm_year += 100;
	tmBuff.tm_mon--;
	startDate = mktime (&tmBuff) - timezone;
	memset ((char *)&tmBuff, 0, sizeof(struct tm));
	sscanf (endStr, AVM_DATE_FORMAT,
		    &tmBuff.tm_mon, &tmBuff.tm_mday, &tmBuff.tm_year,
		    &tmBuff.tm_hour, &tmBuff.tm_min, &tmBuff.tm_sec);
	if (tmBuff.tm_year < 70)
	    tmBuff.tm_year += 100;
	tmBuff.tm_mon--;
	endDate = mktime (&tmBuff) - timezone;
	/*
	Collect the maximum endDate from each record in the file. If there are
	no records in the file then return 0.
	*/
	if (endDate > maxDate)
	    maxDate = endDate;
    }
    return (maxDate);
}
/*
  Search catalog for mathcing table name and return the rest of the data
*/
int catGetEntryByTableName (char *targetTable,
			    time_t *startDate,
			    time_t *endDate,
			    char *diskName,
			    char *pathName,
			    char *onlineFlag)
{
    char startStr[AVM_STARTDATE_SIZE+1], endStr[AVM_ENDDATE_SIZE+1], tableName[AVM_TABLENAME_SIZE+1];
    struct tm tmBuff;
    int len;
    int i;

    firstRec = 1;			/* start at top */
    len = strlen(targetTable);
    for(i = len-1; i >= 0; i--)
    {
	if (targetTable[i] != ' ')
	    break;
	else
	    targetTable[i] = 0;
    }
    while ((currentRec = catGetNextRawEntry (startStr, endStr, tableName, diskName, pathName, onlineFlag)) != 0)
    {
	if (!strcmp(targetTable,tableName))
	{
	    /* convert time strings to time_t */
	    memset ((char *)&tmBuff, 0, sizeof(struct tm));
	    sscanf (startStr, AVM_DATE_FORMAT,
			&tmBuff.tm_mon, &tmBuff.tm_mday, &tmBuff.tm_year,
			&tmBuff.tm_hour, &tmBuff.tm_min, &tmBuff.tm_sec);
	    if (tmBuff.tm_year < 70)
		tmBuff.tm_year += 100;
	    tmBuff.tm_mon--;
	    *startDate = mktime(&tmBuff) - timezone;
	    memset ((char *)&tmBuff, 0, sizeof(struct tm));
	    sscanf (endStr, AVM_DATE_FORMAT,
			&tmBuff.tm_mon, &tmBuff.tm_mday, &tmBuff.tm_year,
			&tmBuff.tm_hour, &tmBuff.tm_min, &tmBuff.tm_sec);
	    if (tmBuff.tm_year < 70)
		tmBuff.tm_year += 100;
	    tmBuff.tm_mon--;
	    *endDate = mktime (&tmBuff) - timezone;
	    break;
	}
    }
  return ((int)currentRec);
}
