/*bldDbRecType.c,v 1.1.1.1 1999/08/03 14:41:55 tis4000 Exp */

/*  @(#)bldDbRecType.c	1.6	6/7/91
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02	mm-dd-yy	iii	Comment
 * 	...
 */

#include <stdio.h>
#include <ctype.h>
#include <sys/file.h>		/* SUN support for disk io */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <dbRecType.h>
#include <sdrHeader.h>
#include <getSdrLine.h>

extern char    *calloc();
int             cmd_opts();
int             errno;

static struct recType *dbRecType = NULL;
static struct recType *pTdbRecType = NULL;
struct lineIn  *plineIn = NULL;

char            inFile[80];
char            outFile[80];
/*
 * BLDDBRECTYPE.C
 *
 * builds SDR file
 *
 * INPUT:     dbRecType.ascii
 *
 * OUTPUT     dbRecType.sdr
 */

main(argc, argv)
    char          **argv;
    int             argc;
{

    FILE           *inFile_fp;
    FILE           *fopen();
    FILE           *outFile_fp;
    long            rec_num;
    long            Ssize = 0;
    long            size = 0;
    long            len;
    long            i;
    long            dim_size;	/* dimension */
    long            rec_entries;/* highest entry number */
    int             rtnval;
    int             def_cnt = 0;
    char            name_buff[MAXBUF];
    char           *addr;
    char           *addr2;
    /* check cmd line usage */
    if (!cmd_opts(argc, argv))
	return -1;
    /* check alignment of structure to 4 bytes */
    if ((sizeof(struct recType) % 4)) {
	printf("bldDbRecType: struct recType size alignment error\n");
	return (-1);
    }
    /* allocate space for the getSdrLine input structure */
    if ((plineIn = (struct lineIn *)
	 calloc(1, (unsigned) sizeof(struct lineIn))) == NULL) {
	printf("bldDbRecType: calloc error\n");
	return (-1);
    }
    plineIn->fname = inFile;
    plineIn->line_no = 0;

    if ((inFile_fp = fopen((char *) inFile, "r")) == NULL) {
	printf("bldDbRecType: fopen error errno=%i\nfile=%s\n", errno, inFile);
	return (-1);
    }
    /* search for DIM control line */
    if ((getSdrLine(inFile_fp, plineIn, NO_QUOTES)) < 0)
	return (-1);
    if (plineIn->eof == EOF) {
	printf("bldDbRecType: Error - Unexpected EOF");
	return (-1);
    }
    if ((sscanf(plineIn->line, "%s %li %li", name_buff, &dim_size,
		&rec_entries)) != 3) {
	printf("bldDbRecType: illegal field in file %s\n", inFile);
	printf("line=%i\n", plineIn->line_no);
	return (-1);
    }
    if ((strcmp("DIM", name_buff)) != 0) {
	printf("bldDbRecType: DIM keyword expected - line=%i\n", plineIn->line_no);
	return (-1);
    }
    /* allocate space for temp structure */
    if ((pTdbRecType = (struct recType *)
	 calloc(1, (unsigned) sizeof(struct recType))) == NULL) {
	printf("bldDbRecType: calloc error\n");
	return (-1);
    }
    size = sizeof(struct recType);
    /* allocate space for temp  ptr array */
    if ((pTdbRecType->papName = (char **)
	 calloc(1, (unsigned) dim_size * sizeof(char *))) == NULL) {
	printf("bldDbRecType: calloc error\n");
	return (-1);
    }
    size += dim_size * sizeof(char *);

    FOREVER {
	if ((getSdrLine(inFile_fp, plineIn, TWO_QUOTES)) < 0)
	    return (-1);
	if (plineIn->eof == EOF) {
	    break;
	}
	if ((rtnval = sscanf(plineIn->line, "%li \"%[^\"]", &rec_num, name_buff)) != 2) {
	    printf("bldDbRecType: illegal field in file %s\n", inFile);
	    printf("line=%i\n", plineIn->line_no);
	    return (-1);
	}
	if (rec_num > dim_size) {
	    printf("bldDbRecType: Error - rec_num > dim_size\n");
	    printf("line=%i\n", plineIn->line_no);
	    return (-1);
	}
	/* check for duplicate entry number */
	if (pTdbRecType->papName[rec_num] != NULL) {
	    printf("bldDbRecType: Error - duplicate entry\n");
	    printf("lineno=%i file=%s\n", plineIn->line_no, inFile);
	}
	/* allocate space for string */
	len = strlen(name_buff);
	if ((pTdbRecType->papName[rec_num] = (char *)
	     calloc(1, (unsigned) len + 1)) == NULL) {
	    printf("bldDbRecType: calloc error\n");
	    return (-1);
	}
	strcpy(pTdbRecType->papName[rec_num], name_buff);
	Ssize += (long) len + 1;
	def_cnt++;
	if (def_cnt == rec_entries)
	    break;
    }
    if ((getc(inFile_fp)) != EOF) {
	printf("bldDbRecType: Error - extra character(s) in file %s\n", inFile);
	return (-1);
    }
    fclose(inFile_fp);

    pTdbRecType->number = dim_size;

    if (rec_entries != def_cnt) {
	printf("bldDbRecType: Error - rec_entries != def_cnt\n");
	printf("rec_entries=%li def_cnt=%li\n", rec_entries, def_cnt);
	return -1;
    }
    /* make sure we have padding to long words */
    Ssize = ((Ssize + sizeof(long) - 1) / sizeof(long)) * sizeof(long);
    size += Ssize;

    /* allocate the structure */
    if ((dbRecType = (struct recType *) calloc(1, (unsigned) size)) == NULL) {
	printf("bldDbRecType: calloc error\n");
	return (-1);
    }
    /* initialize addr to first byte past struct recType */
    addr = (char *) dbRecType + sizeof(struct recType);
    addr2 = (char *) dbRecType + size - Ssize;
    /* initialize the structure elements */
    dbRecType->number = pTdbRecType->number;
    dbRecType->papName = (char **) addr;
    /* copy each defined string */
    for (i = 0; i < dbRecType->number; i++) {
	if (pTdbRecType->papName[i] != NULL) {
	    dbRecType->papName[i] = (char *) addr2;
	    len = strlen(pTdbRecType->papName[i]);
	    strcpy(addr2, pTdbRecType->papName[i]);
	    addr2 += len + 1;
	}
    }

    /* write the SDR structure */
    if ((outFile_fp = fopen((char *) outFile, "w")) == NULL) {
	printf("bldDbRecType: fopen errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, outFile);
	return (-1);
    }

    /* write the header which tells us the type of data */
    fprintf(outFile_fp, "dbRecType\n");

    /* write the number of entries */
    fprintf(outFile_fp, "%d\n", dbRecType -> number);

    /* write each entry */
    for (i = 0; i < dbRecType -> number; i++)
	if (dbRecType -> papName[i])
	    fprintf(outFile_fp, "%s\n", dbRecType -> papName[i]);
	else
	    fprintf(outFile_fp, "\n");

    /* close the output file */
    fclose(outFile_fp);


    return (0);
}
/*
 * CMD_OPTS
 *
 * checks command line arg usage
 *
 */
cmd_opts(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    struct stat    *buf;
    int             c;
    int             err = 0;
    short           aflg = 0;
    short           oflg = 0;
    if ((buf = (struct stat *)
	 calloc(1, (unsigned) sizeof(struct stat))) == NULL) {
	printf("bldDbRecType: calloc error\n");
	return (0);
    }
    while ((c = getopt(argc, argv, "a:o:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile, optarg);
	    if (stat(inFile, buf) == -1) {
		printf("bldDbRecType: Can't stat file=%s errno=%i\n",
		       inFile, errno);
		err++;
	    }
	    break;
	case 'o':
	    oflg++;
	    strcpy(outFile, optarg);
	    break;
	}

    if (!aflg || !oflg || err) {
	printf("Usage: \n");
	printf("    <bldDbRecType -a dbRecType.ascii -o dbRecType.sdr>\n");
	return 0;
    }
    return 1;
}
/* end of bldDbRecType.c */
