/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		CHANNEL ACCESS ALIAS DEFINITIONS		*/
/*								*/
/*	The original implementation of channel access had	*/
/*	several aliases that really were #define macros		*/
/*	that gave the end user several different ways of	*/
/*	calling the same routine. These macros gave a very	*/
/*	small improvement in performance but made debugging	*/
/*	and understanding the system harder. Rather than have	*/
/*	these strange little creatures lurking, we'll make	*/
/*	real entry points out of them.				*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <stdio.h>

#include "dataTypes.h"
#include "cadef.h"

/******
	ca_search, really a call to ca_build_and_connect
******/

int ca_search(name, chanPtr)
char	*name;
void	*chanPtr;
{
    return(ca_build_and_connect(name, AGNT_NOCONN, 0, chanPtr, 0, 0, 0));
}


/******
	ca_build_channel, really a call to ca_build_and_connect
******/

int ca_build_channel(name, type, chanPtr, valPtr)
char	*name;
int	type;
void	*chanPtr;
void	*valPtr;
{
    return(ca_build_and_connect(name, (short) type, 1, chanPtr, valPtr, 0, 0));
}


/******
	ca_array_build, really a call to ca_build_and_connect
******/

int ca_array_build(name, type, count, chanPtr, valPtr)
char	*name;
int	type;
int	count;
void	*chanPtr;
void	*valPtr;
{
    return(ca_build_and_connect(name, (short) type, count, chanPtr, valPtr, 0, 0));
}


/******
	ca_bput, really a call to ca_array_put
******/

int ca_bput(chanPtr, valPtr)
void	*chanPtr;
void	*valPtr;
{
    return(ca_array_put(AGNT_STRING, 1, chanPtr, valPtr));
}


/******
	ca_rput, really a call to ca_array_put
******/

int ca_rput(chanPtr, valPtr)
void	*chanPtr;
void	*valPtr;
{
    return(ca_array_put(AGNT_FLOAT, 1, chanPtr, valPtr));
}


/******
	ca_put, really a call to ca_array_put
******/

int ca_put(type, chanPtr, valPtr)
int	type;
void	*chanPtr;
void	*valPtr;
{
    return(ca_array_put((short) type, 1, chanPtr, valPtr));
}


/******
	ca_bget, really a call to ca_array_get
******/

int ca_bget(chanPtr, valPtr)
void	*chanPtr;
void	*valPtr;
{
    return(ca_array_get((short) AGNT_STRING, 1, chanPtr, valPtr));
}


/******
	ca_rget, really a call to ca_array_get
******/

int ca_rget(chanPtr, valPtr)
void	*chanPtr;
void	*valPtr;
{
    return(ca_array_get(AGNT_FLOAT, 1, chanPtr, valPtr));
}


/******
	ca_get, really a call to ca_array_get
******/

int ca_get(type, chanPtr, valPtr)
int	type;
void	*chanPtr;
void	*valPtr;
{
    return(ca_array_get((short) type, 1, chanPtr, valPtr));
}


/******
	ca_bget_callback, really a call to ca_array_get_callback
******/

int ca_bget_callback(chanPtr, funcPtr, argPtr)
void	*chanPtr;
void	(*funcPtr)(struct event_handler_args);
void	*argPtr;
{
    return(ca_array_get_callback(AGNT_STRING, 1, chanPtr, funcPtr, argPtr));
}


/******
	ca_rget_callback, really a call to ca_array_get_get_callback
******/

int ca_rget_callback(chanPtr, funcPtr, argPtr)
void	*chanPtr;
void	(*funcPtr)(struct event_handler_args);
void	*argPtr;
{
    return(ca_array_get_callback(AGNT_FLOAT, 1, chanPtr, funcPtr, argPtr));
}


/******
	ca_get_callback, really a call to ca_array_get_callback
******/

int ca_get_callback(type, chanPtr, funcPtr, argPtr)
int	type;
void	*chanPtr;
void	(*funcPtr)(struct event_handler_args);
void	*argPtr;
{
    return(ca_array_get_callback((short) type, 1, chanPtr, funcPtr, argPtr));
}


/******
	ca_add_event, really a call to ca_add_array_event
******/

int ca_add_event(type, chanPtr, funcPtr, argPtr, eventPtr)
int	type;
void	*chanPtr;
void	(*funcPtr)(struct event_handler_args);
void	*argPtr;
void	*eventPtr;
{
    return(ca_add_array_event(type, 1, chanPtr, funcPtr, argPtr, (float) 0, (float) 0, (float) 0, eventPtr));
}


/******
	We don't intend to implement ca_add_delta_event
	We don't intend to implement ca_add_general_event
******/


/******
	ca_add_array_event, really a call to ca_add_masked_array_event
******/

int ca_add_array_event(int type, int count, void *chanPtr, void (*funcPtr)(struct event_handler_args), void *argPtr, 
				   float float1, float float2, float float3, void *eventPtr)
{
    return(ca_add_masked_array_event(type, count, chanPtr, funcPtr, argPtr, float1, float2, float3, eventPtr,
		AGNT_VALUE | AGNT_ALARM));
}


/****************************************************************/
/*								*/
/*	FORMER MACROS NOW IMPLEMENTED AS REAL FUNCTIONS		*/
/*								*/
/****************************************************************/

char *ca_name(pChid)
chid	pChid;
{
    return (pChid -> chanName);
}

short ca_field_type(pChid)
chid	pChid;
{
    if (pChid -> chanState != currOpen)
	return (TYPENOTCONN);
    else
    	return (pChid -> nativeType);
}

short ca_element_count(pChid)
chid	pChid;
{
    return (pChid -> nativeCount);
}

void *ca_puser(pChid)
chid	pChid;
{
    return (pChid -> puser);
}

enum channel_state ca_state(pChid)
chid	pChid;
{
    return ((enum channel_state)(pChid -> chanState));
}

int SEVCHK(num, msg)
int	num;
char	*msg;
{
    return 0;
}

int ca_change_puser(pChid, puser)
chid	pChid;
void	*puser;
{
    pChid -> puser = puser;
    return 0;
}

/******
	This routine returns the same text strings as the original
	macro implementation, but the method is different.
******/

static char AGNT_NORMAL_MSG[] = "Normal successful completion";
static char AGNT_ALLOCMEM_MSG[] = "Unable to allocate additional dynamic memory";
static char AGNT_TIMEOUT_MSG[] = "User specified timeout on IO operation expired";
static char AGNT_STRTOBIG_MSG[] = "The supplied string is unusually large";
static char AGNT_BADCHID_MSG[] = "Invalid or corrupt chid";
static char AGNT_BADTYPE_MSG[] = "The type you have requested from this channel is unknown";
static char AGNT_GETFAIL_MSG[] = "Could not perform a database value get for that channel";
static char AGNT_PUTFAIL_MSG[] = "Could not perform a database value put for that channel";
static char AGNT_ADDFAIL_MSG[] = "Could not perform a database event add for that channel";
static char AGNT_BADCOUNT_MSG[] = "Count requested inappropriate for that channel";
static char AGNT_EVDISALLOW_MSG[] = "The CA routine called is inappropriate for use within an event handler";
/* not used
static char AGNT_DISCONN_MSG[] = "Network connection lost";
*/
static char AGNT_NOCACTX_MSG[] = "Attempt to import from a task without a CA context failed";

char *ca_message(status)
int status;
{
    static char	unknownMessage[100];

    if (status == AGNT_NORMAL)
 	return (AGNT_NORMAL_MSG);
    if (status == AGNT_ALLOCMEM)
 	return (AGNT_ALLOCMEM_MSG);
    if (status == AGNT_TIMEOUT)
 	return (AGNT_TIMEOUT_MSG);
    if (status == AGNT_STRTOBIG)
 	return (AGNT_STRTOBIG_MSG);
    if (status == AGNT_BADCHID)
 	return (AGNT_BADCHID_MSG);
    if (status == AGNT_BADTYPE)
 	return (AGNT_BADTYPE_MSG);
    if (status == AGNT_GETFAIL)
 	return (AGNT_GETFAIL_MSG);
    if (status == AGNT_PUTFAIL)
 	return (AGNT_PUTFAIL_MSG);
    if (status == AGNT_ADDFAIL)
 	return (AGNT_ADDFAIL_MSG);
    if (status == AGNT_BADCOUNT)
 	return (AGNT_BADCOUNT_MSG);
    if (status == AGNT_EVDISALLOW)
 	return (AGNT_EVDISALLOW_MSG);
    if (status == AGNT_NOCACTX)
	return (AGNT_NOCACTX_MSG);
    
    sprintf(unknownMessage, "Unknown CA message code %d\n", status);
    return (unknownMessage);
}


/******
	The new CA doesn't have a need for exception event handlers. This is a stub for those
	who may have called one in the past.
******/

int ca_add_exception_event(pfunc, arg)
void	(*pfunc)();
void	*arg;
{
    return (AGNT_NORMAL);
}
