/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	BUILD A CHANNEL, READ IT, AND CREATE A MONITOR		*/
/*								*/
/*	This routine causes a channel to be created, a read	*/
/*	callback for the control variant of the native data	*/
/*	type to be performed, and a monitor for the status	*/
/*	variant to also be created. The purpose of this call	*/
/*	is to reduce the number of "round trips" that must	*/
/*	be made between the workstation from three to one	*/
/*	for graphics to appear on the screen. This is of	*/
/*	significant consequence on the Ethernet when there is	*/
/*	a large number of channels and is always significant	*/
/*	on slower media.					*/
/*								*/
/****************************************************************/

#ifdef vms
#include <selectvms.h>
#endif
#ifdef WIN32
#include <sys/types.h>
#include	<winsock.h>
extern int sendCAQ(short, char *, int);
#else
#ifdef vxWorks
#include <taskLib.h>
#endif vxWorks
#include <types.h>
#include <in.h>
#endif WIN32

#if defined(AXP) || defined(vxWorks) || defined(WIN32)
#include <string.h>
#endif AXP

#include "agentErrs.h"
#include "agentMsgs.h"
#include "cadef.h"
#include "caContext.h"

int ca_build_read_and_monitor(name, connFunc, pUser, readFunc, readArg, 
	monFunc, monArg, pChid, pEvid)
char	*name;
void	(*connFunc)(struct connection_handler_args);
void	*pUser;
void	(*readFunc)(struct event_handler_args);
void	*readArg;
void	(*monFunc)(struct event_handler_args);
void	*monArg;
chid	*pChid;
evid	*pEvid;
{
    chid	currChid;
    evid	currEvid;
    CATX	*ctx;
    short	length;
    short	msgType;
    struct
	{
	CH_BUILD	chBuild;
	char		nameBuff[sizeof(currChid -> chanName)];
	} m;

    /*
    Get our context
    */

    ctx = getContext();

    /*
    Validate that the name is not too big.
    */

    if (strlen(name) > sizeof(currChid -> chanName) - 1)
	return (AGNT_STRTOBIG);

    /*
    Get a chid and put it at the front of the chain.
    */

    currChid = newCHID();
    if (!currChid)
	return (AGNT_ALLOCMEM);
    *pChid = currChid;
    currChid -> pNextChan = ctx -> pFirstChid;
    ctx -> pFirstChid = currChid;
    if (currChid -> pNextChan)
	currChid -> pNextChan -> pPrevChan = (struct chanBlock *) currChid;

    /*
    Set up the chid by storing the name in the chid and record the connection 
    and read callback parameters.
    */

    currChid -> chanState = neverOpened;
    currChid -> nativeType = AGNT_NOCONN;
    strcpy(currChid -> chanName, name);
    currChid -> connFunc = connFunc;
    currChid -> puser = pUser;
    currChid -> callBack = readFunc;
    currChid -> callBackArg = readArg;
#ifdef vxWorks
    currChid -> paddr = (void *) taskIdSelf();
#endif

    /*
    Put an evid on the chid and record the event callback parameters there.
    */

    currEvid = newEVID();
    if (!currEvid)
	return (AGNT_ALLOCMEM);
    *pEvid = currEvid;
    currEvid -> mask = AGNT_VALUE | AGNT_ALARM;
    currEvid -> userFunc = monFunc;
    currEvid -> userArg = monArg;
    currEvid -> pChan = currChid;
    currEvid -> pNextEvnt = currChid -> pFirstEvnt;
    if (currEvid -> pNextEvnt)
	currEvid -> pNextEvnt -> pPrevEvnt = currEvid;
    currChid -> pFirstEvnt = currEvid;

    /*
    Format a message and sent it to chanAcq.
    */

#ifndef POINTER64
    m.chBuild.chHandle = (int) currChid;
#else
    m.chBuild.chHandle = currChid -> chidToken;
#endif
    m.chBuild.chType = htons(AGNT_ADAPTIVE);
    length = strlen(name) + 1;
    m.chBuild.chNameLen = htons(length);
    strcpy(m.nameBuff, name);
    msgType = htons(CHAN_BUILD);
    sendCAQ(msgType, (char *) &m, sizeof(CH_BUILD) + length);
    return (AGNT_NORMAL);
}


/****************************************************************/
/*								*/
/*	COMPLETE THE PROCESSING FOR CA_BUILD_READ_AND_MONITOR	*/
/*								*/
/*	This function must be called very early in the user's	*/
/*	connection handler to set up for the rest of the	*/
/*	processing.						*/
/*								*/
/****************************************************************/

int ca_complete_bld_read_mon(pchid)
chid	pchid;
{
    /*
    Validate the chid
    */

    if (!pchid)
	return (AGNT_BADCHID);

    /*
    When we set up the chid and the evid, we didn't fill in the
    read type, read count, monitor type, or the monitor count.
    Those fields can't be filled in until we know the native
    data type and count. chanAcq will guarantee that the build will
    be acknowledged before the read and monitored data arrive. This
    function's job is to set up the read type, read count, monitor
    type, and monitor count as a function of the native data type.
    */

    pchid -> buildType = dbf_type_to_DBR_CTRL(pchid -> nativeType);
    pchid -> buildCount = pchid -> nativeCount;
    if (pchid -> buildCount == 0)
	pchid -> buildCount = 1;

    pchid -> pFirstEvnt -> type = dbf_type_to_DBR_STS(pchid -> nativeType);
    pchid -> pFirstEvnt -> count = pchid -> nativeCount;
    if (pchid -> pFirstEvnt -> count == 0)
	pchid -> pFirstEvnt -> count = 1;
    return (AGNT_NORMAL);
}
