/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	    MAIN PROGRAM FOR CHANNEL ACCESS AGENTS		*/
/*								*/
/*	This is the main program for the agent architecture	*/
/*	for channel access. It is the command that is placed	*/
/*	into /etc/rc.local to launch the agents. This program	*/
/*	immediately forks and then launches the three active	*/
/*	parts of the system. After doing this, we can just	*/
/*	die silently.						*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <pathNames.h>
#include <fcntl.h>
#ifdef SOLARIS
#include <sys/time.h>
#include <sys/resource.h>
#endif

#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#include <process.h>
#endif WIN32

#include <cadef.h>

main()
{
#ifndef WIN32
    int	i;
#else
    WSADATA wsaData; 
    WORD    wVersionRequested = MAKEWORD(1, 1);
    int	    err;

/******
	Initialize
******/
    err = WSAStartup(wVersionRequested, &wsaData); 
 
    if (err != 0) 
	return 0; 
#endif WIN32

/******
	The next thing to do is to close all the file descriptors
	that we can and then open the three standard ones to a
	harmless I/O device, namely /dev/null. By doing this, we 
	save all of our child processes the trouble of doing this 
	because they will inherit all of our open files.
******/

#ifdef WIN32
    /* don't bother with the close in WIN32, we can control inheritence */
    _spawnl (_P_DETACH, NSNAME, NSNAME, (char *) 0);


    /*
    Channel Acquisition and Control
    */
    _spawnl (_P_DETACH, CANAME, CANAME, (char *) 0);

    /*
    Socket Creation and Maintenance
    */
    Sleep(2000);				/* Give chanAcq a chance to get going. */
    _spawnl (_P_DETACH, SMNAME, SMNAME, (char *) 0);
#else
#ifdef SOLARIS
    {
	struct rlimit rl;
	if (getrlimit (RLIMIT_NOFILE, &rl))
	{
	    /* Just use default */
	    for (i = 19; i >= 0; --i)
		close(i);
	}
	else
	{
	    for (i = rl.rlim_cur - 1; i >= 0; --i)
		close(i);
	}
    }
#else    
    for (i = getdtablesize() - 1; i >= 0; --i)
	close(i);
#endif
    i = open("/dev/null", O_RDWR);	/* stdin  */
    dup(i);				/* stdout */
    dup(i);				/* stderr */

/******
	Now, spawn each of our three children.
******/


    /*
    Name Server
    */

    i = fork();
    if (i == 0)
	{
	execl(NSNAME, NSNAME, (char *) 0);
	exit(1);
    	}


    /*
    Channel Acquisition and Control
    */

    sleep(4);
    i = fork();
    if (i == 0)
	{
	execl(CANAME, CANAME, (char *) 0);
	exit(1);
	}


    /*
    Socket Creation and Maintenance
    */

    sleep(2);				/* Give chanAcq a chance to get going. */
    i = fork();
    if (i == 0)
	{
	execl(SMNAME, SMNAME, (char *) 0);
	exit(1);
	}
#endif WIN32

    /*
    Now, wait 5 seconds and try to become a CA user. If we can't just try again. The
    purpose here is to not terminate this program until Channel Access is ready for
    the world to play with him.
    */

    do
	{
#ifdef WIN32
	    Sleep (5000);
#else
	    sleep(5);
#endif WIN32
	if (ca_task_initialize() == ECA_NORMAL)
    	    exit(0);
	} while (1);

	return 0;
}
