/*msgProc.c,v 1.1.1.1 1999/08/03 14:42:03 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	PROCESS MESSAGES FROM CHANNEL ACQUISITON AND CONTROL	*/
/*								*/
/*	This routine will read the socket and process the	*/
/*	following message types:				*/
/*								*/
/*	CHAN_BLDACK:	Build Acknowledgement			*/
/*	CHAN_NOTFOUND:	Channel Not Found (Build Reject)	*/
/*	CHAN_CONNEVNT:	Channel Connection Event		*/
/*	CHAN_DISCEVNT:	Channel Disconnect Event		*/
/*	CHAN_DRRDATA:	Direct Read Data			*/
/*	CHAN_MONDATA:	Monitor Data				*/
/*	CHAN_TRDATA:	Unsolicited Trend Data			*/
/*	CHAN_RESMODE:	Resource Mode Change			*/
/*	CHAN_PARMREPLY:	Parameter Fetch Reply			*/
/*	CHAN_ALARM:	Unsolicited Alarms			*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifndef vxWorks
#include <bcopy.h>
#endif vxWorks

#ifdef WIN32
#include <sys/types.h>
#include <winsock.h>
#include "NTUnixSubs.h"
#else WIN32
#include <types.h>
#include <in.h>
#ifdef vxWorks
#include <vxWorks.h>
#include <tickLib.h>
#include <selectLib.h>
#include <sys/times.h>
#include <string.h>
#else vxWorks
#ifdef AXP
#include <sys/time.h>
#else AXP
#include <time.h>
#include <sys/time.h>
#endif AXP
#endif vxWorks
#endif WIN32

#include "agentMsgs.h"
#include "dataTypes.h"
#include "agentErrs.h"
#include "cadef.h"
#include "caContext.h"
#include "tisAlarm.h"


int caMsgProc (float timeout, int persist)
{
  short				msgType;
  struct timeval		maxWait;
#ifdef vxWorks
  int				currTime;
  int				exitTime;
#else vxWorks
  struct timeval		currTime;
  struct timeval		exitTime;
#endif vxWorks
  fd_set			ioMask;
  CATX				*ctx;

  /*
  Get our context
  */

  ctx = getContext();
#ifdef vxWorks
  if (ctx -> importedContext)
	return (AGNT_TIMEOUT);
#endif vxWorks

  /*
  Check the port for I/O. Wait no longer than the user told
  us to wait. Once the first message arrives at the port,
  we'll grab all the messages out of the port until it's
  empty and then return. We will stay in this routine for
  the full amount of time that the user specified.
  */

  ctx -> firstSelDone = 0;
  maxWait.tv_sec = (long) timeout;
  maxWait.tv_usec = (long) ((timeout - (double) maxWait.tv_sec) * 1000000.0);

  /*
  If the user specified zero time, then treat it as a poll.
  Set our delay time to one microsecond.
  */

  if ((maxWait.tv_sec == 0) && (maxWait.tv_usec == 0))
	maxWait.tv_usec = 1;

  /*
  Loop while there's time left.
  */

  while ((maxWait.tv_sec >= 0) && (maxWait.tv_usec >= 0))
    {

    /*
    Get the time of day at the start of the loop
    */

#ifdef vxWorks
    currTime = tickGet();
#else vxWorks
#ifndef vms
    gettimeofday(&currTime, NULL);
#else vms
    sys$gettim(currTime);
#endif vms
#endif vxWorks

    ioMask = ctx -> caSocketMask;
    if (select(ctx -> caSocketFd + 1, &ioMask, NULL, NULL, &maxWait) <= 0)
	if (ctx -> firstSelDone)
	    return(AGNT_NORMAL);
	else
	    return(AGNT_TIMEOUT);

    /*
    There are messages. Read the socket and parse all that's there.
    */

    if (!getMore())
	return (AGNT_TIMEOUT);

    ctx -> firstSelDone = 1;
    while (ctx -> nchib)
	{
	if (recvCAQ((char *) &msgType, sizeof(msgType)) <= 0)
	    {

	    /*
	    The socket from Channel Acquisition and Control has failed.
	    Close it and print a message. We will reconnect later.
	    NOTE: Implementation of reconnection is being deferred
	    until we try CA applications in different nodes than
	    the one containing Channel Acquisition and Control. Until
	    that time, it is highly unlikely that our socket will
	    ever get broken.
	    */

	    return (AGNT_TIMEOUT);
	    }

	msgType = ntohs(msgType);
	if (msgType == CHAN_BLDACK)
	    {

	    /*
	    Build Acknowledgement. Get the message and fill out the data
	    fields in the chid.
	    */

	    CH_BLDACK	bldAck;

	    if (recvCAQ((char *) &bldAck, sizeof(bldAck)) <= 0)
		return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &bldAck, (char *) 0);
	    }

	else if (msgType == CHAN_NOTFOUND)
	    {

	    /*
	    Channel not found. Set the native data type in the chid to
	    NOT CONNECTED. The only reason for this message is to let
	    the application know as soon as possible (in a compatible
	    manner) that his connection attempt failed.
	    */

	    CH_NOTFOUND	notFound;

	    if (recvCAQ((char *) &notFound, sizeof(notFound)) <= 0)
		return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &notFound, (char *) 0);
	    }

	else if (msgType == CHAN_CONNEVNT)
	    {

	    /*
	    Channel Connection Event. Get the message.
	    */

	    CH_CONNEVNT		connEvnt;

	    if (recvCAQ((char *) &connEvnt, sizeof(connEvnt)) <= 0)
		return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &connEvnt, (char *) 0);
	    }

	else if (msgType == CHAN_DISCEVNT)
	    {

	    /*
	    Channel Disconnect Event. Get the message.
	    */

	    CH_DISCEVNT		discEvnt;

	    if (recvCAQ((char *) &discEvnt, sizeof(discEvnt)) <= 0)
		return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &discEvnt, (char *) 0);
	    }

	else if (msgType == CHAN_DRRDATA)
	    {

	    /*
	    Data from a direct read. Get the header and the data.
	    */

	    CH_DRRDATA		drData;
	    int			dataLength1;

	    if (recvCAQ((char *) &drData, sizeof(drData)) <= 0)
		return (AGNT_TIMEOUT);
	    dataLength1 = ntohs(drData.chDataLen);
	    while (dataLength1 > ctx -> nchib)
		if (!getMore())
		    return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &drData, ctx -> caDataBuff + ctx -> bufOutPos);
	    ctx -> bufOutPos += dataLength1;
	    ctx -> nchib -= dataLength1;
	    }

	else if (msgType == CHAN_MONDATA)
	    {

	    /*
	    Data from a monitor. Get the header and the data.
	    */

	    CH_MONDATA			monData;
	    int				dataLength2;

	    if (recvCAQ((char *) &monData, sizeof(monData)) <= 0)
		return (AGNT_TIMEOUT);
	    dataLength2 = ntohs(monData.chDataLen);
	    while (dataLength2 > ctx -> nchib)
		if (!getMore())
		    return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &monData, ctx -> caDataBuff + ctx -> bufOutPos);
	    ctx -> bufOutPos += dataLength2;
	    ctx -> nchib -= dataLength2;
	    }

	else if (msgType == CHAN_TRDATA)
	    {

	    /*
	    Data for a trend handler. Get the header and the data.
	    */

	    CH_TRDATA		trendData;
	    int			dataLength3;

	    if (recvCAQ((char *) &trendData, sizeof(trendData)) <= 0)
		return (AGNT_TIMEOUT);
	    dataLength3 = ntohs(trendData.chNameLen) + ntohs(trendData.chDataLen);
	    while (dataLength3 > ctx -> nchib)
		if (!getMore())
		    return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &trendData, ctx -> caDataBuff + ctx -> bufOutPos);
	    ctx -> bufOutPos += dataLength3;
	    ctx -> nchib -= dataLength3;
	    }

	else if (msgType == CHAN_CLOSEACK)
	    {
	    CH_CLOSEACK		closeAck;

	    if (recvCAQ((char *) &closeAck, sizeof(closeAck)) <= 0)
		return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &closeAck, (char *) 0);
	    }

	else if (msgType == CHAN_RESMODE)
	    {

	    /*
	    Resource mode change for a trend or alarm handler. Get the message.
	    */

	    CH_RESMODE		resMode;

	    if (recvCAQ((char *) &resMode, sizeof(resMode)) <= 0)
		return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &resMode, (char *) 0);
	    }

	else if (msgType == CHAN_PARMREPLY)
	    {

	    /*
	    Parameters fetched without channel connections.
	    */

	    CH_PARMREPLY	parmReply;
	    int			dataLength4;

	    if (recvCAQ((char *) &parmReply, sizeof(parmReply)) <= 0)
		return (AGNT_TIMEOUT);
	    dataLength4 = ntohs(parmReply.chDataLen);
	    while (dataLength4 > ctx -> nchib)
		if (!getMore())
		    return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &parmReply, ctx -> caDataBuff + ctx -> bufOutPos);
	    ctx -> bufOutPos += dataLength4;
	    ctx -> nchib -= dataLength4;
	    }

	else if (msgType == CHAN_ALARM)
	    {

	    /*
	    Unsolicited Alarm Data
	    */

	    CH_ALARM		alarm;
	    int			dataLength5;

	    if (recvCAQ((char *) &alarm, sizeof(alarm)) <= 0)
		return (AGNT_TIMEOUT);
	    dataLength5 = ntohs(alarm.alarmDataLen);
	    while (dataLength5 > ctx -> nchib)
		if (!getMore())
			return (AGNT_TIMEOUT);
	    caMsgHandler(msgType, (char *) &alarm, ctx -> caDataBuff + ctx -> bufOutPos);
	    ctx -> bufOutPos += dataLength5;
	    ctx -> nchib -= dataLength5;
	    }
	}

  /*
  See if the caller wants to stay here for the full amount of time or just until I/O is complete.
  */

  if (!persist)
    break;

  /*
  Get time time of day at the end of the loop and subtract the amount of time that we
  were in the loop from the amount of time that the user wanted to be here.
  */

#ifdef vxWorks
  exitTime = tickGet();
  exitTime -= currTime;
  maxWait.tv_sec = exitTime / 60;
  maxWait.tv_usec = (exitTime % 60) * 16667; /* There are 16667 usec per tick */
#else vxWorks
#ifndef vms
  gettimeofday(&exitTime, NULL);
  exitTime.tv_sec -= currTime.tv_sec;
  exitTime.tv_usec -= currTime.tv_usec;
  if (exitTime.tv_usec < 0)
    {
    exitTime.tv_usec += 1000000;
    exitTime.tv_sec--;
    }
  maxWait.tv_sec -= exitTime.tv_sec;
  maxWait.tv_usec -= exitTime.tv_usec;
#else vms
    sys$gettim(exitTime);
    lib$sub_times(exitTime, currTime, deltaTime);
    lib$cvtf_from_internal_time(&opCode, &secsActive, deltaTime);
    maxWait.tv_sec -= (int) secsActive;
    maxWait.tv_usec -= (int)((secsActive - (int) secsActive) * 1000000.0);
#endif vms
  if (maxWait.tv_usec < 0)
    {
    maxWait.tv_usec += 1000000;
    maxWait.tv_sec--;
    }
#endif vxWorks
  }

  /*
  Done Deal
  */

  return (AGNT_NORMAL);
}

static CH_BLDACK	guOldBldack;

int caMsgHandler(msgType, msgBuff, auxBuff)
short	msgType;
char	*msgBuff;
char	*auxBuff;
{
    CATX	*ctx;
    struct event_handler_args	evntHdlrArgs;
    chid	pChid;
    char	dbrBuff[2048];

    ctx = getContext();
    if (msgType == CHAN_BLDACK)
	{
	CH_BLDACK	*bldAck;

	bldAck = (CH_BLDACK *) msgBuff;
guOldBldack = *bldAck;
#ifndef POINTER64
	pChid = (chid) (bldAck -> chHandle);
#else
	pChid = Pointer64fromToken(bldAck -> chHandle);
#endif
#ifdef vxWorks
        ca_import((int)pChid -> paddr);
	if (!ctx)
	    ctx = getContext();
	ctx = ctx -> importedContext;
#endif vxWorks
	if (!caValidChid(pChid, ctx))
	    return 0;
	if (pChid -> chanState == closed)
	    return 0;
	if (pChid -> pPrevChan == pChid)
	    return 0;
	pChid -> nativeType = ntohs(bldAck -> chDataType);
	pChid -> nativeCount = ntohs(bldAck -> chElementCount);
	if (pChid -> nativeCount == 0)
	    pChid -> nativeCount = 1;
	pChid -> handle = bldAck -> chIocHandle;
	pChid -> chanState = currOpen;

	/*
	If there is a connection event handler, call that routine.
	*/

	if (pChid -> connFunc)
	    {
	    struct connection_handler_args	openConnArgs;

	    openConnArgs.chid = pChid;
	    openConnArgs.op = CA_OP_CONN_UP;
	    (*(pChid -> connFunc))(openConnArgs);
	    }
	}

    else if (msgType == CHAN_NOTFOUND)
	{
	CH_NOTFOUND	*notFound;

	notFound = (CH_NOTFOUND *) msgBuff;
#ifndef POINTER64
	pChid = (chid) (notFound -> chHandle);
#else
	pChid = Pointer64fromToken(notFound -> chHandle);
#endif
#ifdef vxWorks
        ca_import((int)pChid -> paddr);
	if (!ctx)
	    ctx = getContext();
	ctx = ctx -> importedContext;
#endif vxWorks
	if (!caValidChid(pChid, ctx))
	    return 0;
	if (pChid -> chanState == closed)
	    return 0;
	if (pChid -> pPrevChan == pChid)
	    return 0;
	pChid -> nativeType = AGNT_NOCONN;
	}

    else if (msgType == CHAN_CONNEVNT)
	{
	CH_CONNEVNT	*connEvnt;

	/*
	Call the user's connection function if he has defined one.
	*/

	connEvnt = (CH_CONNEVNT *) msgBuff;
#ifndef POINTER64
	pChid = (chid) (connEvnt -> chHandle);
#else
	pChid = Pointer64fromToken(connEvnt -> chHandle);
#endif
#ifdef vxWorks
        ca_import((int)pChid -> paddr);
	if (!ctx)
	    ctx = getContext();
	ctx = ctx -> importedContext;
#endif vxWorks
	if (!caValidChid(pChid, ctx))
	    return 0;
	if (pChid -> chanState == closed)
	    return 0;
	if (pChid -> pPrevChan == pChid)
	    return 0;
	pChid -> chanState = currOpen;
	if (pChid -> connFunc)
	    {

	    struct connection_handler_args	connHandArgs;

	    connHandArgs.chid = pChid;
	    connHandArgs.op = CA_OP_CONN_UP;
	    (*(pChid -> connFunc))(connHandArgs);
	    }
	}

    else if (msgType == CHAN_DISCEVNT)
	{
	CH_DISCEVNT	*discEvnt;
	   
	/*
	Call the user's connection function if he has defined one.
	*/

	discEvnt = (CH_DISCEVNT *) msgBuff;
#ifndef POINTER64
	pChid = (chid) (discEvnt -> chHandle);
#else
	pChid = Pointer64fromToken(discEvnt -> chHandle);
#endif
#ifdef vxWorks
        ca_import((int)pChid -> paddr);
	if (!ctx)
	    ctx = getContext();
	ctx = ctx -> importedContext;
#endif vxWorks
	if (!caValidChid(pChid, ctx))
	    return 0;
	if (pChid -> chanState == closed)
	    return 0;
	if (pChid -> pPrevChan == pChid)
	    return 0;
	pChid -> chanState = prevOpened;
	if (pChid -> connFunc)
	    {
	    struct connection_handler_args	connHandArgs;

	    connHandArgs.chid = pChid;
	    connHandArgs.op = CA_OP_CONN_DOWN;
	    (*(pChid -> connFunc))(connHandArgs);
	    }
	}

    else if (msgType == CHAN_DRRDATA)
	{
	CH_DRRDATA	*drData;

	/*
	If there is a callback pointer, set up to call that routine.
	This is precisely the same as handling an event which is 
	duplicated in the processing of a CHAN_MONDATA message.
	*/

	drData = (CH_DRRDATA *) msgBuff;
#ifndef POINTER64
	pChid = (chid) (drData -> chHandle);
#else
	pChid = Pointer64fromToken(drData -> chHandle);
#endif
#ifdef vxWorks
        ca_import((int)pChid -> paddr);
	if (!ctx)
	    ctx = getContext();
	ctx = ctx -> importedContext;
#endif vxWorks
	if (!caValidChid(pChid, ctx))
	    return 0;
	if (pChid -> chanState == closed)
	    return 0;
	if (pChid -> pPrevChan == pChid)
	    return 0;
	if (pChid -> callBack)
	    {
	    evntHdlrArgs.usr = pChid -> callBackArg;
	    evntHdlrArgs.chid = pChid;
	    evntHdlrArgs.type = pChid -> buildType;
	    evntHdlrArgs.count = pChid -> buildCount;
	    evntHdlrArgs.dbr = (void *) dbrBuff;

	    /****** HERE ******/
	    caConvType(auxBuff, ntohs(drData -> chDataType), dbrBuff, 
		(short) pChid -> buildType, pChid -> buildCount);
	    (*(pChid -> callBack))(evntHdlrArgs);
	    pChid -> callBack = 0;
	    }

	else
	    {
	    /*
	    This wasn't a read with a requested callback.
	    If the user has no direct read pending, drop the data
	    on the floor.
	    */

	    if (pChid -> buildValue)
	        {

	        /*
	        Now, convert the data from agent format to the traditional
	        dbr format. This is necessary to insure source compatibility
	        for existing applications.
	        */
	
	        caConvType(auxBuff, ntohs(drData -> chDataType), (char *) (pChid -> buildValue), 
			pChid -> buildType, pChid -> buildCount);
	    
	        /*
	        Finally, indicate that the read is no longer pending. We'll do that by
	        clearing the direct read data address.
	        */

	        pChid -> buildValue = 0;
	        }
	    }
	}

    else if (msgType == CHAN_MONDATA)
	{
	CH_MONDATA	*monData;
	
	/*
	Before we can dispatch the event, we have to make sure that the
	chid has an evid and that the evid contains a valid function pointer.
	*/

	monData = (CH_MONDATA *) msgBuff;
#ifndef POINTER64
	pChid = (chid) (monData -> chHandle);
#else
	pChid = Pointer64fromToken(monData -> chHandle);
#endif
#ifdef vxWorks
        ca_import((int)pChid -> paddr);
	if (!ctx)
	    ctx = getContext();
	ctx = ctx -> importedContext;
#endif vxWorks
	if (!caValidChid(pChid, ctx))
	    return 0;
	if (pChid -> chanState == closed)
	    return 0;
	if (pChid -> pPrevChan == pChid)
	    return 0;
	if ((pChid -> pFirstEvnt) && (pChid -> pFirstEvnt -> userFunc))
	    {
	    /*
	    Build the structure, convert the data, and then call his routine.
	    */

	    evntHdlrArgs.usr = pChid -> pFirstEvnt -> userArg;
	    evntHdlrArgs.chid = pChid;
	    evntHdlrArgs.type = pChid -> pFirstEvnt -> type;
	    evntHdlrArgs.count = pChid -> pFirstEvnt -> count;
	    evntHdlrArgs.dbr = (void *) dbrBuff;

	    caConvType(auxBuff, ntohs(monData -> chDataType), dbrBuff, 
	        (short) pChid -> pFirstEvnt -> type, pChid -> pFirstEvnt -> count);

	    (*(pChid -> pFirstEvnt -> userFunc))(evntHdlrArgs);
	    }
	}

    else if (msgType == CHAN_CLOSEACK)
	{
	CH_CLOSEACK	*closeAck;
	evid		currEvnt;
	evid		oldEvnt;

	/*
	Channel close acknowledgement. It is now safe to unchain this chid
	from the list and put it into the free pool.
	*/

	closeAck = (CH_CLOSEACK *) msgBuff;


#ifndef POINTER64
	pChid = (chid) (closeAck -> chHandle);
#else
	pChid = Pointer64fromToken(closeAck -> chHandle);
#endif
#ifdef vxWorks
        ca_import((int)pChid -> paddr);
	if (!ctx)
	    ctx = getContext();
	ctx = ctx -> importedContext;
#endif vxWorks
	if (!caValidChid(pChid, ctx))
	    return 0;
	if (pChid -> pPrevChan == pChid)
	    return 0;

	/*
	Free up any event headers
	*/

	currEvnt = pChid->pFirstEvnt;

	while (currEvnt)
	    {
	    oldEvnt = currEvnt;
	    currEvnt = currEvnt -> pNextEvnt;
	    deleteEVID(oldEvnt);
	    }

	/*
	Now unchain the chid and free it.
	*/

	if (pChid -> pNextChan)
	    pChid -> pNextChan -> pPrevChan = pChid -> pPrevChan;
	if (pChid -> pPrevChan)
	    pChid -> pPrevChan -> pNextChan = pChid -> pNextChan;
	else
	    ctx -> pFirstChid = pChid -> pNextChan;

	pChid -> pPrevChan = pChid;	/* Indicate closed */
	deleteCHID(pChid);
	}

#ifndef vxWorks
    else if (msgType == CHAN_TRDATA)
	{
	CH_TRDATA	*trendData;
	char		*pName;
	short		oldType;

	/*
	Unsolicited trend data.
	*/

	trendData = (CH_TRDATA *) msgBuff;
	pName = auxBuff;

	/*
	Figure out the destination data type for the user. We could do this by just
	asking off the agent bits, but that makes us more sensitive to changes that
	might be made in the future.
	*/

	trendData -> chDataType = ntohs(trendData -> chDataType);
	oldType = trendData -> chDataType & 0xff;
	if (oldType == AGNT_STRING)
	    oldType = OLD_DBR_TIME_STRING;
	else if (oldType == AGNT_INT)
	    oldType = OLD_DBR_TIME_INT;
	else if (oldType == AGNT_FLOAT)
	    oldType = OLD_DBR_TIME_FLOAT;
	else if (oldType == AGNT_ENUM)
	    oldType = OLD_DBR_TIME_ENUM;
	else if (oldType == AGNT_CHAR)
	    oldType = OLD_DBR_TIME_CHAR;
	else if (oldType == AGNT_LONG)
	    oldType = OLD_DBR_TIME_LONG;
	else if (oldType == AGNT_DOUBLE)
	    oldType = OLD_DBR_TIME_DOUBLE;
	else
	    return 0;

	/*
	Convert the data to that type and then call the user's callback routine to
	give him the data.
	*/

	caConvType(auxBuff + ntohs(trendData -> chNameLen), trendData  -> chDataType, dbrBuff, oldType, 1);
	if (ctx -> trendEventFunc)
	    (*(ctx -> trendEventFunc))(pName, oldType, 1, dbrBuff);
	}

    else if (msgType == CHAN_RESMODE)
	{
	CH_RESMODE	*resMode;

	/*
	Resource mode change. Let the appropriate handler callbacks know about this.
	*/


	resMode = (CH_RESMODE *) msgBuff;
	resMode -> chResStatus = ntohs(resMode -> chResStatus);

	if (ctx -> trendConnFunc)
	    (*(ctx -> trendConnFunc))(resMode -> chResName, resMode -> chResStatus);
	if (ctx -> alarmConnFunc)
	    (*(ctx -> alarmConnFunc))(resMode -> chResName, resMode -> chResStatus);
	}

    else if (msgType == CHAN_PARMREPLY)
	{
	CH_PARMREPLY	*parmReply;

	/*
	Connectionless parameter reply. Convert the data and then present it to the
	callback routine, if one has been defined.
	*/

	parmReply = (CH_PARMREPLY *) msgBuff;
	parmReply -> chParmType = ntohs(parmReply -> chParmType);
	caConvParamData(auxBuff, parmReply -> chParmType, dbrBuff);
	if (ctx -> paramDataFunc)
	    (*(ctx -> paramDataFunc))(parmReply -> chParmType, dbrBuff, parmReply -> chArgument);
	}

    else if (msgType == CHAN_ALARM)
	{
	CH_ALARM	*alarm;
	tisAlarm	alarmData;

	/*
	An unsolicited alarm. Parse the input string and pass it on to the caller.
	*/

	alarm = (CH_ALARM *) msgBuff;
	alarmData.alarmTime.secPastEpoch = ntohl(alarm -> alarmTime.secPastEpoch);
	alarmData.alarmTime.nsec = ntohl(alarm -> alarmTime.nsec);
	alarmData.alarmSerial = alarm -> alarmSerial;
	alarmData.stateLess = alarm -> alarmHandle;
	alarmData.alarmStat = ntohs(alarm -> alarmStat);
	alarmData.alarmSevr = ntohs(alarm -> alarmSevr);
	alarmData.alarmAck = ntohs(alarm -> alarmAck);
	alarmData.alarmEnabled = ntohs(alarm -> alarmEnabled);
	alarmData.alarmOnscan = ntohs(alarm -> alarmOnscan);
	caConvAlarmData(auxBuff, &alarmData);
	if (ctx -> alarmEventFunc)
	    (*(ctx -> alarmEventFunc))(&alarmData);
	}

#endif vxWorks
    return 0;
}

/****************************************************************/
/*								*/
/*	VALIDATE THE CHID TO KEEP US FROM CRASHING		*/
/*								*/
/****************************************************************/

int caValidChid(pChid, ctx)
chid	pChid;
CATX	*ctx;
{
    chid	currChid;

    if (ctx)
    {
	currChid = ctx -> pFirstChid;
	while (currChid)
	{
	    if (currChid == pChid)
		return (1);
	    currChid = currChid -> pNextChan;
	}
    }
    return (0);
}
