/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		NAME SERVER FOR CHANNEL ACCESS			*/
/*								*/
/*	This program resolves channel names to a list of	*/
/*	possible Internet addresses and a resource name.	*/
/*	Other services include returning a list of the		*/
/*	Internet address for all the IOCs that we know		*/
/*	and also returning a list of all the resources		*/
/*	that are in the system.					*/
/*								*/
/*	Endian issues: We may be a different endian from any	*/
/*	user.							*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifndef AXP
#include <bcopy.h>
#endif
#ifdef WIN32
#include <winsock.h>
#else
#include <types.h>
#include <socket.h>
#include <errno.h>
#include <in.h>
#include <tcp.h>
#include <netdb.h>
#if (defined(ultrix) || defined(AXP))
#include <ioctl.h>
#else
#include <filio.h>
#endif
#include <errno.h>
#include <in.h>
#include <tcp.h>
#include <netdb.h>
#ifdef AXP
#include <sys/time.h>
#else
#include <time.h>
#include <sys/time.h>
#endif AXP
#endif WIN32

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#ifdef WIN32
typedef void *caddr_t;
#endif WIN32

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#ifndef	WIN32
#    include "channelControl.h"	/* def of SOCKET */
#endif	WIN32
#include "agentPorts.h"
#include "agentMsgs.h"
#include "pathNames.h"
#include "dbRecType.h"
#include "msgMon.h"


/****************************************************************/
/*								*/
/*		GLOBAL DATA FOR THIS PROCESS			*/
/*								*/
/****************************************************************/
#define MAX_NAME_CLIENTS 100

SOCKET			tcpFd;
SOCKET			udpFd;
int			maxFd;
int			clientMax;
SOCKET			clientSockets [MAX_NAME_CLIENTS];
extern int		errno;
fd_set			fdMask;
short			msgType;

SOCKET			lastFd = INVALID_SOCKET;
char			sockBuff[1024];
int			sockSize = 0;

C4CODE			codeBase;

D4DATA			*pvmap;
F4FIELD			*pvmapName;
F4FIELD			*pvmapResource;
F4FIELD			*pvmapCreationTime;
F4FIELD			*pvmapChecksum;
F4FIELD			*pvmapPointType;
T4TAG			*pvmapTag;

D4DATA			*reslist;
F4FIELD			*reslistResName;
F4FIELD			*reslistIOC1;
F4FIELD			*reslistIOC2;
F4FIELD			*reslistDPA;
T4TAG			*reslistTag;

D4DATA			*ioclist;
F4FIELD			*ioclistIOCName;
T4TAG			*ioclistTag;

struct iocCache
    {
    char		iocName[16];
    struct in_addr	inetAddr;
    struct iocCache	*nextIoc;
    } *iocCache = 0;
int			numIocs = 0;

struct resCache
    {
    char		resName[16];
    struct in_addr	inetAddr1;
    struct in_addr	inetAddr2;
    struct resCache	*nextRes;
    } *resCache = 0;
int			numRes = 0;

struct resCache		*dpaCache = 0;
int			numDpa = 0;
struct timeval		oneSecond = {5, 0};
struct timeval		pollVal = {0, 0};

struct recType		*dbRecType;

#define			REOPEN_TIME	30
int			lastReopen;




/****************************************************************/
/*								*/
/*			FUNCTION PROTOTYPES			*/
/*								*/
/****************************************************************/

void trimCpy (char *destination, char *source);
struct in_addr findIOC (char *iocName);
struct resCache *findRes (char *name);
int sendClient (SOCKET fd, char *buff, int nbytes);
int nameFlush (SOCKET fd, char *buff, int nbytes);
int recvClient (SOCKET fd, char *buff, int nbytes);
void closeClient (SOCKET sock);
int getInt (FILE *fd);
void trimNewLine (char *string);

/****************************************************************/
/*								*/
/*			THE MAIN PROGRAM			*/
/*								*/
/****************************************************************/

main()
{
    int			optVal = 1;
    struct sockaddr_in	portAddr;
    struct iocCache	*currIocNode;
    struct resCache	*currResNode;
    struct resCache	*currDpaNode;
    int			dpaNode = 0;
    struct hostent	*hostent;
    char		tempName[132];
    fd_set		fdList;
    int		fdIndex;
    SOCKET		newFd;
    int			addrSize;
    int			status;
    short		length;
    FILE		*sdrFile;
    int			numRecs;
    int			i;

/******
	Initialize
******/
#ifdef WIN32
    WSADATA wsaData; 
    WORD    wVersionRequested = MAKEWORD(1, 1);
    int	    err;

    err = WSAStartup(wVersionRequested, &wsaData); 
 
    if (err != 0) 
	return 0; 
#endif WIN32

    tisOpenLog("nameSrv", 3, 3, NULL, NULL);
    serverInit(NSPATH, NSLOCK);

/******
	Create the UDP port that IOCs will use to connect to us.
******/

    udpFd = socket(PF_INET, SOCK_DGRAM, 0);
    maxFd = udpFd;
    setsockopt(udpFd, SOL_SOCKET, SO_REUSEADDR, (caddr_t) &optVal, sizeof(optVal));
    portAddr.sin_family = AF_INET;
    portAddr.sin_addr.s_addr = INADDR_ANY;
    portAddr.sin_port = htons(NAM_PORT);
    if (bind(udpFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1)
	{
	tisSysLog(0, 0, "nameSrv can't bind to UDP port: %m");
	exit(1);
	}
    FD_ZERO(&fdMask);
    FD_SET(udpFd, &fdMask);

/******
	Create the TCP port that users will use to connect to us.
******/

    tcpFd = socket(PF_INET, SOCK_STREAM, 0);
    setsockopt(tcpFd, SOL_SOCKET, SO_REUSEADDR, (caddr_t) &optVal, sizeof(optVal));
    portAddr.sin_family = AF_INET;
    portAddr.sin_addr.s_addr = INADDR_ANY;
    portAddr.sin_port = htons(NAM_PORT);
    if (bind(tcpFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1)
	{
	tisSysLog(0, 0, "nameSrv can't bind to TCP port: %m");
	exit(1);
	}
#ifndef WIN32
    if (tcpFd > maxFd)
	maxFd = tcpFd;
#endif WIN32
    listen(tcpFd, 2);
    FD_SET(tcpFd, &fdMask);

/******
	Establish connections to all the dBase files that we need.
******/

    d4init(&codeBase);
#if defined (WIN32) || defined(USECB63)
    codeBase.errFieldName = 0;
    codeBase.errOpen = 0;
    codeBase.errTagName = 0;
    codeBase.readLock = 0;
#else
    codeBase.field_name_error = 0;
    codeBase.open_error = 0;
    codeBase.tag_name_error = 0;
    codeBase.read_lock = 0;
#endif WIN32
#ifndef CB51
    codeBase.wait = 0;
#endif
    /*
    PVMAP.DBF
    */

#ifdef WIN32
    if ((pvmap = d4open(&codeBase, "\\tis4000db\\pvmap.dbf")) == 0)
#else
    if ((pvmap = d4open(&codeBase, "/tis4000db/pvmap.dbf")) == 0)
#endif WIN32
	{
	tisSysLog(0, 0, "nameSrv can't open PVMAP.");
	exit(1);
	}
    if ((pvmapName = d4field(pvmap, "NAME")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find field NAME in PVMAP.");
	exit(1);
	}
    if ((pvmapResource = d4field(pvmap, "RESRCNAME")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find field RESRCNAME in PVMAP.");
	exit(1);
	}
    if ((pvmapCreationTime = d4field(pvmap, "TIMESTAMP")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find field TIMESTAMP in PVMAP.");
	exit(1);
	}
    if ((pvmapChecksum = d4field(pvmap, "CHECKSUM")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find field CHECKSUM in PVMAP.");
	exit(1);
	}
    if ((pvmapPointType = d4field(pvmap, "TYPE")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find field TYPE in PVMAP.");
	exit(1);
	}
    if ((pvmapTag = d4tag(pvmap, "NAME")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find tag NAME in PVMAP.");
	exit(1);
	}
    d4tag_select(pvmap, pvmapTag);
    d4unlock_all(pvmap);
	
    /*
    IOCLIST.DBF
    */

#ifdef WIN32
    if ((ioclist = d4open(&codeBase, "\\tis4000db\\ioclist.dbf")) == 0)
#else
    if ((ioclist = d4open(&codeBase, "/tis4000db/ioclist.dbf")) == 0)
#endif WIN32
	{
	tisSysLog(0, 0, "nameSrv can't open IOCLIST.");
	exit(1);
	}
    if ((ioclistIOCName = d4field(ioclist, "IOCNAME")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find field IOCNAME in IOCLIST.");
	exit(1);
	}
    if ((ioclistTag = d4tag(ioclist, "IOCNAME")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find tag IOCNAME in IOCLIST.");
	exit(1);
	}
    d4tag_select(ioclist, ioclistTag);
    d4unlock_all(ioclist);

    /*
    RESLIST.DBF
    */

#ifdef WIN32
    if ((reslist = d4open(&codeBase, "\\tis4000db\\reslist.dbf")) == 0)
#else
    if ((reslist = d4open(&codeBase, "/tis4000db/reslist.dbf")) == 0)
#endif WIN32
	{
	tisSysLog(0, 0, "nameSrv can't open RESLIST.");
	exit(1);
	}
    if ((reslistResName = d4field(reslist, "RESRCNAME")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find field RESRCNAME in RESLIST.");
	exit(1);
	}
    if ((reslistIOC1 = d4field(reslist, "PRIMARYIOC")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find field PRIMARYIOC in RESLIST.");
	exit(1);
	}
    if ((reslistIOC2 = d4field(reslist, "SECNDRYIOC")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find field SECNDRYIOC in RESLIST.");
	exit(1);
	}
    if ((reslistTag = d4tag(reslist, "RESRCNAME")) == 0)
	{
	tisSysLog(0, 0, "nameSrv can't find tag RESRCNAME in RESLIST.");
	exit(1);
	}
    reslistDPA = d4field(reslist, "DPA");
    d4tag_select(reslist, reslistTag);
    d4unlock_all(reslist);

    /*
    We will cache all the records from IOCLIST and RESLIST. This will keep
    our runtime disk I/O to a minimum.
    */

    d4top(ioclist);
    do
	{
	if (d4eof(ioclist))
	    break;
	d4unlock_all(ioclist);
	if (d4deleted(ioclist))
	    continue;
	if (numIocs)
	    {
	    currIocNode -> nextIoc = (struct iocCache *) calloc(1, sizeof(struct iocCache));
	    currIocNode = currIocNode -> nextIoc;
	    }
	else
	    {
	    iocCache = (struct iocCache *) calloc(1, sizeof(struct iocCache));
	    currIocNode = iocCache;
	    }
	numIocs++;
	trimCpy(currIocNode -> iocName, f4str(ioclistIOCName));
	if ((hostent = gethostbyname(currIocNode -> iocName)) == 0)
	    {
	    tisSysLog(0, 0, "nameSrv can't find IOC %s in the network database.", currIocNode -> iocName);
	    exit(1);
	    }
	currIocNode -> inetAddr.s_addr = ((struct in_addr *)hostent -> h_addr_list[0])->s_addr;
	} while(d4skip(ioclist, (long) 1) != r4eof);
    d4unlock_all(ioclist);

    d4top(reslist);
    do
	{
	if (d4eof(reslist))
	    break;
	d4unlock_all(reslist);
	if (d4deleted(reslist))
	    continue;
	if (numRes)
	    {
	    currResNode -> nextRes = (struct resCache *) calloc(1, sizeof(struct resCache));
	    currResNode = currResNode -> nextRes;
	    }
	else
	    {
	    resCache = (struct resCache *) calloc(1, sizeof(struct resCache));
	    currResNode = resCache;
	    }
	numRes++;
	trimCpy(currResNode -> resName, f4str(reslistResName));
	trimCpy(tempName, f4str(reslistIOC1));
	if (strlen(tempName))
	    {
	    currResNode -> inetAddr1 = findIOC(tempName);
	    if (currResNode -> inetAddr1.s_addr == 0)
		{
		tisSysLog(0, 0, "nameSrv: Resource %s references IOC %s which can't be found.\n",
			currResNode -> resName, tempName);
		exit(1);
		}
	    }
	trimCpy(tempName, f4str(reslistIOC2));
	if (strlen(tempName))
	    {
	    currResNode -> inetAddr2 = findIOC(tempName);
	    if (currResNode -> inetAddr2.s_addr == 0)
		{
		tisSysLog(0, 0, "nameSrv: Resource %s references IOC %s which can't be found.\n",
			currResNode -> resName, tempName);
		exit(1);
		}

	    /*
	    If there is only one IOC, make it inetAddr1 even if the RESLIST had it
	    listed as inetAddr2.
	    */

	    if (currResNode -> inetAddr1.s_addr == 0)
		{
		currResNode -> inetAddr1 = currResNode -> inetAddr2;
		currResNode -> inetAddr2.s_addr = 0;
		}
	    }

	if (reslistDPA)
	    {
	    if (f4true(reslistDPA))
		{
		if (numDpa)
		    {
		    currDpaNode -> nextRes = (struct resCache *) malloc(sizeof(struct resCache));
		    currDpaNode = currDpaNode -> nextRes;
		    }
		else
		    {
		    dpaCache = (struct resCache *) malloc(sizeof(struct resCache));
		    currDpaNode = dpaCache;
		    }
		numDpa++;
		*currDpaNode = *currResNode;
	        currDpaNode -> inetAddr1.s_addr = ++dpaNode;
	        currDpaNode -> inetAddr2.s_addr = ++dpaNode;
		currDpaNode -> nextRes = 0;
		}
	    }
	} while (d4skip(reslist, (long) 1) != r4eof);
    d4unlock_all(reslist);

    /*
    Cache up the array of point type names.
    */

#ifdef WIN32
    if ((sdrFile = fopen("\\tis4000db\\default.dctsdr", "r")) == NULL)
	{
	tisSysLog(0, 0, "nameSrv can't open \\tis4000db\\default.dctsdr");
	exit(1);
	}
#else
    if ((sdrFile = fopen("/tis4000db/default.dctsdr", "r")) == NULL)
	{
	tisSysLog(0, 0, "nameSrv can't open /tis4000db/default.dctsdr");
	exit(1);
	}
#endif WIN32

    fgets(tempName, sizeof(tempName), sdrFile);
    trimNewLine(tempName);
    if (strcmp(tempName, "dbRecType"))
	{
	tisSysLog(0, 0, "nameSrv didn't find dbRecType at head of /tis4000db/default.dctsdr");
	exit(1);
	}

    numRecs = getInt(sdrFile);
    dbRecType = (struct recType *) malloc(sizeof(struct recType));
    dbRecType -> number = numRecs;
    dbRecType -> papName = (char **) calloc(numRecs, sizeof(char *));
    for (i = 0; i < dbRecType -> number; i++)
	{
	fgets(tempName, sizeof(tempName), sdrFile);
	if ((numRecs = strlen(tempName)) > 1)
	    {
	    dbRecType -> papName[i] = malloc(numRecs);
	    trimNewLine(tempName);
	    strcpy(dbRecType -> papName[i], tempName);
	    }
	}
    fclose(sdrFile);

/******
	Now wait for socket I/O to make things happen.
******/

    for (;;)
	{

	/*
	First, wait for something to happen. We'll wait forever.
	*/

	fdList = fdMask;
	select(maxFd + 1, &fdList, NULL, NULL, NULL);

	/*
	Then, process all buffer contents until all sockets have been
	drained.
	*/

	fdList = fdMask;
	while (status = select(maxFd + 1, &fdList, NULL, NULL, &pollVal))
	    {

	    /*
	    Find out which fd caused us to execute.
	    */
	    if (status < 0)
	    {
#ifdef WIN32
		tisSysLog(1, 0, "nameSrv: select error %d", WSAGetLastError ());
#else
		tisSysLog(1, 0, "nameSrv: select error %m");
#endif WIN32
		continue;
	    }

	    /*
	    See if this is the UDP port. Just echo the message back.
	    */

	    if (FD_ISSET(udpFd, &fdList))
		{
		addrSize = sizeof(portAddr);
		recvfrom(udpFd, (char *) &msgType, sizeof(msgType), 0,
			(struct sockaddr *) &portAddr, &addrSize);
		sendto(udpFd, (char *) &msgType, sizeof(msgType), 0,
			(struct sockaddr *) &portAddr, addrSize);
		continue;
		}

	    /*
	    See if this is a new user trying to connect
	    */

	    if (FD_ISSET(tcpFd, &fdList))
		{
		addrSize = sizeof(portAddr);
		newFd = accept(tcpFd, (struct sockaddr *) &portAddr, &addrSize);
		for (fdIndex = 0; fdIndex < clientMax; fdIndex++)
		{
		    if (clientSockets [fdIndex] == INVALID_SOCKET)
			break;
		}
		if (fdIndex < MAX_NAME_CLIENTS)
		{
		    if (fdIndex == clientMax)
			clientMax++;
		    clientSockets [fdIndex] = newFd;
		    setsockopt(newFd, IPPROTO_TCP, TCP_NODELAY, (char *)&optVal, sizeof(optVal));
#ifndef WIN32
		    if (newFd > maxFd)
			maxFd = newFd;
#endif WIn32
		    FD_SET(newFd, &fdMask);
#ifndef SOLARIS
#ifdef WIN32
		    ioctlsocket(newFd, FIONBIO, &optVal);
#else
		    ioctl(newFd, FIONBIO, &optVal);
#endif WIN32
#else
		    sockGrow(newFd);
		    fcntl(newFd, F_SETFL, O_NONBLOCK);
#endif
		    tisSysLog(1, 0, "nameSrv: application added on fd %d", newFd);
		}
		else
		{
		    tisSysLog(1, 0, "nameSrv: no more room for new client application");
#ifdef	WIN32
		    closesocket (newFd);
#else
		    close (newFd);
#endif	WIN32
		}

		}
	    else
		{

		/*
		Scan open sockets. Get the message type so we know
		what to do.
		*/
		for (fdIndex = 0; fdIndex < clientMax; fdIndex++)
		{
		    if (clientSockets [fdIndex] != INVALID_SOCKET
		      && FD_ISSET (clientSockets [fdIndex], &fdList))
		    {
			if (recv(clientSockets [fdIndex], (char *) &msgType, sizeof(msgType), 0) <= 0)
			    {
#ifdef	WIN32
			    closesocket (clientSockets [fdIndex]);
#else
			    close (clientSockets [fdIndex]);
#endif	WIN32
			    FD_CLR(clientSockets [fdIndex], &fdMask);
			    clientSockets [fdIndex] = INVALID_SOCKET;
			    continue;
			    }
			msgType = ntohs(msgType);
			if (msgType == NAME_SRCH)
			    {

			    /*
			    Name Search.
			    */

			    char		*colon;
			    NS_SRCH		request;
			    NS_SRCH_REPLY	reply;

			    if (!recvClient(clientSockets [fdIndex], (char *) &request, sizeof(request)))
				continue;
			    reply.nsHandle = request.nsHandle;	/* This is opaque data! Don't flip it! */
			    reply.nsNumNodes = 0;
			    reply.nsResNameLen = 0;

			    /*
			    Get the name he wants.
			    */

			    if (!recvClient(clientSockets [fdIndex], tempName, (int) ntohs(request.nsNameLen)))
				continue;
			    tisSysLog(3, 0, "nameSrv: Search for %s", tempName);
			    msgType = htons(NAME_SRCH_REPLY);

			    /*
			    PVMAP.DBF doesn't want to know about the dot or any field name. If there is a dot,
			    stamp on it. dBase also wants blank padding, so also provide that.
			    */

			    for (length = 0; tempName[length]; length++)
				if (tempName[length] == '.')
				    {
				    tempName[length] = 0;
				    break;
				    }
			    strcat(tempName, "                               ");
			    status = d4seek(pvmap, tempName);
			    d4unlock_all(pvmap);
			    if (status == 0)
				if (d4deleted(pvmap))
				    status = 1;
			    if (status)
				{

				/*
				Name not found. Check for possible DPA.
				*/

				if ((colon = strchr(tempName, ':')))
				    {
				    *colon = 0;
				    for (currDpaNode = dpaCache; currDpaNode; currDpaNode =
					    currDpaNode -> nextRes)
					if (!strcmp(tempName, currDpaNode -> resName))
					    break;
				    if (currDpaNode)
					{
					if (currDpaNode -> inetAddr1.s_addr)
					    reply.nsNumNodes++;
					if (currDpaNode -> inetAddr2.s_addr)
					    reply.nsNumNodes++;
					reply.nsNumNodes = htons(reply.nsNumNodes);
					length = strlen(currDpaNode -> resName) + 1;
					reply.nsResNameLen = htons(length);
					if (!sendClient(clientSockets [fdIndex], (char *) &msgType, sizeof(msgType)))
					    continue;
					if (!sendClient(clientSockets [fdIndex], (char *) &reply, sizeof(reply)))
					    continue;
					if (currDpaNode -> inetAddr1.s_addr)
					    if (!sendClient(clientSockets [fdIndex], (char *) &currDpaNode -> inetAddr1, sizeof(currResNode -> inetAddr1)))
						continue;
					if (currDpaNode -> inetAddr2.s_addr)
					    if (!sendClient(clientSockets [fdIndex], (char *) &currDpaNode -> inetAddr2, sizeof(currResNode -> inetAddr1)))
						continue;
					tisSysLog(3, 0, "\tfound as a DPA!");
					sendClient(clientSockets [fdIndex], currDpaNode -> resName, (int) length);
					continue;
					}
				    }

				else if ((time(NULL) - lastReopen) > REOPEN_TIME)
				    {

				    /*
				    If the name wasn't found and it's not a DPA, then we'll close and
				    reopen PVMAP and try again. However, don't do this every time there's
				    an invalid name. The tuning constant REOPEN_TIME keeps us from doing
				    this too often. At SNWS, almost every display has at least one point
				    that has not yet been entered into the database.
				    */

				    d4flush_all(pvmap);
				    d4close(pvmap);
				    lastReopen = time(NULL);
#ifdef WIN32
				    if ((pvmap = d4open(&codeBase, "\\tis4000db\\pvmap.dbf")) == 0)
#else
				    if ((pvmap = d4open(&codeBase, "/tis4000db/pvmap.dbf")) == 0)
#endif WIN32
	    
					{
					tisSysLog(0, 0, "nameSrv can't open PVMAP.");
					exit(1);
					}
    				    if ((pvmapName = d4field(pvmap, "NAME")) == 0)
					{
					tisSysLog(0, 0, "nameSrv can't find field NAME in PVMAP.");
					exit(1);
					}
    				    if ((pvmapResource = d4field(pvmap, "RESRCNAME")) == 0)
					{
					tisSysLog(0, 0, "nameSrv can't find field RESRCNAME in PVMAP.");
					exit(1);
					}
    				    if ((pvmapCreationTime = d4field(pvmap, "TIMESTAMP")) == 0)
					{
					tisSysLog(0, 0, "nameSrv can't find field TIMESTAMP in PVMAP.");
					exit(1);
					}
    				    if ((pvmapChecksum = d4field(pvmap, "CHECKSUM")) == 0)
					{
					tisSysLog(0, 0, "nameSrv can't find field CHECKSUM in PVMAP.");
					exit(1);
					}
    				    if ((pvmapPointType = d4field(pvmap, "TYPE")) == 0)
					{
					tisSysLog(0, 0, "nameSrv can't find field TYPE in PVMAP.");
					exit(1);
					}
    				    if ((pvmapTag = d4tag(pvmap, "NAME")) == 0)
					{
					tisSysLog(0, 0, "nameSrv can't find tag NAME in PVMAP.");
					exit(1);
					}
    				    d4tag_select(pvmap, pvmapTag);
    				    d4unlock_all(pvmap);

				    /*
				    Now that the file has been reopenend, try again to find the name.
				    */

				    status = d4seek(pvmap, tempName);
				    d4unlock_all(pvmap);
				    if (status == 0)
			    		if (d4deleted(pvmap))
					    status = 1;
				    if (status)
					{

					/*
					Still not found. Send a rejection.
					*/

					tisSysLog(3, 0, "\tnot found after re-open!\n");
			    		if (!sendClient(clientSockets [fdIndex], (char *) &msgType, sizeof(msgType)))
					    continue;
			    		sendClient(clientSockets [fdIndex], (char *) &reply, sizeof(reply));
			    		    continue;
					}
				    }

				else
				    {
			    	    /*
			    	    Not found, not found on reopen, and not a DPA. Just send him a rejection.
			    	    */

			    	    tisSysLog(3, 0, "\tnot found in nameSrv!");
			    	    if (!sendClient(clientSockets [fdIndex], (char *) &msgType, sizeof(msgType)))
					continue;
			    	    sendClient(clientSockets [fdIndex], (char *) &reply, sizeof(reply));
			    	    continue;
				    }
				}

			    /*
			    Found the name. Now find the resource name
			    */

			    trimCpy(tempName, f4str(pvmapResource));
			    tisSysLog(3, 0, "\tFound as ordinary point");
			    if (!(currResNode = findRes(tempName)))
				{
				if (!sendClient(clientSockets [fdIndex], (char *) &msgType, sizeof(msgType)))
				    continue;
				sendClient(clientSockets [fdIndex], (char *) &reply, sizeof(reply));
				continue;
				}

			    /*
			    Found the resource. Now format the response and ship it out.
			    */

			    reply.nsCreationTime = htonl(f4int(pvmapCreationTime));
			    reply.nsChecksum  = htons((unsigned short)strtol(f4str(pvmapChecksum),NULL,16));
			    trimCpy(tempName, f4str(pvmapPointType));
			    for (i = 0; i < dbRecType -> number; i++)
				{
				if (dbRecType -> papName[i])
				    if (!strcmp(dbRecType -> papName[i], tempName))
					break;
				}
			    reply.nsPointType = htons((short) i);
			    if (currResNode -> inetAddr1.s_addr)
				reply.nsNumNodes++;
			    if (currResNode -> inetAddr2.s_addr)
				reply.nsNumNodes++;
			    reply.nsNumNodes = htons(reply.nsNumNodes);
			    length = strlen(currResNode -> resName) + 1;
			    reply.nsResNameLen = htons(length);
			    if (!sendClient(clientSockets [fdIndex], (char *) &msgType, sizeof(msgType)))
				continue;
			    if (!sendClient(clientSockets [fdIndex], (char *) &reply, sizeof(reply)))
				continue;
			    if (currResNode -> inetAddr1.s_addr)
				if (!sendClient(clientSockets [fdIndex], (char *) &currResNode -> inetAddr1, sizeof(currResNode -> inetAddr1)))
				    continue;
			    if (currResNode -> inetAddr2.s_addr)
				if (!sendClient(clientSockets [fdIndex], (char *) &currResNode -> inetAddr2, sizeof(currResNode -> inetAddr2)))
				    continue;
			    sendClient(clientSockets [fdIndex], currResNode -> resName, (int) length);
			    }

			else if (msgType == NAME_RSRQ)
			    {

			    /*
			    Resource List Request.
			    */

			    NS_RSRQ_REPLY	rsrqReply;

			    msgType = htons(NAME_RSRQ_REPLY);
			    tisSysLog(2, 0, "nameSrv: Resource List Request");
			    if (sendClient(clientSockets [fdIndex], (char *) &msgType, sizeof(msgType)))
				{
				rsrqReply.nsNumNames = htons((short)numRes);
				rsrqReply.nsMaxNameLen = htons(sizeof(currResNode -> resName));
				if (sendClient(clientSockets [fdIndex], (char *) &rsrqReply, sizeof(rsrqReply)))
				    {
				    for (currResNode = resCache; currResNode; currResNode = currResNode -> nextRes)
					{
					if (!sendClient(clientSockets [fdIndex], currResNode -> resName,
						sizeof(currResNode -> resName)))
					    break;
					if (!sendClient(clientSockets [fdIndex], (char *) &currResNode -> inetAddr1, sizeof(currResNode -> inetAddr1)))
					    break;
					if (!sendClient(clientSockets [fdIndex], (char *) &currResNode -> inetAddr2, sizeof(currResNode -> inetAddr2)))
					    break;
					}
				    }
				}
			    }
			else if (msgType == NAME_IOCL)
			    {
			    NS_IOCL_REPLY	ioclReply;

			    /*
			    IOC List Request.
			    */

			    msgType = htons(NAME_IOCL_REPLY);
			    tisSysLog(2, 0, "nameSrv: IOC list request");
			    if (sendClient(clientSockets [fdIndex], (char *) &msgType, sizeof(msgType)))
				{
				ioclReply.nsNumNodes = htons((short)numIocs);
				if (sendClient(clientSockets [fdIndex], (char *) &ioclReply, sizeof(ioclReply)))
				    {
				    for (currIocNode = iocCache; currIocNode; currIocNode = currIocNode -> nextIoc)
					{
					if (!sendClient(clientSockets [fdIndex], (char *) &currIocNode -> inetAddr, sizeof(currIocNode -> inetAddr)))
					    break;
					}
				    }
				}
			    }

			else if (msgType == NAME_DPAL)
			    {
			    NS_DPAL_REPLY	dpalReply;

			    msgType = htons(NAME_DPAL_REPLY);
			    tisSysLog(2, 0, "nameSrv: DPA list request");
			    if (sendClient(clientSockets [fdIndex], (char *) &msgType, sizeof(msgType)))
				{
				dpalReply.nsNumDPAs = htons((short)numDpa);
				dpalReply.nsMaxNameLen = htons(sizeof(currDpaNode -> resName));
				if (sendClient(clientSockets [fdIndex], (char *) &dpalReply, sizeof(dpalReply)))
				    {
				    for (currDpaNode = dpaCache; currDpaNode; currDpaNode =
					    currDpaNode -> nextRes)
					{
					if (!sendClient(clientSockets [fdIndex], currDpaNode -> resName, 
						sizeof(currDpaNode -> resName)))
					    break;
					if (!sendClient(clientSockets [fdIndex], (char *) &currDpaNode -> inetAddr1, sizeof(currDpaNode -> inetAddr1)))
					    break;
					if (!sendClient(clientSockets [fdIndex], (char *) &currDpaNode -> inetAddr2, sizeof(currDpaNode -> inetAddr2)))
					    break;
					}
				    }
				}
			    }
			}
		    }	
		}

	    /*
	    Recopy the fd mask just in case we changed it during this iteration.
	    */

	    fdList = fdMask;
	    }

	/*
	Flush the buffer if anyone is waiting.
	*/

	nameFlush(lastFd, sockBuff, sockSize);
	}
	return 0;
}


/****************************************************************/
/*								*/
/*	COPY ONE STRING TO ANOTHER, STOPPING AT FIRST BLANK	*/
/*								*/
/****************************************************************/

void trimCpy(destination, source)
char	*destination;
char	*source;
{
    while (*destination = *source)
	{
	if (*destination == ' ')
	    {
	    *destination = 0;
	    return;
	    }
	destination++;
	source++;
	}
    return;
}


/****************************************************************/
/*								*/
/*		FIND INTERNET ADDRESS FOR AN IOC		*/
/*								*/
/*	A zero means that we couldn't find it.			*/
/*								*/
/****************************************************************/

struct in_addr findIOC(iocName)
char	*iocName;
{
    struct iocCache	*iocEntry;
    char		tempName[sizeof(iocEntry -> iocName) * 2];
    int			status;
    struct hostent	*hostent;
    struct in_addr	failReturn = {0};

    /*
    We start looking in the table we've cached.
    */

    iocEntry = iocCache;
    for (iocEntry = iocCache; iocEntry; iocEntry = iocEntry -> nextIoc)
	{
	if (!strcmp(iocName, iocEntry -> iocName))
	    return (iocEntry -> inetAddr);
	if (!(iocEntry -> nextIoc))
	    break;
	}

    /*
    Not in the cache. See if we can find it on disk. If we do,
    cache that entry.
    */

    strcpy(tempName, iocName);
    strcat(tempName, "               ");
    status = d4seek(ioclist, tempName);
    d4unlock_all(ioclist);
    if (status == 0)
	if (d4deleted(ioclist))
	    status = 1;
    if (status)
	return failReturn;
    if ((hostent = gethostbyname(iocName)) == 0)
	{
	tisSysLog(0, 600, "nameSrv can't find IOC %s in the network database.\n", iocName);
	return failReturn;
	}
    numIocs++;
    if (iocEntry)
	{
	iocEntry -> nextIoc = (struct iocCache *) calloc(1, sizeof(struct iocCache));
	iocEntry = iocEntry -> nextIoc;
	}
    else
	{
	iocCache = (struct iocCache *) calloc(1, sizeof(struct iocCache));
	iocEntry = iocCache;
	}
    strcpy(iocEntry -> iocName, iocName);
    iocEntry -> inetAddr.s_addr = ((struct in_addr *)hostent -> h_addr_list[0])->s_addr;
    return (iocEntry -> inetAddr);
}


/****************************************************************/
/*								*/
/*		FIND A RESOURCE BY NAME				*/
/*								*/
/*	Return a null pointer if we can't find it.		*/
/*								*/
/****************************************************************/

struct resCache *findRes(name)
char	*name;
{
    struct resCache	*resEntry;
    char		tempName[sizeof(resEntry -> resName) * 2];
    int			status;

    /*
    Start by looking in the names we've already cached.
    */

    resEntry = resCache;
    for (resEntry = resCache; resEntry; resEntry = resEntry -> nextRes)
	{
	if (!strcmp(name, resEntry -> resName))
	    return (resEntry);
	if (!(resEntry -> nextRes))
	    break;
	}

    /*
    Not in the cache. Try to find it on the disk.
    */

    strcpy(tempName, name);
    strcat(tempName, "               ");
    status = d4seek(reslist, tempName);
    d4unlock_all(reslist);
    if (status == 0)
	if (d4deleted(reslist))
	    status = 1;
    if (status)
	return(0);

    /*
    Put this new one into the cache.
    */

    numRes++;
    if (resEntry)
	{
	resEntry -> nextRes = (struct resCache *) calloc(1, sizeof(struct resCache));
	resEntry = resEntry -> nextRes;
	}
    else
	{
	resCache = (struct resCache *) calloc(1, sizeof(struct resCache));
	resEntry = resCache;
	}
    strcpy(resEntry -> resName, name);
    trimCpy(tempName, f4str(reslistIOC1));
    if (strlen(tempName))
	resEntry -> inetAddr1 = findIOC(tempName);
    trimCpy(tempName, f4str(reslistIOC2));
    if (strlen(tempName))
	resEntry -> inetAddr2 = findIOC(tempName);
    return (resEntry);
}


/****************************************************************/
/*								*/
/*		SEND A MESSAGE TO A CLIENT			*/
/*								*/
/*	This routine modified to buffer up the messages to 	*/
/*	make bigger blocks back to the user. nameFlush()	*/
/*	actually does the writing now.				*/
/*								*/
/*	Return a zero on failure.				*/
/*								*/
/****************************************************************/

sendClient(fd, buff, nbytes)
SOCKET	fd;
char	*buff;
int	nbytes;
{
    int	bytesAvail;

    if (nbytes == 0)
	return (0);
    
    /*
    If we changed FDs, flush the previous one.
    */

    if (fd != lastFd)
	{
	nameFlush(lastFd, sockBuff, sockSize);
	lastFd = fd;
	}

    /*
    Move in all the data that will fit. Flush the buffer if there's
    too much to move in all at once.
    */

    while (nbytes)
	{
	bytesAvail = sizeof(sockBuff) - sockSize;
	if (bytesAvail < nbytes)
	    {
	    memcpy(sockBuff + sockSize, buff, bytesAvail);
	    sockSize += bytesAvail;
	    if (nameFlush(lastFd, sockBuff, sockSize) == 0)
		return (0);
	    nbytes -= bytesAvail;
	    buff += bytesAvail;
	    }
	else
	    {
	    memcpy(sockBuff + sockSize, buff, nbytes);
	    sockSize += nbytes;
	    break;
	    }
	}
    
    return (1);
}

/****************************************************************/
/*								*/
/*		FLUSH THE SOCKET BUFFER TO THE CLIENT		*/
/*								*/
/****************************************************************/

nameFlush(fd, buff, nbytes)
SOCKET	fd;
char	*buff;
int	nbytes;
{
    int		bytesSent;
    int		txBuffSize;
    int		intSize;

    if (fd <= 0)
	return (0);
    if (nbytes == 0)
	return (0);

    bytesSent = send(fd, buff, nbytes, 0);
    if (bytesSent == -1)
	{
#ifdef WIN32
	errno = WSAGetLastError ();
	if (errno == WSAEWOULDBLOCK)
#else
	if (errno == EWOULDBLOCK)
#endif WIN32
	    {
	    intSize = sizeof(txBuffSize);
	    getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&txBuffSize, &intSize);
	    txBuffSize += nbytes;
	    if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&txBuffSize, sizeof(txBuffSize)) == -1)
		{
		tisSysLog(1, 0, "nameSrv dropping client on fd %d because of logjam", fd);
		lastFd = INVALID_SOCKET;
		sockSize = 0;
		closeClient (fd);
		return(0);
		}
	    return (nameFlush(fd, buff, nbytes));
	    }
	else
	    {
	    tisSysLog(1, 0, "nameSrv I/O error writing to client on fd %d: %m", fd);
	    lastFd = INVALID_SOCKET;
	    sockSize = 0;
	    closeClient (fd);
	    return (0);
	    }
	}
    else if (bytesSent != nbytes)
	{
	nbytes -= bytesSent;
	return (nameFlush(fd, buff + bytesSent, nbytes));
	}
    sockSize = 0;
    return (1);
}


/****************************************************************/
/*								*/
/*		RECEIVE A MESSAGE FROM A CLIENT			*/
/*								*/
/*	Return a zero on failure.				*/
/*								*/
/****************************************************************/

recvClient(fd, buff, nbytes)
SOCKET	fd;
char	*buff;
int	nbytes;
{
    fd_set	fdList;
    int	numRead;

    FD_ZERO(&fdList);
    FD_SET(fd, &fdList);
    while (nbytes)
	{
	if (select(fd + 1, &fdList, NULL, NULL, &oneSecond) <= 0)
	    {
	    tisSysLog(1, 0, "nameSrv dropping client on fd %d because of timeout", fd);
	    closeClient (fd);
	    return (0);
	    }
	if ((numRead = recv(fd, buff, nbytes, 0)) <= 0)
	    {
	    tisSysLog(1, 0, "nameSrv dropping client on fd %d because of I/O error: %m", fd);
	    closeClient (fd);
	    return (0);
	    }
	nbytes -= numRead;
	buff += numRead;
	}
    return (1);
}

void closeClient (SOCKET sock)
{
    int fdIndex;
    
    for (fdIndex = 0; fdIndex < clientMax; fdIndex++)
    {
	if (clientSockets [fdIndex] == sock)
	    break;
    }
    if (fdIndex < MAX_NAME_CLIENTS)
    {
#ifdef	WIN32
	closesocket (clientSockets [fdIndex]);
#else
	close (clientSockets [fdIndex]);
#endif	WIN32
	FD_CLR(clientSockets [fdIndex], &fdMask);
	clientSockets [fdIndex] = INVALID_SOCKET;
    }
}

/****************************************************************/
/*								*/
/*	GET AN INTEGER FROM THE SDR FILE INPUT STREAM		*/
/*								*/
/****************************************************************/

int getInt(fd)
FILE	*fd;
{
    char	buffer[132];

    fgets(buffer, sizeof(buffer), fd);
    return (atoi(buffer));
}


/****************************************************************/
/*								*/
/*	TRIM THE TRAILING NEW LINE FROM A CHAR STRING		*/
/*								*/
/****************************************************************/

void trimNewLine(string)
char	*string;
{
    char	*newLine;

    if (newLine = strchr(string, '\n'))
	*newLine = 0;
}
