/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	CONVERT CA AGENT DATA TYPES TO DBR DATA TYPES		*/
/*								*/
/*	The agent data types are in standard network byte	*/
/*	order. The dbr types are always host byte order.	*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifdef vxWorks
#include <vxWorks.h>
#endif vxWorks

#ifndef ultrix
#if !(defined(AXP) || defined(vxWorks))
#include <bcopy.h>
#endif AXP
#endif ultrix

#if defined(AXP) || defined(vxWorks)
#include <string.h>
#endif

#include "dbDefs.h"
#include "db_access.h"		/* Traditional EPICS data types */
#include "tsDefs.h"
#include "dataTypes.h"		/* New agent CA data types */
#include "tisByteOrder.h"	/* Macros for floating point conversion */
#ifdef WIN32
#include <sys/types.h>
#include <winsock.h>
#else
#include <types.h>
#include <in.h>			/* Contains htons(), htonl(), etc. */
#endif WIN32
#include <stdio.h>

#ifdef AXP
#include <string.h>
#endif

/****************************************************************/
/*								*/
/*		FUNDAMENTAL DATA SIZES				*/
/*								*/
/****************************************************************/

static short fundamentalSize[] =
    {
    0,			/* AGNT_STRING */
    sizeof(short),	/* AGNT_SHORT and AGNT_SHORT */
    sizeof(float),	/* AGNT_FLOAT */
    sizeof(short),	/* AGNT_ENUM */
    sizeof(char),	/* AGNT_CHAR */
    sizeof(int),	/* AGNT_LONG */
    sizeof(double),	/* AGNT_DOUBLE */
    };

/****************************************************************/
/*								*/
/*		CONVERT THE DATA BUFFER				*/
/*								*/
/*	This logic is very similar to tranformType.c. That	*/
/*	routine transforms data between two agent types. This	*/
/*	version translates between agent types and the old	*/
/*	EPICS dbr types, performing endian translation along	*/
/*	the way.						*/
/*								*/
/****************************************************************/

int caConvType(agentBuff, agentType, dbrBuff, dbrType, count)
char	*agentBuff;
short	agentType;
char	*dbrBuff;
short	dbrType;
int	count;
{
    char	*pEgu = NULL;
    char	*pHopr = NULL;
    char	*pLopr = NULL;
    char	*pStrings;
    short	numStrings;
    short	*pNumStrings;
    char	*pTime = NULL;
    char	*pDesc = NULL;
    char	*pAck = NULL;
    char	*pAenb = NULL;
    char	*pSevr = NULL;
    char	*pStat = NULL;
    char	*pPrec = NULL;
    char	*pDrvh = NULL;
    char	*pDrvl = NULL;
    short	dataSize;

    /*
    Brain damage alert. The following structure is defined
    here to be the same size as a TS_STAMP on the IOC. We
    had to do this because of the DEC Alpha.
    */

    struct
	{
	int	secPastEpoch;
	int	nsec;
	} dbrTs;
    double	dbrTempDouble;
    long	dbrTempLong;
    float	dbrTempFloat;
    short	dbrTempShort;

    /*
    Start off by locating all of the fields in the agent buffer.
    */

    dataSize = fundamentalSize[agentType & 0xff];
    if (agentType & AGNT_STAT)
	{
	pStat = agentBuff;
	agentBuff += sizeof(short);
	pSevr = agentBuff;
	agentBuff += sizeof(short);
	if ((agentType & 0xff) == AGNT_STRING)
	    {
	    pAenb = agentBuff;
	    agentBuff += 4;
	    pAck = agentBuff;
	    agentBuff += 4;
	    pDesc = agentBuff;
	    agentBuff += 28;
	    pEgu = agentBuff;
	    agentBuff += 8;
	    }
	}

    if (agentType & AGNT_TIME)
	{
	pTime = agentBuff;
	agentBuff += sizeof(dbrTs);
	}

    if (agentType & AGNT_GRAPH)
	{
	if ((agentType & 0xff) == AGNT_ENUM)
	    {
	    pNumStrings = (short *) agentBuff;
	    agentBuff += sizeof(short);
	    bcopy((char *)pNumStrings, (char *) &numStrings, sizeof(short));
	    numStrings = ntohs(numStrings);
	    pStrings = agentBuff;
	    agentType &= ~AGNT_CNTRL;
	    }
	else
	    {
	    pHopr = agentBuff;
	    agentBuff += dataSize;
	    pLopr = agentBuff;
	    agentBuff += dataSize;

	    if (((agentType & 0xff) == AGNT_FLOAT) || ((agentType & 0xff) == AGNT_DOUBLE))
		{
		pPrec = agentBuff;
		agentBuff += sizeof(short);
		}
	    }
	}

    if (agentType & AGNT_CNTRL)
	{
	if ((agentType & 0xff) == AGNT_ENUM)
	    {
	    pNumStrings = (short *) agentBuff;
	    agentBuff += sizeof(short);
	    bcopy((char *)pNumStrings, (char *) &numStrings, sizeof(short));
	    numStrings = ntohs(numStrings);
	    pStrings = agentBuff;
	    }

	else
	    {
	    pDrvh = agentBuff;
	    agentBuff += dataSize;
	    pDrvl = agentBuff;
	    agentBuff += dataSize;
	    }
	}

    /*
    Now, create the old type (i.e., dbr) structure. We don't convert
    from one fundamental type to another. That's done in the IOC.
    Another point of interest is that we will copy all data types
    from the input stream to local variables and then convert them
    to the local endian. The reason for this extra copy is to make
    sure that we don't get burned by alignment of data items in
    the input stream.
    */

    if (count <= 0)
	count = 1;

    switch (dbrType)
	{
	case OLD_DBR_STRING:
	    {
	    strcpy(dbrBuff, agentBuff);
	    break;
	    }

	case OLD_DBR_INT:
	/* case OLD_DBR_SHORT:	Same value as OLD_DBR_INT */
	case OLD_DBR_ENUM:
	    {
	    short	*dbrShort = (short *) dbrBuff;

	    while (count--)
		{
	    	bcopy(agentBuff, (char *) &dbrTempShort, sizeof(short));
		agentBuff += sizeof(short);
	    	*dbrShort++ = ntohs(dbrTempShort);
		}
	    break;
	    }

	case OLD_DBR_FLOAT:
	    {
	    float	*dbrFloat = (float *) dbrBuff;

	    while (count--)
		{
	    	bcopy(agentBuff, (char *) &dbrTempFloat, sizeof(float));
		agentBuff += sizeof(float);
	    	ntohf(dbrTempFloat, *dbrFloat);
		dbrFloat++;		/* ++ is dangerous in macros! */
		}
	    break;
	    }

	case OLD_DBR_CHAR:
	    while (count--)
	    	*dbrBuff++ = *agentBuff++;
	    break;

	case OLD_DBR_LONG:
	    {
	    long	*dbrLong = (long *) dbrBuff;

	    while (count--)
		{
	    	bcopy(agentBuff, (char *) &dbrTempLong, sizeof(int));
		agentBuff += sizeof(long);
	    	*dbrLong++ = ntohl(dbrTempLong);
		}
	    break;
	    }

	case OLD_DBR_DOUBLE:
	    {
	    double	*dbrDouble = (double *) dbrBuff;

	    while (count--)
		{
	    	bcopy(agentBuff, (char *) &dbrTempDouble, sizeof(double));
		agentBuff += sizeof(double);
	    	ntohd(dbrTempDouble, *dbrDouble);
		dbrDouble++;		/* ++ is dangerous in macros! */
		}
	    break;
	    }

	case OLD_DBR_STS_STRING:
	case OLD_DBR_GR_STRING:		/* Not actually implemented */
	case OLD_DBR_CTRL_STRING:		/* Not actually implemented */
	    {
	    struct dbr_sts_string	*dbrStsStr = (struct dbr_sts_string *) dbrBuff;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrStsStr -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrStsStr -> severity = ntohs(dbrTempShort);
	    strcpy(dbrStsStr -> value, agentBuff);
	    strncpy(dbrStsStr -> aenb, pAenb, sizeof(dbrStsStr -> aenb));
	    strncpy(dbrStsStr -> ack, pAck, sizeof(dbrStsStr -> ack));
	    strcpy(dbrStsStr -> desc, pDesc);
	    strcpy(dbrStsStr -> egu, pEgu);
	    break;
	    }

	case OLD_DBR_STS_INT:
	/* case OLD_DBR_STS_SHORT:		Same value as OLD_DBR_STS_INT */
	case OLD_DBR_STS_ENUM:
	    {
	    struct dbr_sts_int	*dbrStsInt = (struct dbr_sts_int *) dbrBuff;
	    short		*dbstiv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrStsInt -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrStsInt -> severity = ntohs(dbrTempShort);
	    dbstiv = &(dbrStsInt -> value);
	    while (count--)
		{
	        bcopy(agentBuff, (char *) &dbrTempShort, sizeof(dbrTempShort));
		*dbstiv++ = ntohs(dbrTempShort);
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_STS_FLOAT:
	    {
	    struct dbr_sts_float	*dbrStsFlt = (struct dbr_sts_float *) dbrBuff;
	    float			*dbstfv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrStsFlt -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrStsFlt -> severity = ntohs(dbrTempShort);
	    dbstfv = &(dbrStsFlt -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempFloat, sizeof(dbrTempFloat));
		ntohf(dbrTempFloat, *dbstfv);
		dbstfv++;
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_STS_CHAR:
	    {
	    struct dbr_sts_char	*dbrStsChar = (struct dbr_sts_char *) dbrBuff;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrStsChar -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrStsChar -> severity = ntohs(dbrTempShort);
	    bcopy(agentBuff, (char *) &(dbrStsChar -> value), count * sizeof(char));
	    break;
	    }

	case OLD_DBR_STS_LONG:
	    {
	    struct dbr_sts_long	*dbrStsLong = (struct dbr_sts_long *) dbrBuff;
	    long		*dbstlv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrStsLong -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrStsLong -> severity = ntohs(dbrTempShort);
	    dbstlv = &(dbrStsLong -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempLong, sizeof(dbrTempLong));
		*dbstlv++ = ntohl(dbrTempLong);
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_STS_DOUBLE:
	    {
	    struct dbr_sts_double	*dbrStsDoub = (struct dbr_sts_double *) dbrBuff;
	    double			*dbstdv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrStsDoub -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrStsDoub -> severity = ntohs(dbrTempShort);
	    dbstdv = &(dbrStsDoub -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempDouble, sizeof(dbrTempDouble));
		ntohd(dbrTempDouble, *dbstdv);
		dbstdv++;
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_TIME_STRING:
	    {
	    struct dbr_time_string	*dbrTimeStr = (struct dbr_time_string *) dbrBuff;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeStr -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeStr -> severity = ntohs(dbrTempShort);
	    bcopy(pTime, (char *) &dbrTs, sizeof(dbrTs));
	    dbrTimeStr -> stamp.secPastEpoch = ntohl(dbrTs.secPastEpoch);
	    dbrTimeStr -> stamp.nsec = ntohl(dbrTs.nsec);
	    strcpy(dbrTimeStr -> value, agentBuff);
	    break;
	    }

	case OLD_DBR_TIME_INT:
	/* case OLD_DBR_TIME_SHORT:		Same value as OLD_DBR_TIME_INT */
	case OLD_DBR_TIME_ENUM:
	    {
	    struct dbr_time_short	*dbrTimeInt = (struct dbr_time_short *) dbrBuff;
	    short			*dbtmsv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeInt -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeInt -> severity = ntohs(dbrTempShort);
	    bcopy(pTime, (char *) &dbrTs, sizeof(dbrTs));
	    dbrTimeInt -> stamp.secPastEpoch = ntohl(dbrTs.secPastEpoch);
	    dbrTimeInt -> stamp.nsec = ntohl(dbrTs.nsec);
	    dbtmsv = &(dbrTimeInt -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempShort, sizeof(dbrTempShort));
		*dbtmsv++ = ntohs(dbrTempShort);
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_TIME_FLOAT:
	    {
	    struct dbr_time_float	*dbrTimeFlt = (struct dbr_time_float *) dbrBuff;
	    float			*dbtmfv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeFlt -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeFlt -> severity = ntohs(dbrTempShort);
	    bcopy(pTime, (char *) &dbrTs, sizeof(dbrTs));
	    dbrTimeFlt -> stamp.secPastEpoch = ntohl(dbrTs.secPastEpoch);
	    dbrTimeFlt -> stamp.nsec = ntohl(dbrTs.nsec);
	    dbtmfv = &(dbrTimeFlt -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempFloat, sizeof(dbrTempFloat));
		ntohf(dbrTempFloat, *dbtmfv);
		dbtmfv++;
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_TIME_CHAR:
	    {
	    struct dbr_time_char	*dbrTimeChar = (struct dbr_time_char *) dbrBuff;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeChar -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeChar -> severity = ntohs(dbrTempShort);
	    bcopy(pTime, (char *) &dbrTs, sizeof(dbrTs));
	    dbrTimeChar -> stamp.secPastEpoch = ntohl(dbrTs.secPastEpoch);
	    dbrTimeChar -> stamp.nsec = ntohl(dbrTs.nsec);
	    bcopy(agentBuff, (char *) &(dbrTimeChar -> value), count * sizeof(char));
	    break;
	    }

	case OLD_DBR_TIME_LONG:
	    {
	    struct dbr_time_long	*dbrTimeLong = (struct dbr_time_long *) dbrBuff;
	    long			*dbtmlv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeLong -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeLong -> severity = ntohs(dbrTempShort);
	    bcopy(pTime, (char *) &dbrTs, sizeof(dbrTs));
	    dbrTimeLong -> stamp.secPastEpoch = ntohl(dbrTs.secPastEpoch);
	    dbrTimeLong -> stamp.nsec = ntohl(dbrTs.nsec);
	    dbtmlv = &(dbrTimeLong -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempLong, sizeof(dbrTempLong));
		*dbtmlv++ = ntohl(dbrTempLong);
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_TIME_DOUBLE:
	    {
	    struct dbr_time_double	*dbrTimeDoub = (struct dbr_time_double *) dbrBuff;
	    double			*dbtmdv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeDoub -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrTimeDoub -> severity = ntohs(dbrTempShort);
	    bcopy(pTime, (char *) &dbrTs, sizeof(dbrTs));
	    dbrTimeDoub -> stamp.secPastEpoch = ntohl(dbrTs.secPastEpoch);
	    dbrTimeDoub -> stamp.nsec = ntohl(dbrTs.nsec);
	    dbtmdv = &(dbrTimeDoub -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempDouble, sizeof(dbrTempDouble));
		ntohd(dbrTempDouble, *dbtmdv);
		dbtmdv++;
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_GR_INT:
	/* case OLD_DBR_GR_SHORT:		Same value as OLD_DBR_GT_INT */
	    {
	    struct dbr_gr_int	*dbrGrInt = (struct dbr_gr_int *) dbrBuff;
	    short		*dbgriv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrGrInt -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrGrInt -> severity = ntohs(dbrTempShort);
	    dbrGrInt -> units[0] = 0;
	    bcopy(pHopr, (char *) &dbrTempShort, sizeof(short));
	    dbrGrInt -> upper_disp_limit = ntohs(dbrTempShort);
	    bcopy(pLopr, (char *) &dbrTempShort, sizeof(short));
	    dbrGrInt -> lower_disp_limit = ntohs(dbrTempShort);
	    dbrGrInt -> upper_alarm_limit = 0;
	    dbrGrInt -> upper_warning_limit = 0;
	    dbrGrInt -> lower_warning_limit = 0;
	    dbrGrInt -> lower_alarm_limit = 0;
	    dbgriv = &(dbrGrInt -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempShort, sizeof(dbrTempShort));
		*dbgriv++ = ntohs(dbrTempShort);
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_GR_FLOAT:
	    {
	    struct dbr_gr_float	*dbrGrFloat = (struct dbr_gr_float *) dbrBuff;
	    float		*dbgrfv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrGrFloat -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrGrFloat -> severity = ntohs(dbrTempShort);
	    dbrGrFloat -> units[0] = 0;
	    bcopy(pPrec, (char *) &dbrTempShort, sizeof(short));
	    dbrGrFloat -> precision = ntohs(dbrTempShort);
	    bcopy(pHopr, (char *) &dbrTempFloat, sizeof(float));
	    ntohf(dbrTempFloat, (dbrGrFloat -> upper_disp_limit));
	    bcopy(pLopr, (char *) &dbrTempFloat, sizeof(float));
	    ntohf(dbrTempFloat, (dbrGrFloat -> lower_disp_limit));
	    dbrGrFloat -> upper_alarm_limit = (float)0;
	    dbrGrFloat -> upper_warning_limit = (float)0;
	    dbrGrFloat -> lower_warning_limit = (float)0;
	    dbrGrFloat -> lower_alarm_limit = (float)0;
	    dbgrfv = &(dbrGrFloat -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempFloat, sizeof(dbrTempFloat));
		ntohf(dbrTempFloat, *dbgrfv);
		dbgrfv++;
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_GR_ENUM:
	case OLD_DBR_CTRL_ENUM:
	    {
	    struct dbr_gr_enum	*dbrGrEnum = (struct dbr_gr_enum *) dbrBuff;
	    short		*dbgrev;
	    short		i;
	    short		j;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrGrEnum -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrGrEnum -> severity = ntohs(dbrTempShort);
	    dbrGrEnum -> no_str = numStrings;
	    for (i = 0; i < numStrings; i++)
		{
		j = 0;
		while ((dbrGrEnum -> strs[i][j++] = *agentBuff++));
		}
	    dbgrev = &(dbrGrEnum -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempShort, sizeof(dbrTempShort));
		*dbgrev++ = ntohs(dbrTempShort);
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_GR_CHAR:
	    {
	    struct dbr_gr_char	*dbrGrChar = (struct dbr_gr_char *) dbrBuff;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrGrChar -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrGrChar -> severity = ntohs(dbrTempShort);
	    dbrGrChar -> upper_disp_limit = *pHopr;
	    dbrGrChar -> lower_disp_limit = *pLopr;
	    dbrGrChar -> upper_alarm_limit = 0;
	    dbrGrChar -> upper_warning_limit = 0;
	    dbrGrChar -> lower_warning_limit = 0;
	    dbrGrChar -> lower_alarm_limit = 0;
	    bcopy(agentBuff, &(dbrGrChar -> value), count * sizeof(char));
	    break;
	    }

	case OLD_DBR_GR_LONG:
	    {
	    struct dbr_gr_long	*dbrGrLong = (struct dbr_gr_long *) dbrBuff;
	    long		*dbgrlv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrGrLong -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrGrLong -> severity = ntohs(dbrTempShort);
	    dbrGrLong -> units[0] = 0;
	    bcopy(pHopr, (char *) &dbrTempLong, sizeof(long));
	    dbrGrLong -> upper_disp_limit = ntohl(dbrTempLong);
	    bcopy(pLopr, (char *) &dbrTempLong, sizeof(long));
	    dbrGrLong -> lower_disp_limit = ntohl(dbrTempLong);
	    dbrGrLong -> upper_alarm_limit = 0;
	    dbrGrLong -> upper_warning_limit = 0;
	    dbrGrLong -> lower_warning_limit = 0;
	    dbrGrLong -> lower_alarm_limit = 0;
	    dbgrlv = &(dbrGrLong -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempLong, sizeof(dbrTempLong));
		*dbgrlv++ = ntohl(dbrTempLong);
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_GR_DOUBLE:
	    {
	    struct dbr_gr_double	*dbrGrDoub = (struct dbr_gr_double *) dbrBuff;
	    double			*dbgrdv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrGrDoub -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrGrDoub -> severity = ntohs(dbrTempShort);
	    dbrGrDoub -> units[0] = 0;
	    bcopy(pPrec, (char *) &dbrTempShort, sizeof(short));
	    dbrGrDoub -> precision = ntohs(dbrTempShort);
	    bcopy(pHopr, (char *) &dbrTempDouble, sizeof(double));
	    ntohd(dbrTempDouble, (dbrGrDoub -> upper_disp_limit));
	    bcopy(pLopr, (char *) &dbrTempDouble, sizeof(double));
	    ntohd(dbrTempDouble, (dbrGrDoub -> lower_disp_limit));
	    dbrGrDoub -> upper_alarm_limit = 0;
	    dbrGrDoub -> upper_warning_limit = 0;
	    dbrGrDoub -> lower_warning_limit = 0;
	    dbrGrDoub -> lower_alarm_limit = 0;
	    dbgrdv = &(dbrGrDoub -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempDouble, sizeof(dbrTempDouble));
		ntohd(dbrTempDouble, *dbgrdv);
		dbgrdv++;
		agentBuff += dataSize;
		}
	    break;
	    }


	case OLD_DBR_CTRL_INT:
	/* case OLD_DBR_CTRL_SHORT:		Same value as OLD_DBR_CTRL_INT */
	    {
	    struct dbr_ctrl_int	*dbrCtrlInt = (struct dbr_ctrl_int *) dbrBuff;
	    short		*dbctiv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlInt -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlInt -> severity = ntohs(dbrTempShort);
	    dbrCtrlInt -> units[0] = 0;
	    bcopy(pHopr, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlInt -> upper_disp_limit = ntohs(dbrTempShort);
	    bcopy(pLopr, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlInt -> lower_disp_limit = ntohs(dbrTempShort);
	    dbrCtrlInt -> upper_alarm_limit = 0;
	    dbrCtrlInt -> upper_warning_limit = 0;
	    dbrCtrlInt -> lower_warning_limit = 0;
	    dbrCtrlInt -> lower_alarm_limit = 0;
	    bcopy(pDrvh, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlInt -> upper_ctrl_limit = ntohs(dbrTempShort);
	    bcopy(pDrvl, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlInt -> lower_ctrl_limit = ntohs(dbrTempShort);
	    dbctiv = &(dbrCtrlInt -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempShort, sizeof(dbrTempShort));
		*dbctiv++ = ntohs(dbrTempShort);
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_CTRL_FLOAT:
	    {
	    struct dbr_ctrl_float	*dbrCtrlFloat = (struct dbr_ctrl_float *) dbrBuff;
	    float			*dbctfv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlFloat -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlFloat -> severity = ntohs(dbrTempShort);
	    bcopy(pPrec, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlFloat -> precision = ntohs(dbrTempShort);
	    dbrCtrlFloat -> units[0] = 0;
	    bcopy(pHopr, (char *) &dbrTempFloat, sizeof(float));
	    ntohf(dbrTempFloat, (dbrCtrlFloat -> upper_disp_limit));
	    bcopy(pLopr, (char *) &dbrTempFloat, sizeof(float));
	    ntohf(dbrTempFloat, (dbrCtrlFloat -> lower_disp_limit));
	    dbrCtrlFloat -> upper_alarm_limit = (float) 0.0;
	    dbrCtrlFloat -> upper_warning_limit = (float) 0.0;
	    dbrCtrlFloat -> lower_warning_limit = (float) 0.0;
	    dbrCtrlFloat -> lower_alarm_limit = (float) 0.0;
	    bcopy(pDrvh, (char *) &dbrTempFloat, sizeof(float));
	    ntohf(dbrTempFloat, (dbrCtrlFloat -> upper_ctrl_limit));
	    bcopy(pDrvl, (char *) &dbrTempFloat, sizeof(float));
	    ntohf(dbrTempFloat, (dbrCtrlFloat -> lower_ctrl_limit));
	    dbctfv = &(dbrCtrlFloat -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempFloat, sizeof(dbrTempFloat));
		ntohf(dbrTempFloat, *dbctfv);
		dbctfv++;
		agentBuff += dataSize;
		}
	    break;
	    }


	case OLD_DBR_CTRL_CHAR:
	    {
	    struct dbr_ctrl_char	*dbrCtrlChar = (struct dbr_ctrl_char *) dbrBuff;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlChar -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlChar -> severity = ntohs(dbrTempShort);
	    dbrCtrlChar -> upper_disp_limit = *pHopr;
	    dbrCtrlChar -> lower_disp_limit = *pLopr;
	    dbrCtrlChar -> upper_alarm_limit = 0;
	    dbrCtrlChar -> upper_warning_limit = 0;
	    dbrCtrlChar -> lower_warning_limit = 0;
	    dbrCtrlChar -> lower_alarm_limit = 0;
	    dbrCtrlChar -> upper_ctrl_limit = *pDrvh;
	    dbrCtrlChar -> lower_ctrl_limit = *pDrvl;
	    bcopy(agentBuff, &(dbrCtrlChar -> value), count * sizeof(char));
	    break;
	    }

	case OLD_DBR_CTRL_LONG:
	    {
	    struct dbr_ctrl_long	*dbrCtrlLong = (struct dbr_ctrl_long *) dbrBuff;
	    long			*dbctlv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlLong -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlLong -> severity = ntohs(dbrTempShort);
	    dbrCtrlLong -> units[0] = 0;
	    bcopy(pHopr, (char *) &dbrTempLong, sizeof(long));
	    dbrCtrlLong -> upper_disp_limit = htonl(dbrTempLong);
	    bcopy(pLopr, (char *) &dbrTempLong, sizeof(long));
	    dbrCtrlLong -> lower_disp_limit = htonl(dbrTempLong);
	    dbrCtrlLong -> upper_alarm_limit = 0;
	    dbrCtrlLong -> upper_warning_limit = 0;
	    dbrCtrlLong -> lower_warning_limit = 0;
	    dbrCtrlLong -> lower_alarm_limit = 0;
	    dbrCtrlLong -> upper_ctrl_limit = htonl(dbrTempLong);
	    bcopy(pDrvl, (char *) &dbrTempLong, sizeof(long));
	    dbrCtrlLong -> lower_ctrl_limit = htonl(dbrTempLong);
	    dbctlv = &(dbrCtrlLong -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempLong, sizeof(dbrTempLong));
		*dbctlv++ = ntohl(dbrTempLong);
		agentBuff += dataSize;
		}
	    break;
	    }

	case OLD_DBR_CTRL_DOUBLE:
	    {
	    struct dbr_ctrl_double	*dbrCtrlDoub = (struct dbr_ctrl_double *) dbrBuff;
	    double			*dbctdv;

	    bcopy(pStat, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlDoub -> status = ntohs(dbrTempShort);
	    bcopy(pSevr, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlDoub -> severity = ntohs(dbrTempShort);
	    bcopy(pPrec, (char *) &dbrTempShort, sizeof(short));
	    dbrCtrlDoub -> precision = ntohs(dbrTempShort);
	    dbrCtrlDoub -> units[0] = 0;
	    bcopy(pHopr, (char *) &dbrTempDouble, sizeof(double));
	    ntohd(dbrTempDouble, (dbrCtrlDoub -> upper_disp_limit));
	    bcopy(pLopr, (char *) &dbrTempDouble, sizeof(double));
	    ntohd(dbrTempDouble, (dbrCtrlDoub -> lower_disp_limit));
	    dbrCtrlDoub -> upper_alarm_limit = 0;
	    dbrCtrlDoub -> upper_warning_limit = 0;
	    dbrCtrlDoub -> lower_warning_limit = 0;
	    dbrCtrlDoub -> lower_alarm_limit = 0;
	    bcopy(pDrvh, (char *) &dbrTempDouble, sizeof(double));
	    ntohd(dbrTempDouble, (dbrCtrlDoub -> upper_ctrl_limit));
	    bcopy(pDrvl, (char *) &dbrTempDouble, sizeof(double));
	    ntohd(dbrTempDouble, (dbrCtrlDoub -> lower_ctrl_limit));
	    dbctdv = &(dbrCtrlDoub -> value);
	    while (count--)
		{
		bcopy(agentBuff, (char *) &dbrTempDouble, sizeof(dbrTempDouble));
		ntohd(dbrTempDouble, *dbctdv);
		dbctdv++;
		agentBuff += dataSize;
		}
	    break;
	    }
	}

    return (0);
}
