/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include	<stdio.h>
#ifndef AXP
#include	<string.h>
#else
#include	<strings.h>
#endif
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<errno.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

static CODE4		cb;
static char		pathbuff[128];
static char		sysbuff[256];
static struct stat	statbuff;

static int create_subdir (void);
static int convert_pid (void);

#define	TIS4000DB	"/tis4000db"
#define	DBCNVRT		"t43t44pidcnvrt"



/* t50 pid record definition, need to change t43 field AUTOCLOSE
   from Logical to Char 10
*/
F4FIELD_INFO	pidFieldInfo[] = 
	{
	{"NAME",	'C',	28,	0},
	{"INITLSETPT",	'N',	10,	4},
	{"REVERSEINI",	'L',	1,	0},
	{"SETPTSPEC",	'C',	33,	0},
	{"SETPTCONST",	'N',	10,	4},
	{"SETPTMS",	'L',	1,	0},
	{"SETPTPROC",	'L',	1,	0},
	{"MVALSPEC",	'C',	33,	0},
	{"MVALMS",	'L',	1,	0},
	{"MVALPROC",	'L',	1,	0},
	{"SETCLAMPHI",	'N',	11,	4},
	{"SETCLAMPLO",	'N',	11,	4},
	{"EGUHIGH",	'N',	11,	4},
	{"EGULOW",	'N',	11,	4},
	{"EGUTAG",	'C',	16,	0},
	{"DISPHIGH",	'N',	11,	4},
	{"DISPLOW",	'N',	11,	4},
	{"PRECISION",	'N',	1,	0},
	{"AUTOMATIC",	'L',	1,	0},
	{"CASFIELD",	'C',	33,	0},
	{"CASCADEMS",	'L',	1,	0},
	{"CASPROCINP",	'L',	1,	0},
	{"CASOPENFLG",	'L',	1,	0},
	{"CASSTPTPCT",	'L',	1,	0},
	{"INCREMENTL",	'L',	1,	0},
	{"OUTPUTCHAN",	'C',	33,	0},
	{"OUTCLAMPHI",	'N',	6,	2},
	{"OUTCLAMPLO",	'N',	7,	2},
	{"OUTREVERSE",	'L',	1,	0},
	{"PROP_BAND",	'N',	7,	3},
	{"RESET",	'N',	7,	3},
	{"RATE",	'N',	7,	3},
	{"ERRSQUARED",	'L',	1,	0},
	{"DELTAERROR",	'L',	1,	0},
	{"TRKMODENA",	'L',	1,	0},
	{"HOLDLOSTMV",	'L',	1,	0},
	{"NOMVOUTVAL",	'N',	7,	3},
	{"DEVLIMIT",	'N',	11,	4},
	{"DEVDEADBND",	'N',	11,	4},
	{"DEVSEVER",	'C',	13,	0},
	{"NOINPSEVER",	'C',	13,	0},
	{"NOMVSEVER",	'C',	13,	0},
	{"NOOUTSEVER",	'C',	13,	0},
	{"NOCCSEVER",	'C',	13,	0},
	{"CLDEADBAND",	'N',	11,	4},
	{"CLHISEVER",	'C',	13,	0},
	{"CLLOSEVER",	'C',	13,	0},
	{"TRKMODSEVR",	'C',	13,	0},
	{"MONDEADBND",	'N',	7,	3},
	{"ARCDEADBND",	'N',	7,	3},
	{"AUTOCLOSE",	'C',	10,	0},
	{"INITACT",	'C',	13,	0},
	{"SETPTRAMP",	'N',	11,	4},
	{"SETPTCROSS",	'L',	1,	0},
	{"SETPTGAP",	'N',	11,	4},
	{"UPDATEISTP",	'L',	1,	0},
	{0,0,0,0}
	};

T4TAG_INFO	pidTagInfo[] = 
	{
	{"NAME",	"NAME",	0,	0,	0},
	{0,0,0,0,0}
	};

int main()
{
    fprintf (stdout, "starting t43t44pid conversion\n\n");
    d4init (&cb);

#if defined(WIN32) || defined(USECB63)
    cb.errGo = 0;
    cb.errOpen = 0;
    cb.errCreate = 0;
    cb.errFieldName = 0;
#else
    cb.go_error = 0;
    cb.open_error = 0;
    cb.create_error = 0;
    cb.field_name_error = 0;
#endif WIN32 or USECB63

    if (create_subdir ())
	return (1);
    if (convert_pid ())
	return (1);
    d4close_all (&cb);
    fprintf (stdout, "\nt43t44pid conversion finished\n");
    return (0);
}


static int create_subdir (void)
{
    sprintf (pathbuff, "%s/%s", TIS4000DB, DBCNVRT);
    if (stat (pathbuff, &statbuff))
    {
	if (errno == ENOENT)
	{
	    sprintf (sysbuff, "mkdir %s", pathbuff);
	    system (sysbuff);
	    if (stat (pathbuff, &statbuff))
	    {
		fprintf (stdout, "t43t44pidconv err: failed to create %s\n", pathbuff);
		return (1);
	    }
	    else
	    {
		fprintf (stdout, "t43t44pidconv: %s created\n", pathbuff);
		return (0);
	    }
	}
	else
	{
	    fprintf (stdout, "t43t44pidconv err: stat problem (%d) on %s\n", errno, pathbuff);
	    return (1);
	}
    }
    else
    {
	fprintf (stdout, "t43t44pidconv err: %s already exists\n", pathbuff);
    }
    return (0);
}


static int convert_pid (void)
{
    DATA4	*old_pid, *new_pid;
    FIELD4	*old_outclamplo, *new_outclamplo;
    FIELD4	*old_outclamphi, *new_outclamphi;
    FIELD4	*old_autoclose, *new_autoclose;
    FIELD4	*old_initact, *new_initact;
    FIELD4	*old_setptramp, *new_setptramp;
    FIELD4	*old_setptcross, *new_setptcross;
    FIELD4	*old_setptgap, *new_setptgap;
    FIELD4	*old_updateistp, *new_updateistp;
    int		old_record_width, new_record_width;
    char	*old_rec, *new_rec;
    char	*pold_autoclose, *pold_outclamphi;
    char	*pold_outclamplo, *pnew_outclamplo;

    fprintf (stdout, "...converting pid.dbf\n");

    sprintf (pathbuff, "%s/pid", TIS4000DB);
    if (!(old_pid = d4open (&cb, pathbuff)))
    {
	fprintf (stdout, "t43t44pidconv err: open error on %s\n", pathbuff);
	return (1);
    }
    d4alias_set (old_pid, "old_pid");

    /* save the old dbase files */
    sprintf (sysbuff, "cp -p %s/pid.dbf %s/%s/pid.dbf.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);
    sprintf (sysbuff, "cp -p %s/pid.mdx %s/%s/pid.mdx.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);

    /* create the new dbase file */
    sprintf (pathbuff, "%s/%s/pid", TIS4000DB,DBCNVRT);
    if (!(new_pid = d4create (&cb, pathbuff, pidFieldInfo, pidTagInfo)))
    {
	fprintf (stdout, "t43t44pidconv err: failed to create %s\n", pathbuff);
	d4close (old_pid);
	return (1);
    }

    old_record_width = d4record_width (old_pid);
    new_record_width = d4record_width (new_pid);
    if (old_record_width == new_record_width)
    {
	fprintf (stdout, "t43t44pidconv err: pid has already been converted\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_outclamphi = d4field (old_pid, "OUTCLAMPHI")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: OUTCLAMPHI not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    pold_outclamphi = f4ptr (old_outclamphi);

    if ((new_outclamphi = d4field (new_pid, "OUTCLAMPHI")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: OUTCLAMPHI not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_outclamplo = d4field (old_pid, "OUTCLAMPLO")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: OUTCLAMPLO not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    pold_outclamplo = f4ptr (old_outclamplo);

    if ((new_outclamplo = d4field (new_pid, "OUTCLAMPLO")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: OUTCLAMPLO not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    pnew_outclamplo = f4ptr (new_outclamplo);

    if ((old_autoclose = d4field (old_pid, "AUTOCLOSE")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: AUTOCLOSE not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((f4type (old_autoclose)) != r4log)
    {
	fprintf (stdout, "t43/t44conv err: AUTOCLOSE field is not a logical\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    /* ptr to AUTOCLOSE field in old_pid */
    pold_autoclose = f4ptr (old_autoclose);

    if ((new_autoclose = d4field (new_pid, "AUTOCLOSE")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: AUTOCLOSE not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_initact = d4field (old_pid, "INITACT")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: INITACT not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((new_initact = d4field (new_pid, "INITACT")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: INITACT not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_setptramp = d4field (old_pid, "SETPTRAMP")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: SETPTRAMP not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((new_setptramp = d4field (new_pid, "SETPTRAMP")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: SETPTRAMP not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_setptcross = d4field (old_pid, "SETPTCROSS")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: SETPTCROSS not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((new_setptcross = d4field (new_pid, "SETPTCROSS")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: SETPTCROSS not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_setptgap = d4field (old_pid, "SETPTGAP")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: SETPTGAP not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((new_setptgap = d4field (new_pid, "SETPTGAP")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: SETPTGAP not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_updateistp = d4field (old_pid, "UPDATEISTP")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: UPDATEISTP not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((new_updateistp = d4field (new_pid, "UPDATEISTP")) == 0)
    {
	fprintf (stdout, "t43t44pidconv err: UPDATEISTP not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    old_rec = d4record (old_pid);
    new_rec = d4record (new_pid);

    for (d4top (old_pid); !d4eof (old_pid); d4skip (old_pid, 1))
    {
	/* do not copy if deleted */
	if (d4deleted (old_pid))
	{
	    continue;
	}
	d4append_start (new_pid, 0);
	d4blank (new_pid);
	memcpy (new_rec, old_rec, pold_outclamphi - old_rec);
	f4assign_field (new_outclamphi, old_outclamphi);
	memcpy (pnew_outclamplo, pold_outclamplo, (pold_autoclose - pold_outclamplo));
	if ((f4true (old_autoclose)) == 0)
	{
	    f4assign (new_autoclose, "Automatic");
	}
	else
	{
	    f4assign (new_autoclose, "Manual");
	}
	f4assign_field (new_initact, old_initact);
	f4assign_field (new_setptramp, old_setptramp);
	f4assign_field (new_setptcross, old_setptcross);
	f4assign_field (new_setptgap, old_setptgap);
	f4assign_field (new_updateistp, old_updateistp);

	d4append (new_pid);
	d4flush (new_pid);
    }
    d4close (old_pid);
    d4close (new_pid);

    fprintf (stdout, "...finished pid.dbf\n");
    return (0);
}

