/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include	<stdio.h>
#ifndef AXP
#include	<string.h>
#else
#include	<strings.h>
#endif
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<errno.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include	"pvmap.h"

static CODE4		cb;
static char		pathbuff[128];
static char		sysbuff[256];
static struct stat	statbuff;

static int create_subdir (void);
static int convert_pvmap (void);

#define	TIS4000DB	"/tis4000db"
#define	DBCNVRT		"t51pvrndx"


int main()
{
    fprintf (stdout, "starting T51pvrndx conversion\n\n");
    d4init (&cb);

#if defined(WIN32) || defined(USECB63)
    cb.errGo = 0;
    cb.errOpen = 0;
    cb.errCreate = 0;
    cb.errFieldName = 0;
#else
    cb.go_error = 0;
    cb.open_error = 0;
    cb.create_error = 0;
    cb.field_name_error = 0;
#endif WIN32 or USECB63

    if (create_subdir ())
	return (1);
    if (convert_pvmap ())
	return (1);
    d4close_all (&cb);
    fprintf (stdout, "\nT51pvrndx conversion finished\n");
    return (0);
}


static int create_subdir (void)
{
    sprintf (pathbuff, "%s/%s", TIS4000DB, DBCNVRT);
    if (stat (pathbuff, &statbuff))
    {
	if (errno == ENOENT)
	{
	    sprintf (sysbuff, "mkdir %s", pathbuff);
	    system (sysbuff);
	    if (stat (pathbuff, &statbuff))
	    {
		fprintf (stdout, "t51pvrndx err: failed to create %s\n", pathbuff);
		return (1);
	    }
	    else
	    {
		fprintf (stdout, "t51pvrndx: %s created\n", pathbuff);
		return (0);
	    }
	}
	else
	{
	    fprintf (stdout, "t51pvrndx err: stat problem (%d) on %s\n", errno, pathbuff);
	    return (1);
	}
    }
    else
    {
	fprintf (stdout, "t51pvrndx err: %s already exists\n", pathbuff);
    }
    return (0);
}

int convert_pvmap (void)
{
    DATA4	*old_pvmap, *new_pvmap;
    char	*old_rec;
    long	recCount;
    long	numRecs;
    int		pct;
    char	*new_rec;

    fprintf (stdout, "...converting pvmap.pdbf\n");

    sprintf (pathbuff, "%s/pvmap", TIS4000DB);
    if (!(old_pvmap = d4open (&cb, pathbuff)))
    {
	fprintf (stdout, "t51pvrndx err: open error on %s\n", pathbuff);
	return (1);
    }
    d4alias_set (old_pvmap, "old_pvmap");

    /* save the old dbase files */
    sprintf (sysbuff, "cp -p %s/pvmap.dbf %s/%s/pvmap.dbf.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);
    sprintf (sysbuff, "cp -p %s/pvmap.mdx %s/%s/pvmap.mdx.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);

    /* create the new dbase file */
    sprintf (pathbuff, "%s/%s/pvmap", TIS4000DB,DBCNVRT);
    if (!(new_pvmap = d4create (&cb, pathbuff, pvmapFieldInfo, pvmapTagInfo)))
    {
	fprintf (stdout, "t51pvrndx err: failed to create %s\n", pathbuff);
	d4close (old_pvmap);
	return (1);
    }
    d4alias_set (new_pvmap, "new_pvmap");

    old_rec = d4record (old_pvmap);
    new_rec = d4record (new_pvmap);

    numRecs = d4reccount (old_pvmap);
    recCount = 0;
    for (d4top (old_pvmap); !d4eof (old_pvmap); d4skip (old_pvmap, 1))
    {
	d4append_start (new_pvmap, 0);
	d4blank (new_pvmap);
	/* copy old record into new record, they are same size */
	memcpy (new_rec, old_rec, d4record_width (old_pvmap));
	d4append (new_pvmap);
	d4flush (new_pvmap);
	recCount++;
	if (!((pct = ((recCount / numRecs) * 100)) % 10))
	{
	    fprintf (stdout, "%d%% records completed\n", pct);
	}
    }

    d4close (old_pvmap);
    d4close (new_pvmap);

    fprintf (stdout, "...finished pvmap.dbf\n");
    return (0);
}
