/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include	<stdio.h>
#ifndef AXP
#include	<string.h>
#else
#include	<strings.h>
#endif
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<errno.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include	"ai.h"
#include	"pat.h"
#include	"pvmap.h"
#include	"strngin.h"
#include	"tagvudb.h"

static CODE4		cb;
static char		pathbuff[128];
static char		sysbuff[256];
static char		*old_rec, *new_rec;
static struct stat	statbuff;

static int create_subdir (void);
static int convert_ai (void);
static int convert_pat (void);
static int convert_pvmap (void);
static int convert_strngin (void);
static int convert_tagvu (void);

#define	TIS4000DB	"/tis4000db"
#define	DBCNVRT		"t51t51cnvrt"


int main()
{
    fprintf (stdout, "starting T51T51 conversion\n\n");
    d4init (&cb);

#if defined(WIN32) || defined(USECB63)
    cb.errGo = 0;
    cb.errOpen = 0;
    cb.errCreate = 0;
    cb.errFieldName = 0;
#else
    cb.go_error = 0;
    cb.open_error = 0;
    cb.create_error = 0;
    cb.field_name_error = 0;
#endif WIN32 or USECB63

    if (create_subdir ())
	return (1);
    if (convert_ai ())
	return (1);
    if (convert_pat ())
	return (1);
    if (convert_pvmap ())
	return (1);
    if (convert_strngin ())
	return (1);
    if (convert_tagvu ())
	return (1);
    d4close_all (&cb);
    fprintf (stdout, "\nT51T51 conversion finished\n");
    return (0);
}


static int create_subdir (void)
{
    sprintf (pathbuff, "%s/%s", TIS4000DB, DBCNVRT);
    if (stat (pathbuff, &statbuff))
    {
	if (errno == ENOENT)
	{
	    sprintf (sysbuff, "mkdir %s", pathbuff);
	    system (sysbuff);
	    if (stat (pathbuff, &statbuff))
	    {
		fprintf (stdout, "t51conv err: failed to create %s\n", pathbuff);
		return (1);
	    }
	    else
	    {
		fprintf (stdout, "t51conv: %s created\n", pathbuff);
		return (0);
	    }
	}
	else
	{
	    fprintf (stdout, "t51conv err: stat problem (%d) on %s\n", errno, pathbuff);
	    return (1);
	}
    }
    else
    {
	fprintf (stdout, "t51conv err: %s already exists\n", pathbuff);
    }
    return (0);
}


static int convert_ai (void)
{
    DATA4	*old_ai, *new_ai;
    FIELD4	*devsetpt;
    FIELD4	*gain, *sigtype;

    fprintf (stdout, "...converting ai.dbf\n");

    sprintf (pathbuff, "%s/ai", TIS4000DB);
    if (!(old_ai = d4open (&cb, pathbuff)))
    {
	fprintf (stdout, "t51conv err: open error on %s\n", pathbuff);
	return (1);
    }
    if ((devsetpt = d4field (old_ai, "DEVSETPT")) == 0)
    {
	fprintf (stdout, "t51conv err: %s is not a t51 file\n", pathbuff);
	d4close (old_ai);
	return (1);
    }
    if ((gain = d4field (old_ai, "GAIN")) != 0)
    {
	fprintf (stdout, "t51conv err: %s already converted\n", pathbuff);
	d4close (old_ai);
	return (1);
    }
    d4alias_set (old_ai, "old_ai");

    /* save the old dbase files */
    sprintf (sysbuff, "cp -p %s/ai.dbf %s/%s/ai.dbf.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);
    sprintf (sysbuff, "cp -p %s/ai.mdx %s/%s/ai.mdx.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);

    /* create the new dbase file */
    sprintf (pathbuff, "%s/%s/ai", TIS4000DB,DBCNVRT);
    if (!(new_ai = d4create (&cb, pathbuff, aiFieldInfo, aiTagInfo)))
    {
	fprintf (stdout, "t51conv err: failed to create %s\n", pathbuff);
	d4close (old_ai);
	return (1);
    }

    old_rec = d4record (old_ai);
    new_rec = d4record (new_ai);

    if ((gain = d4field (new_ai, "GAIN")) == 0)
    {
	fprintf (stdout, "t51conv err: GAIN not exist\n");
	d4close (old_ai);
	d4close (new_ai);
	return (1);
    }

    if ((sigtype = d4field (new_ai, "SIGTYPE")) == 0)
    {
	fprintf (stdout, "t51conv err: SIGTYPE not exist\n");
	d4close (old_ai);
	d4close (new_ai);
	return (1);
    }

    for (d4top (old_ai); !d4eof (old_ai); d4skip (old_ai, 1))
    {
	d4append_start (new_ai, 0);
	d4blank (new_ai);
	memcpy (new_rec, old_rec, d4record_width (old_ai));
	f4assign_int (gain, 0);
	f4assign (sigtype, "N/A");
	d4append (new_ai);
	d4flush (new_ai);
    }

    d4close (old_ai);
    d4close (new_ai);

    fprintf (stdout, "...finished ai.dbf\n");
    return (0);
}


static int convert_pat (void)
{
    DATA4	*old_pat, *new_pat;
    FIELD4	*boardtype;
    FIELD4	*rlboard;
    FIELD4	*fbboard;
    char	*old_rec;
    char	*new_rec;

    fprintf (stdout, "...converting pat.pdbf\n");

    sprintf (pathbuff, "%s/pat", TIS4000DB);
    if (!(old_pat = d4open (&cb, pathbuff)))
    {
	fprintf (stdout, "t51conv err: open error on %s\n", pathbuff);
	return (1);
    }
    d4alias_set (old_pat, "old_pat");

    /* save the old dbase files */
    sprintf (sysbuff, "cp -p %s/pat.dbf %s/%s/pat.dbf.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);
    sprintf (sysbuff, "cp -p %s/pat.mdx %s/%s/pat.mdx.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);

    /* create the new dbase file */
    sprintf (pathbuff, "%s/%s/pat", TIS4000DB,DBCNVRT);
    if (!(new_pat = d4create (&cb, pathbuff, patFieldInfo, patTagInfo)))
    {
	fprintf (stdout, "t51conv err: failed to create %s\n", pathbuff);
	d4close (old_pat);
	return (1);
    }

    old_rec = d4record (old_pat);
    new_rec = d4record (new_pat);

    if ((boardtype = d4field (new_pat, "BOARDTYPE")) == 0)
    {
	fprintf (stdout, "t51conv err: BOARDTYPE (new) not exist\n");
	d4close (old_pat);
	d4close (new_pat);
	return (1);
    }

    if ((rlboard = d4field (new_pat, "RLBOARD")) == 0)
    {
	fprintf (stdout, "t51conv err: RLBOARD (new) not exist\n");
	d4close (old_pat);
	d4close (new_pat);
	return (1);
    }

    if ((fbboard = d4field (new_pat, "FBBOARD")) == 0)
    {
	fprintf (stdout, "t51conv err: FBBOARD (new) not exist\n");
	d4close (old_pat);
	d4close (new_pat);
	return (1);
    }

    for (d4top (old_pat); !d4eof (old_pat); d4skip (old_pat, 1))
    {
	d4append_start (new_pat, 0);
	d4blank (new_pat);
	/* copy old record into new record, they are same size */
	memcpy (new_rec, old_rec, d4record_width (old_pat));
	if (strncmp (f4str (boardtype), "REFERENCE", 9) == 0)
	{
	    f4assign (boardtype, "PAT BO REFERENCE");
	}
	if (strncmp (f4str (rlboard), "REFERENCE", 9) == 0)
	{
	    f4assign (rlboard, "PAT BI REFERENCE");
	}
	if (strncmp (f4str (fbboard), "REFERENCE", 9) == 0)
	{
	    f4assign (fbboard, "PAT AI REFERENCE");
	}
	d4append (new_pat);
	d4flush (new_pat);
    }

    d4close (old_pat);
    d4close (new_pat);

    fprintf (stdout, "...finished pat.dbf\n");
    return (0);
}


static int convert_pvmap (void)
{
    DATA4	*old_pvmap, *new_pvmap;
    FIELD4	*res_old, *res_new, *adisable;
    FIELD4	*alarmdelay;
    FIELD4	*devtype;
    FIELD4	*infotag;
    FIELD4	*type;
    char	*pres_old, *pres_new;

    fprintf (stdout, "...converting pvmap.dbf\n");

    sprintf (pathbuff, "%s/pvmap", TIS4000DB);
    if (!(old_pvmap = d4open (&cb, pathbuff)))
    {
	fprintf (stdout, "t51conv err: open error on %s\n", pathbuff);
	return (1);
    }
    if ((alarmdelay = d4field (old_pvmap, "ALARMDELAY")) == 0)
    {
	fprintf (stdout, "t51conv err: %s is not a t51 file\n", pathbuff);
	d4close (old_pvmap);
	return (1);
    }
    if ((adisable = d4field (old_pvmap, "ADISABLE")) != 0)
    {
	fprintf (stdout, "t51conv err: %s already converted\n", pathbuff);
	d4close (old_pvmap);
	return (1);
    }
    d4alias_set (old_pvmap, "old_pvmap");

    /* save the old dbase files */
    sprintf (sysbuff, "cp -p %s/pvmap.dbf %s/%s/pvmap.dbf.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);
    sprintf (sysbuff, "cp -p %s/pvmap.mdx %s/%s/pvmap.mdx.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);

    /* create the new dbase file */
    sprintf (pathbuff, "%s/%s/pvmap", TIS4000DB,DBCNVRT);
    if (!(new_pvmap = d4create (&cb, pathbuff, pvmapFieldInfo, pvmapTagInfo)))
    {
	fprintf (stdout, "t51conv err: failed to create %s\n", pathbuff);
	d4close (old_pvmap);
	return (1);
    }

    old_rec = d4record (old_pvmap);
    new_rec = d4record (new_pvmap);

    if ((res_old = d4field (old_pvmap, "RESRCNAME")) == 0)
    {
	fprintf (stdout, "t51conv err: RESRCNAME (old) not exist\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }
    /* ptr to start of RESRCNAME in old record */
    pres_old = f4ptr (res_old);

    if ((res_new = d4field (new_pvmap, "RESRCNAME")) == 0)
    {
	fprintf (stdout, "t51conv err: RESRCNAME (new) not exist\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }
    /* ptr to start of RESRCNAME field in new record */
    pres_new = f4ptr (res_new);

    if ((devtype = d4field (new_pvmap, "DEVTYPE")) == 0)
    {
	fprintf (stdout, "t51conv err: DEVTYPE (new) not exist\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((infotag = d4field (new_pvmap, "INFOTAG")) == 0)
    {
	fprintf (stdout, "t51conv err: INFOTAG (new) not exist\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((type = d4field (new_pvmap, "TYPE")) == 0)
    {
	fprintf (stdout, "t51conv err: TYPE (new) not exist\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    for (d4top (old_pvmap); !d4eof (old_pvmap); d4skip (old_pvmap, 1))
    {
	d4append_start (new_pvmap, 0);
	d4blank (new_pvmap);
	/* copy old record up to the RESRCNAME field to the new record */
	memcpy (new_rec, old_rec, pres_old - old_rec);
	/* copy from RESRCNAME in old record to new record */
	memcpy (pres_new, pres_old, d4record_width (old_pvmap) -
				      (pres_old - old_rec));
	if (strncmp (f4str (devtype), "TIME/DATE", 9) == 0)
	{
	    f4assign (devtype, "SYSINFO");
	}
	f4assign_char (infotag, 'F');
	if (strncmp (f4str (type), "pat", 3) == 0)
	{
	    if (strncmp (f4str (devtype), "REFERENCE", 9) == 0)
	    {
		f4assign (devtype, "PAT BO REFERENCE");
	    }
	}
	d4append (new_pvmap);
	d4flush (new_pvmap);
    }

    d4close (old_pvmap);
    d4close (new_pvmap);

    fprintf (stdout, "...finished pvmap.dbf\n");
    return (0);
}


static int convert_strngin (void)
{
    DATA4	*old_strngin, *new_strngin;
    FIELD4	*cvsv;
    FIELD4	*bdtype;

    fprintf (stdout, "...converting strngin.dbf\n");

    sprintf (pathbuff, "%s/strngin", TIS4000DB);
    if (!(old_strngin = d4open (&cb, pathbuff)))
    {
	fprintf (stdout, "t51conv err: open error on %s\n", pathbuff);
	return (1);
    }
    if ((cvsv = d4field (old_strngin, "CHGVALSEVR")) != 0)
    {
	fprintf (stdout, "t51conv err: %s already converted\n", pathbuff);
	d4close (old_strngin);
	return (1);
    }
    d4alias_set (old_strngin, "old_strngin");

    /* save the old dbase files */
    sprintf (sysbuff, "cp -p %s/strngin.dbf %s/%s/strngin.dbf.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);
    sprintf (sysbuff, "cp -p %s/strngin.mdx %s/%s/strngin.mdx.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);

    /* create the new dbase file */
    sprintf (pathbuff, "%s/%s/strngin", TIS4000DB,DBCNVRT);
    if (!(new_strngin = d4create (&cb, pathbuff, strnginFieldInfo, strnginTagInfo)))
    {
	fprintf (stdout, "t51conv err: failed to create %s\n", pathbuff);
	d4close (old_strngin);
	return (1);
    }

    old_rec = d4record (old_strngin);
    new_rec = d4record (new_strngin);

    if ((cvsv = d4field (new_strngin, "CHGVALSEVR")) == 0)
    {
	fprintf (stdout, "t51conv err: CHGVALSEVR not exist\n");
	d4close (old_strngin);
	d4close (new_strngin);
	return (1);
    }

    if ((bdtype = d4field (new_strngin, "BOARDTYPE")) == 0)
    {
	fprintf (stdout, "t51conv err: BOARDTYPE not exist\n");
	d4close (old_strngin);
	d4close (new_strngin);
	return (1);
    }

    for (d4top (old_strngin); !d4eof (old_strngin); d4skip (old_strngin, 1))
    {
	d4append_start (new_strngin, 0);
	d4blank (new_strngin);
	memcpy (new_rec, old_rec, d4record_width (old_strngin));
	f4assign (cvsv, "NO ALARM");
	if (strncmp (f4str (bdtype), "TIME/DATE", 9) == 0)
	{
	    f4assign (bdtype, "SYSINFO");
	}
	d4append (new_strngin);
	d4flush (new_strngin);
    }

    d4close (old_strngin);
    d4close (new_strngin);

    fprintf (stdout, "...finished strngin.dbf\n");
    return (0);
}


static int convert_tagvu (void)
{
    DATA4	*old_tagvu, *new_tagvu;
    FIELD4	*oldptname, *newptname;
    FIELD4	*oldloc, *newloc;
    FIELD4	*olddesc, *newdesc;
    FIELD4	*olddate_apl, *newdate_apl;
    FIELD4	*oldtime_apl, *newtime_apl;
    FIELD4	*oldoper_apl, *newoper_apl;
    FIELD4	*oldreason_apl, *newreason_apl;
    FIELD4	*olddate_rm, *newdate_rm;
    FIELD4	*oldtime_rm, *newtime_rm;
    FIELD4	*oldoper_rm, *newoper_rm;
    FIELD4	*oldreason_rm, *newreason_rm;
    FIELD4	*tagtype;
    int		create_tagvu = 0;

    fprintf (stdout, "...converting tagvu.dbf\n");

    sprintf (pathbuff, "%s/tagvu.dbf", TIS4000DB);
    /* since the odl tagvu created tagvu.dbf, there may not be one
       so all we have to do is create it
    */
    if (stat (pathbuff, &statbuff) != 0)
    {
	if (errno == ENOENT)
	    create_tagvu = 1;
	else
	{
	    fprintf (stdout, "t51conv err: stat problem (%d) on %s\n", errno, pathbuff);
	    return (1);
	}
    }
    sprintf (pathbuff, "%s/tagvu", TIS4000DB);
    if (create_tagvu == 0)
    {
	if (!(old_tagvu = d4open (&cb, pathbuff)))
	{
	    fprintf (stdout, "t51conv err: open error on %s\n", pathbuff);
	    return (1);
	}
	if ((tagtype = d4field (old_tagvu, "TAGTYPE")) != 0)
	{
	    fprintf (stdout, "t51conv err: %s already converted\n", pathbuff);
	    d4close (old_tagvu);
	    return (1);
	}
	d4alias_set (old_tagvu, "old_tagvu");

	/* save the old dbase files */
	sprintf (sysbuff, "cp -p %s/tagvu.dbf %s/%s/tagvu.dbf.sav", 
			   TIS4000DB, TIS4000DB, DBCNVRT);
	system (sysbuff);
	sprintf (sysbuff, "cp -p %s/tagvu.mdx %s/%s/tagvu.mdx.sav", 
			   TIS4000DB, TIS4000DB, DBCNVRT);
	system (sysbuff);
    }

    /* create the new dbase file */
    sprintf (pathbuff, "%s/%s/tagvu", TIS4000DB,DBCNVRT);
    if (!(new_tagvu = d4create (&cb, pathbuff, tagvuFieldInfo, tagvuTagInfo)))
    {
	fprintf (stdout, "t51conv err: failed to create %s\n", pathbuff);
	if (create_tagvu == 0)
	    d4close (old_tagvu);
	return (1);
    }
    if (create_tagvu)
    {
	d4close (new_tagvu);
	fprintf (stdout, "...finished tagvu.dbf\n");
	return (0);
    }

    old_rec = d4record (old_tagvu);
    new_rec = d4record (new_tagvu);

    if ((oldptname = d4field (old_tagvu, "POINTNAME")) == 0)
    {
	fprintf (stdout, "t51conv err: POINTNAME not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newptname = d4field (new_tagvu, "POINTNAME")) == 0)
    {
	fprintf (stdout, "t51conv err: POINTNAME not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldloc = d4field (old_tagvu, "LOCATION")) == 0)
    {
	fprintf (stdout, "t51conv err: LOCATION not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newloc = d4field (new_tagvu, "LOCATION")) == 0)
    {
	fprintf (stdout, "t51conv err: LOCATION not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((olddesc = d4field (old_tagvu, "DESCRIPTOR")) == 0)
    {
	fprintf (stdout, "t51conv err: DESCRIPTOR not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newdesc = d4field (new_tagvu, "DESCRIPTOR")) == 0)
    {
	fprintf (stdout, "t51conv err: DESCRIPTOR not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((olddate_apl = d4field (old_tagvu, "DATE_APL")) == 0)
    {
	fprintf (stdout, "t51conv err: DATE_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newdate_apl = d4field (new_tagvu, "DATE_APL")) == 0)
    {
	fprintf (stdout, "t51conv err: DATE_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldtime_apl = d4field (old_tagvu, "TIME_APL")) == 0)
    {
	fprintf (stdout, "t51conv err: TIME_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newtime_apl = d4field (new_tagvu, "TIME_APL")) == 0)
    {
	fprintf (stdout, "t51conv err: TIME_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldoper_apl = d4field (old_tagvu, "OPER_APL")) == 0)
    {
	fprintf (stdout, "t51conv err: OPER_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newoper_apl = d4field (new_tagvu, "OPER_APL")) == 0)
    {
	fprintf (stdout, "t51conv err: OPER_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldreason_apl = d4field (old_tagvu, "REASON_APL")) == 0)
    {
	fprintf (stdout, "t51conv err: REASON_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newreason_apl = d4field (new_tagvu, "REASON_APL")) == 0)
    {
	fprintf (stdout, "t51conv err: REASON_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((olddate_rm = d4field (old_tagvu, "DATE_RMV")) == 0)
    {
	fprintf (stdout, "t51conv err: DATE_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newdate_rm = d4field (new_tagvu, "DATE_RMV")) == 0)
    {
	fprintf (stdout, "t51conv err: DATE_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldtime_rm = d4field (old_tagvu, "TIME_RMV")) == 0)
    {
	fprintf (stdout, "t51conv err: TIME_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newtime_rm = d4field (new_tagvu, "TIME_RMV")) == 0)
    {
	fprintf (stdout, "t51conv err: TIME_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldoper_rm = d4field (old_tagvu, "OPER_RMV")) == 0)
    {
	fprintf (stdout, "t51conv err: OPER_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newoper_rm = d4field (new_tagvu, "OPER_RMV")) == 0)
    {
	fprintf (stdout, "t51conv err: OPER_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldreason_rm = d4field (old_tagvu, "REASON_RMV")) == 0)
    {
	fprintf (stdout, "t51conv err: REASON_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newreason_rm = d4field (new_tagvu, "REASON_RMV")) == 0)
    {
	fprintf (stdout, "t51conv err: REASON_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((tagtype = d4field (new_tagvu, "TAGTYPE")) == 0)
    {
	fprintf (stdout, "t51conv err: TAGTYPE not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    for (d4top (old_tagvu); !d4eof (old_tagvu); d4skip (old_tagvu, 1))
    {
	d4append_start (new_tagvu, 0);
	d4blank (new_tagvu);
	f4assign_field (newptname, oldptname);
	f4assign_field (newloc, oldloc);
	f4assign_field (newdesc, olddesc);
	f4assign_field (newdate_apl, olddate_apl);
	f4assign_field (newtime_apl, oldtime_apl);
	f4assign_field (newoper_apl, oldoper_apl);
	f4assign_field (newreason_apl, oldreason_apl);
	f4assign_field (newdate_rm, olddate_rm);
	f4assign_field (newtime_rm, oldtime_rm);
	f4assign_field (newoper_rm, oldoper_rm);
	f4assign_field (newreason_rm, oldreason_rm);
	f4assign_char (tagtype, 'O');
	d4append (new_tagvu);
	d4flush (new_tagvu);
    }

    d4close (old_tagvu);
    d4close (new_tagvu);

    fprintf (stdout, "...finished tagvu.dbf\n");
    return (0);
}
