#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuAlarmInfoDialogClass.hxx"
#include "vuUtils.hxx"

// filter functions for the guidance and display fields
static int guidanceFilter (filteredText *text, vtextSelection *sel,
			   const vchar **s, long *len);
static int displayFilter (filteredText *text, vtextSelection *sel,
			  const vchar **s, long *len);


/*******************************************************************
*
*	ALARM INFO DIALOG CLASS
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuAlarmInfoDialogClass, vdialog, 
					"vuAlarmInfoDialogClass");

/*******************************************************************
*
*    vuAlarmInfoDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuAlarmInfoDialogClass::LoadInit(vresource res)
{
    dialogModified = vFALSE;
    dialogTitle = NULL;

    guidanceText = filteredTextitem::RequestSubLoad (this,
					vunames_vuAlarmInfoGuidance);
    
    displayText = filteredTextitem::RequestSubLoad (this,
					vunames_vuAlarmInfoDisplay);

    vcontainer *buttoncontainer = vcontainer::RequestSubLoad (this,
					vunames_vuAlarmInfoButtonContainer);

    applyButton = vuAlarmInfoApplyButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuAlarmInfoApplyButton);

    dismissButton = vuAlarmInfoDismissButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuAlarmInfoDismissButton);

    deleteButton = vuAlarmInfoDeleteButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuAlarmInfoDeleteButton);

    // call base class loadInit to finish the job
    vdialog::LoadInit(res);

    // set ptrs
    applyButton -> setAlarmInfoDialog (this);
    dismissButton -> setAlarmInfoDialog (this);
    deleteButton -> setAlarmInfoDialog (this);

    // store dialogModifed
    guidanceText -> SetTextModifiedPtr (&this -> dialogModified);
    displayText -> SetTextModifiedPtr (&this -> dialogModified);

    // set filters
    guidanceText -> SetItemFilterFunc (guidanceFilter);
    displayText -> SetItemFilterFunc (displayFilter);

}


/*******************************************************************
*
*    vuAlarmInfoDialogClass notify function
*
*******************************************************************/

void vuAlarmInfoDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	int result = VU_TRUE;
	if (dialogModified == vTRUE)
	{
	    result = vuMessageDialogCenter (this, VU_MOD_DISMISS_MSG);
	}
	if (result == VU_TRUE)
	{
	    Close();
	    vevent::StopProcessing ();
	}
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuAlarmInfoDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuAlarmInfoDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }

    if (dialogTitle != NULL)
    {
	vstrDestroy (dialogTitle);
    }

    vdialog::Destroy();
}


/*******************************************************************
*
*    vuAlarmInfoDialogClas Open function
*
*******************************************************************/

void vuAlarmInfoDialogClass::vuAlarmInfoOpen(vuMainDialogClass *mainDialg,
					    vchar *pv)
{
    mainDialog = mainDialg;
    pvname = pv;

    // map to database fields
    dataVuApplication *dataVuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = dataVuApp -> getDatabaseClass ();

    nameField     = DB_MapField (db -> vuAlarmInfoTable, FLD_NAME);
    guidanceField = DB_MapField (db -> vuAlarmInfoTable, FLD_GUIDANCE);
    displayField  = DB_MapField (db -> vuAlarmInfoTable, FLD_DISPLAY);


    vresource dialogTitles = mainDialog -> getTitlesResource();
    vstr *aiTitle = vresourceGetString (dialogTitles, vunames_vuAlarmInfoDialogTitle);
    dialogTitle = vstrCloneScribed (vcharScribeFormat (aiTitle, pvname));

    SetTitle (dialogTitle);

    vstrDestroy (aiTitle);

    // load up the dialog if data already exists
    char pvName[PVNAME_SZ + 1];
    char seekBuff[PVNAME_SZ + 1];

    vcharExportBounded (vcharScribeSystemExport (pvname),
			pvName,
			PVNAME_SZ);
    pvName[PVNAME_SZ] = 0;
    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);

    DB_RECORD_PTR rec = DB_Seek (db -> vuAlarmInfoTable, seekBuff);
    // if a record already exists, get data and display it
    if (rec != NULL)
    {
	char textGuidance[ALARMINFO_GUIDANCE_SZ + 1];
	char textDisplay[ALARMINFO_DISPLAY_SZ + 1];
	int  dataType;

	DB_GetFieldValue (rec, guidanceField, textGuidance, &dataType);
	DB_GetFieldValue (rec, displayField, textDisplay, &dataType);

	guidanceText -> SetTextScribed (vcharScribeSystem (textGuidance));
	displayText -> SetTextScribed (vcharScribeSystem (textDisplay));
    }
    // else make sure fields in display are blank
    else
    {
	guidanceText -> SetTextScribed (vcharScribeSystem (""));
	displayText -> SetTextScribed (vcharScribeSystem (""));
    }
    dialogModified = vFALSE;

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    Place(GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuAlarmInfoApplyButtonClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuAlarmInfoApplyButtonClass, vbutton, 
					"vuAlarmInfoApplyButtonClass");

/*******************************************************************
*
*    vuAlarmInfoApplyButtonClass Observer function
*
*******************************************************************/

void vuAlarmInfoApplyButtonClass::ObserveDialogItem (vevent *event)
{
    vuMainDialogClass *mainDialog = alarminfoDialog -> getMainDialog ();
    dataVuApplication *dataVuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = dataVuApp -> getDatabaseClass ();
    vchar *pv = alarminfoDialog -> getPvName ();

    int result = vuMessageDialogCenter (alarminfoDialog, 
					VU_ALARMINFO_UPDATE_REQ_MSG, 
					pv);
    if (result == VU_FALSE)
    {
	return;
    }

    char pvname[PVNAME_SZ + 1];
    vcharExportBounded (vcharScribeSystemExport (pv),
			pvname,
			PVNAME_SZ);
    pvname[PVNAME_SZ] = 0;

    char seekBuff[PVNAME_SZ + 1];
    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvname);

    DB_RECORD_PTR rec = DB_Seek (db -> vuAlarmInfoTable, seekBuff);

    // need to create a new record
    if (rec == 0)
    {
	rec = DB_NewRecord (db -> vuAlarmInfoTable, NULL);
	DB_PutFieldValue (rec, alarminfoDialog -> nameField, pvname);
    }
    char textGuidance[ALARMINFO_GUIDANCE_SZ + 1];
    vstr *text = alarminfoDialog -> guidanceText -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (text),
			textGuidance,
			ALARMINFO_GUIDANCE_SZ);
    textGuidance[ALARMINFO_GUIDANCE_SZ] = 0;
    DB_PutFieldValue (rec, alarminfoDialog -> guidanceField, textGuidance);
    vstrDestroy (text);

    char textDisplay[ALARMINFO_DISPLAY_SZ + 1];
    text = alarminfoDialog -> displayText -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (text),
			textDisplay,
			ALARMINFO_DISPLAY_SZ);
    textDisplay[ALARMINFO_DISPLAY_SZ] = 0;
    DB_PutFieldValue (rec, alarminfoDialog -> displayField, textDisplay);
    vstrDestroy (text);

    // turn off modified flag
    alarminfoDialog -> setDialogModified (vFALSE);

    mainDialog -> vuMainMessageCenter (vunames_vuAlarmInfoUpdateMsg, pv);
}


/*******************************************************************
*
*    vuAlarmInfoDeleteButtonClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuAlarmInfoDeleteButtonClass, vbutton, 
					"vuAlarmInfoDeleteButtonClass");

/*******************************************************************
*
*    vuAlarmInfoDeleteButtonClass Observer function
*
*******************************************************************/

void vuAlarmInfoDeleteButtonClass::ObserveDialogItem (vevent *event)
{
    vuMainDialogClass *mainDialog = alarminfoDialog -> getMainDialog ();
    dataVuApplication *dataVuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = dataVuApp -> getDatabaseClass ();
    vchar *pv = alarminfoDialog -> getPvName ();

    char pvname[PVNAME_SZ + 1];
    vcharExportBounded (vcharScribeSystemExport (pv),
			pvname,
			PVNAME_SZ);
    pvname[PVNAME_SZ] = 0;

    char seekBuff[PVNAME_SZ + 1];
    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvname);

    DB_RECORD_PTR rec = DB_Seek (db -> vuAlarmInfoTable, seekBuff);

    // can't delete what is not in the table
    if (rec == 0)
    {
	vuMessageDialogCenter (alarminfoDialog,
			       VU_ALARMINFO_DELETE_ERR_MSG,
			       pv);
	return;
    }

    int result = vuMessageDialogCenter (alarminfoDialog, 
					VU_ALARMINFO_DELETE_REQ_MSG, 
					pv);
    if (result == VU_FALSE)
    {
	return;
    }

    DB_DeleteRecord (db -> vuAlarmInfoTable, rec);

    // clear the fields
    alarminfoDialog -> guidanceText -> SetTextScribed (vcharScribeSystem (""));
    alarminfoDialog -> displayText -> SetTextScribed (vcharScribeSystem (""));

    // turn off modified flag
    alarminfoDialog -> setDialogModified (vFALSE);

    mainDialog -> vuMainMessageCenter (vunames_vuAlarmInfoDeleteMsg, pv);

}


/*******************************************************************
*
*    vuAlarmInfoDismissButtonClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuAlarmInfoDismissButtonClass, vbutton, 
					"vuAlarmInfoDismissButtonClass");

/*******************************************************************
*
*    vuAlarmInfoDismissButtonClass Observer function
*
*******************************************************************/

void vuAlarmInfoDismissButtonClass::ObserveDialogItem (vevent *event)
{
    int dialogModified = alarminfoDialog -> getDialogModified ();
    int result = VU_TRUE;

    if (dialogModified == VU_TRUE)
    {
	result = vuMessageDialogCenter (alarminfoDialog, VU_MOD_DISMISS_MSG);
    }
    if (result == VU_TRUE)
    {
	alarminfoDialog -> Close ();
	vevent::StopProcessing ();
    }
}



/*******************************************************************
*
*    vuAlarmInfoDialog filter functions
*
*******************************************************************/

static int guidanceFilter (filteredText *text, vtextSelection *sel,
			   const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, ALARMINFO_GUIDANCE_SZ - 1));
}

static int displayFilter (filteredText *text, vtextSelection *sel,
			  const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, ALARMINFO_DISPLAY_SZ - 1));
}

