#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER
#include vlistHEADER
#include vsliderHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbFldTypes.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "dbRecDes.h"
#include "dbRecType.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuUtils.hxx"
#include "vuArReducDialogClass.hxx"

static vchar *valFields[] = 
	    { 
		(vchar *)"VAL ", NULL
	    };
static vchar *biFields[] = 
	    { 
		(vchar *)"VAL ", (vchar *)"TCNT", NULL
	    };
static vchar *accumFields[] = 
	    { 
		(vchar *)"VAL ", (vchar *)"FSMP", (vchar *)"FTOT", (vchar *)"FAVG",
		(vchar *)"FMIN", (vchar *)"FMAX", (vchar *)"RSMP", (vchar *)"RTOT",
		(vchar *)"RAVG", (vchar *)"RMIN", (vchar *)"RMAX", NULL
	    };
static vchar *deadtimeFields[] = 
	    { 
		(vchar *)"VAL ", (vchar *)"CAVE", (vchar *)"CMAX", (vchar *)"CMIN", 
		(vchar *)"CTOT", NULL
	    };
static vchar *pidFields[] = 
	    { 
		(vchar *)"VAL ", (vchar *)"ASTP", (vchar *)"MVAL", NULL
	    };
static vchar *selFields[] = 
	    { 
		(vchar *)"VAL ", (vchar *)"SELM", NULL
	    };
static vchar *fanFields[] = 
	    { 
		(vchar *)"VAL ", (vchar *)"SELM", NULL
	    };
static vchar *udbFields[] = 
	    { 
		(vchar *)"VAL ", (vchar *)"FP01", (vchar *)"FP02", (vchar *)"FP03", 
		(vchar *)"FP04", (vchar *)"FP05", (vchar *)"FP06", (vchar *)"FP07", 
		(vchar *)"FP08", (vchar *)"BL01", (vchar *)"BL02", (vchar *)"BL03", 
		(vchar *)"BL04", (vchar *)"BL05", (vchar *)"BL06", (vchar *)"BL07", 
		(vchar *)"BL08", (vchar *)"NCMD", NULL
	    };
static vchar *patFields[] = 
	    { 
		(vchar *)"VAL ", (vchar *)"ASTP", NULL
	    };
static vchar *pumpFields[] = 
	    {
		(vchar *)"VAL ", (vchar *)"ACRT", NULL
	    };


/***********************************************************************
*
*	Ar Reduc Dialog Class
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION (vuArReducDialogClass, vdialog, 
					"vuArReducDialogClass");

/**********************************************************************
*
*	vuArReducDialogClass override for constructor
*
**********************************************************************/

void vuArReducDialogClass::LoadInit (vresource res)
{
    methodStrs = NULL;
    methodAbrStrs = NULL;
    rateStrs = NULL;
    rateAbrStrs = NULL;
    rateValStrs = NULL;
    numMethodStrs = 0;
    numMethodAbrStrs = 0;
    numRateStrs = 0;
    numRateAbrStrs = 0;
    numRateValStrs = 0;

    method = vuArReducMethodsClass::RequestSubLoad (this,
					vunames_vuArReducMethod);
    rate = vuArReducRatesClass::RequestSubLoad (this,
					vunames_vuArReducRate);
    archive = checkBoxClass::RequestSubLoad (this,
					vunames_vuArReducArchive);
    valuesList = vuArReducValuesListClass::RequestSubLoad (this,
					vunames_vuArReducAssigned);
    fieldsList = vuArReducFieldsListClass::RequestSubLoad (this,
					vunames_vuArReducFields);
    vcontainer *buttoncontainer = vcontainer::RequestSubLoad (this,
					vunames_vuArReducButtonContainer);
    applyButton = vuArReducApplyButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuArReducApplyButton);
    dismissButton = vuArReducDismissButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuArReducDismissButton);
    clearButton = vuArReducClearButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuArReducClearButton);
    vdialog::LoadInit (res);

    method -> SetUserPtr ((void *)this);
    rate -> SetUserPtr ((void *)this);
    valuesList -> storeArReducDialog (this);
    fieldsList -> storeArReducDialog (this);
    applyButton -> storeArReducDialog (this);
    dismissButton -> storeArReducDialog (this);
    clearButton -> storeArReducDialog (this);
	//
	// handle readonly state for the apply and delete buttons
	//
	if(vuReadOnly == vTRUE)
	{
		applyButton->Disable();
		clearButton->Disable();
	}
}


/*********************************************************************
*
*	vuArReducDialogClass window notify function
*
*********************************************************************/

void vuArReducDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*********************************************************************
*
*	vuArReducDialogClass destroy function
*
*********************************************************************/

void vuArReducDialogClass::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
	vevent::StopProcessing ();
    }
    // destroy values list
    vliststr *values = (vliststr *)valuesList -> GetList ();
    if (values != NULL)
    {
	values -> Destroy ();
    }
    // destroy fields list
    vliststr *fields = (vliststr *)fieldsList -> GetList ();
    if (fields != NULL)
    {
	fields -> Destroy ();
    }

    vstrDestroy (dialogTitle);

    size_t i;
    if (methodStrs != NULL)
    {
	for (i = 0; i < numMethodStrs; i++)
	{
	    vstrDestroy (methodStrs[i]);
	}
	vmemFree ((void *)methodStrs);
    }
    if (methodAbrStrs != NULL)
    {
	for (i = 0; i < numMethodAbrStrs; i++)
	{
	    vstrDestroy (methodAbrStrs[i]);
	}
	vmemFree ((void*)methodAbrStrs);
    }
    if (rateStrs != NULL)
    {
	for (i = 0; i < numRateStrs; i++)
	{
	    vstrDestroy (rateStrs[i]);
	}
	vmemFree ((void *)rateStrs);
    }
    if (rateAbrStrs != NULL)
    {
	for (i = 0; i < numRateAbrStrs; i++)
	{
	    vstrDestroy (rateAbrStrs[i]);
	}
	vmemFree ((void *)rateAbrStrs);
    }
    if (rateValStrs != NULL)
    {
	for (i = 0; i < numRateValStrs; i++)
	{
	    vstrDestroy (rateValStrs[i]);
	}
	vmemFree ((void *)rateValStrs);
    }

    vdialog::Destroy ();
}


/********************************************************************
*
*	vuArReducDialogClass open function
*
********************************************************************/

void vuArReducDialogClass::vuArReducOpen ( vuMainDialogClass *mnDialog,
					   char *ptName,
					   int   pointType)
{
    // store for later use
    mainDialog = mnDialog;

    vresource arrayDict = mainDialog -> getArraysResource ();
    methodsRes = vresourceGet (arrayDict, vunames_vuArReducMeth);
    methodsAbrRes = vresourceGet (arrayDict, vunames_vuArReducMethAbrev);
    rateStrRes = vresourceGet (arrayDict, vunames_vuArReducRateStr);
    rateAbrRes = vresourceGet (arrayDict, vunames_vuArReducRateAbr);
    rateValRes = vresourceGet (arrayDict, vunames_vuArReducRateVals);

    numMethodStrs    = vresourceCountComponents (methodsRes);
    numMethodAbrStrs = vresourceCountComponents (methodsAbrRes);
    numRateStrs      = vresourceCountComponents (rateStrRes);
    numRateAbrStrs   = vresourceCountComponents (rateAbrRes);
    numRateValStrs   = vresourceCountComponents (rateValRes);

    methodStrs    = (vstr **)(vmemAllocAndClear ((sizeof (vstr *)) * numMethodStrs));
    methodAbrStrs = (vstr **)(vmemAllocAndClear ((sizeof (vstr *)) * numMethodAbrStrs));
    rateStrs      = (vstr **)(vmemAllocAndClear ((sizeof (vstr *)) * numRateStrs));
    rateAbrStrs   = (vstr **)(vmemAllocAndClear ((sizeof (vstr *)) * numRateAbrStrs));
    rateValStrs   = (vstr **)(vmemAllocAndClear ((sizeof (vstr *)) * numRateValStrs));

    size_t i;
    for (i = 0; i < numMethodStrs; i++)
    {
	methodStrs[i] = vresourceGetStringAt (methodsRes, i);
    }
    for (i = 0; i < numMethodAbrStrs; i++)
    {
	methodAbrStrs[i] = vresourceGetStringAt (methodsAbrRes, i);
    }
    for (i = 0; i < numRateStrs; i++)
    {
	rateStrs[i] = vresourceGetStringAt (rateStrRes, i);
    }
    for (i = 0; i < numRateAbrStrs; i++)
    {
	rateAbrStrs[i] = vresourceGetStringAt (rateAbrRes, i);
    }
    for (i = 0; i < numRateValStrs; i++)
    {
	rateValStrs[i] = vresourceGetStringAt (rateValRes, i);
    }

    method -> buildMenu ();
    rate -> buildMenu ();

    vstr *ptname = vstrCloneScribed (vcharScribeSystem (ptName));
    // save pointname for later use
    vcharCopyBounded (ptname, pointName, PVNAME_SZ);
    pointName[PVNAME_SZ] = 0;

    vresource dialogTitles = mainDialog -> getTitlesResource();
    vstr *arReducTitle = vresourceGetString (dialogTitles, 
				vunames_vuArReducTitle);
    dialogTitle = vstrCloneScribed (vcharScribeFormat (arReducTitle, ptname));

    vstrDestroy (arReducTitle);
    vstrDestroy (ptname);

    SetTitle (dialogTitle);

    if ((fieldsList -> buildFieldsList (pointType)) == vFALSE)
	return;

    // map to database fields
    dataVuApplication *dataVuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = dataVuApp -> getDatabaseClass ();
    arRedChan = DB_MapField (db -> vuArReducTable, FLD_ARCH_CHAN);
    arRedRate = DB_MapField (db -> vuArReducTable, FLD_ARCH_RATE);
    arRedMethod = DB_MapField (db -> vuArReducTable, FLD_ARCH_METHOD);
    arRedArchive = DB_MapField (db -> vuArReducTable, FLD_ARCH_ARCHIVE);
    pvmapTrend = DB_MapField (db -> vuPvmapTable, FLD_TREND);

    valuesList -> buildValuesList ();


    //open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog after dialgo is closed
    Place (vwindow::GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open ();
    MakeTopMostDialog (this, NULL);
    //done here to get the valuesList::ObserveSelect to fire and
    // cause the fieldsList,rate,method and archve to update
    vliststr *valList = (vliststr *)valuesList -> GetList ();
    if (valList != NULL)
    {
	int rowCount = valList -> GetRowCount ();
	if (rowCount > 0)
	{
	    vlistSelection *listSelection = valuesList -> GetSelection ();
	    valList -> SelectRow (listSelection, 0);
	}
    }
    vevent::Process ();
}


/********************************************************************
*
*	vuArReducDialogClass append to value list function
*
********************************************************************/

void vuArReducDialogClass::appendToValuesList (vchar *fieldStr,
					       vchar *rateStr,
					       vchar *methodStr)
{
    vliststr *list = (vliststr *)valuesList -> GetList ();
    if (list == NULL)
    {
	list = new vliststr;
	valuesList -> SetList (list);
	vlistsize *size = valuesList -> GetSizes ();
	valuesList -> SetColumnWidth (vlistviewALL_COLUMNS, size -> GetColumnWidth (0) / 3);
	vlistSelection *listSelection = valuesList -> GetSelection ();
	list -> SetSelectMethod (listSelection, vlistFREE_SELECTION);
	list -> SetExtendRowSelection (listSelection, vTRUE);
    }
    vlistSelection *selection = valuesList -> GetSelection ();
    int numRows = list -> GetRowCount ();
    int row, column;
    if (numRows > 0)
    {
	list -> GetSelectedCell (selection, &row, &column);
	list -> UnselectRow (selection, row);
    }
    list -> SetCellValueScribed (numRows, 0, vcharScribeSystem (fieldStr));
    list -> SetCellValueScribed (numRows, 1, vcharScribeSystem (rateStr));
    list -> SetCellValueScribed (numRows, 2, vcharScribeSystem (methodStr));
    list -> SelectRow (selection, numRows);
}


/********************************************************************
*
*	vuArReducDialogClass find record function
*
********************************************************************/

DB_RECORD_PTR vuArReducDialogClass::findRecord (char *pvname, 
						char *meth,
						char *field)
{
    char key[PVNAME_SZ + ARCHIVE_METH_SZ];
    memset (key, 0, sizeof (key));
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();
    DB_UseKey (db -> vuArReducTable, FLD_ARCH_PTNAME_METH_KEY);
    DB_Top (db -> vuArReducTable);

    if ( ((strlen (field) == 0)) || ((strncmp (field, "VAL", 3)) == 0) )
    {
	sprintf (key, "%-*.*s%*.*s",PVNAME_SZ, PVNAME_SZ, pvname,
				    ARCHIVE_METH_SZ, ARCHIVE_METH_SZ, meth);
    }
    else
    {
	// try with appended field name
	char pvFldName[PVNAME_SZ + 1];
	sprintf (pvFldName, "%s.%s\0", pvname, field);
	sprintf (key, "%-*.*s%*.*s",PVNAME_SZ, PVNAME_SZ, pvFldName,
				    ARCHIVE_METH_SZ, ARCHIVE_METH_SZ, meth);
    }
    DB_RECORD_PTR rec = DB_Seek (db -> vuArReducTable, key);
    return (rec);
}


/********************************************************************
*
*	vuArReducDialogClass update record function
*
********************************************************************/

void vuArReducDialogClass::updateRecord (DB_RECORD_PTR rec,
						  char *pvName, 
						  char *fldName,
						  char *rateStr,
						  char *methodStr,
						  char *archiveVal)
{
    char pv[PVNAME_SZ + 1];

    if ((strlen (fldName)) > 0)
    {
	sprintf (pv, "%s.%s\0", pvName, fldName);
    }
    else
    {
	strcpy (pv, pvName);
    }
    DB_PutFieldValue (rec, arRedChan, pv);
    DB_PutFieldValue (rec, arRedRate, rateStr);
    DB_PutFieldValue (rec, arRedMethod, methodStr);
    DB_PutFieldValue (rec, arRedArchive, archiveVal);
}


/********************************************************************
*
*	vuArReducDialogClass update trend enable
*	find in pvmap and set trend flag
*	also, send trend flag via channel access to the IOC and
*	turn off the download flag
*
********************************************************************/

void vuArReducDialogClass::updateTrendEnable (char *trendValue)
{
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();
    char pvName[PVNAME_SZ + 1];
    char pv[PVNAME_SZ + 1];
    DB_RECORD_PTR rec;

    vcharExportBounded (vcharScribeSystemExport (pointName), pv, PVNAME_SZ);
    pv[PVNAME_SZ] = 0;
    sprintf (pvName, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pv);
    if ((rec = DB_Seek(db -> vuPvmapTable, pv)) != 0)
    {
	VERIFY_INFO	verifyInfo;
	DB_PutFieldValue (rec, pvmapTrend, trendValue);
	int val = (int)(*trendValue);
	vuCaPutValue (vuApp, pv, "TENB", val);
	if ((DB_GetVerifyInfo (db -> vuPvmapTable, pvName, &verifyInfo, 0)) != 0)
	{
	    verifyInfo.viPVStatus = 0;
	    DB_PutVerifyInfo (db -> vuPvmapTable, pvName, &verifyInfo, 0);
	}
    }

}


/********************************************************************
*
*	vuArReducDialogClass update record in  value list function
*
********************************************************************/

void vuArReducDialogClass::updateValuesList (vchar *fieldStr,
					     vchar *rateStr,
					     vchar *methodStr)
{
    int numRows = 0;
    int foundIt = vFALSE;

    vliststr *list = (vliststr *)valuesList -> GetList ();
    if (list == NULL)
    {
	list = new vliststr;
	valuesList -> SetList (list);
	vlistsize *size = valuesList -> GetSizes ();
	valuesList -> SetColumnWidth (vlistviewALL_COLUMNS, size -> GetColumnWidth (0) / 3);
	vlistSelection *listSelection = valuesList -> GetSelection ();
	list -> SetSelectMethod (listSelection, vlistFREE_SELECTION);
	list -> SetExtendRowSelection (listSelection, vTRUE);
	foundIt = vTRUE;
    }
    else
    {
	numRows = list -> GetRowCount ();
	int i;
	vchar *fieldVal, *methodVal;
	for (i = 0; i < numRows; i++)
	{
	    fieldVal = (vchar *)(list -> GetCellValue (i, 0));
	    methodVal = (vchar *)(list -> GetCellValue (i, 2));
	    if ( ((vcharCompare (fieldStr, fieldVal) == 0)) &&
		 ((vcharCompare (methodStr, methodVal)) == 0) )
	    {
		numRows = i;
		foundIt = vTRUE;
		break;
	    }
	}
    }
    if (foundIt == vTRUE)
    {
	vlistSelection *selection = valuesList -> GetSelection ();
	list -> SetCellValueScribed (numRows, 0, vcharScribeSystem (fieldStr));
	list -> SetCellValueScribed (numRows, 1, vcharScribeSystem (rateStr));
	list -> SetCellValueScribed (numRows, 2, vcharScribeSystem (methodStr));
	list -> SelectRow (selection, numRows);
    }
}


/*******************************************************************
*
*	ar reduc methods class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuArReducMethodsClass, optionMenuClass,
					"vuArReducMethodsClass");
/********************************************************************
*
*	vuArReducMethodsClass build methods menu function
*
********************************************************************/

void vuArReducMethodsClass::buildMenu (void)
{
    vuArReducDialogClass *arReducDialog = (vuArReducDialogClass *)GetUserPtr ();
    vstr **strPtrs = arReducDialog -> getMethodStrs ();
    int numMethodStrs = arReducDialog -> getNumMethodStrs ();

    /* create a new menu to attach to the vcontrol item, we are 
       creating a popup menu within a dialog
    */
    vmenu *newMenu = new vmenu;
    vmenuItem *menuItem;
    int i;

    for (i = 0; i < numMethodStrs; i++)
    {
	menuItem = new vmenuItem;
	menuItem -> SetTitleScribed (vcharScribeSystem (strPtrs[i]));
	newMenu -> AppendItem (menuItem);
    }
    // attach the menu to the control
    this -> SetMenu (newMenu);
    // set min and max for the control
    this -> SetMin (0);
    this -> SetMax (numMethodStrs - 1);
    // set value so current menu item will be the 1st one
    this -> SetValue (0);
}


/*******************************************************************
*
*	ar reduc rates class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuArReducRatesClass, optionMenuClass,
					"vuArReducRatesClass");
/********************************************************************
*
*	vuArReducRatesClass build methods menu function
*
********************************************************************/

void vuArReducRatesClass::buildMenu (void)
{
    vuArReducDialogClass *arReducDialog = (vuArReducDialogClass *)GetUserPtr ();
    int numRateStrs = arReducDialog -> getNumRateStrs ();
    vstr **rateStrs = arReducDialog -> getRateStrs ();

    /* create a new menu to attach to the vcontrol item, we are 
       creating a popup menu within a dialog
    */
    vmenu *newMenu = new vmenu;
    vmenuItem *menuItem;
    int i;
    int oneMinuteRate = -1;
    vstr *str;

    for (i = 0; i < numRateStrs; i++)
    {
	menuItem = new vmenuItem;
	str = vstrCloneScribed (vcharScribeSystem (rateStrs[i]));
	menuItem -> SetTitle (str);
	if ((vcharSearch (str, (const vchar *)("1 minute"))) != NULL)
	{
	    oneMinuteRate = i;
	}
	vstrDestroy (str);
	newMenu -> AppendItem (menuItem);
    }
    // attach the menu to the control
    this -> SetMenu (newMenu);
    // set min and max for the control
    this -> SetMin (0);
    this -> SetMax (numRateStrs - 1);
    // set value so current menu item will be the 1st one
    if (oneMinuteRate > -1)
    {
	this -> SetValue (oneMinuteRate);
    }
    else
    {
        this -> SetValue (0);
    }
}


/*******************************************************************
*
*	ar reduc values list class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuArReducValuesListClass, vlistview,
					"vuArReducValuesListClass");

/******************************************************************
*
*	ar reduc values list class observe select
*
******************************************************************/

void vuArReducValuesListClass::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	vlistSelection *listSelection = GetSelection ();
	vliststr *valList = (vliststr *)GetList ();
	int row, column;
	valList -> GetSelectedCell (listSelection, &row, &column);
	if (row != vlistNO_CELLS_SELECTED)
	{
	    vchar *fld, *rt, *meth;
	    fld  = (vchar *)(valList -> GetCellValue (row, 0));
	    rt   = (vchar *)(valList -> GetCellValue (row, 1));
	    meth = (vchar *)(valList -> GetCellValue (row, 2));

	    // updated trend fields with new value
	    vliststr *fldList = (vliststr *)(arReducDialog -> fieldsList -> GetList ());
	    listSelection = arReducDialog -> fieldsList -> GetSelection ();
	    int selectedFieldsRow;
	    fldList -> GetSelectedCell (listSelection, &selectedFieldsRow, &column);
	    const vchar *cellValue;
	    int foundMatch = vFALSE;
	    vlistIterator iterator;
	    iterator.Start (fldList);
	    while (iterator.Next())
	    {
		row = iterator.GetRow ();
		cellValue = (const vchar *)(iterator.GetValue ());
		if ((vcharCompare (cellValue, fld)) == 0)
		{
		    foundMatch = vTRUE;
		    break;
		}
	    }
	    iterator.Finish ();
	    if (foundMatch == vTRUE)
	    {
		if (selectedFieldsRow != vlistNO_CELLS_SELECTED)
		{
		    fldList -> UnselectRow (listSelection, selectedFieldsRow);
		}
		fldList -> SelectRow (listSelection, row);
	    }

	    // find the rate string and set the rate menu to this value

	    int minVal, maxVal;
	    minVal = arReducDialog -> rate -> GetMin ();
	    maxVal = arReducDialog -> rate -> GetMax ();
	    vmenu *menu = arReducDialog -> rate -> GetMenu ();
	    vmenuItem *menuItem;
	    const vchar *menuItemTitle;
	    int i;
	    vstr **strPtrs;

	    for (i = minVal; i <= maxVal; i++)
	    {
		menuItem = menu -> GetItemAt (i);
		menuItemTitle = menuItem -> GetTitle ();
		if ( (vcharCompareBounded (rt, menuItemTitle, 6)) == 0)
		{
		    if (i < (arReducDialog -> getNumRateStrs ()))
		    {
			strPtrs = arReducDialog -> getRateStrs ();
			arReducDialog -> rate -> SetOptionMenuValue (strPtrs[i]);
		    }
		    else
		    {
			arReducDialog -> rate -> SetOptionMenuValue (rt);
		    }
		    break;
		}
	    }

	    int numAbrMethods = arReducDialog -> getNumMethodAbrStrs ();
	    int  foundMethod = vFALSE;
	    strPtrs = arReducDialog -> getMethodAbrStrs ();

	    for (i = 0; i < numAbrMethods; i++)
	    {
		if ((vcharCompareBounded (strPtrs[i], meth, ARCHIVE_METH_SZ)) == 0)
		{
		    foundMatch = vTRUE;
		    break;
		}
	    }
	    if (foundMatch == vTRUE)
	    {
		strPtrs = arReducDialog -> getMethodStrs ();
		arReducDialog -> method -> SetOptionMenuValue (strPtrs[i]);
	    }
	    else
	    {
		arReducDialog -> method -> SetOptionMenuValue (meth);
	    }

	    // set archive value based on per method value
	    DB_RECORD_PTR rec;
	    char pvName[PVNAME_SZ + 1];
	    char methodStr[ARCHIVE_METH_SZ + 1];
	    char fldStr[FLDNAME_SZ + 1];

	    vcharExportBounded (vcharScribeSystemExport (arReducDialog -> getPointName()),
				pvName,
				PVNAME_SZ);
	    pvName[PVNAME_SZ] = 0;
	    vcharExportBounded (vcharScribeSystemExport (meth), methodStr, ARCHIVE_METH_SZ);
	    methodStr[ARCHIVE_METH_SZ] = 0;
	    vcharExportBounded (vcharScribeSystemExport (fld), fldStr, FLDNAME_SZ);
	    fldStr[FLDNAME_SZ] = 0;
	    rec = arReducDialog -> findRecord (pvName, methodStr, fldStr);
	    
	    vuMainDialogClass *mainDialog = arReducDialog -> getMainDialog ();
	    dataVuApplication *vuApp = mainDialog -> getApplication ();
	    vuDatabaseClass *db = vuApp -> getDatabaseClass ();
	    int type;
	    char archiveType;

	    DB_GetFieldValue (rec, arReducDialog -> arRedArchive, &archiveType, &type);
	    arReducDialog -> archive -> SetValue ((int)archiveType);
	}
    }
}


/*******************************************************************
*
*	vuArReducValuesList build function
*
*******************************************************************/

void vuArReducValuesListClass::buildValuesList (void)
{
    vliststr *list = (vliststr *)(GetList ());

    // destroy old list if there is one
    if (list != NULL)
    {
	SetList (NULL);
	list -> Destroy ();
    }
    // new list with 3 columns
    list = new vliststr;

    vuMainDialogClass *mainDialog = arReducDialog -> getMainDialog ();
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();

    char	pvName[PVNAME_SZ +1];
    char	fldName[FLDNAME_SZ + 1];
    char	rate[ARCHIVE_RATE_SZ + 1];
    char	method[ARCHIVE_METH_SZ + 1];
    char	*tk1, *tk2;
    int		dataType;
    int		row = 0;

    vchar	*currPvName = arReducDialog -> getPointName ();
    char	currPv[PVNAME_SZ + 1];
    vcharExportBounded (vcharScribeSystemExport (currPvName), currPv, PVNAME_SZ);
    currPv[PVNAME_SZ] = 0;

    DB_UseKey (db -> vuArReducTable, FLD_ARCH_PTNAME_KEY);
    DB_RECORD_PTR rec = DB_Top (db -> vuArReducTable);
    while (rec)
    {
	DB_GetFieldValue (rec, arReducDialog -> arRedChan, pvName, &dataType);
	// check for blank records
	if (strlen (pvName) > 0)
	{
	    // split name, into name and field names
	    tk1 = strtok (pvName, ".");
	    tk2 = strtok (NULL, ".");
	    fldName[0] = 0;
	    if (tk2)
	    {
		strncpy (fldName, tk2, FLDNAME_SZ);
	    }
	    else
	    {
		strcpy (fldName, "VAL ");
	    }
	    fldName[FLDNAME_SZ] = 0;
	    // have we found a match for our point
	    if ((strcmp (currPv, pvName)) == 0)
	    {
		DB_GetFieldValue (rec, arReducDialog -> arRedRate, rate, &dataType);
		DB_GetFieldValue (rec, arReducDialog -> arRedMethod, method, &dataType);
		// make rate pretty, convert from seconds to seconds or minutes
		vstr *rateStr = vstrCloneScribed (vcharScribeSystem (rate));
		if (isdigit (rate[0]))
		{
		    int numValRates = arReducDialog -> getNumRateValStrs ();
		    int i;
		    vstr **strPtrs = arReducDialog -> getRateValStrs ();
		    for (i = 0; i < numValRates; i++)
		    {
			if ((vcharCompare (strPtrs[i], rateStr)) == 0)
			{
			    vstrDestroy (rateStr);
			    strPtrs = arReducDialog -> getRateAbrStrs ();
			    rateStr = vstrClone (strPtrs[i]);
			    break;
			}
		    }
		}
		list -> SetCellValueScribed (row, 0, vcharScribeSystem (fldName));
		list -> SetCellValueScribed (row, 1, vcharScribeSystem (rateStr));
		list -> SetCellValueScribed (row, 2, vcharScribeSystem (method));
		if (rateStr != NULL)
		{
		    vstrDestroy (rateStr);
		}
		row++;
	    }
	}
	rec = DB_SeekNext (db -> vuArReducTable);
    }
    if (row > 0)
    {
	// tell listview about new list
	SetList (list);
	vlistsize *size = GetSizes ();
	SetColumnWidth (vlistviewALL_COLUMNS, size -> GetColumnWidth (0) / 3);
	vlistSelection *listSelection = GetSelection ();
	list -> SetSelectMethod (listSelection, vlistFREE_SELECTION);
	list -> SetExtendRowSelection (listSelection, vTRUE);
	if(vuReadOnly == vFALSE)
		arReducDialog->clearButton->Enable();
    }
    else
    {
	list -> Destroy ();
	arReducDialog->clearButton->Disable();
    }
}


/******************************************************************
*
*	ar reduc fields list class
*
******************************************************************/

vkindTYPED_FULL_DEFINITION(vuArReducFieldsListClass, vlistview,
					"vuArReducFieldsListClass")


/******************************************************************
*
*	ar reduc fields list class, build list function
*
******************************************************************/

int vuArReducFieldsListClass::buildFieldsList (int pttype)
{
    vchar **arr;

    switch (pttype)
    {
	case VU_AI:
	case VU_AO:
	case VU_BO:
	case VU_CALC:
	case VU_MBBI:
	case VU_MBBO:
	case VU_LEADLAG:
	case VU_RAMP:
	case VU_PIPE:
	case VU_TANK:
	case VU_VALVE:
	case VU_POT:
	case VU_PTO:
	case VU_TD:
	case VU_SIGCHAR:
	case VU_LDFW1:
	case VU_LDFW2:
	case VU_DDC:
	    arr	= valFields;
	    break;
	case VU_BI:
	    arr	= biFields;
	    break;
	case VU_FANOUT:
	    arr = fanFields;
	    break;
	case VU_PID:
	    arr = pidFields;
	    break;
	case VU_SEL:
	    arr = selFields;
	    break;
	case VU_ACCUM:
	    arr = accumFields;
	    break;
	case VU_DEADTIME:
	    arr = deadtimeFields;
	    break;
	case VU_UDB:
	    arr = udbFields;
	    break;
	case VU_PAT:
	    arr = patFields;
	    break;
	case VU_PUMP:
	    arr = pumpFields;
	    break;
	default:
	{
	    vchar *ptName = arReducDialog -> getPointName ();
	    vuMessageDialogCenter (NULL, VU_NOT_TREND_PT_MSG, ptName);
	    return (vFALSE);
	}
    }
    int i = 0;
    int row = 0;
    vliststr *newlist = new vliststr;
    while (arr[i])
    {
	newlist -> SetCellValueScribed (row, 0, vcharScribeSystem (arr[i]));
	i++;
	row++;
    }
    SetList (newlist);
    vlistSelection *listSelection = GetSelection ();
    newlist -> SetSelectMethod (listSelection, vlistONE_ONLY_SELECTION);
    newlist -> SelectRow (listSelection, 0);
    return (vTRUE);
}


/*****************************************************************
*
*	ar reduc apply button class
*
*****************************************************************/

vkindTYPED_FULL_DEFINITION(vuArReducApplyButtonClass, vbutton,
					"vuArReducApplyButtonClass")

/*****************************************************************
*
*	vuArReducApplyButtonClass observer
*
*****************************************************************/

void vuArReducApplyButtonClass::ObserveDialogItem (vevent *event)
{
    char str[128];
    char methodStr[ARCHIVE_METH_SZ + 1];
    char rateStr[ARCHIVE_RATE_SZ + 1];
    char pvName[PVNAME_SZ + 1];
    int  numStrings;
    int  i;
    int  foundMatch;
    vstr **strs;
    vchar methAbr[ARCHIVE_METH_SZ + 1];

    // get the reduction method
    const vchar *method = arReducDialog -> method -> GetOptionMenuValue ();
    strs = arReducDialog -> getMethodStrs ();
    numStrings = arReducDialog -> getNumMethodStrs ();
    for (i = 0; i < numStrings; i++)
    {
	if ((vcharCompare (method, strs[i])) == 0)
	{
	    strs = arReducDialog -> getMethodAbrStrs ();
	    vcharCopyBounded (strs[i], methAbr, ARCHIVE_METH_SZ);
	    methAbr[ARCHIVE_METH_SZ] =0;
	    vcharExportBounded (vcharScribeSystemExport (strs[i]),
				str,
				ARCHIVE_METH_SZ);
	    str[ARCHIVE_METH_SZ] = 0;
	    sprintf (methodStr, "%-*.*s\0", ARCHIVE_METH_SZ, ARCHIVE_METH_SZ, str);
	    break;
	}
    }

    // get the reduction rate
    const vchar *rate = arReducDialog -> rate -> GetOptionMenuValue ();
    vchar rateAbr[ARCHIVE_RATE_SZ * 2 + 1];
    strs = arReducDialog -> getRateStrs ();
    numStrings = arReducDialog -> getNumRateStrs ();
    foundMatch = vFALSE;
    for (i = 0; i < numStrings; i++)
    {
	if ((vcharCompare (rate, strs[i])) == 0)
	{
	    foundMatch = vTRUE;
	    break;
	}
    }
    if (foundMatch == vTRUE)
    {
	strs = arReducDialog -> getRateValStrs ();
	vcharExportBounded (vcharScribeSystemExport (strs[i]),
			    str,
			    ARCHIVE_RATE_SZ);
	str[ARCHIVE_RATE_SZ] = 0;
	sprintf (rateStr, "%-*.*s\0", ARCHIVE_RATE_SZ, ARCHIVE_RATE_SZ, str);
	strs = arReducDialog -> getRateAbrStrs ();
	vcharCopyBounded (strs[i], rateAbr, ARCHIVE_RATE_SZ * 2);
	rateAbr[ARCHIVE_RATE_SZ * 2] = 0;
    }
    else
    {
	vcharExportBounded (vcharScribeSystemExport (rate),
			    str,
			    ARCHIVE_RATE_SZ);
	sprintf (rateStr, "%-*.*s\0", ARCHIVE_RATE_SZ, ARCHIVE_RATE_SZ, str);
	vcharCopyBounded (rate, rateAbr, ARCHIVE_RATE_SZ * 2);
	rateAbr[ARCHIVE_RATE_SZ * 2] = 0;
    }

    // get the selected field
    vliststr *fieldList = (vliststr *)arReducDialog -> fieldsList -> GetList ();
    vlistSelection *selection = arReducDialog -> fieldsList -> GetSelection ();
    int row, column;
    fieldList -> GetSelectedCell (selection, &row, &column);
    vchar *field = (vchar *)fieldList -> GetCellValue (row, 0);
    char fieldName[FLDNAME_SZ + 1];
    if ((vcharCompareBounded ((const vchar *)field, "VAL ", FLDNAME_SZ)) == 0)
    {
	strcpy (fieldName, "");
    }
    else
    {
	vcharExportBounded (vcharScribeSystemExport (field),
			    str,
			    FLDNAME_SZ);
	sprintf (fieldName, "%-*.*s\0", FLDNAME_SZ, FLDNAME_SZ, str);
    }

    // get the archive flag
    int archive = arReducDialog -> archive -> GetValue ();
    char archiveVal;
    if (archive == 1)
	archiveVal = 1;
    else 
	archiveVal = 0;

    // retrieve the pointname
    vcharExportBounded (vcharScribeSystemExport (arReducDialog -> getPointName()),
			str,
			PVNAME_SZ);
    str[PVNAME_SZ]= 0;
    strcpy (pvName, str);
    pvName[PVNAME_SZ] = 0;

    int result;
    vuMainDialogClass *mainDialog = arReducDialog -> getMainDialog ();
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();
    DB_RECORD_PTR rec = arReducDialog -> findRecord (pvName, methodStr, fieldName);
    char trendValue = 1;
    // a new record
    if (rec == 0)
    {
	result = vuMessageDialogCenter (arReducDialog, 
					VU_ARCH_NEW_REQ_MSG, 
					method,
					arReducDialog -> getPointName ());
	if (result == VU_TRUE)
	{
	    rec = DB_NewRecord (db -> vuArReducTable, NULL);
	    arReducDialog -> updateRecord (rec, pvName, fieldName, 
					   rateStr, methodStr, &archiveVal);
	    // set ON TREND
	    arReducDialog -> updateTrendEnable (&trendValue);

	    arReducDialog -> appendToValuesList (field, rateAbr, methAbr);
		if(vuReadOnly == vFALSE)
		    arReducDialog->clearButton->Enable();

	    mainDialog -> vuMainMessageCenter (vunames_vuArReducNewMsg,
					       arReducDialog -> getPointName (),
					       rate,
					       method);
	}
    }
    // update existing record
    else
    {
	result = vuMessageDialogCenter (arReducDialog, 
					VU_ARCH_UPDATE_REQ_MSG, 
					method,
					arReducDialog -> getPointName ());
	if (result == VU_TRUE)
	{
	    arReducDialog -> updateRecord (rec, pvName, fieldName, 
					   rateStr, methodStr, &archiveVal);
	    // set ON TREND
	    arReducDialog -> updateTrendEnable (&trendValue);

	    arReducDialog -> updateValuesList (field, rateAbr, methAbr);

	    mainDialog -> vuMainMessageCenter (vunames_vuArReducUpdateMsg,
					       arReducDialog -> getPointName (),
					       rate,
					       method);
	}
    }
}


/*****************************************************************
*
*	ar reduc dismiss button class
*
*****************************************************************/

vkindTYPED_FULL_DEFINITION(vuArReducDismissButtonClass, vbutton,
					"vuArReducDismissButtonClass")

/*****************************************************************
*
*	vuArReducDismissButtonClass observer
*
*****************************************************************/

void vuArReducDismissButtonClass::ObserveDialogItem (vevent *event)
{
    arReducDialog -> Close ();
    vevent::StopProcessing ();
}



/*****************************************************************
*
*	ar reduc clear button class
*
*****************************************************************/

vkindTYPED_FULL_DEFINITION(vuArReducClearButtonClass, vbutton,
					"vuArReducClearButtonClass")

/*****************************************************************
*
*	vuArReducClearButtonClass observer
*
*****************************************************************/

void vuArReducClearButtonClass::ObserveDialogItem (vevent *event)
{
    vliststr *list = (vliststr *)(arReducDialog -> valuesList -> GetList ());
    if (list != NULL)
    {
	int numRows = list -> GetRowCount ();
	if (numRows == 0)
	{
	    return;
	}
	int result = vuMessageDialogCenter (arReducDialog, VU_ARCH_DELETE_REQ_MSG);
	if (result == VU_FALSE)
	{
	    return;
	}
	vlistSelection *selection = arReducDialog -> valuesList -> GetSelection ();
	vlistIterator iterator;
	int row;
	vchar *fld, *meth;
	char  fieldVal[FLDNAME_SZ + 1];
	char  methodVal[ARCHIVE_METH_SZ + 1];
	char  pvName[PVNAME_SZ + 1];
	DB_RECORD_PTR rec;
	vuMainDialogClass *mainDialog = arReducDialog -> getMainDialog ();
	dataVuApplication *vuApp = mainDialog -> getApplication ();
	vuDatabaseClass *db = vuApp -> getDatabaseClass ();


	vcharExportBounded (vcharScribeSystemExport (arReducDialog -> getPointName()),
			    pvName, PVNAME_SZ);
	pvName[PVNAME_SZ] = 0;

	// traverse all selected items and delete from list and database
	iterator.StartWithSelection (selection);
	int lastRow = -1;
	while (iterator.Next ())
	{
	    row = iterator.GetRow ();
	    if(lastRow == row) continue;
	    lastRow = row;
	    fld  = (vchar *)(list -> GetCellValue (row, 0));
	    meth = (vchar *)(list -> GetCellValue (row, 2));
	    vcharExportBounded (vcharScribeSystemExport (fld),
				fieldVal, FLDNAME_SZ);
	    fieldVal[FLDNAME_SZ] = 0;
	    vcharExportBounded (vcharScribeSystemExport (meth),
				methodVal, ARCHIVE_METH_SZ);
	    methodVal[ARCHIVE_METH_SZ] = 0;
	    if ( (rec = arReducDialog -> findRecord (pvName, methodVal, fieldVal)) != 0)
	    {
		DB_DeleteRecord (db -> vuArReducTable, rec);
	    }
	    //list -> RemoveRow(row);
	}
	iterator.Finish();
	numRows = list -> GetRowCount();
	int i = 0;
	while(numRows)
	{
	    if(list->IsCellSelected(selection,i,0))
	    {
		list->RemoveRow(i);
		numRows--;
	    }
	    else
	    {
		if(++i > numRows) break;
	    }
	}
	mainDialog->vuMainMessageCenter(vunames_vuArReducDeleteMsg,arReducDialog->getPointName ());

	// if no more records in ar_reduc, then turn of trend flag in pvmap
	// we make the assumption that if there are no more rows in the list
	// then there should be no more records in ar_reduc
	numRows = list -> GetRowCount ();
	if (numRows == 0)
	{
	    char key[PVNAME_SZ + 1];
	    sprintf (key, "%-*.*s", PVNAME_SZ, PVNAME_SZ, pvName);
	    if ( (rec = DB_Seek (db -> vuPvmapTable, key)) != 0)
	    {
		char trendValue = 0;
		DB_PutFieldValue (rec, arReducDialog -> pvmapTrend, &trendValue);
		arReducDialog -> updateTrendEnable (&trendValue);
		mainDialog -> vuMainMessageCenter (vunames_vuArReducAllDeleteMsg,
						   arReducDialog -> getPointName ());
		Disable();
	    }
	}
	else
	{
	    list -> SelectRow (selection, 0);
	}
    }

}



/***********************************************************************
*
*	trend edit dialog class
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION (vuTrendEditDialogClass, vdialog, 
					"vuTrendEditDialogClass");

/**********************************************************************
*
*	trend edit dialog class override for constructor
*
**********************************************************************/

void vuTrendEditDialogClass::LoadInit (vresource res)
{
    pointName = filteredTextitem::RequestSubLoad (this,
					vunames_vuTrendEditPvName);
    okButton = vuTrendEditOkButtonClass::RequestSubLoad (this,
					vunames_vuTrendEditOkButton);
    cancelButton = vuTrendEditCancelButtonClass::RequestSubLoad (this,
					vunames_vuTrendEditCancelButton);
    vdialog::LoadInit (res);

    pointName -> SetItemFilterFunc (pvnameFilter);
    okButton -> storeTrendEditDialog (this);
    cancelButton -> storeTrendEditDialog (this);
}


/*********************************************************************
*
*	trend edit dialog class window notify function
*
*********************************************************************/

void vuTrendEditDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	// set to nothing so caller has nothing to process
	vmemSet ((void *)pointName, 0, sizeof (pointName));
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/********************************************************************
*
*	vuTrendEditDialogClass open function
*
********************************************************************/

void vuTrendEditDialogClass::vuTrendEditOpen (vuMainDialogClass *mnDialog)
{
    // store for later use
    mainDialog = mnDialog;

    //open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog after dialgo is closed
    Place (vwindow::GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open ();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*****************************************************************
*
*	trend edit ok button class
*
*****************************************************************/

vkindTYPED_FULL_DEFINITION(vuTrendEditOkButtonClass, vbutton,
					"vuTrendEditOkButtonClass")

/*****************************************************************
*
*	trend edit ok button class observer
*
*****************************************************************/

void vuTrendEditOkButtonClass::ObserveDialogItem (vevent *event)
{
    vstr *ptName = trendEditDialog -> pointName -> GetTextAsString ();
    if ((vcharLength (ptName)) == 0)
    {
	// no point name entered error msg
	vuMessageDialogCenter (trendEditDialog, VU_NO_PTNAME_ENTERED_MSG);
    }
    else
    {
	vchar *pvName = trendEditDialog -> getPvName ();
	vcharCopyBounded (ptName, pvName, PVNAME_SZ);
	pvName[PVNAME_SZ] = 0;
	char pointName[PVNAME_SZ + 1];
	char pv[PVNAME_SZ + 1];
	vcharExportBounded (vcharScribeSystemExport (pvName),
			    pointName,
			    PVNAME_SZ);

	vuMainDialogClass *mainDialog = trendEditDialog -> getMainDialog ();
	dataVuApplication *dataVuApp = mainDialog -> getApplication ();
	vuDatabaseClass *db = dataVuApp -> getDatabaseClass ();
	DB_RECORD_PTR	pntRecord;
	sprintf (pv, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pointName);
	if ((pntRecord = DB_Seek (db -> vuPvmapTable, pv)) == 0)
	{
	    vuMessageDialogCenter (trendEditDialog, VU_PTNOTINPVMAPERR, pvName);
	    vmemSet ((void *)pvName, 0, PVNAME_SZ);
	    pvName[PVNAME_SZ] = 0;
	}
	else
	{
	    trendEditDialog -> Close ();
	    vevent::StopProcessing ();
	}
    }
    vstrDestroy (ptName);
}


/*****************************************************************
*
*	trend edit cancel button class
*
*****************************************************************/

vkindTYPED_FULL_DEFINITION(vuTrendEditCancelButtonClass, vbutton,
					"vuTrendEditCancelButtonClass")

/*****************************************************************
*
*	trend edit cancel button class observer
*
*****************************************************************/

void vuTrendEditCancelButtonClass::ObserveDialogItem (vevent *event)
{
    // set to nothing so caller has nothing to process
    vchar *pvName = trendEditDialog -> getPvName ();
    vmemSet ((void *)pvName, 0, PVNAME_SZ);
    pvName[PVNAME_SZ] = 0;
    trendEditDialog -> Close ();
    vevent::StopProcessing ();
}
