#ifndef	vucommondialogclassHXX
#define	vucommondialogclassHXX

#include "textFilters.hxx"

// define the type of device type some points can have

typedef	enum {
	VU_REFERENCE,
	VU_CONSTANT,
	VU_HARDWARE
	} DTYP_TYPE;

class	vuCommonAlarmInfoButtonClass;
class	vuCommonApplyButtonClass;
class	vuCommonDeleteButtonClass;
class	vuCommonDialogClass;
class	vuCommonDismissButtonClass;
class	vuCommonGoToButtonClass;
class	vuCommonNextButtonClass;
class	vuCommonPreviousButtonClass;
class	vuCommonResourceClass;
class	vuCommonScanTypeClass;
class	vuCommonTrendButtonClass;
class	vuPvNameFilteredTextitem;


/** COMMON DIALOG CLASS *****************************************************

	This is the base class for the point type windows. It will
	define all the functions needed to access the common fields
	and open the point type specific containers. The point type
	specific classes will be sub-classed from this base and be
	in their own .hxx & .cxx files. The nameing convention is as
	follows:

	.hxx files ---- vuXXXDialogClass.hxx
	.cxx files -- vuXXXDialogClass.cxx

	where XXX is the point type name

	e.g.

	vuAiDialogClass.hxx
	vuAiDialogClass.cxx

	Class names will be the same as the .cxx name.

	GCT 10-Jun-1997
*/

class vuCommonDialogClass : public vdialog
{
    vuMainDialogClass	*mainDialogClass;

    // pointers to the common dialog items
    vuPvNameFilteredTextitem	*pvname;
    filteredTextitem		*descriptor;
    filteredTextitem		*location;
    filteredTextitem		*protectionGroup;
    filteredTextitem		*phase;
    filteredTextitem		*event;
    vuCommonResourceClass	*resourceName;
    vuCommonScanTypeClass	*scanType;
    vuPvNameButtonClass		*forwardLinkButton;
    vuFilteredTextitem		*forwardLink;
    vuPvNameButtonClass		*scanDisableButton;
    vuFilteredTextitem		*scanDisable;
    filteredTextitem		*disableValue;
    vuCommonApplyButtonClass	*applyButton;
    vuCommonDismissButtonClass	*dismissButton;
    vuCommonDeleteButtonClass	*deleteButton;
    vuCommonGoToButtonClass	*gotoButton;
    vuCommonNextButtonClass	*nextButton;
    vuCommonPreviousButtonClass *previousButton;
    vuCommonTrendButtonClass	*trendButton;
    vuCommonAlarmInfoButtonClass *alarmInfoButton;

protected:

    int			dialogModified;
    int			pointType;
    void		*dbValue;	// used by vuLoadDialog and vuSaveDialog

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(vuCommonDialogClass);

    // macros for standard versions of the constructors and destructors
    vloadableINLINE_CONSTRUCTORS(vuCommonDialogClass, vdialog);

    // vloadable override
    void LoadInit (vresource resource);

    // vwindow overide
    void ObserveWindow (vevent *event);

    // override 
    void Destroy ();

    // function to reset common dialog
    virtual void resetDialog (void);

    // handle mainDialog functions
    void storeMainDialog (vuMainDialogClass *mD) 
	{ mainDialogClass = mD; }
    vuMainDialogClass *getMainDialog (void) 
	{ return (mainDialogClass); }

    // handle the dialog modify flag
    void setDialogModify (int mod) 
	{ dialogModified = mod; }
    int  getDialogModify (void) 
	{ return (dialogModified); }

    // load common fields from database & update window
    virtual void vuLoadDialog (vchar *pointName);

    // Collect data from window and save to database	
    virtual void vuSaveDialog (vchar *pointName);

    // functions to handle the pvname field
    vuPvNameFilteredTextitem *getPvNameField (void) 
	{ return (pvname); }
    void vuClearPvNameField (void);

    // handle resource name
    vuCommonResourceClass *getResName (void) 
	{ return (resourceName); }

    // handle the pointType
    void setPointType (int pType) 
	{ pointType = pType; }
    int  getPointType (void) 
	{ return (pointType); }

    // functions for the scan rate
    void vuGetScanType (DB_TABLE_PTR table, 
			DB_RECORD_PTR rec, 
			const vchar *optinMenuValue,
			char *dbvalue);
    void vuSetScanType (DB_TABLE_PTR table, 
			DB_RECORD_PTR rec, 
			char *dbvalue);

    // function to handle the forwardlink fields
    vuPvNameButtonClass *getFwdLinkButton (void) 
	{ return (forwardLinkButton); }
    filteredTextitem *getFwdLink (void) 
	{ return (forwardLink); }

    // function to handle the trend button
    vuCommonTrendButtonClass *getTrendButton (void) 
	{ return (trendButton); }
    //
    // This is the area to hold the space-padded point name used in ALL
    // DB_Seek operations. This is needed to insure an exact match
    // during keyed searches
public:
    char seekBuff[PVNAME_SZ+1];
};


// *** COMMON RESOURCE CLASS  ********************************************************

class vuCommonResourceClass : public vcontrolOptionMenu
{
    vuCommonDialogClass	*commonDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(vuCommonResourceClass);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(vuCommonResourceClass, vcontrolOptionMenu);

    // override the control notify
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void storeCommonDialog (vuCommonDialogClass *cd) 
	{ commonDialog = cd; }
    vuCommonDialogClass *getCommonDialog (void) 
	{ return (commonDialog); }

    // function to build resource  menu
    void vuBuildResourceMenu (void);
};


// *** COMMON SCAN TYPE CLASS  ********************************************************

class vuCommonScanTypeClass : public vcontrolOptionMenu
{
    vuCommonDialogClass	*commonDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(vuCommonScanTypeClass);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(vuCommonScanTypeClass, vcontrolOptionMenu);

    // override the control notify
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void storeCommonDialog (vuCommonDialogClass *cd) 
	{ commonDialog = cd; }
    vuCommonDialogClass *getCommonDialog (void) 
	{ return (commonDialog); }

    // function to build Scan Type  menu
    void buildScanTypeMenu (void);
};


// *** PVNAME FILTERED TEXT ITEM subclass ********************************************

class vuPvNameFilteredTextitem : public filteredTextitem
{
public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (vuPvNameFilteredTextitem);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (vuPvNameFilteredTextitem, filteredTextitem);

    // virtual member function override
    void ObserveView (vevent *event, int message);

};


// *** COMMON DIALOG BUTTON CLASSES **************************************************

class vuCommonApplyButtonClass : public vbutton
{
    vuCommonDialogClass	*commonDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (vuCommonApplyButtonClass);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (vuCommonApplyButtonClass, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void storeCommonDialog (vuCommonDialogClass *cd) 
	{ commonDialog = cd; }
    vuCommonDialogClass *getCommonDialog (void) 
	{ return (commonDialog); }
};


class vuCommonDismissButtonClass : public vbutton
{
    vuCommonDialogClass	*commonDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (vuCommonDismissButtonClass);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (vuCommonDismissButtonClass, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void storeCommonDialog (vuCommonDialogClass *cd) 
	{ commonDialog = cd; }
    vuCommonDialogClass *getCommonDialog (void) 
	{ return (commonDialog); }
};

class vuCommonDeleteButtonClass : public vbutton
{
    vuCommonDialogClass	*commonDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (vuCommonDeleteButtonClass);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (vuCommonDeleteButtonClass, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void storeCommonDialog (vuCommonDialogClass *cd) 
	{ commonDialog = cd; }
    vuCommonDialogClass *getCommonDialog (void) 
	{ return (commonDialog); }
};


class vuCommonGoToButtonClass : public vbutton
{
    vuCommonDialogClass	*commonDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (vuCommonGoToButtonClass);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (vuCommonGoToButtonClass, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void storeCommonDialog (vuCommonDialogClass *cd) 
	{ commonDialog = cd; }
    vuCommonDialogClass *getCommonDialog (void) 
	{ return (commonDialog); }
};

class vuCommonNextButtonClass : public vbutton
{
    vuCommonDialogClass	*commonDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (vuCommonNextButtonClass);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (vuCommonNextButtonClass, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void storeCommonDialog (vuCommonDialogClass *cd) 
	{ commonDialog = cd; }
    vuCommonDialogClass *getCommonDialog (void) 
	{ return (commonDialog); }
};

class vuCommonPreviousButtonClass : public vbutton
{
    vuCommonDialogClass	*commonDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (vuCommonPreviousButtonClass);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (vuCommonPreviousButtonClass, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void storeCommonDialog (vuCommonDialogClass *cd) 
	{ commonDialog = cd; }
    vuCommonDialogClass *getCommonDialog (void) 
	{ return (commonDialog); }
};


class vuCommonTrendButtonClass : public vbutton
{
    vuCommonDialogClass	*commonDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (vuCommonTrendButtonClass);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (vuCommonTrendButtonClass, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void storeCommonDialog (vuCommonDialogClass *cd) 
	{ commonDialog = cd; }
    vuCommonDialogClass *getCommonDialog (void) 
	{ return (commonDialog); }
};

class vuCommonAlarmInfoButtonClass : public vbutton
{
    vuCommonDialogClass	*commonDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (vuCommonAlarmInfoButtonClass);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (vuCommonAlarmInfoButtonClass, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void storeCommonDialog (vuCommonDialogClass *cd) 
	{ commonDialog = cd; }
    vuCommonDialogClass *getCommonDialog (void) 
	{ return (commonDialog); }
};

#endif	vucommondialogclassHXX
