#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "choice.h"
#include "choiceGbl.h"
#include "dpacfg.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuDpaDialogClass.hxx"
#include "vuUtils.hxx"

// filter functions unique to launch config
static int vuDpaDataFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len);
static int vuDpaStringFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len);

static char *dTypeStrings [] = {
	"STRING",
	"SHORT",
	"FLOAT",
	"ENUM",
	"CHAR",
	"LONG",
	"DOUBLE",
	0 };


/*******************************************************************
*
*	DPA SELECT DPA DIALOG
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaSelectDpaDialogClass, vdialog, 
						"vuDpaSelectDpaDialogClass");

/*******************************************************************
*
*    vuDpaSelectDpaDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuDpaSelectDpaDialogClass::LoadInit(vresource res)
{
    dpaList = vuDpaSelectDpaListClass::RequestSubLoad (this,
					vunames_vuDpaSelectDpaList);
    gotoButton = vuDpaSelectDpaGoToButtonClass::RequestSubLoad (this,
					vunames_vuDpaSelectDpaGoToButton);
    dismissButton = vuDpaSelectDpaDismissButtonClass::RequestSubLoad (this,
					vunames_vuDpaSelectDpaDismissButton);

    // call base class loadInit to finish the job
    vdialog::LoadInit(res);

    // set ptrs
    dpaList -> setSelectDpaDialog (this);
    gotoButton -> setSelectDpaDialog (this);
    dismissButton -> setSelectDpaDialog (this);

}


/*******************************************************************
*
*    vuDpaSelectDpaDialogClass notify function
*
*******************************************************************/

void vuDpaSelectDpaDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	vmemSet ((void *)selectedDpa, 0, sizeof (selectedDpa));
	Close();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuDpaSelectDpaDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuDpaSelectDpaDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }

    vliststr *list = (vliststr *)(dpaList -> GetList ());
    if (list != NULL)
    {
	list -> Destroy ();
    }

    vdialog::Destroy();
}


/*******************************************************************
*
*    vuDpaSelectDpaDialogClass Open function
*
*******************************************************************/

void vuDpaSelectDpaDialogClass::vuDpaSelectOpen(vuMainDialogClass *mainDialg)
{
    mainDialog = mainDialg;

    dpaList -> buildDpaList ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    Place(GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuDpaSelectDpaGoToButtonClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaSelectDpaGoToButtonClass, vbutton, 
					"vuDpaSelectDpaGoToButtonClass");

/*******************************************************************
*
*    vuDpaSelectDpaGoToButtonClass Observer function
*
*******************************************************************/

void vuDpaSelectDpaGoToButtonClass::ObserveDialogItem (vevent *event)
{
    vliststr *list = (vliststr *)(selectDpaDialog -> dpaList -> GetList ());
    if (list == NULL)
    {
	selectDpaDialog -> Close ();
	vevent::StopProcessing ();
	return;
    }
    int rowCount = list -> GetRowCount ();
    if (rowCount == 0)
    {
	selectDpaDialog -> Close ();
	vevent::StopProcessing ();
	return;
    }
    vlistSelection *selection = selectDpaDialog -> dpaList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	vuMessageDialogCenter (selectDpaDialog, VU_DPA_NODPA_SEL_MSG);
	return;
    }
    vchar *dpaValue = (vchar *)(list -> GetCellValue (row, 0));
    vchar *selectedDpa = selectDpaDialog -> getSelectedDpa ();
    vcharCopyBounded (dpaValue, selectedDpa, RESOURCE_SZ);
    selectedDpa[RESOURCE_SZ] = 0;
    selectDpaDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuDpaSelectDpaDismissButtonClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaSelectDpaDismissButtonClass, vbutton, 
					"vuDpaSelectDpaDismissButtonClass");

/*******************************************************************
*
*    vuDpaSelectDpaDismissButtonClass Observer function
*
*******************************************************************/

void vuDpaSelectDpaDismissButtonClass::ObserveDialogItem (vevent *event)
{
    vchar *selectedDpa = selectDpaDialog -> getSelectedDpa ();
    vmemSet ((void *)selectedDpa, 0, RESOURCE_SZ);
    selectDpaDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuDpaSelectDpaListClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaSelectDpaListClass, vlistview, 
					"vuDpaSelectDpaListClass");

/*******************************************************************
*
*    vuDpaSelectDpaListClass Observe function
*
*******************************************************************/

void vuDpaSelectDpaListClass::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	int clickCount = GetLastPointerClickCount ();
	if (clickCount == 2)
	{
	    vchar	*itemValue;
	    int		row, column;

	    vliststr *list = (vliststr *)(GetList ());
	    vlistSelection *listSelection = GetSelection ();
	    list -> GetSelectedCell (listSelection, &row, &column);
	    itemValue = (vchar *)(list -> GetCellValue (row, 0));

	    vchar *selectedDpa = selectDpaDialog -> getSelectedDpa ();
	    vcharCopyBounded (itemValue, selectedDpa, RESOURCE_SZ);
	    selectedDpa[RESOURCE_SZ] = 0;
				
	    // close the dialog and stop the sub event loop
	    // started in vuDpaSelectOpen ()
	    selectDpaDialog -> Close ();
	    vevent::StopProcessing ();
	}
    }
}


/*******************************************************************
*
*    vuDpaSelectDpaListClass build function
*
*******************************************************************/

void vuDpaSelectDpaListClass::buildDpaList (void)
{
    vliststr *list = (vliststr *)(GetList ());
    if (list != NULL)
    {
	SetList (NULL);
	list -> Destroy ();
    }

    vuMainDialogClass *mainDialog = selectDpaDialog -> getMainDialog ();
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();

    list = vuGetResList (db, VU_ONLY_DPAS);
   
    //  tell list view what to look at
    SetList(list);

    // set so only one resource name can be selected
    selection = GetSelection ();
    list -> SetSelectMethod (selection, vlistONE_ONLY_SELECTION);
}


/*******************************************************************
*
*	DPA SELECT POINT DIALOG
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaSelectPtDialogClass, vdialog, 
						"vuDpaSelectPtDialogClass");

/*******************************************************************
*
*    vuDpaSelectPtDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuDpaSelectPtDialogClass::LoadInit(vresource res)
{
    dpaList = vuDpaSelectPtListClass::RequestSubLoad (this,
					vunames_vuDpaSelectPtList);
    gotoButton = vuDpaSelectPtGoToButtonClass::RequestSubLoad (this,
					vunames_vuDpaSelectPtGoToButton);
    dismissButton = vuDpaSelectPtDismissButtonClass::RequestSubLoad (this,
					vunames_vuDpaSelectPtDismissButton);

    // call base class loadInit to finish the job
    vdialog::LoadInit(res);

    // set ptrs
    dpaList -> setDpaSelectDialog (this);
    gotoButton -> setDpaSelectDialog (this);
    dismissButton -> setDpaSelectDialog (this);

}


/*******************************************************************
*
*    vuDpaSelectPtDialogClass notify function
*
*******************************************************************/

void vuDpaSelectPtDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	vmemSet ((void *)selectedDpa, 0, sizeof (selectedDpa));
	Close();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuDpaSelectPtDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuDpaSelectPtDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }

    vliststr *list = (vliststr *)(dpaList -> GetList ());
    if (list != NULL)
    {
	list -> Destroy ();
    }

    vdialog::Destroy();
}


/*******************************************************************
*
*    vuDpaSelectPtDialogClass Open function
*
*******************************************************************/

void vuDpaSelectPtDialogClass::vuDpaSelectOpen(vuDpaDialogClass *dlg)
{
    dpaDialog = dlg;
    vuMainDialogClass *mainDialog = dpaDialog -> getMainDialog ();

    dpaList -> buildDpaList ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    Place(dpaDialog, vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuDpaSelectPtGoToButtonClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaSelectPtGoToButtonClass, vbutton, 
					"vuDpaSelectPtGoToButtonClass");

/*******************************************************************
*
*    vuDpaSelectPtGoToButtonClass Observer function
*
*******************************************************************/

void vuDpaSelectPtGoToButtonClass::ObserveDialogItem (vevent *event)
{
    vliststr *list = (vliststr *)(selectDpaDialog -> dpaList -> GetList ());
    if (list == NULL)
    {
	selectDpaDialog -> Close ();
	vevent::StopProcessing ();
	return;
    }
    int rowCount = list -> GetRowCount ();
    if (rowCount == 0)
    {
	selectDpaDialog -> Close ();
	vevent::StopProcessing ();
	return;
    }
    vlistSelection *selection = selectDpaDialog -> dpaList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	vuMessageDialogCenter (selectDpaDialog, VU_DPA_NODPA_SEL_MSG);
	return;
    }
    // get the selected DAP point name from the list and save it
    vchar *dpaValue = (vchar *)(list -> GetCellValue (row, 0));
    vchar *selectedDpa = selectDpaDialog -> getSelectedDpa ();
    vcharCopyBounded (dpaValue, selectedDpa, PVNAME_SZ);
    selectedDpa[PVNAME_SZ] = 0;
    selectDpaDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuDpaSelectPtDismissButtonClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaSelectPtDismissButtonClass, vbutton, 
					"vuDpaSelectPtDismissButtonClass");

/*******************************************************************
*
*    vuDpaSelectPtDismissButtonClass Observer function
*
*******************************************************************/

void vuDpaSelectPtDismissButtonClass::ObserveDialogItem (vevent *event)
{
    vchar *selectedDpaPt = selectDpaDialog -> getSelectedDpa ();
    vmemSet ((void *)selectedDpaPt, 0, PVNAME_SZ);
    selectDpaDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuDpaSelectPtListClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaSelectPtListClass, vlistview, 
					"vuDpaSelectPtListClass");

/*******************************************************************
*
*    vuDpaSelectPtListClass Observe function
*
*******************************************************************/

void vuDpaSelectPtListClass::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	int clickCount = GetLastPointerClickCount ();
	if (clickCount == 2)
	{
	    vchar	*itemValue;
	    int		row, column;

	    vliststr *list = (vliststr *)(GetList ());
	    vlistSelection *listSelection = GetSelection ();
	    list -> GetSelectedCell (listSelection, &row, &column);
	    itemValue = (vchar *)(list -> GetCellValue (row, 0));

	    vchar *selectedDpaPt = selectDpaDialog -> getSelectedDpa ();
	    vcharCopyBounded (itemValue, selectedDpaPt, PVNAME_SZ);
	    selectedDpaPt[PVNAME_SZ] = 0;
				
	    // close the dialog and stop the sub event loop
	    // started in vuDpaSelectOpen ()
	    selectDpaDialog -> Close ();
	    vevent::StopProcessing ();
	}
    }
}


/*******************************************************************
*
*    vuDpaSelectPtListClass build function
*
*******************************************************************/

void vuDpaSelectPtListClass::buildDpaList (void)
{
    vliststr *list = (vliststr *)(GetList ());
    if (list != NULL)
    {
	SetList (NULL);
	list -> Destroy ();
    }

    vuDpaDialogClass  *dpaDialog = selectDpaDialog -> getDpaDialog ();
    DB_TABLE_PTR dpaTable = dpaDialog -> getDpaTable ();

    DB_RECORD_PTR rec = DB_Top (dpaTable);
    DB_FIELD_PTR  dpaNameField = DB_MapField (dpaTable, FLD_NAME);
    char dpaName[PVNAME_SZ + 1];
    int  dataType;
    int row = 0;

    if (rec)
    {
	list = new vliststr;
    }
    else
    {
	SetList (NULL);
	return;
    }

    while (rec)
    {
	DB_GetFieldValue (rec, dpaNameField, dpaName, &dataType);
	list -> SetCellValueScribed (row, 0, vcharScribeSystem (dpaName));
	row++;
	rec = DB_SeekNext (dpaTable);
    }
    SetList (list);
}


/*******************************************************************
*
*    DPA CONFIG DIALOG CLASS
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaDialogClass, vdialog, 
					"vuDpaDialogClass");

/*******************************************************************
*
*    vuDpaDialogClass LoadInit constructor
*
*******************************************************************/

void vuDpaDialogClass::LoadInit (vresource res)
{
    vcontainer *pvcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditPvContainer);
    dpaName = filteredTextitem::RequestSubLoad (pvcontainer,
				vunames_vuDpaEditPV);

    vcontainer *desccontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditDESCContainer);
    description = filteredTextitem::RequestSubLoad (desccontainer,
				vunames_vuDpaEditDESC);

    vcontainer *preccontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditPRECContainer);
    precision = filteredTextitem::RequestSubLoad (preccontainer,
				vunames_vuDpaEditPREC);

    vcontainer *elemcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditELEMContainer);
    elements = filteredTextitem::RequestSubLoad (elemcontainer,
				vunames_vuDpaEditELEM);

    vcontainer *egucontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditEGUContainer);
    eguTag = filteredTextitem::RequestSubLoad (egucontainer,
				vunames_vuDpaEditEGU);

    vcontainer *dtypcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditDTYPContainer);
    caDataType = vuDpaDtypeMenuClass::RequestSubLoad (dtypcontainer,
				vunames_vuDpaEditDTYP);

    vcontainer *udatacontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditUDATAContainer);
    userData = filteredTextitem::RequestSubLoad (udatacontainer,
				vunames_vuDpaEditUDATA);

    onAlarm = checkBoxClass::RequestSubLoad (this,
				vunames_vuDpaEditAENB);

    onTrend = checkBoxClass::RequestSubLoad (this,
				vunames_vuDpaEditTENB);

    vcontainer *strscontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditSTRSContainer);

    vcontainer *str1container = vcontainer::RequestSubLoad (strscontainer,
				vunames_vuDpaEditSTR1Container);
    strings[0] = filteredTextitem::RequestSubLoad (str1container,
				vunames_vuDpaEditSTR1);

    vcontainer *str2container = vcontainer::RequestSubLoad (strscontainer,
				vunames_vuDpaEditSTR2Container);
    strings[1] = filteredTextitem::RequestSubLoad (str2container,
				vunames_vuDpaEditSTR2);

    vcontainer *str3container = vcontainer::RequestSubLoad (strscontainer,
				vunames_vuDpaEditSTR3Container);
    strings[2] = filteredTextitem::RequestSubLoad (str3container,
				vunames_vuDpaEditSTR3);

    vcontainer *str4container = vcontainer::RequestSubLoad (strscontainer,
				vunames_vuDpaEditSTR4Container);
    strings[3] = filteredTextitem::RequestSubLoad (str4container,
				vunames_vuDpaEditSTR4);

    vcontainer *str5container = vcontainer::RequestSubLoad (strscontainer,
				vunames_vuDpaEditSTR5Container);
    strings[4] = filteredTextitem::RequestSubLoad (str5container,
				vunames_vuDpaEditSTR5);

    vcontainer *str6container = vcontainer::RequestSubLoad (strscontainer,
				vunames_vuDpaEditSTR6Container);
    strings[5] = filteredTextitem::RequestSubLoad (str6container,
				vunames_vuDpaEditSTR6);

    vcontainer *str7container = vcontainer::RequestSubLoad (strscontainer,
				vunames_vuDpaEditSTR7Container);
    strings[6] = filteredTextitem::RequestSubLoad (str7container,
				vunames_vuDpaEditSTR7);

    vcontainer *str8container = vcontainer::RequestSubLoad (strscontainer,
				vunames_vuDpaEditSTR8Container);
    strings[7] = filteredTextitem::RequestSubLoad (str8container,
				vunames_vuDpaEditSTR8);

    vcontainer *valscontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditVALSContainer);

    vcontainer *loprcontainer = vcontainer::RequestSubLoad (valscontainer,
				vunames_vuDpaEditLOPRContainer);
    dspLow = filteredTextitem::RequestSubLoad (loprcontainer,
				vunames_vuDpaEditLOPR);

    vcontainer *hoprcontainer = vcontainer::RequestSubLoad (valscontainer,
				vunames_vuDpaEditHOPRContainer);
    dspHigh = filteredTextitem::RequestSubLoad (hoprcontainer,
				vunames_vuDpaEditHOPR);

    vcontainer *irlocontainer = vcontainer::RequestSubLoad (valscontainer,
				vunames_vuDpaEditIRLOContainer);
    irLow = filteredTextitem::RequestSubLoad (irlocontainer,
				vunames_vuDpaEditIRLO);

    vcontainer *irhicontainer = vcontainer::RequestSubLoad (valscontainer,
				vunames_vuDpaEditIRHIContainer);
    irHigh = filteredTextitem::RequestSubLoad (irhicontainer,
				vunames_vuDpaEditIRHI);

    vcontainer *alrmcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditALARMContainer);

    vcontainer *hihicontainer = vcontainer::RequestSubLoad (alrmcontainer,
				vunames_vuDpaEditHIHIContainer);
    hihiLimit = filteredTextitem::RequestSubLoad (hihicontainer,
				vunames_vuDpaEditHIHI);

    vcontainer *highcontainer = vcontainer::RequestSubLoad (alrmcontainer,
				vunames_vuDpaEditHIGHContainer);
    highLimit = filteredTextitem::RequestSubLoad (highcontainer,
				vunames_vuDpaEditHIGH);

    vcontainer *lowcontainer = vcontainer::RequestSubLoad (alrmcontainer,
				vunames_vuDpaEditLOWContainer);
    lowLimit = filteredTextitem::RequestSubLoad (lowcontainer,
				vunames_vuDpaEditLOW);

    vcontainer *lolocontainer = vcontainer::RequestSubLoad (alrmcontainer,
				vunames_vuDpaEditLOLOContainer);
    loloLimit = filteredTextitem::RequestSubLoad (lolocontainer,
				vunames_vuDpaEditLOLO);

    vcontainer *hhsvcontainer = vcontainer::RequestSubLoad (alrmcontainer,
				vunames_vuDpaEditHHSVContainer);
    hihiSeverity = vuChoiceGblMenuClass::RequestSubLoad (hhsvcontainer,
				vunames_vuDpaEditHHSV);

    vcontainer *hsvcontainer = vcontainer::RequestSubLoad (alrmcontainer,
				vunames_vuDpaEditHSVContainer);
    highSeverity = vuChoiceGblMenuClass::RequestSubLoad (hsvcontainer,
				vunames_vuDpaEditHSV);

    vcontainer *lsvcontainer = vcontainer::RequestSubLoad (alrmcontainer,
				vunames_vuDpaEditLSVContainer);
    lowSeverity = vuChoiceGblMenuClass::RequestSubLoad (lsvcontainer,
				vunames_vuDpaEditLSV);

    vcontainer *llsvcontainer = vcontainer::RequestSubLoad (alrmcontainer,
				vunames_vuDpaEditLLSVContainer);
    loloSeverity = vuChoiceGblMenuClass::RequestSubLoad (llsvcontainer,
				vunames_vuDpaEditLLSV);

    vcontainer *buttoncontainer = vcontainer::RequestSubLoad (this,
				vunames_vuDpaEditButtonContainer);

    applyButton = vuDpaApplyButtonClass::RequestSubLoad (buttoncontainer,
				vunames_vuDpaEditApplyButton);
    dismissButton = vuDpaDismissButtonClass::RequestSubLoad (buttoncontainer,
				vunames_vuDpaEditDismissButton);
    deleteButton = vuDpaDeleteButtonClass::RequestSubLoad (buttoncontainer,
				vunames_vuDpaEditDeleteButton);
    gotoButton = vuDpaGoToButtonClass::RequestSubLoad (buttoncontainer,
				vunames_vuDpaEditGoToButton);
    nextButton = vuDpaNextButtonClass::RequestSubLoad (buttoncontainer,
				vunames_vuDpaEditNextButton);
    prevButton = vuDpaPrevButtonClass::RequestSubLoad (buttoncontainer,
				vunames_vuDpaEditPrevButton);

    vdialog::LoadInit (res);

    // have buttons remember ptr to dialog
    applyButton -> storeDpaDialog (this);
    dismissButton -> storeDpaDialog (this);
    deleteButton -> storeDpaDialog (this);
    gotoButton -> storeDpaDialog (this);
    nextButton -> storeDpaDialog (this);
    prevButton -> storeDpaDialog (this);

    // store ptr to dialogModified flag
    dpaName -> SetTextModifiedPtr (&this -> dialogModified);
    description -> SetTextModifiedPtr (&this -> dialogModified);
    precision -> SetTextModifiedPtr (&this -> dialogModified);
    elements -> SetTextModifiedPtr (&this -> dialogModified);
    eguTag -> SetTextModifiedPtr (&this -> dialogModified);
    caDataType -> SetModifiedPtr (&this -> dialogModified);
    userData -> SetTextModifiedPtr (&this -> dialogModified);
    onAlarm -> SetModifiedPtr (&this -> dialogModified);
    onTrend -> SetModifiedPtr (&this -> dialogModified);
    strings[0] -> SetTextModifiedPtr (&this -> dialogModified);
    strings[1] -> SetTextModifiedPtr (&this -> dialogModified);
    strings[2] -> SetTextModifiedPtr (&this -> dialogModified);
    strings[3] -> SetTextModifiedPtr (&this -> dialogModified);
    strings[4] -> SetTextModifiedPtr (&this -> dialogModified);
    strings[5] -> SetTextModifiedPtr (&this -> dialogModified);
    strings[6] -> SetTextModifiedPtr (&this -> dialogModified);
    strings[7] -> SetTextModifiedPtr (&this -> dialogModified);
    dspLow -> SetTextModifiedPtr (&this -> dialogModified);
    dspHigh -> SetTextModifiedPtr (&this -> dialogModified);
    irHigh -> SetTextModifiedPtr (&this -> dialogModified);
    irLow -> SetTextModifiedPtr (&this -> dialogModified);
    hihiLimit -> SetTextModifiedPtr (&this -> dialogModified);
    highLimit -> SetTextModifiedPtr (&this -> dialogModified);
    lowLimit -> SetTextModifiedPtr (&this -> dialogModified);
    loloLimit -> SetTextModifiedPtr (&this -> dialogModified);
    hihiSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    highSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    lowSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    loloSeverity -> vuSetModifiedPtr (&this -> dialogModified);

    // set filter function ptrs
    dpaName -> SetItemFilterFunc (pvnameFilter);
    description -> SetItemFilterFunc (descriptorFilter);
    precision -> SetItemFilterFunc (numericFilter);
    elements -> SetItemFilterFunc (numericFilter);
    eguTag -> SetItemFilterFunc (egutagFilter);
    userData -> SetItemFilterFunc (vuDpaDataFilter);
    strings[0] -> SetItemFilterFunc (vuDpaStringFilter);
    strings[1] -> SetItemFilterFunc (vuDpaStringFilter);
    strings[2] -> SetItemFilterFunc (vuDpaStringFilter);
    strings[3] -> SetItemFilterFunc (vuDpaStringFilter);
    strings[4] -> SetItemFilterFunc (vuDpaStringFilter);
    strings[5] -> SetItemFilterFunc (vuDpaStringFilter);
    strings[6] -> SetItemFilterFunc (vuDpaStringFilter);
    strings[7] -> SetItemFilterFunc (vuDpaStringFilter);
    dspLow -> SetItemFilterFunc (floatFilter);
    dspHigh -> SetItemFilterFunc (floatFilter);
    irLow -> SetItemFilterFunc (floatFilter);
    irHigh -> SetItemFilterFunc (floatFilter);
    hihiLimit -> SetItemFilterFunc (floatFilter);
    highLimit -> SetItemFilterFunc (floatFilter);
    lowLimit -> SetItemFilterFunc (floatFilter);
    loloLimit -> SetItemFilterFunc (floatFilter);

    // build menus
    caDataType -> buildMenu ();
    hihiSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    highSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    lowSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    loloSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);
}


/*******************************************************************
*
*    vuDpaDialogClass notify function
*
*******************************************************************/

void vuDpaDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	int result = VU_TRUE;
	if (dialogModified == vTRUE)
	{
	    result = vuMessageDialogCenter (this, VU_MOD_DISMISS_MSG); 
	}
	if (result == VU_TRUE)
	{
	    Close();
	    vevent::StopProcessing ();
	}
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuDpaDialogClass destroy function
*
*******************************************************************/

void vuDpaDialogClass::Destroy(void)
{
    // delete selectDpaDialog but 1st destroy its list
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }

    if (selectDpaDialog != NULL)
	delete selectDpaDialog;

    vdialog::Destroy();
}


/*******************************************************************
*
*    vuDpaDialogClass open function
*
*******************************************************************/

void vuDpaDialogClass::vuDpaOpen(vuMainDialogClass *mnDlg)
{
    mainDialog = mnDlg;
    selectDpaDialog = NULL;
    dialogModified = vFALSE;

    dataVuApplication *dataVuApp = mainDialog -> getApplication ();
    vresource vuRes = dataVuApp -> getDataVuvresource ();

    vresource selRes = vresourceGet(vuRes, vunames_vuDpaSelectDpaDialog);
    vuDpaSelectDpaDialogClass *selDpa = new vuDpaSelectDpaDialogClass (selRes);
    selDpa -> vuDpaSelectOpen (mainDialog);

    vcharCopyBounded (selDpa -> getSelectedDpa (), selectedDpa, RESOURCE_SZ);
    selectedDpa[RESOURCE_SZ] = 0;
    delete selDpa;
    if ((vcharLength (selectedDpa)) == 0)
    {
	return;
    }
    DB_TABLE_PTR table;
    char resName[RESOURCE_SZ + 1];
    vcharExportBounded (vcharScribeSystemExport (selectedDpa),
			resName,
			RESOURCE_SZ);

    if ((table = DB_MapTable (resName)) == 0)
    {
	if ((table = DB_CreateTable (resName,
				     dpaFieldInfo,
				     dpaTagInfo)) == 0)
	{
	    vuMessageDialogCenter (NULL,
				   VU_DPAFILEERR,
				   selectedDpa);
	    return;
	}
	else
	{
	    mainDialog -> vuMainMessageCenter (vunames_vuDpaCreateMsg,
					       selectedDpa);
	}
    }
    vuDpaLoadConfig (table, 1);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    Place(GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuDpaDialogClass load dialog function
*
*******************************************************************/

void vuDpaDialogClass::vuDpaLoadConfig(DB_TABLE_PTR table, int gotoTop)
{
    int  i;
    char strFldName[16];

    // map to database fields
    if (table != dpaTable)
    {
	dpaTable = table;
	dpaNameField = DB_MapField (dpaTable, "NAME");
	dpaDescField = DB_MapField (dpaTable, "DESC");
	dpaPrecField = DB_MapField (dpaTable, "PRECISION");
	dpaElemField = DB_MapField (dpaTable, "COUNT");
	dpaEguField  = DB_MapField (dpaTable, "EGUTAG");
	dpaDtypeField  = DB_MapField (dpaTable, "DTYPE");
	dpaDataField  = DB_MapField (dpaTable, "Userdata");
	dpaAenbField  = DB_MapField (dpaTable, "AENB");
	dpaTenbField  = DB_MapField (dpaTable, "TENB");
	dpaNumStrField  = DB_MapField (dpaTable, "NUMSTRS");
	for (i = 0; i < 8; i++)
	{
	    sprintf (strFldName, "STRING%d\0", i+1);
	    dpaStrsFields[i] = DB_MapField (dpaTable, strFldName);
	}
	dpaLoprField  = DB_MapField (dpaTable, "DISPLOW");
	dpaHoprField  = DB_MapField (dpaTable, "DISPHIGH");
	dpaIrloField  = DB_MapField (dpaTable, "IRLOW");
	dpaIrhiField  = DB_MapField (dpaTable, "IRHIGH");
	dpaHihiField  = DB_MapField (dpaTable, "HIHILIMIT");
	dpaHighField  = DB_MapField (dpaTable, "HIGHLIMIT");
	dpaLowField  = DB_MapField (dpaTable, "LOWLIMIT");
	dpaLoloField  = DB_MapField (dpaTable, "LOLOLIMIT");
	dpaHhsvField  = DB_MapField (dpaTable, "HIHISEVER");
	dpaHsvField  = DB_MapField (dpaTable, "HIGHSEVER");
	dpaLsvField  = DB_MapField (dpaTable, "LOWSEVER");
	dpaLlsvField  = DB_MapField (dpaTable, "LOLOSEVER");
    }
    if ( (((vcharLength (selectedDpa)) != 0) && (gotoTop > 0)) ||
	 (((vcharLength (selectedDpaPt)) != 0) && (gotoTop == 0)) )
    {
	char   buff[256];
	char   state;
	int    dataType;
	double doubleValue;
	int    intValue;
	vstr   *finalValue;
	DB_RECORD_PTR rec;
	
	if (gotoTop)
	{
	    rec = DB_Top (dpaTable);
	}
	else
	{
	    char selectedName[PVNAME_SZ + 1];
	    vcharExportBounded (vcharScribeSystemExport (selectedDpaPt),
				selectedName,
				PVNAME_SZ);
	    selectedName[PVNAME_SZ] = 0;
	    rec = DB_Seek (dpaTable, selectedName);
	}

	DB_GetFieldValue (rec, dpaNameField, (void *)buff, &dataType);
	dpaName -> SetTextScribed (vcharScribeSystem (buff));

	DB_GetFieldValue (rec, dpaDescField, (void *)buff, &dataType);
	description -> SetTextScribed (vcharScribeSystem (buff));

	DB_GetFieldValue (rec, dpaPrecField, (void *)&doubleValue, &dataType);
	intValue = (int)doubleValue;
	finalValue = vstrCloneScribed (vnumScribeInt (intValue));
	precision -> SetText (finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaElemField, (void *)&doubleValue, &dataType);
	intValue = (int)doubleValue;
	finalValue = vstrCloneScribed (vnumScribeInt (intValue));
	elements -> SetText (finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaEguField, (void *)buff, &dataType);
	eguTag -> SetTextScribed (vcharScribeSystem (buff));

	DB_GetFieldValue (rec, dpaDtypeField, (void *)&doubleValue, &dataType);
	intValue = (int)doubleValue;
	finalValue = vstrCloneScribed (vcharScribeSystem (dTypeStrings[intValue]));
	caDataType -> SetOptionMenuValue (finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaDataField, (void *)buff, &dataType);
	userData -> SetTextScribed (vcharScribeSystem (buff));

	DB_GetFieldValue (rec, dpaAenbField, (void *)&state, &dataType);
	if (state & 1)
	{
	    onAlarm -> SetValue (1);
	}
	else
	{
	    onAlarm -> SetValue (0);
	}

	DB_GetFieldValue (rec, dpaTenbField, (void *)&state, &dataType);
	if (state & 1)
	{
	    onTrend -> SetValue (1);
	}
	else
	{
	    onTrend -> SetValue (0);
	}

	for (i = 0; i < 8; i++)
	{
	    DB_GetFieldValue (rec, dpaStrsFields[i], (void *)buff, &dataType);
	    strings[i] -> SetTextScribed (vcharScribeSystem (buff));

	}

	DB_GetFieldValue (rec, dpaLoprField, (void *)&doubleValue, &dataType);
	finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
	dspLow -> SetText (finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaHoprField, (void *)&doubleValue, &dataType);
	finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
	dspHigh -> SetText (finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaIrloField, (void *)&doubleValue, &dataType);
	finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
	irLow -> SetText (finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaIrhiField, (void *)&doubleValue, &dataType);
	finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
	irHigh -> SetText (finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaHihiField, (void *)&doubleValue, &dataType);
	finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
	hihiLimit -> SetText (finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaHighField, (void *)&doubleValue, &dataType);
	finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
	highLimit -> SetText (finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaLowField, (void *)&doubleValue, &dataType);
	finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
	lowLimit -> SetText (finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaLoloField, (void *)&doubleValue, &dataType);
	finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
	loloLimit -> SetText (finalValue);
	vstrDestroy (finalValue);

	// point the list of alarm severities in the SDR
	char **choices = choiceGbl -> papChoiceSet[GBL_ALARM_SEV] -> papChoice;

	DB_GetFieldValue (rec, dpaHhsvField, (void *)&doubleValue, &dataType);
	intValue = (int)doubleValue;
	finalValue = vstrCloneScribed (vcharScribeSystem (choices[intValue]));
	vuSetOptionMenuValue (hihiSeverity, finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaHsvField, (void *)&doubleValue, &dataType);
	intValue = (int)doubleValue;
	finalValue = vstrCloneScribed (vcharScribeSystem (choices[intValue]));
	vuSetOptionMenuValue (highSeverity, finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaLsvField, (void *)&doubleValue, &dataType);
	intValue = (int)doubleValue;
	finalValue = vstrCloneScribed (vcharScribeSystem (choices[intValue]));
	vuSetOptionMenuValue (lowSeverity, finalValue);
	vstrDestroy (finalValue);

	DB_GetFieldValue (rec, dpaLlsvField, (void *)&doubleValue, &dataType);
	intValue = (int)doubleValue;
	finalValue = vstrCloneScribed (vcharScribeSystem (choices[intValue]));
	vuSetOptionMenuValue (loloSeverity, finalValue);
	vstrDestroy (finalValue);
    }
    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuDpaDialogClass reset dialog function
*
*******************************************************************/

void vuDpaDialogClass::vuDpaResetConfig(void)
{
    
    dpaName -> SetTextScribed (vcharScribeSystem (""));
    description -> SetTextScribed (vcharScribeSystem (""));
    precision -> SetTextScribed (vcharScribeSystem (""));
    elements -> SetTextScribed (vcharScribeSystem (""));
    eguTag -> SetTextScribed (vcharScribeSystem (""));
    caDataType -> SetValue (0);
    userData -> SetTextScribed (vcharScribeSystem (""));
    onAlarm -> SetValue (0);
    onTrend -> SetValue (0);

    int i;
    for (i = 0; i< 8; i++)
    {
	strings[i] -> SetTextScribed (vcharScribeSystem (""));
    }

    dspLow -> SetTextScribed (vcharScribeSystem (""));
    dspHigh -> SetTextScribed (vcharScribeSystem (""));
    irLow -> SetTextScribed (vcharScribeSystem (""));
    irHigh -> SetTextScribed (vcharScribeSystem (""));
    hihiLimit -> SetTextScribed (vcharScribeSystem (""));
    highLimit -> SetTextScribed (vcharScribeSystem (""));
    lowLimit -> SetTextScribed (vcharScribeSystem (""));
    loloLimit -> SetTextScribed (vcharScribeSystem (""));
    hihiSeverity -> SetValue (0);
    highSeverity -> SetValue (0);
    highSeverity -> SetValue (0);
    highSeverity -> SetValue (0);

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuDpaDialogClass save dialog function
*
*******************************************************************/

void vuDpaDialogClass::vuDpaSaveConfig(void)
{
    vstr		*text;
    char		name[PVNAME_SZ + 1];
    char		buff[256];
    DB_RECORD_PTR	rec;
    int			newRecord = vFALSE;
    double		number;
    int			value;
    char		trueFalse;
    int			i, numStrs;

    text = dpaName -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (text),
			name,
			PVNAME_SZ);
    name[PVNAME_SZ] = 0;
    sprintf (buff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, name);
    if ((rec = DB_Seek (dpaTable, buff)) == 0)
    {
	rec = DB_NewRecord (dpaTable,NULL);
	newRecord = vTRUE;
	mainDialog -> vuMainMessageCenter (vunames_vuDpaCreatePtMsg,
					   text, selectedDpa);
    }
    else
    {
	mainDialog -> vuMainMessageCenter (vunames_vuDpaModPtMsg,
					   text, selectedDpa);
    }

    DB_PutFieldValue (rec, dpaNameField, name);
    vstrDestroy (text);

    text = description -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (text),
			buff,
			DESCRIPTOR_SZ);
    DB_PutFieldValue (rec, dpaDescField, buff);
    vstrDestroy (text);

    text = precision -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	number = 0;
    }
    else
    {
	number = (double)(vnum::ParseInt (text, NULL));
    }
    DB_PutFieldValue (rec, dpaPrecField, &number);
    vstrDestroy (text);

    text = elements -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	number = 0;
    }
    else
    {
	number = (double)(vnum::ParseInt (text, NULL));
    }
    DB_PutFieldValue (rec, dpaElemField, &number);
    vstrDestroy (text);

    text = eguTag -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (text),
			buff,
			EGUTAG_SZ);
    DB_PutFieldValue (rec, dpaEguField, buff);
    vstrDestroy (text);

    value = caDataType -> GetValue ();
    number = value;
    DB_PutFieldValue (rec, dpaDtypeField, &number);

    text = userData -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (text),
			buff,
			DPA_DATA_SZ);
    DB_PutFieldValue (rec, dpaDataField, buff);
    vstrDestroy (text);

    value = onAlarm -> GetValue ();
    trueFalse = value;
    DB_PutFieldValue (rec, dpaAenbField, &trueFalse);

    value = onTrend -> GetValue ();
    trueFalse = value;
    DB_PutFieldValue (rec, dpaTenbField, &trueFalse);

    for (i = 0, numStrs = 0; i < 8; i++)
    {
	text = strings[i] -> GetTextAsString ();
	if ((vcharLength (text)) != 0)
	{
	    vcharExportBounded (vcharScribeSystemExport (text),
				buff,
				DPA_STRING_SZ);
	    numStrs++;
	    DB_PutFieldValue (rec, dpaStrsFields[i], buff);
	}
	vstrDestroy (text);
    }
    number = numStrs;
    DB_PutFieldValue (rec, dpaNumStrField, &number);

    text = dspLow -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	number = 0;
    }
    else
    {
	number = vnum::ParseFloat (text, NULL);
    }
    DB_PutFieldValue (rec, dpaLoprField, &number);
    vstrDestroy (text);

    text = dspHigh -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	number = 0;
    }
    else
    {
	number = vnum::ParseFloat (text, NULL);
    }
    DB_PutFieldValue (rec, dpaHoprField, &number);
    vstrDestroy (text);

    text = irLow -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	number = 0;
    }
    else
    {
	number = vnum::ParseFloat (text, NULL);
    }
    DB_PutFieldValue (rec, dpaIrloField, &number);
    vstrDestroy (text);

    text = irHigh -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	number = 0;
    }
    else
    {
	number = vnum::ParseFloat (text, NULL);
    }
    DB_PutFieldValue (rec, dpaIrhiField, &number);
    vstrDestroy (text);

    text = hihiLimit -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	number = 0;
    }
    else
    {
	number = vnum::ParseFloat (text, NULL);
    }
    DB_PutFieldValue (rec, dpaHihiField, &number);
    vstrDestroy (text);

    text = highLimit -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	number = 0;
    }
    else
    {
	number = vnum::ParseFloat (text, NULL);
    }
    DB_PutFieldValue (rec, dpaHighField, &number);
    vstrDestroy (text);

    text = lowLimit -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	number = 0;
    }
    else
    {
	number = vnum::ParseFloat (text, NULL);
    }
    DB_PutFieldValue (rec, dpaLowField, &number);
    vstrDestroy (text);

    text = loloLimit -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	number = 0;
    }
    else
    {
	number = vnum::ParseFloat (text, NULL);
    }
    DB_PutFieldValue (rec, dpaLoloField, &number);
    vstrDestroy (text);

    value = hihiSeverity -> GetValue ();
    number = value;
    DB_PutFieldValue (rec, dpaHhsvField, &number);

    value = highSeverity -> GetValue ();
    number = value;
    DB_PutFieldValue (rec, dpaHsvField, &number);

    value = lowSeverity -> GetValue ();
    number = value;
    DB_PutFieldValue (rec, dpaLsvField, &number);

    value = loloSeverity -> GetValue ();
    number = value;
    DB_PutFieldValue (rec, dpaLlsvField, &number);

    dialogModified = vFALSE;

    if (newRecord == vTRUE)
    {
	if (selectDpaDialog != NULL)
	{
	    selectDpaDialog -> dpaList -> buildDpaList ();
	}
    }
}


/*******************************************************************
*
*    vuDpaApplyButtonClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaApplyButtonClass, vbutton, 
					"vuDpaApplyButtonClass");

/*******************************************************************
*
*    vuDpaApplyButtonClass Observer function
*
*******************************************************************/

void vuDpaApplyButtonClass::ObserveDialogItem (vevent *event)
{
    vstr *text;

    text = dpaDialog -> dpaName -> GetTextAsString ();
    if ((vcharLength (text)) == 0)
    {
	vuMessageDialogCenter (dpaDialog, VU_DPA_NO_PT_MSG);
	vstrDestroy (text);
	return;
    }
    int result = vuMessageDialogCenter (dpaDialog, VU_DPA_UPDATE_REQ_MSG);
    if (result == VU_FALSE)
    {
	vstrDestroy (text);
	return;
    }
    vstrDestroy (text);
    dpaDialog -> vuDpaSaveConfig ();
}


/*******************************************************************
*
*    vuDpaDismissButtonClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaDismissButtonClass, vbutton, 
					"vuDpaDismissButtonClass");

/*******************************************************************
*
*    vuDpaDismissButtonClass Observer function
*
*******************************************************************/

void vuDpaDismissButtonClass::ObserveDialogItem (vevent *event)
{
    int result = VU_TRUE;
    if ((dpaDialog -> getDialogModify ()) == vTRUE)
    {
	result = vuMessageDialogCenter (dpaDialog, VU_MOD_DISMISS_MSG); 
    }
    if (result == VU_TRUE)
    {
	dpaDialog -> Close();
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    vuDpaDeleteButtonClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaDeleteButtonClass, vbutton, 
					"vuDpaDeleteButtonClass");

/*******************************************************************
*
*    vuDpaDeleteButtonClass Observer function
*
*******************************************************************/

void vuDpaDeleteButtonClass::ObserveDialogItem (vevent *event)
{
	printf ("Delete button\n");
}


/*******************************************************************
*
*    vuDpaGoToButtonClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaGoToButtonClass, vbutton, 
					"vuDpaGoToButtonClass");

/*******************************************************************
*
*    vuDpaGoToButtonClass Observer function
*
*******************************************************************/

void vuDpaGoToButtonClass::ObserveDialogItem (vevent *event)
{
    // chg to data do we want to overrite with new dpa point
    if ((dpaDialog -> getDialogModify ()) == vTRUE)
    {
	int result = vuMessageDialogCenter (dpaDialog, VU_MOD_DISMISS_MSG);
	if (result == VU_TRUE)
	{
	    return;
	}
    }
    vuDpaSelectPtDialogClass *dpaSelectDialog = dpaDialog -> getSelectDpaDialog ();
    // if first time to call the DPA select DPA point name
    if (dpaSelectDialog == NULL)
    {
	vuMainDialogClass *mainDialog = dpaDialog -> getMainDialog ();
	dataVuApplication *dataVuApp = mainDialog -> getApplication ();
	vresource vuRes = dataVuApp -> getDataVuvresource ();
	vresource selRes = vresourceGet (vuRes, vunames_vuDpaSelectPtDialog);
	dpaSelectDialog = new vuDpaSelectPtDialogClass (selRes);
	dpaDialog -> setSelectDpaDialog (dpaSelectDialog);
    }
    dpaSelectDialog -> vuDpaSelectOpen (dpaDialog);
    vchar *selectedDpa = dpaSelectDialog -> getSelectedDpa ();
    if ((vcharLength (selectedDpa)) != 0)
    {
	vchar *selectedDpaPt = dpaDialog -> getSelectedDpaPt ();
	// copy the choice in to the DPA Dialog class
	vcharCopy (selectedDpa, selectedDpaPt);
	selectedDpaPt[PVNAME_SZ] = 0;
	dpaDialog -> vuDpaResetConfig ();
	DB_TABLE_PTR dpaTable = dpaDialog -> getDpaTable ();
	dpaDialog -> vuDpaLoadConfig (dpaTable, 0);
    }
}



/*******************************************************************
*
*    vuDpaNextButtonClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaNextButtonClass, vbutton, 
					"vuDpaNextButtonClass");

/*******************************************************************
*
*    vuDpaNextButtonClass Observer function
*
*******************************************************************/

void vuDpaNextButtonClass::ObserveDialogItem (vevent *event)
{
    vuDpaSelectPtDialogClass *selectDpaDialog = dpaDialog -> getSelectDpaDialog ();
    if (selectDpaDialog == NULL)
    {
	vuMessageDialogCenter (dpaDialog, VU_NO_LIST_ERR);
	return;
    }
    vliststr *list = (vliststr *)(selectDpaDialog -> dpaList -> GetList ());
    vlistSelection *selection = selectDpaDialog -> dpaList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row != vlistNO_CELLS_SELECTED)
    {
	int numRows = list -> GetRowCount ();
	if ( (row + 1) < numRows)
	{
	    if ( (dpaDialog -> getDialogModify ()) == vTRUE)
	    {
		int result = vuMessageDialogCenter (dpaDialog, VU_MOD_DISMISS_MSG);
		if (result == VU_FALSE)
		{
		    return;
		}
	    }
	    list -> UnselectRow (selection, row);
	    row++;
	    list -> SelectRow (selection, row);
	    vchar *selectedName = (vchar *)(list -> GetCellValue (row, 0));

	    vchar *selectedDpaPt = dpaDialog -> getSelectedDpaPt ();
	    vcharCopy (selectedName, selectedDpaPt);
	    selectedDpaPt[PVNAME_SZ] = 0;
	    dpaDialog -> vuDpaResetConfig ();
	    DB_TABLE_PTR dpaTable = dpaDialog -> getDpaTable ();
	    dpaDialog -> vuDpaLoadConfig (dpaTable, 0);
	}
	else
	{
	    vuMessageDialogCenter (dpaDialog, VU_ENDOFLIST_MSG);
	}
    }
    else
    {
	vuMessageDialogCenter (dpaDialog, VU_DPA_NODPA_SEL_MSG);
    }
}


/*******************************************************************
*
*    vuDpaPrevButtonClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaPrevButtonClass, vbutton, 
					"vuDpaPrevButtonClass");

/*******************************************************************
*
*    vuDpaPrevButtonClass Observer function
*
*******************************************************************/

void vuDpaPrevButtonClass::ObserveDialogItem (vevent *event)
{
    vuDpaSelectPtDialogClass *selectDpaDialog = dpaDialog -> getSelectDpaDialog ();
    if (selectDpaDialog == NULL)
    {
	vuMessageDialogCenter (dpaDialog, VU_NO_LIST_ERR);
	return;
    }
    vliststr *list = (vliststr *)(selectDpaDialog -> dpaList -> GetList ());
    vlistSelection *selection = selectDpaDialog -> dpaList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row != vlistNO_CELLS_SELECTED)
    {
	int numRows = list -> GetRowCount ();
	if ( (row - 1) >= 0)
	{
	    if ( (dpaDialog -> getDialogModify ()) == vTRUE)
	    {
		int result = vuMessageDialogCenter (dpaDialog, VU_MOD_DISMISS_MSG);
		if (result == VU_FALSE)
		{
		    return;
		}
	    }
	    list -> UnselectRow (selection, row);
	    row--;
	    list -> SelectRow (selection, row);
	    vchar *selectedName = (vchar *)(list -> GetCellValue (row, 0));

	    vchar *selectedDpaPt = dpaDialog -> getSelectedDpaPt ();
	    vcharCopy (selectedName, selectedDpaPt);
	    selectedDpaPt[PVNAME_SZ] = 0;
	    dpaDialog -> vuDpaResetConfig ();
	    DB_TABLE_PTR dpaTable = dpaDialog -> getDpaTable ();
	    dpaDialog -> vuDpaLoadConfig (dpaTable, 0);
	}
	else
	{
	    vuMessageDialogCenter (dpaDialog, VU_STARTOFLIST_MSG);
	}
    }
    else
    {
	vuMessageDialogCenter (dpaDialog, VU_DPA_NODPA_SEL_MSG);
    }
}


/*******************************************************************
*
*    vuDpaDtypeMenuClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuDpaDtypeMenuClass, optionMenuClass, 
					"vuDpaDtypeMenuClass");

/*******************************************************************
*
*    vuDpaDtypeMenuClass build menu function
*
*******************************************************************/

void vuDpaDtypeMenuClass::buildMenu (void)
{
    vmenu *newMenu = new vmenu;
    vmenuItem *menuItem;
    int i = 0;
    int numFound = 0;

    while (dTypeStrings[i])
    {
	numFound++;
	menuItem = new vmenuItem;
	menuItem -> SetTitleScribed (vcharScribeSystem ((void *)dTypeStrings[i]));
	newMenu -> AppendItem (menuItem);
	i++;
    }

    // attach the menu to the control
    this -> SetMenu (newMenu);
    // set min and max for the control
    this -> SetMin (0);
    this -> SetMax (numFound - 1);
    // set value so current menu item will be the 1st one
    this -> SetValue (0);
}


/*******************************************************************
*
*    some filter functions unique to dpa
*
*******************************************************************/

static int vuDpaDataFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, DPA_DATA_SZ) );
}

static int vuDpaStringFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, DPA_STRING_SZ) );
}

