#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbFldTypes.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "dbRecDes.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuCommonDialogClass.hxx"
#include "vuFanoutDialogClass.hxx"
#include "vuNavigatorDialogClass.hxx"
#include "vuUtils.hxx"


vkindTYPED_FULL_DEFINITION(vuFanoutDialogClass, vuCommonDialogClass, "vuFanoutDialogClass");

/*******************************************************************
*
*    vuFanoutDialogClass override for constructors init ()
*
*******************************************************************/

void vuFanoutDialogClass::LoadInit (vresource res)
{
    // load in the ai specific fields
    fanoutContainer = vuFanoutContainerClass::RequestSubLoad (this,
					vunames_vuCommonContainer);
    vuCommonDialogClass::LoadInit (res);

    // store dialog modified flag
    fanoutContainer -> pv1 -> SetTextModifiedPtr (&this -> dialogModified);
    fanoutContainer -> pv2 -> SetTextModifiedPtr (&this -> dialogModified);
    fanoutContainer -> pv3 -> SetTextModifiedPtr (&this -> dialogModified);
    fanoutContainer -> pv4 -> SetTextModifiedPtr (&this -> dialogModified);
    fanoutContainer -> pv5 -> SetTextModifiedPtr (&this -> dialogModified);
    fanoutContainer -> pv6 -> SetTextModifiedPtr (&this -> dialogModified);
    fanoutContainer -> selectionNumber -> SetTextModifiedPtr (&this -> dialogModified);
    fanoutContainer -> onScan -> vuSetModifiedPtr (&this -> dialogModified);
    fanoutContainer -> specifySelection -> vuSetModifiedPtr (&this -> dialogModified);

    // store dialog ptr for vuPvNameButtonClass
    fanoutContainer -> pv1Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    fanoutContainer -> pv2Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    fanoutContainer -> pv3Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    fanoutContainer -> pv4Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    fanoutContainer -> pv5Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    fanoutContainer -> pv6Button -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuPvNameFilter
    fanoutContainer -> pv1 -> storeCommonDialog ((vuCommonDialogClass *)this);
    fanoutContainer -> pv2 -> storeCommonDialog ((vuCommonDialogClass *)this);
    fanoutContainer -> pv3 -> storeCommonDialog ((vuCommonDialogClass *)this);
    fanoutContainer -> pv4 -> storeCommonDialog ((vuCommonDialogClass *)this);
    fanoutContainer -> pv5 -> storeCommonDialog ((vuCommonDialogClass *)this);
    fanoutContainer -> pv6 -> storeCommonDialog ((vuCommonDialogClass *)this);

    // set the point type
    setPointType (VU_FANOUT);

    // get the title from the resource file, fanoutTitle must be
    // freed by the class destuctor
    vapplication *app = vapplication::GetCurrent ();
    vresource dialogTitles = vresourceGet (app -> GetResources (),
				vunames_vuDialogTitles);
    fanoutTitle = vresourceGetString (dialogTitles, 
				vunames_vuFanoutDialogTitle);
    SetTitle (fanoutTitle);

    // these are not needed on the fanout
    vuPvNameButtonClass *fwdButton = getFwdLinkButton ();
    filteredTextitem *fwdLink = getFwdLink ();
    fwdButton -> Disable ();
    fwdLink -> Disable ();
}


/*******************************************************************
*
*    vuFanoutDialogClass Destroy function
*
*******************************************************************/

void vuFanoutDialogClass::Destroy (void)
{
    vstrDestroy (fanoutTitle);
    vuCommonDialogClass::Destroy ();
}


/*******************************************************************
*
*    vuFanoutDialogClass, reset fields, overrides base class 
*	vuCommonDialogClass
*
*******************************************************************/

void vuFanoutDialogClass::resetDialog (void)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];

    vresource defRes = mainDialog -> getDefaultsResource ();

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0; 
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	// if not a common dialog field
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    dialogItem = vuFindContainerItem (fanoutContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {

		textItem = (vtextitem *)dialogItem;
		textItem -> SetTextScribed (vcharScribeSystem (""));
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		toggleItem -> SetValue (0);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		optionMenu -> SetValue (0);
	    }
	}  // end if not in common dialog
    }
    iterator.Finish ();

    // call base class reset function with proper point index
    vuCommonDialogClass::resetDialog ();

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuLoadDialog 
*
*******************************************************************/

void vuFanoutDialogClass::vuLoadDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*finalValue;

    char		pvName [PVNAME_SZ + 1];
    char		key [VU_KEY_SIZE + 1];
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dbaseType;	// codebase field type
    int			intValue;
    unsigned int	uIntValue;
    double		doubleValue;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (db -> vuPvmapTable, seekBuff);
    if (pntRecord == (DB_RECORD_PTR)0)
    {
	return;
    }
    else
    {
	int ptType;
	DB_GetFieldValue (pntRecord, db -> pvmapTypeField, dbValue, &dbaseType);
	ptType = vuGetPointTypeIndex ((char *)dbValue);
	// the point name in pvname field is NOT a FANOUT
	if (ptType != VU_FANOUT)
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTTYPEDIALOGERR,
				   pointName, ptType, VU_FANOUT);
	    vuClearPvNameField ();
	    return;
	}
	// the point name entered in pvname field has been deleted
	if (DB_IsDeleted (db -> vuPvmapTable))
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTISDELETEDERR, pointName);
	    vuClearPvNameField ();
	    return;
	}
    }
    // call base class vuLoadDialog first
    vuCommonDialogClass::vuLoadDialog (pointName);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    // only seek if table has changed or new xrefOpId
	    if ( (lastTable != xrefNode -> xrefTable) ||
		 (strlen (xrefNode -> xrefOpId) ) )
	    {
		if (strlen (xrefNode -> xrefOpId))
		{
		    sprintf (key, "%-*.*s%-s\0", 
				   PVNAME_SZ, PVNAME_SZ, pvName,
				   xrefNode -> xrefOpId);
		    pntRecord=  DB_Seek (xrefNode -> xrefTable, key);
		}
		else
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
		    pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    }
	    lastTable = xrefNode -> xrefTable;
	    dialogItem = vuFindContainerItem (fanoutContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
	    doubleValue = 0.0;
	    // GetField actually returns a number in the char array
	    if ( (dbaseType == 'F') || (dbaseType == 'N') )
	    {
		doubleValue = *((double *)dbValue);
	    }
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		    intValue = (int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeInt (intValue));
		    break;

		case DBF_ENUM:
		case DBF_USHORT:
		    uIntValue = (unsigned int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeUInt (uIntValue));
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		    if (dbaseType == 'L')
		    {
			int	logical;

			logical = (*((char*)dbValue) ? 
					xrefNode -> xrefTrue : 
					(xrefNode -> xrefTrue ^1));
			finalValue = vstrCloneScribed (vnumScribeInt (logical));
			break;
		    }
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_FWDLINK:
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		default:
		    continue;
	    }
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;
		textItem -> SetTextScribed (vcharScribeSystem (""));
		textItem -> SetText (finalValue);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		if (finalValue[0] == '0')
		    toggleItem -> SetValue (0);
		else
		    toggleItem -> SetValue (1);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		vuSetOptionMenuValue (optionMenu, finalValue);
		vstrDestroy (finalValue);
	    }
	}  // end if not in commondialog
    }
    iterator.Finish ();

    setDialogModify (vFALSE);
}


/*******************************************************************
*
*    vuSaveDialog 
*
*******************************************************************/

void vuFanoutDialogClass::vuSaveDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    const vchar 	*optionMenuValue;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*stringValue;

    char		pvName [PVNAME_SZ + 1];
    char		key [VU_KEY_SIZE + 1];
    DB_FIELD_PTR	putField;
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    double		doubleValue;
    int			doSave;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    if ( (fanoutContainer -> specifySelection -> GetValue ()) == vTRUE)
    {
	int selectedNumber = fanoutContainer -> selectionNumber -> GetTextAsInt ();
	if ( (selectedNumber < 1) || (selectedNumber > 6) )
	{
	    vuMessageDialogCenter (this, VU_FANOUT_SELN_ERR);
	    return;
	}
	filteredTextitem *pv;
	switch (selectedNumber)
	{
	    case 1:
		pv = fanoutContainer -> pv1;
		break;
	    case 2:
		pv = fanoutContainer -> pv2;
		break;
	    case 3:
		pv = fanoutContainer -> pv3;
		break;
	    case 4:
		pv = fanoutContainer -> pv4;
		break;
	    case 5:
		pv = fanoutContainer -> pv5;
		break;
	    case 6:
		pv = fanoutContainer -> pv6;
		break;
	}
	stringValue = pv -> GetTextAsString ();
	if ( (vcharCompare (stringValue, "")) == 0)
	{
	    vuMessageDialogCenter (this, VU_FANOUT_SELNPV_ERR, selectedNumber);
	    return;
	}
    }

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    doSave = vTRUE;
	    // only seek if table has changed or new xrefOpId
	    if ( (lastTable != xrefNode -> xrefTable) ||
		 (strlen (xrefNode -> xrefOpId) ) )
	    {
		if (strlen (xrefNode -> xrefOpId))
		{
		    sprintf (key, "%-*.*s%-s\0", 
				   PVNAME_SZ, PVNAME_SZ, pvName,
				   xrefNode -> xrefOpId);
		    pntRecord=  DB_Seek (xrefNode -> xrefTable, key);
		}
		else
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
		    pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    }
	    lastTable = xrefNode -> xrefTable;

	    putField = xrefNode -> xrefField;

	    dialogItem = vuFindContainerItem (fanoutContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		case DBF_ENUM:
		case DBF_USHORT:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    doubleValue = atof ((const char *)dbValue);
		    // if specify selection is false, no need to put a number
		    // in the selected number field
		    if ( (vcharCompareBounded (epics, "SELN", FLDNAME_SZ)) == 0)
		    {
			if ( (fanoutContainer -> specifySelection -> GetValue ()) == vFALSE)
			    doubleValue = 0;
		    }
		    DB_PutFieldValue (pntRecord, putField, (void *)&doubleValue);
		    doSave = vFALSE;
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		{
		    if (xrefNode -> xrefField -> field_type == 'L')
		    {
			int	logical;
			toggleItem = (vcontrolToggle *)dialogItem;
			logical = toggleItem -> GetValue ();
			*((char *)dbValue) = (logical ? xrefNode -> xrefTrue : 
								xrefNode -> xrefTrue ^1);
			break;
		    }
		    optionMenu = (vcontrolOptionMenu *)dialogItem;
		    optionMenuValue = vuGetOptionMenuValue (optionMenu);
		    vcharExport (vcharScribeSystemExport (optionMenuValue), dbValue);
		    break;
		}

		case DBF_FWDLINK:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		default:
		    continue;

	    }  // end xrefNode -> xrefEPICSFieldType

	    if (doSave == vTRUE)
	    {
		DB_PutFieldValue (pntRecord, putField, dbValue);
	    }

	}  // end if not in commondialog

    }  // end while iterator

    iterator.Finish ();

    // call base class vuSaveDialog
    vuCommonDialogClass::vuSaveDialog (pointName);
}


vkindTYPED_FULL_DEFINITION(vuFanoutContainerClass, vcontainer, "vuFanoutContainerClass");

/*******************************************************************
*
*    vuFanoutContainerClass override for constructor
*
*******************************************************************/

void vuFanoutContainerClass::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource pointDictRes = vresourceGet (app -> GetResources (), 
				vunames_vuPointDictionary);
    vresource realRes = vresourceGet (pointDictRes, vunames_vuFanoutContainer);

    vcontainer *pv1container = vcontainer::RequestSubLoad (this,
				vunames_vuFanoutLNK1Container);
    pv1Button = vuPvNameButtonClass::RequestSubLoad (pv1container,
				vunames_vuFanoutLNK1Button);
    pv1 = vuFilteredTextitem::RequestSubLoad (pv1container,
				vunames_LNK1);

    vcontainer *pv2container = vcontainer::RequestSubLoad (this,
				vunames_vuFanoutLNK2Container);
    pv2Button = vuPvNameButtonClass::RequestSubLoad (pv2container,
				vunames_vuFanoutLNK2Button);
    pv2 = vuFilteredTextitem::RequestSubLoad (pv2container,
				vunames_LNK2);

    vcontainer *pv3container = vcontainer::RequestSubLoad (this,
				vunames_vuFanoutLNK3Container);
    pv3Button = vuPvNameButtonClass::RequestSubLoad (pv3container,
				vunames_vuFanoutLNK3Button);
    pv3 = vuFilteredTextitem::RequestSubLoad (pv3container,
				vunames_LNK3);

    vcontainer *pv4container = vcontainer::RequestSubLoad (this,
				vunames_vuFanoutLNK4Container);
    pv4Button = vuPvNameButtonClass::RequestSubLoad (pv4container,
				vunames_vuFanoutLNK4Button);
    pv4 = vuFilteredTextitem::RequestSubLoad (pv4container,
				vunames_LNK4);

    vcontainer *pv5container = vcontainer::RequestSubLoad (this,
				vunames_vuFanoutLNK5Container);
    pv5Button = vuPvNameButtonClass::RequestSubLoad (pv5container,
				vunames_vuFanoutLNK5Button);
    pv5 = vuFilteredTextitem::RequestSubLoad (pv5container,
				vunames_LNK5);

    vcontainer *pv6container = vcontainer::RequestSubLoad (this,
				vunames_vuFanoutLNK6Container);
    pv6Button = vuPvNameButtonClass::RequestSubLoad (pv6container,
				vunames_vuFanoutLNK6Button);
    pv6 = vuFilteredTextitem::RequestSubLoad (pv6container,
				vunames_LNK6);

    onScan = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OSCN);

    specifySelection = vuCheckBoxClass::RequestSubLoad (this,
				vunames_SELM);

    vcontainer *selncontainer = vcontainer::RequestSubLoad (this,
				vunames_vuFanoutSELNContainer);
    selectionNumber = filteredTextitem::RequestSubLoad (selncontainer,
				vunames_SELN);

    vcontainer::LoadInit (realRes);

    // store ptr to text field for vuPvNameButtonClass
    pv1Button -> storePvName (pv1);
    pv2Button -> storePvName (pv2);
    pv3Button -> storePvName (pv3);
    pv4Button -> storePvName (pv4);
    pv5Button -> storePvName (pv5);
    pv6Button -> storePvName (pv6);

    // set filters
    selectionNumber -> SetItemFilterFunc (numericFilter);

    pv1 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *pv1Text = (filteredText *)pv1 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    pv1Text -> SetUserPtr ((void *)pv1);

    pv2 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *pv2Text = (filteredText *)pv2 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    pv2Text -> SetUserPtr ((void *)pv2);

    pv3 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *pv3Text = (filteredText *)pv3 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    pv3Text -> SetUserPtr ((void *)pv3);

    pv4 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *pv4Text = (filteredText *)pv4 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    pv4Text -> SetUserPtr ((void *)pv4);

    pv5 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *pv5Text = (filteredText *)pv5 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    pv5Text -> SetUserPtr ((void *)pv5);

    pv6 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *pv6Text = (filteredText *)pv6 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    pv6Text -> SetUserPtr ((void *)pv6);

}

/*******************************************************************
*
*    vuFanoutContainerClass override for destructor
*
*******************************************************************/

void vuFanoutContainerClass::Destroy (void)
{
    vcontainer::Destroy ();
}

