#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuLaunchCfgDialogClass.hxx"
#include "vuUtils.hxx"

// filter functions unique to lauch config
static int vuLaunchCfgKeyIdFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len);
static int vuLaunchCfgKeyFilter (filteredText *text, vtextSelection *sel,
				 const vchar **string, long *len);
static int vuLaunchCfgFileFilter (filteredText *text, vtextSelection *sel,
				  const vchar **string, long *len);
static int vuLaunchCfgParamsFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len);
static int vuLaunchCfgDescFilter (filteredText *text, vtextSelection *sel,
				  const vchar **string, long *len);


/*******************************************************************
*
*	LAUNCH CONFIG DIALOG CLASS
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuLaunchCfgDialogClass, vdialog, 
						"vuLaunchCfgDialogClass");

/*******************************************************************
*
*    vuLaunchCfgDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuLaunchCfgDialogClass::LoadInit(vresource res)
{
    dialogModified = vFALSE;

    selCfgDialog = NULL;

    keyId = vuLaunchCfgKeyIdFilteredTextitem::RequestSubLoad (this,
					vunames_vuLaunchCfgKeyId);

    vcontainer *nkcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuLaunchCfgNextKeyContainer);
    nextKey = filteredTextitem::RequestSubLoad (nkcontainer,
					vunames_vuLaunchCfgNextKey);

    vcontainer *pkcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuLaunchCfgPrevKeyContainer);
    prevKey = filteredTextitem::RequestSubLoad (pkcontainer,
					vunames_vuLaunchCfgPrevKey);

    vcontainer *flcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuLaunchCfgFileContainer);
    file = filteredTextitem::RequestSubLoad (flcontainer,
					vunames_vuLaunchCfgFile);

    vcontainer *pmcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuLaunchCfgParamsContainer);
    params = filteredTextitem::RequestSubLoad (pmcontainer,
					vunames_vuLaunchCfgParams);

    vcontainer *dccontainer = vcontainer::RequestSubLoad (this,
					vunames_vuLaunchCfgDescContainer);
    description = filteredTextitem::RequestSubLoad (dccontainer,
					vunames_vuLaunchCfgDesc);

    vcontainer *btcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuLaunchCfgButtonContainer);

    applyButton = vuLaunchCfgApplyButtonClass::RequestSubLoad (btcontainer,
					vunames_vuLaunchCfgApplyButton);

    dismissButton = vuLaunchCfgDismissButtonClass::RequestSubLoad (btcontainer,
					vunames_vuLaunchCfgDismissButton);

    deleteButton = vuLaunchCfgDeleteButtonClass::RequestSubLoad (btcontainer,
					vunames_vuLaunchCfgDeleteButton);

    newButton = vuLaunchCfgNewButtonClass::RequestSubLoad (btcontainer,
					vunames_vuLaunchCfgNewButton);

    gotoButton = vuLaunchCfgGoToButtonClass::RequestSubLoad (btcontainer,
					vunames_vuLaunchCfgGoToButton);

    nextButton = vuLaunchCfgNextButtonClass::RequestSubLoad (btcontainer,
					vunames_vuLaunchCfgNextButton);

    prevButton = vuLaunchCfgPrevButtonClass::RequestSubLoad (btcontainer,
					vunames_vuLaunchCfgPrevButton);

    // call base class loadInit to finish the job
    vdialog::LoadInit(res);

    // store ptr to dialog in buttons
    applyButton -> storeLaunchCfgDialog (this);
    dismissButton -> storeLaunchCfgDialog (this);
    deleteButton -> storeLaunchCfgDialog (this);
    newButton -> storeLaunchCfgDialog (this);
    gotoButton -> storeLaunchCfgDialog (this);
    nextButton -> storeLaunchCfgDialog (this);
    prevButton -> storeLaunchCfgDialog (this);

    // store ptr to dialog
    keyId -> storeLaunchCfgDialog (this);

    // set filters
    keyId -> SetItemFilterFunc (vuLaunchCfgKeyIdFilter);
    filteredText *idtext = (filteredText *)keyId -> GetTextData ();
    // remember ptr to textitem for later use
    idtext -> SetUserPtr ((void *)keyId);

    nextKey -> SetItemFilterFunc (vuLaunchCfgKeyFilter);
    prevKey -> SetItemFilterFunc (vuLaunchCfgKeyFilter);
    file -> SetItemFilterFunc (vuLaunchCfgFileFilter);
    params -> SetItemFilterFunc (vuLaunchCfgParamsFilter);
    description -> SetItemFilterFunc (vuLaunchCfgDescFilter);

    // store dialogModified ptr
    keyId -> SetTextModifiedPtr (&this -> dialogModified);
    nextKey -> SetTextModifiedPtr (&this -> dialogModified);
    prevKey -> SetTextModifiedPtr (&this -> dialogModified);
    file -> SetTextModifiedPtr (&this -> dialogModified);
    params -> SetTextModifiedPtr (&this -> dialogModified);
    description -> SetTextModifiedPtr (&this -> dialogModified);

    // place window in center of screen
    Place(GetRoot(),vrectPLACE_CENTER,vrectPLACE_CENTER);
    if(vuReadOnly == vTRUE)
    {
	applyButton->Disable();
	deleteButton->Disable();
    }
}


/*******************************************************************
*
*    vuLaunchCfgDialogClass notify function
*
*******************************************************************/

void vuLaunchCfgDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	int result = VU_TRUE;
	if (dialogModified == vTRUE)
	{
	    result = vuMessageDialogCenter (this, VU_MOD_DISMISS_MSG);
	}
	if (result == VU_TRUE)
	{
	    Close();
	    vevent::StopProcessing ();
	}
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuLaunchCfgDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuLaunchCfgDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }

    if (selCfgDialog != NULL)
    {
	delete selCfgDialog;
    }

    vdialog::Destroy();
}


/*******************************************************************
*
*    vuLaunchCfgDialogClass map database fields function
*
*******************************************************************/

void vuLaunchCfgDialogClass::vuMapFields (void)
{
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();

    keyField = DB_MapField (db -> vuLaunchCfgTable, "KEY");
    nextKeyField = DB_MapField (db -> vuLaunchCfgTable, "NEXT");
    prevKeyField = DB_MapField (db -> vuLaunchCfgTable, "PREV");
    fileField = DB_MapField (db -> vuLaunchCfgTable, "FILE");
    paramsField = DB_MapField (db -> vuLaunchCfgTable, "PARAM");
    descField = DB_MapField (db -> vuLaunchCfgTable, "DESC");
    DB_UseKey (db -> vuLaunchCfgTable, "TAG");

}


/*******************************************************************
*
*    vuLaunchCfgDialogClass Open function
*
*******************************************************************/

void vuLaunchCfgDialogClass::vuLaunchCfgOpen(vuMainDialogClass *mainDialg)
{
    mainDialog = mainDialg;

    vuMapFields ();

    vuNewConfig ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuLaunchCfgDialogClass new config  function
*
*******************************************************************/

void vuLaunchCfgDialogClass::vuNewConfig(void)
{
    keyId -> SetTextScribed (vcharScribeSystem (""));
    nextKey -> SetTextScribed (vcharScribeSystem (""));
    prevKey -> SetTextScribed (vcharScribeSystem (""));
    file -> SetTextScribed (vcharScribeSystem (""));
    params -> SetTextScribed (vcharScribeSystem (""));
    description -> SetTextScribed (vcharScribeSystem (""));

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuLaunchCfgDialogClass load config function
*
*******************************************************************/

void vuLaunchCfgDialogClass::vuLoadConfig (vchar *keyidstr)
{
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();
    DB_RECORD_PTR rec;
    char keyid[VU_LAUNCHCFG_KEY_SZ + 1];

    vcharExportBounded (vcharScribeSystemExport (keyidstr),
			keyid,
			VU_LAUNCHCFG_KEY_SZ);
    keyid[VU_LAUNCHCFG_KEY_SZ] = 0;

    if ( (rec = DB_Seek (db -> vuLaunchCfgTable, keyid)) != 0)
    {
	char		text[VU_LAUNCHCFG_PARAM_SZ + 10];
	int		type;

	// keyid field
	keyId -> SetTextScribed (vcharScribeSystem (keyid));

	// next key field
	DB_GetFieldValue (rec, nextKeyField, text, &type);
	nextKey -> SetTextScribed (vcharScribeSystem (text));

	// prev key field
	DB_GetFieldValue (rec, prevKeyField, text, &type);
	prevKey -> SetTextScribed (vcharScribeSystem (text));

	// file field
	DB_GetFieldValue (rec, fileField, text, &type);
	file -> SetTextScribed (vcharScribeSystem (text));

	// file field
	DB_GetFieldValue (rec, paramsField, text, &type);
	params -> SetTextScribed (vcharScribeSystem (text));

	// description field
	DB_GetFieldValue (rec, descField, text, &type);
	description -> SetTextScribed (vcharScribeSystem (text));
    }

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuLaunchCfgApplyButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuLaunchCfgApplyButtonClass, vbutton, 
					"vuLaunchCfgApplyButtonClass");

void vuLaunchCfgApplyButtonClass::ObserveDialogItem (vevent *event)
{
    vstr *dialogText;
    dialogText = launchCfgDialog -> keyId -> GetTextAsString ();
    if ( (vcharLength (dialogText)) == 0)
    {
	vuMessageDialogCenter (launchCfgDialog, VU_LAUNCHCFG_NO_KEY_ERR);
	vstrDestroy (dialogText);
	return;
    }
    char keyid[VU_LAUNCHCFG_KEY_SZ + 1];
    vcharExportBounded (vcharScribeSystemExport (dialogText),
			keyid,
			VU_LAUNCHCFG_KEY_SZ);
    keyid[VU_LAUNCHCFG_KEY_SZ] = 0;

    vuMainDialogClass *mainDialog = launchCfgDialog -> getMainDialog ();
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();
    DB_RECORD_PTR rec;

    int result;
    if ( (rec = DB_Seek (db -> vuLaunchCfgTable, keyid)) == 0)
    {
	result = vuMessageDialogCenter (launchCfgDialog, 
					VU_LAUNCHCFG_NEW_REQ_MSG, 
					dialogText);
	if (result == VU_FALSE)
	{
	    vstrDestroy (dialogText);
	    return;
	}
	rec = (DB_RECORD_PTR)DB_NewRecord (db -> vuLaunchCfgTable, NULL);
	mainDialog -> vuMainMessageCenter (vunames_vuLaunchCfgNewMsg, dialogText);
	vuSelCfgDialogClass *selCfgDialog = launchCfgDialog -> getSelCfgDialog ();
	// add to selection list if it exists
	if (selCfgDialog != NULL)
	{
	    vliststr *list = (vliststr *)(selCfgDialog -> selCfgList -> GetList ());
	    int rowCount = list -> GetRowCount ();
	    list -> SetCellValue (rowCount, 0, (const void *)dialogText);
	    list -> Sort (vlistROW, 0, NULL);
	}
    }
    else
    {
	result = vuMessageDialogCenter (launchCfgDialog, 
					VU_LAUNCHCFG_MOD_REQ_MSG, 
					dialogText);
	if (result == VU_FALSE)
	{
	    vstrDestroy (dialogText);
	    return;
	}
	mainDialog -> vuMainMessageCenter (vunames_vuLaunchCfgModMsg, dialogText);
    }

    vstrDestroy (dialogText);

    char text[VU_LAUNCHCFG_PARAM_SZ + 10];

    // save key id to database
    DB_PutFieldValue (rec, launchCfgDialog -> keyField, keyid);

    // save next key to database
    dialogText = launchCfgDialog -> nextKey -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (dialogText),
			text,
			VU_LAUNCHCFG_KEY_SZ);
    DB_PutFieldValue (rec, launchCfgDialog -> nextKeyField, text);
    vstrDestroy (dialogText);

    // save prev key to database
    dialogText = launchCfgDialog -> prevKey -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (dialogText),
			text,
			VU_LAUNCHCFG_KEY_SZ);
    DB_PutFieldValue (rec, launchCfgDialog -> prevKeyField, text);
    vstrDestroy (dialogText);

    // save file to database
    dialogText = launchCfgDialog -> file -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (dialogText),
			text,
			VU_LAUNCHCFG_FILE_SZ);
    text[VU_LAUNCHCFG_FILE_SZ] = 0;
    DB_PutFieldValue (rec, launchCfgDialog -> fileField, text);
    vstrDestroy (dialogText);

    // save params to database
    dialogText = launchCfgDialog -> params -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (dialogText),
			text,
			VU_LAUNCHCFG_PARAM_SZ);
    text[VU_LAUNCHCFG_PARAM_SZ] = 0;
    DB_PutFieldValue (rec, launchCfgDialog -> paramsField, text);
    vstrDestroy (dialogText);

    // save description to database
    dialogText = launchCfgDialog -> description -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (dialogText),
			text,
			VU_LAUNCHCFG_DESC_SZ);
    text[VU_LAUNCHCFG_DESC_SZ] = 0;
    DB_PutFieldValue (rec, launchCfgDialog -> descField, text);
    vstrDestroy (dialogText);

    launchCfgDialog -> setDialogModify (vFALSE);
}


/*******************************************************************
*
*    vuLaunchCfgDismissButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuLaunchCfgDismissButtonClass, vbutton, 
					"vuLaunchCfgDismissButtonClass");

void vuLaunchCfgDismissButtonClass::ObserveDialogItem (vevent *event)
{
    int modified = launchCfgDialog -> getDialogModify ();
    int result = VU_TRUE;

    if (modified == vTRUE)
    {
	result = vuMessageDialogCenter (launchCfgDialog, VU_MOD_DISMISS_MSG);
    }
    if (result == VU_TRUE)
    {
	launchCfgDialog -> Close ();
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    vuLaunchCfgDeleteButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuLaunchCfgDeleteButtonClass, vbutton, 
					"vuLaunchCfgDeleteButtonClass");

void vuLaunchCfgDeleteButtonClass::ObserveDialogItem (vevent *event)
{
    vstr *keyidstr = launchCfgDialog -> keyId -> GetTextAsString ();
    if ( (vcharLength (keyidstr)) == 0)
    {
	vuMessageDialogCenter (launchCfgDialog, VU_LAUNCHCFG_NO_KEY_ERR);
	vstrDestroy (keyidstr);
	return;
    }
    int result = vuMessageDialogCenter (launchCfgDialog,
					VU_LAUNCHCFG_DEL_REQ_MSG,
					keyidstr);
    if (result == VU_TRUE)
    {
	char keyid[VU_LAUNCHCFG_KEY_SZ + 1];
	vcharExportBounded (vcharScribeSystemExport (keyidstr),
			    keyid,
			    VU_LAUNCHCFG_KEY_SZ);
	keyid[VU_LAUNCHCFG_KEY_SZ] = 0;

	vuMainDialogClass *mainDialog = launchCfgDialog -> getMainDialog ();
	dataVuApplication *vuApp = mainDialog -> getApplication ();
	vuDatabaseClass *db = vuApp -> getDatabaseClass ();
	DB_RECORD_PTR rec;

	if ( (rec = DB_Seek (db -> vuLaunchCfgTable, keyid)) != 0)
	{
	    DB_DeleteRecord (db -> vuLaunchCfgTable, rec);
	    // log this with main
	    mainDialog -> vuMainMessageCenter (vunames_vuLaunchCfgDelMsg,
					       keyidstr);

	    // update key id list from config select dialog, because its list
	    // is used for the Next and Prev button options

	    vuSelCfgDialogClass *selCfgDialog = launchCfgDialog -> getSelCfgDialog ();
	    if (selCfgDialog != NULL)
	    {
		vliststr *list = (vliststr *)(selCfgDialog -> selCfgList -> GetList ());
		vlistIterator listIterator;
		int row;
		vchar *listKeyId;
		listIterator.Start ((vlist *)list);
		while (listIterator.Next ())
		{
		    row = listIterator.GetRow ();
		    listKeyId = (vchar *)(list -> GetCellValue (row, 0));
		    if ( (vcharCompareBounded (listKeyId, 
					       keyidstr, 
					       VU_LAUNCHCFG_KEY_SZ)) == 0)
		    {
			list -> RemoveRow (row);
			break;
		    }
		}
		listIterator.Finish ();
	    }
	}
	else
	{
	    vuMessageDialogCenter (launchCfgDialog, 
				   VU_LAUNCHCFG_NOT_EXIST_MSG, 
				   keyidstr);
	}
    }
    vstrDestroy (keyidstr);
}


/*******************************************************************
*
*    vuLaunchCfgNewButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuLaunchCfgNewButtonClass, vbutton, 
					"vuLaunchCfgNewButtonClass");

void vuLaunchCfgNewButtonClass::ObserveDialogItem (vevent *event)
{
    launchCfgDialog -> vuNewConfig ();
}


/*******************************************************************
*
*    vuLaunchCfgGoToButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuLaunchCfgGoToButtonClass, vbutton, 
					"vuLaunchCfgGoToButtonClass");

void vuLaunchCfgGoToButtonClass::ObserveDialogItem (vevent *event)
{
    vuSelCfgDialogClass *selCfgDialog = launchCfgDialog -> getSelCfgDialog ();
    if (selCfgDialog == NULL)
    {
	vuMainDialogClass *mainDialog = launchCfgDialog -> getMainDialog ();
	dataVuApplication *vuApp = mainDialog -> getApplication ();
	vresource vuRes = vuApp -> getDataVuvresource ();
	vresource selcfgRes = vresourceGet (vuRes, vunames_vuSelectCfgDialog);
	selCfgDialog = new vuSelCfgDialogClass (selcfgRes);
	launchCfgDialog -> storeSelCfgDialog (selCfgDialog);
    }
    selCfgDialog -> vuSelCfgOpen (launchCfgDialog);
    vchar *keyid = selCfgDialog -> vuGetKeyId ();
    if (keyid[0] != 0)
    {
	launchCfgDialog -> vuLoadConfig (keyid);
    }
}


/*******************************************************************
*
*    vuLaunchCfgNextButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuLaunchCfgNextButtonClass, vbutton, 
					"vuLaunchCfgNextButtonClass");

void vuLaunchCfgNextButtonClass::ObserveDialogItem (vevent *event)
{
    vuSelCfgDialogClass *selCfgDialog = launchCfgDialog -> getSelCfgDialog ();
    if (selCfgDialog == NULL)
    {
	vuMessageDialogCenter (launchCfgDialog, VU_NO_LIST_ERR);
	return;
    }
    vliststr *list = (vliststr *)(selCfgDialog -> selCfgList -> GetList ());
    vlistSelection *selection = selCfgDialog -> selCfgList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row != vlistNO_CELLS_SELECTED)
    {
	int numRows = list -> GetRowCount ();
	if ( (row + 1) < numRows)
	{
	    if ( (launchCfgDialog -> getDialogModify ()) == vTRUE)
	    {
		int result = vuMessageDialogCenter (launchCfgDialog, 
						    VU_MOD_DISMISS_MSG);
		if (result == VU_FALSE)
		{
		    return;
		}
	    }
	    list -> UnselectRow (selection, row);
	    row++;
	    list -> SelectRow (selection, row);
	    vchar *keyid = (vchar *)(list -> GetCellValue (row, 0));
	    launchCfgDialog -> vuLoadConfig (keyid);
	}
	else
	{
	    vuMessageDialogCenter (launchCfgDialog, VU_ENDOFLIST_MSG);
	}
    }
    else
    {
	vuMessageDialogCenter (launchCfgDialog, VU_LAUNCHCFG_NO_SELECTED_MSG);
    }
}


/*******************************************************************
*
*    vuLaunchCfgPrevButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuLaunchCfgPrevButtonClass, vbutton, 
					"vuLaunchCfgPrevButtonClass");

void vuLaunchCfgPrevButtonClass::ObserveDialogItem (vevent *event)
{
    vuSelCfgDialogClass *selCfgDialog = launchCfgDialog -> getSelCfgDialog ();
    if (selCfgDialog == NULL)
    {
	vuMessageDialogCenter (launchCfgDialog, VU_NO_LIST_ERR);
	return;
    }
    vliststr *list = (vliststr *)(selCfgDialog -> selCfgList -> GetList ());
    vlistSelection *selection = selCfgDialog -> selCfgList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row != vlistNO_CELLS_SELECTED)
    {
	if ( (row - 1) >= 0)
	{
	    if ( (launchCfgDialog -> getDialogModify ()) == vTRUE)
	    {
		int result = vuMessageDialogCenter (launchCfgDialog, VU_MOD_DISMISS_MSG);
		if (result == VU_FALSE)
		{
		    return;
		}
	    }
	    list -> UnselectRow (selection, row);
	    row--;
	    list -> SelectRow (selection, row);
	    vchar *scriptId = (vchar *)(list -> GetCellValue (row, 0));
	    launchCfgDialog -> vuLoadConfig (scriptId);
	}
	else
	{
	    vuMessageDialogCenter (launchCfgDialog, VU_STARTOFLIST_MSG);
	}
    }
    else
    {
	vuMessageDialogCenter (launchCfgDialog, VU_LAUNCHCFG_NO_SELECTED_MSG);
    }
}


/*******************************************************************
*
*	SELECT CONFIG DIALOG CLASS
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuSelCfgDialogClass, vdialog, 
						"vuSelSCrDialogClass");

/*******************************************************************
*
*    vuSelCfgDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuSelCfgDialogClass::LoadInit(vresource res)
{
    selCfgList = vuSelCfgListClass::RequestSubLoad (this,
					vunames_vuSelCfgList);

    applyButton = vuSelCfgApplyButtonClass::RequestSubLoad (this,
					vunames_vuSelCfgApplyButton);

    dismissButton = vuSelCfgDismissButtonClass::RequestSubLoad (this,
					vunames_vuSelCfgDismissButton);

    // call base class loadInit to finish the job
    vdialog::LoadInit(res);

    selCfgList -> storeSelCfgDialog (this);
    applyButton -> storeSelCfgDialog (this);
    dismissButton -> storeSelCfgDialog (this);

}


/*******************************************************************
*
*    vuSelCfgDialogClass notify function
*
*******************************************************************/

void vuSelCfgDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	// zap keyId so caller will act on NULL string
	vmemSet ((void *)keyId, 0, VU_LAUNCHCFG_KEY_SZ);
	Close();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuSelCfgDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuSelCfgDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    vliststr *list = (vliststr *)(selCfgList -> GetList ());
    selCfgList -> SetList (NULL);
    list -> Destroy ();

    vdialog::Destroy();
}



/*******************************************************************
*
*    vuSelCfgDialogClass Open function
*
*******************************************************************/

void vuSelCfgDialogClass::vuSelCfgOpen(vuLaunchCfgDialogClass *launchCfgDlg)
{
    launchCfgDialog = launchCfgDlg;
    vuMainDialogClass *mainDialog = launchCfgDialog -> getMainDialog ();
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();


    vliststr *list = (vliststr *)(selCfgList -> GetList ());

    // destroy old list if it exists
    if (list != NULL)
    {
	selCfgList -> SetList (NULL);
	list -> Destroy ();
    }

    // new list to populate
    vliststr *newList = new vliststr;

    int type;
    int row = 0;
    char keyIdName[VU_LAUNCHCFG_KEY_SZ + 1];

    DB_RECORD_PTR rec = DB_Top (db -> vuLaunchCfgTable);
    while (rec)
    {
	DB_GetFieldValue (rec, launchCfgDialog -> keyField, keyIdName, &type);
	keyIdName[VU_LAUNCHCFG_KEY_SZ] = 0;
	newList -> SetCellValueScribed (row, 0, vcharScribeSystem (keyIdName));
	row++;
	rec = DB_SeekNext (db -> vuLaunchCfgTable);
    }

    newList -> Sort (vlistROW, 0, NULL);
    selCfgList -> SetList (newList);
    vlistSelection *listSelection = selCfgList -> GetSelection ();
    // only one selection at a time
    newList -> SetSelectMethod (listSelection, vlistONE_ONLY_SELECTION);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    // place window in center of screen
    Place(launchCfgDialog, vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuSelCfgApplyButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuSelCfgApplyButtonClass, vbutton, 
					"vuSelCfgApplyButtonClass");

void vuSelCfgApplyButtonClass::ObserveDialogItem (vevent *event)
{
    vchar *keyid = selCfgDialog -> vuGetKeyId ();
    int row, column;
    const vchar *itemValue;

    vliststr *list = (vliststr *)selCfgDialog -> selCfgList -> GetList ();
    vlistSelection *listSelection = selCfgDialog -> selCfgList -> GetSelection ();
    list -> GetSelectedCell (listSelection, &row, &column);
    if (row != vlistNO_CELLS_SELECTED)
    {
	itemValue = (const vchar *)(list -> GetCellValue (row, column));
	// copy value to script dialog
	vcharCopyBounded (itemValue, keyid, VU_LAUNCHCFG_KEY_SZ);
    }
    // no item selected, then set to zip, so caller 
    // processes NULL string
    else
    {
	vmemSet ((void *)keyid, 0, VU_LAUNCHCFG_KEY_SZ);
    }
    // close dialog and stop sub event loop
    selCfgDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuSelCfgDismissButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuSelCfgDismissButtonClass, vbutton, 
					"vuSelCfgDismissButtonClass");

void vuSelCfgDismissButtonClass::ObserveDialogItem (vevent *event)
{
    vchar *keyid = selCfgDialog -> vuGetKeyId ();
    // zap it so caller of dialog will act on a NULL string
    vmemSet ((void *)keyid, 0, VU_LAUNCHCFG_KEY_SZ);
    selCfgDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*	CONFIG SELECT LIST DIALOG CLASS
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuSelCfgListClass, vlistview,
					"vuSelCfgListClass")

/******************************************************************
*
*    vuSelCfgListClass observe function
*	SELECTIN_HAS_CHANGED = user has release the mouse button
*	and commits the new selection, if we have a double mouse
*	click, then take same action as APPLY BUTTON
*
******************************************************************/

void vuSelCfgListClass::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	int clickCount = GetLastPointerClickCount ();
	if (clickCount == 2)
	{
	    vchar *keyid = selCfgDialog -> vuGetKeyId ();
	    int row, column;
	    const vchar *itemValue;

	    vliststr *list = (vliststr *)GetList ();
	    vlistSelection *listSelection = GetSelection ();
	    list -> GetSelectedCell (listSelection, &row, &column);
	    itemValue = (const vchar *)(list -> GetCellValue (row, column));
	    // copy value to script dialog
	    vcharCopyBounded (itemValue, keyid, VU_LAUNCHCFG_KEY_SZ);
	    // close dialogand stop sub event loop
	    selCfgDialog -> Close ();
	    vevent::StopProcessing ();
	}
    }
}


/**********************************************************************
*
*    class that overrides filteredTextitem class
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(vuLaunchCfgKeyIdFilteredTextitem, filteredTextitem, 
					"vuLaunchCfgKeyIdFilteredTextitem");

/**********************************************************************
*
*    vuLaunchCfgKeyIdFilteredTextitem observe view override
*	we are looking for a loss of focus on the script id field
*
**********************************************************************/

void vuLaunchCfgKeyIdFilteredTextitem::ObserveView (vevent *event, int message)
{    
    if (message == vtextviewACCEPT_NOTIFY)
    {
	vdialogItemList	*list = this -> GetItemListIn ();
	vobjectAttributed *obj = list -> GetFocusItem ();
	if ( ((vobjectAttributed *)this) != obj)
	{
	    vdialogItemList *list = GetItemListIn ();
	    if (list != NULL)
	    {
		vdialog *dialog = list -> DetermineDialog ();
		if (dialog != NULL)
		{
		    vchar keyid [VU_LAUNCHCFG_KEY_SZ + 1];
		    GetText (keyid, VU_LAUNCHCFG_KEY_SZ);
		    if (keyid[0] != 0)
		    {
			vuLaunchCfgDialogClass *launchCfgDlg = 
					(vuLaunchCfgDialogClass *)dialog;
			launchCfgDlg -> vuLoadConfig (keyid);
		    }
		}
	    }
	}
    }
}


/**********************************************************************
*
*	filter functions unique to the launch config editor
*
**********************************************************************/

static int vuLaunchCfgKeyIdFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len)
{
    int		ret;
    vchar	aSpace;

    if (*len > 0)
    {
	aSpace = vcharFromLiteral (' ');
	if (**string == aSpace)
	{
	    long textLen = text -> GetLength ();
	    if (textLen != 0)
	    {
		vwindow::Beep ();
		ret = vFALSE;
	    }
	    else
	    {
		vuLaunchCfgKeyIdFilteredTextitem *textItem = 
			(vuLaunchCfgKeyIdFilteredTextitem *)text -> GetUserPtr ();
		if (textItem != NULL)
		{
		    int ok = vFALSE;
		    vuLaunchCfgDialogClass *launchCfgDialog = 
						textItem -> getLaunchCfgDialog ();
		    if (launchCfgDialog != NULL)
		    {
    			vuSelCfgDialogClass *selCfgDialog = 
						launchCfgDialog -> getSelCfgDialog ();
			if (selCfgDialog == NULL)
			{
			    vuMainDialogClass *mainDialog = 
						launchCfgDialog -> getMainDialog ();
			    dataVuApplication *vuApp = mainDialog -> getApplication ();
			    vresource vuRes = vuApp -> getDataVuvresource ();
			    vresource selcfgRes = vresourceGet (vuRes, 
							vunames_vuSelectCfgDialog);
			    selCfgDialog = new vuSelCfgDialogClass (selcfgRes);
			    launchCfgDialog -> storeSelCfgDialog (selCfgDialog);
			}
			selCfgDialog -> vuSelCfgOpen (launchCfgDialog);
			vchar *keyid = selCfgDialog -> vuGetKeyId ();
			vchar *newString = ((vchar *)(*string));
			if (keyid[0] != 0)
			{
			    *len = vcharLength (keyid);
			    vcharCopyBounded (keyid, newString, VU_LAUNCHCFG_KEY_SZ);
			    ret = NCharFilter (text, sel, string, len, 
					       VU_LAUNCHCFG_KEY_SZ - 1);
			    ok = vTRUE;
			}
			else
			{
			    ok = vFALSE;
			}
		    }
		    if (ok == vFALSE)
		    {
			vwindow::Beep ();
			ret = vFALSE;
		    }
		}
		else
		{
		    vwindow::Beep ();
		    ret = vFALSE;
		}
	    }
	}
	else
	{
	    ret = NCharFilter (text, sel, string, len, VU_LAUNCHCFG_KEY_SZ - 1);
	}
    }
    else
    {
	vwindow::Beep ();
	ret = vFALSE;
    }

    return (ret);
}


static int vuLaunchCfgKeyFilter (filteredText *text, vtextSelection *sel,
				 const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, VU_LAUNCHCFG_KEY_SZ) );
}

static int vuLaunchCfgFileFilter (filteredText *text, vtextSelection *sel,
				  const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, VU_LAUNCHCFG_FILE_SZ) );
}

static int vuLaunchCfgDescFilter (filteredText *text, vtextSelection *sel,
				  const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, VU_LAUNCHCFG_PARAM_SZ - 1) );
}

static int vuLaunchCfgParamsFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, VU_LAUNCHCFG_DESC_SZ - 1) );
}

