#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbFldTypes.h"
#include "choiceGbl.h"
#include "choiceLdfw2.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "dbRecDes.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuCommonDialogClass.hxx"
#include "vuLdfwDialogClass.hxx"
#include "vuNavigatorDialogClass.hxx"
#include "vuUtils.hxx"


vkindTYPED_FULL_DEFINITION(vuLdfw2DialogClass, vuCommonDialogClass, "vuLdfw2DialogClass");

/*******************************************************************
*
*    vuLdfw2DialogClass override for constructors init ()
*
*******************************************************************/

void vuLdfw2DialogClass::LoadInit (vresource res)
{
    int	i;

    // load in the ldfw2 specific fields
    ldfw2Container = vuLdfw2ContainerClass::RequestSubLoad (this,
				vunames_vuCommonContainer);
    vuCommonDialogClass::LoadInit (res);

    // store dialog modified flag
    ldfw2Container -> speedCtrl -> SetTextModifiedPtr (&this -> dialogModified);
    ldfw2Container -> procSpeedCtrl -> vuSetModifiedPtr (&this -> dialogModified);
    ldfw2Container -> msSpeedCtrl -> vuSetModifiedPtr (&this -> dialogModified);
    ldfw2Container -> onScan -> vuSetModifiedPtr (&this -> dialogModified);
    ldfw2Container -> onAlarm -> vuSetModifiedPtr (&this -> dialogModified);
    ldfw2Container -> manual -> vuSetModifiedPtr (&this -> dialogModified);
    ldfw2Container -> ldfw1 -> SetTextModifiedPtr (&this -> dialogModified);
    ldfw2Container -> procLdfw1 -> vuSetModifiedPtr (&this -> dialogModified);
    ldfw2Container -> msLdfw1 -> vuSetModifiedPtr (&this -> dialogModified);
    ldfw2Container -> precision -> SetTextModifiedPtr (&this -> dialogModified);
    ldfw2Container -> pumpSelMode -> vuSetModifiedPtr (&this -> dialogModified);
    ldfw2Container -> pumpStartSpeed -> SetTextModifiedPtr (&this -> dialogModified);
    ldfw2Container -> pumpStopSpeed -> SetTextModifiedPtr (&this -> dialogModified);
    ldfw2Container -> monitorDeadband -> SetTextModifiedPtr (&this -> dialogModified);
    ldfw2Container -> alarmGroup -> SetTextModifiedPtr (&this -> dialogModified);
    ldfw2Container -> alarmDisable -> SetTextModifiedPtr (&this -> dialogModified);

    for (i = 0; i < VU_NUM_LDFW_PUMPS; i++)
    {
	ldfw2Container -> adjFactors[i] -> SetTextModifiedPtr (&this -> dialogModified);
    }

    // store dialog ptr for vuPvNameButtonClass
    ldfw2Container -> speedCtrlButton -> storeCommonDialog ((vuCommonDialogClass *)this);
    ldfw2Container -> ldfw1Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ldfw2Container -> alarmDisableButton -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuPvNameFilter
    ldfw2Container -> speedCtrl -> storeCommonDialog ((vuCommonDialogClass *)this);
    ldfw2Container -> ldfw1 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ldfw2Container -> alarmDisable -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuAlarmGroupFilter
    ldfw2Container -> alarmGroup -> storeCommonDialog ((vuCommonDialogClass *)this);

    // set the point type
    setPointType (VU_LDFW2);

    // get the title from the resource file, aiTitle must be
    // freed by the class destuctor
    vapplication *app = vapplication::GetCurrent ();
    vresource dialogTitles = vresourceGet (app -> GetResources (),
				vunames_vuDialogTitles);
    ldfw2Title = vresourceGetString (dialogTitles, 
				vunames_vuLdfw2DialogTitle);
    SetTitle (ldfw2Title);
}


/*******************************************************************
*
*    vuLdfw2DialogClass Destroy function
*
*******************************************************************/

void vuLdfw2DialogClass::Destroy (void)
{
    vstrDestroy (ldfw2Title);
    vuCommonDialogClass::Destroy ();
}


/*******************************************************************
*
*    vuLdfw2DialogClass, reset fields, overrides base class 
*	vuCommonDialogClass
*
*******************************************************************/

void vuLdfw2DialogClass::resetDialog (void)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];

    vresource defRes = mainDialog -> getDefaultsResource ();

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0; 
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	// if not a common dialog field
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    dialogItem = vuFindContainerItem (ldfw2Container, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;

		// set the default value for the suspend value field
		if ( (vcharCompareBounded (epics, "PREC", FLDNAME_SZ)) == 0)
		{
		    vstr *prec = vresourceGetString (defRes, 
						vunames_vuPRECdefault);
		    textItem -> SetText (prec);
		    vstrDestroy (prec);
		}
		else
		{
		    textItem -> SetTextScribed (vcharScribeSystem (""));
		}
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		toggleItem -> SetValue (0);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		optionMenu -> SetValue (0);
	    }
	}  // end if not in common dialog
    }
    iterator.Finish ();

    // call base class reset function with proper point index
    vuCommonDialogClass::resetDialog ();

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuLoadDialog 
*
*******************************************************************/

void vuLdfw2DialogClass::vuLoadDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*finalValue;

    char		pvName [PVNAME_SZ + 1];
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dbaseType;	// codebase field type
    int			intValue;
    unsigned int	uIntValue;
    long		longValue;
    unsigned long	uLongValue;
    double		doubleValue;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (db -> vuPvmapTable, seekBuff);
    if (pntRecord == (DB_RECORD_PTR)0)
    {
	return;
    }
    else
    {
	int ptType;
	DB_GetFieldValue (pntRecord, db -> pvmapTypeField, dbValue, &dbaseType);
	ptType = vuGetPointTypeIndex ((char *)dbValue);
	// the point name in pvname field is NOT an LDFW2
	if (ptType != VU_LDFW2)
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTTYPEDIALOGERR,
				   pointName, ptType, VU_LDFW2);
	    vuClearPvNameField ();
	    return;
	}
	// the point name entered in pvname field has been deleted
	if (DB_IsDeleted (db -> vuPvmapTable))
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTISDELETEDERR, pointName);
	    vuClearPvNameField ();
	    return;
	}
    }
    // call base class vuLoadDialog first
    vuCommonDialogClass::vuLoadDialog (pointName);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;
	    dialogItem = vuFindContainerItem (ldfw2Container, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
	    doubleValue = 0.0;
	    // GetField actually returns a number in the char array
	    if ( (dbaseType == 'F') || (dbaseType == 'N') )
	    {
		doubleValue = *((double *)dbValue);
	    }
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		    intValue = (int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeInt (intValue));
		    break;

		case DBF_LONG:
		    longValue = (long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeLong (longValue));
		    break;

		case DBF_ENUM:
		case DBF_USHORT:
		    uIntValue = (unsigned int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeUInt (uIntValue));
		    break;

		case DBF_ULONG:
		    uLongValue = (unsigned long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeULong (uLongValue));
		    break;

		case DBF_FLOAT:
		case DBF_DOUBLE:
		    finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
		    break;

		case DBF_STRING:
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		    if (dbaseType == 'L')
		    {
			int	logical;

			logical = (*((char*)dbValue) ? 
					xrefNode -> xrefTrue : 
					(xrefNode -> xrefTrue ^1));
			finalValue = vstrCloneScribed (vnumScribeInt (logical));
			break;
		    }
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_INLINK:
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		default:
		    continue;
	    }
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;
		textItem -> SetTextScribed (vcharScribeSystem (""));
		textItem -> SetText (finalValue);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		if (finalValue[0] == '0')
		    toggleItem -> SetValue (0);
		else
		    toggleItem -> SetValue (1);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		vuSetOptionMenuValue (optionMenu, finalValue);
		vstrDestroy (finalValue);
	    }
	}  // end if not in commondialog
    }
    iterator.Finish ();

    // call base class vuLoadDialog
    vuCommonDialogClass::vuLoadDialog (pointName);
    setDialogModify (vFALSE);
}


/*******************************************************************
*
*    vuSaveDialog 
*
*******************************************************************/

void vuLdfw2DialogClass::vuSaveDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    const vchar 	*optionMenuValue;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*stringValue;

    char		pvName [PVNAME_SZ + 1];
    DB_FIELD_PTR	putField;
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    double		doubleValue;
    int			doSave;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    if ( (validateLdfw1Ref (db)) == vFALSE)
	return;

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    doSave = vTRUE;
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;

	    putField = xrefNode -> xrefField;

	    dialogItem = vuFindContainerItem (ldfw2Container, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		case DBF_LONG:
		case DBF_ENUM:
		case DBF_USHORT:
		case DBF_ULONG:
		case DBF_FLOAT:
		case DBF_DOUBLE:
		    textItem = (vtextitem *)dialogItem;
    		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    doubleValue = atof ((const char *)dbValue);
		    DB_PutFieldValue (pntRecord, putField, (void *)&doubleValue);
		    doSave = vFALSE;
		    break;

		case DBF_STRING:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		{
		    if (xrefNode -> xrefField -> field_type == 'L')
		    {
			int	logical;
			toggleItem = (vcontrolToggle *)dialogItem;
			logical = toggleItem -> GetValue ();
			*((char *)dbValue) = (logical ? xrefNode -> xrefTrue : 
								xrefNode -> xrefTrue ^1);
			break;
		    }
		    optionMenu = (vcontrolOptionMenu *)dialogItem;
		    optionMenuValue = vuGetOptionMenuValue (optionMenu);
		    vcharExport (vcharScribeSystemExport (optionMenuValue), dbValue);
		    break;
		}

		case DBF_INLINK:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		default:
		    continue;

	    }  // end xrefNode -> xrefEPICSFieldType

	    if (doSave == vTRUE)
	    {
		DB_PutFieldValue (pntRecord, putField, dbValue);
	    }

	}  // end if not in commondialog

    }  // end while iterator

    iterator.Finish ();

    // call base class vuSaveDialog
    vuCommonDialogClass::vuSaveDialog (pointName);
}


/*******************************************************************
*
*    vuLdfw2DialogClass, validate function, called from 
*	vuSaveDialog,
*	the ldfw1 ref must be there AND it MUST be a ldfw1 type
*
*******************************************************************/

int vuLdfw2DialogClass::validateLdfw1Ref (vuDatabaseClass *db)
{
    vstr *ldfw1Str;
    int  result = vTRUE;

    ldfw1Str = ldfw2Container -> ldfw1 -> GetTextAsString ();
    if (ldfw1Str[0] == 0)
    {
	vuMessageDialogCenter (this, VU_LDFW2_NO_LDFW1_ERR);
	result = vFALSE;
    }
    else
    {
	char ldfw1Name[PVNAME_SZ + 1];
	vcharExportBounded (vcharScribeSystemExport (ldfw1Str),
			    ldfw1Name,
			    PVNAME_SZ);
	ldfw1Name[PVNAME_SZ] = 0;

    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, ldfw1Name);
	DB_RECORD_PTR pntRecord = DB_Seek (db -> vuPvmapTable, seekBuff);
	if (pntRecord == (DB_RECORD_PTR)0)
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_LDFW2_LDFW1_EXIST_ERR, ldfw1Str);
	    result = vFALSE;
	}
	else
	{
	    int ptType;
	    int dbaseType;
	    char ptTypeStr[POINTTYPE_SZ + 1];
	    DB_GetFieldValue (pntRecord, db -> pvmapTypeField, (void *)&ptTypeStr[0], &dbaseType);
	    ptType = vuGetPointTypeIndex (ptTypeStr);
	    // the point name in the ldfw1 field is NOT an LDFW1
	    if (ptType != VU_LDFW1)
	    {
		vuMessageDialogCenter ((vdialog *)this, VU_PTTYPEDIALOGERR,
					ldfw1Str, ptType, VU_LDFW2);
		result = vFALSE;
	    }
	    // the point name entered in pvname field has been deleted
	    else if (DB_IsDeleted (db -> vuPvmapTable))
	    {
		vuMessageDialogCenter ((vdialog *)this, VU_PTISDELETEDERR, ldfw1Str);
		result = vFALSE;
	    }
	}
    }

    vstrDestroy (ldfw1Str);
    return (result);
}


vkindTYPED_FULL_DEFINITION(vuLdfw2ContainerClass, vcontainer, "vuLdfw2ContainerClass");

/*******************************************************************
*
*    vuLdfw2ContainerClass override for constructor
*
*******************************************************************/

void vuLdfw2ContainerClass::LoadInit (vresource res)
{
    int	i;

    vapplication *app = vapplication::GetCurrent ();
    vresource pointDictRes = vresourceGet (app -> GetResources (), 
				vunames_vuPointDictionary);
    vresource realRes = vresourceGet (pointDictRes, vunames_vuLdfw2Container);

    vcontainer *scntcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2SCNTContainer);
    speedCtrlButton = vuPvNameButtonClass::RequestSubLoad (scntcontainer,
				vunames_vuLdfw2SCNTButton);
    speedCtrl = vuFilteredTextitem::RequestSubLoad (scntcontainer,
				vunames_SCNT);

    procSpeedCtrl = vuCheckBoxClass::RequestSubLoad (this,
				vunames_SCPR);
    msSpeedCtrl = vuCheckBoxClass::RequestSubLoad (this,
				vunames_SCMS);

    onScan = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OSCN);

    onAlarm = vuCheckBoxClass::RequestSubLoad (this,
				vunames_AENB);

    manual = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MODE);

    vcontainer *ldfwcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2LDFWContainer);
    ldfw1Button = vuPvNameButtonClass::RequestSubLoad (ldfwcontainer,
				vunames_vuLdfw2LDFWButton);
    ldfw1 = vuFilteredTextitem::RequestSubLoad (ldfwcontainer,
				vunames_LDFW);

    procLdfw1 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_LDPR);
    msLdfw1 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_LDMS);

    vcontainer *preccontainer = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2PRECContainer);
    precision = filteredTextitem::RequestSubLoad (preccontainer,
				vunames_PREC);

    vcontainer *psmdcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2PSMDContainer);
    pumpSelMode = vuChoiceRecMenuClass::RequestSubLoad (psmdcontainer,
				vunames_PSMD);

    vcontainer *stspcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2STSPContainer);
    pumpStartSpeed = filteredTextitem::RequestSubLoad (stspcontainer,
				vunames_STSP);

    vcontainer *spspcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2SPSPContainer);
    pumpStopSpeed = filteredTextitem::RequestSubLoad (spspcontainer,
				vunames_SPSP);

    vcontainer *mdelcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2MDELContainer);
    monitorDeadband = filteredTextitem::RequestSubLoad (mdelcontainer,
				vunames_MDEL);

    vcontainer *alrmgrpcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2ALGPContainer);
    alarmGroup = vuFilteredTextitem::RequestSubLoad (alrmgrpcontainer,
				vunames_ALGP);

    vcontainer *alarmdiscontainer = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2ADISContainer);
    alarmDisableButton = vuPvNameButtonClass::RequestSubLoad (alarmdiscontainer,
				vunames_vuLdfw2ADISButton);
    alarmDisable = vuFilteredTextitem::RequestSubLoad (alarmdiscontainer,
				vunames_ADIS);

    vcontainer *af01container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF01Container);
    adjFactors[0] = filteredTextitem::RequestSubLoad (af01container,
				vunames_AF01);

    vcontainer *af02container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF02Container);
    adjFactors[1] = filteredTextitem::RequestSubLoad (af02container,
				vunames_AF02);

    vcontainer *af03container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF03Container);
    adjFactors[2] = filteredTextitem::RequestSubLoad (af03container,
				vunames_AF03);

    vcontainer *af04container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF04Container);
    adjFactors[3] = filteredTextitem::RequestSubLoad (af04container,
				vunames_AF04);

    vcontainer *af05container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF05Container);
    adjFactors[4] = filteredTextitem::RequestSubLoad (af05container,
				vunames_AF05);

    vcontainer *af06container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF06Container);
    adjFactors[5] = filteredTextitem::RequestSubLoad (af06container,
				vunames_AF06);

    vcontainer *af07container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF07Container);
    adjFactors[6] = filteredTextitem::RequestSubLoad (af07container,
				vunames_AF07);

    vcontainer *af08container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF08Container);
    adjFactors[7] = filteredTextitem::RequestSubLoad (af08container,
				vunames_AF08);

    vcontainer *af09container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF09Container);
    adjFactors[8] = filteredTextitem::RequestSubLoad (af09container,
				vunames_AF09);

    vcontainer *af10container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF10Container);
    adjFactors[9] = filteredTextitem::RequestSubLoad (af10container,
				vunames_AF10);

    vcontainer *af11container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF11Container);
    adjFactors[10] = filteredTextitem::RequestSubLoad (af11container,
				vunames_AF11);

    vcontainer *af12container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF12Container);
    adjFactors[11] = filteredTextitem::RequestSubLoad (af12container,
				vunames_AF12);

    vcontainer *af13container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF13Container);
    adjFactors[12] = filteredTextitem::RequestSubLoad (af13container,
				vunames_AF13);

    vcontainer *af14container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF14Container);
    adjFactors[13] = filteredTextitem::RequestSubLoad (af14container,
				vunames_AF14);

    vcontainer *af15container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF15Container);
    adjFactors[14] = filteredTextitem::RequestSubLoad (af15container,
				vunames_AF15);

    vcontainer *af16container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF16Container);
    adjFactors[15] = filteredTextitem::RequestSubLoad (af16container,
				vunames_AF16);

    vcontainer *af17container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF17Container);
    adjFactors[16] = filteredTextitem::RequestSubLoad (af17container,
				vunames_AF17);

    vcontainer *af18container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF18Container);
    adjFactors[17] = filteredTextitem::RequestSubLoad (af18container,
				vunames_AF18);

    vcontainer *af19container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF19Container);
    adjFactors[18] = filteredTextitem::RequestSubLoad (af19container,
				vunames_AF19);

    vcontainer *af20container = vcontainer::RequestSubLoad (this,
				vunames_vuLdfw2AF20Container);
    adjFactors[19] = filteredTextitem::RequestSubLoad (af20container,
				vunames_AF20);

    vcontainer::LoadInit (realRes);

    // store ptr to text field for vuPvNameButtonClass
    speedCtrlButton -> storePvName (speedCtrl);
    ldfw1Button -> storePvName (ldfw1);
    alarmDisableButton -> storePvName (alarmDisable);

    // set filters
    precision -> SetItemFilterFunc (numericFilter);
    pumpStartSpeed -> SetItemFilterFunc (floatFilter);
    pumpStopSpeed -> SetItemFilterFunc (floatFilter);
    monitorDeadband -> SetItemFilterFunc (floatFilter);

    speedCtrl -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *scntText = (filteredText *)speedCtrl -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    scntText -> SetUserPtr ((void *)speedCtrl);

    ldfw1 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *ldfwText = (filteredText *)ldfw1 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    ldfwText -> SetUserPtr ((void *)ldfw1);

    alarmGroup -> SetItemFilterFunc (vuAlarmGroupFilter);
    filteredText *alarmText = (filteredText *)alarmGroup -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    alarmText -> SetUserPtr ((void *)alarmGroup);

    alarmDisable -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *alrmDisText = (filteredText *)alarmDisable -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    alrmDisText -> SetUserPtr ((void *)alarmDisable);

    for (i = 0; i < VU_NUM_LDFW_PUMPS; i++)
    {
	adjFactors[i] -> SetItemFilterFunc (floatFilter);
    }

    // build popup menus
    pumpSelMode -> vuBuildOptionMenu (VU_LDFW2, REC_LDFW2_PSMD);

}

/*******************************************************************
*
*    vuLdfw2ContainerClass override for destructor
*
*******************************************************************/

void vuLdfw2ContainerClass::Destroy (void)
{
    vcontainer::Destroy ();
}

