#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbFldTypes.h"
#include "choiceGbl.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "dbRecDes.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuCommonDialogClass.hxx"
#include "vuStringinDialogClass.hxx"
#include "vuNavigatorDialogClass.hxx"
#include "vuUtils.hxx"


vkindTYPED_FULL_DEFINITION(vuStringinDialogClass, vuCommonDialogClass, "vuStringinDialogClass");

/*******************************************************************
*
*    vuStringinDialogClass override for constructors init ()
*
*******************************************************************/

void vuStringinDialogClass::LoadInit (vresource res)
{
    // load in the stringin specific fields
    stringinContainer = vuStringinContainerClass::RequestSubLoad (this,
				vunames_vuCommonContainer);
    vuCommonDialogClass::LoadInit (res);

    // store dialog modified flag
    stringinContainer -> boardType -> vuSetModifiedPtr (&this -> dialogModified);
    stringinContainer -> line -> SetTextModifiedPtr (&this -> dialogModified);
    stringinContainer -> addr -> SetTextModifiedPtr (&this -> dialogModified);
    stringinContainer -> card -> SetTextModifiedPtr (&this -> dialogModified);
    stringinContainer -> signal -> SetTextModifiedPtr (&this -> dialogModified);
    stringinContainer -> onScan -> vuSetModifiedPtr (&this -> dialogModified);
    stringinContainer -> onAlarm -> vuSetModifiedPtr (&this -> dialogModified);
    stringinContainer -> reference -> SetTextModifiedPtr (&this -> dialogModified);
    stringinContainer -> processInput -> vuSetModifiedPtr (&this -> dialogModified);
    stringinContainer -> maximizeSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    stringinContainer -> readSize -> SetModifiedPtr (&this -> dialogModified);
    stringinContainer -> manual -> vuSetModifiedPtr (&this -> dialogModified);
    stringinContainer -> alarmGroup -> SetTextModifiedPtr (&this -> dialogModified);
    stringinContainer -> alarmDisable -> SetTextModifiedPtr (&this -> dialogModified);
    stringinContainer -> chgValSeverity -> vuSetModifiedPtr (&this -> dialogModified);

    // store dialog ptr for vuPvNameButtonClass
    stringinContainer -> referenceButton -> storeCommonDialog ((vuCommonDialogClass *)this);
    stringinContainer -> alarmDisableButton -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuPvNameFilter
    stringinContainer -> reference -> storeCommonDialog ((vuCommonDialogClass *)this);
    stringinContainer -> alarmDisable -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuAlarmGroupFilter
    stringinContainer -> alarmGroup -> storeCommonDialog ((vuCommonDialogClass *)this);

    // set the point type
    setPointType (VU_STRINGIN);

    // get the title from the resource file, biTitle must be
    // freed by the class destuctor
    vapplication *app = vapplication::GetCurrent ();
    vresource dialogTitles = vresourceGet (app -> GetResources (),
				vunames_vuDialogTitles);
    stringinTitle = vresourceGetString (dialogTitles, 
				vunames_vuStringinDialogTitle);
    SetTitle (stringinTitle);

    // trend button is not needed 
    vuCommonTrendButtonClass *trendbutton = getTrendButton ();
    trendbutton -> Disable ();
}


/*******************************************************************
*
*    vuStringinDialogClass Destroy function
*
*******************************************************************/

void vuStringinDialogClass::Destroy (void)
{
    vstrDestroy (stringinTitle);
    vuCommonDialogClass::Destroy ();
}


/*******************************************************************
*
*    vuStringinDialogClass, reset fields, overrides base class 
*	vuCommonDialogClass
*
*******************************************************************/

void vuStringinDialogClass::resetDialog (void)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];
    vspinner		*spinner;

    vresource defRes = mainDialog -> getDefaultsResource ();

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0; 
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	// if not a common dialog field
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    dialogItem = vuFindContainerItem (stringinContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;

		// set the default value for the Bi's line field
		if ( ((vcharCompareBounded (epics, "LINE", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "ADDR", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "CARD", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "SIGN", FLDNAME_SZ)) == 0) )
		{
		    vstr *zero = vresourceGetString (defRes, 
						vunames_vuZeroDefault);
		    textItem -> SetText (zero);
		    vstrDestroy (zero);
		}
		else
		{
		    textItem -> SetTextScribed (vcharScribeSystem (""));
		}
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		toggleItem -> SetValue (0);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		optionMenu -> SetValue (0);
	    }
	    else if ( (dialogItem -> IsKindOf (&vspinner::Kind)) == vTRUE)
	    {
		spinner = (vspinner *)dialogItem;
		spinner -> SetValue (spinner -> GetMin());
	    }
	}  // end if not in common dialog
    }
    iterator.Finish ();

    // call base class reset function with proper point index
    vuCommonDialogClass::resetDialog ();

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuLoadDialog 
*
*******************************************************************/

void vuStringinDialogClass::vuLoadDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*finalValue;
    vstr		*referenceString;
    vstr		*constantString;
    vspinner		*spinner;

    char		pvName [PVNAME_SZ + 1];
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dbaseType;	// codebase field type
    int			intValue;
    unsigned int	uIntValue;
    long		longValue;
    unsigned long	uLongValue;
    double		doubleValue;
    DTYP_TYPE		dType;		// device type indicator

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (db -> vuPvmapTable, seekBuff);
    if (pntRecord == (DB_RECORD_PTR)0)
    {
	return;
    }
    else
    {
	int ptType;
	DB_GetFieldValue (pntRecord, db -> pvmapTypeField, dbValue, &dbaseType);
	ptType = vuGetPointTypeIndex ((char *)dbValue);
	// the point name in pvname field is NOT a STRINGIN
	if (ptType != VU_STRINGIN)
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTTYPEDIALOGERR,
				   pointName, ptType, VU_STRINGIN);
	    vuClearPvNameField ();
	    return;
	}
	// the point name entered in pvname field has been deleted
	if (DB_IsDeleted (db -> vuPvmapTable))
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTISDELETEDERR, pointName);
	    vuClearPvNameField ();
	    return;
	}
    }
    // call base class vuLoadDialog first
    vuCommonDialogClass::vuLoadDialog (pointName);

    // since the stringin can be a hardware point, find what we are for later use
    // by getting the DTYP (device type) dictionary entry
    referenceString = vstrCloneScribed (vcharScribeSystem ("REFERENCE"));
    constantString = vstrCloneScribed (vcharScribeSystem ("CONSTANT"));
    xrefNode = (xrefNodeClass *)(dict -> GetElem (
			vnameGetGlobalScribed (vcharScribeSystem ("DTYP"))));
    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
    if ( (vcharCompare ((const vchar *)dbValue, referenceString)) == 0)
	dType = VU_REFERENCE;
    else if ( (vcharCompare ((const vchar *)dbValue, constantString)) == 0)
	dType = VU_CONSTANT;
    else
	dType = VU_HARDWARE;
    vstrDestroy (referenceString);
    vstrDestroy (constantString);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;
	    dialogItem = vuFindContainerItem (stringinContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
	    doubleValue = 0.0;
	    // GetField actually returns a number in the char array
	    if ( (dbaseType == 'F') || (dbaseType == 'N') )
	    {
		doubleValue = *((double *)dbValue);
	    }
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		    intValue = (int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeInt (intValue));
		    break;

		case DBF_LONG:
		    longValue = (long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeLong (longValue));
		    break;

		case DBF_ENUM:
		case DBF_USHORT:
		    uIntValue = (unsigned int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeUInt (uIntValue));
		    break;

		case DBF_ULONG:
		    uLongValue = (unsigned long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeULong (uLongValue));
		    break;

		case DBF_FLOAT:
		case DBF_DOUBLE:
		    // to output CONSTANT field, device type must be CONSTANT
		    if ( (strncmp (xrefNode -> xrefEpicsName, "CNST", FLDNAME_SZ)) == 0)
		    {
			if (dType != VU_CONSTANT)
			    continue;
		    }
		    finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
		    break;

		case DBF_STRING:
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		    if (dbaseType == 'L')
		    {
			int	logical;

			logical = (*((char*)dbValue) ? 
					xrefNode -> xrefTrue : 
					(xrefNode -> xrefTrue ^1));
			finalValue = vstrCloneScribed (vnumScribeInt (logical));
			break;
		    }
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_INLINK:
		    if ( (strncmp (xrefNode -> xrefEpicsName, "INP ", FLDNAME_SZ)) == 0)
		    {
			// device type must be REFERENCE to output the reference
			if (dType == VU_REFERENCE)
			{
			    DB_GetFieldValue (pntRecord, xrefNode -> xrefRefField, 
					      dbValue, &dbaseType);
			    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
			}
			else
			{
			    finalValue = vstrCloneScribed (vcharScribeSystem (""));
			}
		    }
		    else
			finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		default:
		    continue;
	    }
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;
		textItem -> SetTextScribed (vcharScribeSystem (""));
		textItem -> SetText (finalValue);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		if (finalValue[0] == '0')
		    toggleItem -> SetValue (0);
		else
		    toggleItem -> SetValue (1);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		vuSetOptionMenuValue (optionMenu, finalValue);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vspinner::Kind)) == vTRUE)
	    {
		spinner = (vspinner *)dialogItem;
		spinner -> SetValue (vnum::ParseInt (finalValue, NULL));
		vstrDestroy (finalValue);
	    }
	}  // end if not in commondialog
    }
    iterator.Finish ();

    setDialogModify (vFALSE);
}


/*******************************************************************
*
*    vuSaveDialog 
*
*******************************************************************/

void vuStringinDialogClass::vuSaveDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    const vchar 	*optionMenuValue;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*stringValue;
    vstr		*referenceString;
    vstr		*constantString;
    vspinner		*spinner;

    char		pvName [PVNAME_SZ + 1];
    char		vuCurrentBoard [VU_KEY_SIZE];
    DB_FIELD_PTR	putField;
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dType;		// device type
    double		doubleValue;
    int			doSave;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    // since the stringin can be a hardware point, find what we are for later use
    // by getting the DTYP (BOARD) value from the dialog
    referenceString = vstrCloneScribed (vcharScribeSystem ("REFERENCE"));
    constantString = vstrCloneScribed (vcharScribeSystem ("CONSTANT"));
    vcharCopyScribedBounded (vcharScribeSystem ("DTYP"), epics, FLDNAME_SZ);
    epics[sizeof (epics) - 1] = 0;
    // find the BOARD (DTYP) field and get its value
    dialogItem = vuFindContainerItem (stringinContainer, epics);
    optionMenu = (vcontrolOptionMenu *)dialogItem;
    optionMenuValue = vuGetOptionMenuValue (optionMenu);
    vcharExport (vcharScribeSystemExport (optionMenuValue), (void *)vuCurrentBoard);
    if ( (vcharCompare (optionMenuValue, referenceString)) == 0)
	dType = VU_REFERENCE;
    else if ( (vcharCompare (optionMenuValue, constantString)) == 0)
	dType = VU_CONSTANT;
    else
	dType = VU_HARDWARE;
    vstrDestroy (referenceString);
    vstrDestroy (constantString);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    doSave = vTRUE;
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;

	    putField = xrefNode -> xrefField;

	    dialogItem = vuFindContainerItem (stringinContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		case DBF_LONG:
		case DBF_ENUM:
		case DBF_USHORT:
		case DBF_ULONG:
		case DBF_FLOAT:
		case DBF_DOUBLE:
		    if ( (dialogItem -> IsKindOf (&vspinner::Kind)) == vTRUE)
		    {
			spinner = (vspinner *)dialogItem;
			// buf fix, spinner mau not have proper value
			spinner -> Unfocus ();
			stringValue = spinner -> GetTextAsString ();
		    }
		    else
		    {
			textItem = (vtextitem *)dialogItem;
			stringValue = textItem -> GetTextAsString ();
		    }
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    doubleValue = atof ((const char *)dbValue);
		    if ( (strncmp (xrefNode -> xrefEpicsName, "CNST", FLDNAME_SZ)) == 0)
		    {
			if (dType != VU_CONSTANT)
			    doubleValue = 0;
		    }
		    DB_PutFieldValue (pntRecord, putField, (void *)&doubleValue);
		    doSave = vFALSE;
		    break;

		case DBF_STRING:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		{
		    if (xrefNode -> xrefField -> field_type == 'L')
		    {
			int	logical;
			toggleItem = (vcontrolToggle *)dialogItem;
			logical = toggleItem -> GetValue ();
			*((char *)dbValue) = (logical ? xrefNode -> xrefTrue : 
								xrefNode -> xrefTrue ^1);
			break;
		    }
		    optionMenu = (vcontrolOptionMenu *)dialogItem;
		    optionMenuValue = vuGetOptionMenuValue (optionMenu);
		    vcharExport (vcharScribeSystemExport (optionMenuValue), dbValue);
		    if ( (strncmp (xrefNode -> xrefEpicsName, "DTYP", FLDNAME_SZ)) == 0)
		    {
			// write to DEVTYPE in pvmap
			DB_PutFieldValue (pntRecord, xrefNode -> xrefField, dbValue);
		    }
		    break;
		}

		case DBF_INLINK:
		    if ( (strncmp (xrefNode -> xrefEpicsName, "INP", 3)) == 0)
		    {
			// always write the boardtype from DTYP into the INP field
			DB_PutFieldValue (pntRecord, putField, vuCurrentBoard);
			// if we are a reference then write INP into the REFERENCE field
			if (dType == VU_REFERENCE)
			{
			    putField = xrefNode -> xrefRefField;
			    textItem = (vtextitem *)dialogItem;
			    stringValue = textItem -> GetTextAsString ();
			    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
			    vstrDestroy (stringValue);
			}
			else
			{
			    // make sure to null out the field
			    putField = xrefNode -> xrefRefField;
			    vmemSet (dbValue, 0, VU_KEY_SIZE);
			}
		    }
		    else
		    {
			textItem = (vtextitem *)dialogItem;
			stringValue = textItem -> GetTextAsString ();
			vcharExport (vcharScribeSystemExport (stringValue), dbValue);
			vstrDestroy (stringValue);
		    }
		    break;

		default:
		    continue;

	    }  // end xrefNode -> xrefEPICSFieldType

	    if (doSave == vTRUE)
	    {
		DB_PutFieldValue (pntRecord, putField, dbValue);
	    }

	}  // end if not in commondialog

    }  // end while iterator

    iterator.Finish ();

    // call base class vuSaveDialog
    vuCommonDialogClass::vuSaveDialog (pointName);
}


vkindTYPED_FULL_DEFINITION(vuStringinContainerClass, vcontainer, "vuStringinContainerClass");

/*******************************************************************
*
*    vuStringinContainerClass override for constructor
*
*******************************************************************/

void vuStringinContainerClass::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource pointDictRes = vresourceGet (app -> GetResources (), 
				vunames_vuPointDictionary);
    vresource realRes = vresourceGet (pointDictRes, vunames_vuStringinContainer);

    vcontainer *btcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringinDTYPContainer);
    boardType = vuBoardTypeMenuClass::RequestSubLoad (btcontainer,
				vunames_DTYP);

    vcontainer *linecontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringinLINEContainer);
    line = filteredTextitem::RequestSubLoad (linecontainer,
				vunames_LINE);

    vcontainer *addrcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringinADDRContainer);
    addr = filteredTextitem::RequestSubLoad (addrcontainer,
				vunames_ADDR);

    vcontainer *cardcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringinCARDContainer);
    card = filteredTextitem::RequestSubLoad (cardcontainer,
				vunames_CARD);

    vcontainer *signalcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringinSIGNContainer);
    signal = filteredTextitem::RequestSubLoad (signalcontainer,
				vunames_SIGN);

    onScan = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OSCN);

    onAlarm = vuCheckBoxClass::RequestSubLoad (this,
				vunames_AENB);

    vcontainer *refcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringinINPContainer);
    referenceButton = vuPvNameButtonClass::RequestSubLoad (refcontainer,
				vunames_vuStringinINPButton);
    reference = vuFilteredTextitem::RequestSubLoad (refcontainer,
				vunames_INP);

    processInput = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PRCI);

    maximizeSeverity = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MAXS);

    vcontainer *rdszcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringinRDSZContainer);
    readSize = spinnerClass::RequestSubLoad (rdszcontainer,
				vunames_RDSZ);

    manual = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MODE);

    vcontainer *alrmgrpcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringinALGPContainer);
    alarmGroup = vuFilteredTextitem::RequestSubLoad (alrmgrpcontainer,
				vunames_ALGP);

    vcontainer *alarmdiscontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringinADISContainer);
    alarmDisableButton = vuPvNameButtonClass::RequestSubLoad (alarmdiscontainer,
				vunames_vuStringinADISButton);
    alarmDisable = vuFilteredTextitem::RequestSubLoad (alarmdiscontainer,
				vunames_ADIS);

    vcontainer *cvsvcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringinCVSVContainer);
    chgValSeverity = vuChoiceGblMenuClass::RequestSubLoad (cvsvcontainer,
				vunames_CVSV);

    vcontainer::LoadInit (realRes);

    // store ptr to text field for vuPvNameButtonClass
    referenceButton -> storePvName (reference);
    alarmDisableButton -> storePvName (alarmDisable);

    // set filters
    line -> SetItemFilterFunc (numericFilter);
    addr -> SetItemFilterFunc (numericFilter);
    card -> SetItemFilterFunc (numericFilter);
    signal -> SetItemFilterFunc (numericFilter);

    reference -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *refText = (filteredText *)reference -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    refText -> SetUserPtr ((void *)reference);

    alarmDisable -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *alrmDisText = (filteredText *)alarmDisable -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    alrmDisText -> SetUserPtr ((void *)alarmDisable);

    alarmGroup -> SetItemFilterFunc (vuAlarmGroupFilter);
    filteredText *alarmText = (filteredText *)alarmGroup -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    alarmText -> SetUserPtr ((void *)alarmGroup);

    // build popup menus
    boardType -> vuBuildBoardTypeMenu (VU_STRINGIN);
    chgValSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);

}

/*******************************************************************
*
*    vuStringinContainerClass override for destructor
*
*******************************************************************/

void vuStringinContainerClass::Destroy (void)
{
    vcontainer::Destroy ();
}

