#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER
#include vrulerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"

#include "vuSyncDialogClass.hxx"
#include "vuUtils.hxx"


vkindTYPED_FULL_DEFINITION(vuSyncDialogClass, vuFindDialogClass, 
						"vuSyncDialogClass");

/*******************************************************************
*
*    vuSyncDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuSyncDialogClass::LoadInit(vresource res)
{
    // set to NULL
    dialogTitle = NULL;

    syncContainer = vuSyncContainerClass::RequestSubLoad (this,
				vunames_vuFindCommonContainer);

    vcontainer *buttonContainer = vcontainer::RequestSubLoad (this,
				vunames_vuFindButtonContainer);
    
    dismissButton = vuSyncDismissButtonClass::RequestSubLoad (buttonContainer,
				vunames_vuFindDismissButton);
    findButton = vuSyncFindButtonClass::RequestSubLoad (buttonContainer,
				vunames_vuFindFindButton);

    // call base class loadInit to finish the job
    vuFindDialogClass::LoadInit(res);

    // set pointer to navigator dialog in these items

    syncContainer -> abButton -> setSyncDialog (this);
    syncContainer -> baButton -> setSyncDialog (this);
    syncContainer -> ruler -> setSyncDialog (this);
    findButton -> setSyncDialog (this);
    dismissButton -> setSyncDialog (this);

}


/*******************************************************************
*
*    vuSyncDialogClass notify function
*
*******************************************************************/

void vuSyncDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	Close ();
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    vuSyncDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuSyncDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    // get the list attached to the listview and destroy it, also set the 
    // foundList in the find class to NULL because the base class Destroy will
    // also try to destroy it if it is not NULL
    vliststr *listdata = (vliststr *)(syncContainer -> pointList -> GetList ());
    if (listdata != NULL)
    {
	syncContainer -> pointList -> SetList (NULL);
	listdata -> Destroy ();
	foundList = NULL;
    }

    vstrDestroy (dialogTitle);

    vuFindDialogClass::Destroy();
}


/*******************************************************************
*
*    vuSyncDialogClass Open function
*
*******************************************************************/

void vuSyncDialogClass::vuFindOpen(vchar *label,vuMainDialogClass *mainDialg,int multiSelect)
{

    mainDialog = mainDialg;

    // call base class open
    vuFindDialogClass::vuFindOpen (label, mainDialg,multiSelect);

    if (dialogTitle == NULL)
    {
	vresource titleRes  = mainDialog -> getTitlesResource ();
	dialogTitle = vresourceGetString (titleRes, vunames_vuSyncDialogTitle);
	SetTitle (dialogTitle);
    }

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    vevent::Process ();
}


vkindTYPED_FULL_DEFINITION(vuSyncRulerClass, vruler, "vuSyncRulerClass");

/*******************************************************************
*
*    vuSyncRulerClass  notify function
*
*******************************************************************/

void vuSyncRulerClass::ObserveRuler(vevent *event, vrulerItem *item, int message)
{
    if (message == vrulerITEM_SELECT)
    {
	vuSyncContainerClass *syncContainer = syncDialog -> getSyncContainer ();
	vliststr *list = (vliststr *)(syncContainer -> pointList -> GetList ());
	if (list == NULL)
	{
	    return;
	}
	vlistSelection *selection = syncContainer -> pointList -> GetSelection ();
	list -> UnselectAll (selection);
	int index = item -> GetIndex ();

	// have clicked on PVNAME, so select entire list
	if (index == 0)
	{
	    list -> SelectAll (selection);
	}
	// else select one of the otehr columns
	else
	{
	    list -> SelectColumn (selection, index);
	}
    }
}


vkindTYPED_FULL_DEFINITION(vuSyncABButtonClass, vbutton, 
					"vuSyncABButtonClass");

/*******************************************************************
*
*    vuSyncABButtonClass 
*	Observer function
*
*******************************************************************/

void vuSyncABButtonClass::ObserveDialogItem (vevent *event)
{
    vuSyncContainerClass *syncContainer = syncDialog -> getSyncContainer ();
    vliststr *list = (vliststr *)(syncContainer -> pointList -> GetList ());
    if (list == NULL)
    {
	return;
    }
    vlistSelection *selection = syncContainer -> pointList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	vuMessageDialogCenter (syncDialog, VU_NOTHING_SELECTED_MSG);
	return;
    }
    vuMainDialogClass   *mainDialog = syncDialog -> getMainDialog ();
    dataVuApplication *app = mainDialog -> getApplication();
    vuDatabaseClass *db = app -> getDatabaseClass();

    syncDialog -> Block ();

    iterator.StartWithSelection (selection);
    while (iterator.Next ())
    {
    }
    iterator.Finish ();
    list -> UnselectAll (selection);

    syncDialog -> Unblock ();
}


vkindTYPED_FULL_DEFINITION(vuSyncBAButtonClass, vbutton, 
					"vuSyncBAButtonClass");

/*******************************************************************
*
*    vuSyncBAButtonClass 
*	Observer function
*
*******************************************************************/

void vuSyncBAButtonClass::ObserveDialogItem (vevent *event)
{
    vuSyncContainerClass *syncContainer = syncDialog -> getSyncContainer ();
    vliststr *list = (vliststr *)(syncContainer -> pointList -> GetList ());
    if (list == NULL)
    {
	return;
    }
    vlistSelection *selection = syncContainer -> pointList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	vuMessageDialogCenter (syncDialog, VU_NOTHING_SELECTED_MSG);
	return;
    }
    vuMainDialogClass   *mainDialog = syncDialog -> getMainDialog ();
    dataVuApplication *app = mainDialog->getApplication();
    vuDatabaseClass *db = app -> getDatabaseClass();

    syncDialog -> Block ();

    iterator.StartWithSelection (selection);
    while (iterator.Next ())
    {
    }
    iterator.Finish ();
    list -> UnselectAll (selection);

    syncDialog -> Unblock ();
}


vkindTYPED_FULL_DEFINITION(vuSyncDismissButtonClass, vbutton,
					"vuSyncDismissButtonClass");

/*******************************************************************
*
*    vuSyncDismissButtonClass 
*	Observer function
*
*******************************************************************/

void vuSyncDismissButtonClass::ObserveDialogItem (vevent *event)
{

    // close dialog and stop event loop started in vuFindOpen ()
    syncDialog -> Close ();
    vevent::StopProcessing ();
}


vkindTYPED_FULL_DEFINITION(vuSyncFindButtonClass, vbutton, 
						"vuSyncFindButtonClass");

/*******************************************************************
*
*    vuSyncFindButtonClass 
*	Observer function
*
*******************************************************************/

void vuSyncFindButtonClass::ObserveDialogItem (vevent *event)
{
    vuSyncContainerClass *syncContainer = syncDialog -> getSyncContainer ();
    vuMainDialogClass   *mainDialog = syncDialog -> getMainDialog ();

    // destroy the old point name list since a new one has 
    // been requested
    vliststr *listdata = (vliststr *)(syncContainer -> pointList -> GetList ());
    if (listdata)
    {
	syncContainer -> pointList -> SetList (NULL);
	listdata -> Destroy ();
	// make sure that the list in the find class is set to NULL
	// after we destroy it
	syncDialog -> setFoundList (NULL);
    }

    // ask base class to get a list of points
    syncDialog -> vuFindList ();

    // get the list base class found and attach it to our listview
    listdata = syncDialog -> getFoundList ();
//    int numInList = syncDialog -> getNumInList ();
    if (listdata != NULL)
    {
	syncDialog -> Block ();

	dataVuApplication *app = mainDialog->getApplication();
	vuDatabaseClass *db = app -> getDatabaseClass();
	vlistIterator iterator;

	iterator.Start (listdata);
	while (iterator.Next())
	{
	}
	iterator.Finish ();

	vuSyncContainerClass *syncContainer = syncDialog -> getSyncContainer ();
	syncContainer -> pointList -> SetList (listdata);
	int i;
	int itemCount = syncContainer -> ruler -> GetItemCount ();;
	vrulerItem *rulerItem;
	int rulerItemWidth;
	for (i = 0; i < itemCount; i++)
	{
	    rulerItem = syncContainer -> ruler -> GetItemAt (i);
	    rulerItemWidth = rulerItem -> GetWidth ();
	    syncContainer -> pointList -> SetColumnWidth (i, rulerItemWidth);
	}
	vlistSelection *selection = syncContainer -> pointList -> GetSelection ();
	listdata -> SetSelectMethod (selection, vlistFREE_SELECTION);

	syncDialog -> Unblock ();
    }
}



vkindTYPED_FULL_DEFINITION(vuSyncContainerClass, vcontainer, 
						"vuSyncContainerClass");

/*******************************************************************
*
*    vuSyncContainerClass 
*	override for constructor
*	We get the real resource since what comes in is the resource
*	for the find dialog and we need the resource for the
*	container
*
*******************************************************************/

void vuSyncContainerClass::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource findDictRes = vresourceGet (app -> GetResources (),
					vunames_vuFindDictionary);
    vresource realRes = vresourceGet (findDictRes, 
					vunames_vuSyncContainer);

    vcontainer *acnt = vcontainer::RequestSubLoad (this,
					vunames_vuSyncPathAContainer);
    pathA = vtextitem::RequestSubLoad (acnt,
					vunames_vuSyncPathA);

    vcontainer *bcnt = vcontainer::RequestSubLoad (this,
					vunames_vuSyncPathBContainer);
    pathB = vtextitem::RequestSubLoad (bcnt,
					vunames_vuSyncPathB);

    pointList = vlistview::RequestSubLoad (this,
					vunames_vuSyncPtList);

    pathAreadOnly = vcontrolToggle::RequestSubLoad (this,
					vunames_vuSyncReadA);

    pathBreadOnly = vcontrolToggle::RequestSubLoad (this,
					vunames_vuSyncReadB);

    vcontainer *buttonContainer = vcontainer::RequestSubLoad (this,
				vunames_vuSyncButtonContainer);

    options = vcontainerExclusiveGroup::RequestSubLoad (this,
				vunames_vuSyncOptionsContainer);
    
    abButton = vuSyncABButtonClass::RequestSubLoad (buttonContainer,
				vunames_vuSyncABButton);
    
    baButton = vuSyncBAButtonClass::RequestSubLoad (buttonContainer,
				vunames_vuSyncBAButton);

    ruler = vuSyncRulerClass::RequestSubLoad (this,
				vunames_vuSyncRuler);

    vcontainer::LoadInit (realRes);
}


/*******************************************************************
*
*    vuSyncContainerClass 
*	override 
*
*******************************************************************/

void vuSyncContainerClass::Destroy (void)
{
    vcontainer::Destroy ();
}

