#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbFldTypes.h"
#include "choiceGbl.h"
#include "choiceTd.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "dbRecDes.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuCommonDialogClass.hxx"
#include "vuTdDialogClass.hxx"
#include "vuNavigatorDialogClass.hxx"
#include "vuUtils.hxx"


vkindTYPED_FULL_DEFINITION(vuTdDialogClass, vuCommonDialogClass, "vuTdDialogClass");

/*******************************************************************
*
*    vuTdDialogClass override for constructors init ()
*
*******************************************************************/

void vuTdDialogClass::LoadInit (vresource res)
{
    // load in the ai specific fields
    tdContainer = vuTdContainerClass::RequestSubLoad (this,
				vunames_vuCommonContainer);
    vuCommonDialogClass::LoadInit (res);

    // store dialog modified flag
    tdContainer -> boardType -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> line -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> addr -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> card -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> signal -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> valueLoc -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> output -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> constantValue -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> zeroState -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> oneState -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> tdValue1 -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> tdValue2 -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> dwellTime -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> zeroSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> oneSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> cosSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> onScan -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> onAlarm -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> closedLoop -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> invertOutput -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> initTimer -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> timerChoice -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> edge -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> cycleFirstState -> vuSetModifiedPtr (&this -> dialogModified);
    tdContainer -> alarmDisable -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> alarmDelay -> SetTextModifiedPtr (&this -> dialogModified);
    tdContainer -> alarmGroup -> SetTextModifiedPtr (&this -> dialogModified);

    // store dialog ptr for vuPvNameButtonClass
    tdContainer -> valueLocButton -> storeCommonDialog ((vuCommonDialogClass *)this);
    tdContainer -> outputButton -> storeCommonDialog ((vuCommonDialogClass *)this);
    tdContainer -> alarmDisableButton -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuPvNameFilter
    tdContainer -> valueLoc -> storeCommonDialog ((vuCommonDialogClass *)this);
    tdContainer -> output -> storeCommonDialog ((vuCommonDialogClass *)this);
    tdContainer -> alarmDisable -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuAlarmGroupFilter
    tdContainer -> alarmGroup -> storeCommonDialog ((vuCommonDialogClass *)this);

    // set the point type
    setPointType (VU_TD);

    // get the title from the resource file, tdTitle must be
    // freed by the class destuctor
    vapplication *app = vapplication::GetCurrent ();
    vresource dialogTitles = vresourceGet (app -> GetResources (),
				vunames_vuDialogTitles);
    tdTitle = vresourceGetString (dialogTitles, 
				vunames_vuTdDialogTitle);
    SetTitle (tdTitle);
}


/*******************************************************************
*
*    vuTdDialogClass Destroy function
*
*******************************************************************/

void vuTdDialogClass::Destroy (void)
{
    vstrDestroy (tdTitle);
    vuCommonDialogClass::Destroy ();
}


/*******************************************************************
*
*    vuTdDialogClass, reset fields, overrides base class 
*	vuCommonDialogClass
*
*******************************************************************/

void vuTdDialogClass::resetDialog (void)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];

    vresource defRes = mainDialog -> getDefaultsResource ();

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0; 
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	// if not a common dialog field
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    dialogItem = vuFindContainerItem (tdContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;

		// set the default value for the Bo's line field
		if ( ((vcharCompareBounded (epics, "LINE", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "ADDR", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "CARD", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "SIGN", FLDNAME_SZ)) == 0) )
		{
		    vstr *zero = vresourceGetString (defRes, 
						vunames_vuZeroDefault);
		    textItem -> SetText (zero);
		    vstrDestroy (zero);
		}
		else
		{
		    textItem -> SetTextScribed (vcharScribeSystem (""));
		}
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		toggleItem -> SetValue (0);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		optionMenu -> SetValue (0);
	    }
	}  // end if not in common dialog
    }
    iterator.Finish ();

    // call base class reset function with proper point index
    vuCommonDialogClass::resetDialog ();

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuLoadDialog 
*
*******************************************************************/

void vuTdDialogClass::vuLoadDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*finalValue;
    vstr		*referenceString;
    vstr		*constantString;

    char		pvName [PVNAME_SZ + 1];
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dbaseType;	// codebase field type
    int			intValue;
    unsigned int	uIntValue;
    long		longValue;
    unsigned long	uLongValue;
    double		doubleValue;
    DTYP_TYPE		dType;		// device type indicator

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (db -> vuPvmapTable, seekBuff);
    if (pntRecord == (DB_RECORD_PTR)0)
    {
	return;
    }
    else
    {
	int ptType;
	DB_GetFieldValue (pntRecord, db -> pvmapTypeField, dbValue, &dbaseType);
	ptType = vuGetPointTypeIndex ((char *)dbValue);
	// the point name in pvname field is NOT a TD
	if (ptType != VU_TD)
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTTYPEDIALOGERR,
				   pointName, ptType, VU_TD);
	    vuClearPvNameField ();
	    return;
	}
	// the point name entered in pvname field has been deleted
	if (DB_IsDeleted (db -> vuPvmapTable))
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTISDELETEDERR, pointName);
	    vuClearPvNameField ();
	    return;
	}
    }
    // call base class vuLoadDialog first
    vuCommonDialogClass::vuLoadDialog (pointName);

    // since the td can be a hardware point, find what we are for later use
    // by getting the DTYP (device type) dictionary entry
    referenceString = vstrCloneScribed (vcharScribeSystem ("REFERENCE"));
    constantString = vstrCloneScribed (vcharScribeSystem ("CONSTANT"));
    xrefNode = (xrefNodeClass *)(dict -> GetElem (
				vnameGetGlobalScribed (vcharScribeSystem ("DTYP"))));
    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
    if ( (vcharCompare ((const vchar *)dbValue, referenceString)) == 0)
	dType = VU_REFERENCE;
    else if ( (vcharCompare ((const vchar *)dbValue, constantString)) == 0)
	dType = VU_CONSTANT;
    else
	dType = VU_HARDWARE;
    vstrDestroy (referenceString);
    vstrDestroy (constantString);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;
	    dialogItem = vuFindContainerItem (tdContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
	    doubleValue = 0.0;
	    // GetField actually returns a number in the char array
	    if ( (dbaseType == 'F') || (dbaseType == 'N') )
	    {
		doubleValue = *((double *)dbValue);
	    }
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		    intValue = (int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeInt (intValue));
		    break;

		case DBF_LONG:
		    longValue = (long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeLong (longValue));
		    break;

		case DBF_ENUM:
		case DBF_USHORT:
		    uIntValue = (unsigned int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeUInt (uIntValue));
		    break;

		case DBF_ULONG:
		    uLongValue = (unsigned long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeULong (uLongValue));
		    break;

		case DBF_FLOAT:
		case DBF_DOUBLE:
		    // to output CONSTANT field, device type must be CONSTANT
		    if ( (strncmp (xrefNode -> xrefEpicsName, "CNST", FLDNAME_SZ)) 
										== 0)
		    {
			if (dType != VU_CONSTANT)
			    continue;
		    }
		    finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
		    break;

		case DBF_STRING:
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		    if (dbaseType == 'L')
		    {
			int	logical;

			logical = (*((char*)dbValue) ? 
					xrefNode -> xrefTrue : 
					(xrefNode -> xrefTrue ^1));
			finalValue = vstrCloneScribed (vnumScribeInt (logical));
			break;
		    }
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_OUTLINK:
		    // device type must be REFERENCE to output the reference
		    if (dType == VU_REFERENCE)
		    {
			DB_GetFieldValue (pntRecord, xrefNode -> xrefRefField, 
								dbValue, &dbaseType);
			finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    }
		    else
		    {
			finalValue = vstrCloneScribed (vcharScribeSystem (""));
		    }
		    break;

		case DBF_INLINK:
		    // works for all td inputs, see pdbxref
		    DB_GetFieldValue (pntRecord, xrefNode -> xrefRefField, 
								dbValue, &dbaseType);
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		default:
		    continue;
	    }
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;
		textItem -> SetTextScribed (vcharScribeSystem (""));
		textItem -> SetText (finalValue);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		if (finalValue[0] == '0')
		    toggleItem -> SetValue (0);
		else
		    toggleItem -> SetValue (1);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		vuSetOptionMenuValue (optionMenu, finalValue);
		vstrDestroy (finalValue);
	    }
	}  // end if not in commondialog
    }
    iterator.Finish ();

    setDialogModify (vFALSE);
}


/*******************************************************************
*
*    vuSaveDialog 
*
*******************************************************************/

void vuTdDialogClass::vuSaveDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    const vchar 	*optionMenuValue;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*stringValue;
    vstr		*referenceString;
    vstr		*constantString;

    char		pvName [PVNAME_SZ + 1];
    char		vuCurrentBoard [VU_KEY_SIZE];
    DB_FIELD_PTR	putField;
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dType;		// device type
    double		doubleValue;
    int			doSave;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    // since the td can be a hardware point, find what we are for later use
    // by getting the DTYP (TDARD) value from the dialog
    referenceString = vstrCloneScribed (vcharScribeSystem ("REFERENCE"));
    constantString = vstrCloneScribed (vcharScribeSystem ("CONSTANT"));
    vcharCopyScribedBounded (vcharScribeSystem ("DTYP"), epics, FLDNAME_SZ);
    epics[sizeof (epics) - 1] = 0;
    // find the BOARD (DTYP) field and get its value
    dialogItem = vuFindContainerItem (tdContainer, epics);
    optionMenu = (vcontrolOptionMenu *)dialogItem;
    optionMenuValue = vuGetOptionMenuValue (optionMenu);
    vcharExport (vcharScribeSystemExport (optionMenuValue), (void *)vuCurrentBoard);
    if ( (vcharCompare (optionMenuValue, referenceString)) == 0)
	dType = VU_REFERENCE;
    else if ( (vcharCompare (optionMenuValue, constantString)) == 0)
	dType = VU_CONSTANT;
    else
	dType = VU_HARDWARE;
    vstrDestroy (referenceString);
    vstrDestroy (constantString);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    doSave = vTRUE;
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;

	    putField = xrefNode -> xrefField;

	    dialogItem = vuFindContainerItem (tdContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		case DBF_LONG:
		case DBF_ENUM:
		case DBF_USHORT:
		case DBF_ULONG:
		case DBF_FLOAT:
		case DBF_DOUBLE:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    doubleValue = atof ((const char *)dbValue);
		    if ( (strncmp (xrefNode -> xrefEpicsName, "CNST", FLDNAME_SZ)) 
										== 0)
		    {
			if (dType != VU_CONSTANT)
			    doubleValue = 0;
		    }
		    DB_PutFieldValue (pntRecord, putField, (void *)&doubleValue);
		    doSave = vFALSE;
		    break;

		case DBF_STRING:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		{
		    if (xrefNode -> xrefField -> field_type == 'L')
		    {
			int	logical;
			toggleItem = (vcontrolToggle *)dialogItem;
			logical = toggleItem -> GetValue ();
			*((char *)dbValue) = (logical ? xrefNode -> xrefTrue : 
								xrefNode -> xrefTrue ^1);
			break;
		    }
		    optionMenu = (vcontrolOptionMenu *)dialogItem;
		    optionMenuValue = vuGetOptionMenuValue (optionMenu);
		    vcharExport (vcharScribeSystemExport (optionMenuValue), dbValue);
		    if ( (strncmp (xrefNode -> xrefEpicsName, "DTYP", FLDNAME_SZ)) == 0)
		    {
			// write to DEVTYPE in pvmap
			DB_PutFieldValue (pntRecord, xrefNode -> xrefField, dbValue);
		    }
		    break;
		}

		case DBF_OUTLINK:
		    // there is only 1 output and that is OUT
		    // always write the boardtype from DTYP into the OUT field
		    DB_PutFieldValue (pntRecord, putField, vuCurrentBoard);
		    // if we are a reference then write OUT into the REFERENCE field
		    if (dType == VU_REFERENCE)
		    {
			putField = xrefNode -> xrefRefField;
			textItem = (vtextitem *)dialogItem;
			stringValue = textItem -> GetTextAsString ();
			vcharExport (vcharScribeSystemExport (stringValue), dbValue);
			vstrDestroy (stringValue);
		    }
		    else
		    {
			// make sure to null out the field
			putField = xrefNode -> xrefRefField;
			vmemSet (dbValue, 0, VU_KEY_SIZE);
		    }
		    break;

		case DBF_INLINK:
		    // for DOL, we must put its value to the OUTPUTREF field
		    // other wise it and other INLINKS are the same 
		    if ( (strncmp (xrefNode -> xrefEpicsName, "DOL ", FLDNAME_SZ)) == 0)
		    {
			putField = xrefNode -> xrefRefField;
		    }
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		default:
		    continue;

	    }  // end xrefNode -> xrefEPICSFieldType

	    if (doSave == vTRUE)
	    {
		DB_PutFieldValue (pntRecord, putField, dbValue);
	    }

	}  // end if not in commondialog

    }  // end while iterator

    iterator.Finish ();

    // call base class vuSaveDialog
    vuCommonDialogClass::vuSaveDialog (pointName);
}


vkindTYPED_FULL_DEFINITION(vuTdContainerClass, vcontainer, "vuTdContainerClass");

/*******************************************************************
*
*    vuTdContainerClass override for constructor
*
*******************************************************************/

void vuTdContainerClass::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource pointDictRes = vresourceGet (app -> GetResources (), 
				vunames_vuPointDictionary);
    vresource realRes = vresourceGet (pointDictRes, vunames_vuTdContainer);

    vcontainer *btcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdDTYPContainer);
    boardType = vuBoardTypeMenuClass::RequestSubLoad (btcontainer,
				vunames_DTYP);

    vcontainer *linecontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdLINEContainer);
    line = filteredTextitem::RequestSubLoad (linecontainer,
				vunames_LINE);

    vcontainer *addrcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdADDRContainer);
    addr = filteredTextitem::RequestSubLoad (addrcontainer,
				vunames_ADDR);

    vcontainer *cardcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdCARDContainer);
    card = filteredTextitem::RequestSubLoad (cardcontainer,
				vunames_CARD);

    vcontainer *signalcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdSIGNContainer);
    signal = filteredTextitem::RequestSubLoad (signalcontainer,
				vunames_SIGN);

    vcontainer *valcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdDOLContainer);
    valueLocButton = vuPvNameButtonClass::RequestSubLoad (valcontainer,
				vunames_vuTdDOLButton);
    valueLoc = vuFilteredTextitem::RequestSubLoad (valcontainer,
				vunames_DOL);

    vcontainer *outcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdOUTContainer);
    outputButton = vuPvNameButtonClass::RequestSubLoad (outcontainer,
				vunames_vuTdOUTButton);
    output = vuFilteredTextitem::RequestSubLoad (outcontainer,
				vunames_OUT);

    vcontainer *cnstcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdCNSTContainer);
    constantValue = filteredTextitem::RequestSubLoad (cnstcontainer,
				vunames_CNST);

    vcontainer *zstatecontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdZNAMContainer);
    zeroState = filteredTextitem::RequestSubLoad (zstatecontainer,
				vunames_ZNAM);

    vcontainer *ostatecontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdONAMContainer);
    oneState = filteredTextitem::RequestSubLoad (ostatecontainer,
				vunames_ONAM);

    vcontainer *tdv1container = vcontainer::RequestSubLoad (this,
				vunames_vuTdTDV1Container);
    tdValue1 = filteredTextitem::RequestSubLoad (tdv1container,
				vunames_TDV1);

    vcontainer *tdv2container = vcontainer::RequestSubLoad (this,
				vunames_vuTdTDV2Container);
    tdValue2 = filteredTextitem::RequestSubLoad (tdv2container,
				vunames_TDV2);

    vcontainer *dwelcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdDWELContainer);
    dwellTime = filteredTextitem::RequestSubLoad (dwelcontainer,
				vunames_DWEL);

    vcontainer *alarmcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdAlarmContainer);

    vcontainer *zerosevcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuTdZSVContainer);
    zeroSeverity = vuChoiceGblMenuClass::RequestSubLoad (zerosevcontainer,
				vunames_ZSV);

    vcontainer *onesevcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuTdOSVContainer);
    oneSeverity = vuChoiceGblMenuClass::RequestSubLoad (onesevcontainer,
				vunames_OSV);

    vcontainer *cosvsevcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuTdCOSVContainer);
    cosSeverity = vuChoiceGblMenuClass::RequestSubLoad (cosvsevcontainer,
				vunames_COSV);

    onScan = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OSCN);

    onAlarm = vuCheckBoxClass::RequestSubLoad (this,
				vunames_AENB);

    closedLoop = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OMSL);

    invertOutput = vuCheckBoxClass::RequestSubLoad (this,
				vunames_INVT);

    initTimer = vuCheckBoxClass::RequestSubLoad (this,
				vunames_IACT);

    vcontainer *tdchcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdTDCHContainer);
    timerChoice = vuChoiceRecMenuClass::RequestSubLoad (tdchcontainer,
				vunames_TDCH);

    vcontainer *ltedcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdLTEDContainer);
    edge = vuChoiceRecMenuClass::RequestSubLoad (ltedcontainer,
				vunames_LTED);

    vcontainer *frstcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdFRSTContainer);
    cycleFirstState = vuChoiceRecMenuClass::RequestSubLoad (frstcontainer,
				vunames_FRST);

    vcontainer *alarmdiscontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdADISContainer);
    alarmDisableButton = vuPvNameButtonClass::RequestSubLoad (alarmdiscontainer,
				vunames_vuTdADISButton);
    alarmDisable = vuFilteredTextitem::RequestSubLoad (alarmdiscontainer,
				vunames_ADIS);

    vcontainer *alarmdelcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdALDLContainer);
    alarmDelay = filteredTextitem::RequestSubLoad (alarmdelcontainer,
				vunames_ALDL);

    vcontainer *alrmgrpcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuTdALGPContainer);
    alarmGroup = vuFilteredTextitem::RequestSubLoad (alrmgrpcontainer,
				vunames_ALGP);

    vcontainer::LoadInit (realRes);

    // store ptr to text field for vuPvNameButtonClass
    valueLocButton -> storePvName (valueLoc);
    outputButton -> storePvName (output);
    alarmDisableButton -> storePvName (alarmDisable);

    // set filters
    line -> SetItemFilterFunc (numericFilter);
    addr -> SetItemFilterFunc (numericFilter);
    card -> SetItemFilterFunc (numericFilter);
    signal -> SetItemFilterFunc (numericFilter);
    constantValue -> SetItemFilterFunc (numericFilter);
    zeroState -> SetItemFilterFunc (digitalStateFilter);
    oneState -> SetItemFilterFunc (digitalStateFilter);
    tdValue1 -> SetItemFilterFunc (floatFilter);
    tdValue2 -> SetItemFilterFunc (floatFilter);
    dwellTime -> SetItemFilterFunc (floatFilter);
    alarmDelay -> SetItemFilterFunc (numericFilter);

    valueLoc -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *valText = (filteredText *)valueLoc -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    valText -> SetUserPtr ((void *)valueLoc);

    output -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *outText = (filteredText *)output -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    outText -> SetUserPtr ((void *)output);

    alarmDisable -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *alrmDisText = (filteredText *)alarmDisable -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    alrmDisText -> SetUserPtr ((void *)alarmDisable);

    alarmGroup -> SetItemFilterFunc (vuAlarmGroupFilter);
    filteredText *alarmText = (filteredText *)alarmGroup -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    alarmText -> SetUserPtr ((void *)alarmGroup);

    // build popup menus
    boardType    -> vuBuildBoardTypeMenu (VU_TD);
    zeroSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    oneSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    cosSeverity  -> vuBuildOptionMenu (GBL_ALARM_SEV);
    timerChoice -> vuBuildOptionMenu (VU_TD, REC_TD_CHOICE);
    edge -> vuBuildOptionMenu (VU_TD, REC_TD_EDGE);
    cycleFirstState -> vuBuildOptionMenu (VU_TD, REC_TD_FSTATE);
}

/*******************************************************************
*
*    vuTdContainerClass override for destructor
*
*******************************************************************/

void vuTdContainerClass::Destroy (void)
{
    vcontainer::Destroy ();
}
