#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vugrplistdialogclass.hxx"
#include "vuUtils.hxx"


/*******************************************************************
*
*	ONLINE ADD DIALOG CLASS
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuGrpListDialogClass, vdialog, 
						"vuGrpListDialogClass");

/*******************************************************************
*
*    vuGrpListDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuGrpListDialogClass::LoadInit(vresource res)
{
    grpList = vuGrpListListClass::RequestSubLoad (this,
				vunames_vuGrpListView);
    okButton = vuGrpListOkButtonClass::RequestSubLoad (this,
				vunames_vuGrpListOkButton);
    dismissButton = vuGrpListDismissButtonClass::RequestSubLoad (this,
				vunames_vuGrpListDismissButton);
    // call base class loadInit to finish the job
    vdialog::LoadInit(res);

    // set ptrs
    grpList -> setGrpListDialog (this);
    okButton -> setGrpListDialog (this);
    dismissButton -> setGrpListDialog (this);
}


/*******************************************************************
*
*    vuGrpListDialogClass notify function
*
*******************************************************************/

void vuGrpListDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	// zap buffer so user has nothing to use
	vmemSet ((void *)groupName, 0, sizeof (groupName));
	Close();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuGrpListDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuGrpListDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }

    vliststr *list = (vliststr *)(grpList -> GetList());
    if (list != NULL)
    {
	list -> Destroy ();
    }

    vdialog::Destroy();
}


/*******************************************************************
*
*    vuGrpListDialogClas Open function
*
*******************************************************************/

void vuGrpListDialogClass::vuGrpListOpen(vuMainDialogClass *mainDialg)
{
    mainDialog = mainDialg;

    grpList -> buildList ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    Place(GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuGrpListOkButtonClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuGrpListOkButtonClass, vbutton, 
					"vuGrpListOkButtonClass");

/*******************************************************************
*
*    vuGrpListOkButtonClass Observer function
*
*******************************************************************/

void vuGrpListOkButtonClass::ObserveDialogItem (vevent *event)
{
    vlistSelection *selected = grplistDialog -> grpList -> GetSelection ();
    vliststr *list = (vliststr *)(grplistDialog -> grpList -> GetList ());
    int row, column;
    list -> GetSelectedCell (selected, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	vuMessageDialogCenter (grplistDialog, VU_NOTHING_SELECTED_MSG);
	return;
    }
    vchar *value = (vchar *)(list -> GetCellValue (row, 0));
    vchar *groupName = grplistDialog -> getGroupName ();
    vcharCopyBounded (value, groupName, PROTECTIONGRP_SZ);
    groupName[PROTECTIONGRP_SZ] = 0;
    grplistDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuGrpListDismissButtonClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuGrpListDismissButtonClass, vbutton, 
					"vuGrpListDismissButtonClass");

/*******************************************************************
*
*    vuGrpListDismissButtonClass Observer function
*
*******************************************************************/

void vuGrpListDismissButtonClass::ObserveDialogItem (vevent *event)
{
    vchar *groupName = grplistDialog -> getGroupName ();
    // zap so user has nothing to use
    vmemSet ((void *)groupName, 0, PROTECTIONGRP_SZ);
    grplistDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuGrpListListClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuGrpListListClass, vlistview, 
					"vuGrpListListClass");

/*******************************************************************
*
*    vuGrpListListClass Observe function
*
*******************************************************************/

void vuGrpListListClass::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	int clickCount = GetLastPointerClickCount ();
	if (clickCount == 2)
	{
	    vlistSelection *selected = GetSelection ();
	    vliststr *list = (vliststr *)(GetList ());
	    int row, column;
	    list -> GetSelectedCell (selected, &row, &column);
	    vchar *value = (vchar *)(list -> GetCellValue (row, 0));
	    vchar *groupName = grplistDialog -> getGroupName ();
	    vcharCopyBounded (value, groupName, PROTECTIONGRP_SZ);
	    groupName[PROTECTIONGRP_SZ] = 0;
	    grplistDialog -> Close ();
	    vevent::StopProcessing ();
	}
    }
}


/*******************************************************************
*
*    vuGrpListListClass build function
*
*******************************************************************/

void vuGrpListListClass::buildList (void)
{
    vuMainDialogClass *mainDialog = grplistDialog -> getMainDialog ();
    dataVuApplication *app = mainDialog -> getApplication ();
    vuDatabaseClass *db = app -> getDatabaseClass ();

    vliststr *list = vuGetGroupList (db);
    if (list != NULL)
    {
	SetList (list);
	vlistSelection *selected = GetSelection ();
	list -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	int rowCount = list -> GetRowCount ();
	if (rowCount > 0)
	{
	    list -> SelectRow (selected, 0);
	}
    }
}

