/*dbls.c,v 1.1.1.1 1999/08/03 14:42:33 tis4000 Exp */

/* dbls.c - structure listing program */
/* share/src/db @(#)dbls.c	1.6     6/7/91 */

/*
 *      Author:          Bob Zeiman
 *      Date:            11-7-90
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  mm-dd-yy        iii     Comment
 * .02	03-29-94	ems	make forward refs for func agree with func def
 */


#define SAME 0

#include <vxWorks.h>
#include <stdio.h>
#include <stdioLib.h>
#include <errnoLib.h>
#include <string.h>
#include <types.h>
#include <taskLib.h>

#include "dbRecType.h"
#include "choice.h"
#include "cvtTable.h"
#include "devSup.h"
#include "drvSup.h"
#include "recSup.h"
#include "dbRecDes.h"
#include "dbRecords.h"
#include "errMdef.h"
#include "error.h"
#include "sdrHeader.h"

/* forward references */
static void     DbRecType (FILE *fp, int fflag);
static void     DrvSup (FILE *fp, int fflag);
static void     DevSup (FILE *fp, int fflag);
static void     DbRecDes (FILE *fp, int fflag);
static void     CvtTable (FILE *fp, int fflag);
static void     ChoiceRec (FILE *fp, int fflag);
static void     ChoiceDev (FILE *fp, int fflag);
static void     ChoiceCvt (FILE *fp, int fflag);
static void     ChoiceGbl (FILE *fp, int fflag);
static void     DbErrDes (FILE *fp, int fflag);
static void     DbRecords (FILE *fp, int fflag);
static void     RecSup (FILE *fp, int fflag);

static struct PRTAB {
    char           *msg;	/* selection string */
    void            (*fp) ();	/* print routine */
}               comtab[] =
{
    {
	"dbRecType", DbRecType
    },
    {
	"drvSup", DrvSup
    },
    {
	"devSup", DevSup
    },
    {
	"dbRecDes", DbRecDes
    },
    {
	"cvtTable", CvtTable
    },
    {
	"choiceRec", ChoiceRec
    },
    {
	"choiceDev", ChoiceDev
    },
    {
	"choiceCvt", ChoiceCvt
    },
    {
	"choiceGbl", ChoiceGbl
    },
    {
	"dbErrDes", DbErrDes
    },
    {
	"recSup", RecSup
    },
    {
	"dbRecords", DbRecords
    },
    {
	"FFFF", NULL
    }
};

static int      getSelection (struct PRTAB *pcomtab);

static char     buffer[512];
static char     ibuff[20];
static int      i;
static int      j;
static int      k;
static int      inum;
static int      begNumI;
static int      endNumI;
static int      endNumJ;
static int      begNumJ;

/*
 * LISTSTRUCTURES
 *
 * MAIN PROGRAM
 *
 */
int dbls()
{
    struct PRTAB   *pcomtab = &comtab[0];
    void            (*fup) ();	/* SUBROUTINE POINTER */
    int             pos,
                    fflag;
    FILE           *fp = NULL;
    char            fname[80];
    long            status;


    if (!dbRecType) {
	status = S_sdr_notLoaded;
	errMessage(status, "dbls: Error - dbRecType not loaded");
	return -1;
    }
    begNumI = 0;
    begNumJ = 0;

    pos = getSelection(pcomtab);
    printf("PUT OUTPUT LISTING IN FILE ??? (y|n) : ");
    scanf("%s", ibuff);
    if (ibuff[0] == 'y')
	fflag = 1;
    else
	fflag = 0;

    if (fflag) {
	printf("ENTER FULL_PATH_NAME OF OUTPUT LISTING FILE : ");
	scanf("%s", fname);
	if ((fp = fopen(fname, "a")) == NULL) {
	    printf("dbls: fopen error errno=%d\nfile=%s\n", MYERRNO, fname);
	    printf("TRY:  < cd \"/full_dir_path/.\" > to correct this failure\n");
	    return (-1);
	}
	printf("Warning Note: file %s opened for appending\n", fname);
    }
    fup = *comtab[pos].fp;	/* PREPARE INDIRECT CALL  */
    (*fup) (fp, fflag);		/* AND CALL SUBROUTINE    */
    if (fflag) {
	printf("YOUR OUTPUT IS LOCATED IN FILE %s\n", fname);
	fclose(fp);
    }
    return (0);
}
				/* end of main */
void getRecTypeSel()
{
    endNumI = dbRecType->number;

    printf("DO YOU WANT TO LIST ALL RECORD TYPES ??? (y|n) : ");
    scanf("%s", ibuff);

    if (ibuff[0] != 'y') {
	FOREVER {
	    printf("\nRECTYPE MENU:\n");
	    for (i = 0; i < dbRecType->number; i++) {
		if (dbRecType->papName[i]) {
		    printf("%d\t%s\n", i, dbRecType->papName[i]);
		}
	    }
	    printf("ENTER SELECTION NUMBER  (X-%d) : ", i - 1);
	    scanf("%d", &inum);
	    if ((dbRecType->papName[inum])
		    && (inum >= 0) && (inum < i)) {
		begNumI = inum;
		endNumI = begNumI + 1;
		break;
	    } else {
		printf("Not a selection, Please try again\n");
	    }
	}
    } else {
	for (i = 0; i < dbRecType->number; i++) {
	    /* start at first record definition */
	    if (dbRecType->papName[i]) {
		begNumI = i;
		break;
	    }
	}
    }
    return;
}				/* end of getRecTypeSel */

static int getSelection(pcomtab)
    struct PRTAB   *pcomtab;
{
    int             pos;
    FOREVER {
	printf("\nSTRUCTURE LISTING MENU:\n");
	pos = 0;		/* POS OF FIRST CMD IN TABLE */
	while (strncmp(pcomtab[pos].msg, "FFFF", 4)) {	/* DO UNTIL END OF TABLE    */
	    printf("\t%d\t%s\n", pos + 1, pcomtab[pos].msg);
	    pos++;
	}
	printf("ENTER SELECTION  NUMBER  (1-%d) : ", pos);
	scanf("%d", &inum);

	if (inum > 0 && inum <= pos)
	    break;
	else
	    printf("Not a selection, Please try again\n");

    }
    return (inum - 1);
}				/* end of getSelection */

void bufOut(fp, fflag)
    FILE           *fp;
    int             fflag;
{
    if (!fflag)
	printf("%s\n", buffer);
    else if ((fprintf(fp, "%s\n", buffer)) == EOF) {
	printf("Error writing file with fprintf\n");
#ifdef vxWorks
	taskSuspend (taskIdSelf());
#else
	exit(1);
#endif vxWorks
    }
}				/* end of bufOut */

static void
DbRecType(fp, fflag)
    FILE           *fp;
    int             fflag;
{

    sprintf(buffer, "\n\ndbls: listing the dbRecType structure\n");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]    dbRecType-> recType", (long)&dbRecType, (long)dbRecType);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx\t      number\t    [  %ld  ]\t/* number of types */",
	    (long)&dbRecType->number, dbRecType->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]  **papName\t\t /* pap to record type */",
	    (long)&dbRecType->papName, (long)dbRecType->papName);
    bufOut(fp, fflag);
    for (i = 0; i < dbRecType->number; i++) {
	if (dbRecType->papName[i]) {
	    sprintf(buffer, "%8lx[%8lx]    papName->[%d]->\t\"%s\"",
		    (long)&dbRecType->papName[i],
		    (long)dbRecType->papName[i],
		    i, dbRecType->papName[i]);
	    bufOut(fp, fflag);
	}
    }
}				/* end of DbRecType */

static void
ChoiceGbl(fp, fflag)
    FILE           *fp;
    int             fflag;
{
    if (!choiceGbl) {
	printf("ChoiceGbl: Error - choiceGbl not loaded\n");
	return;
    }
    sprintf(buffer, "\n\ndbls: listing the choiceGbl structure\n");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]    choiceGbl -> arrChoiceSet", (long)&choiceGbl, (long)choiceGbl);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx\t\t  number\t\t[%ld]\t/* number of choice sets */",
	    (long)&choiceGbl->number,
	    choiceGbl->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]\t**papChoiceSet -> choiceSet",
	    (long)&choiceGbl->papChoiceSet,
	    (long)choiceGbl->papChoiceSet);
    bufOut(fp, fflag);
    for (i = 0; i < choiceGbl->number; i++) {
	if (choiceGbl->papChoiceSet[i]) {
	    sprintf(buffer, "%8lx[%8lx]\t  papChoiceSet->[%d]\tINDEX[%d]",
		    (long)&choiceGbl->papChoiceSet[i],
		    (long)choiceGbl->papChoiceSet[i], i, i);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t      number [%ld]\t\t/* number of choices */",
		    (long)&choiceGbl->papChoiceSet[i]->number,
		    choiceGbl->papChoiceSet[i]->number);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t    **papChoice -> \t\"choice string\"",
		    (long)&choiceGbl->papChoiceSet[i]->papChoice,
		    (long)choiceGbl->papChoiceSet[i]->papChoice);
	    bufOut(fp, fflag);
	    for (j = 0; j < choiceGbl->papChoiceSet[i]->number; j++) {
		if (choiceGbl->papChoiceSet[i]->papChoice[j]) {
		    sprintf(buffer, "%8lx[%8lx]\t\t  papChoice->[%d]\t\"%s\"",
			    (long)&choiceGbl->papChoiceSet[i]->papChoice[j],
			    (long)choiceGbl->papChoiceSet[i]->papChoice[j],
			    j,
			    choiceGbl->papChoiceSet[i]->papChoice[j]);
		    bufOut(fp, fflag);
		}
	    }
	}
    }
}				/* end of ChoiceGbl */

static void
ChoiceDev(fp, fflag)
    FILE           *fp;
    int             fflag;
{
    if (!choiceDev) {
	printf("ChoiceDev: Error - choiceDev not loaded\n");
	return;
    }
    getRecTypeSel();
    sprintf(buffer, "\n\ndbls: listing the  choiceDev structure");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]choiceDev-> devChoiceRec ", (long)&choiceDev, (long)choiceDev);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx\t      number\t[%ld]\t\t\t/* number of devChoiceSet */",
	    (long)&choiceDev->number, choiceDev->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]    **papDevChoiceSet -> devChoiceSet ",
	    (long)&choiceDev->papDevChoiceSet,
	    (long)choiceDev->papDevChoiceSet);
    bufOut(fp, fflag);
    for (i = begNumI; i < endNumI; i++) {
	if (choiceDev->papDevChoiceSet[i]) {
	    sprintf(buffer, "\n%8lx[%8lx]\tpapDevChoiceSet->[%d]\t\tRECTYPE \"%s\"",
		    (long)&choiceDev->papDevChoiceSet[i],
		    (long)choiceDev->papDevChoiceSet[i], i,
		    dbRecType->papName[i]);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t   number\t[%ld]\t\t/*number of choices */ ",
		    (long)&choiceDev->papDevChoiceSet[i]->number,
		    choiceDev->papDevChoiceSet[i]->number);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t   **papDevChoice -> devChoice",
		    (long)&choiceDev->papDevChoiceSet[i]->papDevChoice,
		    (long)choiceDev->papDevChoiceSet[i]->papDevChoice);
	    bufOut(fp, fflag);
	    for (j = 0; j < choiceDev->papDevChoiceSet[i]->number; j++) {
		if (choiceDev->papDevChoiceSet[i]->papDevChoice[j]) {
		    sprintf(buffer, "%8lx[%8lx]\t   papDevChoice[%d]",
			 (long)&choiceDev->papDevChoiceSet[i]->papDevChoice[j],
			 (long)choiceDev->papDevChoiceSet[i]->papDevChoice[j], j);
		    bufOut(fp, fflag);
		    sprintf(buffer, "%8lx\t\t       link_type [%ld]",
			    (long)&choiceDev->papDevChoiceSet[i]->papDevChoice[j]->link_type,
			    choiceDev->papDevChoiceSet[i]->papDevChoice[j]->link_type);
		    bufOut(fp, fflag);
		    sprintf(buffer, "%8lx[%8lx]\t\t*pchoice\t\t\"%s\"",
			   (long)&choiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice,
		 	   (long)choiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice,
		 	   choiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice);
		 	   bufOut(fp, fflag);
		}
	    }
	}
    }
}				/* end	ChoiceDev */

static void
ChoiceCvt(fp, fflag)
    FILE           *fp;
    int             fflag;
{

    if (!choiceCvt) {
	printf("ChoiceCvt: Error - choiceCvt not loaded\n");
	return;
    }
    sprintf(buffer, "\n\ndbls: listing the  choiceCvt structure");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx] choiceCvt -> choiceCvt", (long)&choiceCvt, (long)choiceCvt);

    bufOut(fp, fflag);
    sprintf(buffer, "%8lx               number\t    [  %ld  ]\t/* number of choices */",
	    (long)&choiceCvt->number, choiceCvt->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]   **papChoice -> \"choice string\"",
	    (long)&choiceCvt->papChoice, (long)choiceCvt->papChoice);
    bufOut(fp, fflag);
    for (i = 0; i < choiceCvt->number; i++)
	if (choiceCvt->papChoice[i]) {
	    sprintf(buffer, "%8lx[%8lx]     papChoice[%d]\t\"%s\" ",
		    (long)&choiceCvt->papChoice[i], (long)choiceCvt->papChoice[i], i,
		    choiceCvt->papChoice[i]);
	    bufOut(fp, fflag);
	}
    return;
}				/* end of ChoiceCvt */

static void
CvtTable(fp, fflag)
    FILE           *fp;
    int             fflag;
{
    if (!cvtTable) {
	printf("CvtTable: Error - cvtTable not loaded\n");
	return;
    }
    begNumI = 0;
    endNumI = cvtTable->number;

    printf("DO YOU WANT TO LIST ALL CVT TABLES ??? (y|n) : ");
    scanf("%s", ibuff);

    if (ibuff[0] != 'y') {
	FOREVER {
	    printf("\nBRKPOINT MENU:\n");
	    for (i = begNumI; i < endNumI; i++) {
		if (cvtTable->papBrkTable[i]) {
		    sprintf(buffer, "%d\t%s", i, cvtTable->papBrkTable[i]->name);
		    bufOut(fp, 0);
		}
	    }
	    printf("ENTER SELECTION NUMBER  (N-%d) : ", i - 1);
	    scanf("%d", &inum);
	    if ((cvtTable->papBrkTable[inum]) && inum < endNumI) {
		begNumI = inum;
		endNumI = begNumI + 1;
		break;
	    } else {
		printf("Not a selection, Please try again\n");
	    }
	}
    } else {
	for (i = 0; i < cvtTable->number; i++) {
	    if (cvtTable->papBrkTable[i]) {
		begNumI = i;
		break;
	    }
	}
    }
    sprintf(buffer, "\n\ndbls: listing the  cvtTable structure");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx] cvtTable -> arrBrkTable", (long)&cvtTable, (long)cvtTable);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx               number\t    [  %ld  ]\t/* number of break tables */",
	    (long)&cvtTable->number, cvtTable->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]   **papBrkTable -> brkTable",
	    (long)&cvtTable->papBrkTable, (long)cvtTable->papBrkTable);
    bufOut(fp, fflag);
    for (i = begNumI; i < endNumI; i++) {
	if (cvtTable->papBrkTable[i]) {
	    sprintf(buffer, "%8lx[%8lx]     papBrkTable[%d] ",
		    (long)&cvtTable->papBrkTable[i], (long)cvtTable->papBrkTable[i], i);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t *name\t\t\t\"%s\"",
		    (long)&cvtTable->papBrkTable[i]->name,
		    (long)cvtTable->papBrkTable[i]->name,
		    cvtTable->papBrkTable[i]->name);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t  number    [  %ld  ]\t/* number of brkInt in this table */ ",
		    (long)&cvtTable->papBrkTable[i]->number,
		    cvtTable->papBrkTable[i]->number);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t  rawLow\t\t\t%-ld",
		    (long)&cvtTable->papBrkTable[i]->rawLow,
		    cvtTable->papBrkTable[i]->rawLow);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t  rawHigh\t\t\t%-ld",
		    (long)&cvtTable->papBrkTable[i]->rawHigh,
		    cvtTable->papBrkTable[i]->rawHigh);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t**papBrkInt",
		    (long)&cvtTable->papBrkTable[i]->papBrkInt,
		    (long)cvtTable->papBrkTable[i]->papBrkInt);
	    bufOut(fp, fflag);
	    for (j = 0; j < cvtTable->papBrkTable[i]->number; j++) {
		if (cvtTable->papBrkTable[i]->papBrkInt[j]) {
		    sprintf(buffer, "%8lx[%8lx]\t  papBrkInt[%d] -> brkInt",
			    (long)&cvtTable->papBrkTable[i]->papBrkInt[j],
			    (long)cvtTable->papBrkTable[i]->papBrkInt[j], j);
		    bufOut(fp, fflag);
		    sprintf(buffer, "%8lx\t\t      raw\t\t\t%-ld",
			    (long)&cvtTable->papBrkTable[i]->papBrkInt[j]->raw,
			    cvtTable->papBrkTable[i]->papBrkInt[j]->raw);
		    bufOut(fp, fflag);
		    sprintf(buffer, "%8lx\t\t      slope\t\t\t%-.5f",
			    (long)&cvtTable->papBrkTable[i]->papBrkInt[j]->slope,
			    cvtTable->papBrkTable[i]->papBrkInt[j]->slope);
		    bufOut(fp, fflag);
		    sprintf(buffer, "%8lx\t\t      eng\t\t\t%-.5f",
			    (long)&cvtTable->papBrkTable[i]->papBrkInt[j]->eng,
			    cvtTable->papBrkTable[i]->papBrkInt[j]->eng);
		    bufOut(fp, fflag);
		}
	    }
	}
    }
}				/* end of CvtTable */

static void
DevSup(fp, fflag)
    FILE           *fp;
    int             fflag;
{

    if (!devSup) {
	printf("DevSup: Error - devSup not loaded\n");
	return;
    }
    getRecTypeSel();
    sprintf(buffer, "\n\ndbls: listing the devSup structure");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx] devSup -> recDevSup ", (long)&devSup, (long)devSup);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx\t\tnumber\t   [  %ld  ]\t/* number of record types */",
	    (long)&devSup->number, devSup->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]   **papDevSup -> devSup ",
	    (long)&devSup->papDevSup, (long)devSup->papDevSup);
    bufOut(fp, fflag);
    for (i = begNumI; i < endNumI; i++) {
	if (devSup->papDevSup[i]) {
	    sprintf(buffer, "%8lx[%8lx]     papDevSup->[%d]\t\"%s\"",
		    (long)&devSup->papDevSup[i],
		    (long)devSup->papDevSup[i], i,
		    dbRecType->papName[i]);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t   number\t[%ld]\t/* number of dset */",
		    (long)&devSup->papDevSup[i]->number,
		    devSup->papDevSup[i]->number);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]       **dsetName",
		    (long)&devSup->papDevSup[i]->dsetName,
		    (long)devSup->papDevSup[i]->dsetName);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]       **papDset",
		    (long)&devSup->papDevSup[i]->papDset,
		    (long)devSup->papDevSup[i]->papDset);
	    bufOut(fp, fflag);
	    for (j = 0; j < devSup->papDevSup[i]->number; j++) {
		if (devSup->papDevSup[i]->dsetName[j]) {
		    sprintf(buffer, "%8lx[%8lx]\t\tdsetName[%d]\t%s",
			    (long)&devSup->papDevSup[i]->dsetName[j],
			    (long)devSup->papDevSup[i]->dsetName[j],
			    j,
			    devSup->papDevSup[i]->dsetName[j]);
		    bufOut(fp, fflag);
		    sprintf(buffer, "%8lx[%8lx]\t\tpapDset[%d]",
			    (long)&devSup->papDevSup[i]->papDset[j],
			    (long)devSup->papDevSup[i]->papDset[j], j);
		    bufOut(fp, fflag);
		}
	    }
	}
    }
}				/* end of DevSup */

static void
DrvSup(fp, fflag)
    FILE           *fp;
    int             fflag;
{
    if (!drvSup) {
	printf("DrvSup: Error - drvSup not loaded\n");
    }
    sprintf(buffer, "\n\ndbls: listing the drvSup structure");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]  drvSup -> drvSup ", (long)&drvSup, (long)drvSup);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx\t\t number\t[%ld]\t/* number of dset */", (long)&drvSup->number, drvSup->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]     **drvetName ->  \t/* pArr of ptr to drvetName */ ",
	    (long)&drvSup->drvetName, (long)drvSup->drvetName);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]     **papDrvet-> drvet \t/* pArr ptr to drvet */",
	    (long)&drvSup->papDrvet, (long)drvSup->papDrvet);
    bufOut(fp, fflag);
    for (i = 0; i < drvSup->number; i++) {
	if (drvSup->drvetName[i])
	    sprintf(buffer, "%8lx[%8lx]\t   drvetName->[%d]\t\"%s\"",
		    (long)&drvSup->drvetName[i],
		    (long)drvSup->drvetName[i],
		    i, drvSup->drvetName[i]);
	bufOut(fp, fflag);
	sprintf(buffer, "%8lx[%8lx]\t   papDrvet->[%d] /* reserved ptr to drvet */",
		(long)&drvSup->papDrvet[i],
		(long)drvSup->papDrvet[i], i);
	bufOut(fp, fflag);
    }
}				/* end of DrvSup */

static void
DbRecDes(fp, fflag)
    FILE           *fp;
    int             fflag;
{
    char           *ptemp;
    int             allJ = 0;
    if (!dbRecDes) {
	printf("DbRecDes: Error - dbRecDes not loaded\n");
	return;
    }
    getRecTypeSel();
    begNumJ = 0;
    endNumJ = dbRecDes->papRecTypDes[begNumI]->no_fields;
    if ((endNumI - begNumI) < 2) {
	printf("DO YOU WANT TO LIST ALL FIELDS ??? (y|n) : ");
	scanf("%s", ibuff);
	if (ibuff[0] != 'y')
	    FOREVER {
	    printf("\nFIELDNAME MENU FOR RECTYPE %s:\n", dbRecType->papName[begNumI]);
	    for (j = 0; j < endNumJ; j++) {
		printf("%3d %4.4s",
		       dbRecDes->papRecTypDes[begNumI]->sortFldInd[j],
		       (char *)&dbRecDes->papRecTypDes[begNumI]->sortFldName[j]);
		if (!((j + 1) % 8))
		    printf("\n");
	    }
	    printf("\n");
	    printf("ENTER SELECTION NUMBER  (0-%d) : ", j - 1);
	    scanf("%d", &inum);
	    if ((inum >= 0 && inum < j)) {
		begNumJ = inum;
		endNumJ = begNumJ + 1;
		break;
	    } else {
		printf("Not a selection, Please try again\n");
	    }
	    }
    } else
	allJ = 1;

    sprintf(buffer, "\n\ndbls: listing the dbRecDes structure");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]  dbRecDes-> recDes", (long)&dbRecDes, (long)dbRecDes);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx\t\tnumber [%ld] \t/* number of recTypDes */",
	    (long)&dbRecDes->number, dbRecDes->number);
    bufOut(fp, fflag);
    for (i = begNumI; i < endNumI; i++) {

	if (dbRecDes->papRecTypDes[i]) {
	    sprintf(buffer, "\n%8lx[%8lx]    **papRecTypDes-> recTypDes\t\"%s\"",
		    (long)&dbRecDes->papRecTypDes,
		    (long)dbRecDes->papRecTypDes,
		    dbRecType->papName[i]);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t    rec_size [%d]",
		    (long)&dbRecDes->papRecTypDes[i]->rec_size,
		    dbRecDes->papRecTypDes[i]->rec_size);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t    no_fields [%d]",
		    (long)&dbRecDes->papRecTypDes[i]->no_fields,
		    dbRecDes->papRecTypDes[i]->no_fields);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t    no_prompt [%d]",
		    (long)&dbRecDes->papRecTypDes[i]->no_prompt,
		    dbRecDes->papRecTypDes[i]->no_prompt);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t    no_links [%d]",
		    (long)&dbRecDes->papRecTypDes[i]->no_links,
		    dbRecDes->papRecTypDes[i]->no_links);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t    *link_ind",
		    (long)&dbRecDes->papRecTypDes[i]->link_ind,
		    (long)dbRecDes->papRecTypDes[i]->link_ind);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t    *sortFldName",
		    (long)&dbRecDes->papRecTypDes[i]->sortFldName,
		    (long)dbRecDes->papRecTypDes[i]->sortFldName);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t    *sortFldInd",
		    (long)&dbRecDes->papRecTypDes[i]->sortFldInd,
		    (long)dbRecDes->papRecTypDes[i]->sortFldInd);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t   **papFldDes-> fldDes",
		    (long)&dbRecDes->papRecTypDes[i]->papFldDes,
		    (long)dbRecDes->papRecTypDes[i]->papFldDes);
	    bufOut(fp, fflag);

	    if (allJ) {
		endNumJ = dbRecDes->papRecTypDes[i]->no_fields;
		/* expand *link_ind */
		for (j = 0; j < dbRecDes->papRecTypDes[i]->no_links; j++) {
		    sprintf(buffer,"%8lx\tlink_ind[%d] offset=[%d]",
			   (long)&dbRecDes->papRecTypDes[i]->link_ind[j],
			   j,
			   dbRecDes->papRecTypDes[i]->link_ind[j]);
    		    bufOut(fp, fflag);
		}
		/* expand *sortFldName and *sortFldInd */
		ptemp = (char *) dbRecDes->papRecTypDes[i]->sortFldName;
		for (j = 0; j < dbRecDes->papRecTypDes[i]->no_fields; j++) {
		    sprintf(buffer,"[%8lx] sortFldName[%2d]=%4.4s [%8lx] sortFldInd=%2d",
			   (long)ptemp,
			   j,
			   ptemp,
			   (long)&dbRecDes->papRecTypDes[i]->sortFldInd[j],
			   dbRecDes->papRecTypDes[i]->sortFldInd[j]
			);
    		    bufOut(fp, fflag);
		    ptemp += 4;
		}
	    }
	    for (j = begNumJ; j < endNumJ; j++) {
		sprintf(buffer, "\n");
		bufOut(fp, fflag);
		sprintf(buffer, "%8lx[%8lx]\t   **papFldDes->fldDes[%d]",
			(long)&dbRecDes->papRecTypDes[i]->papFldDes[j],
			(long)dbRecDes->papRecTypDes[i]->papFldDes[j], j);
		bufOut(fp, fflag);
		sprintf(buffer, "%8lx\t\t\tfldname\t\t\"%4.4s\"",
			(long)dbRecDes->papRecTypDes[i]->papFldDes[j]->fldname,
			dbRecDes->papRecTypDes[i]->papFldDes[j]->fldname);
		bufOut(fp, fflag);
		sprintf(buffer, "%8lx\t\t\toffset [%d]",
			(long)&dbRecDes->papRecTypDes[i]->papFldDes[j]->offset,
			dbRecDes->papRecTypDes[i]->papFldDes[j]->offset);
		bufOut(fp, fflag);
		sprintf(buffer, "%8lx\t\t\tsize [%d]",
			(long)&dbRecDes->papRecTypDes[i]->papFldDes[j]->size,
			dbRecDes->papRecTypDes[i]->papFldDes[j]->size);
		bufOut(fp, fflag);
		sprintf(buffer, "%8lx\t\t\tspecial [%d]",
			(long)&dbRecDes->papRecTypDes[i]->papFldDes[j]->special,
			dbRecDes->papRecTypDes[i]->papFldDes[j]->special);
		bufOut(fp, fflag);
		sprintf(buffer, "%8lx\t\t\tfield_type [%d]",
			(long)&dbRecDes->papRecTypDes[i]->papFldDes[j]->field_type,
			dbRecDes->papRecTypDes[i]->papFldDes[j]->field_type);
		bufOut(fp, fflag);
		sprintf(buffer, "%8lx\t\t\tdbr_field_type [%d]",
		   (long)&dbRecDes->papRecTypDes[i]->papFldDes[j]->dbr_field_type,
		   dbRecDes->papRecTypDes[i]->papFldDes[j]->dbr_field_type);
		bufOut(fp, fflag);
		sprintf(buffer, "%8lx\t\t\tprocess_passive [%d]",
		   (long)&dbRecDes->papRecTypDes[i]->papFldDes[j]->process_passive,
		   dbRecDes->papRecTypDes[i]->papFldDes[j]->process_passive);
		bufOut(fp, fflag);
		sprintf(buffer, "%8lx\t\t\tchoice_set [%d]",
			(long)&dbRecDes->papRecTypDes[i]->papFldDes[j]->choice_set,
			dbRecDes->papRecTypDes[i]->papFldDes[j]->choice_set);
		bufOut(fp, fflag);
		sprintf(buffer, "%8lx\t\t\tpromptflag [%d]",
			(long)&dbRecDes->papRecTypDes[i]->papFldDes[j]->promptflag,
			dbRecDes->papRecTypDes[i]->papFldDes[j]->promptflag);
		bufOut(fp, fflag);
	    }
	}
    }
}				/* end of DbRecDes */

static void
ChoiceRec(fp, fflag)
    FILE           *fp;
    int             fflag;
{
    if (!choiceRec) {
	printf("ChoiceRec: Error - choiceRec not loaded\n");
	return;
    }
    endNumI = dbRecType->number;

    printf("DO YOU WANT TO LIST ALL RECORD TYPES ??? (y|n) : ");
    scanf("%s", ibuff);

    if (ibuff[0] != 'y') {
	FOREVER {
	    printf("\nRECTYPE MENU:\n");
	    for (i = 0; i < dbRecType->number; i++) {
		if (dbRecType->papName[i]
			&& choiceRec->papArrChoiceSet[i]) {
		    printf("%d\t%s\n", i, dbRecType->papName[i]);
		}
	    }
	    printf("ENTER SELECTION NUMBER  (X-%d) : ", i - 1);
	    scanf("%d", &inum);
	    if ((choiceRec->papArrChoiceSet[inum])
		    && (inum >= 0) && (inum < i)) {
		begNumI = inum;
		endNumI = begNumI + 1;
		break;
	    } else {
		printf("Not a selection, Please try again\n");
	    }
	}
    } else {
	for (i = 0; i < dbRecType->number; i++) {
	    /* start at first record definition */
	    if (choiceRec->papArrChoiceSet[i]) {
		begNumI = i;
		break;
	    }
	}
    }
    sprintf(buffer, "\n\ndbls: listing the  choiceRec structure");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]   choiceRec -> choiceRec", (long)&choiceRec, (long)choiceRec);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx\t\tnumber\t\t[%ld]", (long)&choiceRec->number, choiceRec->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]\t**papArrChoiceSet -> arrChoiceSet",
	    (long)&choiceRec->papArrChoiceSet,
	    (long)choiceRec->papArrChoiceSet);
    bufOut(fp, fflag);

    for (i = begNumI; i < endNumI; i++) {
	if (choiceRec->papArrChoiceSet[i]) {
	    sprintf(buffer, "%8lx[%8lx]\tpapArrChoiceSet[%d] -> arrChoiceSet\t\tRECTYPE \"%s\"",
		    (long)&choiceRec->papArrChoiceSet[i],
		    (long)choiceRec->papArrChoiceSet[i], i,
		    dbRecType->papName[i]);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t    number\t\t[%ld]\t/* number of choice sets */",
		    (long)&choiceRec->papArrChoiceSet[i]->number,
		    choiceRec->papArrChoiceSet[i]->number);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t    **papChoiceSet",
		    (long)&choiceRec->papArrChoiceSet[i]->papChoiceSet,
		    (long)choiceRec->papArrChoiceSet[i]->papChoiceSet);
	    bufOut(fp, fflag);
	    for (j = 0; j < choiceRec->papArrChoiceSet[i]->number; j++) {
		if (choiceRec->papArrChoiceSet[i]->papChoiceSet[j]) {
		    sprintf(buffer, "%8lx[%8lx]\t    papChoiceSet[%d] -> choiceSet \t\tINDEX[%d]",

			   (long)&choiceRec->papArrChoiceSet[i]->papChoiceSet[j],
		    	   (long)choiceRec->papArrChoiceSet[i]->papChoiceSet[j], j, j);
		    bufOut(fp, fflag);
		    sprintf(buffer, "%8lx\t\t\tnumber\t\t[%ld]\t/* number of choices */ ",
		    	    (long)&choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->number,
			    choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->number);
		    bufOut(fp, fflag);
		    sprintf(buffer, "%8lx[%8lx]\t\t**papChoice -> \"string\"",
			    (long)&choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice,
			    (long)choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice);
		    bufOut(fp, fflag);
		    for (k = 0; k < choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->number; k++) {
			if (choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k]) {
			    sprintf(buffer, "%8lx[%8lx]\t\tpapChoice[%d]\t\"%s\"",
				    (long)&choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k],
				    (long)choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k], k,
				    choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k]);
			    bufOut(fp, fflag);
			}
		    }

		}
	    }

	}
    }
}				/* end of ChoiceRec */

static void
DbErrDes(fp, fflag)
    FILE           *fp;
    int             fflag;
{
    if (!dbErrDes) {
	printf("DbErrDes Error - dbErrDes not loaded\n");
	return;
    }
    sprintf(buffer, "\n\ndbls: listing the dbErrDes structure");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx] dbErrDes -> errDes ", (long)&dbErrDes, (long)dbErrDes);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx\t\tnumber\t   [  %ld  ]\t /* dim of err modules */",
	    (long)&dbErrDes->number, dbErrDes->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]   **papErrSet -> errSet ",
	    (long)&dbErrDes->papErrSet, (long)dbErrDes->papErrSet);
    bufOut(fp, fflag);
    for (i = 0; i < dbErrDes->number; i++) {
	if (dbErrDes->papErrSet[i]) {
	    sprintf(buffer, "%8lx[%8lx]     papErrSet->[%d]\tmodule[%d]",
		    (long)&dbErrDes->papErrSet[i],
		    (long)dbErrDes->papErrSet[i], i,
		    i + 501);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t   number\t[%ld]\t/* dim of errSet */",
		    (long)&dbErrDes->papErrSet[i]->number,
		    dbErrDes->papErrSet[i]->number);
	    bufOut(fp, fflag);
	    for (j = 0; j < dbErrDes->papErrSet[i]->number; j++) {
		if (dbErrDes->papErrSet[i]->papName[j]) {
		    sprintf(buffer, "%8lx[%8lx]\t\tpapName[%d]\t\"%s\"",
			    (long)&dbErrDes->papErrSet[i]->papName[j],
			    (long)dbErrDes->papErrSet[i]->papName[j],
			    j,
			    dbErrDes->papErrSet[i]->papName[j]);
		    bufOut(fp, fflag);
		}
	    }
	}
    }
}				/* end of DbErrDes */

static void
DbRecords(fp, fflag)
    FILE           *fp;
    int             fflag;
{
    if (!dbRecords) {
	printf("DbRecords: Error - dbRecords not loaded\n");
	return;
    }
    getRecTypeSel();
    sprintf(buffer, "\n\ndbls: listing the dbRecords structure\n");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]    dbRecords -> recHeader", (long)&dbRecords, (long)dbRecords);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx\t\t  number\t\t[%ld]\t/* number of record types */",
	    (long)&dbRecords->number,
	    dbRecords->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]\t**papRecLoc -> recLoc",
	    (long)&dbRecords->papRecLoc,
	    (long)dbRecords->papRecLoc);
    bufOut(fp, fflag);
    for (i = begNumI; i < endNumI; i++) {
	if (dbRecords->papRecLoc[i]) {
	    sprintf(buffer, "%8lx[%8lx]\t  papRecLoc->[%d]\tRECTYPE[\"%s\"]",
		    (long)&dbRecords->papRecLoc[i],
		    (long)dbRecords->papRecLoc[i],
		    i,
		    dbRecType->papName[i]);
	    bufOut(fp, fflag);

	    sprintf(buffer, "%8lx\t\t      no_records [%ld]\t\t/* no_records of this type */",
		    (long)&dbRecords->papRecLoc[i]->no_records,
		    dbRecords->papRecLoc[i]->no_records);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t      no_used [%ld]\t\t/* no_used of this type */",
		    (long)&dbRecords->papRecLoc[i]->no_used,
		    dbRecords->papRecLoc[i]->no_used);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx\t\t      rec_size [%ld]\t\t/* rec_size in bytes */",
		    (long)&dbRecords->papRecLoc[i]->rec_size,
		    dbRecords->papRecLoc[i]->rec_size);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     *pFirst\t\t\t/* ptr to first record */",
		    (long)&dbRecords->papRecLoc[i]->pFirst,
		    (long)dbRecords->papRecLoc[i]->pFirst);
	    bufOut(fp, fflag);
	}
    }
}				/* end of DbRecords */

static void
RecSup(fp, fflag)
    FILE           *fp;
    int             fflag;
{
    if (!recSup) {
	printf("RecSup: Error - recSup not loaded\n");
	return;
    }
    getRecTypeSel();
    sprintf(buffer, "\n\ndbls: listing the recSup structure\n");
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]   recSup -> recSup", (long)&recSup, (long)recSup);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx\t\t number\t\t[%ld]\t/* number of record types */",
	    (long)&recSup->number,
	    recSup->number);
    bufOut(fp, fflag);
    sprintf(buffer, "%8lx[%8lx]     **papRset -> rset",
	    (long)&recSup->papRset,
	    (long)recSup->papRset);
    bufOut(fp, fflag);
    for (i = begNumI; i < endNumI; i++) {
	if (recSup->papRset[i]) {
	    sprintf(buffer, "%8lx[%8lx]\t papRset->[%d]\t\tRECTYPE[\"%s\"]",
		    (long)&recSup->papRset[i],
		    (long)recSup->papRset[i], i,
		    dbRecType->papName[i]);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     number\t\t/*number of support routines */",
		    (long)&recSup->papRset[i]->number,
		    recSup->papRset[i]->number);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     report\t\t/*print report */",
		    (long)&recSup->papRset[i]->report,
		    (long)recSup->papRset[i]->report);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     init\t\t/*init support */",
		    (long)&recSup->papRset[i]->init,
		    (long)recSup->papRset[i]->init);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     init_record\t/*init record */",
		    (long)&recSup->papRset[i]->init_record,
		    (long)recSup->papRset[i]->init_record);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     process\t\t/*process record */",
		    (long)&recSup->papRset[i]->process,
		    (long)recSup->papRset[i]->process);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     special\t\t/*special processing */",
		    (long)&recSup->papRset[i]->special,
		    (long)recSup->papRset[i]->special);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     get_precision\t/* get_precision of this type */",
		    (long)&recSup->papRset[i]->get_precision,
		    (long)recSup->papRset[i]->get_precision);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     get_value\t\t/*get value field */",
		    (long)&recSup->papRset[i]->get_value,
		    (long)recSup->papRset[i]->get_value);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     cvt_dbaddr\t\t/*cvt dbAddr */",
		    (long)&recSup->papRset[i]->cvt_dbaddr,
		    (long)recSup->papRset[i]->cvt_dbaddr);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     get_array_info\t/* get_array_info of this type */",
		    (long)&recSup->papRset[i]->get_array_info,
		    (long)recSup->papRset[i]->get_array_info);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     put_array_info\t/* put_array_info of this type */",
		    (long)&recSup->papRset[i]->put_array_info,
		    (long)recSup->papRset[i]->put_array_info);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     get_enum_str\t/*get string from enum item*/",
		    (long)&recSup->papRset[i]->get_enum_str,
		    (long)recSup->papRset[i]->get_enum_str);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     get_units \t\t/* get_units of this type */",
		    (long)&recSup->papRset[i]->get_units,
		    (long)recSup->papRset[i]->get_units);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     get_graphic_double /* get_graphic_double of this type */",
		    (long)&recSup->papRset[i]->get_graphic_double,
		    (long)recSup->papRset[i]->get_graphic_double);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     get_control_double /* get_control_double of this type */",
		    (long)&recSup->papRset[i]->get_control_double,
		    (long)recSup->papRset[i]->get_control_double);
	    bufOut(fp, fflag);
	    sprintf(buffer, "%8lx[%8lx]\t     get_enum_strs\t/*get all enum strings */",
		    (long)&recSup->papRset[i]->get_enum_strs,
		    (long)recSup->papRset[i]->get_enum_strs);
	    bufOut(fp, fflag);
	}
    }
}				/* end of recSup */
