// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (C) 1993-1995  Microsoft Corporation.  All Rights Reserved.
//
//  MODULE:   dbClientSvc.c
//
//  PURPOSE:  Implements the body of the service.
//            This starts channelAccess up as a service.
//
//  FUNCTIONS:
//            ServiceStart(DWORD dwArgc, LPTSTR *lpszArgv);
//            ServiceStop( );
//
//  COMMENTS: The functions implemented in dbClientSvc.c are
//            prototyped in service.h
//              
//
//  AUTHOR: Grant Irani
//	    (modified from original by Craig Link - Microsoft Developer Support)
//


#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <tchar.h>
#include "dbClientservice.h"


// this event is signalled when the
// service should end
//
HANDLE  hServerStopEvent = NULL;
HANDLE	dbcHandle;

int	hServerDebugLevel = 0;

/******
	File names for execl
******/

#define DBCNAME	"\\tis4000\\tasks\\dbClient"

//
//  FUNCTION: ServiceStart
//
//  PURPOSE: Actual code of the service
//           that does the work.
//
//  PARAMETERS:
//    dwArgc   - number of command line arguments
//    lpszArgv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    The body of this program is channelAccess's main.
//
VOID ServiceStart (DWORD dwArgc, LPTSTR *lpszArgv)
{
    HANDLE                  hEvents[2] = {NULL, NULL};
    DWORD                   dwWait;

    ///////////////////////////////////////////////////
    //
    // Service initialization
    //

    // report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
        SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        3000))                 // wait hint
        goto cleanup;

    // create the event object. The control handler function signals
    // this event when it receives the "stop" control code.
    //
    hServerStopEvent = CreateEvent(
        NULL,    // no security attributes
        TRUE,    // manual reset event
        FALSE,   // not-signalled
        NULL);   // no name

    if ( hServerStopEvent == NULL)
        goto cleanup;

    hEvents[0] = hServerStopEvent;

    // report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
        SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        3000))                 // wait hint
        goto cleanup;

    /*
    dbClient Manager
    */
    hEvents[1] = dbcHandle = (HANDLE)_spawnv(_P_NOWAIT, DBCNAME, lpszArgv);

    // report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
        SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        3000))                 // wait hint
        goto cleanup;

    // report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
        SERVICE_RUNNING,       // service state
        NO_ERROR,              // exit code
        0))                    // wait hint
        goto cleanup;

    //
    // End of initialization
    //
    ////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////
    //
    // Service is now running, monitor dbClient processes
    //

    // Wait for events. any event indicates shutdown!
    dwWait = WaitForMultipleObjects( 2, hEvents, FALSE, INFINITE );

    // terminate the dbClient processes!
    TerminateProcess (dbcHandle, 0);

  cleanup:

    if (hServerStopEvent)
        CloseHandle(hServerStopEvent);

}


//
//  FUNCTION: ServiceStop
//
//  PURPOSE: Stops the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    If a ServiceStop procedure is going to
//    take longer than 3 seconds to execute,
//    it should spawn a thread to execute the
//    stop code, and return.  Otherwise, the
//    ServiceControlManager will believe that
//    the service has stopped responding.
//    
VOID ServiceStop()
{
    if ( hServerStopEvent )
        SetEvent(hServerStopEvent);
}

//
//  FUNCTION: ServiceDebugIncrement
//
//  PURPOSE: Changes the debug level of the service
//
//  PARAMETERS:
//    change value
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    
VOID ServiceDebugIncrement(int increment)
{
    hServerDebugLevel += increment;
    if (hServerDebugLevel < 0)
	hServerDebugLevel = 0;
}
