/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
	Emulate the operation of 'dbClient' within the IOC. This task will handle
	the 'dbClient' functions of 'downloading' the SDR in this case from PROM
	and the 'downloading' of the point database from E2PROM.

	This tasks looks just like the Host version of dbClient as far as
	'dbserver' is concerned. The messages between the two will be identical
	differing only in the way the data is obtained.

	G.C.T.		29-Mar-1994
*/
#include	<vxWorks.h>
#include	<sockLib.h>
#include	<selectLib.h>
#include	<string.h>
#include	<symLib.h>
#include	<taskLib.h>
#include	<task_params.h>
#include	<stdio.h>
#include	<socket.h>
#include	<ioctl.h>
#include	<ioLib.h>
#include	<netinet/in.h>
#include	<netinet/tcp.h>
#include	<logLib.h>
#include	<systime.h>

#include	"sdrHeader.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"dbRecType.h"
#include	"dbRecDes.h"
#include	"dbRecords.h"
#include	"dbFldTypes.h"
#include	"devSup.h"
#include	"drvSup.h"
#include	"recSup.h"
#include	"cvtTable.h"
#include	"dbPvd.h"
#include	"choice.h"
#include	"error.h"
#include	"link.h"
#include	"redmsgs.h"
#include	"pvd.h"


#ifdef PRIV_AUDIT
/* for the auditing and priviledge functions */
#include "sslShared.h"
#include "iocPriv.h"
#include "auditSystem.h"
extern AUDIT_ID	auditLog;	/* the system log */

static char	X509_dn [X509_MAX_DN]; /* distinguished name from certificate */
#endif PRIV_AUDIT

/*
	global data
*/
int		emSocket = -1;
#ifdef PRIV_AUDIT
SSL		*emConn;
#endif PRIV_AUDIT
int		emListenSocket = -1;
char		*emSDRStartAddr = 0;			/* address of the SDR table in PROM */
char		*emSDREndAddr = 0;			/* address of the SDR table in PROM */
int		emSDRSize;				/* size of SDR table in bytes */
int		emSDROffset = 0;
/*struct timeval	timeOut = {5, 0};*/
extern SYMTAB_ID sysSymTbl;

void		GetRecord (char *pkey, short type, char *buff);
void		PutRecord (char *pkey, short type, char *buff);
void		DelRecord (char *pkey);
short		GetType (char *pkey);
int		initdbe2 (void);
void		DelE2DB (void);

int		dbEmulatorInit (void);
void		dbEmulator (void);
int		emSymHandler (void);
int		emDbRecDelete (void);
unsigned short	DBC_CRCRecord (char *buf, int type_index);
int		emDbDownload (void);
int		emSend (int fd, char *buff, int nbytes);
int		emDbMod (void);
int		emSymHandler (void);
int		emCrcHandler (void);
int		emReadTcp (int fd, char *buff, int nbytes);
char		*emGetRecord (char *name, short *type, short *size);
int		emDbRecDelete (void);
void		emPutRecord (struct dbCommon *rec, short type);
void		emDelRecord (char *name);
void		emDelDatabase (void);

/*
	Defines
*/
#define	EM_TRANSMIT_SIZE	(16384)		/* chg from 1024, ems 03-21-95 */
#define	EM_WINDOW_SIZE		(8)
/*
	typedefs
*/
typedef	struct
	{
	short	recType;
	short	recSize;
	char	*recPtr;
	} EM_RECDESC;
/*
	init function just spawns the main task and then exits
*/

int	dbeTrace;

int	dbEmulatorInit()
{
    /*
	Just spawn the main task from here.
    */
    if (taskSpawn(DBEMULATOR_NAME, DBEMULATOR_PRI, VX_FP_TASK, DBEMULATOR_STACK, (FUNCPTR)dbEmulator,0,0,0,0,0,0,0,0,0,0) == ERROR)
	{
	printErr("Can't spawn dbEmulator.\n");
	return (-1);
	}
    return 0;
}
/*
	This is the 'main' task. Loop around waiting for messages from dbServer
*/
void	dbEmulator()
{
    struct	sockaddr_in sockaddr;
    fd_set	readFds;
    int		tempSocket, stat;
#ifndef PRIV_AUDIT
    int		status;
    int		len;
    int		true = 1;
#endif PRIV_AUDIT
    short	msgType;
    short	oldMsgType = 0;
    /*
    */

#ifdef PRIV_AUDIT
    sslSharedInit ();
#endif PRIV_AUDIT

    emListenSocket = socket(AF_INET,SOCK_STREAM,0);
    sockaddr.sin_family = AF_INET;
    sockaddr.sin_addr.s_addr = INADDR_ANY;
    sockaddr.sin_port = htons(DBSERVER_UDP_PORT);
    bind(emListenSocket,(struct sockaddr *)&sockaddr,sizeof(sockaddr));
    listen(emListenSocket,10);				/* listen for 'dbserver' */

    tempSocket = emListenSocket;
    while(1)
	{
	FD_ZERO(&readFds);
	FD_SET(tempSocket,&readFds);
	stat = select(tempSocket + 1, &readFds, NULL, NULL, NULL);	/* wait for messages from dbServer */
	if(stat > 0)
		{
		if(FD_ISSET(emListenSocket,&readFds))
			{
#ifdef PRIV_AUDIT
			/* promote the new socket to SSL */
			char *hostName;

			if (do_accept (emListenSocket, &emSocket, &hostName) == 0)
			{
			    logMsg ("dbEmulator: accept failed - status = 0x%x\n", errno,0,0,0,0,0);
			    continue;
			}
			if ((emConn = setup_ssl_accept (emSocket)) == NULL)
			{
			    logMsg ("dbEmulator: SSL build failed for %s - status = 0x%x\n",
				      (int)hostName, errno,0,0,0,0);
			    close (emSocket);
			    free (hostName);
			    continue;
			}
			if (!accept_ssl_connection_timed (emConn, 10/*seconds*/))
			{
			    logMsg ("dbEmulator: error engaging SSL with %s\n", (int)hostName,0,0,0,0,0);
			    sslSharedCloseConn (emConn);
			    free (hostName);
			    continue;
			}
			/* validate user privilege */
			GetPeerCN (emConn, X509_dn, sizeof(X509_dn));
			if (GetGeneralPrivHost (X509_dn, "DB_") < 1)
			{
			    char errMsg [80];

			    GetPeerDN (emConn, X509_dn, sizeof(X509_dn));
			    sprintf (errMsg, "db Connect privilege error from %s.", hostName);
			    SecurityAlarm (X509_dn, errMsg, 0);
			    if (auditLog)
				{
				AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
						taskName(taskIdSelf()), errMsg);
				}
			    sslSharedCloseConn (emConn);
			    continue;
			}
			else
			{
			    GetPeerDN (emConn, X509_dn, sizeof(X509_dn));
			    if (auditLog)
			    {
				char msg [80];

				sprintf (msg, "db connect from %s.", hostName);
				AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
					       taskName(taskIdSelf()), msg);
			    }
			}
#else
			len = sizeof(sockaddr);
			emSocket = accept(emListenSocket, (struct sockaddr *)&sockaddr,&len);
			if (dbeTrace)
			    {
			    printf("New dbEmulator socket with IP 0x%lX, port %d\n", 
				    sockaddr.sin_addr.s_addr, sockaddr.sin_port);
			    len = sizeof(sockaddr);
			    getsockname(emSocket, (struct sockaddr *)&sockaddr, &len);
			    printf("My addr is IP 0x%lX, port %d\n", 
				    sockaddr.sin_addr.s_addr, sockaddr.sin_port);
			    }
			status = ioctl(emSocket, FIONBIO, (int)&true);
			if (status == ERROR)
			    logMsg ("dbEmulator can't set non blocking IO on socket to dbServer, %d",
			             errno,0,0,0,0,0);
#endif PRIV_AUDIT
			tempSocket = emSocket;
			continue;
			}
		oldMsgType = msgType;
		if((emReadTcp(emSocket,(char *)&msgType,sizeof(short))) == ERROR)
			{
#ifdef PRIV_AUDIT
			SSL *conn = sslSharedGetConn (emSocket);

			if (auditLog)
			    {
			    GetPeerDN (conn, X509_dn, sizeof(X509_dn));
			    AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
					   taskName(taskIdSelf()), "Connection closed to dbServer");
			    }
			sslSharedCloseConn (conn);
#else
			close(emSocket);
#endif PRIV_AUDIT
			tempSocket = emListenSocket;
			emSocket = -1;
			continue;
			}
		msgType = ntohs(msgType);
		switch(msgType)
			{
			case DBC_MSG_DWNLDREQ:
				if(emDbDownload() == ERROR)
					{
#ifdef PRIV_AUDIT
					SSL *conn = sslSharedGetConn (emSocket);

					if (auditLog)
					    {
					    GetPeerDN (conn, X509_dn, sizeof(X509_dn));
					    AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
							   taskName(taskIdSelf()), "Connection closed to dbServer");
					    }
					sslSharedCloseConn (conn);
#else
					close(emSocket);
#endif PRIV_AUDIT
					tempSocket = emListenSocket;
					emSocket = -1;
					continue;
					}
				break;
			case DBC_MSG_OLDELRESP:
				if(emDbRecDelete() == ERROR)
					{
#ifdef PRIV_AUDIT
					SSL *conn = sslSharedGetConn (emSocket);

					if (auditLog)
					    {
					    GetPeerDN (conn, X509_dn, sizeof(X509_dn));
					    AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
							   taskName(taskIdSelf()), "Connection closed to dbServer");
					    }
					sslSharedCloseConn (conn);
#else
					close(emSocket);
#endif PRIV_AUDIT
					tempSocket = emListenSocket;
					emSocket = -1;
					continue;
					}
				break;
			case DBC_MSG_DBMODREQ:
				if(emDbMod() == ERROR)
					{
#ifdef PRIV_AUDIT
					SSL *conn = sslSharedGetConn (emSocket);

					if (auditLog)
					    {
					    GetPeerDN (conn, X509_dn, sizeof(X509_dn));
					    AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
							   taskName(taskIdSelf()), "Connection closed to dbServer");
					    }
					sslSharedCloseConn (conn);
#else
					close(emSocket);
#endif PRIV_AUDIT
					tempSocket = emListenSocket;
					emSocket = -1;
					continue;
					}
				break;
			case DBC_MSG_SYMRESPONSE:
			case DBC_MSG_NEWSYM:
				if(emSymHandler() == ERROR)
					{
#ifdef PRIV_AUDIT
					SSL *conn = sslSharedGetConn (emSocket);

					if (auditLog)
					    {
					    GetPeerDN (conn, X509_dn, sizeof(X509_dn));
					    AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
							   taskName(taskIdSelf()), "Connection closed to dbServer");
					    }
					sslSharedCloseConn (conn);
#else
					close(emSocket);
#endif PRIV_AUDIT
					tempSocket = emListenSocket;
					emSocket = -1;
					continue;
					}
				break;
			case DBC_MSG_CRCRESPONSE:
				if(emCrcHandler() == ERROR)
					{
#ifdef PRIV_AUDIT
					SSL *conn = sslSharedGetConn (emSocket);

					if (auditLog)
					    {
					    GetPeerDN (conn, X509_dn, sizeof(X509_dn));
					    AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
							   taskName(taskIdSelf()), "Connection closed to dbServer");
					    }
					sslSharedCloseConn (conn);
#else
					close(emSocket);
#endif PRIV_AUDIT
					tempSocket = emListenSocket;
					emSocket = -1;
					continue;
					}
				break;
			default:
			        logMsg ("Invalid msgType %d(%x), previous %d(%x)\n",
					msgType,msgType,oldMsgType,oldMsgType,0,0);
				break;
			}
		}
	    else if (dbeTrace)
	  	perror("dbEmulator main select loop error");
	}
    }

/*
	Send database records obtained from the E2PROM to 'dbServer'. The first request will contain a NULL
	point name. Subsequent requests will contain the name of the last point sent.
*/
int	emDbDownload (void)
{
    DBC_DWNLD_REQUEST	msg;
    DBC_DWNLD_RESP	resp;
    EM_RECDESC		rPtr[EM_WINDOW_SIZE];
    char		*rec, name[PVNAME_SZ+1];
    short	msgType = htons(DBC_MSG_DWNLDRESP);
    int	i;

    resp.msgNumberSent = 0;													/* start with zero */
    if((emReadTcp(emSocket,(char *)&msg,DBC_REQSIZE)) == ERROR)		/* get the message from dbServer */
	return(ERROR);
    strcpy(name,msg.msgLastName);
    rec = emGetRecord(name,
			&rPtr[resp.msgNumberSent].recType,
			&rPtr[resp.msgNumberSent].recSize);		/* "" says get FIRST record */
    while(rec && (resp.msgNumberSent < EM_WINDOW_SIZE))
	{
	rPtr[resp.msgNumberSent].recPtr = rec;
	resp.msgNumberSent++;
	rec = emGetRecord(name,
			&rPtr[resp.msgNumberSent].recType,
			&rPtr[resp.msgNumberSent].recSize);		/* point to next record. 'name' has been modified */
	}
    if (emSend(emSocket,(char *)&msgType,sizeof(short)) == ERROR)
	return (ERROR);							/* send response message type */
    resp.msgNumberSent = htons(resp.msgNumberSent);
    if (emSend(emSocket,(char *)&resp,DBC_RESPSIZE) == ERROR)
	return (ERROR);							/* tell 'em how many points */
    for(i = 0;i < ntohs(resp.msgNumberSent);i++)
	{
	/* convert recType for transmission */
	rPtr[i].recType = htons(rPtr[i].recType);
	if (emSend(emSocket,(char *)&rPtr[i].recType,sizeof(rPtr[i].recType)) == ERROR)
	    return (ERROR);
	if (emSend(emSocket,rPtr[i].recPtr,rPtr[i].recSize) == ERROR)
	    return (ERROR);
	free(rPtr[i].recPtr);
	}
    return(!ERROR);
}

int emSend(fd, buff, nbytes)
int	fd;
char	*buff;
int	nbytes;
{
    int		bytesSent;
#ifdef PRIV_AUDIT
    SSL		*conn = sslSharedGetConn (fd);
#endif PRIV_AUDIT

    while (nbytes)
	{
#ifdef PRIV_AUDIT
	bytesSent = SSL_write (conn, buff, nbytes);
#else
	bytesSent = send(fd, buff, nbytes, 0);
#endif PRIV_AUDIT
	if (bytesSent == ERROR)
	    {
	    if (errno == EWOULDBLOCK)
		{
		int	txBuff;
		int	intSize = sizeof(txBuff);

		getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&txBuff, &intSize);
		txBuff += nbytes;
		if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&txBuff, sizeof(txBuff)) == ERROR)
		    {
		    logMsg("Can't expand dbEmulator socket to dbServer\n",
				0, 0, 0, 0, 0, 0);
		    return (ERROR);
		    }
		continue;
		}
	    else
		{
		logMsg("dbEmulator write error to dbServer, %d",
			errno,0,0,0,0,0);
		return (ERROR);
		}
	    }
	else if (bytesSent == 0)
	    {
	    logMsg("Zero length write from dbEmulator to dbServer\n",
		    0,0,0,0,0,0);
	    return (ERROR);
	    }
	else
	    {
	    nbytes -= bytesSent;
	    buff += bytesSent;
	    }
	}
    return (OK);
}

/*
	Apply changes to supplied fields. This is similar to dbClient only MUCH simpler(i.e NO dBase!)
*/
int	emDbMod (void)
{
    DBC_DBMODHDR	hdr;
    short	typeIndex, recSize;
    int		fieldName;
    char	fieldValue[100], *recBuff;
    struct	fldDes	*fldDesc;
    struct recTypDes	*recDes;
    unsigned short crc;

    if((emReadTcp(emSocket,(char *)&hdr,DBC_MODHDRSIZE)) == ERROR)
	return(ERROR);
    pvdGetRec(hdr.dbmodName,&typeIndex,&recSize);					/* recSize justa dummy here */
    recDes = GET_PRECTYPDES(typeIndex);
    recSize = recDes -> rec_size;
    recBuff = malloc(recSize);
    GetRecord(hdr.dbmodName,typeIndex,recBuff);						/* get the record and the point type index for SDR reference */
    hdr.dbmodNFields = ntohs(hdr.dbmodNFields);
    while(hdr.dbmodNFields--)								/* loop through all supplied fields */
	{
	if((emReadTcp(emSocket,(char *)&fieldName,sizeof(int))) == ERROR)		/* get the EPICS field name (4 bytes) */
		return(ERROR);
	fldDesc = pvdGetFld(typeIndex,(char *)&fieldName);
	if((emReadTcp(emSocket,fieldValue,fldDesc->size)) == ERROR)			/* get the data */
		return(ERROR);
	/* Record is stored as received (in net std order) */
	memcpy(&recBuff[fldDesc->offset],fieldValue,fldDesc->size);			/* MOD the record */
	}
    /*
	  GCT 20-Jun-1997

	  calculate the CRC after a field mod and store in record for later. This
	  is so we can load this record from BBM on those machines that have it and
	  verify the CRCs
    */
    fldDesc = pvdGetFld(typeIndex,"RCRC");
    /* use off the wire CRC (for values in Net std order) */
    crc = DBC_CRCRecord(&recBuff[0],typeIndex);
    crc = htons (crc);
    memcpy(&recBuff[fldDesc->offset],(char *)&crc,fldDesc->size);			/* store the CRC */
    emPutRecord((struct dbCommon *)recBuff,typeIndex);					/* save changes */
    free(recBuff);
    return(!ERROR);
}
/*
	Flush emSocket for symbol table responses
*/
int	emSymHandler()
{
    DBC_SYMRESPHDR	hdr;
    DBC_SYMRESPNAME	name;

    if((emReadTcp(emSocket,(char *)&hdr,sizeof(DBC_SYMRESPHDR))) == ERROR)			/* get sym header */
	return(ERROR);
    hdr.numSymbols = ntohs(hdr.numSymbols);
    while(hdr.numSymbols--)
	{
	if((emReadTcp(emSocket,(char *)&name,sizeof(DBC_SYMRESPNAME))) == ERROR)	/* and name+value */
	    return(ERROR);
	}
    return(!ERROR);
}
/*
	Flush emSocket for Crc table responses
*/
int	emCrcHandler()
{
    DBC_CRCRESPHDR	hdr;
    DBC_CRCINFO	info;

    if((emReadTcp(emSocket,(char *)&hdr,sizeof(DBC_CRCRESPHDR))) == ERROR)			/* get sym header */
	return(ERROR);
    hdr.numCrcs = ntohs(hdr.numCrcs);
    while(hdr.numCrcs--)
	{
	if((emReadTcp(emSocket,(char *)&info,sizeof(DBC_CRCINFO))) == ERROR)	/* and name+value */
	    return(ERROR);
	}
    return(!ERROR);
}
/*
	Read ALL data expected from socket
*/
int	emReadTcp(fd,buff,nbytes)
int		fd;
char		*buff;
int		nbytes;
{
    int		bytesWanted = nbytes;
    int		bytesGot,err;
    fd_set	readFds;
#ifdef PRIV_AUDIT
    SSL		*conn = sslSharedGetConn (fd);
#endif PRIV_AUDIT

    FD_ZERO(&readFds);
    FD_SET(fd,&readFds);		/* the one I am interested in */
    err = !ERROR;
    while (nbytes > 0)
	{
#ifdef PRIV_AUDIT
	/* Must read BEFORE select in case data is in BIO */
        if ((bytesGot = SSL_read (conn, buff, nbytes)) > 0)
	    {
	    buff += bytesGot;
	    nbytes -= bytesGot;
	    nbytes -= bytesGot;
	    buff += bytesGot;
	    err = bytesWanted;
	    continue;
	    }
	else
	    {
	    long err;
	    int status = SSL_get_error (conn, bytesGot);

	    err = ERR_get_error ();
	    if (err != 0)
		logMsg ("%s\n", (int)ERR_error_string (err, NULL),0,0,0,0,0);
	    switch (status)
		{
		    case SSL_ERROR_WANT_WRITE:
		    case SSL_ERROR_NONE:
		    case SSL_ERROR_WANT_READ:
		    case SSL_ERROR_WANT_CONNECT:
		    case SSL_ERROR_WANT_X509_LOOKUP:
			/* benign, just want to write/read for state machine */
			break;
		    case SSL_ERROR_SSL:
			if (errno == EWOULDBLOCK)
			    break;
		    case SSL_ERROR_SYSCALL:
		    case SSL_ERROR_ZERO_RETURN:
		    default:
			if (dbeTrace)
			    perror("dbEmulator recv error in emReadTcp");
			return ERROR;
		}
	    }
#endif PRIV_AUDIT
	if ((bytesGot = select(fd + 1, &readFds, NULL, NULL, NULL)) <= 0)
	    {
	    if (dbeTrace)
		perror ("dbEmulator select error in emReadTcp");
	    err = ERROR;
	    break;
	    }
#ifdef PRIV_AUDIT
        if ((bytesGot = SSL_read (conn, buff, nbytes)) <= 0)
#else
	if ((bytesGot = recv(fd, buff, nbytes, 0)) == 0)
#endif PRIV_AUDIT
	if ((bytesGot = recv(fd, buff, nbytes, 0)) == ERROR)
	    {
	    if (dbeTrace)
		perror("dbEmulator recv error in emReadTcp");
	    err = ERROR;
	    break;
	    }
	if (!bytesGot)
	    {
	    if (dbeTrace)
		perror("dbEmulator zero length read in emReadTcp");
	    err = ERROR;
	    break;
	    }
	nbytes -= bytesGot;
	buff += bytesGot;
	err = bytesWanted;
	}
    return (err);
}
/*
	This is a wrapper function which contains the real function to
	get a pointer to the next record in EEROM
*/
char	*emGetRecord(name,type,size)
char	*name;
short	*type;
short	*size;
{
    char	*rec;
    struct	recTypDes	*recDes;

    if((*type = GetType(name)))
	{
	recDes = GET_PRECTYPDES(*type);
	*size = recDes -> rec_size;
	rec = (char *)malloc(*size);
	GetRecord(name,*type,rec);
	return(rec);
	}
    return(0);
}
int	emDbRecDelete(void)
{
    DBC_DELRESP	delHdr;

    if((emReadTcp(emSocket,(char *)&delHdr,sizeof(DBC_DELRESP))) == ERROR)
	return(ERROR);
    DelRecord(delHdr.deletedName);
    return 0;
}

void	emPutRecord(rec,type)
struct	dbCommon	*rec;
short			type;
{
    PutRecord ((char *)rec, type, (char *)rec);
}

void	emDelRecord(name)
char	*name;
{
    DelRecord(name);
}

void	emDelDatabase (void)
{
    DelE2DB();
}

