/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include "typeindex.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"

static	char	*valFields[]   	= { "",".VAL"};
static	char	*biFields[]	= { "",".VAL",".TCNT"};
static	char	*accumFields[] 	= { "",".VAL",".FSMP",".FTOT",".FAVG",".FMIN",".FMAX",".RSMP",".RTOT",".RAVG",".RMIN",".RMAX"};
static	char	*deadtimeFields[]	= { "",".VAL",".CAVE",".CMAX",".CMIN",".CTOT"};
static	char	*pidFields[]   	= { "",".VAL",".ASTP",".MVAL"};
static	char	*selFields[]   	= { "",".VAL",".SELM"};
static	char	*fanFields[]   	= { "",".VAL",".SELM"};
static	char	*udbFields[]   	= { "",".VAL",".FP01",".FP02",".FP03",".FP04",".FP05",".FP06",".FP07",".FP08",
			       	    ".BL01",".BL02",".BL03",".BL04",".BL05",".BL06",".BL07",".BL08",".NCMD"};
static	char	*patFields[]   	= { "",".VAL",".ASTP"};	/* ems 9/12/95 */
static	char	*pumpFields[]	= { "",".VAL",".ACRT"};

char **dbFindInArreduc(DB_TABLE_PTR arTable,char *pvName,int typeIndex,int *numStrings)
{
  char **arr;
  char **retArray;
  char name[PVNAME_SZ+FLDNAME_SZ+1];
  char searchName[PVNAME_SZ+FLDNAME_SZ+1];
  char foundName[PVNAME_SZ+FLDNAME_SZ+1];
  char searchKey[PVNAME_SZ+3+1];
  char foundRate[6];
  char foundMeth[4];
  char foundStr[80];
  char foundArch;
  char *dot;
  int count;
  int i;
  int j;
  int type;
  int status;
  DB_RECORD_PTR rec;
  DB_FIELD_PTR nameField;
  DB_FIELD_PTR rateField;
  DB_FIELD_PTR methField;
  DB_FIELD_PTR archField;

  strcpy(name,pvName);
  /* if called with name.field then drop field */
  if((dot = strchr(name,'.')) != 0) *dot = 0;
  switch(typeIndex)
    {
    case PT_AI:
    case PT_AO:
    case PT_BO:
    case PT_CALC:
    case PT_MBBI:
    case PT_MBBO:
    case PT_LEADLAG:
    case PT_RAMP:
    case PT_PIPE:
    case PT_TANK:
    case PT_VALVE:
    case PT_POT:
    case PT_PTO:
    case PT_TD:
    case PT_SIGCHAR:
      arr	= valFields;
      count = 2;
      break;
    case PT_BI:
      arr	= biFields;
      count = 3;
      break;
    case PT_FANOUT:
      arr = fanFields;
      count = 3;
      break;
    case PT_PID:
      arr = pidFields;
      count = 4;
      break;
    case PT_SEL:
      arr = selFields;
      count = 3;
      break;
    case PT_ACCUM:
      arr = accumFields;
      count = 12;
      break;
    case PT_DEADTIME:
      arr = deadtimeFields;
      count = 6;
      break;
    case PT_UDB:
      arr = udbFields;
      count = 19;
      break;
    case PT_PAT:			/* ems 9/12/95 */
      arr = patFields;
      count = 3;
      break;
    case PT_PUMP:
      arr = pumpFields;
      count = 3;
      break;
    default:
      arr = 0;
      count = 0;
    }
  /*
    Using all possible fields for this point name, search for records which match.
    First count the number of matches so we can allocate an array of pointers
    to the found strings.
  */
  nameField = DB_MapField(arTable,"POINTNAME");
  rateField = DB_MapField(arTable,"RATE");
  methField = DB_MapField(arTable,"METHOD");
  archField = DB_MapField(arTable,"ARCHIVE");
  DB_UseKey(arTable,"POINTMETH");
  *numStrings = 0;
  retArray = NULL;
  for(i = 0;i < count;i++)
    {
      sprintf(searchName,"%s%s\0",name,arr[i]);
      sprintf(searchKey,"%-*.*s%-*.*s\0",PVNAME_SZ,PVNAME_SZ,searchName,3,3,"");
      rec = (DB_RECORD_PTR)DB_SeekStatus(arTable,searchKey,&status);
      if(status == DB_AFTER) rec = (DB_RECORD_PTR)1;
      while(rec)
	{
	  DB_GetFieldValue(rec,nameField,foundName,&type);
	  if(!strcmp(foundName,searchName))
	  {
	    (*numStrings)++;
	  }
	  else
	  {
	    break;
	  }
	  rec = (DB_RECORD_PTR)DB_SeekNext(arTable);
	}
    }
  /*
    If *numStrings is zero then we're done. else collect the strings into the array
    and return it;
  */
  if(*numStrings != 0)
    {
      retArray = calloc(*numStrings+1,sizeof(char *));	/* allocate and leave room for NULL pointer */
      j = 0;
      for(i = 0;i < count;i++)
	{
	  sprintf(searchName,"%s%s\0",name,arr[i]);
	  sprintf(searchKey,"%-*.*s%-*.*s\0",PVNAME_SZ,PVNAME_SZ,searchName,3,3,"");
	  rec = (DB_RECORD_PTR)DB_SeekStatus(arTable,searchKey,&status);
	  if(status == DB_AFTER) rec = (DB_RECORD_PTR)1;
	  while(rec)
	    {
	      DB_GetFieldValue(rec,nameField,foundName,&type);
	      if(!strcmp(foundName,searchName))
		{
		  DB_GetFieldValue(rec,rateField,foundRate,&type);
		  DB_GetFieldValue(rec,methField,foundMeth,&type);
		  DB_GetFieldValue(rec,archField,&foundArch,&type);
		  sprintf(foundStr,"%s,%s,%s,%s",foundName,foundRate,foundMeth,(foundArch ? "T":"F"));
		  retArray[j] = malloc(strlen(foundStr)+1);
		  sprintf(retArray[j++],"%s\0",foundStr);
		}
	      else
	        {
  		  break;
	        }
	      rec = (DB_RECORD_PTR)DB_SeekNext(arTable);
	    }
	}
      retArray[j] = NULL;				/* last one is a NULL pointer */
    }
return(retArray);
}
