/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include	<DBDefines.h>
void		dbClrDeleteFlag();
/*
  this function return a success (0) if a record could either be created
  by 'd4append_blank' or from a previously deleted(i.e. blanked) record.

  The following is the logic(?) of this function:

  1. If 'name' is not NULL then.
  2. Attempt to locate record in given file. This would succeed in PVMAP
     and fail everywhere else.
  3. If 'name' record is found then see if marked for delete.
  4. If marked for delete then undelete it and return else just return.
  5. If 'name' record is not found, go to top of file and use blank record.
  6. If no blank (deleted) record then create new one (d4append_blank).
  7. If there is a blank record then just undelete it and return.
  8. If 'name' is NULL create new record(d4append_blank) and return.
*/
int    dbObtainRecord(D4DATA *table,char *name)
{
  T4TAG	*tag;
  int  stat;

  if(name)					/* try to find pre-existing record */
    {
      if((stat = d4seek(table,name)) == 0)	/* if found then see if delete flag set */
	{
	  if(dbDeleted(table))		/* if deleted then return record */
	    {
	      if(d4deleted(table))	/* remove dBase delete mark */
		d4recall(table);
	      dbClrDeleteFlag(table);	/* set 'delete flag' to FALSE (if there is one) */
	      if(stat != -1) dbWriteLog(__FILE__,__LINE__,table,"REUSE",0,name,0);
	      d4unlock_all(table);
	      return(stat);
	    }
	}
      else
	{
	  stat = d4top(table);
	  if(stat == 0)
	    {
	      if(dbDeleted(table))	       	/* is record blank ? */
		{
		  if(d4deleted(table))		/* remove dBase delete mark */
		    d4recall(table);
		  dbClrDeleteFlag(table);       /* set 'delete flag' to FALSE (if there is one) */
		}
	      else
		{
		  tag = d4tag_selected(table);	/* get the selected tag */
		  d4tag_select(table,0);        /* record number processing */
		  d4go_eof(table);     	      	/* recalculate EOF(number of records) and position file there */
		  stat = d4append_blank(table); /* append at new EOF */
		  d4tag_select(table,tag);      /* restore original tag */
		  if(stat != 0) stat = -1;     	/* could not append then error */
		}
	    }
	  else if(stat == r4eof)
	    {					/* file empty then create new record */
	      stat = d4append_blank(table);
	      if(stat != 0) stat = -1;		/* could not append then error */
	    }
	  else
	    {
	      stat = -1;
	    }
	}
    }
  else
    {
      stat = d4top(table);
      if(stat == 0)
	{
	  if(dbDeleted(table))	       	/* is record blank ? */
	    {
	      if(d4deleted(table))		/* remove dBase delete mark */
		d4recall(table);
	      dbClrDeleteFlag(table);       /* set 'delete flag' to FALSE (if there is one) */
	    }
	  else
	    {
	      tag = d4tag_selected(table);	/* get the selected tag */
	      d4tag_select(table,0);        /* record number processing */
	      d4go_eof(table);     	      	/* recalculate EOF(number of records) and position file there */
	      stat = d4append_blank(table); /* append at new EOF */
	      d4tag_select(table,tag);      /* restore original tag */
	      if(stat != 0) stat = -1;     	/* could not append then error */
	    }
	}
      else if(stat == r4eof)
	{					/* file empty then create new record */
	  stat = d4append_blank(table);
	  if(stat != 0) stat = -1;		/* could not append then error */
	}
      else
	{
	  stat = -1;
	}
    }
  if(stat != -1) dbWriteLog(__FILE__,__LINE__,table,"NEW",0,name,0);
  d4unlock_all(table);
  return(stat);
}
/*
	This function checks the second byte in the current record for a space.
	If it finds one it returns a 1 is returned else it returns a 0. The
	second byte is used because the first byte is reserved by dBase for their
	own deleted record indicator.

	This function will also be used within the following functions:

		DB_Bottom
		DB_SeekNext
		DB_SeekPrev
		DB_Top
*/
int	dbDeleted(D4DATA *table)
{
F4FIELD	*del;

if((del = d4field(table,FLD_DELETEFLAG)) != 0)	/* must me pvmap */
  return(f4true(del));
else
  return(d4deleted(table));			/* otherwise record is blanked and CB delete flag is set */
}
void	dbClrDeleteFlag(table)
D4DATA	*table;
{
F4FIELD	*del;

if((del = d4field(table,FLD_DELETEFLAG)) != 0)
  f4assign_char(del,'F');		       	/* set delete flag to false */
return;
}
