/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
	Structures and defines for 'dpa' functions

	GCT	17-Oct-1994
*/
#ifndef	INCdpaH
#define	INCdpaH

#include	<sys/types.h>
#include	"lstLib.h"				/* so everyone gets it */
#include	"dbDefs.h"
#include	"tsDefs.h"
#include	"agentMsgs.h"

#define	HSRV_PORT		8595

#define	DPA_MAX_CHARS	4096		/* BIG buffer for socket reads */
#define	DPA_CHAN_DATA_SIZE	4096	/* place to put data and parameters */
#define	DPA_MAX_VALUE_SIZE	80     	/* VAL data size */
#define DPA_USER_DATA_SIZE      256	/* user data area size */

#define	DPA_NORMAL		 0			/* dpaInit return codes */
#define	DPA_NORSRV		-1			/* cant spawn rsrvTask */
#define	DPA_NOFORK		-2			/* can't fork to execl rsrvTask */
#define	DPA_NOCONN		-3			/* can't accept new socket */
#define	DPA_TIMEOUT		-4			/* select timer expired with nothing to read */
#define	DPA_SELECTERR	-5			/* error from select function */
#define	DPA_READERR		-6			/* error from recv */
#define	DPA_ZEROREAD	-7			/* zero length read from socket */
#define	DPA_NULLHANDLE	-8			/* null handle specfied */
#define	DPA_OPENERR		-9			/* can't open config file */
/*
	define message types. Actually these are the same as the
	the corresponding ones in agentMsgs
*/
#define	APP_OPEN		CHAN_BUILD		
#define	APP_READ		CHAN_DIRREAD
#define	APP_WRITE		CHAN_DIRPUT
#define	APP_BUILDACK	CHAN_BLDACK
#define	APP_READRESP	CHAN_DRRDATA
#define	APP_BLDRDMON	CHAN_BLDRDMON
#define	APP_SINGMON		CHAN_SINGMON
#define	APP_DUMPTREND	CHAN_ALQCONN+1				/* no body to message */
#define	APP_DUMPALARM	APP_DUMPTREND+1				/* no body to message */
/*
	define rsrvTask location & name.
*/
#define	DPA_HSRV_TASK	"/tis4000/tasks/hsrvTask"
#define	DPA_HSRV_NAME	"hsrvTask"

typedef	enum
	{
	CHAN_READ,				/* response to CHAN_DIRREAD -> send CHAN_DRRDATA */
	CHAN_MONITOR,			/* response for async application writes */
	CHAN_SINGMONITOR		/* final response to CHAN_BLDRDMON */
	} CHAN_DATA_RESPONSE;
/*
	Field definitions for get/put field ops
*/
typedef	enum
	{
	DPA_VAL=0,
	DPA_EGU,
	DPA_STAT,
	DPA_SEVR,
	DPA_HOPR,
	DPA_LOPR,
	DPA_DRVH,
	DPA_DRVL,
	DPA_HIHI,
	DPA_HIGH,
	DPA_LOW,
	DPA_LOLO,
	DPA_HHSV,
	DPA_HISV,
	DPA_LOSV,
	DPA_LLSV,
	DPA_PREC,
	DPA_AENB,
	DPA_TENB,
	DPA_ACK,
	DPA_DESC,
	DPA_USER,
	NUM_FIELDS
	} FIELD_SEL;
/* #define	NUM_FIELDS	21 */
#define	DPA_USER_DEFINED_FIELD	-1

typedef	struct
{
  char	*appName;				/* pointer to application name from argv[0] */
  int		appFd;					/* FD for rsrvTask communications */
  int		appInet;				/* Inet address of hsrvTask */
  int		hsrvTaskPID;			/* just what it says! */
  short	appMaxHandles;			/* number of handles for this app */
  short	appResourceState;		/* 1 = master 0 = slave */
/*
	User supplied (via dpaInit) functions
*/
  void	(*appOpen)();
  void	(*appRead)();
  void	(*appWrite)();
  void	(*appClose)();
  /*
	Receive buffers and place holders
*/
  int		rxCount;				/* number of chars in rxBuff */
  int		rxNext;					/* index of next char */
  char	rxBuff[DPA_MAX_CHARS];	/* incomming message buffer */
  CHAN_DATA_RESPONSE	appResp;	/* type of response to hsrvTask */
} DPA_INFO,*DPA_INFO_PTR;

typedef	struct	_handles
{
  struct _handles	*nextHandle;	/* next or zero for end */
  int  handle;		       	/* actual handle from CHAN_BUILD or CHAN_BLDRDMON */
  short  requestedType;		/* data type as requested from data client */
  short  requestedCount;	/* amount of data the client wants */
} DPA_HANDLES,*DPA_HANDLES_PTR;
/*
	Field descriptor
*/
typedef	struct	_dpaddr
{
  struct	_dpaddr	*nextField;    		/* pointer to next field in chain...NYI */
  void		*precord;		       	/* pointer to full record */
  void		*pfield;		       	/* pointer to field withing record */
  struct _handles *sendHandles;			/* pointer to list of handles for this point */
  void		*chHandle;		       	/* from CHAN_BUILD & CHAN_BLDRDMON */
  void		*chCaHandle;			/* from CHAN_MONREQ */
  void		*chAppHandle;			/* from CHAN_DIRREAD CHAN_BLDRDMON */
  short		nativeCount;
  short		field_size;		       	/* size of field in bytes */
  FIELD_SEL	field_index;			/* field index DPA_VAL... */
  short		nativeType;			/* MY data type (AGNT..) */
  char		fieldName[5];			/* 4 char field name...just like regular points */
  char		init;			       	/* 1 = initialized & ok to write */
  short		numStrings;		       	/* number of following enum strings */
  char		**enums;		       	/* pointer to enum list */
  double       	lolimit;		       	/* display limits */
  double       	hilimit;
} DPA_ADDR,*DPA_ADDR_PTR;
/*
	Internal channel info table
*/
typedef	struct	_chaninfo
{
  NODE				chanNext;      	/* next NODE in list or 0 if end */
  short				chanDataType;	/* desired data type of value */
  short				chanCaType;    	/* data type returned by Channel Access */
  DPA_ADDR_PTR			chanFields;	/* array of dbAddr's indexed by field index */
  short				chanValueSize;	/* size in bytes of above data */
  char				chanName[PVNAME_SZ+FLDNAME_SZ+50];
  char				chanDesc[32];	/* descriptor */
  char				chanEgu[8];    	/* 7 with NULL terminator */
  short				chanStat;      	/* for status & severity reports */
  short				chanSevr;
  char				chanValue[DPA_MAX_VALUE_SIZE];	/* current value of type chanDataType */
  double		       	chanDispHi;    	/* display range */
  double		       	chanDispLo;
  double		       	chanDrvHi;     	/* instrument range */
  double		       	chanDrvLo;
  double		       	chanHiHi;      	/* alarm limits */
  double		       	chanHi;
  double		       	chanLo;
  double		       	chanLoLo;
  short				chanHHSever;   	/* alarm severity */
  short				chanHSever;
  short				chanLSever;
  short				chanLLSever;
  short				chanPrec;      	/* precision */
  short				chanNumStrs;	/* number of strings in chanEnumList */
  char				**chanEnumList;	/* string list for enumerations */
  short				chanMask;      	/* AGNT_MASK values */
  short				chanAenb;      	/* alarm enable */
  short				chanTenb;      	/* trend enable */
  short				chanAck;       	/* alarm ACK */
  short				valid;	       	/* 0 = invalid */
  /*
    Place to save the last alarm time. This is used during an ACK to check if
    the ACK message is received while another alarm is pending. If that is
    the case then do NOT ACK the alarm.
    */
  struct
  {
    int	secPastEpoch;
    int	nsec;
  } alarm_time;
  char 	last_value[DPA_MAX_VALUE_SIZE];	/* last alarm value of type chanDataType */
  char				chanUserData[DPA_USER_DATA_SIZE];	/* Application dependent user data */
} CHANNEL_INFO,*CHANNEL_INFO_PTR;

/*
  GCT 30-Sep-1997
  expanded message for CH_DRRDATA
*/
typedef struct
{
  CH_DRRDATA  drr;
  void  *appHandle;
} DPA_DRRDATA, DPA_DRRDATA_PTR;
#define DPA_DRRDATA_SIZE sizeof(DPA_DRRDATA)
/*
	conversion function prototypes
*/
float	dpaConvertToFloat(void *,void *,short);
double	dpaConvertToDouble(void *,void *,short);
char	*dpaConvertToEnumStr(void *,void *,short);
short	dpaConvertToInt(void *,void *,short);
char	*dpaConvertToString(void *,void *,short);
char	dpaConvertToChar(void *,void *,short);
long	dpaConvertToLong(void *,void *,short);
int    	dpaSetAlarmAttr(void *,double ,double ,double ,double ,short ,short ,short ,short,short);
int    	dpaGetAlarmAttr(void *,double *,double *,double *,double *,short *,short *,short *,short *);
int    	dpaGetAttributes(void *,short *,short *,double *,double *,double *,double *,short *,char **);
int    	dpaGetFieldAttributes(void *,short *,short *,double *,double *,double *,double *,short *,char **);
int    	dpaSetAttributes(void *,short ,short ,double ,double ,double ,double ,short ,char **);
int    	dpaWhichField(char *);
short	dpaGetField(void *,void **);
int    	dpaPutField(void *,short,void *,short);
void	dpaSplitName(char *,char *,char *);
DPA_ADDR_PTR	dpaNameToAddr(char *,CHANNEL_INFO_PTR);
void	*dpaGetHandle(char *);
int    	dpaWrite(void *,void *,short,short,short,short);
void	*dpaSetUserField(void *,char *,short,short,short,char **,double,double,void *);
void	dpaGetFieldType(void *,short *);
void	dpaTrend(void *);
void	dpaAlarm(void *,void *,short,short,short,TS_STAMP *);
#if defined(SOLARIS) || defined(AXP)
#define bzero(b,n)	memset (b, 0, n)
#define bcopy(s,d,n)	memcpy (d, s, n)
#endif

#endif	INCdpaH
