/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
	These two functions operate on the internal data table. They both affect
	the attributes fields. The dpaSetAttributes function is used to fill in those fields
	with values specified by the user. The dpaGetAttributes function reads the table
	pointed to by the handle parameter.

	GCT 19-OCT-1994

*/
#include	<in.h>

#include	"dbDefs.h"
#include	"dpa.h"
extern	DPA_INFO	dpaInfo;

int	dpaSetAttributes(void *handle,short type,short mask,double disphi,double displo,
					 double drvhi,double drvlo,short prec,char **enumlist)
{
CHANNEL_INFO_PTR	table;
int	i;

if(!handle)
	return(DPA_NULLHANDLE);
table = (CHANNEL_INFO_PTR)handle;


table->chanDataType	= type;
table->chanDispHi	= disphi;
table->chanDispLo	= displo;
table->chanDrvHi	= drvhi;
table->chanDrvLo	= drvlo;
table->chanPrec		= prec;
table->chanMask		= mask;
if(enumlist)
	{
	i = 0;
	while(enumlist[i++]);
	table->chanNumStrs = i - 1;
	table->chanEnumList = (char **)calloc(table->chanNumStrs,sizeof(char *));	/* allocate space for array */
	for(i = 0;i < table->chanNumStrs;i++)
		{
		table->chanEnumList[i] = (char *)calloc(1,strlen(enumlist[i]+1));		/* allocate space for strings */
		strcpy((char *)table->chanEnumList[i],enumlist[i]);					/* copy strings to new place */
		}
	}
return(DPA_NORMAL);
}
int	dpaGetAttributes(void *handle,short *type,short *mask,double *disphi,double *displo,
					 double *drvhi,double *drvlo,short *prec,char **enumlist)
{
CHANNEL_INFO_PTR	table;

if(!handle)
	return(DPA_NULLHANDLE);
table = (CHANNEL_INFO_PTR)handle;

*type		= table->chanDataType;
*mask		= table->chanMask;
*disphi		= table->chanDispHi;
*displo		= table->chanDispLo;
*drvhi		= table->chanDrvHi;
*drvlo		= table->chanDrvLo;
*prec		= table->chanPrec;
enumlist	= table->chanEnumList;
return(DPA_NORMAL);
}
int	dpaGetFieldAttributes(void *handle,short *type,short *mask,double *disphi,double *displo,
					 double *drvhi,double *drvlo,short *prec,char **enumlist)
{
DPA_ADDR	*pAddr;
CHANNEL_INFO_PTR	table;

if(!handle)
	return(DPA_NULLHANDLE);
pAddr = (DPA_ADDR *)handle;
table = (CHANNEL_INFO_PTR)pAddr->precord;

*type		= table->chanDataType;
*mask		= table->chanMask;
*disphi		= table->chanDispHi;
*displo		= table->chanDispLo;
*drvhi		= table->chanDrvHi;
*drvlo		= table->chanDrvLo;
*prec		= table->chanPrec;
enumlist	= table->chanEnumList;
return(DPA_NORMAL);
}
short	dpaIsValid(void *handle)
{
CHANNEL_INFO_PTR	table = (CHANNEL_INFO_PTR)handle;
return(table -> valid);
}

int	dpaGetAlarmAttr(void *handle,double *hihi,double *high,double *low,double *lolo,
					short *hihisever,short *hisever,short *losever,short *lolosever)
{
CHANNEL_INFO_PTR	table;

if(!handle)
	return(DPA_NULLHANDLE);
table = (CHANNEL_INFO_PTR)handle;

*hihi		= table->chanHiHi;
*high		= table->chanHi;
*low		= table->chanLo;
*lolo		= table->chanLoLo;
*hihisever	= table->chanHHSever;
*hisever	= table->chanHSever;
*losever	= table->chanLSever;
*lolosever	= table->chanLLSever;
return(DPA_NORMAL);
}

int dpaSetAlarmAttr(void *handle,double hihi,double high,double low,double lolo,
		    short hihisever,short hisever,short losever,short lolosever,short aenb)
{
CHANNEL_INFO_PTR	table;

if(!handle)
	return(DPA_NULLHANDLE);
table = (CHANNEL_INFO_PTR)handle;

table->chanHiHi		= hihi;
table->chanHi		= high;
table->chanLo		= low;
table->chanLoLo		= lolo;
table->chanHHSever	= hihisever;
table->chanHSever	= hisever;
table->chanLSever	= losever;
table->chanLLSever	= lolosever;
table->chanAenb		= aenb;
return(DPA_NORMAL);
}
/*
	Resource mode GET/SET functions. The SET function also informs hsrvTask any change in mode.
*/
void	dpaSetResMode(short newmode)
{
short	msgType;
CH_RESMODE	resMsg;
if(dpaInfo.appResourceState != newmode)
	{
	dpaInfo.appResourceState = newmode;							/* set new mode */
	resMsg.chResStatus = htons(newmode);						/* new mode... */
	strcpy(resMsg.chResName,dpaInfo.appName);					/* ...and resource(app) name */
	msgType = htons(CHAN_RESMODE);
	send(dpaInfo.appFd,&msgType,sizeof(msgType),0);				/* send the message type... */
	send(dpaInfo.appFd,&resMsg,sizeof(CH_RESMODE),0);			/* followed by the message header... */
	}
}

short	dpaGetResMode()
{
return(dpaInfo.appResourceState);
}
/*
  GCT 12-Nov-1996 Use this function to return an un-initialized DPA_ADDR_PTR object.
  */
void	*dpaGetUserFieldPtr(void *handle,char *field)
{
CHANNEL_INFO_PTR	entry;
DPA_ADDR	*pAddr;
char		padField[5];

if(!handle)
	return((void *)0);
sprintf(padField,"%-4.4s",field);
entry = (CHANNEL_INFO_PTR)handle;
if((pAddr = (DPA_ADDR_PTR)findField(entry->chanFields,padField)) == 0)
  return((void *)0);						/* error no field defined yet */
else
  return((void *)pAddr);
}
/*
	Set attributes of user-defined field. A user-defined field is one that does not
	appear in the static field set as defined in dpa.h
*/
void	*dpaSetUserField(void *handle,char *field,short size,short type,short count,char **enumlist,double lo,double hi,void *data)
{
CHANNEL_INFO_PTR	entry;
DPA_ADDR	*pAddr;
char		padField[5];
short		i;
int numstrs;
if(!handle)
	return((void *)0);
sprintf(padField,"%-4.4s",field);
entry = (CHANNEL_INFO_PTR)handle;
if(count <= 0) count = 1;
if((pAddr = (DPA_ADDR_PTR)findField(entry->chanFields,padField)) == 0)
	return((void *)0);						/* error no field defined yet */
else
	{
	pAddr->field_size      	= size;
	pAddr->nativeType	= type;
	pAddr->nativeCount     	= count;
	pAddr->lolimit	       	= lo;
	pAddr->hilimit	       	= hi;
	/*
	  GCT 15-Mar-96 Add ability for user to specify data area.
	*/
	if(!pAddr->pfield)
	  {
	  if(data == 0)
	    {
	    pAddr->pfield		= (void *)calloc(count,size);	/* get space for 'count' number of 'size' bytes of data */
	    }
	  else
	    {
	    pAddr->pfield = data;	/* use user supplied buffer */
	    }
	  }
	pAddr->field_index		= DPA_USER_DEFINED_FIELD;		/* mark as user-defined field */
	if(enumlist == 0)
	  {
	    enumlist = entry->chanEnumList;	/* if param is zero try the ones from the config file(could also be zero) */
	    numstrs = entry->chanNumStrs;
	  }
	else
	  {
	    i = 0;
	    while(enumlist[i++]);
	    numstrs = i - 1;
	  }
	if(enumlist && !pAddr->enums)
		{
		pAddr->numStrings = numstrs;
		pAddr->enums = (char **)calloc(pAddr->numStrings,sizeof(char *));	/* allocate space for array */
		for(i = 0;i < pAddr->numStrings;i++)
			{
			pAddr->enums[i] = (char *)calloc(1,strlen(enumlist[i])+1);		/* allocate space for strings */
			strcpy((char *)pAddr->enums[i],enumlist[i]);					/* copy strings to new place */
			}
		}
	pAddr->init = 1;							/* OK to write */
	}
return((void *)pAddr);							/* success */
}
void	dpaGetFieldType(void *handle,short *type)
{
DPA_ADDR	*pAddr;
pAddr = (DPA_ADDR_PTR)handle;
*type = pAddr->nativeType;
}
/*
  return a pointer to the user defined data from configuration. If "buff"
  is non-zero then copy user data into buff.
*/
char *dpaGetUserData(void *handle,char *buff,int length)
{
  CHANNEL_INFO_PTR	table;

  if(!handle)
    return(0);
  table = (CHANNEL_INFO_PTR)handle;
  if(length > DPA_USER_DATA_SIZE)
    length = DPA_USER_DATA_SIZE;
  if(buff)
    memcpy(buff,&table->chanUserData[0],length);
  return(&table->chanUserData[0]);
}
/*
  Copy 'length' bytes from user supplied area to chanUserData area.
  'length' will be clipped to DPA_USER_DATA_SIZE (256) if it is
  longer.
*/
void  dpaSetUserData(void *handle,char *buff, int length)
{
  CHANNEL_INFO_PTR	table;

  if(!handle)
    return;
  table = (CHANNEL_INFO_PTR)handle;
  if(length > DPA_USER_DATA_SIZE)
    length = DPA_USER_DATA_SIZE;
  memcpy(&table->chanUserData[0],buff,length);
  return;
}
