#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#ifdef	WIN32
#    include <direct.h>
#    include "NTUnixSubs.h"
     extern char *optarg;
#endif WIN32

#include "index.h"
#include "msgMon.h"

/* Functions defined in this module */

static	void	BuildTable (void);
static	void	DumpTable (void);
static	long	GetDataBase (void);
static	char	*MakeString (char *, int);
static	int	BlankField (char *);
static	void	FreePointList (void);

/* Codebase stuff */

static	C4CODE	Settings;
	
#define	POINTNAMELEN	30
typedef	char	POINTNAME [POINTNAMELEN];

typedef struct rec {
    POINTNAME	PointName;
    long	Next;
    int		Refs;
} REC;

/* Database path */

static	char	DbPath [100];

/* Hashed index of all points */

static	INDEX	Index;

/* Output file */

static	FILE	*OutFile;

/* Linked list of input points */

typedef struct plistnode {
    char	*PointName;
    char	*FwdLink;
    struct plistnode *Next;
} PointListNode;
static	PointListNode *PointList;

/* Last column used on output */

#define	LAST_COL	70

extern	char	*optarg;

/*----------*/
void main (int argc, char **argv)

    {
	int	c;
	char	DbName [100];
	long	NumRecs;

	tisOpenLog ("chains", 6, 2, NULL, NULL);

	/* Set default database directory */

#ifdef	WIN32
	strcpy (DbName, "\\tis4000db");
#else
	strcpy (DbName, "/tis4000db");
#endif	WIN32
	/* Set default output file */

	OutFile = stdout;

	/* Parse the command line arguments */

	while ((c = getopt (argc, argv, "p:o:")) != -1) {
	    switch (c) {
		case 'p':	/* Database directory name */
		    strcpy (DbName, optarg);
		    break;
		case 'o':	/* Output file name */
		    OutFile = fopen (optarg, "w");
		    if (OutFile == NULL) {
			tisSysLog (0,0, "Unable to create %s", optarg);
			exit (1);
		    }
		    break;
	    }
	}

#ifdef	WIN32
	if (_chdir (DbName) != 0)
#else
	if (chdir (DbName) != 0)
#endif	WIN32
	{
	    tisSysLog (0,0, "Unable to find directory %s", DbName);
	    exit (1);
	}

#ifdef	WIN32
	_getcwd (DbPath, sizeof (DbPath));
#else
	getcwd (DbPath, sizeof(DbPath));
#endif	WIN32

	/* Prepare CodeBase for use */

	d4init (&Settings);
#if defined(WIN32) | defined(USECB63)
	Settings.errOpen = 0;
#else
	Settings.open_error = 0;
#endif	WIN32 or USECB63

	/* Open the database and read in all the records that have a */
	/* forward link. */

	NumRecs = GetDataBase ();
	if (NumRecs == -1) {
	    tisSysLog (0,0, "Unable to open PVMAP in directory %s\n", DbPath);
	    exit (1);
	}

	/* Create and open an index */

	if (!IndexCreate ("index.tmp", sizeof (REC), sizeof (POINTNAME), NumRecs * 2)) {
	    tisSysLog (0,0, "Unable to create temporary hash file.\n");
	    exit (-1);
	}

	if (!IndexOpen ("index.tmp", &Index)) {
	    tisSysLog (0,0, "Unable to create hash table\n");
	    exit (-1);
	}

	/* Read the input and build the table */

	BuildTable ();

	/* Dump the table to check it */

	DumpTable ();

	/* Free the point list */

	FreePointList ();

    }

/* end main */

/*----------*/
static long GetDataBase (void)	/* Read in all database records that */
				/* have a forward reference */
    {

	char	DbName [100];
	D4DATA	*DBase;
	F4FIELD	*PointNameField, *FwdLinkField;
	long	i, NumPoints;
	PointListNode	*p;
	char	Name [30], Link [30];

	/* Initialize list to empty */

	PointList = NULL;

	/* Open the database and fields */

	sprintf (DbName, "%s/pvmap", DbPath);
	DBase = d4open (&Settings, DbName);
	if (DBase == 0)
	    return -1;

	PointNameField = d4field (DBase, "NAME");
	FwdLinkField = d4field (DBase, "FLNK");
 
	/* Go through database and extract points */

	NumPoints = 0;
	for (i=1 ; i<=d4reccount (DBase) ; i++) {

	    /* Retrieve the next record */

	    d4go (DBase, i);

	    /* Check if this point has a forward link */

	    strcpy (Name, f4str (PointNameField));
	    strcpy (Link, f4str (FwdLinkField));

	    if (!BlankField (f4str (FwdLinkField))) {

		/* Allocate a new node and put stuff into it */

		p = (PointListNode *) malloc (sizeof (PointListNode));
		p->PointName = MakeString (f4str (PointNameField), 1);
		p->FwdLink = MakeString (f4str (FwdLinkField), 1);
		p->Next = PointList;
		PointList = p;

		NumPoints++;

	    }

	}

	d4close (DBase);

	return NumPoints;
 
    }

/* end GetDataBase */

/*----------*/
static void BuildTable (void)	/* Read the input and build the table */

    {

	long	i, j;
	REC	Rec, FwdRec;
	PointListNode	*p;

	for (p=PointList ; p ; p = p->Next) {

	    /* Lookup the forward link node.  If it exists, increment its */
	    /* reference count.  If it doesn't exist, insert it with a */
	    /* reference count of 1. */

	    if (strlen (p->FwdLink)) {
		strcpy (FwdRec.PointName, p->FwdLink);
		if (IndexLookUp (&Index, &FwdRec, &j) == 0) {	/* Not there */
		    FwdRec.Next = -1;
		    FwdRec.Refs = 1;
		    IndexInsert (&Index, &FwdRec, &j);
		} else {
		    FwdRec.Refs++;
		    IndexWriteDirect (&Index, &FwdRec, j);
		}
	    } else {
		j = -1;
	    }

	    /* Put the location of the forward referenced point in the */
	    /* current record. Place point into the table. */

	    strcpy (Rec.PointName, p->PointName);
	    if (IndexLookUp (&Index, &Rec, &i)) {
		Rec.Next = j;
		IndexInsert (&Index, &Rec, &i);
	    } else {
	        Rec.Next = j;
	        Rec.Refs = 0;
	        IndexInsert (&Index, &Rec, &i);
	    }
	}
     }

/* end BuildTable */

/*----------*/
static void DumpTable (void)

{

	long	First;
	REC	Rec;
	int	Column;

	IndexRewind (&Index);
	while ((First = IndexReadNext (&Index, &Rec)) != -1) {
	    if (Rec.Next != -1) {
		if (Rec.Refs == 0) {
		    fprintf (OutFile, "*");
		    Column = 1;
		    for (;;) {
			if (Column + strlen (Rec.PointName) + 1 > LAST_COL) {
			    Column = 1;
			    fprintf (OutFile, "\n ");
			}
			fprintf (OutFile, "%s ", Rec.PointName);
			Column += strlen (Rec.PointName) + 1;
			if (Rec.Next == -1) break;
			IndexReadDirect (&Index, &Rec, Rec.Next);
		    }
		    fprintf (OutFile, "\n");
		}
	    }
	}

/*	IndexRewind (&Index);
	while ((i = IndexReadNext (&Index, &Rec)) != -1)
	    printf ("%4ld: %-5s %4ld %3d\n", i, Rec.PointName, Rec.Next, Rec.Refs);
*/

}

/*----------*/
static char *MakeString (char *Str, int Strip)	/* Build a string from the heap */

    {

	char	*p, Buf [100];
	int	i;

	/* If no string, return NULL */

	if (Str == NULL)
	    return NULL;

	/* Copy the string to a local buffer */

	strcpy (Buf, Str);

	/* Optionally strip off the trailing blanks */

	if (Strip) {
	    i = strlen (Buf) - 1;
	    while (i >= 0)
		if (isspace (Buf [i]))
		    i--;
		else
		    break;
	    Buf [i+1] = '\0';
	}

	/* Allocate memory for the string and copy it */

	p = malloc (strlen (Buf) + 1);
	strcpy (p, Buf);

	return p;

    }

/* end MakeString */

/*----------*/
static int BlankField (char *s)

    {

	while (*s) {
	    if (!isspace (*s))
		return 0;
	    s++;
	}

	return 1;
    }

/* end BlankField */

/*----------*/
static void FreePointList (void)

    {

	PointListNode	*p;

	while (PointList) {
	    p = PointList;
	    PointList = PointList->Next;
	    free (p->PointName);
	    free (p->FwdLink);
	    free (p);
	}

    }

/* end FreePointList */

	    
