#define	MAXBUFSIZE	50000
/****************************************************************/
/*   Copyright (c) 1988,1989,1990 by SpeleoTechnologies, Inc.	*/
/*			All Rights Reserved			*/
/****************************************************************/
/*
/*	file:	hashindex.c	/* Hashed index file utilities */
/*	author:	Douglas P. Dotson
/*	date:	15 Mar 88
/*	edit:	29 Oct 92	2144	DPD
*/
#define LINT_ARGS

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>

#include "index.h"
#include "msgMon.h"

/* Functions in this module */

	int 	IndexCreate ();
	int	IndexOpen ();
	void	IndexClose ();
	void	IndexDelete ();
	long	IndexSize ();
	long	IndexNumEntries ();
	int	IndexInsert ();
	int	IndexLookUp ();
	void	IndexReadDirect ();
	void	IndexWriteDirect ();
	void	IndexRewind ();
	long	IndexReadNext ();
static	void	GetRecord ();
static	void	PutRecord ();
static	long	Hash ();
static	long	ReHash ();

typedef	struct {
    long	IndexSize;
    int		KeySize;
    int		RecSize;
    long	NumRecs;
} INDEXHEADER;
static	INDEXHEADER	Header;

/* Table of twin primes */

static long Primes [] = {
 	313, 433, 523, 619, 811, 911, 1033, 1153,
	1231, 1303, 1429, 1609, 1723, 1873, 1933,
	2029, 3121, 4021, 5011, 6091, 7129, 8089, 9013, 10009,
	11059, 12043, 13009, 14011, 15139, 16063, 17029, 18043,
	19081, 20023, 21013, 22039, 23029, 24109, 25033, 26113,
	27109, 28099, 29131, 30013, 31081, 32029, 33073, 34033,
	35083, 36013, 37021, 38239, 39163, 40013, 41143, 42019,
	43201, 44029, 45121, 46051, 47389, 48313, 49123, 50023,
	51133, 52183, 53233, 54013, 55051, 56041, 57223, 58111,
	59011, 60103, 61223, 62131, 63031, 64153, 65173, 66109,
	67189, 68113, 69193, 70123, 71263, 72091, 73039, 74203,
	75013, 76003, 77239, 78139, 79153, 80149, 81019, 82009,
	83233, 84061, 85093, 86029, 87013, 88003, 89563, 90019,
	91153, 92179, 93133, 94009, 95089, 96223, 97003, 98323,
	99133, 99991, 0
};

/*----------*/
/* Create hash file */
int IndexCreate (char *PathName, int RecSize, int KeySize, long Entries)
    {

	long	ix, BytesLeft, BytesToWrite;
	size_t	BufSize;
	FILE	*Fil;
	char	*Buf;

	/* Determine hash file size.  The size of the hash file is the */
	/* next prime number greater than 50% larger than the maximum */
	/* number of records expected.  The primes in the table above are */
	/* larger of a pair of twin primes to obtain better speading from */
	/* the Hash and ReHash functions */

	Entries = (long)(Entries * 1.5);
	ix = 0;
	while (Primes [ix]) {
	    if (Primes [ix] >= Entries) break;
	    ix++;
	}
	Header.IndexSize = Primes [ix];

	/* Set up file header */

	Header.RecSize = RecSize;
	Header.KeySize = KeySize;
	Header.NumRecs = 0L;

	/* Create and initialize the Hash file to all empty */

	if ((Fil = fopen (PathName, "w+b")) == NULL)
	    return (0);

	if (fwrite ((char *)&Header, sizeof (Header), 1, Fil) != 1) {
	    fclose (Fil);
	    return (0);
	}

	/* To fill the file with empty records faster, allocate a big */
	/* buffer, fill with -1 and write it as many times as necessary */

	BufSize = MAXBUFSIZE;
	while ((Buf = malloc (BufSize)) == NULL)
	    BufSize /= 2;

	memset (Buf, '\377', BufSize);
	BytesLeft = Header.IndexSize * Header.RecSize;
	while (BytesLeft > 0L) {
	    BytesToWrite = (BytesLeft >= BufSize) ? BufSize : BytesLeft;
	    if (fwrite (Buf, (size_t)BytesToWrite, 1, Fil) == 0) {
		fclose (Fil);
		unlink (PathName);
		free (Buf);
		return (0);
	    }
	    BytesLeft -= MAXBUFSIZE;
	}
	free (Buf);
	fclose (Fil);

	return (1);

    }

/* end IndexCreate */
 
/*----------*/
int IndexOpen (char *IndexPath, INDEX *IndexDesc) /* Open an index file */

    {

	FILE	*Fil;
 
	/* Open the index file */

	Fil = fopen (IndexPath, "r+");
	if (Fil == NULL)
	    return 0;

	/* Read the file header */

	if (fread ((char *)&Header, sizeof (Header), 1, Fil) != 1) {
	    fclose (Fil);
	    return 0;
	}

	/* Place necessary items in the file descriptor */

	IndexDesc->IndexFile = Fil;
	IndexDesc->IndexSize = Header.IndexSize;
	IndexDesc->KeySize = Header.KeySize;
	IndexDesc->RecSize = Header.RecSize;
	IndexDesc->NumRecs = Header.NumRecs;

	/* Allocate buffer for index */

	IndexDesc->Buffer = malloc (Header.RecSize);
	if (IndexDesc->Buffer == NULL) {
	    fclose (Fil);
	    return 0;
	}

	/* Allocate cache */

	IndexDesc->Cache = malloc (Header.RecSize);
	IndexDesc->CachePlace = -1L;
	if (IndexDesc->Cache == NULL) {
	    fclose (Fil);
	    return 0;
	}

	return 1;	
    }

/* end IndexOpen */

/*----------*/
void IndexClose (INDEX *IndexDesc)	/* Close an index file */

    {

	/* Write information back to the header */

	Header.IndexSize = IndexDesc->IndexSize;
	Header.KeySize = IndexDesc->KeySize;
	Header.RecSize = IndexDesc->RecSize;
	Header.NumRecs = IndexDesc->NumRecs;
	fseek (IndexDesc->IndexFile, 0L, 0); 
	fwrite ((char *)&Header, sizeof (Header), 1, IndexDesc->IndexFile);

	/* Release the simple cache and the record buffer */

	free (IndexDesc->Buffer);
	free (IndexDesc->Cache);

	fclose (IndexDesc->IndexFile);

    }

/* end IndexClose */

/*----------*/
void IndexDelete (char *IndexPath)	/* Close and remove the hash file */

    {

	unlink (IndexPath);

    }

/* end IndexDelete */

/*----------*/
long IndexSize (INDEX *IndexDesc)	/* Get size of index file */

    {

	return IndexDesc->IndexSize;

    }

/* end IndexSize */
 

/*----------*/
long IndexNumEntries (INDEX *IndexDesc)  /* Get number of records in the index */

    {

	return IndexDesc->NumRecs;

    }

/* end IndexNumEntries */

/*----------*/
int IndexInsert (INDEX *IndexDesc, char *Buf, long *Place)	/* Write Buf to IndexFile */

    {

	long	i, j;

	/* Calculate Hash and ReHash values */

	i = Hash (Buf, IndexDesc);
	j = ReHash (i);

	/* If key is already in file then return record containing key. */   
	/* If key not in file, insert new record into the file */

	for (;;) {
	    if (i > IndexDesc->IndexSize) {
		tisSysLog (0,0,"Hashing error: index = %ld\n", i);
		exit (-1);
	    }
	    GetRecord (IndexDesc, i);
	    if (*(IndexDesc->Buffer) == -1) break;
	    if (strcmp (Buf, IndexDesc->Buffer) == 0) {
		memcpy (Buf, IndexDesc->Buffer, IndexDesc->RecSize);
		memcpy (IndexDesc->Cache, Buf, IndexDesc->RecSize);
		IndexDesc->CachePlace = i;
		*Place = i;
		return 1;
	    }
	    i = (i + j) % IndexDesc->IndexSize;
	}

	/* Insert record into file */

	PutRecord (IndexDesc, i, Buf);

	/* Place record in the cache */

	memcpy (IndexDesc->Cache, Buf, IndexDesc->RecSize);
	IndexDesc->CachePlace = i;

	*Place = i;
	IndexDesc->NumRecs++;
	return 1;

    }

/* end IndexInsert */
		
/*----------*/
/* Find a record in the Hash file */
int IndexLookUp (INDEX *IndexDesc, char *Buf, long *Place)

    {

	long	i, j;

	/* Check if requested record is in the cache */

	if (strcmp (Buf, IndexDesc->Cache) == 0) {
	    memcpy (Buf, IndexDesc->Cache, IndexDesc->RecSize);
	    *Place = IndexDesc->CachePlace;
	    return (1);
	}

	/* Calculate Hash and Rehash values */

	i = Hash (Buf, IndexDesc);
	j = ReHash (i);

	/* Find an entry in the Hash file */

	for (;;) {
	    GetRecord (IndexDesc, i);
	    if (*(IndexDesc->Buffer) == -1) return 0;
	    if (strcmp (Buf, IndexDesc->Buffer) == 0) {
		memcpy (Buf, IndexDesc->Buffer, IndexDesc->RecSize);
		*Place = i;
		memcpy (IndexDesc->Cache, IndexDesc->Buffer, IndexDesc->RecSize);
		IndexDesc->CachePlace = i;
		return 1;
	    }
	    i = (i + j) % IndexDesc->IndexSize;
	}

    }

/* end IndexLookUp */

/*----------*/
static void GetRecord (INDEX *IndexDesc, long RecNum)	/* Read a record from the hash file */

    {

	fseek (IndexDesc->IndexFile,
		sizeof (INDEXHEADER) + RecNum * IndexDesc->RecSize, 0);
	fread (IndexDesc->Buffer, IndexDesc->RecSize, 1, IndexDesc->IndexFile);

    }

/* end GetRecord */

/*----------*/
/* Write a record from the hash file */
static void PutRecord (INDEX *IndexDesc, long RecNum, char *Buf)

    {

	fseek (IndexDesc->IndexFile,
		sizeof (INDEXHEADER) + RecNum * IndexDesc->RecSize, 0);
	fwrite (Buf, IndexDesc->RecSize, 1, IndexDesc->IndexFile);

    }

/* end PutRecord */

/*----------*/
static long Hash (char *Key, INDEX *IndexDesc)	/* Calculate Hash Index */

    {

	unsigned long	Accum;
	unsigned char	Buf [50], *p;
	int	Size;

	/* Pad key out with spaces */

	Size = IndexDesc->KeySize;
	memset (Buf, ' ', Size);
	memcpy (Buf, Key, strlen (Key));
	Buf [Size] = '\0';
	
	p = Buf;
	Accum = 0;
	while (*p) {
	    Accum = ((Accum << 8) + *p) % IndexDesc->IndexSize;
	    p++;
	}

	return Accum;

    }

/* end Hash */

/*----------*/
static long ReHash (long Key)	/* Calculate rehash index */

    {

	return (long)(256 - (Key & 0xFF));

    }

/* end ReHash */

/*----------*/
void IndexReadDirect (INDEX *IndexDesc, char *Buf, long RecNum)

   {

	/* Check if requested record is in the cache */

	if (RecNum == IndexDesc->CachePlace) {
	    memcpy (Buf, IndexDesc->Cache, IndexDesc->RecSize);
	    return;
	}

	GetRecord (IndexDesc, RecNum);
	memcpy (Buf, IndexDesc->Buffer, IndexDesc->RecSize);

    }

/* end IndexReadDirect */

/*----------*/
void IndexWriteDirect (INDEX *IndexDesc, char *Buf, long RecNum)

   {

	PutRecord (IndexDesc, RecNum, Buf);

	/* Place record just written in the cache */

	memcpy (IndexDesc->Cache, Buf, IndexDesc->RecSize);
	IndexDesc->CachePlace = RecNum;

    }

/* end IndexWriteDirect */

/*----------*/
void IndexRewind (INDEX *IndexDesc)	/* Position at beginning of index file */

    {

	IndexDesc->CurrentRecord = -1L;

    }

/* end IndexRewind */

/*----------*/
long IndexReadNext (INDEX *IndexDesc, char *Buf)	/* Read next non-empty entry */

    {

	do {
	    IndexDesc->CurrentRecord++;
	    if ((IndexDesc->CurrentRecord+1) >= IndexDesc->IndexSize)
		return (-1L);
	    GetRecord (IndexDesc, IndexDesc->CurrentRecord);
	} while (*(IndexDesc->Buffer) == -1);
	memcpy (Buf, IndexDesc->Buffer, IndexDesc->RecSize);
	return (IndexDesc->CurrentRecord);

    }

/* end IndexReadNext */

