/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <hsmain.h>

static char *hostMsgs[] =
{
  "an INVALID",
  "a File Information",
  "a File Request",
  "a Host Wakeup call",
  "a Host Boomerang",
  "an INVALID",
  "an INVALID",
  "an INVALID"
};

/* #define EWOULDBLOCK_NOT_ERROR 1 */

int hsReadChars(HOST_NODE_PTR node,char *buff,int nbytes) /* read TCP buffer for given host */
{
memcpy(buff,&node->hostBuff[node->hostNextChar],nbytes);
node->hostNextChar += nbytes;
node->hostCharCount -= nbytes;
return (nbytes);
}
int hsWriteChars(HOST_NODE_PTR info ,char *buff,int nbytes) /* send chars to host */
{
int		numSent;
#ifdef	EWOULDBLOCK_NOT_ERROR
int		buffSize;
int		intSize = sizeof(int);
#endif	EWOULDBLOCK_NOT_ERROR
if(info->hostFD == -1)
  return(1);					/* not really an error */
numSent = send(info->hostFD,buff,nbytes,0);
info->hostTxTime = time(0);
lastHost = info;					/* last host written to */
if (numSent == -1)
  {
#ifdef EWOULDBLOCK_NOT_ERROR
    if (errno == EWOULDBLOCK)
      {
	getsockopt(info->hostFD, SOL_SOCKET, SO_SNDBUF, (char *)&buffSize, &intSize);
	buffSize += nbytes;
	if (setsockopt(info->hostFD, SOL_SOCKET, SO_SNDBUF, (char *)&buffSize, sizeof(buffSize)) == -1)
	  {
	    VERBOSE(V_DEBUG)(0,0,"Setsockopt error in 'hsWriteChars' %m");
	    return(hsTcpError(info));
	  }
	VERBOSE(V_DEBUG)(0,0,"Socket %d tx buffer bumped by %d bytes to %d\n",info->hostFD,nbytes,buffSize);
      }
    else
      {
	VERBOSE(V_DEBUG)(0,0,"Send error in 'hsWriteChars' %m");
	return(hsTcpError(info));
      }
    return (hsWriteChars(info, buff, nbytes));
#else
    return(hsTcpError(info));
#endif
  }
else if (numSent != nbytes)
  {
    buff += numSent;
    nbytes -= numSent;
    info->hostTxTime = time(0);
    return (hsWriteChars(info, buff, nbytes));
  }
return(1);
}
int hsHaveCompleteMsg(HOST_NODE_PTR node)  	/* check for complete messages */
{
short	mType;
int    	mSize,temp;
INFO_MSG info;
REC_MSG  rec;

if(node->hostCharCount < sizeof(mType))
  return(0);
memcpy(&mType,&node->hostBuff[node->hostNextChar],sizeof(mType));
mSize = HOST_BUFF_SIZE+1;
/*
	switch on the message type. All messages are considered complete if the number of bytes
	in the buffer (hostCharCount) is greater that or equal to the total message size PLUS
	the size of the message type.
*/
switch(ntohs(mType))
  {
  case HS_MSG_BOOMERANG:			/* This message type does not have a body */
    mSize = sizeof(short);
    break;
  case HS_MSG_INFO:
    if(node->hostCharCount > (sizeof(INFO_MSG) + sizeof(mType)))
      {
	memcpy(&info,(INFO_MSG *)&node->hostBuff[node->hostNextChar+2],sizeof(INFO_MSG));
	temp = ntohl(info.infoNameSize);
	mSize = sizeof(INFO_MSG) + (temp+1) + sizeof(mType);
      }
    break;
  case HS_MSG_REQUEST:
    if(node->hostCharCount > (sizeof(REC_MSG) + sizeof(mType)))
      {
	memcpy(&rec,(REC_MSG *)&node->hostBuff[node->hostNextChar+2],sizeof(REC_MSG));
	temp = rec.recNameSize;
	mSize = sizeof(REC_MSG) + ntohl(temp) + sizeof(mType);
      }
    break;
  }
if(mSize > node->hostCharCount)
  return(0);
else
  return(1);
}

int hsReadFD(HOST_NODE_PTR node)		/* drain FD for this host */
{
int		readSize,stat;

if(node->hostFD == -1)
  return(-1);
errno = 0;
/*
	If we are already some bytes into the buffer then slide them down to the
	begining and zero the next char pointer.
*/
if(node->hostNextChar)
  {
    if(node->hostCharCount)
      {
	memcpy(&node->hostBuff[0],&node->hostBuff[node->hostNextChar],node->hostCharCount);
      }
    node->hostNextChar = 0;
  }
readSize = HOST_BUFF_SIZE - node->hostCharCount;
stat = recv(node->hostFD, &node->hostBuff[node->hostCharCount], readSize, 0);
if(stat < 0)
  {
    VERBOSE(V_DEBUG)(0,0,"Read error in 'hsReadFD' %m");
    return(hsTcpError(node));
  }
if(stat == 0)			/* zero length read */
  {
    VERBOSE(V_DEBUG)(0,0,"Zero length read from %s on %d",node->hostName,node->hostFD);
    return(hsTcpError(node));
  }
node->hostCharCount += stat;			/* number actually read into buffer */
node->hostRxTime = time(0);		/* time of this recieve */
return(stat);
}

void hsHandleConnect(int addr,int newfd)		/* other host has connected */
{
int		index;
struct hostent	*ent;
char		hname[100];
HOST_NODE	*node;

index = hsyncFindHost(addr);
node = &hostParent[index];
if((ent = gethostbyaddr((char *)&addr,sizeof(addr),AF_INET)) != 0)
  {
    strcpy(hname,ent->h_name);
    strtok(hname,".");
    strcpy(node->hostName,hname);			/* strip off other stuff from name e.g. "tis4000.com" */
  }
hsDoAccept(addr,node,newfd);
}
void hsDoAccept(int addr,HOST_NODE *node,int newfd)
{
  char trueFlag = 1;

  if(node->hostFD != -1)
    {
      close(node->hostFD);			/* close this socket...might get a new one */
      FD_CLR(node->hostFD,&fdInfo.myFds);
      node->hostFD	= -1;			/* clear socket */
      node->hostState	= HOST_NONET;
    }
  node->hostFD = newfd;				/* this is the socket to use */
  node->hostAddr = addr;
  setsockopt(node->hostFD,IPPROTO_TCP,TCP_NODELAY,(char *)&trueFlag,sizeof(trueFlag));
  maxSocket(node->hostFD);			/* maximize this sockets Rx/Tx buffers GCT 25-Mar-97 */
  hsSetNoBlock(node);
  node->hostState = HOST_OK;		       /* connected now! */
  node->hostExchangeTime = time(0);
  node->hostRxTime = time(0);			/* time of this recieve */
  node->hostCharCount = 0;			/* nothing in buffer yet */
  node->hostNextChar = 0;			/* start of buffer */
  node->hostSendIndex = 0;
  node->hostPass = HOST_PASS1;			/* start with pass one */
  node->hostSendStatus = HS_SEND_INFO;		/* start out by sending info */
  if(node->hostFD > fdInfo.maxFd) fdInfo.maxFd = node->hostFD;
  FD_SET(node->hostFD,&fdInfo.myFds);
  VERBOSE(V_DEBUG)(0,0,"Accepted connection to %s on socket %d",node->hostName,node->hostFD);
  numConnected++;
  if(numConnected > maxConnected) maxConnected = numConnected;
}


int hsTcpError(HOST_NODE_PTR node)				/* in case of error... */
{
if(node->hostFD != -1)
  {
    VERBOSE(V_DEBUG)(0,0,"Closing socket <%d> to HOST <%s>...%m\n",
						node->hostFD,node->hostName);
    close(node->hostFD);					/* close this socket...might get a new one */
    FD_CLR(node->hostFD,&fdInfo.myFds);				/* clear here too */
    node->hostFD	= -1;					/* clear socket */
    node->hostState	= HOST_NONET;
    node->hostCharCount	= 0;
    node->hostNextChar	= 0;
    numConnected--;
  }
return(-1);
}
void hsIncomingMsg(AVL_TREE *ftree,HOST_NODE *node)	       	/* check for incoming messages from hosts */
{
int status,size,sz;
short msgType,type;
FILE_NODE fn,*pfn;
time_t remoteTime;
char fname[MAXPATHLEN],fname2[MAXPATHLEN],rcpCmd[MAXPATHLEN],*dbfPtr;
struct stat stBuff;

if((node->hostFD != -1) && FD_ISSET(node->hostFD,&fdInfo.allFds))   	/* was it for me ? */
  {
    status = hsReadFD(node);					/* drain the FD */
    if(status > 0)
      {
	while(hsHaveCompleteMsg(node))
	  {					               	/* if complete message is in buffer */
	    hsReadChars(node,(char *)&msgType,sizeof(msgType)); /* get the message type */
	    type = ntohs(msgType);
	    VERBOSE(V_STATUS)(0,0,"Got %s message from %s\n",hostMsgs[type&7],node->hostName);
	    switch(type)
	      {
	      case HS_MSG_BOOMERANG:
		break;
	      case HS_MSG_INFO:
		hsReadChars(node,(char *)&remoteTime,sizeof(time_t));/* get the remote time */
		remoteTime = ntohl(remoteTime);
		hsReadChars(node,(char *)&size,sizeof(int));	/* get file name size */
		size = ntohl(size);
		hsReadChars(node,(char *)&fname[0],size);      	/* then file name */
		strcpy(fn.fileName,fname);
		if(((pfn = avl_findsym(ftree,&fn)) != 0) && (pfn->fileStatus == FILE_VALID))
		  {
		    if(stat(pfn->fileName,&stBuff) != -1)	/* check local file time */
		      {
			if(remoteTime > pfn->fileLocalTime)	/* if remote time is more recent then request file */
			  {
			    pfn->fileLocalTime = remoteTime;
			    VERBOSE(V_STATUS)(0,0,"Requesting file %s from %s",
								pfn->fileName,node->hostName);
			    msgType = htons(HS_MSG_REQUEST);
			    hsWriteChars(node,(char *)&msgType,sizeof(msgType)); /* send mesage type */
			    size = strlen(pfn->fileName)+1;			/* include NULL with file name */
			    if(size & 1) size++;			       	/* make sure size is even */
			    sz = size;
			    size = htonl(size);
			    hsWriteChars(node,(char *)&size,sizeof(size));    	/* and size of file name */
			    hsWriteChars(node,pfn->fileName,sz);	       	/* send file name */
			  }
			else if(remoteTime < pfn->fileLocalTime)	/* local time is more recent then send file */
			  {
		    /*
		      If the file name has a ".dbf" extension then make an ".mdx" file
		      name and request that file as well. This is done to avoid a mismatch
		      between the dbf and mdx files which must be together..
		      11-Oct-96 CAREFULL -- some dbase files do not have MDX partners
		      16-Sep-1997 send MDX file first followed by the DBF. This will ensure
		      that the DBF file will have a more recent date for those programs
		      which are semsitive to that sort of thing.
		    */
			    if((dbfPtr = strstr(fname,HSYNC_DBF_FILE_EXT)) != 0)
			      {
				strcpy(dbfPtr,HSYNC_MDX_FILE_EXT);
				strcpy(fname2,pfn->fileName);
				dbfPtr = strstr(fname2,HSYNC_DBF_FILE_EXT);
				strcpy(dbfPtr,HSYNC_MDX_FILE_EXT);
				if((status = stat(fname2,&stBuff)) != -1)
				  {
				    sprintf(rcpCmd,"rcp -p %s %s:%s &",fname2,node->hostName,fname2);
				    VERBOSE(V_DEBUG)(0,0,"RCP command %s\n",rcpCmd);
				    if(system(rcpCmd) < 0)
				      VERBOSE(V_DEBUG)(0,0,"Error from SYSTEM/RCP: %m");
				  }
			      }
			    VERBOSE(V_STATUS)(0,0,"Sending file %s to %s",
							pfn->fileName,node->hostName);
			    sprintf(rcpCmd,"rcp -p %s %s:%s &",pfn->fileName,node->hostName,fname);
			    VERBOSE(V_DEBUG)(0,0,"RCP command %s\n",rcpCmd);
			    if(system(rcpCmd) < 0)
			      VERBOSE(V_DEBUG)(0,0,"Error from SYSTEM/RCP: %m");
			  }
		      }
		    else
		      {
			if(errno == ENOENT)			/* file not found...then request it */
			  {
			    msgType = HS_MSG_REQUEST;
			    msgType = htons(msgType);
			    hsWriteChars(node,(char *)&msgType,sizeof(msgType)); /* send mesage type */
			    size = strlen(pfn->fileName)+1;			/* include NULL with file name */
			    if(size & 1) size++;			       	/* make sure size is even */
			    sz = size;
			    size = htonl(size);
			    hsWriteChars(node,(char *)&size,sizeof(size));    	/* and size of file name */
			    hsWriteChars(node,pfn->fileName,sz);	       	/* send file name */
			  }
		      }
		  }
		break;
	      case HS_MSG_REQUEST:
		hsReadChars(node,(char *)&size,sizeof(size));	/* get file name size */
		size = ntohl(size);
		hsReadChars(node,(char *)&fname[0],size);	/* get the file name */
		strcpy(fn.fileName,fname);
		if((pfn = avl_findsym(ftree,&fn)) != 0)		/* find my local file */
		  {
		    /*
		      If the file name has a ".dbf" extension then make an ".mdx" file
		      name and request that file as well. This is done to avoid a mismatch
		      between the dbf and mdx files which must be together..
		      11-Oct-96 CAREFULL -- some dbase files do not have MDX partners
		      29-Oct-1997 send MDX file first followed by the DBF. This will ensure
		      that the DBF file will have a more recent date for those programs
		      which are semsitive to that sort of thing. This was fixed here to
		      make is consistant with the send-now code.
		    */
		    if((dbfPtr = strstr(fname,HSYNC_DBF_FILE_EXT)) != 0)
		      {
			strcpy(dbfPtr,HSYNC_MDX_FILE_EXT);
			strcpy(fname2,pfn->fileName);
			dbfPtr = strstr(fname2,HSYNC_DBF_FILE_EXT);
			strcpy(dbfPtr,HSYNC_MDX_FILE_EXT);
			if((status = stat(fname2,&stBuff)) != -1)
			  {
			    sprintf(rcpCmd,"rcp -p %s %s:%s &",fname2,node->hostName,fname2);
			    VERBOSE(V_DEBUG)(0,0,"RCP command %s\n",rcpCmd);
			    if(system(rcpCmd) < 0)
			      VERBOSE(V_DEBUG)(0,0,"Error from SYSTEM/RCP: %m");
			  }
		      }
		    sprintf(rcpCmd,"rcp -p %s %s:%s &",pfn->fileName,node->hostName,fname);
		    VERBOSE(V_DEBUG)(0,0,"RCP command %s\n",rcpCmd);
 		    if(system(rcpCmd) < 0)
		      VERBOSE(V_DEBUG)(0,0,"Error from SYSTEM/RCP: %m");
		  }
		else
		  {
		    VERBOSE(V_INFO)(0,0,"File %s not found in tree on %s",
								fname,node->hostName);
		  }
		break;
	      }
	  }
      }
  }
}

void hsSendFileInfo(HOST_NODE_PTR hinfo)     	/* Send file info to other hosts */
{
  int sendSize,count;
  char *buff;

  buff = hsGet20(hinfo,&sendSize,&count);
  if(sendSize > 0)
    {
      hsWriteChars(hinfo,buff,sendSize);
      VERBOSE(V_STATUS)(0,0, "Sending %d messages of file info to <%s> on socket <%d>",
			count,hinfo->hostName,hinfo->hostFD);
    }
}
/*
  gwt file D&T and assemble messages into one large buffer for later transmission
*/
void hsGetFileInfo(void *p1,void *f_info)
{
FILE_NODE_PTR	finfo = (FILE_NODE_PTR)f_info;
struct stat stBuff;
char buff[80],*localBuff;
int totalSize;
#ifdef	IMMEDIATE_UPDATE
int	i, status;
char	fname[MAXPATHLEN],fname2[MAXPATHLEN], rcpCmd[MAXPATHLEN];
char	*dbfPtr;
#endif	IMMEDIATE_UPDATE

localBuff = &buff[0];
if(stat(finfo->fileName,&stBuff) != -1)
  {
#ifdef IMMEDIATE_UPDATE
    if(stBuff.st_mtime > finfo->fileLocalTime)	/* if current file time greater than the last time I looked? */
      {
	totalSize = hsFillSendBuff(localBuff,finfo);	/* yes then tell everybody about it */
	for(i = 0;i < maxConnected;i++)
	  {
	    VERBOSE(V_STATUS)(0,0,"Sending file %s to %s",
						finfo->fileName,hostParent[i].hostName);
	    sprintf(rcpCmd,"rcp -p %s %s:%s &",finfo->fileName,hostParent[i].hostName,fname);
	    VERBOSE(V_DEBUG)(0,0,"RCP command %s\n",rcpCmd);
	    if(system(rcpCmd) < 0)
	      VERBOSE(V_DEBUG)(0,0,"Error from SYSTEM/RCP: %m");
	    /*
	      If the file name has a ".dbf" extension then make an ".mdx" file
	      name and request that file as well. This is done to avoid a mismatch
	      between the dbf and mdx files which must be together..
	      11-Oct-96 CAREFULL -- some dbase files do not have MDX partners
	      */
	    if((dbfPtr = strstr(fname,HSYNC_DBF_FILE_EXT)) != 0)
	      {
		strcpy(dbfPtr,HSYNC_MDX_FILE_EXT);
		strcpy(fname2,finfo->fileName);
		dbfPtr = strstr(fname2,HSYNC_DBF_FILE_EXT);
		strcpy(dbfPtr,HSYNC_MDX_FILE_EXT);
		if((status = stat(fname2,&stBuff)) != -1)
		  {
		    sprintf(rcpCmd,"rcp -p %s %s:%s &",fname2,hostParent[i].hostName,fname2);
		    VERBOSE(V_DEBUG)(0,0,"RCP command %s\n",rcpCmd);
		    if(system(rcpCmd) < 0)
		      VERBOSE(V_DEBUG)(0,0,"Error from SYSTEM/RCP: %m");
		  }
	      }
	  }
      }
#endif
    finfo->fileLocalTime = stBuff.st_mtime;	       	/* get file time */
  }
else
  {
    finfo->fileLocalTime = 0;
  }
totalSize = hsFillSendBuff(sendBuff,finfo);
sendBuff += totalSize;
}
int hsFillSendBuff(char *buff,FILE_NODE_PTR finfo)
{
  int size,sz,totalSize;
  short msgType;
  time_t lt;

  totalSize = 0;
  /* copy message type */
  msgType = htons(HS_MSG_INFO);
  memcpy(buff,&msgType,sizeof(short));
  ((short *)buff)++;		/* assemble message */
  totalSize += sizeof(msgType);
  /* copy time */
  lt = htonl(finfo->fileLocalTime);
  memcpy(buff,(char *)&lt,sizeof(time_t));
  ((time_t *)buff)++;
  totalSize += sizeof(lt);
  /* copy file name length */
  size  = strlen(finfo->fileName)+1;	/* include NULL */
  sz = htonl(size);
  memcpy(buff,(char *)&sz,sizeof(int));
  ((int *)buff)++;
  totalSize += sizeof(sz);
  /* copy name itself */
  memcpy(buff,&finfo->fileName,strlen(finfo->fileName)+1);
  buff += size;
  totalSize += size;
  return(totalSize);
}

void hsSetNoBlock(HOST_NODE_PTR node)
{
#ifdef USE_NON_BLOCK
int  stat;
char trueFlag = 1;
#ifdef SOLARIS
stat = fcntl(node->hostFD,F_GETFL,trueFlag);
if(stat >= 0)
  {
    stat |= O_NDELAY;
    stat = fcntl(node->hostFD,F_SETFL,stat);
  }
#else
stat = ioctl(node->hostFD,FIONBIO,&trueFlag);	/* GCT 2-Feb-1995 set to non-blocking */
#endif
if(stat < 0)
  {
    VERBOSE(V_STATUS)(0,0, "Error setting Non-blocking I/O for %s...%m", node->hostName);
  }
#endif
}
/****************************************************************/
/*								*/
/*	MAX OUT A SOCKET'S SEND AND RECEIVE BUFFERS		*/
/*								*/
/*	The sockets that connect the Socket Creation and	*/
/*	Maintenance program with Channel Acquisition and	*/
/*	Control require a lot of buffer space, especially	*/
/*	for large systems. The maximum transmit and receive	*/
/*	buffers are implementation specific, so we will		*/
/*	bump up each buffer until it can't be bumped any	*/
/*	more. This will guarantee that we always have as	*/
/*	much buffering space as the system will let us have.	*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

maxSocket(int fd)
{
  int		size;
  int		intsize = sizeof(int);

  for (;;)
    {
      getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&size, &intsize); 
      size += 1024;
      if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&size, sizeof(size)) == -1)
	break;
    }
  
  for (;;)
    {
      getsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&size, &intsize); 
      size += 1024;
      if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&size, sizeof(size)) == -1)
	break;
    }
}
/*
  Return pointer to & size of up to hsCount (default = 20) complete messages. After < hsCount
  then reset for next time around.
*/
char *hsGet20(HOST_NODE_PTR node,int *size,int *count)
{
int i,fsize;
char *start;
INFO_MSG info;

*size = 0;
start = &sendOut[node->hostSendIndex];
for(i = 0;i < hsCount;i++)
  {
    if(node->hostSendIndex >= sendBuffSize)
      break;
    memcpy((char *)&info,&sendOut[node->hostSendIndex+sizeof(short)],sizeof(INFO_MSG));
    fsize = ntohl(info.infoNameSize);
    *size += sizeof(short) + sizeof(INFO_MSG) + fsize;
    node->hostSendIndex += (sizeof(short) + sizeof(INFO_MSG) + fsize);
  }
if(i < hsCount)
  {
    node->hostSendIndex = 0;	/* start over */
    node->hostPass = HOST_PASS2;
    node->hostExchangeTime = time(0);	/* this is so we do totalFiles/hsCount messages every hsExchange seconds */
  }					/* or hsCount files every hsExchange seconds */
*count = i;
return(start);
}
