/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include "ims.h"
static char line[80],nextChar;

int imsParseControl(CONTROL_INFO_PTR ctl)
{
char *str,inLine[80];
int secs;
ctl->imsImportPoints = avl_init(imsPointCompare);		/* init trees for both groups */
ctl->imsExportPoints = avl_init(imsPointCompare);
/* Import half */
str = fgets(&ctl->imsImportFileName[0],64,imsControlFile);     	/* get export file name */
ctl->imsImportFileName[strlen(str)-1] = 0;			/* drop \n */
str = fgets(&ctl->imsImportScript[0],64,imsControlFile);
ctl->imsImportScript[strlen(str)-1] = 0;		       	/* drop \n */
ctl->imsMaxRate = -1;
while((str = fgets(&inLine[0],80,imsControlFile)) != NULL)
  {
    inLine[strlen(str)-1] = 0;					/* drop \n */
    if(strlen(inLine) == 0) break;
    imsAddToImportTree(ctl,inLine);
  }
/* Export half */
str = fgets(&ctl->imsExportFileName[0],64,imsControlFile);     	/* get export file name */
ctl->imsExportFileName[strlen(str)-1] = 0;			/* drop \n */
str = fgets(&inLine[0],80,imsControlFile);
inLine[strlen(str)-1] = 0;				       	/* drop \n */
secs = atof(inLine)*60;
ctl->imsExportRate = (secs/5)*5;
ctl->imsRateSave = ctl->imsExportRate;				/* save here to reset */
while((str = fgets(&inLine[0],80,imsControlFile)) != NULL)
  {
    inLine[strlen(str)-1] = 0;					/* drop \n */
    if(strlen(inLine) == 0) break;
    imsAddToExportTree(ctl,inLine);
  }
ctl->imsImportFile = 0;
return(1);
}

/* compare two point names */
int  imsPointCompare(void *e1,void *e2)
  {
    return(strcmp(((POINT_INFO_PTR)e1)->imsPointName,((POINT_INFO_PTR)e2)->imsPointName));
  }

/* Add to the import tree */
void imsAddToImportTree(CONTROL_INFO_PTR ctl,char *pointStr)
{
  char *sp;
  int  castat;
  double rate;
  POINT_INFO  pointInfo,*pInfo;
  evid EVID;

  memset((char *)&pointInfo,0,POINT_INFO_SIZE);
  strcpy(pointInfo.imsPointName,"DUMMY");			/* point name if none on line */
  rate = 0.0;
  sp = strtok(pointStr," ");		       		/* get name from line */
  if(isalpha(*sp))
    strcpy(pointInfo.imsPointName,sp);
  if((sp = strtok(NULL," ")) != NULL)			/* and rate from rest of line */
    rate = atof(sp);
  pointInfo.imsRate = rate * 60.0;
  pointInfo.imsRate = (pointInfo.imsRate/5)*5;       	/* round to nearest 5 seconds */
  if(pointInfo.imsRate > ctl->imsMaxRate) ctl->imsMaxRate = pointInfo.imsRate;
  pointInfo.imsSaveRate = pointInfo.imsRate;		/* save here for reset */
/* insert in tree */
  pInfo = avl_findsym(ctl->imsImportPoints,&pointInfo);
  if(pInfo == NULL)
    {
      if(!(pInfo = avl_newnode(POINT_INFO_SIZE)))
	{
	  tisSysLog(0,0,"No memory for Import tree insert...%s",pointInfo.imsPointName);
	  return;
	}
      *pInfo = pointInfo;
      avl_insert(ctl->imsImportPoints,pInfo);
      /* attach this point to CA */
      if((castat = ca_build_read_and_monitor(pInfo->imsPointName,
					     imsCAConnect,pInfo,
					     imsCARead,pInfo,
					     imsCAMonitor,pInfo,
					     &pInfo->imsCHID,&EVID)) != ECA_NORMAL)
	pInfo->imsCHID = 0;
      else
	{
	  ca_pend_io(0.1);
	  pInfo->imsCHID->paddr = (void *)pInfo;
	}
    }
}

/* Add to the export tree */
void imsAddToExportTree(CONTROL_INFO_PTR ctl,char *pointStr)
{
  char *sp,name[PVNAME_SZ+1];
  int  castat;
  POINT_INFO  pointInfo,*pInfo;
  evid EVID;

  memset((char *)&pointInfo,0,POINT_INFO_SIZE);
  strcpy(pointInfo.imsPointName,"DUMMY");	       	/* point name if none on line */
  sp = strtok(pointStr," ");		       		/* get name from line */
  if(isalpha(*sp))
    {
      strcpy(pointInfo.imsPointName,sp);
      strcpy(name,sp);
    }
  sp = strtok(NULL," ");				/* get data type A or N */
  pointInfo.imsDataType = (toupper(*sp) == 'A' ? 'A' : 'N');
  if((sp = strtok(NULL," ")) != NULL)  			/* get alarm switch A or T */
    pointInfo.imsDoAlarm = (toupper(*sp) == 'A' ? 1 : 0);
  if((sp = strtok(NULL," ")) != NULL)  			/* get trend switch A or T */
    pointInfo.imsDoTrend = (toupper(*sp) == 'T' ? 1 : 0);
/* insert in tree */
  pInfo = avl_findsym(ctl->imsExportPoints,&pointInfo);
  if(pInfo == NULL)
    {
      if(!(pInfo = avl_newnode(POINT_INFO_SIZE)))
	{
	  tisSysLog(0,0,"No memory for Export tree insert...%s",pointInfo.imsPointName);
	  return;
	}
      *pInfo = pointInfo;
      avl_insert(ctl->imsExportPoints,pInfo);
      /* attach this point to CA */
      if((castat = ca_build_read_and_monitor(pInfo->imsPointName,
					     imsCAConnect,pInfo,
					     imsCARead,pInfo,
					     imsCAMonitor,pInfo,
					     &pInfo->imsCHID,&EVID)) != ECA_NORMAL)
	pInfo->imsCHID = 0;
      else
	{
	  ca_pend_io(0.1);
	  pInfo->imsCHID->paddr = (void *)pInfo;
	}
    }
}

int imsParseExport(CONTROL_INFO_PTR ctl)
{
char *str,*status,*severity,*value,*pChoice;
short i;
time_t tt;
POINT_INFO  pointInfo,*pInfo;
struct choiceSet *set;
double val;

tt = time(0);
if((ctl->imsExportFile = fopen(&ctl->imsExportFileName[0],"r")) == NULL)	/* open the export file */
  return(0);
while((str = fgets(line,80,ctl->imsExportFile)) != NULL)
  {
    line[strlen(str)-1] = 0;
    nextChar = 0;
    if((str = imsGetToken()) != NULL)
      {
	strcpy(pointInfo.imsPointName,str);
	pInfo = avl_findsym(ctl->imsExportPoints,&pointInfo);
	if(pInfo != NULL)
	  {
	    value = imsGetToken();
	    if(pInfo->imsDataType == 'A')
	      strcpy(pInfo->imsPointValue,value);
	    else
	      {
		val = (double)atof(value);
		memcpy(&pInfo->imsPointValue[0],(char *)&val,sizeof(double));
	      }
	    status = imsGetToken();
	    if((set = GET_PCHOICE_SET(choiceGbl,GBL_ALARM_STAT)) == 0)
	      i = 0;
	    else
	      {
		for(i = 0;i < set->number; i++)
		  {
		    pChoice = GET_CHOICE(set,i);
		    if(!strcmp(pChoice,status))
		      break;
		  }
	      }
	    pInfo->imsPointStatus = (i >= set->number) ? 0 : i;
	    severity = imsGetToken();
	    if((set = GET_PCHOICE_SET(choiceGbl,GBL_ALARM_SEV)) == 0)
	      i = 0;
	    else
	      {
		for(i = 0;i < set->number; i++)
		  {
		    pChoice = GET_CHOICE(set,i);
		    if(!strcmp(pChoice,severity))
		      break;
		  }
	      }
	    pInfo->imsPointSeverity = (i >= set->number) ? 0 : i;
	  }
      }
  }
fclose(ctl->imsExportFile);
return(1);
}

char *imsGetToken(void)
{
char quote=0,*ptr;
while(isspace(line[nextChar])) nextChar++;
ptr = &line[nextChar];
while(line[nextChar] != 0)
  {
    if(line[nextChar] == '"')
      {
	if(quote ^= 1)
	  ptr++;
	else
	  line[nextChar] = 0;
	nextChar++;
      }
    if((line[nextChar] == ' ') && (quote == 0))
      {
	line[nextChar++] = 0;
	return(ptr);
      }
    nextChar++;
  }
return(ptr);
}
void imsRegisterCA(void *arg, int fd, int cond)	/* Reconnect to CA */
{
  if (cond == 0)
    {
      tisSysLog (0,0, "Connection to CA lost.");
    }
  else
    {
    caSocket = fd;				/* save CA fd here */
    if(caSocket > maxFd) maxFd = caSocket;	/* add to group I'm interested in */
    FD_SET(caSocket,&myFds);
    memcpy((char *)&saveFds,(char *)&myFds,sizeof(fd_set));
    }
}
void imsCAConnect(struct connection_handler_args arg)
{
if(arg.op == CA_OP_CONN_UP)
  ca_complete_bld_read_mon(arg.chid);
}

void imsCAMonitor(struct event_handler_args arg)
{
POINT_INFO_PTR node;
node = arg.usr;					/* address of the point info node for this point */
switch(arg.type)
  {
  case OLD_DBR_STS_DOUBLE:
    memcpy(&node->imsPointValue[0],(char *)&(((struct dbr_sts_double *)arg.dbr)->value),sizeof(double));
    node->imsPointStatus = ((struct dbr_sts_double *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_sts_double *)arg.dbr)->severity;
    break;
  case OLD_DBR_STS_FLOAT:
    memcpy(&node->imsPointValue[0],(char *)&(((struct dbr_sts_float *)arg.dbr)->value),sizeof(double));
    node->imsPointStatus = ((struct dbr_sts_float *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_sts_float *)arg.dbr)->severity;
    break;
  case OLD_DBR_STS_CHAR:
    memcpy(&node->imsPointValue[0],(char *)&(((struct dbr_sts_char *)arg.dbr)->value),sizeof(char));
    node->imsPointStatus = ((struct dbr_sts_char *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_sts_char *)arg.dbr)->severity;
    break;
  case OLD_DBR_STS_LONG:
    memcpy(&node->imsPointValue[0],(char *)&(((struct dbr_sts_long *)arg.dbr)->value),sizeof(int));
    node->imsPointStatus = ((struct dbr_sts_long *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_sts_long *)arg.dbr)->severity;
    break;
  case OLD_DBR_STS_SHORT:
  case OLD_DBR_STS_ENUM:
    memcpy(&node->imsPointValue[0],(char *)&(((struct dbr_sts_short *)arg.dbr)->value),sizeof(short));
    node->imsPointStatus = ((struct dbr_sts_short *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_sts_short *)arg.dbr)->severity;
    break;
  case OLD_DBR_CTRL_STRING:
  case OLD_DBR_STS_STRING:
    strcpy(node->imsPointValue, ((struct dbr_sts_string *)arg.dbr)->value);
    node->imsPointStatus = ((struct dbr_sts_string *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_sts_string *)arg.dbr)->severity;
    break;
  case OLD_DBR_CTRL_DOUBLE:
    memcpy(&node->imsPointValue[0],(char *)&(((struct dbr_ctrl_double *)arg.dbr)->value),sizeof(double));
    node->imsPointStatus = ((struct dbr_ctrl_double *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_ctrl_double *)arg.dbr)->severity;
    break;
  case OLD_DBR_CTRL_FLOAT:
    memcpy(&node->imsPointValue[0],(char *)&(((struct dbr_ctrl_float *)arg.dbr)->value),sizeof(double));
    node->imsPointStatus = ((struct dbr_ctrl_float *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_ctrl_float *)arg.dbr)->severity;
    break;
  case OLD_DBR_CTRL_CHAR:
    memcpy(&node->imsPointValue[0],(char *)&(((struct dbr_ctrl_char *)arg.dbr)->value),sizeof(char));
    node->imsPointStatus = ((struct dbr_ctrl_char *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_ctrl_char *)arg.dbr)->severity;
    break;
  case OLD_DBR_CTRL_LONG:
    memcpy(&node->imsPointValue[0],(char *)&(((struct dbr_ctrl_long *)arg.dbr)->value),sizeof(int));
    node->imsPointStatus = ((struct dbr_ctrl_long *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_ctrl_long *)arg.dbr)->severity;
    break;
  case OLD_DBR_CTRL_SHORT:
  case OLD_DBR_CTRL_ENUM:
    memcpy(&node->imsPointValue[0],(char *)&(((struct dbr_ctrl_short *)arg.dbr)->value),sizeof(short));
    node->imsPointStatus = ((struct dbr_ctrl_short *)arg.dbr)->status;
    node->imsPointSeverity = ((struct dbr_ctrl_short *)arg.dbr)->severity;
    break;
  }
}

void imsCARead(struct event_handler_args arg)
{
imsCAMonitor(arg);
}
