/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/***************************************************************************
 merge_optrak

merge the source op_trak file into the destination op_trak file
****************************************************************************/

#include <stdio.h>
#include <time.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "msgMon.h"
#include "OperPrvDefs.h"
#include "trak.c"

#define TRUE          1
#define FALSE         0
#define PREV_HOURS    24
#define SECS_PER_HOUR 3600
#define MAXPATH       100

int GetRecTime ( D4DATA *FileDesc, char StartBuf[], char StopBuf[] );


/************************************************************************
*			PUBLIC DECLARATIONS				*
************************************************************************/

char		rbuf[20]; /* Read buffer for keeping track of last run - the base time */
time_t		CurEpoch;
time_t		CutOffEpoch;
struct tm	CurTm;
struct tm	CutOffTm;
int		CutOff;		/* Cutoff point for keeping data records */

/*********************************************************************************
 copy the records from the source db file that do not have a match in the 
 destination db file, to the destination db file.
*/

void merge_files(D4DATA *source, D4DATA *destination, char start[], char stop[])
{
    int    found;
    char   source_op[31], source_date[9], source_time[9], source_pt[29], source_fld[5];
    char   dest_op[31], dest_date[9], dest_time[9], tag_string[20], dest_pt[29], dest_fld[5];
    long   srctime, dsttime;

    while (!d4eof(source)) {

	if (!GetRecTime( source, start, stop))  {
	    /* If the record is not with in the start stop date range */
            d4skip(source, 1);
	    d4unlock_all(source);
	    continue;
        }

	found = FALSE;

	/* seek to the destination for the source date/time */

	memset( source_date, 0, sizeof source_date);
	strcpy( source_date, f4str(d4field(source, "TRK_DATE")));

	memset( source_time, 0, sizeof source_time);
	strcpy( source_time, f4str(d4field(source, "TRK_TIME")));

	memset(tag_string, 0, sizeof tag_string);
	sprintf(tag_string, "%s%s", source_date, source_time);

        memset( source_op, 0, sizeof source_op);
        strcpy( source_op, f4str(d4field(source, "OPERATOR")));

        memset( source_pt, 0, sizeof source_pt);
        strcpy( source_pt, f4str(d4field(source, "POINTNAME")));

        memset( source_fld, 0, sizeof source_fld);
        strcpy( source_fld, f4str(d4field(source, "FIELDNAME")));

	srctime = f4long (d4field(source, "UNIX_TIME"));

	d4unlock_all(source);

	if ( d4seek( destination, tag_string ) == 0) { 
	
	   memset( dest_date, 0, sizeof dest_date);
	   strcpy( dest_date, f4str(d4field(destination, "TRK_DATE")));

	   memset( dest_time, 0, sizeof dest_time);
	   strcpy( dest_time, f4str(d4field(destination, "TRK_TIME")));

           memset( dest_op, 0, sizeof dest_op);
           strcpy( dest_op, f4str(d4field(destination, "OPERATOR")));

           memset( dest_pt, 0, sizeof dest_pt);
           strcpy( dest_pt, f4str(d4field(destination, "POINTNAME")));

           memset( dest_fld, 0, sizeof dest_fld);
           strcpy( dest_fld, f4str(d4field(destination, "FIELDNAME")));

	   dsttime = f4long (d4field(destination, "UNIX_TIME"));

	   d4unlock_all(destination);

	   while ( (strcmp(source_date, dest_date) == 0) &&
	           (strcmp(source_time, dest_time) == 0) ){

	      if ( (strcmp(source_op, dest_op) == 0) &&
		   (strcmp(source_pt, dest_pt) == 0) &&
		   (strcmp(source_fld, dest_fld) == 0)) {

	          /* a record is considered to match if the date, time,
	             operator, pointname, and field are the same. */

	          found = TRUE;

                  /* If a match is found but ONLY ONE is marked for deletion, 
                     undelete it. */

                  if (d4deleted(source)) {
                    if (!d4deleted(destination))
                        d4recall(source);
                  }
                  else  {
	            if (d4deleted(destination))
	                d4recall(destination);
                  }
		  d4unlock_all(source);
		  d4unlock_all(destination);

	          break;
             }
             else {
	           d4skip(destination, 1);

	           memset( dest_date, 0, sizeof dest_date);
	           strcpy( dest_date, f4str(d4field(destination, "TRK_DATE")));

	           memset( dest_time, 0, sizeof dest_time);
	           strcpy( dest_time, f4str(d4field(destination, "TRK_TIME")));

                   memset( dest_op, 0, sizeof dest_op);
                   strcpy( dest_op, f4str(d4field(destination, "OPERATOR")));

           	   memset( dest_pt, 0, sizeof dest_pt);
           	   strcpy( dest_pt, f4str(d4field(destination, "POINTNAME")));

           	   memset( dest_fld, 0, sizeof dest_fld);
           	   strcpy( dest_fld, f4str(d4field(destination, "FIELDNAME")));

		   dsttime = f4long (d4field(destination, "UNIX_TIME"));

		   d4unlock_all(source);
		   d4unlock_all(destination);
             }
	  }
       }
       if (!found) 
          /* If the source record is not found in the destination db file */
          AddToList(destination);

      d4skip(source, 1);
      d4top(destination);
      d4unlock_all(source);
      d4unlock_all(destination);
   }
   d4unlock_all(source);
   d4unlock_all(destination);
}

/**************************************************************************************/

void get_mergetimes(char *endtime, char *starttime, int hours)

{

    /*Get the current time to obtain stopping time for merging.*/

	time( &CurEpoch );
	CurTm = *(localtime( &CurEpoch ));
	CutOffEpoch = CurEpoch;
	CutOffTm = *(localtime( &CutOffEpoch ));
	memset( endtime, 0, sizeof rbuf );
	if( CutOffTm.tm_year >= 70 )  {
		sprintf( endtime, "19%2.2d%2.2d%2.2d%2.2d", 
			CutOffTm.tm_year % 100, CutOffTm.tm_mon+1, CutOffTm.tm_mday, CutOffTm.tm_hour );
	}
	else  {
		sprintf( endtime, "20%2.2d%2.2d%2.2d%2.2d", 
			CutOffTm.tm_year % 100, CutOffTm.tm_mon+1, CutOffTm.tm_mday, CutOffTm.tm_hour );
	}

	/* obtain starting time for merging files */

	if (hours)
	    CutOffEpoch = CutOffEpoch - (hours * SECS_PER_HOUR);
        else
	    CutOffEpoch = CutOffEpoch - (PREV_HOURS * SECS_PER_HOUR);

        CutOffTm = *(localtime(&CutOffEpoch));
	memset( starttime, 0, sizeof rbuf);
	if( CutOffTm.tm_year >= 70 )  {
		sprintf( starttime, "19%2.2d%2.2d%2.2d%2.2d", 
			CutOffTm.tm_year % 100, CutOffTm.tm_mon+1, CutOffTm.tm_mday, CutOffTm.tm_hour );
	}
	else  {
		sprintf( starttime, "20%2.2d%2.2d%2.2d%2.2d", 
			CutOffTm.tm_year % 100, CutOffTm.tm_mon+1, CutOffTm.tm_mday, CutOffTm.tm_hour );
	}
}

/**************************************************************************************/
int GetRecTime ( D4DATA *FileDesc, char StartBuf[], char StopBuf[] )
{
	int	yr;
	char	DateTimeStr[20];
	char	CmpDateTime[10];

	memset( DateTimeStr, 0, sizeof DateTimeStr );
	memset( CmpDateTime, 0, sizeof CmpDateTime );
	strcpy( DateTimeStr, f4str( d4field( FileDesc, "TRK_DATE" )));
	yr = atoi( &DateTimeStr[6] );

	/* Get first two characters of time buf - we're only interested in hour value */
	sprintf( &DateTimeStr[strlen( DateTimeStr )], "  %2.2s", 
					f4str( d4field( FileDesc, "TRK_TIME" )));

	d4unlock_all(FileDesc);

	/* Now build comparison buffer */
	if( yr >= 70 )  {
		sprintf( CmpDateTime, "19%2.2s%2.2s%2.2s%2.2s",
			&DateTimeStr[6], DateTimeStr, &DateTimeStr[3], &DateTimeStr[10] );
	}
	else  {
		sprintf( CmpDateTime, "20%2.2s%2.2s%2.2s%2.2s",
			&DateTimeStr[6], DateTimeStr, &DateTimeStr[3], &DateTimeStr[10] );
	}

	if(((strcmp( CmpDateTime, StartBuf )) >= 0 ) && ((strcmp( StopBuf, CmpDateTime )) > 0 )){
		return 1;
	}
	else {
		return 0;
	}
}
/************************************************************************************/
void usage()
{
    printf ("Usage: MergeTrak {source} [hours]\twhere:");
    printf ("\t{File1} = Merge, complete path to op_trak.dbf, source file;");
    printf ("\t[hours] = Optional parameter, number of hours to merge, default is 24;");
}

/***********************************************************************************/

void main(int argc, char *argv[])

{
    D4DATA               *source, *destination;
    char                 starttime[sizeof rbuf], endtime[sizeof rbuf];
    int                  hours = 0;

      if (argc < 2 || argc > 3) {
	  usage();
	  exit( 1 );
      }
      else {
          if (argc ==3) {
	     hours = atoi(argv[2]);
	     if (!hours) {
		 tisSysLog(0,0,"MergeTrak : The number of hours to merge must be an integer greater than 0");
		 exit( 1 );
             }
           }
      }

     get_mergetimes(endtime, starttime, hours);

      /* initialize codebase */ 
       initCodebase();

       if (! (source = openSourceOpTrak(argv[1]))) {
		tisSysLog(0,0, "MergeTrak : Can not open %s, no merging will be done", argv[1]);
		exit( 1 );
       }
#ifdef WIN32
       if (! (destination = openDestOpTrak("\\tis4000db\\op_trak.dbf"))) {
#else
       if (! (destination = openDestOpTrak("/tis4000db/op_trak.dbf"))) {
#endif WIN32
		tisSysLog(0,0,"MergeTrak : Can not open %s, no merging will be done", "/tis4000db/op_trak.dbf");
		closeTrak(source);
		exit( 1 );
       }

       if ( !(d4tag(source, "UNIXTIME")) ||
	    !(d4tag(destination, "UNIXTIME")) ) {

		closeTrak(source);
		closeTrak(destination);

		tisSysLog(0,0, "MergeTrak : tag does not match op_trak file");
		exit (1);
       }
       d4tag_select(source, d4tag(source, "UNIXTIME"));
       d4tag_select(destination, d4tag(destination, "UNIXTIME"));

       d4top(source);
       d4top(destination);

       d4unlock_all(source);
       d4unlock_all(destination);

       /* merge source into destination */
       merge_files( source, destination, starttime, endtime);

       d4close(source);
       UpdateTrak(destination);
       d4close(destination);

       exit(0);
}

