
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
 /* 	FILE NAME   :	OperPrvTrk.c
  *	PURPOSE	    :	Library routines for operator privilege checking and operator tracking.
  *
  *	AUTHOR	    :	Donald E. Kyne, Jr.
  *	DATE	    :	15-Sep-1992
  *
  *					M O D I F I C A T I O N S
  *
  *	DATE		INITIALS	DETAILS
  *
  *	Copyright 1992, Tate Integrated Systems, L.P.  All rights reserved.
  */
#include <stdio.h>
#include <stdlib.h>
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include "genDefs.h"
#include "db_access.h"
#include "sydDefs.h"
#include "cadef.h"
#ifdef WIN32
#include <io.h>
#include <windows.h>
#include <winbase.h>
#else
#include <sys/time.h>
#include <pwd.h>
#endif WIN32
#include <sys/types.h>
#ifdef AXP
#include <sys/stat.h>
#endif
#if defined(SOLARIS) || defined(WIN32)
#include <sys/stat.h>
#endif
#include <time.h>
#include <fcntl.h>
#include "alarm.h"
#include "OperPriv.h"	  /* Operator tracking definitions ( Error codes, structure defs, etc. ) */
#include "OperPrvDefs.h"  /* More operator tracking definitions ( Structure defs, etc. ) */
#include "tisEncrypt.h"
#include "msgMon.h"

#ifndef FALSE
#define	FALSE		(0)
#define	TRUE		(1)
#endif

#define	LOGGEDOFF	0
#define	LOGGEDON	1

/************************************************************************
 *			PUBLIC DECLARATIONS				*
 ************************************************************************/
 /* NOTE:
 *	for size definitions, ex. PVNAME_SZ, see .../share/epicsH/dbdefs.h
 *	which is included in OperPrvDefs.h
 */
int	add_record( char *TxtStr, char *Channel, char *OldValue, char *NewValue );
char	*getDisplayHost (void);
void	GetOperGroup (void);
int	GetProtGrpPrivilege (char *reqPriv);
int	GetUtilPrivilege (char *reqPriv);
int	OperPriv( C4CODE *cBase, char *myGroup );

static int		cbInit = 0;		/* is local codeBase dexcriptor inited */
static	C4CODE		cb;			/* Codebase status configuration block */
/* Codebase file descriptors */
static	D4DATA		*OperPrv = 0;
static	D4DATA		*TrakFile = 0;
static	D4DATA		*FieldPrv = 0;
static	D4DATA		*PointDB = 0;		/* Point Database file descriptor */
static	D4DATA		*UserGrp = 0;		/* Username/Groupname cross reference table */
static	D4DATA		*UtilAccDb = 0;		/* utility access database */
#if 0
static	I4INDEX		*TrakIndex;		/* TrakFile's index file descriptor */
#endif 0
F4FIELD_INFO		*fld_list;
T4TAG_INFO		*tag_list;
static	short		tagdVal;		/* TAGDEPTH value from PVMAP */
static	char		CurrOper[OPERATOR_SZ+1];	/* Name of operator who's modifying channel */
static	char		OperPrvLevel[PRIV_LEVEL_SZ+1];	/* Operator's privilege level */
static	char		OperGroup[GROUPNAME_SZ+1];	/* Operator's assigned group - if assigned */
static	char		ProtGroup[PROTECTIONGRP_SZ+1];	/* Operator's protection group */
static	char		PtName[PVNAME_SZ+1];
static	char		FldName[FLDNAME_SZ+1];
time_t			CurTime;
struct	tm		CurTmElements;
#ifdef	DAILY_FILE
static	int		DayOfLastRec;		/* Day of the year for the previous tracking
						 * record.  Used to start new daily tracking
						 * file.
						 */
#endif	DAILY_FILE
static	char		TrakName[30];		/* Name of operator tracking file */
static	char		FilePath[40];
static	int		usingTIS = 0;
static	int		operNameInitialized = 0;

#ifdef PRIVDIAG
void prvDiag (char *proc, char *text)
{
    FILE *f;
    struct tm *tm_struct;
    time_t now;

    now = time(0);
    tm_struct = localtime (&now);
#ifdef	WIN32
    f = fopen("\\tis4000db\\loginTrace", "a");
#else
    f = fopen("/tis4000db/loginTrace", "a");
#endif	WIN32
    if (f)
    {
	fprintf(f, "%2.2d/%2.2d/%4.4d %2.2d:%2.2d %s (%d): %s usingTIS = %d, Operator name=%s\n",
			tm_struct->tm_mon+1, tm_struct->tm_mday, tm_struct->tm_year+1900,
			tm_struct->tm_hour, tm_struct->tm_min,
			proc, getpid(), text, usingTIS, CurrOper);
	fclose(f);
    }
}
#endif PRIVDIAG

/*********************************************************************************************/
/* Get the operator name from the TIS operator mechanism if installed */
/* IF not installed, get it from UNIX. */

void GetOperName (char *nameBuff)

    {

	struct stat	status;
	static	char	loginFilePath [100];
	static	int	operNameTime;		/* Timestamp of operator name file */
	FILE	*f;

	/* Determine whether or not we are using the TIS operator login mechanism */

	if (!operNameInitialized) {
#ifdef	WIN32
	    sprintf (loginFilePath, "\\tis4000\\operators\\logins\\%s", getDisplayHost ());
	    usingTIS = _access (loginFilePath, 6);
#else
	    sprintf (loginFilePath, "/tis4000/operators/logins/%s", getDisplayHost ());
	    usingTIS = access (loginFilePath, R_OK | F_OK);
#endif	WIN32
	    if (!usingTIS)
	    {
#ifdef	WIN32
		usingTIS = _access ("\\tis4000\\operators\\secureList", 6);
#else
		usingTIS = access ("/tis4000/operators/secureList", R_OK | F_OK);
#endif	WIN32
	    }

	    operNameTime = 0;
	    operNameInitialized = 1;
	}

	/* If we are not using the TIS mechanism, get the username from UNIX */
	/*
 	GCT 26-Oct-1995 Do following only if this is the first time in this
	function. For some reason 'getpwuid' returns a zero the second time 
	it is called.
	*/
	if (!usingTIS)
		{
		if(!strlen(nameBuff))
			{
#ifdef WIN32
			DWORD bufsize = sizeof(CurrOper);
			GetUserName (CurrOper, &bufsize);
			strcpy (nameBuff, CurrOper);
#else
			strcpy (nameBuff, ( getpwuid( geteuid()))->pw_name );
#endif WIN32
			}
		else
			strcpy (CurrOper, nameBuff);
#ifdef PRIVDIAG
		prvDiag ("GetOperName", "Set user name extracted from Unix");
#endif PRIVDIAG
		return;
		}

	/* We are using the TIS mechanism. Check to see if the login file has changed since the */
	/* last time we read it. */

#ifdef	WIN32
	sprintf (loginFilePath, "\\tis4000\\operators\\logins\\%s", getDisplayHost ());
#else
	sprintf (loginFilePath, "/tis4000/operators/logins/%s", getDisplayHost ());
#endif
	stat (loginFilePath, &status);
	if (operNameTime != status.st_mtime) {
	    f = fopen (loginFilePath, "r");
	    if (f != NULL) {
		fgets (CurrOper, sizeof (CurrOper), f);
		CurrOper [strlen (CurrOper) - 1] = '\0';
		fclose (f);
#ifdef PRIVDIAG
		prvDiag ("GetOperName", "Set name read from file.");
#endif PRIVDIAG
		operNameTime = status.st_mtime;
	    } else {	/* No login file, report as "nooperator" */
		strcpy (CurrOper, "nooperator");
#ifdef PRIVDIAG
		prvDiag ("GetOperName", "Set name to noperator because I couldn't read the file.");
#endif PRIVDIAG
	    }
	}
	    
	/* Return the name to the caller */

	strcpy (nameBuff, CurrOper);
 
    }


/*******************************************************************************************/

char *getDisplayHost (void)

    /* Extract the host name from the DISPLAY environment variable. If DISPLAY */
    /* does not contain a host name, get the host name directly. */

    {

	static	char	hostName [50];
		char	*p;
		char	*env;

	/* Get the value of the DISPLAY environment variable and replace the colon */
	/* with a NULL. */

	hostName [0] = 0;
	env = getenv ("DISPLAY");
	if (env != NULL)
	{
	    strcpy (hostName, getenv ("DISPLAY"));
	    p = strchr (hostName, ':');
	    if (p)
		*p = '\0';
	}

	/* If the hostName is zero length then get the actual hostname */

	if (strlen (hostName) == 0)
	    gethostname (hostName, sizeof (hostName));

	return hostName;

    }


/********************************************************************************************/

int	InitOperPriv( OperName )
char		*OperName;
{

	/* Initialize dBase environment, and open PrivTab.dbf 
	 * for point's (NOT field) group level assigment.
	 * 
	 *	RETURNS  :	PRIVOK or PRIVNOCONN
	 */

	if (cbInit == 0)
	{
	    d4init( &cb );		/* Initialize the Codebase settings. */
#if defined(USECB63) || defined(WIN32)
	    cb.errGo = 0;
	    cb.errCreate = 0;
	    cb.errOpen = 0;
	    cb.errTagName = 0;
	    cb.errFieldName = 0;
	    cb.errOff = 1;		/* disbale codebase error msgs */
	    cb.readOnly = 1;
#else
	    cb.go_error = 0;
	    cb.create_error = 0;
	    cb.open_error = 0;
	    cb.tag_name_error = 0;
	    cb.field_name_error = 0;
#ifdef CB51
	    cb.read_only = 1;
#endif CB51
#endif USECB63
	    cbInit = 1;
	}
	GetOperName( CurrOper );	/* Find out who I am */

	if( OperName )  {
	    strncpy( OperName, CurrOper, OPERATOR_SZ );		/* Save the operator name for later */
	}

    if (PointDB == 0)
    {
#ifdef WIN32
	if( !( PointDB = d4open( &cb, "\\tis4000db\\pvmap" )))   {
#else
	if( !( PointDB = d4open( &cb, "/tis4000db/pvmap" )))   {
#endif WIN32
		/* No Point Database was found. */
		return PRIVNOCONN;
	}
	d4unlock_all( PointDB );
    }

    if (OperPrv == 0)
    {
#ifdef WIN32
	if( !( OperPrv = d4open( &cb, "\\tis4000db\\privtab" )))   {
#else
	if( !( OperPrv = d4open( &cb, "/tis4000db/privtab" )))   {
#endif WIN32
		/* No privilege table is found. */
		return PRIVNOCONN;
	}
	d4unlock_all( OperPrv );
    }

    if (FieldPrv == 0)
    {
#ifdef WIN32
	if( !( FieldPrv = d4open( &cb, "\\tis4000db\\fldprott" )))   {
#else
	if( !( FieldPrv = d4open( &cb, "/tis4000db/fldprott" )))   {
#endif WIN32
		/* No Field Protection Table was found. */
		return PRIVNOCONN;
	}
	d4unlock_all( FieldPrv );
    }

    if (UserGrp == 0)
    {
#ifdef WIN32
	UserGrp = d4open( &cb, "\\tis4000db\\sysusers" );
#else
	UserGrp = d4open( &cb, "/tis4000db/sysusers" );
#endif WIN32
	if (UserGrp)
	{
	    /* unlock it if it exists, otherwise don't worry about it */
	    d4unlock_all( UserGrp );
	}
    }

    if (UtilAccDb == 0)
    {
#ifdef WIN32
	UtilAccDb = d4open( &cb, "\\tis4000db\\utilacc" );
#else
	UtilAccDb = d4open( &cb, "/tis4000db/utilacc" );
#endif WIN32
	if (UtilAccDb)
	{
	    /* unlock it if it exists, if not dont worry, for backwords comaptability */
	    d4unlock_all( UtilAccDb );
	}
    }
    return PRIVOK;
}


/*******************************************************************************************/
#define CONVERTSUB
#ifdef WIN32
#define TIS4000DB   "\\tis4000db\\"
#else
#define TIS4000DB   "/tis4000db/"
#endif WIN32
#define DBCNVRT	    "optraksave"
#define POSTPROCESS InsertUnixTime
#define FIELDINFO   TrakFlds
#define TAGINFO	    TrakTag

static void InsertUnixTime (DATA4 *data)
{
    struct tm tmstruct;
    char *stringp;

    /* Convert current record date time to time_t value */
    stringp = f4str( d4field( data, "TRK_DATE" ));
    tmstruct.tm_year = atoi( stringp + 6 );
    tmstruct.tm_mon = atoi( stringp ) - 1;
    tmstruct.tm_mday = atoi( stringp + 3 );
    stringp = f4str( d4field( data, "TRK_TIME" ));
    tmstruct.tm_hour = atoi( stringp );
    tmstruct.tm_min = atoi( stringp + 3 );
    tmstruct.tm_sec = atoi( stringp + 6 );
    tmstruct.tm_isdst = -1;
    f4assign_long ( d4field( data, "UNIX_TIME" ), mktime (&tmstruct));
    
}

#include "convertdb.h"

int	InitOperTracking( OperName )
char		*OperName;
{
    char FullName [80];	/* name + path */

	/* Open tracking file for OperName and announce start of activity.
	 * Current operator is returned in OperName.
	 *
	 *	RETURNS	:	PRIVOK or PRIVNOCONN
	 */

	GetOperName( CurrOper );	/* Find out who I am */

#ifdef DAILY_FILE
	time( &CurTime );
	CurTmElements = *(localtime( &CurTime ));
	DayOfLastRec = CurTmElements.tm_yday;
	memset( TrakName, 0, sizeof( TrakName ));

	/* Construct tracking file name using "occyyddd" format */
	strftime( TrakName, sizeof( TrakName ), "o%Y%j", &CurTmElements );
#else
	strcpy( TrakName, "op_trak" );
#endif
#ifdef CB51
#if defined(USECB63) || defined(WIN32)
	cb.readOnly = 0;
#else
	cb.read_only = 0;
#endif USECB63
#endif CB51

	sprintf (FullName, "%s%s", TIS4000DB, TrakName);
	if( !( TrakFile = d4open( &cb, FullName )))   {
		/* Tracking file must be created. */
#if defined(USECB63) || defined(WIN32)
		cb.errorCode = 0;		/* Clear out the open error */
#else
		cb.error_code = 0;		/* Clear out the open error */
#endif USECB63
		fld_list = TrakFlds;
		tag_list = TrakTag;
		if( !( TrakFile = d4create( &cb, FullName, fld_list, tag_list )))  {
		    return PRIVNOCONN;		/* Can't connect to tracking file */
		}
#ifdef WIN32
		Sleep(1000);		/* Wait a second for file creation to complete */
#else
		sleep(1);		/* Wait a second for file creation to complete */
#endif WIN32

		/* Set file protections for other users in the group */
		memset( FilePath, 0, sizeof FilePath );
		sprintf( FilePath, "%s.dbf", FullName );
#ifndef WIN32
		chmod( FilePath, 0666 );
		sprintf( FilePath, "%s.mdx", FullName );
		chmod( FilePath, 0666 );
#endif WIN32
	}
	else if (d4field (TrakFile, "UNIX_TIME") == 0)   {
		/* Tracking file must be converted. */
		d4close( TrakFile );

		convertDB (TrakName);

		/* Set file protections for other users in the group */
		memset( FilePath, 0, sizeof FilePath );
		sprintf( FilePath, "%s.dbf", TrakName );
#ifndef WIN32
		chmod( FilePath, 0666 );
		sprintf( FilePath, "%s.mdx", TrakName );
		chmod( FilePath, 0666 );
#endif WIN32

		/* Open the converted file */
		if (!( TrakFile = d4open( &cb, FullName )))
		    return PRIVNOCONN;		/* Can't connect to tracking file */
	}
	d4unlock_all( TrakFile );

	GetOperName( CurrOper );	/* Find out who I am */
	if(OperName && !strlen(OperName))
	{
		strncpy( OperName, CurrOper, OPERATOR_SZ );	/* Save the operator name for later */
	}

	return PRIVOK;

}


/*********************************************************************************************/

int	CheckOperPriv( Channel )
char		*Channel;
{
	unsigned int FldPos;
	int	rv;			/* Return value */
	char	PtGrp[PROTECTIONGRP_SZ+1];
	char	PtType[POINTTYPE_SZ+1];	/* Point type place holder */
	char	FldLevel[PRIV_LEVEL_SZ+1];

	if (cbInit == 0)
	    InitOperPriv (NULL);

	/* Determine whether the operator
	 * has privilege to access Channel.
	 *
	 *	RETURNS	:	PRIVOK	(Channel exists in table)
	 *			PRIVNOFIELD (No entry in field protection table)
	 */

	/* First step is to find the point in PVMAP.DBF, and determine the
	 * Protection Group to which it belongs.
	 */
	memset( PtName, 0, sizeof( PtName ));
	memset( PtType, 0, sizeof( PtType ));
	memset( PtGrp, 0, sizeof( PtGrp ));
	memset( FldName, 0, sizeof( FldName ));
	FldPos = strcspn( Channel, "." );
	strncpy( PtName, Channel, FldPos );

	if( FldPos == strlen( Channel ))  {
		strcpy( FldName, "VAL" );
	}
	else  {
		strcpy( FldName, &Channel[FldPos+1] );
	}

/*
	Lookup PtName in PVMAP.DBF and return its PtType & PtGrp.

	GCT 9-Nov-95

	If PV not found then check for ":". If colon found then return PRIVOK
	because this is a DPA point and no further checking is required.
	If the ":" is not found then PV does not exist so return PRIVDBERR.

*/
	if((rv = GetProtGrp( PtType, PtGrp )) != PRIVOK)
		{
		if(strchr(PtName,':'))
			return(PRIVOK);
		return(PRIVDBERR);
		}

	/* The point has been located & its protection group and point type extracted. 
	 * Now, look in Group Protection table to determine if
	 * there is a record for the current operator in the point's
	 * protection group.
	 */
	strcpy( ProtGroup, PtGrp );
	
	if(( rv = GetOperLevel()) != PRIVOK )  {
		return rv;
	}

	/* Now we have the operator's protecton level within the Point Protection Group.
	 * Using the point type extracted from PVMAP.DBF, extract the FIELD's protection
	 * level from the field protection table, and check it against the operator's level
	 * within the Point Protection Group.
	 */
	rv = GetFieldLevel( PtType, FldName, FldLevel );

	/* We now have the following:
	 * 	OperPrvLevel	:	The operator's privilege level within the protection group;
	 *	FldLevel	:	The protection level required to modify the field in question.
	 *
	 * Compare the two to see if the operator is permitted to modify the field in question.
	 */
	if(( strcmp( FldLevel, OperPrvLevel )) > 0 )  {
		return PRIVNOFIELD;
	}

	/* See if the point has been tagged. */
	if( tagdVal > 0 )  {
		return PRIVTAGGED;
	}

	return PRIVOK;

}


/*********************************************************************************************/

int	WriteOperTracking( Channel, OldValue, NewValue )
char		*Channel;
char		*OldValue;
char		*NewValue;
{
	int	rv;	/* Return Value */

	/* Write operator tracking record.
	 *	RETURNS	:	PRIVOK or PRIVNOTRACK
	 */
	rv = add_record( "  OPERATOR TRACKING", Channel, OldValue, NewValue );
	return(( rv ) ? PRIVNOTRACK : PRIVOK );

}


/*********************************************************************************************/

int	OperPrivTrackErr( Code, Message )
int		Code;
char		***Message;
{
	/* Copy error text associated with Code into buffer pointed to by Message.
	 *
	 *	RETURNS	:	PRIVOK or PRVNOCODE
	 */
	switch( Code )
	{
	case PRIVNOCONN:	*Message = InitMsg;	break;
	case PRIVDBERR:		*Message = NoPtMsg;	break;	
	case PRIVNOGROUP:	*Message = NoOprMsg;	break;	
	case PRIVNOFIELD:	*Message = NoOpFldMod;	break;	
	case PRIVNOTRACK:	*Message = NoTrkWrite;	break;
	case PRIVTAGGED:	*Message = FldTagged;	break;
	case PRIVNODEV:		*Message = NoPrinter;	break;
	case PRIVNOACCESS:	*Message = NoAccess;	break;
	case PRIVNONET:		*Message = NoNetAccess;	break;
	default:		*Message = NoErrCode;	break;
	}

	return PRIVOK;

}


/*********************************************************************************************/

int	CloseOperPriv(void)
{
	/* Close files for Operator Privileges
	 *
	 *	RETURNS	:	Yes, it does...
	 */

	/* close pvmap.dbf if it was opened */
	if( PointDB )  {
	    d4unlock_all( PointDB );	/* Just to be on the safe side ... */
	    d4close( PointDB );
	    PointDB = 0;
	}

	/* close the operator privilege table if it was opened */
	if( OperPrv )  {
	    d4unlock_all( OperPrv );	/* Just to be on the safe side ... */
	    d4close( OperPrv );
	    OperPrv = 0;
	}

	/* close the field privilege table if it was opened */
	if( FieldPrv )  {
	    d4unlock_all( FieldPrv );	/* Just to be on the safe side ... */
	    d4close( FieldPrv );
	    FieldPrv = 0;
	}

	/* close the sysusers table if it was opened */
	if( UserGrp )  {
	    d4unlock_all( UserGrp );	/* Just to be on the safe side ... */
	    d4close( UserGrp );
	    UserGrp = 0;
	}

	/* close the utility access table, if it was opened */
	if (UtilAccDb)
	{
	    d4unlock_all (UtilAccDb);
	    d4close (UtilAccDb);
	    UtilAccDb = 0;
	}
	/* the next GetOperName will reset them based on the current login */
	usingTIS = 0;
	operNameInitialized = 0;

	return PRIVOK;

}


/*********************************************************************************************/

int	CloseOperTracking( OperName )
char		*OperName;
{
	/* Close tracking file for OperName
	 *
	 *	RETURNS	:	Yes, it does...
	 */


	if( TrakFile )  {
	    d4unlock_all( TrakFile );	/* Just to be on the safe side ... */
	    d4close( TrakFile );
	    TrakFile = 0;
	}

	return PRIVOK;

}


/*********************************************************************************************/

int GetProtGrp( PtType, PtGrp )
char 		*PtType;
char 		*PtGrp;
{
	T4TAG	*curtag;
	int	rv;


	d4top( PointDB );		/* Start at the beginning of the database */
	curtag = d4tag( PointDB, "POINTNAME" );
	d4tag_select( PointDB, curtag );
	rv = d4seek( PointDB, PtName );	/* Look for point record using expression in using POINTNAME tag */
	d4unlock_all( PointDB );					/* Release the file */

	if( rv )  {
		/* No record of this point. */
		return PRIVDBERR;
	}

	/* See if this record is deleted */

	if (f4true (d4field (PointDB, "DELETEFLAG")))
	    return PRIVDBERR;

	/* The point exists.  Get its type & protection group. */
	strcpy( PtType, f4str( d4field( PointDB, "TYPE" )));		/* Point Type */
	strcpy( PtGrp, f4str( d4field( PointDB, "PRIVGROUP" )));	/* Protection Group for the point */
	if( !strcmp( PtGrp, "                " ))  {
		strcpy( PtGrp, "untitled" );
	}
	tagdVal = atoi( f4str( d4field( PointDB, "TAGDEPTH" )));

	return PRIVOK;
}


/*********************************************************************************************/

int GetOperLevel()
{
	char	SrchBuff[100];
	T4TAG	*curtag;
	int	rv;

	/* Look for user's name in 'sysusers.dbf' and extract the
	 * associated group name.
	 */
	GetOperGroup();

	memset( SrchBuff, 0, sizeof( SrchBuff ));
	if( OperGroup[0] > ' ' )  {
		sprintf( SrchBuff, "%-*.*s%-*.*s", GROUPNAME_SZ, GROUPNAME_SZ, OperGroup, 
						PROTECTIONGRP_SZ, PROTECTIONGRP_SZ, ProtGroup );
	}
	else  {
		sprintf( SrchBuff, "%-*.*s%-*.*s", GROUPNAME_SZ, GROUPNAME_SZ, CurrOper, 
						PROTECTIONGRP_SZ, PROTECTIONGRP_SZ, ProtGroup );
	}
	d4top( OperPrv );		/* Start at the beginning of the database */
	curtag = d4tag( OperPrv, "OPERGROUP" );
	d4tag_select( OperPrv, curtag );
	rv = d4seek( OperPrv, SrchBuff );	/* Look for CurrOper+PtGroup using expression in OPERGROUP tag */
	d4unlock_all( OperPrv );						/* Release the file */

	if( rv )  {
		/* No record of this operator being associated with the point's group. */
		return PRIVNOGROUP;
	}

	/* Extract this operator's Privilege Level.  
	 * (We have the group already since it matched.)
	 */
	strcpy( OperPrvLevel, f4str( d4field( OperPrv, "PRIV_LEVEL" ))); /* Privilege Level */

	if( strcmp( OperPrvLevel, "00" ))  {
		return PRIVOK;
	}
	else  {
		/* Privilege level for operator is zero */
		return PRIVNOACCESS;
	}

}


/*********************************************************************************************/

int GetFieldLevel( PtType, FieldName, FieldPrvLevel )
char 		*PtType;
char 		*FieldName;
char 		*FieldPrvLevel;
{
	char	SrchBuff[100];
	T4TAG	*curtag;
	int	rv;


	memset( SrchBuff, 0, sizeof( SrchBuff ));
	sprintf( SrchBuff, "%-*.*s%-*.*s", POINTTYPE_SZ, POINTTYPE_SZ, PtType, 
					   FLDNAME_SZ, FLDNAME_SZ, FieldName );

	d4top( FieldPrv );		/* Start at the beginning of the database */
	curtag = d4tag( FieldPrv, "TYPEFIELD" );
	d4tag_select( FieldPrv, curtag );
	rv = d4seek( FieldPrv, SrchBuff ); /* Look for PtType+ProtGroup using expression in TYPEGROUP tag */
	d4unlock_all( FieldPrv );					/* Release the file */

	if( rv )  {
		strcpy( FieldPrvLevel, "01" );
	}
	else  {
		/* Extract this operator's Privilege Level.  (We have the group already since it matched.) */
		strcpy( FieldPrvLevel, f4str( d4field( FieldPrv, "PRIVLEVEL" )));	/* Privilege Level */
	}
	return PRIVOK;

}


/*********************************************************************************************/

void GetOperGroup()
{
	char	SrchBuff[OPERATOR_SZ];
	T4TAG	*curtag;
	int	rv;

	GetOperName (CurrOper); /* get current user in case login changed */
	memset( OperGroup, 0, sizeof OperGroup );
	if( !UserGrp )  {
		/* "sysusers.dbf" was inaccessible.  Use the CurrOper value. */
		return;
	}

	memset( SrchBuff, 0, sizeof( SrchBuff ));
	sprintf( SrchBuff, "%-*.*s", OPERATOR_SZ, OPERATOR_SZ, CurrOper );

	d4top( UserGrp );		/* Start at the beginning of the database */
	curtag = d4tag( UserGrp, "NAME" );
	d4tag_select( UserGrp, curtag );
	rv = d4seek( UserGrp, SrchBuff ); /* Look for Operator's group name using expression in NAME tag */
	d4unlock_all( UserGrp );					/* Release the file */

	if( rv )  {
	}
	else  {
		/* Extract this operator's group name. */
		strcpy( OperGroup, f4str( d4field( UserGrp, "OPTYPE" )));
	}
#ifdef	PRIVDIAG
{
char buff[128];
sprintf (buff, "GetOperGroup -%s-",OperGroup);
prvDiag ("GetOperGroup",buff);
}
#endif	PRIVDIAG
	return;

}


/*********************************************************************************************/

int add_record( TxtStr, Channel, OldValue, NewValue )
char 		*TxtStr;
char 		*Channel;
char 		*OldValue;
char 		*NewValue;
{
	char	FmtDate[11];
	char	TimeBuff[9];	/* Allow for NULL termination */
	unsigned int FldPos;

	/*
	 * See if new daily tracking file is needed.
	 */
	time( &CurTime );
	CurTmElements = *(localtime( &CurTime ));

#ifdef DAILY_FILE
	if( DayOfLastRec != CurTmElements.tm_yday )  {
	    d4unlock_all( TrakFile );
	    d4close( TrakFile );
	    DayOfLastRec = CurTmElements.tm_yday;
	    memset( TrakName, 0, sizeof( TrakName ));
#ifdef	WIN32
	    strftime( TrakName, sizeof( TrakName ), "\\tis4000db\\o%Y%j", &CurTmElements );
#else
	    strftime( TrakName, sizeof( TrakName ), "/tis4000db/o%Y%j", &CurTmElements );
#endif	WIN32

	    /* Tracking file must be created. */
	    cb.error_code = 0;		/* Clear out the open error */
	    fld_list = &TrakFlds[0];
	    tag_list = &TrakTag[0];
	    if( !( TrakFile = d4create( &cb, TrakName, fld_list, tag_list )))  {
		return PRIVNOCONN;		/* Can't connect to tracking file */
	    }
	    sleep(1);		/* Wait a second for file creation */

	    /* Set file protections for other users in the group */
	    memset( FilePath, 0, sizeof FilePath );
	    sprintf( FilePath, "%s.dbf", TrakName );
#ifdef WIN32
	    chmod( FilePath, 0666 );
	    sprintf( FilePath, "%s.mdx", TrakName );
	    chmod( FilePath, 0666 );
#endif WIN32
#if 0
	    fchmod( TrakFile->file.hand, 0766 );
	    if(( TrakIndex = d4index( TrakFile, TrakName )) == NULL )  {
		return PRIVNOCONN;
	    }
	    fchmod( TrakIndex->file.hand, 0666 );
	    fcntl( TrakFile->file.hand, F_SETFL, O_SYNC );

	    /* Do the same for the index file - after getting the file information. */
	    fcntl( TrakIndex->file.hand, F_SETFL, O_SYNC );
#endif

/**  This may be a bad thing for here	    d4unlock_all( TrakFile );
 **/
	}
#endif

	d4go_eof (TrakFile);
	d4append_start( TrakFile, 0 );	/* '0' => no memo fields */
	d4blank( TrakFile );

	f4assign_n( d4field( TrakFile, "OPERATOR" ), CurrOper, strlen( CurrOper ));

	/* Store date in MM/DD/CCYY format */

	memset( FmtDate, 0, sizeof( FmtDate ));
	strftime( FmtDate, sizeof( FmtDate ), "%m/%d/%y", &CurTmElements );
	f4assign_n( d4field( TrakFile, "TRK_DATE" ), FmtDate, strlen( FmtDate ));

	/* Store the time */

	memset( TimeBuff, 0, sizeof( TimeBuff ));
	strftime( TimeBuff, sizeof( TimeBuff ), "%H:%M:%S", &CurTmElements );
	f4assign_n( d4field( TrakFile, "TRK_TIME" ), TimeBuff, strlen( TimeBuff ));

	/* Store the descriptor */

	f4assign_n( d4field( TrakFile, "DESCRIPTOR" ), TxtStr, strlen( TxtStr ));

	/* Decompose the channel into name and field */

	memset( PtName, 0, sizeof( PtName ));
	memset( FldName, 0, sizeof( FldName ));
	FldPos = strcspn( Channel, "." );
	strncpy( PtName, Channel, FldPos );
	if( FldPos == strlen( Channel ))  {
		strcpy( FldName, "VAL" );
	}
	else  {
		strcpy( FldName, &Channel[FldPos+1] );
	}
	f4assign_n( d4field( TrakFile, "POINTNAME" ), PtName, strlen( PtName ));
	f4assign_n( d4field( TrakFile, "FIELDNAME" ), FldName, strlen( FldName ));

	f4assign_n( d4field( TrakFile, "OLD_VALUE" ), OldValue, strlen( OldValue ));
	f4assign_n( d4field( TrakFile, "NEW_VALUE" ), NewValue, strlen( NewValue ));
	f4assign_n( d4field( TrakFile, "PRINT_FLAG" ), " ", 1 );
	f4assign_long( d4field( TrakFile, "UNIX_TIME" ), CurTime );

	d4append( TrakFile );
	d4flush_all( TrakFile );
#ifndef CB51
	d4update_header( TrakFile );
#endif
	d4unlock_all( TrakFile );				/* Release the file */

	return PRIVOK;
}


/*********************************************************************************************/

int OperPriv( cBase, reqPriv )
C4CODE	*cBase;				/* Codebase status configuration block */
char	*reqPriv;			/* can be a protection group or a utility privilege */
{
	/* Get my privilege level within 'reqPriv' */
	int	rv;
	int	ItWasOpen = 0;	/* Used to determine whether or not to clean up when done. */

	/* If no codebase descriptor is given, use the initialized one */

	if (cBase == NULL)
	{
	    if (cbInit == 0)
		InitOperPriv (NULL);
	    cBase = &cb;
	}

	if( !OperPrv )  {
	    /* Find out who I am. */
	    GetOperName( CurrOper );

	    /* We need to initialize the Operator Privilege files. */
#ifdef WIN32
	    if( !( OperPrv = d4open( cBase, "\\tis4000db\\privtab" )))
	    {
#else
	    if( !( OperPrv = d4open( cBase, "/tis4000db/privtab" )))
	    {
#endif WIN32
		/* No privilege table is found. */
		return 0;
	    }
	    d4unlock_all( OperPrv );

#ifdef WIN32
	    UserGrp = d4open( cBase, "\\tis4000db\\sysusers" );
#else
	    UserGrp = d4open( cBase, "/tis4000db/sysusers" );
#endif WIN32
	    if( UserGrp )
	    {
		/* Unlock if it exists.  Otherwise, don't worry about it. */
		d4unlock_all( UserGrp );
	    }
	    if( !UtilAccDb )
	    {
	    /* We need to initialize the Utility Privilege files. */
#ifdef WIN32
		UtilAccDb = d4open( cBase, "\\tis4000db\\utilacc" );
#else
		UtilAccDb = d4open( cBase, "/tis4000db/utilacc" );
#endif WIN32
		if (UtilAccDb)
		{
		    /* unlock if exists, else dont worry, for backwards compatability */
		    d4unlock_all( UtilAccDb );
		}
	    }

	    /* Find out who I am. */
	    GetOperName( CurrOper );

	}
	else
	{
	    ItWasOpen = 1;
	}

	/* Look for user's name in 'sysusers.dbf' and extract the
	 * associated group name.
	 */
	GetOperGroup();

	/* if we fail to find the privilege in the utility access db (UtilAccDb)
	 * try to find it in privtab.db
	 */
	if ((rv = GetUtilPrivilege (reqPriv)) == 0)
	{
	    rv = GetProtGrpPrivilege (reqPriv);
	}


	if( !ItWasOpen )
	{
	    d4close(OperPrv);
	    if (UserGrp)
	    {
		d4close(UserGrp);
	    }
	    if (UtilAccDb)
	    {
		d4close(UtilAccDb);
	    }
	    OperPrv = 0;	/* DEK 3/17/94 Setup for next time */
	    UserGrp = 0;	/* DEK 3/17/94 Setup for next time */
	    UtilAccDb = 0;
	}
	ItWasOpen = 0;	/* Same reason we cleared OperPrv */
	return( rv );
}


/*********************************************************************************************/
/*  Utility privilege has form of:
 *	*MMI_AAA_BBB_CCC
 *  We look for each level:
 *	MMI
 *	if they have MMI priv, go no further
 *	MMI_AAA etc
*/

int GetUtilPrivilege (char *reqPriv)
{
    T4TAG	*curtag;
    char	seekBuff[GROUPNAME_SZ + UTILACCPRIV_SZ + 1];
    char	*token, *ptr;
    char	utilPriv [UTILACCPRIV_SZ + 1];
    char	workStr [UTILACCPRIV_SZ + 1];
    int		rv = 0;
    int		seekReturn;
    int		first = 1;
    
    /* if database wasn't opened, no use to go on */
    if (reqPriv [0] != '*')
    {
	return (rv);
    }
    if (UtilAccDb == 0)
    {
	/* an overide if the utility access database is not there */
	if ((ptr = getenv ("UTILITYACCESS")) != NULL)
	{
	    return (99);
	}
	return (rv);
    }

    memset( utilPriv, 0, sizeof( utilPriv ));
    memset( workStr, 0, sizeof( workStr ));

    /* utility access priv should have an * at the beginning */
    ptr = &reqPriv[1];

    /* save copy of requested privilege since strtok would destroy the original string */
    strncpy (utilPriv, ptr, sizeof (utilPriv));
    utilPriv [sizeof (utilPriv) - 1] = 0;
    token = strtok (utilPriv, "_");
    while (token != NULL)
    {
	if (!first)
	{
	    strcat (workStr, "_");
	}
	else
	{
	    first = 0;
	}
	strcat (workStr, token);
	memset (seekBuff, 0, sizeof (seekBuff));
	if( OperGroup[0] > ' ' )
	{
	    sprintf( seekBuff, "%-*.*s%-*.*s", GROUPNAME_SZ, GROUPNAME_SZ, OperGroup, 
					   UTILACCPRIV_SZ, UTILACCPRIV_SZ, workStr );
	}
	else
	{
	    sprintf( seekBuff, "%-*.*s%-*.*s", GROUPNAME_SZ, GROUPNAME_SZ, CurrOper, 
					   UTILACCPRIV_SZ, UTILACCPRIV_SZ, workStr );
	}
	seekBuff [sizeof (seekBuff) - 1] = 0;
	curtag = d4tag (UtilAccDb, "GROUPPRIV");
	d4tag_select (UtilAccDb, curtag);
	seekReturn = d4seek (UtilAccDb, seekBuff);
	/* privilege was found */
	if (seekReturn == 0)
	{
	    strcpy( OperPrvLevel, f4str( d4field( UtilAccDb, "PRIV_LEVEL" )));	/* Privilege Level */
	    d4unlock_all (UtilAccDb);
	    OperPrvLevel [sizeof (OperPrvLevel) - 1] = 0;
	    rv = atoi (OperPrvLevel);
	    break;
	}
	token = strtok (NULL, "_");
   }
   return (rv);
}


/*********************************************************************************************/

int GetProtGrpPrivilege (char *reqPriv)
{
    char	seekBuff[GROUPNAME_SZ + PROTECTIONGRP_SZ + 1];
    T4TAG	*curtag;
    int		rv = 0;
    int		seekReturn;
    
    memset( seekBuff, 0, sizeof( seekBuff ));

    if( OperGroup[0] > ' ' )
    {
	sprintf( seekBuff, "%-*.*s%-*.*s", GROUPNAME_SZ, GROUPNAME_SZ, OperGroup, 
					   PROTECTIONGRP_SZ, PROTECTIONGRP_SZ, reqPriv );
    }
    else
    {
	sprintf( seekBuff, "%-*.*s%-*.*s", GROUPNAME_SZ, GROUPNAME_SZ, CurrOper, 
					   PROTECTIONGRP_SZ, PROTECTIONGRP_SZ, reqPriv );
    }
    seekBuff [sizeof (seekBuff) - 1] = 0;
    d4top( OperPrv );		/* Start at the beginning of the database */
    curtag = d4tag( OperPrv, "OPERGROUP" );
    d4tag_select( OperPrv, curtag );
    /* Look for OperGroupr+PtGroup using expression in OPERGROUP tag */
    seekReturn = d4seek( OperPrv, seekBuff );
    d4unlock_all( OperPrv );		/* Release the file */

    if( seekReturn )
    {
	/* No record of this group being associated with the point's group. */
	return rv;
    }
    /* Extract this group's Privilege Level.  (We have the group already since it matched.) */
    strcpy( OperPrvLevel, f4str( d4field( OperPrv, "PRIV_LEVEL" )));	/* Privilege Level */
    rv = atoi (OperPrvLevel);
    return (rv);
}


/*********************************************************************************************/

int	CheckChannel( Channel )
char		*Channel;
{
#ifdef	NOT_RIGHT_NOW
	int	FldPos;
	F4FIELD	*curfield;
	T4TAG	*curtag;
	int	rv;			/* Return value */
	char	PtGrp[GROUPSIZE+1];
	char	PtType[TYPESIZE+1];	/* Point type place holder */
	char	FldLevel[3];
	char	caChan[CHANSIZE+1];
	chid	ChanID;


	/* Determine whether the operator
	 * has privilege to access Channel.
	 *
	 * RETURNS : TRUE	(Operator has sufficient privilege to access/modify channel.)
	 *	     FALSE 	(Current operator is not permitted to access/modify channel.)
	 */

	if( !PointDB )  {
	    /* We need to initialize the Operator Privilege files. */
	    InitOperPriv(NULL);
	}

	/* First step is to find the point in PVMAP.DBF, and determine the
	 * Protection Group to which it belongs.
	 */
	memset( PtName, 0, sizeof( PtName ));
	memset( PtType, 0, sizeof( PtType ));
	memset( PtGrp, 0, sizeof( PtGrp ));
	memset( FldName, 0, sizeof( FldName ));
	FldPos = strcspn( Channel, "." );
	strncpy( PtName, Channel, FldPos );

	if( FldPos == strlen( Channel ))  {
		strcpy( FldName, "VAL" );
	}
	else  {
		strcpy( FldName, &Channel[FldPos+1] );
	}

	/* Lookup PtName in PVMAP.DBF and return its PtType & PtGrp */
	if( rv = GetProtGrp( PtType, PtGrp )) {
		CloseOperTracking(CurrOper);
		return FALSE;
	}

	/* The point has been located & its protection group and point type extracted. 
	 * Now, look in Group Protection table to determine if
	 * there is a record for the current operator in the point's
	 * protection group.
	 */
	strcpy( ProtGroup, PtGrp );
	
	if(( rv = GetOperLevel()) != PRIVOK )  {
		CloseOperTracking(CurrOper);
		return FALSE;
	}

	/* Now we have the operator's protecton level within the Point Protection Group.
	 * Using the point type extracted from PVMAP.DBF, extract the FIELD's protection
	 * level from the field protection table, and check it against the operator's level
	 * within the Point Protection Group.
	 */
	rv = GetFieldLevel( PtType, FldName, FldLevel );

	/* We now have the following:
	 * 	OperPrvLevel :	The operator's privilege level within the protection group;
	 *	FldLevel     :	The protection level required to modify the field in question.
	 *
	 * Compare the two to see if the operator is permitted to modify the field in question.
	 */
	if(( strcmp( FldLevel, OperPrvLevel )) > 0 )  {
		CloseOperTracking(CurrOper);
		return FALSE;
	}

	/* See if the point has been tagged. */
	if( tagdVal > 0 )  {
		CloseOperTracking(CurrOper);
		return FALSE;
	}
#endif

	if( !PointDB )  {
	    /* We need to initialize the Operator Privilege files. */
	    InitOperPriv(NULL);
	}

	if((CheckOperPriv( Channel )) == PRIVOK )  {
	    CloseOperTracking(CurrOper);
	    return TRUE;
	}

	CloseOperTracking(CurrOper);
	return FALSE;

}


/**************************************************************************************
  SetChannel allows the user of this file to set the PtName for the add_record routine.
  otherwise the last point that permissions were checked for is assumed to be the current
  point.
**************************************************************************************/
void  SetChannel(Channel)

char  *Channel;

{
    strcpy(PtName, Channel);
}


/************************************************
*
*    Verify the password for the login person
*
************************************************/

int VerifyPassword (char *login, char *password)
{
    int		retVal = 1;
    char	seekBuff [OPERATOR_SZ + 1];
    char	encryptPasswd [ENCRYPTPASSWD_SZ + 1];
    char	foundEncryptPasswd [ENCRYPTPASSWD_SZ + 1];
    int		len;
    DATA4	*tisUsers;
    TAG4	*nameTag;
    FIELD4	*encryptPasswdFld;

    if (cbInit == 0)
    {
	InitOperPriv (NULL);
	/* dont need all the files opened */
	CloseOperPriv ();
    }
#ifdef	WIN32
    if ((tisUsers = d4open (&cb, "\\tis4000db\\tisusers")) == NULL)
    {
#else
    if ((tisUsers = d4open (&cb, "/tis4000db/tisusers")) == NULL)
    {
#endif	WIN32
	return (retVal);
    }
    d4unlock (tisUsers);
    
    nameTag = d4tag (tisUsers, "NAME");
    d4tagSelect (tisUsers, nameTag);
    d4top (tisUsers);
    sprintf (seekBuff, "%-*.*s", OPERATOR_SZ, OPERATOR_SZ, login);
    if ((d4seek (tisUsers, seekBuff)) != r4success)
    {
	d4close (tisUsers);
	return (retVal);
    }
    encryptPasswdFld = d4field (tisUsers, "ENCRPASSWD");
    strcpy (foundEncryptPasswd, f4str (encryptPasswdFld));
    foundEncryptPasswd [sizeof (foundEncryptPasswd) - 1] = 0;
    /* remove trailing spaces from database value */
    for (len = strlen (foundEncryptPasswd) - 1;
	 len >= 0;
	 len--)
    {
	if (isspace (foundEncryptPasswd [len]))
	{
	    foundEncryptPasswd [len] = 0;
	}
	else
	{
	    break;
	}
    }
    strcpy (encryptPasswd, password);
    tisEncrypt (encryptPasswd, UNENCRYPTPASSWD_SZ);
    if ((strcmp (encryptPasswd, foundEncryptPasswd)) == 0)
    {
	retVal = 0;
    }
    d4close (tisUsers);
    return (retVal);
}


