/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* Convert all links in a path to the actual effective pathname */

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#ifdef WIN32
#include <stdlib.h>
#endif WIN32

#include "resolveLinks.h"

#ifdef DEBUG_RESOLVE

main (int argc, char *argv[])

    {

	char	origPath [100], resPath [100];

	strcpy (origPath, argv [1]);
	resolveLinks (origPath, resPath);
	printf ("%s\n", resPath);

    }
#endif

/*----------*/
void resolveLinks (char *orig, char *resolved)

    {

	char	*cp, original [100];
#ifndef WIN32
	int	sc, n;
	struct stat statinfo;
	char	pathname [100], link [100];
#else
	char	*newcp;
#endif WIN32

	strcpy (original, orig);

	/* Check if the filename portion of the path is a link. If it is, resolve it. */
	/* If the resulting pathname begins with / then replace the original path */
	/* with the new one. Otherwise replace the filename portion of the original */
	/* path with the resolved path. Whew! */

#ifdef WIN32
	/* in WIN32, check for environment variable with same name as path, substsitute if found */
	/* pathname will have form '\string\...' or '/string/...' */
	if (*original == '\\'
	    && (cp = strchr (&original[1], '\\')) != NULL)
	{
	    *cp++ = 0;
	    newcp = getenv (original);
	    if (newcp)
		strcpy (resolved, newcp);
	    else
		strcpy (resolved, original);
	    strcat (resolved, "\\");
	    strcat (resolved, cp);
	}
	else if (*original == '/'
	    && (cp = strchr (&original[1], '/')) != NULL)
	{
	    *cp++ = 0;
	    newcp = getenv (&original[1]);
	    if (newcp)
		strcpy (resolved, newcp);
	    else /* substitute win/DOS path separators */
	    {
		strcpy (resolved, original);
		if (original[0] == '/')
		    resolved[0] = '\\';
	    }
	    strcat (resolved, "\\");
	    strcat (resolved, cp);
	}
	else
	    strcpy (resolved, orig);
#else
	if (lstat (original, &statinfo) == 0) {	
#ifdef MIPSEL
	    sc = statinfo.st_mode & S_IFLNK; /* st_mode is bit encoded */
#else
	    sc = S_ISLNK (statinfo.st_mode);
#endif
	    if (sc) {
		n = readlink (original, pathname, sizeof (original));
		pathname [n] = '\0';

		if (pathname [0] == '/') {
		    strcpy (original, pathname);
		} else {
		    cp = original + strlen (original) - 1;
		    while ((cp != original) && (*cp != '/'))
			cp--;
		    if (*cp == '/')
			*++cp = '\0';
		    else
			strcpy (original, "");
		    strcat (original, pathname);
	        }
		/* Fall through and resolve the full path */
	    }
	}

	/* Start with an empty path */

	strcpy (resolved, "");

	/* If the original path begins with a / initialize the working path to / */

	cp = original;
	if (*original == '/') {
	    strcpy (resolved, "/");
	    cp++;
	}

	/* Pick up each component of the path and append it to the working path. If the */
	/* resulting path refers to a link, resolve the link and place it at the end */
	/* of the working path. */

	cp = strtok (cp, "/");
	do {
	    strcpy (pathname, resolved);
	    strcat (pathname, cp);
	    if (lstat (pathname, &statinfo) == 0) {
#ifdef MIPSEL
		sc = statinfo.st_mode & S_IFLNK; /* st_mode is bit encoded */
#else
		sc = S_ISLNK (statinfo.st_mode);
#endif
		if (sc) {	/* Is a link */
		    n = readlink (pathname, link, sizeof (link));
		    link [n] = '\0';
		    if (*link == '/')
			strcpy (link, link + 1);  /* Strip initial / */
		    strcat (resolved, link);
		    strcat (resolved, "/");
		} else {
		    strcat (resolved, cp);
		    strcat (resolved, "/");
		}
	    } else { /* Doesn't exist, must be a file. */
		strcat (resolved, cp);
		strcat (resolved, "/");
		break;
	    }
	    cp = strtok (NULL, "/");
	} while (cp);

	/* An extra / ends up on the end, strip it off */

	*(resolved + strlen (resolved) - 1) = '\0';
#endif WIN32
    }

