/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include	<stdio.h>
#include	<signal.h>
#include	<time.h>
#ifndef SOLARIS
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/ioctl.h>
#else
#include	<sys/filio.h>
#endif
#include	<netinet/in.h>
#include	<netinet/tcp.h>
#include	<netdb.h>
#include 	<sys/ioctl.h>
#include	<sys/termios.h>
#include 	<fcntl.h>
#include 	<errno.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include	"error.h"
#include	"msmain.h"
#include	"msgMon.h"

/* Functions defined in this module */

static	void	intHandler (int);
static	void	pipeHandler (int);
static	int	findPartner (char *, char *);
static	void	setupListenFd (void);
static	void	setupBackupLink (void);
static	void	checkForConnect (void);
static	void	connectToPartner (void);	
static	void	processPartnerMessages (void);
static	void	processBackupMessages (void);
static	void	sendStatusToPartner (void);
static	void	checkForPartnerTimeout (void);
static	void	processAppMessages (void);
static	void	sendPartnerMessages (void);
static	void	sendBackupMessages (void);
static	void	sendAppMessages (void);
static	void	sendStateToApps (void);
static	void	sendCurrentStateToApp (int);
static	void	stripBlanks (char *);
static	int	comBufAllocate (ComBuf *, int, int);
static	int	comBufInsert (ComBuf *, char *, int);
static	void	comBufFlush (ComBuf *);
static	int	getMessage (ComBuf *, Msg *);
static	void	printStat (char *func, int value);

/* External functions */

extern	void	serverInit ();

/* Variables */

extern	int	errno;
extern	char	*sys_errlist [];

	int	myPid;
	int	debugLevel = 0;
static	int	heartRate = HEARTRATE;
static	int	partnerTimeout = PARTNER_TIMEOUT;
static	char	databaseDirectory [MAXPATHLENGTH];
static	char	partnerName [MAXHOSTNAMELENGTH+1], hostName [MAXHOSTNAMELENGTH+1];
static	unsigned int	partnerAddr, myAddr;
static	int	partnerFd = -1;
static	int	listenFd = -1;
static	int	connectInProgress = 0;
static	long	lastCommRecvTime;
static	WS_STATE	myState, partnerState;
static	struct timeval masterDelay;
static	AppNode	appInfo [MAX_APPS];
static	int	optval = 1;	
static	char	*msState [] = {"", "SLAVE", "MASTER", "MASTER/NP", "UNKNOWN"};
static	long	startTime, disconnectPending = -1;
static	int	numEntries;
static	int	backupLinkActive = 0;
static	char	backupLinkDevice [100];
static	int	backupLinkFd = -1;
static	int	running;

/* Partner/partner buffers */

static	ComBuf	partnerRxBuffer, partnerTxBuffer;
static	ComBuf	backupRxBuffer, backupTxBuffer;

/*----------*/
main (argc, argv)

    int		argc;				/* get database location */
    char	*argv[];

    {

	extern	char	*optarg;

	int	i, cmd, timeNow, lastTime;
	struct	hostent	*ent;

	serverInit ("msArbiter", "/tis4000/tasks/msArbiter.lock", "lock");

	/* Note start time for state arbitration */

	startTime = time (0);

	/* Get my host name and process id */

	myPid = getpid ();
	gethostname (hostName, sizeof (hostName));


	/* Initialize all applications to non existant */

	for (i=0 ; i<MAX_APPS; i++) {
	    appInfo [i].fd = -1;
	    appInfo [i].name[0] = '\0';
	}
	numEntries = FIRST_FD;

	/* Set up signal handlers to gracefully handle signals */

	signal (SIGINT, intHandler);		/* handle ^C from keyboard */
	signal (SIGPIPE, pipeHandler);		/* go here if write to bad pipe is attempted */

	/* Set the default location of the resource list (/tis4000db). */

	strcpy (databaseDirectory, DEFAULT_DIRECTORY);

	/* Pick up any arguments */

	while ((cmd = getopt (argc, argv, "svdp:r:n:b:")) != -1) {
	    switch (cmd) {
		case 's':
		    debugLevel = 1;
		    break;
		case 'v':
		    debugLevel = 2;
		    break;
		case 'd':
		    debugLevel = 3;
		    break;
		case 'p':
		    if(*optarg != '-') {
			strcpy (databaseDirectory, optarg);
			break;
		    } else
			tisSysLog (0,0, "%s: option requires an argument -- %c", 
				   argv[0], cmd);
		case 'r':
		    if(*optarg != '-') {
			sscanf(optarg, "%d", &heartRate);
			break;
		    } else
			fprintf(stderr, "%s: option requires an argument -- %c\n", argv[0], cmd);
		case 'n':
		    if(*optarg != '-') {
			sscanf(optarg, "%d", &partnerTimeout);
			break;
		    } else
			fprintf(stderr, "%s: option requires an argument -- %c\n", argv[0], cmd);
		case 'b':
		    backupLinkActive = 1;
		    strcpy (backupLinkDevice, optarg);
		    if (debugLevel >= 1)
			tisSysLog (0,0, "msArbiter(%d): Backup link active on %s",
				myPid, backupLinkDevice); 
		    break;
		case '?':
		    tisSysLog (0,0, "Usage is:");
		    tisSysLog (0,0, "\tmsArbiter -p pathname -r retry -n timeout -b device");
		    tisSysLog (0,0, "\tpathname - Location of database files (%s)", 
				DEFAULT_DIRECTORY);
		    tisSysLog (0,0, "\tretry ---- Retry interval (%d)", HEARTRATE);
		    tisSysLog (0,0, "\ttimeout -- Timeout before assuming master (%d)", 
				PARTNER_TIMEOUT);
		    tisSysLog (0,0, "\tdevice --- Backup link device");
		    exit (-1);
	    }
	}

	/* Find my partner in the resource list. If not found then no redundancy is possible. Come up as */
	/* a master with no partner. If a partner is found then determine the addresses of myself and */
	/* my partner. If either cannot be found */

	if (findPartner (hostName, partnerName)) {

	    if (ent = gethostbyname (hostName))
		myAddr = ntohl (*(unsigned int *)*ent->h_addr_list);
	    else
		tisSysLog (0,0, "msArbiter(%d): Unable to resolve address of %s - no redundancy",
				myPid, hostName);

	    if (ent = gethostbyname (partnerName))
		partnerAddr = ntohl (*(unsigned int *)*ent->h_addr_list);
	    else
		tisSysLog (0,0, "msArbiter(%d): Unable to resolve address of %s", 
				myPid, partnerName);

	} else {
	    tisSysLog (0,0, "msArbiter(%d): Unable to find partner in ioclist", myPid);
	    exit (-1);
	}

	/* Initialize buffers for partner/partner communications */

	comBufAllocate (&partnerTxBuffer, BUFFER_SIZE, -1);
	comBufAllocate (&partnerRxBuffer, BUFFER_SIZE, -1);
	if (backupLinkActive) {
	    comBufAllocate (&backupTxBuffer, BACKUP_BUFFER_SIZE, -1);
	    comBufAllocate (&backupRxBuffer, BACKUP_BUFFER_SIZE, -1);
	}

	/* Set up the listen socket for out partner and apps to connect to */

	setupListenFd ();

	/* This is the main processing loop. A select is performed on the listen socket and all */
	/* currently active listen sockets.  Also, we will try to connect to our */
	/* partner if not already connected. */

	/* We will initially be a master with no partner */

	myState = MASTER_NOPARTNER;
	partnerState = UNKNOWN;

	lastTime = 0;
	running = 1;
	while (running) {

	    fd_set	rdFdSet, wrFdSet;
	    int		maxFd, result;

	    /* Check if anyone is trying to connect to us */

	    checkForConnect ();

	    /* Setup up backuplink */

	    setupBackupLink ();

	    /* Process partner messages */

	    processPartnerMessages ();

	    /* Process messages over the backup link */

	    processBackupMessages ();

	    /* Process any pending application messages */

	    processAppMessages ();

	    /* Send a status update every heartRate seconds. Check if our partner has died. */

	    timeNow = time (0);
	    if ((timeNow - lastTime) >= heartRate) {
	        connectToPartner ();
		sendStatusToPartner ();
		checkForPartnerTimeout ();
		if (debugLevel >= 2)
		    tisSysLog (0,0, "msArbiter(%d): My state is %s, My partner state is %s",
				myPid, msState [myState], msState [partnerState]);
		lastTime = timeNow;
	    }

	    /* Send any messages that are queued up for our partner */

	    sendPartnerMessages ();

	    /* Send any messages that are queued up for our partner */

	    sendBackupMessages ();

	    /* Send any messages that are queued up for our apps */

	    sendAppMessages ();

	    /* Wait for any activity on our sockets */

	    masterDelay.tv_sec = heartRate;
	    masterDelay.tv_usec = 0;

	    /* Activate any sockets we are reading or writing */

	    FD_ZERO (&rdFdSet);
	    FD_ZERO (&wrFdSet);
	    maxFd = -1;

	    if ((partnerFd != -1) && !connectInProgress) {	/* Msgs from partner */
		FD_SET (partnerFd, &rdFdSet);
		if (partnerFd > maxFd)
		    maxFd = partnerFd;
	    }

	    if ((partnerFd != -1) && connectInProgress) {	/* Completed connection to partner */
		FD_SET (partnerFd, &wrFdSet);
		if (partnerFd > maxFd)
		    maxFd = partnerFd;
	    }

	    if ((backupLinkFd != -1) && backupLinkActive) {	/* Msgs received on backup link. */
		FD_SET (backupLinkFd, &rdFdSet);
		if (backupLinkFd > maxFd)
		    maxFd = backupLinkFd;
	    }

	    for (i=FIRST_FD ; i<MAX_APPS ; i++) {		/* Applications */
		if (appInfo [i].fd != -1) {
		    FD_SET (appInfo [i].fd, &rdFdSet);
		    if (appInfo [i].fd > maxFd)
			maxFd = appInfo [i].fd;
		}
	    }

	    /* Set at the select until any fd becomes ready */

	    result = select (maxFd+1, &rdFdSet, &wrFdSet, NULL, &masterDelay);
	    printStat ("Select", result);

	    if (result > 0) {	/* One or more fd's are ready */

		/* Check for read ready for our partner */

		int	bytesLeft, bytesRead;

		/* Connection to partner has completed. */

		if ((partnerFd != -1) && connectInProgress) {
		    if (FD_ISSET (partnerFd, &wrFdSet)) {
			connectInProgress = 0;
			lastCommRecvTime = time (0);
		    }
		}

		/* Incoming message from partner */

		if ((partnerFd != -1) && !connectInProgress) {
		    if (FD_ISSET (partnerFd, &rdFdSet)) {
			bytesLeft = partnerRxBuffer.bufSize - partnerRxBuffer.curSize;
			bytesRead = recv (partnerFd, &partnerRxBuffer.buffer [partnerRxBuffer.curSize], bytesLeft, 0); 
			printStat ("recv (partnerFd)", bytesRead);
			if (bytesRead > 0) {
			    partnerRxBuffer.curSize += bytesRead;
			    disconnectPending = -1;
			} else if ((errno != EWOULDBLOCK) && (errno != EINVAL)) {
			    if (errno != ECONNREFUSED)
				if (debugLevel >= 1) {
				    tisSysLog (0,0, "msArbiter(%d): Partner disconnected",
						myPid);
				}
			    close (partnerFd);
			    partnerFd = -1;
			    connectInProgress = 0;
			    comBufFlush (&partnerTxBuffer);
			    comBufFlush (&partnerRxBuffer);
			    if (disconnectPending == -1)
				disconnectPending = time (0);
			}
		    }
		}

		/* Incoming message from backup link */

		if ((backupLinkFd != -1) && backupLinkActive) {
		    if (FD_ISSET (backupLinkFd, &rdFdSet)) {
			bytesLeft = backupRxBuffer.bufSize - backupRxBuffer.curSize;
			bytesRead = read (backupLinkFd, &backupRxBuffer.buffer [backupRxBuffer.curSize], bytesLeft); 
			if (bytesRead > 0) {
			    backupRxBuffer.curSize += bytesRead;
			} else if ((errno != EWOULDBLOCK) && (errno != EINVAL)) {
			    if (errno != ECONNREFUSED)
				if (debugLevel >= 1) {
				    tisSysLog (0,0, "msArbiter(%d): Backup link failed.",
						myPid);
				}
			    close (backupLinkFd);
			    backupLinkFd = -1;
			    comBufFlush (&backupTxBuffer);
			    comBufFlush (&backupRxBuffer);
			}
		    }
		}

		/* Incoming message from application */

		for (i=FIRST_FD ; i<MAX_APPS ; i++) {
		    if (appInfo [i].fd == -1)
			continue;
		    if (FD_ISSET (appInfo [i].fd, &rdFdSet)) {
			bytesLeft = appInfo [i].txBuffer.bufSize - appInfo [i].txBuffer.curSize;
			bytesRead = recv (appInfo [i].fd, &appInfo [i].txBuffer.buffer [appInfo [i].txBuffer.curSize],
					bytesLeft, 0);
			if (debugLevel >= 2)
			    tisSysLog (0,0, "msArbiter(%d): Received %d bytes from ap on fd=%d, ix=%d",
					myPid, bytesRead, appInfo [i].fd, i);
			if (bytesRead > 0) {
			    appInfo [i].txBuffer.curSize += bytesRead;
			} else if (bytesRead == 0) {
			    if (debugLevel >= 1)
				tisSysLog (0,0, "msArbiter(%d): App %s disconnected",
						myPid, appInfo [i].name);
			    close (appInfo [i].fd);
			    appInfo [i].fd = -1;
			    strcpy (appInfo [i].name, "");
			} else if ((errno != EWOULDBLOCK) && (errno != EINVAL)) {
			    if (debugLevel >= 1)
				tisSysLog (0,0, "msArbiter(%d): App %s disconnected",
						myPid, appInfo [i].name);
			    close (appInfo [i].fd);
			    appInfo [i].fd = -1;
			    strcpy (appInfo [i].name, "");
			}
		    }
		}


	    }

	}

    }

/*----------*/
static void intHandler (sig)

    int 		sig;

    {

	if (debugLevel >= 1)
	    tisSysLog (0,0, "msArbiter(%d): INT received.", myPid);
	running = 0;

    }

/*----------*/
static void pipeHandler (sig)

    int 		sig;

    {

	if (debugLevel >= 1)
	    printf ("Broken pipe - ignored\n");

    }

/*----------*/
static int findPartner (host, part)

    char	*host, *part;

    {

	C4CODE	cb;
	F4FIELD	*ioc1Field, *ioc2Field;
	D4DATA	*data;
	char	path [MAXPATHLENGTH+1];
	char	ioc1Name [MAXHOSTNAMELENGTH+1], ioc2Name [MAXHOSTNAMELENGTH+1];

	d4init (&cb);

	sprintf (path, "%s/%s", databaseDirectory, RESLIST_FILE);
	data = d4open (&cb, path);
	if (data == NULL)
	    return 0;
	ioc1Field = d4field (data, FLD_IOC1);
	ioc2Field = d4field (data, FLD_IOC2);

	strcpy (part, "");
	for (d4top (data) ; !d4eof (data) ; d4skip (data, 1L)) {
	    strcpy (ioc1Name, f4str (ioc1Field));
	    stripBlanks (ioc1Name);
	    strcpy (ioc2Name, f4str (ioc2Field));
	    stripBlanks (ioc2Name);

	    if (strcmp (ioc1Name, host) == 0) {
		strcpy (part, ioc2Name);
		break;
	    }
	    if (strcmp (ioc2Name, host) == 0) {
		strcpy (part, ioc1Name);
		break;
	    }
	}

	d4close (data);
	return strlen (part) > 0;

    }

/*----------*/
static void setupListenFd (void)

    {

	int	len;
	struct	sockaddr_in	setAddr;

	/* Open up a listen socket so that our partner and apps can connect to us. */

	if((listenFd = socket (AF_INET,SOCK_STREAM,0)) < 0) {
	    perror ("Can't create socket");
	    exit (-1);
	}

	if ((setsockopt (listenFd, SOL_SOCKET ,SO_REUSEADDR, (caddr_t)&optval, sizeof(optval))) < 0) {
	    perror ("Can't set socket options");
	    exit (-1);
	}
	ioctl (listenFd, FIONBIO, &optval);

	memset (&setAddr, 0, sizeof(setAddr));
	setAddr.sin_family = AF_INET;
	setAddr.sin_addr.s_addr = htonl (INADDR_ANY);
	setAddr.sin_port = htons (LISTEN_PORT);
	len = sizeof (setAddr);
	while (bind (listenFd, (struct sockaddr *)&setAddr, sizeof(setAddr)) < 0) {
	    perror ("Can't bind socket...Retrying!");
	    sleep (5);
	}

	if (listen (listenFd, 10) < 0) {
	    perror ("Can't listen on socket...Exiting!");
	    return;
	}

    }

/*----------*/
static void setupBackupLink (void)	/* Open the backup link if one is specified */

    {

	int	true = 1;
	struct termios	t;

	if ((backupLinkFd != -1) || !backupLinkActive)
	    return;

	/* Try to open up the backup link device.  If it will not open then continue */
	/* to run but without the link. */

	backupLinkFd = open (backupLinkDevice, O_RDWR | O_NOCTTY);
	if (backupLinkFd == -1) {
	    tisSysLog (0,0, "msArbiter(%d): Backup link on %s would not open -- running without backup link.",
			myPid, backupLinkDevice);
	    backupLinkActive = 0;
	    return;
	}

	/* Set link to non-blocking */

	ioctl (backupLinkFd, FIONBIO, &true);

	/* Set up serial port parameters */

	memset ((char *)&t, 0, sizeof (t));
	t.c_iflag = IGNBRK;	/* Ignore break */
	t.c_cflag = B38400 + CS8 + CREAD + CLOCAL + CRTSCTS;
			/* 9600 baud, 8 data bits, receive enabled, local mode (no car detect) */
			/* enable RTS/CTS */
 	ioctl (backupLinkFd, TCSETS, &t);

	ioctl (backupLinkFd, TIOCEXCL, 0);	/* Exclusive use */

#ifndef AXP
	ioctl (backupLinkFd, TIOCGSOFTCAR, &true);
#endif

    }

/*----------*/
static void checkForConnect (void)

    {

	int	acceptFd;
	int	len, i;
	struct	sockaddr_in	setAddr;
	int	optval;

	len = sizeof (setAddr);
	if ((acceptFd = accept (listenFd, (struct sockaddr *)&setAddr, &len)) >= 0) {

	    printStat ("accept", acceptFd);

	    /* Set socket to non-blocking */

	    optval = 1;
	    ioctl (acceptFd, FIONBIO, &optval);

	    /* Check if our partner is trying to connect */

	    if (setAddr.sin_addr.s_addr == htonl (partnerAddr)) {
		if (debugLevel >= 1)
		    tisSysLog (0,0, "msArbiter(%d): Partner connected on fd=%d",
				myPid, acceptFd);
		partnerFd = acceptFd;
		lastCommRecvTime = time (0);
		optval = 50000;
		if ((setsockopt (partnerFd, SOL_SOCKET, SO_RCVBUF, (char *)&optval, sizeof(optval))) < 0)
		    perror ("msArbiter: Unable to increase partner TCP/IP receive buffer size.");
		if ((setsockopt (partnerFd, SOL_SOCKET ,SO_SNDBUF, (char *)&optval, sizeof(optval))) < 0)
		    perror ("msArbiter: Unable to increase partner TCP/IP transmit buffer size.");
		return;
	    }

	    /* A new application is registering. Find an empty entry in the table and insert the */
	    /* new application. */

	    for (i=FIRST_FD ; i<numEntries ; i++)
		if (appInfo [i].fd == -1)
		    break;

	    if (i == MAX_APPS) {
		tisSysLog (0,0, "msArbiter(%d): Maximum number of applications exceeded",
				myPid);
		close (acceptFd);
	    } else {
		appInfo [i].fd = acceptFd;
		comBufAllocate (&appInfo [i].txBuffer, BUFFER_SIZE, acceptFd);
		comBufAllocate (&appInfo [i].rxBuffer, BUFFER_SIZE, acceptFd);
		if (i == numEntries)
		    numEntries++;
	    }

	    if (debugLevel >= 1)
		tisSysLog (0,0, "msArbiter(%d): App connected on fd=%d", myPid, acceptFd);

	}

    }

/*----------*/
static void connectToPartner (void)

    {

	int	result, fd;
	struct	sockaddr_in	setAddr;
	int	optval;

	/* Only the host with the lower IP address will try to connect */

	if (myAddr > partnerAddr)
	    return;

	/* If a connection is in progress, don't try again */

	if (connectInProgress)
	    return;

	/* If not connected to our partner, try to connect.  These attempts will be */
	/* retried every "heartBeat"  seconds. */

	if (partnerFd == -1) {

	    if ((fd = socket (AF_INET, SOCK_STREAM, 0)) >= 0) {
		optval = 1;
		setsockopt (fd, IPPROTO_TCP, TCP_NODELAY, (caddr_t) &optval, sizeof (optval));
		ioctl (fd, FIONBIO, &optval);
		optval = 50000;
		if ((setsockopt (fd, SOL_SOCKET ,SO_RCVBUF, (char *)&optval, sizeof(optval))) < 0)
		    perror ("msArbiter: Unable to increase partner TCP/IP receive buffer size.");
		if ((setsockopt (fd, SOL_SOCKET ,SO_SNDBUF, (char *)&optval, sizeof(optval))) < 0)
		    perror ("msArbiter: Unable to increase partner TCP/IP transmit buffer size.");
		memset (&setAddr, 0, sizeof (setAddr));
		setAddr.sin_family = AF_INET;
		setAddr.sin_addr.s_addr = htonl (partnerAddr);
		setAddr.sin_port = htons (LISTEN_PORT);
		if ((result = connect (fd, (struct sockaddr *)&setAddr, sizeof (setAddr))) == -1) {
		    printStat ("connect", result);
		    if (errno == EINPROGRESS) {
			partnerFd = fd;
			lastCommRecvTime = time (0);
			connectInProgress = 1;
			if (debugLevel >= 1) {
			    tisSysLog (0,0, "msArbiter(%d): Attempting connection to %s, fd=%d",
					myPid, partnerName, fd); 
			}
		    }
		} else {
		    partnerFd = fd;
		    connectInProgress = 0;
		    lastCommRecvTime = time (0);
		    tisSysLog (0,0, "msArbiter(%d): Connected to %s, fd=%d",
						myPid, partnerName, fd);

		}
	    }
	}

    }

/*----------*/
static void processPartnerMessages (void)

    {

	Msg		msg;
	long		partnerUpTime;
	int		i, oldState;

	while (getMessage (&partnerRxBuffer, &msg)) {
	    switch (msg.msgHeader.type) {

		case PARTNER_STATE:	/* Message from my partner */

		    partnerState = ntohl (((StateMsg *)&msg)->state);
		    partnerUpTime = ntohl (((StateMsg *)&msg)->upTime);
		    lastCommRecvTime = time (0);

		    if (debugLevel >= 2)
			tisSysLog (0,0, "msArbiter(%d): State message received from partner -- %s",
					myPid, msState [partnerState]);

		    /* Resolve our state.  The host that has been up the longest is the master */

		    oldState = myState;
		    if ((time (0) - startTime) > partnerUpTime)
			myState = MASTER;
		    else
			myState = SLAVE;

		    if (myState != oldState)
			sendStateToApps ();

		    break;

		case APP_DATA:		/* Messages from app on partner */

		    if (debugLevel >= 2)
			tisSysLog (0,0, "msArbiter(%d): Message type %d received from partner app %s, %d bytes.",
					myPid, msg.msgHeader.type, ((AppDataMsg *)&msg)->name, 
					msg.msgHeader.size);

		    /* Find the entry for the destination app and place the message in its receive buffer */

		    for (i=FIRST_FD ; i<numEntries ; i++) {
			if (appInfo [i].fd != -1) {
			    if (strcmp (appInfo [i].name, ((AppDataMsg *)&msg)->name) == 0) {
				msg.msgHeader.magic = htonl (msg.msgHeader.magic);
				msg.msgHeader.size = htonl (msg.msgHeader.size);
				msg.msgHeader.type = htonl (msg.msgHeader.type);
				comBufInsert (&appInfo [i].rxBuffer, (char *)&msg, ntohl (msg.msgHeader.size));
				break;
			    }
			}
		    }

		    if (i == numEntries)
			if (debugLevel >= 2)
			    tisSysLog (0,0, "msArbiter(%d): App %s not registered, message discarded.",
					myPid, ((AppDataMsg *)&msg)->name);
		    break;

		default:

		    tisSysLog (0,0, "msArbiter(%d): Unknown message type from partner = %d",
				myPid, msg.msgHeader.type);
		    comBufFlush (&partnerRxBuffer);
		    break;
	    }
	}

    }

/*----------*/
static void processBackupMessages (void)

    {

	Msg		msg;
	long		partnerUpTime;
	int		oldState;

	while (getMessage (&backupRxBuffer, &msg)) { 
	    switch (msg.msgHeader.type) {

		case PARTNER_STATE:	/* Message from my partner */

		    partnerState = ntohl (((StateMsg *)&msg)->state);
		    partnerUpTime = ntohl (((StateMsg *)&msg)->upTime);
		    lastCommRecvTime = time (0);

		    if (debugLevel >= 2)
			tisSysLog (0,0, "msArbiter(%d): State message received from partner on backup link - %s",
					myPid, msState [partnerState]);

		    /* Resolve our state.  The host that has been up the longest is the master */

		    oldState = myState;
		    if ((time (0) - startTime) > partnerUpTime)
			myState = MASTER;
		    else
			myState = SLAVE;

		    if (myState != oldState)
			sendStateToApps ();

		    break;

		default:

		    tisSysLog (0,0, "msArbiter(%d): Unknown message type from partner on backup link = %d",
				myPid, msg.msgHeader.type);
		    comBufFlush (&backupRxBuffer);
		    break;
	    }
	}

    }

/*----------*/
static void sendStatusToPartner (void)

    {

	StateMsg	msg;

	msg.magic = htonl (MSG_MAGIC_NUMBER);
	msg.size = htonl (sizeof (StateMsg));
	msg.type = htonl (PARTNER_STATE);
	msg.state = htonl (myState);
	msg.upTime = htonl (time (0) - startTime);

	if ((partnerFd != -1) && !connectInProgress) {
	    comBufInsert (&partnerTxBuffer, (char *)&msg, ntohl (msg.size));
	}

	if ((backupLinkFd != -1) && backupLinkActive) {
	    comBufInsert (&backupTxBuffer, (char *)&msg, ntohl (msg.size));
	}

    }

/*----------*/
static void checkForPartnerTimeout (void)

    {

	/* If we have a disconnect in progress, see if it has been long enough to declare our */
	/* partner as lost. */

	if (disconnectPending != -1) {
	    if ((time (0) - disconnectPending) > DISCONNECT_PEND_TIME) {
		myState = MASTER_NOPARTNER;
		partnerState = UNKNOWN;
		sendStateToApps ();
	    }
	}

	/* Check if a long time has elapsed since we have heard from our partner. */

	if (partnerFd != -1) {
	    if ((time (0) - lastCommRecvTime) > partnerTimeout) {
		if (debugLevel >= 1)
		    tisSysLog (0,0, "msArbiter(%d): Partner timeout -- connection closed",
					myPid);
		close (partnerFd);
		partnerFd = -1;
		connectInProgress = 0;
		comBufFlush (&partnerTxBuffer);
		comBufFlush (&partnerRxBuffer);
		myState = MASTER_NOPARTNER;
		partnerState = UNKNOWN;
		sendStateToApps ();
	    }
	}

    }

/*----------*/
static void processAppMessages (void)

    {

	Msg		msg;
	int		i;

	for (i=FIRST_FD ; i<numEntries ; i++) {
	    if (appInfo [i].fd != -1) {
		while (getMessage (&appInfo [i].txBuffer, &msg)) {
		    switch (msg.msgHeader.type) {

			case APP_REGISTER:	/* Application is registering */

			    if (debugLevel >= 1)
				tisSysLog (0,0, "msArbiter(%d): App %s registered.",
						myPid, ((AppRegisterMsg *)&msg)->name);
			    strcpy (appInfo [i].name, ((AppRegisterMsg *)&msg)->name);

			    /* Send back the current state */

			    sendCurrentStateToApp (i);

			    break;

			case APP_DATA:		/* Send message from our app to partner */

			    if (debugLevel >= 2)
				tisSysLog (0,0, "msArbiter(%d): Message type %d received from local app %s, %d bytes.",
						myPid, msg.msgHeader.type, ((AppDataMsg *)&msg)->name, 
						msg.msgHeader.size);

			    /* Queue message to be sent to partner */

			    if ((partnerFd != -1) && !connectInProgress) {
				msg.msgHeader.magic = htonl (msg.msgHeader.magic);
				msg.msgHeader.size = htonl (msg.msgHeader.size);
				msg.msgHeader.type = htonl (msg.msgHeader.type);
				comBufInsert (&partnerTxBuffer, (char *)&msg, ntohl (msg.msgHeader.size));
			    } else if (debugLevel >= 2)
				tisSysLog (0,0, "msArbiter(%d): No partner connected, app message discarded.", myPid);

			    break;

			default:

			    tisSysLog (0,0, "msArbiter(%d): Unknown app message type = %d",
						myPid, msg.msgHeader.type);
			    comBufFlush (&appInfo [i].txBuffer);
			    break;
		    }
		}
	    }
	}

    }

/*----------*/
static int getMessage (buf, msg)  /* Get next complete message */

    ComBuf	*buf;
    Msg		*msg;

    {
	MsgHeader	*header;
	int		msgSize;
	uint32		magic;

	/* Check if enough bytes in the buffer to contain a message header */

	if (buf->curSize < sizeof (MsgHeader))
	    return 0;

	/* The beginning of the buffer contains a message header */

	header = (MsgHeader *)(buf->buffer);

	/* Check the magic number to see if this is a valid message */

	magic = ntohl (header->magic);
	if (magic != MSG_MAGIC_NUMBER) {
	    tisSysLog (0,0, "msArbiter(%d): Bad magic number in message - buffer flushed.",
			myPid);
	    comBufFlush (buf);
	    return 0;
	}

	/* Check if enough bytes in the buffer to form a complete message */

	msgSize = ntohl (header->size);
	if (msgSize > buf->curSize)
	    return 0;

	/* Extract the message from the buffer */

	memcpy (msg, buf->buffer, msgSize);

	/* Change the byte ordering of the header */

	msg->msgHeader.magic = ntohl (msg->msgHeader.magic);
	msg->msgHeader.size = ntohl (msg->msgHeader.size);
	msg->msgHeader.type = ntohl (msg->msgHeader.type);

	/* Squash out the message out of the buffer */

	memcpy (buf->buffer, buf->buffer + msgSize, buf->curSize - msgSize);
	buf->curSize -= msgSize;
	return 1;

    }

/*----------*/
static void sendPartnerMessages (void)	/* Send any queued messages for our partner */

    {

	if ((partnerFd == -1) || connectInProgress)
	    return;
 
	if (partnerTxBuffer.curSize > 0) {
	    int bytesSent;
	    bytesSent = send (partnerFd, partnerTxBuffer.buffer, partnerTxBuffer.curSize, 0);
	    printStat ("send", bytesSent);
	    if (bytesSent > 0) {
		memcpy (partnerTxBuffer.buffer, partnerTxBuffer.buffer + bytesSent,
				partnerTxBuffer.curSize - bytesSent);
		partnerTxBuffer.curSize -= bytesSent;
	    } else if (bytesSent == -1) {
		if ((errno != EWOULDBLOCK) && (errno != EINVAL)) {
		    if (debugLevel >= 1) {
			tisSysLog (0,0, "msArbiter(%d): Partner disconnected.", myPid);
		    }
		    close (partnerFd);
		    partnerFd = -1;
		    connectInProgress = 0;
		    comBufFlush (&partnerTxBuffer);
		    comBufFlush (&partnerRxBuffer);
		    if (disconnectPending == -1)
			disconnectPending = time (0);
		}
	    }
	}

    }

/*----------*/
static void sendBackupMessages (void)	/* Send any queued backup messages for our partner */

    {


	if (backupTxBuffer.curSize > 0) {
	    int bytesSent;
	    bytesSent = write (backupLinkFd, backupTxBuffer.buffer, backupTxBuffer.curSize);
	    if (bytesSent > 0) {
		memcpy (backupTxBuffer.buffer, backupTxBuffer.buffer + bytesSent,
				backupTxBuffer.curSize - bytesSent);
		backupTxBuffer.curSize -= bytesSent;
	    } else if (bytesSent == -1) {
		if ((errno != EWOULDBLOCK) && (errno != EINVAL)) {
		    if (debugLevel >= 1) {
			tisSysLog (0,0, "msArbiter(%d): Backup link disconnected", myPid);
		    }
		    close (backupLinkFd);
		    backupLinkFd = -1;
		    comBufFlush (&backupTxBuffer);
		    comBufFlush (&backupRxBuffer);
		}
	    }
	}

    }

/*----------*/
static void sendAppMessages (void)		/* Send any queued messages for our apps */

    {

	int	i;

	for (i=FIRST_FD ; i<numEntries ; i++) {
	    if ((appInfo [i].fd != -1) && (appInfo [i].rxBuffer.curSize != 0)) {
		int bytesSent;
		bytesSent = send (appInfo [i].fd, appInfo [i].rxBuffer.buffer, appInfo [i].rxBuffer.curSize, 0);
		if (bytesSent > 0) {
		    memcpy (appInfo [i].rxBuffer.buffer, appInfo [i].rxBuffer.buffer + bytesSent,
				appInfo [i].rxBuffer.curSize - bytesSent);
		    appInfo [i].rxBuffer.curSize -= bytesSent;
		} else if (bytesSent == -1) {
		    if ((errno != EWOULDBLOCK) && (errno != EINVAL)) {
			if (debugLevel >= 1)
			    tisSysLog (0,0, "msArbiter(%d): App %s disconnected.", myPid, appInfo [i].name);
			close (appInfo [i].fd);
			appInfo [i].fd = -1;
		    }
		}
	    }
	}

    }

/*----------*/
static void sendStateToApps (void)	/* Send current state to all connected apps */

    {
	int		i;

	for (i=FIRST_FD ; i<numEntries ; i++)
	    sendCurrentStateToApp (i);

    }

/*----------*/
static void sendCurrentStateToApp (i)

    int		i;

    {

	AppStateMsg	msg;

	msg.magic = htonl (MSG_MAGIC_NUMBER);
	msg.size = htonl (sizeof (AppStateMsg));
	msg.type = htonl (APP_STATE);
	msg.state = htonl (myState);

	if (appInfo [i].fd != -1) {
	    if (strlen (appInfo [i].name) != 0) {
		comBufInsert (&appInfo [i].rxBuffer, (char *)&msg, ntohl (msg.size));
	    }
	}

    }

/*----------*/
static void stripBlanks (buf)	/* Strip blanks from a string */

    char	buf [];

    {

	int	last;

	last = strlen (buf) - 1;
	while (last >= 0) {
	    if (buf [last] != ' ')
		break;
	    last--;
	}
	buf [last+1] = '\0';	/* enpty string */
				
    }

/*----------*/
static int comBufAllocate (buf, size, fd)

    ComBuf	*buf;
    int		size;
    int		fd;

    {

	if (buf->buffer != NULL)
	    free (buf->buffer);
	buf->buffer = malloc (size);
	buf->bufSize = size;
	buf->curSize = 0;
	buf->fd = fd;
	return buf->buffer != NULL;

    }

/*----------*/
static int comBufInsert (buf, data, size)

    ComBuf	*buf;
    char	*data;
    int		size;

    {

	if (size > (buf->bufSize - buf->curSize)) {
	    if (debugLevel >= 2)
		tisSysLog (0,0, "msArbiter(%d): Communication buffer overflow.", myPid);
	    return 0;
	}

	memcpy (&buf->buffer [buf->curSize], data, size);
	buf->curSize += size;
	return 1;

    }

/*----------*/
static void comBufFlush (buf)

    ComBuf	*buf;

    {

	buf->curSize = 0;

    }

/*----------*/
static void printStat (char *func, int value)

    {

/*	printf ("%s returns %d, errno = %d (%s)\n", func, value, errno, sys_errlist [errno]);
*/
    }

