/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include	<stdio.h>
#include	<memory.h>

#include	"msmain.h"
#include	"msgMon.h"

/* Functions defined in this module */

static	int	drainSocket (int appHandle, int microSecs);
	
/* External variables */

extern	int	appHostState;
extern	int	appMyHandle;

/* Local Variables */

static	char	localBuffer [BUFFER_SIZE];
static	int	localBufferIx = 0;	/* Next available slot and also current size */
static	Msg	msg;

/*----------*/
int appRecvData (appHandle, buffer, maxDelay)

    int		appHandle;
    char	*buffer;
    int		maxDelay;

    {

	int		msgSize;
	MsgHeader	*header;

	/* Always drain the socket from the arbiter */

	if (!drainSocket (appHandle, maxDelay))
	    return 0;

	/* Messages in the buffer can be either data messages or status change messages. */
	/* Process all status messages updating the current status until a data message */
	/* is found. */

	for (;;) {

	    /* Check if enough bytes in the buffer to form a complete message. */

	    if (localBufferIx >= sizeof (MsgHeader)) {	/* Enough for a message header */

		header = (MsgHeader *)localBuffer;
		if (ntohl (header->magic) != MSG_MAGIC_NUMBER) {
		    tisSysLog (0,0, "appRecvData(%d): Bad magic number in message from arbiter -- buffer flushed.",
				getpid ());
		    localBufferIx = 0;
		    return 0;
		}
		
		msgSize = ntohl (header->size);
		if (localBufferIx >= msgSize) {
		    memcpy (&msg, localBuffer, msgSize);
		    memcpy (localBuffer, &localBuffer [msgSize], localBufferIx - msgSize);
		    localBufferIx -= msgSize;
		} else {		/* Not enough to satisfy request */
		    return 0;
		}
	    } else
		return 0;

	    switch (ntohl (msg.msgHeader.type)) {

		case APP_DATA:		/* A data message */

		    memcpy (buffer, ((AppDataMsg *)&msg)->data,
				msgSize - sizeof (MsgHeader) - APP_NAME_LEN);
		    return 1;
		
		case APP_STATE:		/* A status update message */

		    appHostState = ntohl (((AppStateMsg *)&msg)->state);
		    break;

		default:

		    fprintf (stderr, "appRecvData(%d): Unknown message received, type=%d, buffer flushed\n",
				getpid (), ntohl (msg.msgHeader.type));
		    localBufferIx = 0;
	    }
	}

    }

/*----------*/
static int drainSocket (int appHandle, int microSecs)

    {

	struct timeval	timeOut;
	int		bytesRead, result, size;
	fd_set		readFds;

	/* Drain the socket of all pending data. */

	timeOut.tv_usec = microSecs % 1000000;
	timeOut.tv_sec = microSecs / 1000000;

	size = BUFFER_SIZE - localBufferIx;

	FD_ZERO (&readFds);
	FD_SET (appHandle, &readFds);

	result = select (appHandle + 1, &readFds, NULL, NULL, &timeOut);
	if (result < 0)
	    return 0;
	if (result == 0)
	   return 1;
	bytesRead = recv (appHandle, &localBuffer [localBufferIx], size, 0);
	if (bytesRead > 0)
	    localBufferIx += bytesRead;
	else if (bytesRead <= 0) {
	    close (appHandle);
	    appMyHandle = -1;
	    return 0;
	}
	return 1;

    }
