/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
   File:   getupdates.c - trend processor update fetch from
		 IOC historic trend function.

   Author: Grant Irani
   Date:   10/27/97
*/

#include <stdio.h>
#include <time.h>
#include <ctype.h>

#if !(defined(MIPSEL) || defined(AXP))
#include "memory.h"
#else
#include <sys/types.h>
#include <sys/stat.h>
extern int lstat ();
extern int readlink ();
#endif

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "trendSizes.h"
#include "trendupdate.h"
#include "msgmon.h"

#define LOG_OPTIONSI	(LOG_INFO | LOG_LOCAL1)
#define LOG_OPTIONSD	(LOG_DEBUG | LOG_LOCAL1)
#define LOG_OPTIONSE	(LOG_ERR | LOG_LOCAL1)

static	C4CODE	cb;		/* Codebase status configuration block */
static	D4DATA	*journal;	/* Codebase file descriptors */

int		myPid;


/* find and install update info for the current time */
static int GetUpdates ()
{
    time_t	recordTime;
    time_t	fileTime;
    struct tm	*tm_struct;
    char	fileName [13];		/* 8.3 file name size */
    char	journalName [100];	/* path + 8.3 file name size */
    int		fieldNum;
    char	pointName [POINTNAME_SIZE];
    short	dataIx, i;
    PointNodePtr p;

    recordTime = recordStartTimeA (300, time (0));

    fileTime = prevMidnite (recordTime);
    tm_struct = gmtime (&fileTime);

    /* Open journal file (return if none)*/
    sprintf (fileName, "TU%02.2d%02.2d%02.2d",
		tm_struct->tm_mon+1, tm_struct->tm_mday, tm_struct->tm_year % 100);
    sprintf (journalName, "%s%s", JOURNAL_PATH, fileName);
    /* Open the database.  If it does not exist then create it */
    resolveLinks (journalName, pathName);
    journal = d4open (&cb, pathName);
    if (journal == NULL)
    {
	return;
    }

    /* locate any records with current information */
    tm_struct = gmtime (&recordTime);
    d4tag_select (journal, d4tag (journal, "POINT_TIME"));
    d4top (journal);
    /* Get base value field */
    fieldNum = d4field_number (journal, "VALUE1");
    while (!d4eof (journal))
    {
	if ((f4int (d4field (data, "YEAR")) == tm_struct->tm_year + 1900)
	  && (f4int (d4field (data, "DAYOFYEAR") == tm_struct->tm_yday)
	  && (f4int (d4field (data, "HOUR") == tm_struct->tm_hour)
	  && (f4int (d4field (data, "OFFSET_MIN") == tm_struct->tm_min))
	{

	    /* this is a current record! Process any new info */
	    strcpy (pointName, f4str (d4field (journal, "POINTNAME");

	    /* Lookup the point in the tree. If found process sample trend */
	    p = findPoint (pointName);
	    if (p != NULL)
	    {
		if (p->trend [0] != NULL)
		{
		    for (dataIx = 0; dataIx < 60; dataIx++)
		    {
			updateStatus = *f4str (d4field_j (journal, fieldNum + (dataIx * 2) + 1));
			if (updateStatus != ' ')
			{
			    /* update into current record! */
			    t->value [dataIx] = f4double (d4field_j (journal, fieldNum + (dataIx * 2))) );
			    t->quality [dataIx] = '!';
			    t->severity [dataIx] = ' ';
			    t->status [dataIx] = ' ';
			}
	    	    }
	        }
	    }
	    else
	    {
	 	if (debugLevel >= 3)
        	    tisSysLog (0,0, "Sample for point not on trend -- %s discarded.\n",
									pointName);
	    }
	}
	d4skip (journal, 1);
    }
    d4close (journal);
}

