/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* Multiple Linked List Utilities */

/* This package supports objects which can simultaneously be */
/* linked onto multiple lists. */

#ifndef _MLIST_H_
#define _MLIST_H_

#include <stdio.h>

/* This is the basic set of forward and backward links for a single list. */
/* An array of these must appear at the beginning of the objects to */
/* appear on the lists. */

typedef struct _mnode {
    void        *next;
    void        *prev;
} MListLinks;

/* This is the list structure. One must be defined for each list. */

typedef struct _mlist {
    void        *front;
    void        *rear;
    int         listNumber;     /* Which link set for this list. */
    int         size;           /* Number of items on the list. */
    void        *current;       /* Iterator for this list. */
    short	reverse;	/* 0=>f>b, 1=>b>f */
} MList, *MListPtr;

/* Functions for manipulating lists */

extern  void    MListInit (MList *, int listNumber);
extern  void    MListInsertRear (MList *, void *);
extern  void    MListInsertFront (MList *, void *);
extern  void    MListRemove (MList *, void *);
extern	void	*MListRemoveFront (MList *);
extern	void	*MListRemoveRear (MList *);
extern  void    MListStartFront (MList *);
extern  void    MListStartRear (MList *);
extern  void    *MListNext (MList *);
extern  void    *MListPrev (MList *);
extern	int	MListSize (MList *);
extern	int	MListEmpty (MList *);
extern	void	*MListGetRear (MList *);
extern	void	*MListGetFront (MList *);


#endif

