/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <time.h>
#ifdef WIN32
#include <stdlib.h>
#include "NTUnixSubs.h"
#endif WIN32
#include "msgMon.h"

#include "trendDefs.h"
#include "trendProcess.h"
#include "dbRequests.h"
#include "rb.h"
#include "utils.h"

#define PRINT_TREND_PTR(x,y,z) if(y != 0){printTrendPtr(x,z);}
extern  int      debugLevel;
extern  int      myPid;
extern  RingBuf  *requestQueue;
extern  int      running;
extern  int      mxSem;
extern  UpToDateInfo *upToDateInfo;
extern  MList    trendList;

static  void     queueRequest (TrendPtr t);
static  void     fillToEnd (TrendPtr t);
static  void     updateUpToDate (TrendPtr t);
        void     resetUpToDatePointers (void);
        void     updateOnly (void);
	void	 printTrendPtr(TrendPtr t,char *str);

extern  void     initializeTrend (TrendPtr, time_t);
extern  PointNodePtr   findPoint (char *);
extern char debugName[40];
extern int flag1,flag2,flag3,flag4,eflag;
/*----------*/
void processStatusChange (TrendPtr t, int status)

    {

        time_t    timeNow, startTime;
        int       i, ix;

	if (!(t->active))
	    return;

        if (debugLevel >= 2)
            tisSysLog (0, 2, "Status change for %s %s, status = %d\n",
		    t->pointNode->pointName, int2methodString (t->method), status);

	/* Ignore status changing to up */

	if (status == 1) {
	    return;
	}

	/* Check to see if the record we currently have is the one that */
	/* contains a bucket for this sample. If not, write out the current*/
	/* record. */

	timeNow = time (0);
	startTime = recordStartTime (t, timeNow);
	if (t->startTime != startTime) {
	    writeTrend (t, timeNow);
        }

	/* We only need to process a lost resource. When the resource is regained the initial */
	/* report will indicate that the resource is again available. */
	/* Fill from the current bucket to now-1 with filled data. Place a connection error in */
	/* the current bucket. */

	ix = bucketIndex (t, timeNow);
	i = t->sampleCnt;
	if ((i < 0 || i > 60) || (ix < 0 || ix >= 60))
	    abort ();
	while (i<ix) {
	    t->value [i] = t->lastValue;
	    t->severity [i] = t->lastSeverity;
	    t->status [i] = t->lastStatus;
	    t->quality [i] = 'F';
	    i++;
	}
	t->quality [i] = ((t->count [i] > 0) ? 'Q' : 'C');
	t->resourceUp = FALSE;
	t->sampleCnt = ix + 1;			/* GCT 5-Aug-96 Update sample count here */
/**/
	PRINT_TREND_PTR(t,flag4,"After status change")
/**/

    }

/*----------*/
void processTrend (TrendPtr t, time_t ts, double value, int status, int severity)

    {

        time_t    startTime;
        long      secsIntoBucket;
	int       i, ix;

	if (!(t->active))
	    return;

        if (debugLevel >= 3)
            tisSysLog (0, 2, "Data for %s %s, t=%s, v=%lf, stat=%d, sevr=%d\n",
		    t->pointNode->pointName, int2methodString (t->method),
		    gmTime (ts), value, status, severity);

	/* Check to see if the record we currently have is the one that */
	/* contains a bucket for this sample. If not, write out the current*/
	/* record. */

	startTime = recordStartTime (t, ts);
	if (t->startTime != startTime) {
	    writeTrend (t, ts);
        }

	/* Get the index of the new sample */

	ix = bucketIndex (t, ts);

	/* Fill up to this bucket */

	i = t->sampleCnt;
	if ((i < 0 || i > 60) || (ix < 0 || ix >= 60))
	    abort ();
	while (i<ix) {
	    t->value [i] = t->lastValue;
	    t->severity [i] = t->lastSeverity;
	    t->status [i] = t->lastStatus;
	    t->quality [i] = t->resourceUp ? 'F' : 'C';
	    i++;
	}

	/* Integrate the new sample into the bucket. If severity of this sample is */
	/* greater than the current severity then update the severity and status. */

	if (severity > t->severity [ix]) {
	    t->severity [ix] = severity;
	    t->status [ix] = status;
	}

	/* Use the appropriate reduction method to integrate the new sample */

	switch (t->method) {
	    case REDUC_MIN:
	        if (t->count [ix] == 0)
		    t->value [ix] = value;
		else if (value < t->value [ix])
		    t->value [ix] = value;
	        break;
	    case REDUC_MAX:
	        if (t->count [ix] == 0)
		    t->value [ix] = value;
		else if (value > t->value [ix])
		    t->value [ix] = value;
	        break;
	    case REDUC_SMP:
	        if (t->count [ix] == 0)
		    t->value [ix] = value;
	        break;
	    case REDUC_AVG:
		secsIntoBucket = secondsIntoBucket (t, ts);
		if (secsIntoBucket == 0) {
		        t->value [ix] = value;
			break;
		}
	        if (t->count [ix] == 0) {
		    t->value [ix] = t->lastValue;
		} else {
		    /* remove excess contribution for previous value */
		    /* by calculating the average for this far into bucket */
		    t->value [ix] = ((t->value [ix] * t->interval) -
				(t->lastValue * (t->interval - secsIntoBucket))) /
				secsIntoBucket;
		}
		/* recalculate average for total interval */
		t->value [ix] = ((t->value [ix] * secsIntoBucket) +
		          (value * (t->interval - secsIntoBucket))) /
			  t->interval;
	        break;
	}

	t->count [ix]++;

	/* If the quality of this bucket is 'C' then this sample represent a reconnection. */
	/* If the quality is 'F' then this bucket was filled as a result of the periodic */
	/* database write. Replace it with a ' ' to indicate good data. */

	if (tolower (t->quality [ix]) != 'q') {
	    if ((t->quality [ix] == 'C') || (t->resourceUp == FALSE))
	        t->quality [ix] = 'q';
	    else if (t->quality [ix] == 'F')
	        t->quality [ix] = ' ';
	}

	/* Update all the 'last...' values for future filling. Also update the bucket count. */
 
	t->lastValue = value;
        t->lastSeverity = severity;
        t->lastStatus = status;
	t->lastTime = ts;
	t->resourceUp = TRUE;
	t->sampleCnt = ix + 1;

	/* If we are maintaining up-to-date data for this trend, update */
	/* the up-to-date values. */

 	updateUpToDate (t);

    }

/*----------*/
void updateUpToDate (TrendPtr t)

    {

        CurrentData    *c;
	int            i;

	sem_wait (mxSem);
        c = t->current;
        if (c) {
            if (debugLevel >= 2)
	        tisSysLog (0, 2, "Updating up-to-date for %s-%s\n",
		        t->pointNode->pointName,
		        int2methodString (t->method));

	    c->interval = t->interval;
	    c->startTime = t->startTime;
	    c->sampleCnt = t->sampleCnt;
	    c->lastValue = t->lastValue;
	    c->lastSeverity = t->lastSeverity;
	    c->lastStatus = t->lastStatus;
	    for (i=0 ; i<t->sampleCnt ; i++) {
	        c->value [i] = t->value [i];
		c->quality [i] = t->quality [i];
		c->severity [i] = t->severity [i];
		c->status [i] = t->status [i];
	    }
	    c->valid = TRUE;
	}

	sem_signal (mxSem);

    }

/*----------*/
void resetUpToDatePointers (void)

    /* If the up-to-date shared memory has been modified then scan */
    /* through all entries and reset all the pointers into the */
    /* up-to-date buffers. */

{
  PointNodePtr    p;
  TrendPtr        t;
  int             i;
  if (upToDateInfo->changed)
    {
      MListStartFront (&trendList);
      while (t = (TrendPtr) MListNext (&trendList))
	t->current = NULL;
      for (i=0 ; i<upToDateInfo->numEntries ; i++)
	{
	  /*
	    GCT 8-Aug-1996
	    Use the pointer stored in the update data structure unless this is the first time
	    this is being used. If so, then get it like before with findPoint
	    */
	  p = (PointNodePtr)upToDateInfo->data [i].p;
	  if(p == NULL)
	    p = (PointNodePtr)findPoint(upToDateInfo->data [i].pointName);
	  if(p)
	    {
	      upToDateInfo->data [i].p = (void *)p;
	      if (t = p->trend [upToDateInfo->data [i].method])
		{
		  if (debugLevel >= 2)
		    tisSysLog (0, 2, "Maintaining up-to-date data for %s-%s\n",
			    upToDateInfo->data [i].pointName,
			    int2methodString (upToDateInfo->data [i].method));
		  t->current = &(upToDateInfo->data [i]);
		  updateUpToDate(t);
		}
	    }
	}
      upToDateInfo->changed = FALSE;
    }
}

void updateOnly (void)

    /* Scan through all entries and update all the */
    /* up-to-date buffers. */

{
  PointNodePtr    p;
  TrendPtr        t;
  int             i;
  for (i=0 ; i<upToDateInfo->numEntries ; i++)
    {
      /*
	GCT 8-Aug-1996
	Use the pointer stored in the update data structure unless this is the first time
	this is being used. Then get it like before with findPoint
	*/
      p = (PointNodePtr)upToDateInfo->data [i].p;
      if(p == NULL)
	p = (PointNodePtr)findPoint(upToDateInfo->data [i].pointName);
      if(p)
	{
	  upToDateInfo->data [i].p = (void *)p;
	  if (t = p->trend [upToDateInfo->data [i].method])
	    {
	      if (debugLevel >= 2)
		tisSysLog (0, 2, "Periodic update for %s-%s\n",
			upToDateInfo->data [i].pointName,
			int2methodString (upToDateInfo->data [i].method));
	      updateUpToDate(t);
	    }
	}
    }
}

/*----------*/
void writeTrend (TrendPtr t, time_t timeNow)

    {

        int     i, ix,saveCnt;
	int	recSpan;
	time_t  startTime, nextTime,saveTime;

	/* Perform any necessary filling. If the last bucket contains data then fill using it. */
	/* If the last bucket contains a connection error, fill to the current bucket with */
	/* connection error. */

	startTime = recordStartTime (t, timeNow);
	saveTime = t->startTime;
	if ((t->interval*BUCKETS_PER_RECORD) >= (24*3600))
	    recSpan = (24*3600);
	else
	    recSpan = (t->interval*BUCKETS_PER_RECORD);
	nextTime = t->startTime + recSpan;  /* desired start time for next record */
	nextTime = recordStartTime(t,nextTime);
	saveCnt = t->sampleCnt;
	if(startTime > t->startTime)
	  {  /* Current time is after this record */
	    /* Fill this record to the end with filled values and queue it for writing. */
	      while(nextTime <= startTime)
		{
/**/
		  PRINT_TREND_PTR(t,flag1,"Before Write, fillToEnd & initializeTrend")
/**/
		  fillToEnd (t);
		  queueRequest (t);
		  initializeTrend (t, nextTime);   /* Initialize a new record with filled values up to the next startTime. */
/**/
		  PRINT_TREND_PTR(t,flag2,"After initializeTrend")
/**/
		  nextTime = t->startTime + recSpan;  /* desired start time for next record */
		  nextTime = recordStartTime(t,nextTime);
		  updateUpToDate(t);
		}
	    ix = bucketIndex (t, timeNow);
	    if (ix < 0 || ix >= 60)
		abort ();
	    for (i=0 ; i<=ix ; i++)
	      {
	        t->value [i] = t->lastValue;
		t->quality [i] = t->resourceUp ? 'F' : 'C';
		t->severity [i] = t->lastSeverity;
		t->status [i] = t->lastStatus;
	      }
	    t->sampleCnt = ix + 1;
	  }

	/* Fill from latest bucket to the present bucket */

	ix = bucketIndex (t, timeNow);
	i = t->sampleCnt;
	if ((i < 0 || i > 60) || (ix < 0 || ix >= 60))
	    abort ();
	while (i<=ix) {
	    if (t->quality [i] != '!')
	    {
		t->quality [i] = t->resourceUp ? 'F' : 'C';
		t->value [i] = t->lastValue;
		t->severity [i] = t->lastSeverity;
		t->status [i] = t->lastStatus;
	    }
	    i++;
	}
	t->sampleCnt = ix + 1;
/**/
	PRINT_TREND_PTR(t,flag3,"After Write")
/**/
	  /* Place the request in the queue */
	    queueRequest (t);

    }

/*----------*/
static void queueRequest (TrendPtr t)

    {

	WriteRequest	request;
        int     i;

	stringCopy (request.pointName, t->pointNode->pointName, POINTNAME_SIZE);
	request.method = t->method;
	request.interval = t->interval;
	strcpy (request.eguTag, t->pointNode->eguTag);
	request.dispLow = t->pointNode->dispLow;
	request.dispHigh = t->pointNode->dispHigh;

	request.startTime = t->startTime;
	request.sampleCnt = t->sampleCnt;
	for (i=0 ; i<request.sampleCnt ; i++) {
	    request.value [i] = t->value [i];
	    request.quality [i] = t->quality [i];
	    request.status [i] = t->status [i];
	    request.severity [i] = t->severity [i];
	}

	/* If the queue is currently full, wait and retry. */

	while (!rbInsert (requestQueue, &request)) {
	    if (debugLevel >= 2)
		tisSysLog (0, 2, "trendServer queue full -- waiting\n");
#ifdef WIN32
	    Sleep (1000);
#else
	    sleep (1);
#endif WIN32
	    if (!running)
		return;
	}

	if (debugLevel >= 2)
	    tisSysLog (0, 2, "Write request queued: %s-%s-%d\n",
		    t->pointNode->pointName,
		    int2methodString (t->method),
		    t->interval);


    }

/*----------*/
static void fillToEnd (TrendPtr t)

    {

        int    numBucks, i;

	numBucks = numBuckets (t, t->startTime);
	for (i=t->sampleCnt ; i<numBucks ; i++) {
	    if (i < 0 || i >= 60)
		abort ();
	    t->value [i] = t->lastValue;
	    t->quality [i] = t->resourceUp ? 'F' : 'C';
	    t->severity [i] = t->lastSeverity;
	    t->status [i] = t->lastStatus;
	}
        t->sampleCnt = numBucks;
}

void printTrendPtr(TrendPtr t,char *str)
{
  char timeStr[30],i,j,k;
  time_t now;
   if(debugName[0])
     {
       if(!strcmp(debugName,t->pointNode->pointName) || !strcmp(debugName,"DUMP_ALL"))
 	{
	  now = time(0);
	  strcpy(timeStr,ctime(&now));
	  printf("TrendPtr data for %s...Reason = %s...@%s",t->pointNode->pointName,str,timeStr);
	  printf("Method %s\n",int2methodString(t->method));
	  strcpy(timeStr,ctime(&t->startTime));
	  printf("Start time %s",timeStr);
	  printf("Sample count %d\n",t->sampleCnt);
	  printf("Resource state %s\n",t->resourceUp ? "UP":"DOWN");
	  strcpy(timeStr,ctime(&t->lastTime));
	  printf("Last bucket value=%f, severity=%c, status=%c, time=%s",
		 t->lastValue,
		 severity2letter(t->lastSeverity),
		 status2letter(t->lastStatus),
		 timeStr);
	  printf("Request que pointer = %08X,upToDateInfo pointer = %08X\n",requestQueue,upToDateInfo);
	  for(i=0;i<BUCKETS_PER_RECORD/4;i++)
	    {
	      for(j=0;j<4;j++)
		{
		  k = i*4+j;
		  printf("(#%02d,%f,\"%c\",\"%c\",\"%c\")",
			 k,t->value[k],
			 severity2letter(t->severity[k]),
			 status2letter(t->status[k]),
			 t->quality[k]);
		}
	      printf("\n");
	    }
 	}
     }
}
