/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* General purpose utilities */

#include <sys/types.h>
#ifndef WIN32
#include <sys/time.h>
#endif WIN32
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <time.h>
#include "assert.h"
#include "msgMon.h"
#include "tsDefs.h"
#include "alarm.h"

#include "mlist.h"
#include "trendDefs.h"

/* Functions defined in this module */

void	stringCopy (char *, char *, int);
void    stripBlanks (char *);
char    *strError (int);
time_t  prevMidnite (time_t);
time_t  recordStartTime (TrendPtr, time_t);
int     bucketIndex (TrendPtr, time_t);
int     numBuckets (TrendPtr, time_t);
int     bucketStartTime (TrendPtr, int);
int     secondsIntoBucket (TrendPtr, time_t);
int     getRecordsPerDay (int);
char    *gmTime (time_t);
int     letter2status (int);
int     status2letter (int);
int     letter2severity (int);
int     severity2letter (int);

/* Mappings for statuses and severity */

static struct statStruct{
    int		stat;
    int		letter;
} statLetters [] = {
		{NO_ALARM,		' '},	/*  0 */
		{READ_ALARM,		'i'},	/*  1 */
		{WRITE_ALARM,		'i'},	/*  2 */
		{HIHI_ALARM,		'H'},	/*  3 */
		{HIGH_ALARM,		'h'},	/*  4 */
		{LOLO_ALARM,		'L'},	/*  5 */
		{LOW_ALARM,		'l'},	/*  6 */
		{SIGNAL_ALARM,		's'},	/*  7 */
		{ROC_ALARM,		'R'},	/*  8 */
		{STATE_ALARM,		'S'},	/*  9 */
		{COS_ALARM,		'C'},	/* 10 */
		{OPENTC_ALARM,		'i'},	/* 11 */
		{TIMEOUT_ALARM,		'i'},	/* 12 */
		{LIMIT_ALARM,		'i'},	/* 13 */
		{CALC_ALARM,		'c'},	/* 14 */
	    	{SCAN_ALARM,		'N'},	/* 15 */
		{LINK_ALARM,		'K'},	/* 16 */
		{SOFT_ALARM,		'f'},	/* 17 */
		{BAD_UDB_ALARM,		'u'},	/* 18 */
		{UDF_ALARM,		'u'},	/* 19 */
		{INHIBIT_ALARM,		'I'},	/* 20 */
		{TAGGED_ALARM,		'T'},	/* 21 */
		{CMD_NOT_ALLOWED,	'a'},	/* 22 */
		{VOLUME,		'V'},	/* 23 */
		{HIHI_LEVEL,		'1'},	/* 24 */
		{TEMPERATURE,		't'},	/* 25 */
		{PRESSURE,		'r'},	/* 26 */
		{BSW,			'B'},	/* 27 */
		{TANK_VALVE,		'v'},	/* 28 */
		{VALVE_OPEN_ALM,	'O'},	/* 29 */
		{VALVE_CLOSE_ALM,	'C'},	/* 30 */
		{VALVE_TORQUE_LAG,	'R'},	/* 31 */
		{VALVE_FAULT_ALM,	'F'},	/* 32 */
		{VALVE_POSITION,	'P'},	/* 33 */
		{AUTOMATIC,		'A'},	/* 34 */
		{VIBRATION,		'v'},	/* 35 */
		{PUMP_START,		'R'},	/* 36 */
		{PUMP_STOP,		'P'},	/* 37 */
		{POT_DROP_READY,	'd'},	/* 38 */
		{PERMISSIVE_ALM,	'P'},	/* 39 */
		{TF_UNCOMMANDED,	'U'},	/* 40 */
		{THRESHOLD_EXCEEDED,	'x'},	/* 41 */
		{PID_ALM_NOMV,		'm'},	/* 42 */
		{PID_ALM_NOOUT,		'o'},	/* 43 */
		{PID_ALM_DEV,		'D'},	/* 44 */
		{PID_ALM_TRACKING,	't'},	/* 45 */
		{PID_ALM_HIGH_STP_CLAMP,'P'},	/* 46 */
		{PID_ALM_LOW_STP_CLAMP,	'P'},	/* 47 */
		{PID_ALM_HIGH_OUT_CLAMP,'O'},	/* 48 */
		{PID_ALM_LOW_OUT_CLAMP,	'O'},	/* 49 */
		{MANUAL_OVERRIDE,	'M'},	/* 50 */
		{SETPT_ERROR,		'p'},	/* 51 */
		{OUT_MODE_ERROR,	'o'},	/* 52 */
		{PARAMETER,		'f'},	/* 53 */
		{HIGH_LEVEL,		'2'},	/* 54 */
		{LOW_LEVEL,		'3'},	/* 55 */
		{LOLO_LEVEL,		'4'},	/* 56 */
		{HIHI_TEMPERATURE,	'5'},	/* 57 */
		{HIGH_TEMPERATURE,	'6'},	/* 58 */
		{LOW_TEMPERATURE,	'7'},	/* 59 */
		{LOLO_TEMPERATURE,	'8'},	/* 60 */
		{NO_INIT_ALARM,		'n'}	/* 61 */
};

static struct severityStruct{
    int		sev;
    int		letter;
} sevLetters [] = {
	    {EVENT_ALARM,	'E'},
	    {INFO_ALARM,	'I'},
	    {MAJOR_ALARM,	'M'},
	    {MINOR_ALARM,	'm'},
	    {VALID_ALARM,	'V'}
};

static  char   *methodStrings [] = {"MIN", "MAX", "AVG", "SMP"};

extern  int    myPid;

/*----------*/
void stringCopy (dest, src, n)

    char	*dest, *src;
    int		n;

	/* String copy with length protection and NULL termination. */

    {

	strncpy (dest, src, n - 1);
	*(dest + n - 1) = '\0';

    }

/*----------*/
void stripBlanks (char *buf)	/* Strip blanks from a string */

    {

	while (*buf) {
	    if (*buf == ' ') {
		*buf = '\0';
		return;
	    }
	    buf++;
	}
	
    }

/*----------*/
char *strError (int errorNum)

    {

#if defined(SOLARIS) || defined (AXP)
	return strerror (errorNum);
#else
	extern char *sys_errlist [];
	return sys_errlist [errorNum];
#endif

    }

/*----------*/
time_t prevMidnite (time_t t)

    {

	struct tm	*tmTime;
	int		secondsPastMidnight;
	time_t          tt = t;

	/* Convert the time to a "struct tm" */

	tmTime = gmtime (&tt);

	/* Convert the hour, minute, and seconds to seconds and subtract */
	/* from the timestamp.  This gives us midnight in seconds relative */
	/* to Unix epoch. */

	secondsPastMidnight = tmTime->tm_hour * 3600 +
			      tmTime->tm_min * 60 +
			      tmTime->tm_sec;

	/* Subtract from sample time */

	return (time_t)t - (time_t)secondsPastMidnight;

    }

/*----------*/
time_t recordStartTime (TrendPtr t, time_t sampleTime)

    {

        time_t  midnite, startTime;
        int     recordNumber;

	midnite = prevMidnite (sampleTime);
	recordNumber = (sampleTime - midnite) / (t->interval * BUCKETS_PER_RECORD);
	startTime = midnite + (recordNumber * t->interval * BUCKETS_PER_RECORD);
	return startTime;
	
    }

/*----------*/
time_t recordStartTimeA (int interval, time_t sampleTime)

    {

        time_t  midnite, startTime;
        int     recordNumber;

	midnite = prevMidnite (sampleTime);
	recordNumber = (sampleTime - midnite) / (interval * BUCKETS_PER_RECORD);
	startTime = midnite + (recordNumber * interval * BUCKETS_PER_RECORD);
	return startTime;
	
    }

/*----------*/
int bucketIndex (TrendPtr t, time_t sampleTime)
 
    {

        return (sampleTime - t->startTime) / t->interval;
    
    }

/*----------*/
int bucketIndexA (time_t startTime, int interval, time_t sampleTime)
 
    {

        return (sampleTime - startTime) / interval;
    
    }

/*----------*/
int bucketStartTime (TrendPtr t, int bucketNumber)

    {

       return t->startTime + (bucketNumber * t->interval);

    }

/*----------*/
int bucketStartTimeA (time_t startTime, int interval, int bucketNumber)

    {

       return startTime + (bucketNumber * interval);

    }

/*----------*/
int secondsIntoBucket (TrendPtr t, time_t sampleTime)

    {

        int     secondsIntoRecord, bucketNumber;
	time_t  bucketStartTime;

        secondsIntoRecord = sampleTime - t->startTime;
        bucketNumber = secondsIntoRecord / t->interval;
	bucketStartTime = t->startTime + (t->interval * bucketNumber);
	return sampleTime - bucketStartTime;

    }

/*----------*/
int secondsIntoBucketA (time_t startTime, int interval, time_t sampleTime)

    {

        int     secondsIntoRecord, bucketNumber;
	time_t  bucketStartTime;

        secondsIntoRecord = sampleTime - startTime;
        bucketNumber = secondsIntoRecord / interval;
	bucketStartTime = startTime + (interval * bucketNumber);
	return sampleTime - bucketStartTime;

    }

/*----------*/
int numBuckets (TrendPtr t, time_t sampleTime)

    {

        time_t    endTime, midNite;
        int       bucketsLeft;
 
	midNite = prevMidnite (t->startTime);
        endTime = t->startTime + (t->interval * BUCKETS_PER_RECORD) - 1;
	if (midNite != prevMidnite (endTime)) {
	    bucketsLeft = (SECONDS_PER_DAY % (t->interval * BUCKETS_PER_RECORD)) /
	                  t->interval;
	} else {
	    bucketsLeft = BUCKETS_PER_RECORD;
	}

	return bucketsLeft;

    }

/*----------*/
int getRecordsPerDay (int interval)

    {

       double    recs;

       recs = (double)(SECONDS_PER_DAY) / (interval * BUCKETS_PER_RECORD);
       return recs == (int)(recs ? recs : recs + 1);


    }

/*----------*/
char *gmTime (time_t clock)

    {

	struct tm *t;
	static char       buf [30];

	t = gmtime (&clock);
	strcpy (buf, asctime (t));
	buf [strlen (buf) - 1] = '\0';
	return buf;

    }

/*----------*/
int severity2letter (int severity)	/* Convert a severity to a letter */

    {

	int	i;

	for (i=0 ; i<(sizeof (sevLetters) / sizeof (sevLetters [0])) ; i++)
	    if (sevLetters [i].sev == severity)
		return sevLetters [i].letter;
	return ' ';

    }

/*----------*/
int letter2severity (int letter)	/* Convert a letter to a severity */

    {

	int	i;

	for (i=0 ; i<(sizeof (sevLetters) / sizeof (sevLetters [0])) ; i++)
	    if (sevLetters [i].letter == letter)
		return sevLetters [i].sev;
	return -1;

    }

/*----------*/
int status2letter (int status)	/* Convert a status to a letter */

    {

        int  i;

	for (i=0 ; i<(sizeof (statLetters) / sizeof (statLetters [0])) ; i++)
	    if (statLetters [i].stat == status)
		return statLetters [i].letter;
	return ' ';

    }

/*----------*/
int letter2status (int letter)	/* Convert a letter to a status */

    {

        int    i;

	for (i=0 ; i<(sizeof (statLetters) / sizeof (statLetters [0])) ; i++)
	    if (statLetters [i].letter == letter)
		return statLetters [i].stat;
	return -1;

    }

/*----------*/
int methodString2Int (char *s)

    {

      int    i;

      for (i=0 ; i<NUM_METHODS ; i++)
	  if (strcmp (s, methodStrings [i]) == 0)
	      return i;

      /* Should never get here. */

      tisSysLog (1,  2, "Invalid reduction method - %s\n", s);
      exit (-1);

    }

/*----------*/
char *int2methodString (int i)

    {

        return methodStrings [i];

    }
