#include <vport.h>

/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vtextitemHEADER
#include vstartupHEADER
#include vtimeHEADER
#include vmemHEADER
#include vcontrolHEADER
#include vtimerHEADER
extern "C"
{
#include "opNoteDb.h"
}
#include "opno.h"
#include "opNoteUtils.hxx"
#include "opNote.hxx"


vkindTYPED_FULL_DEFINITION(opNoteSearchDialog, vdialog, "opNoteSearchDialog");

/*******************************************************************
*
*    opNoteSearchDialog override for constructors init ()
*
*******************************************************************/

void opNoteSearchDialog::LoadInit (vresource res)
{
    int		i;
    vchar	*pvChar = NULL;

    // store the search dialog's resource
    searchDialogResource = res;

    // set dialog status
    setSearchDialogStatus (NO_SEARCH_STATUS);

    oper = vtextitem::RequestSubLoad (this, opno_searchOperatorTag);
    startDate = vtextitem::RequestSubLoad (this, opno_searchStartDateTag);
    startTime = vtextitem::RequestSubLoad (this, opno_searchStartTimeTag);
    endDate = vtextitem::RequestSubLoad (this, opno_searchEndDateTag);
    endTime = vtextitem::RequestSubLoad (this, opno_searchEndTimeTag);
    keyClass = opNoteControl::RequestSubLoad (this, opno_searchKeyClassTag);
    keyValue = vcontrolOptionMenu::RequestSubLoad (this, opno_searchKeyValueTag);
    keyClass -> setKeyValue (keyValue);
    searchString = vtextitem::RequestSubLoad (this, opno_searchStringTag);
    applyButton = vbutton::RequestSubLoad (this, opno_searchApplyButtonTag);
    cancelButton = vbutton::RequestSubLoad (this, opno_searchCancelButtonTag);
    newButton = vbutton::RequestSubLoad (this, opno_searchNewButtonTag);

    // init inherited data
    vdialog::LoadInit (res);

    // store internal names for the search button tags
    commandNames[APPLY_BUTTON]  = opno_searchApplyButtonTag;
    commandNames[NEW_BUTTON]    = opno_searchNewButtonTag;
    commandNames[CANCEL_BUTTON] = opno_searchCancelButtonTag;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    // populate our dictionary with menu commands
    for (i = APPLY_BUTTON; i < SEARCHITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	opNoteSearchCommand *command = new opNoteSearchCommand;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (commandNames[i], command);
    }
    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);

    // build the keyClass menu
    keyClass -> buildKeyClassMenu (1);

    // build keyValue menu based on first item in keyClass list
    vmenu *keymenu = keyClass -> GetMenu ();
    vmenuItem *menuItem = keymenu -> GetItemAt (0);
    pvChar = (vchar *)menuItem -> GetTitle ();
    keyClass -> buildKeyValueMenu (pvChar);
}


/*******************************************************************
*
*    opNoteSearchDialog destroy function
*
*******************************************************************/

void opNoteSearchDialog::Destroy (void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    vdialog::Destroy ();
}

/*******************************************************************
*
*    opNoteSearchDialog notify function
*
*******************************************************************/

void opNoteSearchDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	setSearchDialogStatus (SEARCH_CANCEL);
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    opNoteSearchDialog, reset dialog fields
*
*******************************************************************/

void opNoteSearchDialog::resetDialogFields (void)
{
    oper -> SetTextScribed (vcharScribeSystem (""));
    startDate -> SetTextScribed (vcharScribeSystem (""));
    startTime -> SetTextScribed (vcharScribeSystem (""));
    endDate -> SetTextScribed (vcharScribeSystem (""));
    endTime -> SetTextScribed (vcharScribeSystem (""));

    // build key value menu based on 1st item in the
    // key class menu
    keyClass -> SetValue (0);
    vmenu *keyMenu = keyClass -> GetMenu ();
    vmenuItem *menuItem = keyMenu -> GetItemAt (0);
    vchar *pvChar = (vchar *)menuItem -> GetTitle ();
    keyClass -> buildKeyValueMenu (pvChar);

    searchString -> SetTextScribed (vcharScribeSystem (""));
    vmemSet ((void *)&searchFilter, 0, sizeof (Filter));
}


/*******************************************************************
*
*    opNoteSearchDialog function to start the search dialog
*
*******************************************************************/

void opNoteSearchDialog::startOpNoteSearch (void)
{
    // clear our copy of the filter
    vmemSet ((void *)&searchFilter, 0, sizeof (Filter));
    searchDialogStatus = NO_SEARCH_STATUS;
    // place search dialog in the center of the screen
    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open ();
    MakeTopMostDialog (this, NULL);
    // start an event loop for our dialog, this will hold up
    // the function that called us until we are finished
    vevent::Process();
}


/*******************************************************************
*
*    verifySearchParams ()
*	save date from search dialog in searchFilter
*	OP_DATE is defined in opNoteUtils.h
*	DATE_ERROR and TIME_ERROR defined in opNoteUtils.h
*	OP_SIZE is defiend in opNoteDb.h
*	Filter is a typedef in opNoteDb.h
*
*******************************************************************/

int opNoteSearchDialog::verifySearchParams (void)
{
    vchar	startDateStr[16];
    vchar	startTimeStr[16];
    vchar	endDateStr[16];
    vchar	endTimeStr[16];
    vchar	op[OP_SIZE + 1];
    vchar	note[MAX_NOTE + 1];
    vchar	*pvChar;
    DateTimeErrors  dateTimeReturn;
    struct tm	tim;

    vmemSet ((void *)&tim, 0, sizeof (struct tm));
    vmemSet ((void *)&searchFilter, 0, sizeof (Filter));
    vmemSet ((void *)&startDateStr[0], 0, sizeof (startDateStr));
    vmemSet ((void *)&startTimeStr[0], 0, sizeof (startTimeStr));
    vmemSet ((void *)&endDateStr[0], 0, sizeof (endDateStr));
    vmemSet ((void *)&endTimeStr[0], 0, sizeof (endTimeStr));
    vmemSet ((void *)&op[0], 0, sizeof (op));
    vmemSet ((void *)&note[0], 0, sizeof(note));

    /* get the operator */
    oper -> GetText (op, OP_SIZE);
    vcharExportBounded (vcharScribeSystemExport (op), searchFilter.op, OP_SIZE);
    startDate -> GetText (startDateStr, sizeof (startDateStr));
    startTime -> GetText (startTimeStr, sizeof (startTimeStr));
    if (startDateStr[0] != 0)
    {
	dateTimeReturn = extractDate (startDateStr, &tim);
	if (dateTimeReturn != OPDATE_OK)
	{
	    dateTimeErrorNotice (this, dateTimeReturn);
	    // set keyboard focus to this field, place cursor here
	    this -> SetFocusItem (startDate);
	    return (1);
	}
	dateTimeReturn = extractTime (startTimeStr, &tim);
	if (dateTimeReturn != OPDATE_OK)
	{
	    dateTimeErrorNotice (this, dateTimeReturn);
	    this -> SetFocusItem (startTime);
	    return (1);
	}
#if (defined(MIPSEL) || defined(SOLARIS) || defined (AXP) || defined(WIN32))
	searchFilter.start = mktime (&tim);
#else
	searchFilter.start = timelocal (&tim);
#endif
    }
    else
	searchFilter.start = 0;

    endDate -> GetText (endDateStr, sizeof (endDateStr));
    endTime -> GetText (endTimeStr, sizeof (endTimeStr));
    if (endDateStr[0] != 0)
    {
	vmemSet ((void *)&tim, 0, sizeof (struct tm));
	dateTimeReturn = extractDate (endDateStr, &tim);
	if (dateTimeReturn != OPDATE_OK)
	{
	    dateTimeErrorNotice (this, dateTimeReturn);
	    this -> SetFocusItem (endDate);
	    return (1);
	}
	dateTimeReturn = extractTime (endTimeStr, &tim);
	if (dateTimeReturn != OPDATE_OK)
	{
	    dateTimeErrorNotice (this, dateTimeReturn);
	    SetFocusItem (endTime);
	    return (1);
	}
#if (defined(MIPSEL) || defined(SOLARIS) || defined (AXP) || defined(WIN32))
	searchFilter.end = mktime (&tim);
#else
	searchFilter.end = timelocal (&tim);
#endif
    }
    else
	searchFilter.end = 0;

    if ( (startDateStr[0] != 0) && (endDateStr[0] != 0) )
    {
	if (searchFilter.end < searchFilter.start)
	{
	    dateTimeErrorNotice ((vdialog *)this, OPSTART_END_DATE_ERROR);
	    SetFocusItem (endDate);
	    return (1);
	}
    }

    // get the key class
    int controlValue = keyClass -> GetValue ();
    // get the menu attached to the control item 
    vmenu *keymenu = keyClass -> GetMenu ();
    // return a ptr to the string that is currently displayed on the menu 
    vmenuItem *menuItem = keymenu -> GetItemAt (controlValue);
    pvChar = (vchar *)menuItem -> GetTitle ();
    vstr *keyClassDef = vstrCloneScribed (vcharScribeSystem (KEY_CLASS_DEFAULT));
    if ( (vcharCompare (pvChar, keyClassDef)) != 0)
    {
	vcharExportBounded (vcharScribeSystemExport (pvChar), 
			    searchFilter.keytype, KEY_SIZE);
    }
    vstrDestroy (keyClassDef);

    // get the key value
    controlValue = keyValue -> GetValue ();
    keymenu = keyValue -> GetMenu ();
    menuItem = keymenu -> GetItemAt (controlValue);
    pvChar = (vchar *)menuItem -> GetTitle ();
    vstr *keyValueDef = vstrCloneScribed (vcharScribeSystem (KEY_VALUE_DEFAULT));
    if ( (vcharCompare (pvChar, keyValueDef)) != 0)
    {
	vcharExportBounded (vcharScribeSystemExport(pvChar),
			    searchFilter.keyvalue, KEY_SIZE);
    }
    vstrDestroy (keyValueDef);

    // get the search string
    searchString -> GetText (note, MAX_NOTE);
    vcharExportBounded (vcharScribeSystemExport (note),
			searchFilter.search_text, MAX_NOTE);
    return (0);
}


/*******************************************************************
*
*    opNoteSearch update filter function
*
*******************************************************************/

void opNoteSearchDialog::updateSearchFilter (Filter *filter)
{
    filter -> start = searchFilter.start;
    filter -> end   = searchFilter.end;
    strncpy (filter -> op, searchFilter.op, OP_SIZE);
    filter -> op[OP_SIZE] = 0;
    strncpy (filter -> keytype, searchFilter.keytype, KEY_SIZE);
    filter -> keytype[KEY_SIZE] = 0;
    strncpy (filter -> keyvalue, searchFilter.keyvalue, KEY_SIZE);
    filter -> keyvalue[KEY_SIZE] = 0;
    strncpy (filter -> search_text, searchFilter.search_text, MAX_NOTE);
    filter -> search_text[MAX_NOTE] = 0;
}


vkindTYPED_FULL_DEFINITION(opNoteSearchCommand, vcommandSelector, NULL);

/*******************************************************************
*
*    opNoteSearchCommand menu selector commands
*
*******************************************************************/

int opNoteSearchCommand::Issue (vdict *dict)
{
    opNoteSearchDialog *searchDialog = (opNoteSearchDialog *) dict -> GetElem (vname_Dialog);
    switch (GetSelector ())
    {
	case APPLY_BUTTON:
	    if (searchDialog -> verifySearchParams () == 0)
	    {
		searchDialog -> setSearchDialogStatus (SEARCH_OK);
		searchDialog -> Close ();
		// stops the event loop that was created when the
		// search dialog was Opened
		vevent::StopProcessing ();
	    }
	    break;

	case NEW_BUTTON:
	    searchDialog -> resetDialogFields ();
	    break;

	case CANCEL_BUTTON:
	    searchDialog -> setSearchDialogStatus (SEARCH_CANCEL);
	    searchDialog -> Close ();
	    vevent::StopProcessing ();
	    break;

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}

