/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


// LoginShell.C

#include <Gina/Gina.h>

extern "C"
{
#include <avl.h>
#include "logutils.h"
#include <ctype.h>
#ifndef MIPSEL
#include "memory.h"
#include "tisEncrypt.h"
#endif
}

#include <stdlib.h>
#ifdef SOLARIS
#include <unistd.h>
#endif
#include "loginShell.h"
#include "operDefs.h"
#include "utilDialogs.h"
#include "crypt.h"

extern	LoginShell	*theLoginShell;
extern	int		numOperators;
extern	char		loginFilePath [];
extern	char		displayHost [];
extern	LoginShell	*theLoginShell;
extern	int		menuChange;
void operLoadList(void *,void *);

//----------
LoginShell::LoginShell (GnDocument *doc, int width, int height) : LoginDialog (doc, width, height)

    {

	document = doc;

    }

//----------
LoginShell::~LoginShell (void)

    {

    }

//----------
void LoginShell::after_create (void)

    {
	LoginDialog :: after_create ();

	operatorList.add_browseSelectionCallback (CALLBACK (LoginShell, operSelectionCallback, this));
	operatorList.add_selectionCallback (CALLBACK (LoginShell, selection_changed, &operatorList));

      	passwordField.add_modifyVerifyCallback(CALLBACK(LoginShell,checkPassword,this));
      	passwordField.add_activateCallback(CALLBACK(LoginShell,checkPassword,this));
	passwordField.setR_maxLength(8);
	passwordField.SetValues();

	// Set up callbacks for the buttons

	okButton.add_activateCallback (CALLBACK (LoginShell, okButtonCallback, this));
	okButton.SetSensitive (False);
	cancelButton.add_activateCallback (CALLBACK (LoginShell, cancelButtonCallback, this));

	// Put the operators in the list

	GnPresentableList *list = new GnPresentableList;
	avl_traverse(operInfoTree,INORDER,operLoadList,(void *)list);
        operatorList.SetItemList (list, True);
	delete list;

    }
void operLoadList(void *p1,void *p2)
{
  GnPresentableList *list = (GnPresentableList *)p1;
  OPERATOR_ENTRY_PTR node = (OPERATOR_ENTRY_PTR)p2;
  list->Append (new GnStringItem (strdup (node->name)));
}
//----------
void LoginShell::operSelectionCallback (caddr_t)
{
  okButton.SetSensitive (True);
}

void LoginShell::selection_changed(void)
{
}

void LoginShell::checkPassword(caddr_t call_data)
{
int len;
XmTextVerifyCallbackStruct *cbs = (XmTextVerifyCallbackStruct *)call_data;

if(cbs->reason == XmCR_ACTIVATE)
  {
    return;
  }

if(cbs->startPos < cbs->currInsert)		/* delete any where... delete whole string */
  {
    cbs->startPos = 0;
    passwd[0] = 0;
    return;
  }

if(cbs->text->length > 1)
  {
    cbs->doit = False;
    return;
  }
if(isdigit(cbs->text->ptr[0]))
  {
    strncat(passwd,cbs->text->ptr,cbs->text->length);
    passwd[cbs->endPos + cbs->text->length] = 0;

    for(len = 0;len < cbs->text->length;len++)
      cbs->text->ptr[len] = '*';
  }
else
  {
    cbs->doit = False;
    return;
  }
}

//----------
void LoginShell::okButtonCallback (caddr_t)

    {

	FILE *f;
	OPERATOR_ENTRY entry,*pEntry;
	// Check to see if someone is already logged in. If yes then generate an implicit
	// logout for the logged in operator.

	if ((f = fopen (loginFilePath, "r")) != NULL) {
	    char line [100];
	    fgets (line, sizeof (line), f);
	    fclose (f);
	    line [strlen (line) - 1] = '\0';
	    if (strcmp (line, "nooperator") != 0)
		makeLogEntry (line, displayHost, IMPLICIT_LOGOUT);
	}

	// Get the selection from the listbox

	GnPresentable *item = operatorList.SelectedItem ();
	char name [100],msg[80];
	item->PrintOn (name, 100);

	// Find the selection in the list
	int loggedIn = 0;
	strcpy(entry.name,name);
	pEntry = (OPERATOR_ENTRY_PTR)avl_findsym(operInfoTree,&entry);

	// Check for workstation privledges
	if(!pEntry->opi)
	  {
	    makeLogEntry (name, displayHost, LOGIN_FAILURE);
	    sprintf(msg,"User\n%s\nDoes not have OPI authorization",name);
	    error_dialog (mainForm, msg);
	  }

	// See if the passwords match

	else if(pEntry->password[0] == '*')
	  {
	    if (strlen(passwd) == 0)
	      {
		loggedIn = 1;
	      }
	    else
	      {
		makeLogEntry (name, displayHost, LOGIN_FAILURE);
		error_dialog (mainForm, "Invalid password!");
	      }
	  } 
	else
	  {
	    char cryptedPassword [PASSWORD_SIZE*2+1];
	    strcpy (cryptedPassword, passwd);
	    tisEncrypt (cryptedPassword,PASSWORD_SIZE);
	    if (strcmp (pEntry->password, cryptedPassword) == 0)
	      {
		loggedIn = 1;
	      }
	    else
	      {
		makeLogEntry (name, displayHost, LOGIN_FAILURE);
		error_dialog (mainForm, "Invalid password!");
	      }
	  }

	if (loggedIn) {

	    // Create the login file

	    f = fopen (loginFilePath, "w");
	    fprintf (f, "%s\n", name);
	    fclose (f);

#ifndef AXP
	    // Create the symbolic link for the menus

// 	    if (menuChange) {
// 		char linkPath [100], linkName [100];
// 		sprintf (linkPath, "/tis4000/operators/%s/openwin-menu", name);
// 		sprintf (linkName, "%s/.openwin-menu", getenv ("HOME"));
// 		unlink (linkName);
// 		symlink (linkPath, linkName);
// 	    }
#endif

	    // Write this transaction to the log file

	    makeLogEntry (name, displayHost, LOGIN);
	    exit (-1);

	}

    }

//----------
void LoginShell::cancelButtonCallback (caddr_t)

    {

	exit (-1);

    }

