/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include "tiscard.h"
#include "msgMon.h"

/* // The path to the login file */
struct termios  savetty;
char  loginFilePath[200],displayHost[100];
AVL_TREE  *ticNameTree;
NAME_ENTRY  *currentUser = 0;		/* set after successful login */
int maxFd,ticFD;
int statTimeout = 60;			/* time out count for stat() call */
time_t fileTime = 0;			/* LoginFilePath time stamp */
struct stat statBuff;
fd_set  myFds,saveFds;

main(int argc, char **argv)
{
  struct timeval timeout;
  int status;
  LOGIN_TYPE  idType;
  char  idString[SLM_CARDID_SIZE+1];
  char  warnMsg[200];
  int hours,minutes,seconds;
  FILE *f;
  NAME_ENTRY entry,*pEntry;
  char line [100];

  tisOpenLog ("tisCard", 6, 2, NULL, NULL);

#ifdef SOLARIS
  sigset(SIGHUP,ticSignalHandler);
  sigset(SIGINT,ticSignalHandler);
#else
  signal(SIGHUP,ticSignalHandler);
  signal(SIGINT,ticSignalHandler);
#endif
  openlog("tisCard",LOG_PID,LOG_LOCAL1);
  ticFD = 0;
  if(argc >= 2)
    {
      if((ticFD = ticInitPort(argc,argv[1])) < 0)
	{
	  tisSysLog (0,0, "usage; tisCard ttya|ttyb");
	  exit(1);					/* usage: tisCard ttya|ttyb */
	}
      if(ticFD > maxFd) maxFd = ticFD;
    }
  FD_SET(ticFD,&saveFds);
  ticNameTree = (AVL_TREE *)avl_init(ticNameCompare);
  ticInit();
  /*
    Check the login file to get the current user. This is done so that when
    tisCard is started after somw one has logged in via tisLogin, the current
    user will be known.
  */
  if ((f = fopen (loginFilePath, "r")) != NULL)
    {
      fgets (line, sizeof (line), f);
      fclose (f);
    }
  stat(loginFilePath,&statBuff);
  fileTime = statBuff.st_mtime;	/* set to new time after file is closed */
  line [strlen (line) - 1] = '\0';
  strcpy(entry.ticName,line);
  if((pEntry = avl_findsym(ticNameTree,&entry)) != 0)
    {
      if(!strcmp(line,"nooperator"))
	currentUser = 0;
      else
	currentUser = pEntry;		/* really is already the current user */
      idType = TISLOGIN;
    }
  timeout.tv_usec = 0;
  timeout.tv_sec = 1;
  memcpy((char *)&myFds,(char *)&saveFds,sizeof(fd_set));
  memset(idString,0,SLM_CARDID_SIZE+1);			/* init for first ID string */
  while(1)
  {
    status = select(maxFd+1,&myFds,NULL,NULL,&timeout);
    if(status == 0)
      {
	if(--statTimeout <= 0)				/* is it time to check the login file? */
	  {
	    statTimeout = 60;				/* reset stat time out */
	    if(stat(loginFilePath,&statBuff) != -1)   	/* get current file time */
	      {
		if(statBuff.st_mtime > fileTime)	/* has file changed since last time? */
		  {
		    if ((f = fopen (loginFilePath, "r")) != NULL)
		      {
			fgets (line, sizeof (line), f);
			fclose (f);
		      }
		    stat(loginFilePath,&statBuff);
		    fileTime = statBuff.st_mtime;	/* set to new time after file is closed */
		    line [strlen (line) - 1] = '\0';
		    if(currentUser)
		      {
			if(strcmp(line,currentUser->ticName))	/* login only if current user different */
			   {
			     strcpy(entry.ticName,line);
			     if((pEntry = avl_findsym(ticNameTree,&entry)) != 0)
			       {
				 currentUser = pEntry;
				 makeLogEntry(line,displayHost,LOGIN);
				 idType = TISLOGIN;
			       }
			   }
		      }
		    else		/* NO current user then just do it! */
		      {
			strcpy(entry.ticName,line);
			if((pEntry = avl_findsym(ticNameTree,&entry)) != 0)
			  {
			    currentUser = pEntry;
			    makeLogEntry(line,displayHost,LOGIN);
			    idType = TISLOGIN;
			  }
		      }
		  }
	      }
	  }
	if(currentUser)
	  {
	    if(currentUser->ticTimeout != -5)
	      {
		if((currentUser->ticTimeout == 60) && ((idType == KEYPAD) || (idType == TISLOGIN)))	/* GCT 19-Nov-1996 warn on timeout */
		  {
		    sprintf(warnMsg,"/tis4000/tasks/autoLogoutWarning %s &",currentUser->ticName);
		    system(warnMsg);
		  }
		if((--currentUser->ticTimeout <= 0) && ((idType == KEYPAD) || (idType == TISLOGIN)))	/* timeout only with keypad */
		  {
		    ticLogout(currentUser);
		    currentUser->ticTimeout = -5;
		    sscanf(currentUser->ticLogoutTime,"%d:%d:%d",&hours,&minutes,&seconds); /* convert time */
		    if((hours+minutes+seconds) != 0)
		      currentUser->ticTimeout = seconds + minutes*60 + hours*3600;
		    currentUser = 0;
		  }
	      }
	  }
      }
    else if(status < 0)
      {
	tisSysLog (0,0, "Error from select %m");
      }
    else
      {
	if(FD_ISSET(ticFD,&myFds))
	  {
	    idType = ticReadID(ticFD,idString);		/* get data from card/keypad */
	    if((idType == KEYPAD) || (idType == MAGCARD))
	      {
		if(strlen(idString))
		  ticLogin(idString,idType);
		memset(idString,0,SLM_CARDID_SIZE+1);  	/* init for next ID string */
	      }
	    else if(idType == READ_ERROR)
	      tisSysLog (0,0, "Read error on serial port...%m");
	  }
      }
    memcpy((char *)&myFds,(char *)&saveFds,sizeof(fd_set));
  }
}

void ticSignalHandler(int whichone)
{
switch(whichone)
  {
  case SIGINT:
    avl_empty(ticNameTree,avl_freenode);
    tcsetattr(ticFD,TCSANOW,&savetty);
    close(ticFD);
    exit(2);
  case SIGHUP:
    avl_empty(ticNameTree,avl_freenode);
    ticBuildNameTree();
    tisSysLog(0,0, "Re-initializing");
  }
}

void ticLogin(char *passwd,LOGIN_TYPE type)
{
  char line [100];
  char cryptedPassword [PASSWORD_SIZE*2+1];
/*      traverse tree looking for either keypadID (encryped) or cardID */
  if(type == KEYPAD)
    {
      strcpy(cryptedPassword, passwd);
      if(cryptedPassword[0] != '*')
	tisEncrypt(cryptedPassword,PASSWORD_SIZE);
      avl_traverse(ticNameTree,INORDER,ticCheckPassword,cryptedPassword);
    }
  else
    avl_traverse(ticNameTree,INORDER,ticCheckCard,passwd);
  if(currentUser == 0)
    makeLogEntry (line, displayHost, LOGIN_FAILURE);
}

int ticInit(void)
{
  FILE *f;
/* 	 Make sure the /tis4000/operators directory exists */
  strcpy (loginFilePath, "/tis4000/operators");
  if (access (loginFilePath, R_OK | W_OK) != 0)
    {
      tisSysLog (0,0, "/tis4000/operators directory not found!...%m");
      exit (-1);
    }
/* 	 If this is the first time a login has ever occurred then it is likely that */
/* 	 /tis4000/operators/logins does not exist. Create it if it doesn't. */

  strcat (loginFilePath, "/logins");
  if (access (loginFilePath, R_OK | W_OK) != 0)
    {
      if (mkdir (loginFilePath, 0700) != 0)
	{
	  tisSysLog (0,0, "Unable to create /tis4000/operators/logins...%m");
	  exit (-1);
	}
    }
/* 	 If a login file does not exist for this host. Create one with the "nooperator" */
/* 	 operator.  This indicates that no one is currently logged in. */
  strcat (loginFilePath, "/");
  strcpy (displayHost, getDisplayHost ());
  strcat (loginFilePath, displayHost);
  if (access (loginFilePath, R_OK | W_OK) != 0)
    {
      if ((f = fopen (loginFilePath, "w")) == NULL)
	{
	  tisSysLog (0,0, "Unable to create %s...%m",loginFilePath);
	  exit (-1);
	}
      fputs ("nooperator\n", f);
      fclose (f);
    }
  stat(loginFilePath,&statBuff);
  fileTime = statBuff.st_mtime;		/* get initial time */
  ticBuildNameTree();
}
/* build tree */
void ticBuildNameTree(void)
{
  FILE *f;
  NAME_ENTRY_PTR  pEntry;
  char line[SLM_MAX_LINE_SIZE];
  int hours,minutes,seconds;
/* 	 Build a tree of names and passwords */
  if((f = fopen("/tis4000/operators/secureList","r")) != 0)
    {
      while((fgets(line,SLM_MAX_LINE_SIZE-1,f)) != 0)
	{
	  line[strlen(line)-1] = 0;				/* drop \n and NULL terminate */
	  pEntry = (NAME_ENTRY_PTR)avl_newnode(NAME_ENTRY_SIZE);	/* create new node */
	  strcpy(pEntry->ticName,(char *)strtok(line," "));		/* get name */
	  strcpy(pEntry->ticKeypadID,(char *)strtok(NULL," "));		/* then keypad ID(password) */
	  strcpy(pEntry->ticCardID,(char *)strtok(NULL," "));		/* then card ID */
	  strcpy(pEntry->ticClearID,(char *)strtok(NULL," "));		/* then un-encrypted keypad ID */
	  strcpy(pEntry->ticLogoutTime,(char *)strtok(NULL," "));		/* then logout time */
	  strtok(NULL," ");
	  pEntry->ticIsOPI = (!strcmp((char *)strtok(NULL," "),"NOOPI") ? 0 : 1);	/* OK for this workstation */
	  pEntry->ticTimeout = -5;
	  sscanf(pEntry->ticLogoutTime,"%d:%d:%d",&hours,&minutes,&seconds); /* convert time */
	  if((hours+minutes+seconds) != 0)
	    pEntry->ticTimeout = seconds + minutes*60 + hours*3600;
	  avl_insert(ticNameTree,pEntry);
	}
      fclose(f);
    }
}
/* compare entry names */
int ticNameCompare(void *p1,void *p2)
{
  return(strcmp(((NAME_ENTRY_PTR)p1)->ticName,
		((NAME_ENTRY_PTR)p2)->ticName));
}

void ticCheckPassword(void *p1,void *p2)
{
NAME_ENTRY_PTR node = (NAME_ENTRY_PTR)p2;
char *pass = (char *)p1;

/*      See if the passwords match */
if(!strcmp(node->ticKeypadID, pass) && (node->ticIsOPI))
  {
    ticPerformLogin(node->ticName,node);
  }
}
void ticCheckCard(void *p1,void *p2)
{
NAME_ENTRY_PTR node = (NAME_ENTRY_PTR)p2;
char *pass = (char *)p1;

if(!strcmp(node->ticCardID,pass) && (node->ticIsOPI))
  {
    ticPerformLogin(node->ticName,node);
  }
}
void ticPerformLogin(char *name,NAME_ENTRY_PTR node)
{
  FILE *f;
  char line[20];

/*      Check to see if someone is already logged in. If yes then generate an implicit */
/*      logout for the logged in operator. */
  if ((f = fopen (loginFilePath, "r")) != NULL)
    {
      fgets (line, sizeof (line), f);
      fclose (f);
    }
  line [strlen (line) - 1] = '\0';
  /*
    if the name is the same then just logout and return
  */
  if(!strcmp(line,name))
    {
      f = fopen (loginFilePath, "w");		/*      Create the login file */
      fprintf (f, "nooperator\n", name);	/* logout if name already logged in */
      fclose (f);
      stat(loginFilePath,&statBuff);
      fileTime = statBuff.st_mtime;	/* set to new time after file is closed */
      makeLogEntry (name, displayHost, EXPLICIT_LOGOUT);
      currentUser = 0;				/* logout means NO current user */
      return;
    }
  if (!strcmp (line, "nooperator"))
    makeLogEntry (line, displayHost, IMPLICIT_LOGOUT);
  f = fopen (loginFilePath, "w");	/*      Create the login file */
  fprintf (f, "%s\n", name);
  fclose (f);
  stat(loginFilePath,&statBuff);
  fileTime = statBuff.st_mtime;	/* set to new time after file is closed */
  makeLogEntry (name, displayHost, LOGIN);
  currentUser = node;
/* #ifndef AXP */
/*     Create the symbolic link for the menus */
/*  	    if (menuChange) { */
/*  		char linkPath [100], linkName [100]; */
/*  		sprintf (linkPath, "/tis4000/operators/%s/openwin-menu", name); */
/*  		sprintf (linkName, "%s/.openwin-menu", getenv ("HOME")); */
/*  		unlink (linkName); */
/*  		symlink (linkPath, linkName); */
/*  	    } */
/* #endif */
}
int ticInitPort(int count, char *portName)
{
  char realPort[32];
  int newfd,i;
  struct termios	tinfo;

  if (count != 2)
    {
      tisSysLog(0,0,"usage: tisCard /dev/tty[a,b]");
      return (-1);
    }
  if (strstr(portName, "ttya") != NULL)
    {
#ifdef SOLARIS
      strcpy(realPort,"/dev/cua/a");
#else
      strcpy(realPort,"/dev/cua0");
#endif
    }
  else if (strstr(portName, "ttyb") != NULL)
    {
#ifdef SOLARIS
      strcpy(realPort,"/dev/cua/b");
#else
      strcpy(realPort,"/dev/cua1");
#endif
    }
  else
    {
      tisSysLog(0,0,"usage: tisCard /dev/tty[a,b]");
      return(-1);
    }
  if((newfd = open(realPort,O_RDWR | O_NDELAY)) < 0)
    {
      tisSysLog (0,0,"tisCard: failed to open serial port %s (%m)",realPort);
      return (-1);
    }
  memset ((char *)&tinfo, 0, sizeof (struct termios));
  if (tcgetattr (newfd, &tinfo) < 0)
    {
      tisSysLog (0,0,"tisCard: failed to get port settings (%m)");
      return (1);
    }
  /* save the original settings to restore later */
  savetty = tinfo;
  /* now that the tty is open set up its characteristics */
  /* set input modes */
  tinfo.c_iflag |= IGNBRK;	/* ignore breaks */
  tinfo.c_iflag |= IGNPAR;	/* ignore parity errors */
  tinfo.c_iflag |= ISTRIP;	/* strip the 8th bit */
  /* set output modes */
  tinfo.c_oflag = 0;
  /* set control modes */
  /* set output baud rate, this also sets the input baud rate
     since we are not specify an input baud rate diff from
     the output
     */
  if(cfsetospeed (&tinfo, B9600) < 0)
    {
      tisSysLog (0,0,"tisCard: failed to set output baud rate (%m)");
      return (-1);
    }
  tinfo.c_cflag |= CS8;			/* set chararcter to 7 bit */
  tinfo.c_cflag |= CREAD;		/* enable receiver */
  tinfo.c_cflag |= CLOCAL;		/* direct control, no modem control */
  /* set line discipline modes */
  tinfo.c_lflag = 0;
  /* set controls:
     min chars requested or # chars available
     return w/o waiting for more chars
     */
  tinfo.c_cc[VMIN] = 0;
  tinfo.c_cc[VTIME] = 0;
  /* set new line attributes NOW */
  if (tcsetattr (newfd, TCSANOW, &tinfo) < 0)
    {
      tisSysLog (0,0,"tisCard: failed to set tty characteristics (%m)");
      return (-1);
    }
  /* set exclusive use of this tty, the arg i is ignored for this call */
  if (ioctl (newfd, TIOCEXCL, &i) < 0)
    {
      tisSysLog (0,0,"tisCard: failed to set exclusive tty use (%m)");
      return (-1);
    }
return(newfd);
}
/* read bytes from card/keypad until terminator */
LOGIN_TYPE ticReadID(int fd,char *pw)
{
char inchar[SLM_CARDID_SIZE+1];
int  count,size;

if(ioctl(fd,FIONREAD,&size) != -1)			/* only read as maany as are in buffer */
  {
    if((count = read(fd,inchar,size)) < 0)
      {
	tisSysLog(0,0,"Read error from serial port: %m");
	return(READ_ERROR);
      }
    inchar[size] = 0;					/* NULL terminate */
    if(strchr(inchar,LF))	       			/* CR -> LF. pw has full string now */
      return(KEYPAD);
    else if(strchr(inchar,VT))				/* VT means from card swipe */
      {
	inchar[size-1] = 0;				/* drop VT */
	strcat(pw,inchar);
	return(MAGCARD);
      }
    else
      {
/* 	if(isdigit(inchar[0])) */
	  strcat(pw,inchar);
	return(IDNOT_YET);
      }
  }
else
  {
    tisSysLog(0,0,"IOCTL error from serial port: %m");
    return(IOCTL_ERROR);
  }
}

/* log user out after logout timer has expired */
void ticLogout(NAME_ENTRY_PTR who)
{
  FILE *f;
  /* Log the logout */
  makeLogEntry (who->ticName, displayHost, EXPLICIT_LOGOUT);
  /* Create a new file with "nooperator" as the operator */
  f = fopen(loginFilePath, "w");
  fprintf(f, "nooperator\n");
  fclose(f);
  stat(loginFilePath,&statBuff);
  fileTime = statBuff.st_mtime;	/* set to new time after file is closed */
}
