#dldef.m4,v 1.1.1.1 1999/08/03 14:43:47 tis4000 Exp

#################################################################
#								#
#	This software is the proprietary property of Tate	#
#	Integrated Systems, L.P.  Patents and copyrights are	#
#	pending.						#
#								#
#	This software is furnished under a license and may	#
#	be used and copied only in accordance with the terms	#
#	of such license and with the inclusion of the above	#
#	notice. This software or any other copies thereof	#
#	may not be provided or otherwise made available to	#
#	any other person. No title to and ownership of the	#
#	software is hereby transferred.				#
#								#
#	The information in this software is subject to		#
#	change without notice and should not be construed 	#
#	as a commitment by Tate Integrated Systems L.P. or	#
#	its third party suppliers.				#
#								#
#	Tate Integrated Systems, L.P. (TIS) and its third	#
#	party suppliers assume no responsibility for the	#
#	use or inability to use any of its software. TIS	#
#	software is provided "as is" without warranty of	#
#	any kind and TIS expressly disclaims all implied	#
#	warranties, including but not limited to the implied	#
#	warranties of merchantability and fitness for a		#
#	particular purpose.					#
#								#
#	Notice: Notwithstanding any other lease or license	#
#	that may pertain to or accompany the delivery of this	#
#	computer software, the rights of the Government		#
#	regarding its use, reproduction, and disclosure are	#
#	as set forth in Section 52.227-19 of the FARS		#
#	Computer Software-Restricted Rights clause.		#
#								#
#################################################################
#
#	D L D E F . M 4
#
#	Author 	Jeffrey O. Hill
#		hill@luke.lanl.gov
#		505 665 1831
#
#	Source for all display list C include files
#
#
#	Notes
#	.01 joh 011191	Update major/minor versions if you make a incompatible 
#			change
#
#	
#	History
#	.01 joh 011191	Versions 1.1, 1.2, 1.3
#			Related displays in seperate partition
#			display default background/foreground colors
#			stack of related displays
#			extend label options
#			new color index data type
#			new menu for indicator fill direction
#			strip chart period units
#			color map file name
#			text alignment for static strings
#			created pen structure for strip charts
#
#	.02 joh 013091	Versions 1.4, 1.5, 1.6, 1.7, 1.8
#			RISC padding
#			dynamic attributes extended
#			default sta and dyn attributes in the header
#			new display x,y,ht,wd constraint
#			omisc became ocmap
#			row/column/both for buttons
#			pointer direction on valuators
#			removed dynamic line style and dynamic fill
#
#	.03 joh 022591	Versions 1.9 1.10
#			dynamic visibility moved to independent catagory
#			cartesian plots revised
#
#	.04 joh 032591	Version 1.11 1.12
#			New smith chart overlay element
# 			text format and alignment specification for dynamic strings
#
#	.05 joh 041091	Version 1.13 1.14
#			more bits in the type field
#			related display labels
#			new cartesian plot parameters
#
#
#	Possible Improvements
#	.01 joh 011191	hi/lo display ranges (% full scale?)		
#			How does the run time stuff know hopr and lopr apply
#			to the field they are attached to?
#
#
#
#
#
#

DLDEF_INLINE
/*
 * NOTE: the display list elements currently contain nothing larger
 * than a short. If this changes RISC alignment issues will need to be
 * revisited.
 */

/*
 * display list version
 */
#define DL_MAJOR_VERSION 1
#define DL_MINOR_VERSION 21

/*
 * magic number identifying properly byte swaped dl segments
 */
#define DL_MAGIC 0x12345678

/* round from bytes to words */
/*#define WDRND(A) ( (A)&1 ? ((A)>>1)+1 : ((A)>>1) )*/
#define WDRND(A) ( ((A)>>1) + (((A)&3) ? (((A)&2)? 1 : 2) : 0) )

/*#define NULL				0*/
#define FALSE				0
#define TRUE				1
#define OK				0
#define ERROR				(-1)
#define OFFSET(structure, member)	((int) &(((structure *) 0) -> member))
#define MEMBER_SIZE(structure, member)	(sizeof (((structure *) 0) -> member))
#define NELEMENTS(array)		(sizeof (array) / sizeof ((array) [0]))
#define FOREVER				for (;;)
#define max(x, y)			(((x) < (y)) ? (y) : (x))
#define min(x, y)			(((x) < (y)) ? (x) : (y))

#define XREC_TO_DLOBJ(XREC, DLOBJ)\
{\
(DLOBJ).x = (XREC).x;\
(DLOBJ).y = (XREC).y;\
(DLOBJ).width = (XREC).width;\
(DLOBJ).height = (XREC).height;\
}

#define DLOBJ_TO_XREC(DLOBJ, XREC)\
{\
(XREC).x = (DLOBJ).x;\
(XREC).y = (DLOBJ).y;\
(XREC).width = (DLOBJ).width;\
(XREC).height = (DLOBJ).height;\
}

#define DL_MINCOLOR 0
#define DL_MIDCOLOR (DL_MAXCOLOR>>1)
#define DL_MAXCOLOR ((unsigned char) ~0)

#define MAX_NPOINTS	50
#define MAX_RULE_LINES	16
ENDDLDEF_INLINE

COMMENT
	Data type enum has been omitted as a possible
	display list field type due to the risk that the
	field size might change depending on the enumeraton
	used. Instead, make an integer structure element or
	bit field and assign an enumerated value to it.

	Most compilers allocate 32 bits to an enumeration.
	If we did use the type enumeration for enumerated fields
	our display lists would be much larger.  
ENDCOMMENT
ENUMERATION(dl_data_type)
	ITEM(	dl_char, 		char)
	ITEM(	dl_uchar, 		unsigned char)
	ITEM(	dl_pchar, 		char *)
	ITEM(	dl_short, 		short)
	ITEM(	dl_ushort, 		unsigned short)
	ITEM(	dl_pshort, 		short *)
	ITEM(	dl_int,  		int)
	ITEM(	dl_uint, 		unsigned int)
	ITEM(	dl_pint, 		int *)
	ITEM(	dl_float, 		float)
	ITEM(	dl_pfloat, 		float *)
	ITEM(	dl_double, 		double)
	ITEM(	dl_pdouble, 		double *)
	ITEM(	dl_field, 		int)
	ITEM(	dl_ufield, 		unsigned)
	ITEM(	dl_struct, 		struct)
	ITEM(	dl_pc_index,		char *)
	ITEM(	dl_string_index,	char *)
	ITEM(	dl_color_index,		unsigned char)
	ITEM(	dl_bckgnd_color_index,	unsigned char)
ENDENUMERATION

ENUMERATION(edd_user_input)
	ITEM(	edd_none, 	no input)
	ITEM(	edd_valuator, 	scroll bar)
	ITEM(	edd_text_edit, 	text entry box)
	ITEM(	edd_message, 	unmodified text)
	ITEM(	edd_menu, 	enum from menu)
	ITEM(	edd_button, 	radio button)
	ITEM(	edd_palette, 	color selection)
	ITEM(	edd_form, 	user input form)
	ITEM(	edd_subform, 	pop up user input form)
ENDENUMERATION

ENUMERATION(edd_valuator_constraint_type)
	ITEM(	edd_no_constraint)
	ITEM(	edd_x_pos, )
	ITEM(	edd_y_pos, )
	ITEM(	edd_width, )
	ITEM(	edd_height, )
	ITEM(	edd_x_pos_on_root, )
	ITEM(	edd_y_pos_on_root, )
	ITEM(	edd_width_on_root, )
	ITEM(	edd_height_on_root, )
	ITEM(	edd_begin_angle, )
	ITEM(	edd_line_width, )
	ITEM(	edd_intensity, )
	ITEM(	edd_path_angle, )
ENDENUMERATION

ENUMERATION(dl_dynamic_color_mod)
	ITEM(	dl_clrmod_static, 	static)
	ITEM(	dl_clrmod_rule, 	rule)
ENDENUMERATION

ENUMERATION(dl_color_rule_connector)
	ITEM(	dl_clrrule_use, 	use)
	ITEM(	dl_clrrule_and, 	and)
	ITEM(	dl_clrrule_andnot, 	and not)
ENDENUMERATION

ENUMERATION(dl_dynamic_visibility_mod)
	ITEM(	dl_vismod_static, 	static)
	ITEM(	dl_vismod_not_zero, 	if not zero)
	ITEM(	dl_vismod_zero, 	if zero)
ENDENUMERATION

ENUMERATION(dl_dynamic_io_type)
	ITEM(	dl_io_monitor, 		monitor)
	ITEM(	dl_io_control, 		control)
ENDENUMERATION

ENUMERATION(dl_dynamic_geom_mod)
	ITEM(	dl_geommod_right, 	fill right)
	ITEM(	dl_geommod_up, 		fill up)
	ITEM(	dl_geommod_left, 	fill left)
	ITEM(	dl_geommod_down, 	fill down)
	ITEM(	dl_geommod_radial, 	fill radial)
ENDENUMERATION

ENUMERATION(dl_dynamic_direction)
	ITEM(	dl_increase_right, 	right)
	ITEM(	dl_increase_left, 	left)
	ITEM(	dl_increase_up, 	up)
	ITEM(	dl_increase_down, 	down)
	ITEM(	dl_increase_radial, 	radial)
ENDENUMERATION

ENUMERATION(dl_pointer_direction)
	ITEM(	dl_pointer_down, 	down)
	ITEM(	dl_pointer_up, 		up)
	ITEM(	dl_pointer_right, 	right)
	ITEM(	dl_pointer_left, 	left)
ENDENUMERATION

ENUMERATION(dl_button_stacking)
	ITEM(	dl_stack_column, 	column)
	ITEM(	dl_stack_row, 		row)
	ITEM(	dl_stack_row_column, 	row column)
ENDENUMERATION

ENUMERATION(dl_dynamic_fill_mod)
	ITEM(	dl_edge_fill, 	from edge)
	ITEM(	dl_zero_fill, 	from zero)
ENDENUMERATION

ENUMERATION(dl_dynamic_label_type)
	ITEM(	dl_label_none, 		none)
	ITEM(	dl_label_outline, 	outline)
	ITEM(	dl_label_limits, 	limits)
	ITEM(	dl_label_channel, 	channel)
ENDENUMERATION

ENUMERATION(dl_fill_style)
	ITEM(	dl_fill_outline, 	outline)
	ITEM(	dl_fill_solid,	 	solid)
ENDENUMERATION

ENUMERATION(dl_edge_style)
	ITEM(	dl_edge_solid, 		solid)
	ITEM(	dl_edge_dash, 		dash)
ENDENUMERATION

ENUMERATION(dl_plot_style)
	ITEM(	dl_point_plot,		point plot)
	ITEM(	dl_line_plot, 		line plot)
	ITEM(	dl_fill_plot, 		fill under)
ENDENUMERATION

ENUMERATION(dl_time_units)
	ITEM(	dl_day,			day)
	ITEM(	dl_hour,		hour)
	ITEM(	dl_min,			minute)
	ITEM(	dl_sec,			second)
	ITEM(	dl_msec,		milli second)
	ITEM(	dl_usec,		micro second)
ENDENUMERATION

ENUMERATION(dl_string_alignment)
	ITEM(	dl_horiz_left,		horiz. left)
	ITEM(	dl_horiz_right,		horiz. right)
	ITEM(	dl_horiz_center,	horiz. centered)
	ITEM(	dl_vert_top,		vert. top)
	ITEM(	dl_vert_bottom,		vert. bottom)
	ITEM(	dl_vert_center,		vert. centered)
ENDENUMERATION

ENUMERATION(dl_string_format)
	ITEM(	dl_format_f,		decimal)
	ITEM(	dl_format_e,		decimal- exponential notation)
	ITEM(	dl_format_g,		decimal- compact)
	ITEM(	dl_format_i,		decimal- truncated )
	ITEM(	dl_format_x,		hexidecimal, truncated )
	ITEM(	dl_format_o,		octal, truncated)
ENDENUMERATION

ENUMERATION(dl_element_class)
	ITEM(	dl_class_attribute, 		attribute)
	ITEM(	dl_class_graphic, 		graphic)
	ITEM(	dl_class_control, 		control)
	ITEM(	dl_class_monitor, 		monitor)
	ITEM(	dl_class_misc, 			miscellaneous)
	ITEM(	dl_class_internal, 		internal)
	ITEM(	dl_class_plot, 			plot)
	ITEM(	dl_class_related_display, 	related display)
	ITEM(	dl_class_execute_script, 	execute script)
	ITEM(	dl_class_kill_display,		kill display)
ENDENUMERATION

ENUMERATION(dl_on_off)
	ITEM(	dl_off, off)
	ITEM(	dl_on, 	on)
ENDENUMERATION

COMMENT
	room here for 64 distinct elements each no longer than 1024 words
ENDCOMMENT
STRUCTURE(dl_element_hdr)
	FIELD(	type, 	element type, edd_menu, dl_uchar,, dl_elements)
	FIELD(	pad, available, edd_none, dl_char,)
	FIELD(	nwords, shrt wds to nxt elmnt, 	edd_none, dl_ushort,)
ENDSTRUCTURE

COMMENT
	Identical to the XRectangle structure
ENDCOMMENT
STRUCTURE(dl_object_hdr)
	FIELD(	x, 	horizontal position, 	edd_valuator,	dl_short,,edd_x_pos)
	FIELD(	y, 	vertical position, 	edd_valuator,	dl_short,,edd_y_pos)
	FIELD(	width, 	horizontal size, 	edd_valuator,	dl_short,,edd_width)
	FIELD(	height, vertical size, 		edd_valuator,	dl_short,,edd_height)
ENDSTRUCTURE

COMMENT
	Identical to the XPoint structure
ENDCOMMENT
STRUCTURE(dl_point)
	FIELD(	x, 	horizontal position, 	edd_none,	dl_short,)
	FIELD(	y, 	vertical position, 	edd_none,	dl_short,)
ENDSTRUCTURE

COMMENT
	Common plot parameters
ENDCOMMENT
STRUCTURE(dl_plot_hdr)
	FIELD(	pad, available, edd_none, dl_short)
	FIELD(	title, 	plot title, 	edd_text_edit,	dl_string_index)
	FIELD(	xlabel, x axis title, 	edd_text_edit,	dl_string_index)
	FIELD(	ylabel, y axis title, 	edd_text_edit,	dl_string_index)
	FIELD(	package, plot package, 	edd_text_edit,	dl_string_index)
	FIELD(	clr,	foreground color, edd_palette,  dl_color_index)
	FIELD(	bclr,	background color, edd_palette,  dl_bckgnd_color_index)
ENDSTRUCTURE

COMMENT
	Inserted into all control objects
ENDCOMMENT
STRUCTURE(dl_control_hdr)
	FIELD(	pad, available, edd_none, dl_short)
	FIELD(	ctrl, channel to control, 	edd_text_edit,	dl_pc_index)
	FIELD(	clr, foreground color, 		edd_palette,	dl_color_index)
	FIELD(	bclr, background color,		edd_palette,	dl_bckgnd_color_index)
ENDSTRUCTURE

COMMENT
	Inserted into all monitor objects
ENDCOMMENT
STRUCTURE(dl_monitor_hdr)
	FIELD(	pad, available, edd_none, dl_short)
	FIELD(	rdbk, 	channel to monitor, 	edd_text_edit,	dl_pc_index)
	FIELD(	clr, foreground color, 		edd_palette,	dl_color_index)
	FIELD(	bclr, background color,		edd_palette,	dl_bckgnd_color_index)
ENDSTRUCTURE

COMMENT
	Inserted into dl_basic_attr
ENDCOMMENT
STRUCTURE(dl_sta_attr)
	FIELD(	clr,	color, 	    edd_palette,  dl_color_index)
	FIELD(	style,	edge style, edd_button,   dl_uchar,, dl_edge_style)
	FIELD(	fill,	fill style, edd_button,   dl_uchar,, dl_fill_style)
	FIELD(	width,	line width, edd_valuator, dl_uchar,, edd_line_width)
ENDSTRUCTURE

COMMENT
	for dynamic graphical elements only

	Sorry about extra substructures but the editor must open
	these structures independently
ENDCOMMENT
STRUCTURE(dl_dyn_mod)
	FIELD(	clrmod, color modifier, edd_menu, dl_uchar, , dl_dynamic_color_mod)
	FIELD(	vis, visibility modifier, edd_menu, dl_uchar, , dl_dynamic_visibility_mod)
ENDSTRUCTURE
STRUCTURE(dl_dyn_param)
	FIELD(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	FIELD(	rulecolorbg, unused result of color rule execution, edd_none, dl_color_index)
	FIELD(	chan, channel for dynamics, edd_text_edit, dl_pc_index)
	FIELD(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	FIELD(	clrrule, color rule name, edd_text_edit, dl_string_index)
	FIELD(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
ENDSTRUCTURE
STRUCTURE(dl_dyn_attr)
	FIELD(	mod, dynamic modifiers, edd_subform, dl_struct,,dl_dyn_mod)
	FIELD(	param, dynamic parameters, edd_subform, dl_struct,,dl_dyn_param)
ENDSTRUCTURE


COMMENT
	color map entry- see color map header element below
ENDCOMMENT
STRUCTURE(dl_color)
	FIELD(	r, red intensity, edd_valuator,	dl_uchar,,edd_intensity)
	FIELD(	g, green intensity, edd_valuator,dl_uchar,,edd_intensity)
	FIELD(	b, blue intensity, edd_valuator,dl_uchar,,edd_intensity)
	FIELD(	inten,  greyscale intensity, edd_valuator,dl_uchar,,edd_intensity)
	FIELD(	blink,  blink enable, edd_button,dl_ushort,,dl_on_off)
	FIELD(	pad, available, edd_none, dl_short)
ENDSTRUCTURE

STRUCTURE(dl_color_rule_info)
	FIELD(	chan, channel, edd_text_edit, dl_string_index)
	FIELD(	value, value, edd_text_edit, dl_string_index)
	FIELD(	connector, logical connector, edd_menu, dl_uchar, , dl_color_rule_connector)
	FIELD(	clr, color, edd_palette,  dl_color_index)
	FIELD(	pad, available, edd_none, dl_short)
ENDSTRUCTURE

COMMENT
	color rule entry- see color rules header element below
ENDCOMMENT
STRUCTURE(dl_color_rule)
	FIELD(	name, rule name, edd_text_edit, dl_string_index)
	FIELD(	info, rule information, edd_none, dl_struct, MAX_RULE_LINES, dl_color_rule_info)
	FIELD(	fg_enable,  foreground enable, edd_button,dl_ushort,,dl_on_off)
	FIELD(	bg_enable,  background enable, edd_button,dl_ushort,,dl_on_off)
	FIELD(	default_fg, foreground color, edd_palette,  dl_color_index)
	FIELD(	default_bg, background color, edd_palette,  dl_color_index)
ENDSTRUCTURE

STRUCTURE(dl_pen)
	FIELD(	chan, pen channel, edd_text_edit, dl_pc_index)
	FIELD(	clr, pen color, edd_palette,  dl_color_index)
	FIELD(	pad1, available, 	edd_none, 	dl_char)
	FIELD(	pad2, available, 	edd_none, 	dl_char)
	FIELD(	pad3, available, 	edd_none, 	dl_char)
ENDSTRUCTURE

STRUCTURE(dl_trace)
	FIELD(	xdata, 	x data channel, edd_text_edit,	dl_pc_index)
	FIELD(	ydata, 	y data channel, edd_text_edit,	dl_pc_index)
	FIELD(	data_clr, data color, 	edd_palette, 	dl_color_index)
	FIELD(	pad1, available, 	edd_none, 	dl_char)
	FIELD(	pad2, available, 	edd_none, 	dl_char)
	FIELD(	pad3, available, 	edd_none, 	dl_char)
ENDSTRUCTURE

COMMENT
	related display name and args
ENDCOMMENT
STRUCTURE(dl_rd_info)
	FIELD(	label, menu label, edd_text_edit, dl_string_index)
	FIELD(	name, display name, edd_text_edit, dl_string_index)
	FIELD(	args, display arguments, edd_text_edit, dl_string_index)
ENDSTRUCTURE

COMMENT
	execute script info
ENDCOMMENT
STRUCTURE(dl_es_info)
	FIELD(	label, menu label, edd_text_edit, dl_string_index)
	FIELD(	path, script path, edd_text_edit, dl_string_index)
	FIELD(	args, script arguments, edd_text_edit, dl_string_index)
ENDSTRUCTURE

COMMENT
	The display list header (the `magic' structure memeber should be last)
ENDCOMMENT
STRUCTURE(dl_hdr)
	FIELD(magic, display list magic value, 	edd_none,	dl_uint)
	FIELD(majv, display list major version, edd_none,	dl_ushort)
	FIELD(mnrv, display list minor version, edd_none,	dl_ushort)
	FIELD(ndyn, number of dynamic graphical, edd_none,	dl_ushort)
	FIELD(npc, number of process connections, edd_none,	dl_ushort)
	FIELD(nstr, number of strings,	 	edd_none,	dl_ushort)
	FIELD(nmon, number of monitor elements,	edd_none,	dl_ushort)
	FIELD(nctrl, number of control elements,edd_none,	dl_ushort)
	FIELD(nplot, number of plot elements,	edd_none,	dl_ushort)
	FIELD(nrd, number of related display elements,	edd_none, dl_ushort)
	FIELD(nes, number of execute script elements,	edd_none, dl_ushort)
	FIELD(nkd, number of kill display elements,	edd_none, dl_ushort)
	FIELD(object, object header, edd_subform,dl_struct,,dl_object_hdr)
	FIELD(clr, foreground color, 		edd_palette,dl_color_index)
	FIELD(bclr, background color,		edd_palette,	dl_bckgnd_color_index)
	FIELD(nwords_dspy, shrt wds in display, edd_none, 	dl_uint)
	FIELD(nwords_sta, shrt wds in static element portion, edd_none, dl_uint)
	FIELD(nwords_cmap, shrt wds in clr map portion, edd_none, dl_uint)
	FIELD(nwords_crules, shrt wds in clr rules portion, edd_none, dl_uint)
	FIELD(odyn, shrt wd offset to dynamics, edd_none, 	dl_uint)
	FIELD(osta, shrt wd offset to statics, 	edd_none, 	dl_uint)
	FIELD(omon, shrt wd offset to monitors, edd_none, 	dl_uint)
	FIELD(octrl, shrt wd offset to controls, edd_none, 	dl_uint)
	FIELD(oplot, shrt wd offset to plots, 	edd_none, 	dl_uint)
	FIELD(ord, shrt wd offset to related displays, 	edd_none, dl_uint)
	FIELD(oes, shrt wd offset to execute scripts, 	edd_none, dl_uint)
	FIELD(okd, shrt wd offset to kill displays, 	edd_none, dl_uint)
	FIELD(opc, shrt wd offset to proc conns, edd_none, 	dl_uint)
	FIELD(ostr, shrt wd offset to strings, 	edd_none, 	dl_uint)
	FIELD(ocmap, shrt wd offset to clr map, edd_none, 	dl_uint)
	FIELD(ocrules, shrt wd offset to clr rules, edd_none, 	dl_uint)
	FIELD(style, default edge style, edd_button, dl_uchar,, dl_edge_style)
	FIELD(fill, default fill style, edd_button,   dl_uchar,, dl_fill_style)
	FIELD(width, default line width, edd_valuator, dl_uchar,, edd_line_width)
	FIELD(clrmod, default color modifier, edd_menu,dl_uchar,,dl_dynamic_color_mod)
	FIELD(vismod, visibility modifier, edd_menu,dl_uchar,,dl_dynamic_visibility_mod)
	FIELD(pad1, available, edd_none, dl_char)
	FIELD(pad2, available, edd_none, dl_short)
	FIELD(clrrule, default color rule name, edd_text_edit, dl_string_index)
	FIELD(pv, default channel, 	edd_text_edit,	dl_string_index)
	FIELD(cmap, external color palette display name, edd_text_edit,dl_string_index)
ENDSTRUCTURE

COMMENT
	graphic objects

	NOTE: 2nd argument to the object MACRO becomes an entry in the
	object type transform menu for the editor modify form. If the
	2nd argument is nill then editor users cannot transform to this
	type of object from other object types.
ENDCOMMENT
OBJECTLIST(dl_elements)

OBJECT(dl_anygo, <<graphic subclass>>, dl_class_graphic)
	ATTR(	element, element header, edd_none, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, edd_none, dl_struct,,dl_object_hdr)
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
ENDOBJECT
OBJECT(dl_rec, rectangle, dl_class_graphic)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
ENDOBJECT
COMMENT
	An array of point structures are inserted after this structure.
ENDCOMMENT
OBJECT(dl_line, line, dl_class_graphic)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	npoints, n points following header, edd_none, dl_ushort)	
	ATTR(	point, point parameters, edd_none, dl_struct, MAX_NPOINTS, dl_point)
	ATTR(	orgpoint, original point parameters, edd_none, dl_struct, MAX_NPOINTS, dl_point)
ENDOBJECT
OBJECT(dl_oval, oval, dl_class_graphic)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
ENDOBJECT
OBJECT(dl_arc, arc, dl_class_graphic)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	begin,	begin angle (64's of a degree),edd_valuator,dl_short,,edd_begin_angle)
	ATTR(	path, path length angle (64's of a degree),edd_valuator,dl_short,,edd_path_angle)
ENDOBJECT
OBJECT(dl_text, text, dl_class_graphic)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	pad, available, edd_none, dl_short)
	ATTR(	textix, text string, 	edd_text_edit,	dl_string_index)
	ATTR(	align, text alignment, edd_menu,  dl_ushort,,dl_string_alignment)
ENDOBJECT
OBJECT(dl_user, user defined area, dl_class_graphic)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	pad, available, edd_none, dl_short)
	ATTR(	param,	user parameters, edd_text_edit, dl_string_index,10)
ENDOBJECT

COMMENT
	Monitor
ENDCOMMENT
OBJECT(dl_anymo, <<monitor subclass>>, dl_class_monitor)
	ATTR(	element, element header, edd_none, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_none, dl_struct,,dl_object_hdr)
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	monitor, monitor header, edd_none, dl_struct,,dl_monitor_hdr)
ENDOBJECT
OBJECT(dl_bar, bar, dl_class_monitor)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	monitor, monitor header, edd_subform, dl_struct,,dl_monitor_hdr)	
	ATTR(	label,	label type, edd_menu, dl_uchar,, dl_dynamic_label_type)
	ATTR(	clrmod, color modifier, edd_menu, dl_uchar,,dl_dynamic_color_mod)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	ATTR(	direction, direction of increase, edd_menu, dl_uchar,,dl_dynamic_direction)
	ATTR(	fillmod, fill modifier, edd_button, dl_uchar,, dl_dynamic_fill_mod)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
ENDOBJECT
OBJECT(dl_tank, tank, dl_class_monitor)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	monitor, monitor header, edd_subform, dl_struct,,dl_monitor_hdr)
	ATTR(	label,	label type, edd_menu, dl_uchar,, dl_dynamic_label_type)
	ATTR(	clrmod, color modifier, edd_menu, dl_uchar,,dl_dynamic_color_mod)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	ATTR(	fillmod, fill modifier, edd_button, dl_uchar,, dl_dynamic_fill_mod)
	ATTR(	direction, direction of increase, edd_menu, dl_uchar,,dl_dynamic_direction)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
ENDOBJECT
OBJECT(dl_meter, meter, dl_class_monitor)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	monitor, monitor header, edd_subform, dl_struct,,dl_monitor_hdr)	
	ATTR(	label,	label type, edd_menu, dl_uchar,, dl_dynamic_label_type)
	ATTR(	clrmod, color modifier, edd_menu, dl_uchar,,dl_dynamic_color_mod)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
ENDOBJECT
OBJECT(dl_indicator, indicator, dl_class_monitor)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	monitor, monitor header, edd_subform, dl_struct,,dl_monitor_hdr)	
	ATTR(	label,	label type, edd_menu, dl_uchar,, dl_dynamic_label_type)
	ATTR(	clrmod, color modifier, edd_menu, dl_uchar,,dl_dynamic_color_mod)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	ATTR(	direction, pointer direction, edd_menu, dl_uchar,,dl_pointer_direction)
	ATTR(	pad, available, edd_none, dl_char)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
ENDOBJECT
OBJECT(dl_text_update, text update, dl_class_monitor)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	monitor, monitor header, edd_subform, dl_struct,,dl_monitor_hdr)	
	ATTR(	clrmod, color modifier, edd_menu, dl_uchar,,dl_dynamic_color_mod)
	ATTR(	pad, available, edd_none, dl_char)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	ATTR(	align, text alignment, edd_menu, dl_uchar,,dl_string_alignment)
	ATTR(	format, text format, edd_menu, dl_uchar,,dl_string_format)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
ENDOBJECT
OBJECT(dl_symbol, user defined symbol, dl_class_monitor)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	monitor, monitor header, edd_subform, dl_struct,,dl_monitor_hdr)
	ATTR(	pad, available, edd_none, dl_short)
	ATTR(	tbd, symbol parameters, edd_text_edit, dl_string_index,10)
ENDOBJECT

COMMENT
	Plots
ENDCOMMENT
OBJECT(dl_anypo, <<plot subclass>>, dl_class_plot)
	ATTR(	element, element header, edd_none, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_none, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	plotcom, common plot parameters, edd_none, dl_struct,,dl_plot_hdr)
ENDOBJECT
COMMENT
	basic cartesian plot
ENDCOMMENT
OBJECT(dl_plot, cartesian plot, dl_class_plot)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	plotcom, common plot parameters, edd_subform, dl_struct,,dl_plot_hdr)
	ATTR(	style, plot style, edd_menu, dl_uchar,,dl_plot_style)
	ATTR(	erase_oldest, erase oldest sample, edd_button, dl_uchar,,dl_on_off)
	ATTR(	count, 	sample save count, edd_text_edit, dl_ushort)
	ATTR(	pad, available, edd_none, dl_ushort)
	ATTR(	trace, trace parameters, edd_subform, dl_struct, 8, dl_trace)
ENDOBJECT
COMMENT
	smith chart overlays for cartesian plot
ENDCOMMENT
OBJECT(dl_smith_overlay, smith overlay plot, dl_class_plot)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	plotcom, common plot parameters, edd_subform, dl_struct,,dl_plot_hdr)
	ATTR(	igrid, impedance grid, edd_button, dl_uchar,, dl_on_off)
	ATTR(	agrid, admittance grid, edd_button, dl_uchar,, dl_on_off)
	ATTR(	igrid_clr, impedance grid color, edd_palette, 	dl_color_index)
	ATTR(	agrid_clr, admittance grid color, edd_palette, 	dl_color_index)
	ATTR(	data_clr, data color, 	edd_palette, 	dl_color_index)
	ATTR(	pad, available, edd_none, dl_char)
	ATTR(	xdata, 	x data channel, edd_text_edit,	dl_pc_index)
	ATTR(	ydata, 	y data channel, edd_text_edit,	dl_pc_index)
ENDOBJECT

COMMENT
	surface plot

	still alot of questions about how surface plots will be attached to
	the data base.
ENDCOMMENT
OBJECT(dl_surface_plot, surface plot, dl_class_plot)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	plotcom, common plot parameters, edd_subform, dl_struct,,dl_plot_hdr)
	ATTR(	pad1, available, edd_none, dl_short)
	ATTR(	data, 	2D Z data channel, edd_text_edit, dl_pc_index)
	ATTR(	data_clr, data color, edd_palette,  dl_color_index)
	ATTR(	pad2, available, edd_none, dl_char)
	ATTR(	dis, viewing distance, edd_text_edit, dl_short)
	ATTR(	xyangle, xy viewing angle, edd_text_edit, dl_short)
	ATTR(	zangle, z viewing angle, edd_text_edit, dl_short)
ENDOBJECT

OBJECT(dl_strip, strip chart, dl_class_plot)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	plotcom, common plot parameters, edd_subform,dl_struct,,dl_plot_hdr)
	ATTR(	delay, delay between samples, edd_text_edit, dl_ushort)
	ATTR(	units, delay units, edd_menu, dl_uchar,, dl_time_units)
	ATTR(	pad1, available, edd_none, dl_uchar)
	ATTR(	pad2, available, edd_none, dl_uchar)
	ATTR(	pad3, available, edd_none, dl_uchar)
	ATTR(	pen, pen parameters, edd_subform, dl_struct, 8, dl_pen)
ENDOBJECT

COMMENT
	Controller
ENDCOMMENT
OBJECT(dl_anyco, <<controller subclass>>, dl_class_control)
	ATTR(	element, element header, edd_none, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_none, dl_struct,,dl_object_hdr)
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	control, control header, edd_none, dl_struct,,dl_control_hdr)
ENDOBJECT
OBJECT(dl_valuator, valuator, dl_class_control)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	control, control header, edd_subform, dl_struct,,dl_control_hdr)
	ATTR(	label, label type, edd_menu, 	dl_uchar,,dl_dynamic_label_type)
	ATTR(	clrmod, color modifier, edd_menu,dl_uchar,,dl_dynamic_color_mod)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	ATTR(	direction, pointer direction, edd_menu, dl_uchar,,dl_pointer_direction)
	ATTR(	pad, available, edd_none, dl_char)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
ENDOBJECT
OBJECT(dl_text_entry, text entry, dl_class_control)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	control, control header, edd_subform, dl_struct,,dl_control_hdr)	
	ATTR(	clrmod, color modifier, edd_menu, dl_uchar,,dl_dynamic_color_mod)
	ATTR(	pad, available, edd_none, dl_char)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
ENDOBJECT
OBJECT(dl_button, button, dl_class_control)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	control, control header, edd_subform, dl_struct,,dl_control_hdr)	
	ATTR(	clrmod, color modifier, edd_menu,dl_uchar,,dl_dynamic_color_mod)
	ATTR(	pad1, available, edd_none, dl_char)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	stacking, stacking modifier, edd_menu,dl_uchar,,dl_button_stacking)
	ATTR(	pad2, available, edd_none, dl_char)
ENDOBJECT
OBJECT(dl_choice_button, choice button, dl_class_control)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	control, control header, edd_subform, dl_struct,,dl_control_hdr)	
	ATTR(	clrmod, color modifier, edd_menu,dl_uchar,,dl_dynamic_color_mod)
	ATTR(	pad1, available, edd_none, dl_char)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	stacking, stacking modifier, edd_menu,dl_uchar,,dl_button_stacking)
	ATTR(	pad2, available, edd_none, dl_char)
ENDOBJECT
OBJECT(dl_message_button, message button, dl_class_control)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	control, control header, edd_subform, dl_struct,,dl_control_hdr)	
	ATTR(	pad1, available, edd_none,dl_ushort)
	ATTR(	press_msg, press message, edd_text_edit, dl_string_index)	
	ATTR(	release_msg, release message, edd_text_edit, dl_string_index)	
	ATTR(	clrmod, color modifier, edd_menu,dl_uchar,,dl_dynamic_color_mod)
	ATTR(	pad2, available, edd_none,dl_uchar)
	ATTR(	pad3, available, edd_none,dl_ushort)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
ENDOBJECT
OBJECT(dl_menu, menu, dl_class_control)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	control, control header, edd_subform, dl_struct,,dl_control_hdr)	
	ATTR(	clrmod, color modifier, edd_menu,dl_uchar,,dl_dynamic_color_mod)
	ATTR(	pad1, available, edd_none, dl_char)
	ATTR(	rulecolorfg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	rulecolorbg, result of color rule execution, edd_none, dl_color_index)
	ATTR(	pad2, available, edd_none,dl_ushort)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
ENDOBJECT



OBJECT(dl_2d_control, 2d control, dl_class_control)
	ATTR(	element, element header, edd_subform, dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform, dl_struct,,dl_object_hdr)	
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	control, control header, edd_subform, dl_struct,,dl_control_hdr)	
	ATTR(	pad, available, edd_none,dl_ushort)
	ATTR(	yctrl, y control channel, edd_text_edit,	dl_pc_index)
	ATTR(	clr, foreground color, 	edd_palette, dl_color_index)
	ATTR(	bclr, background color,	edd_palette, dl_bckgnd_color_index)
	ATTR(	label, label type, edd_menu, dl_uchar,,dl_dynamic_label_type)
	ATTR(	clrmod, color modifier, edd_menu,dl_uchar,,dl_dynamic_color_mod)
	ATTR(	clrrule, color rule name, edd_text_edit, dl_string_index)
	ATTR(	clrargs, color rule arguments, edd_text_edit, dl_string_index)
	ATTR(	rulechan, channels for dynamic color rules, edd_none, dl_pc_index, MAX_RULE_LINES)
ENDOBJECT

OBJECT(dl_related_display, related display, dl_class_related_display)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	selfkill, self kill, edd_button, dl_uchar,, dl_on_off)
	ATTR(	pad, available, edd_none,dl_uchar)
	ATTR(	display, display info, edd_subform,dl_struct,8,dl_rd_info)
	ATTR(	clr, foreground color, 	edd_palette,	dl_color_index)
	ATTR(	bclr, background color,	edd_palette,	dl_bckgnd_color_index)
ENDOBJECT

OBJECT(dl_execute_script, execute script, dl_class_execute_script)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	pad, available, edd_none,dl_ushort)
	ATTR(	script, script info, edd_subform,dl_struct,8,dl_es_info)
	ATTR(	clr, foreground color, 	edd_palette,	dl_color_index)
	ATTR(	bclr, background color,	edd_palette,	dl_bckgnd_color_index)
ENDOBJECT

OBJECT(dl_kill_display, kill display, dl_class_kill_display)
	ATTR(	element, element header, edd_subform,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_subform,dl_struct,,dl_object_hdr)
	ATTR(	groupid, id if in a group, edd_none, dl_short,)
	ATTR(	pad, available, edd_none,dl_ushort)
	ATTR(	label, button label, edd_text_edit, dl_string_index)	
	ATTR(	clr, foreground color, 	edd_palette,	dl_color_index)
	ATTR(	bclr, background color,	edd_palette,	dl_bckgnd_color_index)
ENDOBJECT

OBJECT(dl_group, <<internal grouping>>, dl_class_internal)
	ATTR(	element, element header, edd_none,dl_struct,,dl_element_hdr)
	ATTR(	object, object header, 	edd_none,dl_struct,,dl_object_hdr)	
ENDOBJECT

COMMENT
	NULL terminated string inserted after this structure.
	String length is determined from nwords. In some cases 
	there will be two NULLs at the end of the string for
	alignment purposes.
ENDCOMMENT
OBJECT(dl_pc, <<PV name element>>, dl_class_internal)
	ATTR(	element, element header, edd_none,dl_struct,,dl_element_hdr)
	ATTR(	count, n dynamic elements on this conn, edd_none, dl_ushort)	
ENDOBJECT

COMMENT
	NULL terminated string inserted after this structure.
	String length is determined from nwords. In some cases 
	there will be two NULLs	at the end of the string for
	alignment purposes.
ENDCOMMENT
OBJECT(dl_string, <<string name element>>, dl_class_internal)
	ATTR(	element, element header, edd_none,dl_struct,,dl_element_hdr)
	ATTR(	count, n references to this string, edd_none, dl_ushort)	
ENDOBJECT


COMMENT
	The color table header 

	n copies of struct dl_color following
ENDCOMMENT
OBJECT(dl_color_map, <<color map>>, dl_class_internal)
	ATTR(	element, element header, edd_none,dl_struct,,dl_element_hdr)
	ATTR(	ncolors, n colors following header, edd_none, dl_ushort)	
ENDOBJECT

COMMENT
	The color rule table header 

	n copies of struct dl_color_rule following
ENDCOMMENT
OBJECT(dl_color_rule_table, <<color rules>>, dl_class_internal)
	ATTR(	element, element header, edd_none,dl_struct,,dl_element_hdr)
	ATTR(	nrules, n color rules following header, edd_none, dl_ushort)
	ATTR(	pad, available, edd_none, dl_ushort)
ENDOBJECT

OBJECT(dl_basic_attribute, <<basic atribute>>, dl_class_attribute)
	ATTR(	element, element header, edd_none,dl_struct,,dl_element_hdr)
	ATTR(	attr, static attributes, edd_none,dl_struct,,dl_sta_attr)
ENDOBJECT

OBJECT(dl_dynamic_attribute, <<dynamic attribute>>, dl_class_attribute)
	ATTR(	element, element header, edd_none,dl_struct,,dl_element_hdr)
	ATTR(	attr, dynamic attributes, edd_none,dl_struct,,dl_dyn_attr)
ENDOBJECT

ENDOBJECTLIST
