/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/******************************************************************************
 *                                                                            *
 *      path:  ca_strip.c                                                     *
 *                                                                            *
 ******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#ifdef WIN32
#include <winsock.h>

#define HUGE HUGE_VAL
#else
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#endif WIN32
#include "cadef.h"
#include "db_access.h"
#include <math.h>

#include "ca_strip.h"

static void caConnCB (struct connection_handler_args args);
static void rateConnCB (struct connection_handler_args args);
static void rateReadCB (struct event_handler_args args);
static void caReadCB (struct event_handler_args args);
static void caMonCB (struct event_handler_args args);

extern int caUpdateReq;
extern char *mktemp (char *);

#define CS_CONN	1
#define CS_DISC	2


#ifdef DEBUG_PROGRAM
double dataBuffer [512];

main ()
{
    strip_channel *strip;

    ca_task_initialize ();

    strip = OpenStripChannel ("tampa_wave", 512, 5, dataBuffer);
    while (1)
	ca_pend_io (5.0);
    StopMonitor (strip);
}
#endif DEBUG_PROGRAM

strip_channel *OpenStripChannel (char *ca_name, int nsamples, int samplerate, double *dataBuffer)
{
    strip_channel *sp;
    int i;

    if (sp = calloc (1, sizeof(strip_channel)))
    {
	strncpy (sp->name, ca_name, sizeof(sp->name));
	sp->nsamples = nsamples;
	sp->samplerate = samplerate;
	sp->samples = dataBuffer;
	for (i = 0; i < nsamples; i++)
	    sp->samples[i] = HUGE;
	sp->sampletime.secPastEpoch = (time(0) - TS_EPOCH_SEC_PAST_1970)
					 - ((sp->nsamples - 1) * sp->samplerate);
	sp->sampletime.nsec = 0;

	/* open channel */
	ca_build_and_connect (sp->name, TYPENOTCONN, 0, &sp->caChid, NULL, caConnCB, sp);
    }
    return sp;
}

static void caConnCB (struct connection_handler_args args)
{
    strip_channel *sp;
    char ratename [30];
    int i;
    char *cp;
    TS_STAMP new_stamp;
    int shiftcnt;

    sp = ca_puser(args.chid);

    /* if connected */
    if (args.op == CA_OP_CONN_UP)
    {
	if (ca_element_count (args.chid) > 1)
	{
	    /* get rate for array values */
	    strncpy (ratename, sp->name, sizeof(ratename));
	    if ((cp = strchr (ratename, '.')) != NULL)
		*cp = 0;
	    strcat (ratename, ".SCAN");
	    ca_build_and_connect (ratename, TYPENOTCONN, 0, &sp->rateChid, NULL, rateConnCB, sp);
	}
	else
	{
	    /* start monitor for values. */
	    ca_add_masked_array_event (OLD_DBR_TIME_FLOAT, 1, sp->caChid, caMonCB,
				       sp, (float)0., (float)0., (float)0., &sp->evId, DBE_VALUE | DBE_ALARM);
	}
	sp->chid_status = CS_CONN;
    }
    else /* disconnected */
    {
	/* cancel monitors */
	if (sp->evId)
	    ca_clear_event (sp->evId);

	/* shift, and fill samples up to now */
	new_stamp.secPastEpoch = (time(0) - TS_EPOCH_SEC_PAST_1970)
				 - ((sp->nsamples - 1) * sp->samplerate);
	shiftcnt = (new_stamp.secPastEpoch - sp->sampletime.secPastEpoch) / sp->samplerate;
	if (shiftcnt > sp->nsamples)
	{
	    /* fill entire record using newest value */
	    for (i = 0; i < sp->nsamples-1; i++)
		sp->samples[i] = sp->samples[sp->nsamples-1];
	}
	else
	{
	    /* shift data down, and fill */
	    for (i = 0; i < (sp->nsamples - shiftcnt); i++)
		sp->samples[i] = sp->samples[shiftcnt+i];
	    for (;i < sp->nsamples-1; i++)
		sp->samples[i] = sp->samples[sp->nsamples-1];
	 }
	sp->sampletime = new_stamp;

	/* start reporting HUGE values */
	sp->samples[sp->nsamples-1] = HUGE;
	sp->chid_status = CS_DISC;
    }
}

static void rateConnCB (struct connection_handler_args args)
{
    strip_channel *sp;

    sp = ca_puser(args.chid);

    /* if connected */
    if (args.op == CA_OP_CONN_UP)
    {
	ca_array_get_callback (OLD_DBR_SHORT, 1, sp->rateChid, rateReadCB, sp);
    }
    else /* disconnected */
    {

    }
}

static void rateReadCB (struct event_handler_args args)
{
    strip_channel *sp;

    /* get value in and close rate channel */
    /* values based on choiceGbl.ascii
	1	"60 second"
	2	"30 second"
	3	"15 second"
	4	"10 second"
	5	"5 second"
	6	"2 second"
	7	"1 second"
	8	"500 ms"
	9	"200 ms"
	10	"100 ms"
    */
    sp = ca_puser(args.chid);
    switch (*(short *)args.dbr)
    {
    case 1:
	sp->chanrate = 60 * 1000;
	break;
    case 2:
	sp->chanrate = 30 * 1000;
	break;
    case 3:
	sp->chanrate = 15 * 1000;
	break;
    case 4:
	sp->chanrate = 10 * 1000;
	break;
    case 5:
	sp->chanrate = 5 * 1000;
	break;
    case 6:
	sp->chanrate = 2 * 1000;
	break;
    case 7:
	sp->chanrate = 1 * 1000;
	break;
    case 8:
	sp->chanrate = 500;
	break;
    case 9:
	sp->chanrate = 200;
	break;
    case 10:
	sp->chanrate = 100;
	break;
    default:
	sp->chanrate = 0;
	break;
    }
    ca_clear_channel (sp->rateChid);

    if (sp->chanrate)
    {
	/* read first batch of values */
	ca_array_get_callback (OLD_DBR_TIME_FLOAT, ca_element_count (sp->caChid),
			       sp->caChid, caReadCB, sp);
    }
    else
    {
	/* start monitor for values. */
	ca_add_masked_array_event (OLD_DBR_TIME_FLOAT, 1, sp->caChid, caMonCB,
				   sp, (float)0., (float)0., (float)0., &sp->evId, DBE_VALUE | DBE_ALARM);
    }
}

static void caReadCB (struct event_handler_args args)
{
    strip_channel *sp;
    TS_STAMP new_stamp;
    float *argvals;
    int i;
    int j;
    long samptime;

    sp = ca_puser(args.chid);
    argvals = &(((struct dbr_time_float *)args.dbr)->value);

    /* shift, and fill samples up to now */
    CAStripSync (sp);

    /* now introduce values from report */
    new_stamp.nsec = ((struct dbr_time_float *)args.dbr)->stamp.nsec;
    new_stamp.secPastEpoch = ((struct dbr_time_float *)args.dbr)->stamp.secPastEpoch
				- ((args.count - 1) * (sp->chanrate/1000));
    samptime = new_stamp.secPastEpoch;

    for (i = 0; i < args.count; i++)
    {
	j = (samptime - sp->sampletime.secPastEpoch) / sp->samplerate;
	if (j < 0)
	    j = 0;
	if (j < sp->nsamples)
	    sp->samples[j] = argvals[i];
	samptime += (sp->chanrate / 1000);
    }
    for (j++; (0 <= j) && (j < sp->nsamples); j++)
	sp->samples[j] = argvals[i-1];

    /* begin monitoring single values */
    ca_add_masked_array_event (OLD_DBR_TIME_FLOAT, 1, sp->caChid, caMonCB, sp,
				(float)0., (float)0., (float)0., &sp->evId, DBE_VALUE | DBE_ALARM);
}

static void caMonCB (struct event_handler_args args)
{
    strip_channel *sp;
    int i;
    long samptime;
    TS_STAMP new_stamp;

    sp = ca_puser(args.chid);

    /* shift, and fill samples up to now */
    CAStripSync (sp);

    new_stamp.nsec = ((struct dbr_time_float *)args.dbr)->stamp.nsec;
    new_stamp.secPastEpoch = ((struct dbr_time_float *)args.dbr)->stamp.secPastEpoch
				- ((args.count - 1) * (sp->chanrate/1000));
    samptime = new_stamp.secPastEpoch;

    /* install new value(s) into array */
    for (i = (samptime - sp->sampletime.secPastEpoch) / sp->samplerate;
	 (i < sp->nsamples);
	 i++)
    {
	if (i < 0)
	    i = 0; 
	sp->samples[i]  = (((struct dbr_time_float *)args.dbr)->value);
    }
}

void CAStripSync (strip_channel *sp)
{
    TS_STAMP new_stamp;
    int i;
    int shiftcnt;

    /* shift, and fill samples up to now */
    new_stamp.secPastEpoch = (time(0) - TS_EPOCH_SEC_PAST_1970)
				 - ((sp->nsamples - 1) * sp->samplerate);
    shiftcnt = (new_stamp.secPastEpoch - sp->sampletime.secPastEpoch) / sp->samplerate;
    if (shiftcnt > sp->nsamples)
    {
	/* fill entire record using newest value */
	for (i = 0; i < sp->nsamples-1; i++)
	    sp->samples[i] = sp->samples[sp->nsamples-1];
    }
    else if (shiftcnt > 0)
    {
	/* shift data down, and fill */
	for (i = 0; i < (sp->nsamples - shiftcnt); i++)
	    sp->samples[i] = sp->samples[shiftcnt+i];
	for (;i < sp->nsamples-1; i++)
	    sp->samples[i] = sp->samples[sp->nsamples-1];
    }
    sp->sampletime.secPastEpoch = sp->sampletime.secPastEpoch + (shiftcnt * sp->samplerate);
}

void StopMonitor (strip_channel *sp)
{
    /* cancel chid, monitors and all */
    if (sp->evId)
	ca_clear_event (sp->evId);
    sp->evId = 0;
    if (sp->caChid)
	ca_clear_channel (sp->caChid);
    sp->caChid = 0;
    if (sp->rateChid)
	ca_clear_channel (sp->rateChid);
    sp->rateChid = 0;
    free (sp);
}

