/*controllers.c,v 1.1.1.1 1999/08/03 14:43:47 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
#undef DEBUG

#ifdef SOLARIS
#include <stdio.h>
#endif
#include <Xlib.h>      			/* the X library 		*/
#ifdef WIN32
#include <X11/XlibXtra.h>
#include <winsock.h>
#else
#include <syslog.h>
#endif WIN32
#include <Xutil.h>
#include <db_access.h>			/* db types and structs		*/
#include <alarm.h>
#include <cadef.h>
#include <dldef.h>			/* display list defines 	*/
#include <dlinfo.h>			/* more display list defines 	*/
#include <font_table.h>			/* defines for font table	*/
#include <dl_context.h>			/* defines dl context		*/
#include <gizmo.h>			/* defs for gizmos */
#include <dynamic_def.h>		/* rti dynamic defines		*/
#include <op_gsdl_def.h>		/* com_table element def	*/
#include <message_butt.h>		/* defines for type of message button */

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

extern void *tisMalloc();
extern void *tisCalloc();
extern void tisFree();

/* mapping for button orientation from display list to gizmo enums */
unsigned short convert_button_stack[3] = {
	gz_stack_column,
	gz_stack_row,
	gz_stack_row_column
};
/* mapping for valuator pointer direction from display list to gizmo enums */
unsigned short convert_ptr_direc[4] = {
	gz_point_down,
	gz_point_up,
	gz_point_right,
	gz_point_left
};

int	okToChangeValue();
void	logChangedValue();

int	inNormalExit = 0;

extern	DATA4	*launchData;



/***********************************************************************
**
**	CONTROLLERS.C
**	
**	Deb Kerstiens 10/90
**
**
**	Routines for setup and updates of contoller-type objects: gizmos 
**		including menus, buttons, message buttons, text entry, 
**		valuators.  Routines for doing the channel access to send
**		values received by operator input via the gizmos : 
**		send_float, send_long, send_string, send_message.  
**
**	Modification History:
**
**	09/19/91	DK	added value_in_range() to verify values in 
**				range in update_text_entry
**	09/24/91	DK	added dmDisplayPvErr(), logic to replace draw of
**				white rectangle
**	09/30/91	DK	added save of current value so can display "last
**				val" when error occurs
**	10/23/91	DK	took out clamp on ctrl limits until handled properly
**				in the database (value_in_range())
**	10/24/91	DK	changed format from g to f in send_str()
**	11/14/91	DK	fixed bug in value_in_range() to make doubles
**				work
**
**********************************************************************/

/***********************************************************************
**
**	update_controllers
**	
**	Deb Kerstiens 10/90
**
**	Purpose : This routine sets up the "gizmos" that are used for allowing
**			operator control via output channels (process variables 
**			in the database.)  
**			When a connection has been successfully made for the 
**			process variable referenced for a given controller-
**			type object, the appropriate gizmo will be set up 
**			for operator input as well as the event routine that 
**			will handle the input from that time on.
**			When a connection doesn't exist for that process
**			variable, a filled white rectangle will indicated this
**			to the operator.  If a connection is lost, the gizmo
**			will be inactive and will be restarted if the connection
**			is restored.
**			If the process variable to which a controller is attached 
**			has been modified, the controller value will be updated to
**			reflect the change.
**			
**			 
**	
**	Note: 	Since the display manager has no easy way of knowing when a 
**		connection to a process variable is completed or when it is 
**		reestablished after being disconnected,etc., this routine is
**		executed periodically to check all controller-type objects
**		and update the display as necessary.
**	
**	Input :
**	    pdl_context : pointer to Display context structure 
**	    pdl_index	: prt to display list color map
**	    pdl		: ptr to display list info for controllers
**	    pdyn	: ptr to the run-time list for controllers
**	    nctrl	: count of controller elements to display
**	Output :
**	    none
**	Returns
**	    none
**
**********************************************************************/



void update_controllers(pdlc, pdl_index, pdl, prti, nctrl,prec,pDm_llist_el,expose_flag)
#ifndef AXP
int				*pdl_index;
#else
unsigned long			*pdl_index;
#endif
struct dl_context		*pdlc;
void				*pdl;
struct rti_ctrl_info 		*prti;
unsigned short			nctrl;
XRectangle			*prec;
struct dm_llist_el		*pDm_llist_el;
int				expose_flag;

{
 void *dis_ptr;
  unsigned short i;
  struct rti_ctrl_info *prti_ctrl; 
  float fval;
  short pv_err;			/* indicates no ca cnxn or other error */
  short geom_ftype;		/* native field type of chans*/
  int 	*pset_modgeom;		/* ptr to set modified array */
  void send_float();
#ifndef AXP
  unsigned fg,bg;
#else
  unsigned long fg,bg;
#endif

#ifdef DEBUG
printf("~in update_controllers\n");
#endif

/* for each element on the display list */
for (	dis_ptr = pdl, prti_ctrl = prti, i = 0;  
	i < nctrl; 
	dis_ptr = (void *)((short *)dis_ptr + ((struct dl_element_hdr *)dis_ptr)->nwords),   
	prti_ctrl++, i++){
    register struct dl_object_hdr *objptr = 
		&(((struct dl_anygo *)dis_ptr)->object);

    if(((struct dl_element_hdr *)dis_ptr)->type ==dl_related_display ||
       ((struct dl_element_hdr *)dis_ptr)->type ==dl_execute_script  ||
       ((struct dl_element_hdr *)dis_ptr)->type ==dl_kill_display){
    	continue;
    }

    pset_modgeom = (prti_ctrl->pgeom_comel->pMod_bits);
    pv_err=dmChkPvStat(prti_ctrl->pgeom_comel, prti_ctrl->init_flag, prti_ctrl->not_conn);
  
    /* check for valid ca connection-- not valid 
	if marked as not connected or no storage allocated for data */ 
/*    not_conn = (((prti_ctrl->pgeom_comel->pChan ) && 
	    ((geom_ftype = ca_field_type(prti_ctrl->pgeom_comel->pChan))
		== TYPENOTCONN))
	    ||(!prti_ctrl->pgeom_comel->valid_data));
*/

    /* if expose, update exposed area */
    if(expose_flag){
    	if( (objptr->x+objptr->width) < prec->x ) {
	    continue;
	}
    	if( objptr->x > (prec->x + (unsigned int)(prec->width)) ) {
	    continue;
	}
    	if( (objptr->y+objptr->height) < prec->y ) {
	    continue;
	}
   	if( objptr->y > (prec->y + (unsigned int)(prec->height)) ) {
	    continue;
	}
	/* if we can't display a real value for this object, display error state*/
	if(pv_err){
	    dmDisplayPvErr(objptr,pv_err, prti_ctrl->pgeom_comel, pdlc, prti_ctrl->pLast_val);
	    prti_ctrl->init_flag = FALSE;
	}
	    
    }

    /* else this is not an expose, check for modification */
    else{
        if(pv_err){
            /* if the error state is the same as last time, no graphics */
    	    if(prti_ctrl->not_conn && (prti_ctrl->not_conn == pv_err))
		continue;
	    else {
		/* display error condition */
		dmDisplayPvErr(objptr, pv_err, prti_ctrl->pgeom_comel, pdlc, prti_ctrl->pLast_val);
	    	/* clear existing init */
		prti_ctrl->init_flag = FALSE;
	   	if(prti_ctrl->pgz){
		    free_gizmo(prti_ctrl->pgz);
	    	    prti_ctrl->pgz = NULL;
		    if(prti_ctrl->pExt &&
		        /* message buttons get strings at init not from db so don't free */
			(((struct dl_element_hdr *)dis_ptr)->type !=dl_message_button)) {
			tisFree(prti_ctrl->pExt);
			prti_ctrl->pExt = NULL;
		    }	
		}
	    }
        }
        /* we have connection and data area */
    	else{
	    /* check for modifications */
   	    /* if the dynamic chan has not been modified and no color rule, no graphics */
	    if(!(pset_modgeom[prti_ctrl->geom_intix] & ~clear_mod_bit_mask[prti_ctrl->geom_modix]) &&
	        !prti_ctrl->clrrule_has_changed){
	    	continue;
	    }
    	}
    } /* end if not expose */

    if(pv_err==dmConnection_okay){
    /* if we made it here, we need to update the display --
	either exposed or modified */
	if (prti_ctrl->clrrule_has_changed)
		expose_flag = 1;
	prti_ctrl->clrrule_has_changed = 0;
#ifdef DEBUG
printf("~c clearing (%s) at:0x%x\n", ca_name(prti_ctrl->pgeom_comel->pChan), &prti_ctrl->clrrule_has_changed);
#endif
	
    geom_ftype = ca_field_type(prti_ctrl->pgeom_comel->pChan);
  
    switch(((struct dl_element_hdr *)dis_ptr)->type){

	case dl_valuator : {

	    /* update the screen, setting up valuator if necessary */
	    update_valuator(pdlc, objptr, prti_ctrl, &pv_err, geom_ftype, expose_flag, pdl_index);
	    break;
	
	} /* end case dl_valuator */


	case dl_button :
	case dl_choice_button :{

	    /* update the screen, setting up button if necessary */
	    update_button(pdlc, objptr, prti_ctrl, &pv_err, geom_ftype, expose_flag, pdl_index);
	    break;

	}/* end dl_button */


	case dl_message_button :{
	    /* this button is one or two state and provides action on press only or press/release*/
	    /* update the screen, setting up message_button if necessary */
	    update_message_button(pdlc, objptr, prti_ctrl, &pv_err, geom_ftype, expose_flag, pdl_index);
	    break;

	}/* end dl_message_button */



	case dl_menu :{

	    /* update the screen, setting up menu if necessary */
	    update_menu(pdlc, objptr, prti_ctrl, &pv_err, geom_ftype, expose_flag, pdl_index);
	    break;

	}/* end dl_menu */



	case dl_text_entry :{

	    /* update the screen, setting up menu if necessary */
	    update_text_entry(pdlc, objptr, prti_ctrl, &pv_err, geom_ftype, expose_flag, pdl_index);
	    break;

	}/* end dl_text_entry */



	default : {
	    /* should return error or abort ??*/
	    printf("~in setup_controllers--took default \n");
	    break;
	}				
		
    }/* end switch */
    }/* end if we need to update display */

    /* clear modified bit */
	if (pset_modgeom)
	    pset_modgeom[prti_ctrl->geom_intix] = 
			pset_modgeom[prti_ctrl->geom_intix] & clear_mod_bit_mask[prti_ctrl->geom_modix];

    /* reset connection flag */
    prti_ctrl->not_conn=pv_err;

  } /* end for each el on list*/

  XFlush(pdlc->d);
  
 
}/* end setup_controllers */



/************************************************************************************************
*
*	permissive_test
*
*
*	permissive_test checks whether the operator has the permission to change the value
*	of the channel defined at pComel.
*
*	NOTE: this is called indirectly by the controller gizmo
*   
*	Input	: pComel	ptr to the channels com record
*		  pg		ptr to gizmo
*	Returns : TRUE if permission granted to change the value, else FALSE
*
*/
int permissive_test(pComel, pg, informOnError)
struct gsdl_com_record	*pComel;
gizmo			*pg;
int			informOnError;
{
	return okToChangeValue(pg->pgg->s, ca_name(pComel->pChan), informOnError);
}




get_ctrl_hoprnlopr(pcom_el, field_type, pHopr, pLopr, pPrec)
struct gsdl_com_record	*pcom_el;
short					field_type;
float					*pHopr;
float					*pLopr;
unsigned short			*pPrec;
	{
	switch (field_type)
	  {
	  case DBR_FLOAT:
		*pHopr = ((struct dbr_ctrl_float *)pcom_el->pData_struct)->upper_ctrl_limit;
		*pLopr = ((struct dbr_ctrl_float *)pcom_el->pData_struct)->lower_ctrl_limit;
		*pPrec = ((struct dbr_ctrl_float *)pcom_el->pData_struct)->precision;
		break;
	  case DBR_DOUBLE:
		*pHopr = ((struct dbr_ctrl_double *)pcom_el->pData_struct)->upper_ctrl_limit;
		*pLopr = ((struct dbr_ctrl_double *)pcom_el->pData_struct)->lower_ctrl_limit;
		*pPrec = ((struct dbr_ctrl_double *)pcom_el->pData_struct)->precision;
		break;
	  case DBR_ENUM:
		*pLopr = 0.0;
		*pHopr = (float)(((struct dbr_ctrl_enum *)pcom_el->pData_struct)->no_str) - 1.0;	/* high is number of enum values - 1 */
		*pPrec = 0;
		break;
	  case DBR_LONG:
		*pHopr = ((struct dbr_ctrl_long *)pcom_el->pData_struct)->upper_ctrl_limit;
		*pLopr = ((struct dbr_ctrl_long *)pcom_el->pData_struct)->lower_ctrl_limit;
		*pPrec = 0;
		break;
	  case DBR_SHORT:
		*pHopr = ((struct dbr_ctrl_int *)pcom_el->pData_struct)->upper_ctrl_limit;
		*pLopr = ((struct dbr_ctrl_int *)pcom_el->pData_struct)->lower_ctrl_limit;
		*pPrec = 0;
		break;
	  case DBR_CHAR:
		*pHopr = ((struct dbr_ctrl_char *)pcom_el->pData_struct)->upper_ctrl_limit;
		*pLopr = ((struct dbr_ctrl_char *)pcom_el->pData_struct)->lower_ctrl_limit;
		*pPrec = 0;
		break;
	  }			
	}

/************************************************************************************************
*
*	update_valuator
*
*	DK 12/90
*
*	Update_valuator handles the initial setup for a valuator gizmo (once a connection 
*		is established.)  From then on it checks for lost or resumed connection
*		to update/init appropriately and/or changes the color as necessary.
*   
*	Input	: pdlc		ptr to the display list context struct
*		  objptr	ptr to the object struct(ht, wt, x,y) 
*		  prti_ctrl	ptr to run-time struct for the valuator
*		  not_conn	indicates ca connect or not
*		  field_type	native type of process variable controlled
*		  expose_flag	indicates expose event or not
*		  pPv_err	ptr to object's state--based on pv state/type
*	Output	: pPv_err	if db_type = string, set to invalid db_type
*	Returns : nothing
*
*/
update_valuator(pdlc,objptr,prti_ctrl, pPv_err, field_type, expose_flag, pdl_color_map_index)
struct dl_context *pdlc;
XRectangle *objptr;
struct rti_ctrl_info *prti_ctrl;
unsigned short *pPv_err, expose_flag;
short field_type;
int	*pdl_color_map_index;

{
    float cvrt_to_float();
    void send_float();
    register struct dl_valuator *pDl_valuator = 
	((struct dl_valuator *)prti_ctrl->pdl_ctrl);
#ifndef AXP
    unsigned fg,bg;
#else
    unsigned long fg,bg;
#endif
    unsigned short prec; 
    float fval;

    /* if illegal type for valuator object */
    if(field_type == DBR_STRING){
	/* first time thru or expose */
	if(expose_flag || !prti_ctrl->init_flag){
	    /* do something here to indicate icky-poo */
	    *pPv_err = dmInvalid_dbtype;
	    dmDisplayPvErr(objptr, *pPv_err, prti_ctrl->pgeom_comel, pdlc, prti_ctrl->pLast_val);
	    prti_ctrl->init_flag = TRUE;
	    prti_ctrl->pgz = NULL;
	}
	return;
    }


    fval = cvrt_to_float(prti_ctrl->pgeom_comel,field_type);
		
    /* first time thru or expose */
    /* new connection , set up gizmo */
    if (!prti_ctrl->init_flag) {
	/* if valid ptr to data from db, and we have valid data (sequencing implies we've seen the limits) */
	if(prti_ctrl->pgeom_comel->pData_struct != NULL && prti_ctrl->pgeom_comel->valid_data){
	    get_ctrl_hoprnlopr(prti_ctrl->pgeom_comel, field_type, &(prti_ctrl->hopr), &(prti_ctrl->lopr), &prec);
	    /* must have limits for valuator*/
		/*NOTE***** This is a kludge until the database routines handle this better */
		/* create a range so the user can see that something is going on */
	    if(fval > prti_ctrl->hopr || fval < prti_ctrl->lopr){
		if(fval > 100.0 || fval < -100.0){
		    prti_ctrl->hopr = prti_ctrl->lopr = 0.0;
		}
		else{
		    prti_ctrl->hopr = 100.0;
		    prti_ctrl->lopr = (fval < 0.0)? -100.0 : 0.0;
		}
	    }
	    else if(prti_ctrl->hopr == 0.0 && prti_ctrl->lopr == 0.0){
		prti_ctrl->hopr = 100.0;
	    }
	    if(prti_ctrl->hopr > prti_ctrl->lopr ){
		prti_ctrl->pgz = create_valuator_gizmo(
			DefaultScreenOfDisplay(pdlc->d),
			pdlc->w,
			objptr, 
			permissive_test, 
			send_float,
			prti_ctrl->pgeom_comel,
			prti_ctrl->lopr, prti_ctrl->hopr,
			convert_ptr_direc[(unsigned short)pDl_valuator->direction],
			0);
		if(prti_ctrl->pgz!= NULL){
		    XMapWindow(prti_ctrl->pgz->pgg->d, prti_ctrl->pgz->w);
		    set_valuator_gizmo_precision(prti_ctrl->pgz, prec);
			set_valuator_gizmo_label_type(prti_ctrl->pgz, pDl_valuator->label);
		    prti_ctrl->init_flag = TRUE;
		}/* end if gz */

	    }/* end if data */

	}

    }/* end if not inited */

    /* if we have a gizmo update, set color */ 
    if(prti_ctrl->pgz != NULL){
	    /* set foreground and background colors in gc*/
  	    if(get_fgbg(pdl_color_map_index, prti_ctrl->pgeom_comel,
		pDl_valuator->control.clr,
		pDl_valuator->control.bclr,
		pDl_valuator->clrmod,
		pDl_valuator->rulecolorfg,
		pDl_valuator->rulecolorbg,
		&fg,&bg) < 0){
		    fg = WhitePixel(pdlc->d, the_screen);
		    bg = BlackPixel(pdlc->d, the_screen);
	    }
	    set_gizmo_color(prti_ctrl->pgz,fg,bg);
	    set_valuator_gizmo(prti_ctrl->pgz, fval);
    }
    /* save current val as last */
    dmSave_float(&prti_ctrl->pLast_val, fval);

}/* end update_valuator */





/************************************************************************************************
*
*	send_float
*
*	DK 12/90
*
*	Send_float sends the value selected using a valuator gizmo to the appropriate process
*		variable.  
*	Input	: pChan		ptr to the CA id block for a process variable
*		  pg		ptr to valuator gizmo struct 
*		  value		value selected using the valuator gizmo
*	Output	: nothing
*	Returns : nothing
*
*/



void send_float (pComel, pg, value)
struct gsdl_com_record *pComel;
gizmo	*pg;
double	value;
{
	float				fval;
	int				status;

	fval = (float)value;
	status = ca_put(DBR_FLOAT, pComel->pChan, &fval);
	ca_flush_io();
	if(status != ECA_NORMAL){
		/* need to do error handling here */
		printf("~CA error returned--couldn't send value %s\n", ca_message(status));
	}
	else if (valuator_direction_changed(pg)){
		char	new_str[50];
		char	old_str[50];
		float	get_valuator_initial_value();

		sprintf( old_str, "%f", get_valuator_initial_value(pg));
		sprintf( new_str, "%f", fval);
		logChangedValue(pg->pgg->s, ca_name(pComel->pChan), old_str, new_str);
	}
}





/************************************************************************************************
*
*	update_menu
*
*	DK 12/90
*
*	Update_menu handles the initial setup for a menu gizmo (once a connection is
*		established.)  From then on it checks for lost or resumed connection
*		to update/init appropriately and/or changes the color as necessary.
*   
*	Input	: pdlc		ptr to the display list context struct
*		  objptr	ptr to the object struct(ht, wt, x,y) 
*		  prti_ctrl	ptr to run-time struct for the menu
*		  pPv_err	ptr to pv/controller state based on connection/type
*		  field_type	native type of process variable controlled
*		  expose_flag	indicates expose event or not
*	Output	: *pPv_err	if type= string, sets to invalid dbtype
*	Returns : nothing
*
*/




update_menu(pdlc, objptr,prti_ctrl, pPv_err, field_type,expose_flag, pdl_color_map_index)
struct dl_context *pdlc;
XRectangle *objptr;
struct rti_ctrl_info *prti_ctrl;
unsigned short *pPv_err, expose_flag;
short field_type;
int	*pdl_color_map_index;

{
#ifndef AXP
    unsigned fg,bg;
#else
    unsigned long fg,bg;
#endif
    float cvrt_to_float();
    void send_long();
    register struct dbr_ctrl_enum *pCtrl_enum;
    register struct dl_menu *pDl_menu= 
	((struct dl_menu *)prti_ctrl->pdl_ctrl);
    /* if illegal type for menu object */
    if(field_type != DBR_ENUM){
	/* first time thru or expose */
	if(expose_flag || !prti_ctrl->init_flag){
	    /* do something here to indicate icky-poo --white for now */
	    *pPv_err = dmInvalid_dbtype;
	    dmDisplayPvErr(objptr, *pPv_err, prti_ctrl->pgeom_comel, pdlc, prti_ctrl->pLast_val);
	    prti_ctrl->init_flag = TRUE;
	    prti_ctrl->pgz = NULL;
	    return;
/*	    printf("Can't create menu for type\n");*/
	}
    }
		
    /* first time thru or expose */
    /* new connection , set up gizmo */
    else if (!prti_ctrl->init_flag) { 
	/* if valid ptr to data from db */
	if(prti_ctrl->pgeom_comel->pData_struct!= NULL){
	    pCtrl_enum = ((struct dbr_ctrl_enum *)prti_ctrl->pgeom_comel->pData_struct);
	    /* no way to tell if channel info is back yet, create
		gizmo when data is there --is this a valid indication?? */
/*	    if(pCtrl_enum->strs[0][0] != '\0'){ */
	    if(prti_ctrl->pgeom_comel->valid_data){
		/* set up table of string ptrs for menu gizmo */
		/* this will be changed to be done in the gz struct extension later */
		prti_ctrl->pExt = (void *)
			tisCalloc(__FILE__, __LINE__, pCtrl_enum->no_str + 1, sizeof(char *));
		if(prti_ctrl->pExt != NULL){
		    register i;
		    /* set up table of string pointers for gizmo in user extension */
		    for ( i = 0; i < pCtrl_enum->no_str; i++){
			((char **)prti_ctrl->pExt)[i] = pCtrl_enum->strs[i];
		    }
		    /* null str to indicate end of string table */
		    ((char **)prti_ctrl->pExt)[i] = NULL;
		    prti_ctrl->pgz = create_menu_gizmo(
			DefaultScreenOfDisplay(pdlc->d), pdlc->w,
			objptr, 
			permissive_test, 
			send_long,
			prti_ctrl->pgeom_comel,
			prti_ctrl->pExt,
			(unsigned)(sizeof(char *) * (pCtrl_enum->no_str+1)));
		    if(prti_ctrl->pgz != NULL){
		    	prti_ctrl->init_flag = TRUE;
		    	XMapWindow(prti_ctrl->pgz->pgg->d, prti_ctrl->pgz->w);
		    }
		}/* end if pExt */
	    }/* end if strs */
/*	    else printf("Took null str case \n"); maybe error to operator */
	}/* end if pData */
    }/* end if not inited */

    /* if we have a gizmo init/update color */ 
    if(prti_ctrl->pgz != NULL){
	pCtrl_enum = ((struct dbr_ctrl_enum *)prti_ctrl->pgeom_comel->pData_struct);
	/* set foreground and background colors in gc*/
  	if(get_fgbg(pdl_color_map_index, prti_ctrl->pgeom_comel,
	    pDl_menu->control.clr,
	    pDl_menu->control.bclr,
	    pDl_menu->clrmod,
	    pDl_menu->rulecolorfg,
	    pDl_menu->rulecolorbg,
	    &fg,&bg) < 0){
		fg = WhitePixel(pdlc->d, the_screen);
		bg = BlackPixel(pdlc->d, the_screen);
	}
	set_gizmo_color(prti_ctrl->pgz,fg,bg);
	if((pCtrl_enum->value >= 0) && (pCtrl_enum->value < pCtrl_enum->no_str)){
	    /* need to indicated error here */
	    set_menu_gizmo(prti_ctrl->pgz, (unsigned)pCtrl_enum->value);
	}
    }/* end gizmo setup */

    /* save current value as last */
    dmSave_float(&prti_ctrl->pLast_val,cvrt_to_float (prti_ctrl->pgeom_comel,field_type));

}/* end update_menu */





/************************************************************************************************
*
*	send_long
*
*	DK 12/90
*
*	Send_long sends the value selected using a menu gizmo to the appropriate process
*		variable.  
*
*	Note:  Right now, a short is sent because long is not working.  It should be 
*		changed when this is fixed.
*
*	Input	: pChan		ptr to the CA id block for a process variable
*		  pg		ptr to menu gizmo struct 
*		  index		index selected using the menu gizmo
*	Output	: nothing
*	Returns : nothing
*
*/


void send_long(pComel, pg, index, pExt)
struct gsdl_com_record *pComel;
gizmo	*pg;
unsigned short index;
void *pExt;

{
	int status;
	short sval;
	char  old_str[50];

	cvrt_to_str(pComel, ca_field_type(pComel->pChan), old_str, dl_format_i);
	sval = (short) index;
	status = ca_put(DBF_SHORT, pComel->pChan, &sval);
	ca_flush_io();
	if(status != ECA_NORMAL){
		/* need to do error handling here */
		printf("~CA error returned--couldn't send value %s\n", ca_message(status));
	}
	else{
		char	*get_choice_gz_string();
		char	*get_menu_gz_string();
		char	*new_str;

		switch (pg->type)
		{
		case gz_choice:	new_str = get_choice_gz_string(pg, index);	break;
		case gz_menu:	new_str = get_menu_gz_string(pg, index);	break;
		default:	abort(); /* mods needed for another type! */	break;
		}
		logChangedValue(pg->pgg->s, ca_name(pComel->pChan), old_str, new_str);
	}
} /* end send_long */





/************************************************************************************************
*
*	update_text_entry
*
*	DK 12/90
*
*	Update_text_entry handles the initial setup for a text_entry gizmo (once a connection 
*		is established.)  From then on it checks for lost or resumed connection
*		to update/init appropriately and/or changes the color as necessary.
*   
*	Input	: pdlc		ptr to the display list context struct
*		  objptr	ptr to the object struct(ht, wt, x,y) 
*		  prti_ctrl	ptr to run-time struct for the text_entry
*		  pPv_err	ptr to pv/controller state based on connection/type
*		  field_type	native type of process variable controlled
*		  expose_flag	indicates expose event or not
*	Output	: nothing
*	Returns : nothing
*
*/


update_text_entry(pdlc,objptr,prti_ctrl, pPv_err, field_type, expose_flag, pdl_color_map_index)
struct dl_context *pdlc;
XRectangle *objptr;
struct rti_ctrl_info *prti_ctrl;
unsigned short *pPv_err, expose_flag;
short field_type;
int	*pdl_color_map_index;

{
#ifndef AXP
    unsigned fg,bg;
#else
    unsigned long fg,bg;
#endif
    void send_string();
    char data_str[MAX_STRING_SIZE];
    register struct dl_text_entry *pDl_text_entry = 
	((struct dl_text_entry *)prti_ctrl->pdl_ctrl);
		
    /* new connection , set up gizmo */
    if (!prti_ctrl->init_flag) { 
	prti_ctrl->pgz = create_textedit_gizmo(
		DefaultScreenOfDisplay(pdlc->d), pdlc->w,
		objptr, 
		permissive_test, 
		send_string,
		prti_ctrl->pgeom_comel,
		MAX_STRING_SIZE, 0);
	if(prti_ctrl->pgz!= NULL){
	    prti_ctrl->init_flag = TRUE;
	    XMapWindow(prti_ctrl->pgz->pgg->d, prti_ctrl->pgz->w);
	}
    }/* end if not inited */

    /* if we have a gizmo, init/update color and value */ 
    if(prti_ctrl->pgz != NULL){
	/* set foreground and background colors in gc*/
  	if(get_fgbg(pdl_color_map_index, prti_ctrl->pgeom_comel,
	    pDl_text_entry->control.clr,
	    pDl_text_entry->control.bclr,
	    pDl_text_entry->clrmod,
	    pDl_text_entry->rulecolorfg,
	    pDl_text_entry->rulecolorbg,
	    &fg,&bg) < 0){
		fg = WhitePixel(pdlc->d, the_screen);
		bg = BlackPixel(pdlc->d, the_screen);
	    }
	set_gizmo_color(prti_ctrl->pgz,fg,bg);
	/* get data*/
	cvrt_to_str(prti_ctrl->pgeom_comel,field_type,data_str, dl_format_f);
	set_textedit_gizmo(prti_ctrl->pgz, data_str);
    }/* end gizmo setup */

    /* save current val as last */
    dmSave_str(&prti_ctrl->pLast_val, data_str);

}/* end update_text_entry */




/************************************************************************************************
*
*	send_string
*
*	DK 12/90
*
*	Send_string sends the value entered using a text entry gizmo to the appropriate
*		 process variable.
*
*	Input	: pChan		ptr to the CA id block for a process variable
*		  pg		ptr to text entry gizmo struct 
*		  string	string entered using the text entry gizmo
*	Output	: nothing
*	Returns : nothing
*
*/ 



void send_string(pComel, pg, string)
struct gsdl_com_record *pComel;
gizmo	*pg;
char *string;

{
	int  status;
	char old_str[50];
/*
	printf("~value is %s\n",string);
*/
	cvrt_to_str(pComel, ca_field_type(pComel->pChan), old_str, dl_format_f);
	if(value_in_range(pComel, string, pg->pgg->s)){
		status = ca_put(DBR_STRING, pComel->pChan, string);
		if(status != ECA_NORMAL){
	    		/* need to do error handling here */
	   		printf("~CA error returned in send_str--couldn't send value %s\n", ca_message(status));
	   		set_textedit_gizmo(pg, old_str);
		}
		else
			logChangedValue(pg->pgg->s, ca_name(pComel->pChan), old_str, string);
	}
} /* end send_string */



/************************************************************************************************
*
*	update_button
*	DK 1/91
*
*	Update_button handles the initial setup for a button gizmo (once a connection is
*		established.)  From then on it checks for lost or resumed connection
*		to update/init appropriately and/or changes the color as necessary.
*   
*	Input	: pdlc		ptr to the display list context struct
*		  objptr	ptr to the object struct(ht, wt, x,y) 
*		  prti_ctrl	ptr to run-time struct for the menu
*		  field_type	native type of process variable controlled
*		  expose_flag	indicates expose event or not
*		  pPv_err	ptr to object's state--based on pv state/type
*	Output	: pPv_err	if db_type != enum, set to invalid db_type
*	Returns : nothing
*
*	Modification History :
*	08/26/91	DK	added check for valid enum value before setting gizmo val
*	09/27/91		DK	added dmDisplayPvErr(), logic to replace draw of
*					 white rectangle
*
*/


update_button(pdlc, objptr,prti_ctrl, pPv_err, field_type,expose_flag, pdl_color_map_index)
struct dl_context *pdlc;
XRectangle *objptr;
struct rti_ctrl_info *prti_ctrl;
unsigned short *pPv_err, expose_flag;
short field_type;
int	*pdl_color_map_index;

{
#ifndef AXP
    unsigned fg,bg;
#else
    unsigned long fg,bg;
#endif
    float cvrt_to_float();
    void send_long();
    register struct dbr_ctrl_enum *pCtrl_enum;
    register struct dl_choice_button *pDl_button= 
	((struct dl_choice_button *)prti_ctrl->pdl_ctrl);
    /* if illegal type for menu object */
    if(field_type != DBR_ENUM){
	/* first time thru or expose */
	if(expose_flag || !prti_ctrl->init_flag){
	    *pPv_err = dmInvalid_dbtype;
	    dmDisplayPvErr(objptr, *pPv_err, prti_ctrl->pgeom_comel, pdlc, prti_ctrl->pLast_val);
	    prti_ctrl->init_flag = TRUE;
	    prti_ctrl->pgz = NULL;
	}
	return;
    }
		
    /* new connection , set up gizmo */
    else if (!prti_ctrl->init_flag) { 
	/* if valid ptr to data from db */
	if(prti_ctrl->pgeom_comel->pData_struct!= NULL){
	    pCtrl_enum = ((struct dbr_ctrl_enum *)prti_ctrl->pgeom_comel->pData_struct);
	    /* no way to tell if channel info is back yet, create
		gizmo when data is there --is this a valid indication?? */
	    if(prti_ctrl->pgeom_comel->valid_data){
/*	    if(pCtrl_enum->strs[0][0] != '\0'){ */
		/* set up table of string ptrs for menu gizmo */
		/* this will be changed to be done in the gz struct extension later */
		prti_ctrl->pExt = (void *)
			tisCalloc(__FILE__, __LINE__, pCtrl_enum->no_str + 1, sizeof(char *));
		if(prti_ctrl->pExt != NULL){
		    register i;
		    /* set up table of string pointers for gizmo in user extension */
		    for ( i = 0; i < pCtrl_enum->no_str; i++){
			((char **)prti_ctrl->pExt)[i] = pCtrl_enum->strs[i];
		    }
		    /* null str to indicate end of string table */
		    ((char **)prti_ctrl->pExt)[i] = NULL;
		    prti_ctrl->pgz = create_choice_gizmo(
			DefaultScreenOfDisplay(pdlc->d), pdlc->w,
			objptr, 
			permissive_test, 
			send_long,
			prti_ctrl->pgeom_comel,
			prti_ctrl->pExt,
			convert_button_stack[(unsigned short)pDl_button->stacking],
			(unsigned)(sizeof(char *) * (pCtrl_enum->no_str+1)));
		    if(prti_ctrl->pgz != NULL){
		    	prti_ctrl->init_flag = TRUE;
		    	XMapWindow(prti_ctrl->pgz->pgg->d, prti_ctrl->pgz->w);
		    }
		}/* end if pExt */
	    }/* end if strs */
/*	    else printf("Took null str case \n"); error to operator ?*/
	}/* end if pData */
    }/* end if not inited */

    /* if we have a gizmo init/update color */ 
    if(prti_ctrl->pgz != NULL){
	/* set foreground and background colors in gc*/
  	if(get_fgbg(pdl_color_map_index, prti_ctrl->pgeom_comel,
	    pDl_button->control.clr,
	    pDl_button->control.bclr,
	    pDl_button->clrmod,
	    pDl_button->rulecolorfg,
	    pDl_button->rulecolorbg,
	    &fg,&bg) < 0){
		fg = WhitePixel(pdlc->d, the_screen);
		bg = BlackPixel(pdlc->d, the_screen);
	    }
	set_gizmo_color(prti_ctrl->pgz,fg,bg);
	if(prti_ctrl->pgeom_comel->pData_struct!= NULL){
	    pCtrl_enum = ((struct dbr_ctrl_enum *)prti_ctrl->pgeom_comel->pData_struct);
	    /* if we have data that corresponds to the enum strings, display it */
	    if((pCtrl_enum->value >= 0) &&(pCtrl_enum->value < pCtrl_enum->no_str)){
	    	set_choice_gizmo(prti_ctrl->pgz, 
		       (unsigned)((struct dbr_ctrl_enum *)prti_ctrl->pgeom_comel->pData_struct)->value);   
	    }
	}/* end if data struct */
    }/* end gizmo setup */
    /* save current val as last */
    dmSave_float(&prti_ctrl->pLast_val, cvrt_to_float(prti_ctrl->pgeom_comel, field_type));

}/* end update_button*/



/************************************************************************************************
*
*	message_permissive_test
*
*
*	message_permissive_test checks whether the operator has the permission to change the value
*	of the channel defined at pRti_ctrl->pRti_ctrl->pgeom_comel.
*
*	NOTE: this is called indirectly by the controller gizmo
*   
*	Input	: pRti_ctrl	ptr to the channels com record
*		  pg		ptr to gizmo
*	Returns : TRUE if permission granted to change the value, else FALSE
*
*/
int message_permissive_test(pRti_ctrl, pg, informOnError)
struct rti_ctrl_info	*pRti_ctrl;
gizmo			*pg;
int			informOnError;
{
	return okToChangeValue(pg->pgg->s, ca_name(pRti_ctrl->pgeom_comel->pChan), informOnError);
}




/************************************************************************************************
*
*	update_message_button
*	DK 6/91
*
*	Update_message_button handles the initial setup for a message_button gizmo 
*		(once a connection is established.)  During configuration(see edd) the
*		button may be set up to act on press only (string for press message
*		but no string for release message) , both press and release(both strings),
*		or press/release default if no strings and process variable type is enum.
*		From then on it checks for lost or resumed connection
*		to update/init appropriately and/or changes the color as necessary.
*   
*	Input	: pdlc		ptr to the display list context struct
*		  objptr	ptr to the object struct(ht, wt, x,y) 
*		  prti_ctrl	ptr to run-time struct for the menu
*		  not_conn	indicates ca connect or not
*		  field_type	native type of process variable controlled
*		  expose_flag	indicates expose event or not
*		  pPv_err	error state for the message button based on pv state/type
*	Output	: *pPv_err	if type= string, sets to invalid dbtype
*	Returns : nothing
*
*/




update_message_button(pdlc, objptr,prti_ctrl, pPv_err, field_type,expose_flag, pdl_color_map_index)
struct dl_context *pdlc;
XRectangle *objptr;
struct rti_ctrl_info *prti_ctrl;
unsigned short *pPv_err, expose_flag;
short field_type;
int	*pdl_color_map_index;

{
#ifndef AXP
    unsigned fg,bg;
#else
    unsigned long fg,bg;
#endif
    void send_message();
    int message_permissive_test();
    register struct dbr_ctrl_enum *pCtrl_enum;
    register struct dl_message_button *pDl_button= 
	((struct dl_message_button *)prti_ctrl->pdl_ctrl);

    /* first time thru or expose  */
    /* if illegal type for valuator object */
    if(field_type == DBR_STRING){
	/* first time thru or expose */
	if(expose_flag || !prti_ctrl->init_flag){
	    *pPv_err = dmInvalid_dbtype;
	    dmDisplayPvErr(objptr, *pPv_err, prti_ctrl->pgeom_comel, pdlc, prti_ctrl->pLast_val);
	    prti_ctrl->init_flag = TRUE;
	    prti_ctrl->pgz = NULL;
	}
	return;
    }


    /* new connection , set up gizmo */
    if (!prti_ctrl->init_flag) { 
	/* if valid ptr to data from db */
	if(prti_ctrl->pgeom_comel->pData_struct!= NULL){
	    pCtrl_enum = ((struct dbr_ctrl_enum *)prti_ctrl->pgeom_comel->pData_struct);
	    /* no way to tell if channel info is back yet, create
		gizmo when data is there --is this a valid indication?? */

	    if(prti_ctrl->pExt != NULL){
		/* set up gizmo */
	    	prti_ctrl->pgz = create_message_gizmo(
			DefaultScreenOfDisplay(pdlc->d), pdlc->w,
			objptr, 
			message_permissive_test, 
			send_message,
			prti_ctrl,
			((char **)prti_ctrl->pExt),
			pDl_button->behavior,
			pCtrl_enum->strs [0], pCtrl_enum->strs [1],
			pCtrl_enum->value);
	    }

	    if(prti_ctrl->pgz != NULL){
		prti_ctrl->init_flag = TRUE;
		XMapWindow(prti_ctrl->pgz->pgg->d, prti_ctrl->pgz->w);
	    }
	
	}/* end if pData */
    }/* end if not inited */

    /* if we have a gizmo init/update color */ 
    if(prti_ctrl->pgz != NULL){
	/* set foreground and background colors in gc*/
  	if(get_fgbg(pdl_color_map_index, prti_ctrl->pgeom_comel,
	    pDl_button->control.clr,
	    pDl_button->control.bclr,
	    pDl_button->clrmod,
	    pDl_button->rulecolorfg,
	    pDl_button->rulecolorbg,
	    &fg,&bg) < 0){
		fg = WhitePixel(pdlc->d, the_screen);
		bg = BlackPixel(pdlc->d, the_screen);
	}
	set_gizmo_color(prti_ctrl->pgz,fg,bg);
  
      }/* end gizmo setup */

}/* end update_message_button*/



/************************************************************************************************
*
*	send_message
*
*	DK 12/90
*
*	Send_message sends a "1" to the appropriate process variable when a 
*		message button gizmo (controller) is selected by depressing
*		the mouse button.
*		If a "release message" has been specified during display 
*		configuration (see edd), a "0" will be sent to the process
*		variable upon release upon release of the mouse button.
*				
*
*	Input	: pRti_ctrl	pointer to run-time control struct 
*		  pg		ptr to text entry gizmo struct 
*		  depressed	flag to indicate whether the mouse button has
*				been depressed(TRUE) or released (FALSE)
*		  string	press or release string--not used under
*				current implementation
*	Output	: nothing
*	Returns : nothing
*
*/ 



void send_message(pRti_ctrl, pg, depressed, pevent)
struct rti_ctrl_info	*pRti_ctrl;
gizmo			*pg;
int			depressed;
XEvent			*pevent;

{
	struct gsdl_com_record	*pComel	= pRti_ctrl->pgeom_comel;
	int			status;

	if (depressed){
		if ((pRti_ctrl->user == PRESS_RELEASE) || (pRti_ctrl->user == PRESS) ||
			(((struct dl_message_button *)pRti_ctrl->pdl_ctrl)->behavior == 1)) { 
			static short	oneval	= 1;

			status = ca_put(DBR_SHORT, pComel->pChan, &oneval);	/*((char **)pRti_ctrl->pExt)[0]);*/
			ca_flush_io();
			if(status != ECA_NORMAL){
				/* need to do error handling here */
				printf("~CA error returned--couldn't send value %s\n", ca_message(status));
			}
			else
				logChangedValue(pg->pgg->s, ca_name(pComel->pChan), ((char **)pRti_ctrl->pExt)[1],
									    ((char **)pRti_ctrl->pExt)[0]);
		}
 	}
	else
	{
		if ((pRti_ctrl->user == PRESS_RELEASE) || (pRti_ctrl->user == PRESS_RELEASE_DB)) {
			static short	zeroval	= 0;

			status = ca_put(DBR_SHORT, pComel->pChan, &zeroval);	/*((char **)pRti_ctrl->pExt)[1]);*/
			ca_flush_io();
			if(status != ECA_NORMAL){
	    			/* need to do error handling here */
	    			printf("~CA error returned--couldn't send value %s\n", ca_message(status));
	        	}	
			else
				logChangedValue(pg->pgg->s, ca_name(pComel->pChan), ((char **)pRti_ctrl->pExt)[0],
									    ((char **)pRti_ctrl->pExt)[1]);
		}
	}/* end if release */
} /* end send_message */



/***********************************************************************
*
*	Delete_controllers
*	
*	Deb Kerstiens 3/91
*
*	Purpose : This routine deletes the "gizmos" that have been created
*			for operator control.  Delete_controllers is called
*			when a display (dm_llist_el) is to be deleted to delete 
*			each gizmo that is currently active.  This must be
*			done so that when the parent (display) window is
*			deleted, the child window created for each gizmo will
*			already have been deleted.
*			Memory calloced for extension info for gizmos is also 
*			deleted.
*
*	Input :
*	    prti	: ptr to the run-time list for controllers
*	    nctrl	: count of controller elements to display
*	Output :
*	    none
*	Returns
*	    none
*
**********************************************************************/

void delete_controllers(prti, nctrl)
struct rti_ctrl_info 	*prti;
unsigned short		nctrl;

{
  unsigned short i;
  struct rti_ctrl_info *prti_ctrl; 
 
#ifdef DEBUG
printf("~in delete_controllers\n");
#endif

for (	prti_ctrl = prti, i = 0;  
	i < nctrl;   
	prti_ctrl++, i++){
	/* free any mem calloced for each element on the controller list */
	if(prti_ctrl->pExt != NULL) tisFree(prti_ctrl->pExt);
	/* delete gizmo for window deletion in proper order */
	if(prti_ctrl->pgz != NULL) free_gizmo(prti_ctrl->pgz);
}

} /* end delete_controllers */




/************************************************************************************************
*
*	value_in_range
*
*	DK 09/91
*
*	Value_in_range checks to see if the text string entered by the operator using the
*		text_entry gizmo is a valid one for the db_type and the control limits of
*		the pv to which it is attached.
*
*
*	Input	: pComel	ptr to com_record
*		  string	string to check 
*	Output	: none
*	Return	: true if okay, false otherwise
*/

value_in_range(pComel, string, s)
struct gsdl_com_record	*pComel;
char					*string;
Screen					*s;
{
    switch(ca_field_type(pComel->pChan)){
	case DBF_CHAR:
		{
	    char					value;
	    struct dbr_ctrl_char	*pctrl	= ((struct dbr_ctrl_char *)pComel->pData_struct);

	    if (sscanf(string, "%c", &value) == 1)
			{
			if (pctrl->lower_ctrl_limit < pctrl->upper_ctrl_limit)
	    		if ((pctrl->lower_ctrl_limit <= value) && (value <= pctrl->upper_ctrl_limit))
					return TRUE;
				else
					return outOfRangeDialog(s, TRUE, (double)pctrl->lower_ctrl_limit, (double)pctrl->upper_ctrl_limit, NULL);
			else
				return TRUE;
			}
		else
			return outOfRangeDialog(s, TRUE, (double)pctrl->lower_ctrl_limit, (double)pctrl->upper_ctrl_limit, "~bad format");
		}
	case DBF_SHORT:
		{
	    short					value;
	    struct dbr_ctrl_short	*pctrl	= ((struct dbr_ctrl_short *)pComel->pData_struct);

	    if (sscanf(string, "%hd", &value) == 1)
			{
			if (pctrl->lower_ctrl_limit < pctrl->upper_ctrl_limit)
	    		if ((pctrl->lower_ctrl_limit <= value) && (value <= pctrl->upper_ctrl_limit))
					return TRUE;
				else
					return outOfRangeDialog(s, TRUE, (double)pctrl->lower_ctrl_limit, (double)pctrl->upper_ctrl_limit, NULL);
			else
				return TRUE;
			}
		else
			return outOfRangeDialog(s, TRUE, (double)pctrl->lower_ctrl_limit, (double)pctrl->upper_ctrl_limit, "~bad format");
		}
	case DBF_LONG:
		{
	    long					value;
	    struct dbr_ctrl_long	*pctrl	= ((struct dbr_ctrl_long *)pComel->pData_struct);

	    if (sscanf(string, "%ld", &value) == 1)
			{
			if (pctrl->lower_ctrl_limit < pctrl->upper_ctrl_limit)
	    		if ((pctrl->lower_ctrl_limit <= value) && (value <= pctrl->upper_ctrl_limit))
					return TRUE;
				else
					return outOfRangeDialog(s, TRUE, (double)pctrl->lower_ctrl_limit, (double)pctrl->upper_ctrl_limit, NULL);
			else
				return TRUE;
			}
		else
			return outOfRangeDialog(s, TRUE, (double)pctrl->lower_ctrl_limit, (double)pctrl->upper_ctrl_limit, "~bad format");
		}
	case DBF_FLOAT:
		{
	    float					value;
	    struct dbr_ctrl_float	*pctrl	= ((struct dbr_ctrl_float *)pComel->pData_struct);

	    if (sscanf(string, "%f", &value) == 1)
			{
			if (pctrl->lower_ctrl_limit < pctrl->upper_ctrl_limit)
	    		if ((pctrl->lower_ctrl_limit <= value) && (value <= pctrl->upper_ctrl_limit))
					return TRUE;
				else
					return outOfRangeDialog(s, TRUE, (double)pctrl->lower_ctrl_limit, (double)pctrl->upper_ctrl_limit, NULL);
			else
				return TRUE;
			}
		else
			return outOfRangeDialog(s, TRUE, (double)pctrl->lower_ctrl_limit, (double)pctrl->upper_ctrl_limit, "~bad format");
		}
	case DBF_DOUBLE:
		{
	    double					value;
	    struct dbr_ctrl_double	*pctrl	= ((struct dbr_ctrl_double *)pComel->pData_struct);

	    if (sscanf(string, "%lf", &value) == 1)
			{
			if (pctrl->lower_ctrl_limit < pctrl->upper_ctrl_limit)
	    		if ((pctrl->lower_ctrl_limit <= value) && (value <= pctrl->upper_ctrl_limit))
					return TRUE;
				else
					return outOfRangeDialog(s, TRUE, (double)pctrl->lower_ctrl_limit, (double)pctrl->upper_ctrl_limit, NULL);
			else
				return TRUE;
			}
		else
			return outOfRangeDialog(s, TRUE, (double)pctrl->lower_ctrl_limit, (double)pctrl->upper_ctrl_limit, "~bad format");
		}
	case DBR_ENUM:
		{
	    int						i;
	    struct dbr_ctrl_enum	*pctrl = ((struct dbr_ctrl_enum *)pComel->pData_struct);

	    for ( i = 0; i < pctrl->no_str; i++)
			if(!strcmp(string, pctrl->strs[i]))
				return TRUE;

		return outOfRangeDialog(s, FALSE, 0.0, 0.0, "~not a valid choice");
		}
	case DBR_STRING:
	    return TRUE;
    }
}



#include <OperPriv.h>
#include <pwd.h>
#ifdef NeXT
#include <libc.h>
#else
#include <unistd.h>
#endif
#include <signal.h>

/*
 * this points to the users name after the operator priv and tracking is initialized
 */
char	operator[30+1]; /* Name of operator */

void AbortHandler()
{
	/*
	 * we need to close the operator tracking data and then exit()
	 *
	 * NOTE: for this technique of closing the tracking data to work,
	 *	 dm should either exit through NormalExit() or abort() after
	 *	 InitOperTracking() is called
	 *	 (i.e. anytime after the first attempt to do a put)
	 */
	CloseOperPriv();
	CloseOperTracking();
	printf("~dm - aborted.\n");	/* this just lets me know it happened :-( */
	exit(1);
}


NormalExit(exitCode)
int	exitCode;
{
	inNormalExit = 1;
	if(strlen(operator))
		CloseOperTracking();
	if (launchData)
		d4close (launchData);
	exit(exitCode);
}


PrivInitialization(s)
Screen	*s;
{
	if (!strlen(operator)){
		/*
		 * get the effective operators name
		 */
		GetUserName (operator);

		/*
		 * do the required initialization
		 */
		if (InformOfOperPrivTrackError(s, InitOperPriv(operator)) != PRIVOK){
			*operator = 0;
			return FALSE;
		}
	}
	else
	{
		GetUserName (operator); /* update operator name to match current login */
	}
	return TRUE;
}

TrackingInitialization(s)
Screen	*s;
{
	if (!strlen(operator)){
		return FALSE;
	}
	/*
	 * do the required initialization
	 */
	if (InformOfOperPrivTrackError(s, InitOperTracking(poperator)) != PRIVOK){
		return FALSE;
	}
	if (signal(SIGABRT, SIG_IGN) != SIG_IGN ){
		/*
		 * interrupts are not being ignored, so catch them in AbortHandler
		 * to close the tracking data
		 */
		signal(SIGABRT, AbortHandler);
	}
	return TRUE;
}

CloseTracking (s)
Screen	*s;
{
	CloseOperTracking();
}


InformOfOperPrivTrackError(s, errCode)
Screen	*s;
int	errCode;
{
	char	**pmessage;

	if (errCode != PRIVOK && OperPrivTrackErr(errCode, &pmessage) == PRIVOK){
		enum 	choices	{ ok, terminate, nchoices };
		char 	*pchoices[nchoices];

		pchoices[ok] = "~ok";
		pchoices[terminate] = 0;
		perform_dialog(s, 600, 300, 21, pmessage, pchoices, 21);
	}
	return errCode;
}


extern int	PrivCheckOverride;

/*
 * this routine checks that this user has the required priv
 *
 * returns:	1 if ok to send
 *		0 if not
 */
int
okToChangeValue(s, pchan, informOnError)
Screen	*s;
char	*pchan;
int	informOnError;
{
	int	errCode;

	if (PrivCheckOverride)
		/*
		 * allow for the -p flag
		 */
		return;

	if (!PrivInitialization(s))
		/*
		 * do not allow modifies if we cannot check for privs
		 */
		return FALSE;

	errCode = CheckOperPriv(pchan);
	if (errCode != PRIVOK && informOnError == TRUE)
		InformOfOperPrivTrackError( s, errCode);

	return (errCode == PRIVOK);
}


/*
 * this routine logs the change to the channel
 *
 * returns:	nothing
 */
static int trkErrNotified = 0;

void
logChangedValue(s, pchan, pold, pnew)
Screen	*s;
char	*pchan;
char	*pold;
char	*pnew;
{
    char *trkfailmsg [] = {
				"Track log failure. Message sent to syslog",
				"and console. Please check the",
				"directory /tis4000db.\n",
						0
			  };

	if (PrivCheckOverride)
		/*
		 * allow for the -p flag
		 */
		return;

	if (!TrackingInitialization(s)) {
		/* Put change notification in system log */
		syslog (LOG_ALERT|LOG_LOCAL1,
			 "dm(%d): Track log failure. %s changed %s from %s to %s",
			 getpid (), operator, pchan, pold, pnew);
		if (!trkErrNotified) {
			enum 	choices	{ ok, terminate, nchoices };
			char 	*pchoices[nchoices];

			pchoices[ok] = "~ok";
			pchoices[terminate] = 0;
			perform_dialog (s, 600, 300, 21, trkfailmsg, pchoices, 21);
			trkErrNotified = 1;
		}

	}
	else
	{
		InformOfOperPrivTrackError( s, WriteOperTracking(pchan, pold, pnew) );
		CloseTracking ();
		trkErrNotified = 0;
	}
}
