/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/***********************************************************************
**
**	RELATED_DISPLAYS.C
**	
**	Deb Kerstiens 12/90
**
**	Modification History:
**	10/15/91	DK	modified to allow macro substitution to be inherited
**				by sub-displays
**	11/21/91	DK	fixed bug in macro substitution--macro strings from
**				display list had been modified previously, preventing
**				any other related display that accessed that macro
**				string from working.  Added calloc, strcpy for new
**				macro string to modify.
**
*/
#include <syslog.h>

#include <Xlib.h>      			/* the X library 		*/
#include <Xutil.h>
#include <db_access.h>
#include <cadef.h>
#include <dldef.h>			/* display list defines 	*/
#include <dlinfo.h>			/* more display list defines 	*/
#include <dl_context.h>			/* defines dl context		*/
#include <gizmo.h>			/* defs for gizmos */
#include <op_gsdl_def.h>
#include <dynamic_def.h>		/* rti dynamic defines		*/
#include <d4all.h>

#ifdef AXP
#include <string.h>
#endif

extern void *tisMalloc();
extern void *tisCalloc();
extern void tisFree();

void add_related_display();

int	displayAdded;

/* CodeBase stuff to support the launch facility */

extern	C4CODE	codebase;
extern	D4DATA	*launchData;
extern	F4FIELD	*nextFld, *prevFld, *fileFld, *paramFld;
extern	T4TAG	*tag;

static	void	stripTrail ();

/************************************************************************************
*
*	setup_related_display
*
*	DK 12/16/90
*
*	Setup_related_displays retrieves the appropriate info from the display list 
*	and sets up the menu gizmo to allow the operator to select a related display. 
*	From this point on, the screen updates will be handled by the menu gizmo and 
*	the appropriate actions, i.e. new display callup, will be  when the handler 
*	specifed during menu gizmo creation is called(when a menu item is selected.)
*
*
*/
void setup_related_displays(pDm_llist_hdr, pDlc, pDm_llist_el, pDl_related_display, nrel,pdl_index)
struct dm_llist_hdr		*pDm_llist_hdr;
struct dl_context 		*pDlc;
struct dm_llist_el		*pDm_llist_el;
struct dl_related_display 	*pDl_related_display;
#ifndef AXP
int				*pdl_index;
#else
unsigned long			*pdl_index;
#endif
short 				nrel;
	{
     /* set up table of string ptrs for menu gizmo */    
    int				i, j, num;
    static unsigned short	nstrs = 0; /* number of valid related display names */
    char			**pDl_str_table	= pDm_llist_el->pdl_strings;
    struct rti_rel_disp		*pRti_rel_disp;
    struct dl_related_display	*pDl;

    if (nrel <= 0)
	return;
    if ((pDm_llist_el->prti_rel_disp = 
		(struct rti_rel_disp *)tisCalloc(__FILE__, __LINE__, nrel, sizeof (struct rti_rel_disp))) == NULL)
	return;
	
    /* set up table of string pointers for gizmo in user extension */
    for (pRti_rel_disp = pDm_llist_el->prti_rel_disp, pDl = pDl_related_display,i = 0; i < nrel;
				i++, pRti_rel_disp++)
	{
	pRti_rel_disp->pDl_related_display = pDl;
	pRti_rel_disp->pDm_llist_hdr = pDm_llist_hdr; 	/* set up ptr to first el on list for handler to add new el */
	pRti_rel_disp->pDm_llist_el  = pDm_llist_el;
	/* alloc for macro pair pointers */
	/* create string table for gizmo menu */
	for (j = 0, nstrs = 0; j < MAX_REL_DISPLAYS;j++)
		{
			int	name_index	= (int)(pDl->display[j].name);
			int	args_index	= (int)(pDl->display[j].args);
			int	label_index	= (int)(pDl->display[j].label);
			
		
			/* use menu label if available, else use display name */
			if (strlen(pDl_str_table[name_index]) > 0)
				{
	    			((char **)pRti_rel_disp->pDisplays)[nstrs] = pDl_str_table[name_index];
				if (strlen(pDl_str_table[label_index]))
					((char **)pRti_rel_disp->pMenustr)[nstrs] = pDl_str_table[label_index];
				else
					((char **)pRti_rel_disp->pMenustr)[nstrs] = pDl_str_table[name_index];
				if (strlen(pDl_str_table[args_index]) > 0)
					{
					pRti_rel_disp->ppMacros[nstrs] = (char **)tisCalloc(__FILE__, __LINE__, MAX_PAIR_STRS, sizeof(char *));
					pRti_rel_disp->pArgList[nstrs] = pDl_str_table[args_index];
					}
				else if (pDm_llist_el->ppMacro_pairs != NULL)
					{
					pRti_rel_disp->ppMacros[nstrs] = pDm_llist_el->ppMacro_pairs;
					}
    				else
					pRti_rel_disp->ppMacros[nstrs] = NULL;
				nstrs++;
				}
			}
    		((char **)pRti_rel_disp->pMenustr)[nstrs] = NULL;	/* null str to indicate end of string table */ 
		if (nstrs > 0)
			{
			pRti_rel_disp->pgz = create_menu_gizmo(DefaultScreenOfDisplay(pDlc->d),
				pDlc->w,
	    			&(pDl->object),
				ok_to_operate,
				add_related_display,
				pRti_rel_disp,
				pRti_rel_disp->pMenustr,
				(unsigned)(sizeof(char *) * (nstrs+1))); /*num of strs + 1 for NULL*/
    	    		if (pRti_rel_disp->pgz != NULL)
				{
				XMapWindow(pRti_rel_disp->pgz->pgg->d, pRti_rel_disp->pgz->w);
				set_gizmo_color(pRti_rel_disp->pgz, COLOR_LOOKUP(pdl_index, pDl->clr),
						COLOR_LOOKUP(pdl_index, pDl->bclr));
				}
			}
		pDl = (struct dl_related_display *)((short *)pDl + ((struct dl_element_hdr *)pDl)->nwords);
		}
	}

/************************************************************************************
*
*	prestart_related_display
*
*	DK 12/16/90
*
*	Activate any related displays whose names begin with a percent.  The displays
*	are activated and then iconized.
*/

prestart_related_displays (pDm_llist_hdr, pDlc, pDm_llist_el, pDl_related_display, nrel,pdl_index)
struct dm_llist_hdr		*pDm_llist_hdr;
struct dl_context 		*pDlc;
struct dm_llist_el		*pDm_llist_el;
struct dl_related_display 	*pDl_related_display;
#ifndef AXP
int				*pdl_index;
#else
unsigned long			*pdl_index;
#endif
short 				nrel;
{
	struct rti_rel_disp		*pRti_rel_disp;
	struct dl_related_display	*pDl;
	char				**pDl_str_table	= pDm_llist_el->pdl_strings;
	int				i, j, nstrs;

	for (pRti_rel_disp = pDm_llist_el->prti_rel_disp, pDl = pDl_related_display,i = 0; i < nrel;
				i++, pRti_rel_disp++) {
	    for (j = 0, nstrs = 0; j < MAX_REL_DISPLAYS;j++) {
		int	name_index	= (int)(pDl->display[j].name);
		if (strlen(pDl_str_table[name_index]) > 0) {

		    /* Here is the biggest kludge I have ever been forced to deal with.  The leading % */
		    /* means that the display should be iconized on initial startup.  Unfortunately this */
		    /* is a two step process.  The display is added to the display list here with a "new" */
		    /* flag set.  In the main event loop in "display_mgr.c' the display is iconized after */
		    /* being activated.  The problen is if I strip the % off the filename here, then */
		    /* display_mgr.c does not know that the display should be iconized.  If I leave the % */
		    /* on then the file cannot be found since the % is not part of the actual filename. */
		    /* I tried putting an additional flag in the display list structure but that made */
		    /* everything go to shit.  Probably some alignment problem.  Here is the solution: */
		    /* The filename is shifted one character to the left eliminating the %.  A 0x7F is written */
		    /* into the position where the original terminating NULL was.  Easy to find with the */
		    /* strlen function. PFEW! */

		    if (pDl_str_table [name_index][0] == '%') {
			strcpy (&pDl_str_table [name_index][0], &pDl_str_table [name_index][1]);
			pDl_str_table [name_index][strlen (&pDl_str_table [name_index][0]) + 1] = 0x7F;
/*			printf ("Preactivating %s\n", &pDl_str_table [name_index][0]); */
			add_related_display (pRti_rel_disp, pRti_rel_disp->pgz, j);
		    }
		}	
	    }
	    pDl = (struct dl_related_display *)((short *)pDl + ((struct dl_element_hdr *)pDl)->nwords);
	}
}
 
/************************************************************************************
*
*	add_related_display
*
*	DK 12/16/90
*
*	Add_related_displays is the handler specifed during menu gizmo creation 
*	to be called when a menu selection from a related display element is selected.
*	It sets up a dm_llist_el member and fills in the display name. It also sets a
*	flag in the hdr of the linked list to indicate that a new element has been added.
*	During the next pass of the display manager, the display will be read and 
*	initialized(ca_setup, etc.)
*
*
*	Input	: pRti_rel_disp 	ptr to related display struct
*		  pg			ptr to a gizmo
*		  index			index of the string in str table for display
*						selected
*	Output	: Allocs mem for new display element and adds it to display manager's
*			list.
*		  Sets flag to let display manager know there's a new display to 
*			manage.
*	Returns	: nothing
*/

extern long lastLicenseCheck;
extern struct dm_llist_el	*pDm_llist_pos;	/* current display in callup stack */

/*----------*/
void add_related_display (pRti_rel_disp, pg, index)

    struct rti_rel_disp	*pRti_rel_disp;
    gizmo		*pg;
    int 		index;

    {

	struct dm_llist_el *pDm_llist_el;
	char	command [500], option [50], buf [500], key [50], parms [250];
 	char	displayName [259], macroName [50], macroValue [100];
	int	i, j, k;

	displayAdded = 1;

	/* Handle TIS licensing */

#if !(defined(ultrix))
	if ((time(NULL) - lastLicenseCheck) > 14400) {
	    tisHardLicense();
	    lastLicenseCheck = time(NULL);
	}
#endif

	/* Copy the display name and parameters to the final display in case no launch is being used */

	strcpy (displayName, pRti_rel_disp->pDisplays[index]);
	stripTrail (displayName);
	if (pRti_rel_disp->pArgList[index])
	    strcpy (parms, pRti_rel_disp->pArgList[index]);
	else
	    strcpy (parms, "");

	/* If the display name field is empty and self-kill is active, kill this display. */
	/* If no self kill, just return */

	if (strlen (displayName) == 0) {
	    if (pRti_rel_disp->pDl_related_display->selfkill) {
		dmExit_display(pRti_rel_disp->pDm_llist_hdr, pRti_rel_disp->pDm_llist_el, NULL); /* <- note the NULL! */
	    }
	    return;
	}

	if (pRti_rel_disp->pArgList[index]) {
	    if (get_macro_pairs (parms, pRti_rel_disp->ppMacros[index],
				MAX_PAIR_STRS, DefaultScreenOfDisplay(pRti_rel_disp->pDm_llist_el->dl_context.d)) == 0) {
		int	i;
			
		for (i = 0; i < MAX_PAIR_STRS; i++)
		    pRti_rel_disp->ppMacros[index][i] = 0;
		return;
	    }
	}

	/* Check if the launch facility is available. If yes, check the first token of the display name */
	/* for the word launch. "launch" may not appear alone but can be like "/tis4000/tasks/launch". */

	if (launchData) {	/* The grphicdb database is available */

	    /* See if the first token contains the substring "launch" */

	    strcpy (command, displayName);
	    strcpy (buf, strtok (command, " "));
	    if (strstr (buf, "launch")) {	/* Found "launch" substring, launch facility is being used */

		/* Collect the key */

		strcpy (key, strtok (NULL, " "));

		/* Collect the P, N, or M */

		strcpy (option, strtok (NULL, " "));

		/* At this point the key may contain a macro reference that needs a substitution performed. */
		/* A macro begins with a $ and contains only letters or digits. Find the macro reference and */
		/* substitute it's value. */

		strcpy (macroName, "");

		if (strchr (key, '$')) {

		    i = j = k = 0;

		    /* Copy everything up to the $ */

		    while (key [i]) {
			if (key [i] != '$')
			    buf [j++] = key [i];
			else
			    break;
			i++;
		    }
		    buf [j] = '\0';

		    /* Get the name of the macro */

		    i++;
		    while (isdigit (key [i]) || isalpha (key [i]))
			macroName [k++] = key [i++];
		    macroName [k] = '\0';

		    /* Search for the macro in the table */

		    k = 0;
		    while (pRti_rel_disp->ppMacros[index][k]) {
			if (strcmp (macroName, pRti_rel_disp->ppMacros[index][k]) == 0) {
			    strcpy (macroValue, pRti_rel_disp->ppMacros[index][k + 1]);
			    break;
			}
			k += 2;
		    }

		    /* Conactenate the macro value onto the key and the the tail of the original key */

		    strcat (buf, macroValue);
		    strcat (buf, &key [i]);
		    strcpy (key, buf);

		    /* The key might be in quotes due to the earlier implementation. Strip them off if so */

		    if (key [0] == 0x22) {
			strcpy (key, &key [1]);
			key [strlen (key) - 1] = '\0';
		    }

		}

		/* Seek on the key */

		sprintf (buf, "%-6s", key);
		if (d4seek (launchData, buf) == 0) {	/* Found it */
		    switch (tolower (option [0])) {
			case 'n':	/* Next */
			    strcpy (key, f4str (nextFld));
			    if (strcasecmp (key, "NONEXT") == 0)
				return;
			    if (d4seek (launchData, key) != 0) {
				syslog (LOG_ALERT|LOG_LOCAL1, "dm(%d): Unable to find key %s in grphicdb.", getpid (), key);
				strcpy (displayName, "ERROR");
			    } else
				strcpy (displayName, f4str (fileFld));
			    break;
			case 'p':	/* Prev */
			    strcpy (key, f4str (prevFld));
			    if (strcasecmp (key, "NOPREV") == 0)
				return;
			    if (d4seek (launchData, key) != 0) {
				syslog (LOG_ALERT|LOG_LOCAL1, "dm(%d): Unable to find key %s in grphicdb.", getpid (), key);
				strcpy (displayName, "ERROR");
			    } else
				strcpy (displayName, f4str (fileFld));
			    break;
			case 'm':	/* Main */
			    /* Already at proper record */
			    strcpy (displayName, f4str (fileFld));
			    break;
		    }
		} else {
		    syslog (LOG_ALERT|LOG_LOCAL1, "dm(%d): Unable to find key %s in grphicdb.", getpid (), key);
		    strcpy (displayName, "ERROR");
		}
		d4unlock_all (launchData);

		/* Copy the script name into the related_display structure */

		stripTrail (displayName);

		/* If the related display callup specified parameters, use them. */
		/* If not, get them from the database */

		if (strlen (f4str (paramFld)) != 0) {
		    strcpy (parms, f4str (paramFld));
		    stripTrail (parms);	   		    
		}

		/* If a display, add the key to the parameter list. If an execute script, */
		/* build the command line and execute it. */

		if ((displayName [0] != '!') && (displayName [0] != '#')) {
/* changed to fixed name, KEY		    if (strlen (macroName)) { */
			sprintf (buf, "KEY=%s", key);
/* changed to fixed name, KEY			sprintf (buf, "%s=%s", macroName, macroValue); */
			if (strlen (parms)) {
			    strcat (buf, ",");
			    strcat (buf, parms);
			}
			strcpy (parms, buf);
/* changed to fixed name, KEY		    } */
		} else {
		    sprintf (buf, "%s %s &", &displayName [1], parms);
		    system (buf);
		    if (pRti_rel_disp->pDl_related_display->selfkill)
			dmExit_display(pRti_rel_disp->pDm_llist_hdr, pRti_rel_disp->pDm_llist_el, NULL);
		    return;
		}

		/* Convert the macros into macro pairs */

		if (pRti_rel_disp->pArgList[index] || strlen (parms)) {
		    if (get_macro_pairs (parms, pRti_rel_disp->ppMacros[index],
				MAX_PAIR_STRS, DefaultScreenOfDisplay(pRti_rel_disp->pDm_llist_el->dl_context.d)) == 0) {
			int	i;
			
			for (i = 0; i < MAX_PAIR_STRS; i++)
			    pRti_rel_disp->ppMacros[index][i] = 0;
			return;
		    }
		}

	    }
	}
	    
	/* Process execute scripts and self-kills */

	if ((displayName [0] == '!') || (displayName [0] == '#')) {

	    /* treat this as an execute script */

	    runScript(&displayName [1], pRti_rel_disp->pDm_llist_el->display_path, pRti_rel_disp->ppMacros[index]);

	    if (pRti_rel_disp->pDl_related_display->selfkill && (displayName[0] == '#')) {

		 /* kill this display just like hitting the exit from the menu */

		dmExit_display(pRti_rel_disp->pDm_llist_hdr, pRti_rel_disp->pDm_llist_el, NULL); /* <- note the NULL! */

	    }
	    return;
	}
	
	/* process stack operators */
	if ((displayName [0] == '<') || (displayName [1] == '-')) {
	    pDm_llist_el = pRti_rel_disp->pDm_llist_hdr->pNew_llist;
	    if (pDm_llist_el == pDm_llist_pos) /* if viewing first entry in list */
		pDm_llist_pos = NULL;	/* this selects last entry in list! */
	    while (pDm_llist_el != NULL && (pDm_llist_pos != pDm_llist_el->pNext_el))
	        pDm_llist_el = pDm_llist_el->pNext_el;
	    pDm_llist_pos = pDm_llist_el;
	    if (pDm_llist_pos == NULL) /* No previous exists, reselect base graphic */
		pDm_llist_pos = pRti_rel_disp->pDm_llist_hdr->pNew_llist;
	    if (pDm_llist_pos->visible) {
		XRaiseWindow(pDm_llist_pos->dl_context.d, pDm_llist_pos->dl_context.w);
		XSetInputFocus(pDm_llist_pos->dl_context.d, pDm_llist_pos->dl_context.w,
					RevertToNone, CurrentTime);
	    } else {
		XMapWindow(pDm_llist_pos->dl_context.d, pDm_llist_pos->dl_context.w);
		XRaiseWindow(pDm_llist_pos->dl_context.d, pDm_llist_pos->dl_context.w);
	    }
	    return;
	}
	else if ((displayName [0] == '-') || (displayName [1] == '>')) {
	    pDm_llist_pos = pDm_llist_pos->pNext_el;
	    if (pDm_llist_pos == NULL)
		pDm_llist_pos = pRti_rel_disp->pDm_llist_hdr->pNew_llist;
	    if (pDm_llist_pos->visible) {
		XRaiseWindow(pDm_llist_pos->dl_context.d, pDm_llist_pos->dl_context.w);
		XSetInputFocus(pDm_llist_pos->dl_context.d, pDm_llist_pos->dl_context.w,
					RevertToNone, CurrentTime);
	    } else {
		XMapWindow(pDm_llist_pos->dl_context.d, pDm_llist_pos->dl_context.w);
		XRaiseWindow(pDm_llist_pos->dl_context.d, pDm_llist_pos->dl_context.w);
	    }
	    return;
	}
	
	/* run the display list looking for a match */

	pDm_llist_el = pRti_rel_disp->pDm_llist_hdr->pNew_llist;
	while (pDm_llist_el != NULL && !display_matches(pRti_rel_disp, index, pDm_llist_el))
	    pDm_llist_el = pDm_llist_el->pNext_el;

	if (pDm_llist_el != NULL) {

	    /* This display is already in the list */
	    pDm_llist_pos = pDm_llist_el;

	    if (pDm_llist_el->visible) {
		XRaiseWindow(pDm_llist_el->dl_context.d, pDm_llist_el->dl_context.w);
		XSetInputFocus(pDm_llist_el->dl_context.d, pDm_llist_el->dl_context.w,
					RevertToNone, CurrentTime);
	    } else {
		XMapWindow(pDm_llist_el->dl_context.d, pDm_llist_el->dl_context.w);
		XRaiseWindow(pDm_llist_el->dl_context.d, pDm_llist_el->dl_context.w);
	    }
	} else {

	    /* This display is NOT in the list, put it at the end */

	    /* go to the end of the list */

	    pDm_llist_el = pRti_rel_disp->pDm_llist_hdr->pNew_llist;
	    while (pDm_llist_el->pNext_el != NULL)
		pDm_llist_el = pDm_llist_el->pNext_el;
		if ((pDm_llist_el->pNext_el = (struct dm_llist_el *)tisCalloc(__FILE__, __LINE__, 1,
					sizeof(struct dm_llist_el))) != NULL) {
		pRti_rel_disp->pDm_llist_hdr->new_llist_flag = TRUE;

		/* We have to copy one past the end of the string to propogate the iconize flag */

		memcpy(pDm_llist_el->pNext_el->display_name, displayName, strlen (pRti_rel_disp->pDisplays[index]) + 2);

		/* need display for setting up window (in setup_graphics) */
		pDm_llist_el->pNext_el->dl_context.d = pDm_llist_el->dl_context.d;
		pDm_llist_el->pNext_el->win_context = pDm_llist_el->win_context;
		pDm_llist_el->pNext_el->dl_context.pfont_table = pDm_llist_el->dl_context.pfont_table;
		if (pRti_rel_disp->ppMacros[index]) {
		    int	i;
				
		    pDm_llist_el->pNext_el->ppMacro_pairs = (char **)tisCalloc(__FILE__, __LINE__, MAX_PAIR_STRS,
				sizeof(char *));
		    for (i = 0; i < MAX_PAIR_STRS; i++)
			pDm_llist_el->pNext_el->ppMacro_pairs[i] = pRti_rel_disp->ppMacros[index][i];
		} else
		    pDm_llist_el->pNext_el->ppMacro_pairs = NULL;
	    }
	}

	if (pRti_rel_disp->pDl_related_display->selfkill) {

	    /* Kill this display just like hitting the exit from the menu */

	    dmExit_display(pRti_rel_disp->pDm_llist_hdr, pRti_rel_disp->pDm_llist_el, NULL); /* <- note the NULL! */

	}

    }

 
/************************************************************************************
*
*	display_matches
*
*	display_matches compares the display_name and the arg list of the given
*	display and the related display for equivalence
*
*	NOTE: args must be the same AND be given in the same order to be considered equal
*
*	Input	: pRti_rel_disp 	ptr to related display struct
*		: index			index of the desired display in the struct
*		: pDm_llist_el		ptr to the display to check
*	Returns	: return TRUE if the same, otherwise FALSE.
*/

display_matches(pRti_rel_disp, index, pDm_llist_el)
struct rti_rel_disp	*pRti_rel_disp;
int					index;
struct dm_llist_el	*pDm_llist_el;
	{
#define	MATCH	TRUE
#define	NOMATCH	FALSE
    char	**pdisplayMacros		= pDm_llist_el->ppMacro_pairs;
    char	**prelatedDisplayMacros	= pRti_rel_disp->ppMacros[index];

    if (strcmp(pDm_llist_el->display_name, pRti_rel_disp->pDisplays[index]))
		/* display names do not match */
		return NOMATCH;

    if ((pdisplayMacros!=0) ^ (prelatedDisplayMacros!=0))
		/* either the display or the related display has macros, but not both */
		return NOMATCH;

    if (pdisplayMacros && prelatedDisplayMacros)
		{
    	int	i;

		/* both the display and the related display have macros, need to check each one */
		for (i = 0; i < MAX_PAIR_STRS; i++)
			{
			if ((pdisplayMacros[i]!=0) ^ (prelatedDisplayMacros[i]!=0))
				/* either the display or the related display have this arg, but not both */
				return NOMATCH;
			if (pdisplayMacros[i] && prelatedDisplayMacros[i])
				/* they both have an arg in this position, check it */
				if (strcmp(pdisplayMacros[i], prelatedDisplayMacros[i]))
					/* the macro contents are different */
					return NOMATCH;
			}
		}
    /* they are the same */
    return MATCH;
	}

	    
/***********************************************************************
**
**	Delete_related_displays
**	
**	Deb Kerstiens 4/91
**
**	Purpose : This routine deletes the "gizmos" that have been created
**			for operator control.  Delete_controllers is called
**			when a display (dm_llist_el) is to be deleted to delete 
**			each gizmo that is currently active.  This must be
**			done so that when the parent (display) window is
**			deleted, the child window created for each gizmo will
**			already have been deleted.
**
**	Input :
**	    prti	: ptr to the run-time list for related_display
**	    nctrl	: count of related_display elements to display
**	Output :
**	    none
**	Returns
**	    none
**
**********************************************************************/

void delete_related_displays(prti, nrd)
struct rti_rel_disp	*prti;
unsigned short		nrd;
	{
	unsigned short	i, j;
 
#ifdef DEBUG
printf("~in delete_related_displays\n");
#endif

	/* for each element on the display list */
	for (i = 0; i < nrd; prti++, i++)
		{
		if (prti->pgz)
			free_gizmo(prti->pgz);
		for (j = 0; j < MAX_REL_DISPLAYS;j++)
			if (prti->pMenustr[j] && prti->ppMacros[j] && (prti->ppMacros[j] != prti->pDm_llist_el->ppMacro_pairs))
				tisFree(prti->ppMacros[j]);
		}
	}

/*----------*/
static void stripTrail (s)

    char	s [];

    {

	int	i;

	i = strlen (s) - 1;
	while (i >= 0) {
	    if (s [i] != ' ')
		break;
	    i--;
	}
	s [i + 1] = '\0';

    }
