#include "opiStringDefs.h"
extern	char	*opiStringTab[];
/*t_dbiocsubs.c,v 1.1.1.1 1999/08/03 14:43:49 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 * DBIOCSUBS.C
 *
 * contains the routines to process database records
 *
 *
 * Author: 	Bob Dalesio
 * Date:	7-9-87
 * @(#)dbiocsubs.c	1.1	9/22/88
 *
 *	Control System Software for the GTA Project
 *
 *	Copyright 1988, 1989, the Regents of the University of California.
 *
 *	This software was produced under a U.S. Government contract
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory, which is
 *	operated by the University of California for the U.S. Department
 *	of Energy.
 *
 *	Developed by the Controls and Automation Group (AT-8)
 *	Accelerator Technology Division
 *	Los Alamos National Laboratory
 *
 *	Direct inqueries to:
 *	Bob Dalesio, AT-8, Mail Stop H820
 *	Los Alamos National Laboratory
 *	Los Alamos, New Mexico 87545
 *	Phone: (505) 667-3414
 *	E-mail: dalesio@luke.lanl.gov
 *
 *	f_to_str	convert a float to a string
 *	    NOTE:	This only converts numbers < 10,000,000
 *	    args
 *		value		floating point value
 *		pstr		pointer to the string
 *		precision	number of places to the right of the dec point
 *	f_to_estr	convert a float to a e-notation string
 *	    NOTE:	This only converts numbers < 10,000,000
 *	    args
 *		value		floating point value
 *		pstr		pointer to the string
 *		precision	number of places to the right of the dec point
 *		e-resolution	1 - regular e notation 3 - engineering
 *
 *	Modification History:
 *	    11/25/91	DK	modified f_to_estr to elimate leading zeros
 *	    11/26/91	DK	fixed bug in f_to_estr caused by floating
 *				point precision error in divide
 */

#if defined(AXP) || defined(WIN32)
#include <string.h>
#endif

 /* FF_TO_STR
 *
 * Faster Float to Null terminated string conversion routine
 *
 * This routine converts numbers less than 10,000,000. It defers to f_to_str for
 * numbers requiring more than 8 places of precision. There are only eight decimal
 * places in a 32 bit long integer. The primary difference between f_to_str and
 * ff_to_str is that ff_to_str performs its conversions mostly in integer
 * operations.
 * This takes about 167.8 microseconds.
 */
static long	frac_multiplier[] = {1,10,100,1000,10000,100000,1000000,10000000,100000000};
ff_to_str(flt_value,pstr_value,precision)
float			flt_value;
register char		 *pstr_value;
unsigned short		precision;
{
        unsigned short		got_one,i;
        register char   	*pfirst_digit;
	register long		whole,iplace,number,fraction,fplace;
	float			ftemp;

	/* can this routine handle this conversion */
	if (precision > 8)
		return(f_to_str(flt_value,pstr_value,precision));

	/* determine the sign */
        pfirst_digit = pstr_value;
        if (flt_value < 0){
                *pstr_value = '-';
                flt_value = -flt_value;
        }else{
		*pstr_value = ' ';
	}

	/* check upper bound */
        if (flt_value >= 10000000){
                strcpy(pstr_value,opiStringTab[STR_Number_Too_Large            ]);
                return(-1);
        }

	/* space for the sign indicator */                
        pstr_value++;
        pfirst_digit++;

 	/* remove the whole number portion */
	whole = (long)flt_value;
	ftemp = (float)(flt_value - whole);

	/* multiplier to convert fractional portion to integer */
	fplace = frac_multiplier[precision];
	fraction = (long)(ftemp * fplace * 10);	
	fraction = (fraction + 5) / 10;		/* round up */

	/* determine rounding into the whole number portion */
	if ((fraction / fplace) >= 1){
		whole++;
		fraction -= fplace;
	}

	/* whole numbers */
        got_one = 0;
        for (iplace = 1000000; iplace >= 1; iplace /= 10){
                if (whole >= iplace){
                        got_one = 1;
                        number = whole / iplace;
                        whole = whole - (number * iplace);
                        *pstr_value = number + '0';
                        pstr_value++;
                }else if (got_one){
                        *pstr_value = '0';
                        pstr_value++;
                }
        }
        if (!got_one){
                *pstr_value = '0';
                pstr_value++;
        }

        /* fraction */
        if (precision > 0){
		/* convert fractional portional to ASCII */
                *pstr_value = '.';
                pstr_value++;
                for (fplace /= 10, i = precision; i > 0; fplace /= 10,i--){
                        number = fraction / fplace;
                        fraction -= number * fplace;
                        *pstr_value = number + '0';
                        pstr_value++;
                }
        }
        *pstr_value = 0;

        return(0);
}

/*
 * F_TO_STR
 *
 * converts floating point numbers to NULL terminated strings
 * 
 * This routine was written to replace the use of SSCANF which takes roughly
 * 24.7 Milliseconds to perform a conversion. This routine will not convert
 * any number with an absolute value greater than or equal to 10,000,000.00.
 * This takes about 289.7 microseconds.
 */
f_to_str(flt_value,pstr_value,precision)
register float		flt_value;
register char		*pstr_value;
register unsigned short	precision;
{
	register unsigned short	got_one;
	register float		place;
	register short		number;
	register char	*pfirst_digit;

	/* determine the first character */
	pfirst_digit = pstr_value;
	if (flt_value < 0){
		*pstr_value = '-';
		flt_value = -flt_value;
	}else{
		*pstr_value = ' ';
	}

	/* is this value in range */
	if (flt_value >= 10000000){
		strcpy(pstr_value,opiStringTab[STR_Number_Too_Large            ]);
		return(-1);
	}
	pstr_value++;
	pfirst_digit++;

	/* whole numbers */
	got_one = 0;
	for (place = (float)1000000.; place >= 1; place /= 10){
		if (flt_value >= place){
			got_one = 1;
			number = (short)(flt_value / place);
			flt_value = flt_value - (number * place);
			*pstr_value = number + '0';
			pstr_value++;
		}else if (got_one){
			*pstr_value = '0';
			pstr_value++;
		}
	}
	if (!got_one){
		*pstr_value = '0';
		pstr_value++;
	}

	/* fraction */
	if (precision > 0){
		*pstr_value = '.';
		pstr_value++;
		for (place = (float).1; precision > 0; place /= 10, precision--){
			number = (short)(flt_value / place);
			flt_value = flt_value - (number * place);
			*pstr_value = number + '0';
			pstr_value++;
		}
	}
	*pstr_value = 0;

	/* rounding */
	if (flt_value >= (place * 5)){
		number = pstr_value - pfirst_digit;
		pstr_value--;
		while ((pstr_value >= pfirst_digit) && 
		  ((*pstr_value == '9') || (*pstr_value == '.'))){
			if (*pstr_value == '9')
				*pstr_value = '0';
			pstr_value--;
		}
		if (pstr_value < pfirst_digit){
			while (number >= 0){
				*(pfirst_digit + number + 1) = *(pfirst_digit + number);
				number--;
			}
			*pfirst_digit = '1';
		}else{
			*pstr_value += 1;
		}
	}
	return(0);
}

/*
 * F_TO_ESTR
 *
 * converts floating point numbers to NULL terminated strings
 * 
 * This routine was written to replace the use of SSCANF which takes roughly
 * 34.4 Milliseconds to perform a e -notation conversion. This routine will not 
 * convert any number with an absolute value greater than or equal to 
 * 10,000,000.00. Takes about 375 microseconds.
 */

static float round_up[] = { (float).5, (float).05, (float).005,(float).0005,
			    (float).00005,(float).000005,(float).0000005,(float).00000005,
			    (float).000000005,(float).0000000005,(float).00000000005,
			    (float).000000000005,(float).0000000000005,(float).00000000000005,
			    (float).000000000000005,(float).0000000000000005};

f_to_estr(flt_value,pstr_value,precision,e_resolution)
register float		flt_value;
register char		*pstr_value;
register unsigned short	precision,e_resolution;
{
	register float	place,divisor;
	register short	e;
	unsigned short	i;
	short		number;

	if (flt_value < 0){
		*pstr_value = '-';
		pstr_value++;
		flt_value = -flt_value;
	}else if (flt_value == 0.0){
		*pstr_value = ' ';
		pstr_value++;
		*pstr_value = '0';
		pstr_value++;
		/* fraction */
		*pstr_value = '.';
		pstr_value++;
		for (place = (float).1; precision > 0; place /= (float)10.0, precision--){
			*pstr_value = '0';
			pstr_value++;
		}
		*pstr_value = 'e';
		pstr_value++;
		*pstr_value = '+';
		pstr_value++;
		*pstr_value = '0';
		pstr_value++;
		*pstr_value = '0';
		pstr_value++;
		*pstr_value = '\0';
		return(0);

	}else{
		*pstr_value = ' ';
		pstr_value++;
	}

	if (flt_value >= 1000000000000000.0){
		strcpy(pstr_value,opiStringTab[STR_Number_Too_Large            ]);
		return(-1);
	}

	/* determine which 3rd power of 10 to use */
	for (i=0,divisor=(float)1.; i<e_resolution; divisor*=10,i++);
	for (e = 12, place = (float)1000000000000.0;
	    flt_value < place;
	    e -= e_resolution, place /= divisor);
	if (e < -99){
		strcpy(pstr_value,opiStringTab[STR_Number_Too_Small            ]);
		return(-1);
	}

	/* whole numbers */
	flt_value = flt_value/place + round_up[precision];
	for (place = (float)100.0; place >= 1.0; place /= (float)10.0){
		if (flt_value >= place){
			number = (short)(flt_value / place);
			flt_value = flt_value - (number * place);
			*pstr_value = number + '0';
			pstr_value++;
		}
	}

	/* fraction */
	*pstr_value = '.';
	pstr_value++;
	for (place = (float).1; precision > 0; place /= (float)10.0, precision--){
		number =  (short)(flt_value / place) ; 
		flt_value = flt_value - (number * place);
		*pstr_value = number + '0';
		pstr_value++;
	}

	/* exponent portion */
	*pstr_value = 'e';
	pstr_value++;
	if (e < 0){
		*pstr_value = '-';
		e = -e;
	}else{
		*pstr_value = '+';
	}
	pstr_value++;
	number = e / 10;
	*pstr_value = number + '0';
	pstr_value++;
	e -= number * 10;
	*pstr_value = e + '0';
	pstr_value++;
	*pstr_value = 0;

	return(0);
}

