/*update_plots.c,v 1.1.1.1 1999/08/03 14:43:52 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include <Xlib.h>      			/* the X library 		*/
#include <Xutil.h>
#include <db_access.h>			/* db types and structs		*/
#include <alarm.h>
#include <cadef.h>
#include <dldef.h>			/* display list defines 	*/
#include <dlinfo.h>			/* more display list defines 	*/
#include <font_table.h>			/* defines for font table	*/
#include <dl_context.h>			/* defines dl context		*/
#include <gizmo.h>			/* defs for gizmos */
#include <dynamic_def.h>		/* rti dynamic defines		*/
#include <op_gsdl_def.h>		/* com_table element def	*/
#include <pfoPlotDefs.h>		/* defs for opi plot stuff	*/
#include <GraphX.h>

#ifdef AXP
#include <string.h>
#endif

extern void *gsd_sync_init();

unsigned char init_bits[] = { 0x1,
			0x3,
			0x7,
			0xf,
			0x1f,
			0x3f,
			0x7f,
			0xff
};
			

/***********************************************************************
**
**	UPDATE_PLOTS
**	
**	Deb Kerstiens 11/90
**
**	Purpose : This routine sets up the run-time interface for the plots
**		defined using the display editor.  Attributes defined include
**		1 to 7 process variables used as values to plot, the frequency
**		at which to plot each new data set(strip chart), colors for each
**		value line/point, etc. (see ../h/dldef.h)  Plots supported 
**		currently include strip charts,  smith overlay plots, and a 
**		variety of Cartesian plots using points or lines to represent
**		either scalar or array data values.
** 
**		To initialize, channel access connections are attempted for each
**		process variable.  If a connection is not established (or lost at
**		any time), the corresponding trace will reflect the minimum value 
**		defined for the strip chart.  At this time connection management is
**		is not fully implemented for other plots.  The data will not be
**		presented until at least one valid trace is connected, but will 
**		not show disconnect status.  This will be implemented at a later
**		time.  For all plots except strip charts, synchronous data  routines
**		are used to fetch the data.  This will likely be changed to accomodate
**		both sync, nonsync and combinations.
** 
**		Update_plots() is called by the continuous(forever) update loop of
**		display_mgr.  When an expose event occurs on a plot window, the routine
**		that handles the updating of that type of plot is called directly. 	   	
**
**
**	Modification History:
*	    04/16/92	dk	added time clamping to keep plots from updating too fast
*				(dm and ioc cpu usage was excessive in this mode)
*	    04/21/92	dk	fixed bug in do_smith() and do_plot() that caused a call
*				to the plot routines before sync data was available and
*				aborted dm (pfoPlot).
*/
/****************************************************************************************
**	update_plots()
**
**	Pupose	: to initialize and/or update each plot defined in the display list		 	
**	
**	Input :
**	    pdl_context : pointer to Display context structure 
**	    pdl_index	: prt to index for display list color map
**	    pdl		: ptr to display list plot info
**	    prti	: ptr to the rti_dynamic plot list
**	    nplots	: count of dynamic plot elements to display
**	    expose_flag	: indicates whether called as result of expose
**	    ppfdctx	: pointer to file descriptor mgr's context
**	Output :
**	    none
**	Returns
**	    none
**
**********************************************************************/

void update_plots(pdlc, pdl_index, pdl, prti, nplots,prec,expose_flag, ppfdctx,strip_elapsed,plot_elapsed)
#ifndef AXP
int 			*pdl_index;
#else
unsigned long		*pdl_index;
#endif
struct dl_context	*pdlc;
void 			*pdl;
struct rti_update_plot 	*prti;
unsigned short		nplots;
XRectangle		*prec;
unsigned short		expose_flag;
void **			ppfdctx;
float strip_elapsed,plot_elapsed;
{
 void *dis_ptr;
  unsigned short i;
  struct rti_update_plot *pRti_plot; 
  void			init_strip();
  void			do_plot();
  void			do_smith();
#ifdef DEBUG
printf("~in update_plots\n");
#endif

/* for each element on the display list */
for (	dis_ptr = pdl, pRti_plot = prti, i = 0;  
	i < nplots; 
	dis_ptr = (short *)dis_ptr + ((struct dl_element_hdr *)dis_ptr)->nwords,   
	pRti_plot++, i++){
    register struct dl_object_hdr *objptr = 
		&(((struct dl_anygo *)dis_ptr)->object);

    if(pRti_plot->num_traces < 1){
/*	printf("Invalid plot-- no traces\n"); error to operator ??*/
	continue; 
    } 
    /* if we've deleted a plot, don't do anything */
    if ((pRti_plot->pSync_ctrl == NULL)
      && (((struct dl_element_hdr *)dis_ptr)->type != dl_strip)) {
	printf("~in update_plots, pSync_ctrl = NULL\n");
	continue;
    }
	

    /* if expose, update exposed area */
    if(expose_flag){
    	if( (objptr->x+objptr->width) < prec->x )
	    continue;
    	if( objptr->x > (prec->x + (unsigned int)(prec->width)) )
	    continue;
    	if( (objptr->y+objptr->height) < prec->y )
	    continue;
   	if( objptr->y > (prec->y + (unsigned int)(prec->height)) )
	    continue;
	    
    }

    /* else this is not an expose, check for modification ???*/


    /* if we made it here, we need to update the display --
	either exposed or modified */
   
    switch(((struct dl_element_hdr *)dis_ptr)->type){

	case dl_strip : {
	    if(strip_elapsed > UPDATE_STRIP_TIME){ 

	    	/* get ranges for trace vals, set up for op input, etc.) */
	    	init_strip(pRti_plot);
#ifdef OLD_STRIP
	    	gsd_sync_read(NEXTSET_SYNC_DATA,
			pRti_plot->pSync_ctrl,
			pRti_plot->num_traces,ppfdctx,.1);
#endif OLD_STRIP
		}

 	    break;
	
	} /* end case dl_strip */


	case dl_plot :{
	    int data_done;
	    if(plot_elapsed > UPDATE_PLOT_TIME){
	    	data_done = gsd_sync_read(NEXTSET_SYNC_DATA,
			pRti_plot->pSync_ctrl,
			pRti_plot->num_traces,ppfdctx,.1);
	    	/* initiate plotting of any new chans just connected,plot all inited */
	    	do_plot(pRti_plot,expose_flag,data_done);
	    	/* initiate next synchrononous read--will have to set up better timeout(.1)later */
	    	if(data_done)gsd_sync_read(NEXTSET_SYNC_DATA,
			pRti_plot->pSync_ctrl,
			pRti_plot->num_traces,ppfdctx,.1);
	    }
	    break;

	}/* end dl_plot*/


	case dl_smith_overlay :{
	    int data_done;
	    if(plot_elapsed > UPDATE_PLOT_TIME) {
	    		data_done = gsd_sync_read(NEXTSET_SYNC_DATA,
		pRti_plot->pSync_ctrl,
			pRti_plot->num_traces,ppfdctx,.1);
	   	 /* initiate plotting of any new chans just connected,plot all inited */
	    	do_smith(pRti_plot, expose_flag,data_done);
	    	/* initiate next synchrononous read--will have to set up better timeout(.1)later */
	    	if(data_done)gsd_sync_read(NEXTSET_SYNC_DATA,
			pRti_plot->pSync_ctrl,
			pRti_plot->num_traces,ppfdctx,.1);
	    }

	    break;

	}/* end dl_smith_overlay */



	case dl_surface_plot :{

	    printf("~In Update_plots, type dl_surface_plot not implemented yet\n");
	    break;

	}/* end dl_surface_plot */


	case dl_2d_control :{ /* don't know if this should be here?? */

	    printf("~In Update_plots, type dl_surface_plot not implemented yet\n");
	    break;

	}/* end dl_2d_control */


	default : {
	    printf("~in update_plots--took default \n");
	    break;
	}				
		
    }/* end switch */


  } /* end for each el on list*/

  XFlush(pdlc->d);
}/* end update_plots */




/***********************************************************************
**
**	init_strip
**	
**	Deb Kerstiens 1/91
**
**	Purpose :  Determine and set ranges for the traces on the strip chart. 
**		Later will set up operator interaction for pan, zoom, change 
**		process variables, etc.  
** 
**	Note	:
**		To initialize, channel access connections are attempted for each
**		process variable.  If a connection is not established (or lost at
**		any time), the corresponding trace will reflect the minimum value 
**		defined for the strip chart.  
**	   	
**		 	
**	
**	Input :
**	    pRti_plot	: ptr to the rti_dynamic plot element
**	Output :
**	    none
**	Returns
**	    none
**
**********************************************************************/ 


void init_strip(pRti_plot)
  struct rti_update_plot *pRti_plot; 
{
    unsigned char	draw_flag, got_one;
    unsigned short	prec;
    short 	field_type;
    register unsigned	i;
    float hopr,lopr;

    got_one = FALSE;
    draw_flag = (!pRti_plot->init_bits)? TRUE : FALSE;

    for (i = 0; i < pRti_plot->num_traces;i++){
      /* if this trace isn't inited */
      if((pRti_plot->init_bits & (1 << i)) == 0x0){
	/* if pv not connected or mem not allocated , return */
 	if(( (pRti_plot->pYlim[i]->pChan != NULL) &&
            ( (field_type = ca_field_type(pRti_plot->pYlim[i]->pChan))
                == TYPENOTCONN))
            ||(pRti_plot->pYlim[i]->pData_struct == NULL))
		continue;
	/* else */
	get_hoprnlopr(pRti_plot->pYlim[i], field_type, &hopr,
	    &lopr, &prec);
	pRti_plot->stripRange[i].maxVal = hopr;
	pRti_plot->stripRange[i].minVal = lopr;
	/* can't tell if data back yet */	
	if((hopr != 0.0)  || (lopr != 0.0)){
	    pRti_plot->init_bits = pRti_plot->init_bits |  (1 << i);
	    got_one = TRUE;
	    stripSetRange(pRti_plot->pPlot, i, (double)lopr, (double)hopr);
	}
      } /* end if not inited */
    }/* end for */
    /* if we haven't drawn yet and we have a channel, begin the
	 strip chart -- it will be updated periodically from now on */	
    if(draw_flag && got_one){
	 stripDraw(((Strip *)pRti_plot->pPlot));
    }
}



/***********************************************************************
**
**	do_plot
**	Deb Kerstiens 2/91
**
**	Purpose :  Determines ranges for each trace and begins plotting of each
**		trace on the plot when a connection is established.  Traces already
**		inited will be plotted each time through. Traces have been defined
**		as a synchronous data set and checks for synchronicity are done by
**		the pfo plotting routines.
**		done. 
** 
**	Note	: operator interaction for pan, zoom, change process variables, etc. 
**			will be added someday and handled here 
**	
**	   	
**		 	
**	
**	Input :
**	    pRti_plot	: ptr to the rti_dynamic plot element
**	    data_done	: flag to indicate whether sync data take done
**	    expose_flag	: if true, display has been exposed
**	Output :
**	    none
**	Returns
**	    none
**
**********************************************************************/ 

void do_plot(pRti_plot, expose_flag, data_done)
  struct rti_update_plot *pRti_plot;
  unsigned short expose_flag;
  int data_done; 
{
    unsigned char got_one;
    unsigned short	prec;
    short 	field_type;
    register unsigned	i;
    float hopr,lopr;
    int plot_type;

    got_one = FALSE;
    for (i = 0; i < pRti_plot->num_traces;i++){
      /* if this trace isn't inited */
      if((pRti_plot->init_bits & (1 << i)) == 0x0){
	/* if pv not connected or mem not allocated , return */
	if(( (pRti_plot->pYlim[i]->pChan != NULL) &&
            ( (field_type = ca_field_type(pRti_plot->pYlim[i]->pChan))
                == TYPENOTCONN))
            ||(pRti_plot->pYlim[i]->pData_struct == NULL) ||
		( (pRti_plot->ysync_data[i].pChid != NULL) &&
            ( (field_type = ca_field_type(pRti_plot->ysync_data[i].pChid))
                == TYPENOTCONN))
            ||(pRti_plot->ysync_data[i].pSdata == NULL))
		continue;
	else{
	    get_hoprnlopr(pRti_plot->pYlim[i], field_type, &hopr,
	   	 &lopr, &prec);
	}
	/* can't tell if data back yet */	
	if((hopr != 0.0)  || (lopr != 0.0)){
	    pRti_plot->init_bits = pRti_plot->init_bits |  (1 << i);
	    /* add channel to be plotted */
	    pfoAxisOpen(pRti_plot->pPlot,
		(int)(((i == 0)&&(pRti_plot->rti_plottype == 2)) ? 1:0),	/* isX */
		(int)0,			/* fill if no sync data */
		pRti_plot->data_clrs[i],
		&(pRti_plot->ysync_data[i]),
		pRti_plot->pYlim[i]->pData_struct,
		dbf_type_to_DBR_CTRL(field_type));
	    got_one = TRUE;
	    
	}
      } /* end if not inited */
    }/* end for */
    /* if we have any chans and if the x chan is one of them, draw them */
    if (pRti_plot->init_bits != 0x0) {
	/* if we have a new channel or this is an expose event */
	if(got_one || expose_flag){
	    pfoWinRepaint( pRti_plot->pPlot, 
		(enum dl_plot_style)((struct dl_plot *)pRti_plot->pDl_plot)->style,
			 pRti_plot->rti_plottype);
	}
	/* plot data if we have new data or if exposed */
	if(data_done || expose_flag){
	    /* if plot ix y vs x and no x data yet, don't do anything */
	    if(!((pRti_plot->rti_plottype == 2)&&((pRti_plot->init_bits & 0x1) == 0x0))){
  	    	pfoWinPlotSample(pRti_plot->pPlot);
	    }
	}
    }
}/* end do_plot */




/*******************************************************************************************
*
*	setup_xyplot
*
*	2/91	DK
*
*	Setup_xyplot initializes both the rti_update_plot struct for the 
*		plot and the pXY struct (for the  routines to use info
*		from the display list).  It initiates a synchronous read for
*		the set of process variables. 
*
* 
*
*	Input	
*	   pdm_llist_el	: ptr to display manager's llist el for this display 
*	   pdlc  	: ptr to the dl_context struct for this display
*	   pPlot_el	: ptr to the run-time struct for this plot
*	   ppfdctx	: ptr to the file descriptor mgr's context for sync data
*	Output	: none
*	
*	Returns	: nothing
*
*/



setup_xyplot(pdm_llist_el,pdlc, pPlot_el, ppfdctx)
struct dm_llist_el 	*pdm_llist_el;
struct dl_context 	*pdlc;
struct rti_update_plot 	*pPlot_el;
void 			**ppfdctx;
{
  int				num_data_store;	/* number of iterations of data to save (for refresh) */
  register short 		i,j;
  unsigned 			ix;
  struct dl_pc 			*pDl_pc;
  struct dl_object_hdr 		*pRec = &(((struct dl_anypo *)pPlot_el->pDl_plot)->object);
  register unsigned short	x_cnt, y_cnt;	/* num x,y traces for this plot*/
  register struct dl_plot 	*pDl_xyplot;

    pDl_xyplot =  ((struct dl_plot *)pPlot_el->pDl_plot);

    for (i = 0, x_cnt = 0, y_cnt = 0;i < 6; i++){

	/* get x channel info--only do one x for now */
	if(i == 0){
	    ix = (int)(pDl_xyplot->trace[i].xdata);
	    pDl_pc = (struct dl_pc *)(((short *)pdm_llist_el->pdl_hdr)
		 + pdm_llist_el->pdl_hdr->opc);
	    for(j = 0; j < ix;j++){
    		pDl_pc = (struct dl_pc *)((short *)pDl_pc + ((struct dl_element_hdr *)pDl_pc)->nwords);
	    }
	    if(strlen((char *)(pDl_pc + 1)) > 0){    
  		pPlot_el->pYlim[y_cnt] = &(pdm_llist_el->pcom_el->pCom_table[ix]);
  		pPlot_el->ysync_data[y_cnt].pName = (char *)(pDl_pc + 1);
  		pPlot_el->data_clrs[y_cnt++] = COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_xyplot->plotcom.clr);
		pPlot_el->rti_plottype = 2; /* ydata vs xdata */
	    }
	    else {
		pPlot_el->rti_plottype = 4; /* ydata vs bin (waveform)*/
	    }
	}/* end setup x data */

	/* get y channel info */

	    ix = (int)(pDl_xyplot->trace[i].ydata);
	    pDl_pc = (struct dl_pc *)(((short *)pdm_llist_el->pdl_hdr) + pdm_llist_el->pdl_hdr->opc);
            for(j = 0; j < ix;j++){
	    	pDl_pc = (struct dl_pc *)((short *)pDl_pc + ((struct dl_element_hdr *)pDl_pc)->nwords);
	    }
	    if(strlen((char *)(pDl_pc + 1)) > 0){
	    /* set up data fetch/conversion */
	    	pPlot_el->pYlim[y_cnt] = &(pdm_llist_el->pcom_el->pCom_table[ix]);
	    	pPlot_el->ysync_data[y_cnt].pName = (char *)(pDl_pc + 1);
	    	pPlot_el->data_clrs[y_cnt++] = COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_xyplot->trace[i].data_clr);

	    }
	 /*end setup y data */
	
    }/* end for i*/

    /* create a window for the xy plot with info from display list (x,y,ht,wd,clr,bclr)*/
    /* get plot coords (x,y,h,w ) from display list*/
    pPlot_el->window= XCreateSimpleWindow(pdlc->d, pdlc->w,
    	pRec->x,pRec->y, pRec->width,pRec->height,
    	1,	/*border width */
    	COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_xyplot->plotcom.clr),
    	COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_xyplot->plotcom.bclr));

    /* Select exposure events and button press events. */

     XSelectInput(pdlc->d,pPlot_el->window,ExposureMask|ButtonPressMask|
	       ButtonReleaseMask);


    /* Tell the server to make the window visible. */
    XMapWindow(pdlc->d,pPlot_el->window);

    /* set up for handling Xevents on window id and for flagging plots */
    pPlot_el->evwin.plot_win = TRUE;
    pPlot_el->evwin.puser = pPlot_el;
    XSaveContext(pdlc->d,pPlot_el->window,
		pdm_llist_el->win_context, (XPointer)&(pPlot_el->evwin));
	
    /* set up channel access for synchronous reads */
    /* need a ctrl ptr for each xyplot */
    pPlot_el->num_traces = y_cnt;
    pPlot_el->pSync_ctrl = gsd_sync_init(pPlot_el->ysync_data,
	pPlot_el->num_traces, ppfdctx, pdm_llist_el->ppMacro_pairs);
    gsd_sync_read(NEXTSET_SYNC_DATA, pPlot_el->pSync_ctrl,
	 pPlot_el->num_traces, ppfdctx, .1);

    if((enum dl_plot_style)pDl_xyplot->style == dl_point_plot){
	num_data_store = 50;
    }
    else{
	 num_data_store = 0;
    }

    /* init the xy plot */
    pPlot_el->pPlot = (caddr_t) pfoWinOpen(pdlc->d,
	pPlot_el->window,
	COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_xyplot->plotcom.clr),
	COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_xyplot->plotcom.bclr), 
	num_data_store,
	pdm_llist_el->pdl_strings[(int)(pDl_xyplot->plotcom.title)],
	pdm_llist_el->pdl_strings[(int)(pDl_xyplot->plotcom.xlabel)],
	pdm_llist_el->pdl_strings[(int)(pDl_xyplot->plotcom.ylabel)]);
	
    return;
}/* end setup_xyplot */



/***********************************************************************
**
**	do_smith
**	Deb Kerstiens 3/91
**
**	Purpose :  Determines ranges for each trace and begins plotting of each
**		point pair on the plot when a connection is established.  Those already
**		inited will be plotted each time through. Pt pairs have been defined
**		as a synchronous data set and checks for synchronicity are done 
**		by the pfo plotting routines
** 
**	Note	: operator interaction for pan, zoom, change process variables, etc. 
**			will be added someday and handled here 
**	
**	   	
**		 	
**	
**	Input :
**	    pRti_plot	: ptr to the rti_dynamic plot element
**	    data_done	: flag to indicate whether sync data take done
**	    expose_flag	: if true, display has been exposed
**	    data_done	: indicates sync data retrieve done 
**				(got sync data or timed out)
**	Output :
**	    none
**	Returns
**	    none
**
**********************************************************************/ 


void do_smith(pRti_plot, expose_flag, data_done)
  struct rti_update_plot *pRti_plot;
  unsigned short expose_flag;
  int data_done; 
{
    unsigned char got_one;
    unsigned short	prec;
    short 	field_type;
    register unsigned	i;
    float hopr,lopr;
    int plot_type;

    got_one = FALSE;
    for (i = 0; i < pRti_plot->num_traces;i++){
      /* if this trace isn't inited */
      if((pRti_plot->init_bits & (1 << i)) == 0x0){
	/* if pv not connected or mem not allocated , return */
	if(( (pRti_plot->pYlim[i]->pChan != NULL) &&
            ( (field_type = ca_field_type(pRti_plot->pYlim[i]->pChan))
                == TYPENOTCONN))
            ||(pRti_plot->pYlim[i]->pData_struct == NULL) ||
		( (pRti_plot->ysync_data[i].pChid != NULL) &&
            ( (field_type = ca_field_type(pRti_plot->ysync_data[i].pChid))
                == TYPENOTCONN))
            ||(pRti_plot->ysync_data[i].pSdata == NULL))
		continue;
	else{
	    get_hoprnlopr(pRti_plot->pYlim[i], field_type, &hopr,
	   	 &lopr, &prec);
	}
	/* can't tell if data back yet */	
	if((hopr != 0.0)  || (lopr != 0.0)){
	    pRti_plot->init_bits = pRti_plot->init_bits |  (1 << i);
	    /* add channel to be plotted */
	    pfoAxisOpen(pRti_plot->pPlot,
		(int)((i== 0) ? 1:0),		/* isX */
		(int)0,			/* fill if no sync data */
		pRti_plot->data_clrs[i],
		&(pRti_plot->ysync_data[i]),
		pRti_plot->pYlim[i]->pData_struct,
		dbf_type_to_DBR_CTRL(field_type));
	    got_one = TRUE;
	    
	}
      } /* end if not inited */
    }/* end for */
    /* if we have both chans, draw them */
    if (pRti_plot->init_bits == 0x3) {
	/* if we have a new channel or this is an expose event */
	if(got_one || expose_flag){
	    pfoWinRepaint( pRti_plot->pPlot, 
		dl_point_plot,
		pRti_plot->rti_plottype);
	}
	/* plot data if we have new data or if exposed */
	if(data_done || expose_flag){
  	    pfoWinPlotSample(pRti_plot->pPlot);
	}
    }/* end if chans */

}/* end do_smith */




/*******************************************************************************************
*
*	setup_smith
*
*	3/91	DK
*
*	Setup_smith initializes the rti_update_plot struct for the smith overlay point pair
*		plot. It initiates a synchronous read for the set of process variables. 
* 
*
*	Input	: pdm_llist_el-- ptr to display manager's llist el for this display 
*		: pdlc -- ptr to the dl_context struct for this display
*		: pPlot_el-- ptr to the run-time struct for this plot
*		: ppfdctx-- ptr to address of file descriptor mgr's id for channel access
*			(needed for sync read )
*	Output	: fills in rti_update plot stuct for later use
*	
*	Returns	: nothing
*
*/



setup_smith(pdm_llist_el,pdlc, pPlot_el, ppfdctx)
struct dm_llist_el 	*pdm_llist_el;
struct dl_context 	*pdlc;
struct rti_update_plot 	*pPlot_el;
void 			**ppfdctx;
{
  int				num_data_store;	/* number of iterations of data to save (for refresh) */
  short 			index;
  unsigned short		i;
  struct dl_pc 			*pDl_pc;
  struct dl_object_hdr 		*pRec = &(((struct dl_anypo *)pPlot_el->pDl_plot)->object);
  register struct dl_smith_overlay 	*pDl_smith;
  int				color;

pDl_smith =  ((struct dl_smith_overlay *)pPlot_el->pDl_plot);
pDl_pc = (struct dl_pc *)(((short *)pdm_llist_el->pdl_hdr) + pdm_llist_el->pdl_hdr->opc);
for(i = 0; i <  (int)(pDl_smith->xdata); i++){
    pDl_pc = (struct dl_pc *)((short *)pDl_pc + ((struct dl_element_hdr *)pDl_pc)->nwords);
}
if(strlen((char *)(pDl_pc + 1)) > 0){
    
  pPlot_el->pYlim[0] = &(pdm_llist_el->pcom_el->pCom_table[(int)(pDl_smith->xdata)]);
  pPlot_el->ysync_data[0].pName = (char *)(pDl_pc + 1);
  pPlot_el->data_clrs[0] = COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_smith->plotcom.clr);

}
else{
    pPlot_el->rti_plottype = 0;
    /* error message ?? */
    return;
}

/* get y process variable names for ca (sync read) */
pDl_pc = (struct dl_pc *)(((short *)pdm_llist_el->pdl_hdr) + pdm_llist_el->pdl_hdr->opc);
for(i = 0; i <  (int)(pDl_smith->ydata); i++){
    pDl_pc = (struct dl_pc *)((short *)pDl_pc + ((struct dl_element_hdr *)pDl_pc)->nwords);
}
if(strlen((char *)(pDl_pc + 1)) > 0){
/* set up data fetch/conversion */
    pPlot_el->pYlim[1] = &(pdm_llist_el->pcom_el->pCom_table[(int)(pDl_smith->ydata)]);
    pPlot_el->ysync_data[1].pName = (char *)(pDl_pc + 1);
    pPlot_el->data_clrs[1] = COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_smith->data_clr);
}
else{
    pPlot_el->rti_plottype = 0;
    /* error message ?? */
    return;
}
if(pDl_smith->igrid){
    if(pDl_smith->agrid){
	pPlot_el->rti_plottype = 7;	/*immittance*/
    }
    else{
	pPlot_el->rti_plottype = 5;	/*impedance */
    }
}
else{
    pPlot_el->rti_plottype = 6;		/*admittance */
}

/* create a window for the xy plot with info from display list (x,y,ht,wd,clr,bclr)*/
/* get plot coords (x,y,h,w ) from display list*/
pPlot_el->window= XCreateSimpleWindow(pdlc->d, pdlc->w,
    pRec->x,pRec->y, pRec->width,pRec->height,
    1,	/*border width */
    COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_smith->plotcom.clr),
    COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index ,pDl_smith->plotcom.bclr));

    /* Select exposure events and button press events. */

  XSelectInput(pdlc->d,pPlot_el->window,ExposureMask|ButtonPressMask|
	       ButtonReleaseMask);


/* Tell the server to make the window visible. */
XMapWindow(pdlc->d,pPlot_el->window);

/* set up for handling events on window id  */
pPlot_el->evwin.plot_win = TRUE;
pPlot_el->evwin.puser = pPlot_el;
XSaveContext(pdlc->d,pPlot_el->window,
		pdm_llist_el->win_context, (XPointer)&(pPlot_el->evwin));
	
/* set up channel access for synchronous reads */
pPlot_el->num_traces = 2 ;
pPlot_el->pSync_ctrl = gsd_sync_init(pPlot_el->ysync_data,
	pPlot_el->num_traces, ppfdctx, pdm_llist_el->ppMacro_pairs);
gsd_sync_read(NEXTSET_SYNC_DATA, pPlot_el->pSync_ctrl,
	 pPlot_el->num_traces, ppfdctx, .1);

num_data_store = 200;

/* init the smith plot */
pPlot_el->pPlot = (caddr_t)pfoWinOpen(pdlc->d,
	pPlot_el->window,
	COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_smith->plotcom.clr),
	COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_smith->plotcom.bclr), 
	num_data_store,
	pdm_llist_el->pdl_strings[(int)(pDl_smith->plotcom.title)],
	pdm_llist_el->pdl_strings[(int)(pDl_smith->plotcom.xlabel)],
	pdm_llist_el->pdl_strings[(int)(pDl_smith->plotcom.ylabel)]);
color = COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_smith->igrid_clr);
pfoWinSetAttr(((PFO_WIN *)pPlot_el->pPlot),PFO_ATTR_FG1, 0, &color);
color = COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_smith->agrid_clr);
pfoWinSetAttr(((PFO_WIN *)pPlot_el->pPlot),PFO_ATTR_FG2, 0, &color);
	
return;
}/* end setup_smith */


