/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
 /*
 *
 *	R E D R A W _ D L . C
 *
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	redraw from the display list under edit
 *	
 *
 *
 *
 *
 * 	History
 *
 *
 * 	Improvements: 
 *	01 	avoid the XClearScreen() on an expose event?
 *	02
 */

#include <Xlib.h>
#include <Xutil.h>

#include <dldef.h>
#include <dlinfo.h>
#include <edd.h>

#ifdef AXP
#include <string.h>
#endif
#ifdef TOKENPTRS
#include <tokenSupport.h>
#endif

enum caption_rows {
	cap_element_type, 
	cap_pv_name, 
	cap_nrows};

enum plot_caption_rows {
	plotcap_element_type, 
	plotcap_plot_title, 
	plotcap_xlabel, 
	plotcap_ylabel, 
	plotcap_package,
	plotcap_nrows};

enum related_display_rows {
	rd_element_type, 
	rd_name, 
	rd_arguments, 
	rd_nrows};

enum execute_script_rows {
	es_element_type, 
	es_path, 
	es_arguments, 
	es_nrows};

enum kill_display_rows {
	kd_element_type, 
	kd_label, 
	kd_nrows};

static int	draw_dyn();
static int	draw_horiz_dyn();
static int	draw_vert_dyn();
static void	draw_handle();

void
display_selection_status(pdue)
due	*pdue;
{
	char		*ptext;
	XRectangle	bounds;

	if(pdue->selections.count == 1){
		struct	dl_anygo	*pgo	= &(((eue_sel *)(pdue->selections.node.next))->eue.elem.dl_anygo);

		ptext		= dl_elements_einfo[pgo->element.type];
		bounds.x	= pgo->object.x;
		bounds.y	= pgo->object.y;
		bounds.width	= pgo->object.width;
		bounds.height	= pgo->object.height;
	}
	else if(pdue->selections.count > 1){
		ptext		= "~Multiple items selected";
		selection_bounds(pdue, &bounds, TRUE);
	}
	else
		ptext		= "";

	if(*ptext)
		sprintf(pdue->pmsg,
			ELEMENT_FORMAT_STRING,
			ptext,
			bounds.x,
			bounds.y,
			bounds.width,
			bounds.height);
	else
		pdue->pmsg[0] = '\0';

	redraw_msg(pdue);
}


void redraw_dl(pdue, prec)
due		*pdue;		/* edit session	info			*/
XRectangle	*prec;		/* the area of the window to redraw	*/
{
	redraw_objects(pdue, pdue->dl.node.next, prec);
	redraw_selections(pdue, prec);
}


void redraw_objects(pdue, peue, prec)
due		*pdue;		/* edit session	info			*/
register eue	*peue;		/* begin with this element		*/
XRectangle	*prec;		/* the area of the window to redraw	*/
{
	register struct dl_element_hdr 	*phdr;
	struct dl_context		*pdlc = &pdue->dlc;
	XGCValues 			GC_values;
	XRectangle			rec;
	struct dl_basic_attribute 	attr;

	/*
	 * 	include size of select markers
	 */
	if(prec){
		rec.x = prec->x - pdue->xselect_border;
		rec.width = prec->width + (pdue->xselect_border<<1);
		rec.y = prec->y - pdue->yselect_border;
		rec.height = prec->height + (pdue->yselect_border<<1);
	}

	/*
	 * 	redraw the display list
	 * 
	 *	only clear old stuff if starting at the bottom of the dl
@@@@not required if this is an expose@@@@@
	 */
	if(peue == (eue *) pdue->dl.node.next)
		if(prec)
			XClearArea(
				pdlc->d, 
				pdlc->w, 
				rec.x, 
				rec.y,
				rec.width,
				rec.height,
				FALSE);
		else
			XClearWindow(
				pdlc->d, 
				pdlc->w);
	/*
	 * clipping  can be done in draw_elems() however a large performance
	 * improvement occurs if it is done here at the top of the element
	 * loop. draw_elems() limit checking is disabled here by passing a
	 * nill rectrangle pointer to draw_elems().
	 */
	if(prec){
  		XSetClipRectangles(
			pdue->d, 
			pdue->dyn_GC, 
			0, 
			0, 
			&rec, 
			1, 
			YXBanded);
  		XSetClipRectangles(
			pdue->d, 
			pdlc->gc, 
			0, 
			0, 
			&rec, 
			1, 
			YXBanded);
	}else{
    		XSetClipMask(pdlc->d, pdue->dyn_GC, None);
    		XSetClipMask(pdlc->d, pdlc->gc, None);
	}

	for(;peue; peue = (eue *)peue->node.next){

		enum dl_element_class 	class;
		int			edit_mode;

		/*
		 * limit check for out of bounds elements
		 */
		if(prec){
			XRectangle	bounds;

			compute_bounds(peue, &bounds);

   			if( (bounds.x+(int)bounds.width) < rec.x )
				continue;
    			if( bounds.x > (rec.x+(int)rec.width) )
				continue;
    			if( (bounds.y+(int)bounds.height) < rec.y )
				continue;
    			if( bounds.y > (rec.y+(int)rec.height) )
				continue;
		}

		phdr = &peue->elem.dl_anygo.element;
		class = dl_element_class[phdr->type];

		edit_mode = 1 | (2 * peue->sel);
		if(class == dl_class_graphic){
			/*
			 * Setup the display list context
			 * 
			 * Done this way its slower but easier to maintain
			 * and understand.
			 */
			attr.attr = peue->sta;

			attr.element.type = dl_basic_attribute;
			attr.element.nwords = 
				WDRND(sizeof(struct dl_basic_attribute));
			draw_static(	
				pdlc, 
				&attr, 
				attr.element.nwords,
				NULL, edit_mode);

			/*
			 * draw graphical elements using the same code 
			 * employed by the display interpreter
			 */
			draw_static(
				pdlc, 
				phdr, 
				phdr->nwords,
				NULL, edit_mode);
			continue;
		}
		else if(class == dl_class_control){
			/*
			 *	code was duplicated here because
			 *	control and moinitor are identical
			 *	for now but most likely will not
			 *	be forever.
			 */
			struct dl_anyco	*paco = &peue->elem.dl_anyco;
			char 		*pcaptions[cap_nrows];
			unsigned	caption_sz[cap_nrows];
			int		status;

			pcaptions[cap_element_type] = 
				dl_elements_einfo[phdr->type];
			caption_sz[cap_element_type] =
				strlen(dl_elements_einfo[phdr->type]);
	 		stringName(&pdue->ct, 
#ifndef TOKENPTRS
				paco->control.ctrl,
#else
				dmPointer64fromToken (paco->control.ctrl),
#endif
				&pcaptions[cap_pv_name], 
				&caption_sz[cap_pv_name]);
			XSetForeground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, paco->control.clr));
			XSetBackground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, paco->control.bclr));
			draw_dyn(
				pdue, 
				&peue->elem.dl_anygo.object, 
				pcaptions,
				caption_sz,
				cap_nrows);
		}
		else if(class == dl_class_monitor){
			/*
			 *	code was duplicated here because
			 *	control and moinitor are identical
			 *	for now but most likely will not
			 *	be forever.
			 */
			struct dl_anymo	*pamo = &peue->elem.dl_anymo;
			char 		*pcaptions[cap_nrows];
			unsigned	caption_sz[cap_nrows];
			int		status;

			pcaptions[cap_element_type] = 
				dl_elements_einfo[phdr->type];
			caption_sz[cap_element_type] =
				strlen(dl_elements_einfo[phdr->type]);
	 		stringName(	
				&pdue->ct, 
#ifndef TOKENPTRS
				pamo->monitor.rdbk,
#else
				dmPointer64fromToken (pamo->monitor.rdbk), 
#endif
				&pcaptions[cap_pv_name], 
				&caption_sz[cap_pv_name]);
			XSetForeground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, pamo->monitor.clr));
			XSetBackground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, pamo->monitor.bclr));
			draw_dyn(
				pdue, 
				&peue->elem.dl_anygo.object, 
				pcaptions,
				caption_sz,
				cap_nrows);
		}
		else if(class == dl_class_plot){
			struct dl_anypo	*papo = &peue->elem.dl_anypo;
			char 		*pcaptions[plotcap_nrows];
			unsigned	caption_sz[plotcap_nrows];
			int		status;

			pcaptions[plotcap_element_type] = 
				dl_elements_einfo[phdr->type];
			caption_sz[plotcap_element_type] =
				strlen(dl_elements_einfo[phdr->type]);
	 		stringName(	
				&pdue->st, 
#ifndef TOKENPTRS
				papo->plotcom.title, 
#else
				dmPointer64fromToken (papo->plotcom.title), 
#endif
				&pcaptions[plotcap_plot_title], 
				&caption_sz[plotcap_plot_title]);
	 		stringName(	
				&pdue->st, 
#ifndef TOKENPTRS
				papo->plotcom.xlabel,
#else
				dmPointer64fromToken (papo->plotcom.xlabel), 
#endif
				&pcaptions[plotcap_xlabel], 
				&caption_sz[plotcap_xlabel]);
	 		stringName(	
				&pdue->st, 
#ifndef TOKENPTRS
				papo->plotcom.ylabel,
#else
				dmPointer64fromToken (papo->plotcom.ylabel), 
#endif
				&pcaptions[plotcap_ylabel], 
				&caption_sz[plotcap_ylabel]);
	 		stringName(	
				&pdue->st, 
#ifndef TOKENPTRS
				papo->plotcom.package, 
#else
				dmPointer64fromToken (papo->plotcom.package), 
#endif
				&pcaptions[plotcap_package], 
				&caption_sz[plotcap_package]);
			XSetForeground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, papo->plotcom.clr));
			XSetBackground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, papo->plotcom.bclr));
			draw_dyn(
				pdue, 
				&peue->elem.dl_anygo.object, 
				pcaptions,
				caption_sz,
				plotcap_nrows);
		}
		else if(class == dl_class_related_display){
			struct dl_related_display	*prdo; 
			char 		*pcaptions[rd_nrows];
			unsigned	caption_sz[rd_nrows];
			int		status;

			prdo = (struct dl_related_display *)
				&peue->elem.dl_anygo;

			pcaptions[rd_element_type] = 
				dl_elements_einfo[phdr->type];
			caption_sz[rd_element_type] =
				strlen(dl_elements_einfo[phdr->type]);
	 		stringName(	
				&pdue->st, 
#ifndef TOKENPTRS
				prdo->display[0].name, 
#else
				dmPointer64fromToken (prdo->display[0].name), 
#endif
				&pcaptions[rd_name], 
				&caption_sz[rd_name]);
	 		stringName(	
				&pdue->st, 
#ifndef TOKENPTRS
				prdo->display[0].args, 
#else
				dmPointer64fromToken (prdo->display[0].args), 
#endif
				&pcaptions[rd_arguments], 
				&caption_sz[rd_arguments]);
			XSetForeground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, prdo->clr));
			XSetBackground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, prdo->bclr));
			draw_dyn(
				pdue, 
				&peue->elem.dl_anygo.object, 
				pcaptions,
				caption_sz,
				rd_nrows);
		}
		else if(class == dl_class_execute_script){
			struct dl_execute_script	*peso; 
			char 		*pcaptions[es_nrows];
			unsigned	caption_sz[es_nrows];
			int		status;

			peso = (struct dl_execute_script *)
				&peue->elem.dl_anygo;

			pcaptions[es_element_type] = 
				dl_elements_einfo[phdr->type];
			caption_sz[es_element_type] =
				strlen(dl_elements_einfo[phdr->type]);

	 		stringName(	
				&pdue->st, 
#ifndef TOKENPTRS
				peso->script[0].path, 
#else
				dmPointer64fromToken (peso->script[0].path), 
#endif
				&pcaptions[es_path], 
				&caption_sz[es_path]);
	 		stringName(	
				&pdue->st, 
#ifndef TOKENPTRS
				peso->script[0].args, 
#else
				dmPointer64fromToken (peso->script[0].args), 
#endif
				&pcaptions[es_arguments], 
				&caption_sz[es_arguments]);
			XSetForeground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, peso->clr));
			XSetBackground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, peso->bclr));
			draw_dyn(
				pdue, 
				&peue->elem.dl_anygo.object, 
				pcaptions,
				caption_sz,
				es_nrows);
		}
		else if(class == dl_class_kill_display){
			struct dl_kill_display	*pkdo; 
			char 		*pcaptions[2];
			unsigned	caption_sz[2];
			int		status;

			pkdo = (struct dl_kill_display *)
				&peue->elem.dl_anygo;

			pcaptions[kd_element_type]  = dl_elements_einfo[phdr->type];
			caption_sz[kd_element_type] = strlen(dl_elements_einfo[phdr->type]);

	 		stringName(	
				&pdue->st, 
#ifndef TOKENPTRS
				pkdo->label, 
#else
				dmPointer64fromToken (pkdo->label), 
#endif
				&pcaptions[kd_label], 
				&caption_sz[kd_label]);
			XSetForeground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, pkdo->clr));
			XSetBackground(	
				pdue->d, 
				pdue->dyn_GC, 
				CLIPCLRIX(pdue, pkdo->bclr));
			draw_dyn(
				pdue, 
				&peue->elem.dl_anygo.object, 
				pcaptions,
				caption_sz,
				kd_nrows);
		}
		else{
			assert(FALSE);
		}
	}
	display_selection_status(pdue);
}


/*
 *	DRAW_DYN
 *	draw dynamic objects with a label
 *
 */
static int draw_dyn(pdue, pobj, pcaptions, pcaplength, ncaptions)
due			*pdue;
struct dl_object_hdr 	*pobj;
char			**pcaptions;
unsigned		*pcaplength;
unsigned		ncaptions;
{
	XRectangle box;

	{
		/*
		 * Temporarily set GC foreground to its background
		 */
		unsigned long fg = pdue->dyn_GC->values.foreground;
		unsigned long bg = pdue->dyn_GC->values.background;

		XSetForeground(pdue->d, pdue->dyn_GC, bg);
        	XFillRectangle( 
			pdue->dlc.d,
              	  	pdue->dlc.w,
               	 	pdue->dyn_GC,
                	pobj->x,
                	pobj->y,
                	pobj->width, 
                	pobj->height);
		XSetForeground(pdue->d, pdue->dyn_GC, fg);
	}
        XDrawRectangle( 
		pdue->dlc.d,
                pdue->dlc.w,
                pdue->dyn_GC,
                pobj->x,
                pobj->y,
                pobj->width-1, 
                pobj->height-1);

	box.x = pobj->x + (pobj->width>>3);
	box.y = pobj->y + (pobj->height>>3);
	box.width = pobj->width - (pobj->width>>2);
	box.height = pobj->height - (pobj->height>>2);
	if(pobj->width > pobj->height)
		return draw_horiz_dyn(
			pdue, 
			&box, 
			pcaptions, 
			pcaplength, 
			ncaptions);
	else
		return draw_vert_dyn(
			pdue, 
			&box, 
			pcaptions, 
			pcaplength, 
			ncaptions);
}


/*
 *	DRAW_HORIZ_DYN
 *	draw dynamic objects with a label
 *
 */
static int draw_horiz_dyn(pdue, pobj, pcaptions, pcaplength, ncaptions)
due			*pdue;
struct dl_object_hdr 	*pobj;
char			**pcaptions;
unsigned		*pcaplength;
unsigned		ncaptions;
{
	int		i;
	XRectangle 	box;
	font_table	*pft = pdue->dlc.pfont_table;


	box.x = pobj->x;
	box.y = pobj->y;
	box.width = pobj->width;
	box.height = pobj->height/ncaptions;
	if(box.height < pft->min_font_height){
		ncaptions = (unsigned int)(pobj->height) / (unsigned int)(pft->min_font_height);
		box.height = pft->min_font_height;
	}

	box.y += (unsigned int)(pobj->height) % (unsigned int)(box.height)>>1;

	for(i=0; i<ncaptions; i++){
		draw_horizontal_string(
				pdue->dlc.d,
				pdue->dlc.w,
				pdue->dyn_GC,
				&box,
				pdue->dlc.pfont_table,
				pcaptions[i],
				pcaplength[i],
				box_begin);
		box.y += box.height;
	}
}


/*
 *	DRAW_VERT_DYN
 *	draw dynamic objects with a label
 *
 */
static int draw_vert_dyn(pdue, pobj, pcaptions, pcaplength, ncaptions)
due			*pdue;
struct dl_object_hdr 	*pobj;
char			**pcaptions;
unsigned		*pcaplength;
unsigned		ncaptions;
{
	int		i;
	XRectangle 	box;
	font_table	*pft = pdue->dlc.pfont_table;


	box.x = pobj->x;
	box.y = pobj->y;
	box.height = pobj->height;
	box.width = pobj->width/ncaptions;
	if(box.width < pft->min_font_width){
		ncaptions = (unsigned int)(pobj->width) / (unsigned int)(pft->min_font_width);
		box.width = pft->min_font_width;
	}

	box.x += (unsigned int)(pobj->width) % (unsigned int)(box.width) >> 1;

	for(i=0; i<ncaptions; i++){
		draw_vertical_string(
				pdue->dlc.d,
				pdue->dlc.w,
				pdue->dyn_GC,
				&box,
				pdue->dlc.pfont_table,
				pcaptions[i],
				pcaplength[i],
				box_begin);
		box.x += box.width;
	}
}



/*
 *
 *	ADD_MARKERS_TO_BOUNDS
 *	(add selection marker area to bounding rectangle)
 */
void add_markers_to_bounds(pdue, pbounds)
due		*pdue;		/* edit session	info	*/
XRectangle	*pbounds;
{
	int		xborder, yborder;

	xborder = pdue->xselect_border;
	yborder = pdue->yselect_border;

	pbounds->x -= xborder;
	pbounds->width += (xborder<<1);
	pbounds->y -= yborder;
	pbounds->height += (yborder<<1);

}


/*
 *
 * REDRAW_SELECTIONS
 *
 */
void redraw_selections(pdue, prec)
due		*pdue;
XRectangle	*prec;		/* the area of the window to redraw	*/
{
	eue_sel				*psel;
	register struct dl_object_hdr	*objptr;
	int				xborder = pdue->xselect_border;
	int				yborder = pdue->yselect_border;
	XRectangle			rec;

	if(prec){
		rec.x = prec->x - xborder;
		rec.width = prec->width + (xborder<<1);
		rec.y = prec->y - yborder;
		rec.height = prec->height + (yborder<<1);
	}

	/*
	 * redraw the selection indicators
	 */
	for(	psel = (eue_sel *) pdue->selections.node.next;
		psel;
		psel = (eue_sel *) psel->node.next){
		XRectangle bounds;
		
		objptr = &psel->eue.elem.dl_anygo.object;

		DLOBJ_TO_XREC(psel->eue.elem.dl_anygo.object, bounds);

		/*
		 * only redraw if in the exposed area of the screen
		 */
		if(prec){
   			if( bounds.x+(int)bounds.width < rec.x)
				continue;
    			if( bounds.x > rec.x+(int)rec.width )
				continue;
    			if( bounds.y+(int)bounds.height < rec.y )
				continue;
    			if( bounds.y > rec.y+(int)rec.height )
				continue;
		}
		redraw_selection(pdue, &bounds);
		if( psel->eue.elem.dl_anygo.element.type == dl_line )
			redraw_line_endpoint_handles(pdue, &psel->eue.elem.dl_line);
	}
}


/*
 *
 * REDRAW_SELECTION
 *
 */
void redraw_selection(pdue, pbounds)
due		*pdue;
XRectangle	*pbounds;
{
	XPoint	select_icon_points[5];
	int	xborder = pdue->xselect_border;
	int	yborder = pdue->yselect_border;

	select_icon_points[0].x = -(xborder/2);	/* set up relative coordinates */
	select_icon_points[0].y = -(yborder/2);
	select_icon_points[1].x = xborder;
	select_icon_points[1].y = 0;
	select_icon_points[2].x = 0;
	select_icon_points[2].y = yborder;
	select_icon_points[3].x = -xborder;
	select_icon_points[3].y = 0;
	select_icon_points[4].x = 0;
	select_icon_points[4].y = -yborder;

	/* top left corner */
	select_icon_points[0].x += pbounds->x;
	select_icon_points[0].y += pbounds->y;
	draw_handle(pdue, &select_icon_points[0]);

	/* top mid-line */
	select_icon_points[0].x += pbounds->width/2;
	select_icon_points[0].y += 0;
	draw_handle(pdue, &select_icon_points[0]);

	/* top right corner */
	select_icon_points[0].x += pbounds->width/2;
	select_icon_points[0].y += 0;
	draw_handle(pdue, &select_icon_points[0]);

	/* right mid-line */
	select_icon_points[0].x += 0;
	select_icon_points[0].y += pbounds->height/2;
	draw_handle(pdue, &select_icon_points[0]);

	/* bottom right corner */
	select_icon_points[0].x += 0;
	select_icon_points[0].y += pbounds->height/2;
	draw_handle(pdue, &select_icon_points[0]);

	/* bottom mid-line */
	select_icon_points[0].x -= pbounds->width/2;
	select_icon_points[0].y -= 0;
	draw_handle(pdue, &select_icon_points[0]);
			
	/* bottom left corner */
	select_icon_points[0].x -= pbounds->width/2;
	select_icon_points[0].y -= 0;
	draw_handle(pdue, &select_icon_points[0]);
			
	/* left mid-line */
	select_icon_points[0].x -= 0;
	select_icon_points[0].y -= pbounds->height/2;
	draw_handle(pdue, &select_icon_points[0]);
}


/*
 *
 * REDRAW_LINE_ENDPOINT_HANDLES
 *
 */
void redraw_line_endpoint_handles(pdue, pline)
due		*pdue;
struct dl_line	*pline;
{
	XPoint		select_icon_points[5];
	int		xborder = (3*pdue->xselect_border)/4;
	int		yborder = (3*pdue->yselect_border)/4;
	unsigned int	i;

	/* set up relative coordinates - 0th element is set below in loop */
	select_icon_points[1].x = xborder;
	select_icon_points[1].y = 0;
	select_icon_points[2].x = 0;
	select_icon_points[2].y = yborder;
	select_icon_points[3].x = -xborder;
	select_icon_points[3].y = 0;
	select_icon_points[4].x = 0;
	select_icon_points[4].y = -yborder;

	for(i = 0; i < pline->npoints; i++){
		select_icon_points[0].x = pline->point[i].x - (xborder/2);
		select_icon_points[0].y = pline->point[i].y - (yborder/2);
		draw_handle(pdue, &select_icon_points[0]);
	}
}


/*
 *
 * DRAW__HANDLE
 *
 */
static void draw_handle(pdue, ppoints)
due	*pdue;
XPoint	*ppoints;
{
	XGCValues 	GC_values;
		
  	GC_values.foreground = WhitePixelOfScreen(pdue->s);
	XChangeGC(pdue->dlc.d, pdue->select_GC, GCForeground, &GC_values);
	XFillPolygon(pdue->dlc.d, pdue->dlc.w, pdue->select_GC, ppoints, 5, Convex, CoordModePrevious );

	GC_values.foreground = BlackPixelOfScreen(pdue->s);
	XChangeGC(pdue->dlc.d, pdue->select_GC, GCForeground, &GC_values);
	XDrawLines(pdue->dlc.d, pdue->dlc.w, pdue->select_GC, ppoints, 5, CoordModePrevious );
}
