/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *
 *	S E L E C T _ U T I L . C
 *
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	code segments which operate on the current selection
 *	
 *
 *
 *
 *
 * 	History
 *	.00 joh 062191	Added push_field_undo() calls for dl updates here
 *	 
 *
 */
#include <edd.h>
#include <dlinfo.h>
#include <edd_menus.h>

/*
 * given a current handle id and an inversion type, this table will
 * produce the new handle id
 */
static enum edd_handle	new_handle[9][4] =
	{
					/* none         x_coords      y_coords      xy_coords */
/* outside */		{ outside,      outside,      outside,      outside      },
/* top_left */		{ top_left,     top_right,    bottom_left,  bottom_right },
/* top */			{ top,          top,          bottom,       bottom       },
/* top_right */		{ top_right,    top_left,     bottom_right, bottom_left  },
/* right */			{ right,        left,         right,        left         },
/* bottom_right */	{ bottom_right, bottom_left,  top_right,    top_left     },
/* bottom */		{ bottom,       bottom,       top,          top          },
/* bottom_left */	{ bottom_left,  bottom_right, top_left,     top_right    },
/* left */			{ left,         right,        left,         right        },
	};



/*
 *	SELECTION_BOUNDS()
 * 	find a rectangle bounding the current selection
 *	(adjusts for line thickness)
 */
void selection_bounds(pdue, pbounds, init)
due					*pdue;
register XRectangle	*pbounds;
int					init;		/* init *pbounds if TRUE */
	{
	register eue_sel	*psel;
	int					firstloop = init;
	XRectangle			rec;

	for (psel = (eue_sel *) pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		{
		compute_bounds(&psel->eue, &rec);
		if (firstloop)
			{
			*pbounds = rec;
			firstloop = FALSE;
			}
		else
			ADD_TO_MAXBOUNDS((*pbounds), rec);
		}
	}


/*
 *	SELECTION_COORDINATE_BOUNDS()
 * 	find a rectangle bounding the current selection
 *	(dont consider line thickness)
 */
void selection_coordinate_bounds(pdue, pbounds, init)
due					*pdue;
register XRectangle	*pbounds;
int					init;		/* init *pbounds if TRUE */
	{
	register eue_sel	*psel;
	int					firstloop = init;
	XRectangle			rec;

	for (psel = (eue_sel *) pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		{
		DLOBJ_TO_XREC(psel->eue.elem.dl_anygo.object, rec);
		if (firstloop)
			{
			*pbounds = rec;
			firstloop = FALSE;
			}
		else
			ADD_TO_MAXBOUNDS((*pbounds), rec);
		}
	}


enum inversion
	{
	none,
	x_coords,
	y_coords,
	xy_coords
	};

/*
 * given an old inversion type and the current quadrant, this tabe gives the new conversion type
 */
enum inversion	inversion[4][4] =
	{
				/*       none     x_coords     y_coords    xy_coords */
/* none */		{        none,    x_coords,    y_coords,   xy_coords },
/* x_coords */	{    x_coords,        none,    y_coords,    y_coords },
/* y_coords */	{    y_coords,    x_coords,        none,    x_coords },
/* xy_corrds */	{   xy_coords,    y_coords,    x_coords,        none },
	};

/*
 * INVERSION_TYPE
 *
 * determine an inversion type based on the relative values of the
 * (x1,y1) (x2,y2) pairs if (x1,y1) is taken as the origin
 */
enum inversion	inversion_type( x1, y1, x2, y2 )
	{
	int	dx	= x2 - x1;
	int	dy	= y2 - y1;

	if      (dx >= 0 && dy >= 0)	return	none;		/* quad I   */
	else if (dx >= 0 && dy <  0)	return	y_coords;	/* quad II  */
	else if	(dx <  0 && dy >= 0)	return	x_coords;	/* quad IV  */
	else /* (dx <  0 && dy <  0)*/	return	xy_coords;	/* quad III */
	}


/*
 * APPLY_INVERSION_TO_ENDPOINTS
 */
apply_inversion_to_endpoints( pline, inversion_type )
struct dl_line	*pline;
enum inversion	inversion_type;
	{
	unsigned int	i;

	if (inversion_type == none)
		return;

	for (i = 0; i < pline->npoints; i++)
		{
		if (inversion_type == x_coords || inversion_type == xy_coords)
			pline->orgpoint[i].x = 1000 - pline->orgpoint[i].x;
		if (inversion_type == y_coords || inversion_type == xy_coords)
			pline->orgpoint[i].y = 1000 - pline->orgpoint[i].y;
		}
	}

/*
 *	MOVE_SELECTIONS()
 * 	initiate move of items to operator selected position
 */
void move_selections(dummy, pdue)
int			dummy;
due			*pdue;
	{
	static void			update_selections_moved();
	XRectangle			maxrec;
	register eue_sel	*psel;

	/*
	 * find the bounds of the selection
	 * push the current values
	 */
	selection_bounds(pdue, &maxrec, TRUE);
	for (psel = (eue_sel *) pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		{
		struct dl_element_hdr	*pelem	= &psel->eue.elem.dl_anygo.element;
		struct dl_object_hdr	*pobj	= &psel->eue.elem.dl_anygo.object;

		push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[x_index_dl_object_hdr], pobj, 0);
		push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[y_index_dl_object_hdr], pobj, 0);
		if ( pelem->type == dl_line )
			push_line_endpoints(pdue, psel);
		}

	/*
	 * get the new coordinate
	 * reset the selection bounds
	 * and redraw
	 */
	get_coord(pdue->d, pdue->dlc.w, pdue->pick_cursor, maxrec.x, maxrec.y, update_selections_moved, pdue);
	selection_bounds(pdue, &maxrec, TRUE);
	redraw_dl(pdue, &maxrec);
	}


/*
 *	UPDATE_SELECTIONS_MOVED()
 * 	move items to operator selected position
 */
static void update_selections_moved(pdue, x, y)
due	*pdue;
int	x;
int	y;
	{
	XRectangle			maxrec;
	register eue_sel	*psel;

	x = max(0, x);	/* clamp positive */
	y = max(0, y);
	
	/*
	 * find the bounds of the selection
	 */
	selection_bounds(pdue, &maxrec, TRUE);

	/*
	 * offset to the user provided position
	 * add changes to the bounds
	 * redraw only items which changed
	 */
	for (psel = (eue_sel *) pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		{
		struct dl_element_hdr	*pelem	= &psel->eue.elem.dl_anygo.element;
		struct dl_object_hdr	*pobj	= &psel->eue.elem.dl_anygo.object;

		pobj->x += (x - maxrec.x);
		pobj->y += (y - maxrec.y);
		if (pelem->type == dl_line)
			adjust_line_endpoints(&(psel->eue.elem.dl_line), x - maxrec.x, y - maxrec.y);
		}
	selection_bounds(pdue, &maxrec, FALSE);
	redraw_objects(pdue, pdue->dl.node.next, &maxrec);
	}



/*
 *	RESIZE_SELECTIONS()
 * 	resize operator selected objects
 */
struct resize_sel_args
	{
	due				*pdue;
	XRectangle		coord_bounds;
	enum edd_handle	handle;
	enum inversion	last_inversion;
	};

void resize_selections(dummy, pdue)
int			dummy;
due			*pdue;
	{
	static void				update_selections_resized();
	struct resize_sel_args	args;
	eue_sel					*psel;

	/*
	 * find the bounds of the selection
	 * push the current values
	 */
	selection_coordinate_bounds(pdue, &args.coord_bounds, TRUE);
	for (psel = (eue_sel *) pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		{
		struct dl_element_hdr	*pelem	= &psel->eue.elem.dl_anygo.element;
		struct dl_object_hdr	*pobj	= &psel->eue.elem.dl_anygo.object;

		push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[x_index_dl_object_hdr],      pobj, 0);
		push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[y_index_dl_object_hdr],      pobj, 0);
		push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[width_index_dl_object_hdr],  pobj, 0);
		push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[height_index_dl_object_hdr], pobj, 0);
		if (pelem->type == dl_line)
			{
			push_original_line_endpoints(pdue, psel);
			push_line_endpoints(pdue, psel);
			}
		}
	pdue->undo_sequence++;	/* force intermediate changes while getting new coord into another undo sequence */

	args.pdue			= pdue;
	args.handle			= bottom_right;
	args.last_inversion	= none;
	get_coord(pdue->d, 
		pdue->dlc.w, 
		pdue->pick_cursor,
 		args.coord_bounds.x + (int)args.coord_bounds.width, 
		args.coord_bounds.y + (int)args.coord_bounds.height, 
		update_selections_resized, 
		&args);

	/*
	 * redraw selection markers
	 */
	discard_current_sequence(pdue);
	selection_bounds(pdue, &args.coord_bounds, TRUE);
	redraw_dl(pdue, &args.coord_bounds);
	}


/*
 *	UPDATE_SELECTIONS_RESIZED()
 * 	resize operator selected objects
 */
static void update_selections_resized(pargs, x, y)
struct resize_sel_args	*pargs;
int						x;
int						y;
	{
	XRectangle		bounds;
	eue_sel			*psel;
	double			x_magnification;
	double			y_magnification;
	XRectangle		*pbounds	= &pargs->coord_bounds;
	enum inversion	inversion_change;
	enum inversion	new_inversion;
	int				x1, y1, x2, y2;
	int				do_inversion;

	x = max(0, x);	/* clamp positive */
	y = max(0, y);
	
	/*
	 * find the bounds of the selection
	 */
	selection_bounds(pargs->pdue, &bounds, TRUE);

	x1 = x2 = pbounds->x;
	y1 = y2 = pbounds->y;
	switch(pargs->handle)
	  {
	  case top_left:		x1 = x;	y1 = y;					break;
	  case top_right:				y1 = y;	x2 = x;			break;
	  case bottom_right:					x2 = x; y2 = y;	break;
	  case bottom_left:		x1 = x; 				y2 = y;	break;
	  }
	new_inversion = inversion_type(x1, y1, x2, y2);
	inversion_change = inversion[pargs->last_inversion][new_inversion];
	if (do_inversion = (inversion_change != pargs->last_inversion))
		{
		pargs->last_inversion = inversion_change;
		pargs->handle = new_handle[pargs->handle][new_inversion];
		}

	/*
	 * force at least 1 pixel width and height bounds, otherwise they cannot be magnified
	 */
	pbounds->width  = (pbounds->width  == 0) ? 1 : pbounds->width;
	pbounds->height = (pbounds->height == 0) ? 1 : pbounds->height;
	x_magnification	= (double)(x - pbounds->x)/(double)(pbounds->width);
	y_magnification	= (double)(y - pbounds->y)/(double)(pbounds->height);

	/*
	 * undo the last intermediate values (pushed below but during the previous call) to get the originals
	 */
	revert_current_sequence(pargs->pdue);
	for (psel = (eue_sel *) pargs->pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		{
		struct dl_element_hdr	*pelem	= &psel->eue.elem.dl_anygo.element;
		struct dl_object_hdr	*pobj	= &psel->eue.elem.dl_anygo.object;

		/*
		 * save these values for the next revert (see above) during the next intermediate value
		 */
		push_field_undo(pargs->pdue, &psel->eue, &dl_object_hdr_sinfo[x_index_dl_object_hdr],      pobj, 0);
		push_field_undo(pargs->pdue, &psel->eue, &dl_object_hdr_sinfo[y_index_dl_object_hdr],      pobj, 0);
		push_field_undo(pargs->pdue, &psel->eue, &dl_object_hdr_sinfo[width_index_dl_object_hdr],  pobj, 0);
		push_field_undo(pargs->pdue, &psel->eue, &dl_object_hdr_sinfo[height_index_dl_object_hdr], pobj, 0);

		/*
		 * calculate the new values based on the x and y magnifications from the originals
		 */
		x1 = ( (pobj->x - pbounds->x) * x_magnification ) + pbounds->x;
		y1 = ( (pobj->y - pbounds->y) * y_magnification ) + pbounds->y;
		x2 = ( ((pobj->x + pobj->width)  - pbounds->x) * x_magnification ) + pbounds->x;
		y2 = ( ((pobj->y + pobj->height) - pbounds->y) * y_magnification ) + pbounds->y;
		pobj->x      = min(x1,x2);
		pobj->y      = min(y1,y2);
		pobj->width  = posdiff(x1,x2);
		pobj->height = posdiff(y1,y2);
		if (pelem->type == dl_line)
			{
			if (do_inversion)
				apply_inversion_to_endpoints(&(psel->eue.elem.dl_line), new_inversion);
			adjust_line_endpoints(&(psel->eue.elem.dl_line), 0, 0);
			}
		}
	selection_bounds(pargs->pdue, &bounds, FALSE);
	redraw_objects(pargs->pdue, pargs->pdue->dl.node.next, &bounds);
	}


/*
 *	SINK_SELECTIONS()
 * 	move selections to the bottom
 */
void sink_selections(dummy, pdue)
int	dummy;
due	*pdue;
	{
	XRectangle	maxrec;
	eue			*peue;
	eue			*pinsert;
	eue			*pnext;

	pdue->dl_mdt = TRUE;

	/*
	 * find the bounds of the selection
	 */
	selection_bounds(pdue, &maxrec, TRUE);

	pinsert = NULL; /* insert at list head */
	for (peue = (eue *) pdue->dl.node.next; peue; peue = pnext)
		{
		pnext = (eue *) peue->node.next;
		if (peue->sel)
			{
			push_front_back_undo(pdue, peue, peue->node.previous);
			lstDelete(&pdue->dl, peue);
			lstInsert(&pdue->dl, pinsert, peue);
			pinsert = peue;
			}
		}

	/*
	 * redraw items which changed
	 */
	redraw_dl(pdue, &maxrec);
	}


/*
 *	FLOAT_SELECTIONS()
 * 	move selections to the top
 */
void float_selections(dummy, pdue)
int	dummy;
due	*pdue;
	{
	XRectangle	maxrec;
	eue			*peue;
	eue			*pnext;
	eue			*plast;

	/*
	 * find the bounds of the selection
	 */
	selection_bounds(pdue, &maxrec, TRUE);

	plast = (eue *) pdue->dl.node.previous;
	for (peue = (eue *) pdue->dl.node.next; peue != plast; peue = pnext)
		{
		pnext = (eue *) peue->node.next;
		if (peue->sel)
			{
			push_front_back_undo(pdue, peue, peue->node.previous);
			lstDelete(&pdue->dl, peue);
			lstAdd(&pdue->dl, peue);
			}
		}

	/*
	 * redraw items which changed
	 */
	redraw_dl(pdue, &maxrec);
	}

/*
 *
 * IN_A_SELECTED_OBJECT()
 * returns TRUE if the given x,y in interior to any objects bounds, otherwise returns FALSE
 */
in_a_selected_object(pdue, x, y)
due	*pdue;
int	x;
int	y;
	{
	XRectangle	bounds;
	eue_sel		*psel;

	for (psel = (eue_sel *) pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		{
		compute_bounds(&psel->eue, &bounds);
		if (IN_RECTANGLE(x, y, &bounds))
			return TRUE;
		}
	return FALSE;
	}

/*
 * IN_HANDLE()
 * returns the location of the handle that contains the given x,y
 */
enum edd_handle in_handle(pdue, x, y, pbounds)
due			*pdue;
int			x;
int			y;
XRectangle	*pbounds;
	{
	static struct handle_path
		{
		int		x;
		int		y;
		enum edd_handle	handle;
		}
	handle_path[] =
		{
		 0,  0, top_left,	/* start at top left and check clockwise around the object */
		 1,  0, top,
		 1,  0, top_right,
		 0,  1, right,
		 0,  1, bottom_right,
		-1,  0, bottom,
		-1,  0, bottom_left,
		 0, -1, left,
		 0,  0, outside,
		};
	struct handle_path	*ppath;
	XRectangle			rec;

	rec.x		= pbounds->x - (pdue->xselect_border/2);
	rec.y		= pbounds->y - (pdue->yselect_border/2);
	rec.width	= pdue->xselect_border;
	rec.height	= pdue->yselect_border;

	for (ppath = &handle_path[0]; ppath->handle != outside; ppath++)
		{
		rec.x += (ppath->x * (pbounds->width/2));
		rec.y += (ppath->y * (pbounds->height/2));
		if (IN_RECTANGLE(x, y, &rec))
			break;
		}
	return ppath->handle;
	}


/*
 * IN_ENDPOINT_HANDLE
 * returns the index of the endpoint that contains the given x,y in its handle
 */
in_endpoint_handle(pdue, pline, x, y)
due				*pdue;
struct dl_line	*pline;
int				x;
int				y;
	{
	unsigned int	i;
	XRectangle		rec;

	rec.width	= (3*pdue->xselect_border)/4;
	rec.height	= (3*pdue->yselect_border)/4;

	for ( i = 0; i < pline->npoints; i++ )
		{
		rec.x = (pline->point[i].x - (rec.width/2));
		rec.y = (pline->point[i].y - (rec.height/2));
		if (IN_RECTANGLE(x, y, &rec))
			break;
		}
	return (i != pline->npoints) ? i : -1;
	}


/*
 * ADJUST_WITH_HANDLE()
 * adjusts all selected objects if x,y was in any of their handles
 * returns TRUE if they were adjusted
 */
struct adjust_with_handle_args
	{
	due				*pdue;
	eue_sel			*psel;
	enum edd_handle	handle;
	enum inversion	last_inversion;
	int				endpoint;
	};

Bool adjust_with_handle( pdue, x, y )
due			*pdue;
int			x;
int			y;
	{
	XRectangle						bounds;
	eue_sel							*psel;
	struct adjust_with_handle_args	args;

	for (psel = (eue_sel *) pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		{
		struct dl_object_hdr	*pobj	= &psel->eue.elem.dl_anygo.object;

		if (psel->eue.elem.dl_anygo.element.type == dl_line &&
		   (args.endpoint = in_endpoint_handle(pdue, &(psel->eue.elem.dl_line), x, y)) != -1)
			{
			static void 	update_endpoint_with_handle();

			args.pdue	= pdue;
			args.psel	= psel;
			push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[x_index_dl_object_hdr],      pobj, 0);
			push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[y_index_dl_object_hdr],      pobj, 0);
			push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[width_index_dl_object_hdr],  pobj, 0);
			push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[height_index_dl_object_hdr], pobj, 0);
			push_original_line_endpoints(pdue, psel);
			push_line_endpoints(pdue, psel);
			get_coord(pdue->d, pdue->dlc.w, pdue->pick_cursor, x, y, update_endpoint_with_handle, &args);

			/*
			 * redraw selection markers
			 */
			selection_bounds(pdue, &bounds, TRUE);
			redraw_dl(pdue, &bounds);
			return TRUE;
			}
		DLOBJ_TO_XREC(psel->eue.elem.dl_anygo.object, bounds);
		if ((args.handle = in_handle(pdue, x, y, &bounds)) != outside)
			{
			static void 	update_with_handle();

			args.pdue		= pdue;
			args.psel		= psel;
			args.last_inversion	= none;
			push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[x_index_dl_object_hdr],      pobj, 0);
			push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[y_index_dl_object_hdr],      pobj, 0);
			push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[width_index_dl_object_hdr],  pobj, 0);
			push_field_undo(pdue, &psel->eue, &dl_object_hdr_sinfo[height_index_dl_object_hdr], pobj, 0);
			if (psel->eue.elem.dl_anygo.element.type == dl_line)
				{
				push_original_line_endpoints(pdue, psel);
				push_line_endpoints(pdue, psel);
				}
			get_coord(pdue->d, pdue->dlc.w, pdue->pick_cursor, x, y, update_with_handle, &args);

			/*
			 * redraw selection markers
			 */
			selection_bounds(pdue, &bounds, TRUE);
			redraw_dl(pdue, &bounds);
			return TRUE;
			}
		}
	return FALSE;
	}

/*
 *	UPDATE_WITH_HANDLE()
 * 	resize operator selected object with its handle
 *
 */
static void update_with_handle(pargs, x, y)
struct adjust_with_handle_args	*pargs;
int								x;
int								y;
	{
	XRectangle				bounds;
	enum inversion			inversion_change;
	enum inversion			new_inversion;
	int						x1, y1, x2, y2;
	eue_sel					*psel	= pargs->psel;
	struct dl_object_hdr	*pobj	= &psel->eue.elem.dl_anygo.object;
	enum dl_elements		type	= psel->eue.elem.dl_anygo.element.type;

	x = max(0, x);	/* clamp positive */
	y = max(0, y);
	
	selection_bounds(pargs->pdue, &bounds, TRUE);
	
	/*
	 * apply input x,y to the rectangle corners [(x1,y1),(x2,y2)] that the handle controls
	 */
	x1 = pobj->x;
	y1 = pobj->y;
	x2 = pobj->x + pobj->width;
	y2 = pobj->y + pobj->height;
	switch(pargs->handle)
	{
	case top_left:		x1 = x;	y1 = y;					break;
	case top:					y1 = y;					break;
	case top_right:				y1 = y;	x2 = x;			break;
	case right:							x2 = x;			break;
	case bottom_right:					x2 = x; y2 = y;	break;
	case bottom:								y2 = y;	break;
	case bottom_left:	x1 = x; 				y2 = y;	break;
	case left:			x1 = x;							break;
	}

	/*
	 * update the object
	 * apply a possible inversion
	 * fixup the line endpoints if this is a line
	 * add changes to the bounds
	 * redraw items which changed
	 */
	pobj->x      = min(x1,x2);
	pobj->y      = min(y1,y2);
	pobj->width  = posdiff(x1,x2);
	pobj->height = posdiff(y1,y2);

	new_inversion = inversion_type(x1, y1, x2, y2);
	inversion_change = inversion[pargs->last_inversion][new_inversion];
	if (inversion_change != pargs->last_inversion)
		{
		pargs->last_inversion = inversion_change;
		if (type == dl_line)
			apply_inversion_to_endpoints(&(psel->eue.elem.dl_line), new_inversion);
		pargs->handle = new_handle[pargs->handle][new_inversion];
		}
	if (type == dl_line)
		adjust_line_endpoints(&(psel->eue.elem.dl_line), 0, 0);

	selection_bounds(pargs->pdue, &bounds, FALSE);
	redraw_objects(pargs->pdue, pargs->pdue->dl.node.next, &bounds);
	}

/*
 *	UPDATE_ENDPOINT_WITH_HANDLE()
 * 	move a selected lines endpoint with its handle
 */
static void update_endpoint_with_handle(pargs, x, y)
struct adjust_with_handle_args	*pargs;
int				x;
int				y;
	{
	XRectangle				bounds;
	int						x1, y1, x2, y2;
	eue_sel					*psel	= pargs->psel;
	struct dl_line			*pline	= &psel->eue.elem.dl_line;
	struct dl_object_hdr	*pobj	= &pline->object;
	unsigned int			i;

	x = max(0, x);	/* clamp positive */
	y = max(0, y);
	
	selection_bounds(pargs->pdue, &bounds, TRUE);
	
	/*
	 * apply x,y to the lines endpoint
	 */
	pline->point[pargs->endpoint].x = x;
	pline->point[pargs->endpoint].y = y;

	/*
	 * update the object
	 * fixup the line endpoints if this is a line
	 * add changes to the bounds
	 * redraw items which changed
	 */
	x1 = x2 = pline->point[0].x;
	y1 = y2 = pline->point[0].y;
	for (i = 1; i < pline->npoints; i++)
		{
		x1 = min(x1, pline->point[i].x);
		y1 = min(y1, pline->point[i].y);
		x2 = max(x2, pline->point[i].x);
		y2 = max(y2, pline->point[i].y);
		}
	pobj->x = min(x1,x2);
	pobj->y = min(y1,y2);
	pobj->width  = posdiff(x1,x2);
	pobj->height = posdiff(y1,y2);
	adjust_original_line_endpoints(pline);

	selection_bounds(pargs->pdue, &bounds, FALSE);
	redraw_objects(pargs->pdue, pargs->pdue->dl.node.next, &bounds);
	}


/*
 *	ADJUST_LINE_ENDPOINTS
 *
 * translate and scale a line object
 */
void adjust_line_endpoints(pline, deltax, deltay)
struct dl_line	*pline;
int				deltax;
int				deltay;
	{
	double			x_magnification;
	double			y_magnification;
	struct dl_point	*ppoint;
	struct dl_point	*porgpoint;
	unsigned int	i;
			
	/*
	 * translate (apply an offect) each line segments endpoint
	 */
	for (i = 0, ppoint = &(pline->point[0]); i < pline->npoints; i++, ppoint++)
		{
		ppoint->x += deltax;
		ppoint->y += deltay;
		}

	/*
	 * scale (magnify) each line segments endpoint
	 */
	x_magnification	= (double)(pline->object.width)  / (double)(1000.0);
	y_magnification	= (double)(pline->object.height) / (double)(1000.0);
	porgpoint = &(pline->orgpoint[0]);

	for (i = 0, ppoint = &(pline->point[0]); i < pline->npoints; i++, ppoint++, porgpoint++)
		{
		ppoint->x = (porgpoint->x * x_magnification) + pline->object.x;
		ppoint->y = (porgpoint->y * y_magnification) + pline->object.y;
		}
	}


/*
 *	ADJUST_ORIGINAL_LINE_ENDPOINTS
 *
 * inverse scale a line object
 */
void adjust_original_line_endpoints(pline)
struct dl_line	*pline;
	{
	double			x_magnification;
	double			y_magnification;
	struct dl_point	*ppoint;
	struct dl_point	*porgpoint;
	unsigned int	i;
			
	/*
	 * scale (magnify) each line segments endpoint
	 */
	if (pline->object.width)
		x_magnification	= (double)(1000.0) / (double)(pline->object.width);
	else
		x_magnification	= (double)(1000000.0); /* close to infinity in this coordinate system */
	if (pline->object.height)
		y_magnification	= (double)(1000.0) / (double)(pline->object.height);
	else
		y_magnification	= (double)(1000000.0); /* close to infinity in this coordinate system */
	porgpoint = &(pline->orgpoint[0]);

	for (i = 0, ppoint = &(pline->point[0]); i < pline->npoints; i++, ppoint++, porgpoint++)
		{
		porgpoint->x = (ppoint->x - pline->object.x) * x_magnification;
		porgpoint->y = (ppoint->y - pline->object.y) * y_magnification;
		}
	}


/*
 *	PUSH_LINE_ENDPOINTS
 *
 * pushes the line endpoints of a line object
 */
void push_line_endpoints(pdue, psel)
due		*pdue;
eue_sel	*psel;
	{
	struct dl_line	*pline	= &psel->eue.elem.dl_line;
	unsigned int	i;
			
	for (i = 0; i < pline->npoints; i++)
		{
		push_field_undo(pdue, &psel->eue, &dl_point_sinfo[x_index_dl_point], &pline->point[i], 0);
		push_field_undo(pdue, &psel->eue, &dl_point_sinfo[y_index_dl_point], &pline->point[i], 0);
		}
	}


/*
 *	PUSH_ORIGINAL_LINE_ENDPOINTS
 *
 * pushes the original line endpoints of a line object
 */
void push_original_line_endpoints(pdue, psel)
due		*pdue;
eue_sel	*psel;
	{
	struct dl_line	*pline	= &psel->eue.elem.dl_line;
	unsigned int	i;
			
	for (i = 0; i < pline->npoints; i++)
		{
		push_field_undo(pdue, &psel->eue, &dl_point_sinfo[x_index_dl_point], &pline->orgpoint[i], 0);
		push_field_undo(pdue, &psel->eue, &dl_point_sinfo[y_index_dl_point], &pline->orgpoint[i], 0);
		}
	}

/*
 *	GROUP_SELECTIONS()
 */
void group_selections(dummy, pdue)
int	dummy;
due	*pdue;
	{
	eue_sel	*psel;

	/*
	 * assign the next group id to each of the currently selected elements
	 */
	for (psel = (eue_sel *) pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		psel->eue.elem.dl_anygo.groupid = pdue->next_groupid;

	pdue->next_groupid++;
	}

/*
 *	UNGROUP_SELECTIONS()
 */
void ungroup_selections(dummy, pdue)
int	dummy;
due	*pdue;
	{
	eue_sel	*psel;

	/*
	 * clear the group assignment for each of the currently selected elements
	 */
	for (psel = (eue_sel *) pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		psel->eue.elem.dl_anygo.groupid = NO_GROUP;
	}
