#include "opiStringDefs.h"
extern	char	*opiStringTab[];
/*t_distribute_selections.c,v 1.1.1.1 1999/08/03 14:43:54 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 * 	D I S T R I B U T E    S E L E C T I O N S . C
 * 
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 * 
 * 	History
 *	.00 joh 062191	Added push_field_undo() calls for dl updates here
 */

#ifdef AXP
#include <stdlib.h>
#endif

void *tisMalloc();
void *tisCalloc();
void tisFree();

#include <edd.h>
#include <edd_menus.h>
#include <dlinfo.h>

static int	compare_left_edges(eue **,eue **);
static int	compare_top_edges(eue **,eue **);
static int	compare_right_edges(eue **,eue **);
static int	compare_bottom_edges(eue **,eue **);
static int	compare_horizontal_centers(eue **,eue **);
static int	compare_vertical_centers(eue **,eue **);


int (*distribute_compare[dis_last])(eue **,eue **) = {
	compare_left_edges,
	compare_right_edges,
	compare_top_edges,
	compare_bottom_edges,
	compare_horizontal_centers,
	compare_vertical_centers};


/*
 * DISTRIBUTE SELECTIONS
 * (see edd_menus.h)
 */
void distribute_selections(distribute_type, pdue)
enum distribute_type	distribute_type;
due			*pdue;
{
	XRectangle	maxrec;
	eue_sel		*psel;
	eue 		**psort;
	unsigned	i;
	int		begin;
	int		end;
	eue		null_eue;
	eue 		*peue;

	/*
	 * need at least three to distribute
	 */
	if(pdue->selections.count <= 2)
		return;

	if(pdue->selections.count > (unsigned int)(pdue->nselsort)){
		int	size;

		if(pdue->ppselsort){
			/*status = free(pdue->ppselsort);*/
			tisFree(pdue->ppselsort);
		}
		pdue->nselsort = pdue->selections.count;	
		size = sizeof(*pdue->ppselsort)*pdue->nselsort;
		/*pdue->ppselsort = (eue **) calloc(1, size);*/
		pdue->ppselsort = (eue **) tisCalloc(__FILE__, __LINE__, 1, size);
		assert(pdue->ppselsort);
	}

	selection_bounds(pdue, &maxrec, TRUE);

	for(	psel = (eue_sel *) pdue->selections.node.next, psort = pdue->ppselsort;
		psel;
		psel = (eue_sel *) psel->node.next, psort++){


		/*
		 * save old values for the undo facility
		 */
		push_field_undo(	
			pdue, 
			&psel->eue, 
			&dl_object_hdr_sinfo[x_index_dl_object_hdr], 
			&psel->eue.elem.dl_anygo.object, 
			0);

		push_field_undo(	
			pdue, 
			&psel->eue, 
			&dl_object_hdr_sinfo[y_index_dl_object_hdr], 
			&psel->eue.elem.dl_anygo.object, 
			0);
		if ( psel->eue.elem.dl_anygo.element.type == dl_line )
			push_line_endpoints(pdue, psel);

		*psort = (eue *) &psel->eue;
	}

#ifdef DEBUG
	for(i=0; i<pdue->selections.count; i++)
		printf("%d: %x\n", i, pdue->ppselsort[i]);
#endif

	qsort(	pdue->ppselsort, 
		pdue->selections.count, 
		sizeof(*pdue->ppselsort), 
		distribute_compare[distribute_type]);
#ifdef DEBUG
	for(i=0; i<pdue->selections.count; i++)
		printf("%d: %x\n", i, pdue->ppselsort[i]);
#endif

#if defined(SOLARIS) || defined(WIN32)
	memset(&null_eue, 0, sizeof(null_eue));
#else
	bzero((char *)&null_eue, sizeof(null_eue));
#endif
	peue = &null_eue;
	begin = (*distribute_compare[distribute_type])(&pdue->ppselsort[0], &peue);
	end   = (*distribute_compare[distribute_type])(&pdue->ppselsort[pdue->selections.count-1], &peue);

#ifdef DEBUG
	printf(opiStringTab[STR_begin_Hd_end_Hdn            ],begin, end);
#endif
	for(	i=1;
		i<pdue->selections.count-1;
		i++){

		struct dl_object_hdr	*pobj;
		int			position;
		XRectangle		bounds;
		int 			offsetx, orgx;
		int 			offsety, orgy;

		peue = pdue->ppselsort[i];
		compute_bounds(peue, &bounds);

		pobj = &(peue->elem.dl_anygo.object);
		offsetx = (orgx = pobj->x) - bounds.x;
		offsety = (orgy = pobj->y) - bounds.y;

		/*
		 * performing the entire calculation here to avoid round off error
		 */
		position = begin + (i * (end-begin))/(pdue->selections.count-1);

		/*
		 * also tests for invalid type and abort if so
		 */
		switch(distribute_type){
		case dis_left_edges:		pobj->x = position + offsetx;                           break;
		case dis_top_edges:		pobj->y = position + offsety;                           break;
		case dis_right_edges:		pobj->x = position + offsetx - (int)bounds.width;       break;
		case dis_bottom_edges:		pobj->y = position + offsety - (int)bounds.height;      break;
		case dis_horizontal_centers:	pobj->x = position + offsetx - (int)(bounds.width>>1);  break;
		case dis_vertical_centers:	pobj->y = position + offsety - (int)(bounds.height>>1); break;
		default:			assert(FALSE);
		}
		if(peue->elem.dl_anygo.element.type == dl_line)
			adjust_line_endpoints(&(peue->elem.dl_line), pobj->x - orgx, pobj->y - orgy);
	}
	selection_bounds(pdue, &maxrec, FALSE);
	redraw_dl(pdue, &maxrec);
}

static compare_left_edges(eue **left, eue **right)
{
	XRectangle bl, br;

	compute_bounds(*left, &bl);
	compute_bounds(*right, &br);

	return(bl.x - br.x);
}

static compare_top_edges(eue **left, eue **right)
{
	XRectangle bl, br;

	compute_bounds(*left, &bl);
	compute_bounds(*right, &br);

	return(bl.y - br.y);
}

static compare_right_edges(eue **left, eue **right)
{
	XRectangle bl, br;

	compute_bounds(*left, &bl);
	compute_bounds(*right, &br);

	return(bl.x + (int)bl.width - br.x - (int)br.width);
}

static compare_bottom_edges(eue **left, eue **right)
{
	XRectangle bl, br;

	compute_bounds(*left, &bl);
	compute_bounds(*right, &br);

	return(bl.y + (int)bl.height - br.y - (int)br.height);
}

static compare_horizontal_centers(eue **left, eue **right)
{
	XRectangle bl, br;

	compute_bounds(*left, &bl);
	compute_bounds(*right, &br);

	return(bl.x + (int)(bl.width>>1) - br.x - (int)(br.width>>1));
}

static compare_vertical_centers(eue **left, eue **right)
{
	XRectangle bl, br;

	compute_bounds(*left, &bl);
	compute_bounds(*right, &br);

	return(bl.y + (int)(bl.height>>1) - br.y - (int)(br.height>>1));
}
