#include "opiStringDefs.h"
extern	char	*opiStringTab[];
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/* 
 *	load_adl - load ascii display list 
 *	---- Insert standard header here ---------
 *
 *	history
 *	.01 joh 070191	added extra arg to close_structure() & open_structure()
 *	.02 joh 090691	added `TK_' prefix to the tokens so they dont collide
 *			with other common C defines
 */
#include	<stdio.h>
#ifdef NeXT
#include	<types.h>
#endif
#include	<fcntl.h>
#include	<dldef.h>
#include	<dlinfo.h>
#include	<dlcvrt.h>
#include	<string_table.h>
#include	<edd.h>

#ifndef	WIN32
/*
 * this is defined in sdtio.h, but the compiler
 * is not seeing it
 */
extern FILE *fdopen (int, char *);
#endif	!WIN32

#ifdef AXP
#include <string.h>
#endif
#ifdef TOKENPTRS
#include <tokenSupport.h>
#endif

#define		DEFINE_TOKENS
#include	<tokens.h>

/* Special values for pSpec */
#define		NO_STRUCTURE	(union dl_elements_union *)0
#define		NAME_UNKNOWN	(union dl_elements_union *)1
#define		FILE_SPEC	(union dl_elements_union *)2

extern void *tisMalloc();
extern void *tisCalloc();
extern void tisFree();

#ifndef		TRUE
#define		TRUE	1
#define		FALSE	0
#endif

/* This structure retains all info needed during the loading process */
struct	adl_info
{
	struct dl_basic_attribute	bas_attr;	/* tmp for basic attributes */
	struct dl_dynamic_attribute	dyn_attr;	/* tmp for dynam attributes */

	struct dl_color_map		color_map;	/* tmp color map */
	int				color_num;	/* current color number */
	struct dl_color			*pColorTbl;	/* ptr to color tbl entry */
	struct dl_color_rule_table	color_rules;	/* tmp color map */
	int				color_rule_num;	/* current color number */
	struct dl_color_rule		*pColorRuleTbl;	/* ptr to color tbl entry */

	due				*pdue;		/* display under edit */
};
typedef struct adl_info	ADL;
typedef union dl_elements_union	ELEMU;

int			line_num; /* Input file line number */

/**************************************************************************
*  NAME: load_adl
*  FUNCTION: Load an ASCII display list.
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	pdue		due *		I/O	ptr to display under edit
*	fd		int		I	input file descriptor
*
*  RETURNS: status { TK_ERROR | TK_OK }
*
*  NOTES: Assumes *pdue has been cleared of previous display & initialized.
***************************************************************************/
load_adl(pdue, fd)
due		*pdue;
int		fd;
{
	int			status;
	ADL			adl;

	/* Open the input file */
	status = openInput(fd);
	if (status != TK_OK)
		return TK_ERROR;

#if defined(SOLARIS) || defined(WIN32)
	memset(&adl, 0, sizeof(ADL));
#else
	bzero((char *)&adl, sizeof(ADL));
#endif
	adl.pdue = pdue;

	/* Buffer output line-by-line */
#if !(defined(SOLARIS) || defined(WIN32))
	setlinebuf(stdout);
	setlinebuf(stderr);
#endif

	/* Init the header structure */
	open_structure(&dl_hdr_sinfo[0], &adl.pdue->dl_hdr, adl.pdue);

	/* Init the tmp basic attribute structure */
	adl.bas_attr.element.type = dl_basic_attribute;
	open_structure(&dl_basic_attribute_sinfo[0], &adl.bas_attr, adl.pdue);

	/* Init the tmp dynamic attribute structure */
	adl.dyn_attr.element.type = dl_dynamic_attribute;
	open_structure(&dl_dynamic_attribute_sinfo[0], &adl.dyn_attr, adl.pdue);

	/* Init the color table */
	adl.color_map.ncolors = 0;
	adl.color_map.element.type = dl_color_map;
	open_structure(&dl_color_map_sinfo[0], &adl.color_map, adl.pdue);
	adl.color_num = 0;
	adl.pColorTbl = 0;

	/* Init the color rule table */
	adl.color_rules.nrules = 0;
	adl.color_rules.element.type = dl_color_rule_table;
	open_structure(&dl_color_rule_table_sinfo[0], &adl.color_rules, adl.pdue);
	adl.color_rule_num = 0;
	adl.pColorRuleTbl = 0;

	/* Read the ascii display list structures 'til eof */
	for (;;)
	{
		status = readStruct(&adl, NO_STRUCTURE, 0);
		if (status == TK_END || status == TK_ERROR)
			break;
	}

	/* Clean up */
	closeInput(fd);

	return status;
}
/**************************************************************************
*  NAME: readStruct
*  FUNCTION: Read in a structure (element or sub-structure)
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	padl		ADL		I	see desc. of ADL
*	pStruct		ELEMU		I	ptr to current structure
*	pSinfo		sinfo		I	structure info
*
*  RETURNS: status { TK_ERROR | TK_OK }
*
*  NOTES: This routine is called recursively.
*	pStruct = NO_STRUCT means we expect an outer structure
*	pStruct = NAME_UNKNOWN means we are in an error condition (i.e. an
*	invalid structure name).
***************************************************************************/
readStruct(padl, pStruct, pSinfo)
ADL			*padl;
ELEMU			*pStruct;
struct sinfo		*pSinfo;
{
	char			name[MAX_TOKEN_LTH], subscr[MAX_TOKEN_LTH],
				 value[MAX_TOKEN_LTH], tmp[MAX_TOKEN_LTH];
	int			token, status;

	/* Loop until EOF or end of structure */
	for (;;)
	{
		strcpy(subscr, ""); /* default subscript */

		token = readObject(name);
		if (token == TK_END)
		{
			if (pStruct == 0)
				return TK_END;
			else
				return TK_ERROR;
		}
		if (token == TK_R_BRACKET) /* End of structure? */
		{
#ifdef	DEBUG
			fprintf(stderr, opiStringTab[STR_End_of_structuren           ]);
#endif	DEBUG
			if (pStruct != 0)
				return TK_OK;
			else
				return TK_ERROR;
		}
		if (token != TK_NAME && token != TK_STRING)
		{
			printError(opiStringTab[STR_expect_structure_or_variable]);
			return TK_ERROR;
		}

		token = readObject(tmp);
		if (token == TK_END)
		{
			printError(opiStringTab[STR_unexpected_EOF              ]);
			return TK_ERROR;
		}
		if (token == TK_SUBSCRIPT)
		{
			strcpy(subscr, tmp);
			/* Read next token */
			token = readObject(tmp);
			if (token == TK_END)
			{
				printError(opiStringTab[STR_unexpected_EOF              ]);
				return TK_ERROR;
			}
		}

		if (token == TK_EQUAL)
		{
			/* Pick up the value */
			token = readObject(value);
			if (token == TK_END)
			{
				printError(opiStringTab[STR_unexpected_EOF              ]);
				return TK_ERROR;
			}
			if (token != TK_NAME && token != TK_NUMBER && token != TK_STRING)
			{
				printError(opiStringTab[STR_not_a_value                 ]);
				return TK_ERROR;
			}
			/* Add the value into the structure */
			status = addValue(padl, pStruct, pSinfo, name, subscr, value);
			if (status == TK_ERROR)
				return TK_ERROR;
		}

		else if (token == TK_L_BRACKET)
		{
			/* Insert the structure */
			status = addStruct(padl, pStruct, pSinfo, name, subscr);
			if (status == TK_ERROR)
				return TK_ERROR;
		}

		else
		{
			printError(opiStringTab[STR_syntax                      ]);
			return TK_ERROR;
		}
	}
}
/**************************************************************************
*  NAME: addValue
*  FUNCTION: Add a value into a structure
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	padl		ADL		I	see desc. of ADL
*	pStruct		ELEMU *		I	ptr to current structure
*	pSinfo		sinfo		I	start of structure info
*	pName		char *		I	ptr to variable name (string)
*	pSubscr		char *		I	ptr to subscript (string)
*	pValue		char *		I	ptr to value (string)
*
*  RETURNS: status { TK_ERROR | TK_OK }
*
*  NOTES: see notes on pStruct in readStruct()
***************************************************************************/
addValue(padl, pStruct, pSinfo, pName, pSubscr, pValue)
ADL			*padl;
ELEMU			*pStruct;
struct sinfo		*pSinfo;
char			*pName, *pSubscr, *pValue;
{
	struct sinfo		*getSinfo();
	int			type, subscr, nelements;
	due			*pdue;

#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR_Add_variable_HsHsCHsn       ], pName, pSubscr, pValue);
#endif
	pdue = padl->pdue;

	if (pStruct == NO_STRUCTURE) /* Are we in a structure? */
	{
		printError(opiStringTab[STR_no_structure_for_variable__H], pName);
		return TK_OK;
	}

	else if (pStruct == NAME_UNKNOWN)
	{
		fprintf(stderr, opiStringTab[STR_Hs_not_processedn           ], pName);
		return TK_OK;
	}

	else if (pStruct == FILE_SPEC)
	{	/* Should be: name = "something" */
		if (strcmp(pName, "name") != 0)
		{
			printError(opiStringTab[STR_Hs_is_not_in_this_structure ], pName);
			return TK_OK;
		}
		strcpy(pdue->pname, pValue);
		return TK_OK;
	}

	/* Find the structure info for this variable */
	pSinfo = getSinfo(pSinfo, pName);
	if (pSinfo == 0)
	{
		printError(opiStringTab[STR_Hs_is_not_in_this_structure ], pName);
		return TK_OK;
	}

	type = pSinfo->type;
	if (type == dl_struct)
	{
		printError(opiStringTab[STR_Hs_is_a_structure_namen     ], pName);
		return TK_OK;
	}

	/* Check for subscript in range */
	nelements = PSINFO_NELEMENTS(pSinfo);
	subscr = atoi(pSubscr);
	if (subscr < 0 || subscr >= nelements)
	{
		fprintf(stderr, opiStringTab[STR_Subscript_out_of_range_HsHsC],
		 pName, pSubscr, pValue);
		return TK_OK;
	}

	/* Insert the value into the structure */
	pstring_to_dl(pStruct, pSinfo, pValue, &pdue->ct, &pdue->st, subscr);

	return TK_OK;
}
/**************************************************************************
*  NAME: addStruct
*  FUNCTION: Add a structure
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	padl		ADL		I	see desc. of ADL
*	pStruct		ELEMU *		I	ptr to current structure
*	pSinfo		sinfo		I	start of structure info
*	pName		char *		I	ptr to variable name (string)
*	pSubscr		char *		I	ptr to subscript (string)
*
*  RETURNS: status { TK_ERROR | TK_OK }
*
*  NOTES: See notes on pStruct in readStruct()
***************************************************************************/
addStruct(padl, pStruct, pSinfo, pName, pSubscr)
ADL			*padl;
ELEMU			*pStruct;
struct sinfo		*pSinfo;
char			*pName;
char			*pSubscr;
{
	int			status, subscr, nelements;
	unsigned		elem_type;
	struct sinfo		*getSinfo();
	due			*pdue;

#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR_addStruct_nameCHs_subscrCHsn], pName, pSubscr);
#endif	DEBUG
	pdue = padl->pdue;

	if (pStruct == NO_STRUCTURE) /* First level: header or element? */
	{
		if (strcmp(pName, "display") == 0)
		{	/* Read remainder of header struct */
			status = readStruct(padl, &pdue->dl_hdr, &dl_hdr_sinfo[0]);
			return status;
		}

		else if (strcmp(pName, "file") == 0)
		{	/* File specification */
			status = readStruct(padl, FILE_SPEC, 0);
			return status;
		}

		/* Else, an element structure */
		elem_type = get_element_type(pName);
		if (elem_type < 0)
		{
			printError(opiStringTab[STR_Hs_is_not_an_element_name   ], pName);
			status = readStruct(padl, NAME_UNKNOWN, 0);
			return status;
		}

		else if (elem_type == dl_color_map) /* Color map? */
		{	/* Build local color table */
			status = readStruct(padl, &padl->color_map,
			 &dl_color_map_sinfo[0]);
			if (padl->pColorTbl == 0)
			{
				closeColorTbl(padl);
				return TK_ERROR;
			}
			/* Create the new color map */
			edd_create_cmap(pdue, padl->pColorTbl, padl->color_map.ncolors, 0);
			closeColorTbl(padl);
			return TK_OK;
		}

		else if (elem_type == dl_color_rule_table) /* Color rules? */
		{	/* Build local color rules table */
			struct dl_color_rule	*pdl_crules;
			unsigned int	rule, line;

			status = readStruct(padl, &padl->color_rules, &dl_color_rule_table_sinfo[0]);
			if (padl->pColorRuleTbl == 0)
			{
				closeColorRuleTbl(padl);
				return TK_ERROR;
			}
			/* Create the new color rules */
			edd_create_crules(pdue, padl->pColorRuleTbl, padl->color_rules.nrules, 0);

			/* duplicate the strings before closing the rule table */
			pdl_crules = pdue->dlc.pcrules->pdl_crules;
			for(rule = 0; rule < padl->color_rules.nrules; rule++, pdl_crules++)
			{
#ifndef TOKENPTRS
				pdl_crules->name = stringInsert(&pdue->st, padl->pColorRuleTbl[rule].name);
#else
				pdl_crules->name = dmTokenFromPointer64 (stringInsert(&pdue->st,
						dmPointer64fromToken (padl->pColorRuleTbl[rule].name)));
#endif
				for(line = 0; line < MAX_RULE_LINES; line++)
				{
#ifndef TOKENPTRS
					pdl_crules->info[line].chan  = stringInsert(&pdue->st, padl->pColorRuleTbl[rule].info[line].chan);
					pdl_crules->info[line].value = stringInsert(&pdue->st, padl->pColorRuleTbl[rule].info[line].value);
#else
					pdl_crules->info[line].chan  = dmTokenFromPointer64 (stringInsert(&pdue->st,
						dmPointer64fromToken (padl->pColorRuleTbl[rule].info[line].chan)));
					pdl_crules->info[line].value = dmTokenFromPointer64 (stringInsert(&pdue->st,
						dmPointer64fromToken (padl->pColorRuleTbl[rule].info[line].value)));
#endif
				}
			}
			closeColorRuleTbl(padl);
			return TK_OK;
		}

		else if (elem_type == dl_basic_attribute)
		{	/* Save basic attributes */

			open_structure(&dl_basic_attribute_sinfo[0], &padl->bas_attr, pdue); /* use defaults */

			status = readStruct(padl, &padl->bas_attr,
			 &dl_basic_attribute_sinfo[0], FALSE);
			return status;
		}

		else if (elem_type == dl_dynamic_attribute)
		{	/* Save dynamic attributes */
			open_structure(&dl_dynamic_attribute_sinfo[0], &padl->dyn_attr, pdue); /* use defaults */

			status = readStruct(padl, &padl->dyn_attr,
			 &dl_dynamic_attribute_sinfo[0], FALSE);
			return status;
		}

		else
		{	/* Else, create the element */
			status = addElem(padl, pSinfo, elem_type);
			return status;
		}
	}

	else if (pStruct == NAME_UNKNOWN) /* are we in an error condition */
	{
		status = readStruct(padl, NAME_UNKNOWN, 0);
		return status;
	}

	/* Else, must be a sub-structure; search structure info table */
	pSinfo = getSinfo(pSinfo, pName);
	if (pSinfo == 0)
	{	/* Not found (Note: "dl_color" is not in the table) */
		if (strcmp(pName, "dl_color") == 0)
		{	/* Read the color table */
			status = readColorTbl(padl);
			return status;
		}
		else if (strcmp(pName, "dl_color_rule") == 0)
		{	/* Read the color rule table */
			status = readColorRuleTbl(padl);
			return status;
		}
		else
		{
			printError(opiStringTab[STR_Hs_is_not_a_valid_substructu], pName);
			status = readStruct(padl, NAME_UNKNOWN, 0);
			return status;
		}
	}

	/* It's a valid name, but not a sub-structure */
	if (pSinfo->type != dl_struct)
	{
		printError(opiStringTab[STR_Hs_is_not_a_structure_name  ], pName);
		status = readStruct(padl, NAME_UNKNOWN, 0);
		return status;
	}

	/* Check for subscript in range */
	nelements = PSINFO_NELEMENTS(pSinfo);
	subscr = atoi(pSubscr); /* subscript as an integer */
	if (subscr < 0 || subscr >= nelements)
	{
		fprintf(stderr, opiStringTab[STR_Subscript_out_of_range_HsHsn],
		 pName, pSubscr);
		return TK_OK;
	}

	/* Ptr to sub-structure is offset from current structure */
	pStruct = (union dl_elements_union *)((char *)pStruct
	 + pSinfo->offset + subscr*pSinfo->size);

	/* Read the sub-structure */
	pSinfo = pSinfo->pinfo;
	status = readStruct(padl, pStruct, pSinfo);
	return status;
}
/**************************************************************************
*  NAME: addElem
*  FUNCTION: Add an element
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	padl		ADL		I	see desc. of ADL
*	pSinfo		sinfo		I	start of structure info
*	elem_type	unsigned	I	element type (e.g. dl_menu)
*
*  RETURNS: status { TK_ERROR | TK_OK }
*
*  NOTES: The element is first constructed as a local structure, and then
*	inserted into the display list.
***************************************************************************/
addElem(padl, pSinfo, elem_type)
ADL			*padl;
struct sinfo		*pSinfo;
unsigned 		elem_type;
{
	unsigned 		size, class;
	int			status;
	eue			*peue;
	due			*pdue;
	ELEMU			dl_element; /* tmp for this element */

	pdue = padl->pdue;
	pSinfo = dl_elements_ginfo[elem_type].sinfo;
	class = dl_element_class[elem_type];
	size = WDRND(dl_elements_ginfo[elem_type].size)<<1;
	/* Note: type & size are fixed; not read from display list file */
#if defined(SOLARIS) || defined(WIN32)
	memset(&dl_element, 0, sizeof(dl_element));
#else
	bzero((char *)&dl_element, sizeof(dl_element));
#endif
	dl_element.dl_anygo.element.type = elem_type;
	dl_element.dl_anygo.element.nwords = size/2;
#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR___classCHs_sizeCHdn         ],
	 dl_element_class_einfo[class], size);
#endif	DEBUG

	/* Initialize the structure contents */
	open_structure(pSinfo, &dl_element, pdue);

	/* Read in the rest of the element structure */
	status = readStruct(padl, &dl_element, pSinfo);
	if (status == TK_ERROR)
	{
		return TK_ERROR;
	}

	/* Create the new element (i.e. add to display list) */
	if (class == dl_class_monitor ||
	    class == dl_class_plot ||
	    class == dl_class_control ||
	    class == dl_class_related_display ||
	    class == dl_class_execute_script ||
	    class == dl_class_kill_display ||
	    class == dl_class_graphic)
	{
		/* Insert into display list */
		peue = create_element(pdue, &dl_element, size);
	}

	if (class == dl_class_graphic)
	{	/* Copy saved attrubutes into a graphic class element.
		 */
		peue->sta = padl->bas_attr.attr;
		peue->dyn = padl->dyn_attr.attr;
	}

	return TK_OK;
}
/**************************************************************************
*  NAME: readColorTbl
*  FUNCTION: Read the color table
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	padl		ADL		I	see desc. of ADL
*
*  RETURNS: status { TK_ERROR | TK_OK }
*
*  NOTES: The color table is built in the ADL structure.
***************************************************************************/
readColorTbl(padl)
ADL			*padl;
{
	int			status;

#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR_readColorn                  ]);
#endif

	if (padl->pColorTbl == 0)
	{
		/* No color table exists: allocate it */
		status = initColorTbl(padl);
		if (status == TK_ERROR)
			return TK_ERROR;
	}

	/* Read in the color table entries */
	status = readStruct(padl, &padl->pColorTbl[padl->color_num],
	 &dl_color_sinfo[0]);
	padl->color_num++;

	return status;
}

/* initColorTbl - init the color table */
initColorTbl(padl)
ADL			*padl;
{
	unsigned		size;
	int			ncolors, i;
	due			*pdue = padl->pdue;

#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR___initColorTbl_ncolorsCHdn  ], padl->color_map.ncolors);
#endif	DEBUG

	ncolors = padl->color_map.ncolors;
	if (ncolors <= 0)
	{
		printError(opiStringTab[STR_ncolors_not_specified       ]);
		return TK_ERROR;
	}

	size = ncolors*sizeof(struct dl_color);
	/*padl->pColorTbl = (struct dl_color *)calloc(1, size);*/
	padl->pColorTbl = (struct dl_color *)tisCalloc(__FILE__, __LINE__, 1, size);
	padl->color_num = 0;
#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR_Allocated_color_table_ncolor], ncolors);
#endif	DEBUG

	for (i = 0; i < ncolors; i++)
	{
		open_structure(&dl_color_sinfo[0], &padl->pColorTbl[i], pdue);
	}

	return TK_OK;
}

/* closeColorTbl - close the color table */
closeColorTbl(padl)
ADL			*padl;
{
	int			ncolors;
	due			*pdue = padl->pdue;

#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR___closeColorTbln            ]);
#endif	DEBUG

	ncolors = padl->color_map.ncolors;
	if ((ncolors <= 0) || (padl->pColorTbl == 0))
		return 0;

	/*free(padl->pColorTbl);*/
	tisFree(padl->pColorTbl);
	padl->pColorTbl = 0;

	return 0;
}
/**************************************************************************
*  NAME: readColorRuleTbl
*  FUNCTION: Read the color Rule table
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	padl		ADL		I	see desc. of ADL
*
*  RETURNS: status { TK_ERROR | TK_OK }
*
*  NOTES: The color rule table is built in the ADL structure.
***************************************************************************/
readColorRuleTbl(padl)
ADL			*padl;
{
	int			status;

#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR_readColorRulen              ]);
#endif

	if (padl->pColorRuleTbl == 0)
	{
		/* No color table Rule exists: allocate it */
		status = initColorRuleTbl(padl);
		if (status == TK_ERROR)
			return TK_ERROR;
	}

	/* Read in the color table entries */
	status = readStruct(padl, &padl->pColorRuleTbl[padl->color_rule_num], &dl_color_rule_sinfo[0]);
	padl->color_rule_num++;

	return status;
}

/* initColorRuleTbl - init the color Rule table */
initColorRuleTbl(padl)
ADL			*padl;
{
	unsigned		size;
	int			nrules, i;
	due			*pdue = padl->pdue;

#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR___initColorRuleTbl_nrulesCHd], padl->color_rules.nrules);
#endif	DEBUG

	nrules = padl->color_rules.nrules;
	if (nrules <= 0)
	{
		printError(opiStringTab[STR_nrules_not_specified        ]);
		return TK_ERROR;
	}

	size = nrules*sizeof(struct dl_color_rule);
	/*padl->pColorRuleTbl = (struct dl_color_rule *)calloc(1, size);*/
	padl->pColorRuleTbl = (struct dl_color_rule *)tisCalloc(__FILE__, __LINE__, 1, size);
	padl->color_rule_num = 0;
#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR_Allocated_color_table_nrules], nrules);
#endif	DEBUG

	for (i = 0; i < nrules; i++)
	{
		open_structure(&dl_color_rule_sinfo[0], &padl->pColorRuleTbl[i], pdue);
	}
	return TK_OK;
}

/* closeColorRuleTbl - close the color rule table */
closeColorRuleTbl(padl)
ADL			*padl;
{
	int			nrules;
	due			*pdue = padl->pdue;

#ifdef	DEBUG
	fprintf(stderr, opiStringTab[STR___closeColorRuleTbln        ]);
#endif	DEBUG

	nrules = padl->color_rules.nrules;
	if ((nrules <= 0) || (padl->pColorRuleTbl == 0))
		return 0;

	/*free(padl->pColorRuleTbl);*/
	tisFree(padl->pColorRuleTbl);
	padl->pColorRuleTbl = 0;

	return 0;
}
/**************************************************************************
*  NAME: getSinfo
*  FUNCTION: Get the structure info ptr for a given name
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	pSinfo		sinfo		I	start of structure info tbl
*	name		char *		I	ptr to name (string)
*
*  RETURNS: ptr to struct info entry if found, else NULL.
*
*  NOTES: see dlinfo.h
***************************************************************************/
 struct sinfo *
getSinfo(pSinfo, name)
struct sinfo		*pSinfo; /* ptr to first of an array of sinfo blocks */
char			*name; /* structure or variable name */
{
	for (; pSinfo->name != 0; pSinfo++)
	{
		if (strcmp(name, pSinfo->name) == 0)
		{
			return pSinfo;
		}
	}

	return 0; /* not found */
}

/**************************************************************************
*  NAME: get_element_type
*  FUNCTION: Given a graphic element name, return its elemement type
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	pName		char *		I	ptr to element name (string)
*
*  RETURNS: element type (int)
*
*  NOTES: see dlinfo.h
***************************************************************************/
get_element_type(pName)
char		*pName; /* element name */
{
	int			elem_type;

	for (elem_type = 0; dl_elements_einfo[elem_type] != 0; elem_type++)
	{
		if (strcmp(pName,  dl_elements_einfo[elem_type]) == 0)
			return elem_type;
	}
	return -1; /* not found in table */
}
/**************************************************************************
*  NAME: printError
*  FUNCTION: print error message with line number
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	pstr		char *		I	ptr to format string
*	var1		char *		I	optional 1-st value
*	var2		char *		I	optional 2-nd value
*
*  NOTES: var1 & var2 correspond to optional formatted output specs in pstr.
***************************************************************************/
printError(pstr, val1, val2)
char		*pstr, *val1, *val2;
{
	extern int		line_num;

#ifdef WIN32
	printf(opiStringTab[STR_Error_                      ]);
	printf(pstr, val1, val2);
	printf(opiStringTab[STR__line_Hdn                   ], line_num);
#else
	fprintf(stderr, opiStringTab[STR_Error_                      ]);
	fprintf(stderr, pstr, val1, val2);
	fprintf(stderr, opiStringTab[STR__line_Hdn                   ], line_num);
#endif WIN32
	return 0;
}

/**************************************************************************
*  NAME: readObject
*  FUNCTION: read input stream and return an object (token & value)
*
*  CALLING SEQUENCE:
*	argument	type		I/O	description
*	--------------------------------------------------------------------
*	pValue		char *		O	ptr to char array for
*						object's value
*
*  RETURNS: token
*  NOTES: Value is passed by lex throuth the global pLEXvalue.
***************************************************************************/
readObject(pValue)
char			*pValue; /* Returned object's value */
{
	extern char		*pLEXvalue; /* Value returned by LEX */
	int			token;
	extern int		line_num;
	extern char		*tokenName[]; /* for debug only */

	token = yylex(); /* call lex lexical analyser */	
	if (pValue == 0)
		return token;

	if (pLEXvalue != 0)
		strcpy(pValue, pLEXvalue);
	else
		strcpy(pValue, "");

	return token;
}

/* openInput - open input file */
openInput(fd)
int		fd;
{
	extern FILE		*yyin; /* lex input stream */
	extern int		line_num;

	/* Cause lex to read input from the specified file */
#ifdef WIN32
	yyin = _fdopen(fd, "rt");
#else
	yyin = fdopen(fd, "r");
#endif WIN32
	if (yyin == 0)
	{
		fprintf(stderr, opiStringTab[STR_Cannot_open_input_file_fdope]);
		return TK_ERROR;
	}
	line_num = 1;

	return TK_OK;
}

/* closeInput - close input file */
closeInput(fd)
int		fd;
{
	extern FILE		*yyin; /* lex input stream */
	int			status;

	do {
		status = readObject(0);
	} while (status != TK_ERROR && status != TK_END);

	fclose(yyin);

	return 0;
}
