#include "opiStringDefs.h"
extern	char	*opiStringTab[];
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	M O D I F Y . C
 *
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	Form for modifying graphic elements
 *
 *
 *	History
 *	-------
 *
 *
 *	Improvements
 *	01	be more selective about redrawing on changes
 *
 *
 */
#include <string.h>
#include <edd.h>
#include <dlinfo.h>
#include <dlcvrt.h>
#include <gizmo.h>
#include <dbDefs.h>
#include <db_access.h>

#ifdef TOKENPTRS
#include <tokenSupport.h>
#endif

extern void *tisMalloc();
extern void *tisCalloc();
extern void tisFree();

#define EDD_MAXCHAR 250

enum column_headers
	{
	description_col,
	gizmo_col,
	last_col
	};

struct modify_gizmo_ext
	{
	struct sinfo	*psinfo;
	unsigned		offset;
	void			*pstructure;
	unsigned		array_index;
	eue				*peue;
	};

static unsigned 	count_rows();
static void 		menu_update();
static void 		modify_finish();
static void 		palette_update();
static void 		valuator_update();
static void 		textedit_update();
static unsigned 	setup_form_dimensions();
static void			insert_gizmo();
static void 		compute_region_and_redraw();
static void			channel_list_update();
static void			modify_channels_finish();

enum modify_options
	{
	option_quit
	};
	
static char  *modify_choices[] =
	{       
	0,
	0
	};



/*
 * MODIFY_SELECTED_ITEMS
 */
void modify_selected_items(param, pdue, pbe)
void			*param;
due				*pdue;
XButtonEvent	*pbe;
	{
	eue_sel		*psel;

	for (psel = (eue_sel *) pdue->selections.node.next; psel; psel = (eue_sel *) psel->node.next)
		modify_element(pdue, &psel->eue, pbe->x_root - pdue->pixel_tolerence, pbe->y_root - pdue->pixel_tolerence);
	}


/*
 * MODIFY_EDD_DEFAULTS
 */
void modify_edd_defaults(param, pdue, pbe)
void			*param;
due				*pdue;
XButtonEvent	*pbe;
	{
	modify_structure( pdue, &pdue->phdrgz, opiStringTab[STR_Display_Attributes_Modify   ], dl_hdr_sinfo, &pdue->dl_hdr, pbe->x_root, pbe->y_root);
	}


/*
 * MODIFY_ELEMENT
 */
void modify_element(pdue, peue, x_root, y_root)
due		*pdue;
eue		*peue;
int		x_root;
int		y_root;
	{
	unsigned		nrows;
	unsigned		nextrow;
	gizmo			*pformg;
	gizmo			*ppopupg;
	unsigned 		type	= peue->elem.dl_anygo.element.type;
	struct sinfo	*psinfo	= dl_elements_ginfo[type].sinfo;

	if (peue->pg)
		{
		move_gizmo(peue->pg, x_root, y_root);
		return;
		}

	modify_choices[0] = opiStringTab[STR_Quit                        ];
	modify_choices[1] = 0;

	peue->pg = ppopupg = create_popup_gizmo(
				pdue->s, 
				pdue->w,
				opiStringTab[STR_Graphic_Object_Modify       ],
				x_root, 
				y_root,
				pdue->gizmo_height,
				modify_finish,
				&peue->pg,
				modify_choices,
				0);
	set_gizmo_color(peue->pg, pdue->foreground.pixel, pdue->background.pixel);

	nrows = count_rows(psinfo);
	if (dl_element_class[type] == dl_class_graphic)
		{
		nrows += count_rows(dl_sta_attr_sinfo);

		/*
		 *	This makes the PV edit go away if no dynamic modifiers are enabled
		 */
		nrows += count_rows(dynamic_graphical_element(peue) ? dl_dyn_attr_sinfo : dl_dyn_mod_sinfo);
		}
	pformg = create_form_gizmo( pdue->s, ppopupg->w, x_root, y_root, last_col, nrows, pdue->gizmo_pad, 0);

	/*
	 * start with row 0
	 */
	nextrow = setup_form_dimensions(pformg, psinfo, pdue, 0, peue, peue, OFFSET(eue, elem)); 

	/*
	 * start at the next available row if a graphical element
	 */
	if (dl_element_class[peue->elem.dl_anygo.element.type] == dl_class_graphic)
		{
		nextrow = setup_form_dimensions(pformg, dl_sta_attr_sinfo, pdue, nextrow, peue, peue, OFFSET(eue, sta));

		/*
		 *	This makes the PV edit go away if no dynamic modifiers are enabled
		 */
		if (dynamic_graphical_element(peue))
			nextrow = setup_form_dimensions(pformg, dl_dyn_attr_sinfo, pdue, nextrow, peue, peue, OFFSET(eue, dyn));
		else
			nextrow = setup_form_dimensions(pformg, dl_dyn_mod_sinfo, pdue, nextrow, peue, peue, OFFSET(eue, dyn.mod));
		}
	resize_form_gizmo(pformg);
	popup_gizmo_add_gizmo(peue->pg, pformg);
/* redundant with popup_gizmo_add_gizmo! agi
**	XMapWindow(peue->pg->pgg->d, peue->pg->w);
*/
	}



/*
 * MODIFY_STRUCTURE
 */
void modify_structure(pdue, pppopupg, ptitle, psinfo, pstruct, x_root, y_root)
due				*pdue;
gizmo			**pppopupg;
char			*ptitle;
struct sinfo	*psinfo;
void			*pstruct;
int				x_root;
int				y_root;
	{
	unsigned	nrows;
	unsigned	nextrow;
	gizmo		*pformg;

	if (*pppopupg)
		{
		move_gizmo(*pppopupg, x_root, y_root);
		return;
		}

	modify_choices[0] = opiStringTab[STR_Quit                        ];
	modify_choices[1] = 0;

	*pppopupg = create_popup_gizmo(
				pdue->s, 
				pdue->w,
				ptitle,
				x_root, 
				y_root,
				pdue->gizmo_height,
				modify_finish,
				pppopupg,
				modify_choices,
				0);
	set_gizmo_color(*pppopupg, pdue->foreground.pixel, pdue->background.pixel);

	nrows = count_rows(psinfo);
	pformg = create_form_gizmo(pdue->s, (*pppopupg)->w, x_root, y_root, last_col, nrows, pdue->gizmo_pad, 0);

	/*
	 * start with row 0 at an offset of 0 from pstruct
	 */
	nextrow = setup_form_dimensions(pformg, psinfo, pdue, 0, NULL, pstruct, 0); 
	resize_form_gizmo(pformg);
	popup_gizmo_add_gizmo(*pppopupg, pformg);
/* redundant with popup_gizmo_add_gizmo! agi
**	XMapWindow((*pppopupg)->pgg->d, (*pppopupg)->w);
*/
	}


/*
 * COUNT_ROWS
 */
static unsigned count_rows(psinfo)
struct sinfo	*psinfo;
	{
	unsigned nrows	= 0;

	for (; psinfo->name; psinfo++)
		{
		if (psinfo->input != edd_none)
			{
			unsigned	nelements = PSINFO_NELEMENTS(psinfo);
			
			if (psinfo->input == edd_subform)
				nrows += count_rows(psinfo->pinfo) * nelements;
			else
				nrows += nelements;
			}
		}
	return nrows;
	}


/*
 * SETUP_FORM_DIMENSIONS
 */
static unsigned setup_form_dimensions(pformg, psinfo, pdue, row, peue, pstructure, offset)
gizmo 			*pformg;
struct sinfo 	*psinfo;
due				*pdue;
unsigned		row;
eue				*peue;
void			*pstructure;
unsigned		offset;
{
	unsigned	count;
	unsigned	i;
	int 		suboffset;

	for (; psinfo->name; psinfo++)
		{
		if (psinfo->input == edd_none)
			continue;
		
		suboffset = offset + psinfo->offset;
		count = PSINFO_NELEMENTS(psinfo);
		for (i = 0; i < count; i++)
			{
			struct modify_gizmo_ext modifyge;

			if (psinfo->input == edd_subform)
				{
				row = setup_form_dimensions(pformg, psinfo->pinfo, pdue, row, peue, pstructure, suboffset);
				suboffset += psinfo->size;
				continue;
				}
			form_gizmo_add_string(pformg, description_col, row, pdue->gizmo_height, psinfo->pString);

			modifyge.psinfo			= psinfo;
			modifyge.peue			= peue;
			modifyge.offset			= offset;
			modifyge.pstructure		= (void *)(((char *)pstructure) + offset);
			modifyge.array_index	= i;
			insert_gizmo(pformg, pdue, row, gizmo_col, &modifyge);
			row++;
			}
		}
	return row; /* next row added */
	}



/*
 * INSERT_GIZMO
 */
static void insert_gizmo(pformg, pdue, row, column, pmodifyge)
gizmo					*pformg;
due						*pdue;
unsigned				row;
unsigned				column;
struct modify_gizmo_ext *pmodifyge;
	{
	XRectangle	obj;
	gizmo		*pg;
	int			xtol	= mm_to_xpixel(MM_TOLERENCE, pdue->s);
	int			ytol	= mm_to_ypixel(MM_TOLERENCE, pdue->s);
	long		ival;

	/* size known, but position tbd */
	obj.x = 0;
	obj.y = 0;
	obj.height = pdue->gizmo_height;
	obj.width = pdue->gizmo_width;
	switch (pmodifyge->psinfo->input)
	  {
	  case edd_valuator:
		{
		double ll;
		double ul;
		double value;

		switch ((enum edd_valuator_constraint_type)pmodifyge->psinfo->pinfo)
		  {
		  case edd_x_pos:
		  case edd_width:			ll = 0;					ul = WidthOfScreen(pdue->s);	break;
		  case edd_y_pos:
		  case edd_height:			ll = 0;					ul = HeightOfScreen(pdue->s);	break;
		  case edd_x_pos_on_root:
		  case edd_width_on_root:	ll = pdue->gizmo_width;	ul = WidthOfScreen(pdue->s);	break;
		  case edd_y_pos_on_root:
		  case edd_height_on_root:	ll = pdue->gizmo_width;	ul = HeightOfScreen(pdue->s);	break;
		  case edd_begin_angle:		ll = -XDEG_PER_REV;		ul = XDEG_PER_REV;				break;
		  case edd_path_angle:		ll = XDEG_PER_REV/10;	ul = XDEG_PER_REV;				break;
		  case edd_line_width:		ll = 0;					ul = 64;						break;
		  case edd_no_constraint:
		  default:					assert(FALSE);
		  }
		pg = create_valuator_gizmo(
				pformg->pgg->s, 
				pformg->w, 
				&obj,
				ok_to_operate,
				valuator_update,
				pdue,
				ll,
				ul,
				gz_point_down,
				sizeof(struct modify_gizmo_ext));
		value = dl_to_double(pmodifyge->pstructure, pmodifyge->psinfo, pmodifyge->array_index);
		set_valuator_gizmo(pg, value);
		break;
		}
	  case edd_text_edit:
		{
		char 	*pstring;

		pg = create_textedit_gizmo(	
				pformg->pgg->s, 
				pformg->w, 
				&obj,
				ok_to_operate,
				textedit_update,
				pdue,
				EDD_MAXCHAR,
				sizeof(struct modify_gizmo_ext));
		pstring = dl_to_pstring(pmodifyge->pstructure, pmodifyge->psinfo, &pdue->ct, &pdue->st, pmodifyge->array_index);
		set_textedit_gizmo(pg, pstring);
		break;
		}
	  case edd_menu:
		pg = create_menu_gizmo(
				pformg->pgg->s, 
				pformg->w, 
				&obj,
				ok_to_operate,
				menu_update,
				pdue,
				pmodifyge->psinfo->pDisp,
				sizeof(struct modify_gizmo_ext));
		ival = dl_to_long(pmodifyge->pstructure, pmodifyge->psinfo, pmodifyge->array_index);
		set_menu_gizmo(pg, ival);
		break;
	  case edd_button:
		pg = create_choice_gizmo(
				pformg->pgg->s, 
				pformg->w, 
				&obj,
				ok_to_operate,
				menu_update,
				pdue,
				pmodifyge->psinfo->pDisp,
				gz_stack_row,
				sizeof(struct modify_gizmo_ext));
		ival = dl_to_long(pmodifyge->pstructure, pmodifyge->psinfo, pmodifyge->array_index);
		set_choice_gizmo(pg, ival);
		break;
	  case edd_palette:
		{
		/*
		 *	The double pointer "pppixels" is safe for the palette gizmo to use even if the color map is swapped.
		 */
		pg = create_palette_gizmo(	
				pformg->pgg->s, 
				pformg->w, 
				&obj,
				palette_update,
				pdue,
				&pdue->ppixels,
				&pdue->npixels,
				sizeof(struct modify_gizmo_ext));
		ival = dl_to_long(pmodifyge->pstructure, pmodifyge->psinfo, pmodifyge->array_index);
		set_palette_gizmo(pg, ival);
		break;
		}
	  case edd_none:
		pg = NULL;
		break;
	  default:
		assert(FALSE);
	  }

#ifdef DEBUG
	printf(opiStringTab[STR_gizmo_window_Hxn            ], pg->w);
	printf(opiStringTab[STR_gizmo_xy_HdHdn              ], pg->obj.x, pg->obj.y);
	printf(opiStringTab[STR_gizmo_htwt_HdHdn            ], pg->obj.height, pg->obj.width);
#endif

	if (pg)
		{
		struct modify_gizmo_ext	*pmge	= (struct modify_gizmo_ext *) pg->pae;

		*pmge = *pmodifyge;
		form_gizmo_add_gizmo(pformg, column, row, pg);
		}
	}


/*
 *	VALUATOR_UPDATE
 */
static void valuator_update(pdue, pg, value)
due		*pdue;
gizmo	*pg;
double	value;
	{
	eue						oldeue;
	struct modify_gizmo_ext *pmodifyge;
	int						status;

	pmodifyge = (struct modify_gizmo_ext *) pg->pae;

	if (pmodifyge->peue)
		oldeue = *pmodifyge->peue;

	/*
	 * save old value on the undo queue
	 */
	push_field_undo(pdue, pmodifyge->peue, pmodifyge->psinfo, pmodifyge->pstructure, pmodifyge->array_index);

	/*
	 * modify the dl
	 */
	status = double_to_dl(pmodifyge->pstructure, pmodifyge->psinfo, value, pmodifyge->array_index);
	if (status < 0)
		{
		sprintf(pdue->pmsg, opiStringTab[STR_Invalid_analog_value_rejecte]);
		redraw_msg(pdue);
		return;
		}

	if (pmodifyge->peue)
		{
		if (pmodifyge->peue->elem.dl_anygo.element.type == dl_line)
			{
			int	deltax	= pmodifyge->peue->elem.dl_line.object.x - oldeue.elem.dl_line.object.x;
			int	deltay	= pmodifyge->peue->elem.dl_line.object.y - oldeue.elem.dl_line.object.y;
			
			adjust_line_endpoints(&(pmodifyge->peue->elem.dl_line), deltax, deltay);
			}
		compute_region_and_redraw(pdue, pmodifyge->peue, &oldeue);
		}
	else
		configure_due(pdue, pmodifyge->psinfo->name);
	}


/*
 *	PALETTE_UPDATE
 */
static void palette_update(pdue, pg, value)
due		*pdue;
gizmo	*pg;
long	value;
	{
	eue			oldeue;
	struct modify_gizmo_ext *pmodifyge;
	int			status;

	pmodifyge = (struct modify_gizmo_ext *) pg->pae;

	if (pmodifyge->peue)
		oldeue = *pmodifyge->peue;
	
	/*
	 */
	push_field_undo(pdue, pmodifyge->peue, pmodifyge->psinfo, pmodifyge->pstructure, pmodifyge->array_index);

	/*
	 * save old value on the undo queue
	 * modify the dl
	 */
	status = long_to_dl(pmodifyge->pstructure, pmodifyge->psinfo, value, pmodifyge->array_index);
	if (status < 0)
		{
		sprintf(pdue->pmsg, opiStringTab[STR_Invalid_color_index_rejected]);
		redraw_msg(pdue);
		return;
		}

	if (pmodifyge->peue)
		compute_region_and_redraw(pdue, pmodifyge->peue, &oldeue);
	else
		configure_due(pdue, pmodifyge->psinfo->name);
	}


/*
 *	TEXTEDIT_UPDATE
 */
static void textedit_update(pdue, pg, pvalue)
due		*pdue;
gizmo	*pg;
char	*pvalue;
	{
	eue						oldeue;
	struct modify_gizmo_ext *pmodifyge;
	int						status;

	pmodifyge = (struct modify_gizmo_ext *) pg->pae;

	if (pmodifyge->peue)
		oldeue = *pmodifyge->peue;

	/*
	 * save old value on the undo queue
	 * and modify the dl
	 */
	push_field_undo(pdue, pmodifyge->peue, pmodifyge->psinfo, pmodifyge->pstructure, pmodifyge->array_index);
	status = pstring_to_dl(pmodifyge->pstructure, pmodifyge->psinfo, pvalue, &pdue->ct, &pdue->st, pmodifyge->array_index);
	if (status < 0)
		{
		sprintf(pdue->pmsg, opiStringTab[STR_Invalid_text_entry_ignoredn ]);
		redraw_msg(pdue);
		return;
		}

	if (pmodifyge->peue)
		compute_region_and_redraw(pdue, pmodifyge->peue, &oldeue);
	else
		configure_due(pdue, pmodifyge->psinfo->name);
	}


/*
 *	MENU_UPDATE
 */
static void menu_update(pdue, pg, value)
due		*pdue;
gizmo	*pg;
int		value;
	{
	eue						neweue;
	eue						oldeue;
	eue						*peue;
	struct modify_gizmo_ext *pmodifyge;
	int						x;
	int						y;
	int						status;

	pmodifyge = (struct modify_gizmo_ext *) pg->pae;

	/*
	 * save old value on the undo queue
	 * and modify the dl
	 */
	push_field_undo(pdue, pmodifyge->peue, pmodifyge->psinfo, pmodifyge->pstructure, pmodifyge->array_index);
	peue = pmodifyge->peue;
	if (peue)
		{
		x = peue->pg->obj.x;
		y = peue->pg->obj.y;
		neweue = *peue;
		oldeue = *peue;

		/*
		 * convert into a temp eue so I can compare to the unmodified eue
		 */
		status = long_to_dl(((char *)&neweue) + pmodifyge->offset, pmodifyge->psinfo, value, pmodifyge->array_index);
		if (status < 0)
			{
			sprintf(pdue->pmsg, opiStringTab[STR_Menu_index_does_not_match_fi]);
			redraw_msg(pdue);
			return;
			}
		change_element(pdue, peue, &neweue);
		compute_region_and_redraw(pdue, peue, &oldeue);
		}
	else
		{
		status = long_to_dl(pmodifyge->pstructure, pmodifyge->psinfo, value, pmodifyge->array_index);
		if (status < 0)
			{
			sprintf(pdue->pmsg, opiStringTab[STR_Menu_index_does_not_match_fi]);
			redraw_msg(pdue);
			return;
			}
		configure_due(pdue, pmodifyge->psinfo->name);
		}
	}


/*
 *	MODIFY FINISH
 */
static void modify_finish(ppg, pg, value)
gizmo	**ppg;	/* the location saving this gizmo pointer */
gizmo	*pg;
int		value;
	{
	enum modify_options  option = value;

#ifdef DEBUG
	assert(*ppg == pg);
#endif
	switch(option)
	  {
	  case option_quit:
			free_gizmo(*ppg);
			*ppg = NULL;
			break;
	  default:
			assert(FALSE);
	  }
	}


/*
 *	COMPUTE REGION AND REDRAW
 *
 *	NOTE: perhaps sliders are the only type which require a redraw from the bottom
 */
static void compute_region_and_redraw(pdue, pneweue, poldeue)
due	*pdue;
eue	*pneweue;
eue	*poldeue;
	{
	XRectangle	newbounds;
	XRectangle	oldbounds;
	XRectangle	rec;	/* the area of the window to redraw 	*/


	compute_bounds(pneweue, &newbounds);
	compute_bounds(poldeue, &oldbounds);

	/* 
	 * calculate the region modified including possible select markers
	 */
	rec.x		= min(oldbounds.x, newbounds.x);
	rec.y		= min(oldbounds.y, newbounds.y);
	rec.width	= max((int)(oldbounds.x+oldbounds.width), (int)(newbounds.x+newbounds.width))-rec.x;
	rec.height	= max((int)(oldbounds.y+oldbounds.height), (int)(newbounds.y+newbounds.height))-rec.y;

	redraw_dl(pdue, &rec);
	}



enum channel_column_headers
	{
	old_col,
	equal_col,
	new_col,
	chan_last_col
	};

struct channel_list
	{
	NODE	node;
	char	channel[EDD_MAXCHAR];	/* the channel name */
	LIST	list;		/* a list of channel_list_elements */
	};

struct channel_list_element 
	{
	NODE					node;
	struct modify_gizmo_ext	mge;
	char					*channel;	/* the channel name */
	};

struct augmented_mge
	{
	struct modify_gizmo_ext	mge;
	struct channel_list		*pchan;
	};

static unsigned add_channel_to_modify_list(pdue, psinfo, pstruct, peue, channel)
due				*pdue;
struct sinfo	*psinfo;
void			*pstruct;
eue				*peue;
char 			*channel;
	{
	struct channel_list		*pchan;
	struct modify_gizmo_ext	modifyge;
	char					*pname;
	char					*dot;
	unsigned 				length;
	unsigned 				num_channels	= 0;

	modifyge.peue			= peue;
	modifyge.offset			= 0;
	modifyge.array_index	= 0;
	modifyge.psinfo			= psinfo;
	modifyge.pstructure		= pstruct;
	
	assert(modifyge.psinfo->input == edd_text_edit);
	stringName(&pdue->ct, channel, &pname, &length);
	length = (dot = strchr(pname, '.')) ? dot - pname : length;
	if (length != 0)
		{
		struct channel_list_element	*pelem;

		for (pchan = (void *)lstFirst(&pdue->channel_list); pchan; pchan = (void *)lstNext((NODE *)pchan))
			if (strncmp(pchan->channel, pname, length) == 0 && (
			    (pchan->channel[length] == 0 && pname[length] == '.') ||
			    (pchan->channel[length] == '.' && pname[length] == 0) ||
			    (pchan->channel[length] == '.' && pname[length] == '.') ||
			    (pchan->channel[length] == 0 && pname[length] == 0)
				))
				break;

		if (pchan == NULL)
			{
			/*
			 * didn't find this channel, create a list for it
			 *
			 * this gets real tricky...
			 * we don't want to present the channel field name to the user, just the channel name part
			 * but this causes problems with color rule arg strings which allow more then just a single
			 * channel name, i.e. A=John.ACK,B=Paul.ACK,C=George.ACK,D=Ringo.ACK
			 * these multipart strings always include at least one ',' character
			 *
			 * since it's not easy to break apart these multipart strings and present each part, we'll
			 * just put the whole string into the list, i.e. ignore the '.' for strings including ','
			 *
			 * for simple channel names which include field names i.e. contain a '.' we'll put the
			 * channel name in without the '.' and reattach the field name part in channel_list_update()
			 */
			/*pchan = (struct channel_list *)calloc(1, sizeof (struct channel_list));*/
			pchan = (struct channel_list *)
					tisCalloc(__FILE__, __LINE__, 1, sizeof (struct channel_list));
			strcpy(pchan->channel, pname);
			if ((dot = strchr(pchan->channel, '.')) && (strchr(pchan->channel, ',') == NULL))
				*dot = 0;
			lstInit(&pchan->list);
			lstAdd(&pdue->channel_list, &pchan->node);
			num_channels = 1;
			}
		/* add an element to the channel list */
		/*pelem = (struct channel_list_element *)calloc(1, sizeof (struct channel_list_element));*/
		pelem = (struct channel_list_element *)
					tisCalloc(__FILE__, __LINE__, 1, sizeof (struct channel_list_element));
		pelem->mge = modifyge;
		pelem->channel = pname;
		lstAdd(&pchan->list, &pelem->node);
		}
	return num_channels;
	}

/*
 * MODIFY_SELECTED_CHANNEL_NAMES
 */
void modify_selected_channel_names(param, pdue, pbe)
void			*param;
due				*pdue;
XButtonEvent	*pbe;
	{
	int					x_root	= pbe->x_root - pdue->pixel_tolerence;
	int					y_root	= pbe->y_root - pdue->pixel_tolerence;
	int					height	= pdue->gizmo_height;
	eue_sel				*psel	= (eue_sel *) pdue->selections.node.next;
	eue_sel				*ps;
	gizmo				*pformg;
	int					row;
	int	 				num_channels;
	struct channel_list	*pchan;

	if (psel->pg)
		{
		move_gizmo(psel->pg, x_root, y_root);	/* Dont allow two forms at once */
		return;
		}

	/*
	 * build the list of channel lists
	 */
	lstInit(&pdue->channel_list);
	for (ps = psel, num_channels = 0; ps; ps = (eue_sel *) ps->node.next)
		{
		struct sinfo	*psinfo;
		void			*pstruct;
		eue				*peue	= &ps->eue;
		char			 *channel;

		switch (dl_element_class[ps->eue.elem.dl_anygo.element.type])
		  {
		  case dl_class_graphic:
			if (dynamic_graphical_element(peue))
				{
				psinfo	= &dl_dyn_param_sinfo[chan_index_dl_dyn_param];
				pstruct	= &ps->eue.dyn.param;
#ifndef TOKENPTRS
				channel	= ps->eue.dyn.param.chan;
#else
				channel	= dmPointer64fromToken (ps->eue.dyn.param.chan);
#endif
				num_channels += add_channel_to_modify_list(pdue, psinfo, pstruct, peue, channel);
				psinfo	= &dl_dyn_param_sinfo[clrargs_index_dl_dyn_param];
				pstruct	= &ps->eue.dyn.param;
#ifndef TOKENPTRS
				channel	= ps->eue.dyn.param.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.dyn.param.clrargs);
#endif
				num_channels += add_channel_to_modify_list(pdue, psinfo, pstruct, peue, channel);
				}
			break;
	      case dl_class_monitor:
			psinfo	= &dl_monitor_hdr_sinfo[rdbk_index_dl_monitor_hdr];
			pstruct	= &ps->eue.elem.dl_anymo.monitor;
#ifndef TOKENPTRS
			channel	= ps->eue.elem.dl_anymo.monitor.rdbk;
#else
			channel	= dmPointer64fromToken (ps->eue.elem.dl_anymo.monitor.rdbk);
#endif
			num_channels += add_channel_to_modify_list(pdue, psinfo, pstruct, peue, channel);
			switch (ps->eue.elem.dl_anymo.element.type)
			  {
			  case dl_bar:
				psinfo	= &dl_bar_sinfo[clrargs_index_dl_bar];
				pstruct	= &ps->eue.elem.dl_bar;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_bar.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_bar.clrargs);
#endif
				break;
			  case dl_meter:
				psinfo	= &dl_meter_sinfo[clrargs_index_dl_meter];
				pstruct	= &ps->eue.elem.dl_meter;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_meter.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_meter.clrargs);
#endif
				break;
			  case dl_tank:
				psinfo	= &dl_tank_sinfo[clrargs_index_dl_tank];
				pstruct	= &ps->eue.elem.dl_tank;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_tank.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_tank.clrargs);
#endif
				break;
			  case dl_indicator:
				psinfo	= &dl_indicator_sinfo[clrargs_index_dl_indicator];
				pstruct	= &ps->eue.elem.dl_indicator;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_indicator.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_indicator.clrargs);
#endif
				break;
			  case dl_text_update:
				psinfo	= &dl_text_update_sinfo[clrargs_index_dl_text_update];
				pstruct	= &ps->eue.elem.dl_text_update;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_text_update.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_text_update.clrargs);
#endif
				break;
			  }
			num_channels += add_channel_to_modify_list(pdue, psinfo, pstruct, peue, channel);
			break;
	      case dl_class_control:
			psinfo	= &dl_control_hdr_sinfo[ctrl_index_dl_control_hdr];
			pstruct	= &ps->eue.elem.dl_anyco.control;
#ifndef TOKENPTRS
			channel	= ps->eue.elem.dl_anyco.control.ctrl;
#else
			channel	= dmPointer64fromToken (ps->eue.elem.dl_anyco.control.ctrl);
#endif
			num_channels += add_channel_to_modify_list(pdue, psinfo, pstruct, peue, channel);
			switch (ps->eue.elem.dl_anymo.element.type)
			  {
			  case dl_message_button:
				psinfo	= &dl_message_button_sinfo[clrargs_index_dl_message_button];
				pstruct	= &ps->eue.elem.dl_message_button;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_message_button.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_message_button.clrargs);
#endif
				break;
			  case dl_valuator:
				psinfo	= &dl_valuator_sinfo[clrargs_index_dl_valuator];
				pstruct	= &ps->eue.elem.dl_valuator;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_valuator.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_valuator.clrargs);
#endif
				break;
			  case dl_button:
				psinfo	= &dl_button_sinfo[clrargs_index_dl_button];
				pstruct	= &ps->eue.elem.dl_button;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_button.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_button.clrargs);
#endif
				break;
			  case dl_choice_button:
				psinfo	= &dl_choice_button_sinfo[clrargs_index_dl_choice_button];
				pstruct	= &ps->eue.elem.dl_choice_button;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_choice_button.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_choice_button.clrargs);
#endif
				break;
			  case dl_menu:
				psinfo	= &dl_menu_sinfo[clrargs_index_dl_menu];
				pstruct	= &ps->eue.elem.dl_menu;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_menu.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_menu.clrargs);
#endif
				break;
			  case dl_text_entry:
				psinfo	= &dl_text_entry_sinfo[clrargs_index_dl_text_entry];
				pstruct	= &ps->eue.elem.dl_text_entry;
#ifndef TOKENPTRS
				channel	= ps->eue.elem.dl_text_entry.clrargs;
#else
				channel	= dmPointer64fromToken (ps->eue.elem.dl_text_entry.clrargs);
#endif
				break;
			  }
			num_channels += add_channel_to_modify_list(pdue, psinfo, pstruct, peue, channel);
			break;
	      case dl_class_plot:
			if (ps->eue.elem.dl_anygo.element.type == dl_strip)
				{
				unsigned	i;

				psinfo	= &dl_pen_sinfo[chan_index_dl_pen];
				for (i = 0; i < 8; i++)
					{
					pstruct	= &ps->eue.elem.dl_strip.pen[i];
#ifndef TOKENPTRS
					channel	= ps->eue.elem.dl_strip.pen[i].chan;
#else
					channel	= dmPointer64fromToken (ps->eue.elem.dl_strip.pen[i].chan);
#endif
					num_channels += add_channel_to_modify_list(pdue, psinfo, pstruct, peue, channel);
					}
				}
			break;
		  }
		}
	
	if (num_channels == 0)
		{
		/*
		 * none to do, free up the memory and get out
		 */
		for(pchan = (void *)lstFirst(&pdue->channel_list); pchan; pchan = (void *)lstNext((NODE *)pchan))
			lstFree(&((struct channel_list *)pchan)->list);
		lstFree(&pdue->channel_list);
		return;
		}

	modify_choices[0] = opiStringTab[STR_Quit                        ];
	modify_choices[1] = 0;
	
	/*
	 * create and fill in the form
	 */
	psel->pg = create_popup_gizmo(
				pdue->s, 
				pdue->w,
				opiStringTab[STR_Channel_Name_Modify         ],
				x_root, 
				y_root,
				pdue->gizmo_height,
				modify_channels_finish,
				&psel->pg,
				modify_choices,
				sizeof (due *));
	*(due **)(psel->pg->pae) = pdue;
	set_gizmo_color(psel->pg, pdue->foreground.pixel, pdue->background.pixel);
	
	pformg = create_form_gizmo(pdue->s, psel->pg->w, 0, 0, chan_last_col, num_channels + 2, pdue->gizmo_pad, 0);

	row = 1;			/* add column headers */
	form_gizmo_add_string(pformg, old_col, row, height, opiStringTab[STR_old                         ] );
	form_gizmo_add_string(pformg, new_col, row, height, opiStringTab[STR_new                         ] );

	row++;				/* list the channel name lines */
	for (pchan = (void *)lstFirst(&pdue->channel_list); pchan; pchan = (void *)lstNext((NODE *)pchan), row++)
		{
		XRectangle	obj;
		gizmo		*pg;

		obj.x		= 0;
		obj.y		= 0;
		obj.height	= pdue->gizmo_height;
		obj.width	= pdue->gizmo_width;
		pg = create_textedit_gizmo(	
					pformg->pgg->s, 
					pformg->w, 
					&obj,
					ok_to_operate,
					channel_list_update,
					pdue,
					EDD_MAXCHAR,
					sizeof (struct augmented_mge));
		((struct augmented_mge *)(pg->pae))->pchan = pchan;
		set_textedit_gizmo(pg, pchan->channel);

		form_gizmo_add_string(pformg, old_col,   row, height, pchan->channel);
		form_gizmo_add_string(pformg, equal_col, row, height, opiStringTab[STR__becomes_                   ]);
		form_gizmo_add_gizmo( pformg, new_col,   row, pg);
		}
	resize_form_gizmo(pformg);
	popup_gizmo_add_gizmo(psel->pg, pformg);
/* redundant with popup_gizmo_add_gizmo! agi
**	XMapWindow(psel->pg->pgg->d, psel->pg->w);
*/
	}

static void channel_list_update(pdue, pg, pname)
due		*pdue;
gizmo	*pg;
char	*pname;
	{
	NODE				*pnode;
	struct channel_list	*pchan	= ((struct augmented_mge *)(pg->pae))->pchan;

	for(pnode = lstFirst(&pchan->list); pnode; pnode = lstNext(pnode))
		{
		char	*strstr();
		char	new_channel_name[EDD_MAXCHAR];	/* could be a clrarg string instead of a channel name */
		char	*old_string		= ((struct channel_list_element *)pnode)->channel;
		char	*pattern_pos	= strstr(old_string, pchan->channel);
		
		if (!pattern_pos)
			{
			printf(opiStringTab[STR_edd_channel_list_update_fail],
			pchan->channel, old_string);
			return;
			}
		new_channel_name[0] = 0;
		strncpy(new_channel_name, old_string, pattern_pos - old_string);
		strcat(new_channel_name, pname);
		strcat(new_channel_name, pattern_pos + strlen(pchan->channel));
		*((struct modify_gizmo_ext *)(pg->pae)) = ((struct channel_list_element *)pnode)->mge;
		textedit_update(pdue, pg, new_channel_name);
		}
	}

static void modify_channels_finish(ppg, pg, value)
gizmo	**ppg;
gizmo	*pg;
int		value;
	{
	NODE				*pnode;
	enum modify_options	option = value;

	assert(*ppg == pg);
	assert(option == option_quit);

	for(pnode = lstFirst(&((*(due **)(pg->pae))->channel_list)); pnode; pnode = lstNext(pnode))
		lstFree(&((struct channel_list *)pnode)->list);
	lstFree(&((*(due **)(pg->pae))->channel_list));

	free_gizmo(*ppg);
	*ppg = NULL;
	}
