/*gz.c,v 1.1.1.1 1999/08/03 14:43:57 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *
 *	G Z . C
 *
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *
 *	core code for the gizmo user interface library
 *
 * 	History
 *	.01 joh 032891	streamlined gizmo creation 
 *	.02 joh 102291	subtract border with from gizmo size
 */

#include 	<stdio.h>
#include 	<Xlib.h>
#include 	<Xutil.h>
#ifdef AXP
#include	<stdlib.h>
#endif
#include	<cursorfont.h>
#include	<keysym.h>
#include	<gray.h>
#include	<plain_menus.h>
#include	<font_table.h>
#define GIZMOGLOBALDEF
#include	<gizmo.h>
#include	<color_map.h>

void *tisMalloc();
void *tisCalloc();
void tisFree();

static gizmo_global 	*gizmo_init();
void 					move_resize_gizmo();
static void 			align_onscreen();
static gizmo 			*find_best_unused_gizmo();
static void 			retire_gizmo();


#define GIZMO_MIN_EVENT_MASK 	(StructureNotifyMask)


/*
 *	GIZMO_INIT
 */
static gizmo_global *gizmo_init(s)
Screen		*s;		/* screen pointer */
	{
	Display			*d		= DisplayOfScreen(s);
	gizmo_global	*pgg	= NULL;
	Status			status;

	if (!gizmo_ctx)
		gizmo_ctx = XUniqueContext();
	if (!gizmo_parent_ctx)
		gizmo_parent_ctx = XUniqueContext();

	/* one gizmo global for each display */
	status = XFindContext(d, RootWindowOfScreen(s), gizmo_parent_ctx, (caddr_t *)&pgg);
	if (status)
		{
		XGCValues 				GC_values;
		XColor					clr;
		XColor					requested_clr;
		XSetWindowAttributes 	xswa;
		Pixmap					gray_stipple_pixmap;
		int						i;

		pgg = (gizmo_global *) tisCalloc(__FILE__, __LINE__, 1, sizeof(gizmo_global));
					/* calloc(1, sizeof (gizmo_global)); */
		if (!pgg)
			return NULL;

		pgg->s					= s;
		pgg->d					= d;
		pgg->XA_WM_TAKE_FOCUS	= XInternAtom(pgg->d, "WM_TAKE_FOCUS", False);

  		GC_values.line_width		= max(gray_width, gray_height);			/* so stipple lines will be grey */
  		GC_values.foreground		= WhitePixelOfScreen(s);
  		GC_values.background		= BlackPixelOfScreen(s);
		GC_values.cap_style			= CapProjecting;
  		GC_values.fill_style		= FillOpaqueStippled;
  		GC_values.function			= GXinvert;
		GC_values.subwindow_mode	= IncludeInferiors;
  		GC_values.stipple			=
		gray_stipple_pixmap			= XCreateBitmapFromData(d, RootWindowOfScreen(s), gray_bits, gray_width, gray_height);

		/********************************************************
		 * Be careful changing the attributes in these GCs		*
		 * (most attributes are not intended to be changed)		*
		 * (create a new GC if you need a new pen mode other	*
		 * than simple color or font changes)					*
		 ********************************************************/
		/********************************************************
		 * The following GC's are for general drawing			*
		 * (modify the fg and bg color if you wish)				*
		 ********************************************************/
		pgg->fg_GC		= XCreateGC(d, RootWindowOfScreen(s),
							GCForeground|GCBackground|GCCapStyle, &GC_values);
#ifdef GRAYNESS
		pgg->grey_fg_GC	= XCreateGC(d, RootWindowOfScreen(s),
							GCForeground|GCBackground|GCCapStyle|GCFillStyle|GCStipple|GCLineWidth, &GC_values);
#else
		pgg->grey_fg_GC	= XCreateGC(d, RootWindowOfScreen(s),
							GCForeground|GCBackground|GCCapStyle|GCFillStyle|GCLineWidth, &GC_values);
#endif
		pgg->invert_GC	= XCreateGC(d, RootWindowOfScreen(s),
							GCLineWidth|GCFunction|GCSubwindowMode, &GC_values);

		/********************************************************
		 * The following GC's are for drawing in 3d				*
		 * (modify only the fg color in these)					*
		 ********************************************************/
		pgg->normal_GC	= XCreateGC(d, RootWindowOfScreen(s),
							GCForeground|GCBackground|GCCapStyle|GCLineWidth, &GC_values);

  		GC_values.line_width		= 2; 
  		GC_values.stipple			= gray_stipple_pixmap;
  		GC_values.foreground		= BlackPixelOfScreen(s);
  		GC_values.background		= WhitePixelOfScreen(s);
		pgg->bright_GC	= XCreateGC(d, RootWindowOfScreen(s),
							GCForeground|GCBackground|GCCapStyle|GCFillStyle|GCStipple|GCLineWidth, &GC_values);

  		GC_values.foreground		= WhitePixelOfScreen(s);
  		GC_values.background		= BlackPixelOfScreen(s);
		pgg->shadow_GC	= XCreateGC(d, RootWindowOfScreen(s),
							GCForeground|GCBackground|GCCapStyle|GCFillStyle|GCStipple|GCLineWidth, &GC_values);


		/********************************************************
		 * The following GC's are for drawing in 3d				*
		 * (dont modify the color of these)						*
		 ********************************************************/
  		GC_values.foreground		= WhitePixelOfScreen(s);
  		GC_values.background		= BlackPixelOfScreen(s);
		pgg->white_GC	= XCreateGC(d, RootWindowOfScreen(s),
							GCForeground|GCBackground|GCCapStyle|GCLineWidth, &GC_values);

  		GC_values.foreground		= BlackPixelOfScreen(s);
  		GC_values.background		= WhitePixelOfScreen(s);
		pgg->black_GC	= XCreateGC(d, RootWindowOfScreen(s),
							GCForeground|GCBackground|GCCapStyle|GCLineWidth, &GC_values);

		/*
		 * if on a color server, use a color for dialog background (this makes the button look better)
		 */
		status = XAllocNamedColor(d, DefaultColormapOfScreen(s), "dim grey", &clr, &requested_clr);
		pgg->dialog_bg	= (status) ? clr.pixel : BlackPixelOfScreen(s);
		pgg->pft		= init_font_table(d);

		/*
		 * The parent of gizmos that loose their application parents
		 */
		xswa.override_redirect = True;
  		pgg->mother = XCreateWindow(d, RootWindowOfScreen(s),
						0,					/* x */
						0,					/* y */
		    			10,					/* wd */
						10,					/* ht */
						0,					/* zero pixel border */
						CopyFromParent,		/* take depth from parent */
						InputOutput,
						CopyFromParent,		/* take visual from parent */
						CWOverrideRedirect,
						&xswa);
		pgg->cursor = XCreateFontCursor(d, XC_hand2);

		/*
		 * call the gizmo specific init functions
		 */
		for (i = gz_none+1; i < gz_last; i++)
			if (gizmo_init_func[i])
				(*gizmo_init_func[i])(pgg);

		assert(XSaveContext(d, RootWindowOfScreen(s), gizmo_parent_ctx, (XPointer)pgg) == 0);
		}
	return pgg;
	}


/*
 *	CREATE_GIZMO
 */
gizmo *create_gizmo(type, s, w, pobj, func, arg, extsize, evmask)
enum gizmo_types	type;
Screen				*s;		/* parents screen */
Window				w;		/* parent window */
XRectangle			*pobj;		
void				(*func)();
void				*arg;
unsigned			extsize;
unsigned			evmask;
	{
	Display			*d		= DisplayOfScreen(s);
	gizmo			*pg		= NULL;
	gizmo_global	*pgg	= NULL;
	unsigned short 	reqsize	= sizeof(gizmo) + extsize;
	XRectangle		obj;
	Status			status;
	
	if ((unsigned int)(pobj->width)  < (GIZMO_BORDER_WIDTH<<1)+1 ||
		(unsigned int)(pobj->height) < (GIZMO_BORDER_WIDTH<<1)+1 )
		return NULL;

	if (!gizmo_parent_ctx)
		/*
		 * one gizmo global per screen
		 */
		gizmo_parent_ctx = XUniqueContext();

	status = XFindContext(d, RootWindowOfScreen(s), gizmo_parent_ctx, (caddr_t *)&pgg);
	if (status)
		{
		pgg = gizmo_init(s);
		assert(pgg != NULL);
		}
	/*
	 * allow for a one pixel border
	 */
	obj.x		= pobj->x;
	obj.y		= pobj->y;
	obj.width	= pobj->width  - (GIZMO_BORDER_WIDTH<<1);
	obj.height	= pobj->height - (GIZMO_BORDER_WIDTH<<1);

	/*
	 *	Make sure popups dont come up off screen
	 */
	if (w == RootWindowOfScreen(s))
		align_onscreen(s, &obj);

	/*
	 * reuse old gizmo if one with a large enough 
	 * memory allocation can be found
	 */
	pg = find_best_unused_gizmo(pgg, reqsize, w, &obj);
	if (pg)
		{
		XWindowChanges			xwc;
		gizmo					oldgz;
		XSetWindowAttributes 	xswa;

		oldgz = *pg;								/* reuse the window by reparenting it */

		lstDelete(&pgg->free_gizmo_list, pg);		/* take it off the free gizmo list */
		pg->free = FALSE;

		/*
		 * NOTE: All attributes changed in a previous life
		 * must be reset to a normal condition here
		 */
		xswa.event_mask			= evmask|GIZMO_MIN_EVENT_MASK;
		xswa.background_pixel	= BlackPixelOfScreen(s);
		xswa.border_pixel		= BlackPixelOfScreen(s);
		XChangeWindowAttributes(d, oldgz.w, CWEventMask|CWBackPixel|CWBorderPixel, &xswa);

#ifdef SOLARIS
		memset (pg, 0, oldgz.size);
#else
		bzero((char *)pg, oldgz.size);
#endif
		pg->w		= oldgz.w;
		pg->tw		= oldgz.tw;
		pg->size	= oldgz.size;
		pg->obj		= obj;
		pg->pgg		= pgg;
		pg->parent	= w;

		if (oldgz.parent != w)
			{
			XReparentWindow(d, pg->w, w,  pg->obj.x, pg->obj.y);
			XReparentWindow(d, pg->tw, w, pg->obj.x, pg->obj.y);
			}
		else if (oldgz.obj.x != pg->obj.x || oldgz.obj.y != pg->obj.y)
			{
			xwc.x = pg->obj.x;
			xwc.y = pg->obj.y;
			XConfigureWindow(d, pg->w,  CWX|CWY, &xwc);
			XConfigureWindow(d, pg->tw, CWX|CWY, &xwc);
			}
		if (oldgz.obj.width != pg->obj.width || oldgz.obj.height != pg->obj.height)
			{
			xwc.width	= pg->obj.width;
			xwc.height	= pg->obj.height;
			XConfigureWindow(d, pg->w,  CWWidth|CWHeight, &xwc);
			XConfigureWindow(d, pg->tw, CWWidth|CWHeight, &xwc);
			if (w == RootWindowOfScreen(s))
				{
				XSizeHints      xsh;

				xsh.min_width  = pg->obj.width;
				xsh.min_height = pg->obj.height;
				xsh.max_width  = pg->obj.width;
				xsh.max_height = pg->obj.height;
				xsh.flags = PMinSize|PMaxSize;
				XSetNormalHints(d, pg->w,  &xsh);
				XSetNormalHints(d, pg->tw, &xsh);
				}
			}
		}
	else
		{
		XSetWindowAttributes 	xswa;

		/*
		 * create from scratch other wise
		 */
		pg = (gizmo *) tisCalloc(__FILE__, __LINE__, 1, reqsize);
				/* calloc(1, reqsize); */
		if (!pg)
			return NULL;
		pg->size	= reqsize;
		pg->obj		= obj;
		pg->pgg		= pgg;
		pg->parent	= w;

		xswa.border_pixel		= BlackPixelOfScreen(s);
		xswa.background_pixel	= BlackPixelOfScreen(s);
		xswa.event_mask			= evmask|GIZMO_MIN_EVENT_MASK;

  		pg->w = XCreateWindow(d, w,
			pg->obj.x,			/* x */
			pg->obj.y,			/* y */
		    pg->obj.width,		/* wd */
			pg->obj.height,		/* ht */
			GIZMO_BORDER_WIDTH,	/* one pixel border */
			CopyFromParent,		/* take depth from parent */
			InputOutput,
			CopyFromParent,		/* take visual from parent */
			CWBorderPixel|CWEventMask|CWBackPixel,
			&xswa);

		xswa.override_redirect = TRUE;
		pg->tw = XCreateWindow(d, w,		/* create a transparent window for events when above is unmapped */
			pg->obj.x,
			pg->obj.y,
		    pg->obj.width,
			pg->obj.height,
			0,
			0,
			InputOnly,
			CopyFromParent,
			CWEventMask|CWOverrideRedirect,
			&xswa);

		/*
		 * obligatory notification of the window manager
		 */
		if (w == RootWindowOfScreen(s))
			{
   		    XSizeHints      xsh;

        	xsh.x			= pg->obj.x;
       	 	xsh.y			= pg->obj.y;
        	xsh.width		= pg->obj.width;
        	xsh.height		= pg->obj.height;
	   	     xsh.min_width	= pg->obj.width;
        	xsh.min_height	= pg->obj.height;
        	xsh.max_width	= pg->obj.width;
        	xsh.max_height	= pg->obj.height;
        	xsh.flags		= USSize|PMinSize|USPosition|PMaxSize;
        	XSetNormalHints(d, pg->w, &xsh);
			}	
		}
	pg->type	= type;
	pg->func	= func;
	pg->arg		= arg;
	pg->fg		= WhitePixelOfScreen(s);
	pg->bg		= BlackPixelOfScreen(s);

	return pg;
	}


/*
 *	FIND_BEST_UNUSED_GIZMO
 */
static gizmo *find_best_unused_gizmo(pgg, nbytes, parent, pobj)
gizmo_global	*pgg;
unsigned		nbytes;
Window			parent;
XRectangle		*pobj;
	{
	gizmo	*pg;
	gizmo	*pg_nbytes_greater	= NULL;
	gizmo	*pg_parent_match	= NULL;
	gizmo	*pg_pos_match		= NULL;
	gizmo	*pg_size_match		= NULL;

	if (pgg->free_gizmo_list.count == 0)
		return NULL;

	for(pg = (gizmo *)pgg->free_gizmo_list.node.next; pg; pg = (gizmo *)pg->node.next)
		{
		int		nbytes_greater	= pg->size >= nbytes;
		int		parent_match	= parent == pg->parent;
		int		pos_match		= pobj->x == pg->obj.x && pobj->y == pg->obj.y;
		int		size_match		= pobj->width == pg->obj.width && pobj->height == pg->obj.height;

		if (nbytes_greater && parent_match)										return pg;	/* close enough, dont wait around */
		else if (pg->size == nbytes && parent_match && pos_match && size_match)	return pg;	/* exact match */
		else if (nbytes_greater && parent_match && pos_match && size_match)		pg_size_match		= pg;
		else if (nbytes_greater && parent_match && pos_match)					pg_pos_match		= pg;
		else if (nbytes_greater && parent_match)								pg_parent_match		= pg;
		else if (nbytes_greater)												pg_nbytes_greater	= pg;
		}

	if (pg_size_match)		return pg_size_match;	
	if (pg_pos_match)		return pg_pos_match;
	if (pg_parent_match)	return pg_parent_match;
	if (pg_nbytes_greater)	return pg_nbytes_greater;
	return NULL;
	}


/*
 *	FREE_GIZMO
 */
void free_gizmo(pg)
gizmo	*pg;
	{
	int		status;
	gizmo	*pgz_parent;

	/*
	 * This should be done prior to window deletion (if required) so subwindows go first. Otherwise
	 * a delete will delete subwindows which could become bad drawables.
	 *
	 * This should be done prior to placing the window on the free queue to prevent a 
	 * gizmo from being reused with its previous children still mapped.
	 */
	if (gizmo_free_func[pg->type])
		(*gizmo_free_func[pg->type])(pg);

	pg->type = gz_none;		/* disable future events to this gizmo */

	/*
	 * if the parent is not a gizmo it could be deleted and this gizmo's window could
	 * be deleted. This prevents that from happening in most cases.
	 * If it does happen I catch the delete event and retire the gizmo.
	 */
	status = XFindContext(pg->pgg->d, pg->parent, gizmo_ctx, (caddr_t *)&pgz_parent);
	if (status)
		{
		/* does an auto unmap */
		XReparentWindow(pg->pgg->d, pg->w, pg->pgg->mother, 0, 0);
		XReparentWindow(pg->pgg->d, pg->tw, pg->pgg->mother, 0, 0);
		}
	/*
	 *	gizmo cant be placed on the free list until I get unmap notification from the server-
	 *	see filter_gizmo_events()
	 */
	if (pg->mapped)
		XUnmapWindow(pg->pgg->d, pg->w);
	else
		{
		lstInsert(&pg->pgg->free_gizmo_list, NULL, pg);
		pg->free = TRUE;
		}
	}


/*
 *	RETIRE_GIZMO
 */
static void retire_gizmo(pg)
gizmo	*pg;
	{
	int status;

	assert(gizmo_ctx);

	status = XDeleteContext(pg->pgg->d, pg->w, gizmo_ctx);
	assert(status == 0);

	if (pg->free)
		lstDelete(&pg->pgg->free_gizmo_list, pg);

	/*assert(free(pg) >= 0);*/
	tisFree(pg);
	}


/*
 *	SET_GIZMO_COLOR
 */
void set_gizmo_color(pg, foreground, background)
gizmo		*pg;
unsigned	foreground;
unsigned	background;
	{
	if (pg->fg == foreground && pg->bg == background)
		/* redraw only on change */
		return;

	pg->fg = foreground;
	pg->bg = background;
	if (pg->fg == invisibleColor || pg->bg == invisibleColor)
		{
		XUnmapWindow(pg->pgg->d, pg->w);
		XMapWindow(pg->pgg->d, pg->tw);
		return;
		}
	else
		{
		XUnmapWindow(pg->pgg->d, pg->tw);
		XMapWindow(pg->pgg->d, pg->w);
		}
	if (pg->type != gz_palette)
		{
     	XSetWindowBackground(pg->pgg->d, pg->w, background);
     	XSetWindowBorder(pg->pgg->d, pg->w, background);
		}

	if (gizmo_color_change_func[pg->type])
		(*gizmo_color_change_func[pg->type])(pg);

	/*
	 *	This forces an expose event.
	 *
	 * 	(It seems that we have to do this even if the server has not told us the
	 *	window has mapped yet to prevent timing problems)
	 */
	XClearArea(pg->pgg->d, pg->w, 0, 0, 0, 0, True);
	}



/*
 *	FILTER_GIZMO_EVENTS
 */
filter_gizmo_events(pevent)
XEvent	*pevent;
	{
	gizmo	*pg	= NULL;
	Display	*d	= pevent->xany.display;
	Window	w	= pevent->xany.window;
	Status	status;

	if (!gizmo_ctx)
		return FALSE;	/* return, if no gizmos created yet */

	status = XFindContext(d, w, gizmo_ctx, (caddr_t *)&pg);
	if (status)
		return FALSE;

	switch(pevent->type)
	  {
	  case ButtonPress:
		/*
		 * solves ancestor grab problem for menus see XGrabButton-
		 * allways send to parent if parent grabbed even if child also grabs
		 */
		if (pevent->xbutton.subwindow != None)
			{
			pevent->xany.window			= pevent->xbutton.subwindow;
			pevent->xbutton.subwindow	= None;
			XPutBackEvent(d, pevent);
			return TRUE;
			}
		break;
	  case ConfigureNotify:
		if (w == pevent->xconfigure.window)
			{
			pg->obj.x		= pevent->xconfigure.x;
			pg->obj.y		= pevent->xconfigure.y;
			pg->obj.width	= pevent->xconfigure.width;
			pg->obj.height	= pevent->xconfigure.height;
			}
		break;
	  case ReparentNotify:
		if (w == pevent->xconfigure.window)
			{
			pg->obj.x	= pevent->xreparent.x;
			pg->obj.y	= pevent->xreparent.y;
			pg->parent	= pevent->xreparent.parent;
			}
		break;
	  case MapNotify:
		if (w == pevent->xmap.window && pevent->xmap.window != pg->tw)
			pg->mapped = TRUE;
		break;
	  case UnmapNotify:
		if (w == pevent->xunmap.window && pevent->xunmap.window != pg->tw)
			{
			pg->mapped = FALSE;
			if (pg->type == gz_none && !pg->free)
				{
				lstInsert(&pg->pgg->free_gizmo_list, NULL, pg);
				pg->free = TRUE;
				}
			}
		break;
	  case DestroyNotify:
		if (w == pevent->xdestroywindow.window)
			{
			printf("~appl window delete murdered gizmo\n");
			retire_gizmo(pg);
			return TRUE;
			}
		break;
	  case Expose:
#ifdef EXPOSE_REMOVE_TWO
		/* clear out extra expose events */
		while (XCheckWindowEvent(pevent->xany.display, pevent->xany.window, ExposureMask, pevent))
			;
#else
		if (pevent->type == Expose)
			if (pevent->xexpose.count)
				return TRUE;
#endif
		break;

	  case MotionNotify:
		/* clear out extra mouse motion events */
		while (XCheckTypedWindowEvent(pevent->xany.display, pevent->xany.window, MotionNotify, pevent))
			;
		break;
	  }

	if (pg->type != gz_none)
		(*gizmo_actions[pg->type])(pevent, pg);		/* call the gizmo specific function */

	return TRUE;
	}


/*
 *	MOVE_GIZMO
 */
void move_gizmo(pg, x, y)
gizmo 	*pg;
int		x;
int		y;
	{
	XWindowChanges	xwc;
	XRectangle	rec;

	rec = pg->obj;
	rec.x = x;
	rec.y = y;

	if (pg->parent == RootWindowOfScreen(pg->pgg->s))
		align_onscreen(pg->pgg->s, &rec);

	xwc.x = rec.x;
	xwc.y = rec.y;
	XConfigureWindow(pg->pgg->d, pg->w, CWX|CWY, &xwc);
	pg->obj = rec;
	}


/*
 *	MOVE_RESIZE_GIZMO
 *
 *	NOTES:	.01	pg->obj is updated only by the configuration notify event
 */
void move_resize_gizmo(pg, pobj)
gizmo 		*pg;
XRectangle 	*pobj;
	{
	Display		*d		= pg->pgg->d;
	XRectangle 	newobj;

	newobj	= *pobj;

	/*
	 *	Make sure popups dont come up off screen
 	 */
	if (pg->parent == RootWindowOfScreen(pg->pgg->s))
		align_onscreen(pg->pgg->s, &newobj);

	pg->obj = newobj;
	XMoveResizeWindow(d, pg->w, newobj.x, newobj.y, newobj.width, newobj.height);

	/*
	 * obligatory notification of the window manager
	 */
	if(pg->parent == RootWindowOfScreen(pg->pgg->s))
		{
       	XSizeHints      xsh;

        xsh.x			= newobj.x;
       	xsh.y			= newobj.y;
        xsh.width		= newobj.width;
        xsh.height		= newobj.height;
        xsh.min_width	= newobj.width;
        xsh.min_height	= newobj.height;
        xsh.max_width	= newobj.width;
        xsh.max_height	= newobj.height;
        xsh.flags		= USSize|PMinSize|USPosition|PMaxSize;
        XSetNormalHints(d, pg->w, &xsh);
		}	
	
	/*
	 * window size changes generate an expose and associated redraw
	 *
	 * gizmo data structure is also updated by the configure notify event.
	 */
	}



/*
 *	ALIGN_ONSCREEN
 *
 *	Make sure popups dont come up off screen
 */
static void align_onscreen(s, prec)
Screen		*s;
XRectangle	*prec;
	{
	int		x	= min(prec->x, (int) (WidthOfScreen(s)  - prec->width  - HORIZ_GIZMO_ROOT_PAD(s)));
	int		y	= min(prec->y, (int) (HeightOfScreen(s) - prec->height - VERT_GIZMO_ROOT_PAD(s)));
	
	prec->x = max(x, HORIZ_GIZMO_ROOT_PAD(s));
	prec->y = max(y, VERT_GIZMO_ROOT_PAD(s));
	}


/*
 *	OK_TO_OPERATE:		Returns TRUE  for gizmo creators that do not need any premissive checks.
 *	NOT_OK_TO_OPERATE:	Returns FALSE for gizmo creators that want read only access.
 */
int ok_to_operate()
	{
	return TRUE;
	}

int not_ok_to_operate()
	{
	return FALSE;
	}
