#include "opiStringDefs.h"
extern	char	*opiStringTab[];
/*t_form_gz.c,v 1.1.1.1 1999/08/03 14:43:57 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	F O R M _ G Z . C
 *
 *	Author: Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	History
 *	.01 120291 joh	No longer includes vxWorks.h
 *
 *	Improvements
 *	.01	Forms grow in size when new items are are added, but do 
 *		not shrink. Perhaps this should be changed.
 */


#include 	<stdio.h>
#include 	<X11/Xlib.h>
#ifdef WIN32
#include <X11/XlibXtra.h>
#endif WIN32
#include 	<X11/Xutil.h>
#include	<X11/keysym.h>
#ifdef AXP
#include <string.h>
#endif
#include	"plain_menus.h"
#include	"font_table.h"
#include 	"gizmo.h"
#include	"tisXext.h"

/*
 *	per form gizmo structures	
 */
struct column_info
	{
	unsigned short	size;
	unsigned short	offset;
#ifdef AXP
	/* quad alignment */
	int		axpPad;
#endif
	};

struct row_info
	{
	unsigned short	size;
	unsigned short	offset;
#ifndef AXP
	Font		font;
	unsigned short	ycentered;
	unsigned short	char_width;
#else
	/* shuffle to get proper quad alignment */
	unsigned short	ycentered;
	unsigned short	char_width;
	Font		font;
#endif
	};

enum occupant_type {no_oc, gizmo_oc, string_oc, color_oc};

struct form_entry
	{
	enum occupant_type 	type;
	void			*poccupant;
	};

typedef struct
	{
	unsigned short		nrows;
	unsigned short		ncolumns;
	unsigned short		margin;
	short			RISC_pad;
	struct column_info	*pcolumn;
	struct row_info		*prow;
	struct form_entry	*pentries;	/* one for each row column entry */
	} form_gizmo_ext;

#define ENTRY_INDEX(PFGE,ROW,COLUMN)  ((PFGE)->nrows*(COLUMN)+(ROW))

void redraw_form_gz();
static void map_occupants();



/*
 * FORM_INIT()
 */
void form_init(pgg)
gizmo_global *pgg;
	{
	}


/*
 * FORM_FREE()
 */
void form_free(pg)
gizmo	*pg;
	{
	form_gizmo_ext	*pfge = (form_gizmo_ext *)(pg+1);
	unsigned 		row;
	unsigned 		column;

	for (row = 0; row < pfge->nrows; row++)
		for (column = 0; column < pfge->ncolumns; column++)
			{
			unsigned index = ENTRY_INDEX(pfge, row, column);
			
			if (pfge->pentries[index].type == gizmo_oc)
				free_gizmo(pfge->pentries[index].poccupant);
			}
	}


/*
 * CREATE_FORM_GIZMO()
 */
gizmo *create_form_gizmo(s, parent_window, x, y, ncolumns, nrows, margin, ext_size)
Screen		*s;
Window		parent_window;
int			x;
int			y;
int			ncolumns;
int			nrows;
unsigned	ext_size;	/* size of the application extension */
	{
	gizmo			*pg;
	form_gizmo_ext	*pfge;
	XRectangle		obj;		
	unsigned		size;

	obj.x		= x;
	obj.y		= y;
	obj.height	= GIZMO_BORDER_WIDTH<<2;	/* ukn at this point */
	obj.width	= GIZMO_BORDER_WIDTH<<2;	/* ukn at this point */

	size = sizeof (form_gizmo_ext) + 
		sizeof (struct form_entry) * nrows * ncolumns +
		sizeof (struct row_info) * nrows  +
		sizeof (struct column_info) * ncolumns +
		ext_size;

	pg = create_gizmo(gz_form, s, parent_window, &obj, NULL, NULL, size, StructureNotifyMask|ExposureMask);
	if(!pg)
		return NULL;

	pfge = (form_gizmo_ext *)(pg+1);
	pfge->nrows = nrows;
	pfge->ncolumns = ncolumns;
	pfge->margin = margin;
	pfge->pcolumn = (struct column_info *) (pfge+1);
	pfge->prow = (struct row_info *) (pfge->pcolumn + ncolumns);
	pfge->pentries = (struct form_entry *) (pfge->prow + nrows);
	if(ext_size)
		pg->pae = (void *) (pfge->pentries + nrows*ncolumns);

	assert(XSaveContext(pg->pgg->d, pg->w,  gizmo_ctx, (XPointer)pg) == 0);
	assert(XSaveContext(pg->pgg->d, pg->tw, gizmo_ctx, (XPointer)pg) == 0);

	return pg;
	}


/*
 *	FORM_ACTION
 */
void form_action(pevent, pg)
XEvent	*pevent;
gizmo	*pg;
	{

	switch(pevent->type)
	  {
	  case Expose:		redraw_form_gz(pg);		break;
	  case MapNotify:	map_occupants(pg, TRUE);	break;
	  case UnmapNotify:	map_occupants(pg, FALSE);	break;
	  }
	}



/*
 *	REDRAW_FORM_GZ
 */
void redraw_form_gz(pg)
gizmo	*pg;
	{
	form_gizmo_ext	*pfge	= (form_gizmo_ext *)(pg+1);
	Display 		*d		= pg->pgg->d;
	Window 			w		= pg->w;
	unsigned		row;
	unsigned		column;
	Font			last_font=0;

	if (!pg->mapped)
		return;

	XSetForeground(pg->pgg->d, pg->pgg->fg_GC, pg->fg);
	XSetBackground(pg->pgg->d, pg->pgg->fg_GC, pg->bg);
	XSetForeground(pg->pgg->d, pg->pgg->grey_fg_GC, pg->fg);
	XSetBackground(pg->pgg->d, pg->pgg->grey_fg_GC, pg->bg);

	for (row = 0; row < pfge->nrows; row++)
		{
		if (last_font != pfge->prow[row].font)
			{
			last_font = pfge->prow[row].font;
			if(last_font)
				XSetFont(d, pg->pgg->fg_GC, pfge->prow[row].font);
			}

		for (column = 0; column < pfge->ncolumns; column++)
			{
			unsigned index = ENTRY_INDEX(pfge,row,column);

			switch (pfge->pentries[index].type)
			  {
			  case gizmo_oc:
				break;
			  case string_oc:
				if (last_font)
					{
					XSetForeground(d, pg->pgg->fg_GC, pg->fg);
					tisDrawString(d, w, pg->pgg->fg_GC, pfge->pcolumn[column].offset+pfge->prow[row].char_width,
					  pfge->prow[row].offset+pfge->prow[row].ycentered, pfge->pentries[index].poccupant,
					  strlen(pfge->pentries[index].poccupant));
					}
				break;
			  case color_oc:
				{
				long	pixel	= (long)pfge->pentries[index].poccupant;

				XSetForeground(d, pg->pgg->fg_GC, pixel);
				XFillRectangle(d, w, pg->pgg->fg_GC, pfge->pcolumn[column].offset, pfge->prow[row].offset,
				  pfge->pcolumn[column].size, pfge->prow[row].size);
				break;
				}
			  case no_oc:
				break;
			  default:
				assert(FALSE);
			  }
			}
		}
	}


/*
 *	RESIZE_FORM_GIZMO
 *
 *	resize window and set text drawing info
 */
void resize_form_gizmo(pg)
gizmo		*pg;
	{
	form_gizmo_ext	*pfge	= (form_gizmo_ext *)(pg+1);
	unsigned 		row;
	unsigned 		column;
	unsigned		height	= 0;
	unsigned		width	= 0;
	unsigned		index;
	XRectangle		obj;

	/* do this first so the font info is set up before setting col widths */
	for (row = 0; row < pfge->nrows; row++)
		{
		XFontStruct	*pfi;
		unsigned	index;
		unsigned	ft_ht;

		index = min(NELEMENTS(pg->pgg->pft->height_to_pfinfo)-1, pfge->prow[row].size);
		pfi = pg->pgg->pft->height_to_pfinfo[index];
		if(!pfi)
			continue;
		pfge->prow[row].font = pfi->fid;
		ft_ht = pfi->max_bounds.ascent + pfi->max_bounds.descent;
		pfge->prow[row].ycentered = ((pfge->prow[row].size-ft_ht)>>1)+pfi->max_bounds.ascent;
		pfge->prow[row].char_width = pfi->max_bounds.width;
		height += pfge->margin;
		pfge->prow[row].offset = height;
		height += pfge->prow[row].size;
		}

	/* resize to the string length grows, but never shrinks */
	for (row = 0; row < pfge->nrows; row++)
		for (column = 0; column < pfge->ncolumns; column++)
			{
			index = ENTRY_INDEX(pfge, row, column);
			if (pfge->pentries[index].type == string_oc)
				{
				int size = (strlen(pfge->pentries[index].poccupant)+2) * pfge->prow[row].char_width;

				pfge->pcolumn[column].size = max(size, (int)(pfge->pcolumn[column].size));
				}
			}


	for (column = 0; column < pfge->ncolumns; column++)
		{
		width += pfge->margin;
		pfge->pcolumn[column].offset = width;
		width += pfge->pcolumn[column].size;
		}

	height += pfge->margin;
	width += pfge->margin;

	obj.x		= pg->obj.x;
	obj.y		= pg->obj.y;
	obj.width	= width;
	obj.height	= height;
	move_resize_gizmo(pg, &obj);

	/* finally place gizmos in their proper locations */
	for (row = 0; row < pfge->nrows; row++)
		for (column = 0; column < pfge->ncolumns; column++)
			{
			index = ENTRY_INDEX(pfge, row, column);
			if (pfge->pentries[index].type == gizmo_oc)
				{
				XRectangle	obj;
				gizmo		*pog = pfge->pentries[index].poccupant;
				
				obj.x		= pfge->pcolumn[column].offset;
				obj.y		= pfge->prow[row].offset;
				obj.width	= pfge->pcolumn[column].size;
				obj.height	= pfge->prow[row].size;
				move_resize_gizmo(pog, &obj);
				}
			}
	}




/*
 *	MAP_OCCUPAMNTS
 *
 */
static void map_occupants(pg, mapit)
gizmo	*pg;
int		mapit;
	{
	form_gizmo_ext	*pfge = (form_gizmo_ext *)(pg+1);
	unsigned 		row;
	unsigned 		column;
	unsigned		index;

	/* finally place gizmos in their proper locations */
	for (row = 0; row < pfge->nrows; row++)
		for (column = 0; column < pfge->ncolumns; column++)
			{
			index = ENTRY_INDEX(pfge, row, column);
			if (pfge->pentries[index].type == gizmo_oc)
				{
				gizmo *pog = pfge->pentries[index].poccupant;
				
				if(mapit && !pog->mapped)
					XMapWindow(pog->pgg->d, pog->w);
				else if(!mapit && pog->mapped)
					XUnmapWindow(pog->pgg->d, pog->w);
				}
			}
	}


/*
 *	FORM_GIZMO_ITEM_GEOM
 */
void form_gizmo_item_geom(pg, column, row, pgeom)
gizmo		*pg;
unsigned	column;
unsigned	row;
XRectangle	*pgeom;
	{
	form_gizmo_ext	*pfge = (form_gizmo_ext *)(pg+1);

	pgeom->x		= pfge->pcolumn[column].offset;
	pgeom->y		= pfge->prow[row].offset;
	pgeom->width	= pfge->pcolumn[column].size;
	pgeom->height	= pfge->prow[row].size;
	}



/*
 *	FORM_GIZMO_ADD_STRING
 */
void form_gizmo_add_string(pg, column, row, height, pstring)
gizmo		*pg;
unsigned	column;
unsigned	row;
unsigned	height;
char		*pstring;	/* optional- if supplied it replaces width */
	{
	form_gizmo_ext	*pfge = (form_gizmo_ext *)(pg+1);
	unsigned		index;

	assert(column < pfge->ncolumns);
	assert(row    < pfge->nrows);

	pfge->prow[row].size = max(pfge->prow[row].size, height);

	/*
	 * width not calculated now since final height ukn
	 */
	index = ENTRY_INDEX(pfge,row,column);
	pfge->pentries[index].poccupant = (void *)pstring;
	pfge->pentries[index].type = string_oc;
	}



/*
 *	FORM_GIZMO_ADD_GIZMO
 */
void form_gizmo_add_gizmo(pg, column, row, paddg)
gizmo		*pg;
unsigned	column;
unsigned	row;
gizmo		*paddg;
	{
	form_gizmo_ext	*pfge = (form_gizmo_ext *)(pg+1);
	unsigned		index;

	assert(column < pfge->ncolumns);
	assert(row    < pfge->nrows);

	pfge->pcolumn[column].size = max(pfge->pcolumn[column].size, paddg->obj.width);
	pfge->prow[row].size = max(pfge->prow[row].size, paddg->obj.height);
	index = ENTRY_INDEX(pfge,row,column);
	pfge->pentries[index].poccupant = (void *)paddg;
	pfge->pentries[index].type = gizmo_oc;
	set_gizmo_color(paddg, pg->fg, pg->bg);
	}


/*
 *	FORM_GIZMO_ADD_COLOR
 *
 *	setup column, row to display a color
 */
void form_gizmo_add_color(pg, column, row, width, height, pixel)
gizmo		*pg;
unsigned	column;
unsigned	row;
unsigned	width;
unsigned	height;
long		pixel;
	{
	form_gizmo_ext	*pfge = (form_gizmo_ext *)(pg+1);
	unsigned	index;

	assert(column < pfge->ncolumns);
	assert(row    < pfge->nrows);

	pfge->pcolumn[column].size = max(pfge->pcolumn[column].size, width);
	pfge->prow[row].size = max(pfge->prow[row].size, height);
	index = ENTRY_INDEX(pfge,row,column);
	pfge->pentries[index].poccupant = (void *)pixel;
	pfge->pentries[index].type = color_oc;
	}


/*
 *	FORM_GIZMO_SET_COLOR
 *
 *	set solid color for column, row
 */
void form_gizmo_set_color(pg, column, row, pixel)
gizmo		*pg;
unsigned	column;
unsigned	row;
long		pixel;
	{
	form_gizmo_ext	*pfge = (form_gizmo_ext *)(pg+1);
	unsigned		index;

	assert(column < pfge->ncolumns);
	assert(row    < pfge->nrows);

	index = ENTRY_INDEX(pfge,row,column);
	assert(pfge->pentries[index].type == color_oc);

	pfge->pentries[index].poccupant = (void *)pixel;
	pfge->pentries[index].type = color_oc;
	}


/*
 *	FORM_GIZMO_RETURN_PITEM
 *
 *	return pointer to item at row, column
 */
void *form_gizmo_return_pitem(pg, column, row)
gizmo		*pg;
unsigned	column;
unsigned	row;
	{
	form_gizmo_ext	*pfge	= (form_gizmo_ext *)(pg+1);
	unsigned		index	= ENTRY_INDEX(pfge, row, column);

	return pfge->pentries[index].poccupant;
	}


/*
 * FORM_COLOR_CHANGE()
 */
void form_gizmo_color_change(pg)
gizmo	*pg;
	{
	form_gizmo_ext	*pfge = (form_gizmo_ext *)(pg+1);
	unsigned 		row;
	unsigned 		column;
	unsigned		index;
	gizmo			*pgo;

	for (row = 0; row < pfge->nrows; row++)
		for (column = 0; column < pfge->ncolumns; column++)
			{
			index = ENTRY_INDEX(pfge, row, column);
			if (pfge->pentries[index].type == gizmo_oc)
				{
				pgo = pfge->pentries[index].poccupant;
				set_gizmo_color(pgo, pg->fg, pg->bg);
				}
			}

	if (pg->mapped)
		redraw_form_gz(pg);
	}
