#include "opiStringDefs.h"
extern	char	*opiStringTab[];
/*t_menu_gz.c,v 1.1.1.1 1999/08/03 14:43:58 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	M E N U _ G Z . C
 *
 *	Author: Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	gizmo envelope for a plain menu
 *
 *	History
 *	.01 120291 joh	no longer includes vxWorks.h
 */

#include 	<stdio.h>
#include 	<X11/Xlib.h>
#ifdef WIN32
#include <X11/XlibXtra.h>
#endif WIN32
#include 	<X11/Xutil.h>
#include	<X11/keysym.h>
#ifdef AXP
#include	<string.h>
#endif
#include	"plain_menus.h"
#include 	"gizmo.h"
#include	"draw_strings.h"

typedef struct
	{
	XPoint						menu_icon_points[4];
	struct menu_configuration	mc;
	unsigned short				selection;	/* to detect changes 	*/
	int							(*permissive_func)();
	int							permissive_result;
	} menu_gizmo_ext;

static void 	redraw_menu_gz();
static void	resize_menu_gizmo();


/*
 *	MENU_INIT
 */
void menu_init(pgg)
gizmo_global *pgg;
	{
	}


/*
 *	CREATE_MENU_GIZMO
 *
 *	NOTE: multiple level menus not supported (this gizmo only)
 */
gizmo *create_menu_gizmo(s, w, pobj, perm_func, func, arg, pchoices, ext_size)
Screen		*s;		/* parent's screen */
Window		w;		/* parent window */
XRectangle	*pobj;		
int			(*perm_func)();
void		(*func)();
void		*arg;
char		**pchoices;
unsigned	ext_size;
	{
	gizmo			*pg;
	menu_gizmo_ext	*pmge;

	pg = create_gizmo(gz_menu, s, w, pobj, func, arg, sizeof(menu_gizmo_ext) + ext_size,
	  PLAIN_MENU_MIN_EVENT_MASK|ButtonPressMask|EnterWindowMask|LeaveWindowMask|ButtonReleaseMask|ExposureMask);
	if (!pg)
		return NULL;

	pmge = (menu_gizmo_ext *) (pg+1);
	if (ext_size)
		pg->pae = (void *) (pmge+1);

	pmge->mc.pname = "";
	pmge->mc.pchoices = pchoices;
	pmge->mc.pactions = NULL;
	init_menu(pg->pgg->d, &pmge->mc);
	pmge->mc.selection = 0; 	/* for plain menus 	*/
	pmge->selection = 0; 		/* to detect changes 	*/
	pmge->permissive_func = perm_func;
	pmge->permissive_result = FALSE;

	resize_menu_gizmo(pg);
	assert(XSaveContext(pg->pgg->d, pg->w,  gizmo_ctx, (XPointer)pg) == 0);
	assert(XSaveContext(pg->pgg->d, pg->tw, gizmo_ctx, (XPointer)pg) == 0);

	XMapWindow(pg->pgg->d, pg->w);
	return pg;
	}



/*
 * 	MENU_FREE
 */
void menu_free(pg)
gizmo *pg;
	{
	}



/*
 * 	RESIZE_MENU_GIZMO
 */
static void resize_menu_gizmo(pg)
gizmo	*pg;
	{
	menu_gizmo_ext		*pmge = (menu_gizmo_ext *)(pg+1);

	/* first x point set prior to each redraw */
	pmge->menu_icon_points[0].x = 0;
	pmge->menu_icon_points[0].y = pg->obj.height;
	/* the rest are offsets */
	pmge->menu_icon_points[1].x = 0;
	pmge->menu_icon_points[1].y = - pg->obj.height;
	pmge->menu_icon_points[2].x = pg->obj.height>>1;
	pmge->menu_icon_points[2].y = pg->obj.height>>1;
	pmge->menu_icon_points[3].x = -(pg->obj.height>>1);
	pmge->menu_icon_points[3].y = pg->obj.height>>1;
	}


/*
 * 	SET_MENU_GIZMO
 */
void set_menu_gizmo(pg, init)
gizmo 		*pg;
unsigned	init;
	{
	menu_gizmo_ext	*pmge = (menu_gizmo_ext *)(pg+1);

	if (pmge->selection != init)
		{
		pmge->selection = pmge->mc.selection = init;
		redraw_menu_gz(pg);
		}
	}	


/*
 * 	MENU_ACTION
 */
void menu_action(pevent, pg)
XEvent	*pevent;
gizmo	*pg;
	{
	menu_gizmo_ext	*pmge = (menu_gizmo_ext *)(pg+1);
	int				posted;

	switch (pevent->type)
	  {
	  case EnterNotify:
		if ((pmge->permissive_result = (*pmge->permissive_func)(pg->arg, pg, FALSE)) == FALSE)
			break;
		XSetWindowBorder(pg->pgg->d, pg->w, pg->fg);
		XDefineCursor(pg->pgg->d, pg->tw, pg->pgg->cursor);
		break;
	  case LeaveNotify:
		if ( pmge->permissive_result == FALSE )
			break;
		XSetWindowBorder(pg->pgg->d, pg->w, pg->bg);
		XUndefineCursor(pg->pgg->d, pg->tw);
		break;
	  case ButtonPress:
		if ((pmge->permissive_result = (*pmge->permissive_func)(pg->arg, pg, TRUE)) == FALSE)
			break;
		if (pevent->xbutton.button != MENU_BUTTON)
			break;
		posted = post_menu(pevent, &pmge->mc, pg);
		if (!posted)
			/*
			 * execute only if they moused down in the menu
			 */
			break;

		if (pmge->selection != pmge->mc.selection)
			{
			pmge->selection = pmge->mc.selection;
			redraw_menu_gz(pg);
			}
		/*
		 * allways complete all labor for the gizmo prior to calling their routine since they
		 * may delete this gizmo while in it! Call their routine even if the selection has
		 * not changed so the user's action can be detected even if he has taken the default.
		 */
		(*pg->func)(pg->arg, pg, pmge->selection);
		break;
	  case ButtonRelease:
		if (pmge->permissive_result == FALSE)
			break;
		(*pg->func)(pg->arg, pg, pmge->selection);		/* Other buttons are just shortcuts for taking the default */
		break;
	  case Expose:
		redraw_menu_gz(pg);
		break;
	  case ConfigureNotify:
		resize_menu_gizmo(pg);
		break;
	  }
	}



/*
 *	REDRAW_MENU_GZ
 *
 * ##### redo this to reduce network overhead and flicker ? #####
 * 
 * could use draw image string to avoid clearing the window.  (requires a full length string however)
 */
static void redraw_menu_gz(pg)
gizmo *pg;
	{
	menu_gizmo_ext	*pmge		= (menu_gizmo_ext *)(pg+1);
	Display 		*d			= pg->pgg->d;
	Window 			w			= pg->w;
	char 			*pselname	= pmge->mc.pchoices[pmge->selection];
	XRectangle 		rec;

	if (!pg->mapped)
		return;

	XSetForeground(pg->pgg->d, pg->pgg->fg_GC, pg->fg);
	XSetBackground(pg->pgg->d, pg->pgg->fg_GC, pg->bg);
	XSetForeground(pg->pgg->d, pg->pgg->grey_fg_GC, pg->fg);
	XSetBackground(pg->pgg->d, pg->pgg->grey_fg_GC, pg->bg);
	XClearWindow(d, w);

	rec.x		= (pg->obj.height>>1);
	rec.y		= 0;
	rec.width	= pg->obj.width - rec.x;
	rec.height	= pg->obj.height;
	draw_horizontal_string( d, w, pg->pgg->fg_GC, &rec, pg->pgg->pft, pselname, strlen(pselname), box_center);
	XFillPolygon(d, w, pg->pgg->grey_fg_GC, pmge->menu_icon_points, NELEMENTS(pmge->menu_icon_points), Convex, CoordModePrevious);
	}



/*
 *	GET_MENU_GZ_STRING
 */
char	*get_menu_gz_string(pg, index)
gizmo	*pg;
int		index;
	{
	return ((menu_gizmo_ext *)(pg+1))->mc.pchoices[index];
	}
