/*Graph3d.h,v 1.1.1.1 1999/08/03 14:43:59 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 * Graph3d structure definition (allows display of (x,y,z) triples)
 */

#include <X11/Xlib.h>

#include "GraphXMacros.h"


#define GRAPH3D_MARGIN		0.1	/* window fraction for text, &c */
#define GRAPH3D_DRAWINGAREA	0.8	/* 2*MARGIN + DRAWINGAREA == 1 */


/* must define the filename of the temporary xwd file for printing */

#ifndef XWD_TEMP_FILE_STRING
#  define XWD_TEMP_FILE_STRING    "./graph3d.xwd"
#endif



typedef enum {
    Graph3dGrid,		/* XYZ plot (XY defined on grid, color) type */
    Graph3dPoint,		/* XYZ plot (simple (x,y,color)) type */
    Graph3dContour		/* XYZ plot (contour) type */
} Graph3dType;

typedef enum {
    Graph3dInternal,		/* make private copies of all  data */
    Graph3dExternal		/* use the passed-in storage for all data */
} Graph3dStorageMode;

typedef struct {
    double	   x;
    double	   y;
    double	   z;
} XYZdataPoint;

typedef enum {
    Graph3dDefaultRange,	/* do own range calculation: based on data */
    Graph3dSpecRange	/* use user specified ranges */
} Graph3dRangeType;

typedef struct {
    double minVal;
    double maxVal;
} Graph3dRange;



struct _Graph3d {
    Display	    *display;
    int		     screen;
    Window	     window;
    Pixmap	     pixmap;
    GC		     gc;
    Graph3dStorageMode storageMode;

    XYZdataPoint       *dataPoints;
    int              nPoints;

    Graph3dRangeType xRangeType;	/* enumerated type for display range */
    Graph3dRange     xRange;		/* the range itself */
    Graph3dRangeType yRangeType;
    Graph3dRange     yRange;
    Graph3dRangeType zRangeType;
    Graph3dRange     zRange;

    Graph3dType	     type;		/* the enumerated type of Graph3d */
    int		     xBins;		/* the number of X bins (Graph3dBin) */
    int		     yBins;		/* the number of Y bins (Graph3dBin) */
    int		     zBins;		/* the number of Z bins (Graph3dBin) */

    XPoint	    *devicePoints;	/* device coordinates for dataPoints */

    BOOLEAN	     setOwnColors;	/* if TRUE user will fill in pixels */
    unsigned long    forePixel; 	/* foreground color */
    unsigned long    backPixel;		/* background color */
    unsigned long    dataPixel;		/* primary data color */

    unsigned long    *zPixels;		/* color table for Z */

    unsigned long    w;
    unsigned long    h;

    BOOLEAN	     setOwnFonts;	/* if TRUE user will fill in fonts */
    char	    *title;
    char	    *titleFont;
    XFontStruct	    *titleFontStruct;
    int		     titleFontHeight;
    int		     titleFontWidth;
    char	    *xAxisLabel;
    char	    *yAxisLabel;
    char	    *axesFont;
    XFontStruct	    *axesFontStruct;
    int		     axesFontHeight;
    int		     axesFontWidth;

    caddr_t	     userData;		/* pointer to arbitrary user data */

    };
typedef struct _Graph3d Graph3d;


extern Graph3d *graph3dInit();
extern void graph3dDraw();
extern void graph3dPrint();
extern void graph3dRefresh();
extern void graph3dResize();
extern void graph3dSet();
extern void graph3dSetRange();
extern void graph3dSetRangeDefault();
extern void graph3dSetBins();
extern void graph3dSetZBins();
extern void graph3dSetDisplayType();
extern void graph3dTerm();
