/*Surface.h,v 1.1.1.1 1999/08/03 14:44:00 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *  Surface structure definition
 */

#include <X11/Xlib.h>

#include "GraphXMacros.h"

#include "3D.h"


/* must define the filename of the temporary xwd file for printing */

#ifndef XWD_TEMP_FILE_STRING
#  define XWD_TEMP_FILE_STRING    "./surface.xwd"
#endif



typedef enum {
    SurfaceInternal,               /* make private copies of all surface data */
    SurfaceExternal                /* use the passed-in storage for all data */
} SurfaceStorageMode;

typedef enum {
    SurfaceSolid = SOLID,	   /* solid model, hidden line removed */
    SurfaceShaded = SHADED	   /* shaded model */
} SurfaceRenderMode;



struct _Surface {
    Display	   *display;
    int		    screen;
    Window	    window;
    Pixmap	    pixmap;
    GC		    gc;

    SurfaceStorageMode storageMode;
    double	   *mesh, *x, *y;
    int		    nx, ny;

    ThreeD	    *s3d;	    /* pointer to structure for 3D stuff */

    double	    minX;	    /* need these to go from normalized  */
    double	    maxX;	    /*  data [0,1] and back again        */
    double	    deltaX;
    double	    minY;
    double	    maxY;
    double	    deltaY;
    double	    minZ;
    double	    maxZ;
    double	    deltaZ;

    BOOLEAN	    setOwnColors;	/* if TRUE user will fill in pixels */
    unsigned long   forePixel, backPixel, dataPixel;

    unsigned long   w;
    unsigned long   h;

    BOOLEAN	    setOwnFonts;	/* if TRUE user will fill in fonts */
    char	   *title;
    char	   *titleFont;
    XFontStruct	   *titleFontStruct;
    int		    titleFontHeight;
    int		    titleFontWidth;


    caddr_t	    userData;		/* pointer to arbitrary user data */

    };


typedef struct _Surface Surface;


extern Surface *surfaceInit();
extern void surfaceDraw();
extern void surfacePrint();
extern void surfaceRefresh();
extern void surfaceResize();
extern void surfaceSet();
extern void surfaceSetRenderMode();
extern void surfaceSetView();
extern void surfaceTerm();
