/*new_strip.c,v 1.1.1.1 1999/08/03 14:43:59 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/***
 ***   STRIP CHART Plot routines & data structures
 ***
 ***	(MDA) 11 May 1990
 *** Rewritten (and reformatted) using ca_strip.c to maintain data values. (AGI)
 ***
 ***	Modification History :
 ***
 ***	DK	01/28/92	fixed bug in stripTerm that caused XError by freeing font
 ***				that had already been freed.  This occurred when the axes
 ***				font was the same as the title font.  Also commented out
 ***				XFreeFontNames after XFreeFontInfo in stripSet because looks 
 ***				like it may be freeing the same thing twice.
 ***/

#include <memLeak.h>
#include <X11/Xlib.h>

#ifdef AXP
#include <string.h>
#include <stdlib.h>
#endif

#include <math.h>
#include <stdio.h>

#include "Strip.h"

#ifdef SOLARIS
#define random()	(drand48())
#endif

XFontStruct *tisLoadQueryFont();

#define MINBORDER 3		/* minimum distance from window edge (Pixels) */

#define RELATIVETICKSIZE 0.02	/* size of tick mark relative to window */
#define NTICKS 2		/* `preferred' number of ticks for axes */
#define MAXTICKS NTICKS+4	/* `maximum' number of tick based on NTICKS */

#define DATALINEWIDTH 1		/* line width for data lines (Pixels) */
#define BORDERLINEWIDTH 1	/* line width for lines (Pixels) */
#define ALMOSTZERO 1.0e-07	/* close enough to zero */

#define MAXCHANNELS 100		/* largest number of channels to sample */

#define LEGENDOFFSET 1          /* amount of legend window to overlap parent */


#ifdef TIMED
/*
 * modified Strip Chart code to subtract out execution time of update/draw
 *   from user-specified interval.  This more accurately defines sampling
 *   intervals
 */
#  include <sys/times.h>

/* define the granularity of your times(2) function (when used) */
#  define HZ      60              /* times(2) returns 1/60 second (most) */
/*#define HZ    100             /* times(2) returns 1/100 second */

#endif

/*
 * define constants for XListFontsWithInfo call
 */
#define FONTWILDCARD "*"
#define MAXFONTS 1



/* extern char *malloc();*/
extern int getOKStringPosition();
extern int getOKStringPositionXY();



static void updateStrip();


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Initialize the Strip
 * function:  Strip *stripInit()
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ 

Strip *stripInit( display, screen, window)
  Display *display;
  int     screen;
  Window  window;
{
    Strip *strip;
    XWindowAttributes windowAttributes;
    int i, w, h;

    strip = (Strip *) malloc( sizeof( Strip ) );
    if (!strip)
	return(NULL);

    XGetWindowAttributes(display, window, &windowAttributes);
    w = windowAttributes.width;
    h = windowAttributes.height;

    /* 
     * initialize some useful fields
     */
    strip->display = display;
    strip->screen = screen;
    strip->window = window;

    /* 
     * description of extent of strip in Pixel coordinates...
     */
    strip->w      = w;
    strip->h      = h;

    strip->dataX0 =  STRIP_MARGIN*w;
    strip->dataY0 =  STRIP_MARGIN*h;
    strip->dataX1 =  (1.0 - STRIP_MARGIN)*w;
    strip->dataY1 =  (1.0 - STRIP_MARGIN)*h;
    strip->dataWidth = strip->dataX1 - strip->dataX0;
    strip->dataHeight = strip->dataY1 - strip->dataY0;


    /*
     * create the pixmaps
     */
    strip->pixmap = XCreatePixmap(strip->display,strip->window,strip->w,strip->h,
    			DefaultDepth(strip->display,strip->screen));
    strip->dataPixmap = XCreatePixmap(strip->display,strip->window,
			strip->dataWidth,strip->dataHeight,
			DefaultDepth(strip->display,strip->screen));


    /* 
     * create a default, writable GC
     */
    strip->gc = XCreateGC(strip->display,DefaultRootWindow(strip->display),
			0L,(XGCValues *)NULL);

    /*
     * and clear the empty fields
     */

#ifdef GTA
#else
    strip->appContext = (XtAppContext) NULL; 
    /* use default AppContext unless user specifies one */
#endif
    strip->dataBuffer = NULL;
    strip->titleFont = NULL;
    strip->axesFont = NULL;
    strip->title = NULL;
    strip->xAxisLabel = NULL;
    strip->yAxisLabel = NULL;
    strip->bufferPtr = NULL;
    strip->nValues = NULL;
    strip->dataPixel = NULL;
#ifdef OLD_STRIP
    strip->getValue = NULL;
#else
    strip->caStrip = NULL;
#endif OLD_STRIP

    for (i=0; i<MAXCURVES; i++) {
	strip->valueRange[i].minVal = 0.0;
	strip->valueRange[i].maxVal = 10.0;
    }


    /*
     * and initialize the Legend fields
     */
    strip->legend = FALSE;
    strip->legendTitle = NULL;
    strip->legendArray = NULL;

    strip->setOwnColors = FALSE;    /* will user fill in own pixel values? */
    strip->setOwnFonts = FALSE;    /* will user fill in own font values? */

    return(strip);

}


#ifdef GTA

/*
 * GTA doesn't want any Xt level things in here, so the
 *   stripSetAppContext() routine is deleted
 */

#else

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Set the application context to use for the Strip
 * function:  void stripSetAppContext()
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ 

void stripSetAppContext(strip,appContext)
  Strip *strip;
  XtAppContext appContext;
{

    /*
     * update the  application context  field
     */
    strip->appContext = appContext;

    XtRemoveTimeOut(strip->timeoutId);
    strip->timeoutId = XtAppAddTimeOut( strip->appContext,
	max(0,(int)(1000.0*(strip->samplingInterval - strip->adjustedInterval))),
	updateStrip,strip);

}


#endif


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Set/Reset the Interval (Timeout) for the Strip
 * function:  void stripSetInterval()
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ 

void stripSetInterval(strip,newInterval)
  Strip *strip;
  double newInterval;
{
#ifdef GTA
    struct timeval timeval;
    int secs, mSecs, uSecs;
#endif

    /*
     * update the sampling Interval field
     */
    strip->samplingInterval  = newInterval;

    /*
     * register the timeout to reflect the current sampling interval
     */
#ifdef GTA

    fdmgr_clear_timeout(strip->pfdctx,strip->palarm);

    mSecs = max(0,(int)(1000.0*(strip->samplingInterval - 
			strip->adjustedInterval)));

    timeval.tv_sec = mSecs/1000;			/* seconds */
    timeval.tv_usec = (mSecs%1000)*1000;		/* micro-seconds */
    strip->palarm = fdmgr_add_timeout(strip->pfdctx,&timeval,
					updateStrip,strip);

#else
    XtRemoveTimeOut(strip->timeoutId);
 
    if (strip->appContext == NULL) {
	strip->timeoutId = XtAddTimeOut( 
			max(0,(int)(1000.0*(strip->samplingInterval - strip->adjustedInterval))),
			updateStrip,strip);
    } else {
	strip->timeoutId = XtAppAddTimeOut( strip->appContext,
			max(0,(int)(1000.0*(strip->samplingInterval - strip->adjustedInterval))),
			updateStrip,strip);
    }

#endif

}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Set/Reset the data color for the Strip
 * function:  void stripSetDataColor()
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ 

void stripSetDataColor(strip,channelNumber,colorString)
  Strip *strip;
  int channelNumber;
  char *colorString;
{
    XColor color, ignore;

    if (channelNumber > strip->nChannels-1 || channelNumber < 0) {
	fprintf(stderr,"~\nstripSetDataColor:  invalid channel number %d", 
		channelNumber);
	return;
    }

    if(!XAllocNamedColor(strip->display,
	DefaultColormap(strip->display,strip->screen),
        colorString,&color,&ignore)) {
        fprintf(stderr, "~\nstripSetDataColor:couldn't allocate color %s",
		colorString);
        strip->dataPixel[channelNumber]  = WhitePixel(strip->display,
		strip->screen);
    } else {
        strip->dataPixel[channelNumber] = color.pixel;
    }

}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Terminate the Strip
 * function:  void stripTerm()
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ 

void stripTerm(strip)
  Strip *strip;
{
    int i;

    /***
     *** free the space used and referenced by the strip chart
     ***/

    /*
     * free the pixmaps
     */
    XFreePixmap(strip->display,strip->pixmap);
    XFreePixmap(strip->display,strip->dataPixmap);

    /* likewise, free the graphics context and font structures */
    XFreeGC(strip->display,strip->gc);
    XFreeFont(strip->display,strip->titleFontStruct);
#ifndef AXP
    if((int)(strip->axesFontStruct) != (int)(strip->titleFontStruct)){
#else
    if(strip->axesFontStruct != strip->titleFontStruct){
#endif
	XFreeFont(strip->display,strip->axesFontStruct);
    }

    /* and finally free the strip chart space itself */

    if (strip->storageMode == StripInternal) {
	for (i=0; i<strip->nChannels; i++)
	    free( (char *) strip->dataBuffer[i]);
	free( (char *) strip->dataBuffer);
	free( (char *) strip->titleFont);
	free( (char *) strip->axesFont);
	free( (char *) strip->title);
	free( (char *) strip->xAxisLabel);
	free( (char *) strip->yAxisLabel);
    }

    free( (char *) strip->bufferPtr);
    free( (char *) strip->nValues);
    free( (char *) strip->dataPixel);
#ifdef OLD_STRIP
    free( (char *) strip->getValue);
#else
    free( (char *) strip->caStrip);
#endif OLD_STRIP

    /* free the legend space */
    if (strip->legend) {
	free( (char *) strip->legendTitle);
	for (i=0; i<strip->nChannels; i++)
	    free( (char *) strip->legendArray[i]);
	free( (char *) strip->legendArray);
	XDestroyWindow(strip->display, strip->legendWindow);
	strip->legend = FALSE;
    }

    free( (char *) strip);

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Resize the Strip (e.g., change sizes: free old pixmaps, create new ones)
 * function: void stripResize()
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ 

void stripResize(strip)
  Strip  *strip;
{
    XWindowAttributes windowAttributes;

    /*
     * check to see if window has changed size
     *   if so, then Free the pixmaps, fetch new ones, and continue
     */
    XGetWindowAttributes(strip->display, strip->window, &windowAttributes);

    if (strip->w != windowAttributes.width
      || strip->h != windowAttributes.height){

	XSetForeground(strip->display, strip->gc, strip->backPixel);
	strip->w = windowAttributes.width;
	strip->h = windowAttributes.height;
	strip->dataX0 = STRIP_MARGIN*windowAttributes.width;
	strip->dataY0 = STRIP_MARGIN*windowAttributes.height;
	strip->dataX1 = (1.0 - STRIP_MARGIN)*windowAttributes.width;
	strip->dataY1 = (1.0 - STRIP_MARGIN)*windowAttributes.height;
	strip->dataWidth = strip->dataX1 - strip->dataX0;
	strip->dataHeight = strip->dataY1 - strip->dataY0;

	if ( strip->pixmap )
	    XFreePixmap(strip->display,strip->pixmap);
	strip->pixmap = XCreatePixmap(strip->display,strip->window,
				      strip->w,strip->h,
				      DefaultDepth(strip->display,strip->screen));

	if ( strip->dataPixmap )
	    XFreePixmap(strip->display,strip->dataPixmap);
	strip->dataPixmap = XCreatePixmap(strip->display,strip->window,
					  strip->dataWidth,strip->dataHeight,
					  DefaultDepth(strip->display,strip->screen));

	XFillRectangle( strip->display, strip->dataPixmap, strip->gc,
			0, 0, strip->dataWidth, strip->dataHeight);
    }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Draw the Strip
 * function: void stripDraw()
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ 

void stripDraw(strip)
  Strip  *strip;
{

    int i, j, textWidth;
    double timeScale, valueScale[MAXCURVES];
    int timeScaledValue, valueScaledValue;
    double binWidth = (strip->dataWidth/((double)(strip->bufferSize)));
    int	numBins	= strip->dataWidth / binWidth;
    double timeValueScale, valueValueScale, timeInterval, valueInterval[MAXCURVES];
    double valueMin[MAXCURVES], valueMax[MAXCURVES], timeMin, timeMax;
    int tickPosition, xPosition, yPosition;
    int goodNumberOfTicks;
    char valueString[13];	/* a 12 char blank string */
#ifndef AXP
    long time_now;
    long timeValue;
#else
    time_t time_now;
    time_t timeValue;
#endif
    char *timeString;
    int px, py;
    int dx, dy;

    static XPoint pointArray[2048];	/* just need place to put line pts */
    XRectangle clipRect[2];

    int buf, usedSpots[(NTICKS+1)*MAXCURVES], usedCtr;
    int stringPosition, stringPositionX, stringPositionY;
    int used[MAXTICKS], numUsed[MAXTICKS];

    strcpy(valueString, "            ");	/* a 12 char blank string */


    XSetForeground(strip->display, strip->gc, strip->backPixel);
    XFillRectangle( strip->display, strip->pixmap, strip->gc, 
			0, 0, strip->w, strip->h);

    /* 
     * calculate Pixel to value scaling
     */

    /* 
     * perform linear scaling (get nice values) for Time and Value axes
     */

    for (i = 0; i < strip->nChannels; i++) {
	linear_scale(strip->valueRange[i].minVal, strip->valueRange[i].maxVal, NTICKS,
		     &valueMin[i], &valueMax[i], &valueInterval[i]);
    }

    if (strip->samplingInterval != 0.0) {
	linear_scale(-1.0*strip->samplingInterval * strip->bufferSize, 
		     0.0, NTICKS,
		     &timeMin, &timeMax, &timeInterval);
    } else {
	timeMin = 0.0;
	timeMax = (double) strip->bufferSize;
	timeInterval = (double) (strip->bufferSize / NTICKS);
    }


    /*
     * now put up the data 
     */

    XSetLineAttributes( strip->display, strip->gc,
			DATALINEWIDTH,LineSolid,CapButt,JoinMiter);
    timeScale = (STRIP_DRAWINGAREA * strip->w) / (timeMax - timeMin);

    for (i = 0; i < strip->nChannels; i++)
	valueScale[i] = (STRIP_DRAWINGAREA * strip->h)/(valueMax[i] - valueMin[i]);

    /* 
     *  draw no-value (HUGE) indicators from ca_strip buffers
     */
    for (j = 0; j < strip->nChannels; j++) {
	int x0, x1;

        for (i = 0; i < strip->bufferSize; i++)
        {
	    if (strip->dataBuffer[j][i] == HUGE)
	    {
		/* set initial point */
		x0 = (int)(i * binWidth + .5);
		while ((strip->dataBuffer[j][i] == HUGE) && i < strip->bufferSize)
		{
		    i++;
		}
		x1 = (int)(i * binWidth + .5);
		/* add plots */
		XSetForeground( strip->display, strip->gc, WhitePixel(strip->display, strip->screen));
		XFillRectangle ( strip->display, strip->pixmap, strip->gc,
				x0 + strip->dataX0, strip->dataY0, x1 - x0, strip->dataHeight );
	    }
	}
    }
    /* 
     *  draw data from ca_strip buffers
     */
    for (j = 0; j < strip->nChannels; j++) {
        int xix = 0;
	/* add plots */
	XSetForeground( strip->display, strip->gc, strip->dataPixel[j]);
        for (i = 0; i < strip->bufferSize; i++)
        {
	    if (xix == 0)
	    {
		if (strip->dataBuffer[j][i] != HUGE)
		{
		    /* set initial point */
		    px = (int)(i*binWidth + .5);
		    py = (int) ((valueMax[j] - strip->dataBuffer[j][i]) * valueScale[j] + .5);
		    pointArray [xix].x = px + strip->dataX0;
		    pointArray [xix].y = strip->dataY0 + py;
		    xix++;
		}
	    }
	    else
	    {
		if (strip->dataBuffer[j][i] == HUGE)
		{
		    XDrawLines ( strip->display, strip->pixmap , strip->gc,
			pointArray, xix, CoordModePrevious );
		    xix = 0;
		}
		else
		{
		    dx = (int)(i*binWidth + .5) - px;
		    dy = (int) ((valueMax[j] - strip->dataBuffer[j][i]) * valueScale[j] + .5) - py;
		    if ( dx || dy)
		    {
			px += dx;
			py += dy;
			pointArray [xix].x = dx;
			pointArray [xix].y = dy;
			xix++;
		    }
		}
	    }
	}
	/* add plots if any exist */
	if (xix)
	    XDrawLines ( strip->display, strip->pixmap , strip->gc,
			 pointArray, xix, CoordModePrevious );
    }

    /* 
     * Put up Value axis and tick marks
     */
    XSetLineAttributes( strip->display, strip->gc,
			BORDERLINEWIDTH,LineSolid,CapButt,JoinMiter);

    XSetForeground( strip->display, strip->gc, strip->forePixel);

    /* left side axis */
    XDrawLine(strip->display,strip->pixmap,strip->gc,
		strip->dataX0,
		strip->dataY0,
		strip->dataX0,
		strip->dataY1);
    /* right side axis */
    XDrawLine(strip->display,strip->pixmap,strip->gc,
		strip->dataX1,
		strip->dataY0,
		strip->dataX1,
		strip->dataY1);


    usedCtr = 0;
    for (i=0; i<MAXTICKS; i++)
    {
	used[i] = 0;
	numUsed[i] = 0;
    }

    for (buf = 0; buf < strip->nChannels; buf++)
    {
	for (valueValueScale=valueMin[buf]; valueValueScale<=valueMax[buf]; valueValueScale+=valueInterval[buf])
	{

	    XSetForeground(strip->display, strip->gc, strip->forePixel);
	    tickPosition = (int)((valueMax[buf]-valueValueScale)*valueScale[buf]) + strip->dataY0;

	    /* left side ticks (could do grids right around here...) */
	    XDrawLine(strip->display,strip->pixmap,strip->gc,
			strip->dataX0 - (int)(RELATIVETICKSIZE*(min(strip->w,strip->h))),
			tickPosition,
			strip->dataX0,
			tickPosition);

	    /* right side ticks */
	    XDrawLine(strip->display,strip->pixmap,strip->gc,
			strip->dataX1, 
			tickPosition,
			strip->dataX1 + (int)(RELATIVETICKSIZE*(min(strip->w,strip->h))),
			tickPosition);

	    /* 
	     * tick labelling (left side only)
	     *  (move label slightly if there's already a string in that area)
	     */
	    sprintf(valueString,"%.3g",valueValueScale);
	    textWidth = tisTextWidth(strip->axesFontStruct,valueString, strlen(valueString));
	    XSetFont(strip->display,strip->gc,strip->axesFontStruct->fid);
	    stringPosition = getOKStringPosition(tickPosition,usedSpots,&usedCtr, strip->axesFontHeight);
	    tisDrawString(strip->display,strip->pixmap,strip->gc,
	    strip->dataX0 - textWidth -(int)(RELATIVETICKSIZE*(min(strip->w,strip->h))),
	    stringPosition,
	    valueString,strlen(valueString));

	    /* put up a little colored circle next to the scale */
	    XSetForeground( strip->display, strip->gc, strip->dataPixel[buf]);
	    XFillArc(strip->display,strip->pixmap,strip->gc,
 		     (int)(STRIP_MARGIN*strip->w) - textWidth
 	               - (int)(RELATIVETICKSIZE*(min(strip->w,strip->h)))
 	               - MINBORDER - strip->axesFontHeight/3,
 		     stringPosition - strip->axesFontHeight/3,
 		     strip->axesFontHeight/3, strip->axesFontHeight/3, 0, 64*360);

	}
    }

    /* 
     * Put up independent variable axis (X) and tick marks
     */

    XSetForeground( strip->display, strip->gc, strip->forePixel);

    /* top line */
    XDrawLine(strip->display,strip->pixmap,strip->gc,
		strip->dataX0,
		strip->dataY0,
		strip->dataX1,
		strip->dataY0);

    /* bottom line */
    XDrawLine(strip->display,strip->pixmap,strip->gc,
		strip->dataX0,
		strip->dataY1,
		strip->dataX1,
		strip->dataY1);



    /*
     * now X axis labelling...
     */
    timeScale = strip->dataWidth / (timeMax - timeMin);
    time_now = time(0);
    for (timeValueScale=timeMin; timeValueScale<=timeMax; 
		timeValueScale+=timeInterval)
    {

	/* problem near zero, check if `close enough' */
	if (timeValueScale > -ALMOSTZERO && timeValueScale < ALMOSTZERO)
	    timeValueScale = 0.0;

	tickPosition = strip->dataX0 + (int)((timeValueScale - timeMin)*timeScale);

	/* bottom ticks (could do grids right around here...) */
	XDrawLine(strip->display,strip->pixmap,strip->gc,
		  tickPosition,
		  strip->dataY1,
		  tickPosition,
		  strip->dataY1 
			+ (int)(RELATIVETICKSIZE*(min(strip->h,strip->w))));

	/* top ticks */
	XDrawLine(strip->display,strip->pixmap,strip->gc,
		  tickPosition,
		  strip->dataY0,
		  tickPosition,
		  strip->dataY0 
		    - (int)(RELATIVETICKSIZE*(min(strip->h,strip->w))));

	/* 
	 * tick labelling (bottom only)
	 */
	timeValue = time_now + (numBins * strip->samplingInterval * (timeValueScale/(timeMax - timeMin)));
	timeString = ctime(&timeValue);
	timeString[19] = 0;
	textWidth = tisTextWidth(strip->axesFontStruct,&timeString[11],
			strlen(&timeString[11]));
	XSetFont(strip->display,strip->gc,strip->axesFontStruct->fid);
	if (timeValueScale == timeMin)
	    xPosition = tickPosition;
	else if (timeValueScale >= timeMax)
	    xPosition = tickPosition - textWidth;
	else
	    xPosition = tickPosition - textWidth/2;
	tisDrawString(strip->display,strip->pixmap,strip->gc,
		      xPosition,
		      strip->dataY1 
			   + (int)(RELATIVETICKSIZE*(min(strip->h,strip->w)))/2
			   + (strip->axesFontHeight),
		      &timeString[11],strlen(&timeString[11]));
    }

    /*
     * put down title
     */
    textWidth = tisTextWidth(strip->titleFontStruct,strip->title, strlen(strip->title));
    XSetFont(strip->display,strip->gc,strip->titleFontStruct->fid);
    tisDrawString(strip->display,strip->pixmap,strip->gc,
		(int)(strip->w - (1.1*textWidth)),
		(int)(1.1*strip->titleFontHeight),strip->title,strlen(strip->title));


    /*
     * label axes
     */
    XSetFont(strip->display,strip->gc,strip->axesFontStruct->fid);

    /*
     * put down x axis label
     */
    textWidth = tisTextWidth(strip->axesFontStruct,strip->xAxisLabel, strlen(strip->xAxisLabel));

    xPosition = strip->dataX1 + (strip->w - strip->dataX1)/2 - textWidth/2 - (int)(binWidth/2.0);
    xPosition = (xPosition+textWidth > strip->w - MINBORDER ? strip->w - MINBORDER - textWidth : xPosition);
    yPosition = strip->dataY1 + (int)(RELATIVETICKSIZE*(min(strip->h,strip->w)))/2 + (strip->axesFontHeight*2);
    tisDrawString(strip->display,strip->pixmap,strip->gc,
		xPosition,
		yPosition,
		strip->xAxisLabel,strlen(strip->xAxisLabel));

    /*
     * put down y axis label
     */
    textWidth = tisTextWidth(strip->axesFontStruct,strip->yAxisLabel, strlen(strip->yAxisLabel));

    xPosition = strip->dataX0/2 - textWidth/2;
    xPosition = (xPosition < MINBORDER ? MINBORDER : xPosition);
    yPosition = (strip->dataY0 - strip->axesFontHeight)/2  + strip->axesFontHeight/2;
    tisDrawString(strip->display,strip->pixmap,strip->gc,
		xPosition,
		yPosition,
		strip->yAxisLabel,strlen(strip->yAxisLabel));


    /*
     * from regular pixmap to window
     */
    XCopyArea(strip->display,strip->pixmap,strip->window,strip->gc,0,0,strip->w,strip->h,0,0);

    /*
     * and from regular pixmap to dataPixmap 
     */
    XCopyArea(strip->display,strip->pixmap,strip->dataPixmap,strip->gc,
		strip->dataX0, strip->dataY0, strip->dataWidth,strip->dataHeight,0,0);

    XFlush(strip->display);
}

/*
 * Refresh the display
 * (don't recalcuate, just remap the pixmap onto the window
 */

void stripRefresh(strip)
  Strip  *strip;
{
    XCopyArea(strip->display,strip->pixmap,strip->window,strip->gc,0,0,
		strip->w,strip->h,0,0);
    XFlush(strip->display);
}




/*
 *   Add a strip chart 
 *
 *   input parameters:
 *
 *   strip	- the strip chart to be described (as returned from stripInit())
 *   nChannels  - number of channels to be plotted (number of dataBuffers, etc)
 *   bufferSize - size of buffers
 *   dataBuffer - array of array of doubles: dataBuffer[nChannels][bufferSize]
 *   valueMin   - minimum Value scale
 *   valueMax   - maximum Value scale
 *   samplingInterval - sampling freq: how long to wait between updates (sec)
 *   getValue   - array of pointers to functions returning a double 
 *			(is called every samplingInterval period of time)
 *   OR caStrip   - array of pointers to current strip structure 
 *			(updated every samplingInterval period of time)
 *   title	- character string title
 *   titleFont	- character string title font
 *   xAxisLabel	- character string X axis label
 *   yAxisLabel	- character string Y axis label
 *   axesFont	- character string axes font
 *   foreColor  - character string foreground color
 *   backColor  - character string background color
 *   dataColor  - array of character strings:  data colors
 *   storageMode - if StripInternal, make a copy of the data; 
 *			otherwise (StripExternal), just copy the pointers
 */

void stripSet( strip, nChannels, bufferSize, dataBuffer,
			valueRange, samplingInterval, 
#ifdef OLD_STRIP
	getValue,
#else
	caStrip,
#endif
			title, titleFont, 
			xAxisLabel, yAxisLabel,
			axesFont, foreColor, backColor, dataColor, storageMode)
  Strip		*strip;
  int		 nChannels;
  int		 bufferSize;
  double	*dataBuffer[];
  StripRange	*valueRange;
  double	 samplingInterval;
#ifdef OLD_STRIP
  double	(*getValue[])();
#else
  strip_channel *caStrip[];
#endif OLD_STRIP
  char		*title, *titleFont, *xAxisLabel, *yAxisLabel, *axesFont;
  char		*foreColor, *backColor, *dataColor[];
  StripStorageMode storageMode;
{
    int i, j, k;
    double r, g, b;
    int stringLength = 0;
    XColor color, ignore;
    int fontCount;
    char **fontNames;
    XFontStruct *fontInfo;

#ifdef GTA
    struct timeval timeval;
    int mSecs;
#endif


    if (storageMode == StripExternal) {
	strip->storageMode = StripExternal;
	strip->dataBuffer = dataBuffer;
	strip->titleFont = titleFont;
	strip->axesFont = axesFont;
	strip->title = title;
	strip->xAxisLabel = xAxisLabel;
	strip->yAxisLabel = yAxisLabel;
    } else {
	strip->storageMode = StripInternal;
	if (strip->dataBuffer) {
            for (i = 0; i < nChannels; i++)
		free( (char *) strip->dataBuffer[i] );
	    free( (char *) strip->dataBuffer );
	}
	strip->dataBuffer = (double **) malloc((unsigned) nChannels 
				* sizeof( double *));
	for (i = 0; i < nChannels; i++)
	    strip->dataBuffer[i] = (double *)malloc((unsigned) ( bufferSize * sizeof(double)));

	if (strip->titleFont) {
	    free( (char *) strip->titleFont);
	}
	stringLength = strlen(titleFont) + 1;
	strip->titleFont = (char *) malloc(stringLength);
#ifdef SOLARIS
	memcpy( strip->titleFont, titleFont, stringLength);
#else
	bcopy( titleFont, strip->titleFont, stringLength);
#endif
	strip->titleFont[stringLength-1] = '\0';

	if (strip->axesFont) {
	    free( (char *) strip->axesFont);
	}
	stringLength = strlen(axesFont) + 1;
	strip->axesFont = (char *) malloc(stringLength);
#ifdef SOLARIS
	memcpy( strip->axesFont, axesFont, stringLength);
#else
	bcopy( axesFont, strip->axesFont, stringLength);
#endif
	strip->axesFont[stringLength-1] = '\0';

	if (strip->title) {
	    free( (char *) strip->title);
	}
	stringLength = strlen(title) + 1;
	strip->title = (char *) malloc(stringLength);
#ifdef SOLARIS
	memcpy( strip->title, title, stringLength);
#else
	bcopy( title, strip->title, stringLength);
#endif
	strip->title[stringLength-1] = '\0';

	if (strip->xAxisLabel) {
	    free( (char *) strip->xAxisLabel);
	}
	stringLength = strlen(xAxisLabel) + 1;
	strip->xAxisLabel = (char *) malloc(stringLength);
#ifdef SOLARIS
	memcpy( strip->xAxisLabel, xAxisLabel, stringLength);
#else
	bcopy( xAxisLabel, strip->xAxisLabel, stringLength);
#endif
	strip->xAxisLabel[stringLength-1] = '\0';

	if (strip->yAxisLabel) {
	    free( (char *) strip->yAxisLabel);
	}
	stringLength = strlen(yAxisLabel) + 1;
	strip->yAxisLabel = (char *) malloc(stringLength);
#ifdef SOLARIS
	memcpy( strip->yAxisLabel,yAxisLabel,  stringLength);
#else
	bcopy( yAxisLabel, strip->yAxisLabel, stringLength);
#endif
	strip->yAxisLabel[stringLength-1] = '\0';

    }
    strip->nChannels = nChannels;
    strip->bufferSize = bufferSize;

    if (strip->bufferPtr)
	free( (char *) strip->bufferPtr);
    strip->bufferPtr = (unsigned long *) malloc((unsigned) nChannels 
	* sizeof(unsigned long));

    if (strip->nValues)
	free( (char *) strip->nValues);
    strip->nValues = (int *) malloc((unsigned) nChannels * sizeof(int));

    if (strip->dataPixel)
	free( (char *) strip->dataPixel);
    strip->dataPixel = (unsigned long *) malloc((unsigned) nChannels * 
						sizeof(unsigned long));

#ifdef OLD_STRIP
    if (strip->getValue)
	free( (char *) strip->getValue);
    strip->getValue = (double (**)()) malloc((unsigned) nChannels 
					* sizeof( double (**)() ) );
#else
    if (strip->caStrip)
	free( (char *) strip->caStrip);
    strip->caStrip = (strip_channel **) malloc((unsigned) nChannels 
					* sizeof( strip_channel ** ) );
#endif OLD_STRIP
    for (i = 0; i< nChannels; i++) {
	strip->bufferPtr[i] = 0;
	strip->nValues[i]   = 0;
#ifdef OLD_STRIP
	strip->getValue[i]  = getValue[i];
#else
	strip->caStrip[i]  = caStrip[i];
#endif OLD_STRIP

	strip->valueRange[i].minVal = valueRange[i].minVal;
	strip->valueRange[i].maxVal = valueRange[i].maxVal;
    }

    strip->samplingInterval = samplingInterval;
    strip->adjustedInterval = 0.0;



    /* set foreground, background and data colors based on passed-in strings */

    /* use a color for the foreground if possible */
    if ((DisplayCells(strip->display,strip->screen)>2)&&(!(strip->setOwnColors))) {
	if(!XAllocNamedColor(strip->display,
			     DefaultColormap(strip->display,strip->screen),
			     foreColor,&color,&ignore))
	{
	    fprintf(stderr,"~\nstripSet:couldn't allocate color %s",foreColor);
	    strip->forePixel  = WhitePixel(strip->display,strip->screen);
	} else {
	    strip->forePixel = color.pixel;
	}

	if(!XAllocNamedColor(strip->display,
			     DefaultColormap(strip->display,strip->screen),
			     backColor,&color,&ignore))
	{
	    fprintf(stderr,"~\nstripSet:couldn't allocate color %s",backColor);
	    strip->backPixel  = BlackPixel(strip->display,strip->screen);
	} else {
	    strip->backPixel = color.pixel;
	}

	/* if dataColor pointer is NULL, then do own color allocation */
	if (dataColor == NULL)
	{
	    if (XAllocColorCells(strip->display,
                DefaultColormap(strip->display,strip->screen), FALSE,
                NULL, 0, strip->dataPixel,nChannels))
	    {

		for (k = 0; k < nChannels; k++) {
		    /* (MDA) - rainbow( (double)(k+1)/(double)nChannels, 1.0, 1.0, &r, &g, &b); */
		    r = ((double) random())/pow(2.0,31.0);
		    g = ((double) random())/pow(2.0,31.0);
		    b = ((double) random())/pow(2.0,31.0);
		    r *= 65535; g *= 65535; b *= 65535;
		    color.red = (unsigned int) r;
		    color.green = (unsigned int) g;
		    color.blue = (unsigned int) b;
		    color.pixel = strip->dataPixel[k];
		    color.flags = DoRed | DoGreen | DoBlue;
		    XStoreColor(strip->display,
		    DefaultColormap(strip->display,strip->screen), &color);
		}

	    } else {
		fprintf(stderr,"~\nstripSet: couldn't allocate color cells");
	    }
	} else {
	    for (i = 0; i < nChannels; i++)
	    {
		if(!XAllocNamedColor(strip->display,
				     DefaultColormap(strip->display,strip->screen),
				     dataColor[i],&color,&ignore))
		{
		    fprintf(stderr,"~\nstripSet:couldn't allocate color %s",dataColor);
		    strip->dataPixel[i]  = WhitePixel(strip->display,strip->screen);
		} else {
		    strip->dataPixel[i] = color.pixel;
		}
	    }
	}

    } else {
	strip->forePixel  = WhitePixel(strip->display,strip->screen);
	strip->backPixel  = BlackPixel(strip->display,strip->screen);
	for (i = 0; i < nChannels; i++)
	    strip->dataPixel[i]  = WhitePixel(strip->display,strip->screen);
    }


    /*
     * load the fonts
     */
    if (!strip->setOwnFonts)
    {
	if ((strip->titleFontStruct=tisLoadQueryFont(strip->display,strip->titleFont)) 
						== NULL)
	{
	    fprintf(stderr,"~display %s doesn't know font %s\n",
		    DisplayString(strip->display),strip->titleFont);
	    fprintf(stderr,"~\t using first available font\n");
	    fontNames = XListFontsWithInfo(strip->display,FONTWILDCARD,
					MAXFONTS,&fontCount,&fontInfo);
	    strip->titleFontStruct = tisLoadQueryFont(strip->display,fontNames[0]);
	    XFreeFontInfo(fontNames,fontInfo,fontCount); /*XFreeFontNames(fontNames);*/
	}
	strip->titleFontHeight = strip->titleFontStruct->ascent + 
				strip->titleFontStruct->descent;
	strip->titleFontWidth = strip->titleFontStruct->max_bounds.rbearing - 
				strip->titleFontStruct->min_bounds.lbearing;

	if ((strip->axesFontStruct=tisLoadQueryFont(strip->display,strip->axesFont)) 
						== NULL)
	{
	    fprintf(stderr,"~stripSet: display %s doesn't know font %s\n",
		    DisplayString(strip->display),strip->axesFont);
	    fprintf(stderr,"~\t using first available font\n");
	    strip->axesFontStruct = strip->titleFontStruct;
	}
	strip->axesFontHeight = strip->axesFontStruct->ascent + 
				strip->axesFontStruct->descent;
	strip->axesFontWidth = strip->axesFontStruct->max_bounds.rbearing - 
				strip->axesFontStruct->min_bounds.lbearing;
    }

    /* 
     * initialize (to background colour) the two pixmaps
     */
    XSetForeground(strip->display, strip->gc, strip->backPixel);

    XFillRectangle( strip->display, strip->pixmap, strip->gc, 
			0, 0, strip->w, strip->h);
    XFillRectangle( strip->display, strip->dataPixmap, strip->gc, 
			0, 0, strip->dataWidth, strip->dataHeight);


    /***
     *** IMPORTANT:  the Intrinsics must have been initialized for the
     ***	Strip Chart to work, since it relies on TimeOut's
     ***/

    /*
     * register the initial (unadjusted) TimeOut here
     */

#ifdef GTA

    mSecs = (int) (1000.0 * strip->samplingInterval);

    timeval.tv_sec = mSecs/1000;			/* seconds */
    timeval.tv_usec = (mSecs%1000)*1000;		/* micro-seconds */
    strip->palarm = fdmgr_add_timeout(strip->pfdctx,&timeval,
				      updateStrip,strip);

#else

    if (strip->appContext == NULL) {
	strip->timeoutId = XtAddTimeOut(
				(int) (1000.0 * strip->samplingInterval),
				updateStrip, strip );
    } else {
	strip->timeoutId = XtAppAddTimeOut(strip->appContext,
				(int) (1000.0 * strip->samplingInterval),
				updateStrip, strip );
}

#endif

    strip->paused = FALSE;

}


/*
 *   Set the  value axis display range of a Strip chart plot
 *
 *   input parameters:
 *
 *   strip       - pointer to Strip data structure
 *   curveNum	 - curve or data set number
 *   minVal      - minimum display value
 *   maxVal      - maximum display value
 */

void stripSetRange(strip,curveNum,minVal,maxVal)
  Strip *strip;
  double minVal,maxVal;
{
     if (curveNum < 0 || curveNum >= strip->nChannels) {
        fprintf(stderr,
         "~\nstripSetRange: illegal curve (number %d) specified\n",
		curveNum);
        return;
     }

     if (minVal >= maxVal) {
        fprintf(stderr,
         "~\nstripSetRange: improper range specified, using defaults\n");
        return;
     }
     strip->valueRange[curveNum].minVal = minVal;
     strip->valueRange[curveNum].maxVal = maxVal;
}




/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * and here's the timeout routine (updates)
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifdef GTA
static void updateStrip(strip)
  Strip *strip;
#else
static void updateStrip(strip,timeoutId)
  Strip *strip;
  XtIntervalId *timeoutId;
#endif
{
    int i;
    double currentValue;
#ifdef GTA
    struct timeval timeval;
    int mSecs;
#endif

#ifndef TIMED
    /*
     * if the strip chart has been paused then simply return and don't reregister
     * the timeout...
     */
    if (strip->paused)
	return;

#else

    static time_t  startTime = 0;
    static time_t  endTime = 0;
    static struct tms  tms;

    static BOOLEAN takeTimes = TRUE; 

    if (strip->paused)
	return;

    /* we want the time spent in user and system space */
    if ((takeTimes) && (startTime == 0)) {
	times(&tms); 
	startTime = tms.tms_utime + tms.tms_stime;
    }
#endif

    for (i = 0; i < strip->nChannels; i++) {
	CAStripSync (strip->caStrip[i]);
    }

    stripDraw(strip);

    /*
     * and reregister the timeout 
     *
     */

#ifdef TIMED
    /*
     *(add some machine performance dependent stuff
     *   use an effective update interval to subtract out client times and
     *   more accurately sample at the specified frequency/interval
     */

    if ((takeTimes) && (startTime != 0)) {
	times(&tms); 
	endTime = tms.tms_utime + tms.tms_stime;
	strip->adjustedInterval = (double) (endTime - startTime);
	strip->adjustedInterval /= HZ;
	startTime = 0;
	takeTimes = FALSE;
    }
#endif


#ifdef GTA

    /*
     * using GTA fdmgr timeout out facility
     */
    mSecs = max(0,(int)(1000.0*(strip->samplingInterval - 
			strip->adjustedInterval)));
    timeval.tv_sec = mSecs/1000;			/* seconds */
    timeval.tv_usec = (mSecs%1000)*1000;		/* micro-seconds */
    strip->palarm = fdmgr_add_timeout(strip->pfdctx,&timeval,
				     updateStrip,strip);

#else

    /*
     * using Xt timeout facility: this function wants milliseconds
     */
    if (strip->appContext == NULL) {
	strip->timeoutId = XtAddTimeOut( 
			max(0,(int)(1000.0*(strip->samplingInterval - strip->adjustedInterval))),
			updateStrip,strip);
    } else {
	strip->timeoutId = XtAppAddTimeOut( strip->appContext,
			max(0,(int)(1000.0*(strip->samplingInterval - strip->adjustedInterval))),
			updateStrip,strip);
    }

#endif


}






/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * here's the Pause strip chart routine
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void stripPause(strip)
  Strip *strip;
{
  strip->paused = TRUE;
}




/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * and the Resume strip chart routine
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void stripResume(strip)
  Strip *strip;
{
 /* reregister the timeout */
#ifdef GTA
  struct timeval timeval;
  int mSecs;

  strip->paused = FALSE;
  mSecs = max(0,(int)(1000.0*(strip->samplingInterval - 
			strip->adjustedInterval)));

  timeval.tv_sec = mSecs/1000;			/* seconds */
  timeval.tv_usec = (mSecs%1000)*1000;		/* micro-seconds */
  strip->palarm = fdmgr_add_timeout(strip->pfdctx,&timeval,
       updateStrip,strip);

#else
strip->paused = FALSE;
if (strip->appContext == NULL) {
  strip->timeoutId = XtAddTimeOut( 
      max(0,(int)(1000.0*(strip->samplingInterval - strip->adjustedInterval))),
      updateStrip,strip);
} else {
  strip->timeoutId = XtAppAddTimeOut( strip->appContext,
      max(0,(int)(1000.0*(strip->samplingInterval - strip->adjustedInterval))),
      updateStrip,strip);
}
#endif
  
}



/*
 *   Print (dump in xwd format, then pipe to `xpr -device ps' then pipe
 *   to printer defined by PSPRINTER environment variable...)
 *
 *   input parameters:
 *
 *   strip        - pointer to Strip data structure
 */
void stripPrint(strip)
  Strip *strip;
{

  utilPrint(strip->display, strip->window, XWD_TEMP_FILE_STRING);

}




/*
 *   Set the legend parameters
 *
 *   input parameters:
 *
 *   strip        - pointer to Strip data structure
 *   legendTitle - title string
 *   legendArray - pointer to array of legend entry strings
 */
void stripSetLegend(strip, legendTitle, legendArray)
  Strip *strip;
  char *legendTitle;
  char **legendArray;
{
  int i, legendWidth = 0, legendHeight= 0;
  XWindowAttributes parentWindowAttributes;
  XSetWindowAttributes windowAttributes;


  if (strip->legend) XDestroyWindow(strip->display,strip->legendWindow);

  strip->legend = TRUE;

  if (strip->legendTitle != NULL) free( (char *) strip->legendTitle);
  strip->legendTitle = malloc(strlen(legendTitle)+1);
  strcpy(strip->legendTitle,legendTitle);

  if (strip->legendArray != NULL) {
     for (i=0; i<strip->nChannels; i++) free( (char *) strip->legendArray[i]);
     free( (char *) strip->legendArray);
  }
  strip->legendArray = (char **) malloc(strip->nChannels*sizeof(char *));
  for (i=0; i<strip->nChannels; i++){
     strip->legendArray[i] = malloc(strlen(legendArray[i])+1);
     strcpy(strip->legendArray[i],legendArray[i]);
  }


/*
 * loop to determine required size of legend window
 */
  legendWidth = tisTextWidth(strip->titleFontStruct,legendTitle,
                       strlen(legendTitle)) + 2*MINBORDER;
  for (i=0; i<strip->nChannels; i++) {
      legendWidth = max(legendWidth,
                        tisTextWidth(strip->axesFontStruct,legendArray[i],
                        strlen(legendArray[i]))+strip->axesFontHeight+
                        4*MINBORDER);

  }

  legendHeight = 2*MINBORDER + strip->nChannels*MINBORDER + MINBORDER +
                        strip->titleFontHeight +
                        strip->nChannels*strip->axesFontHeight;

/*
 * create a window butting up against strip window.
 * legendWindow is created as child of strip window.
 */
  strip->legendWindow = XCreateSimpleWindow(strip->display,
                                strip->window,
                                strip->w-legendWidth
                                        -LEGENDOFFSET-(2*BORDERLINEWIDTH),
                                LEGENDOFFSET,
                                legendWidth, legendHeight,
                                BORDERLINEWIDTH,
                                strip->forePixel, strip->backPixel);

/*
 * Set backing store attribute from parent window
 *  (Xt and XView differ on the default)
 */
  XGetWindowAttributes(strip->display, strip->window, &parentWindowAttributes);
  windowAttributes.backing_store = parentWindowAttributes.backing_store;
  XChangeWindowAttributes(strip->display, strip->legendWindow,
        CWBackingStore, &windowAttributes);

  XMapRaised(strip->display,strip->legendWindow);
 

}


/*
 *   Draw the legend
 *
 *   input parameters:
 *
 *   strip        - pointer to Strip data structure
 */
void stripDrawLegend(strip)
  Strip *strip;
{
  int i;

  if (!strip->legend) return;

/* title */
  XSetFont(strip->display,strip->gc,strip->titleFontStruct->fid);
  XSetForeground( strip->display, strip->gc, strip->forePixel);
  tisDrawString(strip->display,strip->legendWindow,strip->gc,
        MINBORDER, strip->titleFontHeight + MINBORDER,
        strip->legendTitle,strlen(strip->legendTitle));

/* legend color-key rectangles */
/* (use axesFontHeight as both Cell width and height */
  XSetFont(strip->display,strip->gc,strip->axesFontStruct->fid);
  for (i=0; i<strip->nChannels; i++) {
     XSetForeground( strip->display, strip->gc, strip->dataPixel[i]);
     XFillRectangle( strip->display, strip->legendWindow, strip->gc, MINBORDER,
        3*MINBORDER + strip->titleFontHeight +
           i*(strip->axesFontHeight+MINBORDER),
        strip->axesFontHeight, strip->axesFontHeight);
  }

/* legend strings */
  XSetForeground( strip->display, strip->gc, strip->forePixel);
  for (i=0; i<strip->nChannels; i++) {
     tisDrawString(strip->display,strip->legendWindow,strip->gc,
        strip->axesFontHeight + 2*MINBORDER,
        3*MINBORDER + strip->titleFontHeight + 2*strip->axesFontHeight/3 +
           i*(strip->axesFontHeight+MINBORDER),
        strip->legendArray[i],strlen(strip->legendArray[i]));
  }

  XRaiseWindow(strip->display,strip->legendWindow);

}


