/*plain_menus.c,v 1.1.1.1 1999/08/03 14:44:05 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	P L A I N _ M E N U S . C
 *	Author: 	Jeffrey O. Hill
 *			hill@luke.lanl.gov
 *			505 665 1831
 *
 *	Simple, fast, multilevel menu panes with an easy C interface
 *
 *
 *	History:
 *	.01 joh	05xx90	first pass
 *	.02 joh 010391	prevented menus from popping up off screen
 *	.03 joh 040991	made menu highlight easier to see
 *	.04 joh 040991	source clean up
 *	.05 joh 042591	better cascade symbol
 *	.06 joh 051391	streamlined menu popup/down
 *	.07 joh 081691	make sure grab is turned off while in their callback
 *	.08 joh 100991	use `CurrentTime' to avoid DECWINDOWS problems.
 *	.09 joh 022592	post_menu() and execute_menu() now allways
 *			return the number of times executed
 *
 *
 *	Notes:
 *	.01	The last operation performed is saved with the user's 
 *		"menu_configuration" so menus will allways popup
 *		with the last operation performed for that type
 *		of menu highlighted. A parallel current selection is
 *		saved in the "menu_in_use" structure reflecting
 *		the next operation to be performed if the operator
 *		releases the mouse button in a menu pane.
 *
 *	.02	popup/down grab sequence is somewhat complex, but 
 *		required to prevent unwanted button up events
 *		while the pointer is warped from its orig position
 *		and back. Be carefull changing the grab/warp/map
 *		sequence in this source.
 */

#include	<memLeak.h>
#include        <stdio.h>
#include        <Xlib.h>           
#include        <Xutil.h>

XFontStruct *tisLoadQueryFont ();

#ifdef AXP
#include <string.h>
#include <stdlib.h>
#endif

#include	<lstLib.h>

#define MENU_GLOBAL_SOURCE
#include	<plain_menus.h>

#if     !defined(NULL)
#       define NULL            0
#endif      
 
#if     !defined(FALSE) || (FALSE!=0)
#       define FALSE           0
#endif
 
#if     !defined(TRUE) || (TRUE!=1)
#       define TRUE            1
#endif
 
#if     !defined(OK) || (OK!=0)
#       define OK              0
#endif

#if     !defined(ERROR) || (ERROR!=(-1))
#       define ERROR           (-1)
#endif

#if     !defined(NELEMENTS)
#       define NELEMENTS(ARRAY) \
                (sizeof (ARRAY) / sizeof ((ARRAY) [0]))
#endif

#if     !defined(OFFSET)
#       define OFFSET(STRUCT_DECL, FIELD) \
                ((int) &(((STRUCT_DECL *) 0) -> FIELD))
#endif

#if     !defined(max)
#       define max(x, y)       (((x) < (y)) ? (y) : (x))
#endif

#if     !defined(min)
#       define min(x, y)       (((x) < (y)) ? (x) : (y))
#endif

/*
 * Even though this is really not correct I am doing it so 
 * that we can run on DECWINDOWS. Perhaps this can be removed
 * after the next DECWINDOWS release.
 */
#ifdef DONT_RUN_ON_DECWINDOWS
#	define MENU_TIME (prom->event_time)
#else
#	define MENU_TIME (CurrentTime)
#endif


/*
 *	courier makes the menu panes long, skinny, and hard to 
 *	traverse to subwindows?
 */
static char	*plain_menu_prefered_fonts[] = {
		"-*-*-bold-r-*--*-120-*-*-c-*-iso8859-1*",
		"-*-*-bold-r-*--*-160-*-*-c-*-iso8859-1*",
		"-*-*-bold-r-*--*-140-*-*-c-*-iso8859-1*",
		"-*-*-bold-r-*--*-*-*-*-c-*-iso8859-1*",
		"*--*-*-*-*-c-*-iso8859-1*",
		"fixed"};


#define MENU_BORDER_WIDTH 2		/* the X window border width */
#define PLAIN_MENU_MARGIN 4		/* additional padding within the window */
/* 
 *	n chars in the box in which the cascade sym is drawn
 */
#define NCHARS_CASCADE 1
/* 
 * 	n chars reserved for cascade sym 
 *	(including 1 char of leading space) 
 */
#define NCHARS_CASCADE_RESERVED (NCHARS_CASCADE +1)

/* 
 *	dynamic stuff for each menu window
 * 	one for each level (I reuse the window within a level)
 *
 */
struct menu_in_use{
	NODE				node;
	Window				w;
	/* the menu currently occupying window */
	struct menu_configuration	*pmc;	
	short				x_root;
	short				y_root;
	short				mapped;
	unsigned short			selection;
};

/* 
 *	root for each window tree
 *
 */
struct root_of_menu{
	LIST			window_list;
	Display			*d;
	struct menu_in_use	*last_mapped;
	GC 			foreground_GC;  	
	GC 			background_GC;  
	XFontStruct		*pfont;
	XEvent			*pevent;	/* first button down */
	void			*param;		/* per post parameter */
	Time			event_time;	/* time of last button event */
	unsigned short		execute_count;	/* n times executed */
	unsigned short		pane_ht;
	unsigned short		char_wd;
	unsigned short		char_ht;
	short			x_root_prior;
	short			y_root_prior;
	char			warped;		/* ptr not at orig pos */
};



/*
 *	BEWARE:
 *	You must also change event masks to move the
 *	the menu button.
 *
 */
#define MENUBUTTON 3

/*
 * When the pointer is outside the uppermost menu pane
 */
#define NO_SELECTION 0xffff

#define ANYWINDOW None 

static XContext menu_ctx;

static void 	create_menu_window();
static void 	motion_action();
static void 	save_current_selection();
void		wait_for_selection();
static		init_menu_window();
static		refresh_menu_window();
static int	consume_extra_button_clicks();
static int	execute_menu(); 
static		post_sub_menu();
static		popdown_action();
static		redraw_action();

/*
 *	highlight_pane()
 *
 *	(a macro)
 *
 */
#define highlight_pane(D,W,GC,X,Y,WIDTH,HEIGHT)\
XDrawRectangle(	(D),\
		(W),\
		(GC),\
		(X)+(PLAIN_MENU_MARGIN/2),\
		(Y)+(PLAIN_MENU_MARGIN/2),\
		(WIDTH)-PLAIN_MENU_MARGIN-1,\
		(HEIGHT)-PLAIN_MENU_MARGIN-1)



/*
 *
 *
 *	INIT_MENU
 *
 *
 *
 */
void init_menu(d, pmc)
Display				*d;
struct menu_configuration	*pmc;
{
	XGCValues 		GC_values; 
	struct root_of_menu 	*prom = NULL;
	unsigned 		i;


	if(!menu_ctx)
		menu_ctx = XUniqueContext();

	XFindContext(d, ANYWINDOW, menu_ctx, (caddr_t *)&prom);
	if(!prom){

		prom = (struct root_of_menu *) 
			calloc(1, sizeof(struct root_of_menu));
		if(!prom)
			abort();

		for(i=0; i< NELEMENTS(plain_menu_prefered_fonts); i++){
			prom->pfont = tisLoadQueryFont(
					d, 
					plain_menu_prefered_fonts[i]);
			if(prom->pfont)
				break;

			printf(	"~plain_menus: warning - could not find font: %s\n", 
				plain_menu_prefered_fonts[i]);
		}

		if(!prom->pfont){
			prom->pfont = XQueryFont(
					d, 
					(XID)DefaultGC(d, DefaultScreen(d)));
			if(!prom->pfont)
				exit(1);
			printf("~plain_menus: limping with the default font\n");
		}

		prom->char_wd = prom->pfont->max_bounds.rbearing - 
					prom->pfont->min_bounds.lbearing;
		prom->char_ht = prom->pfont->ascent + prom->pfont->descent;
		prom->pane_ht = prom->char_ht + PLAIN_MENU_MARGIN*2;

  		GC_values.foreground = WhitePixel(d, DefaultScreen(d));
  		GC_values.background = BlackPixel(d, DefaultScreen(d));
		GC_values.font = prom->pfont->fid;
  		prom->foreground_GC = 
			XCreateGC(	d, 
					DefaultRootWindow(d),
					GCForeground|GCBackground|GCFont,
					&GC_values);

  		GC_values.foreground = BlackPixel(d, DefaultScreen(d));
  		GC_values.background = WhitePixel(d, DefaultScreen(d));
		GC_values.font = prom->pfont->fid;
  		prom->background_GC = 
			XCreateGC(	d, 
					DefaultRootWindow(d),
					GCForeground|GCBackground|GCFont,
					&GC_values);
		prom->d = d;

		if(XSaveContext(d, ANYWINDOW, menu_ctx, (XPointer)prom))
			abort();
	}
	init_menu_window(	pmc, 
				prom, 
				0);	/* the base menu level */
}



/*
 *
 *
 *	INIT_MENU_WINDOW
 *
 *
 *
 */
static init_menu_window(pmc, prom, level)
struct menu_configuration	*pmc;
struct root_of_menu 		*prom;
int				level;
{
	char			**p;
	int			maxstringsize=0;
	char			**choices = pmc->pchoices;
	int			pane;

	for(p = choices; *p; p++)
		maxstringsize = max(maxstringsize, strlen(*p));

	pmc->nchoices = (unsigned short)(p-choices);
	/*
	 * extra chars of space here is reserved for the cascade symbol
	 */
	pmc->wd = (maxstringsize+NCHARS_CASCADE_RESERVED)*prom->char_wd + 
			PLAIN_MENU_MARGIN*2;

	pmc->ht = prom->pane_ht * pmc->nchoices;
	pmc->inactive_wd = pmc->wd - (pmc->wd>>2);
	pmc->selection = 0;

 	create_menu_window(prom, level);

	if(pmc->pactions){
		for(pane = 0; pane < (unsigned int)(pmc->nchoices); pane++){
			struct menu_configuration	*psubmc;

			psubmc = (struct menu_configuration *)
					pmc->pactions[pane].arg;

			if(pmc->pactions[pane].func == SUBMENUNEEDED){
				init_menu_window(	psubmc, 
							prom, 
							level+1);

				pmc->total_wd = max(	pmc->total_wd, 
							psubmc->total_wd);

				pmc->total_ht = max(	(unsigned int)(pmc->total_ht),
							(unsigned int)(psubmc->total_ht) +
								(unsigned int)(prom->pane_ht) * pane);

			}
		}
	}

	if(pmc->total_wd)
		pmc->total_wd += pmc->inactive_wd;
	else
		pmc->total_wd = pmc->wd + (MENU_BORDER_WIDTH<<2);

	pmc->total_ht = max((unsigned int)(pmc->total_ht), (unsigned int)(pmc->ht) + (MENU_BORDER_WIDTH<<2));

}



/*
 *
 *
 *	CREATE_MENU_WINDOW
 *
 *
 */
static void create_menu_window(prom, level)
struct root_of_menu 		*prom;
int				level;
{
	Window mw;
	XSetWindowAttributes 	xswa;
	unsigned 		size;
	struct menu_in_use 	*pmiu;
	Display			*d = prom->d;

	if(prom->window_list.count > level)
		return;

	pmiu = (struct menu_in_use *) calloc(1, sizeof(struct menu_in_use));
	if(!pmiu)
		abort();

	xswa.background_pixel = BlackPixel(d,DefaultScreen(d));
	xswa.border_pixel = WhitePixel(d,DefaultScreen(d));
	xswa.event_mask = 	ExposureMask|ButtonPressMask|
				ButtonReleaseMask|ButtonMotionMask;
	xswa.bit_gravity = NorthWestGravity;
	xswa.override_redirect = TRUE;
	xswa.save_under = TRUE;

	/*
	 * NOTE: X bombs if you create a zero width or height window
	 */
  	mw = XCreateWindow(
			d,
			DefaultRootWindow(d),
			0,	/* dont know x yet */
			0,	/* dont know y yet */
		     	10,	/* dont know ht yet */
			10,	/* dont know wd yet */
			MENU_BORDER_WIDTH,
			0,	/* take depth from parent */
			InputOutput,
			CopyFromParent,
			CWBackPixel|CWBorderPixel|CWEventMask|CWBitGravity|
				CWOverrideRedirect|CWSaveUnder,
			&xswa);

	pmiu->w = mw;

	lstAdd(&prom->window_list, (NODE *)pmiu);
	
	return;
}


/*
 *
 *
 *	POST_MENU
 *
 *	returns the number of times the menu was executed
 *
 */
int
post_menu(pevent, pmc, param)
XEvent				*pevent;
struct menu_configuration	*pmc;
void				*param;	/* user available */
{  	
	int 			status;
	struct menu_in_use 	*pmiu;
	int			x,y;
	struct root_of_menu 	*prom = NULL;
	struct menu_in_use 	*pnmiu;
#ifdef DEBUG
	printf("~slow button down consumed \n");
#endif
	if(pevent->type != ButtonPress)
		abort();

	XFindContext(pevent->xany.display, ANYWINDOW, menu_ctx, (caddr_t *)&prom);
	if(!prom)
		return 0;

	prom->pevent = pevent;
	prom->param = param;
	prom->execute_count = 0;
	prom->event_time = pevent->xbutton.time;

	/*
	 * save the init position of the pointer for popdown restore
	 */
	prom->x_root_prior = pevent->xbutton.x_root;
	prom->y_root_prior = pevent->xbutton.y_root;


	pmiu = (struct menu_in_use *) prom->window_list.node.next;

	/*
	 *	Force menus to come up on screen
	 *
	 */
	pmiu->x_root = 
		min(	
			pevent->xbutton.x_root,
			DisplayWidth(prom->d,DefaultScreen(prom->d))
				- (unsigned int)(pmc->total_wd));
	pmiu->y_root = 
		min(	
			pevent->xbutton.y_root, 
			DisplayHeight(prom->d,DefaultScreen(prom->d))
				- (unsigned int)(pmc->total_ht));
	pmiu->x_root = max(
			pmiu->x_root, 
			0);
	pmiu->y_root = max(
			pmiu->y_root, 
			0);

	/*
	 * init and map the menu windows
	 */
	pmiu->pmc = pmc;
	post_sub_menu(prom, pmiu);

	while(TRUE){

		if(!pmiu->pmc->pactions)
			break;
	
		if(pmiu->pmc->pactions[pmiu->selection].func 
			!= SUBMENUNEEDED)
			break;

		pnmiu = (struct menu_in_use *) pmiu->node.next;
		pnmiu->pmc = (struct menu_configuration *) 
			pmiu->pmc->pactions[pmiu->selection].arg;

		pnmiu->x_root = pmiu->x_root + 
				pmiu->pmc->inactive_wd;
		pnmiu->y_root = pmiu->y_root + 
				prom->pane_ht*pmiu->selection;

		post_sub_menu(prom, pnmiu);

		pmiu = pnmiu;
	}

	/*
	 * Move pointer onto current selection
	 */
	prom->warped = TRUE;
	XWarpPointer(	prom->d, 
			None, 
			pmiu->w,
			0,0,0,0,
			prom->char_wd,
			pmiu->selection * prom->pane_ht + 
				PLAIN_MENU_MARGIN+prom->pfont->ascent);

#ifdef DEBUG
	printf("~map warp sync at %x\n", prom->event_time);
#endif
	status = XGrabPointer(
		pevent->xany.display,
		pevent->xany.window,
		False,
		PLAIN_MENU_MIN_EVENT_MASK,
		GrabModeSync,
		GrabModeAsync,
		None,
		None,
		MENU_TIME);
	if(status != GrabSuccess){
		printf("~plain_menus: grab failed %d\n", status);
		popdown_action(prom);
		return 0;
	}

	/*
	 * Any button events in the queue prior to unthawing the grab
	 * are used without regard to their position since we suspect 
	 * they occured prior to the pointer warp
	 */
	status = consume_extra_button_clicks(
				prom, 
				FALSE);
	while(TRUE){
		if(status){
			wait_for_selection(prom);
		}
		status = consume_extra_button_clicks(
				prom, 
				TRUE);
		if(!status){
			break;
		}
	}

	/*
	 * ungrab exit
	 */
	XUngrabPointer(
		prom->d, 
		MENU_TIME);

	/*
	 * force the grab out to the server
	 * so they dont get hung in the debugger
	 * in their menu action routine.
	 */
	XFlush(prom->d);

	/*
	 * call their function only after
	 * the grab is off and the menu
	 * is unmapped
	 */
	status = execute_menu(prom);

#ifdef DEBUG
	printf("~ungrab at %x\n",prom->event_time);
	printf("----------------------\n");
#endif
	return (status);
}


/*
 *	WAIT_FOR_BUTTON_UP()
 */
void wait_for_selection(prom)
struct root_of_menu 	*prom;
{
	int status;

	if(!prom->last_mapped->mapped)
		return;

	/*
	 * Allow the server to process button events
	 */
#ifdef DEBUG
	printf("~selection unsync at %x\n", prom->event_time);
#endif
	XAllowEvents(
		prom->d,
		AsyncPointer,
		MENU_TIME);

	while(TRUE){	
		XEvent			event;
		struct menu_in_use	*plm;
		Window			w = prom->pevent->xany.window;

		XWindowEvent(
			prom->d, 
			w,
			ButtonReleaseMask|ButtonMotionMask,
			&event);

		plm = prom->last_mapped;

		switch(event.type){
		case ButtonRelease:
		{
#ifdef DEBUG
			printf("~slow button up consumed\n");
#endif
			if(event.xbutton.button != MENUBUTTON)
				continue;

			prom->event_time = event.xbutton.time;

                        /*
                         * only execute if they moused up in the menu or they
                         * released the button prior to mapping the menu
                         * (they want to quickly take the default)
                         */
                        if(plm->selection != NO_SELECTION){

                                /*
                                 * insure this type of menu pops up to
                                 * this menu pane next time
                                 */
                                save_current_selection(prom);
#if 0
				prom->execute_count++;
#else
				prom->execute_count=1;
#endif
                        }

        		popdown_action(prom);

#ifdef DEBUG
			printf("~popdown sync at %x\n", prom->event_time);
#endif
       			/*
        		 * stop the delivery of events until
        		 * the extra ones can be consumed
        		 * (extras are events which occur while ptr is moving
        		 * and the window is unmapping)
        		 */
       			 status = XGrabPointer(
               			 prom->d,
               			 prom->pevent->xany.window,
               			 False,
               			 PLAIN_MENU_MIN_EVENT_MASK,
               			 GrabModeSync,
               			 GrabModeAsync,
               			 None,
               			 None,
               			 MENU_TIME);
       			 if(status != GrabSuccess){
               			 abort();
       			 }

			return;
		}
		case MotionNotify:

			if(!(event.xmotion.state & Button3Mask))
				continue;

			if(!prom->warped)
				continue;

			prom->event_time = event.xmotion.time;

			motion_action(&event, prom, plm);

			break;
		default:
			abort();
			break;
		}
	}
}


/*
 *	CONSUME_EXTRA_BUTTON_CLICKS()
 *	returns TRUE if we are waiting for a button up
 *	else FALSE
 *
 */
static int consume_extra_button_clicks(prom, button_up)
struct root_of_menu 		*prom;
int				button_up;
{
	XEvent	event;
	int	status;
	Window	w;

	w = prom->pevent->xany.window;

	/*
	 * remove motion after button click and before
	 * pointer warp and map
	 *
	 * and
 	 *
	 * remove motion after button release and before 
	 * pointer unwarp and unmap
	 */
	while(TRUE){
		status = XCheckTypedWindowEvent(
				prom->d, 
				w, 
				MotionNotify, 
				&event);
		if(status == FALSE)
			break;
	}

	while(TRUE){

		if(button_up){
			status = XCheckTypedWindowEvent(
					prom->d, 
					w, 
					ButtonPress, 
					&event);
			if(status == TRUE && 
				event.xbutton.button == MENUBUTTON){
				prom->event_time = event.xbutton.time;
#ifdef DEBUG
				printf("~button down consumed\n");
#endif
				button_up = FALSE;
			}
			else{
				return FALSE;
			}
		}
		else{
			status = XCheckTypedWindowEvent(
					prom->d, 
					w, 
					ButtonRelease, 
					&event);
			if(status == TRUE && 
				event.xbutton.button == MENUBUTTON){
				prom->event_time = event.xbutton.time;
#ifdef DEBUG
				printf("~button up consumed\n");
#endif
				button_up = TRUE;

				prom->execute_count++;

        			/*
         			 * Allow extra event to be delivered up until
        			 * the map and unwarp is finished - see the
         		  	 * grab sync below (once the sync is on 
				 * we can identify these events
         			 * and consume them)
       				 */
				if(prom->last_mapped->mapped){
#ifdef DEBUG
					printf("~unsync at %d\n", prom->event_time);
#endif
       					XAllowEvents(
               					 prom->d,
               					 AsyncPointer,
               					 MENU_TIME);

       					popdown_action(prom);

#ifdef DEBUG
					printf("~sync at %d\n", prom->event_time);
#endif
     					 /*
       					  * stop the delivery of events until
       					  * the extra ones can be consumed
       					  * (extras are events which occur 
					  * while ptr is moving
       					  * and the window is unmapping)
       					  */
       					status = XGrabPointer(
               					 prom->d,
               					 prom->pevent->xany.window,
               					 False,
               					 PLAIN_MENU_MIN_EVENT_MASK,
               					 GrabModeSync,
              					 GrabModeAsync,
               					 None,
               					 None,
               					 MENU_TIME);
       					if(status != GrabSuccess){
                				abort();
        				}
				}
			}
			else{
				return TRUE;
			}
		}
	}
}


/*
 *
 *
 *	SAVE_CURRENT_SELECTION
 *
 *
 *
 */
static void save_current_selection(prom)
struct root_of_menu	*prom;
{
	struct menu_in_use 	*pmiu;

	for(	pmiu = (struct menu_in_use *) prom->window_list.node.next;
		pmiu;
		pmiu = (struct menu_in_use *) pmiu->node.next){

		pmiu->pmc->selection = pmiu->selection;
		if(pmiu == prom->last_mapped)
			break;
	}
}




/*
 *
 *
 *	EXECUTE_MENU
 *
 *	unroll the menu to the actual pane executed	
 *
 *	return the number of times executed
 */
static int 
execute_menu(prom) 
struct root_of_menu 		*prom;
{
	struct menu_configuration	*pmc = prom->last_mapped->pmc;
	struct jump_table 		*jt;
	int				i;

	/*
	 * Blank menu entry does not execute
	 */
	if(*pmc->pchoices[pmc->selection] == '\0')
		return 0;

	/*
	 * no jump table 
	 */
	if(!pmc->pactions){
		return prom->execute_count;
	}

	jt = &pmc->pactions[pmc->selection];

	while(jt->func == SUBMENUNEEDED){
		struct menu_configuration	*ptmpmc;

		ptmpmc = (struct menu_configuration *)jt->arg;
		jt = &ptmpmc->pactions[ptmpmc->selection];
	}

	if(jt->func){
		for(i=0; i < (unsigned int)(prom->execute_count); i++){
			(*jt->func)(
				jt->arg,
				prom->param,
				prom->pevent);
		}
		return prom->execute_count;
	}
	else{
		printf("~plain_menu: no action installed\n");
		return 0;
	}
}


/*
 *
 *
 *	POST_SUB_MENU
 *
 *
 *
 */
static post_sub_menu(prom, pmiu)
struct root_of_menu	*prom;
struct menu_in_use 	*pmiu;
{  	

	/*
	 * Move window under the pointer
 	 */
	XMoveResizeWindow(	prom->d, 
				pmiu->w, 
				pmiu->x_root,
				pmiu->y_root,
				pmiu->pmc->wd,
				pmiu->pmc->ht
				);

	XMapRaised(prom->d, pmiu->w);
	pmiu->mapped = TRUE;
	pmiu->selection = pmiu->pmc->selection;
	prom->last_mapped = pmiu;
	/*
	 * No need to wait for an expose or map notify
	 * since this is an overrride redirect window
	 */
	redraw_action(prom, pmiu);
}



/*
 *
 *
 *	POPDOWN_ACTION
 *
 *	Questions: why does this flicker the window manager if
 *	the mouseup is outside the menu ?
 *
 */
static popdown_action(prom)
struct root_of_menu	*prom;
{  	
	int			status;
	struct menu_in_use 	*pmiu = 
		(struct menu_in_use *) prom->window_list.node.next;
#ifdef DEBUG
	printf("~umwarp\n");
#endif
	/*
	 * Allways leave the pointer where they had it if leaving the root
	 * menu- do this before the pointer ungrab to avoid race
	 */
	if(prom->warped){
		XWarpPointer(
			prom->d, 
			None, 
			DefaultRootWindow(prom->d),
			0,0,0,0,
			prom->x_root_prior,
			prom->y_root_prior);
		prom->warped = FALSE;
	}

#ifdef DEBUG
	printf("~ummap\n");
#endif
	/*
	 * remove all menu windows
	 */
	for(	;
		pmiu;
		pmiu = (struct menu_in_use *) pmiu->node.next)
		if(pmiu->mapped){
			XUnmapWindow(	prom->d, 
					pmiu->w);
			pmiu->mapped = FALSE;
		}

	return;
}



/*
 *
 *
 *	MOTION_ACTION
 *
 *
 *
 */
static void motion_action(pevent, prom, pmiu)
XEvent			*pevent;
struct root_of_menu	*prom;
struct menu_in_use 	*pmiu;
{  	
	int	selection;
	int	x,y;

#ifdef DEBUG
	if(pevent->xany.type != MotionNotify)
		abort();
#endif
	x = pevent->xmotion.x_root - pmiu->x_root;
	y = pevent->xmotion.y_root - pmiu->y_root;

	if(x < 0){
		struct menu_in_use 	*punder;

		/*
		 *	Remove Uppermost Pane
		 *
		 */
		punder = (struct menu_in_use *) pmiu->node.previous;
	
		if(punder){
			XUnmapWindow(	prom->d, 
					pmiu->w);
			pmiu->mapped = FALSE;
			pmiu = prom->last_mapped = punder;
			motion_action(pevent, prom, pmiu);
			return;
		}
		else{
			selection = NO_SELECTION;
		}
	}
	else if(x > (unsigned int)(pmiu->pmc->wd)){
		selection = NO_SELECTION;
	}
	else{
		selection = (y/(unsigned int)(prom->pane_ht));
		if(selection <0 || selection >= (unsigned int)(pmiu->pmc->nchoices))
			selection = NO_SELECTION;
	}

	if(selection != pmiu->selection){
		int	box_x, box_y;

		box_x = 0;

		/*
		 *
		 * remove the current selection
		 *
		 */
		if(pmiu->selection != NO_SELECTION){
			box_y = prom->pane_ht*pmiu->selection;
			highlight_pane(	
				prom->d, 
				pmiu->w,
				prom->background_GC,
				box_x,
				box_y,
				pmiu->pmc->wd,
				prom->pane_ht);
		}

		/*
		 *
		 * add the new selection
		 *
		 */
		pmiu->selection = selection;
		if(pmiu->selection != NO_SELECTION){
			box_y = prom->pane_ht*pmiu->selection;
			highlight_pane(	
				prom->d, 
				pmiu->w,
				prom->foreground_GC,
				box_x,
				box_y,
				pmiu->pmc->wd,
				prom->pane_ht);
		}
	}

	if(!pmiu->pmc->pactions || selection == NO_SELECTION)
		return;

	if(pmiu->pmc->pactions[selection].func == SUBMENUNEEDED)
		if(x > pmiu->pmc->inactive_wd){
			struct menu_in_use *pnmiu = 
				(struct menu_in_use *) pmiu->node.next;

			pnmiu->pmc = (struct menu_configuration *) 
				pmiu->pmc->pactions[selection].arg;
			pnmiu->x_root = pmiu->x_root + 
				pmiu->pmc->inactive_wd;
			pnmiu->y_root = pmiu->y_root + 
				prom->pane_ht*selection;

			post_sub_menu(prom, pnmiu);
		}

	return;
}





/*
 *
 *
 *	REDRAW_ACTION
 *
 *
 *
 */
static redraw_action(prom, pmiu)
struct root_of_menu	*prom;
struct menu_in_use 	*pmiu;
{
	int	 	x,y;
	char		**choices = pmiu->pmc->pchoices;
	int		i;
	int 		cascade_wd;
	int		cascade_ht;
	/* array to draw menu cascade pointer */
	XPoint		sympoints[4];

	cascade_wd = (prom->pfont->max_bounds.rbearing - 
			prom->pfont->min_bounds.lbearing)*
			NCHARS_CASCADE;
	cascade_ht = prom->pfont->max_bounds.ascent + 
			prom->pfont->max_bounds.descent;

	sympoints[0].x = 	pmiu->pmc->wd - 
				PLAIN_MENU_MARGIN -
				cascade_wd;

	/* the following points are offsets */
	sympoints[1].x = cascade_wd;
	sympoints[1].y = (cascade_ht>>1);

	sympoints[2].x = -cascade_wd;
	sympoints[2].y = (cascade_ht>>1);

	sympoints[3].x = 0;
	sympoints[3].y = -cascade_ht;


	x = 0;
	y = 0;

	for(i=0; i< (unsigned int)(pmiu->pmc->nchoices); i++){

		tisDrawString(	prom->d, 
				pmiu->w,
				prom->foreground_GC,
				PLAIN_MENU_MARGIN,
				y + prom->pfont->ascent + PLAIN_MENU_MARGIN,
				*choices,
				strlen(*choices));	/* very inefficient */


		/*
		 * Draw The Cascade Symbol Next to the right Margin Allways
		 * one char large
		 */
		if(pmiu->pmc->pactions)
		  if(pmiu->pmc->pactions[i].func == SUBMENUNEEDED){
			sympoints[0].y = y + PLAIN_MENU_MARGIN;

			XFillPolygon(	prom->d, 
					pmiu->w,
					prom->foreground_GC,
					sympoints,
					NELEMENTS(sympoints),
					Convex,
					CoordModePrevious);
		}

		if(i == pmiu->selection){
			highlight_pane(	prom->d, 
					pmiu->w,
					prom->foreground_GC,
					0,
					y,
					pmiu->pmc->wd,
					prom->pane_ht);
		}

		y += prom->pane_ht;
		choices++;
	}

	return;
}



/*
 *
 *
 *	REFRESH_MENU
 *
 *	allows dynamic menu panes
 *
 */
void refresh_menu(d, pmc)
Display				*d;
struct menu_configuration	*pmc;
{
	struct root_of_menu 	*prom = NULL;

	if(!menu_ctx)	/* user must call init_menu() first */
		return;

	XFindContext(d, ANYWINDOW, menu_ctx, (caddr_t *)&prom);
	if(!prom)
		abort();

	pmc->total_wd = 0;
	pmc->total_ht = 0;
	refresh_menu_window( pmc, prom, 0);	/* the base menu level */
}



/*
 *
 *
 *	REFRESH_MENU_WINDOW
 *
 *
 *
 */
static refresh_menu_window(pmc, prom, level)
struct menu_configuration	*pmc;
struct root_of_menu 		*prom;
int				level;
{
	char	**p;
	int	maxstringsize	= 0;
	char	**choices	= pmc->pchoices;

	for(p = choices; *p; p++)
		maxstringsize = max(maxstringsize, strlen(*p));

	pmc->nchoices		= (unsigned short)(p - choices);

	/*
	 * extra chars of space here is reserved for the cascade symbol
	 */
	pmc->wd			= (maxstringsize+NCHARS_CASCADE_RESERVED)*prom->char_wd + PLAIN_MENU_MARGIN*2;
	pmc->ht			= prom->pane_ht * pmc->nchoices;
	pmc->inactive_wd	= pmc->wd - (pmc->wd>>2);
	pmc->selection		= 0;

	if(pmc->pactions){
		int	pane;

		for(pane = 0; pane < (unsigned int)(pmc->nchoices); pane++){
			struct menu_configuration	*psubmc;

			psubmc = (struct menu_configuration *)pmc->pactions[pane].arg;
			if(pmc->pactions[pane].func == SUBMENUNEEDED){
				psubmc->total_wd = 0;
				psubmc->total_ht = 0;
				refresh_menu_window(psubmc, prom, level+1);
				pmc->total_wd = max((unsigned int)(pmc->total_wd), psubmc->total_wd);
				pmc->total_ht = max((unsigned int)(pmc->total_ht), psubmc->total_ht + prom->pane_ht*pane);
			}
		}
	}

	if(pmc->total_wd)
		pmc->total_wd += pmc->inactive_wd;
	else
		pmc->total_wd = pmc->wd + (MENU_BORDER_WIDTH<<2);

	pmc->total_ht = max((unsigned int)(pmc->total_ht), pmc->ht + (MENU_BORDER_WIDTH<<2));
}
