/*color_map.c,v 1.1.1.1 1999/08/03 14:44:07 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 * 	C O L O R _ M A P . C
 * 	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 * 
 * routines for loading an OPI color map quickly into/from the X default 
 * color map and freeing allocated entries when done.
 * 
 * 
 * 
 * History
 * .01 joh	07-xx-90	written
 * .02 joh	09-14-90	moved to the OPI utility library for use 
 *				by OPI runtime and editor
 * 
 */

#include 	<fcntl.h>
#ifdef WIN32
#include 	<io.h>
#else
#include 	<file.h>
#endif WIN32
#include 	<X11/Xlib.h>
#ifdef WIN32
#include <X11/XlibXtra.h>
#endif WIN32
#include 	<X11/Xutil.h>
#include 	<X11/Xatom.h>
#ifdef AXP
#include	<stdlib.h>
#endif

#include 	<dldef.h>
#include 	<dlinfo.h>
#include	<color_map.h>
#include	<lstLib.h>

#ifdef SOLARIS
#define	bcopy(s,d,n)	memcpy (d, s, n)
#endif

#define 	MIDINTENSITY ((DL_MIDCOLOR)<<8)
void 		get_color();

int private_colors;

/*
 *
 *
 * 	COLOR_VISUAL_TEST
 * 	Is it a color screen
 *
 */
color_visual_test(d, s)
Display	*d;
int	s;
{

	/*
		I use the default visual since most servers can only 
		load one color map at a time which leads me to use
		the default color map which must be used with the
		default visual.
	*/
	switch(DefaultVisual(d, s)->class){
	    default:
	    case GrayScale:
	    case StaticGray:
		return FALSE;

	    case PseudoColor:
	    case DirectColor:
	    case TrueColor:
	    case StaticColor:
		return TRUE;
	}
}


/*
 *
 *
 * 	CREATE_CMAP
 * 
 *
 *	loads both a greyscale and a color map for portability
 *	across servers 
 *
 *	tries for a writeable color map first and settles for a 
 *	readonly map by calling XAllocColor() otherwise
 * 
 *
 */
struct color_map *create_cmap(d, pdlc, noldcolors, nnewcolors, do_inten, do_color)
Display				*d;
struct dl_color			*pdlc;
unsigned			noldcolors;
unsigned			nnewcolors;
int				do_inten;
int				do_color;
{
	unsigned 		ncolors = nnewcolors + noldcolors;
	struct color_map	*pcmap;
	unsigned		size;
	unsigned		i;
	unsigned long		*ptmp;

	if(!(do_inten || do_color))
		return NULL;

	size = 	sizeof(*pcmap) +
		( 	(do_inten ? sizeof(*pcmap->pcolor_pixels) : 0) +
			(do_color ? sizeof(*pcmap->pinten_pixels) : 0) +
			sizeof(*pcmap->pdl_colors) 	)*ncolors;

	pcmap = (struct color_map *) calloc(1, size);
	if(!pcmap)
		return NULL;

	pcmap->ncolors = ncolors;
	ptmp = (unsigned long *) (pcmap+1);

	/*
	 * Note: if both color and inten pixels are
	 * allocated the storage for there values must be 
	 * contiguous with color pixels first. This is because
	 * both color and intensity are allocated with
	 * one call to XAllocColorCells(). see below.
	 *
	 */
	if(do_color){
		pcmap->pcolor_pixels = ptmp;
		ptmp += ncolors;
	}
	else
		pcmap->pcolor_pixels = NULL;

	if(do_inten){
		pcmap->pinten_pixels = ptmp;
		ptmp += ncolors;
	}
	else
		pcmap->pinten_pixels = NULL;

	pcmap->pdl_colors = (struct dl_color *) (ptmp);

	if(ncolors == 0)
		return pcmap;

	for(i = 0; i < noldcolors; i++, pdlc++){
		pdlc->blink = (pdlc->blink == 1);
#ifdef WIN32
		memcpy(&pcmap->pdl_colors[i], pdlc, sizeof *pdlc);
#else
		bcopy(pdlc, &pcmap->pdl_colors[i], sizeof *pdlc);
#endif WIN32
	}
	for(i = 0, pdlc = pcmap->pdl_colors; i < ncolors; i++, pdlc++){
		XColor	xcolor;

		pdlc->blink = (pdlc->blink == 1);
/*
		xcolor.red	= pdlc->r << NBBY;
		xcolor.green	= pdlc->g << NBBY;
		xcolor.blue	= pdlc->b << NBBY;
*/
		xcolor.red	= (int)((pdlc->r/255.) * 65535);
		xcolor.green	= (int)((pdlc->g/255.) * 65535);
		xcolor.blue	= (int)((pdlc->b/255.) * 65535);
		if(do_color)
			get_color(d, pcmap, &xcolor, i, color_mapping, reuse_pixel);
/*
		xcolor.red	=
		xcolor.green	=
		xcolor.blue	= pdlc->inten << NBBY;
*/
		xcolor.red	=
		xcolor.green	=
		xcolor.blue	= (int)((pdlc->r/255.) * 65535);
		if(do_inten)
			get_color(d, pcmap, &xcolor, i, intensity_mapping, reuse_pixel);
	}
	return pcmap;
}


/*
 *
 *
 * 	GET_COLOR
 * 
 *
 */
typedef struct {
	NODE	node;	/* handle for list library	*/
	XColor	color;	/* the allocated XColor		*/
	int	blink;
	int	use_count;
} allocated_color;

static LIST	color_list;	/* current list of allocated colors */
static LIST	delete_list;	/* list of colors the get_color may delete before trying to allocate new ones */

void get_color(d, pcmap, pxcolor, index, mapping, reuse_option)
Display			*d;
struct color_map	*pcmap;
XColor			*pxcolor;
unsigned		index;
enum mapping 		mapping;
enum get_color_option	reuse_option;
{
	int		scrn = DefaultScreen(d);
	Status		status;
	unsigned long 	*ppixel;
	unsigned long 	plane_mask;
	static int	first_time = 1;
	XColor		alloc_xcolor;
	allocated_color	*p;
	int		private_color = FALSE;

	if(mapping == color_mapping){
		pxcolor->pixel	= pcmap->pcolor_pixels[index];
		ppixel		= &pcmap->pcolor_pixels[index];
	}
	else if(mapping == intensity_mapping){
		pxcolor->pixel	= pcmap->pinten_pixels[index];
		ppixel		= &pcmap->pinten_pixels[index];
	}

	if(DefaultDepth(d, scrn) == 1){
		/* dont bother asking the server if its monochrome */
		unsigned long 	average	= (int)(pxcolor->red + pxcolor->green +
						pxcolor->blue)/3;

		*ppixel = (average > MIDINTENSITY) ?
				WhitePixel(d, scrn) : BlackPixel(d, scrn);
		return;
	}

	if(first_time){
		first_time = 0;
		private_colors = FALSE;
		lstInit(&color_list);
		lstInit(&delete_list);
	}
	
	if (reuse_option == reuse_pixel)
		/*
		 * look for a match before resorting to allocating new colors
		 */
		for(p = (allocated_color *)lstFirst(&color_list); p;
					p = (allocated_color *)lstNext((struct node *)p)){
			if((pxcolor->red   == p->color.red) &&
			   (pxcolor->green == p->color.green) &&
			   (pxcolor->blue  == p->color.blue) &&
			   (pcmap->pdl_colors[index].blink == p->blink)){
				*ppixel = p->color.pixel;
				p->use_count++;
				return;
			}
		}

	/*
	 * delete any colors that are on the delete list
	 */
	for(p = (allocated_color *)lstFirst(&delete_list); p;
					p = (allocated_color *)lstNext((struct node *)p))
	    if (p->color.flags != 0) /* private color */
		XFreeColors(
			d, 
			DefaultColormap(d, DefaultScreen(d)),
			&p->color.pixel,
			1,
			None);
	lstFree(&delete_list);
	
	/*
	 * if it's not blinking, try to get a close matched read only color
	 */
	alloc_xcolor = *pxcolor;
	if (!(pcmap->pdl_colors[index].blink)
	  && XAllocColor (d, DefaultColormap(d, DefaultScreen(d)), &alloc_xcolor)
	  && alloc_xcolor.red == pxcolor->red
	  && alloc_xcolor.green == pxcolor->green
	  && alloc_xcolor.blue == pxcolor->blue) {
	    *ppixel = alloc_xcolor.pixel;
	} else {
	    printf ("Color not in the color palette -- trying to allocate in display.\n");
	    /*
	     * allocate private colors from the server and store the rgb values
	     */	
	    status = XAllocColorCells(	
		    d,
		    DefaultColormap(d, DefaultScreen(d)),
		    False,	/* non contig ok */
		    &plane_mask,
		    0,
		    ppixel,
		    1);
	    if (status) {
		private_colors = private_color = TRUE;
		pxcolor->flags	= DoRed|DoGreen|DoBlue;	
		pxcolor->pixel	= *ppixel;
		XStoreColor(
			d,
			DefaultColormap(d, DefaultScreen(d)),
			pxcolor);
		alloc_xcolor = *pxcolor;
	    } else {
		printf ("No private colors, using closest match, blink disabled.\n");
		/* pcmap->pdl_colors[index].blink = 0; */
		if (XAllocColor (d, DefaultColormap(d, DefaultScreen(d)), &alloc_xcolor)) {
		    *ppixel = alloc_xcolor.pixel;
		} else {
		    printf ("No color matches, quitting.\n");
		    exit (-1);
		}
	    }
	}

	/*
	 * link the color onto the list for future comparisons
	 */
	p = (allocated_color *)calloc(1, sizeof (allocated_color));
	p->color	= alloc_xcolor;
	if (private_color)
	    p->color.flags = 1;
	p->blink	= pcmap->pdl_colors[index].blink;
	p->use_count	= 1;
	lstAdd(&color_list, &p->node);
}




/*
 *
 *
 * 	FREE_COLOR
 * 
 *
 */
void free_color(d, pixel)
Display		*d;
unsigned long	pixel;
{	
	allocated_color	*p;

	for(p = (allocated_color *)lstFirst(&color_list); p; p = (allocated_color *)lstNext((struct node *)p))
		if(p->color.pixel == pixel)
			if(--(p->use_count) == 0){
				lstDelete(&color_list, (struct node *)p);
				lstAdd(&delete_list, (struct node *)p);
			}
}

/*
 *
 *
 * 	DELETE_CMAP
 * 
 *
 */
void delete_cmap(d, pcmap)
Display			*d;
struct color_map	*pcmap;
{	
	unsigned int	i;

	for(i = 0; i < pcmap->ncolors; i++){
		if(pcmap->pcolor_pixels)
			free_color(d, pcmap->pcolor_pixels[i]);
		if(pcmap->pinten_pixels)
			free_color(d, pcmap->pinten_pixels[i]);
	}
	free(pcmap);
}


/*
 *
 *	LOAD_CMAP_FROM_DISPLAY_FILE
 *
 */
struct color_map *load_cmap_from_display_file(d, pname, penvpath, is_editor)
Display	*d;
char	*pname;
char	*penvpath;
int	is_editor;
{
	int 			fd;
	struct dl_hdr		dl_hdr;
	int			status;
	struct color_map 	*pcmap = NULL; 
	char			path_used[64];
	int			color_flag;


	fd = open_display_list_file(
		penvpath, 
		path_used, 
		sizeof(path_used), 
		pname, 
		O_RDONLY);
	if(fd == ERROR){
		printf(	"External clr map display file access problem [%s]\n", 
			pname);
		return NULL;
	}

	/*
	 * Read in the display list header
	 */
#ifdef WIN32
	status = _read(fd, &dl_hdr, sizeof(dl_hdr));
#else
	status = read(fd, &dl_hdr, sizeof(dl_hdr));
#endif WIN32
	if(status == ERROR || dl_hdr.magic != DL_MAGIC || valid_dl_test(&dl_hdr)<0 ||
			(dl_hdr.nwords_cmap<<1) < sizeof(struct dl_color_map)){
		printf(	"Invalid color map display file ignored[%s/%s]\n", path_used, pname);
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		return NULL;
	}
	else {
		struct dl_color_map 	dl_cmap; 
		struct dl_color		*pdlc;
		unsigned		size;

#ifdef WIN32
		status = _lseek(fd, (dl_hdr.ocmap<<1), SEEK_SET);
#else
		status = lseek(fd, (dl_hdr.ocmap<<1), L_SET);
#endif WIN32
		if(status<0){
			printf(	"Invalid display list 0 [%s/%s]\n", path_used, pname);
#ifdef WIN32
			_close(fd);
#else
			close(fd);
#endif WIN32
			return NULL;
		}
#ifdef WIN32
		status = _read(fd, &dl_cmap, sizeof(dl_cmap));
#else
		status = read(fd, &dl_cmap, sizeof(dl_cmap));
#endif WIN32
		if(status<0){
			printf(	"Invalid display list 1 [%s/%s]\n", path_used, pname);
#ifdef WIN32
			_close(fd);
#else
			close(fd);
#endif WIN32
			return NULL;
		}

		
		if(dl_cmap.element.type != dl_color_map){
			printf(	"Display does not have a color map[%s/%s]\n", path_used, pname);
#ifdef WIN32
			_close(fd);
#else
			close(fd);
#endif WIN32
			return NULL;
		}

		size = (dl_cmap.element.nwords<<1) - sizeof(dl_cmap);
		pdlc = (struct dl_color *) calloc(1, size);

#ifdef WIN32
		status = _read(fd, pdlc, size);
#else
		status = read(fd, pdlc, size);
#endif WIN32
		if(status<0){
			printf(	"Invalid display list 2 [%s/%s]\n", path_used, pname);
			free(pdlc);
#ifdef WIN32
			_close(fd);
#else
			close(fd);
#endif WIN32
			return NULL;
		}
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32

		color_flag = color_visual_test(d, DefaultScreen(d));

		pcmap = create_cmap(	
			d, 
			pdlc, 
			dl_cmap.ncolors, 
			0,
			!color_flag || is_editor,
			color_flag  || is_editor);

		free(pdlc);
	}
#ifdef DEBUG
	if(pcmap)
		printf(	"Loaded color map from path: %s name: %s\n", path_used, pname);
#endif
	return pcmap;
}



/*
 *	DO_BLINK
 */
void	do_blink(d, pcmap, state)
Display			*d;
struct color_map	*pcmap;
int			state;
{
	unsigned int	divisor;
	unsigned int	index;
	unsigned int	did_any = FALSE;
#ifndef NeXT
	divisor = state ? 4 : 3;

	for(index = 0; index < pcmap->ncolors; index++){
		XColor		xcolor;
		struct dl_color	*pdlc = &pcmap->pdl_colors[index];

		xcolor.flags = DoRed|DoGreen|DoBlue;
		if(pdlc->blink == TRUE && private_colors){
			if(pcmap->pcolor_pixels){
				xcolor.red	= ((unsigned int)(pdlc->r) << NBBY) * 3 / divisor;
				xcolor.green	= ((unsigned int)(pdlc->g) << NBBY) * 3 / divisor;
				xcolor.blue	= ((unsigned int)(pdlc->b) << NBBY) * 3 / divisor;
				xcolor.pixel	= pcmap->pcolor_pixels[index];
				XStoreColor(d, DefaultColormap(d, DefaultScreen(d)), &xcolor);
			}
			if(pcmap->pinten_pixels){
				xcolor.red	=
				xcolor.green	=
				xcolor.blue	= ((unsigned int)(pdlc->inten) << NBBY) * 3 / divisor;
				xcolor.pixel	= pcmap->pinten_pixels[index];
				XStoreColor(d, DefaultColormap(d, DefaultScreen(d)), &xcolor);
			}
			did_any = TRUE;
		}
	}
	if(did_any)
		XFlush(d);
#endif
}
