/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 * 	C O L O R _ R U L E S . C
 * 	Author	Larry Blische (from color_map.c by: Jeffrey O. Hill)
 * 
 * routines for loading an OPI color rule table.
 */

#ifdef AXP
#include <stdlib.h>
#endif

#include 	<dldef.h>
#include 	<dlinfo.h>
#include	<color_rules.h>
#include	<string_table.h>
#ifdef WIN32
#include	<io.h>
#else
#include 	<file.h>
#endif WIN32
#include 	<fcntl.h>

extern void *tisCalloc();
extern void tisFree();

void	load_crule_strings_from_display_file();

/*
 *
 *
 * 	DELETE_CRULES
 * 
 *
 */
void delete_crules(pcrules)
struct color_rules	*pcrules;
	{	
		tisFree(pcrules);
		/*free(pcrules);*/
	}


/*
 * 	CREATE_CRULES
 */
struct color_rules *create_crules(pdlc, noldrules, nnewrules)
struct dl_color_rule	*pdlc;
unsigned				noldrules;
unsigned				nnewrules;
	{
	struct color_rules		*pcrules;
	struct dl_color_rule	*p;
	unsigned				i;
	unsigned 				nrules	= nnewrules + noldrules;
	unsigned				size	= sizeof *pcrules + (sizeof (struct dl_color_rule) * nrules);

	pcrules = (struct color_rules *) tisCalloc(__FILE__, __LINE__, 1, size);
					/*calloc(1, size);*/
	if (!pcrules)
		return NULL;

	pcrules->nrules = nrules;
	pcrules->pdl_crules = (struct dl_color_rule *) (pcrules+1);

	if (nrules == 0)
		return pcrules;

	for (i = 0, p = pdlc; i < noldrules; i++, p++)
#if defined(SOLARIS) || defined(WIN32)
		memcpy(&pcrules->pdl_crules[i], p, sizeof *p);
#else
		bcopy(p, &pcrules->pdl_crules[i], sizeof *p);
#endif
	return pcrules;
	}


/*
 *	LOAD_CRULE_FROM_DISPLAY_FILE
 */
struct color_rules *load_crules_from_display_file(pname, penvpath, pst)
char	*pname;
char	*penvpath;
LIST	*pst;
	{
	int 				fd;
	struct dl_hdr		dl_hdr;
	int					status;
	struct color_rules 	*pcrules = NULL; 
	char				path_used[64];

#ifdef WIN32
	fd = open_display_list_file(penvpath, path_used, sizeof(path_used), pname, _O_BINARY|_O_RDONLY);
#else
	fd = open_display_list_file(penvpath, path_used, sizeof(path_used), pname, O_RDONLY);
#endif WIN32
	if (fd == ERROR)
		{
		printf("External clr rules display file access problem [%s]\n", pname);
		return NULL;
		}

	/*
	 * Read in the display list header
	 */
#ifdef WIN32
	status = _read(fd, &dl_hdr, sizeof(dl_hdr));
#else
	status = read(fd, &dl_hdr, sizeof(dl_hdr));
#endif WIN32
	if (status == ERROR || dl_hdr.magic != DL_MAGIC || valid_dl_test(&dl_hdr) < 0 ||
	  (dl_hdr.nwords_crules<<1) < sizeof(struct dl_color_rule))
		{
		printf(	"Invalid color rules display file ignored[%s/%s]\n", path_used, pname);
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		return NULL;
		}
	else
		{
		struct dl_color_rule_table 	dl_crules; 
		struct dl_color_rule		*pdlc;
		unsigned					size;
		void						*pbuff;
		int							nwords;

#ifdef WIN32
		status = _lseek(fd, (dl_hdr.ocrules<<1), SEEK_SET);
#else
		status = lseek(fd, (dl_hdr.ocrules<<1), L_SET);
#endif WIN32
		if (status < 0)
			{
			printf("Invalid display list 0 [%s/%s]\n", path_used, pname);
#ifdef WIN32
			_close(fd);
#else
			close(fd);
#endif WIN32
			return NULL;
			}
#ifdef WIN32
		status = _read(fd, &dl_crules, sizeof(dl_crules));
#else
		status = read(fd, &dl_crules, sizeof(dl_crules));
#endif WIN32
		if (status < 0)
			{
			printf("Invalid display list 1 [%s/%s]\n", path_used, pname);
#ifdef WIN32
			_close(fd);
#else
			close(fd);
#endif WIN32
			return NULL;
			}
		if (dl_crules.element.type != dl_color_rule_table)
			{
			printf("Display does not have a color rule[%s/%s]\n", path_used, pname);
#ifdef WIN32
			_close(fd);
#else
			close(fd);
#endif WIN32
			return NULL;
			}
		size = (dl_crules.element.nwords<<1) - sizeof(dl_crules);
		pdlc = (struct dl_color_rule *) tisCalloc(__FILE__, __LINE__, 1, size);
						/*calloc(1, size);*/
		if (!pdlc)
			abort();
#ifdef WIN32
		status = _read(fd, pdlc, size);
#else
		status = read(fd, pdlc, size);
#endif WIN32
		if (status < 0)
			{
			printf("Invalid display list 2 [%s/%s]\n", path_used, pname);
			tisFree(pdlc);
			/*free(pdlc);*/
#ifdef WIN32
			_close(fd);
#else
			close(fd);
#endif WIN32
			return NULL;
			}
		nwords = dl_hdr.nwords_dspy - WDRND(sizeof(dl_hdr));
		pbuff = (void *) tisCalloc(__FILE__, __LINE__, 1, nwords << 1);
				/*calloc(1, nwords<<1);*/
		if (!pbuff)
			abort();

#ifdef WIN32
		status = _lseek(fd, sizeof(dl_hdr), SEEK_SET);
		if (_read(fd, pbuff, nwords<<1) < 0)
			abort();
#else
		status = lseek(fd, sizeof(dl_hdr), L_SET);
		if (read(fd, pbuff, nwords<<1) < 0)
			abort();
#endif WIN32

		load_crule_strings_from_display_file(pst, pbuff, nwords, &dl_hdr, pdlc, dl_crules.nrules);
		tisFree(pbuff);
		/*free(pbuff);*/
		pcrules = create_crules(pdlc, dl_crules.nrules, 0);
		tisFree(pdlc);
		/*free(pdlc);*/
		}
#ifdef DEBUG
	if (pcrules)
		printf("Loaded color rules from path: %s name: %s\n", path_used, pname);
#endif
#ifdef WIN32
	_close(fd);
#else
	close(fd);
#endif WIN32
	return pcrules;
	}


void load_crule_strings_from_display_file(pst, pbuff, nwords, pdl_hdr, pdlc, nrules)
LIST					*pst;
void					*pbuff;
int						nwords;
struct dl_hdr			*pdl_hdr;
struct dl_color_rule	*pdlc;
unsigned				nrules;
	{
	struct dl_anygo		*pgo;
	int					rule, line;
	LIST				foreignStringList;

	/*
	 * create a list of the strings in the foreign display
	 */
	stringInitList(&foreignStringList);
	for (pgo = (struct dl_anygo *)pbuff; nwords > 0; 
	  nwords -= pgo->element.nwords, pgo = (struct dl_anygo *)(((short *)pgo) + pgo->element.nwords))
		{
		if (pgo->element.type == dl_string)
			stringInsert(&foreignStringList, (char *)(((char *)pgo) + sizeof(struct dl_string)));
		}
	/*
	 * find strings in rule table and insert them into the display tables
	 */
	for (rule = 0; rule < nrules; rule++, pdlc++)
		{
#ifndef TOKENPTRS
		pdlc->name = stringInsert(pst, ((StringItem *)lstNth(&foreignStringList, *(int *)&(pdlc->name)+1))->string);
#else
		pdlc->name = dmTokenFromPointer64 (stringInsert(pst, ((StringItem *)lstNth(&foreignStringList, *(int *)&(pdlc->name)+1))->string));
#endif
		for (line = 0; line < MAX_RULE_LINES; line++)
			{
#ifndef TOKENPTRS
			pdlc->info[line].chan  = stringInsert(pst, ((StringItem *)lstNth(&foreignStringList, *(int *)&(pdlc->info[line].chan)+1))->string);
			pdlc->info[line].value = stringInsert(pst, ((StringItem *)lstNth(&foreignStringList, *(int *)&(pdlc->info[line].value)+1))->string);
#else
			pdlc->info[line].chan  = dmTokenFromPointer64 (stringInsert(pst, ((StringItem *)lstNth(&foreignStringList,
 *(int *)&(pdlc->info[line].chan)+1))->string));
			pdlc->info[line].value = dmTokenFromPointer64 (stringInsert(pst, ((StringItem *)lstNth(&foreignStringList,
 *(int *)&(pdlc->info[line].value)+1))->string));
#endif
			}
		}
	/*
	 * delete these strings to get rid of the storage space
	 */
	stringDeleteList(&foreignStringList);
	}
