/*dl_util.c,v 1.1.1.1 1999/08/03 14:44:07 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	O P I   D I S P L A Y    L I S T    U T I L I T I E S
 *
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	
 *	history
 *	.01 joh	 7-26-90	first cut
 *	.02 ajk 12-13-90	modified output format
 *	.03 joh 01-18-91	added arrays of structures
 *	.04 joh 03-22-91	added dump_display_strings()
 *
 *
 */
#ifndef WIN32
#include <memLeak.h>
#endif
#ifdef WIN32
#include <io.h>
#include <string.h>
#include <stdlib.h>
#endif
#include <fcntl.h>
#include <dldef.h>
#include <dlinfo.h>
#include <dlcvrt.h>
#include <string_table.h>
#include <stdio.h>
#ifdef AXP
#include <string.h>
#include <stdlib.h>
#endif

/*char	*calloc();*/
static dump_dl_structure();


/*
 *	DUMP_ELEMENT
 */
dump_element(pgo, pct, pst)
struct dl_anygo		*pgo;
LIST				*pct;
LIST				*pst;
	{
	dump_dl_structure(pgo, dl_elements_ginfo[pgo->element.type].sinfo, pct, pst, 0, dl_elements_einfo[pgo->element.type], stdout);
	return 0;
	}


/*
 * 	DUMP_DISPLAY  
 */
void dump_display(name, pdstfile, dump_name)
char		*name;
FILE		*pdstfile;
char		*dump_name;
	{
	int 				fd;
	int 				status;
	char				*filebuf;
	struct dl_anygo 	*pgo;
	int					nwords;
	struct dl_hdr		dl_hdr;
	char				*p;
	LIST				st;
	LIST				ct;

#ifdef WIN32
	fd = _open(name, _O_BINARY|_O_RDONLY);
#else
	fd = open(name, O_RDONLY);
#endif WIN32
	if (fd == ERROR)
		{
		fprintf(stderr, "file access problem: [%s]\n", name);
		return;
		}
	/*
	 * Read in the display list header
	 */
#ifdef WIN32
	status = _read(fd, &dl_hdr, sizeof(dl_hdr));
#else
	status = read(fd, &dl_hdr, sizeof(dl_hdr));
#endif WIN32
	if (status == ERROR)
		{
		fprintf(stderr, "file read problem ?\n");
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		return;
		}
	if (valid_dl_test(&dl_hdr) < 0)
		{
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		return;
		}
	nwords = dl_hdr.nwords_dspy - (WDRND(sizeof(dl_hdr)));
	filebuf = calloc(1, nwords<<1);
	if (!filebuf)
		{
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		fprintf(stderr, "file will not fit in memory\n");
		return;
		}
	/*
	 * Read in the display list
	 */
#ifdef WIN32
	status = _read(fd, filebuf, nwords<<1);
#else
	status = read(fd, filebuf, nwords<<1);
#endif WIN32
	if (status != (nwords<<1))
		{
		fprintf(stderr, "corrupt display file ?\n");
		fprintf(stderr, "tried %d, read %d\n", nwords*2, status);
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		return;
		}
	fixup_string_pointers(&dl_hdr, filebuf);
	
	/*
	 * set up the string tables
	 */
	stringInitList(&st);
	stringInitList(&ct);
	pgo = (struct dl_anygo *) filebuf;
	for (nwords = dl_hdr.nwords_dspy - (WDRND(sizeof(dl_hdr))); nwords > 0; 
	  nwords -= pgo->element.nwords, pgo = (struct dl_anygo *)(((short *) pgo) + pgo->element.nwords))
		{
		switch (pgo->element.type)
		  {
		  case dl_string:	stringInsert(&st, (char *)pgo + sizeof(struct dl_string));	break;
		  case dl_pc:		stringInsert(&ct, (char *)pgo + sizeof(struct dl_pc));		break;
		  }
		}

	/*
	 * strip off the path name
	 */
	for (p = dump_name; *p; p++)
		if(*p == '/')
			dump_name = p+1;

	fprintf(pdstfile, "file {\n");
	fprintf(pdstfile, "\tname=\"%s\"\n", dump_name);
	fprintf(pdstfile, "}\n");
	dump_dl_structure(&dl_hdr, dl_hdr_sinfo, &ct, &st, 0, "display", pdstfile);

	pgo = (struct dl_anygo *) filebuf;
	for (nwords = dl_hdr.nwords_dspy - (WDRND(sizeof(dl_hdr))); nwords > 0; 
	  nwords -= pgo->element.nwords, pgo = (struct dl_anygo *)(((short *) pgo) + pgo->element.nwords))
		{
		char	pstring[64];
		int		type	= pgo->element.type;

		if (type == dl_string || type == dl_pc)
			continue;

		sprintf(pstring, "\"%s\"", dl_elements_einfo[type]);
		dump_dl_structure(pgo, dl_elements_ginfo[type].sinfo, &ct, &st, 0, pstring, pdstfile);
		}
#ifdef WIN32
	_close(fd);
#else
	close(fd);
#endif WIN32
	free(filebuf);
	stringDeleteList(&st);
	stringDeleteList(&ct);
	}

int fixup_string_pointers(pdl_hdr, filebuf)
struct dl_hdr	*pdl_hdr;
char			*filebuf;
	{
	struct dl_anygo 	*pgo;
	int					nwords;
	unsigned			nstr, npc;
	char				**pst;
	char				**pct;

	/*
	 * build a table of pointers to the strings and to the pcs
	 */
	pst = (char **)calloc(pdl_hdr->nstr, sizeof (char *));
	pct = (char **)calloc(pdl_hdr->npc,  sizeof (char *));
	if (!pst || !pct)
		{
		fprintf(stderr, "strings wont fit in memory\n");
		return ERROR;
		}
	pgo = (struct dl_anygo *) filebuf;
	nstr = npc = 0;
	for (nwords = pdl_hdr->nwords_dspy - (WDRND(sizeof(*pdl_hdr))); nwords > 0; 
	  nwords -= pgo->element.nwords, pgo = (struct dl_anygo *)(((short *) pgo) + pgo->element.nwords))
		{
		switch (pgo->element.type)
		  {
		  case dl_string:	pst[nstr++] = (char *)(((char *)pgo) + sizeof(struct dl_string));	break;
		  case dl_pc:		pct[npc++]  = (char *)(((char *)pgo) + sizeof(struct dl_pc));		break;
		  }
		}
	if (nstr != pdl_hdr->nstr || npc != pdl_hdr->npc)
		{
		fprintf(stderr, "bad number of strings\n");
		return ERROR;
		}

	/*
	 * change the string indexes to pointers in the header and each element
	 */
	convert_string_index_to_ptr(dl_hdr_sinfo, pdl_hdr, pct, pst);

	pgo = (struct dl_anygo *) filebuf;
	for (nwords = pdl_hdr->nwords_dspy - (WDRND(sizeof(*pdl_hdr))); nwords > 0; 
	  nwords -= pgo->element.nwords, pgo = (struct dl_anygo *)(((short *) pgo) + pgo->element.nwords))
		{
		int	type	= pgo->element.type;

		if (type == dl_string || type == dl_pc)
			continue;

		convert_string_index_to_ptr(dl_elements_ginfo[type].sinfo, pgo, pct, pst);
		if (strcmp(dl_elements_einfo[type], "<<color rules>>") == 0)
			{	/* these are extended objects */
			struct dl_color_rule_table	*pdlcrules 	= (struct dl_color_rule_table *)pgo;
			struct dl_color_rule 		*pdlcr 		= (struct dl_color_rule *)((char *)pgo + sizeof(struct dl_color_rule_table));
			int							elemno;

			for (elemno = 0; elemno < pdlcrules->nrules; elemno++)
				convert_string_index_to_ptr(dl_color_rule_sinfo, &pdlcr[elemno], pct, pst);
			}
		}
	free(pst);
	free(pct);
	return OK;
	}



/*
 *	DUMP_DL_STRUCTURE
 */
static dump_dl_structure(pstruct, psinfo, pct, pst, level, name, pdstfile)
char			*pstruct;
struct sinfo 	*psinfo;
LIST			*pct;
LIST			*pst;
unsigned		level;
char			*name;
FILE			*pdstfile;
	{
	char 		pstring[64];
	char		*ptmpstr;
	unsigned	count;
	unsigned	i;
	unsigned	elemno;
	char		*psubstruct;
	char		*p;
	int			length;

	for(i = 0; i < level; i++)
		fprintf(pdstfile, "\t");
	fprintf(pdstfile, "%s {\n", name);

	for ( ; psinfo->name != 0; psinfo++)
		{
		if (strncmp(psinfo->name, "pad", 3) == 0)
			continue;

		if (strcmp(psinfo->name, "element") == 0)
			continue;

		if (strcmp(psinfo->name, "point") == 0 || strcmp(psinfo->name, "orgpoint") == 0)
			count = ((struct dl_line *)pstruct)->npoints;
		else
			count = PSINFO_NELEMENTS(psinfo);

		/*
		 * incremental offset will be wrong if the last element was a bit field
		 */
		psubstruct = psinfo->offset + pstruct;

		for (elemno = 0; elemno < count; elemno++)
			{
			if (psinfo->type == dl_struct)
				{
				char *pstr;

				if (count == 1)
					pstr = psinfo->name;
				else
					{
					sprintf(pstring, "%s[%d]", psinfo->name, elemno);
					pstr = pstring;
					}
	
				dump_dl_structure(psubstruct, psinfo->pinfo, pct, pst, level+1, pstr, pdstfile);
				psubstruct += psinfo->size;
				continue;
				}
			for (i = 0; i < (level+1); i++)
				fprintf(pdstfile, "\t");

			ptmpstr = dl_to_pstring(pstruct, psinfo, pct, pst, elemno);
			p = NULL;
			length = strlen(ptmpstr);
			if (length > 0)
				{	/* add \ before " chars in strings */
				char	*s;

				p = calloc(1, (2 * length) + 1 );	/* need at most this many chars */
				if (!p)
					abort();
				for (s = p; *ptmpstr; *s++ = *ptmpstr++)
					{
					if (*ptmpstr == '\"')
						*s++ = '\\';
					}
				*s = 0;
				ptmpstr = p;	/* point to the new string */
				}
			if(count == 1)
				fprintf(pdstfile, "%s=\"%s\"\n",       psinfo->name, ptmpstr);
			else
				fprintf(pdstfile, "%s[%d] = \"%s\"\n", psinfo->name, elemno, ptmpstr);
			if(p)
				free( p );
			}
		}
	if (strcmp(name, "\"<<color map>>\"") == 0)
		{
		struct dl_color 	*pdlc		= (struct dl_color *)((char *)pstruct + sizeof(struct dl_color_map));
		struct dl_color_map	*pdlcmap	= (struct dl_color_map *)pstruct;
		
		for (elemno = 0; elemno < pdlcmap->ncolors; elemno++)
			dump_dl_structure(&pdlc[elemno], dl_color_sinfo, pct, pst, level+1, "dl_color", pdstfile);
		}
	if (strcmp(name, "\"<<color rules>>\"") == 0)
		{
		struct dl_color_rule 		*pdlcr 		= (struct dl_color_rule *)((char *)pstruct + sizeof(struct dl_color_rule_table));
		struct dl_color_rule_table	*pdlcrules 	= (struct dl_color_rule_table *)pstruct;

		for (elemno = 0; elemno < pdlcrules->nrules; elemno++)
			dump_dl_structure(&pdlcr[elemno], dl_color_rule_sinfo, pct, pst, level+1, "dl_color_rule", pdstfile);
		}
	for (i = 0; i < level; i++)
		fprintf(pdstfile, "\t");
	fprintf(pdstfile, "}\n");
	return 0;
	}


/*
 *	TRAVERSE_PSINFO
 */
void traverse_psinfo(psinfo, offset, pfunc, parg)
register struct sinfo	*psinfo;
register unsigned		offset;
void					(*pfunc)();
void					*parg;
	{
	unsigned 	nelements;
	unsigned	suboffset;
	unsigned	i;

	for ( ; psinfo->name; psinfo++)
		{
		nelements = PSINFO_NELEMENTS(psinfo);

		suboffset = psinfo->offset + offset;

		for (i = 0; i < nelements; i++)
			{
			if (psinfo->type == dl_struct)
				traverse_psinfo(psinfo->pinfo, suboffset, pfunc, parg);
			else
				(*pfunc)(parg, psinfo, suboffset, i);

			suboffset += psinfo->size;
			}
		}
	}


/*
 *	VALID_DL_TEST
 */
int valid_dl_test(pdlh)
struct dl_hdr	*pdlh;
	{
	if (pdlh->magic != DL_MAGIC)
		{
		fprintf(stderr, "Not a binary display file?\n");
		return ERROR;
		}
	if (pdlh->majv != DL_MAJOR_VERSION || pdlh->mnrv != DL_MINOR_VERSION)
		{
		fprintf(stderr, "Binary display file version, %d.%d, does not\n",     pdlh->majv,       pdlh->mnrv);
		fprintf(stderr, "match this program's display list version: %d.%d\n", DL_MAJOR_VERSION, DL_MINOR_VERSION);
		return ERROR;
		}
	return OK;
	}


/*
 * 	DUMP_DISPLAY_STRINGS 
 */
void dump_display_strings(name, pdstfile, dump_name)
char	*name;
FILE	*pdstfile;
char	*dump_name;
	{
	int 				fd;
	int 				status;
	char				*filebuf;
	struct dl_anygo 	*pgo;
	int					nwords;
	struct dl_hdr		dl_hdr;
	unsigned			i;

#ifdef WIN32
	fd = _open(name, _O_BINARY|_O_RDONLY);
#else
	fd = open(name, O_RDONLY);
#endif WIN32
	if (fd == ERROR)
		{
		fprintf(stderr, "file access problem: [%s]\n", name);
		return;
		}
	/*
	 * Read in the display list header
	 */
#ifdef WIN32
	status = _read(fd, &dl_hdr, sizeof(dl_hdr));
#else
	status = read(fd, &dl_hdr, sizeof(dl_hdr));
#endif WIN32
	if (status == ERROR)
		{
		fprintf(stderr, "file read problem ?\n");
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		return;
		}
	if (valid_dl_test(&dl_hdr) < 0)
		{
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		return;
		}
	nwords = dl_hdr.nwords_dspy - (WDRND(sizeof(dl_hdr)));
	filebuf = calloc(1, nwords<<1);
	if (!filebuf)
		{
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		fprintf(stderr, "file will not fit in memory\n");
		return;
		}
	/*
	 * Read in the display list
	 */
#ifdef WIN32
	status = _read(fd, filebuf, nwords<<1);
#else
	status = read(fd, filebuf, nwords<<1);
#endif WIN32
	if (status != (nwords<<1))
		{
		fprintf(stderr, "corrupt display file ?\n");
		fprintf(stderr, "tried %d, read %d\n", nwords*2, status);
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		return;
		}
	fprintf(pdstfile, "\n\nDump of display list strings from %s.\n\n\n", name);
	
	/*
	 * dump the string table
	 */
	i = 0;
	fprintf(pdstfile, "ix\tstring\n");
	pgo = (struct dl_anygo *) filebuf;
	for (nwords = dl_hdr.nwords_dspy - (WDRND(sizeof(dl_hdr))); nwords > 0; 
	  nwords -= pgo->element.nwords, pgo = (struct dl_anygo *)(((short *) pgo) + pgo->element.nwords))
		if (pgo->element.type == dl_string)
			fprintf(pdstfile, "%d\t[%s]\n", i++, (char *)((char *)pgo + sizeof(struct dl_string)));

	/*
	 * dump the PC table
	 */
	i = 0;
	fprintf(pdstfile, "\n\nix\tprocess connection name\n");
	pgo = (struct dl_anygo *) filebuf;
	for (nwords = dl_hdr.nwords_dspy - (WDRND(sizeof(dl_hdr))); nwords > 0; 
	  nwords -= pgo->element.nwords, pgo = (struct dl_anygo *)(((short *) pgo) + pgo->element.nwords))
		if (pgo->element.type == dl_pc)
			fprintf(pdstfile, "%d\t[%s]\n", i++, (char *)((char *)pgo + sizeof(struct dl_pc)));

#ifdef WIN32
	_close(fd);
#else
	close(fd);
#endif WIN32
	free(filebuf);
	}
