/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *
 *	D L C V R T . C
 *
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *
 *	routines for converting to/from display list structures
 *
 * 	History
 *	.01 joh 011791	Fixed nill string to numeric type conversion
 *			in pstring to dl.
 */
#include <stdio.h>
#include <dldef.h>
#define DLINFOGLBL
#include <dlinfo.h>
#include <dlcvrt.h>
#include <string_table.h>
#ifdef WIN32
#include <stdlib.h>
#include <string.h>
#endif WIN32

#ifdef TOKENPTRS
#include <tokenSupport.h>
#endif

#define BITINIT_CHK(PSINFO)	do { if(PSINFO->set_field) (* PSINFO->set_field)(PSINFO), PSINFO->set_field = NULL;} while(0)


double dl_to_double(pstructure, psinfo, element)
char			*pstructure;
struct sinfo	*psinfo;
unsigned		element;
	{
	void 	*pitem	= psinfo->offset + pstructure;
	long	ival;

	if (element >= (unsigned)PSINFO_NELEMENTS(psinfo))
		abort();

	switch (psinfo->type)
	  {
	  default:
	  case dl_struct:		fprintf(stderr, "conversion not done yet\n");	return 0.0;
	  case dl_char:
	  case dl_color_index:
	  case dl_bckgnd_color_index:	return (double)(((char *)pitem)[element]);
	  case dl_uchar:		return (double)(((unsigned char *)pitem)[element]);
	  case dl_pchar:		return (double)(*((char **)pitem)[element]);
	  case dl_short:		return (double)(((short *)pitem)[element]);
	  case dl_pc_index:
	  case dl_string_index:
	  case dl_ushort:		return (double)(((unsigned short *)pitem)[element]);
	  case dl_pshort:		return (double)(*((short **)pitem)[element]);
	  case dl_int:			return (double)(((int *)pitem)[element]);
	  case dl_uint:			return (double)(((unsigned int *)pitem)[element]);
	  case dl_pint:			return (double)(*((int **)pitem)[element]);
	  case dl_float:		return (double)(((float *)pitem)[element]);
	  case dl_pfloat:		return (double)(*((float **)pitem)[element]);
	  case dl_double:		return (double)(((double *)pitem)[element]);
	  case dl_pdouble:		return (double)(*((double **)pitem)[element]);
	  case dl_field:
	  case dl_ufield:		BITINIT_CHK(psinfo); read_bitfield(pstructure, psinfo, &ival); return (double)ival;
	  }
	}


long dl_to_long(pstructure, psinfo, element)
char			*pstructure;
struct sinfo	*psinfo;
unsigned		element;
	{
	void 	*pitem	= psinfo->offset + pstructure;
	long 	ival;

	if (element >= (unsigned)PSINFO_NELEMENTS(psinfo))
		abort();

	switch (psinfo->type)
	  {
	  default:
	  case dl_struct:		fprintf(stderr, "conversion not done yet\n");	return 0;
	  case dl_color_index:
	  case dl_bckgnd_color_index:
	  case dl_char:			return (long)(((char *)pitem)[element]);
	  case dl_uchar:		return (long)(((unsigned char *)pitem)[element]);
	  case dl_pc_index:
	  case dl_string_index:
	  case dl_pchar:
#ifndef TOKENPTRS
					return (long)(*((char **)pitem)[element]);
#else
					return (long)(((int *)pitem)[element]);
#endif
	  case dl_short:		return (long)(((short *)pitem)[element]);
	  case dl_ushort:		return (long)(((unsigned short *)pitem)[element]);
	  case dl_pshort:		return (long)(*((short **)pitem)[element]);
	  case dl_int:			return (long)(((int *)pitem)[element]);
	  case dl_uint:			return (long)(((unsigned int *)pitem)[element]);
	  case dl_pint:			return (long)(*((int **)pitem)[element]);
	  case dl_float:		return (long)(((float *)pitem)[element]);
	  case dl_pfloat:		return (long)(*((float **)pitem)[element]);
	  case dl_double:		return (long)(((double *)pitem)[element]);
	  case dl_pdouble:		return (long)(*((double **)pitem)[element]);
	  case dl_field:
	  case dl_ufield:		BITINIT_CHK(psinfo); read_bitfield(pstructure, psinfo, &ival); return ival;
	  }
	}


/*
 *	DL_TO_PSTRING
 */
char *dl_to_pstring(pstructure, psinfo, pct, pst, element)
char			*pstructure;
struct sinfo	*psinfo;
LIST			*pct;
LIST			*pst;
unsigned		element;
	{
	static char	buf[500];	/* returned string is only valid until the next call */
	void 		*pitem		= psinfo->offset + pstructure;
	char		*pstring	= buf;
	char 		*ptmp;
	unsigned 	length;
	long 		ival;

	if(element >= (unsigned)PSINFO_NELEMENTS(psinfo))
		abort();

	if (psinfo->input == edd_menu || psinfo->input == edd_button)
		{	/* return the string equivalent of enumerated types */
		long 	index	= dl_to_long(pstructure, psinfo, element);
	
		return pstring = (index < 0) ? "<cvrt failed>" : (((char **)psinfo->pinfo)[index]);
		}

	switch (psinfo->type)
	  {
	  default:				return "<ukn data type>";
	  case dl_struct:			return ((struct sinfo *)psinfo->pinfo)->name;
	  case dl_color_index:
	  case dl_bckgnd_color_index:
	  case dl_char:				sprintf(pstring, "%d", ((char *)pitem)[element]);			return pstring;
	  case dl_uchar:			sprintf(pstring, "%u", ((unsigned char *)pitem)[element]);	return pstring;
	  case dl_pchar:
#ifndef TOKENPTRS
						sprintf(pstring, "%d", *((char **)pitem)[element]);
#else
						sprintf(pstring, "%d", ((int *)pitem)[element]);
#endif
						return pstring;
	  case dl_short:			sprintf(pstring, "%d", ((short *)pitem)[element]);			return pstring;
	  case dl_ushort:			sprintf(pstring, "%u", ((unsigned short *)pitem)[element]);	return pstring;
	  case dl_pshort:			sprintf(pstring, "%d", *((short **)pitem)[element]);		return pstring;
	  case dl_int:				sprintf(pstring, "%d", ((int *)pitem)[element]);			return pstring;
	  case dl_uint:				sprintf(pstring, "%u", ((unsigned int *)pitem)[element]);	return pstring;
	  case dl_pint:				sprintf(pstring, "%d", *((int **)pitem)[element]);			return pstring;
	  case dl_float:			sprintf(pstring, "%g", ((float *)pitem)[element]);			return pstring;
	  case dl_pfloat:			sprintf(pstring, "%g", *((float **)pitem)[element]);		return pstring;
	  case dl_double:			sprintf(pstring, "%g", ((double *)pitem)[element]);			return pstring;
	  case dl_pdouble:			sprintf(pstring, "%g", *((double **)pitem)[element]);		return pstring;
	  case dl_pc_index:
#ifndef TOKENPTRS
						return (stringName(pct, ((char **)pitem)[element], &ptmp, &length) == OK) ? ptmp : pstring;
#else
						return (stringName(pct, dmPointer64fromToken(((int *)pitem)[element]), &ptmp, &length) == OK) ? ptmp : pstring;
#endif
	  case dl_string_index:
#ifndef TOKENPTRS
						return (stringName(pst, ((char **)pitem)[element], &ptmp, &length) == OK) ? ptmp : pstring;
#else
						return (stringName(pst, dmPointer64fromToken(((int *)pitem)[element]), &ptmp, &length) == OK) ? ptmp : pstring;
#endif
	  case dl_field:
	  case dl_ufield:			BITINIT_CHK(psinfo); read_bitfield(pstructure, psinfo, &ival); sprintf(pstring, "%ld", ival);	return pstring;
	  }
	}


double_to_dl(pstructure, psinfo, fval, element)
char			*pstructure;
struct sinfo	*psinfo;
double			fval;
unsigned		element;
	{
	void	*pitem = psinfo->offset + pstructure;

	if (element >= (unsigned)PSINFO_NELEMENTS(psinfo))
		return ERROR;

	switch (psinfo->type)
	  {
	  default:
	  case dl_struct:		fprintf(stderr, "conversion failed\n");				return ERROR;
	  case dl_color_index:
	  case dl_bckgnd_color_index:
	  case dl_char:			((char *)pitem)[element]		= (char)fval;		return OK;
	  case dl_uchar:		((unsigned char *)pitem)[element]	= ( unsigned char)fval;	return OK;
	  case dl_pc_index:
	  case dl_string_index:
	  case dl_pchar:		*((char **)pitem)[element]		= (char)fval;		return OK;
	  case dl_short:		((short *)pitem)[element]		= (short)fval;		return OK;
	  case dl_ushort:		((unsigned short *)pitem)[element]	= (unsigned short)fval;	return OK;
	  case dl_pshort:		*((short **)pitem)[element]		= (short)fval;		return OK;
	  case dl_int:			((int *)pitem)[element]			= (int)fval;		return OK;
	  case dl_uint:			((unsigned int *)pitem)[element]	= (unsigned int)fval;	return OK;
	  case dl_pint:			*((int **)pitem)[element]		= (int)fval;		return OK;
	  case dl_float:		((float *)pitem)[element]		= (float)fval;		return OK;
	  case dl_pfloat:		*((float **)pitem)[element]		= (float)fval;		return OK;
	  case dl_double:		((double *)pitem)[element]		= fval;			return OK;
	  case dl_pdouble:		*((double **)pitem)[element]		= fval;			return OK;
	  case dl_field:
	  case dl_ufield:		BITINIT_CHK(psinfo); write_bitfield(pstructure, psinfo, (long)fval);	return OK;
	  }
	}

pstring_to_dl(pstructure, psinfo, pstring, pct, pst, element)
char			*pstructure;
struct sinfo	*psinfo;
char			*pstring;
LIST			*pct;
LIST			*pst;
unsigned		element;
	{
	void 		*pitem = psinfo->offset + pstructure;

	if (element >= (unsigned)PSINFO_NELEMENTS(psinfo))
		return ERROR;

	switch (psinfo->type)
	  {
	  default:
	  case dl_struct:
			fprintf(stderr, "conversion not done yet\n");
			return ERROR;
	  case dl_pchar:
#ifndef TOKENPTRS
			((char **)pitem)[element] = pstring;
#else
			((int *)pitem)[element] = dmTokenFromPointer64 (pstring);
#endif
			return OK;
	  case dl_pc_index:
#ifndef TOKENPTRS
			((char **)pitem)[element] = stringInsert(pct, pstring);
#else
			((int *)pitem)[element] = dmTokenFromPointer64 (stringInsert(pct, pstring));
#endif
			return OK;
	  case dl_string_index:
#ifndef TOKENPTRS
			((char **)pitem)[element] = stringInsert(pst, pstring);
#else
			((int *)pitem)[element] = dmTokenFromPointer64 (stringInsert(pst, pstring));
#endif
			return OK;
	  case dl_color_index:
	  case dl_bckgnd_color_index:
	  case dl_char:
	  case dl_uchar:
	  case dl_short:
	  case dl_ushort:
	  case dl_pshort:
	  case dl_int:
	  case dl_uint:
	  case dl_pint:
	  case dl_float:
	  case dl_pfloat:
	  case dl_double:
	  case dl_pdouble:
	  case dl_field:
	  case dl_ufield:
		if (pstring)
			{
			double	fval;
			long 	enum_val;
			char	**pstr;
			
			if (sscanf(pstring, "%le", &fval) == 1)
				return double_to_dl(pstructure, psinfo, fval, element);

			if (psinfo->pinfo)	
				for (pstr = (char **)psinfo->pinfo, enum_val = 0; *pstr != 0; pstr++, enum_val++)
					if (strcmp(*pstr, pstring) == 0)
						return long_to_dl(pstructure, psinfo, enum_val, element);
			}
		fprintf(stderr, "pstring_to_dl: cvrt failed [%s=\"%s\"]\n", psinfo->name, pstring ? pstring : "");
		return ERROR;
	  }
	}

long_to_dl(pstructure, psinfo, ival, element)
char			*pstructure;
struct sinfo	*psinfo;
long 			ival;
unsigned		element;
	{
	void	*pitem	= psinfo->offset + pstructure;

	if (element >= (unsigned)PSINFO_NELEMENTS(psinfo))
		return ERROR;

	switch (psinfo->type)
	  {
	  default:
	  case dl_struct:		fprintf(stderr, "conversion not done yet\n");	return ERROR;
	  case dl_color_index:
	  case dl_bckgnd_color_index:
	  case dl_char:			((char *)pitem)[element]		= (char)ival;		return OK;
	  case dl_uchar:		((unsigned char *)pitem)[element]	= (unsigned char)ival;	return OK;
	  case dl_pc_index:
	  case dl_string_index:
	  case dl_pchar:		*((char **)pitem)[element]		= (char)ival;		return OK;
	  case dl_short:		((short *)pitem)[element]		= (short)ival;		return OK;
	  case dl_ushort:		((unsigned short *)pitem)[element]	= (unsigned short)ival;	return OK;
	  case dl_pshort:		*((short **)pitem)[element]		= (short)ival;		return OK;
	  case dl_int:			((int *)pitem)[element]			= (int)ival;		return OK;
	  case dl_uint:			((unsigned int *)pitem)[element]	= (unsigned int)ival;	return OK;
	  case dl_pint:			*((int **)pitem)[element]		= ival;		return OK;
	  case dl_float:		((float *)pitem)[element]		= (float)ival;		return OK;
	  case dl_pfloat:		*((float **)pitem)[element]		= (float)ival;		return OK;
	  case dl_double:		((double *)pitem)[element]		= ival;		return OK;
	  case dl_pdouble:		*((double **)pitem)[element]		= ival;		return OK;
	  case dl_field:
	  case dl_ufield:		BITINIT_CHK(psinfo); write_bitfield(pstructure, psinfo, ival);	return OK;
	  }
	}
