/*draw_elems.c,v 1.1.1.1 1999/08/03 14:44:07 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
#include <X11/Xlib.h>      			/* the X library 		*/
#include <X11/Xutil.h>
#include <string.h>
#ifdef WIN32
#include <io.h>
#include <process.h>
#include <stdlib.h>
#include <windows.h>
#include <winbase.h>
#else
#include <strings.h>
#include <unistd.h>
#endif WIN32
#include <malloc.h>
#include "dldef.h"			/* display list defines 	*/
#include "dlinfo.h"			/* more display list defines 	*/
#include "font_table.h"			/* defines for font table	*/
#include "dl_context.h"			/* defines dl context		*/

#include <xpm.h>			/* for image drawing		*/

#define XDEG_PER_REV (360*64)

/* X pixmap image support - list of images installed in graphic */
typedef struct xpmImage {
	struct xpmImage *next;		/* list of active images */
	XImage		*ximage;	/* active image info */
	XImage		*shapeimage;	/* active image info */
	char		filename [160];	/* name of image file original data */
	int		width;		/* size of target on screen */
	int		height;		/* (ditto) */
	struct dl_text *ptext;		/* pointer to parent text object */
} XPMImage;

static XPMImage *images;		/* list of images in this graphic */

/***********************************************************************
**
**	DRAW_STATIC
**	
**	Deb Kerstiens 6/6/90
**
**
**	Purpose: 
**		This routine draws the static elements the a display list
**		into an X Window.
**		Currently the fonts are loaded if not already done.  For
**		efficiency, the fonts should be preloaded.  Note :the code for 
**		loading fonts might be best located elsewhere.  
**	
**	Notes: 	
**		Clipping and bounds checking can be turned off by setting the
**		XRectangle pointer to nil.
**
**	Input :
**		pdlc:		pointer to the display list context
**		dl_ptr:		pointer to a segment of display list (without dl_hdr)
**	    	dl_size: 	size (in shorts) of the display list segment
**		prect:		pointer to an XRecatngle limiting the redraw area.
**	Output:
**	    	none
**	Returns:
**	    	none
**
**	History
**	06-14-90 joh	added pdlc argument- a pointer to the display list context
**	06-21-90 joh 	finished arcs
**	07-20-90 joh	added rectangle ptr argument for selective redraw
**	08-01-90 joh	draw text centered in boxes larger than MAXFONTHEIGHT
**	08-01-90 joh	fetch string given the string index
**	09-17-90 joh	added rising/falling lines
**	09-17-90 joh	added element attributes
**	09-21-90 joh	added filled objects
**	01-15-91 joh	added text format
**	03-27-91 joh	changed to draw_fixed_XXX_string()
**
**********************************************************************/

void draw_static(pdlc, dl_ptr, dl_size, prec, edit_mode)
struct dl_context	*pdlc;
void 			*dl_ptr;
unsigned 		dl_size;	
XRectangle		*prec;	/* the area of the window to redraw */
int				edit_mode;
{
  register void 	*dis_ptr;
  void 			*last_adr;	/* last mem adr in display list */
  register int		 lw;		/* line width from the GC */
  register int		 lwd2;		/* line width from the GC over 2 */
  static int			current_foreground_color;

#ifdef DEBUG
printf("in draw_static\n");
#endif

  if(prec)
  	XSetClipRectangles(pdlc->d, pdlc->gc, 0, 0, prec, 1, YXBanded);

  last_adr = (short *)dl_ptr + dl_size;
   /* for each element on the display list */
  for (	dis_ptr = dl_ptr;  
#ifndef AXP
	(int)dis_ptr < (int)last_adr;
#else
	dis_ptr < last_adr;
#endif
	dis_ptr = (void *)((short *)dis_ptr + ((struct dl_element_hdr *)dis_ptr)->nwords)){

    register struct dl_object_hdr *objptr = 
		&(((struct dl_anygo *)dis_ptr)->object);
    register struct dl_element_hdr *elemptr = 
		&(((struct dl_anygo *)dis_ptr)->element);

   	/*
    	 * 	update the GC from the attributes
	 *	joh	9-17-90
    	 */
    if(dl_element_class[elemptr->type] != dl_class_graphic){
    	switch(elemptr->type){
	case dl_basic_attribute:{
		struct dl_sta_attr *pattr;
		XGCValues 	GC_values;

		pattr = &((struct dl_basic_attribute *)dis_ptr)->attr;

		if(pattr->clr == invisibleColor)
			{
			GC_values.foreground = WhitePixel(pdlc->d, DefaultScreen(pdlc->d));
			current_foreground_color = invisibleColor;
			}
		else
			{
			if(pdlc->color)
  				current_foreground_color = GC_values.foreground = pdlc->pcmap->pcolor_pixels[pattr->clr];
			else
  				current_foreground_color = GC_values.foreground = pdlc->pcmap->pinten_pixels[pattr->clr];
			}
  		GC_values.line_width = pattr->width;
		GC_values.join_style = JoinRound;
		GC_values.cap_style = CapProjecting;
		if(pattr->style == dl_edge_solid)
			GC_values.line_style = LineSolid;
		else if(pattr->style == dl_edge_dash)
			GC_values.line_style = LineOnOffDash;
		else
			GC_values.line_style = LineSolid;

		pdlc->fill_style = pattr->fill;

		/*
		 * Does not send a message to the server unless somthing
		 * changed
		 */
		XChangeGC(	pdlc->d,
				pdlc->gc,
				GCForeground|GCLineWidth|GCLineStyle|GCJoinStyle|GCCapStyle,
				&GC_values);
		break;
	}
	case dl_dynamic_attribute:
		printf("draw_elems: saw dynamic attr?\n");
		break;
	default:
		abort();
	}

	continue;
    }

#ifdef ILLEGAL_X_USAGE
    lw = pdlc->gc->values.line_width;
#else
    {
	XGCValues values;

	if (XGetGCValues (pdlc->d, pdlc->gc, GCLineWidth, &values))
	    lw = values.line_width;
	else
	    lw = 1;
    }
#endif ILLEGAL_X_USAGE
    lwd2 = lw>>1;



    /* 
	joh 7-20-90

	only redraw if in the exposed area of the screen
    */
    if(prec){
	int x1, x2, y1, y2;

    	x1 = objptr->x;
    	y1 = objptr->y;
    	x2 = objptr->x + objptr->width;
    	y2 = objptr->y + objptr->height;

    	if(pdlc->fill_style == dl_fill_outline){
		x1 -= lwd2;
		y1 -= lwd2;
		x2 += lwd2;
		y2 += lwd2;
    	}

   	if( x2 < prec->x )
		continue;
    	if( x1 > (prec->x + (int)prec->width) )
		continue;
    	if( y2 < prec->y )
		continue;
    	if( y1 > (prec->y + (int)prec->height) )
		continue;
    }

#ifdef DEBUG
      printf(	"Type %s\n", dl_elements_einfo[elemptr->type]);
#endif
	if (current_foreground_color != invisibleColor || edit_mode)
    switch(elemptr->type){
      case dl_user: 
      case dl_rec: /* rectangle */
      { 

	/* 
	draw the rectangle 
	joh 9-21-90 added filled arcs
	*/
	if(pdlc->fill_style == dl_fill_outline){
		XDrawRectangle(	
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				objptr->x,
				objptr->y,
				objptr->width, 
				objptr->height);
	}else
		XFillRectangle(	pdlc->d,
				pdlc->w,
				pdlc->gc,
				objptr->x,
				objptr->y,
				objptr->width, 
				objptr->height);
	break;
      }

      case dl_line:{ /* line */
	if(pdlc->fill_style == dl_fill_outline)
		XDrawLines(	pdlc->d,
				pdlc->w,
				pdlc->gc,
				(XPoint *) &(((struct dl_line *)dis_ptr)->point[0]),
				(int) ((struct dl_line *)dis_ptr)->npoints,
				CoordModeOrigin);
	else
		XFillPolygon(	pdlc->d,
				pdlc->w,
				pdlc->gc,
				(XPoint *) &(((struct dl_line *)dis_ptr)->point[0]),
				(int) ((struct dl_line *)dis_ptr)->npoints,
				Complex,
				CoordModeOrigin);
	break;
	}

      case dl_oval:{ /* oval */

	/* 
	draw the oval
	joh 9-21-90 added filled arcs
	*/
	if(pdlc->fill_style == dl_fill_outline){
		XDrawArc(
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				objptr->x,
				objptr->y,
				objptr->width, 
				objptr->height,
				0,
				XDEG_PER_REV);
	}
	else
		XFillArc(
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				objptr->x,
				objptr->y,
				objptr->width,
				objptr->height,
				0,
				XDEG_PER_REV);
	break;

	}
      case dl_arc:{ /* arc */
	/*
	joh 6-21-90 finished arcs

	joh 9-21-90 added filled arcs
	*/
	register struct dl_arc *parc = (struct dl_arc *) dis_ptr;

	if(pdlc->fill_style == dl_fill_outline){
		XDrawArc(
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				parc->object.x,
				parc->object.y,
				parc->object.width,
				parc->object.height,
				parc->begin,
				parc->path);
	}
	else
		XFillArc(
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				parc->object.x,
				parc->object.y,
				parc->object.width,
				parc->object.height,
				parc->begin,
				parc->path);
	
	break;
      }

      case dl_text:{ /* text */
	char 			*display_str;
	unsigned 		display_str_length;
	int			status;
	XRectangle		box;
	enum string_format	frmt;
	int			horiz;
	char			image_error_msg [100];

	register struct dl_text *ptext = (struct dl_text *) dis_ptr;

	/*
	 * joh 7-20-90
	 * 
	 * convert string index to string pointer
	 */
	status = string_index_to_display_str(	pdlc->pstring_table,
						ptext->textix, /* pointer is really an integer! */
						&display_str, 
						&display_str_length);
	if(status < 0)
		abort();

	DLOBJ_TO_XREC(ptext->object, box);

	/* if string is form <<...>> draw xpm image in bounding box */
	if (((strncmp (display_str, "<<", 2) == 0)
	      && (strcmp (display_str + strlen(display_str)-2, ">>") == 0)))
	{
	    XPMImage *curimage;		/* place in image list for searches */
	    char filename [160];	/* name of image file original data */
	    char *cp;

	    for (curimage = images; curimage; curimage = curimage->next)
		if (curimage->ptext == ptext)
		    break;
	    if (curimage)
	    {
		if (strcmp (curimage->filename, display_str)
		  || (curimage->width != box.width)
		  || (curimage->height != box.height))
		{
		    /* don't change image while selected! */
		    if ((edit_mode & 2) == 0)
		    {
			/* delete previous image, doesn't fit anymore */
			if (curimage->ximage)
			    XDestroyImage (curimage->ximage);
			curimage->ximage = NULL;
			if (curimage->shapeimage)
		            XDestroyImage (curimage->shapeimage);
			strncpy (curimage->filename, display_str, sizeof(curimage->filename));
			curimage->width = box.width;
			curimage->height = box.height;
		    }
		}
	    }
	    else
	    {
		curimage = calloc (1, sizeof(XPMImage));
		if (curimage == NULL)
		{
		    sprintf (image_error_msg, "no memory for image structure for %s",
					display_str);
		    display_str = image_error_msg;
		    display_str_length = strlen (image_error_msg);
		}
		else
		{
		    strncpy (curimage->filename, display_str, sizeof(curimage->filename));
		    curimage->width = box.width;
		    curimage->height = box.height;
		    curimage->ptext = ptext;
		    /* stick it on the list */
		    curimage->next = images;
		    images = curimage;
		}
	    }
	    strncpy (filename, display_str+2, strlen(display_str)-4);
	    filename [strlen(display_str)-4] = 0;
	    if (curimage && curimage->ximage == NULL && ((edit_mode & 2) == 0)
		&& (_access( filename, 0 ) != -1))
	    {
		char	command[100];

		if (edit_mode)
		{
		    /* First, use mogrify to fit image to rectangle, and reduce colors, save as xpm file. */
#ifdef SPAWNIT
#ifdef WIN32
		    int pid;
		    char *argv [] =
		    {
			"mogrify",	/* arg 0 */
			"-format",	/* 1 */
			"xpm",		/* 2 */
			"-colors",	/* 3 */
			"32",		/* 4 */
			"-geometry",	/* 5 */
			"",		/* 6 */
			"",		/* 7 */
			0		/* 8 */
		    };

		    sprintf (command, "%dx%d", box.width, box.height);
		    argv [6] = command;
		    argv [7] = filename;
		    pid = _spawnvp (_P_WAIT, argv[0], (const char * const *)argv);
#else WIN32
		    sprintf (command, "%dx%d", box.width, box.height);
		    _spawnl (_P_WAIT, "c:\\tis4000\\tasks\\ImageMagick\\bin\\mogrify.exe", "mogrify", "-format", "xpm",
				      "-colors", "32", "-geometry", command, filename, NULL);
#endif WIN32
#else SPAWNIT
#ifdef WIN32
		    PROCESS_INFORMATION    procInfo;
		    STARTUPINFO	startInfo;
		    long   exitCode;

		    sprintf (command, "mogrify -format xpm -colors 32 -geometry %dx%d %s", box.width, box.height, filename);
		    GetStartupInfo (&startInfo);
		    if (CreateProcess (NULL, command, NULL, NULL, TRUE, DETACHED_PROCESS, NULL, NULL, &startInfo, &procInfo))
		    {
			while (GetExitCodeProcess (procInfo.hProcess, &exitCode) && exitCode == STILL_ACTIVE)
			    /* waiting for finish */;
		    }
		    else
		    {
			LPVOID lpMsgBuf;
 
			FormatMessage( 
			    FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			    NULL,
			    GetLastError(),
			    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
			    (LPTSTR) &lpMsgBuf,
			    0,
			    NULL 
			);

			// Display the string.
//			MessageBox( NULL, lpMsgBuf, "GetLastError", MB_OK|MB_ICONINFORMATION );

			// Free the buffer.
			LocalFree( lpMsgBuf );
 
		    }
/*		    system ( command ); */
#else WIN32
		    sprintf (command, "mogrify -format xpm -colors 32 -geometry %dx%d %s", box.width, box.height, filename);
		    system ( command );
#endif WIN32
#endif SPAWNIT
		}
		/* open xpm file, and copy to window */
		if (cp = strchr(filename, '.'))
		    strcpy (cp+1, "xpm");
		/* Create the X pixmap. */
		if (XpmReadFileToImage (pdlc->d, filename, &curimage->ximage, &curimage->shapeimage,
				   NULL /*XpmAttributes * */) != XpmSuccess)
		{
		    if (curimage->ximage)
			XDestroyImage (curimage->ximage);
		    curimage->ximage = NULL;
		    if (curimage->shapeimage)
		        XDestroyImage (curimage->shapeimage);
		}
	    }
	    if (curimage->ximage)
	    {
		int srcX, srcY;
		    /* Display the X pixmap. in the expose area. */
		    if (prec)
		    {
			srcX = prec->x - box.x; srcY = prec->y - box.y;
		    }
		    else
		    {
			srcX = 0; srcY = 0;
		    }
		    XPutImage (pdlc->d, pdlc->w, pdlc->gc, curimage->ximage, srcX, srcY,
					box.x, box.y, box.width - srcX, box.height - srcY);
		break;
	    }
	    else
	    {
		sprintf (image_error_msg, "Can't open image file %s", display_str);
		display_str = image_error_msg;
		display_str_length = strlen (image_error_msg);
	    }
	}

	switch(ptext->align){
	case dl_horiz_left:
		horiz = TRUE;
		frmt = box_begin;
		break;
	case dl_horiz_right:
		horiz = TRUE;
		frmt = box_end;
		break;
	case dl_horiz_center:
		horiz = TRUE;
		frmt = box_center;
		break;
	case dl_vert_top:
		horiz = FALSE;
		frmt = box_begin;
		break;
	case dl_vert_bottom:
		horiz = FALSE;
		frmt = box_end;
		break;
	case dl_vert_center:
		horiz = FALSE;
		frmt = box_center;
		break;
	default:
		abort();
	}

	/*
	 * using size of extent rectange, determine appropriate font and draw
	 */
	if(horiz)
		draw_fixed_horizontal_string(
			pdlc->d, 
			pdlc->w, 
			pdlc->gc, 
			&box, 
			pdlc->pfont_table, 
			display_str, 
			display_str_length,
			frmt);
	else
		draw_fixed_vertical_string(
			pdlc->d, 
			pdlc->w, 
			pdlc->gc, 
			&box, 
			pdlc->pfont_table, 
			display_str, 
			display_str_length,
			frmt);

	break;
	}

      default:
	printf("Type not matched\n");
	break;
    } /* end switch */
  } /* end for */

  /*
   *	Remove clip mask
   *	(may not be required if this dlc is not shared)
   */
  if(prec)
    XSetClipMask(pdlc->d, pdlc->gc, None);

  return;

} /* end draw_obj */


