/*font_table.c,v 1.1.1.1 1999/08/03 14:44:08 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 * 	F O N T _ T A B L E . C
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 * Loads a table of related fonts ordered by font height.
 * 
 * 
 * 
 * Notes 
 *
 * .01 	XListFontsWithInfo does not set the fid in the info structure.
 * 
 * .02 	All of the XListFonts type routines are very slow. tisLoadFont can use a
 * wild carded name, but it becomes slow also. Attempts to load a batch of
 * names and then parse them for the correct sizes locally provides no
 * noticeable improvement since the time to load N batched names is about as
 * long as the time to do the individually.
 * 
 * .03 	I have settled on determining the names of the fonts once after the
 * server boots or resets and then loading the names from a resource
 * for each client attaching thereafter.
 *
 *
 * History
 *
 */

#include	<memLeak.h>
#include 	<stdio.h>
#include 	<X11/Xlib.h>
#ifdef WIN32
#include <X11/XlibXtra.h>
#endif WIN32
#include 	<X11/Xutil.h>
#include 	<X11/Xatom.h>
#ifdef AXP
#include <string.h>
#include <stdlib.h>
#endif

#include	"dldef.h"
#include	"font_table.h"

#define		ANYWINDOW None 

#define		FONT_TABLE_ATOM_NAME	"EPICS_OPI_FONT_TABLE"

extern XFontStruct *tisLoadQueryFont();

/* something not in the X logical font naming standard */
#define		SKIP_STRING		"#!."	

#ifdef STATIC_FONT_NAME
static char	*static_font_list[] = {
"-misc-fixed-medium-r-normal--8-60-100-100-c-50-iso8859-1",
"-misc-fixed-medium-r-normal--9-80-100-100-c-60-iso8859-1",
"-misc-fixed-medium-r-normal--10-100-75-75-c-60-iso8859-1",
"-misc-fixed-medium-r-normal--13-100-100-100-c-70-iso8859-1",
"-misc-fixed-medium-r-normal--14-110-100-100-c-70-iso8859-1",
"-misc-fixed-medium-r-normal--15-120-100-100-c-90-iso8859-1",
"-sony-fixed-medium-r-normal--16-120-100-100-c-80-iso8859-1",
"-misc-fixed-medium-r-normal--20-140-100-100-c-100-iso8859-1",
"-sony-fixed-medium-r-normal--24-170-100-100-c-120-iso8859-1"};
#endif

static char	*prefered_font_list[] = {
"-adobe-*-bold-r-*--*-*-*-*-m-*-iso8859-1*",
"-adobe-*-bold-r-*--*-*-*-*-m-*-iso8859-1*",
"-adobe-*-bold-r-*--*-*-*-*-c-*-iso8859-1*",
};

static char	*old_prefered_font_list[] = {
"-*-*-bold-r-*--*-*-*-*-m-*-iso8859-1*",
"-*-*-bold-r-*--*-*-*-*-m-*-iso8859-1*",
"-*-*-bold-r-*--*-*-*-*-c-*-iso8859-1*",
};

static char *get_font_names();
static set_font_names();

/*
 * How many fonts fonts I am willing to look at
 * (the first time a client connects to the server only)
 */
#define MAXFONTLIST 300


/*
 * INIT FONT TABLE
 */
font_table *init_font_table(d)
Display				*d;
{
	unsigned		pixel_height;
	unsigned		pixel_width;
	XFontStruct		*pfi = NULL;
	font_table 		*pft = NULL;
	static XContext		font_table_ctx;
	Atom			font_table_atom;
	char			*font_name_string;
	char			*fns;
	Status			status;

	if(!font_table_ctx)
		font_table_ctx = XUniqueContext();

	status = XFindContext(d, ANYWINDOW, font_table_ctx, (caddr_t *)&pft);
	if(status == 0)
		return pft;

	pft = (font_table *) calloc(1, sizeof(*pft));
	if(!pft)
		abort();

	font_table_atom = XInternAtom(d, FONT_TABLE_ATOM_NAME, False);
	if(font_table_atom == None)
		abort();


	font_name_string = get_font_names(d, font_table_atom);
font_name_string = NULL;
	if(!font_name_string){
		set_font_names(d, font_table_atom);
		font_name_string = get_font_names(d, font_table_atom);
		if(!font_name_string)
			abort();
	}

/*	printf("Loading fonts.... "); */
#if !(defined(vxWorks) || defined(WIN32))
	fflush(stdout);
#endif

	fns = font_name_string;
	pft->min_font_height = MAXFONTHEIGHT;
	pft->max_font_height = MINFONTHEIGHT;
	for(	pixel_height = MINFONTHEIGHT, pfi = NULL; 
		pixel_height<NELEMENTS(pft->height_to_pfinfo); 
		pixel_height++){

#ifdef DEBUG
		printf("%s\n", fns);
#endif
		if(*fns){
			if(strcmp(fns, SKIP_STRING)){
				XFontStruct *pfitmp;

				pfitmp = tisLoadQueryFont(d, fns);
				if(pfitmp)
					pfi = pfitmp;
			}
			fns += strlen(fns)+1;
		}
		if(pfi){
			pft->height_to_pfinfo[pixel_height] = pfi;
			pft->max_font_height = 
				max(pixel_height, pft->max_font_height);
			pft->min_font_height =
				min(pft->min_font_height, pixel_height);

			if(pfi->max_bounds.width<NELEMENTS(pft->width_to_pfinfo));
			pft->width_to_pfinfo[pfi->max_bounds.width] = pfi;

		}
	}

	/*
	 * fill in any gaps in the width table
	 */
	pfi = NULL;
	pft->min_font_width = MAXFONTWIDTH;
	pft->max_font_width = MINFONTWIDTH;
	for(	pixel_width = MINFONTWIDTH; 
		pixel_width < NELEMENTS(pft->width_to_pfinfo); 
		pixel_width++){
		if(pft->width_to_pfinfo[pixel_width]){
			pft->max_font_width = max(pft->max_font_width, pixel_width);
			pft->min_font_width = min(pft->min_font_width, pixel_width);

			pfi = pft->width_to_pfinfo[pixel_width];
		}
		else
			pft->width_to_pfinfo[pixel_width] = pfi;
	}

/*	printf("done\n"); */

	if(XSaveContext(d, ANYWINDOW, font_table_ctx, (XPointer)pft))
		abort();

	return pft;
}


/*
 * GET FONT NAMES
 * 
 * property allways returned with an ASCII NULL at the end
 */
static char *get_font_names(d, font_table_atom)
Display		*d;
Atom		font_table_atom;
{
	int		status;
	Atom		actual_type;
	int		actual_format;
	unsigned long	actual_nitems;
	unsigned long	bytes_left;
	unsigned char	*pfont_list;
	font_table 	*pft; /* for NELEMENTS (don't delete, referenced in macro expansion) */


	status = XGetWindowProperty(
		d,
		DefaultRootWindow(d),
		font_table_atom,
		0L, 					/* offset */
		(long)(NELEMENTS(pft->height_to_pfinfo)*50), 	/* fetch len (all of it) */
		False,					/* dont delete */
		font_table_atom,			/* requested type */	
		&actual_type,
		&actual_format,	
		&actual_nitems,
		&bytes_left,
		&pfont_list);
	if(status == Success){
		if(actual_type == font_table_atom)
			if(actual_format == 8)
				if(bytes_left == 0)
					return (char *)pfont_list;
	}
	return NULL;
}





/*
 * SET FONT NAMES
 * 
 *	It is best to search for a font of a particular point size so
 *	that a font matching the screen resolution will be used.
 *
 */
static set_font_names(d, font_table_atom)
Display		*d;
Atom		font_table_atom;
{
	font_table 	*pft; /* for NELEMENTS */
	int		i, j;
	char		*ordered_font_names[NELEMENTS(pft->height_to_pfinfo)];
	struct{
		XFontStruct	*pfi;
		char		**font_names;
		int		nfound;
	} font_stuff[NELEMENTS(prefered_font_list) + NELEMENTS(old_prefered_font_list)];
	  /* This ensures that font_stuff is big enough to hold the biggest list */
	int	oldFonts, numFonts;
	char	**fontList;

	oldFonts = getenv ("DMOLDFONTS") != 0;
	if (oldFonts) {
	    fontList = old_prefered_font_list;
	    numFonts = NELEMENTS(old_prefered_font_list);
	} else {
	    fontList = prefered_font_list;
	    numFonts = NELEMENTS(prefered_font_list);
	}

/*	printf("Searching for fonts....\n"); */
#if !(defined(vxWorks) || defined(WIN32))
	fflush(stdout);
#endif

	for(i=0; i<NELEMENTS(ordered_font_names); i++){
		ordered_font_names[i] = NULL;
	}

#ifdef DEBUG
	printf("sorting whats available\n");
#endif
	for(i=0 ; i<numFonts ; i++){
/*		printf ("%s\n", fontList [i]);  */
		font_stuff[i].font_names = XListFontsWithInfo(
			d, 
			fontList [i], 
			MAXFONTLIST, 
			&font_stuff[i].nfound,
			&font_stuff[i].pfi);

		for(j=0; j<font_stuff[i].nfound; j++){
			unsigned height;

			height = font_stuff[i].pfi[j].max_bounds.ascent + 
				font_stuff[i].pfi[j].max_bounds.descent;
#ifdef DEBUG
			printf(	"pixel height = %d, %s\n", 
				height, 
				font_stuff[i].font_names[j]);
#endif
			if(height < NELEMENTS(ordered_font_names)){
				if(ordered_font_names[height] == NULL){
					ordered_font_names[height] = 
						font_stuff[i].font_names[j];
				}
			}
		}
	}

#ifdef DEBUG
	printf("writing sorted list to the server\n");
#endif
	XChangeProperty(
		d,
		DefaultRootWindow(d),
		font_table_atom,
		font_table_atom,
		8, 			/* char format */
		PropModeReplace,
		NULL,
		0);

/*	printf ("Selected fonts:\n"); */

	for(	i = MINFONTHEIGHT; 
		i<NELEMENTS(ordered_font_names); 
		i++){

		if(ordered_font_names[i] == NULL){
			ordered_font_names[i] = SKIP_STRING;
		} else {
/*		    printf("pixel height = %d, %s\n", i, ordered_font_names[i]); */
		}

		XChangeProperty(
				d,
				DefaultRootWindow(d),
				font_table_atom,
				font_table_atom,
				8, 			/* char format */
				PropModeAppend,
				(unsigned char *)ordered_font_names[i],
				strlen(ordered_font_names[i])+1);
	}

	for(i=0; i<numFonts; i++){
		XFreeFontInfo(	
			font_stuff[i].font_names,
			font_stuff[i].pfi,
			font_stuff[i].nfound);
	}

/*	printf("done\n"); */
return 0;
}

