/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
**	tisXext.c - X extensions to support dm and edd for 16 bit fonts.
**
** Font         tisLoadFont (display, font_name)
** XFontStruct *tisLoadQueryFont (display, font_name)
**              tisDrawString (display, window, gc, x, y, text, textlen)
**              tisDrawImageString (display, window, gc, x, y, text, textlen)
** int          tisTextWidth (font_struct, string, string_len)
**
*/

#include	<memLeak.h>
#include        <stdio.h>
#include        <Xlib.h>           
#include        <Xutil.h>
#include	<string.h>
#ifdef AXP
#include <stdlib.h>
#endif

/*#include	<malloc.h>*/
/*#define		USE_XMB	1 */

/* We will handle this many 8/16 font pairs */
#define MAXFONTS	100

typedef struct {
	Font	fid;
	char	*font8;
#ifdef	USE_XMB
	XFontSet	xmbFid;
	char		*xmbFontPair;
#else
	Font	fid16;
	char	*font16;
#endif
} TisFontList;

TisFontList	fontList [MAXFONTS];		/* room for 100 font relations */
Display		*savedDisplay;			/* remember for TextWidth16 call */

/*  The fontPairs array must be the same size as the fontPairNames list, and 
** fontPairNames must collate in consecutive sequence in the StringTable!
*/
#define MAXFONTPAIRS 25
int fontPairInitialized = 0; /* flag indicating pairs were loaded already */
char *fontPairs [MAXFONTPAIRS+1];

/* fontPairNames - the strConv program picks the strings in this comment
** out and enters them in the StringTable. Add new values as necessary,
** but watch the collation is consecutive in StringTable.
**    "  FONTPAIR01",
**    "  FONTPAIR02",
**    "  FONTPAIR03",
**    "  FONTPAIR04",
**    "  FONTPAIR05",
**    "  FONTPAIR06",
**    "  FONTPAIR07",
**    "  FONTPAIR08",
**    "  FONTPAIR09",
**    "  FONTPAIR10",
**    "  FONTPAIR11",
**    "  FONTPAIR12",
**    "  FONTPAIR13",
**    "  FONTPAIR14",
**    "  FONTPAIR15",
**    "  FONTPAIR16",
**    "  FONTPAIR17",
**    "  FONTPAIR18",
**    "  FONTPAIR19",
**    "  FONTPAIR20",
**    "  FONTPAIR21",
**    "  FONTPAIR22",
**    "  FONTPAIR23",
**    "  FONTPAIR24",
**    "  FONTPAIR25"
*/
/*
** from plain_menus.c
**    "-*-*-bold-r-*--*-120-*-*-c-*-iso8859-1*,-*-*-medium-r-*--*-120-*-*-c-*-ksc5601.1987-0*",
**    "-*-*-bold-r-*--*-160-*-*-c-*-iso8859-1*,-*-*-medium-r-*--*-160-*-*-c-*-ksc5601.1987-0*",
**    "-*-*-bold-r-*--*-140-*-*-c-*-iso8859-1*,-*-*-medium-r-*--*-140-*-*-c-*-ksc5601.1987-0*",
**    "-*-*-bold-r-*--*-*-*-*-c-*-iso8859-1*,-*-*-medium-r-*--*-*-*-*-c-*-ksc5601.1987-0*",
**    "*--*-*-*-*-c-*-iso8859-1*,*--*-*-*-*-c-*-ksc5601.1987-0*",
**    "fixed,*--*-*-*-*-c-*-ksc5601.1987-0*",
** from font_table, prefered_font_list * /
**    "-adobe-courier-bold-r-*--*-*-*-*-m-*-iso8859-1*,-*-*-medium-r-*--*-*-*-*-c-*-ksc5601.1987-1*",
**    "-adobe-*-bold-r-*--*-*-*-*-m-*-iso8859-1*,-*-*-medium-r-*--*-*-*-*-m-*-ksc5601.1987-1*",
**    "-adobe-*-bold-r-*--*-*-*-*-c-*-iso8859-1*,-*-*-medium-r-*--*-*-*-*-c-*-ksc5601.1987-1*",
** from font_table, old_prefered_font_list * /
**    "-*-courier-bold-r-*--*-*-*-*-m-*-iso8859-1*,-*-*-medium-r-*--*-*-*-*-m-*-ksc5601.1987-1*",
**    "-*-*-bold-r-*--*-*-*-*-m-*-iso8859-1*,-*-*-medium-r-*--*-*-*-*-m-*-ksc5601.1987-1*",
**    "-*-*-bold-r-*--*-*-*-*-c-*-iso8859-1*,-*-*-medium-r-*--*-*-*-*-c-*-ksc5601.1987-1*",
** from font_table, static_font_list * /
**    "-misc-fixed-medium-r-normal--8-60-100-100-c-50-iso8859-1,-sun-gothic-medium-r-normal--14-120-75-75-c-120-ksc5601.1987-0",
**    "-misc-fixed-medium-r-normal--9-80-100-100-c-60-iso8859-1,-sun-gothic-medium-r-normal--14-120-75-75-c-120-ksc5601.1987-0",
**    "-misc-fixed-medium-r-normal--10-100-75-75-c-60-iso8859-1,-sun-gothic-medium-r-normal--14-120-75-75-c-120-ksc5601.1987-0",
**    "-misc-fixed-medium-r-normal--13-100-100-100-c-70-iso8859-1,-sun-gothic-medium-r-normal--14-120-75-75-c-120-ksc5601.1987-0",
**    "-misc-fixed-medium-r-normal--14-110-100-100-c-70-iso8859-1,-sun-gothic-medium-r-normal--14-120-75-75-c-120-ksc5601.1987-0",
**    "-misc-fixed-medium-r-normal--15-120-100-100-c-90-iso8859-1,-sun-gothic-medium-r-normal--14-120-75-75-c-120-ksc5601.1987-0",
**    "-sony-fixed-medium-r-normal--16-120-100-100-c-80-iso8859-1,-sun-myoungjo-medium-r-normal--16-140-75-75-c-140-ksc5601.1987-0",
**    "-misc-fixed-medium-r-normal--20-140-100-100-c-100-iso8859-1,-sun-myoungjo-medium-r-normal--16-140-75-75-c-140-ksc5601.1987-0",
**    "-sony-fixed-medium-r-normal--24-170-100-100-c-120-iso8859-1,-sun-myoungjo-medium-r-normal--22-200-75-75-c-200-ksc5601.1987-0",
**    ""
*/

char *default16 = "fixed";

/*
** Font        tisLoadFont (display, font_name)
*/
Font tisLoadFont (display, font_name)
Display *display;
char *font_name;
{
    int ix, jx;
	char	**missing_char_list;
	char	*def_string;
	int		missing_char_count;

    savedDisplay = display;

    /* initialize font list if necessary */
    if (!fontPairInitialized)
		{
	    fontPairs[0] =     "~  FONTPAIR01";
	    fontPairs[1] =     "~  FONTPAIR02";
	    fontPairs[2] =     "~  FONTPAIR03";
	    fontPairs[3] =     "~  FONTPAIR04";
	    fontPairs[4] =     "~  FONTPAIR05";
	    fontPairs[5] =     "~  FONTPAIR06";
	    fontPairs[6] =     "~  FONTPAIR07";
	    fontPairs[7] =	   "~  FONTPAIR08";
	    fontPairs[8] =     "~  FONTPAIR09";
	    fontPairs[9] =     "~  FONTPAIR10";
	    fontPairs[10] =    "~  FONTPAIR11";
	    fontPairs[11] =    "~  FONTPAIR12";
	    fontPairs[12] =    "~  FONTPAIR13";
	    fontPairs[13] =    "~  FONTPAIR14";
	    fontPairs[14] =    "~  FONTPAIR15";
	    fontPairs[15] =    "~  FONTPAIR16";
	    fontPairs[16] =    "~  FONTPAIR17";
	    fontPairs[17] =    "~  FONTPAIR18";
	    fontPairs[18] =    "~  FONTPAIR19";
	    fontPairs[19] =    "~  FONTPAIR20";
	    fontPairs[20] =    "~  FONTPAIR21";
	    fontPairs[21] =    "~  FONTPAIR22";
	    fontPairs[22] =    "~  FONTPAIR23";
	    fontPairs[23] =    "~  FONTPAIR24";
	    fontPairs[24] =    "~  FONTPAIR25";
	    fontPairs[25] =    0;
	    fontPairInitialized = 1;
	    }

    /* try to find font in the font list */
    for (ix = 0; ix < MAXFONTS && fontList[ix].font8; ix++)
	    {
		if (strcmp (fontList[ix].font8, font_name) == 0)
			{
	    	return fontList[ix].fid;
			}
    	}
    if (ix < MAXFONTS)
    	{
	/* ix is pointed at a vacant slot, look for font in pairing list */
		for (jx = 0; fontPairs[jx]; jx++)
			{
		    if (strncmp (font_name, fontPairs[jx], strcspn(fontPairs[jx], ",")) == 0)
				{
			    fontList[ix].fid = XLoadFont (display, font_name);
			    fontList[ix].font8 = malloc (strlen (font_name)+1);
			    strcpy (fontList[ix].font8, font_name);
#ifdef	USE_XMB
			    fontList[ix].xmbFontPair = fontPairs[jx];
			    fontList[ix].xmbFid = XCreateFontSet(display,fontPairs[jx],&missing_char_list,&missing_char_count,&def_string);
#else
			    fontList[ix].font16 = strchr(fontPairs[jx], ',') + 1;
			    fontList[ix].fid16 = XLoadFont (display, fontList[ix].font16);
#endif
			    return fontList[ix].fid;
				}
			}
	    }
/*	printf("Font %s not found!\n",font_name);*/
    return (XLoadFont (display, font_name));
}

/*
** XFontStruct tisLoadQueryFont (display, font_name)
*/
XFontStruct *tisLoadQueryFont (display, font_name)
Display *display;
char *font_name;
{
    return XQueryFont (display, tisLoadFont (display, font_name));
}

/*
**             tisDrawString (display, window, gc, x, y, text, textlen)
*/
tisDrawString (display, window, gc, x, y, text, textlen)
Display *display;
Window window;
GC gc;
int x;
int y;
char *text;
int textlen;
{
    int ix;
    int  segx;
    char *startseg;
    int  seglen;
    GC	gc16;
    XGCValues gcv;
    int mask16 = GCFont;
    static char converttext [400]; /* stuff converted characters here */
    char *convertp;

    /* try to find font fid in the font list */
    for (ix = 0; ix < MAXFONTS && fontList[ix].font8; ix++)
    	{
		if (fontList[ix].fid == gc->values.font)
			{
#ifdef	USE_XMB
			gc16 = XCreateGC(display,window,0,0);
			XCopyGC(display,gc,0x7FFFFFFF,gc16);
			XmbDrawString(display,window,fontList[ix].xmbFid,gc16,x,y,text,textlen);
			XFreeGC(display,gc16);
			return;
#else
	    	/* found it, display the string segments */
		    gcv.font = fontList[ix].fid16;
			gc16 = XCreateGC (display, window, mask16, &gcv);
		    startseg = text;
		    segx = x;
		    while (startseg < text+textlen)
	    		{
				seglen = 0;
				if (*startseg & 0x80) /* 16 bit segment */
					{
		    		convertp = converttext;
		    		while ((*(startseg+seglen) & 0x80) && (startseg+seglen < text+textlen))
		    			{
						/* convert characters for DrawString16 */
						*convertp++ = (*(startseg+seglen) & 0x7f);
						seglen++;
						*convertp++ = (*(startseg+seglen) & 0x7f);
						seglen++;
		    			}
		    		XDrawString16 (display, window, gc16, segx, y, (XChar2b*)converttext, seglen/2);
					segx += XTextWidth16 (XQueryFont (display, fontList[ix].fid16), (XChar2b*)converttext, seglen/2);
					}
				else
					{
		    		while (((*(startseg+seglen) & 0x80) == 0) && (startseg+seglen < text+textlen))
					seglen++;
		    		XDrawString (display, window, gc, segx, y, startseg, seglen);
		    		segx += XTextWidth (XQueryFont (display, fontList[ix].fid), startseg, seglen);
					}
				startseg += seglen;
	    		}
			return;
#endif
			}
	    }
    /* we didn't find a match, just call XDrawString */
    XDrawString (display, window, gc, x, y, text, textlen);
}

/*
** int          tisTextWidth (font_struct, text, text_len)
*/
int tisTextWidth (font_struct, text, textlen)
XFontStruct *font_struct;
char *text;
int textlen;
{
    int ix;
    int  returnsize;
    char *startseg;
    int  seglen;
    static char converttext [400]; /* stuff converted characters here */
    char *convertp;
#ifdef	USE_XMB
	XRectangle	len1,len2;
#endif

    /* try to find font fid in the font list */
    for (ix = 0; ix < MAXFONTS && fontList[ix].font8; ix++)
    	{
		if (fontList[ix].fid == font_struct->fid)
			{
#ifdef	USE_XMB
			XmbTextExtents(fontList[ix].xmbFid,text,textlen,&len1,&len2);
			returnsize = len1.width;
#else
	    	/* found it, size the string segments */
	    	returnsize = 0;
	    	while (startseg < text+textlen)
	    		{
				seglen = 0;
				if (*startseg & 0x80) /* 16 bit segment */
					{
		    		convertp = converttext;
		    		while ((*(startseg+seglen) & 0x80) && (startseg+seglen < text+textlen))
		    			{
						/* convert characters for DrawString16 */
						*convertp++ = (*(startseg+seglen) & 0x7f);
						seglen++;
						*convertp++ = (*(startseg+seglen) & 0x7f);
						seglen++;
		    			}
		    		returnsize += XTextWidth16 (XQueryFont (savedDisplay, fontList[ix].fid16), (XChar2b*)converttext, seglen/2);
					}
				else
					{
		    		while (((*(startseg+seglen) & 0x80) == 0) && (startseg+seglen < text+textlen))
					seglen++;
					returnsize += XTextWidth (font_struct, startseg, seglen);
					}
				startseg += seglen;
	    		}
#endif
			return returnsize;
			}
    	}
    /* we didn't find a match, just call XTextWidth */
    return XTextWidth (font_struct, text, textlen);
}

/*
**             tisDrawImageString (display, window, gc, x, y, text, textlen)
*/
tisDrawImageString (display, window, gc, x, y, text, textlen)
Display *display;
Window window;
GC gc;
int x;
int y;
char *text;
int textlen;
{
    int ix;
    int  segx;
    char *startseg;
    int  seglen;
    GC	gc16;
    XGCValues gcv;
    int mask16 = GCFont;
    static char converttext [400]; /* stuff converted characters here */
    char *convertp;

    /* try to find font fid in the font list */
    for (ix = 0; ix < MAXFONTS && fontList[ix].font8; ix++)
    	{
		if (fontList[ix].fid == gc->values.font)
			{
#ifdef	USE_XMB
			gc16 = XCreateGC(display,window,0,0);
			XCopyGC(display,gc,0x7FFFFFFF,gc16);
			XmbDrawImageString(display,window,fontList[ix].xmbFid,gc16,x,y,text,textlen);
			XFreeGC(display,gc16);
			return;
#else
	    	/* found it, display the string segments */
	    	gcv.font = fontList[ix].fid16;
	    	gc16 = XCreateGC (display, window, mask16, &gcv);
	    	startseg = text;
	    	segx = x;
	    	while (startseg < text+textlen)
	    		{
				seglen = 0;
				if (*startseg & 0x80) /* 16 bit segment */
					{
		    		convertp = converttext;
		    		while ((*(startseg+seglen) & 0x80) && (startseg+seglen < text+textlen))
		    			{
						/* convert characters for DrawString16 */
						*convertp++ = (*(startseg+seglen) & 0x7f);
						seglen++;
						*convertp++ = (*(startseg+seglen) & 0x7f);
						seglen++;
		    			}
					XDrawImageString16 (display, window, gc16, segx, y, (XChar2b*)converttext, seglen/2);
					}
				else
					{
		    		while (((*(startseg+seglen) & 0x80) == 0) && (startseg+seglen < text+textlen))
					seglen++;
		    		XDrawImageString (display, window, gc, segx, y, startseg, seglen);
					segx += XTextWidth (XQueryFont (display, fontList[ix].fid), startseg, seglen);
					}
				startseg += seglen;
	    		}
			return;
#endif
			}
		}
    /* we didn't find a match, just call XDrawString */
    XDrawImageString (display, window, gc, x, y, text, textlen);
}
