/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		TOKEN SUPPORT FOR 64 BIT ARCHITECTURES		*/
/*								*/
/*	This is a set of routines to provide translation between*/
/*	32 bit 'token's and the machine's memory pointers.	*/
/*								*/
/*	Author: Grant Irani					*/
/*								*/
/****************************************************************/

#include <tokenSupport.h>

/******
	Space for the real pointers, kept in an array.
******/

void *dmTokenPointers [MAXTOKENS];	/* Maximum tokens */

long	totalProbes = 0;
long	totalReused = 0;
long	totalPtrs = 0;

/******
	Convert a pointer to a token, the hard way.
******/

int dmTokenFromPointer64 (void *ptr)
{
    int limit;
    unsigned int srchIx = (unsigned int)(((long)ptr / (long)2) & (MAXTOKENS-1));

    /* if NULL comes in, send NULL out */
    if (ptr == 0)
	return 0;

    for (limit = 0; limit < MAXTOKENS; limit ++)
    {
	if (dmTokenPointers [srchIx] == 0)
	{
	    dmTokenPointers [srchIx] = ptr;
	    totalPtrs++;
	    return (srchIx + 500000);
	}
	else if (dmTokenPointers [srchIx] == ptr)
	{
	    totalReused++;
	    return (srchIx + 500000);
	}
	else
	{
	    totalProbes++;
	    srchIx = (srchIx + 1) % MAXTOKENS;
	}
    }
    /* no room at the Inn, bail out */
    printf ("Out of token conversion space.\nPlease contact program support.\n");
    exit (1);
}

void dmTokenFree (void *ptr)
{
    int limit;
    unsigned int srchIx = (unsigned int)(((long)ptr / (long)16) % MAXTOKENS);

    if (ptr == 0)
	return;
    for (limit = 0; limit < MAXTOKENS; limit ++)
    {
	if (dmTokenPointers [srchIx] == 0)
	{
	    return;
	}
	if (dmTokenPointers [srchIx] == ptr)
	{
	    dmTokenPointers [srchIx] = 0;
	    return;
	}
	else
	{
	    srchIx = (srchIx + 1) % MAXTOKENS;
	}
    }
}

void dmTokenFlush (void)
{
    totalProbes = 0;
    totalReused = 0;
    totalPtrs = 0;

    memset (dmTokenPointers, 0, MAXTOKENS*sizeof(void *));
}
