/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
Copyright (c) 1992 Thomas A. Limoncelli, tom_limoncelli@warren.mentorg.com

Permission to use, copy, modify, distribute, and sell this software and
its documentation for any purpose is hereby granted without fee,
provided that (i) the above copyright notices and this permission
notice appear in all copies of the software and related documentation,
and (ii) the name of Mentor Graphics Corp or any company or
organization that I am associated with may not be used in any
advertising or publicity relating to the software without the specific,
prior written permission of that company or orgaization

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

IN NO EVENT SHALL THOMAS A. LIMONCELLI OR MENTOR GRAPHICS CORP BE
LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES
OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON
ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
OR PERFORMANCE OF THIS SOFTWARE.
*/
/******************************************************************************
  modem.c


*******************************************************************************/
/****************************************************************/
/* USER CONFIGURABLE OPTIONS: */

/* this should be "#define" if you use SunOS, or "#undef" if you
** use HPUX.  This controls the name of LOCKDIR and if getpriority()
** is used.  I'm sure more needs to be done, but that's a small start.
*/
#define REAL_OS

#ifdef REAL_OS
#define LOCKDIR	"/var/spool/uucp"
#else
#define LOCKDIR "/usr/spool/locks"
/* That may not be correct */
#endif

#define  MAX_ERROR      1000
/* not talking to the modem correctly?  Try mucking with
the grabmodem() routine.  */

/* END OF USER CONFIGURABLE OPTIONS */
/****************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/termios.h>

#ifdef REAL_OS
#include <sys/time.h>		/* required for <sys/resource.h> */
#include <sys/resource.h>	/* required for getpriority() */
#endif

/* ASCII constants */
#define STX (2)
#define EOT (4)
#define ACK (6)
#define LF (10)
#define CR (13)
#define NAK (21)
#define ESC (27)
#define RS (30)

#define MAX_PACKET	(10000)	/* we'll never get a packet this big */

char *lockname = NULL;

/* print a string without worrying about unprintable charactors */
void 
safeprint(str)
char           *str;
{
	while (*str) {
        	if (isgraph(*str))
			(void) fprintf(stderr, "%c", *str);
		else {
			switch (*str) {
			case LF:
				(void) fprintf(stderr, "\\n");
				break;
			case CR:
				(void) fprintf(stderr, "\\r");
				break;
			case 32:
				(void) fprintf(stderr, "\\s");
				break;
			default:
				(void) fprintf(stderr, "\\%d", *str);
				break;
			}
		}
		str++;
	}
	fflush(stdout);
}


/* open the modem.  You should have done a lockmodem() first */
int grabmodem(dev)
char           *dev;
{
	struct termios  ti;
	int             modem;

	errno = 0;
	/* printf("in grab modem\n"); */
	modem = open(dev, O_RDWR, 777);
	/* printf("%d\n", errno); */
	if (errno) {
		/* printf("#MODOPEN modem can't be opened\n"); */
		return 0;
	}

	/* set tty params to 300bps, even parity, 7-1-e */
	errno = 0;
	/* printf("get the modem params\n"); */
	ioctl(modem, TCGETS, &ti);
	if (errno) {
		close(modem);
		return 0;
	}
	/* input mode */
	ti.c_iflag |= IGNBRK;   /* ignore breaks */
	/* ti.c_iflag |= IGNPAR; *//* ignore parity */
	ti.c_iflag &= ~INPCK;   /* ignore parity errors */
	ti.c_iflag |= ISTRIP;   /* strip 8th bit */
	ti.c_iflag &= ~INLCR;   /* don't cr->nl */
	ti.c_iflag &= ~ICRNL;   /* don't cr->nl */
	ti.c_iflag &= ~IGNCR;   /* don't ignore cr */
	/* ti.c_iflag &= ~IXON; *//* don't do xon */
	/* ti.c_iflag &= ~IXOFF; *//* don't do xoff */

        /* output mode */
	ti.c_oflag &= ~OPOST;   /* don't post-process */

	ti.c_cflag &= ~CBAUD;
	ti.c_cflag |= B300;	/* baud=300 */
	/* ti.c_cflag |= B1200;	/* baud=1200 */

	ti.c_cflag &= ~CSIZE;   /* 7-bit bytes */
	ti.c_cflag |= CS7;
/*	ti.c_cflag |= CS8; */

	ti.c_cflag &= ~CSTOPB;  /* one stop bit */
/*	ti.c_cflag |= CSTOPB;  /* two stop bit */

	ti.c_cflag |= PARENB;  	/* parity */
	ti.c_cflag &= ~PARODD;  /* even parity */
/*	ti.c_cflag |= PARODD;  /* odd parity */

 	 ti.c_cflag |= HUPCL;    /* hang up on last close */
	/* ti.c_cflag |= CRTSCTS; *//* hardware handshaking */
         ti.c_cflag |= CLOCAL;   /* allow multiple calls with one open and close */

	/* control */
	ti.c_cc[VMIN] = 0;  /* read() can get as few as 0 bytes */
	ti.c_cc[VTIME] = 50;    /* time out at 5 seconds no matter what */

        /* line discipline modes */
	ti.c_lflag &= ~ISIG;    /* disable signals */
	ti.c_lflag &= ~ICANON;  /* disable signals */
	ti.c_lflag &= ~ECHO;    /* don't echo */

	errno = 0;
	/* printf("set the modem params\n"); */
	ioctl(modem, TCSETS, &ti);
	if (errno) {
		close(modem);
		return 0;
	}
	return modem;
}

/* send data to the modem */
void send_data(fd, str)
int             fd;
char           *str;
{
	/* printf("Sending: :"); safeprint(str); printf(":\n", str); */
	write(fd, str, strlen(str));
}

/* wait for a particular string from the modem */
int match(mod, str, err) 
	int mod;
	char *str;  /* # of retries permitted */
	int err;
{
	int len;
	char c;

/* printf("MATCHING on :"); safeprint(str); printf(":\n", str); */

	while (1) {
		c = 0;
 /* printf("waiting for :%c:\n", *str);  */
		len = read(mod, &c, 1);
		if (len) {
 /* printf("got=%d:%c\n", c,c);  */
			if (c == *str) {
				str++;
			}
			if (!(*str)) break; /* matched all?  Exit loop */
		}
		if (!err--) {
			/* printf("NOT MATCHED\n"); */
			return 1;
		}
	}
/* printf("MATCHED\n"); */
	return 0;
}


void 
hangup_modem(fd)
int             fd;
{
	sleep(3);
	send_data(fd, "+++");  /* switch from data mode to command mode */
	sleep(3);
	send_data(fd, "ATH\r");/* hang up */
	sleep(1);
}

/* unlock the modem */
void unlockmodem(name)
char           *name;
{
	char lname[200];

	strcpy(lname, LOCKDIR);
	strcat(lname, "/LCK..");
	strcat(lname, 1 + rindex(name, '/'));

	/* printf("Unlocking modem.\n"); */

    (void)unlink(lname);
    return;
}

/* clean up and leave this program */
void bail_out(modem)
int modem;
{
	if (modem) {
		hangup_modem(modem);
	}
}

void close_modem(modem, dev)
int  modem;
char *dev;
{
   close(modem);
   unlockmodem(dev);
}

/* lock the modem OR DIE*/
int  lockmodem(dev, lkn)
char           *dev;
char           *lkn;
{
	int lock, pid;
	int failcnt = 3;
	char waitcnt = 0;
	char lname[200];

	strcpy(lname, LOCKDIR);
	strcat(lname, "/LCK..");
	strcat(lname, 1 + rindex(dev, '/'));

/* printf("Lockfile = %s\n", lname); */
	lkn = strdup(lname);
	while (failcnt--) {
		errno = 0;
		lock = open(lname, O_CREAT | O_WRONLY | O_EXCL, 0777);
		/* printf("%d\n", errno); */
		if (lock == -1) {
#ifdef REAL_OS
			/* printf("Modem is locked, attempting to steal.\n"); */
			/* locked, let's read the cookie in the lock */
			pid = 0;
			if ((lock = open(lname, O_RDONLY)) != -1) {
				(void)read(lock, &pid, sizeof(int) );
				/*printf("Device is locked by process %d\n", pid);*/
				close(lock);
			}
			/* printf("%d\n", errno); */
			/* printf("Lock = %d\n", lock); */
			if (pid < 3) {
				/* printf("#MODOPEN device is locked by pid < 3\n"); */
				bail_out(0);
				return 0;
			}
			/* see if the process still is alive */
			errno = 0;
			(void) getpriority(PRIO_PROCESS, pid);
			if (errno == ESRCH) {   /* lock process dead, let's go! */
				if (unlink(lname)) {
					/* printf("#MODOPEN Can't steal lock.\n"); */
					bail_out(0);
					return 0;
				} else {
					/* printf("Lock is stale, stealing!\n");*/
				}
			} else {
				/* printf("#MODOPEN Valid lock in the way.\n"); */
				bail_out(0);
				return 0;
			}
#else
			/* printf("#MODOPEN it's locked, I'm out of here!\n"); */
			bail_out(0);
			retrun 0;
#endif
		} else {
			/* lock opened, stuff and go */
			pid = getpid();
			write(lock, &pid, sizeof(int));
			close(lock);
			break;
		}
	}
	if (failcnt==-1) {
		/* printf("#MODOPEN Couldn't lock modem after many tries.\n");*/
		bail_out(0);
		return 0;
	}
	return 1;
}

/* Loop until you get a valid packet.  If you get a "message sequence" then
display it.  If you get an invalid pack DIE */

int  get_packet(fd, str)
int fd;
char *str;
{
	int max;
	char c;
	int len;
	char *buf = str;
	int err;

	*str = 0;
	err=10;	/* permit up to 10 message sequences or bad packets */
	while (err--) {
		/* printf("Getting packet\n"); */
		max = MAX_PACKET;
		while (read(fd, &c, 1) == 1) {
			if (c == LF) continue;	/* skip LF's */
			if (c == CR) {
				/* don't actually put CR in the string */
				break;
			}
			*buf++ = c;
			max--;
			if (!max) {
				/* packet was too long */
				/* printf("#PACKLEN packet was too long\n"); */
				return 0;
			}
		}
		*buf = 0;
		len =  buf - str;
		if (len) {
			/* printf("Got packet--length=%d\n", len); */
			if (isgraph(*str)) {
				/* printf("#GOTMESSEQ message sequece=:"); */
				safeprint(str);
				/* printf(":\n"); */
				return 1;		/* ddj 4/30  */
			} else {
				/* print packet contents */
				/* printf("packet is data=:"); safeprint(str); printf(":013:\n"); */
				return 1;
			}
		}
	}
	/* printf("#LONELY It's too quiet in here.  Disconnect.\n");  */
	return 0;
}

/****************************************************************************
  open_modem

******************************************************************************/
int  open_modem(dev)
   char *dev;
{
	char *lockname;	
	int  failcnt, modem;

       /* printf("in modem dev = %s \n", dev); */
       /* lock modem or die */
       if (!lockmodem( dev, lockname ) )
	   return 0;
	/* open modem or die */
	/* printf("opening modem\n"); */
	modem = grabmodem( dev );
	/* printf("back to main\n"); */
	if (!modem) {
	   bail_out(modem);
	   return 0;
        }
        return modem;
}

/*****************************************************************************
  connect_modem

******************************************************************************/
int  connect_modem(modem, dialstring, errormessage)
   int  modem;
   char dialstring[12];
   char errormessage[MAX_ERROR+1];
{
	char *pack = dialstring, c, accessnumber[20];
	int  failcnt;

	/* see if modem is awake or hangup; after 3 tries die */
	failcnt = 3;
	while (failcnt--) {
		send_data(modem, "AT\r");
		if (match(modem, "OK", 6)) {
			/* printf("No response.  Hang up and try again.\n"); */
			hangup_modem(modem);
			while (read(modem, &c, 1)) { /*DONOTHING*/ };
		} else break;
	}
	if (failcnt==-1) {
	    strcpy(errormessage, "Modem not responding.");
	    bail_out(modem);
	    return 0;
        }

/* printf("dialing\n"); */

	/* send the "A" of "ATDT" -- 
	   (AT) attention (D) dial the number using (T) the touch-tone method */
	do {
		send_data(modem, "A");
	} while (match(modem, "A", 2));
	/* send the rest of the dial string */
	sprintf( accessnumber, "TDT%s\r", dialstring );
	send_data(modem, accessnumber);
	(void) match(modem, accessnumber, 10);

	/* wait for the modem to connect */
	/* printf("waiting for CONNECT\n"); */
	if (match(modem, "CONNECT", 30)) {
		/* printf("#NOCONN no connect\n"); */
		sprintf(errormessage,"Dialing %s, pageing service not connecting.", dialstring);
		bail_out(modem);
		return 0;
	}

/* printf("Waiting for ID=\n"); */

	/* send a CR ever second until "ID=" comes back */
	failcnt = 10;
	/* CRs needed by PacTel   DJ */
	send_data(modem, "\r");
	/* end of DJ */
	while (failcnt--) {
		if (match(modem, "ID=", 4)) send_data(modem, "\r");
		else break;
	}
	if (failcnt==-1) { 
	   strcpy(errormessage,"Pageing service not responding with ID.");
	   bail_out(modem);
	   return 0;
        } 

        return modem;
}
