/*time_utility.h,v 1.1.1.1 1999/08/03 14:44:12 tis4000 Exp */

/*time_utility.h,v 1.1.1.1 1999/08/03 14:44:12 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*******************************************************************************************
*   time_utility.h
*
*   This utility contains procedures and functions which 
*
*   K.Grimes  February 1994
********************************************************************************************/

#include <time.h>
#include <stdio.h>
#include <string.h>

#define    FALSE              0
#define    TRUE               1
#define    TIMESTRING_SIZE    8

/********************************************************************************************
     function     :   construct_epoch

     input        :   seconds after the minute (0,59).
                      minutes after the hour (0,59).
                      hours since midnight (0,23).
                      day of the month (1,31) or julian days + 1.
                      months since january (0,11) or 0 for julian.
                      years since 1900.

     output       :   the epoch time of the input.
**********************************************************************************************/

time_t  construct_epoch(second, minute, hour, day, month, year)

int     second, minute, hour, day, month, year;
{

struct tm    thetime;


     thetime.tm_sec = second;
     thetime.tm_min = minute;
     thetime.tm_hour = hour;
     thetime.tm_mday = day;
     thetime.tm_mon = month;
     thetime.tm_year = year;
     thetime.tm_wday = 0;
     thetime.tm_yday = 0;
     thetime.tm_isdst = 0;
     thetime.tm_zone = " ";
     thetime.tm_gmtoff = 0;

     return mktime(&thetime);
}


int  parse_date(datestring, themonth, theday, theyear)

char  datestring[11];
int   *themonth, *theday, *theyear;
{

char         *datepiece, *temp = NULL;
static int   daytab[2][13] = { {0,31,28,31,30,31,30,31,31,30,31,30,31},
                               {0,31,29,31,30,31,30,31,31,30,31,30,31} };
int          leap, valid_date = FALSE, month, day, year;

   /* Determine if the correct date format has been followed while parsing the date string */ 

   if ( (datepiece = strtok(datestring, "/")) != NULL){

	 month = atoi(datepiece);
         if ( (month >= 1 && month <= 12) ) 
	      valid_date = TRUE;

         if ( (datepiece = strtok(temp, "/")) != NULL ) {
	     day = atoi(datepiece);

             if ( (datepiece = strtok(temp, "\0")) != NULL) {
                 
		 if ( (year = atoi(datepiece)) >= 1990) { 
     
                     leap = year%4 == 0 && year%100 != 0 || year%400 == 0;
                     if (day > 0 && day <= daytab[leap][month])
                        valid_date = TRUE;
                     else
	                valid_date = FALSE; 
		 }
             }
         }
    }
    if (valid_date) {
       *theday = day;
       *themonth = month;
       *theyear = year;
    }
    return valid_date;
}

int parse_time_numeric( time, thehour, theminute)
int  time;
int  *thehour;
int  *theminute;
{
   char   *timestring = (char *)malloc(sizeof(time)+1); 
   int    valid_time = FALSE;

   sprintf(timestring, "%d", time);

   if (strlen(timestring) == 4) {
      *thehour = timestring[0]*10 + timestring[1];
      *theminute = timestring[2]*10 + timestring[3];
      if (*thehour >= 0 && *thehour <= 23 && *theminute >= 0 && *theminute <= 59)
         valid_time = TRUE;
   }
   else if (strlen(timestring) == 3){
      *thehour = timestring[0];
      *theminute = timestring[1]*10 + timestring[2];
      if (*thehour >= 0 && *thehour <= 23 && *theminute >= 0 && *theminute <= 59)
         valid_time = TRUE;
   }

   return valid_time;
}



int parse_time_string(timestring, thehour, theminute, thesecond)

char  timestring[10];
int   *thehour, *theminute, *thesecond;
{
char  *timepiece, *temp = NULL;
int   digit, hour, minute,second, valid_time = FALSE;

   /* Determine if the correct time format has been followed while parsing the time string */

   if ( (timepiece = strtok(timestring, ":")) != NULL ) {
	 hour = atoi(timepiece);
	 if (hour == 0) {

            for (digit = TRUE; (digit == TRUE) && (*timepiece != '\0'); timepiece++)
	       digit = isdigit(*timepiece);
            if (digit != FALSE)
		valid_time = TRUE;
         }
	 else if (hour > 0 && hour <= 24)
	     valid_time = TRUE; 

        if ( (timepiece = strtok(temp, ":")) != NULL){
              minute = atoi(timepiece); 
	      if (minute == 0) {

                 for (digit = TRUE; (digit == TRUE) && (*timepiece != '\0'); timepiece++)
		     digit = isdigit(*timepiece);
                 if (!digit)
	 	    valid_time = FALSE; 
              }
	      else if (minute < 0 || minute >= 60)
		  valid_time = FALSE;

              if( (timepiece = strtok(temp, "\0")) != NULL){
		 second = atoi(timepiece);
		 if (second == 0) {

		     for (digit = TRUE; (digit == TRUE) &&(*timepiece != '\0'); timepiece++)
			  digit = isdigit(*timepiece);
                     if (!digit)
			 valid_time = FALSE;
                 }
		 else if (second < 0 || minute >= 60)
		     valid_time = FALSE;
              }
         }
    }
    if (valid_time) {
       *thehour = hour;
       *theminute = minute;
       *thesecond = second;
    }
    return valid_time;
}


/*************************************************************************************************
* function    :  get_time_string 
*
* description :  returns the current time in hh:mm:ss format.
**************************************************************************************************/

void  get_time_string(timestring)

char         timestring[TIMESTRING_SIZE+1];
{
char         hour[3], minute[3], second[3];
time_t       epoch;
struct tm    *timestamp; 

   epoch = time(NULL);
   timestamp = localtime(&epoch);

   if (timestamp->tm_hour < 10)
       sprintf(hour,"0%d", timestamp->tm_hour);
   else
       sprintf(hour, "%d", timestamp->tm_hour);

   if (timestamp->tm_min < 10)
       sprintf(minute, "0%d", timestamp->tm_min);
   else
       sprintf(minute, "%d", timestamp->tm_min);

   if (timestamp->tm_sec < 10)
       sprintf(second, "0%d", timestamp->tm_sec);
   else
       sprintf(second, "%d", timestamp->tm_sec);

       
   sprintf(timestring, "%s:%s:%s", hour, minute, second);

}
