/*procstat.c,v 1.1.1.1 1999/08/03 14:44:17 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/****************************************************************************
 *
 *  MODULE NAME            : procstat.c 
 *  AUTHOR                 : emsteel
 *  DATE                   : 28 Feb 94
 *                                     
 *
 *  DESCRIPTION: print out process stat gathered by point types
 *               each point type uses its type as an index into an
 *               array process_cnts every time it is processed. The
 *               point must be on scan to register
 *       
 *
****************************************************************************/

/****************************************************************************
 *
 * Programmer's Notes:
 *
 *    1 - initProcessStats is called from .../src/db/iocInit and lives in
 *        iocInit, it allocates process_cnts which is declared global in
 *        iocInit.c
 *    2 - args to taskSpawn are defined in .../share/epicsH/task_params.h
 *    3 - vxTicksPerSecond defined in .../share/epicsH/dbDefs.h
 *    4 - process_stats, checked by each point type before the process count
 *        for that point type is incremented
 *
 *
****************************************************************************/


#include <vxWorks.h>
#include <types.h>
#include <stdioLib.h>
#include <lstLib.h>
#include <taskLib.h>
#include <logLib.h>
#include <string.h>

#include "dbDefs.h"
#include "dbRecType.h"
#include "task_params.h"

extern long *process_cnts;
int process_stats = 0;
static int psTaskId;


/****************************************************************************
 *
 *  MODULE NAME            : PrintProcessStats 
 *  AUTHOR                 : emsteel
 *  DATE                   : 28 Feb 94
 *
 *          ARGUMENT(S)    : num_secs - cycle time to print statistics
 *
 *
 *  DESCRIPTION: task that prints the process stats from info placed in
 *               process_cnts. zero process_cnts at start of task and each
 *               cycle time
 *               
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 05 Jul 94
 *  REVISION:	add nsecs = num_secs and set to 1 if <= 0, prevent divide
 *		by 0 later
 *
****************************************************************************/

static void PrintProcessStats (num_secs)
   int   num_secs;
{
   int     num_delay_ticks;
   int     i, j;
   long    tot_process;
   int     ave_process;
   int     rem;
   int     nsecs = num_secs;

   /* default to 1, save a possible divide by 0 later */
   if (nsecs <= 0)
      nsecs = 1;
   num_delay_ticks = nsecs * vxTicksPerSecond;
   rem = dbRecType -> number % 3;
   printf ("PrintProcessStats with cycle time of %d\n", nsecs);
   for (;;)
   {
      bzero ((char *)process_cnts, sizeof (long) * dbRecType -> number);
         /* tell points to start counting */
      process_stats = 1;
         /* count for this long */
      taskDelay (num_delay_ticks);
         /* stop counting */
      process_stats = 0;
      tot_process = 0;
      ave_process = 0;
         /* print counting results */
      for (i = 0; i < dbRecType -> number; i += 3)
      {
	 if ((dbRecType -> number - i) == rem)
         {
            for (j = i; j < dbRecType -> number; j++)
            {
               if (dbRecType -> papName[j] == NULL)
                  continue;
               printf ("%15s - %-5ld    ", dbRecType -> papName[j], process_cnts[j]);
               tot_process += process_cnts[j];
            }
            printf ("\n");
         }
         else
         {
            for (j = i; j < (i + 3); j++)
            {
               if (dbRecType -> papName[j] == NULL)
                  continue;
               printf ("%15s - %-5ld    ", dbRecType -> papName[j], process_cnts[j]);
               tot_process += process_cnts[j];
            }
            printf ("\n");
         }
      }
      printf ("Total process %ld\n", tot_process);
      ave_process = tot_process / nsecs;
      printf ("Averge process %d/sec\n", ave_process);
   }
      
   
} /* end func PrintProcessStats () */


/****************************************************************************
 *
 *  MODULE NAME            : StartPrStat 
 *  AUTHOR                 : emsteel
 *  DATE                   : 28 Feb 94
 *
 *          ARGUMENT(S)    : num_secs - cycle time to print statistics
 *
 *  DESCRIPTION: start the task to print the stats
 *
****************************************************************************/

void StartPrStat (num_secs)
   int   num_secs;
{
   if (process_cnts == NULL)
   {
       printf ("StartPrStat: there is no stats buffer\n");
      return;
   }
   if ((psTaskId = taskSpawn (PROCESS_STATS_NAME, PROCESS_STATS_PRI,
                              PROCESS_STATS_OPT, PROCESS_STATS_STACK,
                              (FUNCPTR)PrintProcessStats, num_secs, 0, 0,
			      0, 0, 0, 0, 0, 0, 0)) == ERROR)
   {
      printf ("failed to start PrintProcessStats (%d)\n", errno);
   }

   return;

} /* end func StartPrStat () */



/****************************************************************************
 *
 *  MODULE NAME            : StopPrStat 
 *  AUTHOR                 : emsteel
 *  DATE                   : 28 Feb 94
 *
 *
 *  DESCRIPTION: kill the task printing stats and zero out process_cnts
 *
****************************************************************************/

void StopPrStat (void)
{
   process_stats = 0;
   if (taskDelete (psTaskId) == ERROR)
      printf ("failed to stop PrintProcessStats (%d)\n", errno);
   return;

} /* end func StopPrStat () */




