/*recAcc.c,v 1.1.1.1 1999/08/03 14:44:18 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/  
/* recAcc.c - Record Support Routines for Accumulator records */
/*
 *      Author: William H. Szostak
 *      Date:   8-04-93
 *
 * Modification Log
 * ----------------
 * .01  01-03-94	ems	ack = 0 in init_record
 *				in process, move check for init to 
 *				 first thing done
 * .02  01-05-94	ems	in monitor,check for VALID_ALARM
 *				before set ACK
 * .03  03-01-94	ems	added support for process stats
 *				removed ack = 0
 * .04	03-29-94	ems	make forward func def agree with actual
 *				func def
 * .05	05-11-94	ems	rewrite init_record, add INI2 to redund list
 *				delete special() and convert(), not used
 *				get_control_double, if field is HYST then
 *				high is hopr-lopr and low is 0
 *				fix alarm() to use hyst properly
 * .06	06-13-94	ems	added support for maximize_severity msev
 *				move call to tsLocalTime inside on scan
 *				moved call to alarm() in side on scan like all
 *				other points
 * .07	07-07-94	ems	in init_record
 *					only calc freezerate if TIMED, check if
 *					aftm is <= 0, can cause divide/0 error
 *					dont init if inp is != DB_LINK
 *					if ASYNC, dont init if afis != DB_LINK
 *				in process
 *					check if aftm <= 0, set VALID/PARAMTER 
 *					alarm, if ASYNC and afis != DB_LINK
 *					set VALID/PARAMETER alarm
 * .08	08-09-94	ems	added field types to get_graphic_double and
 *				get_control_double, removed init message, not needed
 * .09	08-17-94	ems	monitor(),rm DBE_LOG from ack
 *				only archive val if tenb (trending) is set
 * .10	09-16-94	ems	major changes to numerous to mention
 * .11	10-20-94	ems	chg time to latm (last alarm time) added 
 *				added lptm (last process time)
 * .12	12-13-94	ems	made update_crc local to funcs and passed as arg 
 *				rather than static global
 * .13	01-11-95	ems	test_rgrp_reset, do not check if year < TS_EPOCH_YEAR
 *				not needed, also get leap value from tsDetail struct
 * .14	01-13-95	ems	added MLST to redundancy, init_record init latm and lptm
 * .15	02-07-95	ems	make LSAV based on EPICS secs rather than clock ticks
 *				init LSAV with an offset based on the phase field to
 *				load level the amount of running values sent up to 
 *				dbClient, esp critical on a system with many acc'c all
 *				set to save at the same time
 * .16	02-08-95	ems	monitor()
 *				    if we post VAL, also post RTOT, RAVG, RMIN, RMAX, RSMP, ATOT
 *					if not then if VAL != TOTAL, RTOT and RSMP may be 
 *					be changing, so we may need to post RTOT and RSMP, since
 *					we do not have a seperate MDEL for RTOT, use 
 *					(HOPR - LOPR) * 0.0005, YES THIS IS A FUDGE
 *				    the same applies to archiving
 * .17	03-15-95	ems	monitor(), do not set ACK if sevr is EVENT_ALARM
 * .18	03-12-95	ems	1) LSAV calc chg.
 *				use the scan rate as a deadband factor when computing the next
 *				LSAV. Due to fact that block will not scna at esactly the scan
 *				rate, need to compensate
 *				2) chg to calcs for RTOT and RAVG
 *				RTOT, new calc is:
 *					LVAL = (lval * slope) * (time_adj / time_mul)
 *				RAVG, new calc is
 *					LVAL = (lval * slope) * (time_mul / time_adj)
 * .19	03-22-95	ems	calc lsav with delta clock ticks in secs not scan rate because
 *				the block may be passive, also if first pass FPPS and in
 *				delta mode DMDF, do not set RMIN and RMAX, since use time
 *				in their calcs, we may have an invalid first delta time
 *				init_record(), test for missed freeze if MASTER and we
 *				are going to init, if done earlier, we can freeze, reset 
 *				running totals if selected, but if we have to init again
 *				the running totals will be reset by the saved values but
 *				we would not freeze and would not reset the running values
 *				giving bad initial running values
 * .20	04-21-95	ems	process(), do no calc next freeze time unless are TIMED
 * .21	05-04-95	ems	if we freeze post and trend even if value has not chg'd
 *				this keeps trending happy
 * .22	08-10-95	ems	removed unsigned long casts from lsav in init_record, 
 *				float casts for time_adj in process and unsigned long cast on
 *				time_adj in process, the last one was causing a problem on 68000
 *				lsav was never being updated because the cast was causing a
 *				a large number and the code never enter thewhile loop to update
 *				lsav, therefore the running values were being updated every scan
 *				instead of every srvt time
 * .23	09-11-95	ems	monitor(), archive when stat/sevr chg's also
 * .24	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 * .25	02-07-96	ems	remove return from HIHI and LOLO alarms, so will always check 
 *				    all alarm sevrs
 * .26	04-04-96	ems	added code for alarm delay
 * .27	06-19-96	ems	added code for alarm disable ADIS
 * .28	07-01-96	ems	process(), if reset when freeze and save runtime, force a
 *				save but do not mess with LSAV the next freeze time
 */

#include	<vxWorks.h>
#include	<inetLib.h>
#include	<stdlib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<string.h>
#include	<lstLib.h>
#include	<logLib.h>
#include	<tickLib.h>
#include	<ctype.h>

#include	"alarm.h"
#include	"cvtTable.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"recSup.h"
#include	"special.h"
#include	"rec/accumRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"accredfldidx.h"
#include	"restable.h"
#include	"recGbl.h"
#include	"okToUse.h"
#include	"db_addr.h"
#include	"dbEvent.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/

#define report NULL
#define initialize NULL
static long init_record();
static long process();
static long special ();
static long get_value();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units();
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double();
static long get_control_double();
static long get_alarm_double();

struct rset accumRSET = {

	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
	};


/* The following definitions are for redundancy */
 
static struct fldDes	**mylist = 0;

static	char *namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"RSMP",
		"RTOT",
		"RAVG",
		"RMIN",
		"RMAX",
		"FTIM",
		"FSMP",
		"FTOT",
		"FAVG",
		"FMIN",
		"FMAX",
		"LALM",
		"RSEC",
		"INI2",
		"LTCK",
		"LMOD",
		"STOT",
		"SSMP",
		"SMIN",
		"SMAX",
		"SSTT",
		"LARS",
		"ARSF",
		"FSTA",
		"FSEV",
		"NFSA",
		"NFSV",
		"LSAV",
		"FSEC",
		"ATOT",
		"LASF",
		"TALC",
		"NFTM",
		"LATM",
		"MLST",
		"LMTO",
		"HHLL",
		"HILO",
		"RTCT",
		"LARF",
		0
		};
struct accred
{
    unsigned short	stat;
    unsigned short	sevr;
    double		val;
    double		rsmp;
    double		rtot;
    double		ravg;
    double		rmin;
    double		rmax;
    TS_STAMP		ftim;
    double		fsmp;
    double		ftot;
    double		favg;
    double		fmin;
    double		fmax;
    double		lalm;
    short		ini2;
    unsigned long	ltck;
    unsigned short	lmod;
    double		stot;
    double		ssmp;
    double		smin;
    double		smax;
    double		sstt;
    double		lars;
    unsigned short	arsf;
    unsigned short	fsta;
    unsigned short	fsev;
    unsigned short	nfsa;
    unsigned short	nfsv;
    unsigned long	lsav;
    double		atot;
    double		lasf;
    short		talc;
    TS_STAMP		nftm;
    TS_STAMP		latm;
    unsigned long	rsec;
    char		lrst[28];
    double		mlst;
    double		lmto;
    short		hhll;
    short		hilo;
    short		rtct;
    unsigned long	larf;
};

/* structure to define the type of asynchronous group reset */

struct rgrp
{
    short		type;		/* see def's below */
    int			day_of_week;	/* 1 - 7, Sun - Sat */
    int			month;		/* 1 - 12, Jan - Dec */
    int			day_of_month;	/* 1 - 31, 32 = last day on month */
    int			hour;		/* 0 - 23 */
    int			minute;		/* 0 - 59 */
};

static struct rgrp *rgrps[256];

static unsigned	short	type;

/* the following definitions must match those in choiceRec.ascii */

#define	TIMED		0
#define	ASYNCHRONOUS	1

#define	HOURS		0
#define	MINUTES		1

#define	TOTAL		0
#define	AVERAGE		1
#define	MINIMUM		2
#define	MAXIMUM		3

/* defines for type of time factor to calc running values */

#define	ACC_NONE	0
#define	ACC_SECOND	1
#define	ACC_MINUTE	2
#define	ACC_HOUR	3

/* defines for the type of timed asynchronous group reset */

#define	DAY_OF_WEEK_RESET	1
#define	DAY_OF_WEEK_TIME_RESET	2
#define	DAY_OF_MONTH_RESET	3
#define	DAY_OF_MONTH_TIME_RESET	4
#define	MONTH_RESET		5
#define	MONTH_TIME_RESET	6
#define	TIME_RESET		7
#define	DAILY_RESET		8	/* default reset */
#define	WEEKLY_RESET		9	/* default reset */
#define	MONTHLY_RESET		10	/* default reset */
#define	LAST_RESET_TYPE		(MONTHLY_RESET + 1)


extern unsigned short DBC_CRCRecord ();	/* see .../share/src/redund/dbCrcSubs.c */
static void alarm ();
static void firstFreeze ();
static void free_rgrps ();
static void freeze ();
int ld_rgrp ();
static void monitor ();
static int parse_rgrp ();
static void post_and_redund ();
int read_rgrps ();
int rgrp_status ();
static int test_rgrp_reset ();
int set_rgrp_defaults ();
static void save_accum ();


/*------------------------------------------------------------------------*/

static long init_record (pacc)
    struct accumRecord	*pacc;
{
    unsigned	short	i;
    long		status = 0;
    long		options = 0;
    long		nRequest = 1;
    char		*pstr;
    char		timebuf [32];
    char		*ptimebuf = timebuf;
    double		frzratesecs = 0;
    int			ini2_flag = 0;
    int			update_crc;
    int			did_a_freeze = 0;
    int			did_a_reset = 0;
    TS_STAMP		timenow;
    TS_STAMP		next_freeze;
    TS_STAMP		reset_stamp;
    struct accred	old_acc;

    if (!(mylist)) 
    {
	for (type = 1; type < 100; type++) 
	{
	    pstr = GET_PRECTYPE (type);
	    if (strncmp (pstr, "accum", 5) == 0) 
		break;
	}
	mylist = redFldAlloc (type, NUMACCFLDS);

	if (!mylist)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)pacc, 
			       "acc, init_record, mylist failed");
	    return (S_rec_Alloc);
	}

	for (i = 0; i < NUMACCFLDS; i++) 
	    redFldAdd (type, i, namelist [i]);
    }

    /* if no timed resets have been set, init 
       to default values
    */

    if (rgrps[0] == 0)
	set_rgrp_defaults ();

    update_crc = 0;
    tsLocalTime(&timenow);
    if (timenow.secPastEpoch < 110000) 
    {
	logMsg ("Accumulators for %s cannot initialize until Time is set in IOC\n",
	         (int)pacc -> name,0,0,0,0,0);
	return (0);
    }

    if (pacc -> inp.type != DB_LINK)
    {
	recGblRecordError (S_db_badField, (struct dbCommon *)pacc, "acc, init_record, input not link");
	return (S_db_badField);
    }

    /* if the reset group number is out of range */

    if ( (pacc -> rgrp < 0) || (pacc -> rgrp > 255) )
    {
	recGblRecordError (S_db_badField, (struct dbCommon *)pacc,
			   "acc, init_record, rgrp range");
	return (S_db_badField);
    }

    pacc->val = 0;

    /* save current accum values, to test later for redundant update */
    save_accum (pacc, &old_acc);

    /* init current running totals from what was saved in the database */
    pacc->rsmp = pacc -> ssmp;
    pacc->rtot = pacc -> stot;
    pacc -> atot = pacc -> sstt;
    /* average is total / num samples */
    if (pacc -> atot && pacc -> rsmp)
    	pacc -> ravg = pacc -> atot / pacc -> rsmp;
    else
	pacc -> ravg = 0;
    pacc->rmin = pacc -> smin;
    pacc->rmax = pacc -> smax;

    /* NOTE: frozen values fsmp,ftot,favg,fmin,fmax have values saved in 
	     database so do not need to be init'd
	     init last freeze time from epics seconds saved in database
	     and freeze time string
    */
    pacc -> ftim.secPastEpoch = pacc -> fsec;
    pacc -> ftim.nsec = 0;
    if (pacc -> ftim.secPastEpoch)
	tsStampToText (&pacc -> ftim, TS_TEXT_MMDDYY, pacc -> fts0);
    else
	pacc -> fts0[0] = 0;

    /* convert the reset time into a timestamp string */

    if (pacc -> rsec)
    {
	reset_stamp.secPastEpoch = pacc -> rsec;
	reset_stamp.nsec = 0;
	tsStampToText (&reset_stamp, TS_TEXT_MMDDYY, pacc -> lrst);
    }
    else
	pacc -> lrst[0] = 0;

    /* init last tick count and last time running totals saved to
       current tick count
    */
    pacc -> ltck = tickGet();

    /* init with offset */
    pacc -> lsav = timenow.secPastEpoch + (pacc -> srvt * pacc -> phas);

    pacc -> init = TRUE;
    pacc -> fpps = 1;
    pacc -> nsta = NO_ALARM;
    pacc -> nsev = NO_ALARM;
    pacc -> msev = NO_ALARM;
    pacc -> talc = 0;

    /* init alarm delay variables */
    pacc -> larf = 0;
    pacc -> hhll = 0;
    pacc -> hilo = 0;
    pacc -> dvct = 0;
    pacc -> rtct = 0;

    /* init last alarm time and last process time */
    if (pacc -> latm.secPastEpoch == 0)
	tsLocalTime (&pacc -> latm);
    if (pacc -> lptm.secPastEpoch == 0)
	tsLocalTime (&pacc -> lptm);

    if (pacc -> afmd == TIMED)
    {
	if (pacc -> aftm <= 0)
	{
	    recGblRecordError (S_db_badField, (struct dbCommon *)pacc, 
				"acc, init_record, 0 freeze time");
	    return (S_db_badField);
	}

	if (pacc->aftu == MINUTES) 
	{
	    frzratesecs = pacc->aftm * 60;
	}     
	else 
	{
	    frzratesecs = pacc->aftm * 3600;
	}
	/* calc first freeze time as string */
	firstFreeze(pacc, timebuf, frzratesecs);

	/* convert to epics time */
	tsTextToStamp (&next_freeze, &ptimebuf);

	/* get next freeze time */
	while (next_freeze.secPastEpoch <= timenow.secPastEpoch) 
	{
	    next_freeze.secPastEpoch += frzratesecs;
	}
	next_freeze.nsec = 0;	/* zero nanoseconds */

	/* update next_freeze time to be used in process() */
	pacc -> nftm = next_freeze;
    }

    /* else we must be ASYNCHRONOUS freeze */
    else
    {
	if (pacc -> afis.type != DB_LINK)
	{
	    recGblRecordError (S_db_badField, (struct dbCommon *)pacc,
				"acc, init_record, async not link");
	    return (S_db_badField);
	}
	/* is it init'd yet, if it is get its value and init
	   the last async reset value LASF field
	*/
	if (!(okToUse (&pacc -> afis)) )
	{
	    pacc -> nsta = NO_INIT_ALARM;
	    pacc -> nsev = VALID_ALARM;
	}
	else
	{
	    status = 0;
	    if ( (status = dbGetLink (&pacc -> afis.value.db_link, 
				      (struct dbCommon *)pacc, 
				      DBR_DOUBLE,(caddr_t)&pacc -> lasf,
				      &options, &nRequest)) != 0)
	    {
		pacc -> nsta = NO_INIT_ALARM;
		pacc -> nsev = VALID_ALARM;
	    }
	}
    }

    /* if doing async reset, is it init'd, if it is get
       it's value and set lars, the last async reset value
    */

    if (pacc -> aris.type == DB_LINK)
    {
	if (!(okToUse (&pacc -> aris)))
	{
	    pacc -> nsta = NO_INIT_ALARM;
	    pacc -> nsev = VALID_ALARM;
	}
	else
	{
	    status = 0;
	    if ( (status = dbGetLink (&pacc -> aris.value.db_link, 
				      (struct dbCommon *)pacc,
				      DBR_DOUBLE, (caddr_t)&pacc -> lars, 
				      &options, &nRequest)) != 0)
	    {
		pacc -> nsta = NO_INIT_ALARM;
		pacc -> nsev = VALID_ALARM;
	    }
	}
    }


    if (pacc -> nsev == NO_ALARM)
    {
	if ( (pacc -> inp.type == DB_LINK) &&
	     (okToUse (&pacc -> inp)) )
	{
	    status = 0;

	    /* if delta mode was selected, init the old raw value */
	    if (pacc -> dmdf)
	    {
		status = dbGetLink (&pacc -> inp.value.db_link, 
				    (struct dbCommon *)pacc, 
				    DBR_DOUBLE, (caddr_t)&pacc -> oraw, 
				    &options, &nRequest);
	    }
	    if (status == 0)
	    {
		ini2_flag = 1;
	    }
	    else
	    {
		pacc -> nsta = NO_INIT_ALARM;
		pacc -> nsev = VALID_ALARM;
	    }
	}
	else
	{
	    pacc -> nsta = NO_INIT_ALARM;
	    pacc -> nsev = VALID_ALARM;
	}
    }

    /* if we are MASTER post monitors and send to redundant partner */

    if (resGetCaAnswer (pacc -> radr, &i) == RES_OK)
    {
	if (i)
	{
	    /* check if we missed a freeze here, freeze may reset the running
	       totals, and if we did this code above, if we where not master
	       we would reset the running totals from the saved totals, but would
	       not refreeze, thereby giving erroneous values for the running 
	       totals
	    */
	    if ( (pacc -> afmd == TIMED) && (ini2_flag) )
	    {
	        /* if no previous freeze time and no samples, do not test
	           if we missed a freeze
	        */
		if ( (pacc -> fsec != 0) && (pacc -> rsmp != 0) )
		{
	    	    /* if next_freeze time - frzratesec is > than the last 
		       freeze time we may have missed a freeze if the IOC 
		       was down DO IT NOW
	    	    */
	    	    if ( (next_freeze.secPastEpoch - frzratesecs) > 
			 pacc -> ftim.secPastEpoch)
	    	    {
			timenow.nsec = 0;
			freeze (pacc, timenow, &update_crc, &did_a_reset);
			did_a_freeze = 1;
			db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rtot, DBE_VALUE);
			db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> ravg, DBE_VALUE);
			db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rmin, DBE_VALUE);
			db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rmax, DBE_VALUE);
			db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rsmp, DBE_VALUE);
			db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> atot, DBE_VALUE);
	    	    }
		}
	    }

	    /* post monitors, if we will finally init, 
	       set INI2, send all redund fields to slave, when slave 
	       receives INI2, it will be fully init'd also
	    */

	    monitor (pacc);

	    if (ini2_flag)
	    {
		pacc -> ini2 = 1;

		/* post events and update redundant fields on slave */
		post_and_redund (pacc, &old_acc, update_crc, did_a_freeze);
	    }
	}
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static void alarm (pacc)
    struct accumRecord	*pacc;
{
    double		val = pacc -> val;
    int			adis = 0;
    ALARM_LIMITS	alarm_limits;

    /* check for alarm inhibit */

    adis = AlarmDisableLink ((void *)pacc);
    if ( (pacc -> aenb == ALARM_DISABLED) || (adis != 0) )
    {
	if (pacc -> nsta != NO_INIT_ALARM)
	{
	    pacc -> nsta = INHIBIT_ALARM;
	    pacc -> nsev = NO_ALARM;
	}
	return;
    }

    if(pacc -> udf == TRUE ) 
    {
	if (pacc -> nsev < VALID_ALARM) 
	{
	    pacc -> nsta = UDF_ALARM;
	    pacc -> nsev = VALID_ALARM;
	}
	return;
    }

    /* set up for limit alarms */

    alarm_limits.hihi = pacc -> hihi;
    alarm_limits.high = pacc -> high;
    alarm_limits.low  = pacc -> low;
    alarm_limits.lolo = pacc -> lolo;
    alarm_limits.devl = 0;
    alarm_limits.devs = 0;
    alarm_limits.hyst = pacc -> hyst;
    alarm_limits.hhsv = pacc -> hhsv;
    alarm_limits.hsv  = pacc -> hsv;
    alarm_limits.lsv  = pacc -> lsv;
    alarm_limits.llsv = pacc -> llsv;
    alarm_limits.dvsv = NO_ALARM;

    AlarmLimits (&alarm_limits, val, (void *)pacc);

    /* do a return to normal, check for alarm delay ALDL only
       on the listed alarms
    */

    if ( ((pacc -> nsta == NO_ALARM) && (pacc -> nsev == NO_ALARM)) &&
	 ((pacc -> stat == HIHI_ALARM) || (pacc -> stat == HIGH_ALARM) ||
	  (pacc -> stat == LOW_ALARM)  || (pacc -> stat == LOLO_ALARM)) )
    {
	if (pacc -> aldl > 0)
	{
	    if (pacc -> rtct < ALDL_MAX)
		pacc -> rtct++;

	    /* delay return to normal so, continue last alarm */
	    if (pacc -> rtct < pacc -> aldl)
	    {
		pacc -> nsta = pacc -> stat;
		pacc -> nsev = pacc -> sevr;
	    }
	}
    }
    else
	pacc -> rtct = 0;

    return;
}


/*------------------------------------------------------------------------*/

/******
	Calculate the first freeze time of the day. This is done
	by making sure that the first freeze time is less than the
	freeze rate.
******/

static void firstFreeze(pacc, timebuf, rate)
    struct accumRecord	*pacc;
    char		*timebuf;
    double		rate;
{
    int	delta;
    int	first;

    delta = rate;
    first = pacc -> fafh * 3600 + pacc -> fafm * 60;
    first = first % delta;
    sprintf(timebuf, "%02d:%02d:00:000", first/3600, (first % 3600) / 60);
    return;
}


/*------------------------------------------------------------------------*/

/******
	free memory for *rgrps[]
******/

static void free_rgrps ()
{
    int i;

    for (i = 0; i < 256; i++)
    {
	if (rgrps[i] != NULL)
	    free ((char*)rgrps[i]);
    }
    bzero ((char*)rgrps, sizeof (rgrps));
    return;
}


/*------------------------------------------------------------------------*/

/******
	freeze record
******/

static void freeze (pacc, next_freeze, update_crc, did_a_reset)
    struct accumRecord	*pacc;
    TS_STAMP		next_freeze;
    int			*update_crc;
    int			*did_a_reset;
{
    int		reset = 0;
    char	timebuf[30];
    TS_STAMP	freeze_time = next_freeze;
    TS_STAMP	last_freeze = pacc -> ftim;

    /* convert freeze time to a string of form MMDDYY:TIME */
    tsStampToText (&freeze_time, TS_TEXT_MMDDYY, timebuf);

    pacc -> fsmp = pacc -> rsmp;		/* num samples	*/
    pacc -> ftot = pacc -> rtot;		/* total	*/
    pacc -> favg = pacc -> ravg;		/* average	*/
    pacc -> fmin = pacc -> rmin;		/* minimum	*/
    pacc -> fmax = pacc -> rmax;		/* maximum	*/
    pacc -> ftim = freeze_time;			/* freeze time struct */
    pacc -> fsec = freeze_time.secPastEpoch;	/* freeze time secs past epics epic */
    pacc -> fsta = pacc -> nfsa;		/* status	*/
    pacc -> fsev = pacc -> nfsv;		/* severity	*/
    strncpy (pacc -> fts0, timebuf, sizeof (pacc -> fts0));    /* freeze time string */

    /* set flag to update CRC value */
    *update_crc = 1;

    /* if reset on freeze or have an async reset or doing a timed group reset */
    if (pacc -> roff || pacc -> arsf || pacc -> rgrp) 
    {
	if (pacc -> roff || pacc -> arsf)
	    reset = 1;
	else if (pacc -> rgrp)
	{
	    if (test_rgrp_reset (pacc, last_freeze))
		reset = 1;
	}
	if (reset)
	{
	    pacc -> rsmp = 0;
	    pacc -> rtot = 0;
	    pacc -> ravg = 0;
	    pacc -> rmin = 0;
	    pacc -> rmax = 0;
	    pacc -> atot = 0;
	    pacc -> nfsa = NO_ALARM;
	    pacc -> nfsv = NO_ALARM;
	    /* set time of last reset */
	    pacc -> rsec = pacc -> fsec;
	    strncpy (pacc -> lrst, timebuf, sizeof (pacc -> lrst));
	    *did_a_reset = 1;
	}
    }
    pacc -> arsf = 0;
    return;
}


/*------------------------------------------------------------------------*/

static long get_value(pacc,pvdes)
    struct	accumRecord	*pacc;
    struct	valueDes	*pvdes;
{
    pvdes->field_type = DBF_DOUBLE;
    pvdes->no_elements = 1;
    (double *) (pvdes->pvalue) = &pacc->val;
    return (0);
}

/*------------------------------------------------------------------------*/

static long get_units(paddr,units)
    struct dbAddr *paddr;
    char	  *units;
{
    struct accumRecord	*pacc = (struct accumRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct accumRecord *)0) -> rtot) ):
	case ( (int)&(((struct accumRecord *)0) -> ftot) ):
	case ( (int)&(((struct accumRecord *)0) -> stot) ):
	case ( (int)&(((struct accumRecord *)0) -> hopr) ):
	case ( (int)&(((struct accumRecord *)0) -> lopr) ):
	    strncpy (units, pacc -> egu, sizeof (pacc -> egu));
	    break;

	case ( (int)&(((struct accumRecord *)0) -> val) ):
	case ( (int)&(((struct accumRecord *)0) -> hyst) ):
	case ( (int)&(((struct accumRecord *)0) -> adel) ):
	case ( (int)&(((struct accumRecord *)0) -> mdel) ):
	case ( (int)&(((struct accumRecord *)0) -> zdth) ):
	    if (pacc -> afs == TOTAL)
		strncpy (units, pacc -> egu, sizeof (pacc -> egu));
	    else
		strncpy (units, pacc -> megu, sizeof (pacc -> megu));
	    break;

	default:
	    strncpy (units, pacc -> megu, sizeof (pacc -> megu));
	    break;
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static long get_precision (paddr, precision)
    struct	dbAddr	*paddr;
    long		*precision;
{
    struct accumRecord	*pacc = (struct accumRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct accumRecord *)0) -> rtot) ):
	case ( (int)&(((struct accumRecord *)0) -> ftot) ):
	case ( (int)&(((struct accumRecord *)0) -> stot) ):
	case ( (int)&(((struct accumRecord *)0) -> hopr) ):
	case ( (int)&(((struct accumRecord *)0) -> lopr) ):
	    *precision = pacc -> prec;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> val) ):
	case ( (int)&(((struct accumRecord *)0) -> hyst) ):
	case ( (int)&(((struct accumRecord *)0) -> adel) ):
	case ( (int)&(((struct accumRecord *)0) -> mdel) ):
	case ( (int)&(((struct accumRecord *)0) -> zdth) ):
	    if (pacc -> afs == TOTAL)
		*precision = pacc -> prec;
	    else
		*precision = pacc -> mpre;
	    break;

	default:
	    *precision = pacc -> mpre;
	    break;
    }

    return (0);
}

/*------------------------------------------------------------------------*/

static long get_alarm_double (paddr, pad)
    struct	dbAddr		*paddr;
    struct	dbr_alDouble	*pad;
{
    struct accumRecord	*pacc = (struct accumRecord *) paddr->precord;

    pad->upper_alarm_limit = pacc->hihi;
    pad->upper_warning_limit = pacc->high;
    pad->lower_warning_limit = pacc->low;
    pad->lower_alarm_limit = pacc->lolo;
    return (0);
}


/*------------------------------------------------------------------------*/

static long get_graphic_double(paddr,pgd)
    struct	dbAddr		*paddr;
    struct	dbr_grDouble	*pgd;
{
    struct accumRecord	*pacc = (struct accumRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct accumRecord *)0) -> val) ):
	    if (pacc -> afs == TOTAL)
	    {
		pgd -> upper_disp_limit = pacc -> hopr;
		pgd -> lower_disp_limit = pacc -> lopr;
	    }
	    else
	    {
		pgd -> upper_disp_limit = pacc -> mhpr;
		pgd -> lower_disp_limit = pacc -> mlpr;
	    }
	    break;

	case ( (int)&(((struct accumRecord *)0) -> hyst) ):
	case ( (int)&(((struct accumRecord *)0) -> adel) ):
	case ( (int)&(((struct accumRecord *)0) -> mdel) ):
	case ( (int)&(((struct accumRecord *)0) -> zdth) ):
	{
	    float    span;
	    float    hopr, lopr;

	    hopr = pacc -> hopr;
	    lopr = pacc -> lopr;
	    if (pacc -> afs != TOTAL)
	    {
		hopr = pacc -> mhpr;
		lopr = pacc -> mlpr;
	    }
	    span = hopr - lopr;
	    if (span < 0.0)
	        span = -span;
	    pgd -> upper_disp_limit = span;
	    pgd -> lower_disp_limit = 0;
	    break;
	}

	case ( (int)&(((struct accumRecord *)0) -> rtot) ):
	case ( (int)&(((struct accumRecord *)0) -> ftot) ):
	case ( (int)&(((struct accumRecord *)0) -> stot) ):
	case ( (int)&(((struct accumRecord *)0) -> hopr) ):
	case ( (int)&(((struct accumRecord *)0) -> lopr) ):
	    pgd -> upper_disp_limit = pacc -> hopr;
	    pgd -> lower_disp_limit = pacc -> lopr;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> aslo) ):
	case ( (int)&(((struct accumRecord *)0) -> aoff) ):
	case ( (int)&(((struct accumRecord *)0) -> mslo) ):
	case ( (int)&(((struct accumRecord *)0) -> moff) ):
	    pgd -> upper_disp_limit = 32767;
	    pgd -> lower_disp_limit = -32768;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> aftm) ):
	    pgd -> upper_disp_limit = 99;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> fafh) ):
	    pgd -> upper_disp_limit = 23;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> fafm) ):
	    pgd -> upper_disp_limit = 59;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> srvt) ):
	    pgd -> upper_disp_limit = 1440;	/* minutes/day */
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> rlov) ):
	case ( (int)&(((struct accumRecord *)0) -> thld) ):
	case ( (int)&(((struct accumRecord *)0) -> fsec) ):
	    pgd -> upper_disp_limit = 999999999;
	    pgd -> lower_disp_limit = 0;
	    break;

	default:
	    pgd -> upper_disp_limit = pacc -> mhpr;
	    pgd -> lower_disp_limit = pacc -> mlpr;
	    break;
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static long get_control_double(paddr,pcd)
    struct	dbAddr		*paddr;
    struct	dbr_ctrlDouble	*pcd;
{
    struct accumRecord *pacc = (struct accumRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct accumRecord *)0) -> val) ):
	    if (pacc -> afs == TOTAL)
	    {
		pcd -> upper_ctrl_limit = pacc -> hopr;
		pcd -> lower_ctrl_limit = pacc -> lopr;
	    }
	    else
	    {
		pcd -> upper_ctrl_limit = pacc -> mhpr;
		pcd -> lower_ctrl_limit = pacc -> mlpr;
	    }
	    break;

	case ( (int)&(((struct accumRecord *)0) -> hyst) ):
	case ( (int)&(((struct accumRecord *)0) -> adel) ):
	case ( (int)&(((struct accumRecord *)0) -> mdel) ):
	case ( (int)&(((struct accumRecord *)0) -> zdth) ):
	{
	    float    span;
	    float    hopr, lopr;

	    hopr = pacc -> hopr;
	    lopr = pacc -> lopr;
	    if (pacc -> afs != TOTAL)
	    {
		hopr = pacc -> mhpr;
		lopr = pacc -> mlpr;
	    }
	    span = hopr - lopr;
	    if (span < 0.0)
	        span = -span;
	    pcd -> upper_ctrl_limit = span;
	    pcd -> lower_ctrl_limit = 0;
	    break;
	}

	case ( (int)&(((struct accumRecord *)0) -> rtot) ):
	case ( (int)&(((struct accumRecord *)0) -> hopr) ):
	case ( (int)&(((struct accumRecord *)0) -> lopr) ):
	    pcd -> upper_ctrl_limit = pacc -> hopr;
	    pcd -> lower_ctrl_limit = pacc -> lopr;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> aslo) ):
	case ( (int)&(((struct accumRecord *)0) -> aoff) ):
	case ( (int)&(((struct accumRecord *)0) -> mslo) ):
	case ( (int)&(((struct accumRecord *)0) -> moff) ):
	    pcd -> upper_ctrl_limit = 32767;
	    pcd -> lower_ctrl_limit = -32768;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> aftm) ):
	    pcd -> upper_ctrl_limit = 99;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> fafh) ):
	    pcd -> upper_ctrl_limit = 23;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> fafm) ):
	    pcd -> upper_ctrl_limit = 59;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> srvt) ):
	    pcd -> upper_ctrl_limit = 1440;	/* minutes/day */
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct accumRecord *)0) -> rlov) ):
	case ( (int)&(((struct accumRecord *)0) -> thld) ):
	    pcd -> upper_ctrl_limit = 999999999;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	default:
	    pcd -> upper_ctrl_limit = pacc -> mhpr;
	    pcd -> lower_ctrl_limit = pacc -> mlpr;
	    break;
    }

    return (0);
}


/*------------------------------------------------------------------------*/

/*****
    can be called in startup.cmd or at ioc prompt by user to load a
    timed reset group
*****/

int ld_rgrp (rgrp_data)
    char    *rgrp_data;
{
    static int	first_ld = 1;

    if (first_ld && (rgrps[0] == 0))
    {
	set_rgrp_defaults ();
	first_ld = 0;
    }
    if ( (rgrp_data[0] == ' ') || (rgrp_data[0] == 0) )
    {
	logMsg ("dl_rgrp, null string or leading spaces\n",0,0,0,0,0,0);
    }
    else
	parse_rgrp (rgrp_data);
    return (0);
}


/*------------------------------------------------------------------------*/

static void monitor (pacc)
    struct	accumRecord	*pacc;
{
    unsigned short	monitor_mask;
    double		delta;
    short		stat, sevr, nsta, nsev;
    int			set_ack = 1;

    /* get previous stat and sevr  and new stat and sevr*/

    stat = pacc->stat;
    sevr = pacc->sevr;
    nsta = pacc->nsta;
    nsev = pacc->nsev;

    /*set current stat and sevr*/

    pacc->stat = nsta;
    pacc->sevr = nsev;
    pacc->nsta = 0;
    pacc->nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == pacc -> psta) && (nsev == pacc -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and out previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	pacc -> psta = stat;
	pacc -> psev = sevr;
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> psev, DBE_VALUE);
    }

    monitor_mask = 0;

    /* alarm condition changed this scan */

    if ((stat != nsta) || (sevr != nsev)) 
    {
	/* post events for alarm condition change*/

	monitor_mask = DBE_ALARM;

	/* when alarm status chg's, get current time and
	   save VAL in lalm (last value when alarm chg'd)
	*/
	tsLocalTime (&pacc -> latm);
	pacc -> lalm = pacc -> val;

	/* post stat and nsev fields */

	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc->stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc->sevr, DBE_VALUE);
	if ((pacc->sevr != NO_ALARM) && (set_ack) &&
	    (pacc->sevr != EVENT_ALARM) )
	{
	    if (pacc->ack == 0)
	    {
		pacc->ack = 1;	/* new unacknowledged alarm */
		db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> ack, DBE_VALUE);
	    }
	}
    }

    /* check for value change */

    delta = pacc->mlst - pacc->val;
    if (delta < 0.0) 
	delta = -delta;
	
    if (delta > pacc->mdel) 
    {
	/* post events for value change */

	monitor_mask |= DBE_VALUE;

	/* update last value monitored */

	pacc->mlst = pacc->val;
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rtot, DBE_VALUE);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> ravg, DBE_VALUE);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rmin, DBE_VALUE);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rmax, DBE_VALUE);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rsmp, DBE_VALUE);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> atot, DBE_VALUE);
	pacc -> lmto = pacc -> rtot;
    }
    else
    {
	if (pacc -> afs != TOTAL)
	{
	    delta = pacc -> rtot - pacc -> lmto;
	    if (delta < 0.0)
		delta = -delta;
	    if (delta > ( (pacc -> hopr - pacc -> lopr) * 0.0005) )
	    {
		db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rtot, DBE_VALUE);
		db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rsmp, DBE_VALUE);
		pacc -> lmto = pacc -> rtot;
	    }
	}
    }

    /* check for archive change only if tenb (trending) enabled */

    if (pacc -> tenb)
    {
	delta = pacc->alst - pacc->val;
	if (delta < 0.0) 
	    delta = -delta;
       
	/* if exceed archive deadband or there has been a change in alarm status */
	if ( (delta > pacc->adel) || (monitor_mask & DBE_ALARM) )
	{
	    /* post events on value field for archive change */

	    monitor_mask |= DBE_LOG;

	    /* update last archive value monitored */

	    pacc->alst = pacc->val;
	    db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rtot, DBE_LOG);
	    db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> ravg, DBE_LOG);
	    db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rmin, DBE_LOG);
	    db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rmax, DBE_LOG);
	    db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rsmp, DBE_LOG);
	    pacc -> lato = pacc -> rtot;
	}
	else
	{
	    if (pacc -> afs != TOTAL)
	    {
		delta = pacc -> rtot - pacc -> lato;
		if (delta < 0.0)
		    delta = -delta;
		if (delta > ( (pacc -> hopr - pacc -> lopr) * 0.0005) )
		{
		    db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rtot, DBE_LOG);
		    db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rsmp, DBE_LOG);
		    pacc -> lato = pacc -> rtot;
		}
	    }
	}
    }

    /* send out monitors connected to the value field */

    if (monitor_mask) 
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc->val, monitor_mask);
	
    return;
}


/*------------------------------------------------------------------------*/

/******
	called from ld_rgrp() or read_rgrps()
	parse the input into something the program can use
	format of input, I hope!!, NOTE: all fields need not be here
	but still need comma separator
		RGRP NUM, DAY_OF_WEEK, MONTH, DAY_OF_MONTH, TIME

		RGRP = 1 - 255
		DOW  = Sun - Sat
		MON  = Jan - Dec
		DOM  = 1 - 31, 32 mean last day of month
		TIEM = HR:MIN, if use time both value smust be present
******/

static int parse_rgrp (rgrp_data)
    char    *rgrp_data;
{
    int		level;
    int		cnt;
    int		rgrp_idx;
    char	*p;
    char	*colon;
    char	hour[8];
    char	minute[8];
    char	levels[5][32];
    struct rgrp *new_rgrp;	/* see start of file for def */

    bzero ((char*)&levels[0][0], sizeof (levels));
    if ( (new_rgrp = (struct rgrp *)calloc (1, sizeof (struct rgrp))) == NULL)
    {
	logMsg ("parse_rgrp, calloc failed\n",0,0,0,0,0,0);
	return (1);
    }
    p = rgrp_data;
    level = 0;

    /* parse into components */

    while ( ((*p) != NULL) && (level < 5) )
    {
	if ((*p) == ',')
	{
	    level++;
	    p++;
	    continue;
	}
	cnt = 0;
	while ( ((*p) != ',') && ((*p) != NULL) )
	{
	    levels[level][cnt] = *p;
	    p++;
	    cnt++;
	}
    }

    if (level != 4)
    {
	logMsg ("parse_rgrp, missing fields\n",0,0,0,0,0,0);
	free ((char*)new_rgrp);
	return (1);
    }
    rgrp_idx = atoi (levels[0]);
    if ( (rgrp_idx < 1) || (rgrp_idx >  255) )
    {
	logMsg ("parse_rgrp, illegal group (%d) (1-255)\n",rgrp_idx,0,0,0,0,0);
	free ((char*)new_rgrp);
	return (1);
    }

    /* get day of week */

    if (levels[1][0] != NULL)
    {
	p = levels[1];
	while (*p)
	{
	    *p = tolower (*p);
	    p++;
	}
	if ( (strncmp (levels[1], "sun", 3)) == 0)
	    new_rgrp -> day_of_week = 1;
	else if ( (strncmp (levels[1], "mon", 3)) == 0)
	    new_rgrp -> day_of_week = 2;
	else if ( (strncmp (levels[1], "tue", 3)) == 0)
	    new_rgrp -> day_of_week = 3;
	else if ( (strncmp (levels[1], "wed", 3)) == 0)
	    new_rgrp -> day_of_week = 4;
	else if ( (strncmp (levels[1], "thu", 3)) == 0)
	    new_rgrp -> day_of_week = 5;
	else if ( (strncmp (levels[1], "fri", 3)) == 0)
	    new_rgrp -> day_of_week = 6;
	else if ( (strncmp (levels[1], "sat", 3)) == 0)
	    new_rgrp -> day_of_week = 7;
	else
	{
	    logMsg ("parse_rgrp, illegal day_of_week, %3s (sun-sat)\n",(int)levels[1],0,0,0,0,0);
	    free ((char*)new_rgrp);
	    return (1);
	}
    }

    /* get month */

    if (levels[2][0] != NULL)
    {
	p = levels[2];
	while (*p)
	{
	    *p = tolower (*p);
	    p++;
	}
	if ( (strncmp (levels[2], "jan", 3)) == 0)
	    new_rgrp -> month = 1;
	else if ( (strncmp (levels[2], "feb", 3)) == 0)
	    new_rgrp -> month = 2;
	else if ( (strncmp (levels[2], "mar", 3)) == 0)
	    new_rgrp -> month = 3;
	else if ( (strncmp (levels[2], "apr", 3)) == 0)
	    new_rgrp -> month = 4;
	else if ( (strncmp (levels[2], "may", 3)) == 0)
	    new_rgrp -> month = 5;
	else if ( (strncmp (levels[2], "jun", 3)) == 0)
	    new_rgrp -> month = 6;
	else if ( (strncmp (levels[2], "jul", 3)) == 0)
	    new_rgrp -> month = 7;
	else if ( (strncmp (levels[2], "aug", 3)) == 0)
	    new_rgrp -> month = 8;
	else if ( (strncmp (levels[2], "sep", 3)) == 0)
	    new_rgrp -> month = 9;
	else if ( (strncmp (levels[2], "oct", 3)) == 0)
	    new_rgrp -> month = 10;
	else if ( (strncmp (levels[2], "nov", 3)) == 0)
	    new_rgrp -> month = 11;
	else if ( (strncmp (levels[2], "dec", 3)) == 0)
	    new_rgrp -> month = 12;
	else
	{
	    logMsg ("parse_rgrp, illegal month, %3s (jan-dec)\n",(int)levels[2],0,0,0,0,0);
	    free ((char*)new_rgrp);
	    return (1);
	}
    }

    /* get day of month */

    if (levels[3][0] != NULL)
    {
	new_rgrp -> day_of_month = atoi (levels[3]);
	if ( (new_rgrp -> day_of_month < 1) ||
	     (new_rgrp -> day_of_month > 32) )
	{
	    logMsg ("parse_rgrp, illegal month, %3s (jan-dec)\n",(int)levels[2],0,0,0,0,0);
	    free ((char*)new_rgrp);
	    return (1);
	}
    }

    /* init to -1 because 00:00 is legit */

    new_rgrp -> hour = new_rgrp -> minute = -1;

    /* get hours and minutes */

    if (levels[4][0] != NULL)
    {
	if ( (colon = (char *)index (levels[4], ':')) == NULL)
	{
	    logMsg ("parse_rgrp, no colon in time, %5s\n",(int)levels[4],0,0,0,0,0);
	    free ((char*)new_rgrp);
	    return (1);
	}

	/* if hour field > 2 chars */

	if ( (colon - (char*)&levels[4][0]) > 2)
	{
	    logMsg ("parse_rgrp, illegal time, %5s\n",(int)levels[4],0,0,0,0,0);
	    free ((char*)new_rgrp);
	    return (1);
	}
	bzero (hour, sizeof (hour));
	bzero (minute, sizeof (minute));
	strncpy (hour, (char*)&levels[4][0], 2);

	/* if minute field (+ colon) > 3 chars */

	if ( ((strlen (colon)) < 3) || ((strlen (colon)) > 3) )
	{
	    logMsg ("parse_rgrp, illegal time, %5s\n",(int)levels[4],0,0,0,0,0);
	    free ((char*)new_rgrp);
	    return (1);
	}
	colon++;
	strncpy (minute, colon, 2);
	new_rgrp -> hour = atoi (hour);
	new_rgrp -> minute = atoi (minute);
	if ( (new_rgrp -> hour < 0) || (new_rgrp -> hour > 23) )
	{
	    logMsg ("parse_rgrp, illegal hour, %5s (0-23)\n",(int)levels[4],0,0,0,0,0);
	    free ((char*)new_rgrp);
	    return (1);
	}
	if ( (new_rgrp -> minute < 0) || (new_rgrp -> minute > 59) )
	{
	    logMsg ("parse_rgrp, illegal minute, %5s (0-59)\n",(int)levels[4],0,0,0,0,0);
	    free ((char*)new_rgrp);
	    return (1);
	}

    }

    new_rgrp -> type = 0;

    /* now determine if they have defined one of the acceptable combinations
       day of week or day_of week with a time
       month (must have day of month) or month with time
       day of month or day of month with time
       time of day
    */
    if ( (new_rgrp -> day_of_week) && (!new_rgrp -> month) &&
	 (!new_rgrp -> day_of_month) )
    {
	if (new_rgrp -> hour >= 0)
	    new_rgrp -> type = DAY_OF_WEEK_TIME_RESET;
	else
	    new_rgrp -> type = DAY_OF_WEEK_RESET;
    }
    else if ( (new_rgrp -> month) && (!new_rgrp -> day_of_week) )
    {
	if (new_rgrp -> day_of_month == 0)
	{
		logMsg ("parse_rgrp, need dom for month option\n",0,0,0,0,0,0);
		free ((char *)new_rgrp);
		return (1);
	}
	if (new_rgrp -> hour >= 0)
	    new_rgrp -> type = MONTH_TIME_RESET;
	else
	    new_rgrp -> type = MONTH_RESET;
    }
    else if ( (new_rgrp -> day_of_month) && (!new_rgrp -> day_of_week) )
    {
	if (new_rgrp -> hour >= 0)
	    new_rgrp -> type = DAY_OF_MONTH_TIME_RESET;
	else
	    new_rgrp -> type = DAY_OF_MONTH_RESET;
    }
    else if (new_rgrp -> hour >= 0)
    {
	new_rgrp -> type = TIME_RESET;
    }
    else
    {
	logMsg ("parse_rgrp, illegal time reset combination (%d,%d,%d,%d,%d)\n",
		new_rgrp -> day_of_week, new_rgrp -> month,
		new_rgrp -> day_of_month, new_rgrp -> hour,
		new_rgrp -> minute,0);
	free ((char*)new_rgrp);
	return (1);
    }

    /* store for later use, if already in use, free it up */

    if (rgrps[rgrp_idx] != 0)
    {
	free ((char*)rgrps[rgrp_idx]);
	rgrps[rgrp_idx] = 0;
    }
    rgrps[rgrp_idx] = new_rgrp;

    return (0);
}


/*------------------------------------------------------------------------*/

/******
	post_events and update redundant fields
	if update_crc flag is set then call DBC_CRCRecord()
	to update the checksum value due to chgs in
	prompted fields
******/

static void post_and_redund (pacc, old_acc, update_crc, did_a_freeze)
    struct accumRecord	*pacc;
    struct accred	*old_acc;
    int			update_crc;
    int			did_a_freeze;
{
    char		redbuf[REDQ_MAX_MSGLEN];
    unsigned short	stat, sevr, monitor_mask;
    TS_STAMP		mytimestamp;

    if (update_crc)
    {
	pacc -> rcrc = DBC_CRCRecord ((char *)pacc, type);
    }

    redQInit ((struct dbCommon *)pacc, redbuf, sizeof (redbuf), type);

    if (old_acc -> stat != pacc -> stat) 
    {
	redQField (redbuf, mylist [STAT]);
    }

    if (old_acc -> sevr != pacc -> sevr) 
    {
	redQField (redbuf, mylist [SEVR]);
    }

    if (old_acc -> val != pacc -> val) 
    {
	redQField (redbuf, mylist [VAL]);
    }

    if (old_acc -> rsmp != pacc -> rsmp) 
	redQField (redbuf, mylist [RSMP]);

    if (old_acc -> rtot != pacc -> rtot) 
	redQField (redbuf, mylist [RTOT]);

    if (old_acc -> ravg != pacc -> ravg) 
	redQField (redbuf, mylist [RAVG]);

    if (old_acc -> rmin != pacc -> rmin) 
	redQField (redbuf, mylist [RMIN]);

    if (old_acc -> rmax != pacc -> rmax) 
	redQField (redbuf, mylist [RMAX]);

    if (old_acc -> ftim.secPastEpoch != pacc -> ftim.secPastEpoch)
    {
	/* fts0 is not redundant, special on fsec will 
	   recreate it on the slave
	*/
	db_post_events ((struct dbCommon *)pacc, pacc -> fts0, DBE_VALUE);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> fsec, DBE_VALUE);
	redQField (redbuf, mylist [FTIM]);
	redQField (redbuf, mylist [FSEC]);
    }

    if (old_acc -> rsec != pacc -> rsec)
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> rsec, DBE_VALUE);
	redQField (redbuf, mylist [RSEC]);
    }

    if (strncmp (old_acc -> lrst, pacc -> lrst, sizeof (old_acc -> lrst)) != 0)
    {
	/* lrst is not redundant because a special on
	   rsec will recreate it on slave
	*/
	db_post_events ((struct dbCommon *)pacc, pacc -> lrst, DBE_VALUE);
    }

    /* KEEP FROZEN TOGETHER, replace stat and sevr with fsta and fsev,
       if we are trending the frozen values, then we want
       to see the frozen status and severity, we can do
       this because db_post_events grabs stat and sevr when the
       LOG bit is set, also replace pacc->latm with pacc->ftim
       because db_post_events grabs pacc->latm for archive,
       actually it is .../share/src/rsrv/rbMgt.c:caRbHandlerPut
       which is called by .../share/src/rsrv/postEventrs.c:db_post_events
       that grabs stat sevr and time
    */
    stat = pacc -> stat;
    sevr = pacc -> sevr;
    mytimestamp = pacc -> latm;
    pacc -> latm = pacc -> ftim;
    pacc -> stat = pacc -> fsta;
    pacc -> sevr = pacc -> fsev;
    monitor_mask = DBE_VALUE;

    /* only trend if trending is enabled */

    if (pacc -> tenb)
	monitor_mask |= DBE_LOG;

    /* post and redund if we did a freeze whether value has chg'd
       or not, keeps trend happy
    */
    if (did_a_freeze)
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> fsmp, monitor_mask);
	redQField (redbuf, mylist [FSMP]);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> ftot, monitor_mask);
	redQField (redbuf, mylist [FTOT]);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> favg, monitor_mask);
	redQField (redbuf, mylist [FAVG]);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> fmin, monitor_mask);
	redQField (redbuf, mylist [FMIN]);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> fmax, monitor_mask);
	redQField (redbuf, mylist [FMAX]);
    }

    /* return to previous condition */
    pacc -> stat = stat;
    pacc -> sevr = sevr;
    pacc -> latm = mytimestamp;

    /* END KEEP TOGETHER */

    if (old_acc -> lalm != pacc -> lalm) 
    {
	redQField (redbuf, mylist [LALM]);
    }

    if (old_acc -> ini2 != pacc -> ini2) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> ini2, DBE_VALUE);
	redQField (redbuf, mylist [INI2]);
    }

    if (old_acc -> ltck != pacc -> ltck) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> ltck, DBE_VALUE);
	redQField (redbuf, mylist [LTCK]);
    }

    if (old_acc -> lmod != pacc -> lmod) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> lmod, DBE_VALUE);
	redQField (redbuf, mylist [LMOD]);
    }

    if (old_acc -> stot != pacc -> stot) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> stot, DBE_VALUE);
	redQField (redbuf, mylist [STOT]);
    }

    if (old_acc -> ssmp != pacc -> ssmp) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> ssmp, DBE_VALUE);
	redQField (redbuf, mylist [SSMP]);
    }

    if (old_acc -> smax != pacc -> smax) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> smax, DBE_VALUE);
	redQField (redbuf, mylist [SMAX]);
    }

    if (old_acc -> smin != pacc -> smin) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> smin, DBE_VALUE);
	redQField (redbuf, mylist [SMIN]);
    }

    if (old_acc -> sstt != pacc -> sstt) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> sstt, DBE_VALUE);
	redQField (redbuf, mylist [SSTT]);
    }

    if (old_acc -> lars != pacc -> lars) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> lars, DBE_VALUE);
	redQField (redbuf, mylist [LARS]);
    }

    if (old_acc -> arsf != pacc -> arsf) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> arsf, DBE_VALUE);
	redQField (redbuf, mylist [ARSF]);
    }

    if (old_acc -> fsta != pacc -> fsta) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> fsta, DBE_VALUE);
	redQField (redbuf, mylist [FSTA]);
    }

    if (old_acc -> fsev != pacc -> fsev) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> fsev, DBE_VALUE);
	redQField (redbuf, mylist [FSEV]);
    }

    if (old_acc -> nfsa != pacc -> nfsa) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> nfsa, DBE_VALUE);
	redQField (redbuf, mylist [NFSA]);
    }

    if (old_acc -> nfsv != pacc -> nfsv) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> nfsv, DBE_VALUE);
	redQField (redbuf, mylist [NFSV]);
    }

    if (old_acc -> lsav != pacc -> lsav) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> lsav, DBE_VALUE);
	redQField (redbuf, mylist [LSAV]);
    }

    if (old_acc -> atot != pacc -> atot) 
	redQField (redbuf, mylist [ATOT]);

    if (old_acc -> lasf != pacc -> lasf) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> lasf, DBE_VALUE);
	redQField (redbuf, mylist [LASF]);
    }

    if (old_acc -> talc != pacc -> talc) 
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> talc, DBE_VALUE);
	redQField (redbuf, mylist [TALC]);
    }

    if (old_acc -> nftm.secPastEpoch != pacc -> nftm.secPastEpoch)
    {
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> nftm, DBE_VALUE);
	redQField (redbuf, mylist [NFTM]);
    }

    if (TsCmpStampsNE (&old_acc -> latm, &pacc -> latm) )
	redQField (redbuf, mylist [LATM]);

    if (old_acc -> mlst != pacc -> mlst)
	redQField (redbuf, mylist [MLST]);

    if (old_acc -> lmto != pacc -> lmto)
	redQField (redbuf, mylist [LMTO]);

    if (old_acc -> hhll != pacc -> hhll)
	redQField (redbuf, mylist [HHLL]);

    if (old_acc -> hilo != pacc -> hilo)
	redQField (redbuf, mylist [HILO]);

    if (old_acc -> rtct != pacc -> rtct)
	redQField (redbuf, mylist [RTCT]);

    if (old_acc -> larf != pacc -> larf)
	redQField (redbuf, mylist [LARF]);

    redQSend (redbuf);

    return;
}


/*------------------------------------------------------------------------*/

static long process (paddr)
    struct dbAddr	*paddr;
{
    struct accumRecord	*pacc = (struct accumRecord *) (paddr->precord);

    char		freezeflag = 0;
    long		status = 0;
    long		options = 0;
    long		nRequest = 1;
    double		tval;
    double		afiv;
    char		timebuf [30];
    char		*ptimebuf = timebuf;
    TS_STAMP		freezetime;
    TS_STAMP		timenow;
    double		frzratesecs;
    double		new_async_reset;
    unsigned long	thistime = 0;
    double		time_adj;
    double		time_mul;
    float		slope;
    double		lval;
    struct accred	old_acc;
    int			update_crc;
    int			did_a_freeze = 0;
    int			did_a_reset = 0;

    /* if not init'd do nothing */

    if ((pacc->init == 0) || (pacc->ini2 == 0)) 
	return (0);

    if (pacc->pact) 
	return (0);

    /* get time stamp for last time record was processed */

    tsLocalTime (&pacc -> lptm);

    pacc -> msev = NO_ALARM;

    /* init here, will be set to 1 if needed */
    update_crc = 0;
	
    /* save current accum status */
    save_accum (pacc, &old_acc);

    /* get a delta time in seconds and calc a time adjustment */

    thistime = tickGet();
    time_adj = (thistime - pacc -> ltck) / (float)vxTicksPerSecond;
    pacc -> ltck = thistime;

    /* get the current time stamp from global value */
    tsLocalTime(&timenow);

    pacc->pact = TRUE;

    if (pacc -> tagd) 
    {
	pacc->nsta = TAGGED_ALARM;
	pacc->nsev = VALID_ALARM;
    } 
    else if (pacc->oscn == OFF_SCAN) 
    {
	pacc->nsta = SCAN_ALARM;
	pacc->nsev = VALID_ALARM;
    } 
    else 
    {
	pacc -> nsta = NO_ALARM;
	pacc -> nsev = NO_ALARM;

	/* update number of process times */
	if (process_stats)
	    process_cnts[type]++;	

	/* test for asynchronous reset */

	if ( (pacc -> aris.type == DB_LINK) &&
	     ( okToUse (&pacc -> aris)) )
	{
	    status = dbGetLink (&pacc -> aris.value.db_link, 
				(struct dbCommon *)pacc,
				DBR_DOUBLE, (caddr_t)&new_async_reset,
				&options, &nRequest);
	    if (status != 0)
	    {
		if (pacc -> nsev < VALID_ALARM)
		{
		    pacc -> nsev = VALID_ALARM;
		    pacc -> nsta = LINK_ALARM;
		}
	    }
	    else
	    {
		/* on chg from 0, set arsf so that when we freeze, we
		   can test arsf to see if we reset
		*/

		if ( (pacc -> lars == 0) && new_async_reset)
		{
		    update_crc = 1;
		    pacc -> arsf = 1;
		}
		pacc -> lars = new_async_reset;
	    }
	}

	/* if in automatic mode, get input and run calcs */

	if (pacc -> mode == NORMAL)
	{
	    /* on tranisition from MANUAL to AUTOMATIC, reset num samples to 1
	       and set ATOT (total used for RAVG ect) to RAVG (running average)
	    */
	    if (pacc -> lmod != NORMAL)
	    {
		pacc -> rsmp = 1;
		pacc -> atot = pacc -> ravg;
	    }
	    status = dbGetLink (&(pacc->inp.value.db_link), 
				(struct dbCommon *)pacc,
				DBR_DOUBLE, (caddr_t)&pacc -> rval, 
				&options, &nRequest);
	    if (status != 0) 
	    {
	        if (pacc->nsev < VALID_ALARM) 
		{
		    pacc->nsev = VALID_ALARM;
		    pacc->nsta = LINK_ALARM;
		}
	    }

	    /* if delta mode is selected, if rval < oraw then we have gone past
	       the rollover point rlov. if rval > rlov then there is a serious
	       problem, zap rval and set alarm if threshold sevr > NO_ALARM.
	       if the change is > than the threshold value thld, zap rval and
	       set alarm
	    */
	    if (pacc->dmdf) 
	    {
		if (pacc -> rval < pacc -> oraw)
		    tval = (pacc -> rlov - pacc -> oraw) + pacc -> rval;
		else
		    tval = pacc -> rval - pacc -> oraw;
	        pacc->oraw = pacc->rval;
		if ( (pacc -> rval > pacc -> rlov) &&
		     (pacc -> thsv > NO_ALARM) )
		{
		    pacc -> rval = 0.0;
		    if (pacc -> nsev < pacc -> thsv) 
		    {
		    	pacc -> nsev = pacc -> thsv;
		    	pacc -> nsta = THRESHOLD_EXCEEDED;
		    }
		}
		else
		    pacc -> rval = tval;

		/* if there is a threshold value, this is value above
		   which they may want to set an alarm but still use
		   the data, see threshold error below
		 */
		if (pacc -> thld > 0.0)
		{
		    if (pacc -> rval > pacc -> thld) 
		    {
			/* we will only try to generate an alarm if
			   we have have had > threshold count times
			   that we have exceeded the threshold
			*/
			if (pacc -> thct > 0)
			{
			    pacc -> talc++;
			    if (pacc -> talc < 0)
				pacc -> talc = pacc -> thct + 1;
			    if (pacc -> talc > pacc -> thct)
			    {
				if (pacc -> nsev < pacc -> thsv) 
				{
				    pacc -> nsev = pacc -> thsv;
				    pacc -> nsta = THRESHOLD_EXCEEDED;
				}
			    }
			}
			else
			    pacc -> talc = 0;
		    }
		    else
			pacc -> talc = 0;
		}
		else
		    pacc -> talc = 0;
	    }
	    else
		pacc -> talc = 0;

	    /* if in delta mode, and there is a threshold error value
	       and we have exceeded it, then set THRESHOLD alarm and 
	       do not use this value in calculations
	    */
	    if ( (pacc -> dmdf) && (pacc -> terr > 0.0) && 
		 (pacc -> rval > pacc -> terr) )
	    {
		if (pacc -> nsev < pacc -> thsv) 
		{
		    pacc -> nsev = pacc -> thsv;
		    pacc -> nsta = THRESHOLD_EXCEEDED;
		}
	
	    }

	    /* If the Use Zero Data Flag is set and the Value is <= to the
	       Zero Data Threshold then do not accumulate this Value
	    */
	    else if ((pacc -> uzdf == 0) && (pacc -> rval <= pacc -> zdth)) 
	    {
		;
	    } 
	    else 
	    {
		/* select type of time adjustment user want */
		switch (pacc -> tsel)
		{
		    case ACC_NONE:
			time_adj = 1.0;
			time_mul = 1.0;
			break;

		    case ACC_SECOND:
			time_mul = 1.0;
			break;

		    case ACC_MINUTE:
			time_mul = 60.0;
			break;

		    case ACC_HOUR:
			time_mul = 3600.0;
			break;

		    default:
			time_adj = 1.0;
			time_mul = 1.0;
			break;
		}

		/* calc rtot */

		lval = pacc -> rval;
		lval += pacc -> aoff;	/* apply offset */

		/* if in delta mode just apply slope */
		if (pacc -> dmdf)
		{
		    if (pacc -> aslo != 0.0)
			lval *= pacc -> aslo;
		}
		else
		{
		    if (pacc -> aslo == 0.0)
			slope = 1.0;
		    else
			slope = pacc -> aslo;
		    lval = (lval * slope) * (time_adj / time_mul);
		}
		pacc -> rtot += lval;

		/* calc rest of running values */

		lval = pacc -> rval;
		lval += pacc -> moff;	/* apply offset */

		/* if in delta mode, apply time factor and slope */

		if (pacc -> dmdf)
		{
		    if (pacc -> mslo == 0.0)
			slope = 1.0;
		    else
			slope = pacc -> mslo;
			lval = (lval * slope) * (time_mul / time_adj);
		}
		else
		{
		    if (pacc -> mslo)
			lval *= pacc -> mslo;
		}
		pacc -> rsmp++;
		pacc -> atot += lval;
		if (pacc -> rsmp == 1)
		{
		    pacc -> ravg = pacc -> atot;
		    /* if it is the first process pass, and we are
		       doing delta mode,we do not want to calc RMIN
		       and RMAX or no time adjustment was selected
		    */
		    if ( (!pacc -> fpps) || (!pacc -> dmdf) ||
			 (pacc -> tsel == ACC_NONE) )
		    {
			pacc -> rmin = lval;
			pacc -> rmax = lval;
		    }
		}
		else
		{
		    pacc -> ravg = pacc -> atot / pacc -> rsmp;
		    /* if it is the first process pass, and we are
		       doing delta mode,we do not want to calc RMIN
		       and RMAX or no time adjustment was selected
		    */
		    if ( (!pacc -> fpps) || (!pacc -> dmdf) ||
			 (pacc -> tsel == ACC_NONE) )
		    {
			if (lval < pacc -> rmin)
			    pacc -> rmin = lval;
			if (lval > pacc -> rmax)
			    pacc -> rmax = lval;
		    }
		}
	    }

	} /* end if (mode == NORMAL) */

	/* else we are in MANUAL mode, set running frozen status */
	else
	{
	    if (pacc -> nfsa != MANUAL_OVERRIDE)
	    {
	    	pacc -> nfsa = MANUAL_OVERRIDE;
	    	pacc -> nfsv = VALID_ALARM;
	    	update_crc = 1;
	    }
	}

	pacc -> lmod = pacc -> mode;

	switch (pacc->afs) 
	{
	    case TOTAL:
		pacc->val = pacc->rtot;
		break;

	    case AVERAGE:
		pacc->val = pacc->ravg;
		break;

	    case MINIMUM:
		pacc->val = pacc->rmin;
		break;

	    case MAXIMUM:
		pacc->val = pacc->rmax;
		break;

	    default:
		pacc->val = 0;
		break;
	}

	if (pacc->afmd == ASYNCHRONOUS) 
	{
	    if (pacc -> afis.type != DB_LINK)
	    {
		recGblRecordError (S_db_badField, (struct dbCommon *)pacc,
					"acc, process, async not link");
		status = S_db_badField;
		if (pacc -> nsev < VALID_ALARM)
		{
		    pacc -> nsev = VALID_ALARM;
		    pacc -> nsta = PARAMETER;
		}
	    }
	    else
	    {
		status = dbGetLink (&pacc->afis.value.db_link, 
				    (struct dbCommon *)pacc,
				    DBR_DOUBLE, (caddr_t)&afiv, 
				    &options, &nRequest);
		if (afiv == pacc->afiv) 
		{
		    /* only freeze on transition state */
		    if (afiv != pacc -> lasf)
			freezeflag = TRUE;
		}
		pacc -> lasf = afiv;
	    }
	} 
	/* we must be a timed freeze */
	else 
	{
	    /* if we are at or have gone past next freeze time */

	    if (pacc -> nftm.secPastEpoch <= timenow.secPastEpoch)
	    {
		freezeflag = TRUE;
	    }
	}

	alarm (pacc);

	/* process maximize severity */

	if (pacc -> msev > pacc -> nsev)
	{
	    pacc -> nsev = pacc -> msev;
	    pacc -> nsta = LINK_ALARM;
	}

    } /* end on scan */

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (pacc -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&pacc -> latm);
	db_post_events ((struct dbCommon *)pacc, (caddr_t)&pacc -> val, DBE_VALUE | DBE_ALARM);
    }

    /* if it is the 1st process pass 
       init last value when alarm chg'd (lalm)
    */

    if (pacc -> fpps)
    {
	pacc -> fpps = 0;
	pacc -> lalm = pacc -> val;
    }

    monitor (pacc);

    /* update running status and severity during freeze cycle */

    if (pacc -> sevr > pacc -> nfsv)
    {
	pacc -> nfsa = pacc -> stat;
	pacc -> nfsv = pacc -> sevr;
	update_crc = 1;
    }

    /* if it is time to freeze, do it now, done here because we need
       nfsa and nfsv updated with current stat and sevr as set
       in monitor, also calculate the time for the next freeze
    */
    if (freezeflag) 
    {
	timenow.nsec = 0;	/* zero nanoseconds */
	freeze (pacc, timenow, &update_crc, &did_a_reset);
	did_a_freeze = 1;

	/* calculate time for the next freeze if TIMED */

	if (pacc -> afmd == TIMED)
	{
	    if (pacc -> aftm <= 0)
	    {
		recGblRecordError (S_db_badField, (struct dbCommon *)pacc, 
				   "acc, process, 0 freeze time");
		status = S_db_badField;
		if (pacc -> nsev < VALID_ALARM)
		{
		    pacc -> nsev = VALID_ALARM;
		    pacc -> nsta = PARAMETER;
		}
	    }
	    else
	    {
		if (pacc->aftu == MINUTES) 
		{
		    frzratesecs = pacc->aftm * 60;
		}		 
		else 
		{
		    frzratesecs = pacc->aftm * 3600;
		}
		firstFreeze(pacc, timebuf, frzratesecs);
		tsTextToStamp (&freezetime, &ptimebuf);
		while (freezetime.secPastEpoch <= timenow.secPastEpoch) 
		{
		    freezetime.secPastEpoch += frzratesecs;
		}
		freezetime.nsec = 0;
		pacc -> nftm = freezetime;
	    }
	}
    }

    /* is it time to save the running totals,
       srvt is in minutes, lsav is in epics seconds
       do this here, because freeze() may zap 
       running values if we reset on freeze
    */

    if (pacc -> srvt && (!(pacc -> tagd)) && (pacc -> oscn != OFF_SCAN) )
    {
	if ( (pacc -> lptm.secPastEpoch > pacc -> lsav) && 
	     ( (pacc -> lptm.secPastEpoch - pacc -> lsav) / 60) >= pacc -> srvt)
	{
	    pacc -> ssmp = pacc -> rsmp;
	    pacc -> stot = pacc -> rtot;
	    pacc -> smin = pacc -> rmin;
	    pacc -> smax = pacc -> rmax;
	    pacc -> sstt = pacc -> atot;

	    /* use the delta scan time in seconds as a deadband value, since we
	       may not be an exact delta time of the scan rate or we may be
	       passive
	    */
	    while ( (pacc -> lsav + time_adj) < pacc -> lptm.secPastEpoch)
	    {
		pacc -> lsav += (pacc -> srvt * 60);
	    }
	    update_crc = 1;
	}
	/* if we did a reset at freeze time, them force a save, but
	   do NOT mess with the LSAV, the next save time
	*/
	else if (did_a_reset)
	{
	    pacc -> ssmp = pacc -> rsmp;
	    pacc -> stot = pacc -> rtot;
	    pacc -> smin = pacc -> rmin;
	    pacc -> smax = pacc -> rmax;
	    pacc -> sstt = pacc -> atot;
	}
    }

    /* post events and update redundancy */

    post_and_redund (pacc, &old_acc, update_crc, did_a_freeze);

    pacc->pact = FALSE;

    /* process the forward scan link record */

    if (pacc -> flnk.type == DB_LINK) 
	dbScanPassive ((struct dbAddr *)pacc -> flnk.value.db_link.pdbAddr);

    return (status);
}


/*------------------------------------------------------------------------*/

/******
	called startup.cmd or at ioc prompt to read the timed reset file
	and call parse_rgrp()
******/

int read_rgrps ()
{
    FILE	*fd;
    struct rgrp *new_rgrp;
    int		ret = 0;
    int		status = 0;
    int		errors = 0;
    int		lineno = 0;
    char	rbuf[256];
    char	*p;

    free_rgrps ();
    if ( (fd = fopen ("/tis4000db/accumReset.txt", "r")) == NULL)
    {
	logMsg ("read_rgrps, open failed\n",0,0,0,0,0,0);
	status = 1;
	goto setdef;
    }

    ret = 0;
    while ( (fgets (rbuf, sizeof (rbuf), fd)) != NULL)
    {
	lineno++;
	/* zap the newline if it exists */
	if ((p = (char *) index(rbuf, '\n')) )
	    *p = NULL;
	/* ignore blank line */
	if ( (rbuf[0] == ' ') || (rbuf[0] == 0) )
	    continue;
	if ((ret = parse_rgrp (rbuf)))
	{
	    logMsg ("read_rgrps, parse error line %d\n", lineno,0,0,0,0,0);
	    errors++;
	}
    }
    fclose (fd);

    if (errors)
	goto setdef;
    else
    {
	/* set rgrps[0] to indicate we have a valid list */
	if ( (new_rgrp = (struct rgrp *)calloc (1, sizeof (struct rgrp))) == NULL)
	{
	    logMsg ("read_rgrps, calloc failed\n",0,0,0,0,0,0);
	    status = 1;
	    goto setdef;
	}
	else
	    rgrps[0] = new_rgrp;
    }
    return (status);

setdef:
    logMsg ("read_rgrps errors, setting defaults\n",0,0,0,0,0,0);
    set_rgrp_defaults ();
    return (status);
}


/*------------------------------------------------------------------------*/

/******
	from ioc prompt, get status of time reset group data
******/

int rgrp_status ()
{
    struct rgrp	*rgrp;
    char	pline[256];
    char	buff[32];
    int		i;
    static char	*dow[] = {"   ","sun","mon","tue","wed","thu","fri","sat"};
    static char *mon[] = {
	"   ","jan","feb","mar","apr","may","jun","jul","aug","sep","oct","nov","dec"};
    static char *rtypes[] = {
	"          ",
	"d-o-w     ",
	"d-o-w-time",
	"d-o-m     ",
	"d-o-m-time",
	"month     ",
	"month-time",
	"time      ",
	"daily     ",
	"weekly    ",
	"monthly   ",
	};

    for (i = 1; i < 256; i++)
    {
	if (rgrps[i] != 0)
	{
	    bzero (pline, sizeof (pline));
	    rgrp = rgrps[i];
	    sprintf (buff, "%03d,", i);
	    strcat (pline, buff);
	    if ( (rgrp -> type < 0) || (rgrp -> type >= LAST_RESET_TYPE) )
		strcpy (buff, "unknown   ,");
	    else
		sprintf (buff, "%s,", rtypes[rgrp -> type]);
	    strcat (pline, buff);
	    if ( (rgrp -> day_of_week < 0) || (rgrp -> day_of_week > 7) )
		strcpy (buff, "unk,");
	    else
		sprintf (buff, "%s,", dow[rgrp -> day_of_week]);
	    strcat (pline, buff);
	    if ( (rgrp -> month < 0) || (rgrp -> month > 12) )
		strcpy (buff, "unk,");
	    else
		sprintf (buff, "%s,", mon[rgrp -> month]);
	    strcat (pline, buff);
	    if (rgrp -> day_of_month)
		sprintf (buff, "%02d,", rgrp -> day_of_month);
	    else
		strcpy (buff, "  ,");
	    strcat (pline, buff);
	    if (rgrp -> hour == -1)
		strcpy (buff, "  :  ");
	    else
		sprintf (buff, "%02d:%02d", rgrp -> hour, rgrp -> minute);
	    strcat (pline, buff);
	    logMsg ("%s\n", (int)pline,0,0,0,0,0);
	}
    }
    return (0);
}



/*------------------------------------------------------------------------*/

/******
	save current record values for
	redundant update and posting of events
******/

static void save_accum (pacc, old_acc)
    struct accumRecord	*pacc;
    struct accred	*old_acc;
{
    old_acc -> stat = pacc -> stat;
    old_acc -> sevr = pacc -> sevr;
    old_acc -> val = pacc -> val;
    old_acc -> rsmp = pacc -> rsmp;
    old_acc -> rtot = pacc -> rtot;
    old_acc -> ravg = pacc -> ravg;
    old_acc -> rmin = pacc -> rmin;
    old_acc -> rmax = pacc -> rmax;
    old_acc -> ftim = pacc -> ftim;
    old_acc -> fsmp = pacc -> fsmp;
    old_acc -> ftot = pacc -> ftot;
    old_acc -> favg = pacc -> favg;
    old_acc -> fmin = pacc -> fmin;
    old_acc -> fmax = pacc -> fmax;
    old_acc -> lalm = pacc -> lalm;
    old_acc -> ini2 = pacc -> ini2;
    old_acc -> ltck = pacc -> ltck;
    old_acc -> lmod = pacc -> lmod;
    old_acc -> stot = pacc -> stot;
    old_acc -> ssmp = pacc -> ssmp;
    old_acc -> smin = pacc -> smin;
    old_acc -> smax = pacc -> smax;
    old_acc -> sstt = pacc -> sstt;
    old_acc -> lars = pacc -> lars;
    old_acc -> arsf = pacc -> arsf;
    old_acc -> fsta = pacc -> fsta;
    old_acc -> fsev = pacc -> fsev;
    old_acc -> nfsa = pacc -> nfsa;
    old_acc -> nfsv = pacc -> nfsv;
    old_acc -> lsav = pacc -> lsav;
    old_acc -> atot = pacc -> atot;
    old_acc -> lasf = pacc -> lasf;
    old_acc -> talc = pacc -> talc;
    old_acc -> nftm = pacc -> nftm;
    old_acc -> latm = pacc -> latm;
    old_acc -> rsec = pacc -> rsec;
    strncpy (old_acc -> lrst, pacc -> lrst, sizeof (pacc -> lrst));
    old_acc -> mlst = pacc -> mlst;
    old_acc -> hhll = pacc -> hhll;
    old_acc -> hilo = pacc -> hilo;
    old_acc -> rtct = pacc -> rtct;
    old_acc -> larf = pacc -> larf;
    return;
}


/*------------------------------------------------------------------------*/

/******
	set defaults for rgrps[];
		[0] = blank struct, used as indicator that rgrps has valid
			ptrs
		[1] = daily reset, reset on first freeze of new day
		[2] = weekly reset, reset on first freeze of new week, sunday
		[3] = monthly reset, reset on first freeze of new month
******/

int set_rgrp_defaults ()
{
    struct rgrp	*new_rgrp;
    int		i;

    free_rgrps ();
    for (i = 0; i < 4; i++)
    {
	if ( (new_rgrp = (struct rgrp *)calloc (1, sizeof (struct rgrp))) == NULL)
	{
	    logMsg ("set_rgrp_defaults, calloc failed\n",0,0,0,0,0,0);
	    free_rgrps ();
	    return (1);
	}
	if (i == 0)
	    new_rgrp -> type = 0;
	else if (i == 1)
	    new_rgrp -> type = DAILY_RESET;
	else if (i == 2)
	    new_rgrp -> type = WEEKLY_RESET;
	else if (i == 3)
	    new_rgrp -> type = MONTHLY_RESET;
	new_rgrp -> hour = -1;
	new_rgrp -> minute = -1;
	rgrps[i] = new_rgrp;
    }
    return (0);
}



/*------------------------------------------------------------------------*/

static long special (paddr,after)
    struct dbAddr	*paddr;
    int			after;
{
    struct accumRecord	*pacc = (struct accumRecord *)(paddr->precord);
    int           	special_type = paddr->special;
    static double       save_rtot, save_ravg, save_rmin, save_rmax, save_rsmp;
    long		status = 0;
    unsigned short	i;
    int			master = 0;
    TS_STAMP		timenow;

    /* Special is called twice from the put function, once before the field
       is updated and once after. Save running totals, if in AUTOMATIC (NORMAL) 
       mode, do not let update of these fields happen only if MASTER, if slave
       let update happen, specials are called when slave gets
       redundant update from master, and we do not want to 
       miss this data
    */

    if (!after)
    {
       save_rtot = pacc -> rtot;
       save_ravg = pacc -> ravg;
       save_rmin = pacc -> rmin;
       save_rmax = pacc -> rmax;
       save_rsmp = pacc -> rsmp;
       return (status);
    }

    if (resGetCaAnswer (pacc -> radr, &i) == RES_OK)
    {
	if (i)
	    master = 1;
    }

    switch (special_type) 
    {
	case SPC_ACC_RTOT:
	    if ( (pacc -> mode == NORMAL) && master)
	    {
		pacc -> rtot = save_rtot;
		break;
	    }
	    break;

	case SPC_ACC_RAVG:
	    if ( (pacc -> mode == NORMAL) && master)
	    {
		pacc -> ravg = save_ravg;
		break;
	    }
	    break;

	case SPC_ACC_RMIN:
	    if ( (pacc -> mode == NORMAL) && master)
	    {
		pacc -> rmin = save_rmin;
		break;
	    }
	    break;

	case SPC_ACC_RMAX:
	    if ( (pacc -> mode == NORMAL) && master)
	    {
		pacc -> rmax = save_rmax;
		break;
	    }
	    break;

	case SPC_ACC_RSMP:
	    if ( (pacc -> mode == NORMAL) && master)
	    {
		pacc -> rsmp = save_rsmp;
		break;
	    }
	    break;

	case SPC_ACC_LTCK:
	    pacc -> ltck = tickGet ();
	    break;

	case SPC_ACC_NFTM:	/* on fields, aftm,aftu,fafh,fafm */
	    {
		double		frzratesecs;
		char		*ptimebuf;
		char		timebuf[32];
		TS_STAMP	next_freeze;

		if (pacc -> aftm <= 0)
		{
		    recGblRecordError (S_db_badField, (struct dbCommon *)pacc, 
					"acc, special, 0 freeze time");
		    status = S_db_badField;
		    break;
		}

		if (pacc->aftu == MINUTES) 
		{
		    frzratesecs = pacc->aftm * 60;
		}     
		else 
		{
		    frzratesecs = pacc->aftm * 3600;
		}

		/* calc first freeze time as string */
		firstFreeze(pacc, timebuf, frzratesecs);

		/* convert to epics time */
		ptimebuf = timebuf;
		tsTextToStamp (&next_freeze, &ptimebuf);

		/* get next freeze time */
		tsLocalTime(&timenow);
		while (next_freeze.secPastEpoch <= timenow.secPastEpoch) 
		{
		    next_freeze.secPastEpoch += frzratesecs;
		}
		next_freeze.nsec = 0;	/* zero nanoseconds */
		pacc -> nftm = next_freeze;
	    }
	    break;

	case SPC_ACC_FSEC:
	    /* recreate the freeze time string */
	    timenow.secPastEpoch = pacc -> fsec;
	    timenow.nsec = 0;
	    tsStampToText (&timenow, TS_TEXT_MMDDYY, pacc -> fts0);
	    break;

	case SPC_ACC_RSEC:
	    /* recreate the last reset time string */
	    timenow.secPastEpoch = pacc -> rsec;
	    timenow.nsec = 0;
	    tsStampToText (&timenow, TS_TEXT_MMDDYY, pacc -> lrst);
	    break;

	default:
	    recGblDbaddrError (S_db_badChoice, paddr, "accum: special");
	    status = S_db_badChoice;
	    break;
    }
    return (status);
}


/*------------------------------------------------------------------------*/

/******
	if doing a time reset (rgrp), then test if it is time to do a reset
	NOTE:	
		see .../epicsH/tsDefs.h for def of TS_STAMP, TS_EPOCH_WDAY_NUM
			and struct tsDetail
		tsStampToLocal() see .../src/libCom/tsSubr.c, converts a timestamp
			which is in GMT to local time
******/

static int test_rgrp_reset (pacc, last_freeze)
    struct accumRecord	*pacc;
    TS_STAMP		last_freeze;
{
    int			last_day = 0;
    int			last_day_of_week = 0;
    int			last_month = 0;
    int			today, day_of_week, month;
    int			year, hour, minute;
    int			last_day_of_month = 0;
    int			valid_last_freeze = 0;
    int			leap = 0;
    int			ret;
    struct tsDetail	tsD;
    static int	daytab[13] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

    /* legal values 0 - 255 */

    if ( (pacc -> rgrp < 0) || (pacc -> rgrp > 255) )
    {
	logMsg ("test_rgrp_reset, rgrp range (%d)\n", pacc -> rgrp,0,0,0,0,0);
	return (0);
    }

    /* rgrp not defined */

    if (rgrps[pacc -> rgrp] == 0)
	return (0);

    /* if there was a valid last freeze, get vital stuff */

    if (last_freeze.secPastEpoch > 110000)
    {
	tsStampToLocal (last_freeze, &tsD);
	last_month       = tsD.monthNum + 1;
	last_day         = tsD.dayMonth + 1;
	last_day_of_week = tsD.dayOfWeek + 1;
	valid_last_freeze = 1;
    }

    /* get vital stuff from the current freeze */
    
    tsStampToLocal (pacc -> ftim, &tsD);
    month       = tsD.monthNum + 1;
    today       = tsD.dayMonth + 1;
    hour        = tsD.hours;
    minute      = tsD.minutes;
    day_of_week = tsD.dayOfWeek + 1;
    year        = tsD.year;
    leap	= tsD.leapYear;

    last_day_of_month = daytab[month];

    /* if it is feb, handle leap year */
    if (month == 2)
	last_day_of_month += leap;

    switch (rgrps[pacc -> rgrp] -> type)
    {
	case DAY_OF_WEEK_RESET:
	    if (day_of_week == rgrps[pacc -> rgrp] -> day_of_week)
		return (1);
	    break;

	case DAY_OF_WEEK_TIME_RESET:
	    if (day_of_week == rgrps[pacc -> rgrp] -> day_of_week)
	    {
		if (hour == rgrps[pacc -> rgrp] -> hour)
		{
		    if (minute == rgrps[pacc -> rgrp] -> minute)
			return (1);
		}
	    }
	    break;

	case DAY_OF_MONTH_RESET:
	    /* if we are looking for the last day of the month */
	    if (rgrps[pacc -> rgrp] -> day_of_month == 32)
	    {
		if (today == last_day_of_month)
		    return (1);
	    }
	    /* may have asked for the 31st of the month, and we are in a
	       month with less days
	    */
	    else if ( (rgrps[pacc -> rgrp] -> day_of_month > last_day_of_month) &&
		      ( today == last_day_of_month) )
		return (1);
	    else if (today == rgrps[pacc -> rgrp] -> day_of_month)
		return (1);
	    break;

	case DAY_OF_MONTH_TIME_RESET:
	    ret = 0;
	    if (rgrps[pacc -> rgrp] -> day_of_month == 32)
	    {
		if (today == last_day_of_month)
		    ret = 1;
	    }
	    else if ( (rgrps[pacc -> rgrp] -> day_of_month > last_day_of_month) &&
		      (today == last_day_of_month) )
		ret = 1;
	    else if (today == rgrps[pacc -> rgrp] -> day_of_month)
		ret = 1;

	    /* if we have the correct day, is it the correct time */
	    if (ret)
	    {
		if (hour == rgrps[pacc -> rgrp] -> hour)
		{
		    if (minute == rgrps[pacc -> rgrp] -> minute)
			return (1);
		}
	    }
	    break;

	case MONTH_RESET:
	    /* do we have the correct month */
	    if (month == rgrps[pacc -> rgrp] -> month)
	    {
		/* and do we have the correct day in the month,
		   remember for the the month reset we must have
		   a day in the month, see parse_rgrp()
		*/
		if (rgrps[pacc -> rgrp] -> day_of_month == 32)
		{
		    if (today == last_day_of_month)
			return (1);
		}
		else if ( (rgrps[pacc -> rgrp] -> day_of_month > last_day_of_month) &&
			  (today == last_day_of_month) )
		    return (1);
		else if (today == rgrps[pacc -> rgrp] -> day_of_month)
		    return (1);
	    }
	    break;

	case MONTH_TIME_RESET:
	    ret = 0;
	    /* do we have the correct month */
	    if (month == rgrps[pacc -> rgrp] -> month)
	    {
		/* and do we have the correct day in the month,
		   remember for the the month reset we must have
		   a day in the month, see parse_rgrp()
		*/
		if (rgrps[pacc -> rgrp] -> day_of_month == 32)
		{
		    if (today == last_day_of_month)
			ret = 1;
		}
		else if ( (rgrps[pacc -> rgrp] -> day_of_month > last_day_of_month) &&
			  (today == last_day_of_month) )
		    ret = 1;
		else if (today == rgrps[pacc -> rgrp] -> day_of_month)
		    ret = 1;
	    }

	    /* if we have the correct month/day, is it the correct time */
	    if (ret)
	    {
		if (hour == rgrps[pacc -> rgrp] -> hour)
		{
		    if (minute == rgrps[pacc -> rgrp] -> minute)
			return (1);
		}
	    }
	    break;

	case TIME_RESET:
	    if (hour == rgrps[pacc -> rgrp] -> hour)
	    {
		if (minute == rgrps[pacc -> rgrp] -> minute)
		    return (1);
	    }
	    break;

	case DAILY_RESET:
	    if (valid_last_freeze)
	    {
		if (last_day != today)
		    return (1);
	    }
	    break;
	    
	case WEEKLY_RESET:
	    if (valid_last_freeze)
	    {
		/* if last freeze was on saturday and today is sunday */

		if ( (last_day_of_week == 7) && (day_of_week == 1) )
		    return (1);
	    }
	    break;

	case MONTHLY_RESET:
	    if (valid_last_freeze)
	    {
		if (last_month != month)
		    return (1);
	    }
	    break;

	default: 
	    break;
    }

    return (0);
}

