/*recAo.c,v 1.1.1.1 1999/08/03 14:44:18 tis4000 Exp */
  
/* recAo.c - Record Support Routines for Analog Output records */
/*
 *      Original Author: Bob Dalesio
 *      Current Author:  Marty Kraimer
 *      Date:            7-14-89
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  09-26-88        lrd     interface the Ziomek085 card
 * .02  12-12-88        lrd     lock record on entry unlock on exit
 * .03  12-15-88        lrd     Process the forward scan link
 * .04  12-23-88        lrd     Alarm on locked MAX_LOCKED times
 * .05  01-13-89        lrd     deleted db_write_ao
 * .06  01-20-89        lrd     fixed vx includes
 * .07  03-03-89        lrd     mods for closed loop/supervisory control
 * .08  03-17-89        lrd     add read_ao routine and call at initialization
 * .09  03-23-89        lrd     convert AB readbacks
 * .10  03-29-89        lrd     make hardware errors MAJOR
 *                              remove hw severity spec from database
 * .11  04-06-89        lrd     remove signal conversions
 * .12  05-03-89        lrd     removed process mask from arg list
 * .13  05-08-89        lrd     fixed init to unlock on return condition
 * .14  05-25-89        lrd     added rate of change add incremental/absolute
 * .15  01-31-90        lrd     add plc_flag to ab_aodriver
 * .16  03-21-90        lrd     add db_post_events for RVAL and RBV
 * .17  04-11-90        lrd     make locals static
 * .18  07-27-90        lrd     implement the output to a database record
 * .19  10-10-90	mrk	extensible record and device support
 * .20  03-12-93	ems	deleted ack from redundancy
 * .21  11-23-93	ems	added cvnTypes.h
 *				use conversion types as defined in cvnTypes.h
 * .22	12-01-93	ems	added slope_and_offset() to calc eslo and roff
 *				do here instead of in devAo***.c, also do here
 *				for SPC_LINCONV
 *				modified formula for LINEAR conversion
 * .23	01-03-94	ems	ack = 0 in init_record, in process if not init'd
 *				do nothing
 * .24  01-05-94	ems	added SPC_CNST to special func
 * .25  02-25-94	ems	oval = pval = val in init_record
 * .26  03-01-94	ems	added support for process stats
 * .27  03-10-94	ems	alarm chgs, if val is past alarm setting, set that
 *				alarm condition, else if val was in that alarm and
 *				it is still in hyst range of alarm value, then stay
 *				in alarm, remove return from high alarm, let user
 *				determine which is more important.
 * .28	03-29-94	ems	make forward func def agree with actual func def
 * .29  04-15-94	ems	added INI2 to namelist, chg NUMFLDS to NUMAOFLDS
 * .30	04-20-94	ems	major chg to way block inits, added ini2 as a secondary
 *				init flag to pass to slave when master inits, the
 *				combination of init and ini2 flags allows the slave to
 *				fully init so it can take when fail over occurs
 * .31	04-27-94	ems	get_control_double, if field is HYST then low = 0
 *				and high is hopr-lopr
 * .32	06-13-94	ems	added support for maximize severity (msev)
 * .33  07-12-94	ems	init_record, if eslo is 0, set val = roff
 *				slope_and_offset, chg to eguf-egul instead of ==
 * .34	08-05-94	ems	added many fields to get_graphic_double and
 *				get_control_double
 * .35	08-17-94	ems	monitor(), rm DBE_LOG from ack, only archive val if
 *				tenb (trending) is enabled, rm DBE_LOG from oval
 *				only use DBE_VALUE for rval,oval and rbv
 * .36	09-23-94	ems	chg size of redundand buffer,redbuf, from 512 to 128, 
 *				all that is really needed is 56, this reduces the
 *				stack required by init_record() and process()
 * .37	10-12-94	ems	add TIME to redundancy,  added fpps
 *				(1st process pass), do time stamp when stat or sevr chg,
 *				see monitor(),and not in process(),set lalm to val when 
 *				stat or sevr chg,in process() if fpps (first process)
 *				init lalm to val and if time stamp is 0, init it also
 * .38	10-20-94	ems	chg time to latm (last alarm time) added 
 *				added lptm (last process time)
 * .39	01-04-95	ems	get_control_double(), return drvh and drvl for VAL
 * .40	03-15-95	ems	monitor(), do not set ACK if sevr is EVENT_ALARM
 * .41	09-11-95	ems	monitor(), archive when stat/sevr chg's also
 * .42	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 * .43	11-13-95        wep     init_record() was modified to handle a returned status 
 *				value of 3. response is valid, but no conversion is 
 *				done, since the dev level code already converted and 
 *				stored value in VAL. (implemented in support of the 
 *				Tosg2 driver)
 * .44	01-31-96	ems	added support for output on chg only OGAP
 * .47	02-07-96	ems	remove return from HIHI and LOLO alarms, so will always check 
 *				    all alarm sevrs
 * .48	04-02-96	ems	added code to support alarm delay
 * .49	06-19-96	ems	added code for alarm disable ADIS
 */

#include	<vxWorks.h>
#include	<inetLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<string.h>

#include	"alarm.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"special.h"
#include	"recSup.h"
#include	"rec/aoRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"aoredfldidx.h"
#include	"restable.h"
#include	"cvnTypes.h"
#include	"db_addr.h"
#include	"dbEvent.h"
#include	"recGbl.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/

#define report NULL
#define initialize NULL
static long init_record ();
static long process ();
static long special ();
static long get_value ();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units ();
static long get_precision ();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double ();
static long get_control_double ();
static long get_alarm_double ();

struct rset aoRSET = {

	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
	};

struct aodset {				/* analog output dset */
	long		number;
	DEVSUPFUN	dev_report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;	/*returns: (0,2)=>(success,success no convert)*/
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	write_ao;	/*(0,1)=>success and */
					/*(continue, don`t continue) */
	DEVSUPFUN	special_linconv;
	DEVSUPFUN	readback_ao;
	};

/* The following defintions are required for redundancy */

static struct fldDes	**mylist = 0;

static char *namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"LALM",
		"INI2",
		"LATM",
		"MLST",
		"LOVL",
		"HHLL",
		"HILO",
		"RTCT",
		"LARF",
		0
		};

static	unsigned	short	type;

/* the following definitions must match those in choiceRec.ascii */

#define OUTPUT_FULL 		0
#define OUTPUT_INCREMENTAL	1

static void alarm ();
static int convert ();
static void monitor ();
static void slope_and_offset ();
static void test_ogap ();
static void aoReadback ();


/*------------------------------------------------------------------------*/

static long init_record (pao)
    struct	aoRecord	*pao;
{
    struct	aodset	*pdset;
    long		status = 0;
    unsigned	short	i;
    char		*pstr;
    char		redbuf [128];
    int			ini2_flag = 0;
    static int		do_convert;

	if (!(mylist)) 
	{
	    for (type = 1; type < 100; type++) 
	    {
		pstr = GET_PRECTYPE (type);
		if (strcmp (pstr, "ao") == 0) 
		    break;
	    }
	    mylist = redFldAlloc (type, NUMAOFLDS);

	    if (!mylist)
	    {
		recGblRecordError (S_rec_Alloc, (struct dbCommon *)pao,
				   "ao, init_record, mylist failed");
		return (S_rec_Alloc);
	    }

	    for (i = 0; i < NUMAOFLDS; i++) 
		redFldAdd (type, i, namelist [i]);
	}

        if (!(pdset = (struct aodset *) (pao->dset))) 
	{
	    recGblRecordError (S_dev_noDSET, (struct dbCommon *)pao, "ao: init_record");
	    return (S_dev_noDSET);
        }

        /* must have write_ao function defined */

        if ((pdset->number < 6) || (pdset->write_ao == NULL)) 
	{
	    recGblRecordError (S_dev_missingSup, (struct dbCommon *)pao, "ao: init_record");
	    return (S_dev_missingSup);
        }

	pao->init = TRUE;
	pao->fpps = 1;

	/* init alarm delay variables */
	pao -> larf = 0;
	pao -> hhll = 0;
	pao -> hilo = 0;
	pao -> dvct = 0;
	pao -> rtct = 0;

	/* init readback variables */
	pao -> rbfn = aoReadback;

	/* init last process time and last alarm time */

	if (pao->lptm.secPastEpoch == 0)
	{
	    tsLocalTime (&pao->lptm);
	}
	if (pao->latm.secPastEpoch == 0)
	{
	    tsLocalTime (&pao->latm);
	}
	/* init so do not carry to process when init succeeds, if set by previous
	   failure of init
	*/
	pao -> nsta = NO_ALARM;
	pao -> nsev = NO_ALARM;
	pao -> msev = NO_ALARM;

	do_convert = 1;

	/* do not allow conversion if the board type is CONSTANT and output
	   is CONSTANT or REFERENCE, this was done in devSoft.c
	*/

        if (pao -> dtyp <= 1)
	{
	    switch (pao -> out.type)
	    {
		case (CONSTANT):
		case (PV_LINK):
		case (DB_LINK):
		case (CA_LINK):
		    if (pao -> linr != NOCONV)
			pao -> linr = NOCONV;

		    /* make sure we do not do any converison when we return
		       from the device init
		    */
		    do_convert = 0;
		    break;
	    }
        }

	/* calculate conversion slope eslo and offset roff
	   do before pdset->init_record in case there is a dev that
	   has to recalc eslo and roff for some reason
	*/

	slope_and_offset (pao);

        if ( pdset->init_record ) 
	{
	    status = (*pdset->init_record) (pao, process);
	    if ((status) && (status != 3))
	    {
		pao -> nsta = NO_INIT_ALARM;
		pao -> nsev = VALID_ALARM;
	    }
	    else
	    {
		ini2_flag = 1;
                if (status == 3)
                {
                    do_convert = 0;
                }
		if (do_convert)
		{
                    if (pao->linr == LINEAR) 
		    {
			if (pao -> eslo == 0)
			    pao -> val = pao -> roff;
			else
                            pao->val = (pao->rval - pao->roff) / pao->eslo;
		    }
		    else 
		    {
		        if (pao->linr == NOCONV)
                            pao->val = pao->rval;
                    }
		}
	        pao->udf = FALSE;
	    }
        }
	else
	    ini2_flag = 1;

        pao -> lovl = pao->oval = pao->pval = pao->val;

	if (resGetCaAnswer(pao -> radr, &i) == RES_OK)
	{
	    if (i)
	    {
		/* post STAT, SEVR and VAL (monitor), if we will finally init, 
		   because we have read a value from the input device, set INI2,
		   send all redund fields to slave, when slave receives INI2, 
		   it will be fully init'd also
		*/

		alarm (pao);

	    	/* check for maximize severity */

	    	if (pao -> msev > pao -> nsev)
	    	{
		    pao -> nsev = pao -> msev;
		    pao -> nsta = LINK_ALARM;
	    	}

		monitor (pao);

		if (ini2_flag)
		{
		    pao -> ini2 = 1;

		    redQInit ((struct dbCommon *)pao, redbuf, sizeof (redbuf), type);
		    redQField (redbuf, mylist [STAT]);
		    redQField (redbuf, mylist [VAL]);
		    redQField (redbuf, mylist [SEVR]);
		    redQField (redbuf, mylist [LALM]);
		    redQField (redbuf, mylist [INI2]);
		    redQField (redbuf, mylist [LOVL]);
		    redQSend (redbuf);
		}
	    }
	}
        return(0);
}


/*------------------------------------------------------------------------*/

static long process (paddr)

    struct	dbAddr	*paddr;
{
    struct aoRecord	*pao   = (struct aoRecord *)(paddr->precord);
    struct aodset	*pdset = (struct aodset *)(pao->dset);
    long		status = 0;
    unsigned short	stat   = pao -> stat;
    unsigned short	sevr   = pao -> sevr;
    double		val    = pao -> val;
    double		lalm   = pao -> lalm;
    TS_STAMP		latm   = pao -> latm;
    double		mlst   = pao -> mlst;
    double		lovl   = pao -> lovl;
    short		hhll   = pao -> hhll;
    short		hilo   = pao -> hilo;
    short		rtct   = pao -> rtct;
    unsigned long	larf   = pao -> larf;
    int			write_output = 1;
    char		redbuf [128];

    /* if not init'd, do nothing */

    if ((!pao -> init) || (!pao -> ini2))
	return (status);

    if( (pdset == NULL) || (pdset->write_ao == NULL) ) 
    {
	pao->pact=TRUE;
	recGblRecordError(S_dev_missingSup, (struct dbCommon *)pao, "write_ao");
	return(S_dev_missingSup);
    }

    /* get time stamp for last time record was processed */

    tsLocalTime (&pao -> lptm);

    pao -> msev = NO_ALARM;

    if (pao -> tagd)
    {
	pao -> nsta = TAGGED_ALARM;
	pao -> nsev = VALID_ALARM;
	pao -> val = pao -> pval;
    }
    else if (pao -> oscn == OFF_SCAN)
    {
	pao -> nsta = SCAN_ALARM;
	pao -> nsev = VALID_ALARM;
	pao -> val = pao -> pval;
    }
    else
    {
	pao -> nsta = NO_ALARM;
	pao -> nsev = NO_ALARM;

	/* update number of process times */
	if (process_stats)
	    process_cnts[type]++;	

	if (pao->pact == FALSE) 
	{
	    /* convert and if success write*/
	    if (convert (pao) == 0) 
	    {
		/* test for output gap before we can output */
		test_ogap (pao, &write_output);
		if (write_output)
		{
		    status = (*pdset->write_ao) (pao);
		    pao -> lovl = pao -> oval;
		}
	    }
	} 
	else
	{
	    /* test for output gap before we output */
	    test_ogap (pao, &write_output);
	    if (write_output)
	    {
		status = (*pdset->write_ao) (pao);
		pao -> lovl = pao -> oval;
	    }
	}

	pao->pact = TRUE;

	if (status == 1) 
	    return (0);
	    
	if (status == -2) 
	    return (0);
	    

	/* check for alarms */

	alarm (pao);

	/* check for maximize severity */

	if (pao -> msev > pao -> nsev)
	{
	    pao -> nsev = pao -> msev;
	    pao -> nsta = LINK_ALARM;
	}
    }

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (pao -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&pao -> latm);
	db_post_events ((struct dbCommon *)pao, (caddr_t)&pao -> val, DBE_VALUE | DBE_ALARM);
    }

    /* if it is the 1st process pass 
       init last value when alarm chg'd (lalm)
    */

    if (pao -> fpps)
    {
	pao -> fpps = 0;
	pao -> lalm = pao -> val;
    }

    /* check event list */

    monitor (pao);

    redQInit ((struct dbCommon *)pao, redbuf, sizeof (redbuf), type);

    if (stat != pao->stat) 
	redQField (redbuf, mylist [STAT]);
	
    if (val != pao->val)
	redQField (redbuf, mylist [VAL]);
	
    if (sevr != pao->sevr)
	redQField (redbuf, mylist [SEVR]);
	
    if (lalm != pao->lalm)
	redQField (redbuf, mylist [LALM]);
	
    if (TsCmpStampsNE (&latm, &pao -> latm) )
	redQField (redbuf, mylist [LATM]);
	
    if (mlst != pao->mlst)
	redQField (redbuf, mylist [MLST]);
	
    if (lovl != pao->lovl)
	redQField (redbuf, mylist [LOVL]);

    if (hhll != pao -> hhll)
	redQField (redbuf, mylist [HHLL]);

    if (hilo != pao -> hilo)
	redQField (redbuf, mylist [HILO]);

    if (rtct != pao -> rtct)
	redQField (redbuf, mylist [RTCT]);

    if (larf != pao -> larf)
	redQField (redbuf, mylist [LARF]);

    redQSend (redbuf);

    /* process the forward scan link record */

    if (pao->flnk.type == DB_LINK) 
	dbScanPassive ((struct dbAddr *)pao->flnk.value.db_link.pdbAddr);
	
    pao->pact = FALSE;

    return (status);
}


/*------------------------------------------------------------------------*/

static long special (paddr,after)
    struct dbAddr *paddr;
    int           after;
{
    struct aoRecord     *pao = (struct aoRecord *)(paddr->precord);
    int                 special_type = paddr->special;
    short		master;
    static double       save_cnst;

    /* if dtyp and out are not CONSTANT, do not let them 
       update CNST
    */

    if (!after)
    {
	save_cnst = pao -> cnst;
	return (0);
    }

    switch(special_type) 
    {

	case(SPC_LINCONV):
	    slope_and_offset (pao);
	    break;

	case (SPC_CNST):
	    if ((pao -> dtyp <= 1) && (pao -> out.type == CONSTANT))
            {
	        pao -> out.value.value = pao -> cnst;
                pao -> rval = pao -> val = pao -> cnst;
            }
	    else
	        pao -> cnst = save_cnst;
	    break;

	case (SPC_AO_VAL):
	    /* if output gap is set and I am in SUPERVISORY mode 
	       (NOTE: dol!=link and omsl=closed_loop acts like supervisory), and 
	       my output channel is not a reference, and my resource is MASTER , 
	       set LOVL to this value, this will ensure that the VAL the user 
	       input will be output
	    */
	    if ( (pao -> ogap == 0) || 
		 ( (pao -> dol.type == DB_LINK) && (pao -> omsl == CLOSED_LOOP) ) ||
		 ( (pao -> dtyp <= 1) && (pao -> out.type == DB_LINK) ) )
		break;

	    if (resGetCaAnswer (pao -> radr, &master) == RES_OK)
	    {
		if (master)
		{
		    pao -> lovl = -999999999;
		}
	    }
	    break;

	default:
            recGblDbaddrError(S_db_badChoice,paddr,"ao: special");
            return(S_db_badChoice);
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static long get_value(pao,pvdes)
    struct aoRecord		*pao;
    struct valueDes	*pvdes;
{
    pvdes->field_type = DBF_DOUBLE;
    pvdes->no_elements=1;
    (double *)(pvdes->pvalue) = &pao->val;
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_units(paddr,units)
    struct dbAddr *paddr;
    char	  *units;
{
    struct aoRecord	*pao=(struct aoRecord *)paddr->precord;

    strncpy(units,pao->egu,sizeof(pao->egu));
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_precision(paddr,precision)
    struct dbAddr *paddr;
    long	  *precision;
{
    struct aoRecord	*pao=(struct aoRecord *)paddr->precord;

    *precision = pao->prec;
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_alarm_double(paddr,pad)
    struct dbAddr *paddr;
    struct dbr_alDouble	*pad;
{
    struct aoRecord	*pao=(struct aoRecord *)paddr->precord;

    pad->upper_alarm_limit = pao->hihi;
    pad->upper_warning_limit = pao->high;
    pad->lower_warning_limit = pao->low;
    pad->lower_alarm_limit = pao->lolo;
    return(0);
}


/*------------------------------------------------------------------------*/

static long get_graphic_double(paddr,pgd)
    struct dbAddr *paddr;
    struct dbr_grDouble	*pgd;
{
    struct aoRecord	*pao=(struct aoRecord *)paddr->precord;
    float    span;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct aoRecord *)0) -> oroc) ):
	case ( (int)&(((struct aoRecord *)0) -> adel) ):
	case ( (int)&(((struct aoRecord *)0) -> mdel) ):
	    span = pao -> eguf - pao -> egul;
	    if (span < 0.0)
	        span = -span;
	    pgd -> upper_disp_limit = span;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> eguf) ):
	case ( (int)&(((struct aoRecord *)0) -> egul) ):
	case ( (int)&(((struct aoRecord *)0) -> drvh) ):
	case ( (int)&(((struct aoRecord *)0) -> drvl) ):
	case ( (int)&(((struct aoRecord *)0) -> hihi) ):
	case ( (int)&(((struct aoRecord *)0) -> high) ):
	case ( (int)&(((struct aoRecord *)0) -> low) ):
	case ( (int)&(((struct aoRecord *)0) -> lolo) ):
	    pgd -> upper_disp_limit = pao -> eguf;
	    pgd -> lower_disp_limit = pao -> egul;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> hopr) ):
	case ( (int)&(((struct aoRecord *)0) -> lopr) ):
	    pgd -> upper_disp_limit = pao -> hopr;
	    pgd -> lower_disp_limit = pao -> lopr;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> rval) ):

	    /* if software point */
	    if ( (pao -> dtyp <=1 ) &&
		 ( (pao -> out.type == CONSTANT) ||
		   (pao -> out.type == PV_LINK) ||
		   (pao -> out.type == DB_LINK) ||
		   (pao -> out.type == CA_LINK) ) )
	    {
		pgd -> upper_disp_limit = pao -> hopr;
		pgd -> lower_disp_limit = pao -> lopr;
	    }
	    /* are hardware */
	    else
	    {
		pgd -> upper_disp_limit = 65535;
		pgd -> lower_disp_limit = 0;
	    }
	    break;

	case ( (int)&(((struct aoRecord *)0) -> hyst) ):
	    span = pao -> hopr - pao -> lopr;
	    if (span < 0.0)
		span = -span;
	    pgd -> upper_disp_limit = span;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> irhi) ):
	case ( (int)&(((struct aoRecord *)0) -> irlo) ):
	    pgd -> upper_disp_limit = 65535;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> rbdt) ):
	    pgd -> upper_disp_limit = 9999;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> ogap) ):
	    span = pao -> drvh - pao -> drvl;
	    if (span < 0.0)
		span = -span;
	    pgd -> upper_disp_limit = span;
	    pgd -> lower_disp_limit = 0;
	    break;

	default:
	    pgd -> upper_disp_limit = pao -> hopr;
	    pgd -> lower_disp_limit = pao -> lopr;
    }

    return(0);
}


/*------------------------------------------------------------------------*/

static long get_control_double(paddr,pcd)
    struct dbAddr *paddr;
    struct dbr_ctrlDouble *pcd;
{
    struct   aoRecord	*pao=(struct aoRecord *)paddr->precord;
    float    span;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct aoRecord *)0) -> val) ):
	    pcd -> upper_ctrl_limit = pao -> drvh;
	    pcd -> lower_ctrl_limit = pao -> drvl;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> oroc) ):
	case ( (int)&(((struct aoRecord *)0) -> adel) ):
	case ( (int)&(((struct aoRecord *)0) -> mdel) ):
	    span = pao -> eguf - pao -> egul;
	    if (span < 0.0)
	        span = -span;
	    pcd -> upper_ctrl_limit = span;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> eguf) ):
	case ( (int)&(((struct aoRecord *)0) -> egul) ):
	case ( (int)&(((struct aoRecord *)0) -> drvh) ):
	case ( (int)&(((struct aoRecord *)0) -> drvl) ):
	case ( (int)&(((struct aoRecord *)0) -> hihi) ):
	case ( (int)&(((struct aoRecord *)0) -> high) ):
	case ( (int)&(((struct aoRecord *)0) -> low) ):
	case ( (int)&(((struct aoRecord *)0) -> lolo) ):
	case ( (int)&(((struct aoRecord *)0) -> hopr) ):
	case ( (int)&(((struct aoRecord *)0) -> lopr) ):
	    pcd -> upper_ctrl_limit = pao -> eguf;
	    pcd -> lower_ctrl_limit = pao -> egul;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> hyst) ):
	    span = pao -> hopr - pao -> lopr;
	    if (span < 0.0)
	        span = -span;
	    pcd -> upper_ctrl_limit = span;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> irhi) ):
	case ( (int)&(((struct aoRecord *)0) -> irlo) ):
	    pcd -> upper_ctrl_limit = 65535;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> ogap) ):
	    span = pao -> drvh - pao -> drvl;
	    if (span < 0.0)
		span = -span;
	    pcd -> upper_ctrl_limit = span;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct aoRecord *)0) -> rbdt) ):
	    pcd -> upper_ctrl_limit = 9999;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	default:
	    pcd -> upper_ctrl_limit = pao -> hopr;
	    pcd -> lower_ctrl_limit = pao -> lopr;
	    break;
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static void alarm (pao)
    struct aoRecord	*pao;
{
    double		val = pao -> val;
    int			adis = 0;
    ALARM_LIMITS	alarm_limits;

    /* check for alarm inhibit */
    adis = AlarmDisableLink ((void *)pao);
    if ( (pao -> aenb == ALARM_DISABLED) || (adis != 0) )
    {
	if (pao -> nsta != NO_INIT_ALARM)
	{
	    pao -> nsta = INHIBIT_ALARM;
	    pao -> nsev = NO_ALARM;
	}
	return;
    }

    if(pao->udf == TRUE )
    {
	if (pao->nsev < VALID_ALARM)
	{
	    pao->nsta = UDF_ALARM;
	    pao->nsev = VALID_ALARM;
	}
	return;
    }

    /* set up for limit alarms */

    alarm_limits.hihi = pao -> hihi;
    alarm_limits.high = pao -> high;
    alarm_limits.low  = pao -> low;
    alarm_limits.lolo = pao -> lolo;
    alarm_limits.devl = 0;
    alarm_limits.devs = 0;
    alarm_limits.hyst = pao -> hyst;
    alarm_limits.hhsv = pao -> hhsv;
    alarm_limits.hsv  = pao -> hsv;
    alarm_limits.lsv  = pao -> lsv;
    alarm_limits.llsv = pao -> llsv;
    alarm_limits.dvsv = NO_ALARM;

    AlarmLimits (&alarm_limits, val, (void *)pao);

    /* do a return to normal, check for alarm delay ALDL only
       on the listed alarms
    */

    if ( ((pao -> nsta == NO_ALARM) && (pao -> nsev == NO_ALARM)) &&
	 ((pao -> stat == HIHI_ALARM) || (pao -> stat == HIGH_ALARM) ||
	  (pao -> stat == LOW_ALARM)  || (pao -> stat == LOLO_ALARM)) )
    {
	if (pao -> aldl > 0)
	{
	    if (pao -> rtct < ALDL_MAX)
		pao -> rtct++;

	    /* delay return to normal so, continue last alarm */
	    if (pao -> rtct < pao -> aldl)
	    {
		pao -> nsta = pao -> stat;
		pao -> nsev = pao -> sevr;
	    }
	}
    }
    else
	pao -> rtct = 0;

    return;
}


/*------------------------------------------------------------------------*/

static int convert (pao)
    struct aoRecord  *pao;
{
	double		value;

        if ((pao->dol.type == DB_LINK) && (pao->omsl == CLOSED_LOOP))
	{
	    long	nRequest;
	    long	options;
	    short	save_pact;
	    long	status;

	    options=0;
	    nRequest=1;
	    save_pact = pao->pact;

	    /* don't allow dbputs to val field */
	    pao->pact = TRUE;

	    pao->val=pao->pval;
            status = dbGetLink(&pao->dol.value.db_link,
			       (struct dbCommon *)pao,
			       DBR_DOUBLE, (caddr_t)&value,
			       &options,&nRequest);
	    pao->pact = save_pact;
	    if(status) 
	    {
	        if(pao->nsev<VALID_ALARM) 
	        {
	            pao->nsta = LINK_ALARM;
	            pao->nsev=VALID_ALARM;
	        }
		return(1);
	    }
            if (pao->oif == OUTPUT_INCREMENTAL) 
		value += pao->val;
        } 
	else 
	    value = pao->val;

        /* check drive limits */
	if(pao->drvh > pao->drvl) 
	{
       	    if (value > pao->drvh) 
	        value = pao->drvh;
       	    else if (value < pao->drvl) 
	        value = pao->drvl;
	}
	pao->val = value;
	pao->pval = value;
	pao->udf = FALSE;

	/* now set value equal to desired output value */
        /* apply the output rate of change */
        if ( pao->oroc )
	{    /*must be defined and >0*/
	    float	diff;

            diff = value - pao->oval;
            if (diff < 0)
	    {
                if (pao->oroc < -diff) 
		    value = pao->oval - pao->oroc;
            }
	    else if (pao->oroc < diff) 
		value = pao->oval + pao->oroc;
        }
	pao->oval = value;

        /* convert */
        if (pao->linr == LINEAR)
        {
            double temp;

            /* y = mx + b , m = slope = eslo, b = offset = roff
		   calc'd in slope_and_offset ()
            */
            temp = (pao -> eslo * value) + pao -> roff;

            /* rval is defined as a long, so we need to round because of
               truncation
               if >= 0.0, round up using 0.5,
               else round up using 0.499999, because we want the following
                  -2.5 -> -1.5****1 has to round to -2
            */
            if (temp >= 0.0)
               pao -> rval = temp + 0.5;
            else
	       pao -> rval = temp - 0.499999;
        }
        else
            pao->rval = value;

	return(0);
}


/*------------------------------------------------------------------------*/

static void monitor (pao)
    struct aoRecord	*pao;
{
	unsigned short	monitor_mask;
	double		delta;
        short           stat,sevr,nsta,nsev;
	int		set_ack = 1;

        /* get previous stat and sevr  and new stat and sevr*/

        stat = pao->stat;
        sevr = pao->sevr;
        nsta = pao->nsta;
        nsev = pao->nsev;

        /*set current stat and sevr*/

        pao->stat = nsta;
        pao->sevr = nsev;
        pao->nsta = 0;
        pao->nsev = 0;

	/* if we are coming out of VALID_ALARM and our new stat/sevr are
	   the same as when we went into VALID_ALARM, do not set ack
	*/
	if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	     (nsta == pao -> psta) && (nsev == pao -> psev) )
	{
	    set_ack = 0;
	}
	/* if we are going into VALID_ALARM, and out previous stat was not
	   VALID_ALARM, remember previous non VALID_ALARM stat/sevr
	   and we do not want to set ACK,
	*/
	if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
	{
	    set_ack = 0;
	    pao -> psta = stat;
	    pao -> psev = sevr;
	    db_post_events ((struct dbCommon *)pao, (caddr_t)&pao -> psta, DBE_VALUE);
	    db_post_events ((struct dbCommon *)pao, (caddr_t)&pao -> psev, DBE_VALUE);
	}

	monitor_mask = 0;

	/* alarm condition changed this scan */

        if ((stat != nsta) || (sevr != nsev)) 
	{
            /* post events for alarm condition change*/

            monitor_mask = DBE_ALARM;

	    /* when alarm status chg's, get current time and
	       save VAL in lalm (last value when alarm chg'd)
	    */
	    tsLocalTime (&pao -> latm);
	    pao -> lalm = pao -> val;

            /* post stat and nsev fields */

            db_post_events ((struct dbCommon *)pao, (caddr_t)&pao->stat, DBE_VALUE);
            db_post_events ((struct dbCommon *)pao, (caddr_t)&pao->sevr, DBE_VALUE);
            if ((pao->sevr != NO_ALARM) && (set_ack) &&
		(pao->sevr != EVENT_ALARM) ) 
	    {
                if (pao->ack == 0)
		{
                    pao->ack = 1;           /* new unacknowledged alarm */
                    db_post_events ((struct dbCommon *)pao, (caddr_t)&pao->ack, DBE_VALUE);
                }
            }
        }
        /* check for value change */

        delta = pao->mlst - pao->val;
        if (delta < 0.0) 
	    delta = -delta;
	
        if (delta > pao->mdel) 
	{
            /* post events for value change */

            monitor_mask |= DBE_VALUE;

            /* update last value monitored */

            pao->mlst = pao->val;
        }

	/* only archive if trending is enabled */

	if (pao -> tenb)
	{
            /* check for archive change */

            delta = pao->alst - pao->val;
            if (delta < 0.0) 
	    	delta = -delta;
       
	    /* if exceed archive deadband or there has been a change in alarm status */
            if ( (delta > pao->adel) || (monitor_mask & DBE_ALARM) )
	    {
            	/* post events on value field for archive change */

            	monitor_mask |= DBE_LOG;

            	/* update last archive value monitored */

            	pao->alst = pao->val;
	    }
        }

        /* send out monitors connected to the value field */

        if (monitor_mask)
            db_post_events ((struct dbCommon *)pao, (caddr_t)&pao->val, monitor_mask);
	
	if (pao->oval != pao->val) 
	    monitor_mask |= DBE_VALUE;
	
	if (monitor_mask) 
	{
	    db_post_events ((struct dbCommon *)pao, (caddr_t)&pao->oval, DBE_VALUE);
	    if (pao->oraw != pao->rval) 
	    {
                db_post_events ((struct dbCommon *)pao, (caddr_t)&pao->rval, DBE_VALUE);
		pao->oraw = pao->rval;
	    }
	    if (pao->orbv != pao->rbv) 
	    {
                db_post_events ((struct dbCommon *)pao, (caddr_t)&pao->rbv, DBE_VALUE);
		pao->orbv = pao->rbv;
	    }
	}
	return;
}


/*------------------------------------------------------------------------*/

static void slope_and_offset (pao)
   struct aoRecord	*pao;
{
      /* slope = instrument range high - instrument range low /
		 eng units full - eng units low
         first check for divide by 0
      */

   if ((pao -> eguf - pao -> egul) == 0)
      pao -> eslo = 0.0;
   else
      pao -> eslo = (pao -> irhi - pao -> irlo) /
                    (pao -> eguf - pao -> egul);

      /* offset is eng units low - (slope * instrument range low */

   pao -> roff = pao -> irlo - (pao -> eslo * pao -> egul);

   return;
}


/*------------------------------------------------------------------------*/

static void test_ogap (pao, write_output)
    struct aoRecord	*pao;
    int			*write_output;
{
    long	status;
    long	options = 0;
    long	nRequests = 1;
    float	temp;
    double	lovl;

    if (pao -> ogap == 0)
	return;

    /* if my output is a reference, then get its value, and set lovl
       (last oval output) to the refs value. this protects us from the
       case where the output may be a field that has been changed
       from a graphic
    */
    if ( (pao -> dtyp <= 1) && (pao -> out.type == DB_LINK) )
    {
	status = dbGetLink (&pao -> out.value.db_link, 
			    (struct dbCommon *)pao, 
			    DBR_DOUBLE, (caddr_t)&lovl, 
			    &options, &nRequests);
	if (status)
	{
	    *write_output = 0;
	    pao -> nsev = VALID_ALARM;
	    pao -> nsta = LINK_ALARM;
	    return;
	}
	pao -> lovl = lovl;
    }
    temp = pao -> lovl - pao -> oval;
    if (temp < 0.0)
	temp = -temp;
    if (temp <= pao -> ogap)
	*write_output = 0;
    return;
}


/*------------------------------------------------------------------------*/

static void	aoReadback (struct aoRecord *pao)
{
    struct aodset	*pdset;
    double		val = pao -> val;
    long		rval = pao -> rval;
    long		status = 0;

    pdset = (struct aodset *)pao -> dset;

    /* no use if we don't have the ability to readback the output */
    if (pdset -> readback_ao == NULL)
	return;

    status = (*pdset -> readback_ao)(pao);
    if ( (status == 0) || (status == 3) )
    {
	/* no chg, nothing to do */
	if ( (val == pao -> val) && (rval == pao -> rval) )
	    return;
	else
	{
	    unsigned short	stat = pao -> stat;
	    unsigned short	sevr = pao -> sevr;
	    double		lalm = pao -> lalm;
	    TS_STAMP		latm = pao -> latm;
	    double		mlst = pao -> mlst;
	    short		hhll = pao -> hhll;
	    short		hilo = pao -> hilo;
	    short		rtct = pao -> rtct;
	    unsigned long	larf = pao -> larf;
	    char		redbuf [128];

	    /* update last process time */
	    tsLocalTime (&pao -> lptm);
	    if (status == 0)
	    {
		if (pao -> linr == LINEAR)
		{
		    if (pao -> eslo == 0)
			pao -> val = pao -> roff;
		    else
			pao -> val = (pao -> rval - pao -> roff) / pao -> eslo;
		}
		else
		{
		    if (pao -> linr == NOCONV)
			pao -> val = pao -> rval;
		}
	    }
	    alarm (pao);
	    monitor (pao);

	    /* do redundancy */
	    redQInit ((struct dbCommon *)pao, redbuf, sizeof (redbuf), type);

	    if (stat != pao->stat) 
		redQField (redbuf, mylist [STAT]);
	
	    if (val != pao->val)
		redQField (redbuf, mylist [VAL]);
	
	    if (sevr != pao->sevr)
		redQField (redbuf, mylist [SEVR]);
	
	    if (lalm != pao->lalm)
		redQField (redbuf, mylist [LALM]);
	
	    if (TsCmpStampsNE (&latm, &pao -> latm) )
		redQField (redbuf, mylist [LATM]);
	
	    if (mlst != pao->mlst)
		redQField (redbuf, mylist [MLST]);
	
	    if (hhll != pao -> hhll)
		redQField (redbuf, mylist [HHLL]);

	    if (hilo != pao -> hilo)
		redQField (redbuf, mylist [HILO]);

	    if (rtct != pao -> rtct)
		redQField (redbuf, mylist [RTCT]);

	    if (larf != pao -> larf)
		redQField (redbuf, mylist [LARF]);

	    redQSend (redbuf);
	}
    }

    return;
}

