/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* recPat.c */
  
/* recPat.c - Record Support Routines for records */
/*
 *      Original Author: EMSTEEL
 *      Date:            26 APr 95
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	xxx	change
 * .02	09-11-95	ems	monitor(), archive when stat/sevr chg's also
 *				process(),chg monitor/archive ASTP, new fields
 *				LASP (last archive setpt) and LMSP (last monitored)
 *				setpt
 * .03	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 * .04	09-14-95	ems	added LASP and LMSP to redundancy
 * .05	10-27-95	ems	added init of dsets dstr for the remote/local and
 *				dstf for the feedback
 * .06	11-09-95	ems	init_record(), if the feedback is bad then do not init 
 *				init_record(), if the istp (inital setpoint is < 0, then
 *				 init astp from val (feedback) else init astp from istp
 *				init_record(), we are going to let the pat init if its 
 *				 setpoint is gotten from a pid block, since usually the 
 *				 pid is looking to the pat to init, and we get into a 
 *				 situation of neither will init
 *				process(), give the setpoint block time to init so we 
 *				 dont flood the system with many alarms, because of the 
 *				 chg in init_record where if the setpoint is a pid, we 
 *				 let the pat init even if the pid is not init'd yet
 * .07	02-07-96	ems	remove return from HIHI and LOLO alarms, so will 
 *				 always check all alarm sevrs
 * .08	04-12-96	ems	added code to support alarm delay
 * .09	05-06-96	ems	added code to support HAND, manual output control
 * .10	06-19-96	ems	added code for alarm disable ADIS
 * .11	07-01-96	ems	process(), if was slave (WLSV), if RVAL > than 0 and POSO
 *				is set, then write to the positive channel just to make sure
 *				we did not loose something in the SLAVE/MASTER transistion
 *				if RVAL is < 0 and NEGO is set wrtite to the negative channel
 * .12	07-01-96	ems	new field sstp, save setpoint, use it to save the setpoint, 
 *				astp, incase some chg's the setpoint, astp, from outside the 
 *				scan task
 * .13	07-09-96	ems	process(), if in HAND and HPOS is on, was not setting POSO, 
 *				which would cause POS output to stay on unless came out of 
 *				HAND if in gap or in hand/bad input/local and were 
 *				slave (wslv) cancel callbacks and turn off outputs
 *				if in LOCAL and in HAND do NOT allow writes to the outputs
 *				on trans to/from hand, always turn off outputs and move logic
 *				to top of process loop to do in both remote/local
 *				special(), on HAND, only do if slave, fix logic setting LHND and
 *				set WSLV
 * .14	07-17-96	ems	init_record(), init mode to TRACKING, add dont_init and SBAD to 
 *				mode test
 *				process(), add input bad for mode TRACKING test
 *
 */

#include	<vxWorks.h>
#include	<inetLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<strLib.h>
#include	<wdLib.h>
#include	<string.h>
#include	<memLib.h>
#include	<logLib.h>
#include	<taskLib.h>
#include	<tickLib.h>

#include	"alarm.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"special.h"
#include	"recSup.h"
#include	"rec/patRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"patredfldidx.h"
#include	"restable.h"
#include	"cvnTypes.h"
#include	"callbacks.h"
#include	"db_addr.h"
#include	"recGbl.h"
#include	"dbEvent.h"
#include	"okToUse.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/

#define report NULL
#define initialize NULL
static long init_record ();
static long process ();
static long special ();
static long get_value ();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long  get_units ();
static long get_precision ();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double ();
static long get_control_double ();
static long get_alarm_double ();

struct rset patRSET = {

	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
	};

/* NOTE: these dset defsf must agree with the dset defs in .../devCom/devSoft.c
	 and .../XXX/devXXX.c where XXX is the hardware name
*/

struct patbodset {		/* binary output dset, for the pos and neg outputs */
	long	     number;
	DEVSUPFUN    dev_report;
	DEVSUPFUN    init;
	DEVSUPFUN    init_record;    /* returns:(0,2)=>(success,success no convert */
	DEVSUPFUN    get_ioint_info;
	DEVSUPFUN    write_pat;       /* returns:(-1,0,1) */
	};

struct patbidset {		/* binary input dset, for the remote/local switch */
	long		number;
	DEVSUPFUN	dev_report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;	/*returns: (-1,0)=>(failure,success)*/
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_pat;	/*(0,1,2) */
	};

struct pataidset {		/* analogue input dset, for the feedback */
	long		number;
	DEVSUPFUN	dev_report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;	/*returns: (-1,0)=>(failure,success)*/
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_pat;	/*(0,1,2) => */
	DEVSUPFUN	special_linconv;
	};

/* The following defintions are required for redundancy */

static struct fldDes	**mylist = 0;

static char *namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"LALM",
		"INI2",
		"LATM",
		"MLST",
		"RLVL",
		"ASTP",
		"CT",
		"MODE",
		"MOFT",
		"MONT",
		"RVAL",
		"LMSP",
		"LASP",
		"HHLL",
		"HILO",
		"RTCT",
		"LARF",
		"SBCT",
		"HLSC",
		"HPOS",
		"HNEG",
		"HAND",
		"SSTP",
		0
		};

static unsigned short	type;	/* index of point type for a pat block */

/* defines for the calls to init_pat(), write_pat() and read_pat() */

#define	PAT_POS_CHANNEL		0
#define	PAT_NEG_CHANNEL		1
#define	PAT_REMOTE_LOCAL	2
#define	PAT_INPUT		3

/* defines for the callback flag CBFL */

#define	PAT_NO_CALLBACK		0
#define	PAT_START_CALLBACK	1
#define	PAT_DURATION_ON		2
#define	PAT_DURATION_OFF	3

/* defines for the new command flag NCMD */

#define	PAT_NO_CMD	0
#define	PAT_NEW_CMD	1

/* defines for RLVL, remote/local */

#define	PAT_REMOTE	0
#define	PAT_LOCAL	1

/* defines for the mode field */

#define PAT_NOT_TRACKING	0
#define	PAT_TRACKING		1

/* defines for alarm clamp */

#define	DO_HIGH_SETPT_CLAMP	1
#define	DO_LOW_SETPT_CLAMP	2

static void alarm (struct patRecord *, unsigned short);
static void clamp_setpoint (struct patRecord *, unsigned short *, int);
static void convert ();
static void monitor ();
static void patCallback ();
static void slope_and_offset ();

extern void callbackRequest();


/*------------------------------------------------------------------------*/

static long init_record (ppat)
    struct	patRecord	*ppat;
{
    struct patbodset	*ppatbodset;
    struct patbidset	*ppatbidset;
    struct pataidset	*ppataidset;
    long		status = 0;
    CALLBACK_PTR	pcallback;
    unsigned short	i;
    unsigned short	clamp_mask = 0;
    char		*pstr;
    char		redbuf [512];
    int			ini2_flag = 0;
    int			do_convert = 1;
    int			dont_init = 0;
    long		options = 0;
    long		nRequests = 1;

    if (!(mylist)) 
    {
	for (type = 1; type < 100; type++) 
	{
	    pstr = GET_PRECTYPE (type);
	    if (strncmp (pstr, "pat", 3) == 0) 
		break;
	}
	mylist = redFldAlloc (type, NUMPATFLDS);

	if (!mylist)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)ppat, "pat, init_record, mylist failed");
	    return (S_rec_Alloc);
	}

	for (i = 0; i < NUMPATFLDS; i++) 
	    redFldAdd (type, i, namelist [i]);
    }

    /* check out the dset for the outputs, which are bo's,
       this dset is set in db/iocInit.c, the dtyp field
       is used for the outputs
    */
    if (!(ppatbodset = (struct patbodset *) (ppat->dset))) 
    {
	recGblRecordError (S_dev_noDSET, (struct dbCommon *)ppat, "pat: init_record,dset");
	return (S_dev_noDSET);
    }

    /* must have write_pat functions defined */

    if ( (ppatbodset->number < 5) || (ppatbodset->write_pat == NULL) ) 
    {
	recGblRecordError (S_dev_missingSup, (struct dbCommon *)ppat, "pat: init_record,dset");
	return (S_dev_missingSup);
    }

    /* get the dset for the remote/local */
    ppat -> dstr = (struct dset *)GET_PDSET(devSup -> papDevSup[type], ppat -> dtyr);

    if (!(ppatbidset = (struct patbidset *) (ppat -> dstr))) 
    {
	recGblRecordError (S_dev_noDSET, (struct dbCommon *)ppat, "pat: init_record,dstr");
	return(S_dev_noDSET);
    }

    /* must have read_pat function defined */

    if ( (ppatbidset -> number < 5) || (ppatbidset -> read_pat == NULL) ) 
    {
	recGblRecordError (S_dev_missingSup, (struct dbCommon *)ppat, "pat: init_record,dstr");
	return (S_dev_missingSup);
    }

    /* get the dset for the feedback */
    ppat -> dstf = (struct dset *)GET_PDSET(devSup -> papDevSup[type], ppat -> dtyf);

    if (!(ppataidset = (struct pataidset *) (ppat -> dstf))) 
    {
	recGblRecordError(S_dev_noDSET, (struct dbCommon *)ppat, "pat: init_record,dstf");
	return (S_dev_noDSET);
    }

    /* must have read_pat function defined */

    if ( (ppataidset -> number < 6) || (ppataidset -> read_pat == NULL) ) 
    {
	recGblRecordError (S_dev_missingSup, (struct dbCommon *)ppat, "pat: init_record,dstf");
	return (S_dev_missingSup);
    }


    if (!(ppat -> dpvt)) 
    {
	pcallback = (CALLBACK_PTR) (calloc (1, sizeof (CALLBACK)));
	if (!pcallback)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)ppat, "pat, init_record, pcallback failed");
	    return (S_rec_Alloc);
	}
	ppat->dpvt = (caddr_t) pcallback;
	pcallback->callback = patCallback;
	pcallback->priority = priorityMedium;
	if (dbNameToAddr (ppat->name, &(pcallback->dbAddr))) 
	{
	    logMsg("dbNameToAddr failed in init_record for recPat\n", 0,0,0,0,0,0);
	    taskSuspend (taskIdSelf());
	}
    }

    ppat -> init = TRUE;
    ppat -> fpps = 1;
    ppat -> ncmd = PAT_NO_CMD;
    ppat -> cbfl = PAT_NO_CALLBACK;
    ppat -> poso = 0;			/* positive channel on */
    ppat -> nego = 0;			/* negative channel on */
    ppat -> err  = 0;
    ppat -> moft = 0;
    ppat -> mont = 0;
    ppat -> mocb = 0;
    ppat -> svft = 0;
    ppat -> svnt = 0;
    ppat -> mode = PAT_TRACKING;
    ppat -> wslv = 0;
    ppat -> ct = tickGet ();
    ppat -> nsta = NO_ALARM;
    ppat -> nsev = NO_ALARM;
    ppat -> msev = NO_ALARM;
    ppat -> hand = 0;
    ppat -> lhnd = 0;
    ppat -> hpos = 0;
    ppat -> hneg = 0;

    /* init alarm delay variables */
    ppat -> larf = 0;
    ppat -> hhll = 0;
    ppat -> hilo = 0;
    ppat -> dvct = 0;
    ppat -> rtct = 0;
    ppat -> sbct = 0;
    ppat -> hlsc = 0;

    /* init last process time and last alarm time */

    if (ppat -> lptm.secPastEpoch == 0)
	tsLocalTime (&ppat -> lptm);
    if (ppat -> latm.secPastEpoch == 0)
	tsLocalTime (&ppat -> latm);

    /* check if the feedback is a CONSTANT or a LINK */
    if (ppat -> dtyf <= 1)
    {
	switch (ppat -> inp.type)
	{
	    case (CONSTANT):
	    case (PV_LINK):
	    case (DB_LINK):
	    case (CA_LINK):
 		do_convert = 0;      
		break;
	}
    }

    slope_and_offset (ppat);

    /* if no feedback init, do not init block */
    if (ppataidset -> init_record)
    {
	/* check the input, and readback VAL */
	status = (*ppataidset -> init_record) (ppat, PAT_INPUT);
	
	if (!status)
	{
	    /* convert VAL from input if hardware */
	    if (do_convert)
	    {
	    	convert (ppat);
	    }
	}
	else
	    dont_init = 1;

	/* check on REMOTE/LOCAL input */
	if (ppat -> rlin.type != CONSTANT)
	{
	    status = (*ppatbidset -> init_record)(ppat, PAT_REMOTE_LOCAL);
	    if (!status)
	    {
		/* check if R/L invert is in effect */
		if (ppat -> rliv)
		{
		    if (ppat -> rlrv == 0)
			ppat -> rlvl = PAT_LOCAL;
		    else
			ppat -> rlvl = PAT_REMOTE;
		}
		else
		{
		    if (ppat -> rlrv)
			ppat -> rlvl = PAT_LOCAL;
		    else
			ppat -> rlvl = PAT_REMOTE;
		}
	    }
	    else
		dont_init = 1;
	}
	/* the default is REMOTE */
	else
	    ppat -> rlvl = PAT_REMOTE;

	if ( (ppat -> rlvl == PAT_LOCAL) || 
	     (ppat -> omsl == SUPERVISORY) ||
	     (ppat -> stpl.type != DB_LINK) ||
	     (dont_init != 0) ||
	     (ppat -> nsta == SBAD_ALARM) )
	    ppat -> mode = PAT_TRACKING;
	else
	    ppat -> mode = PAT_NOT_TRACKING;

	/* init the 2 outputs */
	status = (*ppatbodset -> init_record)(ppat, PAT_POS_CHANNEL);
	if (status)
	    dont_init = 1;
	status = (*ppatbodset -> init_record)(ppat, PAT_NEG_CHANNEL);
	if (status)
	    dont_init = 1;
    }
    else
	dont_init = 1;

    if (!dont_init)
    {
	ppat -> astp = ppat -> val;
	if ( (ppat -> stpl.type == DB_LINK) && (ppat -> omsl == CLOSED_LOOP) )
	{
	    if ( (dbGetLink (&ppat -> stpl.value.db_link, 
			     (struct dbCommon *)ppat, 
			     DBR_FLOAT, (caddr_t)&ppat -> astp, 
			     &options, &nRequests) != NULL) ||
		 (!okToUse (&ppat -> stpl)) )
	    {
		/* we are going to let the pat init if its setpoint is 
		   gotten from a pid block, since usually the pid is looking
		   to the pat to init, and we get into a situation of neither
		   will init
		*/
		ppat -> astp = ppat -> val;
		pstr = GET_PRECTYPE (((struct dbAddr *)(ppat -> stpl.value.db_link.pdbAddr)) -> record_type);
		if (strcmp(pstr, "pid") != 0)
		    dont_init = 1;
	    }
	}
	clamp_setpoint (ppat, &clamp_mask, 1);
	ppat -> sstp = ppat -> astp;
    }

    /* get the initial tick count */

    ppat -> itck = tickGet();

    if (dont_init)
    {
	ppat -> nsta = NO_INIT_ALARM;
	ppat -> nsev = VALID_ALARM;
    }
    else
	ini2_flag = 1;

    if (resGetCaAnswer(ppat->radr, &i) == RES_OK) 
    {
	if (i) 
	{
	    /* check for maximize severity */

	    if (ppat -> msev > ppat -> nsev)
	    {
		ppat -> nsev = ppat -> msev;
		ppat -> nsta = LINK_ALARM;
	    }

	    monitor (ppat);

	    if (ini2_flag)
	    {
		ppat -> ini2 = 1;

		db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> val, DBE_VALUE);
		db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> astp, DBE_VALUE);
		db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> rlvl, DBE_VALUE);
		db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> mode, DBE_VALUE);

		redQInit ((struct dbCommon *)ppat, redbuf, sizeof (redbuf), type);
		redQField (redbuf, mylist [STAT]);
		redQField (redbuf, mylist [SEVR]);
		redQField (redbuf, mylist [VAL]);
		redQField (redbuf, mylist [LALM]);
		redQField (redbuf, mylist [INI2]);
		redQField (redbuf, mylist [RLVL]);
		redQField (redbuf, mylist [ASTP]);
		redQField (redbuf, mylist [MODE]);
		redQField (redbuf, mylist [RVAL]);
		redQField (redbuf, mylist [SSTP]);

		redQSend (redbuf);
	    }
	}
    }
    return(0);
}


/*------------------------------------------------------------------------*/

static void alarm (ppat, clamp_mask)
    struct patRecord	*ppat;
    unsigned short	clamp_mask;
{
    double		val = ppat->val;
    int			adis = 0;
    ALARM_LIMITS	alarm_limits;

    /* check for alarm inhibit */

    adis = AlarmDisableLink ((void *)ppat);
    if ( (ppat->aenb == ALARM_DISABLED) || (adis != 0) )
    {
	if (ppat -> nsta != NO_INIT_ALARM)
	{
	    ppat->nsta = INHIBIT_ALARM;
	    ppat->nsev = NO_ALARM;
	}
	return;
    }

    if(ppat->udf == TRUE) 
    {
	if (ppat->nsev < VALID_ALARM) 
	{
	    ppat->nsta = UDF_ALARM;
	    ppat->nsev = VALID_ALARM;
	}
	return;
    }

    /* check for SBAD alarm */

    if (ppat -> nsta == SBAD_ALARM)
    {
	if (ppat -> aldl > 0)
	{
	    if (ppat -> sbct < ALDL_MAX)
		ppat -> sbct++;
	    if (ppat -> sbct >= ppat -> aldl)
		ppat -> nsev = MAJOR_ALARM;
	    else
		ppat -> nsev = ppat -> nsta = NO_ALARM;
	}
	else
	{
	    ppat -> nsev = MAJOR_ALARM;
	}
    }
    else
    {
	if (ppat -> aldl > 0)
	{
	    ppat -> sbct = 0;
	}
    }

    /* test for HIGH setpoint clamp alarm */

    if (ppat -> chsv > NO_ALARM)
    {
	if (clamp_mask & DO_HIGH_SETPT_CLAMP)
	{
	    if (ppat -> aldl > 0)
	    {
		/* if we were counting low setpoint clamp, 
		   zero counter since same counter used for
		   HIGH and LOW clamp alarm
		*/
		if (ppat -> larf & LARF_SETPT_LOW)
		    ppat -> hlsc = 0;
		if (ppat -> hlsc < ALDL_MAX)
		    ppat -> hlsc++;
		ppat -> larf |= LARF_SETPT_HIGH;
		if ( (ppat -> hlsc >= ppat ->aldl) &&
		     (ppat -> nsev < ppat -> chsv) )
		{
		    ppat -> nsta = HIGH_SETPT_CLAMP;
		    ppat -> nsev = ppat -> chsv;
		}
	    }
	    else
	    {
		if (ppat -> nsev < ppat -> chsv)
		{
		    ppat -> nsta = HIGH_SETPT_CLAMP;
		    ppat -> nsev = ppat -> chsv;
		}
	    }
	}
	else
	{
	    if (ppat -> aldl > 0)
	    {
		ppat -> larf &= ~LARF_SETPT_HIGH;
		if (!(ppat -> larf & ((unsigned long)LARF_SETPT_LOW)))
		    ppat -> hlsc = 0;
	    }
	}
    }
    else
    {
	if (ppat -> aldl > 0)
	{
	    ppat -> larf &= ~LARF_SETPT_HIGH;
	    if (!(ppat -> larf & ((unsigned long)LARF_SETPT_LOW)))
		ppat -> hlsc = 0;
	}
    }

    /* check for LOW setpoint clamp alarm */

    if (ppat -> clsv > NO_ALARM)
    {
	if (clamp_mask & DO_LOW_SETPT_CLAMP)
	{
	    if (ppat -> aldl > 0)
	    {
		if (ppat -> larf & LARF_SETPT_HIGH)
		    ppat -> hlsc = 0;
		if (ppat -> hlsc < ALDL_MAX)
		    ppat -> hlsc++;
		ppat -> larf |= LARF_SETPT_LOW;
		if ( (ppat -> hlsc >= ppat ->aldl) &&
		     (ppat -> nsev < ppat -> clsv) )
		{
		    ppat -> nsta = LOW_SETPT_CLAMP;
		    ppat -> nsev = ppat -> clsv;
		}
	    }
	    else
	    {
		if (ppat -> nsev < ppat -> clsv)
		{
		    ppat -> nsta = LOW_SETPT_CLAMP;
		    ppat -> nsev = ppat -> clsv;
		}
	    }
	}
	else
	{
	    if (ppat -> aldl > 0)
	    {
		ppat -> larf &= ~LARF_SETPT_LOW;
		if (!(ppat -> larf & ((unsigned long)LARF_SETPT_HIGH)))
		    ppat -> hlsc = 0;
	    }
	}
    }
    else
    {
	if (ppat -> aldl > 0)
	{
	    ppat -> larf &= ~LARF_SETPT_LOW;
	    if (!(ppat -> larf & ((unsigned long)LARF_SETPT_HIGH)))
		ppat -> hlsc = 0;
	}
    }

    /* set up for limit alarms */

    alarm_limits.hihi = ppat -> hihi;
    alarm_limits.high = ppat -> high;
    alarm_limits.low  = ppat -> low;
    alarm_limits.lolo = ppat -> lolo;
    alarm_limits.devl = 0;
    alarm_limits.devs = 0;
    alarm_limits.hyst = ppat -> hyst;
    alarm_limits.hhsv = ppat -> hhsv;
    alarm_limits.hsv  = ppat -> hsv;
    alarm_limits.lsv  = ppat -> lsv;
    alarm_limits.llsv = ppat -> llsv;
    alarm_limits.dvsv = NO_ALARM;

    AlarmLimits (&alarm_limits, val, (void *)ppat);

    /* do a return to normal, check for alarm delay ALDL only
       on the listed alarms
    */

    if ( ((ppat -> nsta == NO_ALARM) && (ppat -> nsev == NO_ALARM)) &&
	 ((ppat -> stat == HIHI_ALARM) || 
	  (ppat -> stat == HIGH_ALARM) ||
	  (ppat -> stat == LOW_ALARM)  || 
	  (ppat -> stat == LOLO_ALARM) ||
	  (ppat -> stat == SBAD_ALARM) ||
	  (ppat -> stat == HIGH_SETPT_CLAMP) ||
	  (ppat -> stat == LOW_SETPT_CLAMP)) )
    {
	if (ppat -> aldl > 0)
	{
	    if (ppat -> rtct < ALDL_MAX)
		ppat -> rtct++;

	    /* delay return to normal so, continue last alarm */
	    if (ppat -> rtct < ppat -> aldl)
	    {
		ppat -> nsta = ppat -> stat;
		ppat -> nsev = ppat -> sevr;
	    }
	}
    }
    else
    {
	if (ppat -> aldl > 0)
	{
	    ppat -> rtct = 0;
	}
    }

    return;
}


/*------------------------------------------------------------------------*/

static void clamp_setpoint (ppat, clamp_mask, set_nsev)
    struct patRecord    *ppat;
    unsigned short	*clamp_mask;
    int			set_nsev;
{

    /* clamp the setpoint in case someone set it too high */
    if (ppat->astp >= ppat->clmh)
    {
	ppat->astp = ppat->clmh;
	if (ppat->nsev < ppat->chsv) 
	{
	    if (set_nsev)
	    {
		ppat->nsta = HIGH_SETPT_CLAMP;
		ppat->nsev = ppat->chsv;
	    }
	    else
		*clamp_mask |= DO_HIGH_SETPT_CLAMP;
	}
    }
    else if (ppat->astp <= ppat->clml)
    {
	ppat->astp = ppat->clml;
	if (ppat->nsev < ppat->clsv) 
	{
	    if (set_nsev)
	    {
		ppat->nsta = LOW_SETPT_CLAMP;
		ppat->nsev = ppat->clsv;
	    }
	    else
		*clamp_mask |= DO_LOW_SETPT_CLAMP;
	}
    }
    return;
}


/*------------------------------------------------------------------------*/

static void convert (ppat)
    struct	patRecord	*ppat;
{
    double	val;

    /* check value in instrument range */

    val = ppat -> irvl;
    if ((ppat -> irhi != 0) || (ppat -> irlo != 0)) 
    {
	if ((val > (ppat -> irhi + ppat -> idev)) || 
	    (val < (ppat -> irlo - ppat -> idev)) ) 
	{
	    if (ppat -> nsev < VALID_ALARM)
	    {
		ppat -> nsta = SBAD_ALARM;
		ppat -> nsev = VALID_ALARM;
	    }
	    return;
	}
    }
    val = (ppat -> eslo * val) + ppat -> roff;
    ppat -> val = val;
	    
    return;
}



/*------------------------------------------------------------------------*/

static long get_value (ppat,pvdes)
    struct patRecord		*ppat;
    struct valueDes		*pvdes;
{
    pvdes -> field_type = DBF_DOUBLE;
    pvdes -> no_elements=1;
    (double *)(pvdes -> pvalue) = &ppat -> val;
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_precision (paddr, precision)
    struct dbAddr *paddr;
    long	  *precision;
{
    struct patRecord	*ppat = (struct patRecord *)paddr -> precord;

    *precision = ppat -> prec;
    return (0);
}

/*------------------------------------------------------------------------*/

static long get_units(paddr,units)
    struct dbAddr *paddr;
    char	  *units;
{
    strcpy (units, "percent");
    return (0);
}

/*------------------------------------------------------------------------*/

static long get_alarm_double (paddr, pad)
    struct dbAddr *paddr;
    struct dbr_alDouble	*pad;
{
    struct patRecord	*ppat =(struct patRecord *) paddr -> precord;

    pad -> upper_alarm_limit = ppat -> hihi;
    pad -> upper_warning_limit = ppat -> high;
    pad -> lower_warning_limit = ppat -> low;
    pad -> lower_alarm_limit = ppat -> lolo;
    return (0);
}


/*------------------------------------------------------------------------*/

static long get_graphic_double (paddr, pgd)
    struct dbAddr *paddr;
    struct dbr_grDouble	*pgd;
{
    struct patRecord	*ppat=(struct patRecord *)paddr -> precord;


    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct patRecord *)0) -> rval) ):
	    pgd -> upper_disp_limit = ppat -> stim * vxTicksPerSecond;
	    pgd -> lower_disp_limit = -(ppat -> stim * vxTicksPerSecond);
	    break;

	default:
	    pgd -> upper_disp_limit = ppat -> hopr;
	    pgd -> lower_disp_limit = ppat -> lopr;
	    break;
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static long get_control_double (paddr, pcd)
    struct dbAddr *paddr;
    struct dbr_ctrlDouble *pcd;
{
    struct patRecord	*ppat = (struct patRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct patRecord *)0) -> hyst) ):
	{
	    float    hyst_span;

	    hyst_span = ppat -> hopr - ppat -> lopr;
	    if (hyst_span < 0.0)
	        hyst_span = -hyst_span;
	    pcd -> upper_ctrl_limit = hyst_span;
	    pcd -> lower_ctrl_limit = 0;
	    break;
	}

	case ( (int)&(((struct patRecord *)0) -> astp) ):
	    pcd -> upper_ctrl_limit = ppat -> clmh;
	    pcd -> lower_ctrl_limit = ppat -> clml;
	    break;

	case ( (int)&(((struct patRecord *)0) -> stim) ):
	    pcd -> upper_ctrl_limit = 1000;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	default:
	    pcd -> upper_ctrl_limit = ppat -> hopr;
	    pcd -> lower_ctrl_limit = ppat -> lopr;
	    break;
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static void monitor (ppat)
    struct patRecord	*ppat;
{
    unsigned short	monitor_mask;
    double		delta;
    short		stat,sevr,nsta,nsev;
    int			set_ack = 1;
	
    /* get previous stat and sevr and new stat and sevr*/

    stat = ppat->stat;
    sevr = ppat->sevr;
    nsta = ppat->nsta;
    nsev = ppat->nsev;

    /*set current stat and sevr*/

    ppat->stat = nsta;
    ppat->sevr = nsev;
    ppat->nsta = 0;
    ppat->nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == ppat -> psta) && (nsev == ppat -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and out previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	ppat -> psta = stat;
	ppat -> psev = sevr;
	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> psev, DBE_VALUE);
    }

    monitor_mask = 0;
	

    /* alarm condition changed this scan */

    if ((stat != nsta) || (sevr != nsev)) 
    {
	/* post events for alarm condition change*/

	monitor_mask = DBE_ALARM;

	/* when alarm status chg's, get current time and
	   save VAL in lalm (last value when alarm chg'd)
	*/
	tsLocalTime (&ppat -> latm);
	ppat -> lalm = ppat -> val;

	/* post stat and nsev fields */

	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat->stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat->sevr, DBE_VALUE);
	if ( (ppat->sevr != NO_ALARM) && (set_ack) &&
	     (ppat->sevr != EVENT_ALARM) )
	{
	    if (ppat->ack == 0) 
	    {
		ppat->ack = 1;           /* new unacknowledged alarm */
		db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat->ack, DBE_VALUE);
	    }
	}
    }

    /* check for value change */

    delta = ppat->mlst - ppat->val;
    if (delta < 0.0) 
	delta = -delta;
	
    if (delta > ppat->mdel) 
    {
	/* post events for value change */

	monitor_mask |= DBE_VALUE;

	/* update last value monitored */

	ppat->mlst = ppat->val;
    }

    /* check for archive change if tenb (trending) is enabled */

    if (ppat -> tenb)
    {
	delta = ppat->alst - ppat->val;
	if (delta < 0.0) 
	    delta = -delta;
       
	/* if exceed archive deadband or there has been a change in alarm status */
	if ( (delta > ppat->adel) || (monitor_mask & DBE_ALARM) )
	{
	    /* post events on value field for archive change */

	    monitor_mask |= DBE_LOG;

	    /* update last archive value monitored */

	    ppat->alst = ppat->val;
	}
    }

    /* send out monitors connected to the value field */

    if (monitor_mask)
	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat->val, monitor_mask);

    return;
}


/*------------------------------------------------------------------------*/

static long process (paddr)
    struct dbAddr	*paddr;
{
    struct patRecord	*ppat = (struct patRecord *) (paddr -> precord);
    struct patbodset	*ppatbodset = (struct patbodset *) (ppat -> dset);
    struct patbidset	*ppatbidset = (struct patbidset *) (ppat -> dstr);
    struct pataidset	*ppataidset = (struct pataidset *) (ppat -> dstf);
    CALLBACK_PTR	pcallback = (CALLBACK_PTR) (ppat -> dpvt);
    int			wait_time = 0;
    long		status = 0;
    float		value;
    float		delta_time;
    unsigned long	ticks_now;
    float		roc_limit;
    unsigned short	stat = ppat -> stat;
    unsigned short	sevr = ppat -> sevr;
    double		val  = ppat -> val;
    double		lalm = ppat -> lalm;
    TS_STAMP		latm = ppat -> latm;
    double		mlst = ppat -> mlst;
    unsigned short	rlvl = ppat -> rlvl;
    float		err  = ppat -> err;
    unsigned short	mode = ppat -> mode;
    unsigned long	ct   = ppat -> ct;
    unsigned long	mont = ppat -> mont;
    unsigned long	moft = ppat -> moft;
    long		rval = ppat -> rval;
    float		lmsp = ppat -> lmsp;
    short		hhll = ppat -> hhll;
    short		hilo = ppat -> hilo;
    short		rtct = ppat -> rtct;
    short		sbct = ppat -> sbct;
    short		hlsc = ppat -> hlsc;
    unsigned long	larf = ppat -> larf;
    unsigned short	hand = ppat -> hand;
    unsigned short	hpos = ppat -> hpos;
    unsigned short	hneg = ppat -> hneg;
    unsigned short	clamp_mask;
    char		redbuf[512];
    long		options = 0;
    long		nRequests = 1;
    int			inp_bad = 0;
    int			prfl = 0;
    int			nrfl = 0;
    unsigned short	monitor_mask;
    double		delta;

    /* if not init'd do nothing */

    if ( (!ppat -> init) || (!ppat -> ini2) )
	return (status);

    if( (ppataidset == NULL) || (ppataidset -> read_pat == NULL) ) 
    {
	ppat->pact = TRUE;
	recGblRecordError(S_dev_missingSup, (struct dbCommon *)ppat, "process, feebback missing");
	return (S_dev_missingSup);
    }

    /* get time stamp for last time record was processed */

    tsLocalTime (&ppat -> lptm);

    ticks_now = tickGet();
    delta_time = (ticks_now - ppat -> ct) / (float)vxTicksPerSecond;
    ppat -> ct = ticks_now;

    ppat-> msev = NO_ALARM;

    clamp_mask = 0;

    if (ppat->tagd) 
    {
	ppat->nsta = TAGGED_ALARM;
	ppat->nsev = VALID_ALARM;
    } 
    else if (ppat->oscn == OFF_SCAN) 
    {
	ppat->nsta = SCAN_ALARM;
	ppat->nsev = VALID_ALARM;
    } 
    else 
    {
	ppat->nsta = NO_ALARM;
	ppat->nsev = NO_ALARM;

	ppat -> pact = TRUE;

	/* update number of process times */

	if (process_stats)
	    process_cnts[type]++;	

	/* if we are here, we have a problem, since the callback
	   should have set cbfl to something other than
	   PAT_START_CALLBACK, so we clean up and start
	   again
	*/
	if (ppat -> cbfl == PAT_START_CALLBACK)
	{
	    wdCancel (pcallback -> wd_id);
	    ppat -> cbfl = PAT_NO_CALLBACK;
	    ppat -> ncmd = PAT_NO_CMD;
	    /*
	    printf ("Pat block %s, process cbfl = 1\n",ppat->name);
	    */
	}

	/* read the feedback, where we actually are, if DB_LINK, read into VAL */
	status = (ppataidset -> read_pat)(ppat, PAT_INPUT);
	/* driver on hold, may be failing over, so do nothing */
	if (status == -2)
	{
	    ppat -> pact = FALSE;
	    return (0);
	}
	else if (status != 0)
	    inp_bad = 1;
	else
	{
	    if (ppat -> inp.type == DB_LINK)
	    {
		if (!okToUse (&ppat -> inp))
		{
		    inp_bad = 1;
		    if (ppat -> nsev < MAJOR_ALARM)
		    {
			ppat -> nsev = MAJOR_ALARM;
			ppat -> nsta = PID_ALM_NOMV;
		    }
		}
	    }
	    else
	    {
		/* convert raw hardware feedback into VAL */
		convert (ppat);
		if (ppat -> nsta == SBAD_ALARM)
		    inp_bad = 1;
	    }
	}

	if (ppat -> rlin.type != CONSTANT)
	{
	    /* read the REMOTE/LOCAL value */
	    status = (*ppatbidset -> read_pat)(ppat, PAT_REMOTE_LOCAL);
	    if (!status)
	    {
		/* check if R/L invert is in effect */
		if (ppat -> rliv)
		{
		    if (ppat -> rlrv == 0)
			ppat -> rlvl = PAT_LOCAL;
		    else
			ppat -> rlvl = PAT_REMOTE;
		}
		else
		{
		    if (ppat -> rlrv)
			ppat -> rlvl = PAT_LOCAL;
		    else
			ppat -> rlvl = PAT_REMOTE;
		}
	    }
	}
	/* the default is REMOTE */
	else
	    ppat -> rlvl = PAT_REMOTE;
	
	/* if we are in LOCAL or hand  and our input is good, then we want
	   to track our setpt, even if it gets overwritten below
	*/
	if ( ((ppat -> rlvl == PAT_LOCAL) || (ppat -> hand)) && (!inp_bad) )
	    ppat -> astp = ppat -> val;

	if ( (ppat -> rlvl == PAT_LOCAL) || 
	     (ppat -> omsl == SUPERVISORY) ||
	     (ppat -> stpl.type != DB_LINK) ||
	     (ppat -> hand != 0) ||
	     (inp_bad) )
	    ppat -> mode = PAT_TRACKING;
	else
	    ppat -> mode = PAT_NOT_TRACKING;

	/* transition to/from hand, make sure outputs are off */
	if (ppat -> lhnd != ppat -> hand)
	{
	    /* cancel the callback, turn off outputs */
	    if (ppat -> cbfl != PAT_NO_CALLBACK)
	    {
		wdCancel (pcallback -> wd_id);
		ppat -> cbfl = PAT_NO_CALLBACK;
	    }
	    if (ppat -> poiv)
		ppat -> rbv = ppat -> pmsk;
	    else
		ppat -> rbv = 0;
	    (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
	    if (ppat -> noiv)
		ppat -> rbv = ppat -> nmsk;
	    else
		ppat -> rbv = 0;
	    (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
	    ppat -> poso = 0;
	    ppat -> nego = 0;
	    ppat -> wslv = 0;
	}

	/* read the SETPOINT, where we want to be */
	if ( (ppat -> stpl.type == DB_LINK) && (ppat -> omsl == CLOSED_LOOP) )
	{
	    value = ppat -> astp;
	    if ( (dbGetLink (&ppat -> stpl.value.db_link, 
			     (struct dbCommon *)ppat, 
			     DBR_FLOAT, (caddr_t)&ppat -> astp, 
			     &options, &nRequests) != NULL) ||
		 (!okToUse (&ppat -> stpl)) )
	    {
		ppat -> astp = value;
		/* give the setpoint block time to init so we dont flood
		   the system with many alarms, because of the chg in init_record
		   where if the setpoint is a pid, we let the pat init even
		   if the pid is not init'd yet
		*/
		if ( ppat -> ct > (ppat -> itck + 1500))
		{
		    if (ppat -> nsev < MAJOR_ALARM)
		    {
			ppat -> nsev = MAJOR_ALARM;
			ppat -> nsta = SETPT_ERROR;
		    }
		}
	    }
	}

	clamp_setpoint (ppat, &clamp_mask, 0);

	/* if roc is less than 1 second durations */
	if ( ((ppat -> prcl / 100) * ppat -> stim) >= 1.0)
	    prfl = 0;
	else
	    prfl = 1;
	if ( ((ppat -> nrcl / 100) * ppat -> stim) >= 1.0)
	    nrfl = 0;
	else
	    nrfl = 1;

	if ( (ppat -> rlvl == PAT_REMOTE) && (!inp_bad) && (!ppat -> hand) )
	{
	    /* how much to move in percent */
	    ppat -> err = ppat -> astp - ppat -> val;
	    value = ppat -> err;
	    if (value < 0)
		value = -value;
	    /* output deadband */
	    if (value > ppat -> gap)
	    {
		/* select the maximum change percent per second */
		if (ppat -> err >= 0)
		    roc_limit = ppat -> prcl;
		else
		    roc_limit = ppat -> nrcl;

		/* rate of change in percent per our scan time */
		roc_limit = roc_limit * delta_time;

		/* calc the percent +/- we need to move */
		if ( (value >= roc_limit) && (roc_limit != 0) )
		{
		    if (ppat -> err >= 0)
		    {
			value = roc_limit;
			if (prfl)
			{
			    ppat -> mont = (int)( ((ppat -> prcl / 100.0) * 
						ppat -> stim) * vxTicksPerSecond);
			    if (ppat -> mont >= 60)
			    {
				ppat -> mont = 0;
				ppat -> moft = 0;
			    }
			    else
				ppat -> moft = 
					(int) (vxTicksPerSecond - ppat -> mont);
			}
			else
			{
			    ppat -> mont = 0;
			    ppat -> moft = 0;
			}
		    }
		    else
		    {
			value = -roc_limit;
			if (nrfl)
			{
			    ppat -> mont = (int)( ((ppat -> nrcl / 100.0) * 
						ppat -> stim) * vxTicksPerSecond);
			    if (ppat -> mont >= 60)
			    {
				ppat -> mont = 0;
				ppat -> moft = 0;
			    }
			    else
				ppat -> moft = (int) (vxTicksPerSecond - ppat -> mont);
			}
			else
			{
			    ppat -> mont = 0;
			    ppat -> moft = 0;
			}
		    }
		}
		else
		{
		    if ( (roc_limit != 0) && (prfl || nrfl) )
		    {
			if (ppat -> err >= 0)
			{
			    /* roc < 1 second and error >= roc, we have to limit
			       on time
			    */
			    if (prfl && (value >= ppat -> prcl) )
			    {
		    		value = ppat -> err;
			    	ppat -> mont = (int)( ((ppat -> prcl / 100.0) * 
						ppat -> stim) * vxTicksPerSecond);
			    	if (ppat -> mont >= 60)
			    	{
				    ppat -> mont = 0;
				    ppat -> moft = 0;
			    	}
			    	else
				    ppat -> moft = (int) (vxTicksPerSecond - 
									ppat -> mont);
			    }
			    else
			    {
		    		value = ppat -> err;
				ppat -> mont = 0;
				ppat -> moft = 0;
			    }
			}
			else
			{
			    /* roc < 1 second and error >= roc, we have to limit
			       on time
			    */
			    if (nrfl && (value >= ppat -> nrcl) )
			    {
		    		value = ppat -> err;
			    	ppat -> mont = (int)( ((ppat -> nrcl / 100.0) * 
						ppat -> stim) * vxTicksPerSecond);
			    	if (ppat -> mont >= 60)
			    	{
				    ppat -> mont = 0;
				    ppat -> moft = 0;
			    	}
			    	else
				    ppat -> moft = (int) (vxTicksPerSecond - 
									ppat -> mont);
			    }
			    else
			    {
		    		value = ppat -> err;
				ppat -> mont = 0;
				ppat -> moft = 0;
			    }
			}
		    }
		    else
		    {
		    	value = ppat -> err;
			ppat -> mont = 0;
			ppat -> moft = 0;
		    }
		}

		/* amount of time to turn on POS/NEG output */
		ppat -> rval = (int)(((value / 100.0) * ppat -> stim) 
								* vxTicksPerSecond);
		/* if we were slave but now are master, do some cleanup,
		   we can not be sure that the outputs were turned on/off
		   at the time of a failover
		*/
		if (ppat -> wslv)
		{
		    ppat -> wslv = 0;
		    if (ppat -> rval == 0)
		    {
			if (ppat -> poiv)
			    ppat -> rbv = ppat -> pmsk;
			else
			    ppat -> rbv = 0;
			(*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
			if (ppat -> noiv)
			    ppat -> rbv = ppat -> nmsk;
			else
			    ppat -> rbv = 0;
			(*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
			ppat -> poso = 0;
			ppat -> nego = 0;
		    }
		    else if (ppat -> rval > 0)
		    {
			/* make sure the neg channel is off */
			if (ppat -> noiv)
			    ppat -> rbv = ppat -> nmsk;
			else
			    ppat -> rbv = 0;
			(*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
			ppat -> nego = 0;
			/* and if we think the pos channel is on
			   turn the pos channel on to be sure
			*/
			if (ppat -> poso)
			{
			    if (ppat -> poiv)
				ppat -> rbv = 0;
			    else
				ppat -> rbv = ppat -> pmsk;
			    (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
			}
		    }
		    /* rval < 0 */
		    else
		    {
			/* make sure pos channel is off */
			if (ppat -> poiv)
			    ppat -> rbv = ppat -> pmsk;
			else
			    ppat -> rbv = 0;
			(*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
			ppat -> poso = 0;
			/* and if we think the neg channel is on
			   turn the neg channel on to be sure
			*/
			if (ppat -> nego)
			{
			    if (ppat -> noiv)
				ppat -> rbv = 0;
			    else
				ppat -> rbv = ppat -> nmsk;
			    (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
			}
		    }
		} /* end if wslv */
	    }
	    /* inside of gap */
	    else
	    {
		ppat -> rval = 0;
		ppat -> mont = 0;
		ppat -> moft = 0;
		if (ppat -> wslv)
		{
		    ppat -> wslv = 0;
		    if (ppat -> cbfl != PAT_NO_CALLBACK)
		    {
			wdCancel (pcallback -> wd_id);
			ppat -> cbfl = PAT_NO_CALLBACK;
		    }
		    if (ppat -> poiv)
			ppat -> rbv = ppat -> pmsk;
		    else
			ppat -> rbv = 0;
		    (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
		    if (ppat -> noiv)
			ppat -> rbv = ppat -> nmsk;
		    else
			ppat -> rbv = 0;
		    (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
		    ppat -> poso = 0;
		    ppat -> nego = 0;
		}
	    }
	}
	/* are local or bad input or hand */
	else
	{
	    ppat -> rval = 0;
	    ppat -> mont = 0;
	    ppat -> moft = 0;
	    ppat -> ncmd = 0;
	    /* cancel the callback, turn off outputs */
	    if (ppat -> cbfl != PAT_NO_CALLBACK)
	    {
		wdCancel (pcallback -> wd_id);
		ppat -> cbfl = PAT_NO_CALLBACK;
		if (ppat -> poiv)
		    ppat -> rbv = ppat -> pmsk;
		else
		    ppat -> rbv = 0;
		(*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
		if (ppat -> noiv)
		    ppat -> rbv = ppat -> nmsk;
		else
		    ppat -> rbv = 0;
		(*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
		ppat -> poso = 0;
		ppat -> nego = 0;
		ppat -> wslv = 0;
	    }
	    if (ppat -> wslv)
	    {
		ppat -> wslv = 0;
		if (ppat -> poiv)
		    ppat -> rbv = ppat -> pmsk;
		else
		    ppat -> rbv = 0;
		(*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
		if (ppat -> noiv)
		    ppat -> rbv = ppat -> nmsk;
		else
		    ppat -> rbv = 0;
		(*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
		ppat -> poso = 0;
		ppat -> nego = 0;
	    }

	    if (ppat -> hand)
	    {
		/* both off or both on, turn off outputs,
		   both can not be on
		*/
		if ( ((!ppat -> hpos) && (!ppat -> hneg)) ||
		     ((ppat -> hpos) && (ppat -> hneg))   || 
		     (ppat -> rlvl == PAT_LOCAL) )
		{
		    if (ppat -> poso)
		    {
			if (ppat -> poiv)
			    ppat -> rbv = ppat -> pmsk;
			else
			    ppat -> rbv = 0;
			(*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
			ppat -> poso = 0;
		    }
		    if (ppat -> nego)
		    {
			if (ppat -> noiv)
			    ppat -> rbv = ppat -> nmsk;
			else
			    ppat -> rbv = 0;
			(*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
			ppat -> nego = 0;
		    }
		}
		else
		{
		    if ( (ppat -> hneg) && (ppat -> rlvl == PAT_REMOTE) )
		    {
			if (ppat -> poso)
			{
			    if (ppat -> poiv)
				ppat -> rbv = ppat -> pmsk;
			    else
				ppat -> rbv = 0;
			    (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
			    ppat -> poso = 0;
			}
			if (ppat -> nego == 0)
			{
			    if (ppat -> noiv)
				ppat -> rbv = 0;
			    else
				ppat -> rbv = ppat -> nmsk;
			    (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
			    ppat -> nego = 1;
			}
		    }
		    else if ( (ppat -> hpos) && (ppat -> rlvl == PAT_REMOTE) )
		    {
			if (ppat -> nego)
			{
			    if (ppat -> noiv)
				ppat -> rbv = ppat -> nmsk;
			    else
				ppat -> rbv = 0;
			    (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
			    ppat -> nego = 0;
			}
			if (ppat -> poso == 0)
			{
			    if (ppat -> poiv)
				ppat -> rbv = 0;
			    else
				ppat -> rbv = ppat -> pmsk;
			    (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
			    ppat -> poso = 1;
			}
		    }
		}
	    }
	}  /* end local, bad input or hand */

	ppat -> lhnd = ppat -> hand;

	if (ppat -> rval || ppat -> cbfl)
	{
	    ppat -> ncmd = PAT_NEW_CMD;
	    if ( ( ( ppat -> cbfl == PAT_DURATION_ON) || 
		   (ppat -> cbfl == PAT_DURATION_OFF)) && (!ppat -> mocb) )
	    {
		wdCancel (pcallback -> wd_id);
		ppat -> cbfl = PAT_NO_CALLBACK;
	    }
	}

	/* if we have someting to do */
	if (ppat -> ncmd)
	{
	    /* if a callback is not already in effect, get on going */
	    if (ppat -> cbfl == PAT_NO_CALLBACK)
	    {
		wait_time = 1;
		ppat -> cbfl = PAT_START_CALLBACK;
		pcallback = (CALLBACK_PTR) (ppat -> dpvt);
		if (pcallback -> wd_id == NULL)
		{
		    pcallback -> wd_id = wdCreate ();
		}
		wdStart (pcallback -> wd_id, wait_time, (FUNCPTR)callbackRequest, (int)pcallback);
	    }
	}
	
	/* check for alarms */

	alarm (ppat, clamp_mask);

	/* check for maximize severity */

	if (ppat -> msev > ppat -> nsev)
	{
	    ppat -> nsev = ppat -> msev;
	    ppat -> nsta = LINK_ALARM;
	}

    } /* end on scan */

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (ppat -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&ppat -> latm);
	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> val, DBE_VALUE | DBE_ALARM);
    }

    /* if it is the 1st process pass 
       init last value when alarm chg'd (lalm)
    */

    if (ppat -> fpps)
    {
	ppat -> fpps = 0;
	ppat -> lalm = ppat -> val;
    }

    /* check event list */

    monitor (ppat);

    /* update redundant fields that have changed */

    redQInit ((struct dbCommon *)ppat, redbuf, sizeof (redbuf), type);

    if (stat != ppat->stat) 
	redQField (redbuf, mylist [STAT]);
	
    if (sevr != ppat->sevr) 
	redQField (redbuf, mylist [SEVR]);
	
    if (val != ppat->val) 
	redQField (redbuf, mylist [VAL]);
	
    if (lalm != ppat->lalm) 
	redQField (redbuf, mylist [LALM]);
	
    if (TsCmpStampsNE (&latm, &ppat -> latm) )
	redQField (redbuf, mylist [LATM]);
	
    if (mlst != ppat -> mlst)
	redQField (redbuf, mylist [MLST]);

    if (rlvl != ppat -> rlvl)
    {
	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> rlvl, DBE_VALUE);
	redQField (redbuf, mylist [RLVL]);
    }

    if (ppat -> sstp != ppat -> astp)
    {
	redQField (redbuf, mylist [ASTP]);
	ppat -> sstp = ppat -> astp;
	redQField (redbuf, mylist [SSTP]);
    }
    monitor_mask = 0;
    if ( (ppat -> stpl.type == DB_LINK) && (ppat -> omsl == CLOSED_LOOP) )
    {
	delta = ppat -> lmsp - ppat -> astp;
	if (delta < 0.0)
	    delta = -delta;
	if (delta > ppat -> mdel)
	{
	    monitor_mask = DBE_VALUE;
	    ppat -> lmsp = ppat -> astp;	/* last monitored setpt */
	}
	if (ppat -> tenb)
	{
	    delta = ppat -> lasp - ppat -> astp;
	    if (delta < 0.0)
		delta = -delta;
	    if (delta > ppat -> adel)
	    {
		monitor_mask |= DBE_LOG;
		ppat -> lasp = ppat -> astp;	/* last archived setpt */
	    }
	}
    }
    else
    {
	if (ppat -> lmsp != ppat -> astp)
	{
	    monitor_mask = DBE_VALUE;
	    ppat -> lmsp = ppat -> astp;
	}
	if (ppat -> tenb)
	{
	    if (ppat -> lasp != ppat -> astp)
	    {
		monitor_mask |= DBE_LOG;
		ppat -> lasp = ppat -> astp;
	    }
	}
    }
    if (monitor_mask)
	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> astp, monitor_mask);

    if (ct != ppat -> ct)
	redQField (redbuf, mylist [CT]);

    if (mode != ppat -> mode)
    {
	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> mode, DBE_VALUE);
	redQField (redbuf, mylist [MODE]);
    }

    if (err != ppat -> err)
	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> err, DBE_VALUE);

    if (moft != ppat -> moft)
	redQField (redbuf, mylist [MOFT]);

    if (mont != ppat -> mont)
	redQField (redbuf, mylist [MONT]);

    if (lmsp != ppat -> lmsp)
	redQField (redbuf, mylist [LMSP]);

    if (hhll != ppat -> hhll)
	redQField (redbuf, mylist [HHLL]);

    if (hilo != ppat -> hilo)
	redQField (redbuf, mylist [HILO]);

    if (rtct != ppat -> rtct)
	redQField (redbuf, mylist [RTCT]);

    if (sbct != ppat -> sbct)
	redQField (redbuf, mylist [SBCT]);

    if (hlsc != ppat -> hlsc)
	redQField (redbuf, mylist [HLSC]);

    if (larf != ppat -> larf)
	redQField (redbuf, mylist [LARF]);

    if (hand != ppat -> hand)
	redQField (redbuf, mylist [HAND]);

    if (hpos != ppat -> hpos)
	redQField (redbuf, mylist [HPOS]);

    if (hneg != ppat -> hneg)
	redQField (redbuf, mylist [HNEG]);

    /* NOTE: RVAL must be sent last */

    if (ppat -> ncmd || (rval != ppat -> rval))
    {
	db_post_events ((struct dbCommon *)ppat, (caddr_t)&ppat -> rval, DBE_VALUE);
	redQField (redbuf, mylist [RVAL]);
    }

    redQSend (redbuf);

    /* process the forward scan link record */

    if (ppat->flnk.type == DB_LINK)  
	dbScanPassive ((struct dbAddr *)ppat->flnk.value.db_link.pdbAddr);

    ppat->pact = FALSE;

    return (status);
}


/*------------------------------------------------------------------------*/
/*
NOTE:	Cycle definition for PAT
	POIV/NOIV are 0 (not invert signal)
      
	DURATION TYPE
        -------------------
        |                 |
        |                 |
        |                 |
    ----|                 |------
       start              end


	POIV/NOIV are 1 (invert signal)

	DURATION TYPE
       start              end
    ----|                 |------
	|                 |
	|                 |
	|-----------------|

	truth table for RBV and POIV/NOIV
		RBV	POIV/NOIV
		0	0	at end of cycle
		1	0	at mid cycle (on for DURATION)
		0	1	at mid cycle (on for DURATION)
		1	1	at end of cycle
*/

static void patCallback (pcallback)
    CALLBACK_PTR	pcallback;
{
    struct patRecord	*ppat = (struct patRecord *) (pcallback->dbAddr.precord);
    struct patbodset	*ppatbodset = (struct patbodset *) (ppat -> dset);
    int			wait_time = 0;
    long		status = 0;
    long		temp;
    int			wslv = 0;
    unsigned short	master = 0;

    /* we must find out if we are SLAVE, so that we do not do 
       any write_pat(), by setting wslv, when the SLAVE becomes
       MASTER, it will allow us to do some cleanup in process
    */
    if (resGetCaAnswer (ppat -> radr, &master) == RES_OK)
    {
	if (!master)
	{
	    ppat -> wslv = 1;
	    wslv = 1;
	}
	else
	    wslv = 0;
    }

    switch (ppat -> cbfl)
    {
	case PAT_START_CALLBACK:
	    /* we were going in the positive direction */
	    if (ppat -> poso)
	    {
		if (ppat -> ncmd)
		{
		/* the callback was cancelled in process(),
		   we have to turn off/on the output to finish
		   the duration and institute the minimum off time
		   if it exists
		*/
		    if (ppat -> rval == 0)
		    {
			if (ppat -> poiv)
			    ppat -> rbv = ppat -> pmsk;
			else
			    ppat -> rbv = 0;
			if(!wslv)
			    status = (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
			ppat -> poso = 0;
			ppat -> cbfl = PAT_NO_CALLBACK;
			ppat -> ncmd = PAT_NO_CMD;
		    }
		    /* we are already going in the pos direction, continue
		       that way
		    */
		    else if (ppat -> rval > 0)
		    {
			ppat -> poso = 1;
			ppat -> cbfl = PAT_DURATION_ON;
		    	temp = ppat -> rval;
		    	if (temp < 0)
			    temp = -temp;
		    	ppat -> pcnt = temp;
		    	ppat -> svft = ppat -> moft;
			ppat -> svnt = ppat -> mont;
		    	if (ppat -> svnt)
		    	{
			    wait_time = (int)ppat -> svnt;
			    ppat -> mocb = 1;
		    	}
		    	else
		    	{
		    	    wait_time = (int)temp;
			    ppat -> mocb = 0;
		    	}
			ppat -> ncmd = PAT_NO_CMD;
		    }
		    /* rval < 0, we have to change direction */
		    else
		    {
			if (ppat -> poiv)
			    ppat -> rbv = ppat -> pmsk;
			else
			    ppat -> rbv = 0;
			ppat -> poso = 0;
			/* write the the off condition */
			if (!wslv)
			    status = (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
			if (!status)
			{
			    if (ppat -> noiv)
				ppat -> rbv = 0;
			    else
				ppat -> rbv = ppat -> nmsk;
			    if (!wslv)
			    	status = (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
			    if (!status)
			    {
				ppat -> nego = 1;
		    		temp = ppat -> rval;
		    		if (temp < 0)
			    	    temp = -temp;
		    		ppat -> pcnt = temp;
		    		ppat -> svft = ppat -> moft;
				ppat -> svnt = ppat -> mont;
		    		if (ppat -> svnt)
		    		{
			    	    wait_time = (int)ppat -> svnt;
			    	    ppat -> mocb = 1;
		    		}
		    		else
		    		{
		    	    	    wait_time = (int)temp;
			    	    ppat -> mocb = 0;
		    		}
				ppat -> cbfl = PAT_DURATION_ON;
			    }
			    else
			    {
				ppat -> nego = 0;
				ppat -> cbfl = PAT_NO_CALLBACK;
			    }
			    ppat -> ncmd = PAT_NO_CMD;
			}
			else
			{
			    ppat -> poso = ppat -> nego = 0;
			    ppat -> ncmd = PAT_NO_CMD;
			    ppat -> cbfl = PAT_NO_CALLBACK;
			}
		    }
		}
		/* error should not be here */
		else
		{
		    ;
		}
	    }
	    /* we were going in the negative direction */
	    else if (ppat -> nego)
	    {
		if (ppat -> ncmd)
		{
		    /* the callback was cancelled in process(),
		       we have to turn off/on the output to finish
		       the duration and institute the minimum off time
		       if it exists
		    */
		    if (ppat -> rval == 0)
		    {
			if (ppat -> noiv)
			    ppat -> rbv = ppat -> nmsk;
			else
			    ppat -> rbv = 0;
			if (!wslv)
			    status = (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
			ppat -> cbfl = PAT_NO_CALLBACK;
			ppat -> ncmd = PAT_NO_CMD;
			ppat -> nego = 0;
		    }
		    /* we are already going in the neg direction, continue
		       that way
		    */
		    else if (ppat -> rval < 0)
		    {
			ppat -> nego = 1;
			ppat -> cbfl = PAT_DURATION_ON;
		    	temp = ppat -> rval;
		    	if (temp < 0)
			    temp = -temp;
		    	ppat -> pcnt = temp;
		    	ppat -> svft = ppat -> moft;
			ppat -> svnt = ppat -> mont;
		    	if (ppat -> svnt)
		    	{
			    wait_time = (int)ppat -> svnt;
			    ppat -> mocb = 1;
		    	}
		    	else
		    	{
		    	    wait_time = (int)temp;
			    ppat -> mocb = 0;
		    	}
			ppat -> ncmd = PAT_NO_CMD;
		    }
		    /* rval > 0, we have to change direction */
		    else
		    {
			if (ppat -> noiv)
			    ppat -> rbv = ppat -> nmsk;
			else
			    ppat -> rbv = 0;
			ppat -> nego = 0;
			/* write the the off condition */
			if (!wslv)
			    status = (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
			if (!status)
			{
			    if (ppat -> poiv)
				ppat -> rbv = 0;
			    else
				ppat -> rbv = ppat -> pmsk;
			    if (!wslv)
			    	status = (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
			    if (!status)
			    {
				ppat -> poso = 1;
		    		temp = ppat -> rval;
		    		if (temp < 0)
			    	    temp = -temp;
		    		ppat -> pcnt = temp;
		    		ppat -> svft = ppat -> moft;
				ppat -> svnt = ppat -> mont;
		    		if (ppat -> svnt)
		    		{
			    	    wait_time = (int)ppat -> svnt;
			    	    ppat -> mocb = 1;
		    		}
		    		else
		    		{
		    	    	    wait_time = (int)temp;
			    	    ppat -> mocb = 0;
		    		}
				ppat -> cbfl = PAT_DURATION_ON;
			    }
			    else
			    {
				ppat -> poso = 0;
				ppat -> cbfl = PAT_NO_CALLBACK;
			    }
			    ppat -> ncmd = PAT_NO_CMD;
			}
			else
			{
			    ppat -> poso = ppat -> nego = 0;
			    ppat -> ncmd = PAT_NO_CMD;
			    ppat -> cbfl = PAT_NO_CALLBACK;
			}
		    }
		}
		/* nothing to do , this is an error, should not be here */
		else
		{
			;
		}
	    }
	    /* we weren't going anywhere */
	    else
	    {
		if (ppat -> rval > 0)
		{
		    if (ppat -> poiv)
			ppat -> rbv = 0;
		    else
			ppat -> rbv = ppat -> pmsk;
		    if (!wslv)
		    	status = (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
		    if (!status)
		    {
			ppat -> poso = 1;
			ppat -> cbfl = PAT_DURATION_ON;
		    }
		}
		else if (ppat -> rval < 0)
		{
		    if (ppat -> noiv)
			ppat -> rbv = 0;
		    else
			ppat -> rbv = ppat -> nmsk;
		    if (!wslv)
		    	status = (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
		    if (!status)
		    {
			ppat -> nego = 1;
			ppat -> cbfl = PAT_DURATION_ON;
		    }
		}
		if (!status)
		{
		    temp = ppat -> rval;
		    if (temp < 0)
			temp = -temp;
		    ppat -> pcnt = temp;
		    ppat -> svft = ppat -> moft;
		    ppat -> svnt = ppat -> mont;
		    if (ppat -> svnt)
		    {
			wait_time = (int)ppat -> svnt;
			ppat -> mocb = 1;
		    }
		    else
		    {
		    	wait_time = (int)temp;
			ppat -> mocb = 0;
		    }
		}
		else
		{
		    ppat -> poso = ppat -> nego = 0;
		    ppat -> mocb = 0;
		    ppat -> cbfl = PAT_NO_CALLBACK;
		}
		ppat -> ncmd = PAT_NO_CMD;
	    }
	    break;

	case PAT_DURATION_ON:
	    /* at the end of a DURATION , turn off/on signal to end duration*/
	    if (ppat -> poso)
	    {
		if (ppat -> poiv)
		    ppat -> rbv = ppat -> pmsk;
		else
		    ppat -> rbv = 0;
		if (!wslv)
		    status = (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
		if (!status)
		{
		    if (ppat -> mocb)
		    {
			wait_time = (int)ppat -> svft;
			ppat -> cbfl = PAT_DURATION_OFF;
		    }
		    else
		    {
			ppat -> cbfl = PAT_NO_CALLBACK;
			ppat -> poso = 0;
		    }
		}
		else
		{
		    ppat -> cbfl = PAT_NO_CALLBACK;
		    ppat -> poso = 0;
		    ppat -> mocb = 0;
		}
	    }
	    else if (ppat -> nego)
	    {
		if (ppat -> noiv)
		    ppat -> rbv = ppat -> nmsk;
		else
		    ppat -> rbv = 0;
		if (!wslv)
		    status = (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
		if (!status)
		{
		    if (ppat -> mocb)
		    {
			wait_time = (int)ppat -> svft;
			ppat -> cbfl = PAT_DURATION_OFF;
		    }
		    else
		    {
			ppat -> cbfl = PAT_NO_CALLBACK;
			ppat -> nego = 0;
		    }
		}
		else
		{
		    ppat -> cbfl = PAT_NO_CALLBACK;
		    ppat -> nego = 0;
		    ppat -> mocb = 0;
		}
	    }
	    else
	    {
		ppat -> cbfl = PAT_NO_CALLBACK;
		ppat -> nego = 0;
		ppat -> mocb = 0;
		ppat -> poso = 0;
	    }
	    break;

	case PAT_DURATION_OFF:
	    if (ppat -> ncmd)
	    {
		if (ppat -> rval > 0)
		{
		    if (ppat -> poiv)
			ppat -> rbv = 0;
		    else
			ppat -> rbv = ppat -> pmsk;
		    if (!wslv)
		    	status = (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
		    if (!status)
		    {
			ppat -> poso = 1;
			ppat -> cbfl = PAT_DURATION_ON;
		    }
		}
		else if (ppat -> rval < 0)
		{
		    if (ppat -> noiv)
			ppat -> rbv = 0;
		    else
			ppat -> rbv = ppat -> nmsk;
		    if (!wslv)
		    	status = (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
		    if (!status)
		    {
			ppat -> nego = 1;
			ppat -> cbfl = PAT_DURATION_ON;
		    }
		}
		if (!status)
		{
		    temp = ppat -> rval;
		    if (temp < 0)
			temp = -temp;
		    ppat -> pcnt = temp;
		    ppat -> svft = ppat -> moft;
		    ppat -> svnt = ppat -> mont;
		    if (ppat -> svnt)
		    {
			wait_time = (int)ppat -> svnt;
			ppat -> mocb = 1;
		    }
		    else
		    {
		    	wait_time = (int)temp;
			ppat -> mocb = 0;
		    }
		}
		else
		{
		    ppat -> poso = 0;
		    ppat -> nego = 0;
		    ppat -> mocb = 0;
		    ppat -> cbfl = PAT_NO_CALLBACK;
		}
		ppat -> ncmd = PAT_NO_CMD;
	    }
	    else
	    {
		if (ppat -> mocb)
		{
		    ppat -> pcnt -= ppat -> svnt;
		    if (ppat -> pcnt <= 0)
		    {
		    	ppat -> cbfl = PAT_NO_CALLBACK;
		    	ppat -> poso = 0;
		    	ppat -> nego = 0;
			ppat -> mocb = 0;
		    }
		    else
		    {
			if (ppat -> poso)
			{
		    	    if (ppat -> poiv)
				ppat -> rbv = 0;
		    	    else
				ppat -> rbv = ppat -> pmsk;
			    if (!wslv)
		    	    	status = (*ppatbodset -> write_pat)(ppat, PAT_POS_CHANNEL);
		    	    if (!status)
		    	    {
				ppat -> poso = 1;
				ppat -> cbfl = PAT_DURATION_ON;
		    	    }
			}
			else if (ppat -> nego)
			{
		    	    if (ppat -> noiv)
				ppat -> rbv = 0;
		    	    else
				ppat -> rbv = ppat -> nmsk;
			    if (!wslv)
		    	    	status = (*ppatbodset -> write_pat)(ppat, PAT_NEG_CHANNEL);
		    	    if (!status)
		    	    {
				ppat -> nego = 1;
				ppat -> cbfl = PAT_DURATION_ON;
		    	    }
			}
			if (!status)
			{
		    	    if (ppat -> pcnt >= ppat -> svnt)
		    	    {
				wait_time = (int)ppat -> svnt;
				ppat -> mocb = 1;
		    	    }
		    	    else
		    	    {
		    		wait_time = (int)ppat -> pcnt;
				ppat -> mocb = 1;
		    	    }
		    	}
		    }
		}
		else
		{
		    ppat -> cbfl = PAT_NO_CALLBACK;
		    ppat -> poso = 0;
		    ppat -> nego = 0;
		    ppat -> mocb = 0;
		}
	    }

	    break;

	default:
	    break;
    }
    if (wait_time)
    {
	if (pcallback -> wd_id == NULL)
	{
	    pcallback -> wd_id = wdCreate ();
	}
	wdStart (pcallback -> wd_id, wait_time, (FUNCPTR)callbackRequest, (int)pcallback);
    }

    return;
}



/*------------------------------------------------------------------------*/

static void slope_and_offset (ppat)
   struct patRecord	*ppat;
{
   if (ppat -> irhi == ppat -> irlo)
      ppat -> eslo = 0.0;
   else
      ppat -> eslo = 100.0 / (ppat -> irhi - ppat -> irlo);

   ppat -> roff = 0.0 - (ppat -> eslo * ppat -> irlo);

   return;
}


/*------------------------------------------------------------------------*/

static long special (paddr, after)
    struct dbAddr	*paddr;
    int			after;
{
    struct patRecord		*ppat = (struct patRecord *)(paddr -> precord);
    CALLBACK_PTR		pcallback = (CALLBACK_PTR)(ppat -> dpvt);
    int				special_type = paddr -> special;
    int				wait_time;
    short			master = 0;

    if (!after)
    {
	return (0);
    }

    resGetCaAnswer (ppat -> radr, &master);

    switch (special_type)
    {
	case SPC_PAT_CT:
	    ppat -> ct = tickGet();
	    break;

	case SPC_PAT_RVAL:
	    /* RVAL was sent by the MASTER, so we may have to start a callback
	       on the slave, to keep, MASTER/SLAVE in sync
	    */
	    /* do not start a callback if we are not fully init'd */
	    if (!ppat -> ini2)
		break;
	
	    if (ppat -> rval || ppat -> cbfl)
	    {
		ppat -> ncmd = PAT_NEW_CMD;
		if ( ( (ppat -> cbfl == PAT_DURATION_ON) ||
		       (ppat -> cbfl == PAT_DURATION_OFF) ) && (!ppat -> mocb) )
		{
		    wdCancel (pcallback -> wd_id);
		    ppat -> cbfl = PAT_NO_CALLBACK;
		}
	    }

	    if (ppat -> ncmd)
	    {
		if (ppat -> cbfl == PAT_NO_CALLBACK)
		{
		    wait_time = 1;
		    ppat -> cbfl = PAT_START_CALLBACK;
		    if (pcallback -> wd_id == NULL)
		    {
			pcallback -> wd_id = wdCreate();
		    }
		    wdStart (pcallback -> wd_id, wait_time, (FUNCPTR)callbackRequest, 
			     (int)pcallback);
		}
	    }
	    break;

	case SPC_PAT_HAND:
	    if (!master)
	    {
		if (ppat -> hand)
		{
		    if (ppat-> cbfl != PAT_NO_CALLBACK)
		    {
			wdCancel (pcallback -> wd_id);
			ppat -> cbfl = PAT_NO_CALLBACK;
			ppat -> poso = 0;
			ppat -> nego = 0;
		    }
		    ppat -> lhnd = ppat -> hand;
		}
		else
		{
		    ppat -> lhnd = ppat -> hand;
		}
		ppat -> wslv = 1;
	    }
	    break;

	default:
	    recGblDbaddrError (S_db_badChoice, paddr, "pat: special");
	    return (S_db_badChoice);
    }

    return (0);

}
