/*recPot.c,v 1.1.1.1 1999/08/03 14:44:20 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************************
 *
 *  MODULE NAME            : recPot.c 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 June 93
 *
 *
 *  DESCRIPTION: record support routines for teh Point-Of-Transfer record
 *
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 23 Feb 94
 *  REVISION:  removed INIT, ALST, MLST from namelist and potred, not needed
 *             also removed time from potred, not needed
 *             only db_post_events if MASTER, in some places the init flag
 *             is checked before posting, if init is set then we were called
 *             from process and are MASTER.
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 02 Mar 94
 *  REVISION: added process stats code
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 29 Mar 94
 *  REVISION:  make forward func def agree with actual func def
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 May 94
 *  REVISION: 	add INI2, PCMD and cmds to redundancy list, move db_post_events
 *		to post_events(), rewrite init_record, FindAPITable now
 *		returns a status instead of void
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 June 94
 *  REVISION:	added support for maximize severity (msev), moved call to
 *		tsLocalTime inside on scan
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 1 Sep 94
 *  REVISION:	added get_control_double(),get_graphic_double() and
 *		get_link_value()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 Oct 94
 *  REVISION:	add TIME to redundancy, added fpps
 *		(1st process pass), do time stamp when stat or sevr chg,
 *		see monitor(),and not in process(),set lalm to val when 
 *		stat or sevr chg,in process() if fpps (first process)
 *		init lalm to val and if time stamp is 0, init it also
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 Oct 94
 *  REVISION:	chg time to latm (last alarm time)
 *		added lptm (last process time)
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 Jan 95
 *  REVISION:	added mlst to redundancy, init latm and lptm in init_record
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 19 Feb 96
 *  REVISION:	made globals into struct for reentranvy
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 Jun 96
 *  REVISION:	added code for alarm disable ADIS
 *
****************************************************************************/


/****************************************************************************
 *
 * Programmer's Notes:
 *
 *
****************************************************************************/



#include <vxWorks.h>
#include <inetLib.h>
#include <types.h>
#include <stdioLib.h>
#include <lstLib.h>
#include <string.h>

#include "alarm.h"
#include "choice.h"
#include "cvtTable.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"
#include "dbFldTypes.h"
#include "dbRecType.h"
#include "devSup.h"
#include "errMdef.h"
#include "recSup.h"
#include "special.h"
#include "rec/potRecord.h"
#include "redflds.h"
#include "redque.h"
#include "potredfldidx.h"
#include "tsDefs.h"          
#include "tankfarm.h"
#include "restable.h"
#include "db_addr.h"
#include "recGbl.h"
#include "dbEvent.h"
#include "okToUse.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table */

#define report NULL
#define initialize NULL
static long init_record();
static long process();
static long special();
static long get_value ();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double ();
static long get_control_double ();
#define get_alarm_double NULL

struct rset potRSET = {

   RSETNUMBER,
   report,
   initialize,
   init_record,
   process,
   special,
   get_value,
   cvt_dbaddr,
   get_array_info,
   put_array_info,
   get_units,
   get_precision,
   get_enum_str,
   get_enum_strs,
   put_enum_str,
   get_graphic_double,
   get_control_double,
   get_alarm_double
   };


/* The folowing definitions are for redundancy */
 
static struct fldDes **mylist = 0;

   /* NOTE: the order and spelling in namelist must be the
            same as in potredfldidx.h
   */

static char *namelist [] = {
               "STAT",
               "SEVR",
               "VAL",
               "STTS",
               "POSV",
               "QVOL",
               "LALM",
               "ESOU",
	       "INI2",
	       "PCMD",
	       "CMDS",
	       "CISS",
	       "CREC",
	       "CSTP",
	       "ISSC",
	       "RESC",
	       "STSC",
	       "LATM",
	       "MLST",
               0
               };

   /* structure to save some of the values in the
      record at start of processing, most are used
      for redunadancy checking
   */

struct potred
{
    unsigned short	stat;
    unsigned short	sevr;
    double		val;
    char		stts[18];
    char		pcmd[18];
    char		cmds[18];
    short		posv;
    double		qvol;
    double		lalm;
    short		esou;
    short		ini2;
    short		ciss;
    short		crec;
    short		cstp;
    short		issc;
    short		resc;
    short		stsc;
    TS_STAMP		latm;
    double		mlst;
};

/* defines for POT commands
   NOTE: these must have these values to
         coincide with potRecord.ascii
*/

#define  POT_ISSUE_CMD     1
#define  POT_RECEIVE_CMD   2
#define  POT_STOP_CMD      3

/* defines for POT status */

#define  POT_ISSUING    0
#define  POT_RECEIVING  1
#define  POT_STOPPED    2

struct pot_vars
{
    short ground_ok;           /* set by read_ground_ok_pv () */
    short transfer_permit_ok;  /* set by read_transfer_permit_pv () */
};

static   unsigned short type;                /* index to pot type */

static void alarm ();
static long check_pot_status (struct potRecord *, struct pot_vars *);
static long enable_transfer ();
static void get_link_value ();
static void monitor ();
static void post_events ();
/*long read_corrected_volume_pv ();*/
static long read_custody_pv ();
/*long read_flow_pv ();*/
static long read_ground_ok_pv (struct potRecord *, struct pot_vars *);
/*long read_meter_pv ();*/
/*long read_pressure_pv ();*/
static long read_sensors (struct potRecord *, struct pot_vars *);
/*long read_temperature_pv ();*/
static long read_transfer_permit_pv (struct potRecord *, struct pot_vars *);
static void reset_pot_commands ();
static long run_commands (struct potRecord *, struct pot_vars *);
static void save_pot ();
static void update_redun ();



/****************************************************************************
 *
 *  MODULE NAME            : init_record 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *  ARGUMENT(S)            : ppot - ptr to pot record  
 *
 *  DESCRIPTION: initialize a pot record
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION:  init PCMD (last command), TF_INIT_CMD defined in
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 03 Jan 94
 *  REVISION:  set ack to 0
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 23 Feb 94
 *  REVISION: only do db_post_events if MASTER, removed ack = 0
 *
****************************************************************************/

static long init_record (ppot)
   struct potRecord	*ppot;

{
    long		status = 0;
    unsigned short	i;
    char		*pstr;
    int			dont_init = 0;
    int			ini2_flag = 0;
    struct potred	pot_old;
    struct pot_vars	potVars;

    if (!(mylist))
    {
         /* find record type for pot */
        for (type = 1, i = 0; type < 100; type++)
        {
            pstr = GET_PRECTYPE (type);     
            if (strncmp (pstr, "pot", 3) == 0)
            {
                i = 1;
                break;
            }
        }

        if (i == 0)
        {
            recGblRecordError (S_db_noRecType, (struct dbCommon *)ppot, "pot:init_record");
            return (S_db_noRecType);
        }
           /* allocate memory for mylist */
        mylist = redFldAlloc (type, NUMPOTFLDS);

	if (!mylist)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)ppot,
			   "pot, init_record, mylist failed");
  	    return (S_rec_Alloc);
 
	}

           /* add fields in namelist to mylist */
        for (i = 0; i < NUMPOTFLDS; i++)
             redFldAdd (type, i, namelist [i]);
    }

      /* now find the API table for this record
         but 1st make sure ptr to API tables is NULL */

    ppot -> papi = NULL;
    status = FindAPITable (ppot -> apit, &ppot -> papi, (int *)&ppot -> anum);
    if (status)
	return (status);
	
	/* save old pot status */
    save_pot (ppot, &pot_old);

    bzero (ppot -> cmds, sizeof (ppot -> cmds));
    ppot -> ciss = 0;
    ppot -> crec = 0;
    ppot -> cstp = 0;
    ppot -> issc = 0;
    ppot -> resc = 0;
    ppot -> stsc = 0;
    ppot -> rate = 0;
    ppot -> amnt = 0;
    ppot -> qvol = 0;
    ppot -> texp = 0;
    ppot -> prec = 2;       /* display precision */
    ppot -> esou = 0;
    strncpy (ppot -> pcmd, TF_INIT_CMD, sizeof (ppot -> pcmd));

       /* set pot's status to STOPPED */
    strncpy (ppot -> stts, ppot -> stps, sizeof (ppot -> stts));
    ppot -> posv = POT_STOPPED;

    ppot -> init = TRUE;
    ppot -> fpps = 1;
    ppot -> nsta = NO_ALARM;
    ppot -> nsev = NO_ALARM;
    ppot -> msev = NO_ALARM;
    /* init last alarm tiem and last process time */
    if (ppot -> latm.secPastEpoch == 0)
	tsLocalTime (&ppot -> latm);
    if (ppot -> lptm.secPastEpoch == 0)
	tsLocalTime (&ppot -> lptm);

    status = read_sensors (ppot, &potVars);
    if (status)
    {
	ppot -> nsta = NO_INIT_ALARM;
	ppot -> nsev = VALID_ALARM;
    }
    else
    {
	ppot -> qvol = ppot -> val;
	if (ppot -> etnf.type != CONSTANT)
	{
	    if (ppot -> etnf.type == DB_LINK)
	    {
		if (!okToUse (&ppot -> etnf) )
		    dont_init = 1;
	    }
	}
	if (dont_init)
	{
	    ppot -> nsta = NO_INIT_ALARM;
	    ppot -> nsev = VALID_ALARM;
	}
	else
	    ini2_flag = 1;
    }

      /* do redundancy if not in slave mode */

    if (resGetCaAnswer (ppot -> radr, &i) == RES_OK)
    {
	if (i)
	{
	    /* check for maximize severity */

	    if (ppot -> msev > ppot -> nsev)
	    {
		ppot -> nsev = ppot -> msev;
		ppot -> nsta = LINK_ALARM;
	    }

	    monitor (ppot);

	    if (ini2_flag)
	    {
		ppot -> ini2 = 1;
		post_events (ppot, &pot_old);
		update_redun (&pot_old, ppot);
	    }
	}
    }

   return (0);

} /* end func init_records () */



/****************************************************************************
 *
 *  MODULE NAME            : alarm 
 *  AUTHOR                 : emsteel
 *  DATE                   : 25 Mar 93
 *
 *  CALLING MODULE(S)      : process ()
 *  ARGUMENT(S)            : ppot
 *
 *  DESCRIPTION: set alarm status. POT must be stopped to check if any
 *                unwanted flow
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 Mar 94
 *  REVISION: alarm chgs, if val is past alarm setting, set that
 *            alarm condition, else if val was in that alarm and
 *            it is still in movd range of alarm value, then stay
 *			in alarm
 *
****************************************************************************/

static void alarm (ppot)
     struct potRecord  *ppot;
{
   int		emergency_stop = 0;
   int		adis = 0;
   double	delta;

   if (ppot -> posv != POT_STOPPED)
      return;

      /* check for inhibited alarms */

   adis = AlarmDisableLink ((void *)ppot);
   if ( (ppot -> aenb == ALARM_DISABLED) || (adis != 0) )
   {
      ppot -> nsta = INHIBIT_ALARM;
      ppot -> nsev = NO_ALARM;
      return;
   }

   if (ppot -> udf == TRUE)
   {
      if (ppot -> nsev < VALID_ALARM)
      {
         ppot -> nsta = UDF_ALARM;
         ppot -> nsev = VALID_ALARM;
      }
      return;
   }

   if (ppot -> nsev == VALID_ALARM)
      return;


      /* test against volume change value */

   if (ppot -> nsev < ppot -> movs)
   {
         /* stopped volume - volume */

      delta = ppot -> qvol - ppot -> val;
      if (delta < 0.0)
         delta = -delta;

         /* alarm if change in volume is > alarm limit or
            pot was in volume alarm and change is still within
            movd of alarm limit
         */
      if ( (delta > ppot -> vold) ||
           ( (ppot -> stat == VOLUME) &&
             (delta > (ppot -> vold - ppot -> movd)) ) )
      {
         ppot -> nsta = VOLUME;
         ppot -> nsev = ppot -> movs;
 
            /* update qvol (stopped volume) for next scan */

         ppot -> qvol = ppot -> val;
         
           /* if this is to init an emergency shutdown (estp) 
              then set esou so the world knows, if they care!
           */

         if (ppot -> estp)
         {
            emergency_stop++;
         }
      }
   }
   if (emergency_stop)
      ppot -> esou = 1;
   else
      ppot -> esou = 0;

   return;

} /* end func alarm () */



/****************************************************************************
 *
 *  MODULE NAME            : check_pot_status 
 *  AUTHOR                 : emsteel
 *  DATE                   : 09 June 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ppot - ptr to pot record
 *
 *  DESCRIPTION: check for drop or ready status, if pot is issuing or
 *                receiving and either transfer_permit_ok or ground_ok are
 *                0, have dropped ready,
 *                   alarm
 *                   stop pot
 *                   send enable_pv a 0
 *                   update qvol (stopped volume) with current volume (val)
 *
****************************************************************************/

static long check_pot_status (ppot, potVars)
   struct potRecord     *ppot;
   struct pot_vars	*potVars;
{
   long           status = 0;
   unsigned short enable;

   if ((ppot -> posv == POT_ISSUING) || (ppot -> posv == POT_RECEIVING))
   {
      if (potVars -> transfer_permit_ok && potVars -> ground_ok)
         return (status);

      TankFarmPostAlarm ((struct dbCommon *)ppot, POT_DROP_READY,
                                                               MAJOR_ALARM);
      strncpy (ppot -> stts, ppot -> stps, sizeof (ppot -> stts));
      ppot -> posv = POT_STOPPED;
      ppot -> qvol = ppot -> val;
      enable = 0;
      enable_transfer (ppot, &enable);
      status = S_tank_PotDropReady;
   }

   return (status);

} /* end func check_pot_status () */



/****************************************************************************
 *
 *  MODULE NAME            : enable_transfer 
 *  AUTHOR                 : emsteel
 *  DATE                   : 09 June 93
 *
 *  CALLING MODULE(S)      : check_pot_status, run_commands
 *          ARGUMENT(S)    : ppot - ptr to pot record
 *                           enable, 0 = disable, 1 = enable transfer
 *
 *  DESCRIPTION: if the enable transfer pv exists then send a 1 to enable
 *               transfer and a 0 to disable. this acts as a permissive
 *               to dispensing hardware
 *
****************************************************************************/

static long enable_transfer (ppot, enable)
   struct potRecord     *ppot;
   unsigned short       *enable;
{
   long status    = 0;
   long nRequests = 1;

   if (ppot -> etnf.type != CONSTANT)
   {
      if (ppot -> etnf.type == DB_LINK)
      {
         status = dbPutLink (&ppot -> etnf.value.db_link,
                             (struct dbCommon *)ppot, DBR_ENUM,
                             (caddr_t)enable, nRequests);
         if (status)
            TankFarmPostAlarm ((struct dbCommon *)ppot, LINK_ALARM,
                                                               VALID_ALARM);
      }
   }
   return (status);

} /* end func enable_transfer () */


/****************************************************************************
 *
 *  MODULE NAME            : get_control_double 
 *  AUTHOR                 : emsteel
 *  DATE                   : 1 Sep 94
 *
 *
 *  DESCRIPTION: return control limits
 *
****************************************************************************/

static long get_control_double(paddr,pcd)
    struct	dbAddr		*paddr;
    struct	dbr_ctrlDouble	*pcd;
{
    struct   potRecord     *ppot = (struct potRecord *) paddr->precord;
    float		    hopr;
    float		    lopr;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct potRecord *)0) -> val) ):
	    hopr = lopr = 0;
	    if (ppot -> cusp.type == DB_LINK)
	    {
	    	get_link_value (ppot -> cusp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ppot -> cusp.value.db_link.pvname, ".LOPR", &lopr);
	    
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		
		pcd -> upper_ctrl_limit = 1000000;
		pcd -> lower_ctrl_limit = 0;
	    }
	    else
	    {
		pcd -> upper_ctrl_limit = hopr;
		pcd -> lower_ctrl_limit = lopr;
	    }
	    break;
		
	case ( (int)&(((struct potRecord *)0) -> vold) ):
	case ( (int)&(((struct potRecord *)0) -> adel) ):
	case ( (int)&(((struct potRecord *)0) -> mdel) ):
	case ( (int)&(((struct potRecord *)0) -> movd) ):
	    hopr = lopr = 0;
	    if (ppot -> cusp.type == DB_LINK)
	    {
	    	get_link_value (ppot -> cusp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ppot -> cusp.value.db_link.pvname, ".LOPR", &lopr);
	    
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		
		pcd -> upper_ctrl_limit = 1000000;
		pcd -> lower_ctrl_limit = 0;
	    }
	    else
	    {
		pcd -> upper_ctrl_limit = hopr - lopr;
		pcd -> lower_ctrl_limit = 0;
	    }
	    break;

	case ( (int)&(((struct potRecord *)0) -> esou) ):
	    pcd -> upper_ctrl_limit = 1;
	    pcd -> lower_ctrl_limit = 0;
	    break;

    }

    return (0);

} /* end func get_control_double () */


/****************************************************************************
 *
 *  MODULE NAME            : get_graphic_double 
 *  AUTHOR                 : emsteel
 *  DATE                   : 1 Sep 94
 *
 *
 *  DESCRIPTION: return graphic limits
 *
****************************************************************************/

static long get_graphic_double(paddr,pgd)
    struct	dbAddr		*paddr;
    struct	dbr_grDouble	*pgd;
{
    struct potRecord    *ppot = (struct potRecord *) paddr->precord;
    float		    hopr;
    float		    lopr;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct potRecord *)0) -> val) ):
	case ( (int)&(((struct potRecord *)0) -> qvol) ):
	    hopr = lopr = 0;
	    if (ppot -> cusp.type == DB_LINK)
	    {
	    	get_link_value (ppot -> cusp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ppot -> cusp.value.db_link.pvname, ".LOPR", &lopr);
	    
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		
		pgd -> upper_disp_limit = 1000000;
		pgd -> lower_disp_limit = 0;
	    }
	    else
	    {
		pgd -> upper_disp_limit = hopr;
		pgd -> lower_disp_limit = lopr;
	    }
	    break;
		
	case ( (int)&(((struct potRecord *)0) -> vold) ):
	case ( (int)&(((struct potRecord *)0) -> adel) ):
	case ( (int)&(((struct potRecord *)0) -> mdel) ):
	case ( (int)&(((struct potRecord *)0) -> movd) ):
	    hopr = lopr = 0;
	    if (ppot -> cusp.type == DB_LINK)
	    {
	    	get_link_value (ppot -> cusp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ppot -> cusp.value.db_link.pvname, ".LOPR", &lopr);
	    
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		
		pgd -> upper_disp_limit = 1000000;
		pgd -> lower_disp_limit = 0;
	    }
	    else
	    {
		pgd -> upper_disp_limit = hopr - lopr;
		pgd -> lower_disp_limit = 0;
	    }
	    break;

	case ( (int)&(((struct potRecord *)0) -> posv) ):
	    pgd -> upper_disp_limit = 2;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct potRecord *)0) -> esou) ):
	    pgd -> upper_disp_limit = 1;
	    pgd -> lower_disp_limit = 0;
	    break;

    }
    return (0);

} /* end func get_graphic_double () */


/****************************************************************************
 *
 *  MODULE NAME            : get_link_value
 *  AUTHOR                 : emsteel
 *  DATE                   : 1 Sep 94
 *
 *
 *  DESCRIPTION: do dbNameToAddr on a record field and return its value
 *
****************************************************************************/

static void get_link_value (recname, recfld, value)
    char	*recname;
    char	*recfld;
    float	*value;
{
    long     status    = 0;
    long     nRequests = 1;
    long     options   = 0;
    char     pvname_fld[36];
    float    val = 0;
    struct   dbAddr   dbAddr;

    strncpy (pvname_fld, recname, 28);
    strcat (pvname_fld, recfld);
    status = dbNameToAddr (pvname_fld, &dbAddr);
    if (status)
	return;
    status = dbGetField (&dbAddr, DBR_FLOAT, (caddr_t)&val,
			 &options, &nRequests, NULL);
    if (status)
	return;
    *value = val;
    return;

} /* end func get_link_value () */


/****************************************************************************
 *
 *  MODULE NAME            : get_precision 
 *  AUTHOR                 : emsteel
 *  DATE                   : 01 July 93
 *
 *
 *  DESCRIPTION: get display precision
 *  NOTE:
 *       set to default of 2 in init_record ()
 *
****************************************************************************/

static long get_precision (paddr, precision)
   struct dbAddr     *paddr;
   long              *precision;
{
   struct potRecord *ppot = (struct potRecord *)paddr -> precord;

   *precision = ppot -> prec;

   return (0);

} /* end func get_precision () */



/***************************************************************************
 *
 *  MODULE NAME            : get_value
 *  AUTHOR                 : emsteel
 *  DATE                   : 09 June 93
 *
 *  DESCRIPTION: return the value field (current xfer volume)
 *
****************************************************************************/

static long get_value (ppot, pvdes)
   struct potRecord    *ppot;
   struct valueDes     *pvdes;
{
   pvdes -> field_type = DBF_DOUBLE;
   pvdes -> no_elements = 1;
   (double *)(pvdes -> pvalue) = &ppot -> val;

   return (0);

} /* end func get_value () */



/****************************************************************************
 *
 *  MODULE NAME            : monitor 
 *  AUTHOR                 : emsteel
 *  DATE                   : 25 Mar 93
 *
 *  CALLING MODULE(S)      : process ()
 *
 *  DESCRIPTION: check event list
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 05 Jan 94
 *  REVISION:  check VALID_ALARM before set ACK
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 30 Aug 94
 *  REVISION:	rm DBE_LOG from ack, only archive if tenb (trending) is
 *		enabled
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 15 Mar 95
 *  REVISION:	do not set ack if sevr is EVENT_ALARM
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 11 Sept 95
 *  REVISION:	archive if stat/sevr also chg's
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 12 Sept 95
 *  REVISION:	chg to way handle VALID_ALARM
 *
****************************************************************************/

static void monitor (ppot)
   struct potRecord *ppot;
{
    unsigned short monitor_mask;
    double         delta;
    short          stat, sevr, nsta, nsev;
    int            set_ack = 1;

      /* get previous stat and sevr and new stat and sevr */

    stat = ppot -> stat;
    sevr = ppot -> sevr;
    nsta = ppot -> nsta;
    nsev = ppot -> nsev;

      /* set current stat and sevr */

    ppot -> stat = nsta;
    ppot -> sevr = nsev;
    ppot -> nsta = 0;
    ppot -> nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == ppot -> psta) && (nsev == ppot -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and out previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	ppot -> psta = stat;
	ppot -> psev = sevr;
	db_post_events ((struct dbCommon *)ppot, (caddr_t)&ppot -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)ppot, (caddr_t)&ppot -> psev, DBE_VALUE);
    }

    monitor_mask = 0;

      /* did alarm condition change this scan */

    if ((stat != nsta) || (sevr != nsev))
    {
	/* post events for alarm condition change */

	monitor_mask |= DBE_ALARM;

	/* when alarm status chg's, get current time and
	   save VAL in lalm (last value when alarm chg'd)
	*/
	tsLocalTime (&ppot -> latm);
	ppot -> lalm = ppot -> val;

	/* post stat and nsev fields */

	db_post_events ((struct dbCommon *)ppot, (caddr_t)&ppot -> stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)ppot, (caddr_t)&ppot -> sevr, DBE_VALUE);
	if ((ppot -> sevr != NO_ALARM) && (set_ack) &&
	    (ppot -> sevr != EVENT_ALARM) )
	{
	    if (ppot -> ack == 0)
	    {
		ppot -> ack = 1;
		db_post_events ((struct dbCommon *)ppot, (caddr_t)&ppot -> ack, DBE_VALUE);
	    }
	}
    }

      /* check for value changes */

    delta = ppot -> mlst - ppot -> val;  /* last monitored val - val */
    if (delta < 0.0)
	delta = -delta;

    if ( delta > ppot -> mdel)   /* monitor deadband */
    {
	/* post events for value change */

	monitor_mask |= DBE_VALUE;

	/* update last value monitored */

	ppot -> mlst = ppot -> val ;
    }

      /* check for archive change if tenb (trending) is enabled */

    if (ppot -> tenb)
    {
	delta = ppot -> alst - ppot -> val;
	if (delta < 0.0)
	    delta = -delta;

	/* if exceed archive deadband or there has been a change in alarm status */
	if ( (delta > ppot -> adel) || (monitor_mask & DBE_ALARM) )
	{
	    /* post events for value change */

	    monitor_mask |= DBE_LOG;

	    /* update last archive value */

	    ppot -> alst = ppot -> val;
	}
    }

    /* send out monitors connected to the value field */

    if (monitor_mask)
    {
	db_post_events ((struct dbCommon *)ppot, (caddr_t)&ppot -> val, monitor_mask);
    }

    return;

} /* end func monitor () */


/****************************************************************************
 *
 *  MODULE NAME            : post_events 
 *  AUTHOR                 : emsteel
 *  DATE                   : 20 May 94
 *
 *  PARAMETERS             : ppot - ptr to current pot record
 *			     ppot_old - ptr to old pot data
 * 
 *  DESCRIPTION: called from init_record and process, if chg between old data
 *		 and new then post for monitors
 *
****************************************************************************/

static void post_events (ppot, ppot_old)
    struct potRecord	*ppot;
    struct potred	*ppot_old;
{
    if (strncmp (ppot -> pcmd, ppot_old -> pcmd, sizeof (ppot -> pcmd)) != 0)
	db_post_events ((struct dbCommon *)ppot, ppot -> pcmd, DBE_VALUE);

    if (strncmp (ppot -> stts, ppot_old -> stts, sizeof (ppot -> stts)) != 0)
	db_post_events ((struct dbCommon *)ppot, ppot -> stts, DBE_VALUE);

    if (ppot -> posv != ppot_old -> posv)
	db_post_events ((struct dbCommon *)ppot, (caddr_t)&ppot -> posv, DBE_VALUE);

    if (ppot -> qvol != ppot_old -> qvol)
	db_post_events ((struct dbCommon *)ppot, (caddr_t)&ppot -> qvol, DBE_VALUE);

    if (ppot -> esou != ppot_old -> esou)
	db_post_events ((struct dbCommon *)ppot, (caddr_t)&ppot -> esou, DBE_VALUE);

    return;

} /* end func post_events () */


/****************************************************************************
 *
 *  MODULE NAME            : process 
 *  AUTHOR                 : emsteel
 *  DATE                   : 18 Mar 93
 *
 *  PARAMETERS             : paddr - ptr to record struct packet            
 * 
 *  DESCRIPTION: called when time to process a pipe record
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 17 Apr 95
 *  REVISION:	added code to process the forward link flnk
 *
****************************************************************************/

static long process (paddr)
       struct dbAddr *paddr;
{
    struct potRecord	*ppot    = (struct potRecord *)(paddr -> precord);
    long		status  = 0; 
    struct potred	pot_old;	/* save initial values for redundancy check */
    struct pot_vars	potVars;

      /* do nothing if we did not init */

    if ((!ppot -> init) || (!ppot -> ini2))
	return (status);

    /* get time stamp for last time record was processed */

    tsLocalTime (&ppot -> lptm);

    save_pot (ppot, &pot_old);

      /* init new status to previous */

    ppot -> nsta = ppot -> stat;
    ppot -> nsev = ppot -> sevr;
    ppot -> msev = NO_ALARM;

      /* off normal operations */

    if (ppot -> tagd)
    {
        ppot -> nsta = TAGGED_ALARM;
        ppot -> nsev = VALID_ALARM;
    }
    else if (ppot -> oscn == OFF_SCAN)
    {
        ppot -> nsta = SCAN_ALARM;
        ppot -> nsev = VALID_ALARM;
    }
    else
    {
        ppot -> nsta = NO_ALARM;
        ppot -> nsev = NO_ALARM;

        /* update number of process times */
        if (process_stats)
            process_cnts[type]++;	

          /* read the sensors */
   
	status = read_sensors (ppot, &potVars);
	if (!status)
	{
               /* set pot to active */
	    ppot -> pact = TRUE;

	    if (check_pot_status (ppot, &potVars))
	    {
		reset_pot_commands (ppot);
	    }
	    else
	    {
		run_commands (ppot, &potVars);
		alarm (ppot);
	    }
	}
	/* check for maximize severity */

	if (ppot -> msev > ppot -> nsev)
	{
           ppot -> nsev = ppot -> msev;
           ppot -> nsta = LINK_ALARM;
	}
   
    } /* end else (oscn != OFF_SCAN) */

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (ppot -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&ppot -> latm);
	db_post_events ((struct dbCommon *)ppot, (caddr_t)&ppot -> val, DBE_VALUE | DBE_ALARM);
    }

    /* if it is the 1st process pass 
       init last value when alarm chg'd (lalm)
    */

    if (ppot -> fpps)
    {
	ppot -> fpps = 0;
	ppot -> lalm = ppot -> val;
    }

       /* check event list */

    monitor (ppot);

       /* the process is no longer active */

    ppot -> pact = FALSE;

       /* post events for monitors */

    post_events (ppot, &pot_old);

       /* now check for redundancy, see programmers notes */

    update_redun (&pot_old, ppot);

   /* process the forward scan link record */

   if (ppot -> flnk.type == DB_LINK) 
      dbScanPassive ((struct dbAddr *)ppot -> flnk.value.db_link.pdbAddr);

    return (0);

} /* end func process () */



/****************************************************************************
 *
 *  MODULE NAME            : read_custody_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 June 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : ppot - ptr to pot record
 *
 *  DESCRIPTION: read the amt of product transfered and update ppot -> val
 *
****************************************************************************/

static long read_custody_pv (ppot)
   struct potRecord    *ppot;
{
    long     status    = 0;
    long     nRequests = 1;
    long     options   = 0;
    double   vol_xfer;

            /* is there a record for the device */

    if (ppot -> cusp.type != CONSTANT)
    {
	if (ppot -> cusp.type == DB_LINK)
	{
	    status = dbGetLink (&ppot -> cusp.value.db_link, 
                                (struct dbCommon *)ppot, DBR_DOUBLE, 
                                (caddr_t)&vol_xfer, &options, &nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppot, LINK_ALARM, 
							      VALID_ALARM);
		return (status);
	    }
	    else if (!ppot -> ini2)
	    {
		if (!okToUse (&ppot -> cusp) )
		    return (-1);
	    }
            ppot -> val = vol_xfer;
	}
     }

    return (status);

} /* end func read_custody_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : read_ground_ok_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 June 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : ppot - ptr to pot record
 *
 *  DESCRIPTION:  read the ground ok pv, a value of 1 indicates pot is
 *                   properly connected
 *                if not connected assume  ground_ok = 1
 *
****************************************************************************/

static long read_ground_ok_pv (ppot, potVars)
   struct potRecord     *ppot;
   struct pot_vars	*potVars;
{
    long     status    = 0;
    long     nRequests = 1;
    long     options   = 0;
    unsigned short groundok;

    potVars -> ground_ok = 0;

            /* is there a record for the device */

    if (ppot -> gokp.type != CONSTANT)
    {
	if (ppot -> gokp.type == DB_LINK)
	{
	    status = dbGetLink (&ppot -> gokp.value.db_link, 
                                (struct dbCommon *)ppot, DBR_ENUM, 
                                (caddr_t)&groundok, &options, &nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppot, LINK_ALARM, 
                                                              VALID_ALARM);
		return (status);
	    }
	    else if (!ppot -> ini2)
	    {
		if (!okToUse (&ppot -> gokp) )
		    return (-1);
	    }
	    if (groundok == 1)
		potVars -> ground_ok = 1;
	}
    }
    else
	potVars -> ground_ok = 1;

    return (status);

} /* end func read_ground_ok_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : read_sensors 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 June 93
 *
 *  CALLING MODULE(S)      : init_record, process
 *          ARGUMENT(S)    : ppot - ptr to pot record
 *
 *  DESCRIPTION: read various sensors
 *  NOTE: several of the sensor readings are commented out for future
 *          development
 *
****************************************************************************/

static long read_sensors (ppot, potVars)
   struct potRecord	*ppot;
   struct pot_vars	*potVars;
{
    long status = 0;

    if ((status = read_transfer_permit_pv (ppot, potVars)) != 0)
	return (status);

    if ((status = read_ground_ok_pv (ppot, potVars)) != 0)
	return (status);

    if ((status = read_custody_pv (ppot)) != 0)
	return (status);

/*  future
    if (status = read_temperature_pv (ppot))
	return (status);

    if (status = read_pressure_pv (ppot))
	return (status);

    if (status = read_meter_pv (ppot))
	return (status);

    if (status = read_flow_pv (ppot))
	return (status);

    if (status = read_corrected_volume_pv (ppot))
	return (status);
*/
    return (status);

} /* end func read_sensors () */



/****************************************************************************
 *
 *  MODULE NAME            : read_transfer_permit_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 June 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : ppot - ptr to pot record
 *
 *  DESCRIPTION: a value of 1 indicates the POT can issue/receive
 *               if not installed assume transfer_permit_ok = 1
 *
****************************************************************************/

static long read_transfer_permit_pv (ppot, potVars)
    struct potRecord	*ppot;
    struct pot_vars	*potVars;
{
    long     status    = 0;
    long     nRequests = 1;
    long     options   = 0;
    unsigned short ready;

    potVars -> transfer_permit_ok = 0;

            /* is there a record for the device */

    if (ppot -> rdyp.type != CONSTANT)
    {
	if (ppot -> rdyp.type == DB_LINK)
	{
	    status = dbGetLink (&ppot -> rdyp.value.db_link, 
                                (struct dbCommon *)ppot, DBR_ENUM, 
                                (caddr_t)&ready, &options, &nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppot, LINK_ALARM, 
                                                               VALID_ALARM);
		return (status);
	    }
	    else if (!ppot -> ini2)
	    {
		if (!okToUse (&ppot -> rdyp) )
		    return (-1);
	    }
            if (ready == 1)
               potVars -> transfer_permit_ok = 1;
	}
    }
    else
      potVars -> transfer_permit_ok = 1;

    return (status);

} /* end func read_transfer_permit_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : reset_pot_commands 
 *  AUTHOR                 : emsteel
 *  DATE                   : 03 Aug 1993
 *
 *          ARGUMENT(S)    : ppot - ptr to pot record
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION:  deleted db_post_events for commands, not needed
 *             added FirstSpecialDigitalCmd
 *
****************************************************************************/

static void reset_pot_commands (ppot)
   struct potRecord    *ppot;
{
    struct tfcmd   tfcmd;

    tfcmd.NumCmds                = 3;
    tfcmd.FirstDigitalCmd        = &ppot -> ciss;
    tfcmd.FirstSpecialDigitalCmd = &ppot -> issc;
    tfcmd.FirstCmdString         = ppot -> siss;
    tfcmd.CmdString              = ppot -> cmds;

    ResetTFCommand (&tfcmd);

    return;
}


/****************************************************************************
 *
 *  MODULE NAME            : run_commands 
 *  AUTHOR                 : emsteel
 *  DATE                   : 09 June 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ppot - ptr to pot record
 *
 *  DESCRIPTION: process any commands if any
 *
****************************************************************************/

static long run_commands (ppot, potVars)
   struct potRecord	*ppot;
   struct pot_vars	*potVars;
{
   long           status    = 0;
   int            cmd;
   int            cmd_type;
   unsigned short enable;
   struct tfcmd   tfcmd;

   tfcmd.NumCmds                = 3;
   tfcmd.FirstDigitalCmd        = &ppot -> ciss;
   tfcmd.FirstSpecialDigitalCmd = &ppot -> issc;
   tfcmd.FirstCmdString         = ppot -> siss;
   tfcmd.CmdString              = ppot -> cmds;

   cmd = GetTFCommand (&tfcmd, &cmd_type);

   if (cmd != NO_TF_COMMAND)
   {
      if ((ppot -> mode == TF_AUTOMATIC) || (ppot -> mode == TF_MANUAL))
      {
         if (cmd == ILLEGAL_TF_STRING_CMD)
         {
               /* no commands allowed when locked */

            TankFarmPostAlarm ((struct dbCommon *)ppot, CMD_NOT_ALLOWED,
                                                               INFO_ALARM);
         }
         else if ( ((ppot -> mode == TF_AUTOMATIC) &&
                    ((cmd_type == TF_STRING_CMD) ||
                     (cmd_type == TF_SPECIAL_DIGITAL_CMD))) ||
                   ((ppot -> mode == TF_MANUAL) &&
                    (cmd_type == TF_DIGITAL_CMD)) )
         {
            switch (cmd)
            {
               case POT_ISSUE_CMD:
                  if (potVars -> transfer_permit_ok && potVars -> ground_ok)
                  {
                     strncpy (ppot -> stts, ppot -> isss,
                                                     sizeof (ppot -> stts));
                     ppot -> posv = POT_ISSUING;

                     enable = 1;
                     status = enable_transfer (ppot, &enable);
  
                        /* update last command */

                     if (ppot -> siss[0])
                        strncpy (ppot -> pcmd, ppot -> siss, 
					       sizeof (ppot -> pcmd));
                     else
                        strncpy (ppot -> pcmd, TF_UNDEF_CMD, 
					       sizeof (ppot -> pcmd));
                  }
                  else
                     TankFarmPostAlarm ((struct dbCommon *)ppot,
                                              CMD_NOT_ALLOWED, INFO_ALARM);
                  break;

               case POT_RECEIVE_CMD:
                  if (potVars -> transfer_permit_ok && potVars -> ground_ok)
                  {
                     strncpy (ppot -> stts, ppot -> recs,
                                                     sizeof (ppot -> stts));
                     ppot -> posv = POT_RECEIVING;
                     enable = 1;
                     status = enable_transfer (ppot, &enable);

                     if (ppot -> srec[0])
                        strncpy (ppot -> pcmd, ppot -> srec, 
					       sizeof (ppot -> pcmd));
                     else
                        strncpy (ppot -> pcmd, TF_UNDEF_CMD, 
					       sizeof (ppot -> pcmd));
                  }
                  else
                     TankFarmPostAlarm ((struct dbCommon *)ppot,
                                               CMD_NOT_ALLOWED, INFO_ALARM);
                  break;

               case POT_STOP_CMD:
                  strncpy (ppot -> stts, ppot -> stps,
                                                     sizeof (ppot -> stts));
                  ppot -> posv = POT_STOPPED;
                  ppot -> qvol = ppot -> val;
                  enable = 0;
                  status = enable_transfer (ppot, &enable);

                  if (ppot -> sstp[0])
                     strncpy (ppot -> pcmd, ppot -> sstp, 
					    sizeof (ppot -> pcmd));
                  else
                     strncpy (ppot -> pcmd, TF_UNDEF_CMD, 
					    sizeof (ppot -> pcmd));
                  break;

               default:
                  TankFarmPostAlarm ((struct dbCommon *)ppot,
                                              CMD_NOT_ALLOWED, INFO_ALARM);
            }
         }
         else
         {
            TankFarmPostAlarm ((struct dbCommon *)ppot, CMD_NOT_ALLOWED,
                                                               INFO_ALARM);
         }

      }
      else if (ppot -> mode == TF_LOCKED)
      {
            /* no commands allowed when locked */

         TankFarmPostAlarm ((struct dbCommon *)ppot, CMD_NOT_ALLOWED,
                                                               INFO_ALARM);
      }
   }

   reset_pot_commands (ppot);

   return (status);

} /* end func run_commands () */



/****************************************************************************
 *
 *  MODULE NAME            : save_pot 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 June 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ppot - ptr to pot record
 *                           ppot_old - save pot data for redundancy update
 *
 *  DESCRIPTION: save the current pot for later redunadancy check
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 23 Feb 94
 *  REVISION: removed time, mlst, alst, not needed
 *
****************************************************************************/

static void save_pot (ppot, ppot_old)
   struct potRecord    *ppot;
   struct potred       *ppot_old;
{
   ppot_old -> stat = ppot -> stat;
   ppot_old -> sevr = ppot -> sevr;
   ppot_old -> val  = ppot -> val;
   strncpy (ppot_old -> stts, ppot -> stts, sizeof (ppot_old -> stts));
   strncpy (ppot_old -> pcmd, ppot -> pcmd, sizeof (ppot_old -> pcmd));
   strncpy (ppot_old -> cmds, ppot -> cmds, sizeof (ppot_old -> cmds));
   ppot_old -> posv = ppot -> posv;
   ppot_old -> qvol = ppot -> qvol;
   ppot_old -> lalm = ppot -> lalm;
   ppot_old -> esou = ppot -> esou;
   ppot_old -> ini2 = ppot -> ini2;
   ppot_old -> ciss = ppot -> ciss;
   ppot_old -> crec = ppot -> crec;
   ppot_old -> cstp = ppot -> cstp;
   ppot_old -> issc = ppot -> issc;
   ppot_old -> resc = ppot -> resc;
   ppot_old -> stsc = ppot -> stsc;
   ppot_old -> latm = ppot -> latm;
   ppot_old -> mlst = ppot -> mlst;

   return;

} /* end func save_pot () */



/****************************************************************************
 *
 *  MODULE NAME            : special 
 *  AUTHOR                 : emsteel
 *  DATE                   : 09 June 93
 *
 *          ARGUMENT(S)    : paddr - ptr to dbAddr as defined in dbAccess.h
 *                           after - not used
 *
 *  DESCRIPTION: special handling for certain fields
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 06 Jan 94
 *  REVISION: check for !after
 *
****************************************************************************/

static long special (paddr, after)
   struct dbAddr    *paddr;
   int              after;
{
    struct potRecord	*ppot        = (struct potRecord *)(paddr -> precord);
    int			special_type = paddr -> special;
    long		status = 0;

    if (!after)
	return (status);

    switch (special_type)
    {
	case SPC_API_TBL:       /* new API table number */
	    status = FindAPITable (ppot -> apit, &ppot -> papi, (int *)&ppot -> anum);
	    break;

	default:
	    recGblDbaddrError (S_db_badChoice, paddr, "pot: special");
	    status = S_db_badChoice;
    }

    return (status);

} /* end func special () */



/****************************************************************************
 *
 *  MODULE NAME            : update_redun 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 June 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ppot - ptr to current pot record
 *                           ppot_old - ptr to last scans data
 *
 *  DESCRIPTION: update redundant fields if they have changed
 *               it is not neccessary to check for MASTER since we are called
 *               from process and are guaranteed to be MASTER at this point
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 23 Feb 94
 *  REVISION: removed time, mlst, alst, not needed
 *
****************************************************************************/

static void update_redun (ppot_old, ppot)
   struct potred        *ppot_old;
   struct potRecord     *ppot;
{
   char  redbuf[400];

   redQInit ((struct dbCommon *)ppot, redbuf, sizeof (redbuf), type);

   if (ppot_old -> stat != ppot -> stat)
      redQField (redbuf, mylist[STAT]);

   if (ppot_old -> sevr != ppot -> sevr)
      redQField (redbuf, mylist[SEVR]);

   if (ppot_old -> val != ppot -> val)
      redQField (redbuf, mylist[VAL]);

   if (strncmp (ppot_old -> stts, ppot -> stts, 
                                  sizeof (ppot_old -> stts)) != 0)
      redQField (redbuf, mylist[STTS]);

   if (strncmp (ppot_old -> pcmd, ppot -> pcmd, 
                                  sizeof (ppot_old -> pcmd)) != 0)
      redQField (redbuf, mylist[PCMD]);

   if (strncmp (ppot_old -> cmds, ppot -> cmds, 
                                  sizeof (ppot_old -> cmds)) != 0)
      redQField (redbuf, mylist[CMDS]);

   if (ppot_old -> posv != ppot -> posv)
      redQField (redbuf, mylist[POSV]);

   if (ppot_old -> qvol != ppot -> qvol)
      redQField (redbuf, mylist[QVOL]);

   if (ppot_old -> lalm != ppot -> lalm)
      redQField (redbuf, mylist[LALM]);

   if (ppot_old -> esou != ppot -> esou)
      redQField (redbuf, mylist[ESOU]);

   if (ppot_old -> ini2 != ppot -> ini2)
      redQField (redbuf, mylist[INI2]);

   if (ppot_old -> ciss != ppot -> ciss)
      redQField (redbuf, mylist[CISS]);

   if (ppot_old -> crec != ppot -> crec)
      redQField (redbuf, mylist[CREC]);

   if (ppot_old -> cstp != ppot -> cstp)
      redQField (redbuf, mylist[CSTP]);

   if (ppot_old -> issc != ppot -> issc)
      redQField (redbuf, mylist[ISSC]);

   if (ppot_old -> resc != ppot -> resc)
      redQField (redbuf, mylist[RESC]);

   if (ppot_old -> stsc != ppot -> stsc)
      redQField (redbuf, mylist[STSC]);

    if (TsCmpStampsNE (&ppot_old -> latm, &ppot -> latm) )
	redQField (redbuf, mylist [LATM]);

   if (ppot_old -> mlst != ppot -> mlst)
      redQField (redbuf, mylist[MLST]);
	
   redQSend (redbuf);

   return;

} /* end func update_redun () */

