/*recPto.c,v 1.1.1.1 1999/08/03 14:44:20 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/* recPto.c */
  
/* recPto.c - Record Support Routines for Pulse Train Output records */
/*
 *      Original Author: William H. Szostak
 *      Date:            8/24/93
 *
 * Modification Log:
 * -----------------
 * .01  08-24-93        whs     creation of record support for PTO
 * .02  01-03-94	ems	set ack to 0 in init_record
 *				in process, if not init'd do nothing
 *				 comment other checks for init
 * .03  01-05-94	ems	in monitor,check VALID_ALARM before
 *				set ACK
 * .04  03-01-94	ems	added support for process stats
 *				removed ack = 0
 * .05	03-29-94	ems	make forward func def agree with actual func def
 * .06	05-18-94	ems	removed #define LINEAR, defined in cvnTypes.h
 *				rewrite init_record, use HYST properly in alarm,
 *				get_control_double i ffield is hyst then 
 *				high=hopr-lopr and low = 0, removed special, no
 *				special linconv in dev drivers
 * .07	06-13-94	ems	removed 2nd call to tsLocalTiem in process() before
 *				handle ptyp, not needed
 * .08	06-13-94	ems	added support for maximzie severity
 * .09	08-18-94	ems	monitor(), rm DBE_LOG from ack
 *				only archive if tenb (tranding) is enabled
 *				set DBE_VALUE only for oval,rval and rbv
 * .10	10-12-94	ems	add TIME to redundancy, added fpps
 *				(1st process pass), do time stamp when stat or sevr chg,
 *				see monitor(),and not in process(),set lalm to val when 
 *				stat or sevr chg,in process() if fpps (first process)
 *				init lalm to val and if time stamp is 0, init it also
 * .11	10-20-94	ems	chg time to latm (last alarm time) added 
 *				added lptm (last process time)
 * .12	01-13-95	wep	add mlst to redundancy, init_record(), init lptm and latm
 * .13	03-15-95	ems	monitor(), do not set ACK if sevr is EVENT_ALARM
 * .14	04-24-95	ems	complete rewrite
 * .15	06-14-95	ems	removed oval, not needed, in special set oraw = rval
 *				reworked init_record so call neg out even if pos status != 0
 *				to make sure nmsk gets set especially on teh slave
 * .16	09-11-95	ems	monitor(), archive when stat/sevr chg's also
 * .17	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 * .18	02-07-96	ems	remove return from HIHI and LOLO alarms, so will always check 
 *				    all alarm sevrs
 * .19	04-10-96	ems	added code to support alarm delay
 * .20	06-20-96	ems	added code for alarm disable ADIS
 *
 */

#include	<vxWorks.h>
#include	<inetLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<strLib.h>
#include	<wdLib.h>
#include	<stdlib.h>
#include	<logLib.h>
#include	<taskLib.h>
#include	<tickLib.h>

#include	"alarm.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"special.h"
#include	"recSup.h"
#include	"rec/ptoRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"ptoredfldidx.h"
#include	"restable.h"
#include	"cvnTypes.h"
#include	"callbacks.h"
#include	"db_addr.h"
#include	"recGbl.h"
#include	"dbEvent.h"
#include	"okToUse.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/

#define report NULL
#define initialize NULL
static long init_record ();
static long process ();
static long special ();
static long get_value ();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units ();
static long get_precision ();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double ();
static long get_control_double ();
static long get_alarm_double ();

struct rset ptoRSET = {

	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
	};

struct ptodset {			/* pulse train output dset */
	long		number;
	DEVSUPFUN	dev_report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;	/*returns: (0,2)=>(success,success no convert)*/
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	write_pto;	/*(0,1)=>success and */
					/*(continue, don`t continue) */
	DEVSUPFUN	special_linconv;
	};

/* The following defintions are required for redundancy */

static struct fldDes	**mylist = 0;

static char *namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"LALM",
		"INI2",
		"LATM",
		"MLST",
		"RVAL",
		"HHLL",
		"HILO",
		"RTCT",
		"LARF",
		"HLOC",
		0
		};

static unsigned short	type;

/* the following definitions must match those in choiceRec.ascii */

#define COUNT 		0
#define DURATION	1

/* defines for the callback flag CBFL */

#define	PTO_NO_CALLBACK		0
#define	PTO_START_CALLBACK	1
#define	PTO_POS_CHANNEL		2
#define	PTO_NEG_CHANNEL		3
#define	PTO_DURATION_ON		4
#define	PTO_DURATION_OFF	5
#define	PTO_DURATION_MIN_NEG	6
#define	PTO_DURATION_MIN_POS	7

/* defines for the new command flag NCMD */

#define	PTO_NO_CMD	0
#define	PTO_NEW_CMD	1

/* defines for alarm clamp */

#define	DO_HIGH_OUTPUT_CLAMP	1
#define	DO_LOW_OUTPUT_CLAMP	2

static void alarm (struct ptoRecord *, unsigned short);
static int convert (struct ptoRecord *, unsigned short *);
static void monitor ();
static void ptoCallback ();



/*------------------------------------------------------------------------*/

static long init_record (ppto)
    struct	ptoRecord	*ppto;
{
    struct ptodset	*pdset;
    long		status = 0;
    CALLBACK_PTR	pcallback;
    unsigned short	i;
    char		*pstr;
    char		redbuf [512];
    int			ini2_flag = 0;
    int			dont_init = 0;
    int			write_pos = 0;
    int			write_neg = 0;

    if (!(mylist)) 
    {
	for (type = 1; type < 100; type++) 
	{
	    pstr = GET_PRECTYPE (type);
	    if (strncmp (pstr, "pto", 3) == 0) 
		break;
	}
	mylist = redFldAlloc (type, NUMPTOFLDS);

	if (!mylist)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)ppto, 
					"pto, init_record, mylist failed");
	    return (S_rec_Alloc);
	}

	for (i = 0; i < NUMPTOFLDS; i++) 
	    redFldAdd (type, i, namelist [i]);
    }

    if (!(pdset = (struct ptodset *) (ppto->dset))) 
    {
	recGblRecordError (S_dev_noDSET, (struct dbCommon *)ppto, "pto: init_record");
	return (S_dev_noDSET);
    }

    /* must have write_pto function defined */

    if ((pdset->number < 5) || (pdset->write_pto == NULL)) 
    {
	recGblRecordError (S_dev_missingSup, (struct dbCommon *)ppto, "pto: init_record");
	return (S_dev_missingSup);
    }

    if (!(ppto -> dpvt)) 
    {
	pcallback = (CALLBACK_PTR) (calloc (1, sizeof (CALLBACK)));
	if (!pcallback)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)ppto, 
					"pto, init_record, pcallback failed");
	    return (S_rec_Alloc);
	}
	ppto->dpvt = (caddr_t) pcallback;
	pcallback->callback = ptoCallback;
	pcallback->priority = priorityMedium;
	if (dbNameToAddr (ppto->name, &(pcallback->dbAddr))) 
	{
	    logMsg("dbNameToAddr failed in init_record for recPto\n",0,0,0,0,0,0);
	    taskSuspend (taskIdSelf());
	}
    }

    ppto -> init = TRUE;
    ppto -> fpps = 1;
    ppto -> cbfl = PTO_NO_CALLBACK;	/* callbacks are in effect */
    ppto -> ncmd = PTO_NO_CMD;		/* new command for callback */
    ppto -> poso = 0;			/* positive channel on */
    ppto -> nego = 0;			/* negative channel on */
    ppto -> pcnt = 0;			/* callback pulse count */
    ppto -> nsta = NO_ALARM;
    ppto -> nsev = NO_ALARM;
    ppto -> msev = NO_ALARM;
    ppto -> wslv = 0;			/* turn off slave flag */
    ppto -> ldir = 0;
    ppto -> sctk = tickGet();

    /* init alarm delay variables */
    ppto -> larf = 0;
    ppto -> hhll = 0;
    ppto -> hilo = 0;
    ppto -> dvct = 0;
    ppto -> hloc = 0;
    ppto -> rtct = 0;

    /* init last process time and last alarm time */

    if (ppto -> lptm.secPastEpoch == 0)
	tsLocalTime (&ppto -> lptm);
    if (ppto -> latm.secPastEpoch == 0)
	tsLocalTime (&ppto -> latm);

    /* get the initial value if dol is a constant*/

    if (ppto->dol.type == CONSTANT ) 
    {
	ppto->udf = FALSE;
	ppto->val = ppto->dol.value.value;
    }

    if (pdset->init_record) 
    {
	status = (*pdset->init_record) (ppto, 0);
	if (!status)
	{
	    /* if pos channel is on, set up to turn off
	       below when we can init
	    */
	    if (ppto -> poiv)
	    {
		if (!ppto -> rbv)
		    write_pos = 1;
	    }
	    else
	    {
		if (ppto -> rbv)
		    write_pos = 1;
	    }
	}
	else
	    dont_init = 1;
	if (ppto->nbit == 2) 
	{
	    status = (*pdset->init_record) (ppto, 1);
	    if (!status)
	    {
		/* if the neg channel is on, set up
		   to turn off below when we can init
		*/
		if (ppto -> noiv)
		{
		    if (!ppto -> rbv)
		        write_neg = 1;
		}
		else
		{
		    if (ppto -> rbv)
			write_neg = 1;
		}
	    }
	    else
		dont_init = 1;
	}
	if (!dont_init)
	{
	    if ( (ppto -> dol.type == DB_LINK) &&
		 (ppto -> omsl == CLOSED_LOOP) )
	    {
		if (!okToUse (&ppto -> dol) )
		    dont_init = 1;
	    }
	}

	if (dont_init)
	{
	    ppto -> nsta = NO_INIT_ALARM;
	    ppto -> nsev = VALID_ALARM;
	}
	else
	    ini2_flag = 1;
    }
    else
	ini2_flag = 1;

    if (resGetCaAnswer(ppto->radr, &i) == RES_OK) 
    {
	if (i) 
	{
	    /* check for maximize severity */

	    if (ppto -> msev > ppto -> nsev)
	    {
		ppto -> nsev = ppto -> msev;
		ppto -> nsta = LINK_ALARM;
	    }

	    monitor (ppto);

	    if (ini2_flag)
	    {
		if (write_pos)
		{
		    if (ppto -> poiv)
			ppto -> rbv = ppto -> pmsk;
		    else
			ppto -> rbv = 0;
		    (*pdset -> write_pto)(ppto, 0);
		}
		if (write_neg)
		{
		    if (ppto -> noiv)
			ppto -> rbv = ppto -> nmsk;
		    else
			ppto -> rbv = 0;
		    (*pdset -> write_pto)(ppto, 1);
		}
		ppto -> ini2 = 1;

		redQInit ((struct dbCommon *)ppto, redbuf, sizeof (redbuf), type);
		redQField (redbuf, mylist [STAT]);
		redQField (redbuf, mylist [SEVR]);
		redQField (redbuf, mylist [VAL]);
		redQField (redbuf, mylist [LALM]);
		redQField (redbuf, mylist [RVAL]);
		redQField (redbuf, mylist [INI2]);

		redQSend (redbuf);
	    }
	}
    }
    return(0);
}


/*------------------------------------------------------------------------*/

static void alarm (ppto, clamp_mask)
    struct ptoRecord	*ppto;
    unsigned short	clamp_mask;
{
    double		val = ppto -> val;
    int			adis = 0;
    ALARM_LIMITS	alarm_limits;

    /* check for alarm inhibit */

    adis = AlarmDisableLink ((void *)ppto);
    if ( (ppto->aenb == ALARM_DISABLED) || (adis != 0) )
    {
	if (ppto -> nsta != NO_INIT_ALARM)
	{
	    ppto->nsta = INHIBIT_ALARM;
	    ppto->nsev = NO_ALARM;
	}
	return;
    }

    if(ppto->udf == TRUE) 
    {
	if (ppto->nsev < VALID_ALARM) 
	{
	    ppto->nsta = UDF_ALARM;
	    ppto->nsev = VALID_ALARM;
	}
	return;
    }

    /* check for HIGH output clamp alarm */

    if (ppto -> chsv > NO_ALARM)
    {
	if (clamp_mask & DO_HIGH_OUTPUT_CLAMP)
	{
	    if (ppto -> aldl > 0)
	    {
		/* if we were counting low output clamp, 
		   zero counter since same counter used for
		   HIGH and LOW clamp alarm
		*/
		if (ppto -> larf & LARF_OUTPUT_LOW)
		    ppto -> hloc = 0;
		if (ppto -> hloc < ALDL_MAX)
		    ppto -> hloc++;
		ppto -> larf |= LARF_OUTPUT_HIGH;
		if ( (ppto -> hloc >= ppto ->aldl) &&
		     (ppto -> nsev < ppto -> chsv) )
		{
		    ppto -> nsta = HIGH_OUTPUT_CLAMP;
		    ppto -> nsev = ppto -> chsv;
		}
	    }
	    else
	    {
		if (ppto -> nsev < ppto -> chsv)
		{
		    ppto -> nsta = HIGH_OUTPUT_CLAMP;
		    ppto -> nsev = ppto -> chsv;
		}
	    }
	}
	else
	{
	    if (ppto -> aldl > 0)
	    {
		ppto -> larf &= ~LARF_OUTPUT_HIGH;
		if (!(ppto -> larf & ((unsigned long)LARF_OUTPUT_LOW)))
		    ppto -> hloc = 0;
	    }
	}
    }
    else
    {
	if (ppto -> aldl > 0)
	{
	    ppto -> larf &= ~LARF_OUTPUT_HIGH;
	    if (!(ppto -> larf & ((unsigned long)LARF_OUTPUT_LOW)))
		ppto -> hloc = 0;
	}
    }

    /* check for LOW output clamp alarm */

    if (ppto -> clsv > NO_ALARM)
    {
	if (clamp_mask & DO_LOW_OUTPUT_CLAMP)
	{
	    if (ppto -> aldl > 0)
	    {
		if (ppto -> larf & LARF_OUTPUT_HIGH)
		    ppto -> hloc = 0;
		if (ppto -> hloc < ALDL_MAX)
		    ppto -> hloc++;
		ppto -> larf |= LARF_OUTPUT_LOW;
		if ( (ppto -> hloc >= ppto ->aldl) &&
		     (ppto -> nsev < ppto -> clsv) )
		{
		    ppto -> nsta = LOW_OUTPUT_CLAMP;
		    ppto -> nsev = ppto -> clsv;
		}
	    }
	    else
	    {
		if (ppto -> nsev < ppto -> clsv)
		{
		    ppto -> nsta = LOW_OUTPUT_CLAMP;
		    ppto -> nsev = ppto -> clsv;
		}
	    }
	}
	else
	{
	    if (ppto -> aldl > 0)
	    {
		ppto -> larf &= ~LARF_OUTPUT_LOW;
		if (!(ppto -> larf & ((unsigned long)LARF_OUTPUT_HIGH)))
		    ppto -> hloc = 0;
	    }
	}
    }
    else
    {
	if (ppto -> aldl > 0)
	{
	    ppto -> larf &= ~LARF_OUTPUT_LOW;
	    if (!(ppto -> larf & ((unsigned long)LARF_OUTPUT_HIGH)))
		ppto -> hloc = 0;
	}
    }

    /* set up for limit alarms */

    alarm_limits.hihi = ppto -> hihi;
    alarm_limits.high = ppto -> high;
    alarm_limits.low  = ppto -> low;
    alarm_limits.lolo = ppto -> lolo;
    alarm_limits.devl = 0;
    alarm_limits.devs = 0;
    alarm_limits.hyst = ppto -> hyst;
    alarm_limits.hhsv = ppto -> hhsv;
    alarm_limits.hsv  = ppto -> hsv;
    alarm_limits.lsv  = ppto -> lsv;
    alarm_limits.llsv = ppto -> llsv;
    alarm_limits.dvsv = NO_ALARM;

    AlarmLimits (&alarm_limits, val, (void *)ppto);

    /* do a return to normal, check for alarm delay ALDL only
       on the listed alarms
    */

    if ( ((ppto -> nsta == NO_ALARM) && (ppto -> nsev == NO_ALARM)) &&
	 ((ppto -> stat == HIHI_ALARM) || 
	  (ppto -> stat == HIGH_ALARM) ||
	  (ppto -> stat == LOW_ALARM)  || 
	  (ppto -> stat == LOLO_ALARM) ||
	  (ppto -> stat == HIGH_OUTPUT_CLAMP) || 
	  (ppto -> stat == LOW_OUTPUT_CLAMP)) )
    {
	if (ppto -> aldl > 0)
	{
	    if (ppto -> rtct < ALDL_MAX)
		ppto -> rtct++;

	    /* delay return to normal so, continue last alarm */
	    if (ppto -> rtct < ppto -> aldl)
	    {
		ppto -> nsta = ppto -> stat;
		ppto -> nsev = ppto -> sevr;
	    }
	}
    }
    else
    {
	if (ppto -> aldl > 0)
	{
	    ppto -> rtct = 0;
	}
    }

    return;
}


/*------------------------------------------------------------------------*/

static int convert (ppto, clamp_mask)
    struct ptoRecord	*ppto;
    unsigned short	*clamp_mask;
{
    double	value;
    long	nRequest = 1;
    long	options = 0;
    short	save_pact = ppto -> pact;
    long	status = 0;

    if ((ppto->dol.type == DB_LINK) && (ppto->omsl == CLOSED_LOOP))
    {
	ppto->pact = TRUE;

	/* don't allow dbputs to val field */

	ppto->val = ppto->pval;
	status = dbGetLink(&ppto->dol.value.db_link, 
			   (struct dbCommon *)ppto, 
			   DBR_DOUBLE, (caddr_t)&value, 
			   &options, &nRequest);
	ppto->pact = save_pact;
	if (status) 
	{
	    if (ppto->nsev < VALID_ALARM) 
	    {
		ppto->nsta = LINK_ALARM;
		ppto->nsev = VALID_ALARM;
	    }
	    return (1);
	}
    } 
    else 
    {
	value = ppto->val;
    }

    /* nothing to do if value < 0 but only can write
       to the POS output
    */
    if ( (value < 0) && (ppto -> nbit == 1) )
	value = 0;

    /* check drive limits */

    if (ppto->clmh > ppto->clml) 
    {
	if ( (value > ppto->clmh) && ppto -> clmh)
	{
	    value = ppto->clmh;
	    if (ppto -> chsv > ppto -> nsev)
	    {
		*clamp_mask |= DO_HIGH_OUTPUT_CLAMP;
	    }
	}
	else if ( (value < ppto->clml) && ppto -> clml)
	{
	    value = ppto->clml;
	    if (ppto -> clsv > ppto -> nsev)
	    {
		*clamp_mask |= DO_LOW_OUTPUT_CLAMP;
	    }
	}
    }
    ppto->val = value;
    ppto->udf = FALSE;

    if (ppto->sclf != 0.0) 
	value /= ppto->sclf;

    /* if pulse type is COUNT, that is input value is a pulse count
       rval is defined as a long, so we need to round because of truncation
       if >= 0.0, round up using 0.5,
       else round up using 0.499999, because we want the following
       -2.5 -> -1.5****1 has to round to -2
    */
    if (ppto -> ptyp == COUNT)
    {
	if (value >= 0.0)
	    ppto -> rval = (int)(value + 0.5);
	else
	    ppto -> rval = (int)(value - 0.499999);
    }
    else
    {
	ppto -> rval = (int)(value * vxTicksPerSecond);
    }
        
    return (0);
}


/*------------------------------------------------------------------------*/

static long get_value (ppto,pvdes)
    struct ptoRecord		*ppto;
    struct valueDes		*pvdes;
{
    pvdes->field_type = DBF_DOUBLE;
    pvdes->no_elements=1;
    (double *)(pvdes->pvalue) = &ppto->val;
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_units (paddr,units)
    struct dbAddr *paddr;
    char	  *units;
{
    struct ptoRecord	*ppto = (struct ptoRecord *)paddr->precord;

    strncpy(units,ppto->egu,sizeof(ppto->egu));
    return (0);
}

/*------------------------------------------------------------------------*/

static long get_precision (paddr, precision)
    struct dbAddr *paddr;
    long	  *precision;
{
    struct ptoRecord	*ppto = (struct ptoRecord *)paddr->precord;

    *precision = ppto->prec;
    return (0);
}


/*------------------------------------------------------------------------*/

static long get_graphic_double (paddr, pgd)
    struct dbAddr *paddr;
    struct dbr_grDouble	*pgd;
{
    struct ptoRecord	*ppto=(struct ptoRecord *)paddr->precord;

    pgd->upper_disp_limit = ppto->hopr;
    pgd->lower_disp_limit = ppto->lopr;
    return (0);
}


/*------------------------------------------------------------------------*/

static long get_control_double (paddr, pcd)
    struct dbAddr *paddr;
    struct dbr_ctrlDouble *pcd;
{
    struct ptoRecord	*ppto = (struct ptoRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct ptoRecord *)0) -> hyst) ):
	{
	    float    hyst_span;

	    hyst_span = ppto -> hopr - ppto -> lopr;
	    if (hyst_span < 0.0)
	        hyst_span = -hyst_span;
	    pcd -> upper_ctrl_limit = hyst_span;
	    pcd -> lower_ctrl_limit = 0;
	    break;
	}

	default:
	    pcd -> upper_ctrl_limit = ppto -> hopr;
	    pcd -> lower_ctrl_limit = ppto -> lopr;
	    break;
    }

    return (0);
}

/*------------------------------------------------------------------------*/

static long get_alarm_double (paddr, pad)
    struct dbAddr *paddr;
    struct dbr_alDouble	*pad;
{
    struct ptoRecord	*ppto =( struct ptoRecord *) paddr->precord;

    pad->upper_alarm_limit = ppto->hihi;
    pad->upper_warning_limit = ppto->high;
    pad->lower_warning_limit = ppto->low;
    pad->lower_alarm_limit = ppto->lolo;
    return (0);
}


/*------------------------------------------------------------------------*/

static void monitor (ppto)
    struct ptoRecord	*ppto;
{
    unsigned short	monitor_mask;
    double		delta;
    short		stat,sevr,nsta,nsev;
    int			set_ack = 1;

    /* get previous stat and sevr and new stat and sevr*/

    stat = ppto->stat;
    sevr = ppto->sevr;
    nsta = ppto->nsta;
    nsev = ppto->nsev;

    /*set current stat and sevr*/

    ppto->stat = nsta;
    ppto->sevr = nsev;
    ppto->nsta = 0;
    ppto->nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == ppto -> psta) && (nsev == ppto -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and out previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	ppto -> psta = stat;
	ppto -> psev = sevr;
	db_post_events ((struct dbCommon *)ppto, (caddr_t)&ppto -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)ppto, (caddr_t)&ppto -> psev, DBE_VALUE);
    }

    monitor_mask = 0;

    /* alarm condition changed this scan */

    if ((stat != nsta) || (sevr != nsev)) 
    {
	/* post events for alarm condition change*/

	monitor_mask = DBE_ALARM;

	/* when alarm status chg's, get current time and
	   save VAL in lalm (last value when alarm chg'd)
	*/
	tsLocalTime (&ppto -> latm);
	ppto -> lalm = ppto -> val;

	/* post stat and nsev fields */

	db_post_events ((struct dbCommon *)ppto, (caddr_t)&ppto->stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)ppto, (caddr_t)&ppto->sevr, DBE_VALUE);
	if ( (ppto->sevr != NO_ALARM) && (set_ack) &&
	     (ppto->sevr != EVENT_ALARM) )
	{
	    if (ppto->ack == 0) 
	    {
		ppto->ack = 1;           /* new unacknowledged alarm */
		db_post_events ((struct dbCommon *)ppto, (caddr_t)&ppto->ack, DBE_VALUE);
	    }
	}
    }

    /* check for value change */

    delta = ppto->mlst - ppto->val;
    if (delta < 0.0) 
	delta = -delta;
	
    if (delta > ppto->mdel) 
    {
	/* post events for value change */

	monitor_mask |= DBE_VALUE;

	/* update last value monitored */

	ppto->mlst = ppto->val;
    }

    /* check for archive change if tenb (trending) is enabled */

    if (ppto -> tenb)
    {
	delta = ppto->alst - ppto->val;
	if (delta < 0.0) 
	    delta = -delta;
	
	/* if exceed archive deadband or there has been a change in alarm status */
	if ( (delta > ppto->adel) || (monitor_mask & DBE_ALARM) )
	{
	    /* post events on value field for archive change */

	    monitor_mask |= DBE_LOG;

	    /* update last archive value monitored */

	    ppto->alst = ppto->val;
	}
    }

    /* send out monitors connected to the value field */

    if (monitor_mask)
	db_post_events ((struct dbCommon *)ppto, (caddr_t)&ppto->val, monitor_mask);
	
    if (ppto->oraw != ppto->rval) 
    {
	db_post_events ((struct dbCommon *)ppto, (caddr_t)&ppto->rval, DBE_VALUE);
	ppto->oraw = ppto->rval;
    }

    return;
}


/*------------------------------------------------------------------------*/
/*
NOTES:
	every time we set ncdm to PTO_NEW_CMD, we want to send RVAL to the
	slave, since it has a special attached to it, it may or may not 
	cause a callback to start on the slave
	if we just transitioned from SLAVE to MASTER, we want to get the 
	outputs in the proper state
*/

static long process (paddr)
    struct dbAddr	*paddr;
{
    struct ptoRecord	*ppto = (struct ptoRecord *) (paddr -> precord);
    struct ptodset	*pdset = (struct ptodset *) (ppto -> dset);
    CALLBACK_PTR	pcallback = (CALLBACK_PTR) (ppto -> dpvt);
    int			wait_time = 0;
    long		status = 0;
    long		temp;
    unsigned short	stat = ppto -> stat;
    unsigned short	sevr = ppto -> sevr;
    double		val  = ppto -> val;
    double		lalm = ppto -> lalm;
    TS_STAMP		latm = ppto -> latm;
    double		mlst = ppto -> mlst;
    long		rval = ppto -> rval;
    short		hhll = ppto -> hhll;
    short		hilo = ppto -> hilo;
    short		rtct = ppto -> rtct;
    unsigned long	larf = ppto -> larf;
    short		hloc = ppto -> hloc;
    unsigned short	clamp_mask;
    int			send_rval = 0;
    char		redbuf[512];

    /* if not init'd do nothing */

    if ( (!ppto -> init) || (!ppto -> ini2) )
	return (status);

    if( (pdset == NULL) || (pdset->write_pto == NULL) ) 
    {
	ppto->pact = TRUE;
	recGblRecordError(S_dev_missingSup, (struct dbCommon *)ppto, "write_pto");
	return (S_dev_missingSup);
    }

    /* get time stamp for last time record was processed */

    tsLocalTime (&ppto -> lptm);

    clamp_mask = 0;

    ppto-> msev = NO_ALARM;

    if (ppto->tagd) 
    {
	ppto->nsta = TAGGED_ALARM;
	ppto->nsev = VALID_ALARM;
	ppto->val = ppto->pval;
    } 
    else if (ppto->oscn == OFF_SCAN) 
    {
	ppto->nsta = SCAN_ALARM;
	ppto->nsev = VALID_ALARM;
	ppto->val = ppto->pval;
    } 
    else 
    {
	ppto->nsta = NO_ALARM;
	ppto->nsev = NO_ALARM;

	ppto -> pact = TRUE;

	/* update number of process times */

	if (process_stats)
	    process_cnts[type]++;	

	/* read VAL compute RVAL do NOT compute RBV */
	if (convert (ppto, &clamp_mask) == 0)
	{
	    /* if RVAL or there is a callback in effect */
	    if (ppto -> rval || ppto -> cbfl)
	    {
		if (ppto -> ptyp == COUNT)
		{
		    temp = ppto -> rval;
		    if (temp < 0)
			temp = -temp;
		    /* do we at least have the minimum number of pulses */
		    if (temp >= ppto -> mnpw)
		    {
			ppto -> ncmd = PTO_NEW_CMD;	
			send_rval = 1;
		    }
		    else
			ppto -> rval = 0;
		
		    /* even though RVAL is 0, this is significant
		       if a callback is in effect
		    */
		    if ( (ppto -> rval == 0) && (ppto -> cbfl) )
		    {
			ppto -> ncmd = PTO_NEW_CMD;
			send_rval = 1;
		    }
		}
		/* are DURATION */
		else
		{
		    temp = ppto -> rval;
		    if (temp < 0)
			temp = -temp;
		    /* do we at least have the minimum pulse duration
		       RVAL is already in ticks see convert()
		    */
		    if (temp >= ((int)(ppto -> pont * vxTicksPerSecond)))
		    {
			ppto -> ncmd = PTO_NEW_CMD;
			send_rval = 1;
		    }
		    else
			ppto -> rval = 0;
		    /* if we are in a DURATION, but not in the minimum off time
		       we want to cancel the callback and start a new on
		    */
		    if (ppto -> cbfl == PTO_DURATION_ON)
		    {
			wdCancel (pcallback -> wd_id);
			ppto -> cbfl = PTO_NO_CALLBACK;
			ppto -> ncmd = PTO_NEW_CMD;
			send_rval = 1;
		    }
		}
	    }
	} /* end convert () */

	/* if we were slave but now are master */
	if (ppto -> wslv)
	{
	    ppto -> wslv = 0;
	    if (ppto -> ptyp == COUNT)
	    {
		if (ppto -> rval == 0)
		{
		    if (ppto -> poiv)
			ppto -> rbv = ppto -> pmsk;
		    else
			ppto -> rbv = 0;
		    (*pdset -> write_pto)(ppto, 0);
		    if (ppto -> nbit == 2)
		    {
			if (ppto -> noiv)
			    ppto -> rbv = ppto -> nmsk;
			else
			    ppto -> rbv = 0;
			(*pdset -> write_pto)(ppto, 1);
		    }
		}
		else if (ppto -> rval > 0)
		{
		    /* make sure neg output channel is off */
		    if (ppto -> nbit == 2)
		    {
			if (ppto -> noiv)
			    ppto -> rbv = ppto -> nmsk;
			else
			    ppto -> rbv = 0;
			(*pdset -> write_pto)(ppto, 1);
		    }
		}
		/* else < 0 */
		else
		{
		    /* make sure pos channel is off */
		    if (ppto -> poiv)
			ppto -> rbv = ppto -> pmsk;
		    else
			ppto -> rbv = 0;
		    (*pdset -> write_pto)(ppto, 0);
		}
	    }
	    /* we are DURATION */
	    else
	    {
		if (ppto -> rval == 0)
		{
		    if (ppto -> poiv)
			ppto -> rbv = ppto -> pmsk;
		    else
			ppto -> rbv = 0;
		    (*pdset -> write_pto)(ppto, 0);
		    if (ppto -> nbit == 2)
		    {
			if (ppto -> noiv)
			    ppto -> rbv = ppto -> nmsk;
			else
			    ppto -> rbv = 0;
			(*pdset -> write_pto)(ppto, 1);
		    }
		}
		else if (ppto -> rval > 0)
		{
		    /* make sure neg output channel is off */
		    if (ppto -> nbit == 2)
		    {
			if (ppto -> noiv)
			    ppto -> rbv = ppto -> nmsk;
			else
			    ppto -> rbv = 0;
			(*pdset -> write_pto)(ppto, 1);
		    }
		    /* given this we have to turn on the pos output chan */
		    if (ppto -> ncmd && (ppto -> cbfl == PTO_NO_CALLBACK) &&
			ppto -> poso)
		    {
			if (ppto -> poiv)
			    ppto -> rbv = 0;
			else
			    ppto -> rbv = ppto -> pmsk;
			(*pdset -> write_pto)(ppto, 0);
		    }
		}
		/* else < 0 */
		else
		{
		    /* make sure pos channel is off */
		    if (ppto -> poiv)
			ppto -> rbv = ppto -> pmsk;
		    else
			ppto -> rbv = 0;
		    (*pdset -> write_pto)(ppto, 0);
		    if (ppto -> nbit == 2)
		    {
			/* given this make sure neg outptu chan is on */
			if (ppto -> ncmd && (ppto -> cbfl == PTO_NO_CALLBACK) &&
			    ppto -> nego)
			{
			    if (ppto -> noiv)
				ppto -> rbv = 0;
			    else
				ppto -> rbv = ppto -> nmsk;
			    (*pdset -> write_pto)(ppto, 1);
			}
		    }
		}
	    }
	}

	/* if we have someting to do */
	if (ppto -> ncmd)
	{
	    /* if a callback is not already in effect, get on going */
	    if (ppto -> cbfl == PTO_NO_CALLBACK)
	    {
		wait_time = 1;
		ppto -> cbfl = PTO_START_CALLBACK;
		pcallback = (CALLBACK_PTR)(ppto -> dpvt);
		if (pcallback -> wd_id == NULL)
		{
		    pcallback -> wd_id = wdCreate ();
		}
		wdStart (pcallback -> wd_id, wait_time, (FUNCPTR)callbackRequest, (int)pcallback);
	    }
	}

	/* check for alarms */

	alarm (ppto, clamp_mask);

	/* check for maximize severity */

	if (ppto -> msev > ppto -> nsev)
	{
	    ppto -> nsev = ppto -> msev;
	    ppto -> nsta = LINK_ALARM;
	}

    } /* end on scan */

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (ppto -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&ppto -> latm);
	db_post_events ((struct dbCommon *)ppto, (caddr_t)&ppto -> val, DBE_VALUE | DBE_ALARM);
    }

    /* if it is the 1st process pass 
       init last value when alarm chg'd (lalm)
    */

    if (ppto -> fpps)
    {
	ppto -> fpps = 0;
	ppto -> lalm = ppto -> val;
    }

    /* check event list */

    monitor (ppto);

    /* update redundant fields that have changed */

    redQInit ((struct dbCommon *)ppto, redbuf, sizeof (redbuf), type);

    if (stat != ppto->stat) 
	redQField (redbuf, mylist [STAT]);
	
    if (sevr != ppto->sevr) 
	redQField (redbuf, mylist [SEVR]);
	
    if (val != ppto->val) 
	redQField (redbuf, mylist [VAL]);
	
    if (lalm != ppto->lalm) 
	redQField (redbuf, mylist [LALM]);
	
    if (TsCmpStampsNE (&latm, &ppto -> latm) )
	redQField (redbuf, mylist [LATM]);
	
    if (mlst != ppto -> mlst)
	redQField (redbuf, mylist [MLST]);

    if ((send_rval) || (rval != ppto -> rval))
	redQField (redbuf, mylist [RVAL]);

    if (hhll != ppto -> hhll)
	redQField (redbuf, mylist [HHLL]);

    if (hilo != ppto -> hilo)
	redQField (redbuf, mylist [HILO]);

    if (rtct != ppto -> rtct)
	redQField (redbuf, mylist [RTCT]);

    if (larf != ppto -> larf)
	redQField (redbuf, mylist [LARF]);

    if (hloc != ppto -> hloc)
	redQField (redbuf, mylist [HLOC]);
	
    redQSend (redbuf);

    /* process the forward scan link record */

    if (ppto->flnk.type == DB_LINK)  
	dbScanPassive ((struct dbAddr *)ppto->flnk.value.db_link.pdbAddr);

    ppto->pact = FALSE;

    return (status);
}


/*------------------------------------------------------------------------*/
/*
NOTE:	Cycle definition for PTO
	POIV/NOIV are 0 (not invert signal)
      
	COUNT TYPE                                DURATION TYPE
        ------     ------     ------              -------------------
        |    |     |    |     |    |              |                 |
        |    |     |    |     |    |              |                 |
        |    |     |    |     |    |              |                 |
    ----|    |-----|    |-----|    |------    ----|                 |------
       start      end


	POIV/NOIV are 1 (invert signal)

	COUNT TYPE                                DURATION TYPE
       start      end
    ----|    |-----|    |-----|    |------    ----|                 |------
	|    |     |    |     |    |              |                 |
	|    |     |    |     |    |              |                 |
	|----|     |----|     |----|              |-----------------|

	truth table for RBV and POIV/NOIV
		RBV	POIV/NOIV
		0	0	at end of cycle
		1	0	at mid cycle (on for DURATION)
		0	1	at mid cycle (on for DURATION)
		1	1	at end of cycle
*/

static void ptoCallback (pcallback)
    CALLBACK_PTR	pcallback;
{
    struct ptoRecord	*ppto = (struct ptoRecord *) (pcallback->dbAddr.precord);
    struct ptodset	*pdset = (struct ptodset *) (ppto -> dset);
    int			wait_time = 0;
    long		status = 0;
    long		temp;
    float		dt;		/* delta time in seconds */
    unsigned short	master = 0;
    int			wslv = 0;

    /* we must find out if we are SLAVE, so that we do not do 
       any write_pto(), by setting wslv, when the SLAVE becomes
       MASTER, it will allow us to do some cleanup in process
    */
    if (resGetCaAnswer (ppto -> radr, &master) == RES_OK)
    {
	if (!master)
	{
	    ppto -> wslv = 1;
	    wslv = 1;
	}
	else
	    wslv = 0;
    }

    switch (ppto -> cbfl)
    {
	case PTO_START_CALLBACK:
	    if (ppto -> ptyp == COUNT)
	    {
		if (ppto -> rval > 0)
		{
		    if (ppto -> poiv)
			ppto -> rbv = 0;
		    else
			ppto -> rbv = ppto -> pmsk;
		    if (!wslv)
		    	status = (*pdset -> write_pto)(ppto, 0);
		    if (!status)
		    {
			ppto -> cbfl = PTO_POS_CHANNEL;
			ppto -> poso = 1;
			ppto -> ldir = 1;
		    }
		}
		else
		{
		    if (ppto -> noiv)
			ppto -> rbv = 0;
		    else
			ppto -> rbv = ppto -> nmsk;
		    if (!wslv)
		    	status = (*pdset -> write_pto)(ppto, 1);
		    if (!status)
		    {
			ppto -> cbfl = PTO_NEG_CHANNEL;
			ppto -> nego = 1;
			ppto -> ldir = -1;
		    }
		}
		if (!status)
		{
		    temp = ppto -> rval;
		    if (temp < 0)
			temp = -temp;
		    ppto -> pcnt = temp;
		    wait_time = (int)(ppto -> pont * vxTicksPerSecond);
		}
		else
		{
		    ppto -> cbfl = PTO_NO_CALLBACK;
		    ppto -> poso = ppto -> nego = 0;
		}
		ppto -> ncmd = PTO_NO_CMD;
	    }
	    /* DURATION */
	    else
	    {
		/* we were going in the positiove direction */
		if (ppto -> poso)
		{
		    if (ppto -> ncmd)
		    {
			/* the callback was cancelled in process(),
			   we have to turn off/on the output to finish
			   the duration and institute the minimum off time
			   if it exists
			*/
			if (ppto -> rval == 0)
			{
			    if (ppto -> poiv)
				ppto -> rbv = ppto -> pmsk;
			    else
				ppto -> rbv = 0;
			    ppto -> sctk = tickGet();
			    if (!wslv)
			    	status = (*pdset -> write_pto)(ppto, 0);
			    if (!status)
			    {
				ppto -> poso = 0;
				if (ppto -> poft)
				{
				    wait_time = (int)(ppto -> poft * vxTicksPerSecond);
				    ppto -> cbfl = PTO_DURATION_OFF;
				}
				else
				{
				    ppto -> cbfl = PTO_NO_CALLBACK;
				    ppto -> poso = 0;
				}
			    }
			    else
			    {
				ppto -> cbfl = PTO_NO_CALLBACK;
				ppto -> poso = 0;
			    }
			}
			/* we are already going in the pos direction, continue
			   that way
			*/
			else if (ppto -> rval > 0)
			{
			    ppto -> poso = 1;
			    ppto -> ldir = 1;
			    ppto -> cbfl = PTO_DURATION_ON;
			    temp = ppto -> rval;
			    if (temp < 0)
				temp = -temp;
			    wait_time = (int)temp;
			}
			/* rval < 0, we have to change direction */
			else
			{
			    if (ppto -> poiv)
				ppto -> rbv = ppto -> pmsk;
			    else
				ppto -> rbv = 0;
			    ppto -> sctk = tickGet();
			    ppto -> poso = 0;
			    /* write the the off condition */
			    if (!wslv)
			    	status = (*pdset -> write_pto)(ppto, 0);
			    if (!status)
			    {
				/* since we are reversing direction, do we have
				   a minimum off time on direction transition
				*/
				if (ppto -> mnpw)
				{
				    wait_time = (int)(ppto -> mnpw * vxTicksPerSecond);
				    ppto -> cbfl = PTO_DURATION_MIN_NEG;
				}
				/* no but do we have have a minimum off time */
				else if (ppto -> poft)
				{
				    wait_time = (int)(ppto -> poft * vxTicksPerSecond);
				    ppto -> cbfl = PTO_DURATION_MIN_NEG;
				}
				/* we got nuttin, so just reverse it */
				else
				{
				    if (ppto -> noiv)
					ppto -> rbv = 0;
				    else
					ppto -> rbv = ppto -> nmsk;
			    	    if (!wslv)
				    	status = (*pdset -> write_pto)(ppto, 1);
				    if (!status)
				    {
					ppto -> nego = 1;
					ppto -> ldir = -1;
					temp = ppto -> rval;
					if (temp < 0)
					    temp = -temp;
					wait_time = (int)temp;
					ppto -> cbfl = PTO_DURATION_ON;
				    }
				    else
				    {
					ppto -> nego = 0;
					ppto -> cbfl = PTO_NO_CALLBACK;
				    }
				}
			    }
			    else
			    {
				ppto -> poso = ppto -> nego = 0;
				ppto -> cbfl = PTO_NO_CALLBACK;
			    }
			}
			ppto -> ncmd = PTO_NO_CMD;
		    }
		    /* nothing to do , this is an error, should not be here */
		    else
		    {
			ppto -> ncmd = PTO_NO_CMD;
		    }
		}
		/* we were going in the negative direction */
		else if (ppto -> nego)
		{
		    if (ppto -> ncmd)
		    {
			/* the callback was cancelled in process(),
			   we have to turn off/on the output to finish
			   the duration and institute the minimum off time
			   if it exists
			*/
			if (ppto -> rval == 0)
			{
			    if (ppto -> noiv)
				ppto -> rbv = ppto -> nmsk;
			    else
				ppto -> rbv = 0;
			    ppto -> sctk = tickGet();
			    if (!wslv)
			    	status = (*pdset -> write_pto)(ppto, 1);
			    if (!status)
			    {
				ppto -> nego = 0;
				if (ppto -> poft)
				{
				    wait_time = (int)(ppto -> poft * vxTicksPerSecond);
				    ppto -> cbfl = PTO_DURATION_OFF;
				}
				else
				{
				    ppto -> cbfl = PTO_NO_CALLBACK;
				    ppto -> nego = 0;
				}
			    }
			    else
			    {
				ppto -> cbfl = PTO_NO_CALLBACK;
				ppto -> nego = 0;
			    }
			}
			/* we are already going in the neg direction, continue
			   that way
			*/
			else if (ppto -> rval < 0)
			{
			    ppto -> nego = 1;
			    ppto -> ldir = -1;
			    ppto -> cbfl = PTO_DURATION_ON;
			    temp = ppto -> rval;
			    if (temp < 0)
				temp = -temp;
			    wait_time = (int)temp;
			}
			/* rval > 0, we have to change direction */
			else
			{
			    if (ppto -> noiv)
				ppto -> rbv = ppto -> nmsk;
			    else
				ppto -> rbv = 0;
			    ppto -> sctk = tickGet();
			    ppto -> nego = 0;
			    /* write the the off condition */
			    if (!wslv)
			    	status = (*pdset -> write_pto)(ppto, 1);
			    if (!status)
			    {
				/* since we are reversing direction, do we have
				   a minimum off time on direction transition
				*/
				if (ppto -> mnpw)
				{
				    wait_time = (int)(ppto -> mnpw * vxTicksPerSecond);
				    ppto -> cbfl = PTO_DURATION_MIN_POS;
				}
				/* no but do we have have a minimum off time */
				else if (ppto -> poft)
				{
				    wait_time = (int)(ppto -> poft * vxTicksPerSecond);
				    ppto -> cbfl = PTO_DURATION_MIN_POS;
				}
				/* we got nuttin, so just reverse it */
				else
				{
				    if (ppto -> poiv)
					ppto -> rbv = 0;
				    else
					ppto -> rbv = ppto -> pmsk;
			    	    if (!wslv)
				    	status = (*pdset -> write_pto)(ppto, 0);
				    if (!status)
				    {
					ppto -> poso = 1;
					ppto -> ldir = 1;
					temp = ppto -> rval;
					if (temp < 0)
					    temp = -temp;
					wait_time = (int)temp;
					ppto -> cbfl = PTO_DURATION_ON;
				    }
				    else
				    {
					ppto -> poso = 0;
					ppto -> cbfl = PTO_NO_CALLBACK;
				    }
				}
			    }
			    else
			    {
				ppto -> poso = ppto -> nego = 0;
				ppto -> cbfl = PTO_NO_CALLBACK;
			    }
			}
			ppto -> ncmd = PTO_NO_CMD;
		    }
		    /* nothing to do , this is an error, should not be here */
		    else
		    {
			ppto -> ncmd = PTO_NO_CMD;
		    }
		}
		/* we weren't going anywhere */
		else
		{
		    if (ppto -> rval > 0)
		    {
			/* if there is a min trans time and we are reversing
			   direction, do the min first
			*/
		    	if (ppto -> mnpw && (ppto -> ldir < 0))
		    	{
			    dt = (tickGet() - ppto -> sctk) / (float)vxTicksPerSecond;
			    if (dt < ppto -> mnpw)
			    {
		 		wait_time = (ppto -> mnpw - dt) * vxTicksPerSecond;
		 		ppto -> cbfl = PTO_DURATION_MIN_POS;
		 		ppto -> ncmd = PTO_NO_CMD;
			    break;
			    }
		    	}
			if (ppto -> poiv)
			    ppto -> rbv = 0;
			else
			    ppto -> rbv = ppto -> pmsk;
			if (!wslv)
			    status = (*pdset -> write_pto)(ppto, 0);
			if (!status)
			{
			    ppto -> poso = 1;
			    ppto -> ldir = 1;
			    ppto -> cbfl = PTO_DURATION_ON;
			}
		    }
		    else if (ppto -> rval < 0)
		    {
			/* if there is a min trans time and we are reversing
			   direction, do the min first
			*/
		    	if (ppto -> mnpw && (ppto -> ldir > 0))
		    	{
			    dt = (tickGet() - ppto -> sctk) / (float)vxTicksPerSecond;
			    if (dt < ppto -> mnpw)
			    {
		 		wait_time = (ppto -> mnpw - dt) * vxTicksPerSecond;
		 		ppto -> cbfl = PTO_DURATION_MIN_NEG;
		 		ppto -> ncmd = PTO_NO_CMD;
			    break;
			    }
		    	}
			if (ppto -> noiv)
			    ppto -> rbv = 0;
			else
			    ppto -> rbv = ppto -> nmsk;
			if (!wslv)
			    status = (*pdset -> write_pto)(ppto, 1);
			if (!status)
			{
			    ppto -> nego = 1;
			    ppto -> ldir = -1;
			    ppto -> cbfl = PTO_DURATION_ON;
			}
		    }
		    if (!status)
		    {
			temp = ppto -> rval;
			if (temp < 0)
			    temp = -temp;
			wait_time = (int)temp;
		    }
		    else
		    {
			ppto -> poso = ppto -> nego = 0;
			ppto -> cbfl = PTO_NO_CALLBACK;
		    }
		    ppto -> ncmd = PTO_NO_CMD;
		}
	    }
	    break;

	case PTO_POS_CHANNEL:
	    if ( (ppto -> poiv && (!ppto -> rbv)) ||
		 ((!ppto -> poiv) && ppto -> rbv) )
	    {
		if (ppto -> poiv)
		    ppto -> rbv = ppto -> pmsk;
		else
		    ppto -> rbv = 0;
		if (!wslv)
		    status = (*pdset -> write_pto)(ppto, 0);
		if (!status)
		    wait_time = (int)(ppto -> poft * vxTicksPerSecond);
		else
		{
		    ppto -> cbfl = PTO_NO_CALLBACK;
		    ppto -> poso = 0;
		    wait_time = 0;
		}
	    }
	    /* at end of a cycle */
	    else
	    {
		/* something new to do */
		if (ppto -> ncmd)
		{
		    /* no more pulses to do */
		    if (ppto -> rval == 0)
		    {
			ppto -> cbfl = PTO_NO_CALLBACK;
			ppto -> poso = ppto -> nego = 0;
			ppto -> pcnt = 0;
			wait_time = 0;
		    }
		    /* already going in POS direction, recompute number 
		       of cycles
		    */
		    else if (ppto -> rval > 0)
		    {
			if (ppto -> poiv)
			    ppto -> rbv = 0;
			else
			    ppto -> rbv = ppto -> pmsk;
			if (!wslv)
			    status = (*pdset -> write_pto)(ppto, 0);
			ppto -> poso = 1;
			ppto -> ldir = 1;
			if (!status)
			{
			    temp = ppto -> rval;
			    if (temp < 0)
				temp = -temp;
			    ppto -> pcnt = temp;
			    wait_time = (int)(ppto -> pont * vxTicksPerSecond);
			}
		    }
		    /* rval < 0, reverse direction */
		    else
		    {
			if (ppto -> noiv)
			    ppto -> rbv = 0;
			else
			    ppto -> rbv = ppto -> nmsk;
			if (!wslv)
			    status = (*pdset -> write_pto)(ppto, 1);
			if (!status)
			{
			    ppto -> nego = 1;
			    ppto -> ldir = -1;
			    ppto -> cbfl = PTO_NEG_CHANNEL;
			    temp = ppto -> rval;
			    if (temp < 0)
				temp = -temp;
			    ppto -> pcnt = temp;
			    wait_time = (int)(ppto -> pont * vxTicksPerSecond);
			}
			else
			{
			    ppto -> cbfl = PTO_NO_CALLBACK;
			    ppto -> pcnt = 0;
			    ppto -> poso = 0;
			}
		    }
		    ppto -> ncmd = PTO_NO_CMD;
		}
		/* nothing new so just continue in the same vain */
		else
		{
		    ppto -> pcnt--;

		    /* we are finished */
		    if (ppto -> pcnt <= 0)
		    {
			ppto -> cbfl = PTO_NO_CALLBACK;
			ppto -> poso = 0;
		    }
		    /* start another cycle */
		    else
		    {
			if (ppto -> poiv)
			    ppto -> rbv = 0;
			else
			    ppto -> rbv = ppto -> pmsk;
			if (!wslv)
			    status = (*pdset -> write_pto)(ppto, 0);
			if (!status)
			{
			    wait_time = (int)(ppto -> pont * vxTicksPerSecond);
			    ppto -> poso = 1;
			    ppto -> ldir = 1;
			}
			else
			{
			    ppto -> cbfl = PTO_NO_CALLBACK;
			    ppto -> poso = 0;
			}
		    }
		}
	    }

	    break;

	case PTO_NEG_CHANNEL:
	    if ( (ppto -> noiv && (!ppto -> rbv)) ||
		 ((!ppto -> noiv) && ppto -> rbv) )
	    {
		if (ppto -> noiv)
		    ppto -> rbv = ppto -> nmsk;
		else
		    ppto -> rbv = 0;
		if (!wslv)
		    status = (*pdset -> write_pto)(ppto, 1);
		if (!status)
		    wait_time = (int)(ppto -> poft * vxTicksPerSecond);
		else
		{
		    ppto -> cbfl = PTO_NO_CALLBACK;
		    ppto -> nego = 0;
		    wait_time = 0;
		}
	    }
	    /* at end of a cycle */
	    else
	    {
		/* something new to do */
		if (ppto -> ncmd)
		{
		    /* no more pulses to do */
		    if (ppto -> rval == 0)
		    {
			ppto -> cbfl = PTO_NO_CALLBACK;
			ppto -> nego = 0;
			ppto -> pcnt = 0;
			wait_time = 0;
		    }
		    /* already going in NEG direction, recompute number 
		       of cycles
		    */
		    else if (ppto -> rval < 0)
		    {
			if (ppto -> noiv)
			    ppto -> rbv = 0;
			else
			    ppto -> rbv = ppto -> nmsk;
			if (!wslv)
			    status = (*pdset -> write_pto)(ppto, 1);
			ppto -> nego = 1;
			ppto -> ldir = -1;
			if (!status)
			{
			    temp = ppto -> rval;
			    if (temp < 0)
				temp = -temp;
			    ppto -> pcnt = temp;
			    wait_time = (int)(ppto -> pont * vxTicksPerSecond);
			}
		    }
		    /* rval > 0, reverse direction */
		    else
		    {
			if (ppto -> poiv)
			    ppto -> rbv = 0;
			else
			    ppto -> rbv = ppto -> pmsk;
			if (!wslv)
			    status = (*pdset -> write_pto)(ppto, 0);
			if (!status)
			{
			    ppto -> poso = 1;
			    ppto -> ldir = 1;
			    ppto -> cbfl = PTO_POS_CHANNEL;
			    temp = ppto -> rval;
			    if (temp < 0)
				temp = -temp;
			    ppto -> pcnt = temp;
			    wait_time = (int)(ppto -> pont * vxTicksPerSecond);
			}
			else
			{
			    ppto -> cbfl = PTO_NO_CALLBACK;
			    ppto -> pcnt = 0;
			    ppto -> nego = 0;
			}
		    }
		    ppto -> ncmd = PTO_NO_CMD;
		}
		/* nothing new so just continue in the same vain */
		else
		{
		    ppto -> pcnt--;
		    /* we are finished */
		    if (ppto -> pcnt <= 0)
		    {
			ppto -> cbfl = PTO_NO_CALLBACK;
			ppto -> nego = 0;
		    }
		    /* start another cycle */
		    else
		    {
			if (ppto -> noiv)
			    ppto -> rbv = 0;
			else
			    ppto -> rbv = ppto -> nmsk;
			if (!wslv)
			     status = (*pdset -> write_pto)(ppto, 1);
			if (!status)
			{
			    wait_time = (int)(ppto -> pont * vxTicksPerSecond);
			    ppto -> nego = 1;
			    ppto -> ldir = -1;
			}
			else
			{
			    ppto -> cbfl = PTO_NO_CALLBACK;
			    ppto -> nego = 0;
			}
		    }
		}
	    }
	    break;

	case PTO_DURATION_ON:
	    /* at the end of a DURATION , turn off/on signal to end duration*/
	    if (ppto -> poso)
	    {
		ppto -> poso = 0;
		if (ppto -> poiv)
		    ppto -> rbv = ppto -> pmsk;
		else
		    ppto -> rbv = 0;
		ppto -> sctk = tickGet();
	        if (!wslv)
		    status = (*pdset -> write_pto)(ppto, 0);
		if (!status)
		{
		    /* is there a minimum off time */
		    if (ppto -> poft)
		    {
			wait_time = (int)(ppto -> poft * vxTicksPerSecond);
			ppto -> cbfl = PTO_DURATION_OFF;
		    }
		    else
			ppto -> cbfl = PTO_NO_CALLBACK;
		}
		else
		    ppto -> cbfl = PTO_NO_CALLBACK;
	    }
	    else if (ppto -> nego)
	    {
		ppto -> nego = 0;
		if (ppto -> noiv)
		    ppto -> rbv = ppto -> nmsk;
		else
		    ppto -> rbv = 0;
		ppto -> sctk = tickGet();
		if (!wslv)
		    status = (*pdset -> write_pto)(ppto, 1);
		if (!status)
		{
		    if (ppto -> poft)
		    {
			wait_time = (int)(ppto -> poft * vxTicksPerSecond);
			ppto -> cbfl = PTO_DURATION_OFF;
		    }
		    else
			ppto -> cbfl = PTO_NO_CALLBACK;
		}
		else
		    ppto -> cbfl = PTO_NO_CALLBACK;
	    }
	    break;

	case PTO_DURATION_OFF:
	    /* we have finished the minimum off time for a duration,
	       but there is someting new to do
	    */
	    if (ppto -> ncmd)
	    {
		if (ppto -> rval > 0)
		{
		    /* if there is a min trans time and we are reversing
		       direction, do the min first
		    */
	    	    if (ppto -> mnpw && (ppto -> ldir < 0))
	    	    {
			dt = (tickGet() - ppto -> sctk) / (float)vxTicksPerSecond;
			if (dt < ppto -> mnpw)
			{
			    wait_time = (ppto -> mnpw - dt) * vxTicksPerSecond;
			    ppto -> cbfl = PTO_DURATION_MIN_POS;
			    ppto -> ncmd = PTO_NO_CMD;
			    break;
			}
	    	    }
		    if (ppto -> poiv)
			ppto -> rbv = 0;
		    else
			ppto -> rbv = ppto -> pmsk;
		    if (!wslv)
		    	status = (*pdset -> write_pto)(ppto, 0);
		    if (!status)
		    {
			ppto -> poso = 1;
			ppto -> ldir = 1;
			ppto -> cbfl = PTO_DURATION_ON;
		    }
		}
		else if (ppto -> rval < 0)
		{
		    /* if there is a min trans time and we are reversing
		       direction, do the min first
		    */
	    	    if (ppto -> mnpw && (ppto -> ldir > 0))
	    	    {
			dt = (tickGet() - ppto -> sctk) / (float)vxTicksPerSecond;
			if (dt < ppto -> mnpw)
			{
			    wait_time = (ppto -> mnpw - dt) * vxTicksPerSecond;
			    ppto -> cbfl = PTO_DURATION_MIN_NEG;
			    ppto -> ncmd = PTO_NO_CMD;
			    break;
			}
	    	    }
		    if (ppto -> noiv)
			ppto -> rbv = 0;
		    else
			ppto -> rbv = ppto -> nmsk;
		    if (!wslv)
		        status = (*pdset -> write_pto)(ppto, 1);
		    if (!status)
		    {
			ppto -> nego = 1;
			ppto -> ldir = -1;
			ppto -> cbfl = PTO_DURATION_ON;
		    }
		}
		if ((!status) && (ppto -> rval))
		{
		    temp = ppto -> rval;
		    if (temp < 0)
			temp = -temp;
		    wait_time = (int)temp;
		}
		else
		{
		    ppto -> poso = ppto -> nego = 0;
		    ppto -> cbfl = PTO_NO_CALLBACK;
		}
		ppto -> ncmd = PTO_NO_CMD;
	    }
	    /* nothing more to do */
	    else
	    {
		ppto -> cbfl = PTO_NO_CALLBACK;
		ppto -> poso = 0;
		ppto -> nego = 0;
	    }

	    break;

	case PTO_DURATION_MIN_POS:
	    /* we have finished the minimum +/- transition time 
	       but there is something new to do 
	    */
	    if (ppto -> ncmd)
	    {
		if (ppto -> rval > 0)
		{
		    if (ppto -> poiv)
			ppto -> rbv = 0;
		    else
			ppto -> rbv = ppto -> pmsk;
		    if (!wslv)
		        status = (*pdset -> write_pto)(ppto, 0);
		    if (!status)
		    {
			ppto -> poso = 1;
			ppto -> ldir = 1;
			ppto -> cbfl = PTO_DURATION_ON;
		    }
		}
		else if (ppto -> rval < 0)
		{
		    if (ppto -> noiv)
			ppto -> rbv = 0;
		    else
			ppto -> rbv = ppto -> nmsk;
		    if (!wslv)
		        status = (*pdset -> write_pto)(ppto, 1);
		    if (!status)
		    {
			ppto -> nego = 1;
			ppto -> ldir = -1;
			ppto -> cbfl = PTO_DURATION_ON;
		    }
		}
		if ((!status) && (ppto -> rval))
		{
		    temp = ppto -> rval;
		    if (temp < 0)
			temp = -temp;
		    wait_time = (int)temp;
		}
		else
		{
		    ppto -> poso = ppto -> nego = 0;
		    ppto -> cbfl = PTO_NO_CALLBACK;
		}
	    }
	    /* nothing new just do the reverse */
	    else
	    {
		if (ppto -> poiv)
		    ppto -> rbv = 0;
		else
		    ppto -> rbv = ppto -> pmsk;
		if (!wslv)
		    status = (*pdset -> write_pto)(ppto, 0);
		if (!status)
		{
		    ppto -> poso = 1;
		    ppto -> ldir = 1;
		    ppto -> cbfl = PTO_DURATION_ON;
		    temp = ppto -> rval;
		    if (temp < 0)
			temp = -temp;
		    wait_time = (int)temp;
		}
		else
		{
		    ppto -> poso = ppto -> nego = 0;
		    ppto -> cbfl = PTO_NO_CALLBACK;
		}
	    }
	    ppto -> ncmd = PTO_NO_CMD;

	    break;

	case PTO_DURATION_MIN_NEG:
	    /* we have finished the -/+ transition time but
	       there is something new to do
	    */
	    if (ppto -> ncmd)
	    {
		if (ppto -> rval > 0)
		{
		    if (ppto -> poiv)
			ppto -> rbv = 0;
		    else
			ppto -> rbv = ppto -> pmsk;
		    if (!wslv)
		        status = (*pdset -> write_pto)(ppto, 0);
		    if (!status)
		    {
			ppto -> poso = 1;
			ppto -> ldir = 1;
			ppto -> cbfl = PTO_DURATION_ON;
		    }
		}
		else if (ppto -> rval < 0)
		{
		    if (ppto -> noiv)
			ppto -> rbv = 0;
		    else
			ppto -> rbv = ppto -> nmsk;
		    if (!wslv)
		        status = (*pdset -> write_pto)(ppto, 1);
		    if (!status)
		    {
			ppto -> nego = 1;
			ppto -> ldir = -1;
			ppto -> cbfl = PTO_DURATION_ON;
		    }
		}
		if ((!status) && (ppto -> rval))
		{
		    temp = ppto -> rval;
		    if (temp < 0)
			temp = -temp;
		    wait_time = (int)temp;
		}
		else
		{
		    ppto -> poso = ppto -> nego = 0;
		    ppto -> cbfl = PTO_NO_CALLBACK;
		}
	    }
	    /* nothing new, just do the reverse */
	    else
	    {
		if (ppto -> noiv)
		    ppto -> rbv = 0;
		else
		    ppto -> rbv = ppto -> nmsk;
		if (!wslv)
		    status = (*pdset -> write_pto)(ppto, 1);
		if (!status)
		{
		    ppto -> nego = 1;
		    ppto -> ldir = -1;
		    ppto -> cbfl = PTO_DURATION_ON;
		    temp = ppto -> rval;
		    if (temp < 0)
			temp = -temp;
		    wait_time = (int)temp;
		}
		else
		{
		    ppto -> poso = ppto -> nego = 0;
		    ppto -> cbfl = PTO_NO_CALLBACK;
		}
	    }
	    ppto -> ncmd = PTO_NO_CMD;

	    break;

	default:
	    break;
    }
    if (wait_time)
    {
	if (pcallback -> wd_id == NULL)
	{
	    pcallback -> wd_id = wdCreate ();
	}
	wdStart (pcallback -> wd_id, wait_time, (FUNCPTR)callbackRequest, (int)pcallback);
    }
    return;

}



/*------------------------------------------------------------------------*/

static long special (paddr, after)
    struct dbAddr	*paddr;
    int			after;
{
    struct ptoRecord	*ppto = (struct ptoRecord *)(paddr -> precord);
    CALLBACK_PTR	pcallback = (CALLBACK_PTR)(ppto -> dpvt);
    int			special_type = paddr -> special;
    int			wait_time;

    if (!after)
	return (0);

    switch (special_type)
    {
	/* RVAL was sent by the MASTER, so we may have to start a callback
	   on the slave, to keep, MASTER/SLAVE in sync
	*/
	case SPC_PTO_RVAL:
	    /* do not start a callback if we are not fully init'd */
	    if (!ppto -> ini2)
		break;

	    /* update old raw value, for posting, see monitor () */
	    ppto -> oraw = ppto -> rval;

	    if (ppto -> ptyp == DURATION)
	    {
		if (ppto -> rval)
		    ppto -> ncmd = PTO_NEW_CMD;
		if (ppto -> cbfl == PTO_DURATION_ON)
		{
		    wdCancel (pcallback -> wd_id);
		    ppto -> cbfl = PTO_NO_CALLBACK;
		    ppto -> ncmd = PTO_NEW_CMD;
		}
	    }
	    else
	    {
		if (ppto -> rval)
		    ppto -> ncmd = PTO_NEW_CMD;
	    }
	    if (ppto -> ncmd)
	    {
		if (ppto -> cbfl == PTO_NO_CALLBACK)
		{
		    wait_time = 1;
		    ppto -> cbfl = PTO_START_CALLBACK;
		    if (pcallback -> wd_id == NULL)
		    {
			pcallback -> wd_id = wdCreate();
		    }
		    wdStart (pcallback -> wd_id, wait_time, (FUNCPTR)callbackRequest, 
			     (int)pcallback);
		}
	    }
	    break;

	default:
	    recGblDbaddrError (S_db_badChoice, paddr, "pto: special");
	    return (S_db_badChoice);
    }

    return (0);

}
