/*recStringin.c,v 1.1.1.1 1999/08/03 14:44:22 tis4000 Exp */

/* recStringin.c - Record Support Routines for Stringin records */
/*
 *      Author: 	Janet Anderson
 *      Date:   	4/23/91
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  mm-dd-yy        iii     Comment
 * .02  01-03-94	ems	set ack to 0 in init_record
 *				in process, if not init"d do nothing
 * .03  03-01-94	ems	added support for process stats
 *				reomved ack = 0
 * .04	03-29-94	ems	make forward func def agree with actual func def
 * .05	05-16-94	ems	added redundancy, rewrite init_record, added ack
 *				to monitor()
 * .06	06-10-94	ems	move call to tsLocalTime inside if on scan but
 *				before call to device
 * .07	06-14-94	ems	added support for maximize severity (msev)
 * .08  08-18-94	ems	monitor(), rm DBE_LOG from ack
 * .09	10-12-94	ems	add TIME to redundancy, added fpps (1st process pass)
 *				and lalm, do time stamp when stat or sevr chg,
 *				see monitor(),and not in process(),set lalm to val when 
 *				stat or sevr chg,in process() if fpps (first process)
 *				init lalm to val and if time stamp is 0, init it also
 * .10	10-20-94	ems	chg time to latm (last alarm time) added 
 *				added lptm (last process time)
 * .11	12-19-94	ems	removed lalm and fpps (first process pass), not needed
 * .12  01-13-95        wep     init lptm and latm
 * .13	03-15-95	ems	monitor(), do not set ACK if sevr is EVENT_ALARM
 * .14	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 * .15	07-11-96	ems	added alarm support and CVSV, chg VAL severity, made
 *				LVAL and OVAL redundant, had to increase redundant
 *				buffer size form 256 to 512
 */


#include        <vxWorks.h>
#include	<inetLib.h>
#include        <types.h>
#include        <stdioLib.h>
#include        <lstLib.h>
#include	<strLib.h>

#include        "alarm.h"
#include        "dbDefs.h"
#include	"dbCommon.h"
#include        "link.h"
#include	"db_field_log.h"
#include        "dbAccess.h"
#include        "dbFldTypes.h"
#include	"dbRecType.h"
#include        "devSup.h"
#include        "errMdef.h"
#include        "recSup.h"
#include	"rec/stringinRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"siredfldidx.h"
#include	"restable.h"
#include	"db_addr.h"
#include	"recGbl.h"
#include	"dbEvent.h"



/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/
#define report NULL
#define initialize NULL
static long init_record();
static long process();
#define special NULL
static long get_value();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
#define get_precision NULL
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
#define get_graphic_double NULL
#define get_control_double NULL
#define get_alarm_double NULL

struct rset stringinRSET={
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double };

struct stringindset { /* stringin input dset */
	long		number;
	DEVSUPFUN	dev_report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record; /*returns: (-1,0)=>(failure,success)*/
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_stringin;/*(0,1)=> success, async */
};

/* The folowing definitions are for redundancy */
 
static struct fldDes	**mylist = 0;

static	char *namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"INI2",
		"LATM",
		"OVAL",
		"LVAL",
		0
		};

static unsigned short	type;

static void alarm ();
static void monitor();


/*------------------------------------------------------------------------*/

static long init_record(pstringin)
    struct stringinRecord	*pstringin;
{
    struct stringindset *pdset;
    long		status;
    char 		*pstr;
    unsigned short	i;
    char		redbuf[512];
    int			ini2_flag = 0;

    if (!(mylist)) 
    {
	for (type = 1; type < 100; type++) 
	{
	    pstr = GET_PRECTYPE (type);
	    if (strncmp (pstr, "stringin", 8) == 0) 
		break;
	}
	mylist = redFldAlloc (type, NUMSIFLDS);

	if (!mylist)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)pstringin,
				"stringin, init_record, mylist failed");
	    return (S_rec_Alloc);
	}

	for (i = 0; i < NUMSIFLDS; i++) 
	    redFldAdd (type, i, namelist [i]);
    }

    if(!(pdset = (struct stringindset *)(pstringin->dset))) 
    {
	recGblRecordError (S_dev_noDSET, (struct dbCommon *)pstringin,
						"stringin: init_record");
	return(S_dev_noDSET);
    }

    /* must have read_stringin function defined */
    if( (pdset->number < 5) || (pdset->read_stringin == NULL) ) 
    {
	recGblRecordError (S_dev_missingSup, (struct dbCommon *)pstringin,
						"stringin: init_record");
	return(S_dev_missingSup);
    }

    pstringin -> init = 1;
    pstringin -> nsta = NO_ALARM;
    pstringin -> nsta = NO_ALARM;
    pstringin -> msev = NO_ALARM;

    /* init last process time and last alarm time */
    if (pstringin -> lptm.secPastEpoch == 0)
        tsLocalTime (&pstringin -> lptm);
    if (pstringin -> latm.secPastEpoch == 0)
        tsLocalTime (&pstringin -> latm);

    if( pdset->init_record )
    {
	status = (*pdset->init_record) (pstringin,process);
	if (status)
	{
	    /* if we are tagged, off scan or in manual, then we will
	       allow the point to init
	    */

	    if ( (!pstringin -> tagd) && (pstringin -> oscn != OFF_SCAN) &&
		 (pstringin -> mode == NORMAL) )
	    {
		pstringin -> nsta = NO_INIT_ALARM;
		pstringin -> nsev = VALID_ALARM;
	    }
	    else
		ini2_flag = 1;
	}
	else
	    ini2_flag = 1;
    }
    else
	ini2_flag = 1;

    if (ini2_flag)
    {
	strncpy (pstringin -> oval, pstringin -> val, sizeof (pstringin -> oval) - 1);
	strncpy (pstringin -> lval, pstringin -> val, sizeof (pstringin -> lval) - 1);
    }

    /* if we are MASTER post VAL and send it to redundant partner */

    if (resGetCaAnswer (pstringin -> radr, &i) == RES_OK)
    {
	if (i)
	{
	    /* post STAT, SEVR and VAL (monitor), if we will finally init, 
	       because we have read a value from the input device, set INI2,
	       send all redund fields to slave, when slave receives INI2, 
	       it will be fully init'd also
	    */
	
	    /* check for maximize severity */

	    if (pstringin -> msev > pstringin -> nsev)
	    {
		pstringin -> nsev = pstringin -> msev;
		pstringin -> nsta = LINK_ALARM;
	    }

	    monitor (pstringin);

	    if (ini2_flag)
	    {
		pstringin -> ini2 = 1;
		redQInit ((struct dbCommon *)pstringin, redbuf, sizeof (redbuf), type);
		redQField (redbuf, mylist [STAT]);
		redQField (redbuf, mylist [SEVR]);
		redQField (redbuf, mylist [VAL]);
		redQField (redbuf, mylist [INI2]);
		redQField (redbuf, mylist [OVAL]);
		redQField (redbuf, mylist [LVAL]);
		
		redQSend (redbuf);
	    }
	}
    }

    return(0);
}


/*------------------------------------------------------------------------*/

static void alarm (pstringin)
    struct stringinRecord	*pstringin;
{

    int		adis;

    /* check for alarm inhibit */

    adis = AlarmDisableLink ((void *)pstringin);

    if ( (pstringin -> aenb == ALARM_DISABLED) || (adis != 0) )
    {
	if (pstringin -> nsta != NO_INIT_ALARM)
	{
	    pstringin->nsta = INHIBIT_ALARM;
	    pstringin->nsev = NO_ALARM;
	}
	return;
    }
    if(pstringin->udf == TRUE ) 
    {
	if (pstringin->nsev < VALID_ALARM) 
	{
	    pstringin->nsta = UDF_ALARM;
	    pstringin->nsev = VALID_ALARM;
	}
	return;
    }

    if (strncmp (pstringin -> val, pstringin -> lval, sizeof (pstringin -> val)) != 0)
    {
	if (pstringin -> cvsv > pstringin -> nsev)
	{
	    pstringin -> nsev = pstringin -> cvsv;
	    pstringin -> nsta = COS_ALARM;
	}
    }

    strncpy (pstringin -> lval, pstringin -> val, sizeof (pstringin -> lval) - 1);

    return;
}


/*------------------------------------------------------------------------*/

static long process(paddr)
    struct dbAddr	*paddr;
{
    struct stringinRecord	*pstringin=(struct stringinRecord *)(paddr->precord);
    struct stringindset		*pdset = (struct stringindset *)(pstringin->dset);
    long			status = 0;
    unsigned short		stat = pstringin -> stat;
    unsigned short		sevr = pstringin -> sevr;
    char			val[80];
    char			lval[80];
    char			oval[80];
    TS_STAMP			latm = pstringin -> latm;
    char			redbuf[512];

    /* if not init'd, do nothing */

    if ((!pstringin -> init) || (!pstringin -> ini2))
	return (0);

    if( (pdset==NULL) || (pdset->read_stringin==NULL) ) 
    {
	pstringin->pact=TRUE;
	recGblRecordError (S_dev_missingSup, (struct dbCommon *)pstringin,"read_stringin");
	return(S_dev_missingSup);
    }

    /* get time stamp for last time record was processed */

    tsLocalTime (&pstringin -> lptm);

    pstringin -> msev = NO_ALARM;
    strncpy (val, pstringin -> val, sizeof (val) - 1);
    strncpy (oval, pstringin -> oval, sizeof (oval) - 1);
    strncpy (lval, pstringin -> lval, sizeof (lval) - 1);

    if (pstringin -> tagd)
    {
	pstringin -> nsta = TAGGED_ALARM;
	pstringin -> nsev = VALID_ALARM;
    }
    else if (pstringin -> oscn == OFF_SCAN)
    {
	pstringin -> nsta = SCAN_ALARM;
	pstringin -> nsev = VALID_ALARM;
    }
    else
    {
	pstringin -> nsta = NO_ALARM;
	pstringin -> nsev = NO_ALARM;

	/* update number of process times */
	if (process_stats)
	    process_cnts[type]++;	

	if (pstringin -> mode == NORMAL)
	{
	    /* read the new value */
	    status = (*pdset->read_stringin)(pstringin); 
	    pstringin->pact = TRUE;

	    /* status is one if an asynchronous record is being processed*/
	    if (status == 1) 
		return(0);
	}

	alarm (pstringin);
	
	/* check for maximize severity */

	if (pstringin -> msev > pstringin -> nsev)
	{
	    pstringin -> nsev = pstringin -> msev;
	    pstringin -> nsta = LINK_ALARM;
	}
    }

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (pstringin -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&pstringin -> latm);
	db_post_events ((struct dbCommon *)pstringin, pstringin -> val, DBE_VALUE | DBE_ALARM);
    }

    /* check event list */
    monitor(pstringin);

    /* process the forward scan link record */
    if (pstringin->flnk.type==DB_LINK) 
	dbScanPassive((struct dbAddr *)pstringin->flnk.value.db_link.pdbAddr);

    /* update redundant fields to slave */

    redQInit ((struct dbCommon *)pstringin, redbuf, sizeof (redbuf), type);

    if (stat != pstringin -> stat)
	redQField (redbuf, mylist [STAT]);

    if (sevr != pstringin -> sevr)
	redQField (redbuf, mylist [SEVR]);

    if (strncmp (val, pstringin -> val, sizeof (val)))
	redQField (redbuf, mylist [VAL]);

    if (TsCmpStampsNE (&latm, &pstringin -> latm) )
	redQField (redbuf, mylist [LATM]);

    if (strncmp (oval, pstringin -> oval, sizeof (oval)))
	redQField (redbuf, mylist [OVAL]);

    if (strncmp (lval, pstringin -> lval, sizeof (lval)))
	redQField (redbuf, mylist [LVAL]);

    redQSend (redbuf);

    pstringin->pact=FALSE;

    return(status);
}


/*------------------------------------------------------------------------*/

static long get_value(pstringin,pvdes)
    struct stringinRecord             *pstringin;
    struct valueDes     *pvdes;
{
    pvdes->field_type = DBF_STRING;
    pvdes->no_elements=1;
    pvdes->pvalue = (caddr_t)(&pstringin->val[0]);
    return(0);
}


/*------------------------------------------------------------------------*/

static void monitor(pstringin)
    struct stringinRecord             *pstringin;
{
    unsigned short  monitor_mask;
    short           stat,sevr,nsta,nsev;
    int		    set_ack = 1;

    /* get previous stat and sevr  and new stat and sevr*/
    stat=pstringin->stat;
    sevr=pstringin->sevr;
    nsta=pstringin->nsta;
    nsev=pstringin->nsev;

    /*set current stat and sevr*/
    pstringin->stat = nsta;
    pstringin->sevr = nsev;
    pstringin->nsta = 0;
    pstringin->nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == pstringin -> psta) && (nsev == pstringin -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and out previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	pstringin -> psta = stat;
	pstringin -> psev = sevr;
	db_post_events ((struct dbCommon *)pstringin, (caddr_t)&pstringin -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)pstringin, (caddr_t)&pstringin -> psev, DBE_VALUE);
    }

    /* Flags which events to fire on the value field */
    monitor_mask = 0;

    /* alarm condition changed this scan */
    if ( (stat != nsta) || (sevr != nsev) )
    {
	/* post events for alarm condition change*/
	monitor_mask = DBE_ALARM;

	/* when alarm status chg's, get current time 
	*/
	tsLocalTime (&pstringin -> latm);

	/* post stat and nsev fields */
	db_post_events ((struct dbCommon *)pstringin, (caddr_t)&pstringin->stat,DBE_VALUE);
	db_post_events ((struct dbCommon *)pstringin,
					(caddr_t)&pstringin->sevr,DBE_VALUE);

	if ( (pstringin -> sevr != NO_ALARM) && 
	     (set_ack) &&
	     (pstringin -> sevr != EVENT_ALARM) )
	{
	    if (pstringin -> ack == 0)
	    {
		pstringin -> ack = 1;
		db_post_events ((struct dbCommon *)pstringin, 
					(caddr_t)&pstringin -> ack, DBE_VALUE);
	    }
	}
    }
    if( (strncmp(pstringin->oval,pstringin->val,sizeof(pstringin->oval))) ||
	( monitor_mask & DBE_ALARM) ) 
    {
        db_post_events ((struct dbCommon *)pstringin, 
					&(pstringin->val[0]), monitor_mask|DBE_VALUE);
	strncpy(pstringin->oval, pstringin->val, sizeof(pstringin->oval));
    }
    return;
}

