/*recTank.c,v 1.1.1.1 1999/08/03 14:44:22 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************************
 *
 *  MODULE NAME            : recTank.c 
 *  AUTHOR                 : emsteel
 *  DATE                   : 13 Apr 93
 *
 *
 *  DESCRIPTION: record support routines for a Tank record
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 23 Feb 94
 *  REVISION:  removed INIT, ALST, MLST, ESIN from namelist and tankred
 *             not needed
 *             removed time from tankred, not needed
 *             in various places check for init flag before doing
 *               db_post_events, assumed we are called from process if true
 *               and are therefore MASTER
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 02 Mar 94
 *  REVISION: added process stats code
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 29 Mar 94
 *  REVISION:  make forward func def agree with actual func def
 *             made save_tank static
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 25 May 94
 *  REVISION:	rewrite init_record, add INI2, PCMD and cmds to redundancy
 *		move db_post_events to post_events
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 June 94
 *  REVISION:	added support for maximize severity (msev), move call to
 *		tsLocalTime inside on scan
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 01 Sep 94
 *  REVISION:	added get_control_double(),get_graphic_double() and
 *		get_link_value()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 Oct 94
 *  REVISION:	add TIME to redundancy, fpas chg'd to frps, added fpps
 *		(1st process pass), do time stamp when stat or sevr chg,
 *		see monitor(),and not in process(),set lalm to val when 
 *		stat or sevr chg,in process() if fpps (first process)
 *		init lalm to val and if time stamp is 0, init it also
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 Oct 94
 *  REVISION:	chg time to latm (last alarm time)
 *		added lptm (last process time)
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 Jan 95
 *  REVISION:	add mlst to redundancy, init latm and lptm in init_record
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 21 MAr 95
 *  REVISION:	update QVOL, stopped volume with NETV net volume not VAL
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 19 Feb 96
 *  REVISION:	make all trip globals into struct so can be reentrant
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 Jun 96
 *  REVISION:	added code for alarm disable ADIS
 *
****************************************************************************/

/****************************************************************************
*
* Programmer's Notes:
*
* ComputeTFFlowRate () - see tankfarm.c
* ComputeTFVolume() - see tankfarm.c
* FindAPIcf () - see tankfarm.c
* FindAPITable () - see tankfarm.c
* FindStrappingTable () - see tankfarm.c
* TankFarmRound () - see tankfarm.c
*
* GET_PARR_CHOICE etc - macros defined in .../epicsH/choice.h
* dbGetField() and dbNameToAddr() - see .../src/db/dbAccess.c
* TsDiffAsDouble - macro in .../epicsH/tsDefs.h
* pvdGetFld () - see .../src/libCom/pvd.c
*
* Numerous calls to dp_post_events, any value in the record that changes
*   due to processing must be posted in case the value is being monitored
*
****************************************************************************/


#include <vxWorks.h>
#include <inetLib.h>
#include <tickLib.h>       /* for tickGet () */
#include <types.h>
#include <stdioLib.h>
#include <string.h>
#include <lstLib.h>

#include "alarm.h"
#include "choice.h"
#include "cvtTable.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"
#include "dbFldTypes.h"
#include "dbRecType.h"
#include "devSup.h"
#include "errMdef.h"
#include "recSup.h"
#include "special.h"
#include "rec/tankRecord.h"
#include "redflds.h"
#include "redque.h"
#include "tankredfldidx.h"
#include "tsDefs.h"          
#include "tankfarm.h"
#include "rec/aiRecord.h"    /* needed by get_level_values () */
#include "rec/calcRecord.h"  /* needed by get_level_values () */
#include "restable.h"
#include "db_addr.h"
#include "recGbl.h"
#include "dbEvent.h"
#include "okToUse.h"
#include "error.h"
#include "pvd.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table */

#define report NULL
#define initialize NULL
static long init_record();
static long process();
static long special();
static long get_value();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double ();
static long get_control_double ();
#define get_alarm_double NULL

struct rset tankRSET = {

    RSETNUMBER,
    report,
    initialize,
    init_record,
    process,
    special,
    get_value,
    cvt_dbaddr,
    get_array_info,
    put_array_info,
    get_units,
    get_precision,
    get_enum_str,
    get_enum_strs,
    put_enum_str,
    get_graphic_double,
    get_control_double,
    get_alarm_double
    };

/* some static funcs */
static void get_link_value (char *recname, char *recfld, float *value);

/* The folowing definitions are for redundancy */
 
static struct fldDes **mylist = 0;

   /* NOTE: the order and spelling in namelist must be the
            same as in tankredfldidx.h
   */

static char *namelist [] = {
               "STAT",
               "SEVR",
               "VAL",
               "V01M",
               "V02M",
               "V03M",
               "V04M",
               "V05M",
               "V06M",
               "V07M",
               "V08M",
               "V09M",
               "V10M",
               "APID",
               "ESOU",
               "TSTS",
               "SGRV",
               "TASV",
               "MLOS",
               "QVOL",
               "TTIM",
               "VTIM",
               "LALM",
               "LLAL",
               "LTAL",
               "LBAL",
               "TVAL",
               "LVHH",
               "LVHI",
               "LVLO",
               "LVLL",
	       "PCMD",
	       "INI2",
	       "CMDS",
	       "CISS",
	       "CREC",
	       "CRUN",
	       "CSTP",
	       "SISS",
	       "SREC",
	       "SRUN",
	       "SSTP",
	       "LATM",
	       "MLST",
               0
               };

static   unsigned short type;       /* index to tank type */

struct tank_trips
{
    int	hi_level_trip;
    int	low_level_trip;
    int	hi_pressure_trip;
    int	low_pressure_trip;
};

struct tankred
{
    unsigned short	stat;
    unsigned short	sevr;
    double		val;
    unsigned short	v01m;
    unsigned short	v02m;
    unsigned short	v03m;
    unsigned short	v04m;
    unsigned short	v05m;
    unsigned short	v06m;
    unsigned short	v07m;
    unsigned short	v08m;
    unsigned short	v09m;
    unsigned short	v10m;
    char		apid[10];
    short		esou;
    char		tsts[18];
    double		sgrv;
    short		tasv;
    unsigned short	mlos;
    double		qvol;
    double		ttim;
    double		vtim;
    double		lalm;
    double		llal;
    double		ltal;
    double		lbal;
    unsigned long	tcks;
    short		tval;
    double		lvhh;
    double		lvhi;
    double		lvlo;
    double		lvll;
    char		pcmd[18];
    double		plev;
    double		bswl;
    double		bswv;
    double		pbsw;
    double		tpav;
    double		dnct;
    double		gcap;
    double		ullg;
    double		avap;
    double		wcap;
    double		netv;
    double		texp;
    double		gvol;
    double		frat;
    char		rsts[18];
    short		esin;
    char		cmds[18];
    short		ciss;
    short		crec;
    short		crun;
    short		cstp;
    short		siss;
    short		srec;
    short		srun;
    short		sstp;
    short		ini2;
    TS_STAMP		latm;
    double		mlst;
};

/* defines for tank commands
   NOTE: these must have these values to
         coincide with tankRecord.ascii
*/

#define  TANK_ISSUE_CMD    1
#define  TANK_RECEIVE_CMD  2
#define  TANK_RUN_CMD      3
#define  TANK_STOP_CMD     4

/* tank operational status */

#define TANK_ISSUING    0
#define TANK_RECEIVING  1
#define TANK_RUNNING    2
#define TANK_STOPPED    3

/* trip sensors */

#define TRIP_NOT_ACTIVE   0
#define TRIP_ACTIVE       1

/* these must be in sync with tankRecord.<ascii,h> */

#define MAX_TANK_VALVES 10
#define MAX_TEMP_PROBES 10

/* defines for which level values were gotten
   from the level pv
*/

#define NO_LEVEL_CHGS    0
#define HIHI_LEVEL_CHG   1
#define HIGH_LEVEL_CHG   2
#define LOW_LEVEL_CHG    4
#define LOLO_LEVEL_CHG   8


static void alarm (struct tankRecord *,
		   struct tank_trips *);
static long check_mode ();
static long compute_some_values ();
static long emergency_stop ();
static long get_level_values ();
static long init_valves ();
static long lock_unlock_valves ();
static void monitor ();
static void post_events ();
static long read_bswpv ();
static long read_levelpv (struct tankRecord *,
			  struct tank_trips *);
static long read_pressurepv (struct tankRecord *,
			     struct tank_trips *);
static long read_sensors (struct tankRecord *, 
			  unsigned long *, 
			  unsigned long *, 
			  struct tank_trips *);
static long read_temppv ();
static void reset_tank_commands ();
static long run_commands ();
static void save_tank ();
static void set_roof_status ();
static void update_redun ();
static void update_qvol ();
static void update_vtim ();
static void valve_status_alarms ();


/****************************************************************************
 *
 *  MODULE NAME            : init_record 
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 apr 93
 *
 *  ARGUMENT(S)            : ptank - ptr to tank record  
 *
 *  DESCRIPTION: initialize a tank record
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION: init PCMD (last command), TF_INIT_CMD defined in
 *            .../share/epicsH/tankfarm.h
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 03 Jan 94
 *  REVISION:  set ack to 0
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 23 Feb 94
 *  REVISION: removed ack = 0, db_post_events only if master
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 May 94
 *  REVISION:	init ttim to -1 not 0, 0 will cause the tank to start timing
 *		after init
 *
****************************************************************************/

static long init_record (ptank)
   struct tankRecord	*ptank;
{
    long		status = 0;
    unsigned short	i;
    char		*pstr;
    unsigned long	curr_ticks = 0;
    unsigned long	last_ticks = 0;
    int			ini2_flag = 0;
    struct tankred	tank_old;
    struct tank_trips	tankTrips;

    if (!(mylist))
    {
         /* find record type for tank */
	for (type = 1, i = 0; type < 100; type++)
        {
            pstr = GET_PRECTYPE (type);     
            if (strncmp (pstr, "tank", 4) == 0)
            {
                i = 1;
                break;
            }
        }

        if (i == 0)
        {
            recGblRecordError (S_db_noRecType, (struct dbCommon *)ptank, 
					    "tank:init_record");
            return (S_db_noRecType);
        }
         /* allocate memory for mylist */
        mylist = redFldAlloc (type, NUMTANKFLDS);

        if (!mylist)
        {
            recGblRecordError (S_rec_Alloc, (struct dbCommon *)ptank,
			   "tank, init_record, mylist failed");
	    return (S_rec_Alloc);
        }

         /* add fields in namelist to mylist */
        for (i = 0; i < NUMTANKFLDS; i++)
            redFldAdd (type, i, namelist [i]);
    }

      /* now find the strapping table for this record 
         but 1st make sure ptr to strapping table is NULL */

    ptank -> pstr = NULL;
    status = FindStrappingTable (ptank -> sttn, &ptank -> pstr, (int *)&ptank -> snum);
    if (status)
	return (status);

      /* now find the API table for this record
         but 1st make sure ptr to API table is NULL */

    ptank -> papi = NULL;
    status = FindAPITable (ptank -> apit, &ptank -> papi, (int *)&ptank -> anum);
    if (status)
	return (status);


    save_tank (ptank, &tank_old);
    ptank -> init = TRUE;
    ptank -> fpps = TRUE;
    ptank -> nsta = NO_ALARM;
    ptank -> nsev = NO_ALARM;
    ptank -> msev = NO_ALARM;
    /* init last time alarmed and last process time */
    if (ptank -> latm.secPastEpoch == 0)
	tsLocalTime (&ptank -> latm);
    if (ptank -> lptm.secPastEpoch == 0)
	tsLocalTime (&ptank -> lptm);

      /* set tank ops state to stopped */

    strncpy (ptank -> tsts, ptank -> stps, sizeof (ptank -> tsts));
    ptank -> tasv = TANK_STOPPED;

      /* init some stuff */

    ptank -> val  = 0;
    ptank -> esin = 0;
    ptank -> esou = 0;
    ptank -> plev = 0;
    ptank -> qvol = 0;
    ptank -> bswl = 0;
    ptank -> pbsw = 0;
    ptank -> gcap = 0;
    ptank -> gvol = 0;
    ptank -> netv = 0;
    ptank -> avap = 0;
    ptank -> ullg = 0;
    ptank -> wcap = 0;
    ptank -> gmas = 0;
    ptank -> nmas = 0;
    ptank -> dnct = 0;
    ptank -> dncr = 0;
    ptank -> frat = 0;
    ptank -> texp = 0;
    ptank -> tpav = 0;
    ptank -> ttim = -1;
    ptank -> vtim = 0;
    ptank -> mlos = ptank -> mode;
    bzero (ptank -> cmds, sizeof (ptank -> cmds));
    ptank -> ciss = 0;
    ptank -> crec = 0;
    ptank -> crun = 0;
    ptank -> cstp = 0;
    ptank -> tcks = 0;
    ptank -> prec = 2;    /* display precision */
    ptank -> tval = 0;    /* tank is in volume alarm condition */
    ptank -> lavl = NO_LEVEL_CHGS;
    strncpy (ptank -> pcmd, TF_INIT_CMD, sizeof (ptank -> pcmd));

      /* get level alarm values from promted fields and place in
         in non-prompted fields, so when update non prompted
         do not have to update checksum and Unix data base will
         maintain original value */

    ptank -> lvhh = ptank -> hhlv;
    ptank -> lvhi = ptank -> hilv;
    ptank -> lvlo = ptank -> lolv;
    ptank -> lvll = ptank -> lllv;

       /* init the valve pv dbAddr's */

    status = init_valves (ptank);
    if (status)
    {
	ptank -> nsta = NO_INIT_ALARM;
	ptank -> nsev = VALID_ALARM;
    }
    else
    {
	/* read all sensors and compute values */

	status = read_sensors (ptank, &curr_ticks, &last_ticks, &tankTrips);
	if (status)
	{
	    ptank -> nsta = NO_INIT_ALARM;
	    ptank -> nsev = VALID_ALARM;
	}
	else
	{
	    ini2_flag = 1;

	       /* set quiescent volume to tank net volume, since we 
		  are stopped 
	       */
	    ptank -> qvol = ptank -> netv;

	       /* set roof status here because current product level
	          is needed
	       */
	    set_roof_status (ptank);
	}
    }

      /* do redundancy if not in slave mode */

    if (resGetCaAnswer (ptank -> radr, &i) == RES_OK)
    {
	if (i)
	{
	    /* check for maximize severity */

	    if (ptank -> msev > ptank -> nsev)
	    {
		ptank -> nsev = ptank -> msev;
		ptank -> nsta = LINK_ALARM;
	    }

	    monitor (ptank);

	    if (ini2_flag)
	    {
		ptank -> ini2 = 1;
		post_events (ptank, &tank_old);
		update_redun (&tank_old, ptank);
	    }
	}
    }

    return (0);

} /* end func init_records () */


/****************************************************************************
 *
 *  MODULE NAME            : alarm 
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 Apr 93
 *
 *  CALLING MODULE(S)      : process ()
 *  ARGUMENT(S)            : ptank
 *
 *  DESCRIPTION: set alarm status
 *               1) get absolute value of diff tween last value alarmed
 *                  and the current value = delta
 *               2) if delta < deadband value, use last value alarmed
 *                  to compare against alarm condition
 *               3) if alarm condition exists and an emergency stop is
 *                  is indicated set local variable emergecy_stop,
 *                  at end of alarm () set ptank -> esou if emergency_stop
 *                  was set
 *               4) update nsta and nsev only if alarm severity is greater
 *                  than current nsev
 *  NOTE:
 *       Do not change order of alarm testing without consultation !!!!
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 Mar 94
 *  REVISION: alarm chgs, if val is past alarm setting, set that
 *            alarm condition, else if val was in that alarm and
 *            it is still in hyst range of alarm value, then stay
 *            in alarm, also changed names of LEVEL and TEMPERATURE alarms
 *
****************************************************************************/

static void alarm (ptank, tankTrips)
   struct tankRecord    *ptank;
   struct tank_trips	*tankTrips;
{
   double	delta;
   int		estop = 0;    /* init emergency stop */
   int		adis = 0;

       /* check for inhibited alarms */

   adis = AlarmDisableLink ((void *)ptank);
   if ( (ptank -> aenb == ALARM_DISABLED) || (adis != 0) )
   {
      ptank -> nsta = INHIBIT_ALARM;
      ptank -> nsev = NO_ALARM;
      return;
   }

      /* if val is undefined */

   if (ptank -> udf == TRUE)
   {
      if (ptank -> nsev < VALID_ALARM)
      {
         ptank -> nsta = UDF_ALARM;
         ptank -> nsev = VALID_ALARM;
      }
      return;
   }

   if (ptank -> nsev == VALID_ALARM)
      return;

      /* check for volume chg alarm if tank is stopped
         val  - current tank volume
         alsv - volume chg severity
         qvol - volume when stopped
         vold - volume change value
         iems - alarm inits emergency shutdown
         hyst - volume change deadband
         first see if we are timing to set quiescent volume
      */

   if (ptank -> tasv == TANK_STOPPED)
   {
         /* we are not timing if < 0 */
      
      if (ptank -> ttim < 0.0)
      {
         delta = ptank -> qvol - ptank -> netv;
         if (delta < 0.0)
            delta = -delta;
         if ( (delta > ptank -> vold) ||
              ( (ptank -> stat == VOLUME) &&
                (delta > (ptank -> vold - ptank -> hyst)) ) )
         {
            if (ptank -> nsev < ptank -> alsv)
            {
               ptank -> nsta = VOLUME;
               ptank -> nsev = ptank -> alsv;
            }
            if (ptank -> iems)
            {
               estop = 1;
            }

               /* reset quiescent volume and start timing again
                  to see if the tank is still changing volume
                  by more than vold, also set tval to indicate
                  tank is in volume alarm condition
               */

            ptank -> qvol = ptank -> netv;
            ptank -> ttim = 0;
            ptank -> tval = 1;
         }
         else
         {
            ptank -> tval = 0;
         }
      }
         /* if the tank was in a volume alarm, continue the alarm
            while we are timing to settle time
         */
      else
      {
         if (ptank -> tval)
         {
            if (ptank -> nsev < ptank -> alsv)
            {
               ptank -> nsta = VOLUME;
               ptank -> nsev = ptank -> alsv;
            }
            if (ptank -> iems)
            {
               estop = 1;
            }
         }
      }
   }

      /* check for the hi trip level
         hlse - hi trip severity
         hlsx - alarm inits emergency shutdown
      */

   if (tankTrips -> hi_level_trip == TRIP_ACTIVE)
   {
      if (ptank -> nsev < ptank -> hlse)
      {
         ptank -> nsta = HIGH_LEVEL;
         ptank -> nsev = ptank -> hlse;
      }   
      if (ptank -> hlsx)
      {
         estop = 1;
      }
   }

      /* current change in level
         llal - last level alarmed
         plev - current product level
         lhst - level chg deadband
      */

      /* check for high high level alarm
         lvhh - hi hi alarm value
         hhle - alarm severity
         hhlx - inits emergency shutdown
      */

   if ( (ptank -> plev > ptank -> lvhh) ||
	( (ptank -> stat == HIHI_LEVEL) &&
	  (ptank -> plev > (ptank -> lvhh - ptank -> lhst)) ) )
   {
      if (ptank -> nsev < ptank -> hhle)
      {
         ptank -> llal = ptank -> plev;
         ptank -> nsev = ptank -> hhle;
         ptank -> nsta = HIHI_LEVEL;
      }
      if (ptank -> hhlx)
      {
         estop = 1;
      }
   }

      /* check for high level alarm
         lvhi - high alarm value
         hile - alarm severity
         hilx - inits emergency shutdown
      */

   if ( (ptank -> plev > ptank -> lvhi) ||
        ( (ptank -> stat == HIGH_LEVEL) &&
          (ptank -> plev > (ptank -> lvhi - ptank -> lhst)) ) )
   {
      if (ptank -> nsev < ptank -> hile)
      {
         ptank -> llal = ptank -> plev;
         ptank -> nsev = ptank -> hile;
         ptank -> nsta = HIGH_LEVEL;
      }
      if (ptank -> hilx)
      {
         estop = 1;
      }
   }

      /* check for low trip level
         llse - low trip severity
         llsx - inits emergency shutdown
      */

   if (tankTrips -> low_level_trip == TRIP_ACTIVE)
   {
      if (ptank -> nsev < ptank -> llse)
      {
         ptank -> nsev = ptank -> llse;
         ptank -> nsta = LOW_LEVEL;
      }
      if (ptank -> llsx)
      {
         estop= 1;
      }
   }

      /* check for low low level alarm
         lvll - low low alarm value
         llle - alarm severity
         lllx - inits emergency shutdown
      */

   if ( (ptank -> plev < ptank -> lvll) ||
        ( (ptank -> stat == LOLO_LEVEL) &&
          (ptank -> plev < (ptank -> lvll + ptank -> lhst)) ) )
   {
      if (ptank -> nsev < ptank -> llle)
      {
         ptank -> llal = ptank -> plev;
         ptank -> nsev = ptank -> llle;
         ptank -> nsta = LOLO_LEVEL;
      }
      if (ptank -> lllx)
      {
         estop = 1;
      }
   }

      /* check for low level alarm
         low  - low alarm value
         lole - alarm severity
         lolx - inits emergency shutdown
      */

   if ( (ptank -> plev < ptank -> lvlo) ||
        ( (ptank -> stat == LOW_LEVEL) &&
          (ptank -> plev < (ptank -> lvlo + ptank -> lhst)) ) )
   {
      if (ptank -> nsev < ptank -> lole)
      {
         ptank -> llal = ptank -> plev;
         ptank -> nsev = ptank -> lole;
         ptank -> nsta = LOW_LEVEL;
      }
      if ( ptank -> lolx)
      {
         estop = 1;
      }
   }

      /* check on tank valve status alarm conditions
         first check to see if we are still timing the
         transition state
      */

   if (ptank -> vtim < 0.0)
      valve_status_alarms (ptank, &estop);

      /* check for BSW alarm
         hibe - alarm severity
         hibx - inits emergency shutdown
         hibv - bsw alarm value
         bhst - bsw chg deadband
         lbal - last bsw alarmed
         bswl - current bsw level
      */

   if ( (ptank -> bswl > ptank -> hibv) ||
        ( (ptank -> stat == BSW) &&
          (ptank ->bswl > (ptank -> hibv - ptank -> bhst)) ) )
   {
      if (ptank -> nsev < ptank -> hibe)
      {
         ptank -> lbal = ptank -> bswl;
         ptank -> nsev = ptank -> hibe;
         ptank -> nsta = BSW;
      }
      if (ptank -> hibx)
      {
         estop = 1;
      }
   }
      /* current change in average temperatue
         tpav - current average temperature
         ltal - last temperature alarmed
         thst - temperature change deadband
      */

      /* check for high high temperature alarm
         hhte - alarm severity
         hhtv - high high alarm value
         hhtx - inits emergency shutdown
      */

   if ( (ptank -> tpav > ptank -> hhtv) ||
        ( (ptank -> stat == HIHI_TEMPERATURE) &&
          (ptank -> tpav > (ptank -> hhtv - ptank -> thst)) ) )
   {
      if (ptank -> nsev < ptank -> hhte)
      {
         ptank -> ltal = ptank -> tpav;
         ptank -> nsev = ptank -> hhte;
         ptank -> nsta = HIHI_TEMPERATURE;
      }
      if (ptank -> hhtx)
      {
         estop = 1;
      }
   }
      /* check for high temperature alarm
         hitv - high alarm value
         hite - alarm severity
         hitx - inits emergency shutdown
      */

   if ( (ptank -> tpav > ptank -> hitv) ||
        ( (ptank -> stat == HIGH_TEMPERATURE) &&
          (ptank -> tpav > (ptank -> hitv - ptank -> thst)) ) )
   {
      if (ptank -> nsev < ptank -> hite)
      {
         ptank -> ltal = ptank -> tpav;
         ptank -> nsev = ptank -> hite;
         ptank -> nsta = HIGH_TEMPERATURE;
      }
      if (ptank -> hitx)
      {
         estop = 1;
      }
   }
      /* check for low low temperature alarm
         llte - alarm severity
         lltx - inits emergency shutdown
         lltv - low low alarm value
      */

   if ( (ptank -> tpav < ptank -> lltv) ||
        ( (ptank -> stat == LOLO_TEMPERATURE) &&
          (ptank -> tpav < (ptank -> lltv + ptank -> thst)) ) )
   {
      if (ptank -> nsev < ptank -> llte)
      {
         ptank -> ltal = ptank -> tpav;
         ptank -> nsev = ptank -> llte;
         ptank -> nsta = LOLO_TEMPERATURE;
      }
      if (ptank -> lltx)
      {
         estop = 1;
      }
   }
      /* check for low temperature alarm
         lotv - low alarm value
         lote - alarm severity
         lotx - inits emergency shutdown
      */

   if ( (ptank -> tpav < ptank -> lotv) ||
        ( (ptank -> stat == LOW_TEMPERATURE) &&
          (ptank -> tpav < (ptank -> lotv + ptank -> thst)) ) )
   {
      if (ptank -> nsev < ptank ->lote)
      {
         ptank -> ltal = ptank -> tpav;
         ptank -> nsev = ptank -> lote;
         ptank -> nsta = LOW_TEMPERATURE;
      }
      if (ptank -> lotx)
      {
         estop = 1;
      }
   }
      /* check for high pressure trip alarm
         hipe - alarm severity
         hipx - inits emergency shutdown
      */

   if (tankTrips -> hi_pressure_trip == TRIP_ACTIVE)
   {
      if (ptank -> nsev < ptank -> hipe)
      {
         ptank -> nsev = ptank -> hipe;
         ptank -> nsta = PRESSURE;
      }
      if (ptank -> hipx)
      {
         estop = 1;
      }
   }
      /* check for low pressure trip alarm
         lope - alarm severity
         lopx - inits emergency shutdown
      */

   if (tankTrips -> low_pressure_trip == TRIP_ACTIVE)
   {
      if (ptank -> nsev < ptank -> lope)
      {
         ptank -> nsev = ptank -> lope;
         ptank -> nsta = PRESSURE;
      }
      if (ptank -> lopx)
      {
         estop = 1;
      }
   }

      /* if any alarm generated an emergency stop action for the tank,
         stop the tank and set tanks associated valves ESIN - 
         emergency stop input
      */

   if (estop)
   {
      ptank -> esou = 1;
      emergency_stop (ptank);
      reset_tank_commands (ptank);
   }
   else
      ptank -> esou = 0;

   return;

} /* end func alarm () */


/****************************************************************************
 *
 *  MODULE NAME            : check_mode 
 *  AUTHOR                 : emsteel
 *  DATE                   : 25 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *
 *  DESCRIPTION: if the tank is LOCKED, then the tanks associated valves
 *               must also be LOCKED if not already locked, if the tank
 *               is not LOCKED and the valves are, then they must be
 *               UNLOCKED
 *               ptank -> mode = current tank mode
 *               ptank -> mlos = previous tank mode
 *
****************************************************************************/

static long check_mode (ptank)
   struct tankRecord   *ptank;
{
   long  status = 0;

   if (ptank -> mode == TF_LOCKED)
   {
      if (ptank -> mlos != TF_LOCKED)
      {
         status = lock_unlock_valves (1);
      }
   }
   else
   {
      if (ptank -> mlos == TF_LOCKED)
      {
         status = lock_unlock_valves (0);
      }
   }

   ptank -> mlos = ptank -> mode;

   return (status);

} /* end func check_mode () */


/****************************************************************************
 *
 *  MODULE NAME            : compute_some_values 
 *  AUTHOR                 : emsteel
 *  DATE                   : 15 Apr 93
 *
 *  CALLING MODULE(S)      : read_senors
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *                           orig_volume - ptr to volume at scan start
 *                           last_time - ptr to TS_STAMP last scan
 *
 *  DESCRIPTION: do some cals
 *
 *  NOTE: some calcs performed in the read... funcs
 *        future calcs will be:
 *                gmas - gross product mass
 *                nmas - net product mass
 *                dncr - corrected density
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 04 Jan 94
 *  REVISION:  compute gvol here instead of in read_bswpv
 *
****************************************************************************/

static long compute_some_values (ptank, orig_volume, curr_ticks, last_ticks)
   struct tankRecord    *ptank;
   double               orig_volume;
   unsigned long        *curr_ticks;
   unsigned long        *last_ticks;
{
   double   lolo_vol = 0;     /* low low volume from strapping table */
   double   lo_vol = 0;       /* low volume, from strapping table */
   long     status = 0;
   double   frat = 0;


       /* gross volume = current volume - bottom sediment volume */

    ptank -> gvol = ptank -> val - ptank -> bswv;

      /* use high high level value to compute the gross capacity
         of the tank, use it to compute ullage
      */

   ComputeTFVolume (ptank -> pstr, ptank -> snum, ptank -> lvhh,
                                                         &ptank -> gcap);
   if (ptank -> gcap < 0.0)
      ptank -> gcap = 0.0;
   ptank -> ullg = ptank -> gcap - ptank -> val;
   if (ptank -> ullg < 0.0)
      ptank -> ullg = 0.0;

      /* use strapping table to compute lolo_vol lo_vol */

   ComputeTFVolume (ptank -> pstr, ptank -> snum, ptank -> lvll, &lolo_vol);
   ComputeTFVolume (ptank -> pstr, ptank -> snum, ptank -> lvlo, &lo_vol);

      /* available product:
         current product volume - low volume
      */

   ptank -> avap = ptank -> val - lo_vol;

      /* working tank capacity:
         gcap - lolo_vol
      */

   ptank -> wcap = ptank -> gcap - lolo_vol;

      /* get the API thermal expansion factor, ptank->texp,
            before we compute net volume, ptank->netv
         net volume:
            gross product volume (gvol) * thermal expansion API (texp)
         gvol = val - bswv
      */

   if (ptank -> papi)
   {
      status = FindAPIcf (ptank -> papi, ptank -> anum, ptank -> tpav,
                          ptank -> sgrv, &ptank -> texp, ptank -> name,
                          "TMPU", ptank -> tmpu);
      if (!status)
      {
         ptank -> netv = ptank -> gvol * ptank -> texp;
      }
      else
	 return (status);
   }


   if (ptank -> ini2)
   {
      status = ComputeTFFlowRate (ptank -> name, "VOLU", "FLRU", ptank -> volu, 
				  ptank -> flru, ptank -> val, orig_volume,
                                  curr_ticks, last_ticks, &frat);
      if (!status)
      {
	 ptank -> frat = frat;
      }
   }

   return (status);
   
} /* end func compute_flow_rate () */


/****************************************************************************
 *
 *  MODULE NAME            : emergency_stop 
 *  AUTHOR                 : emsteel
 *  DATE                   : 13 Aug 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *
 *  DESCRIPTION: stop the tank, if not already stopped and set its
 *               associated valves esin field so valve will start its
 *               emergency stop action
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 27 Feb 95
 *  REVISION:	corrected arg list to dbPutField
 *
****************************************************************************/

static long emergency_stop (ptank)
   struct tankRecord    *ptank;
{
   int      i;
   long     status    = 0;
   char     *pvname   = &ptank -> v01p[0];     /* name of 1st valve  */
   short    esin      = 1;
   long     nRequests = 1;
   char     pvname_fld[36];
   char     errmsg[128];
   struct   dbAddr   dbAddr;

   if (ptank -> tasv != TANK_STOPPED)
   {
      strncpy (ptank -> tsts, ptank -> stps, sizeof (ptank -> tsts));
      ptank -> tasv = TANK_STOPPED;
      ptank -> vtim = 0;
      ptank -> ttim = 0;
      ptank -> tval = 0;
   }

      /* resolve valve name and field ESIN to struct dbAddr,
         for use by dbPutField
      */

   for (i = 0; i < MAX_TANK_VALVES; i++)
   {
      if (*pvname)
      {
         strncpy (pvname_fld, pvname, 28);
         strcat (pvname_fld, ".ESIN");
         if ((status = dbNameToAddr (pvname_fld, &dbAddr)) != 0)
         {
            sprintf (errmsg, "tank: PV:%.28s,valve:%.28s,esin",
                              ptank -> name, pvname);
            errMessage (S_tank_TankValve_Err, errmsg);
            return (S_tank_TankValve_Err);
         }
         if ((status = dbPutField (&dbAddr, DBR_SHORT, (caddr_t)&esin, 
                                  nRequests)) != 0)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM,
                                                               VALID_ALARM);
            return (status);
         }
      }
      pvname += 30;
   }

   return (status);

} /* end func emergency_stop () */


/****************************************************************************
 *
 *  MODULE NAME            : get_control_double 
 *  AUTHOR                 : emsteel
 *  DATE                   : 31 Aug 94
 *
 *
 *  DESCRIPTION: return control limits
 *
****************************************************************************/

static long get_control_double(paddr,pcd)
    struct	dbAddr		*paddr;
    struct	dbr_ctrlDouble	*pcd;
{
    struct tankRecord    *ptank = (struct tankRecord *) paddr->precord;
    struct strappingValue   *stbl;
    double                  top_strap_vol = 0;
    double		    bottom_strap_vol = 0;
    double		    top_strap_level = 0;
    double		    bottom_strap_level = 0;
    float		    hopr;
    float		    lopr;

    if (ptank -> pstr)
    {
	stbl = (struct strappingValue *)ptank -> pstr;
	top_strap_vol = (stbl + (ptank -> snum - 1)) -> volume;
	top_strap_level = (stbl + (ptank -> snum - 1)) -> sensorValue;
	bottom_strap_vol = stbl -> volume;
	bottom_strap_level = stbl -> sensorValue;
    }

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct tankRecord *)0) -> val) ):
	case ( (int)&(((struct tankRecord *)0) -> bswv) ):
	case ( (int)&(((struct tankRecord *)0) -> gcap) ):
	case ( (int)&(((struct tankRecord *)0) -> gvol) ):
	case ( (int)&(((struct tankRecord *)0) -> netv) ):
	case ( (int)&(((struct tankRecord *)0) -> avap) ):
	case ( (int)&(((struct tankRecord *)0) -> wcap) ):

	    /* if there is a strapping table, limit is bottom and top entry */
	    if (ptank -> pstr)
	    {
		pcd -> upper_ctrl_limit = top_strap_vol;
		pcd -> lower_ctrl_limit = bottom_strap_vol;
	    }
	    else
	    {
		pcd -> upper_ctrl_limit = 1000000;
		pcd -> lower_ctrl_limit = 0;
	    }
	    break;

	case ( (int)&(((struct tankRecord *)0) -> stto) ):
	case ( (int)&(((struct tankRecord *)0) -> isto) ):
	case ( (int)&(((struct tankRecord *)0) -> reto) ):
	case ( (int)&(((struct tankRecord *)0) -> ruto) ):
	case ( (int)&(((struct tankRecord *)0) -> tset) ):
	    pcd -> upper_ctrl_limit = 86400;	/* secs/day */
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> sgrv) ):
	case ( (int)&(((struct tankRecord *)0) -> pbsw) ):
	    pcd -> upper_ctrl_limit = 100;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> vold) ):
	case ( (int)&(((struct tankRecord *)0) -> adel) ):
	case ( (int)&(((struct tankRecord *)0) -> mdel) ):
	case ( (int)&(((struct tankRecord *)0) -> hyst) ):
	case ( (int)&(((struct tankRecord *)0) -> ullg) ):
	    if (ptank -> pstr)
		hopr = top_strap_vol;
	    else if (ptank -> gcap > 0)
		hopr = ptank -> gcap;
	    else
		hopr = 1000000;
	    pcd -> upper_ctrl_limit = hopr;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> lllv) ):
	case ( (int)&(((struct tankRecord *)0) -> lolv) ):
	case ( (int)&(((struct tankRecord *)0) -> hhlv) ):
	case ( (int)&(((struct tankRecord *)0) -> hilv) ):
	case ( (int)&(((struct tankRecord *)0) -> plev) ):
	case ( (int)&(((struct tankRecord *)0) -> plva) ):
	case ( (int)&(((struct tankRecord *)0) -> blva) ):
	case ( (int)&(((struct tankRecord *)0) -> rllv) ):
	case ( (int)&(((struct tankRecord *)0) -> rflv) ):
	case ( (int)&(((struct tankRecord *)0) -> htrp) ):
	case ( (int)&(((struct tankRecord *)0) -> ltrp) ):
	case ( (int)&(((struct tankRecord *)0) -> t01l) ):
	case ( (int)&(((struct tankRecord *)0) -> t02l) ):
	case ( (int)&(((struct tankRecord *)0) -> t03l) ):
	case ( (int)&(((struct tankRecord *)0) -> t04l) ):
	case ( (int)&(((struct tankRecord *)0) -> t05l) ):
	case ( (int)&(((struct tankRecord *)0) -> t06l) ):
	case ( (int)&(((struct tankRecord *)0) -> t07l) ):
	case ( (int)&(((struct tankRecord *)0) -> t08l) ):
	case ( (int)&(((struct tankRecord *)0) -> t09l) ):
	case ( (int)&(((struct tankRecord *)0) -> t10l) ):
	case ( (int)&(((struct tankRecord *)0) -> lvhh) ):
	case ( (int)&(((struct tankRecord *)0) -> lvhi) ):
	case ( (int)&(((struct tankRecord *)0) -> lvlo) ):
	case ( (int)&(((struct tankRecord *)0) -> lvll) ):

	    /* for levels use hopr and lopr of level pv, else get from 
	       strapping table, else default
	    */
	    hopr = lopr = 0;
	    if (ptank -> levp.type == DB_LINK)
	    {
		get_link_value (ptank -> levp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> levp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
	    	if (ptank -> pstr)
	    	{
		    hopr = top_strap_level;
		    lopr = bottom_strap_level;
	    	}
	    	else
	    	{
		    hopr = 1000000;
		    lopr = 0;
	    	}
	    }
	    pcd -> upper_ctrl_limit = hopr;
	    pcd -> lower_ctrl_limit = lopr;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> lltv) ):
	case ( (int)&(((struct tankRecord *)0) -> lotv) ):
	case ( (int)&(((struct tankRecord *)0) -> hhtv) ):
	case ( (int)&(((struct tankRecord *)0) -> hitv) ):
	case ( (int)&(((struct tankRecord *)0) -> tpav) ):

	    /* get hopr and lopr of ave temp pv, else from 1st temp probe
	       else default
	    */
	    hopr = lopr = 0;
	    if (ptank -> tmpp.type == DB_LINK)
	    {
		get_link_value (ptank -> tmpp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> tmpp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
	    	if (ptank -> t01v.type == DB_LINK)
	    	{
		    get_link_value (ptank -> t01v.value.db_link.pvname, ".HOPR", &hopr);
		    get_link_value (ptank -> t01v.value.db_link.pvname, ".LOPR", &lopr);
	    	}
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		hopr = 300;
		lopr = 0;
	    }
	    pcd -> upper_ctrl_limit = hopr;
	    pcd -> lower_ctrl_limit = lopr;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> hibv) ):
	case ( (int)&(((struct tankRecord *)0) -> bswl) ):

	    /* get from bswp hopr and lopr, else from strapping table
	       else default
	    */
	    hopr = lopr = 0;
	    if (ptank -> bswp.type == DB_LINK)
	    {
		get_link_value (ptank -> bswp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> bswp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
	    	if (ptank -> pstr)
	    	{
		    hopr = top_strap_level;
		    lopr = bottom_strap_level;
	    	}
	    	else
	    	{
		    hopr = 1000000;
		    lopr = 0;
	    	}
	    }
	    pcd -> upper_ctrl_limit = hopr;
	    pcd -> lower_ctrl_limit = lopr;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> esin) ):
	case ( (int)&(((struct tankRecord *)0) -> esou) ):
	    pcd -> upper_ctrl_limit = 1;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> lhst) ):
	    hopr = lopr = 0;
	    if (ptank -> levp.type == DB_LINK)
	    {
		get_link_value (ptank -> levp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> levp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		if (ptank -> pstr)
		{
		    hopr = top_strap_level;
		    lopr = bottom_strap_level;
		}
		else
		{
		    hopr = 1000000;
		    lopr = 0;
		}
	    }
	    pcd -> upper_ctrl_limit = hopr - lopr;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> thst) ):
	    hopr = lopr = 0;
	    if (ptank -> tmpp.type == DB_LINK)
	    {
		get_link_value (ptank -> tmpp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> tmpp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
	    	if (ptank -> t01v.type == DB_LINK)
	    	{
		    get_link_value (ptank -> t01v.value.db_link.pvname, ".HOPR", &hopr);
		    get_link_value (ptank -> t01v.value.db_link.pvname, ".LOPR", &lopr);
	    	}
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		hopr = 300;
		lopr = 0;
	    }
	    pcd -> upper_ctrl_limit = hopr - lopr;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> bhst) ):
	    hopr = lopr = 0;
	    if (ptank -> bswp.type == DB_LINK)
	    {
		get_link_value (ptank -> bswp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> bswp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		if (ptank -> pstr)
		{
		    hopr = top_strap_level;
		    lopr = bottom_strap_level;
		}
		else
		{
		    hopr = 1000000;
		    lopr = 0;
		}
	    }
	    pcd -> upper_ctrl_limit = hopr - lopr;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> prec) ):
	    pcd -> upper_ctrl_limit = 10;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> dnct) ):
	    hopr = lopr = 0;
	    if (ptank -> denp.type == DB_LINK)
	    {
		get_link_value (ptank -> denp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> denp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		hopr = 1000000;
		lopr = 0;
	    }
	    pcd -> upper_ctrl_limit = hopr;
	    pcd -> lower_ctrl_limit = lopr;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> frat) ):
	case ( (int)&(((struct tankRecord *)0) -> htln) ):
	case ( (int)&(((struct tankRecord *)0) -> diam) ):
	case ( (int)&(((struct tankRecord *)0) -> elev) ):
	case ( (int)&(((struct tankRecord *)0) -> rwgt) ):
	    pcd -> upper_ctrl_limit = 1000000;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> texp) ):
	    pcd -> upper_ctrl_limit = 2;
	    pcd -> lower_ctrl_limit = 0;
	    break;
    }

    return (0);

} /* end func get_control_double () */


/****************************************************************************
 *
 *  MODULE NAME            : get_graphic_double 
 *  AUTHOR                 : emsteel
 *  DATE                   : 31 Aug 94
 *
 *
 *  DESCRIPTION: return graphic limits
 *
****************************************************************************/

static long get_graphic_double(paddr,pgd)
    struct	dbAddr		*paddr;
    struct	dbr_grDouble	*pgd;
{
    struct tankRecord    *ptank = (struct tankRecord *) paddr->precord;
    struct strappingValue   *stbl;
    double                  top_strap_vol = 0;
    double		    bottom_strap_vol = 0;
    double		    top_strap_level = 0;
    double		    bottom_strap_level = 0;
    float		    hopr;
    float		    lopr;

    if (ptank -> pstr)
    {
	stbl = (struct strappingValue *)ptank -> pstr;
	top_strap_vol = (stbl + (ptank -> snum - 1)) -> volume;
	top_strap_level = (stbl + (ptank -> snum - 1)) -> sensorValue;
	bottom_strap_vol = stbl -> volume;
	bottom_strap_level = stbl -> sensorValue;
    }

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct tankRecord *)0) -> val) ):
	case ( (int)&(((struct tankRecord *)0) -> bswv) ):
	case ( (int)&(((struct tankRecord *)0) -> gcap) ):
	case ( (int)&(((struct tankRecord *)0) -> gvol) ):
	case ( (int)&(((struct tankRecord *)0) -> netv) ):
	case ( (int)&(((struct tankRecord *)0) -> avap) ):
	case ( (int)&(((struct tankRecord *)0) -> qvol) ):
	case ( (int)&(((struct tankRecord *)0) -> wcap) ):

	    /* if there is a strapping table, limit is bottom and top entry */
	    if (ptank -> pstr)
	    {
		pgd -> upper_disp_limit = top_strap_vol;
		pgd -> lower_disp_limit = bottom_strap_vol;
	    }
	    else
	    {
		pgd -> upper_disp_limit = 1000000;
		pgd -> lower_disp_limit = 0;
	    }
	    break;

	case ( (int)&(((struct tankRecord *)0) -> v01m) ):
	case ( (int)&(((struct tankRecord *)0) -> v02m) ):
	case ( (int)&(((struct tankRecord *)0) -> v03m) ):
	case ( (int)&(((struct tankRecord *)0) -> v04m) ):
	case ( (int)&(((struct tankRecord *)0) -> v05m) ):
	case ( (int)&(((struct tankRecord *)0) -> v06m) ):
	case ( (int)&(((struct tankRecord *)0) -> v07m) ):
	case ( (int)&(((struct tankRecord *)0) -> v08m) ):
	case ( (int)&(((struct tankRecord *)0) -> v09m) ):
	case ( (int)&(((struct tankRecord *)0) -> v10m) ):
	    pgd -> upper_disp_limit = 2;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> stto) ):
	case ( (int)&(((struct tankRecord *)0) -> isto) ):
	case ( (int)&(((struct tankRecord *)0) -> reto) ):
	case ( (int)&(((struct tankRecord *)0) -> ruto) ):
	case ( (int)&(((struct tankRecord *)0) -> tset) ):
	    pgd -> upper_disp_limit = 86400;	/* secs/day */
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> sgrv) ):
	case ( (int)&(((struct tankRecord *)0) -> pbsw) ):
	    pgd -> upper_disp_limit = 100;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> vold) ):
	case ( (int)&(((struct tankRecord *)0) -> adel) ):
	case ( (int)&(((struct tankRecord *)0) -> mdel) ):
	case ( (int)&(((struct tankRecord *)0) -> hyst) ):
	case ( (int)&(((struct tankRecord *)0) -> ullg) ):
	    if (ptank -> pstr)
		hopr = top_strap_vol;
	    else if (ptank -> gcap > 0)
		hopr = ptank -> gcap;
	    else
		hopr = 1000000;
	    pgd -> upper_disp_limit = hopr;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> lllv) ):
	case ( (int)&(((struct tankRecord *)0) -> lolv) ):
	case ( (int)&(((struct tankRecord *)0) -> hhlv) ):
	case ( (int)&(((struct tankRecord *)0) -> hilv) ):
	case ( (int)&(((struct tankRecord *)0) -> plev) ):
	case ( (int)&(((struct tankRecord *)0) -> plva) ):
	case ( (int)&(((struct tankRecord *)0) -> blva) ):
	case ( (int)&(((struct tankRecord *)0) -> rllv) ):
	case ( (int)&(((struct tankRecord *)0) -> rflv) ):
	case ( (int)&(((struct tankRecord *)0) -> htrp) ):
	case ( (int)&(((struct tankRecord *)0) -> ltrp) ):
	case ( (int)&(((struct tankRecord *)0) -> t01l) ):
	case ( (int)&(((struct tankRecord *)0) -> t02l) ):
	case ( (int)&(((struct tankRecord *)0) -> t03l) ):
	case ( (int)&(((struct tankRecord *)0) -> t04l) ):
	case ( (int)&(((struct tankRecord *)0) -> t05l) ):
	case ( (int)&(((struct tankRecord *)0) -> t06l) ):
	case ( (int)&(((struct tankRecord *)0) -> t07l) ):
	case ( (int)&(((struct tankRecord *)0) -> t08l) ):
	case ( (int)&(((struct tankRecord *)0) -> t09l) ):
	case ( (int)&(((struct tankRecord *)0) -> t10l) ):
	case ( (int)&(((struct tankRecord *)0) -> lvhh) ):
	case ( (int)&(((struct tankRecord *)0) -> lvhi) ):
	case ( (int)&(((struct tankRecord *)0) -> lvlo) ):
	case ( (int)&(((struct tankRecord *)0) -> lvll) ):

	    /* for levels use hopr and lopr of level pv, else get from 
	       strapping table, else default
	    */
	    hopr = lopr = 0;
	    if (ptank -> levp.type == DB_LINK)
	    {
		get_link_value (ptank -> levp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> levp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
	    	if (ptank -> pstr)
	    	{
		    hopr = top_strap_level;
		    lopr = bottom_strap_level;
	    	}
	    	else
	    	{
		    hopr = 1000000;
		    lopr = 0;
	    	}
	    }
	    pgd -> upper_disp_limit = hopr;
	    pgd -> lower_disp_limit = lopr;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> lltv) ):
	case ( (int)&(((struct tankRecord *)0) -> lotv) ):
	case ( (int)&(((struct tankRecord *)0) -> hhtv) ):
	case ( (int)&(((struct tankRecord *)0) -> hitv) ):
	case ( (int)&(((struct tankRecord *)0) -> tpav) ):

	    /* get hopr and lopr of ave temp pv, else from 1st temp probe
	       else default
	    */
	    hopr = lopr = 0;
	    if (ptank -> tmpp.type == DB_LINK)
	    {
		get_link_value (ptank -> tmpp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> tmpp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
	    	if (ptank -> t01v.type == DB_LINK)
	    	{
		    get_link_value (ptank -> t01v.value.db_link.pvname, ".HOPR", &hopr);
		    get_link_value (ptank -> t01v.value.db_link.pvname, ".LOPR", &lopr);
	    	}
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		hopr = 300;
		lopr = 0;
	    }
	    pgd -> upper_disp_limit = hopr;
	    pgd -> lower_disp_limit = lopr;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> hibv) ):
	case ( (int)&(((struct tankRecord *)0) -> bswl) ):

	    /* get from bswp hopr and lopr, else from strapping table
	       else default
	    */
	    hopr = lopr = 0;
	    if (ptank -> bswp.type == DB_LINK)
	    {
		get_link_value (ptank -> bswp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> bswp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
	    	if (ptank -> pstr)
	    	{
		    hopr = top_strap_level;
		    lopr = bottom_strap_level;
	    	}
	    	else
	    	{
		    hopr = 1000000;
		    lopr = 0;
	    	}
	    }
	    pgd -> upper_disp_limit = hopr;
	    pgd -> lower_disp_limit = lopr;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> esin) ):
	case ( (int)&(((struct tankRecord *)0) -> esou) ):
	case ( (int)&(((struct tankRecord *)0) -> tval) ):
	    pgd -> upper_disp_limit = 1;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> lhst) ):
	    hopr = lopr = 0;
	    if (ptank -> levp.type == DB_LINK)
	    {
		get_link_value (ptank -> levp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> levp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		if (ptank -> pstr)
		{
		    hopr = top_strap_level;
		    lopr = bottom_strap_level;
		}
		else
		{
		    hopr = 1000000;
		    lopr = 0;
		}
	    }
	    pgd -> upper_disp_limit = hopr - lopr;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> thst) ):
	    hopr = lopr = 0;
	    if (ptank -> tmpp.type == DB_LINK)
	    {
		get_link_value (ptank -> tmpp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> tmpp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
	    	if (ptank -> t01v.type == DB_LINK)
	    	{
		    get_link_value (ptank -> t01v.value.db_link.pvname, ".HOPR", &hopr);
		    get_link_value (ptank -> t01v.value.db_link.pvname, ".LOPR", &lopr);
	    	}
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		hopr = 300;
		lopr = 0;
	    }
	    pgd -> upper_disp_limit = hopr - lopr;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> bhst) ):
	    hopr = lopr = 0;
	    if (ptank -> bswp.type == DB_LINK)
	    {
		get_link_value (ptank -> bswp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> bswp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		if (ptank -> pstr)
		{
		    hopr = top_strap_level;
		    lopr = bottom_strap_level;
		}
		else
		{
		    hopr = 1000000;
		    lopr = 0;
		}
	    }
	    pgd -> upper_disp_limit = hopr - lopr;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> prec) ):
	    pgd -> upper_disp_limit = 10;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> dnct) ):
	    hopr = lopr = 0;
	    if (ptank -> denp.type == DB_LINK)
	    {
		get_link_value (ptank -> denp.value.db_link.pvname, ".HOPR", &hopr);
		get_link_value (ptank -> denp.value.db_link.pvname, ".LOPR", &lopr);
	    }
	    if ( (hopr == 0) && (lopr == 0) )
	    {
		hopr = 1000000;
		lopr = 0;
	    }
	    pgd -> upper_disp_limit = hopr;
	    pgd -> lower_disp_limit = lopr;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> frat) ):
	case ( (int)&(((struct tankRecord *)0) -> htln) ):
	case ( (int)&(((struct tankRecord *)0) -> diam) ):
	case ( (int)&(((struct tankRecord *)0) -> elev) ):
	case ( (int)&(((struct tankRecord *)0) -> rwgt) ):
	    pgd -> upper_disp_limit = 1000000;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> texp) ):
	case ( (int)&(((struct tankRecord *)0) -> mlos) ):
	    pgd -> upper_disp_limit = 2;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> ttim) ):
	    if (ptank -> tset)
		pgd -> upper_disp_limit = ptank -> tset;
	    else
		pgd -> upper_disp_limit = 86400;
	    pgd -> lower_disp_limit = -1;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> vtim) ):
	    
	    /* set upper to largest valve state transition time, else
	       default, 86400 (secs/day)
	    */
	    hopr = ptank -> stto;
	    if (ptank -> isto > hopr)
		hopr = ptank -> isto;
	    if (ptank -> reto > hopr)
		hopr = ptank -> reto;
	    if (ptank -> ruto > hopr)
		hopr = ptank -> ruto;
	    if (hopr)
		pgd -> upper_disp_limit = hopr;
	    else
		pgd -> upper_disp_limit = 86400;
	    pgd -> lower_disp_limit = -1;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> tasv) ):
	    pgd -> upper_disp_limit = 3;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct tankRecord *)0) -> lavl) ):
	    pgd -> upper_disp_limit = 8;
	    pgd -> lower_disp_limit = 0;
	    break;

    }

    return (0);

} /* end func get_graphic_double () */


/****************************************************************************
 *
 *  MODULE NAME            : get_link_value
 *  AUTHOR                 : emsteel
 *  DATE                   : 01 Sep 94
 *
 *
 *  DESCRIPTION: do dbNameToAddr on a record field and return its value
 *
****************************************************************************/

static void get_link_value (recname, recfld, value)
    char	*recname;
    char	*recfld;
    float	*value;
{
    long     status    = 0;
    long     nRequests = 1;
    long     options   = 0;
    char     pvname_fld[36];
    float    val = 0;
    struct   dbAddr   dbAddr;

    strncpy (pvname_fld, recname, 28);
    strcat (pvname_fld, recfld);
    status = dbNameToAddr (pvname_fld, &dbAddr);
    if (status)
	return;
    status = dbGetField (&dbAddr, DBR_FLOAT, (caddr_t)&val,
			 &options, &nRequests, NULL);
    if (status)
	return;
    *value = val;
    return;

} /* end func get_link_value () */

/****************************************************************************
 *
 *  MODULE NAME            : get_level_values
 *  AUTHOR                 : emsteel
 *  DATE                   : 13 Aug 93
 *
 *  CALLING MODULE(S)      : read_levelpv
 *          ARGUMENT(S)    : ptank = ptr to tank record
 *
 *  DESCRIPTION: get the level value from the level pv's alarm limits
 *  NOTES:
 *        the fields used are not the prompted fields for level but local
 *        this was done so that the original dtabase would not change and
 *        there would be no problem with redundant update
 *        see init_record for initialization of these fields
 *
****************************************************************************/

static long get_level_values (ptank)
   struct   tankRecord  *ptank;
{
   char    errmsg[128];
   char    *pstr;
   caddr_t precord;
   short   rectype;
   double  hihi, high, low, lolo;
   long    status    = 0;
   struct  aiRecord   *pai;
   struct  calcRecord *pcalc;

   precord = 
     ((struct dbAddr *)(ptank -> levp.value.db_link.pdbAddr)) -> precord;
   rectype = 
     ((struct dbAddr *)(ptank -> levp.value.db_link.pdbAddr)) -> record_type;

      /* can't find record type for levelpv */

   if (!(pstr = GET_PRECTYPE (rectype)))
   {
      sprintf (errmsg, "tank: PV:%.28s, levelpv:%.28s, rectype:%s",
                        ptank -> name, ptank -> levp.value.db_link.pvname,
                        pstr);
      errMessage (S_db_noRecType, errmsg);
      return (S_db_noRecType);
   }
   else
   {
      if (strncmp (pstr, "ai", 2) == 0)
      {
         pai = (struct aiRecord *) precord;
         hihi = (double)pai -> hihi;
         high = (double)pai -> high;
         low  = (double)pai -> low;
         lolo = (double)pai -> lolo;
      }
      else if (strncmp (pstr, "calc", 4) == 0)
      {
         pcalc = (struct calcRecord *) precord;
         hihi = (double)pcalc -> hihi;
         high = (double)pcalc -> high;
         low  = (double)pcalc -> low;
         lolo = (double)pcalc -> lolo;
      }
      else
      {
         /* record type is not one to get values from */

         sprintf (errmsg, "tank: PV:%.28s, levelpv:%.28s,ill rectype:%s",
                           ptank -> name, ptank -> levp.value.db_link.pvname,
                           pstr);
         errMessage (S_db_noRecType, errmsg);
         return (S_db_noRecType);
      }
   }

   /* update values if 0 or if updated before */

   if ((ptank -> lvhh == 0) || (ptank -> lavl && HIHI_LEVEL_CHG))
   {
      if (ptank -> lvhh != hihi)
      {
         ptank -> lavl |= HIHI_LEVEL_CHG;
         ptank -> lvhh = hihi;
      }
   }
   if ((ptank -> lvhi == 0) || (ptank -> lavl && HIGH_LEVEL_CHG))
   {
      if (ptank -> lvhi != high)
      {
         ptank -> lavl |= HIGH_LEVEL_CHG;
         ptank -> lvhi = high;
      }
   }
   if ((ptank -> lvlo == 0) || (ptank -> lavl && LOW_LEVEL_CHG))
   {
      if (ptank -> lvlo != low)
      {
         ptank -> lavl |= LOW_LEVEL_CHG;
         ptank -> lvlo = low;
      }
   }
   if ((ptank -> lvll == 0) || (ptank -> lavl && LOLO_LEVEL_CHG))
   {
      if (ptank -> lvll != lolo)
      {
         ptank -> lavl |= LOLO_LEVEL_CHG;
         ptank -> lvll = lolo;
      }
   }

   return (status);

} /* end func get_level_value () */


/****************************************************************************
 *
 *  MODULE NAME            : get_precision 
 *  AUTHOR                 : emsteel
 *  DATE                   : 29 June 93
 *
 *
 *  DESCRIPTION: return display precision
 *
****************************************************************************/

static long get_precision (paddr, precision)
   struct dbAddr     *paddr;
   long              *precision;
{
   struct tankRecord *ptank = (struct tankRecord *)paddr -> precord;

   *precision = ptank -> prec;

   return (0);

} /* end func get_precision () */




/****************************************************************************
 *
 *  MODULE NAME            : get_value 
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 Apr 93
 *
 *
 *  DESCRIPTION: return the value field (current tank volume)
 *
****************************************************************************/

static long get_value (ptank, pvdes)
   struct tankRecord   *ptank;
   struct valueDes     *pvdes;
{
   pvdes -> field_type = DBF_DOUBLE;
   pvdes -> no_elements = 1;
   (double *)(pvdes -> pvalue) = &ptank -> val;

   return (0);

} /* end func get_value () */


/****************************************************************************
 *
 *  MODULE NAME            : init_valves 
 *  AUTHOR                 : emsteel
 *  DATE                   : 23 Apr 93
 *
 *  CALLING MODULE(S)      : init_record
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *
 *  DESCRIPTION: given the name of a valve pv, add .VASV to it and use
 *               dbNameToAddr to resolve it to a struct dbAddr, so that
 *               later the valve's vasv field can be accessed
 *  NOTE:
 *        this will only work if the ptank -> vxxp fields and
 *        ptank -> pvxx_dbAddr fields in tankRecord.ascii stay together
 *        and in the proper order. There is a note in tankRecord.ascii
 *        to this effect
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 25 May 94
 *  REVISION:	return non 0 status if the tanks valves are not init'd
 *
****************************************************************************/

static long init_valves (ptank)
   struct tankRecord   *ptank;
{
   struct dbAddr  *pdbAddr       = &ptank -> pv1_dbAddr;
   struct dbCommon *recHdr;
   char           *pvname        = &ptank -> v01p[0];
   long           status         = 0;
   char           pvname_fld[36];
   char           errmsg[128];
   int            i;
   int            valve_notinit = 0;

   for (i = 0; i < MAX_TANK_VALVES; i++)
   {
         /* init ptr to valve record */

      pdbAddr -> precord = NULL;

         /* if there is a valve defined */

      if (*pvname)
      {
         strncpy (pvname_fld, pvname, 30);
         strcat  (pvname_fld, ".VASV");
         if ((status = dbNameToAddr (pvname_fld, pdbAddr)) != 0)
         {
            sprintf (errmsg, "tank: PV:%.28s, valve:%.28s, init error",
                             ptank -> name, pvname);
            errMessage (S_tank_TankValve_Err, errmsg);
            return (S_tank_TankValve_Err);
         }
	 else
	 {
	    recHdr = (struct dbCommon *) (pdbAddr -> precord);
	    if (recHdr -> stat == NO_INIT_ALARM)
		valve_notinit = 1;
	 }
      }
      pvname += 30;
      pdbAddr++;
   }

   if (valve_notinit)
      status = -1;

   return (status);

} /* end func init_valves () */


/****************************************************************************
 *
 *  MODULE NAME            : lock_unlock_valves 
 *  AUTHOR                 : emsteel
 *  DATE                   : 18 May 93
 *
 *  CALLING MODULE(S)      : check_mode
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *                           lock -  1 = lock, 0 = unlock
 *
 *  DESCRIPTION: get the valves current mode
 *               if we are to lock the tanks associated valves
 *                   1 - save valves mode (save_mode) in ptank -> vxxm
 *                   2 - set valves mode to TF_LOCKED
 *               if we are to unlock the tanks valves
 *                   1 - make sure the valve is still locked
 *                   2 - restore valves mode to ptank -> vxxm
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 27 Feb 95
 *  REVISION:	corrected arg list to dbPutField
 *
****************************************************************************/

static long lock_unlock_valves (ptank, lock)
   struct tankRecord   *ptank;
   int                 lock;
{
   struct dbAddr  dbAddr;
   char           *pvname     = &ptank -> v01p[0];
   unsigned short *valve_mode = &ptank -> v01m;
   unsigned short mode;
   unsigned short save_mode;
   long           status      = 0;
   long           options     = 0;
   long           nRequests   = 1;
   char           errmsg[128];
   char           pvname_fld[36];
   int            i;

   for (i = 0; i < MAX_TANK_VALVES; i++)
   {
      if (*pvname)
      {
         strncpy (pvname_fld, pvname, 30);
         strcat  (pvname_fld, ".MODE");
         if ((status = dbNameToAddr (pvname_fld, &dbAddr)) != 0)
         {
            sprintf (errmsg, "tank: PV:%.28s, valve:%.30s, lock_unlock",
                             ptank -> name, pvname);
            errMessage (S_tank_TankValve_Err, errmsg);
            return (S_tank_TankValve_Err);
         }
         if ((status = dbGetField (&dbAddr, DBR_ENUM, (caddr_t)&mode,
                                  &options, &nRequests, NULL)) != 0)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM,
                                                              VALID_ALARM);
            return (status);
         }
         if (lock)
         {
            if (mode != TF_LOCKED)
            {
               save_mode = mode;
               mode = TF_LOCKED;
               if ((status = dbPutField (&dbAddr, DBR_ENUM, (caddr_t)&mode,
                                        nRequests)) != 0)
               {
                  TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                               VALID_ALARM);
                  return (status);
               }
               *valve_mode = save_mode;
            }
         }
         else
         {
            if (mode == TF_LOCKED)
            {
               mode = *valve_mode;
               if ((status = dbPutField (&dbAddr, DBR_ENUM, (caddr_t)&mode,
                                        nRequests)) != 0)
               {
                  TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                               VALID_ALARM);
                  return (status);
               }
            }
         }

      }
      valve_mode++;
      pvname += 30;
   }

   return (status);

} /* end func lock_unlock_valves () */


/****************************************************************************
 *
 *  MODULE NAME            : monitor 
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 Apr 93
 *
 *  CALLING MODULE(S)      : process ()
 *                           run_commands ()
 *
 *  DESCRIPTION: check event list
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 05 Jan 94
 *  REVISION:  check VALID_ALARM before set ACK
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 30 Aug 94
 *  REVISION:	rm DBE_LOG from ack, only archive if tenb (trending) is
 *		enabled
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 15 Mar 95
 *  REVISION:	do not set ack if sevr is EVENT_ALARM
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 11 Sept 95
 *  REVISION:	archive if stat/sevr also chg's
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 12 Sept 95
 *  REVISION:	chg way handle VALID_ALARM
 *
****************************************************************************/

static void monitor (ptank)
   struct tankRecord *ptank;
{
    unsigned short monitor_mask;
    double         delta;
    short          stat, sevr, nsta, nsev;
    int		   set_ack = 1;

      /* get previous stat and sevr and new stat and sevr */

    stat = ptank -> stat;
    sevr = ptank -> sevr;
    nsta = ptank -> nsta;
    nsev = ptank -> nsev;

      /* set current stat and sevr */

    ptank -> stat = nsta;
    ptank -> sevr = nsev;
    ptank -> nsta = 0;
    ptank -> nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == ptank -> psta) && (nsev == ptank -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and out previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	ptank -> psta = stat;
	ptank -> psev = sevr;
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> psev, DBE_VALUE);
    }

    monitor_mask = 0;

      /* did alarm condition change this scan */

    if ((stat != nsta) || (sevr != nsev))
    {
	/* post events for alarm condition change */

	monitor_mask |= DBE_ALARM;

	/* when alarm status chg's, get current time and
	   save VAL in lalm (last value when alarm chg'd)
	*/
	tsLocalTime (&ptank -> latm);
	ptank -> lalm = ptank -> val;

	/* post stat and nsev fields */

	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> sevr, DBE_VALUE);
	if ((ptank -> sevr != NO_ALARM) && (set_ack) &&
	    (ptank -> sevr != EVENT_ALARM) )
	{
	    if (ptank -> ack == 0)
	    {
		ptank -> ack = 1;
		db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> ack, DBE_VALUE);
	    }
	}
    }
      /* check for value changes */

    delta = ptank -> mlst - ptank -> val;  /* last monitored val - val */
    if (delta < 0.0)
	delta = -delta;

    if ( delta > ptank -> mdel)   /* monitor deadband */
    {
	/* post events for value change */

	monitor_mask |= DBE_VALUE;

	/* update last value monitored */

	ptank -> mlst = ptank -> val ;
    }

      /* check for archive change if tenb (trending) is enabled */

    if (ptank -> tenb)
    {
	delta = ptank -> alst - ptank -> val;
	if (delta < 0.0)
	    delta = -delta;

	if ( (delta > ptank -> adel) || (monitor_mask & DBE_ALARM) )
	{
	    /* post events for value change */

	    monitor_mask |= DBE_LOG;

	    /* update last archive value */

	    ptank -> alst = ptank -> val;
	}
    }

      /* send out monitors connected to the value field */

    if (monitor_mask)
    {
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> val, monitor_mask);
    }

    return;

} /* end func monitor () */


/****************************************************************************
 *
 *  MODULE NAME            : post_events
 *  AUTHOR                 : emsteel
 *  DATE                   : 25 May 94
 *
 *  PARAMETERS             : ptank - ptr to current record
 *                           ptank_old - ptr to previous values
 * 
 *  DESCRIPTION: called when time to process a tank record
 *
****************************************************************************/

static void post_events (ptank, ptank_old)
    struct tankRecord	*ptank;
    struct tankred	*ptank_old;
{
    if (strncmp (ptank_old -> tsts, ptank -> tsts, sizeof (ptank_old -> tsts))  != 0)
	db_post_events ((struct dbCommon *)ptank, ptank -> tsts, DBE_VALUE);

    if (strncmp (ptank_old -> pcmd, ptank -> pcmd, sizeof (ptank_old -> pcmd))  != 0)
	db_post_events ((struct dbCommon *)ptank, ptank -> pcmd, DBE_VALUE);

    if (strncmp (ptank_old -> rsts, ptank -> rsts, sizeof (ptank_old -> rsts))  != 0)
	db_post_events ((struct dbCommon *)ptank, ptank -> rsts, DBE_VALUE);

    if (ptank_old -> esou != ptank -> esou)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> esou, DBE_VALUE);

    if (ptank_old -> tasv != ptank -> tasv)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> tasv, DBE_VALUE);

    if (ptank_old -> mlos != ptank -> mlos)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> mlos, DBE_VALUE);

    if (ptank_old -> qvol != ptank -> qvol)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> qvol, DBE_VALUE);

    if (ptank_old -> ttim != ptank -> ttim)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> ttim, DBE_VALUE);

    if (ptank_old -> vtim != ptank -> vtim)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> vtim, DBE_VALUE);

    if (ptank_old -> tcks != ptank -> tcks)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> tcks, DBE_VALUE);

    if (ptank_old -> tval != ptank -> tval)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> tval, DBE_VALUE);

    if (ptank_old -> lvhh != ptank -> lvhh)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> lvhh, DBE_VALUE);

    if (ptank_old -> lvhi != ptank -> lvhi)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> lvhi, DBE_VALUE);

    if (ptank_old -> lvlo != ptank -> lvlo)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> lvlo, DBE_VALUE);

    if (ptank_old -> lvll != ptank -> lvll)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> lvll, DBE_VALUE);

    if (ptank_old -> plev != ptank -> plev)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> plev, DBE_VALUE);

    if (ptank_old -> bswl != ptank -> bswl)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> bswl, DBE_VALUE);

    if (ptank_old -> bswv != ptank -> bswv)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> bswv, DBE_VALUE);

    if (ptank_old -> pbsw != ptank -> pbsw)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> pbsw, DBE_VALUE);

    if (ptank_old -> tpav != ptank -> tpav)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> tpav, DBE_VALUE);

    if (ptank_old -> dnct != ptank -> dnct)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> dnct, DBE_VALUE);

    if (ptank_old -> gvol != ptank -> gvol)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> gvol, DBE_VALUE);

    if (ptank_old -> gcap != ptank -> gcap)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> gcap, DBE_VALUE);

    if (ptank_old -> ullg != ptank -> ullg)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> ullg, DBE_VALUE);

    if (ptank_old -> avap != ptank -> avap)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> avap, DBE_VALUE);

    if (ptank_old -> wcap != ptank -> wcap)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> wcap, DBE_VALUE);

    if (ptank_old -> netv != ptank -> netv)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> netv, DBE_VALUE);

    if (ptank_old -> texp != ptank -> texp)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> texp, DBE_VALUE);

    if (ptank_old -> esin != ptank -> esin)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> esin, DBE_VALUE);

    if (ptank_old -> frat != ptank -> frat)
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> frat, DBE_VALUE);

    return;

} /* end func post_events () */


/****************************************************************************
 *
 *  MODULE NAME            : process 
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 Apr 93
 *
 *  PARAMETERS             : paddr - ptr to record struct packet            
 * 
 *  DESCRIPTION: called when time to process a tank record
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 17 Apr 95
 *  REVISION:	added code to process the forward link flnk
 *
****************************************************************************/

static long process (paddr)
   struct dbAddr *paddr;
{
    struct tankRecord	*ptank = (struct tankRecord *) (paddr -> precord);
    struct tankred	tank_old;
    struct tank_trips	tankTrips;
    long		status = 0;

      /* do nothing if we did not init */

    if ((!ptank -> init) || (!ptank -> ini2))
	return (0);

    /* get time stamp for last time record was processed */

    tsLocalTime (&ptank -> lptm);

      /* save the current state of the tank for redundancy check */

    save_tank (ptank, &tank_old);

      /* get system tick counter to use for time elapse */

    ptank -> tcks = tickGet ();

      /* init new status to previous */

    ptank -> nsta = ptank -> stat;
    ptank -> nsev = ptank -> sevr;
    ptank -> msev = NO_ALARM;

      /* off normal operations */

    if (ptank -> tagd)
    {
	ptank -> nsta = TAGGED_ALARM;
	ptank -> nsev = VALID_ALARM;
    }
    else if (ptank -> oscn == OFF_SCAN)
    {
	ptank -> nsta = SCAN_ALARM;
	ptank -> nsev = VALID_ALARM;
    }
    else
    {
	ptank -> nsta = NO_ALARM;
	ptank -> nsev = NO_ALARM;

        /* update number of process times */
	if (process_stats)
	    process_cnts[type]++;	

	set_roof_status (ptank);

         /* read all sensors and compute all values */

	status = read_sensors (ptank, &ptank -> tcks, &tank_old.tcks, &tankTrips);
	if (!status)
	{
		/* set tank to active */
   
	    ptank -> pact = TRUE;

		/* check for change in ptank -> mode */

	    check_mode (ptank);

		/* if tank is stopped, see if we are timing to set
		   quiescent volume
		*/

	    if (ptank -> tasv == TANK_STOPPED)
		update_qvol (ptank, &tank_old);

		/* update valve transition timer */

	    update_vtim (ptank, &tank_old);

		/* if an emergency stop has been called for */

	    if (ptank -> esin)
	    {
		status = emergency_stop (ptank);
		reset_tank_commands (ptank);
		ptank -> esin = 0;
	    }
	    else
	    {
		/* get current command if any and nprocess */

		status = run_commands (ptank);
	    }

		/* check for alarm conditions */

	    alarm (ptank, &tankTrips);

	    /* check for maximize severity */

	    if (ptank -> msev > ptank -> nsev)
	    {
		ptank -> nsev = ptank -> msev;
		ptank -> nsta = LINK_ALARM;
	    }

	}

    } /* end else (oscn != OFF_SCAN) */

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (ptank -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&ptank -> latm);
	db_post_events ((struct dbCommon *)ptank, (caddr_t)&ptank -> val, DBE_VALUE | DBE_ALARM);
    }

    /* if it is the 1st process pass 
       init last value when alarm chg'd (lalm)
    */

    if (ptank -> fpps)
    {
	ptank -> fpps = 0;
	ptank -> lalm = ptank -> val;
    }

      /* check event list */

    monitor (ptank);

      /* the process is no longer active */

    ptank -> pact = FALSE;

	/* post events for monitors */

    post_events (ptank, &tank_old);

      /* now check for redundancy, see programmers notes */

    update_redun (&tank_old, ptank);

   /* process the forward scan link record */

   if (ptank -> flnk.type == DB_LINK) 
      dbScanPassive ((struct dbAddr *)ptank -> flnk.value.db_link.pdbAddr);

    return (status);

} /* end func process () */


/****************************************************************************
 *
 *  MODULE NAME            : read_bswpv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 15 Apr 93
 *
 *  CALLING MODULE(S)      : init_record, read_sensors
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *
 *
 *  DESCRIPTION: read the BSW (Bottom Sediment and Water) sensor
 *
 *  Note: The bsw pv might be of the type that measures from the sensor
 *        to the top of the water. This is denoted by the "distance" choice
 *        from ptank->bmea. The bsw level is calculated as
 *             level = sensor level value (ptank->blva) - sensor value
 *        else the method is "level" and the level is as read from the
 *        sensor
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 04 Jan 94
 *  REVISION:  do calc for gvol in compute_some_values
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 29 Mar 94
 *  REVISION:  chg choiceInd from short to unsigned short to agree with what 
 *             bmea is
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 11 July 94
 *  REVISION:	if val is 0 then set pbsw to 0, alleviate possibel divide 
 *		by zero problem
 *
****************************************************************************/

static long read_bswpv (ptank, orig_val)
   struct tankRecord  *ptank;
   double             orig_val;    /* volume at start of proccessing */
{
   long     nRequests = 1;
   long     options   = 0;
   double   curr_bswl = ptank -> bswl;    /* current bsw level */
   double   curr_bswv = ptank -> bswv;    /* current bsw volume */
   long     status = 0;
                                          /* stuff to access a choice set */
   short    choice_set = REC_TANK_MEAS;   /* type of set */
   unsigned short    choiceInd = ptank -> bmea;    /* current choice */
   struct arrChoiceSet  *parrChoiceSet;   /* tank choice sets */
   struct choiceSet     *pchoiceSet;      /* choice set */
   char                 *pchoice_meas;    /* the method of measurement */

   if (ptank -> bswp.type != CONSTANT)
   {
      if (ptank -> bswp.type == DB_LINK)
      {
            /* get value from sensor */
         status = dbGetLink (&ptank -> bswp.value.db_link, 
			     (struct dbCommon *)ptank, DBR_DOUBLE, 
			     (caddr_t)&curr_bswl, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM,
                                                               VALID_ALARM);
            return (status);
         }
	 else if (!ptank -> ini2)
	 {
	    if (!okToUse (&ptank -> bswp) )
	       return (-1);
	 }
               /* compute bsw volume by interpolating from strapping table */

         ComputeTFVolume (ptank -> pstr, ptank -> snum, curr_bswl,
                                                                &curr_bswv);

         if ( (!(parrChoiceSet = GET_PARR_CHOICE_SET (choiceRec, type)))
           || (!(pchoiceSet = GET_PCHOICE_SET (parrChoiceSet, choice_set)))
           || (!(pchoice_meas = GET_CHOICE (pchoiceSet, choiceInd))) )
         {
            recGblRecordError (S_db_badChoice, (struct dbCommon *)ptank, 
			       "tank: read_bswpv, choice set");
            return (S_db_badChoice);
         }
         if (strncmp (pchoice_meas, "distance", 8) == 0)
         {
            curr_bswl = ptank -> blva - curr_bswl;
         }
         else if (strncmp (pchoice_meas, "level", 5) != 0)
         {
            recGblRecordError (S_db_badChoice, (struct dbCommon *)ptank, 
			       "tank: read_bswpv, measurement");
            return (S_db_badChoice);
         }

            /* update current bsw level, volume, % bsw to current
               product volume, gross volume */

         ptank -> bswl = curr_bswl;
         ptank -> bswv = curr_bswv;

            /* compute % BSW */
         
	 if (ptank -> val == 0)
	     ptank -> pbsw = 0;
	 else
             ptank -> pbsw = (ptank -> bswv / ptank -> val) * 100;
      }
   } /* end if ! CONSTANT */


   return (status);

} /* end func read_bswpv () */


/****************************************************************************
 *
 *  MODULE NAME            : read_denspv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 15 Apr 93
 *
 *  CALLING MODULE(S)      : init_record, read_sensors
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *
 *
 *  DESCRIPTION: read the density pv
 *
****************************************************************************/

static long read_denspv (ptank)
   struct tankRecord  *ptank;
{
   long     nRequests = 1;
   long     options   = 0;
   long     status    = 0;
   double   curr_dnct = ptank -> dnct;    /* current product density */

   if (ptank -> denp.type != CONSTANT)
   {
      if (ptank -> denp.type == DB_LINK)
      {
            /* get value from sensor */
         status = dbGetLink (&ptank -> denp.value.db_link, 
			     (struct dbCommon *)ptank, DBR_DOUBLE, 
			     (caddr_t)&curr_dnct, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
							      VALID_ALARM);
	    return (status);
         }
         else if (!ptank -> ini2)
         {
	    if (!okToUse (&ptank -> denp) )
	       return (-1);
	 }
         ptank -> dnct = curr_dnct;
      }
   }

   return (status);

} /* end func read_denspv () */


/****************************************************************************
 *
 *  MODULE NAME            : read_levelpv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 Apr 93
 *
 *  CALLING MODULE(S)      : init_record, read_sensors
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *
 *
 *  DESCRIPTION: read the level pv and volume pv sensors
 *               if there is a level pv then get its value, if there is
 *               a volume pv get its value (the products volume). if there
 *               was no level pv then interpolate level from strapping
 *               table given volume from volume pv. if no volume pv
 *               and there is a level pv then interpolate volume
 *               from strapping table
 *  Note: The level pv might be of the type that measures from the sensor
 *        to the top of the product. This is denoted by the "distance" choice
 *        from ptank->pmea. The product level is calculated as
 *        level = sensor level value (ptank->plva) - sensor value
 *        else the method is "level" and the level is as read from the
 *        sensor
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 04 Jan 94
 *  REVISION:  chg choiceInd to fro short to unsigned short to agree with 
 *             what pmea is
 *
****************************************************************************/

static long read_levelpv (ptank, tankTrips)
   struct tankRecord	*ptank;
   struct tank_trips	*tankTrips;
{
   short    levpv     = 0;                /* = 1 if there is a level pv */
   short    volpv     = 0;                /* = 1 if there is a volume pv */
   long     nRequests = 1;
   long     options   = 0;
   double   curr_plev = ptank -> plev;    /* current value from level pv */
   double   curr_val  = ptank -> val;     /* current value from volume pv */
   unsigned short trip_val = 0;
   long     status    = 0;
                                          /* stuff to access a choice set */
   short    choice_set = REC_TANK_MEAS;   /* type of set */
   unsigned short    choiceInd = ptank -> pmea;    /* current choice */
   struct arrChoiceSet  *parrChoiceSet;   /* tank choice sets */
   struct choiceSet     *pchoiceSet;      /* choice set */
   char                 *pchoice_meas;    /* the method of measurement */

   if (ptank -> levp.type != CONSTANT)
   {
      if (ptank -> levp.type == DB_LINK)
      {
            /* get value from sensor */
         status = dbGetLink (&ptank -> levp.value.db_link, 
			     (struct dbCommon *)ptank, DBR_DOUBLE, 
			     (caddr_t)&curr_plev, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                               VALID_ALARM);
             return (status);
         }
	 else if (!ptank -> ini2)
	 {
	    if (!okToUse (&ptank -> levp) )
	       return (-1);
	 }
         levpv = 1;
      }
   }

   if (ptank -> volp.type != CONSTANT)
   {
      if (ptank -> volp.type == DB_LINK)
      {
         status = dbGetLink (&ptank -> volp.value.db_link, 
			     (struct dbCommon *)ptank, DBR_DOUBLE, 
			     (caddr_t)&curr_val, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                              VALID_ALARM);
            return (status);
         }
         else if (!ptank -> ini2)
         {
	    if (!okToUse (&ptank -> volp) )
	       return (-1);
         }
         volpv = 1;
      }
   }
      /* if there is a level pv, find the type of measurement and adjust
         the level accordingly
      */

   if (levpv)
   {
      if ( (!(parrChoiceSet = GET_PARR_CHOICE_SET (choiceRec, type)))
        || (!(pchoiceSet = GET_PCHOICE_SET (parrChoiceSet, choice_set)))
        || (!(pchoice_meas = GET_CHOICE (pchoiceSet, choiceInd))) )
      {
         recGblRecordError (S_db_badChoice, (struct dbCommon *)ptank,
                            "tank: read_levelpv, choice set");
         return (S_db_badChoice);
      }
      if (strncmp (pchoice_meas, "distance", 8) == 0)
      {
         curr_plev = ptank -> plva - curr_plev;
      }
      else if (strncmp (pchoice_meas, "level", 5) != 0)
      {
         recGblRecordError (S_db_badChoice, (struct dbCommon *)ptank,
                            "tank: read_levelpv, measurement");
         return (S_db_badChoice);
      }
   }
      /* if there was a volume pv but no levelpv, then compute
         level by interpolating from strapping table
      */

   if (volpv && (!levpv))
   {
      ComputeTFLevel (ptank -> pstr, ptank -> snum, &curr_plev, curr_val);
   }
      /* if there was no volume pv but there was a level pv, then
         compute volume by interplating from strapping table
      */

   if ((!volpv) && levpv)
   {
      ComputeTFVolume (ptank -> pstr, ptank -> snum, curr_plev, &curr_val);
   }

      /* update current product level, current product volume */

   ptank -> plev = curr_plev;
   ptank -> val = curr_val;

      /* read the low and hi level trip sensors if they exist */

   tankTrips -> hi_level_trip = TRIP_NOT_ACTIVE;
   tankTrips -> low_level_trip = TRIP_NOT_ACTIVE;

   if (ptank -> llpv.type != CONSTANT)
   {
      if (ptank -> llpv.type == DB_LINK)
      {
            /* get value from sensor */
         status = dbGetLink (&ptank -> llpv.value.db_link, 
                             (struct dbCommon *)ptank, DBR_ENUM, 
                             (caddr_t)&trip_val, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                               VALID_ALARM);
            return (status);
         }
         else if (!ptank -> ini2)
         {
	    if (!okToUse (&ptank -> llpv) )
	       return (-1);
         }
         if (trip_val == 1)
            tankTrips -> low_level_trip = TRIP_ACTIVE;
      }
   }

   if (ptank -> hlpv.type != CONSTANT)
   {
      if (ptank -> hlpv.type == DB_LINK)
      {
            /* get value from sensor */
         status = dbGetLink (&ptank -> hlpv.value.db_link, 
                             (struct dbCommon *)ptank, DBR_ENUM, 
                             (caddr_t)&trip_val, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                               VALID_ALARM);
            return (status);
         }
         else if (!ptank -> ini2)
         {
	    if (!okToUse (&ptank -> hlpv) )
	       return (-1);
         }
         if (trip_val == 1)
            tankTrips -> hi_level_trip = TRIP_ACTIVE;
      }
   }

      /* get HIHI, HIGH, LOLO, LOW level values from level pv
         alarm values, if pv exists */

   if (levpv)
   {
      status = get_level_values (ptank);
   }

   return (status);

} /* end func read_levelpv () */


/****************************************************************************
 *
 *  MODULE NAME            : read_pressurepv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 19 Apr 93
 *
 *  CALLING MODULE(S)      : read_sensors()
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *
 *  DESCRIPTION: read the pressure trip sensors
 *
****************************************************************************/

static long read_pressurepv (ptank, tankTrips)
   struct tankRecord   *ptank;
   struct tank_trips	*tankTrips;
{
   long     status            = 0;
   unsigned short   trip_val  = 0;
   long     nRequests         = 1;
   long     options           = 0;

   tankTrips -> hi_pressure_trip = TRIP_NOT_ACTIVE;
   tankTrips -> low_pressure_trip = TRIP_NOT_ACTIVE;

      /* low pressure trip sensor */

   if (ptank -> ltpv.type != CONSTANT)
   {
      if (ptank -> ltpv.type == DB_LINK)
      {
            /* get value from sensor */
         status = dbGetLink (&ptank -> ltpv.value.db_link, 
                             (struct dbCommon *)ptank, DBR_ENUM, 
                             (caddr_t)&trip_val, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                                VALID_ALARM);
            return (status);
         }
         else if (!ptank -> ini2)
         {
	    if (!okToUse (&ptank -> ltpv) )
	       return (-1);
         }
         if (trip_val == 1)
            tankTrips -> low_pressure_trip = TRIP_ACTIVE;
      }
   }
      /* hi pressure trip sensor */

   if (ptank -> htpv.type != CONSTANT)
   {
      if (ptank -> htpv.type == DB_LINK)
      {
            /* get value from sensor */
         status = dbGetLink (&ptank -> htpv.value.db_link, 
                             (struct dbCommon *)ptank, DBR_ENUM,
                             (caddr_t)&trip_val, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                                VALID_ALARM);
            return (status);
         }
         else if (!ptank -> ini2)
         {
	    if (!okToUse (&ptank -> htpv) )
	       return (-1);
         }
         if (trip_val == 1)
            tankTrips -> hi_pressure_trip = TRIP_ACTIVE;
      }
   }

   return (status);

} /* end func read_pressurepv () */


/****************************************************************************
 *
 *  MODULE NAME            : read_sensors 
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 Apr 93
 *
 *  CALLING MODULE(S)      : init_record, process
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *                           curr_ticks - ptr to current system tiem ticks
 *                           last_ticks - ptr to last scans time ticks
 *
 *  DESCRIPTION: call all functions to read all type of sensors for a
 *               tank record
 *
****************************************************************************/

static long read_sensors (ptank, curr_ticks, last_ticks, tankTrips)
   struct  tankRecord   *ptank;
   unsigned long        *curr_ticks;
   unsigned long        *last_ticks;
   struct tank_trips	*tankTrips;
{
   double   orig_vol = ptank -> val;  /* previous product volume */
   long     status   = 0;

      /* read level and volume */

   if ((status = read_levelpv (ptank, tankTrips)) != 0)
      return (status);

      /* read BSW Bottom Sediment and Water */

   if ((status = read_bswpv (ptank, orig_vol)) != 0)
      return (status);

      /* read temperature(s)
         NOTE: must be read after read_levelpv ()
      */

   if ((status = read_temppv (ptank)) != 0)
      return (status);

      /* read density sensor */

   if ((status = read_denspv (ptank)) != 0)
      return (status);

      /* read the pressure trip switches */

   if ((status = read_pressurepv (ptank, tankTrips)) != 0)
      return (status);

   status = compute_some_values (ptank, orig_vol, curr_ticks, last_ticks);

   return (status);
   
} /* end func read_sensors () */


/****************************************************************************
 *
 *  MODULE NAME            : read_temppv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 15 Apr 93
 *
 *  CALLING MODULE(S)      : init_record, read_sensors
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *
 *
 *  DESCRIPTION: read temperature sensor(s)
 *               if there is an ave temp probe read it
 *               if there are temp probes at various levels, then read
 *                  the temp if it's level is less than current product
 *                  level (plev). add up the temp and get an average
 *
****************************************************************************/

static long read_temppv (ptank)
   struct tankRecord  *ptank;
{
   long     nRequests      = 1;
   long     options        = 0;
   long     status         = 0;
   double   curr_tpav      = ptank -> tpav;   /* current average product temp */
   short    ave_temp_probe = 0;               /* is there an ave temp pv */
   struct link *probeptr   = &ptank -> t01v;  /* first temperature probe */
   double   *probe_level   = &ptank -> t01l;  /* level of first probe */
   short    num_probes     = 0;               /* num temp probes installed */
   double   temp_sum       = 0;               /* sum of temps from probes */
   double   temp           = 0;               /* value returned by temp probe */
   short    i;

   if (ptank -> tmpp.type != CONSTANT)
   {
      if (ptank -> tmpp.type == DB_LINK)
      {
            /* get value from sensor */

         status = dbGetLink (&ptank -> tmpp.value.db_link, 
                             (struct dbCommon *)ptank, DBR_DOUBLE, 
                             (caddr_t)&curr_tpav, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                                VALID_ALARM);
            return (status);
         }
         else if (!ptank -> ini2)
         {
	    if (!okToUse (&ptank -> tmpp) )
	       return (-1);
         }
         ave_temp_probe = 1;
      }
   }

   if (!ave_temp_probe)
   {
      /* loop through all possible temp probes */

      for (i = 0; i < MAX_TEMP_PROBES; i++)
      {
         if (probeptr -> type != CONSTANT)
         {
            if (probeptr -> type == DB_LINK)
            {
               status = dbGetLink (&probeptr -> value.db_link, 
                                   (struct dbCommon *)ptank, DBR_DOUBLE, 
                                   (caddr_t)&temp, &options, &nRequests);
               if (status)
               {
                  TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                               VALID_ALARM);
                  return (status);
               }
	       else if (!ptank -> ini2)
	       {
		  if (!okToUse (probeptr) )
		     return (-1);
	       }
               if (ptank -> plev >= (*probe_level))
               {
                  num_probes++;
                  temp_sum += temp;
               }
            }
         }
         probeptr++;
         probe_level++;
      }
   }

      /* use ave temp from ave temp pv, else calculate from
         immersed temp sensors if they were installed
      */

   if (!ave_temp_probe)
   {
      if (num_probes)
         curr_tpav = temp_sum / num_probes;
   }
   if (ave_temp_probe || num_probes)
   {
      ptank -> tpav = curr_tpav;
   }

   return (status);

} /* end func read_temppv () */


/****************************************************************************
 *
 *  MODULE NAME            : reset_tank_commands 
 *  AUTHOR                 : emsteel
 *  DATE                   : 13 Aug 93
 *
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *
 *  DESCRIPTION: reset tank command fields
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION:  deleted dp_post_events for command fields, not needed
 *
****************************************************************************/

static void reset_tank_commands (ptank)
   struct tankRecord    *ptank;
{
   struct tfcmd   tfcmd;

   tfcmd.NumCmds                = 4;
   tfcmd.FirstDigitalCmd        = &ptank -> ciss;
   tfcmd.FirstSpecialDigitalCmd = &ptank -> siss;
   tfcmd.FirstCmdString         = ptank -> issc;
   tfcmd.CmdString              = ptank -> cmds;

   ResetTFCommand (&tfcmd);

   return;

} /* end func reset_tank_commands () */


/****************************************************************************
 *
 *  MODULE NAME            : run_commands 
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 Apr 93
 *
 *  CALLING MODULE(S)      : process ()
 *  ARGUMENT(S)            : ptank = ptr to tank record
 *
 *  DESCRIPTION: run a command to the tank
 *  NOTE:
 *       ptank -> vtim is the valve transition state timer
 *       ptank -> ttim is the tank settling timer to set
 *                     quiescent volume (ptank -> qvol)
 *       struct tfcmd - see tankfarm.h
 *       GetTFCommand() and ResetTFCommand() - see tankfarm.c
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION: added code to process special digital cmds and update of
 *            PCMD (last command field)
 *
****************************************************************************/

static long run_commands (ptank)
   struct tankRecord   *ptank;
{
   struct tfcmd   tfcmd;
   int            cmd;
   int            cmd_type;

   tfcmd.NumCmds                = 4;
   tfcmd.FirstDigitalCmd        = &ptank -> ciss;
   tfcmd.FirstSpecialDigitalCmd = &ptank -> siss;
   tfcmd.FirstCmdString         = ptank -> issc;
   tfcmd.CmdString              = ptank -> cmds;

   cmd = GetTFCommand (&tfcmd, &cmd_type);

   if (cmd != NO_TF_COMMAND)
   {
      if ((ptank -> mode == TF_AUTOMATIC) || (ptank -> mode == TF_MANUAL))
      {
         if (cmd == ILLEGAL_TF_STRING_CMD)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, CMD_NOT_ALLOWED,
                                                               INFO_ALARM);
         }
         else if ( ((ptank -> mode == TF_AUTOMATIC) &&
                    ((cmd_type == TF_STRING_CMD) ||
                     (cmd_type == TF_SPECIAL_DIGITAL_CMD))) ||
                   ((ptank -> mode == TF_MANUAL) &&
                    (cmd_type == TF_DIGITAL_CMD)) )
         {
            switch (cmd)
            {
               case TANK_ISSUE_CMD:
                  strncpy (ptank -> tsts, ptank -> isss,
                                                     sizeof (ptank -> tsts));
                  ptank -> tasv = TANK_ISSUING;
                  ptank -> ttim = -1.0;
                  ptank -> vtim = 0;
                  ptank -> tval = 0;

                     /* update last command */
  
                  if (ptank -> issc[0])
                     strncpy (ptank -> pcmd, ptank -> issc, sizeof (ptank -> pcmd));
                  else
                     strncpy (ptank -> pcmd, TF_UNDEF_CMD, sizeof (ptank -> pcmd));

                  break;
      
               case TANK_RECEIVE_CMD:
                  strncpy (ptank -> tsts, ptank -> recs,
                                                     sizeof (ptank -> tsts));
                  ptank -> tasv = TANK_RECEIVING;
                  ptank -> ttim = -1.0;
                  ptank -> vtim = 0;
                  ptank -> tval = 0;
  
                  if (ptank -> recc[0])
                     strncpy (ptank -> pcmd, ptank -> recc, sizeof (ptank -> pcmd));
                  else
                     strncpy (ptank -> pcmd, TF_UNDEF_CMD, sizeof (ptank -> pcmd));

                  break;
      
               case TANK_RUN_CMD:
                  strncpy (ptank -> tsts, ptank -> runs,
                                                     sizeof (ptank -> tsts));
                  ptank -> tasv = TANK_RUNNING;
                  ptank -> ttim = -1.0;
                  ptank -> vtim = 0;
                  ptank -> tval = 0;
  
                  if (ptank -> runc[0])
                     strncpy (ptank -> pcmd, ptank -> runc, sizeof (ptank -> pcmd));
                  else
                     strncpy (ptank -> pcmd, TF_UNDEF_CMD, sizeof (ptank -> pcmd));

                  break;
      
               case TANK_STOP_CMD:
                  strncpy (ptank -> tsts, ptank -> stps,
                                                    sizeof (ptank -> tsts));
                  ptank -> tasv = TANK_STOPPED;
                  ptank -> vtim = 0;
                  ptank -> ttim = 0;
                  ptank -> tval = 0;
  
                  if (ptank -> stpc[0])
                     strncpy (ptank -> pcmd, ptank -> stpc, sizeof (ptank -> pcmd));
                  else
                     strncpy (ptank -> pcmd, TF_UNDEF_CMD, sizeof (ptank -> pcmd));

                  break;
      
               default:
                  TankFarmPostAlarm ((struct dbCommon *)ptank, 
                                               CMD_NOT_ALLOWED, INFO_ALARM);
            } /* end switch (cmd) */
         }
         else
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, CMD_NOT_ALLOWED, 
                                                               INFO_ALARM);
         }
      }
      else if (ptank -> mode == TF_LOCKED)
      {
         TankFarmPostAlarm ((struct dbCommon*)ptank, CMD_NOT_ALLOWED, 
                                                               INFO_ALARM);
      }
   }

   reset_tank_commands (ptank);

   return (0);

} /* end func run_command () */


/****************************************************************************
 *
 *  MODULE NAME            : save_tank 
 *  AUTHOR                 : emsteel
 *  DATE                   : 22 Apr 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *                           tank_old ptr to tank redundancy struct
 *
 *  DESCRIPTION: save values from tank record for redundancy check
 *               at end of process
 * 
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 23 Feb 94
 *  REVISION:  removed init, alst, mlst, esin, time not needed
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 Sept 94
 *  REVISION:	ini2 was added
 *
****************************************************************************/

static void save_tank (ptank, ptank_old)
   struct tankRecord  *ptank;
   struct tankred     *ptank_old;
{
   ptank_old -> stat = ptank -> stat;
   ptank_old -> sevr = ptank -> sevr;
   ptank_old -> val  = ptank -> val;
   ptank_old -> v01m = ptank -> v01m;
   ptank_old -> v02m = ptank -> v02m;
   ptank_old -> v03m = ptank -> v03m;
   ptank_old -> v04m = ptank -> v04m;
   ptank_old -> v05m = ptank -> v05m;
   ptank_old -> v06m = ptank -> v06m;
   ptank_old -> v07m = ptank -> v07m;
   ptank_old -> v08m = ptank -> v08m;
   ptank_old -> v09m = ptank -> v09m;
   ptank_old -> v10m = ptank -> v10m;
   strncpy (ptank_old -> apid, ptank -> apid, sizeof (ptank_old -> apid));
   ptank_old -> esou = ptank -> esou;
   strncpy (ptank_old -> tsts, ptank -> tsts, sizeof (ptank_old -> tsts));
   ptank_old -> tasv = ptank -> tasv;
   ptank_old -> mlos = ptank -> mlos;
   ptank_old -> sgrv = ptank -> sgrv;
   ptank_old -> qvol = ptank -> qvol;
   ptank_old -> ttim = ptank -> ttim;
   ptank_old -> vtim = ptank -> vtim;
   ptank_old -> lalm = ptank -> lalm;
   ptank_old -> llal = ptank -> llal;
   ptank_old -> ltal = ptank -> ltal;
   ptank_old -> lbal = ptank -> lbal;
   ptank_old -> tcks = ptank -> tcks;
   ptank_old -> tval = ptank -> tval;
   ptank_old -> lvhh = ptank -> lvhh;
   ptank_old -> lvhi = ptank -> lvhi;
   ptank_old -> lvlo = ptank -> lvlo;
   ptank_old -> lvll = ptank -> lvll;
   strncpy (ptank_old -> pcmd, ptank -> pcmd, sizeof (ptank_old -> pcmd));
   strncpy (ptank_old -> rsts, ptank -> rsts, sizeof (ptank_old -> rsts));
   strncpy (ptank_old -> cmds, ptank -> cmds, sizeof (ptank_old -> cmds));
   ptank_old -> plev = ptank -> plev;
   ptank_old -> bswl = ptank -> bswl;
   ptank_old -> bswv = ptank -> bswv;
   ptank_old -> pbsw = ptank -> pbsw;
   ptank_old -> tpav = ptank -> tpav;
   ptank_old -> dnct = ptank -> dnct;
   ptank_old -> gcap = ptank -> gcap;
   ptank_old -> ullg = ptank -> ullg;
   ptank_old -> avap = ptank -> avap;
   ptank_old -> wcap = ptank -> wcap;
   ptank_old -> netv = ptank -> netv;
   ptank_old -> texp = ptank -> texp;
   ptank_old -> esin = ptank -> esin;
   ptank_old -> ciss = ptank -> ciss;
   ptank_old -> crec = ptank -> crec;
   ptank_old -> crun = ptank -> crun;
   ptank_old -> cstp = ptank -> cstp;
   ptank_old -> siss = ptank -> siss;
   ptank_old -> srec = ptank -> srec;
   ptank_old -> srun = ptank -> srun;
   ptank_old -> sstp = ptank -> sstp;
   ptank_old -> frat = ptank -> frat;
   ptank_old -> ini2 = ptank -> ini2;
   ptank_old -> latm = ptank -> latm;
   ptank_old -> mlst = ptank -> mlst;

   return;

} /* end func save_tank () */


/****************************************************************************
 *
 *  MODULE NAME            : set_roof_status 
 *  AUTHOR                 : emsteel
 *  DATE                   : 15 Apr 93
 *
 *  CALLING MODULE(S)      : init_record, process
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *
 *  DESCRIPTION: set tank roof status field
 *               if product level > roof float level
 *                  roof status is rfls (roof float string)
 *               else if product level < roof landed level)
 *                  roof status is rlls (roos landed string)
 *               else roof is rcrs (roof critical string)
 *  NOTE:
 *       see choice.h to explain the structs and the macros
 *       the variable -type- is static global to this source file and is
 *         an index that represents the tank record type. it is therefore
 *         an index into the choice sets for records.
 *       *pchoice ends up as a pointer to the roof type string
 *       see choiceRec.ascii for the def's of strings to roof types
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 04 Jan 94
 *  REVISION: chg choiceInd from short to unsigned short to agree with what
 *            rtyp is
 *
****************************************************************************/

static void set_roof_status (ptank)
   struct tankRecord   *ptank;
{
   short                choice_set = REC_TANK_ROOF;
   unsigned short       choiceInd  = ptank -> rtyp;
   struct arrChoiceSet  *parrChoiceSet;
   struct choiceSet     *pchoiceSet;
   char                 *pchoice;

   if   ((parrChoiceSet = GET_PARR_CHOICE_SET (choiceRec, type))
      && (pchoiceSet = GET_PCHOICE_SET (parrChoiceSet, choice_set))
      && (pchoice = GET_CHOICE (pchoiceSet, choiceInd)))
   {
      if (strncmp (pchoice, "not installed", 6) == 0)
         strncpy (ptank -> rsts, ptank -> rnis, sizeof (ptank -> rsts));
      else if (strncmp (pchoice, "fixed", 5) == 0)
         strncpy (ptank -> rsts, "fixed", sizeof (ptank -> rsts));

         /* roof type is floating */
      else
      {
         if (ptank -> plev >= ptank -> rflv)
            strncpy (ptank -> rsts, ptank ->rfls, sizeof (ptank -> rsts));
         else if (ptank -> plev <= ptank -> rllv)
            strncpy (ptank -> rsts, ptank -> rlls, sizeof (ptank -> rsts));
         else
            strncpy (ptank -> rsts, ptank -> rcrs, sizeof (ptank -> rsts));
      }
   }
   else
      recGblRecordError (S_db_badChoice, (struct dbCommon *)ptank, 
					"tank: set roof status");

   return;

} /* end func set_roof_status () */


/****************************************************************************
 *
 *  MODULE NAME            : special
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 Apr 93
 *
 *  CALLING MODULE(S)      : 
 *  ARGUMENT(S)            : paddr = ptr to dbAddr as defined in dbAccess.h
 *                         : after = not used here
 *
 *
 *  DESCRIPTION: special handling when a field changes etc
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 06 Jan 94
 *  REVISION: check for !after
 *
****************************************************************************/

static long special (paddr, after)
   struct dbAddr	*paddr;
   int		after;
{
   struct tankRecord	*ptank = (struct tankRecord *)(paddr -> precord);
   int	  special_type         = paddr -> special;
   long   status = 0;

   if (!after)
      return (0);

   switch (special_type)
   {
      case SPC_STRAP_TBL:    /* new strapping table */
         status = FindStrappingTable (ptank -> sttn, &ptank -> pstr, (int *)&ptank -> snum);
         break;

      case SPC_API_TBL:      /* new API table number */
         status = FindAPITable (ptank -> apit, &ptank -> papi, (int *)&ptank -> anum);
         break;

      default:
         recGblDbaddrError (S_db_badChoice, paddr, "tank: special");
         status = S_db_badChoice;
   }

   return (status);

} /* end func special () */


/****************************************************************************
 *
 *  MODULE NAME            : update_redun 
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 Apr 93
 *
 *  CALLING MODULE(S)      : process ()
 *  ARGUMENT(S)            : ptank_old = ptr t tank state at start of process
 *                         : ptank = ptr to tank start at end of process
 *
 *
 *  DESCRIPTION: have any of the fields deemed as being redundant changed
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 23 Feb 94
 *  REVISION:  removed ALST, MLST, ESIN  not needed
 *
****************************************************************************/

static void update_redun (ptank_old, ptank)
   struct tankred    *ptank_old;
   struct tankRecord *ptank;
{
   char  redbuf [REDQ_MAX_MSGLEN];

   redQInit ((struct dbCommon *)ptank, redbuf, sizeof (redbuf), type);

   if (ptank_old -> stat != ptank -> stat)
      redQField (redbuf, mylist[STAT]);

   if (ptank_old -> sevr != ptank -> sevr)
      redQField (redbuf, mylist[SEVR]);

   if (ptank_old -> val != ptank -> val)
      redQField (redbuf, mylist[VAL]);

   if (ptank_old -> v01m != ptank -> v01m)
      redQField (redbuf, mylist[V01M]);

   if (ptank_old -> v02m != ptank -> v02m)
      redQField (redbuf, mylist[V02M]);

   if (ptank_old -> v03m != ptank -> v03m)
      redQField (redbuf, mylist[V03M]);

   if (ptank_old -> v04m != ptank -> v04m)
      redQField (redbuf, mylist[V04M]);

   if (ptank_old -> v05m != ptank -> v05m)
      redQField (redbuf, mylist[V05M]);

   if (ptank_old -> v06m != ptank -> v06m)
      redQField (redbuf, mylist[V06M]);

   if (ptank_old -> v07m != ptank -> v07m)
      redQField (redbuf, mylist[V07M]);

   if (ptank_old -> v08m != ptank -> v08m)
      redQField (redbuf, mylist[V08M]);

   if (ptank_old -> v09m != ptank -> v09m)
      redQField (redbuf, mylist[V09M]);

   if (ptank_old -> v10m != ptank -> v10m)
      redQField (redbuf, mylist[V10M]);

   if (strncmp (ptank_old -> apid, ptank -> apid, 
                                   sizeof (ptank_old -> apid)) != 0)
      redQField (redbuf, mylist[APID]);

   if (ptank_old -> esou != ptank -> esou)
      redQField (redbuf, mylist[ESOU]);

   if (strncmp (ptank_old -> tsts, ptank -> tsts, 
                                   sizeof (ptank_old -> tsts)) != 0)
      redQField (redbuf, mylist[TSTS]);

   if (ptank_old -> tasv != ptank -> tasv)
      redQField (redbuf, mylist[TASV]);

   if (ptank_old -> mlos != ptank -> mlos)
      redQField (redbuf, mylist[MLOS]);

   if (ptank_old -> sgrv != ptank -> sgrv)
      redQField (redbuf, mylist[SGRV]);

   if (ptank_old -> qvol != ptank -> qvol)
      redQField (redbuf, mylist[QVOL]);

   if (ptank_old -> ttim != ptank -> ttim)
      redQField (redbuf, mylist[TTIM]);

   if (ptank_old -> vtim != ptank -> vtim)
      redQField (redbuf, mylist[VTIM]);

   if (ptank_old -> lalm != ptank -> lalm)
      redQField (redbuf, mylist[LALM]);

   if (ptank_old -> llal != ptank -> llal)
      redQField (redbuf, mylist[LLAL]);

   if (ptank_old -> ltal != ptank -> ltal)
      redQField (redbuf, mylist[LTAL]);

   if (ptank_old -> lbal != ptank -> lbal)
      redQField (redbuf, mylist[LBAL]);

   if (ptank_old -> tval != ptank -> tval)
      redQField (redbuf, mylist[TVAL]);

   if (ptank_old -> lvhh != ptank -> lvhh)
      redQField (redbuf, mylist[LVHH]);

   if (ptank_old -> lvhi != ptank -> lvhi)
      redQField (redbuf, mylist[LVHI]);

   if (ptank_old -> lvlo != ptank -> lvlo)
      redQField (redbuf, mylist[LVLO]);

   if (ptank_old -> lvll != ptank -> lvll)
      redQField (redbuf, mylist[LVLL]);

   if (strncmp (ptank_old -> pcmd, ptank -> pcmd, 
                                   sizeof (ptank_old -> pcmd)) != 0)
      redQField (redbuf, mylist[PCMD]);

   if (strncmp (ptank_old -> cmds, ptank -> cmds, 
                                   sizeof (ptank_old -> cmds)) != 0)
      redQField (redbuf, mylist[CMDS]);

   if (ptank_old -> ini2 != ptank -> ini2)
      redQField (redbuf, mylist[INI2]);

   if (ptank_old -> ciss != ptank -> ciss)
      redQField (redbuf, mylist[CISS]);

   if (ptank_old -> crec != ptank -> crec)
      redQField (redbuf, mylist[CREC]);

   if (ptank_old -> crun != ptank -> crun)
      redQField (redbuf, mylist[CRUN]);

   if (ptank_old -> cstp != ptank -> cstp)
      redQField (redbuf, mylist[CSTP]);

   if (ptank_old -> siss != ptank -> siss)
      redQField (redbuf, mylist[SISS]);

   if (ptank_old -> srec != ptank -> srec)
      redQField (redbuf, mylist[SREC]);

   if (ptank_old -> srun != ptank -> srun)
      redQField (redbuf, mylist[SRUN]);

   if (ptank_old -> sstp != ptank -> sstp)
      redQField (redbuf, mylist[SSTP]);

    if (TsCmpStampsNE (&ptank_old -> latm, &ptank -> latm) )
	redQField (redbuf, mylist [LATM]);

    if (ptank_old -> mlst != ptank -> mlst)
	redQField (redbuf, mylist [MLST]);
	
   redQSend (redbuf);

   return;

} /* end func update_redun () */


/****************************************************************************
 *
 *  MODULE NAME            : update_qvol 
 *  AUTHOR                 : emsteel
 *  DATE                   : 17 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *                           ptank_old - ptr to last scan
 *
 *  DESCRIPTION: ptank -> ttim is a timer that counts time from a stop
 *               command until ptank -> tset which is a settle time,
 *               the time for the product in the tank to settle before
 *               the quiescent volume (ptank -> qvol) is updated with the
 *               current product net volume (ptank -> netv)
 *
****************************************************************************/

static void update_qvol (ptank, ptank_old)
   struct tankRecord    *ptank;
   struct tankred       *ptank_old;
{
   double delta_time;

      /* if not timing, do nothing */

   if (ptank -> ttim < 0.0)
      return;

      /* diff in secs tween last scan and current */

   delta_time = TFTicksDiffInSeconds (&ptank -> tcks, &ptank_old -> tcks);

   ptank -> ttim += delta_time;

      /* if = or exceeded the settling time, update qvol */

   if (ptank -> ttim >= ptank -> tset)
   {
      ptank -> ttim = -1.0;

	 /* if not already alarming */

      if (!ptank -> tval)
      {
         ptank -> qvol = ptank -> netv;
      }
   }

   return;

} /* end func update_qvol () */


/****************************************************************************
 *
 *  MODULE NAME            : update_vtim 
 *  AUTHOR                 : emsteel
 *  DATE                   : 18 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *                           ptank_old - ptr to last scan
 *
 *  DESCRIPTION: ptank -> vtim is a timer that times the transition
 *               of the state of the tank before we check on the
 *               status of the tank's associated valves
 *               vtim is set to 0 in run_commands, < 0 means finished
 *               timing. 
 *
****************************************************************************/

static void update_vtim (ptank, ptank_old)
   struct tankRecord    *ptank;
   struct tankred       *ptank_old;
{
   double   delta_time = 0.0;
   double   timeout = 0.0;

      /* if not timing, do nothing */

   if (ptank -> vtim < 0.0)
      return;

      /* diff in secs tween last scan and current */

   delta_time = TFTicksDiffInSeconds (&ptank -> tcks, &ptank_old -> tcks);

   ptank -> vtim += delta_time;

   switch (ptank -> tasv)
   {
      case TANK_ISSUING:
         timeout = ptank -> isto;
         break;

      case TANK_RECEIVING:
         timeout = ptank -> reto;
         break;

      case TANK_RUNNING:
         timeout = ptank -> ruto;
         break;

      case TANK_STOPPED:
         timeout = ptank -> stto;
         break;
   }
      /* if = or exceeded the transition time, finished timing */

   if (ptank -> vtim >= timeout)
      ptank -> vtim = -1.0;

   return;

} /* end func update_vtim () */


/****************************************************************************
 *
 *  MODULE NAME            : valve_status_alarms 
 *  AUTHOR                 : emsteel
 *  DATE                   : 21 Apr 93
 *
 *  CALLING MODULE(S)      : alarm()
 *          ARGUMENT(S)    : ptank - ptr to tank record
 *                           estop - ptr emergency stop flag
 *
 *  DESCRIPTION: the dbAddr for each valve was set in init_valves ()
 *
 *               see if state of tank valves matches state of the tank
 *
 *               vsts is a ptr to a description of what state a valve
 *               should be in based on the state of the tank
 *
 *               don't check stopped tank for num valves opened,
 *               there may or not be a relief valve for example
 *               that may or may not be opened
 *
 *               if (ptank -> eves) set estop to 1
 *
****************************************************************************/

static void valve_status_alarms (ptank, estop)
   struct tankRecord   *ptank;
   int                 *estop;
{
   short             valve_status[MAX_TANK_VALVES];  
   unsigned short    *vsts              = NULL;
   long              nRequests          = 1;
   long              options            = 0;
   long              status             = 0;
   short             num_opened_valves  = 0;
   int               num_valves         = 0;
   short             valve_open_error   = 0;
   short             valve_status_error = 0;
   struct dbAddr     *pdbAddr           = &ptank -> pv1_dbAddr;
   int               i;

      /* get status of tanks' valves from valves' record, or
         set to NO_VALVE if not there
      */

   for (i = 0; i < MAX_TANK_VALVES; i++)
   {
      if (pdbAddr -> precord)
      {
         if ((status = dbGetField (pdbAddr, DBR_SHORT,
                                  (caddr_t)&valve_status[i],
                                  &options, &nRequests, NULL)) != 0)
         {
            TankFarmPostAlarm ((struct dbCommon *)ptank, LINK_ALARM, 
                                                               VALID_ALARM);
            valve_status[i] = NO_VALVE;
         }
         else
         {
            if ((valve_status[i] == VALVE_OPENED) ||
                (valve_status[i] == VALVE_OPENING) ||
                (valve_status[i] == VALVE_JOG_OPEN) ||
                (valve_status[i] == VALVE_POSITION_OPENING))
                  num_opened_valves++;
	    num_valves++;
         }
      }
      else
         valve_status[i] = NO_VALVE;
      pdbAddr++;
   }

      /* if no valves, no need to go on */

   if (!num_valves)
      return;

      /* check for proper number of valves opened given that
            operational state of the tank
         also point to the required status of the 1st valve
      */

   switch (ptank -> tasv)
   {
      case TANK_ISSUING:
         if (num_opened_valves < 1)
            valve_open_error++;
         vsts = &ptank -> v01i;
         break;
      case TANK_RECEIVING:
         if (num_opened_valves < 1)
            valve_open_error++;
         vsts = &ptank -> v01r;
         break;
      case TANK_RUNNING:
         if (num_opened_valves < 2)
            valve_open_error++;
         vsts = &ptank -> v01n;
         break;

      case TANK_STOPPED:
         vsts = &ptank -> v01s;
         break;
   }

   if (vsts)
   {
      for (i = 0; i < MAX_TANK_VALVES; i++, vsts++)
      {
         if (valve_status[i] != NO_VALVE)
         {
            switch (*vsts)
            {
               case TANK_VALVE_CLOSED:
                  if ((valve_status[i] != VALVE_CLOSED) &&
                      (valve_status[i] != VALVE_CLOSING))
                     valve_status_error++;
                  break;
               case TANK_VALVE_OPENED:
                  if ((valve_status[i] != VALVE_OPENED) &&
                      (valve_status[i] != VALVE_OPENING))
                     valve_status_error++;
                  break;
               case TANK_VALVE_UNSPECIFIED:
                  if (valve_status[i] == VALVE_FAULT)
                     valve_status_error++;
                  break;
            }
            if (valve_status_error)
               break;
         }
      }
   }

   if (valve_open_error || valve_status_error)
   {
      if (ptank -> nsev < ptank -> evss)
      {
         ptank -> nsev = ptank -> evss;
         ptank -> nsta = TANK_VALVE;
      }
      if (ptank -> eves)
      {
         *estop = 1;
      }
   }

   return;

} /* end func valve_status_alarms () */

