/*tankfarm.c,v 1.1.1.1 1999/08/03 14:44:23 tis4000 Exp */

/* tankfarm.c */
/* share/src/rec  @(#)tankfarm.c   1.00 3/24/93 */
  
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************************
 *
 *  MODULE NAME            : tankfarm.c
 *  AUTHOR                 : emsteel
 *  DATE                   : 24 Mar 93
 *
 *
 *  DESCRIPTION: routines needed to support tankfarm block
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 10 dec 93
 *  REVISION:  added several sprintfs to errMessage to make them more
 *             meaningful
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 19 May 94
 *  REVISION:  make FindStrappingTable and FindAPITable return long not void
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 11 Jul 94
 *  REVISION:	ComputeTFFlowRate(), if delta_time is 0, set flow rate to 0
 *		ComputeTFLevel(), if (hvol - lowv) == 0, set current level to
 *		    hsen (high sensor value from table)
 *		ComputeTFVolume(), if (hsen - lsen) == 0, set current volume
 *		    to hvol (high volume value from table)
 *		alleviate potential divide by 0 situations
 *
****************************************************************************/

/****************************************************************************
*
* Programmer's Notes:
*
*  strapping table structure and api table structures defined in tanktbls.h
*    which is included in tankfarm.h
*
****************************************************************************/

#include <vxWorks.h>
#include <sysLib.h>
#include <symLib.h>
#include <sysSymTbl.h>     /* for sysSymTbl */
#include <stat.h>          /* for stat ()    */
#include <types.h>
#include <stdioLib.h>
#include <string.h>
#include <lstLib.h>
#include <math.h>

#include "alarm.h"
#include "caeventmask.h"
#include "choice.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"
#include "errMdef.h"
#include "tankfarm.h"
#include "error.h"

typedef unsigned char UTINY;



/****************************************************************************
 *
 *  MODULE NAME            : ComputeTFFlowRate 
 *  AUTHOR                 : emsteel
 *  DATE                   : 26 May 93
 *
 *  ARGUMENT(S) :
 *
 *  DESCRIPTION: compute a flow rate
 *
 *  NOTE:  flow rate = volume change / unit of time
 *         flow rate = (change in volume * time value) / change in time
 *  
 *  NOTE: conversion values used
 *        brls -> gals   = brls * 42
 *        brls -> liters = brls * 42 * 3.7854 (convert to gals then liters)
 *        brls -> f3     = brls * 42 * 0.1337  (convert to gals then f3)
 *        brls -> m3     = brls * 42 * 0.003785 (convert to gals then m3)
 *        gals -> brls   = gals * 0.023809523
 *        gals -> liters = gals * 3.7854
 *        gals -> f3     = gals * 0.1337
 *        gals -> m3     = gals * 0.003785
 *        liters -> brls = liters * 0.2642 * 0.023809523 (gals then brls)
 *        liters -> gals = liters * 0.2642
 *        liters -> f3   = liters * 0.03531
 *        liters -> m3   = liters * 0.001
 *        f3 -> brls     = f3 * 7.48052 * 0.023809523 (gals then brls)
 *        f3 -> gals     = f3 * 7.48052
 *        f3 -> liters   = f3 * 28.32
 *        f3 -> m3       = f3 * 0.02832
 *        m3 -> brls     = m3 * 264.2 * 0.023809523 (gals then brls)
 *        m3 -> gals     = m3 * 264.2
 *        m3 -> liters   = m3 * 1000.0
 *        m3 -> f3       = m3 * 35.31
 *        Conversions taken from Instrument Engineers handbook
 *           Vol 1 table A.1 Conversion Factors
 *
****************************************************************************/

long ComputeTFFlowRate (recname, vol_field_name, flow_field_name, vol_units,
                        flow_units, curr_volume, prev_volume,
                        curr_ticks, last_ticks, flow_rate)

     char               *recname;         /* name of record */
     char               *vol_field_name;  /* name of the volume units field */
     char               *flow_field_name; /* name of the flow rate units field */
     unsigned short     vol_units;        /* index to volume units list */
     unsigned short     flow_units;       /* index to flow rate units list */
     double             curr_volume;     
     double             prev_volume;
     unsigned long      *curr_ticks;      /* system time ticks */
     unsigned long      *last_ticks;      /* system time ticks */
     double             *flow_rate;       /* return the flow rate */
{
   long        status = 0;
   double      delta_vol;     /* volume chg between scans */
   double      delta_time;    /* time in seconds between scans */
   double      time_value;    /* time correction */
   char        *time_units;   /* time units of measure (sec/min/hr) */
   struct dbAddr pdbAddr;     /* structure for record data */
   char        pvname[34];
   char        errmsg[128];
                              /* stuff for choiceSets */
   struct choiceSet  *pchoiceSet;
   short             vol_choice_set;
   short             flow_rate_choice_set;
   char              *pchoice_vol;
   char              *pchoice_flow;


   *flow_rate = 0;

      /* get choice set for the volume units */

   strncpy (pvname, recname, sizeof (pvname));
   strcat (pvname, ".");
   strcat (pvname, vol_field_name);
   if ((status = dbNameToAddr (pvname, &pdbAddr)) != 0)
   {
      sprintf (errmsg, "FlowRate: PV.name:%.32s",pvname);
      errMessage (S_db_notFound, errmsg);
      return (S_db_notFound);
   }
   vol_choice_set = pdbAddr.choice_set;

      /* get the choice set for the flow rate units */

   strncpy (pvname, recname, sizeof (pvname));
   strcat (pvname, ".");
   strcat (pvname, flow_field_name);
   if ((status = dbNameToAddr (pvname, &pdbAddr)) != 0)
   {
      sprintf (errmsg, "FlowRate: PV.name:%.32s",pvname);
      errMessage (S_db_notFound, errmsg);
      return (S_db_notFound);
   }
   flow_rate_choice_set = pdbAddr.choice_set;


      /* get ptr to volume units choice */

   if ( (!(pchoiceSet = GET_PCHOICE_SET(choiceGbl, vol_choice_set)))
     || (!(pchoice_vol = GET_CHOICE (pchoiceSet, vol_units))))
   {
      sprintf (errmsg, "FlowRate, pv %.28s: vol choice", recname);
      errMessage (S_db_badChoice, errmsg);
      return (S_db_badChoice);
   }

      /* get ptr to flow rate units choice */

   if ( (!(pchoiceSet = GET_PCHOICE_SET(choiceGbl, flow_rate_choice_set)))
     || (!(pchoice_flow = GET_CHOICE (pchoiceSet, flow_units))))
   {
      sprintf (errmsg, "FlowRate, pv %.28s: flow rate choice", recname);
      errMessage (S_db_badChoice, errmsg);
      return (S_db_badChoice);
   }
      /* flow rate units form vol/unit of time */

   if ((time_units = strchr (pchoice_flow, '/')) == NULL)
   {
      sprintf (errmsg, "FlowRate, pv %.28s: time units", recname);
      errMessage (S_db_badChoice, errmsg);
      return (S_db_badChoice);
   }
   time_units++;

      /* if the volume units of measure are not the same
         conversions are neccessary
      */

   delta_vol = curr_volume - prev_volume;
   if (delta_vol < 0.0)
      delta_vol = -delta_vol;

   if (strncmp (pchoice_vol, pchoice_flow, 2) != 0)
   {
      if (strncmp (pchoice_vol, "brls", 4) == 0)
      {
         if (strncmp (pchoice_flow, "gals", 4) == 0)
            delta_vol = delta_vol * 42;
         else if (strncmp (pchoice_flow, "liters", 6) == 0)
            delta_vol = delta_vol * 42 * 3.7854;
         else if (strncmp (pchoice_flow, "f3", 2) == 0)
            delta_vol = delta_vol * 42 * 0.1337;
         else if (strncmp (pchoice_flow, "m3", 2) == 0)
            delta_vol = delta_vol * 42 * 0.003785;
         else
         {
            sprintf (errmsg, "FlowRate, pv %.28s: flow", recname);
            errMessage (S_db_badChoice, errmsg);
            return (S_db_badChoice);
         }
      }
      else if (strncmp (pchoice_vol, "gals", 4) == 0)
      {
         if (strncmp (pchoice_flow, "brls", 4) == 0)
            delta_vol = delta_vol * 0.023809523;
         else if (strncmp (pchoice_flow, "liters", 6) == 0)
            delta_vol = delta_vol * 3.7854;
         else if (strncmp (pchoice_flow, "f3", 2) == 0)
            delta_vol = delta_vol * 0.1337;
         else if (strncmp (pchoice_flow, "m3", 2) == 0)
            delta_vol = delta_vol * 0.003785;
         else
         {
            sprintf (errmsg, "FlowRate, pv %.28s: flow", recname);
            errMessage (S_db_badChoice, errmsg);
            return (S_db_badChoice);
         }
      }
      else if (strncmp (pchoice_vol, "liters", 6) == 0)
      {
         if (strncmp (pchoice_flow, "brls", 4) == 0)
            delta_vol = delta_vol * 0.2642 * 0.023809523;
         else if (strncmp (pchoice_flow, "gals", 4) == 0)
            delta_vol = delta_vol * 0.2642;
         else if (strncmp (pchoice_flow, "f3", 2) == 0)
            delta_vol = delta_vol * 0.03531;
         else if (strncmp (pchoice_flow, "m3", 2) == 0)
            delta_vol = delta_vol * 0.001;
         else
         {
            sprintf (errmsg, "FlowRate, pv %.28s: flow", recname);
            errMessage (S_db_badChoice, errmsg);
            return (S_db_badChoice);
         }
      }
      else if (strncmp (pchoice_vol, "f3", 2) == 0)
      {
         if (strncmp (pchoice_flow, "brls", 4) == 0)
            delta_vol = delta_vol * 7.48052 * 0.023809523;
         else if (strncmp (pchoice_flow, "gals", 4) == 0)
            delta_vol = delta_vol * 7.48052;
         else if (strncmp (pchoice_flow, "liters", 6) == 0)
            delta_vol = delta_vol * 28.32;
         else if (strncmp (pchoice_flow, "m3", 2) == 0)
            delta_vol = delta_vol * 0.02832;
         else
         {
            sprintf (errmsg, "FlowRate, pv %.28s: flow", recname);
            errMessage (S_db_badChoice, errmsg);
            return (S_db_badChoice);
         }
      }
      else if (strncmp (pchoice_vol, "m3", 2) == 0)
      {
         if (strncmp (pchoice_flow, "brls", 4) == 0)
            delta_vol = delta_vol * 264.2 * 0.023809523;
         else if (strncmp (pchoice_flow, "gals", 4) == 0)
            delta_vol = delta_vol * 264.2;
         else if (strncmp (pchoice_flow, "liters", 6) == 0)
            delta_vol = delta_vol * 1000.0;
         else if (strncmp (pchoice_flow, "f3", 2) == 0)
            delta_vol = delta_vol * 35.31;
         else
         {
            sprintf (errmsg, "FlowRate, pv %.28s: flow", recname);
            errMessage (S_db_badChoice, errmsg);
            return (S_db_badChoice);
         }
      }
      else
      {
         sprintf (errmsg, "FlowRate, pv %.28s: flow", recname);
         errMessage (S_db_badChoice, errmsg);
         return (S_db_badChoice);
      }

   } /* end if (vol != flow) */

      /* returns the diff in seconds between scans */

   delta_time = TFTicksDiffInSeconds (curr_ticks, last_ticks);

   if (strncmp (time_units, "sec", 3) == 0)
      time_value = 1;
   else if (strncmp (time_units, "min", 3) == 0)
      time_value = 60;     /* 60 seconds/min */
   else if (strncmp (time_units, "hr", 2) == 0)
      time_value = 3600;   /* 60 * 60, 60 sec/min 60 min/hr */
   else
   {
      sprintf (errmsg, "FlowRate, pv %.28s: time value", recname);
      errMessage (S_db_badChoice, errmsg);
      return (S_db_badChoice);
   }

      /* compute curr flow rate */

   if (delta_time == 0)
      *flow_rate = 0;
   else
      *flow_rate = (delta_vol * time_value) / delta_time;

   return (status);

} /* end func ComputeTFFlowRate () */



/****************************************************************************
 *
 *  MODULE NAME            : ComputeTFLevel  
 *  AUTHOR                 : emsteel
 *  DATE                   : 6 Apr 93
 *
 *
 *  DESCRIPTION: given a volume and  a ptr to a strapping table, interpolate
 *               the level
 *
 *  NOTES:
 *
 *   Formula to interpolate volume
 *  
 *   hsen - curr_level   hvol - curr_volume
 *   ------------------ = ------------------ 
 *   hsen - lsen         hvol - lvol      
 *
 *
 *                          hvol - curr_volume
 *  *curr_level = hsen - (( ------------------ ) * (hsen - lsen))
 *                          hvol - lvol
 *
****************************************************************************/

void ComputeTFLevel (pstr, num_entries, curr_level, curr_volume)
     caddr_t   pstr;          /* ptr to strapping table */
     int       num_entries;   /* number of entries in table */
     double    *curr_level;   /* return interpolated level */
     double    curr_volume;   /* value read from volume sensor */
{

   struct strappingValue *stbl = (struct strappingValue *)pstr;
   register int      i;
   register double   lsen;         /* low sensor value */
   register double   lvol;         /* low volume  */
   register double   hsen;         /* high sensor */
   register double   hvol;         /* high volume */
   register double   shsen;        /* save high sensor value */
   register double   shvol;        /* save high volume value */

      /* if no strapping table */

   if (!stbl)
      return;

      /* init to 1st value in strapping table */

   hsen = lsen = stbl -> sensorValue;
   hvol = lvol = stbl -> volume;

      /* if the current volume is less than the 1st entry in the strapping
         table, then return 0.0 for the level
      */

   if (curr_volume < lvol)
   {
      *curr_level = 0.0;
      return;
   }

      /* this will prevent a divide by 0 */

   if (curr_volume == lvol)
   {
      *curr_level = lsen;
      return;
   }
      /* find the low and high values that bracket the current sensor value */

   for (i = 1; i <= num_entries; i++)
   {
      if (curr_volume > hvol)
      {
         /* is there another table value */

         if (( i + 1) <= num_entries)
         {
               /* save current high values */

            shsen = hsen;
            shvol = hvol;

               /* get new high values */

	    stbl++;
            hsen = stbl -> sensorValue;
            hvol = stbl -> volume;

               /* update low values with old high values */

            lsen = shsen;
            lvol = shvol;
         }
      }
      else
         break;
   }
   if ((hvol - lvol) == 0)
      *curr_level = hsen;
   else
      *curr_level = hsen - (((hvol - curr_volume) /
                       (hvol - lvol)) * (hsen - lsen));
   return;

} /* end func ComputeTFLevel () */


/****************************************************************************
 *
 *  MODULE NAME            : ComputeTFVolume 
 *  AUTHOR                 : emsteel
 *  DATE                   : 24 Mar 93
 *
 *
 *  DESCRIPTION: given a level and  a ptr to a strapping table, interpolate
 *               the volume
 *
 *  NOTES:
 *
 *   Formula to interpolate volume
 *  
 *   hsen - curr_sval    hvol - curr_volume
 *   ----------------- = ------------------ 
 *   hsen - lsen         hvol - lvol       
 *                                        
 *  
 *                            hsen - curr_sval
 *   *curr_volume = hvol - (( ----------------- ) * ( hvol - lvol))
 *                            hsen - lsen
 *
 *
****************************************************************************/

void ComputeTFVolume (pstr, num_entries, curr_sval, curr_volume)
     caddr_t   pstr;             /* ptr to stapping table */
     int       num_entries;      /* number entries in table */
     double    curr_sval;        /* current sensor value */
     double    *curr_volume;     /* return the interpolated volume */
{
   struct strappingValue *stbl = (struct strappingValue *)pstr;
   register int      i;
   register double   lsen;         /* low sensor value */
   register double   lvol;         /* low volume  */
   register double   hsen;         /* high sensor */
   register double   hvol;         /* high volume */
   register double   shsen;        /* save high sensor value */
   register double   shvol;        /* save high volume value */

      /* if no strapping table */

   if (!stbl)
      return;

      /* init to 1st value in strapping table */

   hsen = lsen = stbl -> sensorValue;
   hvol = lvol = stbl -> volume;

      /* if current sensor value is less than the 1st entry in the strapping
         table, then return 0.0
      */

   if (curr_sval < lsen)
   {
      *curr_volume = 0.0;
      return;
   }
      /* this will prevent divide by 0 */

   if (curr_sval == lsen)
   {
      *curr_volume = lvol;
      return;
   }

      /* find the low and high values that bracket the current sensor value */

   for (i = 1; i <= num_entries; i++)
   {
      if (curr_sval > hsen)
      {
            /* is there another value in the table */

         if ((i + 1) <= num_entries)
         {
               /* save old high values */

            shsen = hsen;  
            shvol = hvol;

               /* get new high values */

	    stbl++;
            hsen = stbl -> sensorValue;
            hvol = stbl -> volume;

               /* update low values with old high values */

            lsen = shsen;
            lvol = shvol;
         }
         else
            break;
      }
      else
         break;
   }

   if ((hsen - lsen) == 0)
      *curr_volume = hvol;
   else
      *curr_volume = hvol - (((hsen - curr_sval) /
                                         (hsen - lsen)) * (hvol - lvol));

   return;

} /* end func ComputeTFVolume () */



/****************************************************************************
 *
 *  MODULE NAME            : FindAPIcf 
 *  AUTHOR                 : emsteel
 *  DATE                   : 14 May 93
 *
 *
 *  DESCRIPTION: find the API thermal correction factor for a product
 *               of given temperature and specific gravity
 *
****************************************************************************/

long FindAPIcf (api_tbl, num_entries, ave_temp, sp_grav, texp,
                recname, temp_field_name, temp_units)

     caddr_t      api_tbl;          /* ptr to API table */
     int          num_entries;      /* number of table entries */
     double       ave_temp;         /* average product temperature */
     double       sp_grav;          /* specific gravity of product */
     double       *texp;            /* return themal expansion value */
     char         *recname;         /* name of record */
     char         *temp_field_name; /* ptr to name of temp units field */
     short        temp_units;       /* current temperature units */
{
   register          i;           /* loop control */
   double            tpav;        /* ave. temp rounded */
   double            sgrv;        /* spec grav rounded */
   long              status = 0;
   struct APIValue   *APIval;    /* stuff to access API table */
   struct dbAddr     pdbAddr;    /* for record packet */
   char              pvname[34];
                                 /* stuff for choice sets */
   struct choiceSet  *pchoiceSet;   /* temperature list */
   char              *pchoice_temp; /* ptr to temp units */
   char              errmsg[128];

   *texp = 0;
   APIval = (struct APIValue *)api_tbl;

      /* if no API table do nothing */

   if (!APIval)
      return(status);

   strncpy (pvname, recname, sizeof (pvname));
   strcat (pvname, ".");
   strcat (pvname, temp_field_name);

   if ((status = dbNameToAddr (pvname, &pdbAddr)) != 0)
   {
      sprintf (errmsg, "FindAPIcf: pvname %.32s", pvname);
      errMessage (S_db_notFound, errmsg);
      return (S_db_notFound);
   }

   if ( (!(pchoiceSet = GET_PCHOICE_SET (choiceGbl, pdbAddr.choice_set)))
     || (!(pchoice_temp = GET_CHOICE (pchoiceSet, temp_units))) )
   {
      sprintf (errmsg, "FindAPIcf, pvname %.28s: temp units", recname);
      errMessage (S_db_badChoice, errmsg);
      return (S_db_badChoice);
   }

      /* convert the ave temp to degF if not already in degF */

   if (strncmp (pchoice_temp, "degC", 4) == 0)
      tpav = ((ave_temp * 9.0) / 5.0) + 32.0;
   else
      tpav = ave_temp;

      /* round to nearest 0.5 */

   tpav = TankFarmRound (tpav);
   sgrv = TankFarmRound (sp_grav);


   for (i = 0; i < num_entries; i++, APIval++)
   {
      if ((APIval -> APITemperature == tpav) &&
          (APIval -> APIGravity == sgrv))
      {
         *texp = APIval -> APIConversionFactor;
         break;
      }
   }

   return(status);

} /* end func FindAPIcf () */



/****************************************************************************
 *
 *  MODULE NAME            : FindAPITable
 *  AUTHOR                 : emsteel
 *  DATE                   : 26 Apr 93
 *
 *  DESCRIPTION: given an API table number, find the table and init a ptr
 *               to it. the symbol _sizeof... contains the size of the
 *               data in bytes
 *
****************************************************************************/

long FindAPITable (apit, papi, num_entries)
     char      *apit;         /* ptr to API table name   */
     caddr_t   *papi;          /* place ptr to table here */
     int       *num_entries;  /* place num table entries here */
{
   
   long     rtnval;
   char     *pvalue;
   int      *psvalue;
   UTINY    type;
   char     symName[32];      /* name of the symbol name */
   char     errmsg[128];

   *papi = NULL;
   *num_entries = 0;

   if (!apit[0])
      return (0);

   strcpy (symName, "_");
   strcat (symName, apit);
   if ((rtnval = symFindByName (sysSymTbl, symName, &pvalue, &type)) != OK)
   {
      sprintf (errmsg, "Can not find API Tbl -%.32s-", symName);
      errMessage (S_tank_APITbl_Err, errmsg);
      return (S_tank_APITbl_Err);
   }

   strcpy (symName, "_sizeof");
   strcat (symName, apit);

   if ((rtnval = symFindByName (sysSymTbl, symName, (char **)&psvalue, &type)) != OK)
   {
      sprintf (errmsg, "Can not find API Tbl size -%.32s-", symName);
      errMessage (S_tank_APITbl_Err, errmsg);
      return (S_tank_APITbl_Err);
   }

   *papi = pvalue;
   *num_entries = (*psvalue) / sizeof (struct APIValue);

   return (0);

}



/****************************************************************************
 *
 *  MODULE NAME            : FindStrappingTable
 *  AUTHOR                 : emsteel
 *  DATE                   : 6 Apr 93
 *
 *
 *  DESCRIPTION: given the name of a strapping table, find the table and
 *               init a ptr to it. _sizeof... is the symbol name to the
 *               size of the table in bytes
 *
****************************************************************************/

long FindStrappingTable (tablename, pstr, num_entries)
     char      *tablename;
     caddr_t   *pstr;             /* place ptr to table here */
     int       *num_entries;     /* place num table entries here */
{
   long     rtnval;
   char     *pvalue;
   int      *psvalue;
   UTINY    type;
   char     symName[32];      /* name of the symbol size name */
   char     errmsg[128];

   *pstr = NULL;
   *num_entries = 0;

   if (!tablename[0])
      return (0);

   strcpy (symName, "_");
   strcat (symName, tablename);
   if ((rtnval = symFindByName (sysSymTbl, symName, &pvalue, &type)) != OK)
   {
      sprintf (errmsg, "Can not find St Tbl -%.32s-", symName);
      errMessage (S_tank_StTbl_Err, errmsg);
      return (S_tank_StTbl_Err);
   }

   strcpy (symName, "_sizeof");
   strcat (symName, tablename);

   if ((rtnval = symFindByName (sysSymTbl, symName, (char **)&psvalue, &type)) != OK)
   {
      sprintf (errmsg, "Can not find St Tbl size -%.32s-", symName);
      errMessage (S_tank_StTbl_Err, errmsg);
      return (S_tank_StTbl_Err);
   }

   *pstr = pvalue;
   *num_entries = (*psvalue) / sizeof (struct strappingValue);

   return (0);

} /* end func FindStrappingTable () */



/****************************************************************************
 *
 *  MODULE NAME            : GetTFCommand 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 June 93
 *
 *          ARGUMENT(S)    : ptfcmd - ptr to tankfarm command structure
 *                                    see tankfarm.h
 *                         : cmd_type - ptr to type of command
 *
 *  DESCRIPTION: return the command for a tankfarm record
 *               set cmd_type
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 dec 93
 *  REVISION:  added code to process special digital commands
 *             these are cmds set by other points while tf point is in
 *             AUTOMATIC. Normally only string cmds allowed in AUTOMATIC
****************************************************************************/

int GetTFCommand (ptfcmd, cmd_type)
   struct tfcmd   *ptfcmd;
   int            *cmd_type;
{
   int   cmd    = NO_TF_COMMAND;
   short *pdig  = ptfcmd -> FirstDigitalCmd;
   short *psdig = ptfcmd -> FirstSpecialDigitalCmd;
   char  *pstr  = ptfcmd -> FirstCmdString;
   int   i;

   *cmd_type = TF_NO_CMD_TYPE;

      /* if no string command, check for digital */

   if ((*ptfcmd -> CmdString) == NULL)
   {
      for ( i = 1; i <= ptfcmd -> NumCmds; i++, pdig++)
      {
         if (*pdig)
         {
            cmd = i;
            *cmd_type = TF_DIGITAL_CMD;
            break;
         }
      }
         /* check for special digital if not regular digital */

      if ((*cmd_type) != TF_DIGITAL_CMD)
      {
         for ( i = 1; i <= ptfcmd -> NumCmds; i++, psdig++)
         {
            if (*psdig)
            {
               cmd = i;
               *cmd_type = TF_SPECIAL_DIGITAL_CMD;
               break;
            }
         }
      }
   }
      /* match cmd to strings in record */

   else
   {
      for (i = 1; i <= ptfcmd -> NumCmds; i++, pstr += 18)
      {
         if (strncmp (pstr, ptfcmd -> CmdString, 18) == 0)
         {
            *cmd_type = TF_STRING_CMD;
            return (i);
         }
      }
      cmd = ILLEGAL_TF_STRING_CMD;
   }

   return (cmd);

} /* end func GetTFCommand () */



/****************************************************************************
 *
 *  MODULE NAME            : ResetTFCommand 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 June 93
 *
 *          ARGUMENT(S)    : ptfcmd - ptr to tankfarm command structure
 *                                    see tankfarm.h
 *
 *  DESCRIPTION: reset a command for a tankfarm record
 *  Notes:
 *          reset all digital cmds to 0, and zero out the command string
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 dec 93
 *  REVISION:  added code to process special digital commands
****************************************************************************/

void ResetTFCommand (ptfcmd)
   struct tfcmd   *ptfcmd;
{
   int i;
   short *pdig  = ptfcmd -> FirstDigitalCmd;
   short *psdig = ptfcmd -> FirstSpecialDigitalCmd;
   
   for (i = 0; i < ptfcmd -> NumCmds; i++, pdig++)
      *pdig = 0;
   
   for (i = 0; i < ptfcmd -> NumCmds; i++, psdig++)
      *psdig = 0;

   bzero (ptfcmd -> CmdString, 18);

   return;

} /* end func ResetTFCommand () */



/****************************************************************************
 *
 *  MODULE NAME            : TankFarmPostAlarm
 *  AUTHOR                 : emsteel
 *  DATE                   : 26 Mar 93
 *
 *  DESCRIPTION: post an alarm condition for a tank farm block
 *
****************************************************************************/

void TankFarmPostAlarm (precord, stat, sevr)
    struct dbCommon    *precord;      /* ptr to tank farm record */
    unsigned short     stat;          /* alarm status to post */
    unsigned short     sevr;          /* alarm severity to post */
{
    int	adis = 0;

      /* check for inhibited alarms */

   adis = AlarmDisableLink ((void *)precord);
   if ( (precord -> aenb == ALARM_DISABLED) || (adis != 0) )
   {
      precord -> nsta = INHIBIT_ALARM;
      precord -> nsev = NO_ALARM;
   }
   else
   {
      if (precord -> nsev < sevr)
      {
         precord -> nsta = stat;
         precord -> nsev = sevr;
      }
   }

   return;

} /* end func TankFarmPostAlarm () */



/****************************************************************************
 *
 *  MODULE NAME            : TankFarmRound
 *  AUTHOR                 : emsteel
 *  DATE                   : 27 Apr 93
 *
 *
 *  DESCRIPTION: round a value, 
 *           < 0.25 round down to nearest integral
 *           < 0.75 round to .5
 *           else round up to nearest integral
 *
****************************************************************************/

double TankFarmRound (value)
       double    value;
{
   double    diff;

   if (value > 0.0)
   {
      diff = value - floor (value);
      if (diff < 0.25)
         return (floor (value));
      else if (diff < 0.75)
         return (floor (value) + 0.5);
      else
         return (ceil (value));
   }
   else if (value < 0.0)
   {
      diff = value - (ceil (value));
      diff = -diff;
      if (diff < 0.25)
         return (ceil (value));
      else if (diff < 0.75)
     return (ceil (value) - 0.5);
      else
     return (floor (value));
   }
   else 
      return (value);

} /* end func TankFarmRound () */



/****************************************************************************
 *
 *  MODULE NAME            : TFTicksDiffInSeconds
 *  AUTHOR                 : emsteel
 *  DATE                   : 18 June 93
 *
 *  DESCRIPTION: given 2 system tick values, compute in seconds the
 *               difference
 *  NOTE:
 *       TFTicksDiffInSeconds is #define in tankfarm.h
 *
****************************************************************************/

double TFTicksDiffInSeconds (pT1, pT2)
   unsigned long     *pT1;       /* ptr to tick value */
   unsigned long     *pT2;       /* ptr to tick value */
{
   double   secs1     = 0.0;
   double   secs2     = 0.0;
   double   diff_secs = 0.0;

   secs1 = TFTicksToSeconds (pT1);
   secs2 = TFTicksToSeconds (pT2);

   diff_secs = secs1 - secs2;

   return (diff_secs);

} /* end func TFTicksDiffInSeconds () */



/****************************************************************************
 *
 *  MODULE NAME            : TFTicksToSeconds 
 *  AUTHOR                 : emsteel
 *  DATE                   : 21 June 93
 *
 *  DESCRIPTION: convert system time ticks to seconds as a double
 *                vxTicksPerSecond defined in .../share/epicsH/dbDefs.h
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : ems
 *  REVISION DATE          : 11 nov 93
 *  REVISION:  deleted EPICS_CLOCK_RATE, will use vxTicksPerSecond as 
 *             defined in .../share/epicsH/dbDefs.h
****************************************************************************/

double TFTicksToSeconds (pTicks)
   unsigned long *pTicks;
{
   double seconds = 0.0;

   seconds = ((double)(*pTicks)) / vxTicksPerSecond;

   return (seconds);

} /* end func TFTicksToSeconds () */
