/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <d4all.h>
#include <redCbDefs.h>
#include <redCbSlave.h>
#include <types.h>
#include <socket.h>
#include <in.h>
#include <tcp.h>
#include <netdb.h>
#include <time.h>
#include <stdio.h>
#include <stat.h>
#include <sys/signal.h>
#include <syslog.h>
#include <unistd.h>
#include <string.h>

extern	int	myPID;
extern	int	Debug, Trace, Verbose;
extern	int	masterFd;
extern	C4CODE	codeBase;
extern	char	masterName [];

/* File list */

extern	SlaveFileDesc	openFiles [MAX_OPEN_FILES];
extern	int		numOpenFiles;

/* Pointer to the receive message buffer */

extern	char		*msgBuffer;
extern	MsgHeader	*msgHeader;

/* Functions defined in this modules */

void	do_d4append ();
void	do_d4append_blank ();
void	do_d4append_start ();
void	do_d4blank ();
void	do_d4bottom ();
void	do_d4close ();
void	do_d4close_all ();
void	do_d4create ();
void	do_d4delete ();
void	do_d4field ();
void	do_d4flush_all ();
void	do_d4go ();
void	do_d4lock_file ();
void	do_d4open ();
void	do_d4pack ();
void	do_d4seek ();
void	do_d4skip ();
void	do_d4tag_select ();
void	do_d4top ();
void	do_d4unlock_all ();
void	do_f4assign ();
void	do_f4assign_double ();
void	do_f4assign_int ();
void	do_f4assign_long();
void	do_f4assign_n ();

/*----------*/
void do_d4append ()

    {

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4append\n", myPID);

	condExit (d4append (lookupDataHandle (getHandle ())) == 0,
			"Failure in d4append_blank");

    }


/*----------*/
void do_d4append_blank ()

    {

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4append_blank\n", myPID);

	condExit (d4append_blank (lookupDataHandle (getHandle ())) == 0,
			"Failure in d4append_blank");

    }

/*----------*/
void do_d4append_start ()

    {

	D4APPEND_START_MSG	*request;

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4append_start\n", myPID);

	request = (D4APPEND_START_MSG *)msgBuffer;

	condExit (d4append_start (lookupDataHandle ((uint32) ntohl (request->data)),
			ntohl (request->useMemoEntries)) == 0,
			"Failure in d4append_start");

    }
                                                                                                                                               
/*----------*/
void do_d4blank ()

    {

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4blank\n", myPID);

	d4blank (lookupDataHandle (getHandle ()));

    }

/*----------*/
void do_d4bottom ()

    {

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4bottom\n", myPID);

	condExit (d4bottom (lookupDataHandle (getHandle ())) >= 0,
			"Failure in d4bottom");

    }

/*----------*/
void do_d4close ()

    {

	D4DATA		*slaveHandle;
	int		masterHandle;

	masterHandle = getHandle ();

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4close(%x)\n",
			myPID, masterHandle);

	slaveHandle = lookupDataHandle (masterHandle);
	condExit (d4close (slaveHandle) == 0, "Failure in d4close");
	unlockFile (openFiles [getFileIndex (masterHandle)].fileName);
	removeFile (masterHandle);

    }

/*----------*/
void do_d4close_all ()

    {

	int	i;

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4close_all\n", myPID);

	condExit (d4close_all (&codeBase) == 0, "Failure in d4close_all");

	/* Unlock all the files */

	for (i=0 ; i<numOpenFiles ; i++)
	    unlockFile (openFiles [i].fileName);

    }

/*----------*/
void do_d4create ()

    {

	D4CREATE_MSG	*request;
	F4FIELD_MSG	*fieldMsg;
	F4FIELD_INFO	*fieldInfo, *fInfo;
	T4TAG_MSG	*tagMsg;
	T4TAG_INFO	*tagInfo, *tInfo;
	D4DATA		*data;
	uint32		handle;
	int		i, numFields, numTags;
	char		pathName [200], *p, name [FILENAME_LEN+1];

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4create\n", myPID);

	/* Get header which contains the file name, number of */
	/* fields, and number of tags. */

	request = (D4CREATE_MSG *)msgBuffer;
	strcpy (name, request->name);
	handle = ntohl (request->data);
	numFields = ntohl (request->numFields);
	numTags = ntohl (request->numTags);

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Creating %s, fields=%d, tags=%d\n", myPID,
			name, numFields, numTags);

	/* Allocate memory for the field information and read info. */

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d):   Fields are:\n", myPID);
 
	fieldInfo = fInfo = (F4FIELD_INFO *)
		malloc ((numFields + 1) * sizeof (F4FIELD_INFO));
	fieldMsg = (F4FIELD_MSG *)msgBuffer;
	for (i=0 ; i<numFields ; i++, fInfo++) {
	    do {
		drainSocket ();
	    } while (!getMessage ());
	    fInfo->name = (char *) malloc (strlen (fieldMsg->name) + 1);
	    strcpy (fInfo->name, fieldMsg->name);
	    fInfo->type = ntohl (fieldMsg->type);
	    fInfo->len = ntohl (fieldMsg->len);
	    fInfo->dec = ntohl (fieldMsg->dec);
	    if (Trace)
		syslog (LOG_OPTIONS, "redCbSlave(%d):    %-12s %c %2d %d\n", myPID,
			fInfo->name, fInfo->type, fInfo->len, fInfo->dec);
	}

	/* Create terminator record. It's not clear which field is the important one */

	fInfo->name = 0;
	fInfo->type = 0;
	fInfo->len = 0;
	fInfo->dec = 0;

	/* Allocate memory for the tag information and read info.  An extra entry */
	/* is on the end which is a terminator. */

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d):   Tags are:\n", myPID);

	if (numTags > 0) { 
	    tagInfo = tInfo = (T4TAG_INFO *) malloc ((numTags + 1) * sizeof (T4TAG_INFO));
	    tagMsg = (T4TAG_MSG *)msgBuffer;
	    for (i=0 ; i<numTags ; i++, tInfo++) {
		do {
		    drainSocket ();
		} while (!getMessage ());
		tInfo->name = malloc (strlen (tagMsg->name) + 1);
		strcpy (tInfo->name, tagMsg->name);
		tInfo->expression = (char *) malloc (strlen (tagMsg->expression) + 1);
		strcpy (tInfo->expression, tagMsg->expression);
		if (strlen (tagMsg->filter)) {
		    tInfo->filter = (char *) malloc (strlen (tagMsg->filter) + 1);
		    strcpy (tInfo->filter, tagMsg->filter);
		} else
		    tInfo->filter = 0;
		tInfo->unique = ntohl (tagMsg->unique);
		tInfo->descending = ntohl (tagMsg->descending);
		if (Trace)
		    if (tInfo->filter)
			syslog (LOG_OPTIONS, 
				"redCbSlave(%d):    %-12s %-20s %-20s %d %d\n", myPID,
				tInfo->name, tInfo->expression, tInfo->filter,
				tInfo->unique, tInfo->descending);
		    else
			syslog (LOG_OPTIONS, 
				"redCbSlave(%d):    %-12s %-20s %d %d\n", myPID,
				tInfo->name, tInfo->expression,
				tInfo->unique, tInfo->descending);
	    }

	    /* Create terminator record. It's not clear which field is the important one */

	    tInfo->name = 0;
	    tInfo->expression = 0;
	    tInfo->filter = 0;
	    tInfo->unique = 0;
	    tInfo->descending = 0;
	} else
	    tagInfo = NULL;

	/* Delete existing files if they exist */

	p = strrchr (name, '.');
	if (p)
	    *p = '\0';
	sprintf (pathName, "%s.dbf", name);
	unlink (pathName);
	sprintf (pathName, "%s.mdx", name);
	unlink (pathName);
	sprintf (pathName, "%s.dbt", name);
	unlink (pathName);

	/* Lock the file before creating it. */

	lockFile (name);

	/* If the file is locked then it already exists and should not be */
	/* created again. */

	data = d4create (&codeBase, name, fieldInfo, tagInfo);
	condExit (data != NULL, "Failure in d4create");

	/* Release the memory for fields and tags */

	fInfo = fieldInfo;
	for (i=0 ; i<numFields ; i++, fInfo++)
	    free (fInfo->name);
	free (fieldInfo);

	if (numTags > 0) {
	    tInfo = tagInfo;
	    for (i=0 ; i<numTags ; i++, tInfo++) {
		free (tInfo->name);
		free (tInfo->expression);
		if (tInfo->filter)
		    free (tInfo->filter);
	    }
	    free (tagInfo);
	}

	/* Insert file into filelist */

	insertFile (name, handle, data);
 	
    }

/*----------*/
void do_d4delete ()

    {

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4delete\n", myPID);

	d4delete (lookupDataHandle (getHandle ()));

    }

/*----------*/
void do_d4field ()

    {

	D4FIELD_MSG	*request;
	F4FIELD		*field;
	char		fieldName [12];
	D4DATA		*data;

	request = (D4FIELD_MSG *)msgBuffer;
	strcpy (fieldName, (char *)request + sizeof (D4FIELD_MSG));	/* Name follows message */

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4field(%x,\"%s\")\n",
			myPID, ntohl (request->data), fieldName);

	data = lookupDataHandle (ntohl (request->data));
	field = d4field (data, fieldName);
	condExit (field != NULL, "Unable to open field");
	insertField (ntohl (request->data), fieldName,
		ntohl (request->masterHandle), field);
 
    }

/*----------*/
void do_d4flush_all ()

    {

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4flush_all\n", myPID);

	condExit (d4flush_all (lookupDataHandle (getHandle ())) == 0,
			"Failure during d4flush_all");

    }

/*----------*/
void do_d4go ()

    {

	D4GO_MSG	*request;

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4go\n", myPID);

	request = (D4GO_MSG *)msgBuffer;
	condExit (d4go (lookupDataHandle (ntohl (request->data)),
			ntohl (request->record_number)) == 0,
			"Failure during d4go");

    }

/*----------*/
void do_d4lock_file ()

    {

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4lock_file\n", myPID);

	condExit (d4lock_file (lookupDataHandle (getHandle ())) >= 0,
			"Failure during d4lock_file");

    }

/*----------*/
void do_d4open ()

    {

	D4OPEN_MSG	*request;
	D4DATA		*data;
	char		fileName [FILENAME_LEN+1], buf [200], *p;
	int		doCopy;

	request = (D4OPEN_MSG *)msgBuffer;
	strcpy (fileName, (char *)request + sizeof (D4OPEN_MSG));

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4open(\"%s\")\n",
			myPID, fileName);

	sendFileRequest (doCopy = (lockFile (fileName) == 0));

	if (doCopy)
	    getFiles (fileName, ntohl (request->hasMDX), ntohl (request->hasDBT));

	data = d4open (&codeBase, fileName);
	sprintf (buf, "Unable to d4open %s", fileName);
	condExit (data != NULL, buf);
	insertFile (fileName, (uint32) ntohl (request->data), data);

    }

/*----------*/
void do_d4pack ()

    {

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4pack\n", myPID);

	condExit (d4pack (lookupDataHandle (getHandle ())) == 0,
			"failure in d4pack");

    }

/*----------*/
void do_d4seek ()

    {

	D4SEEK_MSG	*request;
	char		key [100];

	request = (D4SEEK_MSG *)msgBuffer;
	strcpy (key, (char *)request + sizeof (D4SEEK_MSG));

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4seek(%x,\"%s\")\n",
			myPID, ntohl (request->data), key);

	d4seek (lookupDataHandle (ntohl (request->data)), key);

    }

/*----------*/
void do_d4skip ()

    {

	D4SKIP_MSG	*request;

	request = (D4SKIP_MSG *)msgBuffer;

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4skip(%x)\n",
			myPID, ntohl (request->data));

	condExit (d4skip (lookupDataHandle (ntohl (request->data)),
			ntohl (request->num_records)) >= 0,
			"Failure in d4skip");

    }

/*----------*/
void do_d4tag_select ()

    {

	D4TAG_SELECT_MSG	*request;
	T4TAG			*tag;
	D4DATA			*data;

	request = (D4TAG_SELECT_MSG *)msgBuffer;

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4tag_select\n", myPID);


	data = lookupDataHandle (ntohl (request->data));
	if (ntohl (request->tag))
	    d4tag_select (data, d4tag (data, request->tagName)); 
	else
	    d4tag_select (data, NULL);

    }

/*----------*/
void do_d4top ()

    {

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4top\n", myPID);

	condExit (d4top (lookupDataHandle (getHandle ())) >= 0,
			"Failure in d4top");

    }

/*----------*/
void do_d4unlock_all ()

    {

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got d4unlock_all\n", myPID);

	condExit (d4unlock_all (lookupDataHandle (getHandle ())) == 0,
			"Failure in d4unlock_all");

    }

/*----------*/
void do_f4assign ()

    {

	F4ASSIGN_MSG	*request;
	char		str [100];

	request = (F4ASSIGN_MSG *)msgBuffer;
	strcpy (str, (char *)request + sizeof (F4ASSIGN_MSG));
	
	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got f4assign\n", myPID);

	f4assign (lookupFieldHandle ((uint32) ntohl (request->field)), str);

    }

/*----------*/
void do_f4assign_double ()

    {

	F4ASSIGN_DOUBLE_MSG	*request;
	double			value;

	request = (F4ASSIGN_DOUBLE_MSG *)msgBuffer;
	value = atof (request->value);
 
	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got f4assign_double\n", myPID);

	f4assign_double (lookupFieldHandle ((uint32) ntohl (request->field)), value);

    }

/*----------*/
void do_f4assign_int ()

    {

	F4ASSIGN_INT_MSG	*request;

	request = (F4ASSIGN_INT_MSG *)msgBuffer;

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got f4assign_int\n", myPID);

	f4assign_int (lookupFieldHandle ((uint32) ntohl (request->field)), ntohl (request->value));

    }

/*----------*/
void do_f4assign_long()

    {

	F4ASSIGN_LONG_MSG	*request;

	request = (F4ASSIGN_LONG_MSG *)msgBuffer;

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got f4assign_long\n", myPID);

	f4assign_long (lookupFieldHandle ((uint32) ntohl (request->field)), ntohl (request->value));

    }

/*----------*/
void do_f4assign_n ()

    {

	F4ASSIGN_N_MSG	*request;
	char		str [100];

	request = (F4ASSIGN_N_MSG *)msgBuffer;
	strcpy (str, (char *)request + sizeof (F4ASSIGN_N_MSG));

	if (Trace)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got f4assign_n\n", myPID);

	f4assign_n (lookupFieldHandle ((uint32) ntohl (request->field)),
			str, request->stringLength);

    }

