/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/****************************************************************/
/*								*/
/*	REDUNDANT CODEBASE DAEMON MESSAGES AND STRUCTURES	*/
/*								*/
/*	This file defines the messages that fly betwen an	*/
/*	application and the daemon. A few control structures	*/
/*	that are internal to the daemon are also defined.	*/
/*								*/
/****************************************************************/

typedef	int		int32;
typedef	unsigned int	uint32;

/* Port to use if running without inetd support */

#define	REDCBPORT	8003	/* Used when inetd didn't start us */

/* Various definitions */

#define	PARTNER_CONNECT_INTERVAL	15
#define	SEND_BLOCK_SIZE			10000	/* For sending files to the slave */
#define	MAGIC_NUMBER			0x12345678
#define	LOG_OPTIONS			LOG_ALERT|LOG_LOCAL1

#define	FILENAME_LEN	100
typedef	char		FileName [FILENAME_LEN+1];

#define	FIELD_LEN	11
typedef	char		Field [FIELD_LEN+1];
#define	MAX_NUM_FIELDS	50

#define	TAG_LEN		11
typedef	char		Tag [TAG_LEN+1];

#define	MAX_KEY_LENGTH	100

/* This is a list of all the open files.  It is used to send the slave the list of */
/* open files when it come alive. */

#define	MAX_OPEN_FILES	100
typedef struct {
    FileName	fileName;
    D4DATA	*masterHandle;
    uint32	slaveHandle;
    uint32	numFields;
    Field	fieldName [MAX_NUM_FIELDS];
    F4FIELD	*masterFields [MAX_NUM_FIELDS];
    uint32	slaveFields [MAX_NUM_FIELDS];
    uint32	currentRecord;
    Tag		currentTag;
    uint32	appendInProgress;
    uint32	redundActive;
    uint32	hasMDX;
    uint32	hasDBT;
    char	pad;
} MasterFileDesc;	

typedef struct {
    FileName	fileName;
    uint32	masterHandle;
    D4DATA	*slaveHandle;
    uint32	numFields;
    Field	fieldName [MAX_NUM_FIELDS];
    uint32	masterFields [MAX_NUM_FIELDS];
    F4FIELD	*slaveFields [MAX_NUM_FIELDS];
    uint32	currentRecord;
    Tag		currentTag;
    uint32	appendInProgress;
    uint32	redundActive;
    uint32	hasMDX;
    uint32	hasDBT;
    char	pad;
} SlaveFileDesc;	

/* Message Types */

/* We don't support redundancy for the entire Codebase package,
	just the calls that are used by the Tankfarm product. */

#define D4APPEND	1	/* Actually append the record after d4append_start() */
#define D4APPEND_BLANK	2	/* Append a blank record to the file */
#define D4APPEND_START	3	/* Allocate an empty record for appending */
#define D4BLANK		4	/* Blank out the current record */
#define D4BOTTOM	5	/* Go to the last record according to current tag */
#define D4CLOSE		6	/* Close a file */
#define D4CLOSE_ALL	7	/* Close all files */
#define D4CREATE	8	/* Create a file */
#define D4DELETE	9	/* Delete the current record */
#define D4FIELD		10	/* Get a field handle */
#define D4FLUSH_ALL	11	/* Flush data record buffer and all index buffers and memo entries */
#define D4GO		12	/* Jump to a given record number */
#define D4INIT		13	/* Initialize access to codebase */
#define	D4LOCK_FILE	14	/* Lock a file */
#define D4OPEN		15	/* Open a file, return handle */
#define D4PACK		16	/* Actually remove all records marked for deletion */
#define D4SEEK		17	/* Seek to record with given tag value */
#define D4SKIP		18	/* Jump forward or backward according to current tag */
#define D4TAG_SELECT	19	/* Select a tag */
#define D4TOP		20	/* Go to first record according to current tag */
#define D4UNLOCK_ALL	21	/* Unlock file and index buffers */
#define F4ASSIGN	22	/* Replace field value with null terminated string */
#define F4ASSIGN_DOUBLE	23	/* Replace field value with a double */
#define F4ASSIGN_INT	24	/* Replace field value with an integer */
#define F4ASSIGN_LONG	25	/* Replace field value with a long */
#define F4ASSIGN_N	26	/* Replace field value with an arbitrary byte stream */

#define	MASTER_LIST	100	/* List of file from master to slave at slave startup */
#define	SEND_REQUEST	101	/* Send files request */
#define	SEND_FILE	102	/* Send a file */
#define	FIELD_DESC	103	/* Field descriptor message */
#define	TAG_DESC	104	/* Tag descriptor message */

/************************************************************************/
/*									*/
/*			MESSAGE FORMATS					*/
/*									*/
/*	The message format follows the TIS tradition of sending a	*/
/*	short which is the message type. After the message type, one	*/
/*	of these structures will follow, depending on the message	*/
/*	type. In some cases, there will be a variable length trailer	*/
/*	after the structure. Those cases will be documented in each	*/
/*	section below.							*/
/*									*/
/************************************************************************/

typedef struct {	/* Each message begins with one of these */
    uint32	msgMagic;
    uint32	msgSize;
    uint32	msgType;
} MsgHeader;

typedef	struct {
    MsgHeader	header;
    uint32	data;
} HANDLE_ONLY_MSG;
   
typedef struct {
    MsgHeader	header;
    uint32	data;
    uint32	useMemoEntries;
} D4APPEND_START_MSG;

typedef struct {
    MsgHeader	header;
    char	name[256];
    uint32	data;
    uint32	numFields;
    uint32	numTags;
} D4CREATE_MSG;

typedef struct {
    MsgHeader	header;
    char	name[16];
    uint32	type;
    uint32	len;
    uint32	dec;
} F4FIELD_MSG;

typedef struct {
    MsgHeader	header;
    char	name[16];
    char	expression[80];
    char	filter[80];
    uint32	unique;
    uint32	descending;
} T4TAG_MSG;

typedef struct {
    MsgHeader	header;
    uint32	data;
    uint32	masterHandle;
} D4FIELD_MSG;

typedef struct {
    MsgHeader	header;
    uint32	data;
    uint32	record_number;
} D4GO_MSG;

typedef struct {
    MsgHeader	header;
    uint32	data;
    uint32	nameLength;
    uint32	hasMDX;
    uint32	hasDBT;
} D4OPEN_MSG;

typedef struct {
    MsgHeader	header;
    uint32	data;
} D4SEEK_MSG;

typedef struct {
    MsgHeader	header;
    uint32	data;
    uint32	num_records;
} D4SKIP_MSG;

typedef struct {
    MsgHeader	header;
    char	tagName [12];
    uint32	data;
    uint32	tag;
} D4TAG_SELECT_MSG;

typedef struct {
    MsgHeader	header;
    uint32	field;
} F4ASSIGN_MSG;

typedef struct {
    MsgHeader	header;
    uint32	field;
    char	value [32];
} F4ASSIGN_DOUBLE_MSG;

typedef struct {
    MsgHeader	header;
    uint32	field;
    int		value;
} F4ASSIGN_INT_MSG;

typedef struct {
    MsgHeader	header;
    uint32	field;
    int32	value;
} F4ASSIGN_LONG_MSG;

typedef struct {
    MsgHeader	header;
    uint32	field;
    uint32	stringLength;
} F4ASSIGN_N_MSG;

typedef struct {
    MsgHeader	header;
    uint32	numFiles;
    char	masterName [32];
} MASTER_LIST_MSG;

typedef struct {
    MsgHeader	header;
    uint32	sendFiles;
} SEND_REQUEST_MSG;

typedef struct {
    MsgHeader	header;
    int32	bytesLeft;
    int32	blockSize;
    char	block [SEND_BLOCK_SIZE];
} FILE_SEND_MSG;

typedef struct {
    MsgHeader	header;
    FileName	fileName;
    uint32	fileHandle;
    uint32	numFields;			  
    Field	fieldName [MAX_NUM_FIELDS];
    uint32	fieldHandles [MAX_NUM_FIELDS];
    uint32	currentRecord;
    Tag		currentTag;
    uint32	appendInProgress;
    uint32	hasMDX;
    uint32	hasDBT;
} FILE_DESC_MSG;	

/* This is used to find out the longest message length */

typedef union {
	HANDLE_ONLY_MSG		handleOnlyMsg;
	D4APPEND_START_MSG	d4appendStartMsg;
	D4CREATE_MSG		d4createMsg;
	F4FIELD_MSG		f4fieldMsg;
	T4TAG_MSG		t4tagMsg;
	D4GO_MSG		d4goMsg;
	D4OPEN_MSG		d4openMsg;
	D4SEEK_MSG		d4seekMsg;
	D4SKIP_MSG		d4skipMsg;
	D4TAG_SELECT_MSG	d4tagSelectMsg;
	F4ASSIGN_MSG		f4assignMsg;
	F4ASSIGN_DOUBLE_MSG	f4assignDoubleMsg;
	F4ASSIGN_INT_MSG	f4assignIntMsg;
	F4ASSIGN_LONG_MSG	f4assignLongMsg;
	F4ASSIGN_N_MSG		f4assignNMsg;
	MASTER_LIST_MSG		masterListReq;
	FILE_SEND_MSG		fileSendMsg;
	FILE_DESC_MSG		fileDescMsg;
} ALL_MSG_UNION;

#define	MAX_MSG_LENGTH		(sizeof (ALL_MSG_UNION) + SEND_BLOCK_SIZE)
