/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/****************************************************************/
/*								*/
/*  	REDUNDANT CODEBASE SERVER DAEMON (Slave Version)	*/
/*								*/
/*	This program is the redundant slave for redCbSlave.	*/
/*	It may only be started by the inetd daemon and 		*/
/*	performs slave functions only.  It accepts commands	*/
/*	from the master and performs them on the slave.		*/
/*								*/
/****************************************************************/

#include <d4all.h>
#include <redCbDefs.h>
#include <sys/types.h>
#include <socket.h>
#include <in.h>
#include <tcp.h>
#include <netdb.h>
#include <sys/time.h>
#include <stdio.h>
#include <stat.h>
#include <sys/signal.h>
#include <syslog.h>
#include <string.h>
#include <fcntl.h>

extern	int	errno;
extern	char	*sys_errlist [];

char	Str [5000]; 

int	masterFd, myPID;
char	masterName [30];
int	Verbose, Trace, Debug, noInetd, debugLevel;
CODE4	codeBase;

/* Receive buffer for messages from our client and the individual message buffer */

static	char		*rxBuffer;
static	int		rxBufferBytes;
#define	RECV_BUFFER_SIZE 56000
	char		*msgBuffer;
static	MsgHeader	*bufferHeader;
	MsgHeader	*msgHeader;

/* File list */

SlaveFileDesc	openFiles [MAX_OPEN_FILES];
int		numOpenFiles;

/* Field list. Since field descriptors are refered to so frequently, we will keep a */
/* list of master vs. slave field handles for rapid access. */

#define	MAX_FIELD_HANDLES	200
typedef struct {
    uint32	masterHandle;
    F4FIELD	*slaveHandle;
} FieldHandleListEntry;
static	FieldHandleListEntry	fieldHandles [MAX_FIELD_HANDLES];
static	int	numFieldHandles;

/* Misc */

static	int	running;
#define	FALSE	0
#define	TRUE	1

/* File descriptor used by the arbiter and mrbiter states */

static	int	arbiterHandle;
#define		SLAVE		1
#define		MASTER		2
#define		MASTERNP	3

/* Functions in this module */

static	void	waitForConnect (void);
static	void	getFileListFromMaster (void);
static	void	openAllFiles (void);
	void	drainSocket (void);
	int	getMessage (void);
static	void	processPendingRequests (void);
	void	getFiles (char *, int, int);
static	void	getSendFileMsg (int i, FILE_DESC_MSG *);
static	void	getFile (char *);
	int	lockFile (char *);
	int	unlockFile (char *);
static	void	flushAllFiles (void);
	int	getHandle (void);
static	void	sendBytes (char *, int);
	void	sendFileRequest (int);
	void	sendOpenAck (void);
	void	condExit (int, char *);
	void	insertFile (char *, uint32, D4DATA *);
static	void	sortFileList (void);
static	int	compare (const void *, const void *);
	void	removeFile (uint32);
	void	insertField (uint32, char *, uint32, F4FIELD *);
	void	setTag (uint32, T4TAG *);
	D4DATA	*lookupDataHandle (uint32);
	int	getFileIndex (uint32);
static	int	bcompare (const void *, const void *);
	void	insertFieldHandle (uint32, F4FIELD *);
static	int	fcompare (const void *, const void *);
	void	removeFieldHandle (uint32);
	F4FIELD	*lookupFieldHandle (uint32);
static	int	bfcompare (const void *, const void *);
static	void	killHandler (int);
static	void	pipeHandler (int);
static	void	usr1Handler (int);
static	void	usr2Handler (int);

/* External functions */

extern	void	do_d4append (void);
extern	void	do_d4append_blank (void);
extern	void	do_d4append_start (void);
extern	void	do_d4blank (void);
extern	void	do_d4bottom (void);
extern	void	do_d4close (void);
extern	void	do_d4close_all (void);
extern	void	do_d4create (void);
extern	void	do_d4delete (void);
extern	void	do_d4field (void);
extern	void	do_d4flush_all (void);
extern	void	do_d4go (void);
extern	void	do_d4lock_file (void);
extern	void	do_d4open (void);
extern	void	do_d4pack (void);
extern	void	do_d4seek (void);
extern	void	do_d4skip (void);
extern	void	do_d4tag (void);
extern	void	do_d4tag_select (void);
extern	void	do_d4top (void);
extern	void	do_d4unlock_all (void);
extern	void	do_f4assign (void);
extern	void	do_f4assign_double (void);
extern	void	do_f4assign_int (void);
extern	void	do_f4assign_long(void);
extern	void	do_f4assign_n (void);

/*----------*/
main (argc, argv)

    int		argc;
    char	*argv[];

    {

	int		status, Cmd;
	struct servent	*sp;
	short		msgType;
	int		optVal = 1;
	struct timeval	timeOut;
	char		*p;
	char		myName [50];
	static	char	*msStates [] = {"", "slave", "master", "master without partner", "unknown"};

	/* Get my PID for status reporting purposes */

	myPID = getpid ();

	/* Register with the arbiter */

	sprintf (myName, "redCbSlave%d", myPID);
	if (!appRegister (myName, &arbiterHandle)) {
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Unable to register with msArbiter!\n", myPID);
	    exit (-1);
	}

	if (Verbose)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Registered with msArbiter!\n", myPID);

	/* If we are currently a master, then this server is not needed */

	appRecvData (arbiterHandle, p, 1);
	appGetStatus (arbiterHandle, &status);
	if (Verbose)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Current state is %s\n", myPID, msStates [status]);
	if (status != SLAVE) {
	    appUnRegister (arbiterHandle);
	    if (Verbose)
		syslog (LOG_OPTIONS, "redCbSlave(%d): Not a slave -- terminating!\n",
				myPID);
	    exit (0);
	}

	/* Parse the command line arguments */

	Verbose = 0;
	Debug = 0;
	Trace = 0;
	noInetd = 0;
	while ((Cmd = getopt (argc,argv,"ivdt")) != -1) {
	    switch (Cmd) {
		case 'i':		/* Not started by inetd */
		    syslog (LOG_OPTIONS, "redCbSlave(%d): No inetd mode on\n", myPID);
		    noInetd = 1;
		    break;		    
		case 'v':		/* Verbose mode */
		    syslog (LOG_OPTIONS, "redCbSlave(%d): Verbose mode on\n", myPID);
		    Verbose = 1;
		    debugLevel = 1;
		    break;
		case 'd':		/* Debug mode */
		    syslog (LOG_OPTIONS, "redCbSlave(%d): Debug mode on\n", myPID);
		    Debug = 1;
		    Verbose = 1;
		    debugLevel = 3;
		    break;
		case 't':		/* Trace mode */
		    syslog (LOG_OPTIONS, "redCbSlave(%d): Trace mode on\n", myPID);
		    Trace = 1;
		    debugLevel = 2;
		    break;
	    }
	}
   
	/* Announce we are running */

	if (Verbose)
	    syslog (LOG_OPTIONS,  
		"redCbSlave(%d): Slave server started.\n", myPID);

	/* If the inetd server started us then our fd is 0.  If we were started manually */
	/* with the -i switch then wait for a connection. */

	if (noInetd)
	    waitForConnect ();
	else {
	    masterFd = dup(0);
	    close (0);
	    close (1);
	    close (2);
	}

	/* Set some options for the master's fd */

	setsockopt (masterFd, SOL_SOCKET, SO_REUSEADDR, (caddr_t) &optVal, sizeof (optVal));
	setsockopt (masterFd, IPPROTO_TCP, TCP_NODELAY, (caddr_t) &optVal, sizeof (optVal));
	setsockopt (masterFd, SOL_SOCKET, SO_KEEPALIVE, (caddr_t) &optVal, sizeof (optVal));

	/* Initialize various things */

	numOpenFiles = 0;
	numFieldHandles = 0;

	/* Init our access to Codebase */

	d4init(&codeBase);
	codeBase.field_name_error = 0;
	codeBase.open_error = 0;
	codeBase.tag_name_error = 0;
	codeBase.read_lock = 0;
	codeBase.create_error = 0;
	codeBase.go_error = 0;
	codeBase.skip_error = 0;

	/* Define a signal handler for KILL signals */

	signal (SIGTERM, killHandler);	/* Intercept kill messages */
	signal (SIGPIPE, pipeHandler);	/* Intercept broken pipe messages */
	signal (SIGUSR1, usr1Handler);	/* Increment debug level */
	signal (SIGUSR2, usr2Handler);	/* Decrement debug level */

	/* Allocate the master receive buffer and the individual message buffer */

	rxBuffer = (char *) malloc (RECV_BUFFER_SIZE);
	bufferHeader = (MsgHeader *) rxBuffer;	/* Beginning of buffer always has a header */

	msgBuffer = (char *) malloc (MAX_MSG_LENGTH);
	msgHeader = (MsgHeader *) msgBuffer;	/* Beginning of buffer always has a header */

	/* Get a fresh copy of all open files from the master. */

	getFileListFromMaster ();
	openAllFiles ();

	/* This is our main loop. We just wait here and let socket I/O tell us */
	/* when to do something. */

	if (Verbose)
	    syslog (LOG_OPTIONS,  "redCbSlave(%d): Ready\n", myPID);

	/* Wait for a message from our master */

	running = TRUE;
	while (running) {
	    drainSocket ();
	    processPendingRequests ();
	}

	if (Verbose)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Flushing and closing all files.\n", myPID);

	flushAllFiles ();

	d4close_all (&codeBase);
	if (masterFd != -1)
	    close (masterFd);
	appUnRegister (arbiterHandle);
	free ((void *)rxBuffer);
	free ((void *)msgBuffer);

	if (Verbose)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Terminating.\n", myPID);

	exit (0);

    }

/*----------*/
static void waitForConnect (void)		/* Wait for the client to connect */

    {

	int			acceptFd, addrSize, optVal = 1, result;
	struct sockaddr_in	portAddr;
	fd_set			fdSet;

	acceptFd = socket(PF_INET, SOCK_STREAM, 0);
	setsockopt(acceptFd, SOL_SOCKET, SO_REUSEADDR, (caddr_t) &optVal, sizeof(optVal));
	setsockopt(acceptFd, IPPROTO_TCP, TCP_NODELAY, (char *)&optVal, sizeof(optVal));
	setsockopt (acceptFd, SOL_SOCKET, SO_KEEPALIVE, (caddr_t) &optVal, sizeof (optVal));
	portAddr.sin_family = AF_INET;
	portAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	portAddr.sin_port = htons(REDCBPORT);
	if (bind(acceptFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1) {
 	    perror("redCbSlave: Can't bind to TCP port");
	    exit(-1);
	}

	listen(acceptFd, 2);

	FD_ZERO (&fdSet);
	FD_SET (acceptFd, &fdSet);
	do {
	    result = select(acceptFd + 1, &fdSet, NULL, NULL, NULL);
	} while (result == -1);

	addrSize = sizeof (portAddr);
	masterFd = accept (acceptFd, (struct sockaddr *) &portAddr, &addrSize);
	close (acceptFd);

    }

/*----------*/
static void getFileListFromMaster (void)

    {

	int			i, copyFiles, numFiles, hasMDX, hasDBT;
	SEND_REQUEST_MSG 	*request;
	MASTER_LIST_MSG		*masterList;
	FILE_DESC_MSG		*fileReqMsg;

	if (Verbose)
	    syslog (LOG_OPTIONS, 
			"redCbSlave(%d): Waiting for file list from master\n", myPID);

	/* Get the MASTERLIST message */

	do {
	    drainSocket ();
	} while (!getMessage ());

	if (msgHeader->msgType != MASTER_LIST) {
	    if (Verbose)
		syslog (LOG_OPTIONS, "redCbSlave(%d): Expected MASTER_LIST, got %d -- terminated.\n",
				myPID, msgHeader->msgType);
	    exit (-1);
	}

	/* Get each file in the master list and request it from the client */

	masterList = (MASTER_LIST_MSG *)msgBuffer;
	numFiles = ntohl (masterList->numFiles);
	strcpy (masterName, masterList->masterName);

	if (Verbose)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Got MASTER_LIST, numFiles = %d\n",
			myPID, numFiles);

	fileReqMsg = (FILE_DESC_MSG *)msgBuffer;
	for (i=0 ; i<numFiles ; i++) {

	    /* Wait for some incoming traffic */

	    do {
		drainSocket ();
	    } while (!getMessage ());

	    /* Got a file request message */

	    getSendFileMsg (i, fileReqMsg);
	    if (Verbose)
		syslog (LOG_OPTIONS, "redCbSlave(%d):   File: %s\n", myPID, openFiles [i].fileName);
		    
	    /* If the file is not currently in use by another slave server, */
	    /* copy it from the master before opening it. */

	    /* Send the send request to the master */

	    copyFiles = lockFile (openFiles [i].fileName) == 0;

	    sendFileRequest (copyFiles);

	    if (copyFiles) {

		/* Retrieve the files */

		getFiles (openFiles [i].fileName, openFiles [i].hasMDX, openFiles [i].hasDBT);
	
	    }
	
	}

	numOpenFiles = numFiles;

    }

/*----------*/
static void getSendFileMsg (int i, FILE_DESC_MSG *msg)

    {

	int	j;

	openFiles [i].numFields = ntohl (msg->numFields);

	for (j=0 ; j<openFiles [i].numFields ; j++) {
	    openFiles [i].masterFields [j] = ntohl (msg->fieldHandles [j]);
	    strcpy (openFiles [i].fieldName [j], msg->fieldName [j]);
	}
	strcpy (openFiles [i].fileName, msg->fileName);
	openFiles [i].masterHandle = ntohl (msg->fileHandle);
	openFiles [i].currentRecord = ntohl (msg->currentRecord);
	openFiles [i].appendInProgress = ntohl (msg->appendInProgress);
	openFiles [i].hasMDX = ntohl (msg->hasMDX);
	openFiles [i].hasDBT = ntohl (msg->hasDBT);

    }

/*----------*/
static void openAllFiles (void)	/* Open the files just received from master */

    {

	int		i, j, result, numCopied;
	D4DATA		*data;
	F4FIELD		*field;
	char		buf [200];

	/* Sort the files so that we can use a binary sort to search the table */

	sortFileList ();

	/* Traverse the file list and open all the files.  Copy it from the master */
	/* first if necessary */

	numCopied = 0;
	for (i=0 ; i<numOpenFiles ; i++) {

	    if (Verbose)
		syslog (LOG_OPTIONS, 
			"redCbSlave(%d): Opening %s\n", myPID, openFiles [i].fileName);

	    /* Open the data file */

	    data = d4open (&codeBase, openFiles [i].fileName);
	    condExit (data != NULL, "Unable to open file during startup.");
	    openFiles [i].slaveHandle = data;
	    if (Verbose)
		syslog (LOG_OPTIONS,
			"redCbSlave(%d):   %s opened.\n", myPID, openFiles [i].fileName);

	    /* Open all of the fields */

	    for (j=0 ; j<openFiles [i].numFields ; j++) {
		field = d4field (data, openFiles [i].fieldName [j]);
		if (Verbose)
		    syslog (LOG_OPTIONS,
			"redCbSlave(%d):   Field %s opened.\n",
			myPID, openFiles [i].fieldName [j]);
		condExit (field != NULL, "Unable to open field during startup");
		openFiles [i].slaveFields [j] = field;
		insertFieldHandle (openFiles [i].masterFields [j], openFiles [i].slaveFields [j]);
	    }

	    /* Select the current tag */

	    if (strlen (openFiles [i].currentTag)) {
		d4tag_select (data, d4tag (data, openFiles [i].currentTag));
		if (Verbose)
			syslog (LOG_OPTIONS,
			"redCbSlave(%d):   Tag %s selected.\n",
			myPID, openFiles [i].currentTag);
	    }

	    /* Position at the proper record number */

	    if ((d4reccount (data) >= openFiles [i].currentRecord) &&
			(openFiles [i].currentRecord >= 1)) {
		condExit (d4unlock_all (data) == 0, "Unable to unlock file during startup");
		if (openFiles [i].currentRecord > 0L) {
		    if (Verbose) 
			syslog (LOG_OPTIONS,
				"redCbSlave(%d):   Positioning at record %ld\n",
				myPID, openFiles [i].currentRecord);
		    result = d4go (data, openFiles [i].currentRecord);
		    condExit (result == 0, "Unable to set position during slave startup");
		}
		if (Verbose)
		    syslog (LOG_OPTIONS,
			"redCbSlave(%d):   Positioned at record %ld\n",
			myPID, openFiles [i].currentRecord);
	    }

	    /* That last call probably locked the record */
	    
	    condExit (d4unlock_all (data) == 0, "Unable to unlock file during startup");
	    if (Verbose)
		syslog (LOG_OPTIONS,
			"redCbSlave(%d):   %s unlocked.\n",
			myPID, openFiles [i].fileName);

	}

	if (Verbose)
	    syslog (LOG_OPTIONS, 
		"redCbSlave(%d): All files opened.\n", myPID);
	    
    }

/*----------*/
void getFiles (fileName, getMDX, getDBT)	/* Get DBF, MDX, DBT files from master */

    char	*fileName;
    int		getMDX, getDBT;

    {

	char	rawFileName [200], *p, fName [200];

	/* Get the data file */

	getFile (fileName);

	/* Strip the extension from the filename */

	strcpy (rawFileName, fileName);
	p = strrchr (rawFileName, '.');
	if (p)
	    *p = '\0';
	
	/* If an index file exists, get it */

	if (getMDX) {
	    sprintf (fName, "%s.mdx", rawFileName);
	    getFile (fName);
	}

	/* If a memo file exists, get it */

	if (getDBT) {
	    sprintf (fName, "%s.dbt", rawFileName);
	    getFile (fName);
	}

    }

/*----------*/
static void getFile (fileName)		/* Get an individual file */

    char	*fileName;

    {

	int		fd, num;
	FILE_SEND_MSG	*fileMsg;

	fd = creat (fileName, 0644);
	if (Debug)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Creating %s\n", myPID, fileName);

	/* Get the file */

	fileMsg = (FILE_SEND_MSG *)msgBuffer;
	num = 0;
	do {

	    /* Wait for the next piece of the file */

	    do {
		drainSocket ();
	    } while (!getMessage ());

	    if (Debug)
		syslog (LOG_OPTIONS, "redCbSlave(%d): Writing block %d, bytes left = %d\n", myPID, num++, ntohl (fileMsg->bytesLeft));

	    /* Write the piece to the file */

	    write (fd, fileMsg->block, ntohl (fileMsg->blockSize));
 
	} while (ntohl (fileMsg->bytesLeft) > 0);

	close (fd);
  
	if (Verbose)
	    syslog (LOG_OPTIONS,
		"redCbLib(%d): File %s was received from master\n", myPID, fileName);
	
    }

/*----------*/
int lockFile (fileName)		/* Increment use count for a file */

    char	*fileName;

    {

	/* If a lock file exists, increment the use count in the file. If the lock */
	/* file does not exist, create one with a count of one. Return the value */
	/* of the use count prior to this increment. */

	int	fd, count;
	char	lockName [200], buf [10], *p, name [FILENAME_LEN+1];

	strcpy (name, fileName);
	p = strchr (name, '.');
	if (p)
	    *p = '\0';

	sprintf (lockName, "%s.LOCK", name);
	fd = open (lockName, O_EXCL | O_RDWR, 0);
	if (fd != -1) {		/* File exists, increment the use count */
	    read (fd, buf, 10);
	    count = atoi (buf);
	    sprintf (buf, "%d", count + 1);
	    lseek (fd, (off_t)0, SEEK_SET);
	    write (fd, buf, 10);
	    close (fd);
	    return count;
	}

	/* File does not exist, create it. */

	fd = open (lockName, O_CREAT | O_EXCL | O_WRONLY | O_TRUNC, 0644);
	sprintf (buf, "%d", 1);
	write (fd, buf, 10);
	close (fd);
	return 0;
	    
    }

/*----------*/
int unlockFile (fileName)	/* Decrement the use count for a file */

    char	*fileName;

    {

	/* Decrement the use count for a file.  If the count goes to zero */
	/* delete the lock file. */

	int	fd, count;
	char	lockName [200], buf [10], *p, name [FILENAME_LEN+1];

	strcpy (name, fileName);
	p = strchr (name, '.');
	if (p)
	    *p = '\0';

	sprintf (lockName, "%s.LOCK", name);
	fd = open (lockName, O_EXCL | O_RDWR, 0);
	if (fd != -1) {		/* File exists, increment the use count */
	    read (fd, buf, 10);
	    count = atoi (buf);
	    if (count > 1) {
		sprintf (buf, "%d", count - 1);
		lseek (fd, (off_t)0, SEEK_SET);
		write (fd, buf, 10);
		close (fd);
		return count;
	    } else {	/* Count has gone to zero, delete the lock file */
		close (fd);
		unlink (lockName);
		return 1;
	    }
	} else {	/* BAD! File is in use but no lock file exists */
	    if (Verbose)
		syslog (LOG_OPTIONS, "redCbSlave(%d): Unable to find %s for unlocking.\n",
			myPID, lockName);
	    return -1;
	}
    
    }

/*----------*/
static void flushAllFiles ()	/* Flush all databases to disk */

    {

	int	i;

	for (i=0 ; i<numOpenFiles ; i++) {
	    d4flush_all(openFiles [i].slaveHandle);
	    d4unlock_all(openFiles [i].slaveHandle);
	}

    }

/*----------*/
void drainSocket (void)		/* Drain all data from socket */

    {

	fd_set	sockFds;
	int	status, maxFd, bytesRead, bytesLeft, i;
	struct	timeval	timeOut;

	timeOut.tv_sec = 1;
	timeOut.tv_usec = 0;

	FD_ZERO(&sockFds);
	FD_SET(masterFd, &sockFds);
	FD_SET(arbiterHandle, &sockFds);
	maxFd = (masterFd > arbiterHandle) ? masterFd : arbiterHandle;

	/* Wait for activity on either the client socket or the arbiter socket */

	status = select (maxFd + 1, &sockFds, NULL, NULL, &timeOut);
	if (status == -1) {
	    if (Verbose)
		syslog (LOG_OPTIONS, "redCbSlave(%d): Error on select -- slave terminated.\n", myPID);
	    running = 0;
	    return;
	}

	/* Check to see if the arbiter is trying to talk to us */

	if (FD_ISSET (arbiterHandle, &sockFds)) {
	    if (!appRecvData (arbiterHandle, NULL, 1)) {
		if (Verbose)
		    syslog (LOG_OPTIONS, "redCbSlave(%d): Lost contact with arbiter -- terminating!\n", myPID);
		running = 0;
	    }
	    appGetStatus (arbiterHandle, &status);
	    if (status != SLAVE) {
		appUnRegister (arbiterHandle);
		if (Verbose)
		    syslog (LOG_OPTIONS, "redCbSlave(%d): No longer a slave -- terminating!\n",myPID);
		running = FALSE;
	    }
	}

	/* If any data from client, drain it into the receive buffer */

	if (FD_ISSET (masterFd, &sockFds)) {
	    bytesLeft = RECV_BUFFER_SIZE - rxBufferBytes;
	    if (bytesLeft == 0)
		return;
	    bytesRead = recv (masterFd, &rxBuffer [rxBufferBytes], bytesLeft, 0);
/*printf ("Rec: (%d @ %d)", bytesRead, rxBufferBytes);
for (i=0 ; i<bytesRead ; i++)
    printf ("%x ", (int)(rxBuffer [rxBufferBytes + i] & 0xff));
printf ("\n");
*/
	    if (bytesRead > 0) {
		rxBufferBytes += bytesRead;
	    } else {
		if (Verbose)
		    syslog (LOG_OPTIONS, "redCbSlave(%d): Client disconnected.\n", myPID);
		running = 0;
	    }
	}

    }

/*----------*/
static void processPendingRequests (void)	/* Process any pending requests from our client */

    {

	while (getMessage ()) {

	    if (Debug)
		syslog (LOG_OPTIONS, "redCbSlave(%d): Message received, type = %d\n",
				myPID, msgHeader->msgType);

	    switch (msgHeader->msgType) {
		case D4APPEND:
		    do_d4append ();
		    break;
		case D4APPEND_BLANK:
		    do_d4append_blank ();
		    break;
		case D4APPEND_START:
		    do_d4append_start ();
		    break;
		case D4BLANK:
		    do_d4blank ();
		    break;
		case D4BOTTOM:
		    do_d4bottom ();
		    break;
		case D4CLOSE:
		    do_d4close ();
		    break;
		case D4CLOSE_ALL:
		    do_d4close_all ();
		    break;
		case D4CREATE:
		    do_d4create ();
		    break;
		case D4DELETE:
		    do_d4delete ();
		    break;
		case D4FIELD:
		    do_d4field ();
		    break;
		case D4FLUSH_ALL:
		    do_d4flush_all ();
		    break;
		case D4GO:
		    do_d4go ();
		    break;
		case D4LOCK_FILE:
		    do_d4lock_file ();
		    break;
		case D4OPEN:
		    do_d4open ();
		    break;
		case D4PACK:
		    do_d4pack ();
		    break;
		case D4SEEK:
		    do_d4seek ();
		    break;
		case D4SKIP:
		    do_d4skip ();
		    break;
		case D4TAG_SELECT:
		    do_d4tag_select ();
		    break;
		case D4TOP:
		    do_d4top ();
		    break;
		case D4UNLOCK_ALL:
		    do_d4unlock_all ();
		    break;
		case F4ASSIGN:
		    do_f4assign ();
		    break;
		case F4ASSIGN_DOUBLE:
		    do_f4assign_double ();
		    break;
		case F4ASSIGN_INT:
		    do_f4assign_int ();
		    break;
		case F4ASSIGN_LONG:
		    do_f4assign_long();
		    break;
		case F4ASSIGN_N:
		    do_f4assign_n ();
		    break;
		default:
		    syslog (LOG_OPTIONS,  
				"redCbSlave(%d): Unknown message code %d received.\n",
				myPID, msgHeader->msgType);
			exit (-1);
	    }
	}

    }

/*----------*/
int getMessage (void)	/* Get next message from the message buffer */

    {

	int	size;

	/* Check if enough bytes in the buffer to make an entire message */

	if (rxBufferBytes < sizeof (MsgHeader))
	    return 0;

	size = ntohl (bufferHeader->msgSize);
	if (size > rxBufferBytes)
	    return 0;

	/* Copy the message into the message buffer squash out the message */

	memcpy (msgBuffer, rxBuffer, size);
	rxBufferBytes -= size;
	memcpy (rxBuffer, &rxBuffer [size], rxBufferBytes);

	/* Fix the byte ordering of the header */

	msgHeader->msgMagic = ntohl (msgHeader->msgMagic);
	msgHeader->msgSize = size;
	msgHeader->msgType = ntohl (msgHeader->msgType);

	/* Check the magic number */

	if (msgHeader->msgMagic != MAGIC_NUMBER) {
	    if (Verbose)
		syslog (LOG_OPTIONS, "redCbSlave(%d): Bad magic number in message from client -- terminating!\n",
				myPID);
	    exit (-1);
	}

	/* Return success */

	if (Debug)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Received message %d\n", myPID, msgHeader->msgType);

	return 1;

    }

/*----------*/
int getHandle ()	/* Get the file handle from the most recent message */

    {

	HANDLE_ONLY_MSG	*msg;

	msg = (HANDLE_ONLY_MSG *)msgBuffer;
	return ntohl (msg->data);

    }

/*----------*/
void sendFileRequest (sendFiles)

    int		sendFiles;

    {

	SEND_REQUEST_MSG	msg;

	msg.header.msgMagic = htonl (MAGIC_NUMBER);
	msg.header.msgSize = htonl (sizeof (msg));
	msg.header.msgType = htonl (SEND_REQUEST);
	msg.sendFiles = htonl (sendFiles);
	sendBytes ((char *)&msg, sizeof (msg));

    }

/*----------*/
static void sendBytes (buff, nbytes)

    char	*buff;
    int		nbytes;

    {

	int	bytesSent;

	if (Debug) {
	    syslog (LOG_OPTIONS, 
		"redCbSlave(%d): Sending %d bytes on fd=%d\n", myPID, nbytes, masterFd);
	}

	bytesSent = send (masterFd, buff, nbytes, 0);
	if (Debug)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): send returned %d\n", myPID, bytesSent);

	condExit (bytesSent == nbytes, "Unable to send message to master\n");

    }

/*----------*/
void condExit (condition, message)	/* Cleanup and exit if false */

    int		condition;
    char	*message;

    {

	if (!condition) {
	    if (Verbose) {
		syslog (LOG_OPTIONS,  "redCbSlave(%d): %s\n", myPID, message);
		syslog (LOG_OPTIONS,  "redCbSlave(%d): Terminated due to error\n", myPID);
	    }
	    exit (-1);
	}

    }

/*----------*/
void insertFile (name, masterHandle, slaveHandle) /* Insert file into open file list */

    char	*name;
    D4DATA	*slaveHandle;
    uint32	masterHandle;

    {

	condExit (numOpenFiles <= MAX_OPEN_FILES, "Too many files open");

	strcpy (openFiles [numOpenFiles].fileName, name);
	openFiles [numOpenFiles].masterHandle = masterHandle;
	openFiles [numOpenFiles].slaveHandle = slaveHandle;
	openFiles [numOpenFiles].numFields = 0;
	strcpy (openFiles [numOpenFiles].currentTag, "");
	numOpenFiles++;

	sortFileList ();

    }

/*----------*/
static void sortFileList (void)

    {

	/* Sort the file by ascending masterHandle so that it can be binary */
	/* searched later. */

	if (numOpenFiles > 1)
	    qsort ((char *)openFiles, numOpenFiles, sizeof (SlaveFileDesc), compare);

    }

/*----------*/
static int compare (e1, e2)

    const void	*e1, *e2;

    {

	unsigned	a = (unsigned)(((SlaveFileDesc *)e1)->masterHandle);
	unsigned	b = (unsigned)(((SlaveFileDesc *)e2)->masterHandle);

	return a - b;

    }

/*----------*/
void removeFile (data)	/* Remove a file from the open file list */

    uint32	data;

    {

	int	i, j;

	for (i=0 ; i<numOpenFiles ; i++)
	    if (openFiles [i].masterHandle == data)
		break; 

	if (i < numOpenFiles) {

	    /* remove entries from the field list */

	    for (j=0 ; j<openFiles [i].numFields ; j++)
		removeFieldHandle (openFiles [i].masterFields [j]);

	    /* Compress remaining list */

	    while (i < (numOpenFiles - 1)) {
		openFiles [i] = openFiles [i+1];
		i++;
	    }
	    numOpenFiles--;
	}

    }

/*----------*/
void insertField (data, name, masterHandle, slaveHandle) /* Insert a field into a file's field list */

    uint32	data;
    char	*name;
    F4FIELD	*slaveHandle;
    uint32	masterHandle;

   {

	int	i;

	for (i=0 ; i<numOpenFiles ; i++)
	    if (openFiles [i].masterHandle == data)
		break;

	if (i != numOpenFiles) {
	    condExit (openFiles [i].numFields <= MAX_FIELD_HANDLES, "Too many fields open");
	    strcpy (openFiles [i].fieldName [openFiles [i].numFields], name);
	    openFiles [i].masterFields [openFiles [i].numFields] = masterHandle;
	    openFiles [i].slaveFields [openFiles [i].numFields] = slaveHandle;
	    openFiles [i].numFields++;
	    insertFieldHandle (masterHandle, slaveHandle);
	} else {
	    syslog (LOG_OPTIONS,  "File is not open for opening of field: %s\n", 
			masterHandle);
	    exit (11);
	}
	    	    
   }

/*----------*/
void setTag (data, tag)

    uint32	data;
    T4TAG	*tag;

    {

	int	i;

	for (i=0 ; i<numOpenFiles ; i++)
	    if (openFiles [i].masterHandle == data)
		break;

	if (i != numOpenFiles)
	    strcpy (openFiles [i].currentTag, tag->alias);
	else {
	    syslog (LOG_OPTIONS,  "redCbSlave(%d): File is not open for setting of tag: %s\n",
			myPID, tag);
	    exit (12);
	}
	
    }

/*----------*/
D4DATA *lookupDataHandle (masterHandle)		/* Get slave handle given master handle */

    uint32		masterHandle;

    {

	SlaveFileDesc	*fileDesc;
	char		str [100];
	int		i;

	
	fileDesc = (SlaveFileDesc *) bsearch ((char *)&masterHandle, (char *)openFiles,
			numOpenFiles, sizeof (SlaveFileDesc), bcompare);
	sprintf (str, "Master(%x)->slave data handle translation failure\n", masterHandle);
	condExit (fileDesc != NULL, str);
	return fileDesc->slaveHandle;

    }

/*----------*/
int getFileIndex (masterHandle)		/* Get slave handle given master handle */

    uint32		masterHandle;

    {

	SlaveFileDesc	*fileDesc;
	char		str [100];
	int		i;

	
	fileDesc = (SlaveFileDesc *) bsearch ((char *)&masterHandle, (char *)openFiles,
			numOpenFiles, sizeof (SlaveFileDesc), bcompare);
	return fileDesc - openFiles;

    }

/*----------*/
static int bcompare (key, entry)

    const void	*key, *entry;

    {

	uint32 keyValue = *((uint32 *)key);
	uint32 handleValue = ((SlaveFileDesc *)entry)->masterHandle;

	return keyValue - handleValue;

    }

/*----------*/
void insertFieldHandle (masterHandle, slaveHandle)

    F4FIELD	*slaveHandle;
    uint32	masterHandle;

    {

	fieldHandles [numFieldHandles].masterHandle = masterHandle;
	fieldHandles [numFieldHandles].slaveHandle = slaveHandle;
	numFieldHandles++;

	/* Sort the file by ascending masterHandle so that it can be binary */
	/* searched later. */

	if (numFieldHandles > 1)
	    qsort ((char *)fieldHandles, numFieldHandles,
			sizeof (FieldHandleListEntry), fcompare);

    }

/*----------*/
static int fcompare (e1, e2)

    const	void	*e1, *e2;

    {

	return ((unsigned)*((uint32 *)e1)) - ((unsigned)*((uint32 *)e2));

    }

/*----------*/
void removeFieldHandle (masterHandle)

    uint32	masterHandle;

    {

	int	i;

	for (i=0 ; i<numFieldHandles ; i++)
	    if (fieldHandles [i].masterHandle == masterHandle)
		break; 

	if (i < numFieldHandles) {
	    while (i < (numFieldHandles - 1)) {
		fieldHandles [i] = fieldHandles [i+1];
		i++;
	    }
	    numFieldHandles--;
	}
    }

/*-----------*/
F4FIELD	*lookupFieldHandle (masterHandle)

    uint32	masterHandle;

    {

	FieldHandleListEntry	*fieldEntry;

	fieldEntry = (FieldHandleListEntry *) bsearch ((char *)&masterHandle,
		(char *)fieldHandles, numFieldHandles,
		sizeof (FieldHandleListEntry), bfcompare);
	condExit (fieldEntry != NULL, "Master->slave field handle translation failure");
	return fieldEntry->slaveHandle;

    }

/*----------*/
static int bfcompare (key, entry)

    const void	*key, *entry;

    {

	uint32	keyValue, handleValue;

	keyValue = *(uint32 *)key;
	handleValue = ((FieldHandleListEntry *)entry)->masterHandle;
	return keyValue - handleValue;

    }

/*----------*/
static void killHandler (sig)		/* Terminate server */

    int sig;

    {

	if (Verbose)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Received -TERM\n", myPID);

	running = FALSE;

    }

/*----------*/
static void pipeHandler (sig)		/* Terminate server */

    int sig;

    {

	if (Verbose)
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Received -PIPE\n", myPID);

	running = FALSE;

    }

/*----------*/
static void usr1Handler (sig)		/* Increment debug level */

    int sig;

    {

	if (debugLevel < 3) {
	    debugLevel++;
	    Debug = Trace = Verbose = FALSE;
	    switch (debugLevel) {
		case 3:
		    Debug = TRUE;
		case 2:
		    Trace = TRUE;
		case 1:
		    Verbose = TRUE;
	    }
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Debug level now %d\n", myPID, debugLevel);
	}

    }

/*----------*/
static void usr2Handler (sig)		/* Decrement debug level */

    int sig;

    {

	if (debugLevel > 0) {
	    debugLevel--;
	    Debug = Trace = Verbose = FALSE;
	    switch (debugLevel) {
		case 3:
		    Debug = TRUE;
		case 2:
		    Trace = TRUE;
		case 1:
		    Verbose = TRUE;
	    }
	    syslog (LOG_OPTIONS, "redCbSlave(%d): Debug level now %d\n", myPID, debugLevel);
	}

    }

