/*emergency.c,v 1.1.1.1 1999/08/03 14:44:27 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/************************************************************************/
/*                                                                      */
/*            EMERGENCY SESSION TASK                                    */
/*                                                                      */
/*    This task sits on a port and waits for a connection.              */
/*                                                                      */
/*                                                                      */
/************************************************************************/

/************************* MOD LOG ***************************************
01/21/92    AGI    410-581-0422    irani@tate.com
        Original entry.
*************************************************************************/

/************************************************************************/
/*                                                                      */
/*            INCLUDE FILES                                             */
/*                                                                      */
/************************************************************************/

#include <vxWorks.h>
#include <socket.h>
#include <in.h>
#include <string.h>
#include <sysSymTbl.h>
#include <errno.h>
#include <bootLib.h>
#include <systime.h>
#include <ioLib.h>
#include <sockLib.h>
#include <stdio.h>

#include "restable.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "task_params.h"        /* For getting task priorities */
#include "dbRecords.h"
#include "redmsgs.h"

/************************************************************************/
/*                                                                      */
/*            GLOBAL DATA WE DEFINE                                     */
/*                                                                      */
/************************************************************************/


/************************************************************************/
/*                                                                      */
/*            GLOBAL DATA WE REFERENCE                                  */
/*                                                                      */
/************************************************************************/
unsigned long ourInetAddr;	/* local box IP address */


/************************************************************************/
/*                                                                      */
/*            OUR PRIVATE DATA                                          */
/*                                                                      */
/************************************************************************/

static int              listenfd;        /* fd for listen socket        */

void emergencyTask ( int port );


/************************************************************************/
/*                                                                      */
/*                INITIALIZE                                            */
/*                                                                      */
/************************************************************************/

int emergencyTaskStart (int port)
{
    /*
    Spawn the emergency task. Use the same
    priority as the shell.
    */

    if (taskSpawn("emergencyTask", 1, VX_FP_TASK, 10000,
                   (FUNCPTR)emergencyTask, port,0,0,0,0,0,0,0,0,0) == ERROR) 
    {
        printErr("Can't spawn emergencyTask.\n");
        return (-1);
    }

    /*
    Initialization complete
    */
    return(0);
}

/****************************************************************/
/*                                                              */
/*            THE EMERGENCY SHELL TASK                          */
/*                                                              */
/*    This is the emergency shell task.                         */
/*                                                              */
/****************************************************************/

void emergencyTask ( int port )
{
    struct sockaddr_in serverAddr;    /* server's address   */
    FAST int        status;
    int		    tempfd;
    int		    addrlen = sizeof(serverAddr);

    if (listenfd != 0 && listenfd != ERROR)
        if ((status = close(listenfd)) == ERROR)
            printErr("emergencyTask: Unable to close open listen socket\n");

    /* Open a stream socket. */
    if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) == ERROR) {
        printErr("emergencyTask: Socket creation error\n");
        taskSuspend(0);
    }
    
    /* Zero the sock_addr structure */
    bfill((char *)&serverAddr, sizeof(serverAddr), 0);
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_port = htons(port);

    /* get server port address */
    if (bind(listenfd, (struct sockaddr *)&serverAddr, sizeof(serverAddr)) == ERROR) {
        printErr("emergencyTask: Bind error\n");
        close(listenfd);
	listenfd = 0;
        taskSuspend(0);
    }
    if (listen(listenfd, 0) == ERROR) {
        printErr("emergencyTask: Listen error\n");
        close(listenfd);
	listenfd = 0;
        taskSuspend(0);
    }

    /* accept a partner */
    while ((tempfd = accept(listenfd, (struct sockaddr *)&serverAddr, &addrlen)) != ERROR)
    {
	char	c;
	char	cmdtext [200];
	char	*argp;
	char	*cmdPtr;
	int	argc;
	int	argv[10];
	int	(*function)();
	char	type;
	int	cmdix;


	/* redirect streams */
	ioTaskStdSet (0, 0, tempfd);
	ioTaskStdSet (0, 1, tempfd);
	ioTaskStdSet (0, 2, tempfd);
	printf ("Emergency shell started - enter command\n");

	/* execute commands from the input stream */
	while (tempfd && !feof(stdin))
	{
	    cmdix = 1;	/* start at 1 to allow '_' addition to routine name */
	    if (read(tempfd, &c, 1) != 1)
	    {
		close (tempfd);
		tempfd = 0;
		break;
	    }
	    while ((c != 0x0d) && (c != 0x0a))
	    {
		if ((c == 0x08) || (c == 0x7f))
		{
		    if (cmdix)
			cmdix --;
		    if (c == 0x7f)
			c = 0x8;
		}
		else
		{
		    if (cmdix < 199)
			cmdtext[cmdix++] = c;
		    write (tempfd, &c, 1);
		}
		if (read(tempfd, &c, 1) != 1)
		{
		    close (tempfd);
		    tempfd = 0;
		    break;
		}
	    }
	    write (tempfd, "\n\r", 2);
	    cmdtext[cmdix] = 0;

	    if (strlen(&cmdtext[1]))
	    {
		/* execute a function */
		printf ("emergency shell - executing %s\n", &cmdtext[1]);

		/* break command into pieces */
		argp = &cmdtext[1];

		/* skip over white space */
		while (*argp && (*argp == ' ' || *argp == '\t'))
		{
		    argp++;
		}
		/* process command text - get routine name, and look it up */
		cmdPtr = argp;
		if ((argp = strchr (argp, ' ')) || (argp = strchr (argp, '\t')))
		{
		    *argp = 0;
		    argp++;
		}
		else
		{
		    argp = argp + strlen(cmdPtr);
		}
		if (strcmp ("exit", cmdPtr) == 0)
		{
		    break;
		}
		else if (symFindByName (sysSymTbl, cmdPtr, (char **)&function, &type) != OK)
		{
		    cmdPtr--;
		    *cmdPtr = '_';
		    if (symFindByName (sysSymTbl, cmdPtr, (char **)&function, &type) != OK)
		    {
			printf ("Unknown function %s cannot execute\n", cmdPtr);
			function = NULL;
		    }

		}

		/* skip to first arg */
		while (*argp && (*argp == ' ' || *argp == '\t'))
		    argp++;

		/* get arguments */
		for (argc = 0; argp && *argp && argc < 10; argc++)
		{
		    /* handle quoted argument (present as string) */
		    if (*argp == '\"')
		    {
			argp++;
			argv [argc] = (int)argp;
			/* go to for terminating '"' */
			argp = strchr (argp, '"');
			if (argp)
			{
			    *argp = 0;
			    argp++;
			    while (*argp && (*argp == ' ' || *argp == '\t'))
				argp++;
		        }
		        else /* missing trailing '"' */
			    break;
		    }
		    else
		    {
			/* set value arg */
			if (*argp == '0')
			{
			    argp++;
			    if ((*argp == 'x') || (*argp == 'X'))
			    {
				argp++; argv[argc] = 0;
				while (*argp && ((('0' <= *argp) && (*argp <= '9'))
				  || (('a' <= *argp) && (*argp <= 'f'))
				  || (('A' <= *argp) && (*argp <= 'F'))))
				{
				    if (('0' <= *argp) && (*argp <= '9'))
					 argv[argc] = (argv[argc] * 16) + *argp - '0';
				    if (('a' <= *argp) && (*argp <= 'f'))
					 argv[argc] = (argv[argc] * 16) + (*argp - 'a') + 10;
				    if (('A' <= *argp) && (*argp <= 'F'))
					 argv[argc] = (argv[argc] * 16) + (*argp - 'A') + 10;
				    argp++;
				}
			    }
			    else
			    {
				argv[argc] = 0;
				while (('0' <= *argp) && (*argp <= '7'))
				{
				    argv[argc] = (argv[argc] * 8) + *argp - '0';
				    argp++;
				}
			    }
			}
			else
		            argv [argc] = atol(argp);
			/* go to next white space - check for blank */
			if ((argp = strchr (argp, ' ')) || (argp = strchr (argp, '\t')))
			{
			    argp++;
			    /* skip to next arg */
			    while (*argp && (*argp == ' ' || *argp == '\t'))
				argp++;
			}
		    }
	    	}
		/* pad argument list to 10 */
		while (argc < 10)
		    argv [argc++] = 0;

		/* submit command */
		if (function)
		{
		    printf ("%s returned %x\n", cmdPtr,
			(*function)(argv[0], argv[1], argv[2], argv[3], argv[4],
			           argv[5], argv[6], argv[7], argv[8], argv[9]));
		}
	    }
	}
	if (tempfd)
	    close (tempfd);
	tempfd = 0;
    }
    close(listenfd);
    listenfd = 0;
}

