/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/************************************************************************/
/*                                                                      */
/*    		        SCRIPT TASK                                     */
/*                                                                      */
/*    This task execs commands from a file. 			        */
/*                                                                      */
/*                                                                      */
/************************************************************************/

/************************* MOD LOG ***************************************
*************************************************************************/

/************************************************************************/
/*                                                                      */
/*            INCLUDE FILES                                             */
/*                                                                      */
/************************************************************************/

#include <vxWorks.h>
#include <logLib.h>
#include <string.h>
#include <stdio.h>
#include <taskLib.h>
#include <symLib.h>
#include <sysSymTbl.h>
#include <usrLib.h>
#include <bootLib.h>
#include <time.h>

/************************************************************************/
/*                                                                      */
/*            GLOBAL DATA WE DEFINE                                     */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*            GLOBAL DATA WE REFERENCE                                  */
/*                                                                      */
/************************************************************************/

/* standard location for bootline in x486 vxWorks */
#define PARAM_STRING_LOCATION (char*)0x1200


/************************************************************************/
/*                                                                      */
/*            OUR PRIVATE DATA                                          */
/*                                                                      */
/************************************************************************/

void scriptTask ( char *scriptfile );


/************************************************************************/
/*                                                                      */
/*                INITIALIZE                                            */
/*                                                                      */
/************************************************************************/

int RunScript (char *filename, int priority)
{
    int runPriority;

    if (priority == 0)
	runPriority = 1;
    else
	runPriority = priority;

    /*
    Spawn the script task. Use the same
    priority as the shell.
    */
    if (taskSpawn("scriptTask", runPriority, VX_FP_TASK, 10000,
                   (FUNCPTR)scriptTask, (int)filename,0,0,0,0,0,0,0,0,0) == ERROR) 
    {
        printErr("Can't spawn scriptTask.\n");
        return (-1);
    }

    /*
    Initialization complete
    */
    return(0);
}

/****************************************************************/
/*                                                              */
/*	            THE SCRIPT TASK 	                        */
/*                                                              */
/*    This is the script executor task.                         */
/*                                                              */
/****************************************************************/

void scriptTask ( char *scriptfile )
{
    char	currentScript [80];
    char	rawtext [200];
    char	cmdtext [200];
    char	*macPtr;
    char	*srcPtr;
    FILE	*inp;
    char	*argp;
    int		argc;
    int		argv[10];
    int		(*function)();
    char	*cmdPtr;
    char	type;
    int		i;
    int		retVal;
    int		conditional;
    BOOT_PARAMS params_struct;
    long	nowUTC;
    struct tm	*tp;

    strncpy (currentScript, scriptfile, sizeof(currentScript));

    /* get TARGET and HOST from the current bootline parameters */

    /* Convert to a structure */
    bootStringToStruct (PARAM_STRING_LOCATION, &params_struct);


    /* Open the input file. */
    if ((inp = fopen (currentScript, "r")) == NULL)
    {
        printErr("scriptTask: Can't open scriptFile %s\n", scriptfile);
        return;
    }
    
    /* execute commands from the input stream */
    cmdtext[0] = '_';
    conditional = 0;
    /* Read into buffer is offset by 1 to allow us to add '_' */
    while (inp && fgets(rawtext, sizeof(rawtext), inp))
    {
	cmdtext[1] = 0;
	/* expand any macros! */
	srcPtr = rawtext;
	while ((macPtr = strstr (srcPtr, "$[")))
	{
	    *macPtr = 0; /* terminate text up to macro */
	    strcat (cmdtext, srcPtr); /* copy into target */
	    *macPtr = '$'; /* restore text for err display */
	    macPtr += 2; /* get to substitution string */
	    if ((srcPtr = strstr (macPtr, "]")))
	    {
		*srcPtr++ = 0; /* terminate the search string */
		if (strcmp (macPtr, "IF") == 0) /* handle built-ins (IF) */
		{
		    *(srcPtr-1) = ']'; /* restore input line for err print */
		    if (conditional == 0)
			conditional = 1;
		    else
		    {
			/* nested IF! */
			logMsg ("Error: cannot nest IF clause!\n%s\n",
				(int)rawtext, 0,0,0,0,0);
			cmdtext[1] = 0; /* prevent further processing of this input line */
			break; 
		    }			
		}
		else if (strcmp (macPtr, "ELSE") == 0) /* handle built-ins (ELSE) */
		{
		    *(srcPtr-1) = ']'; /* restore input line for err print */
		    if (conditional == 3)
			conditional = 2;
		    else if (conditional == 2)
			conditional = 3;
		    else
		    {
			/* misplaced else! */
			logMsg ("Error: ELSE not in IF clause!\n%s\n",
				(int)rawtext, 0,0,0,0,0);
			cmdtext[1] = 0; /* prevent further processing of this input line */
			break; 
		    }			
		}
		else if (strcmp (macPtr, "ENDIF") == 0) /* handle built-ins (ENDIF) */
		{
		    *(srcPtr-1) = ']'; /* restore input line for err print */
		    if ((conditional == 3) || (conditional == 2))
			conditional = 0;
		    else
		    {
			/* misplaced ENDIF! */
			logMsg ("Error: ENDIF not in IF clause!\n%s\n",
				(int)rawtext, 0,0,0,0,0);
			cmdtext[1] = 0; /* prevent further processing of this input line */
			break; 
		    }			
		    conditional = 0;
		}
		else if (strcmp (macPtr, "HOST") == 0) /* handle built-ins (hostname) */
		{
		    *(srcPtr-1) = ']'; /* restore input line for err print */
		    strcat (cmdtext, params_struct.hostName);
		}
		else if (strcmp (macPtr, "TARGET") == 0) /* handle built-ins (iocname) */
		{
		    *(srcPtr-1) = ']'; /* restore input line for err print */
		    strcat (cmdtext, params_struct.targetName);
		}
		else if (strcmp (macPtr, "SDATE") == 0) /* handle built-ins (short date (UTC)) */
		{
		    *(srcPtr-1) = ']'; /* restore input line for err print */
		    nowUTC = time(0);
		    tp = gmtime (&nowUTC);

		    sprintf (&cmdtext[strlen(cmdtext)], "%2.2d%2.2d%2.2d",
			    tp->tm_mon+1, tp->tm_mday, tp->tm_year % 100);
		}
		else if (strcmp (macPtr, "CDAY") == 0) /* handle built-ins (start of current
							 day as unix time) */
		{
		    *(srcPtr-1) = ']'; /* restore input line for err print */
		    nowUTC = time(0);
		    tp = gmtime (&nowUTC);

		    tp->tm_hour = 0;
		    tp->tm_min = 0;
		    tp->tm_sec = 0;
		    sprintf (&cmdtext[strlen(cmdtext)], "%ld", mktime (tp));
		}
		else if (getenv (macPtr))/* look up in env strings */
		{
		    strcat (cmdtext, getenv (macPtr));
		    *(srcPtr-1) = ']'; /* restore input line for err print */
		}
		else /* undefined macro */
		{
		    *(srcPtr-1) = ']'; /* restore input line for err print */
		    logMsg ("Error: undefined string substitution!\n%s\n",
			    (int)rawtext, 0,0,0,0,0);
		    cmdtext[1] = 0; /* prevent further processing of this input line */
		    break; 
		}
	    }
	    else
	    {
		/* unterminated macro! */
		logMsg ("Error: unterminated string substitution!\n%s\n",
			(int)rawtext, 0,0,0,0,0);
		cmdtext[1] = 0; /* prevent further processing of this input line */
		break; 
	    }
	}
	strcat (cmdtext, srcPtr); /* copy into target */

	if (strlen(&cmdtext[1]))
	{
	    if (cmdtext[strlen(&cmdtext[1])] == '\n')
		cmdtext[strlen(&cmdtext[1])] = 0;

	    /* break command into pieces (start after space for '_') */
	    argp = &cmdtext[1];

	    /* skip over white space */
	    while (*argp && (*argp == ' ' || *argp == '\t'))
	    {
		argp++;
	    }
	    /* skip line if comment char */
	    if (*argp == 0 || *argp == '#')
	    {
		continue;
	    }

	    /* execute a function */
	    logMsg ("Script executing: %s\n", (int)&cmdtext[1],0,0,0,0,0);

	    /* process command text - get routine name, and look it up */
	    cmdPtr = argp;
	    while (*argp && (*argp != ' ' && *argp != '\t' && *argp != '('))
	    {
		argp++;
	    }
	    *argp = 0; argp++;

	    /* first check for routine asis */
	    if (symFindByName (sysSymTbl, cmdPtr, (char **)&function, &type) != OK)
	    {
		/* otherwise look for "_<name>" */
		cmdPtr--;
		*cmdPtr = '_';
		if (symFindByName (sysSymTbl, cmdPtr, (char **)&function, &type) != OK)
		{
		    printf ("Unknown function %s cannot execute\n", cmdPtr);
		    function = NULL;
		}

	    }

	    /* skip to first arg  (ignore parens for scripts) */
	    while (*argp && (*argp == ' ' || *argp == '\t' || *argp == '('))
		argp++;

	    /* watch for builtin ld with redirect */
	    if (strcmp ("_ld", cmdPtr) == 0)
	    {
		if (*argp == '<') /* next item is filename */
		{
		    argp++;
		    argv [0] = (int)(argp + 1);
		    while (*argp && (*argp != ' ' || *argp == '\t'))
			argp++;
		    *argp = 0;
		    if ((argv[1] = (int)ld (0,1,(char *)argv[0])) == 0)
		    {
			logMsg ("Load %s failed, aborting script\n", argv[0],0,0,0,0,0);
			return;
		    }
		    else
			logMsg ("Load %s returned %lx\n", argv[0],argv[1],0,0,0,0);
		    continue;
		}
	    }
	    
	    /* pad argument list to 10 with 0 */
	    for (i = 0; i < 10; i++)
		argv [i] = 0;

	    /* get arguments, all separated by commas */
	    for (argc = 0; argp && *argp && *argp != ')' && argc < 10; argc++)
	    {
		/* skip blanks  (ignore parens and commas for scripts) */
		while (*argp && (*argp == ' ' || *argp == '\t'
		  || *argp == '(' || *argp == ')' || *argp == ','))
		    argp++;

		/* Terminate line at comment character */
		if (*argp == '#')
		{
		    break;
		}
		/* handle quoted argument (present it as a string) */
		if (*argp == '\"')
		{
		    argp++;
		    argv [argc] = (int)argp;
		    /* go to terminating '"' */
		    argp = strchr (argp, '\"');
		    if (argp)
		    {
			*argp = 0;
			argp++;
		    }
		    else /* missing trailing '"' */
		    {
			logMsg ("Missing trailing \"\n", 0,0,0,0,0,0);
		    }
		}
		else
		{
		    /* set value arg */
		    if (*argp == '0')
		    {
			argp++;
			if ((*argp == 'x') || (*argp == 'X'))
			{
			    argp++; argv[argc] = 0;
			    while (*argp && ((('0' <= *argp) && (*argp <= '9'))
			      || (('a' <= *argp) && (*argp <= 'f'))
			      || (('A' <= *argp) && (*argp <= 'F'))))
			    {
				if (('0' <= *argp) && (*argp <= '9'))
				     argv[argc] = (argv[argc] * 16) + *argp - '0';
				if (('a' <= *argp) && (*argp <= 'f'))
				     argv[argc] = (argv[argc] * 16) + (*argp - 'a') + 10;
				if (('A' <= *argp) && (*argp <= 'F'))
				     argv[argc] = (argv[argc] * 16) + (*argp - 'A') + 10;
				argp++;
			    }
			}
			else
			{
			    argv[argc] = 0;
			    while (('0' <= *argp) && (*argp <= '7'))
			    {
				argv[argc] = (argv[argc] * 8) + *argp - '0';
				argp++;
			    }
			}
		    }
		    else
		    {
		        argv [argc] = atol(argp);
			/* go to next white space or delimiter - check for blank */
			while (*argp && *argp != ' ' && *argp != '\t' && *argp != ',')
			{
			    argp++;
			}
		    }
		}
	    }
	    /* submit command */
	    if (function)
	    {
		if (conditional != 3)
		    retVal = (*function)(argv[0], argv[1], argv[2], argv[3], argv[4],
			       argv[5], argv[6], argv[7], argv[8], argv[9]);
		else
		{
		   retVal = 0;
		    printf ("Skipping: ");
		}
		printf ("%s (", cmdPtr);
		for (i = 0; i < argc; i++)
		{
		    if (i != 0)
			printf (", ");
		    printf ("%x", argv [i]);
		}
		if (conditional != 3)
		    printf (") returned %x\n", retVal);
		if (conditional == 1)
		{
		    if (retVal)
			conditional = 2; /* execute up to ELSE or ENDIF */
		    else
			conditional = 3; /* skip up to ELSE or ENDIF */
		}
	    }
	}
    }
    if (inp)
	fclose (inp);
}

