/*camessage.c,v 1.1.1.1 1999/08/03 14:44:29 tis4000 Exp */

/*	@(#)camessage.c
 *   @(#)camessage.c	1.2	6/27/91
 *	Author:	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		(505) 665 1831
 *	Date:	5-88
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * 	Modification Log:
 * 	-----------------
 *	.01 joh 012290 	placed break in ca_cancel_event search so entire 
 *			list is not searched.
 *	.02 joh	011091	added missing break to error message send 
 *			chix select switch
 *	.03 joh	071291	now timestanmps channel in use block
 *	.04 joh	071291	code for IOC_CLAIM_CIU command
 *	.05 joh 082691  use db_post_single_event() instead of read_reply()
 *			to avoid deadlock condition between the client
 *			and the server.
 *	.06 joh	110491	lock added for IOC_CLAIM_CIU command
 *	.07 joh	021292	Better diagnostics
 *	.08 joh	021492	use lstFind() to verify chanel in clear_channel()
 *	.09 joh 031692	dont send exeception to the client if bad msg
 *			detected- just disconnect instead
 */

#include <vxWorks.h>
#include <taskLib.h>
#include <types.h>
#include <in.h>

#include "db_access.h"
#include "dbDefs.h"	/* agi */
#include "dbCommon.h"   /* agi */
#include "tsk_params.h"
#include "server.h"
#include "caerr.h"
#include "restable.h"  /* agi */

static struct extmsg nill_msg;

#define MPTOPADDR(MP) (&((struct channel_in_use *)(MP)->m_pciu)->addr)
#define	RECORD_NAME(PADDR) (((struct db_addr *)(PADDR))->precord)

LOCAL void build_reply(FAST struct extmsg *mp, struct client  *client);
LOCAL void read_reply(FAST struct event_ext *pevext, FAST struct db_addr *paddr;,
		      int hold, void *pfl);
LOCAL void read_sync_reply(FAST struct extmsg *mp, struct client  *client);
LOCAL void event_cancel_reply(FAST struct extmsg *mp, struct client  *client);
LOCAL void clear_channel_reply(FAST struct extmsg *mp, struct client  *client);
LOCAL void send_err(struct extmsg  *curp, int status, struct client *client, char *footnote);
LOCAL void log_header (FAST struct extmsg *mp, int mnum);
LOCAL void search_fail_reply(FAST struct extmsg *mp, struct client  *client);


/*
 * CAMESSAGE()
 *
 *
 */
camessage(client, recv)
	struct client  *client;
	struct message_buffer *recv;
{
	int            		nmsg = 0;
	unsigned		msgsize;
	unsigned		bytes_left;
	int             	status;
	FAST struct extmsg 	*mp;
	FAST struct event_ext 	*pevext;


	if (CASDEBUG > 2)
		logMsg("CAS: Parsing %d(decimal) bytes\n", recv->cnt);

	bytes_left = recv->cnt;
	while (bytes_left) {

		if(bytes_left < sizeof(*mp))
			return OK;

		mp = (struct extmsg *) &recv->buf[recv->stk];

		msgsize = mp->m_postsize + sizeof(*mp);

		if(msgsize > bytes_left) 
			return OK;

		nmsg++;

		if (CASDEBUG > 2)
			log_header(mp, nmsg);

		switch (mp->m_cmmd) {
		case IOC_NOOP:	/* verify TCP */
			noop_reply (mp, client);
			break;

		case IOC_EVENT_ADD:

			FASTLOCK(&rsrv_free_eventq_lck);
			pevext = (struct event_ext *) 
				lstGet(&rsrv_free_eventq);
			FASTUNLOCK(&rsrv_free_eventq_lck);
			if (!pevext) {
				int size = db_sizeof_event_block() 
						+ sizeof(*pevext);
				pevext =
					(struct event_ext *) malloc(size);
				if (!pevext) {
					LOCK_CLIENT(client);
					send_err(
						mp,
						ECA_ALLOCMEM, 
						client, 
						RECORD_NAME(MPTOPADDR(mp)));
					UNLOCK_CLIENT(client);
					break;
				}
			}
			pevext->msg = *mp;
			pevext->mp = &pevext->msg;	/* for speed- see
							 * IOC_READ */
			pevext->client = client;
			pevext->send_lock = TRUE;
			pevext->size = (mp->m_count - 1) 
				* dbr_value_size[mp->m_type] +
				dbr_size[mp->m_type];

			status = db_add_event(client->evuser,
					      MPTOPADDR(mp),
					      read_reply,
					      pevext,
			   (unsigned) ((struct monops *) mp)->m_info.m_mask,
					      pevext + 1);
			if (status == ERROR) {
				LOCK_CLIENT(client);
				send_err(
					mp, 
					ECA_ADDFAIL, 
					client, 
					RECORD_NAME(MPTOPADDR(mp)));
				UNLOCK_CLIENT(client);
				FASTLOCK(&rsrv_free_eventq_lck);
				lstAdd(&rsrv_free_eventq, pevext);
				FASTUNLOCK(&rsrv_free_eventq_lck);
				break;
			}

			/*
			 * Only add to the list if we can get enough
			 * memory. If not, attempts to delete it 
			 * from the client will cause a warning message
			 * to be printed since it will not be found on
			 * the list.
			 */
			lstAdd(	&((struct channel_in_use *)mp->m_pciu)->eventq, 
				pevext);

			/*
			 * allways send it once at event add
			 */
			/*
			 * if the client program issues many monitors
			 * in a row then I recv when the send side
			 * of the socket would block. This prevents
			 * a application program initiated deadlock.
			 *
			 * However when I am reconnecting I reissue 
			 * the monitors and I could get deadlocked.
			 * The client is blocked sending and the server
			 * task for the client is blocked sending in
			 * this case. I cant check the recv part of the
			 * socket in the client since I am still handling an
			 * outstanding recv ( they must be processed in order).
			 * I handle this problem in the server by using
			 * post_single_event() below instead of calling
			 * read_reply() in this module. This is a complete
			 * fix since a monitor setup is the only request
			 * soliciting a reply in the client which is 
			 * issued from inside of service.c (from inside
			 * of the part of the ca client which services
			 * messages sent by the server).
			 */

			db_post_single_event(pevext+1);

			break;

		case IOC_EVENT_CANCEL:
			event_cancel_reply(mp, client);
			break;

		case IOC_CLEAR_CHANNEL:
			clear_channel_reply(mp, client);
			break;

		case IOC_READ_NOTIFY:
		case IOC_READ:
			{
				struct event_ext evext;

				pevext = &evext;
				pevext->mp = mp;
				pevext->client = client;
				pevext->send_lock = TRUE;
				if (mp->m_count == 1)
					pevext->size = dbr_size[mp->m_type];
				else
					pevext->size = (mp->m_count - 1) * 
						dbr_value_size[mp->m_type] +
						dbr_size[mp->m_type];

				/*
				 * Arguments to this routine organized in
				 * favor of the standard db event calling
				 * mechanism-  routine(userarg, paddr). See
				 * events added above.
				 * 
				 * Hold argument set true so the send message
				 * buffer is not flushed once each call.
				 */
				read_reply(pevext, MPTOPADDR(mp), TRUE, NULL);
				break;
			}
		case IOC_SEARCH:
		case IOC_BUILD:
			build_reply(mp, client);
			break;
		case IOC_WRITE:
			status = db_put_field(
					      MPTOPADDR(mp),
					      mp->m_type,
					      mp + 1,
					      mp->m_count
				);
			if (status < 0) {
				LOCK_CLIENT(client);
				send_err(
					mp, 
					ECA_PUTFAIL, 
					client, 
					RECORD_NAME(MPTOPADDR(mp)));
				UNLOCK_CLIENT(client);
			}
			break;
		case IOC_EVENTS_ON:
			{
				struct event_ext evext;
				struct channel_in_use *pciu =
				(struct channel_in_use *) & client->addrq;

				client->eventsoff = FALSE;

				LOCK_CLIENT(client);
				while (pciu = (struct channel_in_use *) 
					pciu->node.next) {

					pevext = (struct event_ext *) 
						& pciu->eventq;
					while (pevext = (struct event_ext *) 
						pevext->node.next){

						if (pevext->modified) {
							evext = *pevext;
							evext.send_lock = FALSE;
							read_reply(
								&evext, 
								MPTOPADDR(&pevext->msg), 
								TRUE, 
								NULL);
							pevext->modified = FALSE;
						}
					}
				}
				UNLOCK_CLIENT(client);
				break;
			}
		case IOC_EVENTS_OFF:
			client->eventsoff = TRUE;
			break;
		case IOC_READ_SYNC:
			read_sync_reply(mp, client);
			break;
		case IOC_CLAIM_CIU:
			/*
			 * remove channel in use block from
			 * the UDP client where it could time
			 * out and place it on the client
			 * who is claiming it
			 */

			/*
			 * clients which dont claim their 
			 * channel in use block prior to
			 * timeout must reconnect
			 */
			LOCK_CLIENT(prsrv_cast_client);
               		status = lstFind(
					&prsrv_cast_client->addrq, 
					mp->m_pciu);
			if(status >= 0){
				lstDelete(
					&prsrv_cast_client->addrq, 
					mp->m_pciu);
			}
			UNLOCK_CLIENT(prsrv_cast_client);
			if(status < 0){
       				free_client(client);
				logMsg("CAS: client timeout disconnect\n");
				taskSuspend (taskIdSelf());
			}
			LOCK_CLIENT(client);
			lstAdd(&client->addrq, mp->m_pciu);
			UNLOCK_CLIENT(client);
			break;
		default:
			logMsg("CAS: bad msg detected\n");
			log_header(mp, nmsg);
#if 0	
			/* 
			 *	most clients dont recover
			 *	from this
			 */
			LOCK_CLIENT(client);
			send_err(mp, ECA_INTERNAL, client, "Invalid Msg");
			UNLOCK_CLIENT(client);
#endif
			/*
			 * returning ERROR here disconnects
			 * the client with the bad message
			 */
			logMsg("CAS: forcing disconnect ...\n");
			return ERROR;
		}

		recv->stk += msgsize;
		bytes_left = recv->cnt - recv->stk;
	}


	return OK;
}


/*
 *
 *	read_sync_reply()
 *
 *
 */
LOCAL void
noop_reply(mp, client)
	FAST struct extmsg *mp;
	struct client  *client;
{
	FAST struct extmsg *reply;

	LOCK_CLIENT(client);
	reply = (struct extmsg *) ALLOC_MSG(client, 0);
	if (!reply)
		taskSuspend(0);

	*reply = *mp;

	END_MSG(client);

	UNLOCK_CLIENT(client);

	return;
}


/*
 *
 *	clear_channel_reply()
 *
 *
 */
LOCAL void
clear_channel_reply(mp, client)
FAST struct extmsg *mp;
struct client  *client;
{
        FAST struct extmsg *reply;
        FAST struct event_ext *pevext;
        struct channel_in_use *pciu;
        FAST int        status;

        /*
         *
         * Verify the channel
         *
         */
        pciu = (struct channel_in_use *) mp->m_pciu;
 	status = lstFind(
			&client->addrq, 
			mp->m_pciu);
	if(status < 0){
                logMsg("CAS: Attempt to delete nonexistent channel ignored\n");
                return;
        }

        while (pevext = (struct event_ext *) lstGet(&pciu->eventq)) {

		status = db_cancel_event(pevext + 1);
		if (status == ERROR){
			taskSuspend(0);
		}
		FASTLOCK(&rsrv_free_eventq_lck);
		lstAdd(&rsrv_free_eventq, pevext);
		FASTUNLOCK(&rsrv_free_eventq_lck);
        }


	/*
	 * send delete confirmed message
	 */
	LOCK_CLIENT(client);
	reply = (struct extmsg *) ALLOC_MSG(client, 0);
	if (!reply) {
		UNLOCK_CLIENT(client);
		taskSuspend(0);
	}
	*reply = *mp;

	END_MSG(client);
	lstDelete(&client->addrq, pciu);
	UNLOCK_CLIENT(client);

	FASTLOCK(&rsrv_free_addrq_lck);
	lstAdd(&rsrv_free_addrq, pciu);
	FASTUNLOCK(&rsrv_free_addrq_lck);

	return;
}




/*
 *
 *	event_cancel_reply()
 *
 *
 * Much more efficient now since the event blocks hang off the channel in use
 * blocks not all together off the client block.
 */
LOCAL void
event_cancel_reply(mp, client)
	FAST struct extmsg *mp;
	struct client  *client;
{
	FAST struct extmsg *reply;
	FAST struct event_ext *pevext;
	LIST           *peventq =
	&((struct channel_in_use *) mp->m_pciu)->eventq;
	FAST int        status;

	for (pevext = (struct event_ext *) peventq->node.next;
	     pevext;
	     pevext = (struct event_ext *) pevext->node.next)
		if (pevext->msg.m_available == mp->m_available) {
			status = db_cancel_event(pevext + 1);
			if (status == ERROR)
				taskSuspend(0);
			lstDelete(peventq, pevext);

			/*
			 * send delete confirmed message
			 */
			LOCK_CLIENT(client);
			reply = (struct extmsg *) ALLOC_MSG(client, 0);
			if (!reply) {
				UNLOCK_CLIENT(client);
				taskSuspend(0);
			}
			*reply = pevext->msg;
			reply->m_postsize = 0;

			END_MSG(client);
			UNLOCK_CLIENT(client);

			FASTLOCK(&rsrv_free_eventq_lck);
			lstAdd(&rsrv_free_eventq, pevext);
			FASTUNLOCK(&rsrv_free_eventq_lck);

			return;
		}
	/*
	 * Not Found- return an error message
	 */
	LOCK_CLIENT(client);
	send_err(mp, ECA_BADMONID, client, NULL);
	UNLOCK_CLIENT(client);

	return;
}



/*
 *
 *	read_reply()
 *
 *
 */
LOCAL void
read_reply(pevext, paddr, hold, pfl)
FAST struct event_ext *pevext;
FAST struct db_addr *paddr;
int             hold;	/* more on the way if true */
void		*pfl;
{
	FAST struct extmsg *mp = pevext->mp;
	FAST struct client *client = pevext->client;
	FAST struct extmsg *reply;
	FAST int        status;
	FAST int        strcnt;

	/*
	 * If flow control is on set modified and send for later
	 */
	if (client->eventsoff) {
		pevext->modified = TRUE;
		return;
	}
	if (pevext->send_lock)
		LOCK_CLIENT(client);

	reply = (struct extmsg *) ALLOC_MSG(client, pevext->size);
	if (!reply) {
		send_err(mp, ECA_TOLARGE, client, RECORD_NAME(paddr));
		if (pevext->send_lock)
			UNLOCK_CLIENT(client);
		return;
	}
	*reply = *mp;
	reply->m_postsize = pevext->size;
	reply->m_pciu = (void *) ((struct channel_in_use *) mp->m_pciu)->chid;
	status = db_get_field(
			      paddr,
			      mp->m_type,
			      reply + 1,
			      mp->m_count,
			      pfl);
	if (status < 0) {
		send_err(mp, ECA_GETFAIL, client, RECORD_NAME(paddr));
		log_header(mp, 0);
	}
	else{
		/*
		 * force string message size to be the true size rounded to even
		 * boundary
		 */
		if (mp->m_type == DBR_STRING && mp->m_count == 1) {
			/* add 1 so that the string terminator will be shipped */
			strcnt = strlen(reply + 1) + 1;
			reply->m_postsize = strcnt;
		}
		else if (mp->m_type == DBR_CTRL_ENUM) {
			/* compress strings in message (watch out for screwed alignments)*/
			int strix;
 			char *toptr = ((struct dbr_ctrl_enum *)(reply+1))->strs [0];
			for (strix = 0; strix < ((struct dbr_ctrl_enum *)(reply+1))->no_str; strix++) {
				strcpy (toptr, ((struct dbr_ctrl_enum *)(reply+1))->strs [strix]);
				toptr += strlen(toptr)+1;
			}
			if ((long)toptr & 1)
				toptr++;
			bcopy (&((struct dbr_ctrl_enum *)(reply+1))->value, toptr, 
				pevext->size - sizeof(struct dbr_ctrl_enum) + sizeof(short));
			reply->m_postsize = pevext->size - ((long)(&((struct dbr_ctrl_enum *)(reply+1))->value) - (long)toptr);
		}
		END_MSG(client);
	
		/*
		 * Ensures timely response for events, but does que 
		 * them up like db requests when the OPI does not keep up.
		 */
		if (!hold)
			cas_send_msg(client,FALSE);
	}

	if (pevext->send_lock)
		UNLOCK_CLIENT(client);

	return;
}


/*
 *
 *	read_sync_reply()
 *
 *
 */
LOCAL void
read_sync_reply(mp, client)
	FAST struct extmsg *mp;
	struct client  *client;
{
	FAST struct extmsg *reply;

	LOCK_CLIENT(client);
	reply = (struct extmsg *) ALLOC_MSG(client, 0);
	if (!reply)
		taskSuspend(0);

	*reply = *mp;

	END_MSG(client);

	UNLOCK_CLIENT(client);

	return;
}


/*
 *
 *	build_reply()
 *
 *
 */
LOCAL void
build_reply(mp, client)
	FAST struct extmsg *mp;
	struct client  *client;
{
	LIST           			*addrq = &client->addrq;
	FAST struct extmsg 		*search_reply;
	FAST struct extmsg 		*get_reply;
	/*FAST*/ int        		status;	/*can't take & of reg var... see below*/
	struct db_addr  		tmp_addr;
	FAST struct channel_in_use 	*pchannel;



	/* Exit quickly if channel not on this node */
	status = db_name_to_addr(
			mp->m_cmmd == IOC_BUILD ? mp + 2 : mp + 1, 
			&tmp_addr);
	if (status < 0) {
		if (CASDEBUG > 2)
			logMsg(	"CAS: Lookup for channel \"%s\" failed\n", 
				mp + 1);
		if (mp->m_type == DOREPLY)
			search_fail_reply(mp, client);
		return;
	}

	/* Exit quickly if resource channel access flag clear  - agi */
	if ((resGetCaAnswer(((struct dbCommon *)tmp_addr.precord)->radr, &status) != RES_OK) /*&status illegal if reg var*/
          || (status == 0)) {
		if (CASDEBUG > 2)
			logMsg(	"CAS: Channel \"%s\" in slave mode\n", 
				mp + 1);
		if (mp->m_type == DOREPLY)
			search_fail_reply(mp, client);
		return;
	}

	/* get block off free list if possible */
	FASTLOCK(&rsrv_free_addrq_lck);
	pchannel = (struct channel_in_use *) lstGet(&rsrv_free_addrq);
	FASTUNLOCK(&rsrv_free_addrq_lck);
	if (!pchannel) {
		pchannel = (struct channel_in_use *) calloc(1, sizeof(*pchannel));
		if (!pchannel) {
			LOCK_CLIENT(client);
			send_err(mp, ECA_ALLOCMEM, client, RECORD_NAME(&tmp_addr));
			UNLOCK_CLIENT(client);
			return;
		}
	}
	pchannel->ticks_at_creation = tickGet();
	pchannel->addr = tmp_addr;
	pchannel->chid = (void *) mp->m_pciu;


	/*
	 * UDP reliability schemes rely on both msgs in same reply Therefore
	 * the send buffer locked while both messages are placed
	 */
	LOCK_CLIENT(client);

	/* store the addr block in a Q so it can be deallocated */
	lstAdd(addrq, pchannel);

	if (mp->m_cmmd == IOC_BUILD) {
		FAST short      type = (mp + 1)->m_type;
		FAST unsigned int count = (mp + 1)->m_count;
		FAST unsigned int size;

		/*
		 * must be large enough to hold both the search and the build-get
		 * reply in one UDP message. Hence the extra sizeof(*mp) added 
		 * in below.
		 */
		size = 	sizeof(*mp) + 		/* search reply hdr size 	*/
			sizeof(*mp) +		/* build get reply hdr size 	*/
			(count - 1) * 		/* size of n-1 array elements 	*/
				dbr_value_size[type] 	
			+ dbr_size[type];	/* size of the structure fetched */

		get_reply = (struct extmsg *) ALLOC_MSG(client, size);
		if (!get_reply) {
			/* tell them that their request is to large */
			send_err(mp, ECA_TOLARGE, client, RECORD_NAME(&tmp_addr));
		} else {
			struct event_ext evext;

			evext.mp = mp + 1;
			/* pchannel ukn prior to connect */
			evext.mp->m_pciu = pchannel;
			evext.client = client;
			evext.send_lock = FALSE;
			evext.size = size;

			/*
			 * Arguments to this routine organized in favor of
			 * the standard	db event calling mechanism-
			 * routine(userarg, paddr). See events added above.
			 * Hold argument set true so the send message buffer
			 * is not flushed once each call.
			 */
			read_reply(&evext, &tmp_addr, TRUE, NULL);

			/*
			 * this allows extra build replies 
			 * to be dicarded
			 */
			get_reply->m_cmmd = IOC_READ_BUILD;
		}
	}
	search_reply = (struct extmsg *) ALLOC_MSG(client, 0);
	if (!search_reply)
		taskSuspend();

	*search_reply = *mp;
	search_reply->m_postsize = 0;

	/* this field for rmt machines where paddr invalid */
	search_reply->m_type = tmp_addr.field_type;
	search_reply->m_count = tmp_addr.no_elements;
	search_reply->m_pciu = (void *) pchannel;

	END_MSG(client);
	UNLOCK_CLIENT(client);

	return;
}


/* 	search_fail_reply()
 *
 *	Only when requested by the client 
 *	send search failed reply
 *
 *
 */
LOCAL void
search_fail_reply(mp, client)
	FAST struct extmsg *mp;
	struct client  *client;
{
	FAST struct extmsg *reply;

	LOCK_CLIENT(client);
	reply = (struct extmsg *) ALLOC_MSG(client, 0);
	if (!reply) {
		taskSuspend(0);
	}
	*reply = *mp;
	reply->m_cmmd = IOC_NOT_FOUND;
	reply->m_postsize = 0;

	END_MSG(client);
	UNLOCK_CLIENT(client);

}


/* 	send_err()
 *
 *	reflect error msg back to the client
 *
 *	send buffer lock must be on while in this routine
 *
 */
LOCAL void
send_err(curp, status, client, footnote)
	struct extmsg  *curp;
	int             status;
	struct client  *client;
	char           *footnote;
{
	FAST struct extmsg *reply;
	FAST int        size;

	/*
	 * force string post size to be the true size rounded to even
	 * boundary
	 */
	size = strlen(footnote)+1;
	size += sizeof(*curp);

	reply = (struct extmsg *) ALLOC_MSG(client, size);
	if (!reply){
		printf(	"caserver: Unable to deliver err msg [%s]\n",
			ca_message(status));
		return;
	}

	*reply = nill_msg;
	reply->m_cmmd = IOC_ERROR;
	reply->m_available = status;
	reply->m_postsize = size;
	switch (curp->m_cmmd) {
	case IOC_EVENT_ADD:
	case IOC_EVENT_CANCEL:
	case IOC_READ:
	case IOC_READ_NOTIFY:
	case IOC_SEARCH:
	case IOC_BUILD:
	case IOC_WRITE:
		reply->m_pciu = (void *)
			 ((struct channel_in_use *) curp->m_pciu)->chid;
		break;
	case IOC_EVENTS_ON:
	case IOC_EVENTS_OFF:
	case IOC_READ_SYNC:
	case IOC_SNAPSHOT:
	default:
		reply->m_pciu = (void *) NULL;
		break;
	}
	*(reply + 1) = *curp;
	strcpy(reply + 2, footnote);

	END_MSG(client);

}


/* 	log_header()
 *
 *	Debug aid - print the header part of a message.
 *
 */
LOCAL void
log_header (mp, mnum)
FAST struct extmsg *mp;
{
	logMsg(	"CAS: N=%d cmd=%d type=%d pstsize=%d paddr=%x avail=%x\n",
	  	mnum, 
	  	mp->m_cmmd,
	  	mp->m_type,
	  	mp->m_postsize,
	  	MPTOPADDR(mp),
	  	mp->m_available);
  	if(mp->m_cmmd==IOC_WRITE && mp->m_type==DBF_STRING)
    		logMsg("CAS: The string written: %s \n",mp+1);
}
