/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		CHANNEL ACCESS REMOTE SERVER TASK		*/
/*								*/
/*	This task resides in the IOC. It services two kinds	*/
/*	of events. They are the arrival of a message from	*/
/*	one of our sockets and the presence of data in the	*/
/*	event queue. An I/O slave task will wake us up when	*/
/*	a message arrives by giving a semaphore. The routine	*/
/*	that writes into the event queue will also set that	*/
/*	same semaphore.						*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <string.h>
#include <semLib.h>
#include <rngLib.h>
#include <sockLib.h>
#include <selectLib.h>
#include <stdio.h>
#include <types.h>
#include <sys/times.h>
#include <vxLib.h>
#include <socket.h>
#include <in.h>
#include <net/if.h>
#include <netinet/tcp.h>
#include <ioLib.h>
#include <tickLib.h>
#include <logLib.h>
#include <inetLib.h>

#include "inet_ntoa.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "paramStructs.h"
#include "fast_lock.h"
#include "serverControl.h"
#include "agentMsgs.h"
#include "task_params.h"
#include "agentPorts.h"
#include "dbAccess.h"
#include "dataTypes.h"
#include "dbRecDes.h"
#include "caRing.h"
#include "restable.h"
#include "recSup.h"
/*#include "caXdr.h"*/
#include "tisByteOrder.h"
#include "mapTypes.h"
#include "cadef.h"
#include "rbMgt.h"
#include "getAgentField.h"
#include "pvd.h"
#include "traverseDb.h"
#include "localAddr.h"
#include "rsrv.h"


#ifdef PRIV_AUDIT
/* for the auditing and priviledge functions */
#include "sslShared.h"
#include "iocPriv.h"
#include "auditSystem.h"
extern AUDIT_ID	auditLog;	/* the system log */

static char		X509_dn [X509_MAX_DN]; /* distinguished name from certificate */
#endif PRIV_AUDIT

/*
	Global Data
*/

typedef struct alq
    {
    CH_ALARM		alarmMsg;
    char		alarmData[150];
    RES_TABLE		*radr;
    struct alq		*nextAlq;
    } ALARM_QUE;
ALARM_QUE		*freeAlarms = 0;	/* Free chain */
ALARM_QUE		*alarmHead = 0;		/* Head of Queue */
ALARM_QUE		*alarmTail = 0;		/* Tail of Queue */
int			rsrvPrintingWarnings = 1;	/* Print alarm queue warnings */
int			maxAlarmQue = 100;	/* Max Queue depth */
int			alarmQueSize = 0;	/* Current Queue size */
int			alarmSerial = 0;	/* Last serial # issued */
int			rsrvIoFlag = 0;		/* I/O this call */
SEM_ID			rsrvSem;		/* Our semaphore */
SEM_ID			rsrvIoSem;		/* Slave's semaphore */
fd_set			rsrvFds;		/* Fd mask, all sockets */
int			rsrvUdpFd;		/* Our UDP port */
int			rsrvTcpFd;		/* Out TCP port */
int			rsrvMaxFd;		/* Max fd used */
int			rsrvInet;		/* My Inet addr */
static int		debugging = 0;
int			rsrvRxTimeout = 4200;	/* Default at 70 seconds */
CCB			*firstClient = 0;
#ifdef vxWorks
extern struct timeval	pollTimeout;
#else
struct timeval		pollTimeout = {0, 0};
#endif
static char		buff[2048];	/* General Get/Put buffer */
static char		rsrvOutBuff[2048];	/* Auxiliary put buffer */

/*
	External Data
*/

extern RING_ID		ringId;
extern RES_TABLE	*resTableHead;
extern struct in_addr	caLocalInet;

/*
	Function Prototypes
*/

int	alqConnect(struct in_addr inetAddr);
int	clientRecv(char *buff, int nbytes, CCB *currCCB, fd_set *localFds);
int	closeMCB(MCB *pMCB);
int	clientSend(CCB *currCCB, fd_set *otherFds);
CCB	*findCCB(struct in_addr inetAddr);
int	sendClient(CCB *pCCB, short msgType, char *buff1, int buff1Len, 
		   char *buff2, int buff2Len, fd_set *fdMask);
int	raClose(int fd);
int	raMessageProc(int fd);
int	raSendBytes(int fd, char *buff, int nbytes);
int	raStreamRead(int fd, char *buff, int nbytes);
CCB	*rsrvFindFd(int fd);
void	rsrvIoSlave(void);
void	rsrvTask(void);
int	sendNewAlarm(CH_ALARM *alarmMsg, char *alarmData, CCB *client);
int	sendRedAlarms(int fd, short msgType, char *buff, int nbytes);
int	validMCB(MCB *pMCB, CCB *pCCB, fd_set *pFdset);

/****************************************************************/
/*								*/
/*			INITIALIZE				*/
/*								*/
/****************************************************************/

int rsrv_init(void)
{
    struct sockaddr_in	portAddr;
#ifdef USE_UDP
    int			optVal = 1;
    short		msgType;
#endif USE_UDP

#ifdef PRIV_AUDIT
    if (!sslSharedInit ())
	{
	logMsg ("Cannot initialize SSL subsystem!\n",0,0,0,0,0,0);
	taskSuspend (taskIdSelf ());
	}
#endif PRIV_AUDIT

    /*
    Create the semaphores used for task synchronization and the ring buffer used for task messaging.
    */

    rsrvSem = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
    rsrvIoSem = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
    caRbInit();

    /*
    Create a UDP port, bind it to our well know address, and also broadcast out a message that we are 
    alive. Do this one time and one time only. This is an optimization so that all the clients that 
    are anxiously waiting for us to arrive won't wait until their next scheduled time to invite us
    to connect to them.
    */

#ifdef USE_UDP
    rsrvUdpFd = socket(PF_INET, SOCK_DGRAM, 0);
    setsockopt(rsrvUdpFd, SOL_SOCKET, SO_BROADCAST, (caddr_t) &optVal, sizeof(optVal));
    portAddr.sin_family = AF_INET;
    portAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    portAddr.sin_port = htons(RSV_PORT);
    if (bind(rsrvUdpFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == ERROR)
	perror("Can't bind rsrv UDP port");
    msgType = htons(SOCK_CONN);
    portAddr.sin_family = AF_INET;
    {
	struct ifnet *lkup_net;
	extern struct ifnet * ifnet;

	for (lkup_net = ifnet; ifnet; lkup_net = lkup_net->if_next)
	{
	    if ((lkup_net->if_flags & IFF_BROADCAST) == IFF_BROADCAST)
	    {
		portAddr.sin_addr = ((struct sockaddr_in *)&(lkup_net->if_addrlist->ifa_broadaddr))->sin_addr;
		break;	/* only accept first one found */
	    }
	}
	logMsg ("Using %s as broadcast address\n", (int)INET_NTOA(portAddr.sin_addr),0,0,0,0,0);
    }
    portAddr.sin_port = htons(SCM_PORT);
    sendto(rsrvUdpFd, (char *) &msgType, sizeof(msgType), 0, (struct sockaddr *) &portAddr, sizeof(portAddr));
#endif USE_UDP

    /*
    Set up the FD mask so we can use select to serve multiple sockets from a single thread.
    */

    FD_ZERO(&rsrvFds);
#ifdef USE_UDP
    FD_SET(rsrvUdpFd, &rsrvFds);
    rsrvMaxFd = rsrvUdpFd;
#endif USE_UDP
    /*
    Create a TCP port for alarm queue redundancy support.
    */

    rsrvTcpFd = socket(PF_INET, SOCK_STREAM, 0);
    portAddr.sin_family = AF_INET;
    portAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    portAddr.sin_port = htons(RSV_PORT);
    if (bind(rsrvTcpFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == ERROR)
	perror("Can't bind rsrv TCP port");
    listen(rsrvTcpFd, 2);
    FD_SET(rsrvTcpFd, &rsrvFds);
    if (rsrvTcpFd > rsrvMaxFd)
	rsrvMaxFd = rsrvTcpFd;


    /*
    Find our Inet address so we can ignore our own broadcasts.
    */

    local_addr(0, &portAddr);
    rsrvInet = ntohl(portAddr.sin_addr.s_addr);

    /*
    Launch the I/O slave and then the server task itself.
    */

    taskSpawn("rsrv I/O", CA_SRVR_PRI - 1, CA_SRVR_OPT, 2048, (FUNCPTR)rsrvIoSlave, 0,0,0,0,0,0,0,0,0,0);
    taskSpawn(CA_SRVR_NAME, CA_SRVR_PRI, CA_SRVR_OPT, CA_SRVR_STACK, (FUNCPTR)rsrvTask, 0,0,0,0,0,0,0,0,0,0);
    return 0;
}

/****************************************************************/
/*								*/
/*			I/O SLAVE SUBTASK			*/
/*								*/
/*	The only reason for this task is that the real server	*/
/*	wants to wake up on either one of two events. That is	*/
/*	just about the only feature that I miss from VMS.	*/
/*								*/
/****************************************************************/

void rsrvIoSlave(void)
{
    fd_set	myFds;
    CCB		*currCCB;
    RES_TABLE	*currRes;

    /*
    This case is simpler than the one used for dbServerTask.  We can just wait forever because 
    rsrvTask has no time based events to service.
    */

    for (;;)
	{
	
	/*
	Build the list of sockets to watch by traversing the list of clients and the resource table.
	Then add in the two sockets that we always use.
	*/

	FD_ZERO(&myFds);
	for (currCCB = firstClient; currCCB; currCCB = currCCB -> nextClient)
	    if (currCCB -> clientFd)
		FD_SET(currCCB -> clientFd, &myFds);
	for (currRes = resTableHead; currRes; currRes = currRes -> link)
	    if (currRes -> redAlarmFd)
		FD_SET(currRes -> redAlarmFd, &myFds);
#ifdef USE_UDP
	FD_SET(rsrvUdpFd, &myFds);
#endif USE_UDP
	FD_SET(rsrvTcpFd, &myFds);

	/*
	Now wait for something to happen.
	*/

	rsrvIoFlag = 0;
	if (select(rsrvMaxFd + 1, &myFds, NULL, NULL, 0) == ERROR)
	    perror("Select error in rsrvIoSlave");
	rsrvIoFlag = 1;
	semGive(rsrvSem);
	semTake(rsrvIoSem, WAIT_FOREVER);
	}
}

/****************************************************************/
/*								*/
/*		CHANNEL ACCESS SERVER TASK			*/
/*								*/
/****************************************************************/

void rsrvTask(void)
{
    fd_set		myFds;
    short		msgType;
    struct sockaddr_in	portAddr;
    struct in_addr	inetAddr;
    int			fdIndex;
    CCB			*currCCB;
    FCB			*pFCB;
    MCB			*pMCB;
    int			addrSize;
#ifdef USE_UDP
    int			newFd;
#endif USE_UDP
    int			optVal = 1;
    struct dbCommon	*recHdr;
#ifdef PRIV_AUDIT
    SSL			*newSSL = NULL;
    char		updateMsg [PVNAME_SZ + FLDNAME_SZ + 2 + 20];
#endif PRIV_AUDIT    

    for (;;)
	{

	/*
	Wait for someone to give us some work to do.
	*/

	semTake(rsrvSem, WAIT_FOREVER);

	/*
	If the I/O flag is set, check the sockets.
	*/

	if (rsrvIoFlag)
	    {
	    bcopy((char *) &rsrvFds, (char *) &myFds, sizeof(myFds));
	    while (select(rsrvMaxFd + 1, &myFds, NULL, NULL, &pollTimeout) > 0)
		{
		for (fdIndex = 0; fdIndex <= rsrvMaxFd; fdIndex++)
		    {
		    if (!FD_ISSET(fdIndex, &myFds))
			continue;
		
#ifdef USE_UDP  /* no UDP if SSL based system */
		    /*
		    See if there is any activity on the UDP port.
		    */

		    if (fdIndex == rsrvUdpFd)
			{

			/*
			Read just the first two bytes. We aren't interested in any more than that.
			*/

			addrSize = sizeof(portAddr);
			msgType = -1;
			if (recvfrom(rsrvUdpFd, (char *) &msgType, sizeof(msgType), 0,
				(struct sockaddr *) &portAddr, &addrSize) > 0)
			    {
			    /*
			    Handle according to the message type
			    */

			    msgType = ntohs(msgType);
			    if (msgType == SOCK_CONN_REPLY)
				{
				/*
				A client wants us to connect to him. Get his Internet address and try 
				to connect.
				*/
				inetAddr = portAddr.sin_addr;
			        newFd = socket(PF_INET, SOCK_STREAM, 0);
				if (debugging)
				    logMsg ("%s wants me to connect.\n", (int)INET_NTOA(inetAddr),0,0,0,0,0);
			        portAddr.sin_family = AF_INET;
			        portAddr.sin_port = htons(SCM_PORT);
/*
				EXPERIMENT! SET UP FOR NONBLOCKING PRIOR TO CONNECT!
*/
				ioctl(newFd, FIONBIO, optVal);
			        if (connect(newFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) ==
					ERROR)
				    {
				    /*
				    Check to see if the error code is "Connect in progress" This is normal.
				    */

				    if (errno != EINPROGRESS)
					{

				    	/*
				    	Can't connect. Wait until we are invited again.
				    	*/

				    	if (debugging)
					    perror("Can't connect to him");
				    	close(newFd);
				    	continue;
					}
				    }
				if (debugging)
				    logMsg("Connection to %s completed!\n", (int)INET_NTOA(inetAddr),0,0,0,0,0);
				if (!(currCCB = findCCB(inetAddr)))
				    {

				    /*
				    This guy is new. Create a CCB for him and put him in the chain.
				    */

				    currCCB = (CCB *) calloc(1, sizeof(CCB));
				    currCCB -> nextClient = (struct cCl *) firstClient;
				    firstClient = currCCB;
				    currCCB -> clientInet = inetAddr;
				    currCCB -> clientFd = newFd;
				    }

				else
				    {

				    /*
				    We've seen him before. Close the old socket and replace it with
				    this one. He must have crashed and then restarted.
				    */

				    if (currCCB -> clientFd)
					{
					close(currCCB -> clientFd);
					FD_CLR(currCCB -> clientFd, &myFds);
					FD_CLR(currCCB -> clientFd, &rsrvFds);
					}
				    currCCB -> clientFd = newFd;
				    currCCB -> handlerMsgRcvd = 0;
			    	    currCCB -> trendHandler = 0;
			    	    currCCB -> alarmHandler = 0;
				    }

				/*
				Make sure that TCP won't hang onto buffers for this new socket. We want
				the system to look fast.
				*/

				setsockopt(newFd, IPPROTO_TCP, TCP_NODELAY, (char *)&optVal, sizeof(optVal));
				FD_SET(newFd, &rsrvFds);
				if (newFd > rsrvMaxFd)
				    rsrvMaxFd = newFd;

				/*
				Set up for flow control
				*/
				addrSize = sizeof(currCCB -> clientTxBuff);
				getsockopt(newFd, SOL_SOCKET, SO_SNDBUF, (char *)&(currCCB -> clientTxBuff),
					&addrSize);
				addrSize = sizeof(currCCB -> clientRxBuff);
				getsockopt(newFd, SOL_SOCKET, SO_RCVBUF, (char *)&(currCCB -> clientRxBuff),
					&addrSize);
				currCCB -> clientRxTopped = 0;
				}

			    else if (msgType == SOCK_ALQCONN)
				{
				RES_TABLE	*currRes;
				int		newFd;
				int		closedFd = -1;

				/*
				Our redundant partner wants us to connect. Compare our Inet address
				with his. We do the connecting if our Inet address is greater than his.

				KLUGE ALERT: This message is sent by sockMnt in our redundant partner.
				Once every 30 seconds he looks at the resource table. If we know who
				the redundant partner is but don't have a redundant alarm queue socket,
				then he sends this message.
				*/

				inetAddr = portAddr.sin_addr;
				if (debugging)
				    logMsg("rsrv: UDP alarm connection request from %s\n", (int)INET_NTOA(inetAddr),0,0,0,0,0);
				if (ntohl(inetAddr.s_addr) < ntohl(caLocalInet.s_addr))
				    continue;

				portAddr.sin_port = htons(RSV_PORT);
				newFd = socket(PF_INET, SOCK_STREAM, 0);
				if (connect(newFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == ERROR)
				    {
				    perror("rsrv can't connect for redundant alarm update");
				    close(newFd);
				    continue;
				    }
				if (debugging)
				    printf("Connected to redundant alarm partner.\n");
				ioctl(newFd, FIONBIO, optVal);
				if (newFd > rsrvMaxFd)
				    rsrvMaxFd = newFd;
				FD_SET(newFd, &rsrvFds);

				/*
				Now traverse the resource table to find all the resources that we share.
				Post this fd in each one of them.
				*/

				for (currRes = resTableHead; currRes; currRes = currRes -> link)
				    {
				    if (currRes -> partnerInet.s_addr != inetAddr.s_addr)
					continue;
				    if ((currRes -> redAlarmFd) && (currRes -> redAlarmFd != closedFd))
				        {
				        closedFd = currRes -> redAlarmFd;
				        close(closedFd);
					FD_CLR(closedFd, &rsrvFds);
					FD_CLR(closedFd, &myFds);
				        }
				    currRes -> redAlarmFd = newFd;
				    }
				}
					
			    else
				{
			        printErr("rsrv: got unknown msg type %d on UDP.\n", ntohs(msgType));
				}
			    }

			else
			    {
			    perror("rsrv: recv error on UDP");
			    }
			}
#endif USE_UDP
		    else if (fdIndex == rsrvTcpFd)
			{
			int		newFd;
#ifdef USE_UDP
			RES_TABLE	*currEntry;
			int		closedFd = -1;
#endif USE_UDP

			/*
			A redundant partner  (or a sockMnt somewhere) is trying
			to connect with us (for alq in some cases for updating alarms.)
			*/

			addrSize = sizeof(portAddr);
			newFd = accept(rsrvTcpFd, (struct sockaddr *)&portAddr, &addrSize);
#ifndef USE_UDP
			/*
			A client wants us to connect to him. Get his Internet address and try 
			to connect.
			*/
			/* convert inet addr to primary network interface */
			if (portAddr.sin_addr.s_addr == htonl(0x7f000001))
			    local_addr(0, &portAddr);
			inetAddr = portAddr.sin_addr;
#ifdef PRIV_AUDIT
			newSSL = connect_ssl_connection_timed (inetAddr.s_addr, SCM_PORT, 10/*seconds*/);
			close (newFd); /* close the signalling connection (served it's purpose) */
			if (newSSL)
			    {
			    GetPeerCN (newSSL, X509_dn, sizeof(X509_dn));
			    if (GetGeneralPrivHost (X509_dn, "CA_") < 1)					
				{
				GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
				/*
				No priviledge to connect!
				*/
				logMsg("CA priviledge error, %s not authorized\n",
					(int)X509_dn,0,0,0,0,0);
				SecurityAlarm (X509_dn, "CA priviledge error, Not authorized", NULL);
				if (auditLog)
				    {
				    char msg [80];

				    sprintf (msg, "Client privilege error with %s", INET_NTOA(inetAddr));
				    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
					    taskName(taskIdSelf()), msg);
				    }
				sslSharedCloseConn (newSSL);
				continue;
				}
			    else
				{
				GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
				if (auditLog)
				    {
				    char msg [80];

				    sprintf (msg, "Connected with %s", INET_NTOA(inetAddr));
				    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
					    taskName(taskIdSelf()), msg);
				    }
				}
			    }
			else
			    continue; /* connection failed */

			/* get the underlying fd, and promote new socket to non-blocking IO */
			newFd = SSL_get_fd (newSSL);
#else
			close (newFd); /* close the signalling connection (served it's purpose) */
			inetAddr = portAddr.sin_addr;
		        newFd = socket(PF_INET, SOCK_STREAM, 0);
			if (debugging)
			    logMsg ("%s wants me to connect.\n", (int)INET_NTOA(inetAddr),0,0,0,0,0);
		        portAddr.sin_family = AF_INET;
		        portAddr.sin_port = htons(SCM_PORT);
		        if (connect(newFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) ==
				ERROR)
			    {
			    /*
			    Can't connect. Wait until we are invited again.
			    */
			    if (debugging)
				perror("Can't connect to him");
			    close(newFd);
			    continue;
			    }
#endif PRIV_AUDIT
			ioctl(newFd, FIONBIO, optVal);
			if (debugging)
			    logMsg("Connection to %s completed!\n", (int)INET_NTOA(inetAddr),0,0,0,0,0);
			if (!(currCCB = findCCB(inetAddr)))
			    {

			    /*
			    This guy is new. Create a CCB for him and put him in the chain.
			    */

			    currCCB = (CCB *) calloc(1, sizeof(CCB));
			    currCCB -> nextClient = (struct cCl *) firstClient;
			    firstClient = currCCB;
			    currCCB -> clientInet = inetAddr;
			    currCCB -> clientFd = newFd;
			    }

			else
			    {

			    /*
			    We've seen him before. Close the old socket and replace it with
			    this one. He must have crashed and then restarted.
			    */

			    if (currCCB -> clientFd)
				{
#ifdef PRIV_AUDIT
				/* reuse newSSL, new circuit defined by newFd now */
				newSSL = sslSharedGetConn (currCCB -> clientFd);
				if (auditLog)
				    {
				    GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
				    AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
						    taskName(taskIdSelf()), "Prior connection closed");
				    }
				sslSharedCloseConn (newSSL);
#else PRIV_AUDIT
				close (currCCB -> clientFd);
#endif PRIV_AUDIT
				FD_CLR(currCCB -> clientFd, &myFds);
				FD_CLR(currCCB -> clientFd, &rsrvFds);
				}
			    currCCB -> clientFd = newFd;
			    currCCB -> handlerMsgRcvd = 0;
			    currCCB -> trendHandler = 0;
			    currCCB -> alarmHandler = 0;
			    }

			/*
			Make sure that TCP won't hang onto buffers for this new socket. We want
			the system to look fast.
			*/

			setsockopt(newFd, IPPROTO_TCP, TCP_NODELAY, (char *)&optVal, sizeof(optVal));
			FD_SET(newFd, &rsrvFds);
			if (newFd > rsrvMaxFd)
			    rsrvMaxFd = newFd;

			/*
			Set up for flow control
			*/
			addrSize = sizeof(currCCB -> clientTxBuff);
			getsockopt(newFd, SOL_SOCKET, SO_SNDBUF, (char *)&(currCCB -> clientTxBuff),
				&addrSize);
			addrSize = sizeof(currCCB -> clientRxBuff);
			getsockopt(newFd, SOL_SOCKET, SO_RCVBUF, (char *)&(currCCB -> clientRxBuff),
				&addrSize);
			currCCB -> clientRxTopped = 0;
#else USE_UDP
			inetAddr = portAddr.sin_addr;
			ioctl(newFd, FIONBIO, optVal);
			if (debugging)
			    logMsg("New redundant alarm connection from %s on fd %d\n",
				    (int)INET_NTOA(inetAddr), newFd, 0,0,0,0);
			FD_SET(newFd, &rsrvFds);
			if (newFd > rsrvMaxFd)
			    rsrvMaxFd = newFd;
			/*
			Now post this fd in the resource table.
			*/

			for (currEntry = resTableHead; currEntry; currEntry = currEntry -> link)
			    {
			    if (currEntry -> partnerInet.s_addr != inetAddr.s_addr)
				continue;
			    if ((currEntry -> redAlarmFd) && (currEntry -> redAlarmFd !=  closedFd))
				{
				closedFd = currEntry -> redAlarmFd;
				close(closedFd);
				FD_CLR(closedFd, &rsrvFds);
				FD_CLR(closedFd, &myFds);
				}
			    currEntry -> redAlarmFd = newFd;
			    }

			/*
			Send a TCP message to the other guy to make him perform connection processing.
			*/

			msgType = htons(CHAN_ALQCONN);
			sendRedAlarms(newFd, msgType, (char *) 0, 0);

			/*
			And now perform connection processing ourselves.
			*/

			alqConnect(inetAddr);
#endif USE_UDP
			}

		    else
			{
#ifdef PRIV_AUDIT
			int bytesGot = 0;
#endif PRIV_AUDIT
			/*
			Must be a TCP port. Find the CCB.
			*/

			currCCB = rsrvFindFd(fdIndex);
			if (!currCCB)
			    {
			    RES_TABLE	*currRes;

			    /*
			    No CCB. Therefore, it must be a redundant alarm queue update.
			    First, make sure that it really is one of our redundant alarm
			    sockets. If not, get it out of our socket mask forever.
			    */

			    for (currRes = resTableHead; currRes; currRes = currRes -> link)
				if (currRes -> redAlarmFd == fdIndex)
				    break;
			    if (!currRes)
				{

				/*
				False alarm, or the remains of a previous incarnation. Make
				sure it doesn't bother us again.
				*/

				FD_CLR(fdIndex, &rsrvFds);
				FD_CLR(fdIndex, &myFds);
				continue;
				}

			    /*
			    It's good. Go ahead and work with it.
			    */

			    if (fdIndex > 0)
				raMessageProc(fdIndex);
			    continue;
			    }

			/*
			Handle receive flow control
			*/
			ioctl(fdIndex, FIONREAD, (int)&addrSize);
			if (addrSize == 0)
			    {
			    /* Woke up with no data! */
#ifdef PRIV_AUDIT
			    if (auditLog)
				{
				GetPeerDN (sslSharedGetConn(currCCB -> clientFd), X509_dn, sizeof(X509_dn));
				AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
						taskName(taskIdSelf()), "Connection closed, no data on select");
				}
			    sslSharedCloseConn (sslSharedGetConn(currCCB -> clientFd));
#else
			    close (currCCB -> clientFd);
#endif PRIV_AUDIT
			    FD_CLR(currCCB -> clientFd, &rsrvFds);
			    FD_CLR(currCCB -> clientFd, &myFds);
			    currCCB -> clientFd = 0;
			    currCCB -> handlerMsgRcvd = 0;
			    currCCB -> txBuffSize = 0;
			    currCCB -> alarmHandler = 0;
			    currCCB -> trendHandler = 0;
			    }

			if (!(currCCB -> clientRxTopped))
			    {
			    if (addrSize == currCCB -> clientRxBuff)
				{
				currCCB -> clientRxBuff += 1024;
				if (setsockopt(fdIndex, SOL_SOCKET, SO_RCVBUF, (char *)&(currCCB -> clientRxBuff),
					sizeof(currCCB -> clientRxBuff)) == ERROR)
				    currCCB -> clientRxTopped = 1;
				}
			    }

			/*
			Get the message type to find out what kind of message we have
			*/

#ifdef PRIV_AUDIT
			while (currCCB -> clientFd
			  && ((bytesGot = SSL_read (sslSharedGetConn(currCCB -> clientFd),
								 (char *) &msgType, sizeof(msgType))) > 0))
			    {
			    if (bytesGot != sizeof(msgType))
				{
				/* short read! */
				if (!clientRecv(((char *)&msgType)+1, sizeof(msgType)-bytesGot, currCCB, &myFds))
				    continue;
				}

#else
			if (!clientRecv((char *) &msgType, sizeof(msgType), currCCB, &myFds))
			    continue;
			else
			    {
#endif PRIV_AUDIT
			    msgType = ntohs(msgType);

			    if (msgType == BOOMERANG)
				{

				/*
				A boomerang. If there's anything queued up for this client, just
				send that. If nothing, generate a boomerang.
				*/

				if (debugging)
				    printf("Boomerang.\n");
				if (currCCB -> txBuffSize == 0)
				    {
				    currCCB -> txBuffSize = sizeof(msgType);
				    *(short *) (currCCB -> txBuff) = htons(BOOMERANG);
				    }
				clientSend(currCCB, &myFds);
				}

			    else if (msgType == CHAN_SESS_INQ)
				{
				CH_SESS_REPLY	chSessReply;

				/*
				Client sends this just after he gets started. He wants our session ID to
				see if all the channels we had during a previous connection are still
				valid. Just reply. If the numbers match, he'll do nothing. If they don't
				match (i.e., he crashed), then he'll send a separate transaction to
				set the session ID.
				*/

				if (debugging)
				    printf("Session Inquiry.\n");
				chSessReply.chSession = currCCB -> clientSession;
				msgType = htons(CHAN_SESS_REPLY);
				sendClient(currCCB, msgType, (char *) &chSessReply, sizeof(chSessReply),
					    (char *) 0, 0, &myFds);
				}

			    else if (msgType == CHAN_BUILD)
				{
				CH_BUILD		chanBuild;
				CH_BLDACK		bldAck;
				char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
				struct dbAddr	dbAddr;
				FCB			*prevFCB;
				MCB			*prevMCB;

				/*
				Channel Build. Get the message header and the channel name.
				*/

				if (!clientRecv((char *) &chanBuild, sizeof(chanBuild), currCCB, &myFds))
				    continue;
				if (!clientRecv(chanName, ntohs(chanBuild.chNameLen), currCCB, &myFds))
				    continue;

				/*
				Get the address and all other information for this channel. Fill out all
				the static information in the reply.
				*/

				if (debugging)
				    printf("Build received!\n");
#ifdef PRIV_AUDIT
				GetPeerCN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				if (GetChannelPriv (X509_dn, chanName) == PRIV_NOACCESS)
				    {
				    GetPeerDN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				    /*
				    No priviledge for channel!
				    */
				    logMsg("CA priviledge error, %s not authorized for %s\n",
					    (int)X509_dn,(int)chanName,0,0,0,0);
				    SecurityAlarm (X509_dn, "Channel not allowed", chanName);
				    if (auditLog)
					{
					sprintf (updateMsg, "Channel connect to %s not authorized", chanName);
					AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
						taskName(taskIdSelf()), updateMsg);
					}
				    continue;
				    }
#endif PRIV_AUDIT
				dbNameToAddr(chanName, &dbAddr);
				if ((int) dbAddr.precord < 0)
				    continue;
				if ((int) dbAddr.pfield < 0)
				    continue;
				bldAck.chHandle = chanBuild.chHandle;
				bldAck.chDataType = mapNew2Old(dbAddr.dbr_field_type);
				bldAck.chDataType = htons(bldAck.chDataType);
				bldAck.chElementCount = htons(dbAddr.no_elements);
				if (!bldAck.chElementCount)
				    bldAck.chElementCount = htons(1);

				/*
				It is possible to get duplicate requests as a part of error recovery.
				If so, just return the original handle in an acknowledgement.
				*/

				pMCB = currCCB -> firstChan;
				while (pMCB)
				    {
				    if (dbAddr.precord == pMCB -> fldCntl -> dbAddr.precord &&
					    dbAddr.pfield == pMCB ->fldCntl -> dbAddr.pfield &&
					    chanBuild.chHandle == pMCB -> handle)
					{
					bldAck.chIocHandle = (int) pMCB;
					recHdr = (struct dbCommon *) dbAddr.precord;
					resGetCaAnswer(recHdr -> radr, &bldAck.currState);
					msgType = htons(CHAN_BLDACK);
					sendClient(currCCB, msgType, (char *) &bldAck, sizeof(bldAck),
						(char *) 0,0, &myFds);
					break;
					}
				    pMCB = pMCB -> nextClient;
				    }
				if (pMCB)
				    continue;

				/*
				Create a new MCB for this channel. The address of the MCB will become
				the handle that we send back to the client. Fill in the rest of the
				static data.
				*/

				pMCB = newMCB();
				bldAck.chIocHandle = (int) pMCB;
				pMCB -> handle = chanBuild.chHandle;
				pMCB -> dataType = ntohs(chanBuild.chType);
				pMCB -> numElem = ntohs(chanBuild.chCount);
				pMCB -> client = currCCB;
				if (debugging)
				    printf("Chan build for %s, his = %X, mine = %X\n", chanName,
					    bldAck.chHandle, bldAck.chIocHandle);

				/*
				Put this new MCB at the head of the chain for this client.
				*/

				pMCB -> nextClient = currCCB -> firstChan;
				currCCB -> firstChan = pMCB;
				if (pMCB -> nextClient)
				    pMCB -> nextClient -> prevClient = pMCB;
				    

				/*
				March along the monitor chain to see if this field is already monitored
				elsewhere, regardless of data type.
				*/

				recHdr = (struct dbCommon *) dbAddr.precord;
				pFCB = (FCB *) (recHdr -> mlis.node.next);
				prevFCB = pFCB;
				while (pFCB)
				    {
				    if (pFCB -> dbAddr.pfield == dbAddr.pfield)
					break;
				    prevFCB = pFCB;
				    pFCB = (FCB *) (pFCB -> nextFCB);
				    }

				if (!pFCB)
				    {

				    /*
				    Must create a new FCB. Fill it out and put it at the end of the
				    monitor queue, if there is one. By putting at the end instead of
				    at the start, we don't have to lock the list! Dangerous, but fun.
				    */

				    pFCB = newFCB();
				    pFCB -> dbAddr = dbAddr;
				    pFCB -> firstChan = pMCB;
				    pMCB -> fldCntl = pFCB;
				    if (prevFCB)
					{
					pFCB -> prevFCB = prevFCB;
					prevFCB -> nextFCB = pFCB;
					}
				    else
					recHdr -> mlis.node.next = (struct node *) pFCB;
				    recHdr -> mlis.count++;
				    }
				else
				    {

				    /*
				    Field was already monitored. Put the MCB on the end of the chain
				    of this FCB because this won't require locking. Then, update the
				    event mask in the FCB.
				    */

				    pMCB -> fldCntl = pFCB;
				    prevMCB = pFCB -> firstChan;
				    while (prevMCB)
					{
					if (prevMCB -> nextMCB)
					    {
					    prevMCB = prevMCB -> nextMCB;
					    continue;
					    }
					prevMCB -> nextMCB = pMCB;
					break;
					}
				    if (pMCB -> dataType != AGNT_NOCONN)
					pFCB -> currMask |= pMCB -> chanMask;
				    }

				/*
				Get the status of this record's resource to complete the acknowlegement message
				and then send that back to the client. Endian doesn't matter for the resource
				state because it's a boolean.
				*/

				resGetCaAnswer(recHdr -> radr, &bldAck.currState);
				msgType = htons(CHAN_BLDACK);
				sendClient(currCCB, msgType, (char *) &bldAck, sizeof(bldAck), (char *) 0, 0, 
					    &myFds);
				}

			    else if (msgType == CHAN_CLOSE)
				{
				CH_CLOSE		chanClose;
				CH_CLOSEACK		closeAck;

				/*
				Channel Close. Get the handle so we know what he's talking about.
				*/

				if (clientRecv((char *) &chanClose, sizeof(chanClose), currCCB, &myFds))
				    {
				    if (debugging)
					printf("Channel close, handle = %X\n", chanClose.chHandle);

				    /*
				    If we ever send the workstation a message with a handle that he
				    doesn't know about, he'll tell us to close it. Just to be safe, make
				    sure that we actually have the handle he's talking about before we
				    try to delete it! Send an acknowledgement even if we didn't find
				    the channel just in case the workstation is holding on to resources
				    pending our close ack.
				    */

				    pMCB = currCCB -> firstChan;
				    while (pMCB)
					{
					if (pMCB -> handle == chanClose.chHandle)
					    break;
					pMCB = pMCB -> nextClient;
					}
				    if (pMCB)
					{
					closeMCB(pMCB);
					closeAck.chHandle = pMCB -> handle;
					msgType = htons(CHAN_CLOSEACK);
					sendClient(currCCB, msgType, (char *) &closeAck, sizeof(closeAck),
						    (char *) 0, 0, &myFds);
					}
				    else if (debugging)
					printf("\tREJECTED!\n");
				    }
				}

			    else if (msgType == CHAN_DIRREAD)
				{
				CH_DIRREAD		dirRead;
				CH_DRRDATA		drrResp;

				/*
				Direct Read. Get the requested data and then package it up for shipment.
				*/

				if (clientRecv((char *) &dirRead, sizeof(dirRead), currCCB, &myFds))
				    {
				    if (debugging)
					printf("Direct read for %X\n", dirRead.chHandle);
				    pMCB = (MCB *) dirRead.chHandle;
				    if (!validMCB(pMCB, currCCB, &myFds))
					continue;
				    pFCB = pMCB -> fldCntl;
				    recHdr = (struct dbCommon *) (pFCB -> dbAddr.precord);
				    drrResp.chDataLen = getAgentField(buff, ntohs(dirRead.chDataType),
							    ntohs(dirRead.chNumElem), recHdr -> stat, recHdr -> sevr, 
							    recHdr -> ack, &(recHdr -> lptm), &(pFCB -> dbAddr));
				    drrResp.chDataLen = htons(drrResp.chDataLen);

				    /*
				    Following four fields are transferred directly from the input to the
				    output buffer. This means we don't have to worry about byte order problems.
				    */

				    drrResp.chHandle = dirRead.chAppHandle;
				    drrResp.chDataType = dirRead.chDataType;
				    drrResp.chCount = dirRead.chNumElem;
				    drrResp.chCaHandle = dirRead.chCaHandle;

				    msgType = htons(CHAN_DRRDATA);
				    sendClient(currCCB, msgType, (char *) &drrResp, sizeof(drrResp), buff, 
						    ntohs(drrResp.chDataLen), &myFds);
				    }
				}
			    else if (msgType == CHAN_SESS_SET)
				{
				CH_SESS_SET		sessSet;

				/*
				A client has sent us a session handle. If it doesn't agree with the
				one that we previously had for him, then he crashed and we should close
				all of the channels that we had for him.
				*/

				if (clientRecv((char *) &sessSet, sizeof(sessSet), currCCB, &myFds))
				    {
				    if (debugging)
					printf("Session set.\n");
				    if (currCCB -> clientSession != sessSet.chSession)
					{
					currCCB -> clientSession = sessSet.chSession;
					while (currCCB -> firstChan)
					    closeMCB(currCCB -> firstChan);
					}
				    }
				}

			    else if (msgType == CHAN_TYPE)
				{
				CH_TYPE		chType;

				/*
				Set the data type, count, and mask in a user's MCB. Also, update the
				overall mask in the FCB. We don't need to lock the monitor list because
				we aren't affecting the linkage.
				*/

				if (clientRecv((char *) &chType, sizeof(chType), currCCB, &myFds))
				    {
				    if (debugging)
					printf("Channel type change for %X\n", chType.chHandle);
				    pMCB = (MCB *) chType.chHandle;
				    if (!validMCB(pMCB, currCCB, &myFds))
					continue;
				    pMCB -> dataType = ntohs(chType.chType);
				    pMCB -> numElem = ntohs(chType.chCount);
				    pMCB -> chanMask = ntohs(chType.chMask);
				    msgType = 0;
				    pFCB = pMCB -> fldCntl;
				    pMCB = pFCB -> firstChan;
				    while (pMCB)
					{
					if (pMCB -> dataType != AGNT_NOCONN)
					    msgType |= pMCB -> chanMask;
					pMCB = pMCB -> nextMCB;
					}
				    pFCB -> currMask = msgType;
				    }
				}

			    else if (msgType == CHAN_MONREQ)
				{
				CH_MONREQ		monReq;
				CH_SINGMON		singMon;

				/*
				Request for an immediate monitor report for a point.
				*/

				if (clientRecv((char *) &monReq, sizeof(monReq), currCCB, &myFds))
				    {
				    if (debugging)
					printf("Single monitor request for %X\n", monReq.chHandle);
				    pMCB = (MCB *) monReq.chHandle;
				    if (!validMCB(pMCB, currCCB, &myFds))
					continue;
				    singMon.chHandle = pMCB -> handle;
				    singMon.chCaHandle = monReq.chCaHandle;
				    singMon.chDataType = monReq.chType;
				    singMon.chMask = monReq.mask;
				    singMon.chCount =  htons(pMCB -> numElem);
				    recHdr = (struct dbCommon *) (pMCB -> fldCntl -> dbAddr.precord);
				    singMon.chDataLen = htons(getAgentField(rsrvOutBuff, ntohs(monReq.chType), 
								pMCB -> numElem, recHdr -> stat, recHdr -> sevr, 
								recHdr -> ack, &(recHdr -> lptm), 
								&(pMCB -> fldCntl -> dbAddr)));
				    msgType = htons(CHAN_SINGMON);
				    sendClient(pMCB -> client, msgType, (char *) &singMon, sizeof(singMon), rsrvOutBuff, 
						    ntohs(singMon.chDataLen), &myFds);
				    }
				}

			    else if (msgType == CHAN_DIRPUT)
				{
				CH_DIRPUT		dirPut;
				struct dbAddr	*pField;
				struct dbAddr	splicedField;
				char		splicedName[PVNAME_SZ + FLDNAME_SZ + 2];

				/*
				Direct Put.
				*/

				if (clientRecv((char *) &dirPut, sizeof(dirPut), currCCB, &myFds))
				    if (clientRecv(buff, ntohs(dirPut.chDataSize), currCCB, &myFds))
				    {
					if (debugging)
					    printf("Direct put for %X\n", dirPut.chHandle);
					pMCB = (MCB *) dirPut.chHandle;
					if (!validMCB(pMCB, currCCB, &myFds))
					    continue;

					/*
					At this point, the data type should be one of the fundamentals.
					Make sure of this by trimming off the embellishment bits. Then,
					we have to convert the "old" data type which the client gave
					us into the "new" data type expected by dbPutField(). Also,
					strictly speaking, we should do endian conversion on the data.
					However, we won't do that right now because the Motorola 680x0
					is already in network byte order. Worry about that later if John
					gets his way and we go to an Intel IOC (yuck!).

					Update:	7/27/95:	John got his way.
					*/

					pField = &(pMCB -> fldCntl -> dbAddr);
#ifdef PRIV_AUDIT
					GetPeerCN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
					if (GetChannelPriv (X509_dn, (char *)(pField -> precord)) != PRIV_FULL)
					    {
					    GetPeerDN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
					    /*
					    No priviledge for channel!
					    */
					    logMsg("CA priviledge error, %s not authorized for %s\n",
						    (int)X509_dn,(int)pField -> precord,0,0,0,0);
					    SecurityAlarm (X509_dn, "Channel write not allowed",(char *)pField->precord);
					    if (auditLog)
						{
						sprintf (updateMsg, "Channel write to %s.%4.4s not authorized",
							 pField->precord, pField->pfldDes);
						AuditAddEvent (auditLog, AUDIT_WRITE, AUDIT_FAIL, X509_dn,
							taskName(taskIdSelf()), updateMsg);
						}
					    continue;
					    }
#endif PRIV_AUDIT
					msgType = ntohs(dirPut.chDataType);
					if (msgType & AGNT_SPLICED)
					    {
					    strcpy(splicedName, (char *) (pField -> precord));
					    if (msgType & AGNT_SPL_STAT)
						strcat(splicedName, ".STAT");
					    else if (msgType & AGNT_SPL_SEVR)
						strcat(splicedName, ".SEVR");
					    else if (msgType & AGNT_SPL_ACK)
						strcat(splicedName, ".ACK");
					    pField = &splicedField;
					    dbNameToAddr(splicedName, pField);
					    }

					switch (msgType & 0xff)
					    {
					    case AGNT_SHORT:
					    case AGNT_ENUM:
						*(short *)buff = ntohs(*(short *)buff);
#ifdef PRIV_AUDIT
						sprintf (updateMsg, "%s.%4.4s %d", (char *) (pField -> precord), pField->pfldDes, *(short *)buff);
#endif PRIV_AUDIT
						break;
					    case AGNT_FLOAT:
						ntohf(buff, buff);
#ifdef PRIV_AUDIT
						sprintf (updateMsg, "%s.%4.4s %f", (char *) (pField -> precord), pField->pfldDes, *(float *)buff);
#endif PRIV_AUDIT
						break;
					    case AGNT_LONG:
						*(long *)buff = ntohl(*(long *)buff);
#ifdef PRIV_AUDIT
						sprintf (updateMsg, "%s.%4.4s %ld", (char *) (pField -> precord), pField->pfldDes, *(long *)buff);
#endif PRIV_AUDIT
						break;
					    case AGNT_DOUBLE:
						htond(buff, buff);
#ifdef PRIV_AUDIT
						sprintf (updateMsg, "%s.%4.4s %f", (char *) (pField -> precord), pField->pfldDes, *(double *)buff);
#endif PRIV_AUDIT
						break;
#ifdef PRIV_AUDIT
					    default:
						sprintf (updateMsg, "%s.%4.4s %s", (char *) (pField -> precord), pField->pfldDes, buff);
#endif PRIV_AUDIT

					    }
					msgType = mapOld2New(msgType & 0xff);

					/*
					Finally, stick the new data into the field.
					*/

					dbPutField(pField, msgType, buff, 1);
#ifdef PRIV_AUDIT
					if (auditLog)
					    {
					    AuditAddEvent (auditLog, AUDIT_WRITE, AUDIT_INFO, X509_dn,
						    taskName(taskIdSelf()), updateMsg);
					    }
#endif PRIV_AUDIT
				    }

				}
			    
			    else if (msgType == CHAN_RESMODE)
				{
				CH_RESMODE		resModeInq;
				RES_TABLE		*resEntry;

				/*
				Resource mode. Normally we send this message to a workstation, but he can also send
				it to us to inquire our current state of affairs with respect to this resource.
				Look up the name in the resource table, get the current state, and then send it back
				to the requester.
				*/

				if (clientRecv((char *) &resModeInq, sizeof(resModeInq), currCCB, &myFds))
				    {
				    if (debugging)
					printf("Resource inquiry for %s\n", resModeInq.chResName);
				    if (resGetByName(resModeInq.chResName, &resEntry) == RES_OK)
					{
					resGetCaAnswer(resEntry, &(resModeInq.chResStatus));
					resModeInq.chResStatus = htons(resModeInq.chResStatus);
					msgType = htons(msgType);
					sendClient(currCCB, msgType, (char *) &resModeInq, sizeof(resModeInq), 
						(char *) 0, 0, &myFds);
					}
				    }
				}

			    else if (msgType == CHAN_RESINQ)
				{
				CH_RESMODE		resModeReply;
				RES_TABLE		*resEntry;

				/*
				Resource mode inquiry. The workstation can send this message to us to ask about
				the state of all of our resources. He would usually do this when a connection
				is resumed after a disruption of the communications facility.
				*/

				if (debugging)
				    printf("Query for all resource states\n");
				msgType = htons(CHAN_RESMODE);
				resEntry = resTableHead;
				while (resEntry)
				    {
				    strcpy(resModeReply.chResName, resEntry -> resourceName);
				    resModeReply.chResStatus = resEntry -> caAnswer;
				    resModeReply.chResStatus = htons(resEntry -> caAnswer);
				    sendClient(currCCB, msgType, (char *) &resModeReply, sizeof(resModeReply),
					    (char *) 0, 0, &myFds);
				    resEntry = resEntry -> link;
				    }
				}

			    else if (msgType == CHAN_QBUILD)
				{
				CH_QBLD		chQbld;
				char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
				struct dbAddr	dbAddr;
				CH_QBLDACK		chQbldAck;
				FCB			*prevFCB;
				MCB			*prevMCB;

				/*
				This transaction is a quick channel build. We only respond with
				the identifying information and the status, severity, and value of
				the point. All the rest of the information comes out of the database
				on the workstation. The purpose of this transaction type is to
				reduce the number of bytes on low speed lines.
				*/

				if (!clientRecv((char *) &chQbld, sizeof(chQbld), currCCB, &myFds))
				    continue;
				if (!clientRecv(chanName, ntohs(chQbld.chNameLen), currCCB, &myFds))
				    continue;

				/*
				Get the address and all other information for this channel. Fill out all
				the static information in the reply.
				*/
#ifdef PRIV_AUDIT
				GetPeerCN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				if (GetChannelPriv (X509_dn, chanName) == PRIV_NOACCESS)
				    {
				    GetPeerDN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				    /*
				    No priviledge for channel!
				    */
				    logMsg("CA priviledge error, %s not authorized for %s\n",
					    (int)X509_dn,(int)chanName,0,0,0,0);
				    SecurityAlarm (X509_dn, "Channel access not allowed", chanName);
				    if (auditLog)
					{
					sprintf (updateMsg, "Channel connect to %s not authorized", chanName);
					AuditAddEvent (auditLog, AUDIT_SESSION, AUDIT_FAIL, X509_dn,
						taskName(taskIdSelf()), updateMsg);
					}
				    continue;
				    }
#endif PRIV_AUDIT
				dbNameToAddr(chanName, &dbAddr);
				recHdr = (struct dbCommon *) dbAddr.precord;
				if ((int) dbAddr.precord < 0)
				    continue;
				if ((int) dbAddr.pfield < 0)
				    continue;
				chQbldAck.chHandle = chQbld.chHandle;
				chQbldAck.chCAHandle = chQbld.chCaHandle;
				chQbldAck.chDataLen = 0;

				/*
				It is also possible to get duplicate requests as a part of error recovery.
				If so, just return the original handle in an acknowledgement.
				*/

				pMCB = currCCB -> firstChan;
				while (pMCB)
				    {
				    if (dbAddr.precord == pMCB -> fldCntl -> dbAddr.precord &&
					    dbAddr.pfield == pMCB -> fldCntl -> dbAddr.pfield &&
					    chQbld.chHandle == pMCB -> handle)
					break;
				    pMCB = pMCB -> nextClient;
				    }
			    
				/*
				Allocate a new MCB if we aren't piggybacking on an old one. Fill out
				the information that we know so far.
				*/

				if (!pMCB)
				    {
				    pMCB = newMCB();
				    pMCB -> handle = chQbld.chHandle;
				    pMCB -> dataType = mapNew2Old(dbAddr.dbr_field_type) | AGNT_STAT;
				    pMCB -> numElem = dbAddr.no_elements;
				    pMCB -> chanMask = AGNT_VALUE | AGNT_ALARM;
				    if (!dbAddr.no_elements)
					pMCB -> numElem = 1;
				    pMCB -> client = currCCB;

				    /*
				    This MCB is new, so put it at the head to the chain for this client.
				    */

				    pMCB -> nextClient = currCCB -> firstChan;
				    currCCB -> firstChan = pMCB;
				    if (pMCB -> nextClient)
					pMCB -> nextClient -> prevClient = pMCB;
				    
				    /*
				    March along the monitor chan to see if this field is already montored
				    elsewhere, regardless of data type.
				    */

				    pFCB = (FCB *) (recHdr -> mlis.node.next);
				    prevFCB = pFCB;
				    while (pFCB)
					{
					if (pFCB -> dbAddr.pfield == dbAddr.pfield)
					    break;
					prevFCB = pFCB;
					pFCB = (FCB *) (pFCB -> nextFCB);
					}

				    if (!pFCB)
					{

					/*
					Must create a new FCB. Fill it out and put it at the end of the
					monitor queue, if there is one. By putting it at the end instead
					of the head, we don't have to lock the list. Dangerous, but fun.
					*/

					pFCB = newFCB();
					pFCB -> dbAddr = dbAddr;
					pFCB -> firstChan = pMCB;
					pFCB -> currMask = (AGNT_VALUE | AGNT_ALARM);
					pMCB -> fldCntl = pFCB;
					if (prevFCB)
					    {
					    pFCB -> prevFCB = prevFCB;
					    prevFCB -> nextFCB = pFCB;
					    }
					else
					    recHdr -> mlis.node.next = (struct node *) pFCB;
					recHdr -> mlis.count++;
					}
				    
				    else
					{
					/*
					Field was already montiored. Put the MCB on the end of the
					chain of this FCB because this won't require locking. Then, update
					the event mask in the FCB.
					*/

					pMCB -> fldCntl = pFCB;
					prevMCB = pFCB -> firstChan;
					while (prevMCB)
					    {
					    if (prevMCB -> nextMCB)
						{
						prevMCB = prevMCB -> nextMCB;
						continue;
						}
					    prevMCB -> nextMCB = pMCB;
					    break;
					    }
					pFCB -> currMask |= (AGNT_VALUE | AGNT_ALARM);
					}
				    }
				chQbldAck.chIocHandle = (int) pMCB;
				    
				/*
				Now get the state of the record's resource to complete the acknowledgement
				message and send it on to the client. Endian doesn't matter for the resource
				state because it's a boolean.
				*/

				resGetCaAnswer(recHdr -> radr, &chQbldAck.currState);
				msgType = htons(CHAN_QBLDACK);

				/*
				If we don't own the resource, send the ack and we are done.
				*/

				if (!chQbldAck.currState)
				    {
			    	    sendClient(currCCB, msgType, (char *) &chQbldAck, sizeof(chQbldAck), (char *) 0, 0, 
					&myFds);
				    continue;
				    }

				/*
				Do a read for the STATUS variant of the native field type.
				*/

				chQbldAck.chDataLen = htons(getAgentField(buff, pMCB -> dataType, pMCB -> numElem,
					recHdr -> stat, recHdr -> sevr, recHdr -> ack, &(recHdr -> lptm), &dbAddr));
				msgType = htons(CHAN_QBLDACK);
				sendClient(currCCB, msgType, (char *) &chQbldAck, sizeof(chQbldAck), buff,
					ntohs(chQbldAck.chDataLen), &myFds);
				}

			    else if (msgType == CHAN_BLDRDMON)
				{
				CH_BLDRDMON		chBldRdMon;
				char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
				struct dbAddr	dbAddr;
				CH_BLDACK		bldAck;
				CH_DRRDATA		drrResp;
				CH_SINGMON		singMon;
				FCB			*prevFCB;
				MCB			*prevMCB;

				/*
				This transaction is really three for one. We create a channel, do
				a direct read for the control variant of the native data type with
				the native count, and then generate a single monitor report for
				the status variant of the native data type with the native count.
				The purpose of this transaction is to be able to pop up graphics
				in dm without waiting for at least two and possibly three round
				trips between the workstation and the IOC.

				Start processing by getting the message header and the channel name.
				A lot of this code is very similar to the processing for the
				CHAN_BUILD, CHAN_DIRREAD, and CHAN_MONREQ messages. However, it
				wasn't close enough to warrant making those parts of the overall
				code modules to be called from both here and there.
				*/

				if (!clientRecv((char *) &chBldRdMon, sizeof(chBldRdMon), currCCB, &myFds))
				    continue;
				if (!clientRecv(chanName, ntohs(chBldRdMon.chNameLen), currCCB, &myFds))
				    continue;

				/*
				Get the address and all other information for this channel. Fill out all
				the static information in the reply.
				*/

				if (debugging)
				    printf("BLDRDMON for %s\n", chanName);
#ifdef PRIV_AUDIT
				GetPeerCN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				if (GetChannelPriv (X509_dn, chanName) == PRIV_NOACCESS)
				    {
				    GetPeerDN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				    /*
				    No priviledge for channel!
				    */
				    logMsg("CA priviledge error, %s not authorized for %s\n",
					    (int)X509_dn,(int)chanName,0,0,0,0);
				    SecurityAlarm (X509_dn, "Channel connect not allowed", chanName);
				    if (auditLog)
					{
					sprintf (updateMsg, "Channel connect to %s not authorized", chanName);
					AuditAddEvent (auditLog, AUDIT_SESSION, AUDIT_FAIL, X509_dn,
						taskName(taskIdSelf()), updateMsg);
					}
				    continue;
				    }
#endif PRIV_AUDIT
				dbNameToAddr(chanName, &dbAddr);
				recHdr = (struct dbCommon *) dbAddr.precord;
				if ((int) dbAddr.precord < 0)
				    continue;
				if ((int) dbAddr.pfield < 0)
				    continue;
				bldAck.chHandle = chBldRdMon.chHandle;
				bldAck.chDataType = mapNew2Old(dbAddr.dbr_field_type);
				bldAck.chDataType = htons(bldAck.chDataType);
				bldAck.chElementCount = htons(dbAddr.no_elements);
				if (!bldAck.chElementCount)
				    bldAck.chElementCount = htons(1);
				
				/*
				It is also possible to get duplicate requests as a part of error recovery.
				If so, just return the original handle in an acknowledgement.
				*/

				pMCB = currCCB -> firstChan;
				while (pMCB)
				    {
				    if (dbAddr.precord == pMCB -> fldCntl -> dbAddr.precord &&
					    dbAddr.pfield == pMCB -> fldCntl -> dbAddr.pfield &&
					    chBldRdMon.chHandle == pMCB -> handle)
					break;
				    pMCB = pMCB -> nextClient;
				    }
			    
				/*
				Allocate a new MCB if we aren't piggybacking on an old one. Fill out
				the information that we know so far.
				*/

				if (!pMCB)
				    {
				    pMCB = newMCB();
				    pMCB -> handle = chBldRdMon.chHandle;
				    pMCB -> dataType = mapNew2Old(dbAddr.dbr_field_type) | AGNT_STAT;
				    pMCB -> numElem = dbAddr.no_elements;
				    pMCB -> chanMask = AGNT_VALUE | AGNT_ALARM;
				    if (!dbAddr.no_elements)
					pMCB -> numElem = 1;
				    pMCB -> client = currCCB;

				    /*
				    This MCB is new, so put it at the head to the chain for this client.
				    */

				    pMCB -> nextClient = currCCB -> firstChan;
				    currCCB -> firstChan = pMCB;
				    if (pMCB -> nextClient)
					pMCB -> nextClient -> prevClient = pMCB;
				    
				    /*
				    March along the monitor chan to see if this field is already montored
				    elsewhere, regardless of data type.
				    */

				    pFCB = (FCB *) (recHdr -> mlis.node.next);
				    prevFCB = pFCB;
				    while (pFCB)
					{
					if (pFCB -> dbAddr.pfield == dbAddr.pfield)
					    break;
					prevFCB = pFCB;
					pFCB = (FCB *) (pFCB -> nextFCB);
					}

				    if (!pFCB)
					{

					/*
					Must create a new FCB. Fill it out and put it at the end of the
					monitor queue, if there is one. By putting it at the end instead
					of the head, we don't have to lock the list. Dangerous, but fun.
					*/

					pFCB = newFCB();
					pFCB -> dbAddr = dbAddr;
					pFCB -> firstChan = pMCB;
					pFCB -> currMask = (AGNT_VALUE | AGNT_ALARM);
					pMCB -> fldCntl = pFCB;
					if (prevFCB)
					    {
					    pFCB -> prevFCB = prevFCB;
					    prevFCB -> nextFCB = pFCB;
					    }
					else
					    recHdr -> mlis.node.next = (struct node *) pFCB;
					recHdr -> mlis.count++;
					}
				    
				    else
					{
					/*
					Field was already montiored. Put the MCB on the end of the
					chain of this FCB because this won't require locking. Then, update
					the event mask in the FCB.
					*/

					pMCB -> fldCntl = pFCB;
					prevMCB = pFCB -> firstChan;
					while (prevMCB)
					    {
					    if (prevMCB -> nextMCB)
						{
						prevMCB = prevMCB -> nextMCB;
						continue;
						}
					    prevMCB -> nextMCB = pMCB;
					    break;
					    }
					pFCB -> currMask |= (AGNT_VALUE | AGNT_ALARM);
					}
				    }
				bldAck.chIocHandle = (int) pMCB;
				    
				/*
				Now get the state of the record's resource to complete the acknowledgement
				message and send it on to the client. Endian doesn't matter for the resource
				state because it's a boolean.
				*/

				resGetCaAnswer(recHdr -> radr, &bldAck.currState);
				msgType = htons(CHAN_BLDACK);
				sendClient(currCCB, msgType, (char *) &bldAck, sizeof(bldAck), (char *) 0, 0, 
					&myFds);

				/*
				If we don't own the resource, we are done.
				*/

				if (!bldAck.currState)
				    continue;

				/*
				We own it. Now do a read for the control variant of the native data type and
				the native count.
				*/

				msgType = mapNew2Old(dbAddr.dbr_field_type);
				if (msgType == AGNT_STRING)
				    msgType |= (AGNT_STAT | AGNT_CNTRL);
				else
				    msgType |= (AGNT_STAT | AGNT_GRAPH | AGNT_CNTRL);
				drrResp.chDataLen = getAgentField(buff, msgType, pMCB -> numElem,
					    recHdr -> stat, recHdr -> sevr, recHdr -> ack, &(recHdr -> lptm), &dbAddr);
				drrResp.chDataLen = htons(drrResp.chDataLen);
				drrResp.chHandle = chBldRdMon.chAppHandle;
				drrResp.chDataType = htons(msgType);
				drrResp.chCount = htons(pMCB -> numElem);
				drrResp.chCaHandle = chBldRdMon.chCaHandle;

				msgType = htons(CHAN_DRRDATA);
				sendClient(currCCB, msgType, (char *) &drrResp, sizeof(drrResp), buff,
					ntohs(drrResp.chDataLen), &myFds);
				
				/*
				Now generate a single monitor report for the status variant of the native data
				type and the native count.
				*/

				singMon.chHandle = chBldRdMon.chHandle;
				singMon.chCaHandle = chBldRdMon.chCaHandle;
				singMon.chDataType = htons(pMCB -> dataType);
				singMon.chCount = htons(pMCB -> numElem);
				singMon.chDataLen = htons(getAgentField(buff, pMCB -> dataType, pMCB -> numElem,
					recHdr -> stat, recHdr -> sevr, recHdr -> ack,  &(recHdr -> lptm), &dbAddr));
				msgType = htons(CHAN_SINGMON);
				sendClient(currCCB, msgType, (char *) &singMon, sizeof(singMon), buff,
					ntohs(singMon.chDataLen), &myFds);

				}

			    else if (msgType == CHAN_HANDLERS)
				{
				CH_HANDLERS		handlerMsg;
				int			oldTrendHandler;
				int			oldAlarmHandler;

				/*
				A workstation is sending us this message to update (or confirm)
				his handler status for trends and alarms.
				*/


				if (!clientRecv((char *) &handlerMsg, sizeof(handlerMsg), currCCB, &myFds))
				    continue;
				if (debugging)
				    printf("Handler mode set\n");
				oldTrendHandler = currCCB -> trendHandler;
				oldAlarmHandler = currCCB -> alarmHandler;
				currCCB -> trendHandler = handlerMsg.chTrendHandler;
				currCCB -> alarmHandler = handlerMsg.chAlarmHandler;
				currCCB -> handlerMsgRcvd = 1;

				/*
				If adding a new alarm handler, dump the alarm Queue at him and then start
				the database traversals.
				*/

				if (handlerMsg.chAlarmHandler && !oldAlarmHandler)
				    {
				    ALARM_QUE	*currAlarm;

				    currAlarm = alarmHead;
				    msgType = htons(CHAN_ALARM);
				    while (currAlarm)
					{
	    				sendClient(currCCB, msgType, (char *) &(currAlarm -> alarmMsg), 
					    sizeof(CH_ALARM), (char *) &(currAlarm -> alarmData), 
					    ntohs(currAlarm -> alarmMsg.alarmDataLen), &myFds);
if (debugging)
	printf("ALARM QUEUE SEND %s: ser = %d, ts = %d.%d\n", currAlarm -> alarmData, currAlarm ->
		    alarmMsg.alarmSerial, currAlarm -> alarmMsg.alarmTime.secPastEpoch, currAlarm -> 
		    alarmMsg.alarmTime.nsec);
					currAlarm = currAlarm -> nextAlq;
					}

				    if (!currCCB -> alarmTravPending)
					{
					currCCB -> alarmTravPending = 1;
			    		taskSpawn(TRAVERSE_NAME, TRAVERSE_PRI, TRAVERSE_OPT, TRAVERSE_STACK,
						(FUNCPTR)traverseDb, 2, (int)currCCB, 0, 0, 0, 0, 0, 0, 0, 0);
					}
				    }

				/*
				If adding a new trend handler, spoon feed him some data to get started.
				*/

				if (handlerMsg.chTrendHandler && !oldTrendHandler)
				    {
				    if (!currCCB -> trendTravPending)
					{
					currCCB -> trendTravPending = 1;
					taskSpawn(TRAVERSE_NAME, TRAVERSE_PRI, TRAVERSE_OPT, TRAVERSE_STACK,
						(FUNCPTR)traverseDb, 1, (int)currCCB, 0, 0, 0, 0, 0, 0, 0, 0);
					}
				    }
				}

			    else if (msgType == CHAN_GETPARAMS)
				{
				CH_GETPARAMS	paramReq;
				CH_PARMREPLY	paramReply;
				char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
				struct dbAddr	dbAddr;
				double		hopr;
				double		lopr;
				double		drvh;
				double		drvl;
				long		prec;
				struct dbCommon	*recHdr;
				struct rset		*prset;
				char		egu[18];
				short		dataBytes = 0;
				char		*outPtr = rsrvOutBuff;
				short		tempShort;
				long		tempLong;
				float		tempFloat;
				double		tempDouble;

				/*
				Request for parameters. Get the header and the channel name.
				*/

				if (!clientRecv((char *) &paramReq, sizeof(paramReq), currCCB, &myFds))
				    continue;
				if (!clientRecv(chanName, ntohs(paramReq.chNameLen), currCCB, &myFds))
				    continue;

				/*
				Locate the record. If we can't find it, drop the request on the floor.
				*/

				if (debugging)
				    printf("Parameter fetch for %s\n", chanName);
#ifdef PRIV_AUDIT
				GetPeerCN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				if (GetChannelPriv (X509_dn, chanName) == PRIV_NOACCESS)
				    {
				    GetPeerDN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				    /*
				    No priviledge for channel!
				    */
				    logMsg("CA priviledge error, %s not authorized for %s\n",
					    (int)X509_dn,(int)chanName,0,0,0,0);
				    SecurityAlarm (X509_dn, "Channel access not allowed", chanName);
				    if (auditLog)
					{
					sprintf (updateMsg, "Channel connect to %s not authorized", chanName);
					AuditAddEvent (auditLog, AUDIT_SESSION, AUDIT_FAIL, X509_dn,
						taskName(taskIdSelf()), updateMsg);
					}
				    continue;
				    }
#endif PRIV_AUDIT
				dbNameToAddr(chanName, &dbAddr);
				if ((int) dbAddr.precord < 0)
				    continue;
				if ((int) dbAddr.pfield < 0)
				    continue;

				/*
				Get all the paramters that are necessary as doubles. We can convert them to
				the requested data type later.
				*/

				hopr = 0;
				lopr = 0;
				drvh = 0;
				drvl = 0;
				prec = 0;
				egu[0] = 0;
				recHdr = (struct dbCommon *) dbAddr.precord;
				prset = GET_PRSET(dbAddr.record_type);
				if (prset)
				    {
				    if (prset -> get_units)
					{
					(*(prset -> get_units))(&dbAddr, egu);
					egu[7] = 0;
					}
				    if (prset -> get_graphic_double)
					{
					struct dbr_grDouble		grDouble;

					(*(prset -> get_graphic_double))(&dbAddr, &grDouble);
					hopr = grDouble.upper_disp_limit;
					lopr = grDouble.lower_disp_limit;
					}
				    if (prset -> get_control_double)
					{
					struct dbr_ctrlDouble	ctrlDouble;

					(*(prset -> get_control_double))(&dbAddr, &ctrlDouble);
					drvh = ctrlDouble.upper_ctrl_limit;
					drvl = ctrlDouble.lower_ctrl_limit;
					}
				    }

				paramReq.chParmType = ntohs(paramReq.chParmType);
				if (paramReq.chParmType == DBR_SHORT_PARAM)
				    {
				    tempShort = htons(((short) hopr));
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes = sizeof(short);
				    tempShort = htons(((short) lopr));
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = htons(((short) drvh));
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = htons(((short) drvl));
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = mapNew2Old(dbAddr.field_type);
				    tempShort = htons(tempShort);
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = htons(dbAddr.no_elements);
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = htons(((short) prec));
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    }

				else if (paramReq.chParmType == DBR_ENUM_PARAM)
				    {
				    struct dbr_enumStrs	enumStrs;
				    int			index;


				    bzero((char *) &enumStrs, sizeof(enumStrs));
				    if (prset)
					{
					if (prset -> get_enum_strs)
					    (*(prset -> get_enum_strs))(&dbAddr, &enumStrs);
					}
				    *(short *) outPtr = htons(enumStrs.no_str);
				    outPtr += sizeof(short);
				    dataBytes = sizeof(short);
				    for (index = 0; index < enumStrs.no_str; index++)
					{
					strcpy(outPtr, enumStrs.strs[index]);
					outPtr += (strlen(enumStrs.strs[index]) + 1);
					dataBytes += (strlen(enumStrs.strs[index]) + 1);
					}
				    }

				else if (paramReq.chParmType == DBR_STRING_PARAM)
				    {

				    /*
				    Nothing to do except the common stuff
				    */

				    }

				else if (paramReq.chParmType == DBR_FLOAT_PARAM)
				    {
				    tempFloat = hopr;	/* Watch endian later */
				    htonf(tempFloat, tempFloat);
				    bcopy((char *) &tempFloat, outPtr, sizeof(float));
				    outPtr += sizeof(float);
				    dataBytes = sizeof(float);
				    tempFloat = lopr;	/* Watch endian later */
				    htonf(tempFloat, tempFloat);
				    bcopy((char *) &tempFloat, outPtr, sizeof(float));
				    outPtr += sizeof(float);
				    dataBytes += sizeof(float);
				    tempFloat = drvh;	/* Watch endian later */
				    htonf(tempFloat, tempFloat);
				    bcopy((char *) &tempFloat, outPtr, sizeof(float));
				    outPtr += sizeof(float);
				    dataBytes += sizeof(float);
				    tempFloat = drvl;	/* Watch endian later */
				    htonf(tempFloat, tempFloat);
				    bcopy((char *) &tempFloat, outPtr, sizeof(float));
				    outPtr += sizeof(float);
				    dataBytes += sizeof(float);
				    tempShort = mapNew2Old(dbAddr.field_type);
				    tempShort = htons(tempShort);
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = htons(dbAddr.no_elements);
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = htons(((short) prec));
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    }

				else if (paramReq.chParmType == DBR_LONG_PARAM)
				    {
				    tempLong = htonl(((long) hopr));
				    bcopy((char *) &tempLong, outPtr, sizeof(long));
				    outPtr += sizeof(long);
				    dataBytes = sizeof(long);
				    tempLong = htonl(((long) lopr));
				    bcopy((char *) &tempLong, outPtr, sizeof(long));
				    outPtr += sizeof(long);
				    dataBytes += sizeof(long);
				    tempLong = htonl(((long) drvh));
				    bcopy((char *) &tempLong, outPtr, sizeof(long));
				    outPtr += sizeof(long);
				    dataBytes += sizeof(long);
				    tempLong = htonl(((long) drvl));
				    bcopy((char *) &tempLong, outPtr, sizeof(long));
				    outPtr += sizeof(long);
				    dataBytes += sizeof(long);
				    tempShort = mapNew2Old(dbAddr.field_type);
				    tempShort = htons(tempShort);
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = htons(dbAddr.no_elements);
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = htons(((short) prec));
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    }

				else if (paramReq.chParmType == DBR_DOUBLE_PARAM)
				    {
				    tempDouble = hopr;	/* Watch endian later */
				    htond(tempDouble, tempDouble);
				    bcopy ((char *) &tempDouble, outPtr, sizeof(double));
				    outPtr += sizeof(double);
				    dataBytes = sizeof(double);
				    tempDouble = lopr;	/* Watch endian later */
				    htond(tempDouble, tempDouble);
				    bcopy((char *) &tempDouble, outPtr, sizeof(double));
				    outPtr += sizeof(double);
				    dataBytes += sizeof(double);
				    tempDouble = drvh;	/* Watch endian later */
				    htond(tempDouble, tempDouble);
				    bcopy((char *) &tempDouble, outPtr, sizeof(double));
				    outPtr += sizeof(double);
				    dataBytes += sizeof(double);
				    htond(tempDouble, tempDouble);
				    tempDouble = drvl;	/* Watch endian later */
				    bcopy((char *) &tempDouble, outPtr, sizeof(double));
				    outPtr += sizeof(double);
				    dataBytes += sizeof(double);
				    tempShort = mapNew2Old(dbAddr.field_type);
				    tempShort = htons(tempShort);
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = htons(dbAddr.no_elements);
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    tempShort = htons(((short) prec));
				    bcopy((char *) &tempShort, outPtr, sizeof(short));
				    outPtr += sizeof(short);
				    dataBytes += sizeof(short);
				    }

				/*
				Now put in all of the fields that are common to all types.
				*/

				strcpy(outPtr, chanName);
				outPtr += (strlen(chanName) + 1);
				dataBytes += (strlen(chanName) + 1);
				strcpy(outPtr, egu);
				outPtr += (strlen(egu) + 1);
				dataBytes += (strlen(egu) + 1);
				strcpy(outPtr, recHdr -> resn);
				outPtr += (strlen(recHdr -> resn) + 1);
				dataBytes += (strlen(recHdr -> resn) + 1);
				strcpy(outPtr, recHdr -> desc);
				dataBytes += (strlen(recHdr -> desc) + 1);

				/*
				Send it on the to requester.
				*/

				paramReply.chParmType = htons(paramReq.chParmType);
				paramReply.chDataLen = htons(dataBytes);
				paramReply.chArgument = paramReq.chArgument; /* DON'T FLIP ENDIANS! */
				paramReply.chACB = paramReq.chACB;
				msgType = htons(CHAN_PARMREPLY);
				sendClient(currCCB, msgType, (char *) &paramReply, sizeof(paramReply), rsrvOutBuff,
					dataBytes, &myFds);
				}

			    else if (msgType == CHAN_ALARM_RECV)
				{
				ALARM_QUE		*currAlq;
				ALARM_QUE		*prevAlq;
				struct
				    {
			    	    CH_ALARM_RECV	alarmRecv;
			    	    char		pvName[PVNAME_SZ + FLDNAME_SZ + 2];
				    } rMsg;

				/*
				A workstation alarm manager is telling us that he has received an
				alarm. Delete it from our queue.
				*/

				if (!clientRecv((char *) &rMsg.alarmRecv, sizeof(CH_ALARM_RECV), currCCB, &myFds))
				    continue;
				if (!clientRecv(rMsg.pvName, ntohs(rMsg.alarmRecv.alarmNameLen), currCCB, &myFds))
				    continue;

				/*
				Relay this message to our redundant partner.
				*/

				pvdGetAddr(rMsg.pvName, (char **)&recHdr);
				msgType = htons(CHAN_ALARM_RECV);
				sendRedAlarms(((RES_TABLE *)(recHdr -> radr)) -> redAlarmFd, msgType, (char *) &rMsg,
					sizeof(CH_ALARM_RECV) + ntohs(rMsg.alarmRecv.alarmNameLen));
				
				if (debugging)
			    	    printf("ALARM RECEIPT %s: ser = %d, ts = %d.%d ...", rMsg.pvName, 
					    rMsg.alarmRecv.alarmSerial, ntohl(rMsg.alarmRecv.alarmTime.secPastEpoch), 
						    ntohl(rMsg.alarmRecv.alarmTime.nsec));
				currAlq = alarmHead;
				prevAlq = 0;
				while (currAlq)
				    {
				    if (currAlq -> alarmMsg.alarmSerial == rMsg.alarmRecv.alarmSerial)
					if (ntohl(rMsg.alarmRecv.alarmTime.secPastEpoch) 
						== currAlq -> alarmMsg.alarmTime.secPastEpoch)
					    if (ntohl(rMsg.alarmRecv.alarmTime.nsec) == currAlq -> alarmMsg.alarmTime.nsec)
				    		if (strcmp(rMsg.pvName, currAlq -> alarmData) == 0)
						    break;
				    prevAlq = currAlq;
				    currAlq = currAlq -> nextAlq;
				    }
			    
				/*
				It may be possible that some other workstation already acknowledged this
				alarm. If we can't find the serial number in the queue, then silently
				ignore it.
				*/

				if (!currAlq)
				    {
				    if (debugging)
					printf("NOT FOUND!\n");
				    continue;
				    }
				
				if (debugging)
				    printf("FOUND!\n");

				/*
				Take this one out of the queue, updating our head and tail pointers. Put
				this one on the free list.
				*/

				if (prevAlq)
				    prevAlq -> nextAlq = currAlq -> nextAlq;
				else
				    alarmHead = currAlq -> nextAlq;
				if (alarmTail == currAlq)
				    {
				    if (prevAlq)
					alarmTail = prevAlq;
				    else
					alarmTail = alarmHead;
				    }
				currAlq -> nextAlq = freeAlarms;
				freeAlarms = currAlq;
				}

			    else if (msgType == CHAN_ALARM_ACK)
				{
				CH_ALARM_ACK	alarmAck;
				char		pv[PVNAME_SZ + FLDNAME_SZ + 2];
				char		*dot;
				struct dbAddr	dbAddr;
				struct dbCommon	*recHdr;
				short		ackValue = 0;

				/*
				An operator has acknowledged an alarm. Turn the ACK field from 1 to zero.
				*/

				if (!clientRecv((char *) &alarmAck, sizeof(alarmAck), currCCB, &myFds))
				    continue;
				if (!clientRecv(pv, ntohs(alarmAck.alarmNameLen), currCCB, &myFds))
				    continue;
			    
				/*
				If there is a period in the point name, smash it.
				*/

				if (debugging)
				    printf("Operator alarm ack for %s", pv);
				if ((dot = (char *) strchr(pv, '.')))
				    *dot = 0;
				strcat(pv, ".ACK");
#ifdef PRIV_AUDIT
				GetPeerCN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				if (GetChannelPriv (X509_dn, pv) != PRIV_FULL)
				    {
				    GetPeerDN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				    /*
				    No priviledge for channel!
				    */
				    logMsg("CA priviledge error, %s not authorized for %s\n",
					    (int)X509_dn,(int)pv,0,0,0,0);
				    SecurityAlarm (X509_dn, "Channel ACK  not allowed", pv);
				    if (auditLog)
					{
					sprintf (updateMsg, "Channel ack for %s not authorized", pv);
					AuditAddEvent (auditLog, AUDIT_SESSION, AUDIT_FAIL, X509_dn,
						taskName(taskIdSelf()), updateMsg);
					}
				    continue;
				    }
#endif PRIV_AUDIT
				dbNameToAddr(pv, &dbAddr);
				if ((int) dbAddr.precord < 0)
				    continue;
				if ((int) dbAddr.pfield < 0)
				    continue;
				recHdr = (struct dbCommon *) dbAddr.precord;
				alarmAck.alarmTime.secPastEpoch = ntohl(alarmAck.alarmTime.secPastEpoch);
				alarmAck.alarmTime.nsec = ntohl(alarmAck.alarmTime.nsec);

				/*
				We should drop this ACK on the floor if the point's current alarm state is
				not NO_ALARM and the time in the record is after the time in the ACK msg.
				However, do send the alarm manager that sent us the ACK the most recent
				alarm so that he can get in sync with us.
				*/

				if (recHdr -> stat)
				    {
				    if ((recHdr -> latm.secPastEpoch > alarmAck.alarmTime.secPastEpoch)
					|| ((recHdr -> latm.secPastEpoch == alarmAck.alarmTime.secPastEpoch)
					    && (recHdr -> latm.nsec > alarmAck.alarmTime.nsec)))
					{
		    			CH_ALARM	alarmData;
					struct fldDes	*valFld;

		    			/*
		    			Create an unsolicited alarm message for the alarm handlers in the system.
		    			*/

					if (debugging)
					    printf("..INVALID TIME!");
		    			alarmData.alarmTime.secPastEpoch = htonl(recHdr -> latm.secPastEpoch);
		    			alarmData.alarmTime.nsec = htonl(recHdr -> latm.nsec);
					alarmData.alarmSerial = 0;
		    			alarmData.alarmStat = htons(recHdr -> stat);
		    			alarmData.alarmSevr = htons(recHdr -> sevr);
		    			alarmData.alarmAck = htons(recHdr ->  ack);
		    			alarmData.alarmEnabled = htons(recHdr -> aenb);
		    			alarmData.alarmOnscan = htons(recHdr -> oscn);
					valFld = (struct fldDes *) pvdGetFld(dbAddr.record_type, "LALM");
					if (!valFld)
					    {
					    if (debugging)
						printf("...Can't correct because no LALM field!\n");
					    continue;
					    }
		    			alarmData.alarmDataLen = htons(getAlarmStream(recHdr, dbAddr.record_type,
						((char *) recHdr) + valFld -> offset, rsrvOutBuff));
		    			sendNewAlarm(&alarmData, rsrvOutBuff, currCCB);
					if (debugging)
					    printf("...Sent back to alarm mgr!\n");
					continue;
					}
				    }
				dbPutField(&dbAddr, mapOld2New(AGNT_SHORT), (caddr_t)&ackValue, 1);
				if (debugging)
				    printf("Acknowledged OK\n");
				}

			    else if (msgType == CHAN_ALARM_CTL)
				{
				CH_ALARM_CTL	alarmCtl;
				char		pv[PVNAME_SZ + FLDNAME_SZ + 2];
				char		*dot;
				struct dbAddr	dbAddr;
				short		alarmState = 0;

				/*
				Turn alarming on or off for a single point.
				*/

				if (!clientRecv((char *) &alarmCtl, sizeof(alarmCtl), currCCB, &myFds))
				    continue;
				if (!clientRecv(pv, ntohs(alarmCtl.alarmNameLen), currCCB, &myFds))
				    continue;
				
				/*
				If there is a period in the name, smash it.
				*/

				if ((dot = (char *) strchr(pv, '.')))
				    *dot = 0;
				strcat(pv, ".AENB");
#ifdef PRIV_AUDIT
				GetPeerCN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				if (GetChannelPriv (X509_dn, pv) != PRIV_FULL)
				    {
				    GetPeerDN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
				    /*
				    No priviledge for channel!
				    */
				    logMsg("CA priviledge error, %s not authorized for %s\n",
					    (int)X509_dn,(int)pv,0,0,0,0);
				    SecurityAlarm (X509_dn, "Channel AENB write not allowed", pv);
				    if (auditLog)
					{
					sprintf (updateMsg, "Channel alarm enable %s not authorized", pv);
					AuditAddEvent (auditLog, AUDIT_SESSION, AUDIT_FAIL, X509_dn,
						taskName(taskIdSelf()), updateMsg);
					}
				    continue;
				    }
#endif PRIV_AUDIT
				dbNameToAddr(pv, &dbAddr);
				if ((int) dbAddr.precord < 0)
				    continue;
				if ((int) dbAddr.pfield < 0)
				    continue;
				if (alarmCtl.alarmCtlFunc)
				    alarmState = 1;
				dbPutField(&dbAddr, mapOld2New(AGNT_SHORT), (caddr_t)&alarmState, 1);
#ifdef PRIV_AUDIT
				if (auditLog)
				    {
				    sprintf(updateMsg, "Operator alarm status change for %s %d", pv, alarmState);
				    AuditAddEvent (auditLog, AUDIT_WRITE, AUDIT_INFO, X509_dn,
					    taskName(taskIdSelf()), updateMsg);
				    }
#endif PRIV_AUDIT
				}

			    else
				{
				printErr("Illegal message type %d from %s\n", msgType, INET_NTOA(currCCB -> clientInet));
				continue;
				}

			    }
#ifdef PRIV_AUDIT
			if (currCCB -> clientFd && bytesGot < 0)
			    {
			    int status;
			    long err;

			    /* don't quit on no data, unless an error condition! */
			    status = SSL_get_error (sslSharedGetConn(currCCB -> clientFd), bytesGot);
			    err = ERR_get_error ();
			    if (err != 0)
				logMsg ("%s\n", (int)ERR_error_string (err, NULL),0,0,0,0,0);
			    switch (status)
				{
				case SSL_ERROR_WANT_WRITE:
				case SSL_ERROR_NONE:
				case SSL_ERROR_WANT_CONNECT:
				case SSL_ERROR_WANT_READ:
				case SSL_ERROR_WANT_X509_LOOKUP:
				    /* benign, just want to write/read for state machine */
				    continue;
				case SSL_ERROR_SSL:
				    if (errno == EWOULDBLOCK)
					continue;
				case SSL_ERROR_SYSCALL:
				case SSL_ERROR_ZERO_RETURN:
				default:
				    /* all these close the connection! */
				    if (auditLog)
					{
					GetPeerDN (sslSharedGetConn(currCCB -> clientFd), X509_dn, sizeof(X509_dn));
					AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
							    taskName(taskIdSelf()), "Connection closed due to read error");
					}
				    sslSharedCloseConn (sslSharedGetConn(currCCB -> clientFd));
				    FD_CLR(currCCB -> clientFd, &rsrvFds);
				    FD_CLR(currCCB -> clientFd, &myFds);
				    currCCB -> clientFd = 0;
				    currCCB -> handlerMsgRcvd = 0;
				    currCCB -> txBuffSize = 0;
				    currCCB -> alarmHandler = 0;
				    currCCB -> trendHandler = 0;
				    break;
				}
			    }
#endif PRIV_AUDIT
			}
		    }

		/*
		Refresh our copy of the fd mask just in case it got changed.
		*/

		bcopy((char *) &rsrvFds, (char *) &myFds, sizeof(myFds));
		}
	    }

	/*
	Check the Queue. Do this every time through the loop. We really don't have to do this
	while processing the sockets because we anticipate that the amount of input traffic will
	be much smaller than the amount of the output traffic.
	*/


	while (!rngIsEmpty(ringId))
	    {

	    /*
	    Get the word that will tell us what kind of structure follows.
	    */

	    rngBufGet(ringId, (char *) &msgType, sizeof(msgType));

	    if (msgType == EVENT_HDR)
		{
		eventRBH	eventHdr;
		struct dbAddr	localAddr;
		CH_MONDATA	monData;

		/*
		Event monitor. Get the header and then the data.
		*/

		rngBufGet(ringId, (char *) &eventHdr, sizeof(eventHdr));
		rngBufGet(ringId, buff, eventHdr.length);
		recHdr = (struct dbCommon *) eventHdr.precord;

		/*
		We'll process any FCBs if field and mask match
		*/

		pFCB = (FCB *) (eventHdr.precord -> mlis.node.next);
		while (pFCB)
		    {
		    if (eventHdr.pfield == (pFCB -> dbAddr.pfield))
			if (pFCB -> currMask & eventHdr.mask)
			    {

			    /*
			    Change the dbAddr structure to point to our buffer instead of the data record.
			    We do that because we want to report the value that caused the event, not the one
			    that is currently in the database record. They could be very different!
			    */

			    localAddr = pFCB -> dbAddr;
			    localAddr.pfield = buff;

			    /*
			    Now march down the list of monitor control blocks and format the event monitor
			    message for each client that's interested. The formats may be different!
			    */

			    pMCB = pFCB -> firstChan;
			    msgType = htons(CHAN_MONDATA);
			    while (pMCB)
		    	    	{
		    	    	if ((pMCB -> chanMask) & eventHdr.mask)
				    {
				    monData.chHandle = pMCB -> handle;
				    monData.chDataType = htons(pMCB -> dataType);
				    monData.chMask = htons(eventHdr.mask);
				    monData.chCount =  htons(pMCB -> numElem);
				    monData.chDataLen = htons(getAgentField(rsrvOutBuff, pMCB -> dataType, 
							    pMCB -> numElem, eventHdr.stat, eventHdr.sevr, 
							    eventHdr.ack, &eventHdr.stamp, &localAddr));
				    if (pMCB -> client -> clientFd && pMCB -> client -> handlerMsgRcvd)
					{
#ifdef PRIV_AUDIT
					char	privChannel [40];
					/*
					See if he has priviledge for channel!
					*/
					sprintf (privChannel, "%.28s.%.4s", localAddr.precord, localAddr.pfldDes);
					GetPeerCN (sslSharedGetConn (pMCB->client->clientFd), X509_dn, sizeof(X509_dn));
					if (GetChannelPriv (X509_dn, privChannel) != PRIV_NOACCESS)
#endif PRIV_AUDIT
					    {
					    sendClient(pMCB -> client, msgType, (char *) &monData, sizeof(monData), 
						    rsrvOutBuff, ntohs(monData.chDataLen), (fd_set *) 0);
					    }
					}
				    }
		    	        pMCB = pMCB -> nextMCB;
		    		}
			    break;
			    }
		    pFCB = pFCB -> nextFCB;
		    }

		/*
		Check for log and alarm bits. Events with these bits on should be sent
		to all machines that have the appropriate type of handler.
		*/

		if ((eventHdr.mask & AGNT_ALARM) && (recHdr -> algp[0] != 0))
		    {
		    CH_ALARM		alarmData;

		    if (debugging)
			printf("Got an alarm event!\n");
		    alarmData.alarmTime.secPastEpoch = htonl(eventHdr.stamp.secPastEpoch);
		    alarmData.alarmTime.nsec = htonl(eventHdr.stamp.nsec);
		    alarmData.alarmSerial = ++alarmSerial;
		    alarmData.alarmStat = htons(eventHdr.stat);
		    eventHdr.sevr |= 0x8000;
		    if (recHdr -> ack)
			eventHdr.sevr |= 0x4000;
		    alarmData.alarmSevr = htons(eventHdr.sevr);
		    alarmData.alarmAck = htons(recHdr -> ack);
		    alarmData.alarmEnabled = htons(recHdr -> aenb);
		    alarmData.alarmOnscan = htons(recHdr -> oscn);
		    alarmData.alarmDataLen = getAlarmStream(recHdr, eventHdr.recType, buff, rsrvOutBuff);
		    alarmData.alarmDataLen = htons(alarmData.alarmDataLen);
		    sendNewAlarm(&alarmData, rsrvOutBuff, NULL);
		    }

		if (eventHdr.mask & AGNT_LOG)
		    {
		    CH_TRDATA		trendData;
		    struct fldDes	*pFldDes;
		    short		endianShort;
		    TS_STAMP		endianStamp;

		    /*
		    If the trend enable bit is clear, then ignore the message. Otherwise, Format a message 
		    for the trend handler.
		    */

		    if (!(recHdr -> tenb))
			continue;
		    strcpy(rsrvOutBuff, recHdr -> name);
		    pFldDes = eventHdr.pFldDes;
		    if (strncmp(pFldDes -> fldname, "VAL ", FLDNAME_SZ))
			{
			char		*space;

		    	trendData.chNameLen = strlen(rsrvOutBuff);
			rsrvOutBuff[trendData.chNameLen++] = '.';
			strncpy(rsrvOutBuff + trendData.chNameLen, pFldDes -> fldname, FLDNAME_SZ);
			rsrvOutBuff[trendData.chNameLen + 4] = 0;
			if ((space = (char *)strchr(rsrvOutBuff, ' ')))
			    *space = 0;
			}
		    trendData.chNameLen = strlen(rsrvOutBuff) + 1;	/* Include the Null! */
		    trendData.chDataType = htons(mapNew2Old(pFldDes -> field_type) | AGNT_STAT | AGNT_TIME);
		    endianShort = htons(eventHdr.stat);
		    bcopy((char *) &endianShort, rsrvOutBuff + trendData.chNameLen, sizeof(short));
		    trendData.chDataLen = sizeof(short);
		    eventHdr.sevr |= 0x8000;
		    if (recHdr -> ack)
			eventHdr.sevr |= 0x4000;
		    endianShort = htons(eventHdr.sevr);
		    bcopy((char *) &endianShort, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, 
			    sizeof(short));
		    trendData.chDataLen += sizeof(short);
		    bcopy((char *) &eventHdr.stamp, (char *) &endianStamp, sizeof(TS_STAMP));
		    endianStamp.secPastEpoch = htonl(endianStamp.secPastEpoch);
		    endianStamp.nsec = htonl(endianStamp.nsec);
		    bcopy((char *) &endianStamp, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, 
			    sizeof(TS_STAMP));
		    trendData.chDataLen += sizeof(TS_STAMP);
		    bcopy(buff, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, pFldDes -> size);
		    switch (mapNew2Old(pFldDes -> field_type))
			{
			case AGNT_SHORT:
			case AGNT_ENUM:
			    {
			    short	m68kShort;

			    bcopy(rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, (char *) &m68kShort, sizeof(short));
			    m68kShort = htons(m68kShort);
			    bcopy((char *) &m68kShort, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, sizeof(short));
			    }
			    break;
			case AGNT_FLOAT:
			    {
			    float	m68kFloat;

			    bcopy(rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, (char *) &m68kFloat, sizeof(float));
			    htonf(m68kFloat, m68kFloat);
			    bcopy((char *) &m68kFloat, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, sizeof(float));
			    }
			    break;
			case AGNT_LONG:
			    {
			    long	m68kLong;

			    bcopy(rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, (char *) &m68kLong, sizeof(long));
			    m68kLong = htonl(m68kLong);
			    bcopy((char *) &m68kLong, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, sizeof(long));
			    }
			    break;
			case AGNT_DOUBLE:
			    {
			    double	m68kDouble;

			    bcopy(rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, (char *) &m68kDouble, sizeof(double));
			    htond(m68kDouble, m68kDouble);
			    bcopy((char *) &m68kDouble, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, sizeof(double));
			    }
			    break;
			}
		    trendData.chNameLen = htons(trendData.chNameLen);
		    trendData.chDataLen += pFldDes -> size;
		    trendData.chDataLen = htons(trendData.chDataLen);
		    msgType = htons(CHAN_TRDATA);

		    /*
		    Now send it to all the trend data handlers that we know about.
		    */

		    currCCB = firstClient;
		    while (currCCB)
		        {
		        if (currCCB -> clientFd && currCCB -> trendHandler && currCCB -> handlerMsgRcvd)
			    {
#ifdef PRIV_AUDIT
			    /*
			    See if he has priviledge for channel!
			    */
			    GetPeerCN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
			    if (GetChannelPriv (X509_dn, rsrvOutBuff) != PRIV_NOACCESS)
#endif PRIV_AUDIT
				{
				sendClient(currCCB, msgType, (char *) &trendData, sizeof(trendData), 
					rsrvOutBuff, ntohs(trendData.chNameLen) + ntohs(trendData.chDataLen), 
					(fd_set *) 0);
				}
			    }
		        currCCB = currCCB -> nextClient;
		        }
		    }
		}

	    else if (msgType == RESRC_HDR)
		{
		resrcRBH	resrcHdr;
		CH_RESMODE	resMode;

		if (debugging)
		    printf("Got a resource event!\n");
		rngBufGet(ringId, (char *) &resrcHdr, sizeof(resrcHdr));
		resMode.chResStatus = htons(resrcHdr.newStat);
		strcpy(resMode.chResName, resrcHdr.resName);
		msgType = htons(CHAN_RESMODE);
		currCCB = firstClient;
		while (currCCB)
		    {
		    if (currCCB -> clientFd && currCCB -> handlerMsgRcvd)
			sendClient(currCCB, msgType, (char *) &resMode, sizeof(resMode), (char *) 0, 0,
					(fd_set *) 0);
		    currCCB = currCCB -> nextClient;
		    }
		}

	    else if (msgType == HNDLR_HDR)
		{
		handlerRBH	handlrHdr;

		/*
		We have data for a trend or alarm handler. Get the header and use that to figure out
		the length of the data.
		*/

		rngBufGet(ringId, (char *) &handlrHdr, sizeof(handlrHdr));
		rngBufGet(ringId, buff, handlrHdr.pFldDes -> size);

		/*
		Format the output message. Remember that both log and alarm bits may be on. Each of
		them could (at the time this was written, anyway) have a different format, so handle
		them independently.
		*/

		if (handlrHdr.mask & AGNT_LOG)
		    {
		    CH_TRDATA	trendData;
		    short	endianShort;
		    TS_STAMP	endianStamp;

		    /*
		    For trend data, set up the data buffer and the fixed portion of the message.
		    */

		    strcpy(rsrvOutBuff, handlrHdr.pRecord -> name);
			
			/*
			Be careful. Field names in the SDR are blank padded, not null terminated!
			Make it null terminated to make life easier everywhere else in the system.
			*/

		    if (strncmp(handlrHdr.pFldDes -> fldname, "VAL ", FLDNAME_SZ))
			{
			char	*space;

		    	trendData.chNameLen = strlen(rsrvOutBuff);
			rsrvOutBuff[trendData.chNameLen++] = '.';
			strncpy(rsrvOutBuff + trendData.chNameLen, handlrHdr.pFldDes -> fldname, FLDNAME_SZ);
			rsrvOutBuff[trendData.chNameLen + 4] = 0;
			if ((space = (char *) strchr(rsrvOutBuff, ' ')))
			    *space = 0;
			}
		    trendData.chNameLen = strlen(rsrvOutBuff) + 1;	/* Include the Null! */
		    trendData.chDataType = htons(mapNew2Old(handlrHdr.pFldDes -> field_type) | AGNT_STAT | AGNT_TIME);
		    endianShort = htons(handlrHdr.stat);
		    bcopy((char *) &endianShort, rsrvOutBuff + trendData.chNameLen, sizeof(short));
		    trendData.chDataLen = sizeof(short);
		    handlrHdr.sevr |= 0x8000;
		    if (handlrHdr.pRecord -> ack)
			handlrHdr.sevr |= 0x4000;
		    endianShort = handlrHdr.sevr;
		    bcopy((char *) &endianShort, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, 
			    sizeof(short));
		    trendData.chDataLen += sizeof(short);
		    bcopy((char *) &handlrHdr.stamp, (char *) &endianStamp, sizeof(TS_STAMP));
		    endianStamp.secPastEpoch = htonl(endianStamp.secPastEpoch);
		    endianStamp.nsec = htonl(endianStamp.nsec);
		    bcopy((char *) &endianStamp, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, 
			    sizeof(TS_STAMP));
		    trendData.chDataLen += sizeof(TS_STAMP);
		    bcopy(buff, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, handlrHdr.pFldDes -> size);
		    switch (mapNew2Old(handlrHdr.pFldDes -> field_type))
			{
			case AGNT_SHORT:
			case AGNT_ENUM:
			    {
			    short	m68kShort;

			    bcopy(rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, (char *) &m68kShort, sizeof(short));
			    m68kShort = htons(m68kShort);
			    bcopy((char *) &m68kShort, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, sizeof(short));
			    }
			    break;
			case AGNT_FLOAT:
			    {
			    float	m68kFloat;

			    bcopy(rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, (char *) &m68kFloat, sizeof(float));
			    htonf(m68kFloat, m68kFloat);
			    bcopy((char *) &m68kFloat, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, sizeof(float));
			    }
			    break;
			case AGNT_LONG:
			    {
			    long	m68kLong;

			    bcopy(rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, (char *) &m68kLong, sizeof(long));
			    m68kLong = htonl(m68kLong);
			    bcopy((char *) &m68kLong, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, sizeof(long));
			    }
			    break;
			case AGNT_DOUBLE:
			    {
			    double	m68kDouble;

			    bcopy(rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, (char *) &m68kDouble, sizeof(double));
			    htond(m68kDouble, m68kDouble);
			    bcopy((char *) &m68kDouble, rsrvOutBuff + trendData.chNameLen + trendData.chDataLen, sizeof(double));
			    }
			    break;
			}
		    trendData.chDataLen += handlrHdr.pFldDes -> size;
		    trendData.chNameLen = htons(trendData.chNameLen);
		    trendData.chDataLen = htons(trendData.chDataLen);
		    msgType = htons(CHAN_TRDATA);

		    /*
		    Now send it to all the trend data handlers that we know about.
		    */

		    currCCB = firstClient;
		    while (currCCB)
		        {
			if (handlrHdr.client && (handlrHdr.client != currCCB))
			    {
			    currCCB = currCCB -> nextClient;
			    continue;
			    }
		        if (currCCB -> clientFd && currCCB -> trendHandler && currCCB -> handlerMsgRcvd)
			    {
#ifdef PRIV_AUDIT
			    /*
			    See if he has priviledge for channel!
			    */
			    GetPeerCN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
			    if (GetChannelPriv (X509_dn, rsrvOutBuff) != PRIV_NOACCESS)
#endif PRIV_AUDIT
				{
				sendClient(currCCB, msgType, (char *) &trendData, sizeof(trendData), 
					rsrvOutBuff, ntohs(trendData.chNameLen) + ntohs(trendData.chDataLen), 
					(fd_set *) 0);
				}
			    if (handlrHdr.client)
				break;
			    }
		        currCCB = currCCB -> nextClient;
		        }
		    }
		
		if ((handlrHdr.mask & AGNT_ALARM) && (handlrHdr.pRecord -> algp[0] != 0))
		    {
		    CH_ALARM	alarmData;

		    /*
		    Create an unsolicited alarm message for the alarm handlers in the system.
		    */

		    alarmData.alarmTime.secPastEpoch = htonl(handlrHdr.stamp.secPastEpoch);
		    alarmData.alarmTime.nsec = htonl(handlrHdr.stamp.nsec);
		    if (handlrHdr.client)
			alarmData.alarmSerial = 0;
		    else
		        alarmData.alarmSerial = ++alarmSerial;
		    alarmData.alarmStat = htons(handlrHdr.stat);
		    handlrHdr.sevr |= 0x8000;
		    if (handlrHdr.pRecord -> ack)
			handlrHdr.sevr |= 0x4000;
		    alarmData.alarmSevr = htons(handlrHdr.sevr);
		    alarmData.alarmAck = htons(handlrHdr.pRecord -> ack);
		    alarmData.alarmEnabled = htons(handlrHdr.pRecord -> aenb);
		    alarmData.alarmOnscan = htons(handlrHdr.pRecord -> oscn);
		    alarmData.alarmDataLen = htons(getAlarmStream(handlrHdr.pRecord, handlrHdr.recType, buff, 
			    rsrvOutBuff));
		    sendNewAlarm(&alarmData, rsrvOutBuff, handlrHdr.client);
		    }
		}
	    else
		printErr("Invalid message type in ring buffer!\n");
	    }

	/*
	Now push out anything that is left in the buffers. We don't want to wait to keep the system
	responsive.
	*/

	currCCB = firstClient;
	while (currCCB)
	    {
	    if (currCCB -> txBuffSize)
		clientSend(currCCB, (fd_set *) 0);
	    currCCB = currCCB -> nextClient;
	    }

	/*
	If the I/O slave was waiting for us, wake him back up.
	*/

	if (rsrvIoFlag)
	    semGive(rsrvIoSem);
	}
}


/****************************************************************/
/*								*/
/*	FIND A CCB GIVEN THE CLIENT'S INTERNET ADDRESS		*/
/*								*/
/****************************************************************/

CCB *findCCB(struct in_addr inetAddr)
{
    CCB		*currCCB;

    currCCB = firstClient;
    while (currCCB)
	{
	if (currCCB -> clientInet.s_addr == inetAddr.s_addr)
	    break;
	currCCB = currCCB -> nextClient;
	}
    return (currCCB);
}


/****************************************************************/
/*								*/
/*		FIND A CCB GIVEN THE SOCKET FD			*/
/*								*/
/****************************************************************/

CCB *rsrvFindFd(int fd)
{
    CCB		*currCCB;

    currCCB = firstClient;
    while (currCCB)
	{
	if (currCCB -> clientFd == fd)
	    break;
	currCCB = currCCB -> nextClient;
	}
    return (currCCB);
}


/****************************************************************/
/*								*/
/*		SEND BYTE STREAM TO THE CLIENT			*/
/*								*/
/*	Return a 1 if successful, zero on failure.		*/
/*								*/
/****************************************************************/

int clientSend(CCB *currCCB, fd_set *otherFds)
{
    int		bytesSent;
#ifdef PRIV_AUDIT
    SSL		*conn;
#endif PRIV_AUDIT

    if (currCCB -> clientFd == 0)
	{
	currCCB -> txBuffSize = 0;
	return (0);
	}

#ifdef PRIV_AUDIT
    conn = sslSharedGetConn (currCCB -> clientFd);
    if ((bytesSent = SSL_write (conn, currCCB -> txBuff, currCCB -> txBuffSize)) == ERROR)
#else
    if ((bytesSent = send(currCCB -> clientFd, currCCB -> txBuff, currCCB -> txBuffSize, 0)) == ERROR)
#endif PRIV_AUDIT
	{
	if (errno == EWOULDBLOCK)
	    {
	    currCCB -> clientTxBuff += currCCB -> txBuffSize;
	    if (setsockopt(currCCB -> clientFd, SOL_SOCKET, SO_SNDBUF, (char *)&(currCCB -> clientTxBuff),
		    sizeof(currCCB -> clientTxBuff)) == ERROR)
		{
#ifdef PRIV_AUDIT
		if (auditLog)
		    {
		    GetPeerDN (conn, X509_dn, sizeof(X509_dn));
		    AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
				    taskName(taskIdSelf()), "Connection closed due to send error");
		    }
		sslSharedCloseConn (conn);
#else
		close(currCCB -> clientFd);
#endif PRIV_AUDIT
		FD_CLR(currCCB -> clientFd, &rsrvFds);
		if (otherFds)
		    FD_CLR(currCCB -> clientFd, otherFds);
		currCCB -> clientFd = 0;
		currCCB -> handlerMsgRcvd = 0;
		currCCB -> txBuffSize = 0;
		currCCB -> alarmHandler = 0;
		currCCB -> trendHandler = 0;
		logMsg("Closed socket to %s because of Tx buffer maxout.\n", (int)INET_NTOA(currCCB -> clientInet),0,0,0,0,0);
		return (0);
		}
	    else
		/*printf("Increased %s Tx buffer to %d.\n", pinet(currCCB -> clientInet), currCCB -> clientTxBuff);*/
		;
	    }
	else
	    {
#ifdef PRIV_AUDIT
	    if (auditLog)
		{
		GetPeerDN (conn, X509_dn, sizeof(X509_dn));
		AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
				    taskName(taskIdSelf()), "Connection closed due to send overflow");
		}
	    sslSharedCloseConn (conn);
#else
	    close(currCCB -> clientFd);
#endif PRIV_AUDIT
	    FD_CLR(currCCB -> clientFd, &rsrvFds);
	    if (otherFds)
		FD_CLR(currCCB -> clientFd, otherFds);
	    currCCB -> clientFd = 0;
	    currCCB -> handlerMsgRcvd = 0;
	    currCCB -> txBuffSize = 0;
	    currCCB -> alarmHandler = 0;
	    currCCB -> trendHandler = 0;
	    return (0);
	    }
	return (clientSend(currCCB, otherFds));
	}
    else if (bytesSent != currCCB -> txBuffSize)
	{
	currCCB -> txBuffSize -= bytesSent;
	bcopy(currCCB -> txBuff + bytesSent, currCCB -> txBuff, currCCB -> txBuffSize);
	return (clientSend(currCCB, otherFds));
	}

    currCCB -> txBuffSize = 0;
    return (1);
}


/****************************************************************/
/*								*/
/*		GET BYTES FROM A STREAM SOCKET			*/
/*								*/
/*	Return a 1 if successful, zero on failure.		*/
/*								*/
/****************************************************************/

int clientRecv(char *buff, int nbytes, CCB *currCCB, fd_set *localFds)
{
    int			bytesGot;
    fd_set		oneFd;
    struct timeval	timeOut;
#ifdef PRIV_AUDIT
    SSL			*conn = sslSharedGetConn (currCCB -> clientFd);
#endif PRIV_AUDIT

    if (currCCB -> clientFd == 0)
	return (0);
    timeOut.tv_sec = 5;
    timeOut.tv_usec = 0;
    FD_ZERO(&oneFd);
    FD_SET(currCCB -> clientFd, &oneFd);
    while (nbytes)
	{
#ifdef PRIV_AUDIT
	/* Must read BEFORE select in case data is in BIO */
        if ((bytesGot = SSL_read (conn, buff, nbytes)) > 0)
	    {
	    if (debugging)
		{
		printf("Received %d bytes\n", bytesGot);
		}
	    currCCB -> lastRxTime = (unsigned long) tickGet();
	    buff += bytesGot;
	    nbytes -= bytesGot;
	    continue;
	    }
#endif PRIV_AUDIT
	if (select(currCCB -> clientFd + 1, &oneFd, NULL, NULL, &timeOut) <= 0)
	    {
	    if (debugging)
		printf("Timeout!\n");
#ifdef PRIV_AUDIT
	    if (auditLog)
		{
		GetPeerDN (conn, X509_dn, sizeof(X509_dn));
		AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
				    taskName(taskIdSelf()), "Connection closed due to read timeout");
		}
	    sslSharedCloseConn (conn);
#else
	    close(currCCB -> clientFd);
#endif PRIV_AUDIT
	    FD_CLR(currCCB -> clientFd, &rsrvFds);
	    FD_CLR(currCCB -> clientFd, localFds);
	    currCCB -> clientFd = 0;
	    currCCB -> handlerMsgRcvd = 0;
	    currCCB -> txBuffSize = 0;
	    currCCB -> alarmHandler = 0;
	    currCCB -> trendHandler = 0;
	    return (0);
	    }
#ifdef PRIV_AUDIT
        if ((bytesGot = SSL_read (conn, buff, nbytes)) <= 0)
#else
	if ((bytesGot = recv(currCCB -> clientFd, buff, nbytes, 0)) <= 0)
#endif PRIV_AUDIT
	    {
	    if (debugging)
		{
		if (bytesGot == 0)
		    printf("Zero length receive\n");
		else
		    perror("Read error");
		}
#ifdef PRIV_AUDIT
	    if (auditLog)
		{
		GetPeerDN (conn, X509_dn, sizeof(X509_dn));
		AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
				    taskName(taskIdSelf()), "Connection closed due to read error");
		}
	    sslSharedCloseConn (conn);
#else
	    close(currCCB -> clientFd);
#endif PRIV_AUDIT
	    FD_CLR(currCCB -> clientFd, &rsrvFds);
	    FD_CLR(currCCB -> clientFd, localFds);
	    currCCB -> clientFd = 0;
	    currCCB -> handlerMsgRcvd = 0;
	    currCCB -> txBuffSize = 0;
	    currCCB -> alarmHandler = 0;
	    currCCB -> trendHandler = 0;
	    return (0);
	    }
	if (debugging)
	    {
	    printf("Received %d bytes\n", bytesGot);
	    }
	currCCB -> lastRxTime = (unsigned long) tickGet();
	buff += bytesGot;
	nbytes -= bytesGot;
	}
    return (1);
}

/****************************************************************/
/*								*/
/*		ENQUEUE MESSAGE TO A CLIENT			*/
/*								*/
/*	This routine builds an output message to a client.	*/
/*	A pointer to a CCB tells us which client and also	*/
/*	locates his transmit buffer. The message is made up	*/
/*	from a message type and up to two additional buffer	*/
/*	pointers. If there isn't room in the transmit buffer	*/
/*	for the complete message, then we empty the existing	*/
/*	buffer first. The entire message goes into the buffer	*/
/*	as one chunk to minimize TCP's message fragmenting.	*/
/*	This routine supplies the SOCK_READ wrapper for the	*/
/*	message.						*/
/*								*/
/*	Return a 1 if successful, a 0 on failure.		*/
/*								*/
/****************************************************************/

int sendClient(CCB *pCCB, short msgType, char *buff1, int buff1Len, 
	       char *buff2, int buff2Len, fd_set *fdMask)
{
    SM_READ	readHdr;
    short	sockReadType = htons(SOCK_READ);
    int		bytesRqd;

    /*
    Calculate the total size that we need.
    */

    bytesRqd= sizeof(sockReadType) + sizeof(readHdr) + sizeof(msgType) + buff1Len + buff2Len;
    if (bytesRqd > sizeof(pCCB -> txBuff))
	return (0);				/* Won't ever fit! */
    if ((pCCB -> txBuffSize + bytesRqd) > sizeof(pCCB -> txBuff))
	if (!clientSend(pCCB, fdMask))
	    return (0);
    
    /*
    The message we want to send consists of a SOCK_READ message type, a struct SM_READ, and
    then the user's message type followed by his buffers.
    */

    bcopy((char *) &sockReadType, pCCB -> txBuff + pCCB -> txBuffSize, sizeof(sockReadType));
    pCCB -> txBuffSize += sizeof(sockReadType);
    readHdr.smLength = htons(sizeof(msgType) + buff1Len + buff2Len);
    bcopy((char *) &readHdr, pCCB -> txBuff + pCCB -> txBuffSize, sizeof(readHdr));
    pCCB -> txBuffSize += sizeof(readHdr);
    bcopy((char *) &msgType, pCCB -> txBuff + pCCB -> txBuffSize, sizeof(msgType));
    pCCB -> txBuffSize += sizeof(msgType);
    if (buff1 && buff1Len)
	{
	bcopy(buff1, pCCB -> txBuff + pCCB -> txBuffSize, buff1Len);
	pCCB -> txBuffSize += buff1Len;
	}
    if (buff2 && buff2Len)
	{
	bcopy(buff2, pCCB -> txBuff + pCCB -> txBuffSize, buff2Len);
	pCCB -> txBuffSize += buff2Len;
	}
    return (1);
}


/****************************************************************/
/*								*/
/*		CLOSE A CHANNEL WITH GIVEN MCB			*/
/*								*/
/****************************************************************/

int closeMCB(MCB *pMCB)
{
    struct dbCommon	*recHdr;
    FCB			*pFCB;
    MCB			*currMCB;

    /*
    First disconnect this MCB from the CCB. That won't affect anyone else.
    */

    if (pMCB -> nextClient)
	pMCB -> nextClient -> prevClient = pMCB -> prevClient;
    if (pMCB -> prevClient)
	pMCB -> prevClient -> nextClient = pMCB -> nextClient;
    else
	pMCB -> client -> firstChan = pMCB -> nextClient;

    /*
    Now we will be altering the monitor list. Lock it so that we don't screw up other users.
    */

    recHdr = (struct dbCommon *) (pMCB -> fldCntl -> dbAddr.precord);
    FASTLOCK(&(recHdr -> mlok));

    /*
    Traverse the MCB chain to develop the new overall event mask. While we are doing this,
    unchain the old MCB.
    */

    pFCB = pMCB -> fldCntl;
    pFCB -> currMask = 0;
    currMCB = pFCB -> firstChan;
    if (currMCB == pMCB)
	{

	/*
	Skip over the first one if its the one we want to delete.
	*/

	currMCB = pMCB -> nextMCB;
	pFCB -> firstChan = currMCB;
	}
    while (currMCB)
	{

	/*
	We know that we won't see the MCB to be deleted the first time through the loop. We also
	know that we won't see it any subsequent time through the loop because of the "double pump fake"
	we will do on the forward link at the end of the loop.
	*/

	if (currMCB -> dataType != AGNT_NOCONN)
	    pFCB -> currMask |= currMCB -> chanMask;
	if (currMCB -> nextMCB == pMCB)
	    currMCB -> nextMCB = pMCB -> nextMCB;
	currMCB = currMCB -> nextMCB;
	}

    /*
    If there are no more MCBs chained on to the FCB, then it can also be released.
    */

    if (!(pFCB -> firstChan))
	{
	if (pFCB -> nextFCB)
	    pFCB -> nextFCB -> prevFCB = pFCB -> prevFCB;
	if (pFCB -> prevFCB)
	    pFCB -> prevFCB -> nextFCB = pFCB -> nextFCB;
	else
	    recHdr -> mlis.node.next = (struct node *) (pFCB -> nextFCB);
	recHdr -> mlis.count--;
	}
    else
	pFCB = 0;

    /*
    Unlock the monitor list because we are done.
    */

    FASTUNLOCK(&(recHdr -> mlok));

    /*
    Now release the MCB and the FCB if it was empty.
    */

    deleteMCB(pMCB);
    if (pFCB)
	deleteFCB(pFCB);
    return 0;
}

/****************************************************************/
/*								*/
/*		TURN DEBUGGING ON AND OFF			*/
/*								*/
/*	A simple routine to toggle the state of debugging.	*/
/*								*/
/****************************************************************/

int rsrvDebug()
{
    debugging = !debugging;
    if (debugging)
	{
	printf("RSRV debugging on\n");
	}
    else
	{
	printf("RSRV debugging off\n");
	}
    return (0);
}


/****************************************************************/
/*								*/
/*		DECLARE MAXIMUM ALARM QUEUE SIZE		*/
/*								*/
/****************************************************************/

int setMaxAlarmQue(size)
int	size;
{
    maxAlarmQue = size;
    return (0);
}


/****************************************************************/
/*								*/
/*	PUT AN ALARM MESSAGE INTO THE QUEUE AND SEND IT TO	*/
/*	ALL THE ALARM HANDLERS THAT WE KNOW ABOUT 		*/
/*								*/
/****************************************************************/

int sendNewAlarm(CH_ALARM *alarmMsg, char *alarmData, CCB *client)
{
    short		msgType;
    CCB			*currCCB;
    ALARM_QUE		*newAlarm;
    struct dbCommon	*recHdr;

    /*
    First thing to do is to send this message to all the registered
    alarm handlers if the alarm was not tagged as being dedicated
    to a single client.
    */

    if (debugging)
	printf("New alarm for %s\n", alarmData);
    msgType = htons(CHAN_ALARM);
    currCCB = firstClient;
    while (currCCB)
	{
	if (client && (client != currCCB))
	    {
	    currCCB = currCCB -> nextClient;
	    continue;
	    }
	if (currCCB -> clientFd && currCCB -> alarmHandler && currCCB -> handlerMsgRcvd)
	    {
#ifdef PRIV_AUDIT
	    /*
	    See if he has priviledge for channel!
	    */
	    GetPeerCN (sslSharedGetConn (currCCB->clientFd), X509_dn, sizeof(X509_dn));
	    if (GetChannelPriv (X509_dn, alarmData) != PRIV_NOACCESS)
#endif PRIV_AUDIT
		{
		sendClient(currCCB, msgType, (char *) alarmMsg, sizeof(CH_ALARM),
			alarmData, ntohs(alarmMsg -> alarmDataLen), (fd_set *) 0);
		}
	    /*
	    Alarms destined for a single client do not get put into the queue.
	    */

	    if (client)
		return (0);
	    }
	currCCB = currCCB -> nextClient;
	}

    /*
    Now insert this message into the queue.
    */

    if (freeAlarms)
	{
	newAlarm = freeAlarms;
	freeAlarms = freeAlarms -> nextAlq;
	}
    else if (alarmQueSize < maxAlarmQue)
	{
	newAlarm = (ALARM_QUE *) malloc(sizeof(ALARM_QUE));
	alarmQueSize++;
	}
    else
	{
	ALARM_QUE	*prevAlarm;

	/*
	The alarm queue has grown to its maximum size. The queue is time
	ordered oldest to newest, so the first one we find with a matching
	PV name can be deleted.
	*/

	prevAlarm = 0;
	newAlarm = alarmHead;
	while (newAlarm)
	    {
	    if (!strcmp(newAlarm -> alarmData, alarmData))
		break;
	    prevAlarm = newAlarm;
	    newAlarm = newAlarm -> nextAlq;
	    }
	if (newAlarm)
	    {
	    if (rsrvPrintingWarnings)
	        printf("Oldest duplicate alarm for %s discarded\n", newAlarm -> alarmData);
	    if (prevAlarm)
		{
		prevAlarm -> nextAlq = newAlarm -> nextAlq;
		if (alarmTail == newAlarm)
		    alarmTail = prevAlarm;
		}
	    else
		{
		alarmHead = newAlarm -> nextAlq;
		if (alarmTail == newAlarm)
		    alarmTail = alarmHead;
		}
	    }
	else
	    {

	    /*
	    No duplicates. Drop the oldest alarm in our queue.
	    */

	    if (rsrvPrintingWarnings)
	    	printf("Alarm for %s dropped because of queue overflow.\n", alarmHead -> alarmData);
	    newAlarm = alarmHead;
	    alarmHead = alarmHead -> nextAlq;
	    if (alarmTail == newAlarm)
		alarmTail = alarmHead;
	    }
	}

    /*
    Fill out the data and put the message onto the tail of the queue.
    */

    newAlarm -> alarmMsg = *alarmMsg;
    pvdGetAddr(alarmData, (char **)&recHdr);
    newAlarm -> radr = recHdr -> radr;
    bcopy(alarmData, newAlarm -> alarmData, ntohs(alarmMsg -> alarmDataLen));
    newAlarm -> nextAlq = 0;

    if (alarmTail)
	{
	alarmTail -> nextAlq = newAlarm;
	alarmTail = newAlarm;
	}
    else
	{
	alarmHead = newAlarm;
	alarmTail = newAlarm;
	}

    /*
    Also send this alarm to our redundant partner in case of our untimely death.
    */

    sendRedAlarms(((RES_TABLE *)(recHdr -> radr)) -> redAlarmFd, msgType, (char *) newAlarm,
	    sizeof(CH_ALARM) + ntohs(alarmMsg -> alarmDataLen));
    return 0;
}


/****************************************************************/
/*								*/
/*		TURN OFF ALARM WARNING PRINTING			*/
/*								*/
/****************************************************************/

int alarmWarningPrintOff()
{
    rsrvPrintingWarnings = 0;
    return 0;
}


/****************************************************************/
/*								*/
/*		TURN ON ALARM WARNING PRINTING			*/
/*								*/
/****************************************************************/

int alarmWarningPrintOn()
{
    rsrvPrintingWarnings = 1;
    return 0;
}


/****************************************************************/
/*								*/
/*			PRINT THE ALARM QUEUE			*/
/*								*/
/****************************************************************/

int alarmQueuePrint()
{
    ALARM_QUE		*currAlq;
    char		*desc;
    char		*value;
    char		*egu;

    currAlq = alarmHead;
    while (currAlq)
	{
	printf("Alarm for %s\n", currAlq -> alarmData);
	printf("\ttime = %d.%d, serial = %d\n", ntohl(currAlq -> alarmMsg.alarmTime.secPastEpoch),
		ntohl(currAlq -> alarmMsg.alarmTime.nsec), currAlq -> alarmMsg.alarmSerial);
	printf("\tstat = %d, sevr = %d, ack = %d\n", ntohs(currAlq -> alarmMsg.alarmStat),
		ntohs(currAlq -> alarmMsg.alarmSevr), ntohs(currAlq -> alarmMsg.alarmAck));
	desc = currAlq -> alarmData;
	desc += (strlen(desc) + 1);
	value = desc;
	value += (strlen(value) + 1);
	egu = value;
	egu += (strlen(egu) + 1);
	printf("\tdesc = %s, val = %s, egu = %s\n\n", desc, value, egu);
	currAlq = currAlq -> nextAlq;
	}
    return 0;
}


/****************************************************************/
/*								*/
/*	PROCESS REDUNDANT ALARM QUEUE CONNECTION		*/
/*								*/
/****************************************************************/

int alqConnect(struct in_addr inetAddr)
{
    ALARM_QUE		*currQue;
    ALARM_QUE		*prevQue;
    ALARM_QUE		*nextQue;
    short		msgType;

    /*
    Traverse the alarm queue. If the Inet address for the new connection
    doesn't match, do nothing. If they do match, then check the resource
    state.
    */

    if (debugging)
    	printf("Redundant alarm connection processing entered.\n");
    prevQue = 0;
    currQue = alarmHead;
    while (currQue)
	{
	nextQue = currQue -> nextAlq;
	if (currQue -> radr -> partnerInet.s_addr == inetAddr.s_addr)
	    {
	    if (currQue -> radr -> caAnswer)
		{

		/*
		We own the resource. Send this alarm to our redundant partner.
		*/

		msgType = htons(CHAN_ALARM);
		sendRedAlarms(currQue -> radr -> redAlarmFd, msgType, (char *) currQue,
			sizeof(CH_ALARM) + ntohs(currQue -> alarmMsg.alarmDataLen));
		prevQue = currQue;
		}
	    
	    else
		{

		/*
		We don't own the resource. Delete the alarm from the queue. The other
		node will send us the alarm.
		*/

		if (prevQue)
		    prevQue -> nextAlq = currQue -> nextAlq;
		else
		    alarmHead = currQue -> nextAlq;
		if ((alarmTail = currQue))
		    {
		    if (prevQue)
			alarmTail = prevQue;
		    else
			alarmTail = alarmHead;
		    }
		currQue -> nextAlq = freeAlarms;
		freeAlarms = currQue;
		}
	    }
	else
	    prevQue = currQue;

	currQue = nextQue;
	}
    return 0;
}


/****************************************************************/
/*								*/
/*	SEND A REDUNDANT ALARM TRANSACTION TO OUR PARTNER	*/
/*								*/
/****************************************************************/

int sendRedAlarms(int fd, short msgType, char *buff, int nbytes)
{
    if (!fd)
	return -1;
    if (!raSendBytes(fd, (char *) &msgType, sizeof(msgType)))
	return -1;
    if (buff && nbytes)
	raSendBytes(fd, buff, nbytes);
    return 0;
}


/****************************************************************/
/*								*/
/*	SEND A STREAM OF BYTES TO OUR REDUNDANT ALARM PARTNER	*/
/*								*/
/*	Return a zero on failure, nonzero for success.		*/
/*								*/
/****************************************************************/

int raSendBytes(int fd, char *buff, int nbytes)
{
    int		buffSize;
    int		numSent;
    int		intSize = sizeof(int);
#ifdef PRIV_AUDIT
    SSL		*conn;
#endif PRIV_AUDIT

#ifdef PRIV_AUDIT
    conn = sslSharedGetConn (fd);
    numSent = SSL_write (conn, buff, nbytes);
#else
    numSent = send(fd, buff, nbytes, 0);
#endif PRIV_AUDIT
    if (debugging)
	{
	printf("Sent %d bytes\n", numSent);
	}
    if (numSent == ERROR)
	{
	if (errno == EWOULDBLOCK)
	    {
	    getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&buffSize, &intSize);
	    buffSize += nbytes;
	    if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&buffSize, sizeof(buffSize)) == ERROR)
		{
		if (debugging)
		    printf("Dropping red alarm connection because of maxout\n");
		raClose(fd);
		return (0);
		}
	    }
	else
	    {
	    if (debugging)
		perror("Xmit error on redundant alarm connection");
	    raClose(fd);
	    return(0);
	    }
	return (raSendBytes(fd, buff, nbytes));
	}

    else if (numSent != nbytes)
	{
	buff += numSent;
	nbytes -= numSent;
	return (raSendBytes(fd, buff, nbytes));
	}
    return (1);
}


/****************************************************************/
/*								*/
/*	CLOSE THE SOCKET TO OUR REDUNDANT ALARM PARTNER		*/
/*								*/
/*	Also clear the redAlarmFd member in every row of the	*/
/*	resource table for those entries that were using this	*/
/*	socket.							*/
/*								*/
/****************************************************************/

int raClose(int fd)
{
    RES_TABLE	*currEntry;
#ifdef PRIV_AUDIT
    SSL		*conn = sslSharedGetConn (fd);

    if (auditLog)
	{
	GetPeerDN (conn, X509_dn, sizeof(X509_dn));
	AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
			    taskName(taskIdSelf()), "Connection closed to redundant alarm");
	}
    sslSharedCloseConn (conn);
#else
    close(fd);
#endif PRIV_AUDIT

    currEntry = resTableHead;
    while (currEntry)
	{
	if (currEntry -> redAlarmFd == fd)
	    currEntry -> redAlarmFd = 0;
	currEntry = currEntry -> link;
	}
    FD_CLR(fd, &rsrvFds);
    return 0;
}


/****************************************************************/
/*								*/
/*	PROCESS AN ALARM TRANSACTION FROM REDUNDANT PARTNER	*/
/*								*/
/****************************************************************/

int raMessageProc(int fd)
{
    short	msgType;
    ALARM_QUE	*prevAlq;
    ALARM_QUE	*currAlq;

    /*
    Get the message type and process accordingly
    */

    if (!raStreamRead(fd, (char *) &msgType, sizeof(msgType)))
	return -1;
    msgType = ntohs(msgType);

    if (msgType == CHAN_ALQCONN)
	{
	RES_TABLE	*currRes;

	/*
	Connection established. Find the Inet address of this partner
	and then process the connection.
	*/

	if (debugging)
	    printf("ALARM CONNECTION\n");
	currRes = resTableHead;
	while (currRes)
	    {
	    if (currRes -> redAlarmFd == fd)
		break;
	    currRes = currRes -> link;
	    }
	if (!currRes)
	    return -1;
	alqConnect(currRes -> partnerInet);
	}

    else if (msgType == CHAN_ALARM_RECV)
	{
	struct
	    {
	    CH_ALARM_RECV		recv;
	    char			pvName[PVNAME_SZ + FLDNAME_SZ + 2];
	    } rMsg;

	/*
	A workstation issued a receipt for an alarm. It was deleted from
	the alarm queue on the master IOC. We should also delete it from
	our queue.
	*/

	if (!raStreamRead(fd, (char *) &rMsg.recv, sizeof(CH_ALARM_RECV)))
	    return -1;
	if (!raStreamRead(fd, rMsg.pvName, ntohs(rMsg.recv.alarmNameLen)))
	    return -1;

	if (debugging)
	    printf("ALARM RECEIPT...");
	prevAlq = 0;
	currAlq = alarmHead;
	while (currAlq)
	    {
	    if (currAlq -> alarmMsg.alarmSerial == rMsg.recv.alarmSerial)
	        if (ntohl(rMsg.recv.alarmTime.secPastEpoch) == ntohl(currAlq -> alarmMsg.alarmTime.secPastEpoch))
		    if (ntohl(rMsg.recv.alarmTime.nsec) == ntohl(currAlq -> alarmMsg.alarmTime.nsec))
		        if (strcmp(rMsg.pvName, currAlq -> alarmData) == 0)
			    break;
	    prevAlq = currAlq;
	    currAlq = currAlq -> nextAlq;
	    }
	if (!currAlq)
	    {
	    if (debugging)
		printf("not found!\n");
	    return -1;
	    }
	
	if (debugging)
	    printf("found!\n");
	if (prevAlq)
	    prevAlq -> nextAlq = currAlq -> nextAlq;
	else
	    alarmHead = currAlq -> nextAlq;
	if (alarmTail == currAlq)
	    {
	    if (prevAlq)
		alarmTail = prevAlq;
	    else
		alarmTail = alarmHead;
	    }
	currAlq -> nextAlq = freeAlarms;
	freeAlarms = currAlq;
	}

    else if (msgType == CHAN_ALARM)
	{
	ALARM_QUE		*destination;
	ALARM_QUE		newAlarm;
	struct dbCommon		*recHdr;

	/*
	A new alarm from our redundant partner. First, read it to keep our
	socket working in synchronization.
	*/

	if (!raStreamRead(fd, (char *) &newAlarm.alarmMsg, sizeof(CH_ALARM)))
	    return -1;
	if (!raStreamRead(fd, newAlarm.alarmData, ntohs(newAlarm.alarmMsg.alarmDataLen)))
	    return -1;
	if (debugging)
	    printf("Redundant alarm for %s\n", newAlarm.alarmData);
	
	/*
	Our queue is time ordered. Find the first alarm that is after this one in time.
	The new one goes before that one in the queue.
	*/

	prevAlq = 0;
	currAlq = alarmHead;
	while (currAlq)
	    {
	    if (ntohl(currAlq -> alarmMsg.alarmTime.secPastEpoch) >
		    ntohl(newAlarm.alarmMsg.alarmTime.secPastEpoch))
		break;

	    if ((ntohl(currAlq -> alarmMsg.alarmTime.secPastEpoch) ==
		    ntohl(newAlarm.alarmMsg.alarmTime.secPastEpoch))
		    && (ntohl(currAlq -> alarmMsg.alarmTime.nsec) >
			ntohl(newAlarm.alarmMsg.alarmTime.nsec)))
		break;
	    prevAlq = currAlq;
	    currAlq = currAlq -> nextAlq;
	    }

	/*
	Find an empty buffer for this new alarm.
	*/

	if (freeAlarms)
	     {
	     destination = freeAlarms;
	     freeAlarms = freeAlarms -> nextAlq;
	     }
	else
	    {

	    /*
	    Allocate a new queue member. It is possible that our queue is already at its
	    max size and we'll now be one too big. We'll fix that later if necessary.
	    */

	    destination = (ALARM_QUE *) calloc(1, sizeof(ALARM_QUE));
	    alarmQueSize++;
	    }

	/*
	Put this new alarm into the queue.
	*/

	if (!prevAlq)
	    {

	    /*
	    We broke out of the time search loop on the first comparison, or the alarm
	    queue was empty. Handle those situations.
	    */

	    if (currAlq)
		{

		/*
		The new alarm is older than the oldest one in our queue. Insert before
		the head.
		*/

		destination -> nextAlq = alarmHead;
		alarmHead = destination;
		}

	    else
		{

		/*
		Our queue was empty. Become both the head and the tail.
		*/

		alarmHead = destination;
		alarmTail = destination;
		destination -> nextAlq = 0;
		}
	    }

	else
	    {

	    /*
	    New alarm goes either into the middle of the queue or gets appended after the tail.
	    */

	    if (currAlq)
		{

		/*
		Insert before the one that we found by searching.
		*/

		prevAlq -> nextAlq = destination;
		destination -> nextAlq = currAlq;
		}

	    else
		{

		/*
		Append to the queue after the tail.
		*/

		alarmTail -> nextAlq = destination;
		alarmTail = destination;
		destination -> nextAlq = 0;
		}
	    }

	destination -> alarmMsg = newAlarm.alarmMsg;
	bcopy(newAlarm.alarmData, destination -> alarmData, ntohs(newAlarm.alarmMsg.alarmDataLen));
	pvdGetAddr(newAlarm.alarmData, (char **)&recHdr);
	destination -> radr = recHdr -> radr;

	/*
	If we made the queue too big, reduce it by deleting the oldest previous alarm for
	the point just inserted, or the head of the queue if there was no duplicate.
	*/

	if (alarmQueSize <= maxAlarmQue)
	    return -1;
	alarmQueSize--;
	prevAlq = 0;
	currAlq = alarmHead;
	while (currAlq != destination)
	    {
	    if (!strcmp(currAlq -> alarmData, destination -> alarmData))
		break;
	    prevAlq = currAlq;
	    currAlq = currAlq -> nextAlq;
	    }
	if ((currAlq != destination) && prevAlq)
	    {

	    /*
	    Delete oldest duplicate
	    */

	    prevAlq -> nextAlq = currAlq -> nextAlq;
	    if (alarmHead == currAlq)
		alarmHead = currAlq -> nextAlq;
	    printf("Oldest duplicate alarm for %s discarded\n", currAlq -> alarmData);
	    }

	else
	    {

	    /*
	    Delete the one at the head of the queue
	    */

	    currAlq = alarmHead;
	    alarmHead = alarmHead -> nextAlq;
	    printf("Alarm for %s dropped because of queue overflow.\n", currAlq -> alarmData);
	    }
	free(currAlq);
	}
    return 0;
}


/****************************************************************/
/*								*/
/*	READ A STREAM OF BYTES FROM OUR REDUNDANT ALARM PARTNER	*/
/*								*/
/****************************************************************/
		
int raStreamRead(int fd, char *buff, int nbytes)
{
    int			bytesGot;
    fd_set		oneFd;
    struct timeval	timeOut;
#ifdef PRIV_AUDIT
    SSL			*conn = sslSharedGetConn (fd);
#endif PRIV_AUDIT

    if (!fd)
	return (0);
    timeOut.tv_sec = 5;
    timeOut.tv_usec = 0;
    FD_ZERO(&oneFd);
    FD_SET(fd, &oneFd);

    while (nbytes)
	{
#ifdef PRIV_AUDIT
	/* Must read BEFORE select in case data is in BIO */
        if ((bytesGot = SSL_read (conn, buff, nbytes)) > 0)
	    {
	    if (debugging)
		{
		printf("Received %d bytes\n", bytesGot);
		}
	    buff += bytesGot;
	    nbytes -= bytesGot;
	    continue;
	    }
#endif PRIV_AUDIT
	if (select(fd + 1, &oneFd, NULL, NULL, &timeOut) <= 0)
	    {
	    if (debugging)
		printf("Killing RA connection due to timeout.\n");
	    raClose(fd);
	    return (0);
	    }
#ifdef PRIV_AUDIT
        if ((bytesGot = SSL_read (conn, buff, nbytes)) <= 0)
#else
	if ((bytesGot = recv(fd, buff, nbytes, 0)) <= 0)
#endif PRIV_AUDIT
	    {
	    if (debugging)
		{
		if (bytesGot < 0)
		    perror("Read error on RA connection");
		else
		    printf("Zero length read on RA connection\n");
		}
	    raClose(fd);
	    return (0);
	    }
	buff += bytesGot;
	nbytes -= bytesGot;
	}
    return (1);
}

/****************************************************************/
/*								*/
/*		VALIDATE A POINTER TO AN MCB			*/
/*								*/
/*	If the pointer isn't valid, then we should close	*/
/*	the socket and then put an impossible value in the	*/
/*	session handle. That will force chanAcq at the other	*/
/*	end of this conversation to rebuild all channels	*/
/*	with us from scratch. Print a message when this		*/
/*	happens because this is a fairly severe problem.	*/
/*								*/
/****************************************************************/

int validMCB(MCB *pMCB, CCB *pCCB, fd_set *pFdset)
{
    MCB		*currMCB;
#ifdef PRIV_AUDIT
    SSL		*conn;
#endif PRIV_AUDIT

    if (pMCB)
	{
	currMCB = pCCB -> firstChan;
	while (currMCB)
	    {
	    if (currMCB == pMCB)
		return (1);
	    currMCB = currMCB -> nextClient;
	    }
	}

    /*
    If we get here, then the pointer is invalid. Take corrective
    action and then notify the world that this has happened.
    */

    if (pCCB -> clientFd)
	{
#ifdef PRIV_AUDIT
	conn = sslSharedGetConn (pCCB -> clientFd);
	if (auditLog)
	    {
	    GetPeerDN (conn, X509_dn, sizeof(X509_dn));
	    AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
			   taskName(taskIdSelf()), "Connection closed invalid MCB");
	    }
	sslSharedCloseConn (conn);
#else
	close(pCCB -> clientFd);
#endif PRIV_AUDIT
	FD_CLR(pCCB -> clientFd, &rsrvFds);
	if (pFdset)
	    FD_CLR(pCCB -> clientFd, pFdset);
	pCCB -> clientFd = 0;
	pCCB -> handlerMsgRcvd = 0;
	pCCB -> alarmHandler = 0;
	pCCB -> trendHandler = 0;
	}
    pCCB -> clientSession = 0;
    logMsg("RSRV: Invalid handle from %s\n", (int)INET_NTOA(pCCB -> clientInet),0,0,0,0,0);
    return (0);
}

/****************************************************************/
/*								*/
/*		    SET THE RSRV SOCKET TIMEOUT			*/
/*								*/
/*	If we haven't heard anything from sockMnt within	*/
/*	rsrvRxTimeout seconds, we consider the socket dead	*/
/*	for anyTrendHandlers() or anyAlarmHandlers purposes	*/
/*	regardless of the REAL state of the socket. This	*/
/*	facility has NO influence on how sockets are treated	*/
/*	within the rest of the system.				*/
/*								*/
/****************************************************************/

int setRsrvTimeout(int seconds)
{
    printf("Rsrv Timeout was %d, changed to %d seconds.\n", rsrvRxTimeout / 60, seconds);
    rsrvRxTimeout = seconds * 60;
    return (0);
}
