/*snlmonlib.c,v 1.1.1.1 1999/08/03 14:44:38 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#ifdef HOST
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <malloc.h>
#  ifdef WIN32
#   include <winsock.h>
#   include <fcntl.h>
#   include <errno.h>
#   include "NTUnixSubs.h"
#  else
#   include <sys/types.h>
#   include <sys/socket.h>
#   include <netinet/in.h>
#   include <netdb.h>
#   include <errno.h>
#   ifndef AXP
#      include <sys/filio.h>
#   endif AXP
#   include <sys/time.h>
#   include <sys/ioctl.h>
#  endif WIN32
#include "msgMon.h"
#define	TRUE	1
#define FALSE	0
#define	printErr	printf
#define	taskSuspend	exit
#ifndef	WIN32
#define	SOCKET		int
#define	INVALID_SOCKET	-1
#define	SOCKET_ERROR	-1
#endif	!WIN32
#else
#include <vxWorks.h>
#endif HOST

#include <string.h>
#include "mon_defs.h"

static	SOCKET	Myfd;
static	int	connectPending;
static  int	FdConnected;
static	char	HostName [132];
static	int	DebugOn = 1;
static	long	curSeq;
static	long	nextRxSeq;

#define MAXRXTIME  10 /* 10 seconds to receive the reply message from IOC */
static  int	rxCharsIn;
static  MSG_HEADER rxHdr;
static	char	*rxBody;

/* Functions defined this module */

int	sendMessage ();
int	readreply ();


/*-----------*/
int snlMonConnect (char *Host)
{
    struct hostent	*hostptr;
    struct sockaddr_in	newaddress;
    fd_set		wrSet;
    fd_set		exSet;
    struct timeval	timeOut;
    int			status;
    int			true = TRUE;
    int			error;

    if(DebugOn)
	printErr ("snlMonConnect: FdConnected=%d, connectPending=%d\n",
		   FdConnected, connectPending);

    /* if already connected, just return true */
    if (FdConnected)
	return 1;

    /* If a connect is pending, check for connection complete. */
    if (connectPending) {
	FD_ZERO (&wrSet);
	FD_SET (Myfd, &wrSet);
	FD_ZERO (&exSet);
	FD_SET (Myfd, &exSet);
	timeOut.tv_sec = 5;
	timeOut.tv_usec = 0;
	status = select(Myfd + 1, NULL, &wrSet, &exSet, &timeOut);
	if(DebugOn)
	    printErr("snlMonConnect: select, status=%d\n", status);
	if (status != 0) {
	    if (FD_ISSET (Myfd, &wrSet)) {
		if(DebugOn)
		    printErr("snlMonConnect:connectPending,FD_ISSET wrSet\n");
		FdConnected = TRUE;
		rxCharsIn = 0;
		curSeq = time(NULL);
		if (rxBody)
		    free (rxBody);
		return 1;
	    }
	    else { /* error occurred! */
		return 0;
	    }
	}
	else {
	    return 0;
	}
    }

    /* remember name for reconnects */
    if (Host)
        strcpy (HostName, Host);

    /* Look up the address of the server host */

    if ((hostptr = gethostbyname (HostName)) == NULL) {
	if (DebugOn)
	    printErr ("snlMonConnect: gethostbyname failed\n");
	return 0;
    }
  
    if((Myfd = socket (AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
    {
	if (DebugOn)
            printErr("snlMonConnect: connect socket creation error\n");
#ifdef	HOST
#ifdef	WIN32
	error = WSAGetLastError ();
#else
	error = errno;
#endif	WIN32
	tisSysLog (0,0, "snlMonConnect: socket creation error %d\n",error);
#endif	HOST
        taskSuspend (0);
    }

    if ((setsockopt(Myfd, SOL_SOCKET, SO_KEEPALIVE, (char *)&true, sizeof(true))) == SOCKET_ERROR)
    {
#ifdef	HOST
#ifdef	WIN32
	error = WSAGetLastError ();
#else
	error = errno;
#endif	WIN32
	tisSysLog (0,0, "snlMonConnect: setsockopt error %d\n",error);
#endif	HOST
	taskSuspend (0);
    }

    memset (&newaddress, 0, sizeof newaddress);
    newaddress.sin_family = AF_INET;
    newaddress.sin_addr.s_addr =
	htonl(((struct in_addr *) *(hostptr->h_addr_list))->s_addr);
    newaddress.sin_port = htons(SNLMON_PORT);
#ifdef	WIN32
    ioctlsocket (Myfd, FIONBIO, &true);
#else
    ioctl(Myfd, FIONBIO, &true);
#endif	WIN32
    if ((connect(Myfd, (struct sockaddr *)&newaddress, sizeof(newaddress))) == SOCKET_ERROR)
    {
#ifdef	WIN32
	error = WSAGetLastError ();
	if ((error == WSAEINPROGRESS) || (error == WSAEWOULDBLOCK))
#else
	error = errno;
	if ((error == EINPROGRESS) || (error == EWOULDBLOCK))
#endif	WIN32
	{
	    if (DebugOn)
		printErr ("snlMonConnect: connect, errno = %d,return 0\n",error);
	    connectPending = 1;
	    return 0;
	}
	else
	{
	    if (DebugOn)
		printErr ("snlMonConnect: connect, errno = %d,return 0\n", error);
#ifdef	WIN32
	    closesocket (Myfd);
#else
            close (Myfd);
#endif	WIN32
	    return 0;
	}
    }
    else
    {
	if(DebugOn)
	    printErr ("snlMOnConnect: connected\n");
	FdConnected = TRUE;
	connectPending = 0;
	rxCharsIn = 0;
	curSeq = time(NULL);
	if (rxBody)
	    free (rxBody);
	return 1;
    }
}


/*----------*/
void snlMonDisconnect (void)
{
#ifdef	WIN32
    closesocket (Myfd);
#else
    close (Myfd);
#endif	WIN32
    Myfd = 0;
    connectPending = 0;
    FdConnected = FALSE;
}


/*----------*/
int sendMessage (int Type, char *Body, int Length, int Info)
{
    MSG_HEADER	msgHeader;
    int		i;

    if (FdConnected == FALSE)
    {
	snlMonConnect (NULL);
	if (FdConnected) {
	    if (DebugOn)
		printf ("Reconnected\n");
	}
	else
	    return 0;
    }

    if (DebugOn) {
	printf ("Sending message: Type=%d, Length=%d, Info=%d, Msg=", Type, Length, Info);
	for (i=0 ; i<Length ; i++)
	    printf ("%x ", *(Body + i));
	printf ("\n");
    }

    /* Send the message header */

    msgHeader.Type = Type;
    nextRxSeq = curSeq++;
    msgHeader.seq = nextRxSeq;
    msgHeader.Length = Length;
    msgHeader.Info = Info;
    if ((send (Myfd, (char *)&msgHeader, sizeof (msgHeader), 0)) == SOCKET_ERROR)
    {
	snlMonDisconnect ();
	if (DebugOn)
	    printf ("Disconnected: %d\n", errno);
	return 0;
    }

    /* Send the message body */

    if (Length > 0)
    {
	if ((send (Myfd, Body, Length, 0)) == SOCKET_ERROR)
	{
	    snlMonDisconnect ();
	    if (DebugOn)
		printf ("Disconnected: %d\n", errno);
	    return 0;
	}
    }

    return 1;
}


/*----------*/
int readReply (char **Reply, int *Size, int *Info)
{
    int		i;
    int		bytesGot;
    long	startTime;
    fd_set	wrSet;
    struct timeval timeOut;
    int		status;
    int		error;

    /* Prepare for the worst */
    if (Reply)
	*Reply = NULL;
    if (Size)
	*Size = 0;
    if (Info)
	*Info = 0;

    if (FdConnected == FALSE) {
	return 0;
    }

    startTime = time (NULL);

    /* get the message matching the last send */
    while ((rxCharsIn < sizeof(rxHdr))
      || ((rxCharsIn >= sizeof (rxHdr)) && (rxHdr.seq != nextRxSeq)))
    {
	while (rxCharsIn < sizeof (rxHdr))
	{
	    /* Wait for data or timeout */
	    FD_ZERO (&wrSet);
	    FD_SET (Myfd, &wrSet);
	    timeOut.tv_sec = (startTime + MAXRXTIME) - time (NULL);
	    timeOut.tv_usec = 0;
	    status = select(Myfd + 1, NULL, &wrSet, NULL, &timeOut);
	    if (status != 0)
	    {
		bytesGot = recv (Myfd, ((char *)&rxHdr) + rxCharsIn, sizeof(rxHdr)-rxCharsIn, 0);
		if (bytesGot < 0)
		{
#ifdef	WIN32
		    error = WSAGetLastError ();
		    if (error != WSAEWOULDBLOCK)
#else
		    error = errno;
		    if (error != EWOULDBLOCK)
#endif	WIN32
		    {
			snlMonDisconnect ();
			if (DebugOn)
			    printf ("Disconnected: %d\n", errno);
			return 0;
		    }
		    bytesGot = 0;
		}
	    }
	    if ((startTime + MAXRXTIME) < time (NULL))
	    {
		if (DebugOn)
		    printf ("Timeout waiting for reply\n");
		return 0;
	    }
	    rxCharsIn += bytesGot;
	}

	/* Read the message body if needed */
	while (rxCharsIn < (sizeof (rxHdr) + rxHdr.Length))
	{
	    if (rxBody == NULL)
		rxBody = (char *) malloc (rxHdr.Length);

	    /* Wait for data or timeout */
	    FD_ZERO (&wrSet);
	    FD_SET (Myfd, &wrSet);
	    timeOut.tv_sec = (startTime + MAXRXTIME) - time (NULL);
	    timeOut.tv_usec = 0;
	    status = select(Myfd + 1, NULL, &wrSet, NULL, &timeOut);
	    if (status != 0)
	    {
	        bytesGot = recv (Myfd, rxBody + (rxCharsIn - sizeof(rxHdr)),
				rxHdr.Length - (rxCharsIn - sizeof(rxHdr)), 0);
		if (bytesGot < 0)
		{
#ifdef	WIN32
		    error = WSAGetLastError ();
		    if (error != WSAEWOULDBLOCK)
#else
		    error = errno;
		    if (error != EWOULDBLOCK)
#endif	WIN32
		    {
			snlMonDisconnect ();
			if (DebugOn)
			    printf ("Disconnected: %d\n", errno);
			return 0;
		    }
		    bytesGot = 0;
		}
	    }
	    if ((startTime + MAXRXTIME) < time (NULL))
	    {
		if (DebugOn)
		    printf ("Timeout waiting for reply\n");
		return 0;
	    }
	    rxCharsIn += bytesGot;
	}

	/* we now have the whole message, if it isn't what we want
	** discard it and try again */
	if (rxHdr.seq != nextRxSeq)
	{
	    if (DebugOn)
		printf ("Discarded out of sequence message of %d bytes\n", rxCharsIn);
	    rxCharsIn = 0;
	    if (rxBody)
		free (rxBody);
	}
    }

    if (DebugOn)
	printf ("Reply received: Type = %d, Length = %d, Info = %d\n",
			rxHdr.Type, rxHdr.Length, rxHdr.Info);

    if (DebugOn)
    {
	if (rxBody != NULL)
	{
	    printf ("Message body: ");
	    for (i=0 ; i<rxHdr.Length ; i++)
		printf ("%x ", rxBody [i]);
	    printf ("\n");
	}
    }

    if (Reply != NULL)
	*Reply = rxBody;
    else
	free (rxBody);
    rxBody = NULL;
    if (Size != NULL)
	*Size = rxHdr.Length;
    if (Info != NULL)
	*Info = rxHdr.Info;

    rxCharsIn = 0;
    return 1;
}


/****************************************************************/
/*	Commands to the server					*/
/****************************************************************/

/*----------*/
int snlMon_Ping (int sendValue, int *receiveValue)
{
    *receiveValue = 0;

    if (!sendMessage (PING_MSG_TYPE, NULL, 0, sendValue))
	return 0;

    return readReply (NULL, NULL, receiveValue);
}

/*----------*/
int snlMon_Load (char *ModuleName)
{
    LOAD_MSG	loadMsg;
    int		Info;

    strcpy (loadMsg.ModuleName, ModuleName);
    if (!sendMessage (LOAD_MSG_TYPE, (char *)&loadMsg, sizeof (loadMsg), 0))
	return 0;

    if (!readReply (NULL, NULL, &Info))
	return 0;
    	
    return Info;
}

/*----------*/
int snlMon_Start (char *ProgramName, char *Parameters, int StackSize, 
		  char *Resource, char *ActualName)

{
    START_MSG	startMsg;
    START_REPLY	*startReply;
    int		Info;

    strcpy (startMsg.ProgramName, ProgramName);
    strcpy (startMsg.Parameters, Parameters);
    startMsg.StackSize = StackSize;
    strcpy (startMsg.Resource, Resource);
    if (!sendMessage (START_MSG_TYPE, (char *)&startMsg, sizeof (startMsg), 0))
	return 0;

    if (!readReply ((char **)&startReply, NULL, &Info))
	return 0;
    if (!Info || !startReply)
	return 0;
    strcpy (ActualName, startReply->ProgramName);
    free (startReply);
    return 1;
}


/*----------*/
int snlMon_Kill (char *ProgramName, long Tid, char *Parameters)
{
    KILL_MSG	killMsg;
    int		Info;

    strcpy (killMsg.ProgramName, ProgramName);
    killMsg.TaskId = Tid;    
    strcpy (killMsg.Parameters, Parameters);
    if (!sendMessage (KILL_MSG_TYPE, (char *)&killMsg, sizeof (killMsg), 0))
	return 0;

    if (!readReply (NULL, NULL, &Info))
	return 0;

    return Info;
}

/*----------*/
int snlMon_Unload (char *ModuleName)
{
    UNLOAD_MSG	unloadMsg;
    int		Info;

    strcpy (unloadMsg.ModuleName, ModuleName);
    if (!sendMessage (UNLOAD_MSG_TYPE, (char *)&unloadMsg, sizeof (unloadMsg), 0))
	return 0;

    if (!readReply (NULL, NULL, &Info))
	return 0;
    return Info;
}

/*----------*/
int snlMon_Terminate ()
{
    int		Info;

    if (!sendMessage (TERMINATE_MSG_TYPE, NULL, 0, 0))
	return 0;

    if (!readReply (NULL, NULL, &Info))
	return 0;

    return Info;
}


/*----------*/
int snlMon_ReportModules (REPMOD_REPLY **ModList, int *NumModules)
{
    if (!sendMessage (REPMOD_MSG_TYPE, NULL, 0, 0))
	return 0;

    return readReply ((char **)ModList, NULL, NumModules);
}

/*----------*/
int snlMon_ReportPrograms (REPPROG_REPLY **ProgList, int *NumPrograms)
{
    if (!sendMessage (REPPROG_MSG_TYPE, NULL, 0, 0))
	return 0;

    return readReply ((char **)ProgList, NULL, NumPrograms);
}

/*----------*/
int snlMon_ReportStateSets (char *ProgramName, long Tid, REPSS_REPLY **SSList,
			    int *NumStateSets)
{
    REPSS_MSG		repssMsg;

    strcpy (repssMsg.ProgramName, ProgramName);
    repssMsg.TaskId = Tid;
    if (!sendMessage (REPSS_MSG_TYPE, (char *)&repssMsg,
			sizeof (REPSS_MSG), 0))
	return 0;

    return readReply ((char **)SSList, NULL, NumStateSets);
}


/*----------*/
int snlMon_ReportChannels (char *ProgramName, long Tid, REPCHANS_REPLY **ChanList,
			   int *NumChans)
{

    REPCHANS_MSG	repchansMsg;

    strcpy (repchansMsg.ProgramName, ProgramName);
    repchansMsg.TaskId = Tid;
    if (!sendMessage (REPCHANS_MSG_TYPE, (char *)&repchansMsg,
			sizeof (REPCHANS_MSG), 0))
	return 0;

    return readReply ((char **)ChanList, NULL, NumChans);
}

/*----------*/
int snlMon_ReportVariables (char *ProgramName, long Tid, REPVARS_REPLY **VarList,
			    int *NumVars)
{

    REPVARS_MSG		repvarsMsg;

    strcpy (repvarsMsg.ProgramName, ProgramName);
    repvarsMsg.TaskId = Tid;
    if (!sendMessage (REPVARS_MSG_TYPE, (char *)&repvarsMsg,
			sizeof (REPVARS_MSG), 0))
	return 0;

    return readReply ((char **)VarList, NULL, NumVars);
}

/*----------*/
int snlMon_ReportVariable (char *ProgramName, long Tid, char *VarName,
			   char **VarValue, int *VarLength)
{

    REPVAR_MSG		repvarMsg;

    strcpy (repvarMsg.ProgramName, ProgramName);
    repvarMsg.TaskId = Tid;
    strcpy (repvarMsg.VariableName, VarName);
    if (!sendMessage (REPVAR_MSG_TYPE, (char *)&repvarMsg,
			sizeof (REPVAR_MSG), 0))
	return 0;

    return readReply (VarValue, VarLength, NULL);
}


/*----------*/
int snlMon_ChangeVariable (char *ProgramName, long Tid, char *VarName,
			   char *VarValue, int VarLength)
{

    char	*chvarMsg;
    int		Info;

    chvarMsg = (char *) malloc (sizeof (CHVAR_MSG) + VarLength);

    strcpy (((CHVAR_MSG *)chvarMsg)->ProgramName, ProgramName);
    ((CHVAR_MSG *)chvarMsg)->TaskId = Tid;
    strcpy (((CHVAR_MSG *)chvarMsg)->VariableName, VarName);
    memcpy (chvarMsg + sizeof (CHVAR_MSG), VarValue, VarLength);

    if (!sendMessage (CHVAR_MSG_TYPE, chvarMsg,
			sizeof (CHVAR_MSG) + VarLength, 0))
	return 0;

    free (chvarMsg);

    if (!readReply (NULL, NULL, &Info))
	return 0;

    return Info;
}

/*----------*/
int snlMon_Suspend (char *ProgramName, long Tid, char *TaskName)
{
    SUSPEND_MSG	suspendMsg;
    int		Info;

    strcpy (suspendMsg.ProgramName, ProgramName);
    suspendMsg.TaskId = Tid;
    strcpy (suspendMsg.TaskName, TaskName);    
    if (!sendMessage (SUSPEND_MSG_TYPE, (char *)&suspendMsg, sizeof (suspendMsg), 0))
	return 0;

    if (!readReply (NULL, NULL, &Info))
	return 0;

    return Info;
}

/*----------*/
int snlMon_Resume (char *ProgramName, long Tid, char *TaskName)
{
    RESUME_MSG	resumeMsg;
    int		Info;

    strcpy (resumeMsg.ProgramName, ProgramName);
    resumeMsg.TaskId = Tid;
    strcpy (resumeMsg.TaskName, TaskName);    
    if (!sendMessage (RESUME_MSG_TYPE, (char *)&resumeMsg, sizeof (resumeMsg), 0))
	return 0;

    if (!readReply (NULL, NULL, &Info))
	return 0;

    return Info;
}

