/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
   File:   opprivdata.c - oppriv data server, providing access
		 to the operator tracking & priviledges for
		 remote clients.
   Author: Grant Irani
   Date:   3/17/96 13:21
*/

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
    #include "cb45to63.h"
#endif
    #include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>

#if !(defined(MIPSEL) || defined(AXP))
#include <memory.h>
#ifdef WIN32
#include <sys/types.h>
#include <sys/stat.h>
#endif WIN32
#else !(defined(MIPSEL) || defined(AXP))
#include <sys/types.h>
#include <sys/stat.h>
extern int lstat ();
extern int readlink ();
#endif !(defined(MIPSEL) || defined(AXP))
#ifndef WIN32
#include <pwd.h>
#endif WIN32

#include "server.h"
#include "dbDefs.h"
#include "OpPrivData.h"
#include "OperPriv.h"
#include "msgMon.h"

/* Local routines */
int GetLogResp (LogonMsg *reqmsg);
int GetCheckPrivResp (PrivMsg *reqmsg);
int GetOpPrivResp (PrivMsg *reqmsg);
int GetTrackResp (TrackMsg *reqmsg);
int GetOpenCloseTrackResp (OpenCloseTrackMsg *reqmsg);
int CloseOperPrivResp (OpenClosePrivMsg *reqmsg);


SERVER *server;		/* the server control structure */

#define LOGGEDOFF	0
#define LOGGEDON	1
int	curstatus = LOGGEDOFF;
int	loginApp = FALSE;
char	curoperator [OPERATOR_SZ];
#ifdef WIN32
#else WIN32
uid_t	saved_uid;
gid_t	 saved_gid;
#endif WIN32
char	curstation [80];


char	setString [132];	/* env setting space */

RespUnion    rspmsg;

SERVER *server;

int main (argc, argv)
int	argc;
char**	argv;
{
    ReqUnion	*reqmsg;
    short msglen;
    static char tmpstring [132];

#ifdef WIN32
    WSADATA wsaData; 
    WORD    wVersionRequested = MAKEWORD(1, 1);
    int	err;

    err = WSAStartup(wVersionRequested, &wsaData); 
 
    if (err != 0)
    {
	/* Tell the user that we couldn't find a useable */ 
	/* winsock.dll.     */ 
	tisSysLog (0, 0, "TIS Server library can't open winSock");
	exit (1);
    }
#endif WIN32

    /* open server as synchronous I/O */
    if ((server = OpenTisService ("OpPrivData", TRUE, NULL)) == NULL)
	exit (0);

    tisSysLog (0, 2, "OpPrivData server running");

    sprintf (setString, "DISPLAY=nohost:0.0");
    putenv (setString);
#ifdef DEBUG
    SetSocketNum (server, 8002);
#endif DEBUG

    while (reqmsg = GetReq (server))
    {
	((PrivMsg *)reqmsg)->type = ntohs (((PrivMsg *)reqmsg)->type);
#ifndef DEBUG
	tisSysLog (0, 2, "Processing request %d", ((PrivMsg *)reqmsg)->type);
#endif
	switch (((PrivMsg *)reqmsg)->type)
	{
	    case OP_LOGON:
	    case OP_LOGOFF:
	    case OP_LOGCHECK:
		msglen = GetLogResp ((LogonMsg *)reqmsg);
		break;
	    case OP_CHECK_PRIV:
		msglen = GetCheckPrivResp ((PrivMsg *)reqmsg);
		break;
	    case OP_PRIV:
		msglen = GetOpPrivResp ((PrivMsg *)reqmsg);
		break;
	    case WRITE_TRACK:
		msglen = GetTrackResp ((TrackMsg *)reqmsg);
		break;
	    case OPEN_TRACK:
	    case CLOSE_TRACK:
		msglen = GetOpenCloseTrackResp ((OpenCloseTrackMsg *)reqmsg);
		break;
	    case CLOSE_PRIV:
		msglen = CloseOperPrivResp ((OpenClosePrivMsg *)reqmsg);
		break;
	}
	SendRsp (server, (u_char *)&rspmsg, msglen);
	free (reqmsg);
    }
#ifndef DEBUG
    tisSysLog (0, 2, "Null message from GetReq");
#endif
    /* if login app, logout now! */
    if (curstatus == LOGGEDON && loginApp)
    {
	sprintf (tmpstring, "/tis4000/operators/logins/%s", getDisplayHost ());
	unlink (tmpstring);
    }

    /* cleanup any open files, etc... */
    CloseSvcSession (server);
    tisCloseLog ();
    return 0;
}

#ifdef WIN32
/* local pw authentication, to prevent requiring pwdauthd to be running */
static int pwdauth (char *login, char *passwd)
{
    /* succeeded in logging in */
    return 0;

/*    if ((pwent = getpwnam(login)) != 0)
    {
	if ((VerifyPassword (login, passwd)) == 0)
	{
	    saved_uid = getuid ();
	    saved_gid = getgid ();
	    seteuid (pwent->pw_uid);
	    setegid (pwent->pw_gid);
	    return 0;
	}
    }*/
    /* Failed to log in */
    return 1;
}


/* call to set impersonation (for priviledges, etc... */
static int setpwduid (char *login)
{
    return 0;
}
#else WIN32
/* local pw authentication, to prevent requiring pwdauthd to be running */
static int pwdauth (char *login, char *passwd)
{
#ifdef SOLARIS
#include <shadow.h>
    struct spwd *spw;
#else SOLARIS
    extern char *crypt (char *key, char *salt);
#endif SOLARIS
    struct passwd *pwent;

#ifdef SOLARIS
    if ((spw = getspnam(login)) != 0)
    {
#endif SOLARIS
	if ((pwent = getpwnam(login)) != 0)
	{
	    if ((VerifyPassword (login, passwd)) == 0)
	    {
	 	saved_uid = getuid ();
		saved_gid = getgid ();
		seteuid (pwent->pw_uid);
		setegid (pwent->pw_gid);
		return 0;
	    }
	}
#ifdef SOLARIS
    }
#endif SOLARIS
    return 1;
}


/* call to set impersonation (for priviledges, etc... */
static int setpwduid (char *login)
{
    struct passwd *pwent;

    if ((pwent = getpwnam(login)) != 0)
    {
#ifdef SOLARIS
	setuid (pwent->pw_uid);
	setgid (pwent->pw_gid);
#else
	setreuid (pwent->pw_uid, pwent->pw_uid);
	setregid (pwent->pw_gid, pwent->pw_gid);
#endif SOLARIS
	return 0;
    }
    return 1;
}
#endif WIN32

int GetLogResp (reqmsg)
LogonMsg *reqmsg;
{
    LogonRsp	*logrsp = (LogonRsp *)&rspmsg;
    char	*opPtr = reqmsg->sysname + strlen(reqmsg->sysname) + 1;
    char	*passPtr = opPtr + strlen(opPtr) + 1;
    static char tmpstring [132];
    FILE	*f;
#ifdef WIN32
    struct _stat	status;
#else WIN32
    struct stat	status;
#endif WIN32

    /* set DISPLAY environment var to satisfy operPrvTrack.o */
    sprintf (setString, "DISPLAY=%s", reqmsg->sysname);
    tisSysLog (0,2,"GetLogResp, set DISPLAY=%d",setString);
    putenv (setString);

    logrsp->type = htons (OP_LOGON);
    sprintf (tmpstring, "/tis4000/operators/logins/%s", getDisplayHost ());
#ifdef WIN32
    if (_stat (tmpstring, &status) == 0)
#else WIN32
    if (stat (tmpstring, &status) == 0)
#endif WIN32
	curstatus = LOGGEDON;
    else
	curstatus = LOGGEDOFF;

    tisSysLog (0, 2, "GetLogResp, curstatus = %d of %s",curstatus,tmpstring);
    if (reqmsg->type == OP_LOGON)
    {
	loginApp = TRUE;
	if (curstatus == LOGGEDON)
	{
	    /* already logged on! */
	    logrsp->value = htons ((unsigned short)PRIVNOCODE);
	}
	else if (pwdauth(opPtr, passPtr) == 0)
	{
	    /* record opname in /tis4000/operators/logins/<station> */
	    if (f = fopen (tmpstring, "w"))
	    {
		tisSysLog (0, 2, "GetLogResp, Opened %s", tmpstring);
		fputs (opPtr, f);
		fputs ("\n", f);
		fclose (f);
		/* OPEN OP PRIV DBS */
		logrsp->value = htons ((unsigned short)InitOperPriv (curoperator));
		if (ntohs(logrsp->value) != PRIVOK)
		{
		    tisSysLog (0,2,"GetLogResp, InitOperPriv failed");
		    unlink (tmpstring);
		}
		else
		{
		    strncpy (curoperator, opPtr, OPERATOR_SZ);
		    strncpy (curstation, reqmsg->sysname, 80);
		    curstatus = LOGGEDON;
		}
	    }
	    else
	    {
		logrsp->value = htons ((unsigned short)PRIVNOCODE);
	    }
	}
	else
	{
	    /* priviledge error! */
	    logrsp->value = htons ((unsigned short)PRIVNOCODE);
	}
    }
    else if (reqmsg->type == OP_LOGCHECK)
    {
	if (curstatus != LOGGEDON)
	{
	    /* NOT logged on! */
	    logrsp->value = htons ((unsigned short)PRIVNOCODE);
	}
	else
	{
	    /* compare opname in /tis4000/operators/logins/<station> */
	    if (f = fopen (tmpstring, "r"))
	    {
		fgets (tmpstring, 132, f);
		fclose (f);
		/* trim newline from input */
		if (strchr (tmpstring, '\n'))
		    *strchr (tmpstring, '\n') = 0;
		if ((strcmp (opPtr, tmpstring) == 0)
		  && (setpwduid (opPtr) == 0))
		{
		    /* OPEN OP PRIV DBS */
		    logrsp->value = htons ((unsigned short)InitOperPriv (curoperator));
		    if (ntohs(logrsp->value) == PRIVOK)
		    {
			strncpy (curoperator, opPtr, OPERATOR_SZ);
			strncpy (curstation, reqmsg->sysname, 80);
		    }
		}
		else
		{
		    logrsp->value = htons ((unsigned short)PRIVNOCODE);
		}
	    }
	    else
	    {
		logrsp->value = htons ((unsigned short)PRIVNOCODE);
	    }
	}
    }
    else
    {
	logrsp->type = htons (OP_LOGOFF);
	if (curstatus == LOGGEDOFF)
	{
	    /* already logged off! */
	    logrsp->value = htons ((unsigned short)PRIVNOCODE);
	}
	else
	{
	    unlink (tmpstring);
#ifdef WIN32
#else WIN32
	    /* restore startup user id */
	    setreuid (saved_uid, saved_uid);
	    setregid (saved_gid, saved_gid);
#endif WIN32
	    /* CLOSE OP PRIV DBS */
	    strncpy (curoperator, "nohost", OPERATOR_SZ);
	    strncpy (curstation, "nohost", 80);
	    logrsp->value = htons (PRIVOK);
	    curstatus = LOGGEDOFF;
	}
    }

    return (sizeof(PrivRsp));
}

int GetCheckPrivResp (reqmsg)
PrivMsg *reqmsg;
{
    char	**errMsgPtr;
    char	*dp;

    PrivRsp	*privrsp = (PrivRsp*)&rspmsg;
    privrsp->type = htons (OP_CHECK_PRIV);
    privrsp->value = htons ( (unsigned short)CheckOperPriv (reqmsg->priv));
    if (ntohs (privrsp->value) != PRIVOK)
    {
	/* on errors, return msg text as well */
	OperPrivTrackErr (ntohs (privrsp->value), &errMsgPtr);
	/* this is an array of strings */
	dp = privrsp->errString;
	while (*errMsgPtr)
	{
	    strncpy (dp, *errMsgPtr, (privrsp->errString + 511) - dp);
	    dp = dp + strlen(*errMsgPtr) + 1;
	    errMsgPtr++;
	}
	*dp++ = 0;
	return (sizeof(PrivRsp) - 512 + (dp - privrsp->errString));
    }
    else
    {
	*privrsp->errString = 0;
	return (sizeof(PrivRsp) - 512 + 1);
    }
}

int GetOpPrivResp (reqmsg)
PrivMsg *reqmsg;
{
    PrivRsp	*privrsp = (PrivRsp*)&rspmsg;
    privrsp->type = htons (OP_PRIV);
    privrsp->value = htons ( (unsigned short)OperPriv (NULL, reqmsg->priv));
    return (sizeof(PrivRsp) - 512);
}

int GetTrackResp (reqmsg)
TrackMsg *reqmsg;
{
    char	*oldvaluePtr = reqmsg->channel + strlen(reqmsg->channel) + 1;
    char	*newvaluePtr = oldvaluePtr + strlen(oldvaluePtr) + 1;
    char	**errMsgPtr;
    char	*dp;

    TrackRsp	*trackrsp = (TrackRsp*)&rspmsg;
    trackrsp->type = htons (WRITE_TRACK);
    trackrsp->value = htons ( (unsigned short)WriteOperTracking (reqmsg->channel,
						 oldvaluePtr,
						 newvaluePtr));
    dp = trackrsp->errString;
    if (ntohs (trackrsp->value) != PRIVOK)
    {
	/* on errors, return msg text as well */
	OperPrivTrackErr (ntohs (trackrsp->value), &errMsgPtr);
	/* this is an array of strings */
	while (*errMsgPtr)
	{
	    strncpy (dp, *errMsgPtr, (trackrsp->errString + 511) - dp);
	    dp = dp + strlen(*errMsgPtr) + 1;
	    errMsgPtr++;
	}
    }
    *dp++ = 0;
    return (sizeof(TrackRsp) - 512 + (dp - trackrsp->errString));
}

int GetOpenCloseTrackResp (reqmsg)
OpenCloseTrackMsg *reqmsg;
{
    OpenCloseTrackRsp *ocrsp = (OpenCloseTrackRsp *)&rspmsg;
    char	**errMsgPtr;
    char	*dp;

    ocrsp->value = htons ((unsigned short)PRIVNOCODE);
    if (reqmsg->type == OPEN_TRACK)
    {
	if (curstatus == LOGGEDON)
	{
	    /* Open operator tracking */
	    ocrsp->value = htons ((unsigned short)InitOperTracking( NULL ));
	}
    }
    else if (reqmsg->type == CLOSE_TRACK)
    {
	if (curstatus == LOGGEDON)
	{
	    /* Close operator tracking */
	    ocrsp->value = htons ((unsigned short)CloseOperTracking( NULL ));
	}
    }

    if (ntohs (ocrsp->value) != PRIVOK)
    {
	/* on errors, return msg text as well */
	OperPrivTrackErr (ntohs (ocrsp->value), &errMsgPtr);
	/* this is an array of strings */
	dp = ocrsp->errString;
	while (*errMsgPtr)
	{
	    strncpy (dp, *errMsgPtr, (ocrsp->errString + 511) - dp);
	    dp = dp + strlen(*errMsgPtr) + 1;
	    errMsgPtr++;
	}
	*dp++ = 0;
	return (sizeof(OpenCloseTrackRsp) - 512 + (dp - ocrsp->errString));
    }
    else
    {
	*ocrsp->errString = 0;
	return (sizeof(OpenCloseTrackRsp) - 512 + 1);
    }
}

int CloseOperPrivResp (reqmsg)
OpenClosePrivMsg *reqmsg;
{
    OpenClosePrivRsp *ocrsp = (OpenClosePrivRsp *)&rspmsg;

    ocrsp->value = htons ((unsigned short)PRIVNOCODE);
    if (curstatus == LOGGEDON)
    {
	/* Close operator privileges */
	ocrsp->value = htons ((unsigned short)CloseOperPriv());
    }
    return (sizeof (OpenClosePrivRsp));
}
