/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
    opprivdata.h - definitions for the oppriv data client (and task mgmt)
    Author: Grant Irani
    Date:   March 21, 1996
*/


#define	    OP_LOGON	    1		/* Set op priv security access */
#define	    OP_LOGOFF	    2		/* Log off tis4000 */
#define	    OP_LOGCHECK	    3		/* Logon check tis4000 */
#define	    OP_CHECK_PRIV   4		/* Check priviledges (CheckOperPriv())*/
#define	    OP_PRIV	    5		/* Check priviledges (OperPriv()) */
#define	    WRITE_TRACK	    6		/* Write tracking message (WriteOperTracking()) */
#define	    OPEN_TRACK	    7		/* Open track access */
#define	    CLOSE_TRACK	    8		/* Close track access */
#define	    CLOSE_PRIV	    9		/* Close priviledge access */

#define	    APP_OPEN	    100		/* App name registration */
#define	    APP_EXIT	    101		/* App exiting */
#define	    APP_KILL	    102		/* Terminate an application */
#define	    APP_LIST	    103		/* Get list of apps */

/* Application function status values */
#define	    APP_OK	    0		/* app operation ok */
#define	    APP_DUP	    0		/* app operation ok, multi instance */
#define	    APP_ERROR	    0		/* app operation failed */


/* message used by logon/logoff */
typedef struct {
    short   type;	    /* message type */
    char    sysname [40];   /* system name */
    char    opname [OPERATOR_SZ];    /* operator name */
    char    pass [40];	    /* password information (encrypted?) */
} LogonMsg;
typedef struct {
    short   type;	    /* message type */
    short   value;	    /* response status */
} LogonRsp;

/* message used for check priviledges */
typedef struct {
    short   type;	    /* message type */
    char    priv[120];	    /* priviledge requested */
} PrivMsg;
typedef struct {
    short   type;	    /* message type */
    short   value;	    /* response status */
    char    errString[512]; /* err message space */
} PrivRsp;

/* message used to write tracking */
typedef struct {
    short   type;	    /* message type */
    char    channel[120];   /* channel */
    char    oldvalue[120];   /* old */
    char    newvalue[120];   /* new */
} TrackMsg;
typedef struct {
    short   type;	    /* message type */
    short   value;	    /* response status */
    char    errString[512]; /* err message space */
} TrackRsp;

/* message used to open/close tracking */
typedef struct {
    short   type;	    /* message type */
    char    opname[OPERATOR_SZ];    /* Operator name */
} OpenCloseTrackMsg;
typedef struct {
    short   type;	    /* message type */
    short   value;	    /* response status */
    char    errString[512]; /* err message space */
} OpenCloseTrackRsp;

/* message used to open/close priv */
typedef struct {
    short   type;	    /* message type */
} OpenClosePrivMsg;
typedef struct {
    short   type;	    /* message type */
    short   value;	    /* response status */
} OpenClosePrivRsp;

/* message used to start/exit/kill an app */
typedef struct {
    short   type;	    /* message type */
    char    appname[80];    /* Application name */
} AppControlMsg;
typedef struct {
    short   type;	    /* message type */
    short   value;	    /* response status */
} AppControlRsp;


/* message used to get open apps list */
typedef struct {
    short   type;	    /* message type */
    char    opname[OPERATOR_SZ];    /* Operator name */
} AppListMsg;
typedef struct {
    short   type;	    /* message type */
    short   value;	    /* response status */
    char    appString[1024]; /* app names space */
} AppListRsp;


typedef union {
    LogonMsg	logon;
    PrivMsg	priv;
    TrackMsg	track;
    OpenCloseTrackMsg openclosetrack;
    OpenClosePrivMsg openclosepriv;
    AppControlMsg appcontrol;
    AppListMsg applist;
} ReqUnion;

typedef union {
    LogonRsp	logon;
    PrivRsp	priv;
    TrackRsp	track;
    OpenCloseTrackRsp openclosetrack;
    OpenClosePrivRsp opencloseprivk;
    AppControlRsp appcontrol;
    AppListRsp applist;
} RespUnion;

/* Server function prototypes */
#ifdef TRK_SERVER
int	main ();

int	GetLogonResp ();
int	GetLogoffResp ();
int	GetCheckPrivResp ();
int	GetOpPrivResp ();
int	GetWriteTrackResp ();
 
/* Client function prototypes */
#elif defined(TRK_CLIENT)

int	InitAppControl (char *AppName);	/* set application control name */
int	ExitAppControl (void);		/* closes app control link */
int	CheckAppControl (void);		/* exits if app control link broken */

int	InitOperPriv (char *OperName);
int	InitOperTracking (char *OperName);
int	CloseOperTracking (char *OperName);
int	CheckOperPriv (char *Channel);
int	WriteOperTracking (char *Channel, char* oldValue, char *newValue);
int	OperPrivTrackErr (int code, char ***message);

#endif
