/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
**	client.c - client side of client/server message functions.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef vxWorks
#include <malloc.h>
#endif vxWorks

#ifndef WIN32
#ifndef vxWorks
#include <sys/types.h>
#include <sys/time.h>
#endif !vxWorks
#endif WIN32

#include "client.h"


#ifndef TRUE
#define TRUE	(1==1)
#define FALSE	(1!=1)
#endif TRUE

CLIENT*	    activeClients = NULL;

static char *DefaultInfoProc ();
static void CheckSocketStatus (CLIENT *client);
static int CheckClient (CLIENT *client);
static void Append (void *item, void **list);

/* Create a client */
CLIENT* OpenServer (char *server, int sync, char *connectinfo)
{
    CLIENT* rclient;
    SOCKET  sudp;
    SOCKADDR_IN srva;
    int	    srvalen;
    SvcReq  msg;
    SvcResp respmsg;
    int	    nadverts;
    int	    naccept;
#ifdef WIN32
    WSADATA wsaData; 
    WORD    wVersionRequested = MAKEWORD(1, 1);
#endif WIN32
    int	    err;
    char    *sp;
    int	    msglen;
    long    hostaddr = 0;
    int	    true = 1;
    struct  hostent *hp;
    char    tempserver [40];
    static char serverList [120];
    char *cp;
    char *ep;
    char tc;
 
#ifdef WIN32
    err = WSAStartup(wVersionRequested, &wsaData); 
 
    if (err != 0) 
	/* Tell the user that we couldn't find a useable */ 
	/* winsock.dll.     */ 
	return NULL; 
#endif WIN32
    rclient = malloc(sizeof(CLIENT));
    if (rclient)
    {
	rclient->next = NULL;
        rclient->s = INVALID_SOCKET;
	rclient->created = time(0);
	rclient->lastRx = 0;
	rclient->lastTx = 0;
	rclient->seq = 0;
	rclient->rxlen = 0;
	rclient->inmsg = NULL;
	rclient->resp = NULL;
	rclient->msgs = NULL;
	rclient->txmsgs = NULL;
	rclient->sync = sync;
	strncpy (rclient->SrvID, server, sizeof(rclient->SrvID));
	rclient->ClientInfo = DefaultInfoProc;
	/* Try to find server if necessary */
	if (*server == ':' || strchr (server, ','))
	{
	    sudp = socket(PF_INET, SOCK_DGRAM, 0);
	    srva.sin_family = AF_INET;
	    srva.sin_port = htons(TISSERVERPORT);
#ifdef WIN32
	    if (setsockopt (sudp, SOL_SOCKET, SO_BROADCAST, (char *)&true, sizeof(true))
			== SOCKET_ERROR)
		/* couldn't set broadcast */ ;
	    if (ioctlsocket (sudp, FIONBIO, &true) == SOCKET_ERROR)
#else
	    if (ioctl (sudp, FIONBIO, &true) == SOCKET_ERROR)
#endif WIN32
	    {
		/* couldn't set non-blocking */
		perror ("Couldn't set socket non-blocking");
	    }
	    msg.type = htons(SVCREQ);
	    msg.sftlvl = htonl(SOFTWARE_LEVEL);
	    if (sp = strchr(server, ':'))
	        strncpy (msg.service, sp+1, sizeof(msg.service));
	    else
	    {
		/* error, bad service name! */
		free (rclient);
		return NULL;
	    }
	    nadverts = 0;
	    do
	    {
		if (strchr (server, ','))
		{
		    /* send to list of hostnames */
		    strncpy (serverList, server, sizeof(serverList));
		    cp = serverList;
		    tc = 0;
		    while ((tc != ':')
		      && ((ep = strchr (cp, ',')) || (ep = strchr (cp, ':'))))
		    {
			tc = *ep; /* : means end of a list */
			*ep++ = 0;
#ifdef vxWorks
			if ((hostaddr = inet_addr (cp)) == INADDR_NONE)
			    hostaddr = 0;
#else
			if (hp = gethostbyname (cp))
			    hostaddr = *(long *)hp->h_addr_list[0];
			else if ((hostaddr = inet_addr (cp)) == INADDR_NONE)
			    hostaddr = 0;
#endif vxWorks
			if (hostaddr)
			{
			    srva.sin_addr.s_addr = hostaddr;
			    if (sendto (sudp, (char *)&msg, sizeof(msg),
				0, (SOCKADDR *)&srva, sizeof(srva)) == SOCKET_ERROR)
			    {
				err = WSAGetLastError();
				perror ("Socket send error during list search traversal");
				free (rclient);
				return NULL;
			    }
			}
			cp = ep; /* move to next item */
		    }
		    /* restore hostaddr to 0 (indicates that we haven't found one yet) */
		    hostaddr = 0;
		}
		else
		{
		    /* Send to broadcast for local servers */
		    srva.sin_addr.s_addr = htonl(INADDR_BROADCAST);
		    if (sendto (sudp, (char *)&msg, sizeof(msg),
			0, (SOCKADDR *)&srva, sizeof(srva)) == SOCKET_ERROR)
		    {
			err = WSAGetLastError();
			perror ("Socket send error during broadcast");
			free (rclient);
			return NULL;
		    }
		    /* also send to localhost (127.0.0.1) for local servers */
		    srva.sin_addr.s_addr = htonl(0x7f000001);
		    if (sendto (sudp, (char *)&msg, sizeof(msg),
			0, (SOCKADDR *)&srva, sizeof(srva)) == SOCKET_ERROR)
		    {
			err = WSAGetLastError();
			perror ("Socket send error to local host");
			free (rclient);
			return NULL;
		    }
		}
		naccept = 0;
		do 
		{
		    srvalen = sizeof(srva);
		    if ((msglen = recvfrom (sudp, (char *)&respmsg, sizeof(respmsg),
					0, (SOCKADDR*)&srva, &srvalen)) < 0)
		    {
			err = WSAGetLastError();
			if ((msglen == 0) || ((msglen == SOCKET_ERROR) && (err != WSAEWOULDBLOCK)))
			{
			    perror ("Socket receive error during search");
			    free (rclient);
			    return NULL;
			}
#ifdef WIN32
			Sleep (1000);
#else
#ifdef vxWorks
			taskDelay (60);
#else
			sleep (1);
#endif vxWorks
#endif WIN32
		    }
		    else
		    {
			if ((msglen == sizeof(respmsg))
			  && (ntohs(respmsg.type) == SVCRESP)
			  && (ntohl (respmsg.sftlvl) == SOFTWARE_LEVEL))
			{
			    hostaddr = srva.sin_addr.s_addr;
			}
		    }
		} while ((++naccept < ACCRATE)
		  && (hostaddr == 0));
	    } while ((++nadverts < MAXADVERTS)
	      && (hostaddr == 0));
	    closesocket (sudp);
	}
	else
	{
	    strncpy (tempserver, server, sizeof(tempserver));
	    if (strchr (tempserver, ':') == NULL)
	    {
		/* illegal server specification */
		free (rclient);
		rclient = 0;
		    hostaddr = 0;
	    }
	    else if (connectinfo && strcmp (connectinfo, "APP_CONTROL") == 0)
	    {
		/* use localhost value (127.0.0.1) for this connection */
		hostaddr = inet_addr ("127.0.0.1");
	    }
	    else
	    {
		*(strchr (tempserver, ':')) = 0;
#ifdef vxWorks
		if ((hostaddr = inet_addr (tempserver)) == INADDR_NONE)
		    hostaddr = 0;
#else
		if (hp = gethostbyname (tempserver))
		    hostaddr = *(long *)hp->h_addr_list[0];
		else if ((hostaddr = inet_addr (tempserver)) == INADDR_NONE)
		    hostaddr = 0;
#endif vxWorks
	    }
	}
	if (hostaddr == 0)
	{
	    perror ("No server available");
	    free (rclient);
	    rclient = 0;
	}
	else
	{
	    /* create the TCP socket */
	    rclient->s = socket(PF_INET, SOCK_STREAM, 0);
	    rclient->sa.sin_family = AF_INET;
	    rclient->sa.sin_addr.s_addr = hostaddr;
	    if (connectinfo && strcmp (connectinfo, "DEBUG") == 0)
		rclient->sa.sin_port = htons(8002);
	    else if (connectinfo && strcmp (connectinfo, "APP_CONTROL") == 0)
		rclient->sa.sin_port = htons(8003);
	    else
		rclient->sa.sin_port = htons(TISSERVERPORT);

	    if (connect (rclient->s, (SOCKADDR*)&rclient->sa, sizeof(rclient->sa)))
	    {
		err = WSAGetLastError();
		perror ("Can't connect to server");
		closesocket (rclient->s);
		free (rclient);
		rclient = 0;
	    }
	    else
	    {
		if ((connectinfo == NULL)
		  || ((strcmp (connectinfo, "DEBUG") != 0)
		    && (strcmp (connectinfo, "APP_CONTROL") != 0)))
		{
		    msg.type = htons(SVCREQ);
		    msg.sftlvl = htonl(SOFTWARE_LEVEL);
		    if (sp = strchr(server, ':'))
			strncpy (msg.service, sp+1, sizeof(msg.service));
		    else
		    {
			/* error, bad service name! */
			closesocket (rclient->s);
			free (rclient);
			return NULL;
		    }
		    if (send (rclient->s, (char *)&msg, sizeof(msg), 0) != sizeof(msg))
		    {
			err = WSAGetLastError();
			closesocket (rclient->s);
			free (rclient);
			rclient = NULL;
		    }
		}
#ifdef WIN32
		if (ioctlsocket (rclient->s, FIONBIO, &true) == SOCKET_ERROR)
		{
		    /* couldn't set non-blocking */
#else
		if (ioctl (rclient->s, FIONBIO, &true) == SOCKET_ERROR)
		{
		    /* couldn't set non-blocking */
		    perror ("Couldn't set socket non-blocking %m");
#endif
		    closesocket (rclient->s);
		    free (rclient);
		    rclient = NULL;
		}
	    }
	}
    }
    if (rclient)
	Append (rclient, (void**)&activeClients);
    return rclient;
}

char ReturnInfoText [1024];
char *GetClientInfo ()
{
    CLIENT *acp = activeClients;

    ReturnInfoText [0] = 0;
    while (acp)
    {
	strcat (ReturnInfoText, acp->ClientInfo(acp));
	acp = acp->next;
    }
    return ReturnInfoText;
}

char clientinfo [132];
static char *DefaultInfoProc (CLIENT *client)
{
    struct hostent *haddr;

    if (client)
    {
	haddr = gethostbyaddr ((char *)&client->sa.sin_addr.s_addr, 4, PF_INET);
	if (haddr)
	    sprintf (clientinfo, "Connected to %s for %s\n", haddr->h_name, client->SrvID);
	else
	    sprintf (clientinfo, "Connected to unknown host %s for %s\n",
			    inet_ntoa (client->sa.sin_addr), client->SrvID);
    }
    else
    {
	sprintf (clientinfo, "Not connected for %s\n", client->SrvID);
    }
    return clientinfo;
}

/* Close the client down */
int CloseSession (CLIENT *client)
{
    CLIENT *acp;
    SVRMSG *freep;

    if (CheckClient(client))
    {
	if (client->s != INVALID_SOCKET)
	    closesocket (client->s);
	client->s = INVALID_SOCKET;
	freep = client->txmsgs;
	while (freep)
	{
	    client->txmsgs = freep->next;
	    free (freep);
	}
	if (client->resp)
	    free (client->resp);
	freep = client->msgs;
	while (freep)
	{
	    client->msgs = freep->next;
	    free (freep);
	}
	if (activeClients == client)
	    activeClients = client->next;
	else
	{
	    acp = activeClients;
	    while (acp)
	    {
		if (acp->next == client)
		    acp->next = client->next;
		acp = acp->next;
	    }
	}
	free (client);
    }
    return 0;
}

/* Send a request */
void *SendReq (CLIENT *client, void* msg, short int length)
{
    short int retLen;

    return SendReqLen (client, msg, length, &retLen);
}

void *SendReqLen (CLIENT *client, void* msg, short int length, short int *retLen)
{
    SVRMSG  *txmsg;
    int	    status;

    if (!CheckClient (client) || client->s == INVALID_SOCKET)
    {
	return NULL;
    }
    client->seq++; /* bump seq on each attempt to ensure matches */
    txmsg = malloc (sizeof(SVRMSG));
    if (txmsg)
    {
	txmsg->hdr.type = htons(SRVREQ);
	txmsg->hdr.seq = htons(client->seq);
	txmsg->hdr.userlength = htons((short)length);
	txmsg->data = malloc (sizeof(txmsg->hdr) + length);
	if (txmsg->data)
	{
	    memcpy (txmsg->data, &txmsg->hdr, sizeof(txmsg->hdr));
	    memcpy (txmsg->data+sizeof(txmsg->hdr), msg, length);
	    txmsg->hdr.userlength = sizeof(txmsg->hdr) + length;
	    status = send (client->s,
			txmsg->data, txmsg->hdr.userlength, 0);
	    if (status == SOCKET_ERROR)
	    {
		/* close down session */
		closesocket (client->s);
		client->s = INVALID_SOCKET;
		free (txmsg->data);
		free (txmsg);
		return NULL;
	    }
	    else if (status != txmsg->hdr.userlength)
	    {
		txmsg->hdr.userlength -= status;
		memcpy (txmsg->data, txmsg->data+status,
			txmsg->hdr.userlength);
		Append (txmsg, (void**)&client->txmsgs);
		return NULL;
	    }
	    free (txmsg->data);
	}
	free (txmsg);
    }
    return (GetRspLen (client, retLen));
}

void *GetRsp (CLIENT *client)
{
    short int len;

    return GetRspLen (client, &len);
}

void *GetRspLen (CLIENT *client, short int *len)
{
    void *data = NULL;

    while (data == NULL)
    {
	if (!CheckClient (client) || client->s == INVALID_SOCKET)
	{
	    return NULL;
	}
	/* check socket for send stuff */
	CheckSocketStatus (client);
	/* return response if present */
	if (client->resp)
	{
	    if (client->resp->hdr.seq == client->seq)
	    {
		*len = client->resp->hdr.userlength;
		data = client->resp->data;
	    }
	    free (client->resp);
	    client->resp = NULL;
	}
	if (client->sync == 0)
	    return data;
#ifdef vxWorks
	taskDelay (1);
#else
#ifdef WIN32
	Sleep (100);
#else
	{
	    struct timeval timeout = {0, 10000};

	    select (8, 0, 0, 0, &timeout);
	}
#endif WIN32
#endif vxWorks
    }
    return data;
}

/* Receive a message */
void *GetMsg (CLIENT *client)
{
    void *data = NULL;
    SVRMSG *rxmsg;

    if (!CheckClient (client) || client->s == INVALID_SOCKET)
    {
	return data;
    }
    /* check socket for send stuff */
    CheckSocketStatus (client);
    /* return response if present */
    rxmsg = client->msgs;
    if (rxmsg)
    {
	client->msgs = rxmsg->next;
	data = rxmsg->data;
	free (rxmsg);
    }
    return data;
}

/* check the socket for activity */
static void CheckSocketStatus (CLIENT *client)
{
    int		status;
    SVRMSG	*msg;
    int		err;
#ifdef SOLARIS
    int		dataAvailable;
#endif SOLARIS

    if (!CheckClient (client) || client->s == INVALID_SOCKET)
    {
	return;
    }
    /* check socket for send stuff */
    while (client->txmsgs)
    {
	msg = client->txmsgs;
	status = send (client->s,
		    msg->data, msg->hdr.userlength, 0);
	if (status == SOCKET_ERROR)
	{
	    err = WSAGetLastError();
	    if (err != WSAEWOULDBLOCK)
	    {
		perror ("Socket receive error");
		/* close down session */
		closesocket (client->s);
		client->s = INVALID_SOCKET;
		return;
	    }
	}
	else if (status == msg->hdr.userlength)
	{
	    client->txmsgs = msg->next;
	    free (msg);
	}
	else
	{
	    msg->hdr.userlength -= status;
	    memcpy (msg->data, msg->data+status,
		    msg->hdr.userlength);
	    break; /* out of send pending messages loop */
	}
    }
    /* check socket for receive stuff */
#ifdef SOLARIS
    if (ioctl(client->s, FIONREAD, &dataAvailable) < 0)
    {
	closesocket (client->s);
	client->s = INVALID_SOCKET;
	return;
    }
    if (dataAvailable == 0)
	return;   
#endif SOLARIS
    if (client->rxlen < sizeof(ServHdr))
    {
	status = recv (client->s,
	    	    ((char *)&client->hdr) + client->rxlen,
	    	    sizeof(ServHdr),
		    0);
	if ((client->rxlen + status) == sizeof(ServHdr))
	{
	    client->rxlen += status;
	    client->hdr.type = ntohs(client->hdr.type);
	    client->hdr.seq = ntohs(client->hdr.seq);
	    client->hdr.userlength = ntohs(client->hdr.userlength);
	    client->inmsg = malloc (sizeof(SVRMSG));
	    if (client->inmsg)
	    {
		client->inmsg->hdr = client->hdr;
		client->inmsg->next = NULL;
		client->inmsg->data = malloc (client->hdr.userlength);
		if (client->inmsg->data == NULL)
		{
		    /* error, close session */
		    closesocket (client->s);
		    client->s = INVALID_SOCKET;
		    return;
		}
	    }
	    else
	    {
		/* error, close session */
		closesocket (client->s);
		client->s = INVALID_SOCKET;
		return;
	    }
	}
	else if ((status == SOCKET_ERROR) || (status == 0))
	{
	    err = WSAGetLastError();
	    if ((status == 0) || ((status == SOCKET_ERROR) && (err != WSAEWOULDBLOCK)))
	    {
		perror ("Socket receive error");
		/* close down session */
		closesocket (client->s);
		client->s = INVALID_SOCKET;
		return;
	    }
	}
    }
    if (client->rxlen >= sizeof(ServHdr))
    {
	status = recv (client->s,
		client->inmsg->data + (client->rxlen - sizeof(ServHdr)),
		client->inmsg->hdr.userlength - (client->rxlen - sizeof(ServHdr)),
		0);
	if (status > 0)
	{
	    client->rxlen += status;
	    if ((unsigned)client->inmsg->hdr.userlength
		 == (client->rxlen - sizeof(ServHdr)))
	    {
		switch (client->inmsg->hdr.type)
		{
		case SRVRSP:
		    /* link data to rxmsg chain, return session */
		    if (client->inmsg->hdr.seq == client->seq)
		    {
			if (client->resp)
			    free (client->resp);
			client->resp = client->inmsg;
		    }
		    else
		    {
			free (client->inmsg);
		    }
		    client->inmsg = NULL;
		    client->rxlen = 0;
		    break;
		case SRVMSG:
		    /* link data to rxmsg chain, return session */
		    Append (client->inmsg, (void**)&client->msgs);
		    client->inmsg = NULL;
		    client->rxlen = 0;
		    break;
		}
	    }
	}
	else /* error, close session */
	{
	    err = WSAGetLastError();
	    if ((status == 0) || ((status == SOCKET_ERROR) && (err != WSAEWOULDBLOCK)))
	    {
		perror ("Socket receive error");
		/* close down session */
		closesocket (client->s);
		client->s = INVALID_SOCKET;
		return;
	    }
	}
    }
}

static int CheckClient (CLIENT *client)
{
    CLIENT *acp = activeClients;
    while (acp)
    {
	if (acp == client)
	    return TRUE;
	acp = acp->next;
    }
    return FALSE;
}

static void Append (void *item, void **list)
{
    if (*list == NULL)
    {
	*list = item;
    }
    else
    {
	void **ptr = list;
	while (*ptr)
	    ptr = *ptr;
	*ptr = item;
    }
}
