/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
   File:   trendata.c - trend data server, providing access to the oracle snap
		 files to remote clients.
   Author: Grant Irani
   Date:   4/30/96 13:21
*/

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include "stdio.h"
#include "time.h"
#include "ctype.h"
#include "fcntl.h"

#if !(defined(MIPSEL) || defined(AXP))
#include "memory.h"
#else
#include <sys/types.h>
#include <sys/stat.h>
extern int lstat ();
extern int readlink ();
#endif

#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#include "defines.h"
#include "trendfil.h"
#include "server.h"
#include "trenddata.h"

#define LOG_OPTIONSI	(LOG_INFO | LOG_LOCAL1)
#define LOG_OPTIONSD	(LOG_DEBUG | LOG_LOCAL1)
#define LOG_OPTIONSE	(LOG_ERR | LOG_LOCAL1)

SERVER *server;		/* the server control structure */


RespUnion    rspmsg;

SERVER *server;
int	trendposition; /* track loc in trends file */
char	*trcp;	
int	reducposition; /* track loc in reducs file */
char	*rdcp;	

int main (argc, argv)
int	argc;
char**	argv;
{
    ReqUnion	*reqmsg;
    int msglen;
    int i;
    char msgbuf [512];

    syslog(LOG_OPTIONSI, "TrendData server running");

    /* all args are considered environment vars */
    for (i = 1; i < argc; i++)
    {
	sprintf (msgbuf, "TrendData putenv - %s", argv[i]);
	syslog(LOG_OPTIONSI, msgbuf);
	putenv (argv);
    }

    /* open server as synchronous I/O */
    if ((server = OpenServer ("TrendData", TRUE, NULL)) == NULL)
	exit (0);

#ifdef DEBUG
    SetSocketNum (server, 8002);
#endif DEBUG
    snapinit ();
    trendposition = 0;
    reducposition = 0;

    while (reqmsg = GetReq (server))
    {
#ifndef DEBUG
	syslog(LOG_OPTIONSD, "Processing request %d", ((PrivMsg *)reqmsg)->type);
#endif
	switch (((PrivMsg *)reqmsg)->type)
	{
	    case GET_PRIV:
		msglen = GetPrivResp (reqmsg);
		break;
	    case GET_TRENDNAMES:
		msglen = GetTrNameResp (reqmsg);
		break;
	    case GET_NOTE:
		msglen = GetNoteResp (GET_NOTE, reqmsg);
		break;
	    case GET_PREVNOTE:
		msglen = GetNoteResp (GET_PREVNOTE, reqmsg);
		break;
	    case SAVE_NOTE:
		msglen = GetSaveNoteResp (GET_NOTE, reqmsg);
		break;
	    case DEL_NOTE:
		msglen = GetNoteResp (DEL_NOTE, reqmsg);
		break;
	    case SELECT_FILE:
		msglen = GetFileResp (reqmsg);
		break;
	    case GET_REDUC:
		msglen = GetReducResp (reqmsg);
		break;
	    case GET_TREND:
		msglen = GetTrendResp (reqmsg);
		break;
	    case SAVE_TREND:
		msglen = GetSaveTrendResp (reqmsg);
		break;
	    case DELETE_TREND:
		msglen = GetDelTrendResp (reqmsg);
		break;
	    case SEARCH_DB:
		msglen = GetSearchResp (reqmsg);
		break;
	    case RECORD:
		msglen = GetRecordResp (reqmsg);
		break;
	    case UPDATE:
		msglen = GetUpdateResp (reqmsg);
		break;
	    case UPDREC:
		msglen = GetUpdrecResp (reqmsg);
		break;
	    case DATA:
		msglen = GetDataResp (reqmsg);
		break;
	    case GET_VALUES:
		msglen = GetValuesResp (reqmsg);
		break;
	}
	SendRsp (server, (u_char *)&rspmsg, msglen);
	free (reqmsg);
    }
#ifndef DEBUG
    syslog(LOG_OPTIONSD, "Null message from GetReq");
#endif
    /* cleanup any open files, etc... */
    CloseSession (server);
    closelog ();
}

int GetPrivResp (reqmsg)
PrivMsg *reqmsg;
{
    PrivRsp	*priv = (PrivRsp *)&rspmsg;

    priv->type = GET_PRIV;
    /* should log in as op w/ password to get permissions */
    /* priv->value = OperPriv (reqmsg->group); */
    priv->type = FALSE;
    return (sizeof(PrivRsp));
}

int GetTrNameResp (reqmsg)
TrNamesMsg *reqmsg;
{
    TrNamesRsp	*trendnames;
    int		ix;

    trendnames = (TrNamesRsp*)&rspmsg;
    trendnames->type = GET_TRENDNAMES;
    /* move to position */
    if ((trendposition == 0) || (trendposition != reqmsg->position))
    {
	trcp = FirstTrend ();
	for (trendposition = 0;
	 trcp && trendposition != reqmsg->position;
	 trendposition++)
	    trcp = NextTrend ();
    }
    /* load in names */
    trendnames->type = FALSE;
    for (ix == 0; trcp && ix < 20; ix++)
    {
	strncpy (trendnames->trendDescriptors[ix], trcp, C_TRENDDESCLEN);
	trcp = NextTrend ();
    }
    trendnames->count = ix;
    trendnames->value = TRUE;
    return (sizeof(TrNamesRsp));
}

int GetNoteResp (type, reqmsg)
int	type;
NoteMsg *reqmsg;
{
    NoteRsp	*note = (NoteRsp *)&rspmsg;

    note->type = type;
    strncpy (note->pointname, reqmsg->pointname, sizeof(note->pointname));
    switch (type)
    {
	case GET_NOTE:
	    note->time = GetNote (reqmsg->pointname, reqmsg->time);
	    break;
	case GET_PREVNOTE:
	    note->time = GetNote (reqmsg->pointname, reqmsg->time);
	    break;
	case DEL_NOTE:
	    DeleteNote ();
	    note->time = 0x7fffffff;
	    break;
    }
    if (note->time != 0x7fffffff)
	strncpy (note->note, ReadNote (), sizeof(note->note));
    return (sizeof(NoteRsp));
}

int GetSaveNoteResp (type, reqmsg)
int	type;
NoteSaveMsg *reqmsg;
{
    NoteRsp	*note = (NoteRsp *)&rspmsg;

    SaveNote (reqmsg->pointname, reqmsg->time, reqmsg->note);
    note->type = SAVE_NOTE;
    strncpy (note->pointname, reqmsg->pointname, sizeof(note->pointname));
    note->time = GetNote (reqmsg->pointname, reqmsg->time);
    if (note->time != 0x7fffffff)
	strncpy (note->note, ReadNote (), sizeof(note->note));
    return (sizeof(NoteRsp));
}

int GetFileResp (reqmsg)
FileMsg *reqmsg;
{
    FileRsp	*file = (FileRsp *)&rspmsg;

    file->type = SELECT_FILE;
    file->value = opensnap (reqmsg->filename);
    return (sizeof(FileRsp));
}

int GetReducResp (reqmsg)
ReducMsg *reqmsg;
{
    ReducRsp	*reduc;
    int		ix;
    int		spos;

    reduc = (ReducRsp*)&rspmsg;
    reduc->type = GET_REDUC;
    /* move to position */
    if ((reducposition == 0) || (reducposition != reqmsg->position))
    {
	rdcp = FirstPoint ();
	for (reducposition = 0;
	  rdcp && reducposition != reqmsg->position;
	  reducposition++)
	    rdcp = NextPoint ();
    }

    /* load in names */
    reduc->type = FALSE;
    for (ix = 0, spos = 0;
      rdcp && (spos + strlen(rdcp) + 1) < 1024;
      ix++)
    {
	strncpy (&reduc->names[spos], rdcp, C_TRENDDESCLEN);
	spos += strlen(rdcp) + 1;
	rdcp = NextPoint ();
    }
    reduc->total = GetPointCount ();
    reduc->count = ix;
    reduc->value = TRUE;
    return (sizeof(ReducRsp));
}

int GetTrendResp (reqmsg)
TrendMsg *reqmsg;
{
    TrendRsp *trend = (TrendRsp *)&rspmsg;

    trend->type = GET_TREND;
    strncpy (trend->descr, reqmsg->descr, sizeof(trend->descr));
    trend->value = GetTrendInfo (trend->descr, &trend->zoom, trend->specs);
    return (sizeof(TrendRsp));
}

int GetSaveTrendResp (reqmsg)
TrendSaveMsg *reqmsg;
{
    TrendRsp *trend = (TrendRsp *)&rspmsg;

    trend->type = SAVE_TREND;
    trend->value = PutTrendInfo (reqmsg->descr, reqmsg->zoom,
				 reqmsg->specs, reqmsg->flag);
    strncpy (trend->descr, reqmsg->descr, sizeof(trend->descr));
    GetTrendInfo (trend->descr, &trend->zoom, trend->specs);
    return (sizeof(TrendRsp));
}

int GetDelTrendResp (reqmsg)
TrendMsg *reqmsg;
{
    TrendRsp *trend = (TrendRsp *)&rspmsg;

    trend->type = DELETE_TREND;
    trend->value = DelTrendInfo (reqmsg->descr);
    return (sizeof(TrendRsp));
}

int GetSearchResp (reqmsg)
SearchMsg *reqmsg;
{
    SearchRsp	*search = (SearchRsp *)&rspmsg;

    search->type = SEARCH_DB;
    search->value = TRUE;
    search->time = dbsearch (reqmsg->pointname,
			 reqmsg->dbred, reqmsg->time,
			 &reqmsg->srch);
    return (sizeof(SearchRsp));
}

int GetRecordResp (reqmsg)
RecMsg *reqmsg;
{
    RecRsp *record = (RecRsp *)&rspmsg;

    record->type = RECORD;
    record->result = GetRec (reqmsg->pointname, reqmsg->dbreduc, reqmsg->time);
    if (record->result)
    {
	LoadRec (record->pointname, &record->dbreduc, &record->time,
		 &record->rate, &record->samples, record->tag,
		 &record->dl, &record->dh,
		 record->value, record->status);
    }
    return (sizeof(RecRsp));
}

int GetUpdateResp (reqmsg)
UpdateMsg *reqmsg;
{
    RecRsp *record = (RecRsp *)&rspmsg;

    record->result = UpdateRec (reqmsg->pointname, reqmsg->dbreduc,
				reqmsg->rec, reqmsg->ix,
               			reqmsg->value, reqmsg->status);
    if (record->result)
    { 
	LoadRec (record->pointname, &record->dbreduc, &record->time,
		 &record->rate, &record->samples, record->tag,
		 &record->dl, &record->dh,
		 record->value, record->status);
    }
    else
    {
	strncpy (record->pointname, reqmsg->pointname, sizeof(record->pointname));
	record->dbreduc = reqmsg->dbreduc;
	record->time = 0x7fffffff;
    }
    return (sizeof(RecRsp));
}

int GetUpdrecResp (reqmsg)
UpdateMsg *reqmsg;
{
    RecRsp *record = (RecRsp *)&rspmsg;

    /* not implemented yet */
    record->result = FALSE;

    if (record->result)
    { 
	LoadRec (record->pointname, &record->dbreduc, &record->time,
		 &record->rate, &record->samples, record->tag,
		 &record->dl, &record->dh,
		 record->value, record->status);
    }
    else
    {
	strncpy (record->pointname, reqmsg->pointname, sizeof(record->pointname));
	record->dbreduc = reqmsg->dbreduc;
	record->time = 0x7fffffff;
    }
    return (sizeof(RecRsp));
}

int GetDataResp (reqmsg)
RecMsg *reqmsg;
{
    RecRsp *record = (RecRsp *)&rspmsg;

    record->type = RECORD;
    record->result = GetRec (reqmsg->pointname, reqmsg->dbreduc, reqmsg->time);
    if (record->result)
    {
	/* find time of good (no connect error) data after time requested */
	while ((ItemTime () != 0x7fffffff) && (ItemStatus ()[0] == 'C'))
	{
	    NextItem ();
	}
	if (ItemTime () != 0x7fffffff)
	{
	    LoadRec (record->pointname, &record->dbreduc, &record->time,
		     &record->rate, &record->samples, record->tag,
		     &record->dl, &record->dh,
		     record->value, record->status);
	}
	else
	{
	    strncpy (record->pointname, reqmsg->pointname, sizeof(record->pointname));
	    record->dbreduc = reqmsg->dbreduc;
	    record->time = 0x7fffffff;
	}
    }
    else
    {
	strncpy (record->pointname, reqmsg->pointname, sizeof(record->pointname));
	record->dbreduc = reqmsg->dbreduc;
	record->time = 0x7fffffff;
    }
    return (sizeof(RecRsp));
}

int GetValuesResp (reqmsg)
GetValMsg *reqmsg;
{
    GetValRsp *record = (GetValRsp *)&rspmsg;

    record->type = RECORD;
    record->samples = GetSnapValues (reqmsg->pointname, redStrings[reqmsg->dbreduc],
				    redStrings[reqmsg->reduc],
				    reqmsg->unitsecs, reqmsg->start, &record->dbrate,
				    record->tag, &record->disp_low, &record->disp_high,
				    record->reqnum, record->values, record-status, record->values2);
    record->result = (record->samples > 0);

    return (sizeof(GetValRsp));
}
