/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
**	server.c - server side of server/server message functions.
*/

#include "server.h"
#include <sys/types.h>
#ifdef SOLARIS
#include <sys/time.h>
#endif SOLARIS
#include "msgMon.h"

#ifndef TRUE
#define TRUE	(1==1)
#define FALSE	(1!=1)
#endif TRUE

SERVER*	    activeServers = NULL;

static void CheckSocketStatus (SERVER *server);
static int CheckServer (SERVER *server);
static void Append (void *item, void **list);

/* Create a server */
SERVER* OpenMultiService (char *service, int sync, char *connectinfo, SOCKET svcSocket)
{
    SERVER* rserver;
    int	    true = 1;

    tisOpenLog ("server", 6, 2, NULL, NULL);
    tisSysLog (0, 0, "Server started");

    rserver = malloc(sizeof(SERVER));
    if (rserver)
    {
	rserver->next = NULL;
        rserver->s = svcSocket;
	rserver->created = time(0);
	rserver->lastRx = 0;
	rserver->lastTx = 0;
	rserver->seq = 0;
	rserver->rxlen = 0;
	rserver->inmsg = NULL;
	rserver->msgs = NULL;
	rserver->txmsgs = NULL;
	rserver->sync = sync;
	strncpy (rserver->SrvID, service, sizeof(rserver->SrvID));

#ifdef WIN32
	if (!sync && ioctlsocket (rserver->s, FIONBIO, &true) == SOCKET_ERROR)
	    /* couldn't set non-blocking */ ;
#else
	if (!sync && ioctl (rserver->s, FIONBIO, &true) == SOCKET_ERROR)
	    /* couldn't set non-blocking */
	    tisSysLog (0, 0, "Couldn't set socket non-blocking %m");
#endif
    }
    if (rserver)
	Append (rserver, (void **)&activeServers);
    return rserver;
    tisCloseLog ();
}

/* SetSocketNum - set listen socket and accept client for debug */
void SetSocketNum (SERVER *server, int port)
{
    int	ls;
    struct sockaddr_in sa;
    int salen;
    struct hostent *ha;
    int true = 1;

    /* create the TCP socket */
    ls = socket(PF_INET, SOCK_STREAM, 0);
    ha = gethostbyname("localhost");
    sa.sin_family = AF_INET;
    sa.sin_addr.s_addr = 0;
    sa.sin_port = htons((short)port);
    if (bind (ls, (struct sockaddr *)&sa, sizeof(sa)) == 0)
    {
	if (listen (ls, 2) != 0)
	    exit (0);
	salen = sizeof(sa);
	while ((server->s = accept (ls, (struct sockaddr *)&sa, &salen)) < 0)
	{
#ifdef WIN32
	    errno = WSAGetLastError ();
	    if (errno != WSAEINTR)
#else
	    if (errno != EINTR)
#endif WIN32
	    exit (0);
	}
#ifdef WIN32
	if (ioctlsocket (server->s, FIONBIO, &true) == SOCKET_ERROR)
#else
	if (ioctl (server->s, FIONBIO, &true) == SOCKET_ERROR)
#endif WIN32
	    /* couldn't set non-blocking */
	    tisSysLog (0, 0, "Couldn't set socket non-blocking %m");
    }
}

/* Close the server down */
int CloseSvcSession (SERVER *server)
{
    SERVER *acp;

    if (CheckServer(server))
    {
	tisSysLog (0, 0, "Closing socket");
	closesocket (server->s);
	server->s = INVALID_SOCKET;
	if (activeServers == server)
	    activeServers = server->next;
	else
	{
	    acp = activeServers;
	    while (acp)
	    {
		if (acp->next == server)
		    acp->next = server->next;
		acp = acp->next;
	    }
	}
	free (server);
    }
    return 0;
}

/* Check server status */
int CheckSvcSession (SERVER *server)
{
    return (CheckServer(server));
}

/* Wait for somthing to be ready on the socket */
int ServerWait (SERVER *server, long delay)
{
    fd_set	read_fds;
    fd_set	write_fds;
    fd_set	except_fds;
    struct timeval timeout;
    int		status;

    if (CheckServer(server))
    {
	FD_ZERO (&read_fds);
	FD_ZERO (&write_fds);
	FD_ZERO (&except_fds);
	FD_SET (server->s, &read_fds);
	timeout.tv_sec = delay/1000;
	timeout.tv_usec = (delay%1000)*1000;
	status = select (server->s + 1, &read_fds, &write_fds, &except_fds, &timeout);
	if (status > 0)
	    return 1;
	else
	    return 0;
    }
    return 0;
}

/* Receive a request */
void *GetReq (SERVER *server)
{
    short retLen;
    
    return GetReqLen (server, &retLen);
}

void *GetReqLen (SERVER *server, short *retLen)
{
    void *data = NULL;
    SVRMSG *rxmsg;
    struct timeval timeout = {0, 10000};
    fd_set readfds;
    int status;

    while ((data == NULL) && (CheckServer (server) == TRUE))
    {
	if ((CheckServer (server) == FALSE) || server->s == INVALID_SOCKET)
	{
	    return NULL;
	}
	/* check socket for send stuff */
	CheckSocketStatus (server);
	/* return request if present */
	rxmsg = server->msgs;
	if (rxmsg)
	{
	    server->msgs = rxmsg->next;
	    server->seq	= rxmsg->hdr.seq;
	    data = rxmsg->data;
	    *retLen = rxmsg->hdr.userlength;
	    free (rxmsg);
	    return data;
	}
	if (server->sync == FALSE && data == NULL)
	{
	    *retLen = 0;
	    return NULL;
	}

	FD_ZERO (&readfds);
	FD_SET (server->s, &readfds);
	timeout.tv_sec = 0;
	timeout.tv_usec = 10000;
	status = select ((server->s) + 1, &readfds, 0, 0, &timeout);
#ifdef SOLARIS
	if (status != 0)
	{
	    int dataAvailable;
	    if ((ioctl(server->s, FIONREAD, &dataAvailable) < 0) || (dataAvailable == 0))
	    {
		closesocket (server->s);
		server->s = INVALID_SOCKET;
	    }
	}
#endif SOLARIS
    }
    return data;
}

/* Send a response */
void SendRsp (SERVER *server, u_char* msg, short int length)
{
    SVRMSG  *txmsg;
    int	    status;

    tisSysLog (1, 0, "Send response");
    if ((CheckServer (server) == FALSE) || server->s == INVALID_SOCKET)
    {
	return;
    }
    txmsg = malloc (sizeof(SVRMSG));
    if (txmsg)
    {
	txmsg->hdr.type = htons(SRVRSP);
	txmsg->hdr.seq = htons((short)server->seq);
	txmsg->hdr.userlength = htons(length);
	txmsg->data = malloc (sizeof(txmsg->hdr) + length);
	if (txmsg->data)
	{
	    memcpy (txmsg->data, &txmsg->hdr, sizeof(txmsg->hdr));
	    memcpy (txmsg->data+sizeof(txmsg->hdr), msg, length);
	    txmsg->hdr.userlength = sizeof(txmsg->hdr) + length;
	    status = send (server->s,
			txmsg->data, txmsg->hdr.userlength, 0);
	    if (status == SOCKET_ERROR)
	    {
		/* close down session */
		closesocket (server->s);
		server->s = INVALID_SOCKET;
		free (txmsg->data);
		free (txmsg);
		return;
	    }
	    else if (status != txmsg->hdr.userlength)
	    {
		txmsg->hdr.userlength -= status;
		memcpy (txmsg->data, txmsg->data+status,
			txmsg->hdr.userlength);
		Append (txmsg, (void **)&server->txmsgs);
		return;
	    }
	    free (txmsg->data);
	}
	free (txmsg);
	server->lastTx = time(0);
    }
}

/* Send a message */
void SendMsg (SERVER *server, u_char* msg, short int length)
{
    SVRMSG  *txmsg;
    int	    status;

    tisSysLog (1, 0, "Send message");
    if ((CheckServer (server) == FALSE) || server->s == INVALID_SOCKET)
    {
	return;
    }
    txmsg = malloc (sizeof(SVRMSG));
    if (txmsg)
    {
	txmsg->hdr.type = htons(SRVMSG);
	txmsg->hdr.seq = htons((short)server->seq);
	txmsg->hdr.userlength = htons(length);
	txmsg->data = malloc (sizeof(txmsg->hdr) + length);
	if (txmsg->data)
	{
	    memcpy (txmsg->data, &txmsg->hdr, sizeof(txmsg->hdr));
	    memcpy (txmsg->data+sizeof(txmsg->hdr), msg, length);
	    txmsg->hdr.userlength = sizeof(txmsg->hdr) + length;
	    status = send (server->s,
			txmsg->data, txmsg->hdr.userlength, 0);
	    if (status == SOCKET_ERROR)
	    {
		/* close down session */
		closesocket (server->s);
		server->s = INVALID_SOCKET;
		free (txmsg->data);
		free (txmsg);
		return;
	    }
	    else if (status != txmsg->hdr.userlength)
	    {
		txmsg->hdr.userlength -= status;
		memcpy (txmsg->data, txmsg->data+status,
			txmsg->hdr.userlength);
		Append (txmsg, (void **)&server->txmsgs);
		return;
	    }
	    free (txmsg->data);
	}
	free (txmsg);
	server->lastTx = time(0);
	tisSysLog (1, 0, "TrendData message sent");
    }
}

/* check the socket for activity */
static void CheckSocketStatus (SERVER *server)
{
    int		status;
    SVRMSG	*msg;
    int		err;
#ifdef SOLARIS
    int		dataAvailable;
#endif SOLARIS

    if ((CheckServer (server) == FALSE) || server->s == INVALID_SOCKET)
    {
	return;
    }
    /* check socket for send stuff */
    while (server->txmsgs)
    {
	msg = server->txmsgs;
	status = send (server->s,
		    msg->data, msg->hdr.userlength, 0);
	if (status == SOCKET_ERROR)
	{
	    err = WSAGetLastError();
	    if (err != WSAEWOULDBLOCK)
	    {
		tisSysLog (0, 0, "Socket receive error: %m");
		/* close down session */
		closesocket (server->s);
		server->s = INVALID_SOCKET;
		return;
	    }
	}
	else if (status == msg->hdr.userlength)
	{
	    server->txmsgs = msg->next;
	    free (msg);
	    server->lastTx = time(0);
	    tisSysLog (1, 0, "TrendData message sent");
	}
	else
	{
	    msg->hdr.userlength -= status;
	    memcpy (msg->data, msg->data+status,
		    msg->hdr.userlength);
	    break; /* out of send pending messages loop */
	}
    }
    /* check socket for receive stuff */
#ifdef SOLARIS
    if (ioctl(server->s, FIONREAD, &dataAvailable) < 0)
    {
	closesocket (server->s);
	server->s = INVALID_SOCKET;
	return;
    }
    if (dataAvailable == 0)
	return;   
#endif SOLARIS
    if (server->rxlen < sizeof(ServHdr))
    {
	status = recv (server->s,
	    	    ((char *)&server->hdr) + server->rxlen,
	    	    sizeof(ServHdr),
		    0);
	if ((status > 0) && (server->rxlen + status) == sizeof(ServHdr))
	{
	    server->rxlen += status;
	    server->hdr.type = ntohs(server->hdr.type);
	    server->hdr.seq = ntohs(server->hdr.seq);
	    server->hdr.userlength = ntohs(server->hdr.userlength);
	    tisSysLog (1, 0, "Message header type %d, seq %d, len %d",
				 server->hdr.type, server->hdr.seq, server->hdr.userlength);
	    server->inmsg = malloc (sizeof(SVRMSG));
	    if (server->inmsg)
	    {
		server->inmsg->hdr = server->hdr;
		server->inmsg->next = NULL;
		server->inmsg->data = malloc (server->hdr.userlength);
		tisSysLog (1, 0, "Message header type %d, seq %d, len %d",
			 server->inmsg->hdr.type, server->inmsg->hdr.seq, server->inmsg->hdr.userlength);

		if (server->inmsg->data == NULL)
		{
		    /* error, close session */
		    closesocket (server->s);
		    server->s = INVALID_SOCKET;
		    return;
		}
	    }
	    else
	    {
		/* error, close session */
		closesocket (server->s);
		server->s = INVALID_SOCKET;
		return;
	    }
	}
	else if ((status == 0) || (status == SOCKET_ERROR))
	{
	    err = WSAGetLastError();
	    if ((status == 0) || ((status == SOCKET_ERROR) && (err != WSAEWOULDBLOCK)))
	    {
		tisSysLog (0, 0, "Socket receive error: %m");
		/* close down session */
		closesocket (server->s);
		server->s = INVALID_SOCKET;
		return;
	    }
	}
    }
    if (server->rxlen >= sizeof(ServHdr))
    {
	status = recv (server->s,
		server->inmsg->data + (server->rxlen - sizeof(ServHdr)),
		server->inmsg->hdr.userlength - (server->rxlen - sizeof(ServHdr)),
		0);
	if (status > 0)
	{
	    server->rxlen += status;

	    tisSysLog (1, 0, "Message body %02.2x %02.2x %02.2x %02.2x %02.2x %02.2x %02.2x %02.2x",
			*(unsigned char *)(server->inmsg->data), *(unsigned char *)(server->inmsg->data+1),
			*(unsigned char *)(server->inmsg->data+2), *(unsigned char *)(server->inmsg->data+3),
			*(unsigned char *)(server->inmsg->data+4), *(unsigned char *)(server->inmsg->data+5),
			*(unsigned char *)(server->inmsg->data+6), *(unsigned char *)(server->inmsg->data+7));
	    tisSysLog (1, 0, "Message body %02.2x %02.2x %02.2x %02.2x %02.2x %02.2x %02.2x %02.2x",
			*(unsigned char *)(server->inmsg->data+8), *(unsigned char *)(server->inmsg->data+9),
			*(unsigned char *)(server->inmsg->data+10), *(unsigned char *)(server->inmsg->data+11),
			*(unsigned char *)(server->inmsg->data+12), *(unsigned char *)(server->inmsg->data+13),
			*(unsigned char *)(server->inmsg->data+14), *(unsigned char *)(server->inmsg->data+15));
	    tisSysLog (1, 0, "Number of user bytes received %d", server->rxlen - sizeof(ServHdr));
	    tisSysLog (1, 0, "Message header type %d, seq %d, len %d",
			 server->inmsg->hdr.type, server->inmsg->hdr.seq, server->inmsg->hdr.userlength);

	    if (server->inmsg->hdr.userlength
		 == (server->rxlen - sizeof(ServHdr)))
	    {
		switch (server->inmsg->hdr.type)
		{
		case SRVREQ:
		    /* link data to input msg chain, return session */
		    Append (server->inmsg, (void **)&server->msgs);
		    server->inmsg = NULL;
		    server->rxlen = 0;
		    break;
		}
		server->lastRx = time(0);
		tisSysLog (1, 0, "TrendData message received");
	    }
	}
	else /* error, close session */
	{
	    err = WSAGetLastError();
	    if ((status == 0) || ((status == SOCKET_ERROR) && (err != WSAEWOULDBLOCK)))
	    {
		tisSysLog (0, 0, "Socket receive error: %m");
		/* close down session */
		closesocket (server->s);
		server->s = INVALID_SOCKET;
		return;
	    }
	}
    }
}

static int CheckServer (SERVER *server)
{
    SERVER *acp = activeServers;
    while (acp)
    {
	if (acp == server)
	    return TRUE;
	acp = acp->next;
    }
    return FALSE;
}

static void Append (void *item, void **list)
{
    if (*list == NULL)
    {
	*list = item;
    }
    else
    {
	void **ptr = list;
	while (*ptr)
	    ptr = *ptr;
	*ptr = item;
    }
}
