/*
   File:   serverdemo.c - demo data server, providing a directory listing
*/

#ifdef WIN32
#define pid_t	long
#endif

#include <stdio.h>
#include <memory.h>
#include <io.h>
#include <fcntl.h>
#include "server.h"
#include "demoMsgs.h"

SERVER *server;		    /* the server control structure */
#ifdef WIN32
/* socket handle for NT must be passed to service implementation */
extern SOCKET tisSvcSocket;
#endif WIN32

/*
**  PRIVATE DATA 
*/
int GetFileResp (FileMsg *reqmsg);

RespUnion    rspmsg;

main (int argc, char *argv[])
{
    ReqUnion	*reqmsg;
    int msglen;

#ifdef SRVDEBUG
    int wait = 1;
    while (wait)
	Sleep (1000);
#endif SRVDEBUG

    /* open server as asynchronous I/O */
    if ((server = OpenTisService ("FileData", FALSE, 0)) == NULL)
	exit (0);
#ifdef SRVDEBUG
    SetSocketNum (server, 8002);
#endif SRVDEBUG

    /* 
    **  Sit around forever processing requests (until the connection goes away)
    */
    while (CheckSvcSession (server))
    {
	if (reqmsg = GetReq (server))
	{
#ifdef DEBUG
	    printf ("Processing request %d", ((FileMsg *)reqmsg)->type);
#endif
	    ((FileMsg *)reqmsg)->type = ntohs(((FileMsg *)reqmsg)->type);

	    switch (((FileMsg *)reqmsg)->type)
	    {
		case GET_FILEDATA:
		    msglen = GetFileResp ((FileMsg *)reqmsg);
		    break;
		default:
		    printf ("Unrecognized request %d", ((FileMsg *)reqmsg)->type);
		    break;
	    }
	    SendRsp (server, (unsigned char *)&rspmsg, (short)msglen);
	    free (reqmsg);
	}
	else
	{
	    ServerWait (server, 5000);
	}
    }
#ifdef DEBUG
    printf ("Client session terminated");
#endif
    /* cleanup any open files, etc... */
    CloseSvcSession (server);
    return 0;
}

/* persistent storage */
short		file_result;
int	        num_file_segs = 0;
long		file_seg_size = 0;
int		current_file = -1;
int		current_file_size;

int GetFileResp (reqmsg)
FileMsg *reqmsg;
{
    FileRsp	*filersp;
    long	reqbase;

    filersp = (FileRsp*)&rspmsg;
    filersp->type = htons(GET_FILEDATA);
    /* fill in data */
    reqbase = ntohl (reqmsg->req_filePosition);
    if (reqbase == 0)
    {
	if (current_file != -1)
	    _close (current_file);
	current_file = _open (reqmsg->filename, _O_RDONLY|_O_BINARY);
	if (current_file != -1)
	{
	    reqbase = 0;
	    /* set current file size */
	    current_file_size = _filelength (current_file);
	    filersp->result = htons (TRUE);
	}
    }

    if (current_file == -1)
    {
	filersp->result = htons (FALSE);
    }

    if (ntohs(filersp->result) == TRUE)
    {
	filersp->fileTotalSize = htonl(current_file_size);
	filersp->filePosition = htonl(reqbase);
	filersp->fileDataSize = htons((short)_read (current_file, filersp->fileData, 1024));
	if (ntohs(filersp->fileDataSize) < 1024)
	{
	    _close (current_file);
	    current_file = -1;
	    if (ntohs(filersp->fileDataSize) == -1)
	    {
		filersp->result = htons (FALSE);
	    }
	}
    }
    return (sizeof(FileRsp));
}

