/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* svcmgr.c - service manager for TIS client/server programs.
**	This program is started by inetd, and waits for a message on fd 0.
**	The program will read the message and exec the program found in
**	the tisservice.conf file.
*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>

#define LOG_OPTIONSI	(LOG_INFO | LOG_LOCAL1)
#define LOG_OPTIONSE	(LOG_ERR | LOG_LOCAL1)

#include "tisservice.h"

SvcReq	inmsg;		/* input message information */
char *newargv [20];	/* up to 20 args */

char	inpstring [512];

int main (argc, argv)
{
    int ix, ix2;
    char *sp;
    FILE *fp;
    char *cp;
    char *execname;

    int client_len;
    struct sockaddr_in client_addr;

    openlog ("svcmgr", LOG_PID, LOG_LOCAL1);
    syslog (LOG_OPTIONSI, "Service manager started");

    /* get message, exit if error */
    client_len = sizeof(client_addr);
    for (ix = 0; ix < sizeof(inmsg); ix += ix2)
    {
	ix2 = read (0, ((char*)&inmsg)+ix, sizeof(inmsg)-ix, 0);
	if ( ix2 < 0)
	{
	    syslog (LOG_OPTIONSE, "Error reading message %m");
	    closelog ();
	    exit (1);
	}
    }
    if ((inmsg.type == SVCREQ) && (ntohl (inmsg.sftlvl) == SOFTWARE_LEVEL))
    {
	/* look up requested service, exit if not found */
	fp = fopen ("/tis4000db/tisservice.conf", "rt");
	if (fp)
	{
            while (!feof (fp))
            {
        	ix = 0;
		while (!feof (fp))
		{
                    inpstring [ix] = fgetc(fp);
                    if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		      || (inpstring [ix] == -1))
		    {
                	break;
 		    }
		    ix++;
 		    if (ix == 511)
		    {
			inpstring[511] = 0;
			syslog (LOG_OPTIONSE, "bad string in config file %s", inpstring);
			closelog ();
			exit (1); /* bad string in file. */
		    }
		}
		inpstring [ix] = 0;

		/* process input line */
		sp = inpstring;
		while (*sp && *sp == ' ')
		    sp++;
		if ((*sp == 0) || (*sp == '#'))
		    continue; /* blank line or comment! */
		/* process input- line is formatted as
		** <service_name> <executable> <args-as in inetd.conf>
		*/
		cp = strchr (sp, ' ');
		if (cp)
		{
		    *cp = 0;
		    if (strcmp (sp, inmsg.service) == 0)
		    {
			close (fp);
			/* exec the image */
			sp = cp+1; /* scan the executable name */
			while (*sp && *sp == ' ')
			    sp++;
			if (*sp == 0)
			{
			    /* error, no program image spec */
			    syslog (LOG_OPTIONSE, "Missing program in %s", inpstring);
			    closelog ();
			    exit (1); /* bad string in file. */
			}
			execname = sp;
			while (*sp && *sp != ' ')
			    sp++;
			if (*sp)
			    *sp++ = 0;

			syslog (LOG_OPTIONSI, "About to exec %s with args %s", execname, sp);
			/* build newargv array */
			for (ix = 0; *sp && ix < 19; ix++)
			{
			    while (*sp && *sp == ' ')
				sp++;
			    if (*sp)
				newargv [ix] = sp;
			    while (*sp && *sp != ' ')
			    {
				if (*sp == '"')
				{
				    while (*sp && *sp != '"')
				    {
					if (*sp == '\\')
					    sp++;
					sp++;
				    }
				}
				sp++;
			    }
			    if (*sp)
				*sp++ = 0;
			}
			newargv[ix] = 0;
			/* run it */
			execvp (execname, newargv);
			/* CAN'T GET HERE UNLESS ERROR!! */
			syslog (LOG_OPTIONSE, "Exec %s failed %m", execname);
			closelog ();
			close (0);
			exit (1);
		    }
		}
	    }
	}
	syslog (LOG_OPTIONSE, "Service %s not found", inmsg.service);
	close (fp);
    }
    else
    {
	syslog (LOG_OPTIONSE, "Can't open config file %m");
    }
    close (0);
    closelog ();
    return (1); /* not found */
}
