/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include "slm.h"

Widget w_main,w_toplevel,m_menubar,slmAreYouSure=0,slmInfo=0;
Widget m_exit,m_rtu,m_personnel,m_download;
Widget w_personnel,w_rtu,w_download,w_adm,w_pgm;
String fallbacks[] =
{
  "SLM*background: gray",
  "SLM*foreground: black",
  "SLM*fontList: 8x13bold",
  "SLM*XmMessageBox.background: CadetBlue",
  NULL
};
char slmPortBuff[SLM_PORT_BUFF_SIZE];
AVL_TREE  *slmNameTree;
RTU_ENTRY  *slmRTUNameTree = 0;
DB_TABLE_PTR  slmIocListTable;
struct	dbdatabase	DB_Database;			/* for DB functions */
int maxFd=-1;
int slmCaFd=-1;
int motifFd=-1;
int slmFd=-1;
int autoMode=0;
fd_set  useFds,saveFds;
struct termios	savetty;
struct timeval timeOut;

main(int argc,char *argv[])
{
  int status;
  XEvent slmXEvent;
  char msg[80];
  READ_CODE slmStat;

  XmString done,rtu,personnel,download;
  XtAppContext slmApp;
#ifdef SOLARIS
  XtSetLanguageProc(NULL,NULL,NULL);
#endif
  tisOpenLog ("slm", 6, 2, NULL, NULL);
  DB_InitDatabase("SLM");
  if(ca_task_initialize() == ECA_NORMAL)
    ca_add_fd_registration (slmReconnectCallback, NULL);
  slmNameTree = avl_init(slmNameCompare);
  slmBuildRTUTree();
#ifdef SOLARIS
  done = XmStringCreateLocalized("Exit");
  rtu = XmStringCreateLocalized("RTU");
  personnel = XmStringCreateLocalized("Personnel");
  download = XmStringCreateLocalized("Download");
#else
  done = XmStringCreateSimple("Exit");
  rtu = XmStringCreateSimple("RTU");
  personnel = XmStringCreateSimple("Personnel");
  download = XmStringCreateSimple("Download");
#endif
  w_toplevel = XtVaAppInitialize(&slmApp,"SLM",NULL,0,&argc,argv,fallbacks,NULL);
  w_main = XtVaCreateManagedWidget("main_window",xmMainWindowWidgetClass,w_toplevel,NULL);
  /* create main(only) menu */
  m_menubar = (Widget)XmVaCreateSimpleMenuBar(w_main,"menubar",
				      XmVaCASCADEBUTTON,personnel,'P',
				      XmVaCASCADEBUTTON,rtu,'R',
				      XmVaCASCADEBUTTON,download,'D',
				      XmVaCASCADEBUTTON,done,'E',
				      XmNspacing,20,
				      NULL);
  XmStringFree(done);
  XmStringFree(rtu);
  XmStringFree(personnel);
  XmStringFree(download);
  slmReadMain();
  if((slmFd = slmInitPort(argc,argv[1])) != -1)
    {
      if(slmFd > maxFd) maxFd = slmFd;
      FD_SET(slmFd,&useFds);
    }
  /* get menu children and attach callbacks */
  m_exit = XtNameToWidget(m_menubar,"button_3");
  XtAddCallback(m_exit,XmNactivateCallback,slmExit,(XtPointer)0);
  m_rtu = XtNameToWidget(m_menubar,"button_1");
  XtAddCallback(m_rtu,XmNactivateCallback,slmRTU,(XtPointer)0);
  m_personnel = XtNameToWidget(m_menubar,"button_0");
  XtAddCallback(m_personnel,XmNactivateCallback,slmPersonnel,(XtPointer)0);
  m_download = XtNameToWidget(m_menubar,"button_2");
  XtAddCallback(m_download,XmNactivateCallback,slmDownload,(XtPointer)0);
  /* create all dialog boxes */
  slmCreateRTU(w_toplevel);
  slmCreatePersonnel(w_toplevel);
  slmCreateDownload(w_toplevel);
  slmCreateAdmit(w_toplevel);
  slmCreateProgram(w_toplevel);
   if(slmFd == -1)
     XtUnmanageChild(slmPgmButton);
  XtManageChild(m_menubar);
  XtRealizeWidget(w_toplevel);
  /* get the X socket */
  motifFd = ConnectionNumber(XtDisplay(w_toplevel));
  if(motifFd > maxFd) maxFd = motifFd;
  FD_SET(motifFd,&useFds);
  memcpy((char *)&saveFds,(char *)&useFds,sizeof(fd_set));
  memset((char *)&slmPortBuff,0,SLM_PORT_BUFF_SIZE);
  timeOut.tv_sec = 1;
  timeOut.tv_usec = 0;
  /* process events and any incomming stuff on the serial port */
  while(1)
    {
      XFlush(XtDisplay(w_toplevel));
      if(XtAppPending(slmApp))
	{
	  XtAppNextEvent(slmApp,&slmXEvent);
	  XtDispatchEvent(&slmXEvent);
	}
      else
	{
	  memcpy((char *)&useFds,(char *)&saveFds,sizeof(fd_set));
	  status = select(maxFd+1,&useFds,0,0,&timeOut);		/* see if anything from port */
	  if(status < 0)
	    perror("slm");
	  else if(status > 0)
	    {
	      if((slmCaFd != -1) && FD_ISSET(slmCaFd,&useFds))
		{
		  ca_pend_event(0);					/* CA has data for us */
		}
	      if((slmFd != -1) && FD_ISSET(slmFd,&useFds))
		{
		  if((slmStat = slmReadFd(slmFd,slmPortBuff)) == PROGRAM_OK)	/* read port for msg from programer */
		    {
		      if(autoMode)
			slmPgmNextCard(slmPortBuff);			/* slmPortBuff contains name of last card */
		      else
			slmPgmClearItem(slmPortBuff);			/* deselect name but don't send next one */
		      memset((char *)&slmPortBuff,0,SLM_PORT_BUFF_SIZE);/* reset for next message */
		    }
		  else if((slmStat == IOCTL_ERROR) || (slmStat == READ_ERROR))
		    {
		      sprintf(msg,"I/O error on serial port %s",sys_errlist[errno]);
		      slmCreateInfo(msg,"SLM");
		    }
		}
	    }
	}
    }
}
