#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vstartupHEADER
#include vcontrolHEADER
#include vtimerHEADER
#include vprefHEADER
#include vlistviewHEADER
#include vrulerHEADER
#include vmemHEADER

#include "snlvu.h"
#include "snlvuApp.hxx"
#include "snlvuMain.hxx"
#include "snlvuInstances.hxx"
#include "snlvuEditInstance.hxx"
#include "snlvuUtils.hxx"

static int programNameFilter (filteredText *text, vtextSelection *sel,
			     const vchar **s, long *len);
static int parametersFilter (filteredText *text, vtextSelection *sel,
			     const vchar **s, long *len);
static int resourceFilter (filteredText *text, vtextSelection *sel,
			     const vchar **s, long *len);


vkindTYPED_FULL_DEFINITION(EditInstanceDialog, vdialog, NULL);

/*******************************************************************
*
*    EditInstanceDialog override for constructors init ()
*
*******************************************************************/

void EditInstanceDialog::LoadInit (vresource res)
{
    vmemSet ((void *)&instance, 0, sizeof (Instance));

    program = filteredTextitem::RequestSubLoad (this, snlvu_Program);
    parameters = filteredTextitem::RequestSubLoad (this, snlvu_Parameters);
    resourceName = filteredTextitem::RequestSubLoad (this, snlvu_Resource);
    okButton = EditInstanceOkButton::RequestSubLoad (this, snlvu_OkButton);
    cancelButton = EditInstanceCancelButton::RequestSubLoad (this, snlvu_CancelButton);

    // init inherited data
    vdialog::LoadInit (res);

    // set filters for text entry fields
    program -> SetItemFilterFunc (programNameFilter);
    parameters -> SetItemFilterFunc (parametersFilter);
    resourceName -> SetItemFilterFunc (resourceFilter);

    // store dialog ptr
    okButton -> StoreEditInstanceDialog (this);
    cancelButton -> StoreEditInstanceDialog (this);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    EditInstanceDialog destructor
*
*******************************************************************/

void EditInstanceDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
}

/*******************************************************************
*
*    EditInstanceDialog observe window event, looking for exit
*
*******************************************************************/

void EditInstanceDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	memset ((void *)&instance, 0, sizeof (instance));
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    EditInstanceDialog Open function
*
*******************************************************************/

void EditInstanceDialog::OpenEditInstance (MainDialog *mainDlg, EDITINSTANCE_TYPE typ, 
				Instance *inst)
{
    mainDialog = mainDlg;
    type = typ;
    memset ((void *)&instance, 0, sizeof (instance));
    if (inst != NULL)
    {
	instance = *inst;
    }
    BuildTitle ();

    InitFields ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open ();
    vevent::Process ();
}

/*******************************************************************
*
*    EditInstanceDialog set title function
*
*******************************************************************/

void EditInstanceDialog::BuildTitle ()
{
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource titlesRes = app -> GetTitlesResource ();
    vstr *dlgTitle = NULL;

    switch (type)
    {
	case DEFINE_NEW_INSTANCE:
	    dlgTitle = vresourceGetString (titlesRes, snlvu_DefineInstanceTitle);
	    break;
	case EDIT_INSTANCE:
	    dlgTitle = vresourceGetString (titlesRes, snlvu_EditInstanceTitle);
	    break;
	default:
	    return;
    }
    SetTitle (dlgTitle);
    if (dlgTitle != NULL)
    {
	vstrDestroy (dlgTitle);
    }
}


/*******************************************************************
*
*    EditInstanceDialog populate dialog fields of passed a
*	non-NULL instance pointer
*
*******************************************************************/

void EditInstanceDialog::InitFields ()
{
    if (instance.programName[0] == 0)
    {
	return;
    }
    program -> SetTextScribed (vcharScribeSystem (instance.programName));
    parameters -> SetTextScribed (vcharScribeSystem (instance.parameters));
    resourceName -> SetTextScribed (vcharScribeSystem (instance.resourceName));
}


/*******************************************************************
*
*    EditInstanceDialog funcs to update instance
*
*******************************************************************/

void EditInstanceDialog::SetProgramName (vstr *progName)
{
    vcharExportBounded (vcharScribeSystemExport (progName),
			(void *)instance.programName,
			PROGRAM_NAME_SIZE);
    instance.programName[PROGRAM_NAME_SIZE] = 0;
}

void EditInstanceDialog::SetParameters (vstr *params)
{
    vcharExportBounded (vcharScribeSystemExport (params),
			(void *)instance.parameters,
			PARAMETERS_SIZE);
    instance.parameters[PARAMETERS_SIZE] = 0;
}

void EditInstanceDialog::SetResourceName (vstr *resName)
{
    vcharExportBounded (vcharScribeSystemExport (resName),
			(void *)instance.resourceName,
			RESOURCE_NAME_SIZE);
    instance.resourceName[RESOURCE_NAME_SIZE] = 0;
}


vkindTYPED_FULL_DEFINITION(EditInstanceOkButton, vbutton, NULL);

/*******************************************************************
*
*    EditInstanceOkButton
*
*******************************************************************/

void EditInstanceOkButton::ObserveDialogItem (vevent *event)
{
    vstr *vPtr;

    vPtr = editinstanceDialog -> program -> GetTextAsString ();
    if ((vcharLength (vPtr)) == 0)
    {
	snlvuMessageCenter (editinstanceDialog, SNLVU_INFO_NOTICE,
					snlvu_NeedProgramNameMsg);
	return;
    }
    editinstanceDialog -> SetProgramName (vPtr);
    vstrDestroy (vPtr);

    vPtr = editinstanceDialog -> parameters -> GetTextAsString ();
    editinstanceDialog -> SetParameters (vPtr);
    vstrDestroy (vPtr);

    vPtr = editinstanceDialog -> resourceName -> GetTextAsString ();
    editinstanceDialog -> SetResourceName (vPtr);
    vstrDestroy (vPtr);

    vevent::StopProcessing ();
}



vkindTYPED_FULL_DEFINITION(EditInstanceCancelButton, vbutton, NULL);

/*******************************************************************
*
*    EditInstanceCancelButton
*	destroy data, so caller has no access to it
*
*******************************************************************/

void EditInstanceCancelButton::ObserveDialogItem (vevent *event)
{
    Instance *instance = editinstanceDialog -> GetInstance ();
    memset ((void *)instance, 0, sizeof (Instance));
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    Filters for the text entry field
*
*******************************************************************/

static int programNameFilter (filteredText *text, vtextSelection *sel,
			     const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, PROGRAM_NAME_SIZE));
}

static int parametersFilter (filteredText *text, vtextSelection *sel,
			     const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, PARAMETERS_SIZE));
}

static int resourceFilter (filteredText *text, vtextSelection *sel,
			   const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, RESOURCE_NAME_SIZE));
}

