#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vstartupHEADER
#include vcontrolHEADER
#include vtimerHEADER
#include vprefHEADER
#include vlistviewHEADER
#include vrulerHEADER
#include vmemHEADER

#include "snlvu.h"
#include "snlvuApp.hxx"
#include "snlvuMain.hxx"
#include "snlvuEnter.hxx"
#include "snlvuUtils.hxx"

static int moduleNameFilter (filteredText *text, vtextSelection *sel,
			     const vchar **s, long *len);


vkindTYPED_FULL_DEFINITION(EnterDialog, vdialog, NULL);

/*******************************************************************
*
*    EnterDialog override for constructors init ()
*
*******************************************************************/

void EnterDialog::LoadInit (vresource res)
{
    vmemSet (enteredData, 0, sizeof (enteredData));

    text = filteredTextitem::RequestSubLoad (this, snlvu_EnterText);
    okButton = EnterOkButton::RequestSubLoad (this, snlvu_OkButton);
    cancelButton = EnterCancelButton::RequestSubLoad (this, snlvu_CancelButton);

    // init inherited data
    vdialog::LoadInit (res);

    // store dialog ptr
    okButton -> StoreEnterDialog (this);
    cancelButton -> StoreEnterDialog (this);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    EnterDialog destructor
*
*******************************************************************/

void EnterDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
}

/*******************************************************************
*
*    EnterDialog observe window event, looking for exit
*
*******************************************************************/

void EnterDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vmemSet (enteredData, 0, sizeof (enteredData));
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    EnterDialog Open function
*
*******************************************************************/

void EnterDialog::OpenEnter (MainDialog *mainDlg, ENTER_TYPE typ, char *arg)
{
    mainDialog = mainDlg;
    type = typ;
    BuildTitle (arg);
    SetFilter ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open ();
    vevent::Process ();
}

/*******************************************************************
*
*    EnterDialog set title function
*
*******************************************************************/

void EnterDialog::BuildTitle (char *arg)
{
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource titlesRes = app -> GetTitlesResource ();
    vstr *titleFormat = NULL;
    vstr *argument = NULL;
    vstr *enterTitle;

    switch (type)
    {
	case ENTER_MODULE_NAME:
	    enterTitle = vresourceGetString (titlesRes, snlvu_EnterModuleNameTitle);
	    if (arg != NULL)
	    {
		text -> SetTextScribed (vcharScribeSystem (arg));
	    }
	    break;
	case ENTER_VALUE:
	    titleFormat = vresourceGetString (titlesRes, snlvu_EnterValueTitle);
	    argument = vstrCloneScribed (vcharScribeSystem (arg));
	    enterTitle  = vstrCloneScribed (vcharScribeFormat (titleFormat, argument));
	    break;
	case NEW_IOC:
	    enterTitle = vresourceGetString (titlesRes, snlvu_NewIocTitle);
	    break;
	default:
	    return;
    }
    SetTitle (enterTitle);
    if (titleFormat != NULL)
    {
	vstrDestroy (titleFormat);
    }
    if (argument != NULL)
    {
	vstrDestroy (argument);
    }
    vstrDestroy (enterTitle);
}


/*******************************************************************
*
*    EnterDialog set filter functions for the text entry field
*
*******************************************************************/

void EnterDialog::SetFilter (void)
{
    switch (type)
    {
	case ENTER_MODULE_NAME:
	    text -> SetItemFilterFunc (moduleNameFilter);
	    break;
	case ENTER_VALUE:
	    break;
	case NEW_IOC:
	    text -> SetItemFilterFunc (iocnameFilter);
	    break;
	default:
	    return;
    }
}


vkindTYPED_FULL_DEFINITION(EnterOkButton, vbutton, NULL);

/*******************************************************************
*
*    EnterOkButton
*
*******************************************************************/

void EnterOkButton::ObserveDialogItem (vevent *event)
{
    vchar *enteredData = enterDialog -> GetEnteredData ();
    filteredTextitem *text = enterDialog -> GetFilteredText ();
    vstr *enteredText = text -> GetTextAsString ();
    if ((vcharLength (enteredText)) == 0)
    {
	snlvuMessageCenter (enterDialog, SNLVU_INFO_NOTICE,
			    snlvu_NothingEnteredMsg);
	vstrDestroy (enteredText);
	return;
    }
    ENTER_TYPE type = enterDialog -> GetType ();
    if (type == ENTER_MODULE_NAME)
    {
	vchar *dec;
	if ((dec = vcharSearchChar (enteredText, (int)'.')) == NULL)
	{
	    snlvuMessageCenter (enterDialog, SNLVU_INFO_NOTICE,
				snlvu_ModuleNameErr);
	    vstrDestroy (enteredText);
	    return;
	}
    }
    vcharCopyBounded (enteredText, enteredData, ENTERED_DATA_SIZE);
    enteredData[ENTERED_DATA_SIZE] = 0;
    vstrDestroy (enteredText);
    vevent::StopProcessing ();
}



vkindTYPED_FULL_DEFINITION(EnterCancelButton, vbutton, NULL);

/*******************************************************************
*
*    EnterCancelButton
*	destroy data, so caller has no access to it
*
*******************************************************************/

void EnterCancelButton::ObserveDialogItem (vevent *event)
{
    vchar *enteredData = enterDialog -> GetEnteredData ();
    vmemSet (enteredData, 0, ENTERED_DATA_SIZE);
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    Filters for the text entry field
*
*******************************************************************/

static int moduleNameFilter (filteredText *text, vtextSelection *sel,
			     const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, MODULE_NAME_SIZE));
}

