#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vstartupHEADER
#include vcontrolHEADER
#include vtimerHEADER
#include vprefHEADER
#include vlistviewHEADER
#include vrulerHEADER

extern "C"
{
#include "usrseq.h"
}

#include "snlvu.h"
#include "snlvuApp.hxx"
#include "snlvuMain.hxx"
#include "snlvuEnter.hxx"
#include "snlvuExpand.hxx"
#include "snlvuUtils.hxx"


vkindTYPED_FULL_DEFINITION(ExpandDialog, vdialog, NULL);

/*******************************************************************
*
*    ExpandDialog override for constructors init ()
*
*******************************************************************/

void ExpandDialog::LoadInit (vresource res)
{
    // init stuff
    InUse = vFALSE;

    // RequestSubLoad returns ptr to uninited dialog object
    // gets init'd when dialog gets init'd by LoadInit()

    expandListView = ExpandListView::RequestSubLoad (this, snlvu_ExpandValueListView);

    modifyButton = ExpandModifyButton::RequestSubLoad (this, snlvu_ModifyButton);

    exitButton = ExpandExitButton::RequestSubLoad (this, snlvu_ExitButton);

    // init inherited data
    vdialog::LoadInit (res);

    // store ptrs to main dialog
    expandListView -> StoreExpandDialog (this);
    modifyButton -> StoreExpandDialog (this);
    exitButton -> StoreExpandDialog (this);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    ExpandDialog Destroy function
*
*******************************************************************/

void ExpandDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
}

/*******************************************************************
*
*    ExpandDialog observe window event, looking for exit
*
*******************************************************************/

void ExpandDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	InUse = vTRUE;
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    ExpandDialog 
*
*******************************************************************/

void ExpandDialog::OpenExpand (MainDialog *dlg, char *progName,
			       long tid, REPVARS_REPLY *var)
{
    mainDialog = dlg;
    strncpy (programName, progName, sizeof (programName));
    programName[PROGRAM_NAME_SIZE] = 0;
    TaskId = tid;
    variable = var;

    if (mainDialog -> modifyVariables == vFALSE)
    {
	modifyButton -> Disable ();
    }

    SetExpandTitle ();

    UpdateExpandList ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open ();
    vevent::Process ();
}


/*******************************************************************
*
*    ExpandDialog function to set the dialog title 
*
*******************************************************************/

void ExpandDialog::SetExpandTitle ()
{
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource titleRes = app -> GetTitlesResource ();

    vstr *titleFormat = vresourceGetString (titleRes, snlvu_ExpandValuesTitle);
    vstr *varName = vstrCloneScribed (vcharScribeSystem (variable -> Name));
    vstr *expandTitle = vstrCloneScribed (vcharScribeFormat (titleFormat, varName));
    SetTitle (expandTitle);
    vstrDestroy (varName);
    vstrDestroy (titleFormat);
    vstrDestroy (expandTitle);
}


/*******************************************************************
*
*    ExpandDialog function to update the expand list in the main
*	view
*
*******************************************************************/

void ExpandDialog::UpdateExpandList (void)
{
    char *var;
    int	 length;

    if (snlMon_ReportVariable (programName, TaskId, variable -> Name,
			       &var, &length))
    {
	expandListView -> BuildList (var, length);
	free ((void *)var);
    }
}



vkindTYPED_FULL_DEFINITION(ExpandListView, vlistview, NULL);

/*******************************************************************
*
*    ExpandListView 
*
*******************************************************************/

void ExpandListView::Destroy (void)
{
    vliststr *list = (vliststr *)(GetList ());
    if (list != NULL)
    {
	SetList (NULL);
	delete list;
    }
    vlistview::Destroy ();
}


/*******************************************************************
*
*    ExpandListView, override
*
*******************************************************************/

void ExpandListView::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	expandDialog -> InUse = vTRUE;
    }
}


/*******************************************************************
*
*    ExpandListView, build the list
*
*******************************************************************/

void ExpandListView::BuildList (char *Var, int length)
{
    vliststr *newList = new vliststr;
    vliststr *oldList = (vliststr *)(GetList ());

    REPVARS_REPLY *Variable = expandDialog -> GetVariable ();

    int Count;
    if (Variable -> NumElements == 0)
    {
	Count = 1;
    }
    else
    {
	Count = Variable->NumElements;
    }

    char *q = Var;
    char AValue [50];
    char line[100];
    int  NumChars = 0;
    int  j;
    int  row = 0;

    for (j = 0 ; j < Count ; j++)
    {
	switch (Variable -> Type)
	{
	    case V_CHAR:
		sprintf (AValue, "%c", *q);
		break;
	    case V_INT:
		sprintf (AValue, "%d", *((int *)q));
		break;
	    case V_LONG:
		sprintf (AValue, "%ld", *((long *)q));
		break;
	    case V_SHORT:
		sprintf (AValue, "%d", (int)*((short *)q));
		break;
	    case V_FLOAT:
		sprintf (AValue, "%.2f", *((float *)q));
		break;
	    case V_DOUBLE:
		sprintf (AValue, "%.2lf", *((double *)q));
		break;
	    case V_STRING:
		sprintf (AValue, "%s", q);
		break;
	}
	sprintf (line, "%s [%d] = %-35.35s", Variable -> Name, j, AValue);
	newList -> SetCellValueScribed (row, 0, vcharScribeSystem (line));
	row++;
	q += Variable -> Size;
    }

    SetList (newList);
    if (oldList != NULL)
    {
	delete oldList;
    }
    int rowCount = newList -> GetRowCount ();
    if (rowCount > 0)
    {
	vlistSelection *selected = GetSelection ();
	newList -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
    }
}


vkindTYPED_FULL_DEFINITION(ExpandModifyButton, vbutton, NULL);

/*******************************************************************
*
*    ExpandModifyButton
*
*******************************************************************/

void ExpandModifyButton::ObserveDialogItem (vevent *event)
{
    ExpandListView	*expandListView = expandDialog -> GetExpandListView ();
    vlistSelection	*selection = expandListView -> GetSelection ();
    vliststr		*list = (vliststr *)(expandListView -> GetList ());
    REPVARS_REPLY	*Variable = expandDialog -> GetVariable ();
    char		*programName = expandDialog -> GetProgramName ();
    long		taskId = expandDialog -> GetTaskId ();

    expandDialog -> InUse = vTRUE;
    if (list == NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	expandDialog -> InUse = vFALSE;
	return;
    }

    int rowCount = list -> GetRowCount ();

    if (rowCount == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	expandDialog -> InUse = vFALSE;
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	expandDialog -> InUse = vFALSE;
	return;
    }

    int row, col;
    list -> GetSelectedCell (selection, &row, &col);

    MainDialog *mainDialog = expandDialog -> GetMainDialog ();
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource enterRes = vresourceGet (appRes, snlvu_EnterDialog);
    EnterDialog *enterDialog = new EnterDialog (enterRes);
    char VarName [50];
    sprintf (VarName, "%s [%d]", Variable -> Name, row);
    enterDialog -> OpenEnter (mainDialog, ENTER_VALUE, VarName);
    vchar *enteredData = enterDialog -> GetEnteredData ();
    char newValue[100];
    vcharExportBounded (vcharScribeSystemExport (enteredData),
			newValue,
			sizeof (newValue));
    newValue[sizeof(newValue) - 1] = 0;
    delete (enterDialog);

    char *q = newValue;
    // remove leading blanks
    while (*q)
    {
	if (*q != ' ')
	    break;
	q++;
    }

    // Get the current value from the server, substitute the changed element and
    // send back to server.

    char *Var;
    int	 Length;

    snlMon_ReportVariable (programName, taskId, Variable -> Name, &Var, &Length);

    if (strlen (q))
    {
	char *MsgBuf = Var + row * Variable -> Size;
	switch (Variable->Type)
	{
	    case V_CHAR:
		*MsgBuf = *q;
		break;
	    case V_INT:
		*((int *)MsgBuf) = atoi (newValue);
		break;
	    case V_LONG:
		*((long *)MsgBuf) = atol (newValue);
		break;
	    case V_SHORT:
		*((short *)MsgBuf) = atoi (newValue);
		break;
	    case V_FLOAT:
		*((float *)MsgBuf) = (float)(atof (newValue));
		break;
	    case V_DOUBLE:
		*((double *)MsgBuf) = atof (newValue);
		break;
	    case V_STRING:
		strcpy (MsgBuf, newValue);
		break;
	}
	int update = vTRUE;

	if (snlMon_ChangeVariable (programName, taskId, Variable->Name, Var, Length) == 0)
	{
	    snlvuMessageCenter (expandDialog, SNLVU_INFO_NOTICE,
							snlvu_VariableModifyError);
	    update = vFALSE;
	}
	free ((void *)Var);
	if (update == vTRUE)
	{
	    expandDialog -> UpdateExpandList ();
	}
    }
    expandDialog -> InUse = vFALSE;
}


vkindTYPED_FULL_DEFINITION(ExpandExitButton, vbutton, NULL);

/*******************************************************************
*
*    ExpandExitButton
*
*******************************************************************/

void ExpandExitButton::ObserveDialogItem (vevent *event)
{
    expandDialog -> InUse = vTRUE;
    // stop secondary event loop started in open()
    vevent::StopProcessing ();
}

