#ifndef	INCsnlvuinstancesdialogHXX
#define	INCsnlvuinstancesdialogHXX

typedef	struct
{
    char	programName[PROGRAM_NAME_SIZE + 1];
    char	parameters[PARAMETERS_SIZE + 1];
    char	resourceName[RESOURCE_NAME_SIZE + 1];
} Instance;

class InstancesListView;
class InstancesStartButton;
class InstancesDefineButton;
class InstancesDeleteButton;
class InstancesEditButton;
class InstancesExitButton;

/*
 *    NOTE:
 *	The Instances dialog use a vlist to store all
 *	instances, it is kept in sync with the
 *	list of instances as displayed in the
 *	InstancesDialog -> instancesListView
 */


class InstancesDialog : public vdialog
{
    MainDialog			*mainDialog;
    vruler			*instancesRuler;
    InstancesListView		*instancesListView;
    InstancesStartButton	*startButton;
    InstancesDefineButton	*defineButton;
    InstancesDeleteButton	*deleteButton;
    InstancesEditButton		*editButton;
    InstancesExitButton		*exitButton;

    void	LoadInstances (void);
    vlist	*instances;
    int		instancesModified;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(InstancesDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(InstancesDialog, vdialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override 
    void Destroy ();

    // vwindow overide
    void ObserveWindow (vevent *event);

    // override a dialog reconfigure event
    int HandleConfigure (vevent *event);

    // open 
    void OpenInstances (MainDialog *dlg);

    MainDialog *GetMainDialog (void)
	{ return (mainDialog); }

    InstancesListView *GetInstancesListView (void)
	{ return (instancesListView); }

    vruler *GetInstancesRuler (void)
	{ return (instancesRuler); }

    void UpdateInstancesList (void);

    int VerifyInstance (Instance inst, int row);

    // helper function for instances array
    void AddInstance (Instance inst);
    void DeleteInstance (int row);
    void GetInstance (int instNumber, Instance *inst);
    vlist *GetInstancesList (void)
	{ return (instances); }
    void StoreInstances (void);
    void UpdateInstance (int row, Instance *inst);
};


class  InstancesListView : public vlistview
{
    InstancesDialog	*instancesDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(InstancesListView);
    vloadableINLINE_CONSTRUCTORS(InstancesListView, vlistview);

    void BuildList (void);

    // set columns width based on current ruler item widths
    void SetColumnsWidths (void);

    void StoreInstancesDialog (InstancesDialog *dlg)
	{ instancesDialog = dlg; }
};


class  InstancesStartButton : public vbutton
{
    InstancesDialog	*instancesDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(InstancesStartButton);
    vloadableINLINE_CONSTRUCTORS(InstancesStartButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreInstancesDialog (InstancesDialog *dlg)
	{ instancesDialog = dlg; }
};


class  InstancesDefineButton : public vbutton
{
    InstancesDialog	*instancesDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(InstancesDefineButton);
    vloadableINLINE_CONSTRUCTORS(InstancesDefineButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreInstancesDialog (InstancesDialog *dlg)
	{ instancesDialog = dlg; }
};


class  InstancesDeleteButton : public vbutton
{
    InstancesDialog	*instancesDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(InstancesDeleteButton);
    vloadableINLINE_CONSTRUCTORS(InstancesDeleteButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreInstancesDialog (InstancesDialog *dlg)
	{ instancesDialog = dlg; }
};


class  InstancesEditButton : public vbutton
{
    InstancesDialog	*instancesDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(InstancesEditButton);
    vloadableINLINE_CONSTRUCTORS(InstancesEditButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreInstancesDialog (InstancesDialog *dlg)
	{ instancesDialog = dlg; }
};


class  InstancesExitButton : public vbutton
{
    InstancesDialog	*instancesDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(InstancesExitButton);
    vloadableINLINE_CONSTRUCTORS(InstancesExitButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreInstancesDialog (InstancesDialog *dlg)
	{ instancesDialog = dlg; }
};

#endif	INCsnlvuinstancesdialogHXX

