#ifndef	INCsnlvumaindialogHXX
#define	INCsnlvumaindialogHXX

class ChannelsDialog;	// snlvuChannels.hxx
class ModulesDialog;	// snlvuModules.hxx
class VariablesDialog;	// snlvuVariables.hxx
class ExpandDialog;	// snlvuExpand.hxx
class StateSetsDialog;	// snlvuStateSets.hxx
class MainListView;
class ModulesButton;
class StopButton;
class ChannelsButton;
class VariablesButton;
class SelectButton;
class StateSetsButton;
class InstancesButton;

// enum values used for the vcommandSelector::Issue
enum
{
    EXIT_MENUITEM = 0,
    MENUITEM_COUNT
};
enum ConnectStatus
{
    SNLVU_NOTCONNECTED = 0,
    SNLVU_CONNECTED
};

class MainDialog : public vdialog
{
    // pointer to application
    snlvuApplication	*snlvuApp;

    // resource for the main dialog
    vresource		mainDialogResource;

    // pointers to items in the main dialog
    vmenu		*helpMenu;
    vruler		*mainRuler;
    MainListView	*mainListView;
    ModulesButton	*modulesButton;
    StopButton		*stopButton;
    ChannelsButton	*channelsButton;
    VariablesButton	*variablesButton;
    SelectButton	*selectButton;
    StateSetsButton	*stateSetsButton;
    InstancesButton	*instancesButton;

    // place to store internal names for main dialog menu items
    const vname		*commandNames[MENUITEM_COUNT];

    // other stuff and functions
    void		BuildAboutMenu (void);
    int			loadInitErrors;
    void		SetMainTitle (ConnectStatus connected);
    TisAbout		*tisAbout;

    // remember the current programs displayed
    REPPROG_REPLY	*PrevPrograms;
    int			PrevNumProg;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(MainDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(MainDialog, vdialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override 
    void Destroy ();

    // vwindow overide
    void ObserveWindow (vevent *event);

    // override a dialog reconfigure event
    int HandleConfigure (vevent *event);

    // open 
    void OpenMain (snlvuApplication *app);

    snlvuApplication *GetApplication (void)
	{ return (snlvuApp); }

    // return the snlvuApp applications resource
    vresource GetSnlvuResource (void)
	{ return (snlvuApp -> GetSnlvuResource()); }

    // return the main dialog's resource
    vresource GetMainDialogResource (void)
	{ return (mainDialogResource); }

    MainListView *GetMainListView (void)
	{ return (mainListView); }

    int GetLoadInitErrors (void)
	{ return (loadInitErrors); }

    vruler *GetMainRuler (void)
	{ return (mainRuler); }

    void UpdateProgramList (int ignorePrev);

    // other dialogs
    ChannelsDialog	*channelsDialog;
    ModulesDialog	*modulesDialog;
    VariablesDialog	*variablesDialog;
    ExpandDialog	*expandDialog;
    StateSetsDialog	*stateSetsDialog;

    // privileges
    int	loadModules;
    int	unloadModules;
    int	modifyVariables;
    int	pauseStateSets;
    int	continueStateSets;
    int	startInstance;
    int	defineInstance;
    int	deleteInstance;
    int	editInstance;
};


class  MainCommand : public vcommandSelector
{

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(MainCommand);
    vloadableINLINE_CONSTRUCTORS(MainCommand, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};


class  MainListView : public vlistview
{
    MainDialog	*mainDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(MainListView);
    vloadableINLINE_CONSTRUCTORS(MainListView, vlistview);

    // override 
    void Destroy ();

    // override
    void DrawCellContents (int row, int col, int selected, vrect *cellRect);

    void BuildList (REPPROG_REPLY *programs, int numProgs);

    // set columns width based on current ruler item widths
    void SetColumnsWidths (void);

    void StoreMainDialog (MainDialog *dlg)
	{ mainDialog = dlg; }
};


class  ModulesButton : public vbutton
{
    MainDialog	*mainDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(ModulesButton);
    vloadableINLINE_CONSTRUCTORS(ModulesButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreMainDialog (MainDialog *dlg)
	{ mainDialog = dlg; }
};

class  StopButton : public vbutton
{
    MainDialog	*mainDialog;

public:
    // macros for standard versions of the
    // constructors and destructors
    vkindTYPED_FULL_DECLARATION(StopButton);
    vloadableINLINE_CONSTRUCTORS(StopButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreMainDialog (MainDialog *dlg)
	{ mainDialog = dlg; }
};


class  ChannelsButton : public vbutton
{
    MainDialog	*mainDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(ChannelsButton);
    vloadableINLINE_CONSTRUCTORS(ChannelsButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreMainDialog (MainDialog *dlg)
	{ mainDialog = dlg; }
};

class  VariablesButton : public vbutton
{
    MainDialog	*mainDialog;

public:
    // macros for standard versions of the
    // constructors and destructors
    vkindTYPED_FULL_DECLARATION(VariablesButton);
    vloadableINLINE_CONSTRUCTORS(VariablesButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreMainDialog (MainDialog *dlg)
	{ mainDialog = dlg; }
};


class  SelectButton : public vbutton
{
    MainDialog	*mainDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(SelectButton);
    vloadableINLINE_CONSTRUCTORS(SelectButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreMainDialog (MainDialog *dlg)
	{ mainDialog = dlg; }
};

class  StateSetsButton : public vbutton
{
    MainDialog	*mainDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(StateSetsButton);
    vloadableINLINE_CONSTRUCTORS(StateSetsButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreMainDialog (MainDialog *dlg)
	{ mainDialog = dlg; }
};


class  InstancesButton : public vbutton
{
    MainDialog	*mainDialog;

public:

    // macros for standard versions of the
    // constructors and destructors

    vkindTYPED_FULL_DECLARATION(InstancesButton);
    vloadableINLINE_CONSTRUCTORS(InstancesButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    void StoreMainDialog (MainDialog *dlg)
	{ mainDialog = dlg; }
};


#endif	INCsnlvumaindialogHXX

