#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vstartupHEADER
#include vcontrolHEADER
#include vtimerHEADER
#include vprefHEADER
#include vlistviewHEADER
#include vrulerHEADER

#include "snlvu.h"
#include "snlvuApp.hxx"
#include "snlvuMain.hxx"

#include "snlvuEnter.hxx"
#include "snlvuModules.hxx"
#include "snlvuUtils.hxx"


vkindTYPED_FULL_DEFINITION(ModulesDialog, vdialog, NULL);

/*******************************************************************
*
*    ModulesDialog override for constructors init ()
*
*******************************************************************/

void ModulesDialog::LoadInit (vresource res)
{
    modulesListView = ModulesListView::RequestSubLoad (this,
						snlvu_ModulesListView);
    loadButton = LoadButton::RequestSubLoad (this, snlvu_LoadButton);
    unloadButton = UnloadButton::RequestSubLoad (this, snlvu_UnloadButton);
    exitButton = ExitButton::RequestSubLoad (this, snlvu_ExitButton);

    // init inherited data
    vdialog::LoadInit (res);

    // store dialog ptr
    modulesListView -> StoreModulesDialog (this);
    loadButton -> StoreModulesDialog (this);
    unloadButton -> StoreModulesDialog (this);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    ModulesDialog Destroy function
*
*******************************************************************/

void ModulesDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    ModulesDialog observe window event, looking for exit
*
*******************************************************************/

void ModulesDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
	vevent::StopProcessing ();
}


/*******************************************************************
*
*    ModulesDialog 
*
*******************************************************************/

void ModulesDialog::OpenModules (MainDialog *dlg)
{
    mainDialog = dlg;

    if (mainDialog -> loadModules == vFALSE)
    {
	loadButton -> Disable ();
    }
    if (mainDialog -> unloadModules == vFALSE)
    {
	unloadButton -> Disable ();
    }

    // set the title, need to get the connected IOC name
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource titlesRes = app -> GetTitlesResource ();
    vstr *titleFormat = vresourceGetString (titlesRes, snlvu_ModulesTitle);
    char *iocName = app -> GetIocName ();
    vstr *iocname = vstrCloneScribed (vcharScribeSystem (iocName));
    vstr *moduleTitle = vstrCloneScribed (vcharScribeFormat (titleFormat, iocname));
    SetTitle (moduleTitle);
    vstrDestroy (titleFormat);
    vstrDestroy (iocname);
    vstrDestroy (moduleTitle);

    modulesListView -> BuildList ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open ();
    vevent::Process ();
}


vkindTYPED_FULL_DEFINITION(ModulesListView, vlistview, NULL);

/*******************************************************************
*
*    ModulesListView class
*
*******************************************************************/

void ModulesListView::Destroy (void)
{
    vliststr *list = (vliststr *)(GetList ());
    if (list != NULL)
    {
	SetList (NULL);
	delete list;
    }
    vlistview::Destroy ();
}

/*******************************************************************
*
*    ModulesListView class
*
*******************************************************************/

void ModulesListView::BuildList (void)
{
    vliststr *list = (vliststr *)(GetList ());
    vliststr *newlist = new vliststr;
    REPMOD_REPLY *modules, *p;
    int numModules = 0;
    int i;
    int row = 0;

    // get info fom IOC
    snlMon_ReportModules (&modules, &numModules);
    p = modules;
    for (i = 0; i < numModules; i++, p++)
    {
	newlist -> SetCellValueScribed (row, 0, vcharScribeSystem (p -> ModuleName));
	row++;
    }
    newlist -> Sort (vlistROW, 0, NULL);
    SetList (NULL);
    SetList (newlist);
    if (list != NULL)
    {
	delete list;
    }
}


vkindTYPED_FULL_DEFINITION(LoadButton, vbutton, NULL);

/*******************************************************************
*
*    LoadButton
*
*******************************************************************/

void LoadButton::ObserveDialogItem (vevent *event)
{
    MainDialog *mainDialog = modulesDialog -> GetMainDialog ();
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource enterRes = vresourceGet (appRes, snlvu_EnterDialog);
    EnterDialog *enterDialog = new EnterDialog (enterRes);
    enterDialog -> OpenEnter (mainDialog, ENTER_MODULE_NAME, NULL);
    vchar *enteredData = enterDialog -> GetEnteredData ();
    if (enteredData[0] != 0)
    {
	char moduleName[MODULE_NAME_SIZE + 1];
	vcharExportBounded (vcharScribeSystemExport (enteredData),
			    moduleName,
			    MODULE_NAME_SIZE);
	moduleName[MODULE_NAME_SIZE] = 0;
	if ((snlMon_Load (moduleName)) == 0)
	{
	    snlvuMessageCenter (modulesDialog, SNLVU_INFO_NOTICE, 
				snlvu_ModuleNotLoadedMsg, enteredData);
	    return;
	}
	ModulesListView *view = modulesDialog -> GetListView ();
	view -> BuildList ();
    }
    delete (enterDialog);
}


vkindTYPED_FULL_DEFINITION(UnloadButton, vbutton, NULL);

/*******************************************************************
*
*    UnloadButton
*
*******************************************************************/

void UnloadButton::ObserveDialogItem (vevent *event)
{
    ModulesListView *view = modulesDialog -> GetListView ();
    vlistSelection *selection = view -> GetSelection ();
    vliststr *list = (vliststr *)(view -> GetList ());
    if (list == NULL)
    {
	snlvuMessageCenter (modulesDialog, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }
    int rowCount = list -> GetRowCount ();
    if (rowCount == 0)
    {
	snlvuMessageCenter (modulesDialog, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (modulesDialog, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	return;
    }
    int row, col;
    list -> GetSelectedCell (selection, &row, &col);
    vchar *moduleName = (vchar *)(list -> GetCellValue (row, col));
    char modulename[MODULE_NAME_SIZE + 1];
    vcharExportBounded (vcharScribeSystemExport (moduleName),
			modulename,
			MODULE_NAME_SIZE);
    modulename[MODULE_NAME_SIZE] = 0;
    int result = snlvuMessageCenter (modulesDialog, SNLVU_CAUTION_NOTICE,
				     snlvu_ModuleStoppedMsg, moduleName);
    if (result == vTRUE)
    {
	snlMon_Unload (modulename);
	view -> BuildList ();
    }
}


vkindTYPED_FULL_DEFINITION(ExitButton, vbutton, NULL);

/*******************************************************************
*
*    ExitButton
*
*******************************************************************/

void ExitButton::ObserveDialogItem (vevent *event)
{
    // stop secondary event loop started in open()
    vevent::StopProcessing ();
}

