/****
*
*	File name :		audit.c
*	Author :		Steve Johnson
*	Creation date :	January 21, 1999
*
*	Copyright (c) 1999 Sigma Six Corporation
*
*	Provide the following entry points
*
*		audit_logon
*		audit_logoff
*		audit_write
*		audit_fail
*
*	Audit stuff
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssaudit.h>
#include <iocPriv.h>
#include <auditSystem.h>

extern AUDIT_ID	auditLog;

extern char		*last_ssl_name;

void audit_logon (char *message)
{
	TRACE_S ("audit_logon, message = %s", message);
	if (auditLog) {
		if (last_ssl_name) {
			AuditAddEvent (auditLog
						 , AUDIT_LOGON
						 , AUDIT_SUCCESS
						 , last_ssl_name
						 , taskName (taskIdSelf())
						 , message);
		} else {
			AuditAddEvent (auditLog
						 , AUDIT_LOGON
						 , AUDIT_SUCCESS
						 , "No ssl name"
						 , taskName (taskIdSelf())
						 , message);
		}
	}
}

void audit_logoff (char *message)
{
	TRACE_S ("audit_logoff, message = %s", message);
	if (auditLog) {
		if (last_ssl_name) {
			AuditAddEvent (auditLog
						 , AUDIT_LOGOFF
						 , AUDIT_SUCCESS
						 , last_ssl_name
						 , taskName (taskIdSelf())
						 , message);
		} else {
			AuditAddEvent (auditLog
						 , AUDIT_LOGOFF
						 , AUDIT_SUCCESS
						 , "No ssl name"
						 , taskName (taskIdSelf())
						 , message);
		}
	}
}

void audit_write (char *message)
{
	TRACE_S ("audit_write, message = %s", message);
	if (auditLog) {
		if (last_ssl_name) {
			AuditAddEvent (auditLog
						 , AUDIT_WRITE
						 , AUDIT_SUCCESS
						 , last_ssl_name
						 , taskName (taskIdSelf())
						 , message);
		} else {
			AuditAddEvent (auditLog
						 , AUDIT_WRITE
						 , AUDIT_SUCCESS
						 , "No ssl name"
						 , taskName (taskIdSelf())
						 , message);
		}
	}
}

void audit_fail (char *message)
{
	TRACE_S ("audit_fail, message = %s", message);
	if (auditLog) {
		if (last_ssl_name) {
			AuditAddEvent (auditLog
						 , AUDIT_LOGON
						 , AUDIT_FAIL
						 , last_ssl_name
						 , taskName (taskIdSelf())
						 , message);
		} else {
			AuditAddEvent (auditLog
						 , AUDIT_LOGON
						 , AUDIT_FAIL
						 , "No ssl name"
						 , taskName (taskIdSelf())
						 , message);
		}
	}
}

void audit_password (char *message)
{
	TRACE_S ("audit_password, message = %s", message);
	if (auditLog) {
		if (last_ssl_name) {
			AuditAddEvent (auditLog
						 , AUDIT_WRITE
						 , AUDIT_WARNING
						 , last_ssl_name
						 , taskName (taskIdSelf())
						 , message);
		} else {
			AuditAddEvent (auditLog
						 , AUDIT_WRITE
						 , AUDIT_WARNING
						 , "No ssl name"
						 , taskName (taskIdSelf())
						 , message);
		}
	}
}

void audit_bad_password (char *message)
{
	TRACE_S ("audit_bad_password, message = %s", message);
	if (auditLog) {
		if (last_ssl_name) {
			AuditAddEvent (auditLog
						 , AUDIT_WRITE
						 , AUDIT_WARNING
						 , last_ssl_name
						 , taskName (taskIdSelf())
						 , message);
		} else {
			AuditAddEvent (auditLog
						 , AUDIT_WRITE
						 , AUDIT_WARNING
						 , "No ssl name"
						 , taskName (taskIdSelf())
						 , message);
		}
	}
}

