/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vtextitemHEADER
#include vstartupHEADER
#include vtimeHEADER
#include vmemHEADER
#include vcontrolHEADER
#include vlistviewHEADER
#include vprintHEADER
#include vfixedHEADER
#include "tagvunames.h"
#include "tisgalaxylib.hxx"
#define TAGVU_CLIENT 1
extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include <stdio.h>
#ifdef	WIN32
#  include <time.h>
#else
#  include <sys/time.h>
#endif	WIN32
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "cadef.h"
#include "msgMon.h"
#include "tagvudb.h"
#include "OperPriv.h" /* Operator track defs ( Error codes, structure defs, etc. ) */
}
typedef char *caddr_t;
#include "tagvu.h"
extern "C"
{
#ifdef TAGVU_CLIENT
#include "client.h"
#include "tagvuMsgs.h"
#endif TAGVU_CLIENT
}
#include "textFilters.hxx"
#include "tagvuMainDialogClass.h"
#include "tagvuFilterDialogClass.h"
#include "tagvuEntryDialogClass.h"
#include "tagvuRemoveDialogClass.h"

int untagDisabled;	// global flag set by permissions request in loadinit

void RegisterCAProxy(void *arg,int fd,int condition);
void ConnectionHandlerProxy(connection_handler_args args);
void TagValueEventProxy(event_handler_args args);
/*
	This is the source for the main dialog class
*/
vkindTYPED_FULL_DEFINITION(tagvuMainDialogClass, vdialog, "tagvuMainDialogClass");
vkindTYPED_FULL_DEFINITION(tagvuListView, vlistview, "tagvuListView");

void tagvuMainDialogClass::LoadInit(vresource res)
{
    loadiniterror = vFALSE;
    tisAbout = NULL;
    tagList = NULL;
    currentSelection = NULL;

    tagListView = tagvuListView::RequestSubLoad (this, tagvunames_tagvuListView);

    vmenubar *menubar = vmenubar::RequestSubLoad (this,tagvunames_tagvuMenuBar);

    vmenu *fileMenu = vmenu::RequestSubLoad (menubar, tagvunames_tagvuFileMenu);
    vmenuItem *exitMenu = vmenuItem::RequestSubLoad (fileMenu, 
						tagvunames_tagvuExitMenuItem);

    vmenu *filterMenu = vmenu::RequestSubLoad (menubar, tagvunames_tagvuFilterMenu);

    vmenu *tagMenu = vmenu::RequestSubLoad (menubar, tagvunames_tagvuTagMenu);

    vmenu *untagMenu = vmenu::RequestSubLoad (menubar,tagvunames_tagvuUntagMenu);
    removeItem = vmenuItem::RequestSubLoad (untagMenu,tagvunames_tagvuRemoveItem);

    vmenu *printMenu = vmenu::RequestSubLoad (menubar, tagvunames_tagvuPrintMenu);

    helpMenu = vmenu::RequestSubLoad (menubar, tagvunames_tagvuHelpMenu);

    numFound = vdialogLabelItem::RequestSubLoad(this,tagvunames_tagvuCount);

    vdialog::LoadInit(res);		// call base class loadInit to finish the job

    tagListView->setMainDialog(this);

    // check for privileges

    // init privilege checking, see .../share/src/libtis/OperPrvtrk.c
    if ((InitOperPriv (NULL)) != PRIVOK)
    {
	tisSysLog (0,0, "failed to init operator privileges\n");
	errMessage ("Failed to initialize Operator Privileges");
	errCode = TAGVU_ERROR_TAGDESC;
	return;
    }
    if ((OperPriv (NULL, "*MMI_TAGVU_FILTER_FILTER")) == 0)
    {
	filterMenu -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_TAGVU_TAG_TAG")) == 0)
    {
	tagMenu -> Disable ();
    }
    removeItem->Disable();		// until there is a selection
    untagDisabled = vFALSE;
    if ((OperPriv (NULL, "*MMI_TAGVU_UNTAG_UNTAG")) == 0)
    {
	untagMenu -> Disable ();
	untagDisabled = vTRUE;
    }
    if ((OperPriv (NULL, "*MMI_TAGVU_PRINT_PRINT")) == 0)
    {
	printMenu -> Disable ();
    }

    // close privilege checking, see .../share/src/libtis/OperPrvtrk.c
    CloseOperPriv ();

    // store internal names for the file main menu tags
    commandNames[EXIT_MENU] = tagvunames_tagvuExitMenuItem;
    commandNames[FILTER_MENU] = tagvunames_tagvuFilterMenuItem;
    commandNames[TAG_MENU] = tagvunames_tagvuTagMenuItem;
    commandNames[UNTAG_REMOVE_MENUITEM] = tagvunames_tagvuRemoveItem;
    commandNames[PRINT_MENU] = tagvunames_tagvuPrintMenuItem;
    vcommandSpace *space = new vcommandSpace;	// create a command space

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);
    // populate our dictionary with menu commands
    for (int i = EXIT_MENU; i < MENUITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	tagvuCommand *command = new tagvuCommand;
	command->SetSelector(i);	// set the commands selector value
	command->setMainDialog(this);	// set Main dialog so we can use it later
	// store command in the dictionary using the intern'd tag name
	dict->SetElem (commandNames[i], command);
    }
    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);
    SetSpaceOwned (space);
	printer = vprint::CreatePreferred();
	//
	// Init Codebase stuff here
	//
#ifndef TAGVU_CLIENT
	/*
	 * open database files
	 */
	d4init (&codeBaseDesc);
	codeBaseDesc.readLock = 0;
	codeBaseDesc.errCreate = 0;
	codeBaseDesc.errFieldName = 0;
	codeBaseDesc.errGo = 0;
	codeBaseDesc.errOpen = 0;
	codeBaseDesc.errSkip = 0;
	codeBaseDesc.errTagName = 0;
	codeBaseDesc.autoOpen = 1;		// 19-Aug-1998 force MDX file open now
	tagDesc = 0;
	pvmapDesc = 0;
	tagIndexDesc = 0;
	opTrackDesc = 0;
#ifndef WIN32
	if ( !(tagDesc = d4open (&codeBaseDesc, "/tis4000db/tagvu.dbf")))
#else
	if ( !(tagDesc = d4open (&codeBaseDesc, "\\tis4000db\\tagvu.dbf")))
#endif WIN32
	{
	    tisSysLog(0,0, "Can not open \"/tagvu.dbf\"\n");
	    closeFiles();
	    errCode = TAGVU_ERROR_TAGDESC;
	    return;
	}
	d4tag_select (tagDesc, NULL);		// use sequential access
	d4unlock_all (tagDesc);
#ifndef WIN32
	if ( !(pvmapDesc = d4open (&codeBaseDesc, "/tis4000db/pvmap.dbf")))
#else
	if ( !(pvmapDesc = d4open (&codeBaseDesc, "\\tis4000db\\pvmap.dbf")))
#endif WIN32
	{
	    tisSysLog (0,0, "Can not open  \"pvamp.dbf\"\n");
	    closeFiles();
	    errCode = TAGVU_ERROR_PVMAP;
	    return;
	}
#ifndef WIN32
	if ( !(opTrackDesc = d4open (&codeBaseDesc, "/tis4000db/op_trak.dbf")))
#else
	if ( !(opTrackDesc = d4open (&codeBaseDesc, "\\tis4000db\\op_trak.dbf")))
#endif
	{
	    tisSysLog (0,0, "Can not open operator tracking file \"op_trak.dbf\"\n");
	    closeFiles();
	    errCode = TAGVU_ERROR_OPTRAK;
	    return;
	}
	/*
	 * initialize channel access
	 */
	if(ca_task_initialize () != AGNT_NORMAL)
	{
	    //tisSysLog(0,2, "error in channel access init\n");
	    //closeFiles();
	    //errCode = TAGVU_ERROR_CA;
	    //return;
	}
	pendingRequest = NULL;
#else
	if((client = OpenServer(":tagvuServer",1,NULL)) == 0)
	{
	    errCode = TAGVU_ERROR_SVC;
	    return;
	}
#endif TAGVU_CLIENT
	caTimer = new tagvuTimer(this);
	caTimer->SetRecurrent();
	caTimer->SetPeriod(TAGVU_TIMER_INTERVAL,0);
	caTimer->Start();
	filterData = NULL;
	updateList();
	updateListView();
	operatorName[0] = 0;
	GetOperName(operatorName);
	errCode = TAGVU_ERROR_SUCCESS;
}

/*
  GCT 14-May-1998 new function to convert date from dBase format
  "CCYYMMDD" to "MM/DD/CCYY" for Y2K compliance
  inDate is in CCYYMMDD format
*/
char *tagvuMainDialogClass::dateToOutput(char *inDate)
{
  static char dateString[20];
  sprintf(dateString,"%.2s/%.2s/%.4s",&inDate[4],&inDate[6],&inDate[0]);
  return(dateString);
}

void tagvuMainDialogClass::closeFiles()
{
#ifndef TAGVU_CLIENT
    if(pvmapDesc)	d4close(pvmapDesc);
    if(opTrackDesc)	d4close(opTrackDesc);
    if(tagDesc)		d4close(tagDesc);
	tagDesc = 0;
	pvmapDesc = 0;
	tagIndexDesc = 0;
	opTrackDesc = 0;
#else
	CloseSession(client);
#endif TAGVU_CLIENT
}

void tagvuMainDialogClass::Destroy()
{
	if(tagList)
	{
	    tagList->Destroy();
	}
	if (tisAbout != NULL)
	{
	    delete tisAbout;
	    stopTisGalaxy (tisGalaxyvresource);
	}
	delete printer;
	if (IsOpen ())
	{
	    Close ();
	}
	vdialog::Destroy ();
}

/*******************************************************************
*
*    vuMainDialogClass notify function
*
*******************************************************************/

void tagvuMainDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
}

/*******************************************************************
*
*    vuMainDialogClass open function override
*
*******************************************************************/

void tagvuMainDialogClass::Open (void)
{
    BuildAboutMenu ();
    vdialog::Open();
}

/*******************************************************************
*
*    vuMainDialogClass build about menu function
*
*******************************************************************/

void tagvuMainDialogClass::BuildAboutMenu (void)
{
    vstr *verMsg = vstrCloneScribed(vcharScribeLiteral ("Tagvu Version 6.0"));
    vstr *aboutTitle = vstrCloneScribed (vcharScribeLiteral ("About TIS4000 Tagvu"));
    // init tisGalaxy, see tisgalaxylib.<hxx,cxx>
    tisGalaxyvresource = startTisGalaxy ();
    tisAbout = new TisAbout (tisGalaxyvresource, verMsg);
    vstrDestroy (verMsg);
    tisAbout -> SetAboutTitle (aboutTitle);
    vstrDestroy (aboutTitle);
    AboutMenuItem *aboutItem = tisAbout -> GetAboutMenuItem ();
    helpMenu -> AppendItem (aboutItem);
}

void tagvuMainDialogClass::errMessage(char *message)
{
	vstr *errtitle,*msg;
	errtitle = vstrCloneScribed(vcharScribeSystem("tagVu"));
	msg = vstrCloneScribed(vcharScribeSystem(message));
	infoNotice(this,errtitle,msg);
	vstrDestroy(errtitle);
	vstrDestroy(msg);
}

int tagvuMainDialogClass::firstTagRecord(FILTER_DATA_PTR filter,PntTagRecord *tagRecord)
{
#ifdef TAGVU_CLIENT
    TAGVU_MSG_RECORD_REQ *recMsg = (TAGVU_MSG_RECORD_REQ *)&reqMsg;
    RESP_MSG_UNION *answer;
#endif TAGVU_CLIENT
	memset(tagRecord,0,sizeof(PntTagRecord));
#ifndef TAGVU_CLIENT
	if(d4top(tagDesc) != r4success) return(0);
	filterIt(filter,tagRecord);
#else
	if(filter == NULL)
		memset((char *)&reqMsg.recRequest.newTagFilter,0,sizeof(FILTER_DATA));
	else
		memcpy((char *)&reqMsg.recRequest.newTagFilter,(char *)filter,sizeof(FILTER_DATA));
	recMsg->msgType = FIRSTRECORD_MSG_ID;
	recMsg->msgType = htons(recMsg->msgType);
	answer = (RESP_MSG_UNION *)SendReq(client,recMsg,sizeof(TAGVU_MSG_RECORD_REQ));
	if(answer == NULL) return(0);
	memcpy((char *)tagRecord,&(answer->recResponse.newTagRecord),sizeof(PntTagRecord));
	if(tagRecord->recNumber != -1)
		if(tagRecord->pointName[0] == 0) return(0);
#endif TAGVU_CLIENT
	return(1);
}

int tagvuMainDialogClass::nextTagRecord(FILTER_DATA_PTR filter,PntTagRecord *tagRecord)
{
#ifdef TAGVU_CLIENT
    TAGVU_MSG_RECORD_REQ *recMsg = (TAGVU_MSG_RECORD_REQ *)&reqMsg;
    RESP_MSG_UNION *answer;
#endif TAGVU_CLIENT
	memset(tagRecord,0,sizeof(PntTagRecord));
#ifndef TAGVU_CLIENT
	if(d4skip(tagDesc,1) != 0) return(0);
	filterIt(filter,tagRecord);
#else
	if(filter == NULL)
		memset((char *)&reqMsg.recRequest.newTagFilter,0,sizeof(FILTER_DATA));
	else
		memcpy((char *)&reqMsg.recRequest.newTagFilter,(char *)filter,sizeof(FILTER_DATA));
	recMsg->msgType = NEXTRECORD_MSG_ID;
	recMsg->msgType = htons(recMsg->msgType);
	answer = (RESP_MSG_UNION *)SendReq(client,recMsg,sizeof(TAGVU_MSG_RECORD_REQ));
	if(answer == NULL) return(0);
	memcpy((char *)tagRecord,&(answer->recResponse.newTagRecord),sizeof(PntTagRecord));
	if(tagRecord->recNumber != -1)
		if(tagRecord->pointName[0] == 0) return(0);
#endif TAGVU_CLIENT
	return(1);
}

void tagvuMainDialogClass::filterIt(FILTER_DATA_PTR filter,PntTagRecord *tagRecord)
{
#ifndef TAGVU_CLIENT
	int foundit = 5;
	if(filter != NULL)
	{
		foundit = 0;
		if(strlen(filter->pointName) == 0) foundit++;
		else foundit += (!strncmp(filter->pointName,f4str(d4field(tagDesc,"POINTNAME")),strlen(filter->pointName)) ? 1 : 0);
		if(strlen(filter->location) == 0) foundit++;
		else foundit += (!strncmp(filter->location, f4str(d4field(tagDesc,"LOCATION")),strlen(filter->location)) ? 1 : 0);
		if(strlen(filter->operatorApplied) == 0) foundit++;
		else foundit += (!strncmp(filter->operatorApplied,f4str(d4field(tagDesc,"OPER_APL")),strlen(filter->operatorApplied)) ? 1 : 0);
		if(strlen(filter->dateApplied) == 0) foundit++;
		else
		{
			if(filter->dateDirection == 1)
			{
				foundit += (strncmp(filter->dateApplied,f4str(d4field(tagDesc,"DATE_APL")),strlen(filter->dateApplied)) < 0 ? 1 : 0);
			}
			else
			{
				foundit += (strncmp(filter->dateApplied,f4str(d4field(tagDesc,"DATE_APL")),strlen(filter->dateApplied)) > 0 ? 1 : 0);
			}
		}
		if(filter->tagState == FILTER_SHOW_ALL) foundit++;
		else
		{
			foundit += ((d4deleted(tagDesc) && filter->tagState == FILTER_SHOW_UNTAGGED) ? 1 : 0);
			foundit += ((!d4deleted(tagDesc) && filter->tagState == FILTER_SHOW_TAGGED) ? 1 : 0);
		}
	}
	if(foundit == 5)
	{
		strncpy (tagRecord->pointName, f4str (d4field (tagDesc, "POINTNAME")), PVNAME_SZ);
		strncpy (tagRecord->descriptor, f4str (d4field (tagDesc, "DESCRIPTOR")), DESCRIPTOR_SZ);
		strncpy (tagRecord->location, f4str (d4field (tagDesc, "LOCATION")), LOCATION_SZ);
		strncpy (tagRecord->dateApplied, f4str (d4field (tagDesc, "DATE_APL")), DATE_SIZE);
		strncpy (tagRecord->timeApplied, f4str (d4field (tagDesc, "TIME_APL")), TIME_SIZE);
		strncpy (tagRecord->operatorApplied, f4str (d4field (tagDesc, "OPER_APL")), OPERATOR_SZ);
		strncpy (tagRecord->reasonApplied, f4str (d4field (tagDesc, "REASON_APL")), REASON_SIZE);
		strncpy (tagRecord->dateRemoved, f4str (d4field (tagDesc, "DATE_RMV")), DATE_SIZE);
		strncpy (tagRecord->timeRemoved, f4str (d4field (tagDesc, "TIME_RMV")), TIME_SIZE);
		strncpy (tagRecord->operatorRemoved, f4str (d4field (tagDesc, "OPER_RMV")), OPERATOR_SZ);
		strncpy (tagRecord->reasonRemoved, f4str (d4field (tagDesc, "REASON_RMV")), REASON_SIZE);
		tagRecord->tagType = f4char (d4field (tagDesc, "TAGTYPE"));
		tagRecord->isDeleted = d4deleted(tagDesc);
		tagRecord->recNumber = d4recno(tagDesc);
	}
#endif TAGVU_CLIENT
}

/*
	prepare contents of tagvu file for display in list view
*/
void tagvuMainDialogClass::updateList()
{
	char txtBuf[LINE_SIZE];
	int row;
	tagvuEntry *entry;
	PntTagRecord tagrec;
	timeout = 0;
	caTimer->Stop();
	if(firstTagRecord(filterData,&tagrec) != NULL);
	{
	    row = 0;
	    tagList = (vlist *)tagListView->GetList();
	    if(tagList) tagList->Destroy();
	    tagList = new vlist;
	    do
	    {
			if(tagrec.recNumber == -1) continue;
			if(!strlen(tagrec.pointName)) continue;
			entry = new tagvuEntry();
			entry->setRecordNumber(tagrec.recNumber);
			sprintf (txtBuf, "%-*s  %-*s  %-*s",PVNAME_SZ, tagrec.pointName,DESCRIPTOR_SZ, tagrec.descriptor,LOCATION_SZ, tagrec.location);
			entry->putLineIndexed(LINE_0_INDEX,txtBuf);
			entry->setTagName(tagrec.pointName);
			sprintf (txtBuf, "    %s:           %-*s %-*s  %-*s  %-*s",
				(tagrec.tagType == 'O') ? OUT_SERVICE_TAG : INFO_TAG,
				EXT_DATE_SIZE, dateToOutput(tagrec.dateApplied),/* GCT 14-May-1998 output century date */
				TIME_SIZE, tagrec.timeApplied,
				OPERATOR_SZ, tagrec.operatorApplied,
				REASON_SIZE, tagrec.reasonApplied);
			entry->putLineIndexed(LINE_1_INDEX,txtBuf);
			if (tagrec.isDeleted)
			{
			sprintf (txtBuf, "    %s Untagged:  %-*s %-*s  %-*s  %-*s",
				 (tagrec.tagType == 'O') ? OUT_SERVICE_TAG : INFO_TAG,
				 EXT_DATE_SIZE, dateToOutput(tagrec.dateRemoved),/* GCT 14-May-1998 output century date */
				 TIME_SIZE, tagrec.timeRemoved,
				 OPERATOR_SZ, tagrec.operatorRemoved,
				 REASON_SIZE, tagrec.reasonRemoved);
				entry->putLineIndexed(LINE_2_INDEX,txtBuf);
			}
			else
			{
				entry->putLineIndexed(LINE_2_INDEX,"");
			}
			tagList->SetCellValue(row++,0,entry);
	    }
	    while(nextTagRecord(filterData,&tagrec) != 0);		// skip to next record in file
		if(filterData != NULL)vmemFree(filterData);				// all done with the filter, get rid of it.
		filterData = NULL;
	}
}

//
// Set list in listview to display contents
//
void tagvuMainDialogClass::updateListView()
{
	tagListView->SetList(tagList);
	vfont *font = vfont::FindScribed(vcharScribeLiteral(""));
 	tagListView->SetRowHeight(vlistviewALL_ROWS,(font->Height()+4)*LINES_PER_TAG);
	tagListView->SetColumnWidth(vlistviewALL_COLUMNS,2000);
	setCurrentSelection(NULL);
	removeItem->Disable();
	numFound->SetTitleScribed(vcharScribeF("Number of records: %d",tagList->GetRowCount()));
}
void tagvuMainDialogClass::FillFromTagDesc(TagValueData *tagdata)
{
#ifndef TAGVU_CLIENT
	d4go(tagDesc,tagdata->recordNum);
	strncpy(tagdata->pointName, f4str (d4field (tagDesc, "POINTNAME")),PVNAME_SZ);
	tagdata->pointName[PVNAME_SZ] = 0;
	strncpy(tagdata->descriptor, f4str(d4field (tagDesc, "DESCRIPTOR")),DESCRIPTOR_SZ);
	tagdata->descriptor[DESCRIPTOR_SZ] = 0;
	tagdata->tagType = f4char(d4field (tagDesc, "TAGTYPE"));
	d4top(tagDesc);
	tagdata->tagValue = 0;
	if (tagdata->tagType == 'O')
	{
		do
		{
			if((!d4deleted (tagDesc)) &&
				(f4char(d4field (tagDesc, "TAGTYPE")) == 'O') &&
				(strncmp(tagdata -> pointName,f4str(d4field(tagDesc, "POINTNAME")), PVNAME_SZ) == 0))
			{
				tagdata -> tagValue++;
			}
		}
		while(d4skip (tagDesc, 1) == 0);
		d4unlock_all(tagDesc);
		if(tagdata -> tagValue)
		{
			tagdata -> tagValue--;
		}
	}
	// info tag, can not be set to NO until ALL info tags for the
	//   point have been deleted
	else
	{
		do
		{
			if((!d4deleted(tagDesc)) &&
			 (f4char(d4field(tagDesc, "TAGTYPE")) == 'I') &&
			 (strncmp(tagdata -> pointName,f4str(d4field(tagDesc, "POINTNAME")), PVNAME_SZ) == 0))
			{
				tagdata -> tagValue++;
			}
		}
		while (d4skip (tagDesc, 1) == 0);
		d4unlock_all (tagDesc);

		// if any info tags are not deleted
		if(tagdata -> tagValue)
		{
			tagdata -> tagValue--;
			// if we still have some left, set value to 1,since only 0 and 1 are legal values
			if(tagdata -> tagValue) tagdata -> tagValue = 1;
		}
	}
#endif TAGVU_CLIENT
}

//
// Tell CA to modify tag field.
//
int tagvuMainDialogClass::SendTagValue(TagValueData *tagValueDataPtr)
{
#ifndef TAGVU_CLIENT
	char	pointName[PVNAME_SZ + FLDNAME_SZ + 2];
	char	txtBuf [100];

	strcpy(pointName, tagValueDataPtr -> pointName);
	stripWhiteSpace (pointName);
	if(tagValueDataPtr -> tagType == 'O')
	    strcat(pointName, ".TAGD");
	else
	    strcat(pointName, ".ITAG");
	tagValueDataPtr->timeout = TIMEOUT_VALUE;
	tagValueDataPtr->thisPointer = (void *)this;
	pendingRequest = tagValueDataPtr;
	ca_build_and_connect(pointName, TYPENOTCONN, 0, &(tagValueDataPtr->chId), NULL,ConnectionHandlerProxy, (void *)tagValueDataPtr);
	ca_pend_io((float)0.0);
	/* Display and activity window */
	sprintf(txtBuf, "Setting %s value for %s...", 
			  (tagValueDataPtr -> tagType == 'O') ? OUT_SERVICE_TAG : INFO_TAG,
			  stripWhiteSpace(tagValueDataPtr -> pointName));
	tisSysLog(0,3,txtBuf);
#endif TAGVU_CLIENT
	return(1);
}

char *tagvuMainDialogClass::stripWhiteSpace(char *string)
{
	int	i;
	for (i=strlen (string)-1; i >= 0 && isspace(string[i]); string[i--] = 0);
	for (i=0; string[i] && isspace(string[i]); i++);
	if (i) strcpy (string, &string[i]);
	return (string);
}

void tagvuMainDialogClass::printList()
{
	vrect pageRect;
	vfont *font;
	vlistIterator *index = new vlistIterator();
	tagvuEntry *item;
	int numItems,totalLines,dotsPerItem,totalDots,numPages;
	int maxItemsPerPage,fontHeight,stringWidth;
	char printtitle[80],*nowString;
	time_t nowTime = time (NULL);
	nowString = (char *)ctime(&nowTime);
	nowString[strlen(nowString)-1] = '\0';		// drop \n

	vdraw::GSave ();
	font = vfont::FindScribed(vcharScribeLiteral(""));//family:Courier-face:medium-size:6"));
	vdraw::SetFont (font);
	fontHeight = font->GetSize();
	numItems = tagList->GetRowCount();		// number of lines in current list
#ifdef WIN32
       	if(printer->GetDevice() != NULL)
	  {/* GetDevice */
	    printer->OpenDocument();
	    printer->SetDocumentOwner(this);
	    pageRect.CopyFloat(printer->GetPage());
	    vdraw::RectsClip ((const vrect *)&pageRect, 1);
	    index->Start(tagList);
	    totalLines = numItems*(LINES_PER_TAG)+1;	// number of lines in current list including spacer
	    totalDots = totalLines*fontHeight;		// number of "dots" for whole list
	    dotsPerItem = fontHeight*(LINES_PER_TAG);	// number of "dots" for each item
	    maxItemsPerPage = (pageRect.h/dotsPerItem)-1;	// MAX items per page...could be less
	    numPages = (numItems/maxItemsPerPage)+1;	// number of pages for this list
	    printer->SetFirstPageNumber((unsigned int)0);
	    printer->SetLastPageNumber((unsigned int)numPages);
	    nowTime = time(0);
	    for(int i = 0;i < numPages;i++)
	      {
		printer->OpenPage();			// each page must be opened/closed
		printer->SetPageNumber((unsigned int)i);
		vdraw::MoveTo(pageRect.x,pageRect.h-fontHeight);	// start at top of page - height of text
		sprintf(printtitle,"**** Tagged Point List    Page -%d- %s ****",i+1,nowString);
		vdraw::ScribedShow(vcharScribeSystem(printtitle));
		stringWidth = vfixedToInt(font->FixScribedStringWidthX(vcharScribeSystem(printtitle)));
		vdraw::RMoveTo(-stringWidth,-fontHeight);
		if(numItems <= maxItemsPerPage) maxItemsPerPage = numItems;
		for(int k = 0;k < maxItemsPerPage;k++)
		  {
		    if(index->Next() == vTRUE)
		      {
			item = (tagvuEntry *)index->GetValue();
			for(int j=0;j < LINES_PER_TAG;j++)	// loop for all lines in item
			  {
			    vdraw::ScribedShow(vcharScribeSystem(item->getLineIndexed(j)));
			    stringWidth = vfixedToInt(font->FixScribedStringWidthX(vcharScribeSystem(item->getLineIndexed(j))));
			    vdraw::RMoveTo(-stringWidth,-fontHeight);
			  }
		      }
		  }
		numItems -= maxItemsPerPage;
		printer->ClosePage();
	      }
	    printer->CloseDocument();
	    index -> Finish ();
	  }
	else
	  {
	    errMessage("Printer not configured on client");
	  }
#else
	FILE		*fd;
	int		i, j;
	struct timeval	tp;
	char		outputPath[256];
	char		scriptPath[256];
	char		bigBuff[512];
	char		*cPtr;

	gettimeofday (&tp, NULL);
	sprintf(outputPath, "/tis4000db/tagvu_%ld.txt", tp.tv_sec);
	fd = fopen (outputPath, "w+");
	fprintf (fd, "tagvu list printed on %s\n\n", nowString);
	index -> Start (tagList);
	while (index -> Next())
	{
	    item = (tagvuEntry *)(index -> GetValue ());
	    for (i = 0; i < LINES_PER_TAG; i++)
	    {
		strcpy (bigBuff, item -> getLineIndexed (i));
		/* remove trailing spaces */
		for (j = strlen (bigBuff) - 1; j >=0; j--)
		{
		    if (bigBuff[j] != ' ')
		    {
			bigBuff[j+1] = 0;
			break;
		    }
		}
		if (i > 0)
		{
		    if ((strlen (bigBuff)) > 0)
		    {
			/* we are trying to isolate the first part of the msg
			 * from the reason part which can be 100 char long
			 */
			if ((cPtr = strchr(bigBuff, ':')) != NULL)
			{
			    cPtr++;
			    /* there is space between the : and the rest 
			     * of the line, get past it
			     */
			    while ((*cPtr) == ' ')
			    {
				cPtr++;
			    }
			    cPtr = cPtr + EXT_DATE_SIZE + TIME_SIZE + OPERATOR_SZ;
			    *(cPtr -1) = 0;
			    fprintf (fd, "%s\n", bigBuff);
			    fprintf (fd, "\t%s\n", cPtr);
			}
			else
			{
			    fprintf (fd, "%s\n", bigBuff);
			}
		    }
		    else
		    {
			fprintf (fd, "%s\n", bigBuff);
		    }
		}
		else
		{
		    fprintf (fd, "%s\n", bigBuff);
		}
	    }
	    fprintf (fd, "\n");
	}
	index -> Finish ();
	fclose (fd);
	sprintf (scriptPath, "csh /tis4000db/tagvu_script %s &", outputPath);
	system (scriptPath);

	
#endif WIN32
	vdraw::GRestore();
}

//
// Handle menu commands here
//
vkindTYPED_FULL_DEFINITION(tagvuCommand, vcommandSelector, NULL);
int tagvuCommand::Issue (vdict *dict)
{
    int			command;
	vapplication *app = vapplication::GetCurrent();
	vresource appres = app->GetResources();
	vresource res;
	tagvuFilterDialogClass *filter;
	tagvuEntryDialogClass *entry;
	tagvuRemoveDialogClass *remove;
	TagValueData *tvd;
	char msg[80];
#ifdef TAGVU_CLIENT
    TAGVU_MSG_ADDTAG *addMsg = (TAGVU_MSG_ADDTAG *)getMainDialog()->getReqMsg();
    TAGVU_MSG_REMOVETAG *remMsg = (TAGVU_MSG_REMOVETAG *)getMainDialog()->getReqMsg();
    RESP_MSG_UNION *answer;
#endif TAGVU_CLIENT


    command = GetSelector ();
    switch (command)
    {
    case EXIT_MENU:
		getMainDialog()->closeFiles();
		vevent::StopProcessing();
		break;
    case FILTER_MENU:
		res = vresourceGet(appres,tagvunames_tagvuFilter);	// load the filter dialog
		filter = new tagvuFilterDialogClass(res);
		filter->Open();
		getMainDialog()->setFilterData(filter->getFilterData());	// save in mainDialog too
		getMainDialog()->updateList();
		getMainDialog()->updateListView();
		delete filter;
		break;
    case TAG_MENU:
		res = vresourceGet(appres,tagvunames_tagvuEntry);	// load the filter dialog
		entry = new tagvuEntryDialogClass(res);
		entry->Open();
		tvd = entry->getTagData();
		if(tvd)
		{
#ifndef TAGVU_CLIENT
		    getMainDialog()->SendTagValue(tvd);
#else
			strcpy(addMsg->pointName,tvd->pointName);
			strcpy(addMsg->reason,tvd->reason);
			strcpy(addMsg->operatorApplied,getMainDialog()->operatorName);
			addMsg->tagType = tvd->tagType;
			addMsg->msgType = ADDTAG_MSG_ID;
			addMsg->msgType = htons(addMsg->msgType);
			getMainDialog()->timeout = 5;
			getMainDialog()->caTimer->Start();
			answer = (RESP_MSG_UNION *)SendReq(getMainDialog()->getClient(),addMsg,sizeof(TAGVU_MSG_ADDTAG));
			if(answer != NULL)
			{
				if(ntohs(answer->msgHeader.msgType) == ADDNACK_MSG_ID)
				{
				  sprintf(msg,"Point %s not defined on server",addMsg->pointName);
					getMainDialog()->errMessage(msg);
				}
			}
			vmemFree(tvd);
#endif TAGVU_CLIENT
		}
		delete entry;
		break;
    case UNTAG_REMOVE_MENUITEM:
		res = vresourceGet(appres,tagvunames_tagvuRemove);	// load the filter dialog
		remove = new tagvuRemoveDialogClass(res);
		remove->removeOpen(getMainDialog()->getCurrentSelection()->getTagName());
		tvd = remove->getTagData();
		if(tvd)
		{
		    tvd->recordNum = getMainDialog()->getCurrentSelection()->getRecordNumber();
#ifndef TAGVU_CLIENT
		    getMainDialog()->FillFromTagDesc(tvd);
		    getMainDialog()->SendTagValue(tvd);
#else
			strcpy(remMsg->reason,tvd->reason);
			remMsg->msgType = REMOVETAG_MSG_ID;
			remMsg->msgType = htons(remMsg->msgType);
			remMsg->recordNum = tvd->recordNum;
			getMainDialog()->timeout = 5;
			getMainDialog()->caTimer->Start();
			strcpy(remMsg->operatorApplied,getMainDialog()->operatorName);
			answer = (RESP_MSG_UNION *)SendReq(getMainDialog()->getClient(),remMsg,sizeof(TAGVU_MSG_REMOVETAG));
			if(answer != NULL)
			{
				if(ntohs(answer->msgHeader.msgType) != REMACK_MSG_ID)
				{
					getMainDialog()->errMessage("Invalid ACK message from Server");
				}
			}
			vmemFree(tvd);
#endif TAGVU_CLIENT
		}
		delete remove;
		break;
    case PRINT_MENU:
		//
		// print is always local even in a C/S setup
		getMainDialog()->getPrinter()->ConfirmStyle();
		getMainDialog()->printList();
		break;
	default:
	    return (vFALSE);
    }
    return (vTRUE);
}

char *tagvuEntry::getLineIndexed(int index)
{
	return(&tagLines[index][0]);
}

void tagvuEntry::putLineIndexed(int index, char *newline)
{
	if(index < 0 || index > LINES_PER_TAG-1)
		return;
	strcpy(&tagLines[index][0],newline);
}

//
// ListView overides
//
void tagvuListView::DrawCellContents(int row,int col,int sel,vrect *crect)
{
    tagvuEntry *entry;
    int textHeight;
    int rectHeight;
    int lineHeight;

    vdraw::GSave();
    vdraw::RectsClip(crect, 1);
    vlist *thisList = (vlist *)GetList();
    entry = (tagvuEntry *)thisList->GetCellValue(row,col);
    vfont *font = vfont::FindScribed(vcharScribeLiteral(""));
    textHeight = font->Height();
    rectHeight = crect->GetHeight();
    lineHeight = rectHeight/LINES_PER_TAG;
    for(int i = 0;i < LINES_PER_TAG;i++)
    {
	vdraw::MoveTo(crect->GetX(),(crect->GetY()+(rectHeight - textHeight * (i + 1))));
	vdraw::ScribedShow(vcharScribeSystem (entry->getLineIndexed(i)));
    }
    vdraw::GRestore();
}

void tagvuListView::ObserveSelect(int state)
{
	tagvuEntry *entry;
	int row,col;
	vlist *thisList = GetList();
	if(state == vlistviewSELECTION_HAS_CHANGED)
	{
		vlistSelection *mySelection = GetSelection();
		thisList->GetSelectedCell(mySelection,&row,&col);
		entry = (tagvuEntry *)thisList->GetCellValue(row,col);
		getMainDialog()->setCurrentSelection(entry);
		// only if user has untag privilege
		if (untagDisabled == vFALSE)
		{
		    getMainDialog()->removeItem->Enable();
		}
	}
}

//
// FD registration and its proxy function. The proxy functions are needed
// to resolve the "this" pointer.
//
void ConnectionHandlerProxy(connection_handler_args args)
{
	void *classPointer;
	classPointer = ((TagValueData *)ca_puser(args.chanid))->thisPointer;
	((tagvuMainDialogClass *)classPointer)->ConnectionHandler(args);
}

void TagValueEventProxy(event_handler_args args)
{
	void *classPointer;
	classPointer = ((TagValueData *)ca_puser(args.chanid))->thisPointer;
	((tagvuMainDialogClass *)classPointer)->TagValueEvent(args);
}

//
// Actual callback functions
//
void tagvuMainDialogClass::ConnectionHandler(connection_handler_args args)
{
	TagValueData	*tagValueDataPtr;	/* Get the pointer to the tag information */
	tagValueDataPtr = (TagValueData *)ca_puser(args.chanid);	/* Update the point's tag value */
	ca_put (OLD_DBR_SHORT, args.chanid, &(tagValueDataPtr -> tagValue));	/* Get the value of the tag field using a callback */
	ca_array_get_callback(OLD_DBR_STRING, 1, args.chanid, TagValueEventProxy, (void *)tagValueDataPtr);
	ca_flush_io();
}

void tagvuMainDialogClass::TagValueEvent(event_handler_args args)
{
#ifndef TAGVU_CLIENT
    char	dateStr[EXT_DATE_SIZE+1];
    char	timeStr[TIME_SIZE+1];
    char	tagValue[BUFSIZ];
    time_t	curTime;
    struct tm	*curTimePtr;
    TagValueData	*tagValueDataPtr;
    char txtBuf[32];

    tagValueDataPtr = (TagValueData *)args.usr;
    strcpy(tagValue,(char *)args.dbr);
    stripWhiteSpace(tagValue);
    if(((tagValueDataPtr -> tagType == 'O') &&
	   (atoi(tagValue) == tagValueDataPtr -> tagValue)) ||
	 ((tagValueDataPtr -> tagType == 'I') &&
	   (((tagValueDataPtr -> tagValue == 0) &&
	       (strcmp(tagValue, "NO") == 0) ) ||
	     ((tagValueDataPtr -> tagValue == 1) &&
	       (strcmp(tagValue, "YES") == 0)))))
    {
    curTime = time(0L);
    curTimePtr = localtime(&curTime);
    pendingRequest = NULL;
    /*
      GCT 14-May-1998 Add in century for Y2K compliance
    */
    sprintf (dateStr, "%04d%02d%02d", curTimePtr -> tm_year+1900,curTimePtr -> tm_mon+1,curTimePtr -> tm_mday);
    sprintf (timeStr, "%02d:%02d:%02d", curTimePtr -> tm_hour,curTimePtr -> tm_min, curTimePtr -> tm_sec);
    ca_clear_channel(tagValueDataPtr -> chId);
    ca_flush_io();
    if(tagValueDataPtr->type == ADD_TAG)
    {
	// add new record to tag file
	d4append_blank (tagDesc);
	f4assign(d4field(tagDesc, "POINTNAME"), tagValueDataPtr->pointName);
	f4assign(d4field(tagDesc, "LOCATION"), tagValueDataPtr->location);
	f4assign(d4field(tagDesc, "DESCRIPTOR"), tagValueDataPtr->descriptor);
	f4assign(d4field(tagDesc, "OPER_APL"), operatorName);
	f4assign(d4field(tagDesc, "REASON_APL"), tagValueDataPtr->reason);
	f4assign(d4field(tagDesc, "DATE_APL"), dateStr);
	f4assign(d4field(tagDesc, "TIME_APL"), timeStr);
	f4assign_char(d4field(tagDesc, "TAGTYPE"), tagValueDataPtr->tagType);
	d4append_blank (opTrackDesc);
	f4assign(d4field(opTrackDesc, "OPERATOR"), operatorName);
	f4assign(d4field(opTrackDesc, "TRK_DATE"), dateStr);
	f4assign(d4field(opTrackDesc, "TRK_TIME"), timeStr);
	f4assign(d4field(opTrackDesc, "POINTNAME"), tagValueDataPtr -> pointName);
	f4assign(d4field(opTrackDesc, "DESCRIPTOR"), tagValueDataPtr -> descriptor);
	f4assign(d4field(opTrackDesc, "OLD_VALUE"),(tagValueDataPtr -> tagType == 'O') ? OUT_SERVICE_TAG : INFO_TAG);
	}
	else if(tagValueDataPtr->type == REMOVE_TAG)
	{
	    d4go(tagDesc, tagValueDataPtr -> recordNum);
	    f4assign(d4field(tagDesc, "DATE_RMV"), dateStr);
	    f4assign(d4field(tagDesc, "TIME_RMV"), timeStr);
	    f4assign(d4field(tagDesc, "OPER_RMV"), operatorName);
	    f4assign(d4field(tagDesc, "REASON_RMV"), tagValueDataPtr -> reason);
	    d4delete(tagDesc);
	    d4append_blank(opTrackDesc);
	    f4assign(d4field(opTrackDesc, "OPERATOR"), operatorName);
	    f4assign(d4field(opTrackDesc, "TRK_DATE"), dateStr);
	    f4assign(d4field(opTrackDesc, "TRK_TIME"), timeStr);
	    f4assign(d4field(opTrackDesc, "POINTNAME"), tagValueDataPtr -> pointName);
	    f4assign(d4field(opTrackDesc, "DESCRIPTOR"), tagValueDataPtr -> descriptor);
	    sprintf(txtBuf, "untag %s",(tagValueDataPtr -> tagType == 'O') ? OUT_SERVICE_TAG : INFO_TAG);
	    f4assign(d4field(opTrackDesc, "OLD_VALUE"), txtBuf);
	}
	d4flush_all(tagDesc);
	d4unlock_all(tagDesc);
	d4flush_all(opTrackDesc);
	d4unlock_all(opTrackDesc);
	updateList();
	updateListView();
	vmemFree(tagValueDataPtr);
	}
#endif TAGVU_CLIENT
}

//
// Here is the timer ObserverTimer function. It just calls
// ca_pend_io every 5 Seconds
//
void tagvuTimer::ObserveTimer()
{
#ifndef TAGVU_CLIENT
    char msg[80];
    ca_pend_io((float)0.0);
    tagvuMainDialogClass *main = getMainDialog();
    if(main->pendingRequest)			// if there is a pending request
    {
	if(--(main->pendingRequest->timeout) < 0)	// and no more time to wait
	{
	    sprintf(msg,"CA Timeout for %s",main->pendingRequest->pointName);
	    main->errMessage(msg);
	    vmemFree(main->pendingRequest);
	    main->pendingRequest = NULL;
	}
    }
#else
/*
	As a client this function calls GetMsg to see if any update requests
	have been made.
*/
    RESP_MSG_UNION *answer;
	answer = (RESP_MSG_UNION *)GetMsg(getMainDialog()->getClient());
	if(answer != NULL)
	{
		if(ntohs(answer->msgHeader.msgType) == UPDATELIST_MSG_ID)
		{
			getMainDialog()->updateList();
			getMainDialog()->updateListView();
		}
	}
	if(getMainDialog()->timeout)
	{
		if(--(getMainDialog()->timeout) <= 0)	// and no more time to wait
		{
		    getMainDialog()->errMessage("Tag/Untag timeout");
			getMainDialog()->caTimer->Stop();
		}
	}
#endif TAGVU_CLIENT
}
