/*CSLMacros.c,v 1.1.1.1 1999/08/03 14:44:52 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/* Purpose: Recursively performs macro substitutions on strings.
 *
 * Example: The definition dog=cat causes chan_$(dog)_001 to become chan_cat_001
 */
#include	<stdio.h>
#include	<ctype.h>
#include	<string.h>

void		*calloc();

void		get_macro_pairs();
void		gsdl_main_macro();
static void	gsdl_process_macro();
static char	*gsdl_lookup_sub();
static void	gsdl_concat();




/*
 *	get_macro_pairs sets up the array of string pairs(0-1, 2-3, etc.) as specified
 *	by a list represented by a string, i.e. the string is parsed to produce a set of
 *	ptrs to pairs of strings.  This set (array) is used for specifying the substitution
 *	of the first string in the pair by the second string.
 *
 *	Note :
 *	The macro substitution string will look something like:
 *		rf = RF1, cav = CAV1
 *
 *	and will produce the array:
 *		pMacro_pairs[0] = "rf"		 	
 *		pMacro_pairs[1] = "RF1"		
 *		pMacro_pairs[2] = "cav"		
 *		pMacro_pairs[3] = "CAV1"		
 *		pMacro_pairs[4] = NULL
 *
 *	pStr		ptr to the string containing the substitution specification(s)
 *	pMac_pairs 	ptr to the array of substitution pairs
 *	max_pairs	maximum number to do (pMac_pairs needs to have (max_pairs*2)+1 elements)
 */

void remove_spaces(p)
char  *p;
	{
	for (; *p && isspace(*p); p++)
		{
		strcpy(p, p+1);
		p--;
		}
	}

void get_macro_pairs(pMacstr, pMac_pairs, max_pairs)
char	*pMacstr;
char	*pMac_pairs[];
int		max_pairs;
	{
	char			*p;
	unsigned int	index		= 0;
	unsigned int	max_index	= (2*max_pairs);

	for (p = pMacstr; *p && index < max_index;)
		{
		char	delimiter	= (index & 1) ? ',' : '=';

		pMac_pairs[index] = p;
		if (delimiter == '=')
			remove_spaces(p);
		while (*p && (*p != delimiter))
			{
			if (isspace(*p) && delimiter == '=')
				{
				remove_spaces(p);
				p--;
				}
			p++;
			}
		if (!(*p) && (delimiter == '='))
			break;
		if ((*p == 0) || (*p == delimiter) )
			{
			if (*p)
				*p++ = 0;
			index++;
			}
		}
	if (index & 1)
		index--;
	pMac_pairs[index] = NULL;
	}


/*
 *	gsdl_main_macro(pTo, pFrom, pMacro_pairs, maxlen)
 *
 *		pTo:		ptr to the character array which is to receive
 *				the channel name after all substitutions.
 *		pFrom:		ptr to the input character array which possibly
 *				has substitutions indicated by 
 *				$(stringtobesubstituted).
 *		pMacro_pairs:	an array of character ptr pairs terminated
 *				with a NULL ptr.  The first element points to
 *				the first string which may be substituted by
 *				the string pointed to by the second element.
 *				The third element points to the second string 
 *				which may be substituted by the string pointed to
 *				by the forth element.  And so on, until terminated
 *				by a NULL.
 *
 */
void gsdl_main_macro(pTo, pFrom, pMacro_pairs, maxlen)
char	*pTo;
char	*pFrom;
char	*pMacro_pairs[];
int		maxlen;
	{
	char	*pTo_max	= pTo + (maxlen - 1);

	while (*pFrom && (pTo < pTo_max))
		{
		if (*pFrom == '$' && pMacro_pairs)
			gsdl_process_macro(&pTo, &pFrom, pTo_max, pMacro_pairs, pTo_max - pTo);
		else
			*pTo++ = *pFrom++;
		}
	*pTo = '\0';
	}


static void gsdl_process_macro(pPto, pPfrom, pTo_max, pMacro_pairs, maxlen)
char	**pPto;
char	**pPfrom;
char	*pTo_max;
char	*pMacro_pairs[];
int		maxlen;
	{
	char	*macro_name		= (char *)calloc(maxlen, sizeof(char));
	char	*pMacro_name		= &macro_name[0];
	char	*pMacro_name_max	= (maxlen - 1) + &macro_name[0];
	char	*pSub_name;
	int		i;

	if (macro_name == NULL)
		return;	/* error: calloc failure */

	for (i = 0; i < maxlen; i++)
		macro_name[i] = '\0';

	(*pPfrom)++;			/* inc past $ */
	if (**pPfrom == '\0')
		{
		gsdl_concat(pPto, "$", pTo_max);
		return; /* error: uncompleted macro name */
		}
	if (**pPfrom == '(')
		{
		(*pPfrom)++;		/* inc past ( */
		while (1)
			{
			if (**pPfrom == '$')
				gsdl_process_macro(&pMacro_name, pPfrom, pMacro_name_max, pMacro_pairs, maxlen);

			if (**pPfrom == ')')
				{
				/* macro complete, so look-up and concat */
				(*pPfrom)++;		/* inc past ) */
				pSub_name = gsdl_lookup_sub(&macro_name[0], pMacro_pairs);
				if (pSub_name == NULL)
					{
					gsdl_concat(pPto, "$(", pTo_max);
					gsdl_concat(pPto, &macro_name[0], pTo_max);
					gsdl_concat(pPto, ")", pTo_max);
					}
				else
					{
					gsdl_concat(pPto, pSub_name, pTo_max);
					}
				return;
				}

			if (**pPfrom == '\0')
				{
				gsdl_concat(pPto, "$(", pTo_max);
				gsdl_concat(pPto, &macro_name[0], pTo_max);
				return; /* error: uncompleted */
				}
			if (pMacro_name >= pMacro_name_max)
				{
				*pMacro_name = '\0';
				gsdl_concat(pPto, "$(", pTo_max);
				gsdl_concat(pPto, &macro_name[0], pTo_max);
				return; /* error: macro expand exceeds buffer */
				}
			*pMacro_name = **pPfrom;
			pMacro_name++;
			(*pPfrom)++;
			}
		} /* error: uncompleted macro name */
	gsdl_concat(pPto, "$", pTo_max);
	free(macro_name);
	}


static char *gsdl_lookup_sub(pMacro_name, pMacro_sub_pairs)
char	*pMacro_name;
char	*pMacro_sub_pairs[];
	{
	while (*pMacro_sub_pairs != NULL)
		{
		if (!strcmp(pMacro_name, *pMacro_sub_pairs))
			return (char *) *(++pMacro_sub_pairs);
		pMacro_sub_pairs++;
		pMacro_sub_pairs++;
		}
	return (char *) NULL;
	}


static void gsdl_concat(pPto, pSub_name, pPto_max)
char	**pPto;
char	*pSub_name;
char 	*pPto_max;
	{
	if (pSub_name == NULL)
		return;

	while (*pSub_name && (*pPto < pPto_max))
		{
		**pPto = *pSub_name;
		pSub_name++;
		(*pPto)++;
		}
	}

#ifdef MAIN
main(argc, argv)	/* apply replacements given on command line "x=cat,y=dog..." to stdin producing stdout */
int		argc;
char	*argv[];
	{
	int		numReplacements;		/* number of pairs */
	char	*p;

	/*
	 * get the macro definitions
	 */
	for (p = (argc > 1) ? argv[1] : "", numReplacements = 0; *p; p++)		/* count the number of replacements */
		if (*p == '=')
				numReplacements++;

	if (numReplacements > 0)
		{
		char	**ppMacroPairs;			/* macro pair list */
		char	inBuf[BUFSIZ];
		char	outBuf[BUFSIZ];

		ppMacroPairs = (char **)calloc((numReplacements*2)+1, sizeof(char *));
		get_macro_pairs(argv[1], ppMacroPairs, numReplacements);
		while (gets(inBuf))
			{
			gsdl_main_macro(outBuf, inBuf, ppMacroPairs, sizeof outBuf);
			puts(outBuf);
			}
		}
	else
		{
		char	inBuf[BUFSIZ];
		
		while (gets(inBuf))
			puts(inBuf);
		}
	}
#endif
