/****************************************************************/
/*								*/
/*	This software was developed for DLA by Tate Integrated  */
/*	Systems, L.P. via a contract with Vitro Corporation.	*/
/*	It is designed to work with TIS4000 Release 4.2.	*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	evol_man.c
 *
 *	This program provides a command line interface to the Evolution Manager requests.
 *
 *	usage:  evol_man estop
 *	        evol_man start|stop|pause|resume <evolution #>
 *	        evol_man start|stop|pause|resume <evolution #> <segment #>
 */
#include <string.h>
#include <ctype.h>
#include <EVOLManager.h>

#define	ARG_ERROR	(LAST_EVOLMANAGER_ERROR-1)	/* an error to give useage message */
#define	GET_OK		(LAST_EVOLMANAGER_ERROR-2)	/* an 'error' that's ok */

char			*hostName = "localhost";	/* currently uses only the local host */
MakeRequestResult	*requestResult;			/* a place to store the request result */
GetNextRequestResult	*getResult;


char *lower(s)						/* make the given string lowercase */
char	*s;
	{
	char	*os;
	
	for (os = s; *os = isupper(*os) ? tolower(*os) : *os; os++)
		;
	return s;		/* return the given 'original' pointer */
	}
	

get_all_reg()			/* get and print all of the next requests */
	{
	while ((getResult = EVOLManagerGetNextRequest())->status != NO_REQUEST_AVAILABLE)
		{
		printf("type:%s evolution:%d segment:%d environment:%s\n",
		    EvolutionProcessorRequestText[getResult->type], getResult->evolution,
				getResult->segment, getResult->environmentString);
		}
	return GET_OK;
	}


req_estop(pCommand)					/* make an ESTOP request */
char	*pCommand;
	{
	if (strcmp(pCommand, "estop") == 0)
		{
		requestResult = EVOLManagerPerformESTOP(hostName);
		return requestResult->status;
		}
	return ARG_ERROR;
	}


EvolutionProcessorRequestType	getType(pCommand)	/* parse the command for a matching reqest type */
char	*pCommand;
	{
	if (strcmp(pCommand, "start")  == 0)		return START;
	else if (strcmp(pCommand, "stop")   == 0)	return STOP;
	else if (strcmp(pCommand, "pause")  == 0)	return PAUSE;
	else if (strcmp(pCommand, "resume") == 0)	return RESUME;
	else return SWINGTO;	/* can't swing from here, use this as an error indication */
	}


req_no_segment(pCommand, pEvolution)	/* make a request without specifing the segment number */
char	*pCommand;
char	*pEvolution;
	{
	EvolutionProcessorRequestType	type = getType(pCommand);

	if (type != SWINGTO)
		{
		requestResult = EVOLManagerMakeRequest(hostName, type,
					atoi(pEvolution), NO_SEGMENT);
		return requestResult->status;
		}
	return ARG_ERROR;
	}


req_all(pCommand, pEvolution, pSegment)			/* make a fully-formed request */
char	*pCommand;
char	*pEvolution;
char	*pSegment;
	{
	EvolutionProcessorRequestType	type = getType(pCommand);
	
	if (type != SWINGTO)
		{
		requestResult = EVOLManagerMakeRequest(hostName, type,
					atoi(pEvolution), atoi(pSegment));
		return requestResult->status;
		}
	return ARG_ERROR;
	}


main(argc, argv)
int	argc;
char	*argv[];
	{
	char	*programName	= ((programName = strrchr(argv[0], '/')) == NULL) ? argv[0]: programName + 1;
	int	status;

	EVOLManagerAttach(hostName);	/* allow communications */		
	switch (argc)			/* do the request */
	    {
	    case 1:	status = get_all_reg();					break;
	    case 2:	status = req_estop(lower(argv[1]));			break;
	    case 3:	status = req_no_segment(lower(argv[1]), argv[2]);	break;
	    case 4:	status = req_all(lower(argv[1]), argv[2], argv[3]);	break;
	    default:	status = ARG_ERROR;					break;
	    }
	switch (status)		/* check the result */
	    {
	    default:
	    case TF_ERROR:		printf("%s\n", requestResult->statusText);
	    case GET_OK:	break;
	    case ARG_ERROR:	printf("usage:\t%s estop\n", programName);
				printf("\t%s start|stop|pause|resume <evolution #>\n",
					programName);
				printf("\t%s start|stop|pause|resume <evolution #> <segment #>\n",
					programName);
				break;
	    }
	EVOLManagerDetach();			/* clean up communications */

	return (status == TF_OK) ? 0 : 1;		/* return a useful status */
	}

