%{

/*lex.l,v 1.1.1.1 1999/08/03 14:44:54 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
	csl.l - CSL lexical analyzer
*/

static	int	FindKeyword ();

#include <stdlib.h>
#include <string.h>
#include "y.tab.h"
#include "symtab.h"
#include "csldefs.h"
#include "tree.h"

int	yylexval;
int	TokenType;
Symbol	*yylexptr;
extern	TreeNode	*yylval;
int	MakeTermNode ();

%}

letter			[a-zA-Z_]
digit			[0-9]
letter_or_digit		[a-zA-Z_0-9]+
dlr			[$]
#csequence		{letter_or_digit}*
csequence		{letter_or_digit}+
label			{csequence}":"
macro			({dlr}"("{csequence}")")
#identifier		({csequence}{macro})*
identifier		({csequence}*{macro}*)+
quote			["]
notquote		[^"]
nl			[\n]
notnl			[^\n]
white_space		[ \t\n]
blank			[ \t]
other			[^a-zA-Z0-9\n\t ]

%%

"^"			return EXPONENT;
"*"			return MULTIPLY;
"/"			return DIVIDE;
"+"			return ADD;
"-"			return SUBTRACT;
"<"			return LT;
"<="			return LE;
">"			return GT;
">="			return GE;
"="			return EQ;
"<>"			return NE;
"&"			return AND;
"|"			return OR;

"/""/"{notnl}*{nl}	;

{digit}+:{digit}+:{digit}+"a""m"	{ return MakeTermNode (Time, TIME_NODE, TIMECONST); }
{digit}+:{digit}+:{digit}+"p""m"	{ return MakeTermNode (Time, TIME_NODE, TIMECONST); }
{digit}+:{digit}+:{digit}+		{ return MakeTermNode (Time, TIME_NODE, TIMECONST); } 
{digit}+:{digit}+"a""m"			{ return MakeTermNode (Time, TIME_NODE, TIMECONST); }
{digit}+:{digit}+"p""m"			{ return MakeTermNode (Time, TIME_NODE, TIMECONST); }
{digit}+:{digit}+			{ return MakeTermNode (Time, TIME_NODE, TIMECONST); }
{digit}+				{ return MakeTermNode (Constant, CONSTANT_NODE, INTCONST); }
{digit}+"."{digit}*			{ return MakeTermNode (Constant, CONSTANT_NODE, REALCONST); }
{label}					{ yytext [yyleng-1] = '\0';
					  return MakeTermNode (Label, LABEL_STATEMENT, LABEL); }
{identifier}+"."{identifier}+		{ return MakeTermNode (Channel, CHANNEL_NODE, CHAN); }
@{identifier}+				{ return MakeTermNode (Timer, TIMER_NODE, TIMER); }
{identifier}				{ return FindKeyword(yytext); }
{quote}{notquote}*{quote}		{ yytext [yyleng-1] = '\0';
	  				  strcpy (yytext, &yytext[1]);
					  return MakeTermNode (Literal, LITERAL_NODE, LITERAL); }

{white_space}+				;

{other}					return yytext [0];

%%

/*
	Keyword lookup
*/

static struct kwords {		/* Keyword table */
	char	*Keyword;	/* Pointer to keyword string */
	int	Class;		/* Token class returned by yylex */
	int	Value;		/* Token value returned in yylexval */
} Keywords [] = {	/* Keywords must be alphabetical for binary search */
	"ABS",		Function,	FUNC_ABS,
	"ACOS",		Function,	FUNC_ACOS,
	"ACTIVITY",	ACTIVITY,	0,
	"AFTER",	AFTER,		0,
	"ALARM",	ALARM,		0,
	"AND",		AND,		0,
	"AS",		AS,		0,
	"ASIN",		Function,	FUNC_ASIN,
	"AT",		AT,		0,
	"ATAN",		Function,	FUNC_ATAN,
	"BE",		BE,		0,
	"CAN",		CAN,		0,
	"CANNOT",	CANNOT,		0,
	"CEILING",	Function,	FUNC_CEILING,
	"COMMAND",	COMMAND,	0,
	"COMPLETE",	COMPLETE,	0,
	"COMPLETES",	COMPLETES,	0,
	"CONTINUE",	CONTINUE,	0,
	"COS",		Function,	FUNC_COS,
	"CSL",		CSL,		0,
	"DO",		DO,		0,
	"ELSE",		ELSE,		0,
	"END",		END,		0,
	"EQUAL",	EQUAL,		0,
	"EVENT",	EVENT,		0,
	"EXECUTE",	EXECUTE,	0,
	"EXP",		Function,	FUNC_EXP,
	"EXPIRED",	EXPIRED,	0,
	"EXPIRES",	EXPIRES,	0,
	"FLOOR",	Function,	FUNC_FLOOR,
	"FOR",		FOR,		0,
	"GO",		GO,		0,
	"GOTO",		GOTO,		0,
	"GREATER",	GREATER,	0,
	"HAS",		HAS,		0,
	"HOURS",	HOURS,		0,
	"IF",		IF,		0,
	"IMMEDIATELY",	IMMEDIATELY,	0,
	"IN",		IN,		0,
	"INFORMATIONAL",INFORMATIONAL,	0,
	"IS",		IS,		0,
	"LESS",		LESS,		0,
	"LN",		Function,	FUNC_LN,
	"LOG",		Function,	FUNC_LOG,
	"MAJOR",	MAJOR,		0,
	"MAX",		Function,	FUNC_MAX,
	"MIN",		Function,	FUNC_MIN,
	"MINOR",	MINOR,		0,
	"MINUTES",	MINUTES,	0,
	"MOD",		Function,	FUNC_MOD,
	"NOT",		NOT,		0,
	"OF",		OF,		0,
	"ON",		ON,		0,
	"OR",		OR,		0,
	"REACHED",	REACHED,	0,
	"SCRIPTCOMPLETIONSTATUS", Function, FUNC_COMPSTATUS,
	"SCRIPTRUNSTATUS", Function,	FUNC_RUNSTATUS,
	"SECONDS",	SECONDS,	0,
	"SET",		SET,		0,
	"SIN",		Function,	FUNC_SIN,
	"SQRT",		Function,	FUNC_SQRT,
	"START",	START,		0,
	"STOP",		STOP,		0,
	"TAN",		Function,	FUNC_TAN,
	"THAN",		THAN,		0,
	"THEN",		THEN,		0,
	"TIMEOUT",	TIMEOUT,	0,
	"TO",		TO,		0,
	"UNTIL",	UNTIL,		0,
	"VERIFY",	VERIFY,		0,
	"WAIT",		WAIT,		0,
	"WHILE",	WHILE,		0,
	"WITH",		WITH,		0,
	"WITHIN",	WITHIN,		0,
};

/*----------*/
static int FindKeyword (String)		/* Find keyword using binary search */

    char	*String;

    {

	struct kwords *Low = Keywords,
		*High = Keywords - 1 + sizeof (Keywords) / sizeof (Keywords [0]),
		*Middle;
	int	Diff;

	
	while (Low <= High) {
	    Middle = Low + (High - Low) / 2;
	    if ((Diff = strcasecmp (Middle->Keyword, String)) == 0) {
		TokenType = Middle->Value;
		return Middle->Class;
	    } else if (Diff < 0)
		Low = Middle + 1;
	    else
		High = Middle - 1;
	}
	return MakeTermNode (Identifier, IDENTIFIER_NODE, VARIABLE);

}

/*----------*/
int MakeTermNode (TokType, NodeType, SymType)	/* Make a terminal node */

    int		TokType, SymType, NodeType;

    {

	yylval = NewTermNode (NodeType, LookUp (yytext, SymType));
	return TokType;

    }

/* end MakeTermNode */

