/****************************************************************/
/*								*/
/*	COMMON INTIALIZATION FOR CHANNEL ACCESS SERVERS		*/
/*								*/
/*	Many of the things that we do for the three servers	*/
/*	that make up the agent architecture for channel access	*/
/*	are common. This routine avoids code duplication.	*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <fcntl.h>
#include <unistd.h>
#include <sys/file.h>
#include <termios.h>
#include <signal.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include <string.h>

#include <msmain.h>

void serverInit(appName, dirPath, lockFile)
char	*appName;
char	*dirPath;
char	*lockFile;
{
    int		fd;
    char	buff[10];

/******
	Change our controlling TTY so that we don't pick up any
	stray signals.
******/

#if !(defined(AXP) || defined(SOLARIS))
    fd = open("/dev/tty", O_RDWR);
    ioctl(fd, TIOCNOTTY, 0);
    close(fd);
#endif

/******
	Move to the defined directory so that the core file
	can be found in case we die an unnatural death.
******/

    if (chdir(dirPath) != 0) {
	syslog (LOG_OPTIONS, "%s(%d): Unable to find %s lock directory (%s)",
			appName, getpid (), appName, dirPath);
	if (mkdir (dirPath, 0775) == 0) {
	    syslog (LOG_OPTIONS, "%s(%d): Lock directory %s created.",
			appName, getpid (), dirPath);
	    chdir (dirPath);
	} else {
	    syslog (LOG_OPTIONS, "%s(%d): Lock directory %s could not be created.",
			appName, getpid (), dirPath);
	    exit (-1);
	}
    }

/******
	Make a process group just for ourselves
******/

#ifdef AXP
    tcsetpgrp (0, getpid ());
#else
#ifdef SOLARIS
    setpgrp ();
#else
    setpgrp(0, getpid());
#endif
#endif

/******
	Record our PID in the lock file to make it easy to find
******/

    sprintf(buff, "%d\n", getpid());
    write(fd, buff, strlen(buff));
    close(fd);		/* I think that we keep the lock even if
			   the file is closed. */

}

