/****************************************************************/
/*								*/
/*	DEFINITIONS FOR THE SGS-THOMSON M48T08 TOD CLOCK	*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

/*
Be careful: Some of these data bytes have control bits buried
in them! See the definitions later in this file!
*/

typedef struct
    {
    unsigned char	control;	/* Control register */
    unsigned char	seconds;	/* 00-59, BCD! */
    unsigned char	minutes;	/* 00-59, BCD! */
    unsigned char	hour;		/* 00-23, BCD! */
    unsigned char	weekday;	/* 01-07, BCD! */
    unsigned char	date;		/* 01-31, BCD! */
    unsigned char 	month;		/* 01-12, BCD! */
    unsigned char	year;		/* 00-99, BCD! */
    } M48T08;

/*
Definitions for the control register.

	NOTE: There are other functions which we will
	not be using. For now, we'll ignore calibration.
	The defaults should be accurate within a minute
	per month and that's good enough for our purposes
	at this time.
*/

#define M48T08_CTL_WRITE	0x80	/* Write bit */
#define M48T08_CTL_READ		0x40	/* Read bit */

/*
Definitions for the seconds register.
*/

#define M48T08_SEC_STOP		0x80	/* Stop oscillator bit */
#define M48T08_SEC_10SEC_MASK	0x70	/* Isolate tens of seconds */
#define M48T08_SEC_10SEC_SHFT	4	/* Shift into/out of LSB */
#define M48T08_SEC_01SEC_MASK	0xf	/* Isolate unit seconds */
#define M48T08_SEC_01SEC_SHFT	0	/* Shift into/out of LSB */

/*
Definitions for the minutes register.
*/

#define M48T08_MIN_10MIN_MASK	0x70	/* Isolate tens of minutes */
#define M48T08_MIN_10MIN_SHFT	4	/* Shift into/out of LSB */
#define M48T08_MIN_01MIN_MASK	0xf	/* Isolate unit minutes */
#define M48T08_MIN_01MIN_SHFT	0	/* Shift into/out of LSB */

/*
Definitions for the hour register.
*/

#define M48T08_HOUR_10HR_MASK	0x30	/* Isolate tens of hours */
#define M48T08_HOUR_10HR_SHFT	4	/* Shift into/out of LSB */
#define M48T08_HOUR_01HR_MASK	0xf	/* Isolate unit hours */
#define M48T08_HOUR_01HR_SHFT	0	/* Shift into/out of LSB */

/*
Definitions for the weekday register

	Note: There are other functions which we will
	not be using.
*/

#define M48T08_DAY_MASK		7	/* Isolate day of week */
#define M48T08_DAY_SHFT		0	/* Shift into/out of LSB */

/*
Definitions for the date register.
*/

#define M48T08_DATE_10DAY_MASK	0x30	/* Isolate tens of days of month */
#define M48T08_DATE_10DAY_SHFT	4	/* Shift into/out of LSB */
#define M48T08_DATE_01DAY_MASK	0xf	/* Isolate unit days of month */
#define M48T08_DATE_01DAY_SHFT	0	/* Shift into/out of LSB */

/*
Definitions for the month register.
*/

#define M48T08_MNTH_10MON_MASK	0x10	/* Isolate tens of months */
#define M48T08_MNTH_10MON_SHFT	4	/* Shift into/out of LSB */
#define M48T08_MNTH_01MON_MASK	0xf	/* Isolate unit months */
#define M48T08_MNTH_01MON_SHFT	0	/* Shift into/out of LSB */

/*
Definitions for the year register.
*/

#define M48T08_YEAR_10YR_MASK	0xf0	/* Isolate tens of years */
#define M48T08_YEAR_10YR_SHFT	4	/* Shift into/out of LSB */
#define M48T08_YEAR_01YR_MASK	0xf	/* Isolate unit years */
#define M48T08_YEAR_01YR_SHFT	0	/* Shift into/out of LSB */
