/*nserver.c,v 1.1.1.1 1999/08/03 14:44:59 tis4000 Exp */

/*	@(#)nserver.c	1.2 7/20/91
 *	Author:	Gale Slentz
 *	Date:	08-24-90
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .00	08-24-90	ges	initial version
 * .01	01-09-91	ges	clean up documentation
 * .02	07-20-91	rac	use EPICS_SYSCLK_PORT to get SYSCLK_PORT
 *
 */
/*
***************************************************
* Server Process:
*	This process is designed to run forever on the UNIX
*	file server.  It provides the system time to the client
*	(Master Time IOC) whenever requested.
*
***************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <envDefs.h>
#include <time.h>

#ifdef SOLARIS
#define bzero(b,n)	memset(b, 0, n)
#endif

#define	ERROR	(-1)
#define	SOCK_STREAM	1	/* stream socket */
#define	AF_INET		2	/* internetwork: UDP, TCP, etc. */

struct sockaddr {
	unsigned short sa_family;	/* address family */
	char sa_data[14];		/* up to 14 bytes of direct addr */
};

typedef struct sockaddr SOCKADDR;

struct in_addr {
	unsigned long s_addr;
};

struct sockaddr_in {
	short sin_family;
	unsigned short sin_port;
	struct in_addr sin_addr;
	char sin_zero[8];
};

/*
***************************************************
* This is a server program which communicates with a client
* through a socket. 
* The server runs on UNIX while the clients are IOCs on VxWorks.
***************************************************
*/

main ()
    {
    unsigned short SYSCLK_PORT;
    char		sysclk_port[80];
    unsigned long	sec_past_unix_epoch; /* UNIX time() storage */
    int			sock, snew;	/* socket fd's */
    struct sockaddr_in	serverAddr;	/* server's address */
    struct sockaddr_in 	clientAddr;	/* client's address */
    int			client_len;	/* length of clientAddr */
    char		c;
    extern int		errno;		/* for UNIX error referencing */
    int bindval;
    int first_loop = 1;

    /* Get the port number from the environment variable
     */
    if (envGetConfigParam(&EPICS_SYSCLK_PORT, 80, sysclk_port) == NULL) {
	printf("nserver: %s isn't set\n", EPICS_SYSCLK_PORT.name);
	exit (1);
    }
    if (sscanf(sysclk_port, "%hu", &SYSCLK_PORT) != 1) {
	printf("nserver: error reading %s: %s\n", EPICS_SYSCLK_PORT.name,
					sysclk_port);
	exit (1);
    }

    /* Zero out the sock_addr structures,
     * (must be done before the socket calls)
     */
    memset (&serverAddr, 0,  sizeof (serverAddr));
    memset (&clientAddr, 0, sizeof (clientAddr));

    /* Opening the socket
     * Using ARPA Internet address format and stream sockets.
     */
    do {
	/* can only fail from no memory */
    	sock = socket (AF_INET, SOCK_STREAM, 0);
	if (sock == ERROR) {
		perror("nserver: opening stream socket failed");
		exit (1);
	}

       /* Set up our internet address, and bind it so the client can connect. */
	serverAddr.sin_family = AF_INET;
  	serverAddr.sin_port   = SYSCLK_PORT;

	bindval = bind (sock, (SOCKADDR *)&serverAddr, sizeof (serverAddr));
	if (bindval == ERROR) {
		if (first_loop == 1) {
			first_loop = 0;
			perror("nserver: binding stream socket failed");
		}
		close (sock);
	}
    } while (bindval == ERROR);

    /* Listen for client connection */

    printf ("nserver: Listening for client\n");
    listen(sock, 2);

    /* The client is connected */

    printf ("nserver: ACCEPTING CLIENT\n");

    client_len = sizeof (clientAddr);

    for (;;) {
	    snew = accept (sock, (SOCKADDR *)&clientAddr, &client_len);

	    if (snew == ERROR) { 
		perror ("nserver: accept failed\n");
    	    }
	    else {
	    	printf ("nserver: CLIENT: port = %d: family = %d: addr = %lx:\n", 
			clientAddr.sin_port, clientAddr.sin_family, 
			clientAddr.sin_addr.s_addr);

	        sec_past_unix_epoch = time((time_t *) 0);
	        printf("nserver: sys time seconds = %li\n", sec_past_unix_epoch);
	        send(snew,&sec_past_unix_epoch, sizeof(sec_past_unix_epoch), 0);
	    }
	    close(snew);
    }

    }
