/*timeServer.c,v 1.1.1.1 1999/08/03 14:44:59 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/****************************************************************/
/*								*/
/*		WORKSTATION TIME SERVER PROGRAM			*/
/*								*/
/*	This program runs on the Unix workstations. If it is	*/
/*	in master mode, it will broadcast the time every so	*/
/*	often. These broadcasts are heard by the other work	*/
/*	stations on the network and also by the IOCS. There	*/
/*	is also a set of rules for promotion from slave to	*/
/*	master.							*/
/*								*/
/*	WARNING! This program must be run with all of the	*/
/*	privileges of the superuser because it tries to		*/
/*	change the time of day on the slave computers.		*/
/*								*/
/****************************************************************/

/****************************************************************/
/*								*/
/*		UNIX PORTABILITY ALERT!				*/
/*								*/
/*	The definition of gettimeofday() changes depending on	*/
/*	your operating system. Sun OS needs two pointers, one	*/
/*	to a structure for the time, the other for a structure	*/
/*	for the time zone. Sun Solaris does not use this 2nd	*/
/*	argument. DEC OSF/1 does use the second argument but	*/
/*	only in a mode for supporting obsolete code. We will	*/
/*	always provide the second argument and use other	*/
/*	methods to determine our time zone (i.e., minutes	*/
/*	west of UCT. That is done just prior to shipping the	*/
/*	time synch message out the door. Therefore, the		*/
/*	timezone structure is no longer used.			*/
/*								*/
/*	GMU, 3.17.96						*/
/*								*/
/****************************************************************/

/****************************************************************/
/*								*/
/*			PROGRAM ENHANCEMENT			*/
/*								*/
/*	This program has been enhanced to support the KSI	*/
/*	(a division of Odetics) TPRO-S board. This board	*/
/*	generates a very accurate time (Good!) but that		*/
/*	time does not include the year (Bad!). Therefore,	*/
/*	to make a time stamp that can be used in the Unix	*/
/*	environment, we have to add the number of seconds	*/
/*	since 00:00:00 Jan 1, 1970 to 00:00:00 Jan 1 of the	*/
/*	current year. This is not too bad, but it requires	*/
/*	that we develop a rule about detecting when the		*/
/*	system is running on New Year's Eve. After much		*/
/*	consideration, here's the rule:  When the KSI		*/
/*	board moves backwards in time by more than 300 days,	*/
/*	we assume that we have just crossed New Year's Eve.	*/
/*	At that point, we snatch the current time out of	*/
/*	the workstation. If our local time is 23:59:00 or	*/
/*	later on 31 December, we will add 1 to the current	*/
/*	year number (i.e., we were probably just a little	*/
/*	bit behind the master time source). Otherwise,		*/
/*	we will just use the local system's current		*/
/*	year. This will work well as long as the workstation	*/
/*	and the master time source are within a minute of	*/
/*	each other when we cross New Year's Eve. Another	*/
/*	restriction is that we shouldn't be starting the	*/
/*	timeServer program in the few minutes before a new	*/
/*	year starts. You should be partying then, anyway!	*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#ifdef WIN32
#include <winsock.h>
#include <io.h>
#include "NTUnixSubs.h"
#else
#include <sys/time.h>
#include <sys/socket.h>
#include <in.h>
#include <if.h>
#include <tpro.h>
#include <netdb.h>
#if !(defined(ultrix) || defined(AXP))
#include <sys/sockio.h>
#else
#include <sys/ioctl.h>
#endif
#endif WIN32
#include <tsDefs.h>
#include <fcntl.h>
#include <timeSync.h>

#if defined(SOLARIS) || defined(WIN32)
#define bcopy(s,d,n)	memcpy (d, s, n)
#endif

#ifdef WIN32
static int
local_addr (int fd, struct sockaddr_in *pAddr);
#endif WIN32
/****************************************************************/
/*								*/
/*		DEFAULT TUNING PARAMETERS			*/
/*								*/
/****************************************************************/

int	assume =  180;	/* Take over after 3 minutes of silence */
int	syncrate =   60;	/* Broadcast once a minute */
short	port   = 8586;	/* UDP port number for our socket */

/****************************************************************/
/*								*/
/*		GLOBAL DATA FOR THIS PROCESS			*/
/*								*/
/****************************************************************/

int	bcastSocket;	/* Our UDP socket FD */
int	fileExists;	/* Does /tis4000db/timeMaster exist? */
int	ksiTpro = -1;	/* Assume KSI board not installed */
long	ksiYear;	/* Secs from 00:00:00 Jan 1, 1970 to
			   00:00:00 Jan 1 of the current year */
short	ksiLastDay = 1000;	/* Last day we read. First time,
				   get the year! */

int	mode;		/* Current mode */

#define	MASTER_MODE	0
#define SLAVE_MODE	1
#define BACKING_OFF	2

struct timeval		masterDelay;
struct timeval		slaveDelay;
struct timeval		backoffDelay;

int	verbose = 0;
int	testing = 0;

/****************************************************************/
/*								*/
/*			THE MAIN PROGRAM			*/
/*								*/
/****************************************************************/

main(argc, argv)
int	argc;
char	**argv;
{
	struct sockaddr_in	setPort;
	struct sockaddr_in	bcastAddr;
	struct timeval		*currDelay;
	struct timeval		currTime;
	struct timezone		tz;
#ifdef WIN32
	WSADATA			wsaData; 
	WORD			wVersionRequested = MAKEWORD(1, 1);
	int			err;
	struct timeval		processTime;
	SYSTEMTIME		winTime;
#else
	int			n;
	struct ifreq		*ifr;
	char			buf[1024];
	struct ifconf		ifc;
#endif WIN32
	struct timeval		wakeupTime;
	struct tm		*tm_time;
	struct timeval		newTime;
	struct tisTimeSync	timeMsg;
	fd_set			bcastFdSet;
	int			optionValue = 1;
	int			setPortSize;
	long			holdUsec;
	unsigned long		myAddr;
	int			status;
	extern	char		*optarg;
	int			c, temp;
	static	int		selectstat;
	struct tm		*pTime;
	time_t			clock = 0;
	u_char			hops = 5;

#ifdef WIN32
    err = WSAStartup(wVersionRequested, &wsaData); 
 
    if (err != 0) 
	/* Tell the user that we couldn't find a useable */ 
	/* winsock.dll.     */ 
	return 0; 
#endif WIN32

#ifdef AXP
	/* Detach from the console if running on AXP */

	setpgrp ();
	setpgid (0,0);
#endif

	/******
	See if the user needs help
	******/

	if (argc > 1)
		{
		if (strcmp(argv[1], "help") == 0)
			{
			printf("\n");
			printf("Usage: timeServer -a<ttt> -b<ttt> -p<nnnn>\n");
			printf("\n");
			printf("\t-a = timeout to assume master role. Default is 180 seconds.\n");
			printf("\t-b = broadcast rate. Default is 60 seconds.\n");
			printf("\t-p = UDP port number to use. Default is 8586.\n");
			printf("\t-v = Enable verbose mode.\n");
			printf("\t-v = Enable testing (doesn't send or process received) mode.\n");
			printf("\n");
			printf("\tIn all cases, <ttt> is a time expressed in seconds\n");
			printf("\tand <nnnn> is a positive integer. Do NOT code '<' or '>'.\n");
			printf("\n");
			exit(0);
			}
		}
	
	/******
	Pick up any arguments
	******/

	while ((c = getopt (argc, argv, "a:b:p:vt")) != EOF) {
	    switch (c) {
		case 'a':
		    sscanf (optarg, "%d", &assume);
		    break;
		case 'b':
		    sscanf (optarg, "%d", &syncrate);
		    break;
		case 'p':
		    scanf (optarg, "%d", &temp);
		    port = temp;
		    break;
		case 'v':
		    verbose = 1;
		    break;
		case 't':
		    testing = 1;
		    break;
		case '?':
		    break;
	    }
	}

	/******
	Initialize
	******/

	mode = SLAVE_MODE;
#ifdef WIN32
	_unlink ("\\tis4000db\\timeMaster");
#else
	unlink("/tis4000db/timeMaster");
#endif WIN32
	fileExists = 0;

	if ((bcastSocket = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		{
		perror("Can't create socket");
		exit (-1);
		}

	if (setsockopt(bcastSocket, SOL_SOCKET, SO_BROADCAST,
			(char *) &optionValue, sizeof(optionValue)) == -1)
		{
		perror("Can't set broadcast socket option");
		exit (-1);
		}
	setPort.sin_family = AF_INET;
	setPort.sin_addr.s_addr = htonl(INADDR_ANY);
	setPort.sin_port = htons(port);
	if (bind(bcastSocket, (struct sockaddr *) &setPort, sizeof(setPort)) == -1)
		{
		perror("Can't bind to timeServer UDP port");
		exit (-1);
		}


	masterDelay.tv_sec = syncrate;
	masterDelay.tv_usec = 0;
	slaveDelay.tv_sec = assume + 10;
	slaveDelay.tv_usec = 0;

	/*
	The backoff delay used to avoid colisions on failover is the low order
	byte of our Inet address treated as tenths of seconds.
	*/

	local_addr(bcastSocket, &setPort);
	myAddr = ntohl(setPort.sin_addr.s_addr);
	backoffDelay.tv_sec = (myAddr & 0xff) / 10;
	backoffDelay.tv_usec = ((myAddr & 0xff) % 10) * 100000;

	/*
	Get the broadcast address in the standard manner.
	*/

#ifndef WIN32
	ifc.ifc_len = sizeof(buf);
	ifc.ifc_buf = buf;

	if (ioctl(bcastSocket, SIOCGIFCONF, (char *) &ifc) < 0)
		{
		perror("Can't SIOCGIFCONF");
		exit (-1);
		}

	if (setsockopt (bcastSocket, IPPROTO_IP, IP_MULTICAST_TTL, &hops, sizeof(hops)))
	    perror ("Unable to set hop count for broadcast , will use default.\n");

	ifr = ifc.ifc_req;

	/*for (n = ifc.ifc_len / sizeof(struct ifreq); --n > 0; ifr++)*/
	for (n = ifc.ifc_len / sizeof(struct ifreq); n--; ifr++)
		{
		if (ifr -> ifr_addr.sa_family != AF_INET)
			continue;
		if (ioctl(bcastSocket, SIOCGIFFLAGS, (char *) ifr) < 0)
			{
			perror("Can't SIOCGIFFLAGS");
			exit (-1);
			}
		if ( (ifr -> ifr_flags & IFF_UP) == 0 ||
				(ifr -> ifr_flags & IFF_LOOPBACK) ||
				(ifr -> ifr_flags & (IFF_BROADCAST | IFF_POINTOPOINT)) == 0)
			continue;
		if (ifr -> ifr_flags & IFF_POINTOPOINT)
			{
			if (ioctl(bcastSocket, SIOCGIFDSTADDR, (char *) ifr) < 0)
				{
				perror("Can't SIOCGIFDSTADDR");
				exit (-1);
				}
			bcopy((char *) &(ifr -> ifr_dstaddr), (char *) &bcastAddr,
					sizeof(bcastAddr));
			break;
			}
		else if (ifr -> ifr_flags & IFF_BROADCAST)
			{
			if (ioctl(bcastSocket, SIOCGIFBRDADDR, (char *) ifr) < 0)
				{
				perror("Can't SIOCGIFBRDADDR");
				exit (-1);
				}
			bcopy((char *) &(ifr -> ifr_broadaddr), (char *) &bcastAddr,
					sizeof(bcastAddr));
			break;
			}
		}
#else
	bcastAddr.sin_addr.s_addr = htonl(INADDR_BROADCAST);
#endif WIN32
	bcastAddr.sin_family = AF_INET;
	bcastAddr.sin_port = htons(port);

#ifndef WIN32
	/*
	Try to access the KSI TPRO-S time generator board. If it is there we will
	use it. If not, then we will use the regular system time clock from the
	workstation. Remember that open returns -1 on error.
	*/

	ksiTpro = open("/dev/tpr0", O_RDWR);
#endif WIN32

	/******
	Main processing loop
	******/

	for (;;)
		{

		if (mode == MASTER_MODE) {
			currDelay = &masterDelay;
			if (verbose)
				printf ("timeServer: Master mode.\n");
		} else if (mode == SLAVE_MODE) {
			currDelay = &slaveDelay;
			if (verbose)
				printf ("timeServer: Slave mode.\n");
		} else {
			currDelay = &backoffDelay;
			if (verbose)
				printf ("timeServer: Backoff mode.\n");
		}

		FD_ZERO(&bcastFdSet);
		FD_SET(bcastSocket, &bcastFdSet);
		selectstat = select(bcastSocket + 1, &bcastFdSet, NULL, NULL, currDelay);
		gettimeofday (&wakeupTime, &tz);
		if (selectstat == 0)
			{

			/*
			Timeout. Handle according to mode.
			*/
			if (mode == SLAVE_MODE)
				{

				/*
				In slave mode, back off to avoid collision
				*/

				mode = BACKING_OFF;
				if (verbose)
					printf ("timeServer: Mode changed to Backoff\n");
				continue;
				}
			else if (mode == BACKING_OFF)
				{
				int	fd;

				/*
				If we've already backed off, we're now the boss.
				*/

				mode = MASTER_MODE;
#ifdef WIN32
				fd = _open("\\tis4000db\\timeMaster", O_WRONLY | O_CREAT | O_TRUNC, 0666);
				_close (fd);
#else
				fd = open("/tis4000db/timeMaster", O_WRONLY | O_CREAT | O_TRUNC, 0666);
				close(fd);
#endif WIN32
				fileExists = 1;
				if (verbose)
					printf ("timeServer: Mode changed to Master.\n");
				}
			}
		else
			{

			/*
			A message arrived. Read it to empty the port.
			*/

			if (verbose)
				printf ("timeServer: Message arrived.\n");

			setPortSize = sizeof(setPort);
			status = recvfrom(bcastSocket, (char *)&timeMsg, sizeof(timeMsg), 0,
					(struct sockaddr *)&setPort, &setPortSize);
			if (verbose) {
				printf ("timeServer: Received message, status = %d\n", status);
				printf ("timeServer: From %d.%d.%d.%d:%d\n",
						ntohl(setPort.sin_addr.s_addr) >> 24 & 0xFF,
						ntohl(setPort.sin_addr.s_addr) >> 16 & 0xFF,
						ntohl(setPort.sin_addr.s_addr) >> 8 & 0xFF,
						ntohl(setPort.sin_addr.s_addr) & 0xFF,
						(int) ntohs (setPort.sin_port));
			}

			if (ntohl(setPort.sin_addr.s_addr) == myAddr)
				{
				continue;
				}
			if (mode == MASTER_MODE)
				{

				/*
				MULTIPLE MASTERS! Set our mode to BACKING_OFF
				and cause a time message to be sent immediately.
				This will also make the other master back off.
				Because we have different timers, only one of
				us will eventually revert back to master.
				*/

				mode = BACKING_OFF;
				if (verbose)
					printf ("timeServer: Mode changed to Backoff.\n");

				}
			else
				{

				/*
				A master sent us the time. Use it, and then
				make sure that we stay in slave mode, using
				the standard timer.

				However, if the KSI TPRO-S board is installed,
				use that time instead of the one that we got
				from the time master!
				*/

				if (verbose)
					printf ("timeServer: Received time.\n");

#ifndef WIN32
				if (~ksiTpro)
					readKsi(&newTime);
				else
#endif WIN32
					{
					newTime.tv_sec = ntohl(timeMsg.time.secPastEpoch) +
						TS_EPOCH_SEC_PAST_1970;
					newTime.tv_usec = ntohl(timeMsg.time.nsec) / 1000;
					}
#ifdef WIN32
				gettimeofday (&processTime, &tz);
				/* don't update to newtime if slightly behind local time */
				if (((newTime.tv_sec == processTime.tv_sec)
				  && (newTime.tv_usec < processTime.tv_usec))
				  || ((newTime.tv_sec > processTime.tv_sec - 2)
				  && (newTime.tv_sec < processTime.tv_sec)))
				{
				    if (verbose)
					    printf ("timeServer: Rejecting deadband time.\n");
				    continue;
				}
				if (processTime.tv_usec < wakeupTime.tv_usec)
				{
				    processTime.tv_sec --;
				    processTime.tv_usec += 1000000L;
				}
				processTime.tv_usec -= wakeupTime.tv_usec;
				processTime.tv_sec -= wakeupTime.tv_sec;
				newTime.tv_sec += processTime.tv_sec;
				newTime.tv_usec += processTime.tv_usec;;
				tm_time = gmtime (&newTime.tv_sec);
				winTime.wYear = tm_time->tm_year + 1900;
				winTime.wMonth = tm_time->tm_mon + 1;
				winTime.wDay = tm_time->tm_mday;
				winTime.wHour = tm_time->tm_hour;
				winTime.wMinute = tm_time->tm_min;
				winTime.wSecond = tm_time->tm_sec;
				winTime.wMilliseconds = (newTime.tv_usec)/1000;
				if (testing)
				{
				    printf ( "Setting time to %d/%d/%d %d:%2.2d:%2.2d.%4.4d\n",
					    winTime.wMonth, winTime.wDay, winTime.wYear,
					    winTime.wHour, winTime.wMinute, winTime.wSecond, winTime.wMilliseconds);
				}
				else
				{
				    SetSystemTime (&winTime);
				}
#else
				gettimeofday(&currTime, &tz);
				currTime.tv_sec = newTime.tv_sec - currTime.tv_sec;
				currTime.tv_usec = newTime.tv_usec - currTime.tv_usec;
				adjtime(&currTime, &newTime);
#endif WIN32
				slaveDelay.tv_sec = assume;
				mode = SLAVE_MODE;
				if (fileExists)
					{
#ifdef WIN32
					_unlink("\\tis4000db\\timeMaster");
#else
					unlink("/tis4000db/timeMaster");
#endif WIN32
					fileExists = 0;
					}
				continue;
				}
			}
		/*
		We are either master or backing off. Send out our idea of
		the current time.

		Again, use the KSI TPRO-S board if we have one. If not, then use the
		workstations's system clock.
		*/

		gettimeofday(&currTime, &tz);
#ifndef WIN32
		if (~ksiTpro)
			readKsi(&currTime);
		else
#endif WIN32
			{
			holdUsec = currTime.tv_usec;

			/*
			Loop until we get a time of day with different useconds.
			This will make sure that we get right on the edge of the
			clock resolution. This call will happen once on Sun CPUs,
			and will last about .004 seconds on DEC/MIPS CPUs.
			*/

			do
				{
				gettimeofday(&currTime, &tz);
				} while (currTime.tv_usec == holdUsec);

			}
		/*
		Adjust for difference in epoch. Epics is 1/1/90, Unix is 1/1/70.
		*/

		timeMsg.time.secPastEpoch = htonl(currTime.tv_sec - TS_EPOCH_SEC_PAST_1970);

		/*
		Adjust for different partial seconds. Epics uses nanoseconds.
		*/

		timeMsg.time.nsec = htonl(currTime.tv_usec * 1000);

		/*
		Move in the current number of minutes west of UCT
		*/

		pTime = localtime(&clock);
		timeMsg.minWest = (24 - ((pTime->tm_mday==1 ? 24 : 0) + pTime->tm_hour)) * 60;
		timeMsg.minWest = htonl(timeMsg.minWest);

		if (verbose)
			printf ("timeServer: Broadcasting time.\n");

		if (testing)
		{
		    tm_time = gmtime (&currTime.tv_sec);
		    printf ( "Sending time %d/%d/%d %d:%2.2d:%2.2d.%4.4d\n",
			    tm_time->tm_mon + 1, tm_time->tm_mday, tm_time->tm_year + 1900,
			    tm_time->tm_hour, tm_time->tm_min, tm_time->tm_sec, (currTime.tv_usec)/1000);
		}
		else
		{
		    status = sendto(bcastSocket, (char *)&timeMsg, sizeof(timeMsg), 0,
				    (struct sockaddr *)&bcastAddr, sizeof(bcastAddr));
		    if (status < 0)
		    {
#ifdef WIN32
			printf ("Can't send time: %d\n", WSAGetLastError());
#else
			    perror("Can't send time");
#endif WIN32
		    }
		}

		/*
		If the KSI TPRO board is installed, read the time again and use 
		it to adjust our own workstation clock.
		*/

#ifndef WIN32
		if (~ksiTpro)
			{
			readKsi(&newTime);
			gettimeofday(&currTime, &tz);
			currTime.tv_sec = newTime.tv_sec - currTime.tv_sec;
			currTime.tv_usec = newTime.tv_usec - currTime.tv_usec;
			adjtime(&currTime, &newTime);
			}
#endif WIN32
		}
	return 0;
}


#ifndef WIN32
/****************************************************************/
/*								*/
/*		READ THE KSI/ODETICS TPRO-S BOARD		*/
/*								*/
/*	Also, perform the gyrations that are necessary to 	*/
/*	return a struct timeval that can be used by Unix.	*/
/*	The Unix epoch is 00:00:00 Jan 1, 1970.			*/
/*								*/
/****************************************************************/

readKsi(timeNow)
struct timeval	*timeNow;
{
	struct currtime		ksiTime;
	short			ksiDay;
	struct timeval		tv;
	struct timezone		tz;
	struct tm		tm;

	read(ksiTpro, &ksiTime, sizeof(ksiTime));

	ksiDay = ksiTime.hundays * 100 + ksiTime.tendays * 10 + ksiTime.days;
	if ((ksiLastDay - ksiDay) > 300)
		{

		long	ksiAdjustment;

		/*
		The KSI board just crossed midnite on New Year's Eve.
		Get the current year, bumping by one if we are in the
		last minute of the year. Remember, the month number
		in a struct tm is zero based!
		*/

		gettimeofday(&tv, &tz);	/* Get the local time */
		tm = *gmtime((time_t *)(&tv.tv_sec));	/* Convert to something useable */
		if ((tm.tm_mon == 11)  && (tm.tm_mday == 31) &&
		    (tm.tm_hour == 23) && (tm.tm_min == 59))
			{
			tv.tv_sec += 60-tm.tm_sec;
			}
                else {
		     tv.tv_sec -= (tm.tm_yday*86400 + tm.tm_hour*3600 
				   + tm.tm_min*60 + tm.tm_sec);
                     }

		/********************************************************/
		/*							*/
		/*		BRAIN DAMAGE ALERT!			*/
		/*							*/
		/*	The KSI software computes ksiTime.tv_sec as	*/
		/*	day * 86400 + hour * 3600 + min * 60 + sec.	*/
		/*	This is wrong because the days in the time	*/
		/*	sources are relative to 1! This means that	*/
		/*	we have to subtract one day's worth of seconds	*/
		/*	from the tv_sec that KSI calculated for us.	*/
		/*	Since you can't trust anyone to leave a bug	*/
		/*	unfixed, what we will do is see if the current	*/
		/*	day number * 86400 is greater than tv_sec.	*/
		/*	If that is true, then the driver from KSI was	*/
		/*	fixed. If it is false, then the driver wasn't	*/
		/*	fixed and we have to subtract 86400 from	*/
		/*	tv_sec reported by KSI before using it.		*/
		/*	Rather than doing that every time we read the	*/
		/*	board, we'll make the adjustment once when	*/
		/*	we determine what year it is.			*/
		/*							*/
		/********************************************************/

		if ((((ksiTime.hundays * 100) + (ksiTime.tendays * 10) + ksiTime.days) * 86400)
				> ksiTime.tv_sec)
			ksiAdjustment = 0;
		else
			ksiAdjustment = 86400;

		ksiYear = tv.tv_sec - ksiAdjustment;

		/*
		All of the above code may have taken a non-trivial amount
		of time. Just in case, read the time again to be sure.
		*/

		read(ksiTpro, &ksiTime, sizeof(ksiTime));
		}

	/********************************************************/
	/*							*/
	/*		BRAIN DAMAGE ALERT!			*/
	/*							*/
	/*	The KSI software computes ksiTime.tv_sec as	*/
	/*	day * 86400 + hour * 3600 + min * 60 + sec.	*/
	/*	This is wrong because the days in the time	*/
	/*	sources are relative to 1! This means that	*/
	/*	we have to subtract one day's worth of seconds	*/
	/*	from the tv_sec that KSI calculated for us.	*/
	/*	Rather than doing that every time we read the	*/
	/*	board, we'll make the adjustment once when	*/
	/*	we determine what year it is.			*/
	/*							*/
	/********************************************************/

	timeNow -> tv_sec = ksiTime.tv_sec + ksiYear;
	timeNow -> tv_usec = ksiTime.tv_usec;
	ksiLastDay = ksiDay;
}
#endif WIN32

#ifdef WIN32
/* a utility routine that should be in a library somewhere */
static int
local_addr (int fd, struct sockaddr_in *pAddr)
{
    char		myName[80];
#ifndef vxWorks
    struct hostent	*hostent;
#endif

    gethostname(myName, sizeof(myName));

#ifdef vxWorks
    pAddr -> sin_addr.s_addr = hostGetByName(myName);
#else
    hostent = gethostbyname(myName);
    pAddr -> sin_addr.s_addr = ((struct in_addr *)hostent -> h_addr_list[0])->s_addr;
#endif

    return (0);
}
#endif WIN32
