#ifndef tispermdatabaseHXX
#define	tispermdatabaseHXX

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "dbDefs.h"	// .../share/epicsH
#include "dbRecDes.h"	// .../share/epicsH
// these are here since because they are needed if using a T51 dbDefs.h
#ifdef	T51DBDEF
#define	DESCRIPTION_SZ		30
#define	OPERATOR_SZ		30
#define	GROUPNAME_SZ		OPERATOR_SZ
#define	ENCRYPTPASSWD_SZ	30
#define	UNENCRYPTPASSWD_SZ	14
#define UTILACCPRIV_SZ		45
#endif	T51DBDEF
}

// define max database path
#ifndef WIN32
#define	TISPERM_MAX_DATABASE_PATH	MAXPATHLEN
#else
#define	TISPERM_MAX_DATABASE_PATH	_MAX_PATH
#endif WIN32

// define tasks path
#ifndef	WIN32
#define	TISPERM_TASKS_PATH	"/tis4000/tasks/"
#else
#define	TISPERM_TASKS_PATH	"\\tis4000\\tasks\\"
#endif	WIN32

// define some database names
#define	FLDPROTT_DB		"fldprott.dbf"
#define	PRIVTAB_DB		"privtab.dbf"
#define	SYSUSERS_DB		"sysusers.dbf"
#define	TISGROUPS_DB		"tisgrps.dbf"
#define	TISPROTGRPS_DB		"protgrps.dbf"
#define TISUSERS_DB		"tisusers.dbf"
#define	UTILACC_DB		"utilacc.dbf"
#define	FIELDS_INFO_FILE	"fieldsInfo.txt"

// define some database field names
#define	ALLOW_AUTOLOGOUT_FLD	"ALLOWAUTOL"	// tisusers.dbf
#define	AUTOLOGOUT_TIME_FLD	"AUTOLOGTIM"	// tisusers.dbf
#define	DESCRIPTION_FLD		"DESC"		// tisusers.dbf,tisgrps.dbf,protgrps.dbf
#define	FIELDNAME_FLD		"FIELDNAME"	// fldprott.dbf
#define	GROUPNAME_FLD		"GROUPNAME"	// tisgrps.dbf,utilacc.dbf
#define	NAME_FLD		"NAME"		// sysusers.dbf,tisusers.dbf
#define	OPERATORID_FLD		"OPERATORID"	// privtab.dbf
#define	POINTTYPE_FLD		"POINTTYPE"	// fldprott.dbf
#define	PRIV_LEVEL_FLD		"PRIV_LEVEL"	// privtab.dbf,utilacc.dbf
#define	PRIVLEVEL_FLD		"PRIVLEVEL"	// fldprott.dbf
#define	PRIVILEGE_FLD		"PRIVILEGE"	// utilacc.dbf
#define	PROT_GROUP_FLD		"PROT_GROUP"	// privtab.dbff,protgrps.dbf
#define	SYSTYPE_FLD		"OPTYPE"	// sysusers.dbf
#define	USERPASSWD_FLD		"ENCRPASSWD"	// tisusers.dbf


// define some tagnames
#define	GROUPNAME_TAG		"GROUPNAME"	// tisgroups.mdx, utilacc.mdx
#define	GROUPPRIV_TAG		"GROUPPRIV"	// utilacc.mdx
#define	NAME_TAG		"NAME"		// sysusers.mdx,tisussers.mdx
#define	OPERATORID_TAG		"OPERATORID"	// privtab.mdx
#define	OPERGROUP_TAG		"OPERGROUP"	// privtab.mdx
#define	POINTTYPE_TAG		"POINTTYPE"	// fldprott.mdx
#define	PROTGRP_TAG		"PROT_GROUP"	// protgrps.mdx	
#define	TYPEFIELD_TAG		"TYPEFIELD"	// fldprott.mdx

// define default values for privilege levels
enum DefaultPrivilegeLevels
{
    ViewPriv = 0,
    TraineePriv = 25,
    OperatorPriv = 50,
    SupervisorPriv = 70,
    EngineerPriv = 80,
    AdminPriv = 90
};

// define default values for privilege levels
enum DefaultProtectionLevels
{
    NoAccess = 0,
    ViewOnly = 1,
    FullAccess = 2
};

#define	MAX_PRIVILEGE_LEVEL	    "99"
#define	UNTITLED_DEFAULT	    "untitled"

//define permissions strings
#define	VIEW_STR	"VIEW ONLY"
#define	TRAINEE_STR	"TRAINEE"
#define	OPERATOR_STR	"OPERATOR"
#define	SUPERVISOR_STR	"SUPERVISOR"	
#define	ENGINEER_STR	"ENGINEER"
#define	ADMIN_STR	"ADMINSTRATOR"

// define privilege strings for protection group access
#define	NO_ACCESS_STR	"NO ACCESS"
#define	VIEW_ONLY_STR	"VIEW ONLY"
#define	FULL_ACCESS_STR	"FULL ACCESS"

// define string for unknown access
#define	UNKNOWN_ACCESS_STR	"UNKNOWN ACCESS"

// define a string length for permission strings
#define	PRIV_STR_SZ	(sizeof (ADMIN_STR) * 2)

// enum values for db functions
enum DbFuncs
{
    AddToDb = 0,
    ModTheDb,
    DelFromDb
};

typedef struct	field
{
    char    ptType [POINTTYPE_SZ + 1];
    char    fldName [FLDNAME_SZ + 1];
    int     fldType;
    int     special;
    char    fldComment [PROMPT_SZ + 1];
}TISPERM_FIELD_DEF;

class Tisprivtab;


class Sysusers
{
    TisPermDatabase *db;
    CODE4	    *cb;
    DATA4	    *sysusers;
    TAG4	    *nameTag;

    // field ptrs for sysusers.dbf
    FIELD4	    *sysnameFld;
    FIELD4	    *systypeFld;

    int		    createdSysusers;

    // private functions
    int CreateSysusersDb (char *dbpath);

public:

    // constructors/destructors
    Sysusers (CODE4 *cb);
    ~Sysusers ();

    // public funcs
    void	AddUser (char *userName, char *groupName);
    void	DeleteGroup (char *groupName);
    void	DeleteUser (vchar *name);
    vliststr	*GetGroupsUsers (vchar *groupname);
    int		GroupHasUsers (char *group);
    int		MapSysuserFlds (void);
    int		OpenSysusers (void);
    void	UpdateSysUsers (vliststr *list, vchar *group);
    int		UserInGroup (char *user, char *group);

    // public inline funcs
    int GetCreatedSysUsers (void)
	{ return (createdSysusers); }
    DATA4 *GetSysusersDb (void)
	{ return (sysusers); }
    FIELD4 *GetSysNameFld (void)
	{ return (sysnameFld); }
    FIELD4 *GetSysTypeFld (void)
	{ return (systypeFld); }
    void SetDb (TisPermDatabase *dB)
	{ db = dB; }
};


class Tisgroups
{
    TisPermDatabase *db;
    CODE4	    *cb;
    DATA4	    *tisgroups;
    TAG4	    *groupTag;

    // field ptrs for tisgroups.dbf
    FIELD4	    *groupnameFld;
    FIELD4	    *groupdescFld;

    // private functions
    void    AddGroup (vchar *name, vchar *desc);
    int	    CreateGroupsDb (char *dbpath, Sysusers *sysUsers);
    void    DeleteGroup (vchar *name);
    void    ModifyGroup (vchar *name, vchar *desc);

public:

    // constructors/destructors
    Tisgroups (CODE4 *cb);
    ~Tisgroups ();

    // public funcs
    vliststr	*GetGroupsList (void);
    int		MapGroupsFlds (void);
    int		OpenGroupsDb (Sysusers *sysUsers);
    void	UpdateGroupsDb (DbFuncs todo, vchar *name, vchar *desc);

    // inline funcs
    void SetDb (TisPermDatabase *dB)
	{ db = dB; }
};


class Tisprotgroups
{
    TisPermDatabase *db;
    CODE4	    *cb;
    DATA4	    *tisprotgrps;
    TAG4	    *protGrpTag;

    // field ptrs for tisprotgroups.dbf
    FIELD4	    *protgrpFld;
    FIELD4	    *protgrpdescFld;

    // private functions
    void    AddProtGroup (vchar *name, vchar *desc);
    int	    CreateProtGroupsDb (char *dbpath, Tisprivtab *tisPrivTab);
    void    DeleteProtGroup (vchar *name);
    void    ModifyProtGroup (vchar *name, vchar *desc);

public:

    // constructors/destructors
    Tisprotgroups (CODE4 *cb);
    ~Tisprotgroups ();

    // public funcs
    vliststr	*GetProtGroupsList (void);
    int		MapProtGroupsFlds (void);
    int		OpenProtGroupsDb (Tisprivtab *tisPrivTab);
    int		ProtGrpHasPts (char *protGrpName);
    void	UpdateProtGroupsDb (DbFuncs todo, vchar *name, vchar *desc);

    // inline funcs
    DATA4 *GetProtGroupsDb (void)
	{ return (tisprotgrps); }
    FIELD4 *GetProtGroupName (void)
	{ return (protgrpFld); }
    FIELD4 *GetProtGroupDesc (void)
	{ return (protgrpdescFld); }
    void SetDb (TisPermDatabase *dB)
	{ db = dB; }
};


class Tisfldprott
{
    TisPermDatabase *db;
    CODE4	    *cb;
    DATA4	    *tisfldprott;
    TAG4	    *typefieldTag;
    TAG4	    *pointtypeTag;

    // field ptrs for tisusers.dbf
    FIELD4	    *pointtypeFld;
    FIELD4	    *fieldnameFld;
    FIELD4	    *privlevelFld;
    FIELD4	    *allowautoFld;

    // private functions
    int		CreateFldProttDb (char *dbpath);

public:

    // constructors/destructors
    Tisfldprott (CODE4 *cb);
    ~Tisfldprott ();

    // public funcs
    void    ClearPrivLevel (char *pType, char *fieldName);
    int	    GetPrivLevel (char *pType, char *fieldName, int *pLevel);
    int	    MapFldProttFlds (void);
    int	    OpenFldProttDb (void);
    void    UpdatePrivLevel (char *pType, char *fieldName, 
			     int pLevel);
    // inline funcs
    void SetDb (TisPermDatabase *dB)
	{ db = dB; }
    void SetTypeFieldTag (void)
    {
	d4tag_select (tisfldprott, typefieldTag);
    }
};


class Tisprivtab
{
    TisPermDatabase *db;
    CODE4	    *cb;
    DATA4	    *tisprivtab;
    TAG4	    *operatoridTag;
    TAG4	    *opergroupTag;

    // field ptrs for privtab.dbf
    FIELD4	    *operatoridFld;
    FIELD4	    *prot_groupFld;
    FIELD4	    *priv_levelFld;

    // private functions
    int		CreatePrivTabDb (char *dbpath);
    int		GetNextGroup (char *seekbuff);

    // private field
    int		    createdPrivTab;

public:

    // constructors/destructors
    Tisprivtab (CODE4 *cb);
    ~Tisprivtab ();

    // public funcs
    void	DeleteGroup (char *groupName);
    vliststr	*GetProtGroupList (char *protGroup);
    vliststr	*GetPrGrpPermissions (char *groupName);
    int		GroupInProtectionGroup (char *groupName);
    int		MapPrivTabFlds (void);
    int		OpenPrivTabDb (void);
    int		PrtGrpHasGroups (char *protGrpName);
    void	UpdatePrivTabUtils (vliststr *selectedList, char *groupName);
    void	UpdatePrtGrp (vliststr *list, char *protGroup);
    void	UpdatePrtGrpPrivs (vliststr *list, char *groupname);
    // inline funcs
    int GetCreatedPrivTab (void)
	{ return (createdPrivTab); }
    FIELD4 *GetProtGrpFld (void)
	{ return (prot_groupFld); }
    DATA4 *GetPrvTabDb (void)
	{ return (tisprivtab); }	
    void SetDb (TisPermDatabase *dB)
	{ db = dB; }
    void SetOperatorIdTag (void)
    {
	d4tag_select (tisprivtab, operatoridTag);
    }
    void SetOperGroupTag (void)
    {
	d4tag_select (tisprivtab, opergroupTag);
    }
};


typedef struct user
{
    char    username [OPERATOR_SZ + 1];
    char    descfld [DESCRIPTION_SZ + 1];
    char    passwdfld [ENCRYPTPASSWD_SZ + 1];
    int	    allowAutologout;
    int	    autoLogoutTime;
}TISUSER;

class Tisusers
{
    TisPermDatabase *db;
    CODE4	    *cb;
    DATA4	    *tisusers;
    TAG4	    *nameTag;

    // field ptrs for tisusers.dbf
    FIELD4	    *usernameFld;
    FIELD4	    *userdescFld;
    FIELD4	    *userpasswdFld;
    FIELD4	    *allowautoFld;
    FIELD4	    *autologoutFld;

    // private functions
    void    AddUser (vchar *name, vchar *desc, vchar *encryPasswrd,
		     int allowAuto, int autoLogoutTime);
    int	    CreateTisusersDb (char *dbpath, Sysusers *sysUsers);
    void    DeleteUser (vchar *name);
    void    ModifyUser (vchar *name, vchar *desc, vchar *encryPasswrd,
			int allowAuto, int autoLogoutTime);

public:

    // constructors/destructors
    Tisusers (CODE4 *cb);
    ~Tisusers ();

    // public funcs
    void    DeleteUsersList (vlist *list);
    vlist   *GetUsersList (void);
    int	    MapUsersFlds (void);
    int	    OpenUsersDb (class Sysusers *sysUser);
    void    UpdateUsersDb (DbFuncs todo, vchar *name, vchar *desc, 
			   vchar *encrypasswd, int allowAuto, 
			   int autoLogoutTime);
    // inline funcs
    void SetDb (TisPermDatabase *dB)
	{ db = dB; }
};


class Tisutilacc
{
    TisPermDatabase *db;
    CODE4	    *cb;
    DATA4	    *tisutilacc;
    TAG4	    *groupnameTag;
    TAG4	    *groupprivTag;

    // field ptrs for tisusers.dbf
    FIELD4	    *groupnameFld;
    FIELD4	    *privilegeFld;
    FIELD4	    *priv_levelFld;

    // private functions
    int		CreateUtilAccDb (char *dbpath);
    int		GetNextGroup (char *seekbuff);

public:

    // constructors/destructors
    Tisutilacc (CODE4 *cb);
    ~Tisutilacc ();

    // public funcs
    void	DeleteGroup (char *groupName);
    vliststr	*GetUtilPermissions (char *groupName);
    int		MapUtilAccFlds (void);
    int		OpenUtilAccDb (void);
    void	UpdateUtilAccPrivileges (vliststr *selectedList, char *groupName);
    // inline funcs
    DATA4 *GetUtilAccDb (void)
	{ return (tisutilacc); }	
    void SetDb (TisPermDatabase *dB)
	{ db = dB; }
    void SetGroupNameTag (void)
    {
	d4tag_select (tisutilacc, groupnameTag);
    }
    void SetGroupPrivTag (void)
    {
	d4tag_select (tisutilacc, groupprivTag);
    }
};


class TisPermDatabase
{
    CODE4	cb;
    TisPermApp	*tisPermApp;

    // ptrs to database file classes
    Tisfldprott		*tisFldPrott;
    Tisprivtab		*tisPrivTab;
    Tisusers		*tisUsers;
    Tisgroups		*tisGroups;
    Tisprotgroups	*tisProtGroups;
    Sysusers		*sysUsers;
    Tisutilacc		*utilAcc;

    int			BuildLists (void);
    int			CountPointTypes (char *path, int *numPtTypes);
    vliststr		*ptTypeList;
    vlist		**fieldsLists;
    int			numFieldsLists;
    void		TraceLists (void);

public:

    TisPermDatabase (TisPermApp *app);
    ~TisPermDatabase ();

    int OpenDbFiles (void);

    // some inline funcs
    TisPermApp *GetTisPermApp (void)
	{ return (tisPermApp); }
    Sysusers *GetSysUsers (void)
	{ return (sysUsers); }
    Tisfldprott *GetFldPrott (void)
	{ return (tisFldPrott); }
    Tisprivtab *GetPrivTab (void)
	{ return (tisPrivTab); }
    Tisgroups *GetTisGroups (void)
	{ return (tisGroups); }
    Tisprotgroups *GetTisProtGroups (void)
	{ return (tisProtGroups); }
    Tisusers *GetTisUsers (void)
	{ return (tisUsers); }
    Tisutilacc *GetTisUtilAcc (void)
	{ return (utilAcc); }
    vlist **GetFieldsLists (void)
	{ return (fieldsLists); }
    vliststr *GetPtTypeList (void)
	{ return (ptTypeList); }

    vliststr	*GetUsersNotInGroups (void);
    vliststr	*GetGrpsNotInPrGrp (char *protGroup);

};

#endif tispermdatabaseHXX
