#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstrHEADER
#include vlistviewHEADER
#include vrulerHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vmenuHEADER
#include vlistHEADER
#include vlistitemHEADER

extern "C"
{
#include "dbDefs.h"	    // .../share/epcisH
#include "special.h"	    // .../share/epcisH
#include "dbFldTypes.h"	    // .../share/epicsH
}

#include "tisperm.h"
#include "tisPermApp.hxx"
#include "tisPermDatabase.hxx"
#include "tisPermMain.hxx"
#include "tisPermUtils.hxx"
#include "tisPermFldPerms.hxx"


vkindTYPED_FULL_DEFINITION(FldPermsDialog, vdialog, NULL);

/*******************************************************************
*
*    FldPermsDialog override for constructors init ()
*
*******************************************************************/

void FldPermsDialog::LoadInit (vresource res)
{
    listView = FldsListView::RequestSubLoad (this,
					tisperm_FieldView);

    comment = vtextitem::RequestSubLoad (this,
					tisperm_Comment);

    pointTypes = PointTypes::RequestSubLoad (this,
					tisperm_PointTypes);

    privilegeLevel = FldPrivLevel::RequestSubLoad (this,
					tisperm_PrivilegeLevel);

    applyButton = FldPermsApplyButton::RequestSubLoad (this,
					tisperm_ApplyButton);
   
    dismissButton = FldPermsDismissButton::RequestSubLoad (this,
					tisperm_DismissButton);
     
    clearButton = FldPermsClearButton::RequestSubLoad (this,
					tisperm_ClearButton);
    vdialog::LoadInit (res);

    // give items ptr to this dialog
    listView -> SetFldPermsDialog (this);
    pointTypes -> SetFldPermsDialog (this);
    privilegeLevel -> SetFldPermsDialog (this);
    applyButton -> SetFldPermsDialog (this);
    dismissButton -> SetFldPermsDialog (this);
    clearButton -> SetFldPermsDialog (this);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    FldPermsDialog Destroy function
*
*******************************************************************/

void FldPermsDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }

    vdialog::Destroy ();
}



/*******************************************************************
*
*    FldPermsDialog notify function
*
*******************************************************************/

void FldPermsDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    FldPermsDialog Open function
*
*******************************************************************/

void FldPermsDialog::OpenFldPerms (MainDialog *mnDlg)
{
    mainDialog = mnDlg;

    // build point type mneu
    pointTypes -> BuildMenu ();

    // build the fields list for the first point type
    listView -> SetInUseFlag (vFALSE);
    listView -> SetCalcColumnWidth (vTRUE);
    listView -> GetLists ();
    listView -> BuildFieldsList (0);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


vkindTYPED_FULL_DEFINITION(FldsListView, vlistview, NULL);

/*******************************************************************
*
*    FldsListView observe function
*
*******************************************************************/

void FldsListView::ObserveSelect (int state)
{
    if (state == vlistitemSELECTION_HAS_CHANGED)
    {
	if (inUseFlag == vTRUE)
	{
	    return;
	}
	vliststr *fldList = (vliststr *)(GetList ());
	vlistSelection *listSelection = GetSelection ();
	int row, column;
	fldList -> GetSelectedCell (listSelection, &row, &column);
	if (row == vlistNO_CELLS_SELECTED)
	{
	    return;
	}
	// get the current select point type
	int ptIndex = fldPermsDialog -> pointTypes -> GetValue ();
	// get its field list
	vlist *fldsList = fieldsLists [ptIndex];
	// get the field definition for the current selected field
	TISPERM_FIELD_DEF *fldDef = (TISPERM_FIELD_DEF *)(fldsList -> GetCellValue (row, 0));
	// update the comment field
	fldPermsDialog -> comment -> SetTextScribed (vcharScribeSystem (fldDef -> fldComment));

	vchar *privLevel = (vchar *)(fldList -> GetCellValue (row, 1));
	if ( ((vcharLength (privLevel)) == 0) || (privLevel [0] == ' ') )
	{
	    return;
	}
	else
	{
	    char privLevelStr [PRIV_STR_SZ + 1];
	    vcharExportBounded (vcharScribeSystemExport (privLevel),
				privLevelStr,
				sizeof (privLevelStr));
	    privLevelStr [sizeof (privLevelStr) - 1] = 0;
	    DefaultPrivilegeLevels privLevelValue = GetPrivLevelValue (privLevelStr);
	    fldPermsDialog -> privilegeLevel -> SetPrivilegeButton (privLevelValue);
	}
    }
    return;
}


/*******************************************************************
*
*    FldsListView Destroy function
*
*******************************************************************/

void FldsListView::Destroy (void)
{
    vliststr *list = (vliststr *)(GetList ());
    if (list != NULL)
    {
	SetList (NULL);
	delete list;
    }
    vlistview::Destroy ();
}


/*******************************************************************
*
*    FldsListView build the list function
*	we refernce 2 lists from the TisPermDatabase class
*		1 - fieldsLists array
*		    array of pointers to lists of field definitions
*		2 - point type list
*		    list of strings for each point type
*	both are indexed by the arg ptindex
*
*******************************************************************/

void FldsListView::BuildFieldsList (int ptindex)
{
    // get old list and destroy it
    vliststr *list = (vliststr *)(GetList ());
    SetList (NULL);
    if (list)
    {
	list -> Destroy ();
    }
    list = new vliststr;
    list -> SetColumnCount (2);

    MainDialog *mainDialog = fldPermsDialog -> GetMainDialog ();
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Tisfldprott *tisFldPrott = db -> GetFldPrott ();

    // set up database ptrs, stuff defined in vuDatabaseClass.hxx

    // use proper index key into field protection table
    tisFldPrott -> SetTypeFieldTag ();

    int			row = 0;
    vlist		*fldList = fieldsLists [ptindex];
    vlistIterator	iterator;
    TISPERM_FIELD_DEF	*fldDef;
    char		ptType [POINTTYPE_SZ + 1];
    vchar		*vPtr;

    vPtr = (vchar *)(ptTypeList -> GetCellValue (ptindex, 0));
    vcharExportBounded (vcharScribeSystemExport (vPtr),
			ptType,
			POINTTYPE_SZ);
    ptType [POINTTYPE_SZ] = 0;

    iterator.Start (fldList);
    while (iterator.Next ())
    {
	fldDef = (TISPERM_FIELD_DEF *)(iterator.GetValue ());
	// make sure it's a field accessable to 'dm'
	if ( (fldDef -> special != SPC_NOMOD) && 
	     (fldDef -> fldType != DBF_NOACCESS) &&
	     ((strncmp (fldDef -> fldName, "UDF", 3)) != 0) )
	{
	    char fieldName[FLDNAME_SZ + 1];
	    int  pLevel;
	    // get epics field name
	    strncpy (fieldName, fldDef -> fldName, FLDNAME_SZ);
	    fieldName[FLDNAME_SZ] = 0;
	    // get a protection level from fldprott.dbf
	    if ((tisFldPrott -> GetPrivLevel (ptType, fieldName, &pLevel)) == vTRUE)
	    {
		char privLevelStr [PRIV_STR_SZ];
		// get the visual string that corresponds to the stored numeric value
		GetPrivLevelStr (pLevel, privLevelStr);
		list -> SetCellValueScribed (row, 0, vcharScribeSystem (fieldName));
		list -> SetCellValueScribed (row, 1, vcharScribeSystem (privLevelStr));
		
	    }
	    else
	    {
		list -> SetCellValueScribed (row, 0, vcharScribeSystem (fieldName));
		list -> SetCellValue (row, 1, " ");
	    }
	    row++;
	}
    }
    iterator.Finish ();
    SetList (list);
    if (calcColumnWidth == vTRUE)
    {
	calcColumnWidth = vFALSE;
	vlistsize *size = GetSizes ();
	SetColumnWidth (vlistviewALL_COLUMNS, size -> GetColumnWidth (0) / 2);
    }
    // set list selection method
    vlistSelection *listSelection;
    listSelection = GetSelection ();
    list -> SetSelectMethod (listSelection, vlistFREE_SELECTION);
    list -> SetExtendRowSelection (listSelection, vTRUE);
}


/*******************************************************************
*
*    FldsListView get lists function
*	we refernce 2 lists from the TisPermDatabase class
*		1 - fieldsLists array
*		    array of pointers to lists of field definitions
*		2 - point type list
*		    list of strings for each point type
*	both are indexed by the arg ptindex
*
*******************************************************************/

void FldsListView::GetLists (void)
{
    MainDialog *mainDialog = fldPermsDialog -> GetMainDialog ();
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    fieldsLists = db -> GetFieldsLists ();
    ptTypeList = db -> GetPtTypeList ();
}


vkindTYPED_FULL_DEFINITION(PointTypes, vcontrolOptionMenu, NULL);

/*******************************************************************
*
*    PointTypes observe function
*
*******************************************************************/

void PointTypes::ObserveDialogItem (vevent *event)
{
    // get menu index
    int indexValue = GetValue ();
    // clear out the comment field
    fldPermsDialog -> comment -> SetTextScribed (vcharScribeLiteral (" "));
    // have the list build itself based on the point type
    fldPermsDialog -> listView -> BuildFieldsList (indexValue); 
    return;
}


/*******************************************************************
*
*    PointTypes observe function
*	dbRecType.h defines dbRecType
*
*******************************************************************/

void PointTypes::BuildMenu (void)
{
    vmenu		*menu = new vmenu;
    int			numFound = 0;
    vlistIterator	iterator;

    MainDialog		*mainDialog = fldPermsDialog -> GetMainDialog ();
    TisPermApp		*app = mainDialog -> GetPermApp ();
    TisPermDatabase	*db = app -> GetDb ();
    vliststr		*list = db -> GetPtTypeList ();
    vchar		*ptType;

    iterator.Start (list);
    while (iterator.Next ())
    {
	ptType = (vchar *)(iterator.GetValue ());
	numFound++;
	vmenuItem *item = new vmenuItem;
	item -> SetTitle (ptType);
	menu -> AppendItem (item);
    }
    iterator.Finish ();
    // attach the menu to the control
    SetMenu (menu);
    // set min / max for the control
    SetMin (0);
    SetMax (numFound - 1);
    // set value to curent menu item will be the first on
    SetValue (0);

    return;
}


vkindTYPED_FULL_DEFINITION(FldPrivLevel, vcontainerExclusiveGroup, NULL);

/*******************************************************************
*
*    FldPrivLevel load function
*
*******************************************************************/

void FldPrivLevel::LoadInit (vresource res)
{
    adminToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_Admin);

    engineerToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_Engineer);

    supervisorToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_Supervisor);

    operatorToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_Operator);

    traineeToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_Training);

    viewToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_View);

    vcontainerExclusiveGroup::LoadInit (res);

    return;
}


/*******************************************************************
*
*    FldPrivLevel observe function
*
*******************************************************************/

void FldPrivLevel::ObserveDialogItem (vevent *event)
{
    vliststr *list = (vliststr *)(fldPermsDialog -> listView -> GetList ());
    if (list == NULL)
    {
	return;
    }
    vlistSelection *lselection = fldPermsDialog -> listView -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (lselection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	return;
    }
    char *privLevelStr = GetPrivStr ();
    if (privLevelStr == NULL)
    {
	return;
    }

    vlistIterator iterator;
    iterator.StartWithSelection (lselection);
    while (iterator.Next ())
    {
	row = iterator.GetRow ();
	column = iterator.GetColumn ();
	if (column == 1)
	{
	    list -> SetCellValueScribed (row, 1, vcharScribeSystem (privLevelStr));
	}
    }
    iterator.Finish ();
    return;
}


/*******************************************************************
*
*    FldPrivLevel set radio button function
*
*******************************************************************/

char *FldPrivLevel::GetPrivStr (void)
{
    char *privLevelStr = NULL;
    vcontrolToggle *toggle = (vcontrolToggle *)(GetSelection ());
    if (toggle == adminToggle)
    {
	privLevelStr = ADMIN_STR;
    }
    else if (toggle == engineerToggle)
    {
	privLevelStr = ENGINEER_STR;
    }
    else if (toggle == supervisorToggle)
    {
	privLevelStr = SUPERVISOR_STR;
    }
    else if (toggle == operatorToggle)
    {
	privLevelStr = OPERATOR_STR;
    }
    else if (toggle == traineeToggle)
    {
	privLevelStr = TRAINEE_STR;
    }
    else if (toggle == viewToggle)
    {
	privLevelStr = VIEW_STR;
    }
    return (privLevelStr);
}

/*******************************************************************
*
*    FldPrivLevel set radio button function
*
*******************************************************************/

void FldPrivLevel::SetPrivilegeButton (DefaultPrivilegeLevels privLevel)
{
    switch (privLevel)
    {
	case ViewPriv:
	    SetSelection (viewToggle);
	    break;
	case TraineePriv:
	    SetSelection (traineeToggle);
	    break;
	case OperatorPriv:
	    SetSelection (operatorToggle);
	    break;
	case SupervisorPriv:
	    SetSelection (supervisorToggle);
	    break;
	case EngineerPriv:
	    SetSelection (engineerToggle);
	    break;
	case AdminPriv:
	    SetSelection (adminToggle);
	    break;
	default:
	    break;
    }
}


vkindTYPED_FULL_DEFINITION(FldPermsApplyButton, vbutton, NULL);

/*******************************************************************
*
*    FldPermsApplyButton observe function
*
*******************************************************************/

void FldPermsApplyButton::ObserveDialogItem (vevent *event)
{
    vliststr *list = (vliststr *)(fldPermsDialog -> listView -> GetList ());
    vlistSelection *selection = fldPermsDialog -> listView -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	tisPermMsgCenter (fldPermsDialog, TISPERM_INFO_NOTICE, 
				    tisperm_NothingSelectedMsg);
	return;
    }
    char *privLevelStrPtr = fldPermsDialog -> privilegeLevel -> GetPrivStr ();
    if (privLevelStrPtr == NULL)
    {
	tisPermMsgCenter (fldPermsDialog, TISPERM_INFO_NOTICE,
				    tisperm_NoPrivLevelMsg);
	// turn off all selections in list
	list -> UnselectAll (selection);
	return;
    }
    int result = tisPermMsgCenter (fldPermsDialog, TISPERM_CAUTION_NOTICE,
				    tisperm_UpdateFldPrivReq);
    if (result == vFALSE)
    {
	// rebuild the list
	int ptIdx = fldPermsDialog -> pointTypes -> GetValue ();
	int topRow = fldPermsDialog -> listView -> GetTopRow ();
	fldPermsDialog -> listView -> BuildFieldsList (ptIdx);
	fldPermsDialog -> listView -> ScrollToCell (topRow, 0);
	return;
    }
    int indexValue = fldPermsDialog -> pointTypes -> GetValue ();
    char pointType [POINTTYPE_SZ + 1];
    MainDialog *mainDialog = fldPermsDialog -> GetMainDialog ();
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    vliststr *ptList = db -> GetPtTypeList ();
    vchar *vPtr = (vchar *)(ptList -> GetCellValue (indexValue, 0));
    vcharExportBounded (vcharScribeSystemExport (vPtr),
			pointType,
			POINTTYPE_SZ);
    pointType [POINTTYPE_SZ] = 0;
    // get ptr to fld privilege class
    Tisfldprott *fldprott = db -> GetFldPrott ();
    // set the pointtype + fieldname as the database tag
    fldprott -> SetTypeFieldTag ();
    char fieldname [FDPROTFLD_SZ + 1];
    char privLevelStr [PRIV_STR_SZ + 1];
    vlistIterator iterator;
    // disable the listview's ObserveSelect ()
    fldPermsDialog -> listView -> SetInUseFlag (vTRUE);
    iterator.StartWithSelection (selection);
    while (iterator.Next ())
    {
	row = iterator.GetRow ();
	column = iterator.GetColumn ();
	if (column != 0)
	{
	    continue;
	}
	vchar *fieldValue = (vchar *)(list -> GetCellValue (row, 0));
	vchar *privLevelValue = (vchar *)(list -> GetCellValue (row, 1));
	if ( ((vcharLength (privLevelValue)) == 0) || (privLevelValue [0] == ' ') )
	{
	    list -> SetCellValueScribed (row, 1, vcharScribeSystem (privLevelStrPtr));
	}
	vcharExportBounded (vcharScribeSystemExport (fieldValue),
			    fieldname,
			    sizeof (fieldname));
	fieldname [sizeof (fieldname) - 1] = 0;
	vcharExportBounded (vcharScribeSystemExport (privLevelValue),
			    privLevelStr,
			    sizeof (privLevelStr));
	privLevelStr [sizeof (privLevelStr) - 1] = 0;
	DefaultPrivilegeLevels privlevel = GetPrivLevelValue (privLevelStr);
	fldprott -> UpdatePrivLevel (pointType, fieldname, privlevel);
    }
    iterator.Finish ();
    // re-enable listview's ObserveSelect ()
    fldPermsDialog -> listView -> SetInUseFlag (vFALSE);
    tisPermMsgCenter (fldPermsDialog, TISPERM_INFO_NOTICE, 
				    tisperm_UpdateFldPrivMsg);
    // turn off all selections in list
    list -> UnselectAll (selection);
    // unselect radio button toggle in priv box
    fldPermsDialog -> privilegeLevel -> SetSelection (NULL);
    // clear the comment 
    fldPermsDialog -> comment -> SetTextScribed (vcharScribeLiteral (" "));
    return;
}


vkindTYPED_FULL_DEFINITION(FldPermsDismissButton, vbutton, NULL);

/*******************************************************************
*
*    FldPermsDismissButton observe function
*
*******************************************************************/

void FldPermsDismissButton::ObserveDialogItem (vevent *event)
{
    vevent::StopProcessing ();
    return;
}


vkindTYPED_FULL_DEFINITION(FldPermsClearButton, vbutton, NULL);

/*******************************************************************
*
*    FldPermsClearButton observe function
*
*******************************************************************/

void FldPermsClearButton::ObserveDialogItem (vevent *event)
{
    vliststr *list = (vliststr *)(fldPermsDialog -> listView -> GetList ());
    vlistSelection *selection = fldPermsDialog -> listView -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	tisPermMsgCenter (fldPermsDialog, TISPERM_INFO_NOTICE, 
				    tisperm_NothingSelectedMsg);
	return;
    }
    int result = tisPermMsgCenter (fldPermsDialog, TISPERM_CAUTION_NOTICE,
				    tisperm_ClearFldPrivReq);
    if (result == vFALSE)
    {
	// turn off all selections in list
	list -> UnselectAll (selection);
	return;
    }
    int indexValue = fldPermsDialog -> pointTypes -> GetValue ();
    char pointType [POINTTYPE_SZ + 1];
    MainDialog *mainDialog = fldPermsDialog -> GetMainDialog ();
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    vliststr *ptList = db -> GetPtTypeList ();
    vchar *vPtr = (vchar *)(ptList -> GetCellValue (indexValue, 0));
    vcharExportBounded (vcharScribeSystemExport (vPtr),
			pointType,
			POINTTYPE_SZ);
    pointType [POINTTYPE_SZ] = 0;
    // get ptr to fld privilege class
    Tisfldprott *fldprott = db -> GetFldPrott ();
    // set the pointtype + fieldname as the database tag
    fldprott -> SetTypeFieldTag ();
    char fieldname [FDPROTFLD_SZ + 1];
    vlistIterator iterator;
    // disable the listview's ObserveSelect ()
    fldPermsDialog -> listView -> SetInUseFlag (vTRUE);
    iterator.StartWithSelection (selection);
    while (iterator.Next ())
    {
	row = iterator.GetRow ();
	column = iterator.GetColumn ();
	if (column != 0)
	{
	    continue;
	}
	vchar *fieldValue = (vchar *)(list -> GetCellValue (row, 0));
	vcharExportBounded (vcharScribeSystemExport (fieldValue),
			    fieldname,
			    sizeof (fieldname));
	fieldname [sizeof (fieldname) - 1] = 0;
	fldprott -> ClearPrivLevel (pointType, fieldname);
	list -> SetCellValueScribed (row, 1, vcharScribeSystem (""));
    }
    iterator.Finish ();
    // re-enable listview's ObserveSelect ()
    fldPermsDialog -> listView -> SetInUseFlag (vFALSE);
    tisPermMsgCenter (fldPermsDialog, TISPERM_INFO_NOTICE, 
				    tisperm_ClearFldPrivMsg);
    // turn off all selections in list
    list -> UnselectAll (selection);
    // unselect radio button toggle in priv box
    fldPermsDialog -> privilegeLevel -> SetSelection (NULL);
    // clear the comment
    fldPermsDialog -> comment -> SetTextScribed (vcharScribeLiteral (" "));
    return;
}

