#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstrHEADER
#include vlistviewHEADER
#include vrulerHEADER

#include "tisperm.h"
#include "tisPermApp.hxx"
#include "tisPermDatabase.hxx"
#include "tisPermMain.hxx"
#include "tisPermUtils.hxx"
#include "tisPermGrpMaint.hxx"
#include "tisPermOptions.hxx"
#include "tisPermAccess.hxx"


vkindTYPED_FULL_DEFINITION(GroupMaintDialog, vdialog, NULL);

/*******************************************************************
*
*    Group Manager Dialog override for constructors init ()
*
*******************************************************************/

void GroupMaintDialog::LoadInit (vresource res)
{
    membersListView = MembersListView::RequestSubLoad (this,
					tisperm_MembersList);

    usersListView = UsersListView::RequestSubLoad (this,
					tisperm_UsersList);
   
    addButton = GrpMaintAddButton::RequestSubLoad (this,
					tisperm_AddButton);

    removeButton = GrpMaintRemoveButton::RequestSubLoad (this,
					tisperm_RemoveButton);
   
    okButton = GrpMaintOkButton::RequestSubLoad (this,
					tisperm_OkButton);
   
    cancelButton = GrpMaintCancelButton::RequestSubLoad (this,
					tisperm_CancelButton);
   
    utilAccButton = GrpMaintUtilAccButton::RequestSubLoad (this,
					tisperm_UtilAccessButton);
    vdialog::LoadInit (res);

    // give items ptr to GroupMaintDialog 
    membersListView -> SetGrpMaintDialog (this);
    usersListView   -> SetGrpMaintDialog (this);
    addButton       -> SetGrpMaintDialog (this);
    removeButton    -> SetGrpMaintDialog (this);
    okButton        -> SetGrpMaintDialog (this);
    cancelButton    -> SetGrpMaintDialog (this);
    utilAccButton   -> SetGrpMaintDialog (this);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    GroupMaintDialog Destroy function
*
*******************************************************************/

void GroupMaintDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }

    vdialog::Destroy ();
}


/*******************************************************************
*
*    GroupMaintDialog notify function
*
*******************************************************************/

void GroupMaintDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    GroupMaintDialog Open function
*
*******************************************************************/

void GroupMaintDialog::OpenGrpMaint (MainDialog *mnDlg, vchar *grp)
{
    mainDialog = mnDlg;

    vcharCopyBounded (grp, group, GROUPNAME_SZ);
    group [GROUPNAME_SZ] = 0;

    SetDialogTitle ();

    membersListView -> BuildList (group);
    usersListView -> BuildList (group);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    GroupMaintDialog Set dialog title function,
*	add group name to dialog title
*
*******************************************************************/

void GroupMaintDialog::SetDialogTitle (void)
{
    TisPermApp	*tisPerm = mainDialog -> GetPermApp ();
    vresource	titlesRes = tisPerm -> GetTitlesRes ();
    
    vstr *titleFormat = vresourceGetString (titlesRes, tisperm_GroupMaint);
    vstr *dialogTitle = vstrCloneScribed (vcharScribeFormat (titleFormat, group));

    SetTitle (dialogTitle);

    vstrDestroy (titleFormat);
    vstrDestroy (dialogTitle);
}


vkindTYPED_FULL_DEFINITION(MembersListView, vlistview, NULL);

/*******************************************************************
*
*    MembersListView constructor override and destructor
*
*******************************************************************/

void MembersListView::LoadInit (vresource resource)
{
    modified = vFALSE;
    list = NULL;
    flag = vFALSE;
    vlistview::LoadInit (resource);
}

void MembersListView::Destroy (void)
{
    if (list != NULL)
    {
	delete list;
    }

    vlistview::Destroy ();
}


/*******************************************************************
*
*    MembersListView observe function
*	user has selected the member list, disable ADD button
*	and clear the user list
*	Make sure to re-enable the REMOVE button
*
*******************************************************************/

void MembersListView::ObserveSelect (int state)
{
    int usersFlag = grpMaintDialog -> usersListView -> GetFlag ();
    // users list has been selected, its ObserveSelect will Unselect
    // my list, OK but dont do this code, or my flag is set
    if ( (usersFlag == vTRUE) || (flag == vTRUE) )
    {
	return;
    }
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	flag = vTRUE;
	GrpMaintAddButton *addButton = grpMaintDialog -> addButton;
	if ((addButton -> IsEnabled ()) == vTRUE)
	{
	    addButton -> Disable ();
	}
	UsersListView *usersListView = grpMaintDialog -> usersListView;
	vlistSelection *lselection = usersListView -> GetSelection ();
	vliststr *viewList = (vliststr *)(usersListView -> GetList ());
	if ((viewList -> IsSelectionEmpty (lselection)) == vFALSE)
	{
	    viewList -> UnselectAll (lselection);
	}
	GrpMaintRemoveButton *removeButton = grpMaintDialog -> removeButton;
	if ((removeButton -> IsEnabled ()) == vFALSE)
	{
	    removeButton -> Enable ();
	}
	flag = vFALSE;
    }
}


/*******************************************************************
*
*    MembersListView build list function
*
*******************************************************************/

void MembersListView::BuildList (vchar *group)
{
    MainDialog	    *mainDialog = grpMaintDialog -> GetMainDialog ();
    TisPermApp	    *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Sysusers	    *sysUsers = db -> GetSysUsers ();

    list = sysUsers -> GetGroupsUsers (group);
    SetList (list);
    vlistSelection  *lselection = GetSelection ();
    list -> SetSelectMethod (lselection, vlistFREE_SELECTION);
    int	rowCount = list -> GetRowCount ();
    if (rowCount > 0)
    {
	list -> Sort (vlistROW, 0, NULL);
    }
}


vkindTYPED_FULL_DEFINITION(UsersListView, vlistview, NULL);

/*******************************************************************
*
*    UsersListView override constructor and destructor
*
*******************************************************************/

void UsersListView::LoadInit (vresource resource)
{
    list = NULL;
    flag = vFALSE;
    vlistview::LoadInit (resource);
}

void UsersListView::Destroy (void)
{
    if (list != NULL)
    {
	delete list;
    }

    vlistview::Destroy ();
}


/*******************************************************************
*
*    UsersListView observe function
*	user has selected the user list, disable REMOVE button
*	and clear the members list
*	Make sure to re-enable the ADD button
*
*******************************************************************/

void UsersListView::ObserveSelect (int state)
{
    int membersFlag = grpMaintDialog -> membersListView -> GetFlag ();
    // members list has been selected, its ObserveSelect will Unselect
    // my list, OK but dont do this code, also if my flag is set
    if ( (membersFlag == vTRUE) || (flag == vTRUE) )
    {
	return;
    }
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	flag = vTRUE;
	GrpMaintRemoveButton *removeButton = grpMaintDialog -> removeButton;
	if ((removeButton -> IsEnabled ()) == vTRUE)
	{
	    removeButton -> Disable ();
	}
	MembersListView *membersListView = grpMaintDialog -> membersListView;
	vlistSelection *lselection = membersListView -> GetSelection ();
	vliststr *viewList = (vliststr *)(membersListView -> GetList ());
	if ((viewList -> IsSelectionEmpty (lselection)) == vFALSE)
	{
	    viewList -> UnselectAll (lselection);
	}
	GrpMaintAddButton *addButton = grpMaintDialog -> addButton;
	if ((addButton -> IsEnabled ()) == vFALSE)
	{
	    addButton -> Enable ();
	}
	flag = vFALSE;
    }
}


/*******************************************************************
*
*    UsersListView build list function
*
*******************************************************************/

void UsersListView::BuildList (vchar *group)
{
    MainDialog	    *mainDialog = grpMaintDialog -> GetMainDialog ();
    TisPermApp	    *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();

    list = db -> GetUsersNotInGroups ();
    SetList (list);
    vlistSelection  *lselection = GetSelection ();
    list -> SetSelectMethod (lselection, vlistFREE_SELECTION);
    int	rowCount = list -> GetRowCount ();
    if (rowCount > 0)
    {
	list -> Sort (vlistROW, 0, NULL);
    }
}


vkindTYPED_FULL_DEFINITION(GrpMaintAddButton, vbutton, NULL);

/*******************************************************************
*
*    GrpMaintAddButton observe function
*
*******************************************************************/

void GrpMaintAddButton::ObserveDialogItem (vevent *event)
{
    UsersListView *usersListView = grpMaintDialog -> usersListView;
    vlistSelection *selection = usersListView -> GetSelection ();
    vliststr *usersList = (vliststr *)(usersListView -> GetList ());
    if ((usersList -> IsSelectionEmpty (selection)) == vTRUE)
    {
	tisPermMsgCenter (grpMaintDialog, TISPERM_INFO_NOTICE, 
				   tisperm_NothingSelectedMsg);
    }
    else
    {
	MembersListView *membersListView = grpMaintDialog -> membersListView;
	vliststr *membersList = (vliststr *)(membersListView -> GetList ());
	vchar *listValue;
	int membersRowCount = membersList -> GetRowCount ();

	// disbale ObservSelect for both views
	membersListView -> SetFlag (vTRUE);
	usersListView -> SetFlag (vTRUE);

	vlistIterator iterator;
	iterator.StartWithSelection (selection);
	while (iterator.Next ())
	{
	    listValue = (vchar *)(iterator.GetValue ()); 
	    membersList -> SetCellValueScribed (membersRowCount, 0,
					        vcharScribeSystem (listValue));
	    membersRowCount++;
	}
	iterator.Finish ();
	// remove selected row from the users list, do here instead of 
	// in above iteration, removing a row while in an iteration
	// screws the iteration up
	int numUserRows = usersList -> GetRowCount();
	int i = 0;
	while(numUserRows)
	{
	    if(usersList -> IsCellSelected (selection, i, 0))
	    {
		usersList -> RemoveRow(i);
		numUserRows--;
	    }
	    else
	    {
		if(++i > numUserRows) break;
	    }
	}
	membersList -> Sort (vlistROW, 0, NULL);

	// reenable ObserveSelect for views
	membersListView -> SetFlag (vFALSE);
	usersListView -> SetFlag (vFALSE);
	membersListView -> SetModified (vTRUE);
    }
}


vkindTYPED_FULL_DEFINITION(GrpMaintRemoveButton, vbutton, NULL);

/*******************************************************************
*
*    GrpMaintRemoveButton observe function
*
*******************************************************************/

void GrpMaintRemoveButton::ObserveDialogItem (vevent *event)
{
    MembersListView *membersListView = grpMaintDialog -> membersListView;
    vlistSelection *selection = membersListView -> GetSelection ();
    vliststr *membersList = (vliststr *)(membersListView -> GetList ());
    if ((membersList -> IsSelectionEmpty (selection)) == vTRUE)
    {
	tisPermMsgCenter (grpMaintDialog, TISPERM_INFO_NOTICE, 
				   tisperm_NothingSelectedMsg);
    }
    else
    {
	UsersListView *usersListView = grpMaintDialog -> usersListView;
	vliststr *usersList = (vliststr *)(usersListView -> GetList ());
	vchar *listValue;
	int usersRowCount = usersList -> GetRowCount ();

	// disable ObserveSelect for both views
	membersListView -> SetFlag (vTRUE);
	usersListView -> SetFlag (vTRUE);

	vlistIterator iterator;
	iterator.StartWithSelection (selection);
	while (iterator.Next ())
	{
	    listValue = (vchar *)(iterator.GetValue ()); 
	    usersList -> SetCellValueScribed (usersRowCount, 0,
					        vcharScribeSystem (listValue));
	    usersRowCount++;
	}
	iterator.Finish ();
	// remove selected row from the members list, do here instead of 
	// in above iteration, removing a row while in an iteration
	// screws the iteration up
	int numMemberRows = membersList -> GetRowCount();
	int i = 0;
	while(numMemberRows)
	{
	    if(membersList -> IsCellSelected (selection, i, 0))
	    {
		membersList -> RemoveRow(i);
		numMemberRows--;
	    }
	    else
	    {
		if(++i > numMemberRows) break;
	    }
	}
	usersList -> Sort (vlistROW, 0, NULL);

	// reenable ObserveSelect for both views
	membersListView -> SetFlag (vFALSE);
	usersListView -> SetFlag (vFALSE);
	membersListView -> SetModified (vTRUE);
    }
}


vkindTYPED_FULL_DEFINITION(GrpMaintOkButton, vbutton, NULL);

/*******************************************************************
*
*    GrpMaintOkButton observe function
*
*******************************************************************/

void GrpMaintOkButton::ObserveDialogItem (vevent *event)
{
    MembersListView *membersListView = grpMaintDialog -> membersListView;
    if ((membersListView -> GetModified ()) == vTRUE)
    {
	vchar *group = grpMaintDialog -> GetGroupName ();
	MainDialog *mainDialog = grpMaintDialog -> GetMainDialog ();
	TisPermApp *app = mainDialog -> GetPermApp ();
	TisPermDatabase *db = app -> GetDb ();
	Sysusers *sysusers = db -> GetSysUsers ();
	vliststr *list = (vliststr *)(membersListView -> GetList ());

	sysusers -> UpdateSysUsers (list, group);

	tisPermMsgCenter (grpMaintDialog, TISPERM_INFO_NOTICE, 
				   tisperm_GroupUpdatedMsg, group);
    }
    vevent::StopProcessing ();
    return;
}


vkindTYPED_FULL_DEFINITION(GrpMaintCancelButton, vbutton, NULL);

/*******************************************************************
*
*    GrpMaintCancelButton observe function
*
*******************************************************************/

void GrpMaintCancelButton::ObserveDialogItem (vevent *event)
{
    vevent::StopProcessing ();
    return;
}


vkindTYPED_FULL_DEFINITION(GrpMaintUtilAccButton, vbutton, NULL);

/*******************************************************************
*
*    GrpMaintUtilAccButton observe function
*
*******************************************************************/

void GrpMaintUtilAccButton::ObserveDialogItem (vevent *event)
{
    vchar	*groupName = grpMaintDialog -> GetGroupName ();
    MainDialog	*mainDialog = grpMaintDialog -> GetMainDialog ();
    TisPermApp	*app = mainDialog -> GetPermApp ();
    vresource	tisPermRes = app -> GetPermRes ();
    vresource	accRes = vresourceGet (tisPermRes, tisperm_AccessDialog);
    AccessDialog *accessDialog = new AccessDialog (accRes);
    // allow accessDialog to come to the top of the window stack
    grpMaintDialog -> SetModal (vFALSE);
    accessDialog -> OpenAccess (mainDialog, groupName);
    delete accessDialog;
    // restore modality
    grpMaintDialog -> SetModal (vTRUE);
}

