#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vstrHEADER
#include vlistviewHEADER
#include vrulerHEADER
#include vcomboboxHEADER
#include vimageHEADER

#include "tisperm.h"
#include "tisPermApp.hxx"
#include "tisPermDatabase.hxx"
#include "tisPermMain.hxx"
#include "tisPermGroup.hxx"
#include "tisPermProtGrp.hxx"
#include "tisPermUser.hxx"
#include "tisPermGrpMaint.hxx"
#include "tisPermPrtGrpMaint.hxx"
#include "tisPermFldPerms.hxx"
#include "tisPermOptions.hxx"
#include "tisPermAccess.hxx"
#include "tisPermUtils.hxx"


vkindTYPED_FULL_DEFINITION(MainDialog, vdialog, NULL);

/*******************************************************************
*
*    MainDialog override for constructors init ()
*
*******************************************************************/

void MainDialog::LoadInit (vresource res)
{
    // init
    tisAbout = NULL;

    // store the main dialog's resource
    mainDialogResource = res;

    // RequestSubLoad returns ptr to uninited dialog object
    // gets init'd when dialog gets init'd by LoadInit()
    vmenubar *menubar = vmenubar::RequestSubLoad (this,
					tisperm_MenuBar);

    // handle the File menu
    vmenu *fileMenu = vmenu::RequestSubLoad (menubar,
					tisperm_FileMenu);
    vmenuItem *quitMenuItem = vmenuItem::RequestSubLoad (fileMenu,
					tisperm_QuitItem);
    
    // handle the User Menu
    vmenu *userMenu = vmenu::RequestSubLoad (menubar,
					tisperm_UserMenu);
    vmenuItem *newUser = vmenuItem::RequestSubLoad (userMenu,
					tisperm_NewUser);
    vmenuItem *modifyUser = vmenuItem::RequestSubLoad (userMenu,
					tisperm_ModifyUser);
    vmenuItem *deleteUser = vmenuItem::RequestSubLoad (userMenu,
					tisperm_DeleteUser);
    
    // handle the Group Menu
    vmenu *groupMenu = vmenu::RequestSubLoad (menubar,
					tisperm_GroupMenu);
    vmenuItem *newGroup = vmenuItem::RequestSubLoad (groupMenu,
					tisperm_NewGroup);
    vmenuItem *modifyGroup = vmenuItem::RequestSubLoad (groupMenu,
					tisperm_ModifyGroup);
    vmenuItem *deleteGroup = vmenuItem::RequestSubLoad (groupMenu,
					tisperm_DeleteGroup);
    vmenuItem *groupMaint = vmenuItem::RequestSubLoad (groupMenu,
					tisperm_GroupMaint);
    vmenuItem *assignPerms = vmenuItem::RequestSubLoad (groupMenu,
					tisperm_AssignPerms);
    
    // handle the Protection Group Menu
    vmenu *prtGroupMenu = vmenu::RequestSubLoad (menubar,
					tisperm_PrtGroupMenu);
    vmenuItem *newPrtGroup = vmenuItem::RequestSubLoad (prtGroupMenu,
					tisperm_NewPrtGroup);
    vmenuItem *modifyPrtGroup = vmenuItem::RequestSubLoad (prtGroupMenu,
					tisperm_ModifyPrtGroup);
    vmenuItem *deletePrtGroup = vmenuItem::RequestSubLoad (prtGroupMenu,
					tisperm_DeletePrtGroup);
    vmenuItem *assignGroups = vmenuItem::RequestSubLoad (prtGroupMenu,
					tisperm_AssignGroups);

    // handle the Edit Menu
    vmenu *editMenu = vmenu::RequestSubLoad (menubar, tisperm_EditMenu);
    vmenuItem *fldProtect = vmenuItem::RequestSubLoad (editMenu,
					tisperm_FldProtections);

    // handle the Help Menu
    helpMenu = vmenu::RequestSubLoad (menubar, tisperm_HelpMenu);

    // load the Main dialogs vlistviews and rulers
    userRuler  = vruler::RequestSubLoad (this, tisperm_UserRuler);
    groupRuler = vruler::RequestSubLoad (this, tisperm_GroupRuler);
    prGrpRuler = vruler::RequestSubLoad (this, tisperm_PrGrpRuler);

    userListView  = UserListView::RequestSubLoad (this, tisperm_UserListView);
    groupListView = GroupListView::RequestSubLoad (this, tisperm_GroupListView);
    prGrpListView = ProtGroupListView::RequestSubLoad (this, tisperm_PrGrpListView);

    // init inherited data
    vdialog::LoadInit (res);

    userListView -> StoreMainDialog (this);
    groupListView -> StoreMainDialog (this);
    prGrpListView -> StoreMainDialog (this);

    // store internal names for the file menu tags
    commandNames[QUIT_MENUITEM] = tisperm_QuitItem;

    // store internal names for the user menu tags
    commandNames[NEWUSER_MENUITEM] = tisperm_NewUser;
    commandNames[MODUSER_MENUITEM] = tisperm_ModifyUser;
    commandNames[DELUSER_MENUITEM] = tisperm_DeleteUser;

    // store internal names for the group menu tags
    commandNames[NEWGRP_MENUITEM]      = tisperm_NewGroup;
    commandNames[MODGRP_MENUITEM]      = tisperm_ModifyGroup;
    commandNames[DELGRP_MENUITEM]      = tisperm_DeleteGroup;
    commandNames[GRPMAINT_MENUITEM]    = tisperm_GroupMaint;
    commandNames[ASSIGNPERMS_MENUITEM] = tisperm_AssignPerms;

    // store internal names for the protection group menu tags
    commandNames[NEWPTGRP_MENUITEM]   = tisperm_NewPrtGroup;
    commandNames[MODPTGRP_MENUITEM]   = tisperm_ModifyPrtGroup;
    commandNames[DELPTGRP_MENUITEM]   = tisperm_DeletePrtGroup;
    commandNames[ASSIGNGRPS_MENUITEM] = tisperm_AssignGroups;

    // store internal names for the edit menu tags
    commandNames[FLDPROTECT_MENUITEM] = tisperm_FldProtections;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;
    // populate our dictionary with menu commands
    for (i = QUIT_MENUITEM; i < MENUITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	MainCommand *command = new MainCommand;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (commandNames[i], command);
    }
    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    MainDialog Destroy function
*
*******************************************************************/

void MainDialog::Destroy (void)
{
    if (tisAbout != NULL)
    {
	delete tisAbout;
    }
    if (IsOpen ())
    {
	Close ();
    }

    vdialog::Destroy ();
}


/*******************************************************************
*
*    MainDialog Handle window resize function
*	need to resize the ruler items in the rulers
*
*******************************************************************/

int MainDialog::HandleConfigure (vevent *event)
{
    vwindow::HandleConfigure (event);
    if (event -> GetType () == veventWINDOW_CONFIGURE && event -> IsSize ())
    {
	int	    rulerwidth;
	int	    itemwidth;
	vrulerItem  *item;
	const vrect *rulerRect;

	rulerRect = userRuler -> GetRect ();
	rulerwidth = rulerRect -> GetWidth ();
	item = userRuler -> GetItemAt (0);
	item -> SetWidth (rulerwidth/2);
	itemwidth = item -> GetWidth ();
	item = userRuler -> GetItemAt (1);
	item -> SetWidth (rulerwidth - itemwidth - 2);
	userRuler -> Compact ();

	rulerRect = groupRuler -> GetRect ();
	rulerwidth = rulerRect -> GetWidth ();
	item = groupRuler -> GetItemAt (0);
	item -> SetWidth (rulerwidth/2);
	itemwidth = item -> GetWidth ();
	item = groupRuler -> GetItemAt (1);
	item -> SetWidth (rulerwidth - itemwidth - 2);
	groupRuler -> Compact ();
	
	rulerRect = prGrpRuler -> GetRect ();
	rulerwidth = rulerRect -> GetWidth ();
	item = prGrpRuler -> GetItemAt (0);
	item -> SetWidth (rulerwidth/2);
	itemwidth = item -> GetWidth ();
	item = prGrpRuler -> GetItemAt (1);
	item -> SetWidth (rulerwidth - itemwidth - 2);
	prGrpRuler -> Compact ();

	const vrect *listRect;
	listRect = userListView -> GetListRect ();
	userListView -> SetColumnWidth (vlistviewALL_COLUMNS,
				        listRect -> GetWidth () / 2);
	listRect = groupListView -> GetListRect ();
	groupListView -> SetColumnWidth (vlistviewALL_COLUMNS,
					 listRect -> GetWidth () / 2);
	listRect = prGrpListView -> GetListRect ();
	prGrpListView -> SetColumnWidth (vlistviewALL_COLUMNS,
					 listRect -> GetWidth () / 2);
    }
    return (vTRUE);
}



/*******************************************************************
*
*    MainDialog notify function
*
*******************************************************************/

void MainDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    MainDialog Open function
*
*******************************************************************/

void MainDialog::OpenMain (TisPermApp *app)
{
    // so main will iconify properly to task bar in NT
    AddStyle (vwindowSTYLE_GLOBAL);

    tisPermApp = app;

    BuildAboutMenu ();

    vdialog::Open ();

    userListView -> BuildList ();
    groupListView -> BuildList ();
    prGrpListView -> BuildList ();
}


/*******************************************************************
*
*    MainDialog build about menuitem function
*
*******************************************************************/

void MainDialog::BuildAboutMenu (void)
{
    vstr *verMsg = vresourceGetString (tisPermApp -> GetMessagesRes (), 
					tisperm_VersionMsg);
    tisAbout = new TisAbout (tisPermApp -> GetTisGalaxyRes (), verMsg);
    vstrDestroy (verMsg);
    vstr *aboutTitle = vresourceGetString (tisPermApp -> GetTitlesRes (),
					   tisperm_AboutTitle);
    tisAbout -> SetAboutTitle (aboutTitle);
    vstrDestroy (aboutTitle);
    AboutMenuItem *aboutItem = tisAbout -> GetAboutMenuItem ();
    helpMenu -> AppendItem (aboutItem);
}


vkindTYPED_FULL_DEFINITION(MainCommand, vcommandSelector, NULL);
/*******************************************************************
*
*    MainCommand menu selector commands
*
*******************************************************************/

int MainCommand::Issue (vdict *dict)
{
    MainDialog	*mainDialog  = (MainDialog *) dict -> GetElem (vname_Dialog);
    TisPermApp *tisPermApp   = mainDialog -> GetPermApp ();
    TisPermDatabase *db      = tisPermApp -> GetDb ();
    vresource	tisPermRes   = tisPermApp -> GetPermRes ();
    vresource	titlesRes    = tisPermApp -> GetTitlesRes ();
    vresource	msgsRes	     = tisPermApp -> GetMessagesRes ();
    vliststr	*liststr;
    vlist	*list;
    vlistSelection *selection;
    vchar	*nameStr;
    vchar	*groupStr;
    vstr	*strValue;
    vstr	*todoStr;
    int		row, col;
    int		result;
    TISUSER	*tisUser;

    int command = GetSelector ();

    switch (command)
    {
	// File menu items

	case QUIT_MENUITEM:
	{
	    vevent::StopProcessing ();
	    break;
	}

	case NEWUSER_MENUITEM:
	{
	    vresource userRes = vresourceGet (tisPermRes, tisperm_UsernameDialog);
	    UserDialog *userDialog = new UserDialog (userRes);
	    userDialog -> OpenUser (mainDialog, UserAdd);
	    delete userDialog;
	    break;
	}

	case MODUSER_MENUITEM:
	{
	    selection = mainDialog -> userListView -> GetSelection ();
	    list = mainDialog -> userListView -> GetList ();
	    int rowCount = list -> GetRowCount ();
	    if (rowCount == 0)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_EmptyListMsg);
		break;
	    }
	    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE,
				  tisperm_NothingSelectedMsg);
		break;
	    }
	    vresource userRes = vresourceGet (tisPermRes, tisperm_UsernameDialog);
	    UserDialog *userDialog = new UserDialog (userRes);
	    userDialog -> OpenUser (mainDialog, UserModify);
	    delete userDialog;
	    list -> UnselectAll (selection);
	    break;
	}

	case DELUSER_MENUITEM:
	{
	    list = (mainDialog -> userListView -> GetList ());
	    selection = mainDialog -> userListView -> GetSelection ();
	    int rowCount = list -> GetRowCount ();
	    if (rowCount == 0)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_EmptyListMsg);
		break;
	    }
	    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				   tisperm_NothingSelectedMsg);
	    }
	    else
	    {
		list -> GetSelectedCell (selection, &row, &col);
		tisUser = (TISUSER *)(list -> GetCellValue (row, 0));
		Sysusers *sysUsers = db -> GetSysUsers ();
		char userGroup [GROUPNAME_SZ + 1];
		vstr *nameStr = vstrCloneScribed (vcharScribeSystem (tisUser -> username));
		// can not delete a user who is part of a group, force them to
		// delete user from group
		if ((sysUsers -> UserInGroup (tisUser -> username, userGroup)) == vTRUE)
		{
		    vstr *groupStr = vstrCloneScribed (vcharScribeSystem (userGroup));
		    tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				      tisperm_UserDeleteErr, nameStr, 
				      groupStr, nameStr);
		    vstrDestroy (nameStr);
		    vstrDestroy (groupStr);
		    list -> UnselectAll (selection);
		    break;
		}
		result = tisPermMsgCenter (mainDialog, TISPERM_CAUTION_NOTICE,
					    tisperm_DeleteUserReqMsg, nameStr);
		if (result == vTRUE)
		{
		    // clone nameStr, ModList, deletes entry from list and nameStr
		    // would no longer exist
		    strValue = vstrClone (nameStr);
		    mainDialog -> userListView -> ModList (DelFromList, strValue, 
							   NULL, NULL, 0, 0);
		    todoStr = vresourceGetString (msgsRes, tisperm_DelStr);
		    tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				       tisperm_UserDbUpdateMsg, strValue, todoStr);
		    vstrDestroy (strValue);
		    vstrDestroy (todoStr);
		}
		vstrDestroy (nameStr);
		list -> UnselectAll (selection);
	    }
	    break;
	}	

	case NEWGRP_MENUITEM:
	{
	    vresource groupRes = vresourceGet (tisPermRes, tisperm_GroupnameDialog);
	    GroupDialog *groupDialog = new GroupDialog (groupRes);
	    groupDialog -> OpenGroup (mainDialog, GroupAdd);
	    delete groupDialog;
	    break;
	}

	case MODGRP_MENUITEM:
	{
	    selection = mainDialog -> groupListView -> GetSelection ();
	    liststr = (vliststr *)(mainDialog -> groupListView -> GetList ());
	    int rowCount = liststr -> GetRowCount ();
	    if (rowCount == 0)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_EmptyListMsg);
		break;
	    }
	    if ((liststr -> IsSelectionEmpty (selection)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_NothingSelectedMsg);
		break;
	    }
	    vresource groupRes = vresourceGet (tisPermRes, tisperm_GroupnameDialog);
	    GroupDialog *groupDialog = new GroupDialog (groupRes);
	    groupDialog -> OpenGroup (mainDialog, GroupModify);
	    delete groupDialog;
	    liststr -> UnselectAll (selection);
	    break;
	}

	case DELGRP_MENUITEM:
	{
	    liststr = (vliststr *)(mainDialog -> groupListView -> GetList ());
	    selection = mainDialog -> groupListView -> GetSelection ();
	    int rowCount = liststr -> GetRowCount ();
	    if (rowCount == 0)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_EmptyListMsg);
		break;
	    }
	    if ((liststr -> IsSelectionEmpty (selection)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				   tisperm_NothingSelectedMsg);
		break;
	    }
	    liststr -> GetSelectedCell (selection, &row, &col);
	    groupStr = (vchar *)(liststr -> GetCellValue (row, 0));
	    char	groupName [GROUPNAME_SZ + 1];
	    vcharExportBounded (vcharScribeSystemExport (groupStr),
				groupName,
				sizeof (groupName));
	    groupName [sizeof (groupName) - 1] = 0;
	    Sysusers *sysUsers = db -> GetSysUsers ();
	    Tisprivtab *privtab = db -> GetPrivTab ();
	    // can not delete a group who is part of a protection group(s), 
	    if ((privtab -> GroupInProtectionGroup (groupName)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE,
				  tisperm_GroupPgDeleteErr,
				  groupStr, groupStr);
		liststr -> UnselectAll (selection);
	    }
	    // can not delete a group that has users in it
	    // force them to delete user(s) group
	    else if ((sysUsers -> GroupHasUsers (groupName)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE,
				  tisperm_GroupUserDeleteErr,
				  groupStr, groupStr, groupStr);
		liststr -> UnselectAll (selection);
	    }
	    else
	    {
		result = tisPermMsgCenter (mainDialog, TISPERM_CAUTION_NOTICE,
					    tisperm_DeleteGroupReqMsg, groupStr);
		if (result == vTRUE)
		{
		    // clone groupStr, ModList, deletes entry from list and nameStr
		    // would no longer exist
		    strValue = vstrClone (groupStr);
		    mainDialog -> groupListView -> ModList (DelFromList, strValue, NULL);
		    todoStr = vresourceGetString (msgsRes, tisperm_DelStr);
		    tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				       tisperm_GroupDbUpdateMsg, strValue, todoStr);
		    vstrDestroy (strValue);
		    vstrDestroy (todoStr);
		}
		liststr -> UnselectAll (selection);
	    }
	    break;
	}

	case GRPMAINT_MENUITEM:
	{
	    selection = mainDialog -> groupListView -> GetSelection ();
	    liststr = (vliststr *)(mainDialog -> groupListView -> GetList ());
	    int rowCount = liststr -> GetRowCount ();
	    if (rowCount == 0)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_EmptyListMsg);
		break;
	    }
	    if ((liststr -> IsSelectionEmpty (selection)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_NothingSelectedMsg);
		break;
	    }
	    liststr -> GetSelectedCell (selection, &row, &col);
	    nameStr = (vchar *)(liststr -> GetCellValue (row, 0));
	    vresource grpMaintRes = vresourceGet (tisPermRes, tisperm_GroupMaintDialog);
	    GroupMaintDialog *grpMaintDlg = new GroupMaintDialog (grpMaintRes);
	    grpMaintDlg -> OpenGrpMaint (mainDialog, nameStr);
	    delete grpMaintDlg;
	    liststr -> UnselectAll (selection);
	    break;
	}

	case ASSIGNPERMS_MENUITEM:
	{
	    selection = mainDialog -> groupListView -> GetSelection ();
	    liststr = (vliststr *)(mainDialog -> groupListView -> GetList ());
	    int rowCount = liststr -> GetRowCount ();
	    if (rowCount == 0)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_EmptyListMsg);
		break;
	    }
	    if ((liststr -> IsSelectionEmpty (selection)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_NothingSelectedMsg);
		break;
	    }
	    liststr -> GetSelectedCell (selection, &row, &col);
	    nameStr = (vchar *)(liststr -> GetCellValue (row, 0));
	    vresource accRes = vresourceGet (tisPermRes, tisperm_AccessDialog);
	    AccessDialog *accessDialog = new AccessDialog (accRes);
	    accessDialog -> OpenAccess (mainDialog, nameStr);
	    delete accessDialog;
	    liststr -> UnselectAll (selection);
	    break;
	}

	case NEWPTGRP_MENUITEM:
	{
	    vresource protGrpRes = vresourceGet (tisPermRes, tisperm_PtGrpDialog);
	    ProtGroupDialog *protGroupDialog = new ProtGroupDialog (protGrpRes);
	    protGroupDialog -> OpenProtGroup (mainDialog, ProtGroupAdd);
	    delete protGroupDialog;
	    break;
	}

	case MODPTGRP_MENUITEM:
	{
	    selection = mainDialog -> prGrpListView -> GetSelection ();
	    liststr = (vliststr *)(mainDialog -> prGrpListView -> GetList ());
	    int rowCount = liststr -> GetRowCount ();
	    if (rowCount == 0)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_EmptyListMsg);
		break;
	    }
	    if ((liststr -> IsSelectionEmpty (selection)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_NothingSelectedMsg);
		break;
	    }
	    vresource protGrpRes = vresourceGet (tisPermRes, tisperm_PtGrpDialog);
	    ProtGroupDialog *protGroupDialog = new ProtGroupDialog (protGrpRes);
	    protGroupDialog -> OpenProtGroup (mainDialog, ProtGroupModify);
	    delete protGroupDialog;
	    liststr -> UnselectAll (selection);
	    break;
	}

	case DELPTGRP_MENUITEM:
	{
	    liststr = (vliststr *)(mainDialog -> prGrpListView -> GetList ());
	    selection = mainDialog -> prGrpListView -> GetSelection ();
	    int rowCount = liststr -> GetRowCount ();
	    if (rowCount == 0)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_EmptyListMsg);
		break;
	    }
	    if ((liststr -> IsSelectionEmpty (selection)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				   tisperm_NothingSelectedMsg);
		break;
	    }
	    else
	    {
		char protGrpName [PROTECTIONGRP_SZ + 1];
		liststr -> GetSelectedCell (selection, &row, &col);
		nameStr = (vchar *)(liststr -> GetCellValue (row, 0));
		vcharExportBounded (vcharScribeSystemExport (nameStr),
				    protGrpName,
				    sizeof (protGrpName));
		protGrpName [sizeof (protGrpName) - 1] = 0;
		// do not le them delete protection group untitled
		if ((strcmp (protGrpName, UNTITLED_DEFAULT)) == 0)
		{
		    tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE,
					    tisperm_NoDelUntitledErr,
					    nameStr);
		    liststr -> UnselectAll (selection);
		    break;
		}
		Tisprivtab *privtab = db -> GetPrivTab ();
		// can not delete a protection group if groups are assigned to it
		if ((privtab -> PrtGrpHasGroups (protGrpName)) == vTRUE)
		{
		    tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE,
					    tisperm_PtGrpGrpDeleteErr,
					    nameStr, nameStr, nameStr);
		    liststr -> UnselectAll (selection);
		    break;
		}

		Tisprotgroups *tisProtGroups = db -> GetTisProtGroups ();
		// can not delete a protection group if it has points assigned to it
		result = tisProtGroups -> ProtGrpHasPts (protGrpName);
		if (result == TISPERM_ERROR)
		{
		    liststr -> UnselectAll (selection);
		    break;
		}
		else if (result == vTRUE)
		{
		    tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE,
					    tisperm_PtGrpPtsDeleteErr,
					    nameStr, nameStr, nameStr);
		    liststr -> UnselectAll (selection);
		    break;
		}

		result = tisPermMsgCenter (mainDialog, TISPERM_CAUTION_NOTICE,
					    tisperm_DeletePtGrpReqMsg, nameStr);
		if (result == vTRUE)
		{
		    // clone nameStr, ModList, deletes entry from list and nameStr
		    // would no longer exist
		    strValue = vstrClone (nameStr);
		    mainDialog -> prGrpListView -> ModList (DelFromList, strValue, NULL);
		    todoStr = vresourceGetString (msgsRes, tisperm_DelStr);
		    tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				       tisperm_PtGrpDbUpdateMsg, strValue, todoStr);
		    vstrDestroy (strValue);
		    vstrDestroy (todoStr);
		}
		liststr -> UnselectAll (selection);
	    }
	    break;
	}

	case ASSIGNGRPS_MENUITEM:
	{
	    selection = mainDialog -> prGrpListView -> GetSelection ();
	    liststr = (vliststr *)(mainDialog -> prGrpListView -> GetList ());
	    int rowCount = liststr -> GetRowCount ();
	    if (rowCount == 0)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_EmptyListMsg);
		break;
	    }
	    if ((liststr -> IsSelectionEmpty (selection)) == vTRUE)
	    {
		tisPermMsgCenter (mainDialog, TISPERM_INFO_NOTICE, 
				  tisperm_NothingSelectedMsg);
		break;
	    }
	    liststr -> GetSelectedCell (selection, &row, &col);
	    nameStr = (vchar *)(liststr -> GetCellValue (row, 0));
	    vresource prGrpMaintRes = vresourceGet (tisPermRes, tisperm_PtGrpMaintDialog);
	    PrtGrpMaintDialog *prtGrpMaintDlg = new PrtGrpMaintDialog (prGrpMaintRes);
	    prtGrpMaintDlg -> OpenPrtGrpMaint (mainDialog, nameStr);
	    delete prtGrpMaintDlg;
	    liststr -> UnselectAll (selection);
	    break;
	}

	case FLDPROTECT_MENUITEM:
	{
	    vresource fldprtRes = vresourceGet (tisPermRes, tisperm_FieldPermsDialog);
	    FldPermsDialog *fldPermsDialog = new FldPermsDialog (fldprtRes);
	    fldPermsDialog -> OpenFldPerms (mainDialog);
	    delete fldPermsDialog;
	    break;
	}

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}


vkindTYPED_FULL_DEFINITION(GroupListView, vlistview, NULL);

/*******************************************************************
*
*    Group listview class
*
*******************************************************************/

void GroupListView::LoadInit (vresource res)
{
    list = NULL;

    vlistview::LoadInit (res);
}

void GroupListView::Destroy (void)
{
    if (list != NULL)
    {
	SetList (NULL);
	delete list;
    }
    vlistview::Destroy ( );
}


/*******************************************************************
*
*    Group listview class, build list function
*
*******************************************************************/

void GroupListView::BuildList (void)
{
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Tisgroups *tisGrps = db -> GetTisGroups ();
    if (list != NULL)
    {
	mainDialog -> groupListView -> SetList (NULL);
	delete list;
	list = NULL;
    }

    list = tisGrps -> GetGroupsList ();
    if (list != NULL)
    {
	mainDialog -> groupListView -> SetList (list);
	vlistSelection *selected = mainDialog -> groupListView -> GetSelection ();
	list -> SetColumnCount (2);
	list -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	list -> SetExtendRowSelection (selected, vTRUE);
	const vrect *listRect = mainDialog -> groupListView -> GetListRect ();
	mainDialog -> groupListView -> SetColumnWidth (vlistviewALL_COLUMNS,
						       listRect -> GetWidth () / 2);
	int rowCount = list -> GetRowCount ();
	if (rowCount > 0)
	{
	    list -> Sort (vlistROW, 0, NULL);
	}
    }
}


/*******************************************************************
*
*    Group listview class, does a group exist in the list
*
*******************************************************************/

int GroupListView::GroupExists (vchar *name)
{
    int		    exists = vFALSE;
    vchar	    *value;
    int		    row;
    vlistIterator   iterator;
    vliststr	    *viewList = (vliststr *)(GetList ());

    iterator.Start (viewList);
    while (iterator.Next ())
    {
	row = iterator.GetRow ();
	value = (vchar *)(viewList -> GetCellValue (row, 0));
	if (vcharCompare (value, name) == 0)
	{
	    exists = vTRUE;
	    break;
	}
    }
    iterator.Finish ();
    return (exists);
}


/*******************************************************************
*
*    Group listview class, modify the list
*
*******************************************************************/

void GroupListView::ModList (ListViewFuncs todo, vchar *name, vchar *desc)
{
    vliststr	    *viewList = (vliststr *)(GetList ());
    vlistSelection  *lselection = GetSelection ();
    int		    row, col;
    DbFuncs	    dbFunc;

    switch (todo)
    {
	case AddToList:
	    row = viewList -> GetRowCount ();
	    viewList -> SetCellValueScribed (row, 0, vcharScribeSystem (name));
	    viewList -> SetCellValueScribed (row, 1, vcharScribeSystem (desc));
	    dbFunc = AddToDb;
	    break;

	case ModTheList:
	{
	    viewList -> GetSelectedCell (selection, &row, &col);
	    // a mod group selection was not chosen, probably tried
	    // to add an existing group and decided to update it 
	    // so we find it
	    if (row == vlistNO_CELLS_SELECTED)
	    {
		vlistIterator	iterator;
		vchar		*value;
		iterator.Start (list);
		while (iterator.Next ())
		{
		    row = iterator.GetRow ();
		    value = (vchar *)(viewList -> GetCellValue (row, 0));
		    if (vcharCompare (value, name) == 0)
		    {
			break;
		    }
		}
		iterator.Finish ();
	    }
	    viewList -> SetCellValueScribed (row, 0, vcharScribeSystem (name));
	    viewList -> SetCellValueScribed (row, 1, vcharScribeSystem (desc));
	    dbFunc = ModTheDb;
	    break;
	}

	case DelFromList:
	    viewList -> GetSelectedCell (lselection, &row, &col);
	    viewList -> RemoveRow (row);
	    dbFunc = DelFromDb;
	    break;
    }
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Tisgroups *tisGroups = db -> GetTisGroups ();
    tisGroups -> UpdateGroupsDb (dbFunc, name, desc);
    viewList -> Sort (vlistROW, 0, NULL);
}


/*******************************************************************
*
*    Group listview class, list observer
*
*******************************************************************/

void GroupListView::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	int clickCount = GetLastPointerClickCount ();
	if (clickCount == 2)
	{
	    TisPermApp *app = mainDialog -> GetPermApp ();
	    vresource tisPermRes = app -> GetPermRes ();
	    vresource grpRes = vresourceGet (tisPermRes, tisperm_GroupnameDialog);
	    GroupDialog *groupDialog = new GroupDialog (grpRes);
	    groupDialog -> OpenGroup (mainDialog, GroupModify);
	    delete groupDialog;
	    vlistSelection *lselection = GetSelection ();
	    vlist *viewList = GetList ();
	    viewList -> UnselectAll (lselection);
	}
    }
}


vkindTYPED_FULL_DEFINITION(ProtGroupListView, vlistview, NULL);

/*******************************************************************
*
*    Protection Group listview class
*
*******************************************************************/

void ProtGroupListView::LoadInit (vresource res)
{
    list = NULL;

    vlistview::LoadInit (res);
}

void ProtGroupListView::Destroy (void)
{
    if (list != NULL)
    {
	SetList (NULL);
	delete list;
    }
    vlistview::Destroy ( );
}


/*******************************************************************
*
*    Protection Group listview class, build list function
*
*******************************************************************/

void ProtGroupListView::BuildList (void)
{
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Tisprotgroups *tisProtGrps = db -> GetTisProtGroups ();
    if (list != NULL)
    {
	mainDialog -> prGrpListView -> SetList (NULL);
	delete list;
	list = NULL;
    }

    list = tisProtGrps -> GetProtGroupsList ();
    if (list != NULL)
    {
	mainDialog -> prGrpListView -> SetList (list);
	vlistSelection *selected = mainDialog -> prGrpListView -> GetSelection ();
	list -> SetColumnCount (2);
	list -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	list -> SetExtendRowSelection (selected, vTRUE);
	const vrect *listRect = mainDialog -> prGrpListView -> GetListRect ();
	mainDialog -> prGrpListView -> SetColumnWidth (vlistviewALL_COLUMNS,
						       listRect -> GetWidth () / 2);
	int rowCount = list -> GetRowCount ();
	if (rowCount > 0)
	{
	    list -> Sort (vlistROW, 0, NULL);
	}
    }
}


/*******************************************************************
*
*    Protection Group listview class, does a group exist in the list
*
*******************************************************************/

int ProtGroupListView::ProtGroupExists (vchar *name)
{
    int		    exists = vFALSE;
    vchar	    *value;
    int		    row;
    vlistIterator   iterator;
    vliststr	    *viewList = (vliststr *)(GetList ());

    iterator.Start (list);
    while (iterator.Next ())
    {
	row = iterator.GetRow ();
	value = (vchar *)(viewList -> GetCellValue (row, 0));
	if (vcharCompare (value, name) == 0)
	{
	    exists = vTRUE;
	    break;
	}
    }
    iterator.Finish ();
    return (exists);
}


/*******************************************************************
*
*    Protection Group listview class, modify the list
*
*******************************************************************/

void ProtGroupListView::ModList (ListViewFuncs todo, vchar *name, vchar *desc)
{
    vliststr	    *viewList = (vliststr *)(GetList ());
    vlistSelection  *lselection = GetSelection ();
    int		    row = 0;
    int		    col;
    DbFuncs	    dbFunc;

    switch (todo)
    {
	case AddToList:
	    row = viewList -> GetRowCount ();
	    viewList -> SetCellValueScribed (row, 0, vcharScribeSystem (name));
	    viewList -> SetCellValueScribed (row, 1, vcharScribeSystem (desc));
	    dbFunc = AddToDb;
	    break;

	case ModTheList:
	{
	    viewList -> GetSelectedCell (lselection, &row, &col);
	    viewList -> SetCellValueScribed (row, 0, vcharScribeSystem (name));
	    viewList -> SetCellValueScribed (row, 1, vcharScribeSystem (desc));
	    dbFunc = ModTheDb;
	    break;
	}

	case DelFromList:
	    viewList -> GetSelectedCell (lselection, &row, &col);
	    viewList -> RemoveRow (row);
	    dbFunc = DelFromDb;
	    break;
    }
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Tisprotgroups *tisProtGroups = db -> GetTisProtGroups ();
    tisProtGroups -> UpdateProtGroupsDb (dbFunc, name, desc);
    viewList -> Sort (vlistROW, 0, NULL);
}


/*******************************************************************
*
*    Protection Group listview class, list observer
*
*******************************************************************/

void ProtGroupListView::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	int clickCount = GetLastPointerClickCount ();
	if (clickCount == 2)
	{
	    TisPermApp *app = mainDialog -> GetPermApp ();
	    vresource tisPermRes = app -> GetPermRes ();
	    vresource protGrpRes = vresourceGet (tisPermRes, tisperm_PtGrpDialog);
	    ProtGroupDialog *protGroupDialog = new ProtGroupDialog (protGrpRes);
	    protGroupDialog -> OpenProtGroup (mainDialog, ProtGroupModify);
	    delete protGroupDialog;
	    vlistSelection *lselection = GetSelection ();
	    vlist *viewList = GetList ();
	    viewList -> UnselectAll (lselection);
	}
    }
}


vkindTYPED_FULL_DEFINITION(UserListView, vlistview, NULL);

int CompareUserListValues (const void *frst,const void *scnd)
{
    TISUSER *first = (TISUSER *)frst;
    TISUSER *second = (TISUSER *)scnd;
    vstr *firstUser = vstrCloneScribed (vcharScribeSystem (first -> username));
    vstr *secondUser = vstrCloneScribed (vcharScribeSystem (second -> username));

    int ret = vcharCompareBase (firstUser, secondUser);
    vstrDestroy (firstUser);
    vstrDestroy (secondUser);
    return (ret);
}  

/*******************************************************************
*
*    User listview class
*
*******************************************************************/

void UserListView::LoadInit (vresource res)
{
    list = NULL;

    vlistview::LoadInit (res);
}

void UserListView::Destroy (void)
{
    if (list != NULL)
    {
	SetList (NULL);
	TisPermApp *app = mainDialog -> GetPermApp ();
	TisPermDatabase *db = app -> GetDb ();
	Tisusers *tisUsers = db -> GetTisUsers ();
	tisUsers -> DeleteUsersList (list);
    }
    vlistview::Destroy ( );
}


/*******************************************************************
*
*    User listview class, build list function
*
*******************************************************************/

void UserListView::BuildList (void)
{
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Tisusers *tisUsers = db -> GetTisUsers ();
    if (list != NULL)
    {
	mainDialog -> userListView -> SetList (NULL);
	delete list;
	list = NULL;
    }

    list = tisUsers -> GetUsersList ();
    if (list != NULL)
    {
	mainDialog -> userListView -> SetList (list);
	vlistSelection *selected = mainDialog -> userListView -> GetSelection ();
	list -> SetColumnCount (2);
	list -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	list -> SetExtendRowSelection (selected, vTRUE);
	const vrect *listRect = mainDialog -> userListView -> GetListRect ();
	mainDialog -> userListView -> SetColumnWidth (vlistviewALL_COLUMNS,
						      listRect -> GetWidth () / 2);
	int rowCount = list -> GetRowCount ();
	if (rowCount > 0)
	{
	    list -> Sort (vlistROW, 0, CompareUserListValues);
	}
    }
}


/*******************************************************************
*
*    User listview class, override of DrawCellContents function
*
*******************************************************************/

void UserListView::DrawCellContents (int row, int col, int selected, vrect *cellRect)
{
    TISUSER	*tisUser;
    vlist	*viewList = (vlist *)GetList ();

    vdraw::GSave ();
    vdraw::RectsClip (cellRect, 1);
    tisUser = (TISUSER *)(viewList -> GetCellValue (row, col));
    if (col == 0)
    {
	vdraw::ScribedRectShowCenterY(vcharScribeSystem (tisUser -> username), cellRect);
    }
    else
    {
	if ((strlen (tisUser -> descfld)) == 0)
	{
	    vdraw::ScribedRectShowCenterY(vcharScribeLiteral (" "), cellRect);
	}
	else
	{
	    vdraw::ScribedRectShowCenterY(vcharScribeSystem (tisUser -> descfld), cellRect);
	}
    }
    vdraw::GRestore();
}

/*******************************************************************
*
*    User listview class, modify the list
*
*******************************************************************/

void UserListView::ModList (ListViewFuncs todo, vchar *name, vchar *desc,
			    vchar *encrypasswd, int allowLogout, int logoutTime)
{
    vlist	    *viewList = GetList ();
    vlistSelection  *lselection = GetSelection ();
    int		    row, col;
    DbFuncs	    dbFunc;
    TISUSER	    *tisUser;

    switch (todo)
    {
	case AddToList:
	    tisUser = (TISUSER *)(vmemAllocAndClear (sizeof (TISUSER)));
	    vcharExportBounded (vcharScribeSystemExport (name),
				tisUser -> username,
				OPERATOR_SZ);
	    tisUser -> username [OPERATOR_SZ] = 0;
	    vcharExportBounded (vcharScribeSystemExport (desc),
				tisUser -> descfld,
				DESCRIPTION_SZ);
	    tisUser -> descfld [DESCRIPTION_SZ] = 0;
	    vcharExportBounded (vcharScribeSystemExport (encrypasswd),
				tisUser -> passwdfld,
				ENCRYPTPASSWD_SZ);
	    tisUser -> passwdfld [ENCRYPTPASSWD_SZ] = 0;
	    tisUser -> allowAutologout = allowLogout;
	    tisUser -> autoLogoutTime = logoutTime;
	    row = viewList -> GetRowCount ();
	    viewList -> SetCellValue (row, 0, (const void *)tisUser);
	    viewList -> SetCellValue (row, 1, (const void *)tisUser);
	    dbFunc = AddToDb;
	    break;

	case ModTheList:
	{
	    viewList -> GetSelectedCell (lselection, &row, &col);
	    // a mod group selection was not chosen, probably tried
	    // to add an existing group and decided to update it 
	    // so we find it
	    if (row == vlistNO_CELLS_SELECTED)
	    {
		vlistIterator	iterator;
		TISUSER		*value;
		int		foundIt = vFALSE;
		vstr		*tisuser;
		iterator.Start (viewList);
		while (iterator.Next ())
		{
		    row = iterator.GetRow ();
		    value = (TISUSER *)(viewList -> GetCellValue (row, 0));
		    tisuser = vstrCloneScribed (vcharScribeSystem (value -> username));
		    if (vcharCompare (tisuser, name) == 0)
		    {
			foundIt = vTRUE;;
		    }
		    vstrDestroy (tisuser);
		    if (foundIt == vTRUE)
		    {
			break;
		    }
		}
		iterator.Finish ();
	    }
	    tisUser = (TISUSER *)(viewList -> GetCellValue (row, 0));
	    vcharExportBounded (vcharScribeSystemExport (name),
				tisUser -> username,
				OPERATOR_SZ);
	    tisUser -> username [OPERATOR_SZ] = 0;
	    vcharExportBounded (vcharScribeSystemExport (desc),
				tisUser -> descfld,
				DESCRIPTION_SZ);
	    tisUser -> descfld [DESCRIPTION_SZ] = 0;
	    vcharExportBounded (vcharScribeSystemExport (encrypasswd),
				tisUser -> passwdfld,
				ENCRYPTPASSWD_SZ);
	    tisUser -> passwdfld [ENCRYPTPASSWD_SZ] = 0;
	    tisUser -> allowAutologout = allowLogout;
	    tisUser -> autoLogoutTime = logoutTime;
	    viewList -> SetCellValue (row, 0, (const void *)tisUser);
	    viewList -> SetCellValue (row, 1, (const void *)tisUser);
	    dbFunc = ModTheDb;
	    break;
	}

	case DelFromList:
	    viewList -> GetSelectedCell (lselection, &row, &col);
	    tisUser = (TISUSER *)(viewList -> GetCellValue (row, 0));
	    vmemFree ((void *)tisUser);
	    list -> RemoveRow (row);
	    dbFunc = DelFromDb;
	    break;
    }
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Tisusers *tisUsers = db -> GetTisUsers ();
    tisUsers -> UpdateUsersDb (dbFunc, name, desc, encrypasswd,
			       allowLogout, logoutTime);
    viewList -> Sort (vlistROW, 0, CompareUserListValues);
}


/*******************************************************************
*
*    User listview class, list observer
*
*******************************************************************/

void UserListView::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	int clickCount = GetLastPointerClickCount ();
	if (clickCount == 2)
	{
	    TisPermApp *app = mainDialog -> GetPermApp ();
	    vresource tisPermRes = app -> GetPermRes ();
	    vresource userRes = vresourceGet (tisPermRes, tisperm_UsernameDialog);
	    UserDialog *userDialog = new UserDialog (userRes);
	    userDialog -> OpenUser (mainDialog, UserModify);
	    delete userDialog;
	    vlistSelection *lselection = GetSelection ();
	    vlist *viewList = GetList ();
	    viewList -> UnselectAll (lselection);
	}
    }
}


/*******************************************************************
*
*    User listview class, does a user exist in the list
*
*******************************************************************/

int UserListView::UserExists (vchar *name)
{
    int		    exists = vFALSE;
    TISUSER	    *value;
    vstr	    *valueName;
    int		    row;
    vlistIterator   iterator;
    vliststr	    *viewList = (vliststr *)(GetList ());

    iterator.Start (list);
    while (iterator.Next ())
    {
	row = iterator.GetRow ();
	value = (TISUSER *)(viewList -> GetCellValue (row, 0));
	valueName = vstrCloneScribed (vcharScribeSystem (value -> username));
	if (vcharCompare (valueName, name) == 0)
	{
	    exists = vTRUE;
	}
	vstrDestroy (valueName);
	if (exists == vTRUE)
	{
	    break;
	}
    }
    iterator.Finish ();
    return (exists);
}

