#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vcharHEADER
#include vnameHEADER
#include vmemHEADER

extern "C"
{
#include "dbDefs.h"
#include "resolveLinks.h"
}
#include "tisperm.h"
#include "tisPermApp.hxx"
#include "tisPermDatabase.hxx"
#include "tisPermOptions.hxx"
#include "tisPermUtils.hxx"

/*******************************************************************
*
*    class OptionTree constructor
*
*******************************************************************/

OptionTree::OptionTree (TisPermApp *permApp)
{
    app = permApp;
    treeRoot = NULL;
    selectedList = NULL;
}

/*******************************************************************
*
*    class OptionTree destructor
*
*******************************************************************/

OptionTree::~OptionTree (void)
{
    if (treeRoot != NULL)
    {
	DeleteFwdLink (treeRoot);
    }
    if (selectedList != NULL)
    {
	delete selectedList;
    }
}


/*******************************************************************
*
*    class OptionTree clear the selected flag function
*
*******************************************************************/

void OptionTree::ClearSelected (TREENODE_PTR node)
{
    if (node != NULL)
    {
	node -> selected = vFALSE;
	ClearSelected (node -> sublink);
	ClearSelected (node -> fwdlink);
    }
}


/*******************************************************************
*
*    class OptionTree clear selected flag in all children
*
*******************************************************************/

void OptionTree::ClearChildrenSelected (TREENODE_PTR node, int *changed)
{
    if (node != NULL)
    {
	if (node -> selected == vTRUE)
	{
	    node -> selected = vFALSE;
	    *changed = vTRUE;
	}
	ClearChildrenSelected (node -> sublink, changed);
	ClearChildrenSelected (node -> fwdlink, changed);
    }
}


/*******************************************************************
*
*    class OptionTree clear the selected flag in parents
*
*******************************************************************/

void OptionTree::ClearParentSelected (TREENODE_PTR node, int *changed)
{
    if (node != NULL)
    {
	TREENODE_PTR	parent = node -> parent;
	while (parent != NULL)
	{
	    if (parent -> selected == vTRUE)
	    {
		parent -> selected = vFALSE;
		*changed = vTRUE;
	    }
	    parent = parent -> parent;
	}
    }
}


/*******************************************************************
*
*    class OptionTree Delete Tree function
*
*******************************************************************/

void OptionTree::DeleteFwdLink (TREENODE_PTR node)
{
    TREENODE_PTR    currNode = node;
    TREENODE_PTR    prevNode;

    while (currNode)
    {
	prevNode = currNode;
	if (currNode -> sublink != NULL)
	{
	    DeleteSubLink (currNode -> sublink);
	}
	currNode = currNode -> fwdlink;
	vmemFree ((void *)prevNode);
    }
}


/*******************************************************************
*
*    class OptionTree Delete Tree function
*
*******************************************************************/

void OptionTree::DeleteSubLink (TREENODE_PTR node)
{
    TREENODE_PTR    currNode = node;
    TREENODE_PTR    prevNode;

    while (currNode)
    {
	prevNode = currNode;
	if (currNode -> fwdlink != NULL)
	{
	    DeleteFwdLink (currNode -> fwdlink);
	}
	currNode = currNode -> sublink;
	vmemFree ((void *)prevNode);
    }
}


/*******************************************************************
*
*    OptionTree find a node, look for options at same level (fwdlink)
*	check each fwdlink for options at lower levels (sublink)
*
*******************************************************************/
TREENODE_PTR OptionTree::FindNode (TREENODE_PTR node, char *optionName)
{
    TREENODE_PTR p;
    TREENODE_PTR res = NULL;

    for (p = node; p != NULL; p = p -> fwdlink)
    {
	if ((strncmp (optionName, p -> optionName, OPTIONNAME_SZ))== 0)
	{
	    res = p;
	    break;
	}
	if (p -> sublink)
	{
	    res = FindInSubLink (p-> sublink, optionName);
	    if (res != NULL)
	    {
		break;
	    }
	}
    }

    return (res);
}


/*******************************************************************
*
*    OptionTree find a node in a sublink (option of lower level)
*	check options at same level (fwdlink)
*
*******************************************************************/
TREENODE_PTR OptionTree::FindInSubLink (TREENODE_PTR node, char *optionName)
{
    TREENODE_PTR p;
    TREENODE_PTR res = NULL;

    for (p = node; p != NULL; p = p -> sublink)
    {
	if ((strncmp (optionName, p -> optionName, OPTIONNAME_SZ)) == 0)
	{
	    res = p;
	    break;
	}
	if (p -> fwdlink)
	{
	    res = FindNode (p -> fwdlink, optionName);
	    if (res != NULL)
	    {
		break;
	    }
	}
    }

    return (res);
}



/*******************************************************************
*
*    class OptionTree load tree function
*    File has form:
*	MMI_DATAVU_FILE_OPEN
*	LEVEL	OPTION
*	  1	   MMI
*	  2	   DATAVU
*	  3        FILE
*         4        OPEN
*         4        EXIT
*         3        EDIT
*         2        TAGVU
*   Level 2 = tis utility
*   Level 3s are utility menu options
*   Level 4s are menu option items
*   Level 5 would be a menu itme option that has an attached menu
*
*******************************************************************/

int OptionTree::LoadTree (void)
{
    FILE	    *fd;
    char	    unresolvedPath [TISPERM_MAX_DATABASE_PATH];
    char	    filepath [TISPERM_MAX_DATABASE_PATH];
    int		    level;
    char	    option [UTILACCPRIV_SZ + 1];    // read from file
    char	    privStr [UTILACCPRIV_SZ + 1];   // copy here, strtok corrupts the str
    char	    workStr [UTILACCPRIV_SZ + 1];   // build str from tokens
    char	    *lastUnderscore;
    char	    *token;
    int		    firstRead = vTRUE;
    int		    firstToken;
    TREENODE_PTR    lastLevels [10];
    TREENODE_PTR    newNode;

    for (int i = 0; i < 10; i++)
    {
	lastLevels[i] = NULL;
    }

    sprintf (unresolvedPath, "%s%s", TISPERM_TASKS_PATH, OPTIONFILE_NAME);
    resolveLinks (unresolvedPath, filepath);

    if ((fd = fopen (filepath, "r")) == NULL)
    {
	tisPermMsgCenter (NULL, TISPERM_INFO_NOTICE, 
			    tisperm_OpenErr, filepath);
	return (TISPERM_ERROR);
    }
    while ((fscanf (fd, "%s\n", option)) != EOF)
    {
	strncpy (privStr, option, sizeof (privStr));
	privStr [sizeof (privStr) - 1] = 0;
	token = strtok (privStr, OPTION_DELIMETER_STR);
	memset (workStr, 0, sizeof (workStr));
	firstToken = vTRUE;
	while (token != NULL)
	{
	    if (firstToken == vFALSE)
	    {
		strcat (workStr, OPTION_DELIMETER_STR);
	    }
	    else
	    {
		firstToken = vFALSE;
	    }
	    strcat (workStr, token);
	    level = ReturnLevel (workStr);
	    // if first read, token will be MMI, we need this since it will be
	    // the head of our privielege tree, treeRoot
	    if (firstRead == vTRUE)
	    {
		if ( (level != 0) || 
		     ((strcmp (workStr, TOP_LEVEL_UTIL_PRIVILEGE)) != 0))
		{
		    tisPermMsgCenter (NULL, TISPERM_INFO_NOTICE,
				      tisperm_OptionsFormatErr, filepath);
		    return (TISPERM_ERROR);
		}
		newNode = (TREENODE_PTR)(vmemAllocAndClear (sizeof (TREENODE)));
		treeRoot = newNode;
		treeRoot -> level = level;
		strncpy (treeRoot -> dbOptionName, workStr, 
					    sizeof (treeRoot -> dbOptionName));
		treeRoot -> dbOptionName [sizeof (treeRoot -> dbOptionName) -1] = 0;
		strncpy (treeRoot -> optionName, token, 
					    sizeof (treeRoot -> optionName));
		treeRoot -> optionName [sizeof (treeRoot -> optionName) - 1] = 0;
		treeRoot -> boxType = TISPERM_GROUP_BOX;
		treeRoot -> imageType = TISPERM_MAIN_PARENT;
		lastLevels [0] = treeRoot;
		firstRead = vFALSE;
	    }
	    else
	    {
		// we have already done MMI, no need to even look at it again
		if ((strcmp (workStr, TOP_LEVEL_UTIL_PRIVILEGE)) == 0)
		{
		    token = strtok (NULL, OPTION_DELIMETER_STR);
		    continue;
		}
		lastUnderscore = strrchr (workStr, OPTION_DELIMETER_CHAR);
		if (lastUnderscore == NULL)
		{
		    lastUnderscore = workStr;
		}
		else
		{
		    lastUnderscore++;
		}
		// have we already stored this option, start with parent
		if ((FindNode (lastLevels [level - 1], lastUnderscore)) == NULL)
		{
		    // new node
		    newNode = (TREENODE_PTR)(vmemAllocAndClear (sizeof (TREENODE)));
		    // levels from 0... as are indicies into lastLevel[]
		    newNode -> parent = lastLevels [level - 1];
		    strncpy (newNode -> dbOptionName, workStr, 
						    sizeof (newNode -> dbOptionName));
		    newNode -> dbOptionName [sizeof (newNode -> dbOptionName) -1] = 0;
		    strncpy (newNode -> optionName, lastUnderscore, 
						    sizeof (newNode -> optionName));
		    newNode -> optionName [sizeof (newNode -> optionName) - 1] = 0;
		    newNode -> level = level;
		    newNode -> boxType = TISPERM_NO_GROUP_BOX;
		    newNode -> imageType = TISPERM_LAST_CHILD;
		    // does the parent level have any options yet
		    if (lastLevels [level - 1] -> sublink == NULL)
		    {
			// root has it's own special box type
			if ((IsRoot (lastLevels [level - 1])) == vFALSE)
			{
			    lastLevels [level - 1] -> boxType = TISPERM_GROUP_BOX;
			    // does parent have any siblings, options at same level
			    if (lastLevels [level - 1] -> fwdlink != NULL)
			    {
				lastLevels [level - 1] -> imageType = TISPERM_PARENT;
			    }
			    else
			    {
				lastLevels [level - 1] -> imageType = TISPERM_LAST_PARENT;
			    }
			}
			lastLevels [level - 1] -> sublink = newNode;
			lastLevels [level] = newNode;
		    }
		    else
		    {
			// parent level has options (sublink)
			TREENODE_PTR nextChild = lastLevels [level - 1] -> sublink -> fwdlink;
			TREENODE_PTR lastChild = lastLevels [level - 1] -> sublink;
			// follow fwdlink's to get to last option
			while (nextChild)
			{
			    lastChild = nextChild;
			    nextChild = nextChild -> fwdlink;
			}
			lastChild -> imageType = TISPERM_CHILD;
			// add new option
			lastChild -> fwdlink = newNode;
			if ((IsRoot (lastChild)) == vFALSE)
			{
			    if (lastChild -> sublink != NULL)
			    {
				lastChild -> imageType = TISPERM_PARENT;
			    }
			}
			lastLevels [level] = newNode;
		    }
		}
	    }

	    token = strtok (NULL, OPTION_DELIMETER_STR);

	} // end while (token != NULL)

    } // end while (fgets)
    fclose (fd);
    return (TISPERM_SUCCESS);
}


/*******************************************************************
*
*    class OptionTree find out if a parent is selected
*
*******************************************************************/

int OptionTree::ParentIsSelected (TREENODE_PTR node)
{
    int	retval = vFALSE;

    if (node == treeRoot)
    {
	if (node -> selected == vTRUE)
	{
	    retval = vTRUE;
	}
    }
    else if (node != NULL)
    {
	TREENODE_PTR	parent = node -> parent;
	while (parent != NULL)
	{
	    if (parent -> selected == vTRUE)
	    {
		retval = vTRUE;
	    }
	    parent = parent -> parent;
	}
    }
    return (retval);
}

/*******************************************************************
*
*    OptionTree find the level if input string
*
*******************************************************************/
int OptionTree::ReturnLevel (char *optionName)
{
    int level = 0;
    char *ptr = optionName;
    while (*ptr)
    {
	if ((*ptr) == '_')
	{
	    level++;
	}
	ptr++;
    }
    return (level);
}


/*******************************************************************
*
*    class OptionTree print tree function
*
*******************************************************************/

void OptionTree::printTree (TREENODE_PTR node)
{
    if (node != NULL)
    {
	printf ("level = %d, option = %s, db = %s, node = %0x, par = %0x, sub=%0x, fwd=%0x\n", 
	    node->level,node->optionName,node->dbOptionName, 
	    node,node->parent,node->sublink,node->fwdlink);
	// write menu item options
	printTree (node -> sublink);
	// write utility menu options
	printTree (node -> fwdlink);
    }
}


/*******************************************************************
*
*    class OptionTree set selected flag in all children
*
*******************************************************************/

void OptionTree::SetChildrenSelected (TREENODE_PTR node, int *changed)
{
    if (node != NULL)
    {
	if (node -> selected == vFALSE)
	{
	    node -> selected = vTRUE;
	    *changed = vTRUE;
	}
	SetChildrenSelected (node -> sublink, changed);
	SetChildrenSelected (node -> fwdlink, changed);
    }
}


/*******************************************************************
*
*   class OptionTree
*   Set the children etc. selected if the parent is selected
*
*******************************************************************/

void OptionTree::SetParentsChildrenSelected (TREENODE_PTR node)
{
    if (node != NULL)
    {
	if (node -> parent != NULL)
	{
	    if (node -> parent -> selected == vTRUE)
	    {
		node -> selected = vTRUE;
	    }
	}
	SetParentsChildrenSelected (node -> sublink);
	SetParentsChildrenSelected (node -> fwdlink);
    }
}


/*******************************************************************
*
*   class OptionTree
*   1 - set a parent of node selected if all of parent's children 
*	are selected
*   2 - check up line because if we select parent of node, then 
*	if all of parents siblings are selected then select
*	parents parent etc.
*
*******************************************************************/

void OptionTree::SetParentSelected (TREENODE_PTR node, int *changed)
{
    TREENODE_PTR    parent = node -> parent;
    TREENODE_PTR    child;
    int		    allSelected;

    // parent can be NULL if node is root of tree
    while (parent != NULL) 
    {
	allSelected = vTRUE;
	child = parent -> sublink;
	while (child != NULL)
	{
	    if (child -> selected == vFALSE)
	    {
		allSelected = vFALSE;
		break;
	    }
	    child = child -> fwdlink;
	}
	if (allSelected == vTRUE)
	{
	    parent -> selected = vTRUE;
	    *changed = vTRUE;
	}
	else
	{
	    break;
	}
	parent = parent -> parent;
    }
}


/*******************************************************************
*
*    class OptionTree write tree function
*
*******************************************************************/

void OptionTree::WriteTree (TREENODE_PTR node, char *groupName)
{
    if (node -> selected == vTRUE)
    {
	int select = vFALSE;
	if (node == treeRoot)
	{
	    select = vTRUE;
	}
	else if ((ParentIsSelected (node)) == vFALSE)
	{
	    select = vTRUE;
	}
	if (select == vTRUE)
	{
	    if (selectedList == NULL)
	    {
		selectedList = new vliststr;
		rowCount = 0;
	    }
	    selectedList -> SetCellValueScribed (rowCount, 0, 
					vcharScribeSystem (node -> dbOptionName));
	    rowCount++;
	}
    }
    WriteTreeTraverse (node -> sublink);
    if (selectedList != NULL)
    {
	TisPermDatabase *db = app -> GetDb ();
	Tisutilacc *utilAcc = db -> GetTisUtilAcc ();
	utilAcc -> UpdateUtilAccPrivileges (selectedList, groupName);
    }
}


/*******************************************************************
*
*    class OptionTree write tree function
*
*******************************************************************/

void OptionTree::WriteTreeTraverse (TREENODE_PTR node)
{
    if (node != NULL)
    {
	if (node -> selected == vTRUE)
	{
	    if ((ParentIsSelected (node)) == vFALSE)
	    {
		if (selectedList == NULL)
		{
		    selectedList = new vliststr;
		    rowCount = 0;
		}
		selectedList -> SetCellValueScribed (rowCount, 0, 
					vcharScribeSystem (node -> dbOptionName));
		rowCount++;
	    }
	}
	// write menu item options
	WriteTreeTraverse (node -> sublink);
	// write utility menu options
	WriteTreeTraverse (node -> fwdlink);
    }
}
