#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstrHEADER
#include vlistviewHEADER
#include vrulerHEADER

extern "C"
{
#include "tisEncrypt.h"
}

#include "tisperm.h"
#include "tisPermApp.hxx"
#include "tisPermDatabase.hxx"
#include "tisPermMain.hxx"
#include "tisPermUtils.hxx"
#include "tisPermUser.hxx"

static int userNameFilter (filteredText *text, vtextSelection *sel,
			   const vchar **s, long *len);
static int descriptionFilter (filteredText *text, vtextSelection *sel,
			      const vchar **s, long *len);
static int passwordFilter (passwordText *text, vtextSelection *sel,
			   const vchar **s, long *len);

vkindTYPED_FULL_DEFINITION(UserDialog, vdialog, NULL);

/*******************************************************************
*
*    UserDialog override for constructors init ()
*
*******************************************************************/

void UserDialog::LoadInit (vresource res)
{
    modifiedFlag = vFALSE;
    userName = filteredTextitem::RequestSubLoad (this,
					tisperm_Username);

    description = filteredTextitem::RequestSubLoad (this,
					tisperm_Description);

    password = passwordTextitem::RequestSubLoad (this,
					tisperm_Password);

    verify = passwordTextitem::RequestSubLoad (this,
					tisperm_Verify);

    allowLogout = UserCheckBox::RequestSubLoad (this,
					tisperm_AllowAutologout);

    logoutTime = spinnerClass::RequestSubLoad (this,
					tisperm_Autologout);

    memberOfGroup = vtextitem::RequestSubLoad (this,
					tisperm_GroupMember);
   
    okButton = UserOkButton::RequestSubLoad (this,
					tisperm_OkButton);
   
    cancelButton = UserCancelButton::RequestSubLoad (this,
					tisperm_CancelButton);
    vdialog::LoadInit (res);

    // give these ptr to user dialog
    allowLogout -> SetUserPtr ((void *)this);
    okButton     -> SetUserDialog (this);
    cancelButton -> SetUserDialog (this);

    // set some filters
    userName    -> SetItemFilterFunc (userNameFilter);
    userName    -> SetTextModifiedPtr (&this -> modifiedFlag);
    description -> SetItemFilterFunc (descriptionFilter);
    description -> SetTextModifiedPtr (&this -> modifiedFlag);
    password    -> SetTextModifiedPtr (&this -> modifiedFlag);
    verify      -> SetTextModifiedPtr (&this -> modifiedFlag);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    UserDialog Destroy function
*
*******************************************************************/

void UserDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }

    vdialog::Destroy ();
}



/*******************************************************************
*
*    UserDialog notify function
*
*******************************************************************/

void UserDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    UserDialog Open function
*
*******************************************************************/

void UserDialog::OpenUser (MainDialog *mnDlg, UserFunc usrFunc)
{
    mainDialog = mnDlg;
    userFunc = usrFunc;

    if (userFunc == UserModify)
    {
	UpdateFields ();
	userName -> Disable ();
    }
    else if (userFunc == UserAdd)
    {
	logoutTime -> Disable ();
	memberOfGroup -> Disable ();
    }
    SetDialogTitle ();
    // reset here because UpdateTextFields will set it to true
    modifiedFlag = vFALSE;
    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
//    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    UserDialog Set dialog title function, depends on what
*	function is to be done
*
*******************************************************************/

void UserDialog::SetDialogTitle (void)
{
    vstr	*dialogTitle;
    TisPermApp	*tisPerm = mainDialog -> GetPermApp ();
    vresource	titlesRes = tisPerm -> GetTitlesRes ();

    switch (userFunc)
    {
	case UserAdd:
	    dialogTitle = vresourceGetString (titlesRes, tisperm_NewUser);
	    break;

	case UserModify:
	    dialogTitle = vresourceGetString (titlesRes, tisperm_ModUser);
	    break;
    }

    SetTitle (dialogTitle);
    vstrDestroy (dialogTitle);
}


/*******************************************************************
*
*    UserDialog 
*	update username, description and password fields,
*	and logout fields
*
*******************************************************************/

void UserDialog::UpdateFields (void)
{
    int		    row, col;
    char	    groupName [GROUPNAME_SZ + 1];
    vlistSelection  *selection;
    vlist	    *userList;
    TISUSER	    *tisUser;
    vstr	    *defPassword;
    TisPermApp	    *tisPerm = mainDialog -> GetPermApp ();
    vresource	    titlesRes = tisPerm -> GetTitlesRes ();

    selection = mainDialog -> userListView -> GetSelection ();
    userList  = mainDialog -> userListView -> GetList ();
    userList -> GetSelectedCell (selection, &row, &col);
    tisUser = (TISUSER *)(userList -> GetCellValue (row, 0));
    userName -> SetTextScribed (vcharScribeSystem (tisUser -> username));
    description -> SetTextScribed (vcharScribeSystem (tisUser -> descfld));
    defPassword = vresourceGetString (titlesRes, tisperm_DefaultPassword);
    password -> SetText (defPassword);
    verify -> SetText (defPassword);
    allowLogout -> SetValue (tisUser -> allowAutologout);
    if (tisUser -> allowAutologout == 0)
    {
	logoutTime -> Disable ();
    }
    else
    {
	logoutTime -> SetValue (tisUser -> autoLogoutTime);
    }
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Sysusers *sysUsers = db -> GetSysUsers ();
    if ((sysUsers -> UserInGroup (tisUser -> username, groupName)) == vTRUE)
    {
	memberOfGroup -> SetTextScribed (vcharScribeSystem (groupName));
    }
    else
    {
	memberOfGroup -> Disable ();
    }
    vstrDestroy (defPassword);
}


/*******************************************************************
*
*    UserDialog, verify password function
*    NOTE:
*	the char buffer to hold the encrypted password MUST be
*	2 * the max size + 1, to hold the encrypted version
*
*******************************************************************/

int UserDialog::VerifyPassword (vchar *encryptedPassword)
{
    char	    passwd [ENCRYPTPASSWD_SZ + 1];
    char	    verifypasswd [ENCRYPTPASSWD_SZ + 1];
    int		    len;
    vstr	    *str;
    passwordText    *text;

    memset (passwd, 0, sizeof (passwd));
    memset (verifypasswd, 0, sizeof (verifypasswd));

    text = (passwordText *)(password -> GetTextData ());
    len = text -> realPassword -> GetLength ();
    // no need to go on, no password entered
    // this is true is a user was modified and a new password was
    // not entered
    if (len == 0)
    {
	return (vTRUE);
    }
    // use the unecrypted password to do the compare
    str = vstrCreateSized (len + 1);
    str [len] = vcharNULL;
    text -> realPassword -> GetText (NULL, str, len);

    vcharExportBounded (vcharScribeSystemExport (str),
			passwd,
			UNENCRYPTPASSWD_SZ);
    passwd [UNENCRYPTPASSWD_SZ] = 0;
    vstrDestroy (str);

    text = (passwordText *)(verify -> GetTextData ());
    len = text -> realPassword -> GetLength ();
    str = vstrCreateSized (len + 1);
    str [len] = vcharNULL;
    text -> realPassword -> GetText (NULL, str, len);

    vcharExportBounded (vcharScribeSystemExport (str),
			verifypasswd,
			UNENCRYPTPASSWD_SZ);
    verifypasswd [UNENCRYPTPASSWD_SZ] = 0;
    vstrDestroy (str);

    if ((strncmp (passwd, verifypasswd, UNENCRYPTPASSWD_SZ)) != 0)
    {
	tisPermMsgCenter (this, TISPERM_INFO_NOTICE, tisperm_PasswdConfirmErr);
	return (vFALSE);
    }

    // passwd is big enough to handle double the size
    // of the unencrypted passwword
    tisEncrypt (passwd, UNENCRYPTPASSWD_SZ);

    str = vstrCloneScribed (vcharScribeSystem (passwd));
    vcharCopy (str, encryptedPassword);
    vstrDestroy (str);
    return (vTRUE);
}


vkindTYPED_FULL_DEFINITION(UserOkButton, vbutton, NULL);

/*******************************************************************
*
*    UserOkButton observe function
*
*******************************************************************/

void UserOkButton::ObserveDialogItem (vevent *event)
{
    vstr	*nameStr = userDialog -> userName -> GetTextAsString ();
    vstr	*descStr = userDialog -> description -> GetTextAsString ();
    int		allowAutologout = userDialog -> allowLogout -> GetValue ();
    int		logoutTime = userDialog -> logoutTime -> GetValue ();
    UserFunc	func = userDialog -> GetUserFunction ();
    MainDialog	*mainDialog = userDialog -> GetMainDialog ();
    int		userExists =  mainDialog -> 
			    userListView -> UserExists (nameStr);
    int		todo = TISPERM_ERROR;
    vchar	encryptedPassword [ENCRYPTPASSWD_SZ + 1];

    if ((userDialog -> GetModifiedFlag ()) == vFALSE)
    {
	tisPermMsgCenter (userDialog, TISPERM_INFO_NOTICE, tisperm_NothingModMsg);
	vstrDestroy (nameStr);
	vstrDestroy (descStr);
	return;
    }

    // if no password, blanks to database
    vmemSet ((void *)encryptedPassword, 0, sizeof (encryptedPassword));

    if (vcharLength (nameStr) == 0)
    {
	tisPermMsgCenter (userDialog, TISPERM_INFO_NOTICE, tisperm_NoUserMsg);
	vstrDestroy (nameStr);
	vstrDestroy (descStr);
	return;
    }
    if (func == UserAdd)
    {
	if (userExists == vTRUE)
	{
	    tisPermMsgCenter (userDialog,TISPERM_INFO_NOTICE,
			       tisperm_UserAlreadyExistsMsg, nameStr);
	    vstrDestroy (nameStr);
	    vstrDestroy (descStr);
	    return;
	}
	// use the unecrypted password to do the compare
	else 
	{
	    passwordText *text;
	    text = (passwordText *)(userDialog -> password -> GetTextData ());
	    int len = text -> realPassword -> GetLength ();
	    if (len == 0)
	    {
#ifndef	IGNORE_PASSWD
		tisPermMsgCenter (userDialog,TISPERM_INFO_NOTICE,
							tisperm_NoPasswdErr);
		vstrDestroy (nameStr);
		vstrDestroy (descStr);
		return;
#endif	IGNORE_PASSWD
	    }
	    else 
	    {
		if  ((userDialog -> VerifyPassword (encryptedPassword)) == vFALSE)
		{
		    vstrDestroy (nameStr);
		    vstrDestroy (descStr);
		    return;
		}
	    }
	    mainDialog -> userListView -> ModList (AddToList, nameStr, 
						   descStr, encryptedPassword,
						   allowAutologout, logoutTime);
	    todo = AddToList;
	}
    }
    // modify existing user
    else if (func == UserModify)
    {
	if (userExists == vFALSE)
	{
	    tisPermMsgCenter (userDialog, TISPERM_INFO_NOTICE,
			       tisperm_UserDoesNotExistMsg, nameStr);
	    vstrDestroy (nameStr);
	    vstrDestroy (descStr);
	    return;
	}
	else
	{
	    if (userDialog -> VerifyPassword (encryptedPassword) == vFALSE)
	    {
		vstrDestroy (nameStr);
		vstrDestroy (descStr);
		return;
	    }
	    mainDialog -> userListView -> ModList (ModTheList, nameStr, 
						   descStr, encryptedPassword,
						   allowAutologout, logoutTime);
	    todo = ModTheList;
	}
    }

    vstr    *todoStr;
    TisPermApp *app = mainDialog -> GetPermApp ();
    vresource	msgsRes = app -> GetMessagesRes ();
    switch (todo)
    {
	case TISPERM_ERROR:
	    vstrDestroy (nameStr);
	    vstrDestroy (descStr);
	    return;

	case AddToList:
	    todoStr = vresourceGetString (msgsRes, tisperm_AddedStr);
	    break;

	case ModTheList:
	    todoStr = vresourceGetString (msgsRes, tisperm_ModStr);
	    break;
    }
    tisPermMsgCenter (userDialog, TISPERM_INFO_NOTICE, 
		       tisperm_UserDbUpdateMsg, nameStr, todoStr);
    vstrDestroy (nameStr);
    vstrDestroy (descStr);
    vstrDestroy (todoStr);
    vevent::StopProcessing ();
    return;
}


vkindTYPED_FULL_DEFINITION(UserCancelButton, vbutton, NULL);

/*******************************************************************
*
*    UserCancelButton observe function
*
*******************************************************************/

void UserCancelButton::ObserveDialogItem (vevent *event)
{
    vevent::StopProcessing ();
    return;
}


vkindTYPED_FULL_DEFINITION(UserCheckBox, checkBoxClass, NULL);

/*******************************************************************
*
*    UserCheckBox observe function
*	if use allow auto logout, then enable the logout time
*	spinner box
*
*******************************************************************/

void UserCheckBox::ObserveDialogItem (vevent *event)
{
    UserDialog *userDialog = (UserDialog *)(GetUserPtr ());
    int allowLogout = userDialog -> allowLogout -> GetValue ();

    if (allowLogout)
    {
	userDialog -> logoutTime -> Enable ();
    }
    else
    {
	userDialog -> logoutTime -> Disable ();
    }
}


/*******************************************************************
*
*    Password text class
*
*******************************************************************/
/*******************************************************************
*
*    Password textitem class
*
*******************************************************************/

// macro of vkindtyped overrides
vkindTYPED_FULL_DEFINITION(passwordTextitem, vtextitem, NULL);

void passwordTextitem::LoadInit (vresource resource)
{
    vtextitemText   *oldText;
    passwordText    *newText;

    vtextitem::LoadInit (resource);

    oldText = (vtextitemText *)GetTextData ();
    newText = new passwordText;
    SetTextData (newText);
    SetSelection (newText -> CreateSelection ());

    delete oldText;

    newText -> realPassword = new vtext;
    newText -> realSelection = newText -> realPassword -> CreateSelection ();
    newText -> firstChar = vTRUE;
}

void passwordTextitem::ObserveView (vevent *event, int message)
{
    if (message == vtextviewBUTTON_UP_NOTIFY)
    {
	passwordText *text = (passwordText *)(GetTextData ());
	vtextSelection *sel = GetSelection ();
	int textLen = text -> GetLength ();
	if (text -> firstChar == vTRUE)
	{
	    if (textLen > 0)
	    {
		MoveCursorToEnd (vtextviewLEFT);
		// select the entire string
		text -> SelectRange (sel, 0, UNENCRYPTPASSWD_SZ);
		text -> firstChar = vFALSE;
	    }
	}
    }
}


/*******************************************************************
*
*    Password vtextitemText class
*
*******************************************************************/

// macro of vkindtyped overrides
vkindTYPED_FULL_DEFINITION(passwordText, vtextitemText, NULL);

passwordText::passwordText(vtextitemText &from)
: vtextitemText(vsuppressDONT_INIT)
{
    vtextitemText::CopyInit(&from);
    modified = NULL;
}

void passwordText::Init()
{
    vtextitemText::Init();
    modified = NULL;
}

void passwordText::LoadInit(vresource resource)
{
    vtextitemText::LoadInit(resource);
    modified = NULL;
}

void passwordText::CopyInit(vloadable *orig)
{
    vtextitemText::CopyInit(orig);
    modified = ((filteredText *)orig) -> GetModifiedPtr ();
}

void passwordText::Copy(vloadable *orig)
{
    vtextitemText::Copy(orig);
    modified = ((filteredText *)orig) -> GetModifiedPtr ();
}

void passwordText::Destroy(void)
{
    delete this -> realPassword;
    vtextitemText::Destroy ();
}


/****************************************************
*
*    filteredText delete
*
****************************************************/

void passwordText::DeleteText (vtextSelection *selection)
{
    long    start, end, selectionLen;
    int	    ret = vFALSE;

    // force initial selection of all default "*"s
    start = GetSelectionStartPosition (selection);
    end   = GetSelectionEndPosition (selection);
    selectionLen = end - start;
    if ( (firstChar == vTRUE) && (selectionLen != UNENCRYPTPASSWD_SZ) )
    {
	vwindow::Beep ();
	return;
    }
	
    // delete the  text from our real password buffer
    long textLen = realPassword -> GetLength ();
    if (textLen > 0)
    {
	realPassword -> SelectRange (realSelection,
				     GetSelectionStart (selection),
				     GetSelectionEnd (selection));
	realPassword -> DeleteText (realSelection);
    }
    

    vtextitemText::DeleteText (selection);
}


/****************************************************
*
*    filteredText insert
*	pass the text through the filter function,
*	if there is a filter and if there is text
*	before giving it to the Text manager, if ok	
*
****************************************************/

void passwordText::InsertText (vtextSelection *selection, const vchar *s, long len)
{
    long nuLen = len;

    if ( (len == 0) ||
	 (len > 1)  ||
	 (passwordFilter(this, selection, &s, &nuLen)) ) 
    {
	vstr	*str, *scan;
	int	i, start;

	scan = str = vstrCreateSized (nuLen);
	i = nuLen;
	while (i-- > 0)
	{
	    *(scan++) = '*';
	}

	start = GetSelectionStart (selection);

	// send to parent class to display the "*"s
	vtextitemText::InsertText(selection, str, nuLen);

	if ( (nuLen > 0) && ((*s) != '*') )
	{
	    // insert text into our real password buffer
	    realPassword -> SelectRange (realSelection,start, start);
	    realPassword -> InsertText (realSelection, s, nuLen);
	}

	vstrDestroy (str);
    }
}


/****************************************************
*
*    passwordText ObserveDelete
*	if modified flag != NULL, set it to vTRUE
*	and call the base class func
*
****************************************************/

void passwordText::ObserveDelete(long pos, long delta)
{
    int *mod = GetModifiedPtr ();
    if (mod != NULL)
	SetModifiedFlag ();
    vtext::ObserveDelete (pos, delta);
}

/****************************************************
*
*    passwordText ObserveInsert
*	if modified flag != NULL, set it to vTRUE
*	and call the base class func
*
****************************************************/

void passwordText::ObserveInsert(long pos, long delta)
{
    int *mod = GetModifiedPtr ();
    if (mod != NULL)
	SetModifiedFlag ();
    vtext::ObserveInsert (pos, delta);
}


/*******************************************************************
*
*    Some filter functions,for the UserDialog
*	NCharFilter (see tisgalaxyLib/textFilters.cxx
*
*******************************************************************/

static int userNameFilter (filteredText *text, vtextSelection *sel,
			    const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, OPERATOR_SZ));
}

static int descriptionFilter (filteredText *text, vtextSelection *sel,
			      const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, DESCRIPTION_SZ));
}

static int passwordFilter (passwordText *text, vtextSelection *sel,
			      const vchar **s, long *len)
{
    long    start, end, selectionLen;
    long    textlen;
    int	    ret = vFALSE;

    // we need to do selection length calculations because the 
    // new text is added _before_ the old selection's text is 
    // nuked.

    start = text -> GetSelectionStartPosition (sel);
    end   = text -> GetSelectionEndPosition (sel);
    selectionLen = end - start;
    textlen = text -> GetLength ();

    if (text -> firstChar == vTRUE)
    {
	text -> firstChar = vFALSE;
    }


    *len = vMIN(UNENCRYPTPASSWD_SZ - (textlen - selectionLen), *len);

    if (*len > 0) 
    {
	return (vTRUE);
    } 
    else 
    {
	vwindow::Beep ();
	ret = vFALSE;
    }
    return (ret);
}
