#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vtextHEADER
#include vstdargHEADER
#include vprintHEADER

#include "tisperm.h"
#include "tisPermApp.hxx"
#include "tisPermDatabase.hxx"
#include "tisPermUtils.hxx"


/*******************************************************************
*
*   given a numeric field privilege level, return the 
*   string that corresponds to it
*
*******************************************************************/

void GetPrivLevelStr (int pLevel, char *protectionLevel)
{
    char *cLevel;
    if (pLevel < TraineePriv)
    {
	cLevel = VIEW_STR;
    }
    else if (pLevel < OperatorPriv)
    {
	cLevel = TRAINEE_STR;
    }
    else if (pLevel < SupervisorPriv)
    {
	cLevel = OPERATOR_STR;
    }
    else if (pLevel < EngineerPriv)
    {
	cLevel = SUPERVISOR_STR;
    }
    else if (pLevel < AdminPriv)
    {
	cLevel = ENGINEER_STR;
    }
    else
    {
	cLevel = ADMIN_STR;
    }
    strcpy (protectionLevel, cLevel);
}


/*******************************************************************
*
*   given a string field privilege level, return the 
*   numeric value that corresponds to it
*
*******************************************************************/

DefaultPrivilegeLevels GetPrivLevelValue (char *privLevelStr)
{
    if (((strlen (privLevelStr)) == 0) || (privLevelStr[0] == ' '))
    {
	return (ViewPriv);
    }
    if ((strcmp (privLevelStr, VIEW_STR)) == 0)
    {
	return (ViewPriv);
    }
    if ((strcmp (privLevelStr, TRAINEE_STR)) == 0)
    {
	return (TraineePriv);
    }
    if ((strcmp (privLevelStr, OPERATOR_STR)) == 0)
    {
	return (OperatorPriv);
    }
    if ((strcmp (privLevelStr, SUPERVISOR_STR)) == 0)
    {
	return (SupervisorPriv);
    }
    if ((strcmp (privLevelStr, ENGINEER_STR)) == 0)
    {
	return (EngineerPriv);
    }
    if ((strcmp (privLevelStr, ADMIN_STR)) == 0)
    {
	return (AdminPriv);
    }
    return (ViewPriv);
}


/*******************************************************************
*
*   given a numeric protection privilege level, return the 
*   string that corresponds to it
*
*******************************************************************/

void GetProtectionLevelStr (int pLevel, char *protectionLevel)
{
    char *cLevel;
    if (pLevel == NoAccess)
    {
	cLevel = NO_ACCESS_STR;
    }
    else if (pLevel == ViewOnly)
    {
	cLevel = VIEW_ONLY_STR;
    }
    else if (pLevel == FullAccess)
    {
	cLevel = FULL_ACCESS_STR;
    }
    else
    {
	cLevel = UNKNOWN_ACCESS_STR;
    }
    strcpy (protectionLevel, cLevel);
}


/*******************************************************************
*
*   given a string protection level, return the 
*   numeric value that corresponds to it
*
*******************************************************************/

DefaultProtectionLevels GetProtectionLevelValue (char *privLevelStr)
{
    if (((strlen (privLevelStr)) == 0) || (privLevelStr[0] == ' '))
    {
	return (NoAccess);
    }
    if ((strcmp (privLevelStr, NO_ACCESS_STR)) == 0)
    {
	return (NoAccess);
    }
    if ((strcmp (privLevelStr, VIEW_ONLY_STR)) == 0)
    {
	return (ViewOnly);
    }
    if ((strcmp (privLevelStr, FULL_ACCESS_STR)) == 0)
    {
	return (FullAccess);
    }
    return (NoAccess);
}

/*******************************************************************
*
*    tisPermMsgCenter ()
*
*******************************************************************/

int tisPermMsgCenter (vdialog *dialog, int msgType, const vname *msgName, ...)
{
    int		result = vTRUE;
    va_list	argptr;

    va_start (argptr, msgName);

    vapplication *tisPermApp   = vapplication::GetCurrent ();
    vresource    tisPermRes    = tisPermApp -> GetResources ();
    vresource    titleRes   = vresourceGet (tisPermRes, tisperm_TitlesDictionary);
    vresource    msgRes     = vresourceGet (tisPermRes, tisperm_MessagesDictionary);
    vstr         *infoTitle = vresourceGetString (titleRes, tisperm_InfoNotice);
    vstr         *cautionTitle = vresourceGetString (titleRes, tisperm_CautionNotice);

    vstr *format = vresourceGetString (msgRes, msgName);
    // NOTE: vstrScribe cleans up vstr *format
    vstr *msg = vstrCloneScribed (vcharScribeFormatScribedVarargs (vstrScribe (format), argptr));

    va_end (argptr);

    if (msgType == TISPERM_INFO_NOTICE)
    {
	infoNotice (dialog, infoTitle, msg);
    }
    else
    {
	result = cautionNotice (dialog, cautionTitle, msg);
    }
    vstrDestroy (infoTitle);
    vstrDestroy (cautionTitle);
    vstrDestroy (msg);

    return (result);
}


/*******************************************************************
*
*    StripWhitespace function
*
*******************************************************************/

char *StripWhitespace (char *stringPtr)
{
    int	i;

    for (i=strlen (stringPtr)-1; i > 0 && isspace (*(stringPtr + i)); i--)
    {
	*(stringPtr + i) = 0;
    }
    for (i=0; isspace (*(stringPtr + i)); i++)
	;
    strcpy (stringPtr, stringPtr + i);
    return (stringPtr);
}
