#include <vport.h>		/* for v*HEADER */
#include vstdlibHEADER
#include vdebugHEADER
#include vdialogHEADER

#include "tisgalaxy.h"
#include "tisgalaxylib.hxx"


/***************************************************************
*
*  tisProgressDialog
*   The ObserveWindow function for the class is defined in the
*	subclass, it tests for the open and sets the event
*	handler
*   The event handler is defined by the user app since it must
*	know about the subclass to get at the overriden
*	DoWork function
*   For an example see .../src/dataVuCC/vuExportClass.<hxx,cxx>
*	It's vuExp[ortClass uses tisProgressDialog as a wrapper
*	class for the work the export needs to do
*
****************************************************************/
vkindTYPED_FULL_DEFINITION(tisProgressDialog, vdialog, NULL);

/***********************************************************
*
*   tisProgressDialog LoadInit
*
************************************************************/

void tisProgressDialog::LoadInit(vresource resource)
{
    quitting = vFALSE;
    messageText = vtextitem::RequestSubLoad (this, tisgalaxy_Text);
    progressDomain = ProgressDomain::RequestSubLoad (this, 
					    tisgalaxy_ProgressDomain);
    cancelButton = ProgressCancelButton::RequestSubLoad (this, 
					    tisgalaxy_CancelButton);

    vdialog::LoadInit (resource);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/************************************************************
*
* tisProgressDialog::Close()
*
*************************************************************/

void tisProgressDialog::Close()
{
    // we will come here if a cancel or quit happens
    // break us out of our sub event loop started in
    // OpenProgress ();
    vevent::StopProcessing ();

    vdialog::Close();

    quitting = vTRUE;
}

/************************************************************
*
*   tisProgressDialog::Destroy()
*	Destroy a progress dialog
* 
*************************************************************/

void tisProgressDialog::Destroy()
{
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy();
}


/************************************************************
*
*   tisProgressDialog::DoWork()
*	Virtual function for user to write when subclasses
*	tisProgressDialog
* 
*************************************************************/

vbool tisProgressDialog::DoWork(double *percentCompleted)
{
    return ((vbool)1);
}


/***********************************************************
*
*   tisProgressDialog open function
*
***********************************************************/

void tisProgressDialog::OpenProgress (void)
{
    vdomain *theDomain = progressDomain -> GetDomain ();
    // this statement makes it so that the drawing domain has a
    // finite boundary
    theDomain -> SetVirtualBounds (vFALSE);
    // now give that domain the bounds of the rectangle in
    // the .vr file
    const vrect   *drect = progressDomain -> GetRect ();
    theDomain -> SetBounds ((vrect *)drect);

    // store the color blue to be used for our progress bar
    vcolorSpec	*colorSpec = new vcolorSpec;
    colorSpec -> FSetRGB (0.0, 0.0, 1.0);
    progressDomain -> domainColor = vcolor::Intern(colorSpec);
    delete colorSpec;

    // cancel button has to know about the progress dialog
    cancelButton -> SetProgressDialog (this);

    vdialog::Open ();
    // hold us here while until the work is done or the
    // the app get a cancel or quit
    vevent::Process ();
}


/*************************************************************
*
* Progress dialog function to update the domain rectangle
*
**************************************************************/

void tisProgressDialog::UpdateDomain (double *percentCompleted)
{
    vdomain		*theDomain;
    vdomainObjectSet	*objectSet;
    vdomainObject	*object;
    vrect		*domainBounds;
    vrect		newRect;
    int			rectWidth;

    // I'm not sure why this is done other than the example
    // creates this temporary object set then deletes it later
    theDomain = progressDomain -> GetDomain ();
    objectSet = vdomainCreateObjectSet (NULL);

    // get the domains bounds area
    domainBounds = theDomain -> GetBounds ();
    // calc the width of the rectangle to be drawn
    rectWidth = (int)((*percentCompleted) * (double)(domainBounds -> GetWidth()));
    // leave 1 pixel at start and end of rectangle
    if (rectWidth >= domainBounds -> GetWidth ())
    {
	rectWidth -= 2;
    }
    // if we have not defined a rectangle yet, do so
    object = theDomain -> GetTopObject ();
    if (object == NULL)
    {
	// this creates the object
	object = theDomain -> AddObject ();
    }
    newRect.Set (domainBounds -> GetX () + 1, domainBounds -> GetY () + 1,
		 rectWidth, domainBounds -> GetHeight () - 2);
    // set the rectangle of the object
    theDomain -> SetObjectBounds (vdomainObjectToSet (object), &newRect);
    // not sure if this is needed, but sample code did it, only it
    // stored a piece of data, I use NULL incase I need to explicitly
    // need to set to NULL
    theDomain -> SetObjectData (vdomainObjectToSet (object), NULL);
    // add object to tempoary object set
    vdomainAddToObjectSet (&objectSet, object);
    // cause redraw of domain
    progressDomain -> InvalView ();
    theDomain -> ShowObjects (objectSet);
    // clean up
    vdomainDestroyObjectSet (objectSet);
}


/***************************************************************
*
*  Progress dialog cancel button
*
***************************************************************/
vkindTYPED_FULL_DEFINITION(ProgressCancelButton, vbutton, NULL);

void ProgressCancelButton::ObserveDialogItem (vevent *event)
{
    progressDialog -> SetQuitting ((vbool) 1);
}


/************************************************************
*
*   Progress dialog domain item
*
*************************************************************/
vkindTYPED_FULL_DEFINITION(ProgressDomain, vdomainitem, NULL);

void ProgressDomain::DrawObject (vdomainObject *object)
{
    vdomain *theDomain = GetDomain ();
    vcolor  *oldColor = vdraw::GetColor ();
    vdraw::SetColor (domainColor);
    vrect   *bounds = theDomain -> GetObjectBounds (object);
    vdraw::RectsFill (bounds, 1);
    vdraw::SetColor (oldColor);
}
