/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************************
/* 									    *
/*  THE INFORMATION IN THIS SOFTWARE IS  SUBJECT TO CHANGE WITHOUT NOTICE   *
/*  AND  SHOULD  NOT  BE  CONSTRUED AS  A COMMITMENT BY DIGITAL EQUIPMENT   *
/*  CORPORATION.							    *
/* 									    *
/*  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE  OR  RELIABILITY OF ITS   *
/*  EXAMPLE SOFTWARE PROGRAMS, AS IS PROVIDED HERE.                         *
/* 									    *
/*									    *
/****************************************************************************

/*
 *      getadr.c -- program to acquire ethernet controller addresses
 *              (c) Digital Equipment Corporation, 1989
 *              Author: Jim Lieb, Digital Equipment Corp. suntan::lieb
 *              This program is distributed free as a demonstration
 *              of network access methods.  No warrantee is expressed or
 *              implied.
 *
 *      Application: Use the getadr function to get the hardware address
 *              for the system in applications where the license key is
 *              verified by checking something unique like a system serial
 *              number (not always present on DEC cpus).
 *
 *      NOTE:   There is a difference between the hardware address and the
 *              physical address of Digital manufactured Ethernet controller.
 *              The hardware address is a unique number, typically of the
 *              form:  08-00-2B-xx-xx-xx, as assigned by IEEE authority. The
 *              xx-xx-xx number is fixed in ROM and is unique for each
 *              controller.  This address is guaranteed fixed and unique.
 *              The physical address is programmable and is MODIFIED by the
 *              DECnet routing layer to encode the 16 bit Phase IV DECnet
 *              address in the least significant 2 octets (byte swapped)
 *              of the source address of packets. This address, typically
 *              of the form aa-00-04-00-yy-yy where yy-yy is the DECnet
 *              address, becomes the address used by the controller for
 *              packet receive and transmit.
 *
 *              March 3, 1990 - Jim Lieb, changed to use internal tables
 *                              and clean up struct usage.  Please note!
 *              May 21, 1990 - John Dustin OSG, use status bits rather than
 *                              string compares to get right interface.
 *                              also test for interface turned on and use
 *                              best guess of none on...
 *		September 28, 1993 - Doug Olson, updated to match change in
 *				/usr/include/net/if.h (IFF_MULTICAST).
 */

#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>                      /* error numbers */
#include <stdio.h>

static char getadrSID[] = "@(#)getadr.c 1.5 6/22/90 demo";
extern int errno;

int getadr();

#if 0
/* main -- Shell to demonstrate the call to the getadr function.  This
 *      also prints the address contents to stdout.
 */

main(argc, argv, envp)
int argc;
char **argv, **envp;
{
        struct ifdevea ifr;
        char *interface;

        if(argc < 2)
                interface = NULL;
        else
                interface = argv[1];
        if(getadr(interface, &ifr) < 0) {
                fprintf(stderr, "couldn't get network interface\n");
                exit(1);
        }
        printf("ifr_name = '%s'\n", ifr.ifr_name);
        printf(
           "Ethernet hardware address = (%02x-%02x-%02x-%02x-%02x-%02x)\n",
                 ifr.default_pa[0], ifr.default_pa[1], ifr.default_pa[2],
                ifr.default_pa[3], ifr.default_pa[4], ifr.default_pa[5]);
        printf(
            "Ethernet physical address = (%02x-%02x-%02x-%02x-%02x-%02x)\n",
                ifr.current_pa[0], ifr.current_pa[1], ifr.current_pa[2],
                ifr.current_pa[3], ifr.current_pa[4], ifr.current_pa[5]);
}
#endif

/*      getadr -- get primary ethernet interface address
 *              read list of net devices via ioctl to ifreqs.  Pass over
 *              loopback device and any serial line interfaces.  Use first
 *              ether device found.  This could cause a problem if there are
 *              multiple ether controllers on the system as it will find the
 *              first device configured.  No panic since most have only one
 *              device.  This function does not require setuid(2) priviledges.
 */

#define IFREQCNT 64     /* need enough room for total number of possible
                         * interfaces, including aliases for LAT and DECnet
                         */
int getadr(name, ifrp)
char *name;
struct ifdevea *ifrp;
{
        int s;
        struct ifreq *ifr;
        struct ifreq ifreqs[IFREQCNT];
        struct ifreq tmp_ifr;
        struct ifconf ifc;
        char *maybe = NULL;     /* a device that might be it */

        bzero(ifrp, sizeof(struct ifdevea));
        if((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
                perror("socket");
                return( -1);
        }
        if(name != NULL)
                strcpy(ifrp->ifr_name, name);
        else {
                ifc.ifc_req = ifreqs;
                ifc.ifc_len = sizeof(ifreqs);
                if (ioctl(s, SIOCGIFCONF, &ifc) < 0) {
                        perror("ioctl: SIOCGIFCONF");
                        return(-1);
                }
                for(ifr = ifreqs; ifr < &ifreqs[IFREQCNT]; ifr++) {
                        if(strlen(ifr->ifr_name) == 0) {
                           if (maybe == NULL) {
                             fprintf(stderr, "No network devices configured\n");
                             return(-1);
                           } else
                                   break;  /* check out maybe */
                        } else {
                                (void) strcpy(tmp_ifr.ifr_name,ifr->ifr_name);
                                if (ioctl(s, SIOCGIFFLAGS, &tmp_ifr) < 0) {
                                        perror("ioctl: SIOCGIFFLAGS");
                                        return(-1);
                                }
                                if ((tmp_ifr.ifr_flags & (IFF_POINTOPOINT))
                                         == (IFF_POINTOPOINT)) {
                                        continue; /* skip point to point */
                                }
                                if ((tmp_ifr.ifr_flags
                                    & (IFF_MULTICAST|IFF_BROADCAST))
                                          != (IFF_MULTICAST|IFF_BROADCAST))
{
                                        continue;       /* skip CI, etc. */
                                }
                                if ((tmp_ifr.ifr_flags & (IFF_UP|IFF_RUNNING))
                                         != (IFF_UP|IFF_RUNNING)) {
                                        /* save this name, it might be the
                                         * right one but just isn't up now
                                         */
                                        if (maybe == NULL)
                                                maybe = ifr->ifr_name;
                                        continue;   /* not up and running */
                                }
                                (void)strcpy(ifrp->ifr_name,tmp_ifr.ifr_name);
                                maybe = NULL; /* found real one, clear maybe */
                                break;
                        }
                }
        }
        if (maybe != NULL) {
                (void) strcpy(ifrp->ifr_name, maybe);
        }
        if (ioctl(s, SIOCRPHYSADDR, ifrp) < 0) {
                perror("ioctl: SIOCRPHYSADDR");
                return(-1);
        }
        return 0;
}

