/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		ROUTINES TO VALIDATE LICENSES			*/
/*								*/
/*	The following routines validate the soft and hard	*/
/*	licenses for the TIS 4000. A license must conform	*/
/*	to the tlm structure conventions in all cases.		*/
/*	Programs that use soft licenses will die if their	*/
/*	license is expired. Programs that use hard licenses	*/
/*	will continue to run but display an info box on		*/
/*	the screen letting the user know that his license	*/
/*	has expired, if that is the case. If licenses are	*/
/*	OK, then the programs return silently to the caller.	*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include "msgMon.h"

/****************************************************************/
/*								*/
/*	CONSTANTS THAT DEFINE THE LICENSE ARRANGEMENTS		*/
/*								*/
/****************************************************************/

#define SOFT_FEATURE	"35"
#define HARD_FEATURE	"45"

#define KEY_DIR		"/tis4000/tasks/keys"
#define LDISP_NAME	"/tis4000/tasks/tisLdisp"


/****************************************************************/
/*								*/
/*	TEXT LINES FOR MESSAGES FROM DAEMONS			*/
/*								*/
/****************************************************************/

static char	daemonSoftCorrupt[] =
    "The TIS 4000 soft license for this system is missing or has been corrupted.";

static char	daemonSoftExpired[] =
    "The TIS 4000 soft license for this system has expired.";

static char	daemonIllegalNow[] =
    "Continued use of this program may constitute illegal use of protected software.";

static char	daemonSoftCall[] =
    "Please call Tate Integrated Systems for assistance: 410-581-0422";

static char	daemonHardCorrupt[] =
    "The TIS 4000 hard license for this system is missing or has been corrupted.";

static char	daemonHardExpired[] =
    "The TIS 4000 hard license for this system has expired.";

static char	daemonHardCall[] =
    "Please call Tate Integrated Systems immediately for assistance: 410-581-0422.";


static int tisLicense();


/****************************************************************/
/*								*/
/*		CHECK THE SOFT LICENSED FEATURE			*/
/*								*/
/*	Errors are displayed by invoking tisLdisp.		*/
/*								*/
/****************************************************************/

tisSoftLicense()
{
    return (tisLicense(SOFT_FEATURE, 0));
}


/****************************************************************/
/*								*/
/*		CHECK THE HARD LICENSED FEATURE			*/
/*								*/
/*	Errors are displayed by invoking tisLdisp.		*/
/*								*/
/****************************************************************/

tisHardLicense()
{
    return (tisLicense(HARD_FEATURE, 0));
}


/****************************************************************/
/*								*/
/*		CHECK A SOFT DAEMON LICENSE			*/
/*								*/
/*	Errors are announced via syslog.			*/
/*								*/
/****************************************************************/

tisDaemonSoftLicense()
{
    return (tisLicense(SOFT_FEATURE, 1));
}


/****************************************************************/
/*								*/
/*		CHECK A HARD DAEMON LICENSE			*/
/*								*/
/*	Errors are announced via syslog.			*/
/*								*/
/****************************************************************/

tisDaemonHardLicense()
{
    return (tisLicense(HARD_FEATURE, 1));
}


/****************************************************************/
/*								*/
/*		CHECK AN ARBITRARY LICENSED FEATURE		*/
/*								*/
/****************************************************************/

static tisLicense(char *tisFeature, int daemon)
{
    int status;
    char		whichWindow[4];

    /*
    Validate the key.
    */

    status = tlm(KEY_DIR, tisFeature);

    /*
    Check the return.
    */

    if (status == 1)			/* OK */
	{
	return (0);
	}

    else if (status == -1)		/* Expired */
        {
	if (!daemon)
	    {
	    if (!strcmp(tisFeature, SOFT_FEATURE))
	        strcpy(whichWindow, "2");
	    else
	        strcpy(whichWindow, "5");
	    }

	else
	    {
	    if (!strcmp(tisFeature, SOFT_FEATURE))
		{
		tisSysLog(0,0, daemonSoftExpired);
		tisSysLog(0,0, daemonIllegalNow);
		tisSysLog(0,0, daemonSoftCall);
		}
	    else
		{
		tisSysLog(0,0, daemonHardExpired);
		tisSysLog(0,0, daemonIllegalNow);
		tisSysLog(0,0, daemonHardCall);
		}
	    return (0);
	    }
	}

    else				/* Missing or corrupted license */
        {
	if (!daemon)
	    {
	    if (!strcmp(tisFeature, SOFT_FEATURE))
		strcpy(whichWindow, "1");
	    else
		strcpy(whichWindow, "4");

	    execl(LDISP_NAME, LDISP_NAME, whichWindow, (char *) 0);
	    exit(0);
	    }

	else
	    {
	    if (!strcmp(tisFeature, SOFT_FEATURE))
		{
		tisSysLog(0,0, daemonSoftCorrupt, (char *) 0);
		tisSysLog(0,0, daemonIllegalNow, (char *) 0);
		tisSysLog(0,0, daemonSoftCall, (char *) 0);
		}
	    else
		{
		tisSysLog(0,0, daemonHardCorrupt, (char *) 0);
		tisSysLog(0,0, daemonIllegalNow, (char *) 0);
		tisSysLog(0,0, daemonHardCall, (char *) 0);
		}
	    exit(0);
	    }
	}

    if (!fork())
	    {
	    execl(LDISP_NAME, LDISP_NAME, whichWindow, (char *) 0);
	    }

    /*
    Wait 10 seconds to give the info box time to pop up to the top.
    */

    sleep(10);

    /*
    If this is one of the "soft" programs, then die now.
    Otherwise, just return to the caller.
    */

    if (!strcmp(tisFeature, SOFT_FEATURE))
	exit(0);

    return (0);
}
