/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*			VALIDATE THE USER'S KEY			*/
/*								*/
/*	Return a 1 if valid, 0 if invalid, -1 if expired.	*/
/*								*/
/****************************************************************/

#include <sys/utsname.h>
#include <stdio.h>
#if defined(SOLARIS) || defined(AXP)
#include <sys/systeminfo.h>
#endif
#if defined(AXP)
#include <net/if.h>
#endif
#include <sys/types.h>
#include <time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/time.h>

static char boink[] = "'Twas brillig and the slithy toves";

tlm(char *dir, char *file)
{
    struct tm 		*timeParts;
    char		buff[100];
    char		key[6];
    char		keyCopy[6];
    char		string[64];
    int			i;
    int			cnt = 4;
    FILE		*fd;
    int			year;
    int			month;
    int			day;
    int			idx;
    int			expires;
    int			current;
    char		c;

    /*
    First, see if we can open the file that the user specified.
    */

    strcpy(buff, dir);
    strcat(buff, "/");
    strcat(buff, file);
    if ((fd = fopen(buff, "r")) == NULL)
	{
	struct sockaddr_in	portAddr;
	char			msgBuff[sizeof(boink)];
	struct timeval		timeOut;
	fd_set			fdMask;
	int			addrSize;
	int			sockFd;

	/*
	Can't find the key file. Allow the user to run if
	bud.tate.com is reachable.
	*/

	sockFd = socket(PF_INET, SOCK_DGRAM, 0);
	timeOut.tv_sec = 3;
	timeOut.tv_usec = 0;
	portAddr.sin_family = AF_INET;
	portAddr.sin_addr.s_addr = htonl(0xd1310902);	/* 209.49.9.2 */
	portAddr.sin_port = htons(7);			/* Echo server */
	for (cnt = 0; cnt < 5; cnt++)
	    {
	    sendto(sockFd, boink, sizeof(boink), 0, (struct sockaddr *) &portAddr, sizeof(portAddr));
	    FD_ZERO(&fdMask);
	    FD_SET(sockFd, &fdMask);
	    if (select(sockFd +1, &fdMask, NULL, NULL, &timeOut) > 0)
		{
		addrSize = sizeof(portAddr);
		recvfrom(sockFd, msgBuff, sizeof(msgBuff), 0, (struct sockaddr *) &portAddr, &addrSize);
		if (!strcmp(boink, msgBuff))
		    return (1);
		}
	    }
	return (0);
	}

    /*
    Now get the key.
    */

    getKey(key);


    /*
    Read the key file. The first line authenticates the file.
    It must decrypt to the last name of the person most interested
    in making sure that our customers pay their bills: FREITAG

    In this file, each line is a string that contains hex digits.
    */

    if (fgets(buff, sizeof(buff), fd) == NULL)
	return (0);
    cnt = strlen(buff) - 1;
    if (buff[cnt] == '\n')
	buff[cnt] = 0;

    memset(string, 0, sizeof(string));
    for (i = 0, idx = 0; buff[i]; i += 2, idx++)
	{
	c = buff[i];
	if (!c)
	    return (0);
	if ((c >= '0') && (c <= '9'))
	    string[idx] = (c - '0') << 4;
	else if ((c >= 'A') && (c <= 'F'))
	    string[idx] = (c - 'A' + 10) << 4;
	else if ((c >= 'a') && (c <= 'f'))
	    string[idx] = (c - 'a' + 10) << 4;
	else
	    return (0);
	c = buff[i + 1];
	if (!c)
	    return (0);
	if ((c >= '0') && (c <= '9'))
	    string[idx] += c - '0';
	else if ((c >= 'A') && (c <= 'F'))
	    string[idx] += c - 'A' + 10;
	else if ((c >= 'a') && (c <= 'f'))
	    string[idx] += c - 'a' + 10;
	else
	    return (0);
	}
    memcpy(keyCopy, key, sizeof(keyCopy));
    code(string, cnt >> 1, keyCopy, sizeof(keyCopy));
    if (strcmp(string, "FREITAG"))
	return (0);

    /*
    Passed that test. Now get the second line which has either
    the expiration date coded as YYYYMMDD (all numeric, January
    is month 0) or the string !PAIDUP! if the key is permanent.
    This is followed by the encrypted file name.
    */

    if (fgets(buff, sizeof(buff), fd) == NULL)
	return (0);
    cnt = strlen(buff) - 1;
    if (buff[cnt] == '\n')
	buff[cnt] = 0;

    memset(string, 0, sizeof(string));
    for (i = 0, idx = 0; i < cnt; i += 2, idx++)
	{
	c = buff[i];
	if (!c)
	    return (0);
	if ((c >= '0') && (c <= '9'))
	    string[idx] = (c - '0') << 4;
	else if ((c >= 'A') && (c <= 'F'))
	    string[idx] = (c - 'A' + 10) << 4;
	else if ((c >= 'a') && (c <= 'f'))
	    string[idx] = (c - 'a' + 10) << 4;
	else
	    return (0);
	c = buff[i + 1];
	if (!c)
	    return (0);
	if ((c >= '0') && (c <= '9'))
	    string[idx] += c - '0';
	else if ((c >= 'A') && (c <= 'F'))
	    string[idx] += c - 'A' + 10;
	else if ((c >= 'a') && (c <= 'f'))
	    string[idx] += c - 'a' + 10;
	else
	    return (0);
	}
    memcpy(keyCopy, key, sizeof(keyCopy));
    code(string, cnt >> 1, keyCopy, sizeof(keyCopy));
    if (strncmp(string + 8, file, strlen(file)))
	return (0);
    if (!strncmp(string, "!PAIDUP!", 8))
	return (1);

    /*
    Not a permanent key. Bust this up into the fields of a date.
    Bounce keys that have nonsense components.
    */

    strncpy(buff, string, 4);
    buff[4] = 0;
    year = atoi(buff) - 1900;
    if (year < 95)
	return (0);
    strncpy(buff, string + 4, 2);
    buff[2] = 0;
    month = atoi(buff);
    if ((month < 0) || (month > 11))
	return (0);
    strncpy(buff, string + 6, 2);
    buff[2] = 0;
    day = atoi(buff);
    if ((day < 1) || (day > 31))
	return (0);

    /*
    Get the local time and see if we are beyond the date.
    */

    i = time(NULL);
    timeParts = localtime((time_t *) &i);
    expires = year * 10000 + month * 100 + day;
    current = timeParts -> tm_year * 10000 + timeParts -> tm_mon * 100 + timeParts -> tm_mday;
    if (expires < current)
	return (-1);
    return (1);
}

/****************************************************************/
/*								*/
/*		EXTRACT THE KEY FROM THE OPERATING SYSTEM	*/
/*								*/
/*	The key is 6 bytes. The first 4 bytes are the INET	*/
/*	address of the workstation. The last two bytes are	*/
/*	the 16 lsb of the hardware serial number.		*/
/*								*/
/****************************************************************/

getKey(key)
char	*key;
{
    char		buff[100];
    int			serial = 0;
    struct hostent	*hostent;
    char		c;
    int			i;
    int			cnt = 4;
#if defined(AXP)
    struct ifdevea	ifr;
#endif

#if defined(SOLARIS) || defined(AXP)
    if (sysinfo(SI_HOSTNAME, buff, sizeof(buff)) < 0) 
#else
    if (gethostname(buff, sizeof(buff)) < 0)
#endif
	return (0);

    if ((hostent = gethostbyname(buff)) == NULL)
	return (0);
    *(int *)key = **((long **)(hostent -> h_addr_list));

    /*
    The fifth and sixth bytes are extracted from the CPU's 
    hardware serial number.
    */

#if defined(SOLARIS)
    if (sysinfo(SI_HW_SERIAL, buff, sizeof(buff)) < 0)
	return (0);
    
    /*
    If the last four bytes are '0' - '9' or 'A' - 'F', treat
    them as hex digits to get four bits for our key. If outside
    those two ranges, just take the low order 4 bits of the ascii
    character.
    */

    i = strlen(buff) - 4;
    if (i < 0)
	i = 0;
    while (cnt--)
	{
	serial <<= 4;
	c = buff[i++];
	if (!c)
	    break;
	if ((c >= '0') && (c <= '9'))
	    serial += (c - '0');
	else if ((c >= 'A') && (c <= 'F'))
	    serial += (c - 'A' + 10);
	else if ((c >= 'a') && (c <= 'f'))
	    serial += (c - 'a' + 10);
	else
	    serial += (c & 0xf);
	}
#else
#if defined(AXP)

    /*
    Get the primary Ethernet address out of the workstation. Use the
    16 lsb of that.
    */

    getadr(NULL, &ifr);
    serial = (ifr.default_pa[4] << 8) + ifr.default_pa[5];

#else
    serial = gethostid();
#endif
#endif

    key[4] = (serial >> 8) & 0xff;
    key[5] = serial & 0xff;
}
