/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		CHECK LICENSING FOR X TERMINALS			*/
/*								*/
/*	This routine is to be called by dm after it has		*/
/*	already called the normal TIS licensing routine.	*/
/*	This routine gets the DISPLAY variable out of the	*/
/*	environment. If that string has a length and does	*/
/*	not begin with a colon, then we get the Inet		*/
/*	address for that name. We use the low order 16 bits	*/
/*	of the Inet address as a feature number for		*/
/*	licensing. If that outboard node is not licensed	*/
/*	or the license is expired, then we do not run.		*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/time.h>

#include "msgMon.h"

static char *pinet();
static char message[] = "'Twas brillig and the slithy toves";

xLicense()
{
    char		*dispEnv;
    char		*newEnv;
    struct hostent	*hostent;
    char		featureName[16];
    int			feature;
    int			inetAddr;
    struct sockaddr_in	portAddr;
    int			fd;
    int			ntries;
    char		buff[sizeof(message)];
    struct timeval	timeOut;
    fd_set		readFd;
    int			addrSize;

    /*
    First, get the DISPLAY variable from the environment. If
    there isn't one, then just return. X will crash the program
    all by itself.
    */

    dispEnv = getenv("DISPLAY");
    if (!dispEnv)
	return (0);

    /*
    Copy the environment string to our local memory just in
    case the system has write protected it against us.
    */

    newEnv = malloc(strlen(dispEnv) + 1);
    strcpy(newEnv, dispEnv);

    /*
    Find the ':' and smash that to a null. This will leave just
    the node name as a normal C language string. If that is the
    first byte of the string, then we are on the normal console
    for the workstation and the previous checking is all that
    applies. Just return in that case.
    */

    dispEnv = strchr(newEnv, ':');
    if (!dispEnv)
	return (0);	/* Malformed variable. Let X crash the program. */
    *dispEnv = 0;
    if (!strlen(newEnv))
	return (0);	/* On the normal console. */
    
    /*
    Now we have the node name. Get its Inet address.
    */

    hostent = gethostbyname(newEnv);
    free(newEnv);
    if (!hostent)
	return (0);	/* Unknown node. Let X crash the program. */
    inetAddr = ntohl(**((long **)(hostent -> h_addr_list)));

    /*
    Use the low order 16 bits as the feature number. Look up this key.
    */

    feature = inetAddr & 0xffff;
    sprintf(featureName, "%d", feature);
    feature = tlm("/tis4000/tasks/keys", featureName);
    
    /*
    If OK, we are done.
    */

    if (feature == 1)
	return (0);
    
    else if (feature == -1)
	{

	/*
	Expired license. Display warning and then quit.
	*/

	execl("/tis4000/tasks/tisLdisp", "/tis4000/tasks/tisLdisp", "2", (char *) 0);
	}

    else
	{

	/*
	Either completely unlicensed or somebody tampered with it. See if we can ping
	bud. If so, then assume it's on our floor.
	*/

	portAddr.sin_family = AF_INET;
	portAddr.sin_addr.s_addr = htonl(0xd1310902);	/* 209.49.9.2 */
	portAddr.sin_port = htons(7);			/* Echo server */
	fd = socket(PF_INET, SOCK_DGRAM, 0);
	if (fd < 0)
	    return (0);
	for (ntries = 0; ntries < 3; ntries++)
	    {
	    sendto(fd, message, sizeof(message), 0, (struct sockaddr *) &portAddr, sizeof(portAddr));
	    timeOut.tv_sec = 1;
	    timeOut.tv_usec = 0;
	    FD_ZERO(&readFd);
	    FD_SET(fd, &readFd);
	    if (select(fd + 1, &readFd, NULL, NULL, &timeOut) > 0)
		{
		recvfrom(fd, buff, sizeof(buff), 0, (struct sockaddr *) &portAddr, &addrSize);
		if (!strcmp(buff, message))
		    break;
		}
	    }
	close(fd);
	if (ntries < 3)
		return (0);		/* We're in house */

	/*
	Not in house. Display message and then quit.
	*/

	execl("/tis4000/tasks/tisLdisp", "/tis4000/tasks/tisLdisp", "1", (char *) 0);
	}

    /*
    This code should never get executed, but if it does, then something is wrong
    with tisLdisp. Write an error message to the system log file and then die.
    */

    tisSysLog(0,0, "X Terminal Licensing Error! Node %s", pinet(inetAddr));
    exit(0);
}

static char *pinet(int inet)
{
    static char inString[32];

    sprintf(inString, "%d.%d.%d.%d", ((inet >> 24) & 0xff), ((inet >> 16) & 0xff), ((inet >> 8) & 0xff),
		inet & 0xff);
    return (inString);
}
