/* trendtextdlg.cxx,v 1.1.1.1 1999/08/03 14:45:11 tis4000 Exp
 *
 */

/*
 * TrendTextDialog - trend graph container class
 */

#include <vport.h>		/* for v*HEADER */
#include vstdlibHEADER		/* for random, exit, EXIT_SUCCESS */
#include vlistHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vstartupHEADER
#include vmemHEADER
#include vprefHEADER
#include vtextitemHEADER
#include vbuttonHEADER
#include vcharHEADER
#include vdialogHEADER
#include vnumHEADER

#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#endif WIN32

extern "C"
{
    #include "trendutils.h"
}
#include "trenddocs.h"
#include "trendtextdlg.h"

// Helper functions
class tlCloseButton : public vbutton {
public:
  // macro for overrides from vkindTyped
  vkindTYPED_FULL_DECLARATION(tlCloseButton);
  
  // macro for the standard versions of the 
  // necessary constructors and destructors
  vloadableINLINE_CONSTRUCTORS(tlCloseButton, vbutton);
  
  // virtual member function overrides
  void ObserveDialogItem(vevent *event);

  void SetParent (TrendTextDialog *p) { parent = p; }
private:
    TrendTextDialog *parent;
};

vkindTYPED_FULL_DEFINITION(tlCloseButton, vbutton, NULL);
void tlCloseButton::ObserveDialogItem (vevent *event)
{
    parent->CloseClick();
}

// macro definitions for vkindtyped overrides
vkindTYPED_FULL_DEFINITION(TrendTextDialog, vdialog, NULL);

// vloadable
void TrendTextDialog::LoadInit(vresource resource)
{
    // get handles to interesting elements
    titletext = vdialogLabelItem::RequestSubLoad (this, vnameInternGlobalLiteral("PointNameTag"));
    valuelist = TrendTextList::RequestSubLoad (this, vnameInternGlobalLiteral("TextListTag"));
    close = tlCloseButton::RequestSubLoad (this, vnameInternGlobalLiteral("CloseButton"));

    vdialog::LoadInit(resource);

    SetPopup (TRUE);
    close->SetParent (this);
    valuelist->SetParent (this);
    trendSet = NULL;
}

/*********
 *  TrendTextDialog::ObserveWindow()
 *
 *  The dialog notification procedure gets called whenever any 
 *  major event applies to the dialog as a whole.  We need to
 *  look for the event which tells us that the dialog has
 *  closed.
 *
 **********/
void TrendTextDialog::ObserveWindow (vevent *event)
{
    if (event->GetType() == veventWINDOW_STATE && event->IsClose ())
    {
	CloseNotify ();
    }
}

/*********
 *  TrendTextDialog::CloseClick()
 *
 *  The procedure gets called whenever the close button is clicked 
 *
 **********/
void TrendTextDialog::CloseClick ()
{
    if (IsOpen())
    {
	Close ();
	CloseNotify ();
    }
}

void TrendTextDialog::CloseNotify (void)
{
    // overide if notify is useful
}


void TrendTextDialog::SetTheCursor (int loc)
{
    valuelist->SetTheCursor (loc);
}

void TrendTextDialog::SetTrendSet (TrendSet *set)
{
    char    ptstring [200];
    char    valstring [30];
    int	    i;

    trendSet = set;
    for (i = 0; i < trendSet->numTrends; i++)
    {
	if (trendSet->trends[i].pointname [0] == '>')
	    break;
    }
    if (i < trendSet->numTrends)
    {
        sprintf (ptstring, "  Date     Time  ");
    }
    else
    {
        sprintf (ptstring, " Offset ");
    }
    for (i = 0; i < trendSet->numTrends; i++)
    {
	if (trendSet->trends[i].enable)
	{
	    sprintf (valstring, " %-28.28s", trendSet->trends[i].PointName ());
	    strcat (ptstring, valstring);
	}
    }
    titletext->SetTitle ((unsigned char *)ptstring);// this shows the text

    valuelist->SetTrendSet (set);
}

void TrendTextDialog::ControlClick (int id)
{
    NewCursor (id);
}

void TrendTextDialog::NewCursor (int id) // override with change callback!
{
    // User code should adjust document for new value
}

// macro definitions for vkindtyped overrides

vkindTYPED_FULL_DEFINITION(TrendTextList, vlistview, NULL);

// vloadable
void TrendTextList::LoadInit(vresource resource)
{
    vlistview::LoadInit(resource);
    trendSet = NULL;
}

/*
 * TrendTextList::Destroy
 *
 * Deletes the list
 */
void TrendTextList::Destroy()
{
    // Destroy list
    delete GetList();  // !!! Same as vdialogItem::itemList ??

    vlistview::Destroy();
}

/*
 * TrendTextList::ObserveSelect
 *
 * move cursor value
 */
void TrendTextList::ObserveSelect(int state)
{
    int row;
    int col;
    vlist *list = GetList();

    list->GetSelectedCell(GetSelection (), &row, &col);
    if ((row != vlistNO_CELLS_SELECTED) && (state == vlistviewSELECTION_HAS_CHANGED))
    {
	// set cursor value to row value
	parent->ControlClick (row);
    }
}

/*
 * TrendTextList::DrawCellContents
 *
 * Draw resource entry in list cell
 */
void TrendTextList::DrawCellContents(int row, int col, 
				      int /*selected*/, vrect *cellRect)
{
    char    ptstring [200];
    char    valstring [30];
    int	    i;

    if ((trendSet == NULL) || (row < 0) || (512 <= row))
	return;

    vdrawGSave();
    vdrawRectsClip(cellRect, 1);

    vrect temp = *cellRect;

    vdrawRectClip(temp.GetX(), temp.GetY(), temp.GetWidth(), temp.GetHeight());
    for (i = 0; i < trendSet->numTrends; i++)
    {
	if (trendSet->trends[i].pointname [0] == '>')
	    break;
    }
    if (i < trendSet->numTrends)
    {
        sprintf (ptstring, "%8.8s %8.8s", trendSet->trends[i].PointDate(row),
				    trendSet->trends[i].PointTime(row));
    }
    else
    {
        StrSpan ( (long)(row * trendSet->zoom), ptstring);
    }
    for (i = 0; i < trendSet->numTrends; i++)
    {
	if (trendSet->trends[i].enable)
	{
	    strcat (ptstring, "      ");
	    trendSet->trends[i].GetValueString (valstring, sizeof(valstring), row);
	    strcat (ptstring, valstring);
	}
    }
    vdrawRectShowCenterY ((unsigned char *)ptstring, &temp);// this shows the text

    vdrawGRestore();
}

void TrendTextList::SetTrendSet (TrendSet *set)
{
    int		ix;
    vliststr	*newlist;
    vlist	*oldlist;

    trendSet = set;

    // Destroy old list
    if (oldlist = GetList())
	delete oldlist;

    newlist = new vliststr;
    for (ix = 0; ix < 511; ix++)
        newlist->SetCellValueScribed (ix, 0, vnumScribeInt (ix));

    SetList (newlist);
    newlist->SetSelectMethod (GetSelection (), vlistONE_ONLY_SELECTION);
}

void TrendTextList::SetTheCursor (int newcur)
{
    cursor = newcur;
    GetList()->SelectRow (GetSelection (), cursor);
    ScrollSelectionIntoView ();
}
