/*
 * trendvu - TIS4000 Application control bar client machines
 *
 * Displays a menu bar for executing TIS4000 applications. If a
 * login has not been established with a server, a login dialog
 * box is displayed.
 */

#include <vport.h>		/* for v*HEADER */
#include vstdlibHEADER		/* for random, exit, EXIT_SUCCESS */
#include vapplicationHEADER
#include vmenubarHEADER
#include vmenuHEADER
#include vcontainerHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vmemHEADER
#include vprefHEADER
#include vtextitemHEADER
#include vcharHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vcomboboxHEADER
#include vprintHEADER
#include vmenuHEADER

#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#include <time.h>
#endif WIN32

extern "C"
{
    #include "msgMon.h"
    #include "trendutils.h"
}
#include "trenddocs.h"
#include "trendcontrols.h"
#include "trendgraph.h"
#include "trendlimitlist.h"
#include "trenddetaillist.h"
#include "trendeditdlg.h"
#include "trendlistdlg.h"
#include "trendprefsdlg.h"
#include "trendmacrodlg.h"
#include "trendvaluedlg.h"
#include "trendsaveasdlg.h"
#include "trendsearchdlg.h"
#include "trendselectdlg.h"
#include "trendtextdlg.h"
#include "tisgalaxylib.hxx"
extern "C"
{
    #include "oracleSnap.h"
    #include "trendSizes.h"
    #ifdef ORACLESNAP
    #include "oracleSnapLib.h"
    #else
    #ifdef WIN32
    #pragma pack (push, 1)
    #endif WIN32
    #include "cb45to63.h"
    #include "d4all.h"
    #ifdef WIN32
    #pragma pack (pop)
    #endif WIN32
    #include "OperPriv.h"
    #include "avmDefs.h"
    #include "xbaseSnapLib.h"
    #endif ORACLESNAP
}

class tvTrendGraph;
class tvDetailList;
class tvLimitList;
class tvTrendControls;
class tvTimer;
class tvTextDialog;

class TrendvuDialog : public vdialog
{
    public:
	// the default versions of the necessary constructors/destructor
	vloadableINLINE_CONSTRUCTORS(TrendvuDialog, vdialog);

	// overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(TrendvuDialog);

	// virtual member function overrides
	// vloadable
	virtual void LoadInit(vresource resource);
	// vwindow
	virtual void ObserveWindow(vevent *event);

	void NewView (char *trendsetname);
	int MenuCommand (int selector);
	void NewCursor (int pos);
	void NewRange (int lpos, int rpos);
	void DetailControlClick (int type, int id);
	void NewLimit (int id, int row);
	void ControlsClick (int type);
	void ToggleClick (int type, int value);
	void TextClosed (void);
	void Tick ();
	void BuildAboutMenu (vresource tisres);
	int	    loadiniterror;
	vmenu	    *helpMenu;
	TisAbout    *tisAbout;
    private:
	TrendSet trendSet;	// the working document
	int	cursor;		// location in trendSet
	SEARCH	lastSearch;	// last search spec

	// visual element classes
	tvTrendGraph *trendGraph;
	tvTrendControls *trendControls;
	tvLimitList *trendHiLimits;
	tvLimitList *trendLoLimits;
	tvDetailList *trendDetail;
	tvTimer	    *timer;

	// supporting storage
	long	Colors[8];
	long	Back;
	vcolorSpec vcolors[8];
	vcolorSpec vback;
	char	dbfilename [C_MAXPATHLEN];
	int	textshown;
	time_t	lastcurrent;
	char	**archTables;
	int	opCanAddNotes;
};

// Helper functions
class tvTrendGraph : public TrendGraph {
public:
    // macro for overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(tvTrendGraph);

    // macro for the standard versions of the 
    // necessary constructors and destructors
    vloadableINLINE_CONSTRUCTORS(tvTrendGraph, TrendGraph);

    void SetParent (TrendvuDialog *p) { parent = p; }
    virtual void NewCursor (int pos);
    virtual void NewRange (int lpos, int rpos);
private:
    TrendvuDialog *parent;
};

vkindTYPED_FULL_DEFINITION(tvTrendGraph, TrendGraph, NULL);
void tvTrendGraph::NewCursor (int pos)
{
    parent->NewCursor (pos);
}

void tvTrendGraph::NewRange (int lpos, int rpos)
{
    parent->NewRange (lpos, rpos);
}

class tvDetailList : public TrendDetailList {
public:
    // macro for overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(tvDetailList);

    // macro for the standard versions of the 
    // necessary constructors and destructors
    vloadableINLINE_CONSTRUCTORS(tvDetailList, TrendDetailList);

    void SetParent (TrendvuDialog *p) { parent = p; }
    virtual void ControlClick (int t, int i);
private:
    TrendvuDialog *parent;
};

vkindTYPED_FULL_DEFINITION(tvDetailList, TrendDetailList, NULL);
void tvDetailList::ControlClick (int type, int id)
{
    parent->DetailControlClick (type, id);
}

class tvLimitList : public TrendLimitList {
public:
    // macro for overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(tvLimitList);

    // macro for the standard versions of the 
    // necessary constructors and destructors
    vloadableINLINE_CONSTRUCTORS(tvLimitList, TrendLimitList);

    void SetParent (TrendvuDialog *p) { parent = p; }
    virtual void NewLimit (int t, int r);
private:
    TrendvuDialog *parent;
};

vkindTYPED_FULL_DEFINITION(tvLimitList, TrendLimitList, NULL);
void tvLimitList::NewLimit (int whattype, int row)
{
    parent->NewLimit (whattype, row);
}

class tvTrendControls : public TrendControls {
public:
  // macro for overrides from vkindTyped
  vkindTYPED_FULL_DECLARATION(tvTrendControls);
  
  // macro for the standard versions of the 
  // necessary constructors and destructors
  vloadableINLINE_CONSTRUCTORS(tvTrendControls, TrendControls);
  
  void SetParent (TrendvuDialog *p) { parent = p; }
  virtual void ControlClick (int t);
  virtual void ToggleClick (int t, int v);
private:
    TrendvuDialog *parent;
};

vkindTYPED_FULL_DEFINITION(tvTrendControls, TrendControls, NULL);
void tvTrendControls::ControlClick (int type)
{
    parent->ControlsClick (type);
}

void tvTrendControls::ToggleClick (int type, int value)
{
    parent->ToggleClick (type, value);
}

/*******************************************************************
*    tvTimer Observer
*******************************************************************/
class tvTimer : public vtimer
{
public:
    void SetParent (TrendvuDialog *p) { parent = p; }
    virtual void ObserveTimer (void);

    // start timer sequence
    void StartTimer (int timeDelaySecs, int timeDelayNSecs);
private:
    TrendvuDialog *parent;
};


void tvTimer::ObserveTimer (void)
{
    parent->Tick ();
}

void tvTimer::StartTimer (int timeDelaySecs, int timeDelayNSecs)
{
    // timer restarts as soon as ObserveTimer finishes
    SetRecurrent ();
    // set timer to expire every .1 seconds
    // args are sec, nanosecs
    SetPeriod (timeDelaySecs, timeDelayNSecs);
    Start ();
}

class tvTextDialog : public TrendTextDialog {
public:
    // macro for overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(tvTextDialog);

    // macro for the standard versions of the 
    // necessary constructors and destructors
    vloadableINLINE_CONSTRUCTORS(tvTextDialog, TrendTextDialog);

    void SetParent (TrendvuDialog *p) { parent = p; }
    void NewCursor (int pos);
    void CloseNotify (void);
private:
    TrendvuDialog *parent;
};

vkindTYPED_FULL_DEFINITION(tvTextDialog, TrendTextDialog, NULL);
void tvTextDialog::NewCursor (int pos)
{
    parent->NewCursor (pos);
}

void tvTextDialog::CloseNotify (void)
{
    parent->TextClosed ();
}

// Global values -
TrendvuDialog *maindialog;

// support dialog handles
TrendEditDialog  *editDialog;
TrendListDialog  *listDialog;
TrendPrefsDialog  *prefsDialog;
TrendMacroDialog  *macroDialog;
TrendValueDialog  *valueDialog;
TrendSaveAsDialog  *saveAsDialog;
TrendSearchDialog  *searchDialog;
TrendSelectDialog  *selectDialog;
tvTextDialog	*textDialog;

/*
 * Declare a selector command subclass to override the issue function.
 */
class mySelectorCommand : public vcommandSelector
{
public:
  // overrides from vkindTyped
  vkindTYPED_FULL_DECLARATION(mySelectorCommand);
  
  // the standard versions of the necessary constructors/destructor
  vloadableINLINE_CONSTRUCTORS(mySelectorCommand, vcommandSelector);
  
  // virtual member function override.
  virtual int Issue(vdict *attr2value);
};

/*
 * tag-to-selector bindings
 */
enum {
    MENU_NEW,	    MENU_OPEN,	    MENU_EDIT,	    MENU_SAVE,	    MENU_SAVE_AS,
    MENU_DELETE,    MENU_PRINT,	    MENU_SETFILE,   MENU_PREF,	    MENU_SAVE_PREF,
    MENU_EXIT,      MENU_MATCHRES,  MENU_FILE_ALM,  MENU_FILE_MAX,  MENU_FILE_MIN,
    MENU_SUBST,     MENU_ALM,	    MENU_MAX,	    MENU_MIN,	    MENU_NEXT,
    MENU_PREV,      MENU_READ,	    MENU_ADD,	    MENU_HELP,
    SELECTOR_COUNT // marks the number of valid entries!
};
static vcommandSelectorInitializer selCmdBindings[] = {
    {"NewMenuTag", MENU_NEW },
    {"OpenMenuTag", MENU_OPEN },
    {"EditMenuTag", MENU_EDIT },
    {"SaveMenuTag", MENU_SAVE },
    {"SaveAsMenuTag", MENU_SAVE_AS },
    {"DeleteMenuTag", MENU_DELETE },
    {"PrintMenuTag", MENU_PRINT },
    {"SetFileMenuTag", MENU_SETFILE },
    {"PrefMenuTag", MENU_PREF },
    {"SavePrefMenuTag", MENU_SAVE_PREF },
    {"ExitMenuTag", MENU_EXIT },
    {"MatchResMenuTag", MENU_MATCHRES },
    {"FileAlmMenuTag", MENU_FILE_ALM },
    {"FileMaxMenuTag", MENU_FILE_MAX },
    {"FileMinMenuTag", MENU_FILE_MIN },
    {"SubstMenuTag", MENU_SUBST },
    {"AlmMenuTag", MENU_ALM },
    {"MaxMenuTag", MENU_MAX },
    {"MinMenuTag", MENU_MIN },
    {"NextMenuTag", MENU_NEXT },
    {"PrevMenuTag", MENU_PREV },
    {"ReadMenuTag", MENU_READ },
    {"AddMenuTag", MENU_ADD },
    {"HelpMenuTag", MENU_HELP },
};

/*--------------------------------------------------------------------------
 |                              mySelectorCommand                          |
 --------------------------------------------------------------------------*/
// definitions for overrides from vkindTyped
vkindTYPED_FULL_DEFINITION(mySelectorCommand, vcommandSelector, NULL);

int mySelectorCommand::Issue(vdict *attr2value)
{
    return maindialog->MenuCommand (GetSelector());

}

// macro definitions for vkindtyped overrides

vkindTYPED_FULL_DEFINITION(TrendvuDialog, vdialog, NULL);

/*********
 *  TrendvuDialog::LoadInit()
 *
 *  This method gets called whenever the dialog is loaded from a resource.
 *  Match our vcontainer subclass with the TrendGraph contained in the resource,
 *  and initialize the subclass.
 *
 **********/
void TrendvuDialog::LoadInit(vresource resource)
{
    int i;

    loadiniterror = vFALSE;

    // initialize components

    trendHiLimits = tvLimitList::RequestSubLoad (this, vnameInternGlobalLiteral("HiLimitViewTag"));
    trendLoLimits = tvLimitList::RequestSubLoad (this, vnameInternGlobalLiteral("LoLimitViewTag"));
    trendDetail = tvDetailList::RequestSubLoad (this, vnameInternGlobalLiteral("DetailViewTag"));
    trendGraph = tvTrendGraph::RequestSubLoad (this, vnameInternGlobalLiteral("GraphTag"));
    trendControls = tvTrendControls::RequestSubLoad (this, vnameInternGlobalLiteral("ControlTag"));

    // load the menu's etc to get ptrs for access privilege checking
    vmenubar *menubar = vmenubar::RequestSubLoad (this, vnameInternGlobalLiteral ("MenuBarTag"));
    // the file menu and it's items
    vmenu *fileMenu = vmenu::RequestSubLoad (menubar, vnameInternGlobalLiteral ("FileMenuTag"));
    vmenuItem *newItem = vmenuItem::RequestSubLoad (fileMenu, 
						    vnameInternGlobalLiteral ("NewMenuTag"));
    vmenuItem *openItem = vmenuItem::RequestSubLoad (fileMenu, 
						    vnameInternGlobalLiteral ("OpenMenuTag"));
    vmenuItem *editItem = vmenuItem::RequestSubLoad (fileMenu, 
						    vnameInternGlobalLiteral ("EditMenuTag"));
    vmenuItem *saveItem = vmenuItem::RequestSubLoad (fileMenu, 
						    vnameInternGlobalLiteral ("SaveMenuTag"));
    vmenuItem *saveasItem = vmenuItem::RequestSubLoad (fileMenu, 
						    vnameInternGlobalLiteral ("SaveAsMenuTag"));
    vmenuItem *deleteItem = vmenuItem::RequestSubLoad (fileMenu, 
						    vnameInternGlobalLiteral ("DeleteMenuTag"));
    vmenuItem *setfileItem = vmenuItem::RequestSubLoad (fileMenu, 
						    vnameInternGlobalLiteral ("SetFileMenuTag"));
    vmenuItem *printItem = vmenuItem::RequestSubLoad (fileMenu, 
						    vnameInternGlobalLiteral ("PrintMenuTag"));
    vmenuItem *prefItem = vmenuItem::RequestSubLoad (fileMenu, 
						    vnameInternGlobalLiteral ("PrefMenuTag"));
    vmenuItem *saveprefItem = vmenuItem::RequestSubLoad (fileMenu, 
						    vnameInternGlobalLiteral ("SavePrefMenuTag"));
    // the view menu and it's items
    vmenu *viewMenu = vmenu::RequestSubLoad (menubar, vnameInternGlobalLiteral ("ViewMenuTag"));
    vmenuItem *matchItem = vmenuItem::RequestSubLoad (viewMenu, 
						    vnameInternGlobalLiteral ("MatchResMenuTag"));
    vmenuItem *filealmItem = vmenuItem::RequestSubLoad (viewMenu, 
						    vnameInternGlobalLiteral ("FileAlmMenuTag"));
    vmenuItem *filemaxItem = vmenuItem::RequestSubLoad (viewMenu, 
						    vnameInternGlobalLiteral ("FileMaxMenuTag"));
    vmenuItem *fileminItem = vmenuItem::RequestSubLoad (viewMenu, 
						    vnameInternGlobalLiteral ("FileMinMenuTag"));
    vmenuItem *substItem = vmenuItem::RequestSubLoad (viewMenu, 
						    vnameInternGlobalLiteral ("SubstMenuTag"));
    // the search menu and it's items
    vmenu *searchMenu = vmenu::RequestSubLoad (menubar, vnameInternGlobalLiteral ("SearchMenuTag"));
    vmenuItem *alarmItem = vmenuItem::RequestSubLoad (searchMenu, 
						    vnameInternGlobalLiteral ("AlarmMenuTag"));
    vmenuItem *maxItem = vmenuItem::RequestSubLoad (searchMenu, 
						    vnameInternGlobalLiteral ("MaxMenuTag"));
    vmenuItem *minItem = vmenuItem::RequestSubLoad (searchMenu, 
						    vnameInternGlobalLiteral ("MinMenuTag"));
    vmenuItem *nextItem = vmenuItem::RequestSubLoad (searchMenu, 
						    vnameInternGlobalLiteral ("NextMenuTag"));
    vmenuItem *prevItem = vmenuItem::RequestSubLoad (searchMenu, 
						    vnameInternGlobalLiteral ("PrevMenuTag"));
    // the notes menu and it's items
    vmenu *notesMenu = vmenu::RequestSubLoad (menubar, vnameInternGlobalLiteral ("NotesMenuTag"));
    vmenuItem *readItem = vmenuItem::RequestSubLoad (notesMenu, 
						    vnameInternGlobalLiteral ("ReadMenuTag"));
    vmenuItem *addItem = vmenuItem::RequestSubLoad (notesMenu, 
						    vnameInternGlobalLiteral ("AddMenuTag"));
    // the help menu 
    helpMenu = vmenu::RequestSubLoad (menubar, vnameInternGlobalLiteral ("HelpMenu"));
    vdialog::LoadInit(resource);

    // establish privileges
    int	allDisabled;
    // init privilege checking, see .../share/src/libtis/OperPrvtrk.c
    if ((InitOperPriv (NULL)) != PRIVOK)
    {
	infoNotice (NULL, (unsigned char *)"Trendvu error",
		    (unsigned char *)"Failed to init Operator privileges.");
	loadiniterror = vTRUE;
	return;
    }
    // check for file menu privileges
    if ((OperPriv (NULL, "*MMI_TRENDVU_FILE_NEW")) == 0)
    {
	newItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_FILE_OPEN")) == 0)
    {
	openItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_FILE_EDIT")) == 0)
    {
	editItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_FILE_SAVE")) == 0)
    {
	saveItem -> Disable ();
	saveasItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_FILE_DELETE")) == 0)
    {
	deleteItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_FILE_PRINT")) == 0)
    {
	printItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_FILE_SETFILE")) == 0)
    {
	setfileItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_FILE_PREFERENCES")) == 0)
    {
	prefItem -> Disable ();
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_FILE_SAVEPREFS")) == 0)
    {
	saveprefItem -> Disable ();
    }
    // check for view menu privileges
    allDisabled = 0;
    if ((OperPriv (NULL, "*MMI_TRENDVU_VIEW_MATCHRES")) == 0)
    {
	matchItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_VIEW_FILEALARM")) == 0)
    {
	filealmItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_VIEW_FILEMAX")) == 0)
    {
	filemaxItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_VIEW_FILEMIN")) == 0)
    {
	fileminItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_VIEW_SUBSTRINGS")) == 0)
    {
	substItem -> Disable ();
	allDisabled++;
    }
    if (allDisabled == 5)
    {
	viewMenu -> Disable ();
    }
    // check for search menu privileges
    allDisabled = 0;
    if ((OperPriv (NULL, "*MMI_TRENDVU_SEARCH_ALARMSTATUS")) == 0)
    {
	alarmItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_SEARCH_FINDMAX")) == 0)
    {
	maxItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_SEARCH_FINDMIN")) == 0)
    {
	minItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_SEARCH_FINDNEXT")) == 0)
    {
	nextItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_SEARCH_FINDPREVIOUS")) == 0)
    {
	prevItem -> Disable ();
	allDisabled++;
    }
    if (allDisabled == 5)
    {
	searchMenu -> Disable ();
    }
    // check for notes menu privileges
    allDisabled = 0;
    if ((OperPriv (NULL, "*MMI_TRENDVU_NOTES_READ")) == 0)
    {
	readItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_TRENDVU_NOTES_ADD")) == 0)
    {
	addItem -> Disable ();
	allDisabled++;
    }
    if (allDisabled == 2)
    {
	notesMenu -> Disable ();
    }
    // close privilege checking, see .../share/src/libtis/OperPrvtrk.c
    CloseOperPriv ();

    cursor = 0;

    // establish color values
    GetPreference (Colors, &Back, dbfilename);
    for (i = 0; i < 8; i++)
    {
	vcolors[i].SetRGB ((Red(Colors[i]) * (vcolorCOMPONENT_MAX + 1))/256,
			   (Green(Colors[i]) * (vcolorCOMPONENT_MAX + 1))/256,
			   (Blue(Colors[i]) * (vcolorCOMPONENT_MAX + 1))/256);
    }
    vback.SetRGB ((Red(Back) * (vcolorCOMPONENT_MAX + 1))/256,
		  (Green(Back) * (vcolorCOMPONENT_MAX + 1))/256,
		  (Blue(Back) * (vcolorCOMPONENT_MAX + 1))/256);
    trendGraph->SetPrefs (vcolors, vback);
    trendHiLimits->SetType (CTL_HILIMIT);
    trendHiLimits->SetPrefs (vcolors, vback);
    trendLoLimits->SetType (CTL_LOLIMIT);
    trendLoLimits->SetPrefs (vcolors, vback);
    trendDetail->SetPrefs (vcolors, vback);
    trendHiLimits->SetParent (this);
    trendLoLimits->SetParent (this);
    trendGraph->SetParent (this);
    trendDetail->SetParent (this);
    trendControls->SetParent (this);

    archTables = NULL;
    textshown = vFALSE;
    // start timer
    timer = new tvTimer;
    timer->SetParent (this);
    timer->StartTimer (1, 0);
}

/*********
 *  TrendvuDialog::ObserveWindow()
 *
 *  The dialog notification procedure gets called whenever any 
 *  major event applies to the dialog as a whole.  We need to
 *  look for the event which tells us that the dialog has
 *  closed.
 *
 **********/
void TrendvuDialog::ObserveWindow (vevent *event)
{
    if (event->GetType() == veventWINDOW_STATE && event->IsClose ())
    {
	vevent::StopProcessing ();
	if (timer)
	    delete timer;
    }
}

void TrendvuDialog::NewView (char *trendsetname)
{
    char    tempstring [132];

    // if text is displayed, close the window
    if (textshown)
    {
	textDialog->Close ();
    }
    textshown = 0;

    // read in trendset here
    trendSet.ChangeSet (trendsetname);

    // notify all views
    trendGraph->SetTrendSet (&trendSet);
    trendHiLimits->SetTrendSet (&trendSet);
    trendLoLimits->SetTrendSet (&trendSet);
    trendDetail->SetTrendSet (&trendSet);
    // setTrendSet in trendControls will always set textshown toggle to 0
    trendControls->SetTrendSet (&trendSet);

    // set position and draw
    trendGraph->SetTheCursor (cursor);
    trendDetail->SetTheCursor (cursor);
    trendControls->SetTheCursor (cursor);
    if (textshown)
	textDialog->SetTheCursor (cursor);

    // adjust title of dialog
    sprintf (tempstring, "%s %s",
	 trendsetname,
	 trendAccess?"":"No realtime values");
    SetTitle ((unsigned char *)tempstring);
}

int TrendvuDialog::MenuCommand (int selector)
{
    int i;
    int cnt;
    char **trendnames;
    char *cp;
    long minrate;
    long srchtime;
    long adjustment;
    int nc;

    switch ( selector )
    {
    case MENU_NEW:
	if (editDialog->Alert (NULL) == RESULT_OK)
	{
	    NewView (editDialog->ResultString ());
	}
	return (TRUE);
    case MENU_OPEN:
	cnt = GetTrendCount ();
	trendnames = (char **)calloc (cnt + 1, sizeof(char *));   
	if (cp = FirstTrend ())
	{
	    i = 0;
	    while (cp && i < cnt)
	    {
		trendnames[i] = (char *)calloc (1, strlen (cp)+1);
		if (trendnames[i])
		{
		    strcpy (trendnames [i], cp);
		    i++;
		}
		cp = NextTrend ();
	    }
	    if (listDialog->Alert ("Open Trend Set", trendnames) == RESULT_OK)
	    {
		NewView (listDialog->ResultString ());
	    }
	    for (i = 0; i < cnt; i++)
	    {
		free (trendnames[i]);
	    }
	}
	free ((char *)trendnames);
	return (TRUE);
    case MENU_EDIT:
	if (editDialog->Alert (&trendSet) == RESULT_OK)
	{
	    NewView (editDialog->ResultString ());
	}
	return (TRUE);
    case MENU_SAVE:
        trendSet.SaveSet (trendSet.setname, FALSE);
	return (TRUE);
    case MENU_SAVE_AS:
	if (saveAsDialog->Alert (trendSet.setname) == RESULT_OK)
	{
	    trendSet.SaveSet (saveAsDialog->ResultString (), saveAsDialog->ConvertAbs ());
	}
	return (TRUE);
    case MENU_DELETE:
	cnt = GetTrendCount ();
	trendnames = (char **)calloc (cnt + 1, sizeof(char *));   
	if (cp = FirstTrend ())
	{
	    i = 0;
	    while (cp && i < cnt)
	    {
		trendnames[i] = (char *)calloc (1, strlen (cp)+1);
		if (trendnames[i])
		{
		    strcpy (trendnames [i], cp);
		    i++;
		}
		cp = NextTrend ();
	    }
	    if (listDialog->Alert ("Delete Trend Set", trendnames) == RESULT_OK)
	    {
		DelTrendInfo (listDialog->ResultString ());
	    }
	    for (i = 0; i < cnt; i++)
	    {
		free (trendnames[i]);
	    }
	}
	free ((char *)trendnames);
	return (TRUE);
    case MENU_PRINT:
	{
	    vprint *printer = vprint::CreatePreferred();

	    vdraw::GSave();
	    printer->OpenDocument();
	    printer->OpenPage();

	    vdraw::SetColor(DetermineBackground());
	    vdraw::RectFill(0, 0, GetWidth(), GetHeight());
	    vdraw::SetColor(DetermineForeground());
	    vdraw::SetFont(DetermineFont());
	    vdraw::SetLineWidth(DetermineLineWidth());
	    Draw();

	    printer->ClosePage();
	    printer->CloseDocument();
	    vdraw::GRestore();

	    delete printer;
	}
	return (TRUE);
    case MENU_SETFILE:
	// Ask for file Info
	char currentfile [200];
	char *pathptr;
	char *nameptr;

	GetDefaultFile (&pathptr, &nameptr);
	strcpy (currentfile, pathptr);
	strcat (currentfile, nameptr);
	if (valueDialog->Alert ("Enter archive filename", currentfile) == RESULT_OK)
	{
	    strncpy (currentfile, valueDialog->ResultString (), sizeof(currentfile));
	    // split result into path and name parts
	    nameptr = currentfile;
	    while (pathptr = strchr(nameptr, '\\'))
		nameptr = pathptr+1;
	    pathptr = currentfile;
		// move filename and null terminate the path
	    memcpy (nameptr+1, nameptr, strlen(nameptr));
	    *nameptr = 0;
	    SetDefaultFile (pathptr, nameptr+1);
	}
	return TRUE;
    case MENU_PREF:
	if (prefsDialog->Alert (vcolors, vback) == RESULT_OK)
	{
	    unsigned int r, g, b;

	    prefsDialog->ResultValues (vcolors, &vback);
	    for (i = 0; i < 8; i++)
	    {
		vcolors[i].GetRGB (&r, &g, &b);
		Colors[i] = Color (r*256/(vcolorCOMPONENT_MAX+1),
				   g*256/(vcolorCOMPONENT_MAX+1),
				   b*256/(vcolorCOMPONENT_MAX+1));
	    }
	    vback.GetRGB (&r, &g, &b);
	    Back = Color (r*256/(vcolorCOMPONENT_MAX+1),
			  g*256/(vcolorCOMPONENT_MAX+1),
			  b*256/(vcolorCOMPONENT_MAX+1));

	    trendGraph->SetPrefs (vcolors, vback);
	    trendHiLimits->SetPrefs (vcolors, vback);
	    trendLoLimits->SetPrefs (vcolors, vback);
	    trendDetail->SetPrefs (vcolors, vback);
	}
	return (TRUE);
    case MENU_SAVE_PREF:
        PutPreference (Colors, Back, dbfilename);
	return (TRUE);
    case MENU_EXIT:
	vevent::StopProcessing ();
	if (timer)
	    delete timer;
	timer = NULL;
	return (TRUE);
    case MENU_MATCHRES:
	minrate = C_MAXRES;
        for (i = 0; i < trendSet.numTrends; i++)
        {
            if (trendSet.specs[i][0]
	      && (minrate > trendSet.trends[i].dbrate))
	    {
                minrate = trendSet.trends[i].dbrate;
	    }
        }
	if (minrate == 0)
	    minrate = 1;
	adjustment = (cursor * trendSet.zoom) - (256 * minrate);
	cursor = 256;
	trendSet.zoom = minrate;
	trendSet.UpdateSet (adjustment, trendSet.zoom);
	trendGraph->Redisplay ();
	trendDetail->Redisplay ();
	trendGraph->SetTheCursor (cursor);
	trendDetail->SetTheCursor (cursor);
	trendControls->SetTheCursor (cursor);
	if (textshown)
	    textDialog->SetTheCursor (cursor);
	return (TRUE);
    case MENU_FILE_ALM:
	if (searchDialog->Alert (&trendSet) == RESULT_OK)
	{
	    lastSearch = *searchDialog->ResultStruct();
	    srchtime = dbsearch (trendSet.trends [lastSearch.type [0] - '0'].pointname,
				 trendSet.trends [lastSearch.type [0] - '0'].dbreduc,
    				 trendSet.trends [lastSearch.type [0] - '0'].starttime
					+ trendSet.zoom * cursor,
	    			 &lastSearch);
	    if (srchtime != -1)
	    {
		adjustment = srchtime - trendSet.trends [lastSearch.type [0] - '0'].starttime;
		trendSet.UpdateSet (adjustment - (256 * trendSet.zoom), trendSet.zoom);
		cursor = 256;
		trendGraph->Redisplay ();
		trendDetail->Redisplay ();
		trendGraph->SetTheCursor (cursor);
		trendDetail->SetTheCursor (cursor);
		trendControls->SetTheCursor (cursor);
		if (textshown)
		    textDialog->SetTheCursor (cursor);
	    }
	}
	return (TRUE);
    case MENU_FILE_MAX:
    case MENU_FILE_MIN:
	if (selectDialog->Alert (&trendSet) == RESULT_OK)
	{
	    lastSearch = *selectDialog->ResultStruct();
	    if (selector == MENU_FILE_MAX)
		lastSearch.type [2] = '+';
	    else
		lastSearch.type [2] = '-';
	    srchtime = dbsearch (trendSet.trends [lastSearch.type [0] - '0'].pointname,
				 trendSet.trends [lastSearch.type [0] - '0'].dbreduc,
    				 trendSet.trends [lastSearch.type [0] - '0'].starttime
	    					+ trendSet.zoom * cursor,
	    			 &lastSearch);
	    if (srchtime != -1)
	    {
		adjustment = srchtime - trendSet.trends [lastSearch.type [0] - '0'].starttime;
		trendSet.UpdateSet (adjustment - (256 * trendSet.zoom), trendSet.zoom);
		cursor = 256;
		trendGraph->Redisplay ();
		trendDetail->Redisplay ();
		trendGraph->SetTheCursor (cursor);
		trendDetail->SetTheCursor (cursor);
		trendControls->SetTheCursor (cursor);
		if (textshown)
		    textDialog->SetTheCursor (cursor);
	    }
	}
	return (TRUE);
#ifdef COMPLETE
    case MENU_SUBST:
        MacrosDialog *theMacrosDialog = new MacrosDialog (document);
        theMacrosDialog->create (theTrendShell->trendform, "Macros");
        theMacrosDialog->alert (Cancel);
        delete theMacrosDialog;
    	return (TRUE);
#endif COMPLETE
    case MENU_ALM:
	if (searchDialog->Alert (&trendSet) == RESULT_OK)
	{
	    lastSearch = *searchDialog->ResultStruct();
	    nc = trendSet.trends[lastSearch.type [0] - '0'].Search (&lastSearch, cursor);
	    if (nc != -1)
            {
		cursor = nc;
		trendGraph->Redisplay ();
		trendDetail->Redisplay ();
		trendGraph->SetTheCursor (cursor);
		trendDetail->SetTheCursor (cursor);
		trendControls->SetTheCursor (cursor);
		if (textshown)
		    textDialog->SetTheCursor (cursor);
	    }
	}
	return (TRUE);
    case MENU_MAX:
    case MENU_MIN:
	if (selectDialog->Alert (&trendSet) == RESULT_OK)
	{
	    lastSearch = *selectDialog->ResultStruct();
	    if (selector == MENU_FILE_MAX)
		lastSearch.type [2] = '+';
	    else
		lastSearch.type [2] = '-';
	    nc = trendSet.trends[lastSearch.type [0] - '0'].Search (&lastSearch, cursor);
	    if (nc != -1)
            {
		cursor = nc;
		trendGraph->Redisplay ();
		trendDetail->Redisplay ();
		trendGraph->SetTheCursor (cursor);
		trendDetail->SetTheCursor (cursor);
		trendControls->SetTheCursor (cursor);
		if (textshown)
		    textDialog->SetTheCursor (cursor);
	    }
	}
	return (TRUE);
    case MENU_NEXT:
	lastSearch.type [1] = 'C';
	nc = trendSet.trends[lastSearch.type [0] - '0'].Search (&lastSearch, cursor);
	if (nc != -1)
        {
	    cursor = nc;
	    trendGraph->Redisplay ();
	    trendDetail->Redisplay ();
	    trendGraph->SetTheCursor (cursor);
	    trendDetail->SetTheCursor (cursor);
	    trendControls->SetTheCursor (cursor);
	    if (textshown)
		textDialog->SetTheCursor (cursor);
        }
    	return (TRUE);
    case MENU_PREV:
	lastSearch.type [1] = 'P';
	nc = trendSet.trends[lastSearch.type [0] - '0'].Search (&lastSearch, cursor);
	if (nc != -1)
        {
	    cursor = nc;
	    trendGraph->Redisplay ();
	    trendDetail->Redisplay ();
	    trendGraph->SetTheCursor (cursor);
	    trendDetail->SetTheCursor (cursor);
	    trendControls->SetTheCursor (cursor);
	    if (textshown)
		textDialog->SetTheCursor (cursor);
        }
    	return (TRUE);
#ifdef COMPLETE
    case MENU_READ:
        ReadNotes ();
	RelNotes ();
    	return (TRUE);
    case MENU_ADD:
        for (i = 0; i < nspecs; i++)
	{
            names [i] = (char *)malloc (C_PTNMLEN);
	    if (names [i])
		strncpy (names[i], Points[i].PointName(), C_PTNMLEN);
	}
        for (; i < C_TRENDPTS; i++)
            names [i] = NULL;
        theNoteSelect = new NoteTrendSelect (document, names);
        theNoteSelect->create (theTrendShell->trendform, "SelectNote");
        theNoteSelect->alert (Cancel);
        char *selpoint = theNoteSelect->Result();
        for (i = 0; i < C_TRENDPTS; i++)
	{
	    if (names [i])
		free (names[i]);
	}
        if (selpoint)
        {
            theNoteAdd = new NoteAdd (document,
                                      selpoint+3,
                                      Points[atoi (selpoint)-1].PointTick(htcursor));
            theNoteAdd->create (theTrendShell->trendform, "AddNote");
            theNoteAdd->alert (Cancel);
            delete theNoteAdd;
        }
        delete theNoteSelect;
        RelNotes ();
        theTrendShell->DrawView ();
    	return (TRUE);
#endif COMPLETE
	case MENU_HELP:
	default:
	    return(FALSE);
    }
}

void TrendvuDialog::NewCursor (int newloc)
{
    if (newloc < 0)
	newloc = 0;
    else if (newloc > 511)
	newloc = 511;
    cursor = newloc;
    trendGraph->SetTheCursor (cursor);
    trendDetail->SetTheCursor (cursor);
    trendControls->SetTheCursor (cursor);
    if (textshown)
	textDialog->SetTheCursor (cursor);
}

void TrendvuDialog::NewRange (int lpos, int rpos)
{
    long adjustment = lpos * trendSet.zoom;
    int newzoom = (rpos - lpos) * trendSet.zoom / 512;
    if (newzoom == 0)
	newzoom = 1;
    trendSet.zoom = newzoom;
    cursor = 0;
    trendSet.UpdateSet (adjustment, trendSet.zoom);
    trendGraph->Redisplay ();
    trendDetail->Redisplay ();
    trendControls->SetTheCursor (cursor);
    if (textshown)
	textDialog->SetTheCursor (cursor);
}

void TrendvuDialog::DetailControlClick (int type, int ix)
{
    if (type == ENB_CLICK)
    {
	trendSet.trends[ix].enable = ! trendSet.trends[ix].enable;
	trendHiLimits->Redisplay ();
	trendLoLimits->Redisplay ();
	if (textshown)
	    textDialog->SetTrendSet (&trendSet);
    }
    else
    {
	trendSet.trends[ix].ctlEnable = ! trendSet.trends[ix].ctlEnable;
    }
    trendGraph->Redisplay ();
    trendDetail->Redisplay ();
}

void TrendvuDialog::NewLimit (int id, int row)
{
    char    valText [80];
    double  value;
    // popup value dialog, and get new value
    if (id == CTL_HILIMIT)
    {
	sprintf (valText, "%9s", trendSet.trends[row].PointHiLim ());
	if (valueDialog->Alert ("Enter new upper display value", valText) == RESULT_OK)
	{
	    value = atof (valueDialog->ResultString ());
	    trendSet.trends[row].dispHi = value;
	}
    }
    else if (id == CTL_LOLIMIT)
    {
	sprintf (valText, "%9s", trendSet.trends[row].PointLoLim ());
	if (valueDialog->Alert ("Enter new lower display value", valText) == RESULT_OK)
	{
	    value = atof (valueDialog->ResultString ());
	    trendSet.trends[row].dispLo = value;
	}
    }
    trendHiLimits->Redisplay ();
    trendLoLimits->Redisplay ();
    trendGraph->Redisplay ();
}

void TrendvuDialog::ControlsClick (int type)
{
    int	    newzoom = 0;
    long    adjustment = 0;
    char    valText [80];

    switch (type)
    {
    case tcTIMEDOWN:
	adjustment = -(trendSet.zoom * (long)(C_TRENDSAMPLES / 4));
	break;
    case tcTIME:
	// Ask for time adjustment!
	if (valueDialog->Alert ("Enter time offset (+ or -)", "0d 00:00:00") == RESULT_OK)
	{
	    adjustment = SpanVal(valueDialog->ResultString ());
	}
	break;
    case tcTIMEUP:
	adjustment = trendSet.zoom * (long)(C_TRENDSAMPLES / 4);
	break;
    case tcCURSORDOWN:
	if (cursor == 0)
	    adjustment = - trendSet.zoom;
	else
	    cursor--;
	break;
    case tcCURSOR:
	// Ask for cursor adjustment!
	sprintf (valText, "%d", cursor);
	if (valueDialog->Alert ("Enter new cursor position", valText) == RESULT_OK)
	{
	    int newc = atoi (valueDialog->ResultString ());
	    if (0 <= newc && newc < 512)
		cursor = newc;
	}
	break;
    case tcCURSORUP:
	if (cursor == 511)
	    adjustment = trendSet.zoom;
	else
	    cursor++;
	break;
    case tcZOOMDOWN:
	newzoom = trendSet.zoom / 2;
	if (newzoom == 0)
	    newzoom = 1;
	adjustment = (cursor * trendSet.zoom) - (256 * newzoom);
	cursor = 256;
	break;
    case tcZOOM:
	// Ask for zoom adjustment!
	sprintf (valText, "%d", trendSet.zoom);
	if (valueDialog->Alert ("Enter new data rate (in seconds)", valText) == RESULT_OK)
	{
	    newzoom = atoi (valueDialog->ResultString ());
	    if (newzoom == 0)
		newzoom = 1;
	    adjustment = (cursor * trendSet.zoom) - (256 * newzoom);
	    cursor = 256;
	}
	break;
    case tcZOOMUP:
	newzoom = trendSet.zoom * 2;
	adjustment = (cursor * trendSet.zoom) - (256 * newzoom);
	cursor = 256;
	break;
    }
    if (adjustment || newzoom)
    {
	if (newzoom)
	{
	    trendSet.zoom = newzoom;
	}
	trendSet.UpdateSet (adjustment, trendSet.zoom);
	trendGraph->Redisplay ();
	trendDetail->Redisplay ();
	if (textshown)
	    textDialog->SetTrendSet (&trendSet);
    }
    trendGraph->SetTheCursor (cursor);
    trendDetail->SetTheCursor (cursor);
    trendControls->SetTheCursor (cursor);
    if (textshown)
	textDialog->SetTheCursor (cursor);
}

void TrendvuDialog::ToggleClick (int type, int value)
{
    switch (type)
    {
    case tcTEXTTOGGLE:
	if (!textshown)
	{
	    if (trendSet.numTrends == 0)
	    {
		infoNotice (NULL, (unsigned char *)"Trendvu error",
		    (unsigned char *)"No points for text display.");
		trendControls->SetTextIndicator (0);
		return;
	    }
	    textDialog->SetTrendSet (&trendSet);
	    textDialog->Open ();
	    textDialog->SetTheCursor (cursor);
	    textshown = vTRUE;
	}
	else
	{
	    textDialog->Close ();
	    textshown = vFALSE;
	}
	break;
    case tcAUTOTOGGLE:
	if (value)
	{
	    trendSet.autoupdate = vTRUE;
	    lastcurrent = time(NULL);
	}
	else
	{
	    trendSet.autoupdate = vFALSE;
	}
	break;
    }
}

void TrendvuDialog::TextClosed (void)
{
    // reset control toggle
    trendControls->SetTextIndicator (0);
    textshown = vFALSE;
}

void TrendvuDialog::Tick (void)
{
    if (trendSet.autoupdate)
    {
	if (time(NULL) >= lastcurrent + trendSet.zoom)
	{
	    lastcurrent = time(NULL);
	    trendSet.UpdateSet (trendSet.zoom, trendSet.zoom);
	    trendGraph->Redisplay ();
	    trendDetail->Redisplay ();
	}
    }

    /* Show dialog for missing data tables */
    if (archTables)
    {
	static char offlinemsg [1024];
	sprintf (offlinemsg,
		 "The following tables are not currently online:\n%s\nWould you like to issue a mount request?",
		 archTables);
	if (cautionNotice (NULL, (unsigned char *)"Trendvu error",
		    (unsigned char *)"No points for text display."))
	{
 //	    sprintf (offlinemsg, "csh -c \"archiveMountRequest %s\" &", archTables);
//	    system (offlinemsg);
	}
    }
    archTables = NULL;
}

void TrendvuDialog::BuildAboutMenu (vresource tisres)
{
    tisAbout = new TisAbout (tisres, (unsigned char *)"Trendvu version 6.0");
    AboutMenuItem *aboutMenu = tisAbout -> GetAboutMenuItem ();
    helpMenu -> AppendItem (aboutMenu);
    tisAbout -> SetAboutTitle ((unsigned char *)"About TIS4000 Trendvu");
}

int main(int argc, char *argv[])
{
    tisOpenLog ("trendvu", 6, 2, NULL, NULL);

    vstartup(argc, argv);

    // get the tisgalaxy resource
    vresource tisres = startTisGalaxy ();

    // create a command space and store it in our application
    vapplication *application = vapplication::GetCurrent();
    vcommandSpace *space = new vcommandSpace;
    // associate a selector command with each command tag that can be issued.
    vdict *dict = 
    mySelectorCommand::CreateDictWithInitializersOfKind
	(mySelectorCommand::StaticGetKind(), selCmdBindings, SELECTOR_COUNT);
    space->AddDictOwned(vname_Command, dict);
    application->SetSpaceOwned(space);

    // Get the resource file and load the application dialog
    vresource res = vapplication::GetCurrent()->GetResources();
    vresource appDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("TrendvuTag"));
    maindialog = new TrendvuDialog(appDlgRes);
    if (maindialog -> loadiniterror == vTRUE)
    {
	delete maindialog;
	stopTisGalaxy (tisres);
	exit (EXIT_FAILURE);
    }
    maindialog -> BuildAboutMenu (tisres);

    vresource editDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("EditDialogTag"));
    editDialog = new TrendEditDialog(editDlgRes);
    vresource listDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("ListDialogTag"));
    listDialog = new TrendListDialog(listDlgRes);
    vresource prefsDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("PrefsDialogTag"));
    prefsDialog = new TrendPrefsDialog(prefsDlgRes);
    vresource macroDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("MacroDialogTag"));
    macroDialog = new TrendMacroDialog(macroDlgRes);
    vresource valueDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("ValueDialogTag"));
    valueDialog = new TrendValueDialog(valueDlgRes);
    vresource saveAsDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("SaveAsDialogTag"));
    saveAsDialog = new TrendSaveAsDialog(saveAsDlgRes);
    vresource searchDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("SearchDialogTag"));
    searchDialog = new TrendSearchDialog(searchDlgRes);
    vresource selectDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("SelectDialogTag"));
    selectDialog = new TrendSelectDialog(selectDlgRes);

    vresource textDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("TextDialog"));
    textDialog = new tvTextDialog (textDlgRes);
    textDialog->SetParent (maindialog);


    // read and install startup argument values

    // Open the dialog and start processing
    maindialog->Open();
    // so main will iconify properly to task bar in NT
    maindialog->AddStyle (vwindowSTYLE_GLOBAL);

    vevent::Process();

    delete maindialog -> tisAbout;
    delete maindialog;
    delete editDialog;
    delete listDialog;
    delete prefsDialog;
    delete macroDialog;
    delete valueDialog;
    delete saveAsDialog;
    delete searchDialog;
    delete selectDialog;
    delete textDialog;

    stopTisGalaxy (tisres);

    exit(EXIT_SUCCESS);
    return EXIT_FAILURE;
}

