/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/* 
 *  NOTE: look to parent file for defines of SOCKET etc.
*/
int WaitTillAvailable(SOCKET alarmMgrSocket);
int SharedMemorySegmentAvailable(SOCKET alarmMgrSocket);
int  ack_point(SOCKET alarmmgrSocket, char *pointname, int ackgroup, char *groupname, int debug);


/**********************************************************************************
  create a socket and connect to alarmmgr existing "message" socket.
**********************************************************************************/
int create_and_connect_socket(SOCKET *alarmmgrSocket)
{
    struct hostent	*hostval;
    int			optionValue = 1;
    struct sockaddr_in	setPort;
    short		msgPort = MSGPORT;
    int			err;
#ifdef	WIN32
    WSADATA	wsaData;
    WORD	wVersionRequested = MAKEWORD (1, 1);

    err = WSAStartup (wVersionRequested, &wsaData);
    if (err != 0)
    {
	tisSysLog (0,0, "can't start windows socket, %d\n", err);
	return 0;
    }
#endif	WIN32
 
    /* create the socket */
    if ((*alarmmgrSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
    {
#ifdef	WIN32
	err = WSAGetLastError ();
#else
	err = errno;
#endif	WIN32
	tisSysLog (0,0, "can't create socket, %d", err);
	return 0;
    }
    /* make a connection with the alarmmgr existing socket */
    memset ((char*) &setPort, 0, sizeof(setPort));
    setPort.sin_family = AF_INET;
    hostval = gethostbyname ("localhost");
    if (hostval == NULL) 
    {
#ifdef	WIN32
	err = WSAGetLastError ();
	closesocket (*alarmmgrSocket);
#else
	err = errno;
	close (*alarmmgrSocket);
#endif	WIN32
	tisSysLog (0,0, "can't translate localhost, %d", err);
	return 0;
    }
    setPort.sin_addr.s_addr =**((long **)(hostval -> h_addr_list));
    setPort.sin_port = htons(msgPort);
    if (connect(*alarmmgrSocket, (struct sockaddr *) &setPort, sizeof (setPort)) == SOCKET_ERROR) 
    {
#ifdef	WIN32
	err = WSAGetLastError ();
	closesocket (*alarmmgrSocket);
#else
	err = errno;
	close (*alarmmgrSocket);
#endif	WIN32
	tisSysLog (0,0, "can't connect to TCP port, %d", err);
	return 0;
    }
    /* set the sockets blocking option */
#ifdef	WIN32
    if (ioctlsocket(*alarmmgrSocket, FIONBIO, &optionValue) != 0)
    {
	err = WSAGetLastError ();;
	closesocket(*alarmmgrSocket);
#else
    if (ioctl(*alarmmgrSocket, FIONBIO, &optionValue) == -1) 
    {
	err = errno;
	close(*alarmmgrSocket);
#endif	WIN32
	tisSysLog (0,0, "can't set socket nonblocking, %d", err);
	return 0;
    }
    return 1;
}
/*******************************************************************************
   read alarmMgr socket
*******************************************************************************/
int GetAlarmMgrMessage(SOCKET alarmMgrSocket, int timeRemaining)
{
    int                 update = 0;
    TypeOnlyMessage    *alarmMgrMsg;
    char                message[sizeof(alarmMgrMsg)];
    time_t              epoch;
    time_t              stopEpoch = time(&epoch) + timeRemaining;
    fd_set              chanmask;
    struct timeval      timeout;
    int                 numbytes;
    int			err;

     timeout.tv_sec =  timeRemaining;
     timeout.tv_usec = 0;
     
     alarmMgrMsg = (TypeOnlyMessage *)message;

     while (1) 
     {
	FD_ZERO(&chanmask);
	FD_SET(alarmMgrSocket, &chanmask);
	if (select(alarmMgrSocket+1, &chanmask, NULL, NULL, &timeout) &&
		FD_ISSET(alarmMgrSocket, &chanmask) ) 
	{
	        numbytes = recv(alarmMgrSocket, message, sizeof (alarmMgrMsg), 0);

		if (numbytes ==0 )
		{
		       /* alarmMgr does not exist */
#ifdef	WIN32
		       closesocket (alarmMgrSocket);
#else
		       close (alarmMgrSocket);
#endif	WIN32
		       return 0;
		}
		else if (numbytes == -1)
		{
#ifdef	WIN32
		    err = WSAGetLastError ();
		    if (err != WSAEWOULDBLOCK)
#else
		    err = errno;
		    if (err != EWOULDBLOCK)
#endif	WIN32
		    {
			    tisSysLog (0,0, "receive error from alarm mgr, %d", err);
	       		    return 0;
		    }
		}
		else if ((alarmMgrMsg->msgType == UPDATE) || 
			 (alarmMgrMsg->msgType == UPDATE_NEW_ALARM)) 
		{
		        if ( (timeout.tv_sec = stopEpoch - time(&epoch)) < 1 )
		  		return 1;
			update = 1;
  		} 
        	else if (alarmMgrMsg->msgType == UNAVAILABLE) 
		{

             		/* alarmMgr is reconfiguring the shared memory segment.
                	Send UNAVAILABLE message back to alarmMgr and 
                	wait for an AVAILABLE message before accessing
                	shared memory segment again.                        */

			send(alarmMgrSocket, (char *)alarmMgrMsg, sizeof(alarmMgrMsg), 0);
			if (!WaitTillAvailable(alarmMgrSocket))
				return 0;
		}
	}
	else if (update)
		return 1;
	else
		timeout.tv_sec = timeRemaining;
	
     }  
}
/*******************************************************************************
     reads the alarmMgr socket to determine if alarmMgr has sent out
     an UNAVAILABLE messages. 

     returns 1 when alarmMgr is AVAILABLE again or

     returns 0 if an error has occured which indicates
	that alarmMgr is not communicating.
*******************************************************************************/
int SharedMemorySegmentAvailable(SOCKET alarmMgrSocket)
{
    int                 numbytes;
    TypeOnlyMessage    *alarmMgrMsg;
    char                message[sizeof(alarmMgrMsg)];
    int			err;
     
	alarmMgrMsg = (TypeOnlyMessage *)message;

	numbytes = recv(alarmMgrSocket, message, sizeof(alarmMgrMsg),0);

	if (numbytes == 0) 
	{
	       /* alarmMgr does not exist */
	       return 0;
	}
 	else  if (numbytes == -1)
	{
#ifdef	WIN32
	    err = WSAGetLastError ();
	    if (err != WSAEWOULDBLOCK)
#else
	    err = errno;
	    if (err != EWOULDBLOCK)
#endif	WIN32
	    {
		tisSysLog (0,0, "receive error from alarm mgr, %d", err);
		return 0;
	    }
	    else 
		/* alarmMgr has noting to say */	
		return 1;
	}
	else if (alarmMgrMsg->msgType == UNAVAILABLE)
	{
             /* alarmMgr is reconfiguring the shared memory segment.
                Send UNAVAILABLE message back to alarmMgr */

	    send(alarmMgrSocket, (char *)alarmMgrMsg, sizeof(alarmMgrMsg), 0);

	    /* wait for the available message before continuing */

	    if (WaitTillAvailable(alarmMgrSocket))
		return 1;
	    else
		return 0;

 	}
	else
	    return 1;
}

/********************************************************************
  returns 1 when an AVAILABLE message is received from alarmMgr

  returns 0 if a communication error occures.
********************************************************************/
int WaitTillAvailable(SOCKET alarmMgrSocket)
{
    int			numbytes;
    TypeOnlyMessage	*alarmMgrMsg;
    char		message[sizeof(alarmMgrMsg)];
    int			err;

	alarmMgrMsg = (TypeOnlyMessage *)message;

  	while (1)
	{
	    numbytes = recv(alarmMgrSocket, message, sizeof (alarmMgrMsg), 0);

	    if (numbytes ==0 ) 		
		return 0;	/* alarmMgr does not exist */
	    else  if (numbytes == -1)
	    {
#ifdef	WIN32
		err = WSAGetLastError ();
		if (err != WSAEWOULDBLOCK)
#else
		err = errno;
		if (err != EWOULDBLOCK)
#endif	WIN32
		{
		    tisSysLog (0,0, "receive error from alarm mgr, %d", err);
		    return 0;
		}
	    }
	    else if (alarmMgrMsg->msgType == AVAILABLE)
		return 1;
	}
} 


/************************************************************************************
  send a message to alarmmgr through the opened socket to acknowledge the alarm
************************************************************************************/
int  ack_point(SOCKET alarmmgrSocket, char *pointname, int ackgroup, char *groupname, int debug)
{
    AlarmMsg	controlMsg;
    PvData	*pvdata;
    char	currOperator [OPERATOR_SZ + 1];
    int		err;

    /* build the message for the alarmmgr to acknowledge the alarm */

    memset ((char *) &controlMsg, 0, sizeof (controlMsg));

    controlMsg.msgType = ACK_MSG;

    strcpy(controlMsg.pointName, pointname);

    /* set to 1 is part of an ackgroup acknowledgement
       set to 0 if part of an ackpoint   */

    controlMsg.memberOfGroup = ackgroup;

    if (ackgroup)
	/* when an ack is part of an ackgroup the groupname is expected */
        strcpy(controlMsg.name.groupName, groupname);
    else
    {
	GetOperName (currOperator);
	/* otherwise the operator name is expected */
        strncpy(controlMsg.name.operatorName, currOperator, sizeof (controlMsg.name.operatorName));
    }

    /* determine is the point is valid by checking shared memory */

    if (pvdata = (PvData *)FindPV(NULL, pointname))
    {
	/* send the time of last alarm this is how ack for point
	   is identified. */
    
	controlMsg.timeStamp = pvdata->lastAlarm;

        /* send the message */

        if (send (alarmmgrSocket, (char *)&controlMsg, sizeof(controlMsg), 0) == -1) 
	{
#ifdef	WIN32
	    err = WSAGetLastError ();
	    closesocket(alarmmgrSocket);
#else
	    err = errno;
	    close(alarmmgrSocket);
#endif	WIN32
	    tisSysLog (0,0, "Error sending ACK point message, %d", err);
	    return 0;
        }
	if (debug)
	{
	    tisSysLog (0,0, "sent ack for point %s", pointname);
        }

        return 1;
    }
    if (debug)
    {
	tisSysLog (0,0, "point %s not in configuration", pointname);
    }

    return 0;
}

/***********************************************************************************
  send a message to alarmmgr through the opened socket that a group is being acked.
  alarmMgr uses this message to identifiy that a group is being acked.
  So that a messages can be printed.
***********************************************************************************/
int  ack_group(SOCKET alarmmgrSocket, char *groupname, int debug)

{
    AlarmMsg	controlMsg;
    char	currOperator [OPERATOR_SZ + 1];
    int		err;

    /* build the message for the alarmmgr that a group is being acked */

    memset ((char *) &controlMsg, 0, sizeof (controlMsg));
    controlMsg.msgType = ACK_GROUP_MSG;
    strcpy(controlMsg.pointName, groupname);
    GetOperName (currOperator);
    strncpy(controlMsg.name.operatorName, currOperator, sizeof (controlMsg.name.operatorName));

    if (FindGroup(NULL, groupname))
    {
        /* send the message */

        if (send (alarmmgrSocket, (char *)&controlMsg, sizeof(controlMsg), 0) == -1)
	{
#ifdef	WIN32
	    err = WSAGetLastError ();
	    closesocket (alarmmgrSocket);
#else
	    err = errno;
	    close (alarmmgrSocket);
#endif	WIN32
	    tisSysLog (0,0, "Error sending ACK group message, %d", err);
	    return 0;
        }
	if (debug)
	{
	    tisSysLog (0,0, "sent ack group message for group %s", groupname);
        }

        return 1;
    }
    if (debug)
    {
	tisSysLog (0,0, "group %s not in configuration", groupname);
    }
    return 0;
}
