/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef SOLARIS
#include <sys/types.h>
#include <unistd.h>
#else
#include <strings.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <time.h>
#include <signal.h>


// Functions defined in this module

static	void	scanRules (void);
static	void	readRuleFile (void);
static	void	printError (int lineNum, char *msg);

// External functions

extern "C"
{
#include "error.h"
#include "cadef.h"
#include "msgMon.h"
#include "dbDefs.h"
#ifndef SOLARIS
extern	int	getpid (void);
#endif
extern	void	setupSignalHandlers (void);	
extern	void	serverInit (char *, char *, char *);
}
#include "ChannelList.h"

// Local data

static	int	currentMin;
	int	running;
	int	debugLevel = 0;
static	time_t	ruleFileTime;

// Misc definitions

typedef	char		ChanName [PVNAME_SZ + 1];
#ifdef WIN32
#define	RULE_FILE	"\\tis4000db\\binPvControl.rules"
#else
#define	RULE_FILE	"/tis4000db/binPvControl.rules"
#endif WIN32

// Table of rules

#define	MAX_RULES	1000
enum Operator {AND, OR, THEN, OUTPUT};
struct RuleEntry {
    ChanName	chanName;	// Name of the channel
    int		channelIx;	// Index of channel in the channel table
    int		OMSLchanIx;	// Channel for OMSL field
    int		value;
    Operator	op;
};
static	RuleEntry	rules [MAX_RULES];
static	int		numRules;
	
//----------
main (int argc, char *argv[])

    {

	int	c;
	extern	char	*optarg;
	time_t		clock;
	struct tm	*t;

	tisOpenLog ("binPvControl", 6, 2, NULL, NULL);

	// Process command line options

	while ((c = getopt (argc, argv, "l:dvs")) != -1)
	    switch (c) {
		case 'l':	// Explicitly set debug level
		    debugLevel = atoi (optarg);
		    break;
		case 'd':	// Debugging on
		    debugLevel = 3;
		    break;
		case 'v':	// Verbose mode on
		    debugLevel = 2;
		    break;
		case 's':	// Status mode
		    debugLevel = 1;
		    break;
	    }

#ifdef	WIN32
	serverInit ("binPvControl", "\\tis4000\\tasks\\binPvControl.lock", "lock");
#else
	serverInit ("binPvControl", "/tis4000/tasks/binPvControl.lock", "lock");
#endif	WIN32

	// Set up signal handlers for kill signals

	setupSignalHandlers ();

	// Force initial read of the rule file

	ruleFileTime = 0;

	// Force initial minutely processing

	currentMin = -1;

	// Initialize CA

	if (ca_task_initialize () != ECA_NORMAL) {
	    tisSysLog (0,0, "CA won't initialize\n");
	    exit (-1);
	}

	// Initialize the channel list

	clInitialize ();

	numRules = 0;
	running = 1;
	while (running) {

	    // Scan the rules

	    scanRules ();

	    // Get the current time to determine if it is time to do things.

	    clock = time (0);
	    t = localtime (&clock);

	    if (t->tm_min != currentMin) {

		currentMin = t->tm_min;

		readRuleFile ();

	    }

	    // Wait for any incoming updates

	    ca_pend_io (60.0);

	} while (running);

	// Close all the channels

	clTerminate ();

	// Disconnect from CA

	ca_task_exit ();

	if (debugLevel >= 1)
	    tisSysLog (0,0, "Terminated.\n");

    }

//----------
static void readRuleFile (void)

    {

	struct stat	status;
	FILE		*f;
	char		line [100], chanName [100], *p;
	int		value, lineNum;
	Operator	op;

	// Get the last modification date/time of the file.  
	// If the file has not changed since
	// the last time we loaded it then don't reload.

	stat (RULE_FILE, &status);
	if (ruleFileTime == status.st_mtime)
	    return;
	ruleFileTime = status.st_mtime;

	// Say something if appropriate

	if (debugLevel >= 1)
	    tisSysLog (0,0, "Reading rules file.");

	// The rule file has changed. Close all open channels

	clTerminate ();
	clInitialize ();

	// Read through the rules file and build the rules list

	f = fopen (RULE_FILE, "r");
	if (f == NULL) {
	    tisSysLog (0,0, "Unable to open %s -- program terminated.", RULE_FILE);
	    exit (-1);
	}

	lineNum = 0;
	while (fgets (line, sizeof (line), f)) {

	    lineNum++;

	    // Skip any comments

	    if (line [0] == '#')
		continue;

	    // Get the channel name

	    p = strtok (line, " \t\n");
	    if (p == NULL) continue;
	    strcpy (chanName, p);

	    // Get the channel value

	    p = strtok (NULL, " \t\n");
	    if (p == NULL)
		printError (lineNum, "Invalid value");
	    value = atoi (p);

	    // Get the operator

	    p = strtok (NULL, " \t\n");
	    if (p) {
		if (strcasecmp (p, "and") == 0)
		    op = AND;
		else if (strcasecmp (p, "or")== 0)
		    op = OR;
		else if (strcasecmp (p, "then") == 0)
		    op = THEN;
		else
		    printError (lineNum, "Unknown operator");
	    } else
		op = OUTPUT;


	    if (numRules == MAX_RULES) {
		tisSysLog (0,0, "Too many rules -- terminated.");
		exit (-1);
	    }

	    strcpy (rules [numRules].chanName, chanName);
	    rules [numRules].channelIx = clAdd (chanName);
	    rules [numRules].value = value;
	    rules [numRules].op = op;

	    // If this is an output channel then set up a monitor 
	    // for closed-loop vs. supervisory

	    if (op == OUTPUT) {
		ChanName chan;
		sprintf (chan, "%s.OMSL", chanName);
		rules [numRules].OMSLchanIx = clAdd (chan);
	    }

	    numRules++;
	
	}

	fclose (f);

	// Open all channels and connect

	clConnect ();

	// Say something if appropriate

	if (debugLevel >= 1)
	    tisSysLog (0,0, "Finished reading rules file.");

    }

//----------
static void printError (int lineNum, char *msg)

    {

	tisSysLog (0,0, "Error in binPvControl.rules line %d: %s", lineNum, msg);
	exit (-1);

    }

//----------
static void scanRules (void)

    {

	int		i, boolvar, valid;
	Operator	op;

	if (debugLevel >= 3)
	    tisSysLog (0,0, "** Evaluating rules **\n");

	// Go through the rule list and evaluate each rule

	i = 0;
	while (i < numRules) {

	    // Process the first clause as a special case

	    boolvar = (rules [i].value == clGetValue (rules [i].channelIx));
	    op = rules [i].op;
	    valid = clValid (rules [i].channelIx);
	    if (debugLevel >= 3)
	    {
		if (clValid (rules [i].channelIx))
		    tisSysLog (0,0, "(%s<%d> == %d)", 
			rules [i].chanName, clGetValue (rules [i].channelIx),
			rules [i].value);
		else
		    tisSysLog (0,0, "(%s(invalid) == %d)",
				rules [i].chanName, rules [i].value);
	    }
	    i++;

	    // Process remaining clauses

	    while (op != THEN) {

		switch (op) {
		    case AND:
			boolvar = boolvar && (rules [i].value == clGetValue (rules [i].channelIx));
			if (debugLevel >= 3)
			    tisSysLog (0,0, " AND ");
			break;
		    case OR:
			boolvar = boolvar || (rules [i].value == clGetValue (rules [i].channelIx));
			if (debugLevel >= 3)
			    tisSysLog (0,0, " OR ");
			break;
		}

		if (debugLevel >= 3)
		{
		    if (clValid (rules [i].channelIx))
			tisSysLog (0,0, "(%s<%d> == %d)", 
				rules [i].chanName,
				clGetValue (rules [i].channelIx),
				rules [i].value);
		    else
			tisSysLog (0,0, "(%s(invalid) == %d)", 
				rules [i].chanName, rules [i].value);
		}

		op = rules [i].op;
		valid = valid && clValid (rules [i].channelIx);
		i++;
	    }

	    if (debugLevel >= 3)
	    {
		if (!valid)
		    tisSysLog (0,0, " = Invalid\n");
		else if (clGetValue (rules [i].OMSLchanIx) != CLOSED_LOOP)
		    tisSysLog (0,0, " = Not Changed (%s not closed loop)\n", 
				rules [i].chanName);
		else if (!boolvar)
		    tisSysLog (0,0, " = 0 (no value written to %s)\n", 
				rules [i].chanName);
		else if (clGetValue (rules [i].channelIx) == rules [i].value)
		    tisSysLog (0,0, " = 1 (%s is already %d -- nothing written)\n",
				rules [i].chanName, rules [i].value);
		else
		    tisSysLog (0,0, " = 1 (%d written to %s)\n", 
				rules [i].value, rules [i].chanName);
	    }

	    // If the result is different than the current value of the target then
	    // update the target with the new value.

	    if (valid && boolvar && (clGetValue (rules [i].OMSLchanIx) == CLOSED_LOOP)) {
		if ((clGetValue (rules [i].channelIx) != rules [i].value)) {
		    clPutValue (rules [i].channelIx, rules [i].value);
		    if (debugLevel >= 2)
			tisSysLog (0,0, "%s written to %d\n", 
				rules [i].chanName, rules [i].value);
		}
	    }

	    i++;

	}

    }

