/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <string.h>

#include "dbDefs.h"
#include "dbRecDes.h"
#include "dbRecType.h"
#include "dbFldTypes.h"

/* Functions defined in this module */

static void stripBlanks (char *);

#define	PATH_NAME_SIZE	100
typedef	char	PathName [PATH_NAME_SIZE+1];

FILE		*outFile;
PathName	dbDir;

/*----------*/
main (int argc, char *argv [])

    {

	int		c, i, j;
	extern	char	*optarg;
	PathName	path;
	char		pointType [20], fieldName [6];
	struct fldDes	**fldPtr;
	int		fldType, first;

	tisOpenLog ("makeLinkSpec", 6, 2, NULL, NULL);

	/* Set up the defaults */

#ifdef WIN32
	strcpy (dbDir, "\\tis4000db\\");
#else
	strcpy (dbDir, "/tis4000db/");
#endif WIN32
	outFile = stdout;

	/* Process command line options */

	while ((c = getopt (argc, argv, "p:o:h")) != -1)
	    switch (c) {
		case 'p':	/* Database directory */
		    strcpy (dbDir, optarg);
		    break;
		case 'o':	/* Output file */
		    outFile = fopen (optarg, "w");
		    if (outFile == NULL) {
			tisSysLog (0,0, "Unable to create %s\n", optarg);
			exit (-1);
		    }
		    break;
		case 'h':
		case '?':
		    tisSysLog (0,0, "Usage: pvXref [-o <outputfile>] [-p <directory>]\n");
		    exit (0);
	    }

	/* Load the SDR */

	sprintf (path, "%sdefault.dctsdr", dbDir);
	sdrLoad (path);

	/* Process each record type */

	fieldName [4] = '\0';
	for (i=1 ; i<dbRecDes->number ; i++) {

	    /* Get the character representation of the point type and */
	    /* write it to the output. */

	    strcpy (pointType, GET_PRECTYPE (i));
	    fprintf (outFile, "%s:", pointType);

	    /* Within each point type, go through and find the fields that contain the */
	    /* names of other points. */

	    fldPtr = dbRecDes->papRecTypDes [i]->papFldDes;
	    first = 1;
	    for (j=0 ; j<dbRecDes->papRecTypDes [i]->no_fields ; j++) {
		fldType = fldPtr [j]->field_type;
		if ((fldType == DBF_INLINK) || (fldType == DBF_OUTLINK) || (fldType == DBF_FWDLINK)) {
		    memcpy (fieldName, fldPtr [j]->fldname, 4);
		    fieldName [4] = '\0';
		    stripBlanks (fieldName);
		    if (fldType != DBF_FWDLINK)
			strcat (fieldName, "*");
		    if (!first)
			fprintf (outFile, ",");
		    first = 0;
		    fprintf (outFile, "%s", fieldName);
		}
	    }
	    fprintf (outFile, "\n");
	}

    }

/*----------*/
static void stripBlanks (char *buf)	/* Strip blanks from a string */

    {

	while (*buf) {
	    if (*buf == ' ') {
		*buf = '\0';
		return;
	    }
	    buf++;
	}
	
    }

