/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* Fetch PV Database values into a dBase file */

#include <stdio.h>
#include <string.h>
#include <time.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "cadef.h"
#include "db_access.h"
#include "msgMon.h"

/* Functions defined in this module */

static	void	connectHandler ();
static	void	eventHandler ();

/* CodeBase things */

C4CODE	cb;
D4DATA	*data;
F4FIELD	*timeFld, *dateFld, *chanFld;

/* Local data */

chid	chanId;
int	valueFetched;
char	stringValue [100];
double	floatValue;
int	fieldType;
#define	DEFAULT_TIMEOUT	10
int	timeoutSeconds = DEFAULT_TIMEOUT;

/*----------*/
main (int argc, char *argv[])
    {

	FILE	*f;
	char	dbaseName [100], line [100], buf [100], *p;
	long	clock;
	struct tm *tim;
	int	c;
	extern	char	*optarg;
	extern	int	optind; 
	char	*chan;

	tisOpenLog ("snapDb", 6, 2, NULL, NULL);

	/* Process command line options */

	while ((c = getopt (argc, argv, "t:")) != -1)
	    switch (c) {
		case 't':	/* Explicitly set timeout */
		    timeoutSeconds = atoi (optarg);
		    break;
	    }

	/* Open the specification file */

	f = fopen (argv [optind], "r");
	if (f == NULL) {
	    tisSysLog (0,0, "Unable to open specification file: %s\n", argv [optind]);
	    exit (-1);
	}

	/* The first line of the file contains the database name and strip off the */
	/* trailing newline. */

	fgets (dbaseName, sizeof (dbaseName), f);
	dbaseName [strlen (dbaseName) - 1] = '\0';

	/* Initialize the CodeBase system */

	d4init (&cb);

	/* Get the database name and open the database */

	data = d4open (&cb, dbaseName);
	if (data == NULL) {
	    tisSysLog (0,0, "Unable to open database %s\n", dbaseName);
	    exit (-1);
	}

	/* Initialize Channel Access */

	ca_task_initialize ();

	/* Get the date/time in case we need it. */

	/* Go through the spec file and process the list of fields/points. If */
	/* the keyword "TIME" appears as the field name then get the time and */
	/* place it in the field called "TIME".  The same applies for the */
	/* keyword DATE. */

	d4append_start (data, 0);
	d4blank (data);
	while (fgets (line, sizeof (line), f)) {
	    if (strlen (line) < 5)
		break;
	    p = strtok (line, " \t\n,:");
	    if (strcasecmp (p, "DATE") == 0) {
		dateFld = d4field (data, "DATE");
		if (dateFld == NULL) {
		    tisSysLog (0,0, "%s does not have a field named DATE.\n", dbaseName);
		    exit (-1);
		}
		clock = time (0);
		tim = localtime (&clock);
		if (f4type (dateFld) == 'C')
		    strftime (buf, sizeof (buf), "%D", tim);
		else
		    strftime (buf, sizeof (buf), "%Y%m%d", tim);
		f4assign (dateFld, buf);
	    } else if (strcasecmp (p, "TIME") == 0) {
		clock = time (0);
		tim = localtime (&clock);
		strftime (buf, sizeof (buf), "%T", tim);
		timeFld = d4field (data, "TIME");
		if (timeFld == NULL) {
		    tisSysLog (0,0, "%s does not have a field named TIME.\n", dbaseName);
		    exit (-1);
		}
		f4assign (timeFld, buf);
	    } else {

		/* Open the field to receive the value */

		chanFld = d4field (data, p);
		if (chanFld == NULL) {
		    tisSysLog (0,0, "%s does not contain a field named %s.\n", dbaseName, p);
		    exit (-1);
		}

		/* Fetch the value from the channel */

		floatValue = 0.0;
		strcpy (stringValue, "");
		chan = strtok (NULL, " \t\n:,");
		fieldType = f4type (chanFld);
		switch (fetchValue (chan)) {
		    case 'C':
			f4assign (chanFld, stringValue);
			break;
		    case 'N':
		    case 'F':
			f4assign_double (chanFld, floatValue);
			break;
		    case 'L':
			f4assign_char (chanFld, ((int)floatValue != 0) ? 'Y' : 'N');
			break;
		    default:
			tisSysLog (0,0, "Unable to find %s -- ERR written to database.\n", 
					chan);
			break;
		}
	    }
	}
	d4append (data);
	d4unlock_all (data);

	/* Close the spec file and the database */

	d4close (data);
	fclose (f);

	exit (0);

    }

/*----------*/
int fetchValue (char *channelName)
    {

	long	startTime;

	ca_build_and_connect (channelName, TYPENOTCONN, 0, &chanId, NULL, connectHandler, 0);
	valueFetched = 0;
	startTime = time (0);
	while (!valueFetched) {
	    ca_pend_io (1.0);
	    if ((time (0) - startTime) > timeoutSeconds)
		break;
	}
	return valueFetched ? fieldType : -1;

    }

/*----------*/
static void connectHandler ( struct connection_handler_args args)
    {

	int	type;

	/* Based upon the field type of the dBASE field this value is going */
	/* to be placed in, determine the CA data type needed to fetch the value. */

	switch (fieldType) {
	    case 'C':
		type = OLD_DBR_STRING;
		break;
	    case 'N':
	    case 'F':
	    case 'L':
		type = OLD_DBR_DOUBLE;
		break;
	}
 
	ca_get_callback (type, &chanId, eventHandler, 0);
	    
    }

/*----------*/
static void eventHandler ( struct event_handler_args args)
    {

	valueFetched = 1;
	switch (fieldType) {
	    case 'C':
		strcpy (stringValue, (char *)args.dbr);
		break;
	    case 'N':
	    case 'F':
	    case 'L':
		floatValue = *((double *)args.dbr);
		break;
	}
	ca_clear_channel (chanId);

    }

