#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vbuttonHEADER
#include vcontainerHEADER
#include vmenubarHEADER
#include vmenuHEADER
#include vlistHEADER
#include vlistviewHEADER
#include vfontHEADER
#include vsliderHEADER
#include vprintHEADER
#include vcomboboxHEADER
#include vdateHEADER

extern "C"
{
#ifdef WIN32
#include <stdio.h>
#include <winsock.h>
// this bit of sillyness is needed because windows #defines AppendMenu
// as #define AppendMenu AppendMenuA, which interfers with galaxy class
// vmenubar::AppendMenu
#undef AppendMenu
#endif WIN32
#include <time.h>
#include "cadef.h"
#include "shareMem.h"
#include "alarmDefs.h"
#include "alarmfil.h"
#include "alarmfmt.h"
#include "alarmvu.h"
#include "alpvmap.h"
#include "bell.h"
extern ShareHeader	*shareHeader;
extern TreeNode		*treeRoot;
}
#ifdef	SOLARIS
#define	__cdecl
#endif

#include "alvu.h"
#include "alarmvu.hxx"
#include "ackManyDialog.hxx"
#include "aldlgListView.hxx"
#include "alarmDialogs.hxx"
#include "comboDialog.hxx"
#include "commentDialog.hxx"
#include "filterDialog.hxx"
#include "oneLineDialog.hxx"
#include "reportDialog.hxx"
#include "textDialog.hxx"
#include "alvuUtils.hxx"
#include "utils.hxx"

// extern def's most defined in alarmvu.cxx
extern long	backgroundColor;
extern vcolor	*backgroundVcolor;
extern vcolor	*blackVcolor;
extern int	caUpdateReq;		// see alvuMaiDialog.cxx
extern char	**channelNames;
extern long	CurrUnackSevColor [];
extern int	groupdisplayHeight;
extern char	linePrinter [];
extern char	**groupNames;
extern long	SevColor [];
extern long	UnackSevColor [];

// extern def's for various dialogs, defined in alamrvu.cxx
extern AlarmEventDialog	    *alarmEventDialogPtr; 
extern AlarmHistoryDialog   *alarmHistoryDialogPtr; 
extern AlarmSummaryDialog   *alarmSummaryDialogPtr;
extern AlarmTrackDialog	    *alarmTrackDialogPtr;
extern AlarmWindowDialog    *alarmWindowDialogPtr;
extern CommentDialog	    *commentDialogPtr;
extern DynamicGroupDialog   *dynamicGroupDialogPtr;
extern TextDialog	    *expandDialogPtr;
extern TextDialog	    *guidanceDialogPtr;

// static stuff
static void popupMenuNoteProc (vmenuItem *item, vevent *event);
static char caption[256];
static char dspBuf[1024];
static char txtBuf[1024];
static char timeBuf[40];

// define this here
int DoStrCmp (const char **a, const char **b)
{
    return (strcmp (*a, *b));
}

/*******************************************************************
*
*    AlarmDialog Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AlarmDialog, vdialog, "AlarmDialog");

/*******************************************************************
*
*    AlarmDialog constructor
*
*******************************************************************/

void AlarmDialog::LoadInit (vresource res)
{
    title         = NULL;
    dataList	  = NULL;
    alarmMenuBar  = vmenubar::RequestSubLoad (this, alvu_MenuBar);
    alarmListView = AldlgListView::RequestSubLoad (this, alvu_ListView);
    closeButton   = CloseButton::RequestSubLoad (this, alvu_CloseButton);
    upButton      = UpButton::RequestSubLoad (this, alvu_UpButton);
    topButton     = TopButton::RequestSubLoad (this,alvu_TopButton);
    pauseButton   = PauseButton::RequestSubLoad (this, alvu_PauseButton);
    exitButton    = ExitButton::RequestSubLoad (this, alvu_ExitButton);
    elementCount  = vtextitem::RequestSubLoad (this, alvu_ItemCount);

    vdialog::LoadInit (res);

    // store ptr to dialog in buttons
    closeButton -> StoreAlarmDialog (this);
    upButton    -> StoreAlarmDialog (this);
    topButton   -> StoreAlarmDialog (this);
    pauseButton -> StoreAlarmDialog (this);
    exitButton  -> StoreAlarmDialog (this);
}


/*******************************************************************
*
*    AlarmDialog window observer
*
*******************************************************************/

void AlarmDialog::ObserveWindow (vevent *event)
{
    if (event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	if ((CtrlFunc (WINCLOSE, -1)) == False)
	{
	    vwindow::Beep ();
	}
	else
	{
	    Close ();
	}
    }
}


/*******************************************************************
*
*    AlarmDialog destructor
*
*******************************************************************/

void AlarmDialog::Destroy (void)
{
    if (PopupMenu != NULL)
    {
	delete PopupMenu;
    }
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    if (dataList != NULL)
    {
	delete dataList;
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    AlarmDialog destructor
*
*******************************************************************/

void AlarmDialog::AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt)
{
    app = alvuApp;
    filePos = 0;
    numFileItems = 0;
    paused = False;

    BuildPopupMenu ();

    // init size of dialog, groupdisplayHeight can be changed 
    vrect *tempRect = (vrect *)(GetRect ());
    tempRect -> SetHeight (groupdisplayHeight);
    tempRect -> SetWidth (ALVU_DEFAULT_GRP_DISPLAY_WID);
    SetRect (tempRect);

    // store ptr for listview
    alarmListView -> StoreAlarmDialog (this);
 
    vdialog::Open ();
}


/*******************************************************************
*
*    AlarmDialog add to dataList
*
*******************************************************************/

void AlarmDialog::AddLine (int colorIdx, char *Date, char *Name,
				 char *Desc, TS_STAMP *altime, int Mon)
{
    AddLine (colorIdx, False, Date, Name, Desc, altime, Mon);
}
void AlarmDialog::AddLine (int colorIdx, int AckM, char *Date, char *Name,
			   char *Desc, TS_STAMP *altime, int Mon)
{
    ListEntry *newEntry;
    int rowCount = dataList -> GetRowCount ();
    if (colorIdx == -1)
    {
	newEntry = new ListEntry (rowCount, 0, AckM, Date, Name, Desc, altime, this);
    }
    else
    {
	newEntry = new ListEntry (rowCount, colorIdx, AckM, Date, Name, Desc, altime, this);
    }
    dataList -> SetCellValue (rowCount, 0, (const void *)newEntry);
    if (monitored & Mon)
    {
	newEntry -> Monitor ();
    }
}


/*******************************************************************
*
*    AlarmDialog Build menu function to be written by subclasses
*
*******************************************************************/

void AlarmDialog::BuildMenus (void)
{
}

/*******************************************************************
*
*    AlarmDialog Build handle blinking
*******************************************************************/

void AlarmDialog::UpdateBlink (void)
{
    if (IsOpen ())
    {
	if ((pauseButton -> IsVisible()) && (pauseButton -> IsEnabled ()))
	{
	    pauseButton -> UpdateBlink ();
	}
    }
}


vkindTYPED_FULL_DEFINITION(popupMenuItems, vmenuItem, NULL);
/*******************************************************************
*
*    AlarmDialog build the popup menu for the ack button
*
*******************************************************************/

void AlarmDialog::BuildPopupMenu (void)
{
    vapplication *alvuApp = vapplication::GetCurrent ();
    vresource alvuResource = alvuApp -> GetResources ();
    vresource titlesResource = vresourceGet (alvuResource, alvu_TitlesDictionary);
    vresource popupStrRes = vresourceGet (titlesResource, alvu_PopupStrings);

    PopupMenu = new vmenu ();

    int i;
    popupMenuItems *menuItem;
    for (i = 0; i < POPUP_COUNT; i++)
    {
	menuItem = new popupMenuItems;
	menuItem -> SetTitleScribed (vresourceScribeStringAt (popupStrRes, i));
	menuItem -> SetObserveMenuItemProc (popupMenuNoteProc);
	menuItem -> StoreAlarmDialog (this);
	PopupMenu -> AppendItem (menuItem);
    }
}


/*******************************************************************
*
*    AlarmDialog check for cancel data
*******************************************************************/

int AlarmDialog::CheckCancel (void)
{
    checksCalled++;
    if ( (checksCalled % 100) == 0)
    {
	vstr *formattedMsg = vresourceGetString (app -> titlesResource,
						alvu_ContinueReadDataTitle);
	vstr *newTitle = vstrCloneScribed (vcharScribeFormat (formattedMsg, checksCalled));
	cancelDataContainer -> cancelDataButton -> SetTitle (newTitle);
	vstrDestroy (formattedMsg);
	vstrDestroy (newTitle);
    }
    vevent::FlushProcessing ();
    return (dataCancelled);
}


/*******************************************************************
*
*    AlarmDialog clear out lists
*
*******************************************************************/

void AlarmDialog::ClearLines (void)
{
    // detach list from listview
    alarmListView -> SetList (NULL);
    if (dataList != NULL)
    {
	vlistIterator iterator;
	iterator.Start (dataList);
	while (iterator.Next ())
	{
	    ListEntry *entry = (ListEntry *)(iterator.GetValue ());
	    if (entry -> IsMonitored ())
	    {
		entry -> StopMonitor ();
	    }
	}
	iterator.Finish ();
	delete dataList;
	dataList = NULL;
    }
    dataList = new vlist;
}
 

/*******************************************************************
*
*    AlarmDialog control function to be written by sub-class
*
*******************************************************************/

int AlarmDialog::CtrlFunc (int cmd, int param)
{
    return (True);
}
 

/*******************************************************************
*
*    AlarmDialog function to setup popup menu
*
*******************************************************************/

void AlarmDialog::DoPopup (vevent *event)
{
    ListEntry *entry = alarmListView -> GetListEntry ();
    AckButton *button = entry -> GetButtonItem ();
    const vrect *tempRect = button -> GetRect ();

    if (CtrlFunc (NAMECLICK, entry -> GetRow()))
    {
	// place popup to the right of the ack button that called it
	PopupMenu -> PlaceRight ((vwindow *)this, tempRect);
	PopupMenu -> Popup (event);
    }
}
 

/*******************************************************************
*
*    AlarmDialog handle a window reconfigure event
*
*******************************************************************/

int AlarmDialog::HandleConfigure (vevent *event)
{
    vwindow::HandleConfigure (event);
    if (event -> GetType () == veventWINDOW_CONFIGURE && event -> IsSize ())
    {
	int newSize = alarmListView -> GetListNumElements ();
	if (newSize != displaySize)
	{
	    displaySize = newSize;
	    if (fileBased)
	    {
		if (dataList != NULL)
		{
		    int rowCnt = dataList -> GetRowCount ();
		    if (rowCnt > 0)
		    {
			SetPosition (Pos ());
		    }
		}
	    }
	}
    }
    return (vTRUE);
}


/*******************************************************************
*
*    AlarmDialog some utility functions
*
*******************************************************************/

char *AlarmDialog::ItemDate (int row)
{
    if (row >= NumItems ())
    {
	return ("");
    }
    else
    {
	ListEntry *entry = (ListEntry *)(dataList -> GetCellValue (row, 0));
	if (entry != NULL)
	{
	    return (entry -> GetDate ());
	}
	else
	{
	    return ("");
	}
    }
}

char *AlarmDialog::ItemName (int row)
{
    if (row >= NumItems ())
    {
	return ("");
    }
    else
    {
	ListEntry *entry = (ListEntry *)(dataList -> GetCellValue (row, 0));
	if (entry != NULL)
	{
	    return (entry -> GetName ());
	}
	else
	{
	    return ("");
	}
    }
}


/*******************************************************************
*
*    AlarmDialog some more utility functions
*
*******************************************************************/

char *AlarmDialog::ItemDesc (int row)
{
    if (row >= NumItems ())
    {
	return ("");
    }
    else
    {
	ListEntry *entry = (ListEntry *)(dataList -> GetCellValue (row, 0));
	if (entry != NULL)
	{
	    return (entry -> GetDesc ());
	}
	else
	{
	    return ("");
	}
    }
}

ListEntry *AlarmDialog::GetEntryPtr (int row)
{
    if (row >= NumItems ())
    {
	return (NULL);
    }
    else
    {
	if (dataList != NULL)
	{
	    ListEntry *entry = (ListEntry *)(dataList -> GetCellValue (row, 0));
	    return (entry);
	}
	else
	{
	    return (NULL);
	}
    }
}
 

/*******************************************************************
*
*    AlarmDialog set up monitor
*
*******************************************************************/

void AlarmDialog::MonitorCB (unsigned int ID)
{
    CtrlFunc (CAMONITOR, ID);
}
 

/*******************************************************************
*
*    AlarmDialog return the number of file or dataList items
*
*******************************************************************/

int AlarmDialog::NumItems (void)
{
    int numItems;

    if (fileBased)
    {
	numItems = numFileItems;
    }
    else
    {
	if (dataList != NULL)
	{
	    numItems = dataList -> GetRowCount ();
	}
	else
	{
	    numItems = 0;
	}
    }
    return (numItems);
}
 

/*******************************************************************
*
*    AlarmDialog return position in list
*
*******************************************************************/

int AlarmDialog::Pos ()
{
    if (fileBased)
    {
	return (filePos);
    }
    else
    {
	vlist *list = alarmListView -> GetList ();
	if (list != NULL)
	{
	    return (alarmListView -> GetTopRow ());
	}
	else
	{
	    return (0);
	}
    }
}
 

/*******************************************************************
*
*    AlarmDialog set bottom position
*
*******************************************************************/

void AlarmDialog::SetBottom (void)
{
    int	newPos;

    if (displaySize > 0)
    {
	newPos = NumItems() - displaySize;
    }
    else
    {
	newPos = 0;
    }
    if (newPos < 0)
    {
	newPos = 0;
    }
    SetPosition (newPos);
}


/*******************************************************************
*
*    AlarmDialog set title function to be written by subclasses
*
*******************************************************************/

void AlarmDialog::SetDialogTitle (void)
{
}

/*******************************************************************
*
*    AlarmDialog handle the pause button
*
*******************************************************************/

void AlarmDialog::ShowPaused (int val)
{
    pauseButton -> Enable ();

    paused = val;

    long color;
    vcolor *fg;

    if (paused)
    {
	double red, green, blue;
	color = CurrUnackSevColor [MAJOR_ALARM + NUM_SEV_COLORS];
	GetRGBColor (color, &red, &green, &blue);
	vcolorSpec *spec = new vcolorSpec;
	spec -> FSetRGB (red, green, blue);
	fg = vcolor::Intern (spec);
    }
    else
    {
	fg = blackVcolor;
    }

    pauseButton -> SetForeground (fg);
}
 

/*******************************************************************
*
*    AlarmDialog set position
*
*******************************************************************/

void AlarmDialog::SetPosition (int newPos)
{
    if (!fileBased && (newPos > (NumItems () - displaySize)))
    {
	newPos = NumItems () - displaySize;
    }
    if (newPos < 0)
    {
	newPos = 0;
    }
    if (fileBased)
    {
	CtrlFunc (MOVECLICK, newPos);
	if (Pos () != 0)
	{
	    fileButtonContainer -> firstButton -> Enable ();
	    fileButtonContainer -> pageupButton -> Enable ();
	    fileButtonContainer -> lineupButton -> Enable ();
	}
	else
	{
	    fileButtonContainer -> firstButton -> Disable ();
	    fileButtonContainer -> pageupButton -> Disable ();
	    fileButtonContainer -> lineupButton -> Disable ();
	}
	if ( (Pos () < 90) && (displaySize < NumItems ()))
	{
	    fileButtonContainer -> linedownButton -> Enable ();
	    fileButtonContainer -> pagedownButton -> Enable ();
	    fileButtonContainer -> lastButton -> Enable ();
	}
	else
	{
	    fileButtonContainer -> linedownButton -> Disable ();
	    fileButtonContainer -> pagedownButton -> Disable ();
	    fileButtonContainer -> lastButton -> Disable ();
	}
    }
    else
    {
	if (NumItems () > 0)
	{
	    alarmListView -> ScrollToCell (newPos, 0);
	}
	if (NumItems () == 0)
	{
	    elementCount -> SetTextScribed (vcharScribeSystem (""));
	}
	else
	{
	    vstr *format = vresourceGetString (app -> messagesResource,
					       alvu_ItemCountMsg);
	    vstr *msg = vstrCloneScribed (vcharScribeFormat (format, 
							     newPos + 1, 
							     NumItems()));
	    vstrDestroy (format);
	    elementCount -> SetText (msg);
	    vstrDestroy (msg);
	}
    }
}


/*******************************************************************
*
*    AlarmDialog Start cancel data button function
*******************************************************************/

void AlarmDialog::StartCancelBox (void)
{
    dataCancelled = False;
    checksCalled = 0;
    elementCount -> SetTextScribed (vcharScribeSystem (""));
    cancelDataContainer -> cancelDataButton -> Enable ();
    cancelDataContainer -> cancelDataButton -> Show ();
    vstr *buttonTitle = vresourceGetString (app -> titlesResource,
					    alvu_StartReadDataTitle);
    cancelDataContainer -> cancelDataButton -> SetTitle (buttonTitle);
    vstrDestroy (buttonTitle);
}


/*******************************************************************
*
*    AlarmDialog Stop cancel data button function
*******************************************************************/

void AlarmDialog::StopCancelBox (void)
{
    cancelDataContainer -> cancelDataButton -> SetTitleScribed (vcharScribeSystem (""));
    cancelDataContainer -> cancelDataButton -> Disable ();
    cancelDataContainer -> cancelDataButton -> Hide ();
}


/*******************************************************************
*
*    AlarmDialog Windows print function to be over written by
*	sub classes
*******************************************************************/

void AlarmDialog::WindowsPrint (int whatToPrint)
{
}

/*******************************************************************
*
*    AlarmDialog Unix print function to be over written by
*	sub classes
*******************************************************************/

void AlarmDialog::UnixPrint (int whatToPrint)
{
}


/*******************************************************************
*
*    popupCommands 
*******************************************************************/

static void popupMenuNoteProc (vmenuItem *item, vevent *event)
{
    popupMenuItems *menuItem = (popupMenuItems *)item;
    vmenuItemList *menuItemList = menuItem -> GetItemListIn ();
    int	index = menuItemList -> GetIndex ((vobjectAttributed *)menuItem);
    AlarmDialog *alarmDialog = menuItem -> GetAlarmDialog ();
    ListEntry *entry = alarmDialog -> alarmListView -> GetListEntry ();
    int row = entry -> GetRow ();
    int clickType;

    switch (index)
    {
	case SUMMARY_MENUITEM:
	{
	    clickType = SUMMARYCLICK;
	    break;
	}

	case GRP_MENUITEM:
	{
	    clickType = GROUPCLICK;
	    break;
	}

	case GUIDANCE_MENUITEM:
	{
	    clickType = GUIDECLICK;
	    break;
	}

	case DISPLAY_MENUITEM:
	{
	    clickType = DISPLAYCLICK;
	    break;
	}

	case DETAIL_MENUITEM:
	{
	    clickType = DETAILCLICK;
	    break;
	}

	case SCRIPT_MENUITEM:
	{
	    clickType = COMMANDCLICK;
	    break;
	}

	case ENABLE_MENUITEM:
	{
	    clickType = ENABLECLICK;
	    break;
	}

	case DISABLE_MENUITEM:
	{
	    clickType = DISABLECLICK;
	    break;
	}

	case COMMENT_MENUITEM:
	{
	    clickType = COMMENTCLICK;
	    break;
	}

	case HELP_MENUITEM:
	{
	    clickType = HELPCLICK;
	    break;
	}

	default:
	    return;
    }
    alarmDialog -> CtrlFunc (clickType, row);
}


/*******************************************************************
*
*    AlarmDialog button classes
*
*******************************************************************/
vkindTYPED_FULL_DEFINITION(CloseButton, vbutton, "CloseButton");
vkindTYPED_FULL_DEFINITION(UpButton, vbutton, "UpButton");
vkindTYPED_FULL_DEFINITION(TopButton, vbutton, "TopButton");
vkindTYPED_FULL_DEFINITION(PauseButton, vbutton, "PauseButton");
vkindTYPED_FULL_DEFINITION(ExitButton, vbutton, "ExitButton");

void CloseButton::ObserveDialogItem (vevent *event)
{
    if ((alarmDialog -> CtrlFunc (WINCLOSE, -1)) == False)
    {
	vwindow::Beep ();
    }
    else
    {
	alarmDialog -> Close ();
    }
}

void UpButton::ObserveDialogItem (vevent *event)
{
    alarmDialog -> CtrlFunc (UPCLICK, -1);
}

void TopButton::ObserveDialogItem (vevent *event)
{
    alarmDialog -> CtrlFunc (TOPCLICK, -1);
}

void PauseButton::ObserveDialogItem (vevent *event)
{
    alarmDialog -> CtrlFunc (PAUSECLICK, -1);
}

void PauseButton::UpdateBlink (void)
{
    if (IsEnabled ())
    {
	if (alarmDialog -> IsPaused ())
	{
	    double red, green, blue;
	    long color = CurrUnackSevColor [MAJOR_ALARM];
	    GetRGBColor (color, &red, &green, &blue);
	    vcolorSpec *spec = new vcolorSpec;
	    spec -> FSetRGB (red, green, blue);
	    vcolor *fg = vcolor::Intern (spec);
	    SetForeground (fg);
	}
    }
}

void ExitButton::ObserveDialogItem (vevent *event)
{
    if ((alarmDialog -> CtrlFunc (EXITCLICK, -1)) == False)
    {
	vwindow::Beep ();
    }
    else
    {
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    CancelDataButtonContainer Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(CancelDataButtonContainer, vcontainer,
			   "CancelDataButtonContainer");

void CancelDataButtonContainer::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource contDictRes = vresourceGet (app -> GetResources (),
					    alvu_ContainerDictionary);
    vresource realRes = vresourceGet (contDictRes, 
					    alvu_CancelDataButtonContainer);
    cancelDataButton = CancelDataButton::RequestSubLoad (this,
					    alvu_CancelDataButton);
    vcontainer::LoadInit (realRes);
}


/*******************************************************************
*
*    AlarmDialog Cancel Data Button class
*
*******************************************************************/
vkindTYPED_FULL_DEFINITION(CancelDataButton, vbutton, "CancelDataButton");

void CancelDataButton::ObserveDialogItem (vevent *event)
{
    alarmDialog -> SetDataCancelled (True);
}



/*******************************************************************
*
*    FileButtonContainer Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(FileButtonContainer, vcontainer, "FileButtonContainer");

void FileButtonContainer::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource contDictRes = vresourceGet (app -> GetResources (),
					    alvu_ContainerDictionary);
    vresource realRes = vresourceGet (contDictRes, 
					    alvu_AlarmDialogButtonContainer);
    firstButton = FirstButton::RequestSubLoad (this,
					    alvu_TopButton);
    pageupButton = PageupButton::RequestSubLoad (this,
					    alvu_OnePageUpButton);
    lineupButton = LineupButton::RequestSubLoad (this,
					    alvu_OneLineUpButton);
    linedownButton = LinedownButton::RequestSubLoad (this,
					    alvu_OneLineDownButton);
    pagedownButton = PagedownButton::RequestSubLoad (this,
					    alvu_PageDownButton);
    lastButton = LastButton::RequestSubLoad (this,
					    alvu_BottomButton);
    vcontainer::LoadInit (realRes);
}


/*******************************************************************
*
*    AlarmDialog File Position Button classes
*
*******************************************************************/
vkindTYPED_FULL_DEFINITION(FirstButton, vbutton, "FirstButton");
vkindTYPED_FULL_DEFINITION(PageupButton, vbutton, "PageupButton");
vkindTYPED_FULL_DEFINITION(LineupButton, vbutton, "LineupButton");
vkindTYPED_FULL_DEFINITION(LinedownButton, vbutton, "LinedownButton");
vkindTYPED_FULL_DEFINITION(PagedownButton, vbutton, "PagedownButton");
vkindTYPED_FULL_DEFINITION(LastButton, vbutton, "LastButton");

void FirstButton::ObserveDialogItem (vevent *event)
{
    alarmDialog -> SetPosition (0);
}

void PageupButton::ObserveDialogItem (vevent *event)
{
    if (alarmDialog -> Size() == 1)
    {
	alarmDialog -> SetPosition (49);
    }
    else
    {
	alarmDialog -> SetPosition (25);
    }
}

void LineupButton::ObserveDialogItem (vevent *event)
{
    alarmDialog -> SetPosition (49);
}

void LinedownButton::ObserveDialogItem (vevent *event)
{
    alarmDialog -> SetPosition (51);
}

void PagedownButton::ObserveDialogItem (vevent *event)
{
    if (alarmDialog -> Size () == 1)
    {
	alarmDialog -> SetPosition (51);
    }
    else
    {
	alarmDialog -> SetPosition (75);
    }
}

void LastButton::ObserveDialogItem (vevent *event)
{
    alarmDialog -> SetPosition (100);
}


/*******************************************************************
*
*    Alarm Summary Dialog
*	subclassed from AlarmDialog
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AlarmSummaryDialog, AlarmDialog, "AlarmSummaryDialog");

/*******************************************************************
*
*    Alarm Summary Dialog, LoadInit ()
*
*******************************************************************/

void AlarmSummaryDialog::LoadInit (vresource resource)
{
    AlarmDialog::LoadInit (resource);
}

/*******************************************************************
*
*    Alarm Summary Dialog, destroy function
*
*******************************************************************/

void AlarmSummaryDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    AlarmDialog::Destroy ();
}


/*******************************************************************
*
*    AlarmSummaryDialog Open function
*
*******************************************************************/

void AlarmSummaryDialog::AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt)
{
    app = alvuApp;
    filter = *filt;
    PopupMenu = NULL;
    groupLinkPtr = FindGroup (treeRoot, filter.group);

    BuildMenus ();
    SetDialogTitle ();

    ackable = True;
    hasdate = False;
    hasdesc = False;
    monitored = False;
    fileBased = False;
    pause = False;

    AlarmDialog::AlarmDialogOpen (alvuApp, filt);

    displaySize = alarmListView -> GetListNumElements ();

    RefreshDisplay ();

    ShowPaused (pause);
}


/*******************************************************************
*
*    AlarmSummaryDialog Build menu function 
*
*******************************************************************/

void AlarmSummaryDialog::BuildMenus (void)
{
    vresource titles = app -> titlesResource;
    vmenu	*newMenu;
    vmenuItem	*menuItem;
    vstr	*label;

    // the print menu

    label = vresourceGetString (titles, alvu_PrintMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ThisWindow);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlsumThisWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_EntireSummary);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlsumEntireWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the filter menu

    label = vresourceGetString (titles, alvu_FilterMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ModifyFilter);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlsumModFilter);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the silence menu

    label = vresourceGetString (titles, alvu_SilenceMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_Silence);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlsumSilence);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the ack all menu

    label = vresourceGetString (titles, alvu_AckAllMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_AckAll);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlsumAckAll);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_AckPage);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlsumAckPage);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // load up menuNames with resources
    menuNames [ALSUM_THISWIN_ITEM]   = alvu_AlsumThisWindow;
    menuNames [ALSUM_ENTIRESUM_ITEM] = alvu_AlsumEntireWindow;
    menuNames [ALSUM_MODFILT_ITEM]   = alvu_AlsumModFilter;
    menuNames [ALSUM_SILENCE_ITEM]   = alvu_AlsumSilence;
    menuNames [ALSUM_ACKALL_ITEM]    = alvu_AlsumAckAll;
    menuNames [ALSUM_ACKPAGE_ITEM]   = alvu_AlsumAckPage;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;
    // populate our dictionary with menu commands
    for (i = ALSUM_THISWIN_ITEM; i < ALSUM_ITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	alarmSummaryCommands *command = new alarmSummaryCommands;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (menuNames[i], command);
    }

    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}


/*******************************************************************
*
*    Alarm Summary Dialog, Ctrl Function ()
*	param if >= 0 is row number in list
*
*******************************************************************/
static AlarmLogRecord	alarmSumRecord;

int AlarmSummaryDialog::CtrlFunc ( int cmd, int param)
{
    char	*ptName;
    PvData	*pvdata;
    Filter	tempFilter;
    char	*guidptr;
    char	*dispptr;
    vstr	*pv;

    ListEntry	*entry = NULL;
    // get resource to alarmvu
    vresource alvuRes = app -> alvuResource;

    switch (cmd)
    {
	case ACKCLICK:
	{
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace(ptName)))
	    {
		// alarm is not ack'd and there is a chain of alarms
		// bring up Unack'd alarm dialog
		if (!pvdata -> Ack && (pvdata -> lastAck.secPastEpoch != 0))
		{
		    vresource ackRes = vresourceGet (alvuRes, alvu_AckManyDialog);
		    AckManyDialog *ackdlg = new AckManyDialog (ackRes);
		    ackdlg -> OpenAckMany (app, ptName, &pvdata -> lastAlarm);
		    int okToAck = ackdlg -> okToAck;
		    delete ackdlg;
		    // dont ACK if cancelled
		    if (okToAck == vFALSE)
		    {
			break;
		    }
		}
		if (AckAlarm (pvdata) != True)
		{
		    pv = vstrCloneScribed (vcharScribeSystem (ptName));
		    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
					     alvu_NoAckPrivMsg, pv);
		    vstrDestroy (pv);
		}
	    }
	    break;
	}
	case EXPANDCLICK:
	    if (expandDialogPtr != NULL)
	    {
		delete expandDialogPtr;
		expandDialogPtr = NULL;
	    }
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		vresource expRes = vresourceGet (alvuRes, alvu_TextDialog);
		expandDialogPtr = new TextDialog (expRes);
		expandDialogPtr -> OpenText (app, 
					     EXPAND_TEXT, 
					     AlarmFormat (&expandsummary, pvdata, NULL),
					     NULL);
	    }

	    break;
	case NAMECLICK:
	    // Adjust menu items using status
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		vmenuItem *item;

		GetAlarmInfo (ptName, &guidptr, &dispptr);

		// get Guidance menuitem
		item = PopupMenu -> GetItemAt (GUIDANCE_MENUITEM);
		// if guidance for point or for parent group, enable
		if ( (guidptr != 0) || 
		     HaveScript (pvdata -> parentGroup -> groupParms.guidance) )
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Display menuitem
		item = PopupMenu -> GetItemAt (DISPLAY_MENUITEM);
		if ( (dispptr != 0) || 
		     HaveScript (pvdata -> parentGroup -> groupParms.displayName))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Detail menuitem
		item = PopupMenu -> GetItemAt (DETAIL_MENUITEM);
		item -> Enable ();
		// get Script menuitem
		item = PopupMenu -> GetItemAt (SCRIPT_MENUITEM);
		if ( HaveScript (pvdata -> parentGroup -> groupParms.manualScript) ||
		     HaveScript (pvdata -> parentGroup -> groupParms.auxManualScript))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Enable menuitem
		item = PopupMenu -> GetItemAt (ENABLE_MENUITEM);
		item -> Enable ();
		// get Disable menuitem
		item = PopupMenu -> GetItemAt (DISABLE_MENUITEM);
		item -> Enable ();
		// get Comment menuitem
		item = PopupMenu -> GetItemAt (COMMENT_MENUITEM);
		item -> Disable ();
		// get Help menuitem
		item = PopupMenu -> GetItemAt (HELP_MENUITEM);
		item -> Disable ();
		return True;
	    }
	    else
		return False;

	case SUMMARYCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		strncpy (filter.group, pvdata -> parentGroup -> groupName, 
			 sizeof (filter.group));
		filter.group [sizeof (filter.group)-1] = 0;
		if (pause)
		{
		    pause = False;
		    ShowPaused (pause);
		}
		Resync ();
	    }
	    break;

	case GROUPCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		tempFilter = filter;
		strncpy (tempFilter.group, pvdata -> parentGroup -> groupName, 
			 sizeof (tempFilter.group));
		tempFilter.group [sizeof (tempFilter.group) - 1] = 0;
		if (dynamicGroupDialogPtr != NULL)
		{
		    delete dynamicGroupDialogPtr;
		}
		vresource alarmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
		dynamicGroupDialogPtr = new DynamicGroupDialog (alarmDlgRes);
		dynamicGroupDialogPtr -> AlarmDialogOpen (app, &tempFilter);
	    }
	    break;

	case GUIDECLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		if (guidanceDialogPtr != NULL)
		{
		    delete guidanceDialogPtr;
		    guidanceDialogPtr = NULL;
		}
		// get any guidance on the point
		GetAlarmInfo (ptName, &guidptr, &dispptr);
		if (guidptr == NULL)
		{
		    // does parent alarm group have guidance
		    if (HaveScript (pvdata -> parentGroup -> groupParms.guidance))
		    {
			guidptr = pvdata -> parentGroup -> groupParms.guidance;
		    }
		}
		if (guidptr && *guidptr)
		{
		    vresource guidRes = vresourceGet (alvuRes, alvu_TextDialog);
		    guidanceDialogPtr = new TextDialog (guidRes);
		    guidanceDialogPtr -> OpenText (app, 
						   GUIDANCE_TEXT, 
						   guidptr,
						   ptName);
		}
	    }
	    break;

	case DISPLAYCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		SpawnPointDisplay (pvdata);
	    }
	    break;

	case DETAILCLICK:
	    ptName = ItemName (param);
	    SpawnDetail (ptName);
	    break;

	case COMMANDCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		SpawnScript (pvdata);
	    }
	    break;

	case ENABLECLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		if (EnablePoint (pvdata) != True)
		{
		    pv = vstrCloneScribed (vcharScribeSystem (ptName));
		    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
					     alvu_NoEnablePriv, pv);
		    vstrDestroy (pv);
		}
	    }
	    break;

	case DISABLECLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		if (DisablePoint (pvdata) != True)
		{
		    pv = vstrCloneScribed (vcharScribeSystem (ptName));
		    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
					     alvu_NoDisablePriv, pv);
		    vstrDestroy (pv);
		}
	    }
	    break;

	case COMMENTCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		if (commentDialogPtr != NULL)
		{
		    delete commentDialogPtr;
		    commentDialogPtr = NULL;
		}
		TS_STAMP *evtime = entry -> GetEventTime ();
		if (Log_RecordByTime (ptName, evtime, &alarmSumRecord) != RECFOUND)
		{
		    break;
		}
		sprintf (dspBuf, "%-17s  %-28s  %s",
				 ItemDate (param), ItemName (param), ItemDesc (param));
		vresource commentRes = vresourceGet (alvuRes, alvu_CommentDialog);
		commentDialogPtr = new CommentDialog (commentRes);
		commentDialogPtr -> OpenComment (app, pvdata -> pvName, dspBuf, 
						 alarmSumRecord.comment, 
						 &alarmSumRecord);
	    }
	    break;

	case UPCLICK:
	    if (groupLinkPtr = FindGroup (treeRoot, filter.group))
	    {
		if (groupLinkPtr -> parent)
		{
		    strncpy (filter.group, groupLinkPtr -> parent -> groupName,
			     sizeof (filter.group));
		    filter.group [sizeof (filter.group)-1] = 0;
		    filter.modified = TRUE;
		    SetDialogTitle ();
		}
	    }
	    Resync ();
	    break;

	case TOPCLICK:
	    strncpy (filter.group, treeRoot -> groupName, sizeof (filter.group));
	    filter.group [sizeof (filter.group)-1] = 0;
	    filter.modified = TRUE;
	    SetDialogTitle ();
	    Resync ();
	    break;

	case PAUSECLICK:
	    pause = !pause;
	    ShowPaused (pause);
	    if (!pause)
	    {
		Resync ();
	    }
	    else
	    {
		RefreshDisplay ();
	    }
	    break;
    }
    return True;
}


/*******************************************************************
*
*    Alarm Summary Dialog, RefreshDisplay ()
*
*******************************************************************/

void AlarmSummaryDialog::RefreshDisplay (void)
{
    int		savePos;
    int		color;
    PvData	*pvdata;
    int		ackval;
    savePos = Pos ();
    ClearLines ();
    for (pvdata = shareHeader -> summaryHead; pvdata; pvdata = pvdata -> nextSummaryPv)
    {
	if (IsChannelFiltered (pvdata, &filter, pvdata -> unAckedSeverity))
	{
	    // filtered out of display
	    continue;
	}
	// if return to normal
	if (pvdata -> RTN)
	{
	    // blink, NO_ALARM
	    color = NUM_SEV_COLORS + NO_ALARM;
	}
	// most recent uncked sev > NO_ALARM, and not acked
	else if ( (pvdata -> unAckedSeverity != NO_ALARM) && (!pvdata -> Ack) )
	{
	    // blink, alarm state
	    color = NUM_SEV_COLORS + pvdata -> unAckedSeverity;
	}
	else
	{
	    // no blink, alarm state
	    color = pvdata -> unAckedSeverity;
	}
	// state of ack button display (blank or +)
	ackval = (!pvdata -> Ack && (pvdata -> lastAck.secPastEpoch != 0) );
	AddLine (color, ackval, "", pvdata -> pvName, 
		 AlarmFormat (&displaysummary, pvdata, NULL), &pvdata -> lastAlarm);
    }
    if (dataList != NULL)
    {
	alarmListView -> SetList (dataList);
	vlistSelection *selected = alarmListView -> GetSelection ();
	dataList -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	dataList -> SetAtLeastOneSelection (selected, vFALSE);
	if (pause)
	{
	    SetPosition (savePos);
	}
	else
	{
	    SetBottom ();
	}
    }
}


/*******************************************************************
*
*    Alarm Summary Dialog, Resync ()
*
*******************************************************************/

void AlarmSummaryDialog::Resync (void)
{
    RefreshDisplay ();
}


/*******************************************************************
*
*    AlarmDialog set title function to be written by subclasses
*
*******************************************************************/

void AlarmSummaryDialog::SetDialogTitle (void)
{
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    vstr *format = vresourceGetString (app -> titlesResource, alvu_AlarmSummaryTitle);
    vstr *group = vstrCloneScribed (vcharScribeSystem (filter.group));
    vstr *mod = vstrCloneScribed (vcharScribeLiteral (filter.modified ? "*" : ""));
    vstr *name = vstrCloneScribed (vcharScribeSystem (filter.name));
    title = vstrCloneScribed (vcharScribeFormat (format, group, mod, name));
    SetTitle (title);
    vstrDestroy (format);
    vstrDestroy (group);
    vstrDestroy (mod);
    vstrDestroy (name);
}

/*******************************************************************
*
*    AlarmSummaryDialog handle blinking
*
*******************************************************************/

void AlarmSummaryDialog::UpdateBlink (void)
{
    alarmListView -> Inval (vwindowINVAL_OPAQUE | vwindowINVAL_UPDATE);
    AlarmDialog::UpdateBlink ();
}


/*******************************************************************
*
*    AlarmDialog unix print override of base class function
*
*******************************************************************/

void AlarmSummaryDialog::UnixPrint (int whatToPrint)
{
    FILE    *printFileDesc;
    char    printFileName [30];
    int	    startLine, endLine;
    int	    i;
    vlist   *list = alarmListView -> GetList ();
    int	    rowCount = list -> GetRowCount ();

    if (whatToPrint == PRINT_THIS_WINDOW)
    {
	startLine = alarmListView -> GetTopRow ();
	endLine = alarmListView -> GetBottomRow ();
    }
    else    // print entire list
    {
	startLine = 0;
	endLine = rowCount - 1;
    }

    strcpy (printFileName, PRINT_FILE_TEMP);
    if ( !mktemp (printFileName) || !(printFileDesc = fopen (printFileName, "w")))
    {
	vstr *prtFileTmp = vstrCloneScribed (vcharScribeLiteral (PRINT_FILE_TEMP));
	alvuMessageDialogCenter (this, ALVU_INFO_NOTICE, alvu_TempPrtFileErr, prtFileTmp);
	vstrDestroy (prtFileTmp);
    }
    else
    {
	for (i = startLine; i <= endLine; i++)
	{
	    fprintf (printFileDesc, "%s\n", ItemDesc (i));
	}
	fclose (printFileDesc);
	if (strlen (linePrinter))
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"Alarm Summary\" %s |lpr -P\"%s\"", 
			      printFileName, linePrinter);
#else
	    sprintf (txtBuf, "lpr -p -r -P\"%s\" -T\"Alarm Summary\" %s", 
			      linePrinter, printFileName);
#endif SOLARIS
	}
	else
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"Alarm Summary\" %s |lpr", printFileName);
#else
	    sprintf (txtBuf, "lpr -p -r -T\"Alarm Summary\" %s", printFileName);
#endif SOLARIS
	}
	system (txtBuf);
#ifdef SOLARIS
	sprintf (txtBuf, "rm %s", printFileName);
	system (txtBuf);
#endif SOLARIS
    }
}


/*******************************************************************
*
*    AlarmDialog windows print override of base class function
*
*******************************************************************/

void AlarmSummaryDialog::WindowsPrint (int whatToPrint)
{
    int	    startLine, endLine;
    int	    i;
    vlist   *list = alarmListView -> GetList ();
    int	    rowCount = list -> GetRowCount ();

    if (app -> printer == NULL)
    {
	alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_NoPrinterMsg);
	return;
    }

    if (whatToPrint == PRINT_THIS_WINDOW)
    {
	startLine = alarmListView -> GetTopRow ();
	endLine = alarmListView -> GetBottomRow ();
    }
    else    // print entire list
    {
	startLine = 0;
	endLine = rowCount - 1;
    }

    // save current graphics environment
    vdraw::GSave ();

    app -> printer -> OpenDocument ();

    // get what is the current printer paper size
    vrect pageRect;
    pageRect.CopyFloat (app -> printer -> GetPaper ());
    int pageHeight = pageRect.GetHeight ();
    vdraw::RectsClip ((const vrect *)&pageRect, 1);
    vfont *font = vfont::FindScribed(vcharScribeLiteral("family:Courier-face:medium-size:8"));

    // Galaxy has a scaling problem when going from the screen to the printer
    // this bit of code is meant to compensate for this shortcoming
    int newSize = (8 * 72 + app -> logicalPixelsY/2) / app -> logicalPixelsY;
    static vfont *newFont = font -> Scale (font -> GetFace (), newSize);
    vdraw::SetFont(newFont);
    int fontHeight = newFont -> Height ();

    // need to scale these 2 values also
    int scaledHeight = (fontHeight * 72 + app -> logicalPixelsY/2) / app -> logicalPixelsY;
    int scaledIndent = (72 * 72 + app -> logicalPixelsY/2) / app -> logicalPixelsY; 

    int rectHeight = 0;
    int pageNum = 0;
    vstr *desc;
    vdate *now = new vdate ();
    now -> SetCurrent ();
    const vdateFormat *newdf = vdateFormat::GetTerse ();
    vstr *formattedHeader = vresourceGetString (app -> titlesResource,
						alvu_AlarmSummaryPrintTitle);
    vstr *dateHeader = vstrCloneScribed (now -> ScribeFormatted (newdf));

    for (i = startLine; i <= endLine; i++)
    {
	if (rectHeight == 0)
	{
	    if (i > startLine)
	    {
		app -> printer -> ClosePage ();
		app -> printer -> OpenPage ();
	    }
	    else
	    {
		app -> printer -> OpenPage ();
	    }
	    pageNum++;
	    // come down 1 inch (72pts/inch) from top of page
	    rectHeight = pageHeight - scaledIndent;
	    vstr *headerStr = vstrCloneScribed (vcharScribeFormat (formattedHeader,
								   dateHeader,
								   pageNum));
	    vdraw::MoveTo (scaledIndent, rectHeight - scaledHeight);
	    vdraw::Show (headerStr);
	    vstrDestroy (headerStr);
	    rectHeight -= 2 * scaledHeight;
	}

	// move to X offset indent, Y value
	vdraw::MoveTo (scaledIndent, rectHeight - scaledHeight);
	desc = vstrCloneScribed (vcharScribeSystem (ItemDesc (i)));
	vdraw::Show (desc);
	vstrDestroy (desc);
	rectHeight -= scaledHeight;
	if (rectHeight < scaledHeight)
	{
	    rectHeight = 0;
	}
    }
    app -> printer -> ClosePage ();
    app -> printer -> CloseDocument ();

    // restore graphics environment
    vdraw::GRestore();

    delete now;
    vstrDestroy (formattedHeader);
    vstrDestroy (dateHeader);
    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
			     alvu_PrintFinished, alvu_AlarmSummaryPrint);
}


vkindTYPED_FULL_DEFINITION(alarmSummaryCommands, vcommandSelector, "alarmSummaryCommands");
/*******************************************************************
*
*    menu selector commands for the alarm summary dialog 
*
*******************************************************************/

int alarmSummaryCommands::Issue (vdict *dict)
{
    int			command;
    AlarmSummaryDialog	*alarmSummaryDialog;
    PvData		*pvdata, *nextpvdata;
    int			startLine, endLine, rowCount;
    int			i;

    alarmSummaryDialog = (AlarmSummaryDialog *) dict -> GetElem (vname_Dialog);
    alvuApplication *app = alarmSummaryDialog -> GetApp ();
    vresource alvuRes = app -> alvuResource;
    AldlgListView *view = alarmSummaryDialog -> alarmListView;
    vlist *list = view -> GetList ();
    command = GetSelector ();

    switch (command)
    {
	case ALSUM_THISWIN_ITEM:
	case ALSUM_ENTIRESUM_ITEM:
	{
	    if (list != NULL)
	    {
		rowCount = list -> GetRowCount ();
	    }
	    if ((list == NULL) || (rowCount == 0))
	    {
		alvuMessageDialogCenter (alarmSummaryDialog, ALVU_INFO_NOTICE,
					 alvu_NothingToPrint);
		break;
	    }
#ifdef WIN32
	    if (command == ALSUM_THISWIN_ITEM)
	    {
		alarmSummaryDialog -> WindowsPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmSummaryDialog -> WindowsPrint (PRINT_ENTIRE_LIST);
	    }
#else
	    if (command == ALSUM_THISWIN_ITEM)
	    {
		alarmSummaryDialog -> UnixPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmSummaryDialog -> UnixPrint (PRINT_ENTIRE_LIST);
	    }
#endif WIN32
	    break;
	}

	case ALSUM_MODFILT_ITEM:
	{
	    vresource filtRes = vresourceGet (alvuRes, alvu_FilterDialog);
	    FilterDialog *filterDialog = new FilterDialog (filtRes);
	    filterDialog -> OpenFilter (app, NULL, &alarmSummaryDialog -> filter, MODIFY_FILTER);
	    filterDialog -> NewFilter (&alarmSummaryDialog -> filter);
	    delete filterDialog;
	    alarmSummaryDialog -> SetDialogTitle ();
	    if (alarmSummaryDialog -> pause)
	    {
		alarmSummaryDialog -> pause = FALSE;
		alarmSummaryDialog -> ShowPaused (alarmSummaryDialog -> pause);
	    }
	    alarmSummaryDialog -> Resync ();
	    break;
	}

	case ALSUM_SILENCE_ITEM:
	    BellTurnOff ();
	    break;

	case ALSUM_ACKALL_ITEM:
	    for (pvdata = shareHeader -> summaryHead; pvdata; pvdata = nextpvdata)
	    {
		nextpvdata = pvdata -> nextSummaryPv;

		if (!IsChannelFiltered (pvdata, &alarmSummaryDialog -> filter, pvdata -> unAckedSeverity))
		{
		    // Ack if necesary
		    if (AckAlarm (pvdata) != True)
		    {
			vstr *pv = vstrCloneScribed (vcharScribeSystem (pvdata -> pvName));
			alvuMessageDialogCenter (alarmSummaryDialog, ALVU_INFO_NOTICE,
						 alvu_NoAckPrivMsg, pv);
			vstrDestroy (pv);
		    }
		}
	    }
	    break;

	case ALSUM_ACKPAGE_ITEM:
	{
	    if (list == NULL)
	    {
		break;
	    }
	    char *pvname;
	    ListEntry *entry;
	    startLine = view -> GetTopRow ();
	    endLine = view -> GetBottomRow ();

	    for (i = startLine; i <= endLine; i++)
	    {
		entry = alarmSummaryDialog -> GetEntryPtr (i);
		if (entry)
		{
		    pvname = entry -> GetName ();
		    pvdata = FindChannel (treeRoot, StripWhitespace (pvname));
		    if (pvdata)
		    {
			if (AckAlarm (pvdata) != True)
			{
			    vstr *pv = vstrCloneScribed (vcharScribeSystem (pvname));
			    alvuMessageDialogCenter (alarmSummaryDialog, ALVU_INFO_NOTICE,
						     alvu_NoAckPrivMsg, pv);
			    vstrDestroy (pv);
			}
		    }
		}
	    }
	    break;
	}

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}


/*******************************************************************
*
*    Dynamic Group Dialog Class
*	subclassed from AlarmDialog
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(DynamicGroupDialog, AlarmDialog, "DynamicGroupDialog");

/*******************************************************************
*
*    DynamicGroupDialog, LoadInit ()
*
*******************************************************************/

void DynamicGroupDialog::LoadInit (vresource resource)
{
    AlarmDialog::LoadInit (resource);
}

/*******************************************************************
*
*    DynamicGroupDialog, destroy function
*
*******************************************************************/

void DynamicGroupDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    AlarmDialog::Destroy ();
}


/*******************************************************************
*
*    DynamicGroupDialog Open function
*
*******************************************************************/

void DynamicGroupDialog::AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt)
{
    app = alvuApp;
    filter = *filt;
    PopupMenu = NULL;
    groupLinkPtr = FindGroup (treeRoot, filter.group);

    BuildMenus ();
    SetDialogTitle ();

    ackable = True;
    hasdate = False;
    hasdesc = True;
    monitored = True;
    fileBased = False;
    pauseButton -> Disable ();

    AlarmDialog::AlarmDialogOpen (alvuApp, filt);

    displaySize = alarmListView -> GetListNumElements ();

    RefreshDisplay ();
}


/*******************************************************************
*
*    DynamicGroupDialog Build menu function 
*
*******************************************************************/

void DynamicGroupDialog::BuildMenus (void)
{
    vresource titles = app -> titlesResource;
    vmenu	*newMenu;
    vmenuItem	*menuItem;
    vstr	*label;

    // the print menu

    label = vresourceGetString (titles, alvu_PrintMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ThisWindow);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpThisWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_EntireGroup);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpEntireWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the filter menu

    label = vresourceGetString (titles, alvu_FilterMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ModifyFilter);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpModFilter);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the Group menu

    label = vresourceGetString (titles, alvu_GroupMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_DisplayTitle);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpDisplay);
    newMenu -> AppendItem (menuItem);
    // check if need to disable
    if (*groupLinkPtr -> groupParms.displayName)
	menuItem -> Enable ();
    else
	menuItem -> Disable ();
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ScriptTitle);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpScript);
    newMenu -> AppendItem (menuItem);
    // check if need to disable
    if (groupLinkPtr -> groupParms.manualScript ||
	groupLinkPtr -> groupParms.auxManualScript)
	menuItem -> Enable ();
    else
	menuItem -> Disable ();
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_GuidanceTitle);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpGuidance);
    newMenu -> AppendItem (menuItem);
    // check if need to disable
    if (groupLinkPtr -> groupParms.guidance != NULL)
	menuItem -> Enable ();
    else
	menuItem -> Disable ();
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the ControlMenu

    label = vresourceGetString (titles, alvu_ControlTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_DisablePtTitle);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpDisablePt);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_EnablePtTitle);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpEnablePt);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_DisableGrpTitle);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpDisableGrp);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_EnableGrpTitle);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpEnableGrp);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the ack all menu

    label = vresourceGetString (titles, alvu_AckAllMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_AckAll);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpAckAll);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_AckPage);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlGrpAckPage);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // load up menuNames with resources
    menuNames [ALGRP_THISWIN_ITEM]     = alvu_AlGrpThisWindow;
    menuNames [ALGRP_ENTIRESUM_ITEM]   = alvu_AlGrpEntireWindow;
    menuNames [ALGRP_MODFILT_ITEM]     = alvu_AlGrpModFilter;
    menuNames [ALGRP_DISPLAY_ITEM]     = alvu_AlGrpDisplay;
    menuNames [ALGRP_SCRIPT_ITEM]      = alvu_AlGrpScript;
    menuNames [ALGRP_GUIDANCE_ITEM]    = alvu_AlGrpGuidance;
    menuNames [ALGRP_DISABLE_PT_ITEM]  = alvu_AlGrpDisablePt;
    menuNames [ALGRP_ENABLE_PT_ITEM]   = alvu_AlGrpEnablePt;
    menuNames [ALGRP_DISABLE_GRP_ITEM] = alvu_AlGrpDisableGrp;
    menuNames [ALGRP_ENABLE_GRP_ITEM]  = alvu_AlGrpEnableGrp;
    menuNames [ALGRP_ACKALL_ITEM]      = alvu_AlGrpAckAll;
    menuNames [ALGRP_ACKPAGE_ITEM]     = alvu_AlGrpAckPage;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;
    // populate our dictionary with menu commands
    for (i = ALGRP_THISWIN_ITEM; i < ALGRP_ITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	dynamicGroupCommands *command = new dynamicGroupCommands;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (menuNames[i], command);
    }

    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}


/*******************************************************************
*
*    DynamicGroupDialog, Ctrl Function ()
*	param if >= 0 is row number in list
*
*******************************************************************/

int DynamicGroupDialog::CtrlFunc ( int cmd, int param)
{
    char	*ptName;
    PvData	*pvdata;
    TreeNode	*nodeptr;
    Filter	tempFilter;
    ListEntry	*entry;
    char	*guidptr;
    char	*dispptr;
    vstr	*pv;

    // get resource to alarmvu
    vresource alvuRes = app -> alvuResource;

    switch (cmd)
    {
	case ACKCLICK:
	{
	    ptName = ItemName (param);
	    if (nodeptr = FindGroup (groupLinkPtr, StripWhitespace (ptName)))
	    {
		if (HighestSeverity (nodeptr -> groupData.numUnackedBySeverity) ||
		    nodeptr -> groupData.numUnackedBySeverity [NO_ALARM])
		{
		    AckGroups (nodeptr);
		}
	    }
	    else if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		entry = GetEntryPtr (param);
		if (entry -> GetCurrAck ())
		{
		    if (entry -> AckMany ())
		    {
			vresource ackRes = vresourceGet (alvuRes, alvu_AckManyDialog);
			AckManyDialog *ackdlg = new AckManyDialog (ackRes);
			ackdlg -> OpenAckMany (app, ptName, &pvdata -> lastAlarm);
			int okToAck = ackdlg -> okToAck;
			delete ackdlg;
			// dont ACK if cancelled
			if (okToAck == vFALSE)
			{
			    break;
			}
		    }
		}
		if (AckAlarm (pvdata) != True)
		{
		    pv = vstrCloneScribed (vcharScribeSystem (ptName));
		    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
					     alvu_NoAckPrivMsg, pv);
		    vstrDestroy (pv);
		}
	    }
	    break;
	}
	case EXPANDCLICK:
	    if (expandDialogPtr != NULL)
	    {
		delete expandDialogPtr;
		expandDialogPtr = NULL;
	    }
	    ptName = ItemName (param);
	    if (nodeptr = FindGroup (groupLinkPtr, StripWhitespace (ptName)))
	    {
		// display expansion dialog for group
		vresource expRes = vresourceGet (alvuRes, alvu_TextDialog);
		expandDialogPtr = new TextDialog (expRes);
		expandDialogPtr -> OpenText (app, 
					     EXPAND_TEXT, 
					     AlarmFormat (&expandgroup, nodeptr, NULL),
					     NULL);
	    }
	    else if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		CaData ca;

		entry = GetEntryPtr (param);
		ca.CurStatus = entry -> GetCurrStatus ();
		ca.Value = entry -> GetValue ();
		entry -> SetDesc (AlarmFormat (&expandvalue, pvdata, &ca));

		// display expansion dialog
		vresource expRes = vresourceGet (alvuRes, alvu_TextDialog);
		expandDialogPtr = new TextDialog (expRes);
		expandDialogPtr -> OpenText (app, 
					     EXPAND_TEXT, 
					     AlarmFormat (&expandvalue, pvdata, &ca),
					     NULL);
	    }

	    break;

	case NAMECLICK:
	    ptName = ItemName (param);
	    vmenuItem *item;
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {

		GetAlarmInfo (ptName, &guidptr, &dispptr);

		// get Guidance menuitem
		item = PopupMenu -> GetItemAt (GUIDANCE_MENUITEM);
		// if guidance for point or for parent group, enable
		if ( (guidptr != 0) || 
		     HaveScript (pvdata -> parentGroup -> groupParms.guidance) )
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Display menuitem
		item = PopupMenu -> GetItemAt (DISPLAY_MENUITEM);
		if ( (dispptr != 0) || 
		      HaveScript (pvdata -> parentGroup -> groupParms.displayName))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Detail -enuitem
		item = PopupMenu -> GetItemAt (DETAIL_MENUITEM);
		item -> Enable ();
		// get Script menuitem
		item = PopupMenu -> GetItemAt (SCRIPT_MENUITEM);
		if ( HaveScript (pvdata -> parentGroup -> groupParms.manualScript) ||
		     HaveScript (pvdata -> parentGroup -> groupParms.auxManualScript))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Enable menuitem
		item = PopupMenu -> GetItemAt (ENABLE_MENUITEM);
		item -> Enable ();
		// get Disable menuitem
		item = PopupMenu -> GetItemAt (DISABLE_MENUITEM);
		item -> Enable ();
		// get Comment menuitem
		item = PopupMenu -> GetItemAt (COMMENT_MENUITEM);
		item -> Disable ();
		// get Help menuitem
		item = PopupMenu -> GetItemAt (HELP_MENUITEM);
		item -> Disable ();
	    }
	    else if (nodeptr = FindGroup (groupLinkPtr, StripWhitespace (ptName)))
	    {
		// get Guidance menuitem
		item = PopupMenu -> GetItemAt (GUIDANCE_MENUITEM);
		if (HaveScript (nodeptr -> groupParms.guidance))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Display menuitem
		item = PopupMenu -> GetItemAt (DISPLAY_MENUITEM);
		if (HaveScript (nodeptr -> groupParms.displayName))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Detail menuitem
		item = PopupMenu -> GetItemAt (DETAIL_MENUITEM);
		item -> Enable ();
		// get Script menuitem
		item = PopupMenu -> GetItemAt (SCRIPT_MENUITEM);
		if ( HaveScript (nodeptr -> groupParms.manualScript) ||
		     HaveScript (nodeptr -> groupParms.auxManualScript))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Enable menuitem
		item = PopupMenu -> GetItemAt (ENABLE_MENUITEM);
		item -> Enable ();
		// get Disable menuitem
		item = PopupMenu -> GetItemAt (DISABLE_MENUITEM);
		item -> Enable ();
		// get Comment menuitem
		item = PopupMenu -> GetItemAt (COMMENT_MENUITEM);
		item -> Disable ();
		// get Help menuitem
		item = PopupMenu -> GetItemAt (HELP_MENUITEM);
		item -> Disable ();
	    }
	    else
	    {
		return False;
	    }
	    return True;

	case SUMMARYCLICK:
	{
	    ptName = ItemName (param);
	    tempFilter = filter;
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		strncpy (tempFilter.group, pvdata -> parentGroup -> groupName, 
			 sizeof (tempFilter.group));
		tempFilter.group [sizeof (filter.group) - 1] = 0;
	    }
	    else
	    {
		strncpy (tempFilter.group, ptName, sizeof (tempFilter.group));
		tempFilter.group [sizeof (tempFilter.group) - 1] = 0;
	    }
	    if (alarmSummaryDialogPtr != NULL)
	    {
		delete alarmSummaryDialogPtr;
	    }
	    vresource alrmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
	    alarmSummaryDialogPtr = new AlarmSummaryDialog (alrmDlgRes);
	    alarmSummaryDialogPtr -> AlarmDialogOpen (app, &tempFilter);
	    break;
	}

	case GROUPCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		strncpy (filter.group, pvdata -> parentGroup -> groupName,
			 sizeof (filter.group));
		filter.group [sizeof (filter.group) - 1] = 0;
	    }
	    else
	    {
		strncpy (filter.group, ptName, sizeof (filter.group));
		filter.group [sizeof (filter.group) - 1] = 0;
	    }
	    groupLinkPtr = FindGroup (treeRoot, filter.group);
	    SetDialogTitle ();
	    RefreshDisplay ();
	    break;

	case GUIDECLICK:
	    ptName = ItemName (param);
	    if (guidanceDialogPtr != NULL)
	    {
		delete guidanceDialogPtr;
		guidanceDialogPtr = NULL;
	    }
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		// get any guidance on the point
		GetAlarmInfo (ptName, &guidptr, &dispptr);
		if (guidptr == NULL)
		{
		    guidptr = pvdata -> parentGroup -> groupParms.guidance;
		}
	    }
	    else
	    {
		nodeptr = FindGroup (groupLinkPtr, StripWhitespace (ptName));
		guidptr = nodeptr -> groupParms.guidance;
	    }
	    if (guidptr && *guidptr)
	    {
		vresource guidRes = vresourceGet (alvuRes, alvu_TextDialog);
		guidanceDialogPtr = new TextDialog (guidRes);
		guidanceDialogPtr -> OpenText (app, GUIDANCE_TEXT, 
					       guidptr, ptName);
	    }
	    break;

	case DISPLAYCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		SpawnPointDisplay (pvdata);
	    }
	    else if (nodeptr = FindGroup (treeRoot, StripWhitespace (ptName)))
	    {
		SpawnGroupDisplay (nodeptr);
	    }
	    break;

	case DETAILCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		SpawnDetail (ptName);
	    }
	    break;

	case COMMANDCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		SpawnScript (pvdata);
	    }
	    else if (nodeptr = FindGroup (treeRoot, StripWhitespace (ptName)))
	    {
		SpawnGroupScript (nodeptr);
	    }
	    break;

	case ENABLECLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		if (EnablePoint (pvdata) != True)
		{
		    pv = vstrCloneScribed (vcharScribeSystem (ptName));
		    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
					     alvu_NoEnablePriv, pv);
		    vstrDestroy (pv);
		}
	    }
	    else if (nodeptr = FindGroup (groupLinkPtr, StripWhitespace (ptName)))
	    {
		EnableGroups (nodeptr);
	    }
	    break;

	case DISABLECLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		if (DisablePoint (pvdata) != True)
		{
		    pv = vstrCloneScribed (vcharScribeSystem (ptName));
		    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
					     alvu_NoDisablePriv, pv);
		    vstrDestroy (pv);
		}
	    }
	    else if (nodeptr = FindGroup (groupLinkPtr, StripWhitespace (ptName)))
	    {
		DisableGroups (nodeptr);
	    }
	    break;

	case CAMONITOR:
	    caUpdateReq = True;
	    ptName = ItemName (param);
	    entry = GetEntryPtr (param);
	    pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName));
	    if (pvdata && !IsChannelFiltered (pvdata, &this -> filter, entry -> GetCurrSeverity ()))
	    {
		CaData ca;

		ca.CurStatus = entry -> GetCurrStatus ();
		ca.Value = entry -> GetValue ();
		entry -> SetDesc (AlarmFormat (&displayvalue, pvdata, &ca));
		entry -> SetColor (entry -> GetCurrSeverity () +
				   entry -> GetCurrAck () ? NUM_SEV_COLORS : 0);
	    }
	    break;

	case UPCLICK:
	    if (groupLinkPtr -> parent)
	    {
		groupLinkPtr = groupLinkPtr -> parent;
	    }
	    strncpy (filter.group, groupLinkPtr -> groupName, sizeof (filter.group));
	    filter.group [sizeof (filter.group)-1] = 0;
	    filter.modified = TRUE;
	    SetDialogTitle ();
	    RefreshDisplay ();
	    break;

	case TOPCLICK:
	    groupLinkPtr = treeRoot;
	    strncpy (filter.group, groupLinkPtr -> groupName, sizeof (filter.group));
	    filter.group [sizeof (filter.group)-1] = 0;
	    filter.modified = TRUE;
	    SetDialogTitle ();
	    RefreshDisplay ();
	    break;
    }
    return True;
}


/*******************************************************************
*
*    DynamicGroupDialog, RefreshDisplay ()
*
*******************************************************************/

void DynamicGroupDialog::RefreshDisplay (void)
{
    int		color;
    int		displayAll = False;
    int		itemix = 0;
    PvData	*pvdata;
    TreeNode	*nodeptr;
    ListEntry	*entry;

    for (nodeptr = groupLinkPtr -> leftChild; nodeptr; nodeptr = nodeptr -> rightSibling)
    {
	if (!IsGroupFiltered (nodeptr, &filter))
	{
 	    if (itemix < NumItems())
	    {
		entry = GetEntryPtr (itemix);
		if (entry != NULL)
		{
		    if (strcmp (entry -> GetName (), nodeptr -> groupName))
		    {
			displayAll = True;
		    }
		}
		else
		{
		    displayAll = True;
		}
	    }
	    else
	    {
		displayAll = True;
	    }
	    itemix++;
	}
    }
    for (pvdata = groupLinkPtr -> pvList; pvdata; pvdata = pvdata -> nextPv)
    {
	if (!IsChannelFiltered (pvdata, &filter))
	{
 	    if (itemix < NumItems())
	    {
		entry = GetEntryPtr (itemix);
		if (entry != NULL)
		{
		    if (strcmp (entry -> GetName (), pvdata -> pvName))
		    {
			displayAll = True;
		    }
		}
		else
		{
		    displayAll = True;
		}
	    }
	    else
	    {
		displayAll = True;
	    }
	    itemix++;
	}
    }
    if (itemix != NumItems())
    {
	displayAll = True;
    }

    if (displayAll)
    {
	ClearLines ();
	for (nodeptr = groupLinkPtr -> leftChild; nodeptr; nodeptr = nodeptr -> rightSibling)
	{
	    if (!IsGroupFiltered (nodeptr, &filter))
	    {
		color = 0;
		AddLine ( color, "", nodeptr -> groupName, nodeptr -> groupName,
			  &nodeptr -> groupData.lastStatusChange);
	    }
	}

	for (pvdata = groupLinkPtr -> pvList; pvdata; pvdata = pvdata -> nextPv)
	{
	    if ( !IsChannelFiltered (pvdata, &filter))
	    {
		color = 0;
		AddLine ( color, False, "", pvdata -> pvName, pvdata -> pvName,
			  &pvdata -> lastAlarm, True /*monitor this*/);
	    }
	}

    }
    if (dataList != NULL)
    {
	alarmListView -> SetList (dataList);
	vlistSelection *selected = alarmListView -> GetSelection ();
	dataList -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	dataList -> SetAtLeastOneSelection (selected, vFALSE);
	SetPosition (0);
	Resync ();
    }
}


/*******************************************************************
*
*    Dynamic Group Dialog, Resync ()
*
*******************************************************************/

void DynamicGroupDialog::Resync (void)
{
    TreeNode	*nodeptr;
    PvData	*pvdata;
    int		color;
    int		itemix = 0;
    ListEntry	*entry;

    for (nodeptr = groupLinkPtr -> leftChild;
	 itemix < NumItems() && nodeptr;
	 nodeptr = nodeptr -> rightSibling)
    {
	if (!IsGroupFiltered (nodeptr, &filter))
	{
	    if (nodeptr->groupData.notRTNnotACK[MAJOR_ALARM])
		color = MAJOR_ALARM + NUM_SEV_COLORS;
	    else if (nodeptr->groupData.notRTNandACK[MAJOR_ALARM])
		color = MAJOR_ALARM;
	    else if (nodeptr->groupData.notRTNnotACK[MINOR_ALARM])
		color = MINOR_ALARM + NUM_SEV_COLORS;
	    else if (nodeptr->groupData.notRTNandACK[MINOR_ALARM])
		color = MINOR_ALARM;
	    else if (nodeptr->groupData.notRTNnotACK[INFO_ALARM])
		color = INFO_ALARM + NUM_SEV_COLORS;
	    else if (nodeptr->groupData.notRTNandACK[INFO_ALARM])
		color = INFO_ALARM;
	    else if (nodeptr->groupData.notRTNnotACK[EVENT_ALARM])
		color = EVENT_ALARM + NUM_SEV_COLORS;
	    else if (nodeptr->groupData.notRTNandACK[EVENT_ALARM])
		color = EVENT_ALARM;
	    else if ((nodeptr->groupData.RTNnotACK[MAJOR_ALARM])
	      || (nodeptr->groupData.RTNnotACK[MINOR_ALARM])
	      || (nodeptr->groupData.RTNnotACK[INFO_ALARM])
	      || (nodeptr->groupData.RTNnotACK[EVENT_ALARM]))
		color = NO_ALARM + NUM_SEV_COLORS;
	    else
		color = NO_ALARM;

	    entry = GetEntryPtr (itemix);
	    entry -> SetDesc (AlarmFormat (&displaygroup, nodeptr, NULL));
	    entry -> SetColor (color);
	    itemix++;
	}
    }

    for (pvdata = groupLinkPtr -> pvList; 
	 itemix < NumItems() && pvdata; 
	 pvdata = pvdata -> nextPv)
    {
	entry = GetEntryPtr (itemix);
	if ( !IsChannelFiltered (pvdata, &filter))
	{
	    if (pvdata -> RTN)
	    {
		// blink, NO_ALARM 
		color = NUM_SEV_COLORS + NO_ALARM;
	    }
	    else if ( (pvdata -> unAckedSeverity != NO_ALARM) && (!pvdata -> Ack) )
	    {
		// most recent unacked sev > NO_ALARM, and notacked
		color = NUM_SEV_COLORS + pvdata -> unAckedSeverity;
	    }
	    else
	    {
		// no blink, alarm state
		color = pvdata -> unAckedSeverity;
	    }
	    entry -> SetColor (color);

	    /* mostly done in CA monitor callback */
	    entry -> SetAckMany (!pvdata -> Ack && (pvdata -> lastAck.secPastEpoch != 0));
	    itemix++;
	}
    }
    // cause display items to be redrawn
    alarmListView -> Inval (vwindowINVAL_OPAQUE | vwindowINVAL_UPDATE);
}


/*******************************************************************
*
*    DynamicGroupDialog set title function to be written by subclasses
*
*******************************************************************/

void DynamicGroupDialog::SetDialogTitle (void)
{
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    vstr *format = vresourceGetString (app -> titlesResource, alvu_DynamicGroupTitle);
    vstr *group = vstrCloneScribed (vcharScribeSystem (filter.group));
    vstr *mod = vstrCloneScribed (vcharScribeLiteral (filter.modified ? "*" : ""));
    vstr *name = vstrCloneScribed (vcharScribeSystem (filter.name));
    title = vstrCloneScribed (vcharScribeFormat (format, group, mod, name));
    SetTitle (title);
    vstrDestroy (format);
    vstrDestroy (group);
    vstrDestroy (mod);
    vstrDestroy (name);
}

/*******************************************************************
*
*    DynamicGroupDialog handle blinking
*
*******************************************************************/

void DynamicGroupDialog::UpdateBlink (void)
{
    if (IsOpen ())
    {
    alarmListView -> Inval (vwindowINVAL_OPAQUE | vwindowINVAL_UPDATE);
/*    int i;
    for (i = alarmListView -> GetTopRow (); i <= alarmListView -> GetBottomRow (); i++)
	alarmListView -> InvalViewRow (i);*/
    AlarmDialog::UpdateBlink ();
    }
}


/*******************************************************************
*
*    AlarmDialog unix print override of base class function
*
*******************************************************************/

void DynamicGroupDialog::UnixPrint (int whatToPrint)
{
    FILE    *printFileDesc;
    char    printFileName [30];
    int	    startLine, endLine;
    int	    i;
    vlist   *list = alarmListView -> GetList ();
    int	    rowCount = list -> GetRowCount ();

    if (whatToPrint == PRINT_THIS_WINDOW)
    {
	startLine = alarmListView -> GetTopRow ();
	endLine = alarmListView -> GetBottomRow ();
    }
    else    // print entire list
    {
	startLine = 0;
	endLine = rowCount - 1;
    }

    strcpy (printFileName, PRINT_FILE_TEMP);
    if ( !mktemp (printFileName) || !(printFileDesc = fopen (printFileName, "w")))
    {
	vstr *prtFileTmp = vstrCloneScribed (vcharScribeLiteral (PRINT_FILE_TEMP));
	alvuMessageDialogCenter (this, ALVU_INFO_NOTICE, alvu_TempPrtFileErr, prtFileTmp);
	vstrDestroy (prtFileTmp);
    }
    else
    {
	for (i = startLine; i <= endLine; i++)
	{
	    fprintf (printFileDesc, "%s\n", ItemName (i));
	}
	fclose (printFileDesc);
	vstr *capFormat = vresourceGetString (app -> messagesResource, alvu_GrpUnixPrtTitle);
	vstr *grp = vstrCloneScribed (vcharScribeSystem (groupLinkPtr -> groupName));
	vstr *headerStr = vstrCloneScribed (vcharScribeFormat (capFormat, grp));
	vcharExportBounded (vcharScribeSystemExport (headerStr),
			    caption,
			    sizeof (caption));
	caption [sizeof (caption) - 1] = 0;
	vstrDestroy (capFormat);
	vstrDestroy (grp);
	vstrDestroy (headerStr);
	if (strlen (linePrinter))
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"%s\" %s |lpr -P\"%s\"", 
			      caption, printFileName, linePrinter);
#else
	    sprintf (txtBuf, "lpr -p -r -P\"%s\" -T\"%s\" %s", 
			      linePrinter, caption, printFileName);
#endif SOLARIS
	}
	else
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"%s\" %s |lpr", caption, printFileName);
#else
	    sprintf (txtBuf, "lpr -p -r -T\"%s\" %s", caption, printFileName);
#endif SOLARIS
	}
	system (txtBuf);
#ifdef SOLARIS
	sprintf (txtBuf, "rm %s", printFileName);
	system (txtBuf);
#endif SOLARIS
    }
}


/*******************************************************************
*
*    AlarmDialog windows print override of base class function
*
*******************************************************************/

void DynamicGroupDialog::WindowsPrint (int whatToPrint)
{
    int startLine, endLine;
    int i;
    vlist *list = alarmListView -> GetList ();
    int rowCount = list -> GetRowCount ();

    if (app -> printer == NULL)
    {
	alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_NoPrinterMsg);
	return;
    }

    if (whatToPrint == PRINT_THIS_WINDOW)
    {
	startLine = alarmListView -> GetTopRow ();
	endLine = alarmListView -> GetBottomRow ();
    }
    else    // print entire list
    {
	startLine = 0;
	endLine = rowCount - 1;
    }

    // save current graphics environment
    vdraw::GSave ();

    app -> printer -> OpenDocument ();

    // get what is the current printer paper size
    vrect pageRect;
    pageRect.CopyFloat (app -> printer -> GetPaper ());
    int pageHeight = pageRect.GetHeight ();
    vdraw::RectsClip ((const vrect *)&pageRect, 1);
    vfont *font = vfont::FindScribed(vcharScribeLiteral("family:Courier-face:medium-size:8"));

    // Galaxy has a scaling problem when going from the screen to the printer
    // this bit of code is meant to compensate for this shortcoming
    int newSize = (8 * 72 + app -> logicalPixelsY / 2) / app -> logicalPixelsY;
    static vfont *newFont = font -> Scale (font -> GetFace (), newSize);
    vdraw::SetFont(newFont);
    int fontHeight = newFont -> Height ();

    // need to scale these 2 values also
    int scaledHeight = (fontHeight * 72 + app -> logicalPixelsY / 2) / app -> logicalPixelsY;
    int scaledIndent = (72 * 72 + app -> logicalPixelsY / 2) / app -> logicalPixelsY; 

    int rectHeight = 0;
    int pageNum = 0;
    vstr *desc;
    vdate *now = new vdate ();
    now -> SetCurrent ();
    const vdateFormat *newdf = vdateFormat::GetTerse ();
    vstr *formattedHeader = vresourceGetString (app -> titlesResource,
						alvu_GrpWinPrtTitle);
    vstr *dateHeader = vstrCloneScribed (now -> ScribeFormatted (newdf));

    for (i = startLine; i <= endLine; i++)
    {
	if (rectHeight == 0)
	{
	    if (i > startLine)
	    {
		app -> printer -> ClosePage ();
		app -> printer -> OpenPage ();
	    }
	    else
	    {
		app -> printer -> OpenPage ();
	    }
	    pageNum++;
	    // come down 1 inch (72pts/inch) from top of page
	    rectHeight = pageHeight - scaledIndent;
	    vstr *groupStr = vstrCloneScribed (vcharScribeSystem (groupLinkPtr -> groupName));
	    vstr *headerStr = vstrCloneScribed (vcharScribeFormat (formattedHeader,
								   groupStr,
								   dateHeader,
								   pageNum));
	    vdraw::MoveTo (scaledIndent, rectHeight - scaledHeight);
	    vdraw::Show (headerStr);
	    vstrDestroy (headerStr);
	    vstrDestroy (groupStr);
	    rectHeight -= 2 * scaledHeight;
	}

	// move to X offset indent, Y value
	vdraw::MoveTo (scaledIndent, rectHeight - scaledHeight);
	desc = vstrCloneScribed (vcharScribeSystem (ItemName (i)));
	vdraw::Show (desc);
	vstrDestroy (desc);
	rectHeight -= scaledHeight;
	if (rectHeight < scaledHeight)
	{
	    rectHeight = 0;
	}
    }
    app -> printer -> ClosePage ();
    app -> printer -> CloseDocument ();

    // restore graphics environment
    vdraw::GRestore();

    delete now;
    vstrDestroy (formattedHeader);
    vstrDestroy (dateHeader);
    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
			     alvu_PrintFinished, alvu_DynamicGroupPrint);
}


vkindTYPED_FULL_DEFINITION(dynamicGroupCommands, vcommandSelector, "dynamicGroupCommands");
/*******************************************************************
*
*    menu selector commands for the dynamic group dialog
*
*******************************************************************/

int dynamicGroupCommands::Issue (vdict *dict)
{
    int			command;
    DynamicGroupDialog	*dynamicGroupDialog;
    ComboDialog		*comboDialog;
    PvData		*pvdata;
    TreeNode		*nodeptr;
    int			startLine, endLine, rowCount;
    int			i;
    vchar		selectedItem[256];
    char		item[256];
    vstr		*strPtr;

    dynamicGroupDialog = (DynamicGroupDialog *) dict -> GetElem (vname_Dialog);
    alvuApplication *app = dynamicGroupDialog -> GetApp ();
    vresource alvuRes = app -> alvuResource;
    vresource comboRes = vresourceGet (alvuRes, alvu_ComboDialog);
    AldlgListView *view = dynamicGroupDialog -> alarmListView;
    vlist *list = view -> GetList ();
    command = GetSelector ();

    switch (command)
    {
	case ALGRP_THISWIN_ITEM:
	case ALGRP_ENTIRESUM_ITEM:
	{
	    if (list != NULL)
	    {
		rowCount = list -> GetRowCount ();
	    }
	    if ((list == NULL) || (rowCount == 0))
	    {
		alvuMessageDialogCenter (dynamicGroupDialog, ALVU_INFO_NOTICE,
					 alvu_NothingToPrint);
		break;
	    }
#ifdef WIN32
	    if (command == ALGRP_THISWIN_ITEM)
	    {
		dynamicGroupDialog -> WindowsPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		dynamicGroupDialog -> WindowsPrint (PRINT_ENTIRE_LIST);
	    }
#else
	    if (command == ALGRP_THISWIN_ITEM)
	    {
		dynamicGroupDialog -> UnixPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		dynamicGroupDialog -> UnixPrint (PRINT_ENTIRE_LIST);
	    }
#endif WIN32
	    break;
	}

	case ALGRP_MODFILT_ITEM:
	{
	    vresource filtRes = vresourceGet (alvuRes, alvu_FilterDialog);
	    FilterDialog *filterDialog = new FilterDialog (filtRes);
	    filterDialog -> OpenFilter (app, NULL, &dynamicGroupDialog -> filter, 
					MODIFY_FILTER);
	    filterDialog -> NewFilter (&dynamicGroupDialog -> filter);
	    delete filterDialog;
	    dynamicGroupDialog -> SetDialogTitle ();
	    dynamicGroupDialog -> RefreshDisplay ();
	    break;
	}

	case ALGRP_DISPLAY_ITEM:
	    if (*dynamicGroupDialog -> groupLinkPtr -> groupParms.displayName)
	    {
		SpawnGroupDisplay (dynamicGroupDialog -> groupLinkPtr);
	    }
	    break;

	case ALGRP_SCRIPT_ITEM:
	    SpawnGroupScript (dynamicGroupDialog -> groupLinkPtr);
	    break;

	case ALGRP_GUIDANCE_ITEM:
	    if (guidanceDialogPtr != NULL)
	    {
		delete guidanceDialogPtr;
		guidanceDialogPtr = NULL;
	    }
	    if (dynamicGroupDialog -> groupLinkPtr -> groupParms.guidance)
	    {
		vresource guidRes = vresourceGet (alvuRes, alvu_TextDialog);
		guidanceDialogPtr = new TextDialog (guidRes);
		guidanceDialogPtr -> OpenText (app, 
					       GUIDANCE_TEXT, 
					       dynamicGroupDialog -> groupLinkPtr -> groupParms.guidance,
					       dynamicGroupDialog -> groupLinkPtr -> groupName);
	    }
	    break;

	case ALGRP_DISABLE_PT_ITEM:
	    comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (app, DISABLE_POINT_LIST, channelNames, NULL);
	    vcharCopyBounded (comboDialog -> GetSelected (), selectedItem, 
			      AL_POINTNAME_SIZE);
	    selectedItem [AL_POINTNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedItem[0])
	    {
		vcharExportBounded (vcharScribeSystemExport (selectedItem),
				    item,
				    AL_POINTNAME_SIZE);
		item [AL_POINTNAME_SIZE] = 0;
		if (pvdata = FindChannel (treeRoot, item))
		{
		    if (DisablePoint (pvdata) != True)
		    {
			strPtr = vstrCloneScribed (vcharScribeSystem (item));
			alvuMessageDialogCenter (dynamicGroupDialog, ALVU_INFO_NOTICE,
						 alvu_NoDisablePriv, strPtr);
			vstrDestroy (strPtr);
		    }
		}
	    }
	    break;

	case ALGRP_ENABLE_PT_ITEM:
	    comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (app, ENABLE_POINT_LIST, channelNames, NULL);
	    vcharCopyBounded (comboDialog -> GetSelected (), selectedItem, 
			      AL_POINTNAME_SIZE);
	    selectedItem [AL_POINTNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedItem[0])
	    {
		vcharExportBounded (vcharScribeSystemExport (selectedItem),
				    item,
				    AL_POINTNAME_SIZE);
		item [AL_POINTNAME_SIZE] = 0;
		if (pvdata = FindChannel (treeRoot, item))
		{
		    if (EnablePoint (pvdata) != True)
		    {
			strPtr = vstrCloneScribed (vcharScribeSystem (item));
			alvuMessageDialogCenter (dynamicGroupDialog, ALVU_INFO_NOTICE,
						 alvu_NoEnablePriv, strPtr);
			vstrDestroy (strPtr);
		    }
		}
	    }
	    break;

	case ALGRP_DISABLE_GRP_ITEM:
	    comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (app, DISABLE_GROUP_LIST, 
				      groupNames, 
				      dynamicGroupDialog -> groupLinkPtr -> groupName);
	    vcharCopyBounded (comboDialog -> GetSelected (), selectedItem, 
			      AL_GROUPNAME_SIZE);
	    selectedItem [AL_GROUPNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedItem[0])
	    {
		vcharExportBounded (vcharScribeSystemExport (selectedItem),
				    item,
				    AL_GROUPNAME_SIZE);
		item [AL_GROUPNAME_SIZE] = 0;
		if (nodeptr = FindGroup (treeRoot, item))
		{
		    DisableGroups (nodeptr);
		}
	    }
	    break;

	case ALGRP_ENABLE_GRP_ITEM:
	    comboDialog = new ComboDialog (comboRes);
	    comboDialog -> OpenCombo (app, ENABLE_GROUP_LIST, 
				      groupNames,
				      dynamicGroupDialog -> groupLinkPtr -> groupName);
	    vcharCopyBounded (comboDialog -> GetSelected (), selectedItem, 
			      AL_GROUPNAME_SIZE);
	    selectedItem [AL_GROUPNAME_SIZE] = 0;
	    delete comboDialog;
	    if (selectedItem[0])
	    {
		vcharExportBounded (vcharScribeSystemExport (selectedItem),
				    item,
				    AL_GROUPNAME_SIZE);
		item [AL_GROUPNAME_SIZE] = 0;
		if (nodeptr = FindGroup (treeRoot, item))
		{
		    EnableGroups (nodeptr);
		}
	    }
	    break;

	case ALGRP_ACKALL_ITEM:
	    if (list == NULL)
	    {
		break;
	    }
	    if (HighestSeverity (dynamicGroupDialog -> groupLinkPtr -> groupData.numUnackedBySeverity) ||
		dynamicGroupDialog -> groupLinkPtr -> groupData.numUnackedBySeverity [NO_ALARM])
	    {
		AckGroups (dynamicGroupDialog -> groupLinkPtr);
	    }
	    break;

	case ALGRP_ACKPAGE_ITEM:
	{
	    if (list == NULL)
	    {
		break;
	    }
	    char *pvname;
	    ListEntry *entry;
	    startLine = view -> GetTopRow ();
	    endLine = view -> GetBottomRow ();

	    for (i = startLine; i <= endLine; i++)
	    {
		entry = (ListEntry *)(list -> GetCellValue (i, 0));
		if (entry)
		{
		    pvname = entry -> GetName ();
		    pvdata = FindChannel (treeRoot, StripWhitespace (pvname));
		    if (pvdata)
		    {
			if (AckAlarm (pvdata) != True)
			{
			    vstr *pv = vstrCloneScribed (vcharScribeSystem (pvname));
			    alvuMessageDialogCenter (dynamicGroupDialog, ALVU_INFO_NOTICE,
						     alvu_NoAckPrivMsg, pv);
			    vstrDestroy (pv);
			}
		    }
		}
	    }
	    break;
	}

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}


/*******************************************************************
*
*    Log File Dialog Class
*	subclassed from AlarmDialog
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(LogFileDialog, AlarmDialog, "LogFileDialog");

// gateway function needed to satisfy compiler with call 
// to function Log_RecordsByTime () see alarmfil.c
// used by LogFileDialog and AlarmSystemDialog

int CheckCancelGW (AlarmDialog *dlg)
{
    return (dlg -> CheckCancel());
}

/*******************************************************************
*
*    LogFileDialog, LoadInit ()
*
*******************************************************************/

void LogFileDialog::LoadInit (vresource resource)
{
    displayTitle = NULL;
    printTitle = NULL;
    fileButtonContainer = FileButtonContainer::RequestSubLoad (this,
					    alvu_BlankContainer);
    cancelDataContainer = CancelDataButtonContainer::RequestSubLoad (this,
					    alvu_ButtonContainer);
    AlarmDialog::LoadInit (resource);

    fileButtonContainer -> firstButton    -> StoreAlarmDialog (this);
    fileButtonContainer -> pageupButton   -> StoreAlarmDialog (this);
    fileButtonContainer -> lineupButton   -> StoreAlarmDialog (this);
    fileButtonContainer -> linedownButton -> StoreAlarmDialog (this);
    fileButtonContainer -> pagedownButton -> StoreAlarmDialog (this);
    fileButtonContainer -> lastButton     -> StoreAlarmDialog (this);
    cancelDataContainer -> cancelDataButton -> StoreAlarmDialog (this);
}

/*******************************************************************
*
*    LogFileDialog, destroy function
*
*******************************************************************/

void LogFileDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    if (displayTitle != NULL)
    {
	vstrDestroy (displayTitle);
    }
    if (printTitle != NULL)
    {
	vstrDestroy (printTitle);
    }
    AlarmDialog::Destroy ();
}


/*******************************************************************
*
*    Log File Dialog Class, Open
*
*******************************************************************/

void LogFileDialog::AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt)
{
    app = alvuApp;
    filter = *filt;
    PopupMenu = NULL;

    SetDialogTitle ();

    // the args for base class AlarmDialog
    ackable = False;
    hasdate = False;
    hasdesc = False;
    monitored = False;
    fileBased = True;

    inMove = False;	// dont refresh if already doing so
    inRefresh = True;	// disable refresh during creation
    pause = False;

    Log_SetFilter (&filter, indexFileType); // see alarmfil.c
    groupLinkPtr = FindGroup (treeRoot, filter.group);

    startName [0] = 0;
    // turn off the scroll bar for the listview, we are using
    // buttons to control the listview
    alarmListView -> SetVertBar (vFALSE);

    cancelDataContainer -> cancelDataButton -> Disable ();
    cancelDataContainer -> cancelDataButton -> Hide ();

    AlarmDialog::AlarmDialogOpen (alvuApp, filt);

    displaySize = alarmListView -> GetListNumElements ();
    
    filePosition = 100;
    inRefresh = False;
    Resync ();

    ShowPaused (pause);
}


/*******************************************************************
*
*    LogFileDialog, Ctrl Function ()
*	param if >= 0 is row number in list
*
*******************************************************************/

int LogFileDialog::CtrlFunc ( int cmd, int param)
{
    char	*ptName;
    PvData	*pvdata;
    Filter	tempFilter;
    ListEntry	*entry;
    TS_STAMP	*evtime;
    char	*guidptr;
    char	*dispptr;
    vstr	*pv;

    // get resource to alarmvu
    vresource alvuRes = app -> alvuResource;

    switch (cmd)
    {
	case MOVECLICK:
	    if (inMove == False)
	    {
		inMove = True;
		filePosition = param;
		RefreshDisplay ();
		inMove = False;
	    }
	    break;

	case WINCLOSE:
	    if (inRefresh == True)
	    {
		return (False);
	    }
	    else
	    {
		break;
	    }

	case EXITCLICK:
	    if (inRefresh == True)
	    {
		return (False);
	    }
	    else
	    {
		break;
	    }

	case EXPANDCLICK:
	{
	    if (expandDialogPtr != NULL)
	    {
		delete expandDialogPtr;
		expandDialogPtr = NULL;
	    }
	    ptName = ItemName (param);
	    entry = GetEntryPtr (param);
	    evtime = entry -> GetEventTime ();
	    if (Log_RecordByTime (ptName, evtime,&alarmLogRecord) != RECFOUND)
	    {
		break;
	    }
	    // display expansion dialog
	    vresource expRes = vresourceGet (alvuRes, alvu_TextDialog);
	    expandDialogPtr = new TextDialog (expRes);
	    if (strcmp (alarmLogRecord.status, "ACK") == 0)
	    {
		expandDialogPtr -> OpenText (app, 
					     EXPAND_TEXT, 
					     AlarmFormat (&expandack, NULL, &alarmLogRecord),
					     NULL);
	    }
	    else 
	    {
		expandDialogPtr -> OpenText (app, 
					     EXPAND_TEXT, 
					     AlarmFormat (&expandalarm, NULL, &alarmLogRecord),
					     NULL);
	    }

	    break;
	}

	case NAMECLICK:
	{
   	    ptName = ItemName (param);
	    vmenuItem *item;
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {

		GetAlarmInfo (ptName, &guidptr, &dispptr);

		// get Guidance menuitem
		item = PopupMenu -> GetItemAt (GUIDANCE_MENUITEM);
		// if guidance for point or for parent group, enable
		if ( (guidptr != 0) || 
		     HaveScript (pvdata -> parentGroup -> groupParms.guidance) )
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Display menuitem
		item = PopupMenu -> GetItemAt (DISPLAY_MENUITEM);
		if ( (dispptr != 0) || 
		      HaveScript (pvdata -> parentGroup -> groupParms.displayName))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Detail -enuitem
		item = PopupMenu -> GetItemAt (DETAIL_MENUITEM);
		item -> Enable ();
		// get Script menuitem
		item = PopupMenu -> GetItemAt (SCRIPT_MENUITEM);
		if ( HaveScript (pvdata -> parentGroup -> groupParms.manualScript) ||
		     HaveScript (pvdata -> parentGroup -> groupParms.auxManualScript))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Enable menuitem
		item = PopupMenu -> GetItemAt (ENABLE_MENUITEM);
		item -> Enable ();
		// get Disable menuitem
		item = PopupMenu -> GetItemAt (DISABLE_MENUITEM);
		item -> Enable ();
		// get Comment menuitem
		item = PopupMenu -> GetItemAt (COMMENT_MENUITEM);
		item -> Enable ();
		// get Help menuitem
		item = PopupMenu -> GetItemAt (HELP_MENUITEM);
		item -> Disable ();
	    }
	    else
	    {
		return False;
	    }
	    return True;
	}

	case SUMMARYCLICK:
	{
	    ptName = ItemName (param);
	    tempFilter = filter;
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		strncpy (tempFilter.group, pvdata -> parentGroup -> groupName, 
			 sizeof (tempFilter.group));
		tempFilter.group [sizeof (filter.group) - 1] = 0;
		if (alarmSummaryDialogPtr != NULL)
		{
		    delete alarmSummaryDialogPtr;
		}
		vresource alrmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
		alarmSummaryDialogPtr = new AlarmSummaryDialog (alrmDlgRes);
		alarmSummaryDialogPtr -> AlarmDialogOpen (app, &tempFilter);
	    }
	    break;
	}

	case GROUPCLICK:
	{
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		tempFilter = filter;
		strncpy (tempFilter.group, pvdata -> parentGroup -> groupName,
			 sizeof (tempFilter.group));
		tempFilter.group [sizeof (filter.group) - 1] = 0;
		if (dynamicGroupDialogPtr != NULL)
		{
		    delete dynamicGroupDialogPtr;
		}
		vresource alrmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
		dynamicGroupDialogPtr = new DynamicGroupDialog (alrmDlgRes);
		dynamicGroupDialogPtr -> AlarmDialogOpen (app, &tempFilter);
	    }
	    break;
	}

	case GUIDECLICK:
	{
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		if (guidanceDialogPtr != NULL)
		{
		    delete guidanceDialogPtr;
		    guidanceDialogPtr = NULL;
		}
		// get any guidance on the point
		GetAlarmInfo (ptName, &guidptr, &dispptr);
		if (guidptr == NULL)
		{
		    if (HaveScript (pvdata -> parentGroup -> groupParms.guidance))
		    {
			guidptr = pvdata -> parentGroup -> groupParms.guidance;
		    }
		}
		if (guidptr && *guidptr)
		{
		    vresource guidRes = vresourceGet (alvuRes, alvu_TextDialog);
		    guidanceDialogPtr = new TextDialog (guidRes);
		    guidanceDialogPtr -> OpenText (app, GUIDANCE_TEXT, 
						   guidptr, ptName);
		}
	    }
	    break;
	}

	case DISPLAYCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		SpawnPointDisplay (pvdata);
	    }
	    break;

	case DETAILCLICK:
	    ptName = ItemName (param);
	    SpawnDetail (ptName);
	    break;

	case COMMANDCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		SpawnScript (pvdata);
	    }
	    break;

	case ENABLECLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		if (EnablePoint (pvdata) != True)
		{
		    pv = vstrCloneScribed (vcharScribeSystem (ptName));
		    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
					     alvu_NoEnablePriv, pv);
		    vstrDestroy (pv);
		}
	    }
	    break;

	case DISABLECLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (groupLinkPtr, StripWhitespace (ptName)))
	    {
		if (DisablePoint (pvdata) != True)
		{
		    pv = vstrCloneScribed (vcharScribeSystem (ptName));
		    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
					     alvu_NoDisablePriv, pv);
		    vstrDestroy (pv);
		}
	    }
	    break;

	case COMMENTCLICK:
	{
	    if (commentDialogPtr != NULL)
	    {
		delete commentDialogPtr;
		commentDialogPtr = NULL;
	    }
	    ptName = ItemName (param);
	    entry = GetEntryPtr (param);
	    evtime = entry -> GetEventTime ();
	    if (Log_RecordByTime (ptName, entry -> GetEventTime (), &alarmLogRecord) != RECFOUND)
	    {
		break;
	    }
	    sprintf (dspBuf, "%-17s  %-28s  %s",
		     ItemDate (param), ItemName (param), ItemDesc (param));
	    vresource commentRes = vresourceGet (alvuRes, alvu_CommentDialog);
	    commentDialogPtr = new CommentDialog (commentRes);
	    commentDialogPtr -> OpenComment (app, ptName, dspBuf, 
					     alarmLogRecord.comment,
					     &alarmLogRecord);
	    break;
	}

	case UPCLICK:
	    if (groupLinkPtr = FindGroup (treeRoot, filter.group))
	    {
		if (groupLinkPtr -> parent)
		{
		    groupLinkPtr = groupLinkPtr -> parent;
		}
	    }
	    filter.modified = TRUE;
	    Log_SetFilter (&filter, indexFileType);
	    filePosition = 100;
	    startName[0] = 0;
	    if (strlen (filter.start))
	    {
		GetGMT_TS_STAMP (filter.start, &startTime);
	    }
	    else
	    {
		startTime.secPastEpoch = 0;
	    }
	    startTime.nsec = 0;
	    Resync ();
	    break;

	case TOPCLICK:
	    strncpy (filter.group, treeRoot -> groupName, sizeof (filter.group));
	    filter.group [sizeof (filter.group)-1] = 0;
	    filter.modified = TRUE;
	    filePosition = 100;
	    startName[0] = 0;
	    if (strlen (filter.start))
	    {
		GetGMT_TS_STAMP (filter.start, &startTime);
	    }
	    else
	    {
		startTime.secPastEpoch = 0;
	    }
	    startTime.nsec = 0;
	    Resync ();
	    break;

	case PAUSECLICK:
	    pause = !pause;
	    ShowPaused (pause);
	    if (!pause)
	    {
		Resync ();
	    }
	    else
	    {
		RefreshDisplay ();
	    }
	    break;
    }
    return True;
}


/*******************************************************************
*
*    LogFileDialog RefreshDisplay function
*******************************************************************/
static AlarmLogRecord tempRecords [ALVU_MAXITEMS];

void LogFileDialog::RefreshDisplay (void)
{
    int i;
    int srcix;
    int movix;
    int retloopcnt;
    int eof = False;
    int (AlarmDialog::*callbackFunc)(void) = &AlarmDialog::CheckCancel;

    if (inRefresh)
	return;
    inRefresh = True;

    StartCancelBox ();

    if (startName[0] == 0)
    {
	/* find out id of first record (for navigation button maintainance */
	if (Log_RecordByTime ( "! ", &startTime, &alarmLogRecord) != NORECORD)
	{
            retloopcnt = Log_RecordsByTime (alarmLogRecord.pointName,
		     &alarmLogRecord.eventtime, alarmLogRecords,
		     1, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
	    if (retloopcnt)
	    {
		strncpy (startName, alarmLogRecords[0].pointName, AL_POINTNAME_SIZE+1);
		startTime = alarmLogRecords[0].eventtime;
		strncpy (topName, alarmLogRecords[0].pointName, AL_POINTNAME_SIZE+1);
		topTime = alarmLogRecords[0].eventtime;
	    }
	    else
	    {
		/* no items for us */
		SetFileItems (0);
		SetFilePos (0);
	    }
	}
	else
	{
	    /* no items in data set */
	    SetFileItems (0);
	    SetFilePos (0);
	}
    }

    int loopcnt = ALVU_MAXITEMS;
    if (loopcnt > Size())
	loopcnt = Size();

    if (filePosition == 0)
    {
	/* get screen full starting at first record */
	strncpy (topName, startName, AL_POINTNAME_SIZE+1);
	topTime = startTime;
        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmLogRecords,
		     loopcnt, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled();
	    inRefresh = False;
	    return;
	}
    }
    else if (filePosition == 25)
    {
	/* find records backwards from top record */
        retloopcnt = Log_RecordsByTime (topName, &topTime, tempRecords,
		     -loopcnt, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled();
	    inRefresh = False;
	    return;
	}
	if (retloopcnt)
	{
	    srcix = 0;
	    /* skip any records before startTime */
	    while (((startTime.secPastEpoch > alarmLogRecords[srcix].eventtime.secPastEpoch)
		 || ((startTime.secPastEpoch == alarmLogRecords[srcix].eventtime.secPastEpoch)
			 && (startTime.nsec > alarmLogRecords[srcix].eventtime.nsec)))
	      || (memcmp (&startTime, &alarmLogRecords[srcix].eventtime, sizeof(startTime)) == 0
		 && strncmp (startName, alarmLogRecords[srcix].pointName, AL_POINTNAME_SIZE+1) >= 0))
	    { /* if above start, adjust back to start */
		srcix++;
		retloopcnt--;
	    }
	    /* make room for the new records, and copy them in */
	    for (movix = loopcnt - 1; movix > retloopcnt-srcix; movix--)
		alarmLogRecords [ movix ] = alarmLogRecords [ movix - (retloopcnt-srcix) ];
	    for ( movix = 0; srcix < retloopcnt; srcix++,movix++)
		alarmLogRecords [ movix ] = tempRecords [ srcix ];
	    retloopcnt = loopcnt;
	}
	else /* no records behind us, leave it alone */
	{
	    StopCancelBox ();
	    filePosition = 0;
	    SetFilePos (filePosition);
	    inRefresh = False;
	    return;
	}
    }
    else if (filePosition == 49)
    {
	/* find  1 record backwards from top record */
        retloopcnt = Log_RecordsByTime (topName, &topTime, tempRecords,
		     -2, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled();
	    inRefresh = False;
	    return;
	}
	if (retloopcnt == 2)
	{
	    if (((startTime.secPastEpoch > alarmLogRecords[0].eventtime.secPastEpoch)
		 || ((startTime.secPastEpoch == alarmLogRecords[0].eventtime.secPastEpoch)
			 && (startTime.nsec > alarmLogRecords[0].eventtime.nsec)))
	      || (memcmp (&startTime, &alarmLogRecords[0].eventtime, sizeof(startTime)) == 0
		 && strncmp (startName, alarmLogRecords[0].pointName, AL_POINTNAME_SIZE+1) >= 0))
	    { /* if above start, leave it alone */
		inRefresh = False;
		return;
	    }
	    else
	    {
		/* shift records down, copy in new record */
		for (movix = loopcnt - 1; movix > 0; movix--)
		    alarmLogRecords [ movix ] = alarmLogRecords [ movix - 1 ];
		alarmLogRecords [ 0 ] = tempRecords [ 0 ];
		retloopcnt = loopcnt;
	    }
	}
	else /* no records behind us, leave it alone */
	{
	    StopCancelBox ();
	    filePosition = 0;
	    SetFilePos (filePosition);
	    inRefresh = False;
	    return;
	}
    }
    else if (filePosition == 50)  /* refresh with current displaySize */
    {
        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmLogRecords,
		     	loopcnt, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled();
	    inRefresh = False;
	    return;
	}
        if (retloopcnt < loopcnt)
	{ /* just position at end */
	    /* find records backwards from end of file */
            retloopcnt = Log_RecordsByTime (0, 0, alarmLogRecords,
		     -loopcnt, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
	    if (CheckCancelled ())
	    {
		ShowCancelled();
		inRefresh = False;
		return;
	    }
	    eof = True;
	    if (retloopcnt)
	    {
		if (((startTime.secPastEpoch > alarmLogRecords[0].eventtime.secPastEpoch)
			 || ((startTime.secPastEpoch == alarmLogRecords[0].eventtime.secPastEpoch)
				 && (startTime.nsec > alarmLogRecords[0].eventtime.nsec)))
		  || (memcmp (&startTime, &alarmLogRecords[0].eventtime, sizeof(startTime)) == 0
		    && strncmp (startName, alarmLogRecords[0].pointName, AL_POINTNAME_SIZE+1) >= 0))
		{ /* if above start, reset to start */
		    strncpy (topName, startName, AL_POINTNAME_SIZE+1);
		    topTime = startTime;
	            retloopcnt = Log_RecordsByTime (topName, &topTime, alarmLogRecords,
			     loopcnt, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
		    if (CheckCancelled ())
		    {
			ShowCancelled();
			inRefresh = False;
			return;
		    }
		}
	    }
	}
    }
    else if (filePosition == 51) /* find records forward from second record */
    {
	/* find  1 record forwards from last record */
        retloopcnt = Log_RecordsByTime (lastName, &lastTime, tempRecords,
		     2, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled();
	    inRefresh = False;
	    return;
	}
	if (retloopcnt == 2)
	{
	    /* shift records up, copy in new record */
	    for (movix = 0; movix < loopcnt-1; movix++)
		alarmLogRecords [ movix ] = alarmLogRecords [ movix + 1 ];
	    alarmLogRecords [ loopcnt - 1 ] = tempRecords [ 1 ];
	    retloopcnt = loopcnt;
	}
	else /* no records in front of us, leave it alone */
	{
	    StopCancelBox ();
	    filePosition = 90;
	    SetFilePos (filePosition);
	    inRefresh = False;
	    return;
	}
    }
    else if (filePosition == 75) /* find records forward from last record */
    {
	strncpy (topName, lastName, AL_POINTNAME_SIZE+1);
	topTime = lastTime;
        retloopcnt = Log_RecordsByTime (lastName, &lastTime, tempRecords,
		     loopcnt, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled();
	    inRefresh = False;
	    return;
	}
	if (retloopcnt != loopcnt)
	    eof = True;
	/* shift records up, copy in new records */
	for (movix = 0; movix < loopcnt-retloopcnt; movix++)
	    alarmLogRecords [ movix ] = alarmLogRecords [ movix + retloopcnt - 1 ];
	for (srcix = 0; srcix < retloopcnt; movix++, srcix++)
	    alarmLogRecords [ movix ] = tempRecords [ srcix ];
	retloopcnt = loopcnt;
    }
    else if (filePosition >= 90)
    {
	/* find records backwards from end of file */
        retloopcnt = Log_RecordsByTime (0, 0, alarmLogRecords,
		     -loopcnt, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled();
	    inRefresh = False;
	    return;
	}
	eof = True;
	if (retloopcnt)
	{
	    if (((startTime.secPastEpoch > alarmLogRecords[0].eventtime.secPastEpoch)
		 || ((startTime.secPastEpoch == alarmLogRecords[0].eventtime.secPastEpoch)
		    && (startTime.nsec > alarmLogRecords[0].eventtime.nsec)))
	      || (memcmp (&startTime, &alarmLogRecords[0].eventtime, sizeof(startTime)) == 0
		 && strncmp (startName, alarmLogRecords[0].pointName, AL_POINTNAME_SIZE+1) >= 0))
	    { /* if above start, reset to start */
		strncpy (topName, startName, AL_POINTNAME_SIZE+1);
		topTime = startTime;
	        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmLogRecords,
			     loopcnt, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
		if (CheckCancelled ())
		{
		    ShowCancelled();
		    inRefresh = False;
		    return;
		}
	    }
	}
    }
    else /* find records forward from top record */
    {
        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmLogRecords,
		     loopcnt, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled();
	    inRefresh = False;
	    return;
	}
    }
    if (retloopcnt)
    {
	strncpy (topName, alarmLogRecords[0].pointName, AL_POINTNAME_SIZE+1);
	topTime = alarmLogRecords[0].eventtime;
	strncpy (lastName, alarmLogRecords[retloopcnt-1].pointName, AL_POINTNAME_SIZE+1);
	lastTime = alarmLogRecords[retloopcnt-1].eventtime;
    }
    StopCancelBox ();
    if (CheckCancelled ())
    {
	ShowCancelled();
	inRefresh = False;
	return;
    }

    if (indexFileType != ALRM_INDEX_FILE)
    {
	SetDialogTitle ();
    }
    elementCount -> SetTextScribed (vcharScribeSystem (""));
    ClearLines ();
    SetFileItems (100);
    for (i = 0; i < retloopcnt; i++)
    {
	if(alarmLogRecords[i].comment[0] != 0)  {
	    sprintf (timeBuf, "%s %s *", alarmLogRecords[i].date, alarmLogRecords[i].time);
	}
	else  {
	    sprintf (timeBuf, "%s %s  ", alarmLogRecords[i].date, alarmLogRecords[i].time);
	}
	if (strcmp (alarmLogRecords[i].status, "ACK") == 0)
	{
	    sprintf (txtBuf, "%s",
			 AlarmFormat (&displayack, NULL, &alarmLogRecords[i]));
	}
	else
	{
	    sprintf (txtBuf, "%s",
			 AlarmFormat (&displayalarm, NULL, &alarmLogRecords[i]));
	}
	AddLine ( -1, "", alarmLogRecords[i].pointName, txtBuf, &alarmLogRecords[i].eventtime );
    }
    if (dataList != NULL)
    {
	alarmListView -> SetList (dataList);
	vlistSelection *selected = alarmListView -> GetSelection ();
	dataList -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	dataList -> SetAtLeastOneSelection (selected, vFALSE);
    }
    alarmListView -> Inval (vwindowINVAL_OPAQUE | vwindowINVAL_UPDATE);
    if ((strcmp (startName, topName) == 0) && (memcmp (&startTime, &topTime, sizeof(startTime)) == 0))
    {
	filePosition = 0;
	if (retloopcnt < loopcnt)
	{
	    SetFileItems (retloopcnt);
	}
    }
    else if (eof || retloopcnt < loopcnt-1)
    {
	filePosition = 90;
    }
    else
    {
	filePosition = 50;
    }
    SetFilePos (filePosition);
    inRefresh = False;
}


/*******************************************************************
*
*    LogFileDialog Resync function
*******************************************************************/

void LogFileDialog::Resync (void)
{
    if (pause)
    {
	RefreshDisplay ();
    }
    else
    {
	SetPosition (100);
    }
}


/*******************************************************************
*
*    LogFileDialog Set Indication Msg's
*
*******************************************************************/

void LogFileDialog::SetIndication (void)
{
    vstr *msg = vresourceGetString (app -> messagesResource,
					    alvu_AlarmUpdateMsg);
    elementCount -> SetTextScribed (vcharScribeSystem (msg));
    vstrDestroy (msg);
    fileButtonContainer -> linedownButton -> Enable ();
    fileButtonContainer -> pagedownButton -> Enable ();
    fileButtonContainer -> lastButton -> Enable ();
}

void LogFileDialog::ShowCancelled (void)
{
    vstr *msg = vresourceGetString (app -> messagesResource,
					    alvu_RefreshCancelledMsg);
    elementCount -> SetTextScribed (vcharScribeSystem (msg));
    vstrDestroy (msg);
    StopCancelBox ();
    fileButtonContainer -> linedownButton -> Enable ();
    fileButtonContainer -> pagedownButton -> Enable ();
    fileButtonContainer -> lastButton -> Enable ();
}


/*******************************************************************
*
*    LogFileDialog Set Dialog Title function
*
*******************************************************************/

void LogFileDialog::SetDialogTitle (void)
{
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    if (indexFileType == ALRM_INDEX_FILE)
    {
	title = vresourceGetString (app -> titlesResource, alvu_RecentAlarmTitle);
    }
    else
    {
	vstr *format = vresourceGetString (app -> titlesResource, alvu_LogFileTitle);
	vstr *group = vstrCloneScribed (vcharScribeSystem (filter.group));
	vstr *mod = vstrCloneScribed (vcharScribeLiteral (filter.modified ? "*" : ""));
	vstr *name = vstrCloneScribed (vcharScribeSystem (filter.name));
	title = vstrCloneScribed (vcharScribeFormat (format, displayTitle,
						     group, mod, name));
	vstrDestroy (format);
	vstrDestroy (group);
	vstrDestroy (mod);
	vstrDestroy (name);
    }
    SetTitle (title);
}

/*******************************************************************
*
*    LogFileDialog handle blinking
*
*******************************************************************/

void LogFileDialog::UpdateBlink (void)
{
    AlarmDialog::UpdateBlink ();
}


/*******************************************************************
*
*    LogFileDialog Unix Print function
*
*******************************************************************/

void LogFileDialog::UnixPrint (int whatToPrint)
{
    FILE    *printFileDesc;
    char    printFileName [30];
    int	    numLines, startIndex;
    int	    i;

    StartCancelBox ();
    if (whatToPrint == PRINT_THIS_WINDOW)
    {
        numLines = Log_RecordsByTime (topName, &this -> topTime, alarmLogRecords,
		     Size (), indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
    }
    else    // print entire list
    {
         numLines = Log_RecordsByTime (startName, &this -> startTime, alarmLogRecords,
		     ALVU_MAXITEMS, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
    }
    StopCancelBox ();
    if (CheckCancelled ())
    {
	ShowCancelled ();
	return;
    }

    strcpy (printFileName, PRINT_FILE_TEMP);
    if ( !mktemp (printFileName) || !(printFileDesc = fopen (printFileName, "w")))
    {
	vstr *prtFileTmp = vstrCloneScribed (vcharScribeLiteral (PRINT_FILE_TEMP));
	alvuMessageDialogCenter (this, ALVU_INFO_NOTICE, alvu_TempPrtFileErr, prtFileTmp);
	vstrDestroy (prtFileTmp);
    }
    else
    {
	// 0 at first, 1 rest of time
	startIndex = 0;	
	while (numLines > startIndex)
	{
	    for (i = startIndex; i < numLines; i++)
	    {
		if (strcmp (alarmLogRecords[i].status, "ACK") == 0)
		{
		    fprintf (printFileDesc, "%s\n",
			     AlarmFormat (&printack, NULL, &alarmLogRecords[i]));
		}
		else
		{
		    fprintf (printFileDesc, "%s\n",
			     AlarmFormat (&printalarm, NULL, &alarmLogRecords[i].date));
		}
	    }
	    if (whatToPrint == PRINT_ENTIRE_LIST)
	    {
		StartCancelBox ();
		numLines = Log_RecordsByTime (alarmLogRecords[numLines-1].pointName,
				&alarmLogRecords[numLines-1].eventtime, alarmLogRecords,
				ALVU_MAXITEMS, indexFileType, (int (__cdecl *)(void))CheckCancelGW, this);
		StopCancelBox ();
		if (CheckCancelled ())
		{
		    ShowCancelled ();
		    break;
		}
		startIndex = 1; // skip last record of previous group
	    }
	    else
	    {
		break;
	    }
	}
	fclose (printFileDesc);
	vcharExportBounded (vcharScribeSystemExport (printTitle),
			    caption,
			    sizeof (caption));
	caption [sizeof (caption) - 1] = 0;
	if (strlen (linePrinter))
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"%s\" %s |lpr -P\"%s\"", 
			      caption, printFileName, linePrinter);
#else
	    sprintf (txtBuf, "lpr -p -r -P\"%s\" -T\"Alarm Summary\" %s", 
			      linePrinter, caption, printFileName);
#endif SOLARIS
	}
	else
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"%s\" %s |lpr", caption, printFileName);
#else
	    sprintf (txtBuf, "lpr -p -r -T\"%s\" %s", caption, printFileName);
#endif SOLARIS
	}
	system (txtBuf);
#ifdef SOLARIS
	sprintf (txtBuf, "rm %s", printFileName);
	system (txtBuf);
#endif SOLARIS
    }
}


/*******************************************************************
*
*    LogFileDialog Windows Print function
*
*******************************************************************/

void LogFileDialog::WindowsPrint (int whatToPrint)
{
    if (app -> printer == NULL)
    {
	alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_NoPrinterMsg);
	return;
    }
    return;
}


/*******************************************************************
*
*    Alarm History Dialog Class
*	subclassed from LogFileDialog
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AlarmHistoryDialog, LogFileDialog, "AlarmHistoryDialog");

/*******************************************************************
*
*    Alarm History Dialog Class, Open
*
*******************************************************************/

void AlarmHistoryDialog::AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt)
{
    app = alvuApp;
    displayTitle = vresourceGetString (alvuApp -> titlesResource, alvu_AlarmHistoryTitle);
    printTitle = vresourceGetString (alvuApp -> titlesResource, alvu_AlarmHistoryPrintTitle);
    indexFileType = ALRMLOG_INDEX_FILE;

    BuildMenus ();

    pauseButton -> Disable ();
    pauseButton -> Hide ();

    if (strlen (filt -> start))
    {
	GetGMT_TS_STAMP (filt -> start, &startTime);
    }
    else
    {
	startTime.secPastEpoch = 0;
    }
    startTime.nsec = 0;

    LogFileDialog::AlarmDialogOpen (alvuApp, filt);
}


/*******************************************************************
*
*    Alarm History Dialog Class, Build Menu function
*
*******************************************************************/

void AlarmHistoryDialog::BuildMenus (void)
{
    vresource titles = app -> titlesResource;
    vmenu	*newMenu;
    vmenuItem	*menuItem;
    vstr	*label;

    // the print menu

    label = vresourceGetString (titles, alvu_PrintMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ThisWindow);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlHistThisWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_EntireHistory);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlHistEntireWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the filter menu

    label = vresourceGetString (titles, alvu_FilterMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ModifyFilter);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlHistModFilter);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // load up menuNames with resources
    menuNames [ALRMHIST_THISWIN_ITEM]   = alvu_AlHistThisWindow;
    menuNames [ALRMHIST_ENTIRESUM_ITEM] = alvu_AlHistEntireWindow;
    menuNames [ALRMHIST_MODFILT_ITEM]   = alvu_AlHistModFilter;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;
    // populate our dictionary with menu commands
    for (i = ALRMHIST_THISWIN_ITEM; i < ALRMHIST_ITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	alarmHistoryCommands *command = new alarmHistoryCommands;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (menuNames[i], command);
    }

    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}



vkindTYPED_FULL_DEFINITION(alarmHistoryCommands, vcommandSelector, "alarmHistoryCommands");
/*******************************************************************
*
*    menu selector commands for the dynamic group dialog
*
*******************************************************************/

int alarmHistoryCommands::Issue (vdict *dict)
{
    int			command;
    AlarmHistoryDialog	*alarmHistoryDialog;

    alarmHistoryDialog = (AlarmHistoryDialog *) dict -> GetElem (vname_Dialog);
    alvuApplication *app = alarmHistoryDialog -> GetApp ();
    vresource alvuRes = app -> alvuResource;

    command = GetSelector ();

    switch (command)
    {
	case ALRMHIST_THISWIN_ITEM:
	case ALRMHIST_ENTIRESUM_ITEM:
	{
	    AldlgListView *view = alarmHistoryDialog -> alarmListView;
	    vlist *list = view -> GetList ();
	    int rowCount = 0;
	    if (list != NULL)
	    {
		rowCount = list -> GetRowCount ();
	    }
	    if ((list == NULL) || (rowCount == 0))
	    {
		alvuMessageDialogCenter (alarmHistoryDialog, ALVU_INFO_NOTICE,
					 alvu_NothingToPrint);
		break;
	    }
#ifdef WIN32
	    if (command == ALRMHIST_THISWIN_ITEM)
	    {
		alarmHistoryDialog -> WindowsPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmHistoryDialog -> WindowsPrint (PRINT_ENTIRE_LIST);
	    }
#else
	    if (command == ALRMHIST_THISWIN_ITEM)
	    {
		alarmHistoryDialog -> UnixPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmHistoryDialog -> UnixPrint (PRINT_ENTIRE_LIST);
	    }
#endif WIN32
	    break;
	}

	case ALRMHIST_MODFILT_ITEM:
	{
	    vresource filtRes = vresourceGet (alvuRes, alvu_FilterDialog);
	    FilterDialog *filterDialog = new FilterDialog (filtRes);
	    filterDialog -> OpenFilter (app, NULL, &alarmHistoryDialog -> filter, 
					MODIFY_FILTER);
	    filterDialog -> NewFilter (&alarmHistoryDialog -> filter);
	    delete filterDialog;
	    Log_SetFilter (&alarmHistoryDialog -> filter, alarmHistoryDialog -> indexFileType);
	    alarmHistoryDialog -> filePosition = 100;
	    alarmHistoryDialog -> startName[0] = 0;
	    if (strlen (alarmHistoryDialog -> filter.start))
	    {
		GetGMT_TS_STAMP (alarmHistoryDialog -> filter.start, 
				 &alarmHistoryDialog -> startTime);
	    }
	    else
	    {
		alarmHistoryDialog -> startTime.secPastEpoch = 0;
	    }
	    alarmHistoryDialog -> startTime.nsec = 0;
	    alarmHistoryDialog -> Resync ();
	    break;
	}

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}


/*******************************************************************
*
*    Alarm Event Dialog Class
*	subclassed from LogFileDialog
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AlarmEventDialog, LogFileDialog, "AlarmEventDialog");

/*******************************************************************
*
*    Alarm History Dialog Class, Open
*
*******************************************************************/

void AlarmEventDialog::AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt)
{
    app = alvuApp;
    displayTitle = vresourceGetString (alvuApp -> titlesResource, alvu_AlarmEventTitle);
    printTitle = vresourceGetString (alvuApp -> titlesResource, alvu_AlarmEventPrintTitle);
    indexFileType = ALRMEVT_INDEX_FILE;

    BuildMenus ();

    pauseButton -> Disable ();
    pauseButton -> Hide ();

    if (strlen (filt -> start))
    {
	GetGMT_TS_STAMP (filt -> start, &startTime);
    }
    else
    {
	startTime.secPastEpoch = 0;
    }
    startTime.nsec = 0;

    LogFileDialog::AlarmDialogOpen (alvuApp, filt);
}


/*******************************************************************
*
*    Alarm Event Dialog Class, Build Menu function
*
*******************************************************************/

void AlarmEventDialog::BuildMenus (void)
{
    vresource titles = app -> titlesResource;
    vmenu	*newMenu;
    vmenuItem	*menuItem;
    vstr	*label;

    // the print menu

    label = vresourceGetString (titles, alvu_PrintMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ThisWindow);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlEvntThisWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_EntireHistory);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlEvntEntireWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the filter menu

    label = vresourceGetString (titles, alvu_FilterMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ModifyFilter);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlEvntModFilter);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // load up menuNames with resources
    menuNames [ALRMEVNT_THISWIN_ITEM]   = alvu_AlEvntThisWindow;
    menuNames [ALRMEVNT_ENTIRESUM_ITEM] = alvu_AlEvntEntireWindow;
    menuNames [ALRMEVNT_MODFILT_ITEM]   = alvu_AlEvntModFilter;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;
    // populate our dictionary with menu commands
    for (i = ALRMEVNT_THISWIN_ITEM; i < ALRMEVNT_ITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	alarmEventCommands *command = new alarmEventCommands;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (menuNames[i], command);
    }

    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}


vkindTYPED_FULL_DEFINITION(alarmEventCommands, vcommandSelector, "alarmEventCommands");
/*******************************************************************
*
*    menu selector commands for the dynamic group dialog
*
*******************************************************************/

int alarmEventCommands::Issue (vdict *dict)
{
    int			command;
    AlarmEventDialog	*alarmEventDialog;

    alarmEventDialog = (AlarmEventDialog *) dict -> GetElem (vname_Dialog);
    alvuApplication *app = alarmEventDialog -> GetApp ();
    vresource alvuRes = app -> alvuResource;

    command = GetSelector ();

    switch (command)
    {
	case ALRMEVNT_THISWIN_ITEM:
	case ALRMEVNT_ENTIRESUM_ITEM:
	{
	    AldlgListView *view = alarmEventDialog -> alarmListView;
	    vlist *list = view -> GetList ();
	    int rowCount = 0;
	    if (list != NULL)
	    {
		rowCount = list -> GetRowCount ();
	    }
	    if ((list == NULL) || (rowCount == 0))
	    {
		alvuMessageDialogCenter (alarmEventDialog, ALVU_INFO_NOTICE,
					 alvu_NothingToPrint);
		break;
	    }
#ifdef WIN32
	    if (command == ALRMEVNT_THISWIN_ITEM)
	    {
		alarmEventDialog -> WindowsPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmEventDialog -> WindowsPrint (PRINT_ENTIRE_LIST);
	    }
#else
	    if (command == ALRMEVNT_THISWIN_ITEM)
	    {
		alarmEventDialog -> UnixPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmEventDialog -> UnixPrint (PRINT_ENTIRE_LIST);
	    }
#endif WIN32
	    break;
	}

	case ALRMEVNT_MODFILT_ITEM:
	{
	    vresource filtRes = vresourceGet (alvuRes, alvu_FilterDialog);
	    FilterDialog *filterDialog = new FilterDialog (filtRes);
	    filterDialog -> OpenFilter (app, NULL, &alarmEventDialog -> filter, 
					MODIFY_FILTER);
	    filterDialog -> NewFilter (&alarmEventDialog -> filter);
	    delete filterDialog;
	    Log_SetFilter (&alarmEventDialog -> filter, alarmEventDialog -> indexFileType);
	    alarmEventDialog -> filePosition = 100;
	    alarmEventDialog -> startName[0] = 0;
	    if (strlen (alarmEventDialog -> filter.start))
	    {
		GetGMT_TS_STAMP (alarmEventDialog -> filter.start, 
				 &alarmEventDialog -> startTime);
	    }
	    else
	    {
		alarmEventDialog -> startTime.secPastEpoch = 0;
	    }
	    alarmEventDialog -> startTime.nsec = 0;
	    alarmEventDialog -> Resync ();
	    break;
	}

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}


/*******************************************************************
*
*    Alarm Window Dialog Class
*	subclassed from LogFileDialog
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AlarmWindowDialog, LogFileDialog, "AlarmWindowDialog");

/*******************************************************************
*
*    Alarm Window Dialog Class, Open
*
*******************************************************************/

void AlarmWindowDialog::AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt)
{
    app = alvuApp;
    displayTitle = vresourceGetString (alvuApp -> titlesResource, alvu_AlarmWindowTitle);
    printTitle = vresourceGetString (alvuApp -> titlesResource, alvu_AlarmWindowPrintTitle);
    indexFileType = ALRM_INDEX_FILE;

    BuildMenus ();

    upButton -> Disable ();
    topButton -> Disable ();

    startTime.secPastEpoch = time(0) - TS_EPOCH_SEC_PAST_1970;
    startTime.nsec = 0;

    LogFileDialog::AlarmDialogOpen (alvuApp, filt);
}


/*******************************************************************
*
*    Alarm Window Dialog Class, Build Menu function
*
*******************************************************************/

void AlarmWindowDialog::BuildMenus (void)
{
    vresource titles = app -> titlesResource;
    vmenu	*newMenu;
    vmenuItem	*menuItem;
    vstr	*label;

    // the print menu

    label = vresourceGetString (titles, alvu_PrintMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ThisWindow);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlWndwThisWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_EntireHistory);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlWndwEntireWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the clear menu

    label = vresourceGetString (titles, alvu_ClearMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_Clear);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlWndwClear);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // load up menuNames with resources
    menuNames [ALRMWNDW_THISWIN_ITEM]   = alvu_AlWndwThisWindow;
    menuNames [ALRMWNDW_ENTIRESUM_ITEM] = alvu_AlWndwEntireWindow;
    menuNames [ALRMWNDW_CLEAR_ITEM]     = alvu_AlWndwClear;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;
    // populate our dictionary with menu commands
    for (i = ALRMWNDW_THISWIN_ITEM; i < ALRMWNDW_ITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	alarmWindowCommands *command = new alarmWindowCommands;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (menuNames[i], command);
    }

    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}


vkindTYPED_FULL_DEFINITION(alarmWindowCommands, vcommandSelector, "alarmWindowCommands");
/*******************************************************************
*
*    menu selector commands for the dynamic group dialog
*
*******************************************************************/

int alarmWindowCommands::Issue (vdict *dict)
{
    int			command;
    AlarmWindowDialog	*alarmWindowDialog;

    alarmWindowDialog = (AlarmWindowDialog *) dict -> GetElem (vname_Dialog);
    alvuApplication *app = alarmWindowDialog -> GetApp ();
    vresource alvuRes = app -> alvuResource;

    command = GetSelector ();

    switch (command)
    {
	case ALRMWNDW_THISWIN_ITEM:
	case ALRMWNDW_ENTIRESUM_ITEM:
	{
	    AldlgListView *view = alarmWindowDialog -> alarmListView;
	    vlist *list = view -> GetList ();
	    int rowCount = 0;
	    if (list != NULL)
	    {
		rowCount = list -> GetRowCount ();
	    }
	    if ((list == NULL) || (rowCount == 0))
	    {
		alvuMessageDialogCenter (alarmWindowDialog, ALVU_INFO_NOTICE,
					 alvu_NothingToPrint);
		break;
	    }
#ifdef WIN32
	    if (command == ALRMWNDW_THISWIN_ITEM)
	    {
		alarmWindowDialog -> WindowsPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmWindowDialog -> WindowsPrint (PRINT_ENTIRE_LIST);
	    }
#else
	    if (command == ALRMWNDW_THISWIN_ITEM)
	    {
		alarmWindowDialog -> UnixPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmWindowDialog -> UnixPrint (PRINT_ENTIRE_LIST);
	    }
#endif WIN32
	    break;
	}

	case ALRMWNDW_CLEAR_ITEM:
	{
	    alarmWindowDialog -> ClearLines ();
	    alarmWindowDialog -> startName[0] = 0;
	    alarmWindowDialog -> startTime.secPastEpoch = time (0) - TS_EPOCH_SEC_PAST_1970;
	    alarmWindowDialog -> startTime.nsec = 0;
	    alarmWindowDialog -> SetPosition (0);
	    alarmWindowDialog -> Resync ();
	}

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}


/*******************************************************************
*
*    Operator Tracking Dialog Class
*	subclassed from AlarmDialog
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AlarmTrackDialog, AlarmDialog, "AlarmTrackDialog");

/*******************************************************************
*
*    AlarmTrackDialog, LoadInit ()
*
*******************************************************************/

void AlarmTrackDialog::LoadInit (vresource resource)
{
    fileButtonContainer = FileButtonContainer::RequestSubLoad (this,
					    alvu_BlankContainer);
    AlarmDialog::LoadInit (resource);

    fileButtonContainer -> firstButton    -> StoreAlarmDialog (this);
    fileButtonContainer -> pageupButton   -> StoreAlarmDialog (this);
    fileButtonContainer -> lineupButton   -> StoreAlarmDialog (this);
    fileButtonContainer -> linedownButton -> StoreAlarmDialog (this);
    fileButtonContainer -> pagedownButton -> StoreAlarmDialog (this);
    fileButtonContainer -> lastButton     -> StoreAlarmDialog (this);
}

/*******************************************************************
*
*    DynamicGroupDialog, destroy function
*
*******************************************************************/

void AlarmTrackDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    AlarmDialog::Destroy ();
}


/*******************************************************************
*
*    DynamicGroupDialog Open function
*
*******************************************************************/

void AlarmTrackDialog::AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt)
{
    app = alvuApp;
    filter = *filt;
    PopupMenu = NULL;
    reqPosition = -1;

    BuildMenus ();
    SetDialogTitle ();
    TrkInit ();			    // alarmfil.c
    TrkSetFilter (&this -> filter); // alarmfil.c

    ackable = False;
    hasdate = True;
    hasdesc = True;
    monitored = False;
    fileBased = True;
    pauseButton -> Disable ();

    // turn off the scroll bar for the listview, we are using
    // buttons to control the listview
    alarmListView -> SetVertBar (vFALSE);

    AlarmDialog::AlarmDialogOpen (alvuApp, filt);

    displaySize = alarmListView -> GetListNumElements ();

    SetPosition (100);
}


/*******************************************************************
*
*    AlarmTrackDialog Build menu function 
*
*******************************************************************/

void AlarmTrackDialog::BuildMenus (void)
{
    vresource titles = app -> titlesResource;
    vmenu	*newMenu;
    vmenuItem	*menuItem;
    vstr	*label;

    // the print menu

    label = vresourceGetString (titles, alvu_PrintMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ThisWindow);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlTrkThisWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_EntireHistory);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlTrkEntireWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the filter menu

    label = vresourceGetString (titles, alvu_FilterMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ModifyFilter);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlTrkModFilter);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);


    // load up menuNames with resources
    menuNames [OPTRK_THISWIN_ITEM]     = alvu_AlTrkThisWindow;
    menuNames [OPTRK_ENTIRESUM_ITEM]   = alvu_AlTrkEntireWindow;
    menuNames [OPTRK_MODFILT_ITEM]     = alvu_AlTrkModFilter;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;
    // populate our dictionary with menu commands
    for (i = OPTRK_THISWIN_ITEM; i < OPTRK_ITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	alarmTrackCommands *command = new alarmTrackCommands;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (menuNames[i], command);
    }

    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}


/*******************************************************************
*
*    AlarmTrackDialog, Ctrl Function ()
*	param if >= 0 is row number in list
*
*******************************************************************/

int AlarmTrackDialog::CtrlFunc ( int cmd, int param)
{
    char	*ptName;
    TreeNode	*groupLinkPtr;

    // get resource to alarmvu
    vresource alvuRes = app -> alvuResource;

    switch (cmd)
    {
	case MOVECLICK:
	    reqPosition = param;
	    RefreshDisplay ();
	    break;

	case EXPANDCLICK:
	    if (expandDialogPtr != NULL)
	    {
		delete expandDialogPtr;
		expandDialogPtr = NULL;
	    }
	    ptName = ItemName (param);
	    if (TrkSeek (param + filePosition + 1) == RECFOUND)
	    {
		// display expansion dialog
		vresource expRes = vresourceGet (alvuRes, alvu_TextDialog);
		expandDialogPtr = new TextDialog (expRes);
		expandDialogPtr -> OpenText (app, 
					     EXPAND_TEXT, 
					     AlarmFormat (&expandtrack, NULL, NULL),
					     NULL);
	    }
	    break;

	case NAMECLICK:
	    return False;

	case UPCLICK:
	    if (groupLinkPtr = FindGroup (treeRoot, filter.group))
	    {
		if (groupLinkPtr -> parent)
		{
		    strncpy (filter.group, groupLinkPtr -> parent -> groupName, 
			     sizeof (filter.group));
		    filter.group [sizeof (filter.group)-1] = 0;
		}
	    }
	    filter.modified = TRUE;
	    SetDialogTitle ();
	    TrkSetFilter (&this -> filter);
	    SetPosition (100);
	    break;

	case TOPCLICK:
	    strncpy (filter.group, treeRoot -> groupName, sizeof (filter.group));
	    filter.group [sizeof (filter.group)-1] = 0;
	    filter.modified = TRUE;
	    SetDialogTitle ();
	    TrkSetFilter (&this -> filter);
	    SetPosition (100);
	    break;
    }
    return True;
}


/*******************************************************************
*
*    AlarmTrackDialog Refresh function
*
*******************************************************************/

void AlarmTrackDialog::RefreshDisplay (void)
{
    int		i;

    ClearLines ();
    if (reqPosition == 0)
	filePosition = 0;
    else if (reqPosition == 25)
    {
	if (filePosition > Size())
	    filePosition -= Size();
	else
	    filePosition = 0;
    }
    else if (reqPosition == 49)
    {
	if (filePosition != 0)
	    filePosition--;
    }
    else if (reqPosition == 51)
    {
	filePosition++;
    }
    else if (reqPosition == 75)
    {
	filePosition += Size();
    }
    else if (reqPosition >= 90)
    {
	filePosition = (int)(TrkNumEntries()) - Size();
    }
    reqPosition = -1;
    if (filePosition > ((int)(TrkNumEntries()) - Size()))
    {
	filePosition = (int)(TrkNumEntries()) - Size();
    }

    if (filePosition < 0)
    {
	filePosition = 0;
    }

    if (filePosition == 0)
    {
	SetFilePos (0);
    }
    else if (filePosition >= ((int)(TrkNumEntries()) - Size()))
    {
	SetFilePos (90);
    }
    else
    {
	SetFilePos (50);
    }

    for (i = filePosition; i < (int)(TrkNumEntries ()) && i < (filePosition + Size()); i++)
    {
	if (TrkSeek (i + 1) == RECFOUND)
	{
	    sprintf (timeBuf, "%s %s", TrkDate (), TrkTime ());
	    AddLine ( -1, timeBuf, TrkPointname (), AlarmFormat (&displaytrack, NULL, NULL));
	}
    }
    if (dataList != NULL)
    {
	alarmListView -> SetList (dataList);
	vlistSelection *selected = alarmListView -> GetSelection ();
	dataList -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	dataList -> SetAtLeastOneSelection (selected, vFALSE);
    }
    if ((filePosition == 0) && (i == (int)(TrkNumEntries ())))
    {
	SetFileItems (i);
    }
    else
    {
	SetFileItems (100);
    }
}


/*******************************************************************
*
*    AlarmTrackDialog Resync function
*
*******************************************************************/

void AlarmTrackDialog::Resync (void)
{
    RefreshDisplay ();
}


/*******************************************************************
*
*    AlarmTrackDialog set title function to be written by subclasses
*
*******************************************************************/

void AlarmTrackDialog::SetDialogTitle (void)
{
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    vstr *format = vresourceGetString (app -> titlesResource, alvu_AlarmTrackTitle);
    vstr *op = vstrCloneScribed (vcharScribeSystem (filter.op));
    vstr *mod = vstrCloneScribed (vcharScribeLiteral (filter.modified ? "*" : ""));
    vstr *name = vstrCloneScribed (vcharScribeSystem (filter.name));
    title = vstrCloneScribed (vcharScribeFormat (format, op, mod, name));
    SetTitle (title);
    vstrDestroy (format);
    vstrDestroy (op);
    vstrDestroy (mod);
    vstrDestroy (name);
}


/*******************************************************************
*
*    AlarmTrackDialog Unix Print function
*
*******************************************************************/

void AlarmTrackDialog::UnixPrint (int whatToPrint)
{
    FILE    *printFileDesc;
    char    printFileName [30];
    int	    i;

    strcpy (printFileName, PRINT_FILE_TEMP);
    if ( !mktemp (printFileName) || !(printFileDesc = fopen (printFileName, "w")))
    {
	vstr *prtFileTmp = vstrCloneScribed (vcharScribeLiteral (PRINT_FILE_TEMP));
	alvuMessageDialogCenter (this, ALVU_INFO_NOTICE, alvu_TempPrtFileErr, prtFileTmp);
	vstrDestroy (prtFileTmp);
    }
    else
    {
	if (whatToPrint == PRINT_THIS_WINDOW)
	{
	    for (i = filePosition; i < (int)(TrkNumEntries ()) && i < (filePosition + Size ()); i++)
	    {
		if (TrkSeek (i + 1) == RECFOUND)
		{
		    fprintf (printFileDesc, "%s\n", AlarmFormat (&printtrack, NULL, NULL));
		}
	    }
	}
	else    // print entire list
	{
	    for (i = 1; i <= (int)(TrkNumEntries ()); i++)
	    {
		TrkSeek (i);
		fprintf (printFileDesc, "%s\n", AlarmFormat (&printtrack, NULL, NULL));
	    };
	}

	fclose (printFileDesc);
	if (strlen (linePrinter))
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"Operator Tracking\" %s |lpr -P\"%s\"", 
			      printFileName, linePrinter);
#else
	    sprintf (txtBuf, "lpr -p -r -P\"%s\" -T\"Operator Tracking\" %s", 
			      linePrinter, printFileName);
#endif SOLARIS
	}
	else
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"Operator Tracking\" %s |lpr", printFileName);
#else
	    sprintf (txtBuf, "lpr -p -r -T\"Operator Tracking\" %s", printFileName);
#endif SOLARIS
	}
	system (txtBuf);
#ifdef SOLARIS
	sprintf (txtBuf, "rm %s", printFileName);
	system (txtBuf);
#endif SOLARIS
    }
}


/*******************************************************************
*
*    AlarmTrackDialog Windows Print function
*
*******************************************************************/

void AlarmTrackDialog::WindowsPrint (int whatToPrint)
{
    if (app -> printer == NULL)
    {
	alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_NoPrinterMsg);
	return;
    }
    return;
}


vkindTYPED_FULL_DEFINITION(alarmTrackCommands, vcommandSelector, "alarmTrackCommands");
/*******************************************************************
*
*    menu selector commands for the alarm track dialog
*
*******************************************************************/

int alarmTrackCommands::Issue (vdict *dict)
{
    int			command;
    AlarmTrackDialog	*alarmTrackDialog;

    alarmTrackDialog = (AlarmTrackDialog *) dict -> GetElem (vname_Dialog);
    alvuApplication *app = alarmTrackDialog -> GetApp ();
    vresource alvuRes = app -> alvuResource;

    command = GetSelector ();

    switch (command)
    {
	case OPTRK_THISWIN_ITEM:
	case OPTRK_ENTIRESUM_ITEM:
	{
	    AldlgListView *view = alarmTrackDialog -> alarmListView;
	    vlist *list = view -> GetList ();
	    int rowCount = 0;
	    if (list != NULL)
	    {
		rowCount = list -> GetRowCount ();
	    }
	    if ((list == NULL) || (rowCount == 0))
	    {
		alvuMessageDialogCenter (alarmTrackDialog, ALVU_INFO_NOTICE,
					 alvu_NothingToPrint);
		break;
	    }
#ifdef WIN32
	    if (command == OPTRK_THISWIN_ITEM)
	    {
		alarmTrackDialog -> WindowsPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmTrackDialog -> WindowsPrint (PRINT_ENTIRE_LIST);
	    }
#else
	    if (command == OPTRK_THISWIN_ITEM)
	    {
		alarmTrackDialog -> UnixPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmTrackDialog -> UnixPrint (PRINT_ENTIRE_LIST);
	    }
#endif WIN32
	    break;
	}

	case OPTRK_MODFILT_ITEM:
	{
	    vresource filtRes = vresourceGet (alvuRes, alvu_FilterDialog);
	    FilterDialog *filterDialog = new FilterDialog (filtRes);
	    filterDialog -> OpenFilter (app, NULL, &alarmTrackDialog -> filter, MODIFY_FILTER);
	    filterDialog -> NewFilter (&alarmTrackDialog -> filter);
	    delete filterDialog;
	    alarmTrackDialog -> SetDialogTitle ();
	    TrkSetFilter (&alarmTrackDialog -> filter);
	    alarmTrackDialog -> SetPosition (100);
	}

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}


/*******************************************************************
*
*    Alarm Points Dialog
*	subclassed from AlarmDialog
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AlarmPointsDialog, AlarmDialog, "AlarmPointsDialog");

RepCriteria lastcriteria;

/*******************************************************************
*
*    Alarm Points Dialog, LoadInit ()
*
*******************************************************************/

void AlarmPointsDialog::LoadInit (vresource resource)
{
    AlarmDialog::LoadInit (resource);
}

/*******************************************************************
*
*    Alarm Points Dialog, destroy function
*
*******************************************************************/

void AlarmPointsDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    AlarmDialog::Destroy ();
} 


/*******************************************************************
*
*    AlarmPointsDialog Open function
*
*******************************************************************/

void AlarmPointsDialog::AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt)
{
    app = alvuApp;
    filter = *filt;
    PopupMenu = NULL;
    criteria = lastcriteria;

    vresource appres = alvuApp -> alvuResource;
    vresource rptRes = vresourceGet (appres, alvu_ReportDialog);
    ReportDialog *reportDialog = new ReportDialog (rptRes);
    reportDialog -> OpenReport (alvuApp, &this -> criteria);
    if (reportDialog -> GetResult () == vFALSE)
    {
	delete reportDialog;
 	return;
    }
    RepCriteria *currcriteria = reportDialog -> GetCriteria ();
    criteria = *currcriteria;
    lastcriteria = criteria;
    delete reportDialog;

    BuildMenus ();
    SetDialogTitle ();

    ackable = False;
    hasdate = False;
    hasdesc = False;
    monitored = False;
    fileBased = False;
    pauseButton -> Disable ();

    AlarmDialog::AlarmDialogOpen (alvuApp, filt);

    displaySize = alarmListView -> GetListNumElements ();

    RefreshDisplay ();
    SetPosition (0);
}


/*******************************************************************
*
*    AlarmPointsDialog Build menu function 
*
*******************************************************************/

void AlarmPointsDialog::BuildMenus (void)
{
    vresource titles = app -> titlesResource;
    vmenu	*newMenu;
    vmenuItem	*menuItem;
    vstr	*label;

    // the print menu

    label = vresourceGetString (titles, alvu_PrintMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ThisWindow);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlrptThisWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_EntireReport);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlrptEntireReport);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_SaveRptAs);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlrptSaveRptAs);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the filter menu

    label = vresourceGetString (titles, alvu_FilterMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ModifyFilter);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlrptModFilter);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // load up menuNames with resources
    menuNames [RPTS_THISWIN_ITEM]   = alvu_AlrptThisWindow;
    menuNames [RPTS_ENTIRESUM_ITEM] = alvu_AlrptEntireReport;
    menuNames [RPTS_SAVEAS_ITEM]    = alvu_AlrptSaveRptAs;
    menuNames [RPTS_MODFILT_ITEM]   = alvu_AlrptModFilter;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;
    // populate our dictionary with menu commands
    for (i = RPTS_THISWIN_ITEM; i < RPTS_ITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	alarmPointCommands *command = new alarmPointCommands;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (menuNames[i], command);
    }

    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}


/*******************************************************************
*
*    AlarmPointsDialog Control function
*
*******************************************************************/

int AlarmPointsDialog::CtrlFunc ( int cmd, int param)
{
    char	*ptName;
    PvData	*pvdata;
    Filter	tempFilter;
    TreeNode	*groupLinkPtr;
    char	*guidptr;
    char	*dispptr;
    vstr	*pv;

    ListEntry	*entry = NULL;
    // get resource to alarmvu
    vresource alvuRes = app -> alvuResource;

    switch (cmd)
    {
	case EXPANDCLICK:
	    if (expandDialogPtr != NULL)
	    {
		delete expandDialogPtr;
		expandDialogPtr = NULL;
	    }
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (treeRoot, StripWhitespace (ptName)))
	    {
		vresource expRes = vresourceGet (alvuRes, alvu_TextDialog);
		expandDialogPtr = new TextDialog (expRes);
		expandDialogPtr -> OpenText (app, 
					     EXPAND_TEXT, 
					     AlarmFormat (&expandpoint, pvdata, NULL),
					     NULL);
	    }
	    break;

	case NAMECLICK:
	    // Adjust menu items using status
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (treeRoot, StripWhitespace (ptName)))
	    {
		vmenuItem *item;

		GetAlarmInfo (ptName, &guidptr, &dispptr);

		// get Guidance menuitem
		item = PopupMenu -> GetItemAt (GUIDANCE_MENUITEM);
		// if guidance for point or for parent group, enable
		if ( (guidptr != 0) || 
		     HaveScript (pvdata -> parentGroup -> groupParms.guidance) )
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Display menuitem
		item = PopupMenu -> GetItemAt (DISPLAY_MENUITEM);
		if ( (dispptr != 0) || 
		     HaveScript (pvdata -> parentGroup -> groupParms.displayName))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Detail menuitem
		item = PopupMenu -> GetItemAt (DETAIL_MENUITEM);
		item -> Enable ();
		// get Script menuitem
		item = PopupMenu -> GetItemAt (SCRIPT_MENUITEM);
		if ( HaveScript (pvdata -> parentGroup -> groupParms.manualScript) ||
		     HaveScript (pvdata -> parentGroup -> groupParms.auxManualScript))
		    item -> Enable ();
		else
		    item -> Disable ();
		// get Enable menuitem
		item = PopupMenu -> GetItemAt (ENABLE_MENUITEM);
		item -> Enable ();
		// get Disable menuitem
		item = PopupMenu -> GetItemAt (DISABLE_MENUITEM);
		item -> Enable ();
		// get Comment menuitem
		item = PopupMenu -> GetItemAt (COMMENT_MENUITEM);
		item -> Disable ();
		// get Help menuitem
		item = PopupMenu -> GetItemAt (HELP_MENUITEM);
		item -> Disable ();
		return True;
	    }
	    else
		return False;

	case SUMMARYCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (treeRoot, StripWhitespace (ptName)))
	    {
		tempFilter = filter;
		strncpy (tempFilter.group, pvdata -> parentGroup -> groupName, 
			 sizeof (tempFilter.group));
		tempFilter.group [sizeof (tempFilter.group)-1] = 0;

		if (alarmSummaryDialogPtr != NULL)
		{
		    delete alarmSummaryDialogPtr;
		}
		vresource alrmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
		alarmSummaryDialogPtr = new AlarmSummaryDialog (alrmDlgRes);
		alarmSummaryDialogPtr -> AlarmDialogOpen (app, &tempFilter);

	    }
	    break;

	case GROUPCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (treeRoot, StripWhitespace (ptName)))
	    {
		tempFilter = filter;
		strncpy (tempFilter.group, pvdata -> parentGroup -> groupName, 
			 sizeof (tempFilter.group));
		tempFilter.group [sizeof (tempFilter.group) - 1] = 0;
		if (dynamicGroupDialogPtr != NULL)
		{
		    delete dynamicGroupDialogPtr;
		}
		vresource alarmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
		dynamicGroupDialogPtr = new DynamicGroupDialog (alarmDlgRes);
		dynamicGroupDialogPtr -> AlarmDialogOpen (app, &tempFilter);
	    }
	    break;

	case GUIDECLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (treeRoot, StripWhitespace (ptName)))
	    {
		if (guidanceDialogPtr != NULL)
		{
		    delete guidanceDialogPtr;
		    guidanceDialogPtr = NULL;
		}
		// get any guidance on the point
		GetAlarmInfo (ptName, &guidptr, &dispptr);
		if (guidptr == NULL)
		{
		    // does parent alarm group have guidance
		    if (HaveScript (pvdata -> parentGroup -> groupParms.guidance))
		    {
			guidptr = pvdata -> parentGroup -> groupParms.guidance;
		    }
		}
		if (guidptr && *guidptr)
		{
		    vresource guidRes = vresourceGet (alvuRes, alvu_TextDialog);
		    guidanceDialogPtr = new TextDialog (guidRes);
		    guidanceDialogPtr -> OpenText (app, 
						   GUIDANCE_TEXT, 
						   guidptr,
						   ptName);
		}
	    }
	    break;

	case DISPLAYCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (treeRoot, StripWhitespace (ptName)))
	    {
		SpawnPointDisplay (pvdata);
	    }
	    break;

	case DETAILCLICK:
	    ptName = ItemName (param);
	    SpawnDetail (ptName);
	    break;

	case COMMANDCLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (treeRoot, StripWhitespace (ptName)))
	    {
		SpawnScript (pvdata);
	    }
	    break;

	case ENABLECLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (treeRoot, StripWhitespace (ptName)))
	    {
		if (EnablePoint (pvdata) != True)
		{
		    pv = vstrCloneScribed (vcharScribeSystem (ptName));
		    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
					     alvu_NoEnablePriv, pv);
		    vstrDestroy (pv);
		}
	    }
	    break;

	case DISABLECLICK:
	    ptName = ItemName (param);
	    if (pvdata = FindChannel (treeRoot, StripWhitespace (ptName)))
	    {
		if (DisablePoint (pvdata) != True)
		{
		    pv = vstrCloneScribed (vcharScribeSystem (ptName));
		    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE,
					     alvu_NoDisablePriv, pv);
		    vstrDestroy (pv);
		}
	    }
	    break;

	case UPCLICK:
	    if (groupLinkPtr = FindGroup (treeRoot, filter.group))
	    {
		if (groupLinkPtr -> parent)
		{
		    strncpy (filter.group, groupLinkPtr -> parent -> groupName,
			     sizeof (filter.group));
		    filter.group [sizeof (filter.group)-1] = 0;
		    filter.modified = TRUE;
		    SetDialogTitle ();
		}
	    }
	    Resync ();
	    break;

	case TOPCLICK:
	    strncpy (filter.group, treeRoot -> groupName, sizeof (filter.group));
	    filter.group [sizeof (filter.group)-1] = 0;
	    filter.modified = TRUE;
	    SetDialogTitle ();
	    Resync ();
	    break;
    }
    return True;
}


/*******************************************************************
*
*    AlarmPointsDialog Refresh Display function
*******************************************************************/

void AlarmPointsDialog::RefreshDisplay ()
{
    PvData *channelPtr = shareHeader -> alarmPoints;
    char **selections;
    int selCount = 0;
    int i;
    int	include;
    char typestring[25];

    selections = (char **) calloc (shareHeader -> alarmPointCount, sizeof (char *));
    ClearLines ();
    for (i = 0; i < shareHeader -> alarmPointCount; i++)
    {
	include = False;
	if ((channelPtr -> parentGroup != NULL) // deleted point if null!
	  && (channelPtr -> lastAlarmReceived.secPastEpoch)
	  && (!IsChannelFiltered (channelPtr, &filter, channelPtr -> severity)))
	{
	    switch (channelPtr -> status)
	    {
		 case SCAN_ALARM:
		    if (criteria.reportitems & RPTOFFSCAN)
			include = TRUE;
		    break;

		 case INHIBIT_ALARM:
		    if (criteria.reportitems & RPTDISABLE)
			include = TRUE;
		    break; 

		 case NO_INIT_ALARM:
		 case CALC_ALARM:
		 case LINK_ALARM:
		    if (criteria.reportitems & RPTINITFAIL)
			include = TRUE;
		    break;

		 case READ_ALARM:
		 case WRITE_ALARM:
		    if (criteria.reportitems & RPTIOFAIL)
			include = TRUE;
		    break;

		 case SIGNAL_ALARM:
		    if (criteria.reportitems & RPTBADSENS)
			include = TRUE;
		    break;

		 case TAGGED_ALARM:
		    if (criteria.reportitems & RPTSYSTAG)
			include = TRUE;
		    break;

		 case PERMISSIVE_ALM:
		 case TF_UNCOMMANDED:
		    ResetPVPoint ();
		    strncpy (typestring, GetPVPointInfo (channelPtr->pvName, "TYPE"),20);
		    trimblanks (typestring);
		    if (((strcmp ("pump", typestring) == 0)
			 && (criteria.reportitems & RPTPUMPFAULT))
		      || ((strcmp ("valve",typestring) == 0)
			 && (criteria.reportitems & RPTVALVFAULT)))
			include = TRUE;
		    break;

		 case PUMP_START:
		 case PUMP_STOP:
		 case TEMPERATURE:
		 case PRESSURE:
		 case VIBRATION:
		    if (criteria.reportitems & RPTPUMPFAULT)
			include = TRUE;
		    break;

		 case VALVE_OPEN_ALM:
		 case VALVE_CLOSE_ALM:
		 case VALVE_FAULT_ALM:
		 case VALVE_POSITION:
		 case VALVE_TORQUE_LAG:
		    if (criteria.reportitems & RPTVALVFAULT)
			include = TRUE;
		    break;
	    }
	    if (criteria.reportitems & RPTINFOTAG)
	    {
		ResetPVPoint ();
		if (GetInfoTag (channelPtr->pvName))
		    include = TRUE;
	    }
	    if (criteria.reportitems & RPTMANUAL)
	    {
		ResetPVPoint ();
		// get and see if omsl field is set to supervisory
		if (GetManualMode (channelPtr->pvName)) 
		    include = TRUE;
	    }
	}
	if (include)
	    selections [selCount++] = channelPtr->pvName;
	channelPtr++;
    }
    qsort ((void *)selections, selCount, sizeof (char *), (int (*)(const void *, const void *)) DoStrCmp);

    for (i = 0; i < selCount; i++)
    {
	if (channelPtr = FindChannel (treeRoot, selections[i]))
	{
	    AddLine ( -1, "", selections [i],
		      AlarmFormat (&displaypoint, channelPtr, NULL));
	}
    }
    if (dataList != NULL)
    {
	alarmListView -> SetList (dataList);
	vlistSelection *selected = alarmListView -> GetSelection ();
	dataList -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	dataList -> SetAtLeastOneSelection (selected, vFALSE);
    }
    free ((char *)selections);
    SetPosition (Pos());
}


/*******************************************************************
*
*    AlarmPointsDialog Resync function
*******************************************************************/

void AlarmPointsDialog :: Resync ()
{
    RefreshDisplay ();
}


/*******************************************************************
*
*    AlarmPointsDialog set title function to be written by subclasses
*
*******************************************************************/

void AlarmPointsDialog::SetDialogTitle (void)
{
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    vstr *format = vresourceGetString (app -> titlesResource, alvu_AlarmReportTitle);
    vstr *criterianame = vstrCloneScribed (vcharScribeSystem (criteria.name));
    vstr *mod = vstrCloneScribed (vcharScribeLiteral (filter.modified ? "*" : ""));
    vstr *name = vstrCloneScribed (vcharScribeSystem (filter.name));
    title = vstrCloneScribed (vcharScribeFormat (format, criterianame, mod, name));
    SetTitle (title);
    vstrDestroy (format);
    vstrDestroy (criterianame);
    vstrDestroy (mod);
    vstrDestroy (name);
}


/*******************************************************************
*
*    AlarmPointsDialog Unix Print function
*
*******************************************************************/

void AlarmPointsDialog::UnixPrint (int whatToPrint)
{
    FILE    *printFileDesc;
    vstr    *prtFileTmp;
    char    printFileName [256];
    int	    i;
 
    printFileName [0] = 0;
    if (whatToPrint == PRINT_SAVE_AS)
    {
	    vresource oneLineRes = vresourceGet (app -> alvuResource, alvu_OneLineDialog);
	    OneLineDialog *oneLineDlg = new OneLineDialog (oneLineRes);
	    oneLineDlg -> OpenOneLine (app, ONELINE_FILENAME);
	    vchar *oneLineText = oneLineDlg -> GetOneLineText ();
	    if (vcharLength (oneLineText) != 0)
	    {
		vcharExportBounded (vcharScribeSystemExport (oneLineText),
				    printFileName,
				    sizeof (printFileName));
		printFileName [sizeof (printFileName) - 1] = 0;
	    }
	    delete oneLineDlg;
	    if (strlen (printFileName ) == 0)
	    {
		return;
	    }
    }
    else
    {
	strcpy (printFileName, PRINT_FILE_TEMP);
	if ( !mktemp (printFileName))
	{
	    prtFileTmp = vstrCloneScribed (vcharScribeLiteral (PRINT_FILE_TEMP));
	    alvuMessageDialogCenter (this, ALVU_INFO_NOTICE, alvu_TempPrtFileErr, prtFileTmp);
	    vstrDestroy (prtFileTmp);
	    return;
	}
    }
    if ( !(printFileDesc = fopen (printFileName, "w")))
    {
	prtFileTmp = vstrCloneScribed (vcharScribeLiteral (PRINT_FILE_TEMP));
	alvuMessageDialogCenter (this, ALVU_INFO_NOTICE, alvu_TempPrtFileErr, prtFileTmp);
	vstrDestroy (prtFileTmp);
    }
    else
    {
	vlist	*list = alarmListView -> GetList ();
	int	rowCount = list -> GetRowCount ();
	int	startLine, endLine;

	if (whatToPrint == PRINT_THIS_WINDOW)
	{
	    startLine = alarmListView -> GetTopRow ();
	    endLine = alarmListView -> GetBottomRow ();
	}
	else    // print entire list
	{
	    startLine = 0;
	    endLine = rowCount - 1;
	}

	for (i = startLine; i <= endLine; i++)
	{
	    fprintf (printFileDesc, "%s\n", 
		AlarmFormat (&displaypoint, FindChannel (treeRoot, ItemName (i)), NULL));
	}

	fclose (printFileDesc);
	// finished if saved to file
	if (whatToPrint == PRINT_SAVE_AS)
	{
	    return;
	}
	// else setup to go to printer
	vstr *capFormat = vresourceGetString (app -> messagesResource, alvu_AlmPtsPrtTitle);
	vstr *name = vstrCloneScribed (vcharScribeSystem (criteria.name));
	vstr *filtMod = vstrCloneScribed (vcharScribeSystem (filter.modified ? "*" : " "));
	vstr *filtName = vstrCloneScribed (vcharScribeSystem (filter.name));
	vstr *headerStr = vstrCloneScribed (vcharScribeFormat (capFormat, 
							       name,
							       filtMod, 
							       filtName));
	vcharExportBounded (vcharScribeSystemExport (headerStr),
			    caption,
			    sizeof (caption));
	caption [sizeof (caption) - 1] = 0;
	vstrDestroy (capFormat);
	vstrDestroy (name);
	vstrDestroy (filtMod);
	vstrDestroy (filtName);
	vstrDestroy (headerStr);
	if (strlen (linePrinter))
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"%s\" %s |lpr -P\"%s\"", 
			      caption, printFileName, linePrinter);
#else
	    sprintf (txtBuf, "lpr -p -r -P\"%s\" -T\"%s\" %s", 
			      linePrinter, caption, printFileName);
#endif SOLARIS
	}
	else
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"%s\" %s |lpr", caption, printFileName);
#else
	    sprintf (txtBuf, "lpr -p -r -T\"%s\" %s", caption, printFileName);
#endif SOLARIS
	}
	system (txtBuf);
#ifdef SOLARIS
	sprintf (txtBuf, "rm %s", printFileName);
	system (txtBuf);
#endif SOLARIS
    }
}


/*******************************************************************
*
*    AlarmPointsDialog Windows Print function
*
*******************************************************************/

void AlarmPointsDialog::WindowsPrint (int whatToPrint)
{
    if (app -> printer == NULL)
    {
	alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_NoPrinterMsg);
	return;
    }
    return;
}


vkindTYPED_FULL_DEFINITION(alarmPointCommands, vcommandSelector, "alarmPointCommands");
/*******************************************************************
*
*    menu selector commands for the alarm report dialog
*
*******************************************************************/

int alarmPointCommands::Issue (vdict *dict)
{
    int			command;
    AlarmPointsDialog	*alarmReportDialog;

    alarmReportDialog = (AlarmPointsDialog *) dict -> GetElem (vname_Dialog);
    alvuApplication *app = alarmReportDialog -> GetApp ();
    vresource alvuRes = app -> alvuResource;

    command = GetSelector ();

    switch (command)
    {
	case RPTS_THISWIN_ITEM:
	case RPTS_ENTIRESUM_ITEM:
	case RPTS_SAVEAS_ITEM:
	{
	    AldlgListView *view = alarmReportDialog -> alarmListView;
	    vlist *list = view -> GetList ();
	    int rowCount = 0;
	    if (list != NULL)
	    {
		rowCount = list -> GetRowCount ();
	    }
	    if ((list == NULL) || (rowCount == 0))
	    {
		alvuMessageDialogCenter (alarmReportDialog, ALVU_INFO_NOTICE,
					 alvu_NothingToPrint);
		break;
	    }
#ifdef WIN32
	    if (command == RPTS_THISWIN_ITEM)
	    {
		alarmReportDialog -> WindowsPrint (PRINT_THIS_WINDOW);
	    }
	    else if (command == RPTS_ENTIRESUM_ITEM)
	    {
		alarmReportDialog -> WindowsPrint (PRINT_ENTIRE_LIST);
	    }
	    else
	    {
		alarmReportDialog -> WindowsPrint (PRINT_SAVE_AS);
	    }
#else
	    if (command == RPTS_THISWIN_ITEM)
	    {
		alarmReportDialog -> UnixPrint (PRINT_THIS_WINDOW);
	    }
	    else if (command == RPTS_ENTIRESUM_ITEM)
	    {
		alarmReportDialog -> UnixPrint (PRINT_ENTIRE_LIST);
	    }
	    else
	    {
		alarmReportDialog -> UnixPrint (PRINT_SAVE_AS);
	    }
#endif WIN32
	    break;
	}

	case RPTS_MODFILT_ITEM:
	{
	    vresource filtRes = vresourceGet (alvuRes, alvu_FilterDialog);
	    FilterDialog *filterDialog = new FilterDialog (filtRes);
	    filterDialog -> OpenFilter (app, NULL, &alarmReportDialog -> filter, MODIFY_FILTER);
	    filterDialog -> NewFilter (&alarmReportDialog -> filter);
	    delete filterDialog;
	    alarmReportDialog -> SetDialogTitle ();
	    alarmReportDialog -> RefreshDisplay ();
	    alarmReportDialog -> SetPosition (0);
	}

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}


/*******************************************************************
*
*	Alarm System Dialog Class
*		subclassed from AlarmDialog
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AlarmSystemDialog, AlarmDialog, "AlarmSystemDialog");

AlarmLogRecord	    alarmSysRecord;
AlarmLogRecord	    alarmSysRecords [ALVU_MAXITEMS];

/*******************************************************************
*
*    AlarmSystemDialog, LoadInit ()
*
*******************************************************************/

void AlarmSystemDialog::LoadInit (vresource resource)
{
    fileButtonContainer = FileButtonContainer::RequestSubLoad (this,
					    alvu_BlankContainer);
    cancelDataContainer = CancelDataButtonContainer::RequestSubLoad (this,
					    alvu_ButtonContainer);
    AlarmDialog::LoadInit (resource);

    fileButtonContainer -> firstButton    -> StoreAlarmDialog (this);
    fileButtonContainer -> pageupButton   -> StoreAlarmDialog (this);
    fileButtonContainer -> lineupButton   -> StoreAlarmDialog (this);
    fileButtonContainer -> linedownButton -> StoreAlarmDialog (this);
    fileButtonContainer -> pagedownButton -> StoreAlarmDialog (this);
    fileButtonContainer -> lastButton     -> StoreAlarmDialog (this);
    cancelDataContainer -> cancelDataButton -> StoreAlarmDialog (this);
}

/*******************************************************************
*
*    AlarmSystemDialog, destroy function
*
*******************************************************************/

void AlarmSystemDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    AlarmDialog::Destroy ();
}


/*******************************************************************
*
*    AlarmSystemDialog Class, Open
*
*******************************************************************/

void AlarmSystemDialog::AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt)
{
    app = alvuApp;
    filter = *filt;
    PopupMenu = NULL;

    pauseButton -> Disable ();

    BuildMenus ();
    SetDialogTitle ();

    // the args for base class AlarmDialog
    ackable = False;
    hasdate = False;
    hasdesc = False;
    monitored = False;
    fileBased = True;

    inMove = False;
    inRefresh = False;
    reqPosition = -1;

    int i;
    for (i = 0; i <= ALVU_MAXITEMS; i++)
    {
	srcType [i].type = NOINFO;
	srcType [i].key = 0;
	srcType [i].eventTime.secPastEpoch = 0;
	srcType [i].eventTime.nsec = 0;
    }

    // open all system info sources - alarmlog, op trak, trend notes, op notes
    TrkInit ();
    TrkSetFilter (&this -> filter);
    // start with no entries
    startName [0] = 0;
    if (strlen (filter.start))
    {
	GetGMT_TS_STAMP (filter.start, &this -> startTime);
    }
    else
    {
	startTime.secPastEpoch = 0;
    }
    startTime.nsec = 0;

    Log_SetFilter (&filter, ALRMSYS_INDEX_FILE); // see alarmfil.c
    groupLinkPtr = FindGroup (treeRoot, filter.group);

    startName [0] = 0;
    // turn off the scroll bar for the listview, we are using
    // buttons to control the listview
    alarmListView -> SetVertBar (vFALSE);

    AlarmDialog::AlarmDialogOpen (alvuApp, filt);

    displaySize = alarmListView -> GetListNumElements (); 

    SetPosition (100);
}


/*******************************************************************
*
*    AlarmSystemDialog Build menu function 
*
*******************************************************************/

void AlarmSystemDialog::BuildMenus (void)
{
    vresource titles = app -> titlesResource;
    vmenu	*newMenu;
    vmenuItem	*menuItem;
    vstr	*label;

    // the print menu

    label = vresourceGetString (titles, alvu_PrintMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ThisWindow);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlsysThisWindow);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_EntireSummary);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlsysEntireSummary);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // the filter menu

    label = vresourceGetString (titles, alvu_FilterMenuTitle);
    newMenu = new vmenu;
    newMenu -> SetTitle (label);
    vstrDestroy (label);
    menuItem = new vmenuItem;
    label = vresourceGetString (titles, alvu_ModifyFilter);
    menuItem -> SetTitle (label);
    menuItem -> SetTag (alvu_AlsysModFilter);
    newMenu -> AppendItem (menuItem);
    vstrDestroy (label);
    alarmMenuBar -> AppendMenu (newMenu);

    // load up menuNames with resources
    menuNames [ALSYS_THISWIN_ITEM]   = alvu_AlsysThisWindow;
    menuNames [ALSYS_ENTIRESUM_ITEM] = alvu_AlsysEntireSummary;
    menuNames [ALSYS_MODFILT_ITEM]   = alvu_AlsysModFilter;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;
    // populate our dictionary with menu commands
    for (i = ALSYS_THISWIN_ITEM; i < ALSYS_ITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	alarmSystemCommands *command = new alarmSystemCommands;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (menuNames[i], command);
    }

    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}


/*******************************************************************
*
*    AlarmSystemDialog Control function 
*
*******************************************************************/

int AlarmSystemDialog::CtrlFunc ( int cmd, int param)
{
    // get resource to alarmvu
    vresource alvuRes = app -> alvuResource;

    switch (cmd)
    {
	case MOVECLICK:
	    if (inMove == False)
	    {
		inMove = True;
		filePosition = param;
		RefreshDisplay ();
		inMove = False;
	    }
	    break;

	case WINCLOSE:
	    if (inRefresh == True)
	    {
		return (False);
	    }
	    else
	    {
		break;
	    }

	case EXITCLICK:
	    if (inRefresh == True)
	    {
		return (False);
	    }
	    else
	    {
		break;
	    }

	case EXPANDCLICK:
	{
	    if (expandDialogPtr != NULL)
	    {
		delete expandDialogPtr;
		expandDialogPtr = NULL;
	    }

	    vresource expRes = vresourceGet (alvuRes, alvu_TextDialog);
	    if (srcType [param].type == TRKINFO)
	    {
		if (TrkSeek (srcType [param].key + 1) == RECFOUND)
		{
	    	    expandDialogPtr = new TextDialog (expRes);
		    expandDialogPtr -> OpenText (app, 
					     EXPAND_TEXT, 
					     AlarmFormat (&expandtrack, NULL, NULL),
					     NULL);
		}
	    }
	    else if (srcType [param].type == LOGINFO)
	    {
	    	expandDialogPtr = new TextDialog (expRes);
		if (strcmp (alarmSysRecords [srcType [param].key].status, "ACK") == 0)
		{
		    expandDialogPtr -> OpenText (app, 
					 EXPAND_TEXT, 
					 AlarmFormat (&expandack,
						      NULL,
						      &alarmSysRecords [srcType [param].key]),
					 NULL);
		}
		else
		{
		    expandDialogPtr -> OpenText (app, 
				   EXPAND_TEXT, 
				   AlarmFormat (&expandalarm,
						NULL,
						&alarmSysRecords [srcType [param].key].date),
				   NULL);
		}
	    }

	    break;
	}

	case NAMECLICK:
	{
	    return False;
	}

	case UPCLICK:
	    if (groupLinkPtr = FindGroup (treeRoot, filter.group))
	    {
		if (groupLinkPtr -> parent)
		{
		    strncpy (filter.group, 
			     groupLinkPtr -> parent -> groupName, 
			     sizeof (filter.group));
		    filter.group [sizeof (filter.group)-1] = 0;
		}
	    }
	    filter.modified = TRUE;
	    SetDialogTitle ();
	    TrkSetFilter (&this -> filter);
	    SetPosition (100);
	    break;

	case TOPCLICK:
	    strncpy (filter.group, treeRoot -> groupName, sizeof (filter.group));
	    filter.group [sizeof (filter.group)-1] = 0;
	    filter.modified = TRUE;
	    SetDialogTitle ();
	    TrkSetFilter (&this -> filter);
	    SetPosition (100);
	    break;

    }
    return True;
}


/*******************************************************************
*
*    AlarmSystemDialog RefreshDisplay function 
*
*******************************************************************/

void AlarmSystemDialog :: RefreshDisplay ( )
{
    int i;
    int j;
    int retloopcnt;
    int eof = False;
    int trkIx;
    int sysIx;
    int checkType;
    int checkKey;
    char checkName[AL_POINTNAME_SIZE+1];
    TS_STAMP checkTime;
    int lastItem;

    if (inRefresh == True)
    {
	return;
    }
    inRefresh = True;

    for (lastItem = 0; srcType[lastItem+1].type != NOINFO; lastItem++)
	; /* just identify index of last item */


    /* while we have a handle on the last item, save info for shift operation */
    if ((filePosition == 25) || (filePosition == 49) || (filePosition == 50))
    {
	/* (25) stop collecting when last item is equal to old first item */
	/* (49) stop collecting when second item is equal to old first item */
	/* (50) stop collecting when first item is equal to old first item */
	checkType = srcType [0].type;
	if (checkType == TRKINFO)
	{
	    checkKey = srcType [0].key;
	}
	else
	{
	    strncpy (checkName, topName, AL_POINTNAME_SIZE + 1);
	}
	checkTime = srcType [0].eventTime;
    }
    else if (filePosition == 51)
    {
	/* stop collecting when check item is equal to old second item */
	checkType = srcType [1].type;
	if (checkType == TRKINFO)
	{
	    checkKey = srcType [1].key;
	}
	else
	{
	    strncpy (checkName, alarmSysRecords[1].pointName, AL_POINTNAME_SIZE + 1);
	}
	checkTime = srcType [1].eventTime;
    }
    else if (filePosition == 75)
    {
	/* stop collecting when first item is equal to old last item */
	checkType = srcType [lastItem].type;
	if (checkType == TRKINFO)
	{
	    checkKey = srcType [lastItem].key;
	}
	else
	{
	    strncpy (checkName, lastName, AL_POINTNAME_SIZE + 1);
	}
	checkTime = srcType [lastItem].eventTime;
    }

    ClearLines ();
    SetFileItems (100);
    StartCancelBox ();

    /* Get enough log records to fill screen (if available) */
    if (startName[0] == 0)
    {
	if (Log_RecordByTime ( "! ", &startTime, &alarmSysRecords[0]) != NORECORD)
	{
            retloopcnt = Log_RecordsByTime (alarmSysRecord.pointName,
		     &alarmSysRecord.eventtime, alarmSysRecords,
		     1, ALRMSYS_INDEX_FILE,
		     (int (__cdecl *)(void))CheckCancelGW, this);
	    if (retloopcnt)
	    {
		strncpy (startName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE + 1);
		startTime = alarmSysRecords[0].eventtime;
		strncpy (topName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE + 1);
		topTime = alarmSysRecords[0].eventtime;
	    }
	    else
	    {
		SetFileItems (0);
		SetFilePos (0);
	    }
	}
	else
	{
	    SetFileItems (0);
	    SetFilePos (0);
	}
    }

    int loopcnt = ALVU_MAXITEMS;
    if (loopcnt > Size())
	loopcnt = Size();

    if (filePosition == 0)
    {
	strncpy (topName, startName, AL_POINTNAME_SIZE + 1);
	topTime = startTime;
        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
		     loopcnt, ALRMSYS_INDEX_FILE,
		     (int (__cdecl *)(void))CheckCancelGW, this);
    }
    else if (filePosition == 25)
    {
	/* find records backwards from top record (even if it's from log file) */
	topTime = srcType[0].eventTime;
        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
		     -loopcnt, ALRMSYS_INDEX_FILE,
		     (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled ();
	    inRefresh = False;
	    return;
	}
	if (retloopcnt)
	{
	    if (((startTime.secPastEpoch > alarmSysRecords[0].eventtime.secPastEpoch)
		 || ((startTime.secPastEpoch == alarmSysRecords[0].eventtime.secPastEpoch)
			 && (startTime.nsec > alarmSysRecords[0].eventtime.nsec)))
	      || (memcmp (&startTime, &alarmSysRecords[0].eventtime, sizeof(startTime)) == 0
		 && strncmp (startName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE+1) >= 0))
	    { /* if above start, backup */
		strncpy (topName, startName, AL_POINTNAME_SIZE + 1);
		topTime = startTime;
	        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
			     loopcnt, ALRMSYS_INDEX_FILE,
			     (int (__cdecl *)(void))CheckCancelGW, this);
		if (CheckCancelled ())
		{
		    ShowCancelled ();
		    inRefresh = False;
		    return;
		}
	    }
	    else
	    {
		strncpy (topName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE + 1);
		topTime = alarmSysRecords[0].eventtime;
	    }
	}
	else
	{
	    retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
				loopcnt, ALRMSYS_INDEX_FILE,
				(int (__cdecl *)(void))CheckCancelGW, this);
	    if (CheckCancelled ())
	    {
		ShowCancelled ();
		inRefresh = False;
		return;
	    }
	}
    }
    else if (filePosition == 49)
    {
	/* find  1 record backwards from top record (OR 2 if first rec is LOGINFO) */
	if (srcType[0].type == LOGINFO)
	{
            retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
		     -2, ALRMSYS_INDEX_FILE,
		     (int (__cdecl *)(void))CheckCancelGW, this);
	    if (CheckCancelled ())
	    {
		ShowCancelled ();
		inRefresh = False;
		return;
	    }
	}
	else
	{
	    topTime = srcType[0].eventTime;
            retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
		     -1, ALRMSYS_INDEX_FILE,
		     (int (__cdecl *)(void))CheckCancelGW, this);
	    if (CheckCancelled ())
	    {
		ShowCancelled ();
		inRefresh = False;
		return;
	    }
	}
	if ((retloopcnt == 2) || (srcType[0].type != LOGINFO))
	{
	    if (((startTime.secPastEpoch > alarmSysRecords[0].eventtime.secPastEpoch)
		 || ((startTime.secPastEpoch == alarmSysRecords[0].eventtime.secPastEpoch)
			 && (startTime.nsec > alarmSysRecords[0].eventtime.nsec)))
	      || (memcmp (&startTime, &alarmSysRecords[0].eventtime, sizeof(startTime)) == 0
		 && strncmp (startName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE+1) >= 0))
	    { /* if above start, backup */
		strncpy (topName, startName, AL_POINTNAME_SIZE+1);
		topTime = startTime;
	        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
			     loopcnt, ALRMSYS_INDEX_FILE,
			     (int (__cdecl *)(void))CheckCancelGW, this);
		if (CheckCancelled ())
		{
		    ShowCancelled ();
		    inRefresh = False;
		    return;
		}
	    }
	    else
	    {
		strncpy (topName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE + 1);
		topTime = alarmSysRecords[0].eventtime;
	        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
			     loopcnt, ALRMSYS_INDEX_FILE,
			     (int (__cdecl *)(void))CheckCancelGW, this);
		if (CheckCancelled ())
		{
		    ShowCancelled ();
		    inRefresh = False;
		    return;
		}
	    }
	}
	else
	{
	    retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
				loopcnt, ALRMSYS_INDEX_FILE,
				(int (__cdecl *)(void))CheckCancelGW, this);
	    if (CheckCancelled ())
	    {
		ShowCancelled ();
		inRefresh = False;
		return;
	    }
	}
    }
    else if ((filePosition == 50) ||  // refresh with current displaySize 
	     (filePosition == 51))    // find records forward from second record
    {
	if (srcType[0].type == LOGINFO)
	{
            retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
				2, ALRMSYS_INDEX_FILE,
				(int (__cdecl *)(void))CheckCancelGW, this);
	    if (CheckCancelled ())
	    {
		ShowCancelled ();
		inRefresh = False;
		return;
	    }
	    if ((filePosition == 51) && (retloopcnt == 2))
	    {
		strncpy (topName, alarmSysRecords[1].pointName, AL_POINTNAME_SIZE + 1);
		topTime = alarmSysRecords[1].eventtime;
	    }
	}
	else
	{
	    topTime = srcType[0].eventTime;
	}
        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
		     	loopcnt, ALRMSYS_INDEX_FILE,
			(int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled ();
	    inRefresh = False;
	    return;
	}
        if (retloopcnt < loopcnt)
	{ // just position at end 
	  // find records backwards from end of file
            retloopcnt = Log_RecordsByTime (0, 0, alarmSysRecords,
		     -loopcnt, ALRMSYS_INDEX_FILE,
		     (int (__cdecl *)(void))CheckCancelGW, this);
	    eof = True;
	    if (retloopcnt)
	    {
		if (((startTime.secPastEpoch > alarmSysRecords[0].eventtime.secPastEpoch)
			 || ((startTime.secPastEpoch == alarmSysRecords[0].eventtime.secPastEpoch)
				 && (startTime.nsec > alarmSysRecords[0].eventtime.nsec)))
		  || (memcmp (&startTime, &alarmSysRecords[0].eventtime, sizeof(startTime)) == 0
		    && strncmp (startName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE+1) >= 0))
		{ // if above start, reset to start
		    strncpy (topName, startName, AL_POINTNAME_SIZE+1);
		    topTime = startTime;
	            retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
			     loopcnt, ALRMSYS_INDEX_FILE,
			     (int (__cdecl *)(void))CheckCancelGW, this);
		    if (CheckCancelled ())
		    {
			ShowCancelled ();
			inRefresh = False;
			return;
		    }
		}
		else
		{
		    strncpy (topName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE + 1);
		    topTime = alarmSysRecords[0].eventtime;
		}
	    }
	}
    }
    else if (filePosition == 75) // find records forward from last record
    {
	if (lastName[0])
	{
	    strncpy (topName, lastName, AL_POINTNAME_SIZE+1);
	    topTime = lastTime;
	}
	else
	{
	    topTime = srcType[lastItem].eventTime;
	}
        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
		     loopcnt, ALRMSYS_INDEX_FILE,
		     (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled ();
	    inRefresh = False;
	    return;
	}
       if (retloopcnt < loopcnt)
	{ // just position at end
	  // find records backwards from end of file 
            retloopcnt = Log_RecordsByTime (0, 0, alarmSysRecords,
		     -loopcnt, ALRMSYS_INDEX_FILE,
		     (int (__cdecl *)(void))CheckCancelGW, this);
	    if (CheckCancelled ())
	    {
		ShowCancelled ();
		inRefresh = False;
		return;
	    }
	    eof = True;
	    if (retloopcnt)
	    {
		if (((startTime.secPastEpoch > alarmSysRecords[0].eventtime.secPastEpoch)
			|| ((startTime.secPastEpoch == alarmSysRecords[0].eventtime.secPastEpoch)
			    && (startTime.nsec > alarmSysRecords[0].eventtime.nsec)))
		  || (memcmp (&startTime, &alarmSysRecords[0].eventtime, sizeof(startTime)) == 0
		    && strncmp (startName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE+1) >= 0))
		{ // if above start, reset to start
		    strncpy (topName, startName, AL_POINTNAME_SIZE+1);
		    topTime = startTime;
	            retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
			     loopcnt, ALRMSYS_INDEX_FILE,
			     (int (__cdecl *)(void))CheckCancelGW, this);
		    if (CheckCancelled ())
		    {
			ShowCancelled ();
			inRefresh = False;
			return;
		    }
		}
		else
		{
		    strncpy (topName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE + 1);
		    topTime = alarmSysRecords[0].eventtime;
		}
	    }
	}
    }
    else if (filePosition >= 90)
    {
	// find records backwards from end of file
        retloopcnt = Log_RecordsByTime (0, 0, alarmSysRecords,
		     -loopcnt, ALRMSYS_INDEX_FILE,
		     (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled ();
	    inRefresh = False;
	    return;
	}
	eof = True;
	if (retloopcnt)
	{
	    if (((startTime.secPastEpoch > alarmSysRecords[0].eventtime.secPastEpoch)
		 || ((startTime.secPastEpoch == alarmSysRecords[0].eventtime.secPastEpoch)
		    && (startTime.nsec > alarmSysRecords[0].eventtime.nsec)))
	      || (memcmp (&startTime, &alarmSysRecords[0].eventtime, sizeof(startTime)) == 0
		 && strncmp (startName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE+1) >= 0))
	    { // if above start, reset to start
		strncpy (topName, startName, AL_POINTNAME_SIZE+1);
		topTime = startTime;
	        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
			     loopcnt, ALRMSYS_INDEX_FILE,
			     (int (__cdecl *)(void))CheckCancelGW, this);
		if (CheckCancelled ())
		{
		    ShowCancelled ();
		    inRefresh = False;
		    return;
		}
	    }
	    else
	    {
		strncpy (topName, alarmSysRecords[0].pointName, AL_POINTNAME_SIZE + 1);
		topTime = alarmSysRecords[0].eventtime;
	    }
	}
    }
    else // find records forward from top record
    {
        retloopcnt = Log_RecordsByTime (topName, &topTime, alarmSysRecords,
		     loopcnt, ALRMSYS_INDEX_FILE,
		     (int (__cdecl *)(void))CheckCancelGW, this);
	if (CheckCancelled ())
	{
	    ShowCancelled ();
	    inRefresh = False;
	    return;
	}
    }
    sysIx = 0;

    /* Get enough Track items to fill screen (if available) just position for reads */
    if (filePosition == 0)
	trkPosition = 0;
    else if (filePosition == 25)
    {
	while ( (trkPosition && 
		(TrkSeek (trkPosition+1) == RECFOUND) && 
		( ((unsigned int)TrkEventTime () - TS_EPOCH_SEC_PAST_1970)
		   > srcType[0].eventTime.secPastEpoch) ) )
	{
	    trkPosition--;
	}
	trkPosition = trkPosition - Size();
    }
    else if ((filePosition == 49) && (srcType [0].type == TRKINFO))
    {
	if (trkPosition != 0)
	    trkPosition--;
    }
    else if ((filePosition == 51) && (srcType [0].type == TRKINFO)) 
	trkPosition++;
    else if (filePosition == 75)
    {
	while ( (TrkSeek (trkPosition+1) == RECFOUND) && 
		( ((unsigned int)TrkEventTime () - TS_EPOCH_SEC_PAST_1970)
		   < srcType[lastItem].eventTime.secPastEpoch))
	{
	    trkPosition++;
	}
    }
    else if (filePosition >= 90)
	trkPosition = TrkNumEntries() - Size();

    if (trkPosition > ((int)TrkNumEntries() - Size()))
	trkPosition = TrkNumEntries() - Size();

    if (trkPosition < 0)
	trkPosition = 0;
    trkIx = 0;

    // We have all the input, now identify the record types for each cell

    // use extra array cell to optimize loop for filePosition >= 90
    // (lets me load the element, then do a single shift up )
    for (i = 0; i < loopcnt + 1; i++)
    {
	if ( (TrkSeek (trkPosition + trkIx + 1) == RECFOUND) && 
	     ( (sysIx >= retloopcnt) || 
	       ( ((unsigned int)TrkEventTime () - TS_EPOCH_SEC_PAST_1970)
		  <= alarmSysRecords[sysIx].eventtime.secPastEpoch) ) )
	{
	    srcType [i].type = TRKINFO;
	    srcType [i].key = trkPosition + trkIx;
	    srcType [i].eventTime.secPastEpoch = TrkEventTime () - TS_EPOCH_SEC_PAST_1970;
	    srcType [i].eventTime.nsec = 0;
	    trkIx++;
	}
	/* insert additional data sources here */
	else if (sysIx < retloopcnt)
	{
	    srcType [i].type = LOGINFO;
	    srcType [i].key = sysIx;
	    srcType [i].eventTime = alarmSysRecords[sysIx].eventtime;
	    sysIx++;
	}
	else
	{
	    /* fill in with NOINFO */
	    srcType [i].type = NOINFO;
	    break;
	}
	if (i == loopcnt)
	{
	    if (filePosition == 0)
	    {
		break;
	    }
	    else if (filePosition == 25)
	    {
		// stop collecting when last item is equal to old first item
		if (srcType [loopcnt-1].type == checkType)
		{
		    if ((checkType == TRKINFO) && (checkKey == srcType [loopcnt-1].key))
			break;
		    else if ((checkType == LOGINFO)
		      && (strcmp (checkName, alarmSysRecords[srcType [loopcnt-1].key].pointName) == 0)
			&& (checkTime.secPastEpoch == alarmSysRecords[srcType [loopcnt-1].key].eventtime.secPastEpoch)
			&& (checkTime.nsec == alarmSysRecords[srcType [loopcnt-1].key].eventtime.nsec))
			break;
		}
	    }
	    else if (filePosition == 49)
	    {
		// stop collecting when second item is equal to old first item
		if (srcType [1].type == checkType)
		{
		    if ((checkType == TRKINFO) && (checkKey == srcType [1].key))
			break;
		    else if ((checkType == LOGINFO)
		      && (strcmp (checkName, alarmSysRecords[srcType [1].key].pointName) == 0)
			&& (checkTime.secPastEpoch == alarmSysRecords[srcType [1].key].eventtime.secPastEpoch)
			&& (checkTime.nsec == alarmSysRecords[srcType [1].key].eventtime.nsec))
			break;
		}
	    }
	    else if ((filePosition == 50)
		// stop collecting when first item is equal to old first item
	      || (filePosition == 51)
		// stop collecting when first item is equal to old second item
	      || (filePosition == 75))
	    {
		if (srcType [0].type == checkType)
		{
		    if ((checkType == TRKINFO) && (checkKey == srcType [0].key))
			break;
		    else if ((checkType == LOGINFO)
		      && (strcmp (checkName, alarmSysRecords[srcType [0].key].pointName) == 0)
			&& (checkTime.secPastEpoch == alarmSysRecords[srcType [0].key].eventtime.secPastEpoch)
			&& (checkTime.nsec == alarmSysRecords[srcType [0].key].eventtime.nsec))
			break;
		}
	    }
	    else if (filePosition >= 90)
	    {
		// go until we run out of items
	    }
	    // slide the records up, to show the last of the merged records
	    for (j = 0; j < loopcnt; j++)
	    {
		srcType [j] = srcType [j+1];
	    }
	    i--;
	}	    	
    }
    // if we are at end of loop, we have an extra element
    if (i==(loopcnt + 1))
	i--;
    for ( ; i < ALVU_MAXITEMS + 1; i++)
    {
	// fill out the rest with NOINFO 
	srcType [i].type = NOINFO;
    }
    StopCancelBox ();
    if (CheckCancelled ())
    {
	ShowCancelled ();
	inRefresh = False;
	return;
    }

    /* We have all the input, now display the info */
    /* force recognition of first log record displayed */
    topName [0] = 0;
    lastName [0] = 0;
    for (i = 0; i < loopcnt; i++)
    {
	
	if (srcType [i].type == TRKINFO
	  && (TrkSeek (srcType [i].key + 1) == RECFOUND))
	{
	    sprintf (timeBuf, "%s %s", TrkDate (), TrkTime ());
	    AddLine ( -1, timeBuf, TrkPointname (), AlarmFormat (&displaytrack, NULL, NULL));
	}
	// insert additional data sources here
	else if (srcType [i].type == LOGINFO)
	{
	    // reset topName to first displayed record
	    if (topName [0] == 0)
	    {
		strncpy (topName, alarmSysRecords[srcType [i].key].pointName, AL_POINTNAME_SIZE + 1);
		topTime = alarmSysRecords[srcType [i].key].eventtime;
	    }
	    strncpy (lastName, alarmSysRecords[retloopcnt-1].pointName, AL_POINTNAME_SIZE +1 );
	    lastTime = alarmSysRecords[retloopcnt-1].eventtime;
	    if(alarmSysRecords[srcType [i].key].comment[0] != 0)  {
		sprintf (timeBuf, "%s %s *",
			 alarmSysRecords[srcType [i].key].date,
			 alarmSysRecords[srcType [i].key].time);
	    }
	    else
	    {
		sprintf (timeBuf, "%s %s  ",
			 alarmSysRecords[srcType [i].key].date,
			 alarmSysRecords[srcType [i].key].time);
	    }
	    if (strcmp (alarmSysRecords[srcType [i].key].status, "ACK") == 0)
	    {
		sprintf (txtBuf, "%s",
			 AlarmFormat (&displayack, NULL, &alarmSysRecords[srcType [i].key]));
	    }
	    else
	    {
		sprintf (txtBuf, "%s",
			 AlarmFormat (&displayalarm, NULL, &alarmSysRecords[srcType [i].key]));
	    }
	    AddLine ( -1, "", alarmSysRecords[srcType [i].key].pointName,
			 txtBuf, &alarmSysRecords[srcType [i].key].eventtime );
	}
	else
	{
	    break;
	}
    }
    if (dataList != NULL)
    {
	alarmListView -> SetList (dataList);
	vlistSelection *selected = alarmListView -> GetSelection ();
	dataList -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	dataList -> SetAtLeastOneSelection (selected, vFALSE);
    }
    alarmListView -> Inval (vwindowINVAL_OPAQUE | vwindowINVAL_UPDATE);
    if ( ((srcType [0].type == LOGINFO)
	 && (strcmp (startName, topName) == 0) && (memcmp (&startTime, &topTime, sizeof(startTime)) == 0))
      || ((srcType [0].type == TRKINFO) && (trkPosition == 0)) )

    {
	filePosition = 0;
	if (srcType [loopcnt - 1].type == NOINFO)
	{
	    SetFileItems (i + 1);
	}
    }
    else if ((eof || (retloopcnt < loopcnt-1)) && (TrkSeek (trkPosition + trkIx + 1) != RECFOUND))
    {
	filePosition = 90;
    }
    else
    {
	filePosition = 50;
    }
    SetFilePos (filePosition);
    inRefresh = False;
}


/*******************************************************************
*
*    AlarmSystemDialog Resync function 
*
*******************************************************************/

void AlarmSystemDialog::Resync ( void)
{
    RefreshDisplay ();
}


/*******************************************************************
*
*    AlarmSystemDialog set title function to be written by subclasses
*
*******************************************************************/

void AlarmSystemDialog::SetDialogTitle (void)
{
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    vstr *format = vresourceGetString (app -> titlesResource, alvu_AlarmSystemTitle);
    vstr *filtGrp = vstrCloneScribed (vcharScribeSystem (filter.group));
    vstr *filtOp = vstrCloneScribed (vcharScribeSystem (filter.op));
    vstr *mod = vstrCloneScribed (vcharScribeLiteral (filter.modified ? "*" : ""));
    vstr *name = vstrCloneScribed (vcharScribeSystem (filter.name));
    title = vstrCloneScribed (vcharScribeFormat (format, filtGrp, filtOp, mod, name));
    SetTitle (title);
    vstrDestroy (format);
    vstrDestroy (filtGrp);
    vstrDestroy (filtOp);
    vstrDestroy (mod);
    vstrDestroy (name);
}


/*******************************************************************
*
*    AlarmSystemDialog Show Cancelled Msg's
*
*******************************************************************/

void AlarmSystemDialog::ShowCancelled (void)
{
    vstr *msg = vresourceGetString (app -> messagesResource,
					    alvu_RefreshCancelledMsg);
    elementCount -> SetTextScribed (vcharScribeSystem (msg));
    vstrDestroy (msg);
    StopCancelBox ();
    fileButtonContainer -> linedownButton -> Enable ();
    fileButtonContainer -> pagedownButton -> Enable ();
    fileButtonContainer -> lastButton -> Enable ();
}


/*******************************************************************
*
*    AlarmSystemDialog Windows Print function
*
*******************************************************************/

void AlarmSystemDialog::WindowsPrint (int whatToPrint)
{
    if (app -> printer == NULL)
    {
	alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, alvu_NoPrinterMsg);
	return;
    }
    return;
}


/*******************************************************************
*
*    AlarmPointsDialog Unix Print function
*
*******************************************************************/

void AlarmSystemDialog::UnixPrint (int whatToPrint)
{
    FILE    *printFileDesc;
    char    printFileName [30];

    strcpy (printFileName, PRINT_FILE_TEMP);
    if ( !mktemp (printFileName) || !(printFileDesc = fopen (printFileName, "w")))
    {
	vstr *prtFileTmp = vstrCloneScribed (vcharScribeLiteral (PRINT_FILE_TEMP));
	alvuMessageDialogCenter (this, ALVU_INFO_NOTICE, alvu_TempPrtFileErr, prtFileTmp);
	vstrDestroy (prtFileTmp);
    }
    else
    {
	if (whatToPrint == PRINT_THIS_WINDOW)
	{
	    int	    startLine = alarmListView -> GetTopRow ();
	    int	    endLine = alarmListView -> GetBottomRow ();
	    int	    i;

	    for (i = startLine; i <= endLine; i++)
	    {
		if (srcType [i].type == TRKINFO)
		{
		    if (TrkSeek (srcType [i].key) == RECFOUND)
		    {
			fprintf (printFileDesc, "%s\n",
					     AlarmFormat (&printtrack, NULL, NULL));
		    }
		}
		else if (srcType [i].type == LOGINFO)
		{
		    if (strcmp (alarmSysRecords[srcType [i].key].status, "ACK") == 0)
		    {
			fprintf (printFileDesc, "%s\n",
					     AlarmFormat (&printack, NULL,
							  &alarmSysRecords[srcType [i].key]));
		    }
		    else
		    {
			fprintf (printFileDesc, "%s\n",
					     AlarmFormat (&printalarm, NULL,
							  &alarmSysRecords[srcType [i].key].date));
		    }	
		}
	    }
	}
	else    // print entire list
	{
	    StartCancelBox ();
	    int	trkIx = 1;
	    int numLines = Log_RecordsByTime (startName, &this -> startTime, alarmSysRecords,
						ALVU_MAXITEMS, ALRMSYS_INDEX_FILE,
						(int (__cdecl *)(void))CheckCancelGW, this);
	    StopCancelBox ();
	    if (CheckCancelled ())
	    {
		ShowCancelled();
		fclose (printFileDesc);
#ifdef SOLARIS
		sprintf (txtBuf, "rm %s", printFileName);
		system (txtBuf);
#endif SOLARIS
		return;
	    }
	    int sysIx = 0;
	    while ((trkIx <= (int)TrkNumEntries ()) && (numLines > sysIx))
	    {
		if ((TrkSeek (trkIx) == RECFOUND) && 
		    ((sysIx >= numLines) ||
		     (((unsigned int)TrkEventTime () - TS_EPOCH_SEC_PAST_1970)
				<= alarmSysRecords[sysIx].eventtime.secPastEpoch)))
		{
		    if (TrkSeek (trkIx) == RECFOUND)
		    {
			fprintf (printFileDesc, "%s\n",
					     AlarmFormat (&printtrack, NULL, NULL));
			trkIx++;
		    }
		    // insert additional data sources here
		    else if (sysIx < numLines)
		    {
			fprintf (printFileDesc, "NOT IMPLEMENTED!!\n");
					   /* AlarmFormat (&printalarm, NULL,
						      &alarmSysRecords[srcType [i].key].date));*/
			sysIx++;
			if (sysIx == numLines)
			{
			    numLines = Log_RecordsByTime (alarmSysRecords[numLines-1].pointName,
					&alarmSysRecords[numLines-1].eventtime, alarmSysRecords,
					ALVU_MAXITEMS, ALRMSYS_INDEX_FILE,
					(int (__cdecl *)(void))CheckCancelGW, this);
			    sysIx = 1; // skip last record of previous group
			}
		    }
		}
	    }
	}

	fclose (printFileDesc);
	if (strlen (linePrinter))
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"System History\" %s |lpr -P\"%s\"", 
			      printFileName, linePrinter);
#else
	    sprintf (txtBuf, "lpr -p -r -P\"%s\" -T\"System History\" %s", 
			      linePrinter, printFileName);
#endif SOLARIS
	}
	else
	{
#ifdef SOLARIS
	    sprintf (txtBuf, "pr  -h \"System History\" %s |lpr", printFileName);
#else
	    sprintf (txtBuf, "lpr -p -r -T\"System History\" %s", printFileName);
#endif SOLARIS
	}
	system (txtBuf);
#ifdef SOLARIS
	sprintf (txtBuf, "rm %s", printFileName);
	system (txtBuf);
#endif SOLARIS

    }
}


vkindTYPED_FULL_DEFINITION(alarmSystemCommands, vcommandSelector, "alarmSystemCommands");
/*******************************************************************
*
*    menu selector commands for the alarm system dialog
*
*******************************************************************/

int alarmSystemCommands::Issue (vdict *dict)
{
    int			command;
    AlarmSystemDialog	*alarmSystemDialog;

    alarmSystemDialog = (AlarmSystemDialog *) dict -> GetElem (vname_Dialog);
    alvuApplication *app = alarmSystemDialog -> GetApp ();
    vresource alvuRes = app -> alvuResource;

    command = GetSelector ();

    switch (command)
    {
	case ALSYS_THISWIN_ITEM:
	case ALSYS_ENTIRESUM_ITEM:
	{
	    AldlgListView *view = alarmSystemDialog -> alarmListView;
	    vlist *list = view -> GetList ();
	    int rowCount = 0;
	    if (list != NULL)
	    {
		rowCount = list -> GetRowCount ();
	    }
	    if ((list == NULL) || (rowCount == 0))
	    {
		alvuMessageDialogCenter (alarmSystemDialog, ALVU_INFO_NOTICE,
					 alvu_NothingToPrint);
		break;
	    }
#ifdef WIN32
	    if (command == ALSYS_THISWIN_ITEM)
	    {
		alarmSystemDialog -> WindowsPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmSystemDialog -> WindowsPrint (PRINT_ENTIRE_LIST);
	    }
#else
	    if (command == ALSYS_THISWIN_ITEM)
	    {
		alarmSystemDialog -> UnixPrint (PRINT_THIS_WINDOW);
	    }
	    else
	    {
		alarmSystemDialog -> UnixPrint (PRINT_ENTIRE_LIST);
	    }
#endif WIN32
	    break;
	}

	case ALSYS_MODFILT_ITEM:
	{
	    vresource filtRes = vresourceGet (alvuRes, alvu_FilterDialog);
	    FilterDialog *filterDialog = new FilterDialog (filtRes);
	    filterDialog -> OpenFilter (app, NULL, &alarmSystemDialog -> filter, MODIFY_FILTER);
	    filterDialog -> NewFilter (&alarmSystemDialog -> filter);
	    delete filterDialog;
	    alarmSystemDialog -> SetDialogTitle ();
	    TrkSetFilter (&alarmSystemDialog -> filter);
	    Log_SetFilter (&alarmSystemDialog -> filter, ALRMSYS_INDEX_FILE);
	    alarmSystemDialog -> SetPosition (100);
	}

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}

