/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		CHANNEL ACQUISITION AND CONTROL			*/
/*								*/
/*	This program does all the work of creating channels	*/
/*	and sharing the data between applications.		*/
/*								*/
/*	Endian issues: We may be a different endian from the	*/
/*	name server, the IOC, and the end user application.	*/
/*	We will always be the same endian as Socket Creation	*/
/*	and Maintenance.					*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/*	Things to do:						*/
/*		Be more robust about losing our connection	*/
/*		with the name server.				*/
/*								*/
/*		Be more robust about losing our connection	*/
/*		with socket creation and maintenance.		*/
/*								*/
/****************************************************************/

#if !(defined(AXP) || defined(vxWorks))
#include <bcopy.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if !(defined(vms) || defined(AXP) || defined(vxWorks) || defined(WIN32))
#include <prof.h>
#else
#define MARK(x) ;
#endif vms

#ifndef WIN32
#include <types.h>
#include <socket.h>

#define SOCKET	int
#define INVALID_SOCKET	-1

#ifdef vxWorks
#include <vxWorks.h>
#include <ioLib.h>
#include <sys/times.h>
#include <tickLib.h>
#include <taskLib.h>
#include <memLib.h>
#include <selectLib.h>
#include <time.h>
#include <sockLib.h>
#include <netinet/tcp.h>
#include <inetLib.h>
#include <hostLib.h>
#include <logLib.h>
#else vxWorks
#ifndef vms
#include <syslog.h>
#endif vms
#include <stdio.h>
#if (defined(ultrix) || defined(AXP))
#include <ioctl.h>
#else ultrix or AXP
#include <filio.h>
#endif ultrix
#ifndef AXP
#include <time.h>
#include <sys/time.h>
#endif AXP
#include <sys/types.h>
#include <tcp.h>
#include <netdb.h>
#endif vxWorks

#include <errno.h>
#include <in.h>
#include <fcntl.h>
#else WIN32
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <winsock.h>
#include "NTUnixSubs.h"
#define MARK(x) ;
#define EWOULDBLOCK WSAEWOULDBLOCK
#ifndef caddr_t
typedef char *caddr_t;
#endif caddr_t
#endif WIN32

#include "inet_ntoa.h"
#include "dbDefs.h"
#include "pathNames.h"
#include "agentPorts.h"
#include "agentMsgs.h"
#include "channelControl.h"
#include "dataTypes.h"

#ifndef vxWorks
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include "link.h"
#include "descDataTable.h"  /* required for fieldAttributes.h */
#include "fieldAttributes.h"
#include "msgMon.h"
#include "choice.h"
#include "db_access.h"
#include "caValues.h"
#include "currValues.h"
#include "choiceGbl.h"
#include "cvtMgt.h"
#endif vxWorks

#ifdef vxWorks
#include "dbCommon.h"
#include "dbAccess.h"
#endif vxWorks
#include "transformType.h"
#include "dumpBlocks.h"
#include "cadef.h"
#include "tisByteOrder.h"


#ifdef PRIV_AUDIT
#include "inet_ntoa.h"
#ifndef vxWorks
#define	ERROR	-1
#endif vxWorks
#include "sslShared.h"

/* for the auditing and priviledge functions */
#include "iocPriv.h"
#include "auditSystem.h"
extern AUDIT_ID	auditLog;	/* the system log */

#endif PRIV_AUDIT

/****************************************************************/
/*								*/
/*		TUNING AND SIZING PARAMETERS			*/
/*								*/
/****************************************************************/

#define	HASH_SIZE	(1024)
#define SCM_BUFF_SIZE	(4096)

/****************************************************************/
/*								*/
/*		GLOBAL DATA THIS PROCESS			*/
/*								*/
/****************************************************************/

/*
Every time we run down all of the CABs on a CHB, we first build
a list of them and then run down the list. Each list is a chuck
of 32 addresses and then a pointer to the next chunk of 32. We
do this to make sure that faulty linkage won't send us of into
space and then crash. This logic was useful during the debugging
phase and was left in for protection.
*/

struct cabList
    {
    struct cabList	*nextList;
    CAB			*refCAB[32];
    } firstCabList;
struct cabList	*currCabList;
int		numCABs;
int		cabIndex;

static int	maxFd;			/* Highest fd for select */

#ifndef vxWorks
#ifndef vms
int		scmPid;			/* sockMnt's process ID */
FILE		*traceFd = 0;		/* For debugging */
#endif vms
#endif vxWorks

SOCKET		namFd = INVALID_SOCKET;			/* Socket to/from name server */
int		namRxSize;		/* Receive buffer size for namFd */
int		namTxSize;		/* Transmit buffer size of namFd */
int		namRxTopped;		/* Maxout flag for namFd */

static short	numTrendHandlers = 0;	/* Number of trend handlers supported */
static short	numAlarmHandlers = 0;	/* Number of alarm handlers supported */
#ifndef vxWorks
char		cacMobyBuff[512*1024];	/* Mega buffer for data -> sockMnt */
int		cacRingIn = 0;		/* Input index for ring buffer */
int		cacRingOut = 0;		/* Output index for ring buffer */
int		cacExtBytes = 0;	/* # bytes in mega buffer */
fd_set		scmMask;		/* Only contains scmFd */
fd_set		writeFds;		/* Used when data in mega buffer */
fd_set		*writeReady;		/* Points at writeFds or NULL */
int		caMustResolve = 1;	/* All host platforms must resolve names */
#else
int		caMustResolve = 0;	/* vxWorks doesn't have to resolve (default) */
extern struct in_addr caLocalInet;	/* Our nodes Internet address */
#endif vxWorks

int		caNameSrvTimeout = 15;	/* Name server timeout is 15 seconds */
int		anyTcpOutputs;		/* Anything sent to an IOC? */

SOCKET		caqTcpFd = INVALID_SOCKET; /* Our server socket */

SOCKET		scmFd = INVALID_SOCKET;	/* Fd for socket creation/maintenance */
int		scmRxSize;		/* Receive buffer size for scmFd */
int		scmTxSize;		/* Transmit buffer size for scmFd */
int		scmRxTopped;		/* Receive buffer maxout flag */
int		scmTxTopped;

fd_set		cacfdMask;		/* Overall fd mask for select() */
fd_set		fdList;			/* Temporary copy of cacfdMask */

ISB		*isbList = 0;		/* Chain of Inet Status Blocks */
ACB		*acbList = 0;		/* Chain of Application Control BLocks */
RCB		*rcbList = 0;		/* Chain of Resource Control Blocks */
NSQ		*nsqList = 0;		/* Name Server Queue */
HTB		*pHash;			/* Hash Table Address */

#ifdef vxWorks
extern struct timeval oneSecond;
extern struct timeval pollTimeout;
#else
struct timeval	oneSecond = {5,0};	/* Used for timing out I/O */
struct timeval	pollTimeout = {0, 0};	/* For a "poll" call to select */
#endif vxWorks

int		session;		/* Our session handle */
int		cacDebug = 0;		/* Are we debugging? */

int		scmOutputBytes = 0;	/* # bytes enqueued to Socket Controller */
char		scmBuff[SCM_BUFF_SIZE];	/* Xmit buffer to Socket Controller */
static char	rxBuff[SCM_BUFF_SIZE];	/* Receive buffer for all sockets */
static char	nameRxBuff[2048];	/* Rx buffer for name server */
static char	nameTxBuff[1024];	/* Tx buffer for name server */
static int	namNchib;		/* # bytes in nameRxBuff */
static int	namTxBytes;		/* # bytes in nameTxBuff */
static int	namBufOutPos;		/* Next byte to be used in nameRxBuff */
static int	nchib;			/* Number bytes active in rxBuff */
static int	unregistered = 1;	/* Don't have sockMnt yet */
int		bufOutPos;		/* Current output index in rxBuff */

extern int	errno;			/* Unix error code for failed operations */
int		true = 1;
int		optval = 1;
int		intsize = sizeof(true);

#if !defined(vxWorks) && !defined(vms)
#define	ATTR_NUM_STRS	64		/* Only 64 strings per enumeration */
/* 
   The following define is the same as FA_STRING_SIZE in fieldAttributes.h
   Any change here should be changed there as well. OR ELSE!!!!!
*/
#define ATTR_STR_LEN	32		/* Each string 31 bytes plus trailing null */
char	attrStrings[ATTR_NUM_STRS][ATTR_STR_LEN];
short	attrNativeType;			/* Native data type for field */
short	attrNativeCount;		/* Native array count */
short	attrPrec;			/* Display precision */
short	attrNumStrings;			/* For enumerations */
double	attrHopr;			/* High end of display range */
double	attrLopr;			/* Low end of display range */
double	attrDrvh;			/* Upper control limit */
double	attrDrvl;			/* Lower control limit */

short	attrMap[] =			/* Maps DBR types to AGNT types */
    {
    AGNT_STRING,
    AGNT_CHAR,
    AGNT_CHAR,
    AGNT_SHORT,
    AGNT_SHORT,
    AGNT_LONG,
    AGNT_LONG,
    AGNT_FLOAT,
    AGNT_DOUBLE,
    AGNT_ENUM
    };
#endif vxWorks and vms

/****************************************************************/
/*								*/
/*			FUNCTION PROTOTYPES			*/
/*								*/
/****************************************************************/

int appTx (ACB *pACB, short msgType, char *buffer, int nbytes);
int appFlush (ACB *pACB);
int appRx (ACB *pACB, char *buffer, int nbytes);
int sockRead (char *buff, int nbytes);
int haveCompleteNameMsg (void);
int chGetMore (void);
void delApp (ACB *pACB);
void sendScm (struct in_addr inet, short msgType, char *buff, int nbytes);
int scmFlush (void);
int caSendNam (char *buff, int nbytes);
int caNameFlush (char *buff, int nbytes);
void nameFix (char *chanName);
int nameFind (char *name, CHB **ppCHB, short type, int spliceMask);
ISB *findISB (struct in_addr inetAddr);
RCB *findRCB (char *name);
int hash (char *name);
int updateDataType (CHB	*pCHB);
void checkChannel (CHB *pCHB);
void updateConn (CHB *pCHB, ISB *pISB, int iocHandle, int typeChanged);
void unchainCHB (CHB *pCHB);
void adjustType (CAB *pCAB);
void rejectChan (ISB *pISB, int	handle);
int haveCompleteSockMsg (void);
int haveCompleteAppMsg (ACB *pACB);
void scmRead (char *buff, int nbytes);
int fillSockBuff (void);
int fillAppBuff (ACB *pACB);
int findNumCABs (CHB *pCHB);
void closeNameSocket (void);
NSQ *sameName (char *targetName);
int getSpliceMask (char *name);
int isValField (CHB *pCHB);
#ifndef vxWorks
static int sendMoby (void);
#endif vxWorks

/****************************************************************/
/*								*/
/*			THE MAIN PROGRAM			*/
/*								*/
/****************************************************************/

#ifdef vxWorks
int chanAcq()
#else
int main()
#endif vxWorks

{
#if defined(vxWorks) || defined(vms)
    struct timeval	nameSrvTimeout;
#endif vxWorks or vms

    struct sockaddr_in	portAddr;
    int			addrSize;
    RCB			*currRCB;
    ACB			*currACB;
    ACB			*nextACB;
    CAB			*pCAB;
    CHB			*pCHB;
    CHB			*nextCHB;
#ifndef vxWorks
    struct hostent	*hostent;
    ISB			*currISB;
    NS_IOCL_REPLY	inetListHdr;
    NS_RSRQ_REPLY	resListHdr;
#endif vxWorks
    int			count;
    short		msgType;
    ISB			*pISB;

#ifdef WIN32
    WSADATA wsaData; 
    WORD    wVersionRequested = MAKEWORD(1, 1);
    int	    err;

    err = WSAStartup(wVersionRequested, &wsaData); 
 
    if (err != 0) 
	return 0; 
#endif WIN32

/******
	Create our session handle. This lets us knows if an IOC's handles
	are valid when we connect. If they aren't, we tell him to drop
	them all and start over.
******/

#ifdef vxWorks
    session = tickGet();
#else
    session = time(NULL);
#endif vxWorks

/******
	Initialize
******/

#ifndef vxWorks
#ifndef vms
    if (getenv("chanTrace"))
#ifdef WIN32
	traceFd = fopen("\\tis4000db\\chanTrace.log", "w");
#else
	traceFd = fopen("/tis4000db/chanTrace.log", "w");
#endif WIN32
    serverInit(CAPATH, CALOCK);

#ifdef PRIV_AUDIT
    /* open default audit trail file */
    if (InitSystemLog (NULL, 0) == NULL)
    {
#if !defined(vxWorks) && !defined(vms)
	tisSysLog(0, 0, "sockMnt can't open default system audit file %s!\nNo auditing will occur\n", systemLogFile);
#else
	perror("sockMnt can't open system audit file %s!\nNo auditing will occur\n", systemLogFile);
#endif vxWorks
    }
#endif PRIV_AUDIT

    tisOpenLog("chanAcq", 6, 2, NULL, NULL);

	/* 
	Message levels:
		0 = fatal errors
		1 = Node Up/Down
		    App attach/detach
		    App register
		2 = App open, close, monadd, monrem
		    IOC blkack, closeack, singmon
		3 = App read, write
		4 = Alarms
		5 = Trend Data
		6 = Monitor Events
	*/

    /*
    Initialize the database for point attributes.
    */

    if (FA_InitAttributes())
	{
	tisSysLog(0, 0, "FA_InitAttributes failed.\n");
	exit(1);
	}
    tisSysLog(0, 0, "Database initialization complete.\n");

#endif vms
#endif vxWorks

/******
	Allocate the hash table. Its purpose is to make sure that we can find
	a CHB by name rapidly.
******/

    pHash = (HTB *) calloc(HASH_SIZE, sizeof(HTB));

/******
	Connect to the name server for Unix platforms. Be persistant about it.
******/

    FD_ZERO(&cacfdMask);
#ifndef vxWorks
#ifndef vms
    namFd = socket(PF_INET, SOCK_STREAM, 0);
    if (namFd < 0) 
	{
#ifdef WIN32
	tisSysLog(0, 0, "chanAcq can't create TCP socket to nameSrv: %d", WSAGetLastError());
#else
	tisSysLog(0, 0, "chanAcq can't create TCP socket to nameSrv: %m");
#endif WIN32
	exit(1);
	}
    hostent = gethostbyname("localhost");
    portAddr.sin_family = AF_INET;
    portAddr.sin_addr = *((struct in_addr *)hostent -> h_addr_list[0]);
    portAddr.sin_port = htons(NAM_PORT);

    for (;;)
	{
	if (connect(namFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) != -1)
	    break;
#ifdef WIN32
	Sleep (2000);
	closesocket(namFd);
#else
	sleep(2);
	close(namFd);
#endif WIN32
    	namFd = socket(PF_INET, SOCK_STREAM, 0);
    	if (namFd < 0) 
	    {
	    tisSysLog(0, 0, "chanAcq can't create TCP socket to nameSrv: %m");
	    exit(1);
	    }
	}
#ifndef SOLARIS
#ifdef WIN32
    ioctlsocket(namFd, FIONBIO, &true);
#else
    ioctl(namFd, FIONBIO, &true);
#endif WIN32
#else
    sockGrow(namFd);
    fcntl(namFd, F_SETFL, O_NONBLOCK);
#endif SOLARIS
    setsockopt(namFd, IPPROTO_TCP, TCP_NODELAY, (char *)&true, sizeof(true));
    getsockopt(namFd, SOL_SOCKET, SO_RCVBUF, (char *)&namRxSize, &intsize);
    getsockopt(namFd, SOL_SOCKET, SO_SNDBUF, (char *)&namTxSize, &intsize);
    namRxTopped = 0;
    maxFd = namFd;
    FD_ZERO(&cacfdMask);
    FD_SET(namFd, &cacfdMask);
#endif vms
#endif vxWorks

/******
	Get the list of Internet addresses out of the name server. Allocate the list
	in one chunk and set up the Inet address in each ISB. Assume that the node
	is down until we are told differently by Socket Creation/Maintenance.
******/

#if defined(vxWorks) || defined(vms)
    isbList = 0;
    rcbList = 0;
#else
    msgType = htons(NAME_IOCL);
    if (send(namFd, (char *) &msgType, sizeof(msgType), 0) == -1)
	{
	tisSysLog(0, 0, "chanAcq send error to nameSrv for IOCL: %m");
	exit(1);		/* Be more robust than this in the future! */
	}
    if (!sockRead((char *) &msgType, sizeof(msgType)))
	{
	tisSysLog(0, 0, "chanAcq read error (msgType) from nameSrv: %m");
	exit(1);		/* Be more robust than this in the future! */
	}
    if (!sockRead((char *) &inetListHdr, sizeof(inetListHdr)))
	{
	tisSysLog(0, 0, "chanAcq read error (inetListHdr) from nameSrv: %m");
	exit(1);		/* Be more robust than this in the future! */
	}
    count = ntohs(inetListHdr.nsNumNodes);
    isbList = (ISB *) calloc(count, sizeof(ISB));
    currISB = isbList;
    while (count--)
	{
	if (!sockRead((char *) &(currISB -> Inet), sizeof(currISB -> Inet)))
	    {
	    tisSysLog(0, 0, "chanAcq read error (ISB -> Inet) from nameSrv: %m");
	    exit(1);		/* Be more robust than this in the future! */
	    }
#ifndef vxWorks
	/*
	On workstations, if this is NOT a DPA then set splicable.
	DPAs are Inet addresses 1 -> 255, 0 is illegal.
	*/

	if (ntohl (currISB -> Inet.s_addr) >= 255)
	    currISB -> splicable = 1;
#endif
#ifdef POINTER64
	currISB -> isbToken = TokenFromPointer64(currISB);
#endif
	if (count)
		currISB -> nextISB = (struct IOCSB *) (currISB + 1);
	currISB++;
	}

    /******
	    Get the list of resources out of the name server. This is very similar to
	    handling the list of Internet addresses.
    ******/

	msgType = htons(NAME_RSRQ);
	if (send(namFd, (char *) &msgType, sizeof(msgType), 0) == -1)
	    {
	    tisSysLog(0, 0, "chanAcq send error to nameSrv for RSRQ: %m");
	    exit(1);		/* Be more robust than this in the future! */
	    }
	if (!sockRead((char *) &msgType, sizeof(msgType)))
	    {
	    tisSysLog(0, 0, "chanAcq read error (msgType) from nameSrv: %m");
	    exit(1);		/* Be more robust than this in the future! */
	    }
	if (!sockRead((char *) &resListHdr, sizeof(resListHdr)))
	    {
	    tisSysLog(0, 0, "chanAcq read error (resListHdr) from nameSrv: %m");
	    exit(1);		/* Be more robust than this in the future! */
	    }
	count = ntohs(resListHdr.nsNumNames);
	rcbList = (RCB *) calloc(count, sizeof(RCB));
	currRCB = rcbList;
	addrSize = ntohs(resListHdr.nsMaxNameLen);
	while (count--)
	    {
	    struct in_addr inetAddr;

	    currRCB -> resName = (char *) malloc(addrSize);
	    if (!sockRead(currRCB -> resName, addrSize))
		{
		tisSysLog(0, 0, "chanAcq read error (resName) from nameSrv: %m");
		exit(1);	/* Be more robust than this in the future! */
		}
	    if (!sockRead((char *) &inetAddr, sizeof(inetAddr)))
		{
		tisSysLog(0, 0, "chanAcq read error (res Inet1) from nameSrv: %m");
		exit(1);	/* Be more robust than this in the future! */
		}
	    if (inetAddr.s_addr)
		{
		currRCB -> pISB1 = findISB(inetAddr);
		}
	    if (!sockRead((char *) &inetAddr, sizeof(inetAddr)))
		{
		tisSysLog(0, 0, "chanAcq read error (res Inet2) from nameSrv: %m");
		exit(1);	/* Be more robust than this in the future! */
		}
	    if (inetAddr.s_addr)
		{
		currRCB -> pISB2 = findISB(inetAddr);
		}
	    if (count)
		currRCB -> pNextRCB = (struct RECB *) (currRCB + 1);
	    currRCB++;
	    }
#endif vxWorks or vms

    /******
	    Create a socket so that applications can connect with us.
    ******/

	caqTcpFd = socket(PF_INET, SOCK_STREAM, 0);
	setsockopt(caqTcpFd, SOL_SOCKET, SO_REUSEADDR, (caddr_t) &optval, sizeof(optval));
	portAddr.sin_addr.s_addr = INADDR_ANY;
	portAddr.sin_port = htons(CAM_PORT);
	if (bind(caqTcpFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1)
	    {

#ifndef vxWorks
#ifndef vms
	    tisSysLog(0, 0, "chanAcq can't bind TCP port: %m");
#else
	    perror("chanAcq can't bind to TCP port");
#endif vms
	    perror("chanAcq can't bind to TCP port");
#endif vxWorks

#ifdef vxWorks
	    taskSuspend(taskIdSelf());	/* For IOC debugging help */
#else
	    exit(1);
#endif vxWorks
	    }
	listen(caqTcpFd, 2);
	if (caqTcpFd > maxFd)
	    maxFd = caqTcpFd;
	FD_SET(caqTcpFd, &cacfdMask);

    /******
	    Now wait for something to happen. This main loop looks a little strange, but there is a reason
	    for it. The outermost loop makes a volatile copy of the current FD mask and waits forever. The
	    inner loop then drains all of the sockets that are currently active. After all sockets have
	    been drained, then we tell Socket Creation and Maintenance to do a flush to keep the system
	    responsive. We can't rely on the user's flush because our timing and sequencing is not
	    necessarily in lock step with his.
    ******/

#if defined(vxWorks) || defined(vms)
	nameSrvTimeout.tv_sec = caNameSrvTimeout;
	nameSrvTimeout.tv_usec = 0;
#endif vxWorks or vms

	MARK(SOCKRD);
	for (;;)
	    {
	    bcopy((char *) &cacfdMask, (char *) &fdList, sizeof(fdList));

	    /*
	    In vxWorks and vms, we have to make sure that we haven't lost the connection to nameSrv
	    or that he hasn't died on us. This variant of the select() call includes a timeout for
	    the name server if there is anything in his queue.
	    */

#if defined(vxWorks) || defined(vms)
	    if ((namFd != INVALID_SOCKET) && nsqList)
		{
		count = select(maxFd +1, &fdList, NULL, NULL, &nameSrvTimeout);
		if (count == 0 || time(NULL) > nsqList -> dueTime)
		    {
		    closeNameSocket();
		    continue;
		    }
		}
	    else
#endif vxWorks or vms

#ifdef vxWorks
	    	select(maxFd + 1, &fdList, NULL, NULL, NULL);
#else
		/*
		Check to see if there's data in the moby buffer for sockMnt. If there is, then
		we have to include the write fd in the mask presented to select so we can
		dump the data out when there's buffering available. This happens on all
		platforms except for vxWorks
		*/

		{
		if (cacExtBytes)
		    {
		    writeReady = &writeFds;
		    bcopy((char *) &scmMask, (char *) &writeFds, sizeof(writeFds));
		    }
		else
		    writeReady = NULL;
		select(maxFd + 1, &fdList, writeReady, NULL, NULL);
		if (writeReady)
		    if ((scmFd != INVALID_SOCKET) && FD_ISSET(scmFd, &writeFds))
			sendMoby();
		}
#endif vxWorks

	    anyTcpOutputs = 0;
	    for (;;)
		{
		bcopy((char *) &cacfdMask, (char *) &fdList, sizeof(fdList));

		/*
		All platforms except for vxWorks must check write buffer availability
		for data stacked in the moby buffer, if any.
		*/

#ifdef vxWorks
	        if (select(maxFd + 1, &fdList, NULL, NULL, &pollTimeout) <= 0)
#else
		if (cacExtBytes)
		    {
		    writeReady = &writeFds;
		    bcopy((char *) &scmMask, (char *) &writeFds, sizeof(writeFds));
		    }
		else
		    writeReady = NULL;
		if (select(maxFd + 1, &fdList, writeReady, NULL, &pollTimeout) <= 0)
#endif vxWorks

		    break;

#ifndef vxWorks
		if (writeReady)
		    if ((scmFd != INVALID_SOCKET) && FD_ISSET(scmFd, &writeFds))
			sendMoby();
#endif vxWorks

		/*
		See if a new user is trying to connect with us.
		*/

		if ((caqTcpFd != INVALID_SOCKET) && FD_ISSET(caqTcpFd, &fdList))
		    {
		    SOCKET	newFd;
		    addrSize = sizeof(portAddr);
		    newFd = accept(caqTcpFd, (struct sockaddr *) &portAddr, &addrSize);
#if !defined(vxWorks) && !defined(vms)
		    tisSysLog(1, 0, "chanAcq adding app on fd %d", newFd);
#else
		    if (cacDebug >= 1)
			printf("chanAcq adding app on fd %d\n", newFd);
#endif
		    if (newFd < 0)
			continue;
#ifdef PRIV_AUDIT
		    /* ONLY allow local connects! */
		    if ((portAddr.sin_addr.s_addr != caLocalInet.s_addr)
		      && (portAddr.sin_addr.s_addr != htonl(0x7f000001)))
		    {
			close (newFd);
#ifdef vxWorks
			logMsg("TCP access from non-local IP - %s\n",
				(int)INET_NTOA(portAddr.sin_addr),0,0,0,0,0);
#else
			tisSysLog (0, 0, "TCP access from non-local IP - %s\n",
				INET_NTOA(portAddr.sin_addr));
#endif vxWorks
			SecurityAlarm (INET_NTOA(portAddr.sin_addr), "TCP access from non-local IP - %s\n", NULL);
			if (auditLog)
			{
#ifdef vxWorks
			    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, INET_NTOA(portAddr.sin_addr),
				    taskName(taskIdSelf()), "TCP access from non-local IP - %s\n");
#else
			    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, INET_NTOA(portAddr.sin_addr),
				    "sockMnt", "TCP access from non-local IP - %s\n");
#endif vxWorks
			}
			continue;
		    }
#endif PRIV_AUDIT

		    setsockopt(newFd, IPPROTO_TCP, TCP_NODELAY, (char *)&true, sizeof(true));
		    setsockopt(newFd, SOL_SOCKET, SO_LINGER, (caddr_t) &true, sizeof(true));
		    if (newFd > maxFd)
			maxFd = newFd;
		    FD_SET(newFd, &cacfdMask);

		    /*
		    Allocate an ACB for this new socket. Put it on the front of the chain. It is possible that
		    this connection might be for Socket Creation and Maintenance, who doesn't need an ACB.
		    If that's true, then we'll delete this ACB later.
		    */

		    currACB = newACB();
		    currACB -> pNextACB = acbList;
		    acbList = currACB;
		    currACB -> sockFd = newFd;
#ifndef SOLARIS
#ifdef WIN32
		    ioctlsocket(newFd, FIONBIO, &true);
#else
		    ioctl(newFd, FIONBIO, (int)&true);
#endif WIN32
#else
		    sockGrow(newFd);
		    fcntl(newFd, F_SETFL, O_NONBLOCK);
#endif SOLARIS
		    getsockopt(newFd, SOL_SOCKET, SO_RCVBUF, (char *)&(currACB -> appRxSize), &intsize);
		    getsockopt(newFd, SOL_SOCKET, SO_SNDBUF, (char *)&(currACB -> appTxSize), &intsize);
		    }

		/*
		See if this is a message from Socket Creation and Maintenance.
		All messages from IOCs are processed here.
		*/

		else if ((scmFd != INVALID_SOCKET) && FD_ISSET(scmFd, &fdList))
		  {

		  /*
		  If the receive buffer is full, then expand it so we don't hang up sockMnt in deadly embrace.
		  */

		  if (!scmRxTopped)
		      {
		      int	bytesPending;

#ifdef WIN32
		      ioctlsocket(scmFd, FIONREAD, &bytesPending);
#else
		      ioctl(scmFd, FIONREAD, (int)&bytesPending);
#endif WIN32
		      if (bytesPending == scmRxSize)
			{
			scmRxSize += 1024;
			if (setsockopt(scmFd, SOL_SOCKET, SO_SNDBUF, (char *)&scmRxSize, sizeof(scmRxSize)) == -1)
			    scmRxTopped = 1;
			}
		      }

		  /*
		  Read as much as we can out of the socket and then parse the messages while there is
		  a complete one in the input buffer. We can't wait for additional data to arrive
		  in the middle of a transaction because that penalizes the applications for slowness
		  from sockMnt and that's unfair.
		  */

		  if (!fillSockBuff())
		    {

#ifndef vxWorks
#ifndef vms
		    tisSysLog(0, 0, "chanAcq error while filling buffer from sockMnt: %m");
#ifndef WIN32
		    kill(scmPid, 3);
#endif WIN32
		    abort();
#endif vms
#endif vxWorks

#ifdef vxWorks
		    taskSuspend(taskIdSelf());	/* For IOC debugging help */
#else
		    exit(1);			/* Be more robust than this in the future! */
#endif vxWorks
		    }

		  while (haveCompleteSockMsg())
		    {

		    /*
		    Get the message type. Don't flip bytes because socket creation and maintenance
		    must always be in the same node that we are!
		    */

		    scmRead((char *) &msgType, sizeof(msgType));
		    if (msgType == CHAN_IOCDN)
			{
			CH_IOCDN	iocDown;
			CH_DISCEVNT	discEvnt1;
			ISB		*partners[10];	/* Only 10 possible due to a restriction in resMgr */
			int		numPartners;
			int		index;
			ISB		*partnerISB;

			/*
			IOC down. We have to create a disconnect event for each channel he had.
			Start by getting the Inet address of the IOC that was lost.
			*/

			MARK(IOCDN);
			scmRead((char *) &iocDown, sizeof(iocDown));
#if !defined(vxWorks) && !defined(vms)
			tisSysLog(0, 0, "chanAcq: IOC down: %s", INET_NTOA(iocDown.chInet));
#else
			printf("chanAcq: IOC down: %s\n", INET_NTOA(iocDown.chInet));
#endif
			pISB = findISB(iocDown.chInet);
			pISB -> connected = 0;	/* Don't send him anything until he's back online. */

			/*
			Traverse the list of CHBs that are linked to this ISB. If the IOC that went
			down is the one that had control of the point, generate a disconnect event
			for the user. Ignore CHBs that are delete pending. Also, keep track of the
			number of partners that this IOC had and their Inet addresses.
			*/

			pCHB = (CHB *) (pISB -> pFirstCHB);
			msgType = htons(CHAN_DISCEVNT);
			while (pCHB)
			    {

			    /*
			    If we had a channel build pending, it will have to start over
			    when the IOC comes back on line.
			    */

			    if (pCHB -> pISB1 == pISB)
				{
				nextCHB = pCHB -> pNextIOC1;

				if (pCHB -> handleIOC1 == -1)
				    pCHB -> handleIOC1 = 0;
				}
			    else
				{
				nextCHB = pCHB -> pNextIOC2;

				if (pCHB -> handleIOC2 == -1)
				    pCHB -> handleIOC2 = 0;
				}

			    if (!(pCHB -> pFirstCAB))
				{

				/*
				CHB is delete pending
				*/

				pCHB = nextCHB;
				continue;
				}

			    if (pCHB -> pCurrISB == pISB)
				{

				/*
				Now create a channel disconnect event for each CAB that is attached to
				the CHB.
				*/

				pCHB -> pCurrISB = 0;
				numCABs = findNumCABs(pCHB);
				currCabList = &firstCabList;
				for (cabIndex = 0; cabIndex < numCABs;)
				    {
				    pCAB = currCabList -> refCAB[cabIndex & 0x1f];
				    if (pCAB -> pCHB)
					{
				        discEvnt1.chHandle = pCAB -> handle;
				        if (!(pCAB -> awaitAck))
					    {
				            pCAB -> awaitAck = 1;
				            if (!appTx(pCAB -> pACB, msgType, (char *) &discEvnt1, sizeof(discEvnt1)))
						break;
					    }
					}
				    cabIndex++;
				    if ((cabIndex & 0x1f) == 0)
					currCabList = currCabList -> nextList;
				    }
				}

			    pCHB = nextCHB;
			    }

			/*
			After having traversed all the channels, we must ask the partner IOCs to report
			their current resource states. The reason for this, found during field test, is 
			that if we ever get two masters and the conflict is resolved by killing one of
			the IOCs, then the remaining machine will never report his state because there
			was no transition. If the channels were on the machine that was killed, then
			they will disconnect and never reconnect unless we ask the redundant partners
			to report their states.
			Also, while running down the resource list, report a disconnect to the trend 
			and alarm handlers for each resource that he owned.
			*/

			numPartners = 0;
			currRCB = rcbList;
			while (currRCB)
			    {
			    partnerISB = 0;
			    if (currRCB -> pISB1 == pISB)
				{
				if (currRCB -> pISB2)
				    partnerISB = currRCB -> pISB2;
				}
			    else if (currRCB -> pISB2 == pISB)
				partnerISB = currRCB -> pISB1;
			    if (partnerISB)
				{
				for (index = 0; index < numPartners; index++)
				    if (partners[index] == partnerISB)
					break;
				if (index >= numPartners)
				    {
				    partners[numPartners++] = partnerISB;
				    if (partnerISB -> connected)
					{
					msgType = htons(CHAN_RESINQ);
					sendScm(partnerISB -> Inet, msgType, NULL, 0);
					}
				    }
				}
			
			    if (currRCB -> currISB == pISB)
				{
				currRCB -> currISB = 0;
				currACB = acbList;
				while (currACB)
				    {
				    CH_RESMODE		resMode2;

				    nextACB = currACB -> pNextACB;
				    if (currACB -> trendHandler || currACB -> alarmHandler)
					{
					msgType = htons(CHAN_RESMODE);
					strcpy(resMode2.chResName, currRCB -> resName);
					resMode2.chResStatus = 0;
					appTx(currACB, msgType, (char *)&resMode2, sizeof(resMode2));
					}
				    if (currACB == nextACB)
					break;
				    currACB = nextACB;
				    }
				}
			    currRCB = currRCB -> pNextRCB;
			    }
			}

#if defined(vxWorks) || defined(vms)
		    else if (msgType == CHAN_GOTNMSRV)
			{
			int		intsize;
			CH_GOTNMSRV	gotNmsrv;
			NSQ		*nsqResend;
			short		msgType;
			struct
			    {
			    NS_SRCH	nsSrch;
			    char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];
			    } resendMsg;
			
			/*
			sockMnt found a name server for us. Connect to him.
			*/

			scmRead((char *) &gotNmsrv, sizeof(gotNmsrv));
			portAddr.sin_family = AF_INET;
			portAddr.sin_addr = gotNmsrv.chInet;
			portAddr.sin_port = htons(NAM_PORT);
			namFd = socket(PF_INET, SOCK_STREAM, 0);
			if (namFd < 0)
			    {
			    perror("chanAcq can't create socket to nameSrv");
#ifdef vxWorks
			    taskSuspend(taskIdSelf());	/* For IOC debugging help */
#else
			    exit(1);
#endif vxWorks
			    }
			if (connect(namFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1)
			    {
			    closeNameSocket();
			    continue;
			    }
			setsockopt(namFd, IPPROTO_TCP, TCP_NODELAY, (char *)&true, sizeof(true));
			intsize = sizeof(int);
			getsockopt(namFd, SOL_SOCKET, SO_RCVBUF, (char *)&namRxSize, &intsize);
			intsize = sizeof(int);
			getsockopt(namFd, SOL_SOCKET, SO_SNDBUF, (char *)&namTxSize, &intsize);
			namRxTopped = 0;
			if (namFd > maxFd)
			    maxFd = namFd;
			FD_SET(namFd, &cacfdMask);

			/*
			Name server now connected. Resend all of the stuff in our queue.
			*/

			if (cacDebug >= 1)
			    printf("Name Server resend!\n");

			for (nsqResend = nsqList; nsqResend; nsqResend = nsqResend -> pNextNSQ)
			    {
			    msgType = htons(NAME_SRCH);
			    if (!caSendNam((char *) &msgType, sizeof(msgType)))
				{
				closeNameSocket();
				break;
				}
			    resendMsg.nsSrch.nsHandle = (int) nsqResend;
			    resendMsg.nsSrch.nsNameLen = strlen(nsqResend -> name) + 1;
			    resendMsg.nsSrch.nsNameLen = htons(resendMsg.nsSrch.nsNameLen);
			    nsqResend -> dueTime = time(NULL) + caNameSrvTimeout;
			    strcpy(resendMsg.chanName, nsqResend -> name);
			    if (!caSendNam((char *) &resendMsg,
				    sizeof(resendMsg.nsSrch) + ntohs(resendMsg.nsSrch.nsNameLen)))
				{
				closeNameSocket();
				break;
				}
			    }
			}
#endif vxWorks or vms

		    else if (msgType == CHAN_IOCUP)
			{
			CH_IOCUP	iocUp;

			/*
			IOC just came up. Get the rest of the header so that we know who just came
			knocking. Remember, we don't have to flip endians here.
			*/

			MARK(IOCUP);
			scmRead((char *) &iocUp, sizeof(iocUp));

			/*
			Mark him as connected. We shouldn't ever have to worry about an IOC that
			we don't already know about. If he was in the original list when we first
			started up, then we both have them. If he was added while the system was
			running, then we find out about him before Socket Creation and Maintenance!
			*/

#if !defined(vxWorks) && !defined(vms)
			tisSysLog(0, 0, "chanAcq: IOC up: %s", INET_NTOA(iocUp.chInet));
#else
			printf("chanAcq: IOC up: %s\n", INET_NTOA(iocUp.chInet));
#endif
			pISB = findISB(iocUp.chInet);

			/*
			Send him a session inquiry to see if we have to start over or if we can pick up 
			where we left off.
			*/

			msgType = htons(CHAN_SESS_INQ);
			sendScm(iocUp.chInet, msgType, NULL, 0);
			}

		    else if (msgType == SOCK_READ)
			{
			SM_READ		msgHdr;

			/*
			Message is inside the general socket wrapper. Read the rest of the header
			to get the originating Inet address. Then, we can process the message.
			*/

			scmRead((char *) &msgHdr, sizeof(msgHdr));
			pISB = findISB(msgHdr.smInet);
			
			/*
			Now get the message type that was sent by the IOC. At this point we have to
			start worrying about endian issues again.
			*/

			scmRead((char *) &msgType, sizeof(msgType));
			msgType = ntohs(msgType);

#if !defined(vxWorks) && !defined(vms)
			if (msgType == CHAN_QBLDACK)
			    {
			    CH_BLDACK		qbldBldack;
			    CH_QBLDACK		chQbldAck;
			    char		*where;
			    short		atShort;
			    float		atFloat;
			    double		atDouble;
			    int			atLong;		/* Channel access longs are really ints! */
			    CALINK		*pCurrValue;
			    VALUE_COMMON	*valCommon;
			    struct
				{
				CH_DRRDATA	drrData;
			    	char		qbldDummy[2048];
				} qbldRead;
			    struct
				{
				CH_MONDATA	monData;
				char		qbldStat[2048];
				} qbldMon;


			    /*
			    Quick Build Acknowledgement. The IOC is going to respond with the
			    status, severity and value of the point. All the other parameters
			    will come out of our local database. Get the body of the message.
			    */

			    scmRead((char *) &chQbldAck, sizeof(chQbldAck));

			    /*
			    Only the master will append data to his response. That is indicated by
			    either chQbldAck.currState != 0 or chQblkdAck.chDataLen != 0. These
			    fields might not have different meanings in the future, so be careful!
			    */

			    chQbldAck.chDataLen = ntohs(chQbldAck.chDataLen);
			    if (chQbldAck.currState && chQbldAck.chDataLen)
				{

				scmRead(qbldMon.qbldStat, chQbldAck.chDataLen);

				/*
				The first 4 data bytes contain stat and sever. See if the
				SEVR word contains the splicing flag. If it does, wipe it out
				and set "splicable" in the ISB.
				*/

				if (*(short *)(qbldMon.qbldStat + sizeof(short)) & htons(0x8000))
				    {
				    pISB -> splicable = 1;
				    }
				}

			    /*
			    We need the native data type and array count at this point. Get that
			    from the local database. First, though, find the CHB!
			    */

			    pCHB = (CHB *) (pISB -> pFirstCHB);
			    while (pCHB)
				{
#ifndef POINTER64
				if ((int) pCHB == chQbldAck.chHandle)
#else
				if (pCHB -> chbToken == chQbldAck.chHandle)
#endif
				     break;
				
				if (pISB == pCHB -> pISB1)
				    pCHB = pCHB -> pNextIOC1;
				else
				    pCHB = pCHB -> pNextIOC2;
				if (!pCHB)
				    {
				    tisSysLog(2, 0, "chanAcq: Quickbuild ack for unknown channel.");
				    rejectChan(pISB, chQbldAck.chIocHandle);
				    continue;
				    }
				}

			    /*
			    Store the handle we were given. Do this even if the channel is delete pending.
			    An app may have closed the channel while the Quickbuild Ack had not yet
			    arrived from the IOC. We still have to tell the IOC to close that channel
			    and then, upon receiving his close ACK, remove it from our name space.
			    */

			    if (pISB == pCHB -> pISB1)
				pCHB -> handleIOC1 = chQbldAck.chIocHandle;
			    else
				pCHB -> handleIOC2 = chQbldAck.chIocHandle;
			
			    /*
			    If the channel should be deleted, our normal channel update processing will
			    send a close message to the IOC. Then, we are through with this message.
			    */

			    if (!(pCHB -> pFirstCAB))
				{
				checkChannel(pCHB);
				continue;
				}

			    /*
			    If this reply is not from the master, we still have to fill in all the
			    parameters. We can have the situation where no IOC is the master and both
			    IOCs connect the channel. Only later does one of them claim the resource.
			    */

			    getAttributes(pCHB -> chanName, pCHB -> pointType, ATTR_NUM_STRS, attrStrings, 
						ATTR_STR_LEN, pCHB -> checksum, pCHB -> creationTime,
						&attrNativeType, &attrNativeCount, &attrHopr, &attrLopr,
						&attrDrvh, &attrDrvl, &attrPrec, &attrNumStrings);
			    attrNativeType = attrMap[attrNativeType];
			    pCHB -> nativeType = attrNativeType;
			    pCHB -> currType = attrNativeType | AGNT_STAT;
			    pCHB -> nativeCount = attrNativeCount;

			    /*
			    Now stop processing if the IOC isn't the master.
			    */

			    if (!chQbldAck.currState)
				continue;
			    pCHB -> pCurrISB = pISB;
			    if (pCHB -> pRes)
				pCHB -> pRes -> currISB = pISB;

			    /*
			    Allocate a value entry and point to it in the tree. Then, stuff the
			    stat, sevr, ack and value into it.
			    */

			    if (isValField(pCHB))
				{
			        pCurrValue = newValue(attrNativeType, attrNativeCount);
			        pCurrValue -> pCHB = pCHB;
			        FA_SetValuePtr(pCHB -> chanName, (void *) pCurrValue);
			        valCommon = pCurrValue -> pData;
			        valCommon -> nativeType = attrNativeType;
			        valCommon -> nativeCount = attrNativeCount;
			        gettimeofday(&(valCommon -> lastUpdate), NULL);
			        valCommon -> stat = *(short *)(qbldMon.qbldStat);
			        valCommon -> sevr = *(short *)(qbldMon.qbldStat + sizeof(short));
			        valCommon -> ack = 0;
			        if (valCommon -> sevr & htons(0x4000))
			    	    valCommon -> ack = htons(1);
			        valCommon -> sevr &= htons(0x3fff);
			        bcopy(qbldMon.qbldStat + sizeof(short) + sizeof(short), ((VALUE_STRING *) valCommon) -> value,
				    chQbldAck.chDataLen - sizeof(short) - sizeof(short));
				}
				    

			    /*
			    Format the read response now. It, in theory, could be used multiple
			    times. Everything always starts with STAT and SEVR.
			    */

			    *(short *)(qbldMon.qbldStat + sizeof(short)) &= htons(0x3fff);
			    bcopy(qbldMon.qbldStat, qbldRead.qbldDummy, sizeof(short) + sizeof(short));
			    where = qbldRead.qbldDummy + sizeof(short) + sizeof(short);

			    switch (attrNativeType)
				{
				case AGNT_STRING:
				    memset(where, 0, 44);	/* Just null out the additional data */
				    where += 44;
				    break;
				
				case AGNT_CHAR:
				    *where++ = (char) attrHopr;
				    *where++ = (char) attrLopr;
				    *where++ = (char) attrDrvh;
				    *where++ = (char) attrDrvl;
				    break;

				case AGNT_SHORT:
				    atShort = htons(((short) attrHopr));	/* Lotsa parens in case htons is a macro! */
				    bcopy((char *) &atShort, where, sizeof(short));
				    where += sizeof(short);
				    atShort = htons(((short) attrLopr));
				    bcopy((char *) &atShort, where, sizeof(short));
				    where += sizeof(short);
				    atShort = htons(((short) attrDrvh));
				    bcopy((char *) &atShort, where, sizeof(short));
				    where += sizeof(short);
				    atShort = htons(((short) attrDrvl));
				    bcopy((char *) &atShort, where, sizeof(short));
				    where += sizeof(short);
				    break;

				case AGNT_LONG:			/* In CA, "longs" are really ints! */
				    atLong = htonl(((int) attrHopr));
				    bcopy((char *) &atLong, where, sizeof(int));
				    where += sizeof(int);
				    atLong = htonl(((int) attrLopr));
				    bcopy((char *) &atLong, where, sizeof(int));
				    where += sizeof(int);
				    atLong = htonl(((int) attrDrvh));
				    bcopy((char *) &atLong, where, sizeof(int));
				    where += sizeof(int);
				    atLong = htonl(((int) attrDrvl));
				    bcopy((char *) &atLong, where, sizeof(int));
				    where += sizeof(int);
				    break;

				case AGNT_FLOAT:
				    atFloat = (float) attrHopr;
				    htonf(atFloat, atFloat);
				    bcopy((char *) &atFloat, where, sizeof(float));
				    where += sizeof(float);
				    atFloat = (float) attrLopr;
				    htonf(atFloat, atFloat);
				    bcopy((char *) &atFloat, where, sizeof(float));
				    where += sizeof(float);
				    atShort = htons(attrPrec);
				    bcopy((char *) &atShort, where, sizeof(short));
				    where += sizeof(short);
				    atFloat = (float) attrDrvh;
				    htonf(atFloat, atFloat);
				    bcopy((char *) &atFloat, where, sizeof(float));
				    where += sizeof(float);
				    atFloat = (float) attrDrvl;
				    htonf(atFloat, atFloat);
				    bcopy((char *) &atFloat, where, sizeof(float));
				    where += sizeof(float);
{
float	crashFloat;
bcopy(qbldMon.qbldStat + sizeof(short) + sizeof(short), (char *) &crashFloat, sizeof(crashFloat));
if ((crashFloat > 1.0E29) || (crashFloat < -1.0E29))
	printf("BAD FLOAT 1!\n");
}
				    break;

				case AGNT_DOUBLE:
				    htond(attrHopr, atDouble);
				    bcopy((char *) &atDouble, where, sizeof(double));
				    where += sizeof(double);
				    htond(attrLopr, atDouble);
				    bcopy((char *) &atDouble, where, sizeof(double));
				    where += sizeof(double);
				    atShort = htons(attrPrec);
				    bcopy((char *) &atShort, where, sizeof(double));
				    where += sizeof(short);
				    htond(attrDrvh, atDouble);
				    bcopy((char *) &atDouble, where, sizeof(double));
				    where += sizeof(double);
				    htond(attrDrvl, atDouble);
				    bcopy((char *) &atDouble, where, sizeof(double));
				    where += sizeof(double);
{
double	crashDouble;
bcopy(qbldMon.qbldStat + sizeof(short) + sizeof(short), (char *) &crashDouble, sizeof(crashDouble));
if ((crashDouble > 1.0E29) || (crashDouble < -1.0E29))
	printf("BAD DOUBLE 1!\n");
}
				    break;

				case AGNT_ENUM:
				    atShort = htons(attrNumStrings);
				    bcopy((char *) &atShort, where, sizeof(short));
				    where += sizeof(short);
				    for (atShort = 0; atShort < attrNumStrings; atShort++)
					{
					strcpy(where, attrStrings[atShort]);
					where += (strlen(where) + 1);
					}
				    break;
				}

			    /*
			    Now move in the value and set the pointer past the end of it.
			    */

			    bcopy(qbldMon.qbldStat + sizeof(short) + sizeof(short), where,
				chQbldAck.chDataLen - sizeof(short) - sizeof(short));
			    where += (chQbldAck.chDataLen - sizeof(short) - sizeof(short));

			    /*
			    Now fill in the fields that won't change from user to user.
			    Do this for the DRRDATA and MONDATA messages.
			    */

			    qbldRead.drrData.chCaHandle = 0;
			    qbldRead.drrData.chDataType = htons((short)(attrNativeType | AGNT_STAT | AGNT_GRAPH | AGNT_CNTRL));
			    qbldRead.drrData.chDataLen = where - qbldRead.qbldDummy;
			    qbldRead.drrData.chDataLen = htons(qbldRead.drrData.chDataLen);
			    qbldRead.drrData.chCount = htons(attrNativeCount);
			    qbldMon.monData.chDataType = htons((short)(attrNativeType | AGNT_STAT));
			    qbldMon.monData.chDataLen = htons(chQbldAck.chDataLen);
			    qbldMon.monData.chMask = htons(AGNT_VALUE);
			    qbldMon.monData.chCount = htons(attrNativeCount);

			    /*
			    Now wake up everyone waiting for an ack on this channel.
			    */

			    numCABs = findNumCABs(pCHB);
			    currCabList = &firstCabList;
			    for (cabIndex = 0; cabIndex < numCABs; )
				{
				pCAB = currCabList -> refCAB[cabIndex & 0x1f];
				if (pCAB -> pCHB)
				    {
				    if (pCAB -> awaitAck)
					{
					short		spliceEnumVal;

					if (pCAB -> mask & AGNT_SPLICED)
					    pCAB -> dataType = AGNT_ENUM | AGNT_STAT;
					else
				            pCAB -> dataType = pCHB -> nativeType | AGNT_STAT;
				        pCAB -> count = pCHB -> nativeCount;
					qbldBldack.chHandle = pCAB -> handle;
#ifndef POINTER64
					qbldBldack.chIocHandle = (int) pCAB;
#else
					qbldBldack.chIocHandle = pCAB -> cabToken;
#endif
					if (pCAB -> mask & AGNT_SPLICED)
					    {
					    qbldBldack.chDataType = htons(AGNT_ENUM);
					    qbldBldack.chElementCount = htons(1);
					    }
					else
					    {
					    qbldBldack.chDataType = htons(attrNativeType);
					    qbldBldack.chElementCount = htons(attrNativeCount);
					    }
					msgType = htons(CHAN_BLDACK);
					if (!appTx(pCAB -> pACB, msgType, (char *) &qbldBldack, sizeof(qbldBldack)))
					    break;
					pCAB -> awaitAck = 0;
					
					qbldRead.drrData.chHandle = pCAB -> handle;
					msgType = htons(CHAN_DRRDATA);

					/*
					If the splicing bits are on, send an alternative message.
					*/

					if (pCAB -> mask & AGNT_SPLICED)
					    {
					    struct
						{
						CH_DRRDATA	spliceRead;
						char		spliceVal[1024];
						} spliced;
					    short		spliceChoice;
					    short		index;
					    char		*where;
					    struct choiceSet	*pChoiceSet;

					    spliced.spliceRead = qbldRead.drrData;	/* Copy all the regular stuff */
					    spliced.spliceRead.chDataType = htons(AGNT_ENUM | AGNT_STAT | AGNT_GRAPH | AGNT_CNTRL);
					    where = spliced.spliceVal;
					    if (pCAB -> mask & AGNT_SPL_STAT)
						{
						spliceChoice = GBL_ALARM_STAT;
						spliceEnumVal = valCommon -> stat;
						}
					    else if (pCAB -> mask & AGNT_SPL_SEVR)
						{
						spliceChoice = GBL_ALARM_SEV;
						spliceEnumVal = valCommon -> sevr;
						}
					    else if (pCAB -> mask & AGNT_SPL_ACK)
						{
						spliceChoice = GBL_YES_NO;
						spliceEnumVal = valCommon -> ack;
						}
					    *(short *) where = valCommon -> stat;
					    where += sizeof(short);
					    *(short *) where = valCommon -> sevr;
					    where += sizeof(short);
					    pChoiceSet = GET_PCHOICE_SET(choiceGbl, spliceChoice);
					    spliceChoice = (short)pChoiceSet -> number;
					    spliceChoice = htons(spliceChoice);
					    *(short *) where = spliceChoice;
					    where += sizeof(short);
					    for (index = 0; index < pChoiceSet -> number; index++)
						{
						strcpy(where, GET_CHOICE(pChoiceSet, index));
						where += strlen(where) + 1;
						}
					    bcopy((char *) &spliceEnumVal, where, sizeof(short));
					    where += sizeof(short);
					    spliced.spliceRead.chDataLen = where - spliced.spliceVal;
					    spliced.spliceRead.chDataLen = htons(spliced.spliceRead.chDataLen);
					    if (!appTx(pCAB -> pACB, msgType, (char *) &spliced, sizeof(CH_DRRDATA)
						    + where - spliced.spliceVal))
						break;
					    }
					else if (!appTx(pCAB -> pACB, msgType, (char *) &qbldRead, sizeof(CH_DRRDATA) +
						ntohs(qbldRead.drrData.chDataLen)))
					    break;
					pCAB -> readType = AGNT_NOCONN;

					msgType = htons(CHAN_MONDATA);
					qbldMon.monData.chHandle = pCAB -> handle;
					if (pCAB -> mask & AGNT_SPLICED)
					    {
					    struct
						{
						CH_MONDATA	spliceMon;
						short		spliceStat;
						short		spliceSevr;
						short		spliceVal;
						} spliceMond;

					    spliceMond.spliceMon = qbldMon.monData;
					    spliceMond.spliceStat = valCommon -> stat;
					    spliceMond.spliceSevr = valCommon -> sevr;
					    spliceMond.spliceVal = spliceEnumVal;
					    spliceMond.spliceMon.chDataType = htons(AGNT_ENUM | AGNT_STAT);
					    spliceMond.spliceMon.chDataLen = htons((sizeof(short) + sizeof(short) + sizeof(short)));
					    if (!appTx(pCAB -> pACB, msgType, (char *) &spliceMond, sizeof(spliceMond)))
						break;
					    }
					else if (!appTx(pCAB -> pACB, msgType, (char *) &qbldMon, sizeof(CH_MONDATA) +
						ntohs(qbldMon.monData.chDataLen)))
					    break;
					}
				    }
				cabIndex++;
				if ((cabIndex & 0x1f) == 0)
				    currCabList = currCabList -> nextList;
				}
			    }

			else
#endif
			if (msgType == CHAN_BLDACK)
			    {
			    CH_BLDACK		bldAck1;

			    /*
			    Channel Build acknowledgement. The IOC got our CHAN_BUILD message and is
			    now giving us back the handle for this channel and its resource status.
			    Get the body of the message.
			    */

			    scmRead((char *) &bldAck1, sizeof(bldAck1));
			    
			    /*
			    Find the CHB that he's talking about. We'll search for it because we
			    don't believe a pointer that anyone else gave us to keep ourselves from
			    crashing due to invalid memory addresses.
			    */

			    MARK(BDACK1);
			    pCHB = (CHB *) (pISB -> pFirstCHB);
			    while (pCHB)
				{
#ifndef POINTER64
				if ((int) pCHB == bldAck1.chHandle)
#else
				if (pCHB->chbToken == bldAck1.chHandle)
#endif
				    break;
				if (pISB == pCHB -> pISB1)
				    pCHB = pCHB -> pNextIOC1;
				else
				    pCHB = pCHB -> pNextIOC2;
				}
			    if (!pCHB)
				{

				/*
				I don't think this can happen, but better safe than sorry.
				*/

#if !defined(vxWorks) && !defined(vms)
				tisSysLog(2, 0, "chanAcq: Build ack for unknown channel %X", bldAck1.chHandle);
#else
				if (cacDebug >= 2)
				    printf("chanAcq: Build ack for unknown channel %X\n", bldAck1.chHandle);
#endif
				rejectChan(pISB, bldAck1.chIocHandle);
				continue;
				}

			    /*
			    Store the handle that we were given. Do this even if the channel is delete
			    pending. An app may have closed the channel while the build ack was still
			    outstanding from the IOC. We still have to tell the IOC to close that
			    channel and then, upon his close ack, remove it from our name space.
			    */

			    if (pISB == pCHB -> pISB1)
				{
				pCHB -> handleIOC1 = bldAck1.chIocHandle;
				}
			    else
				{
				pCHB -> handleIOC2 = bldAck1.chIocHandle;
				}

			    MARK(BDACK2);
#if !defined(vxWorks) && !defined(vms)
#ifndef POINTER64
			    tisSysLog(2, 0, "chanAcq: Build ack for %s, CHB = %X, IOC handle = %X", pCHB -> chanName,
				    (int) pCHB, bldAck1.chIocHandle);
#else
			    tisSysLog(2, 0, "chanAcq: Build ack for %s, CHB = %X, IOC handle = %X", pCHB -> chanName,
				    pCHB->chbToken, bldAck1.chIocHandle);
#endif
#else
			    if (cacDebug >= 2)
				printf("chanAcq: Build ack of %s, CHB = %X, IOC handle = %X\n", pCHB -> chanName,
				    (int) pCHB, bldAck1.chIocHandle);
#endif
			    
			    /*
			    If the point should be deleted, our normal channel update processing will
			    send a close message to the IOC. Then we're done with this message.
			    */

			    if (!(pCHB -> pFirstCAB))
				{
				checkChannel(pCHB);
				continue;
				}

			    /*
			    If this is a workstation that supports splicing, put the appropriate data in the
			    tree after we establish a place for it.
			    */

#ifndef vxWorks
			    if (isValField(pCHB))
				{
				CALINK		*caLink;
				VALUE_COMMON	*valCommon;

				caLink = FA_GetValuePtr(pCHB -> chanName);
				if (!caLink)
				    {
			            caLink = newValue(ntohs(bldAck1.chDataType), ntohs(bldAck1.chElementCount));
			            caLink -> pCHB = pCHB;
			            valCommon = (VALUE_COMMON *) (caLink -> pData);
			            valCommon -> nativeType = ntohs(bldAck1.chDataType);
			            valCommon -> nativeCount = ntohs(bldAck1.chElementCount);
			            valCommon -> stat = 0;
			            valCommon -> sevr = 0;
			            valCommon -> ack = 0;
			            valCommon -> lastUpdate.tv_sec = 0;
			            valCommon -> lastUpdate.tv_usec = 0;
			            FA_SetValuePtr(pCHB -> chanName, caLink);
				    }
				}
#endif

			
			    /*
			    If we already know that there should be a monitor on this channel, apply
			    it to the channel.
			    */

			    if (pCHB -> currMask && ((pCHB -> currType & 0xff) != AGNT_ADAPTIVE))
				{
				CH_TYPE	chnType;

				chnType.chHandle = bldAck1.chIocHandle;
				chnType.chType = htons(pCHB -> currType);
				chnType.chCount = htons((short)pCHB -> currCount);
				chnType.chMask = htons(pCHB -> currMask);
				msgType = htons(CHAN_TYPE);
				sendScm(pISB -> Inet, msgType, (char *) &chnType, sizeof(chnType));
				}

			    /*
			    Mark the channel as connected and store the handle. Then, if the IOC owns
			    the resource, run down the list of CABs and send a Build Ack to everyone 
			    who was waiting and didn't want an initial read.  At the same time, if a 
			    direct read is pending because of an initial data sample desired at open 
			    time, issue that read now. Defer the Build Ack until the direct read completes.
			    If a monitor is to be placed on the channel, ask for the initial report also.
			    */

			    pCHB -> nativeType = ntohs(bldAck1.chDataType);
			    pCHB -> nativeCount = ntohs(bldAck1.chElementCount);
			    if (!bldAck1.currState)
				continue;
			    pCHB -> pCurrISB = pISB;
			    if (pCHB -> pRes)
				pCHB -> pRes -> currISB = pISB;
			    numCABs = findNumCABs(pCHB);
			    currCabList = &firstCabList;
			    for (cabIndex = 0; cabIndex < numCABs;)
				{
				pCAB = currCabList -> refCAB[cabIndex & 0x1f];
				if (pCAB -> pCHB)
				    {
				    if (pCAB -> awaitAck)
				        {
				        if (pCAB -> readType == AGNT_NOCONN)
					    {
					    bldAck1.chHandle = pCAB -> handle;
#ifndef POINTER64
					    bldAck1.chIocHandle = (int) pCAB;
#else
					    bldAck1.chIocHandle = pCAB->cabToken;
#endif
					    if (pCAB -> mask & AGNT_SPLICED)
						{
						bldAck1.chDataType = htons(AGNT_ENUM);
						bldAck1.chElementCount = htons(1);
						}
					    else
						{
					        bldAck1.chDataType = htons(pCHB -> nativeType);
					        bldAck1.chElementCount = htons(pCHB -> nativeCount);
						}
					    msgType = htons(CHAN_BLDACK);
					    if (!appTx(pCAB -> pACB, msgType, (char *) &bldAck1, sizeof(bldAck1)))
						break;
					    pCAB -> awaitAck = 0;
					    }
				        }

				    if ((pCAB -> readType != AGNT_NOCONN) && ((pCAB -> readType & 0xff) != AGNT_ADAPTIVE))
				        {
				        CH_DIRREAD	dirRead1;

				        msgType = htons(CHAN_DIRREAD);
				        if (pCHB -> pCurrISB == pCHB -> pISB1)
					    dirRead1.chHandle = pCHB -> handleIOC1;
				        else
					    dirRead1.chHandle = pCHB -> handleIOC2;
#ifndef POINTER64
				        dirRead1.chCaHandle = (int) pCAB;
				        dirRead1.chAppHandle = (int) (pCAB -> pACB);
#else
				        dirRead1.chCaHandle = pCAB->cabToken;
				        dirRead1.chAppHandle = pCAB -> pACB->acbToken;
#endif
				        dirRead1.chDataType = htons(pCAB -> readType);
				        dirRead1.chNumElem = htons(pCAB -> count);
				        sendScm(pISB -> Inet, msgType, (char *) &dirRead1, sizeof(dirRead1));
#if !defined(vxWorks) && !defined(vms)
					tisSysLog(2, 0, "\tSent direct read for connection data type.");
#else
					if (cacDebug >= 2)
					    printf("\tSent direct read for connection data type.\n");
#endif
				        }

				    if ((pCAB -> mask & ~(AGNT_SPLICED | AGNT_SPL_VAL)) && ((pCAB -> readType & 0xff) != AGNT_ADAPTIVE))
				        {
				        CH_MONREQ		reqFirstMon;

				        if (pCHB -> pCurrISB == pCHB -> pISB1)
				    	    reqFirstMon.chHandle = pCHB -> handleIOC1;
				        else
					    reqFirstMon.chHandle = pCHB -> handleIOC2;
#ifndef POINTER64
				        reqFirstMon.chCaHandle = (int) pCAB;
#else
				        reqFirstMon.chCaHandle = pCAB -> cabToken;
#endif

				        reqFirstMon.chType = ntohs(pCAB -> dataType);
				        reqFirstMon.chCount = ntohs(pCAB -> count);
				        reqFirstMon.mask = ntohs((short)(pCAB -> mask & ~(AGNT_SPLICED | AGNT_SPL_VAL)));
				        msgType = htons(CHAN_MONREQ);
				        sendScm(pCHB -> pCurrISB -> Inet, msgType, (char *) &reqFirstMon,
					        sizeof(reqFirstMon));
				        }
				    }

				cabIndex++;
				if ((cabIndex & 0x1f) == 0)
				    currCabList = currCabList -> nextList;
				}
			    }

			else if (msgType == CHAN_DRRDATA)
			    {
			    struct drHold
				{
			        CH_DRRDATA	drrData;
			        char		dataBuffer[2048];
				} drMsg;
			    struct drHold	spliceMsg;
			    int			length1;
			    ACB			*pACB;

			    /*
			    Direct Read Data. We requested an immediate read and this is the answer.
			    Start by reading the message header and the data to keep the sockets in
			    synchronization. Then, find the CAB referenced in the message. We'll
			    search for it because we can't trust a point that anybody else sends to us.
			    Note that the search starts with the ACB instead of looking at CHBs starting
			    at the ISB because the CHB may have been changed by type splicing/promotion.
			    */

			    MARK(RDATA1);
			    scmRead((char *) &drMsg.drrData, sizeof(drMsg.drrData));
			    length1 = ntohs(drMsg.drrData.chDataLen);
			    scmRead(drMsg.dataBuffer, length1);
			    if (ntohs(drMsg.drrData.chDataType) & AGNT_STAT)
				{

				/*
				Status information is present. Look for the splicing flag.
				*/

#ifndef vxWorks
				if (*(short *)(drMsg.dataBuffer + sizeof(short)) & htons(0x8000))
				    {
				    pISB -> splicable = 1;
				    }
#endif
				}

			    pACB = acbList;
			    while (pACB)
				{
#ifndef POINTER64
				if (drMsg.drrData.chHandle == (int) pACB)
#else
				if (drMsg.drrData.chHandle == pACB -> acbToken)
#endif
				    break;
				if (pACB == pACB -> pNextACB)
				    {
				    pACB = 0;
				    break;
				    }
				pACB = pACB -> pNextACB;
				}
			    if (!pACB)
				{
#if !defined(vxWorks) && !defined(vms)
				tisSysLog(3, 0, "chanAcq: DRRDATA discarded because I can't find the ACB.");
				tisSysLog(3, 0, "\tchHandle = %X, chCaHandle = %X\n", drMsg.drrData.chHandle,
				        drMsg.drrData.chCaHandle);
#else
				if (cacDebug >= 3)
				    {
				    printf("chanAcq: DRRDATA discarded because I can't find the ACB.\n");
				    printf("\tchHandle = %X, chCaHandle = %X\n", drMsg.drrData.chHandle,
					    drMsg.drrData.chCaHandle);
				    }
#endif
				continue;
				}
			    
			    /*
			    Find the CAB chained to this ACB that is referenced in the read.
			    */

			    MARK(RDATA2);
			    pCAB = pACB -> pFirstCAB;
			    while (pCAB)
				{
#ifndef POINTER64
				if (drMsg.drrData.chCaHandle == (int) pCAB)
#else
				if (drMsg.drrData.chCaHandle == pCAB -> cabToken)
#endif
				    break;
				pCAB = pCAB -> pNextApp;
				}
			    if (!pCAB)
				{
#if !defined(vxWorks) && !defined(vms)
				tisSysLog(3, 0, "chanAcq: DRRDATA discarded because I can't find CAB.");
				tisSysLog(3, 0, "\tchHandle = %X, chCaHandle = %X", drMsg.drrData.chHandle, 
					drMsg.drrData.chCaHandle);
#else
				if(cacDebug >= 3)
				    {
				    printf("chanAcq: DRRDATA discarded because I can't find CAB.\n");
				    printf("\tchHandle = %X, chCaHandle = %X", drMsg.drrData.chHandle, 
					    drMsg.drrData.chCaHandle);
				    }
#endif
				continue;
				}

			    /*
			    If the CAB doesn't point back to a CHB, then the channel has been closed
			    before the read completed. Discard the data.
			    */

			    if (!(pCAB -> pCHB))
				continue;
#if !defined(vxWorks) && !defined(vms)
			    tisSysLog(3, 0, "chanAcq: DRRDATA for %s", pCAB -> pCHB -> chanName);
#else
			    if (cacDebug >= 3)
			        printf("chanAcq: DRRDATA for %s\n", pCAB -> pCHB -> chanName);
#endif

			    /*
			    Make sure that we have a handle for this IOC because of the following
			    situation: An application builds a channel and causes a read. He closes
			    the channel before the read is fully processed. He opens another channel
			    and that previously used CAB gets assigned to the new one. A build is
			    sent to the IOC but the OLD read response comes back before the build ACK.
			    This makes us generate a BLDACK to the application without us first
			    having the correct data type and element count for the channel.
			    What we will do is to make sure that for every read response that we
			    have a valid handle for the IOC that sent the message. If not, drop it
			    on the floor.
			    */

			    pCHB = pCAB -> pCHB;
			    if (pISB == pCHB -> pISB1)
				{
				if ((pCHB -> handleIOC1 == 0) || (pCHB -> handleIOC1 == -1))
				    continue;
				}
			    else
				{
				if ((pCHB -> handleIOC2 == 0) || (pCHB -> handleIOC2 == -1))
				    continue;
				}

			    /*
			    This read can satisfy the CAB. Now, check to see if the application
			    is awaiting an open acknowledgement. If so send the ack to exchange
			    handles.
			    */

			    if (pCAB -> awaitAck)
				{
				CH_BLDACK	bldAck2;

				msgType = htons(CHAN_BLDACK);
				bldAck2.chHandle = pCAB -> handle;
#ifndef POINTER64
				bldAck2.chIocHandle = (int) pCAB;
#else
				bldAck2.chIocHandle = pCAB -> cabToken;
#endif
				if (pCAB -> mask & AGNT_SPLICED)
				    {
				    bldAck2.chDataType = htons(AGNT_ENUM);
				    bldAck2.chElementCount = htons(1);
				    }
				else
				    {
				    bldAck2.chDataType = htons(pCHB -> nativeType);
				    bldAck2.chElementCount = htons(pCHB -> nativeCount);
				    }
				if (!appTx(pCAB -> pACB, msgType, (char *) &bldAck2, sizeof(bldAck2)))
				    continue;
				pCAB -> awaitAck = 0;
				}
				    
			    /*
			    Change the handle for the user and send it on to the application.
			    The call to transformType was inserted here to handle splicing.
			    */

			    drMsg.drrData.chHandle = pCAB -> handle;

			    spliceMsg.drrData = drMsg.drrData;
			    msgType = pCAB -> dataType;
			    if (msgType == AGNT_ADAPTIVE)
				{
				if (pCAB -> mask & AGNT_SPLICED)
				    msgType = AGNT_ENUM | AGNT_GRAPH | AGNT_STAT | AGNT_CNTRL;
				else
				    msgType = ntohs(drMsg.drrData.chDataType);
				pCAB -> count = pCHB -> nativeCount;
				}
			    spliceMsg.drrData.chDataType = htons(msgType);
			    spliceMsg.drrData.chDataLen = transformType(ntohs(drMsg.drrData.chDataType), ntohs(drMsg.drrData.chCount), 
					msgType, pCAB -> count, drMsg.dataBuffer, spliceMsg.dataBuffer, (short)(pCAB -> mask & AGNT_SPLICED));
			    spliceMsg.drrData.chDataLen = htons(spliceMsg.drrData.chDataLen);
			    msgType = htons(CHAN_DRRDATA);
			    appTx(pCAB -> pACB, msgType, (char *) &spliceMsg, sizeof(CH_DRRDATA) 
					+ ntohs(spliceMsg.drrData.chDataLen));
			    }

			else if (msgType == CHAN_SINGMON)
			    {
			    struct singHold
				{
			        CH_SINGMON	singMon;
				char		dataBuffer[2048];
				} singMsg;
			    struct
				{
				CH_MONDATA	monData;
				char		dataBuffer[2048];
				} singOutput;
			    int			length3;
			    CAB			*smCAB;

			    /*
			    A single application monitor report. This is a report for one and only one
			    application. It is essentially the same as a normal monitor report (i.e.,
			    message type CHAN_MONDATA) except that we get a pointer to a CAB in addition
			    to a pointer to the CHB. That uniquely identifies the application.
			    */

			    MARK(SINGM1);
			    scmRead((char *) &singMsg.singMon, sizeof(singMsg.singMon));

			    length3 = ntohs(singMsg.singMon.chDataLen);
			    scmRead((char *) singMsg.dataBuffer, length3);
#ifndef vxWorks
			    if (ntohs(singMsg.singMon.chDataType) & AGNT_STAT)
				{

				/*
				If splicing is on, wipe out the bits and set ISB flag.
				*/

				if (*(short *)(singMsg.dataBuffer + sizeof(short)) & htons(0x8000))
				    {
				    pISB -> splicable = 1;
				    }
				}
#endif

			    /*
			    Find the CHB he's talking about.
			    */

			    pCHB = (CHB *) (pISB -> pFirstCHB);
			    while (pCHB)
				{
#ifndef POINTER64
				if (singMsg.singMon.chHandle == (int) pCHB)
#else
				if (singMsg.singMon.chHandle == pCHB -> chbToken)
#endif

				    break;
				if (pCHB -> pISB1 == pISB)
				    pCHB = pCHB -> pNextIOC1;
				else
				    pCHB = pCHB -> pNextIOC2;
				}
			    if (!pCHB)
				{
				/*
				rejectChan(pISB, singMsg.singMon.chHandle);
				Don't send this! we don't have his handle!
				*/
				continue;
				}

			    /*
			    If the CHB is delete pending, ignore the message.
			    */

			    if (!(pCHB -> pFirstCAB))
				{
				/*
				rejectChan(pISB, singMsg.singMon.chHandle);
				Don't send this! We need the IOC's handle!
				*/
				continue;
				}

			    MARK(SIGNM2);
#if !defined(vxWorks) && !defined(vms)
			    tisSysLog(6, 0, "Single monitor report for %s", pCHB -> chanName);
#else
			    if (cacDebug >= 6)
				printf("Single monitor report for %s\n", pCHB -> chanName);
#endif
			    
			    /*
			    Ignore the message if the application closed the channel prior
			    to getting this report. Validate that by making sure that the
			    CAB is actually chained onto the CHB.
			    */

#ifndef POINTER64
			    smCAB = (CAB *) singMsg.singMon.chCaHandle;
#else
			    smCAB = (CAB *) Pointer64fromToken (singMsg.singMon.chCaHandle);
#endif
			    pCAB = pCHB -> pFirstCAB;
			    while (pCAB)
				{
				if (pCAB == smCAB)
				    break;
				pCAB = pCAB -> pNextChan;
				}
			    if (!pCAB)
				continue;
			   
			    if (!(pCAB -> pCHB))
				continue;

			    /*
			    Make sure that we have a handle for this IOC because of the following
			    situation: An application builds a channel and creates a monitor. He closes
			    the channel before the monitor is fully processed. He opens another channel
			    and that previously used CAB gets assigned to the new one. A build is
			    sent to the IOC but the OLD SINGMON comes back before the build ACK.
			    What we will do is to make sure that for SINGMON that we have a valid handle 
			    for the IOC that sent the message. If not, drop it on the floor.
			    */

			    if (pISB == pCHB -> pISB1)
				{
				if ((pCHB -> handleIOC1 == 0) || (pCHB -> handleIOC1 == -1))
				    continue;
				}
			    else
				{
				if ((pCHB -> handleIOC2 == 0) || (pCHB -> handleIOC2 == -1))
				    continue;
				}

			    /*
			    It's also possible that a "regular" monitor got here before this
			    special one.
			    */

			    if (!(pCAB -> awaitMon))
				{
#if !defined(vxWorks) && !defined(vms)
				tisSysLog(2, 0, "\tDiscarded because it was a duplicate!\n");
#else
				if (cacDebug >= 2)
				    printf("\tDiscarded because it was a duplicate!\n");
#endif
				continue;
				}

			    /*
			    And also, on workstations only, park this information in the tree if
			    it contains ONLY the value and status information. This is the variant
			    used by dm.
			    */
#ifndef vxWorks
			    {
			    CALINK		*caLink;
			    VALUE_COMMON	*valCommon;

			    if (isValField(pCHB) && ((pCAB -> dataType & 0xff00) == AGNT_STAT))
				{
				caLink = FA_GetValuePtr(pCHB -> chanName);
				if (!caLink)
				    {
			            caLink = newValue(pCHB -> nativeType, pCHB -> nativeCount);
			            caLink -> pCHB = pCHB;
			            valCommon = (VALUE_COMMON *) (caLink -> pData);
			            valCommon -> nativeType = pCHB -> nativeType;
			            valCommon -> nativeCount = pCHB -> nativeCount;
			            valCommon -> stat = 0;
			            valCommon -> sevr = 0;
			            valCommon -> ack = 0;
			            valCommon -> lastUpdate.tv_sec = 0;
			            valCommon -> lastUpdate.tv_usec = 0;
			            FA_SetValuePtr(pCHB -> chanName, caLink);
				    }

				valCommon = caLink -> pData;
				valCommon -> stat = *(short *)singMsg.dataBuffer;
				valCommon -> sevr = *(short *)(singMsg.dataBuffer + sizeof(short));
				if (ntohs(valCommon -> sevr) & 0x4000)
				    valCommon -> ack = htons(1);
				else
				    valCommon -> ack = 0;
			        valCommon -> sevr &= htons(0x3fff);
				    
			        bcopy(singMsg.dataBuffer + sizeof(short) + sizeof(short), ((VALUE_STRING *) valCommon) -> value,
				    ntohs(singMsg.singMon.chDataLen) - sizeof(short) - sizeof(short));
				}
			    }
#endif

			    /*
			    Now create the monitor report.
			    */

			    pCAB -> awaitMon = 0;
			    if ((pCAB -> dataType & 0xff) == AGNT_ADAPTIVE)
				{
				pCHB -> currType = pCHB -> nativeType | AGNT_STAT;
				if (pCAB -> mask & AGNT_SPLICED)
				    pCAB -> dataType = AGNT_ENUM | AGNT_STAT;
				else
			            pCAB -> dataType = pCHB -> nativeType | AGNT_STAT;
			        pCAB -> count = pCHB -> nativeCount;
				pCAB -> readType = AGNT_NOCONN;
				pCHB -> currCount = pCHB -> nativeCount;
				adjustType(pCAB);
				checkChannel(pCHB);
				}
			    singOutput.monData.chHandle = pCAB -> handle;
			    singOutput.monData.chDataType = htons(pCAB -> dataType);
			    if (pCAB -> mask & AGNT_SPLICED)
				singOutput.monData.chDataType = htons(AGNT_ENUM | AGNT_STAT);
			    singOutput.monData.chMask = singMsg.singMon.chMask;
			    singOutput.monData.chCount = htons(pCAB -> count);
			    length3 = transformType(ntohs(singMsg.singMon.chDataType), ntohs(singMsg.singMon.chCount), 
						   pCAB -> dataType, pCAB -> count, singMsg.dataBuffer, 
						   singOutput.dataBuffer, (short)(pCAB -> mask & AGNT_SPLICED));
			    singOutput.monData.chDataLen = htons((short)length3);
			    msgType = htons(CHAN_MONDATA);
			    appTx(pCAB -> pACB, msgType, (char *) &singOutput, sizeof(CH_MONDATA) 
					+ length3);

			    }

			else if (msgType == CHAN_MONDATA)
			    {

			    /*
			    Data from a monitor.  Search for the CHB because we don't trust anybody.
			    */

			    struct monHold
				{
			        CH_MONDATA		monData;
			        char		dataBuffer[2048];
				} monMsg;
			    struct monHold	monOutput;
			    int			length2;
			    short		mask;
#ifndef vxWorks
			    short		spliceStat;
			    short		spliceSevr;
			    short		spliceAck;
#endif vxWorks

			    MARK(MOND1);
			    scmRead((char *) &monMsg.monData, sizeof(monMsg.monData));
			    length2 = ntohs(monMsg.monData.chDataLen);
			    scmRead((char *) monMsg.dataBuffer, length2);
#ifndef vxWorks
			    if (ntohs(monMsg.monData.chDataType) & AGNT_STAT)
				{

				/*
				If splicing flag is on, set flag in ISB and clear sevr.
				*/

				if (*(short *)(monMsg.dataBuffer + sizeof(short)) & htons(0x8000))
				    {
				    pISB -> splicable = 1;
				    }
				}
#endif

			    pCHB = (CHB *) (pISB -> pFirstCHB);
			    while (pCHB)
				{
#ifndef POINTER64
				if (monMsg.monData.chHandle == (int) pCHB)
#else
				if (monMsg.monData.chHandle == pCHB -> chbToken)
#endif
				    break;
				if (pCHB -> pISB1 == pISB)
				    pCHB = pCHB -> pNextIOC1;
				else
				    pCHB = pCHB -> pNextIOC2;
				}
			    if (!pCHB)
				{
				/*
				rejectChan(pISB, monMsg.monData.chHandle);
				Don't send this! We don't have the IOC's handle!
				*/
				continue;
				}
			    if (!(pCHB -> pFirstCAB))
				{
				/*
				rejectChan(pISB, monMsg.monData.chHandle);
				Don't send this! We need the IOC's handle!
				*/
				continue;
				}
			    if ( ((ntohs(monMsg.monData.chDataType) & 0xff) != (pCHB -> currType & 0xff))
				 && ((pCHB -> currType & 0xff) != AGNT_ADAPTIVE) )
			        {
#if !defined(vxWorks) && !defined(vms)
			        tisSysLog(1, 0, "Monitor msg type mismatch for %s discarded", pCHB -> chanName);
#endif
			        continue;
			        }
#if !defined(vxWorks) && !defined(vms)
			    tisSysLog(6, 0, "Event monitor for %s", pCHB -> chanName);
#else
			    if (cacDebug >= 6)
				printf("Event monitor for %s\n", pCHB -> chanName);
#endif

#ifndef vxWorks
			    {
			    CALINK		*caLink;
			    VALUE_COMMON	*valCommon;

			    if (isValField(pCHB) && ((ntohs(monMsg.monData.chDataType) & 0xff00) == AGNT_STAT))
				{
				caLink = FA_GetValuePtr(pCHB -> chanName);
				if (!caLink)
				    {
			            caLink = newValue(pCHB -> nativeType, pCHB -> nativeCount);
			            caLink -> pCHB = pCHB;
			            valCommon = (VALUE_COMMON *) (caLink -> pData);
			            valCommon -> nativeType = pCHB -> nativeType;
			            valCommon -> nativeCount = pCHB -> nativeCount;
			            valCommon -> stat = 0;
			            valCommon -> sevr = 0;
			            valCommon -> ack = 0;
			            valCommon -> lastUpdate.tv_sec = 0;
			            valCommon -> lastUpdate.tv_usec = 0;
			            FA_SetValuePtr(pCHB -> chanName, caLink);
				    }

				valCommon = caLink -> pData;
				valCommon -> stat = *(short *)monMsg.dataBuffer;
				spliceStat = valCommon -> stat;
				valCommon -> sevr = *(short *)(monMsg.dataBuffer + sizeof(short));
				spliceSevr = valCommon -> sevr;
				if (ntohs(valCommon -> sevr) & 0x4000)
				    valCommon -> ack = htons(1);
				else
				    valCommon -> ack = 0;
				spliceAck = valCommon -> ack;
			        valCommon -> sevr &= htons(0x3fff);
				    
			        bcopy(monMsg.dataBuffer + sizeof(short) + sizeof(short), ((VALUE_STRING *) valCommon) -> value,
				    ntohs(monMsg.monData.chDataLen) - sizeof(short) - sizeof(short));
				}
			    }
#endif
			    /*
			    Find all the CABs that can be satisfied by this request. They will be the ones
			    where the mask in the CAB has at least one bit in common with the mask in the
			    report from the IOC. Also, all the "embellishment" bits in the CAB must be set
			    in the data type from the IOC. It's OK if additional ones are also set.
			    */

			    MARK(MOND2);
			    msgType = htons(CHAN_MONDATA);
			    mask = ntohs(monMsg.monData.chMask);
			    numCABs = findNumCABs(pCHB);
			    currCabList = &firstCabList;
			    for (cabIndex = 0; cabIndex < numCABs;)
				{
				pCAB = currCabList -> refCAB[cabIndex & 0x1f];

				if (pCAB -> pCHB)
				    {
				    if (pCAB -> dataType == AGNT_ADAPTIVE)
					{
					if (pCAB -> mask & AGNT_SPLICED)
					    pCAB -> dataType = AGNT_ENUM | AGNT_STAT;
					else
					    pCAB -> dataType = pCHB -> nativeType | AGNT_STAT;
					pCAB -> count = pCHB -> nativeCount;
					}
				    if ((mask & pCAB -> mask) 
					    && (((htons(monMsg.monData.chDataType) & pCAB -> dataType) == pCAB -> dataType)
						|| (pCAB -> mask & AGNT_SPLICED)))
				        {

				        /*
				        We have a match. Set up the response message.
				        */

				        monOutput.monData.chHandle = pCAB -> handle;
				        monOutput.monData.chDataType = htons(pCAB -> dataType);
					if (pCAB -> mask & AGNT_SPLICED)
					    monOutput.monData.chDataType = htons(AGNT_ENUM | AGNT_STAT);
				        monOutput.monData.chMask = monMsg.monData.chMask;
				        monOutput.monData.chCount = htons(pCAB -> count);
				        length2 = transformType(ntohs(monMsg.monData.chDataType), ntohs(monMsg.monData.chCount),
						       pCAB -> dataType, pCAB -> count, monMsg.dataBuffer, 
						       monOutput.dataBuffer, (short)(pCAB -> mask & AGNT_SPLICED));
				        monOutput.monData.chDataLen = htons((short)length2);
				        pCAB -> awaitMon = 0;
				        appTx(pCAB -> pACB, msgType, (char *) &monOutput, sizeof(CH_MONDATA) 
					        + length2);
				        }
				    }
				cabIndex++;
				if ((pCHB -> currType & 0xff) == AGNT_ADAPTIVE)
				    updateDataType(pCHB);
				if ((cabIndex & 0x1f) == 0)
				    currCabList = currCabList -> nextList;
				}
			    /*
			    And also, on workstations only, park this information in the tree if
			    it contains ONLY the value and status information. This is the variant
			    used by dm.
			    */
			    }

			else if (msgType == CHAN_RESMODE)
			    {
			    CH_RESMODE		resMode;
			    RCB			*pRCB;
			    CH_CONNEVNT		connEvnt;
			    CH_DISCEVNT		discEvnt2;
			    short		anyWaiting;

			    /*
			    A resource in the IOC has changed state and he is letting us know about it.
			    Get the message body and find the RCB that corresponds to this resource.
			    */

			    scmRead((char *) &resMode, sizeof(resMode));

			    pRCB = findRCB(resMode.chResName);
#if !defined(vxWorks) && !defined(vms)
			    tisSysLog(1, 0, "chanAcq: Resource mode change from %s for %s, mode %d",
					INET_NTOA(pISB -> Inet), resMode.chResName, resMode.chResStatus);
#else
			    if (cacDebug >= 1)
				printf("chanAcq: Resource mode change from %s for %s, mode %d\n", 
					INET_NTOA(pISB -> Inet), resMode.chResName, resMode.chResStatus);
#endif

			    /*
			    There is a potentially nasty situation where one node is happily master and
			    then his redundant partner grabs the resource, notices his mistake, and then
			    gives the resource back, with apologies. The redundant partner's grab and
			    then give back get reported to us, but the original master never tells us
			    that he got the resource back because, in his mind, he never lost it. This
			    means we can generate disconnects but never reconnect. Make sure this doesn't
			    happen when a resource loss is reported by looking at the RCB for this 
			    resource. If it is redundant, and the redundant partner is online, ask him 
			    about his resource state.
			    */

			    if ((pRCB -> currISB == pISB) && (!resMode.chResStatus))
				{
				msgType = htons(msgType);
				if (pRCB -> pISB1 == pISB)
				    {
				    if (pRCB -> pISB2)
				        if (pRCB -> pISB2 -> connected)
					    sendScm(pRCB -> pISB2 -> Inet, msgType,
						    (char *) &resMode, sizeof(resMode));
				    }
				else if (pRCB -> pISB1)
				    {
				    if (pRCB -> pISB1 -> connected)
				        sendScm(pRCB -> pISB1 -> Inet, msgType, (char *) &resMode,
					        sizeof(resMode));
				    }
				}

			    /*
			    Traverse all of the channels that belong to this resource. Generate connect
			    and/or disconnect events as appropriate. This is another case where we can
			    make a fairly complex decision tree to save a few lines of code here or
			    there. Instead, we will write straightforward code that has a non-negligible
			    probability of working.
			    */

			    for (pCHB = pRCB -> pFirstCHB; pCHB; pCHB = nextCHB)
				{
				nextCHB = pCHB -> pNextRes;

				/*
				Ignore CHBs that are delete pending.
				*/

				if (!(pCHB -> pFirstCAB))
				    continue;

				/*
				Don't do anything if the channel isn't connected to the IOC yet! This means
				that we don't have his handles yet. Anything we're likely to send him at
				this point would have either a zero or an old handle and make him crash.
				That would be rude.
				*/

				if (pCHB -> pISB1 == pISB)
				    {
				    if ((pCHB -> handleIOC1 == -1) || (pCHB -> handleIOC1 == 0))
					continue;
				    }
				else if ((pCHB -> handleIOC2 == -1) || (pCHB -> handleIOC2 == 0))
				    continue;

				if (resMode.chResStatus)
				    {

				    /*
				    Sender just acquired the resource. Find out what the CHB says the state
				    of the point is.
				    */

				    if (pCHB -> pCurrISB == 0)
					{

					/*
					Nobody had the point. Now look at each of the applications hooked up
					to this channel. If they already had a connection to this channel,
					generate a connection event. If any application has an open pending,
					the regular channel maintenance routines can take care of either
					issuing the direct read or acknowledge an outstanding open request.
					*/

					pCHB -> pCurrISB = pISB;
					anyWaiting = 0;
					numCABs = findNumCABs(pCHB);
					currCabList = &firstCabList;

					for (cabIndex = 0; cabIndex < numCABs;)
					    {
					    pCAB = currCabList -> refCAB[cabIndex & 0x1f];

					    if (pCAB -> pCHB)
						{
						/************* TEST! *************/
						/* It is true that we should NOT send the app a connection
						event unless he has the channel built on his end. However,
						that should NOT determine whether or not we might have to
						finish building the channel for someone ELSE. This is the
						original logic. Keep it in case we are wrong.

					        if (pCAB -> awaitAck)
						    anyWaiting = 1;
					        else
						    {
						    msgType = htons(CHAN_CONNEVNT);
						    connEvnt.chHandle = pCAB -> handle;
						    appTx(pCAB -> pACB, msgType, (char *) &connEvnt, 
							    sizeof(connEvnt));
						    }
						END CODE DELETED FOR TEST! */

						/* BEGIN SUBSTITUTED CODE */
						if (!(pCAB -> awaitAck))
						    {
						    msgType = htons(CHAN_CONNEVNT);
						    connEvnt.chHandle = pCAB -> handle;
						    appTx(pCAB -> pACB, msgType, (char *) &connEvnt, 
							    sizeof(connEvnt));
						    }
						/* END SUBSTITUTED CODE */

						/*
						Handle creating the initial sample, if desired.
						*/

						if (pCAB -> readType != AGNT_NOCONN)
						    {
						    CH_DIRREAD		dirRead3;

						    if (pCAB -> readType != AGNT_ADAPTIVE)
							{
							dirRead3.chDataType = htons(pCAB -> readType);
							dirRead3.chNumElem = htons(pCAB -> count);
							}
						    else
							{
							if (pCHB -> nativeType != AGNT_STRING)
							    dirRead3.chDataType =
								    htons((short)(pCHB -> nativeType | AGNT_GRAPH |
								    AGNT_STAT| AGNT_CNTRL));
							else
							    dirRead3.chDataType =
								    htons((short)(pCHB -> nativeType | AGNT_STAT |
								    AGNT_CNTRL));
							if (pCAB -> mask & AGNT_SPLICED)
							    dirRead3.chNumElem = htons(1);
							else
							    dirRead3.chNumElem = htons(pCHB -> nativeCount);
							}
						    if (pISB == pCHB -> pISB1)
							dirRead3.chHandle = pCHB -> handleIOC1;
						    else
							dirRead3.chHandle = pCHB -> handleIOC2;
#ifndef POINTER64
						   dirRead3.chCaHandle = (int) pCAB;
						   dirRead3.chAppHandle = (int) (pCAB -> pACB);
#else
						   dirRead3.chCaHandle = pCAB ->cabToken;
						   dirRead3.chAppHandle = pCAB -> pACB -> acbToken;
#endif
						   msgType = htons(CHAN_DIRREAD);
						   sendScm(pISB -> Inet, msgType, (char *) &dirRead3, sizeof(dirRead3));
						   }

					        if (pCAB -> awaitMon)
						    {

						    /*
						    In a just acquired resource, this CAB hasn't had an
						    initial monitor request satisfied yet. Ask the IOC
						    for one.
						    */

						    CH_MONREQ	resReq1;

						    if (pCHB -> pCurrISB == pCHB -> pISB1)
						        resReq1.chHandle = pCHB -> handleIOC1;
						    else
						        resReq1.chHandle = pCHB -> handleIOC2;
#ifndef POINTER64
						    resReq1.chCaHandle = (int) pCAB;
#else
						    resReq1.chCaHandle = pCAB -> cabToken;
#endif
						    if (pCAB -> dataType != AGNT_ADAPTIVE && !(pCAB -> mask & AGNT_SPLICED))
							{
						        resReq1.chType = htons(pCAB -> dataType);
						        resReq1.chCount = pCAB -> count;
						        resReq1.mask = pCAB -> mask;
							}
						    else
							{
						        resReq1.chType = htons((short)(pCHB -> nativeType | AGNT_STAT));
							resReq1.chCount = htons(pCHB -> nativeCount);
							resReq1.mask = htons(AGNT_VALUE | AGNT_ALARM);
							}
						    msgType = htons(CHAN_MONREQ);
						    sendScm(pISB -> Inet, msgType, (char *) &resReq1,
							    sizeof(resReq1));
						    }
						}
					    cabIndex++;
					    if ((cabIndex & 0x1f) == 0)
						currCabList = currCabList -> nextList;
					    }

					/****** BEGIN TEST DELETED CODE *******
					if (anyWaiting)
					    checkChannel(pCHB);
					****** END TEST DELETED CODE ******/

					/* BEGIN SUBSTITUTE CODE */

					checkChannel(pCHB);

					/* END SUBSTITUTE CODE */
					}

				    else if (pCHB -> pCurrISB == pISB)
					{

					/*
					He already had it. Ask for a new monitor just because we came
					back from a state where we weren't connected.
					*/

					/* DELETE CODE -- DO NOTHING IN THIS CASE
					
					This caused a race condition in dm where we were generating
					events before he was ready for them. Let the application
					decide what to do.
					*/

					}

				    else if (pCHB -> pCurrISB)
					{

					/*
					Redundant partner has it. This is almost the same as the case
					where nobody had it except that we have to generate a disconnect
					followed by a connect.
					*/

					pCHB -> pCurrISB = pISB;
					anyWaiting = 0;
					numCABs = findNumCABs(pCHB);
					currCabList = &firstCabList;

					for (cabIndex = 0; cabIndex < numCABs;)
					    {
					    pCAB = currCabList -> refCAB[cabIndex & 0x1f];

					    if (pCAB -> pCHB)
						{

						/*
						This code was deleted in the case above. I'll leave it here
						because it doesn't do anything in the new scheme of things.
						*/

					        if (pCAB -> awaitAck)
						    anyWaiting = 1;
					        else
						    {
						    }

						/* End "psuedo deletion" */

						/*
						Handle creating the initial sample, if desired.
						*/

						if (pCAB -> readType != AGNT_NOCONN)
						    {
						    CH_DIRREAD		dirRead4;

						    if (pCAB -> readType != AGNT_ADAPTIVE)
							{
							dirRead4.chDataType = htons(pCAB -> readType);
							dirRead4.chNumElem = htons(pCAB -> count);
							}
						    else
							{
							if (pCAB -> mask & AGNT_SPLICED)
							    dirRead4.chDataType =
								    htons(AGNT_ENUM | AGNT_GRAPH | AGNT_STAT |
								    AGNT_CNTRL);
							if (pCHB -> nativeType != AGNT_STRING)
							    dirRead4.chDataType =
								    htons((short)(pCHB -> nativeType | AGNT_GRAPH |
								    AGNT_STAT| AGNT_CNTRL));
							else
							    dirRead4.chDataType =
								    htons((short)(pCHB -> nativeType | AGNT_STAT |
								    AGNT_CNTRL));
							if (pCAB -> mask & AGNT_SPLICED)
							    dirRead4.chNumElem = htons(1);
							else
							    dirRead4.chNumElem = htons(pCHB -> nativeCount);
							}
						    if (pISB == pCHB -> pISB1)
							dirRead4.chHandle = pCHB -> handleIOC1;
						    else
							dirRead4.chHandle = pCHB -> handleIOC2;
#ifndef POINTER64
						   dirRead4.chCaHandle = (int) pCAB;
						   dirRead4.chAppHandle = (int) (pCAB -> pACB);
#else
						   dirRead4.chCaHandle = pCAB -> cabToken;
						   dirRead4.chAppHandle = pCAB -> pACB -> acbToken;
#endif

						   msgType = htons(CHAN_DIRREAD);
						   sendScm(pISB -> Inet, msgType, (char *) &dirRead4, sizeof(dirRead4));
						   }

					        if (pCAB -> awaitMon)
						    {

						    /*
						    In a just acquired resource, this CAB hasn't had an
						    initial monitor request satisfied yet. We probably
						    have asked the parter for one, but the other IOC
						    grabbed the resource before we got the reply. Send
						    the request again.
						    */

						    CH_MONREQ	resReq2;

						    if (pCHB -> pCurrISB == pCHB -> pISB1)
						        resReq2.chHandle = pCHB -> handleIOC1;
						    else
						        resReq2.chHandle = pCHB -> handleIOC2;
#ifndef POINTER64
						    resReq2.chCaHandle = (int) pCAB;
#else
						    resReq2.chCaHandle = pCAB -> cabToken;
#endif
						    if (pCAB -> dataType != AGNT_ADAPTIVE && !(pCAB -> mask & AGNT_SPLICED))
							{
						        resReq2.chType = htons(pCAB -> dataType);
						        resReq2.chCount = pCAB -> count;
						        resReq2.mask = pCAB -> mask;
							}
						    else
							{
						        resReq2.chType = htons((short)(pCHB -> nativeType | AGNT_STAT));
							resReq2.chCount = htons(pCHB -> nativeCount);
							resReq2.mask = htons(AGNT_VALUE | AGNT_ALARM);
							}
						    resReq2.mask = pCAB -> mask;
						    msgType = htons(CHAN_MONREQ);
						    sendScm(pISB -> Inet, msgType, (char *) &resReq2,
							    sizeof(resReq2));
						    pCAB -> awaitMon = 1;
						    }
						}
					    cabIndex++;
					    if ((cabIndex & 0x1f) == 0)
						currCabList = currCabList -> nextList;
					    }

					/* START DELETED CODE FOR TEST

					See comments in previous section about why we are doing this.

					if (anyWaiting)
					    checkChannel(pCHB);

					****** END DELETED CODE */

					/* START SUBSTITUTED CODE */

					checkChannel(pCHB);

					/* END SUBSTITUTED CODE */
					}
				    }
				
				else if (pCHB -> pCurrISB == pISB)
				    {

				    /*
				    We only care about resources that have been lost when the IOC that
				    reported the loss was the owner. We just ignore the message in the
				    other two cases (i.e., the redundant partner has it or nobody has it).
				    Send a disconnect event to everyone that had the channel open. Don't
				    do anything if an open was pending. That will take care of itself
				    when the resource is reclaimed.
				    */

				    msgType = htons(CHAN_DISCEVNT);
				    numCABs = findNumCABs(pCHB);
				    currCabList = &firstCabList;

				    for (cabIndex = 0; cabIndex < numCABs;)
					{
					pCAB = currCabList -> refCAB[cabIndex & 0x1f];

					if (pCAB -> pCHB)
					    {
					    if (!(pCAB -> awaitAck))
					        {
					        discEvnt2.chHandle = pCAB -> handle;
					        appTx(pCAB -> pACB, msgType, (char *) &discEvnt2, 
							    sizeof(discEvnt2));
					        }
					    }
					cabIndex++;
					if ((cabIndex & 0x1f) == 0)
					    currCabList = currCabList -> nextList;
					}
				    pCHB -> pCurrISB = 0;
				    }
				}

			    /*
			    Now that we've taken care of all the channels, figure out what we
			    should do about alarm and trend handlers.
			    */


			    msgType = 0;
			    if (resMode.chResStatus)
				{
			        NSQ			*tempNSQ;

				/*
				This IOC is reporting that he owns the resource. What we do
				depends on our previous idea of the resource's status. If
				nobody owned the resource, report that it's connected. If
				either this IOC or the partner had it, don't tell the trend
				or alarm handlers about it.
				*/

				if (!(pRCB -> currISB))
				    {
				    pRCB -> currISB = pISB;
				    msgType = htons(CHAN_RESMODE);
				    }

				/*
				If there are any outstanding connectionless parameter fetches,
				alarm acknowledgments, or alarm control actions for this resource, 
				do them now.
				*/

				while (pRCB -> pFirstParam)
				    {
				    short		paramType;
				    struct
					{
					CH_GETPARAMS	getParams;
					char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
					} rp;

				    tempNSQ = pRCB -> pFirstParam;
				    rp.getParams.chParmType = htons(tempNSQ -> type);
				    rp.getParams.chNameLen = strlen(tempNSQ -> name) + 1;
				    rp.getParams.chNameLen = htons(rp.getParams.chNameLen);
				    rp.getParams.chArgument = tempNSQ -> handle;
#ifndef POINTER64
				    rp.getParams.chACB = (int) (tempNSQ -> pACB);
#else
				    rp.getParams.chACB = tempNSQ -> pACB -> acbToken;
#endif
				    strcpy(rp.chanName, tempNSQ -> name);
				    paramType = htons(CHAN_GETPARAMS);
				    sendScm(pRCB -> currISB -> Inet, paramType, (char *) &rp,
					    sizeof(CH_GETPARAMS) + ntohs(rp.getParams.chNameLen));
				    pRCB -> pFirstParam = tempNSQ -> pNextNSQ;
				    deleteNSQ(tempNSQ);
				    }

				while (pRCB -> pFirstAck)
				    {
				    short		ackType;
				    struct
					{
					CH_ALARM_ACK	ack;
					char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
					} am;

				    /*
				    Send out the current alarm acknowledgment.
				    */

				    tempNSQ = pRCB -> pFirstAck;
				    am.ack.alarmTime.secPastEpoch = tempNSQ -> ackTime.secPastEpoch;
				    am.ack.alarmTime.nsec = tempNSQ -> ackTime.nsec;
				    am.ack.alarmNameLen = strlen(tempNSQ -> name) + 1;
				    am.ack.alarmNameLen = htons(am.ack.alarmNameLen);
				    strcpy(am.chanName, tempNSQ -> name);
				    ackType = htons(CHAN_ALARM_ACK);
				    sendScm(pRCB -> currISB -> Inet, ackType, (char *) &am,
					    sizeof(CH_ALARM_ACK) + ntohs(am.ack.alarmNameLen));
				    pRCB -> pFirstAck = tempNSQ -> pNextNSQ;
				    deleteNSQ(tempNSQ);
				    }

				while (pRCB -> pFirstCtl)
				    {
				    short		ctlType;
				    struct
					{
					CH_ALARM_CTL	ctl;
					char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
					} ac;


				    /*
				    Send out the current alarm control action.
				    */

				    tempNSQ = pRCB -> pFirstCtl;
				    if (tempNSQ -> type)
					ac.ctl.alarmCtlFunc = htons(1);
				    else
					ac.ctl.alarmCtlFunc = 0;
				    ac.ctl.alarmNameLen = strlen(tempNSQ -> name) + 1;
				    ac.ctl.alarmNameLen = htons(ac.ctl.alarmNameLen);
				    strcpy(ac.chanName, tempNSQ -> name);
				    ctlType = htons(CHAN_ALARM_CTL);
				    sendScm(pRCB -> currISB -> Inet, ctlType, (char *) &ac,
					    sizeof(CH_ALARM_CTL) + ntohs(ac.ctl.alarmNameLen));
				    pRCB -> pFirstCtl = tempNSQ -> pNextNSQ;
				    deleteNSQ(tempNSQ);
				    }
				}

			    else
				{

				/*
				This IOC is reporting that he lost the resource. Report the loss
				if he was the owner. Ignore the message in all other conditions.
				*/

				if (pRCB -> currISB == pISB)
				    {
				    pRCB -> currISB = 0;
				    msgType = htons(CHAN_RESMODE);
				    }
				}

			    /*
			    If the message type is non zero, then we have to report this resource
			    change to all the trend and alarm handlers that we are supporting.
			    */

			    if (msgType)
				{
				currACB = acbList;
				while (currACB)
				    {
				    nextACB = currACB -> pNextACB;
				    if (currACB -> trendHandler || currACB -> alarmHandler)
				        appTx(currACB, msgType, (char *) &resMode, sizeof(resMode));
				    if (currACB == nextACB)
					break;
				    currACB = nextACB;
				    }
				}
			    }

			else if (msgType == CHAN_SESS_REPLY)
			    {
			    CH_SESS_REPLY	sessReply;
			    CHB			*nextCHB;
			    CH_HANDLERS		handlerMsg;

			    /*
			    After an IOC came up, we asked for his session ID to see if it was just a
			    communications outage or if either one of us got restarted.
			    Start by getting the session handle and comparing it to our own.
			    */

			    MARK(SREPLY);
			    scmRead((char *) &sessReply, sizeof(sessReply));

			    pISB -> connected = 1;
			    if (sessReply.chSession == session)
				{

				/*
				Session numbers match.  Now pass over all the CHBs for this IOC and 
				start to connect channels that might have been added while the circuit 
				was down.
				*/

#if !defined(vxWorks) && !defined(vms)
				tisSysLog(1, 0, "chanAcq: Session reply from %s", INET_NTOA(pISB -> Inet));
#else
				if (cacDebug >= 1)
				    printf("chanAcq: Session reply from %s\n", INET_NTOA(pISB -> Inet));
#endif
				msgType = htons(CHAN_BUILD);
				pCHB = (CHB *) (pISB -> pFirstCHB);
				while (pCHB)
				    {
				    int		handle;

				    if (pCHB -> pISB1 == pISB)
					{
					handle = pCHB -> handleIOC1;
					if (handle == -1)
					    pCHB -> handleIOC1 = 0;
				        nextCHB = pCHB -> pNextIOC1;
					}
				    else
					{
					handle = pCHB -> handleIOC2;
					if (handle == -1)
					    pCHB -> handleIOC2 = 0;
				        nextCHB = pCHB -> pNextIOC2;
					}

				    /*
				    checkChannel and the routines that it calls won't send out a
				    build if they think that one is already pending. Our CHB is
				    marked that way. Therefore, we must take special action to
				    get another build sent out upon reconnection.
				    */

				    if (handle == -1)
					{
					updateConn(pCHB, pISB, 0, 0);
					pCHB = nextCHB;
					continue;
					}

				    /*
				    Now see if we have to generate any new connects, data type mods, or
				    perform any direct reads.
				    */

				    checkChannel(pCHB);
				    pCHB = nextCHB;
				    }
				}
			
			    else
				{
				CH_SESS_SET	sessSet;

				/*
				Session handles didn't match. Either we restared or the IOC rebooted
				since the last time we talked. Send him the new session number which 
				will make him trash all his old connections.
				*/

				msgType	= htons(CHAN_SESS_SET);
				sessSet.chSession = session;	/* Endian not important */
				sendScm(pISB -> Inet, msgType, (char *) &sessSet, sizeof(sessSet));

				/*
				Now we have to march through each connection that we had and mark
				them as not connected. Then, try to reconnect them.
				WARNING: In theory, we should never have a case where the channel
				was connected to this IOC by time we reach this state. We should get an
				IOC DOWN message before we got the IOC UP that triggered this response.
				In order to get here, either we or the IOC had to have restarted. If that 
				turns out not to be the case because of some strange combination of states,
				then we'll have to create a disconnect event for each channel that was
				in that state.
				*/

				pCHB = (CHB *) (pISB -> pFirstCHB);
				while (pCHB)
				    {
				    if (pCHB -> pCurrISB == pISB)
					{
					pCHB -> pCurrISB = 0;
					if (pCHB -> pRes)
					    pCHB -> pRes -> currISB = 0;
					}
				    if (pCHB -> pISB1 == pISB)
					{
					nextCHB = pCHB -> pNextIOC1;
					pCHB -> handleIOC1 = 0;
					}
				    else
					{
					nextCHB = pCHB -> pNextIOC2;
					pCHB -> handleIOC2 = 0;
					}
				    checkChannel(pCHB);
				    pCHB = nextCHB;
				    }
				}

			    /*
			    At this point, we've got ourselves going again. Now send him our trend
			    and alarm handler status to keep that data flowing too.
			    */

			    msgType = htons(CHAN_HANDLERS);
			    handlerMsg.chTrendHandler = htons(numTrendHandlers);
			    handlerMsg.chAlarmHandler = htons(numAlarmHandlers);
			    sendScm(pISB -> Inet, msgType, (char *) &handlerMsg, sizeof(handlerMsg));

			    /*
			    Send the ioc a request to report the state of all his resources.
			    We need this so we can keep track of them for trend and alarm
			    purposes.
			    */

			    msgType = htons(CHAN_RESINQ);
			    sendScm(pISB -> Inet, msgType, NULL, 0);
			    }

			else if (msgType == CHAN_CLOSEACK)
			    {
			    CH_CLOSEACK		closeAck;

			    /*
			    The IOC was told to close a channel. This is his acknowledgement that
			    he did so.
			    */

			    MARK(CLACK1);
			    scmRead((char *) &closeAck, sizeof(closeAck));

			    /*
			    The handle is the address of the CHB. Look this up anyway, because we never
			    believe a pointer that someone else gives us. Just because the whole world
			    is actually out to get you doesn't mean that you aren't paranoid.
			    */

			    pCHB = (CHB *) (pISB -> pFirstCHB);
			    while (pCHB)
				{
#ifndef POINTER64
				if ((int) pCHB == closeAck.chHandle)
#else
				if (pCHB -> chbToken == closeAck.chHandle)
#endif
				    break;
				if (pCHB -> pISB1 == pISB)
				    pCHB = pCHB -> pNextIOC1;
				else
				    pCHB = pCHB -> pNextIOC2;
				}
			    if (!pCHB)
				continue;
#if !defined(vxWorks) && !defined(vms)
			    tisSysLog(2, 0, "chanAcq: Close ack for %s", pCHB -> chanName);
#else
			    if (cacDebug >= 2)
				printf("chanAcq: Close ack for %s\n", pCHB -> chanName);
#endif

			    /*
			    Found the one he's talking about. Clear the appropriate connect handle. If both
			    of them are clear, we can unchain the CHB and release it.
			    */

			    MARK(CLACK2);
			    if (pCHB -> pISB1 == pISB)
				pCHB -> handleIOC1 = 0;
			    else if (pCHB -> pISB2 == pISB)
				pCHB -> handleIOC2 = 0;
			    if ((pCHB -> handleIOC1 == 0) && (pCHB -> handleIOC2 == 0))
				unchainCHB(pCHB);
			    }

			else if (msgType == CHAN_NOTFOUND)
			    {
			    CH_NOTFOUND		dpaReject;

			    /*
			    We sent a CHAN_BUILD to a Data Provider Application (DPA), and he
			    rejected it. Send a CHAN_NOTFOUND to the application
			    and then free up the CABs and then the CHB.
			    */

			    scmRead((char *) &dpaReject, sizeof(dpaReject));

			    pCHB = (CHB *) (pISB -> pFirstCHB);
			    while (pCHB)
				{
#ifndef POINTER64
				if ((int) pCHB == dpaReject.chHandle)
#else
				if (pCHB -> chbToken == dpaReject.chHandle)
#endif
				    break;
				if (pISB == pCHB -> pISB1)
				    pCHB = pCHB -> pNextIOC1;
				else
				    pCHB = pCHB -> pNextIOC2;
				}
			    if (!pCHB)
				{
#if !defined(vxWorks) && !defined(vms)
				tisSysLog(2, 0, "chanAcq: DPA reject for %x\n", dpaReject.chHandle);
#else
				if (cacDebug >= 2)
				    printf("chanAcq: DPA reject for %x\n", dpaReject.chHandle);
#endif
				continue;
				}

			    msgType = htons(CHAN_NOTFOUND);
			    while ((pCAB = pCHB -> pFirstCAB))
				{

				/*
				First, remove this one from the ACB linkage. Do that because
				if there is an error, appTx will follow the linkage and put
				our CAB onto the free list. We don't want that to happen
				while we are processing the list here.
				*/

				pCHB -> pFirstCAB = pCAB -> pNextChan;
				if (pCAB -> pPrevApp)
				    pCAB -> pPrevApp -> pNextApp = pCAB -> pNextApp;
				else
				    pCAB -> pACB -> pFirstCAB = pCAB -> pNextApp;
				dpaReject.chHandle = pCAB -> handle;
				appTx(pCAB -> pACB, msgType, (char *) &dpaReject, sizeof(dpaReject));
				deleteCAB(pCAB);
				}

			    /*
			    Now remove this empty CHB from all of its linkage.
			    */

			    unchainCHB(pCHB);
			    }

#ifndef vxWorks
			else if (msgType == CHAN_ALARM)
			    {
			    struct
				{
				CH_ALARM	alarm;
				char		dataBuff[150];
				} alarmMsg;

			    /*
			    An unsolicited alarm. Forward this message to all of the alarm handlers
			    that we know. Also, if the time is greater than the last alarm that we
			    got from this IOC, then post that time in ISB. We'll use it when we
			    reconnect if the connection ever gets disrupted and then restored.
			    */

			    scmRead((char *) &alarmMsg.alarm, sizeof(alarmMsg.alarm));
			    scmRead(alarmMsg.dataBuff, ntohs(alarmMsg.alarm.alarmDataLen));

			    if (ntohs(alarmMsg.alarm.alarmSevr) & 0x8000)
				{

				pISB -> splicable = 1;
				alarmMsg.alarm.alarmSevr &= htons(0x3fff);
				}

#if !defined(vxWorks) && !defined(vms)
			    tisSysLog(4, 0, "chanAcq: Alarm for %s", alarmMsg.dataBuff);
#else
			    if (cacDebug >= 4)
				printf("chanAcq: alarm for %s\n", alarmMsg.dataBuff);
#endif

			    /*
			    Send the alarm to the alarm handlers.
			    */

			    msgType = htons(CHAN_ALARM);
#ifndef POINTER64
			    alarmMsg.alarm.alarmHandle = (int) pISB;
#else
			    alarmMsg.alarm.alarmHandle = pISB -> isbToken;
#endif

			    currACB = acbList;
			    while (currACB)
				{
				ACB	*nextACB;

				nextACB = currACB -> pNextACB;
				if (currACB -> alarmHandler)
				    appTx(currACB, msgType, (char *) &alarmMsg, sizeof(CH_ALARM) 
					    + ntohs(alarmMsg.alarm.alarmDataLen));
				if (currACB == nextACB)
				    break;
				currACB = nextACB;
				}
			    }

			else if (msgType == CHAN_TRDATA)
			    {
			    short	tempSever;
			    struct
				{
			    	CH_TRDATA	trendData;
			    	char		dataBuff[2048];
				} msg;

			    /*
			    Unsolicited trend data. All we really have to do is to forward this message
			    to all of the trend handlers that we know about.
			    */


			    scmRead((char *) &msg.trendData, sizeof(msg.trendData));
			    scmRead(msg.dataBuff, ntohs(msg.trendData.chNameLen) + ntohs(msg.trendData.chDataLen));
			    if (ntohs(msg.trendData.chDataType) & AGNT_STAT)
				{

				/*
				Data is the point name, null terminated, then followed by 
				timed variant. Just a bit more address arithmetic to do.
				*/

				bcopy(msg.dataBuff + ntohs(msg.trendData.chNameLen) + sizeof(short), (char *) &tempSever, sizeof(tempSever));
				if (tempSever & htons(0x8000))
				    {
				    pISB -> splicable = 1;
				    msg.trendData.chDataType &= htons(0x3fff);
				    }
				}

#if !defined(vxWorks) && !defined(vms)
			    tisSysLog(5, 0, "chanAcq: Trend data for %s", msg.dataBuff);
#else
			    if (cacDebug >= 5)
				printf("chanAcq: Trend data for %s\n", msg.dataBuff);
#endif
			    currACB = acbList;
			    msgType = htons(CHAN_TRDATA);
			    while (currACB)
				{
				ACB	*nextACB;

				nextACB = currACB -> pNextACB;
				if (currACB -> trendHandler)
				    appTx(currACB, msgType, (char *) &msg, sizeof(CH_TRDATA) + 
					    ntohs(msg.trendData.chNameLen) + ntohs(msg.trendData.chDataLen));
				if (currACB == nextACB)
				    break;
				currACB = nextACB;
				}
			    }

			else if (msgType == CHAN_PARMREPLY)
			    {
			    struct
				{
				CH_PARMREPLY	parmReply;
				char		dataBuff[2048];
				} pr;

			    /*
			    Connectionless parameter data. We just forward this to the application that
			    requested it. He will parse it and invoke the callback.
			    */

			    scmRead((char *) &pr.parmReply, sizeof(pr.parmReply));
			    scmRead(pr.dataBuff, ntohs(pr.parmReply.chDataLen));

#if !defined(vms)
			    tisSysLog(5, 0, "chanAcq: Connectionless parameter reply from %s\n",
				    INET_NTOA(currISB -> Inet));
#else
			    if (cacDebug >= 5)
				printf("chanAcq: Connectionless parameter reply from %s\n", INET_NTOA(currISB -> Inet));
#endif
			    msgType = htons(CHAN_PARMREPLY);

			    /*
			    Just to make sure, validate the pointer to the ACB in case the application
			    died after requesting the data but before the response from the IOC was
			    processed.
			    */

			    currACB = acbList;
			    while (currACB)
				{
#ifndef POINTER64
				if (currACB == (ACB *)pr.parmReply.chACB)
#else
				if (currACB -> acbToken == pr.parmReply.chACB)
#endif
				    break;
				if (currACB == currACB -> pNextACB)
				    {
				    currACB = 0;
				    break;
				    }
				currACB = currACB -> pNextACB;
				}
			    if (currACB)
				appTx(currACB, msgType, (char *) &pr, sizeof(CH_PARMREPLY) +
					ntohs(pr.parmReply.chDataLen));
			    }
#endif vxWorks

			else
			    {

			    /*
			    Unknown message code. We got out of sync.
			    */

#if !defined(vxWorks) && !defined(vms)
			    tisSysLog(0, 0, "chanAcq: Unknown message code %hd from sockMnt.");
#else
			    fprintf(stderr, "chanAcq: Unknown message code %hd from sockMnt.\n", msgType);
#endif
			    continue;
			    }
			}
		    }
		  }

		/*
		See if this is a message from the name server
		*/

		else if ((namFd != INVALID_SOCKET) && FD_ISSET(namFd, &fdList))
		  {

		  /*
		  Check to see if the receive buffer needs to be expanded
		  */

		  if (!namRxTopped)
		      {
		      int	namBytesPending;

#if !defined(vxWorks) && !defined(vms)
		      tisSysLog(2, 0, "chanAcq: Message from nameServer.");
#else
		      if (cacDebug >= 2)
			printf("chanAcq: Message from nameServer.\n");
#endif
#ifdef WIN32
		      ioctlsocket(namFd, FIONREAD, &namBytesPending);
#else
		      ioctl(namFd, FIONREAD, (int)&namBytesPending);
#endif WIN32
		      if (namBytesPending == namRxSize)
			{
			namRxSize += 1024;
			if (setsockopt(namFd, SOL_SOCKET, SO_RCVBUF, (char *)&namRxSize, sizeof(namRxSize)) == -1)
			    namRxTopped = 1;
			}
		      }

		  /*
		  Once again, drain everything out of the socket and process it.
		  */

		  if (!chGetMore())
		    {

#if defined(vxWorks) || defined(vms)
		    closeNameSocket();
		    continue;
#else
#ifndef WIN32
		    kill(scmPid, 3);
#endif WIN32
		    abort();
#endif vxWorks or vms

		    }
		  
		  while (haveCompleteNameMsg())
		    {

		    /*
		    Get the message type. Remember, we may have a different endian than the name server.
		    */

		    sockRead((char *) &msgType, sizeof(msgType));
		    msgType = htons(msgType);

		    if (msgType == NAME_SRCH_REPLY)
			{

			/*
			Name Search Reply. Start by getting the header. That will let us know if the
			name was found.
			*/

			NS_SRCH_REPLY	srchReply;
			struct in_addr	inetAddr[2];
			char		resName[PVNAME_SZ + 1];
			NSQ		*prevNamQue;
			NSQ		*currNamQue;
			NSQ		*nextNamQue;
			CAB		*allocCAB1;
		        int		status1;

			MARK(NSREP);
			sockRead((char *) &srchReply, sizeof(srchReply));

			/*
			Find the Name Queue Element that corresponded to the request.

			Even though the handle in the request is the address of the Name Queue Element,
			search down the chain until we find it. We always operate in that manner just
			in case some other event causes us to remove an entity from a list. We don't
			want to ever believe a pointer that some other process gave us without first
			verifying it.
			*/

			currNamQue = nsqList;
			prevNamQue = 0;

			while (currNamQue)
			    {
#ifndef POINTER64
			    if (srchReply.nsHandle == (int) currNamQue)
#else
			    if (srchReply.nsHandle == currNamQue -> nsqToken)
#endif
				break;
			    prevNamQue = currNamQue;
			    currNamQue = currNamQue -> pNextNSQ;
			    }

			/*
			Read the rest of the data from the socket to keep in synchronization.
			*/

			srchReply.nsNumNodes = ntohs(srchReply.nsNumNodes);
			if (srchReply.nsNumNodes)
			    {
			    inetAddr[1].s_addr = 0;	/* In case IOC isn't redundant */
			    sockRead((char *) inetAddr, srchReply.nsNumNodes * sizeof(struct in_addr));
			    srchReply.nsResNameLen = ntohs(srchReply.nsResNameLen);
			    sockRead(resName, srchReply.nsResNameLen);

			    /*
			    If the name queue entry wasn't found, then we have nothing more
			    to do here.
			    */

			    if (!currNamQue)
				continue;
			    }

			else 
			    {

			    /*
			    Name wasn't found. If a resouce name was sent, read it to keep the
			    socket in synch.
			    */

			    srchReply.nsResNameLen = ntohs(srchReply.nsResNameLen);
			    if (srchReply.nsResNameLen)
			        sockRead(resName, srchReply.nsResNameLen);

			    /*
			    If we didn't find the name Queue entry, then there is nothing
			    more for us to do. This would happen if we noticed an application
			    terminate while transactions we had pending on his behalf with the
			    name server had not yet completed.
			    */

			    if (!currNamQue)
				continue;

			    /*
			    Unchain the NSQ entry that is the head of this name list from the
			    name server queue.
			    */
			
			    if (prevNamQue)
				prevNamQue -> pNextNSQ = currNamQue -> pNextNSQ;
			    else
				nsqList = currNamQue -> pNextNSQ;

			    while (currNamQue)
			        {
			        CH_NOTFOUND		notFound;

			        /*
			        Inform all the applications waiting on this name.  We don't inform 
				parameter fetches and this can't happen for either alarm acks or alarm 
				controls. The point had to exist for an alarm to happen!
			        */

				if ((currNamQue -> count & 0x7f00) != 0x7f00)
				    {
			            notFound.chHandle = currNamQue -> handle;
			            msgType = htons(CHAN_NOTFOUND);
			            appTx(currNamQue -> pACB, msgType, (char *) &notFound, sizeof(notFound));
				    }
				prevNamQue = currNamQue;
				currNamQue = currNamQue -> nameChain;
				deleteNSQ(prevNamQue);
			        }
			    continue;
			    }

			/*
			Point was found and we have an entry in the name server queue. Unchain
			the entry at the head of the name list from the queue now.
			*/

			if (prevNamQue)
			    prevNamQue -> pNextNSQ = currNamQue -> pNextNSQ;
			else
			    nsqList = currNamQue -> pNextNSQ;

			/*
			Process all of the name queue elements with this point name.
			*/

			nextNamQue = currNamQue;
			while ((currNamQue = nextNamQue))	/* One = is intentional here! */
			    {
			    ISB		*pISB1;
			    ISB		*pISB2;
			    int		spliceMask;

			    nextNamQue = currNamQue -> nameChain;
			    if ((currNamQue -> count & 0x7f00) == 0x7f00)
			        {

			        /*
				If this name search was done because of a connectionless parameter
				fetch, alarm ack, or alarm control action, handle that now.
			        Start by finding the RCB that corresponds to the point.
			        */

			        currRCB = findRCB(resName);

			        if (currNamQue -> count == 0x7fff)
			            {

				    /*
				    Parameter fetch
				    */

			            if (currRCB -> currISB)
				        {
				        struct
				            {
				            CH_GETPARAMS	getParams;
				            char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
				            } nm;

				        /*
				        He's online. Generate the request and then free up this name server
				        queue element.
				        */

				        nm.getParams.chParmType = htons(currNamQue -> type);
				        nm.getParams.chNameLen = strlen(currNamQue -> name) + 1;
				        nm.getParams.chNameLen = htons(nm.getParams.chNameLen);
				        nm.getParams.chArgument = currNamQue -> handle;
#ifndef POINTER64
				        nm.getParams.chACB = (int) (currNamQue -> pACB);
#else
				        nm.getParams.chACB = currNamQue -> pACB -> acbToken;
#endif
				        strcpy(nm.chanName, currNamQue -> name);
				        msgType = htons(CHAN_GETPARAMS);
				        sendScm(currRCB -> currISB -> Inet, msgType, (char *) &nm,
					        sizeof(CH_GETPARAMS) + ntohs(nm.getParams.chNameLen));
				        deleteNSQ(currNamQue);
				        }
			    
			            else
				        {

				        /*
				        He's not online. Put this NSQ on the front of the queue for this
				        resource. This means that parameter requests made when the resource
				        is offline will be honored in reverse order, but we never made any
				        promises about that.
				        */

				        currNamQue -> pNextNSQ = currRCB -> pFirstParam;
				        currRCB -> pFirstParam = currNamQue;
				        }
			            }


			        /*
			        Processing for alarm acks and alarm enable/disable controls is very similar
			        to that for parameter fetches. Look at that code for comments about what is
			        going on here.
			        */

			        else if (currNamQue -> count == 0x7ffe)
				    {

				    /*
				    Alarm Acknowledgment.
				    */

			            if (currRCB -> currISB)
				        {
				        struct
				            {
				            CH_ALARM_ACK	alarmAck;
				            char		pvName[PVNAME_SZ + FLDNAME_SZ + 2];
				            } nm;

				        /*
				        He's online. Generate the ack and then free up this name server
				        queue element.
				        */

				        nm.alarmAck.alarmTime.secPastEpoch = currNamQue -> ackTime.secPastEpoch;
				        nm.alarmAck.alarmTime.nsec = currNamQue -> ackTime.nsec;
				        nm.alarmAck.alarmNameLen = strlen(currNamQue -> name) + 1;
				        nm.alarmAck.alarmNameLen = htons(nm.alarmAck.alarmNameLen);
				        strcpy(nm.pvName, currNamQue -> name);
				        msgType = htons(CHAN_ALARM_ACK);
				        sendScm(currRCB -> currISB -> Inet, msgType, (char *) &nm,
					        sizeof(CH_ALARM_ACK) + ntohs(nm.alarmAck.alarmNameLen));
				        deleteNSQ(currNamQue);
				        }
			    
			            else
				        {

				        /*
				        He's not online. Put this NSQ on the front of the queue for this
				        resource. This means that parameter requests made when the resource
				        is offline will be honored in reverse order, but we never made any
				        promises about that.
				        */

				        currNamQue -> pNextNSQ = currRCB -> pFirstAck;
				        currRCB -> pFirstAck = currNamQue;
				        }
				    }

			        else if (currNamQue -> count == 0x7ffd)
				    {

				    /*
				    Alarm Enable/Disable Control.
				    */

			            if (currRCB -> currISB)
				        {
				        struct
				            {
				            CH_ALARM_CTL	alarmCtl;
				            char		pvName[PVNAME_SZ + FLDNAME_SZ + 2];
				            } nm;

				        /*
				        He's online. Generate the request and then free up this name server
				        queue element.
				        */

				        if (currNamQue -> type)
					    nm.alarmCtl.alarmCtlFunc = htons(1);
				        else
					    nm.alarmCtl.alarmCtlFunc = 0;
				        nm.alarmCtl.alarmNameLen = strlen(currNamQue -> name) + 1;
				        nm.alarmCtl.alarmNameLen = htons(nm.alarmCtl.alarmNameLen);
				        strcpy(nm.pvName, currNamQue -> name);
				        msgType = htons(CHAN_ALARM_CTL);
				        sendScm(currRCB -> currISB -> Inet, msgType, (char *) &nm,
					        sizeof(CH_ALARM_CTL) + ntohs(nm.alarmCtl.alarmNameLen));
				        deleteNSQ(currNamQue);
				        }
			    
			            else
				        {

				        /*
				        He's not online. Put this NSQ on the front of the queue for this
				        resource. This means that parameter requests made when the resource
				        is offline will be honored in reverse order, but we never made any
				        promises about that.
				        */

				        currNamQue -> pNextNSQ = currRCB -> pFirstCtl;
				        currRCB -> pFirstCtl = currNamQue;
				        }
				    }
			        continue;
			        }

			    /*
			    Check to see if this channel is ACK, SEVER, or STAT. If so, we can splice
			    it onto VAL.
			    */

			    pISB1 = findISB(inetAddr[0]);
			    pISB2 = findISB(inetAddr[1]);
			    spliceMask = 0;
			    if (pISB1 -> splicable)
				{
				spliceMask = 1;
				if (pISB2)
				    {
				    if (!(pISB2 -> splicable))
					spliceMask = 0;
				    }
				}
			    if (spliceMask)
			    	spliceMask = getSpliceMask(currNamQue -> name);


			    /*
			    We may have sent this name to the name server two (or even more) times before
			    the first answer came back. Look this name up in the CHBs to see if that
			    happened.
			    */

			    status1 = nameFind(currNamQue -> name, &pCHB, currNamQue -> type, spliceMask);
			    if (status1 <= 0)
			        {

			        /*
			        Point is either brand new or a new data type for an existing point. Create
			        a new CHB from scratch.
			        */

			        pCHB = newCHB();
			        strcpy(pCHB -> chanName, currNamQue -> name);
			        pCHB -> pISB1 = pISB1;
			        pCHB -> pISB2 = pISB2;

				/*
				If this is ACK, STAT, or SEVR and we are splicing it onto VAL,
				then we should make it the native data type for VAL. This is
				the way that the channel might be most likely opened in the
				future!
				*/

				if (spliceMask & AGNT_SPLICED)
				    pCHB -> currType = AGNT_ADAPTIVE;
				else
				    pCHB -> currType = currNamQue -> type;

			        pCHB -> currCount = currNamQue -> count;
			        pCHB -> pNextIOC1 = (struct CHDR *) (pCHB -> pISB1 -> pFirstCHB);
			        pCHB -> pISB1 -> pFirstCHB = pCHB;
			        if (pCHB -> pISB2)
				    {
				    pCHB -> pNextIOC2 = pCHB -> pISB2 -> pFirstCHB;
				    pCHB -> pISB2 -> pFirstCHB = pCHB;
				    }
			        if (pCHB -> pNextIOC1)
				    /* LINKAGE! */
				    {
				    if (pCHB -> pNextIOC1 -> pISB1 == pCHB -> pISB1)
				        pCHB -> pNextIOC1 -> pPrevIOC1 = pCHB;
				    else
				        pCHB -> pNextIOC1 -> pPrevIOC2 = pCHB;
				    }
			        if (pCHB -> pNextIOC2)
				    {
				    if (pCHB -> pNextIOC2 -> pISB1 == pCHB -> pISB2)
				        pCHB -> pNextIOC2 -> pPrevIOC1 = pCHB;
				    else
				        pCHB -> pNextIOC2 -> pPrevIOC2 = pCHB;
				    }
			        pCHB -> pRes = findRCB(resName);
			        pCHB -> pNextRes = pCHB -> pRes -> pFirstCHB;
			        pCHB -> pRes -> pFirstCHB = pCHB;
			        if (pCHB -> pNextRes)
				    pCHB -> pNextRes -> pPrevRes = pCHB;
			        pCHB -> pHash = pHash + hash(pCHB -> chanName);
			        pCHB -> pNextHash = pCHB -> pHash -> pFirstCHB;
			        pCHB -> pHash -> pFirstCHB = pCHB;
			        pCHB -> checksum = ntohs(srchReply.nsChecksum);
			        pCHB -> pointType = ntohs(srchReply.nsPointType);
			        pCHB -> creationTime = ntohl(srchReply.nsCreationTime);
			        }

			    /*
			    Attach a new CAB to the CHB
			    */

			    allocCAB1 = newCAB();
			    allocCAB1 -> awaitAck = 1;
			    if (currNamQue -> type == AGNT_ADAPTIVE)
			        {
			        allocCAB1 -> awaitMon = 1;
			        allocCAB1 -> mask = AGNT_VALUE | AGNT_ALARM;
			        }
			    allocCAB1 -> mask |= spliceMask;
			    allocCAB1 -> pCHB = pCHB;
			    allocCAB1 -> pACB = currNamQue -> pACB;
			    allocCAB1 -> handle = currNamQue -> handle;
			    allocCAB1 -> dataType = currNamQue -> type;
			    allocCAB1 -> readType = currNamQue -> type;
			    allocCAB1 -> count = currNamQue -> count;
			    allocCAB1 -> pNextChan = pCHB -> pFirstCAB;
			    pCHB -> pFirstCAB = allocCAB1;
			    allocCAB1 -> pNextApp = allocCAB1 -> pACB -> pFirstCAB;
			    if (allocCAB1 -> pNextApp)
			        allocCAB1 -> pNextApp -> pPrevApp = allocCAB1;
			    allocCAB1 -> pACB -> pFirstCAB = allocCAB1;
    
			    /*
			    Create the channel or change data type as necessary.
			    */

			    checkChannel(pCHB);

			    /*
			    And finally, we're done with this entry.
			    */

			    deleteNSQ(currNamQue);
			    }
			}
		    }
		  }

		/*
		None of the above -- look this one up in our list of applications.
		*/

		else if (acbList)
		    {
		    
		    /*
		    Why do we have the variable nextACB instead of just saying something like
		    currACB = currACB -> pnextACB in the final clause of the following "for"
		    statement? The reason (found during testing) is that during the loop we
		    might well decide to drop the current ACB out of the list. In that case we might
		    follow a link into oblivion, causing us to either crash or to not properly
		    process all of the ACBs in the chain.
		    */

		    MARK(APSTRT);
		    nextACB = acbList -> pNextACB;
		    for (currACB = acbList; currACB; currACB = nextACB)
			{
			nextACB = currACB -> pNextACB;
			if (nextACB == currACB)
			    nextACB = 0;
			if ((currACB ->sockFd != INVALID_SOCKET) && FD_ISSET(currACB -> sockFd, &fdList))
			  {

			  /*
			  See if the receive buffer for this socket is adequate.
			  */

			  if (!(currACB -> appRxTopped))
			      {
			      int	appBytesPending;

#ifdef WIN32
			      ioctlsocket(currACB -> sockFd, FIONREAD, &appBytesPending);
#else
			      ioctl(currACB -> sockFd, FIONREAD, (int)&appBytesPending);
#endif WIN32
			      if (appBytesPending == currACB -> appRxSize)
				{
				currACB -> appRxSize += 1024;
				if (setsockopt(currACB -> sockFd, SOL_SOCKET, SO_RCVBUF, (char *)&(currACB -> appRxSize),
					sizeof(currACB -> appRxSize)) == -1)
				    currACB -> appRxTopped = 1;
				}
			      }

			  /*
			  Same old technique. Read all there is to get and process it one message
			  at a time. Loop where there are complete messages in the buffer.
			  */

			  if (!fillAppBuff(currACB))
			    {
			    delApp(currACB);
			    continue;
			    }

			  while (haveCompleteAppMsg(currACB))
			    {

			    /*
			    Found one -- get the message type and see what he wants
			    */

			    appRx(currACB, (char *) &msgType, sizeof(msgType));

			    if (msgType == CHAN_RGST)		/* Check before flipping! */
				{

				/*
				Socket Creation/Maintenance registers with us. Delete
				this ACB and remember his fd for special treatment later.
				*/

				MARK(RGST);
#if !defined(vxWorks) && !defined(vms)
				tisSysLog(1, 0, "chanAcq: sockMnt registration.");
#else
				if (cacDebug >= 1)
				    printf("chanAcq: sockMnt registration.\n");
#endif
				scmFd = currACB -> sockFd;
				scmRxSize = currACB -> appRxSize;
				scmTxSize = currACB -> appTxSize;
				delApp(currACB);		/* Should be empty! */
#ifndef vxWorks
				FD_ZERO(&scmMask);
				FD_SET(scmFd, &scmMask);
#ifndef vms
				{
				FILE	*gotcha;

				if ((gotcha = fopen("/tis4000/tasks/scM/sockMnt.lock", "r")) == NULL)
				    tisSysLog(0, 0, "chanAcq get pid of sockMnt\n");
				else
				    {
				    fscanf(gotcha, "%d", &scmPid);
				    tisSysLog(1, 0, "chanAcq: sockMnt's pid is %d\n", scmPid);
				    fclose(gotcha);
				    }
				}
#endif vms
#endif vxWorks

#if defined(vxWorks) || defined(vms)

				/*
				In vxWorks and vms, let sockMnt know that we need a connection to
				the name server. closeNameSocket() is smart enough not to close
				the existing socket when there isn't one.
				*/

				closeNameSocket();
#endif vxWorks or vms

				unregistered = 0;
				continue;
				}

			    else
				{

				/*
				Message is from a channel access application. Do what is indicated
				by the message type.
				*/

			        msgType = ntohs(msgType);	/* He might be different from us */

				if (msgType == DUMP)
				    {
				    ISB		*pDumpISB;
				    CHB		*pDumpCHB;
				    CAB		*pDumpCAB;
				    ACB		*pDumpACB;
				    FILE	*fd;

				    /*
				    Dump all our control blocks in a controlled manner.
				    */

				    MARK(DUMP);
				    printf("DUMP\n");
				    if ((fd = fopen("caqdump", "w")) == NULL)
					continue;
				    fprintf(fd, "CONTROL BLOCK DUMP BY ISB:\n\n");
				    pDumpISB = isbList;
				    while (pDumpISB)
					{
					dumpISB(fd, pDumpISB);
					pDumpCHB = pDumpISB -> pFirstCHB;
					while (pDumpCHB)
					    {
					    dumpCHB(fd, pDumpCHB);
					    pDumpCAB = pDumpCHB -> pFirstCAB;
					    while (pDumpCAB)
						{
						dumpCAB(fd, pDumpCAB);
						pDumpCAB = pDumpCAB -> pNextChan;
						}
					    if (pDumpISB == pDumpCHB -> pISB1)
						pDumpCHB = pDumpCHB -> pNextIOC1;
					    else if (pDumpISB == pDumpCHB -> pISB2)
						pDumpCHB = pDumpCHB -> pNextIOC2;
					    else
						{
						fprintf(fd, "\tTHIS CHB DOESN'T POINT BACK TO ITS ISB! ***\n");
						break;
						}
					   }
					pDumpISB = pDumpISB -> nextISB;
					}

				    fprintf(fd, "CONTROL BLOCK DUMP BY ACB:\n\n");
				    pDumpACB = acbList;
				    while(pDumpACB)
					{
					dumpACB(fd, pDumpACB);
					pDumpCAB = pDumpACB -> pFirstCAB;
					while (pDumpCAB)
					    {
					    dumpCAB(fd, pDumpCAB);
					    pDumpCHB = pDumpCAB -> pCHB;
					    dumpCHB(fd, pDumpCHB);
					    pDumpCAB = pDumpCAB -> pNextApp;
					    }
					if (pDumpACB == pDumpACB -> pNextACB)
					    break;
					pDumpACB = pDumpACB -> pNextACB;
					}
				    fclose(fd);
				    }

				else if (msgType == CHAN_BUILD)
				    {
				    CH_BUILD	chBuild;
				    char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];
#ifndef vxWorks
				    char	spliceName[PVNAME_SZ + FLDNAME_SZ + 2];
#endif
				    int		spliceMask = 0;
				    int		status2;
				    CAB		*allocCAB2;

				    /*
				    Channel Build. This is caused by an application calling the function
				    ca_build_and_connect() or any one of its variants.
				    Get the rest of the header and then the name.
				    */

				    MARK(BUILD);
				    appRx(currACB, (char *) &chBuild, sizeof(chBuild));
				    msgType = ntohs(chBuild.chNameLen);
				    appRx(currACB, chanName, (int) msgType);
#if !defined(vxWorks) && !defined(vms)
				    tisSysLog(2, 0, "chanAcq: build %s\n", chanName);
#else
				    if (cacDebug >= 2)
					printf("chanAcq: build %s\n", chanName);
#endif


				    /*
				    Get the name into a standard format.
				    */

				    nameFix(chanName);
				    chBuild.chType = ntohs(chBuild.chType);

#ifndef vxWorks
				    /*
				    In non-vxWorks systems, see if we can splice this channel onto
				    an existing one. Handle it the old way if we don't know where
				    the point is.
				    */


				    strcpy(spliceName, chanName);
				    spliceMask = getSpliceMask(spliceName);
				    status2 = nameFind(spliceName, &pCHB, chBuild.chType, spliceMask);
				    if (spliceMask && status2 >= 0)
					{

					/*
					It is possible that we can splice this point!
					We know where the point is. See if ALL IOCs that
					hold the point are splicable.
					*/

					if (pCHB -> pISB1 -> splicable)
					    {
					    if (pCHB -> pISB2)
						{
						if (!(pCHB -> pISB2 -> splicable))
						    {

						    /*
						    We could have spliced but the second IOC can't.
						    If we're going to splice, BOTH IOCs must be
						    splicable.
						    */

						    spliceMask = 0;
						    }
						}
					    }
					else
					    {

					    /*
					    We could have spliced but the first IOC can't. We'll have to
					    do it the old way.
					    */

					    spliceMask = 0;
					    }
					}
				    else
					{
					spliceMask = 0;		/* Can't splice onto what we don't have! */
					}

				    if (spliceMask)
					{
					struct
					    {
					    CH_MONDATA		monData;
					    char		spliceStat[2048];
					    } spliceMon;
					CALINK			*caLink;
					VALUE_COMMON		*valCommon;

					/*
					We can and will splice. If the name search was not a perfect
					match (in the splicing sense) then we'll create a new channel
					in the old manner but with the name possibly changed. Let the
					bits in spliceMask stay!
					*/



					if (status2 == 0)
					    strcpy(chanName, spliceName);
					else
					    {
					    CH_BLDACK	bldAck3;

					    /*
					    We are splicing onto an existing channel pointed to by pCHB.
					    Create a new CAB and attach it to that CHB and the ACB.
					    */

					    allocCAB2 = newCAB();
					    allocCAB2 -> pCHB = pCHB;
					    allocCAB2 -> awaitAck = 1;
					    allocCAB2 -> pACB = currACB;
					    allocCAB2 -> handle = chBuild.chHandle;
					    allocCAB2 -> pNextChan = pCHB -> pFirstCAB;
					    pCHB -> pFirstCAB = allocCAB2;
					    allocCAB2 -> pNextApp = currACB -> pFirstCAB;
					    if (allocCAB2 -> pNextApp)
						allocCAB2 -> pNextApp -> pPrevApp = allocCAB2;
					    currACB -> pFirstCAB = allocCAB2;
					    allocCAB2 -> mask = spliceMask;
					    allocCAB2 -> dataType = chBuild.chType;
					    allocCAB2 -> readType = chBuild.chType;
					    allocCAB2 -> count = ntohs(chBuild.chCount);
					    if (chBuild.chType == AGNT_ADAPTIVE)
						{
						allocCAB2 -> awaitMon = 1;
						allocCAB2 -> mask |= AGNT_VALUE | AGNT_ALARM;
						allocCAB2 -> count = 1;
						}

					    /*
					    If the channel we are attached to is online and fully built, do the BLDACK
					    now. Otherwise, we have to wait.
					    */

					    if (!pCHB -> pCurrISB)
						continue;
					    if (pCHB -> pCurrISB == pCHB -> pISB1)
						{
						if ((pCHB -> handleIOC1 == 0) || (pCHB -> handleIOC1 == -1))
						continue;
						}
					    else if ((pCHB -> handleIOC2 == 0) || (pCHB -> handleIOC2 == -1))
						continue;

					    caLink = (CALINK *) FA_GetValuePtr(spliceName);
					    if (!caLink)
						continue;
					    valCommon = (VALUE_COMMON *) (caLink -> pData);
					    bldAck3.chHandle = allocCAB2 -> handle;
					    bldAck3.chIocHandle = (int) allocCAB2;
					    bldAck3.currState = 1;
					    if (spliceMask & AGNT_SPLICED)
					        {
					        bldAck3.chDataType = htons(AGNT_ENUM);
					        bldAck3.chElementCount = htons(1);
					        }
					    else
					        {
			    		        bldAck3.chDataType = htons(pCHB -> nativeType);
			    		        bldAck3.chElementCount = htons(pCHB -> nativeCount);
					        }
					    msgType = htons(CHAN_BLDACK);
					    appTx(currACB, msgType, (char *) &bldAck3, sizeof(bldAck3));
					    allocCAB2 -> awaitAck = 0;


					    /*
					    And if there is an initial read or the ADAPTIVE data type,
					    we can also do that now, too!
					    */

					    if (allocCAB2 -> dataType == AGNT_ADAPTIVE)
					        {
					        short			spliceChoice;
					        short			spliceEnumVal;
					        short			index;
					        char			*where;
						struct choiceSet	*pChoiceSet;
					        struct
						    {
						    CH_DRRDATA		drrData;
						    char		spliceDummy[2048];
						    } spliceRead;


						if (spliceMask & AGNT_SPLICED)	/* An enum */
						    {
						    if (spliceMask & AGNT_SPL_STAT)
						        {
						        spliceChoice = GBL_ALARM_STAT;
						        spliceEnumVal = valCommon -> stat;
						        }
						    else if (spliceMask & AGNT_SPL_SEVR)
						        {
						        spliceChoice = GBL_ALARM_SEV;
						        spliceEnumVal = valCommon -> sevr;
						        }
						    else if (spliceMask & AGNT_SPL_ACK)
						        {
						        spliceChoice = GBL_YES_NO;
						        spliceEnumVal = valCommon -> ack;
						        }
						    pChoiceSet = GET_PCHOICE_SET(choiceGbl, spliceChoice);
						    spliceRead.drrData.chHandle = allocCAB2 -> handle;;
						    spliceRead.drrData.chDataType = htons(AGNT_ENUM | AGNT_STAT | AGNT_CNTRL);
						    spliceRead.drrData.chCount = htons(1);
						    *(short *) (spliceMon.spliceStat) = htons(valCommon -> stat);
						    *(short *) (spliceMon.spliceStat + sizeof(short)) = htons(valCommon -> sevr);
						    *(short *) spliceRead.spliceDummy = *(short *) spliceMon.spliceStat;
						    *(short *) (spliceRead.spliceDummy + sizeof(short)) =
							    *(short *) (spliceMon.spliceStat + sizeof(short));
						    where = spliceRead.spliceDummy + sizeof(short) + sizeof(short);
						    spliceChoice = (short)pChoiceSet -> number;
						    spliceChoice = htons(spliceChoice);
						    *(short *) where = spliceChoice;
						    where += sizeof(short);
						    for (index = 0; index < pChoiceSet -> number; index++)
						        {
						        strcpy(where, GET_CHOICE(pChoiceSet, index));
						        where += strlen(where) + 1;
						        }
						    bcopy((char *) &spliceEnumVal, where, sizeof(short));
						    where += sizeof(short);
						    spliceRead.drrData.chDataLen = htons((short)(where - spliceRead.spliceDummy));
						    msgType = htons(CHAN_DRRDATA);
						    appTx(currACB, msgType, (char *) &spliceRead, sizeof(CH_DRRDATA) +
							    where - spliceRead.spliceDummy);

						    spliceMon.monData.chHandle = allocCAB2 -> handle;
						    allocCAB2 -> dataType = AGNT_ENUM | AGNT_STAT;
						    allocCAB2 -> readType = AGNT_NOCONN;
						    spliceMon.monData.chDataType = htons(AGNT_ENUM | AGNT_STAT);
						    spliceMon.monData.chDataLen = htons(sizeof(short) + sizeof(short) + sizeof(short));
						    spliceMon.monData.chMask = htons(AGNT_VALUE | AGNT_LOG);
						    spliceMon.monData.chCount = htons(1);
						    *(short *) (spliceMon.spliceStat + sizeof(short) + sizeof(short)) = spliceEnumVal;
						    msgType = htons(CHAN_MONDATA);
						    appTx(currACB, msgType, (char *) &spliceMon, sizeof(CH_MONDATA) +
							    sizeof(short) + sizeof(short) + sizeof(short));
						    allocCAB2 -> awaitMon = 0;
						    }
						else
						    {
						    short		dataLen;
						    short		atShort;
						    int			atLong;
						    float		atFloat;
						    double		atDouble;
						    short		moveCnt = pCHB -> nativeCount;

						    /*
						    It's the value. Get all the stuff we need out of the local database.
						    */

						    getAttributes(pCHB -> chanName, pCHB -> pointType, ATTR_NUM_STRS,
							    attrStrings, ATTR_STR_LEN, pCHB -> checksum,
							    pCHB -> creationTime, &attrNativeType, &attrNativeCount,
							    &attrHopr, &attrLopr, &attrDrvh, &attrDrvl, &attrPrec,
							    &attrNumStrings);
			    			    attrNativeType = attrMap[attrNativeType];

						    /*
						    Get stat and sevr into the monitor buffer.
						    */

						    *(short *) (spliceMon.spliceStat) = htons(valCommon -> stat);
						    *(short *) (spliceMon.spliceStat + sizeof(short)) = htons(valCommon -> sevr);
						    bcopy(spliceMon.spliceStat, spliceRead.spliceDummy, sizeof(short) + sizeof(short));

						    /*
						    Move in all the declarations for the read.
						    */

			    			    where = spliceRead.spliceDummy + sizeof(short) + sizeof(short);

			    			    switch (attrNativeType)
						        {
						        case AGNT_STRING:

							    /*
							    Strings are screwy with all kinds of additional data that is
							    not used by the people that do adaptive data type reads. The
							    actual string data is 44 bytes past the start of the data
							    that we have recorded. Null that stuff out. The handling of
							    the char ptr "where" is different for the case of strings
							    because of that add'l data.
							    */

							    strcpy(spliceMon.spliceStat + sizeof(short) + sizeof(short) + 44,
								    ((VALUE_STRING *) valCommon) -> value + 44);
							    dataLen = strlen(spliceMon.spliceStat + sizeof(short) + sizeof(short) + 44);
							    dataLen += 44;
				    			    memset(spliceMon.spliceStat + sizeof(short) + sizeof(short), 0, 44);
				    			    break;
			
						        case AGNT_CHAR:
							    {
							    char	*source = &(((VALUE_CHAR *) valCommon) -> value);
							    char	*dest = (char *)(spliceMon.spliceStat + sizeof(short) + sizeof(short));

							    while (moveCnt--)
								*dest++ = *source++;
							    }

							    dataLen = sizeof(char);
				    			    *where++ = (char) attrHopr;
				    			    *where++ = (char) attrLopr;
				    			    *where++ = (char) attrDrvh;
				    			    *where++ = (char) attrDrvl;
				    			    break;

						        case AGNT_SHORT:
							    {
							    short	*source = &(((VALUE_SHORT *) valCommon) -> value);
							    short	*dest = (short *)(spliceMon.spliceStat + sizeof(short) + sizeof(short));

							    while (moveCnt--)
								*dest++ = *source++;
							    }

							    dataLen = sizeof(short);
				    			    atShort = htons(((short) attrHopr));    /* Lotsa parens in case htons is a macro! */
				    			    bcopy((char *) &atShort, where, sizeof(short));
				    			    where += sizeof(short);
				    			    atShort = htons(((short) attrLopr));
				    			    bcopy((char *) &atShort, where, sizeof(short));
				    			    where += sizeof(short);
				    			    atShort = htons(((short) attrDrvh));
				    			    bcopy((char *) &atShort, where, sizeof(short));
				    			    where += sizeof(short);
				    			    atShort = htons(((short) attrDrvl));
				    			    bcopy((char *) &atShort, where, sizeof(short));
				    			    where += sizeof(short);
				    			    break;

						        case AGNT_LONG:			/* In CA, "longs" are really ints! */
							    {
							    int		*source = &(((VALUE_LONG *) valCommon) -> value);
							    int		*dest = (int *)(spliceMon.spliceStat + sizeof(short) + sizeof(short));

							    while (moveCnt--)
								*dest++ = *source++;
							    }

							    dataLen = sizeof(int);
				    			    atLong = htonl(((int) attrHopr));
				    			    bcopy((char *) &atLong, where, sizeof(int));
				    			    where += sizeof(int);
				    			    atLong = htonl(((int) attrLopr));
				    			    bcopy((char *) &atLong, where, sizeof(int));
				    			    where += sizeof(int);
				    			    atLong = htonl(((int) attrDrvh));
				    			    bcopy((char *) &atLong, where, sizeof(int));
				    			    where += sizeof(int);
				    			    atLong = htonl(((int) attrDrvl));
				    			    bcopy((char *) &atLong, where, sizeof(int));
				    			    where += sizeof(int);
				    			    break;

							case AGNT_FLOAT:
							    {
							    float	*source = &(((VALUE_FLOAT *) valCommon) -> value);
							    float	*dest = (float *)(spliceMon.spliceStat + sizeof(short) + sizeof(short));

							    while (moveCnt--)
								*dest++ = *source++;
							    }

							    dataLen = sizeof(float);
			    				    atFloat = (float) attrHopr;
			    				    htonf(atFloat, atFloat);
			    				    bcopy((char *) &atFloat, where, sizeof(float));
				    			    where += sizeof(float);
				    			    atFloat = (float) attrLopr;
				    			    htonf(atFloat, atFloat);
							    bcopy((char *) &atFloat, where, sizeof(float));
			    				    where += sizeof(float);
			    				    atShort = htons(attrPrec);
			    				    bcopy((char *) &atShort, where, sizeof(short));
				    			    where += sizeof(short);
				    			    atFloat = (float) attrDrvh;
				    			    htonf(atFloat, atFloat);
				   			    bcopy((char *) &atFloat, where, sizeof(float));
				   			    where += sizeof(float);
				    			    atFloat = (float) attrDrvl;
				  			    htonf(atFloat, atFloat);
				    			    bcopy((char *) &atFloat, where, sizeof(float));
							    where += sizeof(float);
{
float	crashFloat;
bcopy(spliceMon.spliceStat + sizeof(short) + sizeof(short), (char *) &crashFloat, sizeof(crashFloat));
if ((crashFloat > 1.0E29) || (crashFloat < -1.0E29))
	printf("BAD FLOAT 2!\n");
}
			    				    break;

						        case AGNT_DOUBLE:
							    {
							    double	*source = &(((VALUE_DOUBLE *) valCommon) -> value);
							    double	*dest = (double *)(spliceMon.spliceStat + sizeof(short) + sizeof(short));

							    while (moveCnt--)
								*dest++ = *source++;
							    }

							    dataLen = sizeof(double);
				    			    htond(attrHopr, atDouble);
				    			    bcopy((char *) &atDouble, where, sizeof(double));
				    			    where += sizeof(double);
				    			    htond(attrLopr, atDouble);
				   			    bcopy((char *) &atDouble, where, sizeof(double));
				   			    where += sizeof(double);
				    			    atShort = htons(attrPrec);
				    			    bcopy((char *) &atShort, where, sizeof(double));
				    			    where += sizeof(short);
				    			    htond(attrDrvh, atDouble);
				    			    bcopy((char *) &atDouble, where, sizeof(double));
				    			    where += sizeof(double);
				    			    htond(attrDrvl, atDouble);
				    			    bcopy((char *) &atDouble, where, sizeof(double));
				    			    where += sizeof(double);
{
double	crashDouble;
bcopy(spliceMon.spliceStat + sizeof(short) + sizeof(short), (char *) &crashDouble, sizeof(crashDouble));
if ((crashDouble > 1.0E29) || (crashDouble < -1.0E29))
	printf("BAD DOUBLE 2!\n");
}
				    			    break;

						        case AGNT_ENUM:
							    {
							    short	*source = &(((VALUE_ENUM *) valCommon) -> value);
							    short	*dest = (short *)(spliceMon.spliceStat + sizeof(short) + sizeof(short));

							    while (moveCnt--)
								*dest++ = *source++;
							    }

							    dataLen = sizeof(short);
				    			    atShort = htons(attrNumStrings);
				    			    bcopy((char *) &atShort, where, sizeof(short));
				    			    where += sizeof(short);
							    for (atShort = 0; atShort < attrNumStrings; atShort++)
							        {
							        strcpy(where, attrStrings[atShort]);
							        where += (strlen(where) + 1);
							        }
				    			    break;
						        }

			    			    /*
			    			    Now move in the value and set the pointer past the end of it.
			    			    */

						    dataLen *= pCHB -> nativeCount;
						    bcopy(spliceMon.spliceStat + sizeof(short) + sizeof(short), where, dataLen);
			    			    where += dataLen;

			    			    /*
			    			    Now fill in the rest of the fields for the DRRDATA, and MONDATA messages.
			    			    */

						    if (allocCAB2 -> mask & AGNT_SPLICED)
							allocCAB2 -> dataType = AGNT_ENUM | AGNT_STAT;
						    else
						        allocCAB2 -> dataType = pCHB -> nativeType | AGNT_STAT;
						    allocCAB2 -> readType = AGNT_NOCONN;
						    spliceRead.drrData.chHandle = allocCAB2 -> handle;
						    spliceRead.drrData.chDataType = ntohs((short)(allocCAB2 -> dataType | AGNT_STAT | AGNT_GRAPH | AGNT_CNTRL));
						    spliceRead.drrData.chDataLen = ntohs((short)(where - spliceRead.spliceDummy));
						    spliceRead.drrData.chCount = htons(pCHB -> nativeCount);
						    msgType = htons(CHAN_DRRDATA);
						    appTx(currACB, msgType, (char *)&spliceRead, sizeof(CH_DRRDATA) + where - spliceRead.spliceDummy);
						    spliceMon.monData.chHandle = allocCAB2 -> handle;
						    spliceMon.monData.chDataType = htons(allocCAB2 -> dataType);
						    spliceMon.monData.chDataLen = htons((short)(dataLen + sizeof(short) + sizeof(short)));
						    spliceMon.monData.chMask = htons((AGNT_VALUE | AGNT_LOG));
			    			    spliceMon.monData.chCount = htons(pCHB -> nativeCount);
						    msgType = htons(CHAN_MONDATA);
						    appTx(currACB, msgType, (char *)&spliceMon, sizeof(CH_MONDATA) + dataLen + sizeof(short) 
							    + sizeof(short));
						    allocCAB2 -> awaitMon = 0;
						    }
						}
					    continue;
					    }
					}
#endif

				    /*
				    See if we already have this name. There are three possiblities.
				    Either we've never seen the name before, or we've seen the name
				    but not the same field & data type, or we find a perfect match.
				    */

				    status2 = nameFind(chanName, &pCHB, chBuild.chType, spliceMask);
				    if (status2 < 0)
					{
					NSQ		*nsQueElem;
					int		pointFound = 0;
					struct
					    {
					    NS_SRCH		nsSrch;
					    char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
					    } nsMsg;

					/*
					This point has never been seen before. Send it off to the
					name server, keeping a copy of the request locally. We'll
					take action when the response comes back from the name
					server.
					*/

#ifdef vxWorks
					if ((namFd == INVALID_SOCKET) && (caMustResolve == 0))
					    {
					    struct dbAddr	dbAddr;

					    if (cacDebug >= 2)
					        printf("chanAcq: Local resolution for %s\n", chanName);
					    dbNameToAddr(chanName, &dbAddr);
					    if ((int) dbAddr.precord < 0 || (int) dbAddr.pfield < 0)
					        {
					        CH_NOTFOUND	noName;

					        /*
					        Not found locally. Send rejection back to app.
					        */

					        noName.chHandle = chBuild.chHandle;
					        msgType = htons(CHAN_NOTFOUND);
					        appTx(currACB, msgType, (char *) &noName, sizeof(noName));
					        continue;
					        }

					    /*
					    Create a local, non-redundant channel

					    NOTE: WE HAVEN'T PUT SUPPORT HERE FOR ADAPTIVE DATA TYPES!
					    It shouldn't be that big a deal, but that type of channel
					    usage is peculiar to workstation programs, not to sequence
					    programs or CA applications running in an IOC.
					    */

					    pCHB = newCHB();
					    strcpy(pCHB -> chanName, chanName);
					    pCHB -> currType = chBuild.chType;	/* Fixed endian earlier! */
					    pCHB -> currCount = ntohs(chBuild.chCount);
					    pCHB -> pISB1 = findISB(caLocalInet);
					    pCHB -> pRes = findRCB(((struct dbCommon *) dbAddr.precord) -> resn);
					    pCHB -> pNextIOC1 = (struct CHDR *) (pCHB -> pISB1 -> pFirstCHB);
					    pCHB -> pISB1 -> pFirstCHB = pCHB;
					    if (pCHB -> pNextIOC1)
					        {
					        if (pCHB -> pNextIOC1 -> pISB1 == pCHB -> pISB1)
						    pCHB -> pNextIOC1 -> pPrevIOC1 = pCHB;
					        else
						    pCHB -> pNextIOC1 -> pPrevIOC2 = pCHB;
					        }
					    pCHB -> pNextRes = pCHB -> pRes -> pFirstCHB;
					    pCHB -> pRes -> pFirstCHB = pCHB;
					    if (pCHB -> pNextRes)
					        pCHB -> pNextRes -> pPrevRes = pCHB;
					    pCHB -> pHash = pHash + hash(pCHB -> chanName);
					    pCHB -> pNextHash = pCHB -> pHash -> pFirstCHB;
					    pCHB -> pHash -> pFirstCHB = pCHB;
					    }
				        else
					    {
#endif vxWorks


#if !defined(vxWorks) && !defined(vms)
					tisSysLog(2, 0, "\tNew name sent to nameSrv.");
#else
					if (cacDebug >= 2)
					    printf("\tNew name sent to nameSrv.\n");
#endif
					
					/*
					Find the end of the list for two reasons. First, this will
					minimize searching when the answer comes back. Second, this
					will keep the queue elements in time order for the vxWorks and vms
					environments so we can determine when the nameServer has died.
					*/

					if ((nsQueElem = sameName(chanName)))
					    {
					    NSQ		*newElement = newNSQ();

					    pointFound = 1;
					    newElement -> nameChain = nsQueElem -> nameChain;
					    nsQueElem -> nameChain = newElement;
					    nsQueElem = newElement;
					    }

					else if ((nsQueElem = nsqList))
					    {
					    while (nsQueElem -> pNextNSQ)
						nsQueElem = nsQueElem -> pNextNSQ;
					    nsQueElem -> pNextNSQ = newNSQ();
					    nsQueElem = nsQueElem -> pNextNSQ;
					    }
					else
					    {
					    nsqList = newNSQ();
					    nsQueElem = nsqList;
					    }

					/*
					Copy all the data from the application's request into the
					queue.
					*/

#if defined(vxWorks) || defined(vms)
					nsQueElem -> dueTime = time(NULL) + caNameSrvTimeout;
#endif vxWorks or vms

					nsQueElem -> handle = chBuild.chHandle;
					nsQueElem -> type = chBuild.chType;
					nsQueElem -> count = ntohs(chBuild.chCount);
					if (nsQueElem -> type == AGNT_ADAPTIVE)
					    nsQueElem -> count = 0;
					nsQueElem -> pACB = currACB;
					strcpy(nsQueElem -> name, chanName);

					/*
					Build a request message and send it to the Name Server.
					*/

					if ((namFd != INVALID_SOCKET) && !pointFound)
					    {
#ifndef POINTER64
					    nsMsg.nsSrch.nsHandle = (int) nsQueElem;
#else
					    nsMsg.nsSrch.nsHandle = nsQueElem -> nsqToken;
#endif
					    nsMsg.nsSrch.nsNameLen = strlen(chanName) + 1;
					    nsMsg.nsSrch.nsNameLen = htons(nsMsg.nsSrch.nsNameLen);
					    strcpy(nsMsg.chanName, chanName);
					    msgType = htons(NAME_SRCH);
					    if (!caSendNam((char *) &msgType, sizeof(msgType)))
					        {

#if defined(vxWorks) || defined(vms)
					        closeNameSocket();
					        continue;
#else
#ifndef WIN32
						kill(scmPid, 3);
#endif WIN32
						abort();
#endif vxWorks or vms

					        }
					    if (!caSendNam((char *) &nsMsg, sizeof(nsMsg.nsSrch) + 
						    ntohs(nsMsg.nsSrch.nsNameLen)))
					        {
#if defined(vxWorks) || defined(vms)
					        closeNameSocket();
					        continue;
#else
#ifndef WIN32
						kill(scmPid, 3);
#endif WIN32
						abort();
#endif vxWorks or vms

					        }
					    }
					continue;
#ifdef vxWorks
					    } /* for the else clause for if((namFd == INVALID_SOCKET) && (caMustResolve == 0)) */
#endif vxWorks
					}

				    else if (status2 == 0)
					{
					CHB	*allocCHB;

					/*
					We've seen this name before, but not this particular field.
					Create a new Channel Header Block (CHB) and set it up.
					*/

					allocCHB = newCHB();
					strcpy(allocCHB -> chanName, chanName);
					allocCHB -> currType = chBuild.chType;
					allocCHB -> currCount = ntohs(chBuild.chCount);

					    /*
					    Just copy the pointers to the IOC Status Blocks.
					    */

					allocCHB -> pISB1 = pCHB -> pISB1;
					allocCHB -> pISB2 = pCHB -> pISB2;

					    /*
					    Copy the PVMAP information too.
					    */

					allocCHB -> checksum = pCHB -> checksum;
					allocCHB -> pointType = pCHB -> pointType;
					allocCHB -> creationTime = pCHB -> creationTime;

					    /*
					    Every point has to live on at least one IOC. Set our new CHB
					    to follow the one that was found for us. This makes handling
					    previous links and start of chains easier.
					    */

					allocCHB -> pNextIOC1 = pCHB -> pNextIOC1;
					if (allocCHB -> pNextIOC1)
					    {
					    if (allocCHB -> pNextIOC1 -> pISB1 == allocCHB -> pISB1)
						allocCHB -> pNextIOC1 -> pPrevIOC1 = allocCHB;
					    else
						allocCHB -> pNextIOC1 -> pPrevIOC2 = allocCHB;
					    }
					pCHB -> pNextIOC1 = allocCHB;
					allocCHB -> pPrevIOC1 = pCHB;

					    /*
					    Repeat previous process if point is on another IOC
					    */

					if (allocCHB -> pISB2)
					    {
					    allocCHB -> pNextIOC2 = pCHB -> pNextIOC2;
					    if (allocCHB -> pNextIOC2)
						{
						if (allocCHB -> pNextIOC2 -> pISB1 == allocCHB -> pISB2)
						    allocCHB -> pNextIOC2 -> pPrevIOC1 = allocCHB;
						else
						    allocCHB -> pNextIOC2 -> pPrevIOC2 = allocCHB;
						}
					    pCHB -> pNextIOC2 = allocCHB;
					    allocCHB -> pPrevIOC2 = pCHB;
					    }

					    /*
					    Jump into the resource chain, also behind existing node.
					    */

					allocCHB -> pNextRes = pCHB -> pNextRes;
					pCHB -> pNextRes = allocCHB;
					if (allocCHB -> pNextRes)
					    allocCHB -> pNextRes -> pPrevRes = allocCHB;
					allocCHB -> pPrevRes = pCHB;
					allocCHB -> pRes = pCHB -> pRes;

					    /*
					    Jump into the hash chain, behind existing node.
					    */

					allocCHB -> pNextHash = pCHB -> pNextHash;
					pCHB -> pNextHash = allocCHB;
					allocCHB -> pHash = pCHB -> pHash;

					    /*
					    Point to the new CHB
					    */

					pCHB = allocCHB;
					}
				
				    /*
				    Put a new Channel Application Block onto the Channel Header Block.
				    pCHB points to the CHB.
				    */

				    allocCAB2 = newCAB();
				    allocCAB2 -> pCHB = pCHB;
				    allocCAB2 -> awaitAck = 1;
				    if (chBuild.chType == AGNT_ADAPTIVE)
					{
					allocCAB2 -> awaitMon = 1;
					allocCAB2 -> mask = AGNT_VALUE | AGNT_ALARM;
					}
				    allocCAB2 -> mask |= spliceMask;
				    allocCAB2 -> pACB = currACB;
				    allocCAB2 -> handle = chBuild.chHandle;
				    allocCAB2 -> dataType = chBuild.chType;
				    allocCAB2 -> readType = chBuild.chType;
				    allocCAB2 -> count = ntohs(chBuild.chCount);
				    allocCAB2 -> pNextChan = pCHB -> pFirstCAB;

				    /*
				    Put this CAB at the start of the application chain.
				    */

				    allocCAB2 -> pNextApp = currACB -> pFirstCAB;
				    if (allocCAB2 -> pNextApp)
					allocCAB2 -> pNextApp -> pPrevApp = allocCAB2;
				    currACB -> pFirstCAB = allocCAB2;
				    pCHB -> pFirstCAB = allocCAB2;

				    /* If we have "spliced" a new user onto an existing channel and he
				    wants an initial sample, ask for it now.
				    */

				    if ((status2 == 1) && (allocCAB2 -> readType != AGNT_NOCONN))
					{
					if (pCHB -> pCurrISB)
					    {
					    if (pCHB -> pCurrISB -> connected)
						{
						CH_DIRREAD	readNow;

						msgType = htons(CHAN_DIRREAD);
						if (pCHB -> pCurrISB == pCHB -> pISB1)
						    readNow.chHandle = pCHB -> handleIOC1;
						else
						    readNow.chHandle = pCHB -> handleIOC2;
						readNow.chDataType = htons(allocCAB2 -> readType);
						readNow.chNumElem = htons(allocCAB2 -> count);
#ifndef POINTER64
						readNow.chCaHandle = (int) allocCAB2;
						readNow.chAppHandle = (int) (allocCAB2 -> pACB);
#else
						readNow.chCaHandle = allocCAB2 -> cabToken;
						readNow.chAppHandle = allocCAB2 -> pACB -> acbToken;
#endif
						if ((readNow.chHandle != 0) && (readNow.chHandle != -1))
						    sendScm(pCHB -> pCurrISB -> Inet, msgType,
							(char *) &readNow, sizeof(readNow));
						}
					    }
					}

				    /*
				    Now see if we have to either change the data type or try to create new
				    channels
				    */

				    checkChannel(pCHB);
				    }

				else if (msgType == CHAN_DIRREAD)
				    {
				    CH_DIRREAD		dirRead2;

				    /*
				    Direct Read. This is caused by an application calling the function ca_get()
				    because he wants to read immediately.
				    */

				    MARK(APRD1);
				    appRx(currACB, (char *) &dirRead2, sizeof(dirRead2));

				    pCAB = currACB -> pFirstCAB;
				    while (pCAB)
					{
#ifndef POINTER64
					if (dirRead2.chHandle == (int) pCAB)
#else
					if (dirRead2.chHandle == pCAB -> cabToken)
#endif
					    break;
					pCAB = pCAB -> pNextApp;
					}
				    if (!pCAB)
					continue;
				    
				    MARK(APRD2);

				    pCHB = pCAB -> pCHB;
				    if (!pCHB)
					continue;
#if !defined(vxWorks) && !defined(vms)
				    tisSysLog(3, 0, "chanAcq: Direct read for %s", pCHB -> chanName);
#else
				    if (cacDebug >= 3)
					printf("chanAcq: Direct read for %s\n", pCHB -> chanName);
#endif

				    /*
				    Now post the read at the IOC if he's online.
				    */

				    if (pCAB -> pCHB -> pCurrISB)
					{
					if (!(pCAB -> pCHB -> pCurrISB -> connected))
					    continue;
					pCAB -> dataType = ntohs(dirRead2.chDataType);
					if (pCAB -> mask & AGNT_SPLICED)
					    dirRead2.chDataType |= htons(AGNT_STAT);
					pCAB -> count = ntohs(dirRead2.chNumElem);
				        msgType = htons(msgType);
					if (pCAB -> pCHB -> pCurrISB == pCAB -> pCHB -> pISB1)
				            dirRead2.chHandle = (int)(pCAB -> pCHB -> handleIOC1);
					else
					    dirRead2.chHandle = (int)(pCAB -> pCHB -> handleIOC2);
					if ((dirRead2.chHandle == 0) || (dirRead2.chHandle == -1))
					    continue;
#ifndef POINTER64
				        dirRead2.chCaHandle = (int) pCAB;
					dirRead2.chAppHandle = (int) (pCAB -> pACB);
#else
				        dirRead2.chCaHandle = pCAB -> cabToken;
					dirRead2.chAppHandle = pCAB -> pACB ->acbToken;
#endif
					sendScm(pCAB -> pCHB -> pCurrISB -> Inet, msgType, (char *) &dirRead2,
						sizeof(dirRead2));
					}
				    }

				else if (msgType == CHAN_DIRPUT)
				    {
				    struct
					{
					CH_DIRPUT		dirPut;
				    	char		dataBuff[1024];
					} dp;
				    short		length3;

				    /*
				    Direct Put. This is caused by an application calling the function ca_put()
				    to write data to a channel. Our involvement in writes is really limited
				    to finding the correct Inet address for the IOC that should process this
				    command and substituting the handle for that IOC.
				    */

				    MARK(APPUT1);
				    appRx(currACB, (char *) &dp.dirPut, sizeof(dp.dirPut));
				    length3 = ntohs(dp.dirPut.chDataSize);
				    appRx(currACB, dp.dataBuff, (int) length3);
				    
				    pCAB = (CAB *) (currACB -> pFirstCAB);
				    while (pCAB)
					{
#ifndef POINTER64
					if ((int) pCAB == dp.dirPut.chHandle)
#else
					if ( pCAB -> cabToken == dp.dirPut.chHandle)
#endif
					    break;
					pCAB = (CAB *) (pCAB -> pNextApp);
					}
				    if (!pCAB)
					continue;
				    
				    /*
				    See if we know where to send this output. If we do, all we have to do is
				    substitute the appropriate IOC's handle for the one that the user gave us.
				    */

				    MARK(APPUT2);
				    pCHB = pCAB -> pCHB;
				    if (!pCHB)
					continue;
#if !defined(vxWorks) && !defined(vms)
				    tisSysLog(3, 0, "chanAcq: Direct put for %s\n", pCHB -> chanName);
#else
				    if (cacDebug >= 3)
					printf("chanAcq: Direct put for %s\n", pCHB -> chanName);
#endif
				    if (pCHB -> pCurrISB == pCHB -> pISB1)
					dp.dirPut.chHandle = pCHB -> handleIOC1;
				    else if (pCHB -> pCurrISB == pCHB -> pISB2)
					dp.dirPut.chHandle = pCHB -> handleIOC2;
				    else
					continue;
				    
				    /*
				    Take care of splicing bits.
				    */

				    dp.dirPut.chDataType = ntohs(dp.dirPut.chDataType);
				    dp.dirPut.chDataType |= (pCAB -> mask & AGNT_SPLICED);
				    dp.dirPut.chDataType = htons(dp.dirPut.chDataType);

				    /*
				    Now send it on to the IOC if he's online.
				    */

				    if ((dp.dirPut.chHandle == 0) || (dp.dirPut.chHandle == -1))
					continue;
				    if (!(pCHB -> pCurrISB -> connected))
					continue;
				    msgType = htons(CHAN_DIRPUT);
				    sendScm(pCHB -> pCurrISB -> Inet, msgType, (char *) &dp, sizeof(CH_DIRPUT) 
					    + length3);
				    }

				else if (msgType == CHAN_MONADD)
				    {
				    CH_MONADD		monAdd;
				    CH_MONREQ		monReq;

				    /*
				    Monitor Add. This is caused by an application calling the function
				    ca_add_masked_array_event() or any one of its equivalents. Start by
				    getting the body of the message and then finding the CAB that was
				    referenced. Search for it in the chain for the application because
				    we never believe a pointer that we get from someone else.
				    */

				    MARK(APMAD1);
				    appRx(currACB, (char *) &monAdd, sizeof(monAdd));

				    pCAB = currACB -> pFirstCAB;
				    while (pCAB)
					{
					if (monAdd.chHandle == pCAB -> handle)
					    break;
					pCAB = pCAB -> pNextApp;
					}
				    if (!pCAB)
					continue;
#if !defined(vxWorks) && !defined(vms)
				    tisSysLog(2, 0, "chanAcq: Monitor add for %s.", pCAB -> pCHB -> chanName);
#else
				    if (cacDebug >= 2)
					printf("chanAcq: Monitor add for %s\n", pCAB -> pCHB -> chanName);
#endif

				    /*
				    Found the CAB. Create the monitor by storing the mask in the CAB.
				    If the data type in the CHB hasn't been declared yet, take care
				    of assigning that.  Then, update the channel in the normal manner.
				    */

				    MARK(APMAD2);
				    pCAB -> awaitMon = 1;
				    pCAB -> mask |= ntohs(monAdd.mask);		/* PRESERVE SPLICE BITS! */
				    pCAB -> dataType = ntohs(monAdd.chType);
				    pCAB -> count = ntohs(monAdd.chCount);
				    pCHB = pCAB -> pCHB;
				    if (!pCHB)
					continue;
				    if (pCAB -> pCHB -> currType == AGNT_NOCONN)
					adjustType(pCAB);
				    checkChannel(pCHB);

				    /*
				    Our CAB may have been spliced onto a different CHB. Make sure
				    we are looking at the correct one.
				    */

				    pCHB  = pCAB -> pCHB;
				    if (!pCHB)
					continue;

				    /*
				    Now, issue a request for an immediate monitor report on the channel.
				    This will cause the application to get it's initial event.
				    */

				    if (!(pCHB -> pCurrISB))
					continue;
				    if (!(pCHB -> pCurrISB -> connected))
					continue;
				    if (pCHB -> pCurrISB == pCHB -> pISB1)
					monReq.chHandle = pCHB -> handleIOC1;
				    else
					monReq.chHandle = pCHB -> handleIOC2;
				    if ((monReq.chHandle == 0) || (monReq.chHandle == -1))
					continue;
#ifndef POINTER64
				    monReq.chCaHandle = (int) pCAB;
#else
				    monReq.chCaHandle = pCAB -> cabToken;
#endif
				    monReq.chType = monAdd.chType;
				    monReq.chCount = monAdd.chCount;
				    monReq.mask = monAdd.mask;
				    msgType = htons(CHAN_MONREQ);
				    sendScm(pCHB -> pCurrISB -> Inet, msgType, (char *) &monReq,
					    sizeof(monReq));
				    }

				else if (msgType == CHAN_MONREM)
				    {
				    CH_MONREM		monRem;

				    /*
				    Monitor Remove. This is caused by an application calling the function
				    ca_clear_event() to take the monitor off of a channel.  Start by
				    getting the body of the message and then finding the ACB that was
				    referenced. Search for it in the chain for the application because
				    we never believe a pointer that we get from someone else.
				    */

				    MARK(APMRM1);
				    appRx(currACB, (char *) &monRem, sizeof(monRem));

				    pCAB = currACB -> pFirstCAB;
				    while (pCAB)
					{
#ifndef POINTER64
					if (monRem.chHandle == (int) pCAB)
#else
					if (monRem.chHandle == pCAB ->cabToken)
#endif
					    break;
					pCAB = pCAB -> pNextApp;
					}
				    if (!pCAB)
					continue;
#if !defined(vxWorks) && !defined(vms)
				    tisSysLog(2, 0, "chanAcq: Monitor remove for %s", pCAB -> pCHB -> chanName);
#else
				    if (cacDebug >= 2)
					printf("chanAcq: Monitor remove for %s\n", pCAB -> pCHB -> chanName);
#endif

				    /*
				    Found the CAB. Remove the monitor by clearing the mask in the CAB.
				    Then, update the channel in the normal manner.
				    */

				    MARK(APMRM2);
				    pCAB -> awaitMon = 0;
				    pCAB -> mask &= 0xff;		/* Preserve splicing bits */
				    pCAB -> dataType = -1;
				    if (pCAB -> pCHB)
				        checkChannel(pCAB -> pCHB);
				    }

				else if (msgType == CHAN_CLOSE)
				    {
				    CH_CLOSE	closeChan;
				    CH_CLOSEACK	closeAck2;
				    CAB		*currCAB;

				    /*
				    Channel Close. This is caused by an application calling the function
				    ca_clear_channel() to break connection with a channel.
				    Start by getting the message that contains the handle. The handle
				    is the address of the CAB.
				    */

				    MARK(APCLS1);
				    appRx(currACB, (char *) &closeChan, sizeof(closeChan));

				    /*
				    If the CAB address is zero, then the user closed the channel before
				    we finished opening it. In this case, use the current ACB pointer
				    currACB and the pointer to the user's chid (in his address space)
				    to find and unchain the name queue entry that corresponds to this
				    channel. It is possible that the name queue entry won't be found.
				    In that case, try to find the "offending" CAB by matching the handles.
				    */

				    if (!closeChan.chHandle)
					{
					NSQ	*prevNamQue2;
					NSQ	*currNamQue2;

					prevNamQue2 = 0;
					currNamQue2 = nsqList;
					while (currNamQue2)
					    {
					    if ((currNamQue2 -> pACB == currACB)
						    && (currNamQue2 -> handle == closeChan.chAppHandle))
						{
						if (currNamQue2 == nsqList)
						    nsqList = currNamQue2 -> pNextNSQ;
						else
						    prevNamQue2 -> pNextNSQ = currNamQue2 -> pNextNSQ;
						deleteNSQ(currNamQue2);
						break;
						}
					    prevNamQue2 = currNamQue2;
					    currNamQue2 = currNamQue2 -> pNextNSQ;
					    }
#if 0
					continue;
#endif
					}

				    /*
				    Deleting a channel consists of unchaining the CAB from the ACB and also
				    from the CHB. Start by finding this CAB by matching the handle that
				    we were given. Search for it because we don't trust anyone else's
				    pointers that we gave them a while ago.
				    */

				    currCAB = currACB -> pFirstCAB;
				    while (currCAB)
					{
#ifndef POINTER64
					if ((int) currCAB == closeChan.chHandle)
#else
					if ( currCAB -> cabToken == closeChan.chHandle)
#endif
					    break;
					if ((int) (currCAB -> handle) == closeChan.chAppHandle)
					    {
#ifndef POINTER64
					    closeChan.chHandle = (int) currCAB;
#else
					    closeChan.chHandle = currCAB -> cabToken;
#endif
					    break;
					    }
					currCAB = currCAB -> pNextApp;
					}
				    if (!currCAB)
					continue;
#if !defined(vxWorks) && !defined(vms)
				    tisSysLog(2, 0, "chanAcq: Close for %s", currCAB -> pCHB -> chanName);
#else
				    if (cacDebug >= 2)
					printf("chanAcq: Close for %s\n", currCAB -> pCHB -> chanName);
#endif

				    /*
				    Now that we have a valid pointer to a CAB, take it out of the ACB chain.
				    */

					/*
					Previous one's next is our next unless we were at the start of the
					chain. In that case, the pointer to the first CAB in the ACB is
					our next one.
					*/

				    MARK(APCLS2);
				    currCAB -> awaitAck = 0;
				    if (currCAB -> pPrevApp)
					currCAB -> pPrevApp -> pNextApp = currCAB -> pNextApp;
				    else
					currACB -> pFirstCAB = currCAB -> pNextApp;

					/*
					If there is one in the chain after us, his previous is our previous.
					*/

				    if (currCAB -> pNextApp)
					currCAB -> pNextApp -> pPrevApp = currCAB -> pPrevApp;

				    /*
				    Now also take this block out of the CHB chain. There are only forward
				    links in this case because we don't anticipate that a single channel
				    will be open more than a few times on a single workstation.
				    */

				    if (currCAB -> pCHB)
					{
				        if (currCAB -> pCHB -> pFirstCAB == currCAB)
					    currCAB -> pCHB -> pFirstCAB = currCAB -> pNextChan;

				        else
					    {

					    /*
					    Find the one before us in the chain.
					    */

					    currCAB = currCAB -> pCHB -> pFirstCAB;
					    if (currCAB)
					      {
#ifndef POINTER64
					      while ( (int) (currCAB -> pNextChan) != closeChan.chHandle)
#else
					      while ( currCAB -> pNextChan -> cabToken != closeChan.chHandle)
#endif
						{
					        currCAB = currCAB -> pNextChan;
						if (!currCAB)
						    break;
						}
					      }

					    /*
					    The forward link for the one before us becomes our forward
					    link. The next line just looks messier than it really is.
					    */

					    if (currCAB)
						{
					        currCAB -> pNextChan = currCAB -> pNextChan -> pNextChan;
						}
#ifndef POINTER64
					    currCAB = (CAB *) closeChan.chHandle;
#else
					    currCAB = (CAB *) Pointer64fromToken (closeChan.chHandle);
#endif
					    }
					}

				    /*
				    First see if the channel can be demoted in type or closed completely.
				    Then free the memory that contained the CAB.
				    */

				    if (currCAB -> pCHB)
				        checkChannel(currCAB -> pCHB);
				    currCAB -> pCHB = 0;
				    deleteCAB(currCAB);

				    /*
				    Let the application know that it's now OK to re-use that handle in 
				    future conversations with us.
				    */

				    closeAck2.chHandle = currCAB -> handle;
				    msgType = htons(CHAN_CLOSEACK);
				    if (!appTx(currACB, msgType, (char *) &closeAck2, sizeof(closeAck2)))
				        break;

				    }

				else if (msgType == SOCK_FLUSH)
				    {

				    /*
				    Buffer flush. This is caused by an application calling any of the ca
				    library functions that cause buffers to get physically written out.
				    This routines include ca_flush_io(), ca_pend_io(), and ca_pend_event().
				    Ignore this because we automatically provide a flush whenever we've
				    drained all of the sockets.
				    */

				    }

#ifndef vxWorks
				else if (msgType == CHAN_ALARM_RECV)
				    {
				    struct
					{
				        CH_ALARM_RECV	alarmRecv;
					char		pvName[PVNAME_SZ + FLDNAME_SZ + 2];
					} rMsg;

				    /*
				    An alarm handler is telling us that he has received and logged an alarm.
				    The IOC can now be told that he can delete the alarm from his queue.
				    All we have to do is to forward this message to the IOC.
				    */

				    appRx(currACB, (char *) &rMsg.alarmRecv, sizeof(CH_ALARM_RECV));
				    appRx(currACB, rMsg.pvName, ntohs(rMsg.alarmRecv.alarmNameLen));
#ifndef POINTER64
				    pISB = (ISB *) rMsg.alarmRecv.alarmHandle;
#else
				    pISB = (ISB *) Pointer64fromToken (rMsg.alarmRecv.alarmHandle);
#endif

				    /*
				    Make sure that the ISB pointer that we got in the message is legal.
				    Also make sure that he's online. If not, just ignore the receipt.
				    */

				    currISB = isbList;
				    while (currISB)
					{
					if (currISB == pISB)
					    break;
					currISB = currISB -> nextISB;
					}
				    if (!currISB)
					continue;
				    if (!(pISB -> connected))
					continue;

				    msgType = htons(CHAN_ALARM_RECV);
				    sendScm(pISB -> Inet, msgType, (char *) &rMsg, sizeof(CH_ALARM_RECV)
					+ ntohs(rMsg.alarmRecv.alarmNameLen));
				    }

				else if (msgType == CHAN_ALARM_ACK)
				    {
				    NSQ			*nsQue;
				    CH_ALARM_ACK	alarmAck;
				    char		pvName[PVNAME_SZ + FLDNAME_SZ + 2];

				    /*
				    An alarm handler is telling us that a human operator has acknowledged 
				    the new alarms for a point. Send the message to the proper IOC.
				    */

				    appRx(currACB, (char *) &alarmAck, sizeof(alarmAck));
				    appRx(currACB, pvName, ntohs(alarmAck.alarmNameLen));

				    /*
				    Try to find the name without pestering the name server. If not, queue up
				    and entry for him.
				    */


				    if (nameFind(pvName, &pCHB, AGNT_NOCONN, 0) < 0)
					{
					int		pointFound = 0;
					struct
					    {
					    NS_SRCH	nsSrch;
					    char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];
					    } nMsg;

					/*
					Have never seen this guy. Send it off to the name server. Use an
					impossible count to indicate that this is a name server search
					caused by an alarm ack instead of a channel build.
					*/

					if (nsQue = sameName(pvName))
					    {
					    NSQ		*newElement = newNSQ();

					    pointFound = 1;
					    newElement -> pNextNSQ = nsQue -> nameChain;
					    nsQue -> nameChain = newElement;
					    nsQue = newElement;
					    }

					else if ((nsQue = nsqList))
					    {
					    while (nsQue -> pNextNSQ)
						nsQue = nsQue -> pNextNSQ;
					    nsQue -> pNextNSQ = newNSQ();
					    nsQue = nsQue -> pNextNSQ;
					    }
					
					else
					    {
					    nsqList = newNSQ();
					    nsQue = nsqList;
					    }
					
					/*
					Fill out the name queue entry.
					*/

#ifdef vms
					nsQue -> dueTime = time(NULL) + caNameSrvTimeout;
#endif vms
					nsQue -> ackTime.secPastEpoch = alarmAck.alarmTime.secPastEpoch;
					nsQue -> ackTime.nsec = alarmAck.alarmTime.nsec;
					nsQue -> count = 0x7ffe;
					strcpy(nsQue -> name, pvName);

					/*
					Build a message and send it to the name server if he's online.
					*/

					if ((namFd != INVALID_SOCKET) && !pointFound)
					    {
#ifndef POINTER64
					    nMsg.nsSrch.nsHandle = (int) nsQue;
#else
					    nMsg.nsSrch.nsHandle = nsQue -> nsqToken;
#endif
					    nMsg.nsSrch.nsNameLen = strlen(pvName) + 1;
					    nMsg.nsSrch.nsNameLen = htons(nMsg.nsSrch.nsNameLen);
					    strcpy(nMsg.chanName, pvName);
					    msgType = htons(NAME_SRCH);

					    if (!caSendNam((char *) &msgType, sizeof(msgType)))
						{
#ifdef vms
						closeNameSocket();
						continue;
#else
#ifndef WIN32
						kill(scmPid, 3);
#endif WIN32
						abort();
#endif vms
						}

					    if (!caSendNam((char *) &nMsg, sizeof(nMsg.nsSrch) +
						    ntohs(nMsg.nsSrch.nsNameLen)))
						{
#ifdef vms
						closeNameSocket();
						continue;
#else
#ifndef WIN32
						kill(scmPid, 3);
#endif WIN32
						abort();
#endif vms
						}
					   }
					continue;
					}
				
				    else
					{

					/*
					Name has been seen. If there is a controlling ISB and he's online,
					send him the request. If not, queue it onto the RCB for that
					resource.
					*/

					if (pCHB -> pCurrISB)
					    {
					    if (pCHB -> pCurrISB -> connected)
						{
						struct
						    {
						    CH_ALARM_ACK	ack;
						    char		ackName[PVNAME_SZ + FLDNAME_SZ + 2];
						    } m;

						m.ack = alarmAck;
						strcpy(m.ackName, pvName);
						msgType = htons(CHAN_ALARM_ACK);
						sendScm(pCHB -> pCurrISB -> Inet, msgType, (char *) &m,
							sizeof(CH_ALARM_ACK) + ntohs(m.ack.alarmNameLen));
						continue;
						}
					    }

					/*
					Can't send message now. Create a queue element and stick it onto
					the resource.
					*/

					nsQue = newNSQ();
					nsQue -> pNextNSQ = pCHB -> pRes -> pFirstAck;
					pCHB -> pRes -> pFirstAck = nsQue;
					nsQue -> ackTime.secPastEpoch = alarmAck.alarmTime.secPastEpoch;
					nsQue -> ackTime.nsec = alarmAck.alarmTime.nsec;
					strcpy(nsQue -> name, pvName);
					}
				    }

				else if (msgType == CHAN_ALARM_CTL)
				    {
				    NSQ			*nsQue;
				    CH_ALARM_CTL	alarmCtl;
				    char		pvName[PVNAME_SZ + FLDNAME_SZ + 2];

				    /*
				    An alarm handler wants to either enable or disable alarming for a point.
				    Send the message to the proper IOC.
				    */

				    appRx(currACB, (char *) &alarmCtl, sizeof(alarmCtl));
				    appRx(currACB, pvName, ntohs(alarmCtl.alarmNameLen));

				    /*
				    Try to find the name without pestering the name server. If not, queue up
				    and entry for him.
				    */


				    if (nameFind(pvName, &pCHB, AGNT_NOCONN, 0) < 0)
					{
					int		pointFound = 0;
					struct
					    {
					    NS_SRCH	nsSrch;
					    char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];
					    } nMsg;

					/*
					Have never seen this guy. Send it off to the name server. Use an
					impossible count to indicate that this is a name server search
					caused by an alarm ack instead of a channel build.
					*/

					if (nsQue = sameName(pvName))
					    {
					    NSQ		*newElement = newNSQ();

					    pointFound = 1;
					    newElement -> pNextNSQ = nsQue -> nameChain;
					    nsQue -> nameChain = newElement;
					    nsQue = newElement;
					    }

					else if ((nsQue = nsqList))
					    {
					    while (nsQue -> pNextNSQ)
						nsQue = nsQue -> pNextNSQ;
					    nsQue -> pNextNSQ = newNSQ();
					    nsQue = nsQue -> pNextNSQ;
					    }
					
					else
					    {
					    nsqList = newNSQ();
					    nsQue = nsqList;
					    }
					
					/*
					Fill out the name queue entry.
					*/

#ifdef vms
					nsQue -> dueTime = time(NULL) + caNameSrvTimeout;
#endif vms
					nsQue -> count = 0x7ffd;
					nsQue -> type = alarmCtl.alarmCtlFunc;
					strcpy(nsQue -> name, pvName);

					/*
					Build a message and send it to the name server if he's online.
					*/

					if ((namFd != INVALID_SOCKET) && !pointFound)
					    {
#ifndef POINTER64
					    nMsg.nsSrch.nsHandle = (int) nsQue;
#else
					    nMsg.nsSrch.nsHandle = nsQue -> nsqToken;
#endif
					    nMsg.nsSrch.nsNameLen = strlen(pvName) + 1;
					    nMsg.nsSrch.nsNameLen = htons(nMsg.nsSrch.nsNameLen);
					    strcpy(nMsg.chanName, pvName);
					    msgType = htons(NAME_SRCH);

					    if (!caSendNam((char *) &msgType, sizeof(msgType)))
						{
#ifdef vms
						closeNameSocket();
						continue;
#else
#ifndef WIN32
						kill(scmPid, 3);
#endif WIN32
						abort();
#endif vms
						}

					    if (!caSendNam((char *) &nMsg, sizeof(nMsg.nsSrch) +
						    ntohs(nMsg.nsSrch.nsNameLen)))
						{
#ifdef vms
						closeNameSocket();
						continue;
#else
#ifndef WIN32
						kill(scmPid, 3);
#endif WIN32
						abort();
#endif vms
						}
					   }
					continue;
					}
				
				    else
					{

					/*
					Name has been seen. If there is a controlling ISB and he's online,
					send him the request. If not, queue it onto the RCB for that
					resource.
					*/

					if (pCHB -> pCurrISB)
					    {
					    if (pCHB -> pCurrISB -> connected)
						{
						struct
						    {
						    CH_ALARM_CTL	ctl;
						    char		ctlName[PVNAME_SZ + FLDNAME_SZ + 2];
						    } m;

						m.ctl = alarmCtl;
						strcpy(m.ctlName, pvName);
						msgType = htons(CHAN_ALARM_CTL);
						sendScm(pCHB -> pCurrISB -> Inet, msgType, (char *) &m,
							sizeof(CH_ALARM_CTL) + ntohs(m.ctl.alarmNameLen));
						continue;
						}
					    }

					/*
					Can't send message now. Create a queue element and stick it onto
					the resource.
					*/

					nsQue = newNSQ();
					nsQue -> pNextNSQ = pCHB -> pRes -> pFirstCtl;
					pCHB -> pRes -> pFirstCtl = nsQue;
					nsQue -> type = alarmCtl.alarmCtlFunc;
					strcpy(nsQue -> name, pvName);
					}
				    }

				else if (msgType == CHAN_HANDLERS)
				    {
				    CH_HANDLERS		handlerMsg;
				    CH_RESMODE		resMode;

				    /*
				    An application has registered as a handler. Mark his handler status in
				    his ACB. Then, bump the counts of the number of handlers we support. If
				    the number is one, then we just added the first handler of that type.
				    Inform all the IOCs in that case.
				    */

				    appRx(currACB, (char *) &handlerMsg, sizeof(handlerMsg));

				    /*
				    These counts are additive. A single application could be both a trend
				    and an alarm handler even if this is unlikely.
				    */

				    currACB -> trendHandler += ntohs(handlerMsg.chTrendHandler);
				    currACB -> alarmHandler += ntohs(handlerMsg.chAlarmHandler);

				    numTrendHandlers += ntohs(handlerMsg.chTrendHandler);
				    numAlarmHandlers += ntohs(handlerMsg.chAlarmHandler);

				    if ((numTrendHandlers == 1) || (numAlarmHandlers == 1))
					{
					ISB		*handlerISB;

					msgType = htons(CHAN_HANDLERS);
					handlerMsg.chTrendHandler = htons(numTrendHandlers);
					handlerMsg.chAlarmHandler = htons(numAlarmHandlers);

					handlerISB = isbList;
					while (handlerISB)
					    {
					    if (handlerISB -> connected)
					        sendScm(handlerISB -> Inet, msgType, (char *) &handlerMsg,
							sizeof(handlerMsg));
					    handlerISB = handlerISB -> nextISB;
					    }
					}

				    /*
				    Now, if he is either kind of handler, run down the list of resources
				    that we know of and tell him about the state of each one.
				    */

				    if (!(handlerMsg.chTrendHandler || handlerMsg.chAlarmHandler))
					continue;
				    msgType = htons(CHAN_RESMODE);
				    currRCB = rcbList;
				    while (currRCB)
					{
					strcpy(resMode.chResName, currRCB -> resName);
					if (currRCB -> currISB)
					    resMode.chResStatus = htons(1);
					else
					    resMode.chResStatus = 0;
				        if (!appTx(currACB, msgType, (char *) &resMode, sizeof(resMode)))
					    break;
					currRCB = currRCB -> pNextRCB;
					}
				    }

				else if (msgType == CHAN_GETPARAMS)
				    {
				    CH_GETPARAMS	getParams;
				    char		paramName[PVNAME_SZ + FLDNAME_SZ + 2];
				    NSQ			*nsQue;

				    /*
				    An application has requested a connectionless set of parameters. See if
				    we already know where the point resides.
				    */

				    appRx(currACB, (char *) &getParams, sizeof(getParams));
				    msgType = ntohs(getParams.chNameLen);
				    appRx(currACB, paramName, (int) msgType);
				    if (nameFind(paramName, &pCHB, AGNT_NOCONN, 0) < 0)
					{
					int		pointFound = 0;
					struct
					    {
					    NS_SRCH	nsSrch;
					    char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];
					    } nMsg;

					/*
					Have never seen this guy. Send it off to the name server. Use an
					impossible count to indicate that this is a name server search
					caused by a parameter fetch instead of a channel build.
					*/

					if (nsQue = sameName(paramName))
					    {
					    NSQ		*newElement = newNSQ();

					    pointFound = 1;
					    newElement -> pNextNSQ = nsQue -> nameChain;
					    nsQue -> nameChain = newElement;
					    nsQue = newElement;
					    }

					else if ((nsQue = nsqList))
					    {
					    while (nsQue -> pNextNSQ)
						nsQue = nsQue -> pNextNSQ;
					    nsQue -> pNextNSQ = newNSQ();
					    nsQue = nsQue -> pNextNSQ;
					    }
					
					else
					    {
					    nsqList = newNSQ();
					    nsQue = nsqList;
					    }
					
					/*
					Fill out the name queue entry.
					*/

#ifdef vms
					nsQue -> dueTime = time(NULL) + caNameSrvTimeout;
#endif vms
					nsQue -> handle = getParams.chArgument;
					nsQue -> type = ntohs(getParams.chParmType);
					nsQue -> count = 0x7fff;
					nsQue -> pACB = currACB;
					strcpy(nsQue -> name, paramName);

					/*
					Build a message and send it to the name server if he's online.
					*/

					if ((namFd != INVALID_SOCKET) && !pointFound)
					    {
#ifndef POINTER64
					    nMsg.nsSrch.nsHandle = (int) nsQue;
#else
					    nMsg.nsSrch.nsHandle = nsQue -> nsqToken;
#endif
					    nMsg.nsSrch.nsNameLen = strlen(paramName) + 1;
					    nMsg.nsSrch.nsNameLen = htons(nMsg.nsSrch.nsNameLen);
					    strcpy(nMsg.chanName, paramName);
					    msgType = htons(NAME_SRCH);

					    if (!caSendNam((char *) &msgType, sizeof(msgType)))
						{
#ifdef vms
						closeNameSocket();
						continue;
#else
#ifndef WIN32
						kill(scmPid, 3);
#endif WIN32
						abort();
#endif vms
						}

					    if (!caSendNam((char *) &nMsg, sizeof(nMsg.nsSrch) +
						    ntohs(nMsg.nsSrch.nsNameLen)))
						{
#ifdef vms
						closeNameSocket();
						continue;
#else
#ifndef WIN32
						kill(scmPid, 3);
#endif WIN32
						abort();
#endif vms
						}
					   }
					continue;
					}
				
				    else
					{

					/*
					Name has been seen. If there is a controlling ISB and he's online,
					send him the request. If not, queue it onto the RCB for that
					resource.
					*/

					if (pCHB -> pCurrISB)
					    {
					    if (pCHB -> pCurrISB -> connected)
						{
						struct
						    {
						    CH_GETPARAMS	gp;
						    char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
						    } m;

						m.gp = getParams;
#ifndef POINTER64
						m.gp.chACB = (int) currACB;
#else
						m.gp.chACB = currACB -> acbToken;
#endif
						strcpy(m.chanName, paramName);
						msgType = htons(CHAN_GETPARAMS);
						sendScm(pCHB -> pCurrISB -> Inet, msgType, (char *) &m,
							sizeof(CH_GETPARAMS) + ntohs(m.gp.chNameLen));
						continue;
						}
					    }

					/*
					Can't send message now. Create a queue element and stick it onto
					the resource.
					*/

					nsQue = newNSQ();
					nsQue -> pNextNSQ = pCHB -> pRes -> pFirstParam;
					pCHB -> pRes -> pFirstParam = nsQue;
					nsQue -> handle = getParams.chArgument;
					nsQue -> type = ntohs(getParams.chParmType);
					nsQue -> pACB = currACB;
					strcpy(nsQue -> name, paramName);
					continue;
					}
				    }
#endif vxWorks

				else
				    {

				    /*
				    Unknown message code. More than likely, we got out of sync with this 
				    application.
				    */

#if !defined(vxWorks) && !defined(vms)
				    tisSysLog(0, 0, "chanAcq: Unknown message code %hd from application", msgType);
#else
				    fprintf(stderr, "Unknown message code %hd from application.\n", msgType);
#endif
				    continue;
				    }
				}
			    }
			  }
			}
		    }
		}

	    /*
	    We've processed all the messages sent to us from all apps, name servers, and our socket
	    manager. If we sent anything to an IOC, make the socket manager flush it now.
	    */

	    if (anyTcpOutputs)
		{
		if ((scmOutputBytes + sizeof(msgType)) > sizeof(scmBuff))
		    scmFlush();
		msgType = SOCK_FLUSH;
		bcopy((char *) &msgType, scmBuff + scmOutputBytes, sizeof(msgType));
		scmOutputBytes += sizeof(msgType);
	    	scmFlush();
		}


	    /*
	    Also, flush anything that has to go out to our applications.
	    */

	    currACB = acbList;
	    while (currACB)
		{
		nextACB = currACB -> pNextACB;
		appFlush(currACB);
		if (nextACB == currACB)
		    break;
		currACB = nextACB;
		}

	    /*
	    And flush the buffer to the name server, too.
	    */


	    if (!caNameFlush(nameTxBuff, namTxBytes))
		{

#if defined(vxWorks) || defined(vms)
		closeNameSocket();
#else
#ifndef WIN32
		kill(scmPid, 3);
#endif WIN32
		abort();
#endif vxWorks or vms

		}

	    }
	    return 0;
}


/****************************************************************/
/*								*/
/*		SEND A MESSAGE TO AN APPLICATION		*/
/*								*/
/*	Return a zero on failure				*/
/*								*/
/****************************************************************/

int appTx(pACB, msgType, buffer, nbytes)
ACB	*pACB;
short	msgType;
char	*buffer;
int	nbytes;
{

#ifndef vxWorks
    if (pACB -> sockFd == INVALID_SOCKET)
	return(0);
    if ((pACB -> buffSize + sizeof(msgType) + nbytes) > sizeof(pACB -> appBuff))
	{
	if (appFlush(pACB) == 0)
	    return (0);
	}
    bcopy((char *) &msgType, pACB -> appBuff + pACB -> buffSize, sizeof(msgType));
    pACB -> buffSize += sizeof(msgType);
    bcopy((char *) buffer, pACB -> appBuff + pACB -> buffSize, nbytes);
    pACB -> buffSize += nbytes;
#else
    char	*auxBuff = 0;
    int		localMsgType;

    localMsgType = ntohs(msgType);
    if (localMsgType == CHAN_DRRDATA)
	auxBuff = buffer + sizeof(CH_DRRDATA);
    else if (localMsgType == CHAN_MONDATA)
	auxBuff = buffer + sizeof(CH_MONDATA);
    caMsgHandler (localMsgType, buffer, auxBuff);
#endif vxWorks

    return (1);
}


/****************************************************************/
/*								*/
/*		FLUSH THE BUFFER TO AN APPLICATION		*/
/*								*/
/*	Return a zero on failure				*/
/*								*/
/****************************************************************/

int appFlush(pACB)
ACB	*pACB;
{
    int		bytesSent;

    if (pACB -> sockFd == INVALID_SOCKET)
    {
	return (0);
    }
    while (pACB -> buffSize)
    {
	bytesSent = send(pACB -> sockFd, pACB -> appBuff, pACB -> buffSize, 0);
#ifndef vxWorks
#ifndef vms
	if (traceFd)
	{
	    fprintf(traceFd, "Tx to app, %d bytes\n", bytesSent);
	}
#endif
#endif

	if (bytesSent == -1)
	{
#ifdef WIN32
	    errno = WSAGetLastError ();
#endif WIN32
	    if ((errno == EWOULDBLOCK) || (errno == EAGAIN))
	    {
		pACB -> appTxSize += pACB -> buffSize;
		if (setsockopt(pACB -> sockFd, SOL_SOCKET, SO_SNDBUF, (char *)&(pACB -> appTxSize), intsize) == -1)
		{
#ifdef WIN32
		    errno = WSAGetLastError ();
#endif WIN32
#if !defined(vxWorks) && !defined(vms)
		    tisSysLog(1, 0, "chanAcq kill app because of transmit buffer maxout");
		    if (pACB -> trendHandler)
		    {
			tisSysLog(0, 0, "chanAcq drop trend handler because of buffer maxout");
		    }
		    if (pACB -> alarmHandler)
		    {
			tisSysLog(0, 0, "chanAcq drop alarm handler because of buffer maxout");
		    }
#else
		    fprintf(stderr, "chanAcq Kill app because of transmit buffer maxout\n");
#endif
		    FD_CLR(pACB -> sockFd, &cacfdMask);
#ifdef WIN32
		    closesocket(pACB -> sockFd);
#else
		    close(pACB -> sockFd);
#endif WIN32
		    pACB -> sockFd = INVALID_SOCKET;
		    delApp(pACB);
		    return (0);
		}
		continue;
	    }
	    else
	    {
#if !defined(vxWorks) && !defined(vms)
		char	errnoMsg[32];

		sprintf(errnoMsg, "errno = %d\n", errno);
		tisSysLog(1, 0, "chanAcq send error to application");
		if (pACB -> trendHandler)
		{
		    tisSysLog(0, 0, "chanAcq send error to trend handler");
		}
		if (pACB -> alarmHandler)
		{
		    tisSysLog(0, 0, "chanAcq send error to alarm handler");
		}
		if (pACB -> trendHandler || pACB -> alarmHandler)
		{
		    tisSysLog(0, 0, errnoMsg);
		}
#else
		perror("chanAcq send error to application");
#endif
		FD_CLR(pACB -> sockFd, &cacfdMask);
#ifdef WIN32
		closesocket(pACB -> sockFd);
#else
		close(pACB -> sockFd);
#endif WIN32
		pACB -> sockFd = INVALID_SOCKET;
		delApp(pACB);
		return (0);
	    }
	}
	else if (bytesSent != pACB -> buffSize)
	{
	    bcopy(pACB -> appBuff + bytesSent, pACB -> appBuff, pACB -> buffSize);
	}
	pACB -> buffSize -= bytesSent;
    }
    return (1);
}


/****************************************************************/
/*								*/
/*		RECEIVE A MESSAGE FROM AN APPLICATION		*/
/*								*/
/*	We always know that there is enough data in the buffer.	*/
/*	haveCompleteAppMsg() has done that for us. Thanks.	*/
/*								*/
/****************************************************************/

int appRx(pACB, buffer, nbytes)
ACB	*pACB;
char	*buffer;
int	nbytes;
{
    bcopy(pACB -> appRxBuff + pACB -> rcvOutPos, buffer, nbytes);
    pACB -> rcvOutPos += nbytes;
    pACB -> nchib -= nbytes;
    return (1);
}


/****************************************************************/
/*								*/
/*		GET A MESSAGE FRAGMENT FROM RX BUFFER		*/
/*								*/
/*	Extract message from the receive buffer.		*/
/*								*/
/*	There is now a subtle twist to this routine. When we	*/
/*	are "up and running," there will always be enough	*/
/*	bytes in the buffer because haveCompleteNameMsg()	*/
/*	will have made sure of that. The test for namNchib	*/
/*	is only necessary for fetching the initial tables	*/
/*	we need to get started. Oh what a tangled web we	*/
/*	weave!							*/
/*								*/
/****************************************************************/

int sockRead(buff, nbytes)
char	*buff;
int	nbytes;
{
    fd_set	fdList;
    while (nbytes > namNchib)
	{
	FD_ZERO(&fdList);
	FD_SET(namFd, &fdList);
	if (select(namFd + 1, &fdList, NULL, NULL, &oneSecond) <= 0)
	    {

#if !defined(vxWorks) && !defined(vms)
	    tisSysLog(0, 0, "chanAcq select error or timeout reading from nameSrv: %m");
#else
	    printf("chanAcq select error or timeout reading from Name Server.\n");
#endif vxWorks

	    return (0);
	    }

	if (!chGetMore())
	    return (0);
	}

    bcopy(nameRxBuff + namBufOutPos, buff, nbytes);
    namBufOutPos += nbytes;
    namNchib -= nbytes;

    return (1);
}


/****************************************************************/
/*								*/
/*   DETERMINE IF WE HAVE A COMPLETE MSG FROM THE NAME SERVER	*/
/*								*/
/****************************************************************/

int haveCompleteNameMsg()
{
    NS_SRCH_REPLY	*srchReply;
    int			bytesNeeded = sizeof(short) + sizeof(NS_SRCH_REPLY);
    short		tempShort;

    /*
    Got to have a message type and a NS_SRCH_REPLY structure.
    */

    if (namNchib < bytesNeeded)
	return (0);

    /*
    Locate the header so we can peek inside it.
    */

    srchReply = (NS_SRCH_REPLY *)(nameRxBuff + namBufOutPos + sizeof(short));

    /*
    There must be at least the number of nodes * sizeof(int) plus
    the number of bytes in the resource name in the buffer.
    */

    bcopy((char *) &(srchReply -> nsNumNodes), (char *) &tempShort, sizeof(tempShort));
    bytesNeeded += ntohs(tempShort) * sizeof(int);
    bcopy((char *) &(srchReply -> nsResNameLen), (char *) &tempShort, sizeof(tempShort));
    bytesNeeded += ntohs(tempShort);
    if (namNchib < bytesNeeded)
	return (0);

    /*
    There is at least one complete message in the buffer!
    */

    return (1);
}

/****************************************************************/
/*								*/
/*		FILL THE NAME SERVER RECEIVE BUFFER		*/
/*								*/
/*	Read as much as we can in one call to minimize trips	*/
/*	through the operating system.				*/
/*								*/
/****************************************************************/

int chGetMore()
{
    int		bytesGot;

    if (namNchib && namBufOutPos)
	bcopy(nameRxBuff + namBufOutPos, nameRxBuff, namNchib);

    if ((bytesGot = recv(namFd, nameRxBuff + namNchib, sizeof(nameRxBuff) - namNchib, 0)) <= 0)
	{
	if (bytesGot == 0)
	    {

#if !defined(vxWorks) && !defined(vms)
	    tisSysLog(0, 1, "chanAcq zero length receive from nameSrv.");
#else
#ifdef vxWorks
	    logMsg ("chanAcq zero length receive from name server on %d.\n", namFd,0,0,0,0,0);
#else vxWorks
	    fprintf(stderr, "chanAcq zero length receive from name server.\n");
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms)

	    }

	else

	    {

#if !defined(vxWorks) && !defined(vms)
	    tisSysLog(0, 1,  "chanAcq recv error from nameSrv: %m");
#else
#ifdef vxWorks
	    logMsg ("chanAcq recv error from Name Server %d on %d\n", errno, namFd,0,0,0,0);
#else vxWorks
	    perror("chanAcq recv error from Name Server");
#endif vxWorks
#endif vxWorks

	    }
	return (0);
	}

#ifndef vxWorks
#ifndef vms
    if (traceFd)
	fprintf(traceFd, "Rx from nameSrv, %d bytes\n", bytesGot);
#endif
#endif

    namNchib += bytesGot;
    namBufOutPos = 0;

    return (1);
}


/****************************************************************/
/*								*/
/*	DELETE AN APPLICATION BECAUSE WE LOST HIS SOCKET CONN	*/
/*								*/
/****************************************************************/

void delApp(pACB)
ACB	*pACB;
{
    CAB		*currCAB;
    CHB		*pCHB;
    CAB		*pCAB;
    ACB		*currACB;

    /*
    Unchain all of this application's CABs from their CHBs
    */

    pACB -> sockFd = INVALID_SOCKET;
    currCAB = pACB -> pFirstCAB;

    while (currCAB)
	{
	if (currCAB -> pPrevApp)
	    {
	    currCAB -> pPrevApp -> pCHB = 0;
	    deleteCAB(currCAB -> pPrevApp);
	    }
	pCHB = currCAB -> pCHB;
	if (!pCHB)
	    {
	    if (currCAB -> pNextApp)
		{
		currCAB = currCAB -> pNextApp;
		continue;
		}
	    else
		{
		deleteCAB(currCAB);
		break;
		}
	    }
	if (pCHB -> pFirstCAB == currCAB)
	    pCHB -> pFirstCAB = currCAB -> pNextChan;
	else if ((pCAB = pCHB -> pFirstCAB))
	    {
	    while (pCAB -> pNextChan != currCAB)
		pCAB = pCAB -> pNextChan;
	    pCAB -> pNextChan = currCAB -> pNextChan;
	    }
        checkChannel(pCHB);
        if (currCAB -> pNextApp)
	    currCAB = currCAB -> pNextApp;
        else
	    {
	    currCAB -> pCHB = 0;
	    deleteCAB(currCAB);
	    break;
	    }
	}

    /*
    Deduct this app's handler status from the number that we support.
    If this number goes to zero, then send our new handler state to
    all the IOCs that we know about.
    */

#ifndef vxWorks
    numTrendHandlers -= pACB -> trendHandler;
    numAlarmHandlers -= pACB -> alarmHandler;
    if ((pACB -> trendHandler || pACB -> alarmHandler)
	    && ((numTrendHandlers == 0) || (numAlarmHandlers == 0)) )
	{
	short		msgType;
	ISB		*handlerISB;
	CH_HANDLERS	handlerMsg;

	msgType = htons(CHAN_HANDLERS);
	handlerMsg.chTrendHandler = numTrendHandlers;
	handlerMsg.chAlarmHandler = numAlarmHandlers;
	handlerISB = isbList;
	while (handlerISB)
	    {
	    if (handlerISB -> connected)
		sendScm(handlerISB -> Inet, msgType, (char *) &handlerMsg, sizeof(handlerMsg));
	    handlerISB = handlerISB -> nextISB;
	    }
	}
#endif vxWorks

    /*
    Now unchain the ACB itself.
    */

    if (acbList == pACB)
	acbList = pACB -> pNextACB;
    else if (acbList)
	{
	currACB = acbList;
	while (currACB -> pNextACB != pACB)
	    {
	    currACB = currACB -> pNextACB;
	    if (!currACB)
		break;
	    }
	if (currACB)
	    currACB -> pNextACB = pACB -> pNextACB;
	}
    deleteACB(pACB);
}


/****************************************************************/
/*								*/
/*	SEND A MESSAGE TO SOCKET CREATION AND MAINTENANCE	*/
/*								*/
/*	This routine buffers up as much as possible and the	*/
/*	mainline code will dump out the buffer when all of the	*/
/*	sockets have been drained. Strictly speaking, this	*/
/*	buffering is not necessary, but the overhead of lots	*/
/*	of little TCP messages is considerable.			*/
/*								*/
/****************************************************************/

void sendScm(inet, msgType, buff, nbytes)
struct in_addr inet;
short	msgType;
char	*buff;
int	nbytes;
{
    SM_WRITE	writeHdr;
    short	writeType;

    /*
    First see if there is room in the buffer for this transaction.
    */

    if ((scmOutputBytes + sizeof(writeType) + sizeof(writeHdr) + sizeof(msgType) + nbytes) > sizeof(scmBuff))
	scmFlush();

    /*
    Stack up everything we need in the output buffer.
    */

    writeType = SOCK_WRITE;
    bcopy((char *) &writeType, scmBuff + scmOutputBytes, sizeof(writeType));
    scmOutputBytes += sizeof(writeType);
    writeHdr.smInet = inet;
    writeHdr.smLength = nbytes + sizeof(msgType);
    bcopy((char *) &writeHdr, scmBuff + scmOutputBytes, sizeof(writeHdr));
    scmOutputBytes += sizeof(writeHdr);
    bcopy((char *) &msgType, scmBuff + scmOutputBytes, sizeof(msgType));
    scmOutputBytes += sizeof(msgType);
    if (buff != NULL)
	{
	bcopy(buff, scmBuff + scmOutputBytes, nbytes);
	scmOutputBytes += nbytes;
	}
    anyTcpOutputs = 1;
}


/****************************************************************/
/*								*/
/*		FLUSH THE BUFFER TO SOCKET MAINTENANCE		*/
/*								*/
/*	Return a zero on failure				*/
/*								*/
/****************************************************************/

int scmFlush()
{
    int		bytesSent;

    while (scmOutputBytes)
    {
#ifndef vxWorks
	if (cacExtBytes)
	{
	    bytesSent = -1;
	    errno = EWOULDBLOCK;
	}
	else
#endif vxWorks
	{
            bytesSent = send(scmFd, scmBuff, scmOutputBytes, 0);
	}
#ifndef vxWorks
#ifndef vms
	if (traceFd)
	{
	    fprintf(traceFd, "Tx to sockMnt, %d bytes\n", bytesSent);
	}
#endif
#endif

    if (bytesSent == -1)
	{
#ifdef WIN32
	errno = WSAGetLastError ();
#endif WIN32
	    if ((errno == EWOULDBLOCK) || (errno == EAGAIN))
	    {
		if (scmTxTopped)
		{

#ifndef vxWorks
		    char	*outBuff = scmBuff;
		    if ((scmOutputBytes + cacExtBytes) > sizeof(cacMobyBuff))
		    {

#ifndef vms
			tisSysLog(0, 0, "chanAcq quits because of moby buffer overflow");
#ifndef WIN32
		 	kill(scmPid, 3);
#endif WIN32
			abort();
#else
			fprintf(stderr, "chanAcq quits because of moby buffer overflow\n");
			exit(1);
#endif vms

		    }
		    cacExtBytes += scmOutputBytes;
		    while (scmOutputBytes > 0)
		    {
			bytesSent = scmOutputBytes;
			if ((cacRingIn + bytesSent) > sizeof(cacMobyBuff))
			{
			    bytesSent = sizeof(cacMobyBuff) - cacRingIn;
			}
			bcopy(outBuff, cacMobyBuff + cacRingIn, bytesSent);
			cacRingIn += bytesSent;
			if (cacRingIn >= sizeof(cacMobyBuff))
			{
			    cacRingIn = 0;
			}
			outBuff += bytesSent;
			scmOutputBytes -= bytesSent;
		    }
		    return (1);

#else
		    printf("chanAcq Quit because of logjam to sockMnt\n");

		    taskSuspend(taskIdSelf());	/* For IOC debugging help */
#endif vxWorks

		}
		scmTxSize += scmOutputBytes;
		if (setsockopt(scmFd, SOL_SOCKET, SO_SNDBUF, (char *)&scmTxSize, sizeof(scmTxSize)) == -1)
		{
		    scmTxTopped = 1;
		}
		continue;
	    }
	    else
	    {

#if !defined(vxWorks) && !defined(vms)
		tisSysLog(0, 0, "chanAcq transmit error to sockMnt: %m");
#ifndef WIN32
		kill(scmPid, 3);
#endif WIN32
		abort();
#else
		perror("chanAcq transmit error to sockMnt");
#endif vxWorks

#ifdef vxWorks
		taskSuspend(taskIdSelf());	/* For IOC debugging help */
#else
		exit(1);	/* Be more robust than this in the future */
#endif vxWorks
	    }
	}
	else if (bytesSent != scmOutputBytes)
	{
	    bcopy(scmBuff + bytesSent, scmBuff, (scmOutputBytes - bytesSent));
	}
	scmOutputBytes -= bytesSent;
    }
    return (1);
}


/****************************************************************/
/*								*/
/*		SEND A MESSAGE TO THE NAME SERVER		*/
/*								*/
/****************************************************************/


int caSendNam(buff, nbytes)
char	*buff;
int	nbytes;
{

    if ((namTxBytes + nbytes) > sizeof(nameTxBuff))
    {
	if(!caNameFlush(nameTxBuff, namTxBytes))
	{
	    return(0);	/* Fixed lack of error check - now lets caller deal with error */
	}
    }
    memcpy(nameTxBuff + namTxBytes, buff, nbytes);
    namTxBytes += nbytes;
    return (1);
}


/****************************************************************/
/*								*/
/*		FLUSH THE BUFFER TO THE NAME SERVER		*/
/*								*/
/****************************************************************/

int caNameFlush(buff, nbytes)
char	*buff;
int	nbytes;
{
    int		bytesSent;

	while (nbytes)
	{
	    bytesSent = send (namFd, buff, nbytes, 0);

#ifndef vxWorks
#ifndef vms

	    if (traceFd)
	    {
		fprintf(traceFd, "Tx to nameSrv, %d bytes\n", bytesSent);
	    }

#endif
#endif

	    if (bytesSent == -1)
	    {
#ifdef WIN32
		errno = WSAGetLastError ();
#endif WIN32
		if ((errno == EWOULDBLOCK) || (errno == EAGAIN))
		{
		    namTxSize += nbytes;
		    if (setsockopt(namFd, SOL_SOCKET, SO_SNDBUF, (char *)&namTxSize, sizeof(namTxSize)) == -1)
		    {

#if !defined(vxWorks) && !defined(vms)

			tisSysLog(0, 0, "chanAcq xmit buffer to nameSrv topped out");
#else
			logMsg("chanAcq xmit buffer to name server topped out.\n",
				0, 0, 0, 0, 0, 0);

#endif vxWorks
			return (0);
		    }
		    continue;
		}
		else
	 	{

#if !defined(vxWorks) && !defined(vms)
		    tisSysLog(0, 0, "chanAcq send error to nameSrv: %m");
#else
		    logMsg("chanAcq send error to name server", 0, 0, 0, 0, 0, 0);
#endif vxWorks
		    return (0);
		}
	    }
	    else if (bytesSent != nbytes)
	    {
		buff += bytesSent;
#ifdef vxWorks
		if (cacDebug >= 1)
		{
		    logMsg ("Partial send to namSrv %d bytes... wanted %d bytes\n",
				bytesSent, nbytes, 0, 0, 0, 0);
		}
#endif vxWorks
	    }
	    nbytes -= bytesSent;
	}
    namTxBytes = 0;
    return (1);
}


/****************************************************************/
/*								*/
/*		STANDARDIZE THE CHANNEL NAME			*/
/*								*/
/*	If the channel name does not specify a field, append	*/
/*	".VAL" If the name ends in a period, append "VAL"	*/
/*								*/
/****************************************************************/

void nameFix(chanName)
char	*chanName;
{
    char	*lastDot;

    while (*chanName)
	{
	if (*chanName == '.')
	    break;
	chanName++;
	}

    if (*chanName)
	{
	if ((lastDot = strrchr(chanName + 1, '.')) != NULL)
	    if (lastDot != chanName)
	    	strcpy(chanName, lastDot);
	if (*(chanName + 1) == 0)
	    strcpy(chanName + 1, "VAL");
	}
    else
	strcpy(chanName, ".VAL");
}


/****************************************************************/
/*								*/
/*	FIND A NAME IN OUR LIST OF KNOWN CHANNELS		*/
/*								*/
/*	Return a -1 if we've never seen the name before.	*/
/*	Return a zero and the address of the CHB if we find	*/
/*	the name but not the field and data type.		*/
/*	Return a 1 and the address of the CHB if we find a	*/
/*	perfect match.						*/
/*								*/
/*	The name is standardized before we get called.		*/
/*								*/
/*	Channel splicing: A name with splicing bits set is	*/
/*	ACK, STAT, of SEVR. That can be hung on VAL of any	*/
/*	data type.						*/
/*								*/
/****************************************************************/

int nameFind(name, ppCHB, type, spliceMask)
char	*name;
CHB	**ppCHB;
short	type;
int	spliceMask;
{
    char	*dot;		/* The location of the dot */
    CHB		*currCHB;	/* The CHB we're working on */
    int		pvLen;		/* Length of PV part of name */


    *ppCHB = 0;
    type &= 0xff;
    dot = (char *) strchr(name, '.');
    pvLen = dot - name;

    currCHB = pHash[hash(name)].pFirstCHB;

    while (currCHB)
	{

	/*
	Don't even look at CHBs that have no CABs. They are delete pending.
	*/

	if (!(currCHB -> pFirstCAB))
	    {
	    currCHB = currCHB -> pNextHash;
	    continue;
	    }

	if (strncmp(currCHB -> chanName, name, pvLen) == 0)
	    {
	    *ppCHB = currCHB;

	    /*
	    Point names match. If the data types are appropriate, we can check
	    for splicing.
	    */

	    if ( (type == (currCHB -> currType & 0xff))
	       || ((type == AGNT_ADAPTIVE) && ((currCHB -> currType & 0xff) == currCHB -> nativeType)) )
		{
	        if ((spliceMask & AGNT_SPLICED) || (strcmp(name + pvLen, ".VAL") == 0))
		    {
		    if (!strcmp(currCHB -> chanName + pvLen, ".VAL"))
		        {
		        if (currCHB -> pISB1 -> splicable)
			    {
			    if (currCHB -> pISB2)
			        {
			        if (currCHB -> pISB2 -> splicable)
				    return (1);
			        }
			    else
			        {
			        return (1);
			        }
			    }
		        }
		    }
		}

	    if ((strcmp(currCHB -> chanName + pvLen, dot) == 0)
	     && ((currCHB -> currType & 0xff) == type))
		{
		if (type == AGNT_ADAPTIVE)	/* Don't know target type yet! */
		    return (0);
		else
		    return (1);
		}
	    }
	currCHB = currCHB -> pNextHash;
	}

    if (*ppCHB)
	return (0);
    else
	return (-1);
}


/****************************************************************/
/*								*/
/*	FIND AN IOC STATUS BLOCK (ISB) GIVEN INET ADDRESS	*/
/*								*/
/*	Create a new one if we can't find the address.		*/
/*								*/
/****************************************************************/

ISB *findISB(inetAddr)
struct in_addr	inetAddr;
{
    ISB		*currISB;

    if (!inetAddr.s_addr)
	return (0);

    currISB = isbList;
    while (currISB)
	{
	if (currISB -> Inet.s_addr == inetAddr.s_addr)
	    return (currISB);
	if (!(currISB -> nextISB))
	    {
	    currISB -> nextISB = (struct IOCSB *) calloc(1, sizeof(ISB));
	    currISB -> nextISB -> Inet = inetAddr;
#ifndef vxWorks
	    /*
	    On workstations, if this is NOT a DPA then set splicable.
	    DPAs are Inet addresses 1 -> 255, 0 is illegal.
	    */

	    if ((currISB -> nextISB -> Inet.s_addr < 0) || (currISB -> nextISB -> Inet.s_addr >= 255))
	        currISB -> nextISB -> splicable = 1;
#endif
#ifdef POINTER64
	    currISB -> nextISB -> isbToken = TokenFromPointer64(currISB);
#endif
	    currISB -> nextISB -> connected = 1;
	    }
	currISB = currISB -> nextISB;
	}

    /*
    Have to create a new one as the very first ISB
    */

    isbList = (struct IOCSB *) calloc(1, sizeof(ISB));
    isbList -> Inet = inetAddr;
    isbList -> connected = 1;
    return (isbList);
}


/****************************************************************/
/*								*/
/*	FIND A RESOURCE CONTROL BLOCK (RCB) GIVEN NAME		*/
/*								*/
/*	Create a new one if we can't find the name.		*/
/*								*/
/****************************************************************/

RCB *findRCB(name)
char	*name;
{
    RCB		*currRCB;

    currRCB = rcbList;
    while (currRCB)
	{
	if (!strcmp(currRCB -> resName, name))
	    return(currRCB);
	if (!currRCB -> pNextRCB)
	    {
	    currRCB -> pNextRCB = (struct RECB *) calloc(1, sizeof(RCB));
	    currRCB -> pNextRCB -> resName = (char *) malloc(strlen(name) + 1);
	    strcpy(currRCB -> pNextRCB -> resName, name);
	    }
	currRCB = currRCB -> pNextRCB;
	}

    /*
    Create the very first one
    */

    rcbList = (struct RECB *) calloc(1, sizeof(RCB));
    rcbList -> resName = (char *) malloc(strlen(name) + 1);
    strcpy(rcbList -> resName, name);
    return (rcbList);
}


/****************************************************************/
/*								*/
/*		CALCULATE HASH VALUE FOR A NAME			*/
/*								*/
/*	The name must be in standardized format. We stop when	*/
/*	we detect the dot. The dot is not included in the	*/
/*	calculation.						*/
/*								*/
/****************************************************************/

static unsigned int hashFactor[] =
    {
    103,   1, 101,   2,  97,   3,  89,   5,  83,   7,
     79,  11,  73,  13,  71,  17,  67,  19,  61,  23,
     59,  29,  53,  31,  47,  37,  43,  41,
    };

int hash(name)
char	*name;
{
    unsigned int total = 0;
    int		i = 0;

    while (*name)
	{
	if (*name == '.')
	    break;
	total += ((unsigned int) *name) * hashFactor[i];
	name++;
	i++;
	}
    return (total % HASH_SIZE);
}


/****************************************************************/
/*								*/
/*	UPDATE THE DATA TYPE, COUNT, AND MASK IN THE CHB	*/
/*								*/
/*	The data type in the CHB (i.e., the one that we get	*/
/*	from the IOC) is the inclusive OR of the data types	*/
/*	in all the CABs. Also, update the element count. The	*/
/*	element count in the CHB is the maximum of the element	*/
/*	counts in all the CABs. The event mask in the CHB is	*/
/*	the inclusive OR of the event mask in all the CABs.	*/
/*	Return a zero if we didn't change the data type, mask,	*/
/*	or element count in the CHB. Return a 1 if we did. 	*/
/*								*/
/*	Extra added attraction for Splicing:			*/
/*								*/
/*	Don't propogate splicing bits to the overall mask for 	*/
/*	the channel.						*/
/*								*/
/****************************************************************/

int updateDataType(pCHB)
CHB	*pCHB;
{
    CAB 	*pCAB;
    short	newType;
    short	newCount;
    short	newMask;
    short	nonSplicedCAB;

    newType = AGNT_NOCONN;
    newCount = 0;
    newMask = 0;
    nonSplicedCAB = 0;

    for (pCAB = pCHB -> pFirstCAB; pCAB; pCAB = pCAB -> pNextChan)
	{
	if (!(pCAB -> pCHB))
	    continue;
	if (pCAB -> pCHB != pCHB)
	    break;

	/*
	Don't include CABS which are just STAT, SEVR, ACK
	in channels that are spliced. However, do remember
	when we see any CAB this is NOT a splice adjunct.
	*/

	if (pCAB -> mask & AGNT_SPLICED)
	    continue;
	nonSplicedCAB = 1;

	newMask |= pCAB -> mask;
	if (pCAB -> dataType != AGNT_NOCONN)
	    {
	    if (newType == AGNT_NOCONN)
		newType = 0;
	    newType |= pCAB -> dataType;
	    }
	if (pCAB -> count > newCount)
	    newCount = pCAB -> count;
	if (pCAB -> pNextChan == pCAB)
	    break;
	if (pCAB -> pNextChan == pCHB -> pFirstCAB)
	    break;
	}
    
    newMask &= ~AGNT_SPLICED;
    newMask &= ~AGNT_SPL_VAL;

    /*
    If there has been NO "real" CAB (as opposed to a
    "real" field) then we have to tell the caller that
    we did make a change to this CHB if it's brand new.
    We tell that by seeing if all possible handles are
    in their "virgin" states. The caller looks for us
    to tell him there's been a change before he'll send
    channel creation/type change messages to an IOC.
    */

    if (!nonSplicedCAB)
	{
	if (pCHB -> handleIOC1 == 0)
	    {
	    if (pCHB -> pISB2)
		{
		if (pCHB -> handleIOC2 == 0)
		    return (1);
		else
		    return (0);
		}
	    return (1);
	    }
	}

    if (newType == pCHB -> currType && newCount == pCHB -> currCount && newMask == pCHB -> currMask)
	return (0);
    if (nonSplicedCAB)
	{
        pCHB -> currType = newType;
        pCHB -> currCount = newCount;
        pCHB -> currMask = newMask;
        return (1);
	}
    else
	return (0);
}


/****************************************************************/
/*								*/
/*		CHECK THE STATE OF THIS CHANNEL			*/
/*								*/
/****************************************************************/

void checkChannel(pCHB)
CHB	*pCHB;
{
    int		status;
    CH_BLDACK	bldAck;
    CAB		*pCAB;
    short	bldType;

    /*
    First see if there are any CABs at all. If there are none, then all users of this
    channel have dropped out. Tell both IOCs to forget about it.
    */

    if (!(pCHB -> pFirstCAB))
	{

	/*
	No further use for this channel. Tell the IOC(s) to let it go.
	If neither one was connected, just unchain the CHB.
	*/

	if ((pCHB -> handleIOC1) || (pCHB -> handleIOC2))
	    {
	    CH_CLOSE	chanClose;

	    /*
	    Send the IOCs messages to close the channel.  Hang onto the CHB until they both acknowledge
	    the message because we don't want this CHB to get used again until its completely purged
	    from the system. If an IOC wasn't connected, then just set its corresponding ISB pointer
	    to zero. That will allow us to unchain the CHB when the only one that was connected gets
	    his close command and acknowledges it.
	    */

	    bldType = htons(CHAN_CLOSE);
#ifndef POINTER64
	    chanClose.chHandle = (int) pCHB;
#else
	    chanClose.chHandle = pCHB -> chbToken;
#endif
	    if ((pCHB -> handleIOC1 != 0) && (pCHB -> handleIOC1 != -1))
		{
		if (pCHB -> pISB1 -> connected)
		    sendScm(pCHB -> pISB1 -> Inet, bldType, (char *) &chanClose, sizeof(chanClose));
		}
	    if ((pCHB -> handleIOC2 != 0) && (pCHB -> handleIOC2 != -1))
		{
		if (pCHB -> pISB2)
		    if (pCHB -> pISB2 -> connected)
			sendScm(pCHB -> pISB2 -> Inet, bldType, (char *) &chanClose, sizeof(chanClose));
		}
	     }

	else
	    {

	    /*
	    No users left. Just delete this CHB from all the chains that referenced it.
	    */

	    unchainCHB(pCHB);
	    }
	return;
	}

    /*
    Still have a use for this one. Find out if the type has to change.
    */

    status = updateDataType(pCHB);

    /*
    We have to either change the data type or create a new channel
    in one or possibly two IOCs.  We could make a very tangled web 
    of code here so that we build each message type only once, but 
    it's much more straightforward (i.e., it might work) to keep our 
    logic neat.
    */

    updateConn(pCHB, pCHB -> pISB1, pCHB -> handleIOC1, status);
    updateConn(pCHB, pCHB -> pISB2, pCHB -> handleIOC2, status);

    /*
    If an initial data sample wasn't wanted, acknowledge the open now!
    */

    if (!(pCHB -> pCurrISB))
	return;
    if (!(pCHB -> pCurrISB -> connected))
	return;

    bldType = htons(CHAN_BLDACK);
    numCABs = findNumCABs(pCHB);
    currCabList = &firstCabList;
    for (cabIndex = 0; cabIndex < numCABs;)
	{
	pCAB = currCabList -> refCAB[cabIndex & 0x1f];
	if (pCAB -> pCHB)
	    {
	    if (pCAB -> awaitAck && pCAB -> readType == AGNT_NOCONN)
	        {
	        bldAck.chHandle = pCAB -> handle;
#ifndef POINTER64
	        bldAck.chIocHandle = (int) pCAB;
#else
	        bldAck.chIocHandle = pCAB -> cabToken;
#endif
		if (pCAB -> mask & AGNT_SPLICED)
		    {
		    bldAck.chDataType = htons(AGNT_ENUM);
		    bldAck.chElementCount = htons(1);
		    }
		else
		    {
	            bldAck.chDataType = htons(pCHB -> nativeType);
	            bldAck.chElementCount = htons(pCHB -> nativeCount);
		    }
	        appTx(pCAB -> pACB, bldType, (char *) &bldAck, sizeof(bldAck));
	        pCAB -> awaitAck = 0;
	        }
	    }
	cabIndex++;
	if ((cabIndex & 0x1f) == 0)
	    currCabList = currCabList -> nextList;
	}
}


/****************************************************************/
/*								*/
/*		UPDATE CHANNEL TYPE OR BUILD CONNECTION		*/
/*								*/
/****************************************************************/

void updateConn(pCHB, pISB, iocHandle, typeChanged)
CHB	*pCHB;
ISB	*pISB;
int	iocHandle;
int	typeChanged;
{
    char	*msgPtr = NULL;
    int		msgSize;
    CH_TYPE	changeType;
    short	msgType = 0;
    struct
	{
	CH_BLDRDMON	chBldRdMon;
	char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
	} bldRdMonMsg;
    struct
	{
	CH_BUILD	chanBuild;
	char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
	} buildMsg;
    struct
	{
	CH_QBLD		chQbld;
	char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
	} quickBldMsg;

    msgSize = 0;

    /*
    See if its possible for this connection to exist
    */

    if (!pISB)
	return;

    /*
    See if just a type change will suffice or if we have to connect.
    */

    if ((iocHandle != 0) && (iocHandle != -1))
	{
	if (typeChanged)
	    {
	    changeType.chHandle = iocHandle;
	    changeType.chType = htons(pCHB -> currType);
	    if ((pCHB -> currType & 0xff) == AGNT_ADAPTIVE)
		changeType.chType = htons((short)(pCHB -> nativeType | AGNT_STAT));
	    changeType.chCount = htons((short)pCHB -> currCount);
	    if ((pCHB -> currType & 0xff) == AGNT_ADAPTIVE)
		changeType.chCount = htons(pCHB -> nativeCount);
	    changeType.chMask = htons(pCHB -> currMask);
	    if ((pCHB -> currType & 0xff) == AGNT_ADAPTIVE)
		changeType.chMask = htons(AGNT_VALUE | AGNT_ALARM);
	    msgType = htons(CHAN_TYPE);
	    msgPtr = (char *) &changeType;
	    msgSize = sizeof(changeType);
#if !defined(vxWorks) && !defined(vms)
	    if (pISB -> connected)
		tisSysLog(2, 0, "\tChanging data type for %s", pCHB -> chanName);
#else
	    if ((cacDebug >= 2) && pISB -> connected)
		printf("\tChanging data type for %s\n", pCHB -> chanName);
#endif
	    }
	}

    else if ((iocHandle == 0) && pISB -> connected)
	{

	/*
	IOC is online. Send him a message to make him create the channel.
	*/

	if ((pCHB -> currType & 0xff) != AGNT_ADAPTIVE)
	    {
	    msgType = htons(CHAN_BUILD);
#ifndef POINTER64
	    buildMsg.chanBuild.chHandle = (int) pCHB;
#else
	    buildMsg.chanBuild.chHandle = pCHB -> chbToken;
#endif
#if !defined(vxWorks) && !defined(vms)
	    if (pISB -> connected)
#ifndef POINTER64
		tisSysLog(2, 0, "\tSending BUILD to IOC for %x, handle = 0x%X", pCHB -> chanName,
			(int) pCHB);
#else
		tisSysLog(2, 0, "\tSending BUILD to IOC for %x, handle = 0x%X", pCHB -> chanName,
			pCHB -> chbToken);
#endif
#else
	    if ((cacDebug >= 2) && pISB -> connected)
	        printf("\tSending BUILD to IOC for %s, handle = 0x%X\n", pCHB -> chanName, (int) pCHB);
#endif
	    buildMsg.chanBuild.chType = htons(pCHB -> currType);
	    buildMsg.chanBuild.chCount = htons((short)pCHB -> currCount);
	    buildMsg.chanBuild.chNameLen = strlen(pCHB -> chanName) + 1;
	    msgPtr = (char *) &buildMsg;
	    msgSize = sizeof(CH_BUILD) + buildMsg.chanBuild.chNameLen;
	    buildMsg.chanBuild.chNameLen = htons(buildMsg.chanBuild.chNameLen);
	    strcpy(buildMsg.chanName, pCHB -> chanName);
	    }

	else if ((pISB -> Inet.s_addr >= 1) && (pISB -> Inet.s_addr <= 255))
	    {

	    /*
	    This is a DPA. Don't use the QUICKBUILD transactions because his points
	    are most likely not in the database. He does understand BLDRDMONs, however.
	    */

	    msgType = htons(CHAN_BLDRDMON);
#ifndef POINTER64
	    bldRdMonMsg.chBldRdMon.chHandle = (int) pCHB;
	    bldRdMonMsg.chBldRdMon.chCaHandle = (int) (pCHB -> pFirstCAB);
	    bldRdMonMsg.chBldRdMon.chAppHandle = (int) (pCHB -> pFirstCAB -> pACB);
#else
	    bldRdMonMsg.chBldRdMon.chHandle = pCHB -> chbToken;
	    bldRdMonMsg.chBldRdMon.chCaHandle = pCHB -> pFirstCAB -> cabToken;
	    bldRdMonMsg.chBldRdMon.chAppHandle = pCHB -> pFirstCAB -> pACB -> acbToken;
#endif
	    bldRdMonMsg.chBldRdMon.chNameLen = strlen(pCHB -> chanName) + 1;
	    msgPtr = (char *) &bldRdMonMsg;
	    msgSize = sizeof(CH_BLDRDMON) + bldRdMonMsg.chBldRdMon.chNameLen;
	    bldRdMonMsg.chBldRdMon.chNameLen = htons(bldRdMonMsg.chBldRdMon.chNameLen);
	    strcpy(bldRdMonMsg.chanName, pCHB -> chanName);
	    }

	else
	    {

	    /*
	    Use the new QUICKBUILD transactions to minimize the bandwidth that we use.
	    */

	    msgType = htons(CHAN_QBUILD);
#ifndef POINTER64
	    quickBldMsg.chQbld.chHandle = (int) pCHB;
	    quickBldMsg.chQbld.chCaHandle = (int) (pCHB -> pFirstCAB);
#else
	    quickBldMsg.chQbld.chHandle = pCHB -> chbToken;
	    quickBldMsg.chQbld.chCaHandle = (int) (pCHB -> pFirstCAB -> cabToken);
#endif
	    
	    quickBldMsg.chQbld.chNameLen = strlen(pCHB -> chanName) + 1;
	    msgSize = sizeof(CH_QBLD) + quickBldMsg.chQbld.chNameLen;
	    msgPtr = (char *) &quickBldMsg;
	    quickBldMsg.chQbld.chNameLen = htons(quickBldMsg.chQbld.chNameLen);
	    strcpy(quickBldMsg.chanName, pCHB -> chanName);
	    }

	/*
	When the handle is set to -1, it means that we have requested a channel from
	the IOC but haven't yet got a BLDACK from him. If this is true, then don't
	request another connection while the first one is still pending.
	*/

	if (pCHB -> pISB1 == pISB)
	    {
	    if (pCHB -> handleIOC1 == -1)
	 	msgSize = 0;
	    else
	        pCHB -> handleIOC1 = -1;
	    }
	else
	    {
	    if (pCHB -> handleIOC2 == -1)
	 	msgSize = 0;
	    else
	        pCHB -> handleIOC2 = -1;
	    }
	}

    /*
    See if a message has to go out.
    */

    if (msgSize && pISB -> connected)
	sendScm(pISB -> Inet, msgType, msgPtr, msgSize);
}


/****************************************************************/
/*								*/
/*	UNCHAIN A CHB FROM EVERYTHING AND THEN DELETE IT	*/
/*								*/
/****************************************************************/

void unchainCHB(pCHB)
CHB 	*pCHB;
{
#ifndef vxWorks
    CALINK		*pCurrValue;
    VALUE_COMMON	*valCommon;
#endif

   /*
   Unchain from the first ISB. If there was no previous CHB, then
   we must have been the start of the list.
   */

   if (pCHB -> pPrevIOC1)
	{
	if (pCHB -> pPrevIOC1 -> pISB1 == pCHB -> pISB1)
	    pCHB -> pPrevIOC1 -> pNextIOC1 = pCHB -> pNextIOC1;
	else
	    pCHB -> pPrevIOC1 -> pNextIOC2 = pCHB -> pNextIOC1;
	}
   else
	pCHB -> pISB1 -> pFirstCHB = pCHB -> pNextIOC1;
   if (pCHB -> pNextIOC1)
	{
	if (pCHB -> pNextIOC1 -> pISB1 == pCHB -> pISB1)
	    pCHB -> pNextIOC1 -> pPrevIOC1 = pCHB -> pPrevIOC1;
	else
	    pCHB -> pNextIOC1 -> pPrevIOC2 = pCHB -> pPrevIOC1;
	}

    /*
    Also unchain from the second ISB, if there is one.
    */

    if (pCHB -> pISB2)
	{
	if (pCHB -> pPrevIOC2)
	    {
	    if (pCHB -> pPrevIOC2 -> pISB1 == pCHB -> pISB2)
		pCHB -> pPrevIOC2 -> pNextIOC1 = pCHB -> pNextIOC2;
	    else
		pCHB -> pPrevIOC2 -> pNextIOC2 = pCHB -> pNextIOC2;
	    }
	else
	    pCHB -> pISB2 -> pFirstCHB = pCHB -> pNextIOC2;
	if (pCHB -> pNextIOC2)
	    {
	    if (pCHB -> pNextIOC2 -> pISB1 == pCHB -> pISB2)
		pCHB -> pNextIOC2 -> pPrevIOC1 = pCHB -> pPrevIOC2;
	    else
		pCHB -> pNextIOC2 -> pPrevIOC2 = pCHB -> pPrevIOC2;
	    }
	}

    /*
    Unchain from the RCB.
    */

    if (pCHB -> pNextRes)
	pCHB -> pNextRes -> pPrevRes = pCHB -> pPrevRes;
    if (pCHB -> pPrevRes)
	pCHB -> pPrevRes -> pNextRes = pCHB -> pNextRes;
    else
	pCHB -> pRes -> pFirstCHB = pCHB -> pNextRes;

    /*
    There is only a single linked list for the hash table because we don't anticipate that there
    will be a lot of points with the same hash value.
    */

    if (pCHB -> pHash -> pFirstCHB == pCHB)
	pCHB -> pHash -> pFirstCHB = pCHB -> pNextHash;
    else
	{
	CHB	*currCHB;

	currCHB = pCHB -> pHash -> pFirstCHB;
	while (currCHB && (currCHB -> pNextHash != pCHB))
	    currCHB = currCHB -> pNextHash;
	if (currCHB)
	    currCHB -> pNextHash = pCHB -> pNextHash;
	}

    /*
    Now free the CHB so that it can be used again.
    */

#ifndef vxWorks
    if (isValField(pCHB))
	{
	pCurrValue = FA_GetValuePtr(pCHB -> chanName);
	if (pCurrValue)
	    {
	    valCommon = pCurrValue -> pData;
	    deleteValue(pCurrValue, valCommon -> nativeType);
	    FA_SetValuePtr(pCHB -> chanName, (void *) 0);
	    }
	}
#endif
    deleteCHB(pCHB);
}


/****************************************************************/
/*								*/
/*	SET CHANNEL TYPE IN A PREVIOUSLY UNCONNECTED CAB	*/
/*								*/
/****************************************************************/

void adjustType(pCAB)
CAB	*pCAB;
{
    CHB		*otherCHB;
    CHB		*pCHB;
    CAB		*otherCAB;

    /*
    First, see if there is another CHB that already has the desired data type on the same channel.
    Make sure to ignore CHBs that are delete pending.
    */

    pCHB = pCAB -> pCHB;
    if (!pCHB)
	return;
    for (otherCHB = pCHB -> pHash -> pFirstCHB; otherCHB; otherCHB = otherCHB -> pNextHash)
	{
	if (pCHB == otherCHB)
	    continue;
	if (!(otherCHB -> pFirstCAB))
	    continue;
	if ((pCAB -> mask & AGNT_SPLICED) && isValField(otherCHB))
	    break;
	if ((otherCHB -> currType & 0xff) != (pCAB -> dataType & 0xff))
	    continue;
	if (!strcmp(otherCHB -> chanName, pCHB -> chanName))
	    break;
	}
    if (otherCHB)
	{

	/*
	There was another CHB. Unchain our ACB from the old CHB and put it onto the new one.
	*/

	if (pCHB -> pFirstCAB == pCAB)
	    pCHB -> pFirstCAB = pCAB -> pNextChan;
	else
	    {
	    otherCAB = pCHB -> pFirstCAB;
	    while (otherCAB -> pNextChan != pCAB)
		otherCAB = otherCAB -> pNextChan;
	    otherCAB -> pNextChan = pCAB -> pNextChan;
	    }
	pCAB -> pNextChan = otherCHB -> pFirstCAB;
	otherCHB -> pFirstCAB = pCAB;
	pCAB -> pCHB = otherCHB;

	/*
	We may have to adjust the data type and/or mask on the CHB we just joined. The
	mainline code that just called us will take care of the old CHB.
	*/

	checkChannel(otherCHB);
	}

    else
	{

	/*
	Another CHB couldn't be found. If we are the only CAB on this CHB, the
	mainline code that just called us will take care of type checking.
	*/

	if ((pCHB -> pFirstCAB != pCAB) || (pCAB -> pNextChan != 0))
	    {

	    /*
	    There is more than one CHB. We have to create a new one and move the CAB
	    from the old CHB to the new one.
	    */

	    otherCHB = newCHB();

	    /*
	    Copy the file name from the original CHB. Set the data type to "not connected"
	    and the count to zero. These will get adjusted when checkChannel gets called.
	    */

	    strcpy(otherCHB -> chanName, pCHB -> chanName);
	    otherCHB -> currType = AGNT_NOCONN;
	    otherCHB -> currCount = 0;

	    /*
	    Just copy the pointers to the ISBs, the current ISB, the native type and the
	    native count.
	    */

	    otherCHB -> pISB1 = pCHB -> pISB1;
	    otherCHB -> pISB2 = pCHB -> pISB2;
	    otherCHB -> pCurrISB = pCHB -> pCurrISB;
	    otherCHB -> nativeType = pCHB -> nativeType;
	    otherCHB -> nativeCount = pCHB -> nativeCount;
	    otherCHB -> checksum = pCHB -> checksum;
	    otherCHB -> pointType = pCHB -> pointType;
	    otherCHB -> creationTime = pCHB -> creationTime;

	    /*
	    Jump into the ISB chains after the old ISB.
	    */

	    otherCHB -> pNextIOC1 = pCHB -> pNextIOC1;
	    if (otherCHB -> pNextIOC1)
		{
		if (otherCHB -> pNextIOC1 -> pISB1 == otherCHB -> pISB1)
		    otherCHB -> pNextIOC1 -> pPrevIOC1 = otherCHB;
		else
		    otherCHB -> pNextIOC1 -> pPrevIOC2 = otherCHB;
		}
	    pCHB -> pNextIOC1 = otherCHB;
	    otherCHB -> pPrevIOC1 = pCHB;

	    if (otherCHB -> pISB2)
		{
		otherCHB -> pNextIOC2 = pCHB -> pNextIOC2;
		if (otherCHB -> pNextIOC2)
		    {
		    if (otherCHB -> pNextIOC2 -> pISB1 == otherCHB -> pISB2)
			otherCHB -> pNextIOC2 -> pPrevIOC1 = otherCHB;
		    else
			otherCHB -> pNextIOC2 -> pPrevIOC2 = otherCHB;
		    }
		pCHB -> pNextIOC2 = otherCHB;
		otherCHB -> pPrevIOC2 = pCHB;
		}

	    /*
	    Jump into the resource chain after the old ISB
	    */

	    otherCHB -> pNextRes = pCHB -> pNextRes;
	    pCHB -> pNextRes = otherCHB;
	    if (otherCHB -> pNextRes)
		otherCHB -> pNextRes -> pPrevRes = otherCHB;
	    otherCHB -> pPrevRes = pCHB;
	    otherCHB -> pRes = pCHB -> pRes;

	    /*
	    Jump into the hash chain.
	    */

	    otherCHB -> pNextHash = pCHB -> pNextHash;
	    pCHB -> pNextHash = otherCHB;
	    otherCHB -> pHash = pCHB -> pHash;

	    /*
	    Unchain the CAB from the old CHB and put it on the new one.
	    */


	    if (pCHB -> pFirstCAB == pCAB)
	        pCHB -> pFirstCAB = pCAB -> pNextChan;
	    else if (pCHB -> pFirstCAB)
	        {
	        otherCAB = pCHB -> pFirstCAB;
	        while (otherCAB -> pNextChan != pCAB)
		    {
		    otherCAB = otherCAB -> pNextChan;
		    if (!otherCAB)
			break;
		    }
		if (otherCAB)
	            otherCAB -> pNextChan = pCAB -> pNextChan;
	        }
	    pCAB -> pNextChan =  0;
	    otherCHB -> pFirstCAB = pCAB;

	    /*
	    Make the CAB point to the new CHB.
	    */

	    pCAB -> pCHB = otherCHB;

	    /*
	    Now establish the new channel in the IOCs as necessary.
	    */

	    checkChannel(otherCHB);
	    }
	}
}


/****************************************************************/
/*								*/
/*			REJECT A CHANNEL			*/
/*								*/
/*	This routine is called whenever we get a CHB address	*/
/*	from an IOC and we can't find that address in our	*/
/*	list of channels. Format a channel close message and	*/
/*	send it to the IOC. There's no point in getting		*/
/*	messages that we can't handle. Shut them off at the	*/
/*	source.							*/
/*								*/
/****************************************************************/

void rejectChan(pISB, handle)
ISB	*pISB;
int	handle;
{
    short 	msgType;
    CH_CLOSE	chanClose;

    if (pISB -> connected)
	{
        chanClose.chHandle = handle;
	msgType = htons(CHAN_CLOSE);
	sendScm(pISB -> Inet, msgType, (char *) &chanClose, sizeof(chanClose));
	}
}


/****************************************************************/
/*								*/
/*		SEND DATA OUT OF THE MOBY BUFFER		*/
/*								*/
/*	The moby buffer is used to contain data that we want	*/
/*	to send to sockMnt but don't have room for in our	*/
/*	socket's transmit buffer.				*/
/*								*/
/****************************************************************/

#ifndef vxWorks
static int sendMoby()
{
    int writeSize;
    int bytesSent;

    writeSize = cacExtBytes;
    if (writeSize > 4096)
	writeSize = 4096;
    if ((writeSize + cacRingOut) > sizeof(cacMobyBuff))
    writeSize = sizeof(cacMobyBuff) - cacRingOut;
    bytesSent = send(scmFd, cacMobyBuff + cacRingOut, writeSize, 0);
    if (bytesSent == -1)
        {
#ifdef WIN32
	errno = WSAGetLastError ();
#endif WIN32
	if (errno = EAGAIN)
	    {
#ifdef WIN32
	    Sleep (1000);
#else
	    sleep(1);
#endif WIN32
	    return 0;
	    }

#ifndef vms
        tisSysLog(0, 0, "chanAcq quits: I/O fail on catch up: %m");
#ifndef WIN32
	kill(scmPid, 3);
#endif WIN32
	abort();
#else
	perror("chanAcq quits: I/O fail on catch up");
        exit(1);
#endif vms

        }
    cacExtBytes -= bytesSent;
    if (cacExtBytes > 0)
        {
        cacRingOut += bytesSent;
        if (cacRingOut >= sizeof(cacMobyBuff))
	    cacRingOut = 0;
        }
    else if (cacExtBytes == 0)
	{
	cacRingOut = 0;
	cacRingIn = 0;
	}
    else
        {

#ifndef vms
        tisSysLog(0, 0, "chanAcq quits: bookkeeping error!");
#ifndef WIN32
	kill(scmPid, 3);
#endif WIN32
	abort();
#else
	fprintf(stderr, "chanAcq quits: bookkeeping error!");
        exit(1);
#endif vms

        }
}
#endif vxWorks

/****************************************************************/
/*								*/
/*	DETERMINE IF WE HAVE A COMPLETE MESSAGE FROM sockMnt	*/
/*								*/
/****************************************************************/

int haveCompleteSockMsg()
{
    short	readBytes;
    SM_READ	*smRead;

    /*
    Must have at least a message type
    */

    if (nchib < sizeof(short))
	return (0);

    /*
    Get the message type to see how much stuff must follow. Don't flip
    endians because we must always be in the same CPU as sockMnt.
    */

    bcopy(rxBuff + bufOutPos, (char *) &readBytes, sizeof(readBytes));
    if (readBytes == CHAN_IOCDN)
	{

	/*
	IOC Down message. Must have sizeof(CH_IOCDN)
	*/

	if (nchib >= (sizeof(short) + sizeof(CH_IOCDN)))
	    return (1);
	else
	    return (0);
	}
 
#if defined(vxWorks) || defined(vms)
    else if (readBytes == CHAN_GOTNMSRV)
	{

	/*
	Name server found. Must have sizeof(CH_GOTNMSRV)
	*/

	if (nchib >= (sizeof(short) + sizeof(CH_GOTNMSRV)))
	    return (1);
	else
	    return (0);
	}
#endif vxWorks or vms

    else if (readBytes == CHAN_IOCUP)
	{

	/*
	IOC Up message. Must have sizeof(CH_IOCUP)
	*/

	if (nchib >= (sizeof(short) + sizeof(CH_IOCUP)))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == SOCK_READ)
	{

	/*
	General socket read wrapper. Must have sizeof(SM_READ)
	plus the number of bytes specified in the wrapper.
	*/

	if (nchib < (sizeof(short) + sizeof(SM_READ)))
	    return (0);
	smRead = (SM_READ *)(rxBuff + bufOutPos + sizeof(short));
	bcopy ((char *) &(smRead -> smLength), (char *)&readBytes, sizeof(readBytes));
	readBytes = ntohs(readBytes);
	if (nchib >= (int)(sizeof(short) + sizeof(SM_READ) + readBytes))
	    return (1);
	else
	    return (0);
	}

    else
	{
	printf("chanAcq sockMnt sent us unknown message code %d!\n", readBytes);
	return (1);
	}
}


/****************************************************************/
/*								*/
/*	DETERMINE IF WE HAVE A COMPLETE MESSAGE FROM AN APP	*/
/*								*/
/****************************************************************/

int haveCompleteAppMsg(pACB)
ACB	*pACB;
{
    short		readBytes;
    CH_BUILD		*chBuild;
    CH_DIRPUT		*chDirput;
    CH_GETPARAMS	*chGetparams;
    CH_ALARM_RECV	*chAlarmRecv;
    CH_ALARM_ACK	*chAlarmAck;
    CH_ALARM_CTL	*chAlarmCtl;

    /*
    Must have at least a message type
    */

    if (pACB -> nchib < sizeof(short))
	return (0);

    /*
    Get the message type, see what must follow
    */

    bcopy(pACB -> appRxBuff + pACB -> rcvOutPos, (char *) &readBytes, sizeof(readBytes));

    if (readBytes == CHAN_RGST)
	return (1);

    readBytes = ntohs(readBytes);
    if (readBytes == DEBUG)
	return (1);

    else if (readBytes == DUMP)
	return (1);

    else if (readBytes == SOCK_FLUSH)
	return (1);

    else if (readBytes == CHAN_DIRREAD)
	{
	if (pACB -> nchib >= (sizeof(short) + sizeof(CH_DIRREAD)))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == CHAN_MONADD)
	{
	if (pACB -> nchib >= (sizeof(short) + sizeof(CH_MONADD)))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == CHAN_MONREM)
	{
	if (pACB -> nchib >= (sizeof(short) + sizeof(CH_MONREM)))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == CHAN_CLOSE)
	{
	if (pACB -> nchib >= (sizeof(short) + sizeof(CH_CLOSE)))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == CHAN_HANDLERS)
	{
	if (pACB -> nchib >= (sizeof(short) + sizeof(CH_HANDLERS)))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == CHAN_ALARM_RECV)
	{
	if (pACB -> nchib < (sizeof(short) + sizeof(CH_ALARM_RECV)))
	    return (0);
	chAlarmRecv = (CH_ALARM_RECV *)(pACB -> appRxBuff + pACB -> rcvOutPos + sizeof(short));
	bcopy((char *) &(chAlarmRecv -> alarmNameLen), (char *) &readBytes, sizeof(readBytes));
	readBytes = ntohs(readBytes);
	if (pACB -> nchib >= (int)(sizeof(short) + sizeof(CH_ALARM_RECV) + readBytes))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == CHAN_BUILD)
	{
	if (pACB -> nchib < (sizeof(short) + sizeof(CH_BUILD)))
	    return (0);
	chBuild = (CH_BUILD *)(pACB -> appRxBuff + pACB -> rcvOutPos + sizeof(short));
	bcopy((char *) &(chBuild -> chNameLen), (char *) &readBytes, sizeof(readBytes));
	readBytes = ntohs(readBytes);
	if (pACB -> nchib >= (int)(sizeof(short) + sizeof(CH_BUILD) + readBytes))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == CHAN_DIRPUT)
	{
	if (pACB -> nchib < (sizeof(short) + sizeof(CH_DIRPUT)))
	    return (0);
	chDirput = (CH_DIRPUT *)(pACB -> appRxBuff + pACB -> rcvOutPos + sizeof(short));
	bcopy((char *) &(chDirput -> chDataSize), (char *) &readBytes, sizeof(readBytes));
	readBytes = ntohs(readBytes);
	if (pACB -> nchib >= (int)(sizeof(short) + sizeof(CH_DIRPUT) + readBytes))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == CHAN_GETPARAMS)
	{
	if (pACB -> nchib < (sizeof(short) + sizeof(CH_GETPARAMS)))
	    return (0);
	chGetparams = (CH_GETPARAMS *) (pACB -> appRxBuff + pACB -> rcvOutPos + sizeof(short));
	bcopy((char *) &(chGetparams -> chNameLen), (char *) &readBytes, sizeof(readBytes));
	readBytes = ntohs(readBytes);
	if (pACB -> nchib >= (int)(sizeof(short) + sizeof(CH_GETPARAMS) + readBytes))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == CHAN_ALARM_ACK)
	{
	if (pACB -> nchib < (sizeof(short) + sizeof(CH_ALARM_ACK)))
	    return (0);
	chAlarmAck = (CH_ALARM_ACK *) (pACB -> appRxBuff + pACB -> rcvOutPos + sizeof(short));
	bcopy((char *) &(chAlarmAck -> alarmNameLen), (char *) &readBytes, sizeof(readBytes));
	readBytes = ntohs(readBytes);
	if (pACB -> nchib >= (int)(sizeof(short) + sizeof(CH_ALARM_ACK) + readBytes))
	    return (1);
	else
	    return (0);
	}

    else if (readBytes == CHAN_ALARM_CTL)
	{
	if (pACB -> nchib < (sizeof(short) + sizeof(CH_ALARM_CTL)))
	    return (0);
	chAlarmCtl = (CH_ALARM_CTL *) (pACB -> appRxBuff + pACB -> rcvOutPos + sizeof(short));
	bcopy((char *) &(chAlarmCtl -> alarmNameLen), (char *) &readBytes, sizeof(readBytes));
	readBytes = ntohs(readBytes);
	if (pACB -> nchib >= (int)(sizeof(short) + sizeof(CH_ALARM_CTL) + readBytes))
	    return (1);
	else
	    return (0);
	}

    else
	{
	printf("chanAcq: Application sent unknown message code %d\n", readBytes);
	return (1);
	}
}


/****************************************************************/
/*								*/
/*		READ BYTES OUT OF BUFFER FROM sockMnt		*/
/*								*/
/*	We don't ever have to worry about having enough		*/
/*	data in the buffer. haveCompleteSockMsg() took care	*/
/*	of that.						*/
/****************************************************************/

void scmRead(buff, nbytes)
char	*buff;
int	nbytes;
{
    bcopy(rxBuff + bufOutPos, buff, nbytes);
    bufOutPos += nbytes;
    nchib -= nbytes;
}


/****************************************************************/
/*								*/
/*		READ THE SOCKET FROM sockMnt			*/
/*								*/
/****************************************************************/

int fillSockBuff()
{
    int		bytesGot;

    /*
    Slide down anything that might already be in the buffer
    */

    if (nchib && bufOutPos)
	bcopy(rxBuff + bufOutPos, rxBuff, nchib);

    /*
    Now read as much as will fit into the remaining room in the
    receive buffer.
    */

    bytesGot = recv(scmFd, rxBuff + nchib, sizeof(rxBuff) - nchib, 0);
    if (bytesGot == -1)
	{

#if !defined(vxWorks) && !defined(vms)
	tisSysLog(0, 0, "chanAcq gets read error from sockMnt: %m");
#else
	perror("chanAcq gets read error from sockMnt");
#endif vxWorks

	return (0);
	}

    else if (bytesGot == 0)
	{

#if !defined(vxWorks) && !defined(vms) && !defined(WIN32)
	syslog(LOG_ALERT | LOG_DAEMON, "chanAcq get zero length read from sockMnt");
#else !defined(vxWorks) && !defined(vms) && !defined(WIN32)
#ifdef vxWorks
	logMsg ("chanAcq gets zero length read from sockMnt\n",0,0,0,0,0,0);
#else vxWorks
	printf("chanAcq gets zero length read from sockMnt\n");
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms) && !defined(WIN32)

	return (0);
	}

#ifndef vxWorks
#ifndef vms
    if (traceFd)
	fprintf(traceFd, "Rx from sockMnt, %d bytes\n", bytesGot);
#endif
#endif

    nchib += bytesGot;
    bufOutPos = 0;
    return (1);
}


/****************************************************************/
/*								*/
/*		READ THE SOCKET FROM AN APPLICATION		*/
/*								*/
/****************************************************************/

int fillAppBuff(pACB)
ACB	*pACB;
{
    int		bytesGot;

    /*
    Slide down anything that might already be in the buffer
    */

    if ((pACB -> nchib) && (pACB -> rcvOutPos))
	bcopy(pACB -> appRxBuff + pACB -> rcvOutPos, pACB -> appRxBuff, pACB -> nchib);

    /*
    Now read as much as will fit into the remaining room in the
    receive buffer.
    */

    if (unregistered)
        bytesGot = recv(pACB -> sockFd, pACB -> appRxBuff + pACB -> nchib, sizeof(short), 0);
    else
        bytesGot = recv(pACB -> sockFd, pACB -> appRxBuff + pACB -> nchib, sizeof(pACB -> appRxBuff) - pACB -> nchib, 0);
    if (bytesGot == -1)
	{
#if !defined(vxWorks) && !defined(vms)
	char	errnoMsg[32];

#ifdef WIN32
	errno = WSAGetLastError ();
#endif WIN32

	sprintf(errnoMsg, "errno = %d", errno);
	tisSysLog(1, 0, "chanAcq gets read error from app: %m");
	if (pACB -> trendHandler)
	    tisSysLog(0, 0, "chanAcq drops trend handler because of read error");
	if (pACB -> alarmHandler)
	    tisSysLog(0, 0, "chanAcq drops alarm handler because of read error");
	if (pACB -> trendHandler || pACB -> alarmHandler)
	    tisSysLog(0, 0, errnoMsg);
#else
	if (cacDebug >= 1)
	    perror("chanAcq gets read error from app");
#endif vxWorks

	FD_CLR(pACB -> sockFd, &cacfdMask);
#ifdef WIN32
	closesocket(pACB -> sockFd);
#else
	close(pACB -> sockFd);
#endif WIN32
	pACB -> sockFd = INVALID_SOCKET;
	return (0);
	}

    else if (bytesGot == 0)
	{
	FD_CLR(pACB -> sockFd, &cacfdMask);
#ifdef WIN32
	closesocket(pACB -> sockFd);
#else
	close(pACB -> sockFd);
#endif WIN32
	pACB -> sockFd = INVALID_SOCKET;
#ifndef vxWorks
	if (pACB -> trendHandler)
	    tisSysLog(0, 0, "chanAcq drops trend handler because of EOF");
	if (pACB -> alarmHandler)
	    tisSysLog(0, 0, "chanAcq drops alarm handler because of EOF");
#endif
	return (0);
	}

#ifndef vxWorks
#ifndef vms
    if (traceFd)
	fprintf(traceFd, "Rx from app, %d bytes\n", bytesGot);
#endif
#endif

    pACB -> nchib += bytesGot;
    pACB -> rcvOutPos = 0;
    return (1);
}


/****************************************************************/
/*								*/
/*	FIND THE NUMBER OF CABs ON A CHB AND BUILD A LIST	*/
/*								*/
/*	It is possible the while we are traversing the		*/
/*	list of CABS on a CHB for the purpose of sending	*/
/*	the application a message (like a connection event)	*/
/*	that something may go wrong with the socket. We don't	*/
/*	want to get lost if appTx() should call delApp()	*/
/*	and change the linkage. Therefore, any time that we	*/
/*	are going through all of the CABs on a CHB we first	*/
/*	build a list and then we run down that list. A change	*/
/*	to the channel close processing and delApp has		*/
/*	made sure that any inactive CAB has pCHB set to zero.	*/
/*	Just don't process those CABs because they are really	*/
/*	on the free list rather than on the CHB. Phew.		*/
/*								*/
/****************************************************************/

int findNumCABs(pCHB)
CHB	*pCHB;
{
    int			numFound = 0;
    struct cabList	*currList = &firstCabList;
    CAB			*pCAB = pCHB -> pFirstCAB;

    while (pCAB)
	{
	if (!(pCAB -> pCHB))
	    break;
	if (pCAB -> pCHB != pCHB)
	    break;
	
	currList -> refCAB[numFound & 0x1f] = pCAB;
	numFound++;
	if ((numFound & 0x1f) == 0)
	    {
	    if (currList -> nextList == 0)
		currList -> nextList = (struct cabList *) malloc(sizeof(struct cabList));
	    currList = currList -> nextList;
	    }
	if (pCAB -> pNextChan == pCAB)
	    break;
	pCAB = pCAB -> pNextChan;
	if (pCAB == pCHB -> pFirstCAB)
	    break;
	}
    return (numFound);
}


/****************************************************************/
/*								*/
/*	CLOSE NAME SERVER SOCKET AND REQUEST ANOTHER ONE	*/
/*								*/
/****************************************************************/

#if defined(vxWorks) || defined(vms)
void closeNameSocket()
{
    short	msgType;
    struct sockaddr_in	portAddr;
    NSQ		*nsqResend;
    struct
	{
	NS_SRCH	nsSrch;
	char	chanName[PVNAME_SZ + FLDNAME_SZ + 2];
	} resendMsg;

    if (namFd != INVALID_SOCKET)
	{
	FD_CLR(namFd, &cacfdMask);
#ifdef WIN32
	closesocket(namFd);
#else
	close(namFd);
#endif WIN32
	namFd = INVALID_SOCKET;
	namNchib = 0;
	namBufOutPos = 0;
	}

    /*
    try local name server. Connect to him.
    */
    portAddr.sin_family = AF_INET;
    portAddr.sin_addr = caLocalInet;
    portAddr.sin_port = htons(NAM_PORT);
    namFd = socket(PF_INET, SOCK_STREAM, 0);
    if (namFd < 0)
	{
	perror("chanAcq can't create socket to nameSrv");
#ifdef vxWorks
	taskSuspend(taskIdSelf());	/* For IOC debugging help */
#else
	exit(1);
#endif vxWorks
	}
    if (connect(namFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1)
	{
#ifdef WIN32
	closesocket(namFd);
#else
	close(namFd);
#endif WIN32
	namFd = INVALID_SOCKET;
	/* try for external name server */
	msgType = CHAN_NEEDNMSRV;
	if ((scmOutputBytes + sizeof(msgType)) > sizeof(scmBuff))
	    scmFlush();
	bcopy((char *) &msgType, scmBuff + scmOutputBytes, sizeof(msgType));
	scmOutputBytes += sizeof(msgType);
	scmFlush();
	}
    else
	{
	/* we're in! */
	setsockopt(namFd, IPPROTO_TCP, TCP_NODELAY, (char *)&true, sizeof(true));
	intsize = sizeof(int);
	getsockopt(namFd, SOL_SOCKET, SO_RCVBUF, (char *)&namRxSize, &intsize);
	intsize = sizeof(int);
	getsockopt(namFd, SOL_SOCKET, SO_SNDBUF, (char *)&namTxSize, &intsize);
	namRxTopped = 0;
	if (namFd > maxFd)
	    maxFd = namFd;
	FD_SET(namFd, &cacfdMask);

	/*
	Name server now connected. Resend all of the stuff in our queue.
	*/

	if (cacDebug >= 1)
	    printf("Name Server resend!\n");

	for (nsqResend = nsqList; nsqResend; nsqResend = nsqResend -> pNextNSQ)
	    {
	    msgType = htons(NAME_SRCH);
	    if (!caSendNam((char *) &msgType, sizeof(msgType)))
		{
		closeNameSocket();
		break;
		}
	    resendMsg.nsSrch.nsHandle = (int) nsqResend;
	    resendMsg.nsSrch.nsNameLen = strlen(nsqResend -> name) + 1;
	    resendMsg.nsSrch.nsNameLen = htons(resendMsg.nsSrch.nsNameLen);
	    nsqResend -> dueTime = time(NULL) + caNameSrvTimeout;
	    strcpy(resendMsg.chanName, nsqResend -> name);
	    if (!caSendNam((char *) &resendMsg,
		    sizeof(resendMsg.nsSrch) + ntohs(resendMsg.nsSrch.nsNameLen)))
		{
		closeNameSocket();
		break;
		}
	    }
	}
}
#endif vxWorks or vms


/****************************************************************/
/*								*/
/*	SEE IF WE CAN FIND THE SAME POINT NAME ALREADY IN	*/
/*		THE NAME SERVER QUEUE.				*/
/*								*/
/****************************************************************/

NSQ *sameName(targetName)
char	*targetName;
{
    char	nameSought[PVNAME_SZ + FLDNAME_SZ + 2];
    char	nameFound[PVNAME_SZ + FLDNAME_SZ + 2];
    NSQ		*currNamQue;
    char	*dot;

    /*
    Copy the target name to a local string and wipe out everything
    after the point name.
    */

    strncpy(nameSought, targetName, PVNAME_SZ + FLDNAME_SZ + 1);
    if ((dot = (char *) strchr(nameSought, '.')))
	*dot = 0;

    /*
    Now loop through the name server queue and see if we can
    find the target point name in there, regardless of field.
    */

    currNamQue = nsqList;
    while (currNamQue)
	{
	strncpy(nameFound, currNamQue -> name, PVNAME_SZ + FLDNAME_SZ + 2);
	if ((dot = (char *) strchr(nameFound, '.')))
	    *dot = 0;
	if (!strcmp(nameSought, nameFound))
	    return (currNamQue);
	currNamQue = currNamQue -> pNextNSQ;
	}
    return (0);
}


/****************************************************************/
/*								*/
/*		GET SPLICING MASK FOR A CHANNEL			*/
/*								*/
/*	If the field name in a channel is .ACK, .SEVR, or	*/
/*	.STAT, return the appproriate mask and change the	*/
/*	name to .VAL. If the field name is none of the above,	*/
/*	just return a zero.					*/
/*								*/
/****************************************************************/

int getSpliceMask(name)
char	*name;
{
    char	*field;
    int		mask = 0;

    field = (char *) strchr(name, '.') + 1;

    if (!strcmp(field, "ACK"))
	mask = AGNT_SPL_ACK;
    else if (!strcmp(field, "SEVR"))
	mask = AGNT_SPL_SEVR;
    else if (!strcmp(field, "STAT"))
	mask = AGNT_SPL_STAT;
    else if (!strcmp(field, "VAL"))
	mask = AGNT_SPL_VAL;
    if (mask & AGNT_SPLICED)
	strcpy(field, "VAL");

    return (mask);
}


/****************************************************************/
/*								*/
/*	DETERMINE IF A CHANNEL IS OPEN ON THE .VAL FIELD	*/
/*								*/
/*	This routine is used to determine if the channel is	*/
/*	really splicable. We'll lie to the caller and only	*/
/*	tell him it's a .VAL field if all of the IOCs that	*/
/*	can possibly contain this point are splicable. Also,	*/
/*	a splicable channel must be open to the native data	*/
/*	type.							*/
/*								*/
/****************************************************************/

int isValField(pCHB)
CHB	*pCHB;
{
    char	*dot;

    if (!(pCHB -> pISB1 -> splicable))
	return (0);
    if (pCHB -> pISB2)
	if (!(pCHB -> pISB2 -> splicable))
	    return (0);
    if ((pCHB -> nativeType & 0xff) != (pCHB -> currType & 0xff))
	return (0);
    dot = (char *) strchr(pCHB -> chanName, '.');
    if (strcmp(dot, ".VAL"))
	return (0);
    else
	return (1);
}

